--
-- PostgreSQL database dump
--

SET statement_timeout = 0;
SET client_encoding = 'UTF8';
SET standard_conforming_strings = on;
SET check_function_bodies = false;
SET client_min_messages = warning;

--
-- Name: plpgsql; Type: EXTENSION; Schema: -; Owner: 
--

CREATE EXTENSION IF NOT EXISTS plpgsql WITH SCHEMA pg_catalog;


--
-- Name: EXTENSION plpgsql; Type: COMMENT; Schema: -; Owner: 
--

COMMENT ON EXTENSION plpgsql IS 'PL/pgSQL procedural language';


--
-- Name: flow; Type: EXTENSION; Schema: -; Owner: 
--

CREATE EXTENSION IF NOT EXISTS flow WITH SCHEMA public;


--
-- Name: EXTENSION flow; Type: COMMENT; Schema: -; Owner: 
--

COMMENT ON EXTENSION flow IS 'data type for cycle of orders';


SET search_path = public, pg_catalog;

--
-- Name: dquantity; Type: DOMAIN; Schema: public; Owner: olivier
--

CREATE DOMAIN dquantity AS bigint
	CONSTRAINT dquantity_check CHECK ((VALUE > 0));


ALTER DOMAIN public.dquantity OWNER TO olivier;

--
-- Name: ymarketaction; Type: TYPE; Schema: public; Owner: olivier
--

CREATE TYPE ymarketaction AS ENUM (
    'init',
    'open',
    'stop',
    'close',
    'start'
);


ALTER TYPE public.ymarketaction OWNER TO olivier;

--
-- Name: ymarketstatus; Type: TYPE; Schema: public; Owner: olivier
--

CREATE TYPE ymarketstatus AS ENUM (
    'INITIALIZING',
    'OPENED',
    'STOPPING',
    'CLOSED',
    'STARTING'
);


ALTER TYPE public.ymarketstatus OWNER TO olivier;

--
-- Name: yresprequote; Type: TYPE; Schema: public; Owner: olivier
--

CREATE TYPE yresprequote AS (
	own integer,
	nr integer,
	np integer,
	qtt_prov bigint,
	qtt_in_min bigint,
	qtt_out_min bigint,
	qtt_in_max bigint,
	qtt_out_max bigint,
	qtt_in_sum bigint,
	qtt_out_sum bigint,
	flows yflow[]
);


ALTER TYPE public.yresprequote OWNER TO olivier;

--
-- Name: fbalance(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fbalance() RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_cnt 		int;
BEGIN
	WITH accounting_order AS (SELECT np,sum(qtt) AS qtt FROM torder GROUP BY np),
	     accounting_mvt   AS (SELECT nat as np,sum(qtt) AS qtt FROM tmvt GROUP BY nat)
	SELECT count(*) INTO _cnt FROM tquality,accounting_order,accounting_mvt
	WHERE tquality.id=accounting_order.np AND tquality.id=accounting_mvt.np
		AND tquality.qtt != accounting_order.qtt + accounting_mvt.qtt;
	RETURN _cnt;
END;		
$$;


ALTER FUNCTION public.fbalance() OWNER TO olivier;

--
-- Name: fchangestatemarket(boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fchangestatemarket(_execute boolean) RETURNS ymarketstatus
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE
	_cnt int;
	_hm tmarket%rowtype;
	_action ymarketaction;
	_prev_status ymarketstatus;
	_res bool;
	_new_status ymarketstatus;
BEGIN

	SELECT market_status INTO _prev_status FROM vmarket;
	IF NOT FOUND THEN 
		_action := 'init';
		_prev_status := 'INITIALIZING';
		_new_status := 'OPENED';
		
	ELSIF (_prev_status = 'STARTING') THEN		
		_action := 'open';
		_new_status := 'OPENED';
		
	ELSIF (_prev_status = 'OPENED') THEN
		_action := 'stop';
		_new_status := 'STOPPING';
		
	ELSIF (_prev_status = 'STOPPING') THEN
		_action := 'close';
		_new_status := 'CLOSED';
		
	ELSE -- _prev_status='CLOSED'
		_action := 'start';
		_new_status := 'STARTING';
		RAISE INFO 'market_session = n->n+1';
	END IF;
	
	RAISE INFO 'market_status %->%',_prev_status,_new_status;
	IF NOT _execute THEN
		RAISE INFO 'The next action will be: %',_action;
		RAISE INFO 'market_status = % is unchanged.',_prev_status;
		RETURN _new_status;
	END IF;
	
	IF (_action = 'init' OR _action = 'open') THEN 		
		GRANT client_opened_role TO client;
		RAISE INFO 'The market is now opened for clients';		
		
	ELSIF (_action = 'stop') THEN
		REVOKE client_opened_role FROM client;
		GRANT  client_stopping_role TO client;	
		RAISE INFO 'The market is now stopping for clients';		
		
	ELSIF (_action = 'close') THEN
		REVOKE client_stopping_role FROM client;
		_res := frenumbertables(false);
		IF NOT _res THEN
			RAISE EXCEPTION USING ERRCODE='YA001';
		END IF;	
					
	ELSE -- _action='start'
		_res := frenumbertables(true);
		IF NOT _res THEN
			RAISE EXCEPTION USING ERRCODE='YA001';
		END IF;			

	END IF;
	
	INSERT INTO tmarket (created) VALUES (statement_timestamp()) RETURNING * INTO _hm;
	RETURN _new_status;
	 
END;
$$;


ALTER FUNCTION public.fchangestatemarket(_execute boolean) OWNER TO olivier;

--
-- Name: fcreate_tmp(integer, yorder, integer, integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fcreate_tmp(_id integer, _ord yorder, _np integer, _nr integer) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_MAXORDERFETCH	 int := fgetconst('MAXORDERFETCH'); 
	_MAXCYCLE 	int := fgetconst('MAXCYCLE');
	_cnt int;
BEGIN
/*	DROP TABLE IF EXISTS _tmp;
	RAISE INFO 'select * from fcreate_tmp(%,yorder_get%,%,%)',_id,_ord,_np,_nr;
	CREATE TEMPORARY TABLE _tmp ON COMMIT DROP  AS (
*/	
	CREATE TEMPORARY TABLE _tmp ON COMMIT DROP AS (
		SELECT A.id,A.ord,A.nr,A.pat FROM (
			WITH RECURSIVE search_backward(id,ord,pat,nr) AS (
				SELECT 	_id,_ord,yflow_get(_ord),_nr
				UNION ALL
				SELECT 	X.id,X.ord,
					yflow_get(X.ord,Y.pat), 
					-- add the order at the beginning of the yflow
					X.nr
					FROM search_backward Y,vorderinsert X
					WHERE   yflow_follow(_MAXCYCLE,X.ord,Y.pat) 
					-- X->Y === X.qtt>0 and X.np=Y[0].nr
					-- Y.pat does not contain X.ord 
					-- len(X.ord+Y.path) <= _MAXCYCLE	
					-- it is not an unexpected cycle: Y[!=-1]|->X === Y[i].np != X.nr with i!= -1
					 
			)
			SELECT id,ord,nr,pat 
			FROM search_backward LIMIT _MAXORDERFETCH 
		) A WHERE  yflow_status(A.pat)=3 --draft
	);
	SELECT COUNT(*) INTO _cnt FROM _tmp;

	RETURN _cnt;
END;
$$;


ALTER FUNCTION public.fcreate_tmp(_id integer, _ord yorder, _np integer, _nr integer) OWNER TO olivier;

--
-- Name: fcreateowner(text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fcreateowner(_name text) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_wid int;
BEGIN
	LOOP
		SELECT id INTO _wid FROM towner WHERE name=_name;
		IF found THEN
			RAISE WARNING 'The owner % was already created',_name;
			return _wid;
		END IF;
		BEGIN
			if(char_length(_name)<1) THEN
				RAISE NOTICE 'Owner s name cannot be empty';
				RAISE EXCEPTION USING ERRCODE='YU001';
			END IF;
			INSERT INTO towner (name) VALUES (_name) RETURNING id INTO _wid;
			RAISE INFO 'owner % created',_name;
			return _wid;
		EXCEPTION WHEN unique_violation THEN
			--
		END;
	END LOOP;
END;
$$;


ALTER FUNCTION public.fcreateowner(_name text) OWNER TO olivier;

--
-- Name: fcreateuser(text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fcreateuser(_name text) RETURNS void
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE
	_user	tuser%rowtype;
	_super	bool;
	_market_status	text;
BEGIN
	IF( _name IN ('admin','client','client_opened_role','client_stopping_role')) THEN
		RAISE WARNING 'The name % is not allowed',_name;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	SELECT * INTO _user FROM tuser WHERE name=_name;
	IF FOUND THEN
		RAISE WARNING 'The user % exists',_name;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	INSERT INTO tuser (name) VALUES (_name);
	
	SELECT rolsuper INTO _super FROM pg_authid where rolname=_name;
	IF NOT FOUND THEN
		EXECUTE 'CREATE ROLE ' || _name;
		EXECUTE 'ALTER ROLE ' || _name || ' NOINHERIT NOSUPERUSER NOCREATEDB NOCREATEROLE NOREPLICATION'; 
		EXECUTE 'ALTER ROLE ' || _name || ' LOGIN CONNECTION LIMIT 1';
	ELSE
		IF(_super) THEN
			-- RAISE INFO 'The role % is a super user.',_name;
			RAISE INFO 'The role is a super user.';
		ELSE
			-- RAISE WARNING 'The user is not found but a role % already exists - unchanged.',_name;
			RAISE WARNING 'The user is not found but the role already exists - unchanged.';
			RAISE EXCEPTION USING ERRCODE='YU001';				
		END IF;
	END IF;
	
	SELECT market_status INTO _market_status FROM vmarket;
	IF(_market_status = 'OPENED') THEN 
		-- RAISE INFO 'The market is opened for this user %', _name;
		RAISE INFO 'The market is opened for this user';
		EXECUTE 'GRANT client_opened_role TO ' || _name;
	ELSIF(_market_status = 'STOPPING') THEN
		-- RAISE INFO 'The market is stopping for this user %', _name;
		RAISE INFO 'The market is stopping for this user ';
		EXECUTE 'GRANT client_stopping_role TO ' || _name;	
	END IF;
	
	RETURN;
		
END;
$$;


ALTER FUNCTION public.fcreateuser(_name text) OWNER TO olivier;

SET default_tablespace = '';

SET default_with_oids = false;

--
-- Name: tquote; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tquote (
    id integer NOT NULL,
    own integer NOT NULL,
    nr integer NOT NULL,
    qtt_requ bigint,
    np integer NOT NULL,
    qtt_prov bigint,
    qtt_in bigint,
    qtt_out bigint,
    flows yflow[],
    created timestamp without time zone NOT NULL,
    removed timestamp without time zone
);


ALTER TABLE public.tquote OWNER TO olivier;

--
-- Name: fexecquote(text, integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fexecquote(_owner text, _idquote integer) RETURNS tquote
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
	
DECLARE
	_wid		int;
	_o		torder%rowtype;
	_idd		int;
	_expected	tquote%rowtype;
	_q		tquote%rowtype;
	_pivot		torder%rowtype;

	_flows		yflow[];
	_ypatmax	yflow;
	_res	        int8[];
	_qtt_prov	int8;
	_qtt_requ	int8;
	_first_mvt	int;
BEGIN
	
	_idd := fverifyquota();
	_wid := fgetowner(_owner,false); -- returns _wid == 0 if not found
	
	SELECT * INTO _q FROM tquote WHERE id=_idquote AND own=_wid;
	IF (NOT FOUND) THEN
		IF(_wid = 0) THEN
			RAISE NOTICE 'the owner % is not found',_owner;
		ELSE
			RAISE NOTICE 'this quote % was not made by owner %',_idquote,_owner;
		END IF;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
		
	-- _q.qtt_requ != 0		
	_qtt_requ := _q.qtt_requ;
	_qtt_prov := _q.qtt_prov;
	
	_o := finsert_toint(_qtt_prov,_q.nr,_q.np,_qtt_requ,_q.own);
	
	_q.id      := _o.id;
	_q.qtt_in  := 0;
	_q.qtt_out := 0;
	_q.flows   := ARRAY[]::yflow[];
	
	FOR _ypatmax IN SELECT _patmax  FROM finsertflows(_o) LOOP
		_first_mvt := fexecute_flow(_ypatmax);
		_res := yflow_qtts(_ypatmax);
		_q.qtt_in  := _q.qtt_in  + _res[1];
		_q.qtt_out := _q.qtt_out + _res[2];
		_q.flows := array_append(_q.flows,_ypatmax);
	END LOOP;
	
	
	IF (	(_q.qtt_in = 0) OR (_qtt_requ = 0) OR
		((_q.qtt_out::double precision)	/(_q.qtt_in::double precision)) > 
		((_qtt_prov::double precision)	/(_qtt_requ::double precision))
	) THEN
		RAISE NOTICE 'Omega of the flows obtained is not limited by the order';
		RAISE EXCEPTION USING ERRCODE='YA003';
	END IF;
	
	PERFORM fremovequote_int(_idquote);	
	PERFORM finvalidate_treltried();
	
	RETURN _q;
	
EXCEPTION WHEN SQLSTATE 'YU001' THEN
	PERFORM fremovequote_int(_idquote); 
	RAISE INFO 'Abort; Quote removed';
	RETURN _q; 

END; 
$$;


ALTER FUNCTION public.fexecquote(_owner text, _idquote integer) OWNER TO olivier;

--
-- Name: fexecute_flow(yflow); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fexecute_flow(_flw yflow) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_commits	int8[][];
	_i		int;
	_next_i		int;
	_nbcommit	int;
	
	_oid		int;
	_w_src		int;
	_w_dst		int;
	_flowr		int8;
	_first_mvt	int;
	_exhausted	bool;
	_mvt_id		int;
	_qtt		int8;
	_cnt 		int;
	_uuid		text;
	_res		text;
BEGIN

	_commits := yflow_to_matrix(_flw);
	-- indices in _commits
	-- 1  2   3  4        5  6        7   8
	-- id,own,nr,qtt_requ,np,qtt_prov,qtt,flowr
	
	_nbcommit := yflow_dim(_flw); -- raise an error when flow->dim not in [2,8]
	_first_mvt := NULL;
	_exhausted := false;
	-- RAISE INFO 'flow of % commits',_nbcommit;
	_i := _nbcommit;	
	FOR _next_i IN 1 .. _nbcommit LOOP
		-- _commits[_next_i] follows _commits[_i]
		_oid	:= _commits[_i][1]::int;
		_w_src	:= _commits[_i][2]::int;
		_w_dst	:= _commits[_next_i][2]::int;
		_flowr	:= _commits[_i][8];
		
		UPDATE torder set qtt = qtt - _flowr ,updated = statement_timestamp()
			WHERE id = _oid AND _flowr <= qtt RETURNING uuid,qtt INTO _uuid,_qtt;
		IF(NOT FOUND) THEN
			RAISE WARNING 'the flow is not in sync with the database torder[%] does not exist or torder.qtt < %',_oid,_flowr ;
			RAISE EXCEPTION USING ERRCODE='YU002';
		END IF;
			
		INSERT INTO tmvt (nb,oruuid,grp,own_src,own_dst,qtt,nat,created) 
			VALUES(_nbcommit,_uuid,_first_mvt,_w_src,_w_dst,_flowr,_commits[_i][5]::int,statement_timestamp())
			RETURNING id INTO _mvt_id;
					
		IF(_first_mvt IS NULL) THEN
			_first_mvt := _mvt_id;
		END IF;
		
		IF(_qtt=0) THEN
			perform fremoveorder_int(_oid);
			_exhausted := true;
		END IF;

		_i := _next_i;
		----------------------------------------------------------------
	END LOOP;
	-- RAISE INFO '_first_mvt=%',_first_mvt;
	UPDATE tmvt SET grp = _first_mvt WHERE id = _first_mvt  AND (grp IS NULL);	
	IF(NOT FOUND) THEN
		RAISE EXCEPTION 'the movement % does not exist',_first_mvt 
			USING ERRCODE='YA003';
	END IF;
	
	IF(NOT _exhausted) THEN
		--  some order should be exhausted 
		RAISE EXCEPTION 'the cycle should exhaust some order' 
			USING ERRCODE='YA003';
	END IF;
	
	PERFORM fupdate_treltried(_commits,_nbcommit);
	
	RETURN _first_mvt;
END;
$$;


ALTER FUNCTION public.fexecute_flow(_flw yflow) OWNER TO olivier;

--
-- Name: fexplodequality(text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fexplodequality(_quality_name text) RETURNS text[]
    LANGUAGE plpgsql
    AS $$
DECLARE
	_e int;
	_q text[];
	_CHECK_QUALITY_OWNERSHIP int := fgetconst('CHECK_QUALITY_OWNERSHIP');
BEGIN
	IF(char_length(_quality_name) <1) THEN
		RAISE NOTICE 'Quality name "%" incorrect: do not len(name)<1',_quality_name;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	IF(_CHECK_QUALITY_OWNERSHIP = 0) THEN
		_q[1] := NULL;
		_q[2] := _quality_name;
		RETURN _q;
	END IF;
	
	_e =position('/' in _quality_name);
	IF(_e < 2) THEN 
		RAISE NOTICE 'Quality name "%" incorrect: <depository>/<quality> expected',_quality_name;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	_q[1] = substring(_quality_name for _e-1);
	_q[2] = substring(_quality_name from _e+1);
	if(char_length(_q[2])<1) THEN
		RAISE NOTICE 'Quality name "%" incorrect: <depository>/<quality> expected',_quality_name;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	RETURN _q;
END;
$$;


ALTER FUNCTION public.fexplodequality(_quality_name text) OWNER TO olivier;

--
-- Name: fget_treltried(integer, integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fget_treltried(_np integer, _nr integer) RETURNS bigint
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_cnt int8;
	_MAXTRY 	int := fgetconst('MAXTRY');
BEGIN
	IF(_MAXTRY=0) THEN
		RETURN 0;
	END IF;
	SELECT cnt into _cnt FROM treltried WHERE np=_np AND nr=_nr;
	IF NOT FOUND THEN
		_cnt := 0;
	END IF;

	RETURN _cnt;
END;
$$;


ALTER FUNCTION public.fget_treltried(_np integer, _nr integer) OWNER TO olivier;

--
-- Name: fgetagr(integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetagr(_grp integer) RETURNS TABLE(_own text, _natp text, _qtt_prov bigint, _qtt_requ bigint, _natr text)
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_fnat	 text;
	_fqtt	 int8;
	_fown	 text;
	_m	 vmvtverif%rowtype;
BEGIN
		_qtt_requ := NULL;
		FOR _m IN SELECT * FROM vmvtverif WHERE grp=_grp ORDER BY id ASC LOOP
			IF(_qtt_requ IS NULL) THEN
				_qtt_requ := _m.qtt;
				SELECT name INTO _natr FROM tquality WHERE _m.nat=id;
				SELECT name INTO _fown FROM towner WHERE _m.own_src=id;
				_fqtt := _m.qtt;
				_fnat := _natr;
			ELSE
				SELECT name INTO _natp FROM tquality WHERE _m.nat=id;
				SELECT name INTO _own FROM towner WHERE _m.own_src=id;
				_qtt_prov := _m.qtt;
				
				RETURN NEXT;
				_qtt_requ := _qtt_prov;
				_natr := _natp;
			END IF;
		END LOOP;
		IF(_qtt_requ IS NOT NULL) THEN
			_own := _fown;
			_natp := _fnat;
			_qtt_prov := _fqtt;
			--_qtt_requ := _qtt_requ;
			--_natr :=  _natr;
			RETURN NEXT;
		END IF;
	RETURN;
END;
$$;


ALTER FUNCTION public.fgetagr(_grp integer) OWNER TO olivier;

--
-- Name: fgetconst(text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetconst(_name text) RETURNS integer
    LANGUAGE plpgsql STABLE
    AS $$
DECLARE
	_ret int;
BEGIN
	SELECT value INTO _ret FROM tconst WHERE name=_name;
	IF(NOT FOUND) THEN
		RAISE 'the const % should be found',_name;
		RAISE EXCEPTION USING ERRCODE='YA002';
	END IF;
	IF(_name = 'MAXCYCLE' AND _ret >8) THEN
		RAISE EXCEPTION 'obCMAXVALUE must be <=8' USING ERRCODE='YA002';
	END IF;
	RETURN _ret;
END; 
$$;


ALTER FUNCTION public.fgetconst(_name text) OWNER TO olivier;

--
-- Name: fgeterrs(boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgeterrs(_details boolean) RETURNS TABLE(_name text, cnt bigint)
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE 
	_i 		int;
	_cnt 		int;
BEGIN		
	_name := 'balance';
	cnt := fbalance();	
	RETURN NEXT;
	
	IF(_details) THEN
	
		_name := 'errors on quantities in mvts';
		cnt := fverifmvt();
		RETURN NEXT;
	
		_name := 'errors on agreements in mvts';
		cnt := fverifmvt2();
		RETURN NEXT;
	END IF;
	RETURN;
END;
$$;


ALTER FUNCTION public.fgeterrs(_details boolean) OWNER TO olivier;

--
-- Name: fgetowner(text, boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetowner(_name text, _insert boolean) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_wid int;
BEGIN

	SELECT id INTO _wid FROM towner WHERE name=_name;
	IF NOT found THEN
		IF (_insert) THEN
			IF (fgetconst('INSERT_OWN_UNKNOWN')=1) THEN
				_wid := fcreateowner(_name);
			ELSE
				RAISE NOTICE 'The owner % is unknown',_name;
				RAISE EXCEPTION USING ERRCODE='YU001';
			END IF;
		ELSE
			_wid := 0;
		END IF;
	END IF;
	return _wid;
END;
$$;


ALTER FUNCTION public.fgetowner(_name text, _insert boolean) OWNER TO olivier;

--
-- Name: fgetprequote(text, text, bigint, text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) RETURNS yresprequote
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
	
DECLARE
	_pivot 		 torder%rowtype;
	_ypatmax	 yflow;
	_flows		 yflow[];
	_res	         int8[];
	_idd		 int;
	_q		 text[];
	_r		 yresprequote;
	_om_min		double precision;
	_om_max		double precision;
	_om		double precision;
BEGIN
	_idd := fverifyquota();
	
	-- quantity must be >0
	IF(_qttprovided<=0) THEN
		RAISE NOTICE 'quantities incorrect: %<=0', _qttprovided;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	_q := fexplodequality(_qualityprovided);
	IF ((_q[1] IS NOT NULL) AND (_q[1] != session_user)) THEN
		RAISE NOTICE 'depository % of quality is not the user %',_q[1],session_user;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	-- qualities are red and inserted if necessary
	_pivot.np := fgetquality(_qualityprovided,true); 
	_pivot.nr := fgetquality(_qualityrequired,true); 
	_pivot.id  := 0; 
	
	-- if does not exists, inserted
	_pivot.own := fgetowner(_owner,true); 
	 
	_pivot.qtt_requ := 0; -- lastignore == true
	_pivot.qtt_prov := _qttprovided; 
	_pivot.qtt := _qttprovided;
	
	_r.own 		:= _pivot.own;
	_r.flows 	:= ARRAY[]::yflow[];
	_r.nr 		:= _pivot.nr;
	_r.np 		:= _pivot.np;
	_r.qtt_prov 	:= _pivot.qtt_prov;
	
	_r.qtt_in_min := 0;	_r.qtt_in_max := 0; 
	_r.qtt_out_min := 0;	_r.qtt_out_max := 0;
	_om_min := 0;		_om_max := 0;
	
	_r.qtt_in_sum := 0;
	_r.qtt_out_sum := 0;
	
	FOR _ypatmax IN SELECT _patmax  FROM finsertflows(_pivot) LOOP
		_r.flows := array_append(_r.flows,_ypatmax);
		_res := yflow_qtts(_ypatmax); -- [in,out] of the last node
		
		_r.qtt_in_sum  := _r.qtt_in_sum + _res[1];
		_r.qtt_out_sum := _r.qtt_out_sum + _res[2];
		
		_om := (_res[2]::double precision)/(_res[1]::double precision);
		
		IF(_om_min = 0 OR _om < _om_min) THEN
			_r.qtt_in_min := _res[1];
			_r.qtt_out_min := _res[2];
			_om_min := _om;
		END IF;
		IF(_om_max = 0 OR _om > _om_max) THEN
			_r.qtt_in_max := _res[1];
			_r.qtt_out_max := _res[2];
			_om_max := _om;
		END IF;
	END LOOP;

	RETURN _r;
END; 
$$;


ALTER FUNCTION public.fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) OWNER TO olivier;

--
-- Name: fgetquality(text, boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetquality(_quality_name text, insert boolean) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_idd int;
	_qlt	tquality%rowtype;
	_q text[];
	_id int;
BEGIN
	LOOP
		SELECT * INTO _qlt FROM tquality WHERE name = _quality_name;
		IF FOUND THEN
			return _qlt.id;
		END IF;
		IF(NOT insert) THEN
			return 0;
		END IF;
		
		BEGIN
			_q := fexplodequality(_quality_name);
			IF(_q[1] IS NOT NULL) THEN 	
				-- _CHECK_QUALITY_OWNERSHIP =1
				SELECT id INTO _idd FROM tuser WHERE name=_q[1];
				IF(NOT FOUND) THEN -- user should exists
					RAISE NOTICE 'The depository "%" is undefined',_q[1] ;
					RAISE EXCEPTION USING ERRCODE='YU001';
				END IF;
			ELSE
				_idd := NULL;
			END IF;
		
			INSERT INTO tquality (name,idd,depository,qtt) VALUES (_quality_name,_idd,_q[1],0)
				RETURNING * INTO _qlt;
			RETURN _qlt.id;
			
		EXCEPTION WHEN unique_violation THEN
			--
		END;
	END LOOP;

END;
$$;


ALTER FUNCTION public.fgetquality(_quality_name text, insert boolean) OWNER TO olivier;

--
-- Name: fgetquote(text, text, bigint, bigint, text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) RETURNS tquote
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
	
DECLARE
	_pivot 		 torder%rowtype;
	_ypatmax	 yflow;
	_flows		 yflow[];
	_res	         int8[];
	_cumul		 int8[];
	_qtt_prov	 int8;
	_qtt_requ	 int8;
	_idd		 int;
	_q		 text[];
	_ret		 tquote%rowtype;
	_r		 tquote%rowtype;
BEGIN
	_idd := fverifyquota();
	
	-- quantities must be >0
	IF(_qttprovided<=0 OR _qttrequired<0) THEN
		RAISE NOTICE 'quantities incorrect: %<=0 or %<0', _qttprovided,_qttrequired;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	_q := fexplodequality(_qualityprovided);
	IF ((_q[1] IS NOT NULL) AND (_q[1] != session_user)) THEN
		RAISE NOTICE 'depository % of quality is not the user %',_q[1],session_user;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	-- qualities are red and inserted if necessary
	_pivot.np := fgetquality(_qualityprovided,true); 
	_pivot.nr := fgetquality(_qualityrequired,true); 
	_pivot.id  := 0; 
	
	-- if does not exists, inserted
	_pivot.own := fgetowner(_owner,true); 
	 
	_pivot.qtt_requ := _qttrequired; -- if _qttrequired==0 then lastignore == true
	_pivot.qtt_prov := _qttprovided; 
	_pivot.qtt := _qttprovided;
	
	_r.id 		:= 0;
	_r.own 		:= _pivot.own;
	_r.flows 	:= ARRAY[]::yflow[];
	_r.nr 		:= _pivot.nr;
	_r.qtt_requ 	:= _pivot.qtt_requ;
	_r.np 		:= _pivot.np;
	_r.qtt_prov 	:= _pivot.qtt_prov;
	
	_cumul[1] := 0; -- in
	_cumul[2] := 0; -- out
	FOR _ypatmax IN SELECT _patmax  FROM finsertflows(_pivot) LOOP
		_r.flows := array_append(_r.flows,_ypatmax);
		_res := yflow_qtts(_ypatmax); -- [in,out] of the last node
		IF(_qttrequired = 0) THEN
			_cumul := yorder_moyen(_cumul[1],_cumul[2],_res[1],_res[2]);
		ELSE
			_cumul[1] := _cumul[1]+_res[1];
			_cumul[2] := _cumul[2]+_res[2];
		END IF;
	END LOOP;
	_r.qtt_in  := _cumul[1];
	_r.qtt_out := _cumul[2];
	
	IF (_qttrequired != 0 AND _r.qtt_out != 0 AND _r.qtt_in != 0) THEN
		INSERT INTO tquote (own,nr,qtt_requ,np,qtt_prov,qtt_in,qtt_out,flows,created,removed) 
			VALUES (_pivot.own,_r.nr,_r.qtt_requ,_r.np,_r.qtt_prov,_r.qtt_in,_r.qtt_out,_r.flows,statement_timestamp(),NULL)
		RETURNING * INTO _ret;
		RETURN _ret;
	ELSE
		RETURN _r;
	END IF;
END; 
$$;


ALTER FUNCTION public.fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) OWNER TO olivier;

--
-- Name: fgetstats(boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetstats(_details boolean) RETURNS TABLE(_name text, cnt bigint)
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE 
	_i 		int;
	_cnt 		int;
BEGIN

	_name := 'number of qualities';
	select count(*) INTO cnt FROM tquality;
	RETURN NEXT;
	
	_name := 'number of owners';
	select count(*) INTO cnt FROM towner;
	RETURN NEXT;
	
	_name := 'number of quotes';
	select count(*) INTO cnt FROM tquote;
	RETURN NEXT;
			
	_name := 'number of orders';
	select count(*) INTO cnt FROM vorderverif;
	RETURN NEXT;
	
	_name := 'number of movements';
	select count(*) INTO cnt FROM vmvtverif;
	RETURN NEXT;
	
	_name := 'number of quotes removed';
	select count(*) INTO cnt FROM tquoteremoved;
	RETURN NEXT;

	_name := 'number of orders removed';
	select count(*) INTO cnt FROM torderremoved;
	RETURN NEXT;
	
	_name := 'number of movements removed';
	select count(*) INTO cnt FROM tmvtremoved;	
	RETURN NEXT;
	
	_name := 'number of agreements';
	select count(distinct grp) INTO cnt FROM vmvtverif where nb!=1;	
	RETURN NEXT;	
	
	_name := 'number of orders rejected';
	select count(distinct grp) INTO cnt FROM vmvtverif where nb=1;	
	RETURN NEXT;

/*	
	FOR _name,cnt IN SELECT name,value FROM tconst LOOP
		RETURN NEXT;
	END LOOP;
*/		
	
	FOR _i,cnt IN select nb,count(distinct grp) FROM vmvtverif where nb!=1 GROUP BY nb LOOP
		_name := 'agreements with ' || _i || ' partners';
		RETURN NEXT;
	END LOOP;

	RETURN;
END;
$$;


ALTER FUNCTION public.fgetstats(_details boolean) OWNER TO olivier;

--
-- Name: fgetstats(timestamp without time zone, timestamp without time zone); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) RETURNS TABLE(_name text, cnt bigint)
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE 
	_i 		int;
	_cnt 		int;
BEGIN
/*
	_name := 'number of qualities';
	select count(*) INTO cnt FROM tquality where updated>=mini and updated<maxi;
	RETURN NEXT;
	
	_name := 'number of owners';
	select count(*) INTO cnt FROM towner where created>=mini and created<maxi;
	RETURN NEXT;
	
	_name := 'number of quotes';
	select count(*) INTO cnt FROM tquote where created>=mini and created<maxi;
	RETURN NEXT;
*/			
	_name := 'number of orders';
	select count(*) INTO cnt FROM vorderverif where  created<maxi;
	RETURN NEXT;
	
	_name := 'number of movements';
	select count(*) INTO cnt FROM vmvtverif where  created<maxi;
	RETURN NEXT;
/*	
	_name := 'number of quotes removed';
	select count(*) INTO cnt FROM tquoteremoved where removed>=mini and removed<maxi;
	RETURN NEXT;

	_name := 'number of orders removed';
	select count(*) INTO cnt FROM torderremoved where updated>=mini and updated<maxi;
	RETURN NEXT;
	
	_name := 'number of movements removed';
	select count(*) INTO cnt FROM tmvtremoved where deleted>=mini and deleted<maxi;	
	RETURN NEXT;
	
	_name := 'number of agreements';
	select count(distinct grp) INTO cnt FROM vmvtverif where nb!=1 and created>=mini and created<maxi;	
	RETURN NEXT;	
*/	
	_name := 'number of orders rejected';
	select count(*) INTO cnt FROM tmvt where nb=1 and created>=mini and created<maxi;	
	RETURN NEXT;	
	
	FOR _i,cnt IN select nb,count(distinct grp) FROM vmvtverif where nb!=1  and created>=mini and created<maxi GROUP BY nb LOOP
		_name := 'agr. with ' || _i || ' partners';
		RETURN NEXT;
	END LOOP;

	RETURN;
END;
$$;


ALTER FUNCTION public.fgetstats(mini timestamp without time zone, maxi timestamp without time zone) OWNER TO olivier;

--
-- Name: fgetuuid(integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fgetuuid(_id integer) RETURNS text
    LANGUAGE plpgsql
    AS $$ 
DECLARE
	_market_session	int;
BEGIN
	SELECT market_session INTO _market_session FROM vmarket;
	-- RETURN lpad(_market_session::text,19,'0') || '-' || lpad(_id::text,19,'0');
	RETURN _market_session::text || '-' || _id::text;
END;
$$;


ALTER FUNCTION public.fgetuuid(_id integer) OWNER TO olivier;

--
-- Name: torder; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE torder (
    id integer NOT NULL,
    uuid text NOT NULL,
    own integer NOT NULL,
    nr integer NOT NULL,
    qtt_requ bigint NOT NULL,
    np integer NOT NULL,
    qtt_prov dquantity NOT NULL,
    qtt bigint NOT NULL,
    start bigint,
    created timestamp without time zone NOT NULL,
    updated timestamp without time zone,
    CONSTRAINT torder_check CHECK ((((qtt >= 0) AND ((qtt_prov)::bigint >= qtt)) AND (qtt_requ >= 0)))
);


ALTER TABLE public.torder OWNER TO olivier;

--
-- Name: TABLE torder; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON TABLE torder IS 'description of orders';


--
-- Name: COLUMN torder.id; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.id IS 'unique id for the session of the market';


--
-- Name: COLUMN torder.uuid; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.uuid IS 'unique id for all sessions';


--
-- Name: COLUMN torder.own; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.own IS 'owner of the value provided';


--
-- Name: COLUMN torder.nr; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.nr IS 'quality required';


--
-- Name: COLUMN torder.qtt_requ; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.qtt_requ IS 'quantity required; used to express omega=qtt_prov/qtt_req';


--
-- Name: COLUMN torder.np; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.np IS 'quality provided';


--
-- Name: COLUMN torder.qtt_prov; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.qtt_prov IS 'quantity offered';


--
-- Name: COLUMN torder.qtt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.qtt IS 'current quantity remaining (<= quantity offered)';


--
-- Name: COLUMN torder.start; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN torder.start IS 'position of treltried[np,nr].cnt when the order is inserted';


--
-- Name: finsert_toint(bigint, integer, integer, bigint, bigint); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION finsert_toint(_qtt_prov bigint, _nr integer, _np integer, _qtt_requ bigint, _own bigint) RETURNS torder
    LANGUAGE plpgsql
    AS $$
DECLARE
	_o		 torder%rowtype;
	_id		 int;
	_uuid		 text;
	_start		 int8;
BEGIN	
	perform fupdate_quality(_np,_qtt_prov);	
		
	INSERT INTO torder (uuid,qtt,nr,np,qtt_prov,qtt_requ,own,created,updated) 
		VALUES ('',_qtt_prov,_nr,_np,_qtt_prov,_qtt_requ,_own,statement_timestamp(),NULL)
		RETURNING id INTO _id;
	
	_uuid := fgetuuid(_id);
	_start := fget_treltried(_np,_nr);
	
	UPDATE torder SET uuid = _uuid,start = _start WHERE id=_id RETURNING * INTO _o;	
	
	RETURN _o;					
END;
$$;


ALTER FUNCTION public.finsert_toint(_qtt_prov bigint, _nr integer, _np integer, _qtt_requ bigint, _own bigint) OWNER TO olivier;

--
-- Name: finsertflows(torder); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION finsertflows(_pivot torder) RETURNS TABLE(_patmax yflow)
    LANGUAGE plpgsql
    AS $$
DECLARE

	_idpivot 	int;
	_cnt 		int;
	_o		torder%rowtype;
	_res	        int8[];
	_start		int8;
	_time_begin	timestamp;
BEGIN
	------------------------------------------------------------------------
	-- _pivot.qtt := _pivot.qtt_prov;
	_time_begin := clock_timestamp();
	
	_cnt := fcreate_tmp(_pivot.id,
			yorder_get(_pivot.id,_pivot.own,_pivot.nr,_pivot.qtt_requ,_pivot.np,_pivot.qtt_prov,_pivot.qtt),
			_pivot.np,_pivot.nr);

	IF(_cnt=0) THEN
		RETURN;
	END IF;
	_cnt := 0;
	
	LOOP		
		SELECT yflow_max(pat) INTO _patmax FROM _tmp  ;
		
		IF (yflow_status(_patmax)!=3) THEN -- status != draft
			EXIT; -- from LOOP
		END IF;
		_cnt := _cnt + 1;
		
		RETURN NEXT;

		UPDATE _tmp SET pat = yflow_reduce(pat,_patmax);
	END LOOP;
	
	DROP TABLE _tmp;
	
	perform fspendquota(_time_begin);
	
 	RETURN;
END; 
$$;


ALTER FUNCTION public.finsertflows(_pivot torder) OWNER TO olivier;

--
-- Name: finsertorder(text, text, bigint, bigint, text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) RETURNS tquote
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
	
DECLARE
	_wid		int;
	_o		torder%rowtype;
	_idd		int;
	_expected	tquote%rowtype;
	_q		tquote%rowtype;
	_pivot		torder%rowtype;
	_qua		text[];

	_flows		yflow[];
	_ypatmax	yflow;
	_res	        int8[];
	_first_mvt	int;
BEGIN
	
	_idd := fverifyquota();
	_q.own := fgetowner(_owner,true); -- inserted if not found
	
	-- quantities must be >0
	IF(_qttprovided<=0 OR _qttrequired<=0) THEN
		RAISE NOTICE 'quantities incorrect: %<=0 or %<=0', _qttprovided,_qttrequired;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	_qua := fexplodequality(_qualityprovided);
	IF ((_qua[1] IS NOT NULL) AND (_qua[1] != session_user)) THEN
		RAISE NOTICE 'depository % of quality is not the user %',_qua[1],session_user;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	-- qualities are red and inserted if necessary
	_q.np := fgetquality(_qualityprovided,true); 
	_q.nr := fgetquality(_qualityrequired,true); 
	-- _q.qtt_requ != 0
		
	_q.qtt_requ := _qttrequired;
	_q.qtt_prov := _qttprovided;
	
	_o := finsert_toint(_qttprovided,_q.nr,_q.np,_qttrequired,_q.own);
	
	_q.id      := _o.id;
	_q.qtt_in  := 0;
	_q.qtt_out := 0;
	_q.flows   := ARRAY[]::yflow[];
	
	FOR _ypatmax IN SELECT _patmax  FROM finsertflows(_o) LOOP
		_first_mvt := fexecute_flow(_ypatmax);
		_res := yflow_qtts(_ypatmax);
		_q.qtt_in  := _q.qtt_in  + _res[1];
		_q.qtt_out := _q.qtt_out + _res[2];
		_q.flows := array_append(_q.flows,_ypatmax);
	END LOOP;
	
	
	IF (	(_q.qtt_in != 0) AND 
		((_q.qtt_out::double precision)		/(_q.qtt_in::double precision)) > 
		((_qttprovided::double precision)	/(_qttrequired::double precision))
	) THEN
		RAISE NOTICE 'Omega of the flows obtained is not limited by the order';
		RAISE EXCEPTION USING ERRCODE='YA003';
	END IF;
		
	PERFORM finvalidate_treltried();
	
	RETURN _q;
	
EXCEPTION WHEN SQLSTATE 'YU001' THEN 
	RAISE INFO 'Abort';
	RETURN _q; 

END; 
$$;


ALTER FUNCTION public.finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) OWNER TO olivier;

--
-- Name: finvalidate_treltried(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION finvalidate_treltried() RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_o 	torder%rowtype;
	_MAXTRY int := fgetconst('MAXTRY');
	_res	int;
	_mvt_id	int;
BEGIN
	IF(_MAXTRY=0) THEN
		RETURN;
	END IF;
	
	FOR _o IN SELECT o.* FROM torder o,treltried r 
		WHERE o.np=r.np AND o.nr=r.nr AND o.start IS NOT NULL AND o.start + _MAXTRY < r.cnt LOOP
		
		INSERT INTO tmvt (nb,oruuid,grp,own_src,own_dst,qtt,nat,created) 
			VALUES(1,_o.uuid,NULL,_o.own,_o.own,_o.qtt,_o.np,statement_timestamp()) 
			RETURNING id INTO _mvt_id;
			
		-- the order order.qtt != 0
		perform fremoveorder_int(_o.id);			
	END LOOP;
	RETURN;
END;
$$;


ALTER FUNCTION public.finvalidate_treltried() OWNER TO olivier;

--
-- Name: fremoveagreement(integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fremoveagreement(_grp integer) RETURNS integer
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE 
	_nat int;
	_scntq	int;
	_cntm	int;
	_scntm	int;
	_qtt int8;
	_qlt tquality%rowtype;
	_CHECK_QUALITY_OWNERSHIP int := fgetconst('CHECK_QUALITY_OWNERSHIP');
BEGIN
	_scntq := 0;_scntm := 0;
	FOR _nat,_qtt,_cntm IN SELECT m.nat,sum(m.qtt),count(m.id) FROM tmvt m, tquality q 
		WHERE m.nat=q.id AND 
		((q.depository=session_user) OR (_CHECK_QUALITY_OWNERSHIP = 0)) AND 
		m.grp=_grp GROUP BY m.nat LOOP
		
		_scntq := _scntq +1;
		_scntm := _scntm + _cntm;
		UPDATE tquality SET qtt = qtt - _qtt WHERE id = _nat RETURNING qtt INTO _qlt;
		-- constraint  tquality.qtt >=0	
	END LOOP;
	
	IF (_scntm=0) THEN
		RAISE WARNING 'The agreement "%" does not exist or no movement of this agreement belongs to the user %',_grp,session_user;
	ELSE
		WITH a AS (DELETE FROM tmvt m USING tquality q WHERE m.nat=q.id AND 
			((q.depository=session_user) OR (_CHECK_QUALITY_OWNERSHIP = 0)) 
			AND m.grp=_grp RETURNING m.*) 
		INSERT INTO tmvtremoved SELECT id,nb,oruuid,grp,own_src,own_dst,qtt,nat,created,statement_timestamp() as deleted FROM a;

	END IF;
	
	RETURN _scntm;

END;
$$;


ALTER FUNCTION public.fremoveagreement(_grp integer) OWNER TO olivier;

--
-- Name: towner; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE towner (
    id integer NOT NULL,
    name text NOT NULL,
    created timestamp without time zone,
    updated timestamp without time zone,
    CONSTRAINT towner_name_check CHECK ((char_length(name) > 0))
);


ALTER TABLE public.towner OWNER TO olivier;

--
-- Name: TABLE towner; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON TABLE towner IS 'owners of values exchanged';


--
-- Name: tquality; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tquality (
    id integer NOT NULL,
    name text NOT NULL,
    idd integer,
    depository text,
    qtt bigint DEFAULT 0,
    created timestamp without time zone,
    updated timestamp without time zone,
    CONSTRAINT tquality_check CHECK ((((char_length(name) > 0) AND (char_length(depository) > 0)) AND (qtt >= 0)))
);


ALTER TABLE public.tquality OWNER TO olivier;

--
-- Name: TABLE tquality; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON TABLE tquality IS 'description of qualities';


--
-- Name: COLUMN tquality.name; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tquality.name IS 'name of depository/name of quality ';


--
-- Name: COLUMN tquality.depository; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tquality.depository IS 'name of depository (user)';


--
-- Name: COLUMN tquality.qtt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tquality.qtt IS 'total quantity on the market for this quality';


--
-- Name: vorder; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vorder AS
    SELECT n.id, n.uuid, w.name AS owner, qr.name AS qua_requ, n.qtt_requ, qp.name AS qua_prov, n.qtt_prov, n.qtt, n.start, n.created, n.updated FROM (((torder n JOIN tquality qr ON ((n.nr = qr.id))) JOIN tquality qp ON ((n.np = qp.id))) JOIN towner w ON ((n.own = w.id)));


ALTER TABLE public.vorder OWNER TO olivier;

--
-- Name: fremoveorder(text); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fremoveorder(_uuid text) RETURNS vorder
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE
	_qtt		int8;
	_o 		torder%rowtype;
	_vo		vorder%rowtype;
	_qlt		tquality%rowtype;
	_CHECK_QUALITY_OWNERSHIP int := fgetconst('CHECK_QUALITY_OWNERSHIP');
BEGIN
	_vo.id = NULL;
	IF(_CHECK_QUALITY_OWNERSHIP != 0) THEN
		SELECT o.* INTO _o FROM torder o,tquality q 
			WHERE 	o.np=q.id AND q.depository=session_user AND o.uuid = _uuid;
		IF NOT FOUND THEN
			RAISE WARNING 'the order % on a quality belonging to % was not found',_uuid,session_user;
			RAISE EXCEPTION USING ERRCODE='YU001';
		END IF;
	ELSE
		SELECT o.* INTO _o FROM torder o WHERE o.uuid = _uuid;
		IF NOT FOUND THEN
			RAISE WARNING 'the order % was not found',_uuid;
			RAISE EXCEPTION USING ERRCODE='YU001';
		END IF;
	END IF;

	SELECT * INTO _vo FROM vorder WHERE id = _o.id;	-- _vo returned
	
		
	UPDATE tquality SET qtt = qtt - _o.qtt WHERE id = _o.np RETURNING qtt INTO _qlt;
	-- check tquality.qtt >=0	
	-- order is removed but is NOT cleared
	perform fremoveorder_int(_o.id);
	
	RETURN _vo;
	
EXCEPTION WHEN SQLSTATE 'YU001' THEN
	RAISE INFO 'ABORTED';
	RETURN _vo; 
END;		
$$;


ALTER FUNCTION public.fremoveorder(_uuid text) OWNER TO olivier;

--
-- Name: fremoveorder_int(integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fremoveorder_int(_id integer) RETURNS void
    LANGUAGE plpgsql
    AS $$
BEGIN		
	WITH a AS (DELETE FROM torder o WHERE o.id=_id RETURNING *) 
	INSERT INTO torderremoved 
		SELECT id,uuid,own,nr,qtt_requ,np,qtt_prov,qtt,start,created,statement_timestamp() 
	FROM a;					
END;
$$;


ALTER FUNCTION public.fremoveorder_int(_id integer) OWNER TO olivier;

--
-- Name: fremovequote_int(integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fremovequote_int(_idquote integer) RETURNS void
    LANGUAGE plpgsql
    AS $$
BEGIN		
	WITH a AS (DELETE FROM tquote o WHERE o.id=_idquote RETURNING *) 
	INSERT INTO tquoteremoved 
		SELECT id,own,nr,qtt_requ,np,qtt_prov,qtt_in,qtt_out,flows,created,statement_timestamp() 
	FROM a;					
END;
$$;


ALTER FUNCTION public.fremovequote_int(_idquote integer) OWNER TO olivier;

--
-- Name: frenumbertables(boolean); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION frenumbertables(exec boolean) RETURNS boolean
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE
	_cnt int;
	_res bool;
BEGIN
	
	SELECT count(*) INTO _cnt FROM tmvt;
	IF (_cnt != 0) THEN
		RAISE INFO 'tmvt must be cleared';
		_res := false;
	ELSE
		_res := true;
	END IF;
		
	IF NOT exec THEN
		RETURN _res;
	END IF;
	
	IF NOT _res THEN
		RAISE WARNING 'tmvt must be cleared';
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	-- desable triggers
	ALTER TABLE towner DISABLE TRIGGER ALL;
	ALTER TABLE tquality DISABLE TRIGGER ALL;
	ALTER TABLE tuser DISABLE TRIGGER ALL;
	
	-- DROP CONSTRAINT ON UPDATE CASCADE on tables tquality,torder,tmvt
    	ALTER TABLE tquality DROP CONSTRAINT ctquality_idd,ADD CONSTRAINT ctquality_idd FOREIGN KEY (idd) references tuser(id) ON UPDATE CASCADE;
	ALTER TABLE tquality DROP CONSTRAINT ctquality_depository,ADD CONSTRAINT ctquality_depository FOREIGN KEY (depository) references tuser(name) ON UPDATE RESTRICT;  -- must not be changed
	  			
	ALTER TABLE torder DROP CONSTRAINT ctorder_own,ADD CONSTRAINT ctorder_own 	FOREIGN KEY (own) references towner(id) ON UPDATE CASCADE;
	ALTER TABLE torder DROP CONSTRAINT ctorder_np,ADD CONSTRAINT ctorder_np 	FOREIGN KEY (np) references tquality(id) ON UPDATE CASCADE;
	ALTER TABLE torder DROP CONSTRAINT ctorder_nr,ADD CONSTRAINT ctorder_nr 	FOREIGN KEY (nr) references tquality(id) ON UPDATE CASCADE;

	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_grp,ADD CONSTRAINT ctmvt_grp 		FOREIGN KEY (grp) references tmvt(id) ON UPDATE CASCADE;
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_own_src,ADD CONSTRAINT ctmvt_own_src 	FOREIGN KEY (own_src) references towner(id) ON UPDATE CASCADE;
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_own_dst,ADD CONSTRAINT ctmvt_own_dst 	FOREIGN KEY (own_dst) references towner(id) ON UPDATE CASCADE;
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_nat,ADD CONSTRAINT ctmvt_nat 		FOREIGN KEY (nat) references tquality(id) ON UPDATE CASCADE;

	TRUNCATE tquote;
	PERFORM setval('tquote_id_seq',1,false);
	
	TRUNCATE tmvt;
	PERFORM setval('tmvt_id_seq',1,false);
	
	-- remove unused qualities
	DELETE FROM tquality q WHERE	q.id NOT IN (SELECT np FROM torder)	
				AND	q.id NOT IN (SELECT nr FROM torder)
				AND 	q.id NOT IN (SELECT nat FROM tmvt);	
	-- renumbering qualities
	PERFORM setval('tquality_id_seq',1,false);
	WITH a AS (SELECT * FROM tquality ORDER BY id ASC)
	UPDATE tquality SET id = nextval('tquality_id_seq') FROM a WHERE a.id = tquality.id;
	
	-- remove unused owners
	DELETE FROM towner o WHERE o.id NOT IN (SELECT idd FROM tquality);
	
	-- renumbering owners
	PERFORM setval('towner_id_seq',1,false);
	WITH a AS (SELECT * FROM towner ORDER BY id ASC)
	UPDATE towner SET id = nextval('towner_id_seq') FROM a WHERE a.id = towner.id;
	
	-- resetting quotas
	UPDATE tuser SET spent = 0;
		
	-- renumbering orders
	PERFORM setval('torder_id_seq',1,false);
	WITH a AS (SELECT * FROM torder ORDER BY id ASC)
	UPDATE torder SET id = nextval('torder_id_seq') FROM a WHERE a.id = torder.id;
		
	TRUNCATE torderremoved; -- does not reset associated sequence if any
	TRUNCATE tmvtremoved;
	TRUNCATE tquoteremoved;
	TRUNCATE treltried;
	
	
	-- reset of constraints
    	ALTER TABLE tquality DROP CONSTRAINT ctquality_idd,ADD CONSTRAINT ctquality_idd FOREIGN KEY (idd) references tuser(id);
    	ALTER TABLE tquality DROP CONSTRAINT ctquality_depository,ADD CONSTRAINT ctquality_depository FOREIGN KEY (depository) references tuser(name);
    		
	ALTER TABLE torder DROP CONSTRAINT ctorder_own,ADD CONSTRAINT ctorder_own 	FOREIGN KEY (own) references towner(id);
	ALTER TABLE torder DROP CONSTRAINT ctorder_np,ADD CONSTRAINT ctorder_np 	FOREIGN KEY (np) references tquality(id);
	ALTER TABLE torder DROP CONSTRAINT ctorder_nr,ADD CONSTRAINT ctorder_nr 	FOREIGN KEY (nr) references tquality(id);

	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_grp,ADD CONSTRAINT ctmvt_grp 		FOREIGN KEY (grp) references tmvt(id);
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_own_src,ADD CONSTRAINT ctmvt_own_src 	FOREIGN KEY (own_src) references towner(id);
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_own_dst,ADD CONSTRAINT ctmvt_own_dst 	FOREIGN KEY (own_dst) references towner(id);
	ALTER TABLE tmvt DROP CONSTRAINT ctmvt_nat,ADD CONSTRAINT ctmvt_nat 		FOREIGN KEY (nat) references tquality(id);
	
	-- enable triggers
	ALTER TABLE towner ENABLE TRIGGER ALL;
	ALTER TABLE tquality ENABLE TRIGGER ALL;
	ALTER TABLE tuser ENABLE TRIGGER ALL;
	
	RAISE INFO 'Run the command:'; 
	RAISE INFO '	VACUUM FULL ANALYZE';
	RAISE INFO 'before starting the market';
	RETURN true;
	 
END;
$$;


ALTER FUNCTION public.frenumbertables(exec boolean) OWNER TO olivier;

--
-- Name: fresetmarket(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fresetmarket() RETURNS void
    LANGUAGE plpgsql SECURITY DEFINER
    AS $$
DECLARE
	_prev_status 	ymarketstatus;
	_new_status 	ymarketstatus;
BEGIN
	LOOP
		_new_status := fchangestatemarket(true); 
		IF(_new_status = 'STOPPING') THEN
			perform fresetmarket_int();
			CONTINUE WHEN true;
		END IF;
		EXIT WHEN _new_status = 'OPENED';
	END LOOP;
	RAISE INFO 'The market is reset and opened';
	RETURN;
END;
$$;


ALTER FUNCTION public.fresetmarket() OWNER TO olivier;

--
-- Name: fresetmarket_int(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fresetmarket_int() RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
	_prev_status ymarketstatus;
BEGIN
	SELECT market_status INTO _prev_status FROM vmarket; 
	IF(_prev_status != 'STOPPING') THEN
		RAISE INFO 'The market state is %!= STOPPING, abort.',_prev_status;
		RAISE EXCEPTION USING ERRCODE='YU001';
	END IF;
	
	TRUNCATE tmvt;
	TRUNCATE torder;
	TRUNCATE towner CASCADE;
	TRUNCATE tquality CASCADE;
	-- TRUNCATE tuser CASCADE;
	-- PERFORM setval('tuser_id_seq',1,false);
	RETURN;
END;
$$;


ALTER FUNCTION public.fresetmarket_int() OWNER TO olivier;

--
-- Name: fspendquota(timestamp without time zone); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fspendquota(_time_begin timestamp without time zone) RETURNS boolean
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_t2	timestamp;
BEGIN
	_t2 := clock_timestamp();
	UPDATE tuser SET spent = spent + extract (microseconds from (_t2-_time_begin)) WHERE name = session_user;
	RETURN true;
END;		
$$;


ALTER FUNCTION public.fspendquota(_time_begin timestamp without time zone) OWNER TO olivier;

--
-- Name: ftime_updated(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION ftime_updated() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
BEGIN
	IF (TG_OP = 'INSERT') THEN
		NEW.created := statement_timestamp();
	ELSE 
		NEW.updated := statement_timestamp();
	END IF;
	RETURN NEW;
END;
$$;


ALTER FUNCTION public.ftime_updated() OWNER TO olivier;

--
-- Name: FUNCTION ftime_updated(); Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON FUNCTION ftime_updated() IS 'trigger updating fields created and updated';


--
-- Name: fupdate_quality(integer, bigint); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fupdate_quality(_qid integer, _qtt bigint) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_qp tquality%rowtype;
	_q text[];
	_id int;
	_qtta int8;
BEGIN

	UPDATE tquality SET qtt = qtt + _qtt 
		WHERE id = _qid RETURNING qtt INTO _qtta;

	IF NOT FOUND THEN
		RAISE EXCEPTION USING ERRCODE='YA004';
	END IF;	
		
	IF (_qtt > 0)   THEN 
		IF (_qtta < _qtt) THEN 
			RAISE WARNING 'Quality "%" owerflows',_quality_name;
			RAISE EXCEPTION USING ERRCODE='YA004';
		END IF; 
	ELSIF (_qtt < 0) THEN
		IF (_qtta > _qtt) THEN 
			RAISE WARNING 'Quality "%" underflows',_quality_name;
			RAISE EXCEPTION USING ERRCODE='YA004';
		END IF;			
	END IF;
	
	RETURN;
END;
$$;


ALTER FUNCTION public.fupdate_quality(_qid integer, _qtt bigint) OWNER TO olivier;

--
-- Name: fupdate_treltried(bigint[], integer); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fupdate_treltried(_commits bigint[], _nbcommit integer) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_i int;
	_np int;
	_nr int;
	_MAXTRY 	int := fgetconst('MAXTRY');
BEGIN
	IF(_MAXTRY=0) THEN
		RETURN;
	END IF;
	
	FOR _i IN 1 .. _nbcommit LOOP
		_nr	:= _commits[_i][3]::int;
		_np	:= _commits[_i][5]::int;
		LOOP
			UPDATE treltried SET cnt = cnt + 1 WHERE np=_np AND nr=_nr;
			IF FOUND THEN
				EXIT;
			ELSE
				BEGIN
					INSERT INTO treltried (np,nr,cnt) VALUES (_np,_nr,1);
				EXCEPTION WHEN check_violation THEN
					-- 
				END;
			END IF;
		END LOOP;
	END LOOP;

	RETURN;
END;
$$;


ALTER FUNCTION public.fupdate_treltried(_commits bigint[], _nbcommit integer) OWNER TO olivier;

--
-- Name: fverifmvt(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fverifmvt() RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_qtt_prov	 int8;
	_qtt		 int8;
	_uuid		 text;
	_qtta		 int8;
	_npa		 int;
	_npb		 int;
	_np		 int;
	_cnterr		 int := 0;
	_iserr		 bool;
BEGIN
	
	FOR _qtt_prov,_qtt,_uuid,_np IN SELECT qtt_prov,qtt,uuid,np FROM vorderverif LOOP
	
		_iserr := false;
	
		SELECT sum(qtt),max(nat),min(nat) INTO _qtta,_npa,_npb 
			FROM vmvtverif WHERE oruuid=_uuid GROUP BY oruuid;
			
		IF(	FOUND ) THEN 
			IF(	(_qtt_prov != _qtta+_qtt) 
				-- NOT vorderverif.qtt_prov == vorderverif.qtt + sum(mvt.qtt)
				OR (_np != _npa)	
				-- NOT mvt.nat == vorderverif.nat 
				OR (_npa != _npb)
				-- NOT all mvt.nat are the same 
			)	THEN 
				_iserr := true;
				
			END IF;	
		END IF;
		
		IF(_iserr) THEN
			_cnterr := _cnterr +1;
			raise INFO 'error on uuid:%',_uuid;
		END IF;
	END LOOP;

	RETURN _cnterr;
END;
$$;


ALTER FUNCTION public.fverifmvt() OWNER TO olivier;

--
-- Name: fverifmvt2(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fverifmvt2() RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_cnterr		 int := 0;
	_cnterrtot	 int := 0;
	_mvt		 tmvt%rowtype;
	_mvtprec	 tmvt%rowtype;
	_mvtfirst	 tmvt%rowtype;
	_uuiderr	 text;
	_cnt		 int;		-- count mvt in agreement
BEGIN
		
	_mvtprec.grp := NULL;_mvtfirst.grp := NULL;
	_uuiderr := NULL;
	FOR _mvt IN SELECT * FROM vmvtverif ORDER BY grp,id ASC  LOOP
		IF(_mvt.grp != _mvtprec.grp) THEN -- first mvt of agreement
			--> finish last agreement
			IF NOT (_mvtprec.grp IS NULL OR _mvtfirst.grp IS NULL) THEN
				_cnterr := _cnterr + fverifmvt2_int(_mvtprec,_mvtfirst);
				_cnt := _cnt +1;
				
				if(_cnt != _mvtprec.nb) THEN
					_cnterr := _cnterr +1;
					RAISE INFO 'wrong number of movements for agreement %',_mvtprec.oruuid;
				END IF;
				-- errors found
				if(_cnterr != 0) THEN
					_cnterrtot := _cnterr + _cnterrtot;
					IF(_uuiderr IS NULL) THEN
						_uuiderr := _mvtprec.oruuid;
					END IF;
				END IF;
			END IF;
			--< A
			_mvtfirst := _mvt;
			_cnt := 0;
			_cnterr := 0;
		ELSE
			_cnterr := _cnterr + fverifmvt2_int(_mvtprec,_mvt);
			_cnt := _cnt +1;
		END IF;
		_mvtprec := _mvt;
	END LOOP;
	--> finish last agreement
	IF NOT (_mvtprec.grp IS NULL OR _mvtfirst.grp IS NULL) THEN
		_cnterr := _cnterr + fverifmvt2_int(_mvtprec,_mvtfirst);
		_cnt := _cnt +1;
		
		if(_cnt != _mvtprec.nb) THEN
			_cnterr := _cnterr +1;
			RAISE INFO 'wrong number of movements for agreement %',_mvtprec.oruuid;
		END IF;
		-- errors found
		if(_cnterr != 0) THEN
			_cnterrtot := _cnterr + _cnterrtot;
			IF(_uuiderr IS NULL) THEN
				_uuiderr := _mvtprec.oruuid;
			END IF;
		END IF;
	END IF;
	--< A
	IF(_cnterrtot != 0) THEN
		RAISE NOTICE 'mvt.oruuid= % is the first agreement where an error is found',_uuiderr;
		RETURN _cnterrtot;
	ELSE
		RETURN 0;
	END IF;
END;
$$;


ALTER FUNCTION public.fverifmvt2() OWNER TO olivier;

--
-- Name: tmvt; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tmvt (
    id integer NOT NULL,
    nb integer NOT NULL,
    oruuid text NOT NULL,
    grp integer,
    own_src integer NOT NULL,
    own_dst integer NOT NULL,
    qtt dquantity NOT NULL,
    nat integer NOT NULL,
    created timestamp without time zone NOT NULL,
    CONSTRAINT tmvt_check CHECK ((((nb = 1) AND (own_src = own_dst)) OR (nb <> 1)))
);


ALTER TABLE public.tmvt OWNER TO olivier;

--
-- Name: TABLE tmvt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON TABLE tmvt IS 'records a change of ownership';


--
-- Name: COLUMN tmvt.nb; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.nb IS 'number of movements of the exchange';


--
-- Name: COLUMN tmvt.oruuid; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.oruuid IS 'order.uuid producing  this movement';


--
-- Name: COLUMN tmvt.grp; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.grp IS 'references the first movement of the exchange';


--
-- Name: COLUMN tmvt.own_src; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.own_src IS 'owner provider';


--
-- Name: COLUMN tmvt.own_dst; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.own_dst IS 'owner receiver';


--
-- Name: COLUMN tmvt.qtt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.qtt IS 'quantity of the value moved';


--
-- Name: COLUMN tmvt.nat; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN tmvt.nat IS 'quality of the value moved';


--
-- Name: fverifmvt2_int(tmvt, tmvt); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fverifmvt2_int(_mvtprec tmvt, _mvt tmvt) RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE
	_o		vorderverif%rowtype;
BEGIN
	SELECT uuid,np,nr,qtt_prov,qtt_requ INTO _o.uuid,_o.np,_o.nr,_o.qtt_prov,_o.qtt_requ FROM vorderverif WHERE uuid = _mvt.oruuid;
	IF (NOT FOUND) THEN
		RAISE INFO 'order not found for vorderverif %',_mvt.oruuid;
		RETURN 1;
	END IF;

	IF(_o.np != _mvt.nat OR _o.nr != _mvtprec.nat) THEN
		RAISE INFO 'mvt.nat != np or mvtprec.nat!=nr';
		RETURN 1;
	END IF;
	
	-- _o.qtt_prov/_o.qtt_requ < _mvt.qtt/_mvtprec.qtt
	IF(((_o.qtt_prov::float8) / (_o.qtt_requ::float8)) < ((_mvt.qtt::float8)/(_mvtprec.qtt::float8))) THEN
		RAISE INFO 'order %->%, with  mvt %->%',_o.qtt_requ,_o.qtt_prov,_mvtprec.qtt,_mvt.qtt;
		RAISE INFO '% < 1; should be >=1',(((_o.qtt_prov::float8) / (_o.qtt_requ::float8)) / ((_mvt.qtt::float8)/(_mvtprec.qtt::float8)));
		RAISE INFO 'order.uuid %, with  mvtid %->%',_o.uuid,_mvtprec.id,_mvt.id;
		RETURN 1;
	END IF;


	RETURN 0;
END;
$$;


ALTER FUNCTION public.fverifmvt2_int(_mvtprec tmvt, _mvt tmvt) OWNER TO olivier;

--
-- Name: fverifyquota(); Type: FUNCTION; Schema: public; Owner: olivier
--

CREATE FUNCTION fverifyquota() RETURNS integer
    LANGUAGE plpgsql
    AS $$
DECLARE 
	_u	tuser%rowtype;
BEGIN
	SELECT * INTO _u FROM tuser WHERE name=session_user;
	IF(_u.id is NULL) THEN
		RAISE WARNING 'the user % is undefined',session_user;
		RAISE EXCEPTION USING ERRCODE='YA005';
	END IF;
	UPDATE tuser SET last_in = statement_timestamp() WHERE name = session_user;
	IF(_u.quota = 0 ) THEN
		RETURN _u.id;
	END IF;

	IF(_u.quota < _u.spent) THEN
		RAISE WARNING 'the quota is reached for the user %',session_user;
		RAISE EXCEPTION USING ERRCODE='YU003';
	END IF;
	RETURN _u.id;

END;		
$$;


ALTER FUNCTION public.fverifyquota() OWNER TO olivier;

--
-- Name: tconst; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tconst (
    name text NOT NULL,
    value integer
);


ALTER TABLE public.tconst OWNER TO olivier;

--
-- Name: tmarket; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tmarket (
    id integer NOT NULL,
    created timestamp without time zone NOT NULL
);


ALTER TABLE public.tmarket OWNER TO olivier;

--
-- Name: tmarket_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE tmarket_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.tmarket_id_seq OWNER TO olivier;

--
-- Name: tmarket_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE tmarket_id_seq OWNED BY tmarket.id;


--
-- Name: tmarket_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('tmarket_id_seq', 1, true);


--
-- Name: tmvt_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE tmvt_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.tmvt_id_seq OWNER TO olivier;

--
-- Name: tmvt_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE tmvt_id_seq OWNED BY tmvt.id;


--
-- Name: tmvt_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('tmvt_id_seq', 44610, true);


--
-- Name: tmvtremoved; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tmvtremoved (
    id integer NOT NULL,
    nb integer NOT NULL,
    oruuid text NOT NULL,
    grp integer NOT NULL,
    own_src integer NOT NULL,
    own_dst integer NOT NULL,
    qtt dquantity NOT NULL,
    nat integer NOT NULL,
    created timestamp without time zone NOT NULL,
    deleted timestamp without time zone NOT NULL
);


ALTER TABLE public.tmvtremoved OWNER TO olivier;

--
-- Name: torder_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE torder_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.torder_id_seq OWNER TO olivier;

--
-- Name: torder_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE torder_id_seq OWNED BY torder.id;


--
-- Name: torder_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('torder_id_seq', 14040, true);


--
-- Name: torderremoved; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE torderremoved (
    id integer NOT NULL,
    uuid text NOT NULL,
    own integer NOT NULL,
    nr integer NOT NULL,
    qtt_requ dquantity NOT NULL,
    np integer NOT NULL,
    qtt_prov dquantity NOT NULL,
    qtt bigint NOT NULL,
    start bigint,
    created timestamp without time zone NOT NULL,
    updated timestamp without time zone
);


ALTER TABLE public.torderremoved OWNER TO olivier;

--
-- Name: towner_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE towner_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.towner_id_seq OWNER TO olivier;

--
-- Name: towner_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE towner_id_seq OWNED BY towner.id;


--
-- Name: towner_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('towner_id_seq', 100, true);


--
-- Name: tquality_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE tquality_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.tquality_id_seq OWNER TO olivier;

--
-- Name: tquality_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE tquality_id_seq OWNED BY tquality.id;


--
-- Name: tquality_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('tquality_id_seq', 100, true);


--
-- Name: tquote_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE tquote_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.tquote_id_seq OWNER TO olivier;

--
-- Name: tquote_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE tquote_id_seq OWNED BY tquote.id;


--
-- Name: tquote_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('tquote_id_seq', 1, false);


--
-- Name: tquoteremoved; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tquoteremoved (
    id integer NOT NULL,
    own integer NOT NULL,
    nr integer NOT NULL,
    qtt_requ bigint,
    np integer NOT NULL,
    qtt_prov bigint,
    qtt_in bigint,
    qtt_out bigint,
    flows yflow[],
    created timestamp without time zone,
    removed timestamp without time zone
);


ALTER TABLE public.tquoteremoved OWNER TO olivier;

--
-- Name: treltried; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE treltried (
    np integer NOT NULL,
    nr integer NOT NULL,
    cnt bigint DEFAULT 0,
    CONSTRAINT treltried_check CHECK (((np <> nr) AND (cnt >= 0)))
);


ALTER TABLE public.treltried OWNER TO olivier;

--
-- Name: tuser; Type: TABLE; Schema: public; Owner: olivier; Tablespace: 
--

CREATE TABLE tuser (
    id integer NOT NULL,
    name text NOT NULL,
    spent bigint DEFAULT 0 NOT NULL,
    quota bigint DEFAULT 0 NOT NULL,
    last_in timestamp without time zone,
    created timestamp without time zone,
    updated timestamp without time zone,
    CONSTRAINT tuser_check CHECK ((((char_length(name) > 0) AND (spent >= 0)) AND (quota >= 0)))
);


ALTER TABLE public.tuser OWNER TO olivier;

--
-- Name: TABLE tuser; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON TABLE tuser IS 'users that have been connected';


--
-- Name: tuser_id_seq; Type: SEQUENCE; Schema: public; Owner: olivier
--

CREATE SEQUENCE tuser_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER TABLE public.tuser_id_seq OWNER TO olivier;

--
-- Name: tuser_id_seq; Type: SEQUENCE OWNED BY; Schema: public; Owner: olivier
--

ALTER SEQUENCE tuser_id_seq OWNED BY tuser.id;


--
-- Name: tuser_id_seq; Type: SEQUENCE SET; Schema: public; Owner: olivier
--

SELECT pg_catalog.setval('tuser_id_seq', 1, true);


--
-- Name: vmarket; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vmarket AS
    SELECT ((tmarket.id + 4) / 4) AS market_session, CASE WHEN (((tmarket.id - 1) % 4) = 0) THEN 'OPENED'::ymarketstatus WHEN (((tmarket.id - 1) % 4) = 1) THEN 'STOPPING'::ymarketstatus WHEN (((tmarket.id - 1) % 4) = 2) THEN 'CLOSED'::ymarketstatus WHEN (((tmarket.id - 1) % 4) = 3) THEN 'STARTING'::ymarketstatus ELSE NULL::ymarketstatus END AS market_status, tmarket.created FROM tmarket ORDER BY tmarket.id DESC LIMIT 1;


ALTER TABLE public.vmarket OWNER TO olivier;

--
-- Name: vmvt; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vmvt AS
    SELECT m.id, m.nb, m.oruuid, m.grp, w_src.name AS provider, q.name AS quality, m.qtt, w_dst.name AS receiver, m.created FROM (((tmvt m JOIN towner w_src ON ((m.own_src = w_src.id))) JOIN towner w_dst ON ((m.own_dst = w_dst.id))) JOIN tquality q ON ((m.nat = q.id)));


ALTER TABLE public.vmvt OWNER TO olivier;

--
-- Name: VIEW vmvt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON VIEW vmvt IS 'List of movements';


--
-- Name: COLUMN vmvt.nb; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.nb IS 'number of movements of the exchange';


--
-- Name: COLUMN vmvt.oruuid; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.oruuid IS 'reference to the exchange';


--
-- Name: COLUMN vmvt.grp; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.grp IS 'id of the first movement of the exchange';


--
-- Name: COLUMN vmvt.provider; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.provider IS 'name of the provider of the movement';


--
-- Name: COLUMN vmvt.quality; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.quality IS 'name of the quality moved';


--
-- Name: COLUMN vmvt.qtt; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.qtt IS 'quantity moved';


--
-- Name: COLUMN vmvt.receiver; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.receiver IS 'name of the receiver of the movement';


--
-- Name: COLUMN vmvt.created; Type: COMMENT; Schema: public; Owner: olivier
--

COMMENT ON COLUMN vmvt.created IS 'time of the transaction';


--
-- Name: vmvtverif; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vmvtverif AS
    SELECT tmvt.id, tmvt.nb, tmvt.oruuid, tmvt.grp, tmvt.own_src, tmvt.own_dst, tmvt.qtt, tmvt.nat, tmvt.created FROM tmvt WHERE (tmvt.grp IS NOT NULL) UNION ALL SELECT tmvtremoved.id, tmvtremoved.nb, tmvtremoved.oruuid, tmvtremoved.grp, tmvtremoved.own_src, tmvtremoved.own_dst, tmvtremoved.qtt, tmvtremoved.nat, tmvtremoved.created FROM tmvtremoved WHERE (tmvtremoved.grp IS NOT NULL);


ALTER TABLE public.vmvtverif OWNER TO olivier;

--
-- Name: vorderinsert; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vorderinsert AS
    SELECT torder.id, yorder_get(torder.id, torder.own, torder.nr, torder.qtt_requ, torder.np, (torder.qtt_prov)::bigint, torder.qtt) AS ord, torder.np, torder.nr FROM torder ORDER BY ((torder.qtt_prov)::double precision / (torder.qtt_requ)::double precision) DESC;


ALTER TABLE public.vorderinsert OWNER TO olivier;

--
-- Name: vorderremoved; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vorderremoved AS
    SELECT n.id, n.uuid, w.name AS owner, qr.name AS qua_requ, n.qtt_requ, qp.name AS qua_prov, n.qtt_prov, n.qtt, n.created, n.updated FROM (((torderremoved n JOIN tquality qr ON ((n.nr = qr.id))) JOIN tquality qp ON ((n.np = qp.id))) JOIN towner w ON ((n.own = w.id)));


ALTER TABLE public.vorderremoved OWNER TO olivier;

--
-- Name: vorderverif; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vorderverif AS
    SELECT torder.id, torder.uuid, torder.own, torder.nr, torder.qtt_requ, torder.np, torder.qtt_prov, torder.qtt, torder.created FROM torder UNION SELECT torderremoved.id, torderremoved.uuid, torderremoved.own, torderremoved.nr, torderremoved.qtt_requ, torderremoved.np, torderremoved.qtt_prov, torderremoved.qtt, torderremoved.created FROM torderremoved;


ALTER TABLE public.vorderverif OWNER TO olivier;

--
-- Name: tquality_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tquality
    ADD CONSTRAINT tquality_pkey PRIMARY KEY (id);


--
-- Name: vstat; Type: VIEW; Schema: public; Owner: olivier
--

CREATE VIEW vstat AS
    SELECT q.name, (sum(d.qtt) - (q.qtt)::numeric) AS delta, q.qtt AS qtt_quality, sum(d.qtt) AS qtt_detail FROM ((SELECT vorderverif.np AS nat, vorderverif.qtt FROM vorderverif UNION ALL SELECT vmvtverif.nat, vmvtverif.qtt FROM vmvtverif) d JOIN tquality q ON ((d.nat = q.id))) GROUP BY q.id ORDER BY q.name;


ALTER TABLE public.vstat OWNER TO olivier;

--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE tmarket ALTER COLUMN id SET DEFAULT nextval('tmarket_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE tmvt ALTER COLUMN id SET DEFAULT nextval('tmvt_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE torder ALTER COLUMN id SET DEFAULT nextval('torder_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE towner ALTER COLUMN id SET DEFAULT nextval('towner_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE tquality ALTER COLUMN id SET DEFAULT nextval('tquality_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE tquote ALTER COLUMN id SET DEFAULT nextval('tquote_id_seq'::regclass);


--
-- Name: id; Type: DEFAULT; Schema: public; Owner: olivier
--

ALTER TABLE tuser ALTER COLUMN id SET DEFAULT nextval('tuser_id_seq'::regclass);


--
-- Data for Name: tconst; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tconst (name, value) FROM stdin;
VERSION-X.y.z	0
VERSION-x.Y.y	3
VERSION-x.y.Z	2
INSERT_OWN_UNKNOWN	1
CHECK_QUALITY_OWNERSHIP	0
MAXTRY	10
MAXCYCLE	8
MAXORDERFETCH	10000
\.


--
-- Data for Name: tmarket; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tmarket (id, created) FROM stdin;
1	2012-08-02 04:43:53.309164
\.


--
-- Data for Name: tmvt; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tmvt (id, nb, oruuid, grp, own_src, own_dst, qtt, nat, created) FROM stdin;
2	4	1-23	1	23	29	4192	11	2012-08-02 04:43:54.314959
3	4	1-70	1	29	11	2445	16	2012-08-02 04:43:54.314959
4	4	1-11	1	11	53	4480	19	2012-08-02 04:43:54.314959
1	4	1-77	1	53	23	827	23	2012-08-02 04:43:54.314959
838	8	1-406	837	18	85	1326	28	2012-08-02 04:44:05.480499
6	6	1-7	5	7	44	3016	12	2012-08-02 04:43:55.251645
7	6	1-56	5	44	63	2114	22	2012-08-02 04:43:55.251645
8	6	1-113	5	63	2	3175	4	2012-08-02 04:43:55.251645
9	6	1-2	5	2	22	7838	3	2012-08-02 04:43:55.251645
10	6	1-139	5	22	45	1563	55	2012-08-02 04:43:55.251645
5	6	1-152	5	45	7	950	13	2012-08-02 04:43:55.251645
12	5	1-100	11	53	46	5801	41	2012-08-02 04:43:55.286995
13	5	1-65	11	46	42	7772	71	2012-08-02 04:43:55.286995
14	5	1-74	11	42	23	5142	76	2012-08-02 04:43:55.286995
15	5	1-111	11	23	67	2534	46	2012-08-02 04:43:55.286995
11	5	1-154	11	67	53	849	86	2012-08-02 04:43:55.286995
17	7	1-62	16	21	42	659	71	2012-08-02 04:43:55.398174
18	7	1-74	16	42	23	518	76	2012-08-02 04:43:55.398174
19	7	1-111	16	23	30	303	46	2012-08-02 04:43:55.398174
20	7	1-32	16	30	7	112	45	2012-08-02 04:43:55.398174
21	7	1-57	16	7	39	493	56	2012-08-02 04:43:55.398174
22	7	1-49	16	39	64	2809	24	2012-08-02 04:43:55.398174
16	7	1-160	16	64	21	959	72	2012-08-02 04:43:55.398174
24	3	1-44	23	36	25	6331	36	2012-08-02 04:43:55.565418
25	3	1-25	23	25	58	1793	35	2012-08-02 04:43:55.565418
23	3	1-173	23	58	36	3117	58	2012-08-02 04:43:55.565418
27	7	1-175	26	46	3	1764	66	2012-08-02 04:43:55.632322
28	7	1-53	26	3	45	2365	65	2012-08-02 04:43:55.632322
29	7	1-102	26	45	51	7927	44	2012-08-02 04:43:55.632322
30	7	1-85	26	51	36	7445	80	2012-08-02 04:43:55.632322
31	7	1-107	26	36	56	1026	28	2012-08-02 04:43:55.632322
32	7	1-136	26	56	80	319	90	2012-08-02 04:43:55.632322
26	7	1-178	26	80	46	1027	41	2012-08-02 04:43:55.632322
34	6	1-72	33	49	35	74	8	2012-08-02 04:43:55.877365
35	6	1-43	33	35	39	14	40	2012-08-02 04:43:55.877365
36	6	1-81	33	39	53	6570	19	2012-08-02 04:43:55.877365
37	6	1-77	33	53	23	770	23	2012-08-02 04:43:55.877365
38	6	1-23	33	23	63	2483	11	2012-08-02 04:43:55.877365
33	6	1-196	33	63	49	1002	28	2012-08-02 04:43:55.877365
40	8	1-136	39	56	80	374	90	2012-08-02 04:43:55.877365
41	8	1-178	39	80	83	964	41	2012-08-02 04:43:55.877365
42	8	1-168	39	83	50	648	77	2012-08-02 04:43:55.877365
43	8	1-128	39	50	61	740	88	2012-08-02 04:43:55.877365
44	8	1-109	39	61	84	4198	87	2012-08-02 04:43:55.877365
45	8	1-165	39	84	14	1470	38	2012-08-02 04:43:55.877365
46	8	1-137	39	14	63	1244	11	2012-08-02 04:43:55.877365
39	8	1-196	39	63	56	1500	28	2012-08-02 04:43:55.877365
48	8	1-18	47	18	22	344	3	2012-08-02 04:43:55.877365
49	8	1-139	47	22	45	56	55	2012-08-02 04:43:55.877365
50	8	1-152	47	45	14	28	13	2012-08-02 04:43:55.877365
51	8	1-138	47	14	6	45	53	2012-08-02 04:43:55.877365
52	8	1-88	47	6	3	56	82	2012-08-02 04:43:55.877365
53	8	1-190	47	3	23	25	23	2012-08-02 04:43:55.877365
54	8	1-23	47	23	63	302	11	2012-08-02 04:43:55.877365
47	8	1-196	47	63	18	458	28	2012-08-02 04:43:55.877365
56	8	1-136	55	56	80	699	90	2012-08-02 04:43:55.877365
57	8	1-178	55	80	83	2719	41	2012-08-02 04:43:55.877365
58	8	1-168	55	83	69	2837	77	2012-08-02 04:43:55.877365
59	8	1-185	55	69	83	227	81	2012-08-02 04:43:55.877365
60	8	1-176	55	83	84	1513	87	2012-08-02 04:43:55.877365
61	8	1-165	55	84	14	800	38	2012-08-02 04:43:55.877365
62	8	1-137	55	14	63	1021	11	2012-08-02 04:43:55.877365
55	8	1-196	55	63	56	1857	28	2012-08-02 04:43:55.877365
64	8	1-3	63	3	7	802	5	2012-08-02 04:43:56.066747
65	8	1-87	63	7	56	1514	81	2012-08-02 04:43:56.066747
66	8	1-90	63	56	26	112	67	2012-08-02 04:43:56.066747
67	8	1-184	63	26	42	216	9	2012-08-02 04:43:56.066747
68	8	1-193	63	42	45	235	65	2012-08-02 04:43:56.066747
69	8	1-102	63	45	29	409	44	2012-08-02 04:43:56.066747
70	8	1-31	63	29	15	2758	25	2012-08-02 04:43:56.066747
63	8	1-206	63	15	3	1105	6	2012-08-02 04:43:56.066747
72	7	1-3	71	3	21	4245	5	2012-08-02 04:43:56.066747
73	7	1-141	71	21	12	828	20	2012-08-02 04:43:56.066747
74	7	1-12	71	12	81	113	12	2012-08-02 04:43:56.066747
75	7	1-157	71	81	45	151	65	2012-08-02 04:43:56.066747
76	7	1-102	71	45	29	445	44	2012-08-02 04:43:56.066747
77	7	1-31	71	29	15	5083	25	2012-08-02 04:43:56.066747
71	7	1-206	71	15	3	3452	6	2012-08-02 04:43:56.066747
79	7	1-103	78	12	84	1232	87	2012-08-02 04:43:56.122278
80	7	1-165	78	84	14	282	38	2012-08-02 04:43:56.122278
81	7	1-137	78	14	63	156	11	2012-08-02 04:43:56.122278
82	7	1-196	78	63	49	123	28	2012-08-02 04:43:56.122278
83	7	1-72	78	49	56	18	8	2012-08-02 04:43:56.122278
84	7	1-197	78	56	50	600	49	2012-08-02 04:43:56.122278
78	7	1-207	78	50	12	176	73	2012-08-02 04:43:56.122278
86	7	1-64	85	42	25	2675	36	2012-08-02 04:43:56.122278
87	7	1-25	85	25	52	1645	35	2012-08-02 04:43:56.122278
88	7	1-99	85	52	63	623	11	2012-08-02 04:43:56.122278
89	7	1-196	85	63	49	699	28	2012-08-02 04:43:56.122278
90	7	1-72	85	49	56	144	8	2012-08-02 04:43:56.122278
91	7	1-197	85	56	50	6874	49	2012-08-02 04:43:56.122278
85	7	1-207	85	50	42	2863	73	2012-08-02 04:43:56.122278
93	8	1-124	92	59	28	1602	11	2012-08-02 04:43:56.178066
94	8	1-30	92	28	80	1204	43	2012-08-02 04:43:56.178066
95	8	1-202	92	80	64	5666	24	2012-08-02 04:43:56.178066
96	8	1-160	92	64	86	1887	72	2012-08-02 04:43:56.178066
97	8	1-179	92	86	6	542	53	2012-08-02 04:43:56.178066
98	8	1-88	92	6	12	739	82	2012-08-02 04:43:56.178066
99	8	1-98	92	12	21	2666	32	2012-08-02 04:43:56.178066
92	8	1-210	92	21	59	2901	17	2012-08-02 04:43:56.178066
101	6	1-124	100	59	63	446	11	2012-08-02 04:43:56.178066
102	6	1-196	100	63	56	792	28	2012-08-02 04:43:56.178066
103	6	1-136	100	56	80	291	90	2012-08-02 04:43:56.178066
104	6	1-178	100	80	61	1104	41	2012-08-02 04:43:56.178066
105	6	1-188	100	61	21	650	32	2012-08-02 04:43:56.178066
100	6	1-210	100	21	59	756	17	2012-08-02 04:43:56.178066
107	8	1-111	106	23	67	640	46	2012-08-02 04:43:56.467323
108	8	1-154	106	67	53	211	86	2012-08-02 04:43:56.467323
109	8	1-100	106	53	46	1425	41	2012-08-02 04:43:56.467323
110	8	1-175	106	46	3	2302	66	2012-08-02 04:43:56.467323
111	8	1-53	106	3	51	2902	65	2012-08-02 04:43:56.467323
112	8	1-71	106	51	13	1584	21	2012-08-02 04:43:56.467323
113	8	1-13	106	13	21	651	3	2012-08-02 04:43:56.467323
106	8	1-225	106	21	23	1316	76	2012-08-02 04:43:56.467323
115	8	1-111	114	23	30	745	46	2012-08-02 04:43:56.467323
116	8	1-32	114	30	82	259	45	2012-08-02 04:43:56.467323
117	8	1-164	114	82	7	470	13	2012-08-02 04:43:56.467323
118	8	1-7	114	7	81	1291	12	2012-08-02 04:43:56.467323
119	8	1-157	114	81	51	2085	65	2012-08-02 04:43:56.467323
120	8	1-71	114	51	13	1284	21	2012-08-02 04:43:56.467323
121	8	1-13	114	13	21	596	3	2012-08-02 04:43:56.467323
114	8	1-225	114	21	23	1358	76	2012-08-02 04:43:56.467323
123	7	1-153	122	79	82	2715	60	2012-08-02 04:43:56.467323
124	7	1-159	122	82	21	355	32	2012-08-02 04:43:56.467323
125	7	1-210	122	21	59	417	17	2012-08-02 04:43:56.467323
126	7	1-124	122	59	63	230	11	2012-08-02 04:43:56.467323
127	7	1-196	122	63	18	383	28	2012-08-02 04:43:56.467323
128	7	1-18	122	18	21	316	3	2012-08-02 04:43:56.467323
122	7	1-225	122	21	79	808	76	2012-08-02 04:43:56.467323
130	8	1-223	129	57	39	350	91	2012-08-02 04:43:56.556197
131	8	1-115	129	39	80	242	90	2012-08-02 04:43:56.556197
132	8	1-178	129	80	46	617	41	2012-08-02 04:43:56.556197
133	8	1-175	129	46	3	840	66	2012-08-02 04:43:56.556197
134	8	1-53	129	3	51	1135	65	2012-08-02 04:43:56.556197
135	8	1-71	129	51	13	522	21	2012-08-02 04:43:56.556197
136	8	1-13	129	13	3	181	3	2012-08-02 04:43:56.556197
129	8	1-228	129	3	57	2032	79	2012-08-02 04:43:56.556197
138	6	1-54	137	42	26	4	67	2012-08-02 04:43:56.600607
139	6	1-184	137	26	42	17	9	2012-08-02 04:43:56.600607
140	6	1-193	137	42	45	41	65	2012-08-02 04:43:56.600607
141	6	1-102	137	45	51	154	44	2012-08-02 04:43:56.600607
142	6	1-85	137	51	87	161	80	2012-08-02 04:43:56.600607
137	6	1-230	137	87	42	73	54	2012-08-02 04:43:56.600607
144	5	1-8	143	8	2	3403	14	2012-08-02 04:43:56.656361
145	5	1-40	143	2	45	4063	55	2012-08-02 04:43:56.656361
146	5	1-152	143	45	7	1839	13	2012-08-02 04:43:56.656361
147	5	1-7	143	7	33	4347	12	2012-08-02 04:43:56.656361
143	5	1-231	143	33	8	2651	15	2012-08-02 04:43:56.656361
149	6	1-50	148	30	45	3699	55	2012-08-02 04:43:56.713905
150	6	1-152	148	45	14	877	13	2012-08-02 04:43:56.713905
151	6	1-138	148	14	6	664	53	2012-08-02 04:43:56.713905
152	6	1-88	148	6	12	392	82	2012-08-02 04:43:56.713905
153	6	1-98	148	12	48	612	32	2012-08-02 04:43:56.713905
148	6	1-234	148	48	30	3169	48	2012-08-02 04:43:56.713905
155	6	1-175	154	46	61	979	66	2012-08-02 04:43:56.76951
156	6	1-158	154	61	84	702	98	2012-08-02 04:43:56.76951
157	6	1-226	154	84	12	2774	20	2012-08-02 04:43:56.76951
158	6	1-144	154	12	2	3224	37	2012-08-02 04:43:56.76951
159	6	1-45	154	2	75	6786	59	2012-08-02 04:43:56.76951
154	6	1-236	154	75	46	453	41	2012-08-02 04:43:56.76951
161	8	1-175	160	46	61	831	66	2012-08-02 04:43:56.858115
162	8	1-158	160	61	84	394	98	2012-08-02 04:43:56.858115
163	8	1-226	160	84	83	1033	20	2012-08-02 04:43:56.858115
164	8	1-162	160	83	2	357	14	2012-08-02 04:43:56.858115
165	8	1-40	160	2	45	389	55	2012-08-02 04:43:56.858115
166	8	1-152	160	45	14	161	13	2012-08-02 04:43:56.858115
167	8	1-138	160	14	52	212	53	2012-08-02 04:43:56.858115
160	8	1-239	160	52	46	581	41	2012-08-02 04:43:56.858115
169	7	1-168	168	83	3	3047	77	2012-08-02 04:43:56.858115
170	7	1-79	168	3	8	892	15	2012-08-02 04:43:56.858115
171	7	1-8	168	8	2	1224	14	2012-08-02 04:43:56.858115
172	7	1-40	168	2	45	1561	55	2012-08-02 04:43:56.858115
173	7	1-152	168	45	14	755	13	2012-08-02 04:43:56.858115
174	7	1-138	168	14	52	1165	53	2012-08-02 04:43:56.858115
168	7	1-239	168	52	83	3732	41	2012-08-02 04:43:56.858115
176	7	1-168	175	83	3	1630	77	2012-08-02 04:43:56.858115
177	7	1-79	175	3	8	596	15	2012-08-02 04:43:56.858115
178	7	1-8	175	8	19	1022	14	2012-08-02 04:43:56.858115
179	7	1-19	175	19	29	1249	29	2012-08-02 04:43:56.858115
180	7	1-78	175	29	86	1261	72	2012-08-02 04:43:56.858115
181	7	1-179	175	86	52	399	53	2012-08-02 04:43:56.858115
175	7	1-239	175	52	83	1598	41	2012-08-02 04:43:56.858115
183	8	1-28	182	4	13	717	40	2012-08-02 04:43:56.858115
184	8	1-201	182	13	33	742	12	2012-08-02 04:43:56.858115
185	8	1-231	182	33	8	632	15	2012-08-02 04:43:56.858115
186	8	1-8	182	8	19	1132	14	2012-08-02 04:43:56.858115
187	8	1-19	182	19	29	1446	29	2012-08-02 04:43:56.858115
188	8	1-78	182	29	86	1526	72	2012-08-02 04:43:56.858115
189	8	1-179	182	86	52	505	53	2012-08-02 04:43:56.858115
182	8	1-239	182	52	4	2114	41	2012-08-02 04:43:56.858115
191	8	1-9	190	9	32	255	16	2012-08-02 04:43:57.028272
192	8	1-198	190	32	26	122	67	2012-08-02 04:43:57.028272
193	8	1-184	190	26	42	479	9	2012-08-02 04:43:57.028272
194	8	1-193	190	42	51	1054	65	2012-08-02 04:43:57.028272
195	8	1-71	190	51	13	645	21	2012-08-02 04:43:57.028272
196	8	1-13	190	13	3	298	3	2012-08-02 04:43:57.028272
197	8	1-228	190	3	23	3496	79	2012-08-02 04:43:57.028272
190	8	1-246	190	23	9	546	17	2012-08-02 04:43:57.028272
199	3	1-133	198	40	4	2222	55	2012-08-02 04:43:57.113934
200	3	1-151	198	4	93	381	64	2012-08-02 04:43:57.113934
198	3	1-250	198	93	40	5650	92	2012-08-02 04:43:57.113934
202	6	1-208	201	19	14	116	13	2012-08-02 04:43:57.169427
203	6	1-138	201	14	6	194	53	2012-08-02 04:43:57.169427
204	6	1-88	201	6	12	254	82	2012-08-02 04:43:57.169427
205	6	1-98	201	12	21	879	32	2012-08-02 04:43:57.169427
206	6	1-21	201	21	76	2526	31	2012-08-02 04:43:57.169427
201	6	1-252	201	76	19	103	5	2012-08-02 04:43:57.169427
208	7	1-178	207	80	61	1680	41	2012-08-02 04:43:57.213833
209	7	1-188	207	61	21	859	32	2012-08-02 04:43:57.213833
210	7	1-210	207	21	59	869	17	2012-08-02 04:43:57.213833
207	7	1-253	207	71	80	509	90	2012-08-02 04:43:57.213833
211	7	1-124	207	59	29	446	11	2012-08-02 04:43:57.213833
212	7	1-70	207	29	39	632	16	2012-08-02 04:43:57.213833
213	7	1-148	207	39	71	359	87	2012-08-02 04:43:57.213833
839	8	1-336	837	85	12	954	20	2012-08-02 04:44:05.480499
215	6	1-165	214	84	14	727	38	2012-08-02 04:43:57.358117
216	6	1-137	214	14	28	485	11	2012-08-02 04:43:57.358117
217	6	1-30	214	28	80	209	43	2012-08-02 04:43:57.358117
218	6	1-202	214	80	41	563	24	2012-08-02 04:43:57.358117
219	6	1-52	214	41	6	297	59	2012-08-02 04:43:57.358117
214	6	1-260	214	6	84	2632	87	2012-08-02 04:43:57.358117
840	8	1-144	837	12	65	948	37	2012-08-02 04:44:05.480499
221	7	1-219	220	34	76	3475	31	2012-08-02 04:43:57.480325
222	7	1-252	220	76	74	76	5	2012-08-02 04:43:57.480325
223	7	1-258	220	74	66	579	4	2012-08-02 04:43:57.480325
224	7	1-119	220	66	29	57	44	2012-08-02 04:43:57.480325
225	7	1-31	220	29	15	423	25	2012-08-02 04:43:57.480325
226	7	1-206	220	15	70	186	6	2012-08-02 04:43:57.480325
220	7	1-264	220	70	34	101	42	2012-08-02 04:43:57.480325
841	8	1-329	837	65	67	1252	46	2012-08-02 04:44:05.480499
228	8	1-126	227	70	62	419	74	2012-08-02 04:43:57.54682
229	8	1-112	227	62	60	4837	52	2012-08-02 04:43:57.54682
230	8	1-262	227	60	50	1614	77	2012-08-02 04:43:57.54682
231	8	1-128	227	50	20	2076	88	2012-08-02 04:43:57.54682
232	8	1-254	227	20	69	1963	1	2012-08-02 04:43:57.54682
233	8	1-125	227	69	14	1979	23	2012-08-02 04:43:57.54682
234	8	1-14	227	14	32	2158	22	2012-08-02 04:43:57.54682
227	8	1-265	227	32	70	460	34	2012-08-02 04:43:57.54682
842	8	1-154	837	67	52	473	86	2012-08-02 04:44:05.480499
236	8	1-170	235	28	41	2055	24	2012-08-02 04:43:57.735435
237	8	1-52	235	41	75	1502	59	2012-08-02 04:43:57.735435
238	8	1-236	235	75	76	70	41	2012-08-02 04:43:57.735435
239	8	1-142	235	76	58	16	62	2012-08-02 04:43:57.735435
240	8	1-93	235	58	26	6	67	2012-08-02 04:43:57.735435
241	8	1-184	235	26	42	20	9	2012-08-02 04:43:57.735435
242	8	1-193	235	42	66	37	65	2012-08-02 04:43:57.735435
235	8	1-269	235	66	28	84	100	2012-08-02 04:43:57.735435
244	8	1-256	243	81	17	979	27	2012-08-02 04:43:57.735435
245	8	1-17	243	17	1	1183	20	2012-08-02 04:43:57.735435
246	8	1-172	243	1	46	359	15	2012-08-02 04:43:57.735435
247	8	1-243	243	46	42	1716	54	2012-08-02 04:43:57.735435
248	8	1-54	243	42	26	85	67	2012-08-02 04:43:57.735435
249	8	1-184	243	26	42	287	9	2012-08-02 04:43:57.735435
250	8	1-193	243	42	66	646	65	2012-08-02 04:43:57.735435
243	8	1-269	243	66	81	1496	100	2012-08-02 04:43:57.735435
252	5	1-256	251	81	17	3362	27	2012-08-02 04:43:57.735435
253	5	1-17	251	17	12	4829	20	2012-08-02 04:43:57.735435
254	5	1-12	251	12	81	804	12	2012-08-02 04:43:57.735435
255	5	1-157	251	81	66	1438	65	2012-08-02 04:43:57.735435
251	5	1-269	251	66	81	3960	100	2012-08-02 04:43:57.735435
257	5	1-21	256	21	76	4958	31	2012-08-02 04:43:57.834927
258	5	1-252	256	76	20	203	5	2012-08-02 04:43:57.834927
259	5	1-20	256	20	10	1628	30	2012-08-02 04:43:57.834927
260	5	1-247	256	10	88	896	97	2012-08-02 04:43:57.834927
256	5	1-270	256	88	21	1712	32	2012-08-02 04:43:57.834927
262	7	1-212	261	14	14	884	13	2012-08-02 04:43:57.912481
263	7	1-138	261	14	75	1184	53	2012-08-02 04:43:57.912481
264	7	1-216	261	75	88	639	97	2012-08-02 04:43:57.912481
265	7	1-270	261	88	21	639	32	2012-08-02 04:43:57.912481
266	7	1-21	261	21	60	967	31	2012-08-02 04:43:57.912481
267	7	1-104	261	60	2	1935	68	2012-08-02 04:43:57.912481
261	7	1-272	261	2	14	809	19	2012-08-02 04:43:57.912481
269	7	1-212	268	14	14	2850	13	2012-08-02 04:43:57.912481
270	7	1-138	268	14	21	4430	53	2012-08-02 04:43:57.912481
271	7	1-37	268	21	45	2255	52	2012-08-02 04:43:57.912481
272	7	1-58	268	45	9	1623	60	2012-08-02 04:43:57.912481
273	7	1-46	268	9	60	1482	31	2012-08-02 04:43:57.912481
274	7	1-104	268	60	2	3994	68	2012-08-02 04:43:57.912481
268	7	1-272	268	2	14	2249	19	2012-08-02 04:43:57.912481
276	6	1-212	275	14	68	1656	13	2012-08-02 04:43:57.912481
277	6	1-238	275	68	45	1238	52	2012-08-02 04:43:57.912481
278	6	1-58	275	45	9	951	60	2012-08-02 04:43:57.912481
279	6	1-46	275	9	60	926	31	2012-08-02 04:43:57.912481
280	6	1-104	275	60	2	2663	68	2012-08-02 04:43:57.912481
275	6	1-272	275	2	14	1600	19	2012-08-02 04:43:57.912481
282	7	1-83	281	2	14	6802	38	2012-08-02 04:43:58.145536
283	7	1-137	281	14	29	4183	11	2012-08-02 04:43:58.145536
284	7	1-70	281	29	32	3369	16	2012-08-02 04:43:58.145536
285	7	1-198	281	32	26	889	67	2012-08-02 04:43:58.145536
286	7	1-184	281	26	42	1930	9	2012-08-02 04:43:58.145536
287	7	1-193	281	42	76	2350	65	2012-08-02 04:43:58.145536
281	7	1-277	281	76	2	1150	79	2012-08-02 04:43:58.145536
289	7	1-83	288	2	26	1299	38	2012-08-02 04:43:58.145536
290	7	1-26	288	26	19	4205	37	2012-08-02 04:43:58.145536
291	7	1-95	288	19	82	3055	60	2012-08-02 04:43:58.145536
292	7	1-159	288	82	2	289	32	2012-08-02 04:43:58.145536
293	7	1-180	288	2	42	113	9	2012-08-02 04:43:58.145536
294	7	1-193	288	42	76	189	65	2012-08-02 04:43:58.145536
288	7	1-277	288	76	2	127	79	2012-08-02 04:43:58.145536
296	7	1-223	295	57	5	497	91	2012-08-02 04:43:58.145536
297	7	1-221	295	5	38	328	63	2012-08-02 04:43:58.145536
298	7	1-48	295	38	58	1150	62	2012-08-02 04:43:58.145536
299	7	1-93	295	58	26	486	67	2012-08-02 04:43:58.145536
300	7	1-184	295	26	42	1667	9	2012-08-02 04:43:58.145536
301	7	1-193	295	42	76	3208	65	2012-08-02 04:43:58.145536
295	7	1-277	295	76	57	2482	79	2012-08-02 04:43:58.145536
303	8	1-223	302	57	5	452	91	2012-08-02 04:43:58.145536
304	8	1-221	302	5	38	371	63	2012-08-02 04:43:58.145536
305	8	1-48	302	38	58	1611	62	2012-08-02 04:43:58.145536
306	8	1-93	302	58	40	844	67	2012-08-02 04:43:58.145536
307	8	1-92	302	40	13	1174	40	2012-08-02 04:43:58.145536
308	8	1-201	302	13	81	1091	12	2012-08-02 04:43:58.145536
309	8	1-157	302	81	76	1897	65	2012-08-02 04:43:58.145536
302	8	1-277	302	76	57	1821	79	2012-08-02 04:43:58.145536
311	8	1-166	310	85	5	496	91	2012-08-02 04:43:58.278954
312	8	1-221	310	5	38	336	63	2012-08-02 04:43:58.278954
310	8	1-281	310	71	85	352	86	2012-08-02 04:43:58.278954
313	8	1-48	310	38	58	1206	62	2012-08-02 04:43:58.278954
314	8	1-93	310	58	40	522	67	2012-08-02 04:43:58.278954
315	8	1-92	310	40	13	600	40	2012-08-02 04:43:58.278954
316	8	1-201	310	13	33	461	12	2012-08-02 04:43:58.278954
317	8	1-231	310	33	71	291	15	2012-08-02 04:43:58.278954
843	8	1-229	837	52	76	367	79	2012-08-02 04:44:05.480499
319	7	1-166	318	85	39	270	91	2012-08-02 04:43:58.278954
320	7	1-115	318	39	80	244	90	2012-08-02 04:43:58.278954
321	7	1-178	318	80	4	814	41	2012-08-02 04:43:58.278954
322	7	1-28	318	4	13	225	40	2012-08-02 04:43:58.278954
323	7	1-201	318	13	33	189	12	2012-08-02 04:43:58.278954
324	7	1-231	318	33	71	131	15	2012-08-02 04:43:58.278954
318	7	1-281	318	71	85	175	86	2012-08-02 04:43:58.278954
844	8	1-161	837	76	92	2173	31	2012-08-02 04:44:05.480499
326	2	1-225	325	21	34	6015	76	2012-08-02 04:43:58.445391
325	2	1-285	325	34	21	3598	3	2012-08-02 04:43:58.445391
328	6	1-135	327	22	51	1240	83	2012-08-02 04:43:58.478561
329	6	1-89	327	51	70	890	34	2012-08-02 04:43:58.478561
330	6	1-126	327	70	62	572	74	2012-08-02 04:43:58.478561
331	6	1-112	327	62	72	4653	52	2012-08-02 04:43:58.478561
332	6	1-129	327	72	56	1731	4	2012-08-02 04:43:58.478561
327	6	1-286	327	56	22	957	33	2012-08-02 04:43:58.478561
334	7	1-135	333	22	51	58	83	2012-08-02 04:43:58.478561
335	7	1-89	333	51	61	67	34	2012-08-02 04:43:58.478561
336	7	1-106	333	61	13	52	40	2012-08-02 04:43:58.478561
337	7	1-201	333	13	44	43	12	2012-08-02 04:43:58.478561
338	7	1-56	333	44	63	25	22	2012-08-02 04:43:58.478561
339	7	1-113	333	63	56	32	4	2012-08-02 04:43:58.478561
333	7	1-286	333	56	22	28	33	2012-08-02 04:43:58.478561
341	8	1-147	340	77	88	706	97	2012-08-02 04:43:58.478561
342	8	1-270	340	88	21	1309	32	2012-08-02 04:43:58.478561
343	8	1-210	340	21	9	1336	17	2012-08-02 04:43:58.478561
344	8	1-9	340	9	32	614	16	2012-08-02 04:43:58.478561
345	8	1-198	340	32	26	288	67	2012-08-02 04:43:58.478561
346	8	1-184	340	26	5	1110	9	2012-08-02 04:43:58.478561
347	8	1-5	340	5	56	728	4	2012-08-02 04:43:58.478561
340	8	1-286	340	56	77	654	33	2012-08-02 04:43:58.478561
349	5	1-147	348	77	88	2104	97	2012-08-02 04:43:58.478561
350	5	1-270	348	88	2	4530	32	2012-08-02 04:43:58.478561
351	5	1-180	348	2	5	2110	9	2012-08-02 04:43:58.478561
352	5	1-5	348	5	56	1608	4	2012-08-02 04:43:58.478561
348	5	1-286	348	56	77	1678	33	2012-08-02 04:43:58.478561
354	8	1-201	353	13	33	315	12	2012-08-02 04:43:58.589497
355	8	1-231	353	33	8	250	15	2012-08-02 04:43:58.589497
356	8	1-8	353	8	2	417	14	2012-08-02 04:43:58.589497
357	8	1-40	353	2	43	647	55	2012-08-02 04:43:58.589497
358	8	1-55	353	43	80	326	43	2012-08-02 04:43:58.589497
359	8	1-202	353	80	15	1649	24	2012-08-02 04:43:58.589497
360	8	1-15	353	15	92	465	4	2012-08-02 04:43:58.589497
353	8	1-287	353	92	13	327	40	2012-08-02 04:43:58.589497
362	8	1-272	361	2	14	1331	19	2012-08-02 04:43:58.755592
363	8	1-212	361	14	68	2051	13	2012-08-02 04:43:58.755592
364	8	1-238	361	68	72	2282	52	2012-08-02 04:43:58.755592
365	8	1-129	361	72	56	1592	4	2012-08-02 04:43:58.755592
366	8	1-286	361	56	22	1649	33	2012-08-02 04:43:58.755592
367	8	1-135	361	22	82	4008	83	2012-08-02 04:43:58.755592
368	8	1-249	361	82	73	961	84	2012-08-02 04:43:58.755592
361	8	1-293	361	73	2	1489	68	2012-08-02 04:43:58.755592
370	8	1-61	369	17	23	694	61	2012-08-02 04:43:58.855999
371	8	1-146	369	23	17	565	27	2012-08-02 04:43:58.855999
372	8	1-17	369	17	83	810	20	2012-08-02 04:43:58.855999
373	8	1-162	369	83	2	304	14	2012-08-02 04:43:58.855999
374	8	1-40	369	2	43	359	55	2012-08-02 04:43:58.855999
375	8	1-55	369	43	80	164	43	2012-08-02 04:43:58.855999
376	8	1-202	369	80	2	632	24	2012-08-02 04:43:58.855999
369	8	1-299	369	2	17	563	70	2012-08-02 04:43:58.855999
378	7	1-32	377	30	7	263	45	2012-08-02 04:43:58.94642
379	7	1-57	377	7	39	895	56	2012-08-02 04:43:58.94642
380	7	1-49	377	39	2	3963	24	2012-08-02 04:43:58.94642
381	7	1-299	377	2	49	2887	70	2012-08-02 04:43:58.94642
382	7	1-68	377	49	44	1464	12	2012-08-02 04:43:58.94642
383	7	1-56	377	44	36	737	22	2012-08-02 04:43:58.94642
377	7	1-301	377	36	30	913	46	2012-08-02 04:43:58.94642
385	6	1-32	384	30	7	9	45	2012-08-02 04:43:58.94642
386	6	1-57	384	7	41	34	56	2012-08-02 04:43:58.94642
387	6	1-149	384	41	13	45	40	2012-08-02 04:43:58.94642
388	6	1-201	384	13	44	36	12	2012-08-02 04:43:58.94642
389	6	1-56	384	44	36	20	22	2012-08-02 04:43:58.94642
384	6	1-301	384	36	30	28	46	2012-08-02 04:43:58.94642
391	8	1-88	390	6	3	717	82	2012-08-02 04:43:59.168413
392	8	1-190	390	3	14	371	23	2012-08-02 04:43:59.168413
393	8	1-14	390	14	32	529	22	2012-08-02 04:43:59.168413
394	8	1-265	390	32	24	148	34	2012-08-02 04:43:59.168413
395	8	1-24	390	24	13	3463	21	2012-08-02 04:43:59.168413
396	8	1-13	390	13	22	1786	3	2012-08-02 04:43:59.168413
397	8	1-139	390	22	39	343	55	2012-08-02 04:43:59.168413
390	8	1-308	390	39	6	492	53	2012-08-02 04:43:59.168413
399	2	1-86	398	44	39	2069	55	2012-08-02 04:43:59.168413
398	2	1-308	398	39	44	3150	53	2012-08-02 04:43:59.168413
401	7	1-258	400	74	56	183	4	2012-08-02 04:43:59.312683
402	7	1-286	400	56	22	102	33	2012-08-02 04:43:59.312683
403	7	1-135	400	22	51	133	83	2012-08-02 04:43:59.312683
404	7	1-89	400	51	70	96	34	2012-08-02 04:43:59.312683
405	7	1-126	400	70	20	62	74	2012-08-02 04:43:59.312683
406	7	1-300	400	20	11	37	91	2012-08-02 04:43:59.312683
400	7	1-312	400	11	74	21	5	2012-08-02 04:43:59.312683
408	7	1-258	407	74	92	2956	4	2012-08-02 04:43:59.312683
409	7	1-287	407	92	13	1319	40	2012-08-02 04:43:59.312683
410	7	1-201	407	13	33	807	12	2012-08-02 04:43:59.312683
411	7	1-231	407	33	71	406	15	2012-08-02 04:43:59.312683
412	7	1-281	407	71	85	392	86	2012-08-02 04:43:59.312683
413	7	1-166	407	85	11	440	91	2012-08-02 04:43:59.312683
407	7	1-312	407	11	74	295	5	2012-08-02 04:43:59.312683
415	8	1-278	414	44	17	317	70	2012-08-02 04:43:59.312683
416	8	1-61	414	17	37	381	61	2012-08-02 04:43:59.312683
417	8	1-47	414	37	22	973	33	2012-08-02 04:43:59.312683
418	8	1-135	414	22	51	2078	83	2012-08-02 04:43:59.312683
419	8	1-89	414	51	70	2457	34	2012-08-02 04:43:59.312683
420	8	1-126	414	70	20	2602	74	2012-08-02 04:43:59.312683
421	8	1-300	414	20	11	2561	91	2012-08-02 04:43:59.312683
414	8	1-312	414	11	44	2397	5	2012-08-02 04:43:59.312683
837	8	1-422	837	92	18	985	54	2012-08-02 04:44:05.480499
423	8	1-240	422	57	15	2369	24	2012-08-02 04:43:59.312683
424	8	1-15	422	15	92	688	4	2012-08-02 04:43:59.312683
425	8	1-287	422	92	13	498	40	2012-08-02 04:43:59.312683
426	8	1-201	422	13	33	495	12	2012-08-02 04:43:59.312683
427	8	1-231	422	33	71	404	15	2012-08-02 04:43:59.312683
428	8	1-281	422	71	85	633	86	2012-08-02 04:43:59.312683
429	8	1-166	422	85	11	1152	91	2012-08-02 04:43:59.312683
422	8	1-312	422	11	57	1255	5	2012-08-02 04:43:59.312683
15858	2	1-4455	15857	21	47	2525	57	2012-08-02 05:13:08.01419
431	6	1-260	430	6	31	5849	87	2012-08-02 04:43:59.412629
432	6	1-131	430	31	66	445	4	2012-08-02 04:43:59.412629
433	6	1-119	430	66	29	94	44	2012-08-02 04:43:59.412629
434	6	1-31	430	29	66	1402	25	2012-08-02 04:43:59.412629
435	6	1-220	430	66	37	62	51	2012-08-02 04:43:59.412629
430	6	1-314	430	37	6	361	59	2012-08-02 04:43:59.412629
437	6	1-34	436	28	30	564	48	2012-08-02 04:43:59.591311
438	6	1-50	436	30	43	1012	55	2012-08-02 04:43:59.591311
439	6	1-55	436	43	80	316	43	2012-08-02 04:43:59.591311
440	6	1-202	436	80	18	988	24	2012-08-02 04:43:59.591311
441	6	1-315	436	18	87	631	70	2012-08-02 04:43:59.591311
436	6	1-318	436	87	28	295	49	2012-08-02 04:43:59.591311
443	8	1-34	442	28	75	565	48	2012-08-02 04:43:59.591311
444	8	1-251	442	75	71	317	15	2012-08-02 04:43:59.591311
445	8	1-281	442	71	85	365	86	2012-08-02 04:43:59.591311
446	8	1-166	442	85	11	488	91	2012-08-02 04:43:59.591311
447	8	1-312	442	11	57	391	5	2012-08-02 04:43:59.591311
448	8	1-240	442	57	18	542	24	2012-08-02 04:43:59.591311
449	8	1-315	442	18	87	423	70	2012-08-02 04:43:59.591311
442	8	1-318	442	87	28	242	49	2012-08-02 04:43:59.591311
451	7	1-207	450	50	42	137	73	2012-08-02 04:43:59.591311
452	7	1-64	450	42	25	172	36	2012-08-02 04:43:59.591311
453	7	1-25	450	25	58	143	35	2012-08-02 04:43:59.591311
454	7	1-173	450	58	53	724	58	2012-08-02 04:43:59.591311
455	7	1-174	450	53	18	440	24	2012-08-02 04:43:59.591311
456	7	1-315	450	18	87	384	70	2012-08-02 04:43:59.591311
450	7	1-318	450	87	50	245	49	2012-08-02 04:43:59.591311
458	3	1-125	457	69	14	2170	23	2012-08-02 04:43:59.712447
459	3	1-14	457	14	53	1421	22	2012-08-02 04:43:59.712447
457	3	1-323	457	53	69	3583	1	2012-08-02 04:43:59.712447
461	7	1-163	460	22	79	3245	44	2012-08-02 04:43:59.712447
462	7	1-304	460	79	25	1697	19	2012-08-02 04:43:59.712447
463	7	1-316	460	25	74	3172	38	2012-08-02 04:43:59.712447
464	7	1-298	460	74	3	1693	82	2012-08-02 04:43:59.712447
465	7	1-190	460	3	14	545	23	2012-08-02 04:43:59.712447
466	7	1-14	460	14	53	482	22	2012-08-02 04:43:59.712447
460	7	1-323	460	53	22	1641	1	2012-08-02 04:43:59.712447
468	4	1-126	467	70	20	1659	74	2012-08-02 04:43:59.867737
469	4	1-300	467	20	39	1660	91	2012-08-02 04:43:59.867737
470	4	1-115	467	39	77	1546	90	2012-08-02 04:43:59.867737
467	4	1-328	467	77	70	1542	34	2012-08-02 04:43:59.867737
472	8	1-170	471	28	2	364	24	2012-08-02 04:43:59.967494
473	8	1-299	471	2	17	217	70	2012-08-02 04:43:59.967494
474	8	1-61	471	17	26	178	61	2012-08-02 04:43:59.967494
475	8	1-101	471	26	76	3479	8	2012-08-02 04:43:59.967494
476	8	1-275	471	76	5	1889	9	2012-08-02 04:43:59.967494
477	8	1-5	471	5	66	860	4	2012-08-02 04:43:59.967494
478	8	1-119	471	66	41	107	44	2012-08-02 04:43:59.967494
471	8	1-331	471	41	28	18	100	2012-08-02 04:43:59.967494
480	7	1-170	479	28	18	2138	24	2012-08-02 04:43:59.967494
481	7	1-315	479	18	87	1425	70	2012-08-02 04:43:59.967494
482	7	1-318	479	87	28	696	49	2012-08-02 04:43:59.967494
483	7	1-34	479	28	75	1795	48	2012-08-02 04:43:59.967494
484	7	1-251	479	75	46	860	15	2012-08-02 04:43:59.967494
485	7	1-313	479	46	41	426	44	2012-08-02 04:43:59.967494
479	7	1-331	479	41	28	77	100	2012-08-02 04:43:59.967494
487	8	1-19	486	19	29	122	29	2012-08-02 04:44:00.144891
488	8	1-78	486	29	68	78	72	2012-08-02 04:44:00.144891
489	8	1-297	486	68	57	12	30	2012-08-02 04:44:00.144891
490	8	1-200	486	57	37	7	51	2012-08-02 04:44:00.144891
491	8	1-314	486	37	6	27	59	2012-08-02 04:44:00.144891
492	8	1-260	486	6	71	296	87	2012-08-02 04:44:00.144891
493	8	1-253	486	71	84	317	90	2012-08-02 04:44:00.144891
486	8	1-334	486	84	19	156	14	2012-08-02 04:44:00.144891
495	4	1-143	494	21	16	2295	26	2012-08-02 04:44:00.233744
496	4	1-16	494	16	15	1800	25	2012-08-02 04:44:00.233744
497	4	1-206	494	15	1	1615	6	2012-08-02 04:44:00.233744
494	4	1-335	494	1	21	4686	96	2012-08-02 04:44:00.233744
499	8	1-293	498	73	2	658	68	2012-08-02 04:44:00.289158
500	8	1-272	498	2	53	602	19	2012-08-02 04:44:00.289158
501	8	1-77	498	53	14	322	23	2012-08-02 04:44:00.289158
502	8	1-14	498	14	32	476	22	2012-08-02 04:44:00.289158
503	8	1-265	498	32	24	137	34	2012-08-02 04:44:00.289158
504	8	1-24	498	24	13	3343	21	2012-08-02 04:44:00.289158
505	8	1-13	498	13	26	1788	3	2012-08-02 04:44:00.289158
498	8	1-337	498	26	73	414	84	2012-08-02 04:44:00.289158
507	3	1-291	506	93	43	489	89	2012-08-02 04:44:00.344457
508	3	1-205	506	43	96	190	29	2012-08-02 04:44:00.344457
506	3	1-338	506	96	93	2105	88	2012-08-02 04:44:00.344457
510	4	1-1	509	1	22	877	1	2012-08-02 04:44:00.455436
511	4	1-163	509	22	51	1610	44	2012-08-02 04:44:00.455436
512	4	1-85	509	51	26	1028	80	2012-08-02 04:44:00.455436
509	4	1-342	509	26	1	918	2	2012-08-02 04:44:00.455436
514	6	1-215	513	90	39	302	91	2012-08-02 04:44:00.455436
515	6	1-115	513	39	84	316	90	2012-08-02 04:44:00.455436
516	6	1-334	513	84	19	240	14	2012-08-02 04:44:00.455436
517	6	1-19	513	19	4	289	29	2012-08-02 04:44:00.455436
518	6	1-227	513	4	26	75	80	2012-08-02 04:44:00.455436
513	6	1-342	513	26	90	118	2	2012-08-02 04:44:00.455436
520	6	1-131	519	31	66	80	4	2012-08-02 04:44:00.754537
519	6	1-350	519	56	31	1245	87	2012-08-02 04:44:00.754537
521	6	1-119	519	66	11	13	44	2012-08-02 04:44:00.754537
522	6	1-325	519	11	79	17	52	2012-08-02 04:44:00.754537
523	6	1-340	519	79	9	408	6	2012-08-02 04:44:00.754537
524	6	1-292	519	9	56	479	57	2012-08-02 04:44:00.754537
846	7	1-406	845	18	85	560	28	2012-08-02 04:44:05.480499
526	8	1-240	525	57	18	597	24	2012-08-02 04:44:00.821366
527	8	1-315	525	18	17	542	70	2012-08-02 04:44:00.821366
528	8	1-61	525	17	23	717	61	2012-08-02 04:44:00.821366
529	8	1-146	525	23	17	675	27	2012-08-02 04:44:00.821366
530	8	1-17	525	17	83	1040	20	2012-08-02 04:44:00.821366
531	8	1-162	525	83	19	450	14	2012-08-02 04:44:00.821366
532	8	1-19	525	19	47	472	29	2012-08-02 04:44:00.821366
525	8	1-351	525	47	57	370	5	2012-08-02 04:44:00.821366
847	7	1-336	845	85	1	403	20	2012-08-02 04:44:05.480499
534	7	1-49	533	39	2	1694	24	2012-08-02 04:44:00.865562
535	7	1-299	533	2	17	1336	70	2012-08-02 04:44:00.865562
536	7	1-61	533	17	3	1081	61	2012-08-02 04:44:00.865562
537	7	1-59	533	3	2	336	68	2012-08-02 04:44:00.865562
538	7	1-272	533	2	53	239	19	2012-08-02 04:44:00.865562
539	7	1-77	533	53	96	74	23	2012-08-02 04:44:00.865562
533	7	1-352	533	96	39	476	56	2012-08-02 04:44:00.865562
541	7	1-49	540	39	18	148	24	2012-08-02 04:44:00.865562
542	7	1-315	540	18	17	98	70	2012-08-02 04:44:00.865562
543	7	1-61	540	17	3	86	61	2012-08-02 04:44:00.865562
544	7	1-59	540	3	2	29	68	2012-08-02 04:44:00.865562
545	7	1-272	540	2	53	17	19	2012-08-02 04:44:00.865562
546	7	1-77	540	53	96	6	23	2012-08-02 04:44:00.865562
540	7	1-352	540	96	39	38	56	2012-08-02 04:44:00.865562
548	5	1-184	547	26	85	2339	9	2012-08-02 04:44:00.965579
549	5	1-268	547	85	3	189	82	2012-08-02 04:44:00.965579
550	5	1-190	547	3	96	81	23	2012-08-02 04:44:00.965579
551	5	1-352	547	96	47	711	56	2012-08-02 04:44:00.965579
547	5	1-353	547	47	26	649	67	2012-08-02 04:44:00.965579
553	7	1-134	552	70	46	41	58	2012-08-02 04:44:01.031584
554	7	1-191	552	46	20	56	74	2012-08-02 04:44:01.031584
555	7	1-300	552	20	39	62	91	2012-08-02 04:44:01.031584
556	7	1-115	552	39	84	63	90	2012-08-02 04:44:01.031584
557	7	1-334	552	84	19	47	14	2012-08-02 04:44:01.031584
558	7	1-19	552	19	30	55	29	2012-08-02 04:44:01.031584
552	7	1-354	552	30	70	39	85	2012-08-02 04:44:01.031584
560	8	1-167	559	58	65	1765	37	2012-08-02 04:44:01.031584
561	8	1-329	559	65	67	2601	46	2012-08-02 04:44:01.031584
562	8	1-154	559	67	85	1096	86	2012-08-02 04:44:01.031584
563	8	1-166	559	85	39	1960	91	2012-08-02 04:44:01.031584
564	8	1-115	559	39	84	2051	90	2012-08-02 04:44:01.031584
565	8	1-334	559	84	19	1562	14	2012-08-02 04:44:01.031584
566	8	1-19	559	19	30	1880	29	2012-08-02 04:44:01.031584
559	8	1-354	559	30	58	1387	85	2012-08-02 04:44:01.031584
568	8	1-19	567	19	30	284	29	2012-08-02 04:44:01.153643
569	8	1-354	567	30	58	214	85	2012-08-02 04:44:01.153643
570	8	1-167	567	58	65	278	37	2012-08-02 04:44:01.153643
571	8	1-329	567	65	67	418	46	2012-08-02 04:44:01.153643
572	8	1-154	567	67	85	180	86	2012-08-02 04:44:01.153643
573	8	1-166	567	85	5	328	91	2012-08-02 04:44:01.153643
574	8	1-221	567	5	49	288	63	2012-08-02 04:44:01.153643
567	8	1-357	567	49	19	231	14	2012-08-02 04:44:01.153643
576	8	1-167	575	58	97	1359	37	2012-08-02 04:44:01.386358
577	8	1-343	575	97	21	259	53	2012-08-02 04:44:01.386358
578	8	1-37	575	21	79	129	52	2012-08-02 04:44:01.386358
579	8	1-340	575	79	9	2380	6	2012-08-02 04:44:01.386358
580	8	1-292	575	9	56	2108	57	2012-08-02 04:44:01.386358
581	8	1-350	575	56	71	4134	87	2012-08-02 04:44:01.386358
582	8	1-253	575	71	36	4204	90	2012-08-02 04:44:01.386358
575	8	1-362	575	36	58	1742	85	2012-08-02 04:44:01.386358
584	8	1-318	583	87	28	644	49	2012-08-02 04:44:01.554638
585	8	1-34	583	28	30	1659	48	2012-08-02 04:44:01.554638
586	8	1-50	583	30	4	4008	55	2012-08-02 04:44:01.554638
587	8	1-151	583	4	36	1493	64	2012-08-02 04:44:01.554638
588	8	1-263	583	36	2	649	68	2012-08-02 04:44:01.554638
589	8	1-272	583	2	25	481	19	2012-08-02 04:44:01.554638
590	8	1-316	583	25	11	1215	38	2012-08-02 04:44:01.554638
583	8	1-367	583	11	87	1022	70	2012-08-02 04:44:01.554638
592	7	1-318	591	87	28	1186	49	2012-08-02 04:44:01.554638
593	7	1-34	591	28	75	3072	48	2012-08-02 04:44:01.554638
594	7	1-251	591	75	46	1911	15	2012-08-02 04:44:01.554638
595	7	1-313	591	46	79	1228	44	2012-08-02 04:44:01.554638
596	7	1-304	591	79	25	872	19	2012-08-02 04:44:01.554638
597	7	1-316	591	25	11	2214	38	2012-08-02 04:44:01.554638
591	7	1-367	591	11	87	1873	70	2012-08-02 04:44:01.554638
599	2	1-35	598	31	68	3360	17	2012-08-02 04:44:01.698628
598	2	1-370	598	68	31	2545	50	2012-08-02 04:44:01.698628
601	7	1-144	600	12	65	733	37	2012-08-02 04:44:01.787236
602	7	1-329	600	65	30	836	46	2012-08-02 04:44:01.787236
603	7	1-32	600	30	7	255	45	2012-08-02 04:44:01.787236
604	7	1-57	600	7	41	916	56	2012-08-02 04:44:01.787236
605	7	1-149	600	41	75	1169	40	2012-08-02 04:44:01.787236
606	7	1-140	600	75	2	2341	95	2012-08-02 04:44:01.787236
600	7	1-371	600	2	12	853	20	2012-08-02 04:44:01.787236
608	8	1-341	607	5	59	119	10	2012-08-02 04:44:01.942108
609	8	1-369	607	59	75	68	48	2012-08-02 04:44:01.942108
610	8	1-251	607	75	38	34	15	2012-08-02 04:44:01.942108
611	8	1-282	607	38	15	115	25	2012-08-02 04:44:01.942108
612	8	1-206	607	15	9	71	6	2012-08-02 04:44:01.942108
613	8	1-292	607	9	56	68	57	2012-08-02 04:44:01.942108
614	8	1-350	607	56	23	144	87	2012-08-02 04:44:01.942108
607	8	1-373	607	23	5	198	18	2012-08-02 04:44:01.942108
616	7	1-370	615	68	67	1701	50	2012-08-02 04:44:02.01968
617	7	1-120	615	67	25	404	36	2012-08-02 04:44:02.01968
618	7	1-25	615	25	29	332	35	2012-08-02 04:44:02.01968
619	7	1-324	615	29	73	1576	84	2012-08-02 04:44:02.01968
620	7	1-293	615	73	59	2041	68	2012-08-02 04:44:02.01968
621	7	1-274	615	59	99	1409	51	2012-08-02 04:44:02.01968
615	7	1-374	615	99	68	518	17	2012-08-02 04:44:02.01968
623	6	1-370	622	68	29	95	50	2012-08-02 04:44:02.01968
624	6	1-248	622	29	7	15	45	2012-08-02 04:44:02.01968
625	6	1-57	622	7	14	60	56	2012-08-02 04:44:02.01968
626	6	1-41	622	14	57	83	30	2012-08-02 04:44:02.01968
627	6	1-200	622	57	99	68	51	2012-08-02 04:44:02.01968
622	6	1-374	622	99	68	27	17	2012-08-02 04:44:02.01968
848	7	1-172	845	1	71	145	15	2012-08-02 04:44:05.480499
629	3	1-319	628	56	57	8220	30	2012-08-02 04:44:02.01968
630	3	1-200	628	57	99	6845	51	2012-08-02 04:44:02.01968
628	3	1-374	628	99	56	2748	17	2012-08-02 04:44:02.01968
632	7	1-370	631	68	67	914	50	2012-08-02 04:44:02.01968
633	7	1-120	631	67	17	242	36	2012-08-02 04:44:02.01968
634	7	1-306	631	17	96	49	23	2012-08-02 04:44:02.01968
635	7	1-352	631	96	14	501	56	2012-08-02 04:44:02.01968
636	7	1-41	631	14	57	715	30	2012-08-02 04:44:02.01968
637	7	1-200	631	57	99	608	51	2012-08-02 04:44:02.01968
631	7	1-374	631	99	68	250	17	2012-08-02 04:44:02.01968
639	8	1-26	638	26	65	979	37	2012-08-02 04:44:02.285657
640	8	1-329	638	65	30	807	46	2012-08-02 04:44:02.285657
641	8	1-32	638	30	82	177	45	2012-08-02 04:44:02.285657
642	8	1-164	638	82	68	203	13	2012-08-02 04:44:02.285657
643	8	1-238	638	68	79	127	52	2012-08-02 04:44:02.285657
644	8	1-340	638	79	9	2140	6	2012-08-02 04:44:02.285657
645	8	1-292	638	9	27	1729	57	2012-08-02 04:44:02.285657
638	8	1-378	638	27	26	360	38	2012-08-02 04:44:02.285657
647	8	1-367	646	11	87	144	70	2012-08-02 04:44:02.285657
648	8	1-318	646	87	28	91	49	2012-08-02 04:44:02.285657
649	8	1-34	646	28	75	234	48	2012-08-02 04:44:02.285657
650	8	1-251	646	75	38	145	15	2012-08-02 04:44:02.285657
651	8	1-282	646	38	15	613	25	2012-08-02 04:44:02.285657
652	8	1-206	646	15	9	468	6	2012-08-02 04:44:02.285657
653	8	1-292	646	9	27	557	57	2012-08-02 04:44:02.285657
646	8	1-378	646	27	11	171	38	2012-08-02 04:44:02.285657
655	8	1-26	654	26	65	2421	37	2012-08-02 04:44:02.285657
656	8	1-329	654	65	9	3075	46	2012-08-02 04:44:02.285657
657	8	1-63	654	9	57	2096	30	2012-08-02 04:44:02.285657
658	8	1-200	654	57	99	1656	51	2012-08-02 04:44:02.285657
659	8	1-374	654	99	19	631	17	2012-08-02 04:44:02.285657
660	8	1-108	654	19	9	1296	6	2012-08-02 04:44:02.285657
661	8	1-292	654	9	27	1797	57	2012-08-02 04:44:02.285657
654	8	1-378	654	27	26	578	38	2012-08-02 04:44:02.285657
663	8	1-298	662	74	3	373	82	2012-08-02 04:44:02.285657
664	8	1-190	662	3	96	186	23	2012-08-02 04:44:02.285657
665	8	1-352	662	96	47	1930	56	2012-08-02 04:44:02.285657
666	8	1-353	662	47	52	2068	67	2012-08-02 04:44:02.285657
667	8	1-73	662	52	15	1060	25	2012-08-02 04:44:02.285657
668	8	1-206	662	15	9	929	6	2012-08-02 04:44:02.285657
669	8	1-292	662	9	27	1273	57	2012-08-02 04:44:02.285657
662	8	1-378	662	27	74	450	38	2012-08-02 04:44:02.285657
671	8	1-26	670	26	97	1141	37	2012-08-02 04:44:02.285657
672	8	1-343	670	97	44	360	53	2012-08-02 04:44:02.285657
673	8	1-86	670	44	43	230	55	2012-08-02 04:44:02.285657
674	8	1-55	670	43	84	119	43	2012-08-02 04:44:02.285657
675	8	1-375	670	84	19	172	17	2012-08-02 04:44:02.285657
676	8	1-108	670	19	9	416	6	2012-08-02 04:44:02.285657
677	8	1-292	670	9	27	610	57	2012-08-02 04:44:02.285657
670	8	1-378	670	27	26	231	38	2012-08-02 04:44:02.285657
679	8	1-163	678	22	79	1576	44	2012-08-02 04:44:02.485329
680	8	1-304	678	79	53	841	19	2012-08-02 04:44:02.485329
681	8	1-77	678	53	96	275	23	2012-08-02 04:44:02.485329
682	8	1-352	678	96	14	1872	56	2012-08-02 04:44:02.485329
683	8	1-41	678	14	72	1795	30	2012-08-02 04:44:02.485329
684	8	1-171	678	72	78	543	47	2012-08-02 04:44:02.485329
685	8	1-150	678	78	64	290	78	2012-08-02 04:44:02.485329
678	8	1-382	678	64	22	780	1	2012-08-02 04:44:02.485329
687	5	1-163	686	22	79	1887	44	2012-08-02 04:44:02.485329
688	5	1-304	686	79	7	1098	19	2012-08-02 04:44:02.485329
689	5	1-33	686	7	78	502	47	2012-08-02 04:44:02.485329
690	5	1-150	686	78	64	292	78	2012-08-02 04:44:02.485329
686	5	1-382	686	64	22	858	1	2012-08-02 04:44:02.485329
692	8	1-182	691	9	2	68	32	2012-08-02 04:44:02.485329
693	8	1-180	691	2	44	31	9	2012-08-02 04:44:02.485329
694	8	1-333	691	44	84	75	43	2012-08-02 04:44:02.485329
695	8	1-375	691	84	59	106	17	2012-08-02 04:44:02.485329
696	8	1-124	691	59	6	63	11	2012-08-02 04:44:02.485329
697	8	1-6	691	6	43	41	10	2012-08-02 04:44:02.485329
698	8	1-271	691	43	64	108	78	2012-08-02 04:44:02.485329
691	8	1-382	691	64	9	461	1	2012-08-02 04:44:02.485329
700	3	1-199	699	60	47	9875	55	2012-08-02 04:44:02.728964
701	3	1-75	699	47	93	299	48	2012-08-02 04:44:02.728964
699	3	1-384	699	93	60	64	75	2012-08-02 04:44:02.728964
703	8	1-290	702	46	76	501	8	2012-08-02 04:44:02.728964
704	8	1-275	702	76	44	397	9	2012-08-02 04:44:02.728964
705	8	1-333	702	44	84	851	43	2012-08-02 04:44:02.728964
706	8	1-375	702	84	59	1055	17	2012-08-02 04:44:02.728964
707	8	1-124	702	59	6	612	11	2012-08-02 04:44:02.728964
708	8	1-6	702	6	59	356	10	2012-08-02 04:44:02.728964
709	8	1-369	702	59	93	294	48	2012-08-02 04:44:02.728964
702	8	1-384	702	93	46	296	75	2012-08-02 04:44:02.728964
711	8	1-365	710	89	50	123	77	2012-08-02 04:44:02.728964
712	8	1-128	710	50	93	201	88	2012-08-02 04:44:02.728964
713	8	1-291	710	93	41	200	89	2012-08-02 04:44:02.728964
714	8	1-224	710	41	86	3393	72	2012-08-02 04:44:02.728964
715	8	1-179	710	86	44	1013	53	2012-08-02 04:44:02.728964
716	8	1-86	710	44	47	610	55	2012-08-02 04:44:02.728964
717	8	1-75	710	47	93	96	48	2012-08-02 04:44:02.728964
710	8	1-384	710	93	89	112	75	2012-08-02 04:44:02.728964
719	6	1-301	718	36	30	362	46	2012-08-02 04:44:02.884051
720	6	1-32	718	30	82	141	45	2012-08-02 04:44:02.884051
721	6	1-164	718	82	68	285	13	2012-08-02 04:44:02.884051
722	6	1-238	718	68	72	315	52	2012-08-02 04:44:02.884051
723	6	1-129	718	72	85	218	4	2012-08-02 04:44:02.884051
718	6	1-387	718	85	36	217	22	2012-08-02 04:44:02.884051
725	8	1-368	724	88	5	2288	18	2012-08-02 04:44:03.016855
726	8	1-341	724	5	59	1946	10	2012-08-02 04:44:03.016855
727	8	1-369	724	59	93	1554	48	2012-08-02 04:44:03.016855
728	8	1-384	724	93	89	1674	75	2012-08-02 04:44:03.016855
729	8	1-365	724	89	22	1923	77	2012-08-02 04:44:03.016855
724	8	1-388	724	11	88	9131	53	2012-08-02 04:44:03.016855
730	8	1-322	724	22	89	731	26	2012-08-02 04:44:03.016855
731	8	1-213	724	89	11	3985	34	2012-08-02 04:44:03.016855
849	7	1-281	845	71	52	200	86	2012-08-02 04:44:05.480499
733	7	1-164	732	82	95	2426	13	2012-08-02 04:44:03.42941
734	7	1-326	732	95	59	1078	10	2012-08-02 04:44:03.42941
735	7	1-369	732	59	93	880	48	2012-08-02 04:44:03.42941
736	7	1-384	732	93	46	969	75	2012-08-02 04:44:03.42941
737	7	1-290	732	46	35	1796	8	2012-08-02 04:44:03.42941
738	7	1-43	732	35	9	1486	40	2012-08-02 04:44:03.42941
732	7	1-395	732	9	82	1237	45	2012-08-02 04:44:03.42941
850	7	1-229	845	52	76	155	79	2012-08-02 04:44:05.480499
740	4	1-251	739	75	71	2470	15	2012-08-02 04:44:03.636911
741	4	1-281	739	71	85	3133	86	2012-08-02 04:44:03.636911
742	4	1-166	739	85	42	4624	91	2012-08-02 04:44:03.636911
739	4	1-399	739	42	75	3991	48	2012-08-02 04:44:03.636911
851	7	1-161	845	76	92	918	31	2012-08-02 04:44:05.480499
744	3	1-315	743	18	17	2380	70	2012-08-02 04:44:03.991125
745	3	1-61	743	17	43	2197	61	2012-08-02 04:44:03.991125
743	3	1-405	743	43	18	3392	24	2012-08-02 04:44:03.991125
845	7	1-422	845	92	18	416	54	2012-08-02 04:44:05.480499
747	4	1-336	746	85	1	3282	20	2012-08-02 04:44:04.013213
748	4	1-172	746	1	46	996	15	2012-08-02 04:44:04.013213
749	4	1-243	746	46	18	4764	54	2012-08-02 04:44:04.013213
746	4	1-406	746	18	85	5407	28	2012-08-02 04:44:04.013213
751	5	1-5	750	5	85	1493	4	2012-08-02 04:44:04.25691
752	5	1-387	750	85	36	1514	22	2012-08-02 04:44:04.25691
753	5	1-301	750	36	9	2577	46	2012-08-02 04:44:04.25691
754	5	1-63	750	9	71	1843	30	2012-08-02 04:44:04.25691
750	5	1-410	750	71	5	1951	9	2012-08-02 04:44:04.25691
756	8	1-384	755	93	89	994	75	2012-08-02 04:44:04.511423
757	8	1-365	755	89	50	643	77	2012-08-02 04:44:04.511423
758	8	1-128	755	50	93	620	88	2012-08-02 04:44:04.511423
759	8	1-291	755	93	41	473	89	2012-08-02 04:44:04.511423
760	8	1-224	755	41	86	4722	72	2012-08-02 04:44:04.511423
761	8	1-179	755	86	44	830	53	2012-08-02 04:44:04.511423
762	8	1-86	755	44	27	294	55	2012-08-02 04:44:04.511423
755	8	1-412	755	27	93	1113	48	2012-08-02 04:44:04.511423
764	6	1-384	763	93	1	1171	75	2012-08-02 04:44:04.511423
765	6	1-76	763	1	18	1500	24	2012-08-02 04:44:04.511423
766	6	1-315	763	18	17	1041	70	2012-08-02 04:44:04.511423
767	6	1-61	763	17	72	951	61	2012-08-02 04:44:04.511423
768	6	1-364	763	72	27	343	55	2012-08-02 04:44:04.511423
763	6	1-412	763	27	93	1529	48	2012-08-02 04:44:04.511423
770	6	1-384	769	93	89	989	75	2012-08-02 04:44:04.511423
771	6	1-365	769	89	22	1050	77	2012-08-02 04:44:04.511423
772	6	1-322	769	22	94	394	26	2012-08-02 04:44:04.511423
773	6	1-311	769	94	44	258	53	2012-08-02 04:44:04.511423
774	6	1-86	769	44	27	150	55	2012-08-02 04:44:04.511423
769	6	1-412	769	27	93	930	48	2012-08-02 04:44:04.511423
776	7	1-251	775	75	46	597	15	2012-08-02 04:44:04.511423
777	7	1-313	775	46	79	468	44	2012-08-02 04:44:04.511423
778	7	1-304	775	79	53	406	19	2012-08-02 04:44:04.511423
779	7	1-77	775	53	96	215	23	2012-08-02 04:44:04.511423
780	7	1-352	775	96	91	2379	56	2012-08-02 04:44:04.511423
781	7	1-218	775	91	27	116	55	2012-08-02 04:44:04.511423
775	7	1-412	775	27	75	787	48	2012-08-02 04:44:04.511423
783	7	1-91	782	57	73	5765	84	2012-08-02 04:44:04.688933
784	7	1-293	782	73	43	5125	68	2012-08-02 04:44:04.688933
785	7	1-360	782	43	30	5227	46	2012-08-02 04:44:04.688933
786	7	1-32	782	30	82	1173	45	2012-08-02 04:44:04.688933
787	7	1-164	782	82	68	1373	13	2012-08-02 04:44:04.688933
788	7	1-238	782	68	25	876	52	2012-08-02 04:44:04.688933
782	7	1-415	782	25	57	9461	11	2012-08-02 04:44:04.688933
790	7	1-70	789	29	39	381	16	2012-08-02 04:44:04.688933
791	7	1-148	789	39	71	144	87	2012-08-02 04:44:04.688933
792	7	1-253	789	71	99	137	90	2012-08-02 04:44:04.688933
793	7	1-392	789	99	47	102	62	2012-08-02 04:44:04.688933
794	7	1-66	789	47	11	42	44	2012-08-02 04:44:04.688933
795	7	1-325	789	11	25	37	52	2012-08-02 04:44:04.688933
789	7	1-415	789	25	29	402	11	2012-08-02 04:44:04.688933
797	8	1-367	796	11	17	185	70	2012-08-02 04:44:04.882974
798	8	1-61	796	17	37	131	61	2012-08-02 04:44:04.882974
799	8	1-47	796	37	45	199	33	2012-08-02 04:44:04.882974
800	8	1-407	796	45	57	486	2	2012-08-02 04:44:04.882974
801	8	1-121	796	57	91	5126	56	2012-08-02 04:44:04.882974
802	8	1-218	796	91	48	128	55	2012-08-02 04:44:04.882974
803	8	1-67	796	48	92	119	51	2012-08-02 04:44:04.882974
796	8	1-418	796	92	11	349	38	2012-08-02 04:44:04.882974
805	7	1-367	804	11	17	186	70	2012-08-02 04:44:04.882974
806	7	1-61	804	17	37	139	61	2012-08-02 04:44:04.882974
807	7	1-47	804	37	45	220	33	2012-08-02 04:44:04.882974
808	7	1-407	804	45	6	562	2	2012-08-02 04:44:04.882974
809	7	1-280	804	6	57	221	30	2012-08-02 04:44:04.882974
810	7	1-200	804	57	92	110	51	2012-08-02 04:44:04.882974
804	7	1-418	804	92	11	337	38	2012-08-02 04:44:04.882974
812	5	1-26	811	26	97	865	37	2012-08-02 04:44:04.882974
813	5	1-343	811	97	44	171	53	2012-08-02 04:44:04.882974
814	5	1-86	811	44	48	68	55	2012-08-02 04:44:04.882974
815	5	1-67	811	48	92	78	51	2012-08-02 04:44:04.882974
811	5	1-418	811	92	26	280	38	2012-08-02 04:44:04.882974
817	5	1-367	816	11	17	1485	70	2012-08-02 04:44:04.882974
818	5	1-61	816	17	72	1356	61	2012-08-02 04:44:04.882974
819	5	1-364	816	72	48	488	55	2012-08-02 04:44:04.882974
820	5	1-67	816	48	92	583	51	2012-08-02 04:44:04.882974
816	5	1-418	816	92	11	2189	38	2012-08-02 04:44:04.882974
822	5	1-298	821	74	11	1526	82	2012-08-02 04:44:04.882974
823	5	1-391	821	11	22	808	19	2012-08-02 04:44:04.882974
824	5	1-295	821	22	59	390	68	2012-08-02 04:44:04.882974
825	5	1-274	821	59	92	319	51	2012-08-02 04:44:04.882974
821	5	1-418	821	92	74	1772	38	2012-08-02 04:44:04.882974
826	1	1-61	\N	17	17	1286	61	2012-08-02 04:44:04.882974
828	7	1-371	827	2	1	1684	20	2012-08-02 04:44:05.225624
829	7	1-172	827	1	23	600	15	2012-08-02 04:44:05.225624
830	7	1-183	827	23	9	196	17	2012-08-02 04:44:05.225624
831	7	1-9	827	9	32	93	16	2012-08-02 04:44:05.225624
827	7	1-420	827	4	2	4032	95	2012-08-02 04:44:05.225624
832	7	1-198	827	32	6	45	67	2012-08-02 04:44:05.225624
833	7	1-411	827	6	4	108	34	2012-08-02 04:44:05.225624
15857	2	1-4566	15857	47	21	9738	97	2012-08-02 05:13:08.01419
835	3	1-190	834	3	78	241	23	2012-08-02 04:44:05.392042
836	3	1-419	834	78	71	169	50	2012-08-02 04:44:05.392042
834	3	1-421	834	71	3	702	82	2012-08-02 04:44:05.392042
853	8	1-406	852	18	49	2018	28	2012-08-02 04:44:05.480499
854	8	1-72	852	49	22	640	8	2012-08-02 04:44:05.480499
855	8	1-22	852	22	45	1371	33	2012-08-02 04:44:05.480499
856	8	1-407	852	45	6	6196	2	2012-08-02 04:44:05.480499
857	8	1-280	852	6	10	4306	30	2012-08-02 04:44:05.480499
858	8	1-247	852	10	72	2552	97	2012-08-02 04:44:05.480499
859	8	1-347	852	72	92	2891	31	2012-08-02 04:44:05.480499
852	8	1-422	852	92	18	1402	54	2012-08-02 04:44:05.480499
861	8	1-406	860	18	30	460	28	2012-08-02 04:44:05.480499
862	8	1-381	860	30	43	230	80	2012-08-02 04:44:05.480499
863	8	1-408	860	43	96	305	26	2012-08-02 04:44:05.480499
864	8	1-404	860	96	20	101	5	2012-08-02 04:44:05.480499
865	8	1-20	860	20	10	895	30	2012-08-02 04:44:05.480499
866	8	1-247	860	10	72	543	97	2012-08-02 04:44:05.480499
867	8	1-347	860	72	92	629	31	2012-08-02 04:44:05.480499
860	8	1-422	860	92	18	312	54	2012-08-02 04:44:05.480499
869	8	1-74	868	42	96	181	76	2012-08-02 04:44:05.735273
870	8	1-361	868	96	37	141	61	2012-08-02 04:44:05.735273
871	8	1-47	868	37	45	264	33	2012-08-02 04:44:05.735273
872	8	1-407	868	45	1	798	2	2012-08-02 04:44:05.735273
873	8	1-1	868	1	1	1114	1	2012-08-02 04:44:05.735273
874	8	1-390	868	1	53	753	19	2012-08-02 04:44:05.735273
875	8	1-77	868	53	31	253	23	2012-08-02 04:44:05.735273
868	8	1-423	868	31	42	338	71	2012-08-02 04:44:05.735273
877	7	1-345	876	50	47	661	29	2012-08-02 04:44:05.735273
878	7	1-351	876	47	20	598	5	2012-08-02 04:44:05.735273
879	7	1-20	876	20	71	5401	30	2012-08-02 04:44:05.735273
880	7	1-410	876	71	85	5695	9	2012-08-02 04:44:05.735273
881	7	1-268	876	85	3	573	82	2012-08-02 04:44:05.735273
882	7	1-190	876	3	31	303	23	2012-08-02 04:44:05.735273
876	7	1-423	876	31	50	633	71	2012-08-02 04:44:05.735273
884	8	1-177	883	41	72	247	61	2012-08-02 04:44:05.735273
885	8	1-364	883	72	27	141	55	2012-08-02 04:44:05.735273
886	8	1-412	883	27	75	1002	48	2012-08-02 04:44:05.735273
887	8	1-251	883	75	46	795	15	2012-08-02 04:44:05.735273
888	8	1-313	883	46	79	652	44	2012-08-02 04:44:05.735273
889	8	1-304	883	79	53	591	19	2012-08-02 04:44:05.735273
890	8	1-77	883	53	31	328	23	2012-08-02 04:44:05.735273
883	8	1-423	883	31	41	722	71	2012-08-02 04:44:05.735273
892	7	1-345	891	50	4	1561	29	2012-08-02 04:44:05.735273
893	7	1-227	891	4	26	431	80	2012-08-02 04:44:05.735273
894	7	1-342	891	26	1	724	2	2012-08-02 04:44:05.735273
895	7	1-1	891	1	1	1303	1	2012-08-02 04:44:05.735273
896	7	1-390	891	1	53	1136	19	2012-08-02 04:44:05.735273
897	7	1-77	891	53	31	635	23	2012-08-02 04:44:05.735273
891	7	1-423	891	31	50	1407	71	2012-08-02 04:44:05.735273
898	1	1-77	\N	53	53	1345	23	2012-08-02 04:44:05.735273
900	6	1-88	899	6	3	116	82	2012-08-02 04:44:05.968063
901	6	1-190	899	3	31	63	23	2012-08-02 04:44:05.968063
902	6	1-423	899	31	50	134	71	2012-08-02 04:44:05.968063
903	6	1-345	899	50	4	143	29	2012-08-02 04:44:05.968063
904	6	1-227	899	4	76	38	80	2012-08-02 04:44:05.968063
899	6	1-427	899	76	6	77	53	2012-08-02 04:44:05.968063
906	6	1-148	905	39	71	1002	87	2012-08-02 04:44:06.345364
907	6	1-253	905	71	99	1569	90	2012-08-02 04:44:06.345364
908	6	1-392	905	99	58	1928	62	2012-08-02 04:44:06.345364
909	6	1-93	905	58	6	1000	67	2012-08-02 04:44:06.345364
910	6	1-411	905	6	19	2573	34	2012-08-02 04:44:06.345364
905	6	1-443	905	19	39	1602	16	2012-08-02 04:44:06.345364
912	3	1-198	911	32	6	427	67	2012-08-02 04:44:06.345364
913	3	1-411	911	6	19	1173	34	2012-08-02 04:44:06.345364
911	3	1-443	911	19	32	780	16	2012-08-02 04:44:06.345364
915	6	1-343	914	97	6	1349	53	2012-08-02 04:44:06.522533
916	6	1-88	914	6	3	1222	82	2012-08-02 04:44:06.522533
917	6	1-430	914	3	29	945	35	2012-08-02 04:44:06.522533
918	6	1-324	914	29	8	3295	84	2012-08-02 04:44:06.522533
919	6	1-241	914	8	60	1553	57	2012-08-02 04:44:06.522533
914	6	1-446	914	60	97	7100	37	2012-08-02 04:44:06.522533
921	3	1-114	920	32	44	6919	9	2012-08-02 04:44:06.589075
922	3	1-333	920	44	41	7929	43	2012-08-02 04:44:06.589075
920	3	1-448	920	41	32	5497	89	2012-08-02 04:44:06.589075
924	7	1-205	923	43	4	1514	29	2012-08-02 04:44:06.589075
925	7	1-227	923	4	26	261	80	2012-08-02 04:44:06.589075
926	7	1-342	923	26	57	274	2	2012-08-02 04:44:06.589075
927	7	1-121	923	57	47	3712	56	2012-08-02 04:44:06.589075
928	7	1-353	923	47	74	2792	67	2012-08-02 04:44:06.589075
929	7	1-436	923	74	41	1402	43	2012-08-02 04:44:06.589075
923	7	1-448	923	41	43	1382	89	2012-08-02 04:44:06.589075
931	6	1-205	930	43	47	3	29	2012-08-02 04:44:06.589075
932	6	1-351	930	47	20	2	5	2012-08-02 04:44:06.589075
933	6	1-20	930	20	75	10	30	2012-08-02 04:44:06.589075
934	6	1-438	930	75	28	3	11	2012-08-02 04:44:06.589075
935	6	1-30	930	28	41	2	43	2012-08-02 04:44:06.589075
930	6	1-448	930	41	43	2	89	2012-08-02 04:44:06.589075
937	3	1-205	936	43	41	2388	29	2012-08-02 04:44:06.589075
938	3	1-437	936	41	41	2385	43	2012-08-02 04:44:06.589075
936	3	1-448	936	41	43	2685	89	2012-08-02 04:44:06.589075
940	8	1-80	939	54	84	1648	78	2012-08-02 04:44:07.264086
941	8	1-452	939	84	98	531	30	2012-08-02 04:44:07.264086
942	8	1-424	939	98	61	357	66	2012-08-02 04:44:07.264086
943	8	1-158	939	61	84	230	98	2012-08-02 04:44:07.264086
944	8	1-226	939	84	12	868	20	2012-08-02 04:44:07.264086
945	8	1-144	939	12	19	905	37	2012-08-02 04:44:07.264086
946	8	1-95	939	19	58	928	60	2012-08-02 04:44:07.264086
939	8	1-456	939	58	54	920	3	2012-08-02 04:44:07.264086
948	8	1-381	947	30	76	855	80	2012-08-02 04:44:07.673115
949	8	1-427	947	76	6	1405	53	2012-08-02 04:44:07.673115
950	8	1-88	947	6	94	1740	82	2012-08-02 04:44:07.673115
951	8	1-383	947	94	42	1347	91	2012-08-02 04:44:07.673115
947	8	1-460	947	59	30	2004	28	2012-08-02 04:44:07.673115
952	8	1-399	947	42	75	1177	48	2012-08-02 04:44:07.673115
953	8	1-251	947	75	71	738	15	2012-08-02 04:44:07.673115
954	8	1-281	947	71	59	947	86	2012-08-02 04:44:07.673115
956	7	1-381	955	30	26	1499	80	2012-08-02 04:44:07.673115
957	7	1-342	955	26	67	2204	2	2012-08-02 04:44:07.673115
958	7	1-321	955	67	85	2139	4	2012-08-02 04:44:07.673115
959	7	1-387	955	85	36	2008	22	2012-08-02 04:44:07.673115
960	7	1-301	955	36	67	3165	46	2012-08-02 04:44:07.673115
961	7	1-154	955	67	59	1332	86	2012-08-02 04:44:07.673115
955	7	1-460	955	59	30	3147	28	2012-08-02 04:44:07.673115
963	6	1-18	962	18	54	1904	3	2012-08-02 04:44:07.673115
964	6	1-80	962	54	84	3520	78	2012-08-02 04:44:07.673115
965	6	1-452	962	84	98	1098	30	2012-08-02 04:44:07.673115
966	6	1-424	962	98	37	763	66	2012-08-02 04:44:07.673115
967	6	1-339	962	37	59	893	86	2012-08-02 04:44:07.673115
962	6	1-460	962	59	18	2194	28	2012-08-02 04:44:07.673115
969	7	1-251	968	75	46	149	15	2012-08-02 04:44:07.861537
970	7	1-313	968	46	11	112	44	2012-08-02 04:44:07.861537
971	7	1-325	968	11	45	166	52	2012-08-02 04:44:07.861537
972	7	1-58	968	45	58	185	60	2012-08-02 04:44:07.861537
973	7	1-456	968	58	26	190	3	2012-08-02 04:44:07.861537
974	7	1-337	968	26	72	42	84	2012-08-02 04:44:07.861537
968	7	1-461	968	72	75	204	48	2012-08-02 04:44:07.861537
976	4	1-205	975	43	30	4150	29	2012-08-02 04:44:08.425838
977	4	1-354	975	30	58	3233	85	2012-08-02 04:44:08.425838
978	4	1-167	975	58	18	4347	37	2012-08-02 04:44:08.425838
975	4	1-468	975	18	43	2387	89	2012-08-02 04:44:08.425838
980	8	1-456	979	58	54	1051	3	2012-08-02 04:44:08.448015
981	8	1-80	979	54	84	2065	78	2012-08-02 04:44:08.448015
982	8	1-452	979	84	71	685	30	2012-08-02 04:44:08.448015
983	8	1-410	979	71	5	742	9	2012-08-02 04:44:08.448015
984	8	1-5	979	5	85	581	4	2012-08-02 04:44:08.448015
985	8	1-387	979	85	24	603	22	2012-08-02 04:44:08.448015
986	8	1-458	979	24	94	2543	7	2012-08-02 04:44:08.448015
979	8	1-469	979	94	58	966	60	2012-08-02 04:44:08.448015
988	7	1-345	987	50	30	612	29	2012-08-02 04:44:08.626719
989	7	1-354	987	30	13	451	85	2012-08-02 04:44:08.626719
990	7	1-96	987	13	30	448	39	2012-08-02 04:44:08.626719
991	7	1-232	987	30	4	4255	41	2012-08-02 04:44:08.626719
992	7	1-28	987	4	9	1267	40	2012-08-02 04:44:08.626719
993	7	1-395	987	9	17	1027	45	2012-08-02 04:44:08.626719
987	7	1-471	987	17	50	631	71	2012-08-02 04:44:08.626719
995	4	1-74	994	42	96	1341	76	2012-08-02 04:44:08.626719
996	4	1-361	994	96	89	1615	61	2012-08-02 04:44:08.626719
997	4	1-439	994	89	17	2492	45	2012-08-02 04:44:08.626719
994	4	1-471	994	17	42	1626	71	2012-08-02 04:44:08.626719
999	7	1-345	998	50	4	2984	29	2012-08-02 04:44:08.626719
1000	7	1-227	998	4	43	769	80	2012-08-02 04:44:08.626719
1001	7	1-408	998	43	89	1033	26	2012-08-02 04:44:08.626719
1002	7	1-213	998	89	61	5596	34	2012-08-02 04:44:08.626719
1003	7	1-106	998	61	9	5091	40	2012-08-02 04:44:08.626719
1004	7	1-395	998	9	17	4403	45	2012-08-02 04:44:08.626719
998	7	1-471	998	17	50	2884	71	2012-08-02 04:44:08.626719
1006	6	1-468	1005	18	43	765	89	2012-08-02 04:44:10.064068
1007	6	1-205	1005	43	4	1218	29	2012-08-02 04:44:10.064068
1008	6	1-227	1005	4	26	306	80	2012-08-02 04:44:10.064068
1009	6	1-342	1005	26	1	466	2	2012-08-02 04:44:10.064068
1010	6	1-1	1005	1	37	760	1	2012-08-02 04:44:10.064068
1005	6	1-482	1005	37	18	1521	37	2012-08-02 04:44:10.064068
1012	6	1-169	1011	70	47	347	56	2012-08-02 04:44:10.086239
1013	6	1-353	1011	47	74	354	67	2012-08-02 04:44:10.086239
1014	6	1-436	1011	74	57	242	43	2012-08-02 04:44:10.086239
1015	6	1-476	1011	57	51	184	65	2012-08-02 04:44:10.086239
1016	6	1-71	1011	51	9	116	21	2012-08-02 04:44:10.086239
1011	6	1-483	1011	9	70	49	99	2012-08-02 04:44:10.086239
1018	8	1-169	1017	70	41	405	56	2012-08-02 04:44:10.086239
1019	8	1-149	1017	41	9	528	40	2012-08-02 04:44:10.086239
1020	8	1-395	1017	9	17	423	45	2012-08-02 04:44:10.086239
1021	8	1-471	1017	17	50	221	71	2012-08-02 04:44:10.086239
1022	8	1-345	1017	50	30	182	29	2012-08-02 04:44:10.086239
1023	8	1-354	1017	30	48	107	85	2012-08-02 04:44:10.086239
1024	8	1-445	1017	48	9	152	21	2012-08-02 04:44:10.086239
1017	8	1-483	1017	9	70	66	99	2012-08-02 04:44:10.086239
1026	8	1-180	1025	2	5	243	9	2012-08-02 04:44:10.639252
1027	8	1-5	1025	5	85	152	4	2012-08-02 04:44:10.639252
1028	8	1-387	1025	85	24	127	22	2012-08-02 04:44:10.639252
1029	8	1-458	1025	24	32	428	7	2012-08-02 04:44:10.639252
1030	8	1-244	1025	32	56	4714	81	2012-08-02 04:44:10.639252
1031	8	1-90	1025	56	6	664	67	2012-08-02 04:44:10.639252
1032	8	1-411	1025	6	8	1496	34	2012-08-02 04:44:10.639252
1025	8	1-487	1025	8	2	633	32	2012-08-02 04:44:10.639252
1034	6	1-204	1033	68	30	191	29	2012-08-02 04:44:10.827218
1035	6	1-354	1033	30	48	147	85	2012-08-02 04:44:10.827218
1036	6	1-445	1033	48	9	277	21	2012-08-02 04:44:10.827218
1037	6	1-483	1033	9	70	136	99	2012-08-02 04:44:10.827218
1038	6	1-169	1033	70	13	1107	56	2012-08-02 04:44:10.827218
1033	6	1-489	1033	13	68	209	39	2012-08-02 04:44:10.827218
1040	5	1-120	1039	67	56	921	36	2012-08-02 04:44:10.860741
1041	5	1-273	1039	56	53	560	72	2012-08-02 04:44:10.860741
1042	5	1-355	1039	53	43	293	68	2012-08-02 04:44:10.860741
1043	5	1-360	1039	43	63	445	46	2012-08-02 04:44:10.860741
1039	5	1-490	1039	63	67	3789	50	2012-08-02 04:44:10.860741
1045	7	1-342	1044	26	1	12	2	2012-08-02 04:44:11.02645
1046	7	1-1	1044	1	37	15	1	2012-08-02 04:44:11.02645
1047	7	1-482	1044	37	18	23	37	2012-08-02 04:44:11.02645
1048	7	1-468	1044	18	41	9	89	2012-08-02 04:44:11.02645
1049	7	1-224	1044	41	68	123	72	2012-08-02 04:44:11.02645
1050	7	1-297	1044	68	72	22	30	2012-08-02 04:44:11.02645
1044	7	1-491	1044	72	26	10	80	2012-08-02 04:44:11.02645
1052	8	1-342	1051	26	1	149	2	2012-08-02 04:44:11.02645
1053	8	1-1	1051	1	37	227	1	2012-08-02 04:44:11.02645
1054	8	1-482	1051	37	19	423	37	2012-08-02 04:44:11.02645
1055	8	1-95	1051	19	58	416	60	2012-08-02 04:44:11.02645
1056	8	1-456	1051	58	54	395	3	2012-08-02 04:44:11.02645
1057	8	1-80	1051	54	84	678	78	2012-08-02 04:44:11.02645
1058	8	1-452	1051	84	72	196	30	2012-08-02 04:44:11.02645
1051	8	1-491	1051	72	26	105	80	2012-08-02 04:44:11.02645
15860	5	1-3605	15859	91	83	5	54	2012-08-02 05:13:08.412048
1060	5	1-5	1059	5	92	1666	4	2012-08-02 04:44:11.413861
1061	5	1-287	1059	92	37	1058	40	2012-08-02 04:44:11.413861
1062	5	1-467	1059	37	14	960	68	2012-08-02 04:44:11.413861
1063	5	1-472	1059	14	6	512	94	2012-08-02 04:44:11.413861
1059	5	1-495	1059	6	5	2457	9	2012-08-02 04:44:11.413861
1065	6	1-172	1064	1	46	492	15	2012-08-02 04:44:11.590666
1066	6	1-313	1064	46	11	384	44	2012-08-02 04:44:11.590666
1067	6	1-325	1064	11	45	590	52	2012-08-02 04:44:11.590666
1068	6	1-58	1064	45	58	681	60	2012-08-02 04:44:11.590666
1069	6	1-456	1064	58	42	723	3	2012-08-02 04:44:11.590666
1064	6	1-496	1064	42	1	1219	20	2012-08-02 04:44:11.590666
1071	5	1-435	1070	9	53	4106	72	2012-08-02 04:44:12.619404
1072	5	1-355	1070	53	59	1916	68	2012-08-02 04:44:12.619404
1073	5	1-274	1070	59	37	1210	51	2012-08-02 04:44:12.619404
1074	5	1-314	1070	37	80	5572	59	2012-08-02 04:44:12.619404
1070	5	1-506	1070	80	9	2103	24	2012-08-02 04:44:12.619404
1076	6	1-435	1075	9	53	1859	72	2012-08-02 04:44:12.619404
1077	6	1-355	1075	53	59	1023	68	2012-08-02 04:44:12.619404
1078	6	1-274	1075	59	92	761	51	2012-08-02 04:44:12.619404
1079	6	1-418	1075	92	52	3857	38	2012-08-02 04:44:12.619404
1080	6	1-217	1075	52	80	1815	59	2012-08-02 04:44:12.619404
1075	6	1-506	1075	80	9	808	24	2012-08-02 04:44:12.619404
1082	4	1-106	1081	61	37	4229	40	2012-08-02 04:44:13.048831
1083	4	1-467	1081	37	43	2903	68	2012-08-02 04:44:13.048831
1084	4	1-360	1081	43	44	3500	46	2012-08-02 04:44:13.048831
1081	4	1-507	1081	44	61	6847	34	2012-08-02 04:44:13.048831
1086	7	1-479	1085	93	99	485	90	2012-08-02 04:44:13.048831
1087	7	1-392	1085	99	58	583	62	2012-08-02 04:44:13.048831
1088	7	1-93	1085	58	40	296	67	2012-08-02 04:44:13.048831
1089	7	1-92	1085	40	37	399	40	2012-08-02 04:44:13.048831
1090	7	1-467	1085	37	43	375	68	2012-08-02 04:44:13.048831
1091	7	1-360	1085	43	44	619	46	2012-08-02 04:44:13.048831
1085	7	1-507	1085	44	93	1656	34	2012-08-02 04:44:13.048831
1093	6	1-487	1092	8	2	298	32	2012-08-02 04:44:13.048831
1094	6	1-180	1092	2	5	141	9	2012-08-02 04:44:13.048831
1095	6	1-5	1092	5	85	109	4	2012-08-02 04:44:13.048831
1096	6	1-387	1092	85	36	112	22	2012-08-02 04:44:13.048831
1097	6	1-301	1092	36	44	193	46	2012-08-02 04:44:13.048831
1092	6	1-507	1092	44	8	571	34	2012-08-02 04:44:13.048831
1099	5	1-190	1098	3	14	649	23	2012-08-02 04:44:13.448081
1100	5	1-14	1098	14	24	897	22	2012-08-02 04:44:13.448081
1101	5	1-458	1098	24	42	3117	7	2012-08-02 04:44:13.448081
1102	5	1-402	1098	42	14	2707	95	2012-08-02 04:44:13.448081
1098	5	1-509	1098	14	3	1450	82	2012-08-02 04:44:13.448081
1104	8	1-214	1103	19	9	108	21	2012-08-02 04:44:13.448081
1105	8	1-483	1103	9	70	49	99	2012-08-02 04:44:13.448081
1106	8	1-169	1103	70	13	372	56	2012-08-02 04:44:13.448081
1107	8	1-489	1103	13	30	65	39	2012-08-02 04:44:13.448081
1108	8	1-232	1103	30	92	597	41	2012-08-02 04:44:13.448081
1109	8	1-237	1103	92	42	775	7	2012-08-02 04:44:13.448081
1110	8	1-402	1103	42	14	767	95	2012-08-02 04:44:13.448081
1103	8	1-509	1103	14	19	417	82	2012-08-02 04:44:13.448081
1112	8	1-207	1111	50	45	136	73	2012-08-02 04:44:13.647161
1113	8	1-296	1111	45	6	236	53	2012-08-02 04:44:13.647161
1114	8	1-88	1111	6	3	256	82	2012-08-02 04:44:13.647161
1115	8	1-430	1111	3	29	237	35	2012-08-02 04:44:13.647161
1116	8	1-324	1111	29	8	993	84	2012-08-02 04:44:13.647161
1117	8	1-241	1111	8	10	561	57	2012-08-02 04:44:13.647161
1118	8	1-485	1111	10	55	274	5	2012-08-02 04:44:13.647161
1111	8	1-510	1111	55	50	276	49	2012-08-02 04:44:13.647161
1120	4	1-207	1119	50	61	1026	73	2012-08-02 04:44:13.647161
1121	4	1-494	1119	61	96	4702	26	2012-08-02 04:44:13.647161
1122	4	1-404	1119	96	55	1415	5	2012-08-02 04:44:13.647161
1119	4	1-510	1119	55	50	1726	49	2012-08-02 04:44:13.647161
1124	8	1-455	1123	83	7	3972	87	2012-08-02 04:44:13.647161
1125	8	1-346	1123	7	99	1638	90	2012-08-02 04:44:13.647161
1126	8	1-392	1123	99	47	2146	62	2012-08-02 04:44:13.647161
1127	8	1-66	1123	47	41	1570	44	2012-08-02 04:44:13.647161
1128	8	1-331	1123	41	97	447	100	2012-08-02 04:44:13.647161
1129	8	1-459	1123	97	10	732	57	2012-08-02 04:44:13.647161
1130	8	1-485	1123	10	55	491	5	2012-08-02 04:44:13.647161
1123	8	1-510	1123	55	83	678	49	2012-08-02 04:44:13.647161
1132	3	1-413	1131	71	37	430	61	2012-08-02 04:44:13.890634
1133	3	1-47	1131	37	76	319	33	2012-08-02 04:44:13.890634
1131	3	1-511	1131	76	71	1604	53	2012-08-02 04:44:13.890634
1135	2	1-254	1134	20	4	1367	1	2012-08-02 04:44:13.968029
1134	2	1-512	1134	4	20	1197	88	2012-08-02 04:44:13.968029
1137	7	1-82	1136	4	29	45	69	2012-08-02 04:44:14.88528
1138	7	1-222	1136	29	14	60	95	2012-08-02 04:44:14.88528
1139	7	1-509	1136	14	3	33	82	2012-08-02 04:44:14.88528
1140	7	1-430	1136	3	55	40	35	2012-08-02 04:44:14.88528
1141	7	1-84	1136	55	45	535	52	2012-08-02 04:44:14.88528
1142	7	1-58	1136	45	44	592	60	2012-08-02 04:44:14.88528
1136	7	1-518	1136	44	4	657	51	2012-08-02 04:44:14.88528
1144	7	1-127	1143	71	56	3017	81	2012-08-02 04:44:15.692403
1145	7	1-90	1143	56	74	468	67	2012-08-02 04:44:15.692403
1146	7	1-436	1143	74	84	321	43	2012-08-02 04:44:15.692403
1147	7	1-375	1143	84	19	414	17	2012-08-02 04:44:15.692403
1148	7	1-108	1143	19	70	897	6	2012-08-02 04:44:15.692403
1149	7	1-264	1143	70	90	928	42	2012-08-02 04:44:15.692403
1143	7	1-522	1143	90	71	1315	70	2012-08-02 04:44:15.692403
1151	7	1-419	1150	78	29	2571	50	2012-08-02 04:44:17.504155
1152	7	1-248	1150	29	82	427	45	2012-08-02 04:44:17.504155
1153	7	1-164	1150	82	68	845	13	2012-08-02 04:44:17.504155
1154	7	1-238	1150	68	40	911	52	2012-08-02 04:44:17.504155
1155	7	1-517	1150	40	42	343	3	2012-08-02 04:44:17.504155
1156	7	1-496	1150	42	4	555	20	2012-08-02 04:44:17.504155
1150	7	1-527	1150	4	78	2475	23	2012-08-02 04:44:17.504155
1158	8	1-321	1157	67	85	172	4	2012-08-02 04:44:18.664455
1159	8	1-387	1157	85	24	155	22	2012-08-02 04:44:18.664455
1160	8	1-458	1157	24	42	566	7	2012-08-02 04:44:18.664455
1161	8	1-402	1157	42	14	516	95	2012-08-02 04:44:18.664455
1157	8	1-532	1157	38	67	198	2	2012-08-02 04:44:18.664455
1162	8	1-509	1157	14	70	259	82	2012-08-02 04:44:18.664455
1163	8	1-515	1157	70	52	604	38	2012-08-02 04:44:18.664455
1164	8	1-217	1157	52	38	283	59	2012-08-02 04:44:18.664455
15861	5	1-4471	15859	83	86	9	84	2012-08-02 05:13:08.412048
1166	6	1-1	1165	1	37	1580	1	2012-08-02 04:44:18.664455
1167	6	1-482	1165	37	23	3043	37	2012-08-02 04:44:18.664455
1168	6	1-492	1165	23	85	5727	51	2012-08-02 04:44:18.664455
1169	6	1-457	1165	85	80	717	26	2012-08-02 04:44:18.664455
1170	6	1-261	1165	80	38	1382	59	2012-08-02 04:44:18.664455
1165	6	1-532	1165	38	1	1007	2	2012-08-02 04:44:18.664455
15862	5	1-2575	15859	86	84	3	23	2012-08-02 05:13:08.412048
1172	8	1-1	1171	1	37	765	1	2012-08-02 04:44:18.664455
1173	8	1-482	1171	37	19	1561	37	2012-08-02 04:44:18.664455
1174	8	1-95	1171	19	58	1678	60	2012-08-02 04:44:18.664455
1175	8	1-456	1171	58	42	1746	3	2012-08-02 04:44:18.664455
1176	8	1-496	1171	42	83	2883	20	2012-08-02 04:44:18.664455
1177	8	1-162	1171	83	61	1414	14	2012-08-02 04:44:18.664455
1178	8	1-484	1171	61	38	596	59	2012-08-02 04:44:18.664455
1171	8	1-532	1171	38	1	460	2	2012-08-02 04:44:18.664455
15863	5	1-3038	15859	84	80	1	12	2012-08-02 05:13:08.412048
1180	7	1-1	1179	1	37	258	1	2012-08-02 04:44:18.664455
1181	7	1-482	1179	37	19	537	37	2012-08-02 04:44:18.664455
1182	7	1-95	1179	19	44	589	60	2012-08-02 04:44:18.664455
1183	7	1-518	1179	44	85	683	51	2012-08-02 04:44:18.664455
1184	7	1-457	1179	85	80	93	26	2012-08-02 04:44:18.664455
1185	7	1-261	1179	80	38	193	59	2012-08-02 04:44:18.664455
1179	7	1-532	1179	38	1	152	2	2012-08-02 04:44:18.664455
15859	5	1-4567	15859	80	91	5	19	2012-08-02 05:13:08.412048
1187	7	1-1	1186	1	6	681	1	2012-08-02 04:44:18.664455
1188	7	1-303	1186	6	2	648	36	2012-08-02 04:44:18.664455
1189	7	1-417	1186	2	14	582	95	2012-08-02 04:44:18.664455
1190	7	1-509	1186	14	70	336	82	2012-08-02 04:44:18.664455
1191	7	1-515	1186	70	52	902	38	2012-08-02 04:44:18.664455
1192	7	1-217	1186	52	38	487	59	2012-08-02 04:44:18.664455
1186	7	1-532	1186	38	1	393	2	2012-08-02 04:44:18.664455
1194	8	1-331	1193	41	38	246	100	2012-08-02 04:44:19.294927
1195	8	1-372	1193	38	82	350	69	2012-08-02 04:44:19.294927
1196	8	1-508	1193	82	45	249	73	2012-08-02 04:44:19.294927
1197	8	1-296	1193	45	71	628	53	2012-08-02 04:44:19.294927
1198	8	1-413	1193	71	23	703	61	2012-08-02 04:44:19.294927
1199	8	1-146	1193	23	17	807	27	2012-08-02 04:44:19.294927
1200	8	1-17	1193	17	26	1369	20	2012-08-02 04:44:19.294927
1193	8	1-534	1193	26	41	818	44	2012-08-02 04:44:19.294927
1202	7	1-324	1201	29	10	3598	84	2012-08-02 04:44:19.705894
1203	7	1-505	1201	10	8	1995	91	2012-08-02 04:44:19.705894
1204	7	1-499	1201	8	89	397	75	2012-08-02 04:44:19.705894
1205	7	1-365	1201	89	50	405	77	2012-08-02 04:44:19.705894
1206	7	1-128	1201	50	93	616	88	2012-08-02 04:44:19.705894
1207	7	1-291	1201	93	13	539	89	2012-08-02 04:44:19.705894
1201	7	1-536	1201	13	29	709	35	2012-08-02 04:44:19.705894
1209	8	1-117	1208	21	13	100	85	2012-08-02 04:44:20.015425
1210	8	1-96	1208	13	11	74	39	2012-08-02 04:44:20.015425
1211	8	1-27	1208	11	77	1984	33	2012-08-02 04:44:20.015425
1212	8	1-147	1208	77	49	1728	97	2012-08-02 04:44:20.015425
1213	8	1-466	1208	49	88	620	18	2012-08-02 04:44:20.015425
1214	8	1-519	1208	88	8	433	34	2012-08-02 04:44:20.015425
1215	8	1-487	1208	8	26	154	32	2012-08-02 04:44:20.015425
1208	8	1-540	1208	26	21	64	92	2012-08-02 04:44:20.015425
1217	5	1-117	1216	21	70	4570	85	2012-08-02 04:44:20.015425
1218	5	1-134	1216	70	46	4007	58	2012-08-02 04:44:20.015425
1219	5	1-191	1216	46	75	4623	74	2012-08-02 04:44:20.015425
1220	5	1-447	1216	75	26	5118	32	2012-08-02 04:44:20.015425
1216	5	1-540	1216	26	21	2515	92	2012-08-02 04:44:20.015425
1222	7	1-117	1221	21	13	76	85	2012-08-02 04:44:20.015425
1223	7	1-96	1221	13	30	67	39	2012-08-02 04:44:20.015425
1224	7	1-232	1221	30	76	522	41	2012-08-02 04:44:20.015425
1225	7	1-433	1221	76	4	1117	51	2012-08-02 04:44:20.015425
1226	7	1-82	1221	4	7	66	69	2012-08-02 04:44:20.015425
1227	7	1-60	1221	7	26	83	32	2012-08-02 04:44:20.015425
1221	7	1-540	1221	26	21	41	92	2012-08-02 04:44:20.015425
1229	8	1-117	1228	21	13	87	85	2012-08-02 04:44:20.015425
1230	8	1-96	1228	13	30	80	39	2012-08-02 04:44:20.015425
1231	8	1-232	1228	30	76	662	41	2012-08-02 04:44:20.015425
1232	8	1-142	1228	76	58	174	62	2012-08-02 04:44:20.015425
1233	8	1-93	1228	58	6	83	67	2012-08-02 04:44:20.015425
1234	8	1-411	1228	6	8	196	34	2012-08-02 04:44:20.015425
1235	8	1-487	1228	8	26	87	32	2012-08-02 04:44:20.015425
1228	8	1-540	1228	26	21	45	92	2012-08-02 04:44:20.015425
1237	8	1-396	1236	53	47	352	29	2012-08-02 04:44:20.015425
1238	8	1-351	1236	47	55	293	5	2012-08-02 04:44:20.015425
1239	8	1-510	1236	55	83	371	49	2012-08-02 04:44:20.015425
1240	8	1-455	1236	83	23	1995	87	2012-08-02 04:44:20.015425
1241	8	1-373	1236	23	88	3795	18	2012-08-02 04:44:20.015425
1242	8	1-519	1236	88	8	3508	34	2012-08-02 04:44:20.015425
1243	8	1-487	1236	8	26	1657	32	2012-08-02 04:44:20.015425
1236	8	1-540	1236	26	53	918	92	2012-08-02 04:44:20.015425
1245	6	1-91	1244	57	10	305	84	2012-08-02 04:44:22.093192
1246	6	1-505	1244	10	37	135	91	2012-08-02 04:44:22.093192
1247	6	1-543	1244	37	3	53	82	2012-08-02 04:44:22.093192
1248	6	1-430	1244	3	55	48	35	2012-08-02 04:44:22.093192
1249	6	1-84	1244	55	60	475	52	2012-08-02 04:44:22.093192
1244	6	1-547	1244	60	57	402	11	2012-08-02 04:44:22.093192
1251	7	1-70	1250	29	32	926	16	2012-08-02 04:44:22.093192
1252	7	1-198	1250	32	40	427	67	2012-08-02 04:44:22.093192
1253	7	1-92	1250	40	9	531	40	2012-08-02 04:44:22.093192
1254	7	1-395	1250	9	82	393	45	2012-08-02 04:44:22.093192
1255	7	1-164	1250	82	68	684	13	2012-08-02 04:44:22.093192
1256	7	1-238	1250	68	60	650	52	2012-08-02 04:44:22.093192
1250	7	1-547	1250	60	29	657	11	2012-08-02 04:44:22.093192
1258	8	1-30	1257	28	57	218	43	2012-08-02 04:44:22.093192
1259	8	1-476	1257	57	76	179	65	2012-08-02 04:44:22.093192
1260	8	1-277	1257	76	57	161	79	2012-08-02 04:44:22.093192
1261	8	1-223	1257	57	37	41	91	2012-08-02 04:44:22.093192
1262	8	1-543	1257	37	3	20	82	2012-08-02 04:44:22.093192
1263	8	1-430	1257	3	55	23	35	2012-08-02 04:44:22.093192
1264	8	1-84	1257	55	60	283	52	2012-08-02 04:44:22.093192
1257	8	1-547	1257	60	28	300	11	2012-08-02 04:44:22.093192
1266	8	1-91	1265	57	8	248	84	2012-08-02 04:44:22.093192
1267	8	1-241	1265	8	10	187	57	2012-08-02 04:44:22.093192
1268	8	1-485	1265	10	55	122	5	2012-08-02 04:44:22.093192
1269	8	1-510	1265	55	26	164	49	2012-08-02 04:44:22.093192
1270	8	1-416	1265	26	26	236	20	2012-08-02 04:44:22.093192
1271	8	1-534	1265	26	11	136	44	2012-08-02 04:44:22.093192
1272	8	1-325	1265	11	60	204	52	2012-08-02 04:44:22.093192
1265	8	1-547	1265	60	57	237	11	2012-08-02 04:44:22.093192
1274	3	1-507	1273	44	93	167	34	2012-08-02 04:44:22.56875
1275	3	1-479	1273	93	37	34	90	2012-08-02 04:44:22.56875
1273	3	1-548	1273	37	44	91	46	2012-08-02 04:44:22.56875
15865	3	1-1674	15864	94	45	40	9	2012-08-02 05:13:08.412048
1277	7	1-526	1276	64	84	1703	43	2012-08-02 04:44:22.56875
1278	7	1-375	1276	84	9	2233	17	2012-08-02 04:44:22.56875
1279	7	1-9	1276	9	32	1099	16	2012-08-02 04:44:22.56875
1280	7	1-198	1276	32	6	552	67	2012-08-02 04:44:22.56875
1281	7	1-411	1276	6	93	1392	34	2012-08-02 04:44:22.56875
1282	7	1-479	1276	93	37	408	90	2012-08-02 04:44:22.56875
1276	7	1-548	1276	37	64	1609	46	2012-08-02 04:44:22.56875
15866	3	1-4397	15864	45	80	26	12	2012-08-02 05:13:08.412048
1284	7	1-317	1283	6	75	3102	83	2012-08-02 04:44:23.850508
1285	7	1-529	1283	75	50	841	77	2012-08-02 04:44:23.850508
1286	7	1-128	1283	50	93	623	88	2012-08-02 04:44:23.850508
1287	7	1-291	1283	93	32	265	89	2012-08-02 04:44:23.850508
1288	7	1-114	1283	32	5	310	9	2012-08-02 04:44:23.850508
1289	7	1-5	1283	5	14	102	4	2012-08-02 04:44:23.850508
1283	7	1-553	1283	14	6	1046	28	2012-08-02 04:44:23.850508
15864	3	1-4567	15864	80	94	133	19	2012-08-02 05:13:08.412048
1291	4	1-381	1290	30	26	1379	80	2012-08-02 04:44:23.850508
1292	4	1-342	1290	26	67	1048	2	2012-08-02 04:44:23.850508
1293	4	1-321	1290	67	14	491	4	2012-08-02 04:44:23.850508
1290	4	1-553	1290	14	30	5601	28	2012-08-02 04:44:23.850508
1295	7	1-310	1294	31	66	1978	25	2012-08-02 04:44:23.850508
1296	7	1-220	1294	66	85	54	51	2012-08-02 04:44:23.850508
1297	7	1-457	1294	85	80	5	26	2012-08-02 04:44:23.850508
1298	7	1-261	1294	80	38	6	59	2012-08-02 04:44:23.850508
1299	7	1-532	1294	38	67	3	2	2012-08-02 04:44:23.850508
1300	7	1-321	1294	67	14	2	4	2012-08-02 04:44:23.850508
1294	7	1-553	1294	14	31	26	28	2012-08-02 04:44:23.850508
15868	3	1-385	15867	72	45	2396	9	2012-08-02 05:13:08.412048
1302	7	1-317	1301	6	75	5004	83	2012-08-02 04:44:23.850508
1303	7	1-529	1301	75	22	1724	77	2012-08-02 04:44:23.850508
1304	7	1-322	1301	22	80	386	26	2012-08-02 04:44:23.850508
1305	7	1-261	1301	80	38	440	59	2012-08-02 04:44:23.850508
1306	7	1-532	1301	38	67	190	2	2012-08-02 04:44:23.850508
1307	7	1-321	1301	67	14	102	4	2012-08-02 04:44:23.850508
1301	7	1-553	1301	14	6	1327	28	2012-08-02 04:44:23.850508
1309	7	1-317	1308	6	38	1005	83	2012-08-02 04:44:23.850508
1310	7	1-470	1308	38	68	432	39	2012-08-02 04:44:23.850508
1311	7	1-204	1308	68	4	222	29	2012-08-02 04:44:23.850508
1312	7	1-227	1308	4	26	34	80	2012-08-02 04:44:23.850508
1313	7	1-342	1308	26	67	31	2	2012-08-02 04:44:23.850508
1314	7	1-321	1308	67	14	18	4	2012-08-02 04:44:23.850508
1308	7	1-553	1308	14	6	250	28	2012-08-02 04:44:23.850508
1316	7	1-72	1315	49	24	384	8	2012-08-02 04:44:23.850508
1317	7	1-462	1315	24	49	974	63	2012-08-02 04:44:23.850508
1318	7	1-357	1315	49	61	670	14	2012-08-02 04:44:23.850508
1319	7	1-484	1315	61	38	204	59	2012-08-02 04:44:23.850508
1320	7	1-532	1315	38	67	114	2	2012-08-02 04:44:23.850508
1321	7	1-321	1315	67	14	79	4	2012-08-02 04:44:23.850508
1315	7	1-553	1315	14	49	1330	28	2012-08-02 04:44:23.850508
1322	1	1-342	\N	26	26	2693	2	2012-08-02 04:44:23.850508
1324	7	1-232	1323	30	76	2094	41	2012-08-02 04:44:25.331686
1325	7	1-142	1323	76	47	456	62	2012-08-02 04:44:25.331686
1326	7	1-66	1323	47	11	238	44	2012-08-02 04:44:25.331686
1327	7	1-325	1323	11	60	262	52	2012-08-02 04:44:25.331686
1328	7	1-547	1323	60	57	224	11	2012-08-02 04:44:25.331686
1329	7	1-91	1323	57	21	171	84	2012-08-02 04:44:25.331686
1323	7	1-557	1323	21	30	307	39	2012-08-02 04:44:25.331686
1331	7	1-204	1330	68	47	425	29	2012-08-02 04:44:25.331686
1332	7	1-351	1330	47	19	307	5	2012-08-02 04:44:25.331686
1333	7	1-208	1330	19	68	314	13	2012-08-02 04:44:25.331686
1334	7	1-238	1330	68	60	325	52	2012-08-02 04:44:25.331686
1335	7	1-547	1330	60	57	309	11	2012-08-02 04:44:25.331686
1336	7	1-91	1330	57	21	264	84	2012-08-02 04:44:25.331686
1330	7	1-557	1330	21	68	527	39	2012-08-02 04:44:25.331686
1338	8	1-204	1337	68	47	1551	29	2012-08-02 04:44:25.331686
1339	8	1-351	1337	47	71	1237	5	2012-08-02 04:44:25.331686
1340	8	1-525	1337	71	68	8889	72	2012-08-02 04:44:25.331686
1341	8	1-297	1337	68	98	2002	30	2012-08-02 04:44:25.331686
1342	8	1-424	1337	98	61	1267	66	2012-08-02 04:44:25.331686
1343	8	1-158	1337	61	52	766	98	2012-08-02 04:44:25.331686
1344	8	1-363	1337	52	21	830	84	2012-08-02 04:44:25.331686
1337	8	1-557	1337	21	68	1829	39	2012-08-02 04:44:25.331686
1346	7	1-332	1345	40	80	2424	79	2012-08-02 04:44:25.331686
1347	7	1-426	1345	80	46	741	15	2012-08-02 04:44:25.331686
1348	7	1-313	1345	46	11	556	44	2012-08-02 04:44:25.331686
1349	7	1-325	1345	11	60	822	52	2012-08-02 04:44:25.331686
1350	7	1-547	1345	60	57	945	11	2012-08-02 04:44:25.331686
1351	7	1-91	1345	57	21	976	84	2012-08-02 04:44:25.331686
1345	7	1-557	1345	21	40	2352	39	2012-08-02 04:44:25.331686
1353	7	1-204	1352	68	46	263	29	2012-08-02 04:44:25.331686
1354	7	1-356	1352	46	84	678	78	2012-08-02 04:44:25.331686
1355	7	1-452	1352	84	98	215	30	2012-08-02 04:44:25.331686
1356	7	1-424	1352	98	61	152	66	2012-08-02 04:44:25.331686
1357	7	1-158	1352	61	52	103	98	2012-08-02 04:44:25.331686
1358	7	1-363	1352	52	21	124	84	2012-08-02 04:44:25.331686
1352	7	1-557	1352	21	68	305	39	2012-08-02 04:44:25.331686
1360	7	1-204	1359	68	47	6	29	2012-08-02 04:44:25.331686
1361	7	1-351	1359	47	55	5	5	2012-08-02 04:44:25.331686
1362	7	1-510	1359	55	21	7	49	2012-08-02 04:44:25.331686
1363	7	1-39	1359	21	60	3	52	2012-08-02 04:44:25.331686
1364	7	1-547	1359	60	57	3	11	2012-08-02 04:44:25.331686
1365	7	1-91	1359	57	21	3	84	2012-08-02 04:44:25.331686
1359	7	1-557	1359	21	68	7	39	2012-08-02 04:44:25.331686
15869	3	1-4397	15867	45	80	1631	12	2012-08-02 05:13:08.412048
1367	8	1-509	1366	14	3	444	82	2012-08-02 04:44:25.862386
1368	8	1-430	1366	3	89	575	35	2012-08-02 04:44:25.862386
1369	8	1-542	1366	89	59	6011	68	2012-08-02 04:44:25.862386
1370	8	1-274	1366	59	4	5106	51	2012-08-02 04:44:25.862386
1371	8	1-82	1366	4	82	374	69	2012-08-02 04:44:25.862386
1372	8	1-508	1366	82	70	256	73	2012-08-02 04:44:25.862386
1373	8	1-441	1366	70	41	321	22	2012-08-02 04:44:25.862386
1366	8	1-558	1366	41	14	773	95	2012-08-02 04:44:25.862386
15867	3	1-4567	15867	80	72	8674	19	2012-08-02 05:13:08.412048
1375	8	1-313	1374	46	11	46	44	2012-08-02 04:44:26.24909
1376	8	1-325	1374	11	60	64	52	2012-08-02 04:44:26.24909
1377	8	1-547	1374	60	57	70	11	2012-08-02 04:44:26.24909
1378	8	1-91	1374	57	21	68	84	2012-08-02 04:44:26.24909
1379	8	1-557	1374	21	40	155	39	2012-08-02 04:44:26.24909
1380	8	1-332	1374	40	76	151	79	2012-08-02 04:44:26.24909
1381	8	1-161	1374	76	45	908	31	2012-08-02 04:44:26.24909
1374	8	1-561	1374	45	46	65	15	2012-08-02 04:44:26.24909
1383	3	1-542	1382	89	2	1005	68	2012-08-02 04:44:27.33196
1384	3	1-272	1382	2	1	238	19	2012-08-02 04:44:27.33196
1382	3	1-564	1382	1	89	373	35	2012-08-02 04:44:27.33196
1386	8	1-84	1385	55	60	1713	52	2012-08-02 04:44:27.33196
1387	8	1-547	1385	60	29	1122	11	2012-08-02 04:44:27.33196
1388	8	1-70	1385	29	32	1024	16	2012-08-02 04:44:27.33196
1389	8	1-198	1385	32	40	306	67	2012-08-02 04:44:27.33196
1390	8	1-92	1385	40	37	246	40	2012-08-02 04:44:27.33196
1391	8	1-467	1385	37	2	138	68	2012-08-02 04:44:27.33196
1392	8	1-272	1385	2	1	68	19	2012-08-02 04:44:27.33196
1385	8	1-564	1385	1	55	223	35	2012-08-02 04:44:27.33196
1394	8	1-84	1393	55	60	541	52	2012-08-02 04:44:27.33196
1395	8	1-547	1393	60	28	428	11	2012-08-02 04:44:27.33196
1396	8	1-30	1393	28	57	233	43	2012-08-02 04:44:27.33196
1397	8	1-476	1393	57	10	130	65	2012-08-02 04:44:27.33196
1398	8	1-283	1393	10	26	69	20	2012-08-02 04:44:27.33196
1399	8	1-534	1393	26	79	26	44	2012-08-02 04:44:27.33196
1400	8	1-304	1393	79	1	15	19	2012-08-02 04:44:27.33196
1393	8	1-564	1393	1	55	58	35	2012-08-02 04:44:27.33196
1402	4	1-155	1401	80	37	2347	40	2012-08-02 04:44:27.33196
1403	4	1-467	1401	37	2	1629	68	2012-08-02 04:44:27.33196
1404	4	1-272	1401	2	1	999	19	2012-08-02 04:44:27.33196
1401	4	1-564	1401	1	80	4038	35	2012-08-02 04:44:27.33196
1406	7	1-99	1405	52	28	260	11	2012-08-02 04:44:27.33196
1407	7	1-30	1405	28	57	214	43	2012-08-02 04:44:27.33196
1408	7	1-476	1405	57	10	181	65	2012-08-02 04:44:27.33196
1409	7	1-283	1405	10	26	145	20	2012-08-02 04:44:27.33196
1410	7	1-534	1405	26	79	82	44	2012-08-02 04:44:27.33196
1411	7	1-304	1405	79	1	71	19	2012-08-02 04:44:27.33196
1405	7	1-564	1405	1	52	423	35	2012-08-02 04:44:27.33196
1413	4	1-383	1412	94	8	3116	91	2012-08-02 04:44:28.536941
1414	4	1-499	1412	8	50	730	75	2012-08-02 04:44:28.536941
1415	4	1-69	1412	50	64	720	74	2012-08-02 04:44:28.536941
1412	4	1-567	1412	64	94	3209	82	2012-08-02 04:44:28.536941
1417	7	1-396	1416	53	41	1036	29	2012-08-02 04:44:28.713823
1418	7	1-437	1416	41	84	1198	43	2012-08-02 04:44:28.713823
1419	7	1-375	1416	84	19	1494	17	2012-08-02 04:44:28.713823
1420	7	1-108	1416	19	70	3130	6	2012-08-02 04:44:28.713823
1421	7	1-264	1416	70	90	3131	42	2012-08-02 04:44:28.713823
1422	7	1-522	1416	90	62	4289	70	2012-08-02 04:44:28.713823
1416	7	1-568	1416	62	53	2837	92	2012-08-02 04:44:28.713823
1424	5	1-462	1423	24	49	3824	63	2012-08-02 04:44:29.266055
1425	5	1-357	1423	49	61	2449	14	2012-08-02 04:44:29.266055
1426	5	1-484	1423	61	80	851	59	2012-08-02 04:44:29.266055
1427	5	1-506	1423	80	13	343	24	2012-08-02 04:44:29.266055
1423	5	1-571	1423	13	24	1322	8	2012-08-02 04:44:29.266055
1429	7	1-429	1428	72	68	1317	72	2012-08-02 04:44:29.266055
1430	7	1-297	1428	68	72	281	30	2012-08-02 04:44:29.266055
1431	7	1-491	1428	72	43	166	80	2012-08-02 04:44:29.266055
1432	7	1-408	1428	43	80	206	26	2012-08-02 04:44:29.266055
1433	7	1-261	1428	80	80	393	59	2012-08-02 04:44:29.266055
1434	7	1-506	1428	80	13	180	24	2012-08-02 04:44:29.266055
1428	7	1-571	1428	13	72	785	8	2012-08-02 04:44:29.266055
1436	7	1-4	1435	4	42	1842	7	2012-08-02 04:44:29.266055
1437	7	1-402	1435	42	14	1760	95	2012-08-02 04:44:29.266055
1438	7	1-509	1435	14	70	925	82	2012-08-02 04:44:29.266055
1439	7	1-515	1435	70	52	2254	38	2012-08-02 04:44:29.266055
1440	7	1-217	1435	52	80	1106	59	2012-08-02 04:44:29.266055
1441	7	1-506	1435	80	13	513	24	2012-08-02 04:44:29.266055
1435	7	1-571	1435	13	4	2274	8	2012-08-02 04:44:29.266055
1443	5	1-69	1442	50	64	2049	74	2012-08-02 04:44:29.520494
1444	5	1-567	1442	64	3	3726	82	2012-08-02 04:44:29.520494
1445	5	1-430	1442	3	68	2014	35	2012-08-02 04:44:29.520494
1446	5	1-123	1442	68	91	1229	93	2012-08-02 04:44:29.520494
1442	5	1-572	1442	91	50	5094	75	2012-08-02 04:44:29.520494
1448	5	1-76	1447	1	41	1806	24	2012-08-02 04:44:29.520494
1449	5	1-52	1447	41	38	936	59	2012-08-02 04:44:29.520494
1450	5	1-532	1447	38	94	380	2	2012-08-02 04:44:29.520494
1451	5	1-394	1447	94	91	360	93	2012-08-02 04:44:29.520494
1447	5	1-572	1447	91	1	1809	75	2012-08-02 04:44:29.520494
1453	6	1-365	1452	89	50	1836	77	2012-08-02 04:44:29.520494
1454	6	1-128	1452	50	93	1701	88	2012-08-02 04:44:29.520494
1455	6	1-291	1452	93	13	906	89	2012-08-02 04:44:29.520494
1456	6	1-536	1452	13	68	726	35	2012-08-02 04:44:29.520494
1457	6	1-123	1452	68	91	562	93	2012-08-02 04:44:29.520494
1452	6	1-572	1452	91	89	2951	75	2012-08-02 04:44:29.520494
1459	8	1-133	1458	40	4	1121	55	2012-08-02 04:44:30.49312
1460	8	1-151	1458	4	93	516	64	2012-08-02 04:44:30.49312
1461	8	1-493	1458	93	37	841	99	2012-08-02 04:44:30.49312
1462	8	1-257	1458	37	47	548	29	2012-08-02 04:44:30.49312
1463	8	1-351	1458	47	19	504	5	2012-08-02 04:44:30.49312
1464	8	1-208	1458	19	85	656	13	2012-08-02 04:44:30.49312
1465	8	1-480	1458	85	44	1530	79	2012-08-02 04:44:30.49312
1458	8	1-577	1458	44	40	1059	92	2012-08-02 04:44:30.49312
1467	7	1-133	1466	40	4	1086	55	2012-08-02 04:44:30.49312
1468	7	1-151	1466	4	93	508	64	2012-08-02 04:44:30.49312
1469	7	1-493	1466	93	70	840	99	2012-08-02 04:44:30.49312
1470	7	1-169	1466	70	13	6778	56	2012-08-02 04:44:30.49312
1471	7	1-489	1466	13	40	1265	39	2012-08-02 04:44:30.49312
1472	7	1-332	1466	40	44	1417	79	2012-08-02 04:44:30.49312
1466	7	1-577	1466	44	40	995	92	2012-08-02 04:44:30.49312
1474	5	1-281	1473	71	52	1250	86	2012-08-02 04:44:31.963077
1475	5	1-229	1473	52	76	868	79	2012-08-02 04:44:31.963077
1476	5	1-161	1473	76	92	4592	31	2012-08-02 04:44:31.963077
1477	5	1-422	1473	92	66	1862	54	2012-08-02 04:44:31.963077
1473	5	1-584	1473	66	71	1014	15	2012-08-02 04:44:31.963077
15871	4	1-4417	15870	60	95	461	15	2012-08-02 05:13:09.230075
1479	8	1-349	1478	98	88	8031	12	2012-08-02 04:44:31.963077
1480	8	1-566	1478	88	94	3083	7	2012-08-02 04:44:31.963077
1481	8	1-469	1478	94	58	1004	60	2012-08-02 04:44:31.963077
1482	8	1-456	1478	58	26	937	3	2012-08-02 04:44:31.963077
1483	8	1-337	1478	26	72	188	84	2012-08-02 04:44:31.963077
1484	8	1-461	1478	72	51	834	48	2012-08-02 04:44:31.963077
1485	8	1-513	1478	51	66	3813	54	2012-08-02 04:44:31.963077
1478	8	1-584	1478	66	98	2292	15	2012-08-02 04:44:31.963077
1487	7	1-313	1486	46	11	845	44	2012-08-02 04:44:31.963077
1488	7	1-325	1486	11	40	1184	52	2012-08-02 04:44:31.963077
1489	7	1-517	1486	40	26	423	3	2012-08-02 04:44:31.963077
1490	7	1-337	1486	26	72	88	84	2012-08-02 04:44:31.963077
1491	7	1-461	1486	72	51	404	48	2012-08-02 04:44:31.963077
1492	7	1-513	1486	51	66	1913	54	2012-08-02 04:44:31.963077
1486	7	1-584	1486	66	46	1191	15	2012-08-02 04:44:31.963077
1494	5	1-119	1493	66	41	1292	44	2012-08-02 04:44:32.703615
1495	5	1-331	1493	41	38	310	100	2012-08-02 04:44:32.703615
1496	5	1-372	1493	38	82	353	69	2012-08-02 04:44:32.703615
1497	5	1-508	1493	82	62	201	73	2012-08-02 04:44:32.703615
1493	5	1-585	1493	62	66	7357	4	2012-08-02 04:44:32.703615
1499	8	1-402	1498	42	2	1396	95	2012-08-02 04:44:32.814162
1500	8	1-371	1498	2	83	431	20	2012-08-02 04:44:32.814162
1501	8	1-473	1498	83	91	808	12	2012-08-02 04:44:32.814162
1502	8	1-582	1498	91	90	1884	42	2012-08-02 04:44:32.814162
1503	8	1-522	1498	90	71	1927	70	2012-08-02 04:44:32.814162
1504	8	1-127	1498	71	55	3192	81	2012-08-02 04:44:32.814162
1505	8	1-544	1498	55	3	3441	34	2012-08-02 04:44:32.814162
1498	8	1-586	1498	3	42	2081	7	2012-08-02 04:44:32.814162
1507	3	1-408	1506	43	88	1469	26	2012-08-02 04:44:34.338852
1508	3	1-478	1506	88	9	2185	60	2012-08-02 04:44:34.338852
1506	3	1-591	1506	9	43	1645	80	2012-08-02 04:44:34.338852
1510	7	1-408	1509	43	80	1140	26	2012-08-02 04:44:34.338852
1511	7	1-261	1509	80	80	2257	59	2012-08-02 04:44:34.338852
1512	7	1-506	1509	80	13	1071	24	2012-08-02 04:44:34.338852
1513	7	1-571	1509	13	4	4342	8	2012-08-02 04:44:34.338852
1514	7	1-4	1509	4	94	3217	7	2012-08-02 04:44:34.338852
1515	7	1-469	1509	94	9	1013	60	2012-08-02 04:44:34.338852
1509	7	1-591	1509	9	43	987	80	2012-08-02 04:44:34.338852
1517	8	1-408	1516	43	96	1181	26	2012-08-02 04:44:34.338852
1518	8	1-404	1516	96	55	406	5	2012-08-02 04:44:34.338852
1519	8	1-510	1516	55	50	567	49	2012-08-02 04:44:34.338852
1520	8	1-207	1516	50	70	386	73	2012-08-02 04:44:34.338852
1521	8	1-441	1516	70	24	480	22	2012-08-02 04:44:34.338852
1522	8	1-458	1516	24	94	1991	7	2012-08-02 04:44:34.338852
1523	8	1-469	1516	94	9	745	60	2012-08-02 04:44:34.338852
1516	8	1-591	1516	9	43	861	80	2012-08-02 04:44:34.338852
1525	8	1-408	1524	43	33	48	26	2012-08-02 04:44:34.338852
1526	8	1-486	1524	33	9	80	17	2012-08-02 04:44:34.338852
1527	8	1-9	1524	9	32	45	16	2012-08-02 04:44:34.338852
1528	8	1-198	1524	32	6	26	67	2012-08-02 04:44:34.338852
1529	8	1-411	1524	6	3	74	34	2012-08-02 04:44:34.338852
1530	8	1-586	1524	3	94	72	7	2012-08-02 04:44:34.338852
1531	8	1-469	1524	94	9	28	60	2012-08-02 04:44:34.338852
1524	8	1-591	1524	9	43	34	80	2012-08-02 04:44:34.338852
1533	2	1-366	1532	58	80	2893	47	2012-08-02 04:44:34.893489
1532	2	1-593	1532	80	58	4641	2	2012-08-02 04:44:34.893489
1535	3	1-482	1534	37	100	2640	37	2012-08-02 04:44:35.601153
1536	3	1-556	1534	100	55	582	64	2012-08-02 04:44:35.601153
1534	3	1-594	1534	55	37	4248	1	2012-08-02 04:44:35.601153
1538	7	1-512	1537	4	93	1658	88	2012-08-02 04:44:35.601153
1539	7	1-291	1537	93	13	1344	89	2012-08-02 04:44:35.601153
1540	7	1-536	1537	13	52	1641	35	2012-08-02 04:44:35.601153
1541	7	1-99	1537	52	80	825	11	2012-08-02 04:44:35.601153
1542	7	1-576	1537	80	4	215	55	2012-08-02 04:44:35.601153
1543	7	1-151	1537	4	55	91	64	2012-08-02 04:44:35.601153
1537	7	1-594	1537	55	4	1832	1	2012-08-02 04:44:35.601153
1545	3	1-532	1544	38	94	1689	2	2012-08-02 04:44:35.821689
1546	3	1-394	1544	94	53	1653	93	2012-08-02 04:44:35.821689
1544	3	1-596	1544	53	38	4015	59	2012-08-02 04:44:35.821689
1548	6	1-506	1547	80	23	50	24	2012-08-02 04:44:35.821689
1549	6	1-569	1547	23	93	39	88	2012-08-02 04:44:35.821689
1550	6	1-291	1547	93	13	29	89	2012-08-02 04:44:35.821689
1551	6	1-536	1547	13	68	34	35	2012-08-02 04:44:35.821689
1552	6	1-123	1547	68	53	37	93	2012-08-02 04:44:35.821689
1547	6	1-596	1547	53	80	130	59	2012-08-02 04:44:35.821689
1553	1	1-532	\N	38	38	2789	2	2012-08-02 04:44:35.821689
1555	3	1-508	1554	82	99	1779	73	2012-08-02 04:44:36.694784
1556	3	1-398	1554	99	15	591	41	2012-08-02 04:44:36.694784
1554	3	1-599	1554	15	82	5872	69	2012-08-02 04:44:36.694784
1558	7	1-174	1557	53	41	1710	24	2012-08-02 04:44:37.125811
1559	7	1-52	1557	41	12	1451	59	2012-08-02 04:44:37.125811
1560	7	1-562	1557	12	64	4329	46	2012-08-02 04:44:37.125811
1561	7	1-526	1557	64	57	4205	43	2012-08-02 04:44:37.125811
1562	7	1-476	1557	57	10	3001	65	2012-08-02 04:44:37.125811
1563	7	1-283	1557	10	58	2026	20	2012-08-02 04:44:37.125811
1557	7	1-602	1557	58	53	2769	58	2012-08-02 04:44:37.125811
1565	8	1-191	1564	46	16	1096	74	2012-08-02 04:44:37.125811
1566	8	1-425	1564	16	72	853	30	2012-08-02 04:44:37.125811
1567	8	1-491	1564	72	43	512	80	2012-08-02 04:44:37.125811
1568	8	1-408	1564	43	96	699	26	2012-08-02 04:44:37.125811
1569	8	1-404	1564	96	55	240	5	2012-08-02 04:44:37.125811
1570	8	1-510	1564	55	26	333	49	2012-08-02 04:44:37.125811
1571	8	1-416	1564	26	58	472	20	2012-08-02 04:44:37.125811
1564	8	1-602	1564	58	46	769	58	2012-08-02 04:44:37.125811
1573	6	1-437	1572	41	57	3805	43	2012-08-02 04:44:37.87784
1572	6	1-608	1572	74	41	3843	29	2012-08-02 04:44:37.87784
1574	6	1-476	1572	57	10	2411	65	2012-08-02 04:44:37.87784
1575	6	1-283	1572	10	58	1445	20	2012-08-02 04:44:37.87784
1576	6	1-602	1572	58	53	1753	58	2012-08-02 04:44:37.87784
1577	6	1-174	1572	53	74	961	24	2012-08-02 04:44:37.87784
15872	4	1-4508	15870	95	74	433	82	2012-08-02 05:13:09.230075
1579	6	1-484	1578	61	80	204	59	2012-08-02 04:44:38.518643
1580	6	1-506	1578	80	74	96	24	2012-08-02 04:44:38.518643
1581	6	1-608	1578	74	46	483	29	2012-08-02 04:44:38.518643
1582	6	1-356	1578	46	84	1195	78	2012-08-02 04:44:38.518643
1583	6	1-452	1578	84	49	364	30	2012-08-02 04:44:38.518643
1578	6	1-610	1578	49	61	505	14	2012-08-02 04:44:38.518643
1585	7	1-534	1584	26	41	310	44	2012-08-02 04:44:39.42509
1586	7	1-331	1584	41	38	92	100	2012-08-02 04:44:39.42509
1587	7	1-372	1584	38	82	130	69	2012-08-02 04:44:39.42509
1588	7	1-508	1584	82	45	91	73	2012-08-02 04:44:39.42509
1589	7	1-296	1584	45	71	228	53	2012-08-02 04:44:39.42509
1590	7	1-413	1584	71	13	252	61	2012-08-02 04:44:39.42509
1584	7	1-613	1584	13	26	524	20	2012-08-02 04:44:39.42509
1592	5	1-526	1591	64	57	460	43	2012-08-02 04:44:39.513477
1593	5	1-476	1591	57	51	348	65	2012-08-02 04:44:39.513477
1594	5	1-71	1591	51	13	218	21	2012-08-02 04:44:39.513477
1595	5	1-13	1591	13	11	103	3	2012-08-02 04:44:39.513477
1591	5	1-614	1591	11	64	446	46	2012-08-02 04:44:39.513477
1597	6	1-526	1596	64	57	378	43	2012-08-02 04:44:39.513477
1598	6	1-476	1596	57	51	301	65	2012-08-02 04:44:39.513477
1599	6	1-71	1596	51	9	198	21	2012-08-02 04:44:39.513477
1600	6	1-483	1596	9	43	88	99	2012-08-02 04:44:39.513477
1601	6	1-546	1596	43	11	77	3	2012-08-02 04:44:39.513477
1596	6	1-614	1596	11	64	349	46	2012-08-02 04:44:39.513477
1603	7	1-508	1602	82	70	393	73	2012-08-02 04:44:39.944705
1604	7	1-441	1602	70	36	453	22	2012-08-02 04:44:39.944705
1605	7	1-301	1602	36	64	719	46	2012-08-02 04:44:39.944705
1606	7	1-526	1602	64	84	773	43	2012-08-02 04:44:39.944705
1607	7	1-375	1602	84	19	1030	17	2012-08-02 04:44:39.944705
1608	7	1-108	1602	19	43	2305	6	2012-08-02 04:44:39.944705
1602	7	1-616	1602	43	82	624	69	2012-08-02 04:44:39.944705
1610	8	1-506	1609	80	74	29	24	2012-08-02 04:44:40.34241
1611	8	1-608	1609	74	10	122	29	2012-08-02 04:44:40.34241
1612	8	1-565	1609	10	18	31	28	2012-08-02 04:44:40.34241
1613	8	1-18	1609	18	11	22	3	2012-08-02 04:44:40.34241
1614	8	1-614	1609	11	64	83	46	2012-08-02 04:44:40.34241
1615	8	1-526	1609	64	84	75	43	2012-08-02 04:44:40.34241
1616	8	1-375	1609	84	69	85	17	2012-08-02 04:44:40.34241
1609	8	1-618	1609	69	80	74	59	2012-08-02 04:44:40.34241
1618	7	1-361	1617	96	13	631	61	2012-08-02 04:44:40.784449
1619	7	1-613	1617	13	58	754	20	2012-08-02 04:44:40.784449
1620	7	1-602	1617	58	53	732	58	2012-08-02 04:44:40.784449
1621	7	1-174	1617	53	23	321	24	2012-08-02 04:44:40.784449
1622	7	1-569	1617	23	27	194	88	2012-08-02 04:44:40.784449
1623	7	1-563	1617	27	17	180	70	2012-08-02 04:44:40.784449
1617	7	1-619	1617	17	96	860	76	2012-08-02 04:44:40.784449
1625	8	1-153	1624	79	82	1891	60	2012-08-02 04:44:40.784449
1626	8	1-159	1624	82	26	216	32	2012-08-02 04:44:40.784449
1627	8	1-540	1624	26	88	104	92	2012-08-02 04:44:40.784449
1628	8	1-521	1624	88	89	64	75	2012-08-02 04:44:40.784449
1629	8	1-365	1624	89	50	60	77	2012-08-02 04:44:40.784449
1630	8	1-128	1624	50	27	82	88	2012-08-02 04:44:40.784449
1631	8	1-563	1624	27	17	101	70	2012-08-02 04:44:40.784449
1624	8	1-619	1624	17	79	645	76	2012-08-02 04:44:40.784449
1633	8	1-475	1632	96	37	3716	66	2012-08-02 04:44:40.784449
1634	8	1-339	1632	37	52	4005	86	2012-08-02 04:44:40.784449
1635	8	1-229	1632	52	44	3104	79	2012-08-02 04:44:40.784449
1636	8	1-577	1632	44	46	1885	92	2012-08-02 04:44:40.784449
1637	8	1-145	1632	46	23	425	24	2012-08-02 04:44:40.784449
1638	8	1-569	1632	23	27	380	88	2012-08-02 04:44:40.784449
1639	8	1-563	1632	27	17	523	70	2012-08-02 04:44:40.784449
1632	8	1-619	1632	17	96	3709	76	2012-08-02 04:44:40.784449
1641	6	1-361	1640	96	89	2933	61	2012-08-02 04:44:40.784449
1642	6	1-439	1640	89	19	4521	45	2012-08-02 04:44:40.784449
1643	6	1-551	1640	19	23	206	24	2012-08-02 04:44:40.784449
1644	6	1-569	1640	23	27	204	88	2012-08-02 04:44:40.784449
1645	6	1-563	1640	27	17	311	70	2012-08-02 04:44:40.784449
1640	6	1-619	1640	17	96	2437	76	2012-08-02 04:44:40.784449
1647	7	1-285	1646	34	11	62	3	2012-08-02 04:44:40.784449
1648	7	1-614	1646	11	30	299	46	2012-08-02 04:44:40.784449
1649	7	1-32	1646	30	19	118	45	2012-08-02 04:44:40.784449
1650	7	1-551	1646	19	23	5	24	2012-08-02 04:44:40.784449
1651	7	1-569	1646	23	27	5	88	2012-08-02 04:44:40.784449
1652	7	1-563	1646	27	17	8	70	2012-08-02 04:44:40.784449
1646	7	1-619	1646	17	34	67	76	2012-08-02 04:44:40.784449
1654	6	1-472	1653	14	49	177	94	2012-08-02 04:44:42.321069
1655	6	1-442	1653	49	64	286	97	2012-08-02 04:44:42.321069
1656	6	1-477	1653	64	88	380	12	2012-08-02 04:44:42.321069
1657	6	1-566	1653	88	42	159	7	2012-08-02 04:44:42.321069
1658	6	1-402	1653	42	39	157	95	2012-08-02 04:44:42.321069
1653	6	1-627	1653	39	14	307	68	2012-08-02 04:44:42.321069
1660	7	1-472	1659	14	49	1229	94	2012-08-02 04:44:42.321069
1661	7	1-442	1659	49	49	2122	97	2012-08-02 04:44:42.321069
1662	7	1-466	1659	49	88	1125	18	2012-08-02 04:44:42.321069
1663	7	1-519	1659	88	3	1116	34	2012-08-02 04:44:42.321069
1664	7	1-586	1659	3	42	1018	7	2012-08-02 04:44:42.321069
1665	7	1-402	1659	42	39	1030	95	2012-08-02 04:44:42.321069
1659	7	1-627	1659	39	14	2075	68	2012-08-02 04:44:42.321069
1667	2	1-582	1666	91	70	6524	42	2012-08-02 04:44:43.337773
1666	2	1-632	1666	70	91	6218	12	2012-08-02 04:44:43.337773
1669	8	1-158	1668	61	84	114	98	2012-08-02 04:44:43.890326
1670	8	1-226	1668	84	58	446	20	2012-08-02 04:44:43.890326
1671	8	1-602	1668	58	46	747	58	2012-08-02 04:44:43.890326
1672	8	1-191	1668	46	16	1094	74	2012-08-02 04:44:43.890326
1673	8	1-425	1668	16	75	876	30	2012-08-02 04:44:43.890326
1674	8	1-438	1668	75	80	279	11	2012-08-02 04:44:43.890326
1675	8	1-576	1668	80	92	92	55	2012-08-02 04:44:43.890326
1668	8	1-633	1668	92	61	160	66	2012-08-02 04:44:43.890326
1677	5	1-378	1676	27	48	335	38	2012-08-02 04:44:44.255143
1678	5	1-570	1676	48	59	229	10	2012-08-02 04:44:44.255143
1676	5	1-635	1676	5	27	1246	57	2012-08-02 04:44:44.255143
1679	5	1-369	1676	59	51	141	48	2012-08-02 04:44:44.255143
1680	5	1-513	1676	51	5	526	54	2012-08-02 04:44:44.255143
15873	4	1-4469	15870	74	2	900	69	2012-08-02 05:13:09.230075
1682	4	1-480	1681	85	38	5	79	2012-08-02 04:44:44.918197
1683	4	1-617	1681	38	8	6	34	2012-08-02 04:44:44.918197
1684	4	1-487	1681	8	76	1	32	2012-08-02 04:44:44.918197
1681	4	1-636	1681	76	85	7	13	2012-08-02 04:44:44.918197
15870	4	1-4569	15870	2	60	564	67	2012-08-02 05:13:09.230075
1686	7	1-480	1685	85	44	4876	79	2012-08-02 04:44:44.918197
1687	7	1-577	1685	44	40	2443	92	2012-08-02 04:44:44.918197
1688	7	1-133	1685	40	48	1870	55	2012-08-02 04:44:44.918197
1689	7	1-67	1685	48	4	2479	51	2012-08-02 04:44:44.918197
1690	7	1-82	1685	4	7	131	69	2012-08-02 04:44:44.918197
1691	7	1-60	1685	7	76	148	32	2012-08-02 04:44:44.918197
1685	7	1-636	1685	76	85	2888	13	2012-08-02 04:44:44.918197
1693	8	1-530	1692	84	66	1559	3	2012-08-02 04:44:44.918197
1694	8	1-630	1692	66	3	1176	34	2012-08-02 04:44:44.918197
1695	8	1-586	1692	3	63	1137	7	2012-08-02 04:44:44.918197
1696	8	1-625	1692	63	32	454	41	2012-08-02 04:44:44.918197
1697	8	1-36	1692	32	4	546	51	2012-08-02 04:44:44.918197
1698	8	1-82	1692	4	7	41	69	2012-08-02 04:44:44.918197
1699	8	1-60	1692	7	76	66	32	2012-08-02 04:44:44.918197
1692	8	1-636	1692	76	84	1834	13	2012-08-02 04:44:44.918197
1701	7	1-86	1700	44	92	605	55	2012-08-02 04:44:45.669831
1702	7	1-633	1700	92	61	806	66	2012-08-02 04:44:45.669831
1703	7	1-158	1700	61	84	438	98	2012-08-02 04:44:45.669831
1704	7	1-226	1700	84	58	1315	20	2012-08-02 04:44:45.669831
1705	7	1-602	1700	58	23	1684	58	2012-08-02 04:44:45.669831
1706	7	1-488	1700	23	68	1987	41	2012-08-02 04:44:45.669831
1700	7	1-639	1700	68	44	1209	53	2012-08-02 04:44:45.669831
1708	5	1-558	1707	41	2	6315	95	2012-08-02 04:44:46.067443
1709	5	1-371	1707	2	4	1740	20	2012-08-02 04:44:46.067443
1710	5	1-527	1707	4	78	4713	23	2012-08-02 04:44:46.067443
1711	5	1-419	1707	78	12	2972	50	2012-08-02 04:44:46.067443
1707	5	1-643	1707	12	41	4590	22	2012-08-02 04:44:46.067443
1713	5	1-301	1712	36	64	709	46	2012-08-02 04:44:46.067443
1714	5	1-526	1712	64	30	705	43	2012-08-02 04:44:46.067443
1715	5	1-267	1712	30	44	124	90	2012-08-02 04:44:46.067443
1716	5	1-552	1712	44	12	211	50	2012-08-02 04:44:46.067443
1712	5	1-643	1712	12	36	483	22	2012-08-02 04:44:46.067443
1718	8	1-226	1717	84	4	257	20	2012-08-02 04:44:46.355044
1719	8	1-527	1717	4	78	1107	23	2012-08-02 04:44:46.355044
1720	8	1-419	1717	78	12	1111	50	2012-08-02 04:44:46.355044
1721	8	1-643	1717	12	32	2732	22	2012-08-02 04:44:46.355044
1722	8	1-265	1717	32	3	723	34	2012-08-02 04:44:46.355044
1723	8	1-586	1717	3	18	621	7	2012-08-02 04:44:46.355044
1724	8	1-533	1717	18	74	101	16	2012-08-02 04:44:46.355044
1717	8	1-644	1717	74	84	73	98	2012-08-02 04:44:46.355044
1726	8	1-226	1725	84	26	126	20	2012-08-02 04:44:46.355044
1727	8	1-534	1725	26	39	65	44	2012-08-02 04:44:46.355044
1728	8	1-497	1725	39	11	15	3	2012-08-02 04:44:46.355044
1729	8	1-614	1725	11	64	65	46	2012-08-02 04:44:46.355044
1730	8	1-526	1725	64	84	68	43	2012-08-02 04:44:46.355044
1731	8	1-375	1725	84	9	96	17	2012-08-02 04:44:46.355044
1732	8	1-9	1725	9	74	47	16	2012-08-02 04:44:46.355044
1725	8	1-644	1725	74	84	34	98	2012-08-02 04:44:46.355044
1734	5	1-226	1733	84	26	1314	20	2012-08-02 04:44:46.355044
1735	5	1-534	1733	26	4	697	44	2012-08-02 04:44:46.355044
1736	5	1-573	1733	4	9	1007	17	2012-08-02 04:44:46.355044
1737	5	1-9	1733	9	74	505	16	2012-08-02 04:44:46.355044
1733	5	1-644	1733	74	84	369	98	2012-08-02 04:44:46.355044
1739	8	1-636	1738	76	84	1095	13	2012-08-02 04:44:46.963184
1740	8	1-530	1738	84	11	668	3	2012-08-02 04:44:46.963184
1741	8	1-614	1738	11	96	2417	46	2012-08-02 04:44:46.963184
1742	8	1-537	1738	96	51	835	48	2012-08-02 04:44:46.963184
1743	8	1-513	1738	51	59	3253	54	2012-08-02 04:44:46.963184
1744	8	1-555	1738	59	38	91	83	2012-08-02 04:44:46.963184
1745	8	1-470	1738	38	3	50	39	2012-08-02 04:44:46.963184
1738	8	1-647	1738	3	76	55	32	2012-08-02 04:44:46.963184
1747	6	1-636	1746	76	84	1553	13	2012-08-02 04:44:46.963184
1748	6	1-530	1746	84	90	1276	3	2012-08-02 04:44:46.963184
1749	6	1-454	1746	90	59	1403	54	2012-08-02 04:44:46.963184
1750	6	1-555	1746	59	38	53	83	2012-08-02 04:44:46.963184
1751	6	1-470	1746	38	3	39	39	2012-08-02 04:44:46.963184
1746	6	1-647	1746	3	76	58	32	2012-08-02 04:44:46.963184
1753	6	1-609	1752	62	16	2366	74	2012-08-02 04:44:47.173171
1754	6	1-425	1752	16	75	1427	30	2012-08-02 04:44:47.173171
1755	6	1-438	1752	75	51	343	11	2012-08-02 04:44:47.173171
1756	6	1-628	1752	51	29	3582	90	2012-08-02 04:44:47.173171
1757	6	1-545	1752	29	38	600	10	2012-08-02 04:44:47.173171
1752	6	1-648	1752	38	62	2752	4	2012-08-02 04:44:47.173171
1759	7	1-287	1758	92	9	1415	40	2012-08-02 04:44:47.173171
1760	7	1-395	1758	9	19	1253	45	2012-08-02 04:44:47.173171
1761	7	1-551	1758	19	23	59	24	2012-08-02 04:44:47.173171
1762	7	1-569	1758	23	90	60	88	2012-08-02 04:44:47.173171
1763	7	1-629	1758	90	5	354	18	2012-08-02 04:44:47.173171
1764	7	1-341	1758	5	38	328	10	2012-08-02 04:44:47.173171
1758	7	1-648	1758	38	92	1954	4	2012-08-02 04:44:47.173171
1766	4	1-442	1765	49	64	119	97	2012-08-02 04:44:48.035275
1767	4	1-477	1765	64	91	150	12	2012-08-02 04:44:48.035275
1768	4	1-582	1765	91	32	491	42	2012-08-02 04:44:48.035275
1765	4	1-649	1765	32	49	77	94	2012-08-02 04:44:48.035275
1770	7	1-577	1769	44	88	845	92	2012-08-02 04:44:49.647927
1771	7	1-521	1769	88	89	681	75	2012-08-02 04:44:49.647927
1772	7	1-365	1769	89	22	818	77	2012-08-02 04:44:49.647927
1773	7	1-322	1769	22	96	348	26	2012-08-02 04:44:49.647927
1774	7	1-404	1769	96	55	123	5	2012-08-02 04:44:49.647927
1775	7	1-510	1769	55	90	177	49	2012-08-02 04:44:49.647927
1769	7	1-653	1769	90	44	1188	79	2012-08-02 04:44:49.647927
1777	8	1-617	1776	38	3	2378	34	2012-08-02 04:44:49.647927
1778	8	1-586	1776	3	63	2295	7	2012-08-02 04:44:49.647927
1779	8	1-625	1776	63	32	915	41	2012-08-02 04:44:49.647927
1780	8	1-36	1776	32	85	1098	51	2012-08-02 04:44:49.647927
1781	8	1-457	1776	85	96	155	26	2012-08-02 04:44:49.647927
1782	8	1-404	1776	96	55	55	5	2012-08-02 04:44:49.647927
1783	8	1-510	1776	55	90	79	49	2012-08-02 04:44:49.647927
1776	8	1-653	1776	90	38	531	79	2012-08-02 04:44:49.647927
1784	1	1-510	\N	55	55	1843	49	2012-08-02 04:44:49.647927
1786	8	1-643	1785	12	41	84	22	2012-08-02 04:44:50.16741
1787	8	1-558	1785	41	35	204	95	2012-08-02 04:44:50.16741
1788	8	1-620	1785	35	92	185	31	2012-08-02 04:44:50.16741
1789	8	1-422	1785	92	62	97	54	2012-08-02 04:44:50.16741
1790	8	1-654	1785	62	18	18	28	2012-08-02 04:44:50.16741
1791	8	1-18	1785	18	11	17	3	2012-08-02 04:44:50.16741
1792	8	1-614	1785	11	78	85	46	2012-08-02 04:44:50.16741
1785	8	1-655	1785	78	12	31	50	2012-08-02 04:44:50.16741
15875	4	1-4028	15874	82	6	5222	36	2012-08-02 05:13:09.230075
1794	6	1-457	1793	85	94	172	26	2012-08-02 04:44:50.357223
1795	6	1-311	1793	94	44	129	53	2012-08-02 04:44:50.357223
1796	6	1-86	1793	44	4	86	55	2012-08-02 04:44:50.357223
1797	6	1-151	1793	4	37	41	64	2012-08-02 04:44:50.357223
1798	6	1-590	1793	37	23	139	7	2012-08-02 04:44:50.357223
1793	6	1-656	1793	23	85	1207	51	2012-08-02 04:44:50.357223
15876	4	1-4548	15874	6	84	8089	53	2012-08-02 05:13:09.230075
1800	6	1-612	1799	95	41	718	22	2012-08-02 04:44:50.456909
1801	6	1-558	1799	41	39	1368	95	2012-08-02 04:44:50.456909
1802	6	1-627	1799	39	14	2259	68	2012-08-02 04:44:50.456909
1803	6	1-472	1799	14	17	1096	94	2012-08-02 04:44:50.456909
1804	6	1-463	1799	17	93	683	93	2012-08-02 04:44:50.456909
1799	6	1-657	1799	93	95	1803	16	2012-08-02 04:44:50.456909
1806	7	1-644	1805	74	54	276	98	2012-08-02 04:44:50.456909
1807	7	1-631	1805	54	41	162	22	2012-08-02 04:44:50.456909
1808	7	1-558	1805	41	39	312	95	2012-08-02 04:44:50.456909
1809	7	1-627	1805	39	14	522	68	2012-08-02 04:44:50.456909
1810	7	1-472	1805	14	17	257	94	2012-08-02 04:44:50.456909
1811	7	1-463	1805	17	93	162	93	2012-08-02 04:44:50.456909
1805	7	1-657	1805	93	74	434	16	2012-08-02 04:44:50.456909
1813	7	1-644	1812	74	61	2671	98	2012-08-02 04:44:50.456909
1814	7	1-266	1812	61	23	1521	58	2012-08-02 04:44:50.456909
1815	7	1-488	1812	23	68	2174	41	2012-08-02 04:44:50.456909
1816	7	1-639	1812	68	29	1602	53	2012-08-02 04:44:50.456909
1817	7	1-450	1812	29	17	1545	94	2012-08-02 04:44:50.456909
1818	7	1-463	1812	17	93	1142	93	2012-08-02 04:44:50.456909
1812	7	1-657	1812	93	74	3580	16	2012-08-02 04:44:50.456909
1820	8	1-198	1819	32	6	234	67	2012-08-02 04:44:50.456909
1821	8	1-411	1819	6	93	632	34	2012-08-02 04:44:50.456909
1822	8	1-479	1819	93	44	206	90	2012-08-02 04:44:50.456909
1823	8	1-552	1819	44	67	401	50	2012-08-02 04:44:50.456909
1824	8	1-120	1819	67	87	113	36	2012-08-02 04:44:50.456909
1825	8	1-233	1819	87	17	171	94	2012-08-02 04:44:50.456909
1826	8	1-463	1819	17	93	130	93	2012-08-02 04:44:50.456909
1819	8	1-657	1819	93	32	435	16	2012-08-02 04:44:50.456909
1827	1	1-198	\N	32	32	316	67	2012-08-02 04:44:50.456909
1829	7	1-9	1828	9	39	1506	16	2012-08-02 04:44:50.899685
1830	7	1-148	1828	39	23	861	87	2012-08-02 04:44:50.899685
1831	7	1-605	1828	23	62	3332	4	2012-08-02 04:44:50.899685
1832	7	1-609	1828	62	10	3151	74	2012-08-02 04:44:50.899685
1833	7	1-645	1828	10	5	2110	54	2012-08-02 04:44:50.899685
1834	7	1-635	1828	5	42	5974	57	2012-08-02 04:44:50.899685
1828	7	1-660	1828	42	9	3283	17	2012-08-02 04:44:50.899685
1836	8	1-616	1835	43	82	12	69	2012-08-02 04:44:52.037977
1837	8	1-508	1835	82	62	8	73	2012-08-02 04:44:52.037977
1838	8	1-585	1835	62	62	357	4	2012-08-02 04:44:52.037977
1839	8	1-609	1835	62	10	405	74	2012-08-02 04:44:52.037977
1840	8	1-645	1835	10	59	315	54	2012-08-02 04:44:52.037977
1841	8	1-555	1835	59	38	12	83	2012-08-02 04:44:52.037977
1842	8	1-470	1835	38	32	9	39	2012-08-02 04:44:52.037977
1835	8	1-668	1835	32	43	41	6	2012-08-02 04:44:52.037977
1844	6	1-609	1843	62	10	396	74	2012-08-02 04:44:52.347639
1845	6	1-645	1843	10	5	263	54	2012-08-02 04:44:52.347639
1846	6	1-635	1843	5	34	741	57	2012-08-02 04:44:52.347639
1847	6	1-42	1843	34	6	507	47	2012-08-02 04:44:52.347639
1848	6	1-524	1843	6	97	60	31	2012-08-02 04:44:52.347639
1843	6	1-671	1843	97	62	421	4	2012-08-02 04:44:52.347639
1850	7	1-609	1849	62	10	1388	74	2012-08-02 04:44:52.347639
1851	7	1-645	1849	10	62	1059	54	2012-08-02 04:44:52.347639
1852	7	1-654	1849	62	24	199	28	2012-08-02 04:44:52.347639
1853	7	1-604	1849	24	7	1707	19	2012-08-02 04:44:52.347639
1854	7	1-33	1849	7	6	1130	47	2012-08-02 04:44:52.347639
1855	7	1-524	1849	6	97	154	31	2012-08-02 04:44:52.347639
1849	7	1-671	1849	97	62	1235	4	2012-08-02 04:44:52.347639
1857	6	1-512	1856	4	27	178	88	2012-08-02 04:44:53.165327
1858	6	1-563	1856	27	17	164	70	2012-08-02 04:44:53.165327
1859	6	1-619	1856	17	34	776	76	2012-08-02 04:44:53.165327
1860	6	1-285	1856	34	11	428	3	2012-08-02 04:44:53.165327
1861	6	1-614	1856	11	98	1229	46	2012-08-02 04:44:53.165327
1856	6	1-673	1856	98	4	308	1	2012-08-02 04:44:53.165327
1863	6	1-289	1862	94	13	238	89	2012-08-02 04:44:53.839678
1864	6	1-536	1862	13	55	346	35	2012-08-02 04:44:53.839678
1865	6	1-84	1862	55	40	4753	52	2012-08-02 04:44:53.839678
1866	6	1-517	1862	40	90	1822	3	2012-08-02 04:44:53.839678
1867	6	1-454	1862	90	85	1940	54	2012-08-02 04:44:53.839678
1862	6	1-676	1862	85	94	1489	96	2012-08-02 04:44:53.839678
1869	7	1-60	1868	7	26	899	32	2012-08-02 04:44:54.128286
1870	7	1-540	1868	26	88	424	92	2012-08-02 04:44:54.128286
1871	7	1-521	1868	88	89	257	75	2012-08-02 04:44:54.128286
1872	7	1-365	1868	89	50	232	77	2012-08-02 04:44:54.128286
1873	7	1-128	1868	50	27	312	88	2012-08-02 04:44:54.128286
1874	7	1-563	1868	27	11	377	70	2012-08-02 04:44:54.128286
1868	7	1-678	1868	11	7	745	69	2012-08-02 04:44:54.128286
1876	6	1-622	1875	42	71	1200	15	2012-08-02 04:44:54.128286
1877	6	1-281	1875	71	52	1781	86	2012-08-02 04:44:54.128286
1878	6	1-229	1875	52	47	1491	79	2012-08-02 04:44:54.128286
1879	6	1-669	1875	47	27	1269	88	2012-08-02 04:44:54.128286
1880	6	1-563	1875	27	11	1885	70	2012-08-02 04:44:54.128286
1875	6	1-678	1875	11	42	4576	69	2012-08-02 04:44:54.128286
1882	6	1-60	1881	7	26	89	32	2012-08-02 04:44:54.128286
1883	6	1-540	1881	26	46	54	92	2012-08-02 04:44:54.128286
1884	6	1-145	1881	46	23	14	24	2012-08-02 04:44:54.128286
1885	6	1-569	1881	23	27	14	88	2012-08-02 04:44:54.128286
1886	6	1-563	1881	27	11	22	70	2012-08-02 04:44:54.128286
1881	6	1-678	1881	11	7	57	69	2012-08-02 04:44:54.128286
15877	4	1-4514	15874	84	2	6545	69	2012-08-02 05:13:09.230075
1888	7	1-373	1887	23	88	1524	18	2012-08-02 04:44:54.249124
1889	7	1-519	1887	88	93	1630	34	2012-08-02 04:44:54.249124
1890	7	1-479	1887	93	44	552	90	2012-08-02 04:44:54.249124
1891	7	1-552	1887	44	67	1158	50	2012-08-02 04:44:54.249124
1892	7	1-120	1887	67	87	353	36	2012-08-02 04:44:54.249124
1893	7	1-233	1887	87	4	574	94	2012-08-02 04:44:54.249124
1887	7	1-679	1887	4	23	692	87	2012-08-02 04:44:54.249124
15874	4	1-4569	15874	2	82	5682	67	2012-08-02 05:13:09.230075
1895	5	1-633	1894	92	61	2152	66	2012-08-02 04:44:55.110977
1896	5	1-158	1894	61	52	993	98	2012-08-02 04:44:55.110977
1897	5	1-363	1894	52	72	820	84	2012-08-02 04:44:55.110977
1898	5	1-461	1894	72	46	2771	48	2012-08-02 04:44:55.110977
1894	5	1-684	1894	46	92	1902	55	2012-08-02 04:44:55.110977
1900	8	1-67	1899	48	85	721	51	2012-08-02 04:44:55.110977
1901	8	1-457	1899	85	17	76	26	2012-08-02 04:44:55.110977
1902	8	1-211	1899	17	52	16	86	2012-08-02 04:44:55.110977
1903	8	1-229	1899	52	47	11	79	2012-08-02 04:44:55.110977
1904	8	1-669	1899	47	9	8	88	2012-08-02 04:44:55.110977
1905	8	1-637	1899	9	72	165	84	2012-08-02 04:44:55.110977
1906	8	1-461	1899	72	46	645	48	2012-08-02 04:44:55.110977
1899	8	1-684	1899	46	48	513	55	2012-08-02 04:44:55.110977
1908	8	1-151	1907	4	80	410	64	2012-08-02 04:44:55.110977
1909	8	1-670	1907	80	68	2834	41	2012-08-02 04:44:55.110977
1910	8	1-639	1907	68	21	2210	53	2012-08-02 04:44:55.110977
1911	8	1-37	1907	21	40	1829	52	2012-08-02 04:44:55.110977
1912	8	1-517	1907	40	26	727	3	2012-08-02 04:44:55.110977
1913	8	1-337	1907	26	72	168	84	2012-08-02 04:44:55.110977
1914	8	1-461	1907	72	46	859	48	2012-08-02 04:44:55.110977
1907	8	1-684	1907	46	4	890	55	2012-08-02 04:44:55.110977
1916	6	1-657	1915	93	41	150	16	2012-08-02 04:44:55.82976
1917	6	1-307	1915	41	97	59	31	2012-08-02 04:44:55.82976
1918	6	1-671	1915	97	92	442	4	2012-08-02 04:44:55.82976
1919	6	1-287	1915	92	13	286	40	2012-08-02 04:44:55.82976
1920	6	1-201	1915	13	36	253	12	2012-08-02 04:44:55.82976
1915	6	1-686	1915	36	93	51	93	2012-08-02 04:44:55.82976
1921	1	1-201	\N	13	13	3355	12	2012-08-02 04:44:55.82976
1923	5	1-569	1922	23	27	1777	88	2012-08-02 04:44:57.144742
1924	5	1-563	1922	27	11	2110	70	2012-08-02 04:44:57.144742
1925	5	1-678	1922	11	7	4092	69	2012-08-02 04:44:57.144742
1926	5	1-60	1922	7	90	4845	32	2012-08-02 04:44:57.144742
1922	5	1-693	1922	90	23	2302	24	2012-08-02 04:44:57.144742
1927	1	1-563	\N	27	27	3199	70	2012-08-02 04:44:57.144742
1929	2	1-256	1928	81	37	20	27	2012-08-02 04:44:59.343953
1928	2	1-703	1928	37	81	32	100	2012-08-02 04:44:59.343953
1931	7	1-372	1930	38	82	2740	69	2012-08-02 04:44:59.343953
1932	7	1-508	1930	82	45	1406	73	2012-08-02 04:44:59.343953
1933	7	1-296	1930	45	21	2551	53	2012-08-02 04:44:59.343953
1934	7	1-37	1930	21	13	1585	52	2012-08-02 04:44:59.343953
1935	7	1-696	1930	13	36	998	1	2012-08-02 04:44:59.343953
1936	7	1-603	1930	36	37	1535	27	2012-08-02 04:44:59.343953
1930	7	1-703	1930	37	38	2671	100	2012-08-02 04:44:59.343953
1938	7	1-565	1937	10	49	158	28	2012-08-02 04:45:00.658781
1939	7	1-72	1937	49	34	44	8	2012-08-02 04:45:00.658781
1940	7	1-702	1937	34	51	25	11	2012-08-02 04:45:00.658781
1941	7	1-628	1937	51	44	278	90	2012-08-02 04:45:00.658781
1942	7	1-552	1937	44	12	462	50	2012-08-02 04:45:00.658781
1943	7	1-643	1937	12	73	1028	22	2012-08-02 04:45:00.658781
1937	7	1-709	1937	73	10	587	29	2012-08-02 04:45:00.658781
1945	7	1-158	1944	61	52	678	98	2012-08-02 04:45:01.951464
1946	7	1-363	1944	52	72	627	84	2012-08-02 04:45:01.951464
1947	7	1-461	1944	72	46	2371	48	2012-08-02 04:45:01.951464
1948	7	1-684	1944	46	43	1821	55	2012-08-02 04:45:01.951464
1949	7	1-55	1944	43	63	701	43	2012-08-02 04:45:01.951464
1950	7	1-704	1944	63	69	1747	87	2012-08-02 04:45:01.951464
1944	7	1-712	1944	69	61	1313	66	2012-08-02 04:45:01.951464
1952	8	1-711	1951	92	42	294	2	2012-08-02 04:45:01.951464
1953	8	1-122	1951	42	67	1044	50	2012-08-02 04:45:01.951464
1954	8	1-120	1951	67	87	291	36	2012-08-02 04:45:01.951464
1955	8	1-233	1951	87	17	432	94	2012-08-02 04:45:01.951464
1956	8	1-463	1951	17	93	324	93	2012-08-02 04:45:01.951464
1957	8	1-657	1951	93	39	1030	16	2012-08-02 04:45:01.951464
1958	8	1-148	1951	39	69	667	87	2012-08-02 04:45:01.951464
1951	8	1-712	1951	69	92	648	66	2012-08-02 04:45:01.951464
1960	8	1-283	1959	10	26	1487	20	2012-08-02 04:45:02.735946
1961	8	1-534	1959	26	11	848	44	2012-08-02 04:45:02.735946
1962	8	1-325	1959	11	13	1314	52	2012-08-02 04:45:02.735946
1963	8	1-696	1959	13	75	1096	1	2012-08-02 04:45:02.735946
1964	8	1-695	1959	75	96	1106	26	2012-08-02 04:45:02.735946
1965	8	1-404	1959	96	19	380	5	2012-08-02 04:45:02.735946
1966	8	1-208	1959	19	41	491	13	2012-08-02 04:45:02.735946
1959	8	1-713	1959	41	10	1844	65	2012-08-02 04:45:02.735946
1967	1	1-325	\N	11	11	2998	52	2012-08-02 04:45:02.735946
1969	8	1-668	1968	32	43	1573	6	2012-08-02 04:45:02.846581
1970	8	1-616	1968	43	7	427	69	2012-08-02 04:45:02.846581
1971	8	1-60	1968	7	90	618	32	2012-08-02 04:45:02.846581
1972	8	1-693	1968	90	74	359	24	2012-08-02 04:45:02.846581
1973	8	1-608	1968	74	10	1795	29	2012-08-02 04:45:02.846581
1974	8	1-565	1968	10	24	539	28	2012-08-02 04:45:02.846581
1975	8	1-604	1968	24	59	4426	19	2012-08-02 04:45:02.846581
1968	8	1-714	1968	59	32	346	39	2012-08-02 04:45:02.846581
1977	4	1-345	1976	50	10	127	29	2012-08-02 04:45:04.095419
1978	4	1-565	1976	10	24	34	28	2012-08-02 04:45:04.095419
1979	4	1-604	1976	24	44	249	19	2012-08-02 04:45:04.095419
1976	4	1-719	1976	44	50	145	71	2012-08-02 04:45:04.095419
1981	7	1-550	1980	50	21	2199	60	2012-08-02 04:45:04.095419
1982	7	1-621	1980	21	34	2526	76	2012-08-02 04:45:04.095419
1983	7	1-285	1980	34	90	2082	3	2012-08-02 04:45:04.095419
1984	7	1-454	1980	90	62	2016	54	2012-08-02 04:45:04.095419
1985	7	1-654	1980	62	24	331	28	2012-08-02 04:45:04.095419
1986	7	1-604	1980	24	44	2576	19	2012-08-02 04:45:04.095419
1980	7	1-719	1980	44	50	1591	71	2012-08-02 04:45:04.095419
1988	6	1-656	1987	23	85	465	51	2012-08-02 04:45:04.871264
1989	6	1-457	1987	85	32	66	26	2012-08-02 04:45:04.871264
1987	6	1-723	1987	19	23	54	7	2012-08-02 04:45:04.871264
1990	6	1-680	1987	32	13	65	52	2012-08-02 04:45:04.871264
1991	6	1-696	1987	13	6	57	1	2012-08-02 04:45:04.871264
1992	6	1-303	1987	6	19	55	36	2012-08-02 04:45:04.871264
1994	7	1-502	1993	25	10	1184	61	2012-08-02 04:45:05.635534
1995	7	1-661	1993	10	90	128	49	2012-08-02 04:45:05.635534
1996	7	1-653	1993	90	47	529	79	2012-08-02 04:45:05.635534
1997	7	1-669	1993	47	9	301	88	2012-08-02 04:45:05.635534
1998	7	1-637	1993	9	50	5455	84	2012-08-02 04:45:05.635534
1999	7	1-683	1993	50	23	3287	90	2012-08-02 04:45:05.635534
1993	7	1-726	1993	23	25	1396	75	2012-08-02 04:45:05.635534
2001	8	1-502	2000	25	10	72	61	2012-08-02 04:45:05.635534
2002	8	1-661	2000	10	90	11	49	2012-08-02 04:45:05.635534
2003	8	1-653	2000	90	47	72	79	2012-08-02 04:45:05.635534
2004	8	1-669	2000	47	90	59	88	2012-08-02 04:45:05.635534
2005	8	1-629	2000	90	88	343	18	2012-08-02 04:45:05.635534
2006	8	1-519	2000	88	93	321	34	2012-08-02 04:45:05.635534
2007	8	1-479	2000	93	23	95	90	2012-08-02 04:45:05.635534
2000	8	1-726	2000	23	25	59	75	2012-08-02 04:45:05.635534
2009	6	1-502	2008	25	10	139	61	2012-08-02 04:45:05.635534
2010	6	1-661	2008	10	90	22	49	2012-08-02 04:45:05.635534
2011	6	1-653	2008	90	38	138	79	2012-08-02 04:45:05.635534
2012	6	1-617	2008	38	93	569	34	2012-08-02 04:45:05.635534
2013	6	1-479	2008	93	23	174	90	2012-08-02 04:45:05.635534
2008	6	1-726	2008	23	25	110	75	2012-08-02 04:45:05.635534
2015	8	1-502	2014	25	13	71	61	2012-08-02 04:45:05.635534
2016	8	1-613	2014	13	83	141	20	2012-08-02 04:45:05.635534
2017	8	1-473	2014	83	36	406	12	2012-08-02 04:45:05.635534
2018	8	1-686	2014	36	93	90	93	2012-08-02 04:45:05.635534
2019	8	1-657	2014	93	39	295	16	2012-08-02 04:45:05.635534
2020	8	1-148	2014	39	7	197	87	2012-08-02 04:45:05.635534
2021	8	1-346	2014	7	23	81	90	2012-08-02 04:45:05.635534
2014	8	1-726	2014	23	25	54	75	2012-08-02 04:45:05.635534
2023	8	1-69	2022	50	10	854	74	2012-08-02 04:45:05.635534
2024	8	1-645	2022	10	85	690	54	2012-08-02 04:45:05.635534
2025	8	1-676	2022	85	94	560	96	2012-08-02 04:45:05.635534
2026	8	1-289	2022	94	13	95	89	2012-08-02 04:45:05.635534
2027	8	1-536	2022	13	52	146	35	2012-08-02 04:45:05.635534
2028	8	1-99	2022	52	51	92	11	2012-08-02 04:45:05.635534
2029	8	1-628	2022	51	23	1280	90	2012-08-02 04:45:05.635534
2022	8	1-726	2022	23	50	873	75	2012-08-02 04:45:05.635534
2031	6	1-502	2030	25	89	1387	61	2012-08-02 04:45:05.635534
2032	6	1-439	2030	89	37	2260	45	2012-08-02 04:45:05.635534
2033	6	1-624	2030	37	63	999	43	2012-08-02 04:45:05.635534
2034	6	1-704	2030	63	7	3444	87	2012-08-02 04:45:05.635534
2035	6	1-346	2030	7	23	1476	90	2012-08-02 04:45:05.635534
2030	6	1-726	2030	23	25	1013	75	2012-08-02 04:45:05.635534
2037	6	1-569	2036	23	90	1011	88	2012-08-02 04:45:06.541699
2038	6	1-629	2036	90	5	5718	18	2012-08-02 04:45:06.541699
2039	6	1-341	2036	5	33	5058	10	2012-08-02 04:45:06.541699
2040	6	1-393	2036	33	21	330	62	2012-08-02 04:45:06.541699
2041	6	1-697	2036	21	97	1258	40	2012-08-02 04:45:06.541699
2036	6	1-727	2036	97	23	1038	24	2012-08-02 04:45:06.541699
2043	5	1-569	2042	23	93	3722	88	2012-08-02 04:45:06.541699
2044	5	1-291	2042	93	13	3622	89	2012-08-02 04:45:06.541699
2045	5	1-536	2042	13	80	5307	35	2012-08-02 04:45:06.541699
2046	5	1-155	2042	80	97	4476	40	2012-08-02 04:45:06.541699
2042	5	1-727	2042	97	23	3757	24	2012-08-02 04:45:06.541699
2048	2	1-398	2047	99	30	2476	41	2012-08-02 04:45:08.276604
2047	2	1-735	2047	30	99	3213	73	2012-08-02 04:45:08.276604
2050	2	1-369	2049	59	28	5308	48	2012-08-02 04:45:08.43164
2049	2	1-736	2049	28	59	8836	10	2012-08-02 04:45:08.43164
2052	6	1-393	2051	33	21	36	62	2012-08-02 04:45:08.43164
2053	6	1-697	2051	21	97	120	40	2012-08-02 04:45:08.43164
2054	6	1-727	2051	97	88	86	24	2012-08-02 04:45:08.43164
2055	6	1-634	2051	88	72	76	84	2012-08-02 04:45:08.43164
2056	6	1-461	2051	72	28	323	48	2012-08-02 04:45:08.43164
2051	6	1-736	2051	28	33	644	10	2012-08-02 04:45:08.43164
2058	8	1-704	2057	63	69	212	87	2012-08-02 04:45:11.40314
2059	8	1-712	2057	69	92	183	66	2012-08-02 04:45:11.40314
2060	8	1-711	2057	92	87	60	2	2012-08-02 04:45:11.40314
2061	8	1-186	2057	87	94	10	13	2012-08-02 04:45:11.40314
2062	8	1-659	2057	94	94	3	26	2012-08-02 04:45:11.40314
2063	8	1-311	2057	94	29	2	53	2012-08-02 04:45:11.40314
2064	8	1-450	2057	29	69	2	94	2012-08-02 04:45:11.40314
2057	8	1-742	2057	69	63	91	43	2012-08-02 04:45:11.40314
2066	8	1-267	2065	30	23	149	90	2012-08-02 04:45:11.40314
2067	8	1-726	2065	23	25	79	75	2012-08-02 04:45:11.40314
2068	8	1-502	2065	25	10	83	61	2012-08-02 04:45:11.40314
2069	8	1-661	2065	10	99	11	49	2012-08-02 04:45:11.40314
2070	8	1-689	2065	99	68	29	41	2012-08-02 04:45:11.40314
2071	8	1-639	2065	68	29	18	53	2012-08-02 04:45:11.40314
2072	8	1-450	2065	29	69	15	94	2012-08-02 04:45:11.40314
2065	8	1-742	2065	69	30	917	43	2012-08-02 04:45:11.40314
2074	8	1-267	2073	30	77	1049	90	2012-08-02 04:45:11.40314
2075	8	1-328	2073	77	70	1229	34	2012-08-02 04:45:11.40314
2076	8	1-741	2073	70	21	54	53	2012-08-02 04:45:11.40314
2077	8	1-37	2073	21	13	46	52	2012-08-02 04:45:11.40314
2078	8	1-696	2073	13	6	40	1	2012-08-02 04:45:11.40314
2079	8	1-303	2073	6	87	38	36	2012-08-02 04:45:11.40314
2080	8	1-233	2073	87	69	61	94	2012-08-02 04:45:11.40314
2073	8	1-742	2073	69	30	4976	43	2012-08-02 04:45:11.40314
2082	8	1-540	2081	26	88	607	92	2012-08-02 04:45:11.911952
2083	8	1-521	2081	88	64	437	75	2012-08-02 04:45:11.911952
2084	8	1-116	2081	64	38	185	83	2012-08-02 04:45:11.911952
2085	8	1-470	2081	38	32	126	39	2012-08-02 04:45:11.911952
2086	8	1-668	2081	32	70	566	6	2012-08-02 04:45:11.911952
2087	8	1-264	2081	70	3	601	42	2012-08-02 04:45:11.911952
2088	8	1-672	2081	3	47	779	93	2012-08-02 04:45:11.911952
2081	8	1-744	2081	47	26	1083	32	2012-08-02 04:45:11.911952
2090	6	1-680	2089	32	13	903	52	2012-08-02 04:45:12.530782
2091	6	1-696	2089	13	36	539	1	2012-08-02 04:45:12.530782
2092	6	1-603	2089	36	37	786	27	2012-08-02 04:45:12.530782
2093	6	1-703	2089	37	97	1296	100	2012-08-02 04:45:12.530782
2094	6	1-459	2089	97	24	1529	57	2012-08-02 04:45:12.530782
2089	6	1-745	2089	24	32	1329	26	2012-08-02 04:45:12.530782
15879	4	1-3668	15878	93	93	2940	10	2012-08-02 05:13:09.638437
2096	7	1-400	2095	24	3	1445	42	2012-08-02 04:45:12.530782
2097	7	1-672	2095	3	47	1774	93	2012-08-02 04:45:12.530782
2098	7	1-744	2095	47	90	2336	32	2012-08-02 04:45:12.530782
2099	7	1-693	2095	90	88	1273	24	2012-08-02 04:45:12.530782
2100	7	1-634	2095	88	8	1181	84	2012-08-02 04:45:12.530782
2101	7	1-241	2095	8	24	806	57	2012-08-02 04:45:12.530782
2095	7	1-745	2095	24	24	939	26	2012-08-02 04:45:12.530782
15880	4	1-4439	15878	93	63	3354	62	2012-08-02 05:13:09.638437
2103	5	1-556	2102	100	80	1220	64	2012-08-02 04:45:12.850928
2104	5	1-670	2102	80	76	6761	41	2012-08-02 04:45:12.850928
2105	5	1-142	2102	76	94	1674	62	2012-08-02 04:45:12.850928
2106	5	1-600	2102	94	27	158	38	2012-08-02 04:45:12.850928
2102	5	1-746	2102	27	100	2035	37	2012-08-02 04:45:12.850928
15881	4	1-2468	15878	63	14	1100	68	2012-08-02 05:13:09.638437
2108	8	1-454	2107	90	85	1526	54	2012-08-02 04:45:13.027823
2109	8	1-676	2107	85	21	1194	96	2012-08-02 04:45:13.027823
2110	8	1-143	2107	21	32	604	26	2012-08-02 04:45:13.027823
2111	8	1-680	2107	32	13	567	52	2012-08-02 04:45:13.027823
2112	8	1-696	2107	13	4	467	1	2012-08-02 04:45:13.027823
2113	8	1-512	2107	4	93	451	88	2012-08-02 04:45:13.027823
2114	8	1-291	2107	93	82	444	89	2012-08-02 04:45:13.027823
2107	8	1-747	2107	82	90	1405	3	2012-08-02 04:45:13.027823
2115	1	1-291	\N	93	93	330	89	2012-08-02 04:45:13.027823
15878	4	1-4570	15878	14	93	2322	75	2012-08-02 05:13:09.638437
2117	4	1-363	2116	52	8	1819	84	2012-08-02 04:45:14.563837
2118	4	1-241	2116	8	24	1458	57	2012-08-02 04:45:14.563837
2119	4	1-745	2116	24	68	1817	26	2012-08-02 04:45:14.563837
2116	4	1-751	2116	68	52	1429	98	2012-08-02 04:45:14.563837
2121	5	1-454	2120	90	59	3015	54	2012-08-02 04:45:14.972351
2122	5	1-555	2120	59	38	108	83	2012-08-02 04:45:14.972351
2123	5	1-470	2120	38	11	76	39	2012-08-02 04:45:14.972351
2124	5	1-27	2120	11	53	2810	33	2012-08-02 04:45:14.972351
2120	5	1-753	2120	53	90	2882	3	2012-08-02 04:45:14.972351
2126	7	1-630	2125	66	70	396	34	2012-08-02 04:45:16.088607
2127	7	1-741	2125	70	29	16	53	2012-08-02 04:45:16.088607
2128	7	1-450	2125	29	69	15	94	2012-08-02 04:45:16.088607
2129	7	1-742	2125	69	84	1130	43	2012-08-02 04:45:16.088607
2130	7	1-375	2125	84	9	1524	17	2012-08-02 04:45:16.088607
2131	7	1-9	2125	9	11	772	16	2012-08-02 04:45:16.088607
2125	7	1-755	2125	11	66	580	3	2012-08-02 04:45:16.088607
2132	1	1-375	\N	84	84	602	17	2012-08-02 04:45:16.088607
2134	8	1-676	2133	85	21	1226	96	2012-08-02 04:45:17.888937
2135	8	1-143	2133	21	24	474	26	2012-08-02 04:45:17.888937
2136	8	1-400	2133	24	3	574	42	2012-08-02 04:45:17.888937
2137	8	1-672	2133	3	93	610	93	2012-08-02 04:45:17.888937
2138	8	1-657	2133	93	41	1520	16	2012-08-02 04:45:17.888937
2139	8	1-307	2133	41	97	508	31	2012-08-02 04:45:17.888937
2140	8	1-671	2133	97	19	3185	4	2012-08-02 04:45:17.888937
2133	8	1-759	2133	19	85	2050	54	2012-08-02 04:45:17.888937
2142	8	1-676	2141	85	30	1127	96	2012-08-02 04:45:17.888937
2143	8	1-302	2141	30	24	1066	57	2012-08-02 04:45:17.888937
2144	8	1-745	2141	24	94	991	26	2012-08-02 04:45:17.888937
2145	8	1-311	2141	94	75	544	53	2012-08-02 04:45:17.888937
2146	8	1-216	2141	75	72	490	97	2012-08-02 04:45:17.888937
2147	8	1-347	2141	72	97	449	31	2012-08-02 04:45:17.888937
2148	8	1-671	2141	97	19	2852	4	2012-08-02 04:45:17.888937
2141	8	1-759	2141	19	85	1860	54	2012-08-02 04:45:17.888937
2150	8	1-676	2149	85	30	70	96	2012-08-02 04:45:17.888937
2151	8	1-302	2149	30	24	73	57	2012-08-02 04:45:17.888937
2152	8	1-745	2149	24	94	75	26	2012-08-02 04:45:17.888937
2153	8	1-311	2149	94	29	45	53	2012-08-02 04:45:17.888937
2154	8	1-450	2149	29	4	39	94	2012-08-02 04:45:17.888937
2155	8	1-679	2149	4	23	38	87	2012-08-02 04:45:17.888937
2156	8	1-605	2149	23	19	145	4	2012-08-02 04:45:17.888937
2149	8	1-759	2149	19	85	104	54	2012-08-02 04:45:17.888937
2158	5	1-521	2157	88	64	4141	75	2012-08-02 04:45:21.258702
2159	5	1-116	2157	64	94	1728	83	2012-08-02 04:45:21.258702
2160	5	1-501	2157	94	82	1544	67	2012-08-02 04:45:21.258702
2161	5	1-758	2157	82	86	2217	98	2012-08-02 04:45:21.258702
2157	5	1-768	2157	86	88	5834	92	2012-08-02 04:45:21.258702
2163	6	1-689	2162	99	68	3431	41	2012-08-02 04:45:24.473522
2164	6	1-639	2162	68	29	1721	53	2012-08-02 04:45:24.473522
2165	6	1-450	2162	29	17	1129	94	2012-08-02 04:45:24.473522
2166	6	1-463	2162	17	47	568	93	2012-08-02 04:45:24.473522
2167	6	1-744	2162	47	41	554	32	2012-08-02 04:45:24.473522
2162	6	1-776	2162	41	99	1626	49	2012-08-02 04:45:24.473522
2169	7	1-416	2168	26	26	3495	20	2012-08-02 04:45:24.473522
2170	7	1-534	2168	26	25	1659	44	2012-08-02 04:45:24.473522
2171	7	1-389	2168	25	70	1432	6	2012-08-02 04:45:24.473522
2172	7	1-264	2168	70	3	1084	42	2012-08-02 04:45:24.473522
2173	7	1-672	2168	3	47	1001	93	2012-08-02 04:45:24.473522
2174	7	1-744	2168	47	41	991	32	2012-08-02 04:45:24.473522
2168	7	1-776	2168	41	26	2953	49	2012-08-02 04:45:24.473522
2176	6	1-416	2175	26	26	189	20	2012-08-02 04:45:24.473522
2177	6	1-534	2175	26	25	102	44	2012-08-02 04:45:24.473522
2178	6	1-389	2175	25	43	113	6	2012-08-02 04:45:24.473522
2179	6	1-616	2175	43	7	28	69	2012-08-02 04:45:24.473522
2180	6	1-60	2175	7	41	37	32	2012-08-02 04:45:24.473522
2175	6	1-776	2175	41	26	141	49	2012-08-02 04:45:24.473522
2182	7	1-428	2181	34	76	734	41	2012-08-02 04:45:24.473522
2183	7	1-142	2181	76	47	236	62	2012-08-02 04:45:24.473522
2184	7	1-66	2181	47	25	181	44	2012-08-02 04:45:24.473522
2185	7	1-389	2181	25	43	249	6	2012-08-02 04:45:24.473522
2186	7	1-616	2181	43	7	76	69	2012-08-02 04:45:24.473522
2187	7	1-60	2181	7	41	123	32	2012-08-02 04:45:24.473522
2181	7	1-776	2181	41	34	585	49	2012-08-02 04:45:24.473522
2189	6	1-207	2188	50	45	170	73	2012-08-02 04:45:24.473522
2190	6	1-296	2188	45	21	427	53	2012-08-02 04:45:24.473522
2191	6	1-37	2188	21	13	367	52	2012-08-02 04:45:24.473522
2192	6	1-696	2188	13	9	320	1	2012-08-02 04:45:24.473522
2193	6	1-182	2188	9	41	50	32	2012-08-02 04:45:24.473522
2188	6	1-776	2188	41	50	239	49	2012-08-02 04:45:24.473522
2195	5	1-682	2194	44	34	1730	8	2012-08-02 04:45:26.630582
2196	5	1-702	2194	34	51	272	11	2012-08-02 04:45:26.630582
2194	5	1-783	2194	66	44	4075	82	2012-08-02 04:45:26.630582
2197	5	1-628	2194	51	44	859	90	2012-08-02 04:45:26.630582
2198	5	1-552	2194	44	66	849	50	2012-08-02 04:45:26.630582
2200	7	1-556	2199	100	37	815	64	2012-08-02 04:45:27.370688
2201	7	1-590	2199	37	5	1814	7	2012-08-02 04:45:27.370688
2202	7	1-757	2199	5	24	1064	57	2012-08-02 04:45:27.370688
2203	7	1-745	2199	24	24	1103	26	2012-08-02 04:45:27.370688
2204	7	1-400	2199	24	3	1509	42	2012-08-02 04:45:27.370688
2205	7	1-672	2199	3	15	1467	93	2012-08-02 04:45:27.370688
2199	7	1-786	2199	15	100	1583	37	2012-08-02 04:45:27.370688
15883	3	1-4333	15882	3	63	935	62	2012-08-02 05:13:09.638437
2207	7	1-468	2206	18	13	409	89	2012-08-02 04:45:27.370688
2208	7	1-536	2206	13	52	529	35	2012-08-02 04:45:27.370688
2209	7	1-99	2206	52	51	282	11	2012-08-02 04:45:27.370688
2210	7	1-628	2206	51	44	3296	90	2012-08-02 04:45:27.370688
2211	7	1-552	2206	44	6	5669	50	2012-08-02 04:45:27.370688
2212	7	1-554	2206	6	15	666	93	2012-08-02 04:45:27.370688
2206	7	1-786	2206	15	18	898	37	2012-08-02 04:45:27.370688
2214	4	1-468	2213	18	13	58	89	2012-08-02 04:45:27.370688
2215	4	1-536	2213	13	68	75	35	2012-08-02 04:45:27.370688
2216	4	1-123	2213	68	15	94	93	2012-08-02 04:45:27.370688
2213	4	1-786	2213	15	18	127	37	2012-08-02 04:45:27.370688
2218	5	1-556	2217	100	93	3600	64	2012-08-02 04:45:27.370688
2219	5	1-493	2217	93	93	5592	99	2012-08-02 04:45:27.370688
2220	5	1-386	2217	93	3	3321	42	2012-08-02 04:45:27.370688
2221	5	1-672	2217	3	15	4072	93	2012-08-02 04:45:27.370688
2217	5	1-786	2217	15	100	5545	37	2012-08-02 04:45:27.370688
2223	8	1-556	2222	100	37	66	64	2012-08-02 04:45:27.370688
2224	8	1-590	2222	37	5	194	7	2012-08-02 04:45:27.370688
2225	8	1-757	2222	5	24	149	57	2012-08-02 04:45:27.370688
2226	8	1-745	2222	24	94	164	26	2012-08-02 04:45:27.370688
2227	8	1-311	2222	94	29	107	53	2012-08-02 04:45:27.370688
2228	8	1-450	2222	29	17	98	94	2012-08-02 04:45:27.370688
2229	8	1-463	2222	17	15	69	93	2012-08-02 04:45:27.370688
2222	8	1-786	2222	15	100	98	37	2012-08-02 04:45:27.370688
2231	5	1-556	2230	100	93	226	64	2012-08-02 04:45:27.370688
2232	5	1-493	2230	93	37	360	99	2012-08-02 04:45:27.370688
2233	5	1-257	2230	37	22	230	29	2012-08-02 04:45:27.370688
2234	5	1-453	2230	22	15	200	93	2012-08-02 04:45:27.370688
2230	5	1-786	2230	15	100	308	37	2012-08-02 04:45:27.370688
2236	4	1-626	2235	83	75	2913	1	2012-08-02 04:45:28.210993
2237	4	1-695	2235	75	32	2609	26	2012-08-02 04:45:28.210993
2238	4	1-680	2235	32	4	2198	52	2012-08-02 04:45:28.210993
2235	4	1-789	2235	4	83	2623	60	2012-08-02 04:45:28.210993
2240	3	1-673	2239	98	4	265	1	2012-08-02 04:45:29.039958
2241	3	1-512	2239	4	42	66	88	2012-08-02 04:45:29.039958
2239	3	1-793	2239	42	98	2443	46	2012-08-02 04:45:29.039958
2243	5	1-453	2242	22	15	681	93	2012-08-02 04:45:29.40452
2244	5	1-786	2242	15	45	913	37	2012-08-02 04:45:29.40452
2245	5	1-681	2242	45	83	790	72	2012-08-02 04:45:29.40452
2246	5	1-663	2242	83	53	1904	25	2012-08-02 04:45:29.40452
2242	5	1-794	2242	53	22	898	29	2012-08-02 04:45:29.40452
2248	3	1-78	2247	29	83	3340	72	2012-08-02 04:45:29.40452
2249	3	1-663	2247	83	53	8095	25	2012-08-02 04:45:29.40452
2247	3	1-794	2247	53	29	3840	29	2012-08-02 04:45:29.40452
2251	5	1-565	2250	10	53	288	28	2012-08-02 04:45:29.40452
2252	5	1-583	2250	53	21	233	62	2012-08-02 04:45:29.40452
2253	5	1-697	2250	21	57	879	40	2012-08-02 04:45:29.40452
2254	5	1-651	2250	57	53	1708	25	2012-08-02 04:45:29.40452
2250	5	1-794	2250	53	10	939	29	2012-08-02 04:45:29.40452
2256	8	1-278	2255	44	17	60	70	2012-08-02 04:45:29.912978
2257	8	1-619	2255	17	74	471	76	2012-08-02 04:45:29.912978
2258	8	1-414	2255	74	37	166	27	2012-08-02 04:45:29.912978
2259	8	1-703	2255	37	70	373	100	2012-08-02 04:45:29.912978
2260	8	1-762	2255	70	97	315	40	2012-08-02 04:45:29.912978
2261	8	1-727	2255	97	88	264	24	2012-08-02 04:45:29.912978
2262	8	1-634	2255	88	55	273	84	2012-08-02 04:45:29.912978
2255	8	1-795	2255	55	44	404	5	2012-08-02 04:45:29.912978
2264	7	1-647	2263	3	41	924	32	2012-08-02 04:45:30.785605
2265	7	1-776	2263	41	50	3258	49	2012-08-02 04:45:30.785605
2266	7	1-207	2263	50	42	1718	73	2012-08-02 04:45:30.785605
2267	7	1-64	2263	42	19	2033	36	2012-08-02 04:45:30.785605
2268	7	1-723	2263	19	5	1473	7	2012-08-02 04:45:30.785605
2269	7	1-757	2263	5	2	971	57	2012-08-02 04:45:30.785605
2263	7	1-796	2263	2	3	802	39	2012-08-02 04:45:30.785605
2271	8	1-647	2270	3	41	217	32	2012-08-02 04:45:30.785605
2272	8	1-776	2270	41	21	914	49	2012-08-02 04:45:30.785605
2273	8	1-39	2270	21	13	348	52	2012-08-02 04:45:30.785605
2274	8	1-696	2270	13	6	269	1	2012-08-02 04:45:30.785605
2275	8	1-303	2270	6	19	234	36	2012-08-02 04:45:30.785605
2276	8	1-723	2270	19	5	203	7	2012-08-02 04:45:30.785605
2277	8	1-757	2270	5	2	160	57	2012-08-02 04:45:30.785605
2270	8	1-796	2270	2	3	158	39	2012-08-02 04:45:30.785605
2279	4	1-386	2278	93	32	477	42	2012-08-02 04:45:30.940765
2280	4	1-649	2278	32	69	46	94	2012-08-02 04:45:30.940765
2281	4	1-742	2278	69	2	2013	43	2012-08-02 04:45:30.940765
2278	4	1-797	2278	2	93	1386	99	2012-08-02 04:45:30.940765
2283	4	1-760	2282	68	73	762	22	2012-08-02 04:45:33.945514
2284	4	1-709	2282	73	46	194	29	2012-08-02 04:45:33.945514
2285	4	1-356	2282	46	32	190	78	2012-08-02 04:45:33.945514
2282	4	1-802	2282	32	68	2267	21	2012-08-02 04:45:33.945514
2287	7	1-227	2286	4	87	679	80	2012-08-02 04:45:38.419212
2288	7	1-230	2286	87	85	221	54	2012-08-02 04:45:38.419212
2289	7	1-676	2286	85	30	115	96	2012-08-02 04:45:38.419212
2290	7	1-302	2286	30	2	94	57	2012-08-02 04:45:38.419212
2291	7	1-796	2286	2	11	66	39	2012-08-02 04:45:38.419212
2292	7	1-27	2286	11	10	1685	33	2012-08-02 04:45:38.419212
2286	7	1-812	2286	10	4	3921	29	2012-08-02 04:45:38.419212
2294	7	1-565	2293	10	49	540	28	2012-08-02 04:45:38.419212
2295	7	1-72	2293	49	4	119	8	2012-08-02 04:45:38.419212
2296	7	1-4	2293	4	5	69	7	2012-08-02 04:45:38.419212
2297	7	1-757	2293	5	2	38	57	2012-08-02 04:45:38.419212
2298	7	1-796	2293	2	11	27	39	2012-08-02 04:45:38.419212
2299	7	1-27	2293	11	10	683	33	2012-08-02 04:45:38.419212
2293	7	1-812	2293	10	10	2009	29	2012-08-02 04:45:38.419212
2300	2	1-814	2300	98	10	4137	56	2012-08-02 04:45:38.949749
2301	2	1-675	2300	10	98	4639	77	2012-08-02 04:45:38.949749
15884	3	1-2468	15882	63	14	321	68	2012-08-02 05:13:09.638437
2303	6	1-79	2302	3	71	375	15	2012-08-02 04:45:39.955096
2304	6	1-804	2302	71	4	705	94	2012-08-02 04:45:39.955096
2305	6	1-679	2302	4	60	693	87	2012-08-02 04:45:39.955096
2306	6	1-667	2302	60	18	1238	78	2012-08-02 04:45:39.955096
2307	6	1-650	2302	18	11	5177	14	2012-08-02 04:45:39.955096
2302	6	1-817	2302	11	3	1194	77	2012-08-02 04:45:39.955096
15882	3	1-4570	15882	14	3	709	75	2012-08-02 05:13:09.638437
2309	7	1-713	2308	41	76	951	65	2012-08-02 04:45:40.452613
2310	7	1-277	2308	76	38	889	79	2012-08-02 04:45:40.452613
2311	7	1-617	2308	38	70	3523	34	2012-08-02 04:45:40.452613
2312	7	1-741	2308	70	29	138	53	2012-08-02 04:45:40.452613
2313	7	1-450	2308	29	49	128	94	2012-08-02 04:45:40.452613
2314	7	1-442	2308	49	10	200	97	2012-08-02 04:45:40.452613
2308	7	1-818	2308	10	41	277	13	2012-08-02 04:45:40.452613
2316	7	1-659	2315	94	32	2810	26	2012-08-02 04:45:40.452613
2317	7	1-680	2315	32	13	2636	52	2012-08-02 04:45:40.452613
2318	7	1-696	2315	13	6	2173	1	2012-08-02 04:45:40.452613
2319	7	1-303	2315	6	87	2012	36	2012-08-02 04:45:40.452613
2320	7	1-233	2315	87	49	3068	94	2012-08-02 04:45:40.452613
2321	7	1-442	2315	49	10	5161	97	2012-08-02 04:45:40.452613
2315	7	1-818	2315	10	94	7720	13	2012-08-02 04:45:40.452613
2323	6	1-659	2322	94	32	267	26	2012-08-02 04:45:40.452613
2324	6	1-680	2322	32	51	259	52	2012-08-02 04:45:40.452613
2325	6	1-701	2322	51	29	253	53	2012-08-02 04:45:40.452613
2326	6	1-450	2322	29	49	263	94	2012-08-02 04:45:40.452613
2327	6	1-442	2322	49	10	458	97	2012-08-02 04:45:40.452613
2322	6	1-818	2322	10	94	708	13	2012-08-02 04:45:40.452613
2329	7	1-583	2328	53	21	120	62	2012-08-02 04:45:42.585046
2330	7	1-697	2328	21	97	305	40	2012-08-02 04:45:42.585046
2331	7	1-727	2328	97	88	167	24	2012-08-02 04:45:42.585046
2332	7	1-634	2328	88	72	113	84	2012-08-02 04:45:42.585046
2333	7	1-461	2328	72	46	367	48	2012-08-02 04:45:42.585046
2334	7	1-684	2328	46	49	241	55	2012-08-02 04:45:42.585046
2328	7	1-822	2328	49	53	233	28	2012-08-02 04:45:42.585046
2336	7	1-583	2335	53	21	271	62	2012-08-02 04:45:42.585046
2337	7	1-697	2335	21	97	713	40	2012-08-02 04:45:42.585046
2338	7	1-727	2335	97	88	406	24	2012-08-02 04:45:42.585046
2339	7	1-634	2335	88	72	286	84	2012-08-02 04:45:42.585046
2340	7	1-461	2335	72	25	963	48	2012-08-02 04:45:42.585046
2341	7	1-692	2335	25	49	503	55	2012-08-02 04:45:42.585046
2335	7	1-822	2335	49	53	505	28	2012-08-02 04:45:42.585046
2343	6	1-583	2342	53	21	1667	62	2012-08-02 04:45:42.585046
2344	6	1-697	2342	21	97	4875	40	2012-08-02 04:45:42.585046
2345	6	1-727	2342	97	88	3083	24	2012-08-02 04:45:42.585046
2346	6	1-634	2342	88	50	2413	84	2012-08-02 04:45:42.585046
2347	6	1-769	2342	50	49	2509	55	2012-08-02 04:45:42.585046
2342	6	1-822	2342	49	53	2797	28	2012-08-02 04:45:42.585046
2349	6	1-583	2348	53	58	924	62	2012-08-02 04:45:42.585046
2350	6	1-93	2348	58	82	422	67	2012-08-02 04:45:42.585046
2351	6	1-758	2348	82	86	547	98	2012-08-02 04:45:42.585046
2352	6	1-768	2348	86	40	1299	92	2012-08-02 04:45:42.585046
2353	6	1-133	2348	40	49	1121	55	2012-08-02 04:45:42.585046
2348	6	1-822	2348	49	53	1391	28	2012-08-02 04:45:42.585046
2355	6	1-583	2354	53	58	958	62	2012-08-02 04:45:42.585046
2356	6	1-93	2354	58	82	480	67	2012-08-02 04:45:42.585046
2357	6	1-758	2354	82	52	682	98	2012-08-02 04:45:42.585046
2358	6	1-363	2354	52	50	761	84	2012-08-02 04:45:42.585046
2359	6	1-769	2354	50	49	967	55	2012-08-02 04:45:42.585046
2354	6	1-822	2354	49	53	1316	28	2012-08-02 04:45:42.585046
2361	7	1-601	2360	12	87	2595	80	2012-08-02 04:45:43.734002
2362	7	1-230	2360	87	85	1264	54	2012-08-02 04:45:43.734002
2363	7	1-676	2360	85	30	987	96	2012-08-02 04:45:43.734002
2364	7	1-302	2360	30	24	1203	57	2012-08-02 04:45:43.734002
2365	7	1-745	2360	24	32	1442	26	2012-08-02 04:45:43.734002
2366	7	1-680	2360	32	20	1350	52	2012-08-02 04:45:43.734002
2360	7	1-823	2360	20	12	3868	85	2012-08-02 04:45:43.734002
2368	8	1-771	2367	29	94	1227	83	2012-08-02 04:45:43.734002
2369	8	1-501	2367	94	74	1251	67	2012-08-02 04:45:43.734002
2370	8	1-436	2367	74	2	976	43	2012-08-02 04:45:43.734002
2371	8	1-797	2367	2	13	1233	99	2012-08-02 04:45:43.734002
2372	8	1-820	2367	13	94	631	13	2012-08-02 04:45:43.734002
2373	8	1-659	2367	94	32	241	26	2012-08-02 04:45:43.734002
2374	8	1-680	2367	32	20	233	52	2012-08-02 04:45:43.734002
2367	8	1-823	2367	20	29	691	85	2012-08-02 04:45:43.734002
2376	7	1-771	2375	29	82	149	83	2012-08-02 04:45:43.734002
2377	7	1-249	2375	82	55	37	84	2012-08-02 04:45:43.734002
2378	7	1-795	2375	55	19	58	5	2012-08-02 04:45:43.734002
2379	7	1-208	2375	19	94	77	13	2012-08-02 04:45:43.734002
2380	7	1-659	2375	94	32	29	26	2012-08-02 04:45:43.734002
2381	7	1-680	2375	32	20	28	52	2012-08-02 04:45:43.734002
2375	7	1-823	2375	20	29	84	85	2012-08-02 04:45:43.734002
2383	6	1-658	2382	14	43	1575	27	2012-08-02 04:45:44.828001
2384	6	1-806	2382	43	32	234	15	2012-08-02 04:45:44.828001
2385	6	1-805	2382	32	38	103	83	2012-08-02 04:45:44.828001
2386	6	1-470	2382	38	11	66	39	2012-08-02 04:45:44.828001
2387	6	1-27	2382	11	96	2208	33	2012-08-02 04:45:44.828001
2382	6	1-827	2382	96	14	1480	6	2012-08-02 04:45:44.828001
2389	8	1-719	2388	44	50	487	71	2012-08-02 04:45:45.899983
2390	8	1-550	2388	50	83	742	60	2012-08-02 04:45:45.899983
2391	8	1-626	2388	83	75	903	1	2012-08-02 04:45:45.899983
2392	8	1-695	2388	75	68	886	26	2012-08-02 04:45:45.899983
2393	8	1-751	2388	68	86	660	98	2012-08-02 04:45:45.899983
2394	8	1-768	2388	86	53	1857	92	2012-08-02 04:45:45.899983
2395	8	1-396	2388	53	62	760	29	2012-08-02 04:45:45.899983
2388	8	1-828	2388	62	44	715	19	2012-08-02 04:45:45.899983
2397	6	1-719	2396	44	50	724	71	2012-08-02 04:45:45.899983
2398	6	1-550	2396	50	5	1190	60	2012-08-02 04:45:45.899983
2399	6	1-788	2396	5	2	1102	43	2012-08-02 04:45:45.899983
2400	6	1-797	2396	2	37	1432	99	2012-08-02 04:45:45.899983
2401	6	1-257	2396	37	62	972	29	2012-08-02 04:45:45.899983
2396	6	1-828	2396	62	44	986	19	2012-08-02 04:45:45.899983
2403	8	1-787	2402	88	68	1443	26	2012-08-02 04:45:45.899983
2404	8	1-751	2402	68	52	1164	98	2012-08-02 04:45:45.899983
2405	8	1-363	2402	52	55	1520	84	2012-08-02 04:45:45.899983
2406	8	1-795	2402	55	19	2430	5	2012-08-02 04:45:45.899983
2407	8	1-208	2402	19	45	3308	13	2012-08-02 04:45:45.899983
2408	8	1-767	2402	45	37	3258	99	2012-08-02 04:45:45.899983
2409	8	1-257	2402	37	62	2221	29	2012-08-02 04:45:45.899983
2402	8	1-828	2402	62	88	2262	19	2012-08-02 04:45:45.899983
15886	5	1-3748	15885	58	30	958	77	2012-08-02 05:13:09.638437
2411	8	1-550	2410	50	9	282	60	2012-08-02 04:45:46.916236
2412	8	1-591	2410	9	87	281	80	2012-08-02 04:45:46.916236
2413	8	1-230	2410	87	94	120	54	2012-08-02 04:45:46.916236
2414	8	1-774	2410	94	49	140	55	2012-08-02 04:45:46.916236
2415	8	1-822	2410	49	53	183	28	2012-08-02 04:45:46.916236
2416	8	1-583	2410	53	58	128	62	2012-08-02 04:45:46.916236
2417	8	1-93	2410	58	55	62	67	2012-08-02 04:45:46.916236
2410	8	1-831	2410	55	50	208	71	2012-08-02 04:45:46.916236
15887	5	1-4532	15885	30	53	1139	74	2012-08-02 05:13:09.638437
2419	8	1-550	2418	50	9	1345	60	2012-08-02 04:45:46.916236
2420	8	1-591	2418	9	87	1494	80	2012-08-02 04:45:46.916236
2421	8	1-230	2418	87	81	710	54	2012-08-02 04:45:46.916236
2422	8	1-734	2418	81	63	465	46	2012-08-02 04:45:46.916236
2423	8	1-748	2418	63	1	576	48	2012-08-02 04:45:46.916236
2424	8	1-826	2418	1	48	267	56	2012-08-02 04:45:46.916236
2425	8	1-705	2418	48	55	237	67	2012-08-02 04:45:46.916236
2418	8	1-831	2418	55	50	891	71	2012-08-02 04:45:46.916236
2426	1	1-93	\N	58	58	245	67	2012-08-02 04:45:46.916236
15888	5	1-4361	15885	53	93	574	24	2012-08-02 05:13:09.638437
2428	8	1-790	2427	23	96	822	76	2012-08-02 04:45:48.010156
2429	8	1-475	2427	96	92	972	66	2012-08-02 04:45:48.010156
2430	8	1-711	2427	92	87	477	2	2012-08-02 04:45:48.010156
2431	8	1-186	2427	87	41	117	13	2012-08-02 04:45:48.010156
2432	8	1-713	2427	41	51	458	65	2012-08-02 04:45:48.010156
2433	8	1-71	2427	51	68	337	21	2012-08-02 04:45:48.010156
2434	8	1-760	2427	68	70	317	22	2012-08-02 04:45:48.010156
2427	8	1-834	2427	70	23	175	46	2012-08-02 04:45:48.010156
2436	7	1-713	2435	41	76	631	65	2012-08-02 04:45:48.253392
2437	7	1-277	2435	76	47	339	79	2012-08-02 04:45:48.253392
2438	7	1-669	2435	47	90	159	88	2012-08-02 04:45:48.253392
2439	7	1-629	2435	90	10	493	18	2012-08-02 04:45:48.253392
2440	7	1-10	2435	10	38	83	10	2012-08-02 04:45:48.253392
2441	7	1-648	2435	38	78	258	4	2012-08-02 04:45:48.253392
2435	7	1-835	2435	78	41	320	13	2012-08-02 04:45:48.253392
2443	3	1-326	2442	95	38	998	10	2012-08-02 04:45:48.253392
2444	3	1-648	2442	38	78	3162	4	2012-08-02 04:45:48.253392
2442	3	1-835	2442	78	95	3970	13	2012-08-02 04:45:48.253392
2446	7	1-659	2445	94	68	85	26	2012-08-02 04:45:48.253392
2447	7	1-751	2445	68	52	51	98	2012-08-02 04:45:48.253392
2448	7	1-363	2445	52	50	49	84	2012-08-02 04:45:48.253392
2449	7	1-683	2445	50	44	36	90	2012-08-02 04:45:48.253392
2450	7	1-552	2445	44	100	56	50	2012-08-02 04:45:48.253392
2451	7	1-765	2445	100	78	161	4	2012-08-02 04:45:48.253392
2445	7	1-835	2445	78	94	295	13	2012-08-02 04:45:48.253392
2453	6	1-530	2452	84	26	178	3	2012-08-02 04:45:48.253392
2454	6	1-337	2452	26	50	34	84	2012-08-02 04:45:48.253392
2455	6	1-683	2452	50	44	27	90	2012-08-02 04:45:48.253392
2456	6	1-552	2452	44	100	44	50	2012-08-02 04:45:48.253392
2457	6	1-765	2452	100	78	134	4	2012-08-02 04:45:48.253392
2452	6	1-835	2452	78	84	261	13	2012-08-02 04:45:48.253392
2459	7	1-767	2458	45	43	1304	99	2012-08-02 04:45:48.253392
2460	7	1-546	2458	43	26	1027	3	2012-08-02 04:45:48.253392
2461	7	1-337	2458	26	50	200	84	2012-08-02 04:45:48.253392
2462	7	1-683	2458	50	44	160	90	2012-08-02 04:45:48.253392
2463	7	1-552	2458	44	100	269	50	2012-08-02 04:45:48.253392
2464	7	1-765	2458	100	78	829	4	2012-08-02 04:45:48.253392
2458	7	1-835	2458	78	45	1648	13	2012-08-02 04:45:48.253392
2466	6	1-752	2465	72	37	894	61	2012-08-02 04:45:49.195487
2467	6	1-47	2465	37	96	1963	33	2012-08-02 04:45:49.195487
2468	6	1-827	2465	96	14	796	6	2012-08-02 04:45:49.195487
2469	6	1-658	2465	14	37	513	27	2012-08-02 04:45:49.195487
2470	6	1-703	2465	37	47	877	100	2012-08-02 04:45:49.195487
2465	6	1-839	2465	47	72	2009	23	2012-08-02 04:45:49.195487
2472	8	1-752	2471	72	37	2272	61	2012-08-02 04:45:49.195487
2473	8	1-47	2471	37	96	4799	33	2012-08-02 04:45:49.195487
2474	8	1-827	2471	96	14	2661	6	2012-08-02 04:45:49.195487
2475	8	1-658	2471	14	43	2343	27	2012-08-02 04:45:49.195487
2476	8	1-806	2471	43	71	288	15	2012-08-02 04:45:49.195487
2477	8	1-804	2471	71	57	450	94	2012-08-02 04:45:49.195487
2478	8	1-798	2471	57	47	1192	100	2012-08-02 04:45:49.195487
2471	8	1-839	2471	47	72	3734	23	2012-08-02 04:45:49.195487
2480	6	1-752	2479	72	23	2585	61	2012-08-02 04:45:49.195487
2481	6	1-146	2479	23	43	2118	27	2012-08-02 04:45:49.195487
2482	6	1-806	2479	43	71	272	15	2012-08-02 04:45:49.195487
2483	6	1-804	2479	71	57	446	94	2012-08-02 04:45:49.195487
2484	6	1-798	2479	57	47	1237	100	2012-08-02 04:45:49.195487
2479	6	1-839	2479	47	72	4057	23	2012-08-02 04:45:49.195487
2486	5	1-836	2485	72	29	3815	90	2012-08-02 04:45:50.697649
2487	5	1-545	2485	29	86	837	10	2012-08-02 04:45:50.697649
2488	5	1-764	2485	86	94	168	54	2012-08-02 04:45:50.697649
2489	5	1-774	2485	94	39	229	55	2012-08-02 04:45:50.697649
2485	5	1-841	2485	39	72	5484	74	2012-08-02 04:45:50.697649
2491	7	1-550	2490	50	83	248	60	2012-08-02 04:45:51.272026
2492	7	1-626	2490	83	4	217	1	2012-08-02 04:45:51.272026
2493	7	1-512	2490	4	90	148	88	2012-08-02 04:45:51.272026
2494	7	1-629	2490	90	10	609	18	2012-08-02 04:45:51.272026
2495	7	1-10	2490	10	86	136	10	2012-08-02 04:45:51.272026
2496	7	1-764	2490	86	80	19	54	2012-08-02 04:45:51.272026
2490	7	1-843	2490	80	50	226	71	2012-08-02 04:45:51.272026
2498	8	1-550	2497	50	83	1386	60	2012-08-02 04:45:51.272026
2499	8	1-626	2497	83	75	1661	1	2012-08-02 04:45:51.272026
2500	8	1-695	2497	75	96	1605	26	2012-08-02 04:45:51.272026
2501	8	1-404	2497	96	19	528	5	2012-08-02 04:45:51.272026
2502	8	1-208	2497	19	95	654	13	2012-08-02 04:45:51.272026
2503	8	1-326	2497	95	86	295	10	2012-08-02 04:45:51.272026
2504	8	1-764	2497	86	80	56	54	2012-08-02 04:45:51.272026
2497	8	1-843	2497	80	50	923	71	2012-08-02 04:45:51.272026
2506	6	1-550	2505	50	83	826	60	2012-08-02 04:45:51.272026
2507	6	1-626	2505	83	36	996	1	2012-08-02 04:45:51.272026
2508	6	1-603	2505	36	84	1957	27	2012-08-02 04:45:51.272026
2509	6	1-575	2505	84	86	172	10	2012-08-02 04:45:51.272026
2510	6	1-764	2505	86	80	33	54	2012-08-02 04:45:51.272026
2505	6	1-843	2505	80	50	546	71	2012-08-02 04:45:51.272026
15889	5	1-1111	15885	93	14	241	68	2012-08-02 05:13:09.638437
2512	4	1-745	2511	24	68	1995	26	2012-08-02 04:45:53.073678
2513	4	1-751	2511	68	52	939	98	2012-08-02 04:45:53.073678
2514	4	1-363	2511	52	10	716	84	2012-08-02 04:45:53.073678
2511	4	1-850	2511	10	24	2675	57	2012-08-02 04:45:53.073678
15885	5	1-4570	15885	14	58	555	75	2012-08-02 05:13:09.638437
2516	6	1-796	2515	2	32	305	39	2012-08-02 04:45:53.073678
2517	6	1-668	2515	32	95	1368	6	2012-08-02 04:45:53.073678
2518	6	1-698	2515	95	11	357	16	2012-08-02 04:45:53.073678
2519	6	1-755	2515	11	26	263	3	2012-08-02 04:45:53.073678
2520	6	1-337	2515	26	10	56	84	2012-08-02 04:45:53.073678
2515	6	1-850	2515	10	2	312	57	2012-08-02 04:45:53.073678
2522	8	1-807	2521	30	43	2569	27	2012-08-02 04:45:53.073678
2523	8	1-806	2521	43	71	406	15	2012-08-02 04:45:53.073678
2524	8	1-804	2521	71	49	818	94	2012-08-02 04:45:53.073678
2525	8	1-442	2521	49	28	1280	97	2012-08-02 04:45:53.073678
2526	8	1-718	2521	28	53	914	33	2012-08-02 04:45:53.073678
2527	8	1-753	2521	53	26	905	3	2012-08-02 04:45:53.073678
2528	8	1-337	2521	26	10	192	84	2012-08-02 04:45:53.073678
2521	8	1-850	2521	10	30	1072	57	2012-08-02 04:45:53.073678
2530	7	1-807	2529	30	43	875	27	2012-08-02 04:45:53.073678
2531	7	1-806	2529	43	71	151	15	2012-08-02 04:45:53.073678
2532	7	1-804	2529	71	31	332	94	2012-08-02 04:45:53.073678
2533	7	1-810	2529	31	53	219	33	2012-08-02 04:45:53.073678
2534	7	1-753	2529	53	26	237	3	2012-08-02 04:45:53.073678
2535	7	1-337	2529	26	10	55	84	2012-08-02 04:45:53.073678
2529	7	1-850	2529	10	30	334	57	2012-08-02 04:45:53.073678
2537	8	1-796	2536	2	32	319	39	2012-08-02 04:45:53.073678
2538	8	1-668	2536	32	70	1593	6	2012-08-02 04:45:53.073678
2539	8	1-264	2536	70	32	1817	42	2012-08-02 04:45:53.073678
2540	8	1-649	2536	32	31	320	94	2012-08-02 04:45:53.073678
2541	8	1-810	2536	31	53	208	33	2012-08-02 04:45:53.073678
2542	8	1-753	2536	53	26	221	3	2012-08-02 04:45:53.073678
2543	8	1-337	2536	26	10	50	84	2012-08-02 04:45:53.073678
2536	8	1-850	2536	10	2	304	57	2012-08-02 04:45:53.073678
2544	1	1-337	\N	26	26	587	84	2012-08-02 04:45:53.073678
2546	3	1-646	2545	2	18	4364	37	2012-08-02 04:45:53.979782
2547	3	1-468	2545	18	4	1759	89	2012-08-02 04:45:53.979782
2545	3	1-851	2545	4	2	1886	85	2012-08-02 04:45:53.979782
2549	7	1-134	2548	70	23	588	58	2012-08-02 04:45:53.979782
2550	7	1-488	2548	23	68	799	41	2012-08-02 04:45:53.979782
2551	7	1-639	2548	68	75	560	53	2012-08-02 04:45:53.979782
2552	7	1-216	2548	75	28	595	97	2012-08-02 04:45:53.979782
2553	7	1-718	2548	28	85	417	33	2012-08-02 04:45:53.979782
2554	7	1-821	2548	85	4	473	89	2012-08-02 04:45:53.979782
2548	7	1-851	2548	4	70	599	85	2012-08-02 04:45:53.979782
2556	6	1-134	2555	70	34	1893	58	2012-08-02 04:45:53.979782
2557	6	1-677	2555	34	88	2937	50	2012-08-02 04:45:53.979782
2558	6	1-730	2555	88	28	1888	97	2012-08-02 04:45:53.979782
2559	6	1-718	2555	28	85	1330	33	2012-08-02 04:45:53.979782
2560	6	1-821	2555	85	4	1513	89	2012-08-02 04:45:53.979782
2555	6	1-851	2555	4	70	1923	85	2012-08-02 04:45:53.979782
2562	6	1-646	2561	2	97	107	37	2012-08-02 04:45:53.979782
2563	6	1-343	2561	97	75	31	53	2012-08-02 04:45:53.979782
2564	6	1-216	2561	75	28	34	97	2012-08-02 04:45:53.979782
2565	6	1-718	2561	28	85	25	33	2012-08-02 04:45:53.979782
2566	6	1-821	2561	85	4	29	89	2012-08-02 04:45:53.979782
2561	6	1-851	2561	4	2	38	85	2012-08-02 04:45:53.979782
2568	6	1-646	2567	2	97	1989	37	2012-08-02 04:45:53.979782
2569	6	1-343	2567	97	29	615	53	2012-08-02 04:45:53.979782
2570	6	1-450	2567	29	31	613	94	2012-08-02 04:45:53.979782
2571	6	1-810	2567	31	85	395	33	2012-08-02 04:45:53.979782
2572	6	1-821	2567	85	4	486	89	2012-08-02 04:45:53.979782
2567	6	1-851	2567	4	2	669	85	2012-08-02 04:45:53.979782
2574	8	1-445	2573	48	90	5649	21	2012-08-02 04:45:53.979782
2575	8	1-520	2573	90	82	376	69	2012-08-02 04:45:53.979782
2576	8	1-508	2573	82	42	264	73	2012-08-02 04:45:53.979782
2577	8	1-64	2573	42	19	414	36	2012-08-02 04:45:53.979782
2578	8	1-723	2573	19	51	398	7	2012-08-02 04:45:53.979782
2579	8	1-829	2573	51	85	1610	33	2012-08-02 04:45:53.979782
2580	8	1-821	2573	85	4	2081	89	2012-08-02 04:45:53.979782
2573	8	1-851	2573	4	48	3008	85	2012-08-02 04:45:53.979782
2581	1	1-450	\N	29	29	5506	94	2012-08-02 04:45:53.979782
2583	8	1-713	2582	41	51	117	65	2012-08-02 04:45:54.653894
2584	8	1-71	2582	51	68	83	21	2012-08-02 04:45:54.653894
2585	8	1-760	2582	68	70	75	22	2012-08-02 04:45:54.653894
2586	8	1-834	2582	70	63	40	46	2012-08-02 04:45:54.653894
2587	8	1-748	2582	63	1	52	48	2012-08-02 04:45:54.653894
2588	8	1-826	2582	1	10	25	56	2012-08-02 04:45:54.653894
2589	8	1-675	2582	10	59	37	77	2012-08-02 04:45:54.653894
2582	8	1-853	2582	59	41	31	13	2012-08-02 04:45:54.653894
2591	8	1-834	2590	70	98	759	46	2012-08-02 04:45:55.043781
2592	8	1-673	2590	98	6	273	1	2012-08-02 04:45:55.043781
2593	8	1-303	2590	6	87	217	36	2012-08-02 04:45:55.043781
2594	8	1-233	2590	87	57	283	94	2012-08-02 04:45:55.043781
2595	8	1-798	2590	57	70	993	100	2012-08-02 04:45:55.043781
2596	8	1-762	2590	70	57	814	40	2012-08-02 04:45:55.043781
2597	8	1-651	2590	57	56	1575	25	2012-08-02 04:45:55.043781
2590	8	1-854	2590	56	70	1519	22	2012-08-02 04:45:55.043781
2599	8	1-558	2598	41	35	286	95	2012-08-02 04:45:55.043781
2600	8	1-620	2598	35	45	218	31	2012-08-02 04:45:55.043781
2601	8	1-561	2598	45	71	15	15	2012-08-02 04:45:55.043781
2602	8	1-804	2598	71	57	28	94	2012-08-02 04:45:55.043781
2603	8	1-798	2598	57	70	99	100	2012-08-02 04:45:55.043781
2604	8	1-762	2598	70	57	74	40	2012-08-02 04:45:55.043781
2605	8	1-651	2598	57	56	143	25	2012-08-02 04:45:55.043781
2598	8	1-854	2598	56	41	140	22	2012-08-02 04:45:55.043781
2607	6	1-709	2606	73	10	279	29	2012-08-02 04:45:55.043781
2608	6	1-565	2606	10	53	83	28	2012-08-02 04:45:55.043781
2609	6	1-583	2606	53	21	62	62	2012-08-02 04:45:55.043781
2610	6	1-697	2606	21	57	225	40	2012-08-02 04:45:55.043781
2611	6	1-651	2606	57	56	423	25	2012-08-02 04:45:55.043781
2606	6	1-854	2606	56	73	443	22	2012-08-02 04:45:55.043781
15891	3	1-4293	15890	5	23	912	56	2012-08-02 05:13:09.638437
2613	6	1-709	2612	73	62	748	29	2012-08-02 04:45:59.824428
2614	6	1-828	2612	62	88	725	19	2012-08-02 04:45:59.824428
2615	6	1-787	2612	88	68	440	26	2012-08-02 04:45:59.824428
2616	6	1-751	2612	68	73	338	98	2012-08-02 04:45:59.824428
2617	6	1-772	2612	73	69	432	88	2012-08-02 04:45:59.824428
2612	6	1-865	2612	69	73	1062	22	2012-08-02 04:45:59.824428
15892	3	1-3680	15890	23	14	264	68	2012-08-02 05:13:09.638437
2619	8	1-634	2618	88	10	134	84	2012-08-02 04:46:00.321812
2620	8	1-850	2618	10	30	819	57	2012-08-02 04:46:00.321812
2621	8	1-807	2618	30	11	2137	27	2012-08-02 04:46:00.321812
2622	8	1-861	2618	11	71	6210	42	2012-08-02 04:46:00.321812
2623	8	1-708	2618	71	63	7515	14	2012-08-02 04:46:00.321812
2624	8	1-863	2618	63	72	314	23	2012-08-02 04:46:00.321812
2625	8	1-752	2618	72	88	268	61	2012-08-02 04:46:00.321812
2618	8	1-868	2618	88	88	123	24	2012-08-02 04:46:00.321812
15890	3	1-4570	15890	14	5	622	75	2012-08-02 05:13:09.638437
2627	8	1-634	2626	88	10	93	84	2012-08-02 04:46:00.321812
2628	8	1-850	2626	10	30	586	57	2012-08-02 04:46:00.321812
2629	8	1-807	2626	30	84	1573	27	2012-08-02 04:46:00.321812
2630	8	1-575	2626	84	86	148	10	2012-08-02 04:46:00.321812
2631	8	1-764	2626	86	80	30	54	2012-08-02 04:46:00.321812
2632	8	1-843	2626	80	41	536	71	2012-08-02 04:46:00.321812
2633	8	1-177	2626	41	88	181	61	2012-08-02 04:46:00.321812
2626	8	1-868	2626	88	88	84	24	2012-08-02 04:46:00.321812
15893	1	1-580	\N	86	86	114	74	2012-08-02 05:13:09.638437
2635	3	1-626	2634	83	36	1477	1	2012-08-02 04:46:00.830273
2636	3	1-603	2634	36	52	543	27	2012-08-02 04:46:00.830273
2634	3	1-869	2634	52	83	6539	60	2012-08-02 04:46:00.830273
2638	7	1-545	2637	29	86	104	10	2012-08-02 04:46:04.310355
2639	7	1-764	2637	86	80	17	54	2012-08-02 04:46:04.310355
2640	7	1-843	2637	80	73	253	71	2012-08-02 04:46:04.310355
2641	7	1-130	2637	73	82	99	69	2012-08-02 04:46:04.310355
2642	7	1-508	2637	82	99	57	73	2012-08-02 04:46:04.310355
2643	7	1-398	2637	99	15	36	41	2012-08-02 04:46:04.310355
2637	7	1-876	2637	15	29	569	90	2012-08-02 04:46:04.310355
2644	1	1-508	\N	82	82	1769	73	2012-08-02 04:46:04.310355
15895	3	1-3680	15894	23	80	1113	68	2012-08-02 05:13:10.887518
2646	5	1-875	2645	17	8	1089	99	2012-08-02 04:46:05.173031
2647	5	1-845	2645	8	14	828	6	2012-08-02 04:46:05.173031
2648	5	1-658	2645	14	11	818	27	2012-08-02 04:46:05.173031
2649	5	1-861	2645	11	20	1905	42	2012-08-02 04:46:05.173031
2645	5	1-878	2645	20	17	1546	21	2012-08-02 04:46:05.173031
15896	3	1-4554	15894	80	14	1582	71	2012-08-02 05:13:10.887518
2651	7	1-843	2650	80	41	5468	71	2012-08-02 04:46:06.885666
2652	7	1-177	2650	41	10	1380	61	2012-08-02 04:46:06.885666
2653	7	1-661	2650	10	21	180	49	2012-08-02 04:46:06.885666
2654	7	1-39	2650	21	20	57	52	2012-08-02 04:46:06.885666
2655	7	1-823	2650	20	12	127	85	2012-08-02 04:46:06.885666
2656	7	1-601	2650	12	24	67	80	2012-08-02 04:46:06.885666
2650	7	1-882	2650	24	80	411	54	2012-08-02 04:46:06.885666
2658	8	1-676	2657	85	18	299	96	2012-08-02 04:46:06.885666
2659	8	1-874	2657	18	19	2954	36	2012-08-02 04:46:06.885666
2660	8	1-723	2657	19	63	2308	7	2012-08-02 04:46:06.885666
2661	8	1-625	2657	63	76	746	41	2012-08-02 04:46:06.885666
2662	8	1-142	2657	76	4	192	62	2012-08-02 04:46:06.885666
2663	8	1-871	2657	4	72	133	30	2012-08-02 04:46:06.885666
2664	8	1-491	2657	72	24	67	80	2012-08-02 04:46:06.885666
2657	8	1-882	2657	24	85	451	54	2012-08-02 04:46:06.885666
2666	7	1-600	2665	94	27	130	38	2012-08-02 04:46:07.339042
2667	7	1-746	2665	27	45	1733	37	2012-08-02 04:46:07.339042
2668	7	1-681	2665	45	53	1460	72	2012-08-02 04:46:07.339042
2669	7	1-355	2665	53	73	759	68	2012-08-02 04:46:07.339042
2670	7	1-710	2665	73	69	1109	17	2012-08-02 04:46:07.339042
2671	7	1-618	2665	69	72	1032	59	2012-08-02 04:46:07.339042
2665	7	1-885	2665	72	94	1324	62	2012-08-02 04:46:07.339042
2673	7	1-871	2672	4	72	2363	30	2012-08-02 04:46:07.339042
2674	7	1-491	2672	72	24	1200	80	2012-08-02 04:46:07.339042
2675	7	1-882	2672	24	60	6497	54	2012-08-02 04:46:07.339042
2676	7	1-862	2672	60	100	7868	39	2012-08-02 04:46:07.339042
2677	7	1-833	2672	100	69	4352	17	2012-08-02 04:46:07.339042
2678	7	1-618	2672	69	72	3271	59	2012-08-02 04:46:07.339042
2672	7	1-885	2672	72	4	4237	62	2012-08-02 04:46:07.339042
2680	4	1-66	2679	47	4	677	44	2012-08-02 04:46:07.339042
2681	4	1-573	2679	4	69	889	17	2012-08-02 04:46:07.339042
2682	4	1-618	2679	69	72	837	59	2012-08-02 04:46:07.339042
2679	4	1-885	2679	72	47	1086	62	2012-08-02 04:46:07.339042
2684	7	1-871	2683	4	72	36	30	2012-08-02 04:46:07.339042
2685	7	1-491	2683	72	24	19	80	2012-08-02 04:46:07.339042
2686	7	1-882	2683	24	85	116	54	2012-08-02 04:46:07.339042
2687	7	1-676	2683	85	21	69	96	2012-08-02 04:46:07.339042
2688	7	1-143	2683	21	80	27	26	2012-08-02 04:46:07.339042
2689	7	1-261	2683	80	72	42	59	2012-08-02 04:46:07.339042
2683	7	1-885	2683	72	4	58	62	2012-08-02 04:46:07.339042
2691	7	1-600	2690	94	27	130	38	2012-08-02 04:46:07.339042
2692	7	1-746	2690	27	64	1892	37	2012-08-02 04:46:07.339042
2693	7	1-858	2690	64	16	9429	93	2012-08-02 04:46:07.339042
2694	7	1-242	2690	16	18	491	78	2012-08-02 04:46:07.339042
2695	7	1-650	2690	18	61	2181	14	2012-08-02 04:46:07.339042
2696	7	1-484	2690	61	72	863	59	2012-08-02 04:46:07.339042
2690	7	1-885	2690	72	94	1210	62	2012-08-02 04:46:07.339042
2698	8	1-871	2697	4	72	206	30	2012-08-02 04:46:07.339042
2699	8	1-491	2697	72	24	116	80	2012-08-02 04:46:07.339042
2700	8	1-882	2697	24	81	761	54	2012-08-02 04:46:07.339042
2701	8	1-734	2697	81	98	426	46	2012-08-02 04:46:07.339042
2702	8	1-673	2697	98	75	149	1	2012-08-02 04:46:07.339042
2703	8	1-695	2697	75	80	123	26	2012-08-02 04:46:07.339042
2704	8	1-261	2697	80	72	213	59	2012-08-02 04:46:07.339042
2697	8	1-885	2697	72	4	304	62	2012-08-02 04:46:07.339042
2706	7	1-871	2705	4	75	485	30	2012-08-02 04:46:07.339042
2707	7	1-438	2705	75	6	142	11	2012-08-02 04:46:07.339042
2708	7	1-6	2705	6	43	90	10	2012-08-02 04:46:07.339042
2709	7	1-271	2705	43	18	229	78	2012-08-02 04:46:07.339042
2710	7	1-650	2705	18	61	1052	14	2012-08-02 04:46:07.339042
2711	7	1-484	2705	61	72	430	59	2012-08-02 04:46:07.339042
2705	7	1-885	2705	72	4	623	62	2012-08-02 04:46:07.339042
15894	3	1-4573	15894	14	23	3133	56	2012-08-02 05:13:10.887518
2713	8	1-66	2712	47	49	338	44	2012-08-02 04:46:07.339042
2714	8	1-750	2712	49	3	338	77	2012-08-02 04:46:07.339042
2715	8	1-79	2712	3	71	118	15	2012-08-02 04:46:07.339042
2716	8	1-804	2712	71	4	247	94	2012-08-02 04:46:07.339042
2717	8	1-679	2712	4	69	271	87	2012-08-02 04:46:07.339042
2718	8	1-712	2712	69	23	261	66	2012-08-02 04:46:07.339042
2719	8	1-444	2712	23	72	327	59	2012-08-02 04:46:07.339042
2712	8	1-885	2712	72	47	480	62	2012-08-02 04:46:07.339042
2721	6	1-380	2720	44	20	730	52	2012-08-02 04:46:08.389417
2722	6	1-823	2720	20	70	1598	85	2012-08-02 04:46:08.389417
2723	6	1-134	2720	70	74	1308	58	2012-08-02 04:46:08.389417
2724	6	1-327	2720	74	37	1225	40	2012-08-02 04:46:08.389417
2725	6	1-467	2720	37	15	952	68	2012-08-02 04:46:08.389417
2720	6	1-887	2720	15	44	616	63	2012-08-02 04:46:08.389417
15898	2	1-4127	15897	28	17	5753	9	2012-08-02 05:13:12.930839
2727	8	1-539	2726	73	52	725	86	2012-08-02 04:46:08.389417
2728	8	1-229	2726	52	47	616	79	2012-08-02 04:46:08.389417
2729	8	1-669	2726	47	69	532	88	2012-08-02 04:46:08.389417
2730	8	1-865	2726	69	73	1262	22	2012-08-02 04:46:08.389417
2731	8	1-709	2726	73	62	873	29	2012-08-02 04:46:08.389417
2732	8	1-828	2726	62	22	817	19	2012-08-02 04:46:08.389417
2733	8	1-295	2726	22	15	392	68	2012-08-02 04:46:08.389417
2726	8	1-887	2726	15	73	326	63	2012-08-02 04:46:08.389417
15897	2	1-4578	15897	17	28	5565	18	2012-08-02 05:13:12.930839
2735	7	1-539	2734	73	52	1129	86	2012-08-02 04:46:08.389417
2736	7	1-229	2734	52	47	965	79	2012-08-02 04:46:08.389417
2737	7	1-669	2734	47	69	839	88	2012-08-02 04:46:08.389417
2738	7	1-865	2734	69	83	2003	22	2012-08-02 04:46:08.389417
2739	7	1-716	2734	83	26	1121	51	2012-08-02 04:46:08.389417
2740	7	1-597	2734	26	15	630	68	2012-08-02 04:46:08.389417
2734	7	1-887	2734	15	73	526	63	2012-08-02 04:46:08.389417
2742	4	1-838	2741	60	29	1623	69	2012-08-02 04:46:08.389417
2743	4	1-222	2741	29	39	2301	95	2012-08-02 04:46:08.389417
2744	4	1-627	2741	39	15	4815	68	2012-08-02 04:46:08.389417
2741	4	1-887	2741	15	60	4197	63	2012-08-02 04:46:08.389417
15900	2	1-4334	15899	82	11	6712	43	2012-08-02 05:13:15.372371
2746	8	1-846	2745	52	85	467	33	2012-08-02 04:46:08.389417
2747	8	1-821	2745	85	64	597	89	2012-08-02 04:46:08.389417
2748	8	1-330	2745	64	32	620	42	2012-08-02 04:46:08.389417
2749	8	1-649	2745	32	57	109	94	2012-08-02 04:46:08.389417
2750	8	1-798	2745	57	70	412	100	2012-08-02 04:46:08.389417
2751	8	1-762	2745	70	37	365	40	2012-08-02 04:46:08.389417
2752	8	1-467	2745	37	15	383	68	2012-08-02 04:46:08.389417
2745	8	1-887	2745	15	52	336	63	2012-08-02 04:46:08.389417
15899	2	1-4584	15899	11	82	6285	48	2012-08-02 05:13:15.372371
2754	8	1-772	2753	73	69	1325	88	2012-08-02 04:46:11.219821
2755	8	1-865	2753	69	83	3213	22	2012-08-02 04:46:11.219821
2756	8	1-716	2753	83	26	1826	51	2012-08-02 04:46:11.219821
2757	8	1-597	2753	26	15	1042	68	2012-08-02 04:46:11.219821
2758	8	1-887	2753	15	52	885	63	2012-08-02 04:46:11.219821
2759	8	1-846	2753	52	85	1193	33	2012-08-02 04:46:11.219821
2760	8	1-821	2753	85	24	1479	89	2012-08-02 04:46:11.219821
2753	8	1-891	2753	24	73	1076	98	2012-08-02 04:46:11.219821
2762	6	1-772	2761	73	18	2177	88	2012-08-02 04:46:11.219821
2763	6	1-811	2761	18	15	1507	68	2012-08-02 04:46:11.219821
2764	6	1-887	2761	15	52	1313	63	2012-08-02 04:46:11.219821
2765	6	1-846	2761	52	85	1816	33	2012-08-02 04:46:11.219821
2766	6	1-821	2761	85	24	2310	89	2012-08-02 04:46:11.219821
2761	6	1-891	2761	24	73	1724	98	2012-08-02 04:46:11.219821
2768	2	1-658	2767	14	90	1084	27	2012-08-02 04:46:14.069413
2767	2	1-896	2767	90	14	2150	6	2012-08-02 04:46:14.069413
2770	8	1-428	2769	34	15	2	41	2012-08-02 04:46:14.555671
2771	8	1-876	2769	15	44	39	90	2012-08-02 04:46:14.555671
2772	8	1-552	2769	44	84	64	50	2012-08-02 04:46:14.555671
2773	8	1-715	2769	84	56	11	25	2012-08-02 04:46:14.555671
2774	8	1-854	2769	56	94	11	22	2012-08-02 04:46:14.555671
2775	8	1-722	2769	94	96	3	33	2012-08-02 04:46:14.555671
2776	8	1-827	2769	96	18	2	6	2012-08-02 04:46:14.555671
2769	8	1-897	2769	18	34	2	49	2012-08-02 04:46:14.555671
2778	7	1-653	2777	90	47	6939	79	2012-08-02 04:46:14.555671
2779	7	1-669	2777	47	76	4010	88	2012-08-02 04:46:14.555671
2780	7	1-892	2777	76	20	6012	42	2012-08-02 04:46:14.555671
2781	7	1-878	2777	20	17	3937	21	2012-08-02 04:46:14.555671
2782	7	1-875	2777	17	8	2239	99	2012-08-02 04:46:14.555671
2783	7	1-845	2777	8	18	1374	6	2012-08-02 04:46:14.555671
2777	7	1-897	2777	18	90	1654	49	2012-08-02 04:46:14.555671
2785	5	1-34	2784	28	1	488	48	2012-08-02 04:46:14.555671
2786	5	1-826	2784	1	13	195	56	2012-08-02 04:46:14.555671
2787	5	1-489	2784	13	32	30	39	2012-08-02 04:46:14.555671
2788	5	1-668	2784	32	18	120	6	2012-08-02 04:46:14.555671
2784	5	1-897	2784	18	28	186	49	2012-08-02 04:46:14.555671
2790	5	1-781	2789	93	1	5458	48	2012-08-02 04:46:14.555671
2791	5	1-826	2789	1	13	2237	56	2012-08-02 04:46:14.555671
2792	5	1-489	2789	13	32	348	39	2012-08-02 04:46:14.555671
2793	5	1-668	2789	32	18	1447	6	2012-08-02 04:46:14.555671
2789	5	1-897	2789	18	93	2292	49	2012-08-02 04:46:14.555671
2795	7	1-781	2794	93	1	3904	48	2012-08-02 04:46:14.555671
2796	7	1-826	2794	1	10	1629	56	2012-08-02 04:46:14.555671
2797	7	1-675	2794	10	59	2083	77	2012-08-02 04:46:14.555671
2798	7	1-853	2794	59	45	1498	13	2012-08-02 04:46:14.555671
2799	7	1-767	2794	45	8	1215	99	2012-08-02 04:46:14.555671
2800	7	1-845	2794	8	18	998	6	2012-08-02 04:46:14.555671
2794	7	1-897	2794	18	93	1610	49	2012-08-02 04:46:14.555671
2802	8	1-428	2801	34	10	976	41	2012-08-02 04:46:14.555671
2803	8	1-895	2801	10	39	550	80	2012-08-02 04:46:14.555671
2804	8	1-782	2801	39	76	457	88	2012-08-02 04:46:14.555671
2805	8	1-892	2801	76	20	940	42	2012-08-02 04:46:14.555671
2806	8	1-878	2801	20	17	844	21	2012-08-02 04:46:14.555671
2807	8	1-875	2801	17	8	659	99	2012-08-02 04:46:14.555671
2808	8	1-845	2801	8	18	554	6	2012-08-02 04:46:14.555671
2801	8	1-897	2801	18	34	915	49	2012-08-02 04:46:14.555671
2810	8	1-440	2809	14	57	627	94	2012-08-02 04:46:14.555671
2811	8	1-798	2809	57	70	2296	100	2012-08-02 04:46:14.555671
2812	8	1-762	2809	70	57	1840	40	2012-08-02 04:46:14.555671
2813	8	1-651	2809	57	56	3717	25	2012-08-02 04:46:14.555671
2814	8	1-854	2809	56	94	3909	22	2012-08-02 04:46:14.555671
2815	8	1-722	2809	94	96	982	33	2012-08-02 04:46:14.555671
2816	8	1-827	2809	96	18	705	6	2012-08-02 04:46:14.555671
2809	8	1-897	2809	18	14	1211	49	2012-08-02 04:46:14.555671
2818	7	1-428	2817	34	32	952	41	2012-08-02 04:46:14.555671
2819	7	1-36	2817	32	26	1057	51	2012-08-02 04:46:14.555671
2820	7	1-597	2817	26	15	584	68	2012-08-02 04:46:14.555671
2821	7	1-887	2817	15	52	480	63	2012-08-02 04:46:14.555671
2822	7	1-846	2817	52	96	626	33	2012-08-02 04:46:14.555671
2823	7	1-827	2817	96	18	465	6	2012-08-02 04:46:14.555671
2817	7	1-897	2817	18	34	828	49	2012-08-02 04:46:14.555671
2825	8	1-440	2824	14	31	526	94	2012-08-02 04:46:14.555671
2826	8	1-810	2824	31	85	359	33	2012-08-02 04:46:14.555671
2827	8	1-821	2824	85	64	467	89	2012-08-02 04:46:14.555671
2828	8	1-330	2824	64	20	494	42	2012-08-02 04:46:14.555671
2829	8	1-878	2824	20	17	519	21	2012-08-02 04:46:14.555671
2830	8	1-875	2824	17	8	474	99	2012-08-02 04:46:14.555671
2831	8	1-845	2824	8	18	467	6	2012-08-02 04:46:14.555671
2824	8	1-897	2824	18	14	903	49	2012-08-02 04:46:14.555671
2833	8	1-884	2832	83	88	1806	24	2012-08-02 04:46:23.227736
2834	8	1-634	2832	88	10	1106	84	2012-08-02 04:46:23.227736
2835	8	1-850	2832	10	2	3880	57	2012-08-02 04:46:23.227736
2836	8	1-796	2832	2	3	2377	39	2012-08-02 04:46:23.227736
2837	8	1-647	2832	3	26	2030	32	2012-08-02 04:46:23.227736
2838	8	1-840	2832	26	23	2120	87	2012-08-02 04:46:23.227736
2839	8	1-605	2832	23	63	5559	4	2012-08-02 04:46:23.227736
2832	8	1-911	2832	63	83	5667	71	2012-08-02 04:46:23.227736
2841	5	1-130	2840	73	7	91	69	2012-08-02 04:46:23.227736
2842	5	1-60	2840	7	26	88	32	2012-08-02 04:46:23.227736
2843	5	1-840	2840	26	23	100	87	2012-08-02 04:46:23.227736
2844	5	1-605	2840	23	63	284	4	2012-08-02 04:46:23.227736
2840	5	1-911	2840	63	73	314	71	2012-08-02 04:46:23.227736
2846	8	1-130	2845	73	42	11	69	2012-08-02 04:46:23.227736
2847	8	1-622	2845	42	46	3	15	2012-08-02 04:46:23.227736
2848	8	1-313	2845	46	25	2	44	2012-08-02 04:46:23.227736
2849	8	1-389	2845	25	18	2	6	2012-08-02 04:46:23.227736
2850	8	1-897	2845	18	47	3	49	2012-08-02 04:46:23.227736
2851	8	1-800	2845	47	23	5	87	2012-08-02 04:46:23.227736
2852	8	1-605	2845	23	63	19	4	2012-08-02 04:46:23.227736
2845	8	1-911	2845	63	73	28	71	2012-08-02 04:46:23.227736
2854	8	1-177	2853	41	23	245	61	2012-08-02 04:46:23.227736
2855	8	1-146	2853	23	100	241	27	2012-08-02 04:46:23.227736
2856	8	1-503	2853	100	69	116	87	2012-08-02 04:46:23.227736
2857	8	1-712	2853	69	92	105	66	2012-08-02 04:46:23.227736
2858	8	1-711	2853	92	42	45	2	2012-08-02 04:46:23.227736
2859	8	1-122	2853	42	100	147	50	2012-08-02 04:46:23.227736
2860	8	1-765	2853	100	63	524	4	2012-08-02 04:46:23.227736
2853	8	1-911	2853	63	41	832	71	2012-08-02 04:46:23.227736
2861	1	1-313	\N	46	46	1110	44	2012-08-02 04:46:23.227736
2862	1	1-897	\N	18	18	288	49	2012-08-02 04:46:23.227736
2864	4	1-757	2863	5	2	2244	57	2012-08-02 04:46:28.497479
2865	4	1-796	2863	2	40	1710	39	2012-08-02 04:46:28.497479
2866	4	1-332	2863	40	79	1322	79	2012-08-02 04:46:28.497479
2863	4	1-921	2863	79	5	3691	7	2012-08-02 04:46:28.497479
2868	7	1-842	2867	70	33	1068	53	2012-08-02 04:46:28.497479
2869	7	1-901	2867	33	8	1752	5	2012-08-02 04:46:28.497479
2870	7	1-864	2867	8	24	863	89	2012-08-02 04:46:28.497479
2871	7	1-891	2867	24	73	577	98	2012-08-02 04:46:28.497479
2872	7	1-772	2867	73	18	653	88	2012-08-02 04:46:28.497479
2873	7	1-881	2867	18	79	497	79	2012-08-02 04:46:28.497479
2867	7	1-921	2867	79	70	1761	7	2012-08-02 04:46:28.497479
2875	6	1-842	2874	70	33	2063	53	2012-08-02 04:46:28.497479
2876	6	1-901	2874	33	44	3922	5	2012-08-02 04:46:28.497479
2877	6	1-278	2874	44	30	633	70	2012-08-02 04:46:28.497479
2878	6	1-785	2874	30	23	922	71	2012-08-02 04:46:28.497479
2879	6	1-913	2874	23	79	715	79	2012-08-02 04:46:28.497479
2874	6	1-921	2874	79	70	2934	7	2012-08-02 04:46:28.497479
2881	5	1-881	2880	18	79	182	79	2012-08-02 04:46:34.507368
2882	5	1-921	2880	79	42	534	7	2012-08-02 04:46:34.507368
2883	5	1-402	2880	42	35	417	95	2012-08-02 04:46:34.507368
2884	5	1-620	2880	35	97	280	31	2012-08-02 04:46:34.507368
2880	5	1-935	2880	97	18	287	88	2012-08-02 04:46:34.507368
2886	3	1-92	2885	40	57	2167	40	2012-08-02 04:46:42.229238
2887	3	1-651	2885	57	38	1718	25	2012-08-02 04:46:42.229238
2885	3	1-948	2885	38	40	3778	67	2012-08-02 04:46:42.229238
2889	5	1-134	2888	70	23	124	58	2012-08-02 04:46:45.023656
2890	5	1-488	2888	23	15	119	41	2012-08-02 04:46:45.023656
2891	5	1-876	2888	15	37	1442	90	2012-08-02 04:46:45.023656
2892	5	1-917	2888	37	28	2239	20	2012-08-02 04:46:45.023656
2888	5	1-952	2888	28	70	178	85	2012-08-02 04:46:45.023656
2894	7	1-445	2893	48	17	1467	21	2012-08-02 04:46:45.023656
2895	7	1-875	2893	17	48	1302	99	2012-08-02 04:46:45.023656
2896	7	1-738	2893	48	32	215	95	2012-08-02 04:46:45.023656
2897	7	1-940	2893	32	15	133	41	2012-08-02 04:46:45.023656
2898	7	1-876	2893	15	37	2541	90	2012-08-02 04:46:45.023656
2899	7	1-917	2893	37	28	6224	20	2012-08-02 04:46:45.023656
2893	7	1-952	2893	28	48	781	85	2012-08-02 04:46:45.023656
2900	1	1-483	\N	9	9	1979	99	2012-08-02 04:46:45.023656
2902	5	1-924	2901	47	100	631	98	2012-08-02 04:46:47.122735
2903	5	1-934	2901	100	42	826	7	2012-08-02 04:46:47.122735
2904	5	1-402	2901	42	32	656	95	2012-08-02 04:46:47.122735
2905	5	1-940	2901	32	37	309	41	2012-08-02 04:46:47.122735
2901	5	1-956	2901	37	47	298	92	2012-08-02 04:46:47.122735
2907	8	1-710	2906	73	69	434	17	2012-08-02 04:46:47.763437
2908	8	1-618	2906	69	21	364	59	2012-08-02 04:46:47.763437
2909	8	1-900	2906	21	53	12	28	2012-08-02 04:46:47.763437
2910	8	1-583	2906	53	47	7	62	2012-08-02 04:46:47.763437
2911	8	1-66	2906	47	1	4	44	2012-08-02 04:46:47.763437
2912	8	1-926	2906	1	55	3	84	2012-08-02 04:46:47.763437
2913	8	1-795	2906	55	73	4	5	2012-08-02 04:46:47.763437
2906	8	1-957	2906	73	73	278	68	2012-08-02 04:46:47.763437
2914	8	1-958	2914	96	29	1301	85	2012-08-02 04:46:48.868228
2915	8	1-771	2914	29	59	1627	83	2012-08-02 04:46:48.868228
2916	8	1-912	2914	59	63	1030	46	2012-08-02 04:46:48.868228
2917	8	1-748	2914	63	25	951	48	2012-08-02 04:46:48.868228
2918	8	1-692	2914	25	59	538	55	2012-08-02 04:46:48.868228
2919	8	1-893	2914	59	37	312	90	2012-08-02 04:46:48.868228
2920	8	1-917	2914	37	83	544	20	2012-08-02 04:46:48.868228
2921	8	1-473	2914	83	96	1130	12	2012-08-02 04:46:48.868228
15902	4	1-2196	15901	14	31	249	4	2012-08-02 05:13:15.372371
2923	6	1-134	2922	70	12	1060	58	2012-08-02 04:46:48.868228
2924	6	1-925	2922	12	45	5557	81	2012-08-02 04:46:48.868228
2925	6	1-886	2922	45	75	919	53	2012-08-02 04:46:48.868228
2926	6	1-216	2922	75	64	861	97	2012-08-02 04:46:48.868228
2927	6	1-477	2922	64	96	958	12	2012-08-02 04:46:48.868228
2922	6	1-958	2922	96	70	1225	85	2012-08-02 04:46:48.868228
15903	4	1-2965	15901	31	99	255	12	2012-08-02 05:13:15.372371
2929	7	1-96	2928	13	93	1982	39	2012-08-02 04:46:48.868228
2930	7	1-938	2928	93	92	2477	66	2012-08-02 04:46:48.868228
2931	7	1-711	2928	92	42	938	2	2012-08-02 04:46:48.868228
2932	7	1-122	2928	42	88	2772	50	2012-08-02 04:46:48.868228
2933	7	1-730	2928	88	64	1579	97	2012-08-02 04:46:48.868228
2934	7	1-477	2928	64	96	1771	12	2012-08-02 04:46:48.868228
2928	7	1-958	2928	96	13	2282	85	2012-08-02 04:46:48.868228
15904	4	1-4557	15901	99	11	1004	43	2012-08-02 05:13:15.372371
2936	7	1-445	2935	48	17	1675	21	2012-08-02 04:46:48.868228
2937	7	1-875	2935	17	66	1230	99	2012-08-02 04:46:48.868228
2938	7	1-288	2935	66	25	183	75	2012-08-02 04:46:48.868228
2939	7	1-502	2935	25	13	203	61	2012-08-02 04:46:48.868228
2940	7	1-613	2935	13	83	340	20	2012-08-02 04:46:48.868228
2941	7	1-473	2935	83	96	821	12	2012-08-02 04:46:48.868228
2935	7	1-958	2935	96	48	1100	85	2012-08-02 04:46:48.868228
2943	8	1-771	2942	29	59	1207	83	2012-08-02 04:46:48.868228
2944	8	1-912	2942	59	98	868	46	2012-08-02 04:46:48.868228
2945	8	1-673	2942	98	75	365	1	2012-08-02 04:46:48.868228
2946	8	1-695	2942	75	94	376	26	2012-08-02 04:46:48.868228
2947	8	1-311	2942	94	75	267	53	2012-08-02 04:46:48.868228
2948	8	1-216	2942	75	64	320	97	2012-08-02 04:46:48.868228
2949	8	1-477	2942	64	96	442	12	2012-08-02 04:46:48.868228
2942	8	1-958	2942	96	29	701	85	2012-08-02 04:46:48.868228
2951	7	1-96	2950	13	32	236	39	2012-08-02 04:46:48.868228
2952	7	1-668	2950	32	66	1173	6	2012-08-02 04:46:48.868228
2953	7	1-766	2950	66	45	370	81	2012-08-02 04:46:48.868228
2954	7	1-886	2950	45	75	78	53	2012-08-02 04:46:48.868228
2955	7	1-216	2950	75	64	93	97	2012-08-02 04:46:48.868228
2956	7	1-477	2950	64	96	132	12	2012-08-02 04:46:48.868228
2950	7	1-958	2950	96	13	215	85	2012-08-02 04:46:48.868228
2958	6	1-771	2957	29	82	383	83	2012-08-02 04:46:48.868228
2959	6	1-249	2957	82	55	97	84	2012-08-02 04:46:48.868228
2960	6	1-795	2957	55	21	153	5	2012-08-02 04:46:48.868228
2961	6	1-141	2957	21	83	43	20	2012-08-02 04:46:48.868228
2962	6	1-473	2957	83	96	128	12	2012-08-02 04:46:48.868228
2957	6	1-958	2957	96	29	212	85	2012-08-02 04:46:48.868228
2964	5	1-889	2963	11	40	848	64	2012-08-02 04:46:53.013082
2965	5	1-51	2963	40	23	230	7	2012-08-02 04:46:53.013082
2966	5	1-656	2963	23	98	926	51	2012-08-02 04:46:53.013082
2967	5	1-919	2963	98	60	2166	45	2012-08-02 04:46:53.013082
2963	5	1-966	2963	60	11	6529	44	2012-08-02 04:46:53.013082
2969	6	1-831	2968	55	41	453	71	2012-08-02 04:46:55.112006
2970	6	1-177	2968	41	13	101	61	2012-08-02 04:46:55.112006
2971	6	1-613	2968	13	83	138	20	2012-08-02 04:46:55.112006
2972	6	1-473	2968	83	96	272	12	2012-08-02 04:46:55.112006
2973	6	1-958	2968	96	79	297	85	2012-08-02 04:46:55.112006
2968	6	1-970	2968	79	55	170	67	2012-08-02 04:46:55.112006
2975	6	1-398	2974	99	37	591	41	2012-08-02 04:46:57.045443
2976	6	1-956	2974	37	47	561	92	2012-08-02 04:46:57.045443
2977	6	1-924	2974	47	52	1165	98	2012-08-02 04:46:57.045443
2978	6	1-363	2974	52	55	1089	84	2012-08-02 04:46:57.045443
2979	6	1-795	2974	55	36	1246	5	2012-08-02 04:46:57.045443
2974	6	1-972	2974	36	99	1051	73	2012-08-02 04:46:57.045443
2981	8	1-64	2980	42	19	54	36	2012-08-02 04:46:57.045443
2982	8	1-723	2980	19	23	46	7	2012-08-02 04:46:57.045443
2983	8	1-656	2980	23	98	305	51	2012-08-02 04:46:57.045443
2984	8	1-919	2980	98	19	1169	45	2012-08-02 04:46:57.045443
2985	8	1-551	2980	19	88	50	24	2012-08-02 04:46:57.045443
2986	8	1-634	2980	88	55	42	84	2012-08-02 04:46:57.045443
2987	8	1-795	2980	55	36	51	5	2012-08-02 04:46:57.045443
2980	8	1-972	2980	36	42	45	73	2012-08-02 04:46:57.045443
2989	8	1-784	2988	67	63	164	91	2012-08-02 04:46:57.045443
2990	8	1-915	2988	63	79	226	85	2012-08-02 04:46:57.045443
2991	8	1-970	2988	79	55	185	67	2012-08-02 04:46:57.045443
2992	8	1-831	2988	55	83	735	71	2012-08-02 04:46:57.045443
2993	8	1-884	2988	83	88	396	24	2012-08-02 04:46:57.045443
2994	8	1-634	2988	88	55	410	84	2012-08-02 04:46:57.045443
2995	8	1-795	2988	55	36	631	5	2012-08-02 04:46:57.045443
2988	8	1-972	2988	36	67	688	73	2012-08-02 04:46:57.045443
2997	7	1-784	2996	67	37	187	91	2012-08-02 04:46:57.045443
2998	7	1-543	2996	37	70	103	82	2012-08-02 04:46:57.045443
2999	7	1-515	2996	70	3	266	38	2012-08-02 04:46:57.045443
3000	7	1-733	2996	3	1	442	44	2012-08-02 04:46:57.045443
3001	7	1-926	2996	1	55	476	84	2012-08-02 04:46:57.045443
3002	7	1-795	2996	55	36	709	5	2012-08-02 04:46:57.045443
2996	7	1-972	2996	36	67	779	73	2012-08-02 04:46:57.045443
3004	6	1-326	3003	95	65	125	10	2012-08-02 04:46:58.106277
3005	6	1-749	3003	65	83	107	20	2012-08-02 04:46:58.106277
3006	6	1-473	3003	83	23	275	12	2012-08-02 04:46:58.106277
3007	6	1-888	3003	23	78	399	47	2012-08-02 04:46:58.106277
3008	6	1-150	3003	78	85	308	78	2012-08-02 04:46:58.106277
3003	6	1-973	3003	85	95	297	13	2012-08-02 04:46:58.106277
3010	3	1-326	3009	95	43	1248	10	2012-08-02 04:46:58.106277
3011	3	1-271	3009	43	85	3015	78	2012-08-02 04:46:58.106277
3009	3	1-973	3009	85	95	2942	13	2012-08-02 04:46:58.106277
3013	6	1-767	3012	45	8	705	99	2012-08-02 04:46:58.106277
3014	6	1-845	3012	8	14	647	6	2012-08-02 04:46:58.106277
3015	6	1-658	3012	14	100	772	27	2012-08-02 04:46:58.106277
3016	6	1-503	3012	100	60	370	87	2012-08-02 04:46:58.106277
3017	6	1-667	3012	60	85	743	78	2012-08-02 04:46:58.106277
3012	6	1-973	3012	85	45	779	13	2012-08-02 04:46:58.106277
15901	4	1-4584	15901	11	14	1218	48	2012-08-02 05:13:15.372371
3019	6	1-304	3018	79	95	68	19	2012-08-02 04:46:58.835141
3020	6	1-964	3018	95	64	517	97	2012-08-02 04:46:58.835141
3021	6	1-477	3018	64	23	520	12	2012-08-02 04:46:58.835141
3022	6	1-888	3018	23	33	611	47	2012-08-02 04:46:58.835141
3023	6	1-931	3018	33	1	326	6	2012-08-02 04:46:58.835141
3018	6	1-975	3018	1	79	109	44	2012-08-02 04:46:58.835141
3025	7	1-926	3024	1	15	723	84	2012-08-02 04:46:58.835141
3026	7	1-928	3024	15	95	143	19	2012-08-02 04:46:58.835141
3027	7	1-964	3024	95	64	1327	97	2012-08-02 04:46:58.835141
3028	7	1-477	3024	64	23	1633	12	2012-08-02 04:46:58.835141
3029	7	1-888	3024	23	33	2345	47	2012-08-02 04:46:58.835141
3030	7	1-931	3024	33	1	1532	6	2012-08-02 04:46:58.835141
3024	7	1-975	3024	1	1	683	44	2012-08-02 04:46:58.835141
15906	5	1-4249	15905	5	22	773	29	2012-08-02 05:13:16.598989
3032	6	1-722	3031	94	96	145	33	2012-08-02 04:47:01.718771
3033	6	1-827	3031	96	1	117	6	2012-08-02 04:47:01.718771
3034	6	1-975	3031	1	31	56	44	2012-08-02 04:47:01.718771
3035	6	1-904	3031	31	77	622	90	2012-08-02 04:47:01.718771
3036	6	1-328	3031	77	82	732	34	2012-08-02 04:47:01.718771
3031	6	1-980	3031	82	94	515	22	2012-08-02 04:47:01.718771
15907	5	1-4487	15905	22	60	797	13	2012-08-02 05:13:16.598989
3038	3	1-327	3037	74	57	326	40	2012-08-02 04:47:02.260036
3039	3	1-651	3037	57	54	85	25	2012-08-02 04:47:02.260036
3037	3	1-982	3037	54	74	2060	58	2012-08-02 04:47:02.260036
15908	5	1-4389	15905	60	21	292	21	2012-08-02 05:13:16.598989
3041	7	1-944	3040	13	33	1054	47	2012-08-02 04:47:02.260036
3042	7	1-931	3040	33	1	437	6	2012-08-02 04:47:02.260036
3043	7	1-975	3040	1	49	113	44	2012-08-02 04:47:02.260036
3044	7	1-750	3040	49	3	66	77	2012-08-02 04:47:02.260036
3045	7	1-79	3040	3	38	13	15	2012-08-02 04:47:02.260036
3046	7	1-282	3040	38	54	39	25	2012-08-02 04:47:02.260036
3040	7	1-982	3040	54	13	4154	58	2012-08-02 04:47:02.260036
3048	7	1-845	3047	8	1	1435	6	2012-08-02 04:47:02.746225
3049	7	1-975	3047	1	49	549	44	2012-08-02 04:47:02.746225
3050	7	1-750	3047	49	22	474	77	2012-08-02 04:47:02.746225
3051	7	1-322	3047	22	67	161	26	2012-08-02 04:47:02.746225
3052	7	1-968	3047	67	62	3920	32	2012-08-02 04:47:02.746225
3053	7	1-779	3047	62	95	2444	81	2012-08-02 04:47:02.746225
3047	7	1-983	3047	95	8	1831	99	2012-08-02 04:47:02.746225
3055	6	1-288	3054	66	89	537	75	2012-08-02 04:47:02.746225
3056	6	1-365	3054	89	22	572	77	2012-08-02 04:47:02.746225
3057	6	1-322	3054	22	67	215	26	2012-08-02 04:47:02.746225
3058	6	1-968	3054	67	62	5787	32	2012-08-02 04:47:02.746225
3059	6	1-779	3054	62	95	3992	81	2012-08-02 04:47:02.746225
3054	6	1-983	3054	95	66	3310	99	2012-08-02 04:47:02.746225
3060	1	1-365	\N	89	89	833	77	2012-08-02 04:47:02.746225
3062	8	1-871	3061	4	13	548	30	2012-08-02 04:47:04.160538
3063	8	1-666	3061	13	98	237	46	2012-08-02 04:47:04.160538
3064	8	1-673	3061	98	6	95	1	2012-08-02 04:47:04.160538
3065	8	1-303	3061	6	19	75	36	2012-08-02 04:47:04.160538
3066	8	1-723	3061	19	23	59	7	2012-08-02 04:47:04.160538
3067	8	1-656	3061	23	98	410	51	2012-08-02 04:47:04.160538
3068	8	1-919	3061	98	61	1868	45	2012-08-02 04:47:04.160538
3061	8	1-985	3061	61	4	789	62	2012-08-02 04:47:04.160538
3070	6	1-66	3069	47	11	217	44	2012-08-02 04:47:04.160538
3071	6	1-889	3069	11	40	51	64	2012-08-02 04:47:04.160538
3072	6	1-51	3069	40	23	25	7	2012-08-02 04:47:04.160538
3073	6	1-656	3069	23	98	181	51	2012-08-02 04:47:04.160538
3074	6	1-919	3069	98	61	767	45	2012-08-02 04:47:04.160538
3069	6	1-985	3069	61	47	337	62	2012-08-02 04:47:04.160538
3076	8	1-130	3075	73	42	1351	69	2012-08-02 04:47:04.657876
3077	8	1-622	3075	42	71	371	15	2012-08-02 04:47:04.657876
3078	8	1-804	3075	71	53	809	94	2012-08-02 04:47:04.657876
3079	8	1-942	3075	53	33	1944	47	2012-08-02 04:47:04.657876
3080	8	1-931	3075	33	1	1437	6	2012-08-02 04:47:04.657876
3081	8	1-975	3075	1	31	663	44	2012-08-02 04:47:04.657876
3082	8	1-904	3075	31	46	7099	90	2012-08-02 04:47:04.657876
3075	8	1-986	3075	46	73	2929	71	2012-08-02 04:47:04.657876
3084	7	1-796	3083	2	3	1459	39	2012-08-02 04:47:05.149686
3085	7	1-647	3083	3	26	1386	32	2012-08-02 04:47:05.149686
3086	7	1-840	3083	26	60	1611	87	2012-08-02 04:47:05.149686
3087	7	1-667	3083	60	85	2146	78	2012-08-02 04:47:05.149686
3088	7	1-973	3083	85	85	1911	13	2012-08-02 04:47:05.149686
3089	7	1-852	3083	85	17	1408	55	2012-08-02 04:47:05.149686
3083	7	1-987	3083	17	2	2140	57	2012-08-02 04:47:05.149686
3091	6	1-807	3090	30	100	822	27	2012-08-02 04:47:05.149686
3092	6	1-503	3090	100	60	280	87	2012-08-02 04:47:05.149686
3093	6	1-667	3090	60	85	400	78	2012-08-02 04:47:05.149686
3094	6	1-973	3090	85	85	369	13	2012-08-02 04:47:05.149686
3095	6	1-852	3090	85	17	281	55	2012-08-02 04:47:05.149686
3090	6	1-987	3090	17	30	459	57	2012-08-02 04:47:05.149686
3097	8	1-796	3096	2	32	50	39	2012-08-02 04:47:05.149686
3098	8	1-668	3096	32	66	191	6	2012-08-02 04:47:05.149686
3099	8	1-766	3096	66	95	47	81	2012-08-02 04:47:05.149686
3100	8	1-983	3096	95	93	33	99	2012-08-02 04:47:05.149686
3101	8	1-386	3096	93	94	16	42	2012-08-02 04:47:05.149686
3102	8	1-947	3096	94	94	26	54	2012-08-02 04:47:05.149686
3103	8	1-774	3096	94	17	32	55	2012-08-02 04:47:05.149686
3096	8	1-987	3096	17	2	59	57	2012-08-02 04:47:05.149686
3105	8	1-42	3104	34	78	1985	47	2012-08-02 04:47:05.149686
3106	8	1-150	3104	78	85	1505	78	2012-08-02 04:47:05.149686
3107	8	1-973	3104	85	45	1425	13	2012-08-02 04:47:05.149686
3108	8	1-767	3104	45	93	1166	99	2012-08-02 04:47:05.149686
3109	8	1-386	3104	93	94	631	42	2012-08-02 04:47:05.149686
3110	8	1-947	3104	94	94	1043	54	2012-08-02 04:47:05.149686
3111	8	1-774	3104	94	17	1352	55	2012-08-02 04:47:05.149686
3104	8	1-987	3104	17	34	2802	57	2012-08-02 04:47:05.149686
3113	5	1-796	3112	2	40	2160	39	2012-08-02 04:47:05.149686
3114	5	1-332	3112	40	72	2073	79	2012-08-02 04:47:05.149686
3115	5	1-959	3112	72	25	1549	48	2012-08-02 04:47:05.149686
3116	5	1-692	3112	25	17	1084	55	2012-08-02 04:47:05.149686
3112	5	1-987	3112	17	2	2285	57	2012-08-02 04:47:05.149686
3118	4	1-439	3117	89	61	716	45	2012-08-02 04:47:06.117672
3117	4	1-990	3117	33	89	556	61	2012-08-02 04:47:06.117672
3119	4	1-985	3117	61	47	293	62	2012-08-02 04:47:06.117672
3120	4	1-66	3117	47	33	176	44	2012-08-02 04:47:06.117672
15909	5	1-3051	15905	21	49	106	49	2012-08-02 05:13:16.598989
3122	6	1-364	3121	72	17	378	55	2012-08-02 04:47:06.117672
3123	6	1-987	3121	17	34	830	57	2012-08-02 04:47:06.117672
3124	6	1-42	3121	34	33	622	47	2012-08-02 04:47:06.117672
3125	6	1-931	3121	33	1	456	6	2012-08-02 04:47:06.117672
3126	6	1-975	3121	1	33	195	44	2012-08-02 04:47:06.117672
3121	6	1-990	3121	33	72	745	61	2012-08-02 04:47:06.117672
3128	6	1-613	3127	13	12	900	20	2012-08-02 04:47:06.117672
3129	6	1-12	3127	12	23	174	12	2012-08-02 04:47:06.117672
3130	6	1-888	3127	23	33	292	47	2012-08-02 04:47:06.117672
3131	6	1-931	3127	33	1	226	6	2012-08-02 04:47:06.117672
3132	6	1-975	3127	1	33	108	44	2012-08-02 04:47:06.117672
3127	6	1-990	3127	33	13	435	61	2012-08-02 04:47:06.117672
3134	7	1-883	3133	49	95	66	19	2012-08-02 04:47:06.604057
3135	7	1-964	3133	95	72	610	97	2012-08-02 04:47:06.604057
3136	7	1-347	3133	72	97	643	31	2012-08-02 04:47:06.604057
3137	7	1-935	3133	97	18	770	88	2012-08-02 04:47:06.604057
3138	7	1-881	3133	18	38	568	79	2012-08-02 04:47:06.604057
3139	7	1-617	3133	38	96	2165	34	2012-08-02 04:47:06.604057
3133	7	1-991	3133	96	49	110	41	2012-08-02 04:47:06.604057
3141	6	1-278	3140	44	71	102	70	2012-08-02 04:47:11.630623
3142	6	1-127	3140	71	95	223	81	2012-08-02 04:47:11.630623
3143	6	1-983	3140	95	13	173	99	2012-08-02 04:47:11.630623
3144	6	1-820	3140	13	41	74	13	2012-08-02 04:47:11.630623
3145	6	1-713	3140	41	23	237	65	2012-08-02 04:47:11.630623
3140	6	1-1000	3140	23	44	781	5	2012-08-02 04:47:11.630623
3147	7	1-401	3146	15	17	52	19	2012-08-02 04:47:11.630623
3148	7	1-992	3146	17	93	97	39	2012-08-02 04:47:11.630623
3149	7	1-938	3146	93	92	146	66	2012-08-02 04:47:11.630623
3150	7	1-711	3146	92	87	67	2	2012-08-02 04:47:11.630623
3151	7	1-186	3146	87	41	15	13	2012-08-02 04:47:11.630623
3152	7	1-713	3146	41	23	56	65	2012-08-02 04:47:11.630623
3146	7	1-1000	3146	23	15	209	5	2012-08-02 04:47:11.630623
3154	5	1-401	3153	15	88	1737	19	2012-08-02 04:47:11.630623
3155	5	1-787	3153	88	17	1035	26	2012-08-02 04:47:11.630623
3156	5	1-211	3153	17	82	280	86	2012-08-02 04:47:11.630623
3157	5	1-955	3153	82	23	1809	65	2012-08-02 04:47:11.630623
3153	5	1-1000	3153	23	15	6859	5	2012-08-02 04:47:11.630623
3159	8	1-432	3158	45	23	8	12	2012-08-02 04:47:11.630623
3160	8	1-888	3158	23	33	13	47	2012-08-02 04:47:11.630623
3161	8	1-931	3158	33	1	9	6	2012-08-02 04:47:11.630623
3162	8	1-975	3158	1	49	4	44	2012-08-02 04:47:11.630623
3163	8	1-750	3158	49	59	4	77	2012-08-02 04:47:11.630623
3164	8	1-853	3158	59	41	3	13	2012-08-02 04:47:11.630623
3165	8	1-713	3158	41	23	9	65	2012-08-02 04:47:11.630623
3158	8	1-1000	3158	23	45	35	5	2012-08-02 04:47:11.630623
3167	7	1-544	3166	55	96	339	34	2012-08-02 04:47:13.045032
3168	7	1-991	3166	96	37	20	41	2012-08-02 04:47:13.045032
3169	7	1-956	3166	37	47	26	92	2012-08-02 04:47:13.045032
3170	7	1-924	3166	47	100	73	98	2012-08-02 04:47:13.045032
3171	7	1-934	3166	100	51	127	7	2012-08-02 04:47:13.045032
3172	7	1-829	3166	51	29	511	33	2012-08-02 04:47:13.045032
3166	7	1-1004	3166	29	55	199	81	2012-08-02 04:47:13.045032
3174	2	1-883	3173	49	41	2446	19	2012-08-02 04:47:15.089031
3173	2	1-1008	3173	41	49	5854	41	2012-08-02 04:47:15.089031
3176	4	1-956	3175	37	53	3318	92	2012-08-02 04:47:15.089031
3177	4	1-396	3175	53	62	1176	29	2012-08-02 04:47:15.089031
3178	4	1-828	3175	62	41	958	19	2012-08-02 04:47:15.089031
3175	4	1-1008	3175	41	37	3127	41	2012-08-02 04:47:15.089031
3180	7	1-940	3179	32	37	396	41	2012-08-02 04:47:18.448843
3181	7	1-956	3179	37	47	353	92	2012-08-02 04:47:18.448843
3182	7	1-924	3179	47	100	689	98	2012-08-02 04:47:18.448843
3183	7	1-934	3179	100	23	832	7	2012-08-02 04:47:18.448843
3184	7	1-656	3179	23	71	4905	51	2012-08-02 04:47:18.448843
3185	7	1-725	3179	71	9	824	66	2012-08-02 04:47:18.448843
3179	7	1-1015	3179	9	32	913	95	2012-08-02 04:47:18.448843
3187	7	1-940	3186	32	37	589	41	2012-08-02 04:47:18.448843
3188	7	1-956	3186	37	53	702	92	2012-08-02 04:47:18.448843
3189	7	1-396	3186	53	62	280	29	2012-08-02 04:47:18.448843
3190	7	1-828	3186	62	17	256	19	2012-08-02 04:47:18.448843
3191	7	1-992	3186	17	93	465	39	2012-08-02 04:47:18.448843
3192	7	1-938	3186	93	9	685	66	2012-08-02 04:47:18.448843
3186	7	1-1015	3186	9	32	1015	95	2012-08-02 04:47:18.448843
3194	6	1-956	3193	37	47	1580	92	2012-08-02 04:47:19.443255
3195	6	1-924	3193	47	68	3377	98	2012-08-02 04:47:19.443255
3196	6	1-866	3193	68	95	1894	81	2012-08-02 04:47:19.443255
3197	6	1-983	3193	95	13	1332	99	2012-08-02 04:47:19.443255
3198	6	1-820	3193	13	40	518	13	2012-08-02 04:47:19.443255
3193	6	1-1018	3193	40	37	1617	41	2012-08-02 04:47:19.443255
3200	5	1-515	3199	70	3	454	38	2012-08-02 04:47:21.906704
3201	5	1-733	3199	3	33	768	44	2012-08-02 04:47:21.906704
3202	5	1-990	3199	33	72	2816	61	2012-08-02 04:47:21.906704
3203	5	1-364	3199	72	3	1465	55	2012-08-02 04:47:21.906704
3199	5	1-1023	3199	3	70	181	82	2012-08-02 04:47:21.906704
3205	6	1-804	3204	71	34	2478	94	2012-08-02 04:47:22.393403
3206	6	1-962	3204	34	9	1034	66	2012-08-02 04:47:22.393403
3207	6	1-1015	3204	9	94	905	95	2012-08-02 04:47:22.393403
3208	6	1-923	3204	94	70	6590	100	2012-08-02 04:47:22.393403
3209	6	1-762	3204	70	67	3198	40	2012-08-02 04:47:22.393403
3204	6	1-1024	3204	67	71	2023	15	2012-08-02 04:47:22.393403
3211	8	1-66	3210	47	49	565	44	2012-08-02 04:47:24.160261
3212	8	1-750	3210	49	59	621	77	2012-08-02 04:47:24.160261
3213	8	1-853	3210	59	45	542	13	2012-08-02 04:47:24.160261
3214	8	1-767	3210	45	93	533	99	2012-08-02 04:47:24.160261
3215	8	1-386	3210	93	94	347	42	2012-08-02 04:47:24.160261
3216	8	1-947	3210	94	81	628	54	2012-08-02 04:47:24.160261
3217	8	1-734	3210	81	8	449	46	2012-08-02 04:47:24.160261
3210	8	1-1026	3210	8	47	729	62	2012-08-02 04:47:24.160261
3218	1	1-66	\N	47	47	3013	44	2012-08-02 04:47:24.160261
3220	7	1-947	3219	94	94	1574	54	2012-08-02 04:47:24.657558
3221	7	1-774	3219	94	3	2133	55	2012-08-02 04:47:24.657558
3222	7	1-1023	3219	3	70	252	82	2012-08-02 04:47:24.657558
3223	7	1-515	3219	70	29	637	38	2012-08-02 04:47:24.657558
3219	7	1-1027	3219	34	94	910	42	2012-08-02 04:47:24.657558
3224	7	1-961	3219	29	63	5349	4	2012-08-02 04:47:24.657558
3225	7	1-967	3219	63	34	406	39	2012-08-02 04:47:24.657558
15905	5	1-4587	15905	49	5	1040	63	2012-08-02 05:13:16.598989
3227	7	1-947	3226	94	94	496	54	2012-08-02 04:47:24.657558
3228	7	1-774	3226	94	3	688	55	2012-08-02 04:47:24.657558
3229	7	1-1023	3226	3	39	85	82	2012-08-02 04:47:24.657558
3230	7	1-574	3226	39	62	65	29	2012-08-02 04:47:24.657558
3231	7	1-828	3226	62	17	63	19	2012-08-02 04:47:24.657558
3232	7	1-992	3226	17	34	120	39	2012-08-02 04:47:24.657558
3226	7	1-1027	3226	34	94	281	42	2012-08-02 04:47:24.657558
3234	8	1-245	3233	55	23	369	12	2012-08-02 04:47:25.154905
3235	8	1-888	3233	23	33	560	47	2012-08-02 04:47:25.154905
3236	8	1-931	3233	33	1	386	6	2012-08-02 04:47:25.154905
3237	8	1-975	3233	1	49	166	44	2012-08-02 04:47:25.154905
3238	8	1-750	3233	49	3	162	77	2012-08-02 04:47:25.154905
3239	8	1-79	3233	3	38	55	15	2012-08-02 04:47:25.154905
3240	8	1-282	3233	38	54	266	25	2012-08-02 04:47:25.154905
3233	8	1-1028	3233	54	55	4269	80	2012-08-02 04:47:25.154905
3241	1	1-975	\N	1	1	1346	44	2012-08-02 04:47:25.154905
15911	5	1-3912	15910	44	42	1216	55	2012-08-02 05:13:16.598989
3243	6	1-1027	3242	34	67	6629	42	2012-08-02 04:47:26.105107
3244	6	1-908	3242	67	4	4071	30	2012-08-02 04:47:26.105107
3245	6	1-848	3242	4	82	727	86	2012-08-02 04:47:26.105107
3246	6	1-955	3242	82	34	2303	65	2012-08-02 04:47:26.105107
3247	6	1-898	3242	34	15	1550	43	2012-08-02 04:47:26.105107
3242	6	1-1030	3242	15	34	3969	39	2012-08-02 04:47:26.105107
3249	6	1-938	3248	93	79	47	66	2012-08-02 04:47:26.105107
3250	6	1-1012	3248	79	72	83	79	2012-08-02 04:47:26.105107
3251	6	1-959	3248	72	25	51	48	2012-08-02 04:47:26.105107
3252	6	1-692	3248	25	43	29	55	2012-08-02 04:47:26.105107
3253	6	1-55	3248	43	15	11	43	2012-08-02 04:47:26.105107
3248	6	1-1030	3248	15	93	42	39	2012-08-02 04:47:26.105107
3255	3	1-950	3254	59	37	3163	45	2012-08-02 04:47:26.105107
3256	3	1-624	3254	37	15	1133	43	2012-08-02 04:47:26.105107
3254	3	1-1030	3254	15	59	5032	39	2012-08-02 04:47:26.105107
3258	6	1-149	3257	41	67	1946	40	2012-08-02 04:47:27.519442
3259	6	1-1024	3257	67	32	1765	15	2012-08-02 04:47:27.519442
3260	6	1-805	3257	32	59	722	83	2012-08-02 04:47:27.519442
3261	6	1-912	3257	59	8	421	46	2012-08-02 04:47:27.519442
3262	6	1-1026	3257	8	9	520	62	2012-08-02 04:47:27.519442
3257	6	1-1033	3257	9	41	1455	56	2012-08-02 04:47:27.519442
3264	8	1-705	3263	48	55	1379	67	2012-08-02 04:47:27.519442
3265	8	1-831	3263	55	73	4361	71	2012-08-02 04:47:27.519442
3266	8	1-130	3263	73	44	1631	69	2012-08-02 04:47:27.519442
3267	8	1-857	3263	44	67	1231	42	2012-08-02 04:47:27.519442
3268	8	1-908	3263	67	13	1268	30	2012-08-02 04:47:27.519442
3269	8	1-666	3263	13	8	528	46	2012-08-02 04:47:27.519442
3270	8	1-1026	3263	8	9	658	62	2012-08-02 04:47:27.519442
3263	8	1-1033	3263	9	48	1854	56	2012-08-02 04:47:27.519442
3272	7	1-705	3271	48	40	2369	67	2012-08-02 04:47:27.519442
3273	7	1-92	3271	40	67	2986	40	2012-08-02 04:47:27.519442
3274	7	1-1024	3271	67	32	2877	15	2012-08-02 04:47:27.519442
3275	7	1-805	3271	32	59	1250	83	2012-08-02 04:47:27.519442
3276	7	1-912	3271	59	8	774	46	2012-08-02 04:47:27.519442
3277	7	1-1026	3271	8	9	1017	62	2012-08-02 04:47:27.519442
3271	7	1-1033	3271	9	48	3021	56	2012-08-02 04:47:27.519442
3279	6	1-675	3278	10	22	610	77	2012-08-02 04:47:27.519442
3280	6	1-322	3278	22	24	233	26	2012-08-02 04:47:27.519442
3281	6	1-971	3278	24	37	330	11	2012-08-02 04:47:27.519442
3282	6	1-516	3278	37	8	96	46	2012-08-02 04:47:27.519442
3283	6	1-1026	3278	8	9	137	62	2012-08-02 04:47:27.519442
3278	6	1-1033	3278	9	10	445	56	2012-08-02 04:47:27.519442
3285	8	1-705	3284	48	82	2207	67	2012-08-02 04:47:27.519442
3286	8	1-758	3284	82	61	3437	98	2012-08-02 04:47:27.519442
3287	8	1-266	3284	61	46	2032	58	2012-08-02 04:47:27.519442
3288	8	1-995	3284	46	64	1430	75	2012-08-02 04:47:27.519442
3289	8	1-116	3284	64	59	647	83	2012-08-02 04:47:27.519442
3290	8	1-912	3284	59	8	464	46	2012-08-02 04:47:27.519442
3291	8	1-1026	3284	8	9	706	62	2012-08-02 04:47:27.519442
3284	8	1-1033	3284	9	48	2430	56	2012-08-02 04:47:27.519442
3293	7	1-724	3292	34	70	81	100	2012-08-02 04:47:27.519442
3294	7	1-762	3292	70	67	70	40	2012-08-02 04:47:27.519442
3295	7	1-1024	3292	67	32	79	15	2012-08-02 04:47:27.519442
3296	7	1-805	3292	32	59	40	83	2012-08-02 04:47:27.519442
3297	7	1-912	3292	59	8	29	46	2012-08-02 04:47:27.519442
3298	7	1-1026	3292	8	9	45	62	2012-08-02 04:47:27.519442
3292	7	1-1033	3292	9	34	156	56	2012-08-02 04:47:27.519442
3300	6	1-428	3299	34	37	537	41	2012-08-02 04:47:28.017119
3301	6	1-956	3299	37	47	654	92	2012-08-02 04:47:28.017119
3302	6	1-924	3299	47	68	1745	98	2012-08-02 04:47:28.017119
3303	6	1-866	3299	68	95	1221	81	2012-08-02 04:47:28.017119
3304	6	1-983	3299	95	79	1071	99	2012-08-02 04:47:28.017119
3299	6	1-1034	3299	79	34	462	49	2012-08-02 04:47:28.017119
3306	6	1-1003	3305	30	73	343	63	2012-08-02 04:47:28.503375
3307	6	1-539	3305	73	24	669	86	2012-08-02 04:47:28.503375
3308	6	1-1007	3305	24	83	473	60	2012-08-02 04:47:28.503375
3309	6	1-626	3305	83	9	524	1	2012-08-02 04:47:28.503375
3310	6	1-182	3305	9	22	70	32	2012-08-02 04:47:28.503375
3305	6	1-1035	3305	22	30	125	57	2012-08-02 04:47:28.503375
3312	5	1-1003	3311	30	73	1706	63	2012-08-02 04:47:28.503375
3313	5	1-539	3311	73	24	3431	86	2012-08-02 04:47:28.503375
3314	5	1-1007	3311	24	82	2499	60	2012-08-02 04:47:28.503375
3315	5	1-159	3311	82	22	328	32	2012-08-02 04:47:28.503375
3311	5	1-1035	3311	22	30	604	57	2012-08-02 04:47:28.503375
3317	7	1-1003	3316	30	2	3038	63	2012-08-02 04:47:28.503375
3318	7	1-937	3316	2	49	3423	28	2012-08-02 04:47:28.503375
3319	7	1-72	3316	49	35	1157	8	2012-08-02 04:47:28.503375
3320	7	1-43	3316	35	67	1023	40	2012-08-02 04:47:28.503375
3321	7	1-1024	3316	67	50	1167	15	2012-08-02 04:47:28.503375
3322	7	1-589	3316	50	22	479	32	2012-08-02 04:47:28.503375
3316	7	1-1035	3316	22	30	973	57	2012-08-02 04:47:28.503375
3324	7	1-983	3323	95	79	324	99	2012-08-02 04:47:29.464544
3325	7	1-1034	3323	79	93	140	49	2012-08-02 04:47:29.464544
3326	7	1-781	3323	93	25	378	48	2012-08-02 04:47:29.464544
3323	7	1-1036	3323	91	95	370	81	2012-08-02 04:47:29.464544
3327	7	1-692	3323	25	3	288	55	2012-08-02 04:47:29.464544
3328	7	1-1023	3323	3	70	34	82	2012-08-02 04:47:29.464544
3329	7	1-515	3323	70	91	88	38	2012-08-02 04:47:29.464544
15912	5	1-3929	15910	42	1	734	23	2012-08-02 05:13:16.598989
3331	5	1-576	3330	80	3	130	55	2012-08-02 04:47:31.032912
3332	5	1-1023	3330	3	70	14	82	2012-08-02 04:47:31.032912
3333	5	1-515	3330	70	27	31	38	2012-08-02 04:47:31.032912
3334	5	1-746	3330	27	90	421	37	2012-08-02 04:47:31.032912
3330	5	1-1038	3330	90	80	476	11	2012-08-02 04:47:31.032912
3336	4	1-592	3335	86	9	1515	66	2012-08-02 04:47:31.949857
3337	4	1-1015	3335	9	32	1935	95	2012-08-02 04:47:31.949857
3338	4	1-940	3335	32	78	967	41	2012-08-02 04:47:31.949857
3335	4	1-1040	3335	78	86	6894	39	2012-08-02 04:47:31.949857
3340	5	1-995	3339	46	50	1706	75	2012-08-02 04:47:34.281021
3341	5	1-69	3339	50	24	1404	74	2012-08-02 04:47:34.281021
3342	5	1-994	3339	24	23	2175	12	2012-08-02 04:47:34.281021
3343	5	1-888	3339	23	12	3058	47	2012-08-02 04:47:34.281021
3339	5	1-1043	3339	12	46	2772	58	2012-08-02 04:47:34.281021
3345	6	1-995	3344	46	60	262	75	2012-08-02 04:47:34.281021
3346	6	1-920	3344	60	37	66	91	2012-08-02 04:47:34.281021
3347	6	1-543	3344	37	3	37	82	2012-08-02 04:47:34.281021
3348	6	1-430	3344	3	89	47	35	2012-08-02 04:47:34.281021
3349	6	1-514	3344	89	12	351	47	2012-08-02 04:47:34.281021
3344	6	1-1043	3344	12	46	368	58	2012-08-02 04:47:34.281021
3351	7	1-983	3350	95	93	6	99	2012-08-02 04:47:36.291462
3352	7	1-386	3350	93	67	3	42	2012-08-02 04:47:36.291462
3353	7	1-908	3350	67	13	4	30	2012-08-02 04:47:36.291462
3354	7	1-666	3350	13	23	2	46	2012-08-02 04:47:36.291462
3355	7	1-790	3350	23	74	6	76	2012-08-02 04:47:36.291462
3356	7	1-414	3350	74	11	2	27	2012-08-02 04:47:36.291462
3350	7	1-1046	3350	11	95	8	81	2012-08-02 04:47:36.291462
3358	5	1-983	3357	95	66	1384	99	2012-08-02 04:47:36.291462
3359	5	1-288	3357	66	25	235	75	2012-08-02 04:47:36.291462
3360	5	1-502	3357	25	23	298	61	2012-08-02 04:47:36.291462
3361	5	1-146	3357	23	11	313	27	2012-08-02 04:47:36.291462
3357	5	1-1046	3357	11	95	1596	81	2012-08-02 04:47:36.291462
3363	8	1-622	3362	42	38	319	15	2012-08-02 04:47:38.047949
3364	8	1-282	3362	38	22	1541	25	2012-08-02 04:47:38.047949
3365	8	1-1025	3362	22	79	1796	66	2012-08-02 04:47:38.047949
3366	8	1-1012	3362	79	72	4041	79	2012-08-02 04:47:38.047949
3367	8	1-959	3362	72	25	3158	48	2012-08-02 04:47:38.047949
3368	8	1-692	3362	25	59	2311	55	2012-08-02 04:47:38.047949
3369	8	1-893	3362	59	15	1433	90	2012-08-02 04:47:38.047949
3362	8	1-1049	3362	15	42	1247	69	2012-08-02 04:47:38.047949
3371	4	1-1003	3370	30	2	4389	63	2012-08-02 04:47:41.361649
3372	4	1-937	3370	2	53	3129	28	2012-08-02 04:47:41.361649
3373	4	1-583	3370	53	13	1616	62	2012-08-02 04:47:41.361649
3370	4	1-1053	3370	13	30	2223	57	2012-08-02 04:47:41.361649
3374	1	1-583	\N	53	53	444	62	2012-08-02 04:47:41.361649
3376	8	1-679	3375	4	69	1241	87	2012-08-02 04:47:43.228679
3377	8	1-712	3375	69	9	1238	66	2012-08-02 04:47:43.228679
3378	8	1-1015	3375	9	56	1917	95	2012-08-02 04:47:43.228679
3379	8	1-1031	3375	56	83	2021	33	2012-08-02 04:47:43.228679
3380	8	1-729	3375	83	78	1742	9	2012-08-02 04:47:43.228679
3381	8	1-1001	3375	78	45	3813	5	2012-08-02 04:47:43.228679
3382	8	1-432	3375	45	73	912	12	2012-08-02 04:47:43.228679
3375	8	1-1056	3375	73	4	1098	94	2012-08-02 04:47:43.228679
3384	7	1-828	3383	62	95	50	19	2012-08-02 04:47:45.93507
3385	7	1-964	3383	95	72	429	97	2012-08-02 04:47:45.93507
3386	7	1-347	3383	72	65	418	31	2012-08-02 04:47:45.93507
3387	7	1-1041	3383	65	17	176	21	2012-08-02 04:47:45.93507
3388	7	1-875	3383	17	14	126	99	2012-08-02 04:47:45.93507
3389	7	1-770	3383	14	91	43	16	2012-08-02 04:47:45.93507
3383	7	1-1059	3383	91	62	63	29	2012-08-02 04:47:45.93507
3391	5	1-951	3390	87	87	1380	36	2012-08-02 04:47:51.878081
3392	5	1-233	3390	87	53	2084	94	2012-08-02 04:47:51.878081
3393	5	1-942	3390	53	12	4568	47	2012-08-02 04:47:51.878081
3394	5	1-1043	3390	12	99	4362	58	2012-08-02 04:47:51.878081
3390	5	1-1068	3390	99	87	4998	74	2012-08-02 04:47:51.878081
3396	7	1-836	3395	72	15	672	90	2012-08-02 04:47:51.878081
3397	7	1-1049	3395	15	42	610	69	2012-08-02 04:47:51.878081
3398	7	1-622	3395	42	71	163	15	2012-08-02 04:47:51.878081
3399	7	1-804	3395	71	53	346	94	2012-08-02 04:47:51.878081
3400	7	1-942	3395	53	12	810	47	2012-08-02 04:47:51.878081
3401	7	1-1043	3395	12	99	825	58	2012-08-02 04:47:51.878081
3395	7	1-1068	3395	99	72	1009	74	2012-08-02 04:47:51.878081
3402	1	1-804	\N	71	71	2354	94	2012-08-02 04:47:51.878081
3404	3	1-391	3403	11	95	89	19	2012-08-02 04:47:52.872378
3405	3	1-964	3403	95	95	540	97	2012-08-02 04:47:52.872378
3403	3	1-1070	3403	95	11	528	82	2012-08-02 04:47:52.872378
3407	8	1-965	3406	75	94	177	54	2012-08-02 04:47:54.253317
3408	8	1-774	3406	94	3	243	55	2012-08-02 04:47:54.253317
3409	8	1-1023	3406	3	70	31	82	2012-08-02 04:47:54.253317
3410	8	1-515	3406	70	3	84	38	2012-08-02 04:47:54.253317
3411	8	1-733	3406	3	1	147	44	2012-08-02 04:47:54.253317
3412	8	1-926	3406	1	55	165	84	2012-08-02 04:47:54.253317
3413	8	1-795	3406	55	77	255	5	2012-08-02 04:47:54.253317
3406	8	1-1073	3406	77	75	164	8	2012-08-02 04:47:54.253317
3414	1	1-795	\N	55	55	958	5	2012-08-02 04:47:54.253317
3416	3	1-908	3415	67	4	781	30	2012-08-02 04:47:56.054253
3417	3	1-848	3415	4	18	252	86	2012-08-02 04:47:56.054253
3415	3	1-1077	3415	18	67	703	42	2012-08-02 04:47:56.054253
3419	7	1-952	3418	28	79	261	85	2012-08-02 04:47:57.25996
3420	7	1-970	3418	79	52	211	67	2012-08-02 04:47:57.25996
3421	7	1-73	3418	52	54	111	25	2012-08-02 04:47:57.25996
3422	7	1-1028	3418	54	39	1838	80	2012-08-02 04:47:57.25996
3423	7	1-782	3418	39	18	1654	88	2012-08-02 04:47:57.25996
3424	7	1-881	3418	18	98	1332	79	2012-08-02 04:47:57.25996
3418	7	1-1079	3418	98	28	2187	20	2012-08-02 04:47:57.25996
3426	2	1-753	3425	53	83	893	3	2012-08-02 04:47:58.298357
3425	2	1-1081	3425	83	53	1179	33	2012-08-02 04:47:58.298357
3428	5	1-1056	3427	73	4	550	94	2012-08-02 04:47:59.347978
3429	5	1-679	3427	4	69	609	87	2012-08-02 04:47:59.347978
3430	5	1-712	3427	69	9	595	66	2012-08-02 04:47:59.347978
3431	5	1-1015	3427	9	61	902	95	2012-08-02 04:47:59.347978
3427	5	1-1084	3427	61	73	466	12	2012-08-02 04:47:59.347978
15913	5	1-3946	15910	1	89	488	10	2012-08-02 05:13:16.598989
3433	7	1-79	3432	3	37	13	15	2012-08-02 04:48:00.298187
3434	7	1-1078	3432	37	73	6	63	2012-08-02 04:48:00.298187
3435	7	1-664	3432	73	91	5	50	2012-08-02 04:48:00.298187
3436	7	1-953	3432	91	73	1	5	2012-08-02 04:48:00.298187
3437	7	1-957	3432	73	73	53	68	2012-08-02 04:48:00.298187
3438	7	1-710	3432	73	22	80	17	2012-08-02 04:48:00.298187
3432	7	1-1086	3432	22	3	62	77	2012-08-02 04:48:00.298187
15914	5	1-3182	15910	89	49	166	49	2012-08-02 05:13:16.598989
3440	5	1-322	3439	22	96	366	26	2012-08-02 04:48:00.298187
3441	5	1-404	3439	96	73	40	5	2012-08-02 04:48:00.298187
3442	5	1-957	3439	73	73	1940	68	2012-08-02 04:48:00.298187
3443	5	1-710	3439	73	22	1969	17	2012-08-02 04:48:00.298187
3439	5	1-1086	3439	22	22	1533	77	2012-08-02 04:48:00.298187
15910	5	1-4587	15910	49	44	1902	63	2012-08-02 05:13:16.598989
3445	5	1-853	3444	59	41	9	13	2012-08-02 04:48:00.298187
3446	5	1-713	3444	41	72	35	65	2012-08-02 04:48:00.298187
3447	5	1-1062	3444	72	70	22	22	2012-08-02 04:48:00.298187
3448	5	1-927	3444	70	22	9	17	2012-08-02 04:48:00.298187
3444	5	1-1086	3444	22	59	11	77	2012-08-02 04:48:00.298187
3450	7	1-853	3449	59	71	814	13	2012-08-02 04:48:00.298187
3451	7	1-1048	3449	71	79	628	99	2012-08-02 04:48:00.298187
3452	7	1-1034	3449	79	34	288	49	2012-08-02 04:48:00.298187
3453	7	1-428	3449	34	37	355	41	2012-08-02 04:48:00.298187
3454	7	1-956	3449	37	81	459	92	2012-08-02 04:48:00.298187
3455	7	1-1042	3449	81	22	693	17	2012-08-02 04:48:00.298187
3449	7	1-1086	3449	22	59	955	77	2012-08-02 04:48:00.298187
3457	8	1-927	3456	70	22	879	17	2012-08-02 04:48:02.364285
3458	8	1-1086	3456	22	59	954	77	2012-08-02 04:48:02.364285
3459	8	1-853	3456	59	85	731	13	2012-08-02 04:48:02.364285
3460	8	1-852	3456	85	59	529	55	2012-08-02 04:48:02.364285
3461	8	1-893	3456	59	15	326	90	2012-08-02 04:48:02.364285
3462	8	1-1049	3456	15	44	247	69	2012-08-02 04:48:02.364285
3463	8	1-857	3456	44	57	186	42	2012-08-02 04:48:02.364285
3456	8	1-1090	3456	57	70	2790	22	2012-08-02 04:48:02.364285
3465	8	1-927	3464	70	22	1227	17	2012-08-02 04:48:02.364285
3466	8	1-1086	3464	22	61	1366	77	2012-08-02 04:48:02.364285
3467	8	1-504	3464	61	60	335	75	2012-08-02 04:48:02.364285
3468	8	1-920	3464	60	5	70	91	2012-08-02 04:48:02.364285
3469	8	1-221	3464	5	73	51	63	2012-08-02 04:48:02.364285
3470	8	1-539	3464	73	18	93	86	2012-08-02 04:48:02.364285
3471	8	1-1077	3464	18	57	247	42	2012-08-02 04:48:02.364285
3464	8	1-1090	3464	57	70	3794	22	2012-08-02 04:48:02.364285
3473	8	1-1012	3472	79	23	4478	79	2012-08-02 04:48:03.546719
3474	8	1-246	3472	23	22	727	17	2012-08-02 04:48:03.546719
3475	8	1-1086	3472	22	61	881	77	2012-08-02 04:48:03.546719
3476	8	1-504	3472	61	60	235	75	2012-08-02 04:48:03.546719
3477	8	1-920	3472	60	100	54	91	2012-08-02 04:48:03.546719
3478	8	1-638	3472	100	41	62	44	2012-08-02 04:48:03.546719
3479	8	1-331	3472	41	15	16	100	2012-08-02 04:48:03.546719
3472	8	1-1092	3472	15	79	2050	66	2012-08-02 04:48:03.546719
3481	7	1-282	3480	38	54	26	25	2012-08-02 04:48:04.441389
3482	7	1-1028	3480	54	39	453	80	2012-08-02 04:48:04.441389
3483	7	1-782	3480	39	69	427	88	2012-08-02 04:48:04.441389
3484	7	1-865	3480	69	46	1054	22	2012-08-02 04:48:04.441389
3485	7	1-1075	3480	46	18	92	28	2012-08-02 04:48:04.441389
3486	7	1-18	3480	18	16	84	3	2012-08-02 04:48:04.441389
3480	7	1-1095	3480	16	38	5	15	2012-08-02 04:48:04.441389
3488	3	1-493	3487	93	43	1528	99	2012-08-02 04:48:06.299242
3489	3	1-546	3487	43	19	723	3	2012-08-02 04:48:06.299242
3487	3	1-1097	3487	19	93	1859	64	2012-08-02 04:48:06.299242
3491	5	1-309	3490	9	57	28	42	2012-08-02 04:48:06.299242
3492	5	1-1090	3490	57	46	480	22	2012-08-02 04:48:06.299242
3493	5	1-1075	3490	46	18	39	28	2012-08-02 04:48:06.299242
3494	5	1-18	3490	18	19	34	3	2012-08-02 04:48:06.299242
3490	5	1-1097	3490	19	9	160	64	2012-08-02 04:48:06.299242
3496	7	1-1047	3495	11	19	588	3	2012-08-02 04:48:06.763128
3497	7	1-1097	3495	19	72	2813	64	2012-08-02 04:48:06.763128
3498	7	1-523	3495	72	72	2249	74	2012-08-02 04:48:06.763128
3499	7	1-836	3495	72	15	1541	90	2012-08-02 04:48:06.763128
3500	7	1-1049	3495	15	47	1363	69	2012-08-02 04:48:06.763128
3501	7	1-1093	3495	47	36	1368	94	2012-08-02 04:48:06.763128
3495	7	1-1098	3495	36	11	687	84	2012-08-02 04:48:06.763128
3503	7	1-303	3502	6	19	440	36	2012-08-02 04:48:08.155104
3504	7	1-723	3502	19	18	426	7	2012-08-02 04:48:08.155104
3505	7	1-533	3502	18	91	78	16	2012-08-02 04:48:08.155104
3506	7	1-1059	3502	91	62	148	29	2012-08-02 04:48:08.155104
3507	7	1-828	3502	62	95	149	19	2012-08-02 04:48:08.155104
3508	7	1-964	3502	95	30	1631	97	2012-08-02 04:48:08.155104
3502	7	1-1101	3502	30	6	453	1	2012-08-02 04:48:08.155104
3509	1	1-828	\N	62	62	1835	19	2012-08-02 04:48:08.155104
3511	7	1-622	3510	42	38	114	15	2012-08-02 04:48:08.641428
3512	7	1-282	3510	38	22	412	25	2012-08-02 04:48:08.641428
3513	7	1-1025	3510	22	23	359	66	2012-08-02 04:48:08.641428
3514	7	1-444	3510	23	88	326	59	2012-08-02 04:48:08.641428
3515	7	1-1032	3510	88	73	40	5	2012-08-02 04:48:08.641428
3516	7	1-957	3510	73	67	2304	68	2012-08-02 04:48:08.641428
3510	7	1-1102	3510	67	42	596	69	2012-08-02 04:48:08.641428
3518	7	1-1093	3517	47	4	837	94	2012-08-02 04:48:08.641428
3519	7	1-679	3517	4	69	702	87	2012-08-02 04:48:08.641428
3520	7	1-712	3517	69	23	520	66	2012-08-02 04:48:08.641428
3521	7	1-444	3517	23	88	499	59	2012-08-02 04:48:08.641428
3522	7	1-1032	3517	88	73	64	5	2012-08-02 04:48:08.641428
3523	7	1-957	3517	73	67	3928	68	2012-08-02 04:48:08.641428
3517	7	1-1102	3517	67	47	1073	69	2012-08-02 04:48:08.641428
3525	7	1-857	3524	44	20	817	42	2012-08-02 04:48:08.641428
3526	7	1-878	3524	20	17	857	21	2012-08-02 04:48:08.641428
3527	7	1-875	3524	17	43	780	99	2012-08-02 04:48:08.641428
3528	7	1-546	3524	43	19	753	3	2012-08-02 04:48:08.641428
3529	7	1-1097	3524	19	36	3954	64	2012-08-02 04:48:08.641428
3530	7	1-263	3524	36	67	2189	68	2012-08-02 04:48:08.641428
3524	7	1-1102	3524	67	44	844	69	2012-08-02 04:48:08.641428
3532	6	1-734	3531	81	23	192	46	2012-08-02 04:48:11.657079
3533	6	1-790	3531	23	34	410	76	2012-08-02 04:48:11.657079
3534	6	1-285	3531	34	19	181	3	2012-08-02 04:48:11.657079
3531	6	1-1107	3531	89	81	600	54	2012-08-02 04:48:11.657079
3535	6	1-1097	3531	19	72	431	64	2012-08-02 04:48:11.657079
3536	6	1-523	3531	72	89	163	74	2012-08-02 04:48:11.657079
3538	6	1-734	3537	81	98	734	46	2012-08-02 04:48:11.657079
3539	6	1-673	3537	98	6	201	1	2012-08-02 04:48:11.657079
3540	6	1-303	3537	6	66	121	36	2012-08-02 04:48:11.657079
3541	6	1-1071	3537	66	99	411	58	2012-08-02 04:48:11.657079
3542	6	1-1068	3537	99	89	333	74	2012-08-02 04:48:11.657079
3537	6	1-1107	3537	89	81	1677	54	2012-08-02 04:48:11.657079
15916	5	1-615	15915	5	43	110	41	2012-08-02 05:13:16.598989
3544	7	1-734	3543	81	98	3103	46	2012-08-02 04:48:11.657079
3545	7	1-673	3543	98	92	988	1	2012-08-02 04:48:11.657079
3546	7	1-872	3543	92	34	515	56	2012-08-02 04:48:11.657079
3547	7	1-1076	3543	34	66	278	36	2012-08-02 04:48:11.657079
3548	7	1-1071	3543	66	99	1099	58	2012-08-02 04:48:11.657079
3549	7	1-1068	3543	99	89	1038	74	2012-08-02 04:48:11.657079
3543	7	1-1107	3543	89	81	6088	54	2012-08-02 04:48:11.657079
3551	6	1-734	3550	81	8	68	46	2012-08-02 04:48:11.657079
3552	6	1-1026	3550	8	22	96	62	2012-08-02 04:48:11.657079
3553	6	1-849	3550	22	22	38	17	2012-08-02 04:48:11.657079
3554	6	1-1086	3550	22	86	50	77	2012-08-02 04:48:11.657079
3555	6	1-580	3550	86	89	15	74	2012-08-02 04:48:11.657079
3550	6	1-1107	3550	89	81	109	54	2012-08-02 04:48:11.657079
3557	5	1-998	3556	64	87	1025	36	2012-08-02 04:48:12.596482
3558	5	1-233	3556	87	4	1663	94	2012-08-02 04:48:12.596482
3559	5	1-679	3556	4	69	2001	87	2012-08-02 04:48:12.596482
3560	5	1-712	3556	69	99	2123	66	2012-08-02 04:48:12.596482
3556	5	1-1109	3556	99	64	989	32	2012-08-02 04:48:12.596482
3562	5	1-784	3561	67	5	1748	91	2012-08-02 04:48:14.794983
3563	5	1-221	3561	5	73	1169	63	2012-08-02 04:48:14.794983
3564	5	1-539	3561	73	16	1974	86	2012-08-02 04:48:14.794983
3565	5	1-1065	3561	16	29	4613	71	2012-08-02 04:48:14.794983
3561	5	1-1112	3561	29	67	9314	73	2012-08-02 04:48:14.794983
3567	5	1-282	3566	38	22	2143	25	2012-08-02 04:48:18.40696
3568	5	1-1025	3566	22	99	2230	66	2012-08-02 04:48:18.40696
3569	5	1-1109	3566	99	22	645	32	2012-08-02 04:48:18.40696
3570	5	1-1035	3566	22	54	1079	57	2012-08-02 04:48:18.40696
3566	5	1-1116	3566	54	38	632	15	2012-08-02 04:48:18.40696
3572	5	1-805	3571	32	59	1242	83	2012-08-02 04:48:18.40696
3573	5	1-912	3571	59	8	886	46	2012-08-02 04:48:18.40696
3574	5	1-1026	3571	8	13	1339	62	2012-08-02 04:48:18.40696
3575	5	1-1053	3571	13	54	2828	57	2012-08-02 04:48:18.40696
3571	5	1-1116	3571	54	32	2484	15	2012-08-02 04:48:18.40696
3577	6	1-976	3576	9	25	904	48	2012-08-02 04:48:19.35699
3578	6	1-692	3576	25	3	627	55	2012-08-02 04:48:19.35699
3579	6	1-1023	3576	3	3	75	82	2012-08-02 04:48:19.35699
3580	6	1-430	3576	3	89	92	35	2012-08-02 04:48:19.35699
3581	6	1-514	3576	89	39	606	47	2012-08-02 04:48:19.35699
3576	6	1-1118	3576	39	9	429	60	2012-08-02 04:48:19.35699
3583	7	1-1067	3582	19	69	1201	87	2012-08-02 04:48:20.251683
3584	7	1-712	3582	69	99	794	66	2012-08-02 04:48:20.251683
3585	7	1-1109	3582	99	64	230	32	2012-08-02 04:48:20.251683
3586	7	1-998	3582	64	66	149	36	2012-08-02 04:48:20.251683
3587	7	1-1071	3582	66	74	514	58	2012-08-02 04:48:20.251683
3588	7	1-327	3582	74	88	418	40	2012-08-02 04:48:20.251683
3582	7	1-1120	3582	88	19	4808	21	2012-08-02 04:48:20.251683
3589	1	1-712	\N	69	69	1834	66	2012-08-02 04:48:20.251683
3591	3	1-1048	3590	71	79	4586	99	2012-08-02 04:48:22.405934
3592	3	1-1034	3590	79	31	1819	49	2012-08-02 04:48:22.405934
3590	3	1-1124	3590	31	71	6863	13	2012-08-02 04:48:22.405934
3594	7	1-1020	3593	54	24	4406	10	2012-08-02 04:48:24.758874
3595	7	1-855	3593	24	47	855	69	2012-08-02 04:48:24.758874
3596	7	1-1093	3593	47	53	724	94	2012-08-02 04:48:24.758874
3597	7	1-942	3593	53	39	1392	47	2012-08-02 04:48:24.758874
3598	7	1-1118	3593	39	37	892	60	2012-08-02 04:48:24.758874
3599	7	1-988	3593	37	52	4388	72	2012-08-02 04:48:24.758874
3593	7	1-1128	3593	52	54	3426	97	2012-08-02 04:48:24.758874
3601	5	1-1101	3600	30	36	603	1	2012-08-02 04:48:24.758874
3602	5	1-603	3600	36	11	1003	27	2012-08-02 04:48:24.758874
3603	5	1-1046	3600	11	33	4375	81	2012-08-02 04:48:24.758874
3604	5	1-844	3600	33	52	3441	72	2012-08-02 04:48:24.758874
3600	5	1-1128	3600	52	30	2755	97	2012-08-02 04:48:24.758874
3606	4	1-347	3605	72	29	2598	31	2012-08-02 04:48:24.758874
3607	4	1-1099	3605	29	45	3042	37	2012-08-02 04:48:24.758874
3608	4	1-681	3605	45	52	2751	72	2012-08-02 04:48:24.758874
3605	4	1-1128	3605	52	72	2418	97	2012-08-02 04:48:24.758874
3610	5	1-1025	3609	22	99	3904	66	2012-08-02 04:48:26.614867
3611	5	1-1109	3609	99	64	1343	32	2012-08-02 04:48:26.614867
3612	5	1-998	3609	64	66	1028	36	2012-08-02 04:48:26.614867
3613	5	1-1071	3609	66	46	4214	58	2012-08-02 04:48:26.614867
3609	5	1-1132	3609	46	22	4008	25	2012-08-02 04:48:26.614867
3615	6	1-1076	3614	34	66	129	36	2012-08-02 04:48:28.782166
3616	6	1-1071	3614	66	46	298	58	2012-08-02 04:48:28.782166
3617	6	1-1132	3614	46	54	160	25	2012-08-02 04:48:28.782166
3618	6	1-1028	3614	54	39	1210	80	2012-08-02 04:48:28.782166
3619	6	1-782	3614	39	37	497	88	2012-08-02 04:48:28.782166
3614	6	1-1135	3614	37	34	407	56	2012-08-02 04:48:28.782166
3621	7	1-1076	3620	34	66	459	36	2012-08-02 04:48:28.782166
3622	7	1-1071	3620	66	13	1912	58	2012-08-02 04:48:28.782166
3623	7	1-944	3620	13	39	686	47	2012-08-02 04:48:28.782166
3624	7	1-1118	3620	39	9	436	60	2012-08-02 04:48:28.782166
3625	7	1-46	3620	9	97	508	31	2012-08-02 04:48:28.782166
3626	7	1-935	3620	97	37	546	88	2012-08-02 04:48:28.782166
3620	7	1-1135	3620	37	34	806	56	2012-08-02 04:48:28.782166
3628	6	1-685	3627	15	78	154	23	2012-08-02 04:48:28.782166
3629	6	1-419	3627	78	88	154	50	2012-08-02 04:48:28.782166
3630	6	1-730	3627	88	72	100	97	2012-08-02 04:48:28.782166
3631	6	1-347	3627	72	97	110	31	2012-08-02 04:48:28.782166
3632	6	1-935	3627	97	37	138	88	2012-08-02 04:48:28.782166
3627	6	1-1135	3627	37	15	236	56	2012-08-02 04:48:28.782166
3634	2	1-444	3633	23	54	5997	59	2012-08-02 04:48:29.245722
3633	2	1-1136	3633	54	23	5850	66	2012-08-02 04:48:29.245722
3636	4	1-1126	3635	92	70	1486	85	2012-08-02 04:48:29.688199
3637	4	1-134	3635	70	46	976	58	2012-08-02 04:48:29.688199
3638	4	1-1132	3635	46	33	723	25	2012-08-02 04:48:29.688199
3635	4	1-1137	3635	33	92	4144	18	2012-08-02 04:48:29.688199
3640	6	1-1126	3639	92	79	1971	85	2012-08-02 04:48:29.688199
3641	6	1-970	3639	79	82	1249	67	2012-08-02 04:48:29.688199
3642	6	1-758	3639	82	61	1486	98	2012-08-02 04:48:29.688199
3643	6	1-266	3639	61	46	671	58	2012-08-02 04:48:29.688199
3644	6	1-1132	3639	46	33	619	25	2012-08-02 04:48:29.688199
3639	6	1-1137	3639	33	92	4414	18	2012-08-02 04:48:29.688199
15917	5	1-4212	15915	43	9	58	34	2012-08-02 05:13:16.598989
3646	2	1-69	3645	50	41	4860	74	2012-08-02 04:48:30.610702
3645	2	1-1139	3645	41	50	8728	75	2012-08-02 04:48:30.610702
15918	5	1-4435	15915	9	21	104	21	2012-08-02 05:13:16.598989
3648	6	1-502	3647	25	53	224	61	2012-08-02 04:48:30.610702
3649	6	1-1138	3647	53	64	199	89	2012-08-02 04:48:30.610702
3650	6	1-330	3647	64	20	173	42	2012-08-02 04:48:30.610702
3651	6	1-878	3647	20	42	150	21	2012-08-02 04:48:30.610702
3652	6	1-996	3647	42	41	75	74	2012-08-02 04:48:30.610702
3647	6	1-1139	3647	41	25	197	75	2012-08-02 04:48:30.610702
15919	5	1-3051	15915	21	49	45	49	2012-08-02 05:13:16.598989
3654	3	1-785	3653	30	63	8942	71	2012-08-02 04:48:33.135428
3655	3	1-377	3653	63	61	7833	43	2012-08-02 04:48:33.135428
3653	3	1-1146	3653	61	30	6524	70	2012-08-02 04:48:33.135428
15915	5	1-4587	15915	49	5	540	63	2012-08-02 05:13:16.598989
3657	5	1-1020	3656	54	24	5438	10	2012-08-02 04:48:34.516382
3658	5	1-855	3656	24	44	690	69	2012-08-02 04:48:34.516382
3659	5	1-857	3656	44	42	336	42	2012-08-02 04:48:34.516382
3660	5	1-1143	3656	42	17	246	80	2012-08-02 04:48:34.516382
3656	5	1-1150	3656	17	54	6464	97	2012-08-02 04:48:34.516382
3662	4	1-1101	3661	30	68	295	1	2012-08-02 04:48:34.516382
3663	4	1-1115	3661	68	42	93	42	2012-08-02 04:48:34.516382
3664	4	1-1143	3661	42	17	72	80	2012-08-02 04:48:34.516382
3661	4	1-1150	3661	17	30	1999	97	2012-08-02 04:48:34.516382
3666	7	1-347	3665	72	45	608	31	2012-08-02 04:48:34.516382
3667	7	1-561	3665	45	37	29	15	2012-08-02 04:48:34.516382
3668	7	1-1078	3665	37	73	14	63	2012-08-02 04:48:34.516382
3669	7	1-539	3665	73	18	18	86	2012-08-02 04:48:34.516382
3670	7	1-1077	3665	18	42	34	42	2012-08-02 04:48:34.516382
3671	7	1-1143	3665	42	17	28	80	2012-08-02 04:48:34.516382
3665	7	1-1150	3665	17	72	852	97	2012-08-02 04:48:34.516382
3673	7	1-765	3672	100	66	146	4	2012-08-02 04:48:35.444308
3674	7	1-1085	3672	66	61	326	14	2012-08-02 04:48:35.444308
3675	7	1-484	3672	61	88	143	59	2012-08-02 04:48:35.444308
3676	7	1-1032	3672	88	77	25	5	2012-08-02 04:48:35.444308
3677	7	1-1073	3672	77	63	16	8	2012-08-02 04:48:35.444308
3678	7	1-690	3672	63	80	22	92	2012-08-02 04:48:35.444308
3672	7	1-1152	3672	80	100	39	50	2012-08-02 04:48:35.444308
3680	7	1-750	3679	49	59	1394	77	2012-08-02 04:48:37.477428
3681	7	1-853	3679	59	85	979	13	2012-08-02 04:48:37.477428
3682	7	1-852	3679	85	3	743	55	2012-08-02 04:48:37.477428
3683	7	1-1023	3679	3	70	78	82	2012-08-02 04:48:37.477428
3684	7	1-515	3679	70	27	174	38	2012-08-02 04:48:37.477428
3685	7	1-746	3679	27	65	2351	37	2012-08-02 04:48:37.477428
3679	7	1-1157	3679	65	49	1575	44	2012-08-02 04:48:37.477428
3687	7	1-926	3686	1	50	2822	84	2012-08-02 04:48:37.477428
3688	7	1-683	3686	50	15	2546	90	2012-08-02 04:48:37.477428
3689	7	1-1049	3686	15	29	2278	69	2012-08-02 04:48:37.477428
3690	7	1-222	3686	29	35	3195	95	2012-08-02 04:48:37.477428
3691	7	1-620	3686	35	29	2716	31	2012-08-02 04:48:37.477428
3692	7	1-1099	3686	29	65	3560	37	2012-08-02 04:48:37.477428
3686	7	1-1157	3686	65	1	2691	44	2012-08-02 04:48:37.477428
3694	6	1-331	3693	41	15	3	100	2012-08-02 04:48:37.477428
3695	6	1-1092	3693	15	23	363	66	2012-08-02 04:48:37.477428
3696	6	1-444	3693	23	21	450	59	2012-08-02 04:48:37.477428
3697	6	1-900	3693	21	92	18	28	2012-08-02 04:48:37.477428
3698	6	1-1130	3693	92	65	13	37	2012-08-02 04:48:37.477428
3693	6	1-1157	3693	65	41	10	44	2012-08-02 04:48:37.477428
3700	5	1-775	3699	76	75	854	53	2012-08-02 04:48:37.477428
3701	5	1-216	3699	75	72	1034	97	2012-08-02 04:48:37.477428
3702	5	1-347	3699	72	29	1276	31	2012-08-02 04:48:37.477428
3703	5	1-1099	3699	29	65	1716	37	2012-08-02 04:48:37.477428
3699	5	1-1157	3699	65	76	1398	44	2012-08-02 04:48:37.477428
3705	7	1-750	3704	49	59	396	77	2012-08-02 04:48:37.477428
3706	7	1-853	3704	59	71	344	13	2012-08-02 04:48:37.477428
3707	7	1-1048	3704	71	14	271	99	2012-08-02 04:48:37.477428
3708	7	1-770	3704	14	81	118	16	2012-08-02 04:48:37.477428
3709	7	1-830	3704	81	29	318	31	2012-08-02 04:48:37.477428
3710	7	1-1099	3704	29	65	435	37	2012-08-02 04:48:37.477428
3704	7	1-1157	3704	65	49	361	44	2012-08-02 04:48:37.477428
3711	1	1-515	\N	70	70	3711	38	2012-08-02 04:48:37.477428
3713	7	1-1109	3712	99	96	641	32	2012-08-02 04:48:38.880866
3714	7	1-907	3712	96	34	1584	76	2012-08-02 04:48:38.880866
3715	7	1-285	3712	34	83	1266	3	2012-08-02 04:48:38.880866
3716	7	1-1081	3712	83	96	2034	33	2012-08-02 04:48:38.880866
3717	7	1-827	3712	96	28	1500	6	2012-08-02 04:48:38.880866
3718	7	1-1147	3712	28	57	318	24	2012-08-02 04:48:38.880866
3712	7	1-1160	3712	57	99	1693	66	2012-08-02 04:48:38.880866
3720	6	1-330	3719	64	20	1633	42	2012-08-02 04:48:41.653454
3721	6	1-878	3719	20	42	1400	21	2012-08-02 04:48:41.653454
3722	6	1-996	3719	42	29	690	74	2012-08-02 04:48:41.653454
3723	6	1-1103	3719	29	65	1029	37	2012-08-02 04:48:41.653454
3724	6	1-1157	3719	65	3	689	44	2012-08-02 04:48:41.653454
3719	6	1-1167	3719	3	64	1895	89	2012-08-02 04:48:41.653454
3726	6	1-891	3725	24	61	173	98	2012-08-02 04:48:41.653454
3727	6	1-266	3725	61	99	89	58	2012-08-02 04:48:41.653454
3728	6	1-1068	3725	99	29	96	74	2012-08-02 04:48:41.653454
3729	6	1-1103	3725	29	65	145	37	2012-08-02 04:48:41.653454
3730	6	1-1157	3725	65	3	98	44	2012-08-02 04:48:41.653454
3725	6	1-1167	3725	3	24	274	89	2012-08-02 04:48:41.653454
3732	6	1-1011	3731	96	22	453	17	2012-08-02 04:48:41.653454
3733	6	1-1086	3731	22	86	548	77	2012-08-02 04:48:41.653454
3734	6	1-580	3731	86	29	165	74	2012-08-02 04:48:41.653454
3735	6	1-1103	3731	29	65	262	37	2012-08-02 04:48:41.653454
3736	6	1-1157	3731	65	3	187	44	2012-08-02 04:48:41.653454
3731	6	1-1167	3731	3	96	547	89	2012-08-02 04:48:41.653454
3738	7	1-891	3737	24	100	864	98	2012-08-02 04:48:41.653454
3739	7	1-934	3737	100	18	1344	7	2012-08-02 04:48:41.653454
3740	7	1-533	3737	18	81	217	16	2012-08-02 04:48:41.653454
3741	7	1-830	3737	81	29	507	31	2012-08-02 04:48:41.653454
3742	7	1-1099	3737	29	65	603	37	2012-08-02 04:48:41.653454
3743	7	1-1157	3737	65	3	435	44	2012-08-02 04:48:41.653454
3737	7	1-1167	3737	3	24	1288	89	2012-08-02 04:48:41.653454
3745	6	1-1167	3744	3	70	5233	89	2012-08-02 04:48:43.123105
3746	6	1-873	3744	70	2	4028	27	2012-08-02 04:48:43.123105
3747	6	1-189	3744	2	39	1754	47	2012-08-02 04:48:43.123105
3748	6	1-1118	3744	39	9	1181	60	2012-08-02 04:48:43.123105
3749	6	1-976	3744	9	34	2367	48	2012-08-02 04:48:43.123105
3744	6	1-1170	3744	34	3	1924	44	2012-08-02 04:48:43.123105
15921	5	1-4288	15920	20	59	599	67	2012-08-02 05:13:16.598989
3751	6	1-750	3750	49	3	2652	77	2012-08-02 04:48:43.123105
3752	6	1-1168	3750	3	16	1753	52	2012-08-02 04:48:43.123105
3753	6	1-449	3750	16	39	1239	47	2012-08-02 04:48:43.123105
3754	6	1-1118	3750	39	9	1020	60	2012-08-02 04:48:43.123105
3755	6	1-976	3750	9	34	2496	48	2012-08-02 04:48:43.123105
3750	6	1-1170	3750	34	49	2479	44	2012-08-02 04:48:43.123105
3757	7	1-1163	3756	8	11	2333	84	2012-08-02 04:48:43.598762
3758	7	1-1047	3756	11	16	1289	3	2012-08-02 04:48:43.598762
3759	7	1-1095	3756	16	37	46	15	2012-08-02 04:48:43.598762
3760	7	1-1078	3756	37	73	22	63	2012-08-02 04:48:43.598762
3761	7	1-539	3756	73	82	30	86	2012-08-02 04:48:43.598762
3762	7	1-955	3756	82	26	120	65	2012-08-02 04:48:43.598762
3756	7	1-1171	3756	26	8	140	90	2012-08-02 04:48:43.598762
3764	5	1-1049	3763	15	29	255	69	2012-08-02 04:48:43.598762
3765	5	1-222	3763	29	79	238	95	2012-08-02 04:48:43.598762
3766	5	1-1127	3763	79	82	72	86	2012-08-02 04:48:43.598762
3767	5	1-955	3763	82	26	318	65	2012-08-02 04:48:43.598762
3763	5	1-1171	3763	26	15	408	90	2012-08-02 04:48:43.598762
3769	6	1-539	3768	73	24	1341	86	2012-08-02 04:48:44.07351
3770	6	1-1007	3768	24	9	788	60	2012-08-02 04:48:44.07351
3771	6	1-976	3768	9	34	1388	48	2012-08-02 04:48:44.07351
3772	6	1-1170	3768	34	1	992	44	2012-08-02 04:48:44.07351
3773	6	1-926	3768	1	41	802	84	2012-08-02 04:48:44.07351
3768	6	1-1172	3768	41	73	826	63	2012-08-02 04:48:44.07351
3775	7	1-380	3774	44	16	2573	52	2012-08-02 04:48:44.07351
3776	7	1-449	3774	16	39	1511	47	2012-08-02 04:48:44.07351
3777	7	1-1118	3774	39	9	1034	60	2012-08-02 04:48:44.07351
3778	7	1-976	3774	9	34	2103	48	2012-08-02 04:48:44.07351
3779	7	1-1170	3774	34	1	1736	44	2012-08-02 04:48:44.07351
3780	7	1-926	3774	1	41	1620	84	2012-08-02 04:48:44.07351
3774	7	1-1172	3774	41	44	1927	63	2012-08-02 04:48:44.07351
3782	7	1-937	3781	2	49	2360	28	2012-08-02 04:48:44.07351
3783	7	1-72	3781	49	6	823	8	2012-08-02 04:48:44.07351
3784	7	1-694	3781	6	47	1781	45	2012-08-02 04:48:44.07351
3785	7	1-1105	3781	47	34	1213	48	2012-08-02 04:48:44.07351
3786	7	1-1170	3781	34	1	1224	44	2012-08-02 04:48:44.07351
3787	7	1-926	3781	1	41	1397	84	2012-08-02 04:48:44.07351
3781	7	1-1172	3781	41	2	2031	63	2012-08-02 04:48:44.07351
3789	6	1-961	3788	29	66	3861	4	2012-08-02 04:48:44.549149
3790	6	1-1085	3788	66	84	6813	14	2012-08-02 04:48:44.549149
3791	6	1-946	3788	84	40	1774	93	2012-08-02 04:48:44.549149
3792	6	1-993	3788	40	42	259	42	2012-08-02 04:48:44.549149
3793	6	1-1143	3788	42	53	290	80	2012-08-02 04:48:44.549149
3788	6	1-1173	3788	53	29	552	38	2012-08-02 04:48:44.549149
3795	6	1-1096	3794	22	59	1895	8	2012-08-02 04:48:44.549149
3796	6	1-1051	3794	59	23	1349	12	2012-08-02 04:48:44.549149
3797	6	1-888	3794	23	39	1826	47	2012-08-02 04:48:44.549149
3798	6	1-1118	3794	39	9	1220	60	2012-08-02 04:48:44.549149
3799	6	1-591	3794	9	53	1169	80	2012-08-02 04:48:44.549149
3794	6	1-1173	3794	53	22	2316	38	2012-08-02 04:48:44.549149
3801	7	1-733	3800	3	1	1920	44	2012-08-02 04:48:44.549149
3802	7	1-926	3800	1	41	1793	84	2012-08-02 04:48:44.549149
3803	7	1-1172	3800	41	60	2133	63	2012-08-02 04:48:44.549149
3804	7	1-838	3800	60	44	692	69	2012-08-02 04:48:44.549149
3805	7	1-857	3800	44	42	550	42	2012-08-02 04:48:44.549149
3806	7	1-1143	3800	42	53	657	80	2012-08-02 04:48:44.549149
3800	7	1-1173	3800	53	3	1333	38	2012-08-02 04:48:44.549149
3808	3	1-1096	3807	22	78	3709	8	2012-08-02 04:48:44.549149
3809	3	1-1131	3807	78	53	2122	80	2012-08-02 04:48:44.549149
3807	3	1-1173	3807	53	22	4365	38	2012-08-02 04:48:44.549149
3811	7	1-1117	3810	66	59	652	58	2012-08-02 04:48:44.549149
3812	7	1-1162	3810	59	88	1236	19	2012-08-02 04:48:44.549149
3813	7	1-787	3810	88	17	648	26	2012-08-02 04:48:44.549149
3814	7	1-211	3810	17	18	154	86	2012-08-02 04:48:44.549149
3815	7	1-1077	3810	18	42	426	42	2012-08-02 04:48:44.549149
3816	7	1-1143	3810	42	53	516	80	2012-08-02 04:48:44.549149
3810	7	1-1173	3810	53	66	1064	38	2012-08-02 04:48:44.549149
3818	7	1-889	3817	11	78	50	64	2012-08-02 04:48:45.234584
3819	7	1-979	3817	78	26	66	65	2012-08-02 04:48:45.234584
3820	7	1-1171	3817	26	15	126	90	2012-08-02 04:48:45.234584
3821	7	1-1049	3817	15	29	117	69	2012-08-02 04:48:45.234584
3822	7	1-222	3817	29	32	163	95	2012-08-02 04:48:45.234584
3823	7	1-940	3817	32	81	99	41	2012-08-02 04:48:45.234584
3817	7	1-1175	3817	81	11	187	44	2012-08-02 04:48:45.234584
3825	5	1-765	3824	100	66	746	4	2012-08-02 04:48:46.537892
3826	5	1-1085	3824	66	51	1519	14	2012-08-02 04:48:46.537892
3827	5	1-1013	3824	51	66	310	3	2012-08-02 04:48:46.537892
3828	5	1-630	3824	66	4	208	34	2012-08-02 04:48:46.537892
3824	5	1-1177	3824	4	100	254	50	2012-08-02 04:48:46.537892
3830	6	1-1166	3829	8	63	185	46	2012-08-02 04:48:46.990717
3831	6	1-748	3829	63	34	174	48	2012-08-02 04:48:46.990717
3832	6	1-1170	3829	34	41	123	44	2012-08-02 04:48:46.990717
3833	6	1-331	3829	41	15	26	100	2012-08-02 04:48:46.990717
3834	6	1-1092	3829	15	21	2620	66	2012-08-02 04:48:46.990717
3829	6	1-1178	3829	21	8	1720	2	2012-08-02 04:48:46.990717
3836	6	1-1166	3835	8	23	154	46	2012-08-02 04:48:46.990717
3837	6	1-790	3835	23	51	694	76	2012-08-02 04:48:46.990717
3838	6	1-1052	3835	51	28	760	6	2012-08-02 04:48:46.990717
3839	6	1-1147	3835	28	57	188	24	2012-08-02 04:48:46.990717
3840	6	1-1160	3835	57	21	1164	66	2012-08-02 04:48:46.990717
3835	6	1-1178	3835	21	8	1045	2	2012-08-02 04:48:46.990717
3842	3	1-902	3841	11	13	641	5	2012-08-02 04:48:48.382795
3843	3	1-1072	3841	13	60	2578	72	2012-08-02 04:48:48.382795
3841	3	1-1181	3841	60	11	4082	73	2012-08-02 04:48:48.382795
3845	3	1-296	3844	45	47	1000	53	2012-08-02 04:48:48.382795
3846	3	1-856	3844	47	60	549	72	2012-08-02 04:48:48.382795
3844	3	1-1181	3844	60	45	1069	73	2012-08-02 04:48:48.382795
15922	5	1-4324	15920	59	45	223	68	2012-08-02 05:13:16.598989
3848	5	1-162	3847	83	51	1065	14	2012-08-02 04:48:49.885796
3849	5	1-1013	3847	51	83	249	3	2012-08-02 04:48:49.885796
3850	5	1-1081	3847	83	29	460	33	2012-08-02 04:48:49.885796
3851	5	1-1004	3847	29	5	161	81	2012-08-02 04:48:49.885796
3847	5	1-1183	3847	5	83	2145	20	2012-08-02 04:48:49.885796
3853	6	1-943	3852	43	50	5161	93	2012-08-02 04:48:51.780659
3854	6	1-203	3852	50	17	1101	21	2012-08-02 04:48:51.780659
3855	6	1-875	3852	17	43	973	99	2012-08-02 04:48:51.780659
3856	6	1-546	3852	43	83	934	3	2012-08-02 04:48:51.780659
3857	6	1-1081	3852	83	50	1765	33	2012-08-02 04:48:51.780659
3852	6	1-1187	3852	50	43	4565	7	2012-08-02 04:48:51.780659
3859	7	1-832	3858	85	88	778	59	2012-08-02 04:48:52.245437
3860	7	1-1032	3858	88	36	138	5	2012-08-02 04:48:52.245437
3861	7	1-1140	3858	36	53	94	94	2012-08-02 04:48:52.245437
3862	7	1-942	3858	53	34	228	47	2012-08-02 04:48:52.245437
3863	7	1-740	3858	34	4	249	34	2012-08-02 04:48:52.245437
3864	7	1-1177	3858	4	46	382	50	2012-08-02 04:48:52.245437
3858	7	1-1188	3858	46	85	1770	88	2012-08-02 04:48:52.245437
3866	7	1-832	3865	85	88	714	59	2012-08-02 04:48:52.245437
3867	7	1-1032	3865	88	77	130	5	2012-08-02 04:48:52.245437
3868	7	1-1073	3865	77	6	84	8	2012-08-02 04:48:52.245437
3869	7	1-694	3865	6	87	181	45	2012-08-02 04:48:52.245437
3870	7	1-1149	3865	87	80	182	92	2012-08-02 04:48:52.245437
3871	7	1-1152	3865	80	46	335	50	2012-08-02 04:48:52.245437
3865	7	1-1188	3865	46	85	1587	88	2012-08-02 04:48:52.245437
3873	2	1-491	3872	72	19	1307	80	2012-08-02 04:48:53.174764
3872	2	1-1190	3872	19	72	4513	30	2012-08-02 04:48:53.174764
3875	4	1-1188	3874	46	18	870	88	2012-08-02 04:48:55.472532
3876	4	1-881	3874	18	98	732	79	2012-08-02 04:48:55.472532
3877	4	1-1079	3874	98	2	1256	20	2012-08-02 04:48:55.472532
3874	4	1-1193	3874	2	46	189	50	2012-08-02 04:48:55.472532
3879	6	1-546	3878	43	83	299	3	2012-08-02 04:48:56.42285
3880	6	1-1081	3878	83	50	536	33	2012-08-02 04:48:56.42285
3881	6	1-1187	3878	50	70	1283	7	2012-08-02 04:48:56.42285
3882	6	1-842	3878	70	75	816	53	2012-08-02 04:48:56.42285
3883	6	1-216	3878	75	74	915	97	2012-08-02 04:48:56.42285
3878	6	1-1195	3878	74	43	337	99	2012-08-02 04:48:56.42285
3885	3	1-114	3884	32	69	737	9	2012-08-02 04:48:58.256865
3886	3	1-1029	3884	69	35	202	25	2012-08-02 04:48:58.256865
3884	3	1-1197	3884	35	32	267	89	2012-08-02 04:48:58.256865
3888	7	1-428	3887	34	81	2655	41	2012-08-02 04:48:59.923148
3889	7	1-1175	3887	81	21	4189	44	2012-08-02 04:48:59.923148
3890	7	1-1017	3887	21	88	850	26	2012-08-02 04:48:59.923148
3891	7	1-1194	3887	88	16	3942	52	2012-08-02 04:48:59.923148
3892	7	1-449	3887	16	30	2265	47	2012-08-02 04:48:59.923148
3893	7	1-816	3887	30	9	1566	14	2012-08-02 04:48:59.923148
3887	7	1-1199	3887	9	34	2657	49	2012-08-02 04:48:59.923148
3895	5	1-1176	3894	75	29	1215	74	2012-08-02 04:48:59.923148
3896	5	1-1103	3894	29	84	1980	37	2012-08-02 04:48:59.923148
3897	5	1-1159	3894	84	39	2852	22	2012-08-02 04:48:59.923148
3898	5	1-909	3894	39	9	511	14	2012-08-02 04:48:59.923148
3894	5	1-1199	3894	9	75	964	49	2012-08-02 04:48:59.923148
3900	5	1-800	3899	47	60	2242	87	2012-08-02 04:48:59.923148
3901	5	1-667	3899	60	41	4547	78	2012-08-02 04:48:59.923148
3902	5	1-903	3899	41	30	646	47	2012-08-02 04:48:59.923148
3903	5	1-816	3899	30	9	525	14	2012-08-02 04:48:59.923148
3899	5	1-1199	3899	9	47	1044	49	2012-08-02 04:48:59.923148
3905	6	1-97	3904	59	24	139	26	2012-08-02 04:48:59.923148
3906	6	1-971	3904	24	57	213	11	2012-08-02 04:48:59.923148
3907	6	1-91	3904	57	41	233	84	2012-08-02 04:48:59.923148
3908	6	1-1172	3904	41	49	329	63	2012-08-02 04:48:59.923148
3909	6	1-357	3904	49	9	265	14	2012-08-02 04:48:59.923148
3904	6	1-1199	3904	9	59	545	49	2012-08-02 04:48:59.923148
3911	3	1-428	3910	34	59	432	41	2012-08-02 04:48:59.923148
3912	3	1-1063	3910	59	9	172	14	2012-08-02 04:48:59.923148
3910	3	1-1199	3910	9	34	355	49	2012-08-02 04:48:59.923148
3914	4	1-1170	3913	34	25	326	44	2012-08-02 04:49:00.398868
3915	4	1-389	3913	25	28	452	6	2012-08-02 04:49:00.398868
3916	4	1-1147	3913	28	89	118	24	2012-08-02 04:49:00.398868
3913	4	1-1200	3913	89	34	319	48	2012-08-02 04:49:00.398868
3918	3	1-912	3917	59	23	2598	46	2012-08-02 04:49:01.326919
3919	3	1-790	3917	23	29	6625	76	2012-08-02 04:49:01.326919
3917	3	1-1202	3917	29	59	6336	83	2012-08-02 04:49:01.326919
3921	5	1-1156	3920	27	59	1074	58	2012-08-02 04:49:01.326919
3922	5	1-1162	3920	59	12	2330	19	2012-08-02 04:49:01.326919
3923	5	1-1083	3920	12	21	982	60	2012-08-02 04:49:01.326919
3924	5	1-621	3920	21	29	1266	76	2012-08-02 04:49:01.326919
3920	5	1-1202	3920	29	27	2124	83	2012-08-02 04:49:01.326919
3926	6	1-888	3925	23	30	635	47	2012-08-02 04:49:02.222038
3927	6	1-816	3925	30	9	400	14	2012-08-02 04:49:02.222038
3928	6	1-1199	3925	9	59	619	49	2012-08-02 04:49:02.222038
3929	6	1-97	3925	59	88	119	26	2012-08-02 04:49:02.222038
3930	6	1-1194	3925	88	37	502	52	2012-08-02 04:49:02.222038
3925	6	1-1204	3925	37	23	513	12	2012-08-02 04:49:02.222038
3932	7	1-743	3931	36	9	1795	14	2012-08-02 04:49:03.625221
3933	7	1-1199	3931	9	59	3368	49	2012-08-02 04:49:03.625221
3934	7	1-97	3931	59	24	783	26	2012-08-02 04:49:03.625221
3935	7	1-971	3931	24	37	1097	11	2012-08-02 04:49:03.625221
3936	7	1-516	3931	37	25	314	46	2012-08-02 04:49:03.625221
3937	7	1-778	3931	25	35	342	45	2012-08-02 04:49:03.625221
3931	7	1-1207	3931	35	36	208	91	2012-08-02 04:49:03.625221
3939	5	1-595	3938	27	59	384	58	2012-08-02 04:49:03.625221
3940	5	1-1162	3938	59	17	846	19	2012-08-02 04:49:03.625221
3941	5	1-992	3938	17	59	1548	39	2012-08-02 04:49:03.625221
3942	5	1-950	3938	59	35	1169	45	2012-08-02 04:49:03.625221
3938	5	1-1207	3938	35	27	735	91	2012-08-02 04:49:03.625221
3944	6	1-96	3943	13	59	1543	39	2012-08-02 04:49:05.010962
3945	6	1-950	3943	59	35	1209	45	2012-08-02 04:49:05.010962
3946	6	1-1207	3943	35	37	830	91	2012-08-02 04:49:05.010962
3947	6	1-543	3943	37	3	484	82	2012-08-02 04:49:05.010962
3948	6	1-430	3943	3	90	647	35	2012-08-02 04:49:05.010962
3943	6	1-1210	3943	90	13	1369	85	2012-08-02 04:49:05.010962
3949	7	1-1211	3949	61	78	321	64	2012-08-02 04:49:05.443119
3950	7	1-979	3949	78	26	423	65	2012-08-02 04:49:05.443119
3951	7	1-1171	3949	26	15	810	90	2012-08-02 04:49:05.443119
3952	7	1-1049	3949	15	47	756	69	2012-08-02 04:49:05.443119
3953	7	1-1093	3949	47	36	802	94	2012-08-02 04:49:05.443119
3954	7	1-1098	3949	36	11	425	84	2012-08-02 04:49:05.443119
3955	7	1-1047	3949	11	61	385	3	2012-08-02 04:49:05.443119
15923	5	1-3842	15920	45	91	140	65	2012-08-02 05:13:16.598989
3957	7	1-368	3956	88	92	23	18	2012-08-02 04:49:06.371081
3958	7	1-1126	3956	92	79	12	85	2012-08-02 04:49:06.371081
3959	7	1-970	3956	79	55	8	67	2012-08-02 04:49:06.371081
3960	7	1-831	3956	55	83	27	71	2012-08-02 04:49:06.371081
3961	7	1-884	3956	83	57	13	24	2012-08-02 04:49:06.371081
3962	7	1-1160	3956	57	53	66	66	2012-08-02 04:49:06.371081
3956	7	1-1213	3956	53	88	101	53	2012-08-02 04:49:06.371081
15924	5	1-4433	15920	91	49	81	49	2012-08-02 05:13:16.598989
3964	5	1-1088	3963	97	71	818	70	2012-08-02 04:49:06.813006
3965	5	1-127	3963	71	5	993	81	2012-08-02 04:49:06.813006
3966	5	1-1183	3963	5	7	6831	20	2012-08-02 04:49:06.813006
3967	5	1-598	3963	7	61	1839	4	2012-08-02 04:49:06.813006
3963	5	1-1214	3963	61	97	1649	18	2012-08-02 04:49:06.813006
3969	6	1-1088	3968	97	71	339	70	2012-08-02 04:49:06.813006
3970	6	1-127	3968	71	55	457	81	2012-08-02 04:49:06.813006
3971	6	1-544	3968	55	4	401	34	2012-08-02 04:49:06.813006
3972	6	1-1177	3968	4	100	321	50	2012-08-02 04:49:06.813006
3973	6	1-765	3968	100	61	618	4	2012-08-02 04:49:06.813006
3968	6	1-1214	3968	61	97	616	18	2012-08-02 04:49:06.813006
3975	6	1-899	3974	15	2	2316	27	2012-08-02 04:49:06.813006
3976	6	1-189	3974	2	34	754	47	2012-08-02 04:49:06.813006
3977	6	1-740	3974	34	4	512	34	2012-08-02 04:49:06.813006
3978	6	1-1177	3974	4	100	488	50	2012-08-02 04:49:06.813006
3979	6	1-765	3974	100	61	1115	4	2012-08-02 04:49:06.813006
3974	6	1-1214	3974	61	15	1320	18	2012-08-02 04:49:06.813006
3981	6	1-1088	3980	97	94	627	70	2012-08-02 04:49:06.813006
3982	6	1-549	3980	94	55	270	81	2012-08-02 04:49:06.813006
3983	6	1-544	3980	55	4	297	34	2012-08-02 04:49:06.813006
3984	6	1-1177	3980	4	100	299	50	2012-08-02 04:49:06.813006
3985	6	1-765	3980	100	61	724	4	2012-08-02 04:49:06.813006
3980	6	1-1214	3980	61	97	906	18	2012-08-02 04:49:06.813006
3987	7	1-1210	3986	90	79	471	85	2012-08-02 04:49:07.277447
3988	7	1-970	3986	79	30	413	67	2012-08-02 04:49:07.277447
3989	7	1-1121	3986	30	75	1021	49	2012-08-02 04:49:07.277447
3990	7	1-1176	3986	75	29	1447	74	2012-08-02 04:49:07.277447
3991	7	1-1103	3986	29	84	2651	37	2012-08-02 04:49:07.277447
3992	7	1-1159	3986	84	8	4293	22	2012-08-02 04:49:07.277447
3986	7	1-1215	3986	8	90	223	35	2012-08-02 04:49:07.277447
3994	5	1-694	3993	6	35	6697	45	2012-08-02 04:49:08.16101
3995	5	1-1207	3993	35	63	3095	91	2012-08-02 04:49:08.16101
3996	5	1-915	3993	63	13	3075	85	2012-08-02 04:49:08.16101
3997	5	1-96	3993	13	80	2333	39	2012-08-02 04:49:08.16101
3993	5	1-1217	3993	80	6	4587	8	2012-08-02 04:49:08.16101
3999	5	1-691	3998	82	35	2256	45	2012-08-02 04:49:08.16101
4000	5	1-1207	3998	35	63	1239	91	2012-08-02 04:49:08.16101
4001	5	1-915	3998	63	13	1463	85	2012-08-02 04:49:08.16101
4002	5	1-96	3998	13	80	1319	39	2012-08-02 04:49:08.16101
3998	5	1-1217	3998	80	82	3082	8	2012-08-02 04:49:08.16101
4004	7	1-1207	4003	35	63	322	91	2012-08-02 04:49:08.625194
4005	7	1-915	4003	63	79	347	85	2012-08-02 04:49:08.625194
4006	7	1-970	4003	79	55	222	67	2012-08-02 04:49:08.625194
4007	7	1-831	4003	55	23	662	71	2012-08-02 04:49:08.625194
4008	7	1-913	4003	23	98	371	79	2012-08-02 04:49:08.625194
4009	7	1-1079	4003	98	5	483	20	2012-08-02 04:49:08.625194
4003	7	1-1218	4003	5	35	641	45	2012-08-02 04:49:08.625194
4011	7	1-1149	4010	87	80	2541	92	2012-08-02 04:49:08.625194
4012	7	1-1152	4010	80	88	4669	50	2012-08-02 04:49:08.625194
4013	7	1-730	4010	88	74	3414	97	2012-08-02 04:49:08.625194
4014	7	1-1195	4010	74	8	1359	99	2012-08-02 04:49:08.625194
4015	7	1-397	4010	8	84	869	47	2012-08-02 04:49:08.625194
4016	7	1-1153	4010	84	5	1407	20	2012-08-02 04:49:08.625194
4010	7	1-1218	4010	5	87	2530	45	2012-08-02 04:49:08.625194
4018	6	1-356	4017	46	41	2447	78	2012-08-02 04:49:09.564436
4019	6	1-903	4017	41	84	366	47	2012-08-02 04:49:09.564436
4020	6	1-1153	4017	84	5	590	20	2012-08-02 04:49:09.564436
4021	6	1-1218	4017	5	87	1057	45	2012-08-02 04:49:09.564436
4022	6	1-1149	4017	87	77	1057	92	2012-08-02 04:49:09.564436
4017	6	1-1220	4017	77	46	900	29	2012-08-02 04:49:09.564436
4024	8	1-1160	4023	57	53	387	66	2012-08-02 04:49:10.039834
4025	8	1-1213	4023	53	75	714	53	2012-08-02 04:49:10.039834
4026	8	1-216	4023	75	74	869	97	2012-08-02 04:49:10.039834
4027	8	1-1195	4023	74	75	341	99	2012-08-02 04:49:10.039834
4028	8	1-1182	4023	75	35	201	25	2012-08-02 04:49:10.039834
4029	8	1-1197	4023	35	24	266	89	2012-08-02 04:49:10.039834
4030	8	1-891	4023	24	25	199	98	2012-08-02 04:49:10.039834
4023	8	1-1221	4023	25	57	61	24	2012-08-02 04:49:10.039834
4031	1	1-216	\N	75	75	1905	97	2012-08-02 04:49:10.039834
4033	6	1-1126	4032	92	79	702	85	2012-08-02 04:49:11.066647
4034	6	1-970	4032	79	55	582	67	2012-08-02 04:49:11.066647
4035	6	1-831	4032	55	23	2247	71	2012-08-02 04:49:11.066647
4036	6	1-913	4032	23	98	1630	79	2012-08-02 04:49:11.066647
4037	6	1-1079	4032	98	80	2748	20	2012-08-02 04:49:11.066647
4032	6	1-1222	4032	80	92	1202	18	2012-08-02 04:49:11.066647
4039	7	1-391	4038	11	88	831	19	2012-08-02 04:49:12.050061
4040	7	1-787	4038	88	88	470	26	2012-08-02 04:49:12.050061
4041	7	1-1194	4038	88	59	2446	52	2012-08-02 04:49:12.050061
4042	7	1-699	4038	59	5	478	95	2012-08-02 04:49:12.050061
4043	7	1-1223	4038	5	42	1800	42	2012-08-02 04:49:12.050061
4044	7	1-1143	4038	42	27	2084	80	2012-08-02 04:49:12.050061
4038	7	1-1225	4038	27	11	1848	82	2012-08-02 04:49:12.050061
4046	6	1-1005	4045	44	25	3290	75	2012-08-02 04:49:12.050061
4047	6	1-502	4045	25	53	4138	61	2012-08-02 04:49:12.050061
4048	6	1-1138	4045	53	64	4078	89	2012-08-02 04:49:12.050061
4049	6	1-330	4045	64	42	3929	42	2012-08-02 04:49:12.050061
4050	6	1-1143	4045	42	27	5220	80	2012-08-02 04:49:12.050061
4045	6	1-1225	4045	27	44	5311	82	2012-08-02 04:49:12.050061
4051	1	1-502	\N	25	25	2121	61	2012-08-02 04:49:12.050061
4053	7	1-282	4052	38	35	451	25	2012-08-02 04:49:13.762418
4052	7	1-1229	4052	27	38	121	15	2012-08-02 04:49:13.762418
4054	7	1-1197	4052	35	70	421	89	2012-08-02 04:49:13.762418
4055	7	1-873	4052	70	2	277	27	2012-08-02 04:49:13.762418
4056	7	1-189	4052	2	84	103	47	2012-08-02 04:49:13.762418
4057	7	1-1153	4052	84	5	116	20	2012-08-02 04:49:13.762418
4058	7	1-1218	4052	5	27	144	45	2012-08-02 04:49:13.762418
15920	5	1-4587	15920	49	20	979	63	2012-08-02 05:13:16.598989
4060	4	1-720	4059	67	84	2134	47	2012-08-02 04:49:13.762418
4061	4	1-1153	4059	84	5	2654	20	2012-08-02 04:49:13.762418
4062	4	1-1218	4059	5	27	3667	45	2012-08-02 04:49:13.762418
4059	4	1-1229	4059	27	67	3416	15	2012-08-02 04:49:13.762418
4064	5	1-589	4063	50	64	1331	32	2012-08-02 04:49:13.762418
4065	5	1-998	4063	64	43	1330	36	2012-08-02 04:49:13.762418
4066	5	1-1203	4063	43	25	2253	46	2012-08-02 04:49:13.762418
4067	5	1-778	4063	25	27	2703	45	2012-08-02 04:49:13.762418
4063	5	1-1229	4063	27	50	3224	15	2012-08-02 04:49:13.762418
15926	3	1-1106	15925	4	89	631	10	2012-08-02 05:13:16.598989
4069	3	1-805	4068	32	25	918	83	2012-08-02 04:49:13.762418
4070	3	1-963	4068	25	27	1413	45	2012-08-02 04:49:13.762418
4068	3	1-1229	4068	27	32	1729	15	2012-08-02 04:49:13.762418
15927	3	1-3182	15925	89	49	254	49	2012-08-02 05:13:16.598989
4072	3	1-1078	4071	37	44	3904	63	2012-08-02 04:49:14.259857
4073	3	1-380	4071	44	69	3219	52	2012-08-02 04:49:14.259857
4071	3	1-1230	4071	69	37	9427	15	2012-08-02 04:49:14.259857
15925	3	1-4587	15925	49	4	3472	63	2012-08-02 05:13:16.598989
4075	3	1-1103	4074	29	84	928	37	2012-08-02 04:49:14.75711
4076	3	1-1159	4074	84	62	1076	22	2012-08-02 04:49:14.75711
4074	3	1-1231	4074	62	29	707	74	2012-08-02 04:49:14.75711
4078	7	1-1227	4077	75	77	213	5	2012-08-02 04:49:16.193065
4079	7	1-1073	4077	77	22	115	8	2012-08-02 04:49:16.193065
4080	7	1-22	4077	22	96	222	33	2012-08-02 04:49:16.193065
4081	7	1-827	4077	96	28	147	6	2012-08-02 04:49:16.193065
4082	7	1-1147	4077	28	41	31	24	2012-08-02 04:49:16.193065
4083	7	1-52	4077	41	3	27	59	2012-08-02 04:49:16.193065
4077	7	1-1234	4077	3	75	15	91	2012-08-02 04:49:16.193065
4085	5	1-1227	4084	75	77	1977	5	2012-08-02 04:49:16.193065
4086	5	1-1073	4084	77	96	1112	8	2012-08-02 04:49:16.193065
4087	5	1-1169	4084	96	41	254	24	2012-08-02 04:49:16.193065
4088	5	1-52	4084	41	3	230	59	2012-08-02 04:49:16.193065
4084	5	1-1234	4084	3	75	130	91	2012-08-02 04:49:16.193065
4090	6	1-915	4089	63	79	124	85	2012-08-02 04:49:16.193065
4091	6	1-970	4089	79	55	96	67	2012-08-02 04:49:16.193065
4092	6	1-831	4089	55	83	344	71	2012-08-02 04:49:16.193065
4093	6	1-884	4089	83	41	175	24	2012-08-02 04:49:16.193065
4094	6	1-52	4089	41	3	163	59	2012-08-02 04:49:16.193065
4089	6	1-1234	4089	3	63	95	91	2012-08-02 04:49:16.193065
4096	7	1-1227	4095	75	36	6607	5	2012-08-02 04:49:16.193065
4097	7	1-1140	4095	36	36	4509	94	2012-08-02 04:49:16.193065
4098	7	1-1098	4095	36	15	2431	84	2012-08-02 04:49:16.193065
4099	7	1-928	4095	15	88	531	19	2012-08-02 04:49:16.193065
4100	7	1-787	4095	88	80	314	26	2012-08-02 04:49:16.193065
4101	7	1-261	4095	80	3	642	59	2012-08-02 04:49:16.193065
4095	7	1-1234	4095	3	75	398	91	2012-08-02 04:49:16.193065
4103	6	1-915	4102	63	79	182	85	2012-08-02 04:49:16.193065
4104	6	1-970	4102	79	30	154	67	2012-08-02 04:49:16.193065
4105	6	1-1121	4102	30	59	369	49	2012-08-02 04:49:16.193065
4106	6	1-97	4102	59	80	94	26	2012-08-02 04:49:16.193065
4107	6	1-261	4102	80	3	199	59	2012-08-02 04:49:16.193065
4102	6	1-1234	4102	3	63	127	91	2012-08-02 04:49:16.193065
4108	1	1-261	\N	80	80	2936	59	2012-08-02 04:49:16.193065
4109	1	1-970	\N	79	79	4498	67	2012-08-02 04:49:16.193065
4111	6	1-1013	4110	51	61	1167	3	2012-08-02 04:49:17.629776
4112	6	1-1211	4110	61	78	995	64	2012-08-02 04:49:17.629776
4113	6	1-979	4110	78	34	1337	65	2012-08-02 04:49:17.629776
4114	6	1-898	4110	34	86	1286	43	2012-08-02 04:49:17.629776
4115	6	1-877	4110	86	89	5018	72	2012-08-02 04:49:17.629776
4110	6	1-1236	4110	89	51	4510	14	2012-08-02 04:49:17.629776
4117	5	1-285	4116	34	61	1546	3	2012-08-02 04:49:18.071331
4118	5	1-1211	4116	61	78	649	64	2012-08-02 04:49:18.071331
4119	5	1-979	4116	78	26	430	65	2012-08-02 04:49:18.071331
4120	5	1-1171	4116	26	90	414	90	2012-08-02 04:49:18.071331
4116	5	1-1237	4116	90	34	3298	76	2012-08-02 04:49:18.071331
4122	5	1-787	4121	88	88	494	26	2012-08-02 04:49:20.932381
4123	5	1-1194	4121	88	59	2901	52	2012-08-02 04:49:20.932381
4124	5	1-699	4121	59	65	722	95	2012-08-02 04:49:20.932381
4125	5	1-1104	4121	65	60	352	30	2012-08-02 04:49:20.932381
4121	5	1-1242	4121	60	88	775	19	2012-08-02 04:49:20.932381
4127	4	1-685	4126	15	78	1459	23	2012-08-02 04:49:22.96518
4128	4	1-419	4126	78	46	565	50	2012-08-02 04:49:22.96518
4129	4	1-1188	4126	46	64	921	88	2012-08-02 04:49:22.96518
4126	4	1-1245	4126	64	15	5792	56	2012-08-02 04:49:22.96518
4131	3	1-282	4130	38	56	641	25	2012-08-02 04:49:28.11433
4132	3	1-854	4130	56	72	378	22	2012-08-02 04:49:28.11433
4130	3	1-1252	4130	72	38	238	15	2012-08-02 04:49:28.11433
4134	6	1-1078	4133	37	44	16	63	2012-08-02 04:49:28.11433
4135	6	1-380	4133	44	70	20	52	2012-08-02 04:49:28.11433
4136	6	1-978	4133	70	29	18	31	2012-08-02 04:49:28.11433
4137	6	1-1099	4133	29	84	19	37	2012-08-02 04:49:28.11433
4138	6	1-1159	4133	84	72	24	22	2012-08-02 04:49:28.11433
4133	6	1-1252	4133	72	37	24	15	2012-08-02 04:49:28.11433
4140	6	1-1243	4139	43	94	5619	62	2012-08-02 04:49:28.11433
4141	6	1-600	4139	94	91	556	38	2012-08-02 04:49:28.11433
4142	6	1-1036	4139	91	55	2047	81	2012-08-02 04:49:28.11433
4143	6	1-544	4139	55	82	2877	34	2012-08-02 04:49:28.11433
4144	6	1-980	4139	82	72	1652	22	2012-08-02 04:49:28.11433
4139	6	1-1252	4139	72	43	1694	15	2012-08-02 04:49:28.11433
4146	6	1-282	4145	38	41	2962	25	2012-08-02 04:49:28.11433
4147	6	1-1058	4145	41	43	1692	55	2012-08-02 04:49:28.11433
4148	6	1-55	4145	43	78	771	43	2012-08-02 04:49:28.11433
4149	6	1-1108	4145	78	68	758	21	2012-08-02 04:49:28.11433
4150	6	1-760	4145	68	72	605	22	2012-08-02 04:49:28.11433
4145	6	1-1252	4145	72	38	647	15	2012-08-02 04:49:28.11433
4152	5	1-1078	4151	37	2	477	63	2012-08-02 04:49:28.11433
4153	5	1-937	4151	2	92	506	28	2012-08-02 04:49:28.11433
4154	5	1-1130	4151	92	84	374	37	2012-08-02 04:49:28.11433
4155	5	1-1159	4151	84	72	552	22	2012-08-02 04:49:28.11433
4151	5	1-1252	4151	72	37	637	15	2012-08-02 04:49:28.11433
4157	6	1-1078	4156	37	8	1544	63	2012-08-02 04:49:28.11433
4158	6	1-1113	4156	8	91	325	38	2012-08-02 04:49:28.11433
4159	6	1-1036	4156	91	55	1418	81	2012-08-02 04:49:28.11433
4160	6	1-544	4156	55	82	2363	34	2012-08-02 04:49:28.11433
4161	6	1-980	4156	82	72	1608	22	2012-08-02 04:49:28.11433
4156	6	1-1252	4156	72	37	1955	15	2012-08-02 04:49:28.11433
15929	5	1-4249	15928	5	49	53	29	2012-08-02 05:13:16.598989
4163	7	1-589	4162	50	64	1458	32	2012-08-02 04:49:28.11433
4164	7	1-998	4162	64	19	1452	36	2012-08-02 04:49:28.11433
4165	7	1-723	4162	19	93	1369	7	2012-08-02 04:49:28.11433
4166	7	1-1239	4162	93	81	1363	41	2012-08-02 04:49:28.11433
4167	7	1-1175	4162	81	54	2613	44	2012-08-02 04:49:28.11433
4168	7	1-997	4162	54	72	2887	22	2012-08-02 04:49:28.11433
4162	7	1-1252	4162	72	50	3543	15	2012-08-02 04:49:28.11433
4170	6	1-1126	4169	92	20	602	85	2012-08-02 04:49:33.439393
4171	6	1-1082	4169	20	30	214	67	2012-08-02 04:49:33.439393
4172	6	1-1121	4169	30	78	432	49	2012-08-02 04:49:33.439393
4173	6	1-1240	4169	78	31	701	60	2012-08-02 04:49:33.439393
4174	6	1-1192	4169	31	16	976	31	2012-08-02 04:49:33.439393
4169	6	1-1261	4169	16	92	1197	18	2012-08-02 04:49:33.439393
4176	4	1-1088	4175	97	28	283	70	2012-08-02 04:49:33.439393
4177	4	1-607	4175	28	81	74	16	2012-08-02 04:49:33.439393
4178	4	1-830	4175	81	16	190	31	2012-08-02 04:49:33.439393
4175	4	1-1261	4175	16	97	273	18	2012-08-02 04:49:33.439393
4180	5	1-1078	4179	37	44	659	63	2012-08-02 04:49:33.936492
4181	5	1-380	4179	44	70	915	52	2012-08-02 04:49:33.936492
4182	5	1-978	4179	70	16	872	31	2012-08-02 04:49:33.936492
4183	5	1-1261	4179	16	11	1113	18	2012-08-02 04:49:33.936492
4179	5	1-1262	4179	11	37	946	15	2012-08-02 04:49:33.936492
4185	7	1-1155	4184	49	3	273	59	2012-08-02 04:49:34.422505
4186	7	1-1234	4184	3	63	173	91	2012-08-02 04:49:34.422505
4187	7	1-915	4184	63	20	244	85	2012-08-02 04:49:34.422505
4188	7	1-1082	4184	20	30	102	67	2012-08-02 04:49:34.422505
4189	7	1-1121	4184	30	57	242	49	2012-08-02 04:49:34.422505
4190	7	1-1244	4184	57	54	444	92	2012-08-02 04:49:34.422505
4184	7	1-1263	4184	54	49	799	100	2012-08-02 04:49:34.422505
4192	4	1-719	4191	44	63	121	71	2012-08-02 04:49:36.212131
4193	4	1-377	4191	63	78	67	43	2012-08-02 04:49:36.212131
4194	4	1-1108	4191	78	78	61	21	2012-08-02 04:49:36.212131
4191	4	1-1266	4191	78	44	274	19	2012-08-02 04:49:36.212131
4196	5	1-787	4195	88	24	1159	26	2012-08-02 04:49:36.212131
4197	5	1-971	4195	24	28	1182	11	2012-08-02 04:49:36.212131
4198	5	1-30	4195	28	78	656	43	2012-08-02 04:49:36.212131
4199	5	1-1108	4195	78	78	615	21	2012-08-02 04:49:36.212131
4195	5	1-1266	4195	78	88	2850	19	2012-08-02 04:49:36.212131
4201	4	1-941	4200	35	28	34	11	2012-08-02 04:49:36.212131
4202	4	1-30	4200	28	78	20	43	2012-08-02 04:49:36.212131
4203	4	1-1108	4200	78	78	19	21	2012-08-02 04:49:36.212131
4200	4	1-1266	4200	78	35	89	19	2012-08-02 04:49:36.212131
4205	2	1-1244	4204	57	94	7066	92	2012-08-02 04:49:37.692629
4204	2	1-1269	4204	94	57	7414	49	2012-08-02 04:49:37.692629
4207	7	1-1240	4206	78	31	300	60	2012-08-02 04:49:37.692629
4208	7	1-1192	4206	31	16	387	31	2012-08-02 04:49:37.692629
4209	7	1-1261	4206	16	11	439	18	2012-08-02 04:49:37.692629
4210	7	1-1262	4206	11	37	331	15	2012-08-02 04:49:37.692629
4211	7	1-1078	4206	37	24	205	63	2012-08-02 04:49:37.692629
4212	7	1-916	4206	24	94	125	92	2012-08-02 04:49:37.692629
4206	7	1-1269	4206	94	78	200	49	2012-08-02 04:49:37.692629
4214	8	1-1240	4213	78	31	752	60	2012-08-02 04:49:37.692629
4215	8	1-1192	4213	31	16	977	31	2012-08-02 04:49:37.692629
4216	8	1-1261	4213	16	11	1117	18	2012-08-02 04:49:37.692629
4217	8	1-1262	4213	11	32	850	15	2012-08-02 04:49:37.692629
4218	8	1-805	4213	32	25	343	83	2012-08-02 04:49:37.692629
4219	8	1-963	4213	25	87	401	45	2012-08-02 04:49:37.692629
4220	8	1-1149	4213	87	94	309	92	2012-08-02 04:49:37.692629
4213	8	1-1269	4213	94	78	497	49	2012-08-02 04:49:37.692629
4222	6	1-1240	4221	78	88	1014	60	2012-08-02 04:49:37.692629
4223	6	1-1014	4221	88	74	465	97	2012-08-02 04:49:37.692629
4224	6	1-1195	4221	74	65	181	99	2012-08-02 04:49:37.692629
4225	6	1-1248	4221	65	24	328	63	2012-08-02 04:49:37.692629
4226	6	1-916	4221	24	94	257	92	2012-08-02 04:49:37.692629
4221	6	1-1269	4221	94	78	526	49	2012-08-02 04:49:37.692629
4228	7	1-546	4227	43	61	406	3	2012-08-02 04:49:38.675835
4229	7	1-1211	4227	61	78	220	64	2012-08-02 04:49:38.675835
4230	7	1-979	4227	78	26	189	65	2012-08-02 04:49:38.675835
4231	7	1-1171	4227	26	90	235	90	2012-08-02 04:49:38.675835
4232	7	1-1237	4227	90	53	2415	76	2012-08-02 04:49:38.675835
4233	7	1-763	4227	53	13	1455	9	2012-08-02 04:49:38.675835
4227	7	1-1271	4227	13	43	673	99	2012-08-02 04:49:38.675835
4235	4	1-1248	4234	65	52	1484	63	2012-08-02 04:49:38.675835
4236	4	1-846	4234	52	83	1819	33	2012-08-02 04:49:38.675835
4237	4	1-729	4234	83	13	1427	9	2012-08-02 04:49:38.675835
4234	4	1-1271	4234	13	65	916	99	2012-08-02 04:49:38.675835
4239	4	1-546	4238	43	83	88	3	2012-08-02 04:49:38.675835
4240	4	1-1081	4238	83	83	166	33	2012-08-02 04:49:38.675835
4241	4	1-729	4238	83	13	145	9	2012-08-02 04:49:38.675835
4238	4	1-1271	4238	13	43	99	99	2012-08-02 04:49:38.675835
4243	4	1-936	4242	48	37	163	95	2012-08-02 04:49:39.957609
4244	4	1-1145	4242	37	46	101	29	2012-08-02 04:49:39.957609
4245	4	1-356	4242	46	89	229	78	2012-08-02 04:49:39.957609
4242	4	1-1274	4242	89	48	435	11	2012-08-02 04:49:39.957609
4247	7	1-891	4246	24	100	324	98	2012-08-02 04:49:43.039873
4248	7	1-934	4246	100	93	553	7	2012-08-02 04:49:43.039873
4249	7	1-1239	4246	93	81	544	41	2012-08-02 04:49:43.039873
4250	7	1-1175	4246	81	21	1031	44	2012-08-02 04:49:43.039873
4251	7	1-1017	4246	21	24	251	26	2012-08-02 04:49:43.039873
4252	7	1-971	4246	24	1	392	11	2012-08-02 04:49:43.039873
4246	7	1-1280	4246	1	24	429	89	2012-08-02 04:49:43.039873
4254	4	1-96	4253	13	40	806	39	2012-08-02 04:49:54.461358
4255	4	1-332	4253	40	98	742	79	2012-08-02 04:49:54.461358
4256	4	1-1079	4253	98	45	1082	20	2012-08-02 04:49:54.461358
4253	4	1-1300	4253	45	13	874	85	2012-08-02 04:49:54.461358
4258	6	1-646	4257	2	24	994	37	2012-08-02 04:49:54.461358
4259	6	1-1289	4257	24	45	5457	31	2012-08-02 04:49:54.461358
4260	6	1-561	4257	45	67	420	15	2012-08-02 04:49:54.461358
4261	6	1-720	4257	67	84	293	47	2012-08-02 04:49:54.461358
4262	6	1-1153	4257	84	45	407	20	2012-08-02 04:49:54.461358
4257	6	1-1300	4257	45	2	372	85	2012-08-02 04:49:54.461358
15930	5	1-4553	15928	49	85	31	97	2012-08-02 05:13:16.598989
4264	3	1-1287	4263	2	65	6686	10	2012-08-02 04:49:54.461358
4265	3	1-749	4263	65	45	6253	20	2012-08-02 04:49:54.461358
4263	3	1-1300	4263	45	2	5743	85	2012-08-02 04:49:54.461358
15931	5	1-4522	15928	85	3	15	91	2012-08-02 05:13:16.598989
4267	7	1-565	4266	10	18	469	28	2012-08-02 04:49:56.902613
4268	7	1-18	4266	18	61	444	3	2012-08-02 04:49:56.902613
4269	7	1-1211	4266	61	78	388	64	2012-08-02 04:49:56.902613
4270	7	1-979	4266	78	26	535	65	2012-08-02 04:49:56.902613
4271	7	1-1171	4266	26	77	1073	90	2012-08-02 04:49:56.902613
4272	7	1-328	4266	77	19	1273	34	2012-08-02 04:49:56.902613
4266	7	1-1303	4266	19	10	1393	29	2012-08-02 04:49:56.902613
15932	5	1-2262	15928	3	49	4	49	2012-08-02 05:13:16.598989
4274	5	1-64	4273	42	42	1202	36	2012-08-02 04:49:57.367116
4275	5	1-860	4273	42	4	591	34	2012-08-02 04:49:57.367116
4276	5	1-1177	4273	4	46	467	50	2012-08-02 04:49:57.367116
4277	5	1-1188	4273	46	60	1113	88	2012-08-02 04:49:57.367116
4273	5	1-1304	4273	60	42	1066	73	2012-08-02 04:49:57.367116
4279	4	1-64	4278	42	2	244	36	2012-08-02 04:49:57.367116
4280	4	1-417	4278	2	100	155	95	2012-08-02 04:49:57.367116
4281	4	1-1258	4278	100	60	168	88	2012-08-02 04:49:57.367116
4278	4	1-1304	4278	60	42	223	73	2012-08-02 04:49:57.367116
4283	4	1-640	4282	37	70	4066	52	2012-08-02 04:49:57.367116
4284	4	1-978	4282	70	97	3686	31	2012-08-02 04:49:57.367116
4285	4	1-935	4282	97	60	4212	88	2012-08-02 04:49:57.367116
4282	4	1-1304	4282	60	37	6528	73	2012-08-02 04:49:57.367116
4287	7	1-1248	4286	65	24	1180	63	2012-08-02 04:50:02.525315
4288	7	1-916	4286	24	81	911	92	2012-08-02 04:50:02.525315
4289	7	1-1042	4286	81	88	1327	17	2012-08-02 04:50:02.525315
4290	7	1-1285	4286	88	1	984	94	2012-08-02 04:50:02.525315
4291	7	1-1288	4286	1	18	688	28	2012-08-02 04:50:02.525315
4292	7	1-18	4286	18	28	618	3	2012-08-02 04:50:02.525315
4286	7	1-1312	4286	28	65	661	99	2012-08-02 04:50:02.525315
4294	2	1-546	4293	43	28	5433	3	2012-08-02 04:50:02.525315
4293	2	1-1312	4293	28	43	5843	99	2012-08-02 04:50:02.525315
4296	4	1-1034	4295	79	78	383	49	2012-08-02 04:50:02.525315
4297	4	1-1240	4295	78	58	736	60	2012-08-02 04:50:02.525315
4298	4	1-456	4295	58	28	791	3	2012-08-02 04:50:02.525315
4295	4	1-1312	4295	28	79	856	99	2012-08-02 04:50:02.525315
4300	5	1-1240	4299	78	58	1765	60	2012-08-02 04:50:03.011659
4301	5	1-456	4299	58	61	1483	3	2012-08-02 04:50:03.011659
4302	5	1-1211	4299	61	40	976	64	2012-08-02 04:50:03.011659
4303	5	1-51	4299	40	27	429	7	2012-08-02 04:50:03.011659
4299	5	1-1313	4299	27	78	1174	49	2012-08-02 04:50:03.011659
4305	7	1-1282	4304	8	13	6947	9	2012-08-02 04:50:03.011659
4306	7	1-1271	4304	13	14	4463	99	2012-08-02 04:50:03.011659
4307	7	1-770	4304	14	40	1668	16	2012-08-02 04:50:03.011659
4308	7	1-1233	4304	40	44	1601	82	2012-08-02 04:50:03.011659
4309	7	1-682	4304	44	4	1232	8	2012-08-02 04:50:03.011659
4310	7	1-4	4304	4	27	973	7	2012-08-02 04:50:03.011659
4304	7	1-1313	4304	27	8	3052	49	2012-08-02 04:50:03.011659
4312	8	1-1253	4311	93	26	466	65	2012-08-02 04:50:03.011659
4313	8	1-1171	4311	26	46	850	90	2012-08-02 04:50:03.011659
4314	8	1-986	4311	46	41	334	71	2012-08-02 04:50:03.011659
4315	8	1-177	4311	41	53	104	61	2012-08-02 04:50:03.011659
4316	8	1-1138	4311	53	24	103	89	2012-08-02 04:50:03.011659
4317	8	1-891	4311	24	100	72	98	2012-08-02 04:50:03.011659
4318	8	1-934	4311	100	27	117	7	2012-08-02 04:50:03.011659
4311	8	1-1313	4311	27	93	387	49	2012-08-02 04:50:03.011659
4320	8	1-1240	4319	78	5	797	60	2012-08-02 04:50:03.011659
4321	8	1-788	4319	5	61	700	43	2012-08-02 04:50:03.011659
4322	8	1-1146	4319	61	28	585	70	2012-08-02 04:50:03.011659
4323	8	1-607	4319	28	40	154	16	2012-08-02 04:50:03.011659
4324	8	1-1233	4319	40	44	163	82	2012-08-02 04:50:03.011659
4325	8	1-682	4319	44	4	139	8	2012-08-02 04:50:03.011659
4326	8	1-4	4319	4	27	121	7	2012-08-02 04:50:03.011659
4319	8	1-1313	4319	27	78	419	49	2012-08-02 04:50:03.011659
4328	8	1-1240	4327	78	5	3483	60	2012-08-02 04:50:03.011659
4329	8	1-788	4327	5	22	3104	43	2012-08-02 04:50:03.011659
4330	8	1-1246	4327	22	28	2342	70	2012-08-02 04:50:03.011659
4331	8	1-607	4327	28	40	625	16	2012-08-02 04:50:03.011659
4332	8	1-1233	4327	40	44	672	82	2012-08-02 04:50:03.011659
4333	8	1-682	4327	44	4	580	8	2012-08-02 04:50:03.011659
4334	8	1-4	4327	4	27	513	7	2012-08-02 04:50:03.011659
4327	8	1-1313	4327	27	78	1805	49	2012-08-02 04:50:03.011659
4336	8	1-39	4335	21	37	1061	52	2012-08-02 04:50:03.011659
4337	8	1-1204	4335	37	73	1495	12	2012-08-02 04:50:03.011659
4338	8	1-1056	4335	73	1	1903	94	2012-08-02 04:50:03.011659
4339	8	1-1288	4335	1	18	1402	28	2012-08-02 04:50:03.011659
4340	8	1-18	4335	18	61	1328	3	2012-08-02 04:50:03.011659
4341	8	1-1211	4335	61	40	1162	64	2012-08-02 04:50:03.011659
4342	8	1-51	4335	40	27	679	7	2012-08-02 04:50:03.011659
4335	8	1-1313	4335	27	21	2476	49	2012-08-02 04:50:03.011659
4343	1	1-18	\N	18	18	3006	3	2012-08-02 04:50:03.011659
4345	7	1-758	4344	82	100	204	98	2012-08-02 04:50:03.520664
4346	7	1-934	4344	100	93	240	7	2012-08-02 04:50:03.520664
4347	7	1-1239	4344	93	99	162	41	2012-08-02 04:50:03.520664
4348	7	1-1311	4344	99	24	287	63	2012-08-02 04:50:03.520664
4349	7	1-916	4344	24	54	154	92	2012-08-02 04:50:03.520664
4350	7	1-1263	4344	54	48	191	100	2012-08-02 04:50:03.520664
4344	7	1-1314	4344	48	82	188	67	2012-08-02 04:50:03.520664
4352	7	1-1309	4351	12	39	442	65	2012-08-02 04:50:03.520664
4353	7	1-1305	4351	39	99	345	73	2012-08-02 04:50:03.520664
4354	7	1-398	4351	99	99	237	41	2012-08-02 04:50:03.520664
4355	7	1-1311	4351	99	24	561	63	2012-08-02 04:50:03.520664
4356	7	1-916	4351	24	54	351	92	2012-08-02 04:50:03.520664
4357	7	1-1263	4351	54	48	509	100	2012-08-02 04:50:03.520664
4351	7	1-1314	4351	48	12	584	67	2012-08-02 04:50:03.520664
4359	8	1-1121	4358	30	93	908	49	2012-08-02 04:50:03.520664
4360	8	1-1253	4358	93	39	1048	65	2012-08-02 04:50:03.520664
4361	8	1-1305	4358	39	37	929	73	2012-08-02 04:50:03.520664
4362	8	1-640	4358	37	40	627	52	2012-08-02 04:50:03.520664
4363	8	1-517	4358	40	61	226	3	2012-08-02 04:50:03.520664
4364	8	1-1211	4358	61	99	172	64	2012-08-02 04:50:03.520664
4365	8	1-1283	4358	99	48	322	100	2012-08-02 04:50:03.520664
4358	8	1-1314	4358	48	30	419	67	2012-08-02 04:50:03.520664
15928	5	1-4587	15928	49	5	54	63	2012-08-02 05:13:16.598989
4367	7	1-1121	4366	30	14	4793	49	2012-08-02 04:50:03.520664
4368	7	1-440	4366	14	36	2519	94	2012-08-02 04:50:03.520664
4369	7	1-1098	4366	36	11	1255	84	2012-08-02 04:50:03.520664
4370	7	1-1047	4366	11	61	1066	3	2012-08-02 04:50:03.520664
4371	7	1-1211	4366	61	99	837	64	2012-08-02 04:50:03.520664
4372	7	1-1283	4366	99	48	1608	100	2012-08-02 04:50:03.520664
4366	7	1-1314	4366	48	30	2151	67	2012-08-02 04:50:03.520664
4374	6	1-998	4373	64	47	1015	36	2012-08-02 04:50:05.976552
4375	6	1-977	4373	47	99	607	64	2012-08-02 04:50:05.976552
4376	6	1-1283	4373	99	48	1101	100	2012-08-02 04:50:05.976552
4377	6	1-1314	4373	48	20	1390	67	2012-08-02 04:50:05.976552
4378	6	1-1123	4373	20	69	1267	60	2012-08-02 04:50:05.976552
4373	6	1-1320	4373	69	64	1164	32	2012-08-02 04:50:05.976552
15934	4	1-2720	15933	68	9	2846	93	2012-08-02 05:13:17.029915
4380	3	1-1228	4379	34	71	3317	51	2012-08-02 04:50:07.445654
4381	3	1-725	4379	71	99	807	66	2012-08-02 04:50:07.445654
4379	3	1-1321	4379	99	34	1039	21	2012-08-02 04:50:07.445654
15935	4	1-3666	15933	9	41	1777	13	2012-08-02 05:13:17.029915
4383	6	1-1057	4382	77	97	2373	34	2012-08-02 04:50:08.870915
4384	6	1-1319	4382	97	22	3179	43	2012-08-02 04:50:08.870915
4385	6	1-1246	4382	22	28	1689	70	2012-08-02 04:50:08.870915
4386	6	1-607	4382	28	44	317	16	2012-08-02 04:50:08.870915
4387	6	1-984	4382	44	21	352	10	2012-08-02 04:50:08.870915
4382	6	1-1324	4382	21	77	861	57	2012-08-02 04:50:08.870915
15936	4	1-4347	15933	41	35	1565	28	2012-08-02 05:13:17.029915
4389	5	1-42	4388	34	84	1498	47	2012-08-02 04:50:08.870915
4390	5	1-1153	4388	84	45	1706	20	2012-08-02 04:50:08.870915
4391	5	1-1300	4388	45	2	1172	85	2012-08-02 04:50:08.870915
4392	5	1-1287	4388	2	21	1020	10	2012-08-02 04:50:08.870915
4388	5	1-1324	4388	21	34	2547	57	2012-08-02 04:50:08.870915
4394	5	1-1057	4393	77	19	1627	34	2012-08-02 04:50:08.870915
4395	5	1-1303	4393	19	22	1578	29	2012-08-02 04:50:08.870915
4396	5	1-453	4393	22	82	1290	93	2012-08-02 04:50:08.870915
4397	5	1-259	4393	82	21	142	10	2012-08-02 04:50:08.870915
4393	5	1-1324	4393	21	77	453	57	2012-08-02 04:50:08.870915
4399	2	1-613	4398	13	14	639	20	2012-08-02 04:50:09.35699
4398	2	1-1325	4398	14	13	445	61	2012-08-02 04:50:09.35699
4401	7	1-868	4400	88	41	155	24	2012-08-02 04:50:09.35699
4402	7	1-52	4400	41	3	124	59	2012-08-02 04:50:09.35699
4403	7	1-1234	4400	3	75	63	91	2012-08-02 04:50:09.35699
4404	7	1-1227	4400	75	36	844	5	2012-08-02 04:50:09.35699
4405	7	1-1140	4400	36	89	451	94	2012-08-02 04:50:09.35699
4406	7	1-1249	4400	89	14	564	20	2012-08-02 04:50:09.35699
4400	7	1-1325	4400	14	88	436	61	2012-08-02 04:50:09.35699
4408	6	1-1250	4407	96	28	4697	6	2012-08-02 04:50:09.35699
4409	6	1-1147	4407	28	20	1061	24	2012-08-02 04:50:09.35699
4410	6	1-739	4407	20	88	1359	17	2012-08-02 04:50:09.35699
4411	6	1-1285	4407	88	89	927	94	2012-08-02 04:50:09.35699
4412	6	1-1249	4407	89	14	1333	20	2012-08-02 04:50:09.35699
4407	6	1-1325	4407	14	96	1185	61	2012-08-02 04:50:09.35699
4414	5	1-1142	4413	31	81	191	92	2012-08-02 04:50:09.35699
4415	5	1-1042	4413	81	88	256	17	2012-08-02 04:50:09.35699
4416	5	1-1285	4413	88	89	175	94	2012-08-02 04:50:09.35699
4417	5	1-1249	4413	89	14	252	20	2012-08-02 04:50:09.35699
4413	5	1-1325	4413	14	31	224	61	2012-08-02 04:50:09.35699
4419	5	1-1115	4418	68	18	1985	42	2012-08-02 04:50:09.887379
4420	5	1-1299	4418	18	37	9044	93	2012-08-02 04:50:09.887379
4421	5	1-1134	4418	37	74	1949	97	2012-08-02 04:50:09.887379
4422	5	1-1195	4418	74	92	751	99	2012-08-02 04:50:09.887379
4418	5	1-1326	4418	92	68	3480	1	2012-08-02 04:50:09.887379
4424	6	1-1284	4423	31	34	3953	47	2012-08-02 04:50:09.887379
4425	6	1-740	4423	34	19	4448	34	2012-08-02 04:50:09.887379
4426	6	1-1303	4423	19	32	4836	29	2012-08-02 04:50:09.887379
4427	6	1-1016	4423	32	74	1041	97	2012-08-02 04:50:09.887379
4428	6	1-1195	4423	74	92	417	99	2012-08-02 04:50:09.887379
4423	6	1-1326	4423	92	31	2009	1	2012-08-02 04:50:09.887379
4430	6	1-470	4429	38	40	229	39	2012-08-02 04:50:10.384596
4431	6	1-332	4429	40	98	203	79	2012-08-02 04:50:10.384596
4432	6	1-1079	4429	98	14	284	20	2012-08-02 04:50:10.384596
4433	6	1-1325	4429	14	53	227	61	2012-08-02 04:50:10.384596
4434	6	1-1138	4429	53	42	194	89	2012-08-02 04:50:10.384596
4429	6	1-1327	4429	42	38	378	83	2012-08-02 04:50:10.384596
4436	7	1-1242	4435	60	17	1462	19	2012-08-02 04:50:11.202161
4437	7	1-992	4435	17	59	1857	39	2012-08-02 04:50:11.202161
4438	7	1-950	4435	59	87	925	45	2012-08-02 04:50:11.202161
4439	7	1-1149	4435	87	80	597	92	2012-08-02 04:50:11.202161
4440	7	1-1152	4435	80	100	705	50	2012-08-02 04:50:11.202161
4441	7	1-765	4435	100	7	1706	4	2012-08-02 04:50:11.202161
4435	7	1-1330	4435	7	60	1054	30	2012-08-02 04:50:11.202161
4443	5	1-847	4442	57	17	2	21	2012-08-02 04:50:11.202161
4444	5	1-875	4442	17	92	1	99	2012-08-02 04:50:11.202161
4445	5	1-1326	4442	92	79	4	1	2012-08-02 04:50:11.202161
4446	5	1-531	4442	79	7	6	4	2012-08-02 04:50:11.202161
4442	5	1-1330	4442	7	57	4	30	2012-08-02 04:50:11.202161
4448	6	1-1294	4447	64	11	425	84	2012-08-02 04:50:11.202161
4449	6	1-1047	4447	11	28	286	3	2012-08-02 04:50:11.202161
4450	6	1-1312	4447	28	92	228	99	2012-08-02 04:50:11.202161
4451	6	1-1326	4447	92	79	787	1	2012-08-02 04:50:11.202161
4452	6	1-531	4447	79	7	1177	4	2012-08-02 04:50:11.202161
4447	6	1-1330	4447	7	64	814	30	2012-08-02 04:50:11.202161
4454	5	1-1242	4453	60	35	263	19	2012-08-02 04:50:11.202161
4455	5	1-941	4453	35	1	126	11	2012-08-02 04:50:11.202161
4456	5	1-1280	4453	1	85	118	89	2012-08-02 04:50:11.202161
4457	5	1-1165	4453	85	7	182	4	2012-08-02 04:50:11.202161
4453	5	1-1330	4453	7	60	146	30	2012-08-02 04:50:11.202161
4459	7	1-1091	4458	82	40	638	16	2012-08-02 04:50:11.202161
4460	7	1-1233	4458	40	44	690	82	2012-08-02 04:50:11.202161
4461	7	1-682	4458	44	75	598	8	2012-08-02 04:50:11.202161
4462	7	1-965	4458	75	53	641	54	2012-08-02 04:50:11.202161
4463	7	1-1185	4458	53	67	1460	2	2012-08-02 04:50:11.202161
4464	7	1-321	4458	67	7	1458	4	2012-08-02 04:50:11.202161
4458	7	1-1330	4458	7	82	1378	30	2012-08-02 04:50:11.202161
15933	4	1-4588	15933	35	68	2354	92	2012-08-02 05:13:17.029915
4466	6	1-848	4465	4	16	321	86	2012-08-02 04:50:11.202161
4467	6	1-1065	4465	16	23	1024	71	2012-08-02 04:50:11.202161
4468	6	1-913	4465	23	98	789	79	2012-08-02 04:50:11.202161
4469	6	1-1079	4465	98	7	1413	20	2012-08-02 04:50:11.202161
4470	6	1-598	4465	7	7	842	4	2012-08-02 04:50:11.202161
4465	6	1-1330	4465	7	4	825	30	2012-08-02 04:50:11.202161
4472	4	1-916	4471	24	54	397	92	2012-08-02 04:50:11.710862
4473	4	1-1263	4471	54	70	379	100	2012-08-02 04:50:11.710862
4474	4	1-762	4471	70	52	174	40	2012-08-02 04:50:11.710862
4471	4	1-1332	4471	52	24	966	63	2012-08-02 04:50:11.710862
4476	7	1-813	4475	4	26	2867	51	2012-08-02 04:50:11.710862
4477	7	1-597	4475	26	72	1397	68	2012-08-02 04:50:11.710862
4478	7	1-1310	4475	72	3	466	77	2012-08-02 04:50:11.710862
4479	7	1-1168	4475	3	37	254	52	2012-08-02 04:50:11.710862
4480	7	1-1204	4475	37	69	289	12	2012-08-02 04:50:11.710862
4481	7	1-1315	4475	69	52	261	40	2012-08-02 04:50:11.710862
4475	7	1-1332	4475	52	4	2320	63	2012-08-02 04:50:11.710862
4483	4	1-1113	4482	8	66	1145	38	2012-08-02 04:50:11.710862
4484	4	1-1117	4482	66	74	688	58	2012-08-02 04:50:11.710862
4485	4	1-327	4482	74	52	723	40	2012-08-02 04:50:11.710862
4482	4	1-1332	4482	52	8	6434	63	2012-08-02 04:50:11.710862
4487	7	1-51	4486	40	93	1375	7	2012-08-02 04:50:14.616612
4488	7	1-1239	4486	93	99	1134	41	2012-08-02 04:50:14.616612
4489	7	1-1311	4486	99	24	2441	63	2012-08-02 04:50:14.616612
4490	7	1-916	4486	24	77	1592	92	2012-08-02 04:50:14.616612
4491	7	1-1220	4486	77	46	1106	29	2012-08-02 04:50:14.616612
4492	7	1-356	4486	46	72	2452	78	2012-08-02 04:50:14.616612
4486	7	1-1337	4486	72	40	2933	64	2012-08-02 04:50:14.616612
4494	7	1-578	4493	29	89	424	94	2012-08-02 04:50:14.616612
4495	7	1-1249	4493	89	14	671	20	2012-08-02 04:50:14.616612
4496	7	1-1325	4493	14	31	657	61	2012-08-02 04:50:14.616612
4497	7	1-1142	4493	31	77	614	92	2012-08-02 04:50:14.616612
4498	7	1-1220	4493	77	46	511	29	2012-08-02 04:50:14.616612
4499	7	1-356	4493	46	72	1356	78	2012-08-02 04:50:14.616612
4493	7	1-1337	4493	72	29	1942	64	2012-08-02 04:50:14.616612
4501	5	1-965	4500	75	53	1192	54	2012-08-02 04:50:16.063709
4502	5	1-1185	4500	53	76	2019	2	2012-08-02 04:50:16.063709
4503	5	1-1094	4500	76	47	1925	53	2012-08-02 04:50:16.063709
4504	5	1-856	4500	47	31	2047	72	2012-08-02 04:50:16.063709
4500	5	1-1339	4500	31	75	1495	8	2012-08-02 04:50:16.063709
4506	2	1-429	4505	72	31	2449	72	2012-08-02 04:50:16.063709
4505	2	1-1339	4505	31	72	1841	8	2012-08-02 04:50:16.063709
4508	3	1-1142	4507	31	80	1220	92	2012-08-02 04:50:18.449453
4509	3	1-1152	4507	80	76	1630	50	2012-08-02 04:50:18.449453
4507	3	1-1344	4507	76	31	1747	61	2012-08-02 04:50:18.449453
4511	7	1-1142	4510	31	9	4478	92	2012-08-02 04:50:18.449453
4512	7	1-1342	4510	9	83	5646	71	2012-08-02 04:50:18.449453
4513	7	1-884	4510	83	89	2808	24	2012-08-02 04:50:18.449453
4514	7	1-1200	4510	89	77	6480	48	2012-08-02 04:50:18.449453
4515	7	1-1212	4510	77	4	3045	34	2012-08-02 04:50:18.449453
4516	7	1-1177	4510	4	76	4161	50	2012-08-02 04:50:18.449453
4510	7	1-1344	4510	76	31	5347	61	2012-08-02 04:50:18.449453
4518	5	1-194	4517	88	26	9	65	2012-08-02 04:50:18.449453
4519	5	1-1171	4517	26	77	15	90	2012-08-02 04:50:18.449453
4520	5	1-328	4517	77	4	16	34	2012-08-02 04:50:18.449453
4521	5	1-1177	4517	4	76	22	50	2012-08-02 04:50:18.449453
4517	5	1-1344	4517	76	88	29	61	2012-08-02 04:50:18.449453
4523	5	1-1142	4522	31	9	1425	92	2012-08-02 04:50:18.449453
4524	5	1-1342	4522	9	83	1858	71	2012-08-02 04:50:18.449453
4525	5	1-642	4522	83	4	877	34	2012-08-02 04:50:18.449453
4526	5	1-1177	4522	4	76	1239	50	2012-08-02 04:50:18.449453
4522	5	1-1344	4522	76	31	1646	61	2012-08-02 04:50:18.449453
4528	5	1-950	4527	59	50	1209	45	2012-08-02 04:50:19.908321
4529	5	1-1119	4527	50	13	1610	62	2012-08-02 04:50:19.908321
4530	5	1-1110	4527	13	1	159	25	2012-08-02 04:50:19.908321
4531	5	1-1333	4527	1	3	61	12	2012-08-02 04:50:19.908321
4527	5	1-1347	4527	3	59	1931	39	2012-08-02 04:50:19.908321
4533	5	1-1054	4532	99	50	255	45	2012-08-02 04:50:19.908321
4534	5	1-1119	4532	50	13	363	62	2012-08-02 04:50:19.908321
4535	5	1-1110	4532	13	1	38	25	2012-08-02 04:50:19.908321
4536	5	1-1333	4532	1	3	16	12	2012-08-02 04:50:19.908321
4532	5	1-1347	4532	3	99	529	39	2012-08-02 04:50:19.908321
4538	5	1-971	4537	24	1	1572	11	2012-08-02 04:50:21.013014
4539	5	1-1280	4537	1	42	1758	89	2012-08-02 04:50:21.013014
4540	5	1-1327	4537	42	25	4269	83	2012-08-02 04:50:21.013014
4541	5	1-963	4537	25	3	6448	45	2012-08-02 04:50:21.013014
4537	5	1-1350	4537	3	24	1012	26	2012-08-02 04:50:21.013014
4543	6	1-1140	4542	36	89	26	94	2012-08-02 04:50:21.974134
4544	6	1-1249	4542	89	45	34	20	2012-08-02 04:50:21.974134
4545	6	1-1300	4542	45	2	26	85	2012-08-02 04:50:21.974134
4546	6	1-646	4542	2	84	63	37	2012-08-02 04:50:21.974134
4547	6	1-1159	4542	84	28	79	22	2012-08-02 04:50:21.974134
4542	6	1-1352	4542	28	36	48	5	2012-08-02 04:50:21.974134
4549	3	1-417	4548	2	35	3457	95	2012-08-02 04:50:23.399057
4550	3	1-620	4548	35	98	3233	31	2012-08-02 04:50:23.399057
4548	3	1-1355	4548	98	2	3716	36	2012-08-02 04:50:23.399057
4552	8	1-1276	4551	53	14	2704	99	2012-08-02 04:50:24.901635
4553	8	1-770	4551	14	40	1176	16	2012-08-02 04:50:24.901635
4554	8	1-1233	4551	40	27	1313	82	2012-08-02 04:50:24.901635
4555	8	1-1010	4551	27	73	1624	12	2012-08-02 04:50:24.901635
4556	8	1-1056	4551	73	34	2071	94	2012-08-02 04:50:24.901635
4557	8	1-962	4551	34	99	1617	66	2012-08-02 04:50:24.901635
4558	8	1-1321	4551	99	13	2127	21	2012-08-02 04:50:24.901635
4551	8	1-1358	4551	13	53	4047	70	2012-08-02 04:50:24.901635
4560	7	1-1305	4559	39	37	883	73	2012-08-02 04:50:25.884942
4561	7	1-640	4559	37	37	609	52	2012-08-02 04:50:25.884942
4562	7	1-1204	4559	37	73	764	12	2012-08-02 04:50:25.884942
4563	7	1-1056	4559	73	89	765	94	2012-08-02 04:50:25.884942
4564	7	1-1249	4559	89	45	992	20	2012-08-02 04:50:25.884942
4565	7	1-1300	4559	45	11	771	85	2012-08-02 04:50:25.884942
4559	7	1-1360	4559	11	39	1106	65	2012-08-02 04:50:25.884942
4567	7	1-1305	4566	39	70	82	73	2012-08-02 04:50:25.884942
4568	7	1-441	4566	70	28	95	22	2012-08-02 04:50:25.884942
4569	7	1-1352	4566	28	36	67	5	2012-08-02 04:50:25.884942
4570	7	1-1140	4566	36	89	42	94	2012-08-02 04:50:25.884942
4571	7	1-1249	4566	89	45	62	20	2012-08-02 04:50:25.884942
4572	7	1-1300	4566	45	11	55	85	2012-08-02 04:50:25.884942
4566	7	1-1360	4566	11	39	90	65	2012-08-02 04:50:25.884942
15938	4	1-4164	15937	13	77	950	71	2012-08-02 05:13:18.234135
4574	8	1-1297	4573	49	59	1743	66	2012-08-02 04:50:25.884942
4575	8	1-1302	4573	59	7	1406	20	2012-08-02 04:50:25.884942
4576	8	1-598	4573	7	7	835	4	2012-08-02 04:50:25.884942
4577	8	1-1330	4573	7	95	816	30	2012-08-02 04:50:25.884942
4578	8	1-1277	4573	95	3	1106	59	2012-08-02 04:50:25.884942
4579	8	1-1234	4573	3	63	731	91	2012-08-02 04:50:25.884942
4580	8	1-915	4573	63	11	1078	85	2012-08-02 04:50:25.884942
4573	8	1-1360	4573	11	49	1958	65	2012-08-02 04:50:25.884942
15939	4	1-4171	15937	77	16	429	19	2012-08-02 05:13:18.234135
4582	6	1-874	4581	18	97	490	36	2012-08-02 04:50:28.723851
4583	6	1-1336	4581	97	2	398	85	2012-08-02 04:50:28.723851
4584	6	1-1287	4581	2	86	482	10	2012-08-02 04:50:28.723851
4585	6	1-764	4581	86	29	96	54	2012-08-02 04:50:28.723851
4586	6	1-1349	4581	29	59	67	56	2012-08-02 04:50:28.723851
4581	6	1-1366	4581	59	18	41	96	2012-08-02 04:50:28.723851
4588	8	1-874	4587	18	19	3267	36	2012-08-02 04:50:28.723851
4589	8	1-723	4587	19	18	3129	7	2012-08-02 04:50:28.723851
4590	8	1-533	4587	18	40	576	16	2012-08-02 04:50:28.723851
4591	8	1-1233	4587	40	44	631	82	2012-08-02 04:50:28.723851
4592	8	1-682	4587	44	75	554	8	2012-08-02 04:50:28.723851
4593	8	1-965	4587	75	29	602	54	2012-08-02 04:50:28.723851
4594	8	1-1349	4587	29	59	429	56	2012-08-02 04:50:28.723851
4587	8	1-1366	4587	59	18	267	96	2012-08-02 04:50:28.723851
4596	6	1-770	4595	14	40	707	16	2012-08-02 04:50:29.983671
4597	6	1-1233	4595	40	44	422	82	2012-08-02 04:50:29.983671
4598	6	1-682	4595	44	4	277	8	2012-08-02 04:50:29.983671
4599	6	1-4	4595	4	44	136	7	2012-08-02 04:50:29.983671
4600	6	1-1357	4595	44	35	343	22	2012-08-02 04:50:29.983671
4595	6	1-1367	4595	35	14	3044	99	2012-08-02 04:50:29.983671
4602	7	1-1349	4601	29	34	437	56	2012-08-02 04:50:30.447714
4603	7	1-1076	4601	34	97	289	36	2012-08-02 04:50:30.447714
4604	7	1-1336	4601	97	11	213	85	2012-08-02 04:50:30.447714
4605	7	1-1360	4601	11	39	340	65	2012-08-02 04:50:30.447714
4606	7	1-1305	4601	39	45	303	73	2012-08-02 04:50:30.447714
4607	7	1-296	4601	45	36	667	53	2012-08-02 04:50:30.447714
4601	7	1-1368	4601	36	29	688	54	2012-08-02 04:50:30.447714
4609	7	1-1349	4608	29	59	631	56	2012-08-02 04:50:30.447714
4610	7	1-1366	4608	59	68	398	96	2012-08-02 04:50:30.447714
4611	7	1-880	4608	68	36	705	5	2012-08-02 04:50:30.447714
4612	7	1-1140	4608	36	34	498	94	2012-08-02 04:50:30.447714
4613	7	1-962	4608	34	53	388	66	2012-08-02 04:50:30.447714
4614	7	1-1213	4608	53	36	737	53	2012-08-02 04:50:30.447714
4608	7	1-1368	4608	36	29	870	54	2012-08-02 04:50:30.447714
4616	6	1-1247	4615	45	11	2767	18	2012-08-02 04:50:30.922791
4617	6	1-1262	4615	11	50	2584	15	2012-08-02 04:50:30.922791
4618	6	1-589	4615	50	64	1021	32	2012-08-02 04:50:30.922791
4619	6	1-998	4615	64	64	1018	36	2012-08-02 04:50:30.922791
4620	6	1-1338	4615	64	28	1055	23	2012-08-02 04:50:30.922791
4615	6	1-1369	4615	28	45	2278	75	2012-08-02 04:50:30.922791
4622	7	1-777	4621	52	3	2413	59	2012-08-02 04:50:30.922791
4623	7	1-1234	4621	3	65	1511	91	2012-08-02 04:50:30.922791
4624	7	1-541	4621	65	70	1354	52	2012-08-02 04:50:30.922791
4625	7	1-978	4621	70	98	1438	31	2012-08-02 04:50:30.922791
4626	7	1-1355	4621	98	64	1550	36	2012-08-02 04:50:30.922791
4627	7	1-1338	4621	64	28	1564	23	2012-08-02 04:50:30.922791
4621	7	1-1369	4621	28	52	3426	75	2012-08-02 04:50:30.922791
4629	7	1-1247	4628	45	11	387	18	2012-08-02 04:50:30.922791
4630	7	1-1262	4628	11	32	369	15	2012-08-02 04:50:30.922791
4631	7	1-805	4628	32	25	187	83	2012-08-02 04:50:30.922791
4632	7	1-963	4628	25	87	274	45	2012-08-02 04:50:30.922791
4633	7	1-1149	4628	87	52	265	92	2012-08-02 04:50:30.922791
4634	7	1-1293	4628	52	28	141	23	2012-08-02 04:50:30.922791
4628	7	1-1369	4628	28	45	312	75	2012-08-02 04:50:30.922791
4636	7	1-777	4635	52	54	1347	59	2012-08-02 04:50:30.922791
4637	7	1-1255	4635	54	3	984	77	2012-08-02 04:50:30.922791
4638	7	1-1168	4635	3	70	646	52	2012-08-02 04:50:30.922791
4639	7	1-978	4635	70	98	708	31	2012-08-02 04:50:30.922791
4640	7	1-1355	4635	98	64	787	36	2012-08-02 04:50:30.922791
4641	7	1-1338	4635	64	28	820	23	2012-08-02 04:50:30.922791
4635	7	1-1369	4635	28	52	1853	75	2012-08-02 04:50:30.922791
4643	5	1-116	4642	64	25	107	83	2012-08-02 04:50:30.922791
4644	5	1-963	4642	25	87	165	45	2012-08-02 04:50:30.922791
4645	5	1-1149	4642	87	52	169	92	2012-08-02 04:50:30.922791
4646	5	1-1293	4642	52	28	95	23	2012-08-02 04:50:30.922791
4642	5	1-1369	4642	28	64	222	75	2012-08-02 04:50:30.922791
4648	7	1-891	4647	24	100	993	98	2012-08-02 04:50:31.894938
4649	7	1-934	4647	100	93	1676	7	2012-08-02 04:50:31.894938
4650	7	1-1239	4647	93	66	1631	41	2012-08-02 04:50:31.894938
4651	7	1-949	4647	66	44	465	52	2012-08-02 04:50:31.894938
4652	7	1-1356	4647	44	85	2170	13	2012-08-02 04:50:31.894938
4653	7	1-852	4647	85	12	1921	55	2012-08-02 04:50:31.894938
4647	7	1-1371	4647	12	24	1366	89	2012-08-02 04:50:31.894938
4655	7	1-1165	4654	85	7	1116	4	2012-08-02 04:50:31.894938
4656	7	1-1330	4654	7	95	1063	30	2012-08-02 04:50:31.894938
4657	7	1-1277	4654	95	54	1414	59	2012-08-02 04:50:31.894938
4658	7	1-1255	4654	54	59	1037	77	2012-08-02 04:50:31.894938
4659	7	1-853	4654	59	85	879	13	2012-08-02 04:50:31.894938
4660	7	1-852	4654	85	12	817	55	2012-08-02 04:50:31.894938
4654	7	1-1371	4654	12	85	600	89	2012-08-02 04:50:31.894938
4662	4	1-934	4661	100	44	457	7	2012-08-02 04:50:32.370327
4663	4	1-1357	4661	44	32	1254	22	2012-08-02 04:50:32.370327
4664	4	1-265	4661	32	97	301	34	2012-08-02 04:50:32.370327
4661	4	1-1372	4661	97	100	321	98	2012-08-02 04:50:32.370327
4666	6	1-1249	4665	89	14	63	20	2012-08-02 04:50:32.834168
4667	6	1-1325	4665	14	88	17	61	2012-08-02 04:50:32.834168
4668	6	1-868	4665	88	89	2	24	2012-08-02 04:50:32.834168
4669	6	1-1200	4665	89	77	3	48	2012-08-02 04:50:32.834168
4670	6	1-1212	4665	77	17	1	34	2012-08-02 04:50:32.834168
4665	6	1-1373	4665	17	89	73	94	2012-08-02 04:50:32.834168
4672	4	1-1098	4671	36	11	171	84	2012-08-02 04:50:32.834168
4673	4	1-1047	4671	11	66	41	3	2012-08-02 04:50:32.834168
4674	4	1-630	4671	66	17	8	34	2012-08-02 04:50:32.834168
4671	4	1-1373	4671	17	36	1204	94	2012-08-02 04:50:32.834168
15940	4	1-4230	15937	16	7	703	3	2012-08-02 05:13:18.234135
4676	3	1-559	4675	52	51	66	83	2012-08-02 04:50:32.834168
4677	3	1-89	4675	51	17	27	34	2012-08-02 04:50:32.834168
4675	3	1-1373	4675	17	52	4637	94	2012-08-02 04:50:32.834168
15937	4	1-4591	15937	7	13	2688	54	2012-08-02 05:13:18.234135
4679	5	1-872	4678	92	59	167	56	2012-08-02 04:50:34.259509
4680	5	1-1366	4678	59	18	76	96	2012-08-02 04:50:34.259509
4681	5	1-874	4678	18	56	664	36	2012-08-02 04:50:34.259509
4682	5	1-273	4678	56	78	358	72	2012-08-02 04:50:34.259509
4678	5	1-1376	4678	78	92	323	1	2012-08-02 04:50:34.259509
4684	6	1-872	4683	92	59	3188	56	2012-08-02 04:50:34.259509
4685	6	1-1366	4683	59	31	1481	96	2012-08-02 04:50:34.259509
4686	6	1-1306	4683	31	99	2564	64	2012-08-02 04:50:34.259509
4687	6	1-1283	4683	99	33	4039	100	2012-08-02 04:50:34.259509
4688	6	1-1361	4683	33	78	6431	72	2012-08-02 04:50:34.259509
4683	6	1-1376	4683	78	92	5964	1	2012-08-02 04:50:34.259509
4690	6	1-1284	4689	31	34	867	47	2012-08-02 04:50:34.259509
4691	6	1-740	4689	34	97	808	34	2012-08-02 04:50:34.259509
4692	6	1-1372	4689	97	25	885	98	2012-08-02 04:50:34.259509
4693	6	1-1221	4689	25	9	230	24	2012-08-02 04:50:34.259509
4694	6	1-435	4689	9	78	514	72	2012-08-02 04:50:34.259509
4689	6	1-1376	4689	78	31	532	1	2012-08-02 04:50:34.259509
4696	6	1-531	4695	79	7	2574	4	2012-08-02 04:50:34.259509
4697	6	1-1330	4695	7	75	2099	30	2012-08-02 04:50:34.259509
4698	6	1-438	4695	75	28	570	11	2012-08-02 04:50:34.259509
4699	6	1-30	4695	28	86	413	43	2012-08-02 04:50:34.259509
4700	6	1-877	4695	86	78	1383	72	2012-08-02 04:50:34.259509
4695	6	1-1376	4695	78	79	1459	1	2012-08-02 04:50:34.259509
4702	6	1-1284	4701	31	34	107	47	2012-08-02 04:50:34.259509
4703	6	1-740	4701	34	97	102	34	2012-08-02 04:50:34.259509
4704	6	1-1372	4701	97	68	114	98	2012-08-02 04:50:34.259509
4705	6	1-866	4701	68	33	72	81	2012-08-02 04:50:34.259509
4706	6	1-844	4701	33	78	61	72	2012-08-02 04:50:34.259509
4701	6	1-1376	4701	78	31	64	1	2012-08-02 04:50:34.259509
4708	2	1-1273	4707	16	67	4493	42	2012-08-02 04:50:35.737094
4707	2	1-1379	4707	67	16	4686	92	2012-08-02 04:50:35.737094
4710	5	1-1342	4709	9	83	1610	71	2012-08-02 04:50:35.737094
4711	5	1-642	4709	83	97	812	34	2012-08-02 04:50:35.737094
4712	5	1-1372	4709	97	51	1030	98	2012-08-02 04:50:35.737094
4713	5	1-1354	4709	51	67	866	42	2012-08-02 04:50:35.737094
4709	5	1-1379	4709	67	9	1155	92	2012-08-02 04:50:35.737094
4715	8	1-1360	4714	11	34	283	65	2012-08-02 04:50:36.205914
4716	8	1-898	4714	34	78	253	43	2012-08-02 04:50:36.205914
4717	8	1-1108	4714	78	85	267	21	2012-08-02 04:50:36.205914
4718	8	1-1037	4714	85	77	154	48	2012-08-02 04:50:36.205914
4719	8	1-1212	4714	77	97	76	34	2012-08-02 04:50:36.205914
4720	8	1-1372	4714	97	100	92	98	2012-08-02 04:50:36.205914
4721	8	1-934	4714	100	16	148	7	2012-08-02 04:50:36.205914
4714	8	1-1380	4714	16	11	171	85	2012-08-02 04:50:36.205914
4722	1	1-934	\N	100	100	1301	7	2012-08-02 04:50:36.205914
4724	6	1-624	4723	37	78	494	43	2012-08-02 04:50:36.681123
4725	6	1-1108	4723	78	34	534	21	2012-08-02 04:50:36.681123
4726	6	1-1228	4723	34	26	1612	51	2012-08-02 04:50:36.681123
4727	6	1-597	4723	26	39	902	68	2012-08-02 04:50:36.681123
4728	6	1-1346	4723	39	6	635	81	2012-08-02 04:50:36.681123
4723	6	1-1381	4723	6	37	1188	45	2012-08-02 04:50:36.681123
4730	6	1-1149	4729	87	52	273	92	2012-08-02 04:50:36.681123
4731	6	1-1293	4729	52	72	143	23	2012-08-02 04:50:36.681123
4732	6	1-752	4729	72	96	119	61	2012-08-02 04:50:36.681123
4733	6	1-1250	4729	96	66	504	6	2012-08-02 04:50:36.681123
4734	6	1-766	4729	66	6	153	81	2012-08-02 04:50:36.681123
4729	6	1-1381	4729	6	87	287	45	2012-08-02 04:50:36.681123
4736	5	1-624	4735	37	78	1305	43	2012-08-02 04:50:36.681123
4737	5	1-1108	4735	78	13	1431	21	2012-08-02 04:50:36.681123
4738	5	1-1358	4735	13	94	2558	70	2012-08-02 04:50:36.681123
4739	5	1-549	4735	94	6	1632	81	2012-08-02 04:50:36.681123
4735	5	1-1381	4735	6	37	3095	45	2012-08-02 04:50:36.681123
4741	6	1-989	4740	24	39	235	68	2012-08-02 04:50:37.145233
4742	6	1-1346	4740	39	6	165	81	2012-08-02 04:50:37.145233
4743	6	1-1381	4740	6	37	280	45	2012-08-02 04:50:37.145233
4744	6	1-624	4740	37	78	106	43	2012-08-02 04:50:37.145233
4745	6	1-1108	4740	78	39	104	21	2012-08-02 04:50:37.145233
4740	6	1-1382	4740	39	24	128	90	2012-08-02 04:50:37.145233
4747	5	1-1368	4746	36	29	2851	54	2012-08-02 04:50:37.609539
4748	5	1-1349	4746	29	59	2026	56	2012-08-02 04:50:37.609539
4749	5	1-1366	4746	59	24	1254	96	2012-08-02 04:50:37.609539
4750	5	1-1331	4746	24	76	1675	63	2012-08-02 04:50:37.609539
4746	5	1-1383	4746	76	36	2474	53	2012-08-02 04:50:37.609539
4752	6	1-669	4751	47	29	591	88	2012-08-02 04:50:42.447879
4753	6	1-1296	4751	29	98	46	31	2012-08-02 04:50:42.447879
4754	6	1-1355	4751	98	97	46	36	2012-08-02 04:50:42.447879
4755	6	1-1336	4751	97	12	34	85	2012-08-02 04:50:42.447879
4756	6	1-601	4751	12	51	21	80	2012-08-02 04:50:42.447879
4751	6	1-1395	4751	51	47	726	79	2012-08-02 04:50:42.447879
4758	3	1-1352	4757	28	45	2029	5	2012-08-02 04:50:48.600754
4759	3	1-432	4757	45	16	190	12	2012-08-02 04:50:48.600754
4757	3	1-1406	4757	16	28	6954	22	2012-08-02 04:50:48.600754
4761	7	1-1352	4760	28	36	1414	5	2012-08-02 04:50:48.600754
4762	7	1-1140	4760	36	31	799	94	2012-08-02 04:50:48.600754
4763	7	1-810	4760	31	83	440	33	2012-08-02 04:50:48.600754
4764	7	1-729	4760	83	69	322	9	2012-08-02 04:50:48.600754
4765	7	1-1029	4760	69	1	73	25	2012-08-02 04:50:48.600754
4766	7	1-1333	4760	1	16	28	12	2012-08-02 04:50:48.600754
4760	7	1-1406	4760	16	28	2235	22	2012-08-02 04:50:48.600754
4768	7	1-1352	4767	28	21	58	5	2012-08-02 04:50:48.600754
4769	7	1-141	4767	21	14	15	20	2012-08-02 04:50:48.600754
4770	7	1-1325	4767	14	72	14	61	2012-08-02 04:50:48.600754
4771	7	1-364	4767	72	3	8	55	2012-08-02 04:50:48.600754
4772	7	1-1023	4767	3	27	1	82	2012-08-02 04:50:48.600754
4773	7	1-1010	4767	27	16	1	12	2012-08-02 04:50:48.600754
4767	7	1-1406	4767	16	28	82	22	2012-08-02 04:50:48.600754
15942	4	1-4164	15941	13	77	1789	71	2012-08-02 05:13:18.234135
4775	6	1-1057	4774	77	97	783	34	2012-08-02 04:50:50.512175
4776	6	1-1372	4774	97	68	1040	98	2012-08-02 04:50:50.512175
4777	6	1-866	4774	68	6	783	81	2012-08-02 04:50:50.512175
4778	6	1-1381	4774	6	67	1573	45	2012-08-02 04:50:50.512175
4779	6	1-1401	4774	67	65	935	94	2012-08-02 04:50:50.512175
4774	6	1-1410	4774	65	77	194	57	2012-08-02 04:50:50.512175
15943	4	1-4171	15941	77	56	815	19	2012-08-02 05:13:18.234135
4781	4	1-1077	4780	18	67	2369	42	2012-08-02 04:50:51.462164
4782	4	1-1379	4780	67	54	1984	92	2012-08-02 04:50:51.462164
4783	4	1-1263	4780	54	73	2235	100	2012-08-02 04:50:51.462164
4780	4	1-1412	4780	73	18	1188	86	2012-08-02 04:50:51.462164
15944	4	1-4407	15941	56	7	1305	3	2012-08-02 05:13:18.234135
4785	7	1-1065	4784	16	83	2610	71	2012-08-02 04:50:51.462164
4786	7	1-642	4784	83	97	1110	34	2012-08-02 04:50:51.462164
4787	7	1-1372	4784	97	51	1187	98	2012-08-02 04:50:51.462164
4788	7	1-1354	4784	51	67	841	42	2012-08-02 04:50:51.462164
4789	7	1-1379	4784	67	54	946	92	2012-08-02 04:50:51.462164
4790	7	1-1263	4784	54	73	1431	100	2012-08-02 04:50:51.462164
4784	7	1-1412	4784	73	16	1022	86	2012-08-02 04:50:51.462164
4792	6	1-1065	4791	16	83	1752	71	2012-08-02 04:50:51.462164
4793	6	1-642	4791	83	97	881	34	2012-08-02 04:50:51.462164
4794	6	1-1372	4791	97	6	1115	98	2012-08-02 04:50:51.462164
4795	6	1-1291	4791	6	54	384	92	2012-08-02 04:50:51.462164
4796	6	1-1263	4791	54	73	687	100	2012-08-02 04:50:51.462164
4791	6	1-1412	4791	73	16	580	86	2012-08-02 04:50:51.462164
4798	5	1-1357	4797	44	28	4569	22	2012-08-02 04:50:52.412428
4799	5	1-1352	4797	28	21	3051	5	2012-08-02 04:50:52.412428
4800	5	1-141	4797	21	14	728	20	2012-08-02 04:50:52.412428
4801	5	1-1325	4797	14	99	629	61	2012-08-02 04:50:52.412428
4797	5	1-1414	4797	99	44	1570	7	2012-08-02 04:50:52.412428
4803	4	1-1007	4802	24	5	2501	60	2012-08-02 04:50:54.798713
4804	4	1-788	4802	5	78	922	43	2012-08-02 04:50:54.798713
4805	4	1-1108	4802	78	49	429	21	2012-08-02 04:50:54.798713
4802	4	1-1418	4802	49	24	7516	86	2012-08-02 04:50:54.798713
4807	4	1-1122	4806	91	74	187	67	2012-08-02 04:50:54.798713
4808	4	1-436	4806	74	78	64	43	2012-08-02 04:50:54.798713
4809	4	1-1108	4806	78	49	32	21	2012-08-02 04:50:54.798713
4806	4	1-1418	4806	49	91	593	86	2012-08-02 04:50:54.798713
4810	1	1-1108	\N	78	78	4962	21	2012-08-02 04:50:54.798713
4812	5	1-748	4811	63	99	4396	48	2012-08-02 04:50:55.284861
4813	5	1-1377	4811	99	74	684	2	2012-08-02 04:50:55.284861
4814	5	1-1370	4811	74	22	333	43	2012-08-02 04:50:55.284861
4815	5	1-1246	4811	22	46	190	70	2012-08-02 04:50:55.284861
4811	5	1-1419	4811	46	63	4475	46	2012-08-02 04:50:55.284861
4817	6	1-1080	4816	77	72	2355	68	2012-08-02 04:50:55.284861
4818	6	1-1310	4816	72	61	806	77	2012-08-02 04:50:55.284861
4819	6	1-504	4816	61	45	206	75	2012-08-02 04:50:55.284861
4820	6	1-1247	4816	45	97	224	18	2012-08-02 04:50:55.284861
4821	6	1-1088	4816	97	46	203	70	2012-08-02 04:50:55.284861
4816	6	1-1419	4816	46	77	5415	46	2012-08-02 04:50:55.284861
4823	5	1-998	4822	64	42	2766	36	2012-08-02 04:50:56.235258
4824	5	1-860	4822	42	97	1712	34	2012-08-02 04:50:56.235258
4825	5	1-1372	4822	97	54	2035	98	2012-08-02 04:50:56.235258
4826	5	1-631	4822	54	66	1367	22	2012-08-02 04:50:56.235258
4822	5	1-1421	4822	66	64	3014	32	2012-08-02 04:50:56.235258
4828	7	1-907	4827	96	74	488	76	2012-08-02 04:50:56.235258
4829	7	1-414	4827	74	2	173	27	2012-08-02 04:50:56.235258
4830	7	1-189	4827	2	34	88	47	2012-08-02 04:50:56.235258
4831	7	1-740	4827	34	97	94	34	2012-08-02 04:50:56.235258
4832	7	1-1372	4827	97	54	117	98	2012-08-02 04:50:56.235258
4833	7	1-631	4827	54	66	83	22	2012-08-02 04:50:56.235258
4827	7	1-1421	4827	66	96	192	32	2012-08-02 04:50:56.235258
4834	1	1-1372	\N	97	97	1706	98	2012-08-02 04:50:56.235258
4836	6	1-989	4835	24	72	1287	68	2012-08-02 04:50:59.825108
4837	6	1-1310	4835	72	3	500	77	2012-08-02 04:50:59.825108
4838	6	1-1168	4835	3	44	318	52	2012-08-02 04:50:59.825108
4839	6	1-1356	4835	44	85	1481	13	2012-08-02 04:50:59.825108
4840	6	1-852	4835	85	51	1307	55	2012-08-02 04:50:59.825108
4835	6	1-1428	4835	51	24	627	90	2012-08-02 04:50:59.825108
4842	5	1-1059	4841	91	22	845	29	2012-08-02 04:51:00.300434
4843	5	1-453	4841	22	51	290	93	2012-08-02 04:51:00.300434
4844	5	1-1423	4841	51	51	231	80	2012-08-02 04:51:00.300434
4845	5	1-1395	4841	51	75	5638	79	2012-08-02 04:51:00.300434
4841	5	1-1429	4841	75	91	1187	16	2012-08-02 04:51:00.300434
4847	6	1-1388	4846	100	32	444	81	2012-08-02 04:51:01.228622
4848	6	1-1184	4846	32	72	338	68	2012-08-02 04:51:01.228622
4849	6	1-1310	4846	72	3	107	77	2012-08-02 04:51:01.228622
4850	6	1-1168	4846	3	44	55	52	2012-08-02 04:51:01.228622
4851	6	1-1356	4846	44	80	207	13	2012-08-02 04:51:01.228622
4846	6	1-1431	4846	80	100	609	57	2012-08-02 04:51:01.228622
4853	6	1-1263	4852	54	73	2115	100	2012-08-02 04:51:01.71434
4854	6	1-1412	4852	73	24	1027	86	2012-08-02 04:51:01.71434
4855	6	1-1007	4852	24	69	467	60	2012-08-02 04:51:01.71434
4856	6	1-1320	4852	69	26	276	32	2012-08-02 04:51:01.71434
4857	6	1-840	4852	26	45	285	87	2012-08-02 04:51:01.71434
4852	6	1-1432	4852	45	54	2056	92	2012-08-02 04:51:01.71434
4859	7	1-1293	4858	52	72	1638	23	2012-08-02 04:51:01.71434
4860	7	1-752	4858	72	88	1242	61	2012-08-02 04:51:01.71434
4861	7	1-194	4858	88	72	308	65	2012-08-02 04:51:01.71434
4862	7	1-1062	4858	72	66	177	22	2012-08-02 04:51:01.71434
4863	7	1-1421	4858	66	26	314	32	2012-08-02 04:51:01.71434
4864	7	1-840	4858	26	45	427	87	2012-08-02 04:51:01.71434
4858	7	1-1432	4858	45	52	4054	92	2012-08-02 04:51:01.71434
4866	7	1-1293	4865	52	93	1336	23	2012-08-02 04:51:01.71434
4867	7	1-803	4865	93	24	279	89	2012-08-02 04:51:01.71434
4868	7	1-891	4865	24	54	169	98	2012-08-02 04:51:01.71434
4869	7	1-631	4865	54	66	100	22	2012-08-02 04:51:01.71434
4870	7	1-1421	4865	66	26	195	32	2012-08-02 04:51:01.71434
4871	7	1-840	4865	26	45	291	87	2012-08-02 04:51:01.71434
4865	7	1-1432	4865	45	52	3020	92	2012-08-02 04:51:01.71434
4872	1	1-891	\N	24	24	1023	98	2012-08-02 04:51:01.71434
4874	6	1-1284	4873	31	34	145	47	2012-08-02 04:51:04.067541
4875	6	1-740	4873	34	4	157	34	2012-08-02 04:51:04.067541
4876	6	1-1177	4873	4	84	239	50	2012-08-02 04:51:04.067541
4873	6	1-1437	4873	87	31	76	1	2012-08-02 04:51:04.067541
4877	6	1-715	4873	84	41	51	25	2012-08-02 04:51:04.067541
4878	6	1-1058	4873	41	87	33	55	2012-08-02 04:51:04.067541
4879	1	1-1177	\N	4	4	1564	50	2012-08-02 04:51:04.067541
15941	4	1-4591	15941	7	13	5026	54	2012-08-02 05:13:18.234135
4881	6	1-1327	4880	42	82	923	83	2012-08-02 04:51:04.542873
4882	6	1-249	4880	82	97	235	84	2012-08-02 04:51:04.542873
4883	6	1-1343	4880	97	6	169	67	2012-08-02 04:51:04.542873
4884	6	1-1363	4880	6	3	93	38	2012-08-02 04:51:04.542873
4885	6	1-733	4880	3	12	164	44	2012-08-02 04:51:04.542873
4880	6	1-1438	4880	12	42	371	89	2012-08-02 04:51:04.542873
4887	7	1-143	4886	21	24	1762	26	2012-08-02 04:51:10.831582
4888	7	1-971	4886	24	48	1802	11	2012-08-02 04:51:10.831582
4889	7	1-1403	4886	48	46	4340	58	2012-08-02 04:51:10.831582
4890	7	1-1132	4886	46	41	3573	25	2012-08-02 04:51:10.831582
4891	7	1-1058	4886	41	87	1558	55	2012-08-02 04:51:10.831582
4892	7	1-1437	4886	87	3	2419	1	2012-08-02 04:51:10.831582
4886	7	1-1447	4886	3	21	5121	96	2012-08-02 04:51:10.831582
4894	6	1-1331	4893	24	76	1443	63	2012-08-02 04:51:10.831582
4895	6	1-1383	4893	76	2	1717	53	2012-08-02 04:51:10.831582
4896	6	1-284	4893	2	16	212	97	2012-08-02 04:51:10.831582
4897	6	1-1378	4893	16	87	271	55	2012-08-02 04:51:10.831582
4898	6	1-1437	4893	87	3	516	1	2012-08-02 04:51:10.831582
4893	6	1-1447	4893	3	24	1341	96	2012-08-02 04:51:10.831582
4900	7	1-1331	4899	24	24	460	63	2012-08-02 04:51:10.831582
4901	7	1-916	4899	24	77	346	92	2012-08-02 04:51:10.831582
4902	7	1-1220	4899	77	32	254	29	2012-08-02 04:51:10.831582
4903	7	1-1016	4899	32	16	47	97	2012-08-02 04:51:10.831582
4904	7	1-1378	4899	16	87	64	55	2012-08-02 04:51:10.831582
4905	7	1-1437	4899	87	3	131	1	2012-08-02 04:51:10.831582
4899	7	1-1447	4899	3	24	365	96	2012-08-02 04:51:10.831582
4907	6	1-880	4906	68	21	490	5	2012-08-02 04:51:10.831582
4908	6	1-141	4906	21	14	120	20	2012-08-02 04:51:10.831582
4909	6	1-1325	4906	14	72	106	61	2012-08-02 04:51:10.831582
4910	6	1-364	4906	72	87	53	55	2012-08-02 04:51:10.831582
4911	6	1-1437	4906	87	3	111	1	2012-08-02 04:51:10.831582
4906	6	1-1447	4906	3	68	318	96	2012-08-02 04:51:10.831582
4913	6	1-1331	4912	24	36	1547	63	2012-08-02 04:51:10.831582
4914	6	1-1186	4912	36	84	1658	50	2012-08-02 04:51:10.831582
4915	6	1-715	4912	84	41	330	25	2012-08-02 04:51:10.831582
4916	6	1-1058	4912	41	87	200	55	2012-08-02 04:51:10.831582
4917	6	1-1437	4912	87	3	432	1	2012-08-02 04:51:10.831582
4912	6	1-1447	4912	3	24	1269	96	2012-08-02 04:51:10.831582
4918	1	1-1325	\N	14	14	3684	61	2012-08-02 04:51:10.831582
4920	7	1-1312	4919	28	14	218	99	2012-08-02 04:51:15.338593
4921	7	1-770	4919	14	40	77	16	2012-08-02 04:51:15.338593
4922	7	1-1233	4919	40	11	78	82	2012-08-02 04:51:15.338593
4923	7	1-391	4919	11	17	36	19	2012-08-02 04:51:15.338593
4924	7	1-992	4919	17	40	59	39	2012-08-02 04:51:15.338593
4925	7	1-332	4919	40	78	60	79	2012-08-02 04:51:15.338593
4919	7	1-1455	4919	78	28	238	3	2012-08-02 04:51:15.338593
4927	8	1-1312	4926	28	66	1136	99	2012-08-02 04:51:15.338593
4928	8	1-288	4926	66	60	207	75	2012-08-02 04:51:15.338593
4929	8	1-920	4926	60	100	52	91	2012-08-02 04:51:15.338593
4930	8	1-638	4926	100	66	67	44	2012-08-02 04:51:15.338593
4931	8	1-1416	4926	66	17	111	19	2012-08-02 04:51:15.338593
4932	8	1-992	4926	17	40	212	39	2012-08-02 04:51:15.338593
4933	8	1-332	4926	40	78	228	79	2012-08-02 04:51:15.338593
4926	8	1-1455	4926	78	28	1057	3	2012-08-02 04:51:15.338593
4935	5	1-1312	4934	28	45	826	99	2012-08-02 04:51:15.338593
4936	5	1-1402	4934	45	36	1123	53	2012-08-02 04:51:15.338593
4937	5	1-1368	4934	36	33	1265	54	2012-08-02 04:51:15.338593
4938	5	1-922	4934	33	78	164	79	2012-08-02 04:51:15.338593
4934	5	1-1455	4934	78	28	766	3	2012-08-02 04:51:15.338593
4940	5	1-646	4939	2	45	760	37	2012-08-02 04:51:15.824937
4941	5	1-681	4939	45	89	694	72	2012-08-02 04:51:15.824937
4942	5	1-1236	4939	89	84	555	14	2012-08-02 04:51:15.824937
4943	5	1-946	4939	84	15	164	93	2012-08-02 04:51:15.824937
4939	5	1-1456	4939	15	2	277	85	2012-08-02 04:51:15.824937
4945	6	1-1076	4944	34	51	1179	36	2012-08-02 04:51:16.774946
4946	6	1-1060	4944	51	24	64	26	2012-08-02 04:51:16.774946
4947	6	1-971	4944	24	48	61	11	2012-08-02 04:51:16.774946
4948	6	1-1403	4944	48	98	135	58	2012-08-02 04:51:16.774946
4949	6	1-1420	4944	98	86	72	98	2012-08-02 04:51:16.774946
4944	6	1-1458	4944	86	34	2642	56	2012-08-02 04:51:16.774946
4951	8	1-1348	4950	85	77	63	46	2012-08-02 04:51:16.774946
4952	8	1-1080	4950	77	72	23	68	2012-08-02 04:51:16.774946
4953	8	1-1310	4950	72	22	7	77	2012-08-02 04:51:16.774946
4954	8	1-322	4950	22	24	2	26	2012-08-02 04:51:16.774946
4955	8	1-971	4950	24	48	2	11	2012-08-02 04:51:16.774946
4956	8	1-1403	4950	48	98	6	58	2012-08-02 04:51:16.774946
4957	8	1-1420	4950	98	86	4	98	2012-08-02 04:51:16.774946
4950	8	1-1458	4950	86	85	160	56	2012-08-02 04:51:16.774946
4959	6	1-1366	4958	59	24	20	96	2012-08-02 04:51:16.774946
4960	6	1-1331	4958	24	8	19	63	2012-08-02 04:51:16.774946
4961	6	1-1113	4958	8	66	3	38	2012-08-02 04:51:16.774946
4962	6	1-1117	4958	66	98	2	58	2012-08-02 04:51:16.774946
4963	6	1-1420	4958	98	86	1	98	2012-08-02 04:51:16.774946
4958	6	1-1458	4958	86	59	44	56	2012-08-02 04:51:16.774946
4965	6	1-685	4964	15	36	1050	23	2012-08-02 04:51:16.774946
4966	6	1-1364	4964	36	35	56	19	2012-08-02 04:51:16.774946
4967	6	1-941	4964	35	48	24	11	2012-08-02 04:51:16.774946
4968	6	1-1403	4964	48	98	65	58	2012-08-02 04:51:16.774946
4969	6	1-1420	4964	98	86	43	98	2012-08-02 04:51:16.774946
4964	6	1-1458	4964	86	15	1939	56	2012-08-02 04:51:16.774946
4971	4	1-1058	4970	41	87	553	55	2012-08-02 04:51:17.283399
4972	4	1-1437	4970	87	79	1198	1	2012-08-02 04:51:17.283399
4973	4	1-531	4970	79	66	2268	4	2012-08-02 04:51:17.283399
4970	4	1-1459	4970	66	41	908	25	2012-08-02 04:51:17.283399
4975	3	1-1058	4974	41	66	99	55	2012-08-02 04:51:17.283399
4976	3	1-1433	4974	66	66	442	4	2012-08-02 04:51:17.283399
4974	3	1-1459	4974	66	41	179	25	2012-08-02 04:51:17.283399
4978	3	1-1356	4977	44	80	4189	13	2012-08-02 04:51:22.696022
4979	3	1-1431	4977	80	31	4455	57	2012-08-02 04:51:22.696022
4977	3	1-1468	4977	31	44	3067	52	2012-08-02 04:51:22.696022
4981	6	1-1410	4980	65	31	70	57	2012-08-02 04:51:23.679401
4980	6	1-1470	4980	38	65	475	94	2012-08-02 04:51:23.679401
4982	6	1-1468	4980	31	44	124	52	2012-08-02 04:51:23.679401
4983	6	1-1356	4980	44	85	435	13	2012-08-02 04:51:23.679401
4984	6	1-1407	4980	85	66	273	4	2012-08-02 04:51:23.679401
4985	6	1-1459	4980	66	38	86	25	2012-08-02 04:51:23.679401
4987	4	1-810	4986	31	83	404	33	2012-08-02 04:51:23.679401
4988	4	1-729	4986	83	69	340	9	2012-08-02 04:51:23.679401
4989	4	1-1029	4986	69	38	89	25	2012-08-02 04:51:23.679401
4986	4	1-1470	4986	38	31	637	94	2012-08-02 04:51:23.679401
15946	4	1-4561	15945	60	19	5182	45	2012-08-02 05:13:19.449411
4991	6	1-1410	4990	65	31	748	57	2012-08-02 04:51:23.679401
4992	6	1-1468	4990	31	64	1822	52	2012-08-02 04:51:23.679401
4993	6	1-1393	4990	64	8	2525	46	2012-08-02 04:51:23.679401
4994	6	1-1026	4990	8	13	3984	62	2012-08-02 04:51:23.679401
4995	6	1-1110	4990	13	38	485	25	2012-08-02 04:51:23.679401
4990	6	1-1470	4990	38	65	3675	94	2012-08-02 04:51:23.679401
4997	7	1-1005	4996	44	39	402	75	2012-08-02 04:51:24.176605
4998	7	1-1415	4996	39	69	731	2	2012-08-02 04:51:24.176605
4999	7	1-1365	4996	69	70	374	73	2012-08-02 04:51:24.176605
5000	7	1-441	4996	70	66	457	22	2012-08-02 04:51:24.176605
5001	7	1-1421	4996	66	26	1065	32	2012-08-02 04:51:24.176605
5002	7	1-840	4996	26	54	1900	87	2012-08-02 04:51:24.176605
4996	7	1-1471	4996	54	44	624	82	2012-08-02 04:51:24.176605
5004	7	1-1328	5003	17	66	537	22	2012-08-02 04:51:25.126746
5005	7	1-1421	5003	66	26	997	32	2012-08-02 04:51:25.126746
5006	7	1-840	5003	26	54	1417	87	2012-08-02 04:51:25.126746
5007	7	1-1471	5003	54	39	371	82	2012-08-02 04:51:25.126746
5008	7	1-574	5003	39	46	222	29	2012-08-02 04:51:25.126746
5009	7	1-356	5003	46	75	460	78	2012-08-02 04:51:25.126746
5003	7	1-1473	5003	75	17	754	56	2012-08-02 04:51:25.126746
5011	6	1-724	5010	34	73	712	100	2012-08-02 04:51:25.126746
5012	6	1-1412	5010	73	91	580	86	2012-08-02 04:51:25.126746
5013	6	1-1122	5010	91	47	393	67	2012-08-02 04:51:25.126746
5014	6	1-1298	5010	47	46	285	29	2012-08-02 04:51:25.126746
5015	6	1-356	5010	46	75	722	78	2012-08-02 04:51:25.126746
5010	6	1-1473	5010	75	34	1441	56	2012-08-02 04:51:25.126746
5017	7	1-631	5016	54	66	887	22	2012-08-02 04:51:25.56895
5018	7	1-1421	5016	66	96	1802	32	2012-08-02 04:51:25.56895
5019	7	1-907	5016	96	53	4023	76	2012-08-02 04:51:25.56895
5020	7	1-763	5016	53	2	3210	9	2012-08-02 04:51:25.56895
5021	7	1-1335	5016	2	71	235	13	2012-08-02 04:51:25.56895
5022	7	1-1267	5016	71	33	167	15	2012-08-02 04:51:25.56895
5016	7	1-1474	5016	33	54	1432	98	2012-08-02 04:51:25.56895
5024	7	1-1292	5023	99	78	2767	47	2012-08-02 04:51:25.56895
5025	7	1-150	5023	78	75	2064	78	2012-08-02 04:51:25.56895
5026	7	1-1473	5023	75	34	3707	56	2012-08-02 04:51:25.56895
5027	7	1-724	5023	34	73	1648	100	2012-08-02 04:51:25.56895
5028	7	1-1412	5023	73	55	1208	86	2012-08-02 04:51:25.56895
5029	7	1-1235	5023	55	33	647	15	2012-08-02 04:51:25.56895
5023	7	1-1474	5023	33	99	5553	98	2012-08-02 04:51:25.56895
5031	4	1-1221	5030	25	92	422	24	2012-08-02 04:51:25.56895
5032	4	1-1323	5030	92	71	237	13	2012-08-02 04:51:25.56895
5033	4	1-1267	5030	71	33	175	15	2012-08-02 04:51:25.56895
5030	4	1-1474	5030	33	25	1562	98	2012-08-02 04:51:25.56895
5035	7	1-1014	5034	88	16	552	97	2012-08-02 04:51:26.055056
5036	7	1-1378	5034	16	66	491	55	2012-08-02 04:51:26.055056
5037	7	1-1433	5034	66	66	1792	4	2012-08-02 04:51:26.055056
5038	7	1-1459	5034	66	38	594	25	2012-08-02 04:51:26.055056
5039	7	1-1470	5034	38	19	2520	94	2012-08-02 04:51:26.055056
5040	7	1-1387	5034	19	78	2470	6	2012-08-02 04:51:26.055056
5034	7	1-1475	5034	78	88	2123	60	2012-08-02 04:51:26.055056
5042	4	1-1320	5041	69	96	1501	32	2012-08-02 04:51:26.055056
5043	4	1-907	5041	96	51	2685	76	2012-08-02 04:51:26.055056
5044	4	1-1052	5041	51	78	2032	6	2012-08-02 04:51:26.055056
5041	4	1-1475	5041	78	69	2099	60	2012-08-02 04:51:26.055056
5046	3	1-621	5045	21	51	2631	76	2012-08-02 04:51:26.055056
5047	3	1-1052	5045	51	78	2241	6	2012-08-02 04:51:26.055056
5045	3	1-1475	5045	78	21	2607	60	2012-08-02 04:51:26.055056
5049	5	1-1122	5048	91	52	404	67	2012-08-02 04:51:26.530111
5050	5	1-73	5048	52	38	192	25	2012-08-02 04:51:26.530111
5051	5	1-1470	5048	38	4	1239	94	2012-08-02 04:51:26.530111
5052	5	1-679	5048	4	70	1232	87	2012-08-02 04:51:26.530111
5048	5	1-1476	5048	70	91	656	86	2012-08-02 04:51:26.530111
5054	4	1-1127	5053	79	24	144	86	2012-08-02 04:51:26.994197
5055	4	1-1007	5053	24	83	89	60	2012-08-02 04:51:26.994197
5056	4	1-933	5053	83	33	63	2	2012-08-02 04:51:26.994197
5053	4	1-1477	5053	33	79	409	95	2012-08-02 04:51:26.994197
5058	4	1-1031	5057	56	8	5578	33	2012-08-02 04:51:26.994197
5059	4	1-1467	5057	8	21	1292	66	2012-08-02 04:51:26.994197
5060	4	1-1178	5057	21	33	947	2	2012-08-02 04:51:26.994197
5057	4	1-1477	5057	33	56	6431	95	2012-08-02 04:51:26.994197
5062	5	1-1127	5061	79	91	1170	86	2012-08-02 04:51:26.994197
5063	5	1-1122	5061	91	20	687	67	2012-08-02 04:51:26.994197
5064	5	1-1123	5061	20	83	587	60	2012-08-02 04:51:26.994197
5065	5	1-933	5061	83	33	448	2	2012-08-02 04:51:26.994197
5061	5	1-1477	5061	33	79	3101	95	2012-08-02 04:51:26.994197
5067	7	1-1343	5066	97	20	1333	67	2012-08-02 04:51:31.423723
5068	7	1-1123	5066	20	88	1156	60	2012-08-02 04:51:31.423723
5069	7	1-1014	5066	88	16	443	97	2012-08-02 04:51:31.423723
5070	7	1-1378	5066	16	87	578	55	2012-08-02 04:51:31.423723
5071	7	1-1437	5066	87	36	1124	1	2012-08-02 04:51:31.423723
5072	7	1-603	5066	36	39	1929	27	2012-08-02 04:51:31.423723
5066	7	1-1484	5066	39	97	2300	84	2012-08-02 04:51:31.423723
5074	6	1-1343	5073	97	6	904	67	2012-08-02 04:51:31.423723
5075	6	1-1363	5073	6	3	433	38	2012-08-02 04:51:31.423723
5076	6	1-733	5073	3	12	669	44	2012-08-02 04:51:31.423723
5077	6	1-1438	5073	12	70	1323	89	2012-08-02 04:51:31.423723
5078	6	1-873	5073	70	39	1111	27	2012-08-02 04:51:31.423723
5073	6	1-1484	5073	39	97	1437	84	2012-08-02 04:51:31.423723
5080	8	1-1467	5079	8	99	5	66	2012-08-02 04:51:31.910011
5081	8	1-1321	5079	99	39	6	21	2012-08-02 04:51:31.910011
5082	8	1-1382	5079	39	46	7	90	2012-08-02 04:51:31.910011
5083	8	1-986	5079	46	23	3	71	2012-08-02 04:51:31.910011
5084	8	1-913	5079	23	78	2	79	2012-08-02 04:51:31.910011
5085	8	1-1455	5079	78	61	8	3	2012-08-02 04:51:31.910011
5079	8	1-1485	5079	49	8	18	33	2012-08-02 04:51:31.910011
5086	8	1-1211	5079	61	49	7	64	2012-08-02 04:51:31.910011
15947	4	1-2566	15945	19	56	743	67	2012-08-02 05:13:19.449411
5088	5	1-1467	5087	8	53	27	66	2012-08-02 04:51:31.910011
5089	5	1-1213	5087	53	36	48	53	2012-08-02 04:51:31.910011
5090	5	1-1368	5087	36	12	52	54	2012-08-02 04:51:31.910011
5091	5	1-1208	5087	12	49	35	64	2012-08-02 04:51:31.910011
5087	5	1-1485	5087	49	8	99	33	2012-08-02 04:51:31.910011
15948	4	1-4500	15945	56	49	203	60	2012-08-02 05:13:19.449411
5093	5	1-977	5092	47	49	2586	64	2012-08-02 04:51:36.726515
5094	5	1-1485	5092	49	44	5741	33	2012-08-02 04:51:36.726515
5095	5	1-1045	5092	44	74	2069	76	2012-08-02 04:51:36.726515
5096	5	1-414	5092	74	2	563	27	2012-08-02 04:51:36.726515
5092	5	1-1494	5092	2	47	5086	36	2012-08-02 04:51:36.726515
15945	4	1-4594	15945	49	60	4390	13	2012-08-02 05:13:19.449411
5098	6	1-1060	5097	51	24	216	26	2012-08-02 04:51:36.726515
5099	6	1-971	5097	24	48	284	11	2012-08-02 04:51:36.726515
5100	6	1-1403	5097	48	13	881	58	2012-08-02 04:51:36.726515
5101	6	1-944	5097	13	42	346	47	2012-08-02 04:51:36.726515
5102	6	1-1452	5097	42	2	273	27	2012-08-02 04:51:36.726515
5097	6	1-1494	5097	2	51	2832	36	2012-08-02 04:51:36.726515
5103	1	1-971	\N	24	24	75	11	2012-08-02 04:51:36.726515
5105	5	1-1476	5104	70	91	2685	86	2012-08-02 04:51:37.212597
5106	5	1-1122	5104	91	20	1355	67	2012-08-02 04:51:37.212597
5107	5	1-1123	5104	20	31	995	60	2012-08-02 04:51:37.212597
5108	5	1-1192	5104	31	11	1163	31	2012-08-02 04:51:37.212597
5104	5	1-1495	5104	11	70	6155	87	2012-08-02 04:51:37.212597
5110	3	1-1471	5109	54	67	487	82	2012-08-02 04:51:37.212597
5111	3	1-409	5109	67	11	310	31	2012-08-02 04:51:37.212597
5109	3	1-1495	5109	11	54	1835	87	2012-08-02 04:51:37.212597
5113	5	1-860	5112	42	82	288	34	2012-08-02 04:51:42.227933
5114	5	1-980	5112	82	66	190	22	2012-08-02 04:51:42.227933
5115	5	1-1421	5112	66	62	436	32	2012-08-02 04:51:42.227933
5116	5	1-779	5112	62	67	319	81	2012-08-02 04:51:42.227933
5112	5	1-1507	5112	67	42	447	36	2012-08-02 04:51:42.227933
5118	6	1-417	5117	2	99	4636	95	2012-08-02 04:51:42.227933
5119	6	1-1351	5117	99	39	1852	21	2012-08-02 04:51:42.227933
5120	6	1-1382	5117	39	24	2410	90	2012-08-02 04:51:42.227933
5121	6	1-989	5117	24	39	4975	68	2012-08-02 04:51:42.227933
5122	6	1-1346	5117	39	67	3696	81	2012-08-02 04:51:42.227933
5117	6	1-1507	5117	67	2	5286	36	2012-08-02 04:51:42.227933
5124	7	1-1424	5123	85	16	837	97	2012-08-02 04:51:46.679714
5125	7	1-1378	5123	16	87	1334	55	2012-08-02 04:51:46.679714
5126	7	1-1437	5123	87	92	3166	1	2012-08-02 04:51:46.679714
5127	7	1-872	5123	92	85	2263	56	2012-08-02 04:51:46.679714
5128	7	1-1348	5123	85	25	1271	46	2012-08-02 04:51:46.679714
5129	7	1-778	5123	25	45	1545	45	2012-08-02 04:51:46.679714
5123	7	1-1516	5123	45	85	367	91	2012-08-02 04:51:46.679714
5131	6	1-1340	5130	23	18	107	42	2012-08-02 04:51:50.194711
5132	6	1-1299	5130	18	15	377	93	2012-08-02 04:51:50.194711
5133	6	1-1456	5130	15	49	543	85	2012-08-02 04:51:50.194711
5134	6	1-1489	5130	49	69	274	9	2012-08-02 04:51:50.194711
5135	6	1-1029	5130	69	61	57	25	2012-08-02 04:51:50.194711
5130	6	1-1522	5130	61	23	109	69	2012-08-02 04:51:50.194711
5137	6	1-1179	5136	33	100	71	90	2012-08-02 04:51:50.194711
5138	6	1-1396	5136	100	15	356	93	2012-08-02 04:51:50.194711
5139	6	1-1456	5136	15	49	664	85	2012-08-02 04:51:50.194711
5140	6	1-1489	5136	49	69	434	9	2012-08-02 04:51:50.194711
5141	6	1-1029	5136	69	61	118	25	2012-08-02 04:51:50.194711
5136	6	1-1522	5136	61	33	290	69	2012-08-02 04:51:50.194711
5143	5	1-1468	5142	31	37	756	52	2012-08-02 04:51:52.94539
5144	5	1-1204	5142	37	73	962	12	2012-08-02 04:51:52.94539
5145	5	1-1056	5142	73	17	1106	94	2012-08-02 04:51:52.94539
5146	5	1-463	5142	17	74	812	93	2012-08-02 04:51:52.94539
5142	5	1-1528	5142	74	31	337	57	2012-08-02 04:51:52.94539
5148	6	1-1116	5147	54	8	614	15	2012-08-02 04:51:52.94539
5149	6	1-773	5147	8	88	353	50	2012-08-02 04:51:52.94539
5150	6	1-1461	5147	88	39	306	21	2012-08-02 04:51:52.94539
5151	6	1-1382	5147	39	100	368	90	2012-08-02 04:51:52.94539
5152	6	1-1396	5147	100	74	1743	93	2012-08-02 04:51:52.94539
5147	6	1-1528	5147	74	54	724	57	2012-08-02 04:51:52.94539
5154	4	1-1471	5153	54	44	262	82	2012-08-02 04:51:54.315396
5155	4	1-682	5153	44	11	236	8	2012-08-02 04:51:54.315396
5156	4	1-1483	5153	11	64	1131	100	2012-08-02 04:51:54.315396
5153	4	1-1531	5153	64	54	747	87	2012-08-02 04:51:54.315396
5158	6	1-655	5157	78	88	770	50	2012-08-02 04:51:57.099444
5159	6	1-1461	5157	88	39	730	21	2012-08-02 04:51:57.099444
5160	6	1-1382	5157	39	100	963	90	2012-08-02 04:51:57.099444
5161	6	1-1396	5157	100	44	4989	93	2012-08-02 04:51:57.099444
5162	6	1-376	5157	44	81	1690	72	2012-08-02 04:51:57.099444
5157	6	1-1537	5157	81	78	2092	46	2012-08-02 04:51:57.099444
5164	6	1-514	5163	89	34	2615	47	2012-08-02 04:51:58.701101
5165	6	1-740	5163	34	82	1270	34	2012-08-02 04:51:58.701101
5166	6	1-980	5163	82	66	386	22	2012-08-02 04:51:58.701101
5167	6	1-1421	5163	66	36	409	32	2012-08-02 04:51:58.701101
5168	6	1-1270	5163	36	77	699	40	2012-08-02 04:51:58.701101
5163	6	1-1539	5163	77	89	782	35	2012-08-02 04:51:58.701101
5170	5	1-581	5169	13	68	99	21	2012-08-02 04:51:58.701101
5171	5	1-760	5169	68	66	41	22	2012-08-02 04:51:58.701101
5172	5	1-1421	5169	66	36	44	32	2012-08-02 04:51:58.701101
5173	5	1-1270	5169	36	77	77	40	2012-08-02 04:51:58.701101
5169	5	1-1539	5169	77	13	87	35	2012-08-02 04:51:58.701101
5175	7	1-514	5174	89	42	5036	47	2012-08-02 04:51:58.701101
5176	7	1-1452	5174	42	39	2216	27	2012-08-02 04:51:58.701101
5177	7	1-1484	5174	39	11	1513	84	2012-08-02 04:51:58.701101
5178	7	1-1047	5174	11	31	661	3	2012-08-02 04:51:58.701101
5179	7	1-1508	5174	31	22	113	73	2012-08-02 04:51:58.701101
5180	7	1-1334	5174	22	77	1161	40	2012-08-02 04:51:58.701101
5174	7	1-1539	5174	77	89	1398	35	2012-08-02 04:51:58.701101
5182	6	1-99	5181	52	86	889	11	2012-08-02 04:51:58.701101
5183	6	1-1505	5181	86	87	392	44	2012-08-02 04:51:58.701101
5184	6	1-1446	5181	87	39	355	65	2012-08-02 04:51:58.701101
5185	6	1-1305	5181	39	22	183	73	2012-08-02 04:51:58.701101
5186	6	1-1334	5181	22	77	2057	40	2012-08-02 04:51:58.701101
5181	6	1-1539	5181	77	52	2721	35	2012-08-02 04:51:58.701101
5187	5	1-1540	5187	57	81	3306	54	2012-08-02 04:51:59.165182
5188	5	1-734	5187	81	53	1652	46	2012-08-02 04:51:59.165182
5189	5	1-1206	5187	53	74	1512	28	2012-08-02 04:51:59.165182
5190	5	1-687	5187	74	51	997	93	2012-08-02 04:51:59.165182
5191	5	1-1423	5187	51	57	910	80	2012-08-02 04:51:59.165182
15950	4	1-808	15949	36	62	9	51	2012-08-02 05:13:19.449411
5193	6	1-1208	5192	12	40	1224	64	2012-08-02 04:51:59.165182
5194	6	1-51	5192	40	43	556	7	2012-08-02 04:51:59.165182
5195	6	1-943	5192	43	15	496	93	2012-08-02 04:51:59.165182
5196	6	1-1456	5192	15	12	753	85	2012-08-02 04:51:59.165182
5197	6	1-601	5192	12	57	474	80	2012-08-02 04:51:59.165182
5192	6	1-1540	5192	57	12	1903	54	2012-08-02 04:51:59.165182
15951	4	1-4300	15949	62	56	4	67	2012-08-02 05:13:19.449411
5199	5	1-1208	5198	12	40	412	64	2012-08-02 04:51:59.165182
5200	5	1-51	5198	40	43	190	7	2012-08-02 04:51:59.165182
5201	5	1-943	5198	43	51	172	93	2012-08-02 04:51:59.165182
5202	5	1-1423	5198	51	57	176	80	2012-08-02 04:51:59.165182
5198	5	1-1540	5198	57	12	719	54	2012-08-02 04:51:59.165182
5203	1	1-51	\N	40	40	2100	7	2012-08-02 04:51:59.165182
5205	5	1-1346	5204	39	67	120	81	2012-08-02 04:51:59.6075
5206	5	1-1507	5204	67	2	133	36	2012-08-02 04:51:59.6075
5207	5	1-417	5204	2	61	91	95	2012-08-02 04:51:59.6075
5208	5	1-1084	5204	61	59	37	12	2012-08-02 04:51:59.6075
5204	5	1-1541	5204	59	39	215	68	2012-08-02 04:51:59.6075
5210	7	1-1310	5209	72	20	1814	77	2012-08-02 04:51:59.6075
5211	7	1-1466	5209	20	74	2511	76	2012-08-02 04:51:59.6075
5212	7	1-414	5209	74	39	740	27	2012-08-02 04:51:59.6075
5213	7	1-1484	5209	39	58	853	84	2012-08-02 04:51:59.6075
5214	7	1-1504	5209	58	37	845	52	2012-08-02 04:51:59.6075
5215	7	1-1204	5209	37	59	926	12	2012-08-02 04:51:59.6075
5209	7	1-1541	5209	59	72	5650	68	2012-08-02 04:51:59.6075
5217	3	1-682	5216	44	75	1817	8	2012-08-02 04:52:00.093292
5218	3	1-965	5216	75	70	1109	54	2012-08-02 04:52:00.093292
5216	3	1-1542	5216	70	44	3685	82	2012-08-02 04:52:00.093292
5220	3	1-1123	5219	20	83	803	60	2012-08-02 04:52:00.546379
5221	3	1-933	5219	83	64	201	2	2012-08-02 04:52:00.546379
5219	3	1-1543	5219	64	20	2860	67	2012-08-02 04:52:00.546379
5223	3	1-1496	5222	41	26	1789	49	2012-08-02 04:52:01.02139
5224	3	1-416	5222	26	70	1599	20	2012-08-02 04:52:01.02139
5222	3	1-1544	5222	70	41	2943	96	2012-08-02 04:52:01.02139
5226	6	1-1331	5225	24	24	874	63	2012-08-02 04:52:01.02139
5227	6	1-1538	5225	24	36	5393	23	2012-08-02 04:52:01.02139
5228	6	1-1364	5225	36	35	364	19	2012-08-02 04:52:01.02139
5229	6	1-941	5225	35	2	195	11	2012-08-02 04:52:01.02139
5230	6	1-1055	5225	2	70	235	20	2012-08-02 04:52:01.02139
5225	6	1-1544	5225	70	24	664	96	2012-08-02 04:52:01.02139
5232	4	1-1363	5231	6	3	973	38	2012-08-02 04:52:01.496654
5233	4	1-733	5231	3	66	1486	44	2012-08-02 04:52:01.496654
5234	4	1-1416	5231	66	26	2161	19	2012-08-02 04:52:01.496654
5231	4	1-1545	5231	26	6	2058	67	2012-08-02 04:52:01.496654
5236	8	1-436	5235	74	86	199	43	2012-08-02 04:52:01.496654
5237	8	1-877	5235	86	81	696	72	2012-08-02 04:52:01.496654
5238	8	1-1537	5235	81	25	758	46	2012-08-02 04:52:01.496654
5239	8	1-778	5235	25	45	817	45	2012-08-02 04:52:01.496654
5240	8	1-1516	5235	45	100	172	91	2012-08-02 04:52:01.496654
5241	8	1-638	5235	100	66	202	44	2012-08-02 04:52:01.496654
5242	8	1-1416	5235	66	26	297	19	2012-08-02 04:52:01.496654
5235	8	1-1545	5235	26	74	286	67	2012-08-02 04:52:01.496654
5244	7	1-1235	5243	55	67	83	15	2012-08-02 04:52:03.054506
5245	7	1-720	5243	67	42	63	47	2012-08-02 04:52:03.054506
5246	7	1-1452	5243	42	39	56	27	2012-08-02 04:52:03.054506
5247	7	1-1484	5243	39	11	77	84	2012-08-02 04:52:03.054506
5248	7	1-1047	5243	11	60	68	3	2012-08-02 04:52:03.054506
5249	7	1-1359	5243	60	18	68	59	2012-08-02 04:52:03.054506
5243	7	1-1547	5243	18	55	137	86	2012-08-02 04:52:03.054506
5251	3	1-1307	5250	47	41	1959	24	2012-08-02 04:52:03.054506
5252	3	1-52	5250	41	18	1993	59	2012-08-02 04:52:03.054506
5250	3	1-1547	5250	18	47	4149	86	2012-08-02 04:52:03.054506
5254	3	1-391	5253	11	26	2526	19	2012-08-02 04:52:05.419795
5255	3	1-1545	5253	26	18	2172	67	2012-08-02 04:52:05.419795
5253	3	1-1553	5253	18	11	5685	82	2012-08-02 04:52:05.419795
5257	6	1-1010	5256	27	59	476	12	2012-08-02 04:52:05.419795
5258	6	1-1541	5256	59	67	3059	68	2012-08-02 04:52:05.419795
5259	6	1-1102	5256	67	23	976	69	2012-08-02 04:52:05.419795
5260	6	1-1340	5256	23	94	1065	42	2012-08-02 04:52:05.419795
5261	6	1-1398	5256	94	18	171	67	2012-08-02 04:52:05.419795
5256	6	1-1553	5256	18	27	470	82	2012-08-02 04:52:05.419795
5263	5	1-358	5262	67	70	181	20	2012-08-02 04:52:05.927995
5264	5	1-1544	5262	70	24	438	96	2012-08-02 04:52:05.927995
5265	5	1-1331	5262	24	24	533	63	2012-08-02 04:52:05.927995
5266	5	1-916	5262	24	14	387	92	2012-08-02 04:52:05.927995
5262	5	1-1555	5262	14	67	2278	69	2012-08-02 04:52:05.927995
5268	7	1-720	5267	67	78	1521	47	2012-08-02 04:52:07.640049
5269	7	1-150	5267	78	7	1308	78	2012-08-02 04:52:07.640049
5270	7	1-1525	5267	7	74	1573	28	2012-08-02 04:52:07.640049
5271	7	1-687	5267	74	15	1400	93	2012-08-02 04:52:07.640049
5272	7	1-1456	5267	15	49	2594	85	2012-08-02 04:52:07.640049
5273	7	1-1489	5267	49	22	1687	9	2012-08-02 04:52:07.640049
5267	7	1-1556	5267	22	67	1944	15	2012-08-02 04:52:07.640049
5275	6	1-687	5274	74	15	364	93	2012-08-02 04:52:08.578907
5276	6	1-1456	5274	15	49	669	85	2012-08-02 04:52:08.578907
5277	6	1-1489	5274	49	22	431	9	2012-08-02 04:52:08.578907
5278	6	1-1556	5274	22	38	493	15	2012-08-02 04:52:08.578907
5279	6	1-1069	5274	38	70	34	45	2012-08-02 04:52:08.578907
5274	6	1-1558	5274	70	74	412	28	2012-08-02 04:52:08.578907
5281	6	1-1466	5280	20	74	1765	76	2012-08-02 04:52:12.666085
5282	6	1-414	5280	74	39	569	27	2012-08-02 04:52:12.666085
5283	6	1-1484	5280	39	58	717	84	2012-08-02 04:52:12.666085
5284	6	1-1504	5280	58	37	777	52	2012-08-02 04:52:12.666085
5285	6	1-1204	5280	37	71	931	12	2012-08-02 04:52:12.666085
5280	6	1-1564	5280	71	20	1166	77	2012-08-02 04:52:12.666085
5287	7	1-792	5286	78	89	246	24	2012-08-02 04:52:12.666085
5288	7	1-1200	5286	89	84	589	48	2012-08-02 04:52:12.666085
5289	7	1-1439	5286	84	86	565	10	2012-08-02 04:52:12.666085
5290	7	1-764	5286	86	70	104	54	2012-08-02 04:52:12.666085
5291	7	1-1542	5286	70	27	560	82	2012-08-02 04:52:12.666085
5292	7	1-1010	5286	27	71	617	12	2012-08-02 04:52:12.666085
5286	7	1-1564	5286	71	78	810	77	2012-08-02 04:52:12.666085
15952	4	1-4500	15949	56	49	2	60	2012-08-02 05:13:19.449411
5294	4	1-1192	5293	31	24	5903	31	2012-08-02 04:52:18.122814
5295	4	1-1514	5293	24	25	2531	71	2012-08-02 04:52:18.122814
5296	4	1-1442	5293	25	45	5451	32	2012-08-02 04:52:18.122814
5293	4	1-1573	5293	45	31	5011	60	2012-08-02 04:52:18.122814
15949	4	1-4594	15949	49	36	56	13	2012-08-02 05:13:19.449411
5298	4	1-46	5297	9	24	1741	31	2012-08-02 04:52:18.122814
5299	4	1-1514	5297	24	25	766	71	2012-08-02 04:52:18.122814
5300	4	1-1442	5297	25	45	1695	32	2012-08-02 04:52:18.122814
5297	4	1-1573	5297	45	9	1600	60	2012-08-02 04:52:18.122814
5302	2	1-842	5301	70	44	584	53	2012-08-02 04:52:20.011937
5301	2	1-1577	5301	44	70	1810	7	2012-08-02 04:52:20.011937
15954	5	1-3346	15953	55	96	848	97	2012-08-02 05:13:19.859507
5304	5	1-943	5303	43	74	1143	93	2012-08-02 04:52:20.011937
5305	5	1-1528	5303	74	90	454	57	2012-08-02 04:52:20.011937
5306	5	1-1224	5303	90	64	302	69	2012-08-02 04:52:20.011937
5307	5	1-1125	5303	64	44	199	53	2012-08-02 04:52:20.011937
5303	5	1-1577	5303	44	43	1152	7	2012-08-02 04:52:20.011937
5309	6	1-1239	5308	93	32	584	41	2012-08-02 04:52:20.011937
5310	6	1-36	5308	32	26	663	51	2012-08-02 04:52:20.011937
5311	6	1-597	5308	26	67	374	68	2012-08-02 04:52:20.011937
5312	6	1-1102	5308	67	64	136	69	2012-08-02 04:52:20.011937
5313	6	1-1125	5308	64	44	97	53	2012-08-02 04:52:20.011937
5308	6	1-1577	5308	44	93	607	7	2012-08-02 04:52:20.011937
5315	6	1-607	5314	28	40	136	16	2012-08-02 04:52:20.973416
5316	6	1-1233	5314	40	91	141	82	2012-08-02 04:52:20.973416
5317	6	1-1196	5314	91	84	341	74	2012-08-02 04:52:20.973416
5318	6	1-1551	5314	84	33	169	100	2012-08-02 04:52:20.973416
5319	6	1-1361	5314	33	57	339	72	2012-08-02 04:52:20.973416
5314	6	1-1579	5314	57	28	530	70	2012-08-02 04:52:20.973416
5321	4	1-549	5320	94	67	366	81	2012-08-02 04:52:20.973416
5322	4	1-1507	5320	67	56	517	36	2012-08-02 04:52:20.973416
5323	4	1-273	5320	56	57	366	72	2012-08-02 04:52:20.973416
5320	4	1-1579	5320	57	94	577	70	2012-08-02 04:52:20.973416
5325	6	1-1569	5324	36	65	2829	95	2012-08-02 04:52:20.973416
5326	6	1-1104	5324	65	75	1309	30	2012-08-02 04:52:20.973416
5327	6	1-438	5324	75	48	405	11	2012-08-02 04:52:20.973416
5328	6	1-1403	5324	48	49	1446	58	2012-08-02 04:52:20.973416
5329	6	1-732	5324	49	57	2363	72	2012-08-02 04:52:20.973416
5324	6	1-1579	5324	57	36	3793	70	2012-08-02 04:52:20.973416
5331	6	1-945	5330	80	44	197	53	2012-08-02 04:52:20.973416
5332	6	1-1577	5330	44	18	1285	7	2012-08-02 04:52:20.973416
5333	6	1-533	5330	18	91	231	16	2012-08-02 04:52:20.973416
5334	6	1-1059	5330	91	29	432	29	2012-08-02 04:52:20.973416
5335	6	1-78	5330	29	57	443	72	2012-08-02 04:52:20.973416
5330	6	1-1579	5330	57	80	726	70	2012-08-02 04:52:20.973416
5337	4	1-1569	5336	36	37	1225	95	2012-08-02 04:52:20.973416
5338	4	1-1145	5336	37	29	913	29	2012-08-02 04:52:20.973416
5339	4	1-78	5336	29	57	951	72	2012-08-02 04:52:20.973416
5336	4	1-1579	5336	57	36	1583	70	2012-08-02 04:52:20.973416
5341	5	1-1556	5340	22	8	5981	15	2012-08-02 04:52:21.448712
5342	5	1-773	5340	8	88	2640	50	2012-08-02 04:52:21.448712
5343	5	1-1461	5340	88	85	1757	21	2012-08-02 04:52:21.448712
5344	5	1-1037	5340	85	23	779	48	2012-08-02 04:52:21.448712
5340	5	1-1580	5340	23	22	7086	9	2012-08-02 04:52:21.448712
5346	6	1-1029	5345	69	1	688	25	2012-08-02 04:52:21.448712
5347	6	1-1333	5345	1	71	284	12	2012-08-02 04:52:21.448712
5348	6	1-1564	5345	71	78	362	77	2012-08-02 04:52:21.448712
5349	6	1-792	5345	78	89	107	24	2012-08-02 04:52:21.448712
5350	6	1-1200	5345	89	23	247	48	2012-08-02 04:52:21.448712
5345	6	1-1580	5345	23	69	2802	9	2012-08-02 04:52:21.448712
5352	5	1-992	5351	17	93	169	39	2012-08-02 04:52:21.901689
5353	5	1-1557	5351	93	29	612	29	2012-08-02 04:52:21.901689
5354	5	1-78	5351	29	89	453	72	2012-08-02 04:52:21.901689
5355	5	1-1236	5351	89	22	294	14	2012-08-02 04:52:21.901689
5351	5	1-1581	5351	22	17	120	19	2012-08-02 04:52:21.901689
5357	5	1-941	5356	35	2	383	11	2012-08-02 04:52:21.901689
5358	5	1-1055	5356	2	70	449	20	2012-08-02 04:52:21.901689
5359	5	1-1544	5356	70	10	1240	96	2012-08-02 04:52:21.901689
5360	5	1-1141	5356	10	22	1396	14	2012-08-02 04:52:21.901689
5356	5	1-1581	5356	22	35	730	19	2012-08-02 04:52:21.901689
5362	7	1-690	5361	63	14	265	92	2012-08-02 04:52:22.842548
5363	7	1-1555	5361	14	23	1876	69	2012-08-02 04:52:22.842548
5364	7	1-1340	5361	23	94	2373	42	2012-08-02 04:52:22.842548
5365	7	1-1398	5361	94	18	442	67	2012-08-02 04:52:22.842548
5366	7	1-1553	5361	18	27	1409	82	2012-08-02 04:52:22.842548
5367	7	1-1010	5361	27	47	1654	12	2012-08-02 04:52:22.842548
5361	7	1-1583	5361	47	63	200	8	2012-08-02 04:52:22.842548
5369	4	1-965	5368	75	70	204	54	2012-08-02 04:52:22.842548
5370	4	1-1542	5368	70	27	1217	82	2012-08-02 04:52:22.842548
5371	4	1-1010	5368	27	47	1486	12	2012-08-02 04:52:22.842548
5368	4	1-1583	5368	47	75	187	8	2012-08-02 04:52:22.842548
5373	3	1-978	5372	70	24	594	31	2012-08-02 04:52:23.317795
5374	3	1-1514	5372	24	65	220	71	2012-08-02 04:52:23.317795
5372	3	1-1584	5372	65	70	889	52	2012-08-02 04:52:23.317795
5376	7	1-1204	5375	37	47	2956	12	2012-08-02 04:52:23.317795
5377	7	1-1583	5375	47	63	290	8	2012-08-02 04:52:23.317795
5378	7	1-690	5375	63	14	310	92	2012-08-02 04:52:23.317795
5379	7	1-1555	5375	14	51	1783	69	2012-08-02 04:52:23.317795
5380	7	1-1205	5375	51	24	1081	31	2012-08-02 04:52:23.317795
5381	7	1-1514	5375	24	65	520	71	2012-08-02 04:52:23.317795
5375	7	1-1584	5375	65	37	2726	52	2012-08-02 04:52:23.317795
5383	6	1-1404	5382	83	6	383	98	2012-08-02 04:52:23.317795
5384	6	1-1291	5382	6	14	116	92	2012-08-02 04:52:23.317795
5385	6	1-1555	5382	14	51	724	69	2012-08-02 04:52:23.317795
5386	6	1-1205	5382	51	24	476	31	2012-08-02 04:52:23.317795
5387	6	1-1514	5382	24	65	249	71	2012-08-02 04:52:23.317795
5382	6	1-1584	5382	65	83	1415	52	2012-08-02 04:52:23.317795
5389	6	1-1393	5388	64	78	3622	46	2012-08-02 04:52:23.317795
5390	6	1-655	5388	78	88	1201	50	2012-08-02 04:52:23.317795
5391	6	1-1461	5388	88	39	1027	21	2012-08-02 04:52:23.317795
5392	6	1-1382	5388	39	46	1219	90	2012-08-02 04:52:23.317795
5393	6	1-986	5388	46	65	471	71	2012-08-02 04:52:23.317795
5388	6	1-1584	5388	65	64	2864	52	2012-08-02 04:52:23.317795
15955	5	1-4356	15953	96	60	423	13	2012-08-02 05:13:19.859507
5395	2	1-1478	5394	13	42	154	66	2012-08-02 04:52:23.815082
5394	2	1-1586	5394	42	13	3903	70	2012-08-02 04:52:23.815082
15956	5	1-4561	15953	60	31	1252	45	2012-08-02 05:13:19.859507
5397	7	1-1503	5396	1	47	1065	49	2012-08-02 04:52:26.212314
5398	7	1-800	5396	47	70	2389	87	2012-08-02 04:52:26.212314
5399	7	1-1476	5396	70	47	1444	86	2012-08-02 04:52:26.212314
5400	7	1-1307	5396	47	41	693	24	2012-08-02 04:52:26.212314
5401	7	1-52	5396	41	45	693	59	2012-08-02 04:52:26.212314
5402	7	1-1493	5396	45	75	531	90	2012-08-02 04:52:26.212314
5396	7	1-1591	5396	75	1	620	26	2012-08-02 04:52:26.212314
5403	1	1-52	\N	41	41	76	59	2012-08-02 04:52:26.212314
5405	6	1-1424	5404	85	2	1614	97	2012-08-02 04:52:26.687425
5406	6	1-1450	5404	2	23	1540	33	2012-08-02 04:52:26.687425
5407	6	1-1570	5404	23	51	1779	93	2012-08-02 04:52:26.687425
5408	6	1-1423	5404	51	39	1317	80	2012-08-02 04:52:26.687425
5409	6	1-782	5404	39	11	735	88	2012-08-02 04:52:26.687425
5404	6	1-1592	5404	11	85	1237	91	2012-08-02 04:52:26.687425
5411	5	1-638	5410	100	87	581	44	2012-08-02 04:52:26.687425
5412	5	1-1446	5410	87	56	613	65	2012-08-02 04:52:26.687425
5413	5	1-1427	5410	56	97	498	31	2012-08-02 04:52:26.687425
5414	5	1-935	5410	97	11	418	88	2012-08-02 04:52:26.687425
5410	5	1-1592	5410	11	100	730	91	2012-08-02 04:52:26.687425
5416	6	1-1461	5415	88	39	915	21	2012-08-02 04:52:29.74751
5417	6	1-1382	5415	39	75	916	90	2012-08-02 04:52:29.74751
5418	6	1-1591	5415	75	1	851	26	2012-08-02 04:52:29.74751
5419	6	1-1503	5415	1	8	1164	49	2012-08-02 04:52:29.74751
5420	6	1-1282	5415	8	81	2315	9	2012-08-02 04:52:29.74751
5415	6	1-1599	5415	81	88	1271	50	2012-08-02 04:52:29.74751
5422	6	1-1061	5421	65	70	578	45	2012-08-02 04:52:30.211563
5423	6	1-1558	5421	70	74	4650	28	2012-08-02 04:52:30.211563
5424	6	1-687	5421	74	74	3442	93	2012-08-02 04:52:30.211563
5425	6	1-1528	5421	74	77	1247	57	2012-08-02 04:52:30.211563
5426	6	1-1057	5421	77	1	3123	34	2012-08-02 04:52:30.211563
5421	6	1-1600	5421	1	65	1928	23	2012-08-02 04:52:30.211563
5428	6	1-419	5427	78	88	1167	50	2012-08-02 04:52:30.211563
5429	6	1-1461	5427	88	39	933	21	2012-08-02 04:52:30.211563
5430	6	1-1382	5427	39	75	1036	90	2012-08-02 04:52:30.211563
5431	6	1-1591	5427	75	78	1069	26	2012-08-02 04:52:30.211563
5432	6	1-1050	5427	78	1	1335	34	2012-08-02 04:52:30.211563
5427	6	1-1600	5427	1	78	1115	23	2012-08-02 04:52:30.211563
5434	6	1-419	5433	78	88	847	50	2012-08-02 04:52:30.211563
5435	6	1-1461	5433	88	59	738	21	2012-08-02 04:52:30.211563
5436	6	1-1129	5433	59	74	583	93	2012-08-02 04:52:30.211563
5437	6	1-1528	5433	74	77	243	57	2012-08-02 04:52:30.211563
5438	6	1-1057	5433	77	1	898	34	2012-08-02 04:52:30.211563
5433	6	1-1600	5433	1	78	817	23	2012-08-02 04:52:30.211563
5440	8	1-946	5439	84	74	884	93	2012-08-02 04:52:31.216982
5441	8	1-1528	5439	74	90	369	57	2012-08-02 04:52:31.216982
5442	8	1-1224	5439	90	23	259	69	2012-08-02 04:52:31.216982
5443	8	1-1340	5439	23	61	313	42	2012-08-02 04:52:31.216982
5444	8	1-1482	5439	61	56	145	86	2012-08-02 04:52:31.216982
5445	8	1-1595	5439	56	61	121	77	2012-08-02 04:52:31.216982
5446	8	1-504	5439	61	80	36	75	2012-08-02 04:52:31.216982
5439	8	1-1602	5439	80	84	2848	14	2012-08-02 04:52:31.216982
5448	8	1-863	5447	63	43	93	23	2012-08-02 04:52:31.216982
5449	8	1-1563	5447	43	97	129	18	2012-08-02 04:52:31.216982
5450	8	1-1088	5447	97	36	138	70	2012-08-02 04:52:31.216982
5451	8	1-1569	5447	36	61	105	95	2012-08-02 04:52:31.216982
5452	8	1-1084	5447	61	71	58	12	2012-08-02 04:52:31.216982
5453	8	1-1564	5447	71	61	83	77	2012-08-02 04:52:31.216982
5454	8	1-504	5447	61	80	25	75	2012-08-02 04:52:31.216982
5447	8	1-1602	5447	80	63	2185	14	2012-08-02 04:52:31.216982
5456	6	1-1576	5455	80	22	873	29	2012-08-02 04:52:32.156544
5457	6	1-453	5455	22	74	716	93	2012-08-02 04:52:32.156544
5458	6	1-1528	5455	74	100	292	57	2012-08-02 04:52:32.156544
5459	6	1-1388	5455	100	32	247	81	2012-08-02 04:52:32.156544
5460	6	1-1184	5455	32	25	219	68	2012-08-02 04:52:32.156544
5455	6	1-1604	5455	25	80	68	83	2012-08-02 04:52:32.156544
5462	6	1-1576	5461	80	22	3914	29	2012-08-02 04:52:32.156544
5463	6	1-453	5461	22	44	3587	93	2012-08-02 04:52:32.156544
5464	6	1-376	5461	44	81	1218	72	2012-08-02 04:52:32.156544
5465	6	1-1537	5461	81	77	1510	46	2012-08-02 04:52:32.156544
5466	6	1-1080	5461	77	25	786	68	2012-08-02 04:52:32.156544
5461	6	1-1604	5461	25	80	273	83	2012-08-02 04:52:32.156544
5468	3	1-1507	5467	67	56	2163	36	2012-08-02 04:52:32.631773
5469	3	1-273	5467	56	44	1542	72	2012-08-02 04:52:32.631773
5467	3	1-1605	5467	44	67	1519	81	2012-08-02 04:52:32.631773
5471	2	1-844	5470	33	44	3241	72	2012-08-02 04:52:32.631773
5470	2	1-1605	5470	44	33	3301	81	2012-08-02 04:52:32.631773
5473	7	1-1557	5472	93	32	3548	29	2012-08-02 04:52:33.570252
5474	7	1-1016	5472	32	2	722	97	2012-08-02 04:52:33.570252
5475	7	1-1450	5472	2	8	1133	33	2012-08-02 04:52:33.570252
5476	7	1-1467	5472	8	59	316	66	2012-08-02 04:52:33.570252
5477	7	1-1302	5472	59	70	240	20	2012-08-02 04:52:33.570252
5478	7	1-1544	5472	70	38	690	96	2012-08-02 04:52:33.570252
5472	7	1-1607	5472	38	93	734	39	2012-08-02 04:52:33.570252
5480	6	1-1402	5479	45	47	508	53	2012-08-02 04:52:34.056628
5481	6	1-856	5479	47	81	658	72	2012-08-02 04:52:34.056628
5482	6	1-1537	5479	81	78	722	46	2012-08-02 04:52:34.056628
5483	6	1-655	5479	78	88	236	50	2012-08-02 04:52:34.056628
5484	6	1-1461	5479	88	26	198	21	2012-08-02 04:52:34.056628
5479	6	1-1608	5479	26	45	407	99	2012-08-02 04:52:34.056628
5486	7	1-1402	5485	45	47	168	53	2012-08-02 04:52:34.056628
5487	7	1-856	5485	47	81	229	72	2012-08-02 04:52:34.056628
5488	7	1-1537	5485	81	91	265	46	2012-08-02 04:52:34.056628
5489	7	1-1322	5485	91	8	178	33	2012-08-02 04:52:34.056628
5490	7	1-1467	5485	8	99	49	66	2012-08-02 04:52:34.056628
5491	7	1-1321	5485	99	26	59	21	2012-08-02 04:52:34.056628
5485	7	1-1608	5485	26	45	128	99	2012-08-02 04:52:34.056628
5493	7	1-1204	5492	37	71	561	12	2012-08-02 04:52:34.553938
5494	7	1-1564	5492	71	58	719	77	2012-08-02 04:52:34.553938
5495	7	1-1425	5492	58	69	488	2	2012-08-02 04:52:34.553938
5496	7	1-1365	5492	69	43	231	73	2012-08-02 04:52:34.553938
5497	7	1-1603	5492	43	25	165	71	2012-08-02 04:52:34.553938
5492	7	1-1609	5492	58	37	456	52	2012-08-02 04:52:34.553938
5498	7	1-1442	5492	25	58	452	32	2012-08-02 04:52:34.553938
15957	5	1-3797	15953	31	72	1001	95	2012-08-02 05:13:19.859507
5500	7	1-1542	5499	70	11	1832	82	2012-08-02 04:52:35.017945
5501	7	1-391	5499	11	35	884	19	2012-08-02 04:52:35.017945
5502	7	1-941	5499	35	57	435	11	2012-08-02 04:52:35.017945
5503	7	1-91	5499	57	58	416	84	2012-08-02 04:52:35.017945
5504	7	1-1504	5499	58	33	449	52	2012-08-02 04:52:35.017945
5505	7	1-1405	5499	33	70	331	32	2012-08-02 04:52:35.017945
5499	7	1-1610	5499	70	70	329	54	2012-08-02 04:52:35.017945
15953	5	1-4595	15953	72	55	1220	70	2012-08-02 05:13:19.859507
5507	6	1-1542	5506	70	3	798	82	2012-08-02 04:52:35.017945
5508	6	1-190	5506	3	43	403	23	2012-08-02 04:52:35.017945
5509	6	1-1563	5506	43	11	519	18	2012-08-02 04:52:35.017945
5510	6	1-1262	5506	11	13	472	15	2012-08-02 04:52:35.017945
5511	6	1-1216	5506	13	70	135	32	2012-08-02 04:52:35.017945
5506	6	1-1610	5506	70	70	139	54	2012-08-02 04:52:35.017945
5513	6	1-1596	5512	90	98	8508	71	2012-08-02 04:52:35.017945
5514	6	1-1479	5512	98	38	2092	70	2012-08-02 04:52:35.017945
5515	6	1-1448	5512	38	58	541	84	2012-08-02 04:52:35.017945
5516	6	1-1504	5512	58	33	693	52	2012-08-02 04:52:35.017945
5517	6	1-1405	5512	33	70	609	32	2012-08-02 04:52:35.017945
5512	6	1-1610	5512	70	90	659	54	2012-08-02 04:52:35.017945
5518	1	1-91	\N	57	57	669	84	2012-08-02 04:52:35.017945
5519	1	1-190	\N	3	3	2638	23	2012-08-02 04:52:35.017945
15959	5	1-4083	15958	96	99	138	12	2012-08-02 05:13:20.278162
5521	5	1-1569	5520	36	61	1334	95	2012-08-02 04:52:35.493284
5522	5	1-1084	5520	61	71	252	12	2012-08-02 04:52:35.493284
5523	5	1-1564	5520	71	78	126	77	2012-08-02 04:52:35.493284
5524	5	1-1566	5520	78	70	31	36	2012-08-02 04:52:35.493284
5520	5	1-1611	5520	70	36	5024	70	2012-08-02 04:52:35.493284
5526	6	1-1448	5525	38	58	468	84	2012-08-02 04:52:35.493284
5527	6	1-1504	5525	58	37	240	52	2012-08-02 04:52:35.493284
5528	6	1-1204	5525	37	71	136	12	2012-08-02 04:52:35.493284
5529	6	1-1564	5525	71	78	81	77	2012-08-02 04:52:35.493284
5530	6	1-1566	5525	78	70	24	36	2012-08-02 04:52:35.493284
5525	6	1-1611	5525	70	38	4531	70	2012-08-02 04:52:35.493284
5531	1	1-1204	\N	37	37	390	12	2012-08-02 04:52:35.493284
5533	2	1-1465	5532	24	18	4092	100	2012-08-02 04:52:35.979463
5532	2	1-1612	5532	18	24	6854	20	2012-08-02 04:52:35.979463
5535	6	1-1608	5534	26	8	8595	99	2012-08-02 04:52:36.918414
5536	6	1-397	5534	8	42	5207	47	2012-08-02 04:52:36.918414
5537	6	1-1452	5534	42	66	4640	27	2012-08-02 04:52:36.918414
5538	6	1-1594	5534	66	69	4065	2	2012-08-02 04:52:36.918414
5539	6	1-1365	5534	69	61	2061	73	2012-08-02 04:52:36.918414
5534	6	1-1614	5534	61	26	3944	21	2012-08-02 04:52:36.918414
5541	7	1-638	5540	100	66	158	44	2012-08-02 04:52:37.404476
5542	7	1-1416	5540	66	35	242	19	2012-08-02 04:52:37.404476
5543	7	1-941	5540	35	2	128	11	2012-08-02 04:52:37.404476
5544	7	1-1055	5540	2	70	152	20	2012-08-02 04:52:37.404476
5545	7	1-1544	5540	70	31	423	96	2012-08-02 04:52:37.404476
5546	7	1-1306	5540	31	42	907	64	2012-08-02 04:52:37.404476
5540	7	1-1615	5540	42	100	129	91	2012-08-02 04:52:37.404476
5548	6	1-1504	5547	58	64	1164	52	2012-08-02 04:52:37.902065
5549	6	1-1393	5547	64	77	1464	46	2012-08-02 04:52:37.902065
5550	6	1-1080	5547	77	25	682	68	2012-08-02 04:52:37.902065
5551	6	1-1604	5547	25	25	225	83	2012-08-02 04:52:37.902065
5552	6	1-963	5547	25	56	330	45	2012-08-02 04:52:37.902065
5547	6	1-1616	5547	56	58	1030	84	2012-08-02 04:52:37.902065
5554	5	1-1386	5553	90	90	20	54	2012-08-02 04:52:39.31594
5555	5	1-1596	5553	90	23	226	71	2012-08-02 04:52:39.31594
5556	5	1-913	5553	23	75	137	79	2012-08-02 04:52:39.31594
5557	5	1-1429	5553	75	91	61	16	2012-08-02 04:52:39.31594
5553	5	1-1619	5553	91	90	52	35	2012-08-02 04:52:39.31594
5559	5	1-691	5558	82	45	1848	45	2012-08-02 04:52:40.26621
5560	5	1-1516	5558	45	5	367	91	2012-08-02 04:52:40.26621
5561	5	1-221	5558	5	24	275	63	2012-08-02 04:52:40.26621
5562	5	1-1538	5558	24	1	1441	23	2012-08-02 04:52:40.26621
5558	5	1-1621	5558	1	82	2451	8	2012-08-02 04:52:40.26621
5564	6	1-1536	5563	91	47	2047	86	2012-08-02 04:52:40.26621
5565	6	1-1307	5563	47	89	690	24	2012-08-02 04:52:40.26621
5566	6	1-1200	5563	89	77	1270	48	2012-08-02 04:52:40.26621
5567	6	1-1212	5563	77	1	476	34	2012-08-02 04:52:40.26621
5568	6	1-1600	5563	1	1	396	23	2012-08-02 04:52:40.26621
5563	6	1-1621	5563	1	91	682	8	2012-08-02 04:52:40.26621
5570	7	1-965	5569	75	70	177	54	2012-08-02 04:52:40.26621
5571	7	1-1542	5569	70	49	746	82	2012-08-02 04:52:40.26621
5572	7	1-707	5569	49	65	376	94	2012-08-02 04:52:40.26621
5573	7	1-1410	5569	65	77	55	57	2012-08-02 04:52:40.26621
5574	7	1-1057	5569	77	1	156	34	2012-08-02 04:52:40.26621
5575	7	1-1600	5569	1	1	131	23	2012-08-02 04:52:40.26621
5569	7	1-1621	5569	1	75	228	8	2012-08-02 04:52:40.26621
5577	3	1-690	5576	63	52	2857	92	2012-08-02 04:52:40.26621
5578	3	1-1293	5576	52	1	1368	23	2012-08-02 04:52:40.26621
5576	3	1-1621	5576	1	63	2413	8	2012-08-02 04:52:40.26621
5580	6	1-1614	5579	61	26	302	21	2012-08-02 04:52:42.619103
5581	6	1-1608	5579	26	66	483	99	2012-08-02 04:52:42.619103
5582	6	1-288	5579	66	64	61	75	2012-08-02 04:52:42.619103
5583	6	1-116	5579	64	80	20	83	2012-08-02 04:52:42.619103
5584	6	1-1576	5579	80	92	199	29	2012-08-02 04:52:42.619103
5579	6	1-1626	5579	92	61	216	73	2012-08-02 04:52:42.619103
5586	6	1-784	5585	67	100	274	91	2012-08-02 04:52:42.619103
5587	6	1-638	5585	100	66	254	44	2012-08-02 04:52:42.619103
5588	6	1-1416	5585	66	17	294	19	2012-08-02 04:52:42.619103
5589	6	1-992	5585	17	93	405	39	2012-08-02 04:52:42.619103
5590	6	1-1557	5585	93	92	1434	29	2012-08-02 04:52:42.619103
5585	6	1-1626	5585	92	67	1558	73	2012-08-02 04:52:42.619103
5592	7	1-1614	5591	61	85	4139	21	2012-08-02 04:52:42.619103
5593	7	1-1037	5591	85	84	1956	48	2012-08-02 04:52:42.619103
5594	7	1-1439	5591	84	29	1557	10	2012-08-02 04:52:42.619103
5595	7	1-1087	5591	29	25	840	68	2012-08-02 04:52:42.619103
5596	7	1-1604	5591	25	80	217	83	2012-08-02 04:52:42.619103
5597	7	1-1576	5591	80	92	2329	29	2012-08-02 04:52:42.619103
5591	7	1-1626	5591	92	61	2727	73	2012-08-02 04:52:42.619103
5599	7	1-1144	5598	56	51	2305	76	2012-08-02 04:52:42.619103
5600	7	1-1052	5598	51	66	2190	6	2012-08-02 04:52:42.619103
5601	7	1-766	5598	66	32	587	81	2012-08-02 04:52:42.619103
5602	7	1-1184	5598	32	25	484	68	2012-08-02 04:52:42.619103
5603	7	1-1604	5598	25	80	140	83	2012-08-02 04:52:42.619103
5604	7	1-1576	5598	80	92	1669	29	2012-08-02 04:52:42.619103
5598	7	1-1626	5598	92	56	2178	73	2012-08-02 04:52:42.619103
15960	5	1-4557	15958	99	42	227	43	2012-08-02 05:13:20.278162
5606	5	1-1603	5605	43	23	557	71	2012-08-02 04:52:42.619103
5607	5	1-913	5605	23	75	422	79	2012-08-02 04:52:42.619103
5608	5	1-1429	5605	75	91	235	16	2012-08-02 04:52:42.619103
5609	5	1-1059	5605	91	92	444	29	2012-08-02 04:52:42.619103
5605	5	1-1626	5605	92	43	691	73	2012-08-02 04:52:42.619103
5610	1	1-714	\N	59	59	513	39	2012-08-02 04:52:42.619103
5612	3	1-963	5611	25	56	252	45	2012-08-02 04:52:43.117058
5613	3	1-1552	5611	56	69	168	94	2012-08-02 04:52:43.117058
5611	3	1-1627	5611	69	25	227	83	2012-08-02 04:52:43.117058
5615	5	1-89	5614	51	1	5106	34	2012-08-02 04:52:43.117058
5616	5	1-1600	5614	1	65	4906	23	2012-08-02 04:52:43.117058
5617	5	1-1061	5614	65	56	2290	45	2012-08-02 04:52:43.117058
5618	5	1-1552	5614	56	69	2047	94	2012-08-02 04:52:43.117058
5614	5	1-1627	5614	69	51	3712	83	2012-08-02 04:52:43.117058
5620	6	1-1598	5619	42	61	189	42	2012-08-02 04:52:44.055987
5621	6	1-1482	5619	61	56	88	86	2012-08-02 04:52:44.055987
5622	6	1-1595	5619	56	61	73	77	2012-08-02 04:52:44.055987
5623	6	1-504	5619	61	80	21	75	2012-08-02 04:52:44.055987
5624	6	1-1602	5619	80	92	1704	14	2012-08-02 04:52:44.055987
5619	6	1-1629	5619	92	42	1127	80	2012-08-02 04:52:44.055987
5626	4	1-1572	5625	67	27	3608	87	2012-08-02 04:52:46.40883
5627	4	1-1634	5625	27	24	2301	14	2012-08-02 04:52:46.40883
5628	4	1-1500	5625	24	67	3194	5	2012-08-02 04:52:46.40883
5625	4	1-1635	5625	67	67	3121	51	2012-08-02 04:52:46.40883
5630	4	1-916	5629	24	52	39	92	2012-08-02 04:52:48.342372
5631	4	1-1293	5629	52	65	15	23	2012-08-02 04:52:48.342372
5632	4	1-1061	5629	65	30	5	45	2012-08-02 04:52:48.342372
5629	4	1-1639	5629	30	24	69	63	2012-08-02 04:52:48.342372
5634	7	1-838	5633	60	23	1138	69	2012-08-02 04:52:48.342372
5635	7	1-1340	5633	23	32	1420	42	2012-08-02 04:52:48.342372
5636	7	1-649	5633	32	69	240	94	2012-08-02 04:52:48.342372
5637	7	1-1627	5633	69	38	424	83	2012-08-02 04:52:48.342372
5638	7	1-470	5633	38	99	307	39	2012-08-02 04:52:48.342372
5639	7	1-1054	5633	99	30	164	45	2012-08-02 04:52:48.342372
5633	7	1-1639	5633	30	60	3047	63	2012-08-02 04:52:48.342372
5640	1	1-857	\N	44	44	5184	42	2012-08-02 04:52:48.342372
5642	7	1-1333	5641	1	73	635	12	2012-08-02 04:52:49.800296
5643	7	1-1056	5641	73	69	806	94	2012-08-02 04:52:49.800296
5644	7	1-1627	5641	69	38	1515	83	2012-08-02 04:52:49.800296
5645	7	1-470	5641	38	99	1166	39	2012-08-02 04:52:49.800296
5646	7	1-1054	5641	99	56	660	45	2012-08-02 04:52:49.800296
5647	7	1-1616	5641	56	77	2315	84	2012-08-02 04:52:49.800296
5641	7	1-1642	5641	77	1	1335	25	2012-08-02 04:52:49.800296
5648	1	1-470	\N	38	38	4658	39	2012-08-02 04:52:49.800296
5650	3	1-1333	5649	1	73	986	12	2012-08-02 04:52:51.258471
5651	3	1-1056	5649	73	33	936	94	2012-08-02 04:52:51.258471
5649	3	1-1645	5649	33	1	2770	25	2012-08-02 04:52:51.258471
5653	7	1-1407	5652	85	67	484	4	2012-08-02 04:52:52.1866
5654	7	1-1643	5652	67	8	372	33	2012-08-02 04:52:52.1866
5655	7	1-1467	5652	8	59	102	66	2012-08-02 04:52:52.1866
5656	7	1-1302	5652	59	54	76	20	2012-08-02 04:52:52.1866
5657	7	1-1443	5652	54	33	159	94	2012-08-02 04:52:52.1866
5658	7	1-1645	5652	33	75	581	25	2012-08-02 04:52:52.1866
5652	7	1-1647	5652	75	85	596	13	2012-08-02 04:52:52.1866
5660	6	1-691	5659	82	30	125	45	2012-08-02 04:52:55.071644
5661	6	1-1639	5659	30	51	1580	63	2012-08-02 04:52:55.071644
5662	6	1-1567	5659	51	67	1436	5	2012-08-02 04:52:55.071644
5663	6	1-1635	5659	67	67	1344	51	2012-08-02 04:52:55.071644
5664	6	1-1572	5659	67	73	1488	87	2012-08-02 04:52:55.071644
5659	6	1-1653	5659	73	82	214	8	2012-08-02 04:52:55.071644
5666	6	1-691	5665	82	30	78	45	2012-08-02 04:52:55.071644
5667	6	1-1639	5665	30	36	1394	63	2012-08-02 04:52:55.071644
5668	6	1-1186	5665	36	88	1504	50	2012-08-02 04:52:55.071644
5669	6	1-1461	5665	88	19	1295	21	2012-08-02 04:52:55.071644
5670	6	1-1067	5665	19	73	464	87	2012-08-02 04:52:55.071644
5665	6	1-1653	5665	73	82	95	8	2012-08-02 04:52:55.071644
5672	6	1-1057	5671	77	82	306	34	2012-08-02 04:52:56.011076
5673	6	1-980	5671	82	6	160	22	2012-08-02 04:52:56.011076
5674	6	1-1578	5671	6	51	134	69	2012-08-02 04:52:56.011076
5675	6	1-1205	5671	51	24	78	31	2012-08-02 04:52:56.011076
5676	6	1-1514	5671	24	94	36	71	2012-08-02 04:52:56.011076
5671	6	1-1655	5671	94	77	102	57	2012-08-02 04:52:56.011076
5678	6	1-1224	5677	90	96	3496	69	2012-08-02 04:52:56.011076
5679	6	1-1597	5677	96	75	2045	1	2012-08-02 04:52:56.011076
5680	6	1-431	5677	75	68	4920	17	2012-08-02 04:52:56.011076
5681	6	1-1649	5677	68	24	3058	31	2012-08-02 04:52:56.011076
5682	6	1-1514	5677	24	94	1624	71	2012-08-02 04:52:56.011076
5677	6	1-1655	5677	94	90	5340	57	2012-08-02 04:52:56.011076
5684	6	1-42	5683	34	42	1008	47	2012-08-02 04:52:56.011076
5685	6	1-1452	5683	42	66	920	27	2012-08-02 04:52:56.011076
5686	6	1-1594	5683	66	69	826	2	2012-08-02 04:52:56.011076
5687	6	1-1365	5683	69	43	429	73	2012-08-02 04:52:56.011076
5688	6	1-1603	5683	43	94	336	71	2012-08-02 04:52:56.011076
5683	6	1-1655	5683	94	34	1243	57	2012-08-02 04:52:56.011076
5690	5	1-760	5689	68	6	3660	22	2012-08-02 04:52:56.949971
5691	5	1-1578	5689	6	19	3037	69	2012-08-02 04:52:56.949971
5692	5	1-1341	5689	19	24	2622	14	2012-08-02 04:52:56.949971
5693	5	1-1500	5689	24	37	3848	5	2012-08-02 04:52:56.949971
5689	5	1-1657	5689	37	68	5275	21	2012-08-02 04:52:56.949971
5695	3	1-760	5694	68	28	303	22	2012-08-02 04:52:56.949971
5696	3	1-1352	5694	28	37	214	5	2012-08-02 04:52:56.949971
5694	3	1-1657	5694	37	68	358	21	2012-08-02 04:52:56.949971
5698	7	1-1208	5697	12	42	764	64	2012-08-02 04:52:58.794864
5699	7	1-1615	5697	42	32	105	91	2012-08-02 04:52:58.794864
5700	7	1-1435	5697	32	1	136	26	2012-08-02 04:52:58.794864
5701	7	1-1503	5697	1	47	218	49	2012-08-02 04:52:58.794864
5702	7	1-800	5697	47	73	445	87	2012-08-02 04:52:58.794864
5703	7	1-1653	5697	73	64	89	8	2012-08-02 04:52:58.794864
5697	7	1-1661	5697	64	12	1187	54	2012-08-02 04:52:58.794864
5704	6	1-1663	5704	33	30	45	45	2012-08-02 04:52:59.700764
5705	6	1-1639	5704	30	76	734	63	2012-08-02 04:52:59.700764
5706	6	1-1383	5704	76	83	900	53	2012-08-02 04:52:59.700764
5707	6	1-1462	5704	83	75	45	79	2012-08-02 04:52:59.700764
5708	6	1-1429	5704	75	91	21	16	2012-08-02 04:52:59.700764
5709	6	1-1619	5704	91	33	18	35	2012-08-02 04:52:59.700764
15961	5	1-4412	15958	42	62	135	51	2012-08-02 05:13:20.278162
5711	8	1-1639	5710	30	51	1362	63	2012-08-02 04:52:59.700764
5712	8	1-1567	5710	51	37	1806	5	2012-08-02 04:52:59.700764
5713	8	1-1657	5710	37	42	3132	21	2012-08-02 04:52:59.700764
5714	8	1-996	5710	42	92	1767	74	2012-08-02 04:52:59.700764
5715	8	1-1565	5710	92	18	152	7	2012-08-02 04:52:59.700764
5716	8	1-533	5710	18	91	26	16	2012-08-02 04:52:59.700764
5717	8	1-1619	5710	91	33	26	35	2012-08-02 04:52:59.700764
5710	8	1-1663	5710	33	30	74	45	2012-08-02 04:52:59.700764
15962	5	1-4300	15958	62	2	41	67	2012-08-02 05:13:20.278162
5719	7	1-1516	5718	45	100	499	91	2012-08-02 04:52:59.700764
5720	7	1-638	5718	100	66	642	44	2012-08-02 04:52:59.700764
5721	7	1-1416	5718	66	35	1032	19	2012-08-02 04:52:59.700764
5722	7	1-941	5718	35	72	573	11	2012-08-02 04:52:59.700764
5723	7	1-1561	5718	72	91	719	16	2012-08-02 04:52:59.700764
5724	7	1-1619	5718	91	33	732	35	2012-08-02 04:52:59.700764
5718	7	1-1663	5718	33	45	2166	45	2012-08-02 04:52:59.700764
15958	5	1-4596	15958	2	96	751	66	2012-08-02 05:13:20.278162
5726	8	1-1552	5725	56	4	801	94	2012-08-02 04:52:59.700764
5727	8	1-679	5725	4	73	962	87	2012-08-02 04:52:59.700764
5728	8	1-1653	5725	73	64	219	8	2012-08-02 04:52:59.700764
5729	8	1-1661	5725	64	33	3301	54	2012-08-02 04:52:59.700764
5730	8	1-922	5725	33	75	451	79	2012-08-02 04:52:59.700764
5731	8	1-1429	5725	75	91	256	16	2012-08-02 04:52:59.700764
5732	8	1-1619	5725	91	33	275	35	2012-08-02 04:52:59.700764
5725	8	1-1663	5725	33	56	858	45	2012-08-02 04:52:59.700764
5733	1	1-679	\N	4	4	1546	87	2012-08-02 04:52:59.700764
5735	6	1-1208	5734	12	77	1677	64	2012-08-02 04:53:01.214471
5736	6	1-1633	5734	77	25	1527	33	2012-08-02 04:53:01.214471
5737	6	1-1651	5734	25	40	1997	17	2012-08-02 04:53:01.214471
5738	6	1-1662	5734	40	49	2545	85	2012-08-02 04:53:01.214471
5739	6	1-1489	5734	49	53	1669	9	2012-08-02 04:53:01.214471
5734	6	1-1664	5734	53	12	2412	54	2012-08-02 04:53:01.214471
5741	4	1-1652	5740	61	94	734	71	2012-08-02 04:53:02.562039
5742	4	1-1655	5740	94	8	1977	57	2012-08-02 04:53:02.562039
5743	4	1-1534	5740	8	7	6625	37	2012-08-02 04:53:02.562039
5740	4	1-1667	5740	7	61	2479	52	2012-08-02 04:53:02.562039
5745	6	1-1405	5744	33	55	209	32	2012-08-02 04:53:02.562039
5746	6	1-1646	5744	55	22	704	62	2012-08-02 04:53:02.562039
5747	6	1-849	5744	22	40	240	17	2012-08-02 04:53:02.562039
5748	6	1-1662	5744	40	2	260	85	2012-08-02 04:53:02.562039
5749	6	1-646	5744	2	7	672	37	2012-08-02 04:53:02.562039
5744	6	1-1667	5744	7	33	294	52	2012-08-02 04:53:02.562039
5751	3	1-1306	5750	31	12	701	64	2012-08-02 04:53:03.478976
5752	3	1-1641	5750	12	13	455	13	2012-08-02 04:53:03.478976
5750	3	1-1669	5750	13	31	389	96	2012-08-02 04:53:03.478976
5754	5	1-1496	5753	41	14	333	49	2012-08-02 04:53:03.478976
5755	5	1-440	5753	14	33	157	94	2012-08-02 04:53:03.478976
5756	5	1-1645	5753	33	75	504	25	2012-08-02 04:53:03.478976
5757	5	1-1647	5753	75	13	454	13	2012-08-02 04:53:03.478976
5753	5	1-1669	5753	13	41	408	96	2012-08-02 04:53:03.478976
5759	7	1-877	5758	86	44	34	72	2012-08-02 04:53:03.932227
5760	7	1-1605	5758	44	32	27	81	2012-08-02 04:53:03.932227
5761	7	1-1184	5758	32	67	21	68	2012-08-02 04:53:03.932227
5762	7	1-1102	5758	67	23	6	69	2012-08-02 04:53:03.932227
5763	7	1-1340	5758	23	42	6	42	2012-08-02 04:53:03.932227
5764	7	1-1143	5758	42	75	7	80	2012-08-02 04:53:03.932227
5758	7	1-1670	5758	75	86	11	43	2012-08-02 04:53:03.932227
5766	6	1-1196	5765	91	84	1404	74	2012-08-02 04:53:04.871429
5767	6	1-1551	5765	84	73	691	100	2012-08-02 04:53:04.871429
5768	6	1-1412	5765	73	1	568	86	2012-08-02 04:53:04.871429
5769	6	1-717	5765	1	13	406	13	2012-08-02 04:53:04.871429
5770	6	1-1669	5765	13	26	418	96	2012-08-02 04:53:04.871429
5765	6	1-1672	5765	26	91	582	82	2012-08-02 04:53:04.871429
5772	7	1-707	5771	49	69	271	94	2012-08-02 04:53:04.871429
5773	7	1-1627	5771	69	27	473	83	2012-08-02 04:53:04.871429
5774	7	1-1156	5771	27	46	233	58	2012-08-02 04:53:04.871429
5775	7	1-1132	5771	46	75	276	25	2012-08-02 04:53:04.871429
5776	7	1-1647	5771	75	13	284	13	2012-08-02 04:53:04.871429
5777	7	1-1669	5771	13	26	293	96	2012-08-02 04:53:04.871429
5771	7	1-1672	5771	26	49	408	82	2012-08-02 04:53:04.871429
5779	5	1-1423	5778	51	75	1486	80	2012-08-02 04:53:05.336402
5780	5	1-1670	5778	75	36	2332	43	2012-08-02 04:53:05.336402
5781	5	1-1526	5778	36	51	415	21	2012-08-02 04:53:05.336402
5782	5	1-1457	5778	51	76	282	41	2012-08-02 04:53:05.336402
5778	5	1-1673	5778	76	51	1304	93	2012-08-02 04:53:05.336402
5784	3	1-1305	5783	39	22	746	73	2012-08-02 04:53:09.876677
5785	3	1-1334	5783	22	39	1916	40	2012-08-02 04:53:09.876677
5783	3	1-1683	5783	39	39	2151	65	2012-08-02 04:53:09.876677
5787	3	1-1575	5786	69	28	1240	57	2012-08-02 04:53:09.876677
5788	3	1-1535	5786	28	39	3019	40	2012-08-02 04:53:09.876677
5786	3	1-1683	5786	39	69	4645	65	2012-08-02 04:53:09.876677
5790	6	1-1422	5789	9	16	1145	97	2012-08-02 04:53:13.080371
5791	6	1-1378	5789	16	51	1783	55	2012-08-02 04:53:13.080371
5792	6	1-1428	5789	51	75	871	90	2012-08-02 04:53:13.080371
5793	6	1-1591	5789	75	91	1033	26	2012-08-02 04:53:13.080371
5794	6	1-1638	5789	91	66	926	79	2012-08-02 04:53:13.080371
5789	6	1-1690	5789	66	9	630	8	2012-08-02 04:53:13.080371
5796	6	1-1483	5795	11	49	5692	100	2012-08-02 04:53:13.080371
5797	6	1-1155	5795	49	45	1998	59	2012-08-02 04:53:13.080371
5798	6	1-1493	5795	45	75	1588	90	2012-08-02 04:53:13.080371
5799	6	1-1591	5795	75	91	1921	26	2012-08-02 04:53:13.080371
5800	6	1-1638	5795	91	66	1756	79	2012-08-02 04:53:13.080371
5795	6	1-1690	5795	66	11	1219	8	2012-08-02 04:53:13.080371
5802	7	1-702	5801	34	72	1377	11	2012-08-02 04:53:13.080371
5803	7	1-1561	5801	72	91	1816	16	2012-08-02 04:53:13.080371
5804	7	1-1619	5801	91	33	1943	35	2012-08-02 04:53:13.080371
5805	7	1-1663	5801	33	11	6041	45	2012-08-02 04:53:13.080371
5806	7	1-1678	5801	11	23	3558	71	2012-08-02 04:53:13.080371
5807	7	1-913	5801	23	66	2744	79	2012-08-02 04:53:13.080371
5801	7	1-1690	5801	66	34	1950	8	2012-08-02 04:53:13.080371
5809	6	1-1450	5808	2	23	309	33	2012-08-02 04:53:13.544515
5810	6	1-1570	5808	23	51	572	93	2012-08-02 04:53:13.544515
5811	6	1-1423	5808	51	42	679	80	2012-08-02 04:53:13.544515
5812	6	1-1598	5808	42	94	115	42	2012-08-02 04:53:13.544515
5813	6	1-1398	5808	94	88	21	67	2012-08-02 04:53:13.544515
5808	6	1-1691	5808	88	2	202	97	2012-08-02 04:53:13.544515
5815	5	1-1450	5814	2	23	966	33	2012-08-02 04:53:13.544515
5816	5	1-1570	5814	23	40	1840	93	2012-08-02 04:53:13.544515
5817	5	1-993	5814	40	94	331	42	2012-08-02 04:53:13.544515
5818	5	1-1398	5814	94	88	61	67	2012-08-02 04:53:13.544515
5814	5	1-1691	5814	88	2	614	97	2012-08-02 04:53:13.544515
5820	6	1-1450	5819	2	8	2662	33	2012-08-02 04:53:13.544515
5821	6	1-1467	5819	8	53	761	66	2012-08-02 04:53:13.544515
5822	6	1-1213	5819	53	48	1390	53	2012-08-02 04:53:13.544515
5823	6	1-1660	5819	48	94	853	42	2012-08-02 04:53:13.544515
5824	6	1-1398	5819	94	88	161	67	2012-08-02 04:53:13.544515
5819	6	1-1691	5819	88	2	1655	97	2012-08-02 04:53:13.544515
5826	3	1-1068	5825	99	92	2278	74	2012-08-02 04:53:16.660226
5827	3	1-1565	5825	92	31	200	7	2012-08-02 04:53:16.660226
5825	3	1-1699	5825	31	99	1824	58	2012-08-02 04:53:16.660226
5829	7	1-1413	5828	64	33	2933	69	2012-08-02 04:53:16.660226
5830	7	1-1179	5828	33	75	737	90	2012-08-02 04:53:16.660226
5831	7	1-1591	5828	75	91	889	26	2012-08-02 04:53:16.660226
5832	7	1-1638	5828	91	66	809	79	2012-08-02 04:53:16.660226
5833	7	1-1690	5828	66	4	560	8	2012-08-02 04:53:16.660226
5834	7	1-4	5828	4	31	501	7	2012-08-02 04:53:16.660226
5828	7	1-1699	5828	31	64	4716	58	2012-08-02 04:53:16.660226
5836	5	1-732	5835	49	44	2017	72	2012-08-02 04:53:16.660226
5837	5	1-1605	5835	44	89	2107	81	2012-08-02 04:53:16.660226
5838	5	1-1318	5835	89	92	1319	74	2012-08-02 04:53:16.660226
5839	5	1-1565	5835	92	31	122	7	2012-08-02 04:53:16.660226
5835	5	1-1699	5835	31	49	1177	58	2012-08-02 04:53:16.660226
5841	6	1-1365	5840	69	47	1551	73	2012-08-02 04:53:17.577085
5842	6	1-859	5840	47	69	478	94	2012-08-02 04:53:17.577085
5843	6	1-1627	5840	69	92	738	83	2012-08-02 04:53:17.577085
5844	6	1-1628	5840	92	43	740	15	2012-08-02 04:53:17.577085
5845	6	1-1243	5840	43	54	2016	62	2012-08-02 04:53:17.577085
5840	6	1-1701	5840	54	69	3491	2	2012-08-02 04:53:17.577085
5847	4	1-186	5846	87	71	582	13	2012-08-02 04:53:17.577085
5848	4	1-1267	5846	71	43	410	15	2012-08-02 04:53:17.577085
5849	4	1-1243	5846	43	54	1369	62	2012-08-02 04:53:17.577085
5846	4	1-1701	5846	54	87	2900	2	2012-08-02 04:53:17.577085
5851	6	1-1094	5850	76	48	3636	53	2012-08-02 04:53:17.577085
5852	6	1-1660	5850	48	94	2160	42	2012-08-02 04:53:17.577085
5853	6	1-1398	5850	94	32	394	67	2012-08-02 04:53:17.577085
5854	6	1-1399	5850	32	55	318	32	2012-08-02 04:53:17.577085
5855	6	1-1646	5850	55	54	1219	62	2012-08-02 04:53:17.577085
5850	6	1-1701	5850	54	76	2957	2	2012-08-02 04:53:17.577085
5857	6	1-1646	5856	55	54	44	62	2012-08-02 04:53:18.494568
5858	6	1-1701	5856	54	87	85	2	2012-08-02 04:53:18.494568
5859	6	1-186	5856	87	13	16	13	2012-08-02 04:53:18.494568
5860	6	1-1669	5856	13	38	13	96	2012-08-02 04:53:18.494568
5861	6	1-1607	5856	38	8	11	39	2012-08-02 04:53:18.494568
5856	6	1-1703	5856	8	55	15	32	2012-08-02 04:53:18.494568
5863	6	1-1646	5862	55	13	2425	62	2012-08-02 04:53:18.494568
5864	6	1-1110	5862	13	75	294	25	2012-08-02 04:53:18.494568
5865	6	1-1647	5862	75	13	309	13	2012-08-02 04:53:18.494568
5866	6	1-1669	5862	13	38	336	96	2012-08-02 04:53:18.494568
5867	6	1-1607	5862	38	8	360	39	2012-08-02 04:53:18.494568
5862	6	1-1703	5862	8	55	624	32	2012-08-02 04:53:18.494568
5869	6	1-1669	5868	13	26	2283	96	2012-08-02 04:53:18.958398
5870	6	1-1672	5868	26	49	3092	82	2012-08-02 04:53:18.958398
5871	6	1-707	5868	49	73	1997	94	2012-08-02 04:53:18.958398
5872	6	1-801	5868	73	32	1290	81	2012-08-02 04:53:18.958398
5873	6	1-1184	5868	32	16	1155	68	2012-08-02 04:53:18.958398
5868	6	1-1704	5868	16	13	2279	13	2012-08-02 04:53:18.958398
5875	6	1-1669	5874	13	41	3344	96	2012-08-02 04:53:18.958398
5876	6	1-1496	5874	41	14	3358	49	2012-08-02 04:53:18.958398
5877	6	1-440	5874	14	73	1955	94	2012-08-02 04:53:18.958398
5878	6	1-801	5874	73	32	1396	81	2012-08-02 04:53:18.958398
5879	6	1-1184	5874	32	16	1383	68	2012-08-02 04:53:18.958398
5874	6	1-1704	5874	16	13	3018	13	2012-08-02 04:53:18.958398
5881	6	1-1253	5880	93	76	154	65	2012-08-02 04:53:19.433578
5882	6	1-1654	5880	76	88	239	50	2012-08-02 04:53:19.433578
5883	6	1-1461	5880	88	51	207	21	2012-08-02 04:53:19.433578
5884	6	1-1457	5880	51	32	145	41	2012-08-02 04:53:19.433578
5885	6	1-36	5880	32	10	159	51	2012-08-02 04:53:19.433578
5880	6	1-1705	5880	10	93	129	49	2012-08-02 04:53:19.433578
5887	6	1-1253	5886	93	39	1184	65	2012-08-02 04:53:19.433578
5888	6	1-1305	5886	39	70	1132	73	2012-08-02 04:53:19.433578
5889	6	1-441	5886	70	28	1382	22	2012-08-02 04:53:19.433578
5890	6	1-1352	5886	28	67	1024	5	2012-08-02 04:53:19.433578
5891	6	1-1635	5886	67	10	1126	51	2012-08-02 04:53:19.433578
5886	6	1-1705	5886	10	93	950	49	2012-08-02 04:53:19.433578
5892	1	1-1461	\N	88	88	1037	21	2012-08-02 04:53:19.433578
5894	5	1-1102	5893	67	23	261	69	2012-08-02 04:53:20.372732
5895	5	1-1340	5893	23	94	334	42	2012-08-02 04:53:20.372732
5896	5	1-1398	5893	94	88	63	67	2012-08-02 04:53:20.372732
5897	5	1-1691	5893	88	22	651	97	2012-08-02 04:53:20.372732
5893	5	1-1707	5893	22	67	697	68	2012-08-02 04:53:20.372732
5899	7	1-1704	5898	16	13	459	13	2012-08-02 04:53:20.372732
5900	7	1-1669	5898	13	94	507	96	2012-08-02 04:53:20.372732
5901	7	1-289	5898	94	64	87	89	2012-08-02 04:53:20.372732
5902	7	1-330	5898	64	94	92	42	2012-08-02 04:53:20.372732
5903	7	1-1398	5898	94	88	18	67	2012-08-02 04:53:20.372732
5904	7	1-1691	5898	88	22	192	97	2012-08-02 04:53:20.372732
5898	7	1-1707	5898	22	16	211	68	2012-08-02 04:53:20.372732
5906	6	1-1619	5905	91	13	47	35	2012-08-02 04:53:20.836771
5907	6	1-581	5905	13	51	116	21	2012-08-02 04:53:20.836771
5908	6	1-1457	5905	51	76	84	41	2012-08-02 04:53:20.836771
5909	6	1-1673	5905	76	16	482	93	2012-08-02 04:53:20.836771
5910	6	1-242	5905	16	17	27	78	2012-08-02 04:53:20.836771
5905	6	1-1708	5905	17	91	47	16	2012-08-02 04:53:20.836771
5912	6	1-1450	5911	2	33	448	33	2012-08-02 04:53:21.279205
5911	6	1-1709	5911	97	2	266	97	2012-08-02 04:53:21.279205
5913	6	1-1353	5911	33	10	467	51	2012-08-02 04:53:21.279205
5914	6	1-1705	5911	10	44	421	49	2012-08-02 04:53:21.279205
5915	6	1-1520	5911	44	19	181	94	2012-08-02 04:53:21.279205
5916	6	1-1387	5911	19	97	326	6	2012-08-02 04:53:21.279205
15964	3	1-969	15963	75	62	690	51	2012-08-02 05:13:20.278162
5918	6	1-992	5917	17	8	509	39	2012-08-02 04:53:22.328328
5919	6	1-1703	5917	8	58	691	32	2012-08-02 04:53:22.328328
5920	6	1-1609	5917	58	51	548	52	2012-08-02 04:53:22.328328
5921	6	1-701	5917	51	47	404	53	2012-08-02 04:53:22.328328
5922	6	1-856	5917	47	93	439	72	2012-08-02 04:53:22.328328
5917	6	1-1711	5917	93	17	345	19	2012-08-02 04:53:22.328328
15965	3	1-4300	15963	62	2	223	67	2012-08-02 05:13:20.278162
5924	6	1-719	5923	44	23	141	71	2012-08-02 04:53:22.328328
5925	6	1-913	5923	23	66	85	79	2012-08-02 04:53:22.328328
5926	6	1-1690	5923	66	11	47	8	2012-08-02 04:53:22.328328
5927	6	1-1483	5923	11	33	175	100	2012-08-02 04:53:22.328328
5928	6	1-1361	5923	33	93	296	72	2012-08-02 04:53:22.328328
5923	6	1-1711	5923	93	44	247	19	2012-08-02 04:53:22.328328
5930	5	1-941	5929	35	63	180	11	2012-08-02 04:53:22.328328
5931	5	1-1648	5929	63	51	417	52	2012-08-02 04:53:22.328328
5932	5	1-701	5929	51	47	346	53	2012-08-02 04:53:22.328328
5933	5	1-856	5929	47	93	422	72	2012-08-02 04:53:22.328328
5929	5	1-1711	5929	93	35	373	19	2012-08-02 04:53:22.328328
5935	5	1-941	5934	35	2	147	11	2012-08-02 04:53:22.328328
5936	5	1-1055	5934	2	24	170	20	2012-08-02 04:53:22.328328
5937	5	1-1465	5934	24	33	157	100	2012-08-02 04:53:22.328328
5938	5	1-1361	5934	33	93	302	72	2012-08-02 04:53:22.328328
5934	5	1-1711	5934	93	35	286	19	2012-08-02 04:53:22.328328
5940	5	1-941	5939	35	2	2730	11	2012-08-02 04:53:22.328328
5941	5	1-1055	5939	2	54	3175	20	2012-08-02 04:53:22.328328
5942	5	1-1443	5939	54	5	6457	94	2012-08-02 04:53:22.328328
5943	5	1-754	5939	5	93	5501	72	2012-08-02 04:53:22.328328
5939	5	1-1711	5939	93	35	5247	19	2012-08-02 04:53:22.328328
5945	5	1-992	5944	17	99	2862	39	2012-08-02 04:53:22.328328
5946	5	1-1054	5944	99	56	1603	45	2012-08-02 04:53:22.328328
5947	5	1-1552	5944	56	5	1470	94	2012-08-02 04:53:22.328328
5948	5	1-754	5944	5	93	1379	72	2012-08-02 04:53:22.328328
5944	5	1-1711	5944	93	17	1449	19	2012-08-02 04:53:22.328328
5949	1	1-992	\N	17	17	1204	39	2012-08-02 04:53:22.328328
5950	1	1-941	\N	35	35	693	11	2012-08-02 04:53:22.328328
5952	4	1-1550	5951	61	17	7239	78	2012-08-02 04:53:23.700336
5953	4	1-1708	5951	17	44	3193	16	2012-08-02 04:53:23.700336
5954	4	1-984	5951	44	78	1239	10	2012-08-02 04:53:23.700336
5951	4	1-1714	5951	78	61	404	38	2012-08-02 04:53:23.700336
5956	5	1-1631	5955	51	75	5727	61	2012-08-02 04:53:23.700336
5957	5	1-700	5955	75	17	2603	78	2012-08-02 04:53:23.700336
5958	5	1-1708	5955	17	44	2078	16	2012-08-02 04:53:23.700336
5959	5	1-984	5955	44	78	1459	10	2012-08-02 04:53:23.700336
5955	5	1-1714	5955	78	51	860	38	2012-08-02 04:53:23.700336
5961	4	1-1631	5960	51	77	3657	61	2012-08-02 04:53:23.700336
5962	4	1-1509	5960	77	6	1210	11	2012-08-02 04:53:23.700336
5963	4	1-6	5960	6	78	500	10	2012-08-02 04:53:23.700336
5960	4	1-1714	5960	78	51	402	38	2012-08-02 04:53:23.700336
5965	2	1-570	5964	48	78	3339	10	2012-08-02 04:53:23.700336
5964	2	1-1714	5964	78	48	3906	38	2012-08-02 04:53:23.700336
5967	3	1-1069	5966	38	70	25	45	2012-08-02 04:53:24.164796
5968	3	1-1558	5966	70	76	233	28	2012-08-02 04:53:24.164796
5966	3	1-1715	5966	76	38	470	15	2012-08-02 04:53:24.164796
5970	5	1-1216	5969	13	58	846	32	2012-08-02 04:53:24.164796
5971	5	1-1609	5969	58	64	814	52	2012-08-02 04:53:24.164796
5972	5	1-1393	5969	64	53	1032	46	2012-08-02 04:53:24.164796
5973	5	1-1206	5969	53	76	1203	28	2012-08-02 04:53:24.164796
5969	5	1-1715	5969	76	13	2997	15	2012-08-02 04:53:24.164796
5975	3	1-1405	5974	33	55	503	32	2012-08-02 04:53:25.070497
5976	3	1-1646	5974	55	34	1208	62	2012-08-02 04:53:25.070497
5974	3	1-1717	5974	34	33	991	52	2012-08-02 04:53:25.070497
5978	5	1-1393	5977	64	96	40	46	2012-08-02 04:53:25.070497
5979	5	1-537	5977	96	84	19	48	2012-08-02 04:53:25.070497
5980	5	1-1439	5977	84	89	20	10	2012-08-02 04:53:25.070497
5981	5	1-1259	5977	89	34	21	62	2012-08-02 04:53:25.070497
5977	5	1-1717	5977	34	64	29	52	2012-08-02 04:53:25.070497
5983	7	1-42	5982	34	78	272	47	2012-08-02 04:53:28.385034
5984	7	1-150	5982	78	17	184	78	2012-08-02 04:53:28.385034
5985	7	1-1708	5982	17	40	255	16	2012-08-02 04:53:28.385034
5986	7	1-1233	5982	40	49	212	82	2012-08-02 04:53:28.385034
5987	7	1-707	5982	49	69	114	94	2012-08-02 04:53:28.385034
5988	7	1-1627	5982	69	90	160	83	2012-08-02 04:53:28.385034
5982	7	1-1725	5982	90	34	430	57	2012-08-02 04:53:28.385034
5990	6	1-1534	5989	8	7	1289	37	2012-08-02 04:53:28.385034
5991	6	1-1667	5989	7	61	534	52	2012-08-02 04:53:28.385034
5992	6	1-1652	5989	61	30	175	71	2012-08-02 04:53:28.385034
5993	6	1-1436	5989	30	69	86	94	2012-08-02 04:53:28.385034
5994	6	1-1627	5989	69	90	125	83	2012-08-02 04:53:28.385034
5989	6	1-1725	5989	90	8	347	57	2012-08-02 04:53:28.385034
5996	4	1-1388	5995	100	32	601	81	2012-08-02 04:53:28.385034
5997	4	1-1184	5995	32	25	573	68	2012-08-02 04:53:28.385034
5998	4	1-1604	5995	25	90	191	83	2012-08-02 04:53:28.385034
5995	4	1-1725	5995	90	100	659	57	2012-08-02 04:53:28.385034
6000	5	1-1224	5999	90	23	2499	69	2012-08-02 04:53:28.385034
6001	5	1-1340	5999	23	32	3137	42	2012-08-02 04:53:28.385034
6002	5	1-649	5999	32	69	533	94	2012-08-02 04:53:28.385034
6003	5	1-1627	5999	69	90	948	83	2012-08-02 04:53:28.385034
5999	5	1-1725	5999	90	90	3327	57	2012-08-02 04:53:28.385034
6004	1	1-559	\N	52	52	366	83	2012-08-02 04:53:28.385034
6006	6	1-996	6005	42	84	2224	74	2012-08-02 04:53:28.848932
6007	6	1-1551	6005	84	53	879	100	2012-08-02 04:53:28.848932
6008	6	1-1720	6005	53	66	830	73	2012-08-02 04:53:28.848932
6009	6	1-706	6005	66	66	238	55	2012-08-02 04:53:28.848932
6010	6	1-1433	6005	66	4	1014	4	2012-08-02 04:53:28.848932
6005	6	1-1726	6005	4	42	4927	21	2012-08-02 04:53:28.848932
6012	6	1-1457	6011	51	76	750	41	2012-08-02 04:53:28.848932
6013	6	1-1673	6011	76	53	3573	93	2012-08-02 04:53:28.848932
6014	6	1-1659	6011	53	10	739	29	2012-08-02 04:53:28.848932
6015	6	1-565	6011	10	85	195	28	2012-08-02 04:53:28.848932
6016	6	1-1719	6011	85	4	241	4	2012-08-02 04:53:28.848932
6011	6	1-1726	6011	4	51	1237	21	2012-08-02 04:53:28.848932
15963	3	1-4596	15963	2	75	4444	66	2012-08-02 05:13:20.278162
6018	6	1-1067	6017	19	73	657	87	2012-08-02 04:53:28.848932
6019	6	1-1653	6017	73	91	118	8	2012-08-02 04:53:28.848932
6020	6	1-1536	6017	91	47	409	86	2012-08-02 04:53:28.848932
6021	6	1-1488	6017	47	85	322	28	2012-08-02 04:53:28.848932
6022	6	1-1719	6017	85	4	401	4	2012-08-02 04:53:28.848932
6017	6	1-1726	6017	4	19	2081	21	2012-08-02 04:53:28.848932
6024	6	1-1591	6023	75	91	1888	26	2012-08-02 04:53:29.887489
6025	6	1-1638	6023	91	31	1220	79	2012-08-02 04:53:29.887489
6026	6	1-1700	6023	31	89	1248	10	2012-08-02 04:53:29.887489
6027	6	1-1259	6023	89	34	900	62	2012-08-02 04:53:29.887489
6028	6	1-1717	6023	34	56	878	52	2012-08-02 04:53:29.887489
6023	6	1-1728	6023	56	75	2207	90	2012-08-02 04:53:29.887489
6030	5	1-1049	6029	15	51	174	69	2012-08-02 04:53:29.887489
6031	5	1-1205	6029	51	58	127	31	2012-08-02 04:53:29.887489
6032	5	1-1480	6029	58	7	114	37	2012-08-02 04:53:29.887489
6033	5	1-1667	6029	7	56	57	52	2012-08-02 04:53:29.887489
6029	5	1-1728	6029	56	15	191	90	2012-08-02 04:53:29.887489
6035	6	1-366	6034	58	78	1171	47	2012-08-02 04:53:31.842909
6036	6	1-150	6034	78	17	739	78	2012-08-02 04:53:31.842909
6037	6	1-1708	6034	17	11	956	16	2012-08-02 04:53:31.842909
6038	6	1-755	6034	11	32	557	3	2012-08-02 04:53:31.842909
6039	6	1-1640	6034	32	69	1377	50	2012-08-02 04:53:31.842909
6034	6	1-1732	6034	69	58	2362	2	2012-08-02 04:53:31.842909
6041	6	1-1619	6040	91	52	511	35	2012-08-02 04:53:32.296277
6042	6	1-99	6040	52	94	185	11	2012-08-02 04:53:32.296277
6043	6	1-1571	6040	94	30	62	45	2012-08-02 04:53:32.296277
6044	6	1-1639	6040	30	51	690	63	2012-08-02 04:53:32.296277
6045	6	1-1567	6040	51	34	553	5	2012-08-02 04:53:32.296277
6040	6	1-1733	6040	34	91	853	16	2012-08-02 04:53:32.296277
6047	4	1-755	6046	11	32	2566	3	2012-08-02 04:53:32.296277
6048	4	1-1640	6046	32	91	7388	50	2012-08-02 04:53:32.296277
6049	4	1-953	6046	91	34	1696	5	2012-08-02 04:53:32.296277
6046	4	1-1733	6046	34	11	3781	16	2012-08-02 04:53:32.296277
6051	6	1-755	6050	11	61	2211	3	2012-08-02 04:53:32.296277
6052	6	1-1211	6050	61	12	1580	64	2012-08-02 04:53:32.296277
6053	6	1-1641	6050	12	13	1094	13	2012-08-02 04:53:32.296277
6054	6	1-1669	6050	13	68	996	96	2012-08-02 04:53:32.296277
6055	6	1-880	6050	68	34	1440	5	2012-08-02 04:53:32.296277
6050	6	1-1733	6050	34	11	3234	16	2012-08-02 04:53:32.296277
6056	1	1-1211	\N	61	61	799	64	2012-08-02 04:53:32.296277
6057	1	1-1669	\N	13	13	388	96	2012-08-02 04:53:32.296277
6059	7	1-1407	6058	85	67	1341	4	2012-08-02 04:53:34.119066
6060	7	1-1643	6058	67	33	1123	33	2012-08-02 04:53:34.119066
6061	7	1-1353	6058	33	10	1172	51	2012-08-02 04:53:34.119066
6062	7	1-1705	6058	10	93	1056	49	2012-08-02 04:53:34.119066
6063	7	1-1253	6058	93	76	1404	65	2012-08-02 04:53:34.119066
6064	7	1-1654	6058	76	38	2422	50	2012-08-02 04:53:34.119066
6058	7	1-1737	6058	38	85	1513	13	2012-08-02 04:53:34.119066
6066	4	1-1482	6065	61	47	876	86	2012-08-02 04:53:39.598482
6067	4	1-1488	6065	47	70	633	28	2012-08-02 04:53:39.598482
6068	4	1-1681	6065	70	41	529	70	2012-08-02 04:53:39.598482
6065	4	1-1750	6065	41	61	2488	42	2012-08-02 04:53:39.598482
6070	7	1-1123	6069	20	54	95	60	2012-08-02 04:53:40.051255
6071	7	1-305	6069	54	78	42	47	2012-08-02 04:53:40.051255
6072	7	1-150	6069	78	17	37	78	2012-08-02 04:53:40.051255
6073	7	1-1708	6069	17	91	67	16	2012-08-02 04:53:40.051255
6074	7	1-1619	6069	91	13	69	35	2012-08-02 04:53:40.051255
6075	7	1-581	6069	13	27	175	21	2012-08-02 04:53:40.051255
6069	7	1-1751	6069	27	20	91	67	2012-08-02 04:53:40.051255
6077	6	1-1253	6076	93	39	530	65	2012-08-02 04:53:40.979233
6078	6	1-1305	6076	39	70	479	73	2012-08-02 04:53:40.979233
6079	6	1-441	6076	70	85	553	22	2012-08-02 04:53:40.979233
6080	6	1-1713	6076	85	23	270	33	2012-08-02 04:53:40.979233
6081	6	1-1570	6076	23	48	487	93	2012-08-02 04:53:40.979233
6076	6	1-1753	6076	48	93	450	49	2012-08-02 04:53:40.979233
6083	6	1-1253	6082	93	34	107	65	2012-08-02 04:53:40.979233
6084	6	1-898	6082	34	36	99	43	2012-08-02 04:53:40.979233
6085	6	1-1526	6082	36	51	22	21	2012-08-02 04:53:40.979233
6086	6	1-1457	6082	51	76	16	41	2012-08-02 04:53:40.979233
6087	6	1-1673	6082	76	48	89	93	2012-08-02 04:53:40.979233
6082	6	1-1753	6082	48	93	87	49	2012-08-02 04:53:40.979233
6089	6	1-1253	6088	93	69	4534	65	2012-08-02 04:53:40.979233
6090	6	1-1575	6088	69	77	2515	57	2012-08-02 04:53:40.979233
6091	6	1-910	6088	77	67	2179	4	2012-08-02 04:53:40.979233
6092	6	1-1643	6088	67	23	1768	33	2012-08-02 04:53:40.979233
6093	6	1-1570	6088	23	48	3485	93	2012-08-02 04:53:40.979233
6088	6	1-1753	6088	48	93	3523	49	2012-08-02 04:53:40.979233
6095	4	1-1022	6094	48	85	1063	22	2012-08-02 04:53:40.979233
6096	4	1-1713	6094	85	23	581	33	2012-08-02 04:53:40.979233
6097	4	1-1570	6094	23	48	1173	93	2012-08-02 04:53:40.979233
6094	4	1-1753	6094	48	48	1223	49	2012-08-02 04:53:40.979233
6099	6	1-1567	6098	51	67	4103	5	2012-08-02 04:53:43.243987
6100	6	1-1635	6098	67	10	3483	51	2012-08-02 04:53:43.243987
6101	6	1-1705	6098	10	21	2270	49	2012-08-02 04:53:43.243987
6102	6	1-39	6098	21	33	707	52	2012-08-02 04:53:43.243987
6103	6	1-1405	6098	33	8	448	32	2012-08-02 04:53:43.243987
6098	6	1-1758	6098	8	51	3953	63	2012-08-02 04:53:43.243987
6105	7	1-1186	6104	36	38	887	50	2012-08-02 04:53:43.243987
6106	7	1-1737	6104	38	95	430	13	2012-08-02 04:53:43.243987
6107	7	1-326	6104	95	89	165	10	2012-08-02 04:53:43.243987
6108	7	1-1259	6104	89	34	134	62	2012-08-02 04:53:43.243987
6109	7	1-1717	6104	34	33	147	52	2012-08-02 04:53:43.243987
6110	7	1-1405	6104	33	8	100	32	2012-08-02 04:53:43.243987
6104	7	1-1758	6104	8	36	948	63	2012-08-02 04:53:43.243987
6112	5	1-1106	6111	4	89	802	10	2012-08-02 04:53:43.243987
6113	5	1-1259	6111	89	34	657	62	2012-08-02 04:53:43.243987
6114	5	1-1717	6111	34	33	729	52	2012-08-02 04:53:43.243987
6115	5	1-1405	6111	33	8	500	32	2012-08-02 04:53:43.243987
6111	5	1-1758	6111	8	4	4782	63	2012-08-02 04:53:43.243987
6117	4	1-1563	6116	43	21	1803	18	2012-08-02 04:53:44.161129
6118	4	1-1738	6116	21	87	4096	2	2012-08-02 04:53:44.161129
6119	4	1-186	6116	87	49	761	13	2012-08-02 04:53:44.161129
6116	4	1-1760	6116	49	43	1687	23	2012-08-02 04:53:44.161129
6121	7	1-1685	6120	70	57	1645	30	2012-08-02 04:53:45.05637
6120	7	1-1762	6120	95	70	2137	91	2012-08-02 04:53:45.05637
6122	7	1-847	6120	57	27	684	21	2012-08-02 04:53:45.05637
6123	7	1-1751	6120	27	88	236	67	2012-08-02 04:53:45.05637
6124	7	1-1691	6120	88	68	1634	97	2012-08-02 04:53:45.05637
6125	7	1-1702	6120	68	40	3159	17	2012-08-02 04:53:45.05637
6126	7	1-1662	6120	40	95	2713	85	2012-08-02 04:53:45.05637
15966	1	1-4300	\N	62	62	30	67	2012-08-02 05:13:20.278162
6128	3	1-598	6127	7	67	934	4	2012-08-02 04:53:45.951164
6129	3	1-1643	6127	67	26	473	33	2012-08-02 04:53:45.951164
6127	3	1-1764	6127	26	7	2597	20	2012-08-02 04:53:45.951164
6131	5	1-1465	6130	24	53	1980	100	2012-08-02 04:53:45.951164
6132	5	1-1720	6130	53	70	1639	73	2012-08-02 04:53:45.951164
6133	5	1-441	6130	70	85	1384	22	2012-08-02 04:53:45.951164
6134	5	1-1713	6130	85	26	495	33	2012-08-02 04:53:45.951164
6130	5	1-1764	6130	26	24	2915	20	2012-08-02 04:53:45.951164
15968	5	1-3556	15967	31	64	24	34	2012-08-02 05:13:22.730917
6136	7	1-1193	6135	2	38	354	50	2012-08-02 04:53:45.951164
6137	7	1-1737	6135	38	49	170	13	2012-08-02 04:53:45.951164
6138	7	1-1760	6135	49	43	388	23	2012-08-02 04:53:45.951164
6139	7	1-1563	6135	43	68	426	18	2012-08-02 04:53:45.951164
6140	7	1-1440	6135	68	67	649	4	2012-08-02 04:53:45.951164
6141	7	1-1643	6135	67	26	418	33	2012-08-02 04:53:45.951164
6135	7	1-1764	6135	26	2	2924	20	2012-08-02 04:53:45.951164
6143	2	1-1323	6142	92	30	2381	13	2012-08-02 04:53:47.762743
6142	2	1-1768	6142	30	92	4105	24	2012-08-02 04:53:47.762743
6145	6	1-1720	6144	53	47	203	73	2012-08-02 04:53:49.166402
6146	6	1-859	6144	47	69	57	94	2012-08-02 04:53:49.166402
6147	6	1-1627	6144	69	90	64	83	2012-08-02 04:53:49.166402
6148	6	1-1725	6144	90	28	139	57	2012-08-02 04:53:49.166402
6149	6	1-1535	6144	28	92	434	40	2012-08-02 04:53:49.166402
6144	6	1-1771	6144	92	53	266	100	2012-08-02 04:53:49.166402
6151	6	1-1769	6150	60	32	3143	42	2012-08-02 04:53:49.166402
6152	6	1-649	6150	32	69	351	94	2012-08-02 04:53:49.166402
6153	6	1-1627	6150	69	90	411	83	2012-08-02 04:53:49.166402
6154	6	1-1725	6150	90	28	918	57	2012-08-02 04:53:49.166402
6155	6	1-1535	6150	28	92	2965	40	2012-08-02 04:53:49.166402
6150	6	1-1771	6150	92	60	1879	100	2012-08-02 04:53:49.166402
6157	5	1-1769	6156	60	61	825	42	2012-08-02 04:53:49.166402
6158	5	1-1482	6156	61	56	319	86	2012-08-02 04:53:49.166402
6159	5	1-1595	6156	56	99	233	77	2012-08-02 04:53:49.166402
6160	5	1-1767	6156	99	92	469	40	2012-08-02 04:53:49.166402
6156	5	1-1771	6156	92	60	383	100	2012-08-02 04:53:49.166402
6162	7	1-1531	6161	64	73	558	87	2012-08-02 04:53:49.166402
6163	7	1-1653	6161	73	64	116	8	2012-08-02 04:53:49.166402
6164	7	1-1661	6161	64	59	1672	54	2012-08-02 04:53:49.166402
6165	7	1-555	6161	59	90	60	83	2012-08-02 04:53:49.166402
6166	7	1-1725	6161	90	28	198	57	2012-08-02 04:53:49.166402
6167	7	1-1535	6161	28	92	944	40	2012-08-02 04:53:49.166402
6161	7	1-1771	6161	92	64	884	100	2012-08-02 04:53:49.166402
6169	6	1-1531	6168	64	73	1820	87	2012-08-02 04:53:49.166402
6170	6	1-1653	6168	73	91	384	8	2012-08-02 04:53:49.166402
6171	6	1-1536	6168	91	56	1566	86	2012-08-02 04:53:49.166402
6172	6	1-1595	6168	56	99	1331	77	2012-08-02 04:53:49.166402
6173	6	1-1767	6168	99	92	3126	40	2012-08-02 04:53:49.166402
6168	6	1-1771	6168	92	64	2975	100	2012-08-02 04:53:49.166402
6174	1	1-965	\N	75	75	4284	54	2012-08-02 04:53:49.166402
6176	5	1-1647	6175	75	49	1040	13	2012-08-02 04:53:49.644863
6177	5	1-1760	6175	49	65	2697	23	2012-08-02 04:53:49.644863
6178	5	1-1061	6175	65	47	1152	45	2012-08-02 04:53:49.644863
6179	5	1-1105	6175	47	33	694	48	2012-08-02 04:53:49.644863
6175	5	1-1772	6175	33	75	1060	25	2012-08-02 04:53:49.644863
6181	4	1-1333	6180	1	71	1152	12	2012-08-02 04:53:49.644863
6182	4	1-1564	6180	71	87	1652	77	2012-08-02 04:53:49.644863
6183	4	1-1632	6180	87	33	1459	48	2012-08-02 04:53:49.644863
6180	4	1-1772	6180	33	1	2473	25	2012-08-02 04:53:49.644863
6185	7	1-1647	6184	75	85	2628	13	2012-08-02 04:53:49.644863
6186	7	1-1407	6184	85	67	2319	4	2012-08-02 04:53:49.644863
6187	7	1-1643	6184	67	7	1935	33	2012-08-02 04:53:49.644863
6188	7	1-1739	6184	7	54	1994	59	2012-08-02 04:53:49.644863
6189	7	1-1255	6184	54	87	1497	77	2012-08-02 04:53:49.644863
6190	7	1-1632	6184	87	33	1356	48	2012-08-02 04:53:49.644863
6184	7	1-1772	6184	33	75	2356	25	2012-08-02 04:53:49.644863
6192	7	1-1653	6191	73	91	194	8	2012-08-02 04:53:51.031792
6193	7	1-1536	6191	91	1	621	86	2012-08-02 04:53:51.031792
6194	7	1-717	6191	1	49	352	13	2012-08-02 04:53:51.031792
6195	7	1-1760	6191	49	43	760	23	2012-08-02 04:53:51.031792
6196	7	1-1563	6191	43	21	791	18	2012-08-02 04:53:51.031792
6197	7	1-1738	6191	21	59	1750	2	2012-08-02 04:53:51.031792
6191	7	1-1775	6191	59	73	1169	87	2012-08-02 04:53:51.031792
6199	6	1-1087	6198	29	25	472	68	2012-08-02 04:53:53.27504
6200	6	1-1604	6198	25	90	137	83	2012-08-02 04:53:53.27504
6201	6	1-1725	6198	90	90	449	57	2012-08-02 04:53:53.27504
6202	6	1-1224	6198	90	96	316	69	2012-08-02 04:53:53.27504
6203	6	1-1597	6198	96	18	180	1	2012-08-02 04:53:53.27504
6198	6	1-1780	6198	18	29	783	10	2012-08-02 04:53:53.27504
6205	7	1-1259	6204	89	34	981	62	2012-08-02 04:53:53.27504
6206	7	1-1717	6204	34	83	1310	52	2012-08-02 04:53:53.27504
6207	7	1-1404	6204	83	54	402	98	2012-08-02 04:53:53.27504
6208	7	1-631	6204	54	6	287	22	2012-08-02 04:53:53.27504
6209	7	1-1578	6204	6	96	307	69	2012-08-02 04:53:53.27504
6210	7	1-1597	6204	96	18	200	1	2012-08-02 04:53:53.27504
6204	7	1-1780	6204	18	89	995	10	2012-08-02 04:53:53.27504
6212	7	1-1087	6211	29	25	346	68	2012-08-02 04:53:53.27504
6213	7	1-1604	6211	25	51	120	83	2012-08-02 04:53:53.27504
6214	7	1-89	6211	51	82	170	34	2012-08-02 04:53:53.27504
6215	7	1-980	6211	82	6	120	22	2012-08-02 04:53:53.27504
6216	7	1-1578	6211	6	96	135	69	2012-08-02 04:53:53.27504
6217	7	1-1597	6211	96	18	92	1	2012-08-02 04:53:53.27504
6211	7	1-1780	6211	18	29	479	10	2012-08-02 04:53:53.27504
6219	2	1-1694	6218	32	54	3313	3	2012-08-02 04:53:55.087733
6218	2	1-1784	6218	54	32	3375	11	2012-08-02 04:53:55.087733
6221	3	1-1561	6220	72	11	2939	16	2012-08-02 04:53:55.087733
6222	3	1-755	6220	11	54	2234	3	2012-08-02 04:53:55.087733
6220	3	1-1784	6220	54	72	2461	11	2012-08-02 04:53:55.087733
6224	5	1-1505	6223	86	87	1703	44	2012-08-02 04:53:55.087733
6225	5	1-1446	6223	87	56	3057	65	2012-08-02 04:53:55.087733
6226	5	1-1427	6223	56	99	4222	31	2012-08-02 04:53:55.087733
6227	5	1-1776	6223	99	54	1595	3	2012-08-02 04:53:55.087733
6223	5	1-1784	6223	54	86	1944	11	2012-08-02 04:53:55.087733
15969	5	1-4343	15967	64	20	33	24	2012-08-02 05:13:22.730917
6229	6	1-1104	6228	65	82	895	30	2012-08-02 04:53:55.994277
6230	6	1-1091	6228	82	91	431	16	2012-08-02 04:53:55.994277
6231	6	1-1619	6228	91	13	463	35	2012-08-02 04:53:55.994277
6232	6	1-581	6228	13	59	1216	21	2012-08-02 04:53:55.994277
6233	6	1-1129	6228	59	69	1064	93	2012-08-02 04:53:55.994277
6228	6	1-1786	6228	69	65	1829	95	2012-08-02 04:53:55.994277
6234	1	1-1619	\N	91	91	2022	35	2012-08-02 04:53:55.994277
15970	5	1-4597	15967	20	4	85	56	2012-08-02 05:13:22.730917
6236	4	1-1735	6235	26	86	4838	39	2012-08-02 04:53:56.889066
6237	4	1-592	6235	86	99	654	66	2012-08-02 04:53:56.889066
6238	4	1-1321	6235	99	48	412	21	2012-08-02 04:53:56.889066
6235	4	1-1788	6235	48	26	4642	96	2012-08-02 04:53:56.889066
15971	5	1-3014	15967	4	76	54	54	2012-08-02 05:13:22.730917
6240	4	1-1607	6239	38	86	762	39	2012-08-02 04:53:56.889066
6241	4	1-592	6239	86	99	121	66	2012-08-02 04:53:56.889066
6242	4	1-1321	6239	99	48	90	21	2012-08-02 04:53:56.889066
6239	4	1-1788	6239	48	38	1190	96	2012-08-02 04:53:56.889066
6244	7	1-1789	6243	86	75	3623	25	2012-08-02 04:53:58.725
6245	7	1-1647	6243	75	25	4004	13	2012-08-02 04:53:58.725
6246	7	1-1151	6243	25	59	645	2	2012-08-02 04:53:58.725
6247	7	1-1775	6243	59	73	583	87	2012-08-02 04:53:58.725
6248	7	1-1653	6243	73	64	131	8	2012-08-02 04:53:58.725
6249	7	1-1661	6243	64	84	1947	54	2012-08-02 04:53:58.725
6243	7	1-1792	6243	84	86	3841	14	2012-08-02 04:53:58.725
6251	5	1-1594	6250	66	59	3866	2	2012-08-02 04:54:01.51251
6252	5	1-1775	6250	59	73	2590	87	2012-08-02 04:54:01.51251
6253	5	1-1653	6250	73	91	431	8	2012-08-02 04:54:01.51251
6254	5	1-1536	6250	91	95	1383	86	2012-08-02 04:54:01.51251
6250	5	1-1798	6250	95	66	5615	27	2012-08-02 04:54:01.51251
6256	4	1-503	6255	100	73	1426	87	2012-08-02 04:54:01.51251
6257	4	1-1653	6255	73	91	249	8	2012-08-02 04:54:01.51251
6258	4	1-1536	6255	91	95	839	86	2012-08-02 04:54:01.51251
6255	4	1-1798	6255	95	100	3575	27	2012-08-02 04:54:01.51251
6259	1	1-1653	\N	73	73	25	8	2012-08-02 04:54:01.51251
6261	5	1-1441	6260	79	99	3554	39	2012-08-02 04:54:01.976318
6262	5	1-1054	6260	99	14	1204	45	2012-08-02 04:54:01.976318
6263	5	1-1795	6260	14	68	1553	35	2012-08-02 04:54:01.976318
6264	5	1-123	6260	68	83	1401	93	2012-08-02 04:54:01.976318
6260	5	1-1799	6260	83	79	5035	44	2012-08-02 04:54:01.976318
6266	7	1-1685	6265	70	57	865	30	2012-08-02 04:54:03.346444
6267	7	1-847	6265	57	27	412	21	2012-08-02 04:54:03.346444
6268	7	1-1751	6265	27	88	163	67	2012-08-02 04:54:03.346444
6269	7	1-1691	6265	88	68	1294	97	2012-08-02 04:54:03.346444
6270	7	1-1702	6265	68	40	2868	17	2012-08-02 04:54:03.346444
6271	7	1-1662	6265	40	3	2823	85	2012-08-02 04:54:03.346444
6265	7	1-1802	6265	3	70	980	91	2012-08-02 04:54:03.346444
6273	6	1-638	6272	100	79	1991	44	2012-08-02 04:54:03.346444
6274	6	1-1441	6272	79	99	2107	39	2012-08-02 04:54:03.346444
6275	6	1-1054	6272	99	14	1071	45	2012-08-02 04:54:03.346444
6276	6	1-1795	6272	14	90	2072	35	2012-08-02 04:54:03.346444
6277	6	1-1210	6272	90	3	3935	85	2012-08-02 04:54:03.346444
6272	6	1-1802	6272	3	100	1652	91	2012-08-02 04:54:03.346444
6279	6	1-1435	6278	32	1	1458	26	2012-08-02 04:54:03.346444
6280	6	1-1503	6278	1	48	2455	49	2012-08-02 04:54:03.346444
6281	6	1-1022	6278	48	6	1972	22	2012-08-02 04:54:03.346444
6282	6	1-1578	6278	6	63	2067	69	2012-08-02 04:54:03.346444
6283	6	1-1453	6278	63	3	2495	85	2012-08-02 04:54:03.346444
6278	6	1-1802	6278	3	32	1085	91	2012-08-02 04:54:03.346444
6285	7	1-638	6284	100	63	528	44	2012-08-02 04:54:03.346444
6286	7	1-1546	6284	63	26	731	30	2012-08-02 04:54:03.346444
6287	7	1-1686	6284	26	94	368	11	2012-08-02 04:54:03.346444
6288	7	1-1571	6284	94	14	212	45	2012-08-02 04:54:03.346444
6289	7	1-1795	6284	14	90	441	35	2012-08-02 04:54:03.346444
6290	7	1-1210	6284	90	3	901	85	2012-08-02 04:54:03.346444
6284	7	1-1802	6284	3	100	407	91	2012-08-02 04:54:03.346444
6292	6	1-1795	6291	14	13	1944	35	2012-08-02 04:54:04.263953
6293	6	1-581	6291	13	27	4055	21	2012-08-02 04:54:04.263953
6294	6	1-1751	6291	27	20	1736	67	2012-08-02 04:54:04.263953
6295	6	1-1123	6291	20	9	1497	60	2012-08-02 04:54:04.263953
6296	6	1-46	6291	9	23	1850	31	2012-08-02 04:54:04.263953
6291	6	1-1804	6291	23	14	1122	45	2012-08-02 04:54:04.263953
6298	6	1-1795	6297	14	13	520	35	2012-08-02 04:54:04.263953
6299	6	1-581	6297	13	27	1123	21	2012-08-02 04:54:04.263953
6300	6	1-1751	6297	27	12	498	67	2012-08-02 04:54:04.263953
6301	6	1-1309	6297	12	56	406	65	2012-08-02 04:54:04.263953
6302	6	1-1427	6297	56	23	462	31	2012-08-02 04:54:04.263953
6297	6	1-1804	6297	23	14	290	45	2012-08-02 04:54:04.263953
6304	6	1-1552	6303	56	19	212	94	2012-08-02 04:54:04.263953
6305	6	1-1387	6303	19	97	322	6	2012-08-02 04:54:04.263953
6306	6	1-1709	6303	97	68	222	97	2012-08-02 04:54:04.263953
6307	6	1-1702	6303	68	68	616	17	2012-08-02 04:54:04.263953
6308	6	1-1649	6303	68	23	417	31	2012-08-02 04:54:04.263953
6303	6	1-1804	6303	23	56	268	45	2012-08-02 04:54:04.263953
6310	6	1-1795	6309	14	52	1194	35	2012-08-02 04:54:04.263953
6311	6	1-99	6309	52	86	672	11	2012-08-02 04:54:04.263953
6312	6	1-1505	6309	86	87	511	44	2012-08-02 04:54:04.263953
6313	6	1-1446	6309	87	56	797	65	2012-08-02 04:54:04.263953
6314	6	1-1427	6309	56	23	956	31	2012-08-02 04:54:04.263953
6309	6	1-1804	6309	23	14	632	45	2012-08-02 04:54:04.263953
6315	1	1-46	\N	9	9	1721	31	2012-08-02 04:54:04.263953
6317	6	1-1738	6316	21	58	219	2	2012-08-02 04:54:05.674368
6318	6	1-366	6316	58	33	122	47	2012-08-02 04:54:05.674368
6319	6	1-931	6316	33	97	74	6	2012-08-02 04:54:05.674368
6320	6	1-1709	6316	97	68	47	97	2012-08-02 04:54:05.674368
6321	6	1-1702	6316	68	15	110	17	2012-08-02 04:54:05.674368
6316	6	1-1807	6316	15	21	93	18	2012-08-02 04:54:05.674368
6323	4	1-1440	6322	68	67	1654	4	2012-08-02 04:54:05.674368
6324	4	1-1643	6322	67	25	1106	33	2012-08-02 04:54:05.674368
6325	4	1-1651	6322	25	15	1210	17	2012-08-02 04:54:05.674368
6322	4	1-1807	6322	15	68	1047	18	2012-08-02 04:54:05.674368
6327	3	1-1440	6326	68	79	6707	4	2012-08-02 04:54:05.674368
6328	3	1-1362	6326	79	15	4704	17	2012-08-02 04:54:05.674368
6326	3	1-1807	6326	15	68	4156	18	2012-08-02 04:54:05.674368
15967	5	1-4602	15967	76	31	28	92	2012-08-02 05:13:22.730917
6330	5	1-1144	6329	56	51	1173	76	2012-08-02 04:54:07.010909
6331	5	1-1052	6329	51	97	973	6	2012-08-02 04:54:07.010909
6332	5	1-1709	6329	97	68	570	97	2012-08-02 04:54:07.010909
6333	5	1-1702	6329	68	45	1233	17	2012-08-02 04:54:07.010909
6329	5	1-1810	6329	45	56	1269	73	2012-08-02 04:54:07.010909
6335	5	1-859	6334	47	19	837	94	2012-08-02 04:54:07.010909
6336	5	1-1387	6334	19	97	1212	6	2012-08-02 04:54:07.010909
6337	5	1-1709	6334	97	68	796	97	2012-08-02 04:54:07.010909
6338	5	1-1702	6334	68	45	1931	17	2012-08-02 04:54:07.010909
6334	5	1-1810	6334	45	47	2229	73	2012-08-02 04:54:07.010909
15973	5	1-1595	15972	56	30	353	77	2012-08-02 05:13:23.968085
6340	6	1-706	6339	66	94	702	55	2012-08-02 04:54:07.010909
6341	6	1-1745	6339	94	71	2110	12	2012-08-02 04:54:07.010909
6342	6	1-1564	6339	71	78	3075	77	2012-08-02 04:54:07.010909
6343	6	1-792	6339	78	20	1037	24	2012-08-02 04:54:07.010909
6344	6	1-739	6339	20	45	1509	17	2012-08-02 04:54:07.010909
6339	6	1-1810	6339	45	66	2134	73	2012-08-02 04:54:07.010909
15974	5	1-4532	15972	30	24	501	74	2012-08-02 05:13:23.968085
6346	7	1-1601	6345	31	7	1395	43	2012-08-02 04:54:07.475009
6347	7	1-1782	6345	7	27	447	21	2012-08-02 04:54:07.475009
6348	7	1-1751	6345	27	88	159	67	2012-08-02 04:54:07.475009
6349	7	1-1691	6345	88	61	1134	97	2012-08-02 04:54:07.475009
6350	7	1-1763	6345	61	53	1063	9	2012-08-02 04:54:07.475009
6351	7	1-1664	6345	53	6	1067	54	2012-08-02 04:54:07.475009
6345	7	1-1811	6345	6	31	840	16	2012-08-02 04:54:07.475009
6353	3	1-1422	6352	9	22	6032	97	2012-08-02 04:54:08.425326
6354	3	1-1707	6352	22	45	5819	68	2012-08-02 04:54:08.425326
6352	3	1-1813	6352	45	9	3671	8	2012-08-02 04:54:08.425326
6356	5	1-702	6355	34	74	491	11	2012-08-02 04:54:08.425326
6357	5	1-1164	6355	74	18	294	1	2012-08-02 04:54:08.425326
6358	5	1-1780	6355	18	29	1484	10	2012-08-02 04:54:08.425326
6359	5	1-1087	6355	29	45	1039	68	2012-08-02 04:54:08.425326
6355	5	1-1813	6355	45	34	726	8	2012-08-02 04:54:08.425326
6361	7	1-1738	6360	21	59	1131	2	2012-08-02 04:54:10.236968
6362	7	1-1775	6360	59	42	882	87	2012-08-02 04:54:10.236968
6363	7	1-737	6360	42	85	290	78	2012-08-02 04:54:10.236968
6364	7	1-1779	6360	85	35	101	56	2012-08-02 04:54:10.236968
6365	7	1-1774	6360	35	87	86	44	2012-08-02 04:54:10.236968
6366	7	1-1446	6360	87	64	131	65	2012-08-02 04:54:10.236968
6360	7	1-1817	6360	64	21	438	18	2012-08-02 04:54:10.236968
6368	3	1-1786	6367	69	79	6700	95	2012-08-02 04:54:13.45176
6369	3	1-1127	6367	79	75	2752	86	2012-08-02 04:54:13.45176
6367	3	1-1825	6367	75	69	4553	93	2012-08-02 04:54:13.45176
6371	6	1-1664	6370	53	62	2514	54	2012-08-02 04:54:16.19132
6372	6	1-654	6370	62	33	411	28	2012-08-02 04:54:16.19132
6373	6	1-1828	6370	33	54	260	60	2012-08-02 04:54:16.19132
6374	6	1-305	6370	54	56	99	47	2012-08-02 04:54:16.19132
6375	6	1-1743	6370	56	88	437	81	2012-08-02 04:54:16.19132
6370	6	1-1831	6370	88	53	1982	9	2012-08-02 04:54:16.19132
6377	5	1-1738	6376	21	59	914	2	2012-08-02 04:54:16.633417
6378	5	1-1775	6376	59	98	605	87	2012-08-02 04:54:16.633417
6379	5	1-1512	6376	98	36	174	29	2012-08-02 04:54:16.633417
6380	5	1-1823	6376	36	72	197	41	2012-08-02 04:54:16.633417
6376	5	1-1832	6376	72	21	417	18	2012-08-02 04:54:16.633417
6382	3	1-1684	6381	46	90	2523	23	2012-08-02 04:54:16.633417
6383	3	1-1754	6381	90	72	3325	41	2012-08-02 04:54:16.633417
6381	3	1-1832	6381	72	46	7301	18	2012-08-02 04:54:16.633417
6385	5	1-1738	6384	21	74	1798	2	2012-08-02 04:54:16.633417
6386	5	1-1370	6384	74	7	1023	43	2012-08-02 04:54:16.633417
6387	5	1-1782	6384	7	51	422	21	2012-08-02 04:54:16.633417
6388	5	1-1457	6384	51	72	279	41	2012-08-02 04:54:16.633417
6384	5	1-1832	6384	72	21	696	18	2012-08-02 04:54:16.633417
6390	6	1-1088	6389	97	13	1440	70	2012-08-02 04:54:16.633417
6391	6	1-1478	6389	13	6	212	66	2012-08-02 04:54:16.633417
6392	6	1-1734	6389	6	17	1105	36	2012-08-02 04:54:16.633417
6393	6	1-306	6389	17	90	258	23	2012-08-02 04:54:16.633417
6394	6	1-1754	6389	90	72	451	41	2012-08-02 04:54:16.633417
6389	6	1-1832	6389	72	97	1318	18	2012-08-02 04:54:16.633417
6396	6	1-1320	6395	69	70	1286	32	2012-08-02 04:54:17.09766
6397	6	1-1610	6395	70	6	1251	54	2012-08-02 04:54:17.09766
6398	6	1-1811	6395	6	31	1340	16	2012-08-02 04:54:17.09766
6399	6	1-1601	6395	31	7	3029	43	2012-08-02 04:54:17.09766
6400	6	1-1782	6395	7	86	1322	21	2012-08-02 04:54:17.09766
6395	6	1-1833	6395	86	69	1308	60	2012-08-02 04:54:17.09766
6402	6	1-1628	6401	92	15	43	15	2012-08-02 04:54:18.036602
6403	6	1-1808	6401	15	62	20	3	2012-08-02 04:54:18.036602
6404	6	1-1636	6401	62	94	18	55	2012-08-02 04:54:18.036602
6405	6	1-1745	6401	94	42	50	12	2012-08-02 04:54:18.036602
6406	6	1-1693	6401	42	96	48	72	2012-08-02 04:54:18.036602
6401	6	1-1835	6401	96	92	31	83	2012-08-02 04:54:18.036602
6408	2	1-954	6407	90	15	580	34	2012-08-02 04:54:18.478502
6407	2	1-1836	6407	15	90	2488	89	2012-08-02 04:54:18.478502
6410	3	1-99	6409	52	97	640	11	2012-08-02 04:54:18.942572
6411	3	1-1790	6409	97	24	1088	59	2012-08-02 04:54:18.942572
6409	3	1-1837	6409	24	52	1295	35	2012-08-02 04:54:18.942572
6413	6	1-1021	6412	60	6	406	66	2012-08-02 04:54:18.942572
6414	6	1-1734	6412	6	47	1722	36	2012-08-02 04:54:18.942572
6415	6	1-977	6412	47	77	992	64	2012-08-02 04:54:18.942572
6416	6	1-1633	6412	77	7	769	33	2012-08-02 04:54:18.942572
6417	6	1-1739	6412	7	24	646	59	2012-08-02 04:54:18.942572
6412	6	1-1837	6412	24	60	817	35	2012-08-02 04:54:18.942572
6419	6	1-123	6418	68	69	336	93	2012-08-02 04:54:18.942572
6420	6	1-1786	6418	69	83	505	95	2012-08-02 04:54:18.942572
6421	6	1-1781	6418	83	15	418	15	2012-08-02 04:54:18.942572
6422	6	1-1808	6418	15	60	193	3	2012-08-02 04:54:18.942572
6423	6	1-1359	6418	60	24	184	59	2012-08-02 04:54:18.942572
6418	6	1-1837	6418	24	68	252	35	2012-08-02 04:54:18.942572
6425	7	1-123	6424	68	48	818	93	2012-08-02 04:54:18.942572
6426	7	1-1753	6424	48	48	822	49	2012-08-02 04:54:18.942572
6427	7	1-1022	6424	48	64	688	22	2012-08-02 04:54:18.942572
6428	7	1-1834	6424	64	63	243	44	2012-08-02 04:54:18.942572
6429	7	1-1546	6424	63	95	324	30	2012-08-02 04:54:18.942572
6430	7	1-1277	6424	95	24	407	59	2012-08-02 04:54:18.942572
6424	7	1-1837	6424	24	68	585	35	2012-08-02 04:54:18.942572
15975	5	1-4217	15972	24	17	167	9	2012-08-02 05:13:23.968085
6432	7	1-809	6431	73	49	534	7	2012-08-02 04:54:18.942572
6433	7	1-1499	6431	49	7	421	78	2012-08-02 04:54:18.942572
6434	7	1-1525	6431	7	76	511	28	2012-08-02 04:54:18.942572
6435	7	1-1715	6431	76	15	1366	15	2012-08-02 04:54:18.942572
6436	7	1-1808	6431	15	60	689	3	2012-08-02 04:54:18.942572
6437	7	1-1359	6431	60	24	714	59	2012-08-02 04:54:18.942572
6431	7	1-1837	6431	24	73	1065	35	2012-08-02 04:54:18.942572
15976	5	1-4578	15972	17	10	168	18	2012-08-02 05:13:23.968085
6439	4	1-1087	6438	29	39	139	68	2012-08-02 04:54:19.439924
6440	4	1-1346	6438	39	88	71	81	2012-08-02 04:54:19.439924
6441	4	1-1831	6438	88	26	257	9	2012-08-02 04:54:19.439924
6438	4	1-1838	6438	26	29	285	10	2012-08-02 04:54:19.439924
6443	7	1-1798	6442	95	100	441	27	2012-08-02 04:54:20.334705
6444	7	1-503	6442	100	42	225	87	2012-08-02 04:54:20.334705
6445	7	1-737	6442	42	7	85	78	2012-08-02 04:54:20.334705
6446	7	1-1525	6442	7	33	107	28	2012-08-02 04:54:20.334705
6447	7	1-1828	6442	33	69	79	60	2012-08-02 04:54:20.334705
6448	7	1-1320	6442	69	59	85	32	2012-08-02 04:54:20.334705
6442	7	1-1840	6442	59	95	81	86	2012-08-02 04:54:20.334705
6450	5	1-1745	6449	94	42	1899	12	2012-08-02 04:54:20.79888
6451	5	1-1693	6449	42	44	929	72	2012-08-02 04:54:20.79888
6452	5	1-1605	6449	44	88	449	81	2012-08-02 04:54:20.79888
6453	5	1-1831	6449	88	1	1123	9	2012-08-02 04:54:20.79888
6449	5	1-1841	6449	1	94	1353	55	2012-08-02 04:54:20.79888
6455	4	1-465	6454	9	56	583	47	2012-08-02 04:54:20.79888
6456	4	1-1743	6454	56	88	1459	81	2012-08-02 04:54:20.79888
6457	4	1-1831	6454	88	1	3769	9	2012-08-02 04:54:20.79888
6454	4	1-1841	6454	1	9	4692	55	2012-08-02 04:54:20.79888
6459	6	1-1839	6458	63	56	186	52	2012-08-02 04:54:21.682801
6460	6	1-1728	6458	56	84	501	90	2012-08-02 04:54:21.682801
6461	6	1-334	6458	84	86	301	14	2012-08-02 04:54:21.682801
6462	6	1-1789	6458	86	42	213	25	2012-08-02 04:54:21.682801
6463	6	1-1800	6458	42	44	1639	40	2012-08-02 04:54:21.682801
6458	6	1-1843	6458	44	63	813	70	2012-08-02 04:54:21.682801
6465	5	1-1842	6464	6	7	1399	37	2012-08-02 04:54:21.682801
6466	5	1-1667	6464	7	33	738	52	2012-08-02 04:54:21.682801
6467	5	1-1405	6464	33	36	632	32	2012-08-02 04:54:21.682801
6468	5	1-1270	6464	36	44	2475	40	2012-08-02 04:54:21.682801
6464	5	1-1843	6464	44	6	1622	70	2012-08-02 04:54:21.682801
6470	4	1-1420	6469	98	15	4180	98	2012-08-02 04:54:23.47259
6471	4	1-1846	6469	15	59	4188	32	2012-08-02 04:54:23.47259
6472	4	1-1840	6469	59	28	3581	86	2012-08-02 04:54:23.47259
6469	4	1-1847	6469	28	98	5226	58	2012-08-02 04:54:23.47259
6474	2	1-1597	6473	96	14	253	1	2012-08-02 04:54:24.378692
6473	2	1-1849	6473	14	96	386	69	2012-08-02 04:54:24.378692
6476	7	1-1827	6475	63	66	337	6	2012-08-02 04:54:25.750015
6477	7	1-766	6475	66	88	81	81	2012-08-02 04:54:25.750015
6478	7	1-1831	6475	88	53	324	9	2012-08-02 04:54:25.750015
6479	7	1-1664	6475	53	6	363	54	2012-08-02 04:54:25.750015
6480	7	1-1811	6475	6	91	318	16	2012-08-02 04:54:25.750015
6481	7	1-1059	6475	91	20	451	29	2012-08-02 04:54:25.750015
6475	7	1-1853	6475	20	63	371	32	2012-08-02 04:54:25.750015
6483	4	1-1610	6482	70	6	559	54	2012-08-02 04:54:25.750015
6484	4	1-1811	6482	6	91	516	16	2012-08-02 04:54:25.750015
6485	4	1-1059	6482	91	20	771	29	2012-08-02 04:54:25.750015
6482	4	1-1853	6482	20	70	667	32	2012-08-02 04:54:25.750015
6487	2	1-1530	6486	42	66	2821	85	2012-08-02 04:54:27.109223
6486	2	1-1856	6486	66	42	4767	42	2012-08-02 04:54:27.109223
6489	7	1-1482	6488	61	28	2090	86	2012-08-02 04:54:27.109223
6490	7	1-1847	6488	28	98	3260	58	2012-08-02 04:54:27.109223
6491	7	1-1420	6488	98	54	2787	98	2012-08-02 04:54:27.109223
6492	7	1-631	6488	54	6	1980	22	2012-08-02 04:54:27.109223
6493	7	1-1578	6488	6	63	2111	69	2012-08-02 04:54:27.109223
6494	7	1-1453	6488	63	66	2593	85	2012-08-02 04:54:27.109223
6488	7	1-1856	6488	66	61	4580	42	2012-08-02 04:54:27.109223
6496	5	1-1482	6495	61	28	46	86	2012-08-02 04:54:27.109223
6497	5	1-1847	6495	28	64	73	58	2012-08-02 04:54:27.109223
6498	5	1-1413	6495	64	63	44	69	2012-08-02 04:54:27.109223
6499	5	1-1453	6495	63	66	55	85	2012-08-02 04:54:27.109223
6495	5	1-1856	6495	66	61	99	42	2012-08-02 04:54:27.109223
6501	4	1-1737	6500	38	1	2393	13	2012-08-02 04:54:29.882214
6502	4	1-1689	6500	1	75	603	1	2012-08-02 04:54:29.882214
6503	4	1-431	6500	75	79	1669	17	2012-08-02 04:54:29.882214
6500	4	1-1862	6500	79	38	3934	50	2012-08-02 04:54:29.882214
6505	6	1-1505	6504	86	12	28	44	2012-08-02 04:54:30.765957
6506	6	1-1438	6504	12	85	59	89	2012-08-02 04:54:30.765957
6507	6	1-1165	6504	85	79	103	4	2012-08-02 04:54:30.765957
6508	6	1-1362	6504	79	45	83	17	2012-08-02 04:54:30.765957
6509	6	1-1810	6504	45	56	111	73	2012-08-02 04:54:30.765957
6504	6	1-1864	6504	56	86	34	11	2012-08-02 04:54:30.765957
6511	5	1-1745	6510	94	42	949	12	2012-08-02 04:54:31.22966
6512	5	1-1693	6510	42	96	727	72	2012-08-02 04:54:31.22966
6513	5	1-1835	6510	96	90	271	83	2012-08-02 04:54:31.22966
6514	5	1-1725	6510	90	42	507	57	2012-08-02 04:54:31.22966
6510	5	1-1865	6510	42	94	602	55	2012-08-02 04:54:31.22966
6516	6	1-1745	6515	94	42	2204	12	2012-08-02 04:54:31.22966
6517	6	1-1693	6515	42	93	2010	72	2012-08-02 04:54:31.22966
6518	6	1-1711	6515	93	7	1582	19	2012-08-02 04:54:31.22966
6519	6	1-33	6515	7	15	834	47	2012-08-02 04:54:31.22966
6520	6	1-1855	6515	15	42	697	57	2012-08-02 04:54:31.22966
6515	6	1-1865	6515	42	94	985	55	2012-08-02 04:54:31.22966
6522	4	1-1557	6521	93	10	468	29	2012-08-02 04:54:33.052479
6523	4	1-565	6521	10	33	145	28	2012-08-02 04:54:33.052479
6524	4	1-1828	6521	33	53	100	60	2012-08-02 04:54:33.052479
6521	4	1-1869	6521	53	93	99	39	2012-08-02 04:54:33.052479
6526	2	1-1872	6525	93	6	2694	53	2012-08-02 04:54:35.825224
6525	2	1-1875	6525	6	93	3020	35	2012-08-02 04:54:35.825224
6528	3	1-1021	6527	60	53	599	66	2012-08-02 04:54:35.825224
6529	3	1-1213	6527	53	6	850	53	2012-08-02 04:54:35.825224
6527	3	1-1875	6527	6	60	1312	35	2012-08-02 04:54:35.825224
6531	3	1-638	6530	100	79	3341	44	2012-08-02 04:54:37.592721
6532	3	1-1441	6530	79	59	3051	39	2012-08-02 04:54:37.592721
6530	3	1-1878	6530	59	100	3214	91	2012-08-02 04:54:37.592721
6534	4	1-1435	6533	32	34	2049	26	2012-08-02 04:54:37.592721
6535	4	1-1527	6533	34	53	1915	60	2012-08-02 04:54:37.592721
6536	4	1-1869	6533	53	59	1635	39	2012-08-02 04:54:37.592721
6533	4	1-1878	6533	59	32	1776	91	2012-08-02 04:54:37.592721
15972	5	1-4605	15972	10	56	434	86	2012-08-02 05:13:23.968085
6538	6	1-638	6537	100	63	621	44	2012-08-02 04:54:37.592721
6539	6	1-1546	6537	63	57	796	30	2012-08-02 04:54:37.592721
6540	6	1-847	6537	57	86	457	21	2012-08-02 04:54:37.592721
6541	6	1-1833	6537	86	53	445	60	2012-08-02 04:54:37.592721
6542	6	1-1869	6537	53	59	426	39	2012-08-02 04:54:37.592721
6537	6	1-1878	6537	59	100	518	91	2012-08-02 04:54:37.592721
6544	5	1-638	6543	100	66	20	44	2012-08-02 04:54:37.592721
6545	5	1-1416	6543	66	12	32	19	2012-08-02 04:54:37.592721
6546	5	1-1083	6543	12	53	13	60	2012-08-02 04:54:37.592721
6547	5	1-1869	6543	53	59	13	39	2012-08-02 04:54:37.592721
6543	5	1-1878	6543	59	100	16	91	2012-08-02 04:54:37.592721
6549	7	1-915	6548	63	66	20	85	2012-08-02 04:54:37.592721
6550	7	1-1856	6548	66	61	36	42	2012-08-02 04:54:37.592721
6551	7	1-1482	6548	61	47	16	86	2012-08-02 04:54:37.592721
6552	7	1-1488	6548	47	33	15	28	2012-08-02 04:54:37.592721
6553	7	1-1828	6548	33	53	11	60	2012-08-02 04:54:37.592721
6554	7	1-1869	6548	53	59	11	39	2012-08-02 04:54:37.592721
6548	7	1-1878	6548	59	63	14	91	2012-08-02 04:54:37.592721
6556	7	1-638	6555	100	63	608	44	2012-08-02 04:54:37.592721
6557	7	1-1546	6555	63	13	816	30	2012-08-02 04:54:37.592721
6558	7	1-666	6555	13	53	410	46	2012-08-02 04:54:37.592721
6559	7	1-1206	6555	53	33	518	28	2012-08-02 04:54:37.592721
6560	7	1-1828	6555	33	53	364	60	2012-08-02 04:54:37.592721
6561	7	1-1869	6555	53	59	379	39	2012-08-02 04:54:37.592721
6555	7	1-1878	6555	59	100	483	91	2012-08-02 04:54:37.592721
6563	7	1-1559	6562	60	99	523	61	2012-08-02 04:54:37.592721
6564	7	1-1414	6562	99	49	1439	7	2012-08-02 04:54:37.592721
6565	7	1-1499	6562	49	7	1105	78	2012-08-02 04:54:37.592721
6566	7	1-1525	6562	7	33	1309	28	2012-08-02 04:54:37.592721
6567	7	1-1828	6562	33	53	921	60	2012-08-02 04:54:37.592721
6568	7	1-1869	6562	53	59	925	39	2012-08-02 04:54:37.592721
6562	7	1-1878	6562	59	60	1182	91	2012-08-02 04:54:37.592721
6570	6	1-1671	6569	98	1	939	13	2012-08-02 04:54:37.592721
6571	6	1-1689	6569	1	92	242	1	2012-08-02 04:54:37.592721
6572	6	1-872	6569	92	59	176	56	2012-08-02 04:54:37.592721
6573	6	1-1366	6569	59	38	112	96	2012-08-02 04:54:37.592721
6574	6	1-1607	6569	38	59	127	39	2012-08-02 04:54:37.592721
6569	6	1-1878	6569	59	98	174	91	2012-08-02 04:54:37.592721
6576	6	1-1546	6575	63	95	214	30	2012-08-02 04:54:38.07966
6577	6	1-1277	6575	95	24	229	59	2012-08-02 04:54:38.07966
6578	6	1-1837	6575	24	60	280	35	2012-08-02 04:54:38.07966
6579	6	1-1021	6575	60	6	135	66	2012-08-02 04:54:38.07966
6580	6	1-1734	6575	6	59	552	36	2012-08-02 04:54:38.07966
6575	6	1-1879	6575	59	63	189	44	2012-08-02 04:54:38.07966
6582	7	1-1546	6581	63	26	377	30	2012-08-02 04:54:38.07966
6583	7	1-1686	6581	26	72	197	11	2012-08-02 04:54:38.07966
6584	7	1-1561	6581	72	31	259	16	2012-08-02 04:54:38.07966
6585	7	1-1601	6581	31	39	648	43	2012-08-02 04:54:38.07966
6586	7	1-1618	6581	39	6	116	66	2012-08-02 04:54:38.07966
6587	7	1-1734	6581	6	59	603	36	2012-08-02 04:54:38.07966
6581	7	1-1879	6581	59	63	262	44	2012-08-02 04:54:38.07966
6589	6	1-1603	6588	43	74	1130	71	2012-08-02 04:54:39.493989
6590	6	1-235	6588	74	56	286	47	2012-08-02 04:54:39.493989
6591	6	1-1743	6588	56	88	818	81	2012-08-02 04:54:39.493989
6592	6	1-1831	6588	88	2	2416	9	2012-08-02 04:54:39.493989
6593	6	1-1335	6588	2	59	118	13	2012-08-02 04:54:39.493989
6588	6	1-1882	6588	59	43	2526	73	2012-08-02 04:54:39.493989
6595	3	1-1574	6594	66	92	733	24	2012-08-02 04:54:39.493989
6596	3	1-1323	6594	92	59	277	13	2012-08-02 04:54:39.493989
6594	3	1-1882	6594	59	66	6273	73	2012-08-02 04:54:39.493989
6598	6	1-986	6597	46	74	458	71	2012-08-02 04:54:39.957878
6599	6	1-235	6597	74	15	210	47	2012-08-02 04:54:39.957878
6600	6	1-1855	6597	15	42	237	57	2012-08-02 04:54:39.957878
6601	6	1-1865	6597	42	94	384	55	2012-08-02 04:54:39.957878
6602	6	1-1745	6597	94	15	1149	12	2012-08-02 04:54:39.957878
6597	6	1-1883	6597	15	46	1071	90	2012-08-02 04:54:39.957878
6604	7	1-1212	6603	77	82	330	34	2012-08-02 04:54:40.918816
6605	7	1-980	6603	82	64	234	22	2012-08-02 04:54:40.918816
6606	7	1-1834	6603	64	63	89	44	2012-08-02 04:54:40.918816
6607	7	1-1546	6603	63	26	129	30	2012-08-02 04:54:40.918816
6608	7	1-1686	6603	26	63	67	11	2012-08-02 04:54:40.918816
6609	7	1-1648	6603	63	95	189	52	2012-08-02 04:54:40.918816
6603	7	1-1885	6603	95	77	606	48	2012-08-02 04:54:40.918816
6611	3	1-1362	6610	79	15	1321	17	2012-08-02 04:54:41.382771
6612	3	1-1807	6610	15	37	1177	18	2012-08-02 04:54:41.382771
6610	3	1-1886	6610	37	79	1868	4	2012-08-02 04:54:41.382771
6614	2	1-1601	6613	31	24	1876	43	2012-08-02 04:54:42.288671
6613	2	1-1888	6613	24	31	838	16	2012-08-02 04:54:42.288671
6616	3	1-1397	6615	21	64	1819	65	2012-08-02 04:54:49.183036
6617	3	1-1817	6615	64	76	6059	18	2012-08-02 04:54:49.183036
6615	3	1-1905	6615	76	21	5902	59	2012-08-02 04:54:49.183036
6619	4	1-1617	6618	95	44	4480	40	2012-08-02 04:54:50.099107
6620	4	1-1843	6618	44	6	1831	70	2012-08-02 04:54:50.099107
6621	4	1-1842	6618	6	71	985	37	2012-08-02 04:54:50.099107
6618	4	1-1907	6618	71	95	2710	44	2012-08-02 04:54:50.099107
6623	4	1-1546	6622	63	15	4286	30	2012-08-02 04:54:50.099107
6624	4	1-1815	6622	15	58	2327	31	2012-08-02 04:54:50.099107
6625	4	1-1480	6622	58	71	1493	37	2012-08-02 04:54:50.099107
6622	4	1-1907	6622	71	63	4468	44	2012-08-02 04:54:50.099107
6627	3	1-1870	6626	72	74	8228	10	2012-08-02 04:54:50.552083
6628	3	1-1761	6626	74	81	2422	5	2012-08-02 04:54:50.552083
6626	3	1-1908	6626	81	72	9808	67	2012-08-02 04:54:50.552083
6630	5	1-1691	6629	88	61	313	97	2012-08-02 04:54:50.552083
6631	5	1-1763	6629	61	26	105	9	2012-08-02 04:54:50.552083
6632	5	1-1838	6629	26	74	41	10	2012-08-02 04:54:50.552083
6633	5	1-1761	6629	74	81	19	5	2012-08-02 04:54:50.552083
6629	5	1-1908	6629	81	88	123	67	2012-08-02 04:54:50.552083
6635	6	1-855	6634	24	51	829	69	2012-08-02 04:54:52.861141
6636	6	1-1205	6634	51	23	391	31	2012-08-02 04:54:52.861141
6637	6	1-1804	6634	23	47	178	45	2012-08-02 04:54:52.861141
6634	6	1-1913	6634	86	24	5437	10	2012-08-02 04:54:52.861141
6638	6	1-1105	6634	47	42	73	48	2012-08-02 04:54:52.861141
6639	6	1-1613	6634	42	86	32	27	2012-08-02 04:54:52.861141
15978	5	1-1307	15977	47	20	175	24	2012-08-02 05:13:23.968085
6641	6	1-1259	6640	89	22	2642	62	2012-08-02 04:54:52.861141
6642	6	1-849	6640	22	84	804	17	2012-08-02 04:54:52.861141
6643	6	1-1765	6640	84	21	61	60	2012-08-02 04:54:52.861141
6644	6	1-621	6640	21	74	61	76	2012-08-02 04:54:52.861141
6645	6	1-414	6640	74	86	17	27	2012-08-02 04:54:52.861141
6640	6	1-1913	6640	86	89	3486	10	2012-08-02 04:54:52.861141
15979	5	1-4597	15977	20	47	484	56	2012-08-02 05:13:23.968085
6647	4	1-1827	6646	63	97	560	6	2012-08-02 04:54:53.314071
6648	4	1-1709	6646	97	22	399	97	2012-08-02 04:54:53.314071
6649	4	1-1707	6646	22	12	389	68	2012-08-02 04:54:53.314071
6646	4	1-1914	6646	12	63	520	32	2012-08-02 04:54:53.314071
6651	6	1-1766	6650	35	64	403	89	2012-08-02 04:54:53.314071
6652	6	1-330	6650	64	94	410	42	2012-08-02 04:54:53.314071
6653	6	1-1398	6650	94	88	77	67	2012-08-02 04:54:53.314071
6654	6	1-1691	6650	88	22	789	97	2012-08-02 04:54:53.314071
6655	6	1-1707	6650	22	12	839	68	2012-08-02 04:54:53.314071
6650	6	1-1914	6650	12	35	1224	32	2012-08-02 04:54:53.314071
6656	1	1-1691	\N	88	88	467	97	2012-08-02 04:54:53.314071
6658	5	1-1739	6657	7	21	306	59	2012-08-02 04:54:57.147574
6659	5	1-1397	6657	21	64	92	65	2012-08-02 04:54:57.147574
6660	5	1-1817	6657	64	97	300	18	2012-08-02 04:54:57.147574
6661	5	1-1088	6657	97	48	273	70	2012-08-02 04:54:57.147574
6657	5	1-1923	6657	48	7	356	33	2012-08-02 04:54:57.147574
6663	5	1-1739	6662	7	54	781	59	2012-08-02 04:54:57.147574
6664	5	1-1255	6662	54	99	508	77	2012-08-02 04:54:57.147574
6665	5	1-1767	6662	99	44	1111	40	2012-08-02 04:54:57.147574
6666	5	1-1843	6662	44	48	643	70	2012-08-02 04:54:57.147574
6662	5	1-1923	6662	48	7	873	33	2012-08-02 04:54:57.147574
6668	6	1-1710	6667	9	12	1891	44	2012-08-02 04:54:59.456325
6669	6	1-1438	6667	12	64	2615	89	2012-08-02 04:54:59.456325
6670	6	1-330	6667	64	61	1699	42	2012-08-02 04:54:59.456325
6671	6	1-1482	6667	61	75	503	86	2012-08-02 04:54:59.456325
6672	6	1-1825	6667	75	48	589	93	2012-08-02 04:54:59.456325
6667	6	1-1928	6667	48	9	3214	61	2012-08-02 04:54:59.456325
6674	5	1-1710	6673	9	32	287	44	2012-08-02 04:54:59.456325
6675	5	1-1510	6673	32	61	248	42	2012-08-02 04:54:59.456325
6676	5	1-1482	6673	61	75	74	86	2012-08-02 04:54:59.456325
6677	5	1-1825	6673	75	48	88	93	2012-08-02 04:54:59.456325
6673	5	1-1928	6673	48	9	483	61	2012-08-02 04:54:59.456325
6679	4	1-1727	6678	68	30	2607	71	2012-08-02 04:54:59.456325
6680	4	1-1436	6678	30	17	1023	94	2012-08-02 04:54:59.456325
6681	4	1-463	6678	17	48	513	93	2012-08-02 04:54:59.456325
6678	4	1-1928	6678	48	68	2829	61	2012-08-02 04:54:59.456325
6683	6	1-737	6682	42	7	1470	78	2012-08-02 04:55:01.756367
6684	6	1-1525	6682	7	85	1483	28	2012-08-02 04:55:01.756367
6685	6	1-1719	6682	85	79	1861	4	2012-08-02 04:55:01.756367
6686	6	1-1362	6682	79	79	1430	17	2012-08-02 04:55:01.756367
6687	6	1-1862	6682	79	56	3090	50	2012-08-02 04:55:01.756367
6682	6	1-1933	6682	56	42	4790	87	2012-08-02 04:55:01.756367
6689	6	1-737	6688	42	7	258	78	2012-08-02 04:55:01.756367
6690	6	1-1525	6688	7	76	281	28	2012-08-02 04:55:01.756367
6691	6	1-1715	6688	76	23	672	15	2012-08-02 04:55:01.756367
6692	6	1-183	6688	23	79	225	17	2012-08-02 04:55:01.756367
6693	6	1-1862	6688	79	56	466	50	2012-08-02 04:55:01.756367
6688	6	1-1933	6688	56	42	779	87	2012-08-02 04:55:01.756367
6695	4	1-328	6694	77	82	2334	34	2012-08-02 04:55:02.629241
6696	4	1-980	6694	82	62	1592	22	2012-08-02 04:55:02.629241
6697	4	1-1231	6694	62	29	1408	74	2012-08-02 04:55:02.629241
6694	4	1-1935	6694	29	77	2047	90	2012-08-02 04:55:02.629241
6699	6	1-1656	6698	93	38	1410	96	2012-08-02 04:55:04.430008
6700	6	1-1607	6698	38	59	1549	39	2012-08-02 04:55:04.430008
6701	6	1-1878	6698	59	70	2046	91	2012-08-02 04:55:04.430008
6702	6	1-1685	6698	70	64	2363	30	2012-08-02 04:55:04.430008
6703	6	1-1294	6698	64	65	1682	84	2012-08-02 04:55:04.430008
6698	6	1-1939	6698	65	93	3632	94	2012-08-02 04:55:04.430008
6705	6	1-288	6704	66	17	373	75	2012-08-02 04:55:04.893728
6706	6	1-1854	6704	17	34	125	8	2012-08-02 04:55:04.893728
6707	6	1-702	6704	34	63	50	11	2012-08-02 04:55:04.893728
6708	6	1-1648	6704	63	95	80	52	2012-08-02 04:55:04.893728
6709	6	1-1885	6704	95	58	146	48	2012-08-02 04:55:04.893728
6704	6	1-1940	6704	58	66	3532	99	2012-08-02 04:55:04.893728
6711	3	1-1019	6710	56	9	843	93	2012-08-02 04:55:04.893728
6712	3	1-1180	6710	9	58	128	48	2012-08-02 04:55:04.893728
6710	3	1-1940	6710	58	56	3502	99	2012-08-02 04:55:04.893728
6714	5	1-171	6713	72	15	1835	47	2012-08-02 04:55:05.359321
6715	5	1-1855	6713	15	42	1768	57	2012-08-02 04:55:05.359321
6716	5	1-1865	6713	42	15	2078	55	2012-08-02 04:55:05.359321
6717	5	1-1801	6713	15	22	1652	45	2012-08-02 04:55:05.359321
6713	5	1-1941	6713	22	72	4971	30	2012-08-02 04:55:05.359321
6719	3	1-1815	6718	15	23	721	31	2012-08-02 04:55:05.359321
6720	3	1-1804	6718	23	22	398	45	2012-08-02 04:55:05.359321
6718	3	1-1941	6718	22	15	1215	30	2012-08-02 04:55:05.359321
6722	4	1-1453	6721	63	49	1774	85	2012-08-02 04:55:05.834241
6723	4	1-1489	6721	49	83	1025	9	2012-08-02 04:55:05.834241
6724	4	1-1921	6721	83	30	1131	4	2012-08-02 04:55:05.834241
6721	4	1-1942	6721	30	63	1613	69	2012-08-02 04:55:05.834241
6726	6	1-1205	6725	51	23	2358	31	2012-08-02 04:55:05.834241
6727	6	1-1804	6725	23	2	1710	45	2012-08-02 04:55:05.834241
6728	6	1-1902	6725	2	1	1818	94	2012-08-02 04:55:05.834241
6729	6	1-1288	6725	1	85	1282	28	2012-08-02 04:55:05.834241
6730	6	1-1719	6725	85	30	1932	4	2012-08-02 04:55:05.834241
6725	6	1-1942	6725	30	51	3128	69	2012-08-02 04:55:05.834241
6732	3	1-702	6731	34	63	1831	11	2012-08-02 04:55:06.74008
6733	3	1-1648	6731	63	33	2610	52	2012-08-02 04:55:06.74008
6731	3	1-1944	6731	33	34	5115	8	2012-08-02 04:55:06.74008
6735	6	1-1833	6734	86	53	2511	60	2012-08-02 04:55:07.193176
6736	6	1-1869	6734	53	93	1224	39	2012-08-02 04:55:07.193176
6737	6	1-1557	6734	93	20	2864	29	2012-08-02 04:55:07.193176
6738	6	1-1853	6734	20	36	1444	32	2012-08-02 04:55:07.193176
6739	6	1-1270	6734	36	85	2616	40	2012-08-02 04:55:07.193176
6734	6	1-1945	6734	85	86	5057	21	2012-08-02 04:55:07.193176
6741	4	1-1833	6740	86	69	780	60	2012-08-02 04:55:07.193176
6742	4	1-1320	6740	69	36	400	32	2012-08-02 04:55:07.193176
6743	4	1-1270	6740	36	85	752	40	2012-08-02 04:55:07.193176
6740	4	1-1945	6740	85	86	1511	21	2012-08-02 04:55:07.193176
15980	5	1-4296	15977	47	48	1875	40	2012-08-02 05:13:23.968085
6745	6	1-1609	6744	58	83	316	52	2012-08-02 04:55:08.518804
6746	6	1-1404	6744	83	53	64	98	2012-08-02 04:55:08.518804
6747	6	1-1449	6744	53	53	60	46	2012-08-02 04:55:08.518804
6748	6	1-1206	6744	53	70	62	28	2012-08-02 04:55:08.518804
6749	6	1-1681	6744	70	37	44	70	2012-08-02 04:55:08.518804
6744	6	1-1948	6744	37	58	472	32	2012-08-02 04:55:08.518804
6751	6	1-1270	6750	36	27	3215	40	2012-08-02 04:55:08.518804
6752	6	1-1893	6750	27	52	469	92	2012-08-02 04:55:08.518804
6753	6	1-1293	6750	52	43	171	23	2012-08-02 04:55:08.518804
6754	6	1-1563	6750	43	97	158	18	2012-08-02 04:55:08.518804
6755	6	1-1088	6750	97	37	112	70	2012-08-02 04:55:08.518804
6750	6	1-1948	6750	37	36	1241	32	2012-08-02 04:55:08.518804
6757	5	1-1747	6756	31	52	448	92	2012-08-02 04:55:08.518804
6758	5	1-1293	6756	52	43	170	23	2012-08-02 04:55:08.518804
6759	5	1-1563	6756	43	97	163	18	2012-08-02 04:55:08.518804
6760	5	1-1088	6756	97	37	120	70	2012-08-02 04:55:08.518804
6756	5	1-1948	6756	37	31	1384	32	2012-08-02 04:55:08.518804
6762	3	1-1260	6761	33	70	428	28	2012-08-02 04:55:08.518804
6763	3	1-1681	6761	70	37	350	70	2012-08-02 04:55:08.518804
6761	3	1-1948	6761	37	33	4252	32	2012-08-02 04:55:08.518804
6764	1	1-1088	\N	97	97	217	70	2012-08-02 04:55:08.518804
6766	4	1-1904	6765	93	70	2052	91	2012-08-02 04:55:08.972212
6767	4	1-1685	6765	70	15	968	30	2012-08-02 04:55:08.972212
6768	4	1-1815	6765	15	92	311	31	2012-08-02 04:55:08.972212
6765	4	1-1949	6765	92	93	9753	35	2012-08-02 04:55:08.972212
6770	6	1-1449	6769	53	53	2368	46	2012-08-02 04:55:09.42556
6771	6	1-1206	6769	53	70	2592	28	2012-08-02 04:55:09.42556
6772	6	1-1681	6769	70	48	2084	70	2012-08-02 04:55:09.42556
6773	6	1-1923	6769	48	44	2283	33	2012-08-02 04:55:09.42556
6774	6	1-1045	6769	44	100	801	76	2012-08-02 04:55:09.42556
6769	6	1-1950	6769	100	53	2368	98	2012-08-02 04:55:09.42556
6776	2	1-344	6775	27	100	2224	76	2012-08-02 04:55:09.42556
6775	2	1-1950	6775	100	27	7152	98	2012-08-02 04:55:09.42556
6778	5	1-1559	6777	60	99	809	61	2012-08-02 04:55:09.88921
6779	5	1-1414	6777	99	49	2202	7	2012-08-02 04:55:09.88921
6780	5	1-1499	6777	49	7	1671	78	2012-08-02 04:55:09.88921
6781	5	1-1525	6777	7	65	1957	28	2012-08-02 04:55:09.88921
6777	5	1-1951	6777	65	60	1850	91	2012-08-02 04:55:09.88921
6783	5	1-1685	6782	70	64	48	30	2012-08-02 04:55:09.88921
6784	5	1-1294	6782	64	65	32	84	2012-08-02 04:55:09.88921
6785	5	1-1939	6782	65	1	68	94	2012-08-02 04:55:09.88921
6786	5	1-1288	6782	1	65	45	28	2012-08-02 04:55:09.88921
6782	5	1-1951	6782	65	70	44	91	2012-08-02 04:55:09.88921
6788	6	1-1685	6787	70	26	2086	30	2012-08-02 04:55:09.88921
6789	6	1-1686	6787	26	25	1081	11	2012-08-02 04:55:09.88921
6790	6	1-1698	6787	25	53	1127	98	2012-08-02 04:55:09.88921
6791	6	1-1449	6787	53	53	1341	46	2012-08-02 04:55:09.88921
6792	6	1-1206	6787	53	65	1745	28	2012-08-02 04:55:09.88921
6787	6	1-1951	6787	65	70	1766	91	2012-08-02 04:55:09.88921
6793	1	1-438	\N	75	75	1136	11	2012-08-02 04:55:09.88921
6795	5	1-1286	6794	85	86	4281	14	2012-08-02 04:55:10.817253
6796	5	1-1789	6794	86	7	3763	25	2012-08-02 04:55:10.817253
6797	5	1-1666	6794	7	89	2199	10	2012-08-02 04:55:10.817253
6798	5	1-1259	6794	89	81	2114	62	2012-08-02 04:55:10.817253
6794	5	1-1953	6794	81	85	2068	16	2012-08-02 04:55:10.817253
6800	6	1-1499	6799	49	7	1747	78	2012-08-02 04:55:11.248188
6801	6	1-1525	6799	7	65	1645	28	2012-08-02 04:55:11.248188
6802	6	1-1951	6799	65	65	1457	91	2012-08-02 04:55:11.248188
6803	6	1-541	6799	65	95	1200	52	2012-08-02 04:55:11.248188
6804	6	1-1885	6799	95	43	2871	48	2012-08-02 04:55:11.248188
6799	6	1-1954	6799	43	49	2863	7	2012-08-02 04:55:11.248188
6806	5	1-1885	6805	95	43	1218	48	2012-08-02 04:55:11.723349
6807	5	1-1954	6805	43	16	1288	7	2012-08-02 04:55:11.723349
6808	5	1-1380	6805	16	49	1296	85	2012-08-02 04:55:11.723349
6809	5	1-1489	6805	49	71	700	9	2012-08-02 04:55:11.723349
6805	5	1-1956	6805	71	95	480	52	2012-08-02 04:55:11.723349
6811	6	1-1956	6810	71	95	275	52	2012-08-02 04:55:13.524003
6812	6	1-1885	6810	95	43	786	48	2012-08-02 04:55:13.524003
6813	6	1-1954	6810	43	49	937	7	2012-08-02 04:55:13.524003
6814	6	1-1499	6810	49	17	684	78	2012-08-02 04:55:13.524003
6815	6	1-1708	6810	17	23	1129	16	2012-08-02 04:55:13.524003
6810	6	1-1960	6810	23	71	356	9	2012-08-02 04:55:13.524003
6817	3	1-33	6816	7	56	1634	47	2012-08-02 04:55:13.965933
6818	3	1-1743	6816	56	43	3576	81	2012-08-02 04:55:13.965933
6816	3	1-1961	6816	43	7	5466	19	2012-08-02 04:55:13.965933
6820	4	1-719	6819	44	74	1427	71	2012-08-02 04:55:13.965933
6821	4	1-235	6819	74	56	414	47	2012-08-02 04:55:13.965933
6822	4	1-1743	6819	56	43	1358	81	2012-08-02 04:55:13.965933
6819	4	1-1961	6819	43	44	3108	19	2012-08-02 04:55:13.965933
6824	4	1-1083	6823	12	54	244	60	2012-08-02 04:55:13.965933
6825	4	1-305	6823	54	56	79	47	2012-08-02 04:55:13.965933
6826	4	1-1743	6823	56	43	295	81	2012-08-02 04:55:13.965933
6823	4	1-1961	6823	43	12	767	19	2012-08-02 04:55:13.965933
6828	6	1-1267	6827	71	8	757	15	2012-08-02 04:55:14.87216
6829	6	1-773	6827	8	16	433	50	2012-08-02 04:55:14.87216
6830	6	1-1934	6827	16	61	116	77	2012-08-02 04:55:14.87216
6831	6	1-504	6827	61	80	32	75	2012-08-02 04:55:14.87216
6832	6	1-1602	6827	80	7	2556	14	2012-08-02 04:55:14.87216
6827	6	1-1963	6827	7	71	969	13	2012-08-02 04:55:14.87216
6834	6	1-1689	6833	1	18	277	1	2012-08-02 04:55:14.87216
6835	6	1-1780	6833	18	89	1374	10	2012-08-02 04:55:14.87216
6836	6	1-1259	6833	89	81	1351	62	2012-08-02 04:55:14.87216
6837	6	1-1953	6833	81	85	1352	16	2012-08-02 04:55:14.87216
6838	6	1-1286	6833	85	7	2863	14	2012-08-02 04:55:14.87216
6833	6	1-1963	6833	7	1	1139	13	2012-08-02 04:55:14.87216
6839	1	1-393	\N	33	33	156	62	2012-08-02 04:55:14.87216
6841	6	1-1893	6840	27	81	91	92	2012-08-02 04:55:15.766864
6842	6	1-1042	6840	81	79	127	17	2012-08-02 04:55:15.766864
6843	6	1-1862	6840	79	16	277	50	2012-08-02 04:55:15.766864
6844	6	1-1934	6840	16	87	75	77	2012-08-02 04:55:15.766864
6845	6	1-1632	6840	87	55	62	48	2012-08-02 04:55:15.766864
6840	6	1-1965	6840	55	27	448	40	2012-08-02 04:55:15.766864
6846	3	1-1973	6846	32	2	1012	45	2012-08-02 04:55:19.279948
6847	3	1-1902	6846	2	19	936	94	2012-08-02 04:55:19.279948
6848	3	1-1387	6846	19	32	1395	6	2012-08-02 04:55:19.279948
15981	5	1-3547	15977	48	10	145	18	2012-08-02 05:13:23.968085
6850	6	1-1119	6849	50	92	4243	62	2012-08-02 04:55:19.279948
6851	6	1-1967	6849	92	22	3307	19	2012-08-02 04:55:19.279948
6852	6	1-295	6849	22	12	1683	68	2012-08-02 04:55:19.279948
6853	6	1-1914	6849	12	63	2517	32	2012-08-02 04:55:19.279948
6854	6	1-1827	6849	63	32	3029	6	2012-08-02 04:55:19.279948
6849	6	1-1973	6849	32	50	2588	45	2012-08-02 04:55:19.279948
15977	5	1-4605	15977	10	47	376	86	2012-08-02 05:13:23.968085
6856	2	1-1845	6855	34	55	6103	48	2012-08-02 04:55:21.025411
6855	2	1-1977	6855	55	34	8169	91	2012-08-02 04:55:21.025411
6858	3	1-1563	6857	43	66	2296	18	2012-08-02 04:55:21.467483
6859	3	1-1679	6857	66	51	744	27	2012-08-02 04:55:21.467483
6857	3	1-1978	6857	51	43	4765	23	2012-08-02 04:55:21.467483
6861	3	1-803	6860	93	70	456	89	2012-08-02 04:55:21.467483
6862	3	1-873	6860	70	51	259	27	2012-08-02 04:55:21.467483
6860	3	1-1978	6860	51	93	2902	23	2012-08-02 04:55:21.467483
6864	3	1-1414	6863	99	16	1146	7	2012-08-02 04:55:23.643904
6865	3	1-1380	6863	16	28	896	85	2012-08-02 04:55:23.643904
6863	3	1-1983	6863	28	99	636	61	2012-08-02 04:55:23.643904
6867	7	1-1710	6866	9	63	1493	44	2012-08-02 04:55:23.643904
6868	7	1-1546	6866	63	13	2063	30	2012-08-02 04:55:23.643904
6869	7	1-666	6866	13	53	1015	46	2012-08-02 04:55:23.643904
6870	7	1-1206	6866	53	65	1208	28	2012-08-02 04:55:23.643904
6871	7	1-1951	6866	65	63	1129	91	2012-08-02 04:55:23.643904
6872	7	1-915	6866	63	28	1604	85	2012-08-02 04:55:23.643904
6866	7	1-1983	6866	28	9	1697	61	2012-08-02 04:55:23.643904
6874	7	1-1509	6873	77	25	141	11	2012-08-02 04:55:23.643904
6875	7	1-1698	6873	25	53	137	98	2012-08-02 04:55:23.643904
6876	7	1-1449	6873	53	53	157	46	2012-08-02 04:55:23.643904
6877	7	1-1206	6873	53	65	198	28	2012-08-02 04:55:23.643904
6878	7	1-1951	6873	65	63	186	91	2012-08-02 04:55:23.643904
6879	7	1-915	6873	63	28	254	85	2012-08-02 04:55:23.643904
6873	7	1-1983	6873	28	77	271	61	2012-08-02 04:55:23.643904
6881	5	1-1142	6880	31	43	456	92	2012-08-02 04:55:23.643904
6882	5	1-1587	6880	43	25	472	50	2012-08-02 04:55:23.643904
6883	5	1-1966	6880	25	90	216	35	2012-08-02 04:55:23.643904
6884	5	1-1210	6880	90	28	439	85	2012-08-02 04:55:23.643904
6880	5	1-1983	6880	28	31	488	61	2012-08-02 04:55:23.643904
6885	1	1-915	\N	63	63	110	85	2012-08-02 04:55:23.643904
6887	7	1-1449	6886	53	66	1204	46	2012-08-02 04:55:25.87539
6888	7	1-1390	6886	66	76	469	28	2012-08-02 04:55:25.87539
6889	7	1-1715	6886	76	8	1288	15	2012-08-02 04:55:25.87539
6890	7	1-773	6886	8	25	806	50	2012-08-02 04:55:25.87539
6891	7	1-1966	6886	25	60	378	35	2012-08-02 04:55:25.87539
6892	7	1-1021	6886	60	2	227	66	2012-08-02 04:55:25.87539
6886	7	1-1988	6886	2	53	1027	98	2012-08-02 04:55:25.87539
6894	6	1-1449	6893	53	77	493	46	2012-08-02 04:55:25.87539
6895	6	1-1080	6893	77	73	256	68	2012-08-02 04:55:25.87539
6896	6	1-1730	6893	73	25	324	50	2012-08-02 04:55:25.87539
6897	6	1-1966	6893	25	60	153	35	2012-08-02 04:55:25.87539
6898	6	1-1021	6893	60	2	92	66	2012-08-02 04:55:25.87539
6893	6	1-1988	6893	2	53	418	98	2012-08-02 04:55:25.87539
6900	6	1-1730	6899	73	25	569	50	2012-08-02 04:55:26.328478
6901	6	1-1966	6899	25	60	267	35	2012-08-02 04:55:26.328478
6902	6	1-1585	6899	60	29	252	74	2012-08-02 04:55:26.328478
6903	6	1-1935	6899	29	85	375	90	2012-08-02 04:55:26.328478
6904	6	1-1852	6899	85	20	4230	5	2012-08-02 04:55:26.328478
6899	6	1-1989	6899	20	73	453	68	2012-08-02 04:55:26.328478
6906	4	1-1749	6905	14	66	4963	45	2012-08-02 04:55:28.970975
6907	4	1-1820	6905	66	83	1233	95	2012-08-02 04:55:28.970975
6908	4	1-1781	6905	83	89	317	15	2012-08-02 04:55:28.970975
6905	4	1-1995	6905	89	14	489	38	2012-08-02 04:55:28.970975
6910	5	1-1919	6909	56	15	1108	67	2012-08-02 04:55:28.970975
6911	5	1-1822	6909	15	47	874	86	2012-08-02 04:55:28.970975
6912	5	1-1488	6909	47	76	409	28	2012-08-02 04:55:28.970975
6913	5	1-1715	6909	76	89	536	15	2012-08-02 04:55:28.970975
6909	5	1-1995	6909	89	56	1420	38	2012-08-02 04:55:28.970975
6915	4	1-1919	6914	56	15	658	67	2012-08-02 04:55:28.970975
6916	4	1-1822	6914	15	55	618	86	2012-08-02 04:55:28.970975
6917	4	1-1235	6914	55	89	225	15	2012-08-02 04:55:28.970975
6914	4	1-1995	6914	89	56	709	38	2012-08-02 04:55:28.970975
6919	7	1-642	6918	83	46	470	34	2012-08-02 04:55:31.63312
6920	7	1-1937	6918	46	84	542	64	2012-08-02 04:55:31.63312
6921	7	1-756	6918	84	43	451	48	2012-08-02 04:55:31.63312
6922	7	1-1954	6918	43	49	595	7	2012-08-02 04:55:31.63312
6923	7	1-1499	6918	49	85	480	78	2012-08-02 04:55:31.63312
6924	7	1-1779	6918	85	85	194	56	2012-08-02 04:55:31.63312
6918	7	1-2001	6918	85	83	901	71	2012-08-02 04:55:31.63312
6926	6	1-1510	6925	32	94	326	42	2012-08-02 04:55:32.075248
6927	6	1-1398	6925	94	94	51	67	2012-08-02 04:55:32.075248
6928	6	1-1991	6925	94	34	356	57	2012-08-02 04:55:32.075248
6929	6	1-42	6925	34	56	192	47	2012-08-02 04:55:32.075248
6930	6	1-1743	6925	56	85	638	81	2012-08-02 04:55:32.075248
6925	6	1-2002	6925	85	32	366	44	2012-08-02 04:55:32.075248
6932	5	1-1510	6931	32	94	462	42	2012-08-02 04:55:32.075248
6933	5	1-1398	6931	94	94	83	67	2012-08-02 04:55:32.075248
6934	5	1-1991	6931	94	100	661	57	2012-08-02 04:55:32.075248
6935	5	1-1388	6931	100	85	526	81	2012-08-02 04:55:32.075248
6931	5	1-2002	6931	85	32	396	44	2012-08-02 04:55:32.075248
6937	3	1-30	6936	28	22	3815	43	2012-08-02 04:55:33.379039
6938	3	1-1246	6936	22	34	2823	70	2012-08-02 04:55:33.379039
6936	3	1-2005	6936	34	28	4662	11	2012-08-02 04:55:33.379039
6940	5	1-1698	6939	25	53	1203	98	2012-08-02 04:55:33.379039
6941	5	1-1449	6939	53	66	1433	46	2012-08-02 04:55:33.379039
6942	5	1-1390	6939	66	70	568	28	2012-08-02 04:55:33.379039
6943	5	1-1681	6939	70	34	653	70	2012-08-02 04:55:33.379039
6939	5	1-2005	6939	34	25	1141	11	2012-08-02 04:55:33.379039
6944	1	1-30	\N	28	28	332	43	2012-08-02 04:55:33.379039
6945	1	1-1390	\N	66	66	329	28	2012-08-02 04:55:33.379039
6947	4	1-1988	6946	2	93	2830	98	2012-08-02 04:55:33.823454
6948	4	1-1417	6946	93	10	1411	51	2012-08-02 04:55:33.823454
6949	4	1-1705	6946	10	46	834	49	2012-08-02 04:55:33.823454
6946	4	1-2007	6946	46	2	938	66	2012-08-02 04:55:33.823454
6951	5	1-1321	6950	99	86	1633	21	2012-08-02 04:55:33.823454
6952	5	1-1833	6950	86	44	1466	60	2012-08-02 04:55:33.823454
6953	5	1-518	6950	44	10	1468	51	2012-08-02 04:55:33.823454
6954	5	1-1705	6950	10	46	1081	49	2012-08-02 04:55:33.823454
6950	5	1-2007	6950	46	99	1515	66	2012-08-02 04:55:33.823454
6956	5	1-1321	6955	99	86	2431	21	2012-08-02 04:55:33.823454
6957	5	1-1833	6955	86	88	2600	60	2012-08-02 04:55:33.823454
6958	5	1-1014	6955	88	24	1208	97	2012-08-02 04:55:33.823454
6959	5	1-1990	6955	24	46	1134	49	2012-08-02 04:55:33.823454
6955	5	1-2007	6955	46	99	1893	66	2012-08-02 04:55:33.823454
15983	4	1-4089	15982	35	30	34	4	2012-08-02 05:13:23.968085
6961	5	1-1388	6960	100	85	1058	81	2012-08-02 04:55:34.251929
6962	5	1-2002	6960	85	12	774	44	2012-08-02 04:55:34.251929
6963	5	1-1438	6960	12	85	1141	89	2012-08-02 04:55:34.251929
6964	5	1-1165	6960	85	64	1734	4	2012-08-02 04:55:34.251929
6960	5	1-2008	6960	64	100	1702	57	2012-08-02 04:55:34.251929
15984	4	1-4011	15982	30	97	42	26	2012-08-02 05:13:23.968085
6966	2	1-910	6965	77	64	3722	4	2012-08-02 04:55:34.251929
6965	2	1-2008	6965	64	77	4838	57	2012-08-02 04:55:34.251929
6968	3	1-1624	6967	9	85	12	89	2012-08-02 04:55:34.251929
6969	3	1-1165	6967	85	64	20	4	2012-08-02 04:55:34.251929
6967	3	1-2008	6967	64	9	27	57	2012-08-02 04:55:34.251929
6971	6	1-1388	6970	100	85	501	81	2012-08-02 04:55:34.251929
6972	6	1-2002	6970	85	74	447	44	2012-08-02 04:55:34.251929
6973	6	1-1394	6970	74	83	299	33	2012-08-02 04:55:34.251929
6974	6	1-729	6970	83	83	272	9	2012-08-02 04:55:34.251929
6975	6	1-1921	6970	83	64	356	4	2012-08-02 04:55:34.251929
6970	6	1-2008	6970	64	100	531	57	2012-08-02 04:55:34.251929
6977	6	1-1874	6976	46	16	907	7	2012-08-02 04:55:36.439407
6978	6	1-1380	6976	16	28	1089	85	2012-08-02 04:55:36.439407
6979	6	1-1983	6976	28	23	1186	61	2012-08-02 04:55:36.439407
6980	6	1-146	6976	23	100	1278	27	2012-08-02 04:55:36.439407
6981	6	1-503	6976	100	99	622	87	2012-08-02 04:55:36.439407
6976	6	1-2013	6976	99	46	640	23	2012-08-02 04:55:36.439407
6983	6	1-1874	6982	46	65	5189	7	2012-08-02 04:55:36.439407
6984	6	1-1752	6982	65	93	2972	24	2012-08-02 04:55:36.439407
6985	6	1-1111	6982	93	44	1624	68	2012-08-02 04:55:36.439407
6986	6	1-2009	6982	44	64	5241	100	2012-08-02 04:55:36.439407
6987	6	1-1531	6982	64	99	3340	87	2012-08-02 04:55:36.439407
6982	6	1-2013	6982	99	46	3547	23	2012-08-02 04:55:36.439407
6989	5	1-1496	6988	41	46	815	49	2012-08-02 04:55:38.637907
6990	5	1-2007	6988	46	95	1183	66	2012-08-02 04:55:38.637907
6991	5	1-1946	6988	95	43	734	48	2012-08-02 04:55:38.637907
6992	5	1-1954	6988	43	80	836	7	2012-08-02 04:55:38.637907
6988	5	1-2018	6988	80	41	949	96	2012-08-02 04:55:38.637907
6994	5	1-1496	6993	41	8	2966	49	2012-08-02 04:55:38.637907
6995	5	1-1487	6993	8	99	1858	87	2012-08-02 04:55:38.637907
6996	5	1-2013	6993	99	46	1871	23	2012-08-02 04:55:38.637907
6997	5	1-1874	6993	46	80	2594	7	2012-08-02 04:55:38.637907
6993	5	1-2018	6993	80	41	3187	96	2012-08-02 04:55:38.637907
6999	6	1-1496	6998	41	35	364	49	2012-08-02 04:55:38.637907
7000	6	1-1906	6998	35	15	330	55	2012-08-02 04:55:38.637907
7001	6	1-1801	6998	15	47	305	45	2012-08-02 04:55:38.637907
7002	6	1-1105	6998	47	43	207	48	2012-08-02 04:55:38.637907
7003	6	1-1954	6998	43	80	275	7	2012-08-02 04:55:38.637907
6998	6	1-2018	6998	80	41	364	96	2012-08-02 04:55:38.637907
7005	4	1-1509	7004	77	63	1089	11	2012-08-02 04:55:39.080116
7006	4	1-1648	7004	63	56	2084	52	2012-08-02 04:55:39.080116
7007	4	1-1728	7004	56	96	5165	90	2012-08-02 04:55:39.080116
7004	4	1-2019	7004	96	77	2825	61	2012-08-02 04:55:39.080116
7009	5	1-1727	7008	68	74	2284	71	2012-08-02 04:55:39.080116
7010	5	1-235	7008	74	59	837	47	2012-08-02 04:55:39.080116
7011	5	1-1909	7008	59	29	2545	74	2012-08-02 04:55:39.080116
7012	5	1-1935	7008	29	96	3049	90	2012-08-02 04:55:39.080116
7008	5	1-2019	7008	96	68	1932	61	2012-08-02 04:55:39.080116
7014	6	1-1745	7013	94	42	181	12	2012-08-02 04:55:39.510942
7015	6	1-1693	7013	42	96	177	72	2012-08-02 04:55:39.510942
7016	6	1-1835	7013	96	27	108	83	2012-08-02 04:55:39.510942
7017	6	1-1156	7013	27	42	49	58	2012-08-02 04:55:39.510942
7018	6	1-1232	7013	42	21	24	76	2012-08-02 04:55:39.510942
7013	6	1-2020	7013	21	94	70	55	2012-08-02 04:55:39.510942
7020	6	1-1801	7019	15	50	1545	45	2012-08-02 04:55:39.510942
7021	6	1-1119	7019	50	92	2132	62	2012-08-02 04:55:39.510942
7022	6	1-1967	7019	92	12	1399	19	2012-08-02 04:55:39.510942
7023	6	1-1083	7019	12	21	515	60	2012-08-02 04:55:39.510942
7024	6	1-621	7019	21	21	640	76	2012-08-02 04:55:39.510942
7019	6	1-2020	7019	21	15	2003	55	2012-08-02 04:55:39.510942
7026	6	1-1469	7025	48	71	804	80	2012-08-02 04:55:39.510942
7027	6	1-1993	7025	71	97	251	11	2012-08-02 04:55:39.510942
7028	6	1-1790	7025	97	54	543	59	2012-08-02 04:55:39.510942
7029	6	1-1255	7025	54	20	399	77	2012-08-02 04:55:39.510942
7030	6	1-1466	7025	20	21	693	76	2012-08-02 04:55:39.510942
7025	6	1-2020	7025	21	48	2324	55	2012-08-02 04:55:39.510942
7032	5	1-1801	7031	15	47	1579	45	2012-08-02 04:55:39.510942
7033	5	1-1105	7031	47	33	1065	48	2012-08-02 04:55:39.510942
7034	5	1-1772	7031	33	45	1825	25	2012-08-02 04:55:39.510942
7035	5	1-939	7031	45	21	507	76	2012-08-02 04:55:39.510942
7031	5	1-2020	7031	21	15	1715	55	2012-08-02 04:55:39.510942
7037	7	1-1925	7036	39	87	99	77	2012-08-02 04:55:43.068309
7038	7	1-1632	7036	87	43	81	48	2012-08-02 04:55:43.068309
7039	7	1-1954	7036	43	80	99	7	2012-08-02 04:55:43.068309
7040	7	1-2018	7036	80	26	120	96	2012-08-02 04:55:43.068309
7041	7	1-960	7036	26	76	71	56	2012-08-02 04:55:43.068309
7042	7	1-1848	7036	76	59	73	95	2012-08-02 04:55:43.068309
7036	7	1-2028	7036	59	39	102	63	2012-08-02 04:55:43.068309
7044	4	1-2006	7043	34	26	3017	48	2012-08-02 04:55:47.431847
7045	4	1-2037	7043	26	52	2522	82	2012-08-02 04:55:47.431847
7046	4	1-2026	7043	52	85	744	87	2012-08-02 04:55:47.431847
7043	4	1-2038	7043	85	34	329	38	2012-08-02 04:55:47.431847
7048	6	1-1919	7047	56	94	275	67	2012-08-02 04:55:47.431847
7049	6	1-1991	7047	94	9	2012	57	2012-08-02 04:55:47.431847
7050	6	1-1624	7047	9	70	819	89	2012-08-02 04:55:47.431847
7051	6	1-873	7047	70	100	664	27	2012-08-02 04:55:47.431847
7052	6	1-503	7047	100	85	290	87	2012-08-02 04:55:47.431847
7047	6	1-2038	7047	85	56	196	38	2012-08-02 04:55:47.431847
7054	3	1-1850	7053	10	14	2980	58	2012-08-02 04:55:47.431847
7055	3	1-1957	7053	14	85	2598	87	2012-08-02 04:55:47.431847
7053	3	1-2038	7053	85	10	1929	38	2012-08-02 04:55:47.431847
15985	4	1-3788	15982	97	10	14	18	2012-08-02 05:13:23.968085
7057	5	1-22	7056	22	7	3407	33	2012-08-02 04:55:47.874093
7058	5	1-1739	7056	7	54	1292	59	2012-08-02 04:55:47.874093
7059	5	1-1255	7056	54	78	357	77	2012-08-02 04:55:47.874093
7060	5	1-1566	7056	78	10	96	36	2012-08-02 04:55:47.874093
7056	5	1-2039	7056	10	22	3911	8	2012-08-02 04:55:47.874093
15982	4	1-4605	15982	10	35	38	86	2012-08-02 05:13:23.968085
7062	6	1-1886	7061	37	30	738	4	2012-08-02 04:55:51.365128
7063	6	1-1942	7061	30	34	1244	69	2012-08-02 04:55:51.365128
7064	6	1-528	7061	34	46	164	34	2012-08-02 04:55:51.365128
7065	6	1-1937	7061	46	84	192	64	2012-08-02 04:55:51.365128
7066	6	1-756	7061	84	17	162	48	2012-08-02 04:55:51.365128
7061	6	1-2047	7061	17	37	385	18	2012-08-02 04:55:51.365128
7068	6	1-1844	7067	100	14	1646	58	2012-08-02 04:55:53.077493
7069	6	1-1957	7067	14	85	1274	87	2012-08-02 04:55:53.077493
7070	6	1-2038	7067	85	91	930	38	2012-08-02 04:55:53.077493
7071	6	1-1036	7067	91	85	3482	81	2012-08-02 04:55:53.077493
7072	6	1-2002	7067	85	79	2860	44	2012-08-02 04:55:53.077493
7067	6	1-2051	7067	79	100	2658	60	2012-08-02 04:55:53.077493
7074	4	1-1991	7073	94	9	2377	57	2012-08-02 04:55:53.530161
7075	4	1-1624	7073	9	90	929	89	2012-08-02 04:55:53.530161
7076	4	1-954	7073	90	2	514	34	2012-08-02 04:55:53.530161
7073	4	1-2052	7073	2	94	338	67	2012-08-02 04:55:53.530161
7078	7	1-1991	7077	94	28	92	57	2012-08-02 04:55:53.530161
7079	7	1-1535	7077	28	27	422	40	2012-08-02 04:55:53.530161
7080	7	1-1893	7077	27	46	83	92	2012-08-02 04:55:53.530161
7081	7	1-1860	7077	46	32	11	91	2012-08-02 04:55:53.530161
7082	7	1-1435	7077	32	78	14	26	2012-08-02 04:55:53.530161
7083	7	1-1050	7077	78	2	17	34	2012-08-02 04:55:53.530161
7077	7	1-2052	7077	2	94	12	67	2012-08-02 04:55:53.530161
7085	6	1-1991	7084	94	100	1853	57	2012-08-02 04:55:53.530161
7086	6	1-1388	7084	100	85	1562	81	2012-08-02 04:55:53.530161
7087	6	1-2002	7084	85	21	1245	44	2012-08-02 04:55:53.530161
7088	6	1-1017	7084	21	78	280	26	2012-08-02 04:55:53.530161
7089	6	1-1050	7084	78	2	337	34	2012-08-02 04:55:53.530161
7084	6	1-2052	7084	2	94	242	67	2012-08-02 04:55:53.530161
7091	6	1-2018	7090	80	26	975	96	2012-08-02 04:55:54.40312
7092	6	1-960	7090	26	15	602	56	2012-08-02 04:55:54.40312
7093	6	1-685	7090	15	26	396	23	2012-08-02 04:55:54.40312
7094	6	1-2030	7090	26	18	261	1	2012-08-02 04:55:54.40312
7095	6	1-1780	7090	18	1	1214	10	2012-08-02 04:55:54.40312
7090	6	1-2054	7090	1	80	824	7	2012-08-02 04:55:54.40312
7097	4	1-1267	7096	71	89	2825	15	2012-08-02 04:55:56.579365
7098	4	1-579	7096	89	98	360	34	2012-08-02 04:55:56.579365
7099	4	1-1970	7096	98	25	176	89	2012-08-02 04:55:56.579365
7096	4	1-2059	7096	25	71	6925	13	2012-08-02 04:55:56.579365
7101	2	1-2049	7100	57	63	4317	74	2012-08-02 04:55:58.291796
7100	2	1-2063	7100	63	57	6091	16	2012-08-02 04:55:58.291796
7103	2	1-1754	7102	90	4	4343	41	2012-08-02 04:55:58.722601
7102	2	1-2064	7102	4	90	3140	23	2012-08-02 04:55:58.722601
7105	5	1-2030	7104	26	18	181	1	2012-08-02 04:55:58.722601
7106	5	1-1780	7104	18	1	781	10	2012-08-02 04:55:58.722601
7107	5	1-2054	7104	1	93	493	7	2012-08-02 04:55:58.722601
7108	5	1-1239	7104	93	4	418	41	2012-08-02 04:55:58.722601
7104	5	1-2064	7104	4	26	315	23	2012-08-02 04:55:58.722601
7110	4	1-2030	7109	26	14	1841	1	2012-08-02 04:55:58.722601
7111	4	1-1849	7109	14	35	2776	69	2012-08-02 04:55:58.722601
7112	4	1-1818	7109	35	4	3245	41	2012-08-02 04:55:58.722601
7109	4	1-2064	7109	4	26	2801	23	2012-08-02 04:55:58.722601
7114	6	1-756	7113	84	17	21	48	2012-08-02 04:55:59.5623
7115	6	1-2047	7113	17	37	40	18	2012-08-02 04:55:59.5623
7116	6	1-1886	7113	37	30	61	4	2012-08-02 04:55:59.5623
7117	6	1-1942	7113	30	22	82	69	2012-08-02 04:55:59.5623
7118	6	1-1930	7113	22	9	22	19	2012-08-02 04:55:59.5623
7113	6	1-2066	7113	9	84	31	64	2012-08-02 04:55:59.5623
7120	5	1-1637	7119	88	26	817	9	2012-08-02 04:55:59.5623
7121	5	1-1838	7119	26	89	1168	10	2012-08-02 04:55:59.5623
7122	5	1-1259	7119	89	92	1053	62	2012-08-02 04:55:59.5623
7123	5	1-1967	7119	92	9	724	19	2012-08-02 04:55:59.5623
7119	5	1-2066	7119	9	88	1126	64	2012-08-02 04:55:59.5623
7125	6	1-756	7124	84	17	2769	48	2012-08-02 04:55:59.5623
7126	6	1-2047	7124	17	65	6511	18	2012-08-02 04:55:59.5623
7127	6	1-1391	7124	65	98	2822	3	2012-08-02 04:55:59.5623
7128	6	1-1741	7124	98	92	2382	62	2012-08-02 04:55:59.5623
7129	6	1-1967	7124	92	9	1868	19	2012-08-02 04:55:59.5623
7124	6	1-2066	7124	9	84	3311	64	2012-08-02 04:55:59.5623
7131	5	1-1388	7130	100	85	2907	81	2012-08-02 04:56:00.435403
7132	5	1-2002	7130	85	79	2612	44	2012-08-02 04:56:00.435403
7133	5	1-2051	7130	79	88	2943	60	2012-08-02 04:56:00.435403
7134	5	1-1014	7130	88	2	1393	97	2012-08-02 04:56:00.435403
7130	5	1-2068	7130	2	100	3060	57	2012-08-02 04:56:00.435403
7136	4	1-1804	7135	23	47	1273	45	2012-08-02 04:56:01.331903
7137	4	1-1105	7135	47	33	874	48	2012-08-02 04:56:01.331903
7138	4	1-1772	7135	33	44	1524	25	2012-08-02 04:56:01.331903
7135	4	1-2070	7135	44	23	1675	31	2012-08-02 04:56:01.331903
7140	6	1-1480	7139	58	2	3422	37	2012-08-02 04:56:05.199571
7141	6	1-2076	7139	2	40	4431	99	2012-08-02 04:56:05.199571
7142	6	1-1929	7139	40	71	4697	80	2012-08-02 04:56:05.199571
7143	6	1-1993	7139	71	63	1398	11	2012-08-02 04:56:05.199571
7144	6	1-1648	7139	63	61	3653	52	2012-08-02 04:56:05.199571
7139	6	1-2079	7139	61	58	3813	31	2012-08-02 04:56:05.199571
7146	6	1-898	7145	34	63	1153	43	2012-08-02 04:56:07.354419
7147	6	1-2058	7145	63	45	2782	69	2012-08-02 04:56:07.354419
7148	6	1-1434	7145	45	65	668	87	2012-08-02 04:56:07.354419
7149	6	1-2011	7145	65	81	851	92	2012-08-02 04:56:07.354419
7150	6	1-1042	7145	81	24	1175	17	2012-08-02 04:56:07.354419
7145	6	1-2084	7145	24	34	1299	65	2012-08-02 04:56:07.354419
7152	6	1-898	7151	34	36	2876	43	2012-08-02 04:56:07.354419
7153	6	1-1526	7151	36	85	613	21	2012-08-02 04:56:07.354419
7154	6	1-1037	7151	85	26	357	48	2012-08-02 04:56:07.354419
7155	6	1-2037	7151	26	82	493	82	2012-08-02 04:56:07.354419
7156	6	1-2017	7151	82	24	2845	17	2012-08-02 04:56:07.354419
7151	6	1-2084	7151	24	34	3192	65	2012-08-02 04:56:07.354419
7158	6	1-898	7157	34	39	599	43	2012-08-02 04:56:07.354419
7159	6	1-1618	7157	39	95	100	66	2012-08-02 04:56:07.354419
7160	6	1-1946	7157	95	26	68	48	2012-08-02 04:56:07.354419
7161	6	1-2037	7157	26	82	96	82	2012-08-02 04:56:07.354419
7162	6	1-2017	7157	82	24	567	17	2012-08-02 04:56:07.354419
7157	6	1-2084	7157	24	34	650	65	2012-08-02 04:56:07.354419
15987	4	1-4089	15986	35	31	1005	4	2012-08-02 05:13:23.968085
7164	6	1-1868	7163	89	38	183	70	2012-08-02 04:56:07.354419
7165	6	1-1448	7163	38	94	47	84	2012-08-02 04:56:07.354419
7166	6	1-2082	7163	94	26	41	48	2012-08-02 04:56:07.354419
7167	6	1-2037	7163	26	82	62	82	2012-08-02 04:56:07.354419
7168	6	1-2017	7163	82	24	390	17	2012-08-02 04:56:07.354419
7163	6	1-2084	7163	24	89	480	65	2012-08-02 04:56:07.354419
15988	4	1-2965	15986	31	56	1104	12	2012-08-02 05:13:23.968085
7170	6	1-1427	7169	56	23	1090	31	2012-08-02 04:56:07.354419
7171	6	1-1804	7169	23	11	828	45	2012-08-02 04:56:07.354419
7172	6	1-1678	7169	11	50	488	71	2012-08-02 04:56:07.354419
7173	6	1-2034	7169	50	81	416	92	2012-08-02 04:56:07.354419
7174	6	1-1042	7169	81	24	641	17	2012-08-02 04:56:07.354419
7169	6	1-2084	7169	24	56	791	65	2012-08-02 04:56:07.354419
7176	6	1-1998	7175	16	17	1987	75	2012-08-02 04:56:08.249672
7177	6	1-1854	7175	17	32	1178	8	2012-08-02 04:56:08.249672
7178	6	1-1757	7175	32	85	1503	87	2012-08-02 04:56:08.249672
7179	6	1-2038	7175	85	3	1216	38	2012-08-02 04:56:08.249672
7180	6	1-733	7175	3	87	2169	44	2012-08-02 04:56:08.249672
7175	6	1-2087	7175	87	16	2543	60	2012-08-02 04:56:08.249672
7182	3	1-171	7181	72	59	1755	47	2012-08-02 04:56:08.702359
7183	3	1-1909	7181	59	59	4224	74	2012-08-02 04:56:08.702359
7181	3	1-2088	7181	59	72	8776	30	2012-08-02 04:56:08.702359
7185	4	1-1906	7184	35	15	833	55	2012-08-02 04:56:09.133452
7186	4	1-1801	7184	15	31	651	45	2012-08-02 04:56:09.133452
7187	4	1-2071	7184	31	11	922	73	2012-08-02 04:56:09.133452
7184	4	1-2089	7184	11	35	1089	49	2012-08-02 04:56:09.133452
7189	5	1-2035	7188	43	90	6496	20	2012-08-02 04:56:09.133452
7190	5	1-1793	7188	90	65	5343	94	2012-08-02 04:56:09.133452
7191	5	1-1410	7188	65	42	978	57	2012-08-02 04:56:09.133452
7192	5	1-2057	7188	42	11	173	73	2012-08-02 04:56:09.133452
7188	5	1-2089	7188	11	43	215	49	2012-08-02 04:56:09.133452
7194	7	1-2035	7193	43	2	1967	20	2012-08-02 04:56:09.133452
7195	7	1-1193	7193	2	68	287	50	2012-08-02 04:56:09.133452
7196	7	1-1947	7193	68	71	178	80	2012-08-02 04:56:09.133452
7197	7	1-1993	7193	71	94	53	11	2012-08-02 04:56:09.133452
7198	7	1-1571	7193	94	31	30	45	2012-08-02 04:56:09.133452
7199	7	1-2071	7193	31	11	47	73	2012-08-02 04:56:09.133452
7193	7	1-2089	7193	11	43	61	49	2012-08-02 04:56:09.133452
7201	6	1-733	7200	3	87	269	44	2012-08-02 04:56:09.575444
7202	6	1-2087	7200	87	21	311	60	2012-08-02 04:56:09.575444
7203	6	1-621	7200	21	29	421	76	2012-08-02 04:56:09.575444
7204	6	1-1712	7200	29	29	476	53	2012-08-02 04:56:09.575444
7205	6	1-2031	7200	29	45	161	21	2012-08-02 04:56:09.575444
7200	6	1-2090	7200	45	3	153	38	2012-08-02 04:56:09.575444
7207	5	1-1294	7206	64	56	435	84	2012-08-02 04:56:10.923331
7208	5	1-1971	7206	56	89	2333	15	2012-08-02 04:56:10.923331
7209	5	1-579	7206	89	2	609	34	2012-08-02 04:56:10.923331
7210	5	1-2052	7206	2	78	476	67	2012-08-02 04:56:10.923331
7206	5	1-2093	7206	78	64	613	30	2012-08-02 04:56:10.923331
7212	7	1-1255	7211	54	59	1374	77	2012-08-02 04:56:11.354042
7213	7	1-1426	7211	59	1	764	10	2012-08-02 04:56:11.354042
7214	7	1-2054	7211	1	49	434	7	2012-08-02 04:56:11.354042
7215	7	1-1499	7211	49	92	265	78	2012-08-02 04:56:11.354042
7216	7	1-1830	7211	92	59	76	95	2012-08-02 04:56:11.354042
7217	7	1-2028	7211	59	77	101	63	2012-08-02 04:56:11.354042
7211	7	1-2094	7211	77	54	2465	59	2012-08-02 04:56:11.354042
7219	6	1-1397	7218	21	72	135	65	2012-08-02 04:56:11.354042
7220	6	1-1062	7218	72	77	65	22	2012-08-02 04:56:11.354042
7221	6	1-2022	7218	77	92	61	78	2012-08-02 04:56:11.354042
7222	6	1-1830	7218	92	59	17	95	2012-08-02 04:56:11.354042
7223	6	1-2028	7218	59	77	19	63	2012-08-02 04:56:11.354042
7218	6	1-2094	7218	77	21	520	59	2012-08-02 04:56:11.354042
7225	7	1-1796	7224	82	5	2168	54	2012-08-02 04:56:11.354042
7226	7	1-1981	7224	5	71	756	80	2012-08-02 04:56:11.354042
7227	7	1-1993	7224	71	94	205	11	2012-08-02 04:56:11.354042
7228	7	1-1571	7224	94	66	104	45	2012-08-02 04:56:11.354042
7229	7	1-1820	7224	66	59	80	95	2012-08-02 04:56:11.354042
7230	7	1-2028	7224	59	77	105	63	2012-08-02 04:56:11.354042
7224	7	1-2094	7224	77	82	2920	59	2012-08-02 04:56:11.354042
7232	4	1-900	7231	21	65	75	28	2012-08-02 04:56:11.354042
7233	4	1-1951	7231	65	5	71	91	2012-08-02 04:56:11.354042
7234	4	1-221	7231	5	77	60	63	2012-08-02 04:56:11.354042
7231	4	1-2094	7231	77	21	1820	59	2012-08-02 04:56:11.354042
7236	6	1-2069	7235	47	65	2680	28	2012-08-02 04:56:12.226952
7237	6	1-1951	7235	65	32	2616	91	2012-08-02 04:56:12.226952
7238	6	1-1435	7235	32	78	3623	26	2012-08-02 04:56:12.226952
7239	6	1-1050	7235	78	2	4704	34	2012-08-02 04:56:12.226952
7240	6	1-2052	7235	2	56	3645	67	2012-08-02 04:56:12.226952
7235	6	1-2096	7235	56	47	3101	63	2012-08-02 04:56:12.226952
7242	6	1-1979	7241	20	27	2419	83	2012-08-02 04:56:12.668953
7243	6	1-1156	7241	27	62	1158	58	2012-08-02 04:56:12.668953
7244	6	1-1890	7241	62	20	2686	29	2012-08-02 04:56:12.668953
7245	6	1-1853	7241	20	99	2835	32	2012-08-02 04:56:12.668953
7246	6	1-2023	7241	99	17	2636	93	2012-08-02 04:56:12.668953
7241	6	1-2097	7241	17	20	4747	23	2012-08-02 04:56:12.668953
7248	6	1-2030	7247	26	90	284	1	2012-08-02 04:56:12.668953
7249	6	1-1858	7247	90	83	204	9	2012-08-02 04:56:12.668953
7250	6	1-1921	7247	83	95	262	4	2012-08-02 04:56:12.668953
7251	6	1-1976	7247	95	99	207	32	2012-08-02 04:56:12.668953
7252	6	1-2023	7247	99	17	211	93	2012-08-02 04:56:12.668953
7247	6	1-2097	7247	17	26	416	23	2012-08-02 04:56:12.668953
7254	6	1-1370	7253	74	39	1716	43	2012-08-02 04:56:13.111435
7255	6	1-1618	7253	39	43	223	66	2012-08-02 04:56:13.111435
7256	6	1-2072	7253	43	80	584	86	2012-08-02 04:56:13.111435
7257	6	1-1812	7253	80	33	350	32	2012-08-02 04:56:13.111435
7258	6	1-1260	7253	33	15	36	28	2012-08-02 04:56:13.111435
7253	6	1-2098	7253	15	74	3546	2	2012-08-02 04:56:13.111435
7260	5	1-1094	7259	76	80	2566	53	2012-08-02 04:56:13.111435
7261	5	1-348	7259	80	12	220	68	2012-08-02 04:56:13.111435
7262	5	1-1914	7259	12	33	249	32	2012-08-02 04:56:13.111435
7263	5	1-1260	7259	33	15	26	28	2012-08-02 04:56:13.111435
7259	5	1-2098	7259	15	76	2620	2	2012-08-02 04:56:13.111435
15989	4	1-4109	15986	56	10	416	18	2012-08-02 05:13:23.968085
7265	6	1-1607	7264	38	86	1312	39	2012-08-02 04:56:13.553003
7266	6	1-592	7264	86	43	277	66	2012-08-02 04:56:13.553003
7267	6	1-2072	7264	43	75	746	86	2012-08-02 04:56:13.553003
7268	6	1-1825	7264	75	17	1071	93	2012-08-02 04:56:13.553003
7269	6	1-2097	7264	17	69	1612	23	2012-08-02 04:56:13.553003
7264	6	1-2099	7264	69	38	1544	96	2012-08-02 04:56:13.553003
15986	4	1-4605	15986	10	35	1132	86	2012-08-02 05:13:23.968085
7271	6	1-960	7270	26	85	935	56	2012-08-02 04:56:13.553003
7272	6	1-2001	7270	85	30	3288	71	2012-08-02 04:56:13.553003
7273	6	1-1436	7270	30	17	1579	94	2012-08-02 04:56:13.553003
7274	6	1-463	7270	17	17	1121	93	2012-08-02 04:56:13.553003
7275	6	1-2097	7270	17	69	1967	23	2012-08-02 04:56:13.553003
7270	6	1-2099	7270	69	26	1899	96	2012-08-02 04:56:13.553003
7277	6	1-2024	7276	34	39	746	43	2012-08-02 04:56:13.553003
7278	6	1-1618	7276	39	43	112	66	2012-08-02 04:56:13.553003
7279	6	1-2072	7276	43	75	338	86	2012-08-02 04:56:13.553003
7280	6	1-1825	7276	75	17	544	93	2012-08-02 04:56:13.553003
7281	6	1-2097	7276	17	69	918	23	2012-08-02 04:56:13.553003
7276	6	1-2099	7276	69	34	985	96	2012-08-02 04:56:13.553003
7283	4	1-960	7282	26	7	201	56	2012-08-02 04:56:13.553003
7284	4	1-1006	7282	7	17	172	93	2012-08-02 04:56:13.553003
7285	4	1-2097	7282	17	69	306	23	2012-08-02 04:56:13.553003
7282	4	1-2099	7282	69	26	347	96	2012-08-02 04:56:13.553003
7286	1	1-463	\N	17	17	1471	93	2012-08-02 04:56:13.553003
7288	4	1-1480	7287	58	2	9	37	2012-08-02 04:56:14.006213
7289	4	1-2076	7287	2	56	10	99	2012-08-02 04:56:14.006213
7290	4	1-1019	7287	56	47	3	93	2012-08-02 04:56:14.006213
7287	4	1-2100	7287	47	58	11	31	2012-08-02 04:56:14.006213
7292	6	1-1776	7291	99	66	2117	3	2012-08-02 04:56:14.006213
7293	6	1-630	7291	66	2	1312	34	2012-08-02 04:56:14.006213
7294	6	1-2052	7291	2	15	868	67	2012-08-02 04:56:14.006213
7295	6	1-1822	7291	15	75	1186	86	2012-08-02 04:56:14.006213
7296	6	1-1825	7291	75	47	1852	93	2012-08-02 04:56:14.006213
7291	6	1-2100	7291	47	99	6915	31	2012-08-02 04:56:14.006213
7298	6	1-2099	7297	69	26	5202	96	2012-08-02 04:56:17.110722
7299	6	1-960	7297	26	76	2701	56	2012-08-02 04:56:17.110722
7300	6	1-1848	7297	76	59	2446	95	2012-08-02 04:56:17.110722
7301	6	1-2028	7297	59	36	2986	63	2012-08-02 04:56:17.110722
7302	6	1-1186	7297	36	7	2850	50	2012-08-02 04:56:17.110722
7297	6	1-2107	7297	7	69	5112	23	2012-08-02 04:56:17.110722
7304	5	1-1399	7303	32	99	835	32	2012-08-02 04:56:17.541405
7305	5	1-2023	7303	99	47	395	93	2012-08-02 04:56:17.541405
7306	5	1-2100	7303	47	97	824	31	2012-08-02 04:56:17.541405
7307	5	1-935	7303	97	8	533	88	2012-08-02 04:56:17.541405
7303	5	1-2108	7303	8	32	2111	67	2012-08-02 04:56:17.541405
7309	4	1-1309	7308	12	56	972	65	2012-08-02 04:56:17.541405
7310	4	1-1427	7308	56	97	661	31	2012-08-02 04:56:17.541405
7311	4	1-935	7308	97	8	465	88	2012-08-02 04:56:17.541405
7308	4	1-2108	7308	8	12	1998	67	2012-08-02 04:56:17.541405
7313	4	1-2093	7312	78	15	43	30	2012-08-02 04:56:17.541405
7314	4	1-1815	7312	15	97	18	31	2012-08-02 04:56:17.541405
7315	4	1-935	7312	97	8	14	88	2012-08-02 04:56:17.541405
7312	4	1-2108	7312	8	78	62	67	2012-08-02 04:56:17.541405
7317	4	1-1434	7316	45	85	849	87	2012-08-02 04:56:20.193022
7318	4	1-2038	7316	85	56	373	38	2012-08-02 04:56:20.193022
7319	4	1-1919	7316	56	47	339	67	2012-08-02 04:56:20.193022
7316	4	1-2114	7316	47	45	5804	69	2012-08-02 04:56:20.193022
7321	5	1-2037	7320	26	82	230	82	2012-08-02 04:56:22.379957
7322	5	1-2017	7320	82	24	1138	17	2012-08-02 04:56:22.379957
7323	5	1-2084	7320	24	84	1094	65	2012-08-02 04:56:22.379957
7324	5	1-1986	7320	84	75	296	31	2012-08-02 04:56:22.379957
7320	5	1-2119	7320	75	26	194	48	2012-08-02 04:56:22.379957
7326	3	1-1772	7325	33	44	170	25	2012-08-02 04:56:22.379957
7327	3	1-2070	7325	44	75	159	31	2012-08-02 04:56:22.379957
7325	3	1-2119	7325	75	33	114	48	2012-08-02 04:56:22.379957
7329	3	1-2037	7328	26	67	1228	82	2012-08-02 04:56:22.379957
7330	3	1-409	7328	67	75	1005	31	2012-08-02 04:56:22.379957
7328	3	1-2119	7328	75	26	828	48	2012-08-02 04:56:22.379957
7332	5	1-1094	7331	76	29	863	53	2012-08-02 04:56:25.009285
7333	5	1-2031	7331	29	28	254	21	2012-08-02 04:56:25.009285
7334	5	1-1880	7331	28	73	96	98	2012-08-02 04:56:25.009285
7335	5	1-772	7331	73	24	109	88	2012-08-02 04:56:25.009285
7331	5	1-2125	7331	24	76	751	2	2012-08-02 04:56:25.009285
7337	5	1-1182	7336	75	44	1904	25	2012-08-02 04:56:25.440481
7338	5	1-2070	7336	44	75	945	31	2012-08-02 04:56:25.440481
7339	5	1-2119	7336	75	55	538	48	2012-08-02 04:56:25.440481
7340	5	1-1965	7336	55	83	1960	40	2012-08-02 04:56:25.440481
7336	5	1-2126	7336	83	75	4722	99	2012-08-02 04:56:25.440481
7342	3	1-2013	7341	99	4	2529	23	2012-08-02 04:56:27.20812
7343	3	1-2124	7341	4	24	3855	85	2012-08-02 04:56:27.20812
7341	3	1-2130	7341	24	99	4465	87	2012-08-02 04:56:27.20812
7345	5	1-2011	7344	65	43	70	92	2012-08-02 04:56:27.20812
7346	5	1-1587	7344	43	7	40	50	2012-08-02 04:56:27.20812
7347	5	1-2107	7344	7	4	49	23	2012-08-02 04:56:27.20812
7348	5	1-2124	7344	4	24	77	85	2012-08-02 04:56:27.20812
7344	5	1-2130	7344	24	65	92	87	2012-08-02 04:56:27.20812
7350	2	1-1606	7349	92	50	1324	89	2012-08-02 04:56:27.650067
7349	2	1-2131	7349	50	92	9194	70	2012-08-02 04:56:27.650067
7352	5	1-1842	7351	6	2	536	37	2012-08-02 04:56:27.650067
7353	5	1-2076	7351	2	76	663	99	2012-08-02 04:56:27.650067
7354	5	1-2015	7351	76	35	289	32	2012-08-02 04:56:27.650067
7355	5	1-1766	7351	35	50	89	89	2012-08-02 04:56:27.650067
7351	5	1-2131	7351	50	6	684	70	2012-08-02 04:56:27.650067
7357	5	1-1845	7356	34	55	984	48	2012-08-02 04:56:30.281322
7358	5	1-1965	7356	55	27	5783	40	2012-08-02 04:56:30.281322
7359	5	1-1893	7356	27	43	950	92	2012-08-02 04:56:30.281322
7360	5	1-1587	7356	43	84	751	50	2012-08-02 04:56:30.281322
7356	5	1-2138	7356	84	34	1159	91	2012-08-02 04:56:30.281322
7362	3	1-221	7361	5	36	2089	63	2012-08-02 04:56:30.281322
7363	3	1-1186	7361	36	84	1897	50	2012-08-02 04:56:30.281322
7361	3	1-2138	7361	84	5	3027	91	2012-08-02 04:56:30.281322
7364	6	1-2139	7364	45	83	3886	95	2012-08-02 04:56:30.734157
7365	6	1-1781	7364	83	89	2740	15	2012-08-02 04:56:30.734157
7366	6	1-579	7364	89	66	558	34	2012-08-02 04:56:30.734157
7367	6	1-2062	7364	66	90	1180	20	2012-08-02 04:56:30.734157
7368	6	1-1793	7364	90	93	829	94	2012-08-02 04:56:30.734157
7369	6	1-1656	7364	93	45	250	96	2012-08-02 04:56:30.734157
7371	5	1-2028	7370	59	36	117	63	2012-08-02 04:56:31.617899
7372	5	1-1186	7370	36	7	87	50	2012-08-02 04:56:31.617899
7373	5	1-2107	7370	7	4	121	23	2012-08-02 04:56:31.617899
7374	5	1-2124	7370	4	22	221	85	2012-08-02 04:56:31.617899
7370	5	1-2141	7370	22	59	123	95	2012-08-02 04:56:31.617899
15991	5	1-4083	15990	96	56	875	12	2012-08-02 05:13:24.388179
7376	5	1-2028	7375	59	73	2783	63	2012-08-02 04:56:31.617899
7377	5	1-664	7375	73	7	1727	50	2012-08-02 04:56:31.617899
7378	5	1-2107	7375	7	4	2523	23	2012-08-02 04:56:31.617899
7379	5	1-2124	7375	4	22	4814	85	2012-08-02 04:56:31.617899
7375	5	1-2141	7375	22	59	2799	95	2012-08-02 04:56:31.617899
7381	3	1-2028	7380	59	93	3233	63	2012-08-02 04:56:31.617899
7382	3	1-1517	7380	93	22	2507	85	2012-08-02 04:56:31.617899
7380	3	1-2141	7380	22	59	2177	95	2012-08-02 04:56:31.617899
7384	6	1-1258	7383	100	24	154	88	2012-08-02 04:56:31.617899
7385	6	1-2125	7383	24	24	1219	2	2012-08-02 04:56:31.617899
7386	6	1-2112	7383	24	92	978	74	2012-08-02 04:56:31.617899
7387	6	1-1565	7383	92	16	89	7	2012-08-02 04:56:31.617899
7388	6	1-1380	7383	16	22	112	85	2012-08-02 04:56:31.617899
7383	6	1-2141	7383	22	100	99	95	2012-08-02 04:56:31.617899
7390	3	1-1747	7389	31	81	2218	92	2012-08-02 04:56:32.059989
7391	3	1-1042	7389	81	11	2714	17	2012-08-02 04:56:32.059989
7389	3	1-2142	7389	11	31	5802	32	2012-08-02 04:56:32.059989
7393	4	1-1260	7392	33	68	340	28	2012-08-02 04:56:32.059989
7394	4	1-1982	7392	68	82	150	82	2012-08-02 04:56:32.059989
7395	4	1-2017	7392	82	11	930	17	2012-08-02 04:56:32.059989
7392	4	1-2142	7392	11	33	2449	32	2012-08-02 04:56:32.059989
7397	3	1-1731	7396	96	37	1531	18	2012-08-02 04:56:32.502071
7398	3	1-1886	7396	37	28	2062	4	2012-08-02 04:56:32.502071
7396	3	1-2143	7396	28	96	2067	79	2012-08-02 04:56:32.502071
7400	5	1-1690	7399	66	22	707	8	2012-08-02 04:56:32.502071
7401	5	1-22	7399	22	83	1285	33	2012-08-02 04:56:32.502071
7402	5	1-729	7399	83	83	1030	9	2012-08-02 04:56:32.502071
7403	5	1-1921	7399	83	28	1185	4	2012-08-02 04:56:32.502071
7399	5	1-2143	7399	28	66	1301	79	2012-08-02 04:56:32.502071
7405	2	1-1744	7404	81	28	4816	4	2012-08-02 04:56:32.502071
7404	2	1-2143	7404	28	81	5569	79	2012-08-02 04:56:32.502071
7407	4	1-2042	7406	14	6	532	70	2012-08-02 04:56:32.944098
7408	4	1-1842	7406	6	93	427	37	2012-08-02 04:56:32.944098
7409	4	1-2102	7406	93	34	325	64	2012-08-02 04:56:32.944098
7406	4	1-2144	7406	34	14	541	98	2012-08-02 04:56:32.944098
7411	7	1-1449	7410	53	74	663	46	2012-08-02 04:56:32.944098
7412	7	1-2092	7410	74	68	887	82	2012-08-02 04:56:32.944098
7413	7	1-1884	7410	68	6	678	22	2012-08-02 04:56:32.944098
7414	7	1-1578	7410	6	22	706	69	2012-08-02 04:56:32.944098
7415	7	1-1930	7410	22	9	219	19	2012-08-02 04:56:32.944098
7416	7	1-2066	7410	9	34	361	64	2012-08-02 04:56:32.944098
7410	7	1-2144	7410	34	53	605	98	2012-08-02 04:56:32.944098
7418	6	1-631	7417	54	8	178	22	2012-08-02 04:56:32.944098
7419	6	1-2140	7417	8	12	236	55	2012-08-02 04:56:32.944098
7420	6	1-1371	7417	12	90	174	89	2012-08-02 04:56:32.944098
7421	6	1-954	7417	90	46	117	34	2012-08-02 04:56:32.944098
7422	6	1-1937	7417	46	34	134	64	2012-08-02 04:56:32.944098
7417	6	1-2144	7417	34	54	240	98	2012-08-02 04:56:32.944098
7424	6	1-1835	7423	96	27	1675	83	2012-08-02 04:56:35.584555
7425	6	1-1156	7423	27	62	719	58	2012-08-02 04:56:35.584555
7426	6	1-1890	7423	62	27	1192	29	2012-08-02 04:56:35.584555
7427	6	1-2086	7423	27	85	714	78	2012-08-02 04:56:35.584555
7428	6	1-1779	7423	85	91	186	56	2012-08-02 04:56:35.584555
7423	6	1-2150	7423	91	96	3651	72	2012-08-02 04:56:35.584555
7430	5	1-1566	7429	78	42	2002	36	2012-08-02 04:56:36.490255
7431	5	1-860	7429	42	66	1190	34	2012-08-02 04:56:36.490255
7432	5	1-2062	7429	66	90	2865	20	2012-08-02 04:56:36.490255
7433	5	1-1793	7429	90	83	2291	94	2012-08-02 04:56:36.490255
7429	5	1-2152	7429	83	78	3201	77	2012-08-02 04:56:36.490255
7435	5	1-2145	7434	90	8	8812	22	2012-08-02 04:56:37.352265
7436	5	1-2140	7434	8	12	8296	55	2012-08-02 04:56:37.352265
7437	5	1-1371	7434	12	90	4345	89	2012-08-02 04:56:37.352265
7438	5	1-954	7434	90	54	2488	34	2012-08-02 04:56:37.352265
7434	5	1-2154	7434	54	90	3951	1	2012-08-02 04:56:37.352265
7440	5	1-2149	7439	57	96	276	61	2012-08-02 04:56:37.352265
7441	5	1-1250	7439	96	5	1070	6	2012-08-02 04:56:37.352265
7442	5	1-1742	7439	5	89	339	15	2012-08-02 04:56:37.352265
7443	5	1-579	7439	89	54	78	34	2012-08-02 04:56:37.352265
7439	5	1-2154	7439	54	57	152	1	2012-08-02 04:56:37.352265
7445	5	1-2149	7444	57	77	1910	61	2012-08-02 04:56:37.352265
7446	5	1-1509	7444	77	1	925	11	2012-08-02 04:56:37.352265
7447	5	1-1280	7444	1	90	909	89	2012-08-02 04:56:37.352265
7448	5	1-954	7444	90	54	533	34	2012-08-02 04:56:37.352265
7444	5	1-2154	7444	54	57	1043	1	2012-08-02 04:56:37.352265
7450	5	1-2149	7449	57	9	703	61	2012-08-02 04:56:37.352265
7451	5	1-1710	7449	9	21	614	44	2012-08-02 04:56:37.352265
7452	5	1-1017	7449	21	78	142	26	2012-08-02 04:56:37.352265
7453	5	1-1050	7449	78	54	175	34	2012-08-02 04:56:37.352265
7449	5	1-2154	7449	54	57	363	1	2012-08-02 04:56:37.352265
7455	5	1-2149	7454	57	72	28	61	2012-08-02 04:56:37.352265
7456	5	1-364	7454	72	12	15	55	2012-08-02 04:56:37.352265
7457	5	1-1371	7454	12	90	11	89	2012-08-02 04:56:37.352265
7458	5	1-954	7454	90	54	7	34	2012-08-02 04:56:37.352265
7454	5	1-2154	7454	54	57	14	1	2012-08-02 04:56:37.352265
7460	5	1-2072	7459	43	1	4317	86	2012-08-02 04:56:39.098049
7461	5	1-717	7459	1	1	2411	13	2012-08-02 04:56:39.098049
7462	5	1-1689	7459	1	57	448	1	2012-08-02 04:56:39.098049
7463	5	1-2149	7459	57	16	493	61	2012-08-02 04:56:39.098049
7459	5	1-2158	7459	16	43	2328	66	2012-08-02 04:56:39.098049
7465	5	1-1321	7464	99	45	459	21	2012-08-02 04:56:39.098049
7466	5	1-2090	7464	45	78	228	38	2012-08-02 04:56:39.098049
7467	5	1-1900	7464	78	57	127	1	2012-08-02 04:56:39.098049
7468	5	1-2149	7464	57	16	141	61	2012-08-02 04:56:39.098049
7464	5	1-2158	7464	16	99	671	66	2012-08-02 04:56:39.098049
15992	5	1-4109	15990	56	10	349	18	2012-08-02 05:13:24.388179
7470	6	1-1213	7469	53	29	268	53	2012-08-02 04:56:39.098049
7471	6	1-2031	7469	29	45	59	21	2012-08-02 04:56:39.098049
7472	6	1-2090	7469	45	78	37	38	2012-08-02 04:56:39.098049
7473	6	1-1900	7469	78	57	26	1	2012-08-02 04:56:39.098049
7474	6	1-2149	7469	57	16	36	61	2012-08-02 04:56:39.098049
7469	6	1-2158	7469	16	53	216	66	2012-08-02 04:56:39.098049
15993	5	1-4605	15990	10	54	1004	86	2012-08-02 05:13:24.388179
7476	6	1-2123	7475	34	34	579	57	2012-08-02 04:56:40.876702
7477	6	1-42	7475	34	46	453	47	2012-08-02 04:56:40.876702
7478	6	1-2103	7475	46	85	705	78	2012-08-02 04:56:40.876702
7479	6	1-1779	7475	85	91	242	56	2012-08-02 04:56:40.876702
7480	6	1-2150	7475	91	72	6254	72	2012-08-02 04:56:40.876702
7475	6	1-2162	7475	72	34	1079	59	2012-08-02 04:56:40.876702
7482	4	1-2071	7481	31	11	3793	73	2012-08-02 04:56:41.749285
7483	4	1-2089	7481	11	18	4662	49	2012-08-02 04:56:41.749285
7484	4	1-2056	7481	18	34	4749	28	2012-08-02 04:56:41.749285
7481	4	1-2164	7481	34	31	2574	45	2012-08-02 04:56:41.749285
7486	5	1-1143	7485	42	39	485	80	2012-08-02 04:56:42.191149
7487	5	1-782	7485	39	24	226	88	2012-08-02 04:56:42.191149
7488	5	1-2125	7485	24	24	1243	2	2012-08-02 04:56:42.191149
7489	5	1-2112	7485	24	55	693	74	2012-08-02 04:56:42.191149
7485	5	1-2165	7485	55	42	696	42	2012-08-02 04:56:42.191149
7491	6	1-2016	7490	73	7	3356	33	2012-08-02 04:56:42.191149
7492	6	1-1739	7490	7	34	2577	59	2012-08-02 04:56:42.191149
7493	6	1-2123	7490	34	34	1299	57	2012-08-02 04:56:42.191149
7494	6	1-42	7490	34	59	954	47	2012-08-02 04:56:42.191149
7495	6	1-1909	7490	59	55	2730	74	2012-08-02 04:56:42.191149
7490	6	1-2165	7490	55	73	4296	42	2012-08-02 04:56:42.191149
7497	6	1-2016	7496	73	78	231	33	2012-08-02 04:56:42.191149
7498	6	1-1999	7496	78	59	65	83	2012-08-02 04:56:42.191149
7499	6	1-912	7496	59	74	44	46	2012-08-02 04:56:42.191149
7500	6	1-2092	7496	74	91	57	82	2012-08-02 04:56:42.191149
7501	6	1-1196	7496	91	55	132	74	2012-08-02 04:56:42.191149
7496	6	1-2165	7496	55	73	248	42	2012-08-02 04:56:42.191149
7503	5	1-1990	7502	24	18	4293	49	2012-08-02 04:56:43.053002
7504	5	1-2056	7502	18	34	4733	28	2012-08-02 04:56:43.053002
7505	5	1-2164	7502	34	31	2776	45	2012-08-02 04:56:43.053002
7506	5	1-2071	7502	31	32	4427	73	2012-08-02 04:56:43.053002
7502	5	1-2167	7502	32	24	4751	97	2012-08-02 04:56:43.053002
7508	4	1-2151	7507	12	27	2839	83	2012-08-02 04:56:43.903774
7509	4	1-1156	7507	27	93	1099	58	2012-08-02 04:56:43.903774
7510	4	1-1748	7507	93	61	1006	50	2012-08-02 04:56:43.903774
7507	4	1-2169	7507	61	12	5255	31	2012-08-02 04:56:43.903774
7512	6	1-2119	7511	75	42	3253	48	2012-08-02 04:56:43.903774
7513	6	1-1613	7511	42	35	1972	27	2012-08-02 04:56:43.903774
7514	6	1-2044	7511	35	100	1175	60	2012-08-02 04:56:43.903774
7515	6	1-1844	7511	100	93	737	58	2012-08-02 04:56:43.903774
7516	6	1-1748	7511	93	61	774	50	2012-08-02 04:56:43.903774
7511	6	1-2169	7511	61	75	4644	31	2012-08-02 04:56:43.903774
7518	6	1-960	7517	26	56	41	56	2012-08-02 04:56:44.765816
7519	6	1-2010	7517	56	44	76	25	2012-08-02 04:56:44.765816
7520	6	1-2070	7517	44	99	79	31	2012-08-02 04:56:44.765816
7521	6	1-1776	7517	99	66	28	3	2012-08-02 04:56:44.765816
7522	6	1-630	7517	66	10	20	34	2012-08-02 04:56:44.765816
7517	6	1-2171	7517	10	26	66	96	2012-08-02 04:56:44.765816
7524	3	1-1021	7523	60	11	2334	66	2012-08-02 04:56:46.964093
7525	3	1-2081	7523	11	83	2659	79	2012-08-02 04:56:46.964093
7523	3	1-2176	7523	83	60	6429	35	2012-08-02 04:56:46.964093
7527	6	1-1607	7526	38	27	2619	39	2012-08-02 04:56:48.278514
7528	6	1-1903	7526	27	22	3510	69	2012-08-02 04:56:48.278514
7529	6	1-1930	7526	22	9	1039	19	2012-08-02 04:56:48.278514
7530	6	1-2066	7526	9	88	1640	64	2012-08-02 04:56:48.278514
7531	6	1-1637	7526	88	53	1209	9	2012-08-02 04:56:48.278514
7526	6	1-2179	7526	53	38	2624	96	2012-08-02 04:56:48.278514
7533	6	1-960	7532	26	35	105	56	2012-08-02 04:56:48.278514
7534	6	1-1774	7532	35	32	92	44	2012-08-02 04:56:48.278514
7535	6	1-1510	7532	32	73	114	42	2012-08-02 04:56:48.278514
7536	6	1-2016	7532	73	83	105	33	2012-08-02 04:56:48.278514
7537	6	1-729	7532	83	53	84	9	2012-08-02 04:56:48.278514
7532	6	1-2179	7532	53	26	183	96	2012-08-02 04:56:48.278514
7539	5	1-1814	7538	57	17	1328	75	2012-08-02 04:56:48.278514
7540	5	1-1854	7538	17	22	740	8	2012-08-02 04:56:48.278514
7541	5	1-22	7538	22	83	1661	33	2012-08-02 04:56:48.278514
7542	5	1-729	7538	83	53	1435	9	2012-08-02 04:56:48.278514
7538	5	1-2179	7538	53	57	3357	96	2012-08-02 04:56:48.278514
7544	7	1-1578	7543	6	64	218	69	2012-08-02 04:56:49.162466
7545	7	1-1125	7543	64	29	152	53	2012-08-02 04:56:49.162466
7546	7	1-2031	7543	29	45	47	21	2012-08-02 04:56:49.162466
7547	7	1-2090	7543	45	56	41	38	2012-08-02 04:56:49.162466
7548	7	1-1919	7543	56	12	63	67	2012-08-02 04:56:49.162466
7549	7	1-1309	7543	12	47	57	65	2012-08-02 04:56:49.162466
7543	7	1-2181	7543	47	6	210	22	2012-08-02 04:56:49.162466
7551	7	1-2022	7550	77	92	3835	78	2012-08-02 04:56:49.162466
7552	7	1-1830	7550	92	99	1459	95	2012-08-02 04:56:49.162466
7553	7	1-1351	7550	99	45	604	21	2012-08-02 04:56:49.162466
7554	7	1-2090	7550	45	56	567	38	2012-08-02 04:56:49.162466
7555	7	1-1919	7550	56	12	923	67	2012-08-02 04:56:49.162466
7556	7	1-1309	7550	12	47	892	65	2012-08-02 04:56:49.162466
7550	7	1-2181	7550	47	77	3522	22	2012-08-02 04:56:49.162466
7558	6	1-2038	7557	85	56	269	38	2012-08-02 04:56:49.604397
7559	6	1-1919	7557	56	56	445	67	2012-08-02 04:56:49.604397
7560	6	1-2096	7557	56	39	397	63	2012-08-02 04:56:49.604397
7561	6	1-1925	7557	39	78	418	77	2012-08-02 04:56:49.604397
7562	6	1-1566	7557	78	74	303	36	2012-08-02 04:56:49.604397
7557	6	1-2182	7557	74	85	338	87	2012-08-02 04:56:49.604397
7564	6	1-1844	7563	100	62	1720	58	2012-08-02 04:56:50.046487
7565	6	1-1890	7563	62	27	4410	29	2012-08-02 04:56:50.046487
7566	6	1-2086	7563	27	85	3254	78	2012-08-02 04:56:50.046487
7567	6	1-1779	7563	85	56	1310	56	2012-08-02 04:56:50.046487
7568	6	1-2010	7563	56	94	2559	25	2012-08-02 04:56:50.046487
7563	6	1-2183	7563	94	100	2305	60	2012-08-02 04:56:50.046487
7570	5	1-1014	7569	88	2	592	97	2012-08-02 04:56:50.046487
7571	5	1-2068	7569	2	33	1311	57	2012-08-02 04:56:50.046487
7572	5	1-2137	7569	33	56	685	56	2012-08-02 04:56:50.046487
7573	5	1-2010	7569	56	94	1358	25	2012-08-02 04:56:50.046487
7569	5	1-2183	7569	94	88	1240	60	2012-08-02 04:56:50.046487
15994	5	1-4110	15990	54	4	825	92	2012-08-02 05:13:24.388179
7575	4	1-1925	7574	39	87	1004	77	2012-08-02 04:56:51.79211
7576	4	1-1632	7574	87	17	810	48	2012-08-02 04:56:51.79211
7577	4	1-2047	7574	17	5	1718	18	2012-08-02 04:56:51.79211
7574	4	1-2187	7574	5	39	1061	63	2012-08-02 04:56:51.79211
15990	5	1-4606	15990	4	96	1750	66	2012-08-02 05:13:24.388179
7579	5	1-1955	7578	42	93	838	35	2012-08-02 04:56:52.233924
7580	5	1-1904	7578	93	98	428	91	2012-08-02 04:56:52.233924
7581	5	1-1671	7578	98	32	2213	13	2012-08-02 04:56:52.233924
7582	5	1-2120	7578	32	53	4163	41	2012-08-02 04:56:52.233924
7578	5	1-2188	7578	53	42	1787	38	2012-08-02 04:56:52.233924
7584	6	1-1884	7583	68	85	535	22	2012-08-02 04:56:53.095975
7585	6	1-1713	7583	85	83	242	33	2012-08-02 04:56:53.095975
7586	6	1-729	7583	83	53	182	9	2012-08-02 04:56:53.095975
7587	6	1-2179	7583	53	34	370	96	2012-08-02 04:56:53.095975
7588	6	1-2024	7583	34	38	275	43	2012-08-02 04:56:53.095975
7583	6	1-2190	7583	38	68	781	82	2012-08-02 04:56:53.095975
15996	2	1-4396	15995	29	74	1212	39	2012-08-02 05:13:28.112962
7590	6	1-574	7589	39	36	3343	29	2012-08-02 04:56:53.095975
7591	6	1-1823	7589	36	53	4466	41	2012-08-02 04:56:53.095975
7592	6	1-2188	7589	53	56	1710	38	2012-08-02 04:56:53.095975
7593	6	1-1919	7589	56	74	2426	67	2012-08-02 04:56:53.095975
7594	6	1-436	7589	74	38	1660	43	2012-08-02 04:56:53.095975
7589	6	1-2190	7589	38	39	4903	82	2012-08-02 04:56:53.095975
7596	5	1-2026	7595	52	85	171	87	2012-08-02 04:56:53.095975
7597	5	1-2038	7595	85	56	120	38	2012-08-02 04:56:53.095975
7598	5	1-1919	7595	56	74	174	67	2012-08-02 04:56:53.095975
7599	5	1-436	7595	74	38	122	43	2012-08-02 04:56:53.095975
7595	5	1-2190	7595	38	52	367	82	2012-08-02 04:56:53.095975
7600	1	1-729	\N	83	83	1942	9	2012-08-02 04:56:53.095975
7602	6	1-2070	7601	44	12	214	31	2012-08-02 04:56:53.957867
7603	6	1-2151	7601	12	59	114	83	2012-08-02 04:56:53.957867
7604	6	1-912	7601	59	74	62	46	2012-08-02 04:56:53.957867
7605	6	1-2092	7601	74	82	65	82	2012-08-02 04:56:53.957867
7606	6	1-2017	7601	82	83	294	17	2012-08-02 04:56:53.957867
7601	6	1-2192	7601	83	44	272	25	2012-08-02 04:56:53.957867
7608	6	1-2183	7607	94	84	101	60	2012-08-02 04:56:53.957867
7609	6	1-2004	7607	84	32	69	44	2012-08-02 04:56:53.957867
7610	6	1-1510	7607	32	73	88	42	2012-08-02 04:56:53.957867
7611	6	1-2016	7607	73	25	84	33	2012-08-02 04:56:53.957867
7612	6	1-1651	7607	25	83	104	17	2012-08-02 04:56:53.957867
7607	6	1-2192	7607	83	94	122	25	2012-08-02 04:56:53.957867
7613	1	1-912	\N	59	59	350	46	2012-08-02 04:56:53.957867
7615	3	1-2141	7614	22	100	1859	95	2012-08-02 04:56:56.134017
7616	3	1-1258	7614	100	77	2666	88	2012-08-02 04:56:56.134017
7614	3	1-2197	7614	77	22	2282	85	2012-08-02 04:56:56.134017
7618	6	1-1371	7617	12	64	542	89	2012-08-02 04:56:56.564843
7619	6	1-2191	7617	64	41	617	37	2012-08-02 04:56:56.564843
7620	6	1-2122	7617	41	23	2385	16	2012-08-02 04:56:56.564843
7621	6	1-1960	7617	23	96	825	9	2012-08-02 04:56:56.564843
7622	6	1-2111	7617	96	72	815	85	2012-08-02 04:56:56.564843
7617	6	1-2198	7617	72	12	739	55	2012-08-02 04:56:56.564843
7624	3	1-2016	7623	73	25	2748	33	2012-08-02 04:56:57.868862
7625	3	1-1651	7623	25	39	3150	17	2012-08-02 04:56:57.868862
7623	3	1-2201	7623	39	73	3122	42	2012-08-02 04:56:57.868862
7627	7	1-2186	7626	76	29	1502	53	2012-08-02 04:56:59.594059
7628	7	1-2031	7626	29	45	501	21	2012-08-02 04:56:59.594059
7629	7	1-2090	7626	45	56	472	38	2012-08-02 04:56:59.594059
7630	7	1-1919	7626	56	78	773	67	2012-08-02 04:56:59.594059
7631	7	1-2093	7626	78	64	1021	30	2012-08-02 04:56:59.594059
7632	7	1-1294	7626	64	53	742	84	2012-08-02 04:56:59.594059
7626	7	1-2205	7626	53	76	1001	44	2012-08-02 04:56:59.594059
7633	1	1-1919	\N	56	56	577	67	2012-08-02 04:56:59.594059
7635	5	1-497	7634	39	62	461	3	2012-08-02 04:57:01.372388
7636	5	1-1636	7634	62	79	412	55	2012-08-02 04:57:01.372388
7637	5	1-2195	7634	79	82	394	82	2012-08-02 04:57:01.372388
7638	5	1-2017	7634	82	21	2271	17	2012-08-02 04:57:01.372388
7634	5	1-2209	7634	21	39	1965	44	2012-08-02 04:57:01.372388
7639	1	1-1023	\N	3	3	277	82	2012-08-02 04:57:01.372388
7641	6	1-1435	7640	32	78	733	26	2012-08-02 04:57:01.814577
7642	6	1-1050	7640	78	10	798	34	2012-08-02 04:57:01.814577
7643	6	1-2171	7640	10	34	2222	96	2012-08-02 04:57:01.814577
7644	6	1-2024	7640	34	20	1605	43	2012-08-02 04:57:01.814577
7645	6	1-2166	7640	20	21	657	70	2012-08-02 04:57:01.814577
7640	6	1-2210	7640	21	32	631	91	2012-08-02 04:57:01.814577
7647	4	1-221	7646	5	47	4086	63	2012-08-02 04:57:01.814577
7648	4	1-2069	7646	47	70	3582	28	2012-08-02 04:57:01.814577
7649	4	1-1681	7646	70	21	3986	70	2012-08-02 04:57:01.814577
7646	4	1-2210	7646	21	5	4630	91	2012-08-02 04:57:01.814577
7651	6	1-2158	7650	16	4	477	66	2012-08-02 04:57:03.968706
7652	6	1-2105	7650	4	70	295	28	2012-08-02 04:57:03.968706
7653	6	1-1681	7650	70	21	290	70	2012-08-02 04:57:03.968706
7654	6	1-2210	7650	21	32	297	91	2012-08-02 04:57:03.968706
7655	6	1-1435	7650	32	84	368	26	2012-08-02 04:57:03.968706
7650	6	1-2215	7650	84	16	61	61	2012-08-02 04:57:03.968706
7657	5	1-1050	7656	78	98	1034	34	2012-08-02 04:57:07.018686
7658	5	1-1970	7656	98	70	934	89	2012-08-02 04:57:07.018686
7659	5	1-873	7656	70	35	781	27	2012-08-02 04:57:07.018686
7660	5	1-2044	7656	35	56	487	60	2012-08-02 04:57:07.018686
7656	5	1-2222	7656	56	78	873	26	2012-08-02 04:57:07.018686
7662	6	1-1513	7661	18	44	639	25	2012-08-02 04:57:07.018686
7663	6	1-2070	7661	44	75	650	31	2012-08-02 04:57:07.018686
7664	6	1-2119	7661	75	42	508	48	2012-08-02 04:57:07.018686
7665	6	1-1613	7661	42	35	343	27	2012-08-02 04:57:07.018686
7666	6	1-2044	7661	35	56	228	60	2012-08-02 04:57:07.018686
7661	6	1-2222	7661	56	18	436	26	2012-08-02 04:57:07.018686
7668	5	1-1638	7667	91	96	69	79	2012-08-02 04:57:07.018686
7669	5	1-1731	7667	96	66	68	18	2012-08-02 04:57:07.018686
7670	5	1-1679	7667	66	35	61	27	2012-08-02 04:57:07.018686
7671	5	1-2044	7667	35	56	41	60	2012-08-02 04:57:07.018686
7667	5	1-2222	7667	56	91	79	26	2012-08-02 04:57:07.018686
7673	6	1-1050	7672	78	65	1013	34	2012-08-02 04:57:07.018686
7674	6	1-1997	7672	65	96	682	79	2012-08-02 04:57:07.018686
7675	6	1-1731	7672	96	66	678	18	2012-08-02 04:57:07.018686
7676	6	1-1679	7672	66	35	609	27	2012-08-02 04:57:07.018686
7677	6	1-2044	7672	35	56	410	60	2012-08-02 04:57:07.018686
7672	6	1-2222	7672	56	78	793	26	2012-08-02 04:57:07.018686
15995	2	1-4615	15995	74	29	4382	85	2012-08-02 05:13:28.112962
7679	6	1-2215	7678	84	16	565	61	2012-08-02 04:57:07.018686
7680	6	1-2158	7678	16	95	4897	66	2012-08-02 04:57:07.018686
7681	6	1-1946	7678	95	42	3382	48	2012-08-02 04:57:07.018686
7682	6	1-1613	7678	42	35	2331	27	2012-08-02 04:57:07.018686
7683	6	1-2044	7678	35	56	1579	60	2012-08-02 04:57:07.018686
7678	6	1-2222	7678	56	84	3077	26	2012-08-02 04:57:07.018686
7685	6	1-630	7684	66	2	1525	34	2012-08-02 04:57:07.483176
7686	6	1-2052	7684	2	78	1034	67	2012-08-02 04:57:07.483176
7687	6	1-2093	7684	78	64	1155	30	2012-08-02 04:57:07.483176
7688	6	1-1294	7684	64	65	709	84	2012-08-02 04:57:07.483176
7689	6	1-1939	7684	65	4	1322	94	2012-08-02 04:57:07.483176
7684	6	1-2223	7684	4	66	2400	3	2012-08-02 04:57:07.483176
15998	4	1-4452	15997	46	85	526	44	2012-08-02 05:13:29.759012
7691	4	1-1636	7690	62	79	954	55	2012-08-02 04:57:07.483176
7692	4	1-2195	7690	79	49	889	82	2012-08-02 04:57:07.483176
7693	4	1-707	7690	49	4	567	94	2012-08-02 04:57:07.483176
7690	4	1-2223	7690	4	62	1093	3	2012-08-02 04:57:07.483176
15999	4	1-3915	15997	85	56	404	97	2012-08-02 05:13:29.759012
7695	4	1-2160	7694	52	20	3903	12	2012-08-02 04:57:12.211161
7696	4	1-2074	7694	20	5	3723	6	2012-08-02 04:57:12.211161
7697	4	1-1742	7694	5	69	792	15	2012-08-02 04:57:12.211161
7694	4	1-2234	7694	69	52	5666	52	2012-08-02 04:57:12.211161
7699	5	1-2181	7698	47	77	3207	22	2012-08-02 04:57:12.6536
7700	5	1-2022	7698	77	85	2715	78	2012-08-02 04:57:12.6536
7701	5	1-1779	7698	85	77	639	56	2012-08-02 04:57:12.6536
7702	5	1-2199	7698	77	49	632	15	2012-08-02 04:57:12.6536
7698	5	1-2235	7698	49	47	1379	65	2012-08-02 04:57:12.6536
7704	6	1-2181	7703	47	77	265	22	2012-08-02 04:57:12.6536
7705	6	1-2022	7703	77	62	192	78	2012-08-02 04:57:12.6536
7706	6	1-1756	7703	62	47	25	63	2012-08-02 04:57:12.6536
7707	6	1-2069	7703	47	76	18	28	2012-08-02 04:57:12.6536
7708	6	1-1715	7703	76	49	33	15	2012-08-02 04:57:12.6536
7703	6	1-2235	7703	49	47	81	65	2012-08-02 04:57:12.6536
7710	4	1-1575	7709	69	33	1504	57	2012-08-02 04:57:12.6536
7711	4	1-2137	7709	33	77	713	56	2012-08-02 04:57:12.6536
7712	4	1-2199	7709	77	49	841	15	2012-08-02 04:57:12.6536
7709	4	1-2235	7709	49	69	2892	65	2012-08-02 04:57:12.6536
7714	5	1-2213	7713	21	95	2595	4	2012-08-02 04:57:12.6536
7715	5	1-1976	7713	95	63	1933	32	2012-08-02 04:57:12.6536
7716	5	1-1827	7713	63	5	2193	6	2012-08-02 04:57:12.6536
7717	5	1-1742	7713	5	49	754	15	2012-08-02 04:57:12.6536
7713	5	1-2235	7713	49	21	2629	65	2012-08-02 04:57:12.6536
7719	6	1-1062	7718	72	77	234	22	2012-08-02 04:57:12.6536
7720	6	1-2022	7718	77	62	245	78	2012-08-02 04:57:12.6536
7721	6	1-1756	7718	62	47	46	63	2012-08-02 04:57:12.6536
7722	6	1-2069	7718	47	76	39	28	2012-08-02 04:57:12.6536
7723	6	1-1715	7718	76	49	103	15	2012-08-02 04:57:12.6536
7718	6	1-2235	7718	49	72	367	65	2012-08-02 04:57:12.6536
7725	5	1-2213	7724	21	95	77	4	2012-08-02 04:57:12.6536
7726	5	1-1976	7724	95	33	59	32	2012-08-02 04:57:12.6536
7727	5	1-1260	7724	33	76	8	28	2012-08-02 04:57:12.6536
7728	5	1-1715	7724	76	49	21	15	2012-08-02 04:57:12.6536
7724	5	1-2235	7724	49	21	76	65	2012-08-02 04:57:12.6536
7730	6	1-2213	7729	21	95	418	4	2012-08-02 04:57:12.6536
7731	6	1-1976	7729	95	35	331	32	2012-08-02 04:57:12.6536
7732	6	1-1766	7729	35	68	112	89	2012-08-02 04:57:12.6536
7733	6	1-2014	7729	68	71	126	13	2012-08-02 04:57:12.6536
7734	6	1-1267	7729	71	49	108	15	2012-08-02 04:57:12.6536
7729	6	1-2235	7729	49	21	399	65	2012-08-02 04:57:12.6536
7736	3	1-579	7735	89	10	929	34	2012-08-02 04:57:13.968161
7737	3	1-2171	7735	10	67	2126	96	2012-08-02 04:57:13.968161
7735	3	1-2238	7735	67	89	5203	15	2012-08-02 04:57:13.968161
7739	7	1-2235	7738	49	72	2124	65	2012-08-02 04:57:13.968161
7740	7	1-1062	7738	72	64	1260	22	2012-08-02 04:57:13.968161
7741	7	1-1834	7738	64	39	423	44	2012-08-02 04:57:13.968161
7742	7	1-497	7738	39	66	96	3	2012-08-02 04:57:13.968161
7743	7	1-630	7738	66	10	64	34	2012-08-02 04:57:13.968161
7744	7	1-2171	7738	10	67	196	96	2012-08-02 04:57:13.968161
7738	7	1-2238	7738	67	49	640	15	2012-08-02 04:57:13.968161
7746	3	1-2093	7745	78	15	1851	30	2012-08-02 04:57:14.399193
7747	3	1-1815	7745	15	59	1381	31	2012-08-02 04:57:14.399193
7745	3	1-2239	7745	59	78	1512	67	2012-08-02 04:57:14.399193
7749	4	1-1716	7748	96	6	722	54	2012-08-02 04:57:14.399193
7750	4	1-1811	7748	6	41	825	16	2012-08-02 04:57:14.399193
7751	4	1-307	7748	41	59	368	31	2012-08-02 04:57:14.399193
7748	4	1-2239	7748	59	96	425	67	2012-08-02 04:57:14.399193
7753	6	1-1298	7752	47	54	64	29	2012-08-02 04:57:14.399193
7754	6	1-2110	7752	54	20	35	24	2012-08-02 04:57:14.399193
7755	6	1-739	7752	20	83	51	17	2012-08-02 04:57:14.399193
7756	6	1-2192	7752	83	44	65	25	2012-08-02 04:57:14.399193
7757	6	1-2070	7752	44	59	70	31	2012-08-02 04:57:14.399193
7752	6	1-2239	7752	59	47	82	67	2012-08-02 04:57:14.399193
7759	6	1-2066	7758	9	34	277	64	2012-08-02 04:57:14.841122
7760	6	1-2144	7758	34	99	386	98	2012-08-02 04:57:14.841122
7761	6	1-1292	7758	99	75	179	47	2012-08-02 04:57:14.841122
7762	6	1-2115	7758	75	66	155	18	2012-08-02 04:57:14.841122
7763	6	1-1679	7758	66	67	113	27	2012-08-02 04:57:14.841122
7758	6	1-2240	7758	67	9	202	19	2012-08-02 04:57:14.841122
7765	3	1-1899	7764	45	55	2947	46	2012-08-02 04:57:14.841122
7766	3	1-1881	7764	55	67	3061	27	2012-08-02 04:57:14.841122
7764	3	1-2240	7764	67	45	5797	19	2012-08-02 04:57:14.841122
7768	5	1-1943	7767	63	93	344	37	2012-08-02 04:57:14.841122
7769	5	1-2102	7767	93	84	261	64	2012-08-02 04:57:14.841122
7770	5	1-756	7767	84	42	201	48	2012-08-02 04:57:14.841122
7771	5	1-1613	7767	42	67	133	27	2012-08-02 04:57:14.841122
7767	5	1-2240	7767	67	63	281	19	2012-08-02 04:57:14.841122
7773	5	1-2231	7772	47	99	3193	98	2012-08-02 04:57:14.841122
7774	5	1-1292	7772	99	75	1804	47	2012-08-02 04:57:14.841122
7775	5	1-2115	7772	75	66	1910	18	2012-08-02 04:57:14.841122
7776	5	1-1679	7772	66	67	1704	27	2012-08-02 04:57:14.841122
7772	5	1-2240	7772	67	47	3699	19	2012-08-02 04:57:14.841122
16000	4	1-4464	15997	56	47	1205	40	2012-08-02 05:13:29.759012
7778	6	1-1731	7777	96	5	1161	18	2012-08-02 04:57:16.12293
7779	6	1-2187	7777	5	47	794	63	2012-08-02 04:57:16.12293
7780	6	1-2069	7777	47	92	708	28	2012-08-02 04:57:16.12293
7781	6	1-1130	7777	92	2	567	37	2012-08-02 04:57:16.12293
7782	6	1-2076	7777	2	83	776	99	2012-08-02 04:57:16.12293
7777	6	1-2243	7777	83	96	1113	79	2012-08-02 04:57:16.12293
15997	4	1-4619	15997	47	46	1280	58	2012-08-02 05:13:29.759012
7784	6	1-1893	7783	27	43	415	92	2012-08-02 04:57:16.996004
7785	6	1-1587	7783	43	25	396	50	2012-08-02 04:57:16.996004
7786	6	1-1966	7783	25	93	167	35	2012-08-02 04:57:16.996004
7787	6	1-1904	7783	93	36	79	91	2012-08-02 04:57:16.996004
7788	6	1-743	7783	36	44	684	14	2012-08-02 04:57:16.996004
7783	6	1-2245	7783	44	27	2089	40	2012-08-02 04:57:16.996004
7790	2	1-2070	7789	44	29	5318	31	2012-08-02 04:57:17.426569
7789	2	1-2246	7789	29	44	5074	25	2012-08-02 04:57:17.426569
16002	4	1-4284	16001	18	14	1418	11	2012-08-02 05:13:29.759012
7792	7	1-2056	7791	18	34	156	28	2012-08-02 04:57:18.288272
7793	7	1-2164	7791	34	31	92	45	2012-08-02 04:57:18.288272
7794	7	1-2071	7791	31	32	147	73	2012-08-02 04:57:18.288272
7795	7	1-2167	7791	32	2	158	97	2012-08-02 04:57:18.288272
7796	7	1-2068	7791	2	33	328	57	2012-08-02 04:57:18.288272
7797	7	1-2137	7791	33	99	161	56	2012-08-02 04:57:18.288272
7791	7	1-2248	7791	99	18	141	49	2012-08-02 04:57:18.288272
7799	5	1-860	7798	42	46	984	34	2012-08-02 04:57:19.183123
7800	5	1-1937	7798	46	34	606	64	2012-08-02 04:57:19.183123
7801	5	1-2144	7798	34	15	580	98	2012-08-02 04:57:19.183123
7802	5	1-1846	7798	15	2	346	32	2012-08-02 04:57:19.183123
7798	5	1-2250	7798	2	42	2699	36	2012-08-02 04:57:19.183123
7804	5	1-2182	7803	74	85	899	87	2012-08-02 04:57:19.183123
7805	5	1-2038	7803	85	78	385	38	2012-08-02 04:57:19.183123
7806	5	1-1900	7803	78	58	219	1	2012-08-02 04:57:19.183123
7807	5	1-2121	7803	58	2	189	32	2012-08-02 04:57:19.183123
7803	5	1-2250	7803	2	74	1498	36	2012-08-02 04:57:19.183123
7809	6	1-2182	7808	74	30	1569	87	2012-08-02 04:57:19.183123
7810	6	1-2230	7808	30	64	3374	22	2012-08-02 04:57:19.183123
7811	6	1-1834	7808	64	87	729	44	2012-08-02 04:57:19.183123
7812	6	1-2087	7808	87	69	479	60	2012-08-02 04:57:19.183123
7813	6	1-1320	7808	69	2	294	32	2012-08-02 04:57:19.183123
7808	6	1-2250	7808	2	74	2470	36	2012-08-02 04:57:19.183123
7815	4	1-1879	7814	59	87	49	44	2012-08-02 04:57:19.183123
7816	4	1-2087	7814	87	69	34	60	2012-08-02 04:57:19.183123
7817	4	1-1320	7814	69	2	22	32	2012-08-02 04:57:19.183123
7814	4	1-2250	7814	2	59	191	36	2012-08-02 04:57:19.183123
7819	5	1-2156	7818	89	62	5451	88	2012-08-02 04:57:22.696334
7820	5	1-2157	7818	62	78	2279	78	2012-08-02 04:57:22.696334
7821	5	1-1974	7818	78	63	479	8	2012-08-02 04:57:22.696334
7822	5	1-690	7818	63	63	631	92	2012-08-02 04:57:22.696334
7818	5	1-2258	7818	63	89	766	66	2012-08-02 04:57:22.696334
7824	6	1-1834	7823	64	87	248	44	2012-08-02 04:57:23.127743
7825	6	1-2087	7823	87	56	285	60	2012-08-02 04:57:23.127743
7826	6	1-2222	7823	56	33	581	26	2012-08-02 04:57:23.127743
7827	6	1-486	7823	33	83	963	17	2012-08-02 04:57:23.127743
7828	6	1-2192	7823	83	94	1231	25	2012-08-02 04:57:23.127743
7823	6	1-2259	7823	94	64	655	22	2012-08-02 04:57:23.127743
7830	6	1-1834	7829	64	79	298	44	2012-08-02 04:57:23.127743
7831	6	1-2051	7829	79	56	335	60	2012-08-02 04:57:23.127743
7832	6	1-2222	7829	56	33	686	26	2012-08-02 04:57:23.127743
7833	6	1-486	7829	33	83	1143	17	2012-08-02 04:57:23.127743
7834	6	1-2192	7829	83	94	1467	25	2012-08-02 04:57:23.127743
7829	6	1-2259	7829	94	64	784	22	2012-08-02 04:57:23.127743
7836	5	1-1036	7835	91	47	1866	81	2012-08-02 04:57:23.558207
7837	5	1-1916	7835	47	66	131	3	2012-08-02 04:57:23.558207
7838	5	1-630	7835	66	10	73	34	2012-08-02 04:57:23.558207
7839	5	1-2171	7835	10	86	183	96	2012-08-02 04:57:23.558207
7835	5	1-2260	7835	86	91	564	38	2012-08-02 04:57:23.558207
7841	6	1-1900	7840	78	58	906	1	2012-08-02 04:57:23.558207
7842	6	1-2121	7840	58	35	1086	32	2012-08-02 04:57:23.558207
7843	6	1-1766	7840	35	90	278	89	2012-08-02 04:57:23.558207
7844	6	1-954	7840	90	10	141	34	2012-08-02 04:57:23.558207
7845	6	1-2171	7840	10	86	364	96	2012-08-02 04:57:23.558207
7840	6	1-2260	7840	86	78	1145	38	2012-08-02 04:57:23.558207
7847	7	1-1117	7846	66	83	1799	58	2012-08-02 04:57:23.558207
7848	7	1-2247	7846	83	45	1614	75	2012-08-02 04:57:23.558207
7849	7	1-1247	7846	45	65	1575	18	2012-08-02 04:57:23.558207
7850	7	1-1391	7846	65	66	516	3	2012-08-02 04:57:23.558207
7851	7	1-630	7846	66	10	341	34	2012-08-02 04:57:23.558207
7852	7	1-2171	7846	10	86	885	96	2012-08-02 04:57:23.558207
7846	7	1-2260	7846	86	66	2795	38	2012-08-02 04:57:23.558207
7853	1	1-630	\N	66	66	1481	34	2012-08-02 04:57:23.558207
7855	3	1-1610	7854	70	84	2960	54	2012-08-02 04:57:25.303843
7856	3	1-1792	7854	84	2	4133	14	2012-08-02 04:57:25.303843
7854	3	1-2264	7854	2	70	3927	32	2012-08-02 04:57:25.303843
7858	2	1-2082	7857	94	95	3748	48	2012-08-02 04:57:25.723937
7857	2	1-2265	7857	95	94	9785	84	2012-08-02 04:57:25.723937
7860	6	1-2164	7859	34	31	279	45	2012-08-02 04:57:27.027656
7861	6	1-2071	7859	31	32	443	73	2012-08-02 04:57:27.027656
7862	6	1-2167	7859	32	3	473	97	2012-08-02 04:57:27.027656
7863	6	1-1926	7859	3	83	455	58	2012-08-02 04:57:27.027656
7864	6	1-2247	7859	83	96	512	75	2012-08-02 04:57:27.027656
7859	6	1-2268	7859	96	34	478	28	2012-08-02 04:57:27.027656
7866	5	1-2141	7865	22	100	184	95	2012-08-02 04:57:28.365897
7867	5	1-1258	7865	100	24	130	88	2012-08-02 04:57:28.365897
7868	5	1-2125	7865	24	24	696	2	2012-08-02 04:57:28.365897
7869	5	1-2112	7865	24	37	377	74	2012-08-02 04:57:28.365897
7865	5	1-2271	7865	37	22	458	85	2012-08-02 04:57:28.365897
7871	5	1-2198	7870	72	47	1139	55	2012-08-02 04:57:28.365897
7872	5	1-1912	7870	47	24	640	88	2012-08-02 04:57:28.365897
7873	5	1-2125	7870	24	24	3537	2	2012-08-02 04:57:28.365897
7874	5	1-2112	7870	24	37	1978	74	2012-08-02 04:57:28.365897
7870	5	1-2271	7870	37	72	2565	85	2012-08-02 04:57:28.365897
7876	4	1-2198	7875	72	79	686	55	2012-08-02 04:57:28.365897
7877	4	1-2195	7875	79	91	444	82	2012-08-02 04:57:28.365897
7878	4	1-1196	7875	91	37	716	74	2012-08-02 04:57:28.365897
7875	4	1-2271	7875	37	72	1198	85	2012-08-02 04:57:28.365897
7880	5	1-2198	7879	72	79	398	55	2012-08-02 04:57:28.365897
7881	5	1-2195	7879	79	68	363	82	2012-08-02 04:57:28.365897
7882	5	1-1884	7879	68	62	262	22	2012-08-02 04:57:28.365897
7883	5	1-1231	7879	62	37	210	74	2012-08-02 04:57:28.365897
7879	5	1-2271	7879	37	72	494	85	2012-08-02 04:57:28.365897
16003	4	1-4155	16001	14	83	514	86	2012-08-02 05:13:29.759012
7885	3	1-1939	7884	65	73	48	94	2012-08-02 04:57:29.238827
7886	3	1-801	7884	73	84	27	81	2012-08-02 04:57:29.238827
7884	3	1-2273	7884	84	65	28	84	2012-08-02 04:57:29.238827
16004	4	1-1896	16001	83	47	524	40	2012-08-02 05:13:29.759012
7888	6	1-2205	7887	53	79	88	44	2012-08-02 04:57:29.238827
7889	6	1-2051	7887	79	88	88	60	2012-08-02 04:57:29.238827
7890	6	1-1014	7887	88	2	37	97	2012-08-02 04:57:29.238827
7891	6	1-2068	7887	2	100	73	57	2012-08-02 04:57:29.238827
7892	6	1-1388	7887	100	84	62	81	2012-08-02 04:57:29.238827
7887	6	1-2273	7887	84	53	72	84	2012-08-02 04:57:29.238827
7894	6	1-2205	7893	53	79	15	44	2012-08-02 04:57:29.238827
7895	6	1-2051	7893	79	56	16	60	2012-08-02 04:57:29.238827
7896	6	1-2222	7893	56	34	31	26	2012-08-02 04:57:29.238827
7897	6	1-195	7893	34	70	10	3	2012-08-02 04:57:29.238827
7898	6	1-2033	7893	70	84	10	81	2012-08-02 04:57:29.238827
7893	6	1-2273	7893	84	53	12	84	2012-08-02 04:57:29.238827
7900	5	1-1939	7899	65	46	1810	94	2012-08-02 04:57:29.238827
7901	5	1-2161	7899	46	67	1140	96	2012-08-02 04:57:29.238827
7902	5	1-2238	7899	67	21	3914	15	2012-08-02 04:57:29.238827
7903	5	1-500	7899	21	84	725	81	2012-08-02 04:57:29.238827
7899	5	1-2273	7899	84	65	878	84	2012-08-02 04:57:29.238827
7905	6	1-1939	7904	65	83	375	94	2012-08-02 04:57:29.238827
7906	6	1-2152	7904	83	78	582	77	2012-08-02 04:57:29.238827
7907	6	1-792	7904	78	15	188	24	2012-08-02 04:57:29.238827
7908	6	1-1917	7904	15	100	157	57	2012-08-02 04:57:29.238827
7909	6	1-1388	7904	100	84	142	81	2012-08-02 04:57:29.238827
7904	6	1-2273	7904	84	65	177	84	2012-08-02 04:57:29.238827
7911	5	1-1816	7910	78	93	858	37	2012-08-02 04:57:29.238827
7912	5	1-2102	7910	93	34	693	64	2012-08-02 04:57:29.238827
7913	5	1-2144	7910	34	68	1223	98	2012-08-02 04:57:29.238827
7914	5	1-866	7910	68	84	895	81	2012-08-02 04:57:29.238827
7910	5	1-2273	7910	84	78	1135	84	2012-08-02 04:57:29.238827
7916	6	1-2205	7915	53	79	1404	44	2012-08-02 04:57:29.238827
7917	6	1-2051	7915	79	54	1589	60	2012-08-02 04:57:29.238827
7918	6	1-305	7915	54	15	725	47	2012-08-02 04:57:29.238827
7919	6	1-1855	7915	15	100	821	57	2012-08-02 04:57:29.238827
7920	6	1-1388	7915	100	84	783	81	2012-08-02 04:57:29.238827
7915	6	1-2273	7915	84	53	1026	84	2012-08-02 04:57:29.238827
7921	1	1-2051	\N	79	79	2178	60	2012-08-02 04:57:29.238827
7923	4	1-2273	7922	84	65	213	84	2012-08-02 04:57:30.586939
7924	4	1-1939	7922	65	83	361	94	2012-08-02 04:57:30.586939
7925	4	1-2152	7922	83	36	447	77	2012-08-02 04:57:30.586939
7922	4	1-2276	7922	36	84	215	81	2012-08-02 04:57:30.586939
7927	6	1-2166	7926	20	87	462	70	2012-08-02 04:57:31.006997
7928	6	1-2270	7926	87	82	206	82	2012-08-02 04:57:31.006997
7929	6	1-2017	7926	82	83	1193	17	2012-08-02 04:57:31.006997
7930	6	1-2192	7926	83	94	1412	25	2012-08-02 04:57:31.006997
7931	6	1-2259	7926	94	39	696	22	2012-08-02 04:57:31.006997
7926	6	1-2277	7926	39	20	987	43	2012-08-02 04:57:31.006997
7933	6	1-2090	7932	45	66	5037	38	2012-08-02 04:57:31.448747
7934	6	1-1117	7932	66	83	3188	58	2012-08-02 04:57:31.448747
7935	6	1-2247	7932	83	96	3271	75	2012-08-02 04:57:31.448747
7936	6	1-2268	7932	96	3	2777	28	2012-08-02 04:57:31.448747
7937	6	1-2180	7932	3	89	2607	57	2012-08-02 04:57:31.448747
7932	6	1-2278	7932	89	45	6182	21	2012-08-02 04:57:31.448747
7939	6	1-1067	7938	19	65	544	87	2012-08-02 04:57:31.448747
7940	6	1-2011	7938	65	63	714	92	2012-08-02 04:57:31.448747
7941	6	1-2258	7938	63	4	845	66	2012-08-02 04:57:31.448747
7942	6	1-2105	7938	4	3	566	28	2012-08-02 04:57:31.448747
7943	6	1-2180	7938	3	89	575	57	2012-08-02 04:57:31.448747
7938	6	1-2278	7938	89	19	1478	21	2012-08-02 04:57:31.448747
7945	5	1-1796	7944	82	84	555	54	2012-08-02 04:57:31.890956
7946	5	1-1792	7944	84	44	791	14	2012-08-02 04:57:31.890956
7947	5	1-2245	7944	44	97	1953	40	2012-08-02 04:57:31.890956
7948	5	1-1936	7944	97	34	185	17	2012-08-02 04:57:31.890956
7944	5	1-2279	7944	34	82	889	59	2012-08-02 04:57:31.890956
7950	5	1-2203	7949	88	96	362	75	2012-08-02 04:57:31.890956
7951	5	1-2268	7949	96	76	271	28	2012-08-02 04:57:31.890956
7952	5	1-1715	7949	76	23	570	15	2012-08-02 04:57:31.890956
7953	5	1-183	7949	23	34	167	17	2012-08-02 04:57:31.890956
7949	5	1-2279	7949	34	88	848	59	2012-08-02 04:57:31.890956
7955	6	1-2203	7954	88	60	1469	75	2012-08-02 04:57:31.890956
7956	6	1-920	7954	60	98	309	91	2012-08-02 04:57:31.890956
7957	6	1-1671	7954	98	1	1324	13	2012-08-02 04:57:31.890956
7958	6	1-1689	7954	1	75	271	1	2012-08-02 04:57:31.890956
7959	6	1-431	7954	75	34	610	17	2012-08-02 04:57:31.890956
7954	6	1-2279	7954	34	88	3261	59	2012-08-02 04:57:31.890956
7961	5	1-247	7960	10	61	450	97	2012-08-02 04:57:32.344008
7962	5	1-1763	7960	61	26	393	9	2012-08-02 04:57:32.344008
7963	5	1-1838	7960	26	74	402	10	2012-08-02 04:57:32.344008
7964	5	1-1761	7960	74	54	487	5	2012-08-02 04:57:32.344008
7960	5	1-2280	7960	54	10	1110	30	2012-08-02 04:57:32.344008
7966	5	1-247	7965	10	77	980	97	2012-08-02 04:57:32.344008
7967	5	1-2218	7965	77	85	1391	22	2012-08-02 04:57:32.344008
7968	5	1-1713	7965	85	25	484	33	2012-08-02 04:57:32.344008
7969	5	1-1759	7965	25	54	1014	5	2012-08-02 04:57:32.344008
7965	5	1-2280	7965	54	10	2363	30	2012-08-02 04:57:32.344008
7971	6	1-2243	7970	83	96	1955	79	2012-08-02 04:57:35.791022
7972	6	1-1731	7970	96	5	1911	18	2012-08-02 04:57:35.791022
7973	6	1-2187	7970	5	59	1225	63	2012-08-02 04:57:35.791022
7974	6	1-2204	7970	59	91	1305	46	2012-08-02 04:57:35.791022
7975	6	1-1322	7970	91	80	878	33	2012-08-02 04:57:35.791022
7970	6	1-2288	7970	80	83	1455	99	2012-08-02 04:57:35.791022
7977	5	1-1992	7976	69	84	1410	81	2012-08-02 04:57:35.791022
7978	5	1-2273	7976	84	53	1846	84	2012-08-02 04:57:35.791022
7979	5	1-2205	7976	53	74	2521	44	2012-08-02 04:57:35.791022
7980	5	1-1394	7976	74	80	1701	33	2012-08-02 04:57:35.791022
7976	5	1-2288	7976	80	69	3047	99	2012-08-02 04:57:35.791022
7982	5	1-1966	7981	25	69	382	35	2012-08-02 04:57:36.641375
7981	5	1-2290	7981	95	25	1151	50	2012-08-02 04:57:36.641375
7983	5	1-2177	7981	69	90	335	1	2012-08-02 04:57:36.641375
7984	5	1-2145	7981	90	39	620	22	2012-08-02 04:57:36.641375
7985	5	1-2277	7981	39	95	668	43	2012-08-02 04:57:36.641375
16001	4	1-4619	16001	47	18	664	58	2012-08-02 05:13:29.759012
7987	5	1-1966	7986	25	69	158	35	2012-08-02 04:57:36.641375
7988	5	1-2177	7986	69	86	165	1	2012-08-02 04:57:36.641375
7989	5	1-1803	7986	86	39	153	22	2012-08-02 04:57:36.641375
7990	5	1-2277	7986	39	95	196	43	2012-08-02 04:57:36.641375
7986	5	1-2290	7986	95	25	401	50	2012-08-02 04:57:36.641375
7992	5	1-2011	7991	65	63	573	92	2012-08-02 04:57:37.945181
7993	5	1-2258	7991	63	11	596	66	2012-08-02 04:57:37.945181
7994	5	1-2081	7991	11	77	927	79	2012-08-02 04:57:37.945181
7995	5	1-1809	7991	77	68	832	93	2012-08-02 04:57:37.945181
7991	5	1-2293	7991	68	65	497	87	2012-08-02 04:57:37.945181
16006	5	1-4452	16005	46	39	1483	44	2012-08-02 05:13:29.759012
7997	4	1-2011	7996	65	63	3641	92	2012-08-02 04:57:37.945181
7998	4	1-2258	7996	63	69	4008	66	2012-08-02 04:57:37.945181
7999	4	1-1740	7996	69	68	4719	93	2012-08-02 04:57:37.945181
7996	4	1-2293	7996	68	65	2986	87	2012-08-02 04:57:37.945181
16007	5	1-497	16005	39	23	374	3	2012-08-02 05:13:29.759012
8001	5	1-1790	8000	97	88	2907	59	2012-08-02 04:57:38.365076
8002	5	1-2203	8000	88	60	864	75	2012-08-02 04:57:38.365076
8003	5	1-920	8000	60	32	120	91	2012-08-02 04:57:38.365076
8004	5	1-1435	8000	32	15	91	26	2012-08-02 04:57:38.365076
8000	5	1-2294	8000	15	97	2496	11	2012-08-02 04:57:38.365076
16008	5	1-4313	16005	23	47	563	56	2012-08-02 05:13:29.759012
8006	6	1-1280	8005	1	64	3932	89	2012-08-02 04:57:38.365076
8007	6	1-2191	8005	64	93	2984	37	2012-08-02 04:57:38.365076
8008	6	1-2102	8005	93	42	1623	64	2012-08-02 04:57:38.365076
8009	6	1-1615	8005	42	32	165	91	2012-08-02 04:57:38.365076
8010	6	1-1435	8005	32	15	156	26	2012-08-02 04:57:38.365076
8005	6	1-2294	8005	15	1	5275	11	2012-08-02 04:57:38.365076
8012	6	1-1713	8011	85	66	2266	33	2012-08-02 04:57:38.818182
8013	6	1-2272	8011	66	74	1371	11	2012-08-02 04:57:38.818182
8014	6	1-1164	8011	74	14	712	1	2012-08-02 04:57:38.818182
8015	6	1-1849	8011	14	22	949	69	2012-08-02 04:57:38.818182
8016	6	1-1930	8011	22	61	266	19	2012-08-02 04:57:38.818182
8011	6	1-2295	8011	61	85	4948	22	2012-08-02 04:57:38.818182
8018	3	1-2288	8017	80	40	4492	99	2012-08-02 04:57:40.110894
8019	3	1-1929	8017	40	12	4403	80	2012-08-02 04:57:40.110894
8017	3	1-2298	8017	12	80	2906	33	2012-08-02 04:57:40.110894
8021	2	1-1258	8020	100	86	4591	88	2012-08-02 04:57:40.552744
8020	2	1-2299	8020	86	100	4385	95	2012-08-02 04:57:40.552744
8023	6	1-2257	8022	87	72	704	72	2012-08-02 04:57:40.552744
8024	6	1-2162	8022	72	34	136	59	2012-08-02 04:57:40.552744
8025	6	1-2123	8022	34	34	77	57	2012-08-02 04:57:40.552744
8026	6	1-42	8022	34	30	64	47	2012-08-02 04:57:40.552744
8027	6	1-1549	8022	30	86	98	88	2012-08-02 04:57:40.552744
8022	6	1-2299	8022	86	87	133	95	2012-08-02 04:57:40.552744
8029	5	1-1824	8028	18	72	583	18	2012-08-02 04:57:40.994612
8030	5	1-2241	8028	72	85	459	16	2012-08-02 04:57:40.994612
8031	5	1-1286	8028	85	44	802	14	2012-08-02 04:57:40.994612
8032	5	1-2245	8028	44	17	2155	40	2012-08-02 04:57:40.994612
8028	5	1-2300	8028	17	18	2476	76	2012-08-02 04:57:40.994612
8034	5	1-2012	8033	24	55	3388	46	2012-08-02 04:57:41.834418
8035	5	1-1881	8033	55	35	1554	27	2012-08-02 04:57:41.834418
8036	5	1-2044	8033	35	53	400	60	2012-08-02 04:57:41.834418
8037	5	1-1869	8033	53	67	154	39	2012-08-02 04:57:41.834418
8033	5	1-2302	8033	67	24	9319	49	2012-08-02 04:57:41.834418
8039	6	1-1937	8038	46	34	3369	64	2012-08-02 04:57:42.265417
8040	6	1-2144	8038	34	15	3475	98	2012-08-02 04:57:42.265417
8041	6	1-1846	8038	15	99	2235	32	2012-08-02 04:57:42.265417
8042	6	1-2023	8038	99	68	1313	93	2012-08-02 04:57:42.265417
8043	6	1-2293	8038	68	56	547	87	2012-08-02 04:57:42.265417
8038	6	1-2303	8038	56	46	5071	34	2012-08-02 04:57:42.265417
8045	6	1-2171	8044	10	57	893	96	2012-08-02 04:57:42.265417
8046	6	1-1814	8044	57	45	234	75	2012-08-02 04:57:42.265417
8047	6	1-1247	8044	45	94	192	18	2012-08-02 04:57:42.265417
8048	6	1-2301	8044	94	52	119	82	2012-08-02 04:57:42.265417
8049	6	1-2026	8044	52	56	40	87	2012-08-02 04:57:42.265417
8044	6	1-2303	8044	56	10	410	34	2012-08-02 04:57:42.265417
8051	6	1-980	8050	82	39	502	22	2012-08-02 04:57:42.265417
8052	6	1-2277	8050	39	92	552	43	2012-08-02 04:57:42.265417
8053	6	1-2202	8050	92	97	308	12	2012-08-02 04:57:42.265417
8054	6	1-2269	8050	97	74	107	36	2012-08-02 04:57:42.265417
8055	6	1-2182	8050	74	56	86	87	2012-08-02 04:57:42.265417
8050	6	1-2303	8050	56	82	995	34	2012-08-02 04:57:42.265417
8057	6	1-1937	8056	46	29	3005	64	2012-08-02 04:57:42.265417
8058	6	1-578	8056	29	83	509	94	2012-08-02 04:57:42.265417
8059	6	1-2152	8056	83	78	619	77	2012-08-02 04:57:42.265417
8060	6	1-1566	8056	78	74	337	36	2012-08-02 04:57:42.265417
8061	6	1-2182	8056	74	56	282	87	2012-08-02 04:57:42.265417
8056	6	1-2303	8056	56	46	3439	34	2012-08-02 04:57:42.265417
8063	3	1-183	8062	23	39	1584	17	2012-08-02 04:57:42.718546
8064	3	1-2201	8062	39	57	1326	42	2012-08-02 04:57:42.718546
8062	3	1-2304	8062	57	23	5736	15	2012-08-02 04:57:42.718546
8066	3	1-2115	8065	75	5	968	18	2012-08-02 04:57:43.149277
8067	3	1-2187	8065	5	25	472	63	2012-08-02 04:57:43.149277
8065	3	1-2305	8065	25	75	1214	47	2012-08-02 04:57:43.149277
8069	6	1-1549	8068	30	86	3426	88	2012-08-02 04:57:43.149277
8070	6	1-2299	8068	86	32	4403	95	2012-08-02 04:57:43.149277
8071	6	1-940	8068	32	10	2516	41	2012-08-02 04:57:43.149277
8072	6	1-895	8068	10	59	1489	80	2012-08-02 04:57:43.149277
8073	6	1-2214	8068	59	25	730	63	2012-08-02 04:57:43.149277
8068	6	1-2305	8068	25	30	2382	47	2012-08-02 04:57:43.149277
8075	6	1-2115	8074	75	66	904	18	2012-08-02 04:57:43.149277
8076	6	1-1679	8074	66	35	778	27	2012-08-02 04:57:43.149277
8077	6	1-2044	8074	35	9	501	60	2012-08-02 04:57:43.149277
8078	6	1-591	8074	9	59	544	80	2012-08-02 04:57:43.149277
8079	6	1-2214	8074	59	25	269	63	2012-08-02 04:57:43.149277
8074	6	1-2305	8074	25	75	886	47	2012-08-02 04:57:43.149277
8081	3	1-2103	8080	46	62	7617	78	2012-08-02 04:57:43.149277
8082	3	1-1756	8080	62	25	1363	63	2012-08-02 04:57:43.149277
8080	3	1-2305	8080	25	46	4538	47	2012-08-02 04:57:43.149277
8084	6	1-2115	8083	75	94	36	18	2012-08-02 04:57:43.149277
8085	6	1-2301	8083	94	38	34	82	2012-08-02 04:57:43.149277
8086	6	1-1797	8083	38	9	15	60	2012-08-02 04:57:43.149277
8087	6	1-591	8083	9	59	18	80	2012-08-02 04:57:43.149277
8088	6	1-2214	8083	59	25	9	63	2012-08-02 04:57:43.149277
8083	6	1-2305	8083	25	75	33	47	2012-08-02 04:57:43.149277
16009	5	1-4296	16005	47	47	2285	40	2012-08-02 05:13:29.759012
8090	5	1-2257	8089	87	72	5542	72	2012-08-02 04:57:44.884357
8091	5	1-2162	8089	72	88	714	59	2012-08-02 04:57:44.884357
8092	5	1-2203	8089	88	45	255	75	2012-08-02 04:57:44.884357
8093	5	1-1247	8089	45	58	210	18	2012-08-02 04:57:44.884357
8089	5	1-2309	8089	58	87	1573	95	2012-08-02 04:57:44.884357
16005	5	1-4619	16005	47	46	2983	58	2012-08-02 05:13:29.759012
8095	5	1-1351	8094	99	45	1595	21	2012-08-02 04:57:44.884357
8096	5	1-2090	8094	45	42	1168	38	2012-08-02 04:57:44.884357
8097	5	1-1955	8094	42	88	438	35	2012-08-02 04:57:44.884357
8098	5	1-2078	8094	88	58	540	18	2012-08-02 04:57:44.884357
8094	5	1-2309	8094	58	99	4934	95	2012-08-02 04:57:44.884357
8100	4	1-1145	8099	37	54	1562	29	2012-08-02 04:57:44.884357
8101	4	1-2110	8099	54	95	676	24	2012-08-02 04:57:44.884357
8102	4	1-1472	8099	95	58	290	18	2012-08-02 04:57:44.884357
8099	4	1-2309	8099	58	37	2676	95	2012-08-02 04:57:44.884357
8104	2	1-1451	8103	23	57	6877	89	2012-08-02 04:57:45.77921
8103	2	1-2311	8103	57	23	9741	66	2012-08-02 04:57:45.77921
8106	5	1-895	8105	10	39	617	80	2012-08-02 04:57:46.198644
8107	5	1-782	8105	39	86	271	88	2012-08-02 04:57:46.198644
8108	5	1-2299	8105	86	87	175	95	2012-08-02 04:57:46.198644
8109	5	1-2257	8105	87	52	439	72	2012-08-02 04:57:46.198644
8105	5	1-2312	8105	52	10	2074	41	2012-08-02 04:57:46.198644
8111	3	1-2188	8110	53	61	168	38	2012-08-02 04:57:46.198644
8112	3	1-1722	8110	61	52	96	72	2012-08-02 04:57:46.198644
8110	3	1-2312	8110	52	53	617	41	2012-08-02 04:57:46.198644
8114	4	1-895	8113	10	12	17	80	2012-08-02 04:57:46.198644
8115	4	1-2298	8113	12	75	11	33	2012-08-02 04:57:46.198644
8116	4	1-1009	8113	75	52	4	72	2012-08-02 04:57:46.198644
8113	4	1-2312	8113	52	10	31	41	2012-08-02 04:57:46.198644
8118	7	1-2258	8117	63	23	1028	66	2012-08-02 04:57:47.922079
8119	7	1-1451	8117	23	64	764	89	2012-08-02 04:57:47.922079
8120	7	1-2191	8117	64	93	770	37	2012-08-02 04:57:47.922079
8121	7	1-2102	8117	93	77	556	64	2012-08-02 04:57:47.922079
8122	7	1-1633	8117	77	66	458	33	2012-08-02 04:57:47.922079
8123	7	1-2272	8117	66	14	287	11	2012-08-02 04:57:47.922079
8117	7	1-2316	8117	14	63	928	92	2012-08-02 04:57:47.922079
8125	5	1-1587	8124	43	68	1196	50	2012-08-02 04:57:47.922079
8126	5	1-1947	8124	68	12	734	80	2012-08-02 04:57:47.922079
8127	5	1-2298	8124	12	66	519	33	2012-08-02 04:57:47.922079
8128	5	1-2272	8124	66	14	348	11	2012-08-02 04:57:47.922079
8124	5	1-2316	8124	14	43	1199	92	2012-08-02 04:57:47.922079
8130	7	1-2258	8129	63	4	1673	66	2012-08-02 04:57:47.922079
8131	7	1-2105	8129	4	92	1131	28	2012-08-02 04:57:47.922079
8132	7	1-1130	8129	92	93	857	37	2012-08-02 04:57:47.922079
8133	7	1-2102	8129	93	77	668	64	2012-08-02 04:57:47.922079
8134	7	1-1633	8129	77	66	593	33	2012-08-02 04:57:47.922079
8135	7	1-2272	8129	66	14	402	11	2012-08-02 04:57:47.922079
8129	7	1-2316	8129	14	63	1400	92	2012-08-02 04:57:47.922079
8137	7	1-1955	8136	42	3	592	35	2012-08-02 04:57:50.551534
8138	7	1-2313	8136	3	68	1267	45	2012-08-02 04:57:50.551534
8139	7	1-2050	8136	68	12	1337	68	2012-08-02 04:57:50.551534
8140	7	1-1927	8136	12	59	1413	31	2012-08-02 04:57:50.551534
8141	7	1-2239	8136	59	12	1658	67	2012-08-02 04:57:50.551534
8142	7	1-1309	8136	12	41	1622	65	2012-08-02 04:57:50.551534
8136	7	1-2322	8136	41	42	1233	38	2012-08-02 04:57:50.551534
8144	4	1-1637	8143	88	71	2875	9	2012-08-02 04:57:50.993558
8145	4	1-1956	8143	71	61	1574	52	2012-08-02 04:57:50.993558
8146	4	1-2079	8143	61	60	1109	31	2012-08-02 04:57:50.993558
8143	4	1-2323	8143	60	88	5448	64	2012-08-02 04:57:50.993558
8148	6	1-1658	8147	53	41	984	37	2012-08-02 04:57:50.993558
8149	6	1-2122	8147	41	23	2663	16	2012-08-02 04:57:50.993558
8150	6	1-1960	8147	23	71	645	9	2012-08-02 04:57:50.993558
8151	6	1-1956	8147	71	61	383	52	2012-08-02 04:57:50.993558
8152	6	1-2079	8147	61	60	293	31	2012-08-02 04:57:50.993558
8147	6	1-2323	8147	60	53	1562	64	2012-08-02 04:57:50.993558
8154	6	1-2144	8153	34	99	641	98	2012-08-02 04:57:50.993558
8155	6	1-1292	8153	99	75	274	47	2012-08-02 04:57:50.993558
8156	6	1-2115	8153	75	94	219	18	2012-08-02 04:57:50.993558
8157	6	1-2301	8153	94	67	153	82	2012-08-02 04:57:50.993558
8158	6	1-409	8153	67	60	91	31	2012-08-02 04:57:50.993558
8153	6	1-2323	8153	60	34	499	64	2012-08-02 04:57:50.993558
8160	5	1-2317	8159	29	90	1440	87	2012-08-02 04:57:51.446642
8161	5	1-1975	8159	90	27	1200	29	2012-08-02 04:57:51.446642
8162	5	1-2086	8159	27	3	807	78	2012-08-02 04:57:51.446642
8163	5	1-2212	8159	3	28	3093	2	2012-08-02 04:57:51.446642
8159	5	1-2324	8159	28	29	1380	81	2012-08-02 04:57:51.446642
8165	5	1-702	8164	34	74	380	11	2012-08-02 04:57:52.761319
8166	5	1-1164	8164	74	14	237	1	2012-08-02 04:57:52.761319
8167	5	1-1849	8164	14	22	380	69	2012-08-02 04:57:52.761319
8168	5	1-1930	8164	22	16	128	19	2012-08-02 04:57:52.761319
8164	5	1-2327	8164	16	34	539	8	2012-08-02 04:57:52.761319
8170	6	1-2192	8169	83	7	4955	25	2012-08-02 04:57:53.214602
8171	6	1-1666	8169	7	1	2497	10	2012-08-02 04:57:53.214602
8172	6	1-2054	8169	1	33	1523	7	2012-08-02 04:57:53.214602
8173	6	1-1866	8169	33	74	466	71	2012-08-02 04:57:53.214602
8174	6	1-235	8169	74	90	172	47	2012-08-02 04:57:53.214602
8169	6	1-2328	8169	90	83	4817	17	2012-08-02 04:57:53.214602
8176	3	1-1765	8175	84	54	362	60	2012-08-02 04:57:53.214602
8177	3	1-305	8175	54	90	139	47	2012-08-02 04:57:53.214602
8175	3	1-2328	8175	90	84	4104	17	2012-08-02 04:57:53.214602
8179	4	1-2164	8178	34	66	277	45	2012-08-02 04:57:53.656472
8180	4	1-1820	8178	66	87	178	95	2012-08-02 04:57:53.656472
8181	4	1-2257	8178	87	19	709	72	2012-08-02 04:57:53.656472
8178	4	1-2329	8178	19	34	625	28	2012-08-02 04:57:53.656472
8183	5	1-2180	8182	3	33	676	57	2012-08-02 04:57:53.656472
8184	5	1-2137	8182	33	76	260	56	2012-08-02 04:57:53.656472
8185	5	1-1848	8182	76	87	219	95	2012-08-02 04:57:53.656472
8186	5	1-2257	8182	87	19	905	72	2012-08-02 04:57:53.656472
8182	5	1-2329	8182	19	3	831	28	2012-08-02 04:57:53.656472
8187	5	1-2330	8187	98	93	3148	57	2012-08-02 04:57:54.087211
8188	5	1-2281	8187	93	6	2405	54	2012-08-02 04:57:54.087211
8189	5	1-1811	8187	6	57	2317	16	2012-08-02 04:57:54.087211
8190	5	1-2049	8187	57	29	1421	74	2012-08-02 04:57:54.087211
8191	5	1-1935	8187	29	98	1752	90	2012-08-02 04:57:54.087211
8193	5	1-2137	8192	33	99	913	56	2012-08-02 04:57:54.087211
8194	5	1-2248	8192	99	75	719	49	2012-08-02 04:57:54.087211
8195	5	1-1176	8192	75	29	866	74	2012-08-02 04:57:54.087211
8196	5	1-1935	8192	29	98	1110	90	2012-08-02 04:57:54.087211
8192	5	1-2330	8192	98	33	2074	57	2012-08-02 04:57:54.087211
8198	3	1-2288	8197	80	83	268	99	2012-08-02 04:57:54.518266
8199	3	1-2243	8197	83	40	298	79	2012-08-02 04:57:54.518266
8197	3	1-2331	8197	40	80	195	33	2012-08-02 04:57:54.518266
8201	6	1-2272	8200	66	1	350	11	2012-08-02 04:57:54.518266
8202	6	1-1280	8200	1	64	368	89	2012-08-02 04:57:54.518266
8203	6	1-2191	8200	64	2	394	37	2012-08-02 04:57:54.518266
8204	6	1-2076	8200	2	83	502	99	2012-08-02 04:57:54.518266
8205	6	1-2243	8200	83	40	670	79	2012-08-02 04:57:54.518266
8200	6	1-2331	8200	40	66	525	33	2012-08-02 04:57:54.518266
8207	6	1-1860	8206	46	84	552	91	2012-08-02 04:57:55.834741
8208	6	1-2206	8206	84	66	827	46	2012-08-02 04:57:55.834741
8209	6	1-2296	8206	66	78	1609	78	2012-08-02 04:57:55.834741
8210	6	1-1974	8206	78	11	360	8	2012-08-02 04:57:55.834741
8211	6	1-1483	8206	11	4	1627	100	2012-08-02 04:57:55.834741
8206	6	1-2334	8206	4	46	3866	92	2012-08-02 04:57:55.834741
8213	6	1-2236	8212	35	79	93	55	2012-08-02 04:57:56.276795
8214	6	1-2195	8212	79	67	93	82	2012-08-02 04:57:56.276795
8215	6	1-409	8212	67	59	74	31	2012-08-02 04:57:56.276795
8216	6	1-2239	8212	59	96	85	67	2012-08-02 04:57:56.276795
8217	6	1-1716	8212	96	34	143	54	2012-08-02 04:57:56.276795
8212	6	1-2335	8212	34	35	41	52	2012-08-02 04:57:56.276795
8219	6	1-1465	8218	24	4	291	100	2012-08-02 04:57:57.138768
8220	6	1-2334	8218	4	18	518	92	2012-08-02 04:57:57.138768
8221	6	1-1778	8218	18	62	145	88	2012-08-02 04:57:57.138768
8222	6	1-2157	8218	62	3	48	78	2012-08-02 04:57:57.138768
8223	6	1-2212	8218	3	15	146	2	2012-08-02 04:57:57.138768
8218	6	1-2337	8218	15	24	393	20	2012-08-02 04:57:57.138768
8225	5	1-2267	8224	15	42	62	38	2012-08-02 04:57:57.138768
8226	5	1-1955	8224	42	12	28	35	2012-08-02 04:57:57.138768
8227	5	1-2275	8224	12	3	43	78	2012-08-02 04:57:57.138768
8228	5	1-2212	8224	3	15	166	2	2012-08-02 04:57:57.138768
8224	5	1-2337	8224	15	15	608	20	2012-08-02 04:57:57.138768
8230	6	1-2267	8229	15	48	153	38	2012-08-02 04:57:57.138768
8231	6	1-570	8229	48	1	150	10	2012-08-02 04:57:57.138768
8232	6	1-2054	8229	1	49	113	7	2012-08-02 04:57:57.138768
8233	6	1-1499	8229	49	3	91	78	2012-08-02 04:57:57.138768
8234	6	1-2212	8229	3	15	379	2	2012-08-02 04:57:57.138768
8229	6	1-2337	8229	15	15	1437	20	2012-08-02 04:57:57.138768
8236	5	1-2291	8235	73	39	1492	22	2012-08-02 04:57:58.431066
8237	5	1-2277	8235	39	95	1806	43	2012-08-02 04:57:58.431066
8238	5	1-2290	8235	95	68	3496	50	2012-08-02 04:57:58.431066
8239	5	1-1947	8235	68	37	1818	80	2012-08-02 04:57:58.431066
8235	5	1-2340	8235	37	73	1214	21	2012-08-02 04:57:58.431066
8241	6	1-2090	8240	45	93	8	38	2012-08-02 04:57:58.431066
8242	6	1-1272	8240	93	4	10	100	2012-08-02 04:57:58.431066
8243	6	1-2334	8240	4	43	22	92	2012-08-02 04:57:58.431066
8244	6	1-1587	8240	43	68	21	50	2012-08-02 04:57:58.431066
8245	6	1-1947	8240	68	37	12	80	2012-08-02 04:57:58.431066
8240	6	1-2340	8240	37	45	9	21	2012-08-02 04:57:58.431066
8247	3	1-2121	8246	58	35	2796	32	2012-08-02 04:58:01.866816
8248	3	1-1766	8246	35	84	614	89	2012-08-02 04:58:01.866816
8246	3	1-2348	8246	84	58	2716	1	2012-08-02 04:58:01.866816
8250	5	1-1803	8249	86	85	566	22	2012-08-02 04:58:01.866816
8251	5	1-1713	8249	85	66	241	33	2012-08-02 04:58:01.866816
8252	5	1-2272	8249	66	1	136	11	2012-08-02 04:58:01.866816
8253	5	1-1280	8249	1	84	121	89	2012-08-02 04:58:01.866816
8249	5	1-2348	8249	84	86	650	1	2012-08-02 04:58:01.866816
8255	5	1-1803	8254	86	62	1145	22	2012-08-02 04:58:01.866816
8256	5	1-1231	8254	62	29	959	74	2012-08-02 04:58:01.866816
8257	5	1-1935	8254	29	27	1320	90	2012-08-02 04:58:01.866816
8258	5	1-1529	8254	27	84	178	89	2012-08-02 04:58:01.866816
8254	5	1-2348	8254	84	86	1122	1	2012-08-02 04:58:01.866816
8260	3	1-2301	8259	94	19	2362	82	2012-08-02 04:58:04.872443
8261	3	1-214	8259	19	32	420	21	2012-08-02 04:58:04.872443
8259	3	1-2355	8259	32	94	3806	18	2012-08-02 04:58:04.872443
8263	4	1-1956	8262	71	52	1346	52	2012-08-02 04:58:05.768353
8264	4	1-2160	8262	52	42	1041	12	2012-08-02 04:58:05.768353
8265	4	1-1693	8262	42	48	706	72	2012-08-02 04:58:05.768353
8262	4	1-2357	8262	48	71	2411	9	2012-08-02 04:58:05.768353
8267	3	1-1599	8266	81	90	2094	50	2012-08-02 04:58:05.768353
8268	3	1-2354	8266	90	48	1080	72	2012-08-02 04:58:05.768353
8266	3	1-2357	8266	48	81	4060	9	2012-08-02 04:58:05.768353
8270	6	1-1727	8269	68	50	3715	71	2012-08-02 04:58:06.198925
8271	6	1-2034	8269	50	43	3125	92	2012-08-02 04:58:06.198925
8272	6	1-1587	8269	43	90	3339	50	2012-08-02 04:58:06.198925
8273	6	1-2354	8269	90	19	2409	72	2012-08-02 04:58:06.198925
8274	6	1-2329	8269	19	69	2959	28	2012-08-02 04:58:06.198925
8269	6	1-2358	8269	69	68	2513	61	2012-08-02 04:58:06.198925
8276	3	1-2310	8275	65	95	6682	24	2012-08-02 04:58:08.364057
8277	3	1-1472	8275	95	64	3010	18	2012-08-02 04:58:08.364057
8275	3	1-2363	8275	64	65	2744	30	2012-08-02 04:58:08.364057
8279	4	1-1277	8278	95	88	2455	59	2012-08-02 04:58:08.364057
8280	4	1-2203	8278	88	45	1348	75	2012-08-02 04:58:08.364057
8281	4	1-1247	8278	45	64	1709	18	2012-08-02 04:58:08.364057
8278	4	1-2363	8278	64	95	1860	30	2012-08-02 04:58:08.364057
8283	5	1-2281	8282	93	2	39	54	2012-08-02 04:58:10.109932
8284	5	1-2345	8282	2	84	195	17	2012-08-02 04:58:10.109932
8285	5	1-1765	8282	84	56	17	60	2012-08-02 04:58:10.109932
8286	5	1-2321	8282	56	71	26	48	2012-08-02 04:58:10.109932
8282	5	1-2367	8282	71	93	50	57	2012-08-02 04:58:10.109932
8288	4	1-2195	8287	79	2	1525	82	2012-08-02 04:58:11.435261
8289	4	1-1389	8287	2	66	437	46	2012-08-02 04:58:11.435261
8290	4	1-2296	8287	66	36	479	78	2012-08-02 04:58:11.435261
8287	4	1-2370	8287	36	79	2723	55	2012-08-02 04:58:11.435261
8292	4	1-2310	8291	65	52	2820	24	2012-08-02 04:58:12.296988
8291	4	1-2372	8291	2	65	1633	30	2012-08-02 04:58:12.296988
8293	4	1-918	8291	52	79	520	55	2012-08-02 04:58:12.296988
8294	4	1-2195	8291	79	2	314	82	2012-08-02 04:58:12.296988
16011	5	1-4469	16010	74	2	842	69	2012-08-02 05:13:30.576546
8296	4	1-171	8295	72	75	326	47	2012-08-02 04:58:12.296988
8297	4	1-2115	8295	75	94	245	18	2012-08-02 04:58:12.296988
8298	4	1-2301	8295	94	2	161	82	2012-08-02 04:58:12.296988
8295	4	1-2372	8295	2	72	959	30	2012-08-02 04:58:12.296988
16012	5	1-4569	16010	2	59	786	67	2012-08-02 05:13:30.576546
8300	6	1-2327	8299	16	84	2925	8	2012-08-02 04:58:12.739132
8301	6	1-2261	8299	84	12	1286	68	2012-08-02 04:58:12.739132
8302	6	1-1914	8299	12	70	1621	32	2012-08-02 04:58:12.739132
8303	6	1-1610	8299	70	41	1440	54	2012-08-02 04:58:12.739132
8304	6	1-2350	8299	41	55	1169	74	2012-08-02 04:58:12.739132
8299	6	1-2373	8299	55	16	841	19	2012-08-02 04:58:12.739132
16013	5	1-4324	16010	59	61	380	68	2012-08-02 05:13:30.576546
8306	3	1-2368	8305	24	62	3725	22	2012-08-02 04:58:12.739132
8307	3	1-1231	8305	62	55	2881	74	2012-08-02 04:58:12.739132
8305	3	1-2373	8305	55	24	2105	19	2012-08-02 04:58:12.739132
8309	5	1-2327	8308	16	84	460	8	2012-08-02 04:58:12.739132
8310	5	1-2261	8308	84	4	210	68	2012-08-02 04:58:12.739132
8311	5	1-2336	8308	4	89	319	81	2012-08-02 04:58:12.739132
8312	5	1-1318	8308	89	55	171	74	2012-08-02 04:58:12.739132
8308	5	1-2373	8308	55	16	128	19	2012-08-02 04:58:12.739132
8314	5	1-2327	8313	16	96	788	8	2012-08-02 04:58:12.739132
8315	5	1-1169	8313	96	22	198	24	2012-08-02 04:58:12.739132
8316	5	1-2237	8313	22	57	349	16	2012-08-02 04:58:12.739132
8317	5	1-2049	8313	57	55	246	74	2012-08-02 04:58:12.739132
8313	5	1-2373	8313	55	16	200	19	2012-08-02 04:58:12.739132
8319	5	1-2327	8318	16	78	1132	8	2012-08-02 04:58:12.739132
8320	5	1-1131	8318	78	37	742	80	2012-08-02 04:58:12.739132
8321	5	1-2340	8318	37	42	600	21	2012-08-02 04:58:12.739132
8322	5	1-996	8318	42	55	345	74	2012-08-02 04:58:12.739132
8318	5	1-2373	8318	55	16	284	19	2012-08-02 04:58:12.739132
8324	5	1-1280	8323	1	84	573	89	2012-08-02 04:58:13.181202
8325	5	1-2348	8323	84	14	2495	1	2012-08-02 04:58:13.181202
8326	5	1-1849	8323	14	35	2504	69	2012-08-02 04:58:13.181202
8327	5	1-1818	8323	35	66	1949	41	2012-08-02 04:58:13.181202
8323	5	1-2374	8323	66	1	797	11	2012-08-02 04:58:13.181202
8329	5	1-2102	8328	93	78	1648	64	2012-08-02 04:58:17.522604
8330	5	1-2346	8328	78	66	3493	45	2012-08-02 04:58:17.522604
8331	5	1-1820	8328	66	83	3094	95	2012-08-02 04:58:17.522604
8332	5	1-1781	8328	83	84	2839	15	2012-08-02 04:58:17.522604
8328	5	1-2384	8328	84	93	2018	37	2012-08-02 04:58:17.522604
8334	6	1-1632	8333	87	71	1172	48	2012-08-02 04:58:18.837208
8335	6	1-2367	8333	71	93	2128	57	2012-08-02 04:58:18.837208
8336	6	1-2281	8333	93	34	1575	54	2012-08-02 04:58:18.837208
8337	6	1-2335	8333	34	52	370	52	2012-08-02 04:58:18.837208
8338	6	1-2160	8333	52	92	351	12	2012-08-02 04:58:18.837208
8333	6	1-2387	8333	92	87	1641	77	2012-08-02 04:58:18.837208
8340	5	1-2364	8339	54	24	3100	81	2012-08-02 04:58:18.837208
8341	5	1-2289	8339	24	77	1154	97	2012-08-02 04:58:18.837208
8342	5	1-2218	8339	77	25	2055	22	2012-08-02 04:58:18.837208
8343	5	1-1787	8339	25	92	567	12	2012-08-02 04:58:18.837208
8339	5	1-2387	8339	92	54	2673	77	2012-08-02 04:58:18.837208
8345	6	1-2364	8344	54	29	3509	81	2012-08-02 04:58:18.837208
8346	6	1-2317	8344	29	5	3297	87	2012-08-02 04:58:18.837208
8347	6	1-1980	8344	5	34	2594	54	2012-08-02 04:58:18.837208
8348	6	1-2335	8344	34	52	627	52	2012-08-02 04:58:18.837208
8349	6	1-2160	8344	52	92	614	12	2012-08-02 04:58:18.837208
8344	6	1-2387	8344	92	54	2958	77	2012-08-02 04:58:18.837208
8351	6	1-1980	8350	5	2	248	54	2012-08-02 04:58:21.466624
8352	6	1-2345	8350	2	39	702	17	2012-08-02 04:58:21.466624
8353	6	1-2201	8350	39	32	390	42	2012-08-02 04:58:21.466624
8354	6	1-649	8350	32	83	33	94	2012-08-02 04:58:21.466624
8355	6	1-2152	8350	83	50	25	77	2012-08-02 04:58:21.466624
8350	6	1-2393	8350	50	5	544	87	2012-08-02 04:58:21.466624
8357	5	1-1980	8356	5	5	4032	54	2012-08-02 04:58:21.466624
8358	5	1-1981	8356	5	59	1236	80	2012-08-02 04:58:21.466624
8359	5	1-2214	8356	59	39	376	63	2012-08-02 04:58:21.466624
8360	5	1-1925	8356	39	50	221	77	2012-08-02 04:58:21.466624
8356	5	1-2393	8356	50	5	5574	87	2012-08-02 04:58:21.466624
8362	6	1-2198	8361	72	79	2463	55	2012-08-02 04:58:23.631867
8363	6	1-2195	8361	79	2	2273	82	2012-08-02 04:58:23.631867
8364	6	1-1389	8361	2	91	1075	46	2012-08-02 04:58:23.631867
8365	6	1-1322	8361	91	66	688	33	2012-08-02 04:58:23.631867
8366	6	1-2272	8361	66	11	439	11	2012-08-02 04:58:23.631867
8361	6	1-2398	8361	11	72	3018	85	2012-08-02 04:58:23.631867
8368	7	1-2194	8367	5	19	244	72	2012-08-02 04:58:23.631867
8369	7	1-2329	8367	19	69	284	28	2012-08-02 04:58:23.631867
8370	7	1-2358	8367	69	9	228	61	2012-08-02 04:58:23.631867
8371	7	1-1710	8367	9	74	206	44	2012-08-02 04:58:23.631867
8372	7	1-1394	8367	74	66	130	33	2012-08-02 04:58:23.631867
8373	7	1-2272	8367	66	11	89	11	2012-08-02 04:58:23.631867
8367	7	1-2398	8367	11	5	649	85	2012-08-02 04:58:23.631867
8375	6	1-2198	8374	72	9	1211	55	2012-08-02 04:58:23.631867
8376	6	1-465	8374	9	75	303	47	2012-08-02 04:58:23.631867
8377	6	1-2115	8374	75	64	331	18	2012-08-02 04:58:23.631867
8378	6	1-2363	8374	64	26	357	30	2012-08-02 04:58:23.631867
8379	6	1-1686	8374	26	11	180	11	2012-08-02 04:58:23.631867
8374	6	1-2398	8374	11	72	1355	85	2012-08-02 04:58:23.631867
8381	2	1-1394	8380	74	57	2747	33	2012-08-02 04:58:24.074213
8380	2	1-2399	8380	57	74	9773	44	2012-08-02 04:58:24.074213
8383	5	1-2350	8382	41	55	1167	74	2012-08-02 04:58:25.377863
8384	5	1-2373	8382	55	12	791	19	2012-08-02 04:58:25.377863
8385	5	1-1083	8382	12	56	274	60	2012-08-02 04:58:25.377863
8386	5	1-2321	8382	56	65	399	48	2012-08-02 04:58:25.377863
8382	5	1-2402	8382	65	41	1526	54	2012-08-02 04:58:25.377863
8388	4	1-1981	8387	5	37	283	80	2012-08-02 04:58:25.377863
8389	4	1-2340	8387	37	85	236	21	2012-08-02 04:58:25.377863
8390	4	1-1037	8387	85	65	147	48	2012-08-02 04:58:25.377863
8387	4	1-2402	8387	65	5	706	54	2012-08-02 04:58:25.377863
8392	6	1-2350	8391	41	84	914	74	2012-08-02 04:58:25.377863
8393	6	1-1551	8391	84	4	477	100	2012-08-02 04:58:25.377863
8394	6	1-2334	8391	4	46	1173	92	2012-08-02 04:58:25.377863
8395	6	1-1860	8391	46	34	173	91	2012-08-02 04:58:25.377863
8396	6	1-1845	8391	34	65	197	48	2012-08-02 04:58:25.377863
8391	6	1-2402	8391	65	41	949	54	2012-08-02 04:58:25.377863
16014	5	1-4538	16010	61	81	1150	14	2012-08-02 05:13:30.576546
8398	6	1-1884	8397	68	31	692	22	2012-08-02 04:58:26.285767
8399	6	1-1851	8397	31	85	387	16	2012-08-02 04:58:26.285767
8400	6	1-1286	8397	85	44	624	14	2012-08-02 04:58:26.285767
8401	6	1-2245	8397	44	1	1550	40	2012-08-02 04:58:26.285767
8402	6	1-2219	8397	1	52	1770	4	2012-08-02 04:58:26.285767
8397	6	1-2404	8397	52	68	1156	82	2012-08-02 04:58:26.285767
16010	5	1-4621	16010	81	74	272	82	2012-08-02 05:13:30.576546
8404	5	1-707	8403	49	83	420	94	2012-08-02 04:58:26.285767
8405	5	1-2152	8403	83	99	513	77	2012-08-02 04:58:26.285767
8406	5	1-1767	8403	99	1	968	40	2012-08-02 04:58:26.285767
8407	5	1-2219	8403	1	52	1149	4	2012-08-02 04:58:26.285767
8403	5	1-2404	8403	52	49	780	82	2012-08-02 04:58:26.285767
8409	6	1-707	8408	49	46	4	94	2012-08-02 04:58:26.285767
8410	6	1-2161	8408	46	4	2	96	2012-08-02 04:58:26.285767
8411	6	1-1938	8408	4	44	3	14	2012-08-02 04:58:26.285767
8412	6	1-2245	8408	44	1	7	40	2012-08-02 04:58:26.285767
8413	6	1-2219	8408	1	52	9	4	2012-08-02 04:58:26.285767
8408	6	1-2404	8408	52	49	6	82	2012-08-02 04:58:26.285767
8415	5	1-2367	8414	71	42	1986	57	2012-08-02 04:58:26.727646
8416	5	1-2057	8414	42	32	397	73	2012-08-02 04:58:26.727646
8417	5	1-2167	8414	32	61	450	97	2012-08-02 04:58:26.727646
8418	5	1-1763	8414	61	48	631	9	2012-08-02 04:58:26.727646
8414	5	1-2405	8414	48	71	867	48	2012-08-02 04:58:26.727646
8420	6	1-1309	8419	12	41	7	65	2012-08-02 04:58:28.031146
8421	6	1-2322	8419	41	93	4	38	2012-08-02 04:58:28.031146
8422	6	1-1272	8419	93	60	4	100	2012-08-02 04:58:28.031146
8423	6	1-2381	8419	60	4	3	23	2012-08-02 04:58:28.031146
8424	6	1-2124	8419	4	46	6	85	2012-08-02 04:58:28.031146
8419	6	1-2408	8419	46	12	9	67	2012-08-02 04:58:28.031146
8426	6	1-1625	8425	71	1	1985	10	2012-08-02 04:58:28.031146
8427	6	1-2054	8425	1	93	1233	7	2012-08-02 04:58:28.031146
8428	6	1-1239	8425	93	66	1030	41	2012-08-02 04:58:28.031146
8429	6	1-2374	8425	66	11	545	11	2012-08-02 04:58:28.031146
8430	6	1-2398	8425	11	46	3460	85	2012-08-02 04:58:28.031146
8425	6	1-2408	8425	46	71	5697	67	2012-08-02 04:58:28.031146
8432	6	1-1910	8431	45	18	1617	76	2012-08-02 04:58:28.031146
8433	6	1-1824	8431	18	64	417	18	2012-08-02 04:58:28.031146
8434	6	1-2363	8431	64	26	399	30	2012-08-02 04:58:28.031146
8435	6	1-1686	8431	26	11	178	11	2012-08-02 04:58:28.031146
8436	6	1-2398	8431	11	46	1192	85	2012-08-02 04:58:28.031146
8431	6	1-2408	8431	46	45	2064	67	2012-08-02 04:58:28.031146
8438	6	1-1309	8437	12	41	217	65	2012-08-02 04:58:28.031146
8439	6	1-2322	8437	41	78	163	38	2012-08-02 04:58:28.031146
8440	6	1-1900	8437	78	90	168	1	2012-08-02 04:58:28.031146
8441	6	1-1858	8437	90	96	119	9	2012-08-02 04:58:28.031146
8442	6	1-2111	8437	96	46	116	85	2012-08-02 04:58:28.031146
8437	6	1-2408	8437	46	12	228	67	2012-08-02 04:58:28.031146
8444	5	1-2341	8443	11	84	692	15	2012-08-02 04:58:29.357537
8445	5	1-2384	8443	84	41	471	37	2012-08-02 04:58:29.357537
8446	5	1-2122	8443	41	23	1746	16	2012-08-02 04:58:29.357537
8447	5	1-1960	8443	23	31	579	9	2012-08-02 04:58:29.357537
8443	5	1-2411	8443	31	11	1044	7	2012-08-02 04:58:29.357537
8449	3	1-2160	8448	52	20	3944	12	2012-08-02 04:58:31.124695
8450	3	1-2074	8448	20	24	5606	6	2012-08-02 04:58:31.124695
8448	3	1-2415	8448	24	52	3842	52	2012-08-02 04:58:31.124695
8452	5	1-2236	8451	35	43	1054	55	2012-08-02 04:58:31.124695
8453	5	1-55	8451	43	92	536	43	2012-08-02 04:58:31.124695
8454	5	1-2202	8451	92	20	396	12	2012-08-02 04:58:31.124695
8455	5	1-2074	8451	20	24	622	6	2012-08-02 04:58:31.124695
8451	5	1-2415	8451	24	35	471	52	2012-08-02 04:58:31.124695
8457	7	1-1797	8456	38	100	968	60	2012-08-02 04:58:31.566653
8458	7	1-1844	8456	100	49	689	58	2012-08-02 04:58:31.566653
8459	7	1-732	8456	49	19	1116	72	2012-08-02 04:58:31.566653
8460	7	1-2329	8456	19	70	1308	28	2012-08-02 04:58:31.566653
8461	7	1-1681	8456	70	13	1420	70	2012-08-02 04:58:31.566653
8462	7	1-1478	8456	13	93	197	66	2012-08-02 04:58:31.566653
8456	7	1-2416	8456	93	38	2186	82	2012-08-02 04:58:31.566653
8464	7	1-1389	8463	2	66	144	46	2012-08-02 04:58:31.566653
8465	7	1-2296	8463	66	78	290	78	2012-08-02 04:58:31.566653
8466	7	1-1974	8463	78	61	67	8	2012-08-02 04:58:31.566653
8467	7	1-2407	8463	61	51	134	65	2012-08-02 04:58:31.566653
8468	7	1-2253	8463	51	60	40	35	2012-08-02 04:58:31.566653
8469	7	1-1021	8463	60	93	24	66	2012-08-02 04:58:31.566653
8463	7	1-2416	8463	93	2	274	82	2012-08-02 04:58:31.566653
8471	6	1-1084	8470	61	42	1008	12	2012-08-02 04:58:32.439432
8472	6	1-1693	8470	42	89	881	72	2012-08-02 04:58:32.439432
8473	6	1-1236	8470	89	44	672	14	2012-08-02 04:58:32.439432
8474	6	1-2245	8470	44	1	1917	40	2012-08-02 04:58:32.439432
8475	6	1-2219	8470	1	41	2513	4	2012-08-02 04:58:32.439432
8470	6	1-2418	8470	41	61	2188	95	2012-08-02 04:58:32.439432
8477	6	1-1781	8476	83	84	552	15	2012-08-02 04:58:32.439432
8478	6	1-2384	8476	84	2	366	37	2012-08-02 04:58:32.439432
8479	6	1-2252	8476	2	34	626	62	2012-08-02 04:58:32.439432
8480	6	1-2314	8476	34	1	466	40	2012-08-02 04:58:32.439432
8481	6	1-2219	8476	1	41	673	4	2012-08-02 04:58:32.439432
8476	6	1-2418	8476	41	83	646	95	2012-08-02 04:58:32.439432
8483	5	1-1084	8482	61	42	1536	12	2012-08-02 04:58:32.439432
8484	5	1-1693	8482	42	19	1562	72	2012-08-02 04:58:32.439432
8485	5	1-2329	8482	19	85	1866	28	2012-08-02 04:58:32.439432
8486	5	1-1719	8482	85	41	2826	4	2012-08-02 04:58:32.439432
8482	5	1-2418	8482	41	61	2865	95	2012-08-02 04:58:32.439432
8488	6	1-2261	8487	84	44	543	68	2012-08-02 04:58:33.743285
8489	6	1-2009	8487	44	4	1433	100	2012-08-02 04:58:33.743285
8490	6	1-2334	8487	4	46	2853	92	2012-08-02 04:58:33.743285
8491	6	1-1860	8487	46	36	341	91	2012-08-02 04:58:33.743285
8492	6	1-743	8487	36	73	2661	14	2012-08-02 04:58:33.743285
8487	6	1-2421	8487	73	84	1286	8	2012-08-02 04:58:33.743285
8494	5	1-275	8493	76	48	1182	9	2012-08-02 04:58:33.743285
8495	5	1-2405	8493	48	71	1345	48	2012-08-02 04:58:33.743285
8496	5	1-2367	8493	71	91	2553	57	2012-08-02 04:58:33.743285
8497	5	1-2371	8493	91	73	2975	14	2012-08-02 04:58:33.743285
8493	5	1-2421	8493	73	76	1493	8	2012-08-02 04:58:33.743285
8499	4	1-2204	8498	59	96	4531	46	2012-08-02 04:58:34.174235
8498	4	1-2422	8498	74	59	5239	63	2012-08-02 04:58:34.174235
8500	4	1-537	8498	96	65	1649	48	2012-08-02 04:58:34.174235
8501	4	1-2402	8498	65	74	6079	54	2012-08-02 04:58:34.174235
8503	5	1-1925	8502	39	54	623	77	2012-08-02 04:58:34.174235
8504	5	1-2364	8502	54	29	792	81	2012-08-02 04:58:34.174235
8505	5	1-2317	8502	29	5	797	87	2012-08-02 04:58:34.174235
8506	5	1-1980	8502	5	74	672	54	2012-08-02 04:58:34.174235
8502	5	1-2422	8502	74	39	673	63	2012-08-02 04:58:34.174235
16016	5	1-3929	16015	42	1	8	23	2012-08-02 05:13:31.405362
8508	5	1-2204	8507	59	96	1003	46	2012-08-02 04:58:34.174235
8509	5	1-537	8507	96	71	456	48	2012-08-02 04:58:34.174235
8510	5	1-2367	8507	71	93	983	57	2012-08-02 04:58:34.174235
8511	5	1-2281	8507	93	74	863	54	2012-08-02 04:58:34.174235
8507	5	1-2422	8507	74	59	929	63	2012-08-02 04:58:34.174235
8513	7	1-1292	8512	99	75	62	47	2012-08-02 04:58:35.135645
8514	7	1-2115	8512	75	64	56	18	2012-08-02 04:58:35.135645
8515	7	1-2363	8512	64	72	50	30	2012-08-02 04:58:35.135645
8516	7	1-491	8512	72	37	25	80	2012-08-02 04:58:35.135645
8517	7	1-2340	8512	37	45	17	21	2012-08-02 04:58:35.135645
8518	7	1-2090	8512	45	25	13	38	2012-08-02 04:58:35.135645
8512	7	1-2424	8512	25	99	128	98	2012-08-02 04:58:35.135645
8520	6	1-772	8519	73	62	56	88	2012-08-02 04:58:35.135645
8521	6	1-2157	8519	62	78	22	78	2012-08-02 04:58:35.135645
8522	6	1-1974	8519	78	61	4	8	2012-08-02 04:58:35.135645
8523	6	1-2407	8519	61	41	8	65	2012-08-02 04:58:35.135645
8524	6	1-2322	8519	41	25	5	38	2012-08-02 04:58:35.135645
8519	6	1-2424	8519	25	73	51	98	2012-08-02 04:58:35.135645
8526	6	1-1292	8525	99	75	305	47	2012-08-02 04:58:35.135645
8527	6	1-2115	8525	75	35	337	18	2012-08-02 04:58:35.135645
8528	6	1-2320	8525	35	55	189	58	2012-08-02 04:58:35.135645
8529	6	1-2101	8525	55	41	57	65	2012-08-02 04:58:35.135645
8530	6	1-2322	8525	41	25	43	38	2012-08-02 04:58:35.135645
8525	6	1-2424	8525	25	99	517	98	2012-08-02 04:58:35.135645
8531	1	1-491	\N	72	72	1741	80	2012-08-02 04:58:35.135645
8532	1	1-2115	\N	75	75	4509	18	2012-08-02 04:58:35.135645
8534	6	1-2014	8533	68	71	376	13	2012-08-02 04:58:35.588934
8535	6	1-1267	8533	71	84	324	15	2012-08-02 04:58:35.588934
8536	6	1-2384	8533	84	41	234	37	2012-08-02 04:58:35.588934
8537	6	1-2122	8533	41	83	923	16	2012-08-02 04:58:35.588934
8538	6	1-2392	8533	83	44	434	82	2012-08-02 04:58:35.588934
8533	6	1-2425	8533	44	68	328	89	2012-08-02 04:58:35.588934
8540	3	1-2424	8539	25	27	3881	98	2012-08-02 04:58:37.323005
8541	3	1-344	8539	27	90	1174	76	2012-08-02 04:58:37.323005
8539	3	1-2429	8539	90	25	426	38	2012-08-02 04:58:37.323005
8543	5	1-2415	8542	24	35	1764	52	2012-08-02 04:58:37.753995
8544	5	1-2236	8542	35	1	3314	55	2012-08-02 04:58:37.753995
8545	5	1-1898	8542	1	21	627	65	2012-08-02 04:58:37.753995
8546	5	1-2213	8542	21	17	529	4	2012-08-02 04:58:37.753995
8542	5	1-2430	8542	17	24	2779	6	2012-08-02 04:58:37.753995
8548	6	1-766	8547	66	29	1451	81	2012-08-02 04:58:37.753995
8549	6	1-2317	8547	29	94	1450	87	2012-08-02 04:58:37.753995
8550	6	1-1873	8547	94	19	616	72	2012-08-02 04:58:37.753995
8551	6	1-2329	8547	19	85	661	28	2012-08-02 04:58:37.753995
8552	6	1-1719	8547	85	17	899	4	2012-08-02 04:58:37.753995
8547	6	1-2430	8547	17	66	5172	6	2012-08-02 04:58:37.753995
8554	6	1-700	8553	75	78	1054	78	2012-08-02 04:58:38.196056
8555	6	1-1974	8553	78	61	247	8	2012-08-02 04:58:38.196056
8556	6	1-2407	8553	61	41	499	65	2012-08-02 04:58:38.196056
8557	6	1-2322	8553	41	25	383	38	2012-08-02 04:58:38.196056
8558	6	1-2424	8553	25	11	4671	98	2012-08-02 04:58:38.196056
8553	6	1-2431	8553	11	75	1006	61	2012-08-02 04:58:38.196056
8560	6	1-2228	8559	94	66	1939	45	2012-08-02 04:58:39.510517
8561	6	1-1820	8559	66	83	1529	95	2012-08-02 04:58:39.510517
8562	6	1-1781	8559	83	84	1249	15	2012-08-02 04:58:39.510517
8563	6	1-2384	8559	84	2	791	37	2012-08-02 04:58:39.510517
8564	6	1-2076	8559	2	99	956	99	2012-08-02 04:58:39.510517
8559	6	1-2434	8559	99	94	2835	21	2012-08-02 04:58:39.510517
8566	5	1-2228	8565	94	2	2643	45	2012-08-02 04:58:39.510517
8567	5	1-1902	8565	2	88	2637	94	2012-08-02 04:58:39.510517
8568	5	1-2342	8565	88	9	1947	18	2012-08-02 04:58:39.510517
8569	5	1-2394	8565	9	99	1224	99	2012-08-02 04:58:39.510517
8565	5	1-2434	8565	99	94	3746	21	2012-08-02 04:58:39.510517
8571	5	1-2228	8570	94	56	565	45	2012-08-02 04:58:39.510517
8572	5	1-1552	8570	56	88	489	94	2012-08-02 04:58:39.510517
8573	5	1-2342	8570	88	9	374	18	2012-08-02 04:58:39.510517
8574	5	1-2394	8570	9	99	244	99	2012-08-02 04:58:39.510517
8570	5	1-2434	8570	99	94	773	21	2012-08-02 04:58:39.510517
8575	1	1-1401	\N	67	67	3459	94	2012-08-02 04:58:39.510517
8577	5	1-1696	8576	55	41	312	37	2012-08-02 04:58:41.709167
8578	5	1-2122	8576	41	83	1194	16	2012-08-02 04:58:41.709167
8579	5	1-2392	8576	83	49	545	82	2012-08-02 04:58:41.709167
8580	5	1-707	8576	49	4	379	94	2012-08-02 04:58:41.709167
8576	5	1-2439	8576	4	55	1509	51	2012-08-02 04:58:41.709167
8582	5	1-2068	8581	2	93	559	57	2012-08-02 04:58:42.150999
8583	5	1-2281	8581	93	74	455	54	2012-08-02 04:58:42.150999
8584	5	1-2422	8581	74	39	454	63	2012-08-02 04:58:42.150999
8585	5	1-1925	8581	39	29	419	77	2012-08-02 04:58:42.150999
8581	5	1-2440	8581	29	2	292	97	2012-08-02 04:58:42.150999
8587	6	1-2435	8586	45	21	1561	65	2012-08-02 04:58:45.189421
8588	6	1-2213	8586	21	41	1557	4	2012-08-02 04:58:45.189421
8589	6	1-2418	8586	41	37	1532	95	2012-08-02 04:58:45.189421
8590	6	1-1145	8586	37	27	1078	29	2012-08-02 04:58:45.189421
8591	6	1-2086	8586	27	13	750	78	2012-08-02 04:58:45.189421
8586	6	1-2447	8586	13	45	669	39	2012-08-02 04:58:45.189421
8593	4	1-2435	8592	45	51	4291	65	2012-08-02 04:58:45.189421
8594	4	1-2253	8592	51	12	1242	35	2012-08-02 04:58:45.189421
8595	4	1-2275	8592	12	13	1971	78	2012-08-02 04:58:45.189421
8592	4	1-2447	8592	13	45	1798	39	2012-08-02 04:58:45.189421
8597	5	1-2272	8596	66	14	3	11	2012-08-02 04:58:48.713405
8598	5	1-2316	8596	14	46	9	92	2012-08-02 04:58:48.713405
8599	5	1-1860	8596	46	16	1	91	2012-08-02 04:58:48.713405
8600	5	1-2397	8596	16	66	2	15	2012-08-02 04:58:48.713405
8596	5	1-2455	8596	66	66	5	33	2012-08-02 04:58:48.713405
8602	6	1-2272	8601	66	97	1402	11	2012-08-02 04:58:48.713405
8603	6	1-1790	8601	97	34	2637	59	2012-08-02 04:58:48.713405
8604	6	1-2123	8601	34	33	1301	57	2012-08-02 04:58:48.713405
8605	6	1-2137	8601	33	77	574	56	2012-08-02 04:58:48.713405
8606	6	1-2199	8601	77	66	630	15	2012-08-02 04:58:48.713405
8601	6	1-2455	8601	66	66	2088	33	2012-08-02 04:58:48.713405
16017	5	1-3946	16015	1	73	6	10	2012-08-02 05:13:31.405362
8608	5	1-1739	8607	7	34	913	59	2012-08-02 04:58:48.713405
8609	5	1-2123	8607	34	33	493	57	2012-08-02 04:58:48.713405
8610	5	1-2137	8607	33	77	238	56	2012-08-02 04:58:48.713405
8611	5	1-2199	8607	77	66	286	15	2012-08-02 04:58:48.713405
8607	5	1-2455	8607	66	7	953	33	2012-08-02 04:58:48.713405
8613	5	1-2272	8612	66	1	877	11	2012-08-02 04:58:48.713405
8614	5	1-1280	8612	1	6	867	89	2012-08-02 04:58:48.713405
8615	5	1-2361	8612	6	16	218	91	2012-08-02 04:58:48.713405
8616	5	1-2397	8612	16	66	387	15	2012-08-02 04:58:48.713405
8612	5	1-2455	8612	66	66	1294	33	2012-08-02 04:58:48.713405
8618	5	1-2272	8617	66	94	259	11	2012-08-02 04:58:48.713405
8619	5	1-1571	8617	94	66	140	45	2012-08-02 04:58:48.713405
8620	5	1-1820	8617	66	83	124	95	2012-08-02 04:58:48.713405
8621	5	1-1781	8617	83	66	105	15	2012-08-02 04:58:48.713405
8617	5	1-2455	8617	66	66	367	33	2012-08-02 04:58:48.713405
8622	1	1-2272	\N	66	66	65	11	2012-08-02 04:58:48.713405
8624	6	1-2290	8623	95	16	4695	50	2012-08-02 04:58:52.33671
8625	6	1-1934	8623	16	61	1372	77	2012-08-02 04:58:52.33671
8626	6	1-504	8623	61	17	415	75	2012-08-02 04:58:52.33671
8627	6	1-1854	8623	17	10	238	8	2012-08-02 04:58:52.33671
8628	6	1-2409	8623	10	68	551	3	2012-08-02 04:58:52.33671
8623	6	1-2463	8623	68	95	1938	43	2012-08-02 04:58:52.33671
8630	5	1-2081	8629	11	78	4234	79	2012-08-02 04:58:53.233223
8631	5	1-2000	8629	78	54	3393	29	2012-08-02 04:58:53.233223
8632	5	1-2110	8629	54	91	1404	24	2012-08-02 04:58:53.233223
8633	5	1-1911	8629	91	6	1141	83	2012-08-02 04:58:53.233223
8629	5	1-2465	8629	6	11	2984	66	2012-08-02 04:58:53.233223
8635	5	1-2220	8634	25	23	775	75	2012-08-02 04:58:53.233223
8636	5	1-2436	8634	23	88	1637	61	2012-08-02 04:58:53.233223
8637	5	1-868	8634	88	91	636	24	2012-08-02 04:58:53.233223
8638	5	1-1911	8634	91	6	580	83	2012-08-02 04:58:53.233223
8634	5	1-2465	8634	6	25	1700	66	2012-08-02 04:58:53.233223
8640	6	1-1130	8639	92	2	1429	37	2012-08-02 04:58:53.689857
8641	6	1-2252	8639	2	34	2134	62	2012-08-02 04:58:53.689857
8642	6	1-2314	8639	34	39	1386	40	2012-08-02 04:58:53.689857
8643	6	1-2438	8639	39	15	620	27	2012-08-02 04:58:53.689857
8644	6	1-2442	8639	15	10	289	71	2012-08-02 04:58:53.689857
8639	6	1-2466	8639	10	92	2213	28	2012-08-02 04:58:53.689857
8646	4	1-2180	8645	3	42	1498	57	2012-08-02 04:58:53.689857
8647	4	1-2057	8645	42	43	262	73	2012-08-02 04:58:53.689857
8648	4	1-1603	8645	43	10	187	71	2012-08-02 04:58:53.689857
8645	4	1-2466	8645	10	3	1565	28	2012-08-02 04:58:53.689857
8650	6	1-2180	8649	3	33	2544	57	2012-08-02 04:58:53.689857
8651	6	1-2137	8649	33	85	1150	56	2012-08-02 04:58:53.689857
8652	6	1-1348	8649	85	55	564	46	2012-08-02 04:58:53.689857
8653	6	1-1881	8649	55	15	625	27	2012-08-02 04:58:53.689857
8654	6	1-2442	8649	15	10	318	71	2012-08-02 04:58:53.689857
8649	6	1-2466	8649	10	3	2662	28	2012-08-02 04:58:53.689857
8656	6	1-2346	8655	78	68	1352	45	2012-08-02 04:58:54.183765
8657	6	1-2050	8655	68	73	1368	68	2012-08-02 04:58:54.183765
8658	6	1-1730	8655	73	16	1646	50	2012-08-02 04:58:54.183765
8659	6	1-1934	8655	16	99	463	77	2012-08-02 04:58:54.183765
8660	6	1-1767	8655	99	5	1099	40	2012-08-02 04:58:54.183765
8655	6	1-2467	8655	5	78	662	64	2012-08-02 04:58:54.183765
8662	3	1-2340	8661	37	73	1876	21	2012-08-02 04:58:55.061199
8663	3	1-2291	8661	73	61	2684	22	2012-08-02 04:58:55.061199
8661	3	1-2469	8661	61	37	2414	80	2012-08-02 04:58:55.061199
8665	6	1-2340	8664	37	94	2276	21	2012-08-02 04:58:55.061199
8666	6	1-2228	8664	94	56	1648	45	2012-08-02 04:58:55.061199
8667	6	1-1552	8664	56	28	1412	94	2012-08-02 04:58:55.061199
8668	6	1-2448	8664	28	24	1634	19	2012-08-02 04:58:55.061199
8669	6	1-2368	8664	24	61	3174	22	2012-08-02 04:58:55.061199
8664	6	1-2469	8664	61	37	2891	80	2012-08-02 04:58:55.061199
8671	5	1-2214	8670	59	39	1551	63	2012-08-02 04:58:55.061199
8672	5	1-1925	8670	39	29	1636	77	2012-08-02 04:58:55.061199
8673	5	1-2440	8670	29	77	1303	97	2012-08-02 04:58:55.061199
8674	5	1-2218	8670	77	61	2899	22	2012-08-02 04:58:55.061199
8670	5	1-2469	8670	61	59	2844	80	2012-08-02 04:58:55.061199
8676	4	1-2296	8675	66	78	3437	78	2012-08-02 04:58:55.503711
8677	4	1-1974	8675	78	78	465	8	2012-08-02 04:58:55.503711
8678	4	1-1131	8675	78	88	184	80	2012-08-02 04:58:55.503711
8675	4	1-2470	8675	88	66	5156	46	2012-08-02 04:58:55.503711
8680	4	1-1589	8679	40	35	217	52	2012-08-02 04:58:55.503711
8681	4	1-2236	8679	35	48	265	55	2012-08-02 04:58:55.503711
8682	4	1-1469	8679	48	88	48	80	2012-08-02 04:58:55.503711
8679	4	1-2470	8679	88	40	2144	46	2012-08-02 04:58:55.503711
8684	5	1-2060	8683	88	78	672	64	2012-08-02 04:58:55.968492
8685	5	1-2346	8683	78	68	1386	45	2012-08-02 04:58:55.968492
8686	5	1-2050	8683	68	12	1416	68	2012-08-02 04:58:55.968492
8687	5	1-1914	8683	12	69	2054	32	2012-08-02 04:58:55.968492
8683	5	1-2471	8683	69	88	1603	26	2012-08-02 04:58:55.968492
8689	6	1-1953	8688	81	83	1826	16	2012-08-02 04:58:56.863457
8690	6	1-2392	8688	83	2	725	82	2012-08-02 04:58:56.863457
8691	6	1-1389	8688	2	91	328	46	2012-08-02 04:58:56.863457
8692	6	1-2284	8688	91	17	333	4	2012-08-02 04:58:56.863457
8693	6	1-2430	8688	17	37	1861	6	2012-08-02 04:58:56.863457
8688	6	1-2473	8688	37	81	2051	62	2012-08-02 04:58:56.863457
8695	4	1-2314	8694	34	5	1780	40	2012-08-02 04:58:56.863457
8696	4	1-2467	8694	5	48	1020	64	2012-08-02 04:58:56.863457
8697	4	1-824	8694	48	37	2072	6	2012-08-02 04:58:56.863457
8694	4	1-2473	8694	37	34	2434	62	2012-08-02 04:58:56.863457
8699	6	1-2403	8698	77	34	2759	62	2012-08-02 04:58:57.305604
8700	6	1-2314	8698	34	5	2161	40	2012-08-02 04:58:57.305604
8701	6	1-2467	8698	5	78	1327	64	2012-08-02 04:58:57.305604
8702	6	1-2346	8698	78	6	2760	45	2012-08-02 04:58:57.305604
8703	6	1-2170	8698	6	1	2327	9	2012-08-02 04:58:57.305604
8698	6	1-2474	8698	1	77	2013	71	2012-08-02 04:58:57.305604
8705	5	1-1814	8704	57	23	648	75	2012-08-02 04:58:58.167363
8706	5	1-2436	8704	23	75	1109	61	2012-08-02 04:58:58.167363
8707	5	1-700	8704	75	13	789	78	2012-08-02 04:58:58.167363
8708	5	1-2447	8704	13	9	506	39	2012-08-02 04:58:58.167363
8704	5	1-2476	8704	9	57	2310	96	2012-08-02 04:58:58.167363
16018	5	1-4551	16015	73	13	11	73	2012-08-02 05:13:31.405362
8710	6	1-1814	8709	57	45	634	75	2012-08-02 04:58:58.167363
8711	6	1-1247	8709	45	94	748	18	2012-08-02 04:58:58.167363
8712	6	1-2301	8709	94	38	669	82	2012-08-02 04:58:58.167363
8713	6	1-1797	8709	38	53	283	60	2012-08-02 04:58:58.167363
8714	6	1-1869	8709	53	9	275	39	2012-08-02 04:58:58.167363
8709	6	1-2476	8709	9	57	1682	96	2012-08-02 04:58:58.167363
8715	1	1-1869	\N	53	53	3587	39	2012-08-02 04:58:58.167363
16019	5	1-4408	16015	13	67	7	1	2012-08-02 05:13:31.405362
8717	2	1-1239	8716	93	65	869	41	2012-08-02 04:58:58.598363
8716	2	1-2477	8716	65	93	1469	7	2012-08-02 04:58:58.598363
8719	6	1-1499	8718	49	3	84	78	2012-08-02 04:58:58.598363
8720	6	1-2212	8718	3	24	251	2	2012-08-02 04:58:58.598363
8721	6	1-2112	8718	24	13	142	74	2012-08-02 04:58:58.598363
8722	6	1-2464	8718	13	35	81	69	2012-08-02 04:58:58.598363
8723	6	1-1818	8718	35	65	71	41	2012-08-02 04:58:58.598363
8718	6	1-2477	8718	65	49	146	7	2012-08-02 04:58:58.598363
8725	6	1-1380	8724	16	92	1135	85	2012-08-02 04:58:58.598363
8726	6	1-2457	8724	92	13	313	78	2012-08-02 04:58:58.598363
8727	6	1-2447	8724	13	27	264	39	2012-08-02 04:58:58.598363
8728	6	1-1903	8724	27	35	355	69	2012-08-02 04:58:58.598363
8729	6	1-1818	8724	35	65	388	41	2012-08-02 04:58:58.598363
8724	6	1-2477	8724	65	16	1008	7	2012-08-02 04:58:58.598363
8731	6	1-1752	8730	65	91	1084	24	2012-08-02 04:58:58.598363
8732	6	1-1911	8730	91	22	1136	83	2012-08-02 04:58:58.598363
8733	6	1-1191	8730	22	27	376	39	2012-08-02 04:58:58.598363
8734	6	1-1903	8730	27	35	547	69	2012-08-02 04:58:58.598363
8735	6	1-1818	8730	35	65	648	41	2012-08-02 04:58:58.598363
8730	6	1-2477	8730	65	65	1871	7	2012-08-02 04:58:58.598363
8737	5	1-2346	8736	78	68	566	45	2012-08-02 04:58:59.913747
8738	5	1-2050	8736	68	39	572	68	2012-08-02 04:58:59.913747
8739	5	1-1346	8736	39	29	408	81	2012-08-02 04:58:59.913747
8740	5	1-2317	8736	29	50	444	87	2012-08-02 04:58:59.913747
8736	5	1-2480	8736	50	78	278	64	2012-08-02 04:58:59.913747
8742	2	1-772	8741	73	91	2628	88	2012-08-02 04:59:01.250663
8741	2	1-2483	8741	91	73	3816	98	2012-08-02 04:59:01.250663
8744	6	1-1846	8743	15	69	896	32	2012-08-02 04:59:01.250663
8745	6	1-2471	8743	69	18	543	26	2012-08-02 04:59:01.250663
8746	6	1-1513	8743	18	94	740	25	2012-08-02 04:59:01.250663
8747	6	1-2259	8743	94	18	293	22	2012-08-02 04:59:01.250663
8748	6	1-2460	8743	18	91	528	88	2012-08-02 04:59:01.250663
8743	6	1-2483	8743	91	15	1069	98	2012-08-02 04:59:01.250663
8750	2	1-1198	8749	69	91	1895	88	2012-08-02 04:59:01.250663
8749	2	1-2483	8749	91	69	4347	98	2012-08-02 04:59:01.250663
8752	5	1-1846	8751	15	69	360	32	2012-08-02 04:59:01.250663
8753	5	1-2471	8751	69	2	270	26	2012-08-02 04:59:01.250663
8754	5	1-2178	8751	2	97	106	31	2012-08-02 04:59:01.250663
8755	5	1-935	8751	97	91	138	88	2012-08-02 04:59:01.250663
8751	5	1-2483	8751	91	15	346	98	2012-08-02 04:59:01.250663
8756	1	1-935	\N	97	97	2240	88	2012-08-02 04:59:01.250663
8758	2	1-146	8757	23	42	1870	27	2012-08-02 04:59:02.55408
8757	2	1-2486	8757	42	23	1797	61	2012-08-02 04:59:02.55408
8760	6	1-2389	8759	40	37	1186	6	2012-08-02 04:59:02.98501
8761	6	1-2473	8759	37	81	1541	62	2012-08-02 04:59:02.98501
8762	6	1-1953	8759	81	53	1617	16	2012-08-02 04:59:02.98501
8763	6	1-2129	8759	53	96	80	75	2012-08-02 04:59:02.98501
8764	6	1-2268	8759	96	32	79	28	2012-08-02 04:59:02.98501
8759	6	1-2487	8759	32	40	525	85	2012-08-02 04:59:02.98501
8766	4	1-1144	8765	56	18	1800	76	2012-08-02 04:59:03.416276
8767	4	1-1824	8765	18	94	270	18	2012-08-02 04:59:03.416276
8768	4	1-2301	8765	94	61	133	82	2012-08-02 04:59:03.416276
8765	4	1-2488	8765	61	56	2827	73	2012-08-02 04:59:03.416276
8770	4	1-859	8769	47	88	1076	94	2012-08-02 04:59:03.416276
8771	4	1-2342	8769	88	94	441	18	2012-08-02 04:59:03.416276
8772	4	1-2301	8769	94	61	218	82	2012-08-02 04:59:03.416276
8769	4	1-2488	8769	61	47	4647	73	2012-08-02 04:59:03.416276
8774	4	1-1864	8773	56	72	46	11	2012-08-02 04:59:03.416276
8775	4	1-1561	8773	72	83	35	16	2012-08-02 04:59:03.416276
8776	4	1-2392	8773	83	61	10	82	2012-08-02 04:59:03.416276
8773	4	1-2488	8773	61	56	240	73	2012-08-02 04:59:03.416276
8778	3	1-1925	8777	39	99	1833	77	2012-08-02 04:59:07.355307
8779	3	1-1767	8777	99	23	1491	40	2012-08-02 04:59:07.355307
8777	3	1-2497	8777	23	39	5339	63	2012-08-02 04:59:07.355307
8781	3	1-2463	8780	68	87	4440	43	2012-08-02 04:59:08.228116
8782	3	1-1895	8780	87	68	531	23	2012-08-02 04:59:08.228116
8780	3	1-2499	8780	68	68	2133	3	2012-08-02 04:59:08.228116
8784	5	1-2338	8783	48	93	1985	24	2012-08-02 04:59:08.228116
8785	5	1-1111	8783	93	44	677	68	2012-08-02 04:59:08.228116
8786	5	1-2009	8783	44	60	1363	100	2012-08-02 04:59:08.228116
8787	5	1-2381	8783	60	68	873	23	2012-08-02 04:59:08.228116
8783	5	1-2499	8783	68	48	3668	3	2012-08-02 04:59:08.228116
8789	5	1-2284	8788	91	41	1990	4	2012-08-02 04:59:12.116739
8790	5	1-2418	8788	41	61	1521	95	2012-08-02 04:59:12.116739
8791	5	1-1084	8788	61	69	615	12	2012-08-02 04:59:12.116739
8792	5	1-1315	8788	69	34	500	40	2012-08-02 04:59:12.116739
8788	5	1-2508	8788	34	91	2270	46	2012-08-02 04:59:12.116739
8794	5	1-1881	8793	55	42	801	27	2012-08-02 04:59:12.116739
8795	5	1-2486	8793	42	72	642	61	2012-08-02 04:59:12.116739
8796	5	1-364	8793	72	61	279	55	2012-08-02 04:59:12.116739
8797	5	1-1385	8793	61	34	175	40	2012-08-02 04:59:12.116739
8793	5	1-2508	8793	34	55	829	46	2012-08-02 04:59:12.116739
8799	4	1-1589	8798	40	35	176	52	2012-08-02 04:59:12.116739
8800	4	1-2236	8798	35	61	335	55	2012-08-02 04:59:12.116739
8801	4	1-1385	8798	61	34	223	40	2012-08-02 04:59:12.116739
8798	4	1-2508	8798	34	40	1117	46	2012-08-02 04:59:12.116739
8803	5	1-1716	8802	96	74	1189	54	2012-08-02 04:59:12.978642
8804	5	1-2422	8802	74	5	1149	63	2012-08-02 04:59:12.978642
8805	5	1-615	8802	5	65	265	41	2012-08-02 04:59:12.978642
8806	5	1-2477	8802	65	8	656	7	2012-08-02 04:59:12.978642
8802	5	1-2510	8802	8	96	803	67	2012-08-02 04:59:12.978642
8808	6	1-1309	8807	12	41	2617	65	2012-08-02 04:59:12.978642
8809	6	1-2322	8807	41	78	1863	38	2012-08-02 04:59:12.978642
8810	6	1-1900	8807	78	90	1823	1	2012-08-02 04:59:12.978642
8811	6	1-1858	8807	90	31	1223	9	2012-08-02 04:59:12.978642
8812	6	1-2411	8807	31	8	2164	7	2012-08-02 04:59:12.978642
8807	6	1-2510	8807	8	12	2897	67	2012-08-02 04:59:12.978642
16015	5	1-4623	16015	67	42	12	55	2012-08-02 05:13:31.405362
8814	4	1-2364	8813	54	29	252	81	2012-08-02 04:59:13.840492
8815	4	1-2317	8813	29	90	213	87	2012-08-02 04:59:13.840492
8816	4	1-1975	8813	90	14	143	29	2012-08-02 04:59:13.840492
8813	4	1-2512	8813	14	54	237	77	2012-08-02 04:59:13.840492
8818	5	1-2021	8817	46	40	1635	85	2012-08-02 04:59:13.840492
8819	5	1-2389	8817	40	92	3096	6	2012-08-02 04:59:13.840492
8820	5	1-2493	8817	92	78	1815	79	2012-08-02 04:59:13.840492
8821	5	1-2000	8817	78	14	1628	29	2012-08-02 04:59:13.840492
8817	5	1-2512	8817	14	46	3054	77	2012-08-02 04:59:13.840492
8823	6	1-2471	8822	69	74	138	26	2012-08-02 04:59:14.283096
8824	6	1-1492	8822	74	99	63	77	2012-08-02 04:59:14.283096
8825	6	1-1767	8822	99	5	155	40	2012-08-02 04:59:14.283096
8826	6	1-2467	8822	5	34	98	64	2012-08-02 04:59:14.283096
8827	6	1-2144	8822	34	59	175	98	2012-08-02 04:59:14.283096
8822	6	1-2513	8822	59	69	171	32	2012-08-02 04:59:14.283096
8829	4	1-2351	8828	16	25	1749	44	2012-08-02 04:59:14.724622
8830	4	1-389	8828	25	37	1366	6	2012-08-02 04:59:14.724622
8831	4	1-2473	8828	37	91	1017	62	2012-08-02 04:59:14.724622
8828	4	1-2514	8828	91	16	2911	89	2012-08-02 04:59:14.724622
8833	6	1-2382	8832	8	97	1376	12	2012-08-02 04:59:14.724622
8834	6	1-2269	8832	97	97	552	36	2012-08-02 04:59:14.724622
8835	6	1-1336	8832	97	40	380	85	2012-08-02 04:59:14.724622
8836	6	1-2389	8832	40	37	579	6	2012-08-02 04:59:14.724622
8837	6	1-2473	8832	37	91	506	62	2012-08-02 04:59:14.724622
8832	6	1-2514	8832	91	8	1698	89	2012-08-02 04:59:14.724622
8839	5	1-2068	8838	2	7	3531	57	2012-08-02 04:59:15.586623
8840	5	1-2503	8838	7	55	1043	46	2012-08-02 04:59:15.586623
8841	5	1-1881	8838	55	15	774	27	2012-08-02 04:59:15.586623
8842	5	1-2442	8838	15	3	264	71	2012-08-02 04:59:15.586623
8838	5	1-2516	8838	3	2	2752	97	2012-08-02 04:59:15.586623
8844	6	1-2155	8843	32	96	809	67	2012-08-02 04:59:15.586623
8845	6	1-1716	8843	96	74	963	54	2012-08-02 04:59:15.586623
8846	6	1-2422	8843	74	98	748	63	2012-08-02 04:59:15.586623
8847	6	1-2263	8843	98	15	576	27	2012-08-02 04:59:15.586623
8848	6	1-2442	8843	15	3	228	71	2012-08-02 04:59:15.586623
8843	6	1-2516	8843	3	32	2763	97	2012-08-02 04:59:15.586623
8850	4	1-1926	8849	3	93	278	58	2012-08-02 04:59:15.586623
8851	4	1-1748	8849	93	8	212	50	2012-08-02 04:59:15.586623
8852	4	1-2055	8849	8	3	25	71	2012-08-02 04:59:15.586623
8849	4	1-2516	8849	3	3	347	97	2012-08-02 04:59:15.586623
8854	5	1-2403	8853	77	18	3778	62	2012-08-02 04:59:16.028706
8855	5	1-2433	8853	18	12	1993	81	2012-08-02 04:59:16.028706
8856	5	1-2491	8853	12	21	811	60	2012-08-02 04:59:16.028706
8857	5	1-621	8853	21	95	805	76	2012-08-02 04:59:16.028706
8853	5	1-2517	8853	95	77	3630	71	2012-08-02 04:59:16.028706
8859	6	1-2403	8858	77	18	125	62	2012-08-02 04:59:16.028706
8860	6	1-2433	8858	18	12	85	81	2012-08-02 04:59:16.028706
8861	6	1-2491	8858	12	100	45	60	2012-08-02 04:59:16.028706
8862	6	1-1844	8858	100	42	32	58	2012-08-02 04:59:16.028706
8863	6	1-1232	8858	42	95	16	76	2012-08-02 04:59:16.028706
8858	6	1-2517	8858	95	77	93	71	2012-08-02 04:59:16.028706
8865	4	1-2034	8864	50	43	2706	92	2012-08-02 04:59:16.028706
8866	4	1-1587	8864	43	78	2824	50	2012-08-02 04:59:16.028706
8867	4	1-914	8864	78	95	551	76	2012-08-02 04:59:16.028706
8864	4	1-2517	8864	95	50	3295	71	2012-08-02 04:59:16.028706
8869	5	1-1436	8868	30	88	568	94	2012-08-02 04:59:16.028706
8870	5	1-2342	8868	88	35	468	18	2012-08-02 04:59:16.028706
8871	5	1-2320	8868	35	42	270	58	2012-08-02 04:59:16.028706
8872	5	1-1232	8868	42	95	143	76	2012-08-02 04:59:16.028706
8868	5	1-2517	8868	95	30	890	71	2012-08-02 04:59:16.028706
8874	3	1-2015	8873	76	99	358	32	2012-08-02 04:59:22.447536
8875	3	1-2023	8873	99	12	288	93	2012-08-02 04:59:22.447536
8873	3	1-2532	8873	12	76	945	99	2012-08-02 04:59:22.447536
8877	4	1-2112	8876	24	13	522	74	2012-08-02 04:59:22.867103
8878	4	1-2464	8876	13	19	396	69	2012-08-02 04:59:22.867103
8879	4	1-1341	8876	19	4	437	14	2012-08-02 04:59:22.867103
8876	4	1-2533	8876	4	24	691	2	2012-08-02 04:59:22.867103
8881	4	1-1336	8880	97	40	209	85	2012-08-02 04:59:23.28683
8882	4	1-2389	8880	40	92	461	6	2012-08-02 04:59:23.28683
8883	4	1-2493	8880	92	45	315	79	2012-08-02 04:59:23.28683
8880	4	1-2534	8880	45	97	257	36	2012-08-02 04:59:23.28683
8885	6	1-2390	8884	32	69	4623	32	2012-08-02 04:59:24.126815
8886	6	1-2471	8884	69	89	2604	26	2012-08-02 04:59:24.126815
8887	6	1-1316	8884	89	66	1502	29	2012-08-02 04:59:24.126815
8888	6	1-2428	8884	66	53	695	30	2012-08-02 04:59:24.126815
8889	6	1-2528	8884	53	33	7198	88	2012-08-02 04:59:24.126815
8884	6	1-2536	8884	33	32	4872	11	2012-08-02 04:59:24.126815
8891	6	1-1571	8890	94	56	606	45	2012-08-02 04:59:24.126815
8892	6	1-1552	8890	56	88	402	94	2012-08-02 04:59:24.126815
8893	6	1-2342	8890	88	64	236	18	2012-08-02 04:59:24.126815
8894	6	1-2363	8890	64	53	188	30	2012-08-02 04:59:24.126815
8895	6	1-2528	8890	53	33	2026	88	2012-08-02 04:59:24.126815
8890	6	1-2536	8890	33	94	1424	11	2012-08-02 04:59:24.126815
8897	6	1-1164	8896	74	93	61	1	2012-08-02 04:59:24.126815
8898	6	1-2417	8896	93	98	87	63	2012-08-02 04:59:24.126815
8899	6	1-2263	8896	98	42	86	27	2012-08-02 04:59:24.126815
8900	6	1-2486	8896	42	9	78	61	2012-08-02 04:59:24.126815
8901	6	1-1502	8896	9	33	134	88	2012-08-02 04:59:24.126815
8896	6	1-2536	8896	33	74	114	11	2012-08-02 04:59:24.126815
8902	1	1-1552	\N	56	56	1040	94	2012-08-02 04:59:24.126815
8904	6	1-2487	8903	32	92	833	85	2012-08-02 04:59:24.55777
8905	6	1-2457	8903	92	17	243	78	2012-08-02 04:59:24.55777
8906	6	1-1708	8903	17	91	421	16	2012-08-02 04:59:24.55777
8907	6	1-1650	8903	91	56	67	86	2012-08-02 04:59:24.55777
8908	6	1-2484	8903	56	14	148	47	2012-08-02 04:59:24.55777
8903	6	1-2537	8903	14	32	133	28	2012-08-02 04:59:24.55777
8910	6	1-2105	8909	4	32	388	28	2012-08-02 04:59:24.977451
8911	6	1-2487	8909	32	40	1526	85	2012-08-02 04:59:24.977451
8912	6	1-2389	8909	40	92	2678	6	2012-08-02 04:59:24.977451
8913	6	1-2493	8909	92	40	1109	79	2012-08-02 04:59:24.977451
8914	6	1-2331	8909	40	45	728	33	2012-08-02 04:59:24.977451
8909	6	1-2538	8909	45	4	916	66	2012-08-02 04:59:24.977451
8916	5	1-2105	8915	4	74	2022	28	2012-08-02 04:59:24.977451
8917	5	1-2478	8915	74	98	3672	97	2012-08-02 04:59:24.977451
8918	5	1-2502	8915	98	74	3014	44	2012-08-02 04:59:24.977451
8919	5	1-1394	8915	74	45	1867	33	2012-08-02 04:59:24.977451
8915	5	1-2538	8915	45	4	3348	66	2012-08-02 04:59:24.977451
16021	5	1-3770	16020	31	69	1491	32	2012-08-02 05:13:31.405362
8921	6	1-2105	8920	4	74	479	28	2012-08-02 04:59:24.977451
8922	6	1-2478	8920	74	74	911	97	2012-08-02 04:59:24.977451
8923	6	1-1195	8920	74	83	353	99	2012-08-02 04:59:24.977451
8924	6	1-2243	8920	83	40	470	79	2012-08-02 04:59:24.977451
8925	6	1-2331	8920	40	45	367	33	2012-08-02 04:59:24.977451
8920	6	1-2538	8920	45	4	723	66	2012-08-02 04:59:24.977451
16022	5	1-4239	16020	69	73	908	10	2012-08-02 05:13:31.405362
8927	3	1-2454	8926	3	40	400	79	2012-08-02 04:59:24.977451
8928	3	1-2331	8926	40	45	315	33	2012-08-02 04:59:24.977451
8926	3	1-2538	8926	45	3	623	66	2012-08-02 04:59:24.977451
16023	5	1-4551	16020	73	13	1865	73	2012-08-02 05:13:31.405362
8930	3	1-1069	8929	38	6	201	45	2012-08-02 04:59:27.178037
8931	3	1-2170	8929	6	23	176	9	2012-08-02 04:59:27.178037
8929	3	1-2543	8929	23	38	2793	15	2012-08-02 04:59:27.178037
16024	5	1-4408	16020	13	67	1353	1	2012-08-02 05:13:31.405362
8933	6	1-2170	8932	6	23	342	9	2012-08-02 04:59:28.040138
8934	6	1-2543	8932	23	84	4614	15	2012-08-02 04:59:28.040138
8935	6	1-2384	8932	84	30	2836	37	2012-08-02 04:59:28.040138
8936	6	1-1723	8932	30	93	275	58	2012-08-02 04:59:28.040138
8937	6	1-1748	8932	93	61	293	50	2012-08-02 04:59:28.040138
8932	6	1-2545	8932	61	6	460	45	2012-08-02 04:59:28.040138
8939	3	1-2170	8938	6	81	2275	9	2012-08-02 04:59:28.040138
8940	3	1-1599	8938	81	61	1567	50	2012-08-02 04:59:28.040138
8938	3	1-2545	8938	61	6	2744	45	2012-08-02 04:59:28.040138
8942	6	1-1820	8941	66	37	320	95	2012-08-02 04:59:28.040138
8943	6	1-1145	8941	37	66	229	29	2012-08-02 04:59:28.040138
8944	6	1-2428	8941	66	64	151	30	2012-08-02 04:59:28.040138
8945	6	1-1294	8941	64	18	105	84	2012-08-02 04:59:28.040138
8946	6	1-2356	8941	18	61	205	50	2012-08-02 04:59:28.040138
8941	6	1-2545	8941	61	66	361	45	2012-08-02 04:59:28.040138
8948	3	1-1236	8947	89	4	1056	14	2012-08-02 04:59:28.470767
8949	3	1-2533	8947	4	6	779	2	2012-08-02 04:59:28.470767
8947	3	1-2546	8947	6	89	2614	72	2012-08-02 04:59:28.470767
8951	5	1-2182	8950	74	50	2089	87	2012-08-02 04:59:28.89117
8952	5	1-2480	8950	50	34	637	64	2012-08-02 04:59:28.89117
8953	5	1-2144	8950	34	15	532	98	2012-08-02 04:59:28.89117
8954	5	1-1846	8950	15	23	277	32	2012-08-02 04:59:28.89117
8950	5	1-2547	8950	23	74	4054	36	2012-08-02 04:59:28.89117
8956	3	1-2182	8955	74	27	3143	87	2012-08-02 04:59:28.89117
8957	3	1-2378	8955	27	23	317	32	2012-08-02 04:59:28.89117
8955	3	1-2547	8955	23	74	5320	36	2012-08-02 04:59:28.89117
8959	4	1-500	8958	21	29	819	81	2012-08-02 04:59:30.194222
8960	4	1-2317	8958	29	90	841	87	2012-08-02 04:59:30.194222
8961	4	1-1975	8958	90	18	689	29	2012-08-02 04:59:30.194222
8958	4	1-2550	8958	18	21	4600	15	2012-08-02 04:59:30.194222
8963	6	1-2224	8962	23	25	1618	44	2012-08-02 04:59:30.763067
8964	6	1-389	8962	25	92	1921	6	2012-08-02 04:59:30.763067
8965	6	1-2493	8962	92	78	1173	79	2012-08-02 04:59:30.763067
8966	6	1-2000	8962	78	66	1096	29	2012-08-02 04:59:30.763067
8967	6	1-2428	8962	66	11	639	30	2012-08-02 04:59:30.763067
8962	6	1-2551	8962	11	23	1771	34	2012-08-02 04:59:30.763067
8969	6	1-2062	8968	66	46	2052	20	2012-08-02 04:59:30.763067
8970	6	1-2525	8968	46	45	3542	25	2012-08-02 04:59:30.763067
8971	6	1-939	8968	45	18	922	76	2012-08-02 04:59:30.763067
8972	6	1-1824	8968	18	64	257	18	2012-08-02 04:59:30.763067
8973	6	1-2363	8968	64	11	267	30	2012-08-02 04:59:30.763067
8968	6	1-2551	8968	11	66	790	34	2012-08-02 04:59:30.763067
8975	6	1-2224	8974	23	54	3068	44	2012-08-02 04:59:30.763067
8976	6	1-997	8974	54	65	3248	22	2012-08-02 04:59:30.763067
8977	6	1-2522	8974	65	88	679	35	2012-08-02 04:59:30.763067
8978	6	1-2078	8974	88	64	1011	18	2012-08-02 04:59:30.763067
8979	6	1-2363	8974	64	11	1052	30	2012-08-02 04:59:30.763067
8974	6	1-2551	8974	11	23	3129	34	2012-08-02 04:59:30.763067
8981	6	1-1997	8980	65	45	593	79	2012-08-02 04:59:30.763067
8982	6	1-2534	8980	45	74	476	36	2012-08-02 04:59:30.763067
8983	6	1-2182	8980	74	90	508	87	2012-08-02 04:59:30.763067
8984	6	1-1975	8980	90	66	446	29	2012-08-02 04:59:30.763067
8985	6	1-2428	8980	66	11	286	30	2012-08-02 04:59:30.763067
8980	6	1-2551	8980	11	65	871	34	2012-08-02 04:59:30.763067
8987	6	1-1970	8986	98	86	1885	89	2012-08-02 04:59:30.763067
8988	6	1-1588	8986	86	16	572	60	2012-08-02 04:59:30.763067
8989	6	1-1998	8986	16	45	431	75	2012-08-02 04:59:30.763067
8990	6	1-1247	8986	45	64	548	18	2012-08-02 04:59:30.763067
8991	6	1-2363	8986	64	11	598	30	2012-08-02 04:59:30.763067
8986	6	1-2551	8986	11	98	1866	34	2012-08-02 04:59:30.763067
8993	3	1-1858	8992	90	83	1888	9	2012-08-02 04:59:31.663968
8994	3	1-1921	8992	83	11	1007	4	2012-08-02 04:59:31.663968
8992	3	1-2553	8992	11	90	6338	1	2012-08-02 04:59:31.663968
8996	6	1-2491	8995	12	16	2218	60	2012-08-02 04:59:32.912346
8997	6	1-1998	8995	16	45	1239	75	2012-08-02 04:59:32.912346
8998	6	1-1247	8995	45	16	836	18	2012-08-02 04:59:32.912346
8999	6	1-2443	8995	16	83	579	99	2012-08-02 04:59:32.912346
9000	6	1-2243	8995	83	88	434	79	2012-08-02 04:59:32.912346
8995	6	1-2556	8995	88	12	7696	81	2012-08-02 04:59:32.912346
9001	1	1-1247	\N	45	45	200	18	2012-08-02 04:59:32.912346
9003	4	1-2301	9002	94	55	2144	82	2012-08-02 04:59:33.751963
9004	4	1-2555	9002	55	2	1275	11	2012-08-02 04:59:33.751963
9005	4	1-1055	9002	2	49	826	20	2012-08-02 04:59:33.751963
9002	4	1-2558	9002	49	94	4342	18	2012-08-02 04:59:33.751963
9007	5	1-2363	9006	64	64	904	30	2012-08-02 04:59:33.751963
9008	5	1-1294	9006	64	53	491	84	2012-08-02 04:59:33.751963
9009	5	1-2205	9006	53	77	364	44	2012-08-02 04:59:33.751963
9010	5	1-2175	9006	77	49	191	20	2012-08-02 04:59:33.751963
9006	5	1-2558	9006	49	64	1090	18	2012-08-02 04:59:33.751963
9012	6	1-1688	9011	66	13	985	74	2012-08-02 04:59:33.751963
9013	6	1-2464	9011	13	35	504	69	2012-08-02 04:59:33.751963
9014	6	1-1818	9011	35	66	395	41	2012-08-02 04:59:33.751963
9015	6	1-2374	9011	66	2	205	11	2012-08-02 04:59:33.751963
9016	6	1-1055	9011	2	49	169	20	2012-08-02 04:59:33.751963
9011	6	1-2558	9011	49	66	1124	18	2012-08-02 04:59:33.751963
9018	5	1-1679	9017	66	42	1440	27	2012-08-02 04:59:33.751963
9019	5	1-2486	9017	42	9	962	61	2012-08-02 04:59:33.751963
9020	5	1-1710	9017	9	77	585	44	2012-08-02 04:59:33.751963
9021	5	1-2175	9017	77	49	358	20	2012-08-02 04:59:33.751963
9017	5	1-2558	9017	49	66	2386	18	2012-08-02 04:59:33.751963
16020	5	1-4623	16020	67	31	2464	55	2012-08-02 05:13:31.405362
9023	6	1-1678	9022	11	30	900	71	2012-08-02 04:59:35.508993
9024	6	1-1436	9022	30	28	493	94	2012-08-02 04:59:35.508993
9025	6	1-2448	9022	28	24	533	19	2012-08-02 04:59:35.508993
9026	6	1-2368	9022	24	61	968	22	2012-08-02 04:59:35.508993
9027	6	1-2469	9022	61	94	824	80	2012-08-02 04:59:35.508993
9022	6	1-2562	9022	94	11	1778	45	2012-08-02 04:59:35.508993
9029	5	1-2050	9028	68	39	3133	68	2012-08-02 04:59:35.508993
9030	5	1-1346	9028	39	12	2232	81	2012-08-02 04:59:35.508993
9031	5	1-2491	9028	12	9	1170	60	2012-08-02 04:59:35.508993
9032	5	1-591	9028	9	94	1323	80	2012-08-02 04:59:35.508993
9028	5	1-2562	9028	94	68	3107	45	2012-08-02 04:59:35.508993
16026	4	1-3573	16025	6	73	368	49	2012-08-02 05:13:31.8252
9034	6	1-2050	9033	68	12	416	68	2012-08-02 04:59:35.508993
9035	6	1-1927	9033	12	29	406	31	2012-08-02 04:59:35.508993
9036	6	1-2246	9033	29	94	372	25	2012-08-02 04:59:35.508993
9037	6	1-2259	9033	94	61	192	22	2012-08-02 04:59:35.508993
9038	6	1-2469	9033	61	94	175	80	2012-08-02 04:59:35.508993
9033	6	1-2562	9033	94	68	421	45	2012-08-02 04:59:35.508993
9039	1	1-591	\N	9	9	755	80	2012-08-02 04:59:35.508993
16027	4	1-3714	16025	73	37	415	36	2012-08-02 05:13:31.8252
9041	5	1-2374	9040	66	14	169	11	2012-08-02 04:59:36.348873
9042	5	1-2316	9040	14	43	511	92	2012-08-02 04:59:36.348873
9043	5	1-1587	9040	43	16	447	50	2012-08-02 04:59:36.348873
9044	5	1-1934	9040	16	98	108	77	2012-08-02 04:59:36.348873
9040	5	1-2564	9040	98	66	320	41	2012-08-02 04:59:36.348873
9046	5	1-2374	9045	66	74	1403	11	2012-08-02 04:59:36.348873
9047	5	1-1164	9045	74	93	722	1	2012-08-02 04:59:36.348873
9048	5	1-2417	9045	93	39	981	63	2012-08-02 04:59:36.348873
9049	5	1-1925	9045	39	98	862	77	2012-08-02 04:59:36.348873
9045	5	1-2564	9045	98	66	2605	41	2012-08-02 04:59:36.348873
9051	4	1-895	9050	10	59	15	80	2012-08-02 04:59:36.348873
9052	4	1-2214	9050	59	39	7	63	2012-08-02 04:59:36.348873
9053	4	1-1925	9050	39	98	7	77	2012-08-02 04:59:36.348873
9050	4	1-2564	9050	98	10	24	41	2012-08-02 04:59:36.348873
9054	1	1-1925	\N	39	39	1961	77	2012-08-02 04:59:36.348873
9056	6	1-2433	9055	18	29	902	81	2012-08-02 04:59:38.867502
9057	6	1-2317	9055	29	50	809	87	2012-08-02 04:59:38.867502
9058	6	1-2480	9055	50	77	418	64	2012-08-02 04:59:38.867502
9059	6	1-1633	9055	77	25	308	33	2012-08-02 04:59:38.867502
9060	6	1-1759	9055	25	43	792	5	2012-08-02 04:59:38.867502
9055	6	1-2570	9055	43	18	1606	62	2012-08-02 04:59:38.867502
9062	6	1-2433	9061	18	89	2592	81	2012-08-02 04:59:38.867502
9063	6	1-1318	9061	89	82	1478	74	2012-08-02 04:59:38.867502
9064	6	1-2563	9061	82	86	830	38	2012-08-02 04:59:38.867502
9065	6	1-1821	9061	86	19	766	70	2012-08-02 04:59:38.867502
9066	6	1-2274	9061	19	43	1648	5	2012-08-02 04:59:38.867502
9061	6	1-2570	9061	43	18	3928	62	2012-08-02 04:59:38.867502
9068	5	1-2568	9067	87	22	1601	24	2012-08-02 04:59:39.718407
9069	5	1-2237	9067	22	91	2888	16	2012-08-02 04:59:39.718407
9070	5	1-1650	9067	91	56	470	86	2012-08-02 04:59:39.718407
9071	5	1-2484	9067	56	15	1063	47	2012-08-02 04:59:39.718407
9067	5	1-2572	9067	15	87	1138	33	2012-08-02 04:59:39.718407
9073	6	1-2568	9072	87	91	2810	24	2012-08-02 04:59:39.718407
9074	6	1-1911	9072	91	82	3047	83	2012-08-02 04:59:39.718407
9075	6	1-249	9072	82	22	768	84	2012-08-02 04:59:39.718407
9076	6	1-1958	9072	22	56	747	86	2012-08-02 04:59:39.718407
9077	6	1-2484	9072	56	15	1747	47	2012-08-02 04:59:39.718407
9072	6	1-2572	9072	15	87	1933	33	2012-08-02 04:59:39.718407
9079	3	1-2167	9078	32	98	841	97	2012-08-02 04:59:40.138294
9080	3	1-2502	9078	98	21	473	44	2012-08-02 04:59:40.138294
9078	3	1-2573	9078	21	32	1290	73	2012-08-02 04:59:40.138294
9082	6	1-1603	9081	43	74	36	71	2012-08-02 04:59:40.138294
9083	6	1-235	9081	74	14	15	47	2012-08-02 04:59:40.138294
9084	6	1-2537	9081	14	74	11	28	2012-08-02 04:59:40.138294
9085	6	1-2478	9081	74	98	20	97	2012-08-02 04:59:40.138294
9086	6	1-2502	9081	98	21	15	44	2012-08-02 04:59:40.138294
9081	6	1-2573	9081	21	43	56	73	2012-08-02 04:59:40.138294
9088	3	1-1864	9087	56	86	1129	11	2012-08-02 04:59:40.138294
9089	3	1-1505	9087	86	21	888	44	2012-08-02 04:59:40.138294
9087	3	1-2573	9087	21	56	3819	73	2012-08-02 04:59:40.138294
9091	6	1-2395	9090	90	93	281	1	2012-08-02 04:59:40.138294
9092	6	1-2417	9090	93	98	430	63	2012-08-02 04:59:40.138294
9093	6	1-2263	9090	98	42	456	27	2012-08-02 04:59:40.138294
9094	6	1-2486	9090	42	9	449	61	2012-08-02 04:59:40.138294
9095	6	1-1710	9090	9	21	403	44	2012-08-02 04:59:40.138294
9090	6	1-2573	9090	21	90	1787	73	2012-08-02 04:59:40.138294
9097	6	1-1460	9096	78	69	547	12	2012-08-02 04:59:40.138294
9098	6	1-1315	9096	69	39	593	40	2012-08-02 04:59:40.138294
9099	6	1-2438	9096	39	42	323	27	2012-08-02 04:59:40.138294
9100	6	1-2486	9096	42	9	331	61	2012-08-02 04:59:40.138294
9101	6	1-1710	9096	9	21	309	44	2012-08-02 04:59:40.138294
9096	6	1-2573	9096	21	78	1425	73	2012-08-02 04:59:40.138294
9103	6	1-1391	9102	65	70	227	3	2012-08-02 04:59:40.580613
9104	6	1-2033	9102	70	12	227	81	2012-08-02 04:59:40.580613
9105	6	1-2491	9102	12	56	124	60	2012-08-02 04:59:40.580613
9106	6	1-2321	9102	56	55	226	48	2012-08-02 04:59:40.580613
9107	6	1-2396	9102	55	52	347	93	2012-08-02 04:59:40.580613
9102	6	1-2574	9102	52	65	530	18	2012-08-02 04:59:40.580613
9109	6	1-2443	9108	16	69	3848	99	2012-08-02 04:59:40.580613
9110	6	1-1992	9108	69	12	1743	81	2012-08-02 04:59:40.580613
9111	6	1-2491	9108	12	56	952	60	2012-08-02 04:59:40.580613
9112	6	1-2321	9108	56	55	1739	48	2012-08-02 04:59:40.580613
9113	6	1-2396	9108	55	52	2682	93	2012-08-02 04:59:40.580613
9108	6	1-2574	9108	52	16	4100	18	2012-08-02 04:59:40.580613
9115	6	1-2241	9114	72	83	608	16	2012-08-02 04:59:40.580613
9116	6	1-2392	9114	83	38	286	82	2012-08-02 04:59:40.580613
9117	6	1-1797	9114	38	56	133	60	2012-08-02 04:59:40.580613
9118	6	1-2321	9114	56	55	248	48	2012-08-02 04:59:40.580613
9119	6	1-2396	9114	55	52	389	93	2012-08-02 04:59:40.580613
9114	6	1-2574	9114	52	72	605	18	2012-08-02 04:59:40.580613
9121	6	1-2320	9120	35	74	864	58	2012-08-02 04:59:40.580613
9122	6	1-327	9120	74	5	1121	40	2012-08-02 04:59:40.580613
9123	6	1-2467	9120	5	84	718	64	2012-08-02 04:59:40.580613
9124	6	1-756	9120	84	55	608	48	2012-08-02 04:59:40.580613
9125	6	1-2396	9120	55	52	960	93	2012-08-02 04:59:40.580613
9120	6	1-2574	9120	52	35	1501	18	2012-08-02 04:59:40.580613
16028	4	1-4539	16025	37	8	854	1	2012-08-02 05:13:31.8252
9127	3	1-2486	9126	42	9	4393	61	2012-08-02 04:59:41.873675
9128	3	1-1502	9126	9	71	4164	88	2012-08-02 04:59:41.873675
9126	3	1-2577	9126	71	42	8645	27	2012-08-02 04:59:41.873675
9130	3	1-1353	9129	33	26	1489	51	2012-08-02 04:59:43.144142
9131	3	1-2376	9129	26	94	284	29	2012-08-02 04:59:43.144142
9129	3	1-2580	9129	94	33	2105	33	2012-08-02 04:59:43.144142
9133	5	1-2568	9132	87	63	1612	24	2012-08-02 04:59:43.144142
9134	5	1-498	9132	63	12	673	81	2012-08-02 04:59:43.144142
9135	5	1-2491	9132	12	16	292	60	2012-08-02 04:59:43.144142
9136	5	1-2347	9132	16	94	167	29	2012-08-02 04:59:43.144142
9132	5	1-2580	9132	94	87	1411	33	2012-08-02 04:59:43.144142
9138	6	1-2568	9137	87	15	3540	24	2012-08-02 04:59:43.144142
9139	6	1-1917	9137	15	7	2466	57	2012-08-02 04:59:43.144142
9140	6	1-2503	9137	7	63	992	46	2012-08-02 04:59:43.144142
9141	6	1-2377	9137	63	90	483	87	2012-08-02 04:59:43.144142
9142	6	1-1975	9137	90	94	353	29	2012-08-02 04:59:43.144142
9137	6	1-2580	9137	94	87	3039	33	2012-08-02 04:59:43.144142
9144	6	1-2535	9143	88	34	1456	59	2012-08-02 04:59:43.144142
9145	6	1-2123	9143	34	7	806	57	2012-08-02 04:59:43.144142
9146	6	1-2503	9143	7	63	389	46	2012-08-02 04:59:43.144142
9147	6	1-2377	9143	63	90	228	87	2012-08-02 04:59:43.144142
9148	6	1-1975	9143	90	94	200	29	2012-08-02 04:59:43.144142
9143	6	1-2580	9143	94	88	2068	33	2012-08-02 04:59:43.144142
9150	6	1-2335	9149	34	35	917	52	2012-08-02 04:59:46.590838
9151	6	1-2236	9149	35	9	1800	55	2012-08-02 04:59:46.590838
9152	6	1-465	9149	9	14	386	47	2012-08-02 04:59:46.590838
9153	6	1-2537	9149	14	32	307	28	2012-08-02 04:59:46.590838
9154	6	1-2487	9149	32	37	1698	85	2012-08-02 04:59:46.590838
9149	6	1-2588	9149	37	34	3737	54	2012-08-02 04:59:46.590838
9156	5	1-2248	9155	99	32	3422	49	2012-08-02 04:59:47.41945
9157	5	1-2472	9155	32	74	3419	11	2012-08-02 04:59:47.41945
9158	5	1-1164	9155	74	90	1393	1	2012-08-02 04:59:47.41945
9159	5	1-1858	9155	90	96	666	9	2012-08-02 04:59:47.41945
9155	5	1-2590	9155	96	99	5623	56	2012-08-02 04:59:47.41945
9161	6	1-2581	9160	22	46	1398	72	2012-08-02 04:59:48.689902
9162	6	1-2549	9160	46	1	2324	13	2012-08-02 04:59:48.689902
9163	6	1-1689	9160	1	90	422	1	2012-08-02 04:59:48.689902
9164	6	1-1858	9160	90	31	217	9	2012-08-02 04:59:48.689902
9165	6	1-2411	9160	31	96	294	7	2012-08-02 04:59:48.689902
9160	6	1-2593	9160	96	22	316	91	2012-08-02 04:59:48.689902
9167	4	1-2530	9166	74	76	1044	8	2012-08-02 04:59:48.689902
9168	4	1-275	9166	76	31	726	9	2012-08-02 04:59:48.689902
9169	4	1-2411	9166	31	96	1013	7	2012-08-02 04:59:48.689902
9166	4	1-2593	9166	96	74	1118	91	2012-08-02 04:59:48.689902
9171	3	1-2530	9170	74	4	2203	8	2012-08-02 04:59:48.689902
9172	3	1-4	9170	4	96	1661	7	2012-08-02 04:59:48.689902
9170	3	1-2593	9170	96	74	2080	91	2012-08-02 04:59:48.689902
9174	5	1-2446	9173	81	82	996	74	2012-08-02 04:59:49.574207
9175	5	1-2563	9173	82	86	414	38	2012-08-02 04:59:49.574207
9176	5	1-1821	9173	86	8	282	70	2012-08-02 04:59:49.574207
9177	5	1-2458	9173	8	7	4250	17	2012-08-02 04:59:49.574207
9173	5	1-2595	9173	7	81	3918	14	2012-08-02 04:59:49.574207
9179	5	1-2439	9178	4	21	880	51	2012-08-02 04:59:51.264292
9180	5	1-2349	9178	21	1	860	13	2012-08-02 04:59:51.264292
9181	5	1-1689	9178	1	58	157	1	2012-08-02 04:59:51.264292
9182	5	1-2121	9178	58	11	180	32	2012-08-02 04:59:51.264292
9178	5	1-2599	9178	11	4	301	94	2012-08-02 04:59:51.264292
9184	4	1-2448	9183	28	47	821	19	2012-08-02 04:59:51.264292
9185	4	1-2231	9183	47	15	572	98	2012-08-02 04:59:51.264292
9186	4	1-1846	9183	15	11	489	32	2012-08-02 04:59:51.264292
9183	4	1-2599	9183	11	28	869	94	2012-08-02 04:59:51.264292
9188	4	1-2448	9187	28	47	1327	19	2012-08-02 04:59:51.264292
9189	4	1-2231	9187	47	59	955	98	2012-08-02 04:59:51.264292
9190	4	1-2513	9187	59	11	741	32	2012-08-02 04:59:51.264292
9187	4	1-2599	9187	11	28	1359	94	2012-08-02 04:59:51.264292
9192	6	1-2342	9191	88	66	1593	18	2012-08-02 04:59:51.264292
9193	6	1-1688	9191	66	82	1527	74	2012-08-02 04:59:51.264292
9194	6	1-2563	9191	82	78	822	38	2012-08-02 04:59:51.264292
9195	6	1-1900	9191	78	58	764	1	2012-08-02 04:59:51.264292
9196	6	1-2121	9191	58	11	1077	32	2012-08-02 04:59:51.264292
9191	6	1-2599	9191	11	88	2219	94	2012-08-02 04:59:51.264292
9198	3	1-2468	9197	63	44	71	68	2012-08-02 04:59:53.796257
9199	3	1-2009	9197	44	64	188	100	2012-08-02 04:59:53.796257
9197	3	1-2605	9197	64	63	185	62	2012-08-02 04:59:53.796257
9201	2	1-1316	9200	89	31	3231	29	2012-08-02 04:59:54.635641
9200	2	1-2607	9200	31	89	7161	26	2012-08-02 04:59:54.635641
9203	5	1-2061	9202	65	11	94	32	2012-08-02 04:59:54.635641
9204	5	1-2599	9202	11	4	179	94	2012-08-02 04:59:54.635641
9205	5	1-2439	9202	4	26	591	51	2012-08-02 04:59:54.635641
9206	5	1-2376	9202	26	31	133	29	2012-08-02 04:59:54.635641
9202	5	1-2607	9202	31	65	441	26	2012-08-02 04:59:54.635641
9208	6	1-1359	9207	60	48	985	59	2012-08-02 04:59:56.337262
9209	6	1-2552	9207	48	93	1077	53	2012-08-02 04:59:56.337262
9210	6	1-2527	9207	93	9	157	39	2012-08-02 04:59:56.337262
9211	6	1-2476	9207	9	58	943	96	2012-08-02 04:59:56.337262
9212	6	1-1829	9207	58	69	1558	90	2012-08-02 04:59:56.337262
9207	6	1-2611	9207	69	60	1031	3	2012-08-02 04:59:56.337262
9214	4	1-2513	9213	59	69	968	32	2012-08-02 04:59:57.165507
9215	4	1-2471	9213	69	36	798	26	2012-08-02 04:59:57.165507
9216	4	1-999	9213	36	2	622	59	2012-08-02 04:59:57.165507
9213	4	1-2613	9213	2	59	966	98	2012-08-02 04:59:57.165507
9218	4	1-2502	9217	98	25	1053	44	2012-08-02 04:59:57.596589
9219	4	1-389	9217	25	92	1421	6	2012-08-02 04:59:57.596589
9220	4	1-2493	9217	92	56	987	79	2012-08-02 04:59:57.596589
9217	4	1-2614	9217	56	98	1090	97	2012-08-02 04:59:57.596589
9222	6	1-2349	9221	21	82	3358	13	2012-08-02 04:59:59.297857
9223	6	1-2256	9221	82	48	574	3	2012-08-02 04:59:59.297857
9224	6	1-2338	9221	48	22	378	24	2012-08-02 04:59:59.297857
9225	6	1-2237	9221	22	17	527	16	2012-08-02 04:59:59.297857
9221	6	1-2618	9221	43	21	3303	51	2012-08-02 04:59:59.297857
9226	6	1-2597	9221	17	43	80	66	2012-08-02 04:59:59.297857
16025	4	1-4624	16025	8	6	1276	87	2012-08-02 05:13:31.8252
9228	5	1-2376	9227	26	54	1098	29	2012-08-02 04:59:59.297857
9229	5	1-2110	9227	54	22	480	24	2012-08-02 04:59:59.297857
9230	5	1-2237	9227	22	17	709	16	2012-08-02 04:59:59.297857
9231	5	1-2597	9227	17	43	115	66	2012-08-02 04:59:59.297857
9227	5	1-2618	9227	43	26	5000	51	2012-08-02 04:59:59.297857
9233	6	1-2438	9232	39	42	64	27	2012-08-02 05:00:00.137569
9234	6	1-2486	9232	42	9	53	61	2012-08-02 05:00:00.137569
9235	6	1-1710	9232	9	77	40	44	2012-08-02 05:00:00.137569
9236	6	1-2175	9232	77	6	30	20	2012-08-02 05:00:00.137569
9237	6	1-2619	9232	6	2	16	73	2012-08-02 05:00:00.137569
9232	6	1-2620	9232	2	39	145	40	2012-08-02 05:00:00.137569
16030	5	1-3232	16029	49	57	4589	16	2012-08-02 05:13:31.8252
9239	6	1-2467	9238	5	84	2288	64	2012-08-02 05:00:00.137569
9240	6	1-756	9238	84	55	1497	48	2012-08-02 05:00:00.137569
9241	6	1-2396	9238	55	43	1824	93	2012-08-02 05:00:00.137569
9242	6	1-2441	9238	43	6	999	20	2012-08-02 05:00:00.137569
9243	6	1-2619	9238	6	2	509	73	2012-08-02 05:00:00.137569
9238	6	1-2620	9238	2	5	4630	40	2012-08-02 05:00:00.137569
9245	6	1-2438	9244	39	77	156	27	2012-08-02 05:00:00.137569
9246	6	1-2524	9244	77	9	62	61	2012-08-02 05:00:00.137569
9247	6	1-1710	9244	9	77	54	44	2012-08-02 05:00:00.137569
9248	6	1-2175	9244	77	6	50	20	2012-08-02 05:00:00.137569
9249	6	1-2619	9244	6	2	29	73	2012-08-02 05:00:00.137569
9244	6	1-2620	9244	2	39	309	40	2012-08-02 05:00:00.137569
9250	1	1-1710	\N	9	9	2281	44	2012-08-02 05:00:00.137569
9252	5	1-2441	9251	43	85	4007	20	2012-08-02 05:00:00.568722
9253	5	1-2385	9251	85	39	3576	100	2012-08-02 05:00:00.568722
9254	5	1-2589	9251	39	14	1004	11	2012-08-02 05:00:00.568722
9255	5	1-2316	9251	14	72	3194	92	2012-08-02 05:00:00.568722
9251	5	1-2621	9251	72	43	6631	93	2012-08-02 05:00:00.568722
9257	6	1-2606	9256	82	7	3367	44	2012-08-02 05:00:01.112078
9258	6	1-2612	9256	7	47	982	89	2012-08-02 05:00:01.112078
9259	6	1-2615	9256	47	25	881	52	2012-08-02 05:00:01.112078
9260	6	1-2518	9256	25	82	290	83	2012-08-02 05:00:01.112078
9261	6	1-249	9256	82	52	46	84	2012-08-02 05:00:01.112078
9256	6	1-2622	9256	52	82	4519	56	2012-08-02 05:00:01.112078
9263	4	1-2146	9262	98	20	2272	23	2012-08-02 05:00:01.112078
9264	4	1-1979	9262	20	82	488	83	2012-08-02 05:00:01.112078
9265	4	1-249	9262	82	52	50	84	2012-08-02 05:00:01.112078
9262	4	1-2622	9262	52	98	5245	56	2012-08-02 05:00:01.112078
9267	5	1-2603	9266	18	59	643	80	2012-08-02 05:00:02.038419
9268	5	1-2214	9266	59	59	353	63	2012-08-02 05:00:02.038419
9269	5	1-2204	9266	59	81	407	46	2012-08-02 05:00:02.038419
9270	5	1-2048	9266	81	88	280	53	2012-08-02 05:00:02.038419
9266	5	1-2624	9266	88	18	334	83	2012-08-02 05:00:02.038419
9272	6	1-2451	9271	53	97	367	12	2012-08-02 05:00:02.877858
9273	6	1-2269	9271	97	76	162	36	2012-08-02 05:00:02.877858
9274	6	1-2427	9271	76	39	437	100	2012-08-02 05:00:02.877858
9275	6	1-2589	9271	39	14	130	11	2012-08-02 05:00:02.877858
9276	6	1-2316	9271	14	72	439	92	2012-08-02 05:00:02.877858
9271	6	1-2626	9271	72	53	343	69	2012-08-02 05:00:02.877858
9278	6	1-2596	9277	89	63	965	5	2012-08-02 05:00:03.319857
9279	6	1-2601	9277	63	18	550	62	2012-08-02 05:00:03.319857
9280	6	1-2433	9277	18	89	315	81	2012-08-02 05:00:03.319857
9281	6	1-1318	9277	89	82	176	74	2012-08-02 05:00:03.319857
9282	6	1-2563	9277	82	21	86	38	2012-08-02 05:00:03.319857
9277	6	1-2627	9277	21	89	285	21	2012-08-02 05:00:03.319857
9284	3	1-996	9283	42	82	877	74	2012-08-02 05:00:03.319857
9285	3	1-2563	9283	82	21	464	38	2012-08-02 05:00:03.319857
9283	3	1-2627	9283	21	42	1679	21	2012-08-02 05:00:03.319857
9287	5	1-1067	9286	19	65	471	87	2012-08-02 05:00:03.319857
9288	5	1-2576	9286	65	66	495	18	2012-08-02 05:00:03.319857
9289	5	1-1688	9286	66	82	520	74	2012-08-02 05:00:03.319857
9290	5	1-2563	9286	82	21	306	38	2012-08-02 05:00:03.319857
9286	5	1-2627	9286	21	19	1235	21	2012-08-02 05:00:03.319857
9292	6	1-1791	9291	33	63	1895	46	2012-08-02 05:00:03.319857
9293	6	1-2377	9291	63	65	1172	87	2012-08-02 05:00:03.319857
9294	6	1-2576	9291	65	66	1298	18	2012-08-02 05:00:03.319857
9295	6	1-1688	9291	66	82	1435	74	2012-08-02 05:00:03.319857
9296	6	1-2563	9291	82	21	890	38	2012-08-02 05:00:03.319857
9291	6	1-2627	9291	21	33	3777	21	2012-08-02 05:00:03.319857
9298	5	1-2555	9297	55	84	236	11	2012-08-02 05:00:04.568435
9299	5	1-2539	9297	84	69	212	12	2012-08-02 05:00:04.568435
9300	5	1-1315	9297	69	5	224	40	2012-08-02 05:00:04.568435
9301	5	1-2467	9297	5	30	135	64	2012-08-02 05:00:04.568435
9297	5	1-2630	9297	30	55	210	82	2012-08-02 05:00:04.568435
9303	6	1-1958	9302	22	56	123	86	2012-08-02 05:00:05.862802
9304	6	1-2484	9302	56	15	276	47	2012-08-02 05:00:05.862802
9305	6	1-1855	9302	15	33	297	57	2012-08-02 05:00:05.862802
9306	6	1-2137	9302	33	99	147	56	2012-08-02 05:00:05.862802
9307	6	1-2248	9302	99	71	130	49	2012-08-02 05:00:05.862802
9302	6	1-2633	9302	71	22	132	84	2012-08-02 05:00:05.862802
9309	6	1-2417	9308	93	73	3673	63	2012-08-02 05:00:06.713494
9310	6	1-664	9308	73	61	3260	50	2012-08-02 05:00:06.713494
9311	6	1-2545	9308	61	19	5608	45	2012-08-02 05:00:06.713494
9312	6	1-2566	9308	19	33	1944	67	2012-08-02 05:00:06.713494
9313	6	1-2383	9308	33	25	2386	59	2012-08-02 05:00:06.713494
9308	6	1-2635	9308	25	93	2409	1	2012-08-02 05:00:06.713494
9315	6	1-1803	9314	86	61	510	22	2012-08-02 05:00:06.713494
9316	6	1-2469	9314	61	94	471	80	2012-08-02 05:00:06.713494
9317	6	1-2562	9314	94	19	1103	45	2012-08-02 05:00:06.713494
9318	6	1-2566	9314	19	33	386	67	2012-08-02 05:00:06.713494
9319	6	1-2383	9314	33	25	477	59	2012-08-02 05:00:06.713494
9314	6	1-2635	9314	25	86	486	1	2012-08-02 05:00:06.713494
9321	4	1-2121	9320	58	99	2749	32	2012-08-02 05:00:06.713494
9322	4	1-2023	9320	99	48	2828	93	2012-08-02 05:00:06.713494
9323	4	1-2027	9320	48	25	1591	59	2012-08-02 05:00:06.713494
9320	4	1-2635	9320	25	58	1718	1	2012-08-02 05:00:06.713494
9325	6	1-1359	9324	60	48	2296	59	2012-08-02 05:00:07.575402
9326	6	1-2552	9324	48	88	2721	53	2012-08-02 05:00:07.575402
9327	6	1-2624	9324	88	79	3140	83	2012-08-02 05:00:07.575402
9328	6	1-2604	9324	79	47	1100	67	2012-08-02 05:00:07.575402
9324	6	1-2637	9324	96	60	2219	3	2012-08-02 05:00:07.575402
9329	6	1-1298	9324	47	96	835	29	2012-08-02 05:00:07.575402
16031	5	1-4316	16029	57	6	4696	29	2012-08-02 05:13:31.8252
9331	6	1-1359	9330	60	25	367	59	2012-08-02 05:00:07.575402
9332	6	1-2635	9330	25	14	383	1	2012-08-02 05:00:07.575402
9333	6	1-1849	9330	14	45	588	69	2012-08-02 05:00:07.575402
9334	6	1-1434	9330	45	90	150	87	2012-08-02 05:00:07.575402
9335	6	1-1975	9330	90	96	133	29	2012-08-02 05:00:07.575402
9330	6	1-2637	9330	96	60	354	3	2012-08-02 05:00:07.575402
16032	5	1-3050	16029	6	37	1315	36	2012-08-02 05:13:31.8252
9337	7	1-2537	9336	14	32	67	28	2012-08-02 05:00:07.99533
9338	7	1-2487	9336	32	37	415	85	2012-08-02 05:00:07.99533
9339	7	1-2588	9336	37	82	1019	54	2012-08-02 05:00:07.99533
9340	7	1-2365	9336	82	4	77	8	2012-08-02 05:00:07.99533
9341	7	1-4	9336	4	96	65	7	2012-08-02 05:00:07.99533
9342	7	1-2593	9336	96	3	91	91	2012-08-02 05:00:07.99533
9336	7	1-2638	9336	3	14	76	47	2012-08-02 05:00:07.99533
9344	5	1-1277	9343	95	25	849	59	2012-08-02 05:00:11.784526
9345	5	1-2635	9343	25	58	711	1	2012-08-02 05:00:11.784526
9346	5	1-2121	9343	58	11	882	32	2012-08-02 05:00:11.784526
9347	5	1-2599	9343	11	45	1601	94	2012-08-02 05:00:11.784526
9343	5	1-2647	9343	45	95	811	30	2012-08-02 05:00:11.784526
9348	1	1-182	\N	9	9	1195	32	2012-08-02 05:00:11.784526
9350	6	1-2091	9349	46	81	2398	14	2012-08-02 05:00:12.215857
9351	6	1-2446	9349	81	82	660	74	2012-08-02 05:00:12.215857
9352	6	1-2563	9349	82	21	296	38	2012-08-02 05:00:12.215857
9353	6	1-2627	9349	21	51	911	21	2012-08-02 05:00:12.215857
9354	6	1-1457	9349	51	37	514	41	2012-08-02 05:00:12.215857
9349	6	1-2648	9349	37	46	1264	11	2012-08-02 05:00:12.215857
9356	5	1-1055	9355	2	70	1095	20	2012-08-02 05:00:12.215857
9357	5	1-2616	9355	70	74	431	26	2012-08-02 05:00:12.215857
9358	5	1-1492	9355	74	98	153	77	2012-08-02 05:00:12.215857
9359	5	1-2564	9355	98	37	427	41	2012-08-02 05:00:12.215857
9355	5	1-2648	9355	37	2	1102	11	2012-08-02 05:00:12.215857
9361	2	1-1320	9360	69	91	2154	32	2012-08-02 05:00:13.486109
9360	2	1-2651	9360	91	69	6977	60	2012-08-02 05:00:13.486109
9363	5	1-2593	9362	96	36	548	91	2012-08-02 05:00:17.264087
9364	5	1-743	9362	36	81	3967	14	2012-08-02 05:00:17.264087
9365	5	1-2446	9362	81	100	1103	74	2012-08-02 05:00:17.264087
9366	5	1-2608	9362	100	8	381	9	2012-08-02 05:00:17.264087
9362	5	1-2660	9362	8	96	489	7	2012-08-02 05:00:17.264087
9368	6	1-2593	9367	96	22	286	91	2012-08-02 05:00:17.264087
9369	6	1-2581	9367	22	89	1519	72	2012-08-02 05:00:17.264087
9370	6	1-1236	9367	89	81	1186	14	2012-08-02 05:00:17.264087
9371	6	1-2446	9367	81	100	379	74	2012-08-02 05:00:17.264087
9372	6	1-2608	9367	100	8	150	9	2012-08-02 05:00:17.264087
9367	6	1-2660	9367	8	96	222	7	2012-08-02 05:00:17.264087
9374	6	1-2593	9373	96	98	238	91	2012-08-02 05:00:17.264087
9375	6	1-1671	9373	98	6	1215	13	2012-08-02 05:00:17.264087
9376	6	1-1521	9373	6	81	627	14	2012-08-02 05:00:17.264087
9377	6	1-2446	9373	81	100	224	74	2012-08-02 05:00:17.264087
9378	6	1-2608	9373	100	8	100	9	2012-08-02 05:00:17.264087
9373	6	1-2660	9373	8	96	165	7	2012-08-02 05:00:17.264087
9380	4	1-2342	9379	88	66	2235	18	2012-08-02 05:00:17.695298
9381	4	1-1688	9379	66	13	2253	74	2012-08-02 05:00:17.695298
9382	4	1-2464	9379	13	70	1676	69	2012-08-02 05:00:17.695298
9379	4	1-2661	9379	70	88	2960	94	2012-08-02 05:00:17.695298
9384	3	1-275	9383	76	8	3496	9	2012-08-02 05:00:19.752018
9385	3	1-2660	9383	8	31	2914	7	2012-08-02 05:00:19.752018
9383	3	1-2666	9383	31	76	7615	8	2012-08-02 05:00:19.752018
9387	4	1-2407	9386	61	51	1432	65	2012-08-02 05:00:19.752018
9388	4	1-2253	9386	51	73	376	35	2012-08-02 05:00:19.752018
9389	4	1-809	9386	73	31	171	7	2012-08-02 05:00:19.752018
9386	4	1-2666	9386	31	61	810	8	2012-08-02 05:00:19.752018
9391	2	1-816	9390	30	5	1751	14	2012-08-02 05:00:21.033463
9390	2	1-2669	9390	5	30	3753	47	2012-08-02 05:00:21.033463
9393	3	1-1855	9392	15	91	3587	57	2012-08-02 05:00:21.033463
9394	3	1-2371	9392	91	5	2733	14	2012-08-02 05:00:21.033463
9392	3	1-2669	9392	5	15	5878	47	2012-08-02 05:00:21.033463
9396	6	1-2417	9395	93	49	1139	63	2012-08-02 05:00:21.442311
9397	6	1-357	9395	49	81	800	14	2012-08-02 05:00:21.442311
9398	6	1-2446	9395	81	100	254	74	2012-08-02 05:00:21.442311
9399	6	1-2608	9395	100	23	100	9	2012-08-02 05:00:21.442311
9400	6	1-2543	9395	23	8	1348	15	2012-08-02 05:00:21.442311
9395	6	1-2670	9395	8	93	823	1	2012-08-02 05:00:21.442311
9402	4	1-2659	9401	82	99	3794	55	2012-08-02 05:00:21.442311
9403	4	1-2426	9401	99	58	3326	16	2012-08-02 05:00:21.442311
9404	4	1-2594	9401	58	8	7111	15	2012-08-02 05:00:21.442311
9401	4	1-2670	9401	8	82	4402	1	2012-08-02 05:00:21.442311
9406	5	1-211	9405	17	85	203	86	2012-08-02 05:00:21.851319
9407	5	1-2652	9405	85	78	2375	78	2012-08-02 05:00:21.851319
9408	5	1-1974	9405	78	10	554	8	2012-08-02 05:00:21.851319
9409	5	1-2409	9405	10	8	1293	3	2012-08-02 05:00:21.851319
9405	5	1-2671	9405	8	17	715	26	2012-08-02 05:00:21.851319
9411	2	1-1722	9410	61	32	3353	72	2012-08-02 05:00:22.293314
9410	2	1-2672	9410	32	61	7133	38	2012-08-02 05:00:22.293314
9413	6	1-2562	9412	94	66	238	45	2012-08-02 05:00:23.121729
9414	6	1-1820	9412	66	12	178	95	2012-08-02 05:00:23.121729
9415	6	1-2366	9412	12	8	117	25	2012-08-02 05:00:23.121729
9416	6	1-2644	9412	8	63	118	71	2012-08-02 05:00:23.121729
9417	6	1-2617	9412	63	52	163	57	2012-08-02 05:00:23.121729
9412	6	1-2674	9412	52	94	115	80	2012-08-02 05:00:23.121729
9419	3	1-1797	9418	38	12	1107	60	2012-08-02 05:00:23.552833
9420	3	1-2639	9418	12	6	2608	10	2012-08-02 05:00:23.552833
9418	3	1-2675	9418	6	38	5620	82	2012-08-02 05:00:23.552833
9422	6	1-2581	9421	22	46	843	72	2012-08-02 05:00:24.812471
9423	6	1-2549	9421	46	82	1549	13	2012-08-02 05:00:24.812471
9424	6	1-2256	9421	82	60	279	3	2012-08-02 05:00:24.812471
9425	6	1-1359	9421	60	25	235	59	2012-08-02 05:00:24.812471
9426	6	1-2635	9421	25	94	199	1	2012-08-02 05:00:24.812471
9421	6	1-2678	9421	94	22	172	91	2012-08-02 05:00:24.812471
9428	6	1-1623	9427	70	90	244	87	2012-08-02 05:00:24.812471
9429	6	1-1975	9427	90	96	194	29	2012-08-02 05:00:24.812471
9430	6	1-2637	9427	96	60	462	3	2012-08-02 05:00:24.812471
9431	6	1-1359	9427	60	25	429	59	2012-08-02 05:00:24.812471
9432	6	1-2635	9427	25	94	401	1	2012-08-02 05:00:24.812471
9427	6	1-2678	9427	94	70	384	91	2012-08-02 05:00:24.812471
16033	5	1-4539	16029	37	8	3142	1	2012-08-02 05:13:31.8252
9434	4	1-2437	9433	47	29	361	31	2012-08-02 05:00:25.254448
9435	4	1-2246	9433	29	94	263	25	2012-08-02 05:00:25.254448
9436	4	1-2259	9433	94	66	103	22	2012-08-02 05:00:25.254448
9433	4	1-2679	9433	66	47	346	12	2012-08-02 05:00:25.254448
16029	5	1-4624	16029	8	49	5442	87	2012-08-02 05:13:31.8252
9438	6	1-1794	9437	67	57	1679	58	2012-08-02 05:00:25.254448
9439	6	1-1996	9437	57	70	1269	69	2012-08-02 05:00:25.254448
9440	6	1-2661	9437	70	45	2044	94	2012-08-02 05:00:25.254448
9441	6	1-2647	9437	45	20	1128	30	2012-08-02 05:00:25.254448
9442	6	1-2585	9437	20	66	1714	22	2012-08-02 05:00:25.254448
9437	6	1-2679	9437	66	67	6594	12	2012-08-02 05:00:25.254448
9444	3	1-2537	9443	14	32	515	28	2012-08-02 05:00:25.696413
9445	3	1-2487	9443	32	21	2000	85	2012-08-02 05:00:25.696413
9443	3	1-2680	9443	21	14	924	47	2012-08-02 05:00:25.696413
16035	5	1-1912	16034	47	98	5503	88	2012-08-02 05:13:32.245191
9447	4	1-2031	9446	29	89	758	21	2012-08-02 05:00:26.105222
9448	4	1-2596	9446	89	63	2193	5	2012-08-02 05:00:26.105222
9449	4	1-2601	9446	63	19	1209	62	2012-08-02 05:00:26.105222
9446	4	1-2681	9446	19	29	2935	53	2012-08-02 05:00:26.105222
9451	4	1-2083	9450	87	79	1139	43	2012-08-02 05:00:28.292494
9452	4	1-2654	9450	79	45	570	79	2012-08-02 05:00:28.292494
9453	4	1-2534	9450	45	11	414	36	2012-08-02 05:00:28.292494
9450	4	1-2686	9450	11	87	4645	48	2012-08-02 05:00:28.292494
9455	5	1-2243	9454	83	81	1751	79	2012-08-02 05:00:30.369318
9456	5	1-1744	9454	81	38	1419	4	2012-08-02 05:00:30.369318
9457	5	1-2521	9454	38	64	2774	100	2012-08-02 05:00:30.369318
9458	5	1-2605	9454	64	25	2597	62	2012-08-02 05:00:30.369318
9454	5	1-2691	9454	25	83	1589	99	2012-08-02 05:00:30.369318
9460	5	1-2667	9459	98	21	461	65	2012-08-02 05:00:30.369318
9461	5	1-2213	9459	21	38	484	4	2012-08-02 05:00:30.369318
9462	5	1-2521	9459	38	64	1228	100	2012-08-02 05:00:30.369318
9463	5	1-2605	9459	64	25	1492	62	2012-08-02 05:00:30.369318
9459	5	1-2691	9459	25	98	1185	99	2012-08-02 05:00:30.369318
9465	5	1-1747	9464	31	72	80	92	2012-08-02 05:00:33.286062
9466	5	1-2626	9464	72	70	69	69	2012-08-02 05:00:33.286062
9467	5	1-2661	9464	70	88	132	94	2012-08-02 05:00:33.286062
9468	5	1-2342	9464	88	88	109	18	2012-08-02 05:00:33.286062
9464	5	1-2698	9464	88	31	167	32	2012-08-02 05:00:33.286062
9470	5	1-2644	9469	8	63	312	71	2012-08-02 05:00:33.727893
9471	5	1-2617	9469	63	52	466	57	2012-08-02 05:00:33.727893
9472	5	1-2674	9469	52	59	356	80	2012-08-02 05:00:33.727893
9473	5	1-2214	9469	59	80	175	63	2012-08-02 05:00:33.727893
9469	5	1-2699	9469	80	8	287	25	2012-08-02 05:00:33.727893
9475	4	1-1566	9474	78	76	432	36	2012-08-02 05:00:34.147947
9476	4	1-2427	9474	76	64	956	100	2012-08-02 05:00:34.147947
9477	4	1-2605	9474	64	11	878	62	2012-08-02 05:00:34.147947
9474	4	1-2700	9474	11	78	795	77	2012-08-02 05:00:34.147947
9479	3	1-2564	9478	98	76	340	41	2012-08-02 05:00:34.147947
9480	3	1-142	9478	76	11	97	62	2012-08-02 05:00:34.147947
9478	3	1-2700	9478	11	98	105	77	2012-08-02 05:00:34.147947
9482	5	1-2608	9481	100	83	99	9	2012-08-02 05:00:34.567688
9483	5	1-1921	9481	83	38	106	4	2012-08-02 05:00:34.567688
9484	5	1-2521	9481	38	64	223	100	2012-08-02 05:00:34.567688
9485	5	1-2605	9481	64	82	225	62	2012-08-02 05:00:34.567688
9481	5	1-2701	9481	82	100	260	74	2012-08-02 05:00:34.567688
9487	4	1-2608	9486	100	23	1505	9	2012-08-02 05:00:34.567688
9488	4	1-2636	9486	23	63	3848	5	2012-08-02 05:00:34.567688
9489	4	1-2601	9486	63	82	2594	62	2012-08-02 05:00:34.567688
9486	4	1-2701	9486	82	100	3436	74	2012-08-02 05:00:34.567688
9491	5	1-2563	9490	82	86	1488	38	2012-08-02 05:00:34.567688
9492	5	1-1821	9490	86	22	1338	70	2012-08-02 05:00:34.567688
9493	5	1-2591	9490	22	98	2502	3	2012-08-02 05:00:34.567688
9494	5	1-1741	9490	98	82	1900	62	2012-08-02 05:00:34.567688
9490	5	1-2701	9490	82	82	2550	74	2012-08-02 05:00:34.567688
9496	3	1-2643	9495	70	61	5256	55	2012-08-02 05:00:35.407405
9497	3	1-1385	9495	61	50	3970	40	2012-08-02 05:00:35.407405
9495	3	1-2703	9495	50	70	5599	46	2012-08-02 05:00:35.407405
9499	5	1-2579	9498	26	56	45	60	2012-08-02 05:00:35.407405
9500	5	1-2321	9498	56	55	76	48	2012-08-02 05:00:35.407405
9501	5	1-2396	9498	55	98	110	93	2012-08-02 05:00:35.407405
9502	5	1-2565	9498	98	50	73	40	2012-08-02 05:00:35.407405
9498	5	1-2703	9498	50	26	103	46	2012-08-02 05:00:35.407405
9504	4	1-2520	9503	10	61	895	95	2012-08-02 05:00:35.407405
9505	4	1-1084	9503	61	69	494	12	2012-08-02 05:00:35.407405
9506	4	1-1315	9503	69	50	549	40	2012-08-02 05:00:35.407405
9503	4	1-2703	9503	50	10	841	46	2012-08-02 05:00:35.407405
9508	6	1-2473	9507	37	11	1387	62	2012-08-02 05:00:35.827545
9509	6	1-2700	9507	11	29	1456	77	2012-08-02 05:00:35.827545
9510	6	1-2440	9507	29	98	1007	97	2012-08-02 05:00:35.827545
9511	6	1-2502	9507	98	77	854	44	2012-08-02 05:00:35.827545
9512	6	1-2175	9507	77	47	714	20	2012-08-02 05:00:35.827545
9507	6	1-2704	9507	47	37	1227	6	2012-08-02 05:00:35.827545
9514	6	1-2493	9513	92	81	381	79	2012-08-02 05:00:35.827545
9515	6	1-1744	9513	81	38	374	4	2012-08-02 05:00:35.827545
9516	6	1-2521	9513	38	39	887	100	2012-08-02 05:00:35.827545
9517	6	1-2589	9513	39	2	268	11	2012-08-02 05:00:35.827545
9518	6	1-1055	9513	2	47	320	20	2012-08-02 05:00:35.827545
9513	6	1-2704	9513	47	92	586	6	2012-08-02 05:00:35.827545
9520	4	1-2661	9519	70	88	1134	94	2012-08-02 05:00:36.258436
9521	4	1-2342	9519	88	16	542	18	2012-08-02 05:00:36.258436
9522	4	1-2443	9519	16	33	302	99	2012-08-02 05:00:36.258436
9519	4	1-2705	9519	33	70	1015	69	2012-08-02 05:00:36.258436
9524	6	1-2506	9523	11	30	159	64	2012-08-02 05:00:36.258436
9525	6	1-2630	9523	30	49	236	82	2012-08-02 05:00:36.258436
9526	6	1-707	9523	49	88	153	94	2012-08-02 05:00:36.258436
9527	6	1-2342	9523	88	16	114	18	2012-08-02 05:00:36.258436
9528	6	1-2443	9523	16	33	99	99	2012-08-02 05:00:36.258436
9523	6	1-2705	9523	33	11	516	69	2012-08-02 05:00:36.258436
9530	5	1-1930	9529	22	24	80	19	2012-08-02 05:00:36.258436
9531	5	1-2368	9529	24	87	153	22	2012-08-02 05:00:36.258436
9532	5	1-2587	9529	87	58	114	56	2012-08-02 05:00:36.258436
9533	5	1-1593	9529	58	33	50	99	2012-08-02 05:00:36.258436
9529	5	1-2705	9529	33	22	263	69	2012-08-02 05:00:36.258436
9535	6	1-2506	9534	11	30	96	64	2012-08-02 05:00:36.258436
9536	6	1-2630	9534	30	68	156	82	2012-08-02 05:00:36.258436
9537	6	1-1884	9534	68	87	128	22	2012-08-02 05:00:36.258436
9538	6	1-2587	9534	87	58	104	56	2012-08-02 05:00:36.258436
9539	6	1-1593	9534	58	33	50	99	2012-08-02 05:00:36.258436
9534	6	1-2705	9534	33	11	287	69	2012-08-02 05:00:36.258436
16036	5	1-4535	16034	98	96	2721	97	2012-08-02 05:13:32.245191
9541	6	1-2678	9540	94	32	496	91	2012-08-02 05:00:36.711722
9542	6	1-1435	9540	32	88	707	26	2012-08-02 05:00:36.711722
9543	6	1-2060	9540	88	30	307	64	2012-08-02 05:00:36.711722
9544	6	1-2630	9540	30	44	496	82	2012-08-02 05:00:36.711722
9545	6	1-2425	9540	44	57	371	89	2012-08-02 05:00:36.711722
9540	6	1-2706	9540	57	94	453	1	2012-08-02 05:00:36.711722
16037	5	1-4356	16034	96	60	820	13	2012-08-02 05:13:32.245191
9547	6	1-2626	9546	72	53	3000	69	2012-08-02 05:00:37.153585
9548	6	1-2451	9546	53	47	2079	12	2012-08-02 05:00:37.153585
9549	6	1-2437	9546	47	57	1740	31	2012-08-02 05:00:37.153585
9550	6	1-2677	9546	57	85	457	28	2012-08-02 05:00:37.153585
9551	6	1-1719	9546	85	49	421	4	2012-08-02 05:00:37.153585
9546	6	1-2707	9546	49	72	5933	92	2012-08-02 05:00:37.153585
9553	5	1-1744	9552	81	41	543	4	2012-08-02 05:00:38.004322
9554	5	1-2418	9552	41	12	531	95	2012-08-02 05:00:38.004322
9555	5	1-2366	9552	12	8	393	25	2012-08-02 05:00:38.004322
9556	5	1-2644	9552	8	72	443	71	2012-08-02 05:00:38.004322
9552	5	1-2709	9552	72	81	547	79	2012-08-02 05:00:38.004322
9558	6	1-2685	9557	43	7	2728	57	2012-08-02 05:00:38.866315
9559	6	1-2503	9557	7	55	1327	46	2012-08-02 05:00:38.866315
9560	6	1-1881	9557	55	15	1620	27	2012-08-02 05:00:38.866315
9561	6	1-2442	9557	15	72	908	71	2012-08-02 05:00:38.866315
9562	6	1-2709	9557	72	42	1157	79	2012-08-02 05:00:38.866315
9557	6	1-2711	9557	42	43	1528	37	2012-08-02 05:00:38.866315
9564	3	1-2612	9563	7	8	4370	89	2012-08-02 05:00:40.147623
9565	3	1-2382	9563	8	64	3072	12	2012-08-02 05:00:40.147623
9563	3	1-2714	9563	64	7	9787	44	2012-08-02 05:00:40.147623
9567	6	1-2520	9566	10	12	2385	95	2012-08-02 05:00:40.567448
9568	6	1-2366	9566	12	8	1662	25	2012-08-02 05:00:40.567448
9569	6	1-2644	9566	8	72	1768	71	2012-08-02 05:00:40.567448
9570	6	1-2709	9566	72	42	2059	79	2012-08-02 05:00:40.567448
9571	6	1-2711	9566	42	59	2487	37	2012-08-02 05:00:40.567448
9566	6	1-2715	9566	59	10	2540	46	2012-08-02 05:00:40.567448
9573	5	1-1719	9572	85	38	398	4	2012-08-02 05:00:43.130348
9574	5	1-2521	9572	38	74	595	100	2012-08-02 05:00:43.130348
9575	5	1-2386	9572	74	39	667	17	2012-08-02 05:00:43.130348
9576	5	1-2201	9572	39	56	456	42	2012-08-02 05:00:43.130348
9572	5	1-2721	9572	56	85	437	28	2012-08-02 05:00:43.130348
9578	4	1-2478	9577	74	98	2259	97	2012-08-02 05:00:43.130348
9579	4	1-2502	9577	98	32	1526	44	2012-08-02 05:00:43.130348
9580	4	1-1510	9577	32	56	1463	42	2012-08-02 05:00:43.130348
9577	4	1-2721	9577	56	74	1663	28	2012-08-02 05:00:43.130348
9582	5	1-1889	9581	63	18	179	29	2012-08-02 05:00:43.130348
9583	5	1-2550	9581	18	23	1026	15	2012-08-02 05:00:43.130348
9584	5	1-183	9581	23	39	305	17	2012-08-02 05:00:43.130348
9585	5	1-2201	9581	39	56	275	42	2012-08-02 05:00:43.130348
9581	5	1-2721	9581	56	63	348	28	2012-08-02 05:00:43.130348
9587	5	1-1889	9586	63	96	524	29	2012-08-02 05:00:43.130348
9588	5	1-2637	9586	96	98	1139	3	2012-08-02 05:00:43.130348
9589	5	1-1741	9586	98	2	762	62	2012-08-02 05:00:43.130348
9590	5	1-2172	9586	2	56	766	42	2012-08-02 05:00:43.130348
9586	5	1-2721	9586	56	63	993	28	2012-08-02 05:00:43.130348
9592	5	1-2559	9591	62	24	557	19	2012-08-02 05:00:43.130348
9593	5	1-2368	9591	24	35	1002	22	2012-08-02 05:00:43.130348
9594	5	1-2689	9591	35	2	1222	62	2012-08-02 05:00:43.130348
9595	5	1-2172	9591	2	56	1339	42	2012-08-02 05:00:43.130348
9591	5	1-2721	9591	56	62	1892	28	2012-08-02 05:00:43.130348
9597	6	1-2711	9596	42	43	178	37	2012-08-02 05:00:43.561703
9598	6	1-2685	9596	43	33	309	57	2012-08-02 05:00:43.561703
9599	6	1-2137	9596	33	99	150	56	2012-08-02 05:00:43.561703
9600	6	1-2248	9596	99	75	130	49	2012-08-02 05:00:43.561703
9601	6	1-1176	9596	75	51	172	74	2012-08-02 05:00:43.561703
9596	6	1-2722	9596	51	42	138	79	2012-08-02 05:00:43.561703
9603	5	1-2631	9602	74	90	3480	87	2012-08-02 05:00:45.251866
9604	5	1-1975	9602	90	96	1780	29	2012-08-02 05:00:45.251866
9605	5	1-2637	9602	96	62	2738	3	2012-08-02 05:00:45.251866
9606	5	1-1636	9602	62	83	1491	55	2012-08-02 05:00:45.251866
9602	5	1-2726	9602	83	74	7271	37	2012-08-02 05:00:45.251866
9607	1	1-1512	\N	98	98	2340	29	2012-08-02 05:00:45.251866
9609	6	1-2713	9608	26	94	1867	21	2012-08-02 05:00:45.705208
9610	6	1-2228	9608	94	61	1124	45	2012-08-02 05:00:45.705208
9611	6	1-985	9608	61	11	453	62	2012-08-02 05:00:45.705208
9612	6	1-2700	9608	11	46	423	77	2012-08-02 05:00:45.705208
9613	6	1-2021	9608	46	95	209	85	2012-08-02 05:00:45.705208
9608	6	1-2727	9608	95	26	1375	69	2012-08-02 05:00:45.705208
9615	5	1-1100	9614	22	35	104	22	2012-08-02 05:00:45.705208
9616	5	1-2689	9614	35	11	120	62	2012-08-02 05:00:45.705208
9617	5	1-2700	9614	11	46	119	77	2012-08-02 05:00:45.705208
9618	5	1-2021	9614	46	95	62	85	2012-08-02 05:00:45.705208
9614	5	1-2727	9614	95	22	431	69	2012-08-02 05:00:45.705208
9620	5	1-1100	9619	22	35	630	22	2012-08-02 05:00:45.705208
9621	5	1-2689	9619	35	11	795	62	2012-08-02 05:00:45.705208
9622	5	1-2700	9619	11	83	855	77	2012-08-02 05:00:45.705208
9623	5	1-1481	9619	83	95	317	85	2012-08-02 05:00:45.705208
9619	5	1-2727	9619	95	22	2398	69	2012-08-02 05:00:45.705208
9625	6	1-2429	9624	90	86	63	38	2012-08-02 05:00:47.36408
9626	6	1-1821	9624	86	6	61	70	2012-08-02 05:00:47.36408
9627	6	1-1842	9624	6	74	51	37	2012-08-02 05:00:47.36408
9628	6	1-2631	9624	74	90	42	87	2012-08-02 05:00:47.36408
9629	6	1-1975	9624	90	81	38	29	2012-08-02 05:00:47.36408
9624	6	1-2731	9624	81	90	131	76	2012-08-02 05:00:47.36408
9631	5	1-1665	9630	48	22	859	70	2012-08-02 05:00:49.496231
9632	5	1-2591	9630	22	60	1706	3	2012-08-02 05:00:49.496231
9633	5	1-1359	9630	60	2	1738	59	2012-08-02 05:00:49.496231
9634	5	1-2613	9630	2	8	2536	98	2012-08-02 05:00:49.496231
9630	5	1-2736	9630	8	48	1408	86	2012-08-02 05:00:49.496231
9636	6	1-2482	9635	63	16	382	60	2012-08-02 05:00:49.496231
9637	6	1-2347	9635	16	96	278	29	2012-08-02 05:00:49.496231
9638	6	1-2637	9635	96	60	762	3	2012-08-02 05:00:49.496231
9639	6	1-1359	9635	60	2	813	59	2012-08-02 05:00:49.496231
9640	6	1-2613	9635	2	8	1243	98	2012-08-02 05:00:49.496231
9635	6	1-2736	9635	8	63	723	86	2012-08-02 05:00:49.496231
16038	5	1-4561	16034	60	3	1466	45	2012-08-02 05:13:32.245191
9642	4	1-15	9641	15	38	715	4	2012-08-02 05:00:50.766846
9643	4	1-2521	9641	38	60	1522	100	2012-08-02 05:00:50.766846
9644	4	1-2381	9641	60	44	1326	23	2012-08-02 05:00:50.766846
9641	4	1-2739	9641	44	15	2871	24	2012-08-02 05:00:50.766846
16034	5	1-4625	16034	3	47	8058	55	2012-08-02 05:13:32.245191
9646	6	1-2453	9645	64	88	1814	61	2012-08-02 05:00:50.766846
9647	6	1-194	9645	88	21	596	65	2012-08-02 05:00:50.766846
9648	6	1-2213	9645	21	38	607	4	2012-08-02 05:00:50.766846
9649	6	1-2521	9645	38	60	1494	100	2012-08-02 05:00:50.766846
9650	6	1-2381	9645	60	44	1505	23	2012-08-02 05:00:50.766846
9645	6	1-2739	9645	44	64	3768	24	2012-08-02 05:00:50.766846
9652	6	1-2564	9651	98	37	2687	41	2012-08-02 05:00:52.037763
9653	6	1-2642	9651	37	34	2670	68	2012-08-02 05:00:52.037763
9654	6	1-1630	9651	34	6	677	70	2012-08-02 05:00:52.037763
9655	6	1-1842	9651	6	43	526	37	2012-08-02 05:00:52.037763
9656	6	1-2685	9651	43	13	869	57	2012-08-02 05:00:52.037763
9651	6	1-2742	9651	13	98	830	77	2012-08-02 05:00:52.037763
9658	5	1-580	9657	86	80	275	74	2012-08-02 05:00:52.037763
9659	5	1-2732	9657	80	8	377	25	2012-08-02 05:00:52.037763
9660	5	1-2644	9657	8	63	453	71	2012-08-02 05:00:52.037763
9661	5	1-2617	9657	63	13	749	57	2012-08-02 05:00:52.037763
9657	5	1-2742	9657	13	86	799	77	2012-08-02 05:00:52.037763
9663	6	1-932	9662	100	57	74	31	2012-08-02 05:00:54.534352
9664	6	1-2677	9662	57	62	32	28	2012-08-02 05:00:54.534352
9665	6	1-2559	9662	62	24	10	19	2012-08-02 05:00:54.534352
9666	6	1-2368	9662	24	87	21	22	2012-08-02 05:00:54.534352
9667	6	1-2587	9662	87	96	16	56	2012-08-02 05:00:54.534352
9662	6	1-2748	9662	96	100	550	15	2012-08-02 05:00:54.534352
9669	6	1-932	9668	100	57	258	31	2012-08-02 05:00:54.534352
9670	6	1-2677	9668	57	62	115	28	2012-08-02 05:00:54.534352
9671	6	1-2559	9668	62	17	40	19	2012-08-02 05:00:54.534352
9672	6	1-2600	9668	17	87	65	22	2012-08-02 05:00:54.534352
9673	6	1-2587	9668	87	96	53	56	2012-08-02 05:00:54.534352
9668	6	1-2748	9668	96	100	1839	15	2012-08-02 05:00:54.534352
9675	6	1-1408	9674	27	11	236	69	2012-08-02 05:00:54.534352
9676	6	1-2506	9674	11	30	80	64	2012-08-02 05:00:54.534352
9677	6	1-2630	9674	30	68	132	82	2012-08-02 05:00:54.534352
9678	6	1-1884	9674	68	87	110	22	2012-08-02 05:00:54.534352
9679	6	1-2587	9674	87	96	91	56	2012-08-02 05:00:54.534352
9674	6	1-2748	9674	96	27	3178	15	2012-08-02 05:00:54.534352
9681	5	1-2521	9680	38	60	782	100	2012-08-02 05:00:54.965229
9682	5	1-2381	9680	60	93	669	23	2012-08-02 05:00:54.965229
9683	5	1-803	9680	93	68	144	89	2012-08-02 05:00:54.965229
9684	5	1-2014	9680	68	3	134	13	2012-08-02 05:00:54.965229
9680	5	1-2749	9680	3	38	374	4	2012-08-02 05:00:54.965229
9686	5	1-2609	9685	55	69	91	90	2012-08-02 05:00:54.965229
9687	5	1-2611	9685	69	48	62	3	2012-08-02 05:00:54.965229
9688	5	1-2338	9685	48	92	50	24	2012-08-02 05:00:54.965229
9689	5	1-1323	9685	92	3	30	13	2012-08-02 05:00:54.965229
9685	5	1-2749	9685	3	55	93	4	2012-08-02 05:00:54.965229
9691	5	1-2609	9690	55	69	1420	90	2012-08-02 05:00:54.965229
9692	5	1-2611	9690	69	48	1013	3	2012-08-02 05:00:54.965229
9693	5	1-2338	9690	48	75	862	24	2012-08-02 05:00:54.965229
9694	5	1-1295	9690	75	3	420	13	2012-08-02 05:00:54.965229
9690	5	1-2749	9690	3	55	1377	4	2012-08-02 05:00:54.965229
9696	4	1-2248	9695	99	75	2623	49	2012-08-02 05:00:55.81654
9697	4	1-1176	9695	75	82	1774	74	2012-08-02 05:00:55.81654
9698	4	1-2563	9695	82	11	520	38	2012-08-02 05:00:55.81654
9695	4	1-2751	9695	11	99	5935	56	2012-08-02 05:00:55.81654
9700	6	1-1348	9699	85	63	417	46	2012-08-02 05:00:55.81654
9701	6	1-2377	9699	63	90	152	87	2012-08-02 05:00:55.81654
9702	6	1-1975	9699	90	81	109	29	2012-08-02 05:00:55.81654
9703	6	1-2731	9699	81	90	231	76	2012-08-02 05:00:55.81654
9704	6	1-2429	9699	90	11	88	38	2012-08-02 05:00:55.81654
9699	6	1-2751	9699	11	85	1247	56	2012-08-02 05:00:55.81654
9706	5	1-1328	9705	17	58	802	22	2012-08-02 05:00:55.81654
9707	5	1-2718	9705	58	66	1944	80	2012-08-02 05:00:55.81654
9708	5	1-1922	9705	66	82	185	74	2012-08-02 05:00:55.81654
9709	5	1-2563	9705	82	11	77	38	2012-08-02 05:00:55.81654
9705	5	1-2751	9705	11	17	1250	56	2012-08-02 05:00:55.81654
9711	5	1-1348	9710	85	63	463	46	2012-08-02 05:00:55.81654
9712	5	1-2377	9710	63	75	205	87	2012-08-02 05:00:55.81654
9713	5	1-2598	9710	75	90	187	76	2012-08-02 05:00:55.81654
9714	5	1-2429	9710	90	11	66	38	2012-08-02 05:00:55.81654
9710	5	1-2751	9710	11	85	1142	56	2012-08-02 05:00:55.81654
9715	1	1-1975	\N	90	90	3580	29	2012-08-02 05:00:55.81654
9717	4	1-2604	9716	79	96	900	67	2012-08-02 05:00:56.68911
9718	4	1-1716	9716	96	2	1130	54	2012-08-02 05:00:56.68911
9719	4	1-2345	9716	2	28	4858	17	2012-08-02 05:00:56.68911
9716	4	1-2753	9716	28	79	3449	83	2012-08-02 05:00:56.68911
9721	5	1-249	9720	82	22	67	84	2012-08-02 05:00:56.68911
9722	5	1-1958	9720	22	48	51	86	2012-08-02 05:00:56.68911
9723	5	1-1665	9720	48	8	25	70	2012-08-02 05:00:56.68911
9724	5	1-2458	9720	8	28	440	17	2012-08-02 05:00:56.68911
9720	5	1-2753	9720	28	82	339	83	2012-08-02 05:00:56.68911
9726	5	1-2113	9725	48	97	11	12	2012-08-02 05:00:56.68911
9727	5	1-2269	9725	97	76	5	36	2012-08-02 05:00:56.68911
9728	5	1-2427	9725	76	74	13	100	2012-08-02 05:00:56.68911
9729	5	1-2386	9725	74	28	22	17	2012-08-02 05:00:56.68911
9725	5	1-2753	9725	28	48	19	83	2012-08-02 05:00:56.68911
9731	5	1-2682	9730	16	4	1189	66	2012-08-02 05:00:57.959709
9732	5	1-2105	9730	4	63	762	28	2012-08-02 05:00:57.959709
9733	5	1-1889	9730	63	96	455	29	2012-08-02 05:00:57.959709
9734	5	1-2637	9730	96	40	1117	3	2012-08-02 05:00:57.959709
9730	5	1-2756	9730	40	16	2010	18	2012-08-02 05:00:57.959709
9736	5	1-2682	9735	16	25	214	66	2012-08-02 05:00:57.959709
9737	5	1-2220	9735	25	17	117	75	2012-08-02 05:00:57.959709
9738	5	1-1854	9735	17	10	68	8	2012-08-02 05:00:57.959709
9739	5	1-2409	9735	10	40	160	3	2012-08-02 05:00:57.959709
9735	5	1-2756	9735	40	16	323	18	2012-08-02 05:00:57.959709
9741	5	1-2244	9740	64	78	1061	73	2012-08-02 05:00:58.390777
9742	5	1-1460	9740	78	97	312	12	2012-08-02 05:00:58.390777
9740	5	1-2757	9740	42	64	1076	91	2012-08-02 05:00:58.390777
9743	5	1-2269	9740	97	6	113	36	2012-08-02 05:00:58.390777
9744	5	1-2692	9740	6	42	3356	10	2012-08-02 05:00:58.390777
9746	4	1-2206	9745	84	26	3678	46	2012-08-02 05:00:58.390777
9747	4	1-2579	9745	26	12	1502	60	2012-08-02 05:00:58.390777
9748	4	1-2639	9745	12	42	7250	10	2012-08-02 05:00:58.390777
9745	4	1-2757	9745	42	84	2700	91	2012-08-02 05:00:58.390777
16040	6	1-4522	16039	85	20	636	91	2012-08-02 05:13:33.449333
9750	5	1-2600	9749	17	35	274	22	2012-08-02 05:01:02.544567
9751	5	1-2689	9749	35	63	343	62	2012-08-02 05:01:02.544567
9752	5	1-2468	9749	63	12	144	68	2012-08-02 05:01:02.544567
9753	5	1-1927	9749	12	7	134	31	2012-08-02 05:01:02.544567
9749	5	1-2767	9749	7	17	191	19	2012-08-02 05:01:02.544567
16041	6	1-3675	16039	20	10	327	61	2012-08-02 05:13:33.449333
9755	4	1-2600	9754	17	25	3484	22	2012-08-02 05:01:02.544567
9756	4	1-1787	9754	25	47	1128	12	2012-08-02 05:01:02.544567
9757	4	1-2437	9754	47	7	1503	31	2012-08-02 05:01:02.544567
9754	4	1-2767	9754	7	17	2282	19	2012-08-02 05:01:02.544567
16042	6	1-3627	16039	10	35	247	28	2012-08-02 05:13:33.449333
9759	5	1-1114	9758	20	39	1619	94	2012-08-02 05:01:02.544567
9760	5	1-2717	9758	39	53	1475	69	2012-08-02 05:01:02.544567
9761	5	1-2451	9758	53	47	1675	12	2012-08-02 05:01:02.544567
9762	5	1-2437	9758	47	7	2297	31	2012-08-02 05:01:02.544567
9758	5	1-2767	9758	7	20	3590	19	2012-08-02 05:01:02.544567
9764	5	1-2600	9763	17	87	4790	22	2012-08-02 05:01:02.544567
9765	5	1-2587	9763	87	58	3801	56	2012-08-02 05:01:02.544567
9766	5	1-1593	9763	58	88	1793	99	2012-08-02 05:01:02.544567
9767	5	1-2628	9763	88	7	1921	31	2012-08-02 05:01:02.544567
9763	5	1-2767	9763	7	17	3025	19	2012-08-02 05:01:02.544567
9769	4	1-2600	9768	17	65	469	22	2012-08-02 05:01:02.544567
9770	4	1-2522	9768	65	23	105	35	2012-08-02 05:01:02.544567
9771	4	1-2163	9768	23	7	177	31	2012-08-02 05:01:02.544567
9768	4	1-2767	9768	7	17	287	19	2012-08-02 05:01:02.544567
9773	5	1-2505	9772	85	37	314	6	2012-08-02 05:01:02.964724
9774	5	1-2473	9772	37	19	329	62	2012-08-02 05:01:02.964724
9775	5	1-2681	9772	19	97	834	53	2012-08-02 05:01:02.964724
9776	5	1-2211	9772	97	62	276	40	2012-08-02 05:01:02.964724
9772	5	1-2768	9772	62	85	359	35	2012-08-02 05:01:02.964724
9778	3	1-123	9777	68	98	4611	93	2012-08-02 05:01:02.964724
9779	3	1-2565	9777	98	62	2756	40	2012-08-02 05:01:02.964724
9777	3	1-2768	9777	62	68	3700	35	2012-08-02 05:01:02.964724
9781	2	1-155	9780	80	62	726	40	2012-08-02 05:01:02.964724
9780	2	1-2768	9780	62	80	976	35	2012-08-02 05:01:02.964724
9783	4	1-2078	9782	88	35	4685	18	2012-08-02 05:01:02.964724
9784	4	1-2320	9782	35	74	2279	58	2012-08-02 05:01:02.964724
9785	4	1-327	9782	74	62	2501	40	2012-08-02 05:01:02.964724
9782	4	1-2768	9782	62	88	3447	35	2012-08-02 05:01:02.964724
9787	4	1-2177	9786	69	82	644	1	2012-08-02 05:01:02.964724
9788	4	1-2659	9786	82	61	568	55	2012-08-02 05:01:02.964724
9789	4	1-1385	9786	61	62	411	40	2012-08-02 05:01:02.964724
9786	4	1-2768	9786	62	69	585	35	2012-08-02 05:01:02.964724
9791	6	1-2442	9790	15	72	1383	71	2012-08-02 05:01:03.804507
9792	6	1-2709	9790	72	45	1554	79	2012-08-02 05:01:03.804507
9793	6	1-2534	9790	45	39	1109	36	2012-08-02 05:01:03.804507
9794	6	1-2500	9790	39	93	1351	58	2012-08-02 05:01:03.804507
9795	6	1-1748	9790	93	43	1442	50	2012-08-02 05:01:03.804507
9790	6	1-2770	9790	43	15	2795	27	2012-08-02 05:01:03.804507
9797	4	1-2442	9796	15	63	753	71	2012-08-02 05:01:03.804507
9798	4	1-2617	9796	63	76	1073	57	2012-08-02 05:01:03.804507
9799	4	1-2388	9796	76	43	772	50	2012-08-02 05:01:03.804507
9796	4	1-2770	9796	43	15	1509	27	2012-08-02 05:01:03.804507
9801	6	1-2365	9800	82	10	81	8	2012-08-02 05:01:04.224383
9802	6	1-2409	9800	10	40	111	3	2012-08-02 05:01:04.224383
9803	6	1-2756	9800	40	16	131	18	2012-08-02 05:01:04.224383
9804	6	1-2682	9800	16	43	51	66	2012-08-02 05:01:04.224383
9805	6	1-2618	9800	43	42	1646	51	2012-08-02 05:01:04.224383
9800	6	1-2771	9800	42	82	1708	54	2012-08-02 05:01:04.224383
9807	4	1-2765	9806	46	77	6662	85	2012-08-02 05:01:04.655402
9808	4	1-2696	9806	77	39	3605	82	2012-08-02 05:01:04.655402
9809	4	1-574	9806	39	75	2397	29	2012-08-02 05:01:04.655402
9806	4	1-2772	9806	75	46	4298	49	2012-08-02 05:01:04.655402
9811	5	1-2765	9810	46	21	804	85	2012-08-02 05:01:04.655402
9812	5	1-2680	9810	21	14	555	47	2012-08-02 05:01:04.655402
9813	5	1-2537	9810	14	63	462	28	2012-08-02 05:01:04.655402
9814	5	1-1889	9810	63	75	270	29	2012-08-02 05:01:04.655402
9810	5	1-2772	9810	75	46	501	49	2012-08-02 05:01:04.655402
9816	5	1-1487	9815	8	50	1770	87	2012-08-02 05:01:04.655402
9817	5	1-2480	9815	50	30	1122	64	2012-08-02 05:01:04.655402
9818	5	1-2630	9815	30	39	1758	82	2012-08-02 05:01:04.655402
9819	5	1-574	9815	39	75	1334	29	2012-08-02 05:01:04.655402
9815	5	1-2772	9815	75	8	2729	49	2012-08-02 05:01:04.655402
9821	6	1-1176	9820	75	51	647	74	2012-08-02 05:01:04.655402
9822	6	1-2722	9820	51	40	518	79	2012-08-02 05:01:04.655402
9823	6	1-2331	9820	40	13	412	33	2012-08-02 05:01:04.655402
9824	6	1-2758	9820	13	47	306	67	2012-08-02 05:01:04.655402
9825	6	1-1298	9820	47	75	225	29	2012-08-02 05:01:04.655402
9820	6	1-2772	9820	75	75	469	49	2012-08-02 05:01:04.655402
9827	4	1-2699	9826	80	8	2456	25	2012-08-02 05:01:05.518544
9828	4	1-2644	9826	8	72	2876	71	2012-08-02 05:01:05.518544
9829	4	1-2709	9826	72	16	3688	79	2012-08-02 05:01:05.518544
9826	4	1-2774	9826	16	80	1390	63	2012-08-02 05:01:05.518544
9831	4	1-1348	9830	85	91	1069	46	2012-08-02 05:01:06.347142
9832	4	1-1322	9830	91	25	684	33	2012-08-02 05:01:06.347142
9833	4	1-1759	9830	25	18	1992	5	2012-08-02 05:01:06.347142
9830	4	1-2776	9830	18	85	2156	56	2012-08-02 05:01:06.347142
9835	3	1-1366	9834	59	68	254	96	2012-08-02 05:01:06.347142
9836	3	1-880	9834	68	18	404	5	2012-08-02 05:01:06.347142
9834	3	1-2776	9834	18	59	447	56	2012-08-02 05:01:06.347142
9838	4	1-2344	9837	58	40	422	79	2012-08-02 05:01:06.347142
9839	4	1-2331	9837	40	25	326	33	2012-08-02 05:01:06.347142
9840	4	1-1759	9837	25	18	980	5	2012-08-02 05:01:06.347142
9837	4	1-2776	9837	18	58	1094	56	2012-08-02 05:01:06.347142
9842	4	1-2523	9841	58	20	2747	43	2012-08-02 05:01:06.347142
9843	4	1-2166	9841	20	19	1402	70	2012-08-02 05:01:06.347142
9844	4	1-2274	9841	19	18	3308	5	2012-08-02 05:01:06.347142
9841	4	1-2776	9841	18	58	4069	56	2012-08-02 05:01:06.347142
16043	6	1-4588	16039	35	4	349	92	2012-08-02 05:13:33.449333
9846	5	1-1998	9845	16	17	2	75	2012-08-02 05:01:06.745092
9847	5	1-1854	9845	17	64	1	8	2012-08-02 05:01:06.745092
9848	5	1-1661	9845	64	2	2	54	2012-08-02 05:01:06.745092
9849	5	1-2345	9845	2	66	4	17	2012-08-02 05:01:06.745092
9845	5	1-2777	9845	66	16	7	60	2012-08-02 05:01:06.745092
16044	6	1-4606	16039	4	63	607	66	2012-08-02 05:13:33.449333
9851	4	1-518	9850	44	42	2292	51	2012-08-02 05:01:06.745092
9852	4	1-2771	9850	42	2	1434	54	2012-08-02 05:01:06.745092
9853	4	1-2345	9850	2	66	3022	17	2012-08-02 05:01:06.745092
9850	4	1-2777	9850	66	44	5371	60	2012-08-02 05:01:06.745092
9855	5	1-2004	9854	84	7	1332	44	2012-08-02 05:01:06.745092
9856	5	1-2612	9854	7	57	644	89	2012-08-02 05:01:06.745092
9857	5	1-2706	9854	57	75	501	1	2012-08-02 05:01:06.745092
9858	5	1-431	9854	75	66	889	17	2012-08-02 05:01:06.745092
9854	5	1-2777	9854	66	84	2830	60	2012-08-02 05:01:06.745092
9860	5	1-2676	9859	36	41	3895	65	2012-08-02 05:01:08.380508
9861	5	1-2322	9859	41	93	2131	38	2012-08-02 05:01:08.380508
9862	5	1-1272	9859	93	39	2158	100	2012-08-02 05:01:08.380508
9863	5	1-2589	9859	39	63	502	11	2012-08-02 05:01:08.380508
9859	5	1-2781	9859	63	36	2168	8	2012-08-02 05:01:08.380508
9865	6	1-2676	9864	36	41	2705	65	2012-08-02 05:01:08.380508
9866	6	1-2322	9864	41	86	1512	38	2012-08-02 05:01:08.380508
9867	6	1-1821	9864	86	87	1105	70	2012-08-02 05:01:08.380508
9868	6	1-2270	9864	87	55	388	82	2012-08-02 05:01:08.380508
9869	6	1-2555	9864	55	63	334	11	2012-08-02 05:01:08.380508
9864	6	1-2781	9864	63	36	1474	8	2012-08-02 05:01:08.380508
9871	6	1-2409	9870	10	8	3228	3	2012-08-02 05:01:08.380508
9872	6	1-2671	9870	8	88	1307	26	2012-08-02 05:01:08.380508
9873	6	1-2060	9870	88	30	417	64	2012-08-02 05:01:08.380508
9874	6	1-2630	9870	30	55	495	82	2012-08-02 05:01:08.380508
9875	6	1-2555	9870	55	63	427	11	2012-08-02 05:01:08.380508
9870	6	1-2781	9870	63	10	1888	8	2012-08-02 05:01:08.380508
9877	5	1-2676	9876	36	89	1057	65	2012-08-02 05:01:08.380508
9878	5	1-1868	9876	89	87	315	70	2012-08-02 05:01:08.380508
9879	5	1-2270	9876	87	55	120	82	2012-08-02 05:01:08.380508
9880	5	1-2555	9876	55	63	112	11	2012-08-02 05:01:08.380508
9876	5	1-2781	9876	63	36	532	8	2012-08-02 05:01:08.380508
9882	5	1-2407	9881	61	89	383	65	2012-08-02 05:01:08.380508
9883	5	1-1868	9881	89	87	119	70	2012-08-02 05:01:08.380508
9884	5	1-2270	9881	87	55	47	82	2012-08-02 05:01:08.380508
9885	5	1-2555	9881	55	63	46	11	2012-08-02 05:01:08.380508
9881	5	1-2781	9881	63	61	230	8	2012-08-02 05:01:08.380508
9887	6	1-275	9886	76	26	320	9	2012-08-02 05:01:08.380508
9888	6	1-1838	9886	26	42	442	10	2012-08-02 05:01:08.380508
9889	6	1-2757	9886	42	37	160	91	2012-08-02 05:01:08.380508
9890	6	1-543	9886	37	55	78	82	2012-08-02 05:01:08.380508
9891	6	1-2555	9886	55	63	78	11	2012-08-02 05:01:08.380508
9886	6	1-2781	9886	63	76	399	8	2012-08-02 05:01:08.380508
9893	6	1-690	9892	63	72	211	92	2012-08-02 05:01:08.380508
9894	6	1-2626	9892	72	11	132	69	2012-08-02 05:01:08.380508
9895	6	1-2506	9892	11	30	33	64	2012-08-02 05:01:08.380508
9896	6	1-2630	9892	30	55	39	82	2012-08-02 05:01:08.380508
9897	6	1-2555	9892	55	63	34	11	2012-08-02 05:01:08.380508
9892	6	1-2781	9892	63	63	177	8	2012-08-02 05:01:08.380508
9899	6	1-1051	9898	59	47	2010	12	2012-08-02 05:01:08.380508
9900	6	1-2437	9898	47	57	2587	31	2012-08-02 05:01:08.380508
9901	6	1-2677	9898	57	68	1044	28	2012-08-02 05:01:08.380508
9902	6	1-1982	9898	68	55	422	82	2012-08-02 05:01:08.380508
9903	6	1-2555	9898	55	63	453	11	2012-08-02 05:01:08.380508
9898	6	1-2781	9898	63	59	2494	8	2012-08-02 05:01:08.380508
9905	5	1-1481	9904	83	60	2811	85	2012-08-02 05:01:10.923799
9906	5	1-2684	9904	60	96	1276	7	2012-08-02 05:01:10.923799
9907	5	1-2593	9904	96	37	1347	91	2012-08-02 05:01:10.923799
9908	5	1-543	9904	37	27	547	82	2012-08-02 05:01:10.923799
9904	5	1-2787	9904	27	83	9654	77	2012-08-02 05:01:10.923799
9910	4	1-504	9909	61	96	15	75	2012-08-02 05:01:10.923799
9911	4	1-2268	9909	96	68	11	28	2012-08-02 05:01:10.923799
9912	4	1-1982	9909	68	27	4	82	2012-08-02 05:01:10.923799
9909	4	1-2787	9909	27	61	67	77	2012-08-02 05:01:10.923799
9914	6	1-2685	9913	43	16	4620	57	2012-08-02 05:01:11.785278
9915	6	1-1445	9913	16	82	1847	59	2012-08-02 05:01:11.785278
9916	6	1-1796	9913	82	11	1465	54	2012-08-02 05:01:11.785278
9917	6	1-2760	9913	11	83	1042	77	2012-08-02 05:01:11.785278
9918	6	1-2743	9913	83	8	868	99	2012-08-02 05:01:11.785278
9913	6	1-2789	9913	8	43	2728	37	2012-08-02 05:01:11.785278
9920	5	1-436	9919	74	20	1494	43	2012-08-02 05:01:12.216516
9921	5	1-2166	9919	20	8	254	70	2012-08-02 05:01:12.216516
9922	5	1-2458	9919	8	39	1889	17	2012-08-02 05:01:12.216516
9923	5	1-2201	9919	39	93	743	42	2012-08-02 05:01:12.216516
9919	5	1-2790	9919	93	74	5613	67	2012-08-02 05:01:12.216516
9925	5	1-700	9924	75	3	11	78	2012-08-02 05:01:16.326107
9926	5	1-2212	9924	3	66	42	2	2012-08-02 05:01:16.326107
9927	5	1-2735	9924	66	25	16	31	2012-08-02 05:01:16.326107
9928	5	1-2475	9924	25	65	24	72	2012-08-02 05:01:16.326107
9924	5	1-2800	9924	65	75	11	61	2012-08-02 05:01:16.326107
9930	4	1-2383	9929	33	2	2267	59	2012-08-02 05:01:16.779156
9931	4	1-2613	9929	2	8	1412	98	2012-08-02 05:01:16.779156
9932	4	1-2736	9929	8	62	335	86	2012-08-02 05:01:16.779156
9929	4	1-2801	9929	62	33	4253	67	2012-08-02 05:01:16.779156
9934	5	1-1298	9933	47	66	1292	29	2012-08-02 05:01:16.779156
9935	5	1-2428	9933	66	64	529	30	2012-08-02 05:01:16.779156
9936	5	1-1294	9933	64	22	236	84	2012-08-02 05:01:16.779156
9937	5	1-1958	9933	22	62	141	86	2012-08-02 05:01:16.779156
9933	5	1-2801	9933	62	47	2691	67	2012-08-02 05:01:16.779156
9939	6	1-2732	9938	80	8	127	25	2012-08-02 05:01:17.640637
9940	6	1-2644	9938	8	72	135	71	2012-08-02 05:01:17.640637
9941	6	1-2709	9938	72	56	158	79	2012-08-02 05:01:17.640637
9942	6	1-2614	9938	56	98	168	97	2012-08-02 05:01:17.640637
9943	6	1-2502	9938	98	56	145	44	2012-08-02 05:01:17.640637
9938	6	1-2803	9938	56	80	104	74	2012-08-02 05:01:17.640637
9945	5	1-2608	9944	100	23	272	9	2012-08-02 05:01:17.640637
9946	5	1-2636	9944	23	18	717	5	2012-08-02 05:01:17.640637
9947	5	1-2776	9944	18	35	851	56	2012-08-02 05:01:17.640637
9948	5	1-1774	9944	35	56	819	44	2012-08-02 05:01:17.640637
9944	5	1-2803	9944	56	100	601	74	2012-08-02 05:01:17.640637
16039	6	1-4628	16039	63	85	1316	97	2012-08-02 05:13:33.449333
9950	5	1-2760	9949	11	83	246	77	2012-08-02 05:01:18.060564
9951	5	1-2743	9949	83	13	205	99	2012-08-02 05:01:18.060564
9952	5	1-820	9949	13	3	98	13	2012-08-02 05:01:18.060564
9953	5	1-2749	9949	3	17	306	4	2012-08-02 05:01:18.060564
9949	5	1-2804	9949	17	11	345	54	2012-08-02 05:01:18.060564
9955	6	1-2231	9954	47	8	752	98	2012-08-02 05:01:20.999047
9956	6	1-2736	9954	8	85	324	86	2012-08-02 05:01:20.999047
9957	6	1-2652	9954	85	62	2813	78	2012-08-02 05:01:20.999047
9958	6	1-1756	9954	62	98	406	63	2012-08-02 05:01:20.999047
9959	6	1-2263	9954	98	24	338	27	2012-08-02 05:01:20.999047
9954	6	1-2811	9954	24	47	1109	19	2012-08-02 05:01:20.999047
16046	6	1-4522	16045	85	63	1688	91	2012-08-02 05:13:33.449333
9961	6	1-2600	9960	17	65	87	22	2012-08-02 05:01:20.999047
9962	6	1-2522	9960	65	69	16	35	2012-08-02 05:01:20.999047
9963	6	1-2177	9960	69	93	16	1	2012-08-02 05:01:20.999047
9964	6	1-2417	9960	93	98	21	63	2012-08-02 05:01:20.999047
9965	6	1-2263	9960	98	24	19	27	2012-08-02 05:01:20.999047
9960	6	1-2811	9960	24	17	67	19	2012-08-02 05:01:20.999047
9967	4	1-1674	9966	94	48	2059	9	2012-08-02 05:01:20.999047
9968	4	1-2405	9966	48	42	2411	48	2012-08-02 05:01:20.999047
9969	4	1-1613	9966	42	24	1491	27	2012-08-02 05:01:20.999047
9966	4	1-2811	9966	24	94	5662	19	2012-08-02 05:01:20.999047
9971	6	1-2642	9970	37	12	1989	68	2012-08-02 05:01:23.098351
9972	6	1-1927	9970	12	57	1758	31	2012-08-02 05:01:23.098351
9973	6	1-2677	9970	57	68	656	28	2012-08-02 05:01:23.098351
9974	6	1-1982	9970	68	55	245	82	2012-08-02 05:01:23.098351
9975	6	1-2555	9970	55	72	244	11	2012-08-02 05:01:23.098351
9970	6	1-2816	9970	72	37	2129	41	2012-08-02 05:01:23.098351
9976	1	1-2555	\N	55	55	5798	11	2012-08-02 05:01:23.098351
9978	5	1-2598	9977	75	90	447	76	2012-08-02 05:01:26.003719
9979	5	1-2429	9977	90	86	116	38	2012-08-02 05:01:26.003719
9980	5	1-1821	9977	86	8	63	70	2012-08-02 05:01:26.003719
9981	5	1-2458	9977	8	64	746	17	2012-08-02 05:01:26.003719
9977	5	1-2823	9977	64	75	660	87	2012-08-02 05:01:26.003719
9983	5	1-2598	9982	75	90	575	76	2012-08-02 05:01:26.003719
9984	5	1-2429	9982	90	93	228	38	2012-08-02 05:01:26.003719
9985	5	1-1272	9982	93	74	265	100	2012-08-02 05:01:26.003719
9986	5	1-2386	9982	74	64	415	17	2012-08-02 05:01:26.003719
9982	5	1-2823	9982	64	75	559	87	2012-08-02 05:01:26.003719
9988	4	1-2480	9987	50	2	500	64	2012-08-02 05:01:26.003719
9989	4	1-2481	9987	2	74	396	100	2012-08-02 05:01:26.003719
9990	4	1-2386	9987	74	64	643	17	2012-08-02 05:01:26.003719
9987	4	1-2823	9987	64	50	896	87	2012-08-02 05:01:26.003719
9992	4	1-2819	9991	97	50	2354	59	2012-08-02 05:01:26.003719
9993	4	1-2763	9991	50	23	2500	15	2012-08-02 05:01:26.003719
9994	4	1-183	9991	23	64	824	17	2012-08-02 05:01:26.003719
9991	4	1-2823	9991	64	97	1160	87	2012-08-02 05:01:26.003719
9996	6	1-2111	9995	96	60	945	85	2012-08-02 05:01:26.821427
9997	6	1-2684	9995	60	96	497	7	2012-08-02 05:01:26.821427
9998	6	1-2593	9995	96	64	682	91	2012-08-02 05:01:26.821427
9999	6	1-2244	9995	64	19	728	73	2012-08-02 05:01:26.821427
10000	6	1-2688	9995	19	50	771	41	2012-08-02 05:01:26.821427
9995	6	1-2825	9995	50	96	1038	9	2012-08-02 05:01:26.821427
10002	6	1-2474	10001	1	30	3559	71	2012-08-02 05:01:26.821427
10003	6	1-1436	10001	30	45	1838	94	2012-08-02 05:01:26.821427
10004	6	1-2647	10001	45	26	984	30	2012-08-02 05:01:26.821427
10005	6	1-1686	10001	26	72	419	11	2012-08-02 05:01:26.821427
10006	6	1-2816	10001	72	50	3571	41	2012-08-02 05:01:26.821427
10001	6	1-2825	10001	50	1	4860	9	2012-08-02 05:01:26.821427
10008	5	1-2543	10007	23	27	577	15	2012-08-02 05:01:26.821427
10009	5	1-1408	10007	27	26	35	69	2012-08-02 05:01:26.821427
10010	5	1-2713	10007	26	51	51	21	2012-08-02 05:01:26.821427
10011	5	1-1457	10007	51	50	32	41	2012-08-02 05:01:26.821427
10007	5	1-2825	10007	50	23	44	9	2012-08-02 05:01:26.821427
10013	6	1-2405	10012	48	55	307	48	2012-08-02 05:01:26.821427
10014	6	1-2396	10012	55	43	424	93	2012-08-02 05:01:26.821427
10015	6	1-2441	10012	43	6	262	20	2012-08-02 05:01:26.821427
10016	6	1-2619	10012	6	19	151	73	2012-08-02 05:01:26.821427
10017	6	1-2688	10012	19	50	174	41	2012-08-02 05:01:26.821427
10012	6	1-2825	10012	50	48	255	9	2012-08-02 05:01:26.821427
10019	6	1-2543	10018	23	96	393	15	2012-08-02 05:01:26.821427
10020	6	1-2784	10018	96	66	206	2	2012-08-02 05:01:26.821427
10021	6	1-2735	10018	66	95	74	31	2012-08-02 05:01:26.821427
10022	6	1-2623	10018	95	19	17	73	2012-08-02 05:01:26.821427
10023	6	1-2688	10018	19	50	19	41	2012-08-02 05:01:26.821427
10018	6	1-2825	10018	50	23	28	9	2012-08-02 05:01:26.821427
10025	6	1-2474	10024	1	30	1100	71	2012-08-02 05:01:26.821427
10026	6	1-1436	10024	30	39	674	94	2012-08-02 05:01:26.821427
10027	6	1-2717	10024	39	26	617	69	2012-08-02 05:01:26.821427
10028	6	1-2713	10024	26	51	1053	21	2012-08-02 05:01:26.821427
10029	6	1-1457	10024	51	50	785	41	2012-08-02 05:01:26.821427
10024	6	1-2825	10024	50	1	1267	9	2012-08-02 05:01:26.821427
10031	6	1-2815	10030	56	93	2877	53	2012-08-02 05:01:26.821427
10032	6	1-2527	10030	93	27	459	39	2012-08-02 05:01:26.821427
10033	6	1-1903	10030	27	26	675	69	2012-08-02 05:01:26.821427
10034	6	1-2713	10030	26	51	1154	21	2012-08-02 05:01:26.821427
10035	6	1-1457	10030	51	50	862	41	2012-08-02 05:01:26.821427
10030	6	1-2825	10030	50	56	1395	9	2012-08-02 05:01:26.821427
10036	1	1-1457	\N	51	51	3137	41	2012-08-02 05:01:26.821427
10038	4	1-2472	10037	32	25	2792	11	2012-08-02 05:01:27.264196
10039	4	1-1698	10037	25	58	2650	98	2012-08-02 05:01:27.264196
10040	4	1-2783	10037	58	65	837	96	2012-08-02 05:01:27.264196
10037	4	1-2826	10037	65	32	2029	49	2012-08-02 05:01:27.264196
10042	2	1-1636	10041	62	94	281	55	2012-08-02 05:01:28.932658
10041	2	1-2830	10041	94	62	362	3	2012-08-02 05:01:28.932658
10044	6	1-2033	10043	70	93	964	81	2012-08-02 05:01:28.932658
10045	6	1-2728	10043	93	42	763	51	2012-08-02 05:01:28.932658
10046	6	1-2771	10043	42	34	1233	54	2012-08-02 05:01:28.932658
10047	6	1-2335	10043	34	35	331	52	2012-08-02 05:01:28.932658
10048	6	1-2236	10043	35	94	710	55	2012-08-02 05:01:28.932658
10043	6	1-2830	10043	94	70	1043	3	2012-08-02 05:01:28.932658
10050	5	1-2756	10049	40	35	3772	18	2012-08-02 05:01:28.932658
10051	5	1-2320	10049	35	13	2044	58	2012-08-02 05:01:28.932658
10052	5	1-944	10049	13	47	914	47	2012-08-02 05:01:28.932658
10053	5	1-2495	10049	47	94	1283	55	2012-08-02 05:01:28.932658
10049	5	1-2830	10049	94	40	1968	3	2012-08-02 05:01:28.932658
16047	6	1-2953	16045	63	64	1880	42	2012-08-02 05:13:33.449333
10055	5	1-2612	10054	7	57	327	89	2012-08-02 05:01:29.771994
10056	5	1-2706	10054	57	75	219	1	2012-08-02 05:01:29.771994
10057	5	1-431	10054	75	64	334	17	2012-08-02 05:01:29.771994
10058	5	1-2823	10054	64	64	410	87	2012-08-02 05:01:29.771994
10054	5	1-2832	10054	64	7	784	44	2012-08-02 05:01:29.771994
16048	6	1-4185	16045	64	35	711	28	2012-08-02 05:13:33.449333
10060	5	1-2612	10059	7	57	1274	89	2012-08-02 05:01:29.771994
10061	5	1-2706	10059	57	94	1371	1	2012-08-02 05:01:29.771994
10062	5	1-2678	10059	94	70	1323	91	2012-08-02 05:01:29.771994
10063	5	1-1623	10059	70	64	849	87	2012-08-02 05:01:29.771994
10059	5	1-2832	10059	64	7	1905	44	2012-08-02 05:01:29.771994
16049	6	1-4588	16045	35	4	951	92	2012-08-02 05:13:33.449333
10065	5	1-2500	10064	39	12	2106	58	2012-08-02 05:01:30.633783
10066	5	1-1486	10064	12	42	1763	51	2012-08-02 05:01:30.633783
10067	5	1-2771	10064	42	11	2876	54	2012-08-02 05:01:30.633783
10068	5	1-2760	10064	11	59	2054	77	2012-08-02 05:01:30.633783
10064	5	1-2834	10064	59	39	1602	36	2012-08-02 05:01:30.633783
16050	6	1-4606	16045	4	63	1566	66	2012-08-02 05:13:33.449333
10070	6	1-543	10069	37	39	454	82	2012-08-02 05:01:31.064532
10071	6	1-574	10069	39	66	334	29	2012-08-02 05:01:31.064532
10072	6	1-2428	10069	66	5	208	30	2012-08-02 05:01:31.064532
10073	6	1-2813	10069	5	26	662	97	2012-08-02 05:01:31.064532
10074	6	1-1430	10069	26	85	171	62	2012-08-02 05:01:31.064532
10069	6	1-2835	10069	85	37	839	91	2012-08-02 05:01:31.064532
10076	7	1-2581	10075	22	65	778	72	2012-08-02 05:01:31.064532
10077	7	1-2800	10075	65	31	346	61	2012-08-02 05:01:31.064532
10078	7	1-1142	10075	31	72	312	92	2012-08-02 05:01:31.064532
10079	7	1-2626	10075	72	12	250	69	2012-08-02 05:01:31.064532
10080	7	1-2391	10075	12	61	56	45	2012-08-02 05:01:31.064532
10081	7	1-985	10075	61	85	27	62	2012-08-02 05:01:31.064532
10075	7	1-2835	10075	85	22	134	91	2012-08-02 05:01:31.064532
10083	6	1-543	10082	37	49	246	82	2012-08-02 05:01:31.064532
10084	6	1-707	10082	49	45	167	94	2012-08-02 05:01:31.064532
10085	6	1-2647	10082	45	5	104	30	2012-08-02 05:01:31.064532
10086	6	1-2813	10082	5	26	337	97	2012-08-02 05:01:31.064532
10087	6	1-1430	10082	26	85	89	62	2012-08-02 05:01:31.064532
10082	6	1-2835	10082	85	37	446	91	2012-08-02 05:01:31.064532
10089	4	1-2647	10088	45	5	1042	30	2012-08-02 05:01:32.755101
10090	4	1-2813	10088	5	98	2708	97	2012-08-02 05:01:32.755101
10091	4	1-2502	10088	98	47	2011	44	2012-08-02 05:01:32.755101
10088	4	1-2839	10088	47	45	2094	94	2012-08-02 05:01:32.755101
10093	5	1-2161	10092	46	65	933	96	2012-08-02 05:01:32.755101
10094	5	1-2826	10092	65	71	1932	49	2012-08-02 05:01:32.755101
10095	5	1-2633	10092	71	53	1590	84	2012-08-02 05:01:32.755101
10096	5	1-2205	10092	53	47	1674	44	2012-08-02 05:01:32.755101
10092	5	1-2839	10092	47	46	1778	94	2012-08-02 05:01:32.755101
10098	6	1-2161	10097	46	65	410	96	2012-08-02 05:01:32.755101
10099	6	1-2826	10097	65	24	873	49	2012-08-02 05:01:32.755101
10100	6	1-2012	10097	24	63	699	46	2012-08-02 05:01:32.755101
10101	6	1-2377	10097	63	64	340	87	2012-08-02 05:01:32.755101
10102	6	1-2832	10097	64	47	697	44	2012-08-02 05:01:32.755101
10097	6	1-2839	10097	47	46	761	94	2012-08-02 05:01:32.755101
10104	6	1-2503	10103	7	63	219	46	2012-08-02 05:01:33.594898
10105	6	1-2377	10103	63	75	116	87	2012-08-02 05:01:33.594898
10106	6	1-2598	10103	75	87	128	76	2012-08-02 05:01:33.594898
10107	6	1-2780	10103	87	22	147	24	2012-08-02 05:01:33.594898
10108	6	1-2237	10103	22	36	240	16	2012-08-02 05:01:33.594898
10103	6	1-2841	10103	36	7	498	57	2012-08-02 05:01:33.594898
10110	5	1-2759	10109	3	64	2368	87	2012-08-02 05:01:35.28515
10111	5	1-2832	10109	64	77	4568	44	2012-08-02 05:01:35.28515
10112	5	1-2175	10109	77	6	3302	20	2012-08-02 05:01:35.28515
10113	5	1-2619	10109	6	80	1623	73	2012-08-02 05:01:35.28515
10109	5	1-2845	10109	80	3	5506	13	2012-08-02 05:01:35.28515
10115	6	1-2699	10114	80	8	3069	25	2012-08-02 05:01:36.533922
10116	6	1-2644	10114	8	85	3516	71	2012-08-02 05:01:36.533922
10117	6	1-2116	10114	85	72	941	92	2012-08-02 05:01:36.533922
10118	6	1-2626	10114	72	12	766	69	2012-08-02 05:01:36.533922
10119	6	1-2391	10114	12	13	174	45	2012-08-02 05:01:36.533922
10114	6	1-2848	10114	13	80	1776	63	2012-08-02 05:01:36.533922
10121	6	1-2263	10120	98	77	1392	27	2012-08-02 05:01:36.533922
10122	6	1-2524	10120	77	31	582	61	2012-08-02 05:01:36.533922
10123	6	1-1142	10120	31	72	562	92	2012-08-02 05:01:36.533922
10124	6	1-2626	10120	72	12	481	69	2012-08-02 05:01:36.533922
10125	6	1-2391	10120	12	13	115	45	2012-08-02 05:01:36.533922
10120	6	1-2848	10120	13	98	1233	63	2012-08-02 05:01:36.533922
10127	4	1-2426	10126	99	36	2238	16	2012-08-02 05:01:37.362291
10128	4	1-2841	10126	36	76	4533	57	2012-08-02 05:01:37.362291
10129	4	1-2388	10126	76	39	3234	50	2012-08-02 05:01:37.362291
10126	4	1-2850	10126	39	99	2579	55	2012-08-02 05:01:37.362291
10131	5	1-2830	10130	94	48	1906	3	2012-08-02 05:01:37.362291
10132	5	1-2338	10130	48	15	1696	24	2012-08-02 05:01:37.362291
10133	5	1-1917	10130	15	76	1486	57	2012-08-02 05:01:37.362291
10134	5	1-2388	10130	76	39	1250	50	2012-08-02 05:01:37.362291
10130	5	1-2850	10130	39	94	1176	55	2012-08-02 05:01:37.362291
10136	5	1-1686	10135	26	94	619	11	2012-08-02 05:01:38.190963
10137	5	1-1571	10135	94	13	222	45	2012-08-02 05:01:38.190963
10138	5	1-2848	10135	13	80	1441	63	2012-08-02 05:01:38.190963
10139	5	1-2699	10135	80	83	1585	25	2012-08-02 05:01:38.190963
10135	5	1-2852	10135	83	26	1971	30	2012-08-02 05:01:38.190963
10141	5	1-1686	10140	26	94	204	11	2012-08-02 05:01:38.190963
10142	5	1-1571	10140	94	13	83	45	2012-08-02 05:01:38.190963
10143	5	1-2848	10140	13	58	615	63	2012-08-02 05:01:38.190963
10144	5	1-2479	10140	58	83	403	25	2012-08-02 05:01:38.190963
10140	5	1-2852	10140	83	26	571	30	2012-08-02 05:01:38.190963
10146	5	1-2585	10145	20	58	1810	22	2012-08-02 05:01:38.190963
10147	5	1-2718	10145	58	66	5346	80	2012-08-02 05:01:38.190963
10148	5	1-1922	10145	66	80	619	74	2012-08-02 05:01:38.190963
10149	5	1-2732	10145	80	83	709	25	2012-08-02 05:01:38.190963
10145	5	1-2852	10145	83	20	1211	30	2012-08-02 05:01:38.190963
10151	5	1-1294	10150	64	53	1078	84	2012-08-02 05:01:38.190963
10152	5	1-2205	10150	53	56	1258	44	2012-08-02 05:01:38.190963
10153	5	1-2803	10150	56	80	818	74	2012-08-02 05:01:38.190963
10154	5	1-2732	10150	80	83	970	25	2012-08-02 05:01:38.190963
10150	5	1-2852	10150	83	64	1717	30	2012-08-02 05:01:38.190963
16045	6	1-4628	16045	63	85	3688	97	2012-08-02 05:13:33.449333
10156	5	1-2813	10155	5	3	3688	97	2012-08-02 05:01:38.190963
10157	5	1-1926	10155	3	93	3272	58	2012-08-02 05:01:38.190963
10158	5	1-1748	10155	93	84	3576	50	2012-08-02 05:01:38.190963
10159	5	1-715	10155	84	83	690	25	2012-08-02 05:01:38.190963
10155	5	1-2852	10155	83	5	1238	30	2012-08-02 05:01:38.190963
10161	6	1-171	10160	72	47	297	47	2012-08-02 05:01:38.190963
10162	6	1-2495	10160	47	94	385	55	2012-08-02 05:01:38.190963
10163	6	1-2830	10160	94	8	546	3	2012-08-02 05:01:38.190963
10164	6	1-2671	10160	8	18	267	26	2012-08-02 05:01:38.190963
10165	6	1-1513	10160	18	83	367	25	2012-08-02 05:01:38.190963
10160	6	1-2852	10160	83	72	663	30	2012-08-02 05:01:38.190963
10167	5	1-171	10166	72	90	191	47	2012-08-02 05:01:38.190963
10168	5	1-2694	10166	90	4	1005	78	2012-08-02 05:01:38.190963
10169	5	1-2814	10166	4	12	278	95	2012-08-02 05:01:38.190963
10170	5	1-2366	10166	12	83	206	25	2012-08-02 05:01:38.190963
10166	5	1-2852	10166	83	72	399	30	2012-08-02 05:01:38.190963
10172	5	1-2585	10171	20	65	3052	22	2012-08-02 05:01:38.190963
10173	5	1-2522	10171	65	60	671	35	2012-08-02 05:01:38.190963
10174	5	1-1585	10171	60	80	629	74	2012-08-02 05:01:38.190963
10175	5	1-2732	10171	80	83	853	25	2012-08-02 05:01:38.190963
10171	5	1-2852	10171	83	20	1725	30	2012-08-02 05:01:38.190963
10176	1	1-1294	\N	64	64	934	84	2012-08-02 05:01:38.190963
10178	6	1-2206	10177	84	63	554	46	2012-08-02 05:01:38.668588
10179	6	1-2377	10177	63	64	336	87	2012-08-02 05:01:38.668588
10180	6	1-2832	10177	64	56	859	44	2012-08-02 05:01:38.668588
10181	6	1-2803	10177	56	100	645	74	2012-08-02 05:01:38.668588
10182	6	1-2608	10177	100	73	298	9	2012-08-02 05:01:38.668588
10177	6	1-2853	10177	73	84	355	91	2012-08-02 05:01:38.668588
10184	4	1-2853	10183	73	3	6657	91	2012-08-02 05:01:40.325948
10185	4	1-2638	10183	3	47	5399	47	2012-08-02 05:01:40.325948
10186	4	1-2495	10183	47	8	7193	55	2012-08-02 05:01:40.325948
10183	4	1-2857	10183	8	73	6162	9	2012-08-02 05:01:40.325948
10188	4	1-2111	10187	96	21	705	85	2012-08-02 05:01:40.325948
10189	4	1-2680	10187	21	47	543	47	2012-08-02 05:01:40.325948
10190	4	1-2495	10187	47	8	781	55	2012-08-02 05:01:40.325948
10187	4	1-2857	10187	8	96	722	9	2012-08-02 05:01:40.325948
10192	6	1-2760	10191	11	83	1473	77	2012-08-02 05:01:43.661842
10193	6	1-2743	10191	83	75	1009	99	2012-08-02 05:01:43.661842
10194	6	1-1182	10191	75	94	449	25	2012-08-02 05:01:43.661842
10195	6	1-2259	10191	94	7	180	22	2012-08-02 05:01:43.661842
10196	6	1-2859	10191	7	78	359	76	2012-08-02 05:01:43.661842
10191	6	1-2865	10191	78	11	2516	54	2012-08-02 05:01:43.661842
10198	6	1-1515	10197	64	46	332	94	2012-08-02 05:01:43.661842
10199	6	1-2161	10197	46	65	180	96	2012-08-02 05:01:43.661842
10200	6	1-2826	10197	65	48	386	49	2012-08-02 05:01:43.661842
10201	6	1-1022	10197	48	7	267	22	2012-08-02 05:01:43.661842
10202	6	1-2859	10197	7	78	566	76	2012-08-02 05:01:43.661842
10197	6	1-2865	10197	78	64	4227	54	2012-08-02 05:01:43.661842
10204	4	1-2856	10203	16	39	1333	50	2012-08-02 05:01:44.081891
10205	4	1-2850	10203	39	10	1209	55	2012-08-02 05:01:44.081891
10206	4	1-2794	10203	10	32	1192	17	2012-08-02 05:01:44.081891
10203	4	1-2866	10203	32	16	1837	90	2012-08-02 05:01:44.081891
10208	4	1-2671	10207	8	17	982	26	2012-08-02 05:01:44.491033
10209	4	1-211	10207	17	63	152	86	2012-08-02 05:01:44.491033
10210	4	1-2482	10207	63	100	44	60	2012-08-02 05:01:44.491033
10207	4	1-2867	10207	100	8	3258	3	2012-08-02 05:01:44.491033
10212	6	1-1226	10211	17	79	1105	43	2012-08-02 05:01:44.491033
10213	6	1-2654	10211	79	42	342	79	2012-08-02 05:01:44.491033
10214	6	1-2711	10211	42	59	251	37	2012-08-02 05:01:44.491033
10215	6	1-2715	10211	59	26	156	46	2012-08-02 05:01:44.491033
10216	6	1-2579	10211	26	100	39	60	2012-08-02 05:01:44.491033
10211	6	1-2867	10211	100	17	2897	3	2012-08-02 05:01:44.491033
10218	6	1-2706	10217	57	86	4081	1	2012-08-02 05:01:48.655442
10219	6	1-1803	10217	86	65	2437	22	2012-08-02 05:01:48.655442
10220	6	1-2522	10217	65	9	403	35	2012-08-02 05:01:48.655442
10221	6	1-2817	10217	9	65	211	96	2012-08-02 05:01:48.655442
10222	6	1-2826	10217	65	6	417	49	2012-08-02 05:01:48.655442
10217	6	1-2877	10217	6	57	6171	89	2012-08-02 05:01:48.655442
10224	5	1-2615	10223	47	83	2672	52	2012-08-02 05:01:48.655442
10225	5	1-1404	10223	83	58	475	98	2012-08-02 05:01:48.655442
10226	5	1-2783	10223	58	65	91	96	2012-08-02 05:01:48.655442
10227	5	1-2826	10223	65	6	134	49	2012-08-02 05:01:48.655442
10223	5	1-2877	10223	6	47	2051	89	2012-08-02 05:01:48.655442
10229	5	1-2500	10228	39	57	85	58	2012-08-02 05:01:49.506306
10230	5	1-1996	10228	57	26	67	69	2012-08-02 05:01:49.506306
10231	5	1-2713	10228	26	89	104	21	2012-08-02 05:01:49.506306
10232	5	1-2596	10228	89	78	344	5	2012-08-02 05:01:49.506306
10228	5	1-2879	10228	78	39	68	36	2012-08-02 05:01:49.506306
10234	4	1-2859	10233	7	87	2968	76	2012-08-02 05:01:51.174685
10235	4	1-2780	10233	87	94	3270	24	2012-08-02 05:01:51.174685
10236	4	1-2782	10233	94	26	1615	92	2012-08-02 05:01:51.174685
10233	4	1-2883	10233	26	7	1342	22	2012-08-02 05:01:51.174685
10238	5	1-1622	10237	64	72	1456	92	2012-08-02 05:01:52.831904
10239	5	1-2626	10237	72	12	1224	69	2012-08-02 05:01:52.831904
10240	5	1-2391	10237	12	86	287	45	2012-08-02 05:01:52.831904
10241	5	1-2824	10237	86	78	1169	97	2012-08-02 05:01:52.831904
10237	5	1-2887	10237	78	64	3705	32	2012-08-02 05:01:52.831904
10243	6	1-2832	10242	64	56	474	44	2012-08-02 05:01:53.671652
10244	6	1-2803	10242	56	51	334	74	2012-08-02 05:01:53.671652
10245	6	1-2722	10242	51	42	266	79	2012-08-02 05:01:53.671652
10246	6	1-2711	10242	42	59	339	37	2012-08-02 05:01:53.671652
10247	6	1-2715	10242	59	13	366	46	2012-08-02 05:01:53.671652
10242	6	1-2889	10242	13	64	198	87	2012-08-02 05:01:53.671652
10249	5	1-2743	10248	83	8	558	99	2012-08-02 05:01:56.532832
10250	5	1-2789	10248	8	59	1588	37	2012-08-02 05:01:56.532832
10251	5	1-2715	10248	59	63	1524	46	2012-08-02 05:01:56.532832
10252	5	1-2377	10248	63	44	770	87	2012-08-02 05:01:56.532832
10248	5	1-2896	10248	44	83	740	77	2012-08-02 05:01:56.532832
10253	1	1-2377	\N	63	63	80	87	2012-08-02 05:01:56.532832
10255	3	1-247	10254	10	78	827	97	2012-08-02 05:01:58.598889
10256	3	1-2887	10254	78	20	2064	32	2012-08-02 05:01:58.598889
10254	3	1-2901	10254	20	10	1647	30	2012-08-02 05:01:58.598889
10258	5	1-2824	10257	86	78	158	97	2012-08-02 05:01:59.007534
10257	5	1-2902	10257	58	86	46	45	2012-08-02 05:01:59.007534
10259	5	1-2887	10257	78	20	423	32	2012-08-02 05:01:59.007534
10260	5	1-2901	10257	20	27	363	30	2012-08-02 05:01:59.007534
10261	5	1-2561	10257	27	58	213	19	2012-08-02 05:01:59.007534
10263	5	1-2901	10262	20	95	1466	30	2012-08-02 05:01:59.825452
10264	5	1-1277	10262	95	82	1587	59	2012-08-02 05:01:59.825452
10265	5	1-1796	10262	82	13	1103	54	2012-08-02 05:01:59.825452
10266	5	1-2827	10262	13	74	2152	42	2012-08-02 05:01:59.825452
10262	5	1-2904	10262	74	20	1770	32	2012-08-02 05:01:59.825452
16052	5	1-4578	16051	17	10	1134	18	2012-08-02 05:13:33.869408
10268	5	1-1622	10267	64	19	920	92	2012-08-02 05:01:59.825452
10269	5	1-2828	10267	19	42	526	51	2012-08-02 05:01:59.825452
10270	5	1-2771	10267	42	13	930	54	2012-08-02 05:01:59.825452
10271	5	1-2827	10267	13	74	2256	42	2012-08-02 05:01:59.825452
10267	5	1-2904	10267	74	64	2305	32	2012-08-02 05:01:59.825452
16053	5	1-4605	16051	10	42	2606	86	2012-08-02 05:13:33.869408
10273	5	1-1309	10272	12	80	1050	65	2012-08-02 05:02:00.665143
10274	5	1-2880	10272	80	9	815	87	2012-08-02 05:02:00.665143
10275	5	1-2833	10272	9	39	2178	40	2012-08-02 05:02:00.665143
10276	5	1-2438	10272	39	3	900	27	2012-08-02 05:02:00.665143
10272	5	1-2906	10272	3	12	1455	67	2012-08-02 05:02:00.665143
16054	5	1-4511	16051	42	55	3594	94	2012-08-02 05:13:33.869408
10278	6	1-1309	10277	12	51	552	65	2012-08-02 05:02:00.665143
10279	6	1-2253	10277	51	69	155	35	2012-08-02 05:02:00.665143
10280	6	1-2177	10277	69	93	182	1	2012-08-02 05:02:00.665143
10281	6	1-2417	10277	93	98	279	63	2012-08-02 05:02:00.665143
10282	6	1-2263	10277	98	3	295	27	2012-08-02 05:02:00.665143
10277	6	1-2906	10277	3	12	604	67	2012-08-02 05:02:00.665143
10284	5	1-436	10283	74	20	355	43	2012-08-02 05:02:00.665143
10285	5	1-2166	10283	20	45	181	70	2012-08-02 05:02:00.665143
10286	5	1-2650	10283	45	42	280	48	2012-08-02 05:02:00.665143
10287	5	1-1613	10283	42	3	203	27	2012-08-02 05:02:00.665143
10283	5	1-2906	10283	3	74	446	67	2012-08-02 05:02:00.665143
10289	5	1-2650	10288	45	42	9	48	2012-08-02 05:02:01.913707
10290	5	1-1613	10288	42	3	6	27	2012-08-02 05:02:01.913707
10291	5	1-2906	10288	3	44	12	67	2012-08-02 05:02:01.913707
10292	5	1-2494	10288	44	32	9	15	2012-08-02 05:02:01.913707
10288	5	1-2909	10288	32	45	6	70	2012-08-02 05:02:01.913707
10294	4	1-1084	10293	61	47	489	12	2012-08-02 05:02:02.753314
10295	4	1-2437	10293	47	25	705	31	2012-08-02 05:02:02.753314
10296	4	1-2475	10293	25	20	1113	72	2012-08-02 05:02:02.753314
10293	4	1-2911	10293	20	61	871	95	2012-08-02 05:02:02.753314
10298	3	1-2896	10297	44	50	3165	77	2012-08-02 05:02:03.173376
10299	3	1-2863	10297	50	41	1479	14	2012-08-02 05:02:03.173376
10297	3	1-2912	10297	41	44	5587	87	2012-08-02 05:02:03.173376
10301	6	1-2802	10300	100	8	1464	99	2012-08-02 05:02:04.001512
10302	6	1-2789	10300	8	59	3753	37	2012-08-02 05:02:04.001512
10303	6	1-2715	10300	59	13	3244	46	2012-08-02 05:02:04.001512
10304	6	1-2889	10300	13	75	1406	87	2012-08-02 05:02:04.001512
10305	6	1-2598	10300	75	99	1325	76	2012-08-02 05:02:04.001512
10300	6	1-2914	10300	99	100	4109	71	2012-08-02 05:02:04.001512
10307	5	1-2617	10306	63	7	878	57	2012-08-02 05:02:04.001512
10308	5	1-2503	10306	7	13	367	46	2012-08-02 05:02:04.001512
10309	5	1-2889	10306	13	75	177	87	2012-08-02 05:02:04.001512
10310	5	1-2598	10306	75	99	185	76	2012-08-02 05:02:04.001512
10306	5	1-2914	10306	99	63	638	71	2012-08-02 05:02:04.001512
10312	5	1-2897	10311	43	75	648	66	2012-08-02 05:02:04.001512
10313	5	1-2266	10311	75	3	400	27	2012-08-02 05:02:04.001512
10314	5	1-2906	10311	3	45	778	67	2012-08-02 05:02:04.001512
10315	5	1-1910	10311	45	99	630	76	2012-08-02 05:02:04.001512
10311	5	1-2914	10311	99	43	2328	71	2012-08-02 05:02:04.001512
10317	4	1-2495	10316	47	99	70	55	2012-08-02 05:02:06.932077
10318	4	1-2426	10316	99	58	32	16	2012-08-02 05:02:06.932077
10319	4	1-2594	10316	58	68	36	15	2012-08-02 05:02:06.932077
10316	4	1-2921	10316	68	47	105	47	2012-08-02 05:02:06.932077
10321	4	1-2634	10320	56	99	214	55	2012-08-02 05:02:06.932077
10322	4	1-2426	10320	99	58	113	16	2012-08-02 05:02:06.932077
10323	4	1-2594	10320	58	68	145	15	2012-08-02 05:02:06.932077
10320	4	1-2921	10320	68	56	491	47	2012-08-02 05:02:06.932077
10325	4	1-2694	10324	90	4	1857	78	2012-08-02 05:02:06.932077
10326	4	1-2814	10324	4	83	308	95	2012-08-02 05:02:06.932077
10327	4	1-1781	10324	83	68	161	15	2012-08-02 05:02:06.932077
10324	4	1-2921	10324	68	90	590	47	2012-08-02 05:02:06.932077
10329	4	1-2572	10328	15	88	729	33	2012-08-02 05:02:06.932077
10330	4	1-2535	10328	88	50	334	59	2012-08-02 05:02:06.932077
10331	4	1-2763	10328	50	68	259	15	2012-08-02 05:02:06.932077
10328	4	1-2921	10328	68	15	1053	47	2012-08-02 05:02:06.932077
10333	4	1-848	10332	4	56	742	86	2012-08-02 05:02:07.783043
10334	4	1-1595	10332	56	6	178	77	2012-08-02 05:02:07.783043
10335	4	1-2913	10332	6	95	28	39	2012-08-02 05:02:07.783043
10332	4	1-2923	10332	95	4	7291	30	2012-08-02 05:02:07.783043
10337	4	1-1815	10336	15	12	170	31	2012-08-02 05:02:07.783043
10338	4	1-2151	10336	12	22	33	83	2012-08-02 05:02:07.783043
10339	4	1-1191	10336	22	95	3	39	2012-08-02 05:02:07.783043
10336	4	1-2923	10336	95	15	796	30	2012-08-02 05:02:07.783043
10341	6	1-1176	10340	75	100	457	74	2012-08-02 05:02:09.031601
10342	6	1-2608	10340	100	73	192	9	2012-08-02 05:02:09.031601
10343	6	1-2853	10340	73	64	207	91	2012-08-02 05:02:09.031601
10344	6	1-2244	10340	64	19	246	73	2012-08-02 05:02:09.031601
10345	6	1-2688	10340	19	49	290	41	2012-08-02 05:02:09.031601
10340	6	1-2926	10340	49	75	360	49	2012-08-02 05:02:09.031601
10347	3	1-2584	10346	12	25	1969	47	2012-08-02 05:02:10.280083
10348	3	1-2791	10346	25	57	1453	36	2012-08-02 05:02:10.280083
10346	3	1-2929	10346	57	12	2185	34	2012-08-02 05:02:10.280083
10350	6	1-2062	10349	66	47	2533	20	2012-08-02 05:02:10.280083
10351	6	1-2704	10349	47	30	4451	6	2012-08-02 05:02:10.280083
10352	6	1-2065	10349	30	21	813	85	2012-08-02 05:02:10.280083
10353	6	1-2680	10349	21	25	571	47	2012-08-02 05:02:10.280083
10354	6	1-2791	10349	25	57	535	36	2012-08-02 05:02:10.280083
10349	6	1-2929	10349	57	66	1021	34	2012-08-02 05:02:10.280083
10356	4	1-1970	10355	98	8	2678	89	2012-08-02 05:02:10.280083
10357	4	1-2382	10355	8	97	3027	12	2012-08-02 05:02:10.280083
10358	4	1-2269	10355	97	57	1377	36	2012-08-02 05:02:10.280083
10355	4	1-2929	10355	57	98	2773	34	2012-08-02 05:02:10.280083
10360	5	1-2870	10359	6	57	4450	33	2012-08-02 05:02:10.280083
10361	5	1-2886	10359	57	13	2655	58	2012-08-02 05:02:10.280083
10362	5	1-944	10359	13	25	1241	47	2012-08-02 05:02:10.280083
10363	5	1-2791	10359	25	57	1300	36	2012-08-02 05:02:10.280083
10359	5	1-2929	10359	57	6	2806	34	2012-08-02 05:02:10.280083
16055	5	1-4104	16051	55	8	2153	95	2012-08-02 05:13:33.869408
10365	3	1-275	10364	76	56	1637	9	2012-08-02 05:02:11.926122
10366	3	1-2815	10364	56	67	2188	53	2012-08-02 05:02:11.926122
10364	3	1-2933	10364	67	76	2724	8	2012-08-02 05:02:11.926122
16051	5	1-4629	16051	8	17	1274	9	2012-08-02 05:13:33.869408
10368	5	1-429	10367	72	20	717	72	2012-08-02 05:02:11.926122
10369	5	1-2911	10367	20	61	419	95	2012-08-02 05:02:11.926122
10370	5	1-2799	10367	61	5	351	17	2012-08-02 05:02:11.926122
10371	5	1-2836	10367	5	67	357	53	2012-08-02 05:02:11.926122
10367	5	1-2933	10367	67	72	533	8	2012-08-02 05:02:11.926122
10373	2	1-2809	10372	29	100	3281	64	2012-08-02 05:02:12.765897
10372	2	1-2935	10372	100	29	5792	91	2012-08-02 05:02:12.765897
16057	5	1-3192	16056	83	9	54	93	2012-08-02 05:13:33.869408
10375	5	1-2855	10374	86	85	166	18	2012-08-02 05:02:12.765897
10376	5	1-2922	10374	85	57	98	36	2012-08-02 05:02:12.765897
10377	5	1-2929	10374	57	46	185	34	2012-08-02 05:02:12.765897
10378	5	1-1937	10374	46	100	190	64	2012-08-02 05:02:12.765897
10374	5	1-2935	10374	100	86	353	91	2012-08-02 05:02:12.765897
16058	5	1-3666	16056	9	60	37	13	2012-08-02 05:13:33.869408
10380	6	1-2654	10379	79	56	2199	79	2012-08-02 05:02:15.240565
10381	6	1-2614	10379	56	26	2432	97	2012-08-02 05:02:15.240565
10382	6	1-1430	10379	26	85	668	62	2012-08-02 05:02:15.240565
10383	6	1-2835	10379	85	33	3490	91	2012-08-02 05:02:15.240565
10384	6	1-1491	10379	33	86	1879	7	2012-08-02 05:02:15.240565
10379	6	1-2941	10379	86	79	3838	43	2012-08-02 05:02:15.240565
16059	5	1-4561	16056	60	31	107	45	2012-08-02 05:13:33.869408
10386	5	1-1191	10385	22	27	215	39	2012-08-02 05:02:16.478267
10387	5	1-1903	10385	27	12	253	69	2012-08-02 05:02:16.478267
10388	5	1-2391	10385	12	13	47	45	2012-08-02 05:02:16.478267
10389	5	1-2848	10385	13	90	395	63	2012-08-02 05:02:16.478267
10385	5	1-2944	10385	90	22	804	83	2012-08-02 05:02:16.478267
16060	5	1-3797	16056	31	8	84	95	2012-08-02 05:13:33.869408
10391	5	1-1651	10390	25	32	1952	17	2012-08-02 05:02:16.908756
10392	5	1-2866	10390	32	84	1521	90	2012-08-02 05:02:16.908756
10393	5	1-334	10390	84	4	596	14	2012-08-02 05:02:16.908756
10394	5	1-2533	10390	4	3	488	2	2012-08-02 05:02:16.908756
10390	5	1-2945	10390	3	25	2943	33	2012-08-02 05:02:16.908756
10396	5	1-2886	10395	57	57	964	58	2012-08-02 05:02:16.908756
10397	5	1-1996	10395	57	19	631	69	2012-08-02 05:02:16.908756
10398	5	1-1341	10395	19	4	388	14	2012-08-02 05:02:16.908756
10399	5	1-2533	10395	4	3	342	2	2012-08-02 05:02:16.908756
10395	5	1-2945	10395	3	57	2219	33	2012-08-02 05:02:16.908756
10401	4	1-2427	10400	76	85	345	100	2012-08-02 05:02:17.74858
10402	4	1-2872	10400	85	36	212	77	2012-08-02 05:02:17.74858
10403	4	1-2276	10400	36	75	81	81	2012-08-02 05:02:17.74858
10400	4	1-2947	10400	75	76	191	36	2012-08-02 05:02:17.74858
10405	4	1-2929	10404	57	6	90	34	2012-08-02 05:02:17.74858
10406	4	1-2870	10404	6	29	95	33	2012-08-02 05:02:17.74858
10407	4	1-1004	10404	29	75	25	81	2012-08-02 05:02:17.74858
10404	4	1-2947	10404	75	57	63	36	2012-08-02 05:02:17.74858
10409	5	1-2847	10408	65	80	5844	65	2012-08-02 05:02:17.74858
10410	5	1-2880	10408	80	44	5153	87	2012-08-02 05:02:17.74858
10411	5	1-2896	10408	44	36	5014	77	2012-08-02 05:02:17.74858
10412	5	1-2276	10408	36	75	2642	81	2012-08-02 05:02:17.74858
10408	5	1-2947	10408	75	65	8496	36	2012-08-02 05:02:17.74858
10414	5	1-1910	10413	45	97	3813	76	2012-08-02 05:02:20.59891
10415	5	1-1736	10413	97	90	2017	50	2012-08-02 05:02:20.59891
10416	5	1-2354	10413	90	20	880	72	2012-08-02 05:02:20.59891
10417	5	1-2911	10413	20	61	410	95	2012-08-02 05:02:20.59891
10413	5	1-2954	10413	61	45	6936	67	2012-08-02 05:02:20.59891
10419	5	1-2853	10418	73	70	249	91	2012-08-02 05:02:21.803018
10420	5	1-1623	10418	70	97	169	87	2012-08-02 05:02:21.803018
10421	5	1-2819	10418	97	45	372	59	2012-08-02 05:02:21.803018
10422	5	1-1493	10418	45	87	277	90	2012-08-02 05:02:21.803018
10418	5	1-2957	10418	87	73	225	9	2012-08-02 05:02:21.803018
10424	3	1-1921	10423	83	55	1593	4	2012-08-02 05:02:21.803018
10425	3	1-2609	10423	55	87	1599	90	2012-08-02 05:02:21.803018
10423	3	1-2957	10423	87	83	1313	9	2012-08-02 05:02:21.803018
10427	5	1-2944	10426	90	18	4571	83	2012-08-02 05:02:22.212404
10428	5	1-2603	10426	18	99	5704	80	2012-08-02 05:02:22.212404
10429	5	1-2890	10426	99	86	1859	51	2012-08-02 05:02:22.212404
10430	5	1-2930	10426	86	84	364	91	2012-08-02 05:02:22.212404
10426	5	1-2958	10426	84	90	2709	63	2012-08-02 05:02:22.212404
10432	6	1-2944	10431	90	22	337	83	2012-08-02 05:02:22.212404
10433	6	1-1191	10431	22	27	77	39	2012-08-02 05:02:22.212404
10434	6	1-1903	10431	27	11	78	69	2012-08-02 05:02:22.212404
10435	6	1-2506	10431	11	100	18	64	2012-08-02 05:02:22.212404
10436	6	1-2935	10431	100	84	25	91	2012-08-02 05:02:22.212404
10431	6	1-2958	10431	84	90	193	63	2012-08-02 05:02:22.212404
10438	5	1-2263	10437	98	20	649	27	2012-08-02 05:02:22.212404
10439	5	1-2492	10437	20	4	145	94	2012-08-02 05:02:22.212404
10440	5	1-2439	10437	4	86	431	51	2012-08-02 05:02:22.212404
10441	5	1-2930	10437	86	84	95	91	2012-08-02 05:02:22.212404
10437	5	1-2958	10437	84	98	792	63	2012-08-02 05:02:22.212404
10443	5	1-1517	10442	93	37	1564	85	2012-08-02 05:02:22.212404
10444	5	1-2588	10442	37	29	3475	54	2012-08-02 05:02:22.212404
10445	5	1-1133	10442	29	86	952	51	2012-08-02 05:02:22.212404
10446	5	1-2930	10442	86	84	242	91	2012-08-02 05:02:22.212404
10442	5	1-2958	10442	84	93	2345	63	2012-08-02 05:02:22.212404
10448	6	1-664	10447	73	90	843	50	2012-08-02 05:02:22.212404
10449	6	1-2354	10447	90	55	515	72	2012-08-02 05:02:22.212404
10450	6	1-2888	10447	55	53	374	21	2012-08-02 05:02:22.212404
10451	6	1-2653	10447	53	100	61	64	2012-08-02 05:02:22.212404
10452	6	1-2935	10447	100	84	108	91	2012-08-02 05:02:22.212404
10447	6	1-2958	10447	84	73	1051	63	2012-08-02 05:02:22.212404
10454	6	1-1567	10453	51	49	339	5	2012-08-02 05:02:22.212404
10455	6	1-2821	10453	49	56	44	44	2012-08-02 05:02:22.212404
10456	6	1-2803	10453	56	15	29	74	2012-08-02 05:02:22.212404
10457	6	1-2951	10453	15	73	27	9	2012-08-02 05:02:22.212404
10458	6	1-2853	10453	73	84	28	91	2012-08-02 05:02:22.212404
10453	6	1-2958	10453	84	51	276	63	2012-08-02 05:02:22.212404
10460	4	1-2690	10459	50	99	114	80	2012-08-02 05:02:22.212404
10461	4	1-2890	10459	99	86	51	51	2012-08-02 05:02:22.212404
10462	4	1-2930	10459	86	84	14	91	2012-08-02 05:02:22.212404
10459	4	1-2958	10459	84	50	143	63	2012-08-02 05:02:22.212404
10463	1	1-664	\N	73	73	2078	50	2012-08-02 05:02:22.212404
10464	1	1-2608	\N	100	100	464	9	2012-08-02 05:02:22.212404
16056	5	1-4629	16056	8	83	59	9	2012-08-02 05:13:33.869408
10466	4	1-2883	10465	26	31	926	22	2012-08-02 05:02:22.654537
10467	4	1-1851	10465	31	58	466	16	2012-08-02 05:02:22.654537
10468	4	1-2594	10465	58	11	756	15	2012-08-02 05:02:22.654537
10465	4	1-2959	10465	11	26	1428	92	2012-08-02 05:02:22.654537
10470	5	1-2955	10469	82	4	1023	94	2012-08-02 05:02:22.654537
10471	5	1-2439	10469	4	86	2785	51	2012-08-02 05:02:22.654537
10472	5	1-2930	10469	86	16	562	91	2012-08-02 05:02:22.654537
10473	5	1-2397	10469	16	11	762	15	2012-08-02 05:02:22.654537
10469	5	1-2959	10469	11	82	1468	92	2012-08-02 05:02:22.654537
10475	5	1-2955	10474	82	46	4028	94	2012-08-02 05:02:22.654537
10476	5	1-2161	10474	46	65	1971	96	2012-08-02 05:02:22.654537
10477	5	1-2826	10474	65	63	3808	49	2012-08-02 05:02:22.654537
10478	5	1-2032	10474	63	11	2541	15	2012-08-02 05:02:22.654537
10474	5	1-2959	10474	11	82	5322	92	2012-08-02 05:02:22.654537
10480	5	1-2425	10479	44	47	777	89	2012-08-02 05:02:23.063517
10481	5	1-2615	10479	47	87	1385	52	2012-08-02 05:02:23.063517
10482	5	1-2200	10479	87	50	1274	59	2012-08-02 05:02:23.063517
10483	5	1-2763	10479	50	32	1198	15	2012-08-02 05:02:23.063517
10479	5	1-2960	10479	32	44	1361	82	2012-08-02 05:02:23.063517
10485	5	1-2942	10484	46	10	1294	84	2012-08-02 05:02:23.063517
10486	5	1-2649	10484	10	73	1003	9	2012-08-02 05:02:23.063517
10487	5	1-2853	10484	73	16	1094	91	2012-08-02 05:02:23.063517
10488	5	1-2397	10484	16	32	2028	15	2012-08-02 05:02:23.063517
10484	5	1-2960	10484	32	46	2776	82	2012-08-02 05:02:23.063517
10490	5	1-409	10489	67	25	994	31	2012-08-02 05:02:23.063517
10491	5	1-2475	10489	25	20	1446	72	2012-08-02 05:02:23.063517
10492	5	1-2911	10489	20	83	1044	95	2012-08-02 05:02:23.063517
10493	5	1-1781	10489	83	32	910	15	2012-08-02 05:02:23.063517
10489	5	1-2960	10489	32	67	1279	82	2012-08-02 05:02:23.063517
10495	5	1-2814	10494	4	12	1523	95	2012-08-02 05:02:23.472456
10496	5	1-2366	10494	12	7	1161	25	2012-08-02 05:02:23.472456
10497	5	1-1666	10494	7	16	703	10	2012-08-02 05:02:23.472456
10498	5	1-2898	10494	16	85	6623	88	2012-08-02 05:02:23.472456
10494	5	1-2961	10494	85	4	5352	78	2012-08-02 05:02:23.472456
10500	3	1-2922	10499	85	43	2952	36	2012-08-02 05:02:25.858944
10501	3	1-2892	10499	43	28	2685	63	2012-08-02 05:02:25.858944
10499	3	1-2967	10499	28	85	5525	18	2012-08-02 05:02:25.858944
10503	6	1-1688	10502	66	15	436	74	2012-08-02 05:02:25.858944
10504	6	1-2951	10502	15	94	441	9	2012-08-02 05:02:25.858944
10505	6	1-2710	10502	94	69	96	35	2012-08-02 05:02:25.858944
10506	6	1-2177	10502	69	93	113	1	2012-08-02 05:02:25.858944
10507	6	1-2417	10502	93	28	175	63	2012-08-02 05:02:25.858944
10502	6	1-2967	10502	28	66	424	18	2012-08-02 05:02:25.858944
10509	3	1-2967	10508	28	16	2216	18	2012-08-02 05:02:27.096169
10510	3	1-2443	10508	16	30	1253	99	2012-08-02 05:02:27.096169
10508	3	1-2970	10508	30	28	1450	63	2012-08-02 05:02:27.096169
10512	4	1-2884	10511	47	21	1636	65	2012-08-02 05:02:27.493789
10513	4	1-2213	10511	21	31	1422	4	2012-08-02 05:02:27.493789
10514	4	1-2831	10511	31	36	2177	10	2012-08-02 05:02:27.493789
10511	4	1-2971	10511	36	47	1738	1	2012-08-02 05:02:27.493789
10516	4	1-2678	10515	94	34	232	91	2012-08-02 05:02:27.493789
10517	4	1-1845	10515	34	84	249	48	2012-08-02 05:02:27.493789
10518	4	1-1439	10515	84	36	248	10	2012-08-02 05:02:27.493789
10515	4	1-2971	10515	36	94	225	1	2012-08-02 05:02:27.493789
10520	5	1-2417	10519	93	36	398	63	2012-08-02 05:02:27.493789
10521	5	1-2851	10519	36	43	476	73	2012-08-02 05:02:27.493789
10522	5	1-1603	10519	43	42	384	71	2012-08-02 05:02:27.493789
10523	5	1-2134	10519	42	36	250	10	2012-08-02 05:02:27.493789
10519	5	1-2971	10519	36	93	245	1	2012-08-02 05:02:27.493789
10525	2	1-1254	10524	35	15	2153	15	2012-08-02 05:02:27.902965
10524	2	1-2972	10524	15	35	9876	91	2012-08-02 05:02:27.902965
10527	5	1-2202	10526	92	15	132	12	2012-08-02 05:02:31.18389
10528	5	1-1883	10526	15	84	116	90	2012-08-02 05:02:31.18389
10529	5	1-334	10526	84	4	88	14	2012-08-02 05:02:31.18389
10530	5	1-2533	10526	4	70	140	2	2012-08-02 05:02:31.18389
10526	5	1-2980	10526	70	92	179	43	2012-08-02 05:02:31.18389
10532	5	1-2931	10531	78	87	169	52	2012-08-02 05:02:31.18389
10533	5	1-2200	10531	87	48	206	59	2012-08-02 05:02:31.18389
10534	5	1-2552	10531	48	98	254	53	2012-08-02 05:02:31.18389
10535	5	1-1901	10531	98	70	95	2	2012-08-02 05:02:31.18389
10531	5	1-2980	10531	70	78	129	43	2012-08-02 05:02:31.18389
10537	6	1-2831	10536	31	36	2549	10	2012-08-02 05:02:31.58168
10538	6	1-2971	10536	36	94	2511	1	2012-08-02 05:02:31.58168
10539	6	1-2678	10536	94	31	2808	91	2012-08-02 05:02:31.58168
10540	6	1-1972	10536	31	26	1766	97	2012-08-02 05:02:31.58168
10541	6	1-1430	10536	26	53	496	62	2012-08-02 05:02:31.58168
10536	6	1-2981	10536	53	31	1346	4	2012-08-02 05:02:31.58168
10543	3	1-2909	10542	32	45	2991	70	2012-08-02 05:02:32.38847
10544	3	1-2650	10542	45	89	2302	48	2012-08-02 05:02:32.38847
10542	3	1-2983	10542	89	32	7534	15	2012-08-02 05:02:32.38847
10546	2	1-523	10545	72	23	2201	74	2012-08-02 05:02:33.570341
10545	2	1-2986	10545	23	72	4532	64	2012-08-02 05:02:33.570341
10548	5	1-2935	10547	100	70	1716	91	2012-08-02 05:02:33.570341
10549	5	1-1623	10547	70	44	1079	87	2012-08-02 05:02:33.570341
10550	5	1-2896	10547	44	86	1069	77	2012-08-02 05:02:33.570341
10551	5	1-580	10547	86	23	316	74	2012-08-02 05:02:33.570341
10547	5	1-2986	10547	23	100	952	64	2012-08-02 05:02:33.570341
10553	5	1-2375	10552	86	57	1384	31	2012-08-02 05:02:33.570341
10554	5	1-2677	10552	57	50	563	28	2012-08-02 05:02:33.570341
10555	5	1-2910	10552	50	66	1994	80	2012-08-02 05:02:33.570341
10556	5	1-1922	10552	66	23	254	74	2012-08-02 05:02:33.570341
10552	5	1-2986	10552	23	86	819	64	2012-08-02 05:02:33.570341
10558	6	1-2906	10557	3	73	674	67	2012-08-02 05:02:34.411944
10559	6	1-2975	10557	73	16	2098	50	2012-08-02 05:02:34.411944
10560	6	1-1934	10557	16	6	527	77	2012-08-02 05:02:34.411944
10561	6	1-2913	10557	6	27	263	39	2012-08-02 05:02:34.411944
10562	6	1-1903	10557	27	19	337	69	2012-08-02 05:02:34.411944
10557	6	1-2988	10557	19	3	364	27	2012-08-02 05:02:34.411944
10564	5	1-2906	10563	3	45	467	67	2012-08-02 05:02:34.411944
10565	5	1-1910	10563	45	10	407	76	2012-08-02 05:02:34.411944
10566	5	1-2940	10563	10	39	204	94	2012-08-02 05:02:34.411944
10567	5	1-2717	10563	39	19	184	69	2012-08-02 05:02:34.411944
10563	5	1-2988	10563	19	3	223	27	2012-08-02 05:02:34.411944
10569	5	1-2906	10568	3	96	692	67	2012-08-02 05:02:34.411944
10570	5	1-2080	10568	96	82	240	92	2012-08-02 05:02:34.411944
10571	5	1-2955	10568	82	39	254	94	2012-08-02 05:02:34.411944
10572	5	1-2717	10568	39	19	243	69	2012-08-02 05:02:34.411944
10568	5	1-2988	10568	19	3	313	27	2012-08-02 05:02:34.411944
10574	5	1-2375	10573	86	25	171	31	2012-08-02 05:02:34.820936
10575	5	1-2475	10573	25	55	172	72	2012-08-02 05:02:34.820936
10576	5	1-2888	10573	55	28	96	21	2012-08-02 05:02:34.820936
10577	5	1-1880	10573	28	81	27	98	2012-08-02 05:02:34.820936
10573	5	1-2989	10573	81	86	142	64	2012-08-02 05:02:34.820936
10579	4	1-2935	10578	100	65	282	91	2012-08-02 05:02:34.820936
10580	4	1-541	10578	65	83	176	52	2012-08-02 05:02:34.820936
10581	4	1-1404	10578	83	81	37	98	2012-08-02 05:02:34.820936
10578	4	1-2989	10578	81	100	202	64	2012-08-02 05:02:34.820936
10583	6	1-2481	10582	2	74	4360	100	2012-08-02 05:02:34.820936
10584	6	1-2386	10582	74	5	5811	17	2012-08-02 05:02:34.820936
10585	6	1-2836	10582	5	58	5411	53	2012-08-02 05:02:34.820936
10586	6	1-2936	10582	58	47	1862	19	2012-08-02 05:02:34.820936
10587	6	1-2231	10582	47	81	1185	98	2012-08-02 05:02:34.820936
10582	6	1-2989	10582	81	2	6713	64	2012-08-02 05:02:34.820936
10589	4	1-2339	10588	49	4	5988	14	2012-08-02 05:02:38.079925
10590	4	1-2533	10588	4	70	6418	2	2012-08-02 05:02:38.079925
10591	4	1-2980	10588	70	57	5563	43	2012-08-02 05:02:38.079925
10588	4	1-2997	10588	57	49	8622	6	2012-08-02 05:02:38.079925
10593	5	1-2625	10592	5	79	215	11	2012-08-02 05:02:38.875532
10594	5	1-2925	10592	79	57	3882	39	2012-08-02 05:02:38.875532
10595	5	1-2292	10592	57	70	1696	3	2012-08-02 05:02:38.875532
10596	5	1-2033	10592	70	85	1742	81	2012-08-02 05:02:38.875532
10592	5	1-2999	10592	85	5	1211	84	2012-08-02 05:02:38.875532
10598	5	1-2688	10597	19	49	1705	41	2012-08-02 05:02:42.565638
10599	5	1-2926	10597	49	44	2299	49	2012-08-02 05:02:42.565638
10600	5	1-1520	10597	44	81	957	94	2012-08-02 05:02:42.565638
10601	5	1-2796	10597	81	39	1294	23	2012-08-02 05:02:42.565638
10597	5	1-3008	10597	39	19	1333	73	2012-08-02 05:02:42.565638
10603	5	1-2826	10602	65	44	287	49	2012-08-02 05:02:44.167298
10604	5	1-1520	10602	44	39	107	94	2012-08-02 05:02:44.167298
10605	5	1-2717	10602	39	53	89	69	2012-08-02 05:02:44.167298
10606	5	1-2451	10602	53	99	92	12	2012-08-02 05:02:44.167298
10602	5	1-3012	10602	99	65	122	96	2012-08-02 05:02:44.167298
10608	5	1-2762	10607	100	53	2703	60	2012-08-02 05:02:44.167298
10609	5	1-2785	10607	53	56	2180	7	2012-08-02 05:02:44.167298
10610	5	1-2934	10607	56	97	5954	23	2012-08-02 05:02:44.167298
10611	5	1-2996	10607	97	99	3034	12	2012-08-02 05:02:44.167298
10607	5	1-3012	10607	99	100	4574	96	2012-08-02 05:02:44.167298
10613	5	1-2166	10612	20	45	1698	70	2012-08-02 05:02:46.144851
10614	5	1-2650	10612	45	40	1724	48	2012-08-02 05:02:46.144851
10615	5	1-2993	10612	40	70	1068	61	2012-08-02 05:02:46.144851
10616	5	1-2864	10612	70	99	2167	16	2012-08-02 05:02:46.144851
10612	5	1-3017	10612	99	20	5101	43	2012-08-02 05:02:46.144851
10618	4	1-2931	10617	78	33	1414	52	2012-08-02 05:02:46.144851
10619	4	1-1405	10617	33	59	915	32	2012-08-02 05:02:46.144851
10620	4	1-2766	10617	59	99	541	16	2012-08-02 05:02:46.144851
10617	4	1-3017	10617	99	78	1450	43	2012-08-02 05:02:46.144851
10622	2	1-1888	10621	24	99	68	16	2012-08-02 05:02:46.144851
10621	2	1-3017	10621	99	24	182	43	2012-08-02 05:02:46.144851
10624	5	1-2320	10623	35	57	30	58	2012-08-02 05:02:49.349011
10625	5	1-1996	10623	57	53	19	69	2012-08-02 05:02:49.349011
10626	5	1-2451	10623	53	99	15	12	2012-08-02 05:02:49.349011
10627	5	1-3012	10623	99	29	16	96	2012-08-02 05:02:49.349011
10623	5	1-3025	10623	29	35	77	18	2012-08-02 05:02:49.349011
10629	6	1-2443	10628	16	53	936	99	2012-08-02 05:02:49.349011
10630	6	1-2837	10628	53	12	1203	31	2012-08-02 05:02:49.349011
10631	6	1-2151	10628	12	48	612	83	2012-08-02 05:02:49.349011
10632	6	1-2113	10628	48	99	280	12	2012-08-02 05:02:49.349011
10633	6	1-3012	10628	99	29	291	96	2012-08-02 05:02:49.349011
10628	6	1-3025	10628	29	16	1428	18	2012-08-02 05:02:49.349011
10635	6	1-2394	10634	9	53	2782	99	2012-08-02 05:02:49.349011
10636	6	1-2837	10634	53	12	3768	31	2012-08-02 05:02:49.349011
10637	6	1-2151	10634	12	48	2022	83	2012-08-02 05:02:49.349011
10638	6	1-2113	10634	48	99	974	12	2012-08-02 05:02:49.349011
10639	6	1-3012	10634	99	29	1068	96	2012-08-02 05:02:49.349011
10634	6	1-3025	10634	29	9	5520	18	2012-08-02 05:02:49.349011
10641	2	1-486	10640	33	56	115	17	2012-08-02 05:02:50.177922
10640	2	1-3027	10640	56	33	86	26	2012-08-02 05:02:50.177922
10643	5	1-1513	10642	18	7	1250	25	2012-08-02 05:02:50.177922
10644	5	1-1666	10642	7	36	654	10	2012-08-02 05:02:50.177922
10645	5	1-2971	10642	36	75	530	1	2012-08-02 05:02:50.177922
10646	5	1-431	10642	75	56	1240	17	2012-08-02 05:02:50.177922
10642	5	1-3027	10642	56	18	958	26	2012-08-02 05:02:50.177922
10648	5	1-2842	10647	25	9	1138	87	2012-08-02 05:02:50.177922
10649	5	1-2833	10647	9	39	3701	40	2012-08-02 05:02:50.177922
10650	5	1-2444	10647	39	74	1345	100	2012-08-02 05:02:50.177922
10651	5	1-2386	10647	74	56	2327	17	2012-08-02 05:02:50.177922
10647	5	1-3027	10647	56	25	1943	26	2012-08-02 05:02:50.177922
10653	3	1-3009	10652	25	61	4125	95	2012-08-02 05:02:50.177922
10654	3	1-2799	10652	61	56	4430	17	2012-08-02 05:02:50.177922
10652	3	1-3027	10652	56	25	3981	26	2012-08-02 05:02:50.177922
10656	3	1-2579	10655	26	16	2354	60	2012-08-02 05:02:51.39313
10657	3	1-1998	10655	16	19	1803	75	2012-08-02 05:02:51.39313
10655	3	1-3030	10655	19	26	5022	46	2012-08-02 05:02:51.39313
10659	5	1-2984	10658	89	47	273	12	2012-08-02 05:02:51.801859
10660	5	1-2437	10658	47	25	376	31	2012-08-02 05:02:51.801859
10661	5	1-2475	10658	25	65	568	72	2012-08-02 05:02:51.801859
10662	5	1-2800	10658	65	66	262	61	2012-08-02 05:02:51.801859
10658	5	1-3031	10658	66	89	397	51	2012-08-02 05:02:51.801859
10664	5	1-2780	10663	87	52	1887	24	2012-08-02 05:02:52.630419
10665	5	1-918	10663	52	98	566	55	2012-08-02 05:02:52.630419
10666	5	1-2891	10663	98	63	845	28	2012-08-02 05:02:52.630419
10667	5	1-1889	10663	63	96	534	29	2012-08-02 05:02:52.630419
10663	5	1-3033	10663	96	87	1515	76	2012-08-02 05:02:52.630419
10668	5	1-3036	10668	26	38	520	89	2012-08-02 05:02:53.835058
10669	5	1-2127	10668	38	70	749	2	2012-08-02 05:02:53.835058
10670	5	1-2980	10668	70	79	973	43	2012-08-02 05:02:53.835058
10671	5	1-2654	10668	79	56	540	79	2012-08-02 05:02:53.835058
10672	5	1-2614	10668	56	26	578	97	2012-08-02 05:02:53.835058
16062	2	1-3884	16061	76	58	1981	48	2012-08-02 05:13:34.27819
10674	6	1-2880	10673	80	9	179	87	2012-08-02 05:02:54.243824
10675	6	1-2833	10673	9	39	529	40	2012-08-02 05:02:54.243824
10676	6	1-2444	10673	39	85	175	100	2012-08-02 05:02:54.243824
10677	6	1-2872	10673	85	6	144	77	2012-08-02 05:02:54.243824
10678	6	1-2913	10673	6	93	69	39	2012-08-02 05:02:54.243824
10673	6	1-3037	10673	93	80	208	65	2012-08-02 05:02:54.243824
16061	2	1-4630	16061	58	76	2896	99	2012-08-02 05:13:34.27819
10680	6	1-1868	10679	89	87	704	70	2012-08-02 05:02:54.243824
10681	6	1-2270	10679	87	46	335	82	2012-08-02 05:02:54.243824
10682	6	1-2942	10679	46	5	169	84	2012-08-02 05:02:54.243824
10683	6	1-2625	10679	5	79	29	11	2012-08-02 05:02:54.243824
10684	6	1-2925	10679	79	93	519	39	2012-08-02 05:02:54.243824
10679	6	1-3037	10679	93	89	1878	65	2012-08-02 05:02:54.243824
10686	4	1-1622	10685	64	61	335	92	2012-08-02 05:02:55.834358
10687	4	1-2792	10685	61	85	361	20	2012-08-02 05:02:55.834358
10688	4	1-2385	10685	85	27	338	100	2012-08-02 05:02:55.834358
10685	4	1-3041	10685	27	64	928	32	2012-08-02 05:02:55.834358
16064	5	1-4544	16063	52	2	2033	98	2012-08-02 05:13:34.27819
10690	4	1-2971	10689	36	94	2691	1	2012-08-02 05:02:56.652062
10691	4	1-2678	10689	94	33	2336	91	2012-08-02 05:02:56.652062
10692	4	1-1491	10689	33	70	998	7	2012-08-02 05:02:56.652062
10689	4	1-3043	10689	70	36	3519	10	2012-08-02 05:02:56.652062
10694	5	1-2971	10693	36	47	563	1	2012-08-02 05:02:56.652062
10695	5	1-2884	10693	47	57	545	65	2012-08-02 05:02:56.652062
10696	5	1-2592	10693	57	85	282	25	2012-08-02 05:02:56.652062
10697	5	1-3028	10693	85	70	181	7	2012-08-02 05:02:56.652062
10693	5	1-3043	10693	70	36	685	10	2012-08-02 05:02:56.652062
10699	5	1-1903	10698	27	19	975	69	2012-08-02 05:02:57.06104
10700	5	1-2988	10698	19	87	794	27	2012-08-02 05:02:57.06104
10701	5	1-2915	10698	87	45	528	70	2012-08-02 05:02:57.06104
10702	5	1-2650	10698	45	77	522	48	2012-08-02 05:02:57.06104
10698	5	1-3044	10698	77	27	1008	39	2012-08-02 05:02:57.06104
10704	5	1-2687	10703	55	36	1531	10	2012-08-02 05:02:57.06104
10705	5	1-2971	10703	36	60	1213	1	2012-08-02 05:02:57.06104
10706	5	1-2807	10703	60	99	346	86	2012-08-02 05:02:57.06104
10707	5	1-2844	10703	99	77	941	48	2012-08-02 05:02:57.06104
10703	5	1-3044	10703	77	55	2310	39	2012-08-02 05:02:57.06104
10709	5	1-2449	10708	66	74	2768	76	2012-08-02 05:02:57.06104
10710	5	1-414	10708	74	87	903	27	2012-08-02 05:02:57.06104
10711	5	1-2915	10708	87	45	814	70	2012-08-02 05:02:57.06104
10712	5	1-2650	10708	45	77	1092	48	2012-08-02 05:02:57.06104
10708	5	1-3044	10708	77	66	2857	39	2012-08-02 05:02:57.06104
10714	5	1-2326	10713	86	74	1476	42	2012-08-02 05:02:57.06104
10715	5	1-2904	10713	74	66	1329	32	2012-08-02 05:02:57.06104
10716	5	1-2908	10713	66	99	221	86	2012-08-02 05:02:57.06104
10717	5	1-2844	10713	99	77	649	48	2012-08-02 05:02:57.06104
10713	5	1-3044	10713	77	86	1717	39	2012-08-02 05:02:57.06104
10719	5	1-2500	10718	39	57	1824	58	2012-08-02 05:02:57.492152
10720	5	1-1996	10718	57	12	1487	69	2012-08-02 05:02:57.492152
10721	5	1-2391	10718	12	86	324	45	2012-08-02 05:02:57.492152
10722	5	1-2824	10718	86	100	1225	97	2012-08-02 05:02:57.492152
10718	5	1-3045	10718	100	39	1399	36	2012-08-02 05:02:57.492152
10724	4	1-2382	10723	8	99	2200	12	2012-08-02 05:02:57.889751
10725	4	1-3012	10723	99	36	1989	96	2012-08-02 05:02:57.889751
10726	4	1-2868	10723	36	57	1097	44	2012-08-02 05:02:57.889751
10723	4	1-3046	10723	57	8	3051	89	2012-08-02 05:02:57.889751
10728	6	1-2615	10727	47	25	2048	52	2012-08-02 05:02:57.889751
10729	6	1-2518	10727	25	48	1234	83	2012-08-02 05:02:57.889751
10730	6	1-2113	10727	48	99	570	12	2012-08-02 05:02:57.889751
10731	6	1-3012	10727	99	36	600	96	2012-08-02 05:02:57.889751
10732	6	1-2868	10727	36	57	385	44	2012-08-02 05:02:57.889751
10727	6	1-3046	10727	57	47	1248	89	2012-08-02 05:02:57.889751
10734	5	1-2817	10733	9	100	505	96	2012-08-02 05:03:00.729277
10735	5	1-2762	10733	100	100	302	60	2012-08-02 05:03:00.729277
10736	5	1-1844	10733	100	49	228	58	2012-08-02 05:03:00.729277
10737	5	1-732	10733	49	44	387	72	2012-08-02 05:03:00.729277
10733	5	1-3053	10733	44	9	516	35	2012-08-02 05:03:00.729277
10739	6	1-2625	10738	5	79	300	11	2012-08-02 05:03:02.739681
10740	6	1-2925	10738	79	66	3874	39	2012-08-02 05:03:02.739681
10741	6	1-2449	10738	66	10	3126	76	2012-08-02 05:03:02.739681
10742	6	1-2940	10738	10	45	1182	94	2012-08-02 05:03:02.739681
10743	6	1-2647	10738	45	84	559	30	2012-08-02 05:03:02.739681
10738	6	1-3058	10738	84	5	2361	84	2012-08-02 05:03:02.739681
10745	5	1-2649	10744	10	6	1780	9	2012-08-02 05:03:02.739681
10746	5	1-2658	10744	6	10	1525	76	2012-08-02 05:03:02.739681
10747	5	1-2940	10744	10	45	753	94	2012-08-02 05:03:02.739681
10748	5	1-2647	10744	45	84	432	30	2012-08-02 05:03:02.739681
10744	5	1-3058	10744	84	10	2216	84	2012-08-02 05:03:02.739681
10750	5	1-1958	10749	22	63	83	86	2012-08-02 05:03:02.739681
10751	5	1-2482	10749	63	16	41	60	2012-08-02 05:03:02.739681
10752	5	1-2347	10749	16	66	27	29	2012-08-02 05:03:02.739681
10753	5	1-2428	10749	66	84	17	30	2012-08-02 05:03:02.739681
10749	5	1-3058	10749	84	22	92	84	2012-08-02 05:03:02.739681
10755	4	1-2897	10754	43	12	948	66	2012-08-02 05:03:04.354577
10756	4	1-2990	10754	12	85	1204	100	2012-08-02 05:03:04.354577
10757	4	1-2872	10754	85	99	1024	77	2012-08-02 05:03:04.354577
10754	4	1-3062	10754	99	43	3573	71	2012-08-02 05:03:04.354577
10759	5	1-2116	10758	85	61	1216	92	2012-08-02 05:03:04.354577
10760	5	1-2792	10758	61	85	1271	20	2012-08-02 05:03:04.354577
10761	5	1-2385	10758	85	85	1277	100	2012-08-02 05:03:04.354577
10762	5	1-2872	10758	85	99	1239	77	2012-08-02 05:03:04.354577
10758	5	1-3062	10758	99	85	4463	71	2012-08-02 05:03:04.354577
10764	6	1-2617	10763	63	85	1924	57	2012-08-02 05:03:04.354577
10765	6	1-837	10763	85	18	411	93	2012-08-02 05:03:04.354577
10766	6	1-2795	10763	18	39	785	40	2012-08-02 05:03:04.354577
10767	6	1-2444	10763	39	85	305	100	2012-08-02 05:03:04.354577
10768	6	1-2872	10763	85	99	303	77	2012-08-02 05:03:04.354577
10763	6	1-3062	10763	99	63	1203	71	2012-08-02 05:03:04.354577
10770	6	1-2874	10769	88	52	97	82	2012-08-02 05:03:04.354577
10771	6	1-2026	10769	52	9	51	87	2012-08-02 05:03:04.354577
10772	6	1-2833	10769	9	39	183	40	2012-08-02 05:03:04.354577
10773	6	1-2444	10769	39	85	73	100	2012-08-02 05:03:04.354577
10774	6	1-2872	10769	85	99	73	77	2012-08-02 05:03:04.354577
10769	6	1-3062	10769	99	88	297	71	2012-08-02 05:03:04.354577
16065	5	1-3632	16063	2	24	2339	8	2012-08-02 05:13:34.27819
10776	5	1-2936	10775	58	47	1408	19	2012-08-02 05:03:07.63729
10777	5	1-2231	10775	47	6	1145	98	2012-08-02 05:03:07.63729
10778	5	1-1291	10775	6	71	365	92	2012-08-02 05:03:07.63729
10779	5	1-3056	10775	71	28	1986	59	2012-08-02 05:03:07.63729
10775	5	1-3070	10775	28	58	3201	53	2012-08-02 05:03:07.63729
16066	5	1-2862	16063	24	24	1997	68	2012-08-02 05:13:34.27819
10781	6	1-2926	10780	49	24	3322	49	2012-08-02 05:03:09.250277
10782	6	1-2012	10780	24	96	2600	46	2012-08-02 05:03:09.250277
10783	6	1-537	10780	96	40	977	48	2012-08-02 05:03:09.250277
10784	6	1-2993	10780	40	88	719	61	2012-08-02 05:03:09.250277
10785	6	1-1729	10780	88	60	98	34	2012-08-02 05:03:09.250277
10780	6	1-3074	10780	60	49	3090	41	2012-08-02 05:03:09.250277
16067	5	1-3309	16063	24	58	596	48	2012-08-02 05:13:34.27819
10787	6	1-2926	10786	49	44	433	49	2012-08-02 05:03:09.250277
10788	6	1-1520	10786	44	81	160	94	2012-08-02 05:03:09.250277
10789	6	1-2796	10786	81	79	192	23	2012-08-02 05:03:09.250277
10790	6	1-2985	10786	79	8	160	95	2012-08-02 05:03:09.250277
10791	6	1-2159	10786	8	60	10	34	2012-08-02 05:03:09.250277
10786	6	1-3074	10786	60	49	362	41	2012-08-02 05:03:09.250277
10793	5	1-2926	10792	49	35	3363	49	2012-08-02 05:03:09.250277
10794	5	1-1906	10792	35	22	2788	55	2012-08-02 05:03:09.250277
10795	5	1-2991	10792	22	8	1065	95	2012-08-02 05:03:09.250277
10796	5	1-2159	10792	8	60	72	34	2012-08-02 05:03:09.250277
10792	5	1-3074	10792	60	49	2671	41	2012-08-02 05:03:09.250277
10798	6	1-1278	10797	95	22	44	84	2012-08-02 05:03:09.250277
10799	6	1-1958	10797	22	99	42	86	2012-08-02 05:03:09.250277
10800	6	1-2844	10797	99	40	140	48	2012-08-02 05:03:09.250277
10801	6	1-2993	10797	40	88	130	61	2012-08-02 05:03:09.250277
10802	6	1-1729	10797	88	60	22	34	2012-08-02 05:03:09.250277
10797	6	1-3074	10797	60	95	891	41	2012-08-02 05:03:09.250277
10803	1	1-537	\N	96	96	734	48	2012-08-02 05:03:09.250277
10805	5	1-2341	10804	11	32	480	15	2012-08-02 05:03:10.488029
10806	5	1-2960	10804	32	52	712	82	2012-08-02 05:03:10.488029
10807	5	1-2026	10804	52	97	373	87	2012-08-02 05:03:10.488029
10808	5	1-2819	10804	97	64	877	59	2012-08-02 05:03:10.488029
10804	5	1-3077	10804	64	11	690	7	2012-08-02 05:03:10.488029
10810	4	1-3011	10809	56	32	1255	15	2012-08-02 05:03:11.327523
10811	4	1-2960	10809	32	44	1731	82	2012-08-02 05:03:11.327523
10812	4	1-2425	10809	44	64	1200	89	2012-08-02 05:03:11.327523
10809	4	1-3079	10809	64	56	1861	47	2012-08-02 05:03:11.327523
10814	4	1-2949	10813	84	56	2011	79	2012-08-02 05:03:12.929741
10815	4	1-2614	10813	56	100	2041	97	2012-08-02 05:03:12.929741
10816	4	1-3045	10813	100	91	2327	36	2012-08-02 05:03:12.929741
10813	4	1-3083	10813	91	84	1581	86	2012-08-02 05:03:12.929741
10818	6	1-2872	10817	85	99	100	77	2012-08-02 05:03:14.54245
10819	6	1-3062	10817	99	88	397	71	2012-08-02 05:03:14.54245
10820	6	1-2874	10817	88	52	126	82	2012-08-02 05:03:14.54245
10821	6	1-2026	10817	52	2	65	87	2012-08-02 05:03:14.54245
10822	6	1-3054	10817	2	2	88	19	2012-08-02 05:03:14.54245
10817	6	1-3087	10817	2	85	103	100	2012-08-02 05:03:14.54245
10824	5	1-2381	10823	60	79	4983	23	2012-08-02 05:03:14.54245
10825	5	1-2985	10823	79	76	4839	95	2012-08-02 05:03:14.54245
10826	5	1-2917	10823	76	2	2815	87	2012-08-02 05:03:14.54245
10827	5	1-3054	10823	2	2	3896	19	2012-08-02 05:03:14.54245
10823	5	1-3087	10823	2	60	4695	100	2012-08-02 05:03:14.54245
10829	4	1-2736	10828	8	10	1755	86	2012-08-02 05:03:15.360152
10830	4	1-3067	10828	10	86	190	45	2012-08-02 05:03:15.360152
10831	4	1-2824	10828	86	29	696	97	2012-08-02 05:03:15.360152
10828	4	1-3089	10828	29	8	3406	98	2012-08-02 05:03:15.360152
10833	6	1-1291	10832	6	82	1567	92	2012-08-02 05:03:15.360152
10834	6	1-2955	10832	82	39	1466	94	2012-08-02 05:03:15.360152
10835	6	1-2717	10832	39	12	1236	69	2012-08-02 05:03:15.360152
10836	6	1-2391	10832	12	86	264	45	2012-08-02 05:03:15.360152
10837	6	1-2824	10832	86	29	981	97	2012-08-02 05:03:15.360152
10832	6	1-3089	10832	29	6	4867	98	2012-08-02 05:03:15.360152
10839	6	1-2920	10838	61	57	730	31	2012-08-02 05:03:15.360152
10840	6	1-2677	10838	57	62	306	28	2012-08-02 05:03:15.360152
10841	6	1-2559	10838	62	58	98	19	2012-08-02 05:03:15.360152
10842	6	1-2902	10838	58	86	24	45	2012-08-02 05:03:15.360152
10843	6	1-2824	10838	86	29	93	97	2012-08-02 05:03:15.360152
10838	6	1-3089	10838	29	61	487	98	2012-08-02 05:03:15.360152
10845	6	1-2783	10844	58	36	176	96	2012-08-02 05:03:15.360152
10846	6	1-2868	10844	36	56	154	44	2012-08-02 05:03:15.360152
10847	6	1-2803	10844	56	51	115	74	2012-08-02 05:03:15.360152
10848	6	1-2722	10844	51	56	93	79	2012-08-02 05:03:15.360152
10849	6	1-2614	10844	56	29	102	97	2012-08-02 05:03:15.360152
10844	6	1-3089	10844	29	58	535	98	2012-08-02 05:03:15.360152
10851	3	1-2649	10850	10	83	978	9	2012-08-02 05:03:16.166885
10852	3	1-1921	10850	83	32	971	4	2012-08-02 05:03:16.166885
10850	3	1-3091	10850	32	10	1503	84	2012-08-02 05:03:16.166885
10854	5	1-1958	10853	22	99	521	86	2012-08-02 05:03:16.166885
10855	5	1-2844	10853	99	40	1552	48	2012-08-02 05:03:16.166885
10856	5	1-2993	10853	40	94	1303	61	2012-08-02 05:03:16.166885
10857	5	1-3026	10853	94	32	333	4	2012-08-02 05:03:16.166885
10853	5	1-3091	10853	32	22	600	84	2012-08-02 05:03:16.166885
10859	3	1-1958	10858	22	61	3336	86	2012-08-02 05:03:16.166885
10860	3	1-1074	10858	61	32	1790	4	2012-08-02 05:03:16.166885
10858	3	1-3091	10858	32	22	3518	84	2012-08-02 05:03:16.166885
10862	5	1-2717	10861	39	53	327	69	2012-08-02 05:03:16.984729
10863	5	1-2451	10861	53	99	276	12	2012-08-02 05:03:16.984729
10864	5	1-3012	10861	99	36	300	96	2012-08-02 05:03:16.984729
10865	5	1-2868	10861	36	100	199	44	2012-08-02 05:03:16.984729
10861	5	1-3093	10861	100	39	483	94	2012-08-02 05:03:16.984729
10867	3	1-2161	10866	46	36	77	96	2012-08-02 05:03:16.984729
10868	3	1-2868	10866	36	100	55	44	2012-08-02 05:03:16.984729
10866	3	1-3093	10866	100	46	146	94	2012-08-02 05:03:16.984729
10870	5	1-2796	10869	81	97	742	23	2012-08-02 05:03:16.984729
10871	5	1-2996	10869	97	99	298	12	2012-08-02 05:03:16.984729
10872	5	1-3012	10869	99	36	354	96	2012-08-02 05:03:16.984729
10873	5	1-2868	10869	36	100	257	44	2012-08-02 05:03:16.984729
10869	5	1-3093	10869	100	81	683	94	2012-08-02 05:03:16.984729
16063	5	1-4630	16063	58	52	893	99	2012-08-02 05:13:34.27819
10875	5	1-2796	10874	81	84	1011	23	2012-08-02 05:03:16.984729
10876	5	1-3038	10874	84	99	318	12	2012-08-02 05:03:16.984729
10877	5	1-3012	10874	99	36	402	96	2012-08-02 05:03:16.984729
10878	5	1-2868	10874	36	100	310	44	2012-08-02 05:03:16.984729
10874	5	1-3093	10874	100	81	876	94	2012-08-02 05:03:16.984729
10880	3	1-1656	10879	93	36	1882	96	2012-08-02 05:03:16.984729
10881	3	1-2868	10879	36	100	1615	44	2012-08-02 05:03:16.984729
10879	3	1-3093	10879	100	93	5083	94	2012-08-02 05:03:16.984729
16069	5	1-4544	16068	52	70	556	98	2012-08-02 05:13:34.27819
10883	4	1-2820	10882	2	70	213	65	2012-08-02 05:03:16.984729
10884	4	1-3061	10882	70	84	266	60	2012-08-02 05:03:16.984729
10885	4	1-2004	10882	84	100	197	44	2012-08-02 05:03:16.984729
10882	4	1-3093	10882	100	2	660	94	2012-08-02 05:03:16.984729
16070	5	1-3292	16068	70	88	336	87	2012-08-02 05:13:34.27819
10887	4	1-1796	10886	82	13	364	54	2012-08-02 05:03:18.619922
10888	4	1-2827	10886	13	78	799	42	2012-08-02 05:03:18.619922
10889	4	1-3066	10886	78	100	101	96	2012-08-02 05:03:18.619922
10886	4	1-3097	10886	100	82	466	59	2012-08-02 05:03:18.619922
16071	5	1-2924	16068	88	80	323	78	2012-08-02 05:13:34.27819
10891	5	1-2782	10890	94	71	1053	92	2012-08-02 05:03:19.437746
10892	5	1-3056	10890	71	53	3350	59	2012-08-02 05:03:19.437746
10893	5	1-2724	10890	53	97	2215	5	2012-08-02 05:03:19.437746
10894	5	1-2664	10890	97	95	1666	88	2012-08-02 05:03:19.437746
10890	5	1-3099	10890	95	94	3436	24	2012-08-02 05:03:19.437746
10896	5	1-1996	10895	57	26	1029	69	2012-08-02 05:03:20.244072
10897	5	1-2713	10895	26	33	1722	21	2012-08-02 05:03:20.244072
10898	5	1-1791	10895	33	91	811	46	2012-08-02 05:03:20.244072
10899	5	1-1322	10895	91	68	558	33	2012-08-02 05:03:20.244072
10895	5	1-3101	10895	68	57	1207	58	2012-08-02 05:03:20.244072
10901	6	1-2649	10900	10	73	372	9	2012-08-02 05:03:20.653182
10902	6	1-2853	10900	73	64	450	91	2012-08-02 05:03:20.653182
10903	6	1-2244	10900	64	43	600	73	2012-08-02 05:03:20.653182
10904	6	1-1603	10900	43	95	493	71	2012-08-02 05:03:20.653182
10905	6	1-3098	10900	95	15	538	21	2012-08-02 05:03:20.653182
10900	6	1-3102	10900	15	10	432	84	2012-08-02 05:03:20.653182
10907	5	1-3076	10906	44	47	2196	67	2012-08-02 05:03:23.514099
10908	5	1-1298	10906	47	54	1670	29	2012-08-02 05:03:23.514099
10909	5	1-2110	10906	54	90	895	24	2012-08-02 05:03:23.514099
10910	5	1-3072	10906	90	10	2252	42	2012-08-02 05:03:23.514099
10906	5	1-3109	10906	10	44	3145	90	2012-08-02 05:03:23.514099
10912	4	1-3076	10911	44	73	1245	67	2012-08-02 05:03:23.514099
10913	4	1-2975	10911	73	67	4635	50	2012-08-02 05:03:23.514099
10914	4	1-2359	10911	67	10	1164	42	2012-08-02 05:03:23.514099
10911	4	1-3109	10911	10	44	1702	90	2012-08-02 05:03:23.514099
10916	5	1-273	10915	56	55	4126	72	2012-08-02 05:03:24.331797
10917	5	1-2888	10915	55	15	3010	21	2012-08-02 05:03:24.331797
10918	5	1-3102	10915	15	26	2030	84	2012-08-02 05:03:24.331797
10919	5	1-3000	10915	26	79	1771	46	2012-08-02 05:03:24.331797
10915	5	1-3111	10915	79	56	6484	36	2012-08-02 05:03:24.331797
10921	5	1-3083	10920	91	48	1610	86	2012-08-02 05:03:24.331797
10922	5	1-1665	10920	48	6	892	70	2012-08-02 05:03:24.331797
10923	5	1-1842	10920	6	59	671	37	2012-08-02 05:03:24.331797
10924	5	1-2715	10920	59	79	671	46	2012-08-02 05:03:24.331797
10920	5	1-3111	10920	79	91	2496	36	2012-08-02 05:03:24.331797
10926	5	1-2572	10925	15	44	593	33	2012-08-02 05:03:24.740824
10927	5	1-1045	10925	44	10	289	76	2012-08-02 05:03:24.740824
10928	5	1-2940	10925	10	81	148	94	2012-08-02 05:03:24.740824
10929	5	1-2796	10925	81	88	200	23	2012-08-02 05:03:24.740824
10925	5	1-3112	10925	88	15	545	47	2012-08-02 05:03:24.740824
10931	5	1-2113	10930	48	15	29	12	2012-08-02 05:03:25.536203
10932	5	1-1883	10930	15	84	24	90	2012-08-02 05:03:25.536203
10933	5	1-334	10930	84	4	17	14	2012-08-02 05:03:25.536203
10934	5	1-2533	10930	4	27	26	2	2012-08-02 05:03:25.536203
10930	5	1-3114	10930	27	48	49	83	2012-08-02 05:03:25.536203
10936	3	1-2759	10935	3	97	2014	87	2012-08-02 05:03:25.944938
10937	3	1-2819	10935	97	7	3909	59	2012-08-02 05:03:25.944938
10935	3	1-3115	10935	7	3	4306	13	2012-08-02 05:03:25.944938
10939	4	1-2891	10938	98	50	2147	28	2012-08-02 05:03:27.171667
10940	4	1-2910	10938	50	99	4882	80	2012-08-02 05:03:27.171667
10941	4	1-2890	10938	99	82	1448	51	2012-08-02 05:03:27.171667
10938	4	1-3118	10938	82	98	2314	55	2012-08-02 05:03:27.171667
10943	5	1-2649	10942	10	37	1498	9	2012-08-02 05:03:28.01093
10944	5	1-2773	10942	37	87	836	68	2012-08-02 05:03:28.01093
10945	5	1-2966	10942	87	27	534	2	2012-08-02 05:03:28.01093
10946	5	1-3114	10942	27	89	402	83	2012-08-02 05:03:28.01093
10942	5	1-3120	10942	89	10	4906	84	2012-08-02 05:03:28.01093
10948	5	1-2401	10947	5	88	976	23	2012-08-02 05:03:28.01093
10949	5	1-3112	10947	88	25	1001	47	2012-08-02 05:03:28.01093
10950	5	1-2791	10947	25	56	389	36	2012-08-02 05:03:28.01093
10951	5	1-3103	10947	56	89	196	83	2012-08-02 05:03:28.01093
10947	5	1-3120	10947	89	5	2460	84	2012-08-02 05:03:28.01093
10953	4	1-2419	10952	34	66	3464	95	2012-08-02 05:03:28.852735
10954	4	1-3039	10952	66	52	4688	52	2012-08-02 05:03:28.852735
10955	4	1-2153	10952	52	28	1748	43	2012-08-02 05:03:28.852735
10952	4	1-3122	10952	28	34	7253	39	2012-08-02 05:03:28.852735
10957	5	1-2449	10956	66	29	267	76	2012-08-02 05:03:28.852735
10958	5	1-1712	10956	29	98	226	53	2012-08-02 05:03:28.852735
10959	5	1-1901	10956	98	70	64	2	2012-08-02 05:03:28.852735
10960	5	1-2980	10956	70	28	65	43	2012-08-02 05:03:28.852735
10956	5	1-3122	10956	28	66	316	39	2012-08-02 05:03:28.852735
10962	5	1-833	10961	100	5	74	17	2012-08-02 05:03:28.852735
10963	5	1-2836	10961	5	98	76	53	2012-08-02 05:03:28.852735
10964	5	1-1901	10961	98	70	22	2	2012-08-02 05:03:28.852735
10965	5	1-2980	10961	70	28	23	43	2012-08-02 05:03:28.852735
10961	5	1-3122	10961	28	100	113	39	2012-08-02 05:03:28.852735
10967	5	1-2326	10966	86	10	181	42	2012-08-02 05:03:28.852735
10968	5	1-3109	10966	10	84	240	90	2012-08-02 05:03:28.852735
10969	5	1-334	10966	84	12	176	14	2012-08-02 05:03:28.852735
10970	5	1-2655	10966	12	28	34	43	2012-08-02 05:03:28.852735
10966	5	1-3122	10966	28	86	201	39	2012-08-02 05:03:28.852735
10972	5	1-833	10971	100	32	341	17	2012-08-02 05:03:28.852735
10973	5	1-2866	10971	32	84	503	90	2012-08-02 05:03:28.852735
10974	5	1-334	10971	84	12	373	14	2012-08-02 05:03:28.852735
10975	5	1-2655	10971	12	28	73	43	2012-08-02 05:03:28.852735
10971	5	1-3122	10971	28	100	433	39	2012-08-02 05:03:28.852735
10976	1	1-334	\N	84	84	10	14	2012-08-02 05:03:28.852735
10978	3	1-2993	10977	40	61	160	61	2012-08-02 05:03:29.261695
10979	3	1-2085	10977	61	72	69	42	2012-08-02 05:03:29.261695
10977	3	1-3123	10977	72	40	537	48	2012-08-02 05:03:29.261695
10981	4	1-2083	10980	87	28	20	43	2012-08-02 05:03:29.261695
10982	4	1-3122	10980	28	86	49	39	2012-08-02 05:03:29.261695
10983	4	1-2326	10980	86	72	19	42	2012-08-02 05:03:29.261695
10980	4	1-3123	10980	72	87	180	48	2012-08-02 05:03:29.261695
10985	5	1-3088	10984	27	71	2728	10	2012-08-02 05:03:29.261695
10986	5	1-3048	10984	71	22	1186	83	2012-08-02 05:03:29.261695
10987	5	1-1191	10984	22	86	246	39	2012-08-02 05:03:29.261695
10988	5	1-2326	10984	86	72	146	42	2012-08-02 05:03:29.261695
10984	5	1-3123	10984	72	27	2175	48	2012-08-02 05:03:29.261695
10990	5	1-2083	10989	87	87	1182	43	2012-08-02 05:03:29.261695
10991	5	1-1895	10989	87	88	349	23	2012-08-02 05:03:29.261695
10992	5	1-3112	10989	88	6	651	47	2012-08-02 05:03:29.261695
10993	5	1-2894	10989	6	72	304	42	2012-08-02 05:03:29.261695
10989	5	1-3123	10989	72	87	5048	48	2012-08-02 05:03:29.261695
10995	4	1-2737	10994	55	47	5	86	2012-08-02 05:03:29.261695
10996	4	1-1307	10994	47	90	2	24	2012-08-02 05:03:29.261695
10997	4	1-3072	10994	90	72	3	42	2012-08-02 05:03:29.261695
10994	4	1-3123	10994	72	55	51	48	2012-08-02 05:03:29.261695
10999	2	1-2974	10998	35	63	2527	51	2012-08-02 05:03:29.681719
10998	2	1-3124	10998	63	35	9015	96	2012-08-02 05:03:29.681719
11001	5	1-2999	11000	85	5	2443	84	2012-08-02 05:03:30.079431
11002	5	1-2401	11000	5	88	1857	23	2012-08-02 05:03:30.079431
11003	5	1-3112	11000	88	100	3647	47	2012-08-02 05:03:30.079431
11004	5	1-2216	11000	100	84	1964	5	2012-08-02 05:03:30.079431
11000	5	1-3125	11000	84	85	5030	81	2012-08-02 05:03:30.079431
11006	5	1-2745	11005	77	63	135	8	2012-08-02 05:03:30.079431
11007	5	1-690	11005	63	71	138	92	2012-08-02 05:03:30.079431
11008	5	1-3056	11005	71	53	633	59	2012-08-02 05:03:30.079431
11009	5	1-2724	11005	53	84	602	5	2012-08-02 05:03:30.079431
11005	5	1-3125	11005	84	77	1628	81	2012-08-02 05:03:30.079431
11011	5	1-2745	11010	77	72	97	8	2012-08-02 05:03:30.079431
11012	5	1-429	11010	72	55	145	72	2012-08-02 05:03:30.079431
11013	5	1-2888	11010	55	89	105	21	2012-08-02 05:03:30.079431
11014	5	1-2596	11010	89	84	335	5	2012-08-02 05:03:30.079431
11010	5	1-3125	11010	84	77	1028	81	2012-08-02 05:03:30.079431
11016	5	1-3081	11015	5	70	82	65	2012-08-02 05:03:30.477471
11017	5	1-3061	11015	70	16	97	60	2012-08-02 05:03:30.477471
11018	5	1-2347	11015	16	59	66	29	2012-08-02 05:03:30.477471
11019	5	1-3052	11015	59	17	59	67	2012-08-02 05:03:30.477471
11015	5	1-3126	11015	17	5	126	25	2012-08-02 05:03:30.477471
11021	6	1-3028	11020	85	16	1458	7	2012-08-02 05:03:30.477471
11022	6	1-1380	11020	16	21	1719	85	2012-08-02 05:03:30.477471
11023	6	1-2680	11020	21	15	1268	47	2012-08-02 05:03:30.477471
11024	6	1-2572	11020	15	13	1312	33	2012-08-02 05:03:30.477471
11025	6	1-2758	11020	13	17	966	67	2012-08-02 05:03:30.477471
11020	6	1-3126	11020	17	85	2064	25	2012-08-02 05:03:30.477471
11027	5	1-3097	11026	100	82	2567	59	2012-08-02 05:03:31.295058
11028	5	1-1796	11026	82	82	1936	54	2012-08-02 05:03:31.295058
11029	5	1-2365	11026	82	72	137	8	2012-08-02 05:03:31.295058
11030	5	1-429	11026	72	97	182	72	2012-08-02 05:03:31.295058
11026	5	1-3128	11026	97	100	671	96	2012-08-02 05:03:31.295058
11032	5	1-2762	11031	100	16	266	60	2012-08-02 05:03:31.295058
11033	5	1-1998	11031	16	17	172	75	2012-08-02 05:03:31.295058
11034	5	1-1854	11031	17	72	85	8	2012-08-02 05:03:31.295058
11035	5	1-429	11031	72	97	127	72	2012-08-02 05:03:31.295058
11031	5	1-3128	11031	97	100	531	96	2012-08-02 05:03:31.295058
11037	4	1-2868	11036	36	11	173	44	2012-08-02 05:03:31.295058
11038	4	1-889	11036	11	11	46	64	2012-08-02 05:03:31.295058
11039	4	1-2540	11036	11	97	46	72	2012-08-02 05:03:31.295058
11036	4	1-3128	11036	97	36	206	96	2012-08-02 05:03:31.295058
11041	5	1-2992	11040	84	17	3296	3	2012-08-02 05:03:31.295058
11042	5	1-1226	11040	17	94	2230	43	2012-08-02 05:03:31.295058
11043	5	1-3003	11040	94	72	511	8	2012-08-02 05:03:31.295058
11044	5	1-429	11040	72	97	864	72	2012-08-02 05:03:31.295058
11040	5	1-3128	11040	97	84	4078	96	2012-08-02 05:03:31.295058
11046	5	1-2970	11045	30	58	7724	63	2012-08-02 05:03:31.726448
11047	5	1-2479	11045	58	5	6840	25	2012-08-02 05:03:31.726448
11048	5	1-3081	11045	5	70	4433	65	2012-08-02 05:03:31.726448
11049	5	1-3061	11045	70	72	5189	60	2012-08-02 05:03:31.726448
11045	5	1-3129	11045	72	30	4260	99	2012-08-02 05:03:31.726448
11051	5	1-2764	11050	57	87	1554	68	2012-08-02 05:03:32.53258
11052	5	1-2966	11050	87	27	2416	2	2012-08-02 05:03:32.53258
11053	5	1-3114	11050	27	94	4419	83	2012-08-02 05:03:32.53258
11054	5	1-1695	11050	94	13	3126	5	2012-08-02 05:03:32.53258
11050	5	1-3131	11050	13	57	2146	11	2012-08-02 05:03:32.53258
11056	5	1-1698	11055	25	58	1927	98	2012-08-02 05:03:32.53258
11057	5	1-2783	11055	58	100	591	96	2012-08-02 05:03:32.53258
11058	5	1-3097	11055	100	53	2678	59	2012-08-02 05:03:32.53258
11059	5	1-2724	11055	53	13	3013	5	2012-08-02 05:03:32.53258
11055	5	1-3131	11055	13	25	2090	11	2012-08-02 05:03:32.53258
11061	5	1-1505	11060	86	56	576	44	2012-08-02 05:03:32.53258
11062	5	1-2803	11060	56	15	401	74	2012-08-02 05:03:32.53258
11063	5	1-2951	11060	15	23	397	9	2012-08-02 05:03:32.53258
11064	5	1-2636	11060	23	13	995	5	2012-08-02 05:03:32.53258
11060	5	1-3131	11060	13	86	722	11	2012-08-02 05:03:32.53258
11066	5	1-1055	11065	2	6	1403	20	2012-08-02 05:03:32.53258
11067	5	1-2619	11065	6	10	868	73	2012-08-02 05:03:32.53258
11068	5	1-3055	11065	10	89	436	21	2012-08-02 05:03:32.53258
11069	5	1-2596	11065	89	13	1549	5	2012-08-02 05:03:32.53258
11065	5	1-3131	11065	13	2	1151	11	2012-08-02 05:03:32.53258
11071	3	1-1298	11070	47	66	884	29	2012-08-02 05:03:32.930771
11072	3	1-2428	11070	66	27	555	30	2012-08-02 05:03:32.930771
11070	3	1-3132	11070	27	47	1199	67	2012-08-02 05:03:32.930771
11074	6	1-2146	11073	98	11	1857	23	2012-08-02 05:03:33.33962
11075	6	1-3119	11073	11	68	1764	33	2012-08-02 05:03:33.33962
11076	6	1-3101	11073	68	57	3816	58	2012-08-02 05:03:33.33962
11077	6	1-1996	11073	57	19	3251	69	2012-08-02 05:03:33.33962
11078	6	1-2988	11073	19	37	3936	27	2012-08-02 05:03:33.33962
11073	6	1-3133	11073	37	98	1822	56	2012-08-02 05:03:33.33962
11079	5	1-3134	11079	22	21	157	85	2012-08-02 05:03:33.737257
11080	5	1-2680	11079	21	15	99	47	2012-08-02 05:03:33.737257
11081	5	1-2572	11079	15	13	88	33	2012-08-02 05:03:33.737257
11082	5	1-2758	11079	13	73	56	67	2012-08-02 05:03:33.737257
11083	5	1-2975	11079	73	22	164	50	2012-08-02 05:03:33.737257
16072	5	1-3603	16068	80	58	136	48	2012-08-02 05:13:34.27819
11085	5	1-3000	11084	26	13	2648	46	2012-08-02 05:03:34.135026
11086	5	1-2889	11084	13	97	1063	87	2012-08-02 05:03:34.135026
11087	5	1-2819	11084	97	28	1748	59	2012-08-02 05:03:34.135026
11088	5	1-3070	11084	28	29	2198	53	2012-08-02 05:03:34.135026
11084	5	1-3135	11084	29	26	3726	84	2012-08-02 05:03:34.135026
16068	5	1-4630	16068	58	52	223	99	2012-08-02 05:13:34.27819
11090	2	1-1820	11089	66	62	1901	95	2012-08-02 05:03:35.328242
11089	2	1-3138	11089	62	66	3953	45	2012-08-02 05:03:35.328242
11092	5	1-985	11091	61	85	457	62	2012-08-02 05:03:35.328242
11093	5	1-2835	11091	85	67	1540	91	2012-08-02 05:03:35.328242
11094	5	1-3121	11091	67	34	1173	39	2012-08-02 05:03:35.328242
11095	5	1-2419	11091	34	62	552	95	2012-08-02 05:03:35.328242
11091	5	1-3138	11091	62	61	1372	45	2012-08-02 05:03:35.328242
11097	5	1-2824	11096	86	100	1840	97	2012-08-02 05:03:35.328242
11098	5	1-3045	11096	100	17	1709	36	2012-08-02 05:03:35.328242
11099	5	1-306	11096	17	79	293	23	2012-08-02 05:03:35.328242
11100	5	1-2985	11096	79	62	208	95	2012-08-02 05:03:35.328242
11096	5	1-3138	11096	62	86	598	45	2012-08-02 05:03:35.328242
11102	5	1-985	11101	61	85	337	62	2012-08-02 05:03:35.328242
11103	5	1-2835	11101	85	65	1482	91	2012-08-02 05:03:35.328242
11104	5	1-541	11101	65	59	1167	52	2012-08-02 05:03:35.328242
11105	5	1-699	11101	59	62	239	95	2012-08-02 05:03:35.328242
11101	5	1-3138	11101	62	61	775	45	2012-08-02 05:03:35.328242
11107	4	1-3071	11106	94	17	5970	36	2012-08-02 05:03:35.328242
11108	4	1-306	11106	17	79	1189	23	2012-08-02 05:03:35.328242
11109	4	1-2985	11106	79	62	982	95	2012-08-02 05:03:35.328242
11106	4	1-3138	11106	62	94	3289	45	2012-08-02 05:03:35.328242
11110	1	1-985	\N	61	61	1817	62	2012-08-02 05:03:35.328242
11112	5	1-1322	11111	91	57	437	33	2012-08-02 05:03:35.759462
11113	5	1-2886	11111	57	57	257	58	2012-08-02 05:03:35.759462
11114	5	1-1996	11111	57	18	228	69	2012-08-02 05:03:35.759462
11115	5	1-2927	11111	18	67	289	68	2012-08-02 05:03:35.759462
11111	5	1-3139	11111	67	91	625	46	2012-08-02 05:03:35.759462
11117	5	1-2889	11116	13	2	2518	87	2012-08-02 05:03:35.759462
11118	5	1-3054	11116	2	94	3327	19	2012-08-02 05:03:35.759462
11119	5	1-1674	11116	94	37	1371	9	2012-08-02 05:03:35.759462
11120	5	1-2773	11116	37	67	2064	68	2012-08-02 05:03:35.759462
11116	5	1-3139	11116	67	13	4478	46	2012-08-02 05:03:35.759462
11122	5	1-2579	11121	26	100	743	60	2012-08-02 05:03:35.759462
11123	5	1-1844	11121	100	57	564	58	2012-08-02 05:03:35.759462
11124	5	1-1996	11121	57	18	513	69	2012-08-02 05:03:35.759462
11125	5	1-2927	11121	18	67	683	68	2012-08-02 05:03:35.759462
11121	5	1-3139	11121	67	26	1554	46	2012-08-02 05:03:35.759462
11126	1	1-1413	\N	64	64	1264	69	2012-08-02 05:03:35.759462
11128	5	1-3085	11127	58	15	24	74	2012-08-02 05:03:37.781213
11129	5	1-2951	11127	15	45	20	9	2012-08-02 05:03:37.781213
11130	5	1-2730	11127	45	32	13	17	2012-08-02 05:03:37.781213
11131	5	1-2866	11127	32	38	16	90	2012-08-02 05:03:37.781213
11127	5	1-3144	11127	38	58	9	38	2012-08-02 05:03:37.781213
11133	5	1-3085	11132	58	51	80	74	2012-08-02 05:03:37.781213
11134	5	1-2722	11132	51	81	57	79	2012-08-02 05:03:37.781213
11135	5	1-1744	11132	81	55	51	4	2012-08-02 05:03:37.781213
11136	5	1-2609	11132	55	38	46	90	2012-08-02 05:03:37.781213
11132	5	1-3144	11132	38	58	28	38	2012-08-02 05:03:37.781213
11138	5	1-3085	11137	58	15	3414	74	2012-08-02 05:03:37.781213
11139	5	1-2951	11137	15	96	3314	9	2012-08-02 05:03:37.781213
11140	5	1-2111	11137	96	89	3009	85	2012-08-02 05:03:37.781213
11141	5	1-2242	11137	89	38	1708	90	2012-08-02 05:03:37.781213
11137	5	1-3144	11137	38	58	1110	38	2012-08-02 05:03:37.781213
11143	5	1-1722	11142	61	44	521	72	2012-08-02 05:03:37.781213
11144	5	1-3053	11142	44	9	631	35	2012-08-02 05:03:37.781213
11145	5	1-2817	11142	9	58	560	96	2012-08-02 05:03:37.781213
11146	5	1-1829	11142	58	38	943	90	2012-08-02 05:03:37.781213
11142	5	1-3144	11142	38	61	615	38	2012-08-02 05:03:37.781213
11148	6	1-2708	11147	84	60	702	35	2012-08-02 05:03:38.985453
11149	6	1-1585	11147	60	15	612	74	2012-08-02 05:03:38.985453
11150	6	1-2951	11147	15	37	600	9	2012-08-02 05:03:38.985453
11151	6	1-2773	11147	37	67	852	68	2012-08-02 05:03:38.985453
11152	6	1-3139	11147	67	12	1744	46	2012-08-02 05:03:38.985453
11147	6	1-3147	11147	12	84	2717	7	2012-08-02 05:03:38.985453
11154	5	1-2341	11153	11	32	829	15	2012-08-02 05:03:38.985453
11155	5	1-2909	11153	32	6	665	70	2012-08-02 05:03:38.985453
11156	5	1-1842	11153	6	59	584	37	2012-08-02 05:03:38.985453
11157	5	1-2715	11153	59	12	681	46	2012-08-02 05:03:38.985453
11153	5	1-3147	11153	12	11	1170	7	2012-08-02 05:03:38.985453
11159	5	1-2614	11158	56	100	172	97	2012-08-02 05:03:40.554418
11160	5	1-3045	11158	100	43	200	36	2012-08-02 05:03:40.554418
11161	5	1-2892	11158	43	2	211	63	2012-08-02 05:03:40.554418
11162	5	1-3107	11158	2	46	226	22	2012-08-02 05:03:40.554418
11158	5	1-3151	11158	46	56	167	79	2012-08-02 05:03:40.554418
11164	6	1-1970	11163	98	64	802	89	2012-08-02 05:03:40.974809
11165	6	1-3079	11163	64	15	1258	47	2012-08-02 05:03:40.974809
11166	6	1-2572	11163	15	67	1297	33	2012-08-02 05:03:40.974809
11167	6	1-3005	11163	67	47	895	20	2012-08-02 05:03:40.974809
11168	6	1-2704	11163	47	31	1644	6	2012-08-02 05:03:40.974809
11163	6	1-3152	11163	31	98	839	34	2012-08-02 05:03:40.974809
11170	4	1-2224	11169	23	77	751	44	2012-08-02 05:03:40.974809
11171	4	1-2175	11169	77	47	706	20	2012-08-02 05:03:40.974809
11172	4	1-2704	11169	47	31	1365	6	2012-08-02 05:03:40.974809
11169	4	1-3152	11169	31	23	733	34	2012-08-02 05:03:40.974809
11174	5	1-3053	11173	44	69	38	35	2012-08-02 05:03:41.383264
11175	5	1-2177	11173	69	75	43	1	2012-08-02 05:03:41.383264
11176	5	1-431	11173	75	39	110	17	2012-08-02 05:03:41.383264
11177	5	1-2201	11173	39	15	118	42	2012-08-02 05:03:41.383264
11173	5	1-3153	11173	15	44	31	72	2012-08-02 05:03:41.383264
11179	4	1-2796	11178	81	93	3170	23	2012-08-02 05:03:41.781072
11180	4	1-803	11178	93	38	732	89	2012-08-02 05:03:41.781072
11181	4	1-2127	11178	38	91	963	2	2012-08-02 05:03:41.781072
11178	4	1-3154	11178	91	81	2616	94	2012-08-02 05:03:41.781072
11183	4	1-2820	11182	2	70	1430	65	2012-08-02 05:03:41.781072
11184	4	1-3061	11182	70	83	1715	60	2012-08-02 05:03:41.781072
11185	4	1-933	11182	83	91	1560	2	2012-08-02 05:03:41.781072
11182	4	1-3154	11182	91	2	4603	94	2012-08-02 05:03:41.781072
16073	1	1-2261	\N	84	84	2905	68	2012-08-02 05:13:34.27819
11187	5	1-2970	11186	30	33	251	63	2012-08-02 05:03:43.007286
11188	5	1-3080	11186	33	43	191	93	2012-08-02 05:03:43.007286
11189	5	1-2441	11186	43	47	95	20	2012-08-02 05:03:43.007286
11190	5	1-2704	11186	47	8	132	6	2012-08-02 05:03:43.007286
11186	5	1-3157	11186	8	30	183	99	2012-08-02 05:03:43.007286
11192	4	1-1019	11191	56	43	159	93	2012-08-02 05:03:43.007286
11193	4	1-2441	11191	43	47	112	20	2012-08-02 05:03:43.007286
11194	4	1-2704	11191	47	8	220	6	2012-08-02 05:03:43.007286
11191	4	1-3157	11191	8	56	429	99	2012-08-02 05:03:43.007286
16075	3	1-3754	16074	40	62	1480	36	2012-08-02 05:13:34.698182
11196	5	1-1415	11195	39	91	594	2	2012-08-02 05:03:43.83615
11197	5	1-3154	11195	91	48	916	94	2012-08-02 05:03:43.83615
11198	5	1-1161	11195	48	16	143	10	2012-08-02 05:03:43.83615
11199	5	1-2898	11195	16	77	695	88	2012-08-02 05:03:43.83615
11195	5	1-3159	11195	77	39	624	75	2012-08-02 05:03:43.83615
16076	3	1-4430	16074	62	46	651	39	2012-08-02 05:13:34.698182
11201	5	1-2406	11200	19	48	184	9	2012-08-02 05:03:43.83615
11202	5	1-2405	11200	48	27	155	48	2012-08-02 05:03:43.83615
11203	5	1-3088	11200	27	16	197	10	2012-08-02 05:03:43.83615
11204	5	1-2898	11200	16	77	1182	88	2012-08-02 05:03:43.83615
11200	5	1-3159	11200	77	19	1308	75	2012-08-02 05:03:43.83615
11206	4	1-2268	11205	96	50	803	28	2012-08-02 05:03:43.83615
11207	4	1-2910	11205	50	39	1932	80	2012-08-02 05:03:43.83615
11208	4	1-782	11205	39	77	1167	88	2012-08-02 05:03:43.83615
11205	4	1-3159	11205	77	96	1309	75	2012-08-02 05:03:43.83615
11210	3	1-2507	11209	56	9	1115	61	2012-08-02 05:03:43.83615
11211	3	1-1502	11209	9	77	1523	88	2012-08-02 05:03:43.83615
11209	3	1-3159	11209	77	56	1917	75	2012-08-02 05:03:43.83615
11213	6	1-2414	11212	66	44	600	72	2012-08-02 05:03:45.858198
11214	6	1-3053	11212	44	5	756	35	2012-08-02 05:03:45.858198
11215	6	1-3042	11212	5	98	1151	55	2012-08-02 05:03:45.858198
11216	6	1-2891	11212	98	63	1718	28	2012-08-02 05:03:45.858198
11217	6	1-1889	11212	63	62	1086	29	2012-08-02 05:03:45.858198
11212	6	1-3164	11212	62	66	5208	51	2012-08-02 05:03:45.858198
11219	5	1-1924	11218	89	98	1452	55	2012-08-02 05:03:49.1507
11220	5	1-2891	11218	98	63	2155	28	2012-08-02 05:03:49.1507
11221	5	1-1889	11218	63	96	1355	29	2012-08-02 05:03:49.1507
11222	5	1-3033	11218	96	8	3821	76	2012-08-02 05:03:49.1507
11218	5	1-3172	11218	8	89	4078	61	2012-08-02 05:03:49.1507
11224	3	1-2939	11223	89	61	4875	98	2012-08-02 05:03:49.55398
11225	3	1-2920	11223	61	94	2717	31	2012-08-02 05:03:49.55398
11223	3	1-3173	11223	94	89	6559	30	2012-08-02 05:03:49.55398
11227	3	1-195	11226	34	70	2123	3	2012-08-02 05:03:50.770866
11228	3	1-2033	11226	70	64	1499	81	2012-08-02 05:03:50.770866
11226	3	1-3176	11226	64	34	9136	26	2012-08-02 05:03:50.770866
11230	5	1-2855	11229	86	85	1303	18	2012-08-02 05:03:51.567005
11231	5	1-3169	11229	85	87	3382	27	2012-08-02 05:03:51.567005
11232	5	1-2915	11229	87	45	2863	70	2012-08-02 05:03:51.567005
11233	5	1-2650	11229	45	29	3604	48	2012-08-02 05:03:51.567005
11229	5	1-3178	11229	29	86	2998	91	2012-08-02 05:03:51.567005
11235	2	1-1845	11234	34	29	593	48	2012-08-02 05:03:51.567005
11234	2	1-3178	11234	29	34	559	91	2012-08-02 05:03:51.567005
11237	2	1-399	11236	42	29	2653	48	2012-08-02 05:03:51.567005
11236	2	1-3178	11236	29	42	2557	91	2012-08-02 05:03:51.567005
11239	2	1-2481	11238	2	28	1161	100	2012-08-02 05:03:52.750688
11238	2	1-3181	11238	28	2	1409	64	2012-08-02 05:03:52.750688
11241	5	1-2540	11240	11	44	1996	72	2012-08-02 05:03:52.750688
11242	5	1-3053	11240	44	60	2390	35	2012-08-02 05:03:52.750688
11243	5	1-1021	11240	60	12	1300	66	2012-08-02 05:03:52.750688
11244	5	1-2990	11240	12	28	1749	100	2012-08-02 05:03:52.750688
11240	5	1-3181	11240	28	11	2167	64	2012-08-02 05:03:52.750688
11246	5	1-2540	11245	11	44	813	72	2012-08-02 05:03:52.750688
11247	5	1-3053	11245	44	60	1095	35	2012-08-02 05:03:52.750688
11248	5	1-1585	11245	60	84	1055	74	2012-08-02 05:03:52.750688
11249	5	1-1551	11245	84	28	563	100	2012-08-02 05:03:52.750688
11245	5	1-3181	11245	28	11	785	64	2012-08-02 05:03:52.750688
11251	5	1-2871	11250	26	3	1238	13	2012-08-02 05:03:53.559694
11252	5	1-2759	11250	3	75	700	87	2012-08-02 05:03:53.559694
11253	5	1-2598	11250	75	8	871	76	2012-08-02 05:03:53.559694
11254	5	1-3172	11250	8	61	976	61	2012-08-02 05:03:53.559694
11250	5	1-3183	11250	61	26	1358	28	2012-08-02 05:03:53.559694
11256	3	1-2507	11255	56	61	686	61	2012-08-02 05:03:54.39071
11257	3	1-2085	11255	61	78	631	42	2012-08-02 05:03:54.39071
11255	3	1-3185	11255	78	56	1217	75	2012-08-02 05:03:54.39071
11259	5	1-1513	11258	18	85	115	25	2012-08-02 05:03:55.208719
11260	5	1-3028	11258	85	16	85	7	2012-08-02 05:03:55.208719
11261	5	1-1380	11258	16	21	105	85	2012-08-02 05:03:55.208719
11262	5	1-2680	11258	21	25	81	47	2012-08-02 05:03:55.208719
11258	5	1-3187	11258	25	18	75	26	2012-08-02 05:03:55.208719
11264	4	1-2522	11263	65	9	214	35	2012-08-02 05:03:56.038449
11265	4	1-2817	11263	9	100	100	96	2012-08-02 05:03:56.038449
11266	4	1-3097	11263	100	20	247	59	2012-08-02 05:03:56.038449
11263	4	1-3189	11263	20	65	2000	22	2012-08-02 05:03:56.038449
11268	4	1-3151	11267	46	40	440	79	2012-08-02 05:03:56.038449
11269	4	1-2331	11267	40	88	219	33	2012-08-02 05:03:56.038449
11270	4	1-2535	11267	88	20	95	59	2012-08-02 05:03:56.038449
11267	4	1-3189	11267	20	46	941	22	2012-08-02 05:03:56.038449
11272	6	1-1561	11271	72	81	1241	16	2012-08-02 05:03:58.016933
11273	6	1-3141	11271	81	21	1356	85	2012-08-02 05:03:58.016933
11274	6	1-2680	11271	21	25	876	47	2012-08-02 05:03:58.016933
11275	6	1-3187	11271	25	18	681	26	2012-08-02 05:03:58.016933
11276	6	1-1513	11271	18	49	871	25	2012-08-02 05:03:58.016933
11271	6	1-3194	11271	49	72	1165	11	2012-08-02 05:03:58.016933
11278	4	1-1505	11277	86	56	2443	44	2012-08-02 05:03:58.016933
11279	4	1-2803	11277	56	80	1675	74	2012-08-02 05:03:58.016933
11280	4	1-2732	11277	80	49	2095	25	2012-08-02 05:03:58.016933
11277	4	1-3194	11277	49	86	3113	11	2012-08-02 05:03:58.016933
11282	5	1-2788	11281	72	13	5337	57	2012-08-02 05:03:58.016933
11283	5	1-981	11281	13	8	1151	6	2012-08-02 05:03:58.016933
11284	5	1-3157	11281	8	75	2052	99	2012-08-02 05:03:58.016933
11285	5	1-1182	11281	75	49	1105	25	2012-08-02 05:03:58.016933
11281	5	1-3194	11281	49	72	1646	11	2012-08-02 05:03:58.016933
11287	5	1-2788	11286	72	46	3010	57	2012-08-02 05:03:58.016933
11288	5	1-2645	11286	46	4	2803	78	2012-08-02 05:03:58.016933
11289	5	1-2814	11286	4	12	779	95	2012-08-02 05:03:58.016933
11290	5	1-2366	11286	12	49	581	25	2012-08-02 05:03:58.016933
11286	5	1-3194	11286	49	72	896	11	2012-08-02 05:03:58.016933
16074	3	1-4631	16074	46	40	9529	6	2012-08-02 05:13:34.698182
11292	5	1-1571	11291	94	94	834	45	2012-08-02 05:03:58.016933
11293	5	1-3071	11291	94	12	1730	36	2012-08-02 05:03:58.016933
11294	5	1-3162	11291	12	57	1563	65	2012-08-02 05:03:58.016933
11295	5	1-2592	11291	57	49	915	25	2012-08-02 05:03:58.016933
11291	5	1-3194	11291	49	94	1429	11	2012-08-02 05:03:58.016933
11297	5	1-2425	11296	44	47	726	89	2012-08-02 05:03:59.234081
11298	5	1-2615	11296	47	87	1118	52	2012-08-02 05:03:59.234081
11299	5	1-2200	11296	87	28	889	59	2012-08-02 05:03:59.234081
11300	5	1-3070	11296	28	6	1057	53	2012-08-02 05:03:59.234081
11296	5	1-3197	11296	6	44	1470	82	2012-08-02 05:03:59.234081
11302	4	1-2425	11301	44	47	87	89	2012-08-02 05:03:59.234081
11303	4	1-2615	11301	47	51	147	52	2012-08-02 05:03:59.234081
11304	4	1-701	11301	51	6	105	53	2012-08-02 05:03:59.234081
11301	4	1-3197	11301	6	44	160	82	2012-08-02 05:03:59.234081
11306	4	1-2425	11305	44	23	2520	89	2012-08-02 05:03:59.234081
11307	4	1-3140	11305	23	5	3015	17	2012-08-02 05:03:59.234081
11308	4	1-2836	11305	5	6	2951	53	2012-08-02 05:03:59.234081
11305	4	1-3197	11305	6	44	4577	82	2012-08-02 05:03:59.234081
11310	4	1-2026	11309	52	75	1567	87	2012-08-02 05:03:59.234081
11311	4	1-2598	11309	75	29	1760	76	2012-08-02 05:03:59.234081
11312	4	1-1712	11309	29	6	1753	53	2012-08-02 05:03:59.234081
11309	4	1-3197	11309	6	52	3321	82	2012-08-02 05:03:59.234081
11314	5	1-2893	11313	5	13	288	57	2012-08-02 05:03:59.234081
11315	5	1-981	11313	13	8	61	6	2012-08-02 05:03:59.234081
11316	5	1-3157	11313	8	45	108	99	2012-08-02 05:03:59.234081
11317	5	1-1402	11313	45	6	137	53	2012-08-02 05:03:59.234081
11313	5	1-3197	11313	6	5	262	82	2012-08-02 05:03:59.234081
11319	5	1-3078	11318	10	96	7574	16	2012-08-02 05:03:59.632599
11320	5	1-3174	11318	96	79	4256	23	2012-08-02 05:03:59.632599
11321	5	1-2985	11318	79	61	3301	95	2012-08-02 05:03:59.632599
11322	5	1-2799	11318	61	92	2957	17	2012-08-02 05:03:59.632599
11318	5	1-3198	11318	92	10	3707	79	2012-08-02 05:03:59.632599
11324	5	1-3078	11323	10	96	782	16	2012-08-02 05:03:59.632599
11325	5	1-3174	11323	96	93	526	23	2012-08-02 05:03:59.632599
11326	5	1-803	11323	93	23	134	89	2012-08-02 05:03:59.632599
11327	5	1-3140	11323	23	92	213	17	2012-08-02 05:03:59.632599
11323	5	1-3198	11323	92	10	320	79	2012-08-02 05:03:59.632599
11329	5	1-1700	11328	31	89	659	10	2012-08-02 05:03:59.632599
11330	5	1-3182	11328	89	35	310	49	2012-08-02 05:03:59.632599
11331	5	1-1906	11328	35	10	282	55	2012-08-02 05:03:59.632599
11332	5	1-2794	11328	10	92	288	17	2012-08-02 05:03:59.632599
11328	5	1-3198	11328	92	31	448	79	2012-08-02 05:03:59.632599
11334	2	1-2645	11333	46	14	13	78	2012-08-02 05:04:00.848431
11333	2	1-3201	11333	14	46	56	57	2012-08-02 05:04:00.848431
11336	4	1-2388	11335	76	22	100	50	2012-08-02 05:04:00.848431
11337	4	1-3134	11335	22	92	77	85	2012-08-02 05:04:00.848431
11338	4	1-2457	11335	92	14	15	78	2012-08-02 05:04:00.848431
11335	4	1-3201	11335	14	76	185	57	2012-08-02 05:04:00.848431
11340	4	1-2674	11339	52	99	1291	80	2012-08-02 05:04:00.848431
11341	4	1-2890	11339	99	31	435	51	2012-08-02 05:04:00.848431
11342	4	1-2952	11339	31	14	185	78	2012-08-02 05:04:00.848431
11339	4	1-3201	11339	14	52	2284	57	2012-08-02 05:04:00.848431
11344	5	1-2977	11343	85	70	108	3	2012-08-02 05:04:00.848431
11345	5	1-2033	11343	70	93	79	81	2012-08-02 05:04:00.848431
11346	5	1-2728	11343	93	31	49	51	2012-08-02 05:04:00.848431
11347	5	1-2952	11343	31	14	23	78	2012-08-02 05:04:00.848431
11343	5	1-3201	11343	14	85	301	57	2012-08-02 05:04:00.848431
11349	4	1-2973	11348	100	57	510	58	2012-08-02 05:04:02.041841
11350	4	1-1996	11348	57	19	352	69	2012-08-02 05:04:02.041841
11351	4	1-2988	11348	19	48	345	27	2012-08-02 05:04:02.041841
11348	4	1-3204	11348	48	100	827	61	2012-08-02 05:04:02.041841
11353	5	1-3031	11352	66	86	67	51	2012-08-02 05:04:02.041841
11354	5	1-2930	11352	86	86	18	91	2012-08-02 05:04:02.041841
11355	5	1-2855	11352	86	85	9	18	2012-08-02 05:04:02.041841
11356	5	1-3169	11352	85	48	22	27	2012-08-02 05:04:02.041841
11352	5	1-3204	11352	48	66	54	61	2012-08-02 05:04:02.041841
11358	5	1-3183	11357	61	62	5145	28	2012-08-02 05:04:02.041841
11359	5	1-2559	11357	62	34	1596	19	2012-08-02 05:04:02.041841
11360	5	1-2861	11357	34	85	494	18	2012-08-02 05:04:02.041841
11361	5	1-3169	11357	85	48	1427	27	2012-08-02 05:04:02.041841
11357	5	1-3204	11357	48	61	4024	61	2012-08-02 05:04:02.041841
11363	4	1-2864	11362	70	17	2830	16	2012-08-02 05:04:02.041841
11364	4	1-2597	11362	17	75	536	66	2012-08-02 05:04:02.041841
11365	4	1-2266	11362	75	48	344	27	2012-08-02 05:04:02.041841
11362	4	1-3204	11362	48	70	990	61	2012-08-02 05:04:02.041841
11367	5	1-2973	11366	100	42	2613	58	2012-08-02 05:04:02.041841
11368	5	1-1232	11366	42	18	1312	76	2012-08-02 05:04:02.041841
11369	5	1-1824	11366	18	85	376	18	2012-08-02 05:04:02.041841
11370	5	1-3169	11366	85	48	1146	27	2012-08-02 05:04:02.041841
11366	5	1-3204	11366	48	100	3410	61	2012-08-02 05:04:02.041841
11371	1	1-1679	\N	66	66	135	27	2012-08-02 05:04:02.041841
11372	1	1-2524	\N	77	77	1647	61	2012-08-02 05:04:02.041841
11374	4	1-3096	11373	83	85	2100	12	2012-08-02 05:04:02.848538
11375	4	1-3191	11373	85	26	3465	92	2012-08-02 05:04:02.848538
11376	4	1-2883	11373	26	64	3134	22	2012-08-02 05:04:02.848538
11373	4	1-3206	11373	64	83	2469	63	2012-08-02 05:04:02.848538
11378	2	1-3107	11377	2	64	7274	22	2012-08-02 05:04:02.848538
11377	2	1-3206	11377	64	2	6290	63	2012-08-02 05:04:02.848538
11380	6	1-3144	11379	38	24	969	38	2012-08-02 05:04:03.677246
11381	6	1-2979	11379	24	64	673	89	2012-08-02 05:04:03.677246
11382	6	1-3079	11379	64	33	1118	47	2012-08-02 05:04:03.677246
11383	6	1-931	11379	33	56	845	6	2012-08-02 05:04:03.677246
11384	6	1-2878	11379	56	79	590	19	2012-08-02 05:04:03.677246
11379	6	1-3208	11379	79	38	1373	90	2012-08-02 05:04:03.677246
11386	5	1-2917	11385	76	75	582	87	2012-08-02 05:04:04.164126
11387	5	1-2598	11385	75	18	412	76	2012-08-02 05:04:04.164126
11388	5	1-1824	11385	18	85	69	18	2012-08-02 05:04:04.164126
11389	5	1-3169	11385	85	64	124	27	2012-08-02 05:04:04.164126
11385	5	1-3209	11385	64	76	1805	95	2012-08-02 05:04:04.164126
11391	6	1-3015	11390	17	18	342	26	2012-08-02 05:04:04.164126
11392	6	1-1513	11390	18	57	417	25	2012-08-02 05:04:04.164126
11393	6	1-2976	11390	57	18	111	76	2012-08-02 05:04:04.164126
11394	6	1-1824	11390	18	85	26	18	2012-08-02 05:04:04.164126
11395	6	1-3169	11390	85	64	49	27	2012-08-02 05:04:04.164126
11390	6	1-3209	11390	64	17	759	95	2012-08-02 05:04:04.164126
16078	4	1-4588	16077	35	52	2715	92	2012-08-02 05:13:35.935381
11397	6	1-2799	11396	61	5	1305	17	2012-08-02 05:04:04.164126
11398	6	1-2836	11396	5	49	1098	53	2012-08-02 05:04:04.164126
11399	6	1-2969	11396	49	18	308	76	2012-08-02 05:04:04.164126
11400	6	1-1824	11396	18	85	58	18	2012-08-02 05:04:04.164126
11401	6	1-3169	11396	85	64	115	27	2012-08-02 05:04:04.164126
11396	6	1-3209	11396	64	61	1880	95	2012-08-02 05:04:04.164126
16079	4	1-4025	16077	52	56	2347	81	2012-08-02 05:13:35.935381
11403	5	1-2366	11402	12	57	999	25	2012-08-02 05:04:04.164126
11404	5	1-2976	11402	57	18	291	76	2012-08-02 05:04:04.164126
11405	5	1-1824	11402	18	85	56	18	2012-08-02 05:04:04.164126
11406	5	1-3169	11402	85	64	116	27	2012-08-02 05:04:04.164126
11402	5	1-3209	11402	64	12	1954	95	2012-08-02 05:04:04.164126
16080	4	1-3778	16077	56	32	896	64	2012-08-02 05:13:35.935381
11408	5	1-1842	11407	6	27	1005	37	2012-08-02 05:04:05.408919
11409	5	1-2315	11407	27	100	336	97	2012-08-02 05:04:05.408919
11410	5	1-3045	11407	100	6	399	36	2012-08-02 05:04:05.408919
11411	5	1-2226	11407	6	98	283	80	2012-08-02 05:04:05.408919
11407	5	1-3212	11407	98	6	1175	70	2012-08-02 05:04:05.408919
16077	4	1-4634	16077	32	35	2245	28	2012-08-02 05:13:35.935381
11413	6	1-2451	11412	53	85	1191	12	2012-08-02 05:04:05.806604
11414	6	1-3191	11412	85	61	1971	92	2012-08-02 05:04:05.806604
11415	6	1-2792	11412	61	85	1931	20	2012-08-02 05:04:05.806604
11416	6	1-2385	11412	85	72	1832	100	2012-08-02 05:04:05.806604
11417	6	1-3073	11412	72	5	1421	60	2012-08-02 05:04:05.806604
11412	6	1-3213	11412	5	53	1236	69	2012-08-02 05:04:05.806604
11419	5	1-3049	11418	95	5	955	35	2012-08-02 05:04:06.635161
11420	5	1-3042	11418	5	98	756	55	2012-08-02 05:04:06.635161
11421	5	1-2891	11418	98	26	586	28	2012-08-02 05:04:06.635161
11422	5	1-2871	11418	26	32	266	13	2012-08-02 05:04:06.635161
11418	5	1-3215	11418	32	95	2951	18	2012-08-02 05:04:06.635161
11424	3	1-2394	11423	9	13	1185	99	2012-08-02 05:04:06.635161
11425	3	1-820	11423	13	32	306	13	2012-08-02 05:04:06.635161
11423	3	1-3215	11423	32	9	3435	18	2012-08-02 05:04:06.635161
11427	4	1-2682	11426	16	4	283	66	2012-08-02 05:04:06.635161
11428	4	1-2105	11426	4	26	114	28	2012-08-02 05:04:06.635161
11429	4	1-2871	11426	26	32	60	13	2012-08-02 05:04:06.635161
11426	4	1-3215	11426	32	16	757	18	2012-08-02 05:04:06.635161
11431	4	1-2697	11430	24	14	533	41	2012-08-02 05:04:07.044529
11432	4	1-2646	11430	14	98	712	80	2012-08-02 05:04:07.044529
11433	4	1-3212	11430	98	91	2976	70	2012-08-02 05:04:07.044529
11430	4	1-3216	11430	91	24	2860	14	2012-08-02 05:04:07.044529
11435	6	1-3026	11434	94	55	506	4	2012-08-02 05:04:08.248435
11436	6	1-2609	11434	55	38	404	90	2012-08-02 05:04:08.248435
11437	6	1-3144	11434	38	58	215	38	2012-08-02 05:04:08.248435
11438	6	1-3085	11434	58	15	544	74	2012-08-02 05:04:08.248435
11439	6	1-2951	11434	15	80	434	9	2012-08-02 05:04:08.248435
11434	6	1-3219	11434	80	94	2374	61	2012-08-02 05:04:08.248435
11441	6	1-2121	11440	58	63	301	32	2012-08-02 05:04:08.657498
11442	6	1-3166	11440	63	78	112	73	2012-08-02 05:04:08.657498
11443	6	1-1460	11440	78	85	40	12	2012-08-02 05:04:08.657498
11444	6	1-3191	11440	85	82	67	92	2012-08-02 05:04:08.657498
11445	6	1-2955	11440	82	71	65	94	2012-08-02 05:04:08.657498
11440	6	1-3220	11440	71	58	205	1	2012-08-02 05:04:08.657498
11447	6	1-2807	11446	60	61	323	86	2012-08-02 05:04:08.657498
11448	6	1-1074	11446	61	31	167	4	2012-08-02 05:04:08.657498
11449	6	1-2965	11446	31	85	183	12	2012-08-02 05:04:08.657498
11450	6	1-3191	11446	85	82	310	92	2012-08-02 05:04:08.657498
11451	6	1-2955	11446	82	71	304	94	2012-08-02 05:04:08.657498
11446	6	1-3220	11446	71	60	984	1	2012-08-02 05:04:08.657498
11453	5	1-2659	11452	82	30	2989	55	2012-08-02 05:04:08.657498
11454	5	1-2775	11452	30	96	563	29	2012-08-02 05:04:08.657498
11455	5	1-3033	11452	96	10	1680	76	2012-08-02 05:04:08.657498
11456	5	1-2940	11452	10	71	874	94	2012-08-02 05:04:08.657498
11452	5	1-3220	11452	71	82	3000	1	2012-08-02 05:04:08.657498
11458	6	1-3147	11457	12	16	1096	7	2012-08-02 05:04:09.066657
11459	6	1-1380	11457	16	21	1303	85	2012-08-02 05:04:09.066657
11460	6	1-2680	11457	21	25	968	47	2012-08-02 05:04:09.066657
11461	6	1-2791	11457	25	12	959	36	2012-08-02 05:04:09.066657
11462	6	1-3162	11457	12	8	887	65	2012-08-02 05:04:09.066657
11457	6	1-3221	11457	8	12	658	46	2012-08-02 05:04:09.066657
11464	6	1-3023	11463	29	9	672	43	2012-08-02 05:04:10.788593
11465	6	1-3146	11463	9	58	336	38	2012-08-02 05:04:10.788593
11466	6	1-3085	11463	58	80	970	74	2012-08-02 05:04:10.788593
11467	6	1-2732	11463	80	49	1135	25	2012-08-02 05:04:10.788593
11468	6	1-3194	11463	49	96	1577	11	2012-08-02 05:04:10.788593
11463	6	1-3225	11463	96	29	2555	6	2012-08-02 05:04:10.788593
11470	4	1-3033	11469	96	8	1622	76	2012-08-02 05:04:16.811242
11471	4	1-3172	11469	8	74	1320	61	2012-08-02 05:04:16.811242
11472	4	1-3013	11469	74	22	371	60	2012-08-02 05:04:16.811242
11469	4	1-3241	11469	22	96	754	29	2012-08-02 05:04:16.811242
11474	6	1-3188	11473	23	50	2244	63	2012-08-02 05:04:16.811242
11475	6	1-2690	11473	50	98	1466	80	2012-08-02 05:04:16.811242
11476	6	1-3212	11473	98	98	5544	70	2012-08-02 05:04:16.811242
11477	6	1-2184	11473	98	72	990	100	2012-08-02 05:04:16.811242
11478	6	1-3073	11473	72	22	688	60	2012-08-02 05:04:16.811242
11473	6	1-3241	11473	22	23	1432	29	2012-08-02 05:04:16.811242
11480	6	1-3188	11479	23	33	1441	63	2012-08-02 05:04:16.811242
11481	6	1-3080	11479	33	43	1207	93	2012-08-02 05:04:16.811242
11482	6	1-2441	11479	43	85	661	20	2012-08-02 05:04:16.811242
11483	6	1-2385	11479	85	72	534	100	2012-08-02 05:04:16.811242
11484	6	1-3073	11479	72	22	393	60	2012-08-02 05:04:16.811242
11479	6	1-3241	11479	22	23	868	29	2012-08-02 05:04:16.811242
11486	4	1-2428	11485	66	5	522	30	2012-08-02 05:04:16.811242
11487	4	1-2932	11485	5	72	425	100	2012-08-02 05:04:16.811242
11488	4	1-3073	11485	72	22	354	60	2012-08-02 05:04:16.811242
11485	4	1-3241	11485	22	66	884	29	2012-08-02 05:04:16.811242
11490	6	1-2539	11489	84	85	396	12	2012-08-02 05:04:18.811052
11491	6	1-3191	11489	85	61	677	92	2012-08-02 05:04:18.811052
11489	6	1-3246	11489	24	84	478	11	2012-08-02 05:04:18.811052
11492	6	1-2792	11489	61	85	707	20	2012-08-02 05:04:18.811052
11493	6	1-2385	11489	85	28	693	100	2012-08-02 05:04:18.811052
11494	6	1-3181	11489	28	24	836	64	2012-08-02 05:04:18.811052
11496	5	1-3225	11495	96	56	202	6	2012-08-02 05:04:18.811052
11497	5	1-2878	11495	56	2	133	19	2012-08-02 05:04:18.811052
11498	5	1-3087	11495	2	28	147	100	2012-08-02 05:04:18.811052
11499	5	1-3181	11495	28	24	188	64	2012-08-02 05:04:18.811052
11495	5	1-3246	11495	24	96	114	11	2012-08-02 05:04:18.811052
16082	5	1-4357	16081	35	82	642	67	2012-08-02 05:13:35.935381
11501	5	1-3098	11500	95	89	184	21	2012-08-02 05:04:20.039039
11502	5	1-2596	11500	89	20	652	5	2012-08-02 05:04:20.039039
11503	5	1-1989	11500	20	87	67	68	2012-08-02 05:04:20.039039
11504	5	1-2966	11500	87	93	112	2	2012-08-02 05:04:20.039039
11500	5	1-3249	11500	93	95	180	71	2012-08-02 05:04:20.039039
11506	5	1-3098	11505	95	15	1840	21	2012-08-02 05:04:20.039039
11507	5	1-3102	11505	15	22	1452	84	2012-08-02 05:04:20.039039
11508	5	1-1958	11505	22	11	1408	86	2012-08-02 05:04:20.039039
11509	5	1-2578	11505	11	93	1019	2	2012-08-02 05:04:20.039039
11505	5	1-3249	11505	93	95	1715	71	2012-08-02 05:04:20.039039
11511	5	1-2833	11510	9	24	961	40	2012-08-02 05:04:20.436835
11512	5	1-3180	11510	24	52	485	52	2012-08-02 05:04:20.436835
11513	5	1-2153	11510	52	24	167	43	2012-08-02 05:04:20.436835
11514	5	1-3229	11510	24	44	295	22	2012-08-02 05:04:20.436835
11510	5	1-3250	11510	44	9	443	87	2012-08-02 05:04:20.436835
11516	6	1-2819	11515	97	34	150	59	2012-08-02 05:04:20.436835
11517	6	1-2123	11515	34	85	73	57	2012-08-02 05:04:20.436835
11518	6	1-2977	11515	85	17	31	3	2012-08-02 05:04:20.436835
11519	6	1-1226	11515	17	24	19	43	2012-08-02 05:04:20.436835
11520	6	1-3229	11515	24	44	35	22	2012-08-02 05:04:20.436835
11515	6	1-3250	11515	44	97	75	87	2012-08-02 05:04:20.436835
11522	6	1-1873	11521	94	55	2830	72	2012-08-02 05:04:20.436835
11523	6	1-2888	11521	55	21	2389	21	2012-08-02 05:04:20.436835
11524	6	1-3051	11521	21	86	1372	49	2012-08-02 05:04:20.436835
11525	6	1-2251	11521	86	24	1086	43	2012-08-02 05:04:20.436835
11526	6	1-3229	11521	24	44	2404	22	2012-08-02 05:04:20.436835
11521	6	1-3250	11521	44	94	5904	87	2012-08-02 05:04:20.436835
11528	3	1-2548	11527	52	8	1073	6	2012-08-02 05:04:22.061062
11529	3	1-3157	11527	8	47	1278	99	2012-08-02 05:04:22.061062
11527	3	1-3254	11527	47	52	7376	55	2012-08-02 05:04:22.061062
11531	4	1-2432	11530	40	85	302	35	2012-08-02 05:04:22.061062
11532	4	1-2505	11530	85	8	219	6	2012-08-02 05:04:22.061062
11533	4	1-3157	11530	8	47	285	99	2012-08-02 05:04:22.061062
11530	4	1-3254	11530	47	40	1798	55	2012-08-02 05:04:22.061062
11535	6	1-2822	11534	71	81	544	16	2012-08-02 05:04:22.470067
11536	6	1-3141	11534	81	89	632	85	2012-08-02 05:04:22.470067
11537	6	1-2242	11534	89	38	345	90	2012-08-02 05:04:22.470067
11538	6	1-3144	11534	38	93	215	38	2012-08-02 05:04:22.470067
11539	6	1-1272	11534	93	29	266	100	2012-08-02 05:04:22.470067
11534	6	1-3255	11534	29	71	1011	71	2012-08-02 05:04:22.470067
11541	5	1-3098	11540	95	15	2438	21	2012-08-02 05:04:22.470067
11542	5	1-3102	11540	15	93	1947	84	2012-08-02 05:04:22.470067
11543	5	1-2881	11540	93	34	949	56	2012-08-02 05:04:22.470067
11544	5	1-724	11540	34	29	511	100	2012-08-02 05:04:22.470067
11540	5	1-3255	11540	29	95	2246	71	2012-08-02 05:04:22.470067
11546	5	1-429	11545	72	20	1219	72	2012-08-02 05:04:27.286656
11547	5	1-2911	11545	20	61	827	95	2012-08-02 05:04:27.286656
11548	5	1-2799	11545	61	92	806	17	2012-08-02 05:04:27.286656
11549	5	1-3198	11545	92	57	1100	79	2012-08-02 05:04:27.286656
11545	5	1-3267	11545	57	72	780	8	2012-08-02 05:04:27.286656
11551	5	1-2341	11550	11	28	2581	15	2012-08-02 05:04:27.684457
11552	5	1-3110	11550	28	5	637	84	2012-08-02 05:04:27.684457
11553	5	1-2401	11550	5	11	426	23	2012-08-02 05:04:27.684457
11554	5	1-3119	11550	11	3	338	33	2012-08-02 05:04:27.684457
11550	5	1-3268	11550	3	11	4621	7	2012-08-02 05:04:27.684457
11556	4	1-2585	11555	20	46	1938	22	2012-08-02 05:04:28.89999
11557	4	1-3151	11555	46	9	1416	79	2012-08-02 05:04:28.89999
11558	4	1-3248	11555	9	26	1380	2	2012-08-02 05:04:28.89999
11555	4	1-3271	11555	26	20	1172	30	2012-08-02 05:04:28.89999
11560	4	1-2939	11559	89	12	629	98	2012-08-02 05:04:28.89999
11561	4	1-3116	11559	12	63	760	40	2012-08-02 05:04:28.89999
11562	4	1-3127	11559	63	26	357	2	2012-08-02 05:04:28.89999
11559	4	1-3271	11559	26	89	313	30	2012-08-02 05:04:28.89999
11564	5	1-2932	11563	5	85	147	100	2012-08-02 05:04:28.89999
11565	5	1-2872	11563	85	61	142	77	2012-08-02 05:04:28.89999
11566	5	1-504	11563	61	14	42	75	2012-08-02 05:04:28.89999
11567	5	1-3211	11563	14	26	186	2	2012-08-02 05:04:28.89999
11563	5	1-3271	11563	26	5	165	30	2012-08-02 05:04:28.89999
11568	1	1-504	\N	61	61	1057	75	2012-08-02 05:04:28.89999
11570	6	1-3051	11569	21	22	704	49	2012-08-02 05:04:29.319558
11571	6	1-2812	11569	22	13	670	5	2012-08-02 05:04:29.319558
11572	6	1-3131	11569	13	96	500	11	2012-08-02 05:04:29.319558
11573	6	1-3225	11569	96	8	906	6	2012-08-02 05:04:29.319558
11574	6	1-3157	11569	8	49	1684	99	2012-08-02 05:04:29.319558
11569	6	1-3272	11569	49	21	1269	21	2012-08-02 05:04:29.319558
11576	5	1-3071	11575	94	12	535	36	2012-08-02 05:04:30.9329
11577	5	1-3162	11575	12	8	301	65	2012-08-02 05:04:30.9329
11578	5	1-3221	11575	8	13	141	46	2012-08-02 05:04:30.9329
11579	5	1-2889	11575	13	9	49	87	2012-08-02 05:04:30.9329
11575	5	1-3276	11575	9	94	427	45	2012-08-02 05:04:30.9329
11581	5	1-3057	11580	69	5	602	85	2012-08-02 05:04:30.9329
11582	5	1-2194	11580	5	20	210	72	2012-08-02 05:04:30.9329
11583	5	1-2911	11580	20	76	142	95	2012-08-02 05:04:30.9329
11584	5	1-2917	11580	76	9	71	87	2012-08-02 05:04:30.9329
11580	5	1-3276	11580	9	69	922	45	2012-08-02 05:04:30.9329
11586	5	1-3244	11585	21	58	239	97	2012-08-02 05:04:30.9329
11587	5	1-3251	11585	58	8	569	65	2012-08-02 05:04:30.9329
11588	5	1-3221	11585	8	13	453	46	2012-08-02 05:04:30.9329
11589	5	1-2889	11585	13	9	265	87	2012-08-02 05:04:30.9329
11585	5	1-3276	11585	9	21	3934	45	2012-08-02 05:04:30.9329
11591	5	1-2242	11590	89	38	1648	90	2012-08-02 05:04:31.739236
11592	5	1-3144	11590	38	72	735	38	2012-08-02 05:04:31.739236
11593	5	1-2938	11590	72	70	644	42	2012-08-02 05:04:31.739236
11594	5	1-3104	11590	70	17	236	33	2012-08-02 05:04:31.739236
11590	5	1-3278	11590	17	89	4228	85	2012-08-02 05:04:31.739236
11596	4	1-2194	11595	5	47	748	72	2012-08-02 05:04:31.739236
11597	4	1-2420	11595	47	11	195	23	2012-08-02 05:04:31.739236
11598	4	1-3119	11595	11	17	135	33	2012-08-02 05:04:31.739236
11595	4	1-3278	11595	17	5	2706	85	2012-08-02 05:04:31.739236
16083	5	1-4028	16081	82	37	496	36	2012-08-02 05:13:35.935381
11600	6	1-1839	11599	63	52	1776	52	2012-08-02 05:04:34.578588
11601	6	1-2153	11599	52	9	703	43	2012-08-02 05:04:34.578588
11602	6	1-3146	11599	9	58	285	38	2012-08-02 05:04:34.578588
11603	6	1-3085	11599	58	15	668	74	2012-08-02 05:04:34.578588
11604	6	1-2951	11599	15	91	493	9	2012-08-02 05:04:34.578588
11599	6	1-3285	11599	91	63	8467	70	2012-08-02 05:04:34.578588
16084	5	1-4539	16081	37	77	951	1	2012-08-02 05:13:35.935381
11606	2	1-1549	11605	30	100	2336	88	2012-08-02 05:04:36.180485
11605	2	1-3289	11605	100	30	1866	47	2012-08-02 05:04:36.180485
11608	5	1-3147	11607	12	11	2416	7	2012-08-02 05:04:37.771315
11609	5	1-2341	11607	11	96	1365	15	2012-08-02 05:04:37.771315
11610	5	1-2784	11607	96	27	652	2	2012-08-02 05:04:37.771315
11611	5	1-3114	11607	27	42	1098	83	2012-08-02 05:04:37.771315
11607	5	1-3293	11607	42	12	1763	46	2012-08-02 05:04:37.771315
11613	4	1-2889	11612	13	6	215	87	2012-08-02 05:04:37.771315
11614	4	1-2663	11612	6	27	115	2	2012-08-02 05:04:37.771315
11615	4	1-3114	11612	27	42	221	83	2012-08-02 05:04:37.771315
11612	4	1-3293	11612	42	13	404	46	2012-08-02 05:04:37.771315
11617	5	1-1524	11616	88	18	353	69	2012-08-02 05:04:37.771315
11618	5	1-2927	11616	18	87	438	68	2012-08-02 05:04:37.771315
11619	5	1-2966	11616	87	27	738	2	2012-08-02 05:04:37.771315
11620	5	1-3114	11616	27	42	1464	83	2012-08-02 05:04:37.771315
11616	5	1-3293	11616	42	88	2772	46	2012-08-02 05:04:37.771315
11622	5	1-2966	11621	87	93	988	2	2012-08-02 05:04:38.566976
11623	5	1-3249	11621	93	71	1539	71	2012-08-02 05:04:38.566976
11624	5	1-2822	11621	71	96	876	16	2012-08-02 05:04:38.566976
11625	5	1-3174	11621	96	77	560	23	2012-08-02 05:04:38.566976
11621	5	1-3295	11621	77	87	604	68	2012-08-02 05:04:38.566976
11627	4	1-2667	11626	98	34	1425	65	2012-08-02 05:04:39.373687
11628	4	1-3084	11626	34	63	2687	80	2012-08-02 05:04:39.373687
11629	4	1-3170	11626	63	10	2238	59	2012-08-02 05:04:39.373687
11626	4	1-3297	11626	10	98	3687	99	2012-08-02 05:04:39.373687
11631	5	1-1019	11630	56	18	264	93	2012-08-02 05:04:39.373687
11632	5	1-2795	11630	18	24	527	40	2012-08-02 05:04:39.373687
11633	5	1-3180	11630	24	87	342	52	2012-08-02 05:04:39.373687
11634	5	1-2200	11630	87	10	418	59	2012-08-02 05:04:39.373687
11630	5	1-3297	11630	10	56	705	99	2012-08-02 05:04:39.373687
11636	3	1-3249	11635	93	42	2069	71	2012-08-02 05:04:40.589326
11637	3	1-2134	11635	42	70	1313	10	2012-08-02 05:04:40.589326
11635	3	1-3300	11635	70	93	1261	2	2012-08-02 05:04:40.589326
11639	5	1-3301	11638	12	96	1698	11	2012-08-02 05:04:41.406215
11640	5	1-3225	11638	96	42	2494	6	2012-08-02 05:04:41.406215
11641	5	1-2544	11638	42	11	629	69	2012-08-02 05:04:41.406215
11642	5	1-2506	11638	11	75	163	64	2012-08-02 05:04:41.406215
11638	5	1-3302	11638	75	12	3194	41	2012-08-02 05:04:41.406215
11644	5	1-3301	11643	12	48	2769	11	2012-08-02 05:04:41.406215
11645	5	1-936	11643	48	17	1030	95	2012-08-02 05:04:41.406215
11646	5	1-3015	11643	17	88	644	26	2012-08-02 05:04:41.406215
11647	5	1-2060	11643	88	75	232	64	2012-08-02 05:04:41.406215
11643	5	1-3302	11643	75	12	4864	41	2012-08-02 05:04:41.406215
11649	5	1-2539	11648	84	15	2888	12	2012-08-02 05:04:42.201903
11650	5	1-1883	11648	15	38	1972	90	2012-08-02 05:04:42.201903
11651	5	1-3144	11648	38	23	1044	38	2012-08-02 05:04:42.201903
11652	5	1-3228	11648	23	77	1641	3	2012-08-02 05:04:42.201903
11648	5	1-3304	11648	77	84	4138	11	2012-08-02 05:04:42.201903
11654	6	1-1698	11653	25	58	1314	98	2012-08-02 05:04:42.201903
11655	6	1-2783	11653	58	58	405	96	2012-08-02 05:04:42.201903
11656	6	1-1829	11653	58	38	667	90	2012-08-02 05:04:42.201903
11657	6	1-3144	11653	38	23	371	38	2012-08-02 05:04:42.201903
11658	6	1-3228	11653	23	77	613	3	2012-08-02 05:04:42.201903
11653	6	1-3304	11653	77	25	1624	11	2012-08-02 05:04:42.201903
11660	5	1-2091	11659	46	12	4821	14	2012-08-02 05:04:42.201903
11661	5	1-2655	11659	12	9	860	43	2012-08-02 05:04:42.201903
11662	5	1-3146	11659	9	23	427	38	2012-08-02 05:04:42.201903
11663	5	1-3228	11659	23	77	768	3	2012-08-02 05:04:42.201903
11659	5	1-3304	11659	77	46	2215	11	2012-08-02 05:04:42.201903
11665	6	1-434	11664	86	89	313	51	2012-08-02 05:04:42.201903
11666	6	1-2984	11664	89	15	199	12	2012-08-02 05:04:42.201903
11667	6	1-1883	11664	15	38	166	90	2012-08-02 05:04:42.201903
11668	6	1-3144	11664	38	23	107	38	2012-08-02 05:04:42.201903
11669	6	1-3228	11664	23	77	206	3	2012-08-02 05:04:42.201903
11664	6	1-3304	11664	77	86	634	11	2012-08-02 05:04:42.201903
11671	4	1-1698	11670	25	58	243	98	2012-08-02 05:04:42.201903
11672	4	1-2783	11670	58	84	83	96	2012-08-02 05:04:42.201903
11673	4	1-2992	11670	84	77	70	3	2012-08-02 05:04:42.201903
11670	4	1-3304	11670	77	25	236	11	2012-08-02 05:04:42.201903
11674	1	1-3144	\N	38	38	1571	38	2012-08-02 05:04:42.201903
11676	6	1-3131	11675	13	57	133	11	2012-08-02 05:04:42.633272
11677	6	1-2764	11675	57	16	100	68	2012-08-02 05:04:42.633272
11678	6	1-2286	11675	16	96	80	13	2012-08-02 05:04:42.633272
11679	6	1-3179	11675	96	71	65	94	2012-08-02 05:04:42.633272
11680	6	1-3220	11675	71	28	202	1	2012-08-02 05:04:42.633272
11675	6	1-3305	11675	28	13	187	5	2012-08-02 05:04:42.633272
11682	5	1-3179	11681	96	71	530	94	2012-08-02 05:04:44.643719
11683	5	1-3220	11681	71	86	1629	1	2012-08-02 05:04:44.643719
11684	5	1-1803	11681	86	91	1614	22	2012-08-02 05:04:44.643719
11685	5	1-3200	11681	91	12	768	85	2012-08-02 05:04:44.643719
11681	5	1-3310	11681	12	96	669	13	2012-08-02 05:04:44.643719
11687	5	1-3240	11686	12	15	91	74	2012-08-02 05:04:44.643719
11688	5	1-2951	11686	15	56	84	9	2012-08-02 05:04:44.643719
11689	5	1-2815	11686	56	9	152	53	2012-08-02 05:04:44.643719
11690	5	1-3307	11686	9	12	216	85	2012-08-02 05:04:44.643719
11686	5	1-3310	11686	12	12	197	13	2012-08-02 05:04:44.643719
11692	6	1-2766	11691	59	96	914	16	2012-08-02 05:04:45.019713
11693	6	1-3174	11691	96	11	520	23	2012-08-02 05:04:45.019713
11694	6	1-3119	11691	11	13	342	33	2012-08-02 05:04:45.019713
11695	6	1-2758	11691	13	96	178	67	2012-08-02 05:04:45.019713
11696	6	1-1716	11691	96	50	254	54	2012-08-02 05:04:45.019713
11691	6	1-3311	11691	50	59	1758	32	2012-08-02 05:04:45.019713
11698	6	1-2908	11697	66	63	738	86	2012-08-02 05:04:45.019713
11699	6	1-2482	11697	63	22	299	60	2012-08-02 05:04:45.019713
11700	6	1-3241	11697	22	59	649	29	2012-08-02 05:04:45.019713
11701	6	1-3052	11697	59	96	473	67	2012-08-02 05:04:45.019713
11702	6	1-1716	11697	96	50	633	54	2012-08-02 05:04:45.019713
11697	6	1-3311	11697	50	66	5085	32	2012-08-02 05:04:45.019713
16085	5	1-4461	16081	77	32	500	64	2012-08-02 05:13:35.935381
11704	4	1-3307	11703	9	12	7749	85	2012-08-02 05:04:45.428471
11705	4	1-3310	11703	12	6	5854	13	2012-08-02 05:04:45.428471
11706	4	1-1521	11703	6	27	2441	14	2012-08-02 05:04:45.428471
11703	4	1-3312	11703	27	9	6074	53	2012-08-02 05:04:45.428471
16081	5	1-4634	16081	32	35	1390	28	2012-08-02 05:13:35.935381
11708	6	1-1946	11707	95	29	2037	48	2012-08-02 05:04:45.837282
11709	6	1-3178	11707	29	86	1854	91	2012-08-02 05:04:45.837282
11710	6	1-2855	11707	86	85	882	18	2012-08-02 05:04:45.837282
11711	6	1-3169	11707	85	87	2503	27	2012-08-02 05:04:45.837282
11712	6	1-2915	11707	87	77	2319	70	2012-08-02 05:04:45.837282
11707	6	1-3313	11707	77	95	3155	66	2012-08-02 05:04:45.837282
11714	6	1-1213	11713	53	49	371	53	2012-08-02 05:04:45.837282
11715	6	1-2969	11713	49	18	159	76	2012-08-02 05:04:45.837282
11716	6	1-1824	11713	18	85	46	18	2012-08-02 05:04:45.837282
11717	6	1-3169	11713	85	87	140	27	2012-08-02 05:04:45.837282
11718	6	1-2915	11713	87	77	140	70	2012-08-02 05:04:45.837282
11713	6	1-3313	11713	77	53	205	66	2012-08-02 05:04:45.837282
11720	3	1-2266	11719	75	87	773	27	2012-08-02 05:04:45.837282
11721	3	1-2915	11719	87	77	778	70	2012-08-02 05:04:45.837282
11719	3	1-3313	11719	77	75	1150	66	2012-08-02 05:04:45.837282
11722	1	1-1824	\N	18	18	65	18	2012-08-02 05:04:45.837282
11724	5	1-2734	11723	6	99	1289	37	2012-08-02 05:04:47.883251
11725	5	1-3303	11723	99	29	812	48	2012-08-02 05:04:47.883251
11726	5	1-3178	11723	29	63	652	91	2012-08-02 05:04:47.883251
11727	5	1-2953	11723	63	2	643	42	2012-08-02 05:04:47.883251
11723	5	1-3318	11723	2	6	1616	92	2012-08-02 05:04:47.883251
11729	3	1-2792	11728	61	76	3941	20	2012-08-02 05:04:47.883251
11730	3	1-3142	11728	76	2	1201	42	2012-08-02 05:04:47.883251
11728	3	1-3318	11728	2	61	3559	92	2012-08-02 05:04:47.883251
11732	5	1-2720	11731	68	18	192	93	2012-08-02 05:04:47.883251
11733	5	1-2795	11731	18	97	358	40	2012-08-02 05:04:47.883251
11734	5	1-1936	11731	97	39	44	17	2012-08-02 05:04:47.883251
11735	5	1-2201	11731	39	2	49	42	2012-08-02 05:04:47.883251
11731	5	1-3318	11731	2	68	147	92	2012-08-02 05:04:47.883251
11737	4	1-2405	11736	48	27	2012	48	2012-08-02 05:04:48.690107
11738	4	1-3088	11736	27	70	2892	10	2012-08-02 05:04:48.690107
11739	4	1-3300	11736	70	10	1986	2	2012-08-02 05:04:48.690107
11736	4	1-3320	11736	10	48	2121	9	2012-08-02 05:04:48.690107
11741	3	1-1838	11740	26	70	4540	10	2012-08-02 05:04:48.690107
11742	3	1-3300	11740	70	10	3314	2	2012-08-02 05:04:48.690107
11740	3	1-3320	11740	10	26	3763	9	2012-08-02 05:04:48.690107
11744	5	1-1698	11743	25	70	361	98	2012-08-02 05:04:50.325221
11745	5	1-3292	11743	70	9	159	87	2012-08-02 05:04:50.325221
11746	5	1-3276	11743	9	69	1665	45	2012-08-02 05:04:50.325221
11747	5	1-3057	11743	69	55	885	85	2012-08-02 05:04:50.325221
11743	5	1-3324	11743	55	25	485	11	2012-08-02 05:04:50.325221
11749	5	1-936	11748	48	76	216	95	2012-08-02 05:04:50.325221
11750	5	1-2917	11748	76	9	106	87	2012-08-02 05:04:50.325221
11751	5	1-3276	11748	9	69	1339	45	2012-08-02 05:04:50.325221
11752	5	1-3057	11748	69	55	853	85	2012-08-02 05:04:50.325221
11748	5	1-3324	11748	55	48	561	11	2012-08-02 05:04:50.325221
11754	3	1-3168	11753	60	5	6050	35	2012-08-02 05:04:50.733847
11755	3	1-3042	11753	5	56	6389	55	2012-08-02 05:04:50.733847
11753	3	1-3325	11753	56	60	8766	4	2012-08-02 05:04:50.733847
11757	5	1-1232	11756	42	10	354	76	2012-08-02 05:04:52.744334
11758	5	1-2940	11756	10	71	180	94	2012-08-02 05:04:52.744334
11759	5	1-3220	11756	71	60	603	1	2012-08-02 05:04:52.744334
11760	5	1-2807	11756	60	69	205	86	2012-08-02 05:04:52.744334
11756	5	1-3330	11756	69	42	694	58	2012-08-02 05:04:52.744334
11762	5	1-2927	11761	18	16	902	68	2012-08-02 05:04:55.174842
11763	5	1-2286	11761	16	12	612	13	2012-08-02 05:04:55.174842
11764	5	1-3240	11761	12	80	231	74	2012-08-02 05:04:55.174842
11765	5	1-2732	11761	80	28	243	25	2012-08-02 05:04:55.174842
11761	5	1-3336	11761	28	18	902	69	2012-08-02 05:04:55.174842
11767	5	1-2285	11766	67	71	267	94	2012-08-02 05:04:55.174842
11768	5	1-3220	11766	71	47	741	1	2012-08-02 05:04:55.174842
11769	5	1-2884	11766	47	57	683	65	2012-08-02 05:04:55.174842
11770	5	1-2592	11766	57	28	337	25	2012-08-02 05:04:55.174842
11766	5	1-3336	11766	28	67	1317	69	2012-08-02 05:04:55.174842
11772	4	1-2988	11771	19	7	1230	27	2012-08-02 05:04:55.174842
11773	4	1-3264	11771	7	16	424	26	2012-08-02 05:04:55.174842
11774	4	1-16	11771	16	28	296	25	2012-08-02 05:04:55.174842
11771	4	1-3336	11771	28	19	1179	69	2012-08-02 05:04:55.174842
11776	5	1-2988	11775	19	37	2253	27	2012-08-02 05:04:55.174842
11777	5	1-3133	11775	37	76	899	56	2012-08-02 05:04:55.174842
11778	5	1-1848	11775	76	12	833	95	2012-08-02 05:04:55.174842
11779	5	1-2366	11775	12	28	540	25	2012-08-02 05:04:55.174842
11775	5	1-3336	11775	28	19	2157	69	2012-08-02 05:04:55.174842
11781	5	1-2043	11780	2	85	1656	100	2012-08-02 05:04:55.174842
11782	5	1-2872	11780	85	86	1469	77	2012-08-02 05:04:55.174842
11783	5	1-580	11780	86	80	449	74	2012-08-02 05:04:55.174842
11784	5	1-2732	11780	80	28	547	25	2012-08-02 05:04:55.174842
11780	5	1-3336	11780	28	2	2353	69	2012-08-02 05:04:55.174842
11786	5	1-2043	11785	2	27	1321	100	2012-08-02 05:04:55.174842
11787	5	1-3041	11785	27	2	3628	32	2012-08-02 05:04:55.174842
11788	5	1-3002	11785	2	84	1978	50	2012-08-02 05:04:55.174842
11789	5	1-715	11785	84	28	395	25	2012-08-02 05:04:55.174842
11785	5	1-3336	11785	28	2	1733	69	2012-08-02 05:04:55.174842
11791	5	1-3180	11790	24	25	1137	52	2012-08-02 05:04:56.843444
11792	5	1-2518	11790	25	42	605	83	2012-08-02 05:04:56.843444
11793	5	1-3293	11790	42	91	744	46	2012-08-02 05:04:56.843444
11794	5	1-1322	11790	91	6	331	33	2012-08-02 05:04:56.843444
11790	5	1-3340	11790	6	24	2866	40	2012-08-02 05:04:56.843444
11796	4	1-1936	11795	97	92	364	17	2012-08-02 05:04:56.843444
11797	4	1-3198	11795	92	40	452	79	2012-08-02 05:04:56.843444
11798	4	1-2331	11795	40	6	307	33	2012-08-02 05:04:56.843444
11795	4	1-3340	11795	6	97	3462	40	2012-08-02 05:04:56.843444
11800	5	1-3180	11799	24	52	578	52	2012-08-02 05:04:56.843444
11801	5	1-2153	11799	52	79	275	43	2012-08-02 05:04:56.843444
11802	5	1-2654	11799	79	40	132	79	2012-08-02 05:04:56.843444
11803	5	1-2331	11799	40	6	93	33	2012-08-02 05:04:56.843444
11799	5	1-3340	11799	6	24	1081	40	2012-08-02 05:04:56.843444
11804	1	1-2331	\N	40	40	2249	33	2012-08-02 05:04:56.843444
16087	3	1-4348	16086	18	62	3583	39	2012-08-02 05:13:35.935381
11806	2	1-1502	11805	9	99	2194	88	2012-08-02 05:04:58.026139
11805	2	1-3343	11805	99	9	1205	61	2012-08-02 05:04:58.026139
16088	3	1-4616	16086	62	32	1318	64	2012-08-02 05:13:35.935381
11808	3	1-2085	11807	61	84	3815	42	2012-08-02 05:04:58.026139
11809	3	1-3257	11807	84	99	4839	88	2012-08-02 05:04:58.026139
11807	3	1-3343	11807	99	61	2831	61	2012-08-02 05:04:58.026139
16086	3	1-4634	16086	32	18	4518	28	2012-08-02 05:13:35.935381
11811	5	1-3180	11810	24	87	2331	52	2012-08-02 05:04:59.627689
11812	5	1-2200	11810	87	20	1109	59	2012-08-02 05:04:59.627689
11813	5	1-1773	11810	20	19	149	90	2012-08-02 05:04:59.627689
11814	5	1-3314	11810	19	25	39	49	2012-08-02 05:04:59.627689
11810	5	1-3347	11810	25	24	9242	40	2012-08-02 05:04:59.627689
11816	6	1-2969	11815	49	8	849	76	2012-08-02 05:05:00.854224
11817	6	1-3172	11815	8	66	925	61	2012-08-02 05:05:00.854224
11818	6	1-3031	11815	66	89	1384	51	2012-08-02 05:05:00.854224
11819	6	1-2984	11815	89	85	939	12	2012-08-02 05:05:00.854224
11820	6	1-3191	11815	85	26	1633	92	2012-08-02 05:05:00.854224
11815	6	1-3350	11815	26	49	1976	53	2012-08-02 05:05:00.854224
11822	3	1-1354	11821	51	84	4623	42	2012-08-02 05:05:01.671839
11823	3	1-3257	11821	84	1	3182	88	2012-08-02 05:05:01.671839
11821	3	1-3352	11821	1	51	9811	98	2012-08-02 05:05:01.671839
11825	4	1-3334	11824	25	86	4151	49	2012-08-02 05:05:02.875754
11826	4	1-2251	11824	86	9	3102	43	2012-08-02 05:05:02.875754
11827	4	1-3146	11824	9	25	1694	38	2012-08-02 05:05:02.875754
11824	4	1-3355	11824	25	25	3894	58	2012-08-02 05:05:02.875754
11829	4	1-1460	11828	78	85	370	12	2012-08-02 05:05:03.273434
11830	4	1-3191	11828	85	61	467	92	2012-08-02 05:05:03.273434
11831	4	1-2792	11828	61	36	389	20	2012-08-02 05:05:03.273434
11828	4	1-3356	11828	36	78	1348	73	2012-08-02 05:05:03.273434
11833	5	1-3055	11832	10	85	473	21	2012-08-02 05:05:03.273434
11834	5	1-1037	11832	85	27	234	48	2012-08-02 05:05:03.273434
11835	5	1-3088	11832	27	22	378	10	2012-08-02 05:05:03.273434
11836	5	1-3290	11832	22	36	295	20	2012-08-02 05:05:03.273434
11832	5	1-3356	11832	36	10	1141	73	2012-08-02 05:05:03.273434
11838	6	1-3217	11837	36	16	262	7	2012-08-02 05:05:03.671329
11839	6	1-1380	11837	16	12	289	85	2012-08-02 05:05:03.671329
11840	6	1-3310	11837	12	96	246	13	2012-08-02 05:05:03.671329
11841	6	1-3179	11837	96	71	191	94	2012-08-02 05:05:03.671329
11842	6	1-3220	11837	71	20	573	1	2012-08-02 05:05:03.671329
11837	6	1-3357	11837	20	36	452	61	2012-08-02 05:05:03.671329
11843	1	1-1380	\N	16	16	157	85	2012-08-02 05:05:03.671329
11845	5	1-3266	11844	51	98	2847	34	2012-08-02 05:05:04.864936
11846	5	1-1970	11844	98	81	1772	89	2012-08-02 05:05:04.864936
11847	5	1-3263	11844	81	88	8284	71	2012-08-02 05:05:04.864936
11848	5	1-2874	11844	88	13	1649	82	2012-08-02 05:05:04.864936
11844	5	1-3360	11844	13	51	3622	18	2012-08-02 05:05:04.864936
11850	4	1-3266	11849	51	23	1573	34	2012-08-02 05:05:04.864936
11851	4	1-2224	11849	23	59	1193	44	2012-08-02 05:05:04.864936
11852	4	1-3069	11849	59	13	636	82	2012-08-02 05:05:04.864936
11849	4	1-3360	11849	13	51	1672	18	2012-08-02 05:05:04.864936
11854	5	1-803	11853	93	81	224	89	2012-08-02 05:05:05.750171
11855	5	1-3263	11853	81	71	1528	71	2012-08-02 05:05:05.750171
11856	5	1-2822	11853	71	81	725	16	2012-08-02 05:05:05.750171
11857	5	1-3141	11853	81	73	856	85	2012-08-02 05:05:05.750171
11853	5	1-3362	11853	73	93	1109	23	2012-08-02 05:05:05.750171
11859	5	1-3119	11858	11	71	898	33	2012-08-02 05:05:05.750171
11860	5	1-2741	11858	71	93	413	58	2012-08-02 05:05:05.750171
11861	5	1-1748	11858	93	22	498	50	2012-08-02 05:05:05.750171
11862	5	1-3134	11858	22	73	575	85	2012-08-02 05:05:05.750171
11858	5	1-3362	11858	73	11	935	23	2012-08-02 05:05:05.750171
11864	5	1-3008	11863	39	43	305	73	2012-08-02 05:05:05.750171
11865	5	1-1603	11863	43	71	239	71	2012-08-02 05:05:05.750171
11866	5	1-2822	11863	71	81	143	16	2012-08-02 05:05:05.750171
11867	5	1-3141	11863	81	73	184	85	2012-08-02 05:05:05.750171
11863	5	1-3362	11863	73	39	300	23	2012-08-02 05:05:05.750171
11869	4	1-3008	11868	39	50	1333	73	2012-08-02 05:05:05.750171
11870	4	1-2413	11868	50	22	689	50	2012-08-02 05:05:05.750171
11871	4	1-3134	11868	22	73	800	85	2012-08-02 05:05:05.750171
11868	4	1-3362	11868	73	39	1309	23	2012-08-02 05:05:05.750171
11873	5	1-2729	11872	71	35	542	61	2012-08-02 05:05:06.976718
11874	5	1-2632	11872	35	92	359	17	2012-08-02 05:05:06.976718
11875	5	1-3198	11872	92	42	561	79	2012-08-02 05:05:06.976718
11876	5	1-2711	11872	42	47	775	37	2012-08-02 05:05:06.976718
11872	5	1-3365	11872	47	71	1212	68	2012-08-02 05:05:06.976718
11878	5	1-2881	11877	93	77	246	56	2012-08-02 05:05:07.363069
11879	5	1-2199	11877	77	32	316	15	2012-08-02 05:05:07.363069
11880	5	1-2909	11877	32	55	251	70	2012-08-02 05:05:07.363069
11881	5	1-3346	11877	55	78	180	97	2012-08-02 05:05:07.363069
11877	5	1-3366	11877	78	93	509	84	2012-08-02 05:05:07.363069
11883	4	1-2951	11882	15	6	1148	9	2012-08-02 05:05:07.760849
11884	4	1-2658	11882	6	8	777	76	2012-08-02 05:05:07.760849
11885	4	1-3172	11882	8	55	612	61	2012-08-02 05:05:07.760849
11882	4	1-3367	11882	55	15	1544	74	2012-08-02 05:05:07.760849
11887	4	1-2951	11886	15	37	988	9	2012-08-02 05:05:07.760849
11888	4	1-2773	11886	37	71	1199	68	2012-08-02 05:05:07.760849
11889	4	1-2729	11886	71	55	415	61	2012-08-02 05:05:07.760849
11886	4	1-3367	11886	55	15	1179	74	2012-08-02 05:05:07.760849
11891	5	1-2722	11890	51	42	1025	79	2012-08-02 05:05:07.760849
11892	5	1-2711	11890	42	47	1133	37	2012-08-02 05:05:07.760849
11893	5	1-3365	11890	47	71	1418	68	2012-08-02 05:05:07.760849
11894	5	1-2729	11890	71	55	507	61	2012-08-02 05:05:07.760849
11890	5	1-3367	11890	55	51	1489	74	2012-08-02 05:05:07.760849
11896	5	1-3229	11895	24	46	2994	22	2012-08-02 05:05:08.578536
11897	5	1-3151	11895	46	42	1906	79	2012-08-02 05:05:08.578536
11898	5	1-2711	11895	42	47	2088	37	2012-08-02 05:05:08.578536
11899	5	1-3365	11895	47	73	2592	68	2012-08-02 05:05:08.578536
11895	5	1-3369	11895	73	24	1662	43	2012-08-02 05:05:08.578536
11901	5	1-3229	11900	24	91	1929	22	2012-08-02 05:05:08.578536
11902	5	1-3200	11900	91	73	873	85	2012-08-02 05:05:08.578536
11903	5	1-3362	11900	73	77	1369	23	2012-08-02 05:05:08.578536
11904	5	1-3295	11900	77	73	1361	68	2012-08-02 05:05:08.578536
11900	5	1-3369	11900	73	24	1012	43	2012-08-02 05:05:08.578536
11906	4	1-3146	11905	9	39	1514	38	2012-08-02 05:05:08.578536
11907	4	1-3021	11905	39	47	2071	37	2012-08-02 05:05:08.578536
11908	4	1-3365	11905	47	73	3212	68	2012-08-02 05:05:08.578536
11905	4	1-3369	11905	73	9	2572	43	2012-08-02 05:05:08.578536
11910	5	1-1888	11909	24	81	1197	16	2012-08-02 05:05:08.578536
11911	5	1-3141	11909	81	73	1575	85	2012-08-02 05:05:08.578536
11912	5	1-3362	11909	73	77	2662	23	2012-08-02 05:05:08.578536
11913	5	1-3295	11909	77	73	3073	68	2012-08-02 05:05:08.578536
11909	5	1-3369	11909	73	24	2463	43	2012-08-02 05:05:08.578536
11915	5	1-3366	11914	78	94	1071	84	2012-08-02 05:05:08.98762
11916	5	1-2082	11914	94	27	615	48	2012-08-02 05:05:08.98762
11917	5	1-3088	11914	27	71	853	10	2012-08-02 05:05:08.98762
11918	5	1-3048	11914	71	6	410	83	2012-08-02 05:05:08.98762
11914	5	1-3370	11914	6	78	565	97	2012-08-02 05:05:08.98762
11920	4	1-3366	11919	78	5	3674	84	2012-08-02 05:05:08.98762
11921	4	1-2401	11919	5	20	2927	23	2012-08-02 05:05:08.98762
11922	4	1-1979	11919	20	6	1151	83	2012-08-02 05:05:08.98762
11919	4	1-3370	11919	6	78	1753	97	2012-08-02 05:05:08.98762
11924	5	1-3251	11923	58	51	1603	65	2012-08-02 05:05:08.98762
11925	5	1-2253	11923	51	23	428	35	2012-08-02 05:05:08.98762
11926	5	1-2163	11923	23	12	634	31	2012-08-02 05:05:08.98762
11927	5	1-2151	11923	12	6	417	83	2012-08-02 05:05:08.98762
11923	5	1-3370	11923	6	58	763	97	2012-08-02 05:05:08.98762
11929	4	1-3366	11928	78	86	1576	84	2012-08-02 05:05:08.98762
11930	4	1-2575	11928	86	20	652	23	2012-08-02 05:05:08.98762
11931	4	1-1979	11928	20	6	319	83	2012-08-02 05:05:08.98762
11928	4	1-3370	11928	6	78	605	97	2012-08-02 05:05:08.98762
11933	5	1-1926	11932	3	74	427	58	2012-08-02 05:05:08.98762
11934	5	1-327	11932	74	24	507	40	2012-08-02 05:05:08.98762
11935	5	1-3180	11932	24	25	301	52	2012-08-02 05:05:08.98762
11936	5	1-2518	11932	25	6	240	83	2012-08-02 05:05:08.98762
11932	5	1-3370	11932	6	3	456	97	2012-08-02 05:05:08.98762
11938	4	1-3251	11937	58	84	1618	65	2012-08-02 05:05:08.98762
11939	4	1-1986	11937	84	12	526	31	2012-08-02 05:05:08.98762
11940	4	1-2151	11937	12	6	369	83	2012-08-02 05:05:08.98762
11937	4	1-3370	11937	6	58	721	97	2012-08-02 05:05:08.98762
11942	5	1-3057	11941	69	12	1579	85	2012-08-02 05:05:09.816388
11943	5	1-3310	11941	12	12	1399	13	2012-08-02 05:05:09.816388
11944	5	1-3240	11941	12	51	625	74	2012-08-02 05:05:09.816388
11945	5	1-2722	11941	51	99	430	79	2012-08-02 05:05:09.816388
11941	5	1-3372	11941	99	69	2628	45	2012-08-02 05:05:09.816388
11947	5	1-2939	11946	89	14	3116	98	2012-08-02 05:05:10.225225
11948	5	1-3237	11946	14	46	1473	22	2012-08-02 05:05:10.225225
11949	5	1-3151	11946	46	99	916	79	2012-08-02 05:05:10.225225
11950	5	1-3372	11946	99	77	5433	45	2012-08-02 05:05:10.225225
11946	5	1-3373	11946	77	89	1878	30	2012-08-02 05:05:10.225225
11952	5	1-3280	11951	42	36	3270	77	2012-08-02 05:05:10.720382
11953	5	1-2276	11951	36	85	1533	81	2012-08-02 05:05:10.720382
11954	5	1-2999	11951	85	18	788	84	2012-08-02 05:05:10.720382
11955	5	1-2356	11951	18	49	1201	50	2012-08-02 05:05:10.720382
11951	5	1-3374	11951	49	42	1487	44	2012-08-02 05:05:10.720382
11957	4	1-1510	11956	32	74	787	42	2012-08-02 05:05:10.720382
11958	4	1-2904	11956	74	2	750	32	2012-08-02 05:05:10.720382
11959	4	1-3002	11956	2	49	397	50	2012-08-02 05:05:10.720382
11956	4	1-3374	11956	49	32	611	44	2012-08-02 05:05:10.720382
11961	4	1-3069	11960	59	46	1149	82	2012-08-02 05:05:10.720382
11962	4	1-2942	11960	46	18	553	84	2012-08-02 05:05:10.720382
11963	4	1-2356	11960	18	49	1062	50	2012-08-02 05:05:10.720382
11960	4	1-3374	11960	49	59	1657	44	2012-08-02 05:05:10.720382
11965	4	1-2870	11964	6	78	3242	33	2012-08-02 05:05:11.167981
11966	4	1-1999	11964	78	6	650	83	2012-08-02 05:05:11.167981
11967	4	1-3370	11964	6	64	1079	97	2012-08-02 05:05:11.167981
11964	4	1-3375	11964	64	6	2519	34	2012-08-02 05:05:11.167981
11969	6	1-1970	11968	98	64	186	89	2012-08-02 05:05:11.167981
11970	6	1-3079	11968	64	25	236	47	2012-08-02 05:05:11.167981
11971	6	1-2791	11968	25	56	138	36	2012-08-02 05:05:11.167981
11972	6	1-3103	11968	56	6	104	83	2012-08-02 05:05:11.167981
11973	6	1-3370	11968	6	64	119	97	2012-08-02 05:05:11.167981
11968	6	1-3375	11968	64	98	326	34	2012-08-02 05:05:11.167981
11975	5	1-1997	11974	65	57	1201	79	2012-08-02 05:05:11.167981
11976	5	1-1680	11974	57	71	875	10	2012-08-02 05:05:11.167981
11977	5	1-3048	11974	71	6	486	83	2012-08-02 05:05:11.167981
11978	5	1-3370	11974	6	64	774	97	2012-08-02 05:05:11.167981
11974	5	1-3375	11974	64	65	2186	34	2012-08-02 05:05:11.167981
11980	5	1-2915	11979	87	91	1039	70	2012-08-02 05:05:12.47042
11981	5	1-3216	11979	91	27	892	14	2012-08-02 05:05:12.47042
11982	5	1-3312	11979	27	72	2457	53	2012-08-02 05:05:12.47042
11983	5	1-3319	11979	72	81	821	78	2012-08-02 05:05:12.47042
11979	5	1-3378	11979	81	87	1170	27	2012-08-02 05:05:12.47042
11985	5	1-3133	11984	37	98	710	56	2012-08-02 05:05:12.47042
11986	5	1-2146	11984	98	77	666	23	2012-08-02 05:05:12.47042
11987	5	1-3295	11984	77	39	684	68	2012-08-02 05:05:12.47042
11988	5	1-3279	11984	39	81	1141	78	2012-08-02 05:05:12.47042
11984	5	1-3378	11984	81	37	1667	27	2012-08-02 05:05:12.47042
11990	5	1-2217	11989	72	69	687	86	2012-08-02 05:05:12.47042
11991	5	1-3330	11989	69	25	2312	58	2012-08-02 05:05:12.47042
11992	5	1-3334	11989	25	61	2456	49	2012-08-02 05:05:12.47042
11993	5	1-3327	11989	61	81	2854	78	2012-08-02 05:05:12.47042
11989	5	1-3378	11989	81	72	4606	27	2012-08-02 05:05:12.47042
11995	5	1-2817	11994	9	100	586	96	2012-08-02 05:05:14.889524
11996	5	1-2762	11994	100	22	301	60	2012-08-02 05:05:14.889524
11997	5	1-3241	11994	22	23	738	29	2012-08-02 05:05:14.889524
11998	5	1-3227	11994	23	12	444	3	2012-08-02 05:05:14.889524
11994	5	1-3384	11994	12	9	692	35	2012-08-02 05:05:14.889524
12000	3	1-2817	11999	9	84	3039	96	2012-08-02 05:05:14.889524
12001	3	1-2992	11999	84	12	2258	3	2012-08-02 05:05:14.889524
11999	3	1-3384	11999	12	9	3555	35	2012-08-02 05:05:14.889524
12003	5	1-2537	12002	14	35	826	28	2012-08-02 05:05:16.127046
12004	5	1-3358	12002	35	66	1417	29	2012-08-02 05:05:16.127046
12005	5	1-2428	12002	66	89	729	30	2012-08-02 05:05:16.127046
12006	5	1-2939	12002	89	44	1189	98	2012-08-02 05:05:16.127046
12002	5	1-3387	12002	44	14	1126	47	2012-08-02 05:05:16.127046
12007	1	1-2428	\N	66	66	1122	30	2012-08-02 05:05:16.127046
12009	5	1-2268	12008	96	35	645	28	2012-08-02 05:05:16.535741
12010	5	1-3358	12008	35	53	1379	29	2012-08-02 05:05:16.535741
12008	5	1-3388	12008	66	96	683	75	2012-08-02 05:05:16.535741
12011	5	1-3186	12008	53	63	5653	14	2012-08-02 05:05:16.535741
12012	5	1-3243	12008	63	66	908	65	2012-08-02 05:05:16.535741
16090	4	1-4588	16089	35	23	695	92	2012-08-02 05:13:35.935381
12014	5	1-3211	12013	14	93	294	2	2012-08-02 05:05:16.535741
12015	5	1-3249	12013	93	100	485	71	2012-08-02 05:05:16.535741
12016	5	1-2802	12013	100	98	225	99	2012-08-02 05:05:16.535741
12017	5	1-2667	12013	98	66	86	65	2012-08-02 05:05:16.535741
12013	5	1-3388	12013	66	14	66	75	2012-08-02 05:05:16.535741
16091	4	1-4048	16089	23	83	569	70	2012-08-02 05:13:35.935381
12019	5	1-2496	12018	4	44	3039	98	2012-08-02 05:05:16.535741
12020	5	1-3387	12018	44	25	3759	47	2012-08-02 05:05:16.535741
12021	5	1-2791	12018	25	12	3978	36	2012-08-02 05:05:16.535741
12022	5	1-3162	12018	12	66	3781	65	2012-08-02 05:05:16.535741
12018	5	1-3388	12018	66	4	2979	75	2012-08-02 05:05:16.535741
12024	5	1-2791	12023	25	43	538	36	2012-08-02 05:05:18.955351
12025	5	1-2892	12023	43	93	511	63	2012-08-02 05:05:18.955351
12026	5	1-1517	12023	93	73	339	85	2012-08-02 05:05:18.955351
12027	5	1-3362	12023	73	5	478	23	2012-08-02 05:05:18.955351
12023	5	1-3394	12023	5	25	611	47	2012-08-02 05:05:18.955351
12029	3	1-3258	12028	63	93	1342	5	2012-08-02 05:05:19.353164
12030	3	1-3359	12028	93	95	429	100	2012-08-02 05:05:19.353164
12028	3	1-3395	12028	95	63	9402	7	2012-08-02 05:05:19.353164
12032	4	1-880	12031	68	93	1605	5	2012-08-02 05:05:20.137914
12033	4	1-3359	12031	93	29	724	100	2012-08-02 05:05:20.137914
12034	4	1-3255	12031	29	53	2318	71	2012-08-02 05:05:20.137914
12031	4	1-3397	12031	53	68	1248	96	2012-08-02 05:05:20.137914
12036	5	1-1938	12035	4	27	308	14	2012-08-02 05:05:20.137914
12037	5	1-3312	12035	27	98	819	53	2012-08-02 05:05:20.137914
12038	5	1-1901	12035	98	93	253	2	2012-08-02 05:05:20.137914
12039	5	1-3249	12035	93	53	353	71	2012-08-02 05:05:20.137914
12035	5	1-3397	12035	53	4	214	96	2012-08-02 05:05:20.137914
12041	4	1-1551	12040	84	27	461	100	2012-08-02 05:05:20.524468
12042	4	1-3041	12040	27	66	891	32	2012-08-02 05:05:20.524468
12043	4	1-2908	12040	66	19	107	86	2012-08-02 05:05:20.524468
12040	4	1-3398	12040	19	84	1373	74	2012-08-02 05:05:20.524468
12045	5	1-2843	12044	22	64	1080	31	2012-08-02 05:05:20.524468
12046	5	1-2818	12044	64	25	120	47	2012-08-02 05:05:20.524468
12047	5	1-2791	12044	25	91	103	36	2012-08-02 05:05:20.524468
12048	5	1-3083	12044	91	19	63	86	2012-08-02 05:05:20.524468
12044	5	1-3398	12044	19	22	1039	74	2012-08-02 05:05:20.524468
12050	2	1-1133	12049	29	88	1736	51	2012-08-02 05:05:20.955505
12049	2	1-3399	12049	88	29	8982	54	2012-08-02 05:05:20.955505
12052	6	1-3342	12051	93	20	359	41	2012-08-02 05:05:22.568362
12053	6	1-3381	12051	20	88	2006	90	2012-08-02 05:05:22.568362
12054	6	1-3287	12051	88	13	1554	57	2012-08-02 05:05:22.568362
12055	6	1-981	12051	13	56	352	6	2012-08-02 05:05:22.568362
12056	6	1-2878	12051	56	6	239	19	2012-08-02 05:05:22.568362
12051	6	1-3403	12051	6	93	401	18	2012-08-02 05:05:22.568362
12058	4	1-3015	12057	17	17	941	26	2012-08-02 05:05:24.556891
12059	4	1-211	12057	17	19	186	86	2012-08-02 05:05:24.556891
12060	4	1-3398	12057	19	2	1773	74	2012-08-02 05:05:24.556891
12057	4	1-3408	12057	2	17	1792	95	2012-08-02 05:05:24.556891
12062	6	1-2366	12061	12	53	568	25	2012-08-02 05:05:24.556891
12063	6	1-3386	12061	53	30	139	97	2012-08-02 05:05:24.556891
12064	6	1-3199	12061	30	66	514	32	2012-08-02 05:05:24.556891
12065	6	1-2908	12061	66	19	63	86	2012-08-02 05:05:24.556891
12066	6	1-3398	12061	19	2	812	74	2012-08-02 05:05:24.556891
12061	6	1-3408	12061	2	12	1119	95	2012-08-02 05:05:24.556891
12068	5	1-2315	12067	27	30	1260	97	2012-08-02 05:05:24.96593
12069	5	1-3199	12067	30	63	6287	32	2012-08-02 05:05:24.96593
12070	5	1-3166	12067	63	11	2202	73	2012-08-02 05:05:24.96593
12071	5	1-902	12067	11	17	982	5	2012-08-02 05:05:24.96593
12067	5	1-3409	12067	17	27	4133	37	2012-08-02 05:05:24.96593
12073	4	1-3079	12072	64	100	173	47	2012-08-02 05:05:26.147891
12074	4	1-2216	12072	100	17	77	5	2012-08-02 05:05:26.147891
12075	4	1-3409	12072	17	13	219	37	2012-08-02 05:05:26.147891
12072	4	1-3412	12072	13	64	176	89	2012-08-02 05:05:26.147891
12077	5	1-3256	12076	86	44	1699	98	2012-08-02 05:05:26.147891
12078	5	1-3387	12076	44	100	1359	47	2012-08-02 05:05:26.147891
12079	5	1-2216	12076	100	17	673	5	2012-08-02 05:05:26.147891
12080	5	1-3409	12076	17	13	2124	37	2012-08-02 05:05:26.147891
12076	5	1-3412	12076	13	86	1899	89	2012-08-02 05:05:26.147891
12082	4	1-3380	12081	30	15	1692	21	2012-08-02 05:05:26.147891
12083	4	1-3102	12081	15	78	1157	84	2012-08-02 05:05:26.147891
12084	4	1-1816	12081	78	13	771	37	2012-08-02 05:05:26.147891
12081	4	1-3412	12081	13	30	915	89	2012-08-02 05:05:26.147891
12086	5	1-3256	12085	86	44	648	98	2012-08-02 05:05:26.147891
12087	5	1-3387	12085	44	14	710	47	2012-08-02 05:05:26.147891
12088	5	1-2537	12085	14	92	603	28	2012-08-02 05:05:26.147891
12089	5	1-1130	12085	92	13	432	37	2012-08-02 05:05:26.147891
12085	5	1-3412	12085	13	86	529	89	2012-08-02 05:05:26.147891
12091	3	1-2043	12090	2	27	585	100	2012-08-02 05:05:26.54592
12092	3	1-3041	12090	27	72	1247	32	2012-08-02 05:05:26.54592
12090	3	1-3413	12090	72	2	1050	69	2012-08-02 05:05:26.54592
12094	6	1-3020	12093	68	24	94	40	2012-08-02 05:05:27.352488
12095	6	1-3180	12093	24	83	55	52	2012-08-02 05:05:27.352488
12096	6	1-1404	12093	83	44	16	98	2012-08-02 05:05:27.352488
12097	6	1-3387	12093	44	15	18	47	2012-08-02 05:05:27.352488
12098	6	1-2572	12093	15	69	18	33	2012-08-02 05:05:27.352488
12093	6	1-3415	12093	69	68	75	100	2012-08-02 05:05:27.352488
12100	5	1-2872	12099	85	6	2334	77	2012-08-02 05:05:27.352488
12101	5	1-2913	12099	6	86	1259	39	2012-08-02 05:05:27.352488
12102	5	1-2326	12099	86	70	1127	42	2012-08-02 05:05:27.352488
12103	5	1-3104	12099	70	69	608	33	2012-08-02 05:05:27.352488
12099	5	1-3415	12099	69	85	2544	100	2012-08-02 05:05:27.352488
12105	3	1-3156	12104	60	69	4628	69	2012-08-02 05:05:27.750218
12106	3	1-2875	12104	69	92	1833	9	2012-08-02 05:05:27.750218
12104	3	1-3416	12104	92	60	4116	34	2012-08-02 05:05:27.750218
12108	6	1-3107	12107	2	18	642	22	2012-08-02 05:05:28.159395
12109	6	1-3277	12107	18	35	515	81	2012-08-02 05:05:28.159395
12110	6	1-3323	12107	35	20	186	6	2012-08-02 05:05:28.159395
12111	6	1-3393	12107	20	19	61	86	2012-08-02 05:05:28.159395
12112	6	1-3398	12107	19	95	934	74	2012-08-02 05:05:28.159395
12107	6	1-3417	12107	95	2	731	63	2012-08-02 05:05:28.159395
12114	5	1-3338	12113	61	70	188	20	2012-08-02 05:05:28.159395
12115	5	1-2616	12113	70	17	73	26	2012-08-02 05:05:28.159395
12116	5	1-211	12113	17	19	16	86	2012-08-02 05:05:28.159395
12117	5	1-3398	12113	19	95	248	74	2012-08-02 05:05:28.159395
12113	5	1-3417	12113	95	61	195	63	2012-08-02 05:05:28.159395
16092	4	1-3602	16089	83	32	114	64	2012-08-02 05:13:35.935381
12119	5	1-3146	12118	9	25	432	38	2012-08-02 05:05:29.385295
12120	5	1-3355	12118	25	93	742	58	2012-08-02 05:05:29.385295
12121	5	1-1748	12118	93	22	675	50	2012-08-02 05:05:29.385295
12122	5	1-3134	12118	22	94	588	85	2012-08-02 05:05:29.385295
12118	5	1-3420	12118	94	9	1011	43	2012-08-02 05:05:29.385295
16089	4	1-4634	16089	32	35	406	28	2012-08-02 05:13:35.935381
12124	3	1-1888	12123	24	81	236	16	2012-08-02 05:05:29.385295
12125	3	1-3141	12123	81	94	272	85	2012-08-02 05:05:29.385295
12123	3	1-3420	12123	94	24	555	43	2012-08-02 05:05:29.385295
12127	4	1-2654	12126	79	91	1485	79	2012-08-02 05:05:29.385295
12128	4	1-2462	12126	91	22	1057	50	2012-08-02 05:05:29.385295
12129	4	1-3134	12126	22	94	1203	85	2012-08-02 05:05:29.385295
12126	4	1-3420	12126	94	79	2707	43	2012-08-02 05:05:29.385295
16094	4	1-3211	16093	14	53	158	2	2012-08-02 05:13:36.764155
12131	4	1-6	12130	6	10	1311	10	2012-08-02 05:05:30.191873
12132	4	1-2740	12130	10	20	304	41	2012-08-02 05:05:30.191873
12133	4	1-3381	12130	20	25	1539	90	2012-08-02 05:05:30.191873
12130	4	1-3422	12130	25	6	2163	11	2012-08-02 05:05:30.191873
12135	3	1-3383	12134	100	74	5981	3	2012-08-02 05:05:30.590179
12136	3	1-2036	12134	74	29	3830	14	2012-08-02 05:05:30.590179
12134	3	1-3423	12134	29	100	7000	12	2012-08-02 05:05:30.590179
12138	5	1-3383	12137	100	70	609	3	2012-08-02 05:05:30.590179
12139	5	1-2033	12137	70	35	550	81	2012-08-02 05:05:30.590179
12140	5	1-3323	12137	35	49	232	6	2012-08-02 05:05:30.590179
12141	5	1-2339	12137	49	29	224	14	2012-08-02 05:05:30.590179
12137	5	1-3423	12137	29	100	540	12	2012-08-02 05:05:30.590179
12143	4	1-1794	12142	67	69	205	58	2012-08-02 05:05:30.590179
12144	4	1-2668	12142	69	91	241	57	2012-08-02 05:05:30.590179
12145	4	1-2371	12142	91	29	305	14	2012-08-02 05:05:30.590179
12142	4	1-3423	12142	29	67	750	12	2012-08-02 05:05:30.590179
12147	4	1-1926	12146	3	93	92	58	2012-08-02 05:05:30.987597
12148	4	1-1748	12146	93	22	72	50	2012-08-02 05:05:30.987597
12149	4	1-3134	12146	22	66	54	85	2012-08-02 05:05:30.987597
12146	4	1-3424	12146	66	3	145	97	2012-08-02 05:05:30.987597
12151	5	1-1926	12150	3	42	3506	58	2012-08-02 05:05:30.987597
12152	5	1-1232	12150	42	97	1540	76	2012-08-02 05:05:30.987597
12153	5	1-1736	12150	97	22	1129	50	2012-08-02 05:05:30.987597
12154	5	1-3134	12150	22	66	1134	85	2012-08-02 05:05:30.987597
12150	5	1-3424	12150	66	3	4120	97	2012-08-02 05:05:30.987597
12156	6	1-3156	12155	60	9	4197	69	2012-08-02 05:05:31.783059
12157	6	1-1158	12155	9	13	1131	37	2012-08-02 05:05:31.783059
12158	6	1-3412	12155	13	30	1466	89	2012-08-02 05:05:31.783059
12159	6	1-3380	12155	30	41	2962	21	2012-08-02 05:05:31.783059
12160	6	1-3354	12155	41	83	7133	54	2012-08-02 05:05:31.783059
12155	6	1-3426	12155	83	60	3418	34	2012-08-02 05:05:31.783059
12162	4	1-3034	12161	55	35	694	28	2012-08-02 05:05:32.20307
12163	4	1-3358	12161	35	89	1280	29	2012-08-02 05:05:32.20307
12164	4	1-3171	12161	89	95	2432	10	2012-08-02 05:05:32.20307
12161	4	1-3427	12161	95	55	918	33	2012-08-02 05:05:32.20307
12166	3	1-3387	12165	44	25	1841	47	2012-08-02 05:05:33.81619
12167	3	1-3187	12165	25	9	521	26	2012-08-02 05:05:33.81619
12165	3	1-3431	12165	9	44	5024	98	2012-08-02 05:05:33.81619
12169	5	1-1354	12168	51	70	299	42	2012-08-02 05:05:33.81619
12170	5	1-3104	12168	70	69	97	33	2012-08-02 05:05:33.81619
12171	5	1-3415	12168	69	26	245	100	2012-08-02 05:05:33.81619
12172	5	1-2948	12168	26	9	35	26	2012-08-02 05:05:33.81619
12168	5	1-3431	12168	9	51	619	98	2012-08-02 05:05:33.81619
12174	5	1-2180	12173	3	91	74	57	2012-08-02 05:05:35.02039
12175	5	1-2371	12173	91	27	89	14	2012-08-02 05:05:35.02039
12176	5	1-3312	12173	27	49	243	53	2012-08-02 05:05:35.02039
12177	5	1-2969	12173	49	49	101	76	2012-08-02 05:05:35.02039
12173	5	1-3434	12173	49	3	80	28	2012-08-02 05:05:35.02039
12179	3	1-3314	12178	19	48	3437	49	2012-08-02 05:05:35.407078
12180	3	1-1022	12178	48	18	1864	22	2012-08-02 05:05:35.407078
12178	3	1-3435	12178	18	19	8123	90	2012-08-02 05:05:35.407078
12182	5	1-2878	12181	56	79	424	19	2012-08-02 05:05:36.224518
12183	5	1-3208	12181	79	19	863	90	2012-08-02 05:05:36.224518
12184	5	1-3314	12181	19	21	498	49	2012-08-02 05:05:36.224518
12185	5	1-39	12181	21	67	182	52	2012-08-02 05:05:36.224518
12181	5	1-3437	12181	67	56	695	6	2012-08-02 05:05:36.224518
12187	5	1-2878	12186	56	79	651	19	2012-08-02 05:05:36.224518
12188	5	1-3208	12186	79	40	1412	90	2012-08-02 05:05:36.224518
12189	5	1-3404	12186	40	63	877	70	2012-08-02 05:05:36.224518
12190	5	1-1839	12186	63	67	245	52	2012-08-02 05:05:36.224518
12186	5	1-3437	12186	67	56	999	6	2012-08-02 05:05:36.224518
12192	5	1-1517	12191	93	66	117	85	2012-08-02 05:05:36.633398
12193	5	1-3424	12191	66	3	328	97	2012-08-02 05:05:36.633398
12194	5	1-1926	12191	3	12	216	58	2012-08-02 05:05:36.633398
12195	5	1-1486	12191	12	36	127	51	2012-08-02 05:05:36.633398
12191	5	1-3438	12191	36	93	227	63	2012-08-02 05:05:36.633398
12197	4	1-1517	12196	93	94	754	85	2012-08-02 05:05:36.633398
12198	4	1-3420	12196	94	61	1386	43	2012-08-02 05:05:36.633398
12199	4	1-3242	12196	61	36	572	51	2012-08-02 05:05:36.633398
12196	4	1-3438	12196	36	93	1224	63	2012-08-02 05:05:36.633398
12201	3	1-2690	12200	50	99	1868	80	2012-08-02 05:05:36.633398
12202	3	1-2890	12200	99	36	878	51	2012-08-02 05:05:36.633398
12200	3	1-3438	12200	36	50	2256	63	2012-08-02 05:05:36.633398
12204	4	1-3241	12203	22	23	4454	29	2012-08-02 05:05:37.055386
12205	4	1-3227	12203	23	6	1989	3	2012-08-02 05:05:37.055386
12206	4	1-3160	12203	6	4	1322	48	2012-08-02 05:05:37.055386
12203	4	1-3439	12203	4	22	2445	60	2012-08-02 05:05:37.055386
12208	5	1-518	12207	44	36	477	51	2012-08-02 05:05:37.055386
12209	5	1-3438	12207	36	33	976	63	2012-08-02 05:05:37.055386
12210	5	1-3080	12207	33	99	779	93	2012-08-02 05:05:37.055386
12211	5	1-2412	12207	99	4	249	48	2012-08-02 05:05:37.055386
12207	5	1-3439	12207	4	44	531	60	2012-08-02 05:05:37.055386
12213	5	1-3213	12212	5	47	1040	69	2012-08-02 05:05:37.055386
12214	5	1-2725	12212	47	85	1199	57	2012-08-02 05:05:37.055386
12215	5	1-2977	12212	85	6	517	3	2012-08-02 05:05:37.055386
12216	5	1-3160	12212	6	4	460	48	2012-08-02 05:05:37.055386
12212	5	1-3439	12212	4	5	1140	60	2012-08-02 05:05:37.055386
12218	4	1-2004	12217	84	28	106	44	2012-08-02 05:05:37.055386
12219	4	1-3284	12217	28	99	179	93	2012-08-02 05:05:37.055386
12220	4	1-2412	12217	99	4	67	48	2012-08-02 05:05:37.055386
12217	4	1-3439	12217	4	84	165	60	2012-08-02 05:05:37.055386
16095	4	1-4545	16093	53	3	48	61	2012-08-02 05:13:36.764155
12222	2	1-1585	12221	60	13	4832	74	2012-08-02 05:05:37.464449
12221	2	1-3440	12221	13	60	5164	35	2012-08-02 05:05:37.464449
16096	4	1-59	16093	3	96	20	68	2012-08-02 05:13:36.764155
12224	6	1-2807	12223	60	69	511	86	2012-08-02 05:05:37.873127
12225	6	1-3330	12223	69	55	1724	58	2012-08-02 05:05:37.873127
12226	6	1-2101	12223	55	4	513	65	2012-08-02 05:05:37.873127
12227	6	1-3371	12223	4	63	710	8	2012-08-02 05:05:37.873127
12228	6	1-690	12223	63	78	949	92	2012-08-02 05:05:37.873127
12223	6	1-3441	12223	78	60	1504	1	2012-08-02 05:05:37.873127
12230	5	1-3357	12229	20	88	905	61	2012-08-02 05:05:37.873127
12231	5	1-194	12229	88	4	310	65	2012-08-02 05:05:37.873127
12232	5	1-3371	12229	4	63	442	8	2012-08-02 05:05:37.873127
12233	5	1-690	12229	63	78	610	92	2012-08-02 05:05:37.873127
12229	5	1-3441	12229	78	20	998	1	2012-08-02 05:05:37.873127
12235	6	1-2773	12234	37	39	118	68	2012-08-02 05:05:38.281941
12236	6	1-3279	12234	39	81	144	78	2012-08-02 05:05:38.281941
12237	6	1-3378	12234	81	87	154	27	2012-08-02 05:05:38.281941
12238	6	1-2915	12234	87	87	129	70	2012-08-02 05:05:38.281941
12239	6	1-2270	12234	87	22	50	82	2012-08-02 05:05:38.281941
12234	6	1-3442	12234	22	37	119	9	2012-08-02 05:05:38.281941
12241	5	1-3361	12240	55	11	813	7	2012-08-02 05:05:38.281941
12242	5	1-2341	12240	11	32	528	15	2012-08-02 05:05:38.281941
12243	5	1-2909	12240	32	87	388	70	2012-08-02 05:05:38.281941
12244	5	1-2270	12240	87	22	173	82	2012-08-02 05:05:38.281941
12240	5	1-3442	12240	22	55	589	9	2012-08-02 05:05:38.281941
12246	4	1-2658	12245	6	49	2154	76	2012-08-02 05:05:38.281941
12247	4	1-3434	12245	49	68	1694	28	2012-08-02 05:05:38.281941
12248	4	1-1982	12245	68	22	698	82	2012-08-02 05:05:38.281941
12245	4	1-3442	12245	22	6	2384	9	2012-08-02 05:05:38.281941
12250	4	1-3171	12249	89	95	4941	10	2012-08-02 05:05:39.872961
12251	4	1-3427	12249	95	69	1635	33	2012-08-02 05:05:39.872961
12252	4	1-3415	12249	69	18	5423	100	2012-08-02 05:05:39.872961
12249	4	1-3446	12249	18	89	2967	29	2012-08-02 05:05:39.872961
12254	6	1-3443	12253	21	72	693	60	2012-08-02 05:05:39.872961
12255	6	1-3129	12253	72	76	470	99	2012-08-02 05:05:39.872961
12256	6	1-3337	12253	76	15	326	47	2012-08-02 05:05:39.872961
12257	6	1-2572	12253	15	69	278	33	2012-08-02 05:05:39.872961
12258	6	1-3415	12253	69	18	979	100	2012-08-02 05:05:39.872961
12253	6	1-3446	12253	18	21	568	29	2012-08-02 05:05:39.872961
12260	6	1-2977	12259	85	70	224	3	2012-08-02 05:05:42.336437
12261	6	1-2033	12259	70	35	206	81	2012-08-02 05:05:42.336437
12262	6	1-3323	12259	35	56	89	6	2012-08-02 05:05:42.336437
12263	6	1-2878	12259	56	6	57	19	2012-08-02 05:05:42.336437
12264	6	1-3403	12259	6	21	91	18	2012-08-02 05:05:42.336437
12259	6	1-3452	12259	21	85	493	57	2012-08-02 05:05:42.336437
12266	3	1-3442	12265	22	69	4447	9	2012-08-02 05:05:43.165196
12267	3	1-1029	12265	69	19	892	25	2012-08-02 05:05:43.165196
12265	3	1-3454	12265	19	22	1689	82	2012-08-02 05:05:43.165196
12269	5	1-3442	12268	22	31	688	9	2012-08-02 05:05:43.165196
12270	5	1-3145	12268	31	13	465	33	2012-08-02 05:05:43.165196
12271	5	1-2758	12268	13	52	280	67	2012-08-02 05:05:43.165196
12272	5	1-73	12268	52	19	121	25	2012-08-02 05:05:43.165196
12268	5	1-3454	12268	19	22	245	82	2012-08-02 05:05:43.165196
12274	4	1-3442	12273	22	83	936	9	2012-08-02 05:05:43.165196
12275	4	1-3192	12273	83	89	729	93	2012-08-02 05:05:43.165196
12276	4	1-3329	12273	89	19	146	25	2012-08-02 05:05:43.165196
12273	4	1-3454	12273	19	22	313	82	2012-08-02 05:05:43.165196
12278	5	1-3437	12277	67	56	4657	6	2012-08-02 05:05:45.938596
12279	5	1-2878	12277	56	25	2243	19	2012-08-02 05:05:45.938596
12280	5	1-3382	12277	25	33	1770	98	2012-08-02 05:05:45.938596
12281	5	1-3432	12277	33	13	1459	23	2012-08-02 05:05:45.938596
12277	5	1-3461	12277	13	67	1545	52	2012-08-02 05:05:45.938596
12283	5	1-3437	12282	67	30	1200	6	2012-08-02 05:05:45.938596
12284	5	1-2065	12282	30	94	237	85	2012-08-02 05:05:45.938596
12285	5	1-3420	12282	94	87	542	43	2012-08-02 05:05:45.938596
12286	5	1-1895	12282	87	13	186	23	2012-08-02 05:05:45.938596
12282	5	1-3461	12282	13	67	280	52	2012-08-02 05:05:45.938596
12288	2	1-3149	12287	89	17	3889	53	2012-08-02 05:05:46.347129
12287	2	1-3462	12287	17	89	6514	9	2012-08-02 05:05:46.347129
12290	5	1-3412	12289	13	30	449	89	2012-08-02 05:05:46.744977
12291	5	1-3380	12289	30	21	720	21	2012-08-02 05:05:46.744977
12292	5	1-3051	12289	21	35	314	49	2012-08-02 05:05:46.744977
12293	5	1-1906	12289	35	25	213	55	2012-08-02 05:05:46.744977
12289	5	1-3463	12289	25	13	436	37	2012-08-02 05:05:46.744977
12295	4	1-3228	12294	23	51	882	3	2012-08-02 05:05:48.390925
12296	4	1-3466	12294	51	77	608	70	2012-08-02 05:05:48.390925
12297	4	1-3313	12294	77	21	656	66	2012-08-02 05:05:48.390925
12294	4	1-3467	12294	21	23	585	38	2012-08-02 05:05:48.390925
12299	4	1-3085	12298	58	13	2078	74	2012-08-02 05:05:48.390925
12300	4	1-3440	12298	13	60	1745	35	2012-08-02 05:05:48.390925
12301	4	1-1021	12298	60	21	814	66	2012-08-02 05:05:48.390925
12298	4	1-3467	12298	21	58	793	38	2012-08-02 05:05:48.390925
12303	5	1-3436	12302	57	70	1975	15	2012-08-02 05:05:49.208845
12304	5	1-3029	12302	70	72	1403	32	2012-08-02 05:05:49.208845
12305	5	1-3413	12302	72	45	1455	69	2012-08-02 05:05:49.208845
12306	5	1-1434	12302	45	76	332	87	2012-08-02 05:05:49.208845
12302	5	1-3469	12302	76	57	621	75	2012-08-02 05:05:49.208845
12308	3	1-2496	12307	4	70	4943	98	2012-08-02 05:05:49.208845
12309	3	1-3292	12307	70	76	2749	87	2012-08-02 05:05:49.208845
12307	3	1-3469	12307	76	4	5388	75	2012-08-02 05:05:49.208845
12311	5	1-3458	12310	3	71	3592	35	2012-08-02 05:05:50.004101
12312	5	1-1454	12310	71	74	2367	17	2012-08-02 05:05:50.004101
12313	5	1-3223	12310	74	18	962	84	2012-08-02 05:05:50.004101
12314	5	1-2356	12310	18	12	1774	50	2012-08-02 05:05:50.004101
12310	5	1-3471	12310	12	3	7242	5	2012-08-02 05:05:50.004101
12316	5	1-2835	12315	85	37	649	91	2012-08-02 05:05:53.572387
12317	5	1-543	12315	37	22	230	82	2012-08-02 05:05:53.572387
12318	5	1-3442	12315	22	65	518	9	2012-08-02 05:05:53.572387
12319	5	1-3035	12315	65	55	310	51	2012-08-02 05:05:53.572387
12315	5	1-3480	12315	55	85	202	62	2012-08-02 05:05:53.572387
12321	5	1-2928	12320	52	86	2407	88	2012-08-02 05:05:53.572387
12322	5	1-3391	12320	86	37	537	39	2012-08-02 05:05:53.572387
12323	5	1-3429	12320	37	54	148	44	2012-08-02 05:05:53.572387
12324	5	1-3195	12320	54	55	137	51	2012-08-02 05:05:53.572387
12320	5	1-3480	12320	55	52	93	62	2012-08-02 05:05:53.572387
16093	4	1-4636	16093	96	14	47	75	2012-08-02 05:13:36.764155
12326	4	1-2835	12325	85	25	712	91	2012-08-02 05:05:53.572387
12327	4	1-3108	12325	25	66	269	61	2012-08-02 05:05:53.572387
12328	4	1-3031	12325	66	55	281	51	2012-08-02 05:05:53.572387
12325	4	1-3480	12325	55	85	202	62	2012-08-02 05:05:53.572387
12330	5	1-2981	12329	53	60	3433	4	2012-08-02 05:05:53.572387
12331	5	1-3168	12329	60	9	3186	35	2012-08-02 05:05:53.572387
12332	5	1-2817	12329	9	35	2746	96	2012-08-02 05:05:53.572387
12333	5	1-2974	12329	35	55	1543	51	2012-08-02 05:05:53.572387
12329	5	1-3480	12329	55	53	1453	62	2012-08-02 05:05:53.572387
16098	3	1-4293	16097	5	23	2346	56	2012-08-02 05:13:36.764155
12335	5	1-2928	12334	52	88	2675	88	2012-08-02 05:05:53.572387
12336	5	1-3348	12334	88	30	662	37	2012-08-02 05:05:53.572387
12337	5	1-1723	12334	30	12	73	58	2012-08-02 05:05:53.572387
12338	5	1-1486	12334	12	55	65	51	2012-08-02 05:05:53.572387
12334	5	1-3480	12334	55	52	67	62	2012-08-02 05:05:53.572387
16099	3	1-3680	16097	23	96	667	68	2012-08-02 05:13:36.764155
12340	4	1-2981	12339	53	17	650	4	2012-08-02 05:05:53.572387
12341	4	1-2804	12339	17	29	771	54	2012-08-02 05:05:53.572387
12342	4	1-1133	12339	29	55	243	51	2012-08-02 05:05:53.572387
12339	4	1-3480	12339	55	53	251	62	2012-08-02 05:05:53.572387
12343	1	1-543	\N	37	37	1558	82	2012-08-02 05:05:53.572387
12345	3	1-1797	12344	38	4	2050	60	2012-08-02 05:05:53.992739
12346	3	1-3163	12344	4	50	1335	42	2012-08-02 05:05:53.992739
12344	3	1-3481	12344	50	38	6724	82	2012-08-02 05:05:53.992739
12348	5	1-409	12347	67	12	352	31	2012-08-02 05:05:53.992739
12349	5	1-2151	12347	12	100	175	83	2012-08-02 05:05:53.992739
12350	5	1-3450	12347	100	39	155	17	2012-08-02 05:05:53.992739
12351	5	1-2201	12347	39	50	122	42	2012-08-02 05:05:53.992739
12347	5	1-3481	12347	50	67	630	82	2012-08-02 05:05:53.992739
12353	5	1-3240	12352	12	13	101	74	2012-08-02 05:05:54.390585
12354	5	1-3440	12352	13	23	105	35	2012-08-02 05:05:54.390585
12355	5	1-2163	12352	23	12	167	31	2012-08-02 05:05:54.390585
12356	5	1-2151	12352	12	2	122	83	2012-08-02 05:05:54.390585
12352	5	1-3482	12352	2	12	207	13	2012-08-02 05:05:54.390585
12358	5	1-3458	12357	3	9	1119	35	2012-08-02 05:05:55.197061
12359	5	1-2817	12357	9	71	933	96	2012-08-02 05:05:55.197061
12360	5	1-3401	12357	71	93	821	98	2012-08-02 05:05:55.197061
12361	5	1-1417	12357	93	65	526	51	2012-08-02 05:05:55.197061
12357	5	1-3484	12357	65	3	2390	5	2012-08-02 05:05:55.197061
12363	5	1-3359	12362	93	96	3131	100	2012-08-02 05:05:55.197061
12364	5	1-3402	12362	96	18	1205	22	2012-08-02 05:05:55.197061
12365	5	1-3277	12362	18	93	1204	81	2012-08-02 05:05:55.197061
12366	5	1-2728	12362	93	65	1028	51	2012-08-02 05:05:55.197061
12362	5	1-3484	12362	65	93	5183	5	2012-08-02 05:05:55.197061
12368	3	1-3132	12367	27	32	1000	67	2012-08-02 05:06:00.963958
12369	3	1-3485	12367	32	91	835	6	2012-08-02 05:06:00.963958
12367	3	1-3498	12367	91	27	1171	30	2012-08-02 05:06:00.963958
12371	6	1-2561	12370	27	6	2014	19	2012-08-02 05:06:00.963958
12372	6	1-3403	12370	6	72	2673	18	2012-08-02 05:06:00.963958
12373	6	1-2241	12370	72	31	2027	16	2012-08-02 05:06:00.963958
12374	6	1-3363	12370	31	32	772	67	2012-08-02 05:06:00.963958
12375	6	1-3485	12370	32	91	1314	6	2012-08-02 05:06:00.963958
12370	6	1-3498	12370	91	27	3760	30	2012-08-02 05:06:00.963958
12377	5	1-2561	12376	27	6	1372	19	2012-08-02 05:06:00.963958
12378	5	1-3403	12376	6	95	1888	18	2012-08-02 05:06:00.963958
12379	5	1-3049	12376	95	85	976	35	2012-08-02 05:06:00.963958
12380	5	1-2505	12376	85	91	832	6	2012-08-02 05:06:00.963958
12376	5	1-3498	12376	91	27	2470	30	2012-08-02 05:06:00.963958
12382	5	1-3155	12381	8	56	211	44	2012-08-02 05:06:00.963958
12383	5	1-2803	12381	56	13	157	74	2012-08-02 05:06:00.963958
12384	5	1-3440	12381	13	85	169	35	2012-08-02 05:06:00.963958
12385	5	1-2505	12381	85	91	181	6	2012-08-02 05:06:00.963958
12381	5	1-3498	12381	91	8	674	30	2012-08-02 05:06:00.963958
12386	1	1-905	\N	78	78	2044	74	2012-08-02 05:06:00.963958
12387	1	1-2803	\N	56	56	2734	74	2012-08-02 05:06:00.963958
12389	4	1-1022	12388	48	46	2334	22	2012-08-02 05:06:01.372957
12390	4	1-3151	12388	46	57	1395	79	2012-08-02 05:06:01.372957
12391	4	1-3267	12388	57	18	844	8	2012-08-02 05:06:01.372957
12388	4	1-3499	12388	18	48	3611	49	2012-08-02 05:06:01.372957
12393	3	1-2610	12392	47	57	547	79	2012-08-02 05:06:01.372957
12394	3	1-3267	12392	57	18	380	8	2012-08-02 05:06:01.372957
12392	3	1-3499	12392	18	47	1869	49	2012-08-02 05:06:01.372957
12396	4	1-1906	12395	35	1	699	55	2012-08-02 05:06:01.372957
12397	4	1-1898	12395	1	4	144	65	2012-08-02 05:06:01.372957
12398	4	1-3371	12395	4	18	179	8	2012-08-02 05:06:01.372957
12395	4	1-3499	12395	18	35	886	49	2012-08-02 05:06:01.372957
12400	6	1-3085	12399	58	13	109	74	2012-08-02 05:06:02.190692
12401	6	1-3440	12399	13	93	104	35	2012-08-02 05:06:02.190692
12402	6	1-1904	12399	93	70	48	91	2012-08-02 05:06:02.190692
12403	6	1-1623	12399	70	76	31	87	2012-08-02 05:06:02.190692
12404	6	1-3469	12399	76	86	59	75	2012-08-02 05:06:02.190692
12399	6	1-3501	12399	86	58	37	38	2012-08-02 05:06:02.190692
12406	5	1-2817	12405	9	35	484	96	2012-08-02 05:06:04.16803
12407	5	1-2974	12405	35	36	249	51	2012-08-02 05:06:04.16803
12408	5	1-3438	12405	36	61	553	63	2012-08-02 05:06:04.16803
12409	5	1-3338	12405	61	27	560	20	2012-08-02 05:06:04.16803
12405	5	1-3506	12405	27	9	615	35	2012-08-02 05:06:04.16803
12411	6	1-1904	12410	93	25	1093	91	2012-08-02 05:06:04.16803
12412	6	1-3108	12410	25	66	541	61	2012-08-02 05:06:04.16803
12413	6	1-3031	12410	66	36	740	51	2012-08-02 05:06:04.16803
12414	6	1-3438	12410	36	61	1791	63	2012-08-02 05:06:04.16803
12415	6	1-3338	12410	61	27	1976	20	2012-08-02 05:06:04.16803
12410	6	1-3506	12410	27	93	2361	35	2012-08-02 05:06:04.16803
12417	4	1-1785	12416	96	95	200	74	2012-08-02 05:06:04.16803
12418	4	1-3417	12416	95	61	188	63	2012-08-02 05:06:04.16803
12419	4	1-3338	12416	61	27	217	20	2012-08-02 05:06:04.16803
12416	4	1-3506	12416	27	96	270	35	2012-08-02 05:06:04.16803
12421	2	1-3349	12420	76	27	4322	20	2012-08-02 05:06:04.16803
12420	2	1-3506	12420	27	76	5501	35	2012-08-02 05:06:04.16803
12422	4	1-3508	12422	81	95	694	74	2012-08-02 05:06:04.963941
12423	4	1-3417	12422	95	36	637	63	2012-08-02 05:06:04.963941
12424	4	1-2851	12422	36	10	675	73	2012-08-02 05:06:04.963941
12425	4	1-3055	12422	10	81	320	21	2012-08-02 05:06:04.963941
16097	3	1-4636	16097	96	5	1629	75	2012-08-02 05:13:36.764155
12427	6	1-3417	12426	95	5	1137	63	2012-08-02 05:06:04.963941
12428	6	1-615	12426	5	20	298	41	2012-08-02 05:06:04.963941
12429	6	1-3381	12426	20	27	1675	90	2012-08-02 05:06:04.963941
12430	6	1-1529	12426	27	30	237	89	2012-08-02 05:06:04.963941
12431	6	1-3380	12426	30	81	490	21	2012-08-02 05:06:04.963941
12426	6	1-3508	12426	81	95	1147	74	2012-08-02 05:06:04.963941
12433	5	1-3403	12432	6	98	944	18	2012-08-02 05:06:05.395208
12434	5	1-3507	12432	98	78	596	92	2012-08-02 05:06:05.395208
12435	5	1-3441	12432	78	42	928	1	2012-08-02 05:06:05.395208
12436	5	1-1985	12432	42	17	398	72	2012-08-02 05:06:05.395208
12432	5	1-3509	12432	17	6	564	19	2012-08-02 05:06:05.395208
16101	4	1-3207	16100	17	79	294	41	2012-08-02 05:13:36.764155
12438	5	1-3478	12437	31	65	904	9	2012-08-02 05:06:05.817803
12439	5	1-3035	12437	65	55	785	51	2012-08-02 05:06:05.817803
12440	5	1-3480	12437	55	53	743	62	2012-08-02 05:06:05.817803
12441	5	1-2981	12437	53	84	1761	4	2012-08-02 05:06:05.817803
12437	5	1-3510	12437	84	31	1845	67	2012-08-02 05:06:05.817803
16102	4	1-4520	16100	79	80	1015	53	2012-08-02 05:13:36.764155
12443	6	1-3230	12442	86	76	100	99	2012-08-02 05:06:06.613203
12444	6	1-3337	12442	76	25	77	47	2012-08-02 05:06:06.613203
12445	6	1-3187	12442	25	25	63	26	2012-08-02 05:06:06.613203
12446	6	1-2842	12442	25	76	35	87	2012-08-02 05:06:06.613203
12447	6	1-3469	12442	76	83	66	75	2012-08-02 05:06:06.613203
12442	6	1-3512	12442	83	86	57	56	2012-08-02 05:06:06.613203
16103	4	1-348	16100	80	96	98	68	2012-08-02 05:13:36.764155
12449	4	1-3417	12448	95	5	849	63	2012-08-02 05:06:08.988473
12450	4	1-615	12448	5	20	202	41	2012-08-02 05:06:08.988473
12451	4	1-3381	12448	20	1	1030	90	2012-08-02 05:06:08.988473
12448	4	1-3518	12448	1	95	944	74	2012-08-02 05:06:08.988473
12453	5	1-3301	12452	12	57	949	11	2012-08-02 05:06:09.784269
12454	5	1-2764	12452	57	39	693	68	2012-08-02 05:06:09.784269
12455	5	1-3279	12452	39	96	1162	78	2012-08-02 05:06:09.784269
12456	5	1-2225	12452	96	10	217	54	2012-08-02 05:06:09.784269
12452	5	1-3520	12452	10	12	1561	41	2012-08-02 05:06:09.784269
12458	5	1-2646	12457	14	63	4082	80	2012-08-02 05:06:09.784269
12459	5	1-3170	12457	63	65	3308	59	2012-08-02 05:06:09.784269
12460	5	1-3474	12457	65	39	931	75	2012-08-02 05:06:09.784269
12461	5	1-1952	12457	39	10	391	54	2012-08-02 05:06:09.784269
12457	5	1-3520	12457	10	14	3052	41	2012-08-02 05:06:09.784269
12463	3	1-2764	12462	57	39	891	68	2012-08-02 05:06:10.19277
12464	3	1-3279	12462	39	23	1715	78	2012-08-02 05:06:10.19277
12462	3	1-3521	12462	23	57	1062	11	2012-08-02 05:06:10.19277
12466	4	1-3467	12465	21	25	40	38	2012-08-02 05:06:11.827787
12467	4	1-3355	12465	25	69	76	58	2012-08-02 05:06:11.827787
12468	4	1-2668	12465	69	86	79	57	2012-08-02 05:06:11.827787
12465	4	1-3525	12465	86	21	40	66	2012-08-02 05:06:11.827787
12470	5	1-2861	12469	34	89	116	18	2012-08-02 05:06:12.612366
12471	5	1-3515	12469	89	32	2320	15	2012-08-02 05:06:12.612366
12472	5	1-2909	12469	32	6	1409	70	2012-08-02 05:06:12.612366
12473	5	1-1842	12469	6	5	936	37	2012-08-02 05:06:12.612366
12469	5	1-3527	12469	5	34	445	19	2012-08-02 05:06:12.612366
12475	5	1-2861	12474	34	89	152	18	2012-08-02 05:06:12.612366
12476	5	1-3515	12474	89	28	3591	15	2012-08-02 05:06:12.612366
12477	5	1-3110	12474	28	78	1281	84	2012-08-02 05:06:12.612366
12478	5	1-1816	12474	78	5	889	37	2012-08-02 05:06:12.612366
12474	5	1-3527	12474	5	34	498	19	2012-08-02 05:06:12.612366
12480	5	1-1926	12479	3	55	670	58	2012-08-02 05:06:12.999133
12481	5	1-2101	12479	55	4	140	65	2012-08-02 05:06:12.999133
12482	5	1-3371	12479	4	65	136	8	2012-08-02 05:06:12.999133
12483	5	1-3233	12479	65	77	84	37	2012-08-02 05:06:12.999133
12479	5	1-3528	12479	77	3	969	97	2012-08-02 05:06:12.999133
12485	4	1-2155	12484	32	96	2332	67	2012-08-02 05:06:12.999133
12486	4	1-2080	12484	96	6	632	92	2012-08-02 05:06:12.999133
12487	4	1-2734	12484	6	77	508	37	2012-08-02 05:06:12.999133
12484	4	1-3528	12484	77	32	6855	97	2012-08-02 05:06:12.999133
12489	4	1-3158	12488	68	2	1949	83	2012-08-02 05:06:13.761464
12490	4	1-3482	12488	2	96	3351	13	2012-08-02 05:06:13.761464
12491	4	1-3179	12488	96	8	2891	94	2012-08-02 05:06:13.761464
12488	4	1-3530	12488	8	68	2579	24	2012-08-02 05:06:13.761464
12493	3	1-2453	12492	64	69	1419	61	2012-08-02 05:06:13.761464
12494	3	1-3503	12492	69	8	2991	94	2012-08-02 05:06:13.761464
12492	3	1-3530	12492	8	64	2791	24	2012-08-02 05:06:13.761464
12496	5	1-3448	12495	49	5	151	23	2012-08-02 05:06:15.341311
12497	5	1-3394	12495	5	15	160	47	2012-08-02 05:06:15.341311
12498	5	1-2572	12495	15	69	123	33	2012-08-02 05:06:15.341311
12499	5	1-3415	12495	69	99	392	100	2012-08-02 05:06:15.341311
12495	5	1-3534	12495	99	49	353	45	2012-08-02 05:06:15.341311
12500	1	1-2572	\N	15	15	2355	33	2012-08-02 05:06:15.341311
12502	4	1-3401	12501	71	33	3559	98	2012-08-02 05:06:17.396442
12503	4	1-3432	12501	33	13	3023	23	2012-08-02 05:06:17.396442
12504	4	1-3461	12501	13	18	3298	52	2012-08-02 05:06:17.396442
12501	4	1-3539	12501	18	71	4935	96	2012-08-02 05:06:17.396442
12506	3	1-3461	12505	13	52	1634	52	2012-08-02 05:06:17.793913
12507	3	1-2153	12505	52	65	517	43	2012-08-02 05:06:17.793913
12505	3	1-3540	12505	65	13	1899	23	2012-08-02 05:06:17.793913
12509	4	1-3394	12508	5	56	2085	47	2012-08-02 05:06:17.793913
12510	4	1-2634	12508	56	43	1200	55	2012-08-02 05:06:17.793913
12511	4	1-55	12508	43	65	439	43	2012-08-02 05:06:17.793913
12508	4	1-3540	12508	65	5	1996	23	2012-08-02 05:06:17.793913
12513	5	1-3394	12512	5	56	2101	47	2012-08-02 05:06:17.793913
12514	5	1-3011	12512	56	28	1078	15	2012-08-02 05:06:17.793913
12515	5	1-3110	12512	28	70	299	84	2012-08-02 05:06:17.793913
12516	5	1-3479	12512	70	65	423	43	2012-08-02 05:06:17.793913
12512	5	1-3540	12512	65	5	1967	23	2012-08-02 05:06:17.793913
12518	4	1-1979	12517	20	6	671	83	2012-08-02 05:06:17.793913
12519	4	1-3370	12517	6	75	995	97	2012-08-02 05:06:17.793913
12520	4	1-3396	12517	75	65	366	43	2012-08-02 05:06:17.793913
12517	4	1-3540	12517	65	20	1754	23	2012-08-02 05:06:17.793913
12522	5	1-3295	12521	77	71	194	68	2012-08-02 05:06:17.793913
12523	5	1-2729	12521	71	51	62	61	2012-08-02 05:06:17.793913
12524	5	1-3495	12521	51	94	29	85	2012-08-02 05:06:17.793913
12525	5	1-3420	12521	94	65	48	43	2012-08-02 05:06:17.793913
12521	5	1-3540	12521	65	77	231	23	2012-08-02 05:06:17.793913
16100	4	1-4636	16100	96	17	266	75	2012-08-02 05:13:36.764155
12527	5	1-803	12526	93	23	26	89	2012-08-02 05:06:17.793913
12528	5	1-3140	12526	23	74	36	17	2012-08-02 05:06:17.793913
12529	5	1-3223	12526	74	70	13	84	2012-08-02 05:06:17.793913
12530	5	1-3479	12526	70	65	22	43	2012-08-02 05:06:17.793913
12526	5	1-3540	12526	65	93	121	23	2012-08-02 05:06:17.793913
12531	1	1-3370	\N	6	6	1474	97	2012-08-02 05:06:17.793913
12533	5	1-3072	12532	90	15	967	42	2012-08-02 05:06:18.214132
12534	5	1-3153	12532	15	17	224	72	2012-08-02 05:06:18.214132
12535	5	1-3509	12532	17	75	269	19	2012-08-02 05:06:18.214132
12536	5	1-3513	12532	75	79	263	37	2012-08-02 05:06:18.214132
12532	5	1-3541	12532	79	90	461	24	2012-08-02 05:06:18.214132
12538	3	1-918	12537	52	25	1401	55	2012-08-02 05:06:18.214132
12539	3	1-3463	12537	25	79	3110	37	2012-08-02 05:06:18.214132
12537	3	1-3541	12537	79	52	5454	24	2012-08-02 05:06:18.214132
12541	3	1-2782	12540	94	6	933	92	2012-08-02 05:06:18.214132
12542	3	1-2734	12540	6	79	868	37	2012-08-02 05:06:18.214132
12540	3	1-3541	12540	79	94	1724	24	2012-08-02 05:06:18.214132
12544	5	1-3092	12543	31	96	66	39	2012-08-02 05:06:19.009614
12545	5	1-3505	12543	96	96	289	100	2012-08-02 05:06:19.009614
12546	5	1-3402	12543	96	18	119	22	2012-08-02 05:06:19.009614
12547	5	1-3277	12543	18	74	126	81	2012-08-02 05:06:19.009614
12543	5	1-3543	12543	74	31	92	26	2012-08-02 05:06:19.009614
12549	5	1-1716	12548	96	83	2054	54	2012-08-02 05:06:23.417491
12550	5	1-3426	12548	83	23	1035	34	2012-08-02 05:06:23.417491
12551	5	1-2224	12548	23	59	1074	44	2012-08-02 05:06:23.417491
12552	5	1-3069	12548	59	77	783	82	2012-08-02 05:06:23.417491
12548	5	1-3554	12548	77	96	1189	67	2012-08-02 05:06:23.417491
12554	5	1-2224	12553	23	42	560	44	2012-08-02 05:06:24.213104
12555	5	1-3280	12553	42	80	1652	77	2012-08-02 05:06:24.213104
12556	5	1-2744	12553	80	63	381	8	2012-08-02 05:06:24.213104
12557	5	1-690	12553	63	31	526	92	2012-08-02 05:06:24.213104
12553	5	1-3556	12553	31	23	533	34	2012-08-02 05:06:24.213104
12558	1	1-690	\N	63	63	1854	92	2012-08-02 05:06:24.213104
12560	5	1-2917	12559	76	94	1272	87	2012-08-02 05:06:26.234883
12561	5	1-1873	12559	94	17	386	72	2012-08-02 05:06:26.234883
12562	5	1-3509	12559	17	25	357	19	2012-08-02 05:06:26.234883
12563	5	1-3382	12559	25	53	258	98	2012-08-02 05:06:26.234883
12559	5	1-3561	12559	53	76	3559	95	2012-08-02 05:06:26.234883
12565	6	1-3082	12564	9	78	481	92	2012-08-02 05:06:26.234883
12566	6	1-3441	12564	78	42	515	1	2012-08-02 05:06:26.234883
12567	6	1-1985	12564	42	17	152	72	2012-08-02 05:06:26.234883
12568	6	1-3509	12564	17	25	148	19	2012-08-02 05:06:26.234883
12569	6	1-3382	12564	25	53	113	98	2012-08-02 05:06:26.234883
12564	6	1-3561	12564	53	9	1646	95	2012-08-02 05:06:26.234883
12571	4	1-2820	12570	2	4	1509	65	2012-08-02 05:06:27.836743
12572	4	1-3371	12570	4	39	2122	8	2012-08-02 05:06:27.836743
12573	4	1-791	12570	39	13	832	52	2012-08-02 05:06:27.836743
12570	4	1-3565	12570	13	2	4696	94	2012-08-02 05:06:27.836743
12575	5	1-2180	12574	3	16	1457	57	2012-08-02 05:06:28.621286
12576	5	1-1445	12574	16	65	410	59	2012-08-02 05:06:28.621286
12577	5	1-3474	12574	65	47	78	75	2012-08-02 05:06:28.621286
12578	5	1-3546	12574	47	43	2423	88	2012-08-02 05:06:28.621286
12574	5	1-3567	12574	43	3	2072	28	2012-08-02 05:06:28.621286
12580	5	1-3344	12579	60	88	841	18	2012-08-02 05:06:28.621286
12581	5	1-3537	12579	88	66	198	65	2012-08-02 05:06:28.621286
12582	5	1-3388	12579	66	47	126	75	2012-08-02 05:06:28.621286
12583	5	1-3546	12579	47	43	4973	88	2012-08-02 05:06:28.621286
12579	5	1-3567	12579	43	60	5366	28	2012-08-02 05:06:28.621286
12585	4	1-2871	12584	26	96	491	13	2012-08-02 05:06:28.621286
12586	4	1-3179	12584	96	40	402	94	2012-08-02 05:06:28.621286
12587	4	1-3234	12584	40	43	473	88	2012-08-02 05:06:28.621286
12584	4	1-3567	12584	43	26	578	28	2012-08-02 05:06:28.621286
12589	4	1-1838	12588	26	22	1074	10	2012-08-02 05:06:29.030211
12590	4	1-3290	12588	22	6	1057	20	2012-08-02 05:06:29.030211
12591	4	1-2619	12588	6	62	682	73	2012-08-02 05:06:29.030211
12588	4	1-3568	12588	62	26	665	9	2012-08-02 05:06:29.030211
12593	5	1-376	12592	44	17	86	72	2012-08-02 05:06:29.450001
12594	5	1-3509	12592	17	25	102	19	2012-08-02 05:06:29.450001
12595	5	1-3382	12592	25	33	95	98	2012-08-02 05:06:29.450001
12596	5	1-3432	12592	33	10	92	23	2012-08-02 05:06:29.450001
12592	5	1-3569	12592	10	44	318	93	2012-08-02 05:06:29.450001
12597	1	1-2231	\N	47	47	980	98	2012-08-02 05:06:29.450001
12599	5	1-2700	12598	11	6	817	77	2012-08-02 05:06:30.757343
12600	5	1-2913	12598	6	96	391	39	2012-08-02 05:06:30.757343
12601	5	1-3505	12598	96	26	1377	100	2012-08-02 05:06:30.757343
12602	5	1-2948	12598	26	43	287	26	2012-08-02 05:06:30.757343
12598	5	1-3572	12598	43	11	833	62	2012-08-02 05:06:30.757343
12604	3	1-2433	12603	18	74	1387	81	2012-08-02 05:06:30.757343
12605	3	1-3543	12603	74	43	816	26	2012-08-02 05:06:30.757343
12603	3	1-3572	12603	43	18	2375	62	2012-08-02 05:06:30.757343
12607	5	1-2681	12606	19	93	3986	53	2012-08-02 05:06:30.757343
12608	5	1-2527	12606	93	96	565	39	2012-08-02 05:06:30.757343
12609	5	1-3505	12606	96	26	2124	100	2012-08-02 05:06:30.757343
12610	5	1-2948	12606	26	43	472	26	2012-08-02 05:06:30.757343
12606	5	1-3572	12606	43	19	1462	62	2012-08-02 05:06:30.757343
12612	2	1-1486	12611	12	33	3112	51	2012-08-02 05:06:31.573119
12611	2	1-3574	12611	33	12	5815	58	2012-08-02 05:06:31.573119
12614	4	1-3187	12613	25	3	4827	26	2012-08-02 05:06:32.81084
12615	4	1-3294	12613	3	25	1288	38	2012-08-02 05:06:32.81084
12616	4	1-3355	12613	25	19	2624	58	2012-08-02 05:06:32.81084
12613	4	1-3577	12613	19	25	5898	47	2012-08-02 05:06:32.81084
12618	5	1-2634	12617	56	31	202	55	2012-08-02 05:06:32.81084
12619	5	1-3549	12617	31	89	417	51	2012-08-02 05:06:32.81084
12620	5	1-2984	12617	89	67	285	12	2012-08-02 05:06:32.81084
12621	5	1-1794	12617	67	19	83	58	2012-08-02 05:06:32.81084
12617	5	1-3577	12617	19	56	248	47	2012-08-02 05:06:32.81084
12622	1	1-1117	\N	66	66	699	58	2012-08-02 05:06:32.81084
12624	4	1-3274	12623	23	19	896	58	2012-08-02 05:06:33.197275
12625	4	1-3577	12623	19	63	2464	47	2012-08-02 05:06:33.197275
12626	4	1-3332	12623	63	88	936	90	2012-08-02 05:06:33.197275
12623	4	1-3578	12623	88	23	566	18	2012-08-02 05:06:33.197275
12628	5	1-3452	12627	21	16	1530	57	2012-08-02 05:06:33.197275
12629	5	1-1445	12627	16	65	674	59	2012-08-02 05:06:33.197275
12630	5	1-3474	12627	65	47	200	75	2012-08-02 05:06:33.197275
12631	5	1-3552	12627	47	88	369	90	2012-08-02 05:06:33.197275
12627	5	1-3578	12627	88	21	253	18	2012-08-02 05:06:33.197275
16105	3	1-1854	16104	17	24	2403	8	2012-08-02 05:13:36.764155
12633	4	1-3573	12632	6	35	451	49	2012-08-02 05:06:33.595514
12634	4	1-1906	12632	35	25	322	55	2012-08-02 05:06:33.595514
12635	4	1-3463	12632	25	63	696	37	2012-08-02 05:06:33.595514
12632	4	1-3579	12632	63	6	1651	87	2012-08-02 05:06:33.595514
16106	3	1-2862	16104	24	96	1806	68	2012-08-02 05:13:36.764155
12637	5	1-3477	12636	4	13	54	52	2012-08-02 05:06:33.595514
12638	5	1-3565	12636	13	67	261	94	2012-08-02 05:06:33.595514
12639	5	1-2956	12636	67	92	52	28	2012-08-02 05:06:33.595514
12640	5	1-1130	12636	92	63	36	37	2012-08-02 05:06:33.595514
12636	5	1-3579	12636	63	4	91	87	2012-08-02 05:06:33.595514
12642	5	1-2924	12641	88	23	1797	78	2012-08-02 05:06:33.595514
12643	5	1-3521	12641	23	80	996	11	2012-08-02 05:06:33.595514
12644	5	1-576	12641	80	25	299	55	2012-08-02 05:06:33.595514
12645	5	1-3463	12641	25	63	742	37	2012-08-02 05:06:33.595514
12641	5	1-3579	12641	63	88	2020	87	2012-08-02 05:06:33.595514
12646	1	1-737	\N	42	42	980	78	2012-08-02 05:06:33.595514
12648	3	1-2873	12647	85	90	2377	5	2012-08-02 05:06:33.992955
12649	3	1-3488	12647	90	52	567	86	2012-08-02 05:06:33.992955
12647	3	1-3580	12647	52	85	8326	38	2012-08-02 05:06:33.992955
12651	4	1-3322	12650	83	4	2000	66	2012-08-02 05:06:34.390902
12652	4	1-3536	12650	4	94	2419	25	2012-08-02 05:06:34.390902
12653	4	1-2259	12650	94	10	817	22	2012-08-02 05:06:34.390902
12650	4	1-3581	12650	10	83	3059	36	2012-08-02 05:06:34.390902
12655	3	1-2847	12654	65	72	2112	65	2012-08-02 05:06:34.390902
12656	3	1-1062	12654	72	10	955	22	2012-08-02 05:06:34.390902
12654	3	1-3581	12654	10	65	3815	36	2012-08-02 05:06:34.390902
12658	5	1-3322	12657	83	4	386	66	2012-08-02 05:06:34.390902
12659	5	1-3536	12657	4	5	526	25	2012-08-02 05:06:34.390902
12660	5	1-3081	12657	5	72	261	65	2012-08-02 05:06:34.390902
12661	5	1-1062	12657	72	10	125	22	2012-08-02 05:06:34.390902
12657	5	1-3581	12657	10	83	525	36	2012-08-02 05:06:34.390902
12663	6	1-3322	12662	83	23	1065	66	2012-08-02 05:06:34.390902
12664	6	1-3555	12662	23	84	1563	60	2012-08-02 05:06:34.390902
12665	6	1-2004	12662	84	59	881	44	2012-08-02 05:06:34.390902
12666	6	1-3069	12662	59	68	491	82	2012-08-02 05:06:34.390902
12667	6	1-1884	12662	68	10	306	22	2012-08-02 05:06:34.390902
12662	6	1-3581	12662	10	83	1365	36	2012-08-02 05:06:34.390902
12668	1	1-1062	\N	72	72	1917	22	2012-08-02 05:06:34.390902
12670	5	1-2634	12669	56	25	800	55	2012-08-02 05:06:34.799816
12671	5	1-3463	12669	25	68	2185	37	2012-08-02 05:06:34.799816
12672	5	1-3497	12669	68	35	2173	11	2012-08-02 05:06:34.799816
12673	5	1-3224	12669	35	27	1413	81	2012-08-02 05:06:34.799816
12669	5	1-3582	12669	27	56	953	47	2012-08-02 05:06:34.799816
12675	4	1-3480	12674	55	2	700	62	2012-08-02 05:06:35.208731
12676	4	1-2172	12674	2	84	804	42	2012-08-02 05:06:35.208731
12677	4	1-3257	12674	84	21	930	88	2012-08-02 05:06:35.208731
12674	4	1-3583	12674	21	55	726	51	2012-08-02 05:06:35.208731
12679	4	1-2984	12678	89	47	317	12	2012-08-02 05:06:35.208731
12680	4	1-2437	12678	47	74	425	31	2012-08-02 05:06:35.208731
12681	4	1-3560	12678	74	21	580	88	2012-08-02 05:06:35.208731
12678	4	1-3583	12678	21	89	474	51	2012-08-02 05:06:35.208731
12683	4	1-3480	12682	55	58	5117	62	2012-08-02 05:06:35.208731
12684	4	1-3239	12682	58	82	5017	17	2012-08-02 05:06:35.208731
12685	4	1-2900	12682	82	21	5564	88	2012-08-02 05:06:35.208731
12682	4	1-3583	12682	21	55	4801	51	2012-08-02 05:06:35.208731
12687	6	1-3562	12686	85	62	3925	1	2012-08-02 05:06:35.617497
12688	6	1-3167	12686	62	5	297	35	2012-08-02 05:06:35.617497
12689	6	1-3042	12686	5	39	425	55	2012-08-02 05:06:35.617497
12690	6	1-3550	12686	39	13	440	52	2012-08-02 05:06:35.617497
12691	6	1-3565	12686	13	47	2238	94	2012-08-02 05:06:35.617497
12686	6	1-3584	12686	47	85	3877	53	2012-08-02 05:06:35.617497
12693	5	1-3562	12692	85	42	4654	1	2012-08-02 05:06:35.617497
12694	5	1-1985	12692	42	65	2099	72	2012-08-02 05:06:35.617497
12695	5	1-2800	12692	65	69	1001	61	2012-08-02 05:06:35.617497
12696	5	1-3503	12692	69	47	2085	94	2012-08-02 05:06:35.617497
12692	5	1-3584	12692	47	85	4075	53	2012-08-02 05:06:35.617497
12698	6	1-3534	12697	99	59	4194	45	2012-08-02 05:06:36.413033
12699	6	1-3472	12697	59	74	3847	61	2012-08-02 05:06:36.413033
12700	6	1-3013	12697	74	84	1213	60	2012-08-02 05:06:36.413033
12701	6	1-2004	12697	84	42	727	44	2012-08-02 05:06:36.413033
12702	6	1-3280	12697	42	71	1720	77	2012-08-02 05:06:36.413033
12697	6	1-3586	12697	71	99	4012	100	2012-08-02 05:06:36.413033
12704	6	1-3534	12703	99	59	805	45	2012-08-02 05:06:36.413033
12705	6	1-3472	12703	59	1	757	61	2012-08-02 05:06:36.413033
12706	6	1-2136	12703	1	35	159	56	2012-08-02 05:06:36.413033
12707	6	1-1774	12703	35	42	129	44	2012-08-02 05:06:36.413033
12708	6	1-3280	12703	42	71	314	77	2012-08-02 05:06:36.413033
12703	6	1-3586	12703	71	99	751	100	2012-08-02 05:06:36.413033
12710	5	1-3472	12709	59	69	163	61	2012-08-02 05:06:37.230609
12711	5	1-3503	12709	69	47	315	94	2012-08-02 05:06:37.230609
12712	5	1-3584	12709	47	93	569	53	2012-08-02 05:06:37.230609
12713	5	1-2527	12709	93	97	86	39	2012-08-02 05:06:37.230609
12709	5	1-3588	12709	97	59	154	45	2012-08-02 05:06:37.230609
12715	5	1-3280	12714	42	80	773	77	2012-08-02 05:06:37.62853
12716	5	1-2744	12714	80	72	158	8	2012-08-02 05:06:37.62853
12717	5	1-429	12714	72	17	249	72	2012-08-02 05:06:37.62853
12718	5	1-3509	12714	17	74	329	19	2012-08-02 05:06:37.62853
12714	5	1-3589	12714	74	42	296	44	2012-08-02 05:06:37.62853
12720	5	1-3578	12719	88	51	676	18	2012-08-02 05:06:38.424002
12721	5	1-3266	12719	51	86	808	34	2012-08-02 05:06:38.424002
12722	5	1-3190	12719	86	31	1355	16	2012-08-02 05:06:38.424002
12723	5	1-3363	12719	31	23	627	67	2012-08-02 05:06:38.424002
12719	5	1-3591	12719	23	88	1083	90	2012-08-02 05:06:38.424002
12725	6	1-2032	12724	63	32	285	15	2012-08-02 05:06:38.833046
12726	6	1-2909	12724	32	87	194	70	2012-08-02 05:06:38.833046
12727	6	1-2270	12724	87	77	80	82	2012-08-02 05:06:38.833046
12728	6	1-3554	12724	77	96	105	67	2012-08-02 05:06:38.833046
12729	6	1-1716	12724	96	65	157	54	2012-08-02 05:06:38.833046
12724	6	1-3592	12724	65	63	362	49	2012-08-02 05:06:38.833046
12731	6	1-2812	12730	22	90	791	5	2012-08-02 05:06:38.833046
12732	6	1-3488	12730	90	61	443	86	2012-08-02 05:06:38.833046
12733	6	1-1074	12730	61	84	219	4	2012-08-02 05:06:38.833046
12734	6	1-3510	12730	84	96	233	67	2012-08-02 05:06:38.833046
12735	6	1-1716	12730	96	65	364	54	2012-08-02 05:06:38.833046
12730	6	1-3592	12730	65	22	881	49	2012-08-02 05:06:38.833046
16104	3	1-4636	16104	96	17	4905	75	2012-08-02 05:13:36.764155
12737	5	1-2812	12736	22	90	1580	5	2012-08-02 05:06:38.833046
12738	5	1-3488	12736	90	61	960	86	2012-08-02 05:06:38.833046
12739	5	1-1074	12736	61	17	516	4	2012-08-02 05:06:38.833046
12740	5	1-2804	12736	17	65	616	54	2012-08-02 05:06:38.833046
12736	5	1-3592	12736	65	22	1620	49	2012-08-02 05:06:38.833046
12742	3	1-3550	12741	39	10	3585	52	2012-08-02 05:06:39.252859
12743	3	1-3376	12741	10	98	4033	53	2012-08-02 05:06:39.252859
12741	3	1-3593	12741	98	39	3595	55	2012-08-02 05:06:39.252859
12745	5	1-3546	12744	47	88	1707	88	2012-08-02 05:06:39.650504
12746	5	1-3348	12744	88	68	261	37	2012-08-02 05:06:39.650504
12747	5	1-3497	12744	68	74	156	11	2012-08-02 05:06:39.650504
12748	5	1-1164	12744	74	64	58	1	2012-08-02 05:06:39.650504
12744	5	1-3594	12744	64	47	59	75	2012-08-02 05:06:39.650504
12750	5	1-3552	12749	47	88	5502	90	2012-08-02 05:06:39.650504
12751	5	1-3578	12749	88	98	3172	18	2012-08-02 05:06:39.650504
12752	5	1-3507	12749	98	78	1783	92	2012-08-02 05:06:39.650504
12753	5	1-3441	12749	78	64	2471	1	2012-08-02 05:06:39.650504
12749	5	1-3594	12749	64	47	3544	75	2012-08-02 05:06:39.650504
12755	6	1-2507	12754	56	69	1138	61	2012-08-02 05:06:39.650504
12756	6	1-3503	12754	69	8	2017	94	2012-08-02 05:06:39.650504
12757	6	1-3530	12754	8	94	1582	24	2012-08-02 05:06:39.650504
12758	6	1-2782	12754	94	78	797	92	2012-08-02 05:06:39.650504
12759	6	1-3441	12754	78	64	1110	1	2012-08-02 05:06:39.650504
12754	6	1-3594	12754	64	56	1601	75	2012-08-02 05:06:39.650504
12761	5	1-3552	12760	47	88	643	90	2012-08-02 05:06:39.650504
12762	5	1-3578	12760	88	95	391	18	2012-08-02 05:06:39.650504
12763	5	1-3049	12760	95	69	231	35	2012-08-02 05:06:39.650504
12764	5	1-2177	12760	69	64	259	1	2012-08-02 05:06:39.650504
12760	5	1-3594	12760	64	47	392	75	2012-08-02 05:06:39.650504
12766	5	1-3512	12765	83	76	676	56	2012-08-02 05:06:39.650504
12767	5	1-1848	12765	76	9	690	95	2012-08-02 05:06:39.650504
12768	5	1-3082	12765	9	78	277	92	2012-08-02 05:06:39.650504
12769	5	1-3441	12765	78	64	409	1	2012-08-02 05:06:39.650504
12765	5	1-3594	12765	64	83	625	75	2012-08-02 05:06:39.650504
12771	6	1-3512	12770	83	86	2401	56	2012-08-02 05:06:39.650504
12772	6	1-3230	12770	86	79	5659	99	2012-08-02 05:06:39.650504
12773	6	1-3262	12770	79	96	2338	67	2012-08-02 05:06:39.650504
12774	6	1-2080	12770	96	78	781	92	2012-08-02 05:06:39.650504
12775	6	1-3441	12770	78	64	1245	1	2012-08-02 05:06:39.650504
12770	6	1-3594	12770	64	83	2054	75	2012-08-02 05:06:39.650504
12777	4	1-777	12776	52	12	126	59	2012-08-02 05:06:39.650504
12778	4	1-3516	12776	12	74	161	11	2012-08-02 05:06:39.650504
12779	4	1-1164	12776	74	64	100	1	2012-08-02 05:06:39.650504
12776	4	1-3594	12776	64	52	169	75	2012-08-02 05:06:39.650504
12781	5	1-2540	12780	11	17	2166	72	2012-08-02 05:06:40.081959
12782	5	1-3509	12780	17	79	2332	19	2012-08-02 05:06:40.081959
12783	5	1-3208	12780	79	16	3998	90	2012-08-02 05:06:40.081959
12784	5	1-2856	12780	16	9	2171	50	2012-08-02 05:06:40.081959
12780	5	1-3595	12780	9	11	2915	64	2012-08-02 05:06:40.081959
12786	5	1-2540	12785	11	17	680	72	2012-08-02 05:06:40.081959
12787	5	1-3509	12785	17	89	861	19	2012-08-02 05:06:40.081959
12788	5	1-1894	12785	89	16	772	90	2012-08-02 05:06:40.081959
12789	5	1-2856	12785	16	9	493	50	2012-08-02 05:06:40.081959
12785	5	1-3595	12785	9	11	778	64	2012-08-02 05:06:40.081959
12791	5	1-2128	12790	75	100	1124	83	2012-08-02 05:06:40.081959
12792	5	1-3450	12790	100	74	1351	17	2012-08-02 05:06:40.081959
12793	5	1-3223	12790	74	18	556	84	2012-08-02 05:06:40.081959
12794	5	1-2356	12790	18	9	1039	50	2012-08-02 05:06:40.081959
12790	5	1-3595	12790	9	75	1762	64	2012-08-02 05:06:40.081959
12796	4	1-3548	12795	5	65	1962	34	2012-08-02 05:06:40.081959
12797	4	1-1997	12795	65	91	1340	79	2012-08-02 05:06:40.081959
12798	4	1-2462	12795	91	9	966	50	2012-08-02 05:06:40.081959
12795	4	1-3595	12795	9	5	1719	64	2012-08-02 05:06:40.081959
12800	5	1-3548	12799	5	60	630	34	2012-08-02 05:06:40.081959
12801	5	1-3156	12799	60	69	830	69	2012-08-02 05:06:40.081959
12802	5	1-2875	12799	69	81	385	9	2012-08-02 05:06:40.081959
12803	5	1-1599	12799	81	9	281	50	2012-08-02 05:06:40.081959
12799	5	1-3595	12799	9	5	526	64	2012-08-02 05:06:40.081959
12805	5	1-2646	12804	14	94	1855	80	2012-08-02 05:06:43.727664
12806	5	1-3526	12804	94	96	439	39	2012-08-02 05:06:43.727664
12807	5	1-3505	12804	96	68	1489	100	2012-08-02 05:06:43.727664
12808	5	1-3020	12804	68	26	1599	40	2012-08-02 05:06:43.727664
12804	5	1-3604	12804	26	14	1701	41	2012-08-02 05:06:43.727664
12810	5	1-3359	12809	93	99	747	100	2012-08-02 05:06:44.523334
12811	5	1-3534	12809	99	59	900	45	2012-08-02 05:06:44.523334
12812	5	1-3472	12809	59	100	952	61	2012-08-02 05:06:44.523334
12813	5	1-2973	12809	100	82	706	58	2012-08-02 05:06:44.523334
12809	5	1-3606	12809	82	93	1310	5	2012-08-02 05:06:44.523334
12815	4	1-2244	12814	64	62	4983	73	2012-08-02 05:06:45.705654
12816	4	1-3568	12814	62	2	4632	9	2012-08-02 05:06:45.705654
12817	4	1-3504	12814	2	47	5218	69	2012-08-02 05:06:45.705654
12814	4	1-3609	12814	47	64	4016	91	2012-08-02 05:06:45.705654
12819	4	1-2953	12818	63	74	283	42	2012-08-02 05:06:45.705654
12820	4	1-2904	12818	74	72	273	32	2012-08-02 05:06:45.705654
12821	4	1-3413	12818	72	47	310	69	2012-08-02 05:06:45.705654
12818	4	1-3609	12818	47	63	239	91	2012-08-02 05:06:45.705654
12823	2	1-2505	12822	85	94	2757	6	2012-08-02 05:06:46.103181
12822	2	1-3610	12822	94	85	2610	35	2012-08-02 05:06:46.103181
12825	5	1-2658	12824	6	97	818	76	2012-08-02 05:06:47.307689
12826	5	1-1736	12824	97	9	693	50	2012-08-02 05:06:47.307689
12827	5	1-3595	12824	9	11	1241	64	2012-08-02 05:06:47.307689
12828	5	1-2540	12824	11	42	1229	72	2012-08-02 05:06:47.307689
12824	5	1-3613	12824	42	6	867	9	2012-08-02 05:06:47.307689
12830	5	1-3520	12829	10	14	1115	41	2012-08-02 05:06:48.887145
12831	5	1-2646	12829	14	39	1429	80	2012-08-02 05:06:48.887145
12832	5	1-782	12829	39	88	1274	88	2012-08-02 05:06:48.887145
12833	5	1-3348	12829	88	62	298	37	2012-08-02 05:06:48.887145
12829	5	1-3617	12829	62	10	149	54	2012-08-02 05:06:48.887145
12835	3	1-3081	12834	5	66	1416	65	2012-08-02 05:06:49.284951
12836	3	1-3388	12834	66	26	1115	75	2012-08-02 05:06:49.284951
12834	3	1-3618	12834	26	5	2026	25	2012-08-02 05:06:49.284951
12838	2	1-2504	12837	11	51	7223	88	2012-08-02 05:06:49.682692
12837	2	1-3619	12837	51	11	9934	96	2012-08-02 05:06:49.682692
16108	5	1-4388	16107	3	97	1819	52	2012-08-02 05:13:37.184443
12840	4	1-3322	12839	83	23	2381	66	2012-08-02 05:06:50.080579
12841	4	1-3555	12839	23	83	2770	60	2012-08-02 05:06:50.080579
12842	4	1-933	12839	83	34	2063	2	2012-08-02 05:06:50.080579
12839	4	1-3620	12839	34	83	2956	36	2012-08-02 05:06:50.080579
12844	4	1-1879	12843	59	42	274	44	2012-08-02 05:06:50.080579
12845	4	1-3280	12843	42	58	663	77	2012-08-02 05:06:50.080579
12846	4	1-1425	12843	58	34	391	2	2012-08-02 05:06:50.080579
12843	4	1-3620	12843	34	59	771	36	2012-08-02 05:06:50.080579
12848	5	1-3083	12847	91	61	420	86	2012-08-02 05:06:50.080579
12849	5	1-1074	12847	61	31	206	4	2012-08-02 05:06:50.080579
12850	5	1-2831	12847	31	70	340	10	2012-08-02 05:06:50.080579
12851	5	1-3300	12847	70	34	299	2	2012-08-02 05:06:50.080579
12847	5	1-3620	12847	34	91	634	36	2012-08-02 05:06:50.080579
12853	3	1-3083	12852	91	11	1213	86	2012-08-02 05:06:50.080579
12854	3	1-2578	12852	11	34	809	2	2012-08-02 05:06:50.080579
12852	3	1-3620	12852	34	91	1773	36	2012-08-02 05:06:50.080579
12856	5	1-2255	12855	74	9	42	50	2012-08-02 05:06:50.86507
12857	5	1-3595	12855	9	77	61	64	2012-08-02 05:06:50.86507
12858	5	1-1633	12855	77	13	45	33	2012-08-02 05:06:50.86507
12859	5	1-2758	12855	13	85	28	67	2012-08-02 05:06:50.86507
12855	5	1-3622	12855	85	74	74	8	2012-08-02 05:06:50.86507
12861	6	1-3233	12860	65	5	3877	37	2012-08-02 05:06:50.86507
12862	6	1-3527	12860	5	74	2118	19	2012-08-02 05:06:50.86507
12863	6	1-3589	12860	74	59	1876	44	2012-08-02 05:06:50.86507
12864	6	1-3069	12860	59	77	1210	82	2012-08-02 05:06:50.86507
12865	6	1-3554	12860	77	85	1626	67	2012-08-02 05:06:50.86507
12860	6	1-3622	12860	85	65	4889	8	2012-08-02 05:06:50.86507
12867	5	1-3166	12866	63	62	203	73	2012-08-02 05:06:53.671591
12868	5	1-3568	12866	62	31	114	9	2012-08-02 05:06:53.671591
12869	5	1-3145	12866	31	55	57	33	2012-08-02 05:06:53.671591
12870	5	1-3034	12866	55	30	30	28	2012-08-02 05:06:53.671591
12866	5	1-3629	12866	30	63	886	32	2012-08-02 05:06:53.671591
12872	5	1-3002	12871	2	84	2693	50	2012-08-02 05:06:53.671591
12873	5	1-3502	12871	84	31	779	9	2012-08-02 05:06:53.671591
12874	5	1-3145	12871	31	55	425	33	2012-08-02 05:06:53.671591
12875	5	1-3034	12871	55	30	238	28	2012-08-02 05:06:53.671591
12871	5	1-3629	12871	30	2	7710	32	2012-08-02 05:06:53.671591
12877	4	1-1897	12876	7	49	171	53	2012-08-02 05:06:53.671591
12878	4	1-2969	12876	49	49	61	76	2012-08-02 05:06:53.671591
12879	4	1-3434	12876	49	30	42	28	2012-08-02 05:06:53.671591
12876	4	1-3629	12876	30	7	1389	32	2012-08-02 05:06:53.671591
12881	5	1-3505	12880	96	99	257	100	2012-08-02 05:06:55.284557
12882	5	1-3534	12880	99	77	204	45	2012-08-02 05:06:55.284557
12883	5	1-3373	12880	77	95	55	30	2012-08-02 05:06:55.284557
12884	5	1-1277	12880	95	32	45	59	2012-08-02 05:06:55.284557
12880	5	1-3633	12880	32	96	97	39	2012-08-02 05:06:55.284557
12886	5	1-3588	12885	97	49	7446	45	2012-08-02 05:06:55.284557
12887	5	1-3448	12885	49	10	2888	23	2012-08-02 05:06:55.284557
12888	5	1-3569	12885	10	48	7874	93	2012-08-02 05:06:55.284557
12889	5	1-2027	12885	48	32	2783	59	2012-08-02 05:06:55.284557
12885	5	1-3633	12885	32	97	6154	39	2012-08-02 05:06:55.284557
12891	4	1-3588	12890	97	77	1232	45	2012-08-02 05:06:55.284557
12892	4	1-3373	12890	77	95	384	30	2012-08-02 05:06:55.284557
12893	4	1-1277	12890	95	32	366	59	2012-08-02 05:06:55.284557
12890	4	1-3633	12890	32	97	907	39	2012-08-02 05:06:55.284557
12895	4	1-3505	12894	96	26	853	100	2012-08-02 05:06:55.284557
12896	4	1-2948	12894	26	36	166	26	2012-08-02 05:06:55.284557
12897	4	1-999	12894	36	32	97	59	2012-08-02 05:06:55.284557
12894	4	1-3633	12894	32	96	259	39	2012-08-02 05:06:55.284557
12899	5	1-3429	12898	37	59	104	44	2012-08-02 05:06:55.284557
12900	5	1-3069	12898	59	77	65	82	2012-08-02 05:06:55.284557
12901	5	1-3554	12898	77	33	84	67	2012-08-02 05:06:55.284557
12902	5	1-2383	12898	33	32	94	59	2012-08-02 05:06:55.284557
12898	5	1-3633	12898	32	37	281	39	2012-08-02 05:06:55.284557
12904	3	1-2442	12903	15	53	843	71	2012-08-02 05:06:56.102545
12905	3	1-3397	12903	53	40	503	96	2012-08-02 05:06:56.102545
12903	3	1-3635	12903	40	15	1790	27	2012-08-02 05:06:56.102545
12907	6	1-3531	12906	21	97	68	76	2012-08-02 05:06:56.102545
12908	6	1-1736	12906	97	84	52	50	2012-08-02 05:06:56.102545
12909	6	1-3502	12906	84	1	21	9	2012-08-02 05:06:56.102545
12910	6	1-2474	12906	1	53	17	71	2012-08-02 05:06:56.102545
12911	6	1-3397	12906	53	40	11	96	2012-08-02 05:06:56.102545
12906	6	1-3635	12906	40	21	41	27	2012-08-02 05:06:56.102545
12913	6	1-3450	12912	100	39	742	17	2012-08-02 05:06:57.308487
12914	6	1-2201	12912	39	15	710	42	2012-08-02 05:06:57.308487
12915	6	1-3153	12912	15	42	167	72	2012-08-02 05:06:57.308487
12916	6	1-3613	12912	42	1	99	9	2012-08-02 05:06:57.308487
12917	6	1-2474	12912	1	33	72	71	2012-08-02 05:06:57.308487
12912	6	1-3638	12912	33	100	692	83	2012-08-02 05:06:57.308487
12919	5	1-2113	12918	48	70	2506	12	2012-08-02 05:06:57.308487
12920	5	1-2109	12918	70	34	511	2	2012-08-02 05:06:57.308487
12921	5	1-3620	12918	34	8	1077	36	2012-08-02 05:06:57.308487
12922	5	1-2662	12918	8	33	416	71	2012-08-02 05:06:57.308487
12918	5	1-3638	12918	33	48	4287	83	2012-08-02 05:06:57.308487
12924	5	1-3450	12923	100	74	167	17	2012-08-02 05:06:57.308487
12925	5	1-3223	12923	74	86	72	84	2012-08-02 05:06:57.308487
12926	5	1-2575	12923	86	3	31	23	2012-08-02 05:06:57.308487
12927	5	1-2805	12923	3	33	12	71	2012-08-02 05:06:57.308487
12923	5	1-3638	12923	33	100	132	83	2012-08-02 05:06:57.308487
12929	5	1-3130	12928	74	87	318	50	2012-08-02 05:06:57.308487
12930	5	1-3389	12928	87	62	241	73	2012-08-02 05:06:57.308487
12931	5	1-3568	12928	62	1	241	9	2012-08-02 05:06:57.308487
12932	5	1-2474	12928	1	33	219	71	2012-08-02 05:06:57.308487
12928	5	1-3638	12928	33	74	2598	83	2012-08-02 05:06:57.308487
12934	5	1-2851	12933	36	62	271	73	2012-08-02 05:06:58.501621
12935	5	1-3568	12933	62	37	219	9	2012-08-02 05:06:58.501621
12936	5	1-2773	12933	37	12	279	68	2012-08-02 05:06:58.501621
12937	5	1-1927	12933	12	25	240	31	2012-08-02 05:06:58.501621
12933	5	1-3641	12933	25	36	278	63	2012-08-02 05:06:58.501621
12939	4	1-3608	12938	83	9	3914	95	2012-08-02 05:06:58.501621
12940	4	1-3082	12938	9	43	1442	92	2012-08-02 05:06:58.501621
12941	4	1-3511	12938	43	25	6865	31	2012-08-02 05:06:58.501621
12938	4	1-3641	12938	25	83	8032	63	2012-08-02 05:06:58.501621
16109	5	1-4440	16107	97	4	1347	92	2012-08-02 05:13:37.184443
12943	5	1-2851	12942	36	66	218	73	2012-08-02 05:06:58.501621
12944	5	1-1574	12942	66	94	43	24	2012-08-02 05:06:58.501621
12945	5	1-2782	12942	94	43	24	92	2012-08-02 05:06:58.501621
12946	5	1-3511	12942	43	25	137	31	2012-08-02 05:06:58.501621
12942	5	1-3641	12942	25	36	188	63	2012-08-02 05:06:58.501621
16110	5	1-4606	16107	4	96	2681	66	2012-08-02 05:13:37.184443
12948	3	1-2687	12947	55	70	956	10	2012-08-02 05:06:59.695173
12949	3	1-3300	12947	70	53	814	2	2012-08-02 05:06:59.695173
12947	3	1-3644	12947	53	55	1367	39	2012-08-02 05:06:59.695173
16111	5	1-4083	16107	96	87	1257	12	2012-08-02 05:13:37.184443
12951	4	1-2687	12950	55	22	2626	10	2012-08-02 05:06:59.695173
12952	4	1-3290	12950	22	60	2583	20	2012-08-02 05:06:59.695173
12953	4	1-3445	12950	60	53	1697	2	2012-08-02 05:06:59.695173
12950	4	1-3644	12950	53	55	3271	39	2012-08-02 05:06:59.695173
12955	5	1-3380	12954	30	81	2381	21	2012-08-02 05:07:00.899119
12956	5	1-3508	12954	81	95	4397	74	2012-08-02 05:07:00.899119
12957	5	1-3417	12954	95	33	3437	63	2012-08-02 05:07:00.899119
12958	5	1-3080	12954	33	13	2807	93	2012-08-02 05:07:00.899119
12954	5	1-3647	12954	13	30	1461	89	2012-08-02 05:07:00.899119
12960	5	1-2127	12959	38	34	341	2	2012-08-02 05:07:00.899119
12961	5	1-3620	12959	34	43	655	36	2012-08-02 05:07:00.899119
12962	5	1-2892	12959	43	33	597	63	2012-08-02 05:07:00.899119
12963	5	1-3080	12959	33	13	515	93	2012-08-02 05:07:00.899119
12959	5	1-3647	12959	13	38	283	89	2012-08-02 05:07:00.899119
12965	5	1-2127	12964	38	53	925	2	2012-08-02 05:07:00.899119
12966	5	1-3644	12964	53	90	1591	39	2012-08-02 05:07:00.899119
12967	5	1-3281	12964	90	55	1848	68	2012-08-02 05:07:00.899119
12968	5	1-3345	12964	55	13	1163	93	2012-08-02 05:07:00.899119
12964	5	1-3647	12964	13	38	701	89	2012-08-02 05:07:00.899119
12970	3	1-3380	12969	30	59	1480	21	2012-08-02 05:07:00.899119
12971	3	1-1129	12969	59	13	1178	93	2012-08-02 05:07:00.899119
12969	3	1-3647	12969	13	30	746	89	2012-08-02 05:07:00.899119
12973	5	1-3390	12972	80	17	1215	72	2012-08-02 05:07:00.899119
12974	5	1-3509	12972	17	74	1787	19	2012-08-02 05:07:00.899119
12975	5	1-3589	12972	74	28	1791	44	2012-08-02 05:07:00.899119
12976	5	1-3284	12972	28	13	3466	93	2012-08-02 05:07:00.899119
12972	5	1-3647	12972	13	80	2353	89	2012-08-02 05:07:00.899119
12978	5	1-2193	12977	50	12	262	68	2012-08-02 05:07:00.899119
12979	5	1-1927	12977	12	25	280	31	2012-08-02 05:07:00.899119
12980	5	1-3641	12977	25	33	401	63	2012-08-02 05:07:00.899119
12981	5	1-3080	12977	33	13	437	93	2012-08-02 05:07:00.899119
12977	5	1-3647	12977	13	50	304	89	2012-08-02 05:07:00.899119
12983	5	1-3240	12982	12	22	582	74	2012-08-02 05:07:02.125972
12984	5	1-2843	12982	22	25	663	31	2012-08-02 05:07:02.125972
12985	5	1-3641	12982	25	36	840	63	2012-08-02 05:07:02.125972
12986	5	1-2851	12982	36	54	896	73	2012-08-02 05:07:02.125972
12982	5	1-3650	12982	54	12	1317	13	2012-08-02 05:07:02.125972
12988	4	1-3501	12987	86	58	726	38	2012-08-02 05:07:03.747369
12989	4	1-3085	12987	58	51	1325	74	2012-08-02 05:07:03.747369
12990	4	1-2722	12987	51	83	606	79	2012-08-02 05:07:03.747369
12987	4	1-3654	12987	83	86	1910	75	2012-08-02 05:07:03.747369
12992	4	1-3552	12991	47	1	1863	90	2012-08-02 05:07:03.747369
12993	4	1-3518	12991	1	51	1110	74	2012-08-02 05:07:03.747369
12994	4	1-2722	12991	51	83	538	79	2012-08-02 05:07:03.747369
12991	4	1-3654	12991	83	47	1793	75	2012-08-02 05:07:03.747369
12996	5	1-3648	12995	13	59	5372	44	2012-08-02 05:07:03.747369
12997	5	1-3069	12995	59	68	2611	82	2012-08-02 05:07:03.747369
12998	5	1-1884	12995	68	46	1420	22	2012-08-02 05:07:03.747369
12999	5	1-3151	12995	46	83	745	79	2012-08-02 05:07:03.747369
12995	5	1-3654	12995	83	13	2883	75	2012-08-02 05:07:03.747369
13001	6	1-3512	13000	83	86	742	56	2012-08-02 05:07:03.747369
13002	6	1-3230	13000	86	49	918	99	2012-08-02 05:07:03.747369
13003	6	1-3272	13000	49	81	373	21	2012-08-02 05:07:03.747369
13004	6	1-3508	13000	81	51	464	74	2012-08-02 05:07:03.747369
13005	6	1-2722	13000	51	83	203	79	2012-08-02 05:07:03.747369
13000	6	1-3654	13000	83	83	859	75	2012-08-02 05:07:03.747369
13007	5	1-3648	13006	13	11	1422	44	2012-08-02 05:07:03.747369
13008	5	1-889	13006	11	5	302	64	2012-08-02 05:07:03.747369
13009	5	1-3548	13006	5	65	273	34	2012-08-02 05:07:03.747369
13010	5	1-1997	13006	65	83	148	79	2012-08-02 05:07:03.747369
13006	5	1-3654	13006	83	13	661	75	2012-08-02 05:07:03.747369
13012	4	1-3512	13011	83	17	676	56	2012-08-02 05:07:03.747369
13013	4	1-1328	13011	17	46	364	22	2012-08-02 05:07:03.747369
13014	4	1-3151	13011	46	83	165	79	2012-08-02 05:07:03.747369
13011	4	1-3654	13011	83	83	740	75	2012-08-02 05:07:03.747369
13016	6	1-2994	13015	27	11	386	3	2012-08-02 05:07:03.747369
13017	6	1-3645	13015	11	20	835	59	2012-08-02 05:07:03.747369
13018	6	1-1773	13015	20	1	223	90	2012-08-02 05:07:03.747369
13019	6	1-3518	13015	1	51	182	74	2012-08-02 05:07:03.747369
13020	6	1-2722	13015	51	83	121	79	2012-08-02 05:07:03.747369
13015	6	1-3654	13015	83	27	555	75	2012-08-02 05:07:03.747369
13021	1	1-2722	\N	51	51	3169	79	2012-08-02 05:07:03.747369
13023	6	1-1328	13022	17	3	3143	22	2012-08-02 05:07:04.178756
13024	6	1-3597	13022	3	62	2408	73	2012-08-02 05:07:04.178756
13025	6	1-3568	13022	62	19	2049	9	2012-08-02 05:07:04.178756
13026	6	1-3315	13022	19	59	992	45	2012-08-02 05:07:04.178756
13027	6	1-3472	13022	59	20	975	61	2012-08-02 05:07:04.178756
13022	6	1-3655	13022	20	17	3878	56	2012-08-02 05:07:04.178756
13029	5	1-3452	13028	21	16	2016	57	2012-08-02 05:07:05.826012
13030	5	1-1445	13028	16	20	754	59	2012-08-02 05:07:05.826012
13031	5	1-1773	13028	20	1	221	90	2012-08-02 05:07:05.826012
13032	5	1-3518	13028	1	81	198	74	2012-08-02 05:07:05.826012
13028	5	1-3659	13028	81	21	393	18	2012-08-02 05:07:05.826012
13034	6	1-2682	13033	16	21	1553	66	2012-08-02 05:07:05.826012
13035	6	1-3467	13033	21	74	1800	38	2012-08-02 05:07:05.826012
13036	6	1-3476	13033	74	23	723	67	2012-08-02 05:07:05.826012
13037	6	1-3591	13033	23	1	1246	90	2012-08-02 05:07:05.826012
13038	6	1-3518	13033	1	81	1200	74	2012-08-02 05:07:05.826012
13033	6	1-3659	13033	81	16	2551	18	2012-08-02 05:07:05.826012
13040	3	1-3049	13039	95	96	1572	35	2012-08-02 05:07:05.826012
13041	3	1-1785	13039	96	81	1186	74	2012-08-02 05:07:05.826012
13039	3	1-3659	13039	81	95	2561	18	2012-08-02 05:07:05.826012
13043	5	1-2361	13042	6	67	521	91	2012-08-02 05:07:06.234873
13042	5	1-3660	13042	71	6	1852	89	2012-08-02 05:07:06.234873
13044	5	1-3121	13042	67	96	604	39	2012-08-02 05:07:06.234873
13045	5	1-3505	13042	96	68	2551	100	2012-08-02 05:07:06.234873
13046	5	1-3020	13042	68	71	3411	40	2012-08-02 05:07:06.234873
16107	5	1-4637	16107	87	3	2057	87	2012-08-02 05:13:37.184443
13048	5	1-3610	13047	94	5	225	35	2012-08-02 05:07:09.794264
13049	5	1-3042	13047	5	39	321	55	2012-08-02 05:07:09.794264
13050	5	1-3550	13047	39	18	332	52	2012-08-02 05:07:09.794264
13051	5	1-3231	13047	18	88	210	20	2012-08-02 05:07:09.794264
13047	5	1-3669	13047	88	94	260	6	2012-08-02 05:07:09.794264
13053	4	1-3551	13052	25	6	3094	11	2012-08-02 05:07:09.794264
13054	4	1-6	13052	6	22	2044	10	2012-08-02 05:07:09.794264
13055	4	1-3290	13052	22	88	1928	20	2012-08-02 05:07:09.794264
13052	4	1-3669	13052	88	25	2542	6	2012-08-02 05:07:09.794264
13057	3	1-3610	13056	94	76	2162	35	2012-08-02 05:07:09.794264
13058	3	1-3349	13056	76	88	1740	20	2012-08-02 05:07:09.794264
13056	3	1-3669	13056	88	94	2321	6	2012-08-02 05:07:09.794264
13060	5	1-3610	13059	94	69	3105	35	2012-08-02 05:07:09.794264
13061	5	1-2177	13059	69	60	3944	1	2012-08-02 05:07:09.794264
13062	5	1-3261	13059	60	22	2191	10	2012-08-02 05:07:09.794264
13063	5	1-3290	13059	22	88	2219	20	2012-08-02 05:07:09.794264
13059	5	1-3669	13059	88	94	3141	6	2012-08-02 05:07:09.794264
13065	6	1-3497	13064	68	35	6035	11	2012-08-02 05:07:11.020452
13066	6	1-3224	13064	35	74	3838	81	2012-08-02 05:07:11.020452
13067	6	1-3543	13064	74	60	2707	26	2012-08-02 05:07:11.020452
13068	6	1-1887	13064	60	86	843	56	2012-08-02 05:07:11.020452
13069	6	1-3230	13064	86	44	1982	99	2012-08-02 05:07:11.020452
13064	6	1-3672	13064	44	68	6202	37	2012-08-02 05:07:11.020452
13071	6	1-3527	13070	5	38	195	19	2012-08-02 05:07:11.020452
13072	6	1-3286	13070	38	59	155	45	2012-08-02 05:07:11.020452
13073	6	1-3472	13070	59	1	173	61	2012-08-02 05:07:11.020452
13074	6	1-2136	13070	1	86	43	56	2012-08-02 05:07:11.020452
13075	6	1-3230	13070	86	44	102	99	2012-08-02 05:07:11.020452
13070	6	1-3672	13070	44	5	320	37	2012-08-02 05:07:11.020452
13077	5	1-3472	13076	59	96	1355	61	2012-08-02 05:07:11.429588
13078	5	1-3406	13076	96	69	2579	17	2012-08-02 05:07:11.429588
13079	5	1-3652	13076	69	91	2947	57	2012-08-02 05:07:11.429588
13080	5	1-2371	13076	91	75	2586	14	2012-08-02 05:07:11.429588
13076	5	1-3673	13076	75	59	1897	45	2012-08-02 05:07:11.429588
13082	5	1-3472	13081	59	100	259	61	2012-08-02 05:07:11.429588
13083	5	1-2973	13081	100	69	181	58	2012-08-02 05:07:11.429588
13084	5	1-2668	13081	69	91	206	57	2012-08-02 05:07:11.429588
13085	5	1-2371	13081	91	75	253	14	2012-08-02 05:07:11.429588
13081	5	1-3673	13081	75	59	259	45	2012-08-02 05:07:11.429588
13087	3	1-1816	13086	78	62	2023	37	2012-08-02 05:07:12.622715
13088	3	1-3617	13086	62	31	1025	54	2012-08-02 05:07:12.622715
13086	3	1-3676	13086	31	78	2812	84	2012-08-02 05:07:12.622715
13090	5	1-2994	13089	27	11	575	3	2012-08-02 05:07:16.654937
13091	5	1-3645	13089	11	12	1573	59	2012-08-02 05:07:16.654937
13092	5	1-3516	13089	12	57	1973	11	2012-08-02 05:07:16.654937
13093	5	1-2764	13089	57	41	1611	68	2012-08-02 05:07:16.654937
13089	5	1-3686	13089	41	27	655	75	2012-08-02 05:07:16.654937
13095	4	1-2842	13094	25	98	1804	87	2012-08-02 05:07:17.063891
13096	4	1-3653	13094	98	49	5733	76	2012-08-02 05:07:17.063891
13097	4	1-3434	13094	49	47	4357	28	2012-08-02 05:07:17.063891
13094	4	1-3687	13094	47	25	3103	26	2012-08-02 05:07:17.063891
13099	5	1-1521	13098	6	75	307	14	2012-08-02 05:07:17.461566
13100	5	1-3673	13098	75	59	350	45	2012-08-02 05:07:17.461566
13101	5	1-3472	13098	59	10	389	61	2012-08-02 05:07:17.461566
13102	5	1-3627	13098	10	30	350	28	2012-08-02 05:07:17.461566
13098	5	1-3688	13098	30	6	583	13	2012-08-02 05:07:17.461566
13104	5	1-3609	13103	47	70	290	91	2012-08-02 05:07:18.257344
13105	5	1-1623	13103	70	98	187	87	2012-08-02 05:07:18.257344
13106	5	1-3653	13103	98	29	586	76	2012-08-02 05:07:18.257344
13107	5	1-1712	13103	29	66	577	53	2012-08-02 05:07:18.257344
13103	5	1-3690	13103	66	47	403	69	2012-08-02 05:07:18.257344
13109	4	1-3664	13108	5	77	3910	64	2012-08-02 05:07:19.837065
13110	4	1-1633	13108	77	13	3269	33	2012-08-02 05:07:19.837065
13111	4	1-2758	13108	13	69	2289	67	2012-08-02 05:07:19.837065
13108	4	1-3694	13108	69	5	3111	3	2012-08-02 05:07:19.837065
13113	6	1-3502	13112	84	55	1469	9	2012-08-02 05:07:21.052233
13114	6	1-3361	13112	55	84	1875	7	2012-08-02 05:07:21.052233
13115	6	1-2708	13112	84	12	494	35	2012-08-02 05:07:21.052233
13116	6	1-2275	13112	12	57	695	78	2012-08-02 05:07:21.052233
13117	6	1-3685	13112	57	56	598	24	2012-08-02 05:07:21.052233
13112	6	1-3697	13112	56	84	3342	50	2012-08-02 05:07:21.052233
13119	5	1-2860	13118	16	3	1225	26	2012-08-02 05:07:22.22319
13120	5	1-3294	13118	3	23	377	38	2012-08-02 05:07:22.22319
13121	5	1-3228	13118	23	11	677	3	2012-08-02 05:07:22.22319
13122	5	1-3645	13118	11	72	1581	59	2012-08-02 05:07:22.22319
13118	5	1-3700	13118	72	16	2215	60	2012-08-02 05:07:22.22319
13124	4	1-3494	13123	58	14	4875	48	2012-08-02 05:07:24.200778
13125	4	1-2196	13123	14	84	724	4	2012-08-02 05:07:24.200778
13126	4	1-3510	13123	84	59	545	67	2012-08-02 05:07:24.200778
13123	4	1-3705	13123	59	58	4814	1	2012-08-02 05:07:24.200778
13128	6	1-3689	13127	57	82	194	58	2012-08-02 05:07:24.200778
13129	6	1-3606	13127	82	90	269	5	2012-08-02 05:07:24.200778
13130	6	1-3488	13127	90	61	117	86	2012-08-02 05:07:24.200778
13131	6	1-1074	13127	61	84	45	4	2012-08-02 05:07:24.200778
13132	6	1-3510	13127	84	59	37	67	2012-08-02 05:07:24.200778
13127	6	1-3705	13127	59	57	355	1	2012-08-02 05:07:24.200778
13134	5	1-3390	13133	80	17	145	72	2012-08-02 05:07:24.9742
13135	5	1-3509	13133	17	91	165	19	2012-08-02 05:07:24.9742
13136	5	1-3605	13133	91	10	161	54	2012-08-02 05:07:24.9742
13137	5	1-3520	13133	10	58	1007	41	2012-08-02 05:07:24.9742
13133	5	1-3707	13133	58	80	362	89	2012-08-02 05:07:24.9742
13139	5	1-3042	13138	5	30	874	55	2012-08-02 05:07:26.145305
13140	5	1-2775	13138	30	16	154	29	2012-08-02 05:07:26.145305
13141	5	1-3646	13138	16	77	1488	44	2012-08-02 05:07:26.145305
13142	5	1-2175	13138	77	11	1329	20	2012-08-02 05:07:26.145305
13138	5	1-3710	13138	11	5	585	35	2012-08-02 05:07:26.145305
13144	5	1-3479	13143	70	43	2882	43	2012-08-02 05:07:29.349072
13145	5	1-3564	13143	43	19	624	62	2012-08-02 05:07:29.349072
13146	5	1-2681	13143	19	49	1745	53	2012-08-02 05:07:29.349072
13147	5	1-2969	13143	49	92	693	76	2012-08-02 05:07:29.349072
13143	5	1-3718	13143	92	70	1604	84	2012-08-02 05:07:29.349072
13149	4	1-3479	13148	70	94	3084	43	2012-08-02 05:07:29.349072
13150	4	1-3003	13148	94	68	703	8	2012-08-02 05:07:29.349072
13151	4	1-3090	13148	68	92	654	76	2012-08-02 05:07:29.349072
13148	4	1-3718	13148	92	70	1612	84	2012-08-02 05:07:29.349072
16113	5	1-4578	16112	17	10	1779	18	2012-08-02 05:13:37.990737
13153	6	1-3713	13152	12	17	3459	56	2012-08-02 05:07:29.735992
13154	6	1-1328	13152	17	18	3063	22	2012-08-02 05:07:29.735992
13155	6	1-3277	13152	18	35	3044	81	2012-08-02 05:07:29.735992
13156	6	1-3323	13152	35	34	1362	6	2012-08-02 05:07:29.735992
13157	6	1-3715	13152	34	50	3580	77	2012-08-02 05:07:29.735992
13152	6	1-3719	13152	50	12	6410	62	2012-08-02 05:07:29.735992
16114	5	1-4605	16112	10	78	4114	86	2012-08-02 05:13:37.990737
13159	4	1-2892	13158	43	33	116	63	2012-08-02 05:07:30.918246
13160	4	1-3080	13158	33	18	106	93	2012-08-02 05:07:30.918246
13161	4	1-2795	13158	18	17	178	40	2012-08-02 05:07:30.918246
13158	4	1-3722	13158	17	43	120	36	2012-08-02 05:07:30.918246
13163	5	1-3488	13162	90	75	6	86	2012-08-02 05:07:32.486979
13164	5	1-1825	13162	75	95	7	93	2012-08-02 05:07:32.486979
13165	5	1-3725	13162	95	87	2	31	2012-08-02 05:07:32.486979
13166	5	1-3596	13162	87	48	1	1	2012-08-02 05:07:32.486979
13162	5	1-3726	13162	48	90	16	5	2012-08-02 05:07:32.486979
13168	5	1-3488	13167	90	47	3114	86	2012-08-02 05:07:32.486979
13169	5	1-1307	13167	47	94	997	24	2012-08-02 05:07:32.486979
13170	5	1-2782	13167	94	78	387	92	2012-08-02 05:07:32.486979
13171	5	1-3441	13167	78	48	415	1	2012-08-02 05:07:32.486979
13167	5	1-3726	13167	48	90	7587	5	2012-08-02 05:07:32.486979
13173	5	1-2664	13172	97	88	1179	88	2012-08-02 05:07:32.486979
13174	5	1-3348	13172	88	68	208	37	2012-08-02 05:07:32.486979
13175	5	1-3497	13172	68	74	143	11	2012-08-02 05:07:32.486979
13176	5	1-1164	13172	74	48	61	1	2012-08-02 05:07:32.486979
13172	5	1-3726	13172	48	97	1177	5	2012-08-02 05:07:32.486979
13178	3	1-3628	13177	35	95	5694	89	2012-08-02 05:07:33.271549
13179	3	1-3321	13177	95	89	2198	51	2012-08-02 05:07:33.271549
13177	3	1-3728	13177	89	35	3468	91	2012-08-02 05:07:33.271549
13181	3	1-3675	13180	20	66	847	61	2012-08-02 05:07:33.271549
13182	3	1-3031	13180	66	89	980	51	2012-08-02 05:07:33.271549
13180	3	1-3728	13180	89	20	1809	91	2012-08-02 05:07:33.271549
13184	3	1-3568	13183	62	55	1837	9	2012-08-02 05:07:34.066828
13185	3	1-3361	13183	55	14	1395	7	2012-08-02 05:07:34.066828
13183	3	1-3730	13183	14	62	3661	73	2012-08-02 05:07:34.066828
13187	4	1-3692	13186	7	31	1160	67	2012-08-02 05:07:34.066828
13188	4	1-3478	13186	31	55	477	9	2012-08-02 05:07:34.066828
13189	4	1-3361	13186	55	14	529	7	2012-08-02 05:07:34.066828
13186	4	1-3730	13186	14	7	2032	73	2012-08-02 05:07:34.066828
13191	5	1-3248	13190	9	7	941	2	2012-08-02 05:07:35.238379
13192	5	1-3699	13190	7	92	713	76	2012-08-02 05:07:35.238379
13193	5	1-3718	13190	92	26	1316	84	2012-08-02 05:07:35.238379
13194	5	1-3000	13190	26	75	957	46	2012-08-02 05:07:35.238379
13190	5	1-3733	13190	75	9	1254	79	2012-08-02 05:07:35.238379
13196	5	1-3673	13195	75	59	1288	45	2012-08-02 05:07:38.366413
13197	5	1-3472	13195	59	3	1060	61	2012-08-02 05:07:38.366413
13198	5	1-59	13195	3	41	424	68	2012-08-02 05:07:38.366413
13199	5	1-3686	13195	41	88	126	75	2012-08-02 05:07:38.366413
13195	5	1-3741	13195	88	75	1523	14	2012-08-02 05:07:38.366413
13201	3	1-3243	13200	63	66	504	65	2012-08-02 05:07:38.366413
13202	3	1-3388	13200	66	88	304	75	2012-08-02 05:07:38.366413
13200	3	1-3741	13200	88	63	3909	14	2012-08-02 05:07:38.366413
13204	5	1-3243	13203	63	4	288	65	2012-08-02 05:07:38.366413
13205	5	1-3371	13203	4	24	380	8	2012-08-02 05:07:38.366413
13206	5	1-2862	13203	24	41	314	68	2012-08-02 05:07:38.366413
13207	5	1-3686	13203	41	88	120	75	2012-08-02 05:07:38.366413
13203	5	1-3741	13203	88	63	1855	14	2012-08-02 05:07:38.366413
13209	5	1-2735	13208	66	87	3551	31	2012-08-02 05:07:39.548438
13210	5	1-3596	13208	87	60	2503	1	2012-08-02 05:07:39.548438
13211	5	1-3261	13208	60	22	1348	10	2012-08-02 05:07:39.548438
13212	5	1-3290	13208	22	61	1323	20	2012-08-02 05:07:39.548438
13208	5	1-3744	13208	61	66	8838	2	2012-08-02 05:07:39.548438
13214	3	1-3743	13213	51	24	6219	58	2012-08-02 05:07:41.990288
13215	3	1-1783	13213	24	78	600	87	2012-08-02 05:07:41.990288
13213	3	1-3750	13213	78	51	4552	93	2012-08-02 05:07:41.990288
13217	3	1-3646	13216	16	41	3283	44	2012-08-02 05:07:42.376874
13218	3	1-3470	13216	41	94	1241	10	2012-08-02 05:07:42.376874
13216	3	1-3751	13216	94	16	425	29	2012-08-02 05:07:42.376874
13220	4	1-3646	13219	16	41	407	44	2012-08-02 05:07:42.376874
13221	4	1-331	13219	41	2	107	100	2012-08-02 05:07:42.376874
13222	4	1-3702	13219	2	94	108	10	2012-08-02 05:07:42.376874
13219	4	1-3751	13219	94	16	44	29	2012-08-02 05:07:42.376874
13224	2	1-3171	13223	89	94	1293	10	2012-08-02 05:07:42.376874
13223	2	1-3751	13223	94	89	581	29	2012-08-02 05:07:42.376874
13226	5	1-1074	13225	61	17	216	4	2012-08-02 05:07:42.774635
13227	5	1-2804	13225	17	10	225	54	2012-08-02 05:07:42.774635
13228	5	1-3520	13225	10	58	1565	41	2012-08-02 05:07:42.774635
13229	5	1-3707	13225	58	69	624	89	2012-08-02 05:07:42.774635
13225	5	1-3752	13225	69	61	459	86	2012-08-02 05:07:42.774635
13231	3	1-3686	13230	41	57	458	75	2012-08-02 05:07:43.946041
13232	3	1-3436	13230	57	27	1612	15	2012-08-02 05:07:43.946041
13230	3	1-3755	13230	27	41	1162	68	2012-08-02 05:07:43.946041
13234	3	1-936	13233	48	9	338	95	2012-08-02 05:07:44.354554
13235	3	1-3082	13233	9	47	108	92	2012-08-02 05:07:44.354554
13233	3	1-3756	13233	47	48	1047	11	2012-08-02 05:07:44.354554
13237	4	1-3516	13236	12	57	2557	11	2012-08-02 05:07:45.149898
13238	4	1-2764	13236	57	55	1145	68	2012-08-02 05:07:45.149898
13239	4	1-3345	13236	55	43	443	93	2012-08-02 05:07:45.149898
13236	4	1-3758	13236	43	12	3719	59	2012-08-02 05:07:45.149898
13241	3	1-3698	13240	3	5	2438	64	2012-08-02 05:07:47.12749
13242	3	1-3548	13240	5	86	2891	34	2012-08-02 05:07:47.12749
13240	3	1-3763	13240	86	3	6146	1	2012-08-02 05:07:47.12749
13244	5	1-3467	13243	21	72	3292	38	2012-08-02 05:07:49.094187
13245	5	1-3298	13243	72	31	1235	54	2012-08-02 05:07:49.094187
13246	5	1-3676	13243	31	88	3381	84	2012-08-02 05:07:49.094187
13247	5	1-3737	13243	88	53	3624	93	2012-08-02 05:07:49.094187
13243	5	1-3768	13243	53	21	2810	66	2012-08-02 05:07:49.094187
13249	5	1-3720	13248	46	18	929	22	2012-08-02 05:07:49.094187
13250	5	1-3277	13248	18	85	917	81	2012-08-02 05:07:49.094187
13251	5	1-2999	13248	85	88	590	84	2012-08-02 05:07:49.094187
13252	5	1-3737	13248	88	53	637	93	2012-08-02 05:07:49.094187
13248	5	1-3768	13248	53	46	498	66	2012-08-02 05:07:49.094187
16115	5	1-1498	16112	78	20	980	24	2012-08-02 05:13:37.990737
13254	2	1-1740	13253	69	53	2141	93	2012-08-02 05:07:49.094187
13253	2	1-3768	13253	53	69	1680	66	2012-08-02 05:07:49.094187
16116	5	1-4597	16112	20	35	2419	56	2012-08-02 05:13:37.990737
13256	5	1-3555	13255	23	3	232	60	2012-08-02 05:07:49.094187
13257	5	1-3460	13255	3	9	287	95	2012-08-02 05:07:49.094187
13258	5	1-3082	13255	9	68	121	92	2012-08-02 05:07:49.094187
13259	5	1-2720	13255	68	53	159	93	2012-08-02 05:07:49.094187
13255	5	1-3768	13255	53	23	126	66	2012-08-02 05:07:49.094187
13261	5	1-2200	13260	87	61	930	59	2012-08-02 05:07:49.49205
13262	5	1-3703	13260	61	32	841	15	2012-08-02 05:07:49.49205
13263	5	1-2909	13260	32	68	621	70	2012-08-02 05:07:49.49205
13264	5	1-3001	13260	68	75	874	88	2012-08-02 05:07:49.49205
13260	5	1-3769	13260	75	87	825	52	2012-08-02 05:07:49.49205
13266	4	1-2518	13265	25	100	2363	83	2012-08-02 05:07:49.49205
13267	4	1-3450	13265	100	82	2914	17	2012-08-02 05:07:49.49205
13268	4	1-2900	13265	82	75	3042	88	2012-08-02 05:07:49.49205
13265	4	1-3769	13265	75	25	2910	52	2012-08-02 05:07:49.49205
13270	5	1-3376	13269	10	66	2608	53	2012-08-02 05:07:49.49205
13271	5	1-3690	13269	66	79	1957	69	2012-08-02 05:07:49.49205
13272	5	1-2104	13269	79	97	1587	5	2012-08-02 05:07:49.49205
13273	5	1-2664	13269	97	75	2177	88	2012-08-02 05:07:49.49205
13269	5	1-3769	13269	75	10	2098	52	2012-08-02 05:07:49.49205
13275	4	1-3502	13274	84	97	909	9	2012-08-02 05:07:51.447838
13276	4	1-3753	13274	97	89	263	10	2012-08-02 05:07:51.447838
13277	4	1-3182	13274	89	63	76	49	2012-08-02 05:07:51.447838
13274	4	1-3774	13274	63	84	3121	50	2012-08-02 05:07:51.447838
13279	5	1-3389	13278	87	7	2413	73	2012-08-02 05:07:51.447838
13280	5	1-3692	13278	7	71	1236	67	2012-08-02 05:07:51.447838
13281	5	1-1625	13278	71	89	351	10	2012-08-02 05:07:51.447838
13282	5	1-3182	13278	89	63	110	49	2012-08-02 05:07:51.447838
13278	5	1-3774	13278	63	87	4853	50	2012-08-02 05:07:51.447838
13284	3	1-2459	13283	50	94	2040	10	2012-08-02 05:07:51.845501
13285	3	1-3751	13283	94	42	638	29	2012-08-02 05:07:51.845501
13283	3	1-3775	13283	42	50	4880	82	2012-08-02 05:07:51.845501
13287	5	1-59	13286	3	41	181	68	2012-08-02 05:07:52.254268
13288	5	1-3686	13286	41	88	51	75	2012-08-02 05:07:52.254268
13289	5	1-3741	13286	88	75	579	14	2012-08-02 05:07:52.254268
13290	5	1-3673	13286	75	40	462	45	2012-08-02 05:07:52.254268
13286	5	1-3776	13286	40	3	481	61	2012-08-02 05:07:52.254268
13292	4	1-3503	13291	69	100	2479	94	2012-08-02 05:07:52.254268
13293	4	1-3113	13291	100	75	755	14	2012-08-02 05:07:52.254268
13294	4	1-3673	13291	75	40	848	45	2012-08-02 05:07:52.254268
13291	4	1-3776	13291	40	69	1245	61	2012-08-02 05:07:52.254268
13296	3	1-3646	13295	16	77	1427	44	2012-08-02 05:07:53.812011
13297	3	1-2175	13295	77	39	441	20	2012-08-02 05:07:53.812011
13295	3	1-3780	13295	39	16	426	29	2012-08-02 05:07:53.812011
13299	4	1-3443	13298	21	84	1293	60	2012-08-02 05:07:53.812011
13300	4	1-2004	13298	84	77	708	44	2012-08-02 05:07:53.812011
13301	4	1-2175	13298	77	39	501	20	2012-08-02 05:07:53.812011
13298	4	1-3780	13298	39	21	1109	29	2012-08-02 05:07:53.812011
13303	5	1-3443	13302	21	3	86	60	2012-08-02 05:07:53.812011
13304	5	1-3460	13302	3	9	89	95	2012-08-02 05:07:53.812011
13305	5	1-3082	13302	9	61	31	92	2012-08-02 05:07:53.812011
13306	5	1-2792	13302	61	39	29	20	2012-08-02 05:07:53.812011
13302	5	1-3780	13302	39	21	69	29	2012-08-02 05:07:53.812011
13308	4	1-3188	13307	23	4	1667	63	2012-08-02 05:07:53.812011
13309	4	1-1106	13307	4	22	327	10	2012-08-02 05:07:53.812011
13310	4	1-3290	13307	22	39	302	20	2012-08-02 05:07:53.812011
13307	4	1-3780	13307	39	23	843	29	2012-08-02 05:07:53.812011
13312	5	1-1315	13311	69	71	1714	40	2012-08-02 05:07:54.972474
13313	5	1-3660	13311	71	83	901	89	2012-08-02 05:07:54.972474
13314	5	1-3392	13311	83	74	1046	42	2012-08-02 05:07:54.972474
13315	5	1-2904	13311	74	39	1010	32	2012-08-02 05:07:54.972474
13311	5	1-3783	13311	39	69	1639	12	2012-08-02 05:07:54.972474
13317	5	1-3228	13316	23	70	2445	3	2012-08-02 05:07:58.087669
13318	5	1-2033	13316	70	35	2158	81	2012-08-02 05:07:58.087669
13319	5	1-3323	13316	35	94	892	6	2012-08-02 05:07:58.087669
13320	5	1-3610	13316	94	61	754	35	2012-08-02 05:07:58.087669
13316	5	1-3791	13316	61	23	1320	38	2012-08-02 05:07:58.087669
13322	5	1-3228	13321	23	52	699	3	2012-08-02 05:07:58.087669
13323	5	1-1275	13321	52	35	465	81	2012-08-02 05:07:58.087669
13324	5	1-3323	13321	35	94	206	6	2012-08-02 05:07:58.087669
13325	5	1-3610	13321	94	61	187	35	2012-08-02 05:07:58.087669
13321	5	1-3791	13321	61	23	352	38	2012-08-02 05:07:58.087669
13326	1	1-1275	\N	52	52	2134	81	2012-08-02 05:07:58.087669
13328	5	1-3597	13327	3	7	688	73	2012-08-02 05:07:58.496521
13329	5	1-3692	13327	7	8	504	67	2012-08-02 05:07:58.496521
13330	5	1-3691	13327	8	86	652	39	2012-08-02 05:07:58.496521
13331	5	1-2326	13327	86	6	608	42	2012-08-02 05:07:58.496521
13327	5	1-3792	13327	6	3	812	22	2012-08-02 05:07:58.496521
13333	5	1-3200	13332	91	35	1559	85	2012-08-02 05:07:58.496521
13334	5	1-3711	13332	35	57	3764	78	2012-08-02 05:07:58.496521
13335	5	1-3685	13332	57	49	3679	24	2012-08-02 05:07:58.496521
13336	5	1-3270	13332	49	6	2211	42	2012-08-02 05:07:58.496521
13332	5	1-3792	13332	6	91	3011	22	2012-08-02 05:07:58.496521
13338	4	1-2155	13337	32	31	588	67	2012-08-02 05:07:59.668069
13339	4	1-3478	13337	31	8	303	9	2012-08-02 05:07:59.668069
13340	4	1-3760	13337	8	49	495	28	2012-08-02 05:07:59.668069
13337	4	1-3795	13337	49	32	1473	97	2012-08-02 05:07:59.668069
13342	3	1-3382	13341	25	61	1560	98	2012-08-02 05:08:00.827619
13343	3	1-2920	13341	61	83	548	31	2012-08-02 05:08:00.827619
13341	3	1-3798	13341	83	25	6073	19	2012-08-02 05:08:00.827619
13345	3	1-2764	13344	57	71	722	68	2012-08-02 05:08:01.236844
13346	3	1-2729	13344	71	72	273	61	2012-08-02 05:08:01.236844
13344	3	1-3799	13344	72	57	1020	11	2012-08-02 05:08:01.236844
13348	4	1-576	13347	80	95	1966	55	2012-08-02 05:08:01.236844
13349	4	1-3681	13347	95	40	1290	45	2012-08-02 05:08:01.236844
13350	4	1-3776	13347	40	72	1741	61	2012-08-02 05:08:01.236844
13347	4	1-3799	13347	72	80	6716	11	2012-08-02 05:08:01.236844
13352	4	1-2851	13351	36	54	1224	73	2012-08-02 05:08:02.805654
13353	4	1-3650	13351	54	96	855	13	2012-08-02 05:08:02.805654
13354	4	1-3179	13351	96	15	324	94	2012-08-02 05:08:02.805654
13351	4	1-3803	13351	15	36	2417	63	2012-08-02 05:08:02.805654
13356	5	1-3096	13355	83	36	3300	12	2012-08-02 05:08:02.805654
13357	5	1-686	13355	36	9	640	93	2012-08-02 05:08:02.805654
13358	5	1-3666	13355	9	96	385	13	2012-08-02 05:08:02.805654
13359	5	1-3179	13355	96	15	287	94	2012-08-02 05:08:02.805654
13355	5	1-3803	13355	15	83	4216	63	2012-08-02 05:08:02.805654
16112	5	1-4639	16112	35	17	1988	9	2012-08-02 05:13:37.990737
13361	3	1-3610	13360	94	61	825	35	2012-08-02 05:08:03.192177
13362	3	1-3791	13360	61	34	1328	38	2012-08-02 05:08:03.192177
13360	3	1-3804	13360	34	94	1060	6	2012-08-02 05:08:03.192177
13364	6	1-3721	13363	71	42	4173	72	2012-08-02 05:08:03.192177
13365	6	1-3613	13363	42	55	2913	9	2012-08-02 05:08:03.192177
13366	6	1-3361	13363	55	84	4162	7	2012-08-02 05:08:03.192177
13367	6	1-2708	13363	84	61	1125	35	2012-08-02 05:08:03.192177
13368	6	1-3791	13363	61	34	2166	38	2012-08-02 05:08:03.192177
13363	6	1-3804	13363	34	71	2068	6	2012-08-02 05:08:03.192177
16118	3	1-4481	16117	54	23	1814	3	2012-08-02 05:13:37.990737
13370	4	1-3690	13369	66	47	1708	69	2012-08-02 05:08:03.601096
13371	4	1-3609	13369	47	65	1387	91	2012-08-02 05:08:03.601096
13372	4	1-541	13369	65	69	1298	52	2012-08-02 05:08:03.601096
13369	4	1-3805	13369	69	66	2171	53	2012-08-02 05:08:03.601096
13374	6	1-3789	13373	66	50	3340	14	2012-08-02 05:08:07.205386
13375	6	1-3400	13373	50	83	2535	5	2012-08-02 05:08:07.205386
13376	6	1-3729	13373	83	94	639	85	2012-08-02 05:08:07.205386
13377	6	1-3420	13373	94	43	1360	43	2012-08-02 05:08:07.205386
13378	6	1-3564	13373	43	32	298	62	2012-08-02 05:08:07.205386
13373	6	1-3814	13373	32	66	910	60	2012-08-02 05:08:07.205386
13380	5	1-3404	13379	40	69	1481	70	2012-08-02 05:08:08.012107
13381	5	1-3815	13379	69	50	2575	77	2012-08-02 05:08:08.012107
13382	5	1-2863	13379	50	75	2172	14	2012-08-02 05:08:08.012107
13383	5	1-3673	13379	75	68	2249	45	2012-08-02 05:08:08.012107
13379	5	1-3816	13379	68	40	2477	90	2012-08-02 05:08:08.012107
13385	5	1-3692	13384	7	69	542	67	2012-08-02 05:08:08.409814
13386	5	1-3694	13384	69	5	723	3	2012-08-02 05:08:08.409814
13387	5	1-3664	13384	5	12	891	64	2012-08-02 05:08:08.409814
13388	5	1-1641	13384	12	68	651	13	2012-08-02 05:08:08.409814
13384	5	1-3817	13384	68	7	812	73	2012-08-02 05:08:08.409814
13390	5	1-3192	13389	83	51	1165	93	2012-08-02 05:08:09.216437
13391	5	1-3743	13389	51	42	3183	58	2012-08-02 05:08:09.216437
13392	5	1-1232	13389	42	29	1618	76	2012-08-02 05:08:09.216437
13393	5	1-1712	13389	29	44	1753	53	2012-08-02 05:08:09.216437
13389	5	1-3819	13389	44	83	1250	9	2012-08-02 05:08:09.216437
13395	5	1-2730	13394	45	82	598	17	2012-08-02 05:08:09.216437
13396	5	1-2900	13394	82	75	646	88	2012-08-02 05:08:09.216437
13397	5	1-3769	13394	75	69	639	52	2012-08-02 05:08:09.216437
13398	5	1-3805	13394	69	44	1045	53	2012-08-02 05:08:09.216437
13394	5	1-3819	13394	44	45	747	9	2012-08-02 05:08:09.216437
13400	5	1-2474	13399	1	53	324	71	2012-08-02 05:08:09.216437
13401	5	1-3696	13399	53	39	259	55	2012-08-02 05:08:09.216437
13402	5	1-3550	13399	39	69	299	52	2012-08-02 05:08:09.216437
13403	5	1-3805	13399	69	44	500	53	2012-08-02 05:08:09.216437
13399	5	1-3819	13399	44	1	365	9	2012-08-02 05:08:09.216437
13405	5	1-837	13404	85	18	408	93	2012-08-02 05:08:10.420925
13406	5	1-2795	13404	18	17	685	40	2012-08-02 05:08:10.420925
13407	5	1-3722	13404	17	43	460	36	2012-08-02 05:08:10.420925
13408	5	1-2892	13404	43	41	444	63	2012-08-02 05:08:10.420925
13404	5	1-3822	13404	41	85	2228	57	2012-08-02 05:08:10.420925
13410	2	1-1774	13409	35	51	4569	44	2012-08-02 05:08:12.851028
13409	2	1-3828	13409	51	35	9238	56	2012-08-02 05:08:12.851028
13412	2	1-3031	13411	66	95	4984	51	2012-08-02 05:08:13.248897
13411	2	1-3829	13411	95	66	9258	61	2012-08-02 05:08:13.248897
13414	5	1-3765	13413	4	56	11	56	2012-08-02 05:08:14.055607
13415	5	1-3563	13413	56	51	11	93	2012-08-02 05:08:14.055607
13416	5	1-3743	13413	51	12	29	58	2012-08-02 05:08:14.055607
13417	5	1-1486	13413	12	42	24	51	2012-08-02 05:08:14.055607
13413	5	1-3831	13413	42	4	13	6	2012-08-02 05:08:14.055607
13419	5	1-3715	13418	34	60	1129	77	2012-08-02 05:08:14.055607
13420	5	1-1264	13418	60	87	372	27	2012-08-02 05:08:14.055607
13421	5	1-2915	13418	87	80	377	70	2012-08-02 05:08:14.055607
13422	5	1-3679	13418	80	42	707	51	2012-08-02 05:08:14.055607
13418	5	1-3831	13418	42	34	415	6	2012-08-02 05:08:14.055607
13424	5	1-3715	13423	34	21	4433	77	2012-08-02 05:08:14.055607
13425	5	1-3587	13423	21	100	2142	37	2012-08-02 05:08:14.055607
13426	5	1-3830	13423	100	12	2983	58	2012-08-02 05:08:14.055607
13427	5	1-1486	13423	12	42	2690	51	2012-08-02 05:08:14.055607
13423	5	1-3831	13423	42	34	1603	6	2012-08-02 05:08:14.055607
13429	4	1-3721	13428	71	55	2205	72	2012-08-02 05:08:14.055607
13430	4	1-2888	13428	55	90	1906	21	2012-08-02 05:08:14.055607
13431	4	1-3553	13428	90	42	1713	51	2012-08-02 05:08:14.055607
13428	4	1-3831	13428	42	71	1038	6	2012-08-02 05:08:14.055607
13433	4	1-3721	13432	71	65	1357	72	2012-08-02 05:08:14.055607
13434	4	1-2800	13432	65	66	655	61	2012-08-02 05:08:14.055607
13435	4	1-3031	13432	66	42	1036	51	2012-08-02 05:08:14.055607
13432	4	1-3831	13432	42	71	633	6	2012-08-02 05:08:14.055607
13437	5	1-3733	13436	75	57	3192	79	2012-08-02 05:08:15.754458
13438	5	1-1680	13436	57	94	1705	10	2012-08-02 05:08:15.754458
13439	5	1-3751	13436	94	21	448	29	2012-08-02 05:08:15.754458
13440	5	1-3443	13436	21	75	403	60	2012-08-02 05:08:15.754458
13436	5	1-3835	13436	75	75	2269	46	2012-08-02 05:08:15.754458
13442	3	1-3371	13441	4	24	337	8	2012-08-02 05:08:18.529785
13443	3	1-2862	13441	24	45	297	68	2012-08-02 05:08:18.529785
13441	3	1-3842	13441	45	4	240	65	2012-08-02 05:08:18.529785
13445	5	1-3591	13444	23	24	1801	90	2012-08-02 05:08:19.700831
13446	5	1-3566	13444	24	75	1474	88	2012-08-02 05:08:19.700831
13447	5	1-3769	13444	75	69	1249	52	2012-08-02 05:08:19.700831
13448	5	1-3805	13444	69	85	1749	53	2012-08-02 05:08:19.700831
13444	5	1-3845	13444	85	23	1153	67	2012-08-02 05:08:19.700831
13450	4	1-3694	13449	69	11	1218	3	2012-08-02 05:08:19.700831
13451	4	1-3645	13449	11	48	3193	59	2012-08-02 05:08:19.700831
13452	4	1-3306	13449	48	85	1125	53	2012-08-02 05:08:19.700831
13449	4	1-3845	13449	85	69	841	67	2012-08-02 05:08:19.700831
13454	3	1-3832	13453	80	45	1830	99	2012-08-02 05:08:19.700831
13455	3	1-1402	13453	45	85	2548	53	2012-08-02 05:08:19.700831
13453	3	1-3845	13453	85	80	1998	67	2012-08-02 05:08:19.700831
13456	1	1-701	\N	51	51	6816	53	2012-08-02 05:08:19.700831
13458	4	1-3832	13457	80	36	3874	99	2012-08-02 05:08:20.098783
13457	4	1-3846	13457	35	80	4580	67	2012-08-02 05:08:20.098783
13459	4	1-3837	13457	36	11	4838	23	2012-08-02 05:08:20.098783
13460	4	1-3746	13457	11	35	5397	80	2012-08-02 05:08:20.098783
16119	3	1-4313	16117	23	35	2628	56	2012-08-02 05:13:37.990737
13462	5	1-3691	13461	8	69	1776	39	2012-08-02 05:08:20.098783
13463	5	1-1891	13461	69	17	2823	40	2012-08-02 05:08:20.098783
13464	5	1-3722	13461	17	6	2160	36	2012-08-02 05:08:20.098783
13465	5	1-2226	13461	6	35	1508	80	2012-08-02 05:08:20.098783
13461	5	1-3846	13461	35	8	1352	67	2012-08-02 05:08:20.098783
16117	3	1-4639	16117	35	54	2560	9	2012-08-02 05:13:37.990737
13467	5	1-3554	13466	77	69	1495	67	2012-08-02 05:08:21.292559
13468	5	1-3694	13466	69	5	2160	3	2012-08-02 05:08:21.292559
13469	5	1-3664	13466	5	78	2886	64	2012-08-02 05:08:21.292559
13470	5	1-1857	13466	78	46	654	92	2012-08-02 05:08:21.292559
13466	5	1-3849	13466	46	77	1034	82	2012-08-02 05:08:21.292559
13472	6	1-1554	13471	72	44	958	99	2012-08-02 05:08:21.700964
13473	6	1-3672	13471	44	62	2753	37	2012-08-02 05:08:21.700964
13474	6	1-3617	13471	62	68	1343	54	2012-08-02 05:08:21.700964
13475	6	1-3839	13471	68	57	2455	78	2012-08-02 05:08:21.700964
13476	6	1-3685	13471	57	9	2214	24	2012-08-02 05:08:21.700964
13471	6	1-3850	13471	9	72	1801	38	2012-08-02 05:08:21.700964
13478	5	1-3021	13477	39	62	323	37	2012-08-02 05:08:21.700964
13479	5	1-3617	13477	62	68	165	54	2012-08-02 05:08:21.700964
13480	5	1-3839	13477	68	57	315	78	2012-08-02 05:08:21.700964
13481	5	1-3685	13477	57	9	297	24	2012-08-02 05:08:21.700964
13477	5	1-3850	13477	9	39	253	38	2012-08-02 05:08:21.700964
13483	4	1-3721	13482	71	65	358	72	2012-08-02 05:08:22.120982
13484	4	1-2800	13482	65	100	146	61	2012-08-02 05:08:22.120982
13485	4	1-2973	13482	100	69	101	58	2012-08-02 05:08:22.120982
13482	4	1-3851	13482	69	71	197	6	2012-08-02 05:08:22.120982
13487	4	1-3551	13486	25	74	2224	11	2012-08-02 05:08:22.120982
13488	4	1-1164	13486	74	57	1258	1	2012-08-02 05:08:22.120982
13489	4	1-3689	13486	57	69	905	58	2012-08-02 05:08:22.120982
13486	4	1-3851	13486	69	25	1886	6	2012-08-02 05:08:22.120982
13490	1	1-1164	\N	74	74	1132	1	2012-08-02 05:08:22.120982
13492	5	1-2540	13491	11	42	13	72	2012-08-02 05:08:24.485229
13493	5	1-3613	13491	42	8	6	9	2012-08-02 05:08:24.485229
13494	5	1-3760	13491	8	49	8	28	2012-08-02 05:08:24.485229
13495	5	1-3795	13491	49	8	15	97	2012-08-02 05:08:24.485229
13491	5	1-3857	13491	8	11	22	64	2012-08-02 05:08:24.485229
13497	5	1-2375	13496	86	64	1295	31	2012-08-02 05:08:24.485229
13498	5	1-2818	13496	64	14	163	47	2012-08-02 05:08:24.485229
13499	5	1-2537	13496	14	49	145	28	2012-08-02 05:08:24.485229
13500	5	1-3795	13496	49	8	430	97	2012-08-02 05:08:24.485229
13496	5	1-3857	13496	8	86	754	64	2012-08-02 05:08:24.485229
13502	5	1-3857	13501	8	75	5915	64	2012-08-02 05:08:25.280733
13503	5	1-2128	13501	75	75	4099	83	2012-08-02 05:08:25.280733
13504	5	1-3639	13501	75	65	1464	98	2012-08-02 05:08:25.280733
13505	5	1-3601	13501	65	87	1584	55	2012-08-02 05:08:25.280733
13501	5	1-3859	13501	87	8	3226	97	2012-08-02 05:08:25.280733
13507	5	1-3857	13506	8	5	103	64	2012-08-02 05:08:25.280733
13508	5	1-3548	13506	5	65	120	34	2012-08-02 05:08:25.280733
13509	5	1-1997	13506	65	30	84	79	2012-08-02 05:08:25.280733
13510	5	1-3351	13506	30	87	27	55	2012-08-02 05:08:25.280733
13506	5	1-3859	13506	87	8	56	97	2012-08-02 05:08:25.280733
13512	4	1-3550	13511	39	33	269	52	2012-08-02 05:08:25.70064
13513	4	1-1405	13511	33	39	219	32	2012-08-02 05:08:25.70064
13514	4	1-3783	13511	39	28	370	12	2012-08-02 05:08:25.70064
13511	4	1-3860	13511	28	39	236	55	2012-08-02 05:08:25.70064
13516	3	1-3549	13515	31	89	4019	51	2012-08-02 05:08:25.70064
13517	3	1-2984	13515	89	28	2826	12	2012-08-02 05:08:25.70064
13515	3	1-3860	13515	28	31	1896	55	2012-08-02 05:08:25.70064
13518	1	1-1405	\N	33	33	3779	32	2012-08-02 05:08:25.70064
13520	2	1-669	13519	47	91	931	88	2012-08-02 05:08:27.711842
13519	2	1-3865	13519	91	47	1058	79	2012-08-02 05:08:27.711842
13522	5	1-3801	13521	97	27	6673	15	2012-08-02 05:08:27.711842
13523	5	1-3755	13521	27	12	4795	68	2012-08-02 05:08:27.711842
13524	5	1-1927	13521	12	74	4806	31	2012-08-02 05:08:27.711842
13525	5	1-3560	13521	74	91	6604	88	2012-08-02 05:08:27.711842
13521	5	1-3865	13521	91	97	7510	79	2012-08-02 05:08:27.711842
13527	5	1-3554	13526	77	8	4174	67	2012-08-02 05:08:28.882463
13528	5	1-3691	13526	8	82	4618	39	2012-08-02 05:08:28.882463
13529	5	1-3864	13526	82	60	2262	85	2012-08-02 05:08:28.882463
13530	5	1-2684	13526	60	51	1195	7	2012-08-02 05:08:28.882463
13526	5	1-3868	13526	51	77	3345	82	2012-08-02 05:08:28.882463
13532	4	1-1389	13531	2	75	934	46	2012-08-02 05:08:28.882463
13533	4	1-3733	13531	75	53	1604	79	2012-08-02 05:08:28.882463
13534	4	1-2946	13531	53	51	592	7	2012-08-02 05:08:28.882463
13531	4	1-3868	13531	51	2	1888	82	2012-08-02 05:08:28.882463
13536	5	1-541	13535	65	69	2436	52	2012-08-02 05:08:29.678128
13537	5	1-3805	13535	69	66	3281	53	2012-08-02 05:08:29.678128
13538	5	1-3690	13535	66	47	2080	69	2012-08-02 05:08:29.678128
13539	5	1-2725	13535	47	19	2227	57	2012-08-02 05:08:29.678128
13535	5	1-3870	13535	19	65	3230	91	2012-08-02 05:08:29.678128
13541	5	1-2244	13540	64	7	2156	73	2012-08-02 05:08:29.678128
13542	5	1-3692	13540	7	23	1349	67	2012-08-02 05:08:29.678128
13543	5	1-3591	13540	23	88	2063	90	2012-08-02 05:08:29.678128
13544	5	1-3287	13540	88	19	1360	57	2012-08-02 05:08:29.678128
13540	5	1-3870	13540	19	64	2008	91	2012-08-02 05:08:29.678128
13546	4	1-2855	13545	86	23	399	18	2012-08-02 05:08:29.678128
13547	4	1-3274	13545	23	69	579	58	2012-08-02 05:08:29.678128
13548	4	1-2668	13545	69	19	616	57	2012-08-02 05:08:29.678128
13545	4	1-3870	13545	19	86	915	91	2012-08-02 05:08:29.678128
13550	4	1-3675	13549	20	72	115	61	2012-08-02 05:08:30.098002
13551	4	1-3799	13549	72	48	392	11	2012-08-02 05:08:30.098002
13552	4	1-936	13549	48	71	138	95	2012-08-02 05:08:30.098002
13549	4	1-3871	13549	71	20	234	91	2012-08-02 05:08:30.098002
13554	5	1-2855	13553	86	16	478	18	2012-08-02 05:08:30.098002
13555	5	1-2682	13553	16	23	278	66	2012-08-02 05:08:30.098002
13556	5	1-3555	13553	23	3	472	60	2012-08-02 05:08:30.098002
13557	5	1-3460	13553	3	71	540	95	2012-08-02 05:08:30.098002
13553	5	1-3871	13553	71	86	1025	91	2012-08-02 05:08:30.098002
13559	5	1-541	13558	65	87	687	52	2012-08-02 05:08:30.098002
13560	5	1-2200	13558	87	12	778	59	2012-08-02 05:08:30.098002
13561	5	1-3516	13558	12	48	922	11	2012-08-02 05:08:30.098002
13562	5	1-936	13558	48	71	387	95	2012-08-02 05:08:30.098002
13558	5	1-3871	13558	71	65	782	91	2012-08-02 05:08:30.098002
16121	4	1-3522	16120	35	83	2930	5	2012-08-02 05:13:37.990737
13564	4	1-623	13563	3	74	1196	81	2012-08-02 05:08:30.098002
13565	4	1-3543	13563	74	25	820	26	2012-08-02 05:08:30.098002
13566	4	1-3009	13563	25	71	829	95	2012-08-02 05:08:30.098002
13563	4	1-3871	13563	71	3	1701	91	2012-08-02 05:08:30.098002
16122	4	1-4633	16120	83	23	2186	3	2012-08-02 05:13:37.990737
13568	4	1-3087	13567	2	68	1967	100	2012-08-02 05:08:30.518041
13569	4	1-3020	13567	68	26	2439	40	2012-08-02 05:08:30.518041
13570	4	1-3604	13567	26	67	2996	41	2012-08-02 05:08:30.518041
13567	4	1-3872	13567	67	2	1837	19	2012-08-02 05:08:30.518041
16123	4	1-4313	16120	23	35	3351	56	2012-08-02 05:13:37.990737
13572	4	1-3382	13571	25	12	860	98	2012-08-02 05:08:30.518041
13573	4	1-3116	13571	12	26	1013	40	2012-08-02 05:08:30.518041
13574	4	1-3604	13571	26	67	1274	41	2012-08-02 05:08:30.518041
13571	4	1-3872	13571	67	25	800	19	2012-08-02 05:08:30.518041
13576	5	1-3589	13575	74	76	1796	44	2012-08-02 05:08:30.518041
13577	5	1-3493	13575	76	69	2346	12	2012-08-02 05:08:30.518041
13578	5	1-1315	13575	69	26	2413	40	2012-08-02 05:08:30.518041
13579	5	1-3604	13575	26	67	3045	41	2012-08-02 05:08:30.518041
13575	5	1-3872	13575	67	74	1917	19	2012-08-02 05:08:30.518041
13581	5	1-3382	13580	25	2	555	98	2012-08-02 05:08:30.518041
13582	5	1-3632	13580	2	35	681	8	2012-08-02 05:08:30.518041
13583	5	1-43	13580	35	26	591	40	2012-08-02 05:08:30.518041
13584	5	1-3604	13580	26	67	769	41	2012-08-02 05:08:30.518041
13580	5	1-3872	13580	67	25	499	19	2012-08-02 05:08:30.518041
13586	5	1-3780	13585	39	21	3	29	2012-08-02 05:08:34.474999
13587	5	1-3443	13585	21	16	4	60	2012-08-02 05:08:34.474999
13588	5	1-2860	13585	16	64	2	26	2012-08-02 05:08:34.474999
13589	5	1-3328	13585	64	90	1	54	2012-08-02 05:08:34.474999
13585	5	1-3882	13585	90	39	1	20	2012-08-02 05:08:34.474999
13591	5	1-3780	13590	39	21	1423	29	2012-08-02 05:08:34.474999
13592	5	1-3443	13590	21	3	1677	60	2012-08-02 05:08:34.474999
13593	5	1-3460	13590	3	61	1628	95	2012-08-02 05:08:34.474999
13594	5	1-2319	13590	61	90	497	54	2012-08-02 05:08:34.474999
13590	5	1-3882	13590	90	39	636	20	2012-08-02 05:08:34.474999
13596	4	1-3833	13595	46	47	6521	28	2012-08-02 05:08:34.474999
13597	4	1-3687	13595	47	64	4468	26	2012-08-02 05:08:34.474999
13598	4	1-3328	13595	64	90	3410	54	2012-08-02 05:08:34.474999
13595	4	1-3882	13595	90	46	5051	20	2012-08-02 05:08:34.474999
13600	5	1-3744	13599	61	23	919	2	2012-08-02 05:08:34.474999
13601	5	1-1819	13599	23	74	156	81	2012-08-02 05:08:34.474999
13602	5	1-3543	13599	74	64	106	26	2012-08-02 05:08:34.474999
13603	5	1-3328	13599	64	90	89	54	2012-08-02 05:08:34.474999
13599	5	1-3882	13599	90	61	144	20	2012-08-02 05:08:34.474999
13605	5	1-3710	13604	11	61	782	35	2012-08-02 05:08:34.474999
13606	5	1-3791	13604	61	39	1512	38	2012-08-02 05:08:34.474999
13607	5	1-3021	13604	39	62	1986	37	2012-08-02 05:08:34.474999
13608	5	1-3617	13604	62	90	1039	54	2012-08-02 05:08:34.474999
13604	5	1-3882	13604	90	11	1718	20	2012-08-02 05:08:34.474999
13610	4	1-3833	13609	46	92	540	28	2012-08-02 05:08:34.474999
13611	4	1-1130	13609	92	62	421	37	2012-08-02 05:08:34.474999
13612	4	1-3617	13609	62	90	222	54	2012-08-02 05:08:34.474999
13609	4	1-3882	13609	90	46	371	20	2012-08-02 05:08:34.474999
13614	5	1-3780	13613	39	74	3150	29	2012-08-02 05:08:34.474999
13615	5	1-2808	13613	74	21	1177	66	2012-08-02 05:08:34.474999
13616	5	1-3467	13613	21	72	1490	38	2012-08-02 05:08:34.474999
13617	5	1-3298	13613	72	90	604	54	2012-08-02 05:08:34.474999
13613	5	1-3882	13613	90	39	1043	20	2012-08-02 05:08:34.474999
13619	5	1-3844	13618	87	94	3039	10	2012-08-02 05:08:35.271091
13620	5	1-3751	13618	94	21	1421	29	2012-08-02 05:08:35.271091
13621	5	1-3443	13618	21	72	2275	60	2012-08-02 05:08:35.271091
13622	5	1-3129	13618	72	76	2025	99	2012-08-02 05:08:35.271091
13618	5	1-3884	13618	76	87	1600	48	2012-08-02 05:08:35.271091
13624	4	1-2800	13623	65	100	1334	61	2012-08-02 05:08:36.486166
13625	4	1-2973	13623	100	69	981	58	2012-08-02 05:08:36.486166
13626	4	1-2668	13623	69	74	1172	57	2012-08-02 05:08:36.486166
13623	4	1-3887	13623	74	65	3054	72	2012-08-02 05:08:36.486166
13628	5	1-3613	13627	42	19	1164	9	2012-08-02 05:08:36.486166
13629	5	1-3315	13627	19	68	634	45	2012-08-02 05:08:36.486166
13630	5	1-3816	13627	68	88	765	90	2012-08-02 05:08:36.486166
13631	5	1-3287	13627	88	74	600	57	2012-08-02 05:08:36.486166
13627	5	1-3887	13627	74	42	1646	72	2012-08-02 05:08:36.486166
13633	5	1-3467	13632	21	24	1053	38	2012-08-02 05:08:37.712429
13634	5	1-2979	13632	24	69	248	89	2012-08-02 05:08:37.712429
13635	5	1-3752	13632	69	47	72	86	2012-08-02 05:08:37.712429
13636	5	1-1307	13632	47	84	12	24	2012-08-02 05:08:37.712429
13632	5	1-3890	13632	84	21	2496	66	2012-08-02 05:08:37.712429
13638	5	1-2105	13637	4	62	874	28	2012-08-02 05:08:37.712429
13639	5	1-2559	13637	62	25	108	19	2012-08-02 05:08:37.712429
13640	5	1-3382	13637	25	25	76	98	2012-08-02 05:08:37.712429
13641	5	1-1221	13637	25	84	15	24	2012-08-02 05:08:37.712429
13637	5	1-3890	13637	84	4	3363	66	2012-08-02 05:08:37.712429
13643	5	1-3645	13642	11	61	350	59	2012-08-02 05:08:38.541081
13644	5	1-3703	13642	61	46	333	15	2012-08-02 05:08:38.541081
13645	5	1-3712	13642	46	76	212	22	2012-08-02 05:08:38.541081
13646	5	1-2964	13642	76	69	29	56	2012-08-02 05:08:38.541081
13642	5	1-3892	13642	69	11	129	3	2012-08-02 05:08:38.541081
13648	4	1-1074	13647	61	53	746	4	2012-08-02 05:08:39.347613
13649	4	1-3636	13647	53	94	256	85	2012-08-02 05:08:39.347613
13650	4	1-3420	13647	94	34	328	43	2012-08-02 05:08:39.347613
13647	4	1-3894	13647	34	61	2504	86	2012-08-02 05:08:39.347613
13652	5	1-3780	13651	39	23	215	29	2012-08-02 05:08:39.734563
13653	5	1-3188	13651	23	36	396	63	2012-08-02 05:08:39.734563
13654	5	1-2851	13651	36	66	407	73	2012-08-02 05:08:39.734563
13655	5	1-1574	13651	66	54	71	24	2012-08-02 05:08:39.734563
13651	5	1-3895	13651	54	39	82	20	2012-08-02 05:08:39.734563
13657	5	1-3880	13656	74	37	3382	39	2012-08-02 05:08:40.52989
13658	5	1-3429	13656	37	76	1223	44	2012-08-02 05:08:40.52989
13659	5	1-3493	13656	76	28	1432	12	2012-08-02 05:08:40.52989
13660	5	1-3860	13656	28	86	803	55	2012-08-02 05:08:40.52989
13656	5	1-3897	13656	86	74	3066	26	2012-08-02 05:08:40.52989
13662	4	1-3880	13661	74	90	3148	39	2012-08-02 05:08:40.52989
13663	4	1-3281	13661	90	20	3605	68	2012-08-02 05:08:40.52989
13664	4	1-2515	13661	20	86	686	55	2012-08-02 05:08:40.52989
13661	4	1-3897	13661	86	74	2735	26	2012-08-02 05:08:40.52989
13666	5	1-1887	13665	60	86	63	56	2012-08-02 05:08:40.52989
13667	5	1-3230	13665	86	76	136	99	2012-08-02 05:08:40.52989
13668	5	1-3337	13665	76	56	95	47	2012-08-02 05:08:40.52989
13669	5	1-2634	13665	56	86	62	55	2012-08-02 05:08:40.52989
13665	5	1-3897	13665	86	60	255	26	2012-08-02 05:08:40.52989
16120	4	1-4639	16120	35	35	3475	9	2012-08-02 05:13:37.990737
13671	4	1-3880	13670	74	50	571	39	2012-08-02 05:08:40.52989
13672	4	1-2665	13670	50	56	149	47	2012-08-02 05:08:40.52989
13673	4	1-2634	13670	56	86	114	55	2012-08-02 05:08:40.52989
13670	4	1-3897	13670	86	74	475	26	2012-08-02 05:08:40.52989
13674	1	1-2634	\N	56	56	3373	55	2012-08-02 05:08:40.52989
13676	6	1-3819	13675	44	8	935	9	2012-08-02 05:08:40.950097
13677	6	1-3760	13675	8	49	1573	28	2012-08-02 05:08:40.950097
13678	6	1-3795	13675	49	28	4812	97	2012-08-02 05:08:40.950097
13679	6	1-3529	13675	28	40	818	96	2012-08-02 05:08:40.950097
13680	6	1-3635	13675	40	21	3429	27	2012-08-02 05:08:40.950097
13675	6	1-3898	13675	21	44	1330	53	2012-08-02 05:08:40.950097
13682	4	1-2758	13681	13	33	1021	67	2012-08-02 05:08:43.767476
13683	4	1-2383	13681	33	12	1235	59	2012-08-02 05:08:43.767476
13684	4	1-3516	13681	12	31	1448	11	2012-08-02 05:08:43.767476
13681	4	1-3905	13681	31	13	1402	33	2012-08-02 05:08:43.767476
13686	5	1-3780	13685	39	62	646	29	2012-08-02 05:08:44.165059
13687	5	1-3368	13685	62	61	545	59	2012-08-02 05:08:44.165059
13688	5	1-3703	13685	61	27	464	15	2012-08-02 05:08:44.165059
13689	5	1-3755	13685	27	94	305	68	2012-08-02 05:08:44.165059
13685	5	1-3906	13685	94	39	246	20	2012-08-02 05:08:44.165059
13691	5	1-3899	13690	6	36	1166	26	2012-08-02 05:08:44.165059
13692	5	1-999	13690	36	61	881	59	2012-08-02 05:08:44.165059
13693	5	1-3703	13690	61	27	844	15	2012-08-02 05:08:44.165059
13694	5	1-3755	13690	27	94	625	68	2012-08-02 05:08:44.165059
13690	5	1-3906	13690	94	6	567	20	2012-08-02 05:08:44.165059
13696	5	1-1665	13695	48	74	2101	70	2012-08-02 05:08:44.97138
13697	5	1-3901	13695	74	5	2160	60	2012-08-02 05:08:44.97138
13698	5	1-3213	13695	5	2	1906	69	2012-08-02 05:08:44.97138
13699	5	1-2043	13695	2	56	1263	100	2012-08-02 05:08:44.97138
13695	5	1-3908	13695	56	48	3932	86	2012-08-02 05:08:44.97138
13701	5	1-1665	13700	48	74	957	70	2012-08-02 05:08:44.97138
13702	5	1-3901	13700	74	71	1014	60	2012-08-02 05:08:44.97138
13703	5	1-3796	13700	71	2	791	69	2012-08-02 05:08:44.97138
13704	5	1-2043	13700	2	56	541	100	2012-08-02 05:08:44.97138
13700	5	1-3908	13700	56	48	1736	86	2012-08-02 05:08:44.97138
13706	4	1-2482	13705	63	71	1665	60	2012-08-02 05:08:44.97138
13707	4	1-3796	13705	71	2	1383	69	2012-08-02 05:08:44.97138
13708	4	1-2043	13705	2	56	1006	100	2012-08-02 05:08:44.97138
13705	4	1-3908	13705	56	63	3438	86	2012-08-02 05:08:44.97138
13710	5	1-1307	13709	47	9	365	24	2012-08-02 05:08:44.97138
13711	5	1-3850	13709	9	34	313	38	2012-08-02 05:08:44.97138
13712	5	1-3804	13709	34	90	295	6	2012-08-02 05:08:44.97138
13713	5	1-3869	13709	90	56	225	100	2012-08-02 05:08:44.97138
13709	5	1-3908	13709	56	47	793	86	2012-08-02 05:08:44.97138
13715	4	1-3776	13714	40	10	464	61	2012-08-02 05:08:46.164989
13716	4	1-3627	13714	10	49	303	28	2012-08-02 05:08:46.164989
13717	4	1-3795	13714	49	42	686	97	2012-08-02 05:08:46.164989
13714	4	1-3911	13714	42	40	429	45	2012-08-02 05:08:46.164989
13719	4	1-3776	13718	40	89	5297	61	2012-08-02 05:08:46.164989
13720	4	1-1924	13718	89	87	1970	55	2012-08-02 05:08:46.164989
13721	4	1-3859	13718	87	42	4053	97	2012-08-02 05:08:46.164989
13718	4	1-3911	13718	42	40	3523	45	2012-08-02 05:08:46.164989
13723	4	1-3550	13722	39	69	289	52	2012-08-02 05:08:46.562602
13724	4	1-3805	13722	69	49	447	53	2012-08-02 05:08:46.562602
13725	4	1-3904	13722	49	44	333	63	2012-08-02 05:08:46.562602
13722	4	1-3912	13722	44	39	270	55	2012-08-02 05:08:46.562602
13727	4	1-3550	13726	39	10	647	52	2012-08-02 05:08:46.562602
13728	4	1-3376	13726	10	49	831	53	2012-08-02 05:08:46.562602
13729	4	1-3904	13726	49	44	659	63	2012-08-02 05:08:46.562602
13726	4	1-3912	13726	44	39	569	55	2012-08-02 05:08:46.562602
13731	4	1-3404	13730	40	55	365	70	2012-08-02 05:08:48.529158
13732	4	1-3346	13730	55	42	204	97	2012-08-02 05:08:48.529158
13733	4	1-3911	13730	42	59	141	45	2012-08-02 05:08:48.529158
13730	4	1-3917	13730	59	40	691	90	2012-08-02 05:08:48.529158
13735	4	1-3899	13734	6	16	5997	26	2012-08-02 05:08:50.486898
13736	4	1-16	13734	16	81	4588	25	2012-08-02 05:08:50.486898
13737	4	1-3459	13734	81	68	4437	16	2012-08-02 05:08:50.486898
13734	4	1-3922	13734	68	6	3179	20	2012-08-02 05:08:50.486898
13738	1	1-16	\N	16	16	893	25	2012-08-02 05:08:50.486898
13740	2	1-3804	13739	34	82	4430	6	2012-08-02 05:08:50.900001
13739	2	1-3923	13739	82	34	4983	38	2012-08-02 05:08:50.900001
13742	5	1-3842	13741	45	8	1146	65	2012-08-02 05:08:52.105147
13743	5	1-3221	13741	8	75	887	46	2012-08-02 05:08:52.105147
13744	5	1-3733	13741	75	57	1613	79	2012-08-02 05:08:52.105147
13745	5	1-1680	13741	57	99	1492	10	2012-08-02 05:08:52.105147
13741	5	1-3926	13741	99	45	1425	68	2012-08-02 05:08:52.105147
13747	4	1-3848	13746	21	46	383	22	2012-08-02 05:08:52.504551
13748	4	1-3151	13746	46	9	268	79	2012-08-02 05:08:52.504551
13749	4	1-3248	13746	9	24	250	2	2012-08-02 05:08:52.504551
13746	4	1-3927	13746	24	21	372	54	2012-08-02 05:08:52.504551
13751	5	1-3848	13750	21	65	192	22	2012-08-02 05:08:52.504551
13752	5	1-2522	13750	65	8	41	35	2012-08-02 05:08:52.504551
13753	5	1-3065	13750	8	70	15	47	2012-08-02 05:08:52.504551
13754	5	1-3913	13750	70	24	106	2	2012-08-02 05:08:52.504551
13750	5	1-3927	13750	24	21	171	54	2012-08-02 05:08:52.504551
13756	4	1-3921	13755	15	27	2080	81	2012-08-02 05:08:54.10865
13757	4	1-3582	13755	27	70	1212	47	2012-08-02 05:08:54.10865
13758	4	1-3913	13755	70	98	7536	2	2012-08-02 05:08:54.10865
13755	4	1-3931	13755	98	15	2054	91	2012-08-02 05:08:54.10865
13760	4	1-3862	13759	75	60	2181	97	2012-08-02 05:08:54.905048
13761	4	1-3095	13759	60	24	1691	2	2012-08-02 05:08:54.905048
13762	4	1-3927	13759	24	25	2681	54	2012-08-02 05:08:54.905048
13759	4	1-3933	13759	25	75	2399	34	2012-08-02 05:08:54.905048
13764	4	1-1997	13763	65	9	95	79	2012-08-02 05:08:54.905048
13765	4	1-3248	13763	9	24	96	2	2012-08-02 05:08:54.905048
13766	4	1-3927	13763	24	25	154	54	2012-08-02 05:08:54.905048
13763	4	1-3933	13763	25	65	140	34	2012-08-02 05:08:54.905048
13768	5	1-3880	13767	74	69	1215	39	2012-08-02 05:08:55.303548
13767	5	1-3934	13767	20	74	954	26	2012-08-02 05:08:55.303548
13769	5	1-1891	13767	69	17	1920	40	2012-08-02 05:08:55.303548
13770	5	1-3722	13767	17	43	1460	36	2012-08-02 05:08:55.303548
13771	5	1-2892	13767	43	20	1593	63	2012-08-02 05:08:55.303548
16125	5	1-4371	16124	51	47	3860	40	2012-08-02 05:13:39.62595
13773	5	1-2506	13772	11	5	347	64	2012-08-02 05:08:55.701865
13774	5	1-3548	13772	5	86	344	34	2012-08-02 05:08:55.701865
13775	5	1-3763	13772	86	60	612	1	2012-08-02 05:08:55.701865
13776	5	1-2807	13772	60	81	179	86	2012-08-02 05:08:55.701865
13772	5	1-3935	13772	81	11	1230	69	2012-08-02 05:08:55.701865
16126	5	1-4619	16124	47	55	3926	58	2012-08-02 05:13:39.62595
13778	6	1-3443	13777	21	53	2849	60	2012-08-02 05:08:56.111726
13779	6	1-2785	13777	53	33	1939	7	2012-08-02 05:08:56.111726
13780	6	1-1866	13777	33	53	514	71	2012-08-02 05:08:56.111726
13781	6	1-3397	13777	53	40	318	96	2012-08-02 05:08:56.111726
13782	6	1-3635	13777	40	82	970	27	2012-08-02 05:08:56.111726
13777	6	1-3936	13777	82	21	2579	29	2012-08-02 05:08:56.111726
16127	5	1-2101	16124	55	91	940	65	2012-08-02 05:13:39.62595
13784	5	1-3789	13783	66	13	5315	14	2012-08-02 05:08:57.308036
13785	5	1-3909	13783	13	13	2525	33	2012-08-02 05:08:57.308036
13786	5	1-2758	13783	13	69	1520	67	2012-08-02 05:08:57.308036
13787	5	1-3694	13783	69	41	1330	3	2012-08-02 05:08:57.308036
13783	5	1-3939	13783	41	66	2280	60	2012-08-02 05:08:57.308036
13789	5	1-3163	13788	4	6	3477	42	2012-08-02 05:08:57.308036
13790	5	1-3792	13788	6	76	4010	22	2012-08-02 05:08:57.308036
13791	5	1-2964	13788	76	69	461	56	2012-08-02 05:08:57.308036
13792	5	1-3892	13788	69	41	1727	3	2012-08-02 05:08:57.308036
13788	5	1-3939	13788	41	4	4263	60	2012-08-02 05:08:57.308036
13794	5	1-305	13793	54	25	721	47	2012-08-02 05:08:57.308036
13795	5	1-3187	13793	25	60	604	26	2012-08-02 05:08:57.308036
13796	5	1-1887	13793	60	69	170	56	2012-08-02 05:08:57.308036
13797	5	1-3892	13793	69	41	684	3	2012-08-02 05:08:57.308036
13793	5	1-3939	13793	41	54	1811	60	2012-08-02 05:08:57.308036
13799	4	1-3899	13798	6	60	914	26	2012-08-02 05:08:58.491302
13800	4	1-1887	13798	60	98	246	56	2012-08-02 05:08:58.491302
13801	4	1-2146	13798	98	90	222	23	2012-08-02 05:08:58.491302
13798	4	1-3942	13798	90	6	518	20	2012-08-02 05:08:58.491302
13803	4	1-3899	13802	6	25	1276	26	2012-08-02 05:08:58.491302
13804	4	1-3009	13802	25	85	1181	95	2012-08-02 05:08:58.491302
13805	4	1-1463	13802	85	90	294	23	2012-08-02 05:08:58.491302
13802	4	1-3942	13802	90	6	704	20	2012-08-02 05:08:58.491302
13807	5	1-3826	13806	43	37	2796	49	2012-08-02 05:08:58.491302
13808	5	1-3658	13806	37	100	1881	71	2012-08-02 05:08:58.491302
13809	5	1-2802	13806	100	36	801	99	2012-08-02 05:08:58.491302
13810	5	1-3837	13806	36	90	979	23	2012-08-02 05:08:58.491302
13806	5	1-3942	13806	90	43	2419	20	2012-08-02 05:08:58.491302
13812	5	1-3885	13811	87	42	335	72	2012-08-02 05:08:59.298633
13813	5	1-3613	13811	42	93	225	9	2012-08-02 05:08:59.298633
13814	5	1-3465	13811	93	43	193	43	2012-08-02 05:08:59.298633
13815	5	1-3564	13811	43	92	43	62	2012-08-02 05:08:59.298633
13811	5	1-3944	13811	92	87	332	64	2012-08-02 05:08:59.298633
13817	4	1-3766	13816	61	94	621	85	2012-08-02 05:08:59.298633
13818	4	1-3420	13816	94	43	1382	43	2012-08-02 05:08:59.298633
13819	4	1-3564	13816	43	92	317	62	2012-08-02 05:08:59.298633
13816	4	1-3944	13816	92	61	2476	64	2012-08-02 05:08:59.298633
13821	4	1-3445	13820	60	24	1011	2	2012-08-02 05:09:00.514587
13822	4	1-3927	13820	24	68	1688	54	2012-08-02 05:09:00.514587
13823	4	1-3839	13820	68	72	3407	78	2012-08-02 05:09:00.514587
13820	4	1-3947	13820	72	60	1537	20	2012-08-02 05:09:00.514587
13825	4	1-3432	13824	33	90	116	23	2012-08-02 05:09:00.9126
13826	4	1-3942	13824	90	70	226	20	2012-08-02 05:09:00.9126
13827	4	1-2616	13824	70	97	81	26	2012-08-02 05:09:00.9126
13824	4	1-3948	13824	97	33	136	98	2012-08-02 05:09:00.9126
13829	5	1-3601	13828	65	42	811	55	2012-08-02 05:09:00.9126
13830	5	1-3929	13828	42	90	538	23	2012-08-02 05:09:00.9126
13831	5	1-3942	13828	90	70	1185	20	2012-08-02 05:09:00.9126
13832	5	1-2616	13828	70	97	481	26	2012-08-02 05:09:00.9126
13828	5	1-3948	13828	97	65	916	98	2012-08-02 05:09:00.9126
13834	4	1-3632	13833	2	19	2268	8	2012-08-02 05:09:00.9126
13835	4	1-3210	13833	19	74	1621	81	2012-08-02 05:09:00.9126
13836	4	1-3543	13833	74	97	1005	26	2012-08-02 05:09:00.9126
13833	4	1-3948	13833	97	2	2060	98	2012-08-02 05:09:00.9126
13838	5	1-3632	13837	2	19	27	8	2012-08-02 05:09:00.9126
13839	5	1-3210	13837	19	27	20	81	2012-08-02 05:09:00.9126
13840	5	1-3582	13837	27	25	12	47	2012-08-02 05:09:00.9126
13841	5	1-3187	13837	25	97	11	26	2012-08-02 05:09:00.9126
13837	5	1-3948	13837	97	2	23	98	2012-08-02 05:09:00.9126
13843	3	1-3837	13842	36	11	1334	23	2012-08-02 05:09:02.912851
13844	3	1-3746	13842	11	59	1414	80	2012-08-02 05:09:02.912851
13842	3	1-3953	13842	59	36	1124	99	2012-08-02 05:09:02.912851
13846	5	1-3272	13845	49	88	1256	21	2012-08-02 05:09:02.912851
13847	5	1-3940	13845	88	17	3946	40	2012-08-02 05:09:02.912851
13848	5	1-3722	13845	17	6	2924	36	2012-08-02 05:09:02.912851
13849	5	1-2226	13845	6	59	1978	80	2012-08-02 05:09:02.912851
13845	5	1-3953	13845	59	49	1693	99	2012-08-02 05:09:02.912851
13851	4	1-305	13850	54	25	1062	47	2012-08-02 05:09:03.310996
13852	4	1-3187	13850	25	25	698	26	2012-08-02 05:09:03.310996
13853	4	1-3009	13850	25	13	516	95	2012-08-02 05:09:03.310996
13850	4	1-3954	13850	13	54	4997	60	2012-08-02 05:09:03.310996
13855	5	1-933	13854	83	24	78	2	2012-08-02 05:09:03.310996
13856	5	1-3927	13854	24	68	72	54	2012-08-02 05:09:03.310996
13857	5	1-3839	13854	68	4	81	78	2012-08-02 05:09:03.310996
13858	5	1-2814	13854	4	13	13	95	2012-08-02 05:09:03.310996
13854	5	1-3954	13854	13	83	148	60	2012-08-02 05:09:03.310996
13860	4	1-1320	13859	69	66	959	32	2012-08-02 05:09:03.310996
13861	4	1-2908	13859	66	63	105	86	2012-08-02 05:09:03.310996
13862	4	1-3820	13859	63	13	130	95	2012-08-02 05:09:03.310996
13859	4	1-3954	13859	13	69	1545	60	2012-08-02 05:09:03.310996
13864	4	1-3448	13863	49	6	731	23	2012-08-02 05:09:04.50511
13865	4	1-3856	13863	6	24	952	2	2012-08-02 05:09:04.50511
13866	4	1-3927	13863	24	32	853	54	2012-08-02 05:09:04.50511
13863	4	1-3957	13863	32	49	2232	45	2012-08-02 05:09:04.50511
13868	5	1-3917	13867	59	42	1582	90	2012-08-02 05:09:04.50511
13869	5	1-3886	13867	42	92	386	28	2012-08-02 05:09:04.50511
13870	5	1-1130	13867	92	62	230	37	2012-08-02 05:09:04.50511
13871	5	1-3617	13867	62	32	93	54	2012-08-02 05:09:04.50511
13867	5	1-3957	13867	32	59	340	45	2012-08-02 05:09:04.50511
13873	5	1-3917	13872	59	24	2170	90	2012-08-02 05:09:04.50511
13874	5	1-3566	13872	24	88	1601	88	2012-08-02 05:09:04.50511
13875	5	1-3348	13872	88	62	305	37	2012-08-02 05:09:04.50511
13876	5	1-3617	13872	62	32	125	54	2012-08-02 05:09:04.50511
13872	5	1-3957	13872	32	59	461	45	2012-08-02 05:09:04.50511
13878	3	1-3797	13877	31	61	156	95	2012-08-02 05:09:04.50511
13879	3	1-2319	13877	61	32	54	54	2012-08-02 05:09:04.50511
13877	3	1-3957	13877	32	31	225	45	2012-08-02 05:09:04.50511
13881	4	1-3707	13880	58	69	1066	89	2012-08-02 05:09:06.103116
13882	4	1-3752	13880	69	81	543	86	2012-08-02 05:09:06.103116
13883	4	1-3935	13880	81	15	2625	69	2012-08-02 05:09:06.103116
13880	4	1-3961	13880	15	58	3867	41	2012-08-02 05:09:06.103116
13885	5	1-3784	13884	75	94	931	10	2012-08-02 05:09:06.103116
13886	5	1-3751	13884	94	21	394	29	2012-08-02 05:09:06.103116
13887	5	1-3443	13884	21	71	571	60	2012-08-02 05:09:06.103116
13888	5	1-3796	13884	71	15	474	69	2012-08-02 05:09:06.103116
13884	5	1-3961	13884	15	75	1083	41	2012-08-02 05:09:06.103116
13890	5	1-2319	13889	61	68	619	54	2012-08-02 05:09:06.501418
13891	5	1-3839	13889	68	90	820	78	2012-08-02 05:09:06.501418
13892	5	1-3678	13889	90	93	267	71	2012-08-02 05:09:06.501418
13893	5	1-3914	13889	93	24	199	38	2012-08-02 05:09:06.501418
13889	5	1-3962	13889	24	61	2321	95	2012-08-02 05:09:06.501418
13895	5	1-3207	13894	17	67	2297	41	2012-08-02 05:09:07.285542
13896	5	1-3872	13894	67	25	1229	19	2012-08-02 05:09:07.285542
13897	5	1-3421	13894	25	59	846	30	2012-08-02 05:09:07.285542
13898	5	1-2485	13894	59	20	836	73	2012-08-02 05:09:07.285542
13894	5	1-3964	13894	20	17	2219	75	2012-08-02 05:09:07.285542
13900	5	1-3785	13899	88	21	4190	27	2012-08-02 05:09:07.285542
13901	5	1-3898	13899	21	49	1349	53	2012-08-02 05:09:07.285542
13902	5	1-3904	13899	49	36	869	63	2012-08-02 05:09:07.285542
13903	5	1-2851	13899	36	20	819	73	2012-08-02 05:09:07.285542
13899	5	1-3964	13899	20	88	2191	75	2012-08-02 05:09:07.285542
13905	5	1-3668	13904	93	99	443	10	2012-08-02 05:09:07.285542
13906	5	1-3926	13904	99	94	346	68	2012-08-02 05:09:07.285542
13907	5	1-3906	13904	94	6	257	20	2012-08-02 05:09:07.285542
13908	5	1-2619	13904	6	20	135	73	2012-08-02 05:09:07.285542
13904	5	1-3964	13904	20	93	365	75	2012-08-02 05:09:07.285542
13910	5	1-3785	13909	88	20	2975	27	2012-08-02 05:09:07.285542
13911	5	1-2492	13909	20	85	777	94	2012-08-02 05:09:07.285542
13912	5	1-3867	13909	85	87	1164	50	2012-08-02 05:09:07.285542
13913	5	1-3389	13909	87	20	612	73	2012-08-02 05:09:07.285542
13909	5	1-3964	13909	20	88	1749	75	2012-08-02 05:09:07.285542
13915	5	1-3207	13914	17	58	1245	41	2012-08-02 05:09:07.285542
13916	5	1-3707	13914	58	31	508	89	2012-08-02 05:09:07.285542
13917	5	1-3738	13914	31	87	569	50	2012-08-02 05:09:07.285542
13918	5	1-3389	13914	87	20	370	73	2012-08-02 05:09:07.285542
13914	5	1-3964	13914	20	17	1086	75	2012-08-02 05:09:07.285542
13920	5	1-3914	13919	93	24	6155	38	2012-08-02 05:09:08.512433
13921	5	1-2979	13919	24	69	3295	89	2012-08-02 05:09:08.512433
13922	5	1-3752	13919	69	63	2162	86	2012-08-02 05:09:08.512433
13923	5	1-2482	13919	63	91	868	60	2012-08-02 05:09:08.512433
13919	5	1-3967	13919	91	93	7061	71	2012-08-02 05:09:08.512433
13925	5	1-3953	13924	59	13	3695	99	2012-08-02 05:09:08.91028
13926	5	1-820	13924	13	96	1849	13	2012-08-02 05:09:08.91028
13927	5	1-3179	13924	96	85	1582	94	2012-08-02 05:09:08.91028
13928	5	1-3867	13924	85	25	3234	50	2012-08-02 05:09:08.91028
13924	5	1-3968	13924	25	59	4194	80	2012-08-02 05:09:08.91028
13930	5	1-3642	13929	5	4	5676	80	2012-08-02 05:09:09.727831
13931	5	1-3919	13929	4	59	2859	29	2012-08-02 05:09:09.727831
13932	5	1-3052	13929	59	31	2507	67	2012-08-02 05:09:09.727831
13933	5	1-3478	13929	31	48	1280	9	2012-08-02 05:09:09.727831
13929	5	1-3970	13929	48	5	9665	44	2012-08-02 05:09:09.727831
13935	5	1-3389	13934	87	7	156	73	2012-08-02 05:09:10.910397
13936	5	1-3692	13934	7	33	115	67	2012-08-02 05:09:10.910397
13937	5	1-2383	13934	33	61	146	59	2012-08-02 05:09:10.910397
13938	5	1-3703	13934	61	100	137	15	2012-08-02 05:09:10.910397
13934	5	1-3973	13934	100	87	217	50	2012-08-02 05:09:10.910397
13940	5	1-3968	13939	25	4	2271	80	2012-08-02 05:09:10.910397
13941	5	1-3919	13939	4	62	1206	29	2012-08-02 05:09:10.910397
13942	5	1-3368	13939	62	61	1137	59	2012-08-02 05:09:10.910397
13943	5	1-3703	13939	61	100	1079	15	2012-08-02 05:09:10.910397
13939	5	1-3973	13939	100	25	1732	50	2012-08-02 05:09:10.910397
13945	2	1-3940	13944	88	41	3732	40	2012-08-02 05:09:12.103257
13944	2	1-3976	13944	41	88	4713	21	2012-08-02 05:09:12.103257
13947	5	1-3211	13946	14	98	670	2	2012-08-02 05:09:14.467415
13948	5	1-3931	13946	98	3	203	91	2012-08-02 05:09:14.467415
13949	5	1-3600	13946	3	25	108	19	2012-08-02 05:09:14.467415
13950	5	1-3421	13946	25	16	91	30	2012-08-02 05:09:14.467415
13946	5	1-3982	13946	16	14	152	75	2012-08-02 05:09:14.467415
13952	2	1-3892	13951	69	19	2331	3	2012-08-02 05:09:14.865564
13951	2	1-3983	13951	19	69	754	56	2012-08-02 05:09:14.865564
13954	2	1-1154	13953	79	95	6428	14	2012-08-02 05:09:15.66109
13953	2	1-3985	13953	95	79	5066	38	2012-08-02 05:09:15.66109
13956	4	1-3550	13955	39	87	569	52	2012-08-02 05:09:16.44549
13957	4	1-2200	13955	87	12	617	59	2012-08-02 05:09:16.44549
13958	4	1-3516	13955	12	74	701	11	2012-08-02 05:09:16.44549
13955	4	1-3987	13955	74	39	547	55	2012-08-02 05:09:16.44549
13960	4	1-3692	13959	7	33	621	67	2012-08-02 05:09:16.832209
13961	4	1-2383	13959	33	61	740	59	2012-08-02 05:09:16.832209
13962	4	1-3703	13959	61	48	656	15	2012-08-02 05:09:16.832209
13959	4	1-3988	13959	48	7	889	73	2012-08-02 05:09:16.832209
13964	3	1-2782	13963	94	68	1097	92	2012-08-02 05:09:17.2301
13965	3	1-2720	13963	68	37	824	93	2012-08-02 05:09:17.2301
13963	3	1-3989	13963	37	94	3439	24	2012-08-02 05:09:17.2301
13967	5	1-3850	13966	9	39	1791	38	2012-08-02 05:09:17.2301
13968	5	1-3021	13966	39	99	1593	37	2012-08-02 05:09:17.2301
13969	5	1-3303	13966	99	84	830	48	2012-08-02 05:09:17.2301
13970	5	1-3625	13966	84	37	603	93	2012-08-02 05:09:17.2301
13966	5	1-3989	13966	37	9	3029	24	2012-08-02 05:09:17.2301
13972	4	1-3920	13971	46	74	3794	91	2012-08-02 05:09:17.627959
13973	4	1-2786	13971	74	94	1975	68	2012-08-02 05:09:17.627959
13974	4	1-3906	13971	94	60	959	20	2012-08-02 05:09:17.627959
13971	4	1-3990	13971	60	46	7980	66	2012-08-02 05:09:17.627959
13976	6	1-3710	13975	11	12	653	35	2012-08-02 05:09:18.821021
13975	6	1-3993	13975	59	11	1664	20	2012-08-02 05:09:18.821021
13977	6	1-2275	13975	12	90	890	78	2012-08-02 05:09:18.821021
13978	6	1-3678	13975	90	53	370	71	2012-08-02 05:09:18.821021
13979	6	1-3397	13975	53	40	225	96	2012-08-02 05:09:18.821021
13980	6	1-3635	13975	40	59	818	27	2012-08-02 05:09:18.821021
16128	5	1-4433	16124	91	11	553	49	2012-08-02 05:13:39.62595
13982	4	1-2616	13981	70	17	2733	26	2012-08-02 05:09:18.821021
13983	4	1-211	13981	17	99	713	86	2012-08-02 05:09:18.821021
13984	4	1-3992	13981	99	59	2609	27	2012-08-02 05:09:18.821021
13981	4	1-3993	13981	59	70	5857	20	2012-08-02 05:09:18.821021
16124	5	1-4643	16124	11	51	1296	68	2012-08-02 05:13:39.62595
13986	5	1-2851	13985	36	87	2942	73	2012-08-02 05:09:19.249594
13987	5	1-3852	13985	87	65	1017	8	2012-08-02 05:09:19.249594
13988	5	1-3233	13985	65	100	748	37	2012-08-02 05:09:19.249594
13989	5	1-3830	13985	100	54	854	58	2012-08-02 05:09:19.249594
13985	5	1-3994	13985	54	36	3048	63	2012-08-02 05:09:19.249594
13991	4	1-3934	13990	20	97	2936	26	2012-08-02 05:09:19.249594
13992	4	1-3788	13990	97	23	924	18	2012-08-02 05:09:19.249594
13993	4	1-3274	13990	23	54	1422	58	2012-08-02 05:09:19.249594
13990	4	1-3994	13990	54	20	5425	63	2012-08-02 05:09:19.249594
13995	4	1-2851	13994	36	11	1263	73	2012-08-02 05:09:19.249594
13996	4	1-3812	13994	11	71	679	33	2012-08-02 05:09:19.249594
13997	4	1-2741	13994	71	54	294	58	2012-08-02 05:09:19.249594
13994	4	1-3994	13994	54	36	1171	63	2012-08-02 05:09:19.249594
13999	5	1-3912	13998	44	47	154	55	2012-08-02 05:09:19.249594
14000	5	1-1912	13998	47	88	163	88	2012-08-02 05:09:19.249594
14001	5	1-3348	13998	88	100	38	37	2012-08-02 05:09:19.249594
14002	5	1-3830	13998	100	54	48	58	2012-08-02 05:09:19.249594
13998	5	1-3994	13998	54	44	192	63	2012-08-02 05:09:19.249594
14004	5	1-3564	14003	43	32	269	62	2012-08-02 05:09:20.854661
14005	5	1-3814	14003	32	71	894	60	2012-08-02 05:09:20.854661
14006	5	1-3796	14003	71	69	794	69	2012-08-02 05:09:20.854661
14007	5	1-2875	14003	69	2	357	9	2012-08-02 05:09:20.854661
14003	5	1-3998	14003	2	43	1127	43	2012-08-02 05:09:20.854661
14009	4	1-1037	14008	85	87	165	48	2012-08-02 05:09:24.038235
14010	4	1-3844	14008	87	99	311	10	2012-08-02 05:09:24.038235
14011	4	1-3926	14008	99	99	306	68	2012-08-02 05:09:24.038235
14008	4	1-4006	14008	99	85	260	21	2012-08-02 05:09:24.038235
14013	5	1-2808	14012	74	53	673	66	2012-08-02 05:09:24.457976
14014	5	1-1213	14012	53	49	1217	53	2012-08-02 05:09:24.457976
14015	5	1-3904	14012	49	44	948	63	2012-08-02 05:09:24.457976
14016	5	1-3912	14012	44	56	804	55	2012-08-02 05:09:24.457976
14012	5	1-4007	14012	56	74	1884	29	2012-08-02 05:09:24.457976
14018	5	1-3368	14017	62	61	1966	59	2012-08-02 05:09:24.457976
14019	5	1-3703	14017	61	48	1871	15	2012-08-02 05:09:24.457976
14020	5	1-3988	14017	48	66	2721	73	2012-08-02 05:09:24.457976
14021	5	1-706	14017	66	56	874	55	2012-08-02 05:09:24.457976
14017	5	1-4007	14017	56	62	2081	29	2012-08-02 05:09:24.457976
14023	4	1-3982	14022	16	39	2005	75	2012-08-02 05:09:24.844702
14024	4	1-1952	14022	39	21	523	54	2012-08-02 05:09:24.844702
14025	4	1-3848	14022	21	56	364	22	2012-08-02 05:09:24.844702
14022	4	1-4008	14022	56	16	1923	30	2012-08-02 05:09:24.844702
14027	4	1-1277	14026	95	61	635	59	2012-08-02 05:09:24.844702
14028	4	1-3703	14026	61	46	373	15	2012-08-02 05:09:24.844702
14029	4	1-3712	14026	46	56	146	22	2012-08-02 05:09:24.844702
14026	4	1-4008	14026	56	95	782	30	2012-08-02 05:09:24.844702
14031	5	1-2485	14030	59	7	2482	73	2012-08-02 05:09:24.844702
14032	5	1-3692	14030	7	44	1410	67	2012-08-02 05:09:24.844702
14033	5	1-2494	14030	44	46	845	15	2012-08-02 05:09:24.844702
14034	5	1-3712	14030	46	56	408	22	2012-08-02 05:09:24.844702
14030	5	1-4008	14030	56	59	2677	30	2012-08-02 05:09:24.844702
14036	4	1-610	14035	49	89	1170	14	2012-08-02 05:09:24.844702
14037	4	1-3253	14035	89	21	134	54	2012-08-02 05:09:24.844702
14038	4	1-3848	14035	21	56	128	22	2012-08-02 05:09:24.844702
14035	4	1-4008	14035	56	49	931	30	2012-08-02 05:09:24.844702
14040	4	1-3749	14039	90	61	1479	59	2012-08-02 05:09:27.021231
14041	4	1-3703	14039	61	96	1421	15	2012-08-02 05:09:27.021231
14042	4	1-2784	14039	96	89	832	2	2012-08-02 05:09:27.021231
14039	4	1-4013	14039	89	90	1321	22	2012-08-02 05:09:27.021231
14044	5	1-3979	14043	30	30	4033	28	2012-08-02 05:09:27.021231
14045	5	1-3688	14043	30	79	6982	13	2012-08-02 05:09:27.021231
14046	5	1-3695	14043	79	6	1401	23	2012-08-02 05:09:27.021231
14047	5	1-3856	14043	6	89	3469	2	2012-08-02 05:09:27.021231
14043	5	1-4013	14043	89	30	5640	22	2012-08-02 05:09:27.021231
14049	3	1-3923	14048	82	72	4381	38	2012-08-02 05:09:27.817116
14050	3	1-2938	14048	72	56	4607	42	2012-08-02 05:09:27.817116
14048	3	1-4015	14048	56	82	4645	6	2012-08-02 05:09:27.817116
14052	4	1-1320	14051	69	42	358	32	2012-08-02 05:09:28.225583
14053	4	1-3881	14051	42	46	136	66	2012-08-02 05:09:28.225583
14054	4	1-3720	14051	46	9	231	22	2012-08-02 05:09:28.225583
14051	4	1-4016	14051	9	69	389	60	2012-08-02 05:09:28.225583
14056	5	1-3163	14055	4	74	358	42	2012-08-02 05:09:28.225583
14057	5	1-2904	14055	74	42	327	32	2012-08-02 05:09:28.225583
14058	5	1-3881	14055	42	46	129	66	2012-08-02 05:09:28.225583
14059	5	1-3720	14055	46	9	230	22	2012-08-02 05:09:28.225583
14055	5	1-4016	14055	9	4	404	60	2012-08-02 05:09:28.225583
14061	4	1-3789	14060	66	89	392	14	2012-08-02 05:09:28.225583
14062	4	1-3253	14060	89	21	51	54	2012-08-02 05:09:28.225583
14063	4	1-3848	14060	21	9	56	22	2012-08-02 05:09:28.225583
14060	4	1-4016	14060	9	66	103	60	2012-08-02 05:09:28.225583
14065	5	1-2004	14064	84	76	1814	44	2012-08-02 05:09:28.225583
14066	5	1-3493	14064	76	3	2463	12	2012-08-02 05:09:28.225583
14067	5	1-4014	14064	3	17	1468	56	2012-08-02 05:09:28.225583
14068	5	1-1328	14064	17	9	1330	22	2012-08-02 05:09:28.225583
14064	5	1-4016	14064	9	84	2526	60	2012-08-02 05:09:28.225583
14070	4	1-3658	14069	37	93	1078	71	2012-08-02 05:09:30.203365
14071	4	1-3914	14069	93	86	952	38	2012-08-02 05:09:30.203365
14072	4	1-1821	14069	86	89	734	70	2012-08-02 05:09:30.203365
14069	4	1-4021	14069	89	37	1865	49	2012-08-02 05:09:30.203365
14074	5	1-3840	14073	95	13	1437	33	2012-08-02 05:09:30.203365
14075	5	1-2758	14073	13	23	892	67	2012-08-02 05:09:30.203365
14076	5	1-3591	14073	23	40	1317	90	2012-08-02 05:09:30.203365
14077	5	1-3404	14073	40	89	700	70	2012-08-02 05:09:30.203365
14073	5	1-4021	14073	89	95	1822	49	2012-08-02 05:09:30.203365
14078	1	1-2758	\N	13	13	338	67	2012-08-02 05:09:30.203365
14080	5	1-2908	14079	66	63	427	86	2012-08-02 05:09:32.202942
14079	5	1-4026	14079	70	66	2775	32	2012-08-02 05:09:32.202942
14081	5	1-3820	14079	63	9	748	95	2012-08-02 05:09:32.202942
14082	5	1-3082	14079	9	46	271	92	2012-08-02 05:09:32.202942
14083	5	1-3849	14079	46	70	369	82	2012-08-02 05:09:32.202942
14085	4	1-3542	14084	12	64	502	96	2012-08-02 05:09:32.202942
14086	4	1-3449	14084	64	46	232	92	2012-08-02 05:09:32.202942
14087	4	1-3849	14084	46	70	338	82	2012-08-02 05:09:32.202942
14084	4	1-4026	14084	70	12	2713	32	2012-08-02 05:09:32.202942
16130	3	1-3842	16129	45	91	948	65	2012-08-02 05:13:39.62595
14089	4	1-3545	14088	34	7	1254	2	2012-08-02 05:09:35.009584
14090	4	1-3699	14088	7	29	1075	76	2012-08-02 05:09:35.009584
14091	4	1-1712	14088	29	21	963	53	2012-08-02 05:09:35.009584
14088	4	1-4033	14088	21	34	887	21	2012-08-02 05:09:35.009584
14093	3	1-2708	14092	84	44	28	35	2012-08-02 05:09:36.191363
14094	3	1-3333	14092	44	24	10	100	2012-08-02 05:09:36.191363
14092	3	1-4036	14092	24	84	144	7	2012-08-02 05:09:36.191363
14096	5	1-3878	14095	24	55	4101	33	2012-08-02 05:09:36.191363
14097	5	1-3034	14095	55	62	2380	28	2012-08-02 05:09:36.191363
14098	5	1-2559	14095	62	91	517	19	2012-08-02 05:09:36.191363
14099	5	1-2489	14095	91	24	199	100	2012-08-02 05:09:36.191363
14095	5	1-4036	14095	24	24	3445	7	2012-08-02 05:09:36.191363
14101	5	1-3878	14100	24	29	2955	33	2012-08-02 05:09:36.191363
14102	5	1-1004	14100	29	35	783	81	2012-08-02 05:09:36.191363
14103	5	1-3323	14100	35	90	250	6	2012-08-02 05:09:36.191363
14104	5	1-3869	14100	90	24	136	100	2012-08-02 05:09:36.191363
14100	5	1-4036	14100	24	24	2415	7	2012-08-02 05:09:36.191363
14106	5	1-3327	14105	61	23	3498	78	2012-08-02 05:09:37.749313
14107	5	1-3521	14105	23	74	1299	11	2012-08-02 05:09:37.749313
14108	5	1-3987	14105	74	56	688	55	2012-08-02 05:09:37.749313
14109	5	1-4007	14105	56	99	1020	29	2012-08-02 05:09:37.749313
14105	5	1-4040	14105	99	61	4805	49	2012-08-02 05:09:37.749313
14111	4	1-3747	14110	98	65	583	98	2012-08-02 05:09:37.749313
14112	4	1-3601	14110	65	56	400	55	2012-08-02 05:09:37.749313
14113	4	1-4007	14110	56	99	600	29	2012-08-02 05:09:37.749313
14110	4	1-4040	14110	99	98	2870	49	2012-08-02 05:09:37.749313
14115	5	1-1130	14114	92	62	306	37	2012-08-02 05:09:38.147286
14116	5	1-3617	14114	62	68	158	54	2012-08-02 05:09:38.147286
14117	5	1-3839	14114	68	72	307	78	2012-08-02 05:09:38.147286
14118	5	1-3947	14114	72	14	133	20	2012-08-02 05:09:38.147286
14114	5	1-4041	14114	14	92	401	28	2012-08-02 05:09:38.147286
14120	5	1-1130	14119	92	99	1885	37	2012-08-02 05:09:38.147286
14121	5	1-3303	14119	99	30	1491	48	2012-08-02 05:09:38.147286
14122	5	1-3331	14119	30	94	827	68	2012-08-02 05:09:38.147286
14123	5	1-3906	14119	94	14	755	20	2012-08-02 05:09:38.147286
14119	5	1-4041	14119	14	92	2371	28	2012-08-02 05:09:38.147286
14125	4	1-3203	14124	85	12	4318	13	2012-08-02 05:09:39.362781
14126	4	1-3240	14124	12	73	2137	74	2012-08-02 05:09:39.362781
14127	4	1-3955	14124	73	12	4813	65	2012-08-02 05:09:39.362781
14124	4	1-4044	14124	12	85	6651	36	2012-08-02 05:09:39.362781
14129	4	1-3966	14128	39	88	388	84	2012-08-02 05:09:40.158205
14130	4	1-3737	14128	88	48	349	93	2012-08-02 05:09:40.158205
14131	4	1-2027	14128	48	42	153	59	2012-08-02 05:09:40.158205
14128	4	1-4046	14128	42	39	560	43	2012-08-02 05:09:40.158205
14133	5	1-3966	14132	39	26	18	84	2012-08-02 05:09:40.158205
14134	5	1-3000	14132	26	91	16	46	2012-08-02 05:09:40.158205
14135	5	1-1322	14132	91	45	10	33	2012-08-02 05:09:40.158205
14136	5	1-3291	14132	45	42	6	59	2012-08-02 05:09:40.158205
14132	5	1-4046	14132	42	39	24	43	2012-08-02 05:09:40.158205
14138	5	1-3003	14137	94	97	214	8	2012-08-02 05:09:40.158205
14139	5	1-3557	14137	97	24	251	7	2012-08-02 05:09:40.158205
14140	5	1-3878	14137	24	88	339	33	2012-08-02 05:09:40.158205
14141	5	1-2535	14137	88	42	227	59	2012-08-02 05:09:40.158205
14137	5	1-4046	14137	42	94	974	43	2012-08-02 05:09:40.158205
14143	5	1-3966	14142	39	90	4544	84	2012-08-02 05:09:40.158205
14144	5	1-1533	14142	90	91	2320	46	2012-08-02 05:09:40.158205
14145	5	1-1322	14142	91	45	1622	33	2012-08-02 05:09:40.158205
14146	5	1-3291	14142	45	42	1156	59	2012-08-02 05:09:40.158205
14142	5	1-4046	14142	42	39	5273	43	2012-08-02 05:09:40.158205
14148	5	1-1888	14147	24	17	204	16	2012-08-02 05:09:40.158205
14149	5	1-2597	14147	17	46	41	66	2012-08-02 05:09:40.158205
14150	5	1-3720	14147	46	90	81	22	2012-08-02 05:09:40.158205
14151	5	1-3749	14147	90	42	90	59	2012-08-02 05:09:40.158205
14147	5	1-4046	14147	42	24	419	43	2012-08-02 05:09:40.158205
14153	5	1-1895	14152	87	45	642	23	2012-08-02 05:09:40.158205
14154	5	1-3937	14152	45	13	795	14	2012-08-02 05:09:40.158205
14155	5	1-3909	14152	13	45	501	33	2012-08-02 05:09:40.158205
14156	5	1-3291	14152	45	42	374	59	2012-08-02 05:09:40.158205
14152	5	1-4046	14152	42	87	1781	43	2012-08-02 05:09:40.158205
14157	1	1-1895	\N	87	87	1552	23	2012-08-02 05:09:40.158205
14159	4	1-3601	14158	65	56	2120	55	2012-08-02 05:09:41.782328
14160	4	1-4007	14158	56	23	3977	29	2012-08-02 05:09:41.782328
14161	4	1-3227	14158	23	84	2121	3	2012-08-02 05:09:41.782328
14158	4	1-4050	14158	84	65	2479	98	2012-08-02 05:09:41.782328
14163	3	1-3814	14162	32	26	7732	60	2012-08-02 05:09:42.191185
14164	3	1-2053	14162	26	66	1203	23	2012-08-02 05:09:42.191185
14162	3	1-4051	14162	66	32	4223	62	2012-08-02 05:09:42.191185
14166	3	1-3713	14165	12	98	1035	56	2012-08-02 05:09:42.191185
14167	3	1-2146	14165	98	66	698	23	2012-08-02 05:09:42.191185
14165	3	1-4051	14165	66	12	2866	62	2012-08-02 05:09:42.191185
14169	4	1-1891	14168	69	29	2276	40	2012-08-02 05:09:42.987019
14170	4	1-3971	14168	29	87	1106	52	2012-08-02 05:09:42.987019
14171	4	1-2200	14168	87	11	1198	59	2012-08-02 05:09:42.987019
14168	4	1-4053	14168	11	69	1552	39	2012-08-02 05:09:42.987019
14173	3	1-3956	14172	4	30	402	28	2012-08-02 05:09:45.373138
14174	3	1-3688	14172	30	98	694	13	2012-08-02 05:09:45.373138
14172	3	1-4059	14172	98	4	378	43	2012-08-02 05:09:45.373138
14176	5	1-3397	14175	53	40	543	96	2012-08-02 05:09:46.146606
14177	5	1-3635	14175	40	81	1988	27	2012-08-02 05:09:46.146606
14178	5	1-2918	14175	81	28	1534	93	2012-08-02 05:09:46.146606
14179	5	1-4027	14175	28	44	871	1	2012-08-02 05:09:46.146606
14175	5	1-4061	14175	44	53	885	71	2012-08-02 05:09:46.146606
14181	5	1-3397	14180	53	71	1040	96	2012-08-02 05:09:46.146606
14182	5	1-3401	14180	71	61	1062	98	2012-08-02 05:09:46.146606
14183	5	1-2920	14180	61	87	1674	31	2012-08-02 05:09:46.146606
14184	5	1-3596	14180	87	44	1187	1	2012-08-02 05:09:46.146606
14180	5	1-4061	14180	44	53	1430	71	2012-08-02 05:09:46.146606
16131	3	1-4433	16129	91	11	581	49	2012-08-02 05:13:39.62595
14186	5	1-3397	14185	53	64	1050	96	2012-08-02 05:09:46.146606
14187	5	1-3449	14185	64	31	558	92	2012-08-02 05:09:46.146606
14188	5	1-3556	14185	31	86	561	34	2012-08-02 05:09:46.146606
14189	5	1-3763	14185	86	44	1188	1	2012-08-02 05:09:46.146606
14185	5	1-4061	14185	44	53	1437	71	2012-08-02 05:09:46.146606
16129	3	1-4643	16129	11	45	1421	68	2012-08-02 05:13:39.62595
14191	5	1-2256	14190	82	84	110	3	2012-08-02 05:09:46.555255
14192	5	1-4050	14190	84	65	107	98	2012-08-02 05:09:46.555255
14193	5	1-3601	14190	65	95	76	55	2012-08-02 05:09:46.555255
14194	5	1-3681	14190	95	77	36	45	2012-08-02 05:09:46.555255
14190	5	1-4062	14190	77	82	748	13	2012-08-02 05:09:46.555255
14196	4	1-3411	14195	58	44	715	63	2012-08-02 05:09:46.555255
14197	4	1-3912	14195	44	95	498	55	2012-08-02 05:09:46.555255
14198	4	1-3681	14195	95	77	292	45	2012-08-02 05:09:46.555255
14195	4	1-4062	14195	77	58	7348	13	2012-08-02 05:09:46.555255
14200	5	1-3240	14199	12	91	196	74	2012-08-02 05:09:46.555255
14201	5	1-3764	14199	91	35	71	81	2012-08-02 05:09:46.555255
14202	5	1-3323	14199	35	77	31	6	2012-08-02 05:09:46.555255
14203	5	1-3665	14199	77	77	14	45	2012-08-02 05:09:46.555255
14199	5	1-4062	14199	77	12	436	13	2012-08-02 05:09:46.555255
14205	4	1-4004	14204	11	63	1552	100	2012-08-02 05:09:47.350952
14206	4	1-3786	14204	63	82	1448	58	2012-08-02 05:09:47.350952
14207	4	1-3606	14204	82	97	2365	5	2012-08-02 05:09:47.350952
14204	4	1-4064	14204	97	11	2396	93	2012-08-02 05:09:47.350952
14209	5	1-2837	14208	53	74	56	31	2012-08-02 05:09:49.350954
14210	5	1-3560	14208	74	21	54	88	2012-08-02 05:09:49.350954
14211	5	1-3583	14208	21	89	31	51	2012-08-02 05:09:49.350954
14212	5	1-2984	14208	89	41	15	12	2012-08-02 05:09:49.350954
14208	5	1-4069	14208	41	53	45	99	2012-08-02 05:09:49.350954
14214	4	1-3884	14213	76	77	2781	48	2012-08-02 05:09:49.350954
14215	4	1-3965	14213	77	76	1203	44	2012-08-02 05:09:49.350954
14216	4	1-3493	14213	76	41	1406	12	2012-08-02 05:09:49.350954
14213	4	1-4069	14213	41	76	4284	99	2012-08-02 05:09:49.350954
14218	4	1-2196	14217	14	26	774	4	2012-08-02 05:09:50.910689
14219	4	1-3339	14217	26	99	466	86	2012-08-02 05:09:50.910689
14220	4	1-3992	14217	99	78	1600	27	2012-08-02 05:09:50.910689
14217	4	1-4073	14217	78	14	3856	48	2012-08-02 05:09:50.910689
14222	5	1-3844	14221	87	99	3142	10	2012-08-02 05:09:50.910689
14223	5	1-3926	14221	99	94	3107	68	2012-08-02 05:09:50.910689
14224	5	1-3906	14221	94	49	2919	20	2012-08-02 05:09:50.910689
14225	5	1-4049	14221	49	78	575	27	2012-08-02 05:09:50.910689
14221	5	1-4073	14221	78	87	1642	48	2012-08-02 05:09:50.910689
14227	4	1-3923	14226	82	72	478	38	2012-08-02 05:09:52.081625
14228	4	1-1554	14226	72	92	232	99	2012-08-02 05:09:52.081625
14229	4	1-3667	14226	92	46	134	30	2012-08-02 05:09:52.081625
14226	4	1-4076	14226	46	82	465	6	2012-08-02 05:09:52.081625
14231	5	1-4043	14230	43	63	4668	34	2012-08-02 05:09:52.081625
14232	5	1-2995	14230	63	88	1483	18	2012-08-02 05:09:52.081625
14233	5	1-2698	14230	88	60	1844	32	2012-08-02 05:09:52.081625
14234	5	1-4068	14230	60	46	529	30	2012-08-02 05:09:52.081625
14230	5	1-4076	14230	46	43	1847	6	2012-08-02 05:09:52.081625
14236	5	1-3551	14235	25	74	4450	11	2012-08-02 05:09:52.081625
14237	5	1-3987	14235	74	95	3548	55	2012-08-02 05:09:52.081625
14238	5	1-3681	14235	95	77	2411	45	2012-08-02 05:09:52.081625
14239	5	1-3373	14235	77	46	973	30	2012-08-02 05:09:52.081625
14235	5	1-4076	14235	46	25	3788	6	2012-08-02 05:09:52.081625
14241	4	1-3765	14240	4	58	2718	56	2012-08-02 05:09:52.081625
14242	4	1-1593	14240	58	92	1222	99	2012-08-02 05:09:52.081625
14243	4	1-3667	14240	92	46	802	30	2012-08-02 05:09:52.081625
14240	4	1-4076	14240	46	4	3178	6	2012-08-02 05:09:52.081625
14244	1	1-1593	\N	58	58	1132	99	2012-08-02 05:09:52.081625
14246	4	1-3867	14245	85	8	5042	50	2012-08-02 05:09:52.501621
14247	4	1-2055	14245	8	100	615	71	2012-08-02 05:09:52.501621
14248	4	1-2802	14245	100	55	185	99	2012-08-02 05:09:52.501621
14245	4	1-4077	14245	55	85	3754	94	2012-08-02 05:09:52.501621
14250	5	1-1161	14249	48	89	326	10	2012-08-02 05:09:52.501621
14251	5	1-3182	14249	89	37	138	49	2012-08-02 05:09:52.501621
14252	5	1-3658	14249	37	100	93	71	2012-08-02 05:09:52.501621
14253	5	1-2802	14249	100	55	40	99	2012-08-02 05:09:52.501621
14249	5	1-4077	14249	55	48	1154	94	2012-08-02 05:09:52.501621
14255	3	1-2808	14254	74	3	1591	66	2012-08-02 05:09:53.308598
14256	3	1-2454	14254	3	90	1008	79	2012-08-02 05:09:53.308598
14254	3	1-4079	14254	90	74	4660	29	2012-08-02 05:09:53.308598
14258	5	1-3052	14257	59	32	3931	67	2012-08-02 05:09:53.308598
14259	5	1-1399	14257	32	42	3229	32	2012-08-02 05:09:53.308598
14260	5	1-3881	14257	42	3	1364	66	2012-08-02 05:09:53.308598
14261	5	1-2454	14257	3	90	900	79	2012-08-02 05:09:53.308598
14257	5	1-4079	14257	90	59	4335	29	2012-08-02 05:09:53.308598
14263	4	1-3407	14262	39	86	3101	40	2012-08-02 05:09:54.10386
14264	4	1-4054	14262	86	42	2042	31	2012-08-02 05:09:54.10386
14265	4	1-4010	14262	42	49	3231	87	2012-08-02 05:09:54.10386
14262	4	1-4081	14262	49	39	2976	69	2012-08-02 05:09:54.10386
14267	4	1-2506	14266	11	86	821	64	2012-08-02 05:09:54.10386
14268	4	1-2375	14266	86	42	1430	31	2012-08-02 05:09:54.10386
14269	4	1-4010	14266	42	49	2518	87	2012-08-02 05:09:54.10386
14266	4	1-4081	14266	49	11	2582	69	2012-08-02 05:09:54.10386
14271	5	1-2984	14270	89	3	402	12	2012-08-02 05:09:55.2862
14272	5	1-4014	14270	3	17	213	56	2012-08-02 05:09:55.2862
14273	5	1-1328	14270	17	90	172	22	2012-08-02 05:09:55.2862
14274	5	1-3749	14270	90	85	168	59	2012-08-02 05:09:55.2862
14270	5	1-4084	14270	85	89	663	51	2012-08-02 05:09:55.2862
14276	4	1-3809	14275	76	4	1669	87	2012-08-02 05:09:55.2862
14277	4	1-3477	14275	4	87	1068	52	2012-08-02 05:09:55.2862
14278	4	1-2200	14275	87	85	1201	59	2012-08-02 05:09:55.2862
14275	4	1-4084	14275	85	76	5111	51	2012-08-02 05:09:55.2862
14279	1	1-1490	\N	13	13	814	59	2012-08-02 05:09:55.2862
14281	4	1-3841	14280	20	11	3942	25	2012-08-02 05:09:57.661387
14282	4	1-2380	14280	11	23	1119	54	2012-08-02 05:09:57.661387
14283	4	1-4047	14280	23	35	1997	42	2012-08-02 05:09:57.661387
14280	4	1-4090	14280	35	20	8402	84	2012-08-02 05:09:57.661387
14285	5	1-2082	14284	94	14	285	48	2012-08-02 05:09:57.661387
14286	5	1-2196	14284	14	17	54	4	2012-08-02 05:09:57.661387
14287	5	1-2804	14284	17	23	54	54	2012-08-02 05:09:57.661387
14288	5	1-4047	14284	23	35	97	42	2012-08-02 05:09:57.661387
14284	5	1-4090	14284	35	94	411	84	2012-08-02 05:09:57.661387
14290	5	1-3240	14289	12	56	644	74	2012-08-02 05:09:59.351679
14291	5	1-3941	14289	56	91	631	46	2012-08-02 05:09:59.351679
14292	5	1-1322	14289	91	45	416	33	2012-08-02 05:09:59.351679
14293	5	1-3291	14289	45	91	251	59	2012-08-02 05:09:59.351679
14289	5	1-4094	14289	91	12	1596	13	2012-08-02 05:09:59.351679
16133	3	1-4006	16132	99	21	1702	21	2012-08-02 05:13:39.62595
14295	5	1-3240	14294	12	26	142	74	2012-08-02 05:09:59.351679
14296	5	1-3855	14294	26	21	34	54	2012-08-02 05:09:59.351679
14297	5	1-3848	14294	21	90	39	22	2012-08-02 05:09:59.351679
14298	5	1-3749	14294	90	91	43	59	2012-08-02 05:09:59.351679
14294	5	1-4094	14294	91	12	295	13	2012-08-02 05:09:59.351679
16134	3	1-3051	16132	21	11	873	49	2012-08-02 05:13:39.62595
14300	3	1-3616	14299	44	44	2810	1	2012-08-02 05:10:01.318292
14301	3	1-4061	14299	44	89	3277	71	2012-08-02 05:10:01.318292
14299	3	1-4099	14299	89	44	6501	36	2012-08-02 05:10:01.318292
14303	4	1-4017	14302	44	44	475	61	2012-08-02 05:10:03.273669
14304	4	1-2461	14302	44	55	71	99	2012-08-02 05:10:03.273669
14305	4	1-4077	14302	55	55	1544	94	2012-08-02 05:10:03.273669
14302	4	1-4104	14302	55	44	792	95	2012-08-02 05:10:03.273669
14307	5	1-4017	14306	44	10	1596	61	2012-08-02 05:10:04.069331
14308	5	1-3627	14306	10	67	1353	28	2012-08-02 05:10:04.069331
14309	5	1-4078	14306	67	82	1233	58	2012-08-02 05:10:04.069331
14310	5	1-3606	14306	82	97	2265	5	2012-08-02 05:10:04.069331
14306	5	1-4106	14306	97	44	2416	95	2012-08-02 05:10:04.069331
14312	6	1-3216	14311	91	50	1758	14	2012-08-02 05:10:04.46703
14313	6	1-3400	14311	50	97	1331	5	2012-08-02 05:10:04.46703
14314	6	1-4106	14311	97	9	1387	95	2012-08-02 05:10:04.46703
14315	6	1-3082	14311	9	46	553	92	2012-08-02 05:10:04.46703
14316	6	1-3849	14311	46	29	833	82	2012-08-02 05:10:04.46703
14311	6	1-4107	14311	29	91	1951	70	2012-08-02 05:10:04.46703
14318	6	1-4092	14317	64	33	6693	7	2012-08-02 05:10:04.46703
14319	6	1-1866	14317	33	53	2434	71	2012-08-02 05:10:04.46703
14320	6	1-3397	14317	53	64	1709	96	2012-08-02 05:10:04.46703
14321	6	1-3449	14317	64	46	900	92	2012-08-02 05:10:04.46703
14322	6	1-3849	14317	46	29	1444	82	2012-08-02 05:10:04.46703
14317	6	1-4107	14317	29	64	3605	70	2012-08-02 05:10:04.46703
14323	1	1-3397	\N	53	53	602	96	2012-08-02 05:10:04.46703
14325	4	1-4000	14324	24	65	1729	98	2012-08-02 05:10:06.501623
14326	4	1-3601	14324	65	95	1926	55	2012-08-02 05:10:06.501623
14327	4	1-3681	14324	95	32	1435	45	2012-08-02 05:10:06.501623
14324	4	1-4112	14324	32	24	4212	60	2012-08-02 05:10:06.501623
14329	5	1-4024	14328	96	3	974	12	2012-08-02 05:10:06.888585
14330	5	1-4014	14328	3	98	472	56	2012-08-02 05:10:06.888585
14331	5	1-2146	14328	98	6	396	23	2012-08-02 05:10:06.888585
14332	5	1-3856	14328	6	30	761	2	2012-08-02 05:10:06.888585
14328	5	1-4113	14328	30	96	970	72	2012-08-02 05:10:06.888585
14334	5	1-3874	14333	12	91	403	59	2012-08-02 05:10:06.888585
14335	5	1-4094	14333	91	79	2303	13	2012-08-02 05:10:06.888585
14336	5	1-3695	14333	79	6	376	23	2012-08-02 05:10:06.888585
14337	5	1-3856	14333	6	30	757	2	2012-08-02 05:10:06.888585
14333	5	1-4113	14333	30	12	1011	72	2012-08-02 05:10:06.888585
14339	5	1-4024	14338	96	9	3746	12	2012-08-02 05:10:06.888585
14340	5	1-3202	14338	9	21	1910	53	2012-08-02 05:10:06.888585
14341	5	1-4033	14338	21	34	1811	21	2012-08-02 05:10:06.888585
14342	5	1-3545	14338	34	30	2638	2	2012-08-02 05:10:06.888585
14338	5	1-4113	14338	30	96	3542	72	2012-08-02 05:10:06.888585
14344	5	1-1718	14343	43	43	463	6	2012-08-02 05:10:06.888585
14345	5	1-4043	14343	43	65	1454	34	2012-08-02 05:10:06.888585
14346	5	1-1997	14343	65	9	1050	79	2012-08-02 05:10:06.888585
14347	5	1-3248	14343	9	30	1130	2	2012-08-02 05:10:06.888585
14343	5	1-4113	14343	30	43	1868	72	2012-08-02 05:10:06.888585
14348	1	1-685	\N	15	15	1867	23	2012-08-02 05:10:06.888585
14350	3	1-880	14349	68	97	3479	5	2012-08-02 05:10:08.855118
14351	3	1-4106	14349	97	3	2589	95	2012-08-02 05:10:08.855118
14349	3	1-4118	14349	3	68	3096	96	2012-08-02 05:10:08.855118
14353	4	1-2992	14352	84	84	928	3	2012-08-02 05:10:08.855118
14354	4	1-4050	14352	84	60	1204	98	2012-08-02 05:10:08.855118
14355	4	1-3876	14352	60	3	960	95	2012-08-02 05:10:08.855118
14352	4	1-4118	14352	3	84	1281	96	2012-08-02 05:10:08.855118
14357	4	1-1938	14356	4	50	1875	14	2012-08-02 05:10:08.855118
14358	4	1-3400	14356	50	97	1185	5	2012-08-02 05:10:08.855118
14359	4	1-4106	14356	97	3	1030	95	2012-08-02 05:10:08.855118
14356	4	1-4118	14356	3	4	1438	96	2012-08-02 05:10:08.855118
14361	3	1-3401	14360	71	60	1898	98	2012-08-02 05:10:08.855118
14362	3	1-3876	14360	60	3	1642	95	2012-08-02 05:10:08.855118
14360	3	1-4118	14360	3	71	2375	96	2012-08-02 05:10:08.855118
14364	5	1-3323	14363	35	43	169	6	2012-08-02 05:10:09.650739
14365	5	1-4043	14363	43	65	503	34	2012-08-02 05:10:09.650739
14366	5	1-1997	14363	65	72	345	79	2012-08-02 05:10:09.650739
14367	5	1-1523	14363	72	57	333	5	2012-08-02 05:10:09.650739
14363	5	1-4120	14363	57	35	368	81	2012-08-02 05:10:09.650739
14369	3	1-4097	14368	45	83	4089	52	2012-08-02 05:10:11.230541
14370	3	1-1404	14368	83	67	1005	98	2012-08-02 05:10:11.230541
14368	3	1-4124	14368	67	45	8594	76	2012-08-02 05:10:11.230541
14372	5	1-3291	14371	45	91	108	59	2012-08-02 05:10:12.015332
14373	5	1-4094	14371	91	79	700	13	2012-08-02 05:10:12.015332
14374	5	1-3695	14371	79	99	129	23	2012-08-02 05:10:12.015332
14375	5	1-3218	14371	99	47	62	47	2012-08-02 05:10:12.015332
14371	5	1-4126	14371	47	45	159	33	2012-08-02 05:10:12.015332
14377	3	1-2032	14376	63	46	592	15	2012-08-02 05:10:14.370014
14378	3	1-3712	14376	46	81	357	22	2012-08-02 05:10:14.370014
14376	3	1-4132	14376	81	63	695	49	2012-08-02 05:10:14.370014
14380	5	1-3840	14379	95	45	720	33	2012-08-02 05:10:14.370014
14381	5	1-3291	14379	45	61	531	59	2012-08-02 05:10:14.370014
14382	5	1-3703	14379	61	46	519	15	2012-08-02 05:10:14.370014
14383	5	1-3712	14379	46	81	340	22	2012-08-02 05:10:14.370014
14379	5	1-4132	14379	81	95	716	49	2012-08-02 05:10:14.370014
14385	5	1-3840	14384	95	88	1440	33	2012-08-02 05:10:14.370014
14386	5	1-2535	14384	88	61	1059	59	2012-08-02 05:10:14.370014
14387	5	1-3703	14384	61	46	1036	15	2012-08-02 05:10:14.370014
14388	5	1-3712	14384	46	81	678	22	2012-08-02 05:10:14.370014
14384	5	1-4132	14384	81	95	1431	49	2012-08-02 05:10:14.370014
14389	1	1-2535	\N	88	88	751	59	2012-08-02 05:10:14.370014
14391	5	1-1997	14390	65	46	345	79	2012-08-02 05:10:15.552451
14390	5	1-4135	14390	1	65	661	34	2012-08-02 05:10:15.552451
14392	5	1-3823	14390	46	63	60	100	2012-08-02 05:10:15.552451
14393	5	1-3786	14390	63	69	51	58	2012-08-02 05:10:15.552451
14394	5	1-2668	14390	69	1	48	57	2012-08-02 05:10:15.552451
16132	3	1-4643	16132	11	99	2298	68	2012-08-02 05:13:39.62595
14396	5	1-3862	14395	75	75	1638	97	2012-08-02 05:10:15.552451
14397	5	1-3396	14395	75	94	735	43	2012-08-02 05:10:15.552451
14398	5	1-3003	14395	94	6	134	8	2012-08-02 05:10:15.552451
14399	5	1-3486	14395	6	1	136	57	2012-08-02 05:10:15.552451
14395	5	1-4135	14395	1	75	1924	34	2012-08-02 05:10:15.552451
14401	4	1-1402	14400	45	21	160	53	2012-08-02 05:10:15.972473
14402	4	1-4033	14400	21	90	112	21	2012-08-02 05:10:15.972473
14403	4	1-3553	14400	90	20	61	51	2012-08-02 05:10:15.972473
14400	4	1-4136	14400	20	45	187	99	2012-08-02 05:10:15.972473
16136	5	1-3556	16135	31	9	133	34	2012-08-02 05:13:40.056933
14405	5	1-3884	14404	76	84	296	48	2012-08-02 05:10:15.972473
14406	5	1-3625	14404	84	48	203	93	2012-08-02 05:10:15.972473
14407	5	1-2027	14404	48	85	70	59	2012-08-02 05:10:15.972473
14408	5	1-4084	14404	85	20	198	51	2012-08-02 05:10:15.972473
14404	5	1-4136	14404	20	76	616	99	2012-08-02 05:10:15.972473
14410	5	1-3337	14409	76	15	750	47	2012-08-02 05:10:15.972473
14411	5	1-1855	14409	15	16	528	57	2012-08-02 05:10:15.972473
14412	5	1-1445	14409	16	85	144	59	2012-08-02 05:10:15.972473
14413	5	1-4084	14409	85	20	416	51	2012-08-02 05:10:15.972473
14409	5	1-4136	14409	20	76	1325	99	2012-08-02 05:10:15.972473
14415	4	1-3337	14414	76	30	2149	47	2012-08-02 05:10:15.972473
14416	4	1-1549	14414	30	21	2165	88	2012-08-02 05:10:15.972473
14417	4	1-3583	14414	21	20	1180	51	2012-08-02 05:10:15.972473
14414	4	1-4136	14414	20	76	3775	99	2012-08-02 05:10:15.972473
14419	3	1-3272	14418	49	90	856	21	2012-08-02 05:10:15.972473
14420	3	1-3553	14418	90	20	499	51	2012-08-02 05:10:15.972473
14418	3	1-4136	14418	20	49	1654	99	2012-08-02 05:10:15.972473
14422	5	1-4112	14421	32	100	1001	60	2012-08-02 05:10:16.381474
14423	5	1-1844	14421	100	65	690	58	2012-08-02 05:10:16.381474
14424	5	1-3972	14421	65	96	663	13	2012-08-02 05:10:16.381474
14425	5	1-3179	14421	96	48	543	94	2012-08-02 05:10:16.381474
14421	5	1-4137	14421	48	32	370	45	2012-08-02 05:10:16.381474
14427	5	1-4024	14426	96	56	3952	12	2012-08-02 05:10:17.972547
14428	5	1-4109	14426	56	16	1576	18	2012-08-02 05:10:17.972547
14429	5	1-2682	14426	16	23	1042	66	2012-08-02 05:10:17.972547
14430	5	1-3555	14426	23	3	2017	60	2012-08-02 05:10:17.972547
14426	5	1-4141	14426	3	96	3093	72	2012-08-02 05:10:17.972547
14432	4	1-3940	14431	88	91	1528	40	2012-08-02 05:10:18.358838
14433	4	1-4100	14431	91	68	260	45	2012-08-02 05:10:18.358838
14434	4	1-3816	14431	68	84	312	90	2012-08-02 05:10:18.358838
14431	4	1-4142	14431	84	88	472	21	2012-08-02 05:10:18.358838
14436	4	1-3961	14435	15	75	1129	41	2012-08-02 05:10:19.676322
14437	4	1-3784	14435	75	89	798	10	2012-08-02 05:10:19.676322
14438	4	1-3182	14435	89	84	281	49	2012-08-02 05:10:19.676322
14435	4	1-4145	14435	84	15	600	69	2012-08-02 05:10:19.676322
14440	5	1-4055	14439	90	96	4691	78	2012-08-02 05:10:19.676322
14441	5	1-2225	14439	96	21	815	54	2012-08-02 05:10:19.676322
14442	5	1-3848	14439	21	81	781	22	2012-08-02 05:10:19.676322
14443	5	1-4132	14439	81	84	1346	49	2012-08-02 05:10:19.676322
14439	5	1-4145	14439	84	90	3134	69	2012-08-02 05:10:19.676322
14445	4	1-4073	14444	78	77	2488	48	2012-08-02 05:10:20.480499
14446	4	1-3965	14444	77	76	1577	44	2012-08-02 05:10:20.480499
14447	4	1-3493	14444	76	11	2199	12	2012-08-02 05:10:20.480499
14444	4	1-4147	14444	11	78	895	27	2012-08-02 05:10:20.480499
14449	5	1-3849	14448	46	29	364	82	2012-08-02 05:10:20.877917
14450	5	1-4107	14448	29	80	725	70	2012-08-02 05:10:20.877917
14451	5	1-3679	14448	80	89	1063	51	2012-08-02 05:10:20.877917
14452	5	1-2984	14448	89	42	572	12	2012-08-02 05:10:20.877917
14448	5	1-4148	14448	42	46	285	92	2012-08-02 05:10:20.877917
14454	5	1-3707	14453	58	47	970	89	2012-08-02 05:10:22.435986
14455	5	1-4098	14453	47	2	2342	9	2012-08-02 05:10:22.435986
14456	5	1-3998	14453	2	4	6886	43	2012-08-02 05:10:22.435986
14457	5	1-3956	14453	4	60	6591	28	2012-08-02 05:10:22.435986
14453	5	1-4152	14453	60	58	2276	41	2012-08-02 05:10:22.435986
14459	5	1-3784	14458	75	99	1048	10	2012-08-02 05:10:22.435986
14460	5	1-3926	14458	99	94	1016	68	2012-08-02 05:10:22.435986
14461	5	1-3906	14458	94	14	935	20	2012-08-02 05:10:22.435986
14462	5	1-4041	14458	14	60	2961	28	2012-08-02 05:10:22.435986
14458	5	1-4152	14458	60	75	1118	41	2012-08-02 05:10:22.435986
14464	4	1-3924	14463	42	1	2563	25	2012-08-02 05:10:23.220218
14465	4	1-3916	14463	1	100	2508	83	2012-08-02 05:10:23.220218
14466	4	1-3450	14463	100	18	2312	17	2012-08-02 05:10:23.220218
14463	4	1-4154	14463	18	42	2373	79	2012-08-02 05:10:23.220218
14468	5	1-1658	14467	53	68	1113	37	2012-08-02 05:10:24.004685
14469	5	1-3497	14467	68	74	1031	11	2012-08-02 05:10:24.004685
14470	5	1-3987	14467	74	87	836	55	2012-08-02 05:10:24.004685
14471	5	1-3859	14467	87	63	1625	97	2012-08-02 05:10:24.004685
14467	5	1-4156	14467	63	53	1312	64	2012-08-02 05:10:24.004685
14473	4	1-3246	14472	24	74	112	11	2012-08-02 05:10:24.004685
14474	4	1-3987	14472	74	87	97	55	2012-08-02 05:10:24.004685
14475	4	1-3859	14472	87	63	201	97	2012-08-02 05:10:24.004685
14472	4	1-4156	14472	63	24	173	64	2012-08-02 05:10:24.004685
14477	5	1-868	14476	88	21	604	24	2012-08-02 05:10:25.540632
14478	5	1-4063	14476	21	18	1047	15	2012-08-02 05:10:25.540632
14479	5	1-3806	14476	18	42	627	12	2012-08-02 05:10:25.540632
14480	5	1-4148	14476	42	10	372	92	2012-08-02 05:10:25.540632
14476	5	1-4160	14476	10	88	1433	61	2012-08-02 05:10:25.540632
14482	4	1-3874	14481	12	11	1041	59	2012-08-02 05:10:25.949281
14483	4	1-4053	14481	11	86	1481	39	2012-08-02 05:10:25.949281
14484	4	1-2326	14481	86	86	1446	42	2012-08-02 05:10:25.949281
14481	4	1-4161	14481	86	12	2190	72	2012-08-02 05:10:25.949281
14486	5	1-1894	14485	89	88	2	90	2012-08-02 05:10:26.722573
14487	5	1-3287	14485	88	1	1	57	2012-08-02 05:10:26.722573
14488	5	1-4135	14485	1	1	12	34	2012-08-02 05:10:26.722573
14489	5	1-1826	14485	1	46	5	25	2012-08-02 05:10:26.722573
14485	5	1-4163	14485	46	89	3	19	2012-08-02 05:10:26.722573
14491	5	1-4001	14490	57	94	19	85	2012-08-02 05:10:26.722573
14492	5	1-3420	14490	94	30	39	43	2012-08-02 05:10:26.722573
14493	5	1-4072	14490	30	57	42	65	2012-08-02 05:10:26.722573
14494	5	1-2592	14490	57	46	24	25	2012-08-02 05:10:26.722573
14490	5	1-4163	14490	46	57	20	19	2012-08-02 05:10:26.722573
14496	4	1-3382	14495	25	60	740	98	2012-08-02 05:10:26.722573
14497	4	1-3876	14495	60	12	838	95	2012-08-02 05:10:26.722573
14498	4	1-2366	14495	12	46	671	25	2012-08-02 05:10:26.722573
14495	4	1-4163	14495	46	25	627	19	2012-08-02 05:10:26.722573
14500	4	1-2862	14499	24	41	411	68	2012-08-02 05:10:27.905031
14501	4	1-3686	14499	41	14	160	75	2012-08-02 05:10:27.905031
14502	4	1-3211	14499	14	25	695	2	2012-08-02 05:10:27.905031
14499	4	1-4166	14499	25	24	487	8	2012-08-02 05:10:27.905031
14504	5	1-2943	14503	49	91	99	59	2012-08-02 05:10:28.291593
14505	5	1-4094	14503	91	96	654	13	2012-08-02 05:10:28.291593
14506	5	1-3179	14503	96	48	549	94	2012-08-02 05:10:28.291593
14507	5	1-4137	14503	48	100	384	45	2012-08-02 05:10:28.291593
14503	5	1-4167	14503	100	49	628	50	2012-08-02 05:10:28.291593
14509	5	1-4119	14508	51	69	4869	35	2012-08-02 05:10:29.065002
14510	5	1-2177	14508	69	28	3930	1	2012-08-02 05:10:29.065002
14511	5	1-3305	14508	28	57	2592	5	2012-08-02 05:10:29.065002
14512	5	1-4120	14508	57	17	1911	81	2012-08-02 05:10:29.065002
14508	5	1-4169	14508	17	51	8809	94	2012-08-02 05:10:29.065002
14513	1	1-2177	\N	69	69	202	1	2012-08-02 05:10:29.065002
14515	5	1-3605	14514	91	21	1877	54	2012-08-02 05:10:29.893876
14516	5	1-3848	14514	21	81	2127	22	2012-08-02 05:10:29.893876
14517	5	1-4132	14514	81	37	4340	49	2012-08-02 05:10:29.893876
14518	5	1-3658	14514	37	77	3167	71	2012-08-02 05:10:29.893876
14514	5	1-4171	14514	77	91	1525	19	2012-08-02 05:10:29.893876
14520	5	1-2752	14519	53	21	1411	53	2012-08-02 05:10:30.269145
14521	5	1-4033	14519	21	53	1511	21	2012-08-02 05:10:30.269145
14522	5	1-4035	14519	53	1	1215	1	2012-08-02 05:10:30.269145
14523	5	1-4030	14519	1	26	2271	88	2012-08-02 05:10:30.269145
14519	5	1-4172	14519	26	53	997	35	2012-08-02 05:10:30.269145
14525	4	1-3702	14524	2	89	4546	10	2012-08-02 05:10:30.667105
14526	4	1-3182	14524	89	47	1938	49	2012-08-02 05:10:30.667105
14527	4	1-2610	14524	47	41	597	79	2012-08-02 05:10:30.667105
14524	4	1-4173	14524	41	2	4346	100	2012-08-02 05:10:30.667105
14529	4	1-4072	14528	30	12	2389	65	2012-08-02 05:10:31.827278
14530	4	1-4044	14528	12	56	2317	36	2012-08-02 05:10:31.827278
14531	4	1-3103	14528	56	4	2236	83	2012-08-02 05:10:31.827278
14528	4	1-4176	14528	4	30	2811	43	2012-08-02 05:10:31.827278
14533	3	1-4130	14532	47	56	4766	36	2012-08-02 05:10:31.827278
14534	3	1-3103	14532	56	4	5121	83	2012-08-02 05:10:31.827278
14532	3	1-4176	14532	4	47	7170	43	2012-08-02 05:10:31.827278
14536	3	1-3601	14535	65	95	688	55	2012-08-02 05:10:33.384674
14537	3	1-3681	14535	95	55	378	45	2012-08-02 05:10:33.384674
14535	3	1-4180	14535	55	65	837	98	2012-08-02 05:10:33.384674
14539	4	1-3449	14538	64	68	1027	92	2012-08-02 05:10:33.771495
14540	4	1-2720	14538	68	18	1235	93	2012-08-02 05:10:33.771495
14541	4	1-2795	14538	18	78	2116	40	2012-08-02 05:10:33.771495
14538	4	1-4181	14538	78	64	2222	96	2012-08-02 05:10:33.771495
14543	5	1-2682	14542	16	46	450	66	2012-08-02 05:10:34.147189
14544	5	1-3720	14542	46	90	622	22	2012-08-02 05:10:34.147189
14545	5	1-3749	14542	90	91	491	59	2012-08-02 05:10:34.147189
14546	5	1-4094	14542	91	87	2375	13	2012-08-02 05:10:34.147189
14542	5	1-4182	14542	87	16	977	18	2012-08-02 05:10:34.147189
14548	5	1-3274	14547	23	69	421	58	2012-08-02 05:10:34.147189
14549	5	1-2668	14547	69	16	412	57	2012-08-02 05:10:34.147189
14550	5	1-1445	14547	16	91	135	59	2012-08-02 05:10:34.147189
14551	5	1-4094	14547	91	87	709	13	2012-08-02 05:10:34.147189
14547	5	1-4182	14547	87	23	316	18	2012-08-02 05:10:34.147189
14553	4	1-4127	14552	28	2	325	9	2012-08-02 05:10:34.147189
14554	4	1-3998	14552	2	33	986	43	2012-08-02 05:10:34.147189
14555	4	1-3657	14552	33	87	523	13	2012-08-02 05:10:34.147189
14552	4	1-4182	14552	87	28	291	18	2012-08-02 05:10:34.147189
14557	5	1-4127	14556	28	8	1994	9	2012-08-02 05:10:34.147189
14558	5	1-3760	14556	8	67	3336	28	2012-08-02 05:10:34.147189
14559	5	1-4078	14556	67	65	3150	58	2012-08-02 05:10:34.147189
14560	5	1-3972	14556	65	87	3120	13	2012-08-02 05:10:34.147189
14556	5	1-4182	14556	87	28	1758	18	2012-08-02 05:10:34.147189
14562	4	1-580	14561	86	73	659	74	2012-08-02 05:10:34.556486
14563	4	1-3955	14561	73	49	1442	65	2012-08-02 05:10:34.556486
14564	4	1-3781	14561	49	97	1090	1	2012-08-02 05:10:34.556486
14561	4	1-4183	14561	97	86	1981	77	2012-08-02 05:10:34.556486
14566	4	1-4177	14565	60	30	3972	43	2012-08-02 05:10:34.556486
14567	4	1-4072	14565	30	49	4625	65	2012-08-02 05:10:34.556486
14568	4	1-3781	14565	49	97	3506	1	2012-08-02 05:10:34.556486
14565	4	1-4183	14565	97	60	6386	77	2012-08-02 05:10:34.556486
14570	2	1-1125	14569	64	12	344	53	2012-08-02 05:10:35.793471
14569	2	1-4186	14569	12	64	882	69	2012-08-02 05:10:35.793471
14572	2	1-1279	14571	70	12	111	53	2012-08-02 05:10:35.793471
14571	2	1-4186	14571	12	70	308	69	2012-08-02 05:10:35.793471
14574	3	1-3449	14573	64	46	827	92	2012-08-02 05:10:37.738004
14575	3	1-3849	14573	46	90	859	82	2012-08-02 05:10:37.738004
14573	3	1-4191	14573	90	64	2502	96	2012-08-02 05:10:37.738004
14577	5	1-3401	14576	71	2	181	98	2012-08-02 05:10:37.738004
14578	5	1-3632	14576	2	82	170	8	2012-08-02 05:10:37.738004
14579	5	1-1806	14576	82	46	56	92	2012-08-02 05:10:37.738004
14580	5	1-3849	14576	46	90	69	82	2012-08-02 05:10:37.738004
14576	5	1-4191	14576	90	71	239	96	2012-08-02 05:10:37.738004
14582	5	1-2504	14581	11	87	1015	88	2012-08-02 05:10:37.738004
14583	5	1-3810	14581	87	42	239	12	2012-08-02 05:10:37.738004
14584	5	1-4148	14581	42	46	125	92	2012-08-02 05:10:37.738004
14585	5	1-3849	14581	46	90	168	82	2012-08-02 05:10:37.738004
14581	5	1-4191	14581	90	11	632	96	2012-08-02 05:10:37.738004
14587	6	1-1089	14586	99	25	623	2	2012-08-02 05:10:37.738004
14588	6	1-4166	14586	25	59	431	8	2012-08-02 05:10:37.738004
14589	6	1-1051	14586	59	42	357	12	2012-08-02 05:10:37.738004
14590	6	1-4148	14586	42	46	217	92	2012-08-02 05:10:37.738004
14591	6	1-3849	14586	46	90	340	82	2012-08-02 05:10:37.738004
14586	6	1-4191	14586	90	99	1492	96	2012-08-02 05:10:37.738004
14592	1	1-3849	\N	46	46	2092	82	2012-08-02 05:10:37.738004
14594	4	1-3950	14593	70	47	3003	47	2012-08-02 05:10:38.544947
14595	4	1-4126	14593	47	55	6521	33	2012-08-02 05:10:38.544947
14596	4	1-3259	14593	55	30	2217	52	2012-08-02 05:10:38.544947
14593	4	1-4193	14593	30	70	4834	24	2012-08-02 05:10:38.544947
14598	3	1-2616	14597	70	36	206	26	2012-08-02 05:10:39.306953
14599	3	1-999	14597	36	84	117	59	2012-08-02 05:10:39.306953
14597	3	1-4195	14597	84	70	549	20	2012-08-02 05:10:39.306953
14601	3	1-3847	14600	1	62	732	29	2012-08-02 05:10:39.306953
14602	3	1-3368	14600	62	84	534	59	2012-08-02 05:10:39.306953
14600	3	1-4195	14600	84	1	2579	20	2012-08-02 05:10:39.306953
14604	5	1-3591	14603	23	40	611	90	2012-08-02 05:10:40.480207
14605	5	1-3404	14603	40	69	326	70	2012-08-02 05:10:40.480207
14606	5	1-3815	14603	69	60	505	77	2012-08-02 05:10:40.480207
14607	5	1-1264	14603	60	98	135	27	2012-08-02 05:10:40.480207
14603	5	1-4198	14603	98	23	413	67	2012-08-02 05:10:40.480207
14609	5	1-3691	14608	8	37	132	39	2012-08-02 05:10:40.480207
14610	5	1-3429	14608	37	76	54	44	2012-08-02 05:10:40.480207
14611	5	1-3493	14608	76	11	72	12	2012-08-02 05:10:40.480207
14612	5	1-4147	14608	11	98	28	27	2012-08-02 05:10:40.480207
14608	5	1-4198	14608	98	8	103	67	2012-08-02 05:10:40.480207
14614	5	1-3026	14613	94	31	1903	4	2012-08-02 05:10:41.253409
14615	5	1-2965	14613	31	11	2237	12	2012-08-02 05:10:41.253409
14616	5	1-4147	14613	11	98	927	27	2012-08-02 05:10:41.253409
14617	5	1-4198	14613	98	93	3642	67	2012-08-02 05:10:41.253409
14613	5	1-4200	14613	93	94	6580	61	2012-08-02 05:10:41.253409
14619	2	1-4001	14618	57	6	1207	85	2012-08-02 05:10:42.026742
14618	2	1-4202	14618	6	57	7213	19	2012-08-02 05:10:42.026742
14621	5	1-775	14620	76	85	1646	53	2012-08-02 05:10:42.435677
14622	5	1-4012	14620	85	100	3075	37	2012-08-02 05:10:42.435677
14623	5	1-3830	14620	100	65	3870	58	2012-08-02 05:10:42.435677
14624	5	1-3972	14620	65	73	3610	13	2012-08-02 05:10:42.435677
14620	5	1-4203	14620	73	76	2992	44	2012-08-02 05:10:42.435677
14626	5	1-2597	14625	17	46	1041	66	2012-08-02 05:10:43.220047
14627	5	1-3920	14625	46	20	950	91	2012-08-02 05:10:43.220047
14628	5	1-3675	14625	20	96	604	61	2012-08-02 05:10:43.220047
14629	5	1-3406	14625	96	1	1850	17	2012-08-02 05:10:43.220047
14625	5	1-4205	14625	1	17	5046	16	2012-08-02 05:10:43.220047
14631	5	1-4198	14630	98	44	4354	67	2012-08-02 05:10:43.606709
14632	5	1-2494	14630	44	46	2576	15	2012-08-02 05:10:43.606709
14633	5	1-3712	14630	46	90	1227	22	2012-08-02 05:10:43.606709
14634	5	1-3749	14630	90	99	1018	59	2012-08-02 05:10:43.606709
14630	5	1-4206	14630	99	98	1531	27	2012-08-02 05:10:43.606709
14636	5	1-4198	14635	98	93	145	67	2012-08-02 05:10:43.606709
14637	5	1-4200	14635	93	89	198	61	2012-08-02 05:10:43.606709
14638	5	1-1924	14635	89	24	57	55	2012-08-02 05:10:43.606709
14639	5	1-2379	14635	24	99	31	59	2012-08-02 05:10:43.606709
14635	5	1-4206	14635	99	98	49	27	2012-08-02 05:10:43.606709
14641	5	1-4180	14640	55	60	745	98	2012-08-02 05:10:44.004535
14642	5	1-3876	14640	60	44	788	95	2012-08-02 05:10:44.004535
14643	5	1-4017	14640	44	51	538	61	2012-08-02 05:10:44.004535
14644	5	1-3495	14640	51	1	329	85	2012-08-02 05:10:44.004535
14640	5	1-4207	14640	1	55	262	45	2012-08-02 05:10:44.004535
14646	5	1-3636	14645	53	1	3077	85	2012-08-02 05:10:44.402339
14647	5	1-4207	14645	1	55	2105	45	2012-08-02 05:10:44.402339
14648	5	1-4180	14645	55	59	5148	98	2012-08-02 05:10:44.402339
14649	5	1-2513	14645	59	48	4152	32	2012-08-02 05:10:44.402339
14645	5	1-4208	14645	48	53	5928	4	2012-08-02 05:10:44.402339
14651	5	1-3677	14650	8	46	1133	25	2012-08-02 05:10:45.230841
14652	5	1-4163	14650	46	38	830	19	2012-08-02 05:10:45.230841
14653	5	1-3286	14650	38	55	535	45	2012-08-02 05:10:45.230841
14654	5	1-4180	14650	55	87	1277	98	2012-08-02 05:10:45.230841
14650	5	1-4210	14650	87	8	1211	18	2012-08-02 05:10:45.230841
14656	4	1-3677	14655	8	46	3835	25	2012-08-02 05:10:45.230841
14657	4	1-4163	14655	46	25	3191	19	2012-08-02 05:10:45.230841
14658	4	1-3382	14655	25	87	3355	98	2012-08-02 05:10:45.230841
14655	4	1-4210	14655	87	8	3611	18	2012-08-02 05:10:45.230841
14660	5	1-3274	14659	23	69	2110	58	2012-08-02 05:10:45.230841
14661	5	1-2668	14659	69	71	2640	57	2012-08-02 05:10:45.230841
14662	5	1-3684	14659	71	75	3122	83	2012-08-02 05:10:45.230841
14663	5	1-3639	14659	75	87	1079	98	2012-08-02 05:10:45.230841
14659	5	1-4210	14659	87	23	1236	18	2012-08-02 05:10:45.230841
14665	4	1-2795	14664	18	91	218	40	2012-08-02 05:10:46.40216
14666	4	1-4100	14664	91	19	29	45	2012-08-02 05:10:46.40216
14667	4	1-2566	14664	19	65	8	67	2012-08-02 05:10:46.40216
14664	4	1-4213	14664	65	18	147	93	2012-08-02 05:10:46.40216
14669	5	1-4027	14668	28	60	4503	1	2012-08-02 05:10:46.40216
14670	5	1-3261	14668	60	94	1868	10	2012-08-02 05:10:46.40216
14671	5	1-3751	14668	94	59	657	29	2012-08-02 05:10:46.40216
14672	5	1-3052	14668	59	65	474	67	2012-08-02 05:10:46.40216
14668	5	1-4213	14668	65	28	8727	93	2012-08-02 05:10:46.40216
14674	4	1-2412	14673	99	14	66	48	2012-08-02 05:10:46.40216
14675	4	1-2196	14673	14	84	12	4	2012-08-02 05:10:46.40216
14676	4	1-3510	14673	84	65	11	67	2012-08-02 05:10:46.40216
14673	4	1-4213	14673	65	99	200	93	2012-08-02 05:10:46.40216
14678	5	1-3958	14677	5	24	1885	55	2012-08-02 05:10:46.811088
14679	5	1-2379	14677	24	99	1058	59	2012-08-02 05:10:46.811088
14680	5	1-4206	14677	99	21	1695	27	2012-08-02 05:10:46.811088
14681	5	1-3531	14677	21	53	2539	76	2012-08-02 05:10:46.811088
14677	5	1-4214	14677	53	5	5249	86	2012-08-02 05:10:46.811088
14683	5	1-3618	14682	26	46	2520	25	2012-08-02 05:10:47.18689
14684	5	1-4163	14682	46	91	1535	19	2012-08-02 05:10:47.18689
14685	5	1-3605	14682	91	23	1289	54	2012-08-02 05:10:47.18689
14686	5	1-4047	14682	23	100	1904	42	2012-08-02 05:10:47.18689
14682	5	1-4215	14682	100	26	2103	75	2012-08-02 05:10:47.18689
14688	3	1-1523	14687	72	57	3929	5	2012-08-02 05:10:49.893745
14689	3	1-4120	14687	57	74	3824	81	2012-08-02 05:10:49.893745
14687	3	1-4222	14687	74	72	4617	79	2012-08-02 05:10:49.893745
14691	5	1-1680	14690	57	89	1644	10	2012-08-02 05:10:49.893745
14692	5	1-3182	14690	89	59	719	49	2012-08-02 05:10:49.893745
14693	5	1-3960	14690	59	57	1216	5	2012-08-02 05:10:49.893745
14694	5	1-4120	14690	57	74	1354	81	2012-08-02 05:10:49.893745
14690	5	1-4222	14690	74	57	1791	79	2012-08-02 05:10:49.893745
14696	2	1-4110	14695	54	28	5184	92	2012-08-02 05:10:50.689072
14695	2	1-4224	14695	28	54	6275	86	2012-08-02 05:10:50.689072
14698	5	1-3617	14697	62	21	391	54	2012-08-02 05:10:51.075766
14699	5	1-3848	14697	21	90	418	22	2012-08-02 05:10:51.075766
14700	5	1-3749	14697	90	99	436	59	2012-08-02 05:10:51.075766
14701	5	1-4206	14697	99	28	823	27	2012-08-02 05:10:51.075766
14697	5	1-4225	14697	28	62	783	37	2012-08-02 05:10:51.075766
14702	1	1-3617	\N	62	62	689	54	2012-08-02 05:10:51.075766
14704	5	1-3466	14703	51	8	3534	70	2012-08-02 05:10:52.6229
14703	5	1-4229	14703	89	51	3676	3	2012-08-02 05:10:52.6229
14705	5	1-4165	14703	8	81	5660	46	2012-08-02 05:10:52.6229
14706	5	1-2048	14703	81	49	3838	53	2012-08-02 05:10:52.6229
14707	5	1-3904	14703	49	89	3069	63	2012-08-02 05:10:52.6229
14709	5	1-4050	14708	84	2	2331	98	2012-08-02 05:10:52.6229
14710	5	1-3632	14708	2	95	3006	8	2012-08-02 05:10:52.6229
14711	5	1-2501	14708	95	30	2448	83	2012-08-02 05:10:52.6229
14712	5	1-4096	14708	30	89	1251	63	2012-08-02 05:10:52.6229
14708	5	1-4229	14708	89	84	1526	3	2012-08-02 05:10:52.6229
14714	4	1-4117	14713	4	53	750	35	2012-08-02 05:10:52.6229
14715	4	1-2752	14713	53	49	1100	53	2012-08-02 05:10:52.6229
14716	4	1-3904	14713	49	89	902	63	2012-08-02 05:10:52.6229
14713	4	1-4229	14713	89	4	1107	3	2012-08-02 05:10:52.6229
14718	5	1-706	14717	66	24	565	55	2012-08-02 05:10:53.81571
14719	5	1-2379	14717	24	99	357	59	2012-08-02 05:10:53.81571
14720	5	1-4206	14717	99	21	643	27	2012-08-02 05:10:53.81571
14721	5	1-3531	14717	21	27	1083	76	2012-08-02 05:10:53.81571
14717	5	1-4232	14717	27	66	1961	73	2012-08-02 05:10:53.81571
14723	3	1-902	14722	11	100	1865	5	2012-08-02 05:10:53.81571
14724	3	1-4101	14722	100	27	1727	76	2012-08-02 05:10:53.81571
14722	3	1-4232	14722	27	11	3598	73	2012-08-02 05:10:53.81571
14726	4	1-3582	14725	27	47	1796	47	2012-08-02 05:10:55.749295
14727	4	1-1861	14725	47	85	946	53	2012-08-02 05:10:55.749295
14728	4	1-4012	14725	85	9	1436	37	2012-08-02 05:10:55.749295
14725	4	1-4237	14725	9	27	3647	81	2012-08-02 05:10:55.749295
14730	3	1-3778	14729	56	53	2449	64	2012-08-02 05:10:55.749295
14731	3	1-1658	14729	53	9	1860	37	2012-08-02 05:10:55.749295
14729	3	1-4237	14729	9	56	5400	81	2012-08-02 05:10:55.749295
14733	4	1-3924	14732	42	1	1338	25	2012-08-02 05:10:56.135998
14734	4	1-3916	14732	1	100	1247	83	2012-08-02 05:10:56.135998
14735	4	1-3450	14732	100	52	1094	17	2012-08-02 05:10:56.135998
14732	4	1-4238	14732	52	42	1301	79	2012-08-02 05:10:56.135998
14737	5	1-1680	14736	57	99	1838	10	2012-08-02 05:10:56.135998
14738	5	1-3926	14736	99	71	1529	68	2012-08-02 05:10:56.135998
14739	5	1-2729	14736	71	96	578	61	2012-08-02 05:10:56.135998
14740	5	1-3406	14736	96	52	1502	17	2012-08-02 05:10:56.135998
14736	5	1-4238	14736	52	57	2283	79	2012-08-02 05:10:56.135998
14742	4	1-3533	14741	23	58	1875	41	2012-08-02 05:10:56.135998
14743	4	1-3707	14741	58	23	753	89	2012-08-02 05:10:56.135998
14744	4	1-3140	14741	23	52	1154	17	2012-08-02 05:10:56.135998
14741	4	1-4238	14741	52	23	1761	79	2012-08-02 05:10:56.135998
14746	4	1-3801	14745	97	18	148	15	2012-08-02 05:10:56.135998
14747	4	1-3806	14745	18	56	92	12	2012-08-02 05:10:56.135998
14748	4	1-3838	14745	56	52	100	17	2012-08-02 05:10:56.135998
14745	4	1-4238	14745	52	97	168	79	2012-08-02 05:10:56.135998
14750	3	1-4152	14749	60	40	428	41	2012-08-02 05:10:57.704946
14751	3	1-3963	14749	40	89	214	14	2012-08-02 05:10:57.704946
14749	3	1-4242	14749	89	60	1557	28	2012-08-02 05:10:57.704946
14753	5	1-4150	14752	37	48	127	93	2012-08-02 05:10:57.704946
14754	5	1-4131	14752	48	85	85	95	2012-08-02 05:10:57.704946
14755	5	1-1463	14752	85	45	19	23	2012-08-02 05:10:57.704946
14756	5	1-3937	14752	45	89	18	14	2012-08-02 05:10:57.704946
14752	5	1-4242	14752	89	37	140	28	2012-08-02 05:10:57.704946
14758	5	1-4119	14757	51	41	983	35	2012-08-02 05:10:58.091705
14759	5	1-3724	14757	41	3	1286	12	2012-08-02 05:10:58.091705
14760	5	1-4014	14757	3	12	751	56	2012-08-02 05:10:58.091705
14761	5	1-3996	14757	12	42	974	68	2012-08-02 05:10:58.091705
14757	5	1-4243	14757	42	51	1216	94	2012-08-02 05:10:58.091705
14763	4	1-4230	14762	16	4	2411	3	2012-08-02 05:10:58.865119
14764	4	1-4117	14762	4	41	1441	35	2012-08-02 05:10:58.865119
14765	4	1-3724	14762	41	83	1780	12	2012-08-02 05:10:58.865119
14762	4	1-4245	14762	83	16	1503	19	2012-08-02 05:10:58.865119
14767	5	1-4230	14766	16	19	5008	3	2012-08-02 05:10:58.865119
14768	5	1-3983	14766	19	79	2396	56	2012-08-02 05:10:58.865119
14769	5	1-3891	14766	79	31	2504	4	2012-08-02 05:10:58.865119
14770	5	1-2965	14766	31	83	2926	12	2012-08-02 05:10:58.865119
14766	5	1-4245	14766	83	16	2777	19	2012-08-02 05:10:58.865119
14772	4	1-4121	14771	57	47	1024	47	2012-08-02 05:11:00.44522
14773	4	1-1861	14771	47	49	752	53	2012-08-02 05:11:00.44522
14774	4	1-3904	14771	49	5	619	63	2012-08-02 05:11:00.44522
14771	4	1-4249	14771	5	57	754	29	2012-08-02 05:11:00.44522
14776	4	1-4205	14775	1	53	1490	16	2012-08-02 05:11:02.013767
14777	4	1-2129	14775	53	14	69	75	2012-08-02 05:11:02.013767
14778	4	1-3211	14775	14	55	301	2	2012-08-02 05:11:02.013767
14775	4	1-4253	14775	55	1	586	17	2012-08-02 05:11:02.013767
14780	3	1-2454	14779	3	23	2640	79	2012-08-02 05:11:02.787434
14781	3	1-3533	14779	23	46	2442	41	2012-08-02 05:11:02.787434
14779	3	1-4255	14779	46	3	4704	66	2012-08-02 05:11:02.787434
14783	5	1-3920	14782	46	74	560	91	2012-08-02 05:11:02.787434
14784	5	1-2786	14782	74	42	513	68	2012-08-02 05:11:02.787434
14785	5	1-4243	14782	42	45	628	94	2012-08-02 05:11:02.787434
14786	5	1-3451	14782	45	46	277	41	2012-08-02 05:11:02.787434
14782	5	1-4255	14782	46	46	637	66	2012-08-02 05:11:02.787434
14788	5	1-4012	14787	85	99	890	37	2012-08-02 05:11:04.322863
14789	5	1-3303	14787	99	84	439	48	2012-08-02 05:11:04.322863
14790	5	1-3625	14787	84	48	302	93	2012-08-02 05:11:04.322863
14791	5	1-4131	14787	48	9	159	95	2012-08-02 05:11:04.322863
14787	5	1-4259	14787	9	85	693	53	2012-08-02 05:11:04.322863
14793	4	1-4033	14792	21	59	2361	21	2012-08-02 05:11:04.322863
14794	4	1-1129	14792	59	48	1298	93	2012-08-02 05:11:04.322863
14795	4	1-4131	14792	48	9	709	95	2012-08-02 05:11:04.322863
14792	4	1-4259	14792	9	21	3215	53	2012-08-02 05:11:04.322863
14797	4	1-3904	14796	49	44	196	63	2012-08-02 05:11:04.322863
14798	4	1-3912	14796	44	22	144	55	2012-08-02 05:11:04.322863
14799	4	1-2991	14796	22	9	50	95	2012-08-02 05:11:04.322863
14796	4	1-4259	14796	9	49	292	53	2012-08-02 05:11:04.322863
14801	3	1-4112	14800	32	4	701	60	2012-08-02 05:11:07.143361
14802	3	1-3163	14800	4	94	553	42	2012-08-02 05:11:07.143361
14800	3	1-4266	14800	94	32	301	45	2012-08-02 05:11:07.143361
14804	4	1-4097	14803	45	88	1026	52	2012-08-02 05:11:07.960665
14805	4	1-3558	14803	88	25	825	2	2012-08-02 05:11:07.960665
14806	4	1-4166	14803	25	37	466	8	2012-08-02 05:11:07.960665
14803	4	1-4268	14803	37	45	2169	76	2012-08-02 05:11:07.960665
14808	5	1-4207	14807	1	32	153	45	2012-08-02 05:11:08.358442
14809	5	1-4112	14807	32	54	345	60	2012-08-02 05:11:08.358442
14807	5	1-4269	14807	95	1	238	85	2012-08-02 05:11:08.358442
14810	5	1-305	14807	54	47	120	47	2012-08-02 05:11:08.358442
14811	5	1-4126	14807	47	95	255	33	2012-08-02 05:11:08.358442
16137	5	1-4435	16135	9	17	322	21	2012-08-02 05:13:40.056933
14813	5	1-4207	14812	1	32	195	45	2012-08-02 05:11:08.767462
14814	5	1-4112	14812	32	3	567	60	2012-08-02 05:11:08.767462
14815	5	1-4141	14812	3	43	865	72	2012-08-02 05:11:08.767462
14816	5	1-1718	14812	43	18	209	6	2012-08-02 05:11:08.767462
14812	5	1-4270	14812	18	1	234	85	2012-08-02 05:11:08.767462
16138	5	1-4140	16135	17	74	198	26	2012-08-02 05:13:40.056933
14818	4	1-4120	14817	57	56	751	81	2012-08-02 05:11:09.176099
14819	4	1-3778	14817	56	86	354	64	2012-08-02 05:11:09.176099
14820	4	1-2375	14817	86	2	565	31	2012-08-02 05:11:09.176099
14817	4	1-4271	14817	2	57	761	5	2012-08-02 05:11:09.176099
16139	5	1-3880	16135	74	12	266	39	2012-08-02 05:13:40.056933
14822	4	1-4053	14821	11	24	231	39	2012-08-02 05:11:09.596145
14823	4	1-3877	14821	24	86	117	42	2012-08-02 05:11:09.596145
14824	4	1-4161	14821	86	30	110	72	2012-08-02 05:11:09.596145
14821	4	1-4272	14821	30	11	258	59	2012-08-02 05:11:09.596145
14826	3	1-4084	14825	85	66	2337	51	2012-08-02 05:11:09.596145
14827	3	1-2414	14825	66	30	221	72	2012-08-02 05:11:09.596145
14825	3	1-4272	14825	30	85	655	59	2012-08-02 05:11:09.596145
14829	3	1-4206	14828	99	44	327	27	2012-08-02 05:11:09.596145
14830	3	1-3064	14828	44	30	57	72	2012-08-02 05:11:09.596145
14828	3	1-4272	14828	30	99	186	59	2012-08-02 05:11:09.596145
14832	5	1-1773	14831	20	84	792	90	2012-08-02 05:11:09.596145
14833	5	1-4142	14831	84	57	1043	21	2012-08-02 05:11:09.596145
14834	5	1-3575	14831	57	30	520	2	2012-08-02 05:11:09.596145
14835	5	1-4113	14831	30	30	783	72	2012-08-02 05:11:09.596145
14831	5	1-4272	14831	30	20	2759	59	2012-08-02 05:11:09.596145
14837	4	1-3236	14836	22	38	231	89	2012-08-02 05:11:09.596145
14838	4	1-2127	14836	38	30	332	2	2012-08-02 05:11:09.596145
14839	4	1-4113	14836	30	30	534	72	2012-08-02 05:11:09.596145
14836	4	1-4272	14836	30	22	2010	59	2012-08-02 05:11:09.596145
14841	4	1-1826	14840	1	37	827	25	2012-08-02 05:11:10.43567
14842	4	1-3623	14840	37	74	452	11	2012-08-02 05:11:10.43567
14843	4	1-3987	14840	74	46	284	55	2012-08-02 05:11:10.43567
14840	4	1-4274	14840	46	1	2113	34	2012-08-02 05:11:10.43567
14845	5	1-1826	14844	1	1	2211	25	2012-08-02 05:11:10.43567
14846	5	1-3916	14844	1	30	2651	83	2012-08-02 05:11:10.43567
14847	5	1-4096	14844	30	44	982	63	2012-08-02 05:11:10.43567
14848	5	1-3912	14844	44	46	631	55	2012-08-02 05:11:10.43567
14844	5	1-4274	14844	46	1	4731	34	2012-08-02 05:11:10.43567
14850	2	1-4268	14849	37	67	4190	76	2012-08-02 05:11:10.855766
14849	2	1-4275	14849	67	37	2976	8	2012-08-02 05:11:10.855766
14852	4	1-4269	14851	95	12	1611	85	2012-08-02 05:11:11.651126
14853	4	1-3310	14851	12	73	934	13	2012-08-02 05:11:11.651126
14854	4	1-4203	14851	73	99	513	44	2012-08-02 05:11:11.651126
14851	4	1-4277	14851	99	95	2235	33	2012-08-02 05:11:11.651126
14856	4	1-3034	14855	55	26	3608	28	2012-08-02 05:11:11.651126
14857	4	1-2871	14855	26	73	2152	13	2012-08-02 05:11:11.651126
14858	4	1-4203	14855	73	99	1287	44	2012-08-02 05:11:11.651126
14855	4	1-4277	14855	99	55	6115	33	2012-08-02 05:11:11.651126
14860	5	1-3987	14859	74	31	624	55	2012-08-02 05:11:14.468292
14861	5	1-3770	14859	31	14	316	32	2012-08-02 05:11:14.468292
14862	5	1-4248	14859	14	63	807	100	2012-08-02 05:11:14.468292
14863	5	1-3786	14859	63	18	652	58	2012-08-02 05:11:14.468292
14859	5	1-4284	14859	18	74	1011	11	2012-08-02 05:11:14.468292
14865	5	1-4155	14864	14	80	1357	86	2012-08-02 05:11:14.468292
14866	5	1-1812	14864	80	14	856	32	2012-08-02 05:11:14.468292
14867	5	1-4248	14864	14	63	2744	100	2012-08-02 05:11:14.468292
14868	5	1-3786	14864	63	18	2424	58	2012-08-02 05:11:14.468292
14864	5	1-4284	14864	18	14	4110	11	2012-08-02 05:11:14.468292
14870	5	1-3456	14869	54	82	704	15	2012-08-02 05:11:14.89939
14871	5	1-4037	14869	82	40	745	90	2012-08-02 05:11:14.89939
14872	5	1-3404	14869	40	69	392	70	2012-08-02 05:11:14.89939
14873	5	1-3815	14869	69	31	601	77	2012-08-02 05:11:14.89939
14869	5	1-4285	14869	31	54	2263	50	2012-08-02 05:11:14.89939
14875	5	1-933	14874	83	7	1749	2	2012-08-02 05:11:16.976527
14876	5	1-3699	14874	7	27	1532	76	2012-08-02 05:11:16.976527
14877	5	1-4232	14874	27	11	2595	73	2012-08-02 05:11:16.976527
14878	5	1-3812	14874	11	75	1260	33	2012-08-02 05:11:16.976527
14874	5	1-4290	14874	75	83	2239	60	2012-08-02 05:11:16.976527
14880	4	1-3796	14879	71	19	1660	69	2012-08-02 05:11:16.976527
14881	4	1-1341	14879	19	13	1771	14	2012-08-02 05:11:16.976527
14882	4	1-3909	14879	13	75	1009	33	2012-08-02 05:11:16.976527
14879	4	1-4290	14879	75	71	2001	60	2012-08-02 05:11:16.976527
14884	5	1-3796	14883	71	22	647	69	2012-08-02 05:11:16.976527
14885	5	1-1100	14883	22	81	183	22	2012-08-02 05:11:16.976527
14886	5	1-4132	14883	81	95	370	49	2012-08-02 05:11:16.976527
14887	5	1-3840	14883	95	75	356	33	2012-08-02 05:11:16.976527
14883	5	1-4290	14883	75	71	742	60	2012-08-02 05:11:16.976527
14889	4	1-3796	14888	71	86	902	69	2012-08-02 05:11:16.976527
14890	4	1-1281	14888	86	13	820	14	2012-08-02 05:11:16.976527
14891	4	1-3909	14888	13	75	493	33	2012-08-02 05:11:16.976527
14888	4	1-4290	14888	75	71	1031	60	2012-08-02 05:11:16.976527
14893	5	1-4264	14892	86	44	8429	27	2012-08-02 05:11:17.793868
14894	5	1-3704	14892	44	45	4612	68	2012-08-02 05:11:17.793868
14895	5	1-3842	14892	45	57	3763	65	2012-08-02 05:11:17.793868
14896	5	1-2592	14892	57	30	2299	25	2012-08-02 05:11:17.793868
14892	5	1-4292	14892	30	86	2877	24	2012-08-02 05:11:17.793868
14898	3	1-4200	14897	93	66	26	61	2012-08-02 05:11:21.086207
14899	3	1-3031	14897	66	62	31	51	2012-08-02 05:11:21.086207
14897	3	1-4300	14897	62	93	19	67	2012-08-02 05:11:21.086207
14901	5	1-4200	14900	93	6	3007	61	2012-08-02 05:11:21.086207
14902	5	1-3447	14900	6	40	1889	90	2012-08-02 05:11:21.086207
14903	5	1-3404	14900	40	80	1220	70	2012-08-02 05:11:21.086207
14904	5	1-3679	14900	80	62	2242	51	2012-08-02 05:11:21.086207
14900	5	1-4300	14900	62	93	1757	67	2012-08-02 05:11:21.086207
14906	5	1-2494	14905	44	82	1539	15	2012-08-02 05:11:21.086207
14907	5	1-4037	14905	82	40	2024	90	2012-08-02 05:11:21.086207
14908	5	1-3404	14905	40	80	1325	70	2012-08-02 05:11:21.086207
14909	5	1-3679	14905	80	62	2468	51	2012-08-02 05:11:21.086207
14905	5	1-4300	14905	62	44	1961	67	2012-08-02 05:11:21.086207
14911	4	1-2454	14910	3	42	735	79	2012-08-02 05:11:21.914763
14912	4	1-3924	14910	42	1	908	25	2012-08-02 05:11:21.914763
14910	4	1-4302	14910	38	3	1323	66	2012-08-02 05:11:21.914763
14913	4	1-3916	14910	1	38	1017	83	2012-08-02 05:11:21.914763
16135	5	1-4644	16135	12	31	130	92	2012-08-02 05:13:40.056933
14915	4	1-3555	14914	23	56	786	60	2012-08-02 05:11:21.914763
14916	4	1-1209	14914	56	34	396	80	2012-08-02 05:11:21.914763
14917	4	1-4294	14914	34	38	366	83	2012-08-02 05:11:21.914763
14914	4	1-4302	14914	38	23	490	66	2012-08-02 05:11:21.914763
14919	5	1-2105	14918	4	67	20	28	2012-08-02 05:11:21.914763
14920	5	1-4078	14918	67	69	17	58	2012-08-02 05:11:21.914763
14921	5	1-2668	14918	69	71	20	57	2012-08-02 05:11:21.914763
14922	5	1-3684	14918	71	38	22	83	2012-08-02 05:11:21.914763
14918	5	1-4302	14918	38	4	31	66	2012-08-02 05:11:21.914763
14924	5	1-1178	14923	21	25	1815	2	2012-08-02 05:11:21.914763
14925	5	1-4166	14923	25	6	1192	8	2012-08-02 05:11:21.914763
14926	5	1-3486	14923	6	71	1412	57	2012-08-02 05:11:21.914763
14927	5	1-3684	14923	71	38	1534	83	2012-08-02 05:11:21.914763
14923	5	1-4302	14923	38	21	2198	66	2012-08-02 05:11:21.914763
14929	5	1-3920	14928	46	27	3601	91	2012-08-02 05:11:21.914763
14930	5	1-595	14928	27	69	1929	58	2012-08-02 05:11:21.914763
14931	5	1-2668	14928	69	71	2370	57	2012-08-02 05:11:21.914763
14932	5	1-3684	14928	71	38	2751	83	2012-08-02 05:11:21.914763
14928	5	1-4302	14928	38	46	4211	66	2012-08-02 05:11:21.914763
14934	5	1-3544	14933	1	60	605	77	2012-08-02 05:11:21.914763
14935	5	1-4177	14933	60	47	381	43	2012-08-02 05:11:21.914763
14936	5	1-4130	14933	47	56	315	36	2012-08-02 05:11:21.914763
14937	5	1-3103	14933	56	38	421	83	2012-08-02 05:11:21.914763
14933	5	1-4302	14933	38	1	667	66	2012-08-02 05:11:21.914763
14938	1	1-2668	\N	69	69	164	57	2012-08-02 05:11:21.914763
14940	4	1-4107	14939	29	69	671	70	2012-08-02 05:11:22.335142
14941	4	1-3815	14939	69	21	1148	77	2012-08-02 05:11:22.335142
14942	4	1-3587	14939	21	47	487	37	2012-08-02 05:11:22.335142
14939	4	1-4303	14939	47	29	295	82	2012-08-02 05:11:22.335142
14944	5	1-4081	14943	49	22	251	69	2012-08-02 05:11:22.733128
14945	5	1-1100	14943	22	81	58	22	2012-08-02 05:11:22.733128
14946	5	1-4132	14943	81	10	95	49	2012-08-02 05:11:22.733128
14947	5	1-3193	14943	10	39	110	88	2012-08-02 05:11:22.733128
14943	5	1-4304	14943	39	49	295	87	2012-08-02 05:11:22.733128
14949	3	1-3573	14948	6	10	2056	49	2012-08-02 05:11:22.733128
14950	3	1-3193	14948	10	39	2525	88	2012-08-02 05:11:22.733128
14948	3	1-4304	14948	39	6	7170	87	2012-08-02 05:11:22.733128
14952	4	1-4055	14951	90	52	3811	78	2012-08-02 05:11:23.561707
14953	4	1-3405	14951	52	96	2288	7	2012-08-02 05:11:23.561707
14954	4	1-3626	14951	96	49	964	73	2012-08-02 05:11:23.561707
14951	4	1-4306	14951	49	90	2478	69	2012-08-02 05:11:23.561707
14956	4	1-3407	14955	39	98	134	40	2012-08-02 05:11:23.561707
14957	4	1-4250	14955	98	96	115	7	2012-08-02 05:11:23.561707
14958	4	1-3626	14955	96	49	49	73	2012-08-02 05:11:23.561707
14955	4	1-4306	14955	49	39	128	69	2012-08-02 05:11:23.561707
14960	4	1-2725	14959	47	46	2808	57	2012-08-02 05:11:23.561707
14961	4	1-2840	14959	46	69	699	83	2012-08-02 05:11:23.561707
14962	4	1-4020	14959	69	49	760	73	2012-08-02 05:11:23.561707
14959	4	1-4306	14959	49	47	2206	69	2012-08-02 05:11:23.561707
14964	5	1-3266	14963	51	41	565	34	2012-08-02 05:11:23.970397
14965	5	1-4052	14963	41	41	113	79	2012-08-02 05:11:23.970397
14966	5	1-4173	14963	41	68	714	100	2012-08-02 05:11:23.970397
14967	5	1-3020	14963	68	71	605	40	2012-08-02 05:11:23.970397
14963	5	1-4307	14963	71	51	705	18	2012-08-02 05:11:23.970397
14969	3	1-2698	14968	88	35	4986	32	2012-08-02 05:11:23.970397
14970	3	1-4058	14968	35	71	3360	40	2012-08-02 05:11:23.970397
14968	3	1-4307	14968	71	88	4567	18	2012-08-02 05:11:23.970397
14972	5	1-2682	14971	16	46	86	66	2012-08-02 05:11:23.970397
14973	5	1-3920	14971	46	20	59	91	2012-08-02 05:11:23.970397
14974	5	1-3675	14971	20	54	28	61	2012-08-02 05:11:23.970397
14975	5	1-4279	14971	54	71	120	40	2012-08-02 05:11:23.970397
14971	5	1-4307	14971	71	16	172	18	2012-08-02 05:11:23.970397
14977	5	1-4127	14976	28	83	1305	9	2012-08-02 05:11:23.970397
14978	5	1-3192	14976	83	73	987	93	2012-08-02 05:11:23.970397
14979	5	1-4283	14976	73	54	209	61	2012-08-02 05:11:23.970397
14980	5	1-4279	14976	54	71	935	40	2012-08-02 05:11:23.970397
14976	5	1-4307	14976	71	28	1391	18	2012-08-02 05:11:23.970397
14982	3	1-4269	14981	95	35	1873	85	2012-08-02 05:11:24.401631
14983	3	1-3711	14981	35	97	2447	78	2012-08-02 05:11:24.401631
14981	3	1-4308	14981	97	95	2935	33	2012-08-02 05:11:24.401631
14985	4	1-4290	14984	75	52	2631	60	2012-08-02 05:11:24.401631
14986	4	1-4029	14984	52	12	554	35	2012-08-02 05:11:24.401631
14987	4	1-2275	14984	12	97	761	78	2012-08-02 05:11:24.401631
14984	4	1-4308	14984	97	75	1452	33	2012-08-02 05:11:24.401631
14989	5	1-3034	14988	55	67	1365	28	2012-08-02 05:11:24.401631
14990	5	1-4078	14988	67	12	1306	58	2012-08-02 05:11:24.401631
14991	5	1-1486	14988	12	31	1146	51	2012-08-02 05:11:24.401631
14992	5	1-2952	14988	31	97	704	78	2012-08-02 05:11:24.401631
14988	5	1-4308	14988	97	55	1551	33	2012-08-02 05:11:24.401631
14994	4	1-4251	14993	95	55	2000	23	2012-08-02 05:11:26.069752
14995	4	1-4309	14993	55	65	1122	58	2012-08-02 05:11:26.069752
14996	4	1-3972	14993	65	72	809	13	2012-08-02 05:11:26.069752
14993	4	1-4312	14993	72	95	1978	99	2012-08-02 05:11:26.069752
14998	5	1-3891	14997	79	31	767	4	2012-08-02 05:11:26.467572
14999	5	1-2965	14997	31	83	711	12	2012-08-02 05:11:26.467572
15000	5	1-4245	14997	83	16	535	19	2012-08-02 05:11:26.467572
15001	5	1-4230	14997	16	23	765	3	2012-08-02 05:11:26.467572
14997	5	1-4313	14997	23	79	926	56	2012-08-02 05:11:26.467572
15003	5	1-2199	15002	77	18	538	15	2012-08-02 05:11:26.467572
15004	5	1-3806	15002	18	83	303	12	2012-08-02 05:11:26.467572
15005	5	1-4245	15002	83	16	245	19	2012-08-02 05:11:26.467572
15006	5	1-4230	15002	16	23	377	3	2012-08-02 05:11:26.467572
15002	5	1-4313	15002	23	77	491	56	2012-08-02 05:11:26.467572
15008	5	1-2344	15007	58	42	212	79	2012-08-02 05:11:26.467572
15009	5	1-3924	15007	42	46	296	25	2012-08-02 05:11:26.467572
15010	5	1-4163	15007	46	16	247	19	2012-08-02 05:11:26.467572
15011	5	1-4230	15007	16	23	402	3	2012-08-02 05:11:26.467572
15007	5	1-4313	15007	23	58	556	56	2012-08-02 05:11:26.467572
15013	4	1-3678	15012	90	89	822	71	2012-08-02 05:11:26.887509
15014	4	1-4099	15012	89	35	1180	36	2012-08-02 05:11:26.887509
15015	4	1-4032	15012	35	15	530	88	2012-08-02 05:11:26.887509
15012	4	1-4314	15012	15	90	2503	78	2012-08-02 05:11:26.887509
15017	5	1-2225	15016	96	21	494	54	2012-08-02 05:11:26.887509
15018	5	1-3848	15016	21	81	409	22	2012-08-02 05:11:26.887509
15019	5	1-4132	15016	81	10	610	49	2012-08-02 05:11:26.887509
15020	5	1-3193	15016	10	15	641	88	2012-08-02 05:11:26.887509
15016	5	1-4314	15016	15	96	3285	78	2012-08-02 05:11:26.887509
16141	5	1-4629	16140	8	45	754	9	2012-08-02 05:13:40.465888
15022	5	1-3603	15021	80	84	398	48	2012-08-02 05:11:26.887509
15023	5	1-3625	15021	84	28	325	93	2012-08-02 05:11:26.887509
15024	5	1-4027	15021	28	1	162	1	2012-08-02 05:11:26.887509
15025	5	1-4030	15021	1	15	238	88	2012-08-02 05:11:26.887509
15021	5	1-4314	15021	15	80	1263	78	2012-08-02 05:11:26.887509
15026	1	1-3848	\N	21	21	4081	22	2012-08-02 05:11:26.887509
16142	5	1-4397	16140	45	87	656	12	2012-08-02 05:13:40.465888
15028	4	1-4207	15027	1	32	857	45	2012-08-02 05:11:29.340321
15029	4	1-4112	15027	32	53	2109	60	2012-08-02 05:11:29.340321
15030	4	1-2785	15027	53	39	1433	7	2012-08-02 05:11:29.340321
15027	4	1-4320	15027	39	1	1216	85	2012-08-02 05:11:29.340321
16143	5	1-4637	16140	87	49	1081	87	2012-08-02 05:13:40.465888
15032	3	1-4217	15031	24	19	1381	9	2012-08-02 05:11:29.737945
15033	3	1-3315	15031	19	39	383	45	2012-08-02 05:11:29.737945
15031	3	1-4321	15031	39	24	7542	74	2012-08-02 05:11:29.737945
16144	5	1-3232	16140	49	75	889	16	2012-08-02 05:13:40.465888
15035	5	1-3996	15034	12	20	991	68	2012-08-02 05:11:30.544854
15036	5	1-2515	15034	20	46	216	55	2012-08-02 05:11:30.544854
15037	5	1-4274	15034	46	86	2212	34	2012-08-02 05:11:30.544854
15038	5	1-3190	15034	86	16	4004	16	2012-08-02 05:11:30.544854
15034	5	1-4323	15034	16	12	726	56	2012-08-02 05:11:30.544854
15040	5	1-4249	15039	5	23	1401	29	2012-08-02 05:11:31.792834
15041	5	1-3227	15039	23	51	921	3	2012-08-02 05:11:31.792834
15042	5	1-3466	15039	51	80	851	70	2012-08-02 05:11:31.792834
15043	5	1-3679	15039	80	41	1550	51	2012-08-02 05:11:31.792834
15039	5	1-4326	15039	41	5	1232	63	2012-08-02 05:11:31.792834
15045	5	1-3096	15044	83	83	1184	12	2012-08-02 05:11:31.792834
15046	5	1-4245	15044	83	74	1116	19	2012-08-02 05:11:31.792834
15047	5	1-3589	15044	74	54	1098	44	2012-08-02 05:11:31.792834
15048	5	1-3195	15044	54	41	1560	51	2012-08-02 05:11:31.792834
15044	5	1-4326	15044	41	83	1274	63	2012-08-02 05:11:31.792834
15050	5	1-4249	15049	5	59	2166	29	2012-08-02 05:11:31.792834
15051	5	1-3052	15049	59	82	2093	67	2012-08-02 05:11:31.792834
15052	5	1-4028	15049	82	45	2119	36	2012-08-02 05:11:31.792834
15053	5	1-4057	15049	45	41	2080	51	2012-08-02 05:11:31.792834
15049	5	1-4326	15049	41	5	1775	63	2012-08-02 05:11:31.792834
15055	3	1-3331	15054	30	42	4155	68	2012-08-02 05:11:32.214797
15056	3	1-4243	15054	42	37	5502	94	2012-08-02 05:11:32.214797
15054	3	1-4327	15054	37	30	7380	48	2012-08-02 05:11:32.214797
15058	5	1-4299	15057	33	42	835	38	2012-08-02 05:11:33.82785
15059	5	1-2798	15057	42	100	743	15	2012-08-02 05:11:33.82785
15060	5	1-3973	15057	100	97	941	50	2012-08-02 05:11:33.82785
15061	5	1-4159	15057	97	83	482	23	2012-08-02 05:11:33.82785
15057	5	1-4331	15057	83	33	1577	30	2012-08-02 05:11:33.82785
15063	4	1-4299	15062	33	72	3878	38	2012-08-02 05:11:33.82785
15064	4	1-1554	15062	72	95	1824	99	2012-08-02 05:11:33.82785
15065	4	1-4251	15062	95	83	2110	23	2012-08-02 05:11:33.82785
15062	4	1-4331	15062	83	33	7110	30	2012-08-02 05:11:33.82785
15067	4	1-4072	15066	30	90	1650	65	2012-08-02 05:11:35.076412
15068	4	1-4281	15066	90	94	1929	84	2012-08-02 05:11:35.076412
15069	4	1-2082	15066	94	82	1333	48	2012-08-02 05:11:35.076412
15066	4	1-4334	15066	82	30	1698	43	2012-08-02 05:11:35.076412
15071	5	1-4197	15070	75	29	162	82	2012-08-02 05:11:35.496064
15072	5	1-4107	15070	29	64	397	70	2012-08-02 05:11:35.496064
15073	5	1-4092	15070	64	39	699	7	2012-08-02 05:11:35.496064
15074	5	1-4320	15070	39	97	661	85	2012-08-02 05:11:35.496064
15070	5	1-4335	15070	97	75	223	47	2012-08-02 05:11:35.496064
15076	4	1-4134	15075	38	55	48	23	2012-08-02 05:11:35.905033
15077	4	1-4309	15075	55	65	39	58	2012-08-02 05:11:35.905033
15078	4	1-3972	15075	65	82	41	13	2012-08-02 05:11:35.905033
15075	4	1-4336	15075	82	38	47	62	2012-08-02 05:11:35.905033
15080	5	1-4134	15079	38	22	4135	23	2012-08-02 05:11:35.905033
15081	5	1-4211	15079	22	8	2034	9	2012-08-02 05:11:35.905033
15082	5	1-3760	15079	8	26	3637	28	2012-08-02 05:11:35.905033
15083	5	1-2871	15079	26	82	3413	13	2012-08-02 05:11:35.905033
15079	5	1-4336	15079	82	38	3944	62	2012-08-02 05:11:35.905033
15085	4	1-3535	15084	88	3	784	12	2012-08-02 05:11:36.325114
15086	4	1-4014	15084	3	12	416	56	2012-08-02 05:11:36.325114
15087	4	1-3996	15084	12	42	491	68	2012-08-02 05:11:36.325114
15084	4	1-4337	15084	42	88	2302	73	2012-08-02 05:11:36.325114
15089	5	1-4278	15088	41	8	3307	39	2012-08-02 05:11:36.325114
15090	5	1-3269	15088	8	23	338	3	2012-08-02 05:11:36.325114
15091	5	1-4313	15088	23	12	488	56	2012-08-02 05:11:36.325114
15092	5	1-3996	15088	12	42	635	68	2012-08-02 05:11:36.325114
15088	5	1-4337	15088	42	41	3286	73	2012-08-02 05:11:36.325114
15094	4	1-4246	15093	51	66	320	73	2012-08-02 05:11:36.744591
15095	4	1-706	15093	66	46	95	55	2012-08-02 05:11:36.744591
15096	4	1-4274	15093	46	1	875	34	2012-08-02 05:11:36.744591
15093	4	1-4338	15093	1	51	1396	76	2012-08-02 05:11:36.744591
15098	5	1-2275	15097	12	97	1373	78	2012-08-02 05:11:37.153576
15099	5	1-4308	15097	97	71	2783	33	2012-08-02 05:11:37.153576
15100	5	1-2741	15097	71	11	1177	58	2012-08-02 05:11:37.153576
15101	5	1-3790	15097	11	99	154	18	2012-08-02 05:11:37.153576
15097	5	1-4339	15097	99	12	940	35	2012-08-02 05:11:37.153576
15103	5	1-3252	15102	48	29	4847	40	2012-08-02 05:11:37.153576
15104	5	1-3971	15102	29	83	2457	52	2012-08-02 05:11:37.153576
15105	5	1-1404	15102	83	87	733	98	2012-08-02 05:11:37.153576
15106	5	1-4210	15102	87	99	812	18	2012-08-02 05:11:37.153576
15102	5	1-4339	15102	99	48	5195	35	2012-08-02 05:11:37.153576
15108	5	1-1904	15107	93	20	184	91	2012-08-02 05:11:37.153576
15109	5	1-3675	15107	20	54	113	61	2012-08-02 05:11:37.153576
15110	5	1-4279	15107	54	48	631	40	2012-08-02 05:11:37.153576
15111	5	1-3547	15107	48	99	53	18	2012-08-02 05:11:37.153576
15107	5	1-4339	15107	99	93	357	35	2012-08-02 05:11:37.153576
15113	5	1-2275	15112	12	90	2375	78	2012-08-02 05:11:37.153576
15114	5	1-3678	15112	90	77	1203	71	2012-08-02 05:11:37.153576
15115	5	1-4171	15112	77	34	602	19	2012-08-02 05:11:37.153576
15116	5	1-2861	15112	34	99	207	18	2012-08-02 05:11:37.153576
15112	5	1-4339	15112	99	12	1432	35	2012-08-02 05:11:37.153576
15118	5	1-791	15117	39	88	205	52	2012-08-02 05:11:37.982681
15117	5	1-4341	15117	20	39	532	8	2012-08-02 05:11:37.982681
15119	5	1-3558	15117	88	30	209	2	2012-08-02 05:11:37.982681
15120	5	1-4113	15117	30	30	337	72	2012-08-02 05:11:37.982681
15121	5	1-4272	15117	30	20	1275	59	2012-08-02 05:11:37.982681
15123	5	1-3651	15122	25	46	454	25	2012-08-02 05:11:39.219903
15124	5	1-4163	15122	46	70	403	19	2012-08-02 05:11:39.219903
15125	5	1-3428	15122	70	69	281	70	2012-08-02 05:11:39.219903
15126	5	1-3815	15122	69	10	533	77	2012-08-02 05:11:39.219903
15122	5	1-4344	15122	10	25	1996	50	2012-08-02 05:11:39.219903
15128	4	1-1595	15127	56	10	1653	77	2012-08-02 05:11:39.639552
15129	4	1-4344	15127	10	8	4667	50	2012-08-02 05:11:39.639552
15130	4	1-2055	15127	8	15	654	71	2012-08-02 05:11:39.639552
15127	4	1-4345	15127	15	56	2513	86	2012-08-02 05:11:39.639552
15132	5	1-3820	15131	63	76	1830	95	2012-08-02 05:11:39.639552
15133	5	1-2917	15131	76	88	831	87	2012-08-02 05:11:39.639552
15134	5	1-2924	15131	88	90	646	78	2012-08-02 05:11:39.639552
15135	5	1-3678	15131	90	15	257	71	2012-08-02 05:11:39.639552
15131	5	1-4345	15131	15	63	1073	86	2012-08-02 05:11:39.639552
15137	4	1-4089	15136	35	17	808	4	2012-08-02 05:11:39.639552
15138	4	1-2804	15136	17	13	799	54	2012-08-02 05:11:39.639552
15139	4	1-4164	15136	13	15	250	71	2012-08-02 05:11:39.639552
15136	4	1-4345	15136	15	35	1064	86	2012-08-02 05:11:39.639552
15141	5	1-4089	15140	35	84	1543	4	2012-08-02 05:11:39.639552
15142	5	1-3510	15140	84	31	1475	67	2012-08-02 05:11:39.639552
15143	5	1-3478	15140	31	1	659	9	2012-08-02 05:11:39.639552
15144	5	1-2474	15140	1	15	476	71	2012-08-02 05:11:39.639552
15140	5	1-4345	15140	15	35	2029	86	2012-08-02 05:11:39.639552
15145	1	1-1074	\N	61	61	9	4	2012-08-02 05:11:39.639552
15147	5	1-3589	15146	74	54	309	44	2012-08-02 05:11:41.275087
15148	5	1-3195	15146	54	41	420	51	2012-08-02 05:11:41.275087
15149	5	1-4326	15146	41	83	328	63	2012-08-02 05:11:41.275087
15150	5	1-3096	15146	83	71	285	12	2012-08-02 05:11:41.275087
15146	5	1-4349	15146	71	74	329	19	2012-08-02 05:11:41.275087
15152	3	1-3877	15151	24	86	133	42	2012-08-02 05:11:43.716353
15153	3	1-4161	15151	86	70	191	72	2012-08-02 05:11:43.716353
15151	3	1-4355	15151	70	24	171	39	2012-08-02 05:11:43.716353
15155	4	1-3784	15154	75	71	1218	10	2012-08-02 05:11:46.599823
15156	4	1-3048	15154	71	69	773	83	2012-08-02 05:11:46.599823
15157	4	1-4020	15154	69	68	780	73	2012-08-02 05:11:46.599823
15154	4	1-4362	15154	68	75	1466	41	2012-08-02 05:11:46.599823
15159	5	1-4212	15158	43	64	2656	34	2012-08-02 05:11:46.599823
15160	5	1-4343	15158	64	68	3583	24	2012-08-02 05:11:46.599823
15161	5	1-3158	15158	68	69	2543	83	2012-08-02 05:11:46.599823
15162	5	1-4020	15158	69	68	2376	73	2012-08-02 05:11:46.599823
15158	5	1-4362	15158	68	43	4609	41	2012-08-02 05:11:46.599823
15164	5	1-4020	15163	69	66	575	73	2012-08-02 05:11:46.997372
15165	5	1-1574	15163	66	21	111	24	2012-08-02 05:11:46.997372
15166	5	1-4063	15163	21	82	202	15	2012-08-02 05:11:46.997372
15167	5	1-4037	15163	82	24	263	90	2012-08-02 05:11:46.997372
15163	5	1-4363	15163	24	69	526	83	2012-08-02 05:11:46.997372
15169	5	1-4096	15168	30	20	2917	63	2012-08-02 05:11:46.997372
15170	5	1-4288	15168	20	44	2363	67	2012-08-02 05:11:46.997372
15171	5	1-2494	15168	44	82	1940	15	2012-08-02 05:11:46.997372
15172	5	1-4037	15168	82	24	2670	90	2012-08-02 05:11:46.997372
15168	5	1-4363	15168	24	30	5655	83	2012-08-02 05:11:46.997372
15174	5	1-4250	15173	98	22	3209	7	2012-08-02 05:11:48.21263
15175	5	1-4291	15173	22	13	2146	5	2012-08-02 05:11:48.21263
15176	5	1-4085	15173	13	62	1717	51	2012-08-02 05:11:48.21263
15177	5	1-4300	15173	62	11	1335	67	2012-08-02 05:11:48.21263
15173	5	1-4366	15173	11	98	3379	40	2012-08-02 05:11:48.21263
15179	5	1-3820	15178	63	9	424	95	2012-08-02 05:11:51.129193
15180	5	1-4259	15178	9	83	2140	53	2012-08-02 05:11:51.129193
15181	5	1-1462	15178	83	41	93	79	2012-08-02 05:11:51.129193
15182	5	1-4173	15178	41	22	531	100	2012-08-02 05:11:51.129193
15178	5	1-4373	15178	22	63	277	86	2012-08-02 05:11:51.129193
15184	4	1-3787	15183	100	41	829	34	2012-08-02 05:11:51.129193
15185	4	1-4052	15183	41	41	201	79	2012-08-02 05:11:51.129193
15186	4	1-4173	15183	41	22	1548	100	2012-08-02 05:11:51.129193
15183	4	1-4373	15183	22	100	1035	86	2012-08-02 05:11:51.129193
15188	5	1-3991	15187	53	94	1696	84	2012-08-02 05:11:51.129193
15189	5	1-2082	15187	94	84	1389	48	2012-08-02 05:11:51.129193
15190	5	1-3625	15187	84	11	1481	93	2012-08-02 05:11:51.129193
15191	5	1-4004	15187	11	22	1123	100	2012-08-02 05:11:51.129193
15187	5	1-4373	15187	22	53	787	86	2012-08-02 05:11:51.129193
15193	4	1-3599	15192	67	40	3719	21	2012-08-02 05:11:52.753707
15194	4	1-2529	15192	40	57	1025	23	2012-08-02 05:11:52.753707
15195	4	1-4129	15192	57	48	2295	16	2012-08-02 05:11:52.753707
15192	4	1-4377	15192	48	67	9049	92	2012-08-02 05:11:52.753707
15197	5	1-1425	15196	58	25	1129	2	2012-08-02 05:11:53.184688
15198	5	1-4166	15196	25	24	823	8	2012-08-02 05:11:53.184688
15199	5	1-2862	15196	24	45	722	68	2012-08-02 05:11:53.184688
15200	5	1-3842	15196	45	35	580	65	2012-08-02 05:11:53.184688
15196	5	1-4378	15196	35	58	1605	77	2012-08-02 05:11:53.184688
15202	5	1-1938	15201	4	50	257	14	2012-08-02 05:11:54.001868
15203	5	1-3400	15201	50	65	218	5	2012-08-02 05:11:54.001868
15204	5	1-4330	15201	65	34	565	57	2012-08-02 05:11:54.001868
15205	5	1-3813	15201	34	60	378	11	2012-08-02 05:11:54.001868
15201	5	1-4380	15201	60	4	146	96	2012-08-02 05:11:54.001868
15207	3	1-4218	15206	55	6	4885	87	2012-08-02 05:11:54.39977
15208	3	1-3573	15206	6	25	542	49	2012-08-02 05:11:54.39977
15206	3	1-4381	15206	25	55	9167	18	2012-08-02 05:11:54.39977
15210	4	1-3909	15209	13	95	544	33	2012-08-02 05:11:55.649591
15211	4	1-4269	15209	95	60	144	85	2012-08-02 05:11:55.649591
15212	4	1-2684	15209	60	57	20	7	2012-08-02 05:11:55.649591
15209	4	1-4384	15209	57	13	4024	14	2012-08-02 05:11:55.649591
15214	4	1-2655	15213	12	94	332	43	2012-08-02 05:11:55.649591
15215	4	1-3003	15213	94	97	29	8	2012-08-02 05:11:55.649591
15216	4	1-3557	15213	97	57	13	7	2012-08-02 05:11:55.649591
15213	4	1-4384	15213	57	12	4380	14	2012-08-02 05:11:55.649591
15218	5	1-3968	15217	25	39	3188	80	2012-08-02 05:11:58.067591
15219	5	1-782	15217	39	87	2353	88	2012-08-02 05:11:58.067591
15220	5	1-3810	15217	87	71	526	12	2012-08-02 05:11:58.067591
15221	5	1-4349	15217	71	53	495	19	2012-08-02 05:11:58.067591
15217	5	1-4390	15217	53	25	3112	50	2012-08-02 05:11:58.067591
15222	1	1-782	\N	39	39	289	88	2012-08-02 05:11:58.067591
15223	4	1-4391	15223	72	88	1733	73	2012-08-02 05:11:58.465321
15224	4	1-3535	15223	88	71	579	12	2012-08-02 05:11:58.465321
15225	4	1-4349	15223	71	91	604	19	2012-08-02 05:11:58.465321
15226	4	1-3605	15223	91	72	646	54	2012-08-02 05:11:58.465321
16140	5	1-4645	16140	75	8	1077	95	2012-08-02 05:13:40.465888
15228	4	1-1460	15227	78	71	1842	12	2012-08-02 05:11:58.465321
15229	4	1-4349	15227	71	91	1927	19	2012-08-02 05:11:58.465321
15230	4	1-3605	15227	91	72	2065	54	2012-08-02 05:11:58.465321
15227	4	1-4391	15227	72	78	5552	73	2012-08-02 05:11:58.465321
15232	5	1-3538	15231	72	64	556	34	2012-08-02 05:11:59.283308
15233	5	1-4343	15231	64	21	538	24	2012-08-02 05:11:59.283308
15234	5	1-4063	15231	21	70	601	15	2012-08-02 05:11:59.283308
15235	5	1-3029	15231	70	77	290	32	2012-08-02 05:11:59.283308
15231	5	1-4393	15231	77	72	995	35	2012-08-02 05:11:59.283308
16146	5	1-3603	16145	80	58	697	48	2012-08-02 05:13:40.874746
15237	2	1-1568	15236	57	77	1248	32	2012-08-02 05:11:59.283308
15236	2	1-4393	15236	77	57	4757	35	2012-08-02 05:11:59.283308
15239	6	1-2529	15238	40	57	199	23	2012-08-02 05:12:03.007891
15240	6	1-4129	15238	57	57	673	16	2012-08-02 05:12:03.007891
15241	6	1-4316	15238	57	57	716	29	2012-08-02 05:12:03.007891
15242	6	1-4121	15238	57	75	965	47	2012-08-02 05:12:03.007891
15243	6	1-4197	15238	75	82	739	82	2012-08-02 05:12:03.007891
15238	6	1-4402	15238	82	40	488	21	2012-08-02 05:12:03.007891
15245	3	1-4343	15244	64	68	2822	24	2012-08-02 05:12:04.22313
15246	3	1-3158	15244	68	13	771	83	2012-08-02 05:12:04.22313
15244	3	1-4405	15244	13	64	4867	34	2012-08-02 05:12:04.22313
15248	4	1-2938	15247	72	86	391	42	2012-08-02 05:12:04.620995
15249	4	1-4161	15247	86	30	554	72	2012-08-02 05:12:04.620995
15250	4	1-4272	15247	30	99	1969	59	2012-08-02 05:12:04.620995
15247	4	1-4406	15247	99	72	298	38	2012-08-02 05:12:04.620995
15252	5	1-4067	15251	11	3	7859	31	2012-08-02 05:12:05.041225
15253	5	1-3949	15251	3	87	4491	88	2012-08-02 05:12:05.041225
15254	5	1-3810	15251	87	71	1198	12	2012-08-02 05:12:05.041225
15255	5	1-4349	15251	71	56	1347	19	2012-08-02 05:12:05.041225
15251	5	1-4407	15251	56	11	2134	3	2012-08-02 05:12:05.041225
15257	5	1-3466	15256	51	68	335	70	2012-08-02 05:12:05.041225
15258	5	1-3001	15256	68	87	515	88	2012-08-02 05:12:05.041225
15259	5	1-3810	15256	87	71	153	12	2012-08-02 05:12:05.041225
15260	5	1-4349	15256	71	56	192	19	2012-08-02 05:12:05.041225
15256	5	1-4407	15256	56	51	339	3	2012-08-02 05:12:05.041225
15262	2	1-1774	15261	35	82	2314	44	2012-08-02 05:12:06.621868
15261	2	1-4411	15261	82	35	3360	56	2012-08-02 05:12:06.621868
15264	5	1-3996	15263	12	42	1301	68	2012-08-02 05:12:06.621868
15265	5	1-4243	15263	42	37	1358	94	2012-08-02 05:12:06.621868
15266	5	1-4327	15263	37	77	1437	48	2012-08-02 05:12:06.621868
15267	5	1-3965	15263	77	82	727	44	2012-08-02 05:12:06.621868
15263	5	1-4411	15263	82	12	1198	56	2012-08-02 05:12:06.621868
15269	4	1-3996	15268	12	16	1495	68	2012-08-02 05:12:06.621868
15270	4	1-2286	15268	16	73	1071	13	2012-08-02 05:12:06.621868
15271	4	1-4203	15268	73	82	800	44	2012-08-02 05:12:06.621868
15268	4	1-4411	15268	82	12	1348	56	2012-08-02 05:12:06.621868
15273	4	1-3563	15272	56	9	401	93	2012-08-02 05:12:06.621868
15274	4	1-3666	15272	9	73	260	13	2012-08-02 05:12:06.621868
15275	4	1-4203	15272	73	82	218	44	2012-08-02 05:12:06.621868
15272	4	1-4411	15272	82	56	411	56	2012-08-02 05:12:06.621868
15277	5	1-4378	15276	35	10	1427	77	2012-08-02 05:12:08.566287
15278	5	1-4344	15276	10	54	2622	50	2012-08-02 05:12:08.566287
15279	5	1-3456	15276	54	82	496	15	2012-08-02 05:12:08.566287
15280	5	1-4037	15276	82	50	319	90	2012-08-02 05:12:08.566287
15276	5	1-4416	15276	50	35	1070	65	2012-08-02 05:12:08.566287
15282	5	1-2592	15281	57	30	854	25	2012-08-02 05:12:08.566287
15283	5	1-4292	15281	30	21	631	24	2012-08-02 05:12:08.566287
15284	5	1-4063	15281	21	82	709	15	2012-08-02 05:12:08.566287
15285	5	1-4037	15281	82	50	567	90	2012-08-02 05:12:08.566287
15281	5	1-4416	15281	50	57	2366	65	2012-08-02 05:12:08.566287
15287	5	1-4037	15286	82	84	1978	90	2012-08-02 05:12:08.953247
15288	5	1-4142	15286	84	90	2724	21	2012-08-02 05:12:08.953247
15289	5	1-3553	15286	90	62	2135	51	2012-08-02 05:12:08.953247
15290	5	1-4300	15286	62	60	1539	67	2012-08-02 05:12:08.953247
15286	5	1-4417	15286	60	82	1658	15	2012-08-02 05:12:08.953247
15292	5	1-4037	15291	82	24	473	90	2012-08-02 05:12:08.953247
15293	5	1-3566	15291	24	21	452	88	2012-08-02 05:12:08.953247
15294	5	1-3583	15291	21	62	379	51	2012-08-02 05:12:08.953247
15295	5	1-4300	15291	62	60	302	67	2012-08-02 05:12:08.953247
15291	5	1-4417	15291	60	82	359	15	2012-08-02 05:12:08.953247
15297	5	1-4132	15296	81	47	94	49	2012-08-02 05:12:09.339916
15298	5	1-2610	15296	47	97	26	79	2012-08-02 05:12:09.339916
15299	5	1-3801	15296	97	82	20	15	2012-08-02 05:12:09.339916
15300	5	1-4037	15296	82	59	22	90	2012-08-02 05:12:09.339916
15296	5	1-4418	15296	59	81	54	22	2012-08-02 05:12:09.339916
15302	5	1-4016	15301	9	54	427	60	2012-08-02 05:12:09.339916
15303	5	1-305	15301	54	56	159	47	2012-08-02 05:12:09.339916
15304	5	1-3011	15301	56	82	96	15	2012-08-02 05:12:09.339916
15305	5	1-4037	15301	82	59	107	90	2012-08-02 05:12:09.339916
15301	5	1-4418	15301	59	9	263	22	2012-08-02 05:12:09.339916
15307	5	1-3979	15306	30	35	198	28	2012-08-02 05:12:09.339916
15308	5	1-4357	15306	35	60	102	67	2012-08-02 05:12:09.339916
15309	5	1-4417	15306	60	82	108	15	2012-08-02 05:12:09.339916
15310	5	1-4037	15306	82	59	126	90	2012-08-02 05:12:09.339916
15306	5	1-4418	15306	59	30	325	22	2012-08-02 05:12:09.339916
15311	1	1-4037	\N	82	82	22	90	2012-08-02 05:12:09.339916
15313	5	1-4011	15312	30	34	2213	26	2012-08-02 05:12:10.909044
15314	5	1-195	15312	34	11	554	3	2012-08-02 05:12:10.909044
15315	5	1-4067	15312	11	2	1683	31	2012-08-02 05:12:10.909044
15316	5	1-4271	15312	2	21	1964	5	2012-08-02 05:12:10.909044
15312	5	1-4422	15312	21	30	2229	4	2012-08-02 05:12:10.909044
15318	5	1-4011	15317	30	88	447	26	2012-08-02 05:12:10.909044
15319	5	1-2060	15317	88	86	162	64	2012-08-02 05:12:10.909044
15320	5	1-2375	15317	86	2	249	31	2012-08-02 05:12:10.909044
15321	5	1-4271	15317	2	21	322	5	2012-08-02 05:12:10.909044
15317	5	1-4422	15317	21	30	406	4	2012-08-02 05:12:10.909044
15323	5	1-2609	15322	55	84	1383	90	2012-08-02 05:12:10.909044
15324	5	1-4142	15322	84	53	1862	21	2012-08-02 05:12:10.909044
15325	5	1-4035	15322	53	28	1319	1	2012-08-02 05:12:10.909044
15326	5	1-3305	15322	28	21	1156	5	2012-08-02 05:12:10.909044
15322	5	1-4422	15322	21	55	1504	4	2012-08-02 05:12:10.909044
16147	5	1-4630	16145	58	52	1063	99	2012-08-02 05:13:40.874746
15328	3	1-4367	15327	66	28	2310	1	2012-08-02 05:12:10.909044
15329	3	1-3305	15327	28	21	2051	5	2012-08-02 05:12:10.909044
15327	3	1-4422	15327	21	66	2702	4	2012-08-02 05:12:10.909044
16148	5	1-4544	16145	52	49	2463	98	2012-08-02 05:13:40.874746
15331	3	1-4011	15330	30	50	3570	26	2012-08-02 05:12:10.909044
15332	3	1-3683	15330	50	21	2252	5	2012-08-02 05:12:10.909044
15330	3	1-4422	15330	21	30	3032	4	2012-08-02 05:12:10.909044
15333	1	1-404	\N	96	96	30	5	2012-08-02 05:12:10.909044
16149	5	1-3674	16145	49	61	1549	54	2012-08-02 05:13:40.874746
15335	5	1-3396	15334	75	42	226	43	2012-08-02 05:12:11.31814
15336	5	1-4412	15334	42	41	309	51	2012-08-02 05:12:11.31814
15337	5	1-4326	15334	41	5	218	63	2012-08-02 05:12:11.31814
15338	5	1-4249	15334	5	60	220	29	2012-08-02 05:12:11.31814
15334	5	1-4423	15334	60	75	527	97	2012-08-02 05:12:11.31814
15340	3	1-4156	15339	63	79	3708	64	2012-08-02 05:12:11.31814
15341	3	1-3863	15339	79	60	2012	29	2012-08-02 05:12:11.31814
15339	3	1-4423	15339	60	63	4945	97	2012-08-02 05:12:11.31814
15343	2	1-2529	15342	40	62	2691	23	2012-08-02 05:12:11.715846
15342	2	1-4424	15342	62	40	9071	21	2012-08-02 05:12:11.715846
15345	5	1-4299	15344	33	72	1126	38	2012-08-02 05:12:12.51113
15346	5	1-4088	15344	72	84	931	4	2012-08-02 05:12:12.51113
15347	5	1-3510	15344	84	32	1012	67	2012-08-02 05:12:12.51113
15348	5	1-1399	15344	32	74	800	32	2012-08-02 05:12:12.51113
15344	5	1-4426	15344	74	33	1778	30	2012-08-02 05:12:12.51113
15350	5	1-2932	15349	5	22	2885	100	2012-08-02 05:12:12.51113
15351	5	1-4373	15349	22	63	1948	86	2012-08-02 05:12:12.51113
15352	5	1-3820	15349	63	100	3863	95	2012-08-02 05:12:12.51113
15353	5	1-4385	15349	100	74	1527	32	2012-08-02 05:12:12.51113
15349	5	1-4426	15349	74	5	3395	30	2012-08-02 05:12:12.51113
15355	4	1-2932	15354	5	22	322	100	2012-08-02 05:12:12.51113
15356	4	1-4373	15354	22	80	220	86	2012-08-02 05:12:12.51113
15357	4	1-1812	15354	80	74	167	32	2012-08-02 05:12:12.51113
15354	4	1-4426	15354	74	5	375	30	2012-08-02 05:12:12.51113
15359	4	1-4207	15358	1	68	2501	45	2012-08-02 05:12:13.704601
15360	4	1-3816	15358	68	84	2544	90	2012-08-02 05:12:13.704601
15361	4	1-4142	15358	84	54	3258	21	2012-08-02 05:12:13.704601
15358	4	1-4429	15358	54	1	3652	85	2012-08-02 05:12:13.704601
15363	5	1-4397	15362	45	42	54	12	2012-08-02 05:12:14.897829
15364	5	1-4148	15362	42	78	29	92	2012-08-02 05:12:14.897829
15365	5	1-3441	15362	78	1	40	1	2012-08-02 05:12:14.897829
15366	5	1-4030	15362	1	14	66	88	2012-08-02 05:12:14.897829
15362	5	1-4432	15362	14	45	70	9	2012-08-02 05:12:14.897829
15368	4	1-3192	15367	83	28	1645	93	2012-08-02 05:12:14.897829
15369	4	1-4027	15367	28	1	983	1	2012-08-02 05:12:14.897829
15370	4	1-4030	15367	1	14	1733	88	2012-08-02 05:12:14.897829
15367	4	1-4432	15367	14	83	1946	9	2012-08-02 05:12:14.897829
15372	5	1-4397	15371	45	71	3104	12	2012-08-02 05:12:14.897829
15373	5	1-4349	15371	71	70	3498	19	2012-08-02 05:12:14.897829
15374	5	1-3428	15371	70	68	2303	70	2012-08-02 05:12:14.897829
15375	5	1-3001	15371	68	14	3182	88	2012-08-02 05:12:14.897829
15371	5	1-4432	15371	14	45	3699	9	2012-08-02 05:12:14.897829
15376	1	1-512	\N	4	4	1150	88	2012-08-02 05:12:14.897829
15378	5	1-3315	15377	19	68	2258	45	2012-08-02 05:12:16.842046
15379	5	1-3816	15377	68	59	1495	90	2012-08-02 05:12:16.842046
15380	5	1-4418	15377	59	9	2368	22	2012-08-02 05:12:16.842046
15381	5	1-4016	15377	9	24	2480	60	2012-08-02 05:12:16.842046
15377	5	1-4437	15377	24	19	7559	9	2012-08-02 05:12:16.842046
15383	4	1-4397	15382	45	71	439	12	2012-08-02 05:12:16.842046
15384	4	1-4349	15382	71	12	417	19	2012-08-02 05:12:16.842046
15385	4	1-1083	15382	12	24	141	60	2012-08-02 05:12:16.842046
15382	4	1-4437	15382	24	45	621	9	2012-08-02 05:12:16.842046
15387	5	1-2602	15386	77	51	43	73	2012-08-02 05:12:16.842046
15388	5	1-2185	15386	51	59	7	90	2012-08-02 05:12:16.842046
15389	5	1-4418	15386	59	9	19	22	2012-08-02 05:12:16.842046
15390	5	1-4016	15386	9	24	33	60	2012-08-02 05:12:16.842046
15386	5	1-4437	15386	24	77	167	9	2012-08-02 05:12:16.842046
15392	3	1-4134	15391	38	1	4029	23	2012-08-02 05:12:17.637768
15393	3	1-3946	15391	1	93	3402	10	2012-08-02 05:12:17.637768
15391	3	1-4439	15391	93	38	4254	62	2012-08-02 05:12:17.637768
15395	3	1-4292	15394	30	70	2678	24	2012-08-02 05:12:18.797669
15396	3	1-3950	15394	70	10	1991	47	2012-08-02 05:12:18.797669
15394	3	1-4442	15394	10	30	2279	25	2012-08-02 05:12:18.797669
15398	4	1-870	15397	47	16	154	19	2012-08-02 05:12:19.19552
15399	4	1-4230	15397	16	45	254	3	2012-08-02 05:12:19.19552
15400	4	1-4340	15397	45	86	696	37	2012-08-02 05:12:19.19552
15397	4	1-4443	15397	86	47	4395	83	2012-08-02 05:12:19.19552
15402	5	1-4098	15401	47	14	5918	9	2012-08-02 05:12:20.377749
15403	5	1-2519	15401	14	7	2305	81	2012-08-02 05:12:20.377749
15404	5	1-4394	15401	7	60	3285	11	2012-08-02 05:12:20.377749
15405	5	1-4380	15401	60	99	1251	96	2012-08-02 05:12:20.377749
15401	5	1-4446	15401	99	47	2259	89	2012-08-02 05:12:20.377749
15407	4	1-2193	15406	50	15	2578	68	2012-08-02 05:12:20.377749
15408	4	1-4328	15406	15	60	4233	11	2012-08-02 05:12:20.377749
15409	4	1-4380	15406	60	99	1640	96	2012-08-02 05:12:20.377749
15406	4	1-4446	15406	99	50	3012	89	2012-08-02 05:12:20.377749
15411	4	1-3816	15410	68	59	1635	90	2012-08-02 05:12:21.570873
15412	4	1-4418	15410	59	91	4450	22	2012-08-02 05:12:21.570873
15413	4	1-2919	15410	91	45	1115	1	2012-08-02 05:12:21.570873
15410	4	1-4449	15410	45	68	1436	45	2012-08-02 05:12:21.570873
15415	3	1-2241	15414	72	57	514	16	2012-08-02 05:12:23.128493
15416	3	1-4316	15414	57	10	67	29	2012-08-02 05:12:23.128493
15414	3	1-4453	15414	10	72	4191	18	2012-08-02 05:12:23.128493
15418	5	1-3752	15417	69	53	2787	86	2012-08-02 05:12:23.526698
15419	5	1-3991	15417	53	86	4826	84	2012-08-02 05:12:23.526698
15420	5	1-2575	15417	86	11	1665	23	2012-08-02 05:12:23.526698
15421	5	1-3746	15417	11	8	1560	80	2012-08-02 05:12:23.526698
15417	5	1-4454	15417	8	69	4197	89	2012-08-02 05:12:23.526698
15423	5	1-4448	15422	73	98	1280	53	2012-08-02 05:12:23.526698
15424	5	1-1901	15422	98	25	371	2	2012-08-02 05:12:23.526698
15425	5	1-4166	15422	25	55	213	8	2012-08-02 05:12:23.526698
15426	5	1-4240	15422	55	8	196	80	2012-08-02 05:12:23.526698
15422	5	1-4454	15422	8	73	531	89	2012-08-02 05:12:23.526698
15428	5	1-3392	15427	83	86	271	42	2012-08-02 05:12:23.526698
15429	5	1-4161	15427	86	47	330	72	2012-08-02 05:12:23.526698
15430	5	1-2420	15427	47	11	100	23	2012-08-02 05:12:23.526698
15431	5	1-3746	15427	11	8	97	80	2012-08-02 05:12:23.526698
15427	5	1-4454	15427	8	83	274	89	2012-08-02 05:12:23.526698
16145	5	1-4646	16145	61	80	1781	78	2012-08-02 05:13:40.874746
15433	2	1-3991	15432	53	88	2211	84	2012-08-02 05:12:24.31087
15432	2	1-4456	15432	88	53	1427	86	2012-08-02 05:12:24.31087
15435	4	1-4089	15434	35	66	3427	4	2012-08-02 05:12:24.31087
15436	4	1-4367	15434	66	75	3254	1	2012-08-02 05:12:24.31087
15437	4	1-4421	15434	75	88	4158	84	2012-08-02 05:12:24.31087
15434	4	1-4456	15434	88	35	3770	86	2012-08-02 05:12:24.31087
15439	5	1-3787	15438	100	64	474	34	2012-08-02 05:12:24.31087
15440	5	1-4343	15438	64	21	777	24	2012-08-02 05:12:24.31087
15441	5	1-4063	15438	21	28	1473	15	2012-08-02 05:12:24.31087
15442	5	1-3110	15438	28	88	581	84	2012-08-02 05:12:24.31087
15438	5	1-4456	15438	88	100	542	86	2012-08-02 05:12:24.31087
15444	4	1-4423	15443	60	96	2231	97	2012-08-02 05:12:24.708904
15445	4	1-4356	15443	96	60	737	13	2012-08-02 05:12:24.708904
15446	4	1-4389	15443	60	54	353	21	2012-08-02 05:12:24.708904
15443	4	1-4457	15443	54	60	967	29	2012-08-02 05:12:24.708904
15448	5	1-3227	15447	23	4	535	3	2012-08-02 05:12:24.708904
15449	5	1-4117	15447	4	72	265	35	2012-08-02 05:12:24.708904
15450	5	1-3538	15447	72	9	187	34	2012-08-02 05:12:24.708904
15451	5	1-4435	15447	9	54	331	21	2012-08-02 05:12:24.708904
15447	5	1-4457	15447	54	23	1043	29	2012-08-02 05:12:24.708904
15453	5	1-3368	15452	62	20	1020	59	2012-08-02 05:12:24.708904
15454	5	1-4341	15452	20	24	415	8	2012-08-02 05:12:24.708904
15455	5	1-2862	15452	24	99	350	68	2012-08-02 05:12:24.708904
15456	5	1-4006	15452	99	54	278	21	2012-08-02 05:12:24.708904
15452	5	1-4457	15452	54	62	1117	29	2012-08-02 05:12:24.708904
15458	3	1-4433	15457	91	47	2075	49	2012-08-02 05:12:25.50476
15459	3	1-2610	15457	47	48	242	79	2012-08-02 05:12:25.50476
15457	3	1-4459	15457	48	91	8000	65	2012-08-02 05:12:25.50476
15461	5	1-3548	15460	5	9	2647	34	2012-08-02 05:12:26.288802
15462	5	1-4435	15460	9	57	6303	21	2012-08-02 05:12:26.288802
15463	5	1-3575	15460	57	91	3717	2	2012-08-02 05:12:26.288802
15464	5	1-4458	15460	91	77	3337	1	2012-08-02 05:12:26.288802
15460	5	1-4461	15460	77	5	2246	64	2012-08-02 05:12:26.288802
15466	5	1-4250	15465	98	39	1798	7	2012-08-02 05:12:27.450695
15467	5	1-4320	15465	39	1	1566	85	2012-08-02 05:12:27.450695
15468	5	1-4207	15465	1	27	1133	45	2012-08-02 05:12:27.450695
15469	5	1-3997	15465	27	56	660	97	2012-08-02 05:12:27.450695
15465	5	1-4464	15465	56	98	2080	40	2012-08-02 05:12:27.450695
15471	5	1-4251	15470	95	57	1334	23	2012-08-02 05:12:28.235096
15472	5	1-4129	15470	57	91	4157	16	2012-08-02 05:12:28.235096
15473	5	1-4368	15470	91	7	2322	26	2012-08-02 05:12:28.235096
15474	5	1-3974	15470	7	60	816	65	2012-08-02 05:12:28.235096
15470	5	1-4466	15470	60	95	986	99	2012-08-02 05:12:28.235096
15476	5	1-4129	15475	57	91	2371	16	2012-08-02 05:12:28.633231
15477	5	1-4368	15475	91	7	1135	26	2012-08-02 05:12:28.633231
15478	5	1-3974	15475	7	60	342	65	2012-08-02 05:12:28.633231
15479	5	1-4466	15475	60	22	354	99	2012-08-02 05:12:28.633231
15475	5	1-4467	15475	22	57	888	23	2012-08-02 05:12:28.633231
15481	4	1-4309	15480	55	55	172	58	2012-08-02 05:12:28.633231
15482	4	1-2101	15480	55	60	53	65	2012-08-02 05:12:28.633231
15483	4	1-4466	15480	60	22	68	99	2012-08-02 05:12:28.633231
15480	4	1-4467	15480	22	55	212	23	2012-08-02 05:12:28.633231
15485	5	1-3946	15484	1	41	2926	10	2012-08-02 05:12:28.633231
15486	5	1-4465	15484	41	1	3201	55	2012-08-02 05:12:28.633231
15487	5	1-1898	15484	1	60	766	65	2012-08-02 05:12:28.633231
15488	5	1-4466	15484	60	22	999	99	2012-08-02 05:12:28.633231
15484	5	1-4467	15484	22	1	3156	23	2012-08-02 05:12:28.633231
15490	4	1-3004	15489	71	75	1780	1	2012-08-02 05:12:29.041971
15491	4	1-4421	15489	75	88	1830	84	2012-08-02 05:12:29.041971
15492	4	1-4456	15489	88	8	1335	86	2012-08-02 05:12:29.041971
15489	4	1-4468	15489	8	71	4333	62	2012-08-02 05:12:29.041971
15494	5	1-2468	15493	63	45	339	68	2012-08-02 05:12:29.041971
15495	5	1-3842	15493	45	90	237	65	2012-08-02 05:12:29.041971
15496	5	1-4281	15493	90	88	293	84	2012-08-02 05:12:29.041971
15497	5	1-4456	15493	88	8	235	86	2012-08-02 05:12:29.041971
15493	5	1-4468	15493	8	63	836	62	2012-08-02 05:12:29.041971
15499	3	1-4446	15498	99	83	1630	89	2012-08-02 05:12:29.881785
15500	3	1-3392	15498	83	55	1146	42	2012-08-02 05:12:29.881785
15498	3	1-4470	15498	55	99	1563	96	2012-08-02 05:12:29.881785
15502	2	1-4441	15501	89	83	3606	54	2012-08-02 05:12:30.279439
15501	2	1-4471	15501	83	89	5474	84	2012-08-02 05:12:30.279439
15504	5	1-4456	15503	88	35	992	86	2012-08-02 05:12:30.279439
15505	5	1-4089	15503	35	30	769	4	2012-08-02 05:12:30.279439
15506	5	1-4011	15503	30	64	839	26	2012-08-02 05:12:30.279439
15507	5	1-3328	15503	64	83	613	54	2012-08-02 05:12:30.279439
15503	5	1-4471	15503	83	88	1284	84	2012-08-02 05:12:30.279439
15509	5	1-2082	15508	94	77	308	48	2012-08-02 05:12:30.279439
15510	5	1-3965	15508	77	82	168	44	2012-08-02 05:12:30.279439
15511	5	1-4411	15508	82	4	299	56	2012-08-02 05:12:30.279439
15512	5	1-3014	15508	4	83	194	54	2012-08-02 05:12:30.279439
15508	5	1-4471	15508	83	94	421	84	2012-08-02 05:12:30.279439
15514	4	1-4412	15513	42	87	799	51	2012-08-02 05:12:30.677156
15515	4	1-3772	15513	87	30	162	25	2012-08-02 05:12:30.677156
15516	4	1-4292	15513	30	86	191	24	2012-08-02 05:12:30.677156
15513	4	1-4472	15513	86	42	523	43	2012-08-02 05:12:30.677156
15518	2	1-1756	15517	62	12	99	63	2012-08-02 05:12:36.963181
15517	2	1-4488	15517	12	62	1489	78	2012-08-02 05:12:36.963181
15520	4	1-3521	15519	23	14	2056	11	2012-08-02 05:12:36.963181
15521	4	1-4155	15519	14	74	444	86	2012-08-02 05:12:36.963181
15522	4	1-4233	15519	74	12	226	63	2012-08-02 05:12:36.963181
15519	4	1-4488	15519	12	23	5787	78	2012-08-02 05:12:36.963181
15524	4	1-3632	15523	2	37	27	8	2012-08-02 05:12:37.361002
15525	4	1-4268	15523	37	51	167	76	2012-08-02 05:12:37.361002
15526	4	1-4246	15523	51	31	43	73	2012-08-02 05:12:37.361002
15523	4	1-4489	15523	31	2	21	98	2012-08-02 05:12:37.361002
15528	4	1-3997	15527	27	75	4847	97	2012-08-02 05:12:37.770056
15529	4	1-3396	15527	75	42	2466	43	2012-08-02 05:12:37.770056
15530	4	1-4412	15527	42	85	3989	51	2012-08-02 05:12:37.770056
15527	4	1-4490	15527	85	27	7207	45	2012-08-02 05:12:37.770056
15532	3	1-4358	15531	74	91	5337	8	2012-08-02 05:12:38.156484
15533	3	1-4123	15531	91	98	3982	17	2012-08-02 05:12:38.156484
15531	3	1-4491	15531	98	74	4926	66	2012-08-02 05:12:38.156484
15535	4	1-4358	15534	74	59	791	8	2012-08-02 05:12:38.156484
15536	4	1-1051	15534	59	56	538	12	2012-08-02 05:12:38.156484
15537	4	1-3838	15534	56	98	478	17	2012-08-02 05:12:38.156484
15534	4	1-4491	15534	98	74	657	66	2012-08-02 05:12:38.156484
16151	2	1-936	16150	48	39	1347	95	2012-08-02 05:13:42.111828
15539	5	1-4410	15538	87	22	757	100	2012-08-02 05:12:38.156484
15540	5	1-4373	15538	22	11	455	86	2012-08-02 05:12:38.156484
15541	5	1-2578	15538	11	55	271	2	2012-08-02 05:12:38.156484
15542	5	1-4253	15538	55	98	459	17	2012-08-02 05:12:38.156484
15538	5	1-4491	15538	98	87	673	66	2012-08-02 05:12:38.156484
16150	2	1-4649	16150	39	48	3151	11	2012-08-02 05:13:42.111828
15544	3	1-4083	15543	96	56	517	12	2012-08-02 05:12:38.156484
15545	3	1-3838	15543	56	98	591	17	2012-08-02 05:12:38.156484
15543	3	1-4491	15543	98	96	1044	66	2012-08-02 05:12:38.156484
15547	5	1-4458	15546	91	45	418	1	2012-08-02 05:12:38.55443
15548	5	1-4449	15546	45	27	435	45	2012-08-02 05:12:38.55443
15549	5	1-3997	15546	27	28	217	97	2012-08-02 05:12:38.55443
15550	5	1-3529	15546	28	79	28	96	2012-08-02 05:12:38.55443
15546	5	1-4492	15546	79	91	626	2	2012-08-02 05:12:38.55443
16153	4	1-4557	16152	99	15	1733	43	2012-08-02 05:13:42.531996
15552	4	1-4458	15551	91	68	210	1	2012-08-02 05:12:38.55443
15553	4	1-1115	15551	68	78	102	42	2012-08-02 05:12:38.55443
15554	4	1-3066	15551	78	79	12	96	2012-08-02 05:12:38.55443
15551	4	1-4492	15551	79	91	284	2	2012-08-02 05:12:38.55443
16154	4	1-3975	16152	15	27	633	97	2012-08-02 05:13:42.531996
15556	5	1-3618	15555	26	30	954	25	2012-08-02 05:12:38.952447
15557	5	1-4292	15555	30	9	1181	24	2012-08-02 05:12:38.952447
15558	5	1-3850	15555	9	9	1075	38	2012-08-02 05:12:38.952447
15559	5	1-3709	15555	9	50	560	94	2012-08-02 05:12:38.952447
15555	5	1-4493	15555	50	26	534	75	2012-08-02 05:12:38.952447
16155	4	1-4525	16152	27	21	579	81	2012-08-02 05:13:42.531996
15561	5	1-4140	15560	17	64	86	26	2012-08-02 05:12:39.338888
15562	5	1-3328	15560	64	23	68	54	2012-08-02 05:12:39.338888
15563	5	1-4047	15560	23	86	136	42	2012-08-02 05:12:39.338888
15564	5	1-4161	15560	86	82	184	72	2012-08-02 05:12:39.338888
15560	5	1-4494	15560	82	17	158	21	2012-08-02 05:12:39.338888
15566	5	1-4429	15565	54	29	2562	85	2012-08-02 05:12:39.338888
15567	5	1-4396	15565	29	24	1587	39	2012-08-02 05:12:39.338888
15568	5	1-3877	15565	24	86	1299	42	2012-08-02 05:12:39.338888
15569	5	1-4161	15565	86	82	1960	72	2012-08-02 05:12:39.338888
15565	5	1-4494	15565	82	54	1876	21	2012-08-02 05:12:39.338888
15570	1	1-3153	\N	15	15	50	72	2012-08-02 05:12:39.338888
15572	4	1-4075	15571	70	64	5289	27	2012-08-02 05:12:39.736796
15573	4	1-4346	15571	64	78	2671	26	2012-08-02 05:12:39.736796
15574	4	1-3742	15571	78	19	2656	97	2012-08-02 05:12:39.736796
15571	4	1-4495	15571	19	70	6969	79	2012-08-02 05:12:39.736796
15576	5	1-1680	15575	57	41	572	10	2012-08-02 05:12:39.736796
15577	5	1-4465	15575	41	51	583	55	2012-08-02 05:12:39.736796
15578	5	1-4482	15575	51	85	276	44	2012-08-02 05:12:39.736796
15579	5	1-3915	15575	85	19	242	97	2012-08-02 05:12:39.736796
15575	5	1-4495	15575	19	57	653	79	2012-08-02 05:12:39.736796
15581	4	1-1429	15580	75	91	360	16	2012-08-02 05:12:39.736796
15582	4	1-4368	15580	91	78	213	26	2012-08-02 05:12:39.736796
15583	4	1-3742	15580	78	19	230	97	2012-08-02 05:12:39.736796
15580	4	1-4495	15580	19	75	654	79	2012-08-02 05:12:39.736796
15585	4	1-3903	15584	60	42	1693	12	2012-08-02 05:12:41.294667
15586	4	1-4148	15584	42	31	720	92	2012-08-02 05:12:41.294667
15587	4	1-3556	15584	31	88	474	34	2012-08-02 05:12:41.294667
15584	4	1-4499	15584	88	60	6945	21	2012-08-02 05:12:41.294667
15589	4	1-4140	15588	17	17	1339	26	2012-08-02 05:12:41.294667
15590	4	1-211	15588	17	100	313	86	2012-08-02 05:12:41.294667
15591	4	1-3787	15588	100	88	181	34	2012-08-02 05:12:41.294667
15588	4	1-4499	15588	88	17	2687	21	2012-08-02 05:12:41.294667
15593	5	1-59	15592	3	87	1855	68	2012-08-02 05:12:42.079104
15594	5	1-4360	15592	87	50	1112	94	2012-08-02 05:12:42.079104
15595	5	1-4493	15592	50	17	992	75	2012-08-02 05:12:42.079104
15596	5	1-3207	15592	17	74	1203	41	2012-08-02 05:12:42.079104
15592	5	1-4501	15592	74	3	3632	61	2012-08-02 05:12:42.079104
15598	4	1-2802	15597	100	76	442	99	2012-08-02 05:12:42.874592
15599	4	1-3337	15597	76	7	218	47	2012-08-02 05:12:42.874592
15600	4	1-4427	15597	7	86	119	6	2012-08-02 05:12:42.874592
15597	4	1-4503	15597	86	100	1702	71	2012-08-02 05:12:42.874592
15602	4	1-1524	15601	88	79	781	69	2012-08-02 05:12:44.045482
15603	4	1-2104	15601	79	65	639	5	2012-08-02 05:12:44.045482
15604	4	1-4330	15601	65	52	1573	57	2012-08-02 05:12:44.045482
15601	4	1-4506	15601	52	88	6091	46	2012-08-02 05:12:44.045482
15606	5	1-4466	15605	60	22	1178	99	2012-08-02 05:12:44.432493
15607	5	1-4467	15605	22	22	3125	23	2012-08-02 05:12:44.432493
15608	5	1-4211	15605	22	85	1267	9	2012-08-02 05:12:44.432493
15609	5	1-4297	15605	85	17	1075	95	2012-08-02 05:12:44.432493
15605	5	1-4507	15605	17	60	1307	65	2012-08-02 05:12:44.432493
15611	5	1-4281	15610	90	94	220	84	2012-08-02 05:12:44.432493
15612	5	1-2082	15610	94	84	163	48	2012-08-02 05:12:44.432493
15613	5	1-3625	15610	84	48	157	93	2012-08-02 05:12:44.432493
15614	5	1-4131	15610	48	17	116	95	2012-08-02 05:12:44.432493
15610	5	1-4507	15610	17	90	176	65	2012-08-02 05:12:44.432493
15616	4	1-3781	15615	49	45	222	1	2012-08-02 05:12:44.432493
15617	4	1-4449	15615	45	31	278	45	2012-08-02 05:12:44.432493
15618	4	1-3797	15615	31	17	203	95	2012-08-02 05:12:44.432493
15615	4	1-4507	15615	17	49	318	65	2012-08-02 05:12:44.432493
15620	5	1-4466	15619	60	76	1181	99	2012-08-02 05:12:44.432493
15621	5	1-3884	15619	76	84	829	48	2012-08-02 05:12:44.432493
15622	5	1-3625	15619	84	48	831	93	2012-08-02 05:12:44.432493
15623	5	1-4131	15619	48	17	638	95	2012-08-02 05:12:44.432493
15619	5	1-4507	15619	17	60	1008	65	2012-08-02 05:12:44.432493
15625	5	1-4466	15624	60	49	4028	99	2012-08-02 05:12:44.432493
15626	5	1-3272	15624	49	59	3010	21	2012-08-02 05:12:44.432493
15627	5	1-1129	15624	59	48	2442	93	2012-08-02 05:12:44.432493
15628	5	1-4131	15624	48	17	1969	95	2012-08-02 05:12:44.432493
15624	5	1-4507	15624	17	60	3272	65	2012-08-02 05:12:44.432493
15630	5	1-4378	15629	35	60	2622	77	2012-08-02 05:12:44.432493
15631	5	1-1264	15629	60	81	838	27	2012-08-02 05:12:44.432493
15632	5	1-2918	15629	81	48	732	93	2012-08-02 05:12:44.432493
15633	5	1-4131	15629	48	17	591	95	2012-08-02 05:12:44.432493
15629	5	1-4507	15629	17	35	984	65	2012-08-02 05:12:44.432493
15635	3	1-4469	15634	74	46	5515	69	2012-08-02 05:12:44.908109
15636	3	1-4351	15634	46	95	2510	15	2012-08-02 05:12:44.908109
15634	3	1-4508	15634	95	74	2500	82	2012-08-02 05:12:44.908109
15638	4	1-4469	15637	74	11	997	69	2012-08-02 05:12:44.908109
15639	4	1-2506	15637	11	81	325	64	2012-08-02 05:12:44.908109
15640	4	1-1501	15637	81	95	242	15	2012-08-02 05:12:44.908109
15637	4	1-4508	15637	95	74	330	82	2012-08-02 05:12:44.908109
15642	5	1-2942	15641	46	77	313	84	2012-08-02 05:12:44.908109
15643	5	1-1642	15641	77	30	180	25	2012-08-02 05:12:44.908109
15644	5	1-4292	15641	30	21	226	24	2012-08-02 05:12:44.908109
15645	5	1-4063	15641	21	95	433	15	2012-08-02 05:12:44.908109
15641	5	1-4508	15641	95	46	611	82	2012-08-02 05:12:44.908109
15647	4	1-2725	15646	47	52	369	57	2012-08-02 05:12:47.272179
15648	4	1-4506	15646	52	81	1112	46	2012-08-02 05:12:47.272179
15649	4	1-2048	15646	81	84	569	53	2012-08-02 05:12:47.272179
15646	4	1-4514	15646	84	47	371	69	2012-08-02 05:12:47.272179
15651	5	1-4104	15650	55	17	408	95	2012-08-02 05:12:47.658858
15652	5	1-4507	15650	17	49	702	65	2012-08-02 05:12:47.658858
15653	5	1-3781	15650	49	77	538	1	2012-08-02 05:12:47.658858
15654	5	1-4461	15650	77	1	356	64	2012-08-02 05:12:47.658858
15650	5	1-4515	15650	1	55	557	94	2012-08-02 05:12:47.658858
15656	5	1-4493	15655	50	5	331	75	2012-08-02 05:12:47.658858
15657	5	1-4293	15655	5	40	619	56	2012-08-02 05:12:47.658858
15658	5	1-4065	15655	40	56	414	81	2012-08-02 05:12:47.658858
15659	5	1-3778	15655	56	1	222	64	2012-08-02 05:12:47.658858
15655	5	1-4515	15655	1	50	349	94	2012-08-02 05:12:47.658858
15661	3	1-4408	15660	13	1	2082	1	2012-08-02 05:12:48.06776
15662	3	1-4030	15660	1	21	4109	88	2012-08-02 05:12:48.06776
15660	3	1-4516	15660	21	13	2463	73	2012-08-02 05:12:48.06776
15664	2	1-4368	15663	91	37	207	26	2012-08-02 05:12:48.852108
15663	2	1-4518	15663	37	91	601	16	2012-08-02 05:12:48.852108
15666	5	1-4316	15665	57	22	3776	29	2012-08-02 05:12:48.852108
15667	5	1-4487	15665	22	60	4970	13	2012-08-02 05:12:48.852108
15668	5	1-4389	15665	60	17	2329	21	2012-08-02 05:12:48.852108
15669	5	1-4140	15665	17	37	1121	26	2012-08-02 05:12:48.852108
15665	5	1-4518	15665	37	57	4512	16	2012-08-02 05:12:48.852108
15671	5	1-3787	15670	100	64	1146	34	2012-08-02 05:12:49.238913
15672	5	1-4343	15670	64	21	953	24	2012-08-02 05:12:49.238913
15673	5	1-4063	15670	21	100	916	15	2012-08-02 05:12:49.238913
15674	5	1-3973	15670	100	65	1081	50	2012-08-02 05:12:49.238913
15670	5	1-4519	15670	65	100	1820	86	2012-08-02 05:12:49.238913
15676	4	1-4511	15675	42	48	3588	94	2012-08-02 05:12:49.238913
15677	4	1-4137	15675	48	100	1607	45	2012-08-02 05:12:49.238913
15678	4	1-4167	15675	100	65	1688	50	2012-08-02 05:12:49.238913
15675	4	1-4519	15675	65	42	3421	86	2012-08-02 05:12:49.238913
15680	5	1-4514	15679	84	67	543	69	2012-08-02 05:12:49.62569
15681	5	1-2285	15679	67	50	133	94	2012-08-02 05:12:49.62569
15682	5	1-4493	15679	50	17	125	75	2012-08-02 05:12:49.62569
15683	5	1-3207	15679	17	79	160	41	2012-08-02 05:12:49.62569
15679	5	1-4520	15679	79	84	637	53	2012-08-02 05:12:49.62569
15685	5	1-3713	15684	12	47	1175	56	2012-08-02 05:12:50.034625
15686	5	1-4296	15684	47	98	3912	40	2012-08-02 05:12:50.034625
15687	5	1-4250	15684	98	96	3223	7	2012-08-02 05:12:50.034625
15688	5	1-3626	15684	96	86	1326	73	2012-08-02 05:12:50.034625
15684	5	1-4521	15684	86	12	2493	62	2012-08-02 05:12:50.034625
15690	5	1-2172	15689	2	94	1619	42	2012-08-02 05:12:50.034625
15691	5	1-4266	15689	94	77	1087	45	2012-08-02 05:12:50.034625
15692	5	1-3373	15689	77	59	471	30	2012-08-02 05:12:50.034625
15693	5	1-2485	15689	59	86	579	73	2012-08-02 05:12:50.034625
15689	5	1-4521	15689	86	2	1298	62	2012-08-02 05:12:50.034625
15695	5	1-3651	15694	25	30	718	25	2012-08-02 05:12:50.830074
15696	5	1-4292	15694	30	86	888	24	2012-08-02 05:12:50.830074
15697	5	1-4472	15694	86	15	2544	43	2012-08-02 05:12:50.830074
15698	5	1-3975	15694	15	22	1375	97	2012-08-02 05:12:50.830074
15694	5	1-4523	15694	22	25	3090	50	2012-08-02 05:12:50.830074
15700	5	1-4210	15699	87	55	697	18	2012-08-02 05:12:52.034621
15701	5	1-4218	15699	55	3	1110	87	2012-08-02 05:12:52.034621
15702	5	1-4388	15699	3	97	1006	52	2012-08-02 05:12:52.034621
15703	5	1-4440	15699	97	9	764	92	2012-08-02 05:12:52.034621
15699	5	1-4526	15699	9	87	610	98	2012-08-02 05:12:52.034621
15705	4	1-2468	15704	63	87	1302	68	2012-08-02 05:12:52.421062
15706	4	1-4360	15704	87	37	827	94	2012-08-02 05:12:52.421062
15707	4	1-4327	15704	37	92	1082	48	2012-08-02 05:12:52.421062
15704	4	1-4527	15704	92	63	2800	62	2012-08-02 05:12:52.421062
15709	2	1-3725	15708	95	10	1141	31	2012-08-02 05:12:52.820695
15708	2	1-4528	15708	10	95	5166	93	2012-08-02 05:12:52.820695
15711	5	1-4184	15710	46	82	164	44	2012-08-02 05:12:54.046988
15712	5	1-4411	15710	82	96	268	56	2012-08-02 05:12:54.046988
15713	5	1-3902	15710	96	52	289	57	2012-08-02 05:12:54.046988
15714	5	1-4506	15710	52	20	914	46	2012-08-02 05:12:54.046988
15710	5	1-4531	15710	20	46	506	81	2012-08-02 05:12:54.046988
15716	5	1-4361	15715	53	9	294	24	2012-08-02 05:12:54.444689
15717	5	1-3850	15715	9	86	235	38	2012-08-02 05:12:54.444689
15718	5	1-1821	15715	86	69	207	70	2012-08-02 05:12:54.444689
15719	5	1-3815	15715	69	30	359	77	2012-08-02 05:12:54.444689
15715	5	1-4532	15715	30	53	499	74	2012-08-02 05:12:54.444689
15721	5	1-3907	15720	34	11	1206	35	2012-08-02 05:12:54.444689
15722	5	1-4480	15720	11	48	1810	86	2012-08-02 05:12:54.444689
15723	5	1-1665	15720	48	69	1125	70	2012-08-02 05:12:54.444689
15724	5	1-3815	15720	69	30	2145	77	2012-08-02 05:12:54.444689
15720	5	1-4532	15720	30	34	3285	74	2012-08-02 05:12:54.444689
15726	5	1-4361	15725	53	6	2129	24	2012-08-02 05:12:54.444689
15727	5	1-3326	15725	6	51	1094	3	2012-08-02 05:12:54.444689
15728	5	1-3466	15725	51	69	1056	70	2012-08-02 05:12:54.444689
15729	5	1-3815	15725	69	30	2056	77	2012-08-02 05:12:54.444689
15725	5	1-4532	15725	30	53	3214	74	2012-08-02 05:12:54.444689
15731	5	1-4525	15730	27	46	789	81	2012-08-02 05:12:54.887207
15732	5	1-4184	15730	46	82	328	44	2012-08-02 05:12:54.887207
15733	5	1-4411	15730	82	56	682	56	2012-08-02 05:12:54.887207
15734	5	1-3563	15730	56	30	732	93	2012-08-02 05:12:54.887207
15730	5	1-4533	15730	30	27	576	97	2012-08-02 05:12:54.887207
15736	5	1-4347	15735	41	37	6509	28	2012-08-02 05:12:55.284483
15737	5	1-4150	15735	37	30	6109	93	2012-08-02 05:12:55.284483
15738	5	1-4533	15735	30	85	3875	97	2012-08-02 05:12:55.284483
15735	5	1-4534	15735	84	41	8005	13	2012-08-02 05:12:55.284483
15739	5	1-4522	15735	85	84	1840	91	2012-08-02 05:12:55.284483
16152	4	1-4650	16152	21	99	579	12	2012-08-02 05:13:42.531996
15741	5	1-4389	15740	60	17	267	21	2012-08-02 05:12:55.284483
15742	5	1-4140	15740	17	78	155	26	2012-08-02 05:12:55.284483
15743	5	1-3742	15740	78	85	163	97	2012-08-02 05:12:55.284483
15744	5	1-4522	15740	85	84	92	91	2012-08-02 05:12:55.284483
15740	5	1-4534	15740	84	60	473	13	2012-08-02 05:12:55.284483
15746	4	1-4361	15745	53	21	1797	24	2012-08-02 05:12:56.113135
15747	4	1-4063	15745	21	100	2569	15	2012-08-02 05:12:56.113135
15748	4	1-3973	15745	100	10	3172	50	2012-08-02 05:12:56.113135
15745	4	1-4536	15745	10	53	3584	74	2012-08-02 05:12:56.113135
15750	5	1-4361	15749	53	86	1581	24	2012-08-02 05:12:56.113135
15751	5	1-4472	15749	86	15	3543	43	2012-08-02 05:12:56.113135
15752	5	1-3975	15749	15	22	1498	97	2012-08-02 05:12:56.113135
15753	5	1-4523	15749	22	10	2635	50	2012-08-02 05:12:56.113135
15749	5	1-4536	15749	10	53	3064	74	2012-08-02 05:12:56.113135
15755	5	1-4217	15754	24	1	278	9	2012-08-02 05:12:56.113135
15756	5	1-2474	15754	1	77	215	71	2012-08-02 05:12:56.113135
15757	5	1-4171	15754	77	53	92	19	2012-08-02 05:12:56.113135
15758	5	1-4390	15754	53	10	661	50	2012-08-02 05:12:56.113135
15754	5	1-4536	15754	10	24	864	74	2012-08-02 05:12:56.113135
15760	5	1-3689	15759	57	12	47	58	2012-08-02 05:12:57.273451
15761	5	1-1486	15759	12	62	40	51	2012-08-02 05:12:57.273451
15762	5	1-4300	15759	62	82	30	67	2012-08-02 05:12:57.273451
15763	5	1-4028	15759	82	37	28	36	2012-08-02 05:12:57.273451
15759	5	1-4539	15759	37	57	64	1	2012-08-02 05:12:57.273451
15765	5	1-2523	15764	58	47	255	43	2012-08-02 05:12:58.058132
15766	5	1-4130	15764	47	69	215	36	2012-08-02 05:12:58.058132
15767	5	1-4473	15764	69	68	102	70	2012-08-02 05:12:58.058132
15768	5	1-3001	15764	68	25	154	88	2012-08-02 05:12:58.058132
15764	5	1-4541	15764	25	58	382	56	2012-08-02 05:12:58.058132
15770	5	1-3902	15769	96	83	4202	57	2012-08-02 05:12:58.058132
15771	5	1-1148	15769	83	41	935	10	2012-08-02 05:12:58.058132
15772	5	1-4465	15769	41	47	1030	55	2012-08-02 05:12:58.058132
15773	5	1-1912	15769	47	25	1207	88	2012-08-02 05:12:58.058132
15769	5	1-4541	15769	25	96	3021	56	2012-08-02 05:12:58.058132
15775	5	1-4065	15774	40	95	866	81	2012-08-02 05:12:58.058132
15776	5	1-3808	15774	95	21	570	21	2012-08-02 05:12:58.058132
15777	5	1-3051	15774	21	10	336	49	2012-08-02 05:12:58.058132
15778	5	1-3193	15774	10	25	502	88	2012-08-02 05:12:58.058132
15774	5	1-4541	15774	25	40	1259	56	2012-08-02 05:12:58.058132
15780	5	1-4210	15779	87	93	1689	18	2012-08-02 05:12:59.250974
15781	5	1-3342	15779	93	37	1511	41	2012-08-02 05:12:59.250974
15782	5	1-4517	15779	37	77	2167	70	2012-08-02 05:12:59.250974
15783	5	1-4382	15779	77	52	609	99	2012-08-02 05:12:59.250974
15779	5	1-4544	15779	52	87	1483	98	2012-08-02 05:12:59.250974
15785	4	1-2085	15784	61	78	2013	42	2012-08-02 05:12:59.637879
15786	4	1-3066	15784	78	79	227	96	2012-08-02 05:12:59.637879
15787	4	1-4492	15784	79	53	5468	2	2012-08-02 05:12:59.637879
15784	4	1-4545	15784	53	61	1831	61	2012-08-02 05:12:59.637879
15789	5	1-2461	15788	44	52	99	99	2012-08-02 05:12:59.637879
15790	5	1-4544	15788	52	58	221	98	2012-08-02 05:12:59.637879
15791	5	1-2783	15788	58	79	67	96	2012-08-02 05:12:59.637879
15792	5	1-4492	15788	79	53	1755	2	2012-08-02 05:12:59.637879
15788	5	1-4545	15788	53	44	636	61	2012-08-02 05:12:59.637879
15794	5	1-3447	15793	6	59	324	90	2012-08-02 05:12:59.637879
15795	5	1-4418	15793	59	9	886	22	2012-08-02 05:12:59.637879
15796	5	1-4016	15793	9	83	1604	60	2012-08-02 05:12:59.637879
15797	5	1-933	15793	83	53	1406	2	2012-08-02 05:12:59.637879
15793	5	1-4545	15793	53	6	535	61	2012-08-02 05:12:59.637879
15799	5	1-4225	15798	28	99	659	37	2012-08-02 05:13:00.433604
15800	5	1-3303	15798	99	92	383	48	2012-08-02 05:13:00.433604
15801	5	1-4527	15798	92	2	736	62	2012-08-02 05:13:00.433604
15802	5	1-2172	15798	2	34	680	42	2012-08-02 05:13:00.433604
15798	5	1-4547	15798	34	28	875	27	2012-08-02 05:13:00.433604
15804	4	1-3704	15803	44	66	900	68	2012-08-02 05:13:00.433604
15805	4	1-4484	15803	66	23	746	54	2012-08-02 05:13:00.433604
15806	4	1-4047	15803	23	34	1369	42	2012-08-02 05:13:00.433604
15803	4	1-4547	15803	34	44	1995	27	2012-08-02 05:13:00.433604
15808	5	1-2918	15807	81	99	1209	93	2012-08-02 05:13:00.433604
15809	5	1-2412	15807	99	92	438	48	2012-08-02 05:13:00.433604
15810	5	1-4527	15807	92	2	975	62	2012-08-02 05:13:00.433604
15811	5	1-2172	15807	2	34	1043	42	2012-08-02 05:13:00.433604
15807	5	1-4547	15807	34	81	1555	27	2012-08-02 05:13:00.433604
15813	5	1-4311	15812	11	96	2116	7	2012-08-02 05:13:00.433604
15814	5	1-3626	15812	96	86	919	73	2012-08-02 05:13:00.433604
15815	5	1-4521	15812	86	2	1823	62	2012-08-02 05:13:00.433604
15816	5	1-2172	15812	2	34	2012	42	2012-08-02 05:13:00.433604
15812	5	1-4547	15812	34	11	3096	27	2012-08-02 05:13:00.433604
15818	5	1-2918	15817	81	9	605	93	2012-08-02 05:13:00.433604
15819	5	1-3666	15817	9	82	390	13	2012-08-02 05:13:00.433604
15820	5	1-4336	15817	82	2	399	62	2012-08-02 05:13:00.433604
15821	5	1-2172	15817	2	34	456	42	2012-08-02 05:13:00.433604
15817	5	1-4547	15817	34	81	728	27	2012-08-02 05:13:00.433604
15823	3	1-4408	15822	13	60	145	1	2012-08-02 05:13:02.024505
15824	3	1-3261	15822	60	73	76	10	2012-08-02 05:13:02.024505
15822	3	1-4551	15822	73	13	177	73	2012-08-02 05:13:02.024505
15826	5	1-4408	15825	13	75	1798	1	2012-08-02 05:13:02.024505
15827	5	1-4421	15825	75	86	2318	84	2012-08-02 05:13:02.024505
15828	5	1-2575	15825	86	1	1014	23	2012-08-02 05:13:02.024505
15829	5	1-3946	15825	1	73	908	10	2012-08-02 05:13:02.024505
15825	5	1-4551	15825	73	13	2150	73	2012-08-02 05:13:02.024505
15831	5	1-2802	15830	100	52	603	99	2012-08-02 05:13:03.217633
15832	5	1-4544	15830	52	2	1322	98	2012-08-02 05:13:03.217633
15833	5	1-3632	15830	2	24	1465	8	2012-08-02 05:13:03.217633
15834	5	1-2862	15830	24	80	1130	68	2012-08-02 05:13:03.217633
15830	5	1-4554	15830	80	100	1467	71	2012-08-02 05:13:03.217633
15836	5	1-4412	15835	42	62	296	51	2012-08-02 05:13:04.377756
15837	5	1-4300	15835	62	31	194	67	2012-08-02 05:13:04.377756
15838	5	1-3478	15835	31	45	86	9	2012-08-02 05:13:04.377756
15839	5	1-4397	15835	45	99	63	12	2012-08-02 05:13:04.377756
15835	5	1-4557	15835	99	42	228	43	2012-08-02 05:13:04.377756
15841	5	1-3003	15840	94	37	201	8	2012-08-02 05:13:04.377756
15842	5	1-4268	15840	37	19	1215	76	2012-08-02 05:13:04.377756
15843	5	1-4074	15840	19	45	209	9	2012-08-02 05:13:04.377756
15844	5	1-4397	15840	45	99	190	12	2012-08-02 05:13:04.377756
15840	5	1-4557	15840	99	94	846	43	2012-08-02 05:13:04.377756
15846	4	1-2375	15845	86	95	597	31	2012-08-02 05:13:05.552814
15847	4	1-2623	15845	95	86	97	73	2012-08-02 05:13:05.552814
15848	4	1-4521	15845	86	51	178	62	2012-08-02 05:13:05.552814
15845	4	1-4560	15845	51	86	401	64	2012-08-02 05:13:05.552814
15850	3	1-1641	15849	12	82	3301	13	2012-08-02 05:13:05.552814
15851	3	1-4336	15849	82	51	2501	62	2012-08-02 05:13:05.552814
15849	3	1-4560	15849	51	12	5894	64	2012-08-02 05:13:05.552814
15853	5	1-1501	15852	81	95	1155	15	2012-08-02 05:13:05.552814
15854	5	1-4508	15852	95	50	1287	82	2012-08-02 05:13:05.552814
15855	5	1-2459	15852	50	93	630	10	2012-08-02 05:13:05.552814
15856	5	1-4439	15852	93	51	678	62	2012-08-02 05:13:05.552814
15852	5	1-4560	15852	51	81	1896	64	2012-08-02 05:13:05.552814
16157	5	1-4589	16156	23	5	2740	63	2012-08-02 05:13:42.531996
16158	5	1-4249	16156	5	49	2684	29	2012-08-02 05:13:42.531996
16159	5	1-4553	16156	49	27	1408	97	2012-08-02 05:13:42.531996
16160	5	1-4525	16156	27	21	1581	81	2012-08-02 05:13:42.531996
16156	5	1-4650	16156	21	23	1938	12	2012-08-02 05:13:42.531996
16162	5	1-4295	16161	62	67	923	1	2012-08-02 05:13:42.95171
16163	5	1-4623	16161	67	47	1709	55	2012-08-02 05:13:42.95171
16164	5	1-1912	16161	47	3	1710	88	2012-08-02 05:13:42.95171
16165	5	1-4568	16161	3	56	1283	60	2012-08-02 05:13:42.95171
16161	5	1-4651	16161	56	62	829	98	2012-08-02 05:13:42.95171
16167	5	1-4239	16166	69	89	608	10	2012-08-02 05:13:43.360686
16168	5	1-3182	16166	89	11	277	49	2012-08-02 05:13:43.360686
16169	5	1-4643	16166	11	61	809	68	2012-08-02 05:13:43.360686
16170	5	1-4538	16166	61	36	2401	14	2012-08-02 05:13:43.360686
16166	5	1-4652	16166	36	69	865	32	2012-08-02 05:13:43.360686
16172	5	1-3059	16171	9	3	5643	88	2012-08-02 05:13:45.006756
16173	5	1-4568	16171	3	56	4516	60	2012-08-02 05:13:45.006756
16174	5	1-4651	16171	56	71	3110	98	2012-08-02 05:13:45.006756
16175	5	1-4444	16171	71	79	2765	74	2012-08-02 05:13:45.006756
16171	5	1-4656	16171	79	9	2153	38	2012-08-02 05:13:45.006756
16177	3	1-3059	16176	9	27	1738	88	2012-08-02 05:13:45.006756
16178	3	1-4289	16176	27	79	787	74	2012-08-02 05:13:45.006756
16176	3	1-4656	16176	79	9	637	38	2012-08-02 05:13:45.006756
16180	5	1-4593	16179	40	64	2246	42	2012-08-02 05:13:45.415398
16181	5	1-4185	16179	64	35	1077	28	2012-08-02 05:13:45.415398
16182	5	1-4588	16179	35	31	1828	92	2012-08-02 05:13:45.415398
16183	5	1-3556	16179	31	47	1814	34	2012-08-02 05:13:45.415398
16179	5	1-4657	16179	47	40	1897	31	2012-08-02 05:13:45.415398
16185	4	1-4611	16184	69	79	2802	74	2012-08-02 05:13:45.824255
16186	4	1-4656	16184	79	39	1471	38	2012-08-02 05:13:45.824255
16187	4	1-3021	16184	39	45	1242	37	2012-08-02 05:13:45.824255
16184	4	1-4658	16184	45	69	2124	96	2012-08-02 05:13:45.824255
16189	5	1-4228	16188	81	9	82	34	2012-08-02 05:13:45.824255
16190	5	1-4435	16188	9	17	146	21	2012-08-02 05:13:45.824255
16191	5	1-4140	16188	17	8	66	26	2012-08-02 05:13:45.824255
16192	5	1-4462	16188	8	45	120	37	2012-08-02 05:13:45.824255
16188	5	1-4658	16188	45	81	248	96	2012-08-02 05:13:45.824255
16194	4	1-4611	16193	69	34	3006	74	2012-08-02 05:13:45.824255
16195	4	1-3907	16193	34	67	906	35	2012-08-02 05:13:45.824255
16196	4	1-4257	16193	67	45	802	37	2012-08-02 05:13:45.824255
16193	4	1-4658	16193	45	69	1768	96	2012-08-02 05:13:45.824255
16198	5	1-4352	16197	60	91	1764	50	2012-08-02 05:13:45.824255
16199	5	1-953	16197	91	83	338	5	2012-08-02 05:13:45.824255
16200	5	1-4633	16197	83	45	171	3	2012-08-02 05:13:45.824255
16201	5	1-4340	16197	45	45	424	37	2012-08-02 05:13:45.824255
16197	5	1-4658	16197	45	60	980	96	2012-08-02 05:13:45.824255
16203	5	1-4446	16202	99	37	174	89	2012-08-02 05:13:45.824255
16204	5	1-2838	16202	37	70	58	79	2012-08-02 05:13:45.824255
16205	5	1-4075	16202	70	28	44	27	2012-08-02 05:13:45.824255
16206	5	1-4225	16202	28	45	42	37	2012-08-02 05:13:45.824255
16202	5	1-4658	16202	45	99	105	96	2012-08-02 05:13:45.824255
16208	5	1-4318	16207	57	79	1359	56	2012-08-02 05:13:45.824255
16209	5	1-4252	16207	79	83	660	5	2012-08-02 05:13:45.824255
16210	5	1-4633	16207	83	45	393	3	2012-08-02 05:13:45.824255
16211	5	1-4340	16207	45	45	1056	37	2012-08-02 05:13:45.824255
16207	5	1-4658	16207	45	57	2643	96	2012-08-02 05:13:45.824255
16213	5	1-4612	16212	67	52	5115	79	2012-08-02 05:13:46.244541
16214	5	1-4138	16212	52	82	2132	72	2012-08-02 05:13:46.244541
16215	5	1-4494	16212	82	54	1527	21	2012-08-02 05:13:46.244541
16216	5	1-4429	16212	54	36	1561	85	2012-08-02 05:13:46.244541
16212	5	1-4659	16212	36	67	2056	69	2012-08-02 05:13:46.244541
16218	4	1-4569	16217	2	82	254	67	2012-08-02 05:13:46.244541
16219	4	1-4028	16217	82	61	212	36	2012-08-02 05:13:46.244541
16220	4	1-4342	16217	61	36	217	85	2012-08-02 05:13:46.244541
16217	4	1-4659	16217	36	2	322	69	2012-08-02 05:13:46.244541
16222	5	1-4023	16221	23	8	217	95	2012-08-02 05:13:46.244541
16223	5	1-4629	16221	8	31	163	9	2012-08-02 05:13:46.244541
16224	5	1-3145	16221	31	95	146	33	2012-08-02 05:13:46.244541
16225	5	1-4269	16221	95	36	180	85	2012-08-02 05:13:46.244541
16221	5	1-4659	16221	36	23	322	69	2012-08-02 05:13:46.244541
16227	4	1-3600	16226	3	56	1115	19	2012-08-02 05:13:47.029058
16228	4	1-4407	16226	56	51	1809	3	2012-08-02 05:13:47.029058
16229	4	1-3466	16226	51	12	1640	70	2012-08-02 05:13:47.029058
16226	4	1-4661	16226	12	3	2157	91	2012-08-02 05:13:47.029058
16231	5	1-2953	16230	63	64	187	42	2012-08-02 05:13:47.029058
16232	5	1-4185	16230	64	35	88	28	2012-08-02 05:13:47.029058
16233	5	1-4588	16230	35	23	145	92	2012-08-02 05:13:47.029058
16234	5	1-4048	16230	23	12	115	70	2012-08-02 05:13:47.029058
16230	5	1-4661	16230	12	63	156	91	2012-08-02 05:13:47.029058
16236	2	1-3386	16235	53	100	324	97	2012-08-02 05:13:49.794874
16235	2	1-4668	16235	100	53	1127	25	2012-08-02 05:13:49.794874
16238	4	1-4623	16237	67	39	2523	55	2012-08-02 05:13:50.629918
16239	4	1-3550	16237	39	52	1941	52	2012-08-02 05:13:50.629918
16240	4	1-2153	16237	52	55	737	43	2012-08-02 05:13:50.629918
16237	4	1-4670	16237	55	67	1770	1	2012-08-02 05:13:50.629918
16242	4	1-4624	16241	8	3	1525	87	2012-08-02 05:13:50.629918
16243	4	1-4388	16241	3	52	1051	52	2012-08-02 05:13:50.629918
16244	4	1-2153	16241	52	55	440	43	2012-08-02 05:13:50.629918
16241	4	1-4670	16241	55	8	1167	1	2012-08-02 05:13:50.629918
16246	5	1-4623	16245	67	42	1341	55	2012-08-02 05:13:50.629918
16247	5	1-3929	16245	42	91	865	23	2012-08-02 05:13:50.629918
16248	5	1-4379	16245	91	86	127	24	2012-08-02 05:13:50.629918
16249	5	1-4472	16245	86	55	286	43	2012-08-02 05:13:50.629918
16245	5	1-4670	16245	55	67	804	1	2012-08-02 05:13:50.629918
16251	4	1-4461	16250	77	12	896	64	2012-08-02 05:13:50.629918
16252	4	1-1641	16250	12	98	733	13	2012-08-02 05:13:50.629918
16253	4	1-4059	16250	98	55	384	43	2012-08-02 05:13:50.629918
16250	4	1-4670	16250	55	77	1358	1	2012-08-02 05:13:50.629918
16255	4	1-3869	16254	90	63	3280	100	2012-08-02 05:13:54.319987
16256	4	1-3786	16254	63	46	2540	58	2012-08-02 05:13:54.319987
16257	4	1-4452	16254	46	66	866	44	2012-08-02 05:13:54.319987
16254	4	1-4679	16254	66	90	5979	6	2012-08-02 05:13:54.319987
16259	5	1-4317	16258	97	84	429	4	2012-08-02 05:13:54.319987
16260	5	1-3510	16258	84	8	353	67	2012-08-02 05:13:54.319987
16261	5	1-3691	16258	8	37	333	39	2012-08-02 05:13:54.319987
16262	5	1-3429	16258	37	66	101	44	2012-08-02 05:13:54.319987
16258	5	1-4679	16258	66	97	713	6	2012-08-02 05:13:54.319987
16264	4	1-3754	16263	40	62	1222	36	2012-08-02 05:13:54.319987
16265	4	1-4430	16263	62	37	1402	39	2012-08-02 05:13:54.319987
16266	4	1-3429	16263	37	66	425	44	2012-08-02 05:13:54.319987
16263	4	1-4679	16263	66	40	3015	6	2012-08-02 05:13:54.319987
16268	5	1-4313	16267	23	4	604	56	2012-08-02 05:13:55.150491
16269	5	1-3014	16267	4	23	136	54	2012-08-02 05:13:55.150491
16270	5	1-4047	16267	23	70	167	42	2012-08-02 05:13:55.150491
16271	5	1-3104	16267	70	89	29	33	2012-08-02 05:13:55.150491
16267	5	1-4681	16267	89	23	721	3	2012-08-02 05:13:55.150491
16273	3	1-1508	16272	31	11	1090	73	2012-08-02 05:13:55.150491
16274	3	1-3812	16272	11	89	245	33	2012-08-02 05:13:55.150491
16272	3	1-4681	16272	89	31	7799	3	2012-08-02 05:13:55.150491
16276	5	1-2843	16275	22	40	6583	31	2012-08-02 05:13:55.572734
16277	5	1-4593	16275	40	70	6292	42	2012-08-02 05:13:55.572734
16278	5	1-3104	16275	70	35	2937	33	2012-08-02 05:13:55.572734
16279	5	1-4608	16275	35	13	6119	83	2012-08-02 05:13:55.572734
16275	5	1-4682	16275	13	22	6532	74	2012-08-02 05:13:55.572734
16281	3	1-4613	16280	11	1	440	25	2012-08-02 05:13:55.572734
16282	3	1-3916	16280	1	13	510	83	2012-08-02 05:13:55.572734
16280	3	1-4682	16280	13	11	571	74	2012-08-02 05:13:55.572734
16284	5	1-1821	16283	86	12	2020	70	2012-08-02 05:13:57.585237
16285	5	1-4661	16283	12	3	1849	91	2012-08-02 05:13:57.585237
16286	5	1-3600	16283	3	53	665	19	2012-08-02 05:13:57.585237
16287	5	1-4390	16283	53	46	3555	50	2012-08-02 05:13:57.585237
16283	5	1-4687	16283	46	86	3111	38	2012-08-02 05:13:57.585237
16289	4	1-2938	16288	72	94	1989	42	2012-08-02 05:13:57.585237
16290	4	1-4266	16288	94	100	1111	45	2012-08-02 05:13:57.585237
16291	4	1-4167	16288	100	46	1570	50	2012-08-02 05:13:57.585237
16288	4	1-4687	16288	46	72	1732	38	2012-08-02 05:13:57.585237
16293	4	1-570	16292	48	73	1178	10	2012-08-02 05:13:57.585237
16294	4	1-4551	16292	73	50	2492	73	2012-08-02 05:13:57.585237
16295	4	1-2413	16292	50	46	1157	50	2012-08-02 05:13:57.585237
16292	4	1-4687	16292	46	48	1362	38	2012-08-02 05:13:57.585237
16297	5	1-3298	16296	72	23	912	54	2012-08-02 05:13:57.585237
16298	5	1-4047	16296	23	94	1890	42	2012-08-02 05:13:57.585237
16299	5	1-4266	16296	94	100	1202	45	2012-08-02 05:13:57.585237
16300	5	1-4167	16296	100	46	1933	50	2012-08-02 05:13:57.585237
16296	5	1-4687	16296	46	72	2428	38	2012-08-02 05:13:57.585237
16302	4	1-4624	16301	8	3	567	87	2012-08-02 05:13:58.00761
16303	4	1-4388	16301	3	88	526	52	2012-08-02 05:13:58.00761
16304	4	1-3558	16301	88	94	545	2	2012-08-02 05:13:58.00761
16301	4	1-4688	16301	94	8	323	1	2012-08-02 05:13:58.00761
16306	3	1-4372	16305	84	39	1527	7	2012-08-02 05:13:58.394296
16307	3	1-4320	16305	39	42	1043	85	2012-08-02 05:13:58.394296
16305	3	1-4689	16305	42	84	2715	52	2012-08-02 05:13:58.394296
16309	2	1-4225	16308	28	100	7035	37	2012-08-02 05:13:58.781256
16308	2	1-4690	16308	100	28	7124	27	2012-08-02 05:13:58.781256
16311	4	1-4601	16310	74	53	203	64	2012-08-02 05:13:59.599451
16312	4	1-1658	16310	53	100	172	37	2012-08-02 05:13:59.599451
16313	4	1-4690	16310	100	19	170	27	2012-08-02 05:13:59.599451
16310	4	1-4692	16310	19	74	185	41	2012-08-02 05:13:59.599451
16315	3	1-4653	16314	46	42	6711	32	2012-08-02 05:14:03.433183
16316	3	1-3881	16314	42	4	2520	66	2012-08-02 05:14:03.433183
16314	3	1-4702	16314	4	46	4862	65	2012-08-02 05:14:03.433183
16318	2	1-3239	16317	58	31	3770	17	2012-08-02 05:14:03.819956
16317	2	1-4703	16317	31	58	4609	62	2012-08-02 05:14:03.819956
16320	5	1-3713	16319	12	56	130	56	2012-08-02 05:14:03.819956
16321	5	1-3563	16319	56	52	129	93	2012-08-02 05:14:03.819956
16322	5	1-4276	16319	52	55	95	2	2012-08-02 05:14:03.819956
16323	5	1-4253	16319	55	31	181	17	2012-08-02 05:14:03.819956
16319	5	1-4703	16319	31	12	245	62	2012-08-02 05:14:03.819956
16325	5	1-3106	16324	69	24	1553	5	2012-08-02 05:14:03.819956
16326	5	1-4618	16324	24	52	1412	93	2012-08-02 05:14:03.819956
16327	5	1-4276	16324	52	55	1098	2	2012-08-02 05:14:03.819956
16328	5	1-4253	16324	55	31	2223	17	2012-08-02 05:14:03.819956
16324	5	1-4703	16324	31	69	3198	62	2012-08-02 05:14:03.819956
16330	4	1-4642	16329	74	28	400	11	2012-08-02 05:14:04.219187
16331	4	1-4635	16329	28	86	266	84	2012-08-02 05:14:04.219187
16332	4	1-2575	16329	86	24	102	23	2012-08-02 05:14:04.219187
16329	4	1-4704	16329	24	74	529	17	2012-08-02 05:14:04.219187
16334	3	1-4392	16333	2	93	2779	23	2012-08-02 05:14:04.994695
16335	3	1-803	16333	93	24	351	89	2012-08-02 05:14:04.994695
16333	3	1-4706	16333	24	2	9258	33	2012-08-02 05:14:04.994695
16337	5	1-868	16336	88	86	178	24	2012-08-02 05:14:05.408713
16338	5	1-4472	16336	86	55	353	43	2012-08-02 05:14:05.408713
16339	5	1-4670	16336	55	6	875	1	2012-08-02 05:14:05.408713
16340	5	1-3945	16336	6	79	1010	40	2012-08-02 05:14:05.408713
16336	5	1-4707	16336	79	88	561	61	2012-08-02 05:14:05.408713
16342	3	1-59	16341	3	51	638	68	2012-08-02 05:14:05.408713
16343	3	1-4371	16341	51	79	2031	40	2012-08-02 05:14:05.408713
16341	3	1-4707	16341	79	3	1380	61	2012-08-02 05:14:05.408713
16345	5	1-3447	16344	6	59	141	90	2012-08-02 05:14:05.408713
16346	5	1-4418	16344	59	66	369	22	2012-08-02 05:14:05.408713
16347	5	1-4509	16344	66	68	285	100	2012-08-02 05:14:05.408713
16348	5	1-3020	16344	68	79	343	40	2012-08-02 05:14:05.408713
16344	5	1-4707	16344	79	6	245	61	2012-08-02 05:14:05.408713
16350	5	1-868	16349	88	20	384	24	2012-08-02 05:14:05.408713
16351	5	1-4597	16349	20	43	1131	56	2012-08-02 05:14:05.408713
16352	5	1-4370	16349	43	68	842	100	2012-08-02 05:14:05.408713
16353	5	1-3020	16349	68	79	1108	40	2012-08-02 05:14:05.408713
16349	5	1-4707	16349	79	88	864	61	2012-08-02 05:14:05.408713
16355	5	1-4675	16354	25	21	1593	21	2012-08-02 05:14:05.85144
16356	5	1-3051	16354	21	11	938	49	2012-08-02 05:14:05.85144
16357	5	1-4643	16354	11	66	2838	68	2012-08-02 05:14:05.85144
16358	5	1-4484	16354	66	62	2887	54	2012-08-02 05:14:05.85144
16354	5	1-4708	16354	62	25	1583	30	2012-08-02 05:14:05.85144
16360	3	1-2513	16359	59	59	1716	32	2012-08-02 05:14:06.674676
16361	3	1-2766	16359	59	85	1134	16	2012-08-02 05:14:06.674676
16359	3	1-4710	16359	85	59	2072	98	2012-08-02 05:14:06.674676
16363	4	1-4444	16362	71	11	3705	74	2012-08-02 05:14:06.674676
16364	4	1-4613	16362	11	81	2711	25	2012-08-02 05:14:06.674676
16365	4	1-3459	16362	81	85	2301	16	2012-08-02 05:14:06.674676
16362	4	1-4710	16362	85	71	4825	98	2012-08-02 05:14:06.674676
16367	4	1-3601	16366	65	31	1750	55	2012-08-02 05:14:06.674676
16368	4	1-3770	16366	31	59	1104	32	2012-08-02 05:14:06.674676
16369	4	1-2766	16366	59	85	766	16	2012-08-02 05:14:06.674676
16366	4	1-4710	16366	85	65	1787	98	2012-08-02 05:14:06.674676
16371	4	1-4517	16370	37	12	1176	70	2012-08-02 05:14:07.492441
16372	4	1-4661	16370	12	15	1380	91	2012-08-02 05:14:07.492441
16373	4	1-3921	16370	15	89	1345	81	2012-08-02 05:14:07.492441
16370	4	1-4712	16370	89	37	947	41	2012-08-02 05:14:07.492441
16375	5	1-4520	16374	79	70	6035	53	2012-08-02 05:14:07.492441
16376	5	1-4627	16374	70	31	4645	31	2012-08-02 05:14:07.492441
16377	5	1-4401	16374	31	27	1336	97	2012-08-02 05:14:07.492441
16378	5	1-4525	16374	27	89	1801	81	2012-08-02 05:14:07.492441
16374	5	1-4712	16374	89	79	1500	41	2012-08-02 05:14:07.492441
16380	4	1-4505	16379	55	98	2713	88	2012-08-02 05:14:07.492441
16381	4	1-4535	16379	98	27	2260	97	2012-08-02 05:14:07.492441
16382	4	1-4525	16379	27	89	3068	81	2012-08-02 05:14:07.492441
16379	4	1-4712	16379	89	55	2572	41	2012-08-02 05:14:07.492441
16384	5	1-2315	16383	27	85	447	97	2012-08-02 05:14:07.913196
16385	5	1-4522	16383	85	3	181	91	2012-08-02 05:14:07.913196
16386	5	1-3600	16383	3	53	69	19	2012-08-02 05:14:07.913196
16387	5	1-4390	16383	53	66	392	50	2012-08-02 05:14:07.913196
16383	5	1-4713	16383	66	27	1851	37	2012-08-02 05:14:07.913196
16389	5	1-4048	16388	23	12	1163	70	2012-08-02 05:14:08.310591
16390	5	1-4661	16388	12	74	1423	91	2012-08-02 05:14:08.310591
16391	5	1-2786	16388	74	61	1219	68	2012-08-02 05:14:08.310591
16392	5	1-4538	16388	61	76	3214	14	2012-08-02 05:14:08.310591
16388	5	1-4714	16388	76	23	1628	92	2012-08-02 05:14:08.310591
16394	5	1-3714	16393	73	37	759	36	2012-08-02 05:14:09.990021
16395	5	1-4539	16393	37	77	1350	1	2012-08-02 05:14:09.990021
16396	5	1-4461	16393	77	86	659	64	2012-08-02 05:14:09.990021
16397	5	1-2375	16393	86	100	877	31	2012-08-02 05:14:09.990021
16393	5	1-4718	16393	100	73	779	49	2012-08-02 05:14:09.990021
16399	4	1-3714	16398	73	6	389	36	2012-08-02 05:14:09.990021
16400	4	1-4548	16398	6	70	536	53	2012-08-02 05:14:09.990021
16401	4	1-4627	16398	70	100	346	31	2012-08-02 05:14:09.990021
16398	4	1-4718	16398	100	73	350	49	2012-08-02 05:14:09.990021
16403	5	1-3960	16402	59	83	4128	5	2012-08-02 05:14:09.990021
16404	5	1-4633	16402	83	4	2708	3	2012-08-02 05:14:09.990021
16405	5	1-4117	16402	4	23	1605	35	2012-08-02 05:14:09.990021
16406	5	1-2163	16402	23	100	2361	31	2012-08-02 05:14:09.990021
16402	5	1-4718	16402	100	59	2575	49	2012-08-02 05:14:09.990021
16408	4	1-2474	16407	1	38	3	71	2012-08-02 05:14:11.205548
16409	4	1-4691	16407	38	96	3	35	2012-08-02 05:14:11.205548
16410	4	1-1785	16407	96	62	2	74	2012-08-02 05:14:11.205548
16407	4	1-4721	16407	62	1	5	9	2012-08-02 05:14:11.205548
16412	3	1-3465	16411	93	23	5519	43	2012-08-02 05:14:11.205548
16413	3	1-4162	16411	23	62	2294	74	2012-08-02 05:14:11.205548
16411	3	1-4721	16411	62	93	6849	9	2012-08-02 05:14:11.205548
16415	5	1-3753	16414	97	41	244	10	2012-08-02 05:14:11.205548
16416	5	1-4465	16414	41	51	245	55	2012-08-02 05:14:11.205548
16417	5	1-1428	16414	51	36	114	90	2012-08-02 05:14:11.205548
16418	5	1-4719	16414	36	62	179	74	2012-08-02 05:14:11.205548
16414	5	1-4721	16414	62	97	569	9	2012-08-02 05:14:11.205548
16420	4	1-2412	16419	99	58	439	48	2012-08-02 05:14:13.260582
16421	4	1-4630	16419	58	76	410	99	2012-08-02 05:14:13.260582
16422	4	1-3337	16419	76	41	207	47	2012-08-02 05:14:13.260582
16419	4	1-4726	16419	41	99	1825	93	2012-08-02 05:14:13.260582
16424	5	1-4533	16423	30	87	1909	97	2012-08-02 05:14:13.260582
16425	5	1-4428	16423	87	39	1308	7	2012-08-02 05:14:13.260582
16426	5	1-4320	16423	39	97	995	85	2012-08-02 05:14:13.260582
16427	5	1-4335	16423	97	41	270	47	2012-08-02 05:14:13.260582
16423	5	1-4726	16423	41	30	3209	93	2012-08-02 05:14:13.260582
16429	5	1-3666	16428	9	60	968	13	2012-08-02 05:14:13.260582
16430	5	1-4389	16428	60	54	452	21	2012-08-02 05:14:13.260582
16431	5	1-4429	16428	54	97	488	85	2012-08-02 05:14:13.260582
16432	5	1-4335	16428	97	41	138	47	2012-08-02 05:14:13.260582
16428	5	1-4726	16428	41	9	1716	93	2012-08-02 05:14:13.260582
16434	5	1-4714	16433	76	52	371	92	2012-08-02 05:14:14.089556
16435	5	1-4025	16433	52	21	406	81	2012-08-02 05:14:14.089556
16436	5	1-4650	16433	21	23	541	12	2012-08-02 05:14:14.089556
16437	5	1-4589	16433	23	83	831	63	2012-08-02 05:14:14.089556
16433	5	1-4728	16433	83	76	714	14	2012-08-02 05:14:14.089556
16439	5	1-4652	16438	36	50	624	32	2012-08-02 05:14:14.089556
16440	5	1-4143	16438	50	21	690	81	2012-08-02 05:14:14.089556
16441	5	1-4650	16438	21	23	1029	12	2012-08-02 05:14:14.089556
16442	5	1-4589	16438	23	83	1768	63	2012-08-02 05:14:14.089556
16438	5	1-4728	16438	83	36	1699	14	2012-08-02 05:14:14.089556
16444	5	1-4110	16443	54	68	669	92	2012-08-02 05:14:15.304693
16445	5	1-2720	16443	68	99	825	93	2012-08-02 05:14:15.304693
16446	5	1-2412	16443	99	58	316	48	2012-08-02 05:14:15.304693
16447	5	1-4630	16443	58	15	469	99	2012-08-02 05:14:15.304693
16443	5	1-4731	16443	15	54	914	86	2012-08-02 05:14:15.304693
16449	4	1-2918	16448	81	48	2895	93	2012-08-02 05:14:15.713559
16450	4	1-4131	16448	48	8	1144	95	2012-08-02 05:14:15.713559
16451	4	1-4629	16448	8	99	392	9	2012-08-02 05:14:15.713559
16448	4	1-4732	16448	99	81	6781	27	2012-08-02 05:14:15.713559
16452	5	1-4737	16452	41	73	614	49	2012-08-02 05:14:17.770404
16453	5	1-3714	16452	73	37	575	36	2012-08-02 05:14:17.770404
16454	5	1-4539	16452	37	77	986	1	2012-08-02 05:14:17.770404
16455	5	1-4461	16452	77	87	463	64	2012-08-02 05:14:17.770404
16456	5	1-3885	16452	87	41	353	72	2012-08-02 05:14:17.770404
16458	5	1-3714	16457	73	62	2080	36	2012-08-02 05:14:17.770404
16459	5	1-4430	16457	62	62	3098	39	2012-08-02 05:14:17.770404
16460	5	1-4616	16457	62	87	1076	64	2012-08-02 05:14:17.770404
16461	5	1-3885	16457	87	41	950	72	2012-08-02 05:14:17.770404
16457	5	1-4737	16457	41	73	1916	49	2012-08-02 05:14:17.770404
16463	5	1-4390	16462	53	66	45	50	2012-08-02 05:14:18.179222
16464	5	1-4713	16462	66	30	139	37	2012-08-02 05:14:18.179222
16465	5	1-1723	16462	30	18	7	58	2012-08-02 05:14:18.179222
16466	5	1-4284	16462	18	3	7	11	2012-08-02 05:14:18.179222
16462	5	1-4738	16462	3	53	12	19	2012-08-02 05:14:18.179222
16468	4	1-4220	16467	27	75	929	16	2012-08-02 05:14:18.179222
16469	4	1-4645	16467	75	39	605	95	2012-08-02 05:14:18.179222
16470	4	1-4649	16467	39	3	757	11	2012-08-02 05:14:18.179222
16467	4	1-4738	16467	3	27	1487	19	2012-08-02 05:14:18.179222
16472	5	1-4390	16471	53	49	719	50	2012-08-02 05:14:18.179222
16473	5	1-2943	16471	49	53	65	59	2012-08-02 05:14:18.179222
16474	5	1-4705	16471	53	18	48	58	2012-08-02 05:14:18.179222
16475	5	1-4284	16471	18	3	57	11	2012-08-02 05:14:18.179222
16471	5	1-4738	16471	3	53	118	19	2012-08-02 05:14:18.179222
16477	4	1-1894	16476	89	22	3417	90	2012-08-02 05:14:18.179222
16478	4	1-4586	16476	22	38	2770	88	2012-08-02 05:14:18.179222
16479	4	1-4592	16476	38	3	1952	11	2012-08-02 05:14:18.179222
16476	4	1-4738	16476	3	89	4958	19	2012-08-02 05:14:18.179222
16481	4	1-2861	16480	34	23	358	18	2012-08-02 05:14:18.179222
16482	4	1-3274	16480	23	18	422	58	2012-08-02 05:14:18.179222
16483	4	1-4284	16480	18	3	615	11	2012-08-02 05:14:18.179222
16480	4	1-4738	16480	3	34	1583	19	2012-08-02 05:14:18.179222
16484	1	1-2861	\N	34	34	2125	18	2012-08-02 05:14:18.179222
16486	5	1-4630	16485	58	76	3717	99	2012-08-02 05:14:18.599673
16487	5	1-2015	16485	76	50	1721	32	2012-08-02 05:14:18.599673
16488	5	1-4143	16485	50	21	1834	81	2012-08-02 05:14:18.599673
16489	5	1-4650	16485	21	21	2717	12	2012-08-02 05:14:18.599673
16485	5	1-4739	16485	21	58	2326	48	2012-08-02 05:14:18.599673
16491	5	1-2786	16490	74	80	3535	68	2012-08-02 05:14:20.146163
16492	5	1-4554	16490	80	38	5335	71	2012-08-02 05:14:20.146163
16493	5	1-4691	16490	38	8	5563	35	2012-08-02 05:14:20.146163
16494	5	1-3065	16490	8	50	2165	47	2012-08-02 05:14:20.146163
16490	5	1-4743	16490	50	74	3560	91	2012-08-02 05:14:20.146163
16496	4	1-4597	16495	20	33	5788	56	2012-08-02 05:14:21.714887
16497	4	1-2810	16495	33	22	1971	90	2012-08-02 05:14:21.714887
16498	4	1-4586	16495	22	77	1504	88	2012-08-02 05:14:21.714887
16495	4	1-4747	16495	77	20	3047	24	2012-08-02 05:14:21.714887
16500	4	1-1111	16499	93	15	495	68	2012-08-02 05:14:21.714887
16501	4	1-4328	16499	15	17	636	11	2012-08-02 05:14:21.714887
16502	4	1-4740	16499	17	77	437	88	2012-08-02 05:14:21.714887
16499	4	1-4747	16499	77	93	1128	24	2012-08-02 05:14:21.714887
16504	4	1-3326	16503	6	51	530	3	2012-08-02 05:14:21.714887
16505	4	1-3466	16503	51	68	410	70	2012-08-02 05:14:21.714887
16506	4	1-3001	16503	68	77	494	88	2012-08-02 05:14:21.714887
16503	4	1-4747	16503	77	6	1286	24	2012-08-02 05:14:21.714887
16508	4	1-4102	16507	31	21	250	21	2012-08-02 05:14:21.714887
16509	4	1-3051	16507	21	10	117	49	2012-08-02 05:14:21.714887
16510	4	1-3193	16507	10	77	139	88	2012-08-02 05:14:21.714887
16507	4	1-4747	16507	77	31	366	24	2012-08-02 05:14:21.714887
16512	4	1-498	16511	63	7	131	81	2012-08-02 05:14:21.714887
16513	4	1-4394	16511	7	17	155	11	2012-08-02 05:14:21.714887
16514	4	1-4740	16511	17	77	111	88	2012-08-02 05:14:21.714887
16511	4	1-4747	16511	77	63	299	24	2012-08-02 05:14:21.714887
16516	5	1-4554	16515	80	38	143	71	2012-08-02 05:14:22.941468
16517	5	1-4691	16515	38	11	93	35	2012-08-02 05:14:22.941468
16518	5	1-4480	16515	11	83	91	86	2012-08-02 05:14:22.941468
16519	5	1-1896	16515	83	74	60	40	2012-08-02 05:14:22.941468
16515	5	1-4750	16515	74	80	151	68	2012-08-02 05:14:22.941468
16521	4	1-4677	16520	85	41	1207	81	2012-08-02 05:14:22.941468
16522	4	1-4749	16520	41	6	1718	1	2012-08-02 05:14:22.941468
16523	4	1-3945	16520	6	74	2142	40	2012-08-02 05:14:22.941468
16520	4	1-4750	16520	74	85	6377	68	2012-08-02 05:14:22.941468
16525	4	1-4725	16524	54	25	3617	19	2012-08-02 05:14:23.770277
16526	4	1-3421	16524	25	79	2274	30	2012-08-02 05:14:23.770277
16527	4	1-3222	16524	79	91	694	36	2012-08-02 05:14:23.770277
16524	4	1-4752	16524	91	54	5323	77	2012-08-02 05:14:23.770277
16529	3	1-4605	16528	10	78	237	86	2012-08-02 05:14:24.201306
16530	3	1-1498	16528	78	38	41	24	2012-08-02 05:14:24.201306
16528	3	1-4753	16528	38	10	140	18	2012-08-02 05:14:24.201306
16532	5	1-2698	16531	88	69	66	32	2012-08-02 05:14:24.201306
16533	5	1-4239	16531	69	73	38	10	2012-08-02 05:14:24.201306
16534	5	1-4551	16531	73	66	74	73	2012-08-02 05:14:24.201306
16535	5	1-1574	16531	66	38	12	24	2012-08-02 05:14:24.201306
16531	5	1-4753	16531	38	88	54	18	2012-08-02 05:14:24.201306
16537	5	1-4647	16536	48	48	2755	69	2012-08-02 05:14:24.201306
16538	5	1-1962	16536	48	79	1248	40	2012-08-02 05:14:24.201306
16539	5	1-4707	16536	79	88	834	61	2012-08-02 05:14:24.201306
16540	5	1-868	16536	88	38	318	24	2012-08-02 05:14:24.201306
16536	5	1-4753	16536	38	48	1491	18	2012-08-02 05:14:24.201306
16542	4	1-3049	16541	95	11	146	35	2012-08-02 05:14:24.201306
16543	4	1-4480	16541	11	78	197	86	2012-08-02 05:14:24.201306
16544	4	1-1498	16541	78	38	51	24	2012-08-02 05:14:24.201306
16541	4	1-4753	16541	38	95	254	18	2012-08-02 05:14:24.201306
16546	5	1-2682	16545	16	4	346	66	2012-08-02 05:14:24.201306
16547	5	1-4702	16545	4	67	756	65	2012-08-02 05:14:24.201306
16548	5	1-3999	16545	67	88	222	61	2012-08-02 05:14:24.201306
16549	5	1-868	16545	88	38	99	24	2012-08-02 05:14:24.201306
16545	5	1-4753	16545	38	16	543	18	2012-08-02 05:14:24.201306
16550	1	1-1498	\N	78	78	1291	24	2012-08-02 05:14:24.201306
16552	5	1-4727	16551	10	51	4244	76	2012-08-02 05:14:24.632523
16553	5	1-4246	16551	51	41	1043	73	2012-08-02 05:14:24.632523
16551	5	1-4754	16551	58	10	1093	42	2012-08-02 05:14:24.632523
16554	5	1-4278	16551	41	12	1067	39	2012-08-02 05:14:24.632523
16555	5	1-4644	16551	12	58	497	92	2012-08-02 05:14:24.632523
16557	5	1-4045	16556	55	20	3636	46	2012-08-02 05:14:25.803474
16558	5	1-4531	16556	20	41	1995	81	2012-08-02 05:14:25.803474
16559	5	1-4749	16556	41	77	2979	1	2012-08-02 05:14:25.803474
16560	5	1-4461	16556	77	38	1563	64	2012-08-02 05:14:25.803474
16556	5	1-4757	16556	38	55	1665	96	2012-08-02 05:14:25.803474
16562	5	1-4045	16561	55	20	646	46	2012-08-02 05:14:25.803474
16563	5	1-4531	16561	20	89	374	81	2012-08-02 05:14:25.803474
16564	5	1-4712	16561	89	74	259	41	2012-08-02 05:14:25.803474
16565	5	1-4601	16561	74	38	250	64	2012-08-02 05:14:25.803474
16561	5	1-4757	16561	38	55	281	96	2012-08-02 05:14:25.803474
16567	5	1-4611	16566	69	91	1303	74	2012-08-02 05:14:25.803474
16568	5	1-3764	16566	91	89	503	81	2012-08-02 05:14:25.803474
16569	5	1-4712	16566	89	74	413	41	2012-08-02 05:14:25.803474
16570	5	1-4601	16566	74	38	473	64	2012-08-02 05:14:25.803474
16566	5	1-4757	16566	38	69	630	96	2012-08-02 05:14:25.803474
16572	5	1-4673	16571	22	73	899	49	2012-08-02 05:14:25.803474
16573	5	1-3714	16571	73	62	1195	36	2012-08-02 05:14:25.803474
16574	5	1-4430	16571	62	62	1970	39	2012-08-02 05:14:25.803474
16575	5	1-4616	16571	62	38	757	64	2012-08-02 05:14:25.803474
16571	5	1-4757	16571	38	22	1024	96	2012-08-02 05:14:25.803474
16577	4	1-4335	16576	97	50	512	47	2012-08-02 05:14:27.007793
16578	4	1-4743	16576	50	74	270	91	2012-08-02 05:14:27.007793
16579	4	1-2786	16576	74	22	86	68	2012-08-02 05:14:27.007793
16576	4	1-4760	16576	22	97	4438	85	2012-08-02 05:14:27.007793
16581	3	1-3714	16580	73	56	393	36	2012-08-02 05:14:27.416911
16582	3	1-3103	16580	56	30	435	83	2012-08-02 05:14:27.416911
16580	3	1-4761	16580	30	73	359	49	2012-08-02 05:14:27.416911
16584	5	1-3714	16583	73	41	1557	36	2012-08-02 05:14:27.416911
16585	5	1-4748	16583	41	82	5169	78	2012-08-02 05:14:27.416911
16586	5	1-4671	16583	82	69	915	94	2012-08-02 05:14:27.416911
16587	5	1-4485	16583	69	30	1467	83	2012-08-02 05:14:27.416911
16583	5	1-4761	16583	30	73	1313	49	2012-08-02 05:14:27.416911
16589	5	1-3960	16588	59	83	3001	5	2012-08-02 05:14:27.416911
16590	5	1-4633	16588	83	1	1956	3	2012-08-02 05:14:27.416911
16591	5	1-2998	16588	1	8	817	68	2012-08-02 05:14:27.416911
16592	5	1-4745	16588	8	30	2101	83	2012-08-02 05:14:27.416911
16588	5	1-4761	16588	30	59	1884	49	2012-08-02 05:14:27.416911
16594	5	1-3193	16593	10	21	2272	88	2012-08-02 05:14:27.416911
16595	5	1-4516	16593	21	11	1288	73	2012-08-02 05:14:27.416911
16596	5	1-3812	16593	11	35	710	33	2012-08-02 05:14:27.416911
16597	5	1-4608	16593	35	30	1735	83	2012-08-02 05:14:27.416911
16593	5	1-4761	16593	30	10	1648	49	2012-08-02 05:14:27.416911
16599	4	1-3960	16598	59	65	978	5	2012-08-02 05:14:27.416911
16600	4	1-4330	16598	65	46	2385	57	2012-08-02 05:14:27.416911
16601	4	1-2840	16598	46	30	590	83	2012-08-02 05:14:27.416911
16598	4	1-4761	16598	30	59	570	49	2012-08-02 05:14:27.416911
16603	5	1-4516	16602	21	41	814	73	2012-08-02 05:14:28.986154
16604	5	1-4278	16602	41	63	857	39	2012-08-02 05:14:28.986154
16605	5	1-4617	16602	63	73	429	89	2012-08-02 05:14:28.986154
16606	5	1-4448	16602	73	20	1324	53	2012-08-02 05:14:28.986154
16602	5	1-4765	16602	20	21	1348	88	2012-08-02 05:14:28.986154
16608	4	1-2869	16607	32	50	850	63	2012-08-02 05:14:30.156955
16609	4	1-2690	16607	50	34	198	80	2012-08-02 05:14:30.156955
16610	4	1-4294	16607	34	22	58	83	2012-08-02 05:14:30.156955
16607	4	1-4768	16607	22	32	9388	49	2012-08-02 05:14:30.156955
16612	3	1-4388	16611	3	84	3311	52	2012-08-02 05:14:30.544053
16613	3	1-4372	16611	84	22	1404	7	2012-08-02 05:14:30.544053
16611	3	1-4769	16611	22	3	6866	87	2012-08-02 05:14:30.544053
16615	3	1-3921	16614	15	41	2085	81	2012-08-02 05:14:30.941649
16616	3	1-4749	16614	41	68	2022	1	2012-08-02 05:14:30.941649
16614	3	1-4770	16614	68	15	3526	91	2012-08-02 05:14:30.941649
16618	3	1-623	16617	3	41	633	81	2012-08-02 05:14:30.941649
16619	3	1-4749	16617	41	68	716	1	2012-08-02 05:14:30.941649
16617	3	1-4770	16617	68	3	1455	91	2012-08-02 05:14:30.941649
16621	4	1-499	16620	8	14	309	75	2012-08-02 05:14:30.941649
16622	4	1-3211	16620	14	94	1161	2	2012-08-02 05:14:30.941649
16623	4	1-4688	16620	94	68	571	1	2012-08-02 05:14:30.941649
16620	4	1-4770	16620	68	8	1614	91	2012-08-02 05:14:30.941649
16625	5	1-1623	16624	70	3	194	87	2012-08-02 05:14:30.941649
16626	5	1-4388	16624	3	88	168	52	2012-08-02 05:14:30.941649
16627	5	1-3558	16624	88	94	163	2	2012-08-02 05:14:30.941649
16628	5	1-4688	16624	94	68	90	1	2012-08-02 05:14:30.941649
16624	5	1-4770	16624	68	70	288	91	2012-08-02 05:14:30.941649
16630	6	1-4702	16629	4	46	866	65	2012-08-02 05:14:31.75932
16631	6	1-4653	16629	46	69	902	32	2012-08-02 05:14:31.75932
16632	6	1-4239	16629	69	41	417	10	2012-08-02 05:14:31.75932
16633	6	1-4465	16629	41	80	293	55	2012-08-02 05:14:31.75932
16634	6	1-3496	16629	80	25	36	75	2012-08-02 05:14:31.75932
16629	6	1-4772	16629	25	4	594	66	2012-08-02 05:14:31.75932
16636	5	1-3544	16635	1	59	6287	77	2012-08-02 05:14:31.75932
16637	5	1-1426	16635	59	41	3170	10	2012-08-02 05:14:31.75932
16638	5	1-4465	16635	41	80	2752	55	2012-08-02 05:14:31.75932
16639	5	1-3496	16635	80	25	419	75	2012-08-02 05:14:31.75932
16635	5	1-4772	16635	25	1	8548	66	2012-08-02 05:14:31.75932
16641	4	1-3668	16640	93	41	1910	10	2012-08-02 05:14:33.383205
16642	4	1-4465	16640	41	40	1656	55	2012-08-02 05:14:33.383205
16643	4	1-2432	16640	40	97	331	35	2012-08-02 05:14:33.383205
16640	4	1-4776	16640	97	93	1597	75	2012-08-02 05:14:33.383205
16645	5	1-2507	16644	56	75	3384	61	2012-08-02 05:14:33.383205
16646	5	1-700	16644	75	90	2891	78	2012-08-02 05:14:33.383205
16647	5	1-3678	16644	90	38	1184	71	2012-08-02 05:14:33.383205
16648	5	1-4691	16644	38	97	1004	35	2012-08-02 05:14:33.383205
16644	5	1-4776	16644	97	56	4987	75	2012-08-02 05:14:33.383205
16650	5	1-3207	16649	17	71	2217	41	2012-08-02 05:14:33.383205
16651	5	1-4199	16649	71	61	1159	52	2012-08-02 05:14:33.383205
16652	5	1-1652	16649	61	38	419	71	2012-08-02 05:14:33.383205
16653	5	1-4691	16649	38	97	375	35	2012-08-02 05:14:33.383205
16649	5	1-4776	16649	97	17	1963	75	2012-08-02 05:14:33.383205
16655	4	1-4757	16654	38	99	1896	96	2012-08-02 05:14:35.018568
16656	4	1-4446	16654	99	68	2624	89	2012-08-02 05:14:35.018568
16654	4	1-4780	16654	77	38	1818	64	2012-08-02 05:14:35.018568
16657	4	1-2014	16654	68	77	2269	13	2012-08-02 05:14:35.018568
16659	5	1-4757	16658	38	99	181	96	2012-08-02 05:14:35.018568
16660	5	1-4446	16658	99	50	287	89	2012-08-02 05:14:35.018568
16661	5	1-2193	16658	50	16	213	68	2012-08-02 05:14:35.018568
16662	5	1-2286	16658	16	77	164	13	2012-08-02 05:14:35.018568
16658	5	1-4780	16658	77	38	151	64	2012-08-02 05:14:35.018568
16664	2	1-4128	16663	79	12	642	6	2012-08-02 05:14:37.040092
16663	2	1-4785	16663	12	79	2639	66	2012-08-02 05:14:37.040092
16666	5	1-4702	16665	4	8	595	65	2012-08-02 05:14:37.040092
16667	5	1-3221	16665	8	20	337	46	2012-08-02 05:14:37.040092
16668	5	1-4531	16665	20	35	172	81	2012-08-02 05:14:37.040092
16669	5	1-3323	16665	35	12	59	6	2012-08-02 05:14:37.040092
16665	5	1-4785	16665	12	4	364	66	2012-08-02 05:14:37.040092
16671	5	1-3544	16670	1	54	2204	77	2012-08-02 05:14:37.040092
16672	5	1-4725	16670	54	61	1827	19	2012-08-02 05:14:37.040092
16673	5	1-2229	16670	61	35	857	81	2012-08-02 05:14:37.040092
16674	5	1-3323	16670	35	12	357	6	2012-08-02 05:14:37.040092
16670	5	1-4785	16670	12	1	2706	66	2012-08-02 05:14:37.040092
16675	1	1-3323	\N	35	35	169	6	2012-08-02 05:14:37.040092
16677	5	1-4651	16676	56	25	437	98	2012-08-02 05:14:37.449131
16678	5	1-1221	16676	25	38	134	24	2012-08-02 05:14:37.449131
16679	5	1-4753	16676	38	41	754	18	2012-08-02 05:14:37.449131
16680	5	1-4438	16676	41	22	940	50	2012-08-02 05:14:37.449131
16676	5	1-4786	16676	22	56	593	60	2012-08-02 05:14:37.449131
16682	5	1-2918	16681	81	95	843	93	2012-08-02 05:14:37.846907
16683	5	1-3725	16681	95	40	368	31	2012-08-02 05:14:37.846907
16684	5	1-4593	16681	40	10	409	42	2012-08-02 05:14:37.846907
16685	5	1-4727	16681	10	90	1521	76	2012-08-02 05:14:37.846907
16681	5	1-4787	16681	90	81	989	27	2012-08-02 05:14:37.846907
16687	4	1-4692	16686	19	43	478	41	2012-08-02 05:14:37.846907
16688	4	1-4212	16686	43	1	334	34	2012-08-02 05:14:37.846907
16689	4	1-4338	16686	1	90	589	76	2012-08-02 05:14:37.846907
16686	4	1-4787	16686	90	19	412	27	2012-08-02 05:14:37.846907
16691	5	1-2492	16690	20	8	343	94	2012-08-02 05:14:37.846907
16692	5	1-4762	16690	8	83	302	89	2012-08-02 05:14:37.846907
16693	5	1-3392	16690	83	10	376	42	2012-08-02 05:14:37.846907
16694	5	1-4727	16690	10	90	1537	76	2012-08-02 05:14:37.846907
16690	5	1-4787	16690	90	20	1098	27	2012-08-02 05:14:37.846907
16696	4	1-2196	16695	14	55	540	4	2012-08-02 05:14:40.815418
16697	4	1-2609	16695	55	36	502	90	2012-08-02 05:14:40.815418
16698	4	1-4719	16695	36	74	749	74	2012-08-02 05:14:40.815418
16695	4	1-4794	16695	74	14	2664	48	2012-08-02 05:14:40.815418
16700	4	1-4146	16699	48	77	2805	28	2012-08-02 05:14:40.815418
16701	4	1-2118	16699	77	36	749	90	2012-08-02 05:14:40.815418
16702	4	1-4719	16699	36	74	1270	74	2012-08-02 05:14:40.815418
16699	4	1-4794	16699	74	48	5138	48	2012-08-02 05:14:40.815418
16704	3	1-2953	16703	63	10	1048	42	2012-08-02 05:14:41.227675
16705	3	1-4727	16703	10	40	2260	76	2012-08-02 05:14:41.227675
16703	3	1-4795	16703	40	63	1562	91	2012-08-02 05:14:41.227675
16707	5	1-2786	16706	74	94	929	68	2012-08-02 05:14:41.227675
16708	5	1-4777	16706	94	52	1298	93	2012-08-02 05:14:41.227675
16709	5	1-4276	16706	52	7	928	2	2012-08-02 05:14:41.227675
16710	5	1-3699	16706	7	40	894	76	2012-08-02 05:14:41.227675
16706	5	1-4795	16706	40	74	1041	91	2012-08-02 05:14:41.227675
16712	5	1-1623	16711	70	3	54	87	2012-08-02 05:14:41.227675
16713	5	1-4388	16711	3	88	51	52	2012-08-02 05:14:41.227675
16714	5	1-3558	16711	88	7	54	2	2012-08-02 05:14:41.227675
16715	5	1-3699	16711	7	40	57	76	2012-08-02 05:14:41.227675
16711	5	1-4795	16711	40	70	74	91	2012-08-02 05:14:41.227675
16717	5	1-3675	16716	20	96	505	61	2012-08-02 05:14:41.227675
16718	5	1-3406	16716	96	77	1552	17	2012-08-02 05:14:41.227675
16719	5	1-3938	16716	77	7	560	2	2012-08-02 05:14:41.227675
16720	5	1-3699	16716	7	40	606	76	2012-08-02 05:14:41.227675
16716	5	1-4795	16716	40	20	792	91	2012-08-02 05:14:41.227675
16722	4	1-4505	16721	55	98	2418	88	2012-08-02 05:14:41.638569
16723	4	1-4535	16721	98	87	1965	97	2012-08-02 05:14:41.638569
16724	4	1-4428	16721	87	32	1721	7	2012-08-02 05:14:41.638569
16721	4	1-4796	16721	32	55	2350	41	2012-08-02 05:14:41.638569
16726	5	1-4601	16725	74	38	1247	64	2012-08-02 05:14:41.638569
16727	5	1-4757	16725	38	45	1664	96	2012-08-02 05:14:41.638569
16728	5	1-3117	16725	45	64	433	70	2012-08-02 05:14:41.638569
16729	5	1-4092	16725	64	32	791	7	2012-08-02 05:14:41.638569
16725	5	1-4796	16725	32	74	1087	41	2012-08-02 05:14:41.638569
16731	5	1-4199	16730	71	97	278	52	2012-08-02 05:14:41.638569
16732	5	1-4440	16730	97	23	217	92	2012-08-02 05:14:41.638569
16733	5	1-4048	16730	23	64	177	70	2012-08-02 05:14:41.638569
16734	5	1-4092	16730	64	32	330	7	2012-08-02 05:14:41.638569
16730	5	1-4796	16730	32	71	462	41	2012-08-02 05:14:41.638569
16736	5	1-4430	16735	62	38	1703	39	2012-08-02 05:14:42.025573
16737	5	1-4354	16735	38	20	1617	46	2012-08-02 05:14:42.025573
16738	5	1-4531	16735	20	89	951	81	2012-08-02 05:14:42.025573
16739	5	1-4712	16735	89	12	670	41	2012-08-02 05:14:42.025573
16735	5	1-4797	16735	12	62	1246	36	2012-08-02 05:14:42.025573
16741	5	1-4748	16740	41	72	379	78	2012-08-02 05:14:42.025573
16742	5	1-3947	16740	72	1	155	20	2012-08-02 05:14:42.025573
16743	5	1-3847	16740	1	23	52	29	2012-08-02 05:14:42.025573
16744	5	1-4701	16740	23	12	56	41	2012-08-02 05:14:42.025573
16740	5	1-4797	16740	12	41	112	36	2012-08-02 05:14:42.025573
16746	5	1-4430	16745	62	12	230	39	2012-08-02 05:14:42.025573
16747	5	1-4644	16745	12	52	104	92	2012-08-02 05:14:42.025573
16748	5	1-4025	16745	52	89	109	81	2012-08-02 05:14:42.025573
16749	5	1-4712	16745	89	12	79	41	2012-08-02 05:14:42.025573
16745	5	1-4797	16745	12	62	164	36	2012-08-02 05:14:42.025573
16751	5	1-4748	16750	41	89	3478	78	2012-08-02 05:14:42.025573
16752	5	1-4504	16750	89	23	2878	2	2012-08-02 05:14:42.025573
16753	5	1-1819	16750	23	89	507	81	2012-08-02 05:14:42.025573
16754	5	1-4712	16750	89	12	430	41	2012-08-02 05:14:42.025573
16750	5	1-4797	16750	12	41	940	36	2012-08-02 05:14:42.025573
16756	5	1-4473	16755	69	12	648	70	2012-08-02 05:14:42.025573
16757	5	1-4661	16755	12	3	928	91	2012-08-02 05:14:42.025573
16758	5	1-623	16755	3	89	694	81	2012-08-02 05:14:42.025573
16759	5	1-4712	16755	89	12	594	41	2012-08-02 05:14:42.025573
16755	5	1-4797	16755	12	69	1348	36	2012-08-02 05:14:42.025573
16761	2	1-4659	16760	36	76	2817	69	2012-08-02 05:14:44.456345
16760	2	1-4803	16760	76	36	2077	85	2012-08-02 05:14:44.456345
16763	3	1-4667	16762	55	56	894	59	2012-08-02 05:14:45.240543
16764	3	1-4543	16762	56	72	1060	79	2012-08-02 05:14:45.240543
16762	3	1-4805	16762	72	55	4708	88	2012-08-02 05:14:45.240543
16766	5	1-4535	16765	98	21	11	97	2012-08-02 05:14:45.240543
16767	5	1-4455	16765	21	16	8	57	2012-08-02 05:14:45.240543
16768	5	1-1445	16765	16	56	3	59	2012-08-02 05:14:45.240543
16769	5	1-4543	16765	56	72	3	79	2012-08-02 05:14:45.240543
16765	5	1-4805	16765	72	98	15	88	2012-08-02 05:14:45.240543
16771	5	1-4651	16770	56	15	1072	98	2012-08-02 05:14:45.638414
16772	5	1-3341	16770	15	77	557	13	2012-08-02 05:14:45.638414
16773	5	1-4780	16770	77	87	425	64	2012-08-02 05:14:45.638414
16774	5	1-3885	16770	87	64	338	72	2012-08-02 05:14:45.638414
16770	5	1-4806	16770	64	56	1973	60	2012-08-02 05:14:45.638414
16776	4	1-4029	16775	52	11	218	35	2012-08-02 05:14:45.638414
16777	4	1-4480	16775	11	86	257	86	2012-08-02 05:14:45.638414
16778	4	1-4236	16775	86	64	186	72	2012-08-02 05:14:45.638414
16775	4	1-4806	16775	64	52	1140	60	2012-08-02 05:14:45.638414
16780	4	1-3448	16779	49	24	1671	23	2012-08-02 05:14:46.047468
16781	4	1-4704	16779	24	74	6964	17	2012-08-02 05:14:46.047468
16782	4	1-4642	16779	74	7	4235	11	2012-08-02 05:14:46.047468
16779	4	1-4807	16779	7	49	3937	45	2012-08-02 05:14:46.047468
16784	4	1-4729	16783	83	20	2312	53	2012-08-02 05:14:46.047468
16785	4	1-4765	16783	20	38	2262	88	2012-08-02 05:14:46.047468
16786	4	1-4592	16783	38	7	2318	11	2012-08-02 05:14:46.047468
16783	4	1-4807	16783	7	83	2987	45	2012-08-02 05:14:46.047468
16788	5	1-4685	16787	89	62	517	29	2012-08-02 05:14:46.047468
16789	5	1-3368	16787	62	53	479	59	2012-08-02 05:14:46.047468
16790	5	1-4705	16787	53	18	464	58	2012-08-02 05:14:46.047468
16791	5	1-4284	16787	18	7	974	11	2012-08-02 05:14:46.047468
16787	5	1-4807	16787	7	89	1262	45	2012-08-02 05:14:46.047468
16793	5	1-2812	16792	22	24	2261	5	2012-08-02 05:14:47.262559
16794	5	1-4618	16792	24	48	1971	93	2012-08-02 05:14:47.262559
16795	5	1-4131	16792	48	8	1584	95	2012-08-02 05:14:47.262559
16796	5	1-4629	16792	8	38	1105	9	2012-08-02 05:14:47.262559
16792	5	1-4810	16792	38	22	2403	49	2012-08-02 05:14:47.262559
16798	6	1-4440	16797	97	58	2251	92	2012-08-02 05:14:47.671524
16799	6	1-4754	16797	58	71	4067	42	2012-08-02 05:14:47.671524
16800	6	1-3893	16797	71	43	1453	72	2012-08-02 05:14:47.671524
16801	6	1-1718	16797	43	12	280	6	2012-08-02 05:14:47.671524
16802	6	1-4785	16797	12	41	1914	66	2012-08-02 05:14:47.671524
16797	6	1-4811	16797	41	97	3595	52	2012-08-02 05:14:47.671524
16804	5	1-4629	16803	8	45	223	9	2012-08-02 05:14:48.47809
16805	5	1-4397	16803	45	11	168	12	2012-08-02 05:14:48.47809
16806	5	1-4147	16803	11	80	56	27	2012-08-02 05:14:48.47809
16807	5	1-4800	16803	80	59	49	99	2012-08-02 05:14:48.47809
16803	5	1-4813	16803	59	8	368	95	2012-08-02 05:14:48.47809
16809	5	1-1463	16808	85	24	209	23	2012-08-02 05:14:48.47809
16810	5	1-4704	16808	24	96	1064	17	2012-08-02 05:14:48.47809
16811	5	1-4746	16808	96	75	252	30	2012-08-02 05:14:48.47809
16812	5	1-4413	16808	75	59	112	99	2012-08-02 05:14:48.47809
16808	5	1-4813	16808	59	85	863	95	2012-08-02 05:14:48.47809
16814	5	1-2319	16813	61	62	166	54	2012-08-02 05:14:48.47809
16815	5	1-4708	16813	62	71	83	30	2012-08-02 05:14:48.47809
16816	5	1-4497	16813	71	80	54	27	2012-08-02 05:14:48.47809
16817	5	1-4800	16813	80	59	53	99	2012-08-02 05:14:48.47809
16813	5	1-4813	16813	59	61	441	95	2012-08-02 05:14:48.47809
16819	3	1-4549	16818	82	37	5670	11	2012-08-02 05:14:48.898023
16820	3	1-516	16818	37	52	935	46	2012-08-02 05:14:48.898023
16818	3	1-4814	16818	52	82	8783	62	2012-08-02 05:14:48.898023
16822	5	1-4363	16821	24	30	2566	83	2012-08-02 05:14:49.317875
16823	5	1-4761	16821	30	98	2477	49	2012-08-02 05:14:49.317875
16824	5	1-3747	16821	98	41	774	98	2012-08-02 05:14:49.317875
16825	5	1-4114	16821	41	93	261	6	2012-08-02 05:14:49.317875
16821	5	1-4815	16821	93	24	1297	90	2012-08-02 05:14:49.317875
16827	5	1-4522	16826	85	74	244	91	2012-08-02 05:14:49.715583
16828	5	1-2786	16826	74	94	207	68	2012-08-02 05:14:49.715583
16829	5	1-4777	16826	94	48	276	93	2012-08-02 05:14:49.715583
16830	5	1-4131	16826	48	35	202	95	2012-08-02 05:14:49.715583
16826	5	1-4816	16826	35	85	481	97	2012-08-02 05:14:49.715583
16832	4	1-3910	16831	76	7	715	11	2012-08-02 05:14:49.715583
16833	4	1-4807	16831	7	31	878	45	2012-08-02 05:14:49.715583
16834	4	1-3797	16831	31	35	657	95	2012-08-02 05:14:49.715583
16831	4	1-4816	16831	35	76	1657	97	2012-08-02 05:14:49.715583
16836	5	1-4522	16835	85	3	26	91	2012-08-02 05:14:49.715583
16837	5	1-3600	16835	3	27	13	19	2012-08-02 05:14:49.715583
16838	5	1-4220	16835	27	75	15	16	2012-08-02 05:14:49.715583
16839	5	1-4645	16835	75	35	18	95	2012-08-02 05:14:49.715583
16835	5	1-4816	16835	35	85	47	97	2012-08-02 05:14:49.715583
16841	4	1-4692	16840	19	12	2097	41	2012-08-02 05:14:50.135498
16842	4	1-4797	16840	12	8	4251	36	2012-08-02 05:14:50.135498
16843	4	1-2662	16840	8	88	1686	71	2012-08-02 05:14:50.135498
16840	4	1-4817	16840	88	19	1979	27	2012-08-02 05:14:50.135498
16845	3	1-4472	16844	86	94	161	43	2012-08-02 05:14:52.124
16846	3	1-3003	16844	94	48	39	8	2012-08-02 05:14:52.124
16844	3	1-4822	16844	48	86	55	24	2012-08-02 05:14:52.124
16848	3	1-4664	16847	32	62	5083	59	2012-08-02 05:14:52.521889
16849	3	1-4599	16847	62	53	2888	39	2012-08-02 05:14:52.521889
16847	3	1-4823	16847	53	32	9390	24	2012-08-02 05:14:52.521889
16851	2	1-3106	16850	69	46	1988	5	2012-08-02 05:14:53.74809
16850	2	1-4826	16850	46	69	5011	62	2012-08-02 05:14:53.74809
16853	5	1-4404	16852	98	54	4248	77	2012-08-02 05:14:54.996669
16854	5	1-4725	16852	54	27	3001	19	2012-08-02 05:14:54.996669
16855	5	1-4220	16852	27	75	2767	16	2012-08-02 05:14:54.996669
16856	5	1-4645	16852	75	85	2657	95	2012-08-02 05:14:54.996669
16852	5	1-4829	16852	85	98	7750	61	2012-08-02 05:14:54.996669
16858	4	1-3447	16857	6	36	917	90	2012-08-02 05:14:54.996669
16859	4	1-4719	16857	36	88	1401	74	2012-08-02 05:14:54.996669
16860	4	1-4460	16857	88	85	453	95	2012-08-02 05:14:54.996669
16857	4	1-4829	16857	85	6	1569	61	2012-08-02 05:14:54.996669
16862	5	1-868	16861	88	54	136	24	2012-08-02 05:14:54.996669
16863	5	1-3895	16861	54	35	168	20	2012-08-02 05:14:54.996669
16864	5	1-4683	16861	35	75	75	16	2012-08-02 05:14:54.996669
16865	5	1-4645	16861	75	85	89	95	2012-08-02 05:14:54.996669
16861	5	1-4829	16861	85	88	318	61	2012-08-02 05:14:54.996669
16866	1	1-868	\N	88	88	1092	24	2012-08-02 05:14:54.996669
16868	2	1-4601	16867	74	16	5632	64	2012-08-02 05:14:56.200925
16867	2	1-4832	16867	16	74	7009	41	2012-08-02 05:14:56.200925
16870	4	1-4338	16869	1	40	2810	76	2012-08-02 05:14:56.598715
16871	4	1-4795	16869	40	63	2383	91	2012-08-02 05:14:56.598715
16872	4	1-2953	16869	63	94	1962	42	2012-08-02 05:14:56.598715
16869	4	1-4833	16869	94	1	2407	34	2012-08-02 05:14:56.598715
16874	5	1-4691	16873	38	48	699	35	2012-08-02 05:14:57.813942
16875	5	1-3252	16873	48	79	572	40	2012-08-02 05:14:57.813942
16876	5	1-4707	16873	79	35	380	61	2012-08-02 05:14:57.813942
16877	5	1-4103	16873	35	3	627	46	2012-08-02 05:14:57.813942
16873	5	1-4836	16873	3	38	824	71	2012-08-02 05:14:57.813942
16879	5	1-4353	16878	82	11	475	67	2012-08-02 05:14:57.813942
16880	5	1-4366	16878	11	79	1064	40	2012-08-02 05:14:57.813942
16881	5	1-4707	16878	79	35	727	61	2012-08-02 05:14:57.813942
16882	5	1-4103	16878	35	3	1235	46	2012-08-02 05:14:57.813942
16878	5	1-4836	16878	3	82	1669	71	2012-08-02 05:14:57.813942
16884	2	1-3943	16883	28	3	425	46	2012-08-02 05:14:57.813942
16883	2	1-4836	16883	3	28	580	71	2012-08-02 05:14:57.813942
16886	2	1-2924	16885	88	31	1029	78	2012-08-02 05:14:58.642616
16885	2	1-4838	16885	31	88	1593	87	2012-08-02 05:14:58.642616
16888	5	1-3232	16887	49	75	58	16	2012-08-02 05:14:58.642616
16889	5	1-4645	16887	75	61	68	95	2012-08-02 05:14:58.642616
16890	5	1-2319	16887	61	61	27	54	2012-08-02 05:14:58.642616
16891	5	1-4646	16887	61	31	33	78	2012-08-02 05:14:58.642616
16887	5	1-4838	16887	31	49	72	87	2012-08-02 05:14:58.642616
16893	5	1-3477	16892	4	61	1725	52	2012-08-02 05:14:58.642616
16894	5	1-1652	16892	61	27	727	71	2012-08-02 05:14:58.642616
16895	5	1-4790	16892	27	10	896	65	2012-08-02 05:14:58.642616
16896	5	1-4837	16892	10	31	1055	78	2012-08-02 05:14:58.642616
16892	5	1-4838	16892	31	4	2490	87	2012-08-02 05:14:58.642616
16897	1	1-3477	\N	4	4	460	52	2012-08-02 05:14:58.642616
16899	5	1-3392	16898	83	94	514	42	2012-08-02 05:14:59.051711
16900	5	1-4833	16898	94	1	710	34	2012-08-02 05:14:59.051711
16901	5	1-4338	16898	1	40	933	76	2012-08-02 05:14:59.051711
16902	5	1-4795	16898	40	21	890	91	2012-08-02 05:14:59.051711
16898	5	1-4839	16898	21	83	565	89	2012-08-02 05:14:59.051711
16904	4	1-4448	16903	73	49	1421	53	2012-08-02 05:14:59.051711
16905	4	1-2969	16903	49	40	570	76	2012-08-02 05:14:59.051711
16906	4	1-4795	16903	40	21	660	91	2012-08-02 05:14:59.051711
16903	4	1-4839	16903	21	73	509	89	2012-08-02 05:14:59.051711
16908	5	1-3137	16907	23	94	614	42	2012-08-02 05:14:59.051711
16909	5	1-4833	16907	94	1	1047	34	2012-08-02 05:14:59.051711
16910	5	1-4338	16907	1	40	1698	76	2012-08-02 05:14:59.051711
16911	5	1-4795	16907	40	21	2000	91	2012-08-02 05:14:59.051711
16907	5	1-4839	16907	21	23	1568	89	2012-08-02 05:14:59.051711
16913	4	1-2862	16912	24	8	675	68	2012-08-02 05:14:59.47151
16914	4	1-4745	16912	8	30	1702	83	2012-08-02 05:14:59.47151
16915	4	1-4761	16912	30	63	1497	49	2012-08-02 05:14:59.47151
16912	4	1-4840	16912	63	24	876	8	2012-08-02 05:14:59.47151
16916	1	1-1604	\N	25	25	2132	83	2012-08-02 05:14:59.47151
16918	5	1-1429	16917	75	71	2517	16	2012-08-02 05:15:00.289234
16919	5	1-4841	16917	71	72	1080	38	2012-08-02 05:15:00.289234
16920	5	1-1554	16917	72	15	487	99	2012-08-02 05:15:00.289234
16921	5	1-4731	16917	15	84	811	86	2012-08-02 05:15:00.289234
16917	5	1-4842	16917	84	75	5910	79	2012-08-02 05:15:00.289234
16923	5	1-4805	16922	72	21	1108	88	2012-08-02 05:15:00.289234
16924	5	1-4516	16922	21	32	554	73	2012-08-02 05:15:00.289234
16925	5	1-105	16922	32	26	60	4	2012-08-02 05:15:00.289234
16926	5	1-3339	16922	26	84	35	86	2012-08-02 05:15:00.289234
16922	5	1-4842	16922	84	72	274	79	2012-08-02 05:15:00.289234
16928	5	1-4821	16927	43	20	1377	91	2012-08-02 05:15:00.289234
16929	5	1-3675	16927	20	44	807	61	2012-08-02 05:15:00.289234
16930	5	1-2461	16927	44	15	134	99	2012-08-02 05:15:00.289234
16931	5	1-4731	16927	15	84	273	86	2012-08-02 05:15:00.289234
16927	5	1-4842	16927	84	43	2444	79	2012-08-02 05:15:00.289234
16933	4	1-3663	16932	40	7	1208	47	2012-08-02 05:15:00.720726
16934	4	1-4427	16932	7	93	972	6	2012-08-02 05:15:00.720726
16935	4	1-4815	16932	93	79	4085	90	2012-08-02 05:15:00.720726
16932	4	1-4843	16932	79	40	4232	20	2012-08-02 05:15:00.720726
16937	4	1-4661	16936	12	21	1356	91	2012-08-02 05:15:02.355599
16938	4	1-4839	16936	21	37	1106	89	2012-08-02 05:15:02.355599
16939	4	1-2838	16936	37	70	386	79	2012-08-02 05:15:02.355599
16936	4	1-4847	16936	70	12	1009	70	2012-08-02 05:15:02.355599
16941	2	1-4811	16940	41	32	4166	52	2012-08-02 05:15:02.764098
16940	2	1-4848	16940	32	41	4320	66	2012-08-02 05:15:02.764098
16943	3	1-4816	16942	35	75	2744	97	2012-08-02 05:15:03.581848
16944	3	1-3396	16942	75	20	689	43	2012-08-02 05:15:03.581848
16942	3	1-4850	16942	20	35	2010	95	2012-08-02 05:15:03.581848
16946	4	1-4845	16945	76	80	1892	80	2012-08-02 05:15:03.581848
16947	4	1-4060	16945	80	12	1131	14	2012-08-02 05:15:03.581848
16948	4	1-2655	16945	12	20	185	43	2012-08-02 05:15:03.581848
16945	4	1-4850	16945	20	76	852	95	2012-08-02 05:15:03.581848
16950	5	1-2319	16949	61	23	439	54	2012-08-02 05:15:03.581848
16951	5	1-4047	16949	23	71	815	42	2012-08-02 05:15:03.581848
16952	5	1-3893	16949	71	35	306	72	2012-08-02 05:15:03.581848
16953	5	1-4779	16949	35	20	260	43	2012-08-02 05:15:03.581848
16949	5	1-4850	16949	20	61	1282	95	2012-08-02 05:15:03.581848
16955	5	1-2379	16954	24	62	1617	59	2012-08-02 05:15:04.001638
16956	5	1-4599	16954	62	24	1832	39	2012-08-02 05:15:04.001638
16957	5	1-3877	16954	24	94	1525	42	2012-08-02 05:15:04.001638
16958	5	1-4266	16954	94	90	1050	45	2012-08-02 05:15:04.001638
16954	5	1-4851	16954	90	24	2216	55	2012-08-02 05:15:04.001638
16960	3	1-4800	16959	80	92	5406	99	2012-08-02 05:15:04.399384
16961	3	1-3667	16959	92	6	3265	30	2012-08-02 05:15:04.399384
16959	3	1-4852	16959	6	80	5871	27	2012-08-02 05:15:04.399384
16963	5	1-3855	16962	26	23	44	54	2012-08-02 05:15:06.016019
16964	5	1-4047	16962	23	94	83	42	2012-08-02 05:15:06.016019
16965	5	1-4833	16962	94	1	131	34	2012-08-02 05:15:06.016019
16962	5	1-4856	16962	74	26	209	74	2012-08-02 05:15:06.016019
16966	5	1-4338	16962	1	74	196	76	2012-08-02 05:15:06.016019
16968	5	1-4404	16967	98	60	1135	77	2012-08-02 05:15:10.235872
16969	5	1-1264	16967	60	80	324	27	2012-08-02 05:15:10.235872
16970	5	1-4800	16967	80	15	295	99	2012-08-02 05:15:10.235872
16971	5	1-4731	16967	15	69	545	86	2012-08-02 05:15:10.235872
16967	5	1-4867	16967	69	98	1835	61	2012-08-02 05:15:10.235872
16973	3	1-2461	16972	44	15	480	99	2012-08-02 05:15:10.235872
16974	3	1-4731	16972	15	69	907	86	2012-08-02 05:15:10.235872
16972	3	1-4867	16972	69	44	3118	61	2012-08-02 05:15:10.235872
16976	4	1-4746	16975	96	13	1127	30	2012-08-02 05:15:11.782617
16977	4	1-4855	16975	13	34	4546	80	2012-08-02 05:15:11.782617
16978	4	1-4294	16975	34	70	3340	83	2012-08-02 05:15:11.782617
16975	4	1-4871	16975	70	96	6080	17	2012-08-02 05:15:11.782617
16980	4	1-4746	16979	96	13	183	30	2012-08-02 05:15:11.782617
16981	4	1-4855	16979	13	9	995	80	2012-08-02 05:15:11.782617
16982	4	1-4153	16979	9	70	298	83	2012-08-02 05:15:11.782617
16979	4	1-4871	16979	70	96	732	17	2012-08-02 05:15:11.782617
16984	5	1-1129	16983	59	9	1171	93	2012-08-02 05:15:13.727019
16985	5	1-3666	16983	9	77	774	13	2012-08-02 05:15:13.727019
16986	5	1-4780	16983	77	87	754	64	2012-08-02 05:15:13.727019
16987	5	1-3885	16983	87	22	764	72	2012-08-02 05:15:13.727019
16983	5	1-4876	16983	22	59	1477	21	2012-08-02 05:15:13.727019
16989	3	1-4429	16988	54	5	989	85	2012-08-02 05:15:13.727019
16990	3	1-2194	16988	5	22	376	72	2012-08-02 05:15:13.727019
16988	3	1-4876	16988	22	54	754	21	2012-08-02 05:15:13.727019
16992	4	1-4626	16991	67	63	752	84	2012-08-02 05:15:13.727019
16993	4	1-4696	16991	63	87	722	64	2012-08-02 05:15:13.727019
16994	4	1-3885	16991	87	22	783	72	2012-08-02 05:15:13.727019
16991	4	1-4876	16991	22	67	1622	21	2012-08-02 05:15:13.727019
16995	1	1-2560	\N	52	52	1793	21	2012-08-02 05:15:13.727019
16997	5	1-4731	16996	15	69	618	86	2012-08-02 05:15:14.898584
16998	5	1-4867	16996	69	75	1309	61	2012-08-02 05:15:14.898584
16999	5	1-700	16996	75	13	729	78	2012-08-02 05:15:14.898584
17000	5	1-4853	16996	13	39	1039	7	2012-08-02 05:15:14.898584
16996	5	1-4879	16996	39	15	531	99	2012-08-02 05:15:14.898584
17002	4	1-4731	17001	15	69	871	86	2012-08-02 05:15:14.898584
17003	4	1-4867	17001	69	36	2259	61	2012-08-02 05:15:14.898584
17004	4	1-3217	17001	36	39	977	7	2012-08-02 05:15:14.898584
17001	4	1-4879	17001	39	15	611	99	2012-08-02 05:15:14.898584
17006	5	1-4731	17005	15	54	102	86	2012-08-02 05:15:14.898584
17007	5	1-4110	17005	54	23	69	92	2012-08-02 05:15:14.898584
17008	5	1-4048	17005	23	64	47	70	2012-08-02 05:15:14.898584
17009	5	1-4092	17005	64	39	72	7	2012-08-02 05:15:14.898584
17005	5	1-4879	17005	39	15	57	99	2012-08-02 05:15:14.898584
17011	5	1-820	17010	13	77	584	13	2012-08-02 05:15:14.898584
17012	5	1-4780	17010	77	73	584	64	2012-08-02 05:15:14.898584
17013	5	1-4093	17010	73	13	477	78	2012-08-02 05:15:14.898584
17014	5	1-4853	17010	13	39	1256	7	2012-08-02 05:15:14.898584
17010	5	1-4879	17010	39	13	1144	99	2012-08-02 05:15:14.898584
17016	5	1-4731	17015	15	86	3940	86	2012-08-02 05:15:14.898584
17017	5	1-4562	17015	86	85	2212	44	2012-08-02 05:15:14.898584
17018	5	1-3915	17015	85	87	2075	97	2012-08-02 05:15:14.898584
17019	5	1-4428	17015	87	39	1881	7	2012-08-02 05:15:14.898584
17015	5	1-4879	17015	39	15	1803	99	2012-08-02 05:15:14.898584
17021	6	1-2492	17020	20	8	1021	94	2012-08-02 05:15:15.296323
17022	6	1-4762	17020	8	3	808	89	2012-08-02 05:15:15.296323
17023	6	1-4717	17020	3	38	900	9	2012-08-02 05:15:15.296323
17024	6	1-4810	17020	38	98	1895	49	2012-08-02 05:15:15.296323
17025	6	1-3747	17020	98	24	569	98	2012-08-02 05:15:15.296323
17020	6	1-4880	17020	24	20	3644	27	2012-08-02 05:15:15.296323
17027	4	1-4692	17026	19	71	13	41	2012-08-02 05:15:15.296323
17028	4	1-4199	17026	71	83	7	52	2012-08-02 05:15:15.296323
17029	4	1-1404	17026	83	24	2	98	2012-08-02 05:15:15.296323
17026	4	1-4880	17026	24	19	13	27	2012-08-02 05:15:15.296323
17031	5	1-4800	17030	80	57	3336	99	2012-08-02 05:15:15.296323
17032	5	1-3385	17030	57	38	565	9	2012-08-02 05:15:15.296323
17033	5	1-4810	17030	38	98	1313	49	2012-08-02 05:15:15.296323
17034	5	1-3747	17030	98	24	436	98	2012-08-02 05:15:15.296323
17030	5	1-4880	17030	24	80	3081	27	2012-08-02 05:15:15.296323
17036	5	1-4856	17035	74	100	1919	74	2012-08-02 05:15:18.091409
17037	5	1-4791	17035	100	10	836	49	2012-08-02 05:15:18.091409
17038	5	1-3193	17035	10	98	1173	88	2012-08-02 05:15:18.091409
17039	5	1-4535	17035	98	34	937	97	2012-08-02 05:15:18.091409
17035	5	1-4887	17035	34	74	1606	76	2012-08-02 05:15:18.091409
17041	4	1-4745	17040	8	70	939	83	2012-08-02 05:15:18.875846
17042	4	1-4871	17040	70	77	2103	17	2012-08-02 05:15:18.875846
17043	4	1-3938	17040	77	96	607	2	2012-08-02 05:15:18.875846
17040	4	1-4889	17040	96	8	396	68	2012-08-02 05:15:18.875846
17045	5	1-4863	17044	22	65	2420	94	2012-08-02 05:15:19.273615
17046	5	1-4870	17044	65	98	1925	49	2012-08-02 05:15:19.273615
17047	5	1-3747	17044	98	24	459	98	2012-08-02 05:15:19.273615
17048	5	1-4880	17044	24	7	2336	27	2012-08-02 05:15:19.273615
17044	5	1-4890	17044	7	22	3227	60	2012-08-02 05:15:19.273615
17050	4	1-623	17049	3	21	685	81	2012-08-02 05:15:20.411447
17051	4	1-4650	17049	21	23	830	12	2012-08-02 05:15:20.411447
17052	4	1-4589	17049	23	26	1159	63	2012-08-02 05:15:20.411447
17049	4	1-4893	17049	26	3	1144	91	2012-08-02 05:15:20.411447
17054	5	1-4839	17053	21	3	692	89	2012-08-02 05:15:20.411447
17055	5	1-4717	17053	3	31	824	9	2012-08-02 05:15:20.411447
17056	5	1-3145	17053	31	83	711	33	2012-08-02 05:15:20.411447
17057	5	1-4583	17053	83	26	697	63	2012-08-02 05:15:20.411447
17053	5	1-4893	17053	26	21	829	91	2012-08-02 05:15:20.411447
17059	4	1-4876	17058	22	17	868	21	2012-08-02 05:15:23.549453
17060	4	1-4140	17058	17	34	492	26	2012-08-02 05:15:23.549453
17061	4	1-195	17058	34	7	153	3	2012-08-02 05:15:23.549453
17058	4	1-4901	17058	7	22	444	72	2012-08-02 05:15:23.549453
17063	6	1-4840	17062	63	48	2778	8	2012-08-02 05:15:24.344987
17064	6	1-4822	17062	48	6	3734	24	2012-08-02 05:15:24.344987
17065	6	1-3326	17062	6	4	1832	3	2012-08-02 05:15:24.344987
17066	6	1-4117	17062	4	11	1160	35	2012-08-02 05:15:24.344987
17067	6	1-4480	17062	11	64	1697	86	2012-08-02 05:15:24.344987
17062	6	1-4903	17062	64	63	4332	49	2012-08-02 05:15:24.344987
17069	4	1-4754	17068	58	94	2294	42	2012-08-02 05:15:24.731268
17070	4	1-4266	17068	94	89	632	45	2012-08-02 05:15:24.731268
17068	4	1-4904	17068	12	58	2496	92	2012-08-02 05:15:24.731268
17071	4	1-4685	17068	89	12	109	29	2012-08-02 05:15:24.731268
17073	2	1-4894	17072	87	82	6360	60	2012-08-02 05:15:28.189303
17072	2	1-4913	17072	82	87	7195	77	2012-08-02 05:15:28.189303
17075	3	1-4898	17074	100	17	2957	75	2012-08-02 05:15:28.575814
17076	3	1-3207	17074	17	64	2450	41	2012-08-02 05:15:28.575814
17074	3	1-4914	17074	64	100	4296	96	2012-08-02 05:15:28.575814
17078	4	1-4898	17077	100	58	2835	75	2012-08-02 05:15:28.575814
17079	4	1-3748	17077	58	84	5688	77	2012-08-02 05:15:28.575814
17080	4	1-2410	17077	84	64	1218	41	2012-08-02 05:15:28.575814
17077	4	1-4914	17077	64	100	2966	96	2012-08-02 05:15:28.575814
17082	4	1-4409	17081	3	1	2013	20	2012-08-02 05:15:28.575814
17083	4	1-3847	17081	1	23	678	29	2012-08-02 05:15:28.575814
17084	4	1-4701	17081	23	64	725	41	2012-08-02 05:15:28.575814
17081	4	1-4914	17081	64	3	1805	96	2012-08-02 05:15:28.575814
17086	4	1-4551	17085	73	78	1696	73	2012-08-02 05:15:28.97407
17087	4	1-1460	17085	78	40	365	12	2012-08-02 05:15:28.97407
17088	4	1-4403	17085	40	65	342	57	2012-08-02 05:15:28.97407
17085	4	1-4915	17085	65	73	1271	10	2012-08-02 05:15:28.97407
17090	4	1-2740	17089	10	14	57	41	2012-08-02 05:15:28.97407
17091	4	1-4513	17089	14	68	21	43	2012-08-02 05:15:28.97407
17092	4	1-4632	17089	68	65	44	57	2012-08-02 05:15:28.97407
17089	4	1-4915	17089	65	10	252	10	2012-08-02 05:15:28.97407
17094	4	1-4221	17093	61	14	99	41	2012-08-02 05:15:28.97407
17095	4	1-4513	17093	14	68	44	43	2012-08-02 05:15:28.97407
17096	4	1-4632	17093	68	65	106	57	2012-08-02 05:15:28.97407
17093	4	1-4915	17093	65	61	725	10	2012-08-02 05:15:28.97407
17098	5	1-3270	17097	49	94	274	42	2012-08-02 05:15:29.37165
17099	5	1-4266	17097	94	90	183	45	2012-08-02 05:15:29.37165
17100	5	1-4851	17097	90	31	375	55	2012-08-02 05:15:29.37165
17101	5	1-3770	17097	31	26	262	32	2012-08-02 05:15:29.37165
17097	5	1-4916	17097	26	49	454	24	2012-08-02 05:15:29.37165
17103	3	1-4762	17102	8	73	1750	89	2012-08-02 05:15:29.7694
17104	3	1-4448	17102	73	17	3623	53	2012-08-02 05:15:29.7694
17102	3	1-4917	17102	17	8	3026	94	2012-08-02 05:15:29.7694
17106	5	1-4762	17105	8	78	1261	89	2012-08-02 05:15:29.7694
17107	5	1-4809	17105	78	87	1256	97	2012-08-02 05:15:29.7694
17108	5	1-4428	17105	87	4	1003	7	2012-08-02 05:15:29.7694
17109	5	1-4912	17105	4	17	1494	53	2012-08-02 05:15:29.7694
17105	5	1-4917	17105	17	8	1649	94	2012-08-02 05:15:29.7694
17111	5	1-4093	17110	73	89	3831	78	2012-08-02 05:15:32.100499
17112	5	1-4504	17110	89	95	2011	2	2012-08-02 05:15:32.100499
17113	5	1-4892	17110	95	2	1549	33	2012-08-02 05:15:32.100499
17114	5	1-4392	17110	2	31	612	23	2012-08-02 05:15:32.100499
17110	5	1-4923	17110	31	73	7026	64	2012-08-02 05:15:32.100499
17116	5	1-578	17115	29	48	33	94	2012-08-02 05:15:32.100499
17117	5	1-4137	17115	48	90	17	45	2012-08-02 05:15:32.100499
17118	5	1-4851	17115	90	42	26	55	2012-08-02 05:15:32.100499
17119	5	1-3929	17115	42	31	16	23	2012-08-02 05:15:32.100499
17115	5	1-4923	17115	31	29	204	64	2012-08-02 05:15:32.100499
17121	5	1-4025	17120	52	85	2749	81	2012-08-02 05:15:32.487324
17122	5	1-2999	17120	85	63	1511	84	2012-08-02 05:15:32.487324
17123	5	1-4696	17120	63	16	1176	64	2012-08-02 05:15:32.487324
17124	5	1-4832	17120	16	67	1722	41	2012-08-02 05:15:32.487324
17120	5	1-4924	17120	67	52	2784	92	2012-08-02 05:15:32.487324
17126	5	1-4025	17125	52	95	456	81	2012-08-02 05:15:32.487324
17127	5	1-3808	17125	95	17	257	21	2012-08-02 05:15:32.487324
17128	5	1-4140	17125	17	47	135	26	2012-08-02 05:15:32.487324
17129	5	1-4834	17125	47	67	245	41	2012-08-02 05:15:32.487324
17125	5	1-4924	17125	67	52	427	92	2012-08-02 05:15:32.487324
17131	5	1-4754	17130	58	20	1041	42	2012-08-02 05:15:32.487324
17132	5	1-4070	17130	20	63	213	84	2012-08-02 05:15:32.487324
17133	5	1-4696	17130	63	16	182	64	2012-08-02 05:15:32.487324
17134	5	1-4832	17130	16	67	293	41	2012-08-02 05:15:32.487324
17130	5	1-4924	17130	67	58	519	92	2012-08-02 05:15:32.487324
17136	4	1-4170	17135	43	17	1454	21	2012-08-02 05:15:32.487324
17137	4	1-4140	17135	17	47	822	26	2012-08-02 05:15:32.487324
17138	4	1-4834	17135	47	67	1599	41	2012-08-02 05:15:32.487324
17135	4	1-4924	17135	67	43	3002	92	2012-08-02 05:15:32.487324
17140	5	1-3218	17139	99	7	1638	47	2012-08-02 05:15:32.898334
17141	5	1-4427	17139	7	20	1671	6	2012-08-02 05:15:32.898334
17142	5	1-3393	17139	20	60	730	86	2012-08-02 05:15:32.898334
17143	5	1-4899	17139	60	97	3723	7	2012-08-02 05:15:32.898334
17139	5	1-4925	17139	97	99	3120	23	2012-08-02 05:15:32.898334
17145	5	1-4139	17144	52	96	1719	2	2012-08-02 05:15:36.024888
17146	5	1-4889	17144	96	94	1283	68	2012-08-02 05:15:36.024888
17147	5	1-4777	17144	94	30	1934	93	2012-08-02 05:15:36.024888
17148	5	1-4533	17144	30	94	1415	97	2012-08-02 05:15:36.024888
17144	5	1-4933	17144	94	52	2143	1	2012-08-02 05:15:36.024888
17150	2	1-3668	17149	93	12	4508	10	2012-08-02 05:15:36.809367
17149	2	1-4935	17149	12	93	6383	75	2012-08-02 05:15:36.809367
17152	5	1-4622	17151	67	90	89	65	2012-08-02 05:15:36.809367
17153	5	1-4281	17151	90	63	76	84	2012-08-02 05:15:36.809367
17154	5	1-4696	17151	63	100	43	64	2012-08-02 05:15:36.809367
17155	5	1-4881	17151	100	12	58	10	2012-08-02 05:15:36.809367
17151	5	1-4935	17151	12	67	103	75	2012-08-02 05:15:36.809367
17157	4	1-3877	17156	24	94	936	42	2012-08-02 05:15:37.582481
17158	4	1-4833	17156	94	41	880	34	2012-08-02 05:15:37.582481
17159	4	1-4052	17156	41	69	112	79	2012-08-02 05:15:37.582481
17156	4	1-4937	17156	69	24	2260	39	2012-08-02 05:15:37.582481
17161	3	1-4617	17160	63	37	1987	89	2012-08-02 05:15:37.582481
17162	3	1-2838	17160	37	69	374	79	2012-08-02 05:15:37.582481
17160	3	1-4937	17160	69	63	7695	39	2012-08-02 05:15:37.582481
17164	5	1-4889	17163	96	66	408	68	2012-08-02 05:15:38.356255
17165	5	1-4484	17163	66	76	378	54	2012-08-02 05:15:38.356255
17166	5	1-4602	17163	76	47	220	92	2012-08-02 05:15:38.356255
17167	5	1-4698	17163	47	83	195	87	2012-08-02 05:15:38.356255
17163	5	1-4939	17163	83	96	551	2	2012-08-02 05:15:38.356255
17169	5	1-4892	17168	95	55	843	33	2012-08-02 05:15:38.356255
17170	5	1-3259	17168	55	97	351	52	2012-08-02 05:15:38.356255
17171	5	1-4440	17168	97	47	266	92	2012-08-02 05:15:38.356255
17172	5	1-4698	17168	47	83	245	87	2012-08-02 05:15:38.356255
17168	5	1-4939	17168	83	95	722	2	2012-08-02 05:15:38.356255
17174	5	1-4892	17173	95	75	4271	33	2012-08-02 05:15:38.356255
17175	5	1-1009	17173	75	16	1668	72	2012-08-02 05:15:38.356255
17176	5	1-2526	17173	16	30	1125	5	2012-08-02 05:15:38.356255
17177	5	1-4736	17173	30	83	1165	87	2012-08-02 05:15:38.356255
17173	5	1-4939	17173	83	95	3545	2	2012-08-02 05:15:38.356255
17179	5	1-3583	17178	21	29	410	51	2012-08-02 05:15:38.742695
17180	5	1-3925	17178	29	48	41	8	2012-08-02 05:15:38.742695
17181	5	1-4822	17178	48	54	45	24	2012-08-02 05:15:38.742695
17182	5	1-3895	17178	54	35	46	20	2012-08-02 05:15:38.742695
17178	5	1-4940	17178	35	21	621	88	2012-08-02 05:15:38.742695
17184	4	1-4592	17183	38	19	3314	11	2012-08-02 05:15:38.742695
17185	4	1-4793	17183	19	31	539	9	2012-08-02 05:15:38.742695
17186	4	1-4280	17183	31	35	236	20	2012-08-02 05:15:38.742695
17183	4	1-4940	17183	35	38	3571	88	2012-08-02 05:15:38.742695
17188	5	1-4575	17187	97	100	389	95	2012-08-02 05:15:38.742695
17189	5	1-4385	17187	100	26	145	32	2012-08-02 05:15:38.742695
17190	5	1-4916	17187	26	54	224	24	2012-08-02 05:15:38.742695
17191	5	1-3895	17187	54	35	263	20	2012-08-02 05:15:38.742695
17187	5	1-4940	17187	35	97	4000	88	2012-08-02 05:15:38.742695
17193	4	1-4289	17192	27	24	551	74	2012-08-02 05:15:38.742695
17194	4	1-4217	17192	24	31	183	9	2012-08-02 05:15:38.742695
17195	4	1-4280	17192	31	35	83	20	2012-08-02 05:15:38.742695
17192	4	1-4940	17192	35	27	1296	88	2012-08-02 05:15:38.742695
17197	5	1-3900	17196	37	75	93	16	2012-08-02 05:15:41.880488
17198	5	1-4645	17196	75	100	108	95	2012-08-02 05:15:41.880488
17199	5	1-4385	17196	100	26	42	32	2012-08-02 05:15:41.880488
17200	5	1-4916	17196	26	17	68	24	2012-08-02 05:15:41.880488
17196	5	1-4948	17196	17	37	96	60	2012-08-02 05:15:41.880488
17202	5	1-2785	17201	53	39	347	7	2012-08-02 05:15:41.880488
17203	5	1-4879	17201	39	76	331	99	2012-08-02 05:15:41.880488
17204	5	1-2015	17201	76	26	159	32	2012-08-02 05:15:41.880488
17205	5	1-4916	17201	26	17	281	24	2012-08-02 05:15:41.880488
17201	5	1-4948	17201	17	53	432	60	2012-08-02 05:15:41.880488
17207	4	1-4494	17206	82	54	505	21	2012-08-02 05:15:42.267241
17208	4	1-4429	17206	54	29	640	85	2012-08-02 05:15:42.267241
17209	4	1-4396	17206	29	24	368	39	2012-08-02 05:15:42.267241
17206	4	1-4949	17206	24	82	568	72	2012-08-02 05:15:42.267241
17211	3	1-4792	17210	6	24	2645	39	2012-08-02 05:15:42.665257
17212	3	1-4949	17210	24	13	3582	72	2012-08-02 05:15:42.665257
17210	3	1-4950	17210	13	6	6634	42	2012-08-02 05:15:42.665257
17214	2	1-4916	17213	26	72	4367	24	2012-08-02 05:15:43.074054
17213	2	1-4951	17213	72	26	3137	32	2012-08-02 05:15:43.074054
17216	5	1-4342	17215	61	97	3525	85	2012-08-02 05:15:45.437797
17217	5	1-4335	17215	97	41	1231	47	2012-08-02 05:15:45.437797
17218	5	1-4932	17215	41	55	4118	70	2012-08-02 05:15:45.437797
17219	5	1-3346	17215	55	68	2864	97	2012-08-02 05:15:45.437797
17215	5	1-4957	17215	68	61	2965	36	2012-08-02 05:15:45.437797
17221	5	1-3601	17220	65	57	110	55	2012-08-02 05:15:46.996014
17222	5	1-3317	17220	57	83	61	33	2012-08-02 05:15:46.996014
17223	5	1-4583	17220	83	26	61	63	2012-08-02 05:15:46.996014
17224	5	1-4893	17220	26	97	74	91	2012-08-02 05:15:46.996014
17220	5	1-4961	17220	97	65	100	98	2012-08-02 05:15:46.996014
17226	5	1-3876	17225	60	6	3146	95	2012-08-02 05:15:46.996014
17227	5	1-4496	17225	6	83	1691	33	2012-08-02 05:15:46.996014
17228	5	1-4583	17225	83	26	1702	63	2012-08-02 05:15:46.996014
17229	5	1-4893	17225	26	97	2078	91	2012-08-02 05:15:46.996014
17225	5	1-4961	17225	97	60	2825	98	2012-08-02 05:15:46.996014
17231	3	1-3975	17230	15	8	627	97	2012-08-02 05:15:48.155647
17232	3	1-4956	17230	8	31	347	61	2012-08-02 05:15:48.155647
17230	3	1-4964	17230	31	15	2101	43	2012-08-02 05:15:48.155647
17234	4	1-4944	17233	30	82	2377	62	2012-08-02 05:15:48.155647
17235	4	1-4474	17233	82	30	936	20	2012-08-02 05:15:48.155647
17236	4	1-4824	17233	30	31	401	61	2012-08-02 05:15:48.155647
17233	4	1-4964	17233	31	30	2817	43	2012-08-02 05:15:48.155647
17238	5	1-3666	17237	9	77	2022	13	2012-08-02 05:15:49.735428
17239	5	1-4780	17237	77	79	1512	64	2012-08-02 05:15:49.735428
17240	5	1-3863	17237	79	33	673	29	2012-08-02 05:15:49.735428
17241	5	1-4818	17237	33	32	544	94	2012-08-02 05:15:49.735428
17237	5	1-4968	17237	32	9	3989	93	2012-08-02 05:15:49.735428
17243	5	1-4812	17242	25	71	627	30	2012-08-02 05:15:49.735428
17244	5	1-4497	17242	71	7	321	27	2012-08-02 05:15:49.735428
17245	5	1-4890	17242	7	22	444	60	2012-08-02 05:15:49.735428
17246	5	1-4863	17242	22	32	333	94	2012-08-02 05:15:49.735428
17242	5	1-4968	17242	32	25	2495	93	2012-08-02 05:15:49.735428
17248	5	1-4869	17247	48	72	516	38	2012-08-02 05:15:50.144595
17249	5	1-4088	17247	72	84	407	4	2012-08-02 05:15:50.144595
17250	5	1-3510	17247	84	11	423	67	2012-08-02 05:15:50.144595
17251	5	1-4366	17247	11	34	991	40	2012-08-02 05:15:50.144595
17247	5	1-4969	17247	34	48	466	100	2012-08-02 05:15:50.144595
17253	4	1-4930	17252	88	70	922	79	2012-08-02 05:15:50.531131
17254	4	1-4847	17252	70	24	2553	70	2012-08-02 05:15:50.531131
17255	4	1-3761	17252	24	2	671	59	2012-08-02 05:15:50.531131
17252	4	1-4970	17252	2	88	822	5	2012-08-02 05:15:50.531131
17257	5	1-3048	17256	71	66	1073	83	2012-08-02 05:15:52.100088
17258	5	1-4931	17256	66	62	488	79	2012-08-02 05:15:52.100088
17259	5	1-3032	17256	62	55	278	2	2012-08-02 05:15:52.100088
17260	5	1-4253	17256	55	67	498	17	2012-08-02 05:15:52.100088
17256	5	1-4974	17256	67	71	1725	10	2012-08-02 05:15:52.100088
17262	5	1-1595	17261	56	87	2713	77	2012-08-02 05:15:52.486725
17263	5	1-4894	17261	87	22	2579	60	2012-08-02 05:15:52.486725
17264	5	1-4863	17261	22	39	2138	94	2012-08-02 05:15:52.486725
17265	5	1-4764	17261	39	6	5055	15	2012-08-02 05:15:52.486725
17261	5	1-4975	17261	6	56	3758	86	2012-08-02 05:15:52.486725
17267	5	1-4899	17266	60	38	276	7	2012-08-02 05:15:52.486725
17268	5	1-2987	17266	38	48	35	8	2012-08-02 05:15:52.486725
17269	5	1-4822	17266	48	21	44	24	2012-08-02 05:15:52.486725
17270	5	1-4063	17266	21	6	75	15	2012-08-02 05:15:52.486725
17266	5	1-4975	17266	6	60	62	86	2012-08-02 05:15:52.486725
17272	4	1-2482	17271	63	22	78	60	2012-08-02 05:15:52.486725
17273	4	1-4863	17271	22	39	73	94	2012-08-02 05:15:52.486725
17274	4	1-4764	17271	39	6	195	15	2012-08-02 05:15:52.486725
17271	4	1-4975	17271	6	63	164	86	2012-08-02 05:15:52.486725
17276	5	1-3699	17275	7	74	1783	76	2012-08-02 05:15:54.420207
17275	5	1-4980	17275	92	7	1777	2	2012-08-02 05:15:54.420207
17277	5	1-4856	17275	74	26	2111	74	2012-08-02 05:15:54.420207
17278	5	1-3855	17275	26	61	491	54	2012-08-02 05:15:54.420207
17279	5	1-4646	17275	61	92	582	78	2012-08-02 05:15:54.420207
17281	3	1-4869	17280	48	78	448	38	2012-08-02 05:15:54.818026
17282	3	1-3825	17280	78	82	72	26	2012-08-02 05:15:54.818026
17280	3	1-4981	17280	82	48	759	100	2012-08-02 05:15:54.818026
17284	4	1-3959	17283	22	60	571	13	2012-08-02 05:15:54.818026
17285	4	1-4389	17283	60	17	238	21	2012-08-02 05:15:54.818026
17286	4	1-4140	17283	17	82	102	26	2012-08-02 05:15:54.818026
17283	4	1-4981	17283	82	22	1614	100	2012-08-02 05:15:54.818026
17287	1	1-4140	\N	17	17	424	26	2012-08-02 05:15:54.818026
17289	3	1-4786	17288	22	52	250	60	2012-08-02 05:15:56.066553
17290	3	1-4029	17288	52	32	45	35	2012-08-02 05:15:56.066553
17288	3	1-4984	17288	32	22	542	50	2012-08-02 05:15:56.066553
17292	5	1-3456	17291	54	6	2109	15	2012-08-02 05:15:56.066553
17293	5	1-4975	17291	6	27	1837	86	2012-08-02 05:15:56.066553
17294	5	1-4928	17291	27	94	1668	9	2012-08-02 05:15:56.066553
17295	5	1-2710	17291	94	32	359	35	2012-08-02 05:15:56.066553
17291	5	1-4984	17291	32	54	5643	50	2012-08-02 05:15:56.066553
17297	3	1-4669	17296	92	35	843	61	2012-08-02 05:15:56.464244
17298	3	1-4103	17296	35	33	1662	46	2012-08-02 05:15:56.464244
17296	3	1-4985	17296	33	92	2262	31	2012-08-02 05:15:56.464244
17300	5	1-3959	17299	22	77	753	13	2012-08-02 05:15:58.000139
17301	5	1-4780	17299	77	16	560	64	2012-08-02 05:15:58.000139
17302	5	1-4832	17299	16	71	721	41	2012-08-02 05:15:58.000139
17303	5	1-4199	17299	71	12	308	52	2012-08-02 05:15:58.000139
17299	5	1-4989	17299	12	22	2123	100	2012-08-02 05:15:58.000139
17305	3	1-4901	17304	7	35	4362	72	2012-08-02 05:15:58.784306
17306	3	1-4779	17304	35	38	2810	43	2012-08-02 05:15:58.784306
17304	3	1-4991	17304	38	7	2211	3	2012-08-02 05:15:58.784306
17308	5	1-4404	17307	98	60	342	77	2012-08-02 05:15:59.170989
17309	5	1-4177	17307	60	38	192	43	2012-08-02 05:15:59.170989
17310	5	1-4991	17307	38	21	209	3	2012-08-02 05:15:59.170989
17311	5	1-4115	17307	21	93	133	47	2012-08-02 05:15:59.170989
17307	5	1-4992	17307	93	98	534	61	2012-08-02 05:15:59.170989
17313	3	1-3495	17312	51	97	1248	85	2012-08-02 05:15:59.170989
17314	3	1-4335	17312	97	93	439	47	2012-08-02 05:15:59.170989
17312	3	1-4992	17312	93	51	1975	61	2012-08-02 05:15:59.170989
17316	4	1-3863	17315	79	56	1923	29	2012-08-02 05:15:59.568852
17317	4	1-4907	17315	56	90	1389	82	2012-08-02 05:15:59.568852
17318	4	1-4849	17315	90	16	3455	74	2012-08-02 05:15:59.568852
17315	4	1-4993	17315	16	79	4388	64	2012-08-02 05:15:59.568852
17320	5	1-156	17319	51	7	211	47	2012-08-02 05:15:59.568852
17321	5	1-4427	17319	7	93	178	6	2012-08-02 05:15:59.568852
17322	5	1-4815	17319	93	36	784	90	2012-08-02 05:15:59.568852
17323	5	1-4719	17319	36	16	1133	74	2012-08-02 05:15:59.568852
17319	5	1-4993	17319	16	51	1735	64	2012-08-02 05:15:59.568852
17325	4	1-4931	17324	66	72	406	79	2012-08-02 05:16:01.225997
17326	4	1-4805	17324	72	27	2000	88	2012-08-02 05:16:01.225997
17327	4	1-4289	17324	27	1	950	74	2012-08-02 05:16:01.225997
17324	4	1-4997	17324	1	66	775	83	2012-08-02 05:16:01.225997
17329	5	1-3599	17328	67	67	16	21	2012-08-02 05:16:02.783704
17330	5	1-4626	17328	67	63	6	84	2012-08-02 05:16:02.783704
17331	5	1-4696	17328	63	16	4	64	2012-08-02 05:16:02.783704
17332	5	1-4832	17328	16	32	6	41	2012-08-02 05:16:02.783704
17328	5	1-5001	17328	32	67	30	92	2012-08-02 05:16:02.783704
17334	5	1-3932	17333	23	47	2053	69	2012-08-02 05:16:02.783704
17335	5	1-2725	17333	47	54	2498	57	2012-08-02 05:16:02.783704
17336	5	1-4945	17333	54	47	367	26	2012-08-02 05:16:02.783704
17337	5	1-4834	17333	47	32	700	41	2012-08-02 05:16:02.783704
17333	5	1-5001	17333	32	23	4096	92	2012-08-02 05:16:02.783704
17339	4	1-3824	17338	27	46	3467	81	2012-08-02 05:16:02.783704
17340	4	1-4184	17338	46	70	1232	44	2012-08-02 05:16:02.783704
17341	4	1-4878	17338	70	32	868	41	2012-08-02 05:16:02.783704
17338	4	1-5001	17338	32	27	5333	92	2012-08-02 05:16:02.783704
17343	5	1-3060	17342	86	93	518	47	2012-08-02 05:16:03.194659
17344	5	1-4992	17342	93	3	2327	61	2012-08-02 05:16:03.194659
17345	5	1-59	17342	3	15	1258	68	2012-08-02 05:16:03.194659
17346	5	1-4328	17342	15	83	2005	11	2012-08-02 05:16:03.194659
17342	5	1-5002	17342	83	86	3340	50	2012-08-02 05:16:03.194659
17348	6	1-4589	17347	23	65	419	63	2012-08-02 05:16:04.36569
17349	6	1-4926	17347	65	50	245	39	2012-08-02 05:16:04.36569
17350	6	1-2665	17347	50	55	48	47	2012-08-02 05:16:04.36569
17351	6	1-4565	17347	55	32	57	52	2012-08-02 05:16:04.36569
17352	6	1-4848	17347	32	47	100	66	2012-08-02 05:16:04.36569
17347	6	1-5005	17347	47	23	353	12	2012-08-02 05:16:04.36569
17354	5	1-4864	17353	4	28	5201	93	2012-08-02 05:16:04.36569
17355	5	1-4755	17353	28	71	1421	41	2012-08-02 05:16:04.36569
17356	5	1-4199	17353	71	32	608	52	2012-08-02 05:16:04.36569
17357	5	1-4848	17353	32	47	1093	66	2012-08-02 05:16:04.36569
17353	5	1-5005	17353	47	4	3965	12	2012-08-02 05:16:04.36569
17359	6	1-4936	17358	53	51	810	55	2012-08-02 05:16:04.36569
17360	6	1-4482	17358	51	70	300	44	2012-08-02 05:16:04.36569
17361	6	1-4878	17358	70	71	180	41	2012-08-02 05:16:04.36569
17362	6	1-4199	17358	71	32	80	52	2012-08-02 05:16:04.36569
17363	6	1-4848	17358	32	47	148	66	2012-08-02 05:16:04.36569
17358	6	1-5005	17358	47	53	553	12	2012-08-02 05:16:04.36569
17365	5	1-4147	17364	11	19	1594	27	2012-08-02 05:16:04.36569
17366	5	1-4692	17364	19	71	1426	41	2012-08-02 05:16:04.36569
17367	5	1-4199	17364	71	32	662	52	2012-08-02 05:16:04.36569
17368	5	1-4848	17364	32	47	1291	66	2012-08-02 05:16:04.36569
17364	5	1-5005	17364	47	11	5081	12	2012-08-02 05:16:04.36569
17370	5	1-4847	17369	70	41	1992	70	2012-08-02 05:16:04.763589
17371	5	1-4614	17369	41	40	1888	46	2012-08-02 05:16:04.763589
17372	5	1-1589	17369	40	12	306	52	2012-08-02 05:16:04.763589
17373	5	1-4989	17369	12	59	2490	100	2012-08-02 05:16:04.763589
17369	5	1-5006	17369	59	70	869	79	2012-08-02 05:16:04.763589
17375	5	1-3900	17374	37	30	1714	16	2012-08-02 05:16:05.957841
17376	5	1-4954	17374	30	32	833	52	2012-08-02 05:16:05.957841
17377	5	1-4848	17374	32	87	2050	66	2012-08-02 05:16:05.957841
17378	5	1-4410	17374	87	52	2714	100	2012-08-02 05:16:05.957841
17374	5	1-5009	17374	52	37	1664	60	2012-08-02 05:16:05.957841
17380	5	1-4979	17379	45	6	4106	15	2012-08-02 05:16:06.753799
17381	5	1-4975	17379	6	80	2676	86	2012-08-02 05:16:06.753799
17379	5	1-5011	17379	93	45	3806	28	2012-08-02 05:16:06.753799
17382	5	1-1812	17379	80	54	1529	32	2012-08-02 05:16:06.753799
17383	5	1-4874	17379	54	93	839	61	2012-08-02 05:16:06.753799
17385	5	1-4979	17384	45	8	1216	15	2012-08-02 05:16:06.753799
17386	5	1-5003	17384	8	85	349	44	2012-08-02 05:16:06.753799
17387	5	1-3915	17384	85	8	245	97	2012-08-02 05:16:06.753799
17388	5	1-4956	17384	8	93	218	61	2012-08-02 05:16:06.753799
17384	5	1-5011	17384	93	45	1056	28	2012-08-02 05:16:06.753799
17390	5	1-4078	17389	67	46	2648	58	2012-08-02 05:16:06.753799
17391	5	1-4452	17389	46	85	1063	44	2012-08-02 05:16:06.753799
17392	5	1-3915	17389	85	8	798	97	2012-08-02 05:16:06.753799
17393	5	1-4956	17389	8	93	648	61	2012-08-02 05:16:06.753799
17389	5	1-5011	17389	93	67	3346	28	2012-08-02 05:16:06.753799
17395	4	1-4722	17394	92	31	592	55	2012-08-02 05:16:06.753799
17396	4	1-3770	17394	31	54	379	32	2012-08-02 05:16:06.753799
17397	4	1-4874	17394	54	93	265	61	2012-08-02 05:16:06.753799
17394	4	1-5011	17394	93	92	1536	28	2012-08-02 05:16:06.753799
17399	4	1-3286	17398	38	90	296	45	2012-08-02 05:16:08.797785
17400	4	1-4851	17398	90	10	611	55	2012-08-02 05:16:08.797785
17401	4	1-2794	17398	10	13	612	17	2012-08-02 05:16:08.797785
17398	4	1-5016	17398	13	38	372	19	2012-08-02 05:16:08.797785
17403	3	1-3589	17402	74	54	184	44	2012-08-02 05:16:09.206626
17404	3	1-3195	17402	54	28	257	51	2012-08-02 05:16:09.206626
17402	3	1-5017	17402	28	74	190	19	2012-08-02 05:16:09.206626
17406	5	1-4952	17405	76	78	4541	26	2012-08-02 05:16:09.206626
17407	5	1-3742	17405	78	68	4942	97	2012-08-02 05:16:09.206626
17408	5	1-4957	17405	68	45	5215	36	2012-08-02 05:16:09.206626
17409	5	1-4891	17405	45	28	5204	51	2012-08-02 05:16:09.206626
17405	5	1-5017	17405	28	76	3969	19	2012-08-02 05:16:09.206626
17411	5	1-4881	17410	100	48	2447	10	2012-08-02 05:16:11.283852
17412	5	1-3040	17410	48	38	107	43	2012-08-02 05:16:11.283852
17413	5	1-4991	17410	38	78	106	3	2012-08-02 05:16:11.283852
17414	5	1-4434	17410	78	70	79	29	2012-08-02 05:16:11.283852
17410	5	1-5022	17410	70	100	1363	64	2012-08-02 05:16:11.283852
17416	5	1-156	17415	51	93	425	47	2012-08-02 05:16:11.283852
17417	5	1-4992	17415	93	5	1613	61	2012-08-02 05:16:11.283852
17418	5	1-4498	17415	5	1	639	20	2012-08-02 05:16:11.283852
17419	5	1-3847	17415	1	70	198	29	2012-08-02 05:16:11.283852
17415	5	1-5022	17415	70	51	3571	64	2012-08-02 05:16:11.283852
17421	5	1-1501	17420	81	19	149	15	2012-08-02 05:16:11.283852
17422	5	1-3489	17420	19	38	16	43	2012-08-02 05:16:11.283852
17423	5	1-4991	17420	38	78	16	3	2012-08-02 05:16:11.283852
17424	5	1-4434	17420	78	70	13	29	2012-08-02 05:16:11.283852
17420	5	1-5022	17420	70	81	232	64	2012-08-02 05:16:11.283852
17426	3	1-2728	17425	93	28	2179	51	2012-08-02 05:16:11.703732
17427	3	1-5017	17425	28	46	1289	19	2012-08-02 05:16:11.703732
17425	3	1-5023	17425	46	93	3266	81	2012-08-02 05:16:11.703732
17429	5	1-4984	17428	32	14	1128	50	2012-08-02 05:16:12.908004
17430	5	1-2232	17428	14	38	190	88	2012-08-02 05:16:12.908004
17431	5	1-4592	17428	38	28	154	11	2012-08-02 05:16:12.908004
17432	5	1-4635	17428	28	66	90	84	2012-08-02 05:16:12.908004
17428	5	1-5026	17428	66	32	89	35	2012-08-02 05:16:12.908004
17434	5	1-4972	17433	93	68	107	70	2012-08-02 05:16:12.908004
17435	5	1-3001	17433	68	38	144	88	2012-08-02 05:16:12.908004
17436	5	1-4592	17433	38	28	137	11	2012-08-02 05:16:12.908004
17437	5	1-4635	17433	28	66	94	84	2012-08-02 05:16:12.908004
17433	5	1-5026	17433	66	93	110	35	2012-08-02 05:16:12.908004
17439	4	1-2275	17438	12	23	463	78	2012-08-02 05:16:12.908004
17440	4	1-3521	17438	23	28	281	11	2012-08-02 05:16:12.908004
17441	4	1-4635	17438	28	66	217	84	2012-08-02 05:16:12.908004
17438	4	1-5026	17438	66	12	284	35	2012-08-02 05:16:12.908004
17443	5	1-2275	17442	12	59	766	78	2012-08-02 05:16:12.908004
17444	5	1-4700	17442	59	27	192	71	2012-08-02 05:16:12.908004
17445	5	1-4790	17442	27	90	238	65	2012-08-02 05:16:12.908004
17446	5	1-4281	17442	90	66	347	84	2012-08-02 05:16:12.908004
17442	5	1-5026	17442	66	12	462	35	2012-08-02 05:16:12.908004
17448	4	1-4698	17447	47	83	1024	87	2012-08-02 05:16:13.736439
17449	4	1-4939	17447	83	7	2841	2	2012-08-02 05:16:13.736439
17450	4	1-3699	17447	7	56	2741	76	2012-08-02 05:16:13.736439
17447	4	1-5028	17447	56	47	1179	92	2012-08-02 05:16:13.736439
17452	5	1-2485	17451	59	78	606	73	2012-08-02 05:16:14.134343
17453	5	1-1460	17451	78	53	200	12	2012-08-02 05:16:14.134343
17454	5	1-4936	17451	53	42	338	55	2012-08-02 05:16:14.134343
17455	5	1-3929	17451	42	87	237	23	2012-08-02 05:16:14.134343
17451	5	1-5029	17451	87	59	576	30	2012-08-02 05:16:14.134343
17457	5	1-4675	17456	25	60	828	21	2012-08-02 05:16:14.134343
17458	5	1-3903	17456	60	53	275	12	2012-08-02 05:16:14.134343
17459	5	1-4936	17456	53	42	492	55	2012-08-02 05:16:14.134343
17460	5	1-3929	17456	42	87	364	23	2012-08-02 05:16:14.134343
17456	5	1-5029	17456	87	25	936	30	2012-08-02 05:16:14.134343
17462	4	1-2485	17461	59	11	905	73	2012-08-02 05:16:14.134343
17463	4	1-3812	17461	11	2	498	33	2012-08-02 05:16:14.134343
17464	4	1-4392	17461	2	87	289	23	2012-08-02 05:16:14.134343
17461	4	1-5029	17461	87	59	777	30	2012-08-02 05:16:14.134343
17466	4	1-4497	17465	71	11	2550	27	2012-08-02 05:16:14.134343
17467	4	1-4311	17465	11	97	1867	7	2012-08-02 05:16:14.134343
17468	4	1-4925	17465	97	87	1434	23	2012-08-02 05:16:14.134343
17465	4	1-5029	17465	87	71	3865	30	2012-08-02 05:16:14.134343
17470	5	1-4722	17469	92	31	1349	55	2012-08-02 05:16:14.543307
17471	5	1-3770	17469	31	60	913	32	2012-08-02 05:16:14.543307
17472	5	1-4068	17469	60	13	302	30	2012-08-02 05:16:14.543307
17473	5	1-4855	17469	13	3	1750	80	2012-08-02 05:16:14.543307
17469	5	1-5030	17469	3	92	3313	28	2012-08-02 05:16:14.543307
17475	5	1-3932	17474	23	23	212	69	2012-08-02 05:16:14.941123
17476	5	1-4023	17474	23	44	139	95	2012-08-02 05:16:14.941123
17477	5	1-4788	17474	44	3	131	80	2012-08-02 05:16:14.941123
17478	5	1-5030	17474	3	46	253	28	2012-08-02 05:16:14.941123
17474	5	1-5031	17474	46	23	347	92	2012-08-02 05:16:14.941123
17480	5	1-2965	17479	31	53	1968	12	2012-08-02 05:16:15.338991
17481	5	1-4936	17479	53	10	2288	55	2012-08-02 05:16:15.338991
17482	5	1-2794	17479	10	67	1343	17	2012-08-02 05:16:15.338991
17483	5	1-4974	17479	67	67	4121	10	2012-08-02 05:16:15.338991
17479	5	1-5032	17479	67	31	2925	4	2012-08-02 05:16:15.338991
17485	4	1-4189	17484	47	10	98	38	2012-08-02 05:16:15.338991
17486	4	1-1409	17484	10	67	121	17	2012-08-02 05:16:15.338991
17487	4	1-4974	17484	67	67	438	10	2012-08-02 05:16:15.338991
17484	4	1-5032	17484	67	47	368	4	2012-08-02 05:16:15.338991
17489	5	1-2286	17488	16	98	53	13	2012-08-02 05:16:18.155912
17490	5	1-4059	17488	98	38	26	43	2012-08-02 05:16:18.155912
17491	5	1-4991	17488	38	7	29	3	2012-08-02 05:16:18.155912
17492	5	1-4901	17488	7	23	81	72	2012-08-02 05:16:18.155912
17488	5	1-5039	17488	23	16	66	68	2012-08-02 05:16:18.155912
17494	5	1-4484	17493	66	23	364	54	2012-08-02 05:16:18.155912
17495	5	1-4047	17493	23	6	767	42	2012-08-02 05:16:18.155912
17496	5	1-4792	17493	6	24	331	39	2012-08-02 05:16:18.155912
17497	5	1-4949	17493	24	23	486	72	2012-08-02 05:16:18.155912
17493	5	1-5039	17493	23	66	411	68	2012-08-02 05:16:18.155912
17499	5	1-4988	17498	88	12	182	14	2012-08-02 05:16:19.371326
17500	5	1-2655	17498	12	38	36	43	2012-08-02 05:16:19.371326
17501	5	1-4991	17498	38	78	43	3	2012-08-02 05:16:19.371326
17502	5	1-4434	17498	78	80	38	29	2012-08-02 05:16:19.371326
17498	5	1-5042	17498	80	88	26	27	2012-08-02 05:16:19.371326
17504	4	1-4647	17503	48	86	3124	69	2012-08-02 05:16:20.984519
17505	4	1-1281	17503	86	81	2878	14	2012-08-02 05:16:20.984519
17506	4	1-4621	17503	81	78	668	82	2012-08-02 05:16:20.984519
17503	4	1-5046	17503	78	48	1583	18	2012-08-02 05:16:20.984519
17508	5	1-4218	17507	55	41	3434	87	2012-08-02 05:16:20.984519
17509	5	1-4960	17507	41	14	2942	15	2012-08-02 05:16:20.984519
17510	5	1-4996	17507	14	81	3665	14	2012-08-02 05:16:20.984519
17511	5	1-4621	17507	81	78	865	82	2012-08-02 05:16:20.984519
17507	5	1-5046	17507	78	55	2083	18	2012-08-02 05:16:20.984519
17513	5	1-497	17512	39	78	679	3	2012-08-02 05:16:21.393145
17514	5	1-4434	17512	78	56	615	29	2012-08-02 05:16:21.393145
17515	5	1-4907	17512	56	78	584	82	2012-08-02 05:16:21.393145
17516	5	1-5046	17512	78	24	1369	18	2012-08-02 05:16:21.393145
17512	5	1-5047	17512	24	39	2903	44	2012-08-02 05:16:21.393145
17518	4	1-4965	17517	53	81	1749	14	2012-08-02 05:16:21.393145
17519	4	1-4621	17517	81	78	382	82	2012-08-02 05:16:21.393145
17520	4	1-5046	17517	78	24	851	18	2012-08-02 05:16:21.393145
17517	4	1-5047	17517	24	53	1809	44	2012-08-02 05:16:21.393145
17522	5	1-3915	17521	85	8	2037	97	2012-08-02 05:16:21.393145
17523	5	1-4956	17521	8	35	2039	61	2012-08-02 05:16:21.393145
17524	5	1-4103	17521	35	19	4060	46	2012-08-02 05:16:21.393145
17525	5	1-4866	17521	19	24	966	18	2012-08-02 05:16:21.393145
17521	5	1-5047	17521	24	85	2199	44	2012-08-02 05:16:21.393145
17527	5	1-4875	17526	12	27	1800	25	2012-08-02 05:16:22.199817
17528	5	1-4564	17526	27	56	1360	59	2012-08-02 05:16:22.199817
17529	5	1-4543	17526	56	70	1574	79	2012-08-02 05:16:22.199817
17530	5	1-4847	17526	70	37	3813	70	2012-08-02 05:16:22.199817
17526	5	1-5049	17526	37	12	1170	96	2012-08-02 05:16:22.199817
17532	5	1-4059	17531	98	38	448	43	2012-08-02 05:16:22.597615
17533	5	1-4991	17531	38	21	279	3	2012-08-02 05:16:22.597615
17534	5	1-4115	17531	21	93	102	47	2012-08-02 05:16:22.597615
17535	5	1-4992	17531	93	76	234	61	2012-08-02 05:16:22.597615
17531	5	1-5050	17531	76	98	1664	13	2012-08-02 05:16:22.597615
17537	5	1-808	17536	36	81	1069	51	2012-08-02 05:16:22.597615
17538	5	1-3265	17536	81	21	121	3	2012-08-02 05:16:22.597615
17539	5	1-4115	17536	21	93	80	47	2012-08-02 05:16:22.597615
17540	5	1-4992	17536	93	76	330	61	2012-08-02 05:16:22.597615
17536	5	1-5050	17536	76	36	4219	13	2012-08-02 05:16:22.597615
17542	4	1-4478	17541	9	52	82	79	2012-08-02 05:16:22.597615
17543	4	1-4138	17541	52	100	42	72	2012-08-02 05:16:22.597615
17544	4	1-5021	17541	100	76	31	61	2012-08-02 05:16:22.597615
17541	4	1-5050	17541	76	9	404	13	2012-08-02 05:16:22.597615
17546	4	1-5016	17545	13	22	882	19	2012-08-02 05:16:24.210907
17547	4	1-295	17545	22	8	408	68	2012-08-02 05:16:24.210907
17548	4	1-4745	17545	8	70	1072	83	2012-08-02 05:16:24.210907
17545	4	1-5054	17545	70	13	1589	17	2012-08-02 05:16:24.210907
17550	3	1-3287	17549	88	54	633	57	2012-08-02 05:16:25.823833
17551	3	1-4945	17549	54	72	115	26	2012-08-02 05:16:25.823833
17549	3	1-5058	17549	72	88	782	90	2012-08-02 05:16:25.823833
17553	4	1-4939	17552	83	94	316	2	2012-08-02 05:16:26.654127
17554	4	1-4688	17552	94	6	165	1	2012-08-02 05:16:26.654127
17555	4	1-3945	17552	6	87	242	40	2012-08-02 05:16:26.654127
17552	4	1-5060	17552	87	83	119	87	2012-08-02 05:16:26.654127
17557	2	1-4543	17556	56	16	4180	79	2012-08-02 05:16:27.858154
17556	2	1-5063	17556	16	56	7430	59	2012-08-02 05:16:27.858154
17559	3	1-4970	17558	2	88	509	5	2012-08-02 05:16:27.858154
17560	3	1-4930	17558	88	16	319	79	2012-08-02 05:16:27.858154
17558	3	1-5063	17558	16	2	744	59	2012-08-02 05:16:27.858154
17561	1	1-1032	\N	88	88	645	5	2012-08-02 05:16:27.858154
17563	3	1-4781	17562	95	46	2671	28	2012-08-02 05:16:28.255966
17564	3	1-5031	17562	46	67	2659	92	2012-08-02 05:16:28.255966
17562	3	1-5064	17562	67	95	5904	85	2012-08-02 05:16:28.255966
17566	5	1-4824	17565	30	98	1902	61	2012-08-02 05:16:28.653743
17567	5	1-4404	17565	98	62	1381	77	2012-08-02 05:16:28.653743
17568	5	1-4990	17565	62	62	1720	98	2012-08-02 05:16:28.653743
17569	5	1-4295	17565	62	5	2240	1	2012-08-02 05:16:28.653743
17565	5	1-5065	17565	5	30	3591	20	2012-08-02 05:16:28.653743
17571	3	1-4986	17570	80	38	1085	43	2012-08-02 05:16:29.062886
17572	3	1-4991	17570	38	25	813	3	2012-08-02 05:16:29.062886
17570	3	1-5066	17570	25	80	1532	71	2012-08-02 05:16:29.062886
17574	5	1-5033	17573	29	20	6438	59	2012-08-02 05:16:29.062886
17575	5	1-4341	17573	20	48	2485	8	2012-08-02 05:16:29.062886
17576	5	1-4822	17573	48	6	3162	24	2012-08-02 05:16:29.062886
17577	5	1-3326	17573	6	25	1468	3	2012-08-02 05:16:29.062886
17573	5	1-5066	17573	25	29	4107	71	2012-08-02 05:16:29.062886
17579	3	1-4970	17578	2	65	2152	5	2012-08-02 05:16:29.471472
17580	3	1-4330	17578	65	72	3063	57	2012-08-02 05:16:29.471472
17578	3	1-5067	17578	72	2	3201	59	2012-08-02 05:16:29.471472
17582	5	1-4599	17581	62	24	171	39	2012-08-02 05:16:29.471472
17583	5	1-4949	17581	24	16	198	72	2012-08-02 05:16:29.471472
17584	5	1-2526	17581	16	65	94	5	2012-08-02 05:16:29.471472
17585	5	1-4330	17581	65	72	167	57	2012-08-02 05:16:29.471472
17581	5	1-5067	17581	72	62	219	59	2012-08-02 05:16:29.471472
17587	3	1-4864	17586	4	95	2635	93	2012-08-02 05:16:31.526488
17588	3	1-3725	17586	95	12	517	31	2012-08-02 05:16:31.526488
17586	3	1-5072	17586	12	4	3412	12	2012-08-02 05:16:31.526488
17590	3	1-3202	17589	9	70	1367	53	2012-08-02 05:16:31.526488
17591	3	1-4627	17589	70	12	543	31	2012-08-02 05:16:31.526488
17589	3	1-5072	17589	12	9	4375	12	2012-08-02 05:16:31.526488
17593	5	1-4922	17592	29	17	1915	24	2012-08-02 05:16:32.343791
17594	5	1-4948	17592	17	37	2796	60	2012-08-02 05:16:32.343791
17595	5	1-3900	17592	37	75	2817	16	2012-08-02 05:16:32.343791
17596	5	1-4645	17592	75	25	3411	95	2012-08-02 05:16:32.343791
17592	5	1-5074	17592	25	29	3910	94	2012-08-02 05:16:32.343791
17598	5	1-5058	17597	72	36	2449	90	2012-08-02 05:16:32.741741
17599	5	1-4719	17597	36	24	3007	74	2012-08-02 05:16:32.741741
17600	5	1-4217	17597	24	44	796	9	2012-08-02 05:16:32.741741
17601	5	1-3782	17597	44	96	449	39	2012-08-02 05:16:32.741741
17597	5	1-5075	17597	96	72	504	26	2012-08-02 05:16:32.741741
17603	5	1-195	17602	34	25	195	3	2012-08-02 05:16:32.741741
17604	5	1-5066	17602	25	29	483	71	2012-08-02 05:16:32.741741
17605	5	1-5033	17602	29	62	669	59	2012-08-02 05:16:32.741741
17606	5	1-4599	17602	62	96	596	39	2012-08-02 05:16:32.741741
17602	5	1-5075	17602	96	34	740	26	2012-08-02 05:16:32.741741
17608	3	1-999	17607	36	62	991	59	2012-08-02 05:16:32.741741
17609	3	1-4599	17607	62	96	1007	39	2012-08-02 05:16:32.741741
17607	3	1-5075	17607	96	36	1427	26	2012-08-02 05:16:32.741741
17611	2	1-3092	17610	31	96	4652	39	2012-08-02 05:16:32.741741
17610	2	1-5075	17610	96	31	6932	26	2012-08-02 05:16:32.741741
17613	4	1-4048	17612	23	37	1138	70	2012-08-02 05:16:35.183293
17614	4	1-5049	17612	37	100	397	96	2012-08-02 05:16:35.183293
17615	4	1-4898	17612	100	69	434	75	2012-08-02 05:16:35.183293
17612	4	1-5081	17612	69	23	1377	92	2012-08-02 05:16:35.183293
17617	5	1-2432	17616	40	11	964	35	2012-08-02 05:16:35.63646
17618	5	1-4480	17616	11	64	1058	86	2012-08-02 05:16:35.63646
17619	5	1-4903	17616	64	30	2027	49	2012-08-02 05:16:35.63646
17620	5	1-5014	17616	30	93	1168	66	2012-08-02 05:16:35.63646
17616	5	1-5082	17616	93	40	5386	55	2012-08-02 05:16:35.63646
17622	5	1-3510	17621	84	32	876	67	2012-08-02 05:16:36.862576
17623	5	1-1399	17621	32	78	692	32	2012-08-02 05:16:36.862576
17624	5	1-5053	17621	78	57	1374	62	2012-08-02 05:16:36.862576
17625	5	1-4908	17621	57	13	1470	36	2012-08-02 05:16:36.862576
17621	5	1-5085	17621	13	84	805	4	2012-08-02 05:16:36.862576
17626	1	1-3510	\N	84	84	680	67	2012-08-02 05:16:36.862576
17628	5	1-4583	17627	83	26	508	63	2012-08-02 05:16:37.68008
17629	5	1-4893	17627	26	20	549	91	2012-08-02 05:16:37.68008
17630	5	1-3675	17627	20	98	306	61	2012-08-02 05:16:37.68008
17631	5	1-4404	17627	98	32	198	77	2012-08-02 05:16:37.68008
17627	5	1-5087	17627	32	83	570	33	2012-08-02 05:16:37.68008
17632	1	1-3108	\N	25	25	2484	61	2012-08-02 05:16:37.68008
17634	4	1-4817	17633	88	7	2271	27	2012-08-02 05:16:39.70207
17635	4	1-4890	17633	7	56	3430	60	2012-08-02 05:16:39.70207
17636	4	1-1209	17633	56	26	1651	80	2012-08-02 05:16:39.70207
17633	4	1-5092	17633	26	88	2217	71	2012-08-02 05:16:39.70207
17638	5	1-4817	17637	88	11	2328	27	2012-08-02 05:16:39.70207
17639	5	1-4311	17637	11	97	1721	7	2012-08-02 05:16:39.70207
17640	5	1-4925	17637	97	11	1230	23	2012-08-02 05:16:39.70207
17641	5	1-3746	17637	11	26	1409	80	2012-08-02 05:16:39.70207
17637	5	1-5092	17637	26	88	2073	71	2012-08-02 05:16:39.70207
17643	4	1-5033	17642	29	20	330	59	2012-08-02 05:16:39.70207
17644	4	1-4341	17642	20	55	128	8	2012-08-02 05:16:39.70207
17645	4	1-4240	17642	55	26	137	80	2012-08-02 05:16:39.70207
17642	4	1-5092	17642	26	29	209	71	2012-08-02 05:16:39.70207
17647	5	1-4790	17646	27	91	1159	65	2012-08-02 05:16:39.70207
17648	5	1-4433	17646	91	98	809	49	2012-08-02 05:16:39.70207
17649	5	1-3747	17646	98	62	248	98	2012-08-02 05:16:39.70207
17650	5	1-4999	17646	62	26	644	80	2012-08-02 05:16:39.70207
17646	5	1-5092	17646	26	27	1016	71	2012-08-02 05:16:39.70207
17652	4	1-2822	17651	71	75	682	16	2012-08-02 05:16:39.70207
17653	4	1-4645	17651	75	44	817	95	2012-08-02 05:16:39.70207
17654	4	1-4788	17651	44	26	745	80	2012-08-02 05:16:39.70207
17651	4	1-5092	17651	26	71	1182	71	2012-08-02 05:16:39.70207
17656	5	1-4680	17655	80	85	300	44	2012-08-02 05:16:40.519532
17657	5	1-3915	17655	85	94	227	97	2012-08-02 05:16:40.519532
17658	5	1-4933	17655	94	6	285	1	2012-08-02 05:16:40.519532
17659	5	1-3945	17655	6	12	387	40	2012-08-02 05:16:40.519532
17655	5	1-5094	17655	12	80	457	68	2012-08-02 05:16:40.519532
17661	3	1-4777	17660	94	98	5480	93	2012-08-02 05:16:40.519532
17662	3	1-2565	17660	98	12	3343	40	2012-08-02 05:16:40.519532
17660	3	1-5094	17660	12	94	4164	68	2012-08-02 05:16:40.519532
17664	4	1-4745	17663	8	66	2446	83	2012-08-02 05:16:40.519532
17665	4	1-4931	17663	66	81	1121	79	2012-08-02 05:16:40.519532
17666	4	1-3981	17663	81	12	748	40	2012-08-02 05:16:40.519532
17663	4	1-5094	17663	12	8	958	68	2012-08-02 05:16:40.519532
17668	4	1-5070	17667	27	88	1795	5	2012-08-02 05:16:40.519532
17669	4	1-4930	17667	88	81	1794	79	2012-08-02 05:16:40.519532
17670	4	1-3981	17667	81	12	1229	40	2012-08-02 05:16:40.519532
17667	4	1-5094	17667	12	27	1615	68	2012-08-02 05:16:40.519532
17671	1	1-2565	\N	98	98	933	40	2012-08-02 05:16:40.519532
17673	4	1-4887	17672	34	53	5513	76	2012-08-02 05:16:42.232304
17674	4	1-5010	17672	53	39	3528	7	2012-08-02 05:16:42.232304
17675	4	1-4879	17672	39	31	3120	99	2012-08-02 05:16:42.232304
17672	4	1-5098	17672	31	34	3313	97	2012-08-02 05:16:42.232304
17677	5	1-4933	17676	94	68	893	1	2012-08-02 05:16:42.232304
17678	5	1-1115	17676	68	70	520	42	2012-08-02 05:16:42.232304
17679	5	1-3104	17676	70	73	301	33	2012-08-02 05:16:42.232304
17680	5	1-5093	17676	73	31	513	99	2012-08-02 05:16:42.232304
17676	5	1-5098	17676	31	94	583	97	2012-08-02 05:16:42.232304
17682	3	1-4428	17681	87	39	621	7	2012-08-02 05:16:42.232304
17683	3	1-4879	17681	39	31	595	99	2012-08-02 05:16:42.232304
17681	3	1-5098	17681	31	87	685	97	2012-08-02 05:16:42.232304
17685	5	1-4957	17684	68	61	1535	36	2012-08-02 05:16:42.232304
17686	5	1-4342	17684	61	60	1902	85	2012-08-02 05:16:42.232304
17687	5	1-2684	17684	60	39	1250	7	2012-08-02 05:16:42.232304
17688	5	1-4879	17684	39	31	1217	99	2012-08-02 05:16:42.232304
17684	5	1-5098	17684	31	68	1423	97	2012-08-02 05:16:42.232304
17690	4	1-4952	17689	76	47	866	26	2012-08-02 05:16:43.845218
17691	4	1-4834	17689	47	6	1490	41	2012-08-02 05:16:43.845218
17689	4	1-5102	17689	60	76	914	19	2012-08-02 05:16:43.845218
17692	4	1-3861	17689	6	60	504	32	2012-08-02 05:16:43.845218
17694	5	1-4220	17693	27	81	280	16	2012-08-02 05:16:43.845218
17695	5	1-4542	17693	81	39	236	7	2012-08-02 05:16:43.845218
17696	5	1-4879	17693	39	76	223	99	2012-08-02 05:16:43.845218
17697	5	1-2015	17693	76	60	106	32	2012-08-02 05:16:43.845218
17693	5	1-5102	17693	60	27	231	19	2012-08-02 05:16:43.845218
17699	5	1-5023	17698	46	77	3800	81	2012-08-02 05:16:43.845218
17700	5	1-2745	17698	77	48	417	8	2012-08-02 05:16:43.845218
17701	5	1-4822	17698	48	72	587	24	2012-08-02 05:16:43.845218
17702	5	1-4951	17698	72	60	532	32	2012-08-02 05:16:43.845218
17698	5	1-5102	17698	60	46	1164	19	2012-08-02 05:16:43.845218
17704	5	1-4907	17703	56	27	2947	82	2012-08-02 05:16:44.276653
17705	5	1-4087	17703	27	45	2040	100	2012-08-02 05:16:44.276653
17706	5	1-5068	17703	45	81	1575	52	2012-08-02 05:16:44.276653
17707	5	1-4862	17703	81	78	6923	37	2012-08-02 05:16:44.276653
17703	5	1-5103	17703	78	56	4780	29	2012-08-02 05:16:44.276653
17709	3	1-4783	17708	78	60	403	85	2012-08-02 05:16:44.276653
17710	3	1-4982	17708	60	78	444	37	2012-08-02 05:16:44.276653
17708	3	1-5103	17708	78	78	495	29	2012-08-02 05:16:44.276653
17712	2	1-3140	17711	23	28	1250	17	2012-08-02 05:16:44.696333
17711	2	1-5104	17711	28	23	1279	89	2012-08-02 05:16:44.696333
17714	2	1-1011	17713	96	28	3669	17	2012-08-02 05:16:44.696333
17713	2	1-5104	17713	28	96	4927	89	2012-08-02 05:16:44.696333
17716	2	1-3016	17715	71	28	1961	17	2012-08-02 05:16:44.696333
17715	2	1-5104	17715	28	71	2956	89	2012-08-02 05:16:44.696333
17718	4	1-4311	17717	11	97	1490	7	2012-08-02 05:16:45.491583
17719	4	1-4925	17717	97	84	1216	23	2012-08-02 05:16:45.491583
17720	4	1-3038	17717	84	49	452	12	2012-08-02 05:16:45.491583
17717	4	1-5106	17717	49	11	1914	27	2012-08-02 05:16:45.491583
17722	3	1-578	17721	29	29	910	94	2012-08-02 05:16:45.889547
17723	3	1-4922	17721	29	10	409	24	2012-08-02 05:16:45.889547
17721	3	1-5107	17721	10	29	4669	64	2012-08-02 05:16:45.889547
17725	5	1-4837	17724	10	42	1615	78	2012-08-02 05:16:47.082585
17726	5	1-4860	17724	42	46	1499	28	2012-08-02 05:16:47.082585
17727	5	1-5031	17724	46	27	2079	92	2012-08-02 05:16:47.082585
17728	5	1-3824	17724	27	56	1543	81	2012-08-02 05:16:47.082585
17724	5	1-5110	17724	56	10	1411	65	2012-08-02 05:16:47.082585
17730	3	1-4932	17729	41	41	1729	70	2012-08-02 05:16:47.491806
17731	3	1-4614	17729	41	45	1518	46	2012-08-02 05:16:47.491806
17729	3	1-5111	17729	45	41	651	47	2012-08-02 05:16:47.491806
17733	4	1-4743	17732	50	97	1607	91	2012-08-02 05:16:47.491806
17734	4	1-4961	17732	97	38	1333	98	2012-08-02 05:16:47.491806
17735	4	1-5086	17732	38	45	3600	46	2012-08-02 05:16:47.491806
17732	4	1-5111	17732	45	50	1609	47	2012-08-02 05:16:47.491806
17737	5	1-4427	17736	7	93	301	6	2012-08-02 05:16:47.491806
17738	5	1-4815	17736	93	36	1038	90	2012-08-02 05:16:47.491806
17739	5	1-4719	17736	36	79	1174	74	2012-08-02 05:16:47.491806
17740	5	1-4419	17736	79	45	942	46	2012-08-02 05:16:47.491806
17736	5	1-5111	17736	45	7	455	47	2012-08-02 05:16:47.491806
17742	4	1-4197	17741	75	90	2157	82	2012-08-02 05:16:47.491806
17743	4	1-4849	17741	90	79	6134	74	2012-08-02 05:16:47.491806
17744	4	1-4419	17741	79	45	5955	46	2012-08-02 05:16:47.491806
17741	4	1-5111	17741	45	75	3479	47	2012-08-02 05:16:47.491806
17746	5	1-4743	17745	50	97	1079	91	2012-08-02 05:16:47.491806
17747	5	1-4961	17745	97	71	1287	98	2012-08-02 05:16:47.491806
17748	5	1-4444	17745	71	79	1094	74	2012-08-02 05:16:47.491806
17749	5	1-4419	17745	79	45	1168	46	2012-08-02 05:16:47.491806
17745	5	1-5111	17745	45	50	751	47	2012-08-02 05:16:47.491806
17751	5	1-4684	17750	47	76	16	19	2012-08-02 05:16:47.911406
17752	5	1-4952	17750	76	72	16	26	2012-08-02 05:16:47.911406
17753	5	1-5058	17750	72	79	91	90	2012-08-02 05:16:47.911406
17754	5	1-4843	17750	79	6	99	20	2012-08-02 05:16:47.911406
17750	5	1-5112	17750	6	47	100	30	2012-08-02 05:16:47.911406
17756	5	1-4413	17755	75	31	32	99	2012-08-02 05:16:47.911406
17757	5	1-5098	17755	31	94	33	97	2012-08-02 05:16:47.911406
17758	5	1-4933	17755	94	5	45	1	2012-08-02 05:16:47.911406
17759	5	1-5065	17755	5	6	64	20	2012-08-02 05:16:47.911406
17755	5	1-5112	17755	6	75	68	30	2012-08-02 05:16:47.911406
17761	5	1-4941	17760	48	54	2493	57	2012-08-02 05:16:47.911406
17762	5	1-4945	17760	54	72	403	26	2012-08-02 05:16:47.911406
17763	5	1-5058	17760	72	79	2444	90	2012-08-02 05:16:47.911406
17764	5	1-4843	17760	79	6	2821	20	2012-08-02 05:16:47.911406
17760	5	1-5112	17760	6	48	3023	30	2012-08-02 05:16:47.911406
17766	5	1-4962	17765	51	8	442	22	2012-08-02 05:16:49.922629
17767	5	1-4546	17765	8	71	355	42	2012-08-02 05:16:49.922629
17768	5	1-3893	17765	71	82	136	72	2012-08-02 05:16:49.922629
17769	5	1-4494	17765	82	67	111	21	2012-08-02 05:16:49.922629
17765	5	1-5117	17765	67	51	265	34	2012-08-02 05:16:49.922629
17771	5	1-4375	17770	66	29	930	85	2012-08-02 05:16:49.922629
17772	5	1-4396	17770	29	24	584	39	2012-08-02 05:16:49.922629
17773	5	1-4949	17770	24	82	984	72	2012-08-02 05:16:49.922629
17774	5	1-4494	17770	82	67	955	21	2012-08-02 05:16:49.922629
17770	5	1-5117	17770	67	66	2700	34	2012-08-02 05:16:49.922629
17776	4	1-4786	17775	22	37	979	60	2012-08-02 05:16:52.75043
17777	4	1-3900	17775	37	75	846	16	2012-08-02 05:16:52.75043
17778	4	1-4645	17775	75	78	879	95	2012-08-02 05:16:52.75043
17775	4	1-5124	17775	78	22	1881	50	2012-08-02 05:16:52.75043
17780	4	1-3753	17779	97	67	1691	10	2012-08-02 05:16:54.750482
17781	4	1-5032	17779	67	55	1019	4	2012-08-02 05:16:54.750482
17782	4	1-2609	17779	55	75	710	90	2012-08-02 05:16:54.750482
17779	4	1-5129	17779	75	97	5557	9	2012-08-02 05:16:54.750482
17784	5	1-5081	17783	69	23	2608	92	2012-08-02 05:16:59.226372
17785	5	1-4048	17783	23	41	1835	70	2012-08-02 05:16:59.226372
17786	5	1-4614	17783	41	19	1780	46	2012-08-02 05:16:59.226372
17787	5	1-4866	17783	19	18	367	18	2012-08-02 05:16:59.226372
17783	5	1-5140	17783	18	69	965	75	2012-08-02 05:16:59.226372
17789	5	1-4846	17788	89	98	2422	1	2012-08-02 05:16:59.226372
17790	5	1-4998	17788	98	69	2132	23	2012-08-02 05:16:59.226372
17791	5	1-3745	17788	69	19	1527	46	2012-08-02 05:16:59.226372
17792	5	1-4866	17788	19	18	324	18	2012-08-02 05:16:59.226372
17788	5	1-5140	17788	18	89	880	75	2012-08-02 05:16:59.226372
17794	4	1-3640	17793	57	41	875	70	2012-08-02 05:16:59.226372
17795	4	1-4614	17793	41	19	975	46	2012-08-02 05:16:59.226372
17796	4	1-4866	17793	19	18	231	18	2012-08-02 05:16:59.226372
17793	4	1-5140	17793	18	57	698	75	2012-08-02 05:16:59.226372
17798	5	1-4034	17797	85	13	1554	17	2012-08-02 05:16:59.226372
17799	5	1-5016	17797	13	90	938	19	2012-08-02 05:16:59.226372
17800	5	1-4187	17797	90	97	754	26	2012-08-02 05:16:59.226372
17801	5	1-3788	17797	97	18	267	18	2012-08-02 05:16:59.226372
17797	5	1-5140	17797	18	85	807	75	2012-08-02 05:16:59.226372
17803	3	1-5126	17802	92	39	542	56	2012-08-02 05:16:59.646337
17804	3	1-5052	17802	39	14	983	22	2012-08-02 05:16:59.646337
17802	3	1-5141	17802	14	92	809	93	2012-08-02 05:16:59.646337
17806	5	1-5112	17805	6	52	256	30	2012-08-02 05:17:00.054913
17807	5	1-1308	17805	52	36	138	90	2012-08-02 05:17:00.054913
17808	5	1-4719	17805	36	1	211	74	2012-08-02 05:17:00.054913
17809	5	1-4997	17805	1	16	153	83	2012-08-02 05:17:00.054913
17805	5	1-5142	17805	16	6	240	20	2012-08-02 05:17:00.054913
17811	5	1-5112	17810	6	9	1238	30	2012-08-02 05:17:00.054913
17812	5	1-4799	17810	9	85	862	28	2012-08-02 05:17:00.054913
17813	5	1-4942	17810	85	1	695	74	2012-08-02 05:17:00.054913
17814	5	1-4997	17810	1	16	573	83	2012-08-02 05:17:00.054913
17810	5	1-5142	17810	16	6	1019	20	2012-08-02 05:17:00.054913
17816	5	1-5070	17815	27	88	524	5	2012-08-02 05:17:00.883969
17817	5	1-4930	17815	88	43	370	79	2012-08-02 05:17:00.883969
17818	5	1-4821	17815	43	97	142	91	2012-08-02 05:17:00.883969
17819	5	1-4961	17815	97	96	122	98	2012-08-02 05:17:00.883969
17815	5	1-5144	17815	96	27	667	68	2012-08-02 05:17:00.883969
17821	6	1-4328	17820	15	83	1807	11	2012-08-02 05:17:00.883969
17822	6	1-5002	17820	83	86	2158	50	2012-08-02 05:17:00.883969
17823	6	1-3060	17820	86	50	240	47	2012-08-02 05:17:00.883969
17824	6	1-4743	17820	50	97	275	91	2012-08-02 05:17:00.883969
17825	6	1-4961	17820	97	96	262	98	2012-08-02 05:17:00.883969
17820	6	1-5144	17820	96	15	1582	68	2012-08-02 05:17:00.883969
17827	6	1-4360	17826	87	65	289	94	2012-08-02 05:17:00.883969
17828	6	1-4870	17826	65	32	258	49	2012-08-02 05:17:00.883969
17829	6	1-2869	17826	32	26	72	63	2012-08-02 05:17:00.883969
17830	6	1-4893	17826	26	97	71	91	2012-08-02 05:17:00.883969
17831	6	1-4961	17826	97	96	79	98	2012-08-02 05:17:00.883969
17826	6	1-5144	17826	96	87	549	68	2012-08-02 05:17:00.883969
17833	5	1-4360	17832	87	8	1840	94	2012-08-02 05:17:00.883969
17834	5	1-4762	17832	8	6	1405	89	2012-08-02 05:17:00.883969
17835	5	1-2361	17832	6	97	357	91	2012-08-02 05:17:00.883969
17836	5	1-4961	17832	97	96	426	98	2012-08-02 05:17:00.883969
17832	5	1-5144	17832	96	87	3226	68	2012-08-02 05:17:00.883969
17838	6	1-4360	17837	87	29	1944	94	2012-08-02 05:17:00.883969
17839	6	1-4922	17837	29	53	894	24	2012-08-02 05:17:00.883969
17840	6	1-4188	17837	53	50	252	47	2012-08-02 05:17:00.883969
17841	6	1-4743	17837	50	97	366	91	2012-08-02 05:17:00.883969
17842	6	1-4961	17837	97	96	442	98	2012-08-02 05:17:00.883969
17837	6	1-5144	17837	96	87	3377	68	2012-08-02 05:17:00.883969
17844	3	1-4509	17843	66	45	4243	100	2012-08-02 05:17:01.73457
17845	3	1-5068	17843	45	12	4794	52	2012-08-02 05:17:01.73457
17843	3	1-5146	17843	12	66	5549	22	2012-08-02 05:17:01.73457
17847	3	1-3765	17846	4	39	3281	56	2012-08-02 05:17:04.573768
17848	3	1-5052	17846	39	68	2935	22	2012-08-02 05:17:04.573768
17846	3	1-5153	17846	68	4	8627	6	2012-08-02 05:17:04.573768
17850	5	1-4372	17849	84	73	4421	7	2012-08-02 05:17:04.982905
17851	5	1-5131	17849	73	83	4702	58	2012-08-02 05:17:04.982905
17852	5	1-4884	17849	83	44	2301	95	2012-08-02 05:17:04.982905
17853	5	1-4788	17849	44	84	1507	80	2012-08-02 05:17:04.982905
17849	5	1-5154	17849	84	84	6442	52	2012-08-02 05:17:04.982905
17855	4	1-4328	17854	15	83	391	11	2012-08-02 05:17:05.779432
17856	4	1-5002	17854	83	22	665	50	2012-08-02 05:17:05.779432
17857	4	1-4786	17854	22	11	426	60	2012-08-02 05:17:05.779432
17854	4	1-5156	17854	11	15	241	68	2012-08-02 05:17:05.779432
17859	3	1-4942	17858	85	79	1166	74	2012-08-02 05:17:06.961744
17860	3	1-4419	17858	79	90	1097	46	2012-08-02 05:17:06.961744
17858	3	1-5159	17858	90	85	1894	28	2012-08-02 05:17:06.961744
17862	3	1-4942	17861	85	56	2913	74	2012-08-02 05:17:06.961744
17863	3	1-3941	17861	56	90	2706	46	2012-08-02 05:17:06.961744
17861	3	1-5159	17861	90	85	4702	28	2012-08-02 05:17:06.961744
17865	5	1-4942	17864	85	52	630	74	2012-08-02 05:17:06.961744
17866	5	1-5083	17864	52	62	408	59	2012-08-02 05:17:06.961744
17867	5	1-4599	17864	62	38	416	39	2012-08-02 05:17:06.961744
17868	5	1-4354	17864	38	90	431	46	2012-08-02 05:17:06.961744
17864	5	1-5159	17864	90	85	872	28	2012-08-02 05:17:06.961744
17870	3	1-4190	17869	91	67	3379	10	2012-08-02 05:17:07.746004
17871	3	1-5032	17869	67	14	2674	4	2012-08-02 05:17:07.746004
17869	3	1-5161	17869	14	91	7740	30	2012-08-02 05:17:07.746004
17873	5	1-5044	17872	34	87	818	66	2012-08-02 05:17:07.746004
17874	5	1-4410	17872	87	95	1015	100	2012-08-02 05:17:07.746004
17875	5	1-5020	17872	95	67	328	10	2012-08-02 05:17:07.746004
17876	5	1-5032	17872	67	14	274	4	2012-08-02 05:17:07.746004
17872	5	1-5161	17872	14	34	840	30	2012-08-02 05:17:07.746004
17878	3	1-5092	17877	26	67	2690	71	2012-08-02 05:17:08.541531
17879	3	1-2542	17877	67	73	353	40	2012-08-02 05:17:08.541531
17877	3	1-5163	17877	73	26	6488	80	2012-08-02 05:17:08.541531
17881	3	1-5114	17880	10	61	105	55	2012-08-02 05:17:08.541531
17882	3	1-1385	17880	61	73	34	40	2012-08-02 05:17:08.541531
17880	3	1-5163	17880	73	10	1039	80	2012-08-02 05:17:08.541531
17884	2	1-4353	17883	82	59	422	67	2012-08-02 05:17:09.712619
17883	2	1-5166	17883	59	82	5691	71	2012-08-02 05:17:09.712619
17886	5	1-4171	17885	77	27	466	19	2012-08-02 05:17:09.712619
17887	5	1-4220	17885	27	71	368	16	2012-08-02 05:17:09.712619
17888	5	1-4841	17885	71	74	133	38	2012-08-02 05:17:09.712619
17889	5	1-3476	17885	74	59	37	67	2012-08-02 05:17:09.712619
17885	5	1-5166	17885	59	77	1468	71	2012-08-02 05:17:09.712619
17891	5	1-4790	17890	27	91	775	65	2012-08-02 05:17:09.712619
17892	5	1-4433	17890	91	32	392	49	2012-08-02 05:17:09.712619
17893	5	1-2869	17890	32	3	91	63	2012-08-02 05:17:09.712619
17894	5	1-5158	17890	3	59	23	67	2012-08-02 05:17:09.712619
17890	5	1-5166	17890	59	27	938	71	2012-08-02 05:17:09.712619
17896	4	1-2015	17895	76	69	1605	32	2012-08-02 05:17:10.09962
17897	4	1-5055	17895	69	6	4262	61	2012-08-02 05:17:10.09962
17895	4	1-5167	17895	51	76	3486	99	2012-08-02 05:17:10.09962
17898	4	1-3447	17895	6	51	2664	90	2012-08-02 05:17:10.09962
17900	5	1-3547	17899	48	18	587	18	2012-08-02 05:17:10.497344
17901	5	1-5140	17899	18	4	1666	75	2012-08-02 05:17:10.497344
17902	5	1-4872	17899	4	16	1988	83	2012-08-02 05:17:10.497344
17903	5	1-5142	17899	16	97	3213	20	2012-08-02 05:17:10.497344
17899	5	1-5168	17899	97	48	7419	40	2012-08-02 05:17:10.497344
17905	5	1-5093	17904	73	76	291	99	2012-08-02 05:17:10.89544
17906	5	1-2015	17904	76	60	123	32	2012-08-02 05:17:10.89544
17907	5	1-5102	17904	60	25	238	19	2012-08-02 05:17:10.89544
17908	5	1-3421	17904	25	14	182	30	2012-08-02 05:17:10.89544
17904	5	1-5169	17904	14	73	192	33	2012-08-02 05:17:10.89544
17910	3	1-5093	17909	73	92	626	99	2012-08-02 05:17:10.89544
17911	3	1-3667	17909	92	14	388	30	2012-08-02 05:17:10.89544
17909	3	1-5169	17909	14	73	412	33	2012-08-02 05:17:10.89544
17913	4	1-2810	17912	33	51	1121	90	2012-08-02 05:17:12.07713
17914	4	1-5167	17912	51	53	1304	99	2012-08-02 05:17:12.07713
17915	4	1-2837	17912	53	18	2053	31	2012-08-02 05:17:12.07713
17912	4	1-5172	17912	18	33	2407	56	2012-08-02 05:17:12.07713
17917	5	1-4252	17916	79	88	1879	5	2012-08-02 05:17:12.07713
17918	5	1-4930	17916	88	52	1846	79	2012-08-02 05:17:12.07713
17919	5	1-4138	17916	52	90	909	72	2012-08-02 05:17:12.07713
17920	5	1-4857	17916	90	18	2887	31	2012-08-02 05:17:12.07713
17916	5	1-5172	17916	18	79	3529	56	2012-08-02 05:17:12.07713
17922	5	1-4639	17921	35	31	984	9	2012-08-02 05:17:12.07713
17923	5	1-3145	17921	31	75	776	33	2012-08-02 05:17:12.07713
17924	5	1-1009	17921	75	90	272	72	2012-08-02 05:17:12.07713
17925	5	1-4857	17921	90	18	874	31	2012-08-02 05:17:12.07713
17921	5	1-5172	17921	18	35	1079	56	2012-08-02 05:17:12.07713
17927	5	1-5052	17926	39	91	677	22	2012-08-02 05:17:12.07713
17928	5	1-2919	17926	91	42	175	1	2012-08-02 05:17:12.07713
17929	5	1-1985	17926	42	90	74	72	2012-08-02 05:17:12.07713
17930	5	1-4857	17926	90	18	253	31	2012-08-02 05:17:12.07713
17926	5	1-5172	17926	18	39	331	56	2012-08-02 05:17:12.07713
17932	4	1-2810	17931	33	51	515	90	2012-08-02 05:17:12.07713
17933	4	1-5167	17931	51	88	684	99	2012-08-02 05:17:12.07713
17934	4	1-2628	17931	88	18	724	31	2012-08-02 05:17:12.07713
17931	4	1-5172	17931	18	33	969	56	2012-08-02 05:17:12.07713
17936	5	1-4773	17935	92	14	336	30	2012-08-02 05:17:12.07713
17937	5	1-5169	17935	14	75	411	33	2012-08-02 05:17:12.07713
17938	5	1-1009	17935	75	90	164	72	2012-08-02 05:17:12.07713
17939	5	1-4857	17935	90	18	600	31	2012-08-02 05:17:12.07713
17935	5	1-5172	17935	18	92	844	56	2012-08-02 05:17:12.07713
17941	4	1-5087	17940	32	78	3232	33	2012-08-02 05:17:12.85116
17942	4	1-1999	17940	78	3	478	83	2012-08-02 05:17:12.85116
17943	4	1-5115	17940	3	8	242	43	2012-08-02 05:17:12.85116
17940	4	1-5174	17940	8	32	2105	77	2012-08-02 05:17:12.85116
17945	4	1-4894	17944	87	37	225	60	2012-08-02 05:17:12.85116
17946	4	1-3900	17944	37	30	123	16	2012-08-02 05:17:12.85116
17947	4	1-5136	17944	30	8	39	43	2012-08-02 05:17:12.85116
17944	4	1-5174	17944	8	87	369	77	2012-08-02 05:17:12.85116
17949	4	1-1426	17948	59	71	678	10	2012-08-02 05:17:12.85116
17950	4	1-3048	17948	71	3	281	83	2012-08-02 05:17:12.85116
17951	4	1-5115	17948	3	8	174	43	2012-08-02 05:17:12.85116
17948	4	1-5174	17948	8	59	1849	77	2012-08-02 05:17:12.85116
17953	3	1-4565	17952	55	61	3880	52	2012-08-02 05:17:13.635396
17954	3	1-1652	17952	61	51	1067	71	2012-08-02 05:17:13.635396
17952	3	1-5176	17952	51	55	3454	47	2012-08-02 05:17:13.635396
17956	5	1-4743	17955	50	74	1545	91	2012-08-02 05:17:13.635396
17957	5	1-2786	17955	74	66	1217	68	2012-08-02 05:17:13.635396
17958	5	1-4484	17955	66	13	975	54	2012-08-02 05:17:13.635396
17959	5	1-4164	17955	13	51	301	71	2012-08-02 05:17:13.635396
17955	5	1-5176	17955	51	50	1185	47	2012-08-02 05:17:13.635396
17961	5	1-4743	17960	50	97	683	91	2012-08-02 05:17:13.635396
17962	5	1-4961	17960	97	49	749	98	2012-08-02 05:17:13.635396
17963	5	1-3674	17960	49	13	416	54	2012-08-02 05:17:13.635396
17964	5	1-4164	17960	13	51	130	71	2012-08-02 05:17:13.635396
17960	5	1-5176	17960	51	50	518	47	2012-08-02 05:17:13.635396
17966	5	1-4427	17965	7	7	1527	6	2012-08-02 05:17:13.635396
17967	5	1-4920	17965	7	66	1543	68	2012-08-02 05:17:13.635396
17968	5	1-4484	17965	66	13	1279	54	2012-08-02 05:17:13.635396
17969	5	1-4164	17965	13	51	408	71	2012-08-02 05:17:13.635396
17965	5	1-5176	17965	51	7	1660	47	2012-08-02 05:17:13.635396
17971	5	1-4565	17970	55	84	1001	52	2012-08-02 05:17:13.635396
17972	5	1-4372	17970	84	97	706	7	2012-08-02 05:17:13.635396
17973	5	1-4925	17970	97	3	494	23	2012-08-02 05:17:13.635396
17974	5	1-2805	17970	3	51	164	71	2012-08-02 05:17:13.635396
17970	5	1-5176	17970	51	55	687	47	2012-08-02 05:17:13.635396
17975	1	1-4961	\N	97	97	301	98	2012-08-02 05:17:13.635396
17976	1	1-3952	\N	93	93	128	98	2012-08-02 05:17:13.635396
17978	5	1-5046	17977	78	18	169	18	2012-08-02 05:17:14.806893
17979	5	1-5140	17977	18	14	338	75	2012-08-02 05:17:14.806893
17980	5	1-3211	17977	14	94	1009	2	2012-08-02 05:17:14.806893
17981	5	1-4688	17977	94	59	394	1	2012-08-02 05:17:14.806893
17977	5	1-5179	17977	59	78	108	82	2012-08-02 05:17:14.806893
17983	5	1-5046	17982	78	18	1086	18	2012-08-02 05:17:14.806893
17984	5	1-5140	17982	18	30	2539	75	2012-08-02 05:17:14.806893
17985	5	1-4581	17982	30	94	1570	97	2012-08-02 05:17:14.806893
17986	5	1-4933	17982	94	59	1852	1	2012-08-02 05:17:14.806893
17982	5	1-5179	17982	59	78	591	82	2012-08-02 05:17:14.806893
17988	5	1-5046	17987	78	18	855	18	2012-08-02 05:17:14.806893
17989	5	1-5140	17987	18	67	2037	75	2012-08-02 05:17:14.806893
17990	5	1-4622	17987	67	49	2319	65	2012-08-02 05:17:14.806893
17991	5	1-3781	17987	49	59	1404	1	2012-08-02 05:17:14.806893
17987	5	1-5179	17987	59	78	457	82	2012-08-02 05:17:14.806893
17992	1	1-1415	\N	39	39	4389	2	2012-08-02 05:17:14.806893
17994	5	1-4920	17993	7	16	5913	68	2012-08-02 05:17:16.386824
17995	5	1-2286	17993	16	60	3243	13	2012-08-02 05:17:16.386824
17996	5	1-4389	17993	60	19	1188	21	2012-08-02 05:17:16.386824
17997	5	1-5138	17993	19	38	5675	37	2012-08-02 05:17:16.386824
17993	5	1-5183	17993	38	7	8656	6	2012-08-02 05:17:16.386824
17999	4	1-3869	17998	90	48	577	100	2012-08-02 05:17:16.386824
18000	4	1-5099	17998	48	75	179	97	2012-08-02 05:17:16.386824
18001	4	1-5181	17998	75	38	685	37	2012-08-02 05:17:16.386824
17998	4	1-5183	17998	38	90	1097	6	2012-08-02 05:17:16.386824
18003	5	1-3274	18002	23	83	1399	58	2012-08-02 05:17:19.093627
18004	5	1-4884	18002	83	8	642	95	2012-08-02 05:17:19.093627
18005	5	1-4629	18002	8	38	299	9	2012-08-02 05:17:19.093627
18006	5	1-4810	18002	38	44	434	49	2012-08-02 05:17:19.093627
18002	5	1-5190	18002	44	23	1257	18	2012-08-02 05:17:19.093627
18008	5	1-3677	18007	8	81	2025	25	2012-08-02 05:17:19.093627
18009	5	1-3459	18007	81	81	1866	16	2012-08-02 05:17:19.093627
18010	5	1-4542	18007	81	84	1373	7	2012-08-02 05:17:19.093627
18011	5	1-4900	18007	84	44	714	49	2012-08-02 05:17:19.093627
18007	5	1-5190	18007	44	8	2381	18	2012-08-02 05:17:19.093627
18013	4	1-2241	18012	72	81	1215	16	2012-08-02 05:17:19.093627
18014	4	1-4542	18012	81	84	802	7	2012-08-02 05:17:19.093627
18015	4	1-4900	18012	84	44	444	49	2012-08-02 05:17:19.093627
18012	4	1-5190	18012	44	72	1575	18	2012-08-02 05:17:19.093627
18017	5	1-2775	18016	30	56	548	29	2012-08-02 05:17:20.288562
18018	5	1-4907	18016	56	78	491	82	2012-08-02 05:17:20.288562
18019	5	1-5046	18016	78	24	1030	18	2012-08-02 05:17:20.288562
18020	5	1-5047	18016	24	54	2059	44	2012-08-02 05:17:20.288562
18016	5	1-5193	18016	54	30	3354	55	2012-08-02 05:17:20.288562
18022	5	1-3770	18021	31	60	1303	32	2012-08-02 05:17:20.288562
18023	5	1-5102	18021	60	92	2605	19	2012-08-02 05:17:20.288562
18024	5	1-890	18021	92	20	1422	88	2012-08-02 05:17:20.288562
18025	5	1-5109	18021	20	54	1180	44	2012-08-02 05:17:20.288562
18021	5	1-5193	18021	54	31	2011	55	2012-08-02 05:17:20.288562
18027	5	1-4692	18026	19	6	1825	41	2012-08-02 05:17:20.675574
18028	5	1-5180	18026	6	32	1493	67	2012-08-02 05:17:20.675574
18029	5	1-1399	18026	32	50	765	32	2012-08-02 05:17:20.675574
18030	5	1-4143	18026	50	73	508	81	2012-08-02 05:17:20.675574
18026	5	1-5194	18026	73	19	2616	27	2012-08-02 05:17:20.675574
18032	4	1-2918	18031	81	92	1581	93	2012-08-02 05:17:20.675574
18033	4	1-5126	18031	92	40	876	56	2012-08-02 05:17:20.675574
18034	4	1-4065	18031	40	73	414	81	2012-08-02 05:17:20.675574
18031	4	1-5194	18031	73	81	2472	27	2012-08-02 05:17:20.675574
18036	4	1-2217	18035	72	80	376	86	2012-08-02 05:17:20.675574
18037	4	1-1812	18035	80	50	293	32	2012-08-02 05:17:20.675574
18038	4	1-4143	18035	50	73	311	81	2012-08-02 05:17:20.675574
18035	4	1-5194	18035	73	72	2562	27	2012-08-02 05:17:20.675574
18040	5	1-4220	18039	27	71	4742	16	2012-08-02 05:17:21.08454
18041	5	1-4841	18039	71	84	2677	38	2012-08-02 05:17:21.08454
18042	5	1-4091	18039	84	82	2678	72	2012-08-02 05:17:21.08454
18043	5	1-4494	18039	82	4	2571	21	2012-08-02 05:17:21.08454
18039	5	1-5195	18039	4	27	3854	19	2012-08-02 05:17:21.08454
18045	3	1-3421	18044	25	38	808	30	2012-08-02 05:17:21.08454
18046	3	1-3377	18044	38	4	615	21	2012-08-02 05:17:21.08454
18044	3	1-5195	18044	4	25	924	19	2012-08-02 05:17:21.08454
18047	1	1-4494	\N	82	82	1270	21	2012-08-02 05:17:21.08454
18049	4	1-4781	18048	95	85	94	28	2012-08-02 05:17:21.869166
18050	4	1-4942	18048	85	52	53	74	2012-08-02 05:17:21.869166
18051	4	1-5083	18048	52	89	27	59	2012-08-02 05:17:21.869166
18048	4	1-5197	18048	89	95	205	85	2012-08-02 05:17:21.869166
18053	4	1-3324	18052	55	83	1794	11	2012-08-02 05:17:21.869166
18054	4	1-5002	18052	83	49	2465	50	2012-08-02 05:17:21.869166
18055	4	1-2943	18052	49	89	332	59	2012-08-02 05:17:21.869166
18052	4	1-5197	18052	89	55	2899	85	2012-08-02 05:17:21.869166
18057	4	1-4898	18056	100	17	909	75	2012-08-02 05:17:22.266787
18058	4	1-1854	18056	17	8	350	8	2012-08-02 05:17:22.266787
18059	4	1-4529	18056	8	55	593	31	2012-08-02 05:17:22.266787
18056	4	1-5198	18056	55	100	1261	96	2012-08-02 05:17:22.266787
18061	4	1-4897	18060	7	50	470	89	2012-08-02 05:17:22.266787
18062	4	1-2193	18060	50	12	268	68	2012-08-02 05:17:22.266787
18063	4	1-1927	18060	12	55	189	31	2012-08-02 05:17:22.266787
18060	4	1-5198	18060	55	7	407	96	2012-08-02 05:17:22.266787
18065	5	1-4898	18064	100	57	2522	75	2012-08-02 05:17:22.266787
18066	5	1-3640	18064	57	41	2179	70	2012-08-02 05:17:22.266787
18067	5	1-4614	18064	41	33	1673	46	2012-08-02 05:17:22.266787
18068	5	1-4985	18064	33	55	1577	31	2012-08-02 05:17:22.266787
18064	5	1-5198	18064	55	100	3425	96	2012-08-02 05:17:22.266787
18070	5	1-4875	18069	12	27	1359	25	2012-08-02 05:17:22.266787
18071	5	1-4564	18069	27	20	834	59	2012-08-02 05:17:22.266787
18072	5	1-4341	18069	20	8	257	8	2012-08-02 05:17:22.266787
18073	5	1-4529	18069	8	55	472	31	2012-08-02 05:17:22.266787
18069	5	1-5198	18069	55	12	1087	96	2012-08-02 05:17:22.266787
18074	1	1-289	\N	94	94	265	89	2012-08-02 05:17:22.266787
18076	2	1-4342	18075	61	39	2514	85	2012-08-02 05:17:22.664804
18075	2	1-5199	18075	39	61	6076	36	2012-08-02 05:17:22.664804
18078	5	1-5085	18077	13	47	1534	4	2012-08-02 05:17:22.664804
18079	5	1-3767	18077	47	60	652	32	2012-08-02 05:17:22.664804
18080	5	1-5102	18077	60	57	1016	19	2012-08-02 05:17:22.664804
18081	5	1-4001	18077	57	39	725	85	2012-08-02 05:17:22.664804
18077	5	1-5199	18077	39	13	3654	36	2012-08-02 05:17:22.664804
18083	3	1-4480	18082	11	86	2251	86	2012-08-02 05:17:23.449374
18084	3	1-4236	18082	86	4	1114	72	2012-08-02 05:17:23.449374
18082	3	1-5201	18082	4	11	2795	35	2012-08-02 05:17:23.449374
18086	3	1-5109	18085	20	11	2321	44	2012-08-02 05:17:23.857879
18087	3	1-889	18085	11	54	589	64	2012-08-02 05:17:23.857879
18085	3	1-5202	18085	54	20	2771	88	2012-08-02 05:17:23.857879
18089	5	1-4985	18088	33	18	314	31	2012-08-02 05:17:24.23376
18090	5	1-5172	18088	18	92	375	56	2012-08-02 05:17:24.23376
18091	5	1-4773	18088	92	14	127	30	2012-08-02 05:17:24.23376
18092	5	1-5169	18088	14	80	132	33	2012-08-02 05:17:24.23376
18088	5	1-5203	18088	80	33	261	46	2012-08-02 05:17:24.23376
18094	5	1-5051	18093	22	83	6070	58	2012-08-02 05:17:24.23376
18095	5	1-4884	18093	83	8	4012	95	2012-08-02 05:17:24.23376
18096	5	1-4629	18093	8	31	2694	9	2012-08-02 05:17:24.23376
18097	5	1-3145	18093	31	80	2162	33	2012-08-02 05:17:24.23376
18093	5	1-5203	18093	80	22	4487	46	2012-08-02 05:17:24.23376
18099	5	1-5051	18098	22	42	1761	58	2012-08-02 05:17:24.23376
18100	5	1-5189	18098	42	50	1156	28	2012-08-02 05:17:24.23376
18101	5	1-5170	18098	50	6	1080	95	2012-08-02 05:17:24.23376
18102	5	1-4496	18098	6	80	558	33	2012-08-02 05:17:24.23376
18098	5	1-5203	18098	80	22	1228	46	2012-08-02 05:17:24.23376
18104	4	1-5068	18103	45	97	1173	52	2012-08-02 05:17:24.620499
18103	4	1-5204	18103	74	45	1209	100	2012-08-02 05:17:24.620499
18105	4	1-4440	18103	97	47	694	92	2012-08-02 05:17:24.620499
18106	4	1-4698	18103	47	74	499	87	2012-08-02 05:17:24.620499
18108	5	1-5009	18107	52	11	3773	60	2012-08-02 05:17:24.620499
18109	5	1-5156	18107	11	27	1972	68	2012-08-02 05:17:24.620499
18110	5	1-5070	18107	27	30	2250	5	2012-08-02 05:17:24.620499
18111	5	1-4736	18107	30	74	2165	87	2012-08-02 05:17:24.620499
18107	5	1-5204	18107	74	52	6451	100	2012-08-02 05:17:24.620499
18113	2	1-1938	18112	4	10	2738	14	2012-08-02 05:17:24.996472
18112	2	1-5205	18112	10	4	4462	96	2012-08-02 05:17:24.996472
18115	3	1-5167	18114	51	31	1168	99	2012-08-02 05:17:25.780973
18116	3	1-5098	18114	31	5	985	97	2012-08-02 05:17:25.780973
18114	3	1-5207	18114	5	51	1161	90	2012-08-02 05:17:25.780973
18118	3	1-5139	18117	79	31	1247	99	2012-08-02 05:17:25.780973
18119	3	1-5098	18117	31	5	1255	97	2012-08-02 05:17:25.780973
18117	3	1-5207	18117	5	79	1766	90	2012-08-02 05:17:25.780973
18121	5	1-1051	18120	59	64	771	12	2012-08-02 05:17:26.200539
18122	5	1-4486	18120	64	92	795	78	2012-08-02 05:17:26.200539
18123	5	1-4980	18120	92	94	2192	2	2012-08-02 05:17:26.200539
18124	5	1-4688	18120	94	29	1118	1	2012-08-02 05:17:26.200539
18120	5	1-5208	18120	29	59	1027	8	2012-08-02 05:17:26.200539
18126	5	1-4822	18125	48	17	751	24	2012-08-02 05:17:26.200539
18127	5	1-4948	18125	17	83	1091	60	2012-08-02 05:17:26.200539
18128	5	1-933	18125	83	94	976	2	2012-08-02 05:17:26.200539
18129	5	1-4688	18125	94	29	552	1	2012-08-02 05:17:26.200539
18125	5	1-5208	18125	29	48	562	8	2012-08-02 05:17:26.200539
18131	5	1-3195	18130	54	31	2344	51	2012-08-02 05:17:27.36063
18132	5	1-2952	18130	31	52	1421	78	2012-08-02 05:17:27.36063
18133	5	1-3405	18130	52	84	963	7	2012-08-02 05:17:27.36063
18134	5	1-4900	18130	84	99	657	49	2012-08-02 05:17:27.36063
18130	5	1-5211	18130	99	54	1690	44	2012-08-02 05:17:27.36063
18136	3	1-1426	18135	59	67	168	10	2012-08-02 05:17:28.145034
18137	3	1-5032	18135	67	56	145	4	2012-08-02 05:17:28.145034
18135	3	1-5213	18135	56	59	282	77	2012-08-02 05:17:28.145034
18139	3	1-3877	18138	24	71	2007	42	2012-08-02 05:17:28.907328
18140	3	1-3893	18138	71	93	509	72	2012-08-02 05:17:28.907328
18138	3	1-5215	18138	93	24	4344	39	2012-08-02 05:17:28.907328
18142	3	1-4617	18141	63	80	1021	89	2012-08-02 05:17:28.907328
18143	3	1-3390	18141	80	93	335	72	2012-08-02 05:17:28.907328
18141	3	1-5215	18141	93	63	3210	39	2012-08-02 05:17:28.907328
18145	4	1-2665	18144	50	75	375	47	2012-08-02 05:17:28.907328
18146	4	1-4197	18144	75	13	226	82	2012-08-02 05:17:28.907328
18147	4	1-5135	18144	13	93	143	72	2012-08-02 05:17:28.907328
18144	4	1-5215	18144	93	50	1682	39	2012-08-02 05:17:28.907328
18149	5	1-2665	18148	50	7	195	47	2012-08-02 05:17:28.907328
18150	5	1-4427	18148	7	20	174	6	2012-08-02 05:17:28.907328
18151	5	1-3393	18148	20	86	67	86	2012-08-02 05:17:28.907328
18152	5	1-4236	18148	86	93	56	72	2012-08-02 05:17:28.907328
18148	5	1-5215	18148	93	50	761	39	2012-08-02 05:17:28.907328
18153	1	1-3893	\N	71	71	1763	72	2012-08-02 05:17:28.907328
18155	4	1-5214	18154	2	59	1940	8	2012-08-02 05:17:30.089492
18156	4	1-1051	18154	59	49	1019	12	2012-08-02 05:17:30.089492
18157	4	1-5106	18154	49	11	2585	27	2012-08-02 05:17:30.089492
18154	4	1-5218	18154	11	2	4457	45	2012-08-02 05:17:30.089492
18159	3	1-5156	18158	11	87	576	68	2012-08-02 05:17:31.647284
18160	3	1-4360	18158	87	88	182	94	2012-08-02 05:17:31.647284
18158	3	1-5222	18158	88	11	2086	60	2012-08-02 05:17:31.647284
18162	5	1-3900	18161	37	71	1321	16	2012-08-02 05:17:31.647284
18163	5	1-4841	18161	71	78	655	38	2012-08-02 05:17:31.647284
18164	5	1-3825	18161	78	68	196	26	2012-08-02 05:17:31.647284
18165	5	1-4009	18161	68	88	68	94	2012-08-02 05:17:31.647284
18161	5	1-5222	18161	88	37	1410	60	2012-08-02 05:17:31.647284
18167	5	1-3900	18166	37	81	146	16	2012-08-02 05:17:31.647284
18168	5	1-4542	18166	81	66	116	7	2012-08-02 05:17:31.647284
18169	5	1-4830	18166	66	29	32	64	2012-08-02 05:17:31.647284
18170	5	1-578	18166	29	88	7	94	2012-08-02 05:17:31.647284
18166	5	1-5222	18166	88	37	148	60	2012-08-02 05:17:31.647284
18172	5	1-5156	18171	11	27	2051	68	2012-08-02 05:17:31.647284
18173	5	1-5070	18171	27	4	2281	5	2012-08-02 05:17:31.647284
18174	5	1-4231	18171	4	67	843	69	2012-08-02 05:17:31.647284
18175	5	1-2285	18171	67	88	190	94	2012-08-02 05:17:31.647284
18171	5	1-5222	18171	88	11	4025	60	2012-08-02 05:17:31.647284
18177	5	1-4735	18176	62	75	176	97	2012-08-02 05:17:31.647284
18178	5	1-5181	18176	75	40	949	37	2012-08-02 05:17:31.647284
18179	5	1-3288	18176	40	67	70	69	2012-08-02 05:17:31.647284
18180	5	1-2285	18176	67	88	16	94	2012-08-02 05:17:31.647284
18176	5	1-5222	18176	88	62	347	60	2012-08-02 05:17:31.647284
18182	5	1-4580	18181	76	81	371	16	2012-08-02 05:17:31.647284
18183	5	1-4542	18181	81	66	310	7	2012-08-02 05:17:31.647284
18184	5	1-4830	18181	66	29	91	64	2012-08-02 05:17:31.647284
18185	5	1-578	18181	29	88	20	94	2012-08-02 05:17:31.647284
18181	5	1-5222	18181	88	76	462	60	2012-08-02 05:17:31.647284
18187	3	1-4330	18186	65	54	1344	57	2012-08-02 05:17:32.045506
18188	3	1-4945	18186	54	72	234	26	2012-08-02 05:17:32.045506
18186	3	1-5223	18186	72	65	543	5	2012-08-02 05:17:32.045506
18190	5	1-4790	18189	27	90	4662	65	2012-08-02 05:17:34.796151
18191	5	1-4281	18189	90	63	3892	84	2012-08-02 05:17:34.796151
18192	5	1-4696	18189	63	54	2180	64	2012-08-02 05:17:34.796151
18193	5	1-5202	18189	54	87	6471	88	2012-08-02 05:17:34.796151
18189	5	1-5230	18189	87	27	6592	71	2012-08-02 05:17:34.796151
18195	4	1-4790	18194	27	91	541	65	2012-08-02 05:17:34.796151
18196	4	1-4433	18194	91	10	326	49	2012-08-02 05:17:34.796151
18197	4	1-3193	18194	10	87	387	88	2012-08-02 05:17:34.796151
18194	4	1-5230	18194	87	27	549	71	2012-08-02 05:17:34.796151
18199	5	1-4817	18198	88	72	1602	27	2012-08-02 05:17:34.796151
18200	5	1-2217	18198	72	64	230	86	2012-08-02 05:17:34.796151
18201	5	1-4903	18198	64	10	580	49	2012-08-02 05:17:34.796151
18202	5	1-3193	18198	10	87	802	88	2012-08-02 05:17:34.796151
18198	5	1-5230	18198	87	88	1324	71	2012-08-02 05:17:34.796151
18204	5	1-4450	18203	59	52	657	78	2012-08-02 05:17:34.796151
18205	5	1-3405	18203	52	66	467	7	2012-08-02 05:17:34.796151
18206	5	1-4830	18203	66	54	141	64	2012-08-02 05:17:34.796151
18207	5	1-5202	18203	54	87	723	88	2012-08-02 05:17:34.796151
18203	5	1-5230	18203	87	59	1277	71	2012-08-02 05:17:34.796151
18209	5	1-4809	18208	78	5	754	97	2012-08-02 05:17:37.215775
18210	5	1-5207	18208	5	57	1229	90	2012-08-02 05:17:37.215775
18211	5	1-5217	18208	57	59	642	77	2012-08-02 05:17:37.215775
18212	5	1-1426	18208	59	12	413	10	2012-08-02 05:17:37.215775
18208	5	1-5236	18208	12	78	658	89	2012-08-02 05:17:37.215775
18214	5	1-5213	18213	56	56	1171	77	2012-08-02 05:17:38.773627
18215	5	1-5237	18213	56	35	1852	84	2012-08-02 05:17:38.773627
18216	5	1-5025	18213	35	94	1004	2	2012-08-02 05:17:38.773627
18217	5	1-4688	18213	94	52	583	1	2012-08-02 05:17:38.773627
18213	5	1-5240	18213	52	56	568	4	2012-08-02 05:17:38.773627
18219	4	1-2459	18218	50	21	2011	10	2012-08-02 05:17:43.468646
18220	4	1-5038	18218	21	6	1509	41	2012-08-02 05:17:43.468646
18221	4	1-5180	18218	6	71	1471	67	2012-08-02 05:17:43.468646
18218	4	1-5252	18218	71	50	4574	82	2012-08-02 05:17:43.468646
18223	4	1-3656	18222	24	21	479	10	2012-08-02 05:17:43.468646
18224	4	1-5038	18222	21	6	376	41	2012-08-02 05:17:43.468646
18225	4	1-5180	18222	6	71	384	67	2012-08-02 05:17:43.468646
18222	4	1-5252	18222	71	24	1252	82	2012-08-02 05:17:43.468646
18227	5	1-5128	18226	35	75	3640	62	2012-08-02 05:17:44.25317
18228	5	1-5221	18226	75	80	5568	89	2012-08-02 05:17:44.25317
18229	5	1-3390	18226	80	82	2941	72	2012-08-02 05:17:44.25317
18230	5	1-4641	18226	82	97	2015	1	2012-08-02 05:17:44.25317
18226	5	1-5254	18226	97	35	1632	38	2012-08-02 05:17:44.25317
18232	4	1-4986	18231	80	68	1372	43	2012-08-02 05:17:44.650985
18233	4	1-4632	18231	68	34	2640	57	2012-08-02 05:17:44.650985
18234	4	1-3813	18231	34	41	1406	11	2012-08-02 05:17:44.650985
18231	4	1-5255	18231	41	80	1457	71	2012-08-02 05:17:44.650985
18236	4	1-4986	18235	80	30	1162	43	2012-08-02 05:17:44.650985
18237	4	1-4944	18235	30	82	1025	62	2012-08-02 05:17:44.650985
18238	4	1-4549	18235	82	41	1086	11	2012-08-02 05:17:44.650985
18235	4	1-5255	18235	41	80	1179	71	2012-08-02 05:17:44.650985
18240	4	1-3599	18239	67	67	1123	21	2012-08-02 05:17:48.585702
18241	4	1-5117	18239	67	47	2842	34	2012-08-02 05:17:48.585702
18242	4	1-4657	18239	47	50	2422	31	2012-08-02 05:17:48.585702
18239	4	1-5265	18239	50	67	2013	92	2012-08-02 05:17:48.585702
18244	4	1-4170	18243	43	67	444	21	2012-08-02 05:17:48.585702
18245	4	1-5117	18243	67	47	1103	34	2012-08-02 05:17:48.585702
18246	4	1-4657	18243	47	50	1041	31	2012-08-02 05:17:48.585702
18243	4	1-5265	18243	50	43	959	92	2012-08-02 05:17:48.585702
18248	5	1-5220	18247	22	12	3301	14	2012-08-02 05:17:50.950278
18249	5	1-2655	18247	12	30	605	43	2012-08-02 05:17:50.950278
18250	5	1-4944	18247	30	57	552	62	2012-08-02 05:17:50.950278
18251	5	1-4908	18247	57	85	557	36	2012-08-02 05:17:50.950278
18247	5	1-5271	18247	85	22	1399	68	2012-08-02 05:17:50.950278
18253	3	1-3906	18252	94	56	2617	20	2012-08-02 05:17:50.950278
18254	3	1-4886	18252	56	85	1086	36	2012-08-02 05:17:50.950278
18252	3	1-5271	18252	85	94	2983	68	2012-08-02 05:17:50.950278
18256	5	1-5251	18255	40	20	475	4	2012-08-02 05:17:50.950278
18257	5	1-5100	18255	20	3	247	75	2012-08-02 05:17:50.950278
18258	5	1-4333	18255	3	57	431	62	2012-08-02 05:17:50.950278
18259	5	1-4908	18255	57	85	479	36	2012-08-02 05:17:50.950278
18255	5	1-5271	18255	85	40	1324	68	2012-08-02 05:17:50.950278
18261	5	1-5247	18260	55	56	1277	81	2012-08-02 05:17:54.032476
18262	5	1-5110	18260	56	10	1200	65	2012-08-02 05:17:54.032476
18263	5	1-4837	18260	10	80	1411	78	2012-08-02 05:17:54.032476
18264	5	1-3603	18260	80	63	607	48	2012-08-02 05:17:54.032476
18260	5	1-5279	18260	63	55	1034	89	2012-08-02 05:17:54.032476
18266	3	1-5113	18265	45	77	8403	70	2012-08-02 05:17:54.806106
18267	3	1-4382	18265	77	97	1939	99	2012-08-02 05:17:54.806106
18265	3	1-5281	18265	97	45	8718	79	2012-08-02 05:17:54.806106
18269	5	1-2946	18268	53	35	23	7	2012-08-02 05:17:54.806106
18270	5	1-5091	18268	35	5	14	97	2012-08-02 05:17:54.806106
18271	5	1-5207	18268	5	79	19	90	2012-08-02 05:17:54.806106
18272	5	1-5139	18268	79	97	13	99	2012-08-02 05:17:54.806106
18268	5	1-5281	18268	97	53	67	79	2012-08-02 05:17:54.806106
18274	2	1-4997	18273	1	19	5103	83	2012-08-02 05:17:55.203696
18273	2	1-5282	18273	19	1	8030	74	2012-08-02 05:17:55.203696
18276	3	1-4791	18275	100	91	280	49	2012-08-02 05:17:55.203696
18277	3	1-5219	18275	91	19	334	83	2012-08-02 05:17:55.203696
18275	3	1-5282	18275	19	100	642	74	2012-08-02 05:17:55.203696
18279	5	1-5009	18278	52	88	139	60	2012-08-02 05:17:56.772515
18280	5	1-5248	18278	88	37	119	76	2012-08-02 05:17:56.772515
18281	5	1-4977	18278	37	8	53	95	2012-08-02 05:17:56.772515
18282	5	1-4629	18278	8	53	40	9	2012-08-02 05:17:56.772515
18278	5	1-5286	18278	53	52	216	100	2012-08-02 05:17:56.772515
18283	1	1-4629	\N	8	8	398	9	2012-08-02 05:17:56.772515
18285	5	1-5083	18284	52	93	2400	59	2012-08-02 05:17:57.170312
18286	5	1-5241	18284	93	14	1688	30	2012-08-02 05:17:57.170312
18287	5	1-5169	18284	14	83	1653	33	2012-08-02 05:17:57.170312
18288	5	1-4583	18284	83	27	1350	63	2012-08-02 05:17:57.170312
18284	5	1-5287	18284	27	52	4145	74	2012-08-02 05:17:57.170312
18290	4	1-4944	18289	30	82	2272	62	2012-08-02 05:17:57.568102
18291	4	1-4474	18289	82	97	1060	20	2012-08-02 05:17:57.568102
18292	4	1-5168	18289	97	87	1994	40	2012-08-02 05:17:57.568102
18289	4	1-5288	18289	87	30	2885	43	2012-08-02 05:17:57.568102
18294	5	1-4944	18293	30	75	760	62	2012-08-02 05:17:57.568102
18295	5	1-5221	18293	75	24	890	89	2012-08-02 05:17:57.568102
18296	5	1-5043	18293	24	11	298	67	2012-08-02 05:17:57.568102
18297	5	1-4366	18293	11	87	618	40	2012-08-02 05:17:57.568102
18293	5	1-5288	18293	87	30	929	43	2012-08-02 05:17:57.568102
18299	5	1-4944	18298	30	57	1079	62	2012-08-02 05:17:57.568102
18300	5	1-4908	18298	57	44	1116	36	2012-08-02 05:17:57.568102
18301	5	1-3616	18298	44	6	447	1	2012-08-02 05:17:57.568102
18302	5	1-3945	18298	6	87	671	40	2012-08-02 05:17:57.568102
18298	5	1-5288	18298	87	30	1153	43	2012-08-02 05:17:57.568102
18304	4	1-5208	18303	29	24	180	8	2012-08-02 05:17:57.965978
18305	4	1-2862	18303	24	55	145	68	2012-08-02 05:17:57.965978
18306	4	1-3345	18303	55	81	94	93	2012-08-02 05:17:57.965978
18303	4	1-5289	18303	81	29	183	1	2012-08-02 05:17:57.965978
18308	5	1-5240	18307	52	26	2759	4	2012-08-02 05:17:57.965978
18309	5	1-3339	18307	26	27	1802	86	2012-08-02 05:17:57.965978
18310	5	1-4928	18307	27	83	1628	9	2012-08-02 05:17:57.965978
18311	5	1-3192	18307	83	81	1457	93	2012-08-02 05:17:57.965978
18307	5	1-5289	18307	81	52	2926	1	2012-08-02 05:17:57.965978
18313	5	1-5254	18312	97	78	2919	38	2012-08-02 05:17:57.965978
18314	5	1-3825	18312	78	72	944	26	2012-08-02 05:17:57.965978
18315	5	1-5223	18312	72	24	2158	5	2012-08-02 05:17:57.965978
18316	5	1-4618	18312	24	81	1894	93	2012-08-02 05:17:57.965978
18312	5	1-5289	18312	81	97	3865	1	2012-08-02 05:17:57.965978
18318	3	1-5251	18317	40	56	1852	4	2012-08-02 05:17:58.761532
18319	3	1-5213	18317	56	16	3530	77	2012-08-02 05:17:58.761532
18317	3	1-5291	18317	16	40	5378	68	2012-08-02 05:17:58.761532
18321	5	1-587	18320	63	44	544	48	2012-08-02 05:18:00.330551
18322	5	1-5192	18320	44	31	1229	22	2012-08-02 05:18:00.330551
18323	5	1-1851	18320	31	81	868	16	2012-08-02 05:18:00.330551
18324	5	1-4542	18320	81	28	683	7	2012-08-02 05:18:00.330551
18320	5	1-5295	18320	28	63	880	91	2012-08-02 05:18:00.330551
18326	5	1-2498	18325	44	63	928	48	2012-08-02 05:18:01.11501
18327	5	1-5279	18325	63	78	1462	89	2012-08-02 05:18:01.11501
18328	5	1-4809	18325	78	5	1531	97	2012-08-02 05:18:01.11501
18329	5	1-5207	18325	5	34	2284	90	2012-08-02 05:18:01.11501
18325	5	1-5297	18325	34	44	6095	13	2012-08-02 05:18:01.11501
18331	6	1-4930	18330	88	62	2492	79	2012-08-02 05:18:02.263588
18332	6	1-3032	18330	62	94	1604	2	2012-08-02 05:18:02.263588
18333	6	1-4688	18330	94	52	944	1	2012-08-02 05:18:02.263588
18334	6	1-5240	18330	52	20	934	4	2012-08-02 05:18:02.263588
18335	6	1-5100	18330	20	15	498	75	2012-08-02 05:18:02.263588
18330	6	1-5300	18330	15	88	2286	5	2012-08-02 05:18:02.263588
18337	6	1-4736	18336	30	6	2563	87	2012-08-02 05:18:02.263588
18338	6	1-2663	18336	6	94	1511	2	2012-08-02 05:18:02.263588
18339	6	1-4688	18336	94	52	917	1	2012-08-02 05:18:02.263588
18340	6	1-5240	18336	52	20	935	4	2012-08-02 05:18:02.263588
18341	6	1-5100	18336	20	15	514	75	2012-08-02 05:18:02.263588
18336	6	1-5300	18336	15	30	2432	5	2012-08-02 05:18:02.263588
18343	5	1-5208	18342	29	97	2021	8	2012-08-02 05:18:02.661599
18344	5	1-3557	18342	97	4	2226	7	2012-08-02 05:18:02.661599
18345	5	1-4912	18342	4	34	3229	53	2012-08-02 05:18:02.661599
18346	5	1-5243	18342	34	28	1796	60	2012-08-02 05:18:02.661599
18342	5	1-5301	18342	28	29	2174	1	2012-08-02 05:18:02.661599
18348	5	1-4420	18347	68	83	765	11	2012-08-02 05:18:03.435294
18349	5	1-5002	18347	83	22	1321	50	2012-08-02 05:18:03.435294
18350	5	1-4786	18347	22	28	860	60	2012-08-02 05:18:03.435294
18351	5	1-5301	18347	28	28	1234	1	2012-08-02 05:18:03.435294
18347	5	1-5303	18347	28	68	1484	61	2012-08-02 05:18:03.435294
18353	3	1-5114	18352	10	10	1457	55	2012-08-02 05:18:05.413565
18354	3	1-2794	18352	10	72	1251	17	2012-08-02 05:18:05.413565
18352	3	1-5308	18352	72	10	6864	80	2012-08-02 05:18:05.413565
18356	3	1-3770	18355	31	69	1496	32	2012-08-02 05:18:06.187246
18357	3	1-5055	18355	69	58	1503	61	2012-08-02 05:18:06.187246
18355	3	1-5310	18355	58	31	5804	55	2012-08-02 05:18:06.187246
18359	2	1-4767	18358	82	58	557	61	2012-08-02 05:18:06.187246
18358	2	1-5310	18358	58	82	2742	55	2012-08-02 05:18:06.187246
18361	4	1-4286	18360	35	60	860	85	2012-08-02 05:18:06.58488
18362	4	1-2684	18360	60	97	509	7	2012-08-02 05:18:06.58488
18363	4	1-4925	18360	97	5	383	23	2012-08-02 05:18:06.58488
18360	4	1-5311	18360	5	35	750	25	2012-08-02 05:18:06.58488
18365	5	1-4792	18364	6	38	1981	39	2012-08-02 05:18:06.971459
18366	5	1-4354	18364	38	33	2212	46	2012-08-02 05:18:06.971459
18367	5	1-4985	18364	33	50	3018	31	2012-08-02 05:18:06.971459
18368	5	1-5265	18364	50	28	3060	92	2012-08-02 05:18:06.971459
18364	5	1-5312	18364	28	6	4122	42	2012-08-02 05:18:06.971459
18370	3	1-4185	18369	64	46	472	28	2012-08-02 05:18:06.971459
18371	3	1-5031	18369	46	28	681	92	2012-08-02 05:18:06.971459
18369	3	1-5312	18369	28	64	953	42	2012-08-02 05:18:06.971459
18373	4	1-4444	18372	71	100	544	74	2012-08-02 05:18:07.756143
18374	4	1-4791	18372	100	99	179	49	2012-08-02 05:18:07.756143
18375	4	1-5211	18372	99	64	345	44	2012-08-02 05:18:07.756143
18372	4	1-5314	18372	64	71	785	98	2012-08-02 05:18:07.756143
18377	4	1-2783	18376	58	22	20	96	2012-08-02 05:18:07.756143
18378	4	1-4673	18376	22	99	14	49	2012-08-02 05:18:07.756143
18379	4	1-5211	18376	99	64	30	44	2012-08-02 05:18:07.756143
18376	4	1-5314	18376	64	58	74	98	2012-08-02 05:18:07.756143
18381	5	1-2040	18380	56	27	456	86	2012-08-02 05:18:07.756143
18382	5	1-4928	18380	27	38	361	9	2012-08-02 05:18:07.756143
18383	5	1-4810	18380	38	99	684	49	2012-08-02 05:18:07.756143
18384	5	1-5211	18380	99	64	1512	44	2012-08-02 05:18:07.756143
18380	5	1-5314	18380	64	56	3944	98	2012-08-02 05:18:07.756143
18386	4	1-5231	18385	19	23	2270	72	2012-08-02 05:18:07.756143
18387	4	1-5039	18385	23	80	1718	68	2012-08-02 05:18:07.756143
18388	4	1-4680	18385	80	64	1185	44	2012-08-02 05:18:07.756143
18385	4	1-5314	18385	64	19	3203	98	2012-08-02 05:18:07.756143
18390	5	1-5236	18389	12	78	2994	89	2012-08-02 05:18:08.16499
18391	5	1-4809	18389	78	5	2029	97	2012-08-02 05:18:08.16499
18392	5	1-5207	18389	5	79	1957	90	2012-08-02 05:18:08.16499
18393	5	1-4843	18389	79	58	1514	20	2012-08-02 05:18:08.16499
18389	5	1-5315	18389	58	12	3177	10	2012-08-02 05:18:08.16499
18395	3	1-3048	18394	71	16	232	83	2012-08-02 05:18:08.16499
18396	3	1-5142	18394	16	58	248	20	2012-08-02 05:18:08.16499
18394	3	1-5315	18394	58	71	530	10	2012-08-02 05:18:08.16499
18398	5	1-5236	18397	12	23	1878	89	2012-08-02 05:18:08.16499
18399	5	1-3137	18397	23	100	747	42	2012-08-02 05:18:08.16499
18400	5	1-4678	18397	100	16	242	83	2012-08-02 05:18:08.16499
18401	5	1-5142	18397	16	58	395	20	2012-08-02 05:18:08.16499
18397	5	1-5315	18397	58	12	1285	10	2012-08-02 05:18:08.16499
18403	5	1-5073	18402	66	38	373	39	2012-08-02 05:18:08.93837
18404	5	1-4354	18402	38	33	411	46	2012-08-02 05:18:08.93837
18405	5	1-4985	18402	33	50	553	31	2012-08-02 05:18:08.93837
18406	5	1-5265	18402	50	76	552	92	2012-08-02 05:18:08.93837
18402	5	1-5317	18402	76	66	525	77	2012-08-02 05:18:08.93837
18408	3	1-4586	18407	22	30	270	88	2012-08-02 05:18:10.507181
18409	3	1-5209	18407	30	52	158	53	2012-08-02 05:18:10.507181
18407	3	1-5321	18407	52	22	242	90	2012-08-02 05:18:10.507181
18411	5	1-5142	18410	16	33	4634	20	2012-08-02 05:18:10.905002
18412	5	1-5177	18410	33	4	3027	56	2012-08-02 05:18:10.905002
18413	5	1-3014	18410	4	61	1842	54	2012-08-02 05:18:10.905002
18414	5	1-4646	18410	61	53	1861	78	2012-08-02 05:18:10.905002
18410	5	1-5322	18410	53	16	3288	83	2012-08-02 05:18:10.905002
18415	4	1-5323	18415	12	21	2194	10	2012-08-02 05:18:11.291708
18416	4	1-5038	18415	21	6	1710	41	2012-08-02 05:18:11.291708
18417	4	1-5180	18415	6	82	1732	67	2012-08-02 05:18:11.291708
18418	4	1-4028	18415	82	12	1271	36	2012-08-02 05:18:11.291708
18420	5	1-4295	18419	62	98	1339	1	2012-08-02 05:18:12.463057
18421	5	1-4998	18419	98	82	1349	23	2012-08-02 05:18:12.463057
18422	5	1-2629	18419	82	34	951	53	2012-08-02 05:18:12.463057
18423	5	1-5243	18419	34	42	621	60	2012-08-02 05:18:12.463057
18419	5	1-5326	18419	42	62	1027	98	2012-08-02 05:18:12.463057
18425	5	1-5301	18424	28	29	293	1	2012-08-02 05:18:12.849471
18426	5	1-5208	18424	29	57	278	8	2012-08-02 05:18:12.849471
18427	5	1-2876	18424	57	68	170	70	2012-08-02 05:18:12.849471
18428	5	1-3001	18424	68	80	225	88	2012-08-02 05:18:12.849471
18424	5	1-5327	18424	80	28	237	60	2012-08-02 05:18:12.849471
18430	5	1-5301	18429	28	98	476	1	2012-08-02 05:18:12.849471
18431	5	1-4998	18429	98	22	420	23	2012-08-02 05:18:12.849471
18432	5	1-4211	18429	22	10	178	9	2012-08-02 05:18:12.849471
18433	5	1-5274	18429	10	80	361	88	2012-08-02 05:18:12.849471
18429	5	1-5327	18429	80	28	383	60	2012-08-02 05:18:12.849471
18435	4	1-3557	18434	97	28	3673	7	2012-08-02 05:18:13.238202
18436	4	1-5295	18434	28	63	3692	91	2012-08-02 05:18:13.238202
18437	4	1-587	18434	63	37	1779	48	2012-08-02 05:18:13.238202
18434	4	1-5328	18434	37	97	3989	8	2012-08-02 05:18:13.238202
18439	3	1-1051	18438	59	21	1779	12	2012-08-02 05:18:13.238202
18440	3	1-4739	18438	21	37	1151	48	2012-08-02 05:18:13.238202
18438	3	1-5328	18438	37	59	2688	8	2012-08-02 05:18:13.238202
18442	5	1-2862	18441	24	66	146	68	2012-08-02 05:18:13.238202
18443	5	1-4484	18441	66	61	135	54	2012-08-02 05:18:13.238202
18444	5	1-4646	18441	61	80	157	78	2012-08-02 05:18:13.238202
18445	5	1-3603	18441	80	37	62	48	2012-08-02 05:18:13.238202
18441	5	1-5328	18441	37	24	178	8	2012-08-02 05:18:13.238202
18447	5	1-3557	18446	97	73	1308	7	2012-08-02 05:18:13.238202
18448	5	1-5131	18446	73	11	2044	58	2012-08-02 05:18:13.238202
18449	5	1-3790	18446	11	73	287	18	2012-08-02 05:18:13.238202
18450	5	1-4758	18446	73	37	358	48	2012-08-02 05:18:13.238202
18446	5	1-5328	18446	37	97	1068	8	2012-08-02 05:18:13.238202
18452	5	1-5328	18451	37	5	165	8	2012-08-02 05:18:14.023034
18453	5	1-1506	18451	5	2	47	33	2012-08-02 05:18:14.023034
18454	5	1-4392	18451	2	5	22	23	2012-08-02 05:18:14.023034
18455	5	1-5311	18451	5	57	42	25	2012-08-02 05:18:14.023034
18451	5	1-5330	18451	57	37	70	48	2012-08-02 05:18:14.023034
18457	5	1-5192	18456	44	91	6479	22	2012-08-02 05:18:14.023034
18458	5	1-2919	18456	91	98	1406	1	2012-08-02 05:18:14.023034
18459	5	1-4998	18456	98	5	1116	23	2012-08-02 05:18:14.023034
18460	5	1-5311	18456	5	57	1897	25	2012-08-02 05:18:14.023034
18456	5	1-5330	18456	57	44	3326	48	2012-08-02 05:18:14.023034
18462	4	1-4708	18461	62	49	1016	30	2012-08-02 05:18:14.431958
18463	4	1-610	18461	49	56	1582	14	2012-08-02 05:18:14.431958
18464	4	1-5280	18461	56	19	1059	85	2012-08-02 05:18:14.431958
18461	4	1-5331	18461	19	62	1845	54	2012-08-02 05:18:14.431958
18466	5	1-5213	18465	56	66	3100	77	2012-08-02 05:18:14.829328
18467	5	1-5073	18465	66	50	1617	39	2012-08-02 05:18:14.829328
18468	5	1-2665	18465	50	7	330	47	2012-08-02 05:18:14.829328
18469	5	1-4427	18465	7	86	234	6	2012-08-02 05:18:14.829328
18465	5	1-5332	18465	86	56	2112	4	2012-08-02 05:18:14.829328
18471	5	1-3767	18470	47	69	1085	32	2012-08-02 05:18:14.829328
18472	5	1-5055	18470	69	96	2292	61	2012-08-02 05:18:14.829328
18473	5	1-1920	18470	96	7	323	47	2012-08-02 05:18:14.829328
18474	5	1-4427	18470	7	86	245	6	2012-08-02 05:18:14.829328
18470	5	1-5332	18470	86	47	2377	4	2012-08-02 05:18:14.829328
18476	4	1-5100	18475	20	26	1289	75	2012-08-02 05:18:14.829328
18477	4	1-3618	18475	26	98	2037	25	2012-08-02 05:18:14.829328
18478	4	1-1867	18475	98	86	242	6	2012-08-02 05:18:14.829328
18475	4	1-5332	18475	86	20	2706	4	2012-08-02 05:18:14.829328
18480	3	1-4143	18479	50	7	1845	81	2012-08-02 05:18:16.81839
18481	3	1-4394	18479	7	23	1732	11	2012-08-02 05:18:16.81839
18479	3	1-5337	18479	23	50	2540	32	2012-08-02 05:18:16.81839
18483	4	1-5102	18482	60	25	689	19	2012-08-02 05:18:16.81839
18484	4	1-3421	18482	25	55	421	30	2012-08-02 05:18:16.81839
18485	4	1-3723	18482	55	23	284	11	2012-08-02 05:18:16.81839
18482	4	1-5337	18482	23	60	446	32	2012-08-02 05:18:16.81839
18487	4	1-4143	18486	50	46	1152	81	2012-08-02 05:18:16.81839
18488	4	1-5336	18486	46	55	726	30	2012-08-02 05:18:16.81839
18489	4	1-3723	18486	55	23	609	11	2012-08-02 05:18:16.81839
18486	4	1-5337	18486	23	50	1190	32	2012-08-02 05:18:16.81839
18491	5	1-4936	18490	53	57	3052	55	2012-08-02 05:18:17.227055
18492	5	1-3317	18490	57	83	1560	33	2012-08-02 05:18:17.227055
18493	5	1-4583	18490	83	3	1434	63	2012-08-02 05:18:17.227055
18494	5	1-5035	18490	3	94	935	96	2012-08-02 05:18:17.227055
18490	5	1-5338	18490	94	53	1663	12	2012-08-02 05:18:17.227055
18496	4	1-4786	18495	22	28	3120	60	2012-08-02 05:18:18.022422
18497	4	1-5301	18495	28	28	4479	1	2012-08-02 05:18:18.022422
18498	4	1-5303	18495	28	76	5386	61	2012-08-02 05:18:18.022422
18495	4	1-5340	18495	76	22	4791	50	2012-08-02 05:18:18.022422
18500	5	1-5315	18499	58	96	4804	10	2012-08-02 05:18:18.42019
18501	5	1-5294	18499	96	19	4305	76	2012-08-02 05:18:18.42019
18502	5	1-4074	18499	19	53	492	9	2012-08-02 05:18:18.42019
18503	5	1-5286	18499	53	100	1747	100	2012-08-02 05:18:18.42019
18499	5	1-5341	18499	100	58	2086	20	2012-08-02 05:18:18.42019
18505	4	1-5177	18504	33	41	258	56	2012-08-02 05:18:18.42019
18506	4	1-5212	18504	41	3	286	51	2012-08-02 05:18:18.42019
18507	4	1-5277	18504	3	100	287	100	2012-08-02 05:18:18.42019
18504	4	1-5341	18504	100	33	406	20	2012-08-02 05:18:18.42019
18509	5	1-5182	18508	39	25	2547	52	2012-08-02 05:18:19.580327
18510	5	1-4994	18508	25	19	969	85	2012-08-02 05:18:19.580327
18511	5	1-5331	18508	19	61	1363	54	2012-08-02 05:18:19.580327
18512	5	1-4646	18508	61	83	1405	78	2012-08-02 05:18:19.580327
18508	5	1-5344	18508	83	39	1175	29	2012-08-02 05:18:19.580327
18514	5	1-4529	18513	8	17	2771	31	2012-08-02 05:18:19.967122
18515	5	1-5342	18513	17	69	2985	36	2012-08-02 05:18:19.967122
18516	5	1-4473	18513	69	66	1326	70	2012-08-02 05:18:19.967122
18517	5	1-5242	18513	66	79	494	98	2012-08-02 05:18:19.967122
18513	5	1-5345	18513	79	8	1159	8	2012-08-02 05:18:19.967122
18519	4	1-4882	18518	59	2	607	62	2012-08-02 05:18:20.762816
18520	4	1-2172	18518	2	64	629	42	2012-08-02 05:18:20.762816
18518	4	1-5347	18518	19	59	574	95	2012-08-02 05:18:20.762816
18521	4	1-4185	18518	64	19	250	28	2012-08-02 05:18:20.762816
18523	4	1-4496	18522	6	24	1795	33	2012-08-02 05:18:20.762816
18524	4	1-5307	18522	24	99	1378	38	2012-08-02 05:18:20.762816
18525	4	1-5080	18522	99	19	1368	28	2012-08-02 05:18:20.762816
18522	4	1-5347	18522	19	6	3586	95	2012-08-02 05:18:20.762816
18527	4	1-5124	18526	78	7	3197	50	2012-08-02 05:18:20.762816
18528	4	1-5147	18526	7	64	996	42	2012-08-02 05:18:20.762816
18529	4	1-4185	18526	64	19	466	28	2012-08-02 05:18:20.762816
18526	4	1-5347	18526	19	78	1260	95	2012-08-02 05:18:20.762816
18531	5	1-5100	18530	20	69	3	75	2012-08-02 05:18:23.502386
18532	5	1-5081	18530	69	43	8	92	2012-08-02 05:18:23.502386
18533	5	1-4170	18530	43	67	4	21	2012-08-02 05:18:23.502386
18534	5	1-5117	18530	67	10	9	34	2012-08-02 05:18:23.502386
18530	5	1-5354	18530	10	20	6	4	2012-08-02 05:18:23.502386
18536	5	1-5100	18535	20	15	481	75	2012-08-02 05:18:23.502386
18537	5	1-5300	18535	15	13	1914	5	2012-08-02 05:18:23.502386
18538	5	1-4085	18535	13	22	1366	51	2012-08-02 05:18:23.502386
18539	5	1-5132	18535	22	10	1549	34	2012-08-02 05:18:23.502386
18535	5	1-5354	18535	10	20	1043	4	2012-08-02 05:18:23.502386
18541	5	1-5100	18540	20	85	186	75	2012-08-02 05:18:23.502386
18542	5	1-4034	18540	85	92	337	17	2012-08-02 05:18:23.502386
18543	5	1-5270	18540	92	67	198	21	2012-08-02 05:18:23.502386
18544	5	1-5117	18540	67	10	510	34	2012-08-02 05:18:23.502386
18540	5	1-5354	18540	10	20	372	4	2012-08-02 05:18:23.502386
18546	4	1-4758	18545	73	81	5870	48	2012-08-02 05:18:24.287011
18547	4	1-4585	18545	81	68	5173	87	2012-08-02 05:18:24.287011
18548	4	1-5238	18545	68	99	5257	76	2012-08-02 05:18:24.287011
18545	4	1-5356	18545	99	73	4846	18	2012-08-02 05:18:24.287011
18550	5	1-4819	18549	79	19	927	28	2012-08-02 05:18:26.629311
18551	5	1-5347	18549	19	77	2530	95	2012-08-02 05:18:26.629311
18552	5	1-5155	18549	77	39	1268	17	2012-08-02 05:18:26.629311
18553	5	1-5298	18549	39	92	1257	27	2012-08-02 05:18:26.629311
18549	5	1-5362	18549	92	79	2423	79	2012-08-02 05:18:26.629311
18555	4	1-4317	18554	97	55	2576	4	2012-08-02 05:18:28.18679
18556	4	1-2609	18554	55	79	2775	90	2012-08-02 05:18:28.18679
18557	4	1-5139	18554	79	25	2259	99	2012-08-02 05:18:28.18679
18554	4	1-5366	18554	25	97	2962	6	2012-08-02 05:18:28.18679
18559	2	1-3916	18558	1	51	1328	83	2012-08-02 05:18:29.380288
18558	2	1-5369	18558	51	1	1177	25	2012-08-02 05:18:29.380288
18561	4	1-4286	18560	35	92	2817	85	2012-08-02 05:18:29.380288
18562	4	1-2457	18560	92	53	886	78	2012-08-02 05:18:29.380288
18563	4	1-5322	18560	53	51	1984	83	2012-08-02 05:18:29.380288
18560	4	1-5369	18560	51	35	2207	25	2012-08-02 05:18:29.380288
18565	4	1-5220	18564	22	56	2318	14	2012-08-02 05:18:29.766702
18566	4	1-5280	18564	56	60	1270	85	2012-08-02 05:18:29.766702
18567	4	1-2684	18564	60	39	683	7	2012-08-02 05:18:29.766702
18564	4	1-5370	18564	39	22	1028	68	2012-08-02 05:18:29.766702
18569	4	1-5329	18568	9	68	818	87	2012-08-02 05:18:29.766702
18570	4	1-5238	18568	68	53	875	76	2012-08-02 05:18:29.766702
18571	4	1-5010	18568	53	39	612	7	2012-08-02 05:18:29.766702
18568	4	1-5370	18568	39	9	1116	68	2012-08-02 05:18:29.766702
18573	5	1-5227	18572	29	78	323	11	2012-08-02 05:18:32.097692
18574	5	1-5299	18572	78	87	142	88	2012-08-02 05:18:32.097692
18575	5	1-5230	18572	87	88	126	71	2012-08-02 05:18:32.097692
18576	5	1-4817	18572	88	41	82	27	2012-08-02 05:18:32.097692
18572	5	1-5376	18572	41	29	513	67	2012-08-02 05:18:32.097692
18578	5	1-1399	18577	32	12	427	32	2012-08-02 05:18:32.097692
18579	5	1-3542	18577	12	94	50	96	2012-08-02 05:18:32.097692
18580	5	1-5338	18577	94	49	54	12	2012-08-02 05:18:32.097692
18581	5	1-5106	18577	49	41	129	27	2012-08-02 05:18:32.097692
18577	5	1-5376	18577	41	32	897	67	2012-08-02 05:18:32.097692
18583	4	1-4500	18582	56	28	716	60	2012-08-02 05:18:32.097692
18584	4	1-5301	18582	28	29	566	1	2012-08-02 05:18:32.097692
18585	4	1-5285	18582	29	41	259	27	2012-08-02 05:18:32.097692
18582	4	1-5376	18582	41	56	1819	67	2012-08-02 05:18:32.097692
18587	4	1-5227	18586	29	41	535	11	2012-08-02 05:18:32.097692
18588	4	1-5255	18586	41	88	403	71	2012-08-02 05:18:32.097692
18589	4	1-4817	18586	88	41	174	27	2012-08-02 05:18:32.097692
18586	4	1-5376	18586	41	29	1274	67	2012-08-02 05:18:32.097692
18591	3	1-3425	18590	70	41	1368	11	2012-08-02 05:18:32.893564
18592	3	1-5255	18590	41	67	1773	71	2012-08-02 05:18:32.893564
18590	3	1-5378	18590	67	70	3749	54	2012-08-02 05:18:32.893564
18594	4	1-3339	18593	26	27	971	86	2012-08-02 05:18:33.28023
18595	4	1-4928	18593	27	89	848	9	2012-08-02 05:18:33.28023
18596	4	1-3149	18593	89	2	765	53	2012-08-02 05:18:33.28023
18593	4	1-5379	18593	2	26	1538	4	2012-08-02 05:18:33.28023
18598	4	1-5100	18597	20	44	316	75	2012-08-02 05:18:33.28023
18599	4	1-3500	18597	44	30	342	5	2012-08-02 05:18:33.28023
18600	4	1-4640	18597	30	2	297	53	2012-08-02 05:18:33.28023
18597	4	1-5379	18597	2	20	620	4	2012-08-02 05:18:33.28023
18602	4	1-5305	18601	38	50	1556	31	2012-08-02 05:18:34.451225
18603	4	1-5265	18601	50	47	1332	92	2012-08-02 05:18:34.451225
18604	4	1-4698	18601	47	71	1056	87	2012-08-02 05:18:34.451225
18601	4	1-5382	18601	71	38	954	38	2012-08-02 05:18:34.451225
18606	5	1-5234	18605	90	42	1898	58	2012-08-02 05:18:34.451225
18607	5	1-5189	18605	42	46	1152	28	2012-08-02 05:18:34.451225
18608	5	1-5031	18605	46	47	1456	92	2012-08-02 05:18:34.451225
18609	5	1-4698	18605	47	71	1243	87	2012-08-02 05:18:34.451225
18605	5	1-5382	18605	71	90	1209	38	2012-08-02 05:18:34.451225
18611	5	1-5305	18610	38	17	665	31	2012-08-02 05:18:34.451225
18612	5	1-5342	18610	17	34	754	36	2012-08-02 05:18:34.451225
18613	5	1-2343	18610	34	30	310	5	2012-08-02 05:18:34.451225
18614	5	1-4736	18610	30	71	318	87	2012-08-02 05:18:34.451225
18610	5	1-5382	18610	71	38	342	38	2012-08-02 05:18:34.451225
18616	4	1-3825	18615	78	72	727	26	2012-08-02 05:18:34.451225
18617	4	1-5223	18615	72	30	1780	5	2012-08-02 05:18:34.451225
18618	4	1-4736	18615	30	71	1888	87	2012-08-02 05:18:34.451225
18615	4	1-5382	18615	71	78	2098	38	2012-08-02 05:18:34.451225
18620	3	1-4480	18619	11	64	1364	86	2012-08-02 05:18:35.622476
18621	3	1-4903	18619	64	93	2313	49	2012-08-02 05:18:35.622476
18619	3	1-5385	18619	93	11	1403	35	2012-08-02 05:18:35.622476
18623	4	1-4480	18622	11	27	87	86	2012-08-02 05:18:35.622476
18624	4	1-4928	18622	27	38	62	9	2012-08-02 05:18:35.622476
18625	4	1-4810	18622	38	93	106	49	2012-08-02 05:18:35.622476
18622	4	1-5385	18622	93	11	76	35	2012-08-02 05:18:35.622476
18627	5	1-4442	18626	10	8	829	25	2012-08-02 05:18:36.049046
18628	5	1-5284	18626	8	63	1026	91	2012-08-02 05:18:36.049046
18629	5	1-587	18626	63	63	679	48	2012-08-02 05:18:36.049046
18630	5	1-5279	18626	63	26	1137	89	2012-08-02 05:18:36.049046
18626	5	1-5386	18626	26	10	715	47	2012-08-02 05:18:36.049046
18632	5	1-2592	18631	57	8	241	25	2012-08-02 05:18:36.440203
18633	5	1-5284	18631	8	21	299	91	2012-08-02 05:18:36.440203
18634	5	1-4839	18631	21	55	244	89	2012-08-02 05:18:36.440203
18635	5	1-5247	18631	55	33	286	81	2012-08-02 05:18:36.440203
18631	5	1-5387	18631	33	57	414	65	2012-08-02 05:18:36.440203
18637	5	1-3781	18636	49	59	926	1	2012-08-02 05:18:37.622421
18638	5	1-5179	18636	59	54	356	82	2012-08-02 05:18:37.622421
18639	5	1-5320	18636	54	63	718	84	2012-08-02 05:18:37.622421
18640	5	1-4696	18636	63	28	637	64	2012-08-02 05:18:37.622421
18636	5	1-5390	18636	28	49	1294	65	2012-08-02 05:18:37.622421
18642	5	1-1409	18641	10	74	3039	17	2012-08-02 05:18:39.179944
18643	5	1-4642	18641	74	7	2299	11	2012-08-02 05:18:39.179944
18644	5	1-5262	18641	7	57	2335	61	2012-08-02 05:18:39.179944
18645	5	1-5116	18641	57	56	1926	84	2012-08-02 05:18:39.179944
18641	5	1-5394	18641	56	10	2332	38	2012-08-02 05:18:39.179944
18647	4	1-5071	18646	14	76	1344	92	2012-08-02 05:18:39.179944
18648	4	1-5317	18646	76	56	1230	77	2012-08-02 05:18:39.179944
18649	4	1-5237	18646	56	56	1909	84	2012-08-02 05:18:39.179944
18646	4	1-5394	18646	56	14	2624	38	2012-08-02 05:18:39.179944
18651	4	1-4091	18650	84	100	3604	72	2012-08-02 05:18:39.179944
18652	4	1-5021	18650	100	57	2784	61	2012-08-02 05:18:39.179944
18653	4	1-5116	18650	57	56	2636	84	2012-08-02 05:18:39.179944
18650	4	1-5394	18650	56	84	3663	38	2012-08-02 05:18:39.179944
18655	4	1-4167	18654	100	39	1202	50	2012-08-02 05:18:39.577832
18656	4	1-5316	18654	39	72	288	17	2012-08-02 05:18:39.577832
18657	4	1-5308	18654	72	4	1391	80	2012-08-02 05:18:39.577832
18654	4	1-5395	18654	4	100	1129	45	2012-08-02 05:18:39.577832
18659	3	1-5036	18658	61	7	2381	82	2012-08-02 05:18:39.577832
18660	3	1-3682	18658	7	4	4488	80	2012-08-02 05:18:39.577832
18658	3	1-5395	18658	4	61	5425	45	2012-08-02 05:18:39.577832
18662	5	1-5246	18661	72	41	821	11	2012-08-02 05:18:40.760271
18663	5	1-5255	18661	41	88	1043	71	2012-08-02 05:18:40.760271
18664	5	1-4817	18661	88	92	1353	27	2012-08-02 05:18:40.760271
18665	5	1-5362	18661	92	88	2645	79	2012-08-02 05:18:40.760271
18661	5	1-5398	18661	88	72	1941	100	2012-08-02 05:18:40.760271
18667	5	1-5349	18666	41	19	2631	85	2012-08-02 05:18:41.955176
18668	5	1-5331	18666	19	76	3531	54	2012-08-02 05:18:41.955176
18669	5	1-4602	18666	76	53	1744	92	2012-08-02 05:18:41.955176
18670	5	1-5367	18666	53	97	1476	91	2012-08-02 05:18:41.955176
18666	5	1-5401	18666	97	41	977	8	2012-08-02 05:18:41.955176
18672	5	1-5349	18671	41	95	2141	85	2012-08-02 05:18:41.955176
18673	5	1-4781	18671	95	3	1138	28	2012-08-02 05:18:41.955176
18674	5	1-5334	18671	3	26	1082	63	2012-08-02 05:18:41.955176
18675	5	1-4893	18671	26	97	1086	91	2012-08-02 05:18:41.955176
18671	5	1-5401	18671	97	41	756	8	2012-08-02 05:18:41.955176
18677	5	1-5349	18676	41	29	575	85	2012-08-02 05:18:41.955176
18678	5	1-4396	18676	29	12	361	39	2012-08-02 05:18:41.955176
18679	5	1-4644	18676	12	53	176	92	2012-08-02 05:18:41.955176
18680	5	1-5367	18676	53	97	193	91	2012-08-02 05:18:41.955176
18676	5	1-5401	18676	97	41	165	8	2012-08-02 05:18:41.955176
18682	4	1-5169	18681	14	24	3864	33	2012-08-02 05:18:44.275269
18683	4	1-5307	18681	24	38	3131	38	2012-08-02 05:18:44.275269
18684	4	1-5305	18681	38	39	6083	31	2012-08-02 05:18:44.275269
18681	4	1-5407	18681	39	14	3255	30	2012-08-02 05:18:44.275269
18686	5	1-623	18685	3	33	1260	81	2012-08-02 05:18:45.059638
18687	5	1-5387	18685	33	91	1682	65	2012-08-02 05:18:45.059638
18688	5	1-4433	18685	91	10	1109	49	2012-08-02 05:18:45.059638
18689	5	1-3193	18685	10	41	1440	88	2012-08-02 05:18:45.059638
18685	5	1-5409	18685	41	3	1937	91	2012-08-02 05:18:45.059638
18691	4	1-5401	18690	97	57	1713	8	2012-08-02 05:18:45.059638
18692	4	1-2876	18690	57	68	1104	70	2012-08-02 05:18:45.059638
18693	4	1-3001	18690	68	41	1546	88	2012-08-02 05:18:45.059638
18690	4	1-5409	18690	41	97	2188	91	2012-08-02 05:18:45.059638
18695	5	1-5044	18694	34	96	2124	66	2012-08-02 05:18:45.844248
18696	5	1-4083	18694	96	53	975	12	2012-08-02 05:18:45.844248
18697	5	1-4936	18694	53	42	1799	55	2012-08-02 05:18:45.844248
18698	5	1-3929	18694	42	66	1374	23	2012-08-02 05:18:45.844248
18694	5	1-5411	18694	66	34	2151	30	2012-08-02 05:18:45.844248
18700	4	1-4799	18699	9	19	492	28	2012-08-02 05:18:45.844248
18701	4	1-5347	18699	19	85	1405	95	2012-08-02 05:18:45.844248
18702	4	1-1463	18699	85	66	408	23	2012-08-02 05:18:45.844248
18699	4	1-5411	18699	66	9	706	30	2012-08-02 05:18:45.844248
18704	5	1-5319	18703	59	66	5255	70	2012-08-02 05:18:47.026346
18705	5	1-5242	18703	66	79	1323	98	2012-08-02 05:18:47.026346
18706	5	1-5345	18703	79	81	2093	8	2012-08-02 05:18:47.026346
18707	5	1-5372	18703	81	97	1499	18	2012-08-02 05:18:47.026346
18703	5	1-5414	18703	97	59	9543	82	2012-08-02 05:18:47.026346
18709	4	1-5237	18708	56	63	943	84	2012-08-02 05:18:47.424294
18710	4	1-4696	18708	63	28	640	64	2012-08-02 05:18:47.424294
18711	4	1-5390	18708	28	97	994	65	2012-08-02 05:18:47.424294
18708	4	1-5415	18708	97	56	847	77	2012-08-02 05:18:47.424294
18713	5	1-4146	18712	48	3	1431	28	2012-08-02 05:18:47.810841
18714	5	1-5334	18712	3	85	1623	63	2012-08-02 05:18:47.810841
18715	5	1-5224	18712	85	48	1319	94	2012-08-02 05:18:47.810841
18716	5	1-4137	18712	48	62	973	45	2012-08-02 05:18:47.810841
18712	5	1-5416	18712	62	48	2590	48	2012-08-02 05:18:47.810841
18718	4	1-5158	18717	3	29	2233	67	2012-08-02 05:18:48.197908
18719	4	1-5227	18717	29	41	2359	11	2012-08-02 05:18:48.197908
18720	4	1-5255	18717	41	46	2558	71	2012-08-02 05:18:48.197908
18717	4	1-5417	18717	46	3	7173	63	2012-08-02 05:18:48.197908
18722	5	1-4893	18721	26	70	165	91	2012-08-02 05:18:48.197908
18723	5	1-1623	18721	70	71	101	87	2012-08-02 05:18:48.197908
18724	5	1-5382	18721	71	46	93	38	2012-08-02 05:18:48.197908
18725	5	1-5249	18721	46	46	52	71	2012-08-02 05:18:48.197908
18721	5	1-5417	18721	46	26	161	63	2012-08-02 05:18:48.197908
18727	4	1-4413	18726	75	25	2196	99	2012-08-02 05:18:48.960157
18728	4	1-5366	18726	25	18	2285	6	2012-08-02 05:18:48.960157
18726	4	1-5419	18726	29	75	5195	30	2012-08-02 05:18:48.960157
18729	4	1-4270	18726	18	29	2051	85	2012-08-02 05:18:48.960157
18731	4	1-3723	18730	55	7	1216	11	2012-08-02 05:18:48.960157
18732	4	1-5262	18730	7	51	1165	61	2012-08-02 05:18:48.960157
18733	4	1-3495	18730	51	29	608	85	2012-08-02 05:18:48.960157
18730	4	1-5419	18730	29	55	1559	30	2012-08-02 05:18:48.960157
18735	3	1-5344	18734	83	39	2393	29	2012-08-02 05:18:49.772118
18736	3	1-5182	18734	39	31	5171	52	2012-08-02 05:18:49.772118
18734	3	1-5421	18734	31	83	2871	78	2012-08-02 05:18:49.772118
18738	4	1-4820	18737	89	20	8148	88	2012-08-02 05:18:50.946782
18739	4	1-5109	18737	20	56	5377	44	2012-08-02 05:18:50.946782
18740	4	1-5343	18737	56	53	7654	20	2012-08-02 05:18:50.946782
18737	4	1-5424	18737	53	89	4781	96	2012-08-02 05:18:50.946782
18742	5	1-4820	18741	89	30	1038	88	2012-08-02 05:18:50.946782
18743	5	1-5209	18741	30	52	541	53	2012-08-02 05:18:50.946782
18744	5	1-5321	18741	52	79	739	90	2012-08-02 05:18:50.946782
18745	5	1-4843	18741	79	53	772	20	2012-08-02 05:18:50.946782
18741	5	1-5424	18741	53	89	542	96	2012-08-02 05:18:50.946782
18747	2	1-4409	18746	3	53	2540	20	2012-08-02 05:18:50.946782
18746	2	1-5424	18746	53	3	2121	96	2012-08-02 05:18:50.946782
18749	4	1-5237	18748	56	77	2669	84	2012-08-02 05:18:51.357374
18750	4	1-5368	18748	77	50	1813	63	2012-08-02 05:18:51.357374
18751	4	1-2690	18748	50	33	1359	80	2012-08-02 05:18:51.357374
18748	4	1-5425	18748	33	56	1980	77	2012-08-02 05:18:51.357374
18753	4	1-5352	18752	100	82	65	67	2012-08-02 05:18:51.357374
18754	4	1-4028	18752	82	6	61	36	2012-08-02 05:18:51.357374
18755	4	1-2226	18752	6	33	42	80	2012-08-02 05:18:51.357374
18752	4	1-5425	18752	33	100	68	77	2012-08-02 05:18:51.357374
18757	5	1-5407	18756	39	55	119	30	2012-08-02 05:18:52.132259
18758	5	1-3723	18756	55	7	99	11	2012-08-02 05:18:52.132259
18759	5	1-5262	18756	7	89	101	61	2012-08-02 05:18:52.132259
18760	5	1-1924	18756	89	59	33	55	2012-08-02 05:18:52.132259
18756	5	1-5427	18756	59	39	254	31	2012-08-02 05:18:52.132259
18762	4	1-4401	18761	31	8	409	97	2012-08-02 05:18:52.132259
18763	4	1-4956	18761	8	89	414	61	2012-08-02 05:18:52.132259
18764	4	1-1924	18761	89	59	156	55	2012-08-02 05:18:52.132259
18761	4	1-5427	18761	59	31	1400	31	2012-08-02 05:18:52.132259
18766	4	1-4565	18765	55	97	2193	52	2012-08-02 05:18:52.531287
18767	4	1-4440	18765	97	53	1306	92	2012-08-02 05:18:52.531287
18768	4	1-5367	18765	53	89	1066	91	2012-08-02 05:18:52.531287
18765	4	1-5428	18765	89	55	1696	47	2012-08-02 05:18:52.531287
18770	4	1-4967	18769	16	56	4525	76	2012-08-02 05:18:52.531287
18771	4	1-5028	18769	56	53	1871	92	2012-08-02 05:18:52.531287
18772	4	1-5367	18769	53	89	1760	91	2012-08-02 05:18:52.531287
18769	4	1-5428	18769	89	16	3228	47	2012-08-02 05:18:52.531287
18774	5	1-5062	18773	49	66	2042	23	2012-08-02 05:18:52.531287
18775	5	1-5411	18773	66	53	3445	30	2012-08-02 05:18:52.531287
18776	5	1-4733	18773	53	53	1950	92	2012-08-02 05:18:52.531287
18777	5	1-5367	18773	53	89	2121	91	2012-08-02 05:18:52.531287
18773	5	1-5428	18773	89	49	4426	47	2012-08-02 05:18:52.531287
18779	5	1-5417	18778	46	3	845	63	2012-08-02 05:18:53.339494
18780	5	1-5035	18778	3	89	443	96	2012-08-02 05:18:53.339494
18781	5	1-4820	18778	89	36	768	88	2012-08-02 05:18:53.339494
18782	5	1-5381	18778	36	90	699	2	2012-08-02 05:18:53.339494
18778	5	1-5430	18778	90	46	345	71	2012-08-02 05:18:53.339494
18784	5	1-5417	18783	46	3	530	63	2012-08-02 05:18:53.339494
18785	5	1-5035	18783	3	7	281	96	2012-08-02 05:18:53.339494
18786	5	1-4897	18783	7	42	359	89	2012-08-02 05:18:53.339494
18787	5	1-5296	18783	42	90	428	2	2012-08-02 05:18:53.339494
18783	5	1-5430	18783	90	46	214	71	2012-08-02 05:18:53.339494
18789	3	1-4599	18788	62	50	2581	39	2012-08-02 05:18:53.72762
18790	3	1-2665	18788	50	94	595	47	2012-08-02 05:18:53.72762
18788	3	1-5431	18788	94	62	2868	59	2012-08-02 05:18:53.72762
18792	5	1-4946	18791	33	51	1547	34	2012-08-02 05:18:53.72762
18793	5	1-4962	18791	51	27	2511	22	2012-08-02 05:18:53.72762
18794	5	1-5396	18791	27	21	1406	3	2012-08-02 05:18:53.72762
18795	5	1-4115	18791	21	94	845	47	2012-08-02 05:18:53.72762
18791	5	1-5431	18791	94	33	4216	59	2012-08-02 05:18:53.72762
18797	5	1-1574	18796	66	72	13	24	2012-08-02 05:18:54.13765
18798	5	1-4951	18796	72	50	12	32	2012-08-02 05:18:54.13765
18799	5	1-4143	18796	50	33	13	81	2012-08-02 05:18:54.13765
18800	5	1-5387	18796	33	93	19	65	2012-08-02 05:18:54.13765
18796	5	1-5432	18796	93	66	67	73	2012-08-02 05:18:54.13765
18802	3	1-2684	18801	60	66	1134	7	2012-08-02 05:18:54.549444
18803	3	1-4830	18801	66	83	311	64	2012-08-02 05:18:54.549444
18801	3	1-5433	18801	83	60	1920	85	2012-08-02 05:18:54.549444
18805	5	1-3324	18804	55	7	3108	11	2012-08-02 05:18:54.549444
18806	5	1-5262	18804	7	36	3534	61	2012-08-02 05:18:54.549444
18807	5	1-3217	18804	36	66	2250	7	2012-08-02 05:18:54.549444
18808	5	1-4830	18804	66	83	650	64	2012-08-02 05:18:54.549444
18804	5	1-5433	18804	83	55	4223	85	2012-08-02 05:18:54.549444
18810	5	1-5192	18809	44	27	682	22	2012-08-02 05:18:54.947725
18811	5	1-5396	18809	27	21	465	3	2012-08-02 05:18:54.947725
18812	5	1-4115	18809	21	14	341	47	2012-08-02 05:18:54.947725
18813	5	1-2537	18809	14	1	328	28	2012-08-02 05:18:54.947725
18809	5	1-5434	18809	1	44	275	48	2012-08-02 05:18:54.947725
18815	4	1-5106	18814	49	51	2998	27	2012-08-02 05:18:55.357779
18816	4	1-4386	18814	51	22	1524	51	2012-08-02 05:18:55.357779
18817	4	1-5132	18814	22	93	1241	34	2012-08-02 05:18:55.357779
18814	4	1-5435	18814	93	49	1153	12	2012-08-02 05:18:55.357779
18819	5	1-3838	18818	56	39	4669	17	2012-08-02 05:18:55.357779
18820	5	1-5298	18818	39	51	4148	27	2012-08-02 05:18:55.357779
18821	5	1-4386	18818	51	22	3013	51	2012-08-02 05:18:55.357779
18822	5	1-5132	18818	22	93	3503	34	2012-08-02 05:18:55.357779
18818	5	1-5435	18818	93	56	4649	12	2012-08-02 05:18:55.357779
18824	4	1-5416	18823	62	44	930	48	2012-08-02 05:18:55.756153
18825	4	1-5192	18823	44	66	1457	22	2012-08-02 05:18:55.756153
18826	4	1-4509	18823	66	86	816	100	2012-08-02 05:18:55.756153
18823	4	1-5436	18823	86	62	537	45	2012-08-02 05:18:55.756153
18828	4	1-5416	18827	62	44	6011	48	2012-08-02 05:18:55.756153
18829	4	1-4995	18827	44	53	1263	9	2012-08-02 05:18:55.756153
18830	4	1-5286	18827	53	86	4641	100	2012-08-02 05:18:55.756153
18827	4	1-5436	18827	86	62	3254	45	2012-08-02 05:18:55.756153
18832	5	1-5416	18831	62	44	407	48	2012-08-02 05:18:55.756153
18833	5	1-4995	18831	44	65	124	9	2012-08-02 05:18:55.756153
18834	5	1-3035	18831	65	3	119	51	2012-08-02 05:18:55.756153
18835	5	1-5277	18831	3	86	151	100	2012-08-02 05:18:55.756153
18831	5	1-5436	18831	86	62	153	45	2012-08-02 05:18:55.756153
18837	4	1-3797	18836	31	78	1072	95	2012-08-02 05:18:55.756153
18838	4	1-5124	18836	78	69	2842	50	2012-08-02 05:18:55.756153
18839	4	1-4858	18836	69	86	1256	100	2012-08-02 05:18:55.756153
18836	4	1-5436	18836	86	31	1299	45	2012-08-02 05:18:55.756153
18841	5	1-5338	18840	94	40	319	12	2012-08-02 05:18:57.292673
18842	5	1-4403	18840	40	34	539	57	2012-08-02 05:18:57.292673
18843	5	1-3813	18840	34	7	355	11	2012-08-02 05:18:57.292673
18844	5	1-5262	18840	7	65	418	61	2012-08-02 05:18:57.292673
18840	5	1-5440	18840	65	94	165	96	2012-08-02 05:18:57.292673
18846	3	1-5247	18845	55	69	4815	81	2012-08-02 05:18:58.077291
18847	3	1-5375	18845	69	15	4100	43	2012-08-02 05:18:58.077291
18845	3	1-5442	18845	15	55	7744	89	2012-08-02 05:18:58.077291
18849	5	1-5385	18848	93	9	876	35	2012-08-02 05:18:58.475296
18850	5	1-3800	18848	9	11	377	67	2012-08-02 05:18:58.475296
18851	5	1-4366	18848	11	87	956	40	2012-08-02 05:18:58.475296
18852	5	1-5288	18848	87	98	1757	43	2012-08-02 05:18:58.475296
18848	5	1-5443	18848	98	93	953	49	2012-08-02 05:18:58.475296
18854	4	1-5385	18853	93	48	581	35	2012-08-02 05:18:58.475296
18855	4	1-3252	18853	48	87	573	40	2012-08-02 05:18:58.475296
18856	4	1-5288	18853	87	98	1089	43	2012-08-02 05:18:58.475296
18853	4	1-5443	18853	98	93	611	49	2012-08-02 05:18:58.475296
18858	5	1-4698	18857	47	71	1283	87	2012-08-02 05:18:59.668851
18859	5	1-5382	18857	71	14	1258	38	2012-08-02 05:18:59.668851
18860	5	1-5261	18857	14	5	536	23	2012-08-02 05:18:59.668851
18861	5	1-5311	18857	5	35	1028	25	2012-08-02 05:18:59.668851
18857	5	1-5446	18857	35	47	1491	92	2012-08-02 05:18:59.668851
18863	5	1-5367	18862	53	21	837	91	2012-08-02 05:18:59.668851
18864	5	1-4839	18862	21	26	690	89	2012-08-02 05:18:59.668851
18865	5	1-5386	18862	26	10	439	47	2012-08-02 05:18:59.668851
18866	5	1-4442	18862	10	35	516	25	2012-08-02 05:18:59.668851
18862	5	1-5446	18862	35	53	803	92	2012-08-02 05:18:59.668851
18868	4	1-3824	18867	27	33	112	81	2012-08-02 05:18:59.668851
18869	4	1-5387	18867	33	57	164	65	2012-08-02 05:18:59.668851
18870	4	1-2592	18867	57	35	97	25	2012-08-02 05:18:59.668851
18867	4	1-5446	18867	35	27	152	92	2012-08-02 05:18:59.668851
18872	5	1-5352	18871	100	29	1762	67	2012-08-02 05:19:00.07895
18873	5	1-5227	18871	29	86	1699	11	2012-08-02 05:19:00.07895
18874	5	1-434	18871	86	3	711	51	2012-08-02 05:19:00.07895
18875	5	1-5277	18871	3	68	702	100	2012-08-02 05:19:00.07895
18871	5	1-5447	18871	68	100	2254	77	2012-08-02 05:19:00.07895
18877	5	1-3475	18876	16	39	2374	17	2012-08-02 05:19:00.07895
18878	5	1-5298	18876	39	51	2056	27	2012-08-02 05:19:00.07895
18879	5	1-4386	18876	51	3	1456	51	2012-08-02 05:19:00.07895
18880	5	1-5277	18876	3	68	1589	100	2012-08-02 05:19:00.07895
18876	5	1-5447	18876	68	16	5646	77	2012-08-02 05:19:00.07895
18882	4	1-5073	18881	66	78	454	39	2012-08-02 05:19:00.07895
18883	4	1-4868	18881	78	34	367	40	2012-08-02 05:19:00.07895
18884	4	1-4969	18881	34	68	178	100	2012-08-02 05:19:00.07895
18881	4	1-5447	18881	68	66	678	77	2012-08-02 05:19:00.07895
18886	4	1-5412	18885	75	33	2417	80	2012-08-02 05:19:01.249943
18887	4	1-5425	18885	33	100	3902	77	2012-08-02 05:19:01.249943
18888	4	1-5352	18885	100	40	3733	67	2012-08-02 05:19:01.249943
18885	4	1-5450	18885	40	75	8174	14	2012-08-02 05:19:01.249943
18890	5	1-5007	18889	57	10	2566	65	2012-08-02 05:19:01.637224
18891	5	1-4837	18889	10	83	2838	78	2012-08-02 05:19:01.637224
18892	5	1-5344	18889	83	56	2731	29	2012-08-02 05:19:01.637224
18893	5	1-4907	18889	56	64	2664	82	2012-08-02 05:19:01.637224
18889	5	1-5451	18889	64	57	7684	46	2012-08-02 05:19:01.637224
18895	4	1-5275	18894	54	7	1102	11	2012-08-02 05:19:02.432852
18896	4	1-5262	18894	7	36	1256	61	2012-08-02 05:19:02.432852
18897	4	1-3217	18894	36	10	802	7	2012-08-02 05:19:02.432852
18894	4	1-5453	18894	10	54	742	75	2012-08-02 05:19:02.432852
18899	4	1-4622	18898	67	10	1316	65	2012-08-02 05:19:02.432852
18900	4	1-4837	18898	10	52	1511	78	2012-08-02 05:19:02.432852
18901	4	1-3405	18898	52	10	1010	7	2012-08-02 05:19:02.432852
18898	4	1-5453	18898	10	67	951	75	2012-08-02 05:19:02.432852
18903	5	1-2653	18902	53	100	1187	64	2012-08-02 05:19:02.830427
18904	5	1-4881	18902	100	21	2620	10	2012-08-02 05:19:02.830427
18905	5	1-5038	18902	21	47	2700	41	2012-08-02 05:19:02.830427
18906	5	1-5143	18902	47	58	2327	26	2012-08-02 05:19:02.830427
18902	5	1-5454	18902	58	53	6346	21	2012-08-02 05:19:02.830427
18908	5	1-5453	18907	10	3	147	75	2012-08-02 05:19:03.217248
18909	5	1-4333	18907	3	82	251	62	2012-08-02 05:19:03.217248
18910	5	1-4549	18907	82	28	296	11	2012-08-02 05:19:03.217248
18911	5	1-4635	18907	28	41	216	84	2012-08-02 05:19:03.217248
18907	5	1-5455	18907	41	10	163	7	2012-08-02 05:19:03.217248
18913	4	1-5453	18912	10	54	2763	75	2012-08-02 05:19:03.217248
18914	4	1-5275	18912	54	28	4346	11	2012-08-02 05:19:03.217248
18915	4	1-4635	18912	28	41	3439	84	2012-08-02 05:19:03.217248
18912	4	1-5455	18912	41	10	2822	7	2012-08-02 05:19:03.217248
18917	5	1-5304	18916	17	86	2537	100	2012-08-02 05:19:03.615025
18918	5	1-5436	18916	86	31	2428	45	2012-08-02 05:19:03.615025
18919	5	1-3797	18916	31	59	1854	95	2012-08-02 05:19:03.615025
18920	5	1-4882	18916	59	44	2229	62	2012-08-02 05:19:03.615025
18916	5	1-5456	18916	44	17	3529	80	2012-08-02 05:19:03.615025
18922	4	1-3234	18921	40	45	5298	88	2012-08-02 05:19:04.012918
18923	4	1-5438	18921	45	9	3966	68	2012-08-02 05:19:04.012918
18924	4	1-5329	18921	9	60	2589	87	2012-08-02 05:19:04.012918
18921	4	1-5457	18921	60	40	4635	94	2012-08-02 05:19:04.012918
18926	5	1-5445	18925	88	10	5931	73	2012-08-02 05:19:04.012918
18927	5	1-5108	18925	10	73	470	18	2012-08-02 05:19:04.012918
18928	5	1-4758	18925	73	81	558	48	2012-08-02 05:19:04.012918
18929	5	1-4585	18925	81	60	482	87	2012-08-02 05:19:04.012918
18925	5	1-5457	18925	60	88	903	94	2012-08-02 05:19:04.012918
18931	4	1-2690	18930	50	17	1874	80	2012-08-02 05:19:05.200332
18932	4	1-5304	18930	17	86	1220	100	2012-08-02 05:19:05.200332
18933	4	1-5436	18930	86	46	1057	45	2012-08-02 05:19:05.200332
18930	4	1-5460	18930	46	50	2515	63	2012-08-02 05:19:05.200332
18935	5	1-5035	18934	3	35	33	96	2012-08-02 05:19:05.200332
18936	5	1-2974	18934	35	3	18	51	2012-08-02 05:19:05.200332
18937	5	1-5277	18934	3	86	22	100	2012-08-02 05:19:05.200332
18938	5	1-5436	18934	86	46	20	45	2012-08-02 05:19:05.200332
18934	5	1-5460	18934	46	3	49	63	2012-08-02 05:19:05.200332
18940	5	1-4926	18939	65	78	578	39	2012-08-02 05:19:05.200332
18941	5	1-4868	18939	78	34	498	40	2012-08-02 05:19:05.200332
18942	5	1-4969	18939	34	86	258	100	2012-08-02 05:19:05.200332
18943	5	1-5436	18939	86	46	257	45	2012-08-02 05:19:05.200332
18939	5	1-5460	18939	46	65	703	63	2012-08-02 05:19:05.200332
18945	5	1-3368	18944	62	54	427	59	2012-08-02 05:19:06.39582
18946	5	1-5365	18944	54	8	336	4	2012-08-02 05:19:06.39582
18947	5	1-5278	18944	8	54	105	82	2012-08-02 05:19:06.39582
18948	5	1-5320	18944	54	43	194	84	2012-08-02 05:19:06.39582
18944	5	1-5463	18944	43	62	640	29	2012-08-02 05:19:06.39582
18950	5	1-4818	18949	33	48	1800	94	2012-08-02 05:19:06.39582
18951	5	1-4137	18949	48	61	994	45	2012-08-02 05:19:06.39582
18952	5	1-5036	18949	61	54	356	82	2012-08-02 05:19:06.39582
18953	5	1-5320	18949	54	43	589	84	2012-08-02 05:19:06.39582
18949	5	1-5463	18949	43	33	2105	29	2012-08-02 05:19:06.39582
18955	4	1-4434	18954	78	23	655	29	2012-08-02 05:19:09.876135
18956	4	1-4701	18954	23	6	642	41	2012-08-02 05:19:09.876135
18957	4	1-3861	18954	6	84	217	32	2012-08-02 05:19:09.876135
18954	4	1-5472	18954	84	78	852	3	2012-08-02 05:19:09.876135
18959	5	1-4434	18958	78	33	455	29	2012-08-02 05:19:09.876135
18960	5	1-4818	18958	33	29	435	94	2012-08-02 05:19:09.876135
18961	5	1-4922	18958	29	72	189	24	2012-08-02 05:19:09.876135
18962	5	1-4951	18958	72	84	145	32	2012-08-02 05:19:09.876135
18958	5	1-5472	18958	84	78	580	3	2012-08-02 05:19:09.876135
18964	5	1-4115	18963	21	7	2260	47	2012-08-02 05:19:09.876135
18965	5	1-4427	18963	7	97	2013	6	2012-08-02 05:19:09.876135
18966	5	1-4317	18963	97	47	1543	4	2012-08-02 05:19:09.876135
18967	5	1-3767	18963	47	84	825	32	2012-08-02 05:19:09.876135
18963	5	1-5472	18963	84	21	3504	3	2012-08-02 05:19:09.876135
18969	3	1-4117	18968	4	57	1051	35	2012-08-02 05:19:09.876135
18970	3	1-1568	18968	57	84	389	32	2012-08-02 05:19:09.876135
18968	3	1-5472	18968	84	4	1710	3	2012-08-02 05:19:09.876135
18972	4	1-4434	18971	78	36	1588	29	2012-08-02 05:19:09.876135
18973	4	1-4716	18971	36	47	653	4	2012-08-02 05:19:09.876135
18974	4	1-3767	18971	47	84	380	32	2012-08-02 05:19:09.876135
18971	4	1-5472	18971	84	78	1755	3	2012-08-02 05:19:09.876135
18976	5	1-4434	18975	78	49	110	29	2012-08-02 05:19:10.285444
18977	5	1-4553	18975	49	5	65	97	2012-08-02 05:19:10.285444
18978	5	1-5207	18975	5	79	97	90	2012-08-02 05:19:10.285444
18979	5	1-5139	18975	79	9	72	99	2012-08-02 05:19:10.285444
18975	5	1-5473	18975	9	78	127	3	2012-08-02 05:19:10.285444
18981	5	1-4401	18980	31	8	172	97	2012-08-02 05:19:10.694063
18982	5	1-4956	18980	8	36	155	61	2012-08-02 05:19:10.694063
18983	5	1-3217	18980	36	73	91	7	2012-08-02 05:19:10.694063
18984	5	1-5131	18980	73	77	128	58	2012-08-02 05:19:10.694063
18980	5	1-5474	18980	77	31	665	31	2012-08-02 05:19:10.694063
18986	4	1-4401	18985	31	8	432	97	2012-08-02 05:19:10.694063
18987	4	1-4956	18985	8	100	404	61	2012-08-02 05:19:10.694063
18988	4	1-2973	18985	100	77	297	58	2012-08-02 05:19:10.694063
18985	4	1-5474	18985	77	31	1605	31	2012-08-02 05:19:10.694063
18990	4	1-4262	18989	14	37	1551	83	2012-08-02 05:19:10.694063
18991	4	1-4590	18989	37	73	346	7	2012-08-02 05:19:10.694063
18992	4	1-5131	18989	73	77	539	58	2012-08-02 05:19:10.694063
18989	4	1-5474	18989	77	14	3119	31	2012-08-02 05:19:10.694063
18994	4	1-5374	18993	6	51	2348	34	2012-08-02 05:19:11.48989
18995	4	1-4962	18993	51	23	4124	22	2012-08-02 05:19:11.48989
18996	4	1-4676	18993	23	52	1173	64	2012-08-02 05:19:11.48989
18993	4	1-5476	18993	52	6	2061	55	2012-08-02 05:19:11.48989
18998	5	1-3929	18997	42	66	1124	23	2012-08-02 05:19:11.48989
18999	5	1-5411	18997	66	34	1721	30	2012-08-02 05:19:11.48989
19000	5	1-5044	18997	34	99	1663	66	2012-08-02 05:19:11.48989
19001	5	1-5441	18997	99	52	852	64	2012-08-02 05:19:11.48989
18997	5	1-5476	18997	52	42	1503	55	2012-08-02 05:19:11.48989
19003	4	1-4085	19002	13	3	1013	51	2012-08-02 05:19:12.285434
19004	4	1-5277	19002	3	48	1297	100	2012-08-02 05:19:12.285434
19005	4	1-5099	19002	48	76	613	97	2012-08-02 05:19:12.285434
19002	4	1-5478	19002	76	13	1210	5	2012-08-02 05:19:12.285434
19007	5	1-4447	19006	28	8	5679	79	2012-08-02 05:19:13.058903
19008	5	1-5371	19006	8	92	2750	93	2012-08-02 05:19:13.058903
19009	5	1-5126	19006	92	41	1701	56	2012-08-02 05:19:13.058903
19010	5	1-5212	19006	41	13	1878	51	2012-08-02 05:19:13.058903
19006	5	1-5480	19006	13	28	5670	35	2012-08-02 05:19:13.058903
19012	3	1-5188	19011	47	19	1878	92	2012-08-02 05:19:13.058903
19013	3	1-2828	19011	19	13	860	51	2012-08-02 05:19:13.058903
19011	3	1-5480	19011	13	47	2687	35	2012-08-02 05:19:13.058903
19015	5	1-4396	19014	29	12	267	39	2012-08-02 05:19:14.671889
19016	5	1-4644	19014	12	27	114	92	2012-08-02 05:19:14.671889
19017	5	1-3824	19014	27	88	77	81	2012-08-02 05:19:14.671889
19018	5	1-5466	19014	88	71	315	14	2012-08-02 05:19:14.671889
19014	5	1-5484	19014	71	29	485	85	2012-08-02 05:19:14.671889
19020	4	1-2457	19019	92	28	1305	78	2012-08-02 05:19:14.671889
19021	4	1-5350	19019	28	22	1091	68	2012-08-02 05:19:14.671889
19022	4	1-5220	19019	22	71	2774	14	2012-08-02 05:19:14.671889
19019	4	1-5484	19019	71	92	4508	85	2012-08-02 05:19:14.671889
19024	4	1-3929	19023	42	5	2179	23	2012-08-02 05:19:15.467543
19025	4	1-5311	19023	5	35	3988	25	2012-08-02 05:19:15.467543
19026	4	1-5446	19023	35	50	5509	92	2012-08-02 05:19:15.467543
19023	4	1-5486	19023	50	42	3065	55	2012-08-02 05:19:15.467543
19027	1	1-3929	\N	42	42	907	23	2012-08-02 05:19:15.467543
19029	4	1-3600	19028	3	90	540	19	2012-08-02 05:19:15.865354
19030	4	1-4187	19028	90	58	340	26	2012-08-02 05:19:15.865354
19031	4	1-5454	19028	58	59	728	21	2012-08-02 05:19:15.865354
19028	4	1-5487	19028	59	3	1268	91	2012-08-02 05:19:15.865354
19033	4	1-4839	19032	21	55	2931	89	2012-08-02 05:19:15.865354
19034	4	1-5247	19032	55	95	3230	81	2012-08-02 05:19:15.865354
19035	4	1-3808	19032	95	59	1882	21	2012-08-02 05:19:15.865354
19032	4	1-5487	19032	59	21	3821	91	2012-08-02 05:19:15.865354
19037	5	1-5401	19036	97	81	868	8	2012-08-02 05:19:15.865354
19038	5	1-5372	19036	81	98	886	18	2012-08-02 05:19:15.865354
19039	5	1-5057	19036	98	38	819	30	2012-08-02 05:19:15.865354
19040	5	1-3377	19036	38	59	559	21	2012-08-02 05:19:15.865354
19036	5	1-5487	19036	59	97	1140	91	2012-08-02 05:19:15.865354
19042	5	1-1972	19041	31	5	592	97	2012-08-02 05:19:15.865354
19043	5	1-5207	19041	5	79	886	90	2012-08-02 05:19:15.865354
19044	5	1-5139	19041	79	49	666	99	2012-08-02 05:19:15.865354
19045	5	1-3272	19041	49	59	489	21	2012-08-02 05:19:15.865354
19041	5	1-5487	19041	59	31	1031	91	2012-08-02 05:19:15.865354
19047	3	1-623	19046	3	95	312	81	2012-08-02 05:19:15.865354
19048	3	1-3808	19046	95	59	197	21	2012-08-02 05:19:15.865354
19046	3	1-5487	19046	59	3	434	91	2012-08-02 05:19:15.865354
19050	3	1-5223	19049	72	98	2621	5	2012-08-02 05:19:16.274495
19051	3	1-5404	19049	98	93	1303	6	2012-08-02 05:19:16.274495
19049	3	1-5488	19049	93	72	1695	26	2012-08-02 05:19:16.274495
19053	4	1-3742	19052	78	76	813	97	2012-08-02 05:19:16.274495
19054	4	1-5478	19052	76	98	1186	5	2012-08-02 05:19:16.274495
19055	4	1-5404	19052	98	93	676	6	2012-08-02 05:19:16.274495
19052	4	1-5488	19052	93	78	1008	26	2012-08-02 05:19:16.274495
19057	5	1-311	19056	94	52	420	53	2012-08-02 05:19:16.274495
19058	5	1-5321	19056	52	79	583	90	2012-08-02 05:19:16.274495
19059	5	1-5139	19056	79	25	387	99	2012-08-02 05:19:16.274495
19060	5	1-5366	19056	25	93	414	6	2012-08-02 05:19:16.274495
19056	5	1-5488	19056	93	94	690	26	2012-08-02 05:19:16.274495
19062	4	1-5292	19061	15	76	303	97	2012-08-02 05:19:16.274495
19063	4	1-5478	19061	76	98	570	5	2012-08-02 05:19:16.274495
19064	4	1-5404	19061	98	93	419	6	2012-08-02 05:19:16.274495
19061	4	1-5488	19061	93	15	805	26	2012-08-02 05:19:16.274495
19065	1	1-5139	\N	79	79	3466	99	2012-08-02 05:19:16.274495
19067	5	1-4317	19066	97	8	3424	4	2012-08-02 05:19:16.685433
19068	5	1-5278	19066	8	54	1326	82	2012-08-02 05:19:16.685433
19069	5	1-5320	19066	54	41	2511	84	2012-08-02 05:19:16.685433
19070	5	1-5455	19066	41	38	1735	7	2012-08-02 05:19:16.685433
19066	5	1-5489	19066	38	97	4621	6	2012-08-02 05:19:16.685433
19072	5	1-4410	19071	87	22	4308	100	2012-08-02 05:19:17.083133
19073	5	1-3959	19071	22	31	1858	13	2012-08-02 05:19:17.083133
19074	5	1-5216	19071	31	20	786	99	2012-08-02 05:19:17.083133
19075	5	1-5318	19071	20	52	2705	17	2012-08-02 05:19:17.083133
19071	5	1-5490	19071	52	87	3668	66	2012-08-02 05:19:17.083133
19077	4	1-5441	19076	99	52	54	64	2012-08-02 05:19:17.083133
19078	4	1-5476	19076	52	10	97	55	2012-08-02 05:19:17.083133
19079	4	1-2794	19076	10	52	80	17	2012-08-02 05:19:17.083133
19076	4	1-5490	19076	52	99	116	66	2012-08-02 05:19:17.083133
19081	2	1-3210	19080	19	29	6275	81	2012-08-02 05:19:17.481055
19080	2	1-5491	19080	29	19	7498	8	2012-08-02 05:19:17.481055
19083	3	1-5453	19082	10	67	3199	75	2012-08-02 05:19:17.878589
19084	3	1-4622	19082	67	77	3928	65	2012-08-02 05:19:17.878589
19082	3	1-5492	19082	77	10	3991	7	2012-08-02 05:19:17.878589
19086	5	1-5370	19085	39	83	4596	68	2012-08-02 05:19:17.878589
19087	5	1-5245	19085	83	23	2976	22	2012-08-02 05:19:17.878589
19088	5	1-4676	19085	23	28	943	64	2012-08-02 05:19:17.878589
19089	5	1-5390	19085	28	77	2092	65	2012-08-02 05:19:17.878589
19085	5	1-5492	19085	77	39	2543	7	2012-08-02 05:19:17.878589
19091	5	1-2457	19090	92	28	107	78	2012-08-02 05:19:19.049905
19092	5	1-5350	19090	28	9	89	68	2012-08-02 05:19:19.049905
19093	5	1-5329	19090	9	60	60	87	2012-08-02 05:19:19.049905
19094	5	1-5457	19090	60	39	110	94	2012-08-02 05:19:19.049905
19090	5	1-5495	19090	39	92	375	85	2012-08-02 05:19:19.049905
19096	5	1-3324	19095	55	86	245	11	2012-08-02 05:19:19.049905
19097	5	1-434	19095	86	76	132	51	2012-08-02 05:19:19.049905
19098	5	1-3809	19095	76	60	46	87	2012-08-02 05:19:19.049905
19099	5	1-5457	19095	60	39	90	94	2012-08-02 05:19:19.049905
19095	5	1-5495	19095	39	55	325	85	2012-08-02 05:19:19.049905
19101	5	1-4610	19100	91	17	709	80	2012-08-02 05:19:19.447924
19102	5	1-5304	19100	17	86	429	100	2012-08-02 05:19:19.447924
19103	5	1-5436	19100	86	61	345	45	2012-08-02 05:19:19.447924
19104	5	1-5036	19100	61	7	129	82	2012-08-02 05:19:19.447924
19100	5	1-5496	19100	7	91	1993	21	2012-08-02 05:19:19.447924
19106	5	1-5000	19105	23	8	2237	68	2012-08-02 05:19:19.447924
19107	5	1-5363	19105	8	82	1547	72	2012-08-02 05:19:19.447924
19108	5	1-4641	19105	82	59	976	1	2012-08-02 05:19:19.447924
19109	5	1-5179	19105	59	7	384	82	2012-08-02 05:19:19.447924
19105	5	1-5496	19105	7	23	7125	21	2012-08-02 05:19:19.447924
19111	3	1-5350	19110	28	66	3752	68	2012-08-02 05:19:21.204485
19112	3	1-4484	19110	66	91	2260	54	2012-08-02 05:19:21.204485
19110	3	1-5500	19110	91	28	7018	78	2012-08-02 05:19:21.204485
19114	5	1-5344	19113	83	24	773	29	2012-08-02 05:19:21.204485
19115	5	1-5448	19113	24	48	1352	16	2012-08-02 05:19:21.204485
19116	5	1-4445	19113	48	46	347	62	2012-08-02 05:19:21.204485
19117	5	1-5444	19113	46	91	266	54	2012-08-02 05:19:21.204485
19113	5	1-5500	19113	91	83	1022	78	2012-08-02 05:19:21.204485
19119	5	1-4868	19118	78	87	840	40	2012-08-02 05:19:22.795272
19120	5	1-5060	19118	87	60	392	87	2012-08-02 05:19:22.795272
19121	5	1-5457	19118	60	88	647	94	2012-08-02 05:19:22.795272
19122	5	1-5445	19118	88	72	3743	73	2012-08-02 05:19:22.795272
19118	5	1-5504	19118	72	78	1142	39	2012-08-02 05:19:22.795272
19124	5	1-4995	19123	44	10	1385	9	2012-08-02 05:19:23.601842
19125	5	1-5274	19123	10	30	2836	88	2012-08-02 05:19:23.601842
19126	5	1-5209	19123	30	52	1629	53	2012-08-02 05:19:23.601842
19127	5	1-5321	19123	52	43	2453	90	2012-08-02 05:19:23.601842
19123	5	1-5506	19123	43	44	5028	48	2012-08-02 05:19:23.601842
19129	5	1-5385	19128	93	53	922	35	2012-08-02 05:19:23.999593
19130	5	1-2752	19128	53	52	1179	53	2012-08-02 05:19:23.999593
19131	5	1-5321	19128	52	57	1757	90	2012-08-02 05:19:23.999593
19132	5	1-5217	19128	57	83	800	77	2012-08-02 05:19:23.999593
19128	5	1-5507	19128	83	93	1083	49	2012-08-02 05:19:23.999593
19134	4	1-5219	19133	91	38	5898	83	2012-08-02 05:19:23.999593
19135	4	1-2325	19133	38	33	2097	80	2012-08-02 05:19:23.999593
19136	4	1-5425	19133	33	83	3422	77	2012-08-02 05:19:23.999593
19133	4	1-5507	19133	83	91	4973	49	2012-08-02 05:19:23.999593
19138	3	1-4640	19137	30	14	6406	53	2012-08-02 05:19:25.689571
19139	3	1-5477	19137	14	94	5672	34	2012-08-02 05:19:25.689571
19137	3	1-5511	19137	94	30	8671	5	2012-08-02 05:19:25.689571
19141	2	1-4023	19140	23	47	2720	95	2012-08-02 05:19:27.656399
19140	2	1-5516	19140	47	23	5563	69	2012-08-02 05:19:27.656399
19143	3	1-5427	19142	59	35	5704	31	2012-08-02 05:19:30.064724
19142	3	1-5522	19142	94	59	800	55	2012-08-02 05:19:30.064724
19144	3	1-4178	19142	35	94	214	91	2012-08-02 05:19:30.064724
19146	3	1-5377	19145	55	69	5854	62	2012-08-02 05:19:30.064724
19147	3	1-5514	19145	69	94	1519	91	2012-08-02 05:19:30.064724
19145	3	1-5522	19145	94	55	6341	55	2012-08-02 05:19:30.064724
19149	5	1-5320	19148	54	41	1356	84	2012-08-02 05:19:30.871616
19150	5	1-5455	19148	41	73	712	7	2012-08-02 05:19:30.871616
19151	5	1-5131	19148	73	77	735	58	2012-08-02 05:19:30.871616
19152	5	1-5474	19148	77	55	2816	31	2012-08-02 05:19:30.871616
19148	5	1-5524	19148	55	54	943	82	2012-08-02 05:19:30.871616
19154	4	1-5320	19153	54	47	2340	84	2012-08-02 05:19:30.871616
19155	4	1-5503	19153	47	23	1766	35	2012-08-02 05:19:30.871616
19156	4	1-2163	19153	23	55	2609	31	2012-08-02 05:19:30.871616
19153	4	1-5524	19153	55	54	1192	82	2012-08-02 05:19:30.871616
19157	1	1-2942	\N	46	46	1345	84	2012-08-02 05:19:30.871616
19159	5	1-4585	19158	81	99	1234	87	2012-08-02 05:19:31.258079
19160	5	1-5173	19158	99	8	789	79	2012-08-02 05:19:31.258079
19161	5	1-5371	19158	8	92	286	93	2012-08-02 05:19:31.258079
19162	5	1-5126	19158	92	55	132	56	2012-08-02 05:19:31.258079
19158	5	1-5525	19158	55	81	2312	48	2012-08-02 05:19:31.258079
19164	3	1-1668	19163	86	33	487	20	2012-08-02 05:19:31.258079
19165	3	1-5177	19163	33	55	273	56	2012-08-02 05:19:31.258079
19163	3	1-5525	19163	55	86	5694	48	2012-08-02 05:19:31.258079
19167	5	1-5457	19166	60	48	88	94	2012-08-02 05:19:32.429172
19168	5	1-4137	19166	48	39	57	45	2012-08-02 05:19:32.429172
19169	5	1-4579	19166	39	52	40	53	2012-08-02 05:19:32.429172
19170	5	1-5321	19166	52	30	59	90	2012-08-02 05:19:32.429172
19166	5	1-5528	19166	30	60	50	87	2012-08-02 05:19:32.429172
19172	5	1-5457	19171	60	39	117	94	2012-08-02 05:19:32.429172
19173	5	1-5495	19171	39	18	376	85	2012-08-02 05:19:32.429172
19174	5	1-2582	19171	18	52	54	53	2012-08-02 05:19:32.429172
19175	5	1-5321	19171	52	30	79	90	2012-08-02 05:19:32.429172
19171	5	1-5528	19171	30	60	67	87	2012-08-02 05:19:32.429172
19177	5	1-5380	19176	60	54	1793	59	2012-08-02 05:19:34.009303
19178	5	1-5365	19176	54	20	1607	4	2012-08-02 05:19:34.009303
19179	5	1-5100	19176	20	26	836	75	2012-08-02 05:19:34.009303
19180	5	1-3618	19176	26	13	1443	25	2012-08-02 05:19:34.009303
19176	5	1-5532	19176	13	60	1726	76	2012-08-02 05:19:34.009303
19182	4	1-5455	19181	41	35	445	7	2012-08-02 05:19:34.815583
19183	4	1-5091	19181	35	75	231	97	2012-08-02 05:19:34.815583
19184	4	1-5181	19181	75	25	1039	37	2012-08-02 05:19:34.815583
19181	4	1-5534	19181	25	41	723	84	2012-08-02 05:19:34.815583
19186	5	1-5437	19185	23	77	183	70	2012-08-02 05:19:34.815583
19187	5	1-4382	19185	77	49	53	99	2012-08-02 05:19:34.815583
19188	5	1-3272	19185	49	19	41	21	2012-08-02 05:19:34.815583
19189	5	1-5138	19185	19	25	314	37	2012-08-02 05:19:34.815583
19185	5	1-5534	19185	25	23	282	84	2012-08-02 05:19:34.815583
19191	4	1-2101	19190	55	93	494	65	2012-08-02 05:19:35.98668
19192	4	1-5432	19190	93	72	1488	73	2012-08-02 05:19:35.98668
19193	4	1-5504	19190	72	28	450	39	2012-08-02 05:19:35.98668
19190	4	1-5537	19190	28	55	1971	58	2012-08-02 05:19:35.98668
19195	4	1-2101	19194	55	97	497	65	2012-08-02 05:19:35.98668
19196	4	1-5415	19194	97	66	551	77	2012-08-02 05:19:35.98668
19197	4	1-5073	19194	66	28	369	39	2012-08-02 05:19:35.98668
19194	4	1-5537	19194	28	55	1791	58	2012-08-02 05:19:35.98668
19199	5	1-5335	19198	61	55	1398	69	2012-08-02 05:19:36.384778
19200	5	1-5533	19198	55	33	1030	81	2012-08-02 05:19:36.384778
19201	5	1-5387	19198	33	93	1264	65	2012-08-02 05:19:36.384778
19202	5	1-5432	19198	93	34	3753	73	2012-08-02 05:19:36.384778
19198	5	1-5538	19198	34	61	1677	89	2012-08-02 05:19:36.384778
19204	4	1-1279	19203	70	52	1354	53	2012-08-02 05:19:39.533148
19205	4	1-5321	19203	52	43	1346	90	2012-08-02 05:19:39.533148
19206	4	1-5506	19203	43	8	1820	48	2012-08-02 05:19:39.533148
19203	4	1-5546	19203	8	70	3559	69	2012-08-02 05:19:39.533148
19208	5	1-5533	19207	55	46	3205	81	2012-08-02 05:19:39.533148
19209	5	1-5336	19207	46	9	2217	30	2012-08-02 05:19:39.533148
19210	5	1-4799	19207	9	1	1457	28	2012-08-02 05:19:39.533148
19211	5	1-5434	19207	1	8	1159	48	2012-08-02 05:19:39.533148
19207	5	1-5546	19207	8	55	3679	69	2012-08-02 05:19:39.533148
19213	4	1-2104	19212	79	24	1069	5	2012-08-02 05:19:39.533148
19214	4	1-4618	19212	24	99	960	93	2012-08-02 05:19:39.533148
19215	4	1-2412	19212	99	8	400	48	2012-08-02 05:19:39.533148
19212	4	1-5546	19212	8	79	1290	69	2012-08-02 05:19:39.533148
19217	4	1-5478	19216	76	24	1700	5	2012-08-02 05:19:40.715348
19218	4	1-4618	19216	24	92	1112	93	2012-08-02 05:19:40.715348
19219	4	1-5126	19216	92	93	627	56	2012-08-02 05:19:40.715348
19216	4	1-5549	19216	93	76	1212	97	2012-08-02 05:19:40.715348
19221	4	1-5478	19220	76	98	1058	5	2012-08-02 05:19:40.715348
19222	4	1-5404	19220	98	4	585	6	2012-08-02 05:19:40.715348
19223	4	1-3765	19220	4	93	383	56	2012-08-02 05:19:40.715348
19220	4	1-5549	19220	93	76	747	97	2012-08-02 05:19:40.715348
19225	4	1-5184	19224	54	54	363	24	2012-08-02 05:19:40.715348
19226	4	1-3895	19224	54	33	472	20	2012-08-02 05:19:40.715348
19227	4	1-5177	19224	33	93	351	56	2012-08-02 05:19:40.715348
19224	4	1-5549	19224	93	54	877	97	2012-08-02 05:19:40.715348
19229	5	1-5478	19228	76	24	27	5	2012-08-02 05:19:40.715348
19230	5	1-4618	19228	24	10	24	93	2012-08-02 05:19:40.715348
19231	5	1-4219	19228	10	33	8	20	2012-08-02 05:19:40.715348
19232	5	1-5177	19228	33	93	6	56	2012-08-02 05:19:40.715348
19228	5	1-5549	19228	93	76	15	97	2012-08-02 05:19:40.715348
19234	5	1-4956	19233	8	65	2537	61	2012-08-02 05:19:40.715348
19235	5	1-5440	19233	65	3	986	96	2012-08-02 05:19:40.715348
19236	5	1-4409	19233	3	33	1200	20	2012-08-02 05:19:40.715348
19237	5	1-5177	19233	33	93	955	56	2012-08-02 05:19:40.715348
19233	5	1-5549	19233	93	8	2554	97	2012-08-02 05:19:40.715348
19239	5	1-4286	19238	35	44	1194	85	2012-08-02 05:19:41.135679
19240	5	1-5529	19238	44	76	1443	44	2012-08-02 05:19:41.135679
19241	5	1-775	19238	76	52	765	53	2012-08-02 05:19:41.135679
19242	5	1-5321	19238	52	56	1117	90	2012-08-02 05:19:41.135679
19238	5	1-5550	19238	56	35	1100	25	2012-08-02 05:19:41.135679
19244	5	1-4759	19243	12	44	243	85	2012-08-02 05:19:41.135679
19245	5	1-5529	19243	44	76	301	44	2012-08-02 05:19:41.135679
19246	5	1-775	19243	76	52	164	53	2012-08-02 05:19:41.135679
19247	5	1-5321	19243	52	56	245	90	2012-08-02 05:19:41.135679
19243	5	1-5550	19243	56	12	248	25	2012-08-02 05:19:41.135679
19249	3	1-5007	19248	57	77	943	65	2012-08-02 05:19:41.920024
19250	3	1-5492	19248	77	94	709	7	2012-08-02 05:19:41.920024
19248	3	1-5552	19248	94	57	4313	46	2012-08-02 05:19:41.920024
19252	3	1-1589	19251	40	84	318	52	2012-08-02 05:19:41.920024
19253	3	1-4372	19251	84	94	226	7	2012-08-02 05:19:41.920024
19251	3	1-5552	19251	94	40	1912	46	2012-08-02 05:19:41.920024
19255	3	1-5476	19254	52	36	5718	55	2012-08-02 05:19:44.350769
19256	3	1-3670	19254	36	35	2183	11	2012-08-02 05:19:44.350769
19254	3	1-5558	19254	35	52	5925	64	2012-08-02 05:19:44.350769
19258	5	1-3677	19257	8	35	832	25	2012-08-02 05:19:45.543475
19259	5	1-5446	19257	35	23	1276	92	2012-08-02 05:19:45.543475
19260	5	1-3932	19257	23	23	773	69	2012-08-02 05:19:45.543475
19261	5	1-5152	19257	23	18	719	36	2012-08-02 05:19:45.543475
19257	5	1-5561	19257	18	8	748	18	2012-08-02 05:19:45.543475
19263	4	1-5401	19262	97	57	338	8	2012-08-02 05:19:46.363286
19264	4	1-2876	19262	57	68	202	70	2012-08-02 05:19:46.363286
19265	4	1-3001	19262	68	76	263	88	2012-08-02 05:19:46.363286
19262	4	1-5563	19262	76	97	465	91	2012-08-02 05:19:46.363286
19267	5	1-4839	19266	21	61	224	89	2012-08-02 05:19:46.363286
19268	5	1-5335	19266	61	69	203	69	2012-08-02 05:19:46.363286
19269	5	1-2875	19266	69	10	82	9	2012-08-02 05:19:46.363286
19270	5	1-5274	19266	10	76	165	88	2012-08-02 05:19:46.363286
19266	5	1-5563	19266	76	21	298	91	2012-08-02 05:19:46.363286
19272	4	1-3600	19271	3	90	284	19	2012-08-02 05:19:46.363286
19273	4	1-4187	19271	90	70	220	26	2012-08-02 05:19:46.363286
19274	4	1-5121	19271	70	76	279	88	2012-08-02 05:19:46.363286
19271	4	1-5563	19271	76	3	543	91	2012-08-02 05:19:46.363286
19276	5	1-5266	19275	84	74	4609	42	2012-08-02 05:19:46.363286
19277	5	1-5402	19275	74	90	3235	18	2012-08-02 05:19:46.363286
19278	5	1-4695	19275	90	78	3631	11	2012-08-02 05:19:46.363286
19279	5	1-5299	19275	78	76	3258	88	2012-08-02 05:19:46.363286
19275	5	1-5563	19275	76	84	6849	91	2012-08-02 05:19:46.363286
19281	2	1-4063	19280	21	49	1463	15	2012-08-02 05:19:48.363219
19280	2	1-5568	19280	49	21	822	24	2012-08-02 05:19:48.363219
19283	3	1-4001	19282	57	44	5151	85	2012-08-02 05:19:49.158616
19284	3	1-5529	19282	44	8	3878	44	2012-08-02 05:19:49.158616
19282	3	1-5570	19282	8	57	9457	19	2012-08-02 05:19:49.158616
19286	3	1-4818	19285	33	39	4029	94	2012-08-02 05:19:50.340398
19287	3	1-5495	19285	39	85	8286	85	2012-08-02 05:19:50.340398
19285	3	1-5573	19285	85	33	6322	29	2012-08-02 05:19:50.340398
19289	2	1-4783	19288	78	85	2248	85	2012-08-02 05:19:50.340398
19288	2	1-5573	19288	85	78	2855	29	2012-08-02 05:19:50.340398
19291	5	1-5258	19290	48	100	1774	72	2012-08-02 05:19:51.522708
19292	5	1-5021	19290	100	65	1135	61	2012-08-02 05:19:51.522708
19293	5	1-5440	19290	65	94	368	96	2012-08-02 05:19:51.522708
19294	5	1-5338	19290	94	31	583	12	2012-08-02 05:19:51.522708
19290	5	1-5576	19290	31	48	2436	60	2012-08-02 05:19:51.522708
19296	5	1-4735	19295	62	8	281	97	2012-08-02 05:19:51.522708
19297	5	1-4956	19295	8	65	243	61	2012-08-02 05:19:51.522708
19298	5	1-5440	19295	65	94	82	96	2012-08-02 05:19:51.522708
19299	5	1-5338	19295	94	31	136	12	2012-08-02 05:19:51.522708
19295	5	1-5576	19295	31	62	595	60	2012-08-02 05:19:51.522708
19301	5	1-5248	19300	88	60	958	76	2012-08-02 05:19:51.522708
19302	5	1-5380	19300	60	61	933	59	2012-08-02 05:19:51.522708
19303	5	1-4648	19300	61	62	286	20	2012-08-02 05:19:51.522708
19304	5	1-5458	19300	62	31	276	12	2012-08-02 05:19:51.522708
19300	5	1-5576	19300	31	88	1267	60	2012-08-02 05:19:51.522708
19306	5	1-5156	19305	11	22	13	68	2012-08-02 05:19:51.522708
19307	5	1-5220	19305	22	88	34	14	2012-08-02 05:19:51.522708
19308	5	1-5559	19305	88	84	14	23	2012-08-02 05:19:51.522708
19309	5	1-3038	19305	84	31	5	12	2012-08-02 05:19:51.522708
19305	5	1-5576	19305	31	11	24	60	2012-08-02 05:19:51.522708
19311	5	1-4735	19310	62	88	347	97	2012-08-02 05:19:51.522708
19312	5	1-5548	19310	88	72	153	3	2012-08-02 05:19:51.522708
19313	5	1-4929	19310	72	62	125	20	2012-08-02 05:19:51.522708
19314	5	1-5458	19310	62	31	132	12	2012-08-02 05:19:51.522708
19310	5	1-5576	19310	31	62	638	60	2012-08-02 05:19:51.522708
19316	5	1-5258	19315	48	82	1818	72	2012-08-02 05:19:51.522708
19317	5	1-4641	19315	82	98	1189	1	2012-08-02 05:19:51.522708
19318	5	1-4998	19315	98	84	1155	23	2012-08-02 05:19:51.522708
19319	5	1-3038	19315	84	31	421	12	2012-08-02 05:19:51.522708
19315	5	1-5576	19315	31	48	2095	60	2012-08-02 05:19:51.522708
19321	5	1-5402	19320	74	90	3089	18	2012-08-02 05:19:53.478609
19322	5	1-4695	19320	90	75	3218	11	2012-08-02 05:19:53.478609
19323	5	1-5520	19320	75	46	4280	73	2012-08-02 05:19:53.478609
19324	5	1-4859	19320	46	26	5201	50	2012-08-02 05:19:53.478609
19320	5	1-5581	19320	26	74	4741	42	2012-08-02 05:19:53.478609
19326	3	1-4023	19325	23	78	392	95	2012-08-02 05:19:53.876387
19327	3	1-5124	19325	78	12	680	50	2012-08-02 05:19:53.876387
19325	3	1-5582	19325	12	23	899	69	2012-08-02 05:19:53.876387
19329	4	1-5533	19328	55	33	423	81	2012-08-02 05:19:53.876387
19330	4	1-5387	19328	33	9	501	65	2012-08-02 05:19:53.876387
19331	4	1-4895	19328	9	12	380	50	2012-08-02 05:19:53.876387
19328	4	1-5582	19328	12	55	596	69	2012-08-02 05:19:53.876387
19333	4	1-5152	19332	23	18	851	36	2012-08-02 05:19:53.876387
19334	4	1-5561	19332	18	41	732	18	2012-08-02 05:19:53.876387
19335	4	1-4438	19332	41	12	729	50	2012-08-02 05:19:53.876387
19332	4	1-5582	19332	12	23	1149	69	2012-08-02 05:19:53.876387
19337	4	1-4803	19336	76	94	4744	85	2012-08-02 05:19:53.876387
19338	4	1-4943	19336	94	46	3144	73	2012-08-02 05:19:53.876387
19339	4	1-4859	19336	46	12	3420	50	2012-08-02 05:19:53.876387
19336	4	1-5582	19336	12	76	5831	69	2012-08-02 05:19:53.876387
19341	5	1-4764	19340	39	49	3661	15	2012-08-02 05:19:55.11399
19342	5	1-5568	19340	49	15	2217	24	2012-08-02 05:19:55.11399
19343	5	1-3739	19340	15	25	1383	37	2012-08-02 05:19:55.11399
19344	5	1-5534	19340	25	84	1268	84	2012-08-02 05:19:55.11399
19340	5	1-5585	19340	84	39	1232	94	2012-08-02 05:19:55.11399
19346	3	1-5274	19345	10	76	72	88	2012-08-02 05:19:56.704835
19347	3	1-5563	19345	76	28	88	91	2012-08-02 05:19:56.704835
19345	3	1-5589	19345	28	10	53	9	2012-08-02 05:19:56.704835
19349	3	1-3932	19348	23	69	1988	69	2012-08-02 05:19:57.102693
19350	3	1-2875	19348	69	88	672	9	2012-08-02 05:19:57.102693
19348	3	1-5590	19348	88	23	4376	92	2012-08-02 05:19:57.102693
19352	5	1-3824	19351	27	88	925	81	2012-08-02 05:19:57.102693
19353	5	1-5466	19351	88	88	3529	14	2012-08-02 05:19:57.102693
19354	5	1-5559	19351	88	22	1189	23	2012-08-02 05:19:57.102693
19355	5	1-4211	19351	22	88	309	9	2012-08-02 05:19:57.102693
19351	5	1-5590	19351	88	27	2254	92	2012-08-02 05:19:57.102693
19356	1	1-3824	\N	27	27	527	81	2012-08-02 05:19:57.102693
19358	5	1-4945	19357	54	5	276	26	2012-08-02 05:19:57.909205
19359	5	1-5572	19357	5	31	1186	13	2012-08-02 05:19:57.909205
19360	5	1-5216	19357	31	25	540	99	2012-08-02 05:19:57.909205
19361	5	1-5366	19357	25	100	649	6	2012-08-02 05:19:57.909205
19357	5	1-5592	19357	100	54	1656	57	2012-08-02 05:19:57.909205
19363	4	1-4818	19362	33	69	66	94	2012-08-02 05:19:59.102593
19364	4	1-4485	19362	69	3	101	83	2012-08-02 05:19:59.102593
19365	4	1-5115	19362	3	79	90	43	2012-08-02 05:19:59.102593
19362	4	1-5595	19362	79	33	69	29	2012-08-02 05:19:59.102593
19367	3	1-5448	19366	24	30	3461	16	2012-08-02 05:19:59.102593
19368	3	1-5136	19366	30	79	1944	43	2012-08-02 05:19:59.102593
19366	3	1-5595	19366	79	24	1618	29	2012-08-02 05:19:59.102593
19370	4	1-4818	19369	33	55	1275	94	2012-08-02 05:19:59.102593
19371	4	1-4104	19369	55	100	932	95	2012-08-02 05:19:59.102593
19372	4	1-5531	19369	100	79	1283	43	2012-08-02 05:19:59.102593
19369	4	1-5595	19369	79	33	1146	29	2012-08-02 05:19:59.102593
19374	5	1-4146	19373	48	50	188	28	2012-08-02 05:20:00.295661
19375	5	1-5170	19373	50	59	147	95	2012-08-02 05:20:00.295661
19376	5	1-4882	19373	59	44	152	62	2012-08-02 05:20:00.295661
19377	5	1-5456	19373	44	2	208	80	2012-08-02 05:20:00.295661
19373	5	1-5598	19373	2	48	425	48	2012-08-02 05:20:00.295661
19379	4	1-4146	19378	48	50	757	28	2012-08-02 05:20:00.295661
19380	4	1-5170	19378	50	44	685	95	2012-08-02 05:20:00.295661
19381	4	1-4788	19378	44	2	622	80	2012-08-02 05:20:00.295661
19378	4	1-5598	19378	2	48	1473	48	2012-08-02 05:20:00.295661
19383	4	1-2060	19382	88	28	651	64	2012-08-02 05:20:00.715708
19384	4	1-5390	19382	28	9	990	65	2012-08-02 05:20:00.715708
19385	4	1-4895	19382	9	93	658	50	2012-08-02 05:20:00.715708
19382	4	1-5599	19382	93	88	2185	26	2012-08-02 05:20:00.715708
19387	4	1-3009	19386	25	48	292	95	2012-08-02 05:20:00.715708
19388	4	1-5034	19386	48	31	106	89	2012-08-02 05:20:00.715708
19389	4	1-3738	19386	31	93	102	50	2012-08-02 05:20:00.715708
19386	4	1-5599	19386	93	25	368	26	2012-08-02 05:20:00.715708
19391	5	1-3121	19390	67	28	772	39	2012-08-02 05:20:01.12449
19392	5	1-5537	19390	28	11	4061	58	2012-08-02 05:20:01.12449
19393	5	1-3790	19390	11	8	588	18	2012-08-02 05:20:01.12449
19394	5	1-3677	19390	8	95	679	25	2012-08-02 05:20:01.12449
19390	5	1-5600	19390	95	67	660	91	2012-08-02 05:20:01.12449
19396	5	1-5359	19395	35	56	1313	90	2012-08-02 05:20:01.919853
19397	5	1-5550	19395	56	27	1418	25	2012-08-02 05:20:01.919853
19398	5	1-4564	19395	27	93	1154	59	2012-08-02 05:20:01.919853
19399	5	1-5241	19395	93	24	950	30	2012-08-02 05:20:01.919853
19395	5	1-5602	19395	24	35	1260	60	2012-08-02 05:20:01.919853
19401	5	1-4580	19400	76	58	1151	16	2012-08-02 05:20:01.919853
19402	5	1-5510	19400	58	60	1200	76	2012-08-02 05:20:01.919853
19403	5	1-5380	19400	60	93	1266	59	2012-08-02 05:20:01.919853
19404	5	1-5241	19400	93	24	1065	30	2012-08-02 05:20:01.919853
19400	5	1-5602	19400	24	76	1444	60	2012-08-02 05:20:01.919853
19406	5	1-5177	19405	33	41	860	56	2012-08-02 05:20:02.317621
19407	5	1-5212	19405	41	3	1119	51	2012-08-02 05:20:02.317621
19408	5	1-5277	19405	3	41	1178	100	2012-08-02 05:20:02.317621
19409	5	1-5535	19405	41	71	1844	34	2012-08-02 05:20:02.317621
19405	5	1-5603	19405	71	33	1293	20	2012-08-02 05:20:02.317621
19411	4	1-5177	19410	33	41	916	56	2012-08-02 05:20:02.317621
19412	4	1-5212	19410	41	22	1215	51	2012-08-02 05:20:02.317621
19413	4	1-5132	19410	22	71	1514	34	2012-08-02 05:20:02.317621
19410	4	1-5603	19410	71	33	1209	20	2012-08-02 05:20:02.317621
19415	5	1-5562	19414	20	98	400	18	2012-08-02 05:20:02.317621
19416	5	1-5057	19414	98	5	409	30	2012-08-02 05:20:02.317621
19417	5	1-2932	19414	5	41	374	100	2012-08-02 05:20:02.317621
19418	5	1-5535	19414	41	71	636	34	2012-08-02 05:20:02.317621
19414	5	1-5603	19414	71	20	542	20	2012-08-02 05:20:02.317621
19420	4	1-2519	19419	14	88	212	81	2012-08-02 05:20:02.726631
19421	4	1-5466	19419	88	88	985	14	2012-08-02 05:20:02.726631
19422	4	1-5559	19419	88	69	405	23	2012-08-02 05:20:02.726631
19419	4	1-5604	19419	69	14	542	9	2012-08-02 05:20:02.726631
19424	5	1-5560	19423	75	20	8744	20	2012-08-02 05:20:03.919805
19425	5	1-5562	19423	20	73	4367	18	2012-08-02 05:20:03.919805
19426	5	1-4149	19423	73	38	1230	19	2012-08-02 05:20:03.919805
19427	5	1-3286	19423	38	84	666	45	2012-08-02 05:20:03.919805
19423	5	1-5607	19423	84	75	5561	35	2012-08-02 05:20:03.919805
19429	5	1-5097	19428	77	76	852	61	2012-08-02 05:20:03.919805
19430	5	1-5340	19428	76	12	526	50	2012-08-02 05:20:03.919805
19431	5	1-5582	19428	12	12	897	69	2012-08-02 05:20:03.919805
19432	5	1-2391	19428	12	84	181	45	2012-08-02 05:20:03.919805
19428	5	1-5607	19428	84	77	1585	35	2012-08-02 05:20:03.919805
19434	3	1-3252	19433	48	91	1132	40	2012-08-02 05:20:03.919805
19435	3	1-4100	19433	91	84	153	45	2012-08-02 05:20:03.919805
19433	3	1-5607	19433	84	48	1469	35	2012-08-02 05:20:03.919805
19437	5	1-5307	19436	24	89	1471	38	2012-08-02 05:20:04.748447
19438	5	1-5162	19436	89	18	1195	36	2012-08-02 05:20:04.748447
19439	5	1-5561	19436	18	98	1172	18	2012-08-02 05:20:04.748447
19440	5	1-5057	19436	98	65	1078	30	2012-08-02 05:20:04.748447
19436	5	1-5609	19436	65	24	1998	33	2012-08-02 05:20:04.748447
19442	4	1-4622	19441	67	9	145	65	2012-08-02 05:20:05.146721
19443	4	1-4895	19441	9	12	102	50	2012-08-02 05:20:05.146721
19444	4	1-5582	19441	12	72	148	69	2012-08-02 05:20:05.146721
19441	4	1-5610	19441	72	67	136	75	2012-08-02 05:20:05.146721
19446	5	1-5275	19445	54	75	1679	11	2012-08-02 05:20:05.146721
19447	5	1-5520	19445	75	34	2156	73	2012-08-02 05:20:05.146721
19448	5	1-5538	19445	34	61	1090	89	2012-08-02 05:20:05.146721
19449	5	1-5335	19445	61	72	1029	69	2012-08-02 05:20:05.146721
19445	5	1-5610	19445	72	54	1190	75	2012-08-02 05:20:05.146721
19451	5	1-5526	19450	15	61	7063	93	2012-08-02 05:20:05.146721
19452	5	1-5459	19450	61	48	6406	95	2012-08-02 05:20:05.146721
19453	5	1-5034	19450	48	61	2924	89	2012-08-02 05:20:05.146721
19454	5	1-5335	19450	61	72	2951	69	2012-08-02 05:20:05.146721
19450	5	1-5610	19450	72	15	3495	75	2012-08-02 05:20:05.146721
19456	5	1-5492	19455	77	66	1372	7	2012-08-02 05:20:06.772059
19457	5	1-4830	19455	66	100	296	64	2012-08-02 05:20:06.772059
19458	5	1-4881	19455	100	96	573	10	2012-08-02 05:20:06.772059
19459	5	1-5294	19455	96	100	561	76	2012-08-02 05:20:06.772059
19455	5	1-5614	19455	100	77	1569	65	2012-08-02 05:20:06.772059
19461	5	1-5415	19460	97	100	1324	77	2012-08-02 05:20:06.772059
19462	5	1-5352	19460	100	71	1223	67	2012-08-02 05:20:06.772059
19463	5	1-1625	19460	71	96	417	10	2012-08-02 05:20:06.772059
19464	5	1-5294	19460	96	100	457	76	2012-08-02 05:20:06.772059
19460	5	1-5614	19460	100	97	1351	65	2012-08-02 05:20:06.772059
19466	4	1-4895	19465	9	26	5086	50	2012-08-02 05:20:06.772059
19467	4	1-5581	19465	26	26	4711	42	2012-08-02 05:20:06.772059
19468	4	1-4194	19465	26	100	1941	76	2012-08-02 05:20:06.772059
19465	4	1-5614	19465	100	9	5758	65	2012-08-02 05:20:06.772059
19470	4	1-4844	19469	74	30	164	90	2012-08-02 05:20:06.772059
19471	4	1-5528	19469	30	68	145	87	2012-08-02 05:20:06.772059
19472	4	1-5238	19469	68	100	141	76	2012-08-02 05:20:06.772059
19469	4	1-5614	19469	100	74	422	65	2012-08-02 05:20:06.772059
19474	4	1-2067	19473	64	100	49	76	2012-08-02 05:20:10.452713
19475	4	1-5614	19473	100	91	117	65	2012-08-02 05:20:10.452713
19476	4	1-4433	19473	91	39	65	49	2012-08-02 05:20:10.452713
19473	4	1-5623	19473	39	64	321	66	2012-08-02 05:20:10.452713
19478	5	1-969	19477	75	83	1460	51	2012-08-02 05:20:10.452713
19479	5	1-5575	19477	83	29	1394	74	2012-08-02 05:20:10.452713
19480	5	1-5187	19477	29	31	1200	22	2012-08-02 05:20:10.452713
19481	5	1-5566	19477	31	39	860	49	2012-08-02 05:20:10.452713
19477	5	1-5623	19477	39	75	4644	66	2012-08-02 05:20:10.452713
19483	5	1-5402	19482	74	24	644	18	2012-08-02 05:20:11.281546
19484	5	1-5346	19482	24	58	983	16	2012-08-02 05:20:11.281546
19485	5	1-5510	19482	58	60	916	76	2012-08-02 05:20:11.281546
19486	5	1-5380	19482	60	6	864	59	2012-08-02 05:20:11.281546
19482	5	1-5625	19482	6	74	1079	42	2012-08-02 05:20:11.281546
19488	5	1-5624	19487	15	60	4235	85	2012-08-02 05:20:12.938805
19489	5	1-4982	19487	60	25	4324	37	2012-08-02 05:20:12.938805
19490	5	1-5534	19487	25	84	3651	84	2012-08-02 05:20:12.938805
19491	5	1-5585	19487	84	25	2966	94	2012-08-02 05:20:12.938805
19487	5	1-5629	19487	25	15	1309	1	2012-08-02 05:20:12.938805
19493	5	1-5575	19492	83	87	2411	74	2012-08-02 05:20:13.347451
19494	5	1-5523	19492	87	43	1801	72	2012-08-02 05:20:13.347451
19495	5	1-1718	19492	43	100	308	6	2012-08-02 05:20:13.347451
19496	5	1-5592	19492	100	84	598	57	2012-08-02 05:20:13.347451
19492	5	1-5630	19492	84	83	2877	51	2012-08-02 05:20:13.347451
19498	4	1-2952	19497	31	36	908	78	2012-08-02 05:20:13.347451
19499	4	1-5554	19497	36	33	479	23	2012-08-02 05:20:13.347451
19500	4	1-5012	19497	33	84	394	57	2012-08-02 05:20:13.347451
19497	4	1-5630	19497	84	31	1962	51	2012-08-02 05:20:13.347451
19502	4	1-5575	19501	83	52	401	74	2012-08-02 05:20:13.347451
19503	4	1-5083	19501	52	34	212	59	2012-08-02 05:20:13.347451
19504	4	1-2123	19501	34	84	90	57	2012-08-02 05:20:13.347451
19501	4	1-5630	19501	84	83	455	51	2012-08-02 05:20:13.347451
19506	4	1-3809	19505	76	71	882	87	2012-08-02 05:20:13.347451
19507	4	1-4784	19505	71	63	354	46	2012-08-02 05:20:13.347451
19508	4	1-5617	19505	63	84	427	57	2012-08-02 05:20:13.347451
19505	4	1-5630	19505	84	76	2707	51	2012-08-02 05:20:13.347451
19510	4	1-5627	19509	64	97	4069	54	2012-08-02 05:20:13.778682
19511	4	1-4463	19509	97	52	999	17	2012-08-02 05:20:13.778682
19512	4	1-5490	19509	52	97	1452	66	2012-08-02 05:20:13.778682
19509	4	1-5631	19509	97	64	550	18	2012-08-02 05:20:13.778682
19514	4	1-4392	19513	2	69	54	23	2012-08-02 05:20:14.176531
19515	4	1-5604	19513	69	38	63	9	2012-08-02 05:20:14.176531
19516	4	1-4810	19513	38	68	127	49	2012-08-02 05:20:14.176531
19513	4	1-5632	19513	68	2	100	33	2012-08-02 05:20:14.176531
19518	4	1-3524	19517	51	69	959	23	2012-08-02 05:20:14.176531
19519	4	1-5604	19517	69	38	1240	9	2012-08-02 05:20:14.176531
19520	4	1-4810	19517	38	68	2748	49	2012-08-02 05:20:14.176531
19517	4	1-5632	19517	68	51	2382	33	2012-08-02 05:20:14.176531
19522	5	1-4445	19521	48	57	442	62	2012-08-02 05:20:14.58528
19523	5	1-4908	19521	57	18	498	36	2012-08-02 05:20:14.58528
19524	5	1-5561	19521	18	64	530	18	2012-08-02 05:20:14.58528
19525	5	1-5627	19521	64	35	4146	54	2012-08-02 05:20:14.58528
19521	5	1-5633	19521	35	48	1314	16	2012-08-02 05:20:14.58528
19527	4	1-5375	19526	69	79	1160	43	2012-08-02 05:20:16.574087
19528	4	1-5595	19526	79	49	1035	29	2012-08-02 05:20:16.574087
19529	4	1-4553	19526	49	45	646	97	2012-08-02 05:20:16.574087
19526	4	1-5638	19526	45	69	706	81	2012-08-02 05:20:16.574087
19531	4	1-5345	19530	79	35	5797	8	2012-08-02 05:20:18.164651
19532	4	1-43	19530	35	88	3897	40	2012-08-02 05:20:18.164651
19533	4	1-5393	19530	88	12	3327	15	2012-08-02 05:20:18.164651
19530	4	1-5642	19530	12	79	3114	98	2012-08-02 05:20:18.164651
19535	4	1-4295	19534	62	6	931	1	2012-08-02 05:20:18.164651
19536	4	1-3945	19534	6	88	1177	40	2012-08-02 05:20:18.164651
19537	4	1-5393	19534	88	12	1015	15	2012-08-02 05:20:18.164651
19534	4	1-5642	19534	12	62	960	98	2012-08-02 05:20:18.164651
19539	5	1-5276	19538	51	36	2956	88	2012-08-02 05:20:18.562542
19540	5	1-5381	19538	36	90	3673	2	2012-08-02 05:20:18.562542
19541	5	1-5430	19538	90	5	2476	71	2012-08-02 05:20:18.562542
19542	5	1-5577	19538	5	82	5444	70	2012-08-02 05:20:18.562542
19538	5	1-5643	19538	82	51	4799	41	2012-08-02 05:20:18.562542
19544	3	1-4744	19543	63	84	2902	84	2012-08-02 05:20:19.40235
19545	3	1-5585	19543	84	20	2230	94	2012-08-02 05:20:19.40235
19543	3	1-5645	19543	20	63	8941	63	2012-08-02 05:20:19.40235
19547	5	1-5610	19546	72	75	912	75	2012-08-02 05:20:19.788803
19548	5	1-5229	19546	75	15	431	1	2012-08-02 05:20:19.788803
19549	5	1-5624	19546	15	94	1147	85	2012-08-02 05:20:19.788803
19550	5	1-4943	19546	94	97	622	73	2012-08-02 05:20:19.788803
19546	5	1-5646	19546	97	72	1040	69	2012-08-02 05:20:19.788803
19552	3	1-4803	19551	76	94	2729	85	2012-08-02 05:20:19.788803
19553	3	1-4943	19551	94	97	1751	73	2012-08-02 05:20:19.788803
19551	3	1-5646	19551	97	76	3466	69	2012-08-02 05:20:19.788803
19555	5	1-2104	19554	79	30	348	5	2012-08-02 05:20:19.788803
19556	5	1-4640	19554	30	34	307	53	2012-08-02 05:20:19.788803
19557	5	1-5243	19554	34	6	188	60	2012-08-02 05:20:19.788803
19558	5	1-5290	19554	6	97	188	73	2012-08-02 05:20:19.788803
19554	5	1-5646	19554	97	79	430	69	2012-08-02 05:20:19.788803
19560	5	1-1279	19559	70	34	1680	53	2012-08-02 05:20:19.788803
19561	5	1-5243	19559	34	93	1036	60	2012-08-02 05:20:19.788803
19562	5	1-5596	19559	93	94	1601	85	2012-08-02 05:20:19.788803
19563	5	1-4943	19559	94	97	1191	73	2012-08-02 05:20:19.788803
19559	5	1-5646	19559	97	70	2733	69	2012-08-02 05:20:19.788803
19565	4	1-1279	19564	70	34	314	53	2012-08-02 05:20:19.788803
19566	4	1-5243	19564	34	6	200	60	2012-08-02 05:20:19.788803
19567	4	1-5290	19564	6	97	208	73	2012-08-02 05:20:19.788803
19564	4	1-5646	19564	97	70	494	69	2012-08-02 05:20:19.788803
19569	4	1-5562	19568	20	64	134	18	2012-08-02 05:20:20.628695
19570	4	1-5627	19568	64	34	1024	54	2012-08-02 05:20:20.628695
19571	4	1-2335	19568	34	7	286	52	2012-08-02 05:20:20.628695
19568	4	1-5648	19568	7	20	189	20	2012-08-02 05:20:20.628695
19573	5	1-4998	19572	98	69	1879	23	2012-08-02 05:20:21.04861
19574	5	1-5604	19572	69	31	2297	9	2012-08-02 05:20:21.04861
19575	5	1-3145	19572	31	69	1639	33	2012-08-02 05:20:21.04861
19576	5	1-5127	19572	69	82	1121	7	2012-08-02 05:20:21.04861
19572	5	1-5649	19572	82	98	2329	1	2012-08-02 05:20:21.04861
19578	4	1-5285	19577	29	92	36	27	2012-08-02 05:20:21.04861
19579	4	1-5362	19577	92	53	63	79	2012-08-02 05:20:21.04861
19580	4	1-2946	19577	53	82	22	7	2012-08-02 05:20:21.04861
19577	4	1-5649	19577	82	29	50	1	2012-08-02 05:20:21.04861
19582	5	1-5303	19581	28	48	179	61	2012-08-02 05:20:21.04861
19583	5	1-5584	19581	48	55	48	47	2012-08-02 05:20:21.04861
19584	5	1-4565	19581	55	84	79	52	2012-08-02 05:20:21.04861
19585	5	1-4372	19581	84	82	63	7	2012-08-02 05:20:21.04861
19581	5	1-5649	19581	82	28	157	1	2012-08-02 05:20:21.04861
19587	5	1-5303	19586	28	48	303	61	2012-08-02 05:20:21.04861
19588	5	1-5584	19586	48	55	81	47	2012-08-02 05:20:21.04861
19589	5	1-4565	19586	55	2	136	52	2012-08-02 05:20:21.04861
19590	5	1-5423	19586	2	82	105	7	2012-08-02 05:20:21.04861
19586	5	1-5649	19586	82	28	263	1	2012-08-02 05:20:21.04861
19592	5	1-3019	19591	13	100	284	64	2012-08-02 05:20:23.015189
19593	5	1-4881	19591	100	82	530	10	2012-08-02 05:20:23.015189
19594	5	1-5452	19591	82	22	536	68	2012-08-02 05:20:23.015189
19595	5	1-5220	19591	22	76	1202	14	2012-08-02 05:20:23.015189
19591	5	1-5654	19591	76	13	2177	95	2012-08-02 05:20:23.015189
19597	3	1-4788	19596	44	80	1670	80	2012-08-02 05:20:23.015189
19598	3	1-4060	19596	80	76	1088	14	2012-08-02 05:20:23.015189
19596	3	1-5654	19596	76	44	2022	95	2012-08-02 05:20:23.015189
19600	4	1-3019	19599	13	52	285	64	2012-08-02 05:20:23.015189
19601	4	1-5476	19599	52	51	499	55	2012-08-02 05:20:23.015189
19602	4	1-5130	19599	51	76	1036	14	2012-08-02 05:20:23.015189
19599	4	1-5654	19599	76	13	2025	95	2012-08-02 05:20:23.015189
19604	5	1-5151	19603	50	76	752	14	2012-08-02 05:20:23.810941
19605	5	1-5654	19603	76	59	1560	95	2012-08-02 05:20:23.810941
19606	5	1-4882	19603	59	47	1909	62	2012-08-02 05:20:23.810941
19607	5	1-5567	19603	47	80	2053	54	2012-08-02 05:20:23.810941
19603	5	1-5656	19603	80	50	263	19	2012-08-02 05:20:23.810941
19609	5	1-5649	19608	82	52	3567	1	2012-08-02 05:20:24.617596
19610	5	1-5240	19608	52	31	2462	4	2012-08-02 05:20:24.617596
19611	5	1-5041	19608	31	57	1024	90	2012-08-02 05:20:24.617596
19612	5	1-5217	19608	57	43	361	77	2012-08-02 05:20:24.617596
19608	5	1-5658	19608	43	82	2026	7	2012-08-02 05:20:24.617596
19614	5	1-5649	19613	82	97	925	1	2012-08-02 05:20:24.617596
19615	5	1-5254	19613	97	42	550	38	2012-08-02 05:20:24.617596
19616	5	1-4896	19613	42	57	203	90	2012-08-02 05:20:24.617596
19617	5	1-5217	19613	57	43	78	77	2012-08-02 05:20:24.617596
19613	5	1-5658	19613	43	82	481	7	2012-08-02 05:20:24.617596
19619	5	1-4319	19618	28	96	2609	39	2012-08-02 05:20:24.617596
19620	5	1-5605	19618	96	35	2161	60	2012-08-02 05:20:24.617596
19621	5	1-5359	19618	35	57	2009	90	2012-08-02 05:20:24.617596
19622	5	1-5217	19618	57	43	878	77	2012-08-02 05:20:24.617596
19618	5	1-5658	19618	43	28	6117	7	2012-08-02 05:20:24.617596
19624	5	1-2104	19623	79	48	1151	5	2012-08-02 05:20:25.412925
19625	5	1-5250	19623	48	70	490	22	2012-08-02 05:20:25.412925
19626	5	1-4258	19623	70	48	120	60	2012-08-02 05:20:25.412925
19627	5	1-5258	19623	48	55	78	72	2012-08-02 05:20:25.412925
19623	5	1-5660	19623	55	79	2619	69	2012-08-02 05:20:25.412925
19629	5	1-2875	19628	69	44	264	9	2012-08-02 05:20:25.412925
19630	5	1-3782	19628	44	96	114	39	2012-08-02 05:20:25.412925
19631	5	1-5605	19628	96	48	61	60	2012-08-02 05:20:25.412925
19632	5	1-5258	19628	48	55	30	72	2012-08-02 05:20:25.412925
19628	5	1-5660	19628	55	69	1056	69	2012-08-02 05:20:25.412925
19634	3	1-3738	19633	31	27	5101	50	2012-08-02 05:20:28.208126
19635	3	1-5616	19633	27	26	8124	83	2012-08-02 05:20:28.208126
19633	3	1-5667	19633	26	31	7450	89	2012-08-02 05:20:28.208126
19637	5	1-5561	19636	18	24	2157	18	2012-08-02 05:20:29.434492
19638	5	1-5346	19636	24	72	3001	16	2012-08-02 05:20:29.434492
19639	5	1-5583	19636	72	82	1871	10	2012-08-02 05:20:29.434492
19640	5	1-5452	19636	82	44	1814	68	2012-08-02 05:20:29.434492
19636	5	1-5670	19636	44	18	2485	36	2012-08-02 05:20:29.434492
19642	5	1-5561	19641	18	72	817	18	2012-08-02 05:20:29.434492
19643	5	1-2241	19641	72	72	762	16	2012-08-02 05:20:29.434492
19644	5	1-5583	19641	72	82	561	10	2012-08-02 05:20:29.434492
19645	5	1-5452	19641	82	44	588	68	2012-08-02 05:20:29.434492
19641	5	1-5670	19641	44	18	871	36	2012-08-02 05:20:29.434492
19647	3	1-4032	19646	35	45	1319	88	2012-08-02 05:20:29.434492
19648	3	1-5438	19646	45	44	1104	68	2012-08-02 05:20:29.434492
19646	3	1-5670	19646	44	35	1918	36	2012-08-02 05:20:29.434492
19650	2	1-4352	19649	60	64	5036	50	2012-08-02 05:20:30.726926
19649	2	1-5673	19649	64	60	2141	96	2012-08-02 05:20:30.726926
19652	4	1-5338	19651	94	51	878	12	2012-08-02 05:20:30.726926
19653	4	1-5206	19651	51	31	805	89	2012-08-02 05:20:30.726926
19654	4	1-3738	19651	31	64	1004	50	2012-08-02 05:20:30.726926
19651	4	1-5673	19651	64	94	464	96	2012-08-02 05:20:30.726926
19656	5	1-5180	19655	6	40	305	67	2012-08-02 05:20:31.124843
19657	5	1-5450	19655	40	76	479	14	2012-08-02 05:20:31.124843
19658	5	1-5654	19655	76	59	708	95	2012-08-02 05:20:31.124843
19659	5	1-4882	19655	59	2	616	62	2012-08-02 05:20:31.124843
19655	5	1-5674	19655	2	6	331	41	2012-08-02 05:20:31.124843
19661	4	1-5143	19660	47	25	443	26	2012-08-02 05:20:31.124843
19662	4	1-3009	19660	25	59	475	95	2012-08-02 05:20:31.124843
19663	4	1-4882	19660	59	2	620	62	2012-08-02 05:20:31.124843
19660	4	1-5674	19660	2	47	499	41	2012-08-02 05:20:31.124843
19665	4	1-5305	19664	38	70	290	31	2012-08-02 05:20:32.296147
19666	4	1-5663	19664	70	60	356	32	2012-08-02 05:20:32.296147
19667	4	1-4068	19664	60	32	106	30	2012-08-02 05:20:32.296147
19664	4	1-5677	19664	32	38	172	38	2012-08-02 05:20:32.296147
19669	4	1-4091	19668	84	67	182	72	2012-08-02 05:20:32.296147
19670	4	1-4955	19668	67	27	54	19	2012-08-02 05:20:32.296147
19671	4	1-5639	19668	27	32	95	30	2012-08-02 05:20:32.296147
19668	4	1-5677	19668	32	84	181	38	2012-08-02 05:20:32.296147
19673	5	1-3825	19672	78	5	10	26	2012-08-02 05:20:32.296147
19674	5	1-5572	19672	5	31	45	13	2012-08-02 05:20:32.296147
19675	5	1-5216	19672	31	92	22	99	2012-08-02 05:20:32.296147
19676	5	1-3667	19672	92	32	15	30	2012-08-02 05:20:32.296147
19672	5	1-5677	19672	32	78	29	38	2012-08-02 05:20:32.296147
19678	5	1-4928	19677	27	85	63	9	2012-08-02 05:20:32.704855
19679	5	1-4297	19677	85	59	64	95	2012-08-02 05:20:32.704855
19680	5	1-4882	19677	59	57	69	62	2012-08-02 05:20:32.704855
19681	5	1-4908	19677	57	84	66	36	2012-08-02 05:20:32.704855
19677	5	1-5678	19677	84	27	78	86	2012-08-02 05:20:32.704855
19683	5	1-5185	19682	39	50	96	28	2012-08-02 05:20:32.704855
19684	5	1-5170	19682	50	59	89	95	2012-08-02 05:20:32.704855
19685	5	1-4882	19682	59	57	109	62	2012-08-02 05:20:32.704855
19686	5	1-4908	19682	57	84	121	36	2012-08-02 05:20:32.704855
19682	5	1-5678	19682	84	39	164	86	2012-08-02 05:20:32.704855
19688	5	1-4110	19687	54	50	414	92	2012-08-02 05:20:32.704855
19689	5	1-5486	19687	50	55	274	55	2012-08-02 05:20:32.704855
19690	5	1-5377	19687	55	57	290	62	2012-08-02 05:20:32.704855
19691	5	1-4908	19687	57	84	342	36	2012-08-02 05:20:32.704855
19687	5	1-5678	19687	84	54	496	86	2012-08-02 05:20:32.704855
19693	5	1-1819	19692	23	33	74	81	2012-08-02 05:20:33.909122
19694	5	1-5387	19692	33	91	81	65	2012-08-02 05:20:33.909122
19695	5	1-4433	19692	91	68	44	49	2012-08-02 05:20:33.909122
19696	5	1-5632	19692	68	66	29	33	2012-08-02 05:20:33.909122
19692	5	1-5681	19692	66	23	564	2	2012-08-02 05:20:33.909122
19698	5	1-4973	19697	48	36	957	78	2012-08-02 05:20:33.909122
19699	5	1-5554	19697	36	69	507	23	2012-08-02 05:20:33.909122
19700	5	1-5604	19697	69	31	499	9	2012-08-02 05:20:33.909122
19701	5	1-3145	19697	31	66	323	33	2012-08-02 05:20:33.909122
19697	5	1-5681	19697	66	48	6368	2	2012-08-02 05:20:33.909122
19703	5	1-4983	19702	9	64	240	50	2012-08-02 05:20:33.909122
19704	5	1-5673	19702	64	22	91	96	2012-08-02 05:20:33.909122
19705	5	1-4673	19702	22	68	64	49	2012-08-02 05:20:33.909122
19706	5	1-5632	19702	68	66	46	33	2012-08-02 05:20:33.909122
19702	5	1-5681	19702	66	9	981	2	2012-08-02 05:20:33.909122
19707	1	1-1256	\N	2	2	621	49	2012-08-02 05:20:33.909122
19709	5	1-5598	19708	2	7	4903	48	2012-08-02 05:20:34.308821
19710	5	1-2693	19708	7	100	532	95	2012-08-02 05:20:34.308821
19711	5	1-5531	19708	100	79	626	43	2012-08-02 05:20:34.308821
19712	5	1-5595	19708	79	3	479	29	2012-08-02 05:20:34.308821
19708	5	1-5682	19708	3	2	2309	80	2012-08-02 05:20:34.308821
19714	5	1-5598	19713	2	55	2897	48	2012-08-02 05:20:34.308821
19715	5	1-2737	19713	55	42	350	86	2012-08-02 05:20:34.308821
19716	5	1-5482	19713	42	93	606	70	2012-08-02 05:20:34.308821
19717	5	1-5594	19713	93	3	242	29	2012-08-02 05:20:34.308821
19713	5	1-5682	19713	3	2	1261	80	2012-08-02 05:20:34.308821
19719	4	1-4153	19718	9	3	1290	83	2012-08-02 05:20:34.308821
19720	4	1-5115	19718	3	79	1250	43	2012-08-02 05:20:34.308821
19721	4	1-5595	19718	79	3	916	29	2012-08-02 05:20:34.308821
19718	4	1-5682	19718	3	9	4790	80	2012-08-02 05:20:34.308821
19723	4	1-5304	19722	17	72	1029	100	2012-08-02 05:20:34.308821
19724	4	1-5246	19722	72	12	413	11	2012-08-02 05:20:34.308821
19725	4	1-5547	19722	12	3	264	29	2012-08-02 05:20:34.308821
19722	4	1-5682	19722	3	17	1425	80	2012-08-02 05:20:34.308821
19727	5	1-5559	19726	88	69	2382	23	2012-08-02 05:20:34.740556
19728	5	1-5604	19726	69	45	2252	9	2012-08-02 05:20:34.740556
19729	5	1-2730	19726	45	39	1299	17	2012-08-02 05:20:34.740556
19730	5	1-5298	19726	39	8	931	27	2012-08-02 05:20:34.740556
19726	5	1-5683	19726	8	88	8314	14	2012-08-02 05:20:34.740556
19732	5	1-5651	19731	14	69	464	59	2012-08-02 05:20:34.740556
19733	5	1-5620	19731	69	26	363	28	2012-08-02 05:20:34.740556
19734	5	1-2871	19731	26	94	269	13	2012-08-02 05:20:34.740556
19735	5	1-5403	19731	94	8	162	27	2012-08-02 05:20:34.740556
19731	5	1-5683	19731	8	14	1659	14	2012-08-02 05:20:34.740556
19736	1	1-2871	\N	26	26	1146	13	2012-08-02 05:20:34.740556
19737	1	1-4211	\N	22	22	179	9	2012-08-02 05:20:34.740556
19739	4	1-5589	19738	28	10	1567	9	2012-08-02 05:20:35.138248
19740	4	1-5274	19738	10	69	2444	88	2012-08-02 05:20:35.138248
19741	4	1-5659	19738	69	83	1015	49	2012-08-02 05:20:35.138248
19738	4	1-5684	19738	83	28	2249	91	2012-08-02 05:20:35.138248
19743	5	1-5589	19742	28	31	1702	9	2012-08-02 05:20:35.138248
19744	5	1-3145	19742	31	69	1371	33	2012-08-02 05:20:35.138248
19745	5	1-5127	19742	69	84	941	7	2012-08-02 05:20:35.138248
19746	5	1-4900	19742	84	83	614	49	2012-08-02 05:20:35.138248
19742	5	1-5684	19742	83	28	1822	91	2012-08-02 05:20:35.138248
19748	3	1-1491	19747	33	84	263	7	2012-08-02 05:20:35.138248
19749	3	1-4900	19747	84	83	174	49	2012-08-02 05:20:35.138248
19747	3	1-5684	19747	83	33	525	91	2012-08-02 05:20:35.138248
19751	5	1-623	19750	3	33	531	81	2012-08-02 05:20:35.138248
19752	5	1-5387	19750	33	4	757	65	2012-08-02 05:20:35.138248
19753	5	1-5601	19750	4	69	435	88	2012-08-02 05:20:35.138248
19754	5	1-5659	19750	69	83	250	49	2012-08-02 05:20:35.138248
19750	5	1-5684	19750	83	3	764	91	2012-08-02 05:20:35.138248
19756	2	1-2262	19755	3	83	686	49	2012-08-02 05:20:35.138248
19755	2	1-5684	19755	83	3	2212	91	2012-08-02 05:20:35.138248
19757	1	1-118	\N	65	65	78	81	2012-08-02 05:20:35.138248
19758	1	1-611	\N	74	74	1455	81	2012-08-02 05:20:35.138248
19760	2	1-5350	19759	28	20	1601	68	2012-08-02 05:20:35.547306
19759	2	1-5685	19759	20	28	2764	78	2012-08-02 05:20:35.547306
19762	3	1-3405	19761	52	39	44	7	2012-08-02 05:20:35.547306
19763	3	1-5370	19761	39	20	53	68	2012-08-02 05:20:35.547306
19761	3	1-5685	19761	20	52	93	78	2012-08-02 05:20:35.547306
19765	4	1-5344	19764	83	93	37	29	2012-08-02 05:20:35.547306
19766	4	1-4909	19764	93	82	28	10	2012-08-02 05:20:35.547306
19767	4	1-5452	19764	82	20	25	68	2012-08-02 05:20:35.547306
19764	4	1-5685	19764	20	83	49	78	2012-08-02 05:20:35.547306
19769	4	1-5322	19768	53	37	1881	83	2012-08-02 05:20:35.547306
19770	4	1-4590	19768	37	39	343	7	2012-08-02 05:20:35.547306
19771	4	1-5370	19768	39	20	498	68	2012-08-02 05:20:35.547306
19768	4	1-5685	19768	20	53	1065	78	2012-08-02 05:20:35.547306
19773	5	1-5554	19772	36	69	1015	23	2012-08-02 05:20:35.547306
19774	5	1-5604	19772	69	97	1263	9	2012-08-02 05:20:35.547306
19775	5	1-3753	19772	97	82	544	10	2012-08-02 05:20:35.547306
19776	5	1-5452	19772	82	20	616	68	2012-08-02 05:20:35.547306
19772	5	1-5685	19772	20	36	1514	78	2012-08-02 05:20:35.547306
19778	4	1-4812	19777	25	32	510	30	2012-08-02 05:20:38.353369
19779	4	1-5677	19777	32	99	714	38	2012-08-02 05:20:38.353369
19780	4	1-5080	19777	99	26	562	28	2012-08-02 05:20:38.353369
19777	4	1-5692	19777	26	25	1989	93	2012-08-02 05:20:38.353369
19782	5	1-5388	19781	26	100	3182	5	2012-08-02 05:20:38.353369
19783	5	1-4101	19781	100	60	2082	76	2012-08-02 05:20:38.353369
19784	5	1-5380	19781	60	69	1614	59	2012-08-02 05:20:38.353369
19785	5	1-5620	19781	69	26	1121	28	2012-08-02 05:20:38.353369
19781	5	1-5692	19781	26	26	4590	93	2012-08-02 05:20:38.353369
19787	5	1-3725	19786	95	70	563	31	2012-08-02 05:20:38.353369
19788	5	1-5663	19786	70	97	735	32	2012-08-02 05:20:38.353369
19789	5	1-5686	19786	97	99	313	38	2012-08-02 05:20:38.353369
19790	5	1-5080	19786	99	26	302	28	2012-08-02 05:20:38.353369
19786	5	1-5692	19786	26	95	1314	93	2012-08-02 05:20:38.353369
19792	4	1-4399	19791	69	45	2804	88	2012-08-02 05:20:39.149373
19793	4	1-5438	19791	45	8	1725	68	2012-08-02 05:20:39.149373
19794	4	1-5363	19791	8	85	940	72	2012-08-02 05:20:39.149373
19791	4	1-5694	19791	85	69	7641	94	2012-08-02 05:20:39.149373
19796	4	1-5121	19795	70	30	3842	88	2012-08-02 05:20:39.56919
19797	4	1-5209	19795	30	14	2011	53	2012-08-02 05:20:39.56919
19798	4	1-5477	19795	14	63	1822	34	2012-08-02 05:20:39.56919
19795	4	1-5695	19795	63	70	3516	26	2012-08-02 05:20:39.56919
19800	5	1-5121	19799	70	20	206	88	2012-08-02 05:20:39.56919
19801	5	1-5109	19799	20	76	164	44	2012-08-02 05:20:39.56919
19802	5	1-775	19799	76	14	88	53	2012-08-02 05:20:39.56919
19803	5	1-5477	19799	14	63	85	34	2012-08-02 05:20:39.56919
19799	5	1-5695	19799	63	70	176	26	2012-08-02 05:20:39.56919
19805	4	1-3092	19804	31	28	54	39	2012-08-02 05:20:39.56919
19806	4	1-5537	19804	28	84	260	58	2012-08-02 05:20:39.56919
19807	4	1-5256	19804	84	63	39	34	2012-08-02 05:20:39.56919
19804	4	1-5695	19804	63	31	84	26	2012-08-02 05:20:39.56919
19809	5	1-3788	19808	97	24	264	18	2012-08-02 05:20:39.56919
19810	5	1-5047	19808	24	76	574	44	2012-08-02 05:20:39.56919
19811	5	1-775	19808	76	14	331	53	2012-08-02 05:20:39.56919
19812	5	1-5477	19808	14	63	347	34	2012-08-02 05:20:39.56919
19808	5	1-5695	19808	63	97	775	26	2012-08-02 05:20:39.56919
19814	5	1-3980	19813	37	97	129	73	2012-08-02 05:20:39.56919
19815	5	1-5646	19813	97	70	301	69	2012-08-02 05:20:39.56919
19816	5	1-1279	19813	70	14	187	53	2012-08-02 05:20:39.56919
19817	5	1-5477	19813	14	63	201	34	2012-08-02 05:20:39.56919
19813	5	1-5695	19813	63	37	459	26	2012-08-02 05:20:39.56919
19819	5	1-5263	19818	43	76	3808	54	2012-08-02 05:20:39.56919
19820	5	1-4602	19818	76	50	2322	92	2012-08-02 05:20:39.56919
19821	5	1-5486	19818	50	6	1459	55	2012-08-02 05:20:39.56919
19822	5	1-5374	19818	6	63	1792	34	2012-08-02 05:20:39.56919
19818	5	1-5695	19818	63	43	4105	26	2012-08-02 05:20:39.56919
19824	5	1-5572	19823	5	9	97	13	2012-08-02 05:20:39.56919
19825	5	1-4478	19823	9	30	21	79	2012-08-02 05:20:39.56919
19826	5	1-3351	19823	30	6	7	55	2012-08-02 05:20:39.56919
19827	5	1-5374	19823	6	63	9	34	2012-08-02 05:20:39.56919
19823	5	1-5695	19823	63	5	21	26	2012-08-02 05:20:39.56919
19829	4	1-3767	19828	47	54	636	32	2012-08-02 05:20:40.398139
19830	4	1-4874	19828	54	74	200	61	2012-08-02 05:20:40.398139
19831	4	1-5470	19828	74	92	120	68	2012-08-02 05:20:40.398139
19828	4	1-5697	19828	92	47	2628	4	2012-08-02 05:20:40.398139
19833	4	1-5100	19832	20	56	137	75	2012-08-02 05:20:40.398139
19834	4	1-2507	19832	56	74	46	61	2012-08-02 05:20:40.398139
19835	4	1-5470	19832	74	92	28	68	2012-08-02 05:20:40.398139
19832	4	1-5697	19832	92	20	620	4	2012-08-02 05:20:40.398139
19837	4	1-5278	19836	8	13	877	82	2012-08-02 05:20:40.398139
19838	4	1-5135	19836	13	23	349	72	2012-08-02 05:20:40.398139
19839	4	1-5039	19836	23	92	151	68	2012-08-02 05:20:40.398139
19836	4	1-5697	19836	92	8	3996	4	2012-08-02 05:20:40.398139
19841	4	1-5278	19840	8	24	42	82	2012-08-02 05:20:40.398139
19842	4	1-3656	19840	24	82	11	10	2012-08-02 05:20:40.398139
19843	4	1-5452	19840	82	92	7	68	2012-08-02 05:20:40.398139
19840	4	1-5697	19840	92	8	184	4	2012-08-02 05:20:40.398139
19845	4	1-5631	19844	97	41	872	18	2012-08-02 05:20:41.646374
19846	4	1-4438	19844	41	64	1057	50	2012-08-02 05:20:41.646374
19847	4	1-5673	19844	64	98	480	96	2012-08-02 05:20:41.646374
19844	4	1-5700	19844	98	97	2336	66	2012-08-02 05:20:41.646374
19849	5	1-5380	19848	60	54	3068	59	2012-08-02 05:20:44.408457
19850	5	1-5365	19848	54	8	2469	4	2012-08-02 05:20:44.408457
19851	5	1-5278	19848	8	59	956	82	2012-08-02 05:20:44.408457
19852	5	1-5319	19848	59	78	716	70	2012-08-02 05:20:44.408457
19848	5	1-5707	19848	78	60	3290	76	2012-08-02 05:20:44.408457
19854	4	1-5674	19853	2	47	2193	41	2012-08-02 05:20:46.772284
19855	4	1-5143	19853	47	2	1321	26	2012-08-02 05:20:46.772284
19856	4	1-2178	19853	2	27	468	31	2012-08-02 05:20:46.772284
19853	4	1-5713	19853	27	2	3287	62	2012-08-02 05:20:46.772284
19858	4	1-4908	19857	57	69	1936	36	2012-08-02 05:20:46.772284
19859	4	1-4473	19857	69	37	642	70	2012-08-02 05:20:46.772284
19860	4	1-4715	19857	37	27	334	31	2012-08-02 05:20:46.772284
19857	4	1-5713	19857	27	57	2391	62	2012-08-02 05:20:46.772284
19862	5	1-4474	19861	82	20	853	20	2012-08-02 05:20:46.772284
19863	5	1-5562	19861	20	98	482	18	2012-08-02 05:20:46.772284
19864	5	1-5057	19861	98	15	377	30	2012-08-02 05:20:46.772284
19865	5	1-1815	19861	15	27	231	31	2012-08-02 05:20:46.772284
19861	5	1-5713	19861	27	82	1802	62	2012-08-02 05:20:46.772284
19867	4	1-5014	19866	30	74	3808	66	2012-08-02 05:20:47.578829
19866	4	1-5715	19866	1	30	5648	49	2012-08-02 05:20:47.578829
19868	4	1-5704	19866	74	12	5366	17	2012-08-02 05:20:47.578829
19869	4	1-5439	19866	12	1	2629	15	2012-08-02 05:20:47.578829
19871	4	1-5219	19870	91	66	2177	83	2012-08-02 05:20:47.578829
19872	4	1-4931	19870	66	97	1018	79	2012-08-02 05:20:47.578829
19873	4	1-3801	19870	97	1	846	15	2012-08-02 05:20:47.578829
19870	4	1-5715	19870	1	91	1949	49	2012-08-02 05:20:47.578829
19875	4	1-5239	19874	48	78	1016	39	2012-08-02 05:20:47.578829
19876	4	1-4868	19874	78	88	866	40	2012-08-02 05:20:47.578829
19877	4	1-5393	19874	88	1	954	15	2012-08-02 05:20:47.578829
19874	4	1-5715	19874	1	48	2360	49	2012-08-02 05:20:47.578829
19879	4	1-5254	19878	97	57	2780	38	2012-08-02 05:20:48.794687
19880	4	1-5512	19878	57	34	7340	40	2012-08-02 05:20:48.794687
19881	4	1-4969	19878	34	58	3594	100	2012-08-02 05:20:48.794687
19878	4	1-5718	19878	58	97	3812	1	2012-08-02 05:20:48.794687
19883	5	1-5449	19882	59	97	20	1	2012-08-02 05:20:49.203208
19884	5	1-5254	19882	97	72	11	38	2012-08-02 05:20:49.203208
19885	5	1-1554	19882	72	25	5	99	2012-08-02 05:20:49.203208
19886	5	1-5366	19882	25	84	4	6	2012-08-02 05:20:49.203208
19882	5	1-5719	19882	84	59	44	28	2012-08-02 05:20:49.203208
19888	5	1-5334	19887	3	3	1935	63	2012-08-02 05:20:49.203208
19889	5	1-5158	19887	3	97	700	67	2012-08-02 05:20:49.203208
19890	5	1-2168	19887	97	25	99	99	2012-08-02 05:20:49.203208
19891	5	1-5366	19887	25	84	120	6	2012-08-02 05:20:49.203208
19887	5	1-5719	19887	84	3	1715	28	2012-08-02 05:20:49.203208
19893	4	1-5449	19892	59	42	2526	1	2012-08-02 05:20:49.203208
19894	4	1-1985	19892	42	43	1110	72	2012-08-02 05:20:49.203208
19895	4	1-1718	19892	43	84	265	6	2012-08-02 05:20:49.203208
19892	4	1-5719	19892	84	59	3980	28	2012-08-02 05:20:49.203208
19897	4	1-5371	19896	8	61	773	93	2012-08-02 05:20:49.623292
19898	4	1-5459	19896	61	100	724	95	2012-08-02 05:20:49.623292
19899	4	1-5531	19896	100	95	1032	43	2012-08-02 05:20:49.623292
19896	4	1-5720	19896	95	8	1226	79	2012-08-02 05:20:49.623292
19901	5	1-3222	19900	79	84	711	36	2012-08-02 05:20:50.816447
19902	5	1-5678	19900	84	42	719	86	2012-08-02 05:20:50.816447
19903	5	1-5482	19900	42	78	977	70	2012-08-02 05:20:50.816447
19904	5	1-5707	19900	78	91	3709	76	2012-08-02 05:20:50.816447
19900	5	1-5723	19900	91	79	2372	30	2012-08-02 05:20:50.816447
19906	4	1-5602	19905	24	76	4618	60	2012-08-02 05:20:50.816447
19907	4	1-4580	19905	76	58	3695	16	2012-08-02 05:20:50.816447
19908	4	1-5510	19905	58	91	3864	76	2012-08-02 05:20:50.816447
19905	4	1-5723	19905	91	24	3395	30	2012-08-02 05:20:50.816447
19910	5	1-5602	19909	24	76	923	60	2012-08-02 05:20:50.816447
19911	5	1-4580	19909	76	72	742	16	2012-08-02 05:20:50.816447
19912	5	1-5583	19909	72	96	573	10	2012-08-02 05:20:50.816447
19913	5	1-5294	19909	96	91	764	76	2012-08-02 05:20:50.816447
19909	5	1-5723	19909	91	24	675	30	2012-08-02 05:20:50.816447
19915	5	1-1815	19914	15	70	164	31	2012-08-02 05:20:50.816447
19916	5	1-5663	19914	70	69	235	32	2012-08-02 05:20:50.816447
19917	5	1-4239	19914	69	96	168	10	2012-08-02 05:20:50.816447
19918	5	1-5294	19914	96	91	228	76	2012-08-02 05:20:50.816447
19914	5	1-5723	19914	91	15	205	30	2012-08-02 05:20:50.816447
19920	4	1-5609	19919	65	82	1590	33	2012-08-02 05:20:50.816447
19921	4	1-5644	19919	82	68	854	8	2012-08-02 05:20:50.816447
19922	4	1-3090	19919	68	91	842	76	2012-08-02 05:20:50.816447
19919	4	1-5723	19919	91	65	763	30	2012-08-02 05:20:50.816447
19923	1	1-4580	\N	76	76	1049	16	2012-08-02 05:20:50.816447
19925	2	1-4740	19924	17	89	846	88	2012-08-02 05:20:51.236637
19924	2	1-5724	19924	89	17	1311	11	2012-08-02 05:20:51.236637
19927	4	1-5437	19926	23	41	2112	70	2012-08-02 05:20:52.473877
19928	4	1-4614	19926	41	63	1844	46	2012-08-02 05:20:52.473877
19929	4	1-5617	19926	63	32	1846	57	2012-08-02 05:20:52.473877
19926	4	1-5727	19926	32	23	4179	84	2012-08-02 05:20:52.473877
19931	3	1-1533	19930	90	63	865	46	2012-08-02 05:20:52.473877
19932	3	1-5617	19930	63	32	892	57	2012-08-02 05:20:52.473877
19930	3	1-5727	19930	32	90	2079	84	2012-08-02 05:20:52.473877
19934	5	1-5377	19933	55	57	113	62	2012-08-02 05:20:52.904843
19935	5	1-4908	19933	57	35	119	36	2012-08-02 05:20:52.904843
19936	5	1-4032	19933	35	89	74	88	2012-08-02 05:20:52.904843
19937	5	1-5724	19933	89	90	141	11	2012-08-02 05:20:52.904843
19933	5	1-5728	19933	90	55	119	55	2012-08-02 05:20:52.904843
19939	5	1-5377	19938	55	2	338	62	2012-08-02 05:20:52.904843
19940	5	1-5674	19938	2	51	270	41	2012-08-02 05:20:52.904843
19941	5	1-5276	19938	51	89	166	88	2012-08-02 05:20:52.904843
19942	5	1-5724	19938	89	90	348	11	2012-08-02 05:20:52.904843
19938	5	1-5728	19938	90	55	325	55	2012-08-02 05:20:52.904843
19943	1	1-1064	\N	8	8	250	55	2012-08-02 05:20:52.904843
19945	5	1-5538	19944	34	50	18	89	2012-08-02 05:20:53.336057
19946	5	1-2193	19944	50	20	12	68	2012-08-02 05:20:53.336057
19947	5	1-5685	19944	20	36	26	78	2012-08-02 05:20:53.336057
19948	5	1-5554	19944	36	27	15	23	2012-08-02 05:20:53.336057
19944	5	1-5729	19944	27	34	41	73	2012-08-02 05:20:53.336057
19950	5	1-1279	19949	70	70	572	53	2012-08-02 05:20:54.186713
19951	5	1-4627	19949	70	70	433	31	2012-08-02 05:20:54.186713
19952	5	1-5663	19949	70	97	603	32	2012-08-02 05:20:54.186713
19953	5	1-5686	19949	97	78	246	38	2012-08-02 05:20:54.186713
19949	5	1-5731	19949	78	70	915	69	2012-08-02 05:20:54.186713
19954	1	1-1279	\N	70	70	1604	53	2012-08-02 05:20:54.186713
19956	4	1-5268	19955	80	70	2471	31	2012-08-02 05:20:54.653925
19957	4	1-5663	19955	70	60	3443	32	2012-08-02 05:20:54.653925
19958	4	1-4068	19955	60	84	1164	30	2012-08-02 05:20:54.653925
19955	4	1-5732	19955	84	80	3495	25	2012-08-02 05:20:54.653925
19960	5	1-5069	19959	19	47	1778	92	2012-08-02 05:20:56.837819
19961	5	1-4698	19959	47	6	1341	87	2012-08-02 05:20:56.837819
19962	5	1-2663	19959	6	90	615	2	2012-08-02 05:20:56.837819
19963	5	1-5430	19959	90	99	325	71	2012-08-02 05:20:56.837819
19959	5	1-5737	19959	99	19	6367	57	2012-08-02 05:20:56.837819
19965	2	1-5605	19964	96	83	1888	60	2012-08-02 05:21:00.571757
19964	2	1-5746	19964	83	96	4320	39	2012-08-02 05:21:00.571757
19967	4	1-4295	19966	62	15	189	1	2012-08-02 05:21:03.92126
19968	4	1-5624	19966	15	66	648	85	2012-08-02 05:21:03.92126
19969	4	1-3517	19966	66	91	168	73	2012-08-02 05:21:03.92126
19966	4	1-5754	19966	91	62	164	98	2012-08-02 05:21:03.92126
19971	5	1-761	19970	30	20	272	68	2012-08-02 05:21:03.92126
19972	5	1-5685	19970	20	36	649	78	2012-08-02 05:21:03.92126
19973	5	1-5554	19970	36	27	423	23	2012-08-02 05:21:03.92126
19974	5	1-5729	19970	27	91	1318	73	2012-08-02 05:21:03.92126
19970	5	1-5754	19970	91	30	1303	98	2012-08-02 05:21:03.92126
19976	5	1-3390	19975	80	43	125	72	2012-08-02 05:21:05.180877
19977	5	1-1718	19975	43	84	30	6	2012-08-02 05:21:05.180877
19978	5	1-5719	19975	84	1	460	28	2012-08-02 05:21:05.180877
19979	5	1-5434	19975	1	73	381	48	2012-08-02 05:21:05.180877
19975	5	1-5757	19975	73	80	242	89	2012-08-02 05:21:05.180877
19981	5	1-5320	19980	54	23	1071	84	2012-08-02 05:21:05.749486
19982	5	1-5437	19980	23	31	546	70	2012-08-02 05:21:05.749486
19983	5	1-3661	19980	31	27	185	50	2012-08-02 05:21:05.749486
19984	5	1-5616	19980	27	82	427	83	2012-08-02 05:21:05.749486
19980	5	1-5758	19980	82	54	626	82	2012-08-02 05:21:05.749486
19986	5	1-5662	19985	86	23	2468	84	2012-08-02 05:21:05.749486
19987	5	1-5437	19985	23	31	1339	70	2012-08-02 05:21:05.749486
19988	5	1-3661	19985	31	27	482	50	2012-08-02 05:21:05.749486
19989	5	1-5616	19985	27	82	1185	83	2012-08-02 05:21:05.749486
19985	5	1-5758	19985	82	86	1848	82	2012-08-02 05:21:05.749486
19991	4	1-5451	19990	64	63	1868	46	2012-08-02 05:21:05.749486
19992	4	1-5617	19990	63	46	2055	57	2012-08-02 05:21:05.749486
19993	4	1-2840	19990	46	82	453	83	2012-08-02 05:21:05.749486
19990	4	1-5758	19990	82	64	723	82	2012-08-02 05:21:05.749486
19995	4	1-1884	19994	68	31	841	22	2012-08-02 05:21:05.749486
19996	4	1-5566	19994	31	91	655	49	2012-08-02 05:21:05.749486
19997	4	1-5219	19994	91	82	714	83	2012-08-02 05:21:05.749486
19994	4	1-5758	19994	82	68	1175	82	2012-08-02 05:21:05.749486
19999	5	1-5657	19998	79	14	1800	53	2012-08-02 05:21:06.176608
20000	5	1-5477	19998	14	71	1852	34	2012-08-02 05:21:06.176608
20001	5	1-5603	19998	71	20	1466	20	2012-08-02 05:21:06.176608
20002	5	1-5562	19998	20	78	1007	18	2012-08-02 05:21:06.176608
19998	5	1-5759	19998	78	79	1639	90	2012-08-02 05:21:06.176608
20004	4	1-5482	20003	42	59	631	70	2012-08-02 05:21:06.960805
20005	4	1-4018	20003	59	29	191	74	2012-08-02 05:21:06.960805
20006	4	1-5187	20003	29	28	157	22	2012-08-02 05:21:06.960805
20003	4	1-5761	20003	28	42	431	86	2012-08-02 05:21:06.960805
20008	5	1-5482	20007	42	37	6264	70	2012-08-02 05:21:06.960805
20009	5	1-4715	20007	37	55	3998	31	2012-08-02 05:21:06.960805
20010	5	1-5524	20007	55	68	1759	82	2012-08-02 05:21:06.960805
20011	5	1-1884	20007	68	28	1238	22	2012-08-02 05:21:06.960805
20007	5	1-5761	20007	28	42	3814	86	2012-08-02 05:21:06.960805
20012	1	1-1884	\N	68	68	1561	22	2012-08-02 05:21:06.960805
20014	4	1-5585	20013	84	39	203	94	2012-08-02 05:21:08.165299
20015	4	1-4764	20013	39	14	428	15	2012-08-02 05:21:08.165299
20016	4	1-4996	20013	14	82	381	14	2012-08-02 05:21:08.165299
20013	4	1-5764	20013	82	84	296	84	2012-08-02 05:21:08.165299
20018	4	1-5368	20017	77	63	2222	63	2012-08-02 05:21:08.165299
20019	4	1-4600	20017	63	14	1995	15	2012-08-02 05:21:08.165299
20020	4	1-4996	20017	14	82	2509	14	2012-08-02 05:21:08.165299
20017	4	1-5764	20017	82	77	2753	84	2012-08-02 05:21:08.165299
20022	4	1-5585	20021	84	80	1008	94	2012-08-02 05:21:08.165299
20023	4	1-5574	20021	80	83	968	63	2012-08-02 05:21:08.165299
20024	4	1-4728	20021	83	82	942	14	2012-08-02 05:21:08.165299
20021	4	1-5764	20021	82	84	1037	84	2012-08-02 05:21:08.165299
20026	6	1-5538	20025	34	42	187	89	2012-08-02 05:21:08.562971
20027	6	1-5296	20025	42	90	295	2	2012-08-02 05:21:08.562971
20028	6	1-5430	20025	90	5	195	71	2012-08-02 05:21:08.562971
20029	6	1-5577	20025	5	66	421	70	2012-08-02 05:21:08.562971
20030	6	1-5242	20025	66	69	165	98	2012-08-02 05:21:08.562971
20025	6	1-5765	20025	69	34	343	73	2012-08-02 05:21:08.562971
20032	3	1-1554	20031	72	25	546	99	2012-08-02 05:21:08.949598
20033	3	1-5366	20031	25	3	668	6	2012-08-02 05:21:08.949598
20031	3	1-5766	20031	3	72	981	38	2012-08-02 05:21:08.949598
20035	5	1-5751	20034	48	74	146	66	2012-08-02 05:21:14.12016
20036	5	1-5704	20034	74	98	200	17	2012-08-02 05:21:14.12016
20037	5	1-5703	20034	98	100	163	45	2012-08-02 05:21:14.12016
20038	5	1-4167	20034	100	12	229	50	2012-08-02 05:21:14.12016
20034	5	1-5779	20034	12	48	699	87	2012-08-02 05:21:14.12016
20040	5	1-5270	20039	92	93	189	21	2012-08-02 05:21:16.539619
20041	5	1-4666	20039	93	20	56	20	2012-08-02 05:21:16.539619
20042	5	1-5562	20039	20	64	42	18	2012-08-02 05:21:16.539619
20043	5	1-5627	20039	64	1	341	54	2012-08-02 05:21:16.539619
20039	5	1-5785	20039	1	92	293	17	2012-08-02 05:21:16.539619
20045	4	1-5758	20044	82	55	5572	82	2012-08-02 05:21:16.93694
20046	4	1-4966	20044	55	12	5017	11	2012-08-02 05:21:16.93694
20047	4	1-5547	20044	12	15	3271	29	2012-08-02 05:21:16.93694
20044	4	1-5786	20044	15	82	3095	83	2012-08-02 05:21:16.93694
20049	3	1-5115	20048	3	79	1380	43	2012-08-02 05:21:16.93694
20050	3	1-5595	20048	79	15	1274	29	2012-08-02 05:21:16.93694
20048	3	1-5786	20048	15	3	1276	83	2012-08-02 05:21:16.93694
20052	5	1-2163	20051	23	55	531	31	2012-08-02 05:21:18.108254
20053	5	1-5524	20051	55	8	203	82	2012-08-02 05:21:18.108254
20054	5	1-5717	20051	8	100	180	64	2012-08-02 05:21:18.108254
20055	5	1-4881	20051	100	97	303	10	2012-08-02 05:21:18.108254
20051	5	1-5789	20051	97	23	429	35	2012-08-02 05:21:18.108254
20057	5	1-1785	20056	96	29	2181	74	2012-08-02 05:21:18.108254
20058	5	1-5187	20056	29	23	2268	22	2012-08-02 05:21:18.108254
20059	5	1-4676	20056	23	100	696	64	2012-08-02 05:21:18.108254
20060	5	1-4881	20056	100	97	1520	10	2012-08-02 05:21:18.108254
20056	5	1-5789	20056	97	96	2795	35	2012-08-02 05:21:18.108254
20062	4	1-4972	20061	93	93	1310	70	2012-08-02 05:21:18.108254
20063	4	1-5594	20061	93	93	587	29	2012-08-02 05:21:18.108254
20064	4	1-4909	20061	93	97	616	10	2012-08-02 05:21:18.108254
20061	4	1-5789	20061	97	93	1170	35	2012-08-02 05:21:18.108254
20066	5	1-5526	20065	15	59	343	93	2012-08-02 05:21:18.915135
20067	5	1-5763	20065	59	60	1036	21	2012-08-02 05:21:18.915135
20068	5	1-3903	20065	60	40	367	12	2012-08-02 05:21:18.915135
20069	5	1-4403	20065	40	27	593	57	2012-08-02 05:21:18.915135
20065	5	1-5791	20065	27	15	168	75	2012-08-02 05:21:18.915135
20071	2	1-4060	20070	80	50	943	14	2012-08-02 05:21:19.721592
20070	2	1-5793	20070	50	80	5175	80	2012-08-02 05:21:19.721592
20073	5	1-5242	20072	66	83	1596	98	2012-08-02 05:21:22.0965
20074	5	1-5530	20072	83	81	2560	26	2012-08-02 05:21:22.0965
20075	5	1-4105	20072	81	32	1645	87	2012-08-02 05:21:22.0965
20076	5	1-5641	20072	32	27	2764	63	2012-08-02 05:21:22.0965
20072	5	1-5799	20072	27	66	4693	70	2012-08-02 05:21:22.0965
20078	5	1-1478	20077	13	74	450	66	2012-08-02 05:21:22.0965
20079	5	1-5704	20077	74	98	714	17	2012-08-02 05:21:22.0965
20080	5	1-5703	20077	98	46	669	45	2012-08-02 05:21:22.0965
20081	5	1-5460	20077	46	27	1772	63	2012-08-02 05:21:22.0965
20077	5	1-5799	20077	27	13	3304	70	2012-08-02 05:21:22.0965
20083	4	1-5672	20082	73	33	3437	81	2012-08-02 05:21:24.07404
20084	4	1-5387	20082	33	89	3601	65	2012-08-02 05:21:24.07404
20085	4	1-5706	20082	89	57	1366	72	2012-08-02 05:21:24.07404
20082	4	1-5804	20082	57	73	2430	3	2012-08-02 05:21:24.07404
20087	4	1-4929	20086	72	62	2615	20	2012-08-02 05:21:24.07404
20088	4	1-5458	20086	62	58	2644	12	2012-08-02 05:21:24.07404
20089	4	1-5076	20086	58	57	1318	72	2012-08-02 05:21:24.07404
20086	4	1-5804	20086	57	72	3197	3	2012-08-02 05:21:24.07404
20091	4	1-5700	20090	98	74	369	66	2012-08-02 05:21:26.09599
20092	4	1-5704	20090	74	39	341	17	2012-08-02 05:21:26.09599
20093	4	1-5298	20090	39	45	190	27	2012-08-02 05:21:26.09599
20090	4	1-5809	20090	45	98	133	96	2012-08-02 05:21:26.09599
20095	4	1-5338	20094	94	56	596	12	2012-08-02 05:21:26.09599
20096	4	1-3838	20094	56	39	526	17	2012-08-02 05:21:26.09599
20097	4	1-5298	20094	39	45	411	27	2012-08-02 05:21:26.09599
20094	4	1-5809	20094	45	94	401	96	2012-08-02 05:21:26.09599
20099	4	1-4897	20098	7	26	4773	89	2012-08-02 05:21:26.09599
20100	4	1-5386	20098	26	61	2488	47	2012-08-02 05:21:26.09599
20101	4	1-5383	20098	61	45	3492	27	2012-08-02 05:21:26.09599
20098	4	1-5809	20098	45	7	3518	96	2012-08-02 05:21:26.09599
20103	5	1-4409	20102	3	30	648	20	2012-08-02 05:21:26.09599
20104	5	1-5714	20102	30	1	1144	54	2012-08-02 05:21:26.09599
20105	5	1-5785	20102	1	39	781	17	2012-08-02 05:21:26.09599
20106	5	1-5298	20102	39	45	638	27	2012-08-02 05:21:26.09599
20102	5	1-5809	20102	45	3	652	96	2012-08-02 05:21:26.09599
20108	5	1-5338	20107	94	51	917	12	2012-08-02 05:21:26.09599
20109	5	1-5206	20107	51	26	706	89	2012-08-02 05:21:26.09599
20110	5	1-5386	20107	26	61	381	47	2012-08-02 05:21:26.09599
20111	5	1-5383	20107	61	45	553	27	2012-08-02 05:21:26.09599
20107	5	1-5809	20107	45	94	577	96	2012-08-02 05:21:26.09599
20113	3	1-5338	20112	94	11	636	12	2012-08-02 05:21:26.09599
20114	3	1-4147	20112	11	45	259	27	2012-08-02 05:21:26.09599
20112	3	1-5809	20112	45	94	329	96	2012-08-02 05:21:26.09599
20116	4	1-5338	20115	94	98	1311	12	2012-08-02 05:21:26.09599
20117	4	1-4987	20115	98	71	723	30	2012-08-02 05:21:26.09599
20118	4	1-4497	20115	71	45	517	27	2012-08-02 05:21:26.09599
20115	4	1-5809	20115	45	94	667	96	2012-08-02 05:21:26.09599
20120	5	1-5530	20119	83	70	1953	26	2012-08-02 05:21:26.913775
20121	5	1-5121	20119	70	45	1642	88	2012-08-02 05:21:26.913775
20122	5	1-5438	20119	45	20	857	68	2012-08-02 05:21:26.913775
20123	5	1-5685	20119	20	87	1427	78	2012-08-02 05:21:26.913775
20119	5	1-5811	20119	87	83	1658	98	2012-08-02 05:21:26.913775
20125	5	1-4295	20124	62	29	1414	1	2012-08-02 05:21:26.913775
20126	5	1-5285	20124	29	45	1015	27	2012-08-02 05:21:26.913775
20127	5	1-5809	20124	45	47	1138	96	2012-08-02 05:21:26.913775
20128	5	1-5353	20124	47	87	763	78	2012-08-02 05:21:26.913775
20124	5	1-5811	20124	87	62	1245	98	2012-08-02 05:21:26.913775
20130	5	1-5530	20129	83	52	526	26	2012-08-02 05:21:26.913775
20131	5	1-1931	20129	52	22	468	93	2012-08-02 05:21:26.913775
20132	5	1-5269	20129	22	12	103	35	2012-08-02 05:21:26.913775
20133	5	1-2275	20129	12	87	163	78	2012-08-02 05:21:26.913775
20129	5	1-5811	20129	87	83	291	98	2012-08-02 05:21:26.913775
20135	4	1-4999	20134	62	17	1141	80	2012-08-02 05:21:26.913775
20136	4	1-5304	20134	17	100	874	100	2012-08-02 05:21:26.913775
20137	4	1-3136	20134	100	87	225	78	2012-08-02 05:21:26.913775
20134	4	1-5811	20134	87	62	413	98	2012-08-02 05:21:26.913775
20139	5	1-1221	20138	25	54	1078	24	2012-08-02 05:21:26.913775
20140	5	1-3895	20138	54	62	1443	20	2012-08-02 05:21:26.913775
20141	5	1-5458	20138	62	64	1546	12	2012-08-02 05:21:26.913775
20142	5	1-4486	20138	64	87	1870	78	2012-08-02 05:21:26.913775
20138	5	1-5811	20138	87	25	3460	98	2012-08-02 05:21:26.913775
20144	5	1-5732	20143	84	95	1447	25	2012-08-02 05:21:27.333921
20145	5	1-5600	20143	95	28	813	91	2012-08-02 05:21:27.333921
20146	5	1-5589	20143	28	85	474	9	2012-08-02 05:21:27.333921
20147	5	1-4297	20143	85	37	331	95	2012-08-02 05:21:27.333921
20143	5	1-5812	20143	37	84	820	30	2012-08-02 05:21:27.333921
20149	5	1-5609	20148	65	69	4356	33	2012-08-02 05:21:27.333921
20150	5	1-5127	20148	69	84	2502	7	2012-08-02 05:21:27.333921
20151	5	1-4900	20148	84	90	1365	49	2012-08-02 05:21:27.333921
20152	5	1-5721	20148	90	37	830	95	2012-08-02 05:21:27.333921
20148	5	1-5812	20148	37	65	2757	30	2012-08-02 05:21:27.333921
20154	5	1-4497	20153	71	45	1743	27	2012-08-02 05:21:27.333921
20155	5	1-5809	20153	45	22	1811	96	2012-08-02 05:21:27.333921
20156	5	1-4673	20153	22	90	1313	49	2012-08-02 05:21:27.333921
20157	5	1-5721	20153	90	37	852	95	2012-08-02 05:21:27.333921
20153	5	1-5812	20153	37	71	3025	30	2012-08-02 05:21:27.333921
20159	4	1-5044	20158	34	96	625	66	2012-08-02 05:21:27.333921
20160	4	1-4083	20158	96	17	269	12	2012-08-02 05:21:27.333921
20161	4	1-5134	20158	17	37	179	95	2012-08-02 05:21:27.333921
20158	4	1-5812	20158	37	34	674	30	2012-08-02 05:21:27.333921
20163	4	1-3723	20162	55	90	978	11	2012-08-02 05:21:27.333921
20164	4	1-5728	20162	90	22	807	55	2012-08-02 05:21:27.333921
20165	4	1-2991	20162	22	37	297	95	2012-08-02 05:21:27.333921
20162	4	1-5812	20162	37	55	1147	30	2012-08-02 05:21:27.333921
20166	1	1-4083	\N	96	96	624	12	2012-08-02 05:21:27.333921
20168	5	1-2232	20167	14	89	865	88	2012-08-02 05:21:28.562718
20169	5	1-5724	20167	89	12	1780	11	2012-08-02 05:21:28.562718
20170	5	1-5547	20167	12	15	1199	29	2012-08-02 05:21:28.562718
20171	5	1-5786	20167	15	82	1173	83	2012-08-02 05:21:28.562718
20167	5	1-5815	20167	82	14	3934	50	2012-08-02 05:21:28.562718
20173	5	1-5605	20172	96	93	3022	60	2012-08-02 05:21:28.959996
20174	5	1-5596	20172	93	36	4162	85	2012-08-02 05:21:28.959996
20175	5	1-5564	20172	36	92	1658	66	2012-08-02 05:21:28.959996
20176	5	1-3260	20172	92	24	1169	88	2012-08-02 05:21:28.959996
20172	5	1-5816	20172	24	96	3646	39	2012-08-02 05:21:28.959996
20178	5	1-5730	20177	94	93	1227	55	2012-08-02 05:21:29.369037
20179	5	1-5794	20177	93	63	1226	46	2012-08-02 05:21:29.369037
20180	5	1-5617	20177	63	27	1379	57	2012-08-02 05:21:29.369037
20181	5	1-5791	20177	27	66	355	75	2012-08-02 05:21:29.369037
20177	5	1-5817	20177	66	94	1541	76	2012-08-02 05:21:29.369037
20183	3	1-5592	20182	100	95	6057	57	2012-08-02 05:21:30.57304
20184	3	1-5755	20182	95	45	5706	40	2012-08-02 05:21:30.57304
20182	3	1-5820	20182	45	100	4239	6	2012-08-02 05:21:30.57304
20186	3	1-5766	20185	3	57	1139	38	2012-08-02 05:21:30.57304
20187	3	1-5512	20185	57	45	1762	40	2012-08-02 05:21:30.57304
20185	3	1-5820	20185	45	3	1359	6	2012-08-02 05:21:30.57304
20189	4	1-3869	20188	90	58	1275	100	2012-08-02 05:21:30.57304
20190	4	1-5718	20188	58	6	1204	1	2012-08-02 05:21:30.57304
20191	4	1-3945	20188	6	45	1653	40	2012-08-02 05:21:30.57304
20188	4	1-5820	20188	45	90	1938	6	2012-08-02 05:21:30.57304
20193	4	1-3754	20192	40	84	345	36	2012-08-02 05:21:30.57304
20194	4	1-5678	20192	84	83	466	86	2012-08-02 05:21:30.57304
20195	4	1-1896	20192	83	45	460	40	2012-08-02 05:21:30.57304
20192	4	1-5820	20192	45	40	627	6	2012-08-02 05:21:30.57304
20197	4	1-554	20196	6	59	98	93	2012-08-02 05:21:32.59489
20198	4	1-5763	20196	59	19	298	21	2012-08-02 05:21:32.59489
20199	4	1-5138	20196	19	22	2183	37	2012-08-02 05:21:32.59489
20196	4	1-5825	20196	22	6	737	50	2012-08-02 05:21:32.59489
20201	5	1-4341	20200	20	72	53	8	2012-08-02 05:21:32.981537
20202	5	1-5798	20200	72	71	105	34	2012-08-02 05:21:32.981537
20203	5	1-5603	20200	71	33	76	20	2012-08-02 05:21:32.981537
20204	5	1-5177	20200	33	30	52	56	2012-08-02 05:21:32.981537
20200	5	1-5826	20200	30	20	145	59	2012-08-02 05:21:32.981537
20206	4	1-4946	20205	33	71	789	34	2012-08-02 05:21:32.981537
20207	4	1-5603	20205	71	33	668	20	2012-08-02 05:21:32.981537
20208	4	1-5177	20205	33	30	546	56	2012-08-02 05:21:32.981537
20205	4	1-5826	20205	30	33	1772	59	2012-08-02 05:21:32.981537
20209	1	1-5177	\N	33	33	136	56	2012-08-02 05:21:32.981537
20211	5	1-5822	20210	7	47	864	92	2012-08-02 05:21:33.390577
20212	5	1-4698	20210	47	87	447	87	2012-08-02 05:21:33.390577
20213	5	1-2490	20210	87	55	110	31	2012-08-02 05:21:33.390577
20214	5	1-5524	20210	55	7	30	82	2012-08-02 05:21:33.390577
20210	5	1-5827	20210	7	7	309	81	2012-08-02 05:21:33.390577
20216	5	1-4168	20215	33	44	149	94	2012-08-02 05:21:33.390577
20217	5	1-5783	20215	44	8	190	48	2012-08-02 05:21:33.390577
20218	5	1-5819	20215	8	55	288	31	2012-08-02 05:21:33.390577
20219	5	1-5524	20215	55	7	116	82	2012-08-02 05:21:33.390577
20215	5	1-5827	20215	7	33	1242	81	2012-08-02 05:21:33.390577
20221	5	1-5625	20220	6	10	4446	42	2012-08-02 05:21:34.197076
20222	5	1-5384	20220	10	59	2876	100	2012-08-02 05:21:34.197076
20223	5	1-5006	20220	59	53	1166	79	2012-08-02 05:21:34.197076
20224	5	1-2946	20220	53	5	449	7	2012-08-02 05:21:34.197076
20220	5	1-5829	20220	5	6	3186	59	2012-08-02 05:21:34.197076
20226	4	1-4648	20225	61	93	825	20	2012-08-02 05:21:34.197076
20227	4	1-5823	20225	93	73	667	35	2012-08-02 05:21:34.197076
20228	4	1-809	20225	73	5	340	7	2012-08-02 05:21:34.197076
20225	4	1-5829	20225	5	61	2449	59	2012-08-02 05:21:34.197076
20230	3	1-4697	20229	18	82	3593	14	2012-08-02 05:21:36.605814
20231	3	1-5764	20229	82	1	2312	84	2012-08-02 05:21:36.605814
20229	3	1-5835	20229	1	18	6326	35	2012-08-02 05:21:36.605814
20233	4	1-2752	20232	53	32	913	53	2012-08-02 05:21:36.605814
20234	4	1-3598	20232	32	82	465	14	2012-08-02 05:21:36.605814
20235	4	1-5764	20232	82	1	327	84	2012-08-02 05:21:36.605814
20232	4	1-5835	20232	1	53	978	35	2012-08-02 05:21:36.605814
20237	4	1-4972	20236	93	91	1761	70	2012-08-02 05:21:36.605814
20238	4	1-3216	20236	91	82	1146	14	2012-08-02 05:21:36.605814
20239	4	1-5764	20236	82	1	813	84	2012-08-02 05:21:36.605814
20236	4	1-5835	20236	1	93	2453	35	2012-08-02 05:21:36.605814
20241	5	1-5752	20240	29	92	254	66	2012-08-02 05:21:37.003449
20242	5	1-3260	20240	92	89	175	88	2012-08-02 05:21:37.003449
20243	5	1-5724	20240	89	90	303	11	2012-08-02 05:21:37.003449
20244	5	1-5728	20240	90	38	234	55	2012-08-02 05:21:37.003449
20240	5	1-5836	20240	38	29	562	68	2012-08-02 05:21:37.003449
20246	4	1-5562	20245	20	78	469	18	2012-08-02 05:21:39.798349
20247	4	1-5759	20245	78	30	837	90	2012-08-02 05:21:39.798349
20248	4	1-5528	20245	30	66	821	87	2012-08-02 05:21:39.798349
20245	4	1-5843	20245	66	20	623	20	2012-08-02 05:21:39.798349
20250	4	1-5817	20249	66	94	3751	76	2012-08-02 05:21:41.168326
20251	4	1-5730	20249	94	82	3290	55	2012-08-02 05:21:41.168326
20252	4	1-4767	20249	82	94	1288	61	2012-08-02 05:21:41.168326
20249	4	1-5846	20249	94	66	840	75	2012-08-02 05:21:41.168326
20254	3	1-5843	20253	66	56	4050	20	2012-08-02 05:21:42.781261
20255	3	1-4886	20253	56	25	1112	36	2012-08-02 05:21:42.781261
20253	3	1-5850	20253	25	66	8585	87	2012-08-02 05:21:42.781261
20257	4	1-5760	20256	40	89	1107	88	2012-08-02 05:21:44.007471
20258	4	1-5724	20256	89	38	2075	11	2012-08-02 05:21:44.007471
20259	4	1-5821	20256	38	32	1950	52	2012-08-02 05:21:44.007471
20256	4	1-5853	20256	32	40	2954	53	2012-08-02 05:21:44.007471
20261	5	1-5452	20260	82	8	408	68	2012-08-02 05:21:45.609331
20262	5	1-5363	20260	8	67	292	72	2012-08-02 05:21:45.609331
20263	5	1-4955	20260	67	27	83	19	2012-08-02 05:21:45.609331
20264	5	1-5639	20260	27	51	140	30	2012-08-02 05:21:45.609331
20260	5	1-5857	20260	51	82	345	10	2012-08-02 05:21:45.609331
20266	4	1-4023	20265	23	100	1202	95	2012-08-02 05:21:46.82457
20267	4	1-5531	20265	100	33	1469	43	2012-08-02 05:21:46.82457
20268	4	1-3657	20265	33	90	716	13	2012-08-02 05:21:46.82457
20265	4	1-5860	20265	90	23	2095	69	2012-08-02 05:21:46.82457
20270	5	1-5650	20269	91	98	1190	17	2012-08-02 05:21:46.82457
20271	5	1-5703	20269	98	81	1049	45	2012-08-02 05:21:46.82457
20272	5	1-5461	20269	81	5	158	26	2012-08-02 05:21:46.82457
20273	5	1-5572	20269	5	90	648	13	2012-08-02 05:21:46.82457
20269	5	1-5860	20269	90	91	1929	69	2012-08-02 05:21:46.82457
20275	5	1-1783	20274	24	99	950	87	2012-08-02 05:21:47.233426
20276	5	1-5173	20274	99	8	920	79	2012-08-02 05:21:47.233426
20277	5	1-5371	20274	8	59	505	93	2012-08-02 05:21:47.233426
20278	5	1-5763	20274	59	64	1411	21	2012-08-02 05:21:47.233426
20274	5	1-5861	20274	64	24	5446	58	2012-08-02 05:21:47.233426
20280	5	1-3538	20279	72	71	3095	34	2012-08-02 05:21:48.824493
20281	5	1-5603	20279	71	62	2049	20	2012-08-02 05:21:48.824493
20282	5	1-5458	20279	62	40	1702	12	2012-08-02 05:21:48.824493
20283	5	1-4403	20279	40	58	2237	57	2012-08-02 05:21:48.824493
20279	5	1-5865	20279	58	72	4210	35	2012-08-02 05:21:48.824493
20285	5	1-1454	20284	71	98	3761	17	2012-08-02 05:21:48.824493
20286	5	1-5703	20284	98	2	3372	45	2012-08-02 05:21:48.824493
20287	5	1-5214	20284	2	6	2215	8	2012-08-02 05:21:48.824493
20288	5	1-3486	20284	6	58	2649	57	2012-08-02 05:21:48.824493
20284	5	1-5865	20284	58	71	5779	35	2012-08-02 05:21:48.824493
20290	5	1-4502	20289	66	89	1747	88	2012-08-02 05:21:52.448072
20291	5	1-5724	20289	89	80	3479	11	2012-08-02 05:21:52.448072
20292	5	1-5499	20289	80	18	2143	13	2012-08-02 05:21:52.448072
20293	5	1-5842	20289	18	2	1931	31	2012-08-02 05:21:52.448072
20289	5	1-5874	20289	2	66	2410	64	2012-08-02 05:21:52.448072
20295	5	1-5373	20294	84	69	2144	59	2012-08-02 05:21:52.448072
20296	5	1-5620	20294	69	1	2013	28	2012-08-02 05:21:52.448072
20297	5	1-5434	20294	1	8	1615	48	2012-08-02 05:21:52.448072
20298	5	1-5819	20294	8	2	3019	31	2012-08-02 05:21:52.448072
20294	5	1-5874	20294	2	84	3818	64	2012-08-02 05:21:52.448072
20300	5	1-5763	20299	59	93	6208	21	2012-08-02 05:21:52.85687
20301	5	1-4666	20299	93	93	1393	20	2012-08-02 05:21:52.85687
20302	5	1-5823	20299	93	49	867	35	2012-08-02 05:21:52.85687
20303	5	1-3662	20299	49	16	364	60	2012-08-02 05:21:52.85687
20299	5	1-5875	20299	16	59	3454	93	2012-08-02 05:21:52.85687
20305	5	1-4812	20304	25	51	25	30	2012-08-02 05:21:52.85687
20306	5	1-5857	20304	51	27	40	10	2012-08-02 05:21:52.85687
20307	5	1-5391	20304	27	96	17	39	2012-08-02 05:21:52.85687
20308	5	1-5605	20304	96	16	11	60	2012-08-02 05:21:52.85687
20304	5	1-5875	20304	16	25	113	93	2012-08-02 05:21:52.85687
20310	4	1-5672	20309	73	7	1238	81	2012-08-02 05:21:54.491972
20311	4	1-5822	20309	7	43	3442	92	2012-08-02 05:21:54.491972
20312	4	1-4170	20309	43	3	1322	21	2012-08-02 05:21:54.491972
20309	4	1-5879	20309	3	73	818	3	2012-08-02 05:21:54.491972
20314	3	1-3875	20313	45	59	125	93	2012-08-02 05:21:54.491972
20315	3	1-5763	20313	59	3	329	21	2012-08-02 05:21:54.491972
20313	3	1-5879	20313	3	45	228	3	2012-08-02 05:21:54.491972
20317	3	1-5672	20316	73	95	879	81	2012-08-02 05:21:54.491972
20318	3	1-3808	20316	95	3	551	21	2012-08-02 05:21:54.491972
20316	3	1-5879	20316	3	73	445	3	2012-08-02 05:21:54.491972
20320	5	1-4929	20319	72	62	1230	20	2012-08-02 05:21:54.491972
20321	5	1-5458	20319	62	17	1323	12	2012-08-02 05:21:54.491972
20322	5	1-5134	20319	17	30	1023	95	2012-08-02 05:21:54.491972
20323	5	1-5710	20319	30	3	1686	21	2012-08-02 05:21:54.491972
20319	5	1-5879	20319	3	72	1413	3	2012-08-02 05:21:54.491972
20325	3	1-4977	20324	37	59	795	95	2012-08-02 05:21:56.900663
20326	3	1-4882	20324	59	60	951	62	2012-08-02 05:21:56.900663
20324	3	1-5885	20324	60	37	1970	76	2012-08-02 05:21:56.900663
20328	5	1-4618	20327	24	25	1824	93	2012-08-02 05:21:59.297931
20329	5	1-4812	20327	25	51	486	30	2012-08-02 05:21:59.297931
20330	5	1-5857	20327	51	96	954	10	2012-08-02 05:21:59.297931
20331	5	1-5294	20327	96	21	998	76	2012-08-02 05:21:59.297931
20327	5	1-5891	20327	21	24	2568	5	2012-08-02 05:21:59.297931
20333	5	1-5250	20332	48	31	259	22	2012-08-02 05:21:59.297931
20334	5	1-1851	20332	31	72	181	16	2012-08-02 05:21:59.297931
20335	5	1-5583	20332	72	96	129	10	2012-08-02 05:21:59.297931
20336	5	1-5294	20332	96	21	160	76	2012-08-02 05:21:59.297931
20332	5	1-5891	20332	21	48	486	5	2012-08-02 05:21:59.297931
20338	4	1-4874	20337	54	74	739	61	2012-08-02 05:21:59.720189
20339	4	1-5470	20337	74	16	916	68	2012-08-02 05:21:59.720189
20340	4	1-2286	20337	16	8	670	13	2012-08-02 05:21:59.720189
20337	4	1-5892	20337	8	54	1144	32	2012-08-02 05:21:59.720189
20342	4	1-5686	20341	97	78	673	38	2012-08-02 05:21:59.720189
20343	4	1-3825	20341	78	5	206	26	2012-08-02 05:21:59.720189
20344	4	1-5572	20341	5	8	862	13	2012-08-02 05:21:59.720189
20341	4	1-5892	20341	8	97	1599	32	2012-08-02 05:21:59.720189
20346	4	1-5890	20345	9	8	946	79	2012-08-02 05:22:00.526392
20347	4	1-5371	20345	8	25	364	93	2012-08-02 05:22:00.526392
20348	4	1-4812	20345	25	19	78	30	2012-08-02 05:22:00.526392
20345	4	1-5894	20345	19	9	1097	16	2012-08-02 05:22:00.526392
20350	5	1-5136	20349	30	79	2114	43	2012-08-02 05:22:00.526392
20351	5	1-5595	20349	79	49	1565	29	2012-08-02 05:22:00.526392
20352	5	1-4553	20349	49	79	641	97	2012-08-02 05:22:00.526392
20353	5	1-5845	20349	79	19	366	30	2012-08-02 05:22:00.526392
20349	5	1-5894	20349	19	30	5348	16	2012-08-02 05:22:00.526392
20355	5	1-1221	20354	25	84	236	24	2012-08-02 05:22:00.946358
20356	5	1-5636	20354	84	84	566	36	2012-08-02 05:22:00.946358
20357	5	1-5678	20354	84	42	750	86	2012-08-02 05:22:00.946358
20358	5	1-5482	20354	42	61	1216	70	2012-08-02 05:22:00.946358
20354	5	1-5895	20354	61	25	891	98	2012-08-02 05:22:00.946358
20360	3	1-5742	20359	21	88	2878	67	2012-08-02 05:22:03.354865
20361	3	1-5493	20359	88	42	921	34	2012-08-02 05:22:03.354865
20359	3	1-5901	20359	42	21	8545	90	2012-08-02 05:22:03.354865
20363	4	1-4875	20362	12	36	1843	25	2012-08-02 05:22:04.138898
20364	4	1-5123	20362	36	90	813	2	2012-08-02 05:22:04.138898
20365	4	1-5430	20362	90	59	497	71	2012-08-02 05:22:04.138898
20362	4	1-5903	20362	59	12	1161	96	2012-08-02 05:22:04.138898
20367	5	1-5698	20366	42	94	1217	50	2012-08-02 05:22:04.138898
20368	5	1-5705	20366	94	83	1017	60	2012-08-02 05:22:04.138898
20369	5	1-933	20366	83	90	881	2	2012-08-02 05:22:04.138898
20370	5	1-5430	20366	90	59	540	71	2012-08-02 05:22:04.138898
20366	5	1-5903	20366	59	42	1263	96	2012-08-02 05:22:04.138898
20372	5	1-4409	20371	3	51	2289	20	2012-08-02 05:22:04.138898
20373	5	1-5462	20371	51	67	1368	18	2012-08-02 05:22:04.138898
20374	5	1-5756	20371	67	46	1440	38	2012-08-02 05:22:04.138898
20375	5	1-5249	20371	46	59	820	71	2012-08-02 05:22:04.138898
20371	5	1-5903	20371	59	3	1969	96	2012-08-02 05:22:04.138898
20377	4	1-5353	20376	47	89	1876	78	2012-08-02 05:22:04.138898
20378	4	1-4504	20376	89	90	1519	2	2012-08-02 05:22:04.138898
20379	4	1-5430	20376	90	59	999	71	2012-08-02 05:22:04.138898
20376	4	1-5903	20376	59	47	2510	96	2012-08-02 05:22:04.138898
20381	4	1-5845	20380	79	51	620	30	2012-08-02 05:22:04.956708
20382	4	1-5857	20380	51	96	1352	10	2012-08-02 05:22:04.956708
20383	4	1-5294	20380	96	93	1573	76	2012-08-02 05:22:04.956708
20380	4	1-5905	20380	93	79	1028	97	2012-08-02 05:22:04.956708
20385	5	1-3937	20384	45	82	1091	14	2012-08-02 05:22:05.374907
20386	5	1-5764	20384	82	23	1133	84	2012-08-02 05:22:05.374907
20387	5	1-5437	20384	23	82	661	70	2012-08-02 05:22:05.374907
20388	5	1-5643	20384	82	99	553	41	2012-08-02 05:22:05.374907
20384	5	1-5906	20384	99	45	1011	23	2012-08-02 05:22:05.374907
20390	4	1-2805	20389	3	5	338	71	2012-08-02 05:22:05.374907
20391	4	1-5577	20389	5	82	664	70	2012-08-02 05:22:05.374907
20392	4	1-5643	20389	82	99	524	41	2012-08-02 05:22:05.374907
20389	4	1-5906	20389	99	3	967	23	2012-08-02 05:22:05.374907
20394	5	1-5012	20393	33	27	5249	57	2012-08-02 05:22:05.374907
20395	5	1-5791	20393	27	3	1561	75	2012-08-02 05:22:05.374907
20396	5	1-4333	20393	3	2	2840	62	2012-08-02 05:22:05.374907
20397	5	1-5674	20393	2	99	2265	41	2012-08-02 05:22:05.374907
20393	5	1-5906	20393	99	33	4655	23	2012-08-02 05:22:05.374907
20399	5	1-5911	20398	94	9	1254	68	2012-08-02 05:22:08.691983
20400	5	1-5329	20398	9	66	889	87	2012-08-02 05:22:08.691983
20401	5	1-5843	20398	66	62	649	20	2012-08-02 05:22:08.691983
20402	5	1-5458	20398	62	37	680	12	2012-08-02 05:22:08.691983
20398	5	1-5914	20398	37	94	1345	27	2012-08-02 05:22:08.691983
20404	4	1-5532	20403	13	33	5125	76	2012-08-02 05:22:09.100844
20405	4	1-5348	20403	33	92	1777	66	2012-08-02 05:22:09.100844
20406	4	1-3260	20403	92	84	1225	88	2012-08-02 05:22:09.100844
20403	4	1-5915	20403	84	13	4938	25	2012-08-02 05:22:09.100844
20408	5	1-5042	20407	80	94	3306	27	2012-08-02 05:22:11.995275
20409	5	1-5911	20407	94	46	3034	68	2012-08-02 05:22:11.995275
20410	5	1-4003	20407	46	76	655	24	2012-08-02 05:22:11.995275
20411	5	1-5597	20407	76	6	2067	58	2012-08-02 05:22:11.995275
20407	5	1-5922	20407	6	80	4823	29	2012-08-02 05:22:11.995275
20413	5	1-5917	20412	76	20	5676	91	2012-08-02 05:22:12.823799
20414	5	1-3675	20412	20	74	2731	61	2012-08-02 05:22:12.823799
20415	5	1-5470	20412	74	29	3137	68	2012-08-02 05:22:12.823799
20416	5	1-5752	20412	29	5	1312	66	2012-08-02 05:22:12.823799
20412	5	1-5924	20412	5	76	3474	93	2012-08-02 05:22:12.823799
20418	5	1-5917	20417	76	67	97	91	2012-08-02 05:22:12.823799
20419	5	1-3121	20417	67	82	96	39	2012-08-02 05:22:12.823799
20420	5	1-3864	20417	82	36	48	85	2012-08-02 05:22:12.823799
20421	5	1-5564	20417	36	5	19	66	2012-08-02 05:22:12.823799
20417	5	1-5924	20417	5	76	54	93	2012-08-02 05:22:12.823799
20423	5	1-5919	20422	55	80	953	86	2012-08-02 05:22:12.823799
20424	5	1-1812	20422	80	42	683	32	2012-08-02 05:22:12.823799
20425	5	1-5518	20422	42	36	1199	85	2012-08-02 05:22:12.823799
20426	5	1-5564	20422	36	5	496	66	2012-08-02 05:22:12.823799
20422	5	1-5924	20422	5	55	1509	93	2012-08-02 05:22:12.823799
20428	5	1-4283	20427	73	94	17	61	2012-08-02 05:22:12.823799
20429	5	1-5846	20427	94	66	11	75	2012-08-02 05:22:12.823799
20430	5	1-5817	20427	66	33	53	76	2012-08-02 05:22:12.823799
20431	5	1-5348	20427	33	5	21	66	2012-08-02 05:22:12.823799
20427	5	1-5924	20427	5	73	68	93	2012-08-02 05:22:12.823799
20433	5	1-5919	20432	55	86	3054	86	2012-08-02 05:22:12.823799
20434	5	1-4236	20432	86	23	2821	72	2012-08-02 05:22:12.823799
20435	5	1-5039	20432	23	29	2447	68	2012-08-02 05:22:12.823799
20436	5	1-5752	20432	29	5	1301	66	2012-08-02 05:22:12.823799
20432	5	1-5924	20432	5	55	4375	93	2012-08-02 05:22:12.823799
20438	4	1-3425	20437	70	90	1946	11	2012-08-02 05:22:14.083462
20439	4	1-5728	20437	90	61	1525	55	2012-08-02 05:22:14.083462
20440	4	1-1385	20437	61	7	1048	40	2012-08-02 05:22:14.083462
20437	4	1-5927	20437	7	70	6371	54	2012-08-02 05:22:14.083462
20442	5	1-5272	20441	82	29	150	74	2012-08-02 05:22:14.083462
20443	5	1-5187	20441	29	91	161	22	2012-08-02 05:22:14.083462
20444	5	1-2919	20441	91	6	44	1	2012-08-02 05:22:14.083462
20445	5	1-3945	20441	6	7	74	40	2012-08-02 05:22:14.083462
20441	5	1-5927	20441	7	82	534	54	2012-08-02 05:22:14.083462
20447	4	1-5817	20446	66	37	258	76	2012-08-02 05:22:14.503512
20448	4	1-4977	20446	37	100	90	95	2012-08-02 05:22:14.503512
20449	4	1-5531	20446	100	8	101	43	2012-08-02 05:22:14.503512
20446	4	1-5928	20446	8	66	64	75	2012-08-02 05:22:14.503512
20451	5	1-4333	20450	3	82	2922	62	2012-08-02 05:22:14.503512
20452	5	1-4474	20450	82	23	1851	20	2012-08-02 05:22:14.503512
20453	5	1-5833	20450	23	100	1342	95	2012-08-02 05:22:14.503512
20454	5	1-5531	20450	100	8	1929	43	2012-08-02 05:22:14.503512
20450	5	1-5928	20450	8	3	1574	75	2012-08-02 05:22:14.503512
20456	5	1-5620	20455	69	92	2332	28	2012-08-02 05:22:15.321079
20457	5	1-4722	20455	92	6	985	55	2012-08-02 05:22:15.321079
20458	5	1-5374	20455	6	84	1231	34	2012-08-02 05:22:15.321079
20459	5	1-5676	20455	84	51	1006	81	2012-08-02 05:22:15.321079
20455	5	1-5930	20455	51	69	2439	59	2012-08-02 05:22:15.321079
20461	3	1-39	20460	21	68	815	52	2012-08-02 05:22:15.751915
20462	3	1-5921	20460	68	83	930	53	2012-08-02 05:22:15.751915
20460	3	1-5931	20460	83	21	4166	49	2012-08-02 05:22:15.751915
20464	5	1-5897	20463	78	50	3402	10	2012-08-02 05:22:17.022646
20465	5	1-5888	20463	50	99	3350	65	2012-08-02 05:22:17.022646
20466	5	1-5622	20463	99	20	2087	83	2012-08-02 05:22:17.022646
20467	5	1-5797	20463	20	84	2350	89	2012-08-02 05:22:17.022646
20463	5	1-5934	20463	84	78	4526	96	2012-08-02 05:22:17.022646
20469	5	1-4673	20468	22	90	1327	49	2012-08-02 05:22:17.022646
20470	5	1-5721	20468	90	12	1033	95	2012-08-02 05:22:17.022646
20471	5	1-5418	20468	12	20	687	83	2012-08-02 05:22:17.022646
20472	5	1-5797	20468	20	84	783	89	2012-08-02 05:22:17.022646
20468	5	1-5934	20468	84	22	1526	96	2012-08-02 05:22:17.022646
20474	4	1-5396	20473	27	21	1387	3	2012-08-02 05:22:18.679652
20475	4	1-4115	20473	21	55	949	47	2012-08-02 05:22:18.679652
20476	4	1-4565	20473	55	84	1534	52	2012-08-02 05:22:18.679652
20473	4	1-5938	20473	84	27	2176	22	2012-08-02 05:22:18.679652
20478	4	1-5636	20477	84	46	2230	36	2012-08-02 05:22:20.756743
20479	4	1-5884	20477	46	59	3464	28	2012-08-02 05:22:20.756743
20480	4	1-5449	20477	59	2	1032	1	2012-08-02 05:22:20.756743
20477	4	1-5943	20477	2	84	1876	24	2012-08-02 05:22:20.756743
20482	5	1-5636	20481	84	84	864	36	2012-08-02 05:22:20.756743
20483	5	1-5678	20481	84	86	940	86	2012-08-02 05:22:20.756743
20484	5	1-4236	20481	86	82	508	72	2012-08-02 05:22:20.756743
20485	5	1-4641	20481	82	2	195	1	2012-08-02 05:22:20.756743
20481	5	1-5943	20481	2	84	439	24	2012-08-02 05:22:20.756743
20487	3	1-5636	20486	84	44	559	36	2012-08-02 05:22:20.756743
20488	3	1-3616	20486	44	2	151	1	2012-08-02 05:22:20.756743
20486	3	1-5943	20486	2	84	360	24	2012-08-02 05:22:20.756743
20490	3	1-5597	20489	76	84	3329	58	2012-08-02 05:22:20.756743
20491	3	1-3621	20489	84	2	418	1	2012-08-02 05:22:20.756743
20489	3	1-5943	20489	2	76	1292	24	2012-08-02 05:22:20.756743
20493	5	1-3895	20492	54	93	1028	20	2012-08-02 05:22:20.756743
20494	5	1-5823	20492	93	44	700	35	2012-08-02 05:22:20.756743
20495	5	1-3333	20492	44	58	292	100	2012-08-02 05:22:20.756743
20496	5	1-5718	20492	58	2	281	1	2012-08-02 05:22:20.756743
20492	5	1-5943	20492	2	54	921	24	2012-08-02 05:22:20.756743
20498	4	1-3270	20497	49	10	1099	42	2012-08-02 05:22:20.756743
20499	4	1-5384	20497	10	58	626	100	2012-08-02 05:22:20.756743
20500	4	1-5718	20497	58	2	619	1	2012-08-02 05:22:20.756743
20497	4	1-5943	20497	2	49	2085	24	2012-08-02 05:22:20.756743
20502	5	1-5859	20501	96	21	745	3	2012-08-02 05:22:21.18795
20503	5	1-4115	20501	21	55	475	47	2012-08-02 05:22:21.18795
20504	5	1-4565	20501	55	84	715	52	2012-08-02 05:22:21.18795
20505	5	1-5938	20501	84	85	946	22	2012-08-02 05:22:21.18795
20501	5	1-5944	20501	85	96	608	94	2012-08-02 05:22:21.18795
20507	4	1-5674	20506	2	6	1285	41	2012-08-02 05:22:22.016445
20508	4	1-5180	20506	6	63	1303	67	2012-08-02 05:22:22.016445
20509	4	1-5666	20506	63	86	553	25	2012-08-02 05:22:22.016445
20506	4	1-5946	20506	86	2	2171	62	2012-08-02 05:22:22.016445
20511	4	1-4474	20510	82	51	409	20	2012-08-02 05:22:22.016445
20512	4	1-5462	20510	51	8	209	18	2012-08-02 05:22:22.016445
20513	4	1-3677	20510	8	86	195	25	2012-08-02 05:22:22.016445
20510	4	1-5946	20510	86	82	824	62	2012-08-02 05:22:22.016445
20515	5	1-5831	20514	21	37	543	12	2012-08-02 05:22:22.016445
20516	5	1-5914	20514	37	94	903	27	2012-08-02 05:22:22.016445
20517	5	1-5911	20514	94	75	708	68	2012-08-02 05:22:22.016445
20518	5	1-5655	20514	75	86	339	25	2012-08-02 05:22:22.016445
20514	5	1-5946	20514	86	21	1470	62	2012-08-02 05:22:22.016445
20520	5	1-4861	20519	18	53	1392	79	2012-08-02 05:22:22.016445
20521	5	1-2946	20519	53	39	556	7	2012-08-02 05:22:22.016445
20522	5	1-5370	20519	39	75	1013	68	2012-08-02 05:22:22.016445
20523	5	1-5655	20519	75	86	595	25	2012-08-02 05:22:22.016445
20519	5	1-5946	20519	86	18	3176	62	2012-08-02 05:22:22.016445
20525	4	1-5678	20524	84	86	349	86	2012-08-02 05:22:22.856415
20526	4	1-4236	20524	86	23	304	72	2012-08-02 05:22:22.856415
20527	4	1-5039	20524	23	51	248	68	2012-08-02 05:22:22.856415
20524	4	1-5948	20524	51	84	267	36	2012-08-02 05:22:22.856415
20529	5	1-2662	20528	8	59	196	71	2012-08-02 05:22:22.856415
20530	5	1-5903	20528	59	74	497	96	2012-08-02 05:22:22.856415
20531	5	1-5818	20528	74	74	260	61	2012-08-02 05:22:22.856415
20532	5	1-5470	20528	74	51	392	68	2012-08-02 05:22:22.856415
20528	5	1-5948	20528	51	8	454	36	2012-08-02 05:22:22.856415
20534	5	1-5099	20533	48	79	591	97	2012-08-02 05:22:23.275992
20535	5	1-5845	20533	79	65	424	30	2012-08-02 05:22:23.275992
20536	5	1-5609	20533	65	78	893	33	2012-08-02 05:22:23.275992
20537	5	1-5571	20533	78	30	1551	80	2012-08-02 05:22:23.275992
20533	5	1-5949	20533	30	48	1224	100	2012-08-02 05:22:23.275992
20539	5	1-5721	20538	90	100	30	95	2012-08-02 05:22:24.104661
20540	5	1-5531	20538	100	8	42	43	2012-08-02 05:22:24.104661
20541	5	1-5928	20538	8	85	33	75	2012-08-02 05:22:24.104661
20542	5	1-4034	20538	85	34	65	17	2012-08-02 05:22:24.104661
20538	5	1-5951	20538	34	90	38	49	2012-08-02 05:22:24.104661
20544	4	1-4751	20543	94	7	276	40	2012-08-02 05:22:24.104661
20545	4	1-5927	20543	7	1	2005	54	2012-08-02 05:22:24.104661
20546	4	1-5785	20543	1	34	1717	17	2012-08-02 05:22:24.104661
20543	4	1-5951	20543	34	94	1018	49	2012-08-02 05:22:24.104661
20548	5	1-5530	20547	83	70	455	26	2012-08-02 05:22:25.764045
20549	5	1-5121	20547	70	45	583	88	2012-08-02 05:22:25.764045
20550	5	1-5438	20547	45	75	464	68	2012-08-02 05:22:25.764045
20551	5	1-5655	20547	75	80	258	25	2012-08-02 05:22:25.764045
20547	5	1-5955	20547	80	83	254	98	2012-08-02 05:22:25.764045
20553	5	1-5572	20552	5	94	544	13	2012-08-02 05:22:29.497884
20554	5	1-5403	20552	94	94	364	27	2012-08-02 05:22:29.497884
20555	5	1-5911	20552	94	29	310	68	2012-08-02 05:22:29.497884
20556	5	1-5752	20552	29	35	132	66	2012-08-02 05:22:29.497884
20552	5	1-5964	20552	35	5	150	26	2012-08-02 05:22:29.497884
20558	4	1-4909	20557	93	33	983	10	2012-08-02 05:22:30.73537
20559	4	1-5750	20557	33	53	1498	44	2012-08-02 05:22:30.73537
20560	4	1-4965	20557	53	73	1533	14	2012-08-02 05:22:30.73537
20557	4	1-5967	20557	73	93	975	29	2012-08-02 05:22:30.73537
20562	3	1-5948	20561	51	35	3441	36	2012-08-02 05:22:31.132961
20563	3	1-4032	20561	35	37	1386	88	2012-08-02 05:22:31.132961
20561	3	1-5968	20561	37	51	4998	68	2012-08-02 05:22:31.132961
20565	3	1-5752	20564	29	92	1293	66	2012-08-02 05:22:31.132961
20566	3	1-3260	20564	92	37	764	88	2012-08-02 05:22:31.132961
20564	3	1-5968	20564	37	29	3331	68	2012-08-02 05:22:31.132961
20568	4	1-5803	20567	15	31	3890	4	2012-08-02 05:22:32.381622
20569	4	1-5041	20567	31	30	2125	90	2012-08-02 05:22:32.381622
20570	4	1-5528	20567	30	13	1850	87	2012-08-02 05:22:32.381622
20567	4	1-5971	20567	13	15	4324	79	2012-08-02 05:22:32.381622
20572	4	1-2946	20571	53	39	881	7	2012-08-02 05:22:32.381622
20573	4	1-5370	20571	39	9	1478	68	2012-08-02 05:22:32.381622
20574	4	1-5329	20571	9	13	995	87	2012-08-02 05:22:32.381622
20571	4	1-5971	20571	13	53	2398	79	2012-08-02 05:22:32.381622
20576	3	1-3981	20575	81	87	77	40	2012-08-02 05:22:32.381622
20577	3	1-5060	20575	87	13	42	87	2012-08-02 05:22:32.381622
20575	3	1-5971	20575	13	81	105	79	2012-08-02 05:22:32.381622
20579	4	1-5729	20578	27	53	2710	73	2012-08-02 05:22:32.790361
20580	4	1-5856	20578	53	51	577	98	2012-08-02 05:22:32.790361
20581	4	1-1354	20578	51	11	386	42	2012-08-02 05:22:32.790361
20578	4	1-5972	20578	11	27	1021	23	2012-08-02 05:22:32.790361
20583	2	1-5883	20582	85	11	1943	42	2012-08-02 05:22:32.790361
20582	2	1-5972	20582	11	85	5452	23	2012-08-02 05:22:32.790361
20585	4	1-5729	20584	27	50	2964	73	2012-08-02 05:22:32.790361
20586	4	1-2413	20584	50	7	1315	50	2012-08-02 05:22:32.790361
20587	4	1-5147	20584	7	11	355	42	2012-08-02 05:22:32.790361
20584	4	1-5972	20584	11	27	1025	23	2012-08-02 05:22:32.790361
20589	4	1-3732	20588	39	87	661	40	2012-08-02 05:22:32.790361
20590	4	1-5060	20588	87	36	332	87	2012-08-02 05:22:32.790361
20591	4	1-5702	20588	36	11	307	42	2012-08-02 05:22:32.790361
20588	4	1-5972	20588	11	39	946	23	2012-08-02 05:22:32.790361
20593	2	1-5572	20592	5	44	1659	13	2012-08-02 05:22:33.210406
20592	2	1-5973	20592	44	5	8645	26	2012-08-02 05:22:33.210406
20595	5	1-5890	20594	9	88	1061	79	2012-08-02 05:22:34.028107
20596	5	1-5398	20594	88	58	731	100	2012-08-02 05:22:34.028107
20594	5	1-5975	20594	93	9	806	16	2012-08-02 05:22:34.028107
20597	5	1-5718	20594	58	6	796	1	2012-08-02 05:22:34.028107
20598	5	1-3945	20594	6	93	1261	40	2012-08-02 05:22:34.028107
20600	5	1-5179	20599	59	3	1025	82	2012-08-02 05:22:34.447693
20601	5	1-5814	20599	3	68	1640	49	2012-08-02 05:22:34.447693
20602	5	1-5632	20599	68	80	1406	33	2012-08-02 05:22:34.447693
20603	5	1-5203	20599	80	28	3054	46	2012-08-02 05:22:34.447693
20599	5	1-5976	20599	28	59	2567	1	2012-08-02 05:22:34.447693
20605	5	1-4868	20604	78	88	378	40	2012-08-02 05:22:34.845457
20606	5	1-5393	20604	88	14	389	15	2012-08-02 05:22:34.845457
20607	5	1-4996	20604	14	82	435	14	2012-08-02 05:22:34.845457
20608	5	1-5764	20604	82	33	424	84	2012-08-02 05:22:34.845457
20604	5	1-5977	20604	33	78	476	39	2012-08-02 05:22:34.845457
20610	3	1-5662	20609	86	33	5394	84	2012-08-02 05:22:35.243312
20611	3	1-5977	20609	33	40	4820	39	2012-08-02 05:22:35.243312
20609	3	1-5978	20609	40	86	4682	82	2012-08-02 05:22:35.243312
20613	4	1-5717	20612	8	100	529	64	2012-08-02 05:22:35.243312
20614	4	1-4881	20612	100	27	949	10	2012-08-02 05:22:35.243312
20615	4	1-5391	20612	27	40	515	39	2012-08-02 05:22:35.243312
20612	4	1-5978	20612	40	8	558	82	2012-08-02 05:22:35.243312
20617	4	1-5810	20616	88	12	1648	8	2012-08-02 05:22:35.243312
20618	4	1-5933	20616	12	67	1511	91	2012-08-02 05:22:35.243312
20619	4	1-3121	20616	67	40	1659	39	2012-08-02 05:22:35.243312
20616	4	1-5978	20616	40	88	2096	82	2012-08-02 05:22:35.243312
20621	4	1-5524	20620	55	59	30	82	2012-08-02 05:22:37.331277
20622	4	1-5319	20620	59	82	21	70	2012-08-02 05:22:37.331277
20623	4	1-5643	20620	82	52	15	41	2012-08-02 05:22:37.331277
20620	4	1-5983	20620	52	55	73	31	2012-08-02 05:22:37.331277
20625	5	1-5965	20624	74	66	3153	87	2012-08-02 05:22:37.331277
20626	5	1-5843	20624	66	51	2287	20	2012-08-02 05:22:37.331277
20627	5	1-5462	20624	51	93	1415	18	2012-08-02 05:22:37.331277
20628	5	1-3342	20624	93	52	1277	41	2012-08-02 05:22:37.331277
20624	5	1-5983	20624	52	74	7535	31	2012-08-02 05:22:37.331277
20630	4	1-5524	20629	55	3	528	82	2012-08-02 05:22:37.331277
20631	4	1-5814	20629	3	4	890	49	2012-08-02 05:22:37.331277
20632	4	1-4425	20629	4	52	168	41	2012-08-02 05:22:37.331277
20629	4	1-5983	20629	52	55	1027	31	2012-08-02 05:22:37.331277
20634	2	1-1938	20633	4	27	3189	14	2012-08-02 05:22:37.751465
20633	2	1-5984	20633	27	4	3226	96	2012-08-02 05:22:37.751465
20636	2	1-1141	20635	10	27	2166	14	2012-08-02 05:22:37.751465
20635	2	1-5984	20635	27	10	2601	96	2012-08-02 05:22:37.751465
20638	5	1-2998	20637	1	29	1369	68	2012-08-02 05:22:38.160119
20639	5	1-5752	20637	29	35	529	66	2012-08-02 05:22:38.160119
20640	5	1-5964	20637	35	55	548	26	2012-08-02 05:22:38.160119
20641	5	1-5782	20637	55	25	1937	58	2012-08-02 05:22:38.160119
20637	5	1-5985	20637	25	1	4078	3	2012-08-02 05:22:38.160119
20643	5	1-4929	20642	72	51	3012	20	2012-08-02 05:22:38.160119
20644	5	1-5462	20642	51	67	1518	18	2012-08-02 05:22:38.160119
20645	5	1-5756	20642	67	90	1346	38	2012-08-02 05:22:38.160119
20646	5	1-5234	20642	90	25	1874	58	2012-08-02 05:22:38.160119
20642	5	1-5985	20642	25	72	4393	3	2012-08-02 05:22:38.160119
20648	5	1-1508	20647	31	53	397	73	2012-08-02 05:22:38.160119
20649	5	1-5856	20647	53	83	90	98	2012-08-02 05:22:38.160119
20650	5	1-5530	20647	83	55	137	26	2012-08-02 05:22:38.160119
20651	5	1-5782	20647	55	25	554	58	2012-08-02 05:22:38.160119
20647	5	1-5985	20647	25	31	1335	3	2012-08-02 05:22:38.160119
20652	1	1-5462	\N	51	51	1690	18	2012-08-02 05:22:38.160119
20654	4	1-5438	20653	45	9	125	68	2012-08-02 05:22:39.430825
20655	4	1-5329	20653	9	32	66	87	2012-08-02 05:22:39.430825
20656	4	1-5641	20653	32	15	94	63	2012-08-02 05:22:39.430825
20653	4	1-5988	20653	15	45	208	88	2012-08-02 05:22:39.430825
20658	4	1-5173	20657	99	81	1647	79	2012-08-02 05:22:39.8396
20659	4	1-3981	20657	81	93	1258	40	2012-08-02 05:22:39.8396
20660	4	1-5975	20657	93	80	856	16	2012-08-02 05:22:39.8396
20657	4	1-5989	20657	80	99	1492	87	2012-08-02 05:22:39.8396
20662	4	1-5530	20661	83	55	2051	26	2012-08-02 05:22:41.099334
20663	4	1-5782	20661	55	84	5009	58	2012-08-02 05:22:41.099334
20664	4	1-5256	20661	84	59	409	34	2012-08-02 05:22:41.099334
20661	4	1-5992	20661	59	83	2228	98	2012-08-02 05:22:41.099334
20666	4	1-5765	20665	69	66	490	73	2012-08-02 05:22:41.099334
20667	4	1-706	20665	66	6	91	55	2012-08-02 05:22:41.099334
20668	4	1-5374	20665	6	59	65	34	2012-08-02 05:22:41.099334
20665	4	1-5992	20665	59	69	411	98	2012-08-02 05:22:41.099334
20670	4	1-5530	20669	83	19	1834	26	2012-08-02 05:22:41.099334
20671	4	1-5175	20669	19	17	675	74	2012-08-02 05:22:41.099334
20672	4	1-5408	20669	17	59	224	34	2012-08-02 05:22:41.099334
20669	4	1-5992	20669	59	83	1558	98	2012-08-02 05:22:41.099334
20674	5	1-3794	20673	42	95	170	25	2012-08-02 05:22:41.099334
20675	5	1-5600	20673	95	97	118	91	2012-08-02 05:22:41.099334
20676	5	1-5401	20673	97	72	70	8	2012-08-02 05:22:41.099334
20677	5	1-5798	20673	72	59	116	34	2012-08-02 05:22:41.099334
20673	5	1-5992	20673	59	42	906	98	2012-08-02 05:22:41.099334
20679	4	1-5214	20678	2	14	2216	8	2012-08-02 05:22:41.938874
20680	4	1-5556	20678	14	69	2525	59	2012-08-02 05:22:41.938874
20681	4	1-5620	20678	69	54	2337	28	2012-08-02 05:22:41.938874
20678	4	1-5994	20678	54	2	3177	45	2012-08-02 05:22:41.938874
20683	5	1-5118	20682	84	82	354	14	2012-08-02 05:22:41.938874
20684	5	1-5764	20682	82	32	377	84	2012-08-02 05:22:41.938874
20685	5	1-5873	20682	32	46	64	36	2012-08-02 05:22:41.938874
20686	5	1-5884	20682	46	54	212	28	2012-08-02 05:22:41.938874
20682	5	1-5994	20682	54	84	295	45	2012-08-02 05:22:41.938874
20688	5	1-5118	20687	84	73	3282	14	2012-08-02 05:22:41.938874
20689	5	1-5967	20687	73	6	2093	29	2012-08-02 05:22:41.938874
20690	5	1-3050	20687	6	46	591	36	2012-08-02 05:22:41.938874
20691	5	1-5884	20687	46	54	1949	28	2012-08-02 05:22:41.938874
20687	5	1-5994	20687	54	84	2723	45	2012-08-02 05:22:41.938874
20693	4	1-5676	20692	84	95	1801	81	2012-08-02 05:22:43.618306
20694	4	1-3808	20692	95	3	1137	21	2012-08-02 05:22:43.618306
20695	4	1-5879	20692	3	13	924	3	2012-08-02 05:22:43.618306
20692	4	1-5998	20692	13	84	2231	34	2012-08-02 05:22:43.618306
20697	5	1-5179	20696	59	3	919	82	2012-08-02 05:22:44.447082
20698	5	1-5814	20696	3	68	1515	49	2012-08-02 05:22:44.447082
20699	5	1-5632	20696	68	78	1339	33	2012-08-02 05:22:44.447082
20700	5	1-5571	20696	78	97	2239	80	2012-08-02 05:22:44.447082
20696	5	1-6000	20696	97	59	2233	1	2012-08-02 05:22:44.447082
20702	5	1-5877	20701	7	25	2238	50	2012-08-02 05:22:44.447082
20703	5	1-3651	20701	25	80	522	25	2012-08-02 05:22:44.447082
20704	5	1-5955	20701	80	62	533	98	2012-08-02 05:22:44.447082
20705	5	1-4999	20701	62	97	1473	80	2012-08-02 05:22:44.447082
20701	5	1-6000	20701	97	7	1487	1	2012-08-02 05:22:44.447082
20707	5	1-4909	20706	93	33	1413	10	2012-08-02 05:22:44.85593
20708	5	1-5750	20706	33	42	2034	44	2012-08-02 05:22:44.85593
20709	5	1-3793	20706	42	60	617	21	2012-08-02 05:22:44.85593
20710	5	1-3903	20706	60	87	212	12	2012-08-02 05:22:44.85593
20706	5	1-6001	20706	87	93	1484	29	2012-08-02 05:22:44.85593
20712	5	1-4553	20711	49	45	2304	97	2012-08-02 05:22:44.85593
20713	5	1-5638	20711	45	95	2442	81	2012-08-02 05:22:44.85593
20714	5	1-3808	20711	95	60	1502	21	2012-08-02 05:22:44.85593
20715	5	1-3903	20711	60	87	530	12	2012-08-02 05:22:44.85593
20711	5	1-6001	20711	87	49	3803	29	2012-08-02 05:22:44.85593
20717	5	1-5815	20716	82	94	218	50	2012-08-02 05:22:45.275967
20718	5	1-5705	20716	94	48	162	60	2012-08-02 05:22:45.275967
20719	5	1-5258	20716	48	23	117	72	2012-08-02 05:22:45.275967
20720	5	1-5039	20716	23	71	85	68	2012-08-02 05:22:45.275967
20716	5	1-6002	20716	71	82	79	83	2012-08-02 05:22:45.275967
20722	3	1-3210	20721	19	7	331	81	2012-08-02 05:22:47.761713
20723	3	1-5822	20721	7	18	264	92	2012-08-02 05:22:47.761713
20721	3	1-6008	20721	18	19	1881	8	2012-08-02 05:22:47.761713
20725	4	1-5798	20724	72	84	4137	34	2012-08-02 05:22:47.761713
20726	4	1-5676	20724	84	7	866	81	2012-08-02 05:22:47.761713
20727	4	1-5822	20724	7	18	821	92	2012-08-02 05:22:47.761713
20724	4	1-6008	20724	18	72	6960	8	2012-08-02 05:22:47.761713
20729	5	1-6000	20728	97	27	3498	1	2012-08-02 05:22:48.601506
20730	5	1-3010	20728	27	93	2727	40	2012-08-02 05:22:48.601506
20731	5	1-5975	20728	93	9	1407	16	2012-08-02 05:22:48.601506
20732	5	1-5890	20728	9	85	1497	79	2012-08-02 05:22:48.601506
20728	5	1-6010	20728	85	97	4517	80	2012-08-02 05:22:48.601506
20734	3	1-5949	20733	30	59	1006	100	2012-08-02 05:22:48.601506
20735	3	1-5006	20733	59	85	386	79	2012-08-02 05:22:48.601506
20733	3	1-6010	20733	85	30	1409	80	2012-08-02 05:22:48.601506
20737	4	1-6009	20736	96	10	347	42	2012-08-02 05:22:48.601506
20738	4	1-5384	20736	10	59	222	100	2012-08-02 05:22:48.601506
20739	4	1-5006	20736	59	85	89	79	2012-08-02 05:22:48.601506
20736	4	1-6010	20736	85	96	339	80	2012-08-02 05:22:48.601506
20740	1	1-1697	\N	68	68	977	40	2012-08-02 05:22:48.601506
20741	1	1-3010	\N	27	27	6786	40	2012-08-02 05:22:48.601506
20743	3	1-3016	20742	71	92	620	17	2012-08-02 05:22:49.452457
20744	3	1-5270	20742	92	41	283	21	2012-08-02 05:22:49.452457
20742	3	1-6012	20742	41	71	1155	89	2012-08-02 05:22:49.452457
20746	3	1-5296	20745	42	25	8473	2	2012-08-02 05:22:49.452457
20747	3	1-5802	20745	25	41	1700	21	2012-08-02 05:22:49.452457
20745	3	1-6012	20745	41	42	7172	89	2012-08-02 05:22:49.452457
20749	4	1-5335	20748	61	91	474	69	2012-08-02 05:22:49.452457
20750	4	1-5650	20748	91	92	260	17	2012-08-02 05:22:49.452457
20751	4	1-5270	20748	92	41	130	21	2012-08-02 05:22:49.452457
20748	4	1-6012	20748	41	61	586	89	2012-08-02 05:22:49.452457
20753	5	1-5908	20752	64	28	1757	92	2012-08-02 05:22:49.452457
20754	5	1-3410	20752	28	25	271	61	2012-08-02 05:22:49.452457
20755	5	1-5926	20752	25	82	291	82	2012-08-02 05:22:49.452457
20756	5	1-4402	20752	82	41	175	21	2012-08-02 05:22:49.452457
20752	5	1-6012	20752	41	64	897	89	2012-08-02 05:22:49.452457
20758	5	1-5518	20757	42	44	2467	85	2012-08-02 05:22:50.305201
20759	5	1-5529	20757	44	42	1710	44	2012-08-02 05:22:50.305201
20760	5	1-3793	20757	42	19	401	21	2012-08-02 05:22:50.305201
20761	5	1-5138	20757	19	18	1535	37	2012-08-02 05:22:50.305201
20757	5	1-6014	20757	18	42	1737	32	2012-08-02 05:22:50.305201
20763	3	1-5518	20762	42	60	2244	85	2012-08-02 05:22:50.305201
20764	3	1-4982	20762	60	18	1540	37	2012-08-02 05:22:50.305201
20762	3	1-6014	20762	18	42	1987	32	2012-08-02 05:22:50.305201
20766	4	1-4143	20765	50	7	597	81	2012-08-02 05:22:50.305201
20767	4	1-5822	20765	7	6	1276	92	2012-08-02 05:22:50.305201
20768	4	1-2734	20765	6	18	732	37	2012-08-02 05:22:50.305201
20765	4	1-6014	20765	18	50	946	32	2012-08-02 05:22:50.305201
20770	3	1-5869	20769	18	83	741	3	2012-08-02 05:22:50.305201
20771	3	1-4694	20769	83	18	1517	37	2012-08-02 05:22:50.305201
20769	3	1-6014	20769	18	18	2179	32	2012-08-02 05:22:50.305201
20773	4	1-6001	20772	87	78	2682	29	2012-08-02 05:22:50.747215
20774	4	1-4783	20772	78	53	1386	85	2012-08-02 05:22:50.747215
20775	4	1-5771	20772	53	11	528	54	2012-08-02 05:22:50.747215
20772	4	1-6015	20772	11	87	579	12	2012-08-02 05:22:50.747215
20777	4	1-5134	20776	17	59	1822	95	2012-08-02 05:22:50.747215
20778	4	1-4882	20776	59	47	1972	62	2012-08-02 05:22:50.747215
20779	4	1-5567	20776	47	11	1876	54	2012-08-02 05:22:50.747215
20776	4	1-6015	20776	11	17	2818	12	2012-08-02 05:22:50.747215
20781	5	1-4486	20780	64	27	3998	78	2012-08-02 05:22:50.747215
20782	5	1-5918	20780	27	54	940	24	2012-08-02 05:22:50.747215
20783	5	1-3895	20780	54	30	1137	20	2012-08-02 05:22:50.747215
20784	5	1-5714	20780	30	11	2260	54	2012-08-02 05:22:50.747215
20780	5	1-6015	20780	11	64	3655	12	2012-08-02 05:22:50.747215
20786	5	1-5995	20785	77	94	2369	61	2012-08-02 05:22:53.255377
20787	5	1-5846	20785	94	30	1136	75	2012-08-02 05:22:53.255377
20788	5	1-4581	20785	30	75	647	97	2012-08-02 05:22:53.255377
20789	5	1-5181	20785	75	46	2663	37	2012-08-02 05:22:53.255377
20785	5	1-6021	20785	46	77	3289	5	2012-08-02 05:22:53.255377
20791	4	1-5995	20790	77	74	1133	61	2012-08-02 05:22:53.255377
20792	4	1-5470	20790	74	42	1369	68	2012-08-02 05:22:53.255377
20793	4	1-4827	20790	42	46	958	37	2012-08-02 05:22:53.255377
20790	4	1-6021	20790	46	77	1364	5	2012-08-02 05:22:53.255377
20795	5	1-5871	20794	72	3	1881	6	2012-08-02 05:22:53.66409
20796	5	1-5766	20794	3	78	2180	38	2012-08-02 05:22:53.66409
20797	5	1-5731	20794	78	67	6931	69	2012-08-02 05:22:53.66409
20798	5	1-2285	20794	67	10	1292	94	2012-08-02 05:22:53.66409
20794	5	1-6022	20794	10	72	4362	24	2012-08-02 05:22:53.66409
20800	5	1-4948	20799	17	83	25	60	2012-08-02 05:22:53.66409
20801	5	1-5669	20799	83	29	13	74	2012-08-02 05:22:53.66409
20802	5	1-5187	20799	29	85	11	22	2012-08-02 05:22:53.66409
20803	5	1-5944	20799	85	10	6	94	2012-08-02 05:22:53.66409
20799	5	1-6022	20799	10	17	21	24	2012-08-02 05:22:53.66409
20805	5	1-5543	20804	66	38	1434	11	2012-08-02 05:22:53.66409
20806	5	1-5821	20804	38	84	1373	52	2012-08-02 05:22:53.66409
20807	5	1-5938	20804	84	85	1885	22	2012-08-02 05:22:53.66409
20808	5	1-5944	20804	85	10	1258	94	2012-08-02 05:22:53.66409
20804	5	1-6022	20804	10	66	5228	24	2012-08-02 05:22:53.66409
20810	3	1-5765	20809	69	51	4988	73	2012-08-02 05:22:56.172295
20811	3	1-2185	20809	51	95	450	90	2012-08-02 05:22:56.172295
20809	3	1-6028	20809	95	69	5048	98	2012-08-02 05:22:56.172295
20813	4	1-3237	20812	14	70	103	22	2012-08-02 05:22:56.172295
20814	4	1-4258	20812	70	35	30	60	2012-08-02 05:22:56.172295
20815	4	1-5359	20812	35	95	20	90	2012-08-02 05:22:56.172295
20812	4	1-6028	20812	95	14	282	98	2012-08-02 05:22:56.172295
20817	5	1-5909	20816	89	95	370	43	2012-08-02 05:22:58.635536
20818	5	1-5720	20816	95	85	322	79	2012-08-02 05:22:58.635536
20819	5	1-6010	20816	85	96	945	80	2012-08-02 05:22:58.635536
20820	5	1-6009	20816	96	29	747	42	2012-08-02 05:22:58.635536
20816	5	1-6034	20816	29	89	497	19	2012-08-02 05:22:58.635536
20822	2	1-5929	20821	18	27	2317	31	2012-08-02 05:23:00.337027
20821	2	1-6038	20821	27	18	6962	42	2012-08-02 05:23:00.337027
20824	4	1-5976	20823	28	56	55	1	2012-08-02 05:23:00.767776
20825	4	1-5947	20823	56	52	24	41	2012-08-02 05:23:00.767776
20826	4	1-5983	20823	52	14	130	31	2012-08-02 05:23:00.767776
20823	4	1-6039	20823	14	28	69	46	2012-08-02 05:23:00.767776
20828	4	1-5963	20827	10	85	3012	9	2012-08-02 05:23:02.049405
20829	4	1-4297	20827	85	85	3111	95	2012-08-02 05:23:02.049405
20830	4	1-1463	20827	85	34	751	23	2012-08-02 05:23:02.049405
20827	4	1-6042	20827	34	10	4880	8	2012-08-02 05:23:02.049405
20832	5	1-5750	20831	33	42	226	44	2012-08-02 05:23:02.46912
20833	5	1-3793	20831	42	3	74	21	2012-08-02 05:23:02.46912
20834	5	1-5879	20831	3	83	61	3	2012-08-02 05:23:02.46912
20835	5	1-4694	20831	83	56	317	37	2012-08-02 05:23:02.46912
20831	5	1-6043	20831	56	33	146	10	2012-08-02 05:23:02.46912
20837	4	1-5903	20836	59	47	1344	96	2012-08-02 05:23:02.878025
20838	4	1-5353	20836	47	36	829	78	2012-08-02 05:23:02.878025
20839	4	1-5554	20836	36	73	486	23	2012-08-02 05:23:02.878025
20836	4	1-6044	20836	73	59	648	71	2012-08-02 05:23:02.878025
20841	5	1-6034	20840	29	27	4332	19	2012-08-02 05:23:03.286719
20842	5	1-5639	20840	27	79	5233	30	2012-08-02 05:23:03.286719
20843	5	1-3222	20840	79	1	1533	36	2012-08-02 05:23:03.286719
20844	5	1-5259	20840	1	86	4776	37	2012-08-02 05:23:03.286719
20840	5	1-6045	20840	86	29	6960	42	2012-08-02 05:23:03.286719
20846	4	1-6034	20845	29	90	692	19	2012-08-02 05:23:03.286719
20847	4	1-4187	20845	90	8	408	26	2012-08-02 05:23:03.286719
20848	4	1-4462	20845	8	86	710	37	2012-08-02 05:23:03.286719
20845	4	1-6045	20845	86	29	1073	42	2012-08-02 05:23:03.286719
20850	4	1-5661	20849	56	81	2612	51	2012-08-02 05:23:03.286719
20851	4	1-3265	20849	81	45	293	3	2012-08-02 05:23:03.286719
20852	4	1-4340	20849	45	86	777	37	2012-08-02 05:23:03.286719
20849	4	1-6045	20849	86	56	1461	42	2012-08-02 05:23:03.286719
20854	4	1-1851	20853	31	9	230	16	2012-08-02 05:23:05.762544
20855	4	1-5890	20853	9	15	327	79	2012-08-02 05:23:05.762544
20856	4	1-5803	20853	15	34	333	4	2012-08-02 05:23:05.762544
20853	4	1-6051	20853	34	31	294	22	2012-08-02 05:23:05.762544
20858	5	1-6044	20857	73	45	3131	71	2012-08-02 05:23:06.160818
20859	5	1-5077	20857	45	12	625	8	2012-08-02 05:23:06.160818
20860	5	1-5933	20857	12	1	541	91	2012-08-02 05:23:06.160818
20861	5	1-5996	20857	1	60	1185	11	2012-08-02 05:23:06.160818
20857	5	1-6052	20857	60	73	2166	23	2012-08-02 05:23:06.160818
20863	5	1-3745	20862	69	30	93	46	2012-08-02 05:23:06.160818
20864	5	1-6007	20862	30	95	28	25	2012-08-02 05:23:06.160818
20865	5	1-5600	20862	95	1	26	91	2012-08-02 05:23:06.160818
20866	5	1-5996	20862	1	60	62	11	2012-08-02 05:23:06.160818
20862	5	1-6052	20862	60	69	122	23	2012-08-02 05:23:06.160818
20868	5	1-6051	20867	34	85	144	22	2012-08-02 05:23:06.956293
20869	5	1-5944	20867	85	96	106	94	2012-08-02 05:23:06.956293
20870	5	1-5859	20867	96	45	148	3	2012-08-02 05:23:06.956293
20871	5	1-4340	20867	45	61	442	37	2012-08-02 05:23:06.956293
20867	5	1-6054	20867	61	34	165	4	2012-08-02 05:23:06.956293
20873	3	1-5826	20872	30	2	949	59	2012-08-02 05:23:07.342959
20874	3	1-4723	20872	2	55	61	53	2012-08-02 05:23:07.342959
20872	3	1-6055	20872	55	30	662	56	2012-08-02 05:23:07.342959
20876	4	1-4370	20875	43	58	2585	100	2012-08-02 05:23:07.342959
20877	4	1-5718	20875	58	98	1377	1	2012-08-02 05:23:07.342959
20878	4	1-5555	20875	98	55	636	53	2012-08-02 05:23:07.342959
20875	4	1-6055	20875	55	43	7252	56	2012-08-02 05:23:07.342959
20880	4	1-6005	20879	82	27	697	96	2012-08-02 05:23:07.342959
20881	4	1-6025	20879	27	98	336	82	2012-08-02 05:23:07.342959
20882	4	1-2571	20879	98	55	92	53	2012-08-02 05:23:07.342959
20879	4	1-6055	20879	55	82	1191	56	2012-08-02 05:23:07.342959
20884	5	1-5920	20883	58	44	2642	9	2012-08-02 05:23:08.569649
20885	5	1-3782	20883	44	96	1315	39	2012-08-02 05:23:08.569649
20886	5	1-5605	20883	96	35	812	60	2012-08-02 05:23:08.569649
20887	5	1-5359	20883	35	91	562	90	2012-08-02 05:23:08.569649
20883	5	1-6058	20883	91	58	6724	78	2012-08-02 05:23:08.569649
20889	5	1-5594	20888	93	15	487	29	2012-08-02 05:23:09.354225
20890	5	1-5786	20888	15	82	456	83	2012-08-02 05:23:09.354225
20891	5	1-5815	20888	82	94	1465	50	2012-08-02 05:23:09.354225
20892	5	1-5705	20888	94	4	1265	60	2012-08-02 05:23:09.354225
20888	5	1-6060	20888	4	93	1169	70	2012-08-02 05:23:09.354225
20894	4	1-4167	20893	100	94	1546	50	2012-08-02 05:23:09.763065
20895	4	1-5705	20893	94	40	1312	60	2012-08-02 05:23:09.763065
20896	4	1-6020	20893	40	93	2203	59	2012-08-02 05:23:09.763065
20893	4	1-6061	20893	93	100	967	45	2012-08-02 05:23:09.763065
20898	5	1-5676	20897	84	95	794	81	2012-08-02 05:23:10.16052
20899	5	1-3808	20897	95	65	465	21	2012-08-02 05:23:10.16052
20900	5	1-5907	20897	65	57	441	96	2012-08-02 05:23:10.16052
20901	5	1-4318	20897	57	85	254	56	2012-08-02 05:23:10.16052
20897	5	1-6062	20897	85	84	1060	34	2012-08-02 05:23:10.16052
20903	3	1-5600	20902	95	13	3585	91	2012-08-02 05:23:11.762746
20904	3	1-5733	20902	13	19	5021	37	2012-08-02 05:23:11.762746
20902	3	1-6066	20902	19	95	5127	25	2012-08-02 05:23:11.762746
20906	4	1-5955	20905	80	44	262	98	2012-08-02 05:23:11.762746
20907	4	1-5848	20905	44	60	384	85	2012-08-02 05:23:11.762746
20908	4	1-4982	20905	60	19	337	37	2012-08-02 05:23:11.762746
20905	4	1-6066	20905	19	80	351	25	2012-08-02 05:23:11.762746
20910	5	1-3943	20909	28	30	5955	46	2012-08-02 05:23:12.171373
20911	5	1-6007	20909	30	12	1777	25	2012-08-02 05:23:12.171373
20912	5	1-4759	20909	12	36	1813	85	2012-08-02 05:23:12.171373
20913	5	1-5564	20909	36	67	787	66	2012-08-02 05:23:12.171373
20909	5	1-6067	20909	67	28	7471	71	2012-08-02 05:23:12.171373
20915	5	1-5964	20914	35	5	1015	26	2012-08-02 05:23:14.126826
20916	5	1-5572	20914	5	1	3349	13	2012-08-02 05:23:14.126826
20917	5	1-5913	20914	1	9	1229	16	2012-08-02 05:23:14.126826
20918	5	1-5890	20914	9	55	1220	79	2012-08-02 05:23:14.126826
20914	5	1-6072	20914	55	35	979	66	2012-08-02 05:23:14.126826
20920	4	1-5964	20919	35	5	193	26	2012-08-02 05:23:14.126826
20921	4	1-5572	20919	5	9	790	13	2012-08-02 05:23:14.126826
20922	4	1-4478	20919	9	55	152	79	2012-08-02 05:23:14.126826
20919	4	1-6072	20919	55	35	151	66	2012-08-02 05:23:14.126826
20924	5	1-3260	20923	92	34	1229	88	2012-08-02 05:23:14.126826
20925	5	1-6019	20923	34	100	849	95	2012-08-02 05:23:14.126826
20926	5	1-5531	20923	100	95	1176	43	2012-08-02 05:23:14.126826
20927	5	1-5720	20923	95	55	1357	79	2012-08-02 05:23:14.126826
20923	5	1-6072	20923	55	92	1507	66	2012-08-02 05:23:14.126826
20929	4	1-5938	20928	84	31	1382	22	2012-08-02 05:23:14.535686
20930	4	1-1851	20928	31	44	707	16	2012-08-02 05:23:14.535686
20931	4	1-5900	20928	44	54	854	51	2012-08-02 05:23:14.535686
20928	4	1-6073	20928	54	84	1377	52	2012-08-02 05:23:14.535686
20933	4	1-3558	20932	88	90	1388	2	2012-08-02 05:23:14.535686
20934	4	1-5430	20932	90	51	839	71	2012-08-02 05:23:14.535686
20935	4	1-5767	20932	51	54	673	51	2012-08-02 05:23:14.535686
20932	4	1-6073	20932	54	88	1465	52	2012-08-02 05:23:14.535686
20937	4	1-5648	20936	7	35	1563	20	2012-08-02 05:23:14.535686
20938	4	1-4683	20936	35	44	688	16	2012-08-02 05:23:14.535686
20939	4	1-5900	20936	44	54	1129	51	2012-08-02 05:23:14.535686
20936	4	1-6073	20936	54	7	2476	52	2012-08-02 05:23:14.535686
20941	4	1-5528	20940	30	32	1545	87	2012-08-02 05:23:15.344229
20942	4	1-5641	20940	32	43	2838	63	2012-08-02 05:23:15.344229
20943	4	1-5306	20940	43	76	577	9	2012-08-02 05:23:15.344229
20940	4	1-6075	20940	76	30	1712	90	2012-08-02 05:23:15.344229
20945	5	1-5507	20944	83	21	3598	49	2012-08-02 05:23:16.15105
20946	5	1-5862	20944	21	42	3406	58	2012-08-02 05:23:16.15105
20947	5	1-5189	20944	42	1	2367	28	2012-08-02 05:23:16.15105
20948	5	1-5434	20944	1	58	2006	48	2012-08-02 05:23:16.15105
20944	5	1-6077	20944	58	83	2304	77	2012-08-02 05:23:16.15105
20950	4	1-5955	20949	80	95	2967	98	2012-08-02 05:23:17.333397
20951	4	1-5774	20949	95	80	1178	16	2012-08-02 05:23:17.333397
20952	4	1-5989	20949	80	23	1825	87	2012-08-02 05:23:17.333397
20949	4	1-6080	20949	23	80	3232	25	2012-08-02 05:23:17.333397
20954	3	1-5364	20953	41	80	548	16	2012-08-02 05:23:17.333397
20955	3	1-5989	20953	80	23	793	87	2012-08-02 05:23:17.333397
20953	3	1-6080	20953	23	41	1503	25	2012-08-02 05:23:17.333397
20957	4	1-5578	20956	69	46	1030	36	2012-08-02 05:23:18.924053
20958	4	1-5884	20956	46	3	1590	28	2012-08-02 05:23:18.924053
20959	4	1-5334	20956	3	34	841	63	2012-08-02 05:23:18.924053
20956	4	1-6084	20956	34	69	4560	97	2012-08-02 05:23:18.924053
20961	5	1-5638	20960	45	51	1252	81	2012-08-02 05:23:18.924053
20962	5	1-5930	20960	51	93	1449	59	2012-08-02 05:23:18.924053
20963	5	1-6061	20960	93	46	322	45	2012-08-02 05:23:18.924053
20964	5	1-5460	20960	46	34	428	63	2012-08-02 05:23:18.924053
20960	5	1-6084	20960	34	45	2379	97	2012-08-02 05:23:18.924053
20966	4	1-5389	20965	31	55	57	79	2012-08-02 05:23:18.924053
20967	4	1-6072	20965	55	87	31	66	2012-08-02 05:23:18.924053
20968	4	1-6006	20965	87	34	24	63	2012-08-02 05:23:18.924053
20965	4	1-6084	20965	34	31	141	97	2012-08-02 05:23:18.924053
20970	5	1-5964	20969	35	50	1964	26	2012-08-02 05:23:19.333043
20971	5	1-3683	20969	50	77	1243	5	2012-08-02 05:23:19.333043
20972	5	1-5995	20969	77	74	1144	61	2012-08-02 05:23:19.333043
20973	5	1-5470	20969	74	4	1533	68	2012-08-02 05:23:19.333043
20969	5	1-6085	20969	4	35	1503	66	2012-08-02 05:23:19.333043
20975	4	1-5481	20974	30	15	983	63	2012-08-02 05:23:20.526512
20976	4	1-5988	20974	15	69	1858	88	2012-08-02 05:23:20.526512
20977	4	1-5659	20974	69	50	702	49	2012-08-02 05:23:20.526512
20974	4	1-6088	20974	50	30	3340	58	2012-08-02 05:23:20.526512
20979	4	1-3603	20978	80	73	877	48	2012-08-02 05:23:21.741402
20980	4	1-5757	20978	73	80	399	89	2012-08-02 05:23:21.741402
20981	4	1-3390	20978	80	85	177	72	2012-08-02 05:23:21.741402
20978	4	1-6091	20978	85	80	2400	78	2012-08-02 05:23:21.741402
20983	4	1-5554	20982	36	84	3901	23	2012-08-02 05:23:21.741402
20984	4	1-3038	20982	84	58	1136	12	2012-08-02 05:23:21.741402
20985	4	1-5076	20982	58	85	468	72	2012-08-02 05:23:21.741402
20982	4	1-6091	20982	85	36	6928	78	2012-08-02 05:23:21.741402
20987	4	1-5418	20986	12	82	1446	83	2012-08-02 05:23:22.150458
20988	4	1-5815	20986	82	94	3493	50	2012-08-02 05:23:22.150458
20989	4	1-5705	20986	94	96	2269	60	2012-08-02 05:23:22.150458
20986	4	1-6092	20986	96	12	2997	95	2012-08-02 05:23:22.150458
20991	5	1-5125	20990	14	5	47	80	2012-08-02 05:23:22.150458
20992	5	1-6053	20990	5	10	76	92	2012-08-02 05:23:22.150458
20993	5	1-6056	20990	10	94	70	50	2012-08-02 05:23:22.150458
20994	5	1-5705	20990	94	96	63	60	2012-08-02 05:23:22.150458
20990	5	1-6092	20990	96	14	115	95	2012-08-02 05:23:22.150458
20996	5	1-5911	20995	94	4	432	68	2012-08-02 05:23:22.968002
20997	5	1-6085	20995	4	35	420	66	2012-08-02 05:23:22.968002
20998	5	1-5964	20995	35	8	543	26	2012-08-02 05:23:22.968002
20999	5	1-4462	20995	8	53	1166	37	2012-08-02 05:23:22.968002
20995	5	1-6094	20995	53	94	510	27	2012-08-02 05:23:22.968002
21001	5	1-4442	21000	10	12	2763	25	2012-08-02 05:23:24.912574
21002	5	1-4759	21000	12	36	2931	85	2012-08-02 05:23:24.912574
21003	5	1-5564	21000	36	74	1322	66	2012-08-02 05:23:24.912574
21004	5	1-5704	21000	74	57	2151	17	2012-08-02 05:23:24.912574
21000	5	1-6099	21000	57	10	2358	47	2012-08-02 05:23:24.912574
21006	4	1-5805	21005	36	55	95	79	2012-08-02 05:23:24.912574
21007	4	1-6072	21005	55	74	107	66	2012-08-02 05:23:24.912574
21008	4	1-5704	21005	74	57	179	17	2012-08-02 05:23:24.912574
21005	4	1-6099	21005	57	36	203	47	2012-08-02 05:23:24.912574
21010	4	1-5586	21009	8	10	5044	92	2012-08-02 05:23:24.912574
21011	4	1-6056	21009	10	39	4851	50	2012-08-02 05:23:24.912574
21012	4	1-5316	21009	39	57	1922	17	2012-08-02 05:23:24.912574
21009	4	1-6099	21009	57	8	2235	47	2012-08-02 05:23:24.912574
21014	5	1-5318	21013	20	98	3572	17	2012-08-02 05:23:25.310338
21015	5	1-5703	21013	98	61	3092	45	2012-08-02 05:23:25.310338
21016	5	1-5036	21013	61	88	1277	82	2012-08-02 05:23:25.310338
21017	5	1-5810	21013	88	69	938	8	2012-08-02 05:23:25.310338
21013	5	1-6100	21013	69	20	1030	99	2012-08-02 05:23:25.310338
21019	5	1-5318	21018	20	34	2485	17	2012-08-02 05:23:25.310338
21020	5	1-5951	21018	34	68	1362	49	2012-08-02 05:23:25.310338
21021	5	1-5632	21018	68	82	1141	33	2012-08-02 05:23:25.310338
21022	5	1-5644	21018	82	69	566	8	2012-08-02 05:23:25.310338
21018	5	1-6100	21018	69	20	668	99	2012-08-02 05:23:25.310338
21024	5	1-5473	21023	9	61	9129	3	2012-08-02 05:23:25.310338
21025	5	1-6081	21023	61	27	5944	96	2012-08-02 05:23:25.310338
21026	5	1-6025	21023	27	88	5198	82	2012-08-02 05:23:25.310338
21027	5	1-5810	21023	88	69	4184	8	2012-08-02 05:23:25.310338
21023	5	1-6100	21023	69	9	5037	99	2012-08-02 05:23:25.310338
21029	3	1-5821	21028	38	2	2985	52	2012-08-02 05:23:29.321029
21030	3	1-5423	21028	2	56	2056	7	2012-08-02 05:23:29.321029
21028	3	1-6110	21028	56	38	3265	11	2012-08-02 05:23:29.321029
21032	5	1-4450	21031	59	58	1921	78	2012-08-02 05:23:32.1491
21033	5	1-5920	21031	58	76	1065	9	2012-08-02 05:23:32.1491
21034	5	1-6075	21031	76	30	3023	90	2012-08-02 05:23:32.1491
21035	5	1-5528	21031	30	63	2611	87	2012-08-02 05:23:32.1491
21031	5	1-6117	21031	63	59	4189	71	2012-08-02 05:23:32.1491
21037	4	1-2542	21036	67	93	1647	40	2012-08-02 05:23:32.1491
21038	4	1-5975	21036	93	80	1082	16	2012-08-02 05:23:32.1491
21039	4	1-5989	21036	80	63	1823	87	2012-08-02 05:23:32.1491
21036	4	1-6117	21036	63	67	3176	71	2012-08-02 05:23:32.1491
21041	4	1-3489	21040	19	8	830	43	2012-08-02 05:23:32.546825
21042	4	1-5928	21040	8	66	486	75	2012-08-02 05:23:32.546825
21043	4	1-5840	21040	66	90	1193	85	2012-08-02 05:23:32.546825
21040	4	1-6118	21040	90	19	9177	15	2012-08-02 05:23:32.546825
21045	3	1-1409	21044	10	57	1149	17	2012-08-02 05:23:33.331153
21046	3	1-6099	21044	57	54	764	47	2012-08-02 05:23:33.331153
21044	3	1-6120	21044	54	10	1298	38	2012-08-02 05:23:33.331153
21048	5	1-5080	21047	99	50	14	28	2012-08-02 05:23:33.331153
21049	5	1-5160	21047	50	1	8	54	2012-08-02 05:23:33.331153
21050	5	1-5785	21047	1	57	6	17	2012-08-02 05:23:33.331153
21051	5	1-6099	21047	57	54	6	47	2012-08-02 05:23:33.331153
21047	5	1-6120	21047	54	99	15	38	2012-08-02 05:23:33.331153
21053	5	1-5261	21052	14	96	1707	23	2012-08-02 05:23:33.331153
21054	5	1-6098	21052	96	1	2296	54	2012-08-02 05:23:33.331153
21055	5	1-5785	21052	1	57	1688	17	2012-08-02 05:23:33.331153
21056	5	1-6099	21052	57	54	1660	47	2012-08-02 05:23:33.331153
21052	5	1-6120	21052	54	14	4174	38	2012-08-02 05:23:33.331153
21058	5	1-2873	21057	85	3	1420	5	2012-08-02 05:23:33.331153
21059	5	1-5738	21057	3	1	1182	54	2012-08-02 05:23:33.331153
21060	5	1-5785	21057	1	57	875	17	2012-08-02 05:23:33.331153
21061	5	1-6099	21057	57	54	867	47	2012-08-02 05:23:33.331153
21057	5	1-6120	21057	54	85	2196	38	2012-08-02 05:23:33.331153
21063	3	1-6106	21062	15	45	3733	39	2012-08-02 05:23:36.115655
21064	3	1-6096	21062	45	35	1360	29	2012-08-02 05:23:36.115655
21062	3	1-6127	21062	35	15	1960	95	2012-08-02 05:23:36.115655
21066	3	1-6106	21065	15	67	5086	39	2012-08-02 05:23:36.115655
21067	3	1-5849	21065	67	35	1828	29	2012-08-02 05:23:36.115655
21065	3	1-6127	21065	35	15	2652	95	2012-08-02 05:23:36.115655
21069	3	1-5639	21068	27	51	1174	30	2012-08-02 05:23:36.524562
21070	3	1-5857	21068	51	62	1097	10	2012-08-02 05:23:36.524562
21068	3	1-6128	21068	62	27	1840	19	2012-08-02 05:23:36.524562
21072	3	1-3421	21071	25	51	909	30	2012-08-02 05:23:36.524562
21073	3	1-5857	21071	51	62	1071	10	2012-08-02 05:23:36.524562
21071	3	1-6128	21071	62	25	2262	19	2012-08-02 05:23:36.524562
21075	4	1-5626	21074	95	65	5489	51	2012-08-02 05:23:36.524562
21076	4	1-6083	21074	65	56	6894	37	2012-08-02 05:23:36.524562
21077	4	1-6043	21074	56	62	1773	10	2012-08-02 05:23:36.524562
21074	4	1-6128	21074	62	95	4492	19	2012-08-02 05:23:36.524562
21078	1	1-3421	\N	25	25	65	30	2012-08-02 05:23:36.524562
21080	4	1-5690	21079	37	61	1214	70	2012-08-02 05:23:37.320014
21081	4	1-5895	21079	61	25	861	98	2012-08-02 05:23:37.320014
21082	4	1-1221	21079	25	85	221	24	2012-08-02 05:23:37.320014
21079	4	1-6130	21079	85	37	1781	64	2012-08-02 05:23:37.320014
21084	3	1-5278	21083	8	5	980	82	2012-08-02 05:23:37.718045
21085	3	1-2893	21083	5	14	704	57	2012-08-02 05:23:37.718045
21083	3	1-6131	21083	14	8	3689	4	2012-08-02 05:23:37.718045
21087	3	1-6089	21086	82	36	1278	90	2012-08-02 05:23:37.718045
21088	3	1-3444	21086	36	14	504	57	2012-08-02 05:23:37.718045
21086	3	1-6131	21086	14	82	3370	4	2012-08-02 05:23:37.718045
21090	3	1-2412	21089	99	8	1400	48	2012-08-02 05:23:38.922388
21091	3	1-5819	21089	8	53	2119	31	2012-08-02 05:23:38.922388
21089	3	1-6134	21089	53	99	4173	93	2012-08-02 05:23:38.922388
21093	5	1-5388	21092	26	77	2004	5	2012-08-02 05:23:38.922388
21094	5	1-5995	21092	77	28	1706	61	2012-08-02 05:23:38.922388
21095	5	1-5483	21092	28	8	895	48	2012-08-02 05:23:38.922388
21096	5	1-5819	21092	8	53	1433	31	2012-08-02 05:23:38.922388
21092	5	1-6134	21092	53	26	2986	93	2012-08-02 05:23:38.922388
21098	5	1-4812	21097	25	12	394	30	2012-08-02 05:23:38.922388
21099	5	1-5828	21097	12	44	229	94	2012-08-02 05:23:38.922388
21100	5	1-5783	21097	44	8	332	48	2012-08-02 05:23:38.922388
21101	5	1-5819	21097	8	53	571	31	2012-08-02 05:23:38.922388
21097	5	1-6134	21097	53	25	1278	93	2012-08-02 05:23:38.922388
21103	4	1-6074	21102	63	28	4678	46	2012-08-02 05:23:39.717743
21104	4	1-5976	21102	28	92	3788	1	2012-08-02 05:23:39.717743
21105	4	1-5485	21102	92	87	1211	8	2012-08-02 05:23:39.717743
21102	4	1-6136	21102	87	63	3061	17	2012-08-02 05:23:39.717743
21107	5	1-5270	21106	92	65	845	21	2012-08-02 05:23:39.717743
21108	5	1-5907	21106	65	27	826	96	2012-08-02 05:23:39.717743
21109	5	1-6025	21106	27	88	707	82	2012-08-02 05:23:39.717743
21110	5	1-5810	21106	88	87	557	8	2012-08-02 05:23:39.717743
21106	5	1-6136	21106	87	92	1426	17	2012-08-02 05:23:39.717743
21112	4	1-5916	21111	86	69	357	8	2012-08-02 05:23:40.513518
21113	4	1-6100	21111	69	25	379	99	2012-08-02 05:23:40.513518
21114	4	1-5366	21111	25	14	413	6	2012-08-02 05:23:40.513518
21111	4	1-6138	21111	14	86	1191	41	2012-08-02 05:23:40.513518
21115	5	1-6142	21115	32	64	2348	54	2012-08-02 05:23:42.139499
21116	5	1-1515	21115	64	44	228	94	2012-08-02 05:23:42.139499
21117	5	1-5783	21115	44	58	369	48	2012-08-02 05:23:42.139499
21118	5	1-6077	21115	58	92	414	77	2012-08-02 05:23:42.139499
21119	5	1-5855	21115	92	32	400	5	2012-08-02 05:23:42.139499
21121	3	1-5742	21120	21	63	2797	67	2012-08-02 05:23:43.741195
21122	3	1-5666	21120	63	92	1626	25	2012-08-02 05:23:43.741195
21120	3	1-6146	21120	92	21	3943	90	2012-08-02 05:23:43.741195
21124	4	1-5418	21123	12	20	662	83	2012-08-02 05:23:44.150091
21125	4	1-5797	21123	20	80	604	89	2012-08-02 05:23:44.150091
21126	4	1-3390	21123	80	30	245	72	2012-08-02 05:23:44.150091
21123	4	1-6147	21123	30	12	1244	95	2012-08-02 05:23:44.150091
21128	5	1-6105	21127	82	28	335	92	2012-08-02 05:23:44.150091
21129	5	1-3410	21127	28	74	56	61	2012-08-02 05:23:44.150091
21130	5	1-5470	21127	74	8	82	68	2012-08-02 05:23:44.150091
21131	5	1-5363	21127	8	30	59	72	2012-08-02 05:23:44.150091
21127	5	1-6147	21127	30	82	373	95	2012-08-02 05:23:44.150091
21132	1	1-110	\N	45	45	733	72	2012-08-02 05:23:44.150091
21134	4	1-5855	21133	92	32	1243	5	2012-08-02 05:23:45.740812
21135	4	1-6142	21133	32	82	6171	54	2012-08-02 05:23:45.740812
21136	4	1-5272	21133	82	57	1466	74	2012-08-02 05:23:45.740812
21133	4	1-6151	21133	57	92	1519	77	2012-08-02 05:23:45.740812
21138	5	1-5855	21137	92	77	205	5	2012-08-02 05:23:45.740812
21139	5	1-5995	21137	77	5	180	61	2012-08-02 05:23:45.740812
21140	5	1-4498	21137	5	42	73	20	2012-08-02 05:23:45.740812
21141	5	1-6114	21137	42	57	237	74	2012-08-02 05:23:45.740812
21137	5	1-6151	21137	57	92	248	77	2012-08-02 05:23:45.740812
21143	2	1-5443	21142	98	79	411	49	2012-08-02 05:23:46.149732
21142	2	1-6152	21142	79	98	2997	43	2012-08-02 05:23:46.149732
21145	5	1-5959	21144	26	42	899	20	2012-08-02 05:23:47.321085
21146	5	1-6114	21144	42	57	2811	74	2012-08-02 05:23:47.321085
21147	5	1-6151	21144	57	65	2842	77	2012-08-02 05:23:47.321085
21148	5	1-5637	21144	65	79	1514	3	2012-08-02 05:23:47.321085
21144	5	1-6155	21144	79	26	1302	2	2012-08-02 05:23:47.321085
21150	4	1-1989	21149	20	4	413	68	2012-08-02 05:23:49.321171
21151	4	1-6085	21149	4	35	345	66	2012-08-02 05:23:49.321171
21152	4	1-5964	21149	35	5	384	26	2012-08-02 05:23:49.321171
21149	4	1-6160	21149	5	20	5043	5	2012-08-02 05:23:49.321171
21154	5	1-4888	21153	28	87	1838	40	2012-08-02 05:23:49.321171
21155	5	1-5060	21153	87	48	853	87	2012-08-02 05:23:49.321171
21156	5	1-5751	21153	48	35	180	66	2012-08-02 05:23:49.321171
21157	5	1-5964	21153	35	5	217	26	2012-08-02 05:23:49.321171
21153	5	1-6160	21153	5	28	3075	5	2012-08-02 05:23:49.321171
21159	5	1-6136	21158	87	92	1473	17	2012-08-02 05:23:50.13312
21160	5	1-5270	21158	92	65	856	21	2012-08-02 05:23:50.13312
21161	5	1-5907	21158	65	27	821	96	2012-08-02 05:23:50.13312
21162	5	1-6025	21158	27	60	689	82	2012-08-02 05:23:50.13312
21158	5	1-6162	21158	60	87	586	8	2012-08-02 05:23:50.13312
21164	5	1-6100	21163	69	49	265	99	2012-08-02 05:23:50.13312
21165	5	1-3272	21163	49	65	213	21	2012-08-02 05:23:50.13312
21166	5	1-5907	21163	65	27	229	96	2012-08-02 05:23:50.13312
21167	5	1-6025	21163	27	60	215	82	2012-08-02 05:23:50.13312
21163	5	1-6162	21163	60	69	205	8	2012-08-02 05:23:50.13312
21169	4	1-4038	21168	4	46	394	36	2012-08-02 05:23:54.12682
21170	4	1-5884	21168	46	1	1276	28	2012-08-02 05:23:54.12682
21171	4	1-5434	21168	1	6	1019	48	2012-08-02 05:23:54.12682
21168	4	1-6172	21168	6	4	1797	3	2012-08-02 05:23:54.12682
21173	4	1-4115	21172	21	54	555	47	2012-08-02 05:23:54.12682
21174	4	1-6120	21172	54	63	1580	38	2012-08-02 05:23:54.12682
21175	4	1-4414	21172	63	6	438	48	2012-08-02 05:23:54.12682
21172	4	1-6172	21172	6	21	784	3	2012-08-02 05:23:54.12682
21177	5	1-6051	21176	34	31	1449	22	2012-08-02 05:23:55.344443
21178	5	1-1851	21176	31	80	755	16	2012-08-02 05:23:55.344443
21179	5	1-5989	21176	80	63	893	87	2012-08-02 05:23:55.344443
21180	5	1-6117	21176	63	66	1092	71	2012-08-02 05:23:55.344443
21176	5	1-6175	21176	66	34	2454	4	2012-08-02 05:23:55.344443
21182	4	1-6133	21181	45	85	915	95	2012-08-02 05:23:55.344443
21183	4	1-1463	21181	85	73	236	23	2012-08-02 05:23:55.344443
21184	4	1-6044	21181	73	66	352	71	2012-08-02 05:23:55.344443
21181	4	1-6175	21181	66	45	1048	4	2012-08-02 05:23:55.344443
21186	4	1-6138	21185	14	5	5426	41	2012-08-02 05:23:56.164426
21187	4	1-5889	21185	5	53	6162	37	2012-08-02 05:23:56.164426
21188	4	1-6094	21185	53	38	2201	27	2012-08-02 05:23:56.164426
21185	4	1-6177	21185	38	14	2174	6	2012-08-02 05:23:56.164426
21190	4	1-3754	21189	40	1	213	36	2012-08-02 05:23:56.164426
21191	4	1-5259	21189	1	53	832	37	2012-08-02 05:23:56.164426
21192	4	1-6094	21189	53	38	357	27	2012-08-02 05:23:56.164426
21189	4	1-6177	21189	38	40	423	6	2012-08-02 05:23:56.164426
21194	5	1-5766	21193	3	79	1451	38	2012-08-02 05:23:56.164426
21195	5	1-1154	21193	79	73	2241	14	2012-08-02 05:23:56.164426
21196	5	1-5967	21193	73	80	1313	29	2012-08-02 05:23:56.164426
21197	5	1-5042	21193	80	38	843	27	2012-08-02 05:23:56.164426
21193	5	1-6177	21193	38	3	1034	6	2012-08-02 05:23:56.164426
21199	5	1-5766	21198	3	3	1558	38	2012-08-02 05:23:56.164426
21200	5	1-3487	21198	3	28	1416	46	2012-08-02 05:23:56.164426
21201	5	1-5976	21198	28	29	1135	1	2012-08-02 05:23:56.164426
21202	5	1-5285	21198	29	38	839	27	2012-08-02 05:23:56.164426
21198	5	1-6177	21198	38	3	1037	6	2012-08-02 05:23:56.164426
21204	5	1-5468	21203	66	65	159	77	2012-08-02 05:23:56.164426
21205	5	1-5637	21203	65	21	99	3	2012-08-02 05:23:56.164426
21206	5	1-4115	21203	21	61	69	47	2012-08-02 05:23:56.164426
21207	5	1-5383	21203	61	38	117	27	2012-08-02 05:23:56.164426
21203	5	1-6177	21203	38	66	153	6	2012-08-02 05:23:56.164426
21209	5	1-3765	21208	4	73	1305	56	2012-08-02 05:23:56.164426
21210	5	1-3811	21208	73	73	2457	14	2012-08-02 05:23:56.164426
21211	5	1-5967	21208	73	80	1595	29	2012-08-02 05:23:56.164426
21212	5	1-5042	21208	80	38	1106	27	2012-08-02 05:23:56.164426
21208	5	1-6177	21208	38	4	1465	6	2012-08-02 05:23:56.164426
21214	2	1-1089	21213	99	60	2121	2	2012-08-02 05:23:56.599925
21213	2	1-6178	21213	60	99	9443	96	2012-08-02 05:23:56.599925
21216	4	1-4324	21215	59	71	2409	68	2012-08-02 05:23:58.986599
21217	4	1-2729	21215	71	66	909	61	2012-08-02 05:23:58.986599
21218	4	1-6182	21215	66	70	1874	86	2012-08-02 05:23:58.986599
21215	4	1-6184	21215	70	59	5810	67	2012-08-02 05:23:58.986599
21220	5	1-1625	21219	71	50	721	10	2012-08-02 05:23:58.986599
21221	5	1-5888	21219	50	67	715	65	2012-08-02 05:23:58.986599
21222	5	1-3999	21219	67	66	211	61	2012-08-02 05:23:58.986599
21223	5	1-6182	21219	66	70	496	86	2012-08-02 05:23:58.986599
21219	5	1-6184	21219	70	71	1748	67	2012-08-02 05:23:58.986599
21225	4	1-6062	21224	85	84	3248	34	2012-08-02 05:23:59.429971
21226	4	1-5676	21224	84	94	2589	81	2012-08-02 05:23:59.429971
21227	4	1-5990	21224	94	3	1673	14	2012-08-02 05:23:59.429971
21224	4	1-6185	21224	3	85	730	56	2012-08-02 05:23:59.429971
21229	2	1-5659	21228	69	58	2909	49	2012-08-02 05:23:59.817307
21228	2	1-6186	21228	58	69	5842	88	2012-08-02 05:23:59.817307
21231	4	1-4289	21230	27	29	1393	74	2012-08-02 05:23:59.817307
21232	4	1-5187	21230	29	31	1455	22	2012-08-02 05:23:59.817307
21233	4	1-5566	21230	31	58	1264	49	2012-08-02 05:23:59.817307
21230	4	1-6186	21230	58	27	3024	88	2012-08-02 05:23:59.817307
21235	4	1-5926	21234	25	60	2685	82	2012-08-02 05:24:00.215601
21236	4	1-6162	21234	60	69	1731	8	2012-08-02 05:24:00.215601
21237	4	1-6100	21234	69	56	1514	99	2012-08-02 05:24:00.215601
21234	4	1-6187	21234	56	25	3253	61	2012-08-02 05:24:00.215601
21239	5	1-6182	21238	66	86	1007	86	2012-08-02 05:24:00.215601
21240	5	1-4236	21238	86	30	728	72	2012-08-02 05:24:00.215601
21241	5	1-6147	21238	30	31	3601	95	2012-08-02 05:24:00.215601
21242	5	1-2045	21238	31	56	231	99	2012-08-02 05:24:00.215601
21238	5	1-6187	21238	56	66	548	61	2012-08-02 05:24:00.215601
21244	5	1-5470	21243	74	8	851	68	2012-08-02 05:24:00.215601
21245	5	1-5363	21243	8	30	554	72	2012-08-02 05:24:00.215601
21246	5	1-6147	21243	30	31	3166	95	2012-08-02 05:24:00.215601
21247	5	1-2045	21243	31	56	235	99	2012-08-02 05:24:00.215601
21243	5	1-6187	21243	56	74	645	61	2012-08-02 05:24:00.215601
21249	3	1-5959	21248	26	42	1751	20	2012-08-02 05:24:01.188856
21250	3	1-6114	21248	42	5	5629	74	2012-08-02 05:24:01.188856
21248	3	1-6189	21248	5	26	2466	2	2012-08-02 05:24:01.188856
21252	5	1-5620	21251	69	47	1111	28	2012-08-02 05:24:01.586912
21253	5	1-6087	21251	47	90	468	11	2012-08-02 05:24:01.586912
21254	5	1-5728	21251	90	93	230	55	2012-08-02 05:24:01.586912
21255	5	1-5794	21251	93	29	140	46	2012-08-02 05:24:01.586912
21251	5	1-6190	21251	29	69	2170	59	2012-08-02 05:24:01.586912
21257	4	1-4648	21256	61	42	343	20	2012-08-02 05:24:01.586912
21258	4	1-6114	21256	42	73	710	74	2012-08-02 05:24:01.586912
21259	4	1-3858	21256	73	29	115	46	2012-08-02 05:24:01.586912
21256	4	1-6190	21256	29	61	1849	59	2012-08-02 05:24:01.586912
21261	4	1-5365	21260	54	47	914	4	2012-08-02 05:24:01.586912
21262	4	1-4189	21260	47	3	176	38	2012-08-02 05:24:01.586912
21263	4	1-3487	21260	3	29	97	46	2012-08-02 05:24:01.586912
21260	4	1-6190	21260	29	54	1671	59	2012-08-02 05:24:01.586912
21265	4	1-5000	21264	23	71	803	68	2012-08-02 05:24:02.383118
21266	4	1-6002	21264	71	20	835	83	2012-08-02 05:24:02.383118
21267	4	1-5797	21264	20	39	879	89	2012-08-02 05:24:02.383118
21264	4	1-6192	21264	39	23	2657	21	2012-08-02 05:24:02.383118
21269	4	1-5907	21268	65	27	131	96	2012-08-02 05:24:02.383118
21270	4	1-6025	21268	27	42	117	82	2012-08-02 05:24:02.383118
21271	4	1-5519	21268	42	39	40	89	2012-08-02 05:24:02.383118
21268	4	1-6192	21268	39	65	129	21	2012-08-02 05:24:02.383118
21273	5	1-6162	21272	60	12	2678	8	2012-08-02 05:24:05.194268
21274	5	1-5933	21272	12	28	2513	91	2012-08-02 05:24:05.194268
21275	5	1-5589	21272	28	9	2443	9	2012-08-02 05:24:05.194268
21276	5	1-4906	21272	9	49	2869	40	2012-08-02 05:24:05.194268
21272	5	1-6199	21272	49	60	3017	82	2012-08-02 05:24:05.194268
21278	4	1-6162	21277	60	10	3913	8	2012-08-02 05:24:05.194268
21279	4	1-5963	21277	10	9	3357	9	2012-08-02 05:24:05.194268
21280	4	1-4906	21277	9	49	4023	40	2012-08-02 05:24:05.194268
21277	4	1-6199	21277	49	60	4318	82	2012-08-02 05:24:05.194268
21282	4	1-5814	21281	3	47	297	49	2012-08-02 05:24:05.194268
21283	4	1-2610	21281	47	82	98	79	2012-08-02 05:24:05.194268
21284	4	1-6174	21281	82	49	166	40	2012-08-02 05:24:05.194268
21281	4	1-6199	21281	49	3	180	82	2012-08-02 05:24:05.194268
21286	4	1-4087	21285	27	59	205	100	2012-08-02 05:24:05.194268
21287	4	1-5006	21285	59	82	87	79	2012-08-02 05:24:05.194268
21288	4	1-6174	21285	82	49	153	40	2012-08-02 05:24:05.194268
21285	4	1-6199	21285	49	27	173	82	2012-08-02 05:24:05.194268
21290	5	1-3895	21289	54	30	2128	20	2012-08-02 05:24:05.703273
21291	5	1-4824	21289	30	94	775	61	2012-08-02 05:24:05.703273
21292	5	1-5846	21289	94	30	243	75	2012-08-02 05:24:05.703273
21293	5	1-4581	21289	30	46	135	97	2012-08-02 05:24:05.703273
21289	5	1-6200	21289	46	54	3465	24	2012-08-02 05:24:05.703273
21295	4	1-4188	21294	53	16	944	47	2012-08-02 05:24:05.703273
21296	4	1-4967	21294	16	93	703	76	2012-08-02 05:24:05.703273
21297	4	1-5905	21294	93	46	249	97	2012-08-02 05:24:05.703273
21294	4	1-6200	21294	46	53	6439	24	2012-08-02 05:24:05.703273
21299	4	1-5389	21298	31	55	3598	79	2012-08-02 05:24:06.104698
21300	4	1-6149	21298	55	48	5609	11	2012-08-02 05:24:06.104698
21301	4	1-6108	21298	48	76	5050	45	2012-08-02 05:24:06.104698
21298	4	1-6201	21298	76	31	4897	97	2012-08-02 05:24:06.104698
21303	5	1-5352	21302	100	40	189	67	2012-08-02 05:24:06.53598
21304	5	1-5450	21302	40	3	371	14	2012-08-02 05:24:06.53598
21305	5	1-6185	21302	3	96	143	56	2012-08-02 05:24:06.53598
21306	5	1-3902	21302	96	85	166	57	2012-08-02 05:24:06.53598
21302	5	1-6202	21302	85	100	220	77	2012-08-02 05:24:06.53598
21308	4	1-2863	21307	50	3	3500	14	2012-08-02 05:24:06.53598
21309	4	1-6185	21307	3	96	1620	56	2012-08-02 05:24:06.53598
21310	4	1-3902	21307	96	85	2268	57	2012-08-02 05:24:06.53598
21307	4	1-6202	21307	85	50	3621	77	2012-08-02 05:24:06.53598
21312	4	1-4824	21311	30	74	485	61	2012-08-02 05:24:09.806612
21313	4	1-5470	21311	74	89	674	68	2012-08-02 05:24:09.806612
21314	4	1-6104	21311	89	1	382	32	2012-08-02 05:24:09.806612
21311	4	1-6210	21311	1	30	982	20	2012-08-02 05:24:09.806612
21316	5	1-5854	21315	99	93	192	76	2012-08-02 05:24:11.883834
21317	5	1-5905	21315	93	69	125	97	2012-08-02 05:24:11.883834
21318	5	1-5578	21315	69	46	53	36	2012-08-02 05:24:11.883834
21319	5	1-5884	21315	46	47	151	28	2012-08-02 05:24:11.883834
21315	5	1-6215	21315	47	99	498	44	2012-08-02 05:24:11.883834
21321	5	1-5004	21320	17	39	226	89	2012-08-02 05:24:11.883834
21322	5	1-6192	21320	39	40	719	21	2012-08-02 05:24:11.883834
21323	5	1-2529	21320	40	100	282	23	2012-08-02 05:24:11.883834
21324	5	1-5898	21320	100	47	297	28	2012-08-02 05:24:11.883834
21320	5	1-6215	21320	47	17	1090	44	2012-08-02 05:24:11.883834
21326	4	1-4139	21325	52	7	1437	2	2012-08-02 05:24:12.712933
21327	4	1-6166	21325	7	50	1677	10	2012-08-02 05:24:12.712933
21328	4	1-5888	21325	50	78	1292	65	2012-08-02 05:24:12.712933
21325	4	1-6217	21325	78	52	2206	1	2012-08-02 05:24:12.712933
21330	5	1-5957	21329	98	67	2223	60	2012-08-02 05:24:12.712933
21331	5	1-6158	21329	67	30	1643	62	2012-08-02 05:24:12.712933
21332	5	1-6112	21329	30	67	995	75	2012-08-02 05:24:12.712933
21333	5	1-4622	21329	67	78	1358	65	2012-08-02 05:24:12.712933
21329	5	1-6217	21329	78	98	2592	1	2012-08-02 05:24:12.712933
21335	5	1-5042	21334	80	38	434	27	2012-08-02 05:24:13.133182
21336	5	1-6177	21334	38	100	324	6	2012-08-02 05:24:13.133182
21337	5	1-5592	21334	100	85	490	57	2012-08-02 05:24:13.133182
21338	5	1-6202	21334	85	92	421	77	2012-08-02 05:24:13.133182
21334	5	1-6218	21334	92	80	1112	29	2012-08-02 05:24:13.133182
21340	4	1-4553	21339	49	21	29	97	2012-08-02 05:24:13.133182
21341	4	1-4455	21339	21	85	17	57	2012-08-02 05:24:13.133182
21342	4	1-6202	21339	85	92	19	77	2012-08-02 05:24:13.133182
21339	4	1-6218	21339	92	49	64	29	2012-08-02 05:24:13.133182
21344	5	1-4553	21343	49	79	821	97	2012-08-02 05:24:13.133182
21345	5	1-5845	21343	79	48	460	30	2012-08-02 05:24:13.133182
21346	5	1-4941	21343	48	85	337	57	2012-08-02 05:24:13.133182
21347	5	1-6202	21343	85	92	420	77	2012-08-02 05:24:13.133182
21343	5	1-6218	21343	92	49	1613	29	2012-08-02 05:24:13.133182
21349	4	1-3050	21348	6	30	843	36	2012-08-02 05:24:13.133182
21350	4	1-4660	21348	30	85	600	57	2012-08-02 05:24:13.133182
21351	4	1-6202	21348	85	92	813	77	2012-08-02 05:24:13.133182
21348	4	1-6218	21348	92	6	3387	29	2012-08-02 05:24:13.133182
21353	5	1-6127	21352	35	48	1041	95	2012-08-02 05:24:13.133182
21354	5	1-5780	21352	48	44	66	94	2012-08-02 05:24:13.133182
21355	5	1-5783	21352	44	58	100	48	2012-08-02 05:24:13.133182
21356	5	1-6077	21352	58	92	106	77	2012-08-02 05:24:13.133182
21352	5	1-6218	21352	92	35	479	29	2012-08-02 05:24:13.133182
21358	4	1-3611	21357	26	85	1008	57	2012-08-02 05:24:13.984491
21359	4	1-6202	21357	85	92	1528	77	2012-08-02 05:24:13.984491
21360	4	1-5855	21357	92	67	1432	5	2012-08-02 05:24:13.984491
21357	4	1-6220	21357	67	26	4166	81	2012-08-02 05:24:13.984491
21362	4	1-5822	21361	7	71	1001	92	2012-08-02 05:24:13.984491
21363	4	1-2227	21361	71	85	112	38	2012-08-02 05:24:13.984491
21364	4	1-2873	21361	85	67	85	5	2012-08-02 05:24:13.984491
21361	4	1-6220	21361	67	7	261	81	2012-08-02 05:24:13.984491
21366	5	1-6168	21365	34	94	1090	87	2012-08-02 05:24:15.279394
21367	5	1-5870	21365	94	94	707	50	2012-08-02 05:24:15.279394
21368	5	1-5705	21365	94	35	577	60	2012-08-02 05:24:15.279394
21369	5	1-5359	21365	35	38	492	90	2012-08-02 05:24:15.279394
21365	5	1-6223	21365	38	34	738	93	2012-08-02 05:24:15.279394
21371	5	1-6168	21370	34	25	5396	87	2012-08-02 05:24:15.279394
21372	5	1-6018	21370	25	10	904	8	2012-08-02 05:24:15.279394
21373	5	1-5963	21370	10	76	698	9	2012-08-02 05:24:15.279394
21374	5	1-6075	21370	76	38	1937	90	2012-08-02 05:24:15.279394
21370	5	1-6223	21370	38	34	3259	93	2012-08-02 05:24:15.279394
21376	3	1-6165	21375	5	47	2316	92	2012-08-02 05:24:16.108728
21377	3	1-5993	21375	47	68	526	87	2012-08-02 05:24:16.108728
21375	3	1-6225	21375	68	5	761	80	2012-08-02 05:24:16.108728
21379	4	1-6165	21378	5	12	4299	92	2012-08-02 05:24:16.108728
21380	4	1-4665	21378	12	80	533	16	2012-08-02 05:24:16.108728
21381	4	1-5989	21378	80	68	822	87	2012-08-02 05:24:16.108728
21378	4	1-6225	21378	68	5	1297	80	2012-08-02 05:24:16.108728
21383	4	1-6009	21382	96	56	976	42	2012-08-02 05:24:16.108728
21384	4	1-5661	21382	56	76	1816	51	2012-08-02 05:24:16.108728
21385	4	1-3809	21382	76	68	594	87	2012-08-02 05:24:16.108728
21382	4	1-6225	21382	68	96	983	80	2012-08-02 05:24:16.108728
21387	5	1-5354	21386	10	47	4367	4	2012-08-02 05:24:16.517329
21388	5	1-4189	21386	47	24	844	38	2012-08-02 05:24:16.517329
21389	5	1-2979	21386	24	39	348	89	2012-08-02 05:24:16.517329
21390	5	1-6192	21386	39	27	684	21	2012-08-02 05:24:16.517329
21386	5	1-6226	21386	27	10	9385	34	2012-08-02 05:24:16.517329
21392	4	1-6127	21391	35	82	3994	95	2012-08-02 05:24:16.926451
21393	4	1-6105	21391	82	19	3314	92	2012-08-02 05:24:16.926451
21394	4	1-2828	21391	19	27	1701	51	2012-08-02 05:24:16.926451
21391	4	1-6227	21391	27	35	1917	29	2012-08-02 05:24:16.926451
21396	4	1-6171	21395	9	88	2512	85	2012-08-02 05:24:16.926451
21397	4	1-5832	21395	88	79	2159	15	2012-08-02 05:24:16.926451
21398	4	1-5982	21395	79	27	2439	51	2012-08-02 05:24:16.926451
21395	4	1-6227	21395	27	9	2803	29	2012-08-02 05:24:16.926451
21400	4	1-5189	21399	42	47	525	28	2012-08-02 05:24:17.335656
21401	4	1-6215	21399	47	99	1999	44	2012-08-02 05:24:17.335656
21402	4	1-5854	21399	99	57	889	76	2012-08-02 05:24:17.335656
21399	4	1-6228	21399	57	42	782	58	2012-08-02 05:24:17.335656
21404	5	1-3092	21403	31	96	874	39	2012-08-02 05:24:17.755998
21405	5	1-5605	21403	96	35	747	60	2012-08-02 05:24:17.755998
21406	5	1-5359	21403	35	38	717	90	2012-08-02 05:24:17.755998
21407	5	1-6223	21403	38	72	1212	93	2012-08-02 05:24:17.755998
21403	5	1-6229	21403	72	31	1415	26	2012-08-02 05:24:17.755998
21409	4	1-3788	21408	97	78	1712	18	2012-08-02 05:24:17.755998
21410	4	1-5759	21408	78	38	2664	90	2012-08-02 05:24:17.755998
21411	4	1-6223	21408	38	72	4546	93	2012-08-02 05:24:17.755998
21408	4	1-6229	21408	72	97	5359	26	2012-08-02 05:24:17.755998
21413	2	1-1931	21412	52	72	1984	93	2012-08-02 05:24:17.755998
21412	2	1-6229	21412	72	52	2381	26	2012-08-02 05:24:17.755998
21415	5	1-4162	21414	23	29	1364	74	2012-08-02 05:24:18.187209
21416	5	1-5187	21414	29	85	1087	22	2012-08-02 05:24:18.187209
21417	5	1-5944	21414	85	44	589	94	2012-08-02 05:24:18.187209
21418	5	1-5783	21414	44	66	711	48	2012-08-02 05:24:18.187209
21414	5	1-6230	21414	66	23	3829	43	2012-08-02 05:24:18.187209
21420	2	1-5363	21419	8	24	1983	72	2012-08-02 05:24:18.60721
21419	2	1-6231	21419	24	8	5669	68	2012-08-02 05:24:18.60721
21422	5	1-4900	21421	84	90	148	49	2012-08-02 05:24:19.457921
21423	5	1-5721	21421	90	12	100	95	2012-08-02 05:24:19.457921
21424	5	1-5418	21421	12	20	58	83	2012-08-02 05:24:19.457921
21425	5	1-5797	21421	20	3	58	89	2012-08-02 05:24:19.457921
21421	5	1-6233	21421	3	84	243	7	2012-08-02 05:24:19.457921
21427	5	1-6110	21426	56	48	248	11	2012-08-02 05:24:19.457921
21428	5	1-6108	21426	48	99	225	45	2012-08-02 05:24:19.457921
21429	5	1-5149	21426	99	20	34	83	2012-08-02 05:24:19.457921
21430	5	1-5797	21426	20	3	35	89	2012-08-02 05:24:19.457921
21426	5	1-6233	21426	3	56	153	7	2012-08-02 05:24:19.457921
21432	5	1-6110	21431	56	90	2058	11	2012-08-02 05:24:19.457921
21433	5	1-5728	21431	90	70	1889	55	2012-08-02 05:24:19.457921
21434	5	1-4301	21431	70	24	341	38	2012-08-02 05:24:19.457921
21435	5	1-2979	21431	24	3	236	89	2012-08-02 05:24:19.457921
21431	5	1-6233	21431	3	56	1148	7	2012-08-02 05:24:19.457921
21437	4	1-5529	21436	44	64	1696	44	2012-08-02 05:24:19.877509
21438	4	1-5687	21436	64	52	855	42	2012-08-02 05:24:19.877509
21439	4	1-5293	21436	52	80	638	15	2012-08-02 05:24:19.877509
21436	4	1-6234	21436	80	44	1503	85	2012-08-02 05:24:19.877509
21441	5	1-6037	21440	65	85	2486	22	2012-08-02 05:24:19.877509
21442	5	1-5944	21440	85	69	1782	94	2012-08-02 05:24:19.877509
21443	5	1-4399	21440	69	67	869	88	2012-08-02 05:24:19.877509
21444	5	1-6145	21440	67	80	1743	15	2012-08-02 05:24:19.877509
21440	5	1-6234	21440	80	65	4978	85	2012-08-02 05:24:19.877509
21446	4	1-5976	21445	28	92	454	1	2012-08-02 05:24:20.717637
21447	4	1-5485	21445	92	72	157	8	2012-08-02 05:24:20.717637
21448	4	1-5798	21445	72	84	367	34	2012-08-02 05:24:20.717637
21445	4	1-6236	21445	84	28	518	46	2012-08-02 05:24:20.717637
21450	5	1-6112	21449	30	85	1999	75	2012-08-02 05:24:23.180936
21451	5	1-5776	21449	85	93	2124	18	2012-08-02 05:24:23.180936
21452	5	1-3342	21449	93	6	1939	41	2012-08-02 05:24:23.180936
21453	5	1-5180	21449	6	80	2601	67	2012-08-02 05:24:23.180936
21449	5	1-6242	21449	80	30	2844	62	2012-08-02 05:24:23.180936
21455	5	1-6112	21454	30	85	324	75	2012-08-02 05:24:23.180936
21456	5	1-5776	21454	85	78	347	18	2012-08-02 05:24:23.180936
21457	5	1-5759	21454	78	21	603	90	2012-08-02 05:24:23.180936
21458	5	1-5742	21454	21	80	416	67	2012-08-02 05:24:23.180936
21454	5	1-6242	21454	80	30	458	62	2012-08-02 05:24:23.180936
21460	4	1-6242	21459	80	47	3069	62	2012-08-02 05:24:24.009581
21461	4	1-5567	21459	47	1	2063	54	2012-08-02 05:24:24.009581
21462	4	1-5785	21459	1	8	1040	17	2012-08-02 05:24:24.009581
21459	4	1-6244	21459	8	80	4652	67	2012-08-02 05:24:24.009581
21464	3	1-2168	21463	97	20	109	99	2012-08-02 05:24:24.009581
21465	3	1-5318	21463	20	8	262	17	2012-08-02 05:24:24.009581
21463	3	1-6244	21463	8	97	1193	67	2012-08-02 05:24:24.009581
21467	5	1-4417	21466	60	79	1209	15	2012-08-02 05:24:24.009581
21468	5	1-5982	21466	79	29	1121	51	2012-08-02 05:24:24.009581
21469	5	1-3925	21466	29	87	111	8	2012-08-02 05:24:24.009581
21470	5	1-6136	21466	87	8	231	17	2012-08-02 05:24:24.009581
21466	5	1-6244	21466	8	60	1314	67	2012-08-02 05:24:24.009581
21472	4	1-4324	21471	59	16	28	68	2012-08-02 05:24:24.009581
21473	4	1-2286	21471	16	11	19	13	2012-08-02 05:24:24.009581
21474	4	1-5881	21471	11	8	13	17	2012-08-02 05:24:24.009581
21471	4	1-6244	21471	8	59	77	67	2012-08-02 05:24:24.009581
21476	4	1-2168	21475	97	56	70	99	2012-08-02 05:24:24.009581
21477	4	1-6187	21475	56	35	176	61	2012-08-02 05:24:24.009581
21478	4	1-2632	21475	35	8	93	17	2012-08-02 05:24:24.009581
21475	4	1-6244	21475	8	97	571	67	2012-08-02 05:24:24.009581
21479	1	1-2494	\N	44	44	182	15	2012-08-02 05:24:24.009581
21481	4	1-5736	21480	66	38	1038	39	2012-08-02 05:24:24.871937
21482	4	1-4354	21480	38	28	1116	46	2012-08-02 05:24:24.871937
21483	4	1-5976	21480	28	68	931	1	2012-08-02 05:24:24.871937
21480	4	1-6246	21480	68	66	1832	27	2012-08-02 05:24:24.871937
21485	4	1-4951	21484	72	1	571	32	2012-08-02 05:24:25.280997
21486	4	1-6210	21484	1	30	925	20	2012-08-02 05:24:25.280997
21487	4	1-4824	21484	30	66	288	61	2012-08-02 05:24:25.280997
21484	4	1-6247	21484	66	72	1069	24	2012-08-02 05:24:25.280997
21489	4	1-4102	21488	31	65	1031	21	2012-08-02 05:24:25.280997
21490	4	1-5907	21488	65	74	828	96	2012-08-02 05:24:25.280997
21491	4	1-5818	21488	74	66	329	61	2012-08-02 05:24:25.280997
21488	4	1-6247	21488	66	31	1592	24	2012-08-02 05:24:25.280997
21493	5	1-4102	21492	31	3	4	21	2012-08-02 05:24:25.280997
21494	5	1-5879	21492	3	61	3	3	2012-08-02 05:24:25.280997
21495	5	1-6081	21492	61	74	2	96	2012-08-02 05:24:25.280997
21496	5	1-5818	21492	74	66	1	61	2012-08-02 05:24:25.280997
21492	5	1-6247	21492	66	31	5	24	2012-08-02 05:24:25.280997
21498	4	1-6172	21497	6	79	4494	3	2012-08-02 05:24:26.938157
21499	4	1-6155	21497	79	7	2354	2	2012-08-02 05:24:26.938157
21500	4	1-6166	21497	7	54	1821	10	2012-08-02 05:24:26.938157
21497	4	1-6251	21497	54	6	4890	48	2012-08-02 05:24:26.938157
21502	5	1-5757	21501	73	39	107	89	2012-08-02 05:24:26.938157
21503	5	1-6192	21501	39	57	224	21	2012-08-02 05:24:26.938157
21504	5	1-3575	21501	57	7	83	2	2012-08-02 05:24:26.938157
21505	5	1-6166	21501	7	54	80	10	2012-08-02 05:24:26.938157
21501	5	1-6251	21501	54	73	268	48	2012-08-02 05:24:26.938157
21507	5	1-5757	21506	73	39	244	89	2012-08-02 05:24:26.938157
21508	5	1-6192	21506	39	2	530	21	2012-08-02 05:24:26.938157
21509	5	1-4540	21506	2	7	171	2	2012-08-02 05:24:26.938157
21510	5	1-6166	21506	7	54	171	10	2012-08-02 05:24:26.938157
21506	5	1-6251	21506	54	73	592	48	2012-08-02 05:24:26.938157
21512	4	1-6175	21511	66	34	3447	4	2012-08-02 05:24:27.766833
21513	4	1-6051	21511	34	31	2379	22	2012-08-02 05:24:27.766833
21514	4	1-5566	21511	31	1	1690	49	2012-08-02 05:24:27.766833
21511	4	1-6253	21511	1	66	1312	71	2012-08-02 05:24:27.766833
21516	3	1-6187	21515	56	76	2466	61	2012-08-02 05:24:28.175566
21517	3	1-5340	21515	76	66	1311	50	2012-08-02 05:24:28.175566
21515	3	1-6254	21515	66	56	1302	99	2012-08-02 05:24:28.175566
21519	5	1-5796	21518	76	1	3205	49	2012-08-02 05:24:28.175566
21520	5	1-6253	21518	1	96	2265	71	2012-08-02 05:24:28.175566
21521	5	1-6197	21518	96	18	2013	24	2012-08-02 05:24:28.175566
21522	5	1-5591	21518	18	66	2013	50	2012-08-02 05:24:28.175566
21518	5	1-6254	21518	66	76	2382	99	2012-08-02 05:24:28.175566
21524	4	1-5796	21523	76	96	3621	49	2012-08-02 05:24:28.175566
21525	4	1-6024	21523	96	7	1697	1	2012-08-02 05:24:28.175566
21526	4	1-5877	21523	7	66	1992	50	2012-08-02 05:24:28.175566
21523	4	1-6254	21523	66	76	2519	99	2012-08-02 05:24:28.175566
21528	4	1-6233	21527	3	84	2476	7	2012-08-02 05:24:28.606795
21529	4	1-4900	21527	84	90	1006	49	2012-08-02 05:24:28.606795
21530	4	1-5721	21527	90	4	455	95	2012-08-02 05:24:28.606795
21527	4	1-6255	21527	4	3	882	89	2012-08-02 05:24:28.606795
21532	5	1-6192	21531	39	3	3352	21	2012-08-02 05:24:28.606795
21533	5	1-5879	21531	3	31	2121	3	2012-08-02 05:24:28.606795
21534	5	1-1508	21531	31	98	587	73	2012-08-02 05:24:28.606795
21535	5	1-6235	21531	98	4	514	95	2012-08-02 05:24:28.606795
21531	5	1-6255	21531	4	39	1323	89	2012-08-02 05:24:28.606795
21537	4	1-5908	21536	64	18	5607	92	2012-08-02 05:24:28.606795
21538	4	1-1778	21536	18	34	1828	88	2012-08-02 05:24:28.606795
21539	4	1-6019	21536	34	4	1077	95	2012-08-02 05:24:28.606795
21536	4	1-6255	21536	4	64	3070	89	2012-08-02 05:24:28.606795
21541	5	1-2361	21540	6	26	325	91	2012-08-02 05:24:28.606795
21542	5	1-6183	21540	26	89	822	11	2012-08-02 05:24:28.606795
21543	5	1-5405	21540	89	98	358	73	2012-08-02 05:24:28.606795
21544	5	1-6235	21540	98	4	385	95	2012-08-02 05:24:28.606795
21540	5	1-6255	21540	4	6	1216	89	2012-08-02 05:24:28.606795
21546	4	1-6037	21545	65	31	2065	22	2012-08-02 05:24:33.202506
21547	4	1-1851	21545	31	44	1247	16	2012-08-02 05:24:33.202506
21548	4	1-5900	21545	44	27	1778	51	2012-08-02 05:24:33.202506
21545	4	1-6266	21545	27	65	5150	85	2012-08-02 05:24:33.202506
21550	5	1-6166	21549	7	27	4926	10	2012-08-02 05:24:35.29057
21551	5	1-5391	21549	27	70	3383	39	2012-08-02 05:24:35.29057
21552	5	1-5867	21549	70	98	2659	73	2012-08-02 05:24:35.29057
21553	5	1-6235	21549	98	74	3104	95	2012-08-02 05:24:35.29057
21549	5	1-6271	21549	74	7	3166	2	2012-08-02 05:24:35.29057
21555	4	1-5859	21554	96	58	4002	3	2012-08-02 05:24:36.848189
21556	4	1-3986	21554	58	3	4346	14	2012-08-02 05:24:36.848189
21557	4	1-6185	21554	3	89	1965	56	2012-08-02 05:24:36.848189
21554	4	1-6274	21554	89	96	2894	94	2012-08-02 05:24:36.848189
21559	2	1-2343	21558	34	55	542	5	2012-08-02 05:24:38.120899
21558	2	1-6277	21558	55	34	8622	36	2012-08-02 05:24:38.120899
21561	4	1-6202	21560	85	92	931	77	2012-08-02 05:24:39.413539
21562	4	1-5855	21560	92	67	518	5	2012-08-02 05:24:39.413539
21563	4	1-6220	21560	67	8	895	81	2012-08-02 05:24:39.413539
21560	4	1-6280	21560	8	85	1034	57	2012-08-02 05:24:39.413539
21565	5	1-6202	21564	85	65	1033	77	2012-08-02 05:24:39.413539
21566	5	1-5637	21564	65	99	473	3	2012-08-02 05:24:39.413539
21567	5	1-6267	21564	99	38	474	53	2012-08-02 05:24:39.413539
21568	5	1-6237	21564	38	8	654	81	2012-08-02 05:24:39.413539
21564	5	1-6280	21564	8	85	931	57	2012-08-02 05:24:39.413539
21570	4	1-5791	21569	27	44	257	75	2012-08-02 05:24:39.413539
21571	4	1-3500	21569	44	67	247	5	2012-08-02 05:24:39.413539
21572	4	1-6220	21569	67	8	619	81	2012-08-02 05:24:39.413539
21569	4	1-6280	21569	8	27	1038	57	2012-08-02 05:24:39.413539
21574	5	1-5079	21573	18	13	2247	84	2012-08-02 05:24:39.413539
21575	5	1-5893	21573	13	68	454	52	2012-08-02 05:24:39.413539
21576	5	1-5921	21573	68	38	1073	53	2012-08-02 05:24:39.413539
21577	5	1-6237	21573	38	8	2008	81	2012-08-02 05:24:39.413539
21573	5	1-6280	21573	8	18	3870	57	2012-08-02 05:24:39.413539
21579	5	1-6237	21578	38	17	60	81	2012-08-02 05:24:40.253037
21580	5	1-5887	21578	17	82	24	4	2012-08-02 05:24:40.253037
21581	5	1-6089	21578	82	38	8	90	2012-08-02 05:24:40.253037
21582	5	1-6223	21578	38	32	13	93	2012-08-02 05:24:40.253037
21578	5	1-6282	21578	32	38	37	53	2012-08-02 05:24:40.253037
21584	5	1-2527	21583	93	96	39	39	2012-08-02 05:24:40.253037
21585	5	1-5605	21583	96	35	35	60	2012-08-02 05:24:40.253037
21586	5	1-5359	21583	35	38	35	90	2012-08-02 05:24:40.253037
21587	5	1-6223	21583	38	32	63	93	2012-08-02 05:24:40.253037
21583	5	1-6282	21583	32	93	258	53	2012-08-02 05:24:40.253037
21589	5	1-6258	21588	70	1	592	36	2012-08-02 05:24:41.932477
21590	5	1-5259	21588	1	61	1774	37	2012-08-02 05:24:41.932477
21591	5	1-6054	21588	61	34	438	4	2012-08-02 05:24:41.932477
21592	5	1-6051	21588	34	37	253	22	2012-08-02 05:24:41.932477
21588	5	1-6286	21588	37	70	485	33	2012-08-02 05:24:41.932477
21594	4	1-6258	21593	70	13	3898	36	2012-08-02 05:24:41.932477
21595	4	1-5085	21593	13	34	1804	4	2012-08-02 05:24:41.932477
21596	4	1-6051	21593	34	37	1220	22	2012-08-02 05:24:41.932477
21593	4	1-6286	21593	37	70	2730	33	2012-08-02 05:24:41.932477
21598	5	1-5644	21597	82	72	1550	8	2012-08-02 05:24:41.932477
21599	5	1-5798	21597	72	10	3047	34	2012-08-02 05:24:41.932477
21600	5	1-5354	21597	10	34	2013	4	2012-08-02 05:24:41.932477
21601	5	1-6051	21597	34	37	1456	22	2012-08-02 05:24:41.932477
21597	5	1-6286	21597	37	82	3483	33	2012-08-02 05:24:41.932477
21603	5	1-6258	21602	70	35	617	36	2012-08-02 05:24:41.932477
21604	5	1-4032	21602	35	27	374	88	2012-08-02 05:24:41.932477
21605	5	1-4289	21602	27	29	157	74	2012-08-02 05:24:41.932477
21606	5	1-5187	21602	29	37	149	22	2012-08-02 05:24:41.932477
21602	5	1-6286	21602	37	70	380	33	2012-08-02 05:24:41.932477
21608	5	1-5644	21607	82	14	540	8	2012-08-02 05:24:41.932477
21609	5	1-5556	21607	14	54	597	59	2012-08-02 05:24:41.932477
21610	5	1-5365	21607	54	34	516	4	2012-08-02 05:24:41.932477
21611	5	1-6051	21607	34	37	413	22	2012-08-02 05:24:41.932477
21607	5	1-6286	21607	37	82	1096	33	2012-08-02 05:24:41.932477
21613	5	1-6258	21612	70	92	2512	36	2012-08-02 05:24:41.932477
21614	5	1-6095	21612	92	45	1606	93	2012-08-02 05:24:41.932477
21615	5	1-5866	21612	45	34	686	4	2012-08-02 05:24:41.932477
21616	5	1-6051	21612	34	37	553	22	2012-08-02 05:24:41.932477
21612	5	1-6286	21612	37	70	1476	33	2012-08-02 05:24:41.932477
21618	3	1-5774	21617	95	30	1120	16	2012-08-02 05:24:42.772559
21619	3	1-5136	21617	30	44	510	43	2012-08-02 05:24:42.772559
21617	3	1-6288	21617	44	95	2972	98	2012-08-02 05:24:42.772559
21621	5	1-3237	21620	14	85	1024	22	2012-08-02 05:24:42.772559
21622	5	1-5944	21620	85	96	585	94	2012-08-02 05:24:42.772559
21623	5	1-5859	21620	96	17	636	3	2012-08-02 05:24:42.772559
21624	5	1-1226	21620	17	44	352	43	2012-08-02 05:24:42.772559
21620	5	1-6288	21620	44	14	2177	98	2012-08-02 05:24:42.772559
21626	4	1-5848	21625	44	65	720	85	2012-08-02 05:24:42.772559
21627	4	1-6037	21625	65	83	238	22	2012-08-02 05:24:42.772559
21628	4	1-5619	21625	83	44	68	43	2012-08-02 05:24:42.772559
21625	4	1-6288	21625	44	44	436	98	2012-08-02 05:24:42.772559
21630	4	1-5231	21629	19	67	1267	72	2012-08-02 05:24:42.772559
21631	4	1-4955	21629	67	89	315	19	2012-08-02 05:24:42.772559
21632	4	1-5909	21629	89	44	268	43	2012-08-02 05:24:42.772559
21629	4	1-6288	21629	44	19	1796	98	2012-08-02 05:24:42.772559
21633	2	1-6292	21633	12	52	3906	86	2012-08-02 05:24:44.463222
21634	2	1-4374	21633	52	12	1825	16	2012-08-02 05:24:44.463222
21636	2	1-6125	21635	12	96	6926	58	2012-08-02 05:24:44.872395
21635	2	1-6293	21635	96	12	6323	8	2012-08-02 05:24:44.872395
21638	5	1-5859	21637	96	79	4435	3	2012-08-02 05:24:46.926828
21639	5	1-6155	21637	79	26	3587	2	2012-08-02 05:24:46.926828
21640	5	1-5959	21637	26	17	2331	20	2012-08-02 05:24:46.926828
21641	5	1-6294	21637	17	16	596	99	2012-08-02 05:24:46.926828
21637	5	1-6298	21637	16	96	4116	94	2012-08-02 05:24:46.926828
21643	5	1-5783	21642	44	58	1183	48	2012-08-02 05:24:46.926828
21644	5	1-6077	21642	58	100	1108	77	2012-08-02 05:24:46.926828
21645	5	1-5352	21642	100	97	939	67	2012-08-02 05:24:46.926828
21646	5	1-2168	21642	97	16	118	99	2012-08-02 05:24:46.926828
21642	5	1-6298	21642	16	44	875	94	2012-08-02 05:24:46.926828
21648	5	1-4104	21647	55	74	2142	95	2012-08-02 05:24:46.926828
21649	5	1-6271	21647	74	26	1911	2	2012-08-02 05:24:46.926828
21650	5	1-5959	21647	26	17	1415	20	2012-08-02 05:24:46.926828
21651	5	1-6294	21647	17	16	412	99	2012-08-02 05:24:46.926828
21647	5	1-6298	21647	16	55	3245	94	2012-08-02 05:24:46.926828
21653	5	1-3451	21652	45	6	382	41	2012-08-02 05:24:46.926828
21654	5	1-3861	21652	6	1	143	32	2012-08-02 05:24:46.926828
21655	5	1-6210	21652	1	17	361	20	2012-08-02 05:24:46.926828
21656	5	1-6294	21652	17	16	109	99	2012-08-02 05:24:46.926828
21652	5	1-6298	21652	16	45	884	94	2012-08-02 05:24:46.926828
21658	3	1-5833	21657	23	74	2745	95	2012-08-02 05:24:47.755777
21659	3	1-6271	21657	74	92	2320	2	2012-08-02 05:24:47.755777
21657	3	1-6300	21657	92	23	4621	20	2012-08-02 05:24:47.755777
21661	4	1-5714	21660	30	1	2419	54	2012-08-02 05:24:47.755777
21662	4	1-5785	21660	1	77	1836	17	2012-08-02 05:24:47.755777
21663	4	1-3938	21660	77	92	580	2	2012-08-02 05:24:47.755777
21660	4	1-6300	21660	92	30	1232	20	2012-08-02 05:24:47.755777
21665	4	1-4683	21664	35	80	38	16	2012-08-02 05:24:47.755777
21666	4	1-5989	21664	80	6	64	87	2012-08-02 05:24:47.755777
21667	4	1-2663	21664	6	92	36	2	2012-08-02 05:24:47.755777
21664	4	1-6300	21664	92	35	82	20	2012-08-02 05:24:47.755777
21669	4	1-6006	21668	87	15	1401	63	2012-08-02 05:24:49.391277
21670	4	1-5988	21668	15	93	4096	88	2012-08-02 05:24:49.391277
21671	4	1-3995	21668	93	11	391	7	2012-08-02 05:24:49.391277
21668	4	1-6304	21668	11	87	911	66	2012-08-02 05:24:49.391277
21673	4	1-5566	21672	31	60	478	49	2012-08-02 05:24:51.898893
21674	4	1-6303	21672	60	34	726	88	2012-08-02 05:24:51.898893
21675	4	1-6019	21672	34	66	452	95	2012-08-02 05:24:51.898893
21672	4	1-6310	21672	66	31	605	22	2012-08-02 05:24:51.898893
21677	4	1-6286	21676	37	59	6	33	2012-08-02 05:24:51.898893
21678	4	1-6195	21676	59	98	2	73	2012-08-02 05:24:51.898893
21679	4	1-6235	21676	98	66	2	95	2012-08-02 05:24:51.898893
21676	4	1-6310	21676	66	37	3	22	2012-08-02 05:24:51.898893
21681	3	1-6153	21680	26	40	1885	41	2012-08-02 05:24:53.975941
21682	3	1-3963	21680	40	51	971	14	2012-08-02 05:24:53.975941
21680	3	1-6315	21680	51	26	2759	2	2012-08-02 05:24:53.975941
21684	4	1-5981	21683	48	10	142	47	2012-08-02 05:24:53.975941
21685	4	1-4442	21683	10	17	158	25	2012-08-02 05:24:53.975941
21686	4	1-5744	21683	17	51	97	14	2012-08-02 05:24:53.975941
21683	4	1-6315	21683	51	48	337	2	2012-08-02 05:24:53.975941
21688	5	1-5989	21687	80	68	1316	87	2012-08-02 05:24:54.395896
21689	5	1-6225	21687	68	99	1075	80	2012-08-02 05:24:54.395896
21690	5	1-5966	21687	99	15	208	63	2012-08-02 05:24:54.395896
21691	5	1-5988	21687	15	25	292	88	2012-08-02 05:24:54.395896
21687	5	1-6316	21687	25	80	1648	16	2012-08-02 05:24:54.395896
21693	4	1-5136	21692	30	98	313	43	2012-08-02 05:24:54.395896
21694	4	1-5443	21692	98	60	108	49	2012-08-02 05:24:54.395896
21695	4	1-6303	21692	60	25	112	88	2012-08-02 05:24:54.395896
21692	4	1-6316	21692	25	30	836	16	2012-08-02 05:24:54.395896
21697	5	1-6023	21696	13	90	354	38	2012-08-02 05:24:54.395896
21698	5	1-5234	21696	90	30	381	58	2012-08-02 05:24:54.395896
21699	5	1-5481	21696	30	15	111	63	2012-08-02 05:24:54.395896
21700	5	1-5988	21696	15	25	206	88	2012-08-02 05:24:54.395896
21696	5	1-6316	21696	25	13	1549	16	2012-08-02 05:24:54.395896
21702	5	1-5391	21701	27	96	594	39	2012-08-02 05:24:55.677723
21703	5	1-5605	21701	96	38	582	60	2012-08-02 05:24:55.677723
21704	5	1-4552	21701	38	44	140	30	2012-08-02 05:24:55.677723
21705	5	1-6207	21701	44	74	114	24	2012-08-02 05:24:55.677723
21701	5	1-6319	21701	74	27	868	10	2012-08-02 05:24:55.677723
21707	3	1-3342	21706	93	29	918	41	2012-08-02 05:24:56.108512
21708	3	1-6297	21706	29	14	907	36	2012-08-02 05:24:56.108512
21706	3	1-6320	21706	14	93	1138	18	2012-08-02 05:24:56.108512
21710	3	1-5756	21709	67	89	2110	38	2012-08-02 05:24:56.108512
21711	3	1-5162	21709	89	14	1682	36	2012-08-02 05:24:56.108512
21709	3	1-6320	21709	14	67	2137	18	2012-08-02 05:24:56.108512
21713	5	1-2523	21712	58	8	231	43	2012-08-02 05:24:56.528368
21714	5	1-5928	21712	8	85	183	75	2012-08-02 05:24:56.528368
21715	5	1-5776	21712	85	78	197	18	2012-08-02 05:24:56.528368
21716	5	1-5759	21712	78	65	343	90	2012-08-02 05:24:56.528368
21712	5	1-6321	21712	65	58	349	56	2012-08-02 05:24:56.528368
21718	5	1-6274	21717	89	80	218	94	2012-08-02 05:24:56.528368
21719	5	1-5574	21717	80	43	207	63	2012-08-02 05:24:56.528368
21720	5	1-5306	21717	43	76	45	9	2012-08-02 05:24:56.528368
21721	5	1-6075	21717	76	65	143	90	2012-08-02 05:24:56.528368
21717	5	1-6321	21717	65	89	147	56	2012-08-02 05:24:56.528368
21723	5	1-4812	21722	25	44	342	30	2012-08-02 05:24:58.583216
21724	5	1-6207	21722	44	74	183	24	2012-08-02 05:24:58.583216
21725	5	1-6319	21722	74	33	913	10	2012-08-02 05:24:58.583216
21726	5	1-5750	21722	33	3	939	44	2012-08-02 05:24:58.583216
21722	5	1-6326	21722	3	25	1498	93	2012-08-02 05:24:58.583216
21728	5	1-4283	21727	73	25	513	61	2012-08-02 05:24:58.583216
21729	5	1-5926	21727	25	7	569	82	2012-08-02 05:24:58.583216
21730	5	1-3682	21727	7	47	1080	80	2012-08-02 05:24:58.583216
21731	5	1-6069	21727	47	3	922	44	2012-08-02 05:24:58.583216
21727	5	1-6326	21727	3	73	2068	93	2012-08-02 05:24:58.583216
21733	5	1-1226	21732	17	95	332	43	2012-08-02 05:24:59.003113
21734	5	1-5720	21732	95	82	375	79	2012-08-02 05:24:59.003113
21735	5	1-6174	21732	82	49	621	40	2012-08-02 05:24:59.003113
21736	5	1-6199	21732	49	33	661	82	2012-08-02 05:24:59.003113
21732	5	1-6327	21732	33	17	489	3	2012-08-02 05:24:59.003113
21738	5	1-5957	21737	98	6	2307	60	2012-08-02 05:24:59.85385
21737	5	1-6329	21737	33	98	2276	1	2012-08-02 05:24:59.85385
21739	5	1-5290	21737	6	98	2373	73	2012-08-02 05:24:59.85385
21740	5	1-6235	21737	98	66	2744	95	2012-08-02 05:24:59.85385
21741	5	1-6310	21737	66	33	4047	22	2012-08-02 05:24:59.85385
21743	5	1-5877	21742	7	23	261	50	2012-08-02 05:24:59.85385
21744	5	1-6093	21742	23	45	209	4	2012-08-02 05:24:59.85385
21745	5	1-6133	21742	45	66	201	95	2012-08-02 05:24:59.85385
21746	5	1-6310	21742	66	33	302	22	2012-08-02 05:24:59.85385
21742	5	1-6329	21742	33	7	173	1	2012-08-02 05:24:59.85385
21748	5	1-5760	21747	40	67	709	88	2012-08-02 05:25:00.693959
21749	5	1-6145	21747	67	79	1209	15	2012-08-02 05:25:00.693959
21750	5	1-5982	21747	79	27	1241	51	2012-08-02 05:25:00.693959
21751	5	1-6227	21747	27	90	1297	29	2012-08-02 05:25:00.693959
21747	5	1-6331	21747	90	40	2035	53	2012-08-02 05:25:00.693959
21753	5	1-3524	21752	51	73	458	23	2012-08-02 05:25:01.113546
21754	5	1-6044	21752	73	96	576	71	2012-08-02 05:25:01.113546
21755	5	1-6197	21752	96	72	539	24	2012-08-02 05:25:01.113546
21756	5	1-4951	21752	72	28	375	32	2012-08-02 05:25:01.113546
21752	5	1-6332	21752	28	51	1435	33	2012-08-02 05:25:01.113546
21758	4	1-6258	21757	70	70	571	36	2012-08-02 05:25:01.113546
21759	4	1-4825	21757	70	73	303	2	2012-08-02 05:25:01.113546
21760	4	1-6262	21757	73	28	89	32	2012-08-02 05:25:01.113546
21757	4	1-6332	21757	28	70	349	33	2012-08-02 05:25:01.113546
21762	4	1-1009	21761	75	23	280	72	2012-08-02 05:25:01.113546
21763	4	1-5039	21761	23	89	245	68	2012-08-02 05:25:01.113546
21764	4	1-6104	21761	89	28	147	32	2012-08-02 05:25:01.113546
21761	4	1-6332	21761	28	75	726	33	2012-08-02 05:25:01.113546
21766	3	1-6327	21765	33	61	1117	3	2012-08-02 05:25:02.373071
21767	3	1-6081	21765	61	77	644	96	2012-08-02 05:25:02.373071
21765	3	1-6335	21765	77	33	1726	82	2012-08-02 05:25:02.373071
21769	5	1-1389	21768	2	17	507	46	2012-08-02 05:25:02.373071
21770	5	1-6325	21768	17	4	400	6	2012-08-02 05:25:02.373071
21771	5	1-3765	21768	4	82	339	56	2012-08-02 05:25:02.373071
21772	5	1-6005	21768	82	77	351	96	2012-08-02 05:25:02.373071
21768	5	1-6335	21768	77	2	1034	82	2012-08-02 05:25:02.373071
21774	5	1-6327	21773	33	15	1879	3	2012-08-02 05:25:02.373071
21775	5	1-5406	21773	15	65	894	90	2012-08-02 05:25:02.373071
21776	5	1-6321	21773	65	82	853	56	2012-08-02 05:25:02.373071
21777	5	1-6005	21773	82	77	887	96	2012-08-02 05:25:02.373071
21773	5	1-6335	21773	77	33	2626	82	2012-08-02 05:25:02.373071
21779	3	1-5753	21778	54	23	1613	13	2012-08-02 05:25:02.373071
21780	3	1-6240	21778	23	77	624	96	2012-08-02 05:25:02.373071
21778	3	1-6335	21778	77	54	1900	82	2012-08-02 05:25:02.373071
21782	5	1-4297	21781	85	66	254	95	2012-08-02 05:25:03.645736
21783	5	1-6310	21781	66	33	296	22	2012-08-02 05:25:03.645736
21784	5	1-6329	21781	33	68	131	1	2012-08-02 05:25:03.645736
21785	5	1-6246	21781	68	92	211	27	2012-08-02 05:25:03.645736
21781	5	1-6338	21781	92	85	269	9	2012-08-02 05:25:03.645736
21787	5	1-5925	21786	17	89	1808	56	2012-08-02 05:25:03.645736
21788	5	1-6274	21786	89	25	2348	94	2012-08-02 05:25:03.645736
21789	5	1-5629	21786	25	68	978	1	2012-08-02 05:25:03.645736
21790	5	1-6246	21786	68	92	1785	27	2012-08-02 05:25:03.645736
21786	5	1-6338	21786	92	17	2593	9	2012-08-02 05:25:03.645736
21792	4	1-3035	21791	65	27	696	51	2012-08-02 05:25:03.645736
21793	4	1-6227	21791	27	80	795	29	2012-08-02 05:25:03.645736
21794	4	1-5042	21791	80	92	517	27	2012-08-02 05:25:03.645736
21791	4	1-6338	21791	92	65	782	9	2012-08-02 05:25:03.645736
21796	5	1-2730	21795	45	34	925	17	2012-08-02 05:25:03.645736
21797	5	1-5951	21795	34	96	544	49	2012-08-02 05:25:03.645736
21798	5	1-6024	21795	96	68	329	1	2012-08-02 05:25:03.645736
21799	5	1-6246	21795	68	92	683	27	2012-08-02 05:25:03.645736
21795	5	1-6338	21795	92	45	1130	9	2012-08-02 05:25:03.645736
21801	5	1-6270	21800	13	24	1830	65	2012-08-02 05:25:04.507616
21802	5	1-6239	21800	24	55	2523	79	2012-08-02 05:25:04.507616
21803	5	1-6072	21800	55	35	2823	66	2012-08-02 05:25:04.507616
21804	5	1-5964	21800	35	25	4080	26	2012-08-02 05:25:04.507616
21800	5	1-6340	21800	25	13	3576	86	2012-08-02 05:25:04.507616
21805	1	1-211	\N	17	17	712	86	2012-08-02 05:25:04.507616
21807	4	1-5437	21806	23	61	1059	70	2012-08-02 05:25:04.927538
21808	4	1-5895	21806	61	61	581	98	2012-08-02 05:25:04.927538
21809	4	1-2920	21806	61	86	583	31	2012-08-02 05:25:04.927538
21806	4	1-6341	21806	86	23	4109	84	2012-08-02 05:25:04.927538
21811	3	1-6075	21810	76	30	2581	90	2012-08-02 05:25:05.359362
21812	3	1-5528	21810	30	99	892	87	2012-08-02 05:25:05.359362
21810	3	1-6342	21810	99	76	2273	9	2012-08-02 05:25:05.359362
21814	3	1-4906	21813	9	87	1082	40	2012-08-02 05:25:05.359362
21815	3	1-5060	21813	87	99	357	87	2012-08-02 05:25:05.359362
21813	3	1-6342	21813	99	9	1493	9	2012-08-02 05:25:05.359362
21817	5	1-5881	21816	11	98	560	17	2012-08-02 05:25:10.210421
21818	5	1-5703	21816	98	93	410	45	2012-08-02 05:25:10.210421
21819	5	1-5740	21816	93	47	350	80	2012-08-02 05:25:10.210421
21820	5	1-6069	21816	47	17	236	44	2012-08-02 05:25:10.210421
21816	5	1-6354	21816	17	11	807	13	2012-08-02 05:25:10.210421
21822	5	1-5881	21821	11	55	864	17	2012-08-02 05:25:10.210421
21823	5	1-5787	21821	55	68	409	87	2012-08-02 05:25:10.210421
21824	5	1-6225	21821	68	47	534	80	2012-08-02 05:25:10.210421
21825	5	1-6069	21821	47	17	362	44	2012-08-02 05:25:10.210421
21821	5	1-6354	21821	17	11	1240	13	2012-08-02 05:25:10.210421
21827	5	1-5842	21826	18	74	692	31	2012-08-02 05:25:10.210421
21828	5	1-5965	21826	74	68	238	87	2012-08-02 05:25:10.210421
21829	5	1-6225	21826	68	47	337	80	2012-08-02 05:25:10.210421
21830	5	1-6069	21826	47	17	247	44	2012-08-02 05:25:10.210421
21826	5	1-6354	21826	17	18	919	13	2012-08-02 05:25:10.210421
21832	5	1-6240	21831	23	10	2270	96	2012-08-02 05:25:10.210421
21833	5	1-1141	21831	10	36	2511	14	2012-08-02 05:25:10.210421
21834	5	1-6063	21831	36	47	1804	80	2012-08-02 05:25:10.210421
21835	5	1-6069	21831	47	17	1492	44	2012-08-02 05:25:10.210421
21831	5	1-6354	21831	17	23	6252	13	2012-08-02 05:25:10.210421
21837	5	1-6044	21836	73	96	4895	71	2012-08-02 05:25:11.525567
21838	5	1-6197	21836	96	36	5474	24	2012-08-02 05:25:11.525567
21839	5	1-3464	21836	36	68	1452	1	2012-08-02 05:25:11.525567
21840	5	1-6246	21836	68	58	2757	27	2012-08-02 05:25:11.525567
21836	5	1-6357	21836	58	73	3253	23	2012-08-02 05:25:11.525567
21842	3	1-3038	21841	84	37	655	12	2012-08-02 05:25:11.525567
21843	3	1-5914	21841	37	58	1335	27	2012-08-02 05:25:11.525567
21841	3	1-6357	21841	58	84	1738	23	2012-08-02 05:25:11.525567
21845	5	1-6020	21844	40	54	4000	59	2012-08-02 05:25:12.33188
21846	5	1-5365	21844	54	45	2485	4	2012-08-02 05:25:12.33188
21847	5	1-6133	21844	45	66	1595	95	2012-08-02 05:25:12.33188
21848	5	1-6310	21844	66	60	1601	22	2012-08-02 05:25:12.33188
21844	5	1-6359	21844	60	40	3330	60	2012-08-02 05:25:12.33188
21850	3	1-5669	21849	83	29	952	74	2012-08-02 05:25:12.33188
21851	3	1-5187	21849	29	60	779	22	2012-08-02 05:25:12.33188
21849	3	1-6359	21849	60	83	1860	60	2012-08-02 05:25:12.33188
21853	5	1-6011	21852	4	61	3522	37	2012-08-02 05:25:12.33188
21854	5	1-6054	21852	61	45	1164	4	2012-08-02 05:25:12.33188
21855	5	1-6133	21852	45	66	1005	95	2012-08-02 05:25:12.33188
21856	5	1-6310	21852	66	60	1357	22	2012-08-02 05:25:12.33188
21852	5	1-6359	21852	60	4	3797	60	2012-08-02 05:25:12.33188
21858	4	1-6250	21857	51	15	38	63	2012-08-02 05:25:13.547315
21859	4	1-5988	21857	15	67	111	88	2012-08-02 05:25:13.547315
21860	4	1-6145	21857	67	86	219	15	2012-08-02 05:25:13.547315
21857	4	1-6362	21857	86	51	36	34	2012-08-02 05:25:13.547315
21862	3	1-3500	21861	44	4	3854	5	2012-08-02 05:25:15.547029
21863	3	1-4231	21861	4	45	747	69	2012-08-02 05:25:15.547029
21861	3	1-6367	21861	45	44	7070	75	2012-08-02 05:25:15.547029
21865	4	1-2507	21864	56	66	290	61	2012-08-02 05:25:15.547029
21866	4	1-6182	21864	66	54	398	86	2012-08-02 05:25:15.547029
21867	4	1-4019	21864	54	45	54	69	2012-08-02 05:25:15.547029
21864	4	1-6367	21864	45	56	613	75	2012-08-02 05:25:15.547029
21869	5	1-5848	21868	44	60	1228	85	2012-08-02 05:25:16.353446
21870	5	1-4982	21868	60	61	1333	37	2012-08-02 05:25:16.353446
21871	5	1-6054	21868	61	45	443	4	2012-08-02 05:25:16.353446
21872	5	1-6133	21868	45	32	384	95	2012-08-02 05:25:16.353446
21868	5	1-6369	21868	32	44	676	98	2012-08-02 05:25:16.353446
21874	5	1-1417	21873	93	65	813	51	2012-08-02 05:25:16.353446
21875	5	1-6083	21873	65	61	1782	37	2012-08-02 05:25:16.353446
21876	5	1-6054	21873	61	45	637	4	2012-08-02 05:25:16.353446
21877	5	1-6133	21873	45	32	595	95	2012-08-02 05:25:16.353446
21873	5	1-6369	21873	32	93	1128	98	2012-08-02 05:25:16.353446
21879	4	1-6345	21878	62	60	710	49	2012-08-02 05:25:16.353446
21880	4	1-6303	21878	60	34	1226	88	2012-08-02 05:25:16.353446
21881	4	1-6019	21878	34	32	868	95	2012-08-02 05:25:16.353446
21878	4	1-6369	21878	32	62	1717	98	2012-08-02 05:25:16.353446
21883	5	1-5966	21882	99	15	847	63	2012-08-02 05:25:19.270626
21884	5	1-5988	21882	15	34	2009	88	2012-08-02 05:25:19.270626
21885	5	1-6019	21882	34	32	1098	95	2012-08-02 05:25:19.270626
21886	5	1-6369	21882	32	53	1675	98	2012-08-02 05:25:19.270626
21882	5	1-6376	21882	53	99	2584	80	2012-08-02 05:25:19.270626
21888	5	1-6053	21887	5	26	1291	92	2012-08-02 05:25:19.270626
21889	5	1-6350	21887	26	37	948	76	2012-08-02 05:25:19.270626
21890	5	1-4977	21887	37	32	345	95	2012-08-02 05:25:19.270626
21891	5	1-6369	21887	32	53	567	98	2012-08-02 05:25:19.270626
21887	5	1-6376	21887	53	5	941	80	2012-08-02 05:25:19.270626
21893	5	1-6053	21892	5	18	1933	92	2012-08-02 05:25:19.270626
21894	5	1-1778	21892	18	34	673	88	2012-08-02 05:25:19.270626
21895	5	1-6019	21892	34	32	423	95	2012-08-02 05:25:19.270626
21896	5	1-6369	21892	32	53	743	98	2012-08-02 05:25:19.270626
21892	5	1-6376	21892	53	5	1318	80	2012-08-02 05:25:19.270626
21898	5	1-6069	21897	47	12	2382	44	2012-08-02 05:25:19.270626
21899	5	1-6363	21897	12	37	1660	64	2012-08-02 05:25:19.270626
21900	5	1-5690	21897	37	61	1410	70	2012-08-02 05:25:19.270626
21901	5	1-5895	21897	61	53	1247	98	2012-08-02 05:25:19.270626
21897	5	1-6376	21897	53	47	2530	80	2012-08-02 05:25:19.270626
21903	4	1-5687	21902	64	56	718	42	2012-08-02 05:25:20.507878
21904	4	1-5661	21902	56	65	886	51	2012-08-02 05:25:20.507878
21905	4	1-6083	21902	65	49	1242	37	2012-08-02 05:25:20.507878
21902	4	1-6379	21902	49	64	1866	44	2012-08-02 05:25:20.507878
21907	5	1-6220	21906	67	17	1457	81	2012-08-02 05:25:23.678507
21908	5	1-5887	21906	17	47	823	4	2012-08-02 05:25:23.678507
21909	5	1-4189	21906	47	90	255	38	2012-08-02 05:25:23.678507
21910	5	1-5234	21906	90	65	422	58	2012-08-02 05:25:23.678507
21906	5	1-6387	21906	65	67	515	5	2012-08-02 05:25:23.678507
21912	4	1-5788	21911	41	32	638	95	2012-08-02 05:25:23.678507
21913	4	1-6369	21911	32	74	1180	98	2012-08-02 05:25:23.678507
21914	4	1-4383	21911	74	65	1090	58	2012-08-02 05:25:23.678507
21911	4	1-6387	21911	65	41	1333	5	2012-08-02 05:25:23.678507
21916	5	1-2413	21915	50	23	612	50	2012-08-02 05:25:25.269635
21917	5	1-6093	21915	23	45	494	4	2012-08-02 05:25:25.269635
21918	5	1-6133	21915	45	66	477	95	2012-08-02 05:25:25.269635
21919	5	1-6310	21915	66	40	720	22	2012-08-02 05:25:25.269635
21915	5	1-6391	21915	40	50	1154	73	2012-08-02 05:25:25.269635
21921	4	1-6246	21920	68	92	1956	27	2012-08-02 05:25:26.91776
21922	4	1-6338	21920	92	94	2913	9	2012-08-02 05:25:26.91776
21923	4	1-2710	21920	94	60	600	35	2012-08-02 05:25:26.91776
21920	4	1-6395	21920	60	68	1045	1	2012-08-02 05:25:26.91776
21925	5	1-5788	21924	41	66	722	95	2012-08-02 05:25:27.315553
21926	5	1-6310	21924	66	66	1024	22	2012-08-02 05:25:27.315553
21927	5	1-4509	21924	66	4	841	100	2012-08-02 05:25:27.315553
21928	5	1-6180	21924	4	10	891	87	2012-08-02 05:25:27.315553
21924	5	1-6396	21924	10	41	1651	5	2012-08-02 05:25:27.315553
21930	4	1-5250	21929	48	66	2242	22	2012-08-02 05:25:27.315553
21931	4	1-4509	21929	66	4	1817	100	2012-08-02 05:25:27.315553
21932	4	1-6180	21929	4	10	2070	87	2012-08-02 05:25:27.315553
21929	4	1-6396	21929	10	48	4128	5	2012-08-02 05:25:27.315553
21934	4	1-6295	21933	1	86	1695	94	2012-08-02 05:25:27.713289
21935	4	1-5942	21933	86	57	1611	76	2012-08-02 05:25:27.713289
21936	4	1-6228	21933	57	37	1384	58	2012-08-02 05:25:27.713289
21933	4	1-6397	21933	37	1	1931	17	2012-08-02 05:25:27.713289
21938	5	1-6155	21937	79	26	813	2	2012-08-02 05:25:28.111411
21939	5	1-5959	21937	26	85	424	20	2012-08-02 05:25:28.111411
21940	5	1-6337	21937	85	100	246	6	2012-08-02 05:25:28.111411
21941	5	1-5592	21937	100	76	423	57	2012-08-02 05:25:28.111411
21937	5	1-6398	21937	76	79	1251	3	2012-08-02 05:25:28.111411
21943	4	1-5406	21942	15	65	1608	90	2012-08-02 05:25:28.111411
21944	4	1-6321	21942	65	96	1213	56	2012-08-02 05:25:28.111411
21945	4	1-3902	21942	96	76	1225	57	2012-08-02 05:25:28.111411
21942	4	1-6398	21942	76	15	4280	3	2012-08-02 05:25:28.111411
21947	5	1-1226	21946	17	88	126	43	2012-08-02 05:25:28.111411
21948	5	1-5351	21946	88	65	36	90	2012-08-02 05:25:28.111411
21949	5	1-6321	21946	65	96	34	56	2012-08-02 05:25:28.111411
21950	5	1-3902	21946	96	76	44	57	2012-08-02 05:25:28.111411
21946	5	1-6398	21946	76	17	193	3	2012-08-02 05:25:28.111411
21952	5	1-6248	21951	63	60	1518	35	2012-08-02 05:25:29.735097
21953	5	1-6395	21951	60	29	2666	1	2012-08-02 05:25:29.735097
21954	5	1-5285	21951	29	58	1970	27	2012-08-02 05:25:29.735097
21955	5	1-6357	21951	58	13	2310	23	2012-08-02 05:25:29.735097
21951	5	1-6402	21951	13	63	1618	19	2012-08-02 05:25:29.735097
21957	4	1-6271	21956	74	25	1005	2	2012-08-02 05:25:31.337389
21958	4	1-5802	21956	25	19	224	21	2012-08-02 05:25:31.337389
21959	4	1-6216	21956	19	92	123	1	2012-08-02 05:25:31.337389
21956	4	1-6406	21956	92	74	1217	95	2012-08-02 05:25:31.337389
21961	3	1-6334	21960	89	96	311	49	2012-08-02 05:25:31.337389
21962	3	1-6024	21960	96	92	177	1	2012-08-02 05:25:31.337389
21960	3	1-6406	21960	92	89	2008	95	2012-08-02 05:25:31.337389
21964	4	1-2045	21963	31	76	159	99	2012-08-02 05:25:31.337389
21965	4	1-5796	21963	76	96	286	49	2012-08-02 05:25:31.337389
21966	4	1-6024	21963	96	92	167	1	2012-08-02 05:25:31.337389
21963	4	1-6406	21963	92	31	1958	95	2012-08-02 05:25:31.337389
21968	5	1-5125	21967	14	47	1227	80	2012-08-02 05:25:31.337389
21969	5	1-6069	21967	47	42	1124	44	2012-08-02 05:25:31.337389
21970	5	1-3793	21967	42	19	367	21	2012-08-02 05:25:31.337389
21971	5	1-6216	21967	19	92	246	1	2012-08-02 05:25:31.337389
21967	5	1-6406	21967	92	14	2957	95	2012-08-02 05:25:31.337389
21973	4	1-5234	21972	90	42	1326	58	2012-08-02 05:25:32.563533
21974	4	1-5189	21972	42	22	813	28	2012-08-02 05:25:32.563533
21975	4	1-6204	21972	22	75	577	57	2012-08-02 05:25:32.563533
21972	4	1-6409	21972	75	90	836	38	2012-08-02 05:25:32.563533
21977	5	1-3487	21976	3	17	383	46	2012-08-02 05:25:32.563533
21978	5	1-6325	21976	17	4	296	6	2012-08-02 05:25:32.563533
21979	5	1-3765	21976	4	96	246	56	2012-08-02 05:25:32.563533
21980	5	1-3902	21976	96	75	307	57	2012-08-02 05:25:32.563533
21976	5	1-6409	21976	75	3	448	38	2012-08-02 05:25:32.563533
21982	3	1-5162	21981	89	30	5164	36	2012-08-02 05:25:32.563533
21983	3	1-4660	21981	30	75	4004	57	2012-08-02 05:25:32.563533
21981	3	1-6409	21981	75	89	6058	38	2012-08-02 05:25:32.563533
21985	4	1-5249	21984	46	96	144	71	2012-08-02 05:25:32.563533
21986	4	1-6197	21984	96	15	175	24	2012-08-02 05:25:32.563533
21987	4	1-1917	21984	15	75	150	57	2012-08-02 05:25:32.563533
21984	4	1-6409	21984	75	46	239	38	2012-08-02 05:25:32.563533
21989	5	1-5505	21988	96	28	31	32	2012-08-02 05:25:32.563533
21990	5	1-6332	21988	28	82	156	33	2012-08-02 05:25:32.563533
21991	5	1-5644	21988	82	6	83	8	2012-08-02 05:25:32.563533
21992	5	1-3486	21988	6	75	109	57	2012-08-02 05:25:32.563533
21988	5	1-6409	21988	75	96	175	38	2012-08-02 05:25:32.563533
21994	2	1-3021	21993	39	71	493	37	2012-08-02 05:25:33.370486
21993	2	1-6411	21993	71	39	461	38	2012-08-02 05:25:33.370486
21996	4	1-4590	21995	37	56	123	7	2012-08-02 05:25:35.359137
21997	4	1-6110	21995	56	74	205	11	2012-08-02 05:25:35.359137
21998	4	1-5806	21995	74	20	123	84	2012-08-02 05:25:35.359137
21995	4	1-6416	21995	20	37	591	83	2012-08-02 05:25:35.359137
22000	5	1-5914	21999	37	92	256	27	2012-08-02 05:25:35.779043
22001	5	1-6338	21999	92	31	409	9	2012-08-02 05:25:35.779043
22002	5	1-4280	21999	31	77	199	20	2012-08-02 05:25:35.779043
22003	5	1-6273	21999	77	74	79	82	2012-08-02 05:25:35.779043
21999	5	1-6417	21999	74	37	131	12	2012-08-02 05:25:35.779043
22005	4	1-6323	22004	22	95	1039	35	2012-08-02 05:25:35.779043
22006	4	1-6400	22004	95	25	1463	61	2012-08-02 05:25:35.779043
22007	4	1-5926	22004	25	74	1764	82	2012-08-02 05:25:35.779043
22004	4	1-6417	22004	74	22	3046	12	2012-08-02 05:25:35.779043
22009	2	1-5745	22008	69	65	3202	42	2012-08-02 05:25:36.187976
22008	2	1-6418	22008	65	69	3141	36	2012-08-02 05:25:36.187976
22011	4	1-5249	22010	46	96	682	71	2012-08-02 05:25:36.972537
22012	4	1-6197	22010	96	74	807	24	2012-08-02 05:25:36.972537
22013	4	1-6319	22010	74	92	5878	10	2012-08-02 05:25:36.972537
22010	4	1-6420	22010	92	46	1162	38	2012-08-02 05:25:36.972537
22015	5	1-6101	22014	87	100	221	45	2012-08-02 05:25:37.414738
22016	5	1-6412	22014	100	55	269	79	2012-08-02 05:25:37.414738
22017	5	1-6072	22014	55	97	262	66	2012-08-02 05:25:37.414738
22018	5	1-5631	22014	97	61	88	18	2012-08-02 05:25:37.414738
22014	5	1-6421	22014	61	87	137	61	2012-08-02 05:25:37.414738
22020	3	1-6183	22019	26	74	5285	11	2012-08-02 05:25:39.568559
22021	3	1-5806	22019	74	75	3438	84	2012-08-02 05:25:39.568559
22019	3	1-6426	22019	75	26	1949	91	2012-08-02 05:25:39.568559
22023	3	1-2148	22022	96	45	345	55	2012-08-02 05:25:40.364011
22024	3	1-4730	22022	45	19	400	37	2012-08-02 05:25:40.364011
22022	3	1-6428	22022	19	96	2245	63	2012-08-02 05:25:40.364011
22026	5	1-5738	22025	3	53	1437	54	2012-08-02 05:25:40.761748
22027	5	1-6135	22025	53	61	632	70	2012-08-02 05:25:40.761748
22028	5	1-5895	22025	61	53	513	98	2012-08-02 05:25:40.761748
22029	5	1-6376	22025	53	1	998	80	2012-08-02 05:25:40.761748
22025	5	1-6429	22025	1	3	1600	5	2012-08-02 05:25:40.761748
22031	4	1-4888	22030	28	87	1435	40	2012-08-02 05:25:40.761748
22032	4	1-5060	22030	87	68	766	87	2012-08-02 05:25:40.761748
22033	4	1-6225	22030	68	1	1290	80	2012-08-02 05:25:40.761748
22030	4	1-6429	22030	1	28	2089	5	2012-08-02 05:25:40.761748
22035	4	1-5401	22034	97	24	3740	8	2012-08-02 05:25:41.159644
22036	4	1-5397	22034	24	60	1776	35	2012-08-02 05:25:41.159644
22037	4	1-6395	22034	60	23	2736	1	2012-08-02 05:25:41.159644
22034	4	1-6430	22034	23	97	5567	91	2012-08-02 05:25:41.159644
22039	5	1-5733	22038	13	49	407	37	2012-08-02 05:25:41.159644
22040	5	1-6379	22038	49	42	810	44	2012-08-02 05:25:41.159644
22041	5	1-3793	22038	42	19	218	21	2012-08-02 05:25:41.159644
22042	5	1-6216	22038	19	23	120	1	2012-08-02 05:25:41.159644
22038	5	1-6430	22038	23	13	252	91	2012-08-02 05:25:41.159644
22044	5	1-6183	22043	26	80	658	11	2012-08-02 05:25:41.159644
22045	5	1-5499	22043	80	60	378	13	2012-08-02 05:25:41.159644
22046	5	1-4389	22043	60	19	199	21	2012-08-02 05:25:41.159644
22047	5	1-6216	22043	19	23	119	1	2012-08-02 05:25:41.159644
22043	5	1-6430	22043	23	26	271	91	2012-08-02 05:25:41.159644
22048	5	1-6431	22048	93	17	1707	74	2012-08-02 05:25:41.568516
22049	5	1-5408	22048	17	84	882	34	2012-08-02 05:25:41.568516
22050	5	1-6236	22048	84	17	1169	46	2012-08-02 05:25:41.568516
22051	5	1-6325	22048	17	4	974	6	2012-08-02 05:25:41.568516
22052	5	1-3765	22048	4	93	832	56	2012-08-02 05:25:41.568516
22054	5	1-6311	22053	47	23	2769	50	2012-08-02 05:25:41.568516
22055	5	1-6093	22053	23	82	2240	4	2012-08-02 05:25:41.568516
22056	5	1-6089	22053	82	65	1130	90	2012-08-02 05:25:41.568516
22057	5	1-6321	22053	65	93	1163	56	2012-08-02 05:25:41.568516
22053	5	1-6431	22053	93	47	2563	74	2012-08-02 05:25:41.568516
22059	4	1-6183	22058	26	80	1553	11	2012-08-02 05:25:41.977529
22060	4	1-5499	22058	80	60	856	13	2012-08-02 05:25:41.977529
22061	4	1-4389	22058	60	47	431	21	2012-08-02 05:25:41.977529
22058	4	1-6432	22058	47	26	667	91	2012-08-02 05:25:41.977529
22063	4	1-5996	22062	1	80	80	11	2012-08-02 05:25:41.977529
22064	4	1-5499	22062	80	60	45	13	2012-08-02 05:25:41.977529
22065	4	1-4389	22062	60	47	23	21	2012-08-02 05:25:41.977529
22062	4	1-6432	22062	47	1	37	91	2012-08-02 05:25:41.977529
22067	5	1-5925	22066	17	93	24	56	2012-08-02 05:25:42.386378
22068	5	1-6431	22066	93	47	52	74	2012-08-02 05:25:42.386378
22069	5	1-6311	22066	47	23	55	50	2012-08-02 05:25:42.386378
22070	5	1-6093	22066	23	35	43	4	2012-08-02 05:25:42.386378
22066	5	1-6433	22066	35	17	31	9	2012-08-02 05:25:42.386378
22072	5	1-5091	22071	35	79	308	97	2012-08-02 05:25:47.180911
22073	5	1-5845	22071	79	71	202	30	2012-08-02 05:25:47.180911
22074	5	1-4497	22071	71	58	133	27	2012-08-02 05:25:47.180911
22075	5	1-6357	22071	58	8	160	23	2012-08-02 05:25:47.180911
22071	5	1-6445	22071	8	35	488	7	2012-08-02 05:25:47.180911
22077	5	1-6304	22076	11	19	2415	66	2012-08-02 05:25:47.180911
22078	5	1-5939	22076	19	59	890	71	2012-08-02 05:25:47.180911
22079	5	1-4450	22076	59	36	447	78	2012-08-02 05:25:47.180911
22080	5	1-5554	22076	36	8	317	23	2012-08-02 05:25:47.180911
22076	5	1-6445	22076	8	11	1015	7	2012-08-02 05:25:47.180911
22082	4	1-5264	22081	66	37	3778	58	2012-08-02 05:25:47.987605
22083	4	1-6397	22081	37	55	3795	17	2012-08-02 05:25:47.987605
22084	4	1-5787	22081	55	94	1694	87	2012-08-02 05:25:47.987605
22081	4	1-6447	22081	94	66	4310	96	2012-08-02 05:25:47.987605
22086	4	1-5818	22085	74	27	61	61	2012-08-02 05:25:47.987605
22087	4	1-6241	22085	27	4	55	100	2012-08-02 05:25:47.987605
22088	4	1-6180	22085	4	94	53	87	2012-08-02 05:25:47.987605
22085	4	1-6447	22085	94	74	150	96	2012-08-02 05:25:47.987605
22090	3	1-5234	22089	90	37	3086	58	2012-08-02 05:25:50.760794
22091	3	1-6397	22089	37	98	2025	17	2012-08-02 05:25:50.760794
22089	3	1-6454	22089	98	90	3869	38	2012-08-02 05:25:50.760794
22093	4	1-5249	22092	46	82	1864	71	2012-08-02 05:25:50.760794
22094	4	1-6370	22092	82	35	2384	95	2012-08-02 05:25:50.760794
22095	4	1-6259	22092	35	98	1173	17	2012-08-02 05:25:50.760794
22092	4	1-6454	22092	98	46	3897	38	2012-08-02 05:25:50.760794
22097	5	1-4648	22096	61	17	1054	20	2012-08-02 05:25:52.329491
22098	5	1-6294	22096	17	91	338	99	2012-08-02 05:25:52.329491
22099	5	1-6427	22096	91	8	477	23	2012-08-02 05:25:52.329491
22100	5	1-6445	22096	8	4	1519	7	2012-08-02 05:25:52.329491
22096	5	1-6458	22096	4	61	3178	59	2012-08-02 05:25:52.329491
22102	3	1-5571	22101	78	80	2227	80	2012-08-02 05:25:53.547147
22103	3	1-4060	22101	80	25	1617	14	2012-08-02 05:25:53.547147
22101	3	1-6461	22101	25	78	1380	33	2012-08-02 05:25:53.547147
22105	2	1-6113	22104	100	38	3025	15	2012-08-02 05:25:55.535426
22104	2	1-6466	22104	38	100	2872	59	2012-08-02 05:25:55.535426
22107	4	1-5889	22106	5	71	2012	37	2012-08-02 05:25:56.308787
22108	4	1-6411	22106	71	3	1640	38	2012-08-02 05:25:56.308787
22109	4	1-3487	22106	3	4	1060	46	2012-08-02 05:25:56.308787
22106	4	1-6468	22106	4	5	1874	41	2012-08-02 05:25:56.308787
22111	4	1-5916	22110	86	72	98	8	2012-08-02 05:25:56.308787
22112	4	1-5798	22110	72	84	174	34	2012-08-02 05:25:56.308787
22113	4	1-6236	22110	84	4	186	46	2012-08-02 05:25:56.308787
22110	4	1-6468	22110	4	86	362	41	2012-08-02 05:25:56.308787
22115	4	1-3861	22114	6	1	732	32	2012-08-02 05:25:56.308787
22116	4	1-6210	22114	1	77	1786	20	2012-08-02 05:25:56.308787
22117	4	1-6419	22114	77	4	892	46	2012-08-02 05:25:56.308787
22114	4	1-6468	22114	4	6	2026	41	2012-08-02 05:25:56.308787
22119	5	1-5130	22118	51	25	3337	14	2012-08-02 05:25:57.082457
22120	5	1-6461	22118	25	51	1327	33	2012-08-02 05:25:57.082457
22121	5	1-3524	22118	51	100	373	23	2012-08-02 05:25:57.082457
22122	5	1-5898	22118	100	89	182	28	2012-08-02 05:25:57.082457
22118	5	1-6470	22118	89	51	2137	55	2012-08-02 05:25:57.082457
22124	4	1-4767	22123	82	66	533	61	2012-08-02 05:25:57.082457
22125	4	1-6182	22123	66	39	629	86	2012-08-02 05:25:57.082457
22126	4	1-5185	22123	39	89	189	28	2012-08-02 05:25:57.082457
22123	4	1-6470	22123	89	82	2466	55	2012-08-02 05:25:57.082457
22128	4	1-5733	22127	13	71	740	37	2012-08-02 05:25:59.46881
22129	4	1-6411	22127	71	90	564	38	2012-08-02 05:25:59.46881
22130	4	1-5234	22127	90	67	636	58	2012-08-02 05:25:59.46881
22127	4	1-6476	22127	67	13	585	91	2012-08-02 05:25:59.46881
22132	4	1-3121	22131	67	38	2952	39	2012-08-02 05:25:59.46881
22133	4	1-4354	22131	38	24	2596	46	2012-08-02 05:25:59.46881
22134	4	1-6455	22131	24	67	2196	58	2012-08-02 05:25:59.46881
22131	4	1-6476	22131	67	67	2804	91	2012-08-02 05:25:59.46881
22136	4	1-5996	22135	1	37	3814	11	2012-08-02 05:25:59.46881
22137	4	1-516	22135	37	24	1155	46	2012-08-02 05:25:59.46881
22138	4	1-6455	22135	24	67	1193	58	2012-08-02 05:25:59.46881
22135	4	1-6476	22135	67	1	1616	91	2012-08-02 05:25:59.46881
22140	5	1-6475	22139	17	3	1030	5	2012-08-02 05:26:00.242302
22141	5	1-5738	22139	3	53	709	54	2012-08-02 05:26:00.242302
22142	5	1-6135	22139	53	61	229	70	2012-08-02 05:26:00.242302
22143	5	1-5895	22139	61	21	142	98	2012-08-02 05:26:00.242302
22139	5	1-6478	22139	21	17	174	97	2012-08-02 05:26:00.242302
22145	5	1-5845	22144	79	59	1040	30	2012-08-02 05:26:00.242302
22146	5	1-2485	22144	59	9	1299	73	2012-08-02 05:26:00.242302
22147	5	1-6408	22144	9	61	999	70	2012-08-02 05:26:00.242302
22148	5	1-5895	22144	61	21	865	98	2012-08-02 05:26:00.242302
22144	5	1-6478	22144	21	79	1473	97	2012-08-02 05:26:00.242302
22150	4	1-4199	22149	71	28	2487	52	2012-08-02 05:26:01.435501
22151	4	1-6453	22149	28	8	3617	9	2012-08-02 05:26:01.435501
22152	4	1-6380	22149	8	27	2460	100	2012-08-02 05:26:01.435501
22149	4	1-6481	22149	27	71	4389	41	2012-08-02 05:26:01.435501
22154	5	1-4513	22153	14	8	686	43	2012-08-02 05:26:01.435501
22155	5	1-5928	22153	8	85	526	75	2012-08-02 05:26:01.435501
22156	5	1-5776	22153	85	94	546	18	2012-08-02 05:26:01.435501
22157	5	1-6232	22153	94	27	892	100	2012-08-02 05:26:01.435501
22153	5	1-6481	22153	27	14	1629	41	2012-08-02 05:26:01.435501
22159	5	1-3861	22158	6	1	41	32	2012-08-02 05:26:01.435501
22160	5	1-6210	22158	1	77	112	20	2012-08-02 05:26:01.435501
22161	5	1-6273	22158	77	27	45	82	2012-08-02 05:26:01.435501
22162	5	1-4087	22158	27	27	53	100	2012-08-02 05:26:01.435501
22158	5	1-6481	22158	27	6	101	41	2012-08-02 05:26:01.435501
22164	4	1-5131	22163	73	67	1041	58	2012-08-02 05:26:01.855646
22165	4	1-6476	22163	67	55	697	91	2012-08-02 05:26:01.855646
22166	4	1-6265	22163	55	35	1365	15	2012-08-02 05:26:01.855646
22163	4	1-6482	22163	35	73	1412	7	2012-08-02 05:26:01.855646
22168	4	1-6304	22167	11	87	3882	66	2012-08-02 05:26:01.855646
22169	4	1-6006	22167	87	63	3739	63	2012-08-02 05:26:01.855646
22170	4	1-4600	22167	63	35	2000	15	2012-08-02 05:26:01.855646
22167	4	1-6482	22167	35	11	2661	7	2012-08-02 05:26:01.855646
22172	5	1-4900	22171	84	21	141	49	2012-08-02 05:26:01.855646
22173	5	1-5862	22171	21	67	85	58	2012-08-02 05:26:01.855646
22174	5	1-6476	22171	67	55	79	91	2012-08-02 05:26:01.855646
22175	5	1-6265	22171	55	35	213	15	2012-08-02 05:26:01.855646
22171	5	1-6482	22171	35	84	306	7	2012-08-02 05:26:01.855646
22177	3	1-6169	22176	7	4	1337	100	2012-08-02 05:26:02.242599
22178	3	1-6180	22176	4	37	393	87	2012-08-02 05:26:02.242599
22176	3	1-6483	22176	37	7	6165	71	2012-08-02 05:26:02.242599
22180	5	1-2578	22179	11	26	1334	2	2012-08-02 05:26:03.446696
22181	5	1-5959	22179	26	85	1008	20	2012-08-02 05:26:03.446696
22182	5	1-6337	22179	85	66	846	6	2012-08-02 05:26:03.446696
22183	5	1-5468	22179	66	88	828	77	2012-08-02 05:26:03.446696
22179	5	1-6486	22179	88	11	2047	86	2012-08-02 05:26:03.446696
22185	5	1-3958	22184	5	93	1277	55	2012-08-02 05:26:03.446696
22186	5	1-5794	22184	93	17	1389	46	2012-08-02 05:26:03.446696
22187	5	1-6325	22184	17	66	1115	6	2012-08-02 05:26:03.446696
22188	5	1-5468	22184	66	88	1141	77	2012-08-02 05:26:03.446696
22184	5	1-6486	22184	88	5	2952	86	2012-08-02 05:26:03.446696
22190	4	1-6478	22189	21	40	3807	97	2012-08-02 05:26:05.037834
22191	4	1-3759	22189	40	68	698	32	2012-08-02 05:26:05.037834
22192	4	1-5078	22189	68	91	451	69	2012-08-02 05:26:05.037834
22189	4	1-6490	22189	91	21	3510	98	2012-08-02 05:26:05.037834
22194	4	1-3341	22193	15	10	2690	13	2012-08-02 05:26:05.037834
22195	4	1-6464	22193	10	68	834	32	2012-08-02 05:26:05.037834
22196	4	1-5078	22193	68	91	604	69	2012-08-02 05:26:05.037834
22193	4	1-6490	22193	91	15	5267	98	2012-08-02 05:26:05.037834
22198	5	1-2431	22197	11	25	321	61	2012-08-02 05:26:08.26521
22199	5	1-5926	22197	25	74	257	82	2012-08-02 05:26:08.26521
22200	5	1-6417	22197	74	64	295	12	2012-08-02 05:26:08.26521
22201	5	1-4486	22197	64	30	240	78	2012-08-02 05:26:08.26521
22197	5	1-6498	22197	30	11	2236	98	2012-08-02 05:26:08.26521
22203	4	1-3341	22202	15	23	658	13	2012-08-02 05:26:08.26521
22204	4	1-6240	22202	23	47	201	96	2012-08-02 05:26:08.26521
22205	4	1-5353	22202	47	30	116	78	2012-08-02 05:26:08.26521
22202	4	1-6498	22202	30	15	1218	98	2012-08-02 05:26:08.26521
22207	5	1-3794	22206	42	17	1285	25	2012-08-02 05:26:08.26521
22208	5	1-6268	22206	17	23	681	61	2012-08-02 05:26:08.26521
22209	5	1-6205	22206	23	46	602	55	2012-08-02 05:26:08.26521
22210	5	1-6278	22206	46	30	531	78	2012-08-02 05:26:08.26521
22206	5	1-6498	22206	30	42	5969	98	2012-08-02 05:26:08.26521
22212	3	1-6110	22211	56	80	1159	11	2012-08-02 05:26:09.49156
22213	3	1-5499	22211	80	77	537	13	2012-08-02 05:26:09.49156
22211	3	1-6501	22211	77	56	889	7	2012-08-02 05:26:09.49156
22215	5	1-5091	22214	35	17	1428	97	2012-08-02 05:26:09.49156
22216	5	1-6475	22214	17	36	8712	5	2012-08-02 05:26:09.49156
22217	5	1-5800	22214	36	1	2922	86	2012-08-02 05:26:09.49156
22218	5	1-717	22214	1	77	1656	13	2012-08-02 05:26:09.49156
22214	5	1-6501	22214	77	35	2837	7	2012-08-02 05:26:09.49156
22220	5	1-4900	22219	84	60	551	49	2012-08-02 05:26:09.49156
22221	5	1-6303	22219	60	23	857	88	2012-08-02 05:26:09.49156
22222	5	1-6480	22219	23	16	508	68	2012-08-02 05:26:09.49156
22223	5	1-2286	22219	16	77	407	13	2012-08-02 05:26:09.49156
22219	5	1-6501	22219	77	84	851	7	2012-08-02 05:26:09.49156
22225	4	1-6223	22224	38	25	16	93	2012-08-02 05:26:09.911619
22226	4	1-4812	22224	25	44	4	30	2012-08-02 05:26:09.911619
22227	4	1-6207	22224	44	51	3	24	2012-08-02 05:26:09.911619
22224	4	1-6502	22224	51	38	11	90	2012-08-02 05:26:09.911619
22229	5	1-6261	22228	75	13	1490	76	2012-08-02 05:26:09.911619
22230	5	1-6441	22228	13	46	1552	38	2012-08-02 05:26:09.911619
22231	5	1-5249	22228	46	96	735	71	2012-08-02 05:26:09.911619
22232	5	1-6197	22228	96	51	702	24	2012-08-02 05:26:09.911619
22228	5	1-6502	22228	51	75	2613	90	2012-08-02 05:26:09.911619
22234	4	1-6321	22233	65	92	378	56	2012-08-02 05:26:09.911619
22235	4	1-4773	22233	92	44	132	30	2012-08-02 05:26:09.911619
22236	4	1-6207	22233	44	51	96	24	2012-08-02 05:26:09.911619
22233	4	1-6502	22233	51	65	411	90	2012-08-02 05:26:09.911619
22238	4	1-6386	22237	9	55	3699	79	2012-08-02 05:26:10.740538
22239	4	1-6072	22237	55	97	2784	66	2012-08-02 05:26:10.740538
22240	4	1-5631	22237	97	71	723	18	2012-08-02 05:26:10.740538
22237	4	1-6504	22237	71	9	4662	27	2012-08-02 05:26:10.740538
22242	3	1-5840	22241	66	53	4437	85	2012-08-02 05:26:11.580013
22243	3	1-5771	22241	53	64	1597	54	2012-08-02 05:26:11.580013
22241	3	1-6506	22241	64	66	2287	75	2012-08-02 05:26:11.580013
22245	5	1-5698	22244	42	23	1510	50	2012-08-02 05:26:12.375431
22246	5	1-6093	22244	23	34	979	4	2012-08-02 05:26:12.375431
22247	5	1-6051	22244	34	40	682	22	2012-08-02 05:26:12.375431
22248	5	1-6391	22244	40	80	878	73	2012-08-02 05:26:12.375431
22244	5	1-6508	22244	80	42	1781	96	2012-08-02 05:26:12.375431
22250	5	1-5818	22249	74	66	1683	61	2012-08-02 05:26:12.375431
22251	5	1-6182	22249	66	63	3621	86	2012-08-02 05:26:12.375431
22252	5	1-2482	22249	63	6	1704	60	2012-08-02 05:26:12.375431
22253	5	1-5290	22249	6	80	1605	73	2012-08-02 05:26:12.375431
22249	5	1-6508	22249	80	74	3621	96	2012-08-02 05:26:12.375431
22255	4	1-5962	22254	2	27	632	10	2012-08-02 05:26:12.375431
22256	4	1-5391	22254	27	70	394	39	2012-08-02 05:26:12.375431
22257	4	1-5867	22254	70	80	281	73	2012-08-02 05:26:12.375431
22254	4	1-6508	22254	80	2	650	96	2012-08-02 05:26:12.375431
22259	5	1-6446	22258	4	62	1215	79	2012-08-02 05:26:12.795513
22260	5	1-3032	22258	62	73	807	2	2012-08-02 05:26:12.795513
22261	5	1-6262	22258	73	1	308	32	2012-08-02 05:26:12.795513
22262	5	1-6210	22258	1	71	862	20	2012-08-02 05:26:12.795513
22258	5	1-6509	22258	71	4	1471	7	2012-08-02 05:26:12.795513
22264	5	1-3377	22263	38	74	1784	21	2012-08-02 05:26:13.20429
22265	5	1-6485	22263	74	82	1611	56	2012-08-02 05:26:13.20429
22266	5	1-6005	22263	82	22	1823	96	2012-08-02 05:26:13.20429
22267	5	1-4673	22263	22	67	1638	49	2012-08-02 05:26:13.20429
22263	5	1-6510	22263	67	38	2327	30	2012-08-02 05:26:13.20429
22269	5	1-6261	22268	75	57	1053	76	2012-08-02 05:26:13.624157
22270	5	1-6228	22268	57	67	667	58	2012-08-02 05:26:13.624157
22271	5	1-6476	22268	67	1	680	91	2012-08-02 05:26:13.624157
22272	5	1-5996	22268	1	68	1209	11	2012-08-02 05:26:13.624157
22268	5	1-6511	22268	68	75	2025	90	2012-08-02 05:26:13.624157
22274	5	1-6154	22273	3	15	1242	98	2012-08-02 05:26:13.624157
22275	5	1-3341	22273	15	77	666	13	2012-08-02 05:26:13.624157
22276	5	1-6501	22273	77	56	1151	7	2012-08-02 05:26:13.624157
22277	5	1-6110	22273	56	68	1568	11	2012-08-02 05:26:13.624157
22273	5	1-6511	22273	68	3	2770	90	2012-08-02 05:26:13.624157
22279	4	1-5528	22278	30	68	662	87	2012-08-02 05:26:13.624157
22280	4	1-6225	22278	68	84	1003	80	2012-08-02 05:26:13.624157
22281	4	1-6459	22278	84	68	480	11	2012-08-02 05:26:13.624157
22278	4	1-6511	22278	68	30	965	90	2012-08-02 05:26:13.624157
22283	4	1-6321	22282	65	58	1186	56	2012-08-02 05:26:13.624157
22284	4	1-2344	22282	58	55	430	79	2012-08-02 05:26:13.624157
22285	4	1-6149	22282	55	68	656	11	2012-08-02 05:26:13.624157
22282	4	1-6511	22282	68	65	1337	90	2012-08-02 05:26:13.624157
22287	3	1-6410	22286	87	24	115	64	2012-08-02 05:26:13.624157
22288	3	1-3246	22286	24	68	69	11	2012-08-02 05:26:13.624157
22286	3	1-6511	22286	68	87	148	90	2012-08-02 05:26:13.624157
22290	5	1-5702	22289	36	56	1084	42	2012-08-02 05:26:14.862037
22291	5	1-5661	22289	56	28	1491	51	2012-08-02 05:26:14.862037
22292	5	1-6401	22289	28	74	88	24	2012-08-02 05:26:14.862037
22293	5	1-6319	22289	74	30	460	10	2012-08-02 05:26:14.862037
22289	5	1-6514	22289	30	36	1517	87	2012-08-02 05:26:14.862037
22295	5	1-5896	22294	80	44	717	16	2012-08-02 05:26:14.862037
22296	5	1-5900	22294	44	28	1072	51	2012-08-02 05:26:14.862037
22297	5	1-6401	22294	28	74	76	24	2012-08-02 05:26:14.862037
22298	5	1-6319	22294	74	30	477	10	2012-08-02 05:26:14.862037
22294	5	1-6514	22294	30	80	1892	87	2012-08-02 05:26:14.862037
22300	3	1-5896	22299	80	72	1220	16	2012-08-02 05:26:14.862037
22301	3	1-5583	22299	72	30	800	10	2012-08-02 05:26:14.862037
22299	3	1-6514	22299	30	80	3196	87	2012-08-02 05:26:14.862037
22303	4	1-6225	22302	68	89	1733	80	2012-08-02 05:26:14.862037
22304	4	1-6403	22302	89	97	464	9	2012-08-02 05:26:14.862037
22305	4	1-3753	22302	97	30	212	10	2012-08-02 05:26:14.862037
22302	4	1-6514	22302	30	68	995	87	2012-08-02 05:26:14.862037
22307	4	1-6396	22306	10	41	2858	5	2012-08-02 05:26:14.862037
22308	4	1-5788	22306	41	86	1458	95	2012-08-02 05:26:14.862037
22309	4	1-2755	22306	86	30	274	10	2012-08-02 05:26:14.862037
22306	4	1-6514	22306	30	10	1321	87	2012-08-02 05:26:14.862037
22311	4	1-5145	22310	20	49	575	37	2012-08-02 05:26:16.11233
22312	4	1-6379	22310	49	2	1319	44	2012-08-02 05:26:16.11233
22313	4	1-5257	22310	2	41	39	41	2012-08-02 05:26:16.11233
22310	4	1-6517	22310	41	20	525	39	2012-08-02 05:26:16.11233
22315	5	1-5541	22314	51	64	2900	44	2012-08-02 05:26:16.11233
22316	5	1-5687	22314	64	78	1757	42	2012-08-02 05:26:16.11233
22317	5	1-3066	22314	78	17	237	96	2012-08-02 05:26:16.11233
22318	5	1-6358	22314	17	41	236	41	2012-08-02 05:26:16.11233
22314	5	1-6517	22314	41	51	3271	39	2012-08-02 05:26:16.11233
22320	4	1-5145	22319	20	56	522	37	2012-08-02 05:26:16.11233
22321	4	1-6043	22319	56	61	240	10	2012-08-02 05:26:16.11233
22322	4	1-4221	22319	61	41	32	41	2012-08-02 05:26:16.11233
22319	4	1-6517	22319	41	20	454	39	2012-08-02 05:26:16.11233
22324	5	1-2189	22323	45	100	434	36	2012-08-02 05:26:16.11233
22325	5	1-5084	22323	100	82	131	56	2012-08-02 05:26:16.11233
22326	5	1-6005	22323	82	17	148	96	2012-08-02 05:26:16.11233
22327	5	1-6358	22323	17	41	151	41	2012-08-02 05:26:16.11233
22323	5	1-6517	22323	41	45	2166	39	2012-08-02 05:26:16.11233
22329	5	1-6433	22328	35	49	866	9	2012-08-02 05:26:17.361149
22330	5	1-6415	22328	49	62	503	75	2012-08-02 05:26:17.361149
22331	5	1-5712	22328	62	20	387	84	2012-08-02 05:26:17.361149
22332	5	1-6416	22328	20	12	1871	83	2012-08-02 05:26:17.361149
22328	5	1-6520	22328	12	35	1266	4	2012-08-02 05:26:17.361149
22334	2	1-5923	22333	3	87	1483	62	2012-08-02 05:26:19.868796
22333	2	1-6526	22333	87	3	4099	74	2012-08-02 05:26:19.868796
22336	3	1-6472	22335	42	38	1488	30	2012-08-02 05:26:20.277474
22337	3	1-3377	22335	38	4	964	21	2012-08-02 05:26:20.277474
22335	3	1-6527	22335	4	42	2022	78	2012-08-02 05:26:20.277474
22339	4	1-6472	22338	42	44	3169	30	2012-08-02 05:26:20.277474
22340	4	1-6207	22338	44	31	2345	24	2012-08-02 05:26:20.277474
22341	4	1-4102	22338	31	4	1810	21	2012-08-02 05:26:20.277474
22338	4	1-6527	22338	4	42	4044	78	2012-08-02 05:26:20.277474
22343	5	1-4728	22342	83	25	5320	14	2012-08-02 05:26:21.537143
22344	5	1-6461	22342	25	75	3646	33	2012-08-02 05:26:21.537143
22345	5	1-1009	22342	75	44	1088	72	2012-08-02 05:26:21.537143
22346	5	1-6305	22342	44	8	1107	28	2012-08-02 05:26:21.537143
22342	5	1-6530	22342	8	83	7268	63	2012-08-02 05:26:21.537143
22348	4	1-5783	22347	44	73	1666	48	2012-08-02 05:26:23.603351
22349	4	1-5757	22347	73	73	836	89	2012-08-02 05:26:23.603351
22350	4	1-6313	22347	73	99	1447	14	2012-08-02 05:26:23.603351
22347	4	1-6535	22347	99	44	1688	94	2012-08-02 05:26:23.603351
22352	3	1-5574	22351	80	83	827	63	2012-08-02 05:26:23.603351
22353	3	1-4728	22351	83	99	660	14	2012-08-02 05:26:23.603351
22351	3	1-6535	22351	99	80	1050	94	2012-08-02 05:26:23.603351
22355	3	1-5574	22354	80	49	5383	63	2012-08-02 05:26:23.603351
22356	3	1-357	22354	49	99	3868	14	2012-08-02 05:26:23.603351
22354	3	1-6535	22354	99	80	6488	94	2012-08-02 05:26:23.603351
22358	4	1-6451	22357	87	48	373	87	2012-08-02 05:26:23.603351
22359	4	1-5751	22357	48	99	94	66	2012-08-02 05:26:23.603351
22360	4	1-6405	22357	99	99	264	14	2012-08-02 05:26:23.603351
22357	4	1-6535	22357	99	87	510	94	2012-08-02 05:26:23.603351
22362	2	1-5635	22361	44	20	447	92	2012-08-02 05:26:24.829477
22361	2	1-6538	22361	20	44	4608	5	2012-08-02 05:26:24.829477
22364	5	1-6339	22363	67	53	1697	59	2012-08-02 05:26:25.238686
22365	5	1-4705	22363	53	67	1480	58	2012-08-02 05:26:25.238686
22366	5	1-6476	22363	67	13	2003	91	2012-08-02 05:26:25.238686
22367	5	1-5733	22363	13	25	3431	37	2012-08-02 05:26:25.238686
22363	5	1-6539	22363	25	67	1805	38	2012-08-02 05:26:25.238686
22369	5	1-3782	22368	44	41	476	39	2012-08-02 05:26:26.055808
22370	5	1-6528	22368	41	67	565	58	2012-08-02 05:26:26.055808
22371	5	1-6476	22368	67	55	452	91	2012-08-02 05:26:26.055808
22372	5	1-6265	22368	55	73	1057	15	2012-08-02 05:26:26.055808
22368	5	1-6541	22368	73	44	1085	9	2012-08-02 05:26:26.055808
22374	4	1-5726	22373	84	71	6234	37	2012-08-02 05:26:26.055808
22375	4	1-6411	22373	71	57	4900	38	2012-08-02 05:26:26.055808
22376	4	1-6107	22373	57	73	4970	15	2012-08-02 05:26:26.055808
22373	4	1-6541	22373	73	84	6051	9	2012-08-02 05:26:26.055808
22378	3	1-4906	22377	9	88	1603	40	2012-08-02 05:26:26.055808
22379	3	1-5393	22377	88	73	1336	15	2012-08-02 05:26:26.055808
22377	3	1-6541	22377	73	9	1786	9	2012-08-02 05:26:26.055808
22381	3	1-6488	22380	77	100	676	59	2012-08-02 05:26:26.055808
22382	3	1-6113	22380	100	73	564	15	2012-08-02 05:26:26.055808
22380	3	1-6541	22380	73	77	754	9	2012-08-02 05:26:26.055808
22384	4	1-5363	22383	8	71	1780	72	2012-08-02 05:26:26.453943
22385	4	1-6346	22383	71	36	1330	47	2012-08-02 05:26:26.453943
22386	4	1-6505	22383	36	64	1720	36	2012-08-02 05:26:26.453943
22383	4	1-6542	22383	64	8	2878	68	2012-08-02 05:26:26.453943
22388	4	1-5329	22387	9	36	843	87	2012-08-02 05:26:26.453943
22389	4	1-5702	22387	36	65	757	42	2012-08-02 05:26:26.453943
22390	4	1-6418	22387	65	64	777	36	2012-08-02 05:26:26.453943
22387	4	1-6542	22387	64	9	1341	68	2012-08-02 05:26:26.453943
22392	5	1-6002	22391	71	20	1430	83	2012-08-02 05:26:26.453943
22393	5	1-5797	22391	20	23	1567	89	2012-08-02 05:26:26.453943
22394	5	1-3137	22391	23	65	630	42	2012-08-02 05:26:26.453943
22395	5	1-6418	22391	65	64	703	36	2012-08-02 05:26:26.453943
22391	5	1-6542	22391	64	71	1319	68	2012-08-02 05:26:26.453943
22397	5	1-2286	22396	16	53	1437	13	2012-08-02 05:26:26.453943
22398	5	1-6371	22396	53	48	1238	2	2012-08-02 05:26:26.453943
22399	5	1-5981	22396	48	36	557	47	2012-08-02 05:26:26.453943
22400	5	1-6505	22396	36	64	844	36	2012-08-02 05:26:26.453943
22396	5	1-6542	22396	64	16	1655	68	2012-08-02 05:26:26.453943
22402	5	1-6104	22401	89	56	60	32	2012-08-02 05:26:26.453943
22403	5	1-5899	22401	56	48	70	2	2012-08-02 05:26:26.453943
22404	5	1-5981	22401	48	36	32	47	2012-08-02 05:26:26.453943
22405	5	1-6505	22401	36	64	49	36	2012-08-02 05:26:26.453943
22401	5	1-6542	22401	64	89	98	68	2012-08-02 05:26:26.453943
22407	4	1-6039	22406	14	24	1131	46	2012-08-02 05:26:26.885069
22408	4	1-6455	22406	24	65	1087	58	2012-08-02 05:26:26.885069
22409	4	1-6387	22406	65	42	1239	5	2012-08-02 05:26:26.885069
22406	4	1-6543	22406	42	14	2197	31	2012-08-02 05:26:26.885069
22411	5	1-6101	22410	87	100	846	45	2012-08-02 05:26:27.293887
22412	5	1-6412	22410	100	55	1008	79	2012-08-02 05:26:27.293887
22413	5	1-6149	22410	55	74	1498	11	2012-08-02 05:26:27.293887
22414	5	1-5806	22410	74	96	823	84	2012-08-02 05:26:27.293887
22410	5	1-6544	22410	96	87	537	61	2012-08-02 05:26:27.293887
22416	3	1-5926	22415	25	86	674	82	2012-08-02 05:26:27.293887
22417	3	1-5662	22415	86	96	951	84	2012-08-02 05:26:27.293887
22415	3	1-6544	22415	96	25	645	61	2012-08-02 05:26:27.293887
22419	4	1-6101	22418	87	46	8	45	2012-08-02 05:26:27.293887
22420	4	1-5460	22418	46	63	21	63	2012-08-02 05:26:27.293887
22421	4	1-4744	22418	63	96	7	84	2012-08-02 05:26:27.293887
22418	4	1-6544	22418	96	87	5	61	2012-08-02 05:26:27.293887
22423	3	1-5818	22422	74	66	271	61	2012-08-02 05:26:28.111468
22424	3	1-6182	22422	66	99	517	86	2012-08-02 05:26:28.111468
22422	3	1-6546	22422	99	74	656	96	2012-08-02 05:26:28.111468
22426	4	1-5264	22425	66	65	1320	58	2012-08-02 05:26:28.111468
22427	4	1-6387	22425	65	36	1641	5	2012-08-02 05:26:28.111468
22428	4	1-5800	22425	36	99	711	86	2012-08-02 05:26:28.111468
22425	4	1-6546	22425	99	66	1089	96	2012-08-02 05:26:28.111468
22430	3	1-5468	22429	66	45	3191	77	2012-08-02 05:26:28.962583
22431	3	1-6383	22429	45	84	3309	37	2012-08-02 05:26:28.962583
22429	3	1-6548	22429	84	66	3750	6	2012-08-02 05:26:28.962583
22433	4	1-4317	22432	97	35	680	4	2012-08-02 05:26:28.962583
22434	4	1-6433	22432	35	84	442	9	2012-08-02 05:26:28.962583
22435	4	1-5726	22432	84	84	653	37	2012-08-02 05:26:28.962583
22432	4	1-6548	22432	84	97	897	6	2012-08-02 05:26:28.962583
22437	5	1-5777	22436	23	30	651	16	2012-08-02 05:26:28.962583
22438	5	1-4954	22436	30	28	302	52	2012-08-02 05:26:28.962583
22439	5	1-6453	22436	28	84	434	9	2012-08-02 05:26:28.962583
22440	5	1-5726	22436	84	84	657	37	2012-08-02 05:26:28.962583
22436	5	1-6548	22436	84	23	925	6	2012-08-02 05:26:28.962583
22442	4	1-6039	22441	14	17	1294	46	2012-08-02 05:26:30.211032
22443	4	1-6325	22441	17	49	1088	6	2012-08-02 05:26:30.211032
22444	4	1-2339	22441	49	42	1157	14	2012-08-02 05:26:30.211032
22441	4	1-6551	22441	42	14	2237	31	2012-08-02 05:26:30.211032
22446	4	1-6445	22445	8	4	907	7	2012-08-02 05:26:30.608534
22447	4	1-6446	22445	4	55	341	79	2012-08-02 05:26:30.608534
22448	4	1-6072	22445	55	39	177	66	2012-08-02 05:26:30.608534
22445	4	1-6552	22445	39	8	602	23	2012-08-02 05:26:30.608534
22450	4	1-6402	22449	13	22	2443	19	2012-08-02 05:26:30.608534
22451	4	1-295	22449	22	4	935	68	2012-08-02 05:26:30.608534
22452	4	1-6085	22449	4	39	765	66	2012-08-02 05:26:30.608534
22449	4	1-6552	22449	39	13	4229	23	2012-08-02 05:26:30.608534
22453	1	1-5665	\N	19	19	708	66	2012-08-02 05:26:30.608534
22454	1	1-5752	\N	29	29	541	66	2012-08-02 05:26:30.608534
22456	5	1-4095	22455	5	13	1045	25	2012-08-02 05:26:31.437153
22457	5	1-5532	22455	13	13	1209	76	2012-08-02 05:26:31.437153
22458	5	1-6441	22455	13	57	1496	38	2012-08-02 05:26:31.437153
22459	5	1-6107	22455	57	7	1923	15	2012-08-02 05:26:31.437153
22455	5	1-6554	22455	7	5	1045	75	2012-08-02 05:26:31.437153
22461	4	1-4622	22460	67	24	499	65	2012-08-02 05:26:31.437153
22462	4	1-6239	22460	24	97	659	79	2012-08-02 05:26:31.437153
22463	4	1-3801	22460	97	7	581	15	2012-08-02 05:26:31.437153
22460	4	1-6554	22460	7	67	358	75	2012-08-02 05:26:31.437153
22465	5	1-4095	22464	5	80	1689	25	2012-08-02 05:26:31.437153
22466	5	1-5679	22464	80	13	1155	76	2012-08-02 05:26:31.437153
22467	5	1-6441	22464	13	57	1512	38	2012-08-02 05:26:31.437153
22468	5	1-6107	22464	57	7	2273	15	2012-08-02 05:26:31.437153
22464	5	1-6554	22464	7	5	1445	75	2012-08-02 05:26:31.437153
22470	5	1-4442	22469	10	17	1440	25	2012-08-02 05:26:31.83508
22471	5	1-6268	22469	17	25	966	61	2012-08-02 05:26:31.83508
22472	5	1-5926	22469	25	8	1179	82	2012-08-02 05:26:31.83508
22473	5	1-5717	22469	8	12	1430	64	2012-08-02 05:26:31.83508
22469	5	1-6555	22469	12	10	1159	47	2012-08-02 05:26:31.83508
22475	3	1-4451	22474	15	25	506	2	2012-08-02 05:26:32.652602
22476	3	1-5802	22474	25	66	73	21	2012-08-02 05:26:32.652602
22474	3	1-6557	22474	66	15	5379	50	2012-08-02 05:26:32.652602
22478	4	1-6417	22477	74	64	1185	12	2012-08-02 05:26:33.072707
22479	4	1-4486	22477	64	72	1126	78	2012-08-02 05:26:33.072707
22480	4	1-6164	22477	72	50	654	52	2012-08-02 05:26:33.072707
22477	4	1-6558	22477	50	74	883	82	2012-08-02 05:26:33.072707
22482	4	1-4402	22481	82	4	1471	21	2012-08-02 05:26:33.072707
22483	4	1-6527	22481	4	72	3001	78	2012-08-02 05:26:33.072707
22484	4	1-6164	22481	72	50	1850	52	2012-08-02 05:26:33.072707
22481	4	1-6558	22481	50	82	2647	82	2012-08-02 05:26:33.072707
22486	5	1-3656	22485	24	30	160	10	2012-08-02 05:26:33.072707
22487	5	1-6514	22485	30	80	774	87	2012-08-02 05:26:33.072707
22488	5	1-5896	22485	80	30	355	16	2012-08-02 05:26:33.072707
22489	5	1-4954	22485	30	50	180	52	2012-08-02 05:26:33.072707
22485	5	1-6558	22485	50	24	311	82	2012-08-02 05:26:33.072707
22491	5	1-5702	22490	36	52	2199	42	2012-08-02 05:26:35.503161
22492	5	1-5293	22490	52	7	2034	15	2012-08-02 05:26:35.503161
22493	5	1-6554	22490	7	54	1324	75	2012-08-02 05:26:35.503161
22494	5	1-5275	22490	54	10	2045	11	2012-08-02 05:26:35.503161
22490	5	1-6564	22490	10	36	2113	87	2012-08-02 05:26:35.503161
22496	4	1-5398	22495	88	61	4080	100	2012-08-02 05:26:35.934367
22497	4	1-6393	22495	61	87	3457	90	2012-08-02 05:26:35.934367
22498	4	1-6410	22495	87	98	2576	64	2012-08-02 05:26:35.934367
22495	4	1-6565	22495	98	88	6358	79	2012-08-02 05:26:35.934367
22500	4	1-4967	22499	16	13	1614	76	2012-08-02 05:26:36.751571
22501	4	1-6441	22499	13	42	1826	38	2012-08-02 05:26:36.751571
22502	4	1-2798	22499	42	78	1782	15	2012-08-02 05:26:36.751571
22499	4	1-6567	22499	78	16	1185	47	2012-08-02 05:26:36.751571
22504	2	1-5621	22503	73	78	5239	15	2012-08-02 05:26:36.751571
22503	2	1-6567	22503	78	73	3862	47	2012-08-02 05:26:36.751571
22506	4	1-6159	22505	7	87	1543	66	2012-08-02 05:26:36.751571
22507	4	1-6006	22505	87	63	2389	63	2012-08-02 05:26:36.751571
22508	4	1-4600	22505	63	78	2054	15	2012-08-02 05:26:36.751571
22505	4	1-6567	22505	78	7	1557	47	2012-08-02 05:26:36.751571
22510	4	1-4442	22509	10	27	527	25	2012-08-02 05:26:36.751571
22511	4	1-4564	22509	27	100	457	59	2012-08-02 05:26:36.751571
22512	4	1-6113	22509	100	78	534	15	2012-08-02 05:26:36.751571
22509	4	1-6567	22509	78	10	426	47	2012-08-02 05:26:36.751571
22514	4	1-6546	22513	99	81	3971	96	2012-08-02 05:26:37.547286
22515	4	1-4228	22513	81	5	1541	34	2012-08-02 05:26:37.547286
22516	4	1-6356	22513	5	37	1541	65	2012-08-02 05:26:37.547286
22513	4	1-6569	22513	37	99	2821	86	2012-08-02 05:26:37.547286
22518	4	1-3958	22517	5	6	466	55	2012-08-02 05:26:37.547286
22519	4	1-5374	22517	6	5	528	34	2012-08-02 05:26:37.547286
22520	4	1-6356	22517	5	37	573	65	2012-08-02 05:26:37.547286
22517	4	1-6569	22517	37	5	1062	86	2012-08-02 05:26:37.547286
22522	5	1-6520	22521	12	45	709	4	2012-08-02 05:26:38.353881
22523	5	1-6133	22521	45	35	556	95	2012-08-02 05:26:38.353881
22524	5	1-6259	22521	35	64	282	17	2012-08-02 05:26:38.353881
22525	5	1-6462	22521	64	82	233	89	2012-08-02 05:26:38.353881
22521	5	1-6571	22521	82	12	1195	83	2012-08-02 05:26:38.353881
22527	5	1-6407	22526	27	89	1386	56	2012-08-02 05:26:40.397758
22528	5	1-6274	22526	89	87	2014	94	2012-08-02 05:26:40.397758
22529	5	1-6451	22526	87	36	1465	87	2012-08-02 05:26:40.397758
22530	5	1-5702	22526	36	58	1491	42	2012-08-02 05:26:40.397758
22526	5	1-6576	22526	58	27	1567	1	2012-08-02 05:26:40.397758
22532	5	1-6407	22531	27	43	1701	56	2012-08-02 05:26:40.397758
22533	5	1-4370	22531	43	4	1329	100	2012-08-02 05:26:40.397758
22534	5	1-6180	22531	4	36	1650	87	2012-08-02 05:26:40.397758
22535	5	1-5702	22531	36	58	1728	42	2012-08-02 05:26:40.397758
22531	5	1-6576	22531	58	27	1869	1	2012-08-02 05:26:40.397758
22537	5	1-3689	22536	57	67	312	58	2012-08-02 05:26:40.397758
22538	5	1-6476	22536	67	70	456	91	2012-08-02 05:26:40.397758
22539	5	1-1623	22536	70	36	340	87	2012-08-02 05:26:40.397758
22540	5	1-5702	22536	36	58	359	42	2012-08-02 05:26:40.397758
22536	5	1-6576	22536	58	57	392	1	2012-08-02 05:26:40.397758
22541	1	1-1623	\N	70	70	1702	87	2012-08-02 05:26:40.397758
22543	3	1-6567	22542	78	8	162	47	2012-08-02 05:26:40.819686
22544	3	1-5586	22542	8	29	270	92	2012-08-02 05:26:40.819686
22542	3	1-6577	22542	29	78	273	15	2012-08-02 05:26:40.819686
22546	3	1-4319	22545	28	38	150	39	2012-08-02 05:26:43.681353
22547	3	1-4354	22545	38	94	164	46	2012-08-02 05:26:43.681353
22545	3	1-6584	22545	94	28	310	7	2012-08-02 05:26:43.681353
22549	4	1-6568	22548	40	23	5683	61	2012-08-02 05:26:44.079019
22550	4	1-6205	22548	23	36	3814	55	2012-08-02 05:26:44.079019
22551	4	1-3670	22548	36	21	1811	11	2012-08-02 05:26:44.079019
22548	4	1-6585	22548	21	40	3067	99	2012-08-02 05:26:44.079019
22553	3	1-6568	22552	40	68	2275	61	2012-08-02 05:26:44.079019
22554	3	1-4420	22552	68	21	713	11	2012-08-02 05:26:44.079019
22552	3	1-6585	22552	21	40	1218	99	2012-08-02 05:26:44.079019
22556	5	1-6427	22555	91	8	2530	23	2012-08-02 05:26:44.079019
22557	5	1-6445	22555	8	35	5243	7	2012-08-02 05:26:44.079019
22558	5	1-5091	22555	35	63	2251	97	2012-08-02 05:26:44.079019
22559	5	1-6047	22555	63	21	1576	11	2012-08-02 05:26:44.079019
22555	5	1-6585	22555	21	91	2757	99	2012-08-02 05:26:44.079019
22561	4	1-6427	22560	91	100	80	23	2012-08-02 05:26:44.079019
22562	4	1-5898	22560	100	47	65	28	2012-08-02 05:26:44.079019
22563	4	1-6087	22560	47	21	38	11	2012-08-02 05:26:44.079019
22560	4	1-6585	22560	21	91	76	99	2012-08-02 05:26:44.079019
22565	5	1-6527	22564	4	72	226	78	2012-08-02 05:26:44.509925
22566	5	1-5722	22564	72	13	123	76	2012-08-02 05:26:44.509925
22567	5	1-6441	22564	13	39	114	38	2012-08-02 05:26:44.509925
22568	5	1-6579	22564	39	54	110	87	2012-08-02 05:26:44.509925
22564	5	1-6586	22564	54	4	133	21	2012-08-02 05:26:44.509925
22570	4	1-6432	22569	47	11	874	91	2012-08-02 05:26:44.509925
22571	4	1-5834	22569	11	76	1023	51	2012-08-02 05:26:44.509925
22572	4	1-3809	22569	76	54	330	87	2012-08-02 05:26:44.509925
22569	4	1-6586	22569	54	47	526	21	2012-08-02 05:26:44.509925
22574	5	1-5879	22573	3	31	94	3	2012-08-02 05:26:44.509925
22575	5	1-5565	22573	31	25	81	37	2012-08-02 05:26:44.509925
22576	5	1-6539	22573	25	39	48	38	2012-08-02 05:26:44.509925
22577	5	1-6579	22573	39	54	67	87	2012-08-02 05:26:44.509925
22573	5	1-6586	22573	54	3	114	21	2012-08-02 05:26:44.509925
22579	5	1-6548	22578	84	20	1249	6	2012-08-02 05:26:45.747196
22580	5	1-3393	22578	20	99	502	86	2012-08-02 05:26:45.747196
22581	5	1-6546	22578	99	47	753	96	2012-08-02 05:26:45.747196
22582	5	1-5353	22578	47	52	538	78	2012-08-02 05:26:45.747196
22578	5	1-6589	22578	52	84	920	37	2012-08-02 05:26:45.747196
22584	5	1-5650	22583	91	1	1610	17	2012-08-02 05:26:46.553776
22585	5	1-6295	22583	1	72	1144	94	2012-08-02 05:26:46.553776
22586	5	1-6487	22583	72	3	860	43	2012-08-02 05:26:46.553776
22587	5	1-6347	22583	3	18	660	98	2012-08-02 05:26:46.553776
22583	5	1-6591	22583	18	91	2958	69	2012-08-02 05:26:46.553776
22589	5	1-6249	22588	25	65	2905	77	2012-08-02 05:26:46.553776
22590	5	1-5637	22588	65	17	1513	3	2012-08-02 05:26:46.553776
22591	5	1-1226	22588	17	3	861	43	2012-08-02 05:26:46.553776
22592	5	1-6347	22588	3	18	682	98	2012-08-02 05:26:46.553776
22588	5	1-6591	22588	18	25	3156	69	2012-08-02 05:26:46.553776
22594	5	1-6249	22593	25	80	2338	77	2012-08-02 05:26:46.553776
22595	5	1-2744	22593	80	80	497	8	2012-08-02 05:26:46.553776
22596	5	1-6360	22593	80	79	573	90	2012-08-02 05:26:46.553776
22597	5	1-6549	22593	79	18	496	98	2012-08-02 05:26:46.553776
22593	5	1-6591	22593	18	25	2416	69	2012-08-02 05:26:46.553776
22599	5	1-6495	22598	88	13	3104	16	2012-08-02 05:26:46.974043
22600	5	1-6023	22598	13	88	772	38	2012-08-02 05:26:46.974043
22601	5	1-5260	22598	88	13	203	23	2012-08-02 05:26:46.974043
22602	5	1-6402	22598	13	5	104	19	2012-08-02 05:26:46.974043
22598	5	1-6592	22598	5	88	5116	1	2012-08-02 05:26:46.974043
22604	3	1-6410	22603	87	33	3458	64	2012-08-02 05:26:47.393721
22605	3	1-6196	22603	33	92	1382	81	2012-08-02 05:26:47.393721
22603	3	1-6593	22603	92	87	6158	90	2012-08-02 05:26:47.393721
22607	4	1-6261	22606	75	93	1582	76	2012-08-02 05:26:47.393721
22608	4	1-5905	22606	93	45	843	97	2012-08-02 05:26:47.393721
22609	4	1-5638	22606	45	92	659	81	2012-08-02 05:26:47.393721
22606	4	1-6593	22606	92	75	3095	90	2012-08-02 05:26:47.393721
22611	3	1-6432	22610	47	11	2299	91	2012-08-02 05:26:48.211305
22612	3	1-5834	22610	11	40	2310	51	2012-08-02 05:26:48.211305
22610	3	1-6595	22610	40	47	1611	21	2012-08-02 05:26:48.211305
22614	4	1-6485	22613	74	93	1504	56	2012-08-02 05:26:48.211305
22615	4	1-6431	22613	93	3	3021	74	2012-08-02 05:26:48.211305
22616	4	1-5693	22613	3	40	2282	51	2012-08-02 05:26:48.211305
22613	4	1-6595	22613	40	74	1844	21	2012-08-02 05:26:48.211305
22618	3	1-6485	22617	74	41	51	56	2012-08-02 05:26:48.211305
22619	3	1-5212	22617	41	40	70	51	2012-08-02 05:26:48.211305
22617	3	1-6595	22617	40	74	60	21	2012-08-02 05:26:48.211305
22621	4	1-6216	22620	19	88	1325	1	2012-08-02 05:26:48.211305
22622	4	1-6495	22620	88	44	1215	16	2012-08-02 05:26:48.211305
22623	4	1-5900	22620	44	40	2207	51	2012-08-02 05:26:48.211305
22620	4	1-6595	22620	40	19	1962	21	2012-08-02 05:26:48.211305
22625	5	1-6294	22624	17	91	924	99	2012-08-02 05:26:49.073561
22626	5	1-6427	22624	91	13	1363	23	2012-08-02 05:26:49.073561
22627	5	1-6402	22624	13	95	1072	19	2012-08-02 05:26:49.073561
22628	5	1-5626	22624	95	23	2408	51	2012-08-02 05:26:49.073561
22624	5	1-6597	22624	23	17	2749	20	2012-08-02 05:26:49.073561
22630	2	1-5188	22629	47	56	1566	92	2012-08-02 05:26:49.482262
22629	2	1-6598	22629	56	47	5044	35	2012-08-02 05:26:49.482262
22632	4	1-6355	22631	63	46	372	55	2012-08-02 05:26:50.288688
22633	4	1-6278	22631	46	52	361	78	2012-08-02 05:26:50.288688
22634	4	1-6589	22631	52	21	583	37	2012-08-02 05:26:50.288688
22631	4	1-6600	22631	21	63	1606	6	2012-08-02 05:26:50.288688
22636	5	1-6512	22635	11	13	876	76	2012-08-02 05:26:51.095279
22637	5	1-6441	22635	13	46	1091	38	2012-08-02 05:26:51.095279
22638	5	1-5249	22635	46	10	618	71	2012-08-02 05:26:51.095279
22639	5	1-6550	22635	10	100	695	9	2012-08-02 05:26:51.095279
22635	5	1-6602	22635	100	11	1198	89	2012-08-02 05:26:51.095279
22641	4	1-6571	22640	82	38	2590	83	2012-08-02 05:26:51.095279
22642	4	1-2325	22640	38	89	927	80	2012-08-02 05:26:51.095279
22643	4	1-6403	22640	89	100	242	9	2012-08-02 05:26:51.095279
22640	4	1-6602	22640	100	82	429	89	2012-08-02 05:26:51.095279
22645	5	1-6471	22644	78	44	1614	72	2012-08-02 05:26:51.504147
22646	5	1-6305	22644	44	62	1852	28	2012-08-02 05:26:51.504147
22647	5	1-5652	22644	62	49	2398	37	2012-08-02 05:26:51.504147
22648	5	1-6379	22644	49	55	4990	44	2012-08-02 05:26:51.504147
22644	5	1-6603	22644	55	78	2731	97	2012-08-02 05:26:51.504147
22650	3	1-6536	22649	52	22	4388	96	2012-08-02 05:26:51.913099
22651	3	1-4673	22649	22	4	2381	49	2012-08-02 05:26:51.913099
22649	3	1-6604	22649	4	52	9666	24	2012-08-02 05:26:51.913099
22653	5	1-6548	22652	84	71	2207	6	2012-08-02 05:26:52.321906
22654	5	1-6529	22652	71	4	1516	12	2012-08-02 05:26:52.321906
22655	5	1-6109	22652	4	91	478	99	2012-08-02 05:26:52.321906
22656	5	1-6427	22652	91	58	678	23	2012-08-02 05:26:52.321906
22652	5	1-6605	22652	58	84	1560	37	2012-08-02 05:26:52.321906
22658	5	1-5079	22657	18	96	1169	84	2012-08-02 05:26:52.730726
22659	5	1-6544	22657	96	94	592	61	2012-08-02 05:26:52.730726
22660	5	1-5846	22657	94	67	264	75	2012-08-02 05:26:52.730726
22661	5	1-4622	22657	67	78	253	65	2012-08-02 05:26:52.730726
22657	5	1-6606	22657	78	18	2956	57	2012-08-02 05:26:52.730726
22663	4	1-5079	22662	18	96	961	84	2012-08-02 05:26:52.730726
22664	4	1-6544	22662	96	88	573	61	2012-08-02 05:26:52.730726
22665	4	1-194	22662	88	78	150	65	2012-08-02 05:26:52.730726
22662	4	1-6606	22662	78	18	2063	57	2012-08-02 05:26:52.730726
22667	4	1-6282	22666	32	32	5894	53	2012-08-02 05:26:53.150554
22666	4	1-6607	22666	70	32	1769	93	2012-08-02 05:26:53.150554
22668	4	1-3598	22666	32	42	3516	14	2012-08-02 05:26:53.150554
22669	4	1-6551	22666	42	70	3848	31	2012-08-02 05:26:53.150554
22671	2	1-2730	22670	45	40	2632	17	2012-08-02 05:26:53.5484
22670	2	1-6608	22670	40	45	4243	9	2012-08-02 05:26:53.5484
22673	4	1-3035	22672	65	23	3070	51	2012-08-02 05:26:53.5484
22674	4	1-6597	22672	23	30	2811	20	2012-08-02 05:26:53.5484
22675	4	1-6580	22672	30	40	2212	17	2012-08-02 05:26:53.5484
22672	4	1-6608	22672	40	65	3849	9	2012-08-02 05:26:53.5484
22677	3	1-4280	22676	31	30	24	20	2012-08-02 05:26:53.5484
22678	3	1-6580	22676	30	40	23	17	2012-08-02 05:26:53.5484
22676	3	1-6608	22676	40	31	48	9	2012-08-02 05:26:53.5484
22680	4	1-4648	22679	61	77	614	20	2012-08-02 05:26:53.946204
22681	4	1-6273	22679	77	74	234	82	2012-08-02 05:26:53.946204
22682	4	1-6417	22679	74	30	370	12	2012-08-02 05:26:53.946204
22679	4	1-6609	22679	30	61	1947	59	2012-08-02 05:26:53.946204
22684	5	1-6441	22683	13	39	199	38	2012-08-02 05:26:55.128483
22685	5	1-6579	22683	39	36	242	87	2012-08-02 05:26:55.128483
22686	5	1-5702	22683	36	65	218	42	2012-08-02 05:26:55.128483
22687	5	1-6418	22683	65	59	224	36	2012-08-02 05:26:55.128483
22683	5	1-6612	22683	59	13	172	76	2012-08-02 05:26:55.128483
22689	2	1-6550	22688	10	55	6517	9	2012-08-02 05:26:55.968025
22688	2	1-6614	22688	55	10	8560	71	2012-08-02 05:26:55.968025
22691	5	1-6427	22690	91	13	528	23	2012-08-02 05:26:57.18316
22692	5	1-6402	22690	13	90	387	19	2012-08-02 05:26:57.18316
22693	5	1-4187	22690	90	25	300	26	2012-08-02 05:26:57.18316
22694	5	1-6340	22690	25	84	252	86	2012-08-02 05:26:57.18316
22690	5	1-6617	22690	84	91	384	99	2012-08-02 05:26:57.18316
22696	5	1-5796	22695	76	75	2142	49	2012-08-02 05:26:57.18316
22697	5	1-6211	22695	75	41	1116	73	2012-08-02 05:26:57.18316
22698	5	1-6399	22695	41	55	1098	93	2012-08-02 05:26:57.18316
22699	5	1-5919	22695	55	84	762	86	2012-08-02 05:26:57.18316
22695	5	1-6617	22695	84	76	1189	99	2012-08-02 05:26:57.18316
22701	4	1-2628	22700	88	70	2734	31	2012-08-02 05:26:57.18316
22702	4	1-6607	22700	70	55	2231	93	2012-08-02 05:26:57.18316
22703	4	1-5919	22700	55	84	1583	86	2012-08-02 05:26:57.18316
22700	4	1-6617	22700	84	88	2521	99	2012-08-02 05:26:57.18316
22705	4	1-4602	22704	76	57	703	92	2012-08-02 05:26:58.023143
22706	4	1-6479	22704	57	85	202	1	2012-08-02 05:26:58.023143
22707	4	1-6439	22704	85	15	522	45	2012-08-02 05:26:58.023143
22704	4	1-6619	22704	15	76	1138	54	2012-08-02 05:26:58.023143
22709	5	1-3221	22708	8	94	813	46	2012-08-02 05:26:58.851617
22710	5	1-6584	22708	94	35	1540	7	2012-08-02 05:26:58.851617
22711	5	1-5091	22708	35	45	1006	97	2012-08-02 05:26:58.851617
22712	5	1-5638	22708	45	13	1084	81	2012-08-02 05:26:58.851617
22708	5	1-6621	22708	13	8	1075	65	2012-08-02 05:26:58.851617
22714	3	1-6480	22713	23	89	3362	68	2012-08-02 05:26:59.271525
22715	3	1-6104	22713	89	65	1890	32	2012-08-02 05:26:59.271525
22713	3	1-6622	22713	65	23	5592	88	2012-08-02 05:26:59.271525
22717	3	1-6543	22716	42	95	401	31	2012-08-02 05:26:59.691364
22718	3	1-2623	22716	95	72	99	73	2012-08-02 05:26:59.691364
22716	3	1-6623	22716	72	42	203	5	2012-08-02 05:26:59.691364
22720	5	1-6564	22719	10	25	3161	87	2012-08-02 05:27:01.315608
22721	5	1-6018	22719	25	72	450	8	2012-08-02 05:27:01.315608
22722	5	1-5798	22719	72	5	786	34	2012-08-02 05:27:01.315608
22723	5	1-6356	22719	5	98	662	65	2012-08-02 05:27:01.315608
22719	5	1-6627	22719	98	10	4141	11	2012-08-02 05:27:01.315608
22725	2	1-6387	22724	65	74	3327	5	2012-08-02 05:27:02.530692
22724	2	1-6630	22724	74	65	2612	58	2012-08-02 05:27:02.530692
22727	3	1-6610	22726	67	49	2907	9	2012-08-02 05:27:03.757139
22728	3	1-6415	22726	49	9	1429	75	2012-08-02 05:27:03.757139
22726	3	1-6633	22726	9	67	3679	91	2012-08-02 05:27:03.757139
22730	3	1-1254	22729	35	7	2085	15	2012-08-02 05:27:03.757139
22731	3	1-6554	22729	7	9	1141	75	2012-08-02 05:27:03.757139
22729	3	1-6633	22729	9	35	3157	91	2012-08-02 05:27:03.757139
22733	5	1-6610	22732	67	17	253	9	2012-08-02 05:27:03.757139
22734	5	1-5925	22732	17	58	188	56	2012-08-02 05:27:03.757139
22735	5	1-2523	22732	58	8	117	43	2012-08-02 05:27:03.757139
22736	5	1-5928	22732	8	9	88	75	2012-08-02 05:27:03.757139
22732	5	1-6633	22732	9	67	269	91	2012-08-02 05:27:03.757139
22738	4	1-5266	22737	84	52	1355	42	2012-08-02 05:27:03.757139
22739	4	1-5293	22737	52	7	1172	15	2012-08-02 05:27:03.757139
22740	4	1-6554	22737	7	9	713	75	2012-08-02 05:27:03.757139
22737	4	1-6633	22737	9	84	2195	91	2012-08-02 05:27:03.757139
22742	4	1-1972	22741	31	49	334	97	2012-08-02 05:27:03.757139
22743	4	1-4865	22741	49	7	298	15	2012-08-02 05:27:03.757139
22744	4	1-6554	22741	7	9	183	75	2012-08-02 05:27:03.757139
22741	4	1-6633	22741	9	31	571	91	2012-08-02 05:27:03.757139
22746	5	1-3861	22745	6	28	309	32	2012-08-02 05:27:04.585958
22747	5	1-6332	22745	28	82	1505	33	2012-08-02 05:27:04.585958
22748	5	1-5644	22745	82	12	780	8	2012-08-02 05:27:04.585958
22749	5	1-5933	22745	12	70	748	91	2012-08-02 05:27:04.585958
22745	5	1-6635	22745	70	6	775	41	2012-08-02 05:27:04.585958
22751	5	1-4600	22750	63	7	791	15	2012-08-02 05:27:05.417057
22752	5	1-6554	22750	7	30	354	75	2012-08-02 05:27:05.417057
22753	5	1-4581	22750	30	79	198	97	2012-08-02 05:27:05.417057
22754	5	1-5845	22750	79	16	96	30	2012-08-02 05:27:05.417057
22750	5	1-6637	22750	16	63	1287	63	2012-08-02 05:27:05.417057
22756	4	1-4600	22755	63	59	759	15	2012-08-02 05:27:05.417057
22757	4	1-5956	22755	59	42	146	78	2012-08-02 05:27:05.417057
22758	4	1-6472	22755	42	16	89	30	2012-08-02 05:27:05.417057
22755	4	1-6637	22755	16	63	1208	63	2012-08-02 05:27:05.417057
22760	5	1-5840	22759	66	53	796	85	2012-08-02 05:27:06.987115
22761	5	1-5771	22759	53	76	411	54	2012-08-02 05:27:06.987115
22762	5	1-4602	22759	76	29	215	92	2012-08-02 05:27:06.987115
22763	5	1-6577	22759	29	100	239	15	2012-08-02 05:27:06.987115
22759	5	1-6641	22759	100	66	286	75	2012-08-02 05:27:06.987115
22765	5	1-5840	22764	66	48	2163	85	2012-08-02 05:27:06.987115
22766	5	1-6515	22764	48	63	482	36	2012-08-02 05:27:06.987115
22767	5	1-5749	22764	63	29	583	92	2012-08-02 05:27:06.987115
22768	5	1-6577	22764	29	100	649	15	2012-08-02 05:27:06.987115
22764	5	1-6641	22764	100	66	778	75	2012-08-02 05:27:06.987115
22770	5	1-4581	22769	30	12	4018	97	2012-08-02 05:27:06.987115
22771	5	1-6243	22769	12	63	2781	36	2012-08-02 05:27:06.987115
22772	5	1-5749	22769	63	29	3612	92	2012-08-02 05:27:06.987115
22773	5	1-6577	22769	29	100	4317	15	2012-08-02 05:27:06.987115
22769	5	1-6641	22769	100	30	5553	75	2012-08-02 05:27:06.987115
22775	5	1-4095	22774	5	80	1554	25	2012-08-02 05:27:06.987115
22776	5	1-5679	22774	80	58	995	76	2012-08-02 05:27:06.987115
22777	5	1-6121	22774	58	29	867	92	2012-08-02 05:27:06.987115
22778	5	1-6577	22774	29	100	1069	15	2012-08-02 05:27:06.987115
22774	5	1-6641	22774	100	5	1419	75	2012-08-02 05:27:06.987115
22780	3	1-775	22779	76	40	575	53	2012-08-02 05:27:07.396153
22781	3	1-5760	22779	40	54	219	88	2012-08-02 05:27:07.396153
22779	3	1-6642	22779	54	76	1030	44	2012-08-02 05:27:07.396153
22783	4	1-5079	22782	18	96	1519	84	2012-08-02 05:27:08.180332
22784	4	1-6544	22782	96	68	1190	61	2012-08-02 05:27:08.180332
22785	4	1-4420	22782	68	21	609	11	2012-08-02 05:27:08.180332
22782	4	1-6644	22782	21	18	2483	57	2012-08-02 05:27:08.180332
22787	4	1-6241	22786	27	61	2170	100	2012-08-02 05:27:09.019983
22788	4	1-6393	22786	61	57	1999	90	2012-08-02 05:27:09.019983
22789	4	1-5217	22786	57	62	982	77	2012-08-02 05:27:09.019983
22786	4	1-6646	22786	62	27	1965	61	2012-08-02 05:27:09.019983
22791	4	1-6205	22790	23	87	1649	55	2012-08-02 05:27:09.019983
22792	4	1-6361	22790	87	74	1659	26	2012-08-02 05:27:09.019983
22793	4	1-1492	22790	74	62	745	77	2012-08-02 05:27:09.019983
22790	4	1-6646	22790	62	23	1534	61	2012-08-02 05:27:09.019983
22795	3	1-6581	22794	66	53	2682	13	2012-08-02 05:27:09.428845
22796	3	1-6371	22794	53	47	2176	2	2012-08-02 05:27:09.428845
22794	3	1-6647	22794	47	66	2675	18	2012-08-02 05:27:09.428845
22798	4	1-5756	22797	67	39	685	38	2012-08-02 05:27:09.428845
22799	4	1-6579	22797	39	6	912	87	2012-08-02 05:27:09.428845
22800	4	1-2663	22797	6	47	506	2	2012-08-02 05:27:09.428845
22797	4	1-6647	22797	47	67	640	18	2012-08-02 05:27:09.428845
22802	4	1-2241	22801	72	30	51	16	2012-08-02 05:27:09.428845
22803	4	1-4954	22801	30	29	25	52	2012-08-02 05:27:09.428845
22804	4	1-6167	22801	29	47	40	2	2012-08-02 05:27:09.428845
22801	4	1-6647	22801	47	72	52	18	2012-08-02 05:27:09.428845
22805	1	1-3143	\N	57	57	2087	2	2012-08-02 05:27:09.428845
22807	5	1-6536	22806	52	17	169	96	2012-08-02 05:27:10.600498
22808	5	1-6358	22806	17	6	139	41	2012-08-02 05:27:10.600498
22809	5	1-3861	22806	6	28	46	32	2012-08-02 05:27:10.600498
22810	5	1-6332	22806	28	73	186	33	2012-08-02 05:27:10.600498
22806	5	1-6650	22806	73	52	282	24	2012-08-02 05:27:10.600498
22812	5	1-6573	22811	88	23	1075	84	2012-08-02 05:27:10.600498
22813	5	1-5437	22811	23	91	568	70	2012-08-02 05:27:10.600498
22814	5	1-3216	22811	91	25	457	14	2012-08-02 05:27:10.600498
22815	5	1-6461	22811	25	73	332	33	2012-08-02 05:27:10.600498
22811	5	1-6650	22811	73	88	504	24	2012-08-02 05:27:10.600498
22817	4	1-1295	22816	75	10	1757	13	2012-08-02 05:27:10.600498
22818	4	1-6464	22816	10	28	636	32	2012-08-02 05:27:10.600498
22819	4	1-6332	22816	28	73	2665	33	2012-08-02 05:27:10.600498
22816	4	1-6650	22816	73	75	4184	24	2012-08-02 05:27:10.600498
22821	4	1-6574	22820	85	73	2019	56	2012-08-02 05:27:10.600498
22822	4	1-3811	22820	73	25	3268	14	2012-08-02 05:27:10.600498
22823	4	1-6461	22820	25	73	2117	33	2012-08-02 05:27:10.600498
22820	4	1-6650	22820	73	85	3414	24	2012-08-02 05:27:10.600498
22825	5	1-6574	22824	85	23	588	56	2012-08-02 05:27:10.600498
22826	5	1-3680	22824	23	89	196	68	2012-08-02 05:27:10.600498
22827	5	1-6104	22824	89	28	107	32	2012-08-02 05:27:10.600498
22828	5	1-6332	22824	28	73	478	33	2012-08-02 05:27:10.600498
22824	5	1-6650	22824	73	85	801	24	2012-08-02 05:27:10.600498
22829	1	1-3680	\N	23	23	147	68	2012-08-02 05:27:10.600498
22831	4	1-6439	22830	85	93	3870	45	2012-08-02 05:27:11.009527
22832	4	1-5740	22830	93	89	4073	80	2012-08-02 05:27:11.009527
22833	4	1-6403	22830	89	1	1005	9	2012-08-02 05:27:11.009527
22830	4	1-6651	22830	1	85	1629	1	2012-08-02 05:27:11.009527
22835	4	1-6495	22834	88	30	3739	16	2012-08-02 05:27:11.009527
22836	4	1-4954	22834	30	28	1764	52	2012-08-02 05:27:11.009527
22837	4	1-6453	22834	28	1	2577	9	2012-08-02 05:27:11.009527
22834	4	1-6651	22834	1	88	4354	1	2012-08-02 05:27:11.009527
22839	5	1-5819	22838	8	71	1784	31	2012-08-02 05:27:11.804678
22840	5	1-6144	22838	71	29	906	51	2012-08-02 05:27:11.804678
22841	5	1-3925	22838	29	80	112	8	2012-08-02 05:27:11.804678
22842	5	1-6360	22838	80	40	130	90	2012-08-02 05:27:11.804678
22838	5	1-6653	22838	40	8	984	48	2012-08-02 05:27:11.804678
22844	4	1-2737	22843	55	84	261	86	2012-08-02 05:27:11.804678
22845	4	1-6617	22843	84	72	411	99	2012-08-02 05:27:11.804678
22846	4	1-5863	22843	72	40	252	90	2012-08-02 05:27:11.804678
22843	4	1-6653	22843	40	55	1997	48	2012-08-02 05:27:11.804678
22848	5	1-5706	22847	89	71	1114	72	2012-08-02 05:27:12.213706
22849	5	1-6346	22847	71	8	551	47	2012-08-02 05:27:12.213706
22850	5	1-5586	22847	8	29	676	92	2012-08-02 05:27:12.213706
22851	5	1-6577	22847	29	87	504	15	2012-08-02 05:27:12.213706
22847	5	1-6654	22847	87	89	3657	65	2012-08-02 05:27:12.213706
22853	4	1-5706	22852	89	44	1054	72	2012-08-02 05:27:12.213706
22854	4	1-6305	22852	44	47	886	28	2012-08-02 05:27:12.213706
22855	4	1-5770	22852	47	87	374	15	2012-08-02 05:27:12.213706
22852	4	1-6654	22852	87	89	3067	65	2012-08-02 05:27:12.213706
22857	4	1-6609	22856	30	53	2228	59	2012-08-02 05:27:12.60058
22858	4	1-4705	22856	53	46	1404	58	2012-08-02 05:27:12.60058
22859	4	1-4452	22856	46	48	441	44	2012-08-02 05:27:12.60058
22856	4	1-6655	22856	48	30	625	12	2012-08-02 05:27:12.60058
22861	5	1-5914	22860	37	8	1697	27	2012-08-02 05:27:12.60058
22862	5	1-6141	22860	8	8	1032	93	2012-08-02 05:27:12.60058
22863	5	1-3282	22860	8	54	265	88	2012-08-02 05:27:12.60058
22864	5	1-6642	22860	54	48	882	44	2012-08-02 05:27:12.60058
22860	5	1-6655	22860	48	37	1292	12	2012-08-02 05:27:12.60058
22866	4	1-6302	22865	49	18	4437	92	2012-08-02 05:27:12.60058
22867	4	1-1778	22865	18	54	1164	88	2012-08-02 05:27:12.60058
22868	4	1-6642	22865	54	48	4029	44	2012-08-02 05:27:12.60058
22865	4	1-6655	22865	48	49	6133	12	2012-08-02 05:27:12.60058
22870	5	1-6499	22869	73	90	15	29	2012-08-02 05:27:12.60058
22871	5	1-6331	22869	90	40	20	53	2012-08-02 05:27:12.60058
22872	5	1-5760	22869	40	54	6	88	2012-08-02 05:27:12.60058
22873	5	1-6642	22869	54	48	20	44	2012-08-02 05:27:12.60058
22869	5	1-6655	22869	48	73	31	12	2012-08-02 05:27:12.60058
22875	4	1-5276	22874	51	54	761	88	2012-08-02 05:27:13.782862
22876	4	1-6642	22874	54	76	2767	44	2012-08-02 05:27:13.782862
22877	4	1-5019	22874	76	24	993	51	2012-08-02 05:27:13.782862
22874	4	1-6658	22874	24	51	1753	41	2012-08-02 05:27:13.782862
22879	3	1-6297	22878	29	45	1854	36	2012-08-02 05:27:13.782862
22880	3	1-4057	22878	45	24	1274	51	2012-08-02 05:27:13.782862
22878	3	1-6658	22878	24	29	2274	41	2012-08-02 05:27:13.782862
22882	3	1-6169	22881	7	5	4541	100	2012-08-02 05:27:14.578475
22883	3	1-6523	22881	5	83	3983	41	2012-08-02 05:27:14.578475
22881	3	1-6660	22881	83	7	7423	71	2012-08-02 05:27:14.578475
22885	4	1-6550	22884	10	8	688	9	2012-08-02 05:27:14.578475
22886	4	1-6380	22884	8	5	375	100	2012-08-02 05:27:14.578475
22887	4	1-6523	22884	5	83	366	41	2012-08-02 05:27:14.578475
22884	4	1-6660	22884	83	10	759	71	2012-08-02 05:27:14.578475
22889	5	1-3616	22888	44	85	305	1	2012-08-02 05:27:16.975876
22890	5	1-6439	22888	85	51	655	45	2012-08-02 05:27:16.975876
22891	5	1-6013	22888	51	49	373	37	2012-08-02 05:27:16.975876
22892	5	1-6379	22888	49	1	741	44	2012-08-02 05:27:16.975876
22888	5	1-6666	22888	1	44	831	36	2012-08-02 05:27:16.975876
22893	1	1-3616	\N	44	44	469	1	2012-08-02 05:27:16.975876
22895	4	1-6248	22894	63	60	1116	35	2012-08-02 05:27:17.78253
22896	4	1-6395	22894	60	85	1374	1	2012-08-02 05:27:17.78253
22897	4	1-6439	22894	85	3	2298	45	2012-08-02 05:27:17.78253
22894	4	1-6668	22894	3	63	1697	19	2012-08-02 05:27:17.78253
22899	5	1-6046	22898	38	10	817	8	2012-08-02 05:27:17.78253
22900	5	1-5963	22898	10	1	532	9	2012-08-02 05:27:17.78253
22901	5	1-6651	22898	1	85	704	1	2012-08-02 05:27:17.78253
22902	5	1-6439	22898	85	3	1365	45	2012-08-02 05:27:17.78253
22898	5	1-6668	22898	3	38	1168	19	2012-08-02 05:27:17.78253
22904	5	1-4187	22903	90	74	2289	26	2012-08-02 05:27:17.78253
22905	5	1-1492	22903	74	62	941	77	2012-08-02 05:27:17.78253
22906	5	1-6646	22903	62	87	1772	61	2012-08-02 05:27:17.78253
22907	5	1-6101	22903	87	3	2982	45	2012-08-02 05:27:17.78253
22903	5	1-6668	22903	3	90	3114	19	2012-08-02 05:27:17.78253
22909	5	1-5909	22908	89	8	1072	43	2012-08-02 05:27:17.78253
22910	5	1-5928	22908	8	56	796	75	2012-08-02 05:27:17.78253
22911	5	1-2507	22908	56	87	612	61	2012-08-02 05:27:17.78253
22912	5	1-6101	22908	87	3	1065	45	2012-08-02 05:27:17.78253
22908	5	1-6668	22908	3	89	1150	19	2012-08-02 05:27:17.78253
22914	4	1-5008	22913	40	82	1105	56	2012-08-02 05:27:17.78253
22915	4	1-6005	22913	82	19	1240	96	2012-08-02 05:27:17.78253
22916	4	1-6518	22913	19	3	886	45	2012-08-02 05:27:17.78253
22913	4	1-6668	22913	3	40	1033	19	2012-08-02 05:27:17.78253
22918	3	1-5928	22917	8	4	825	75	2012-08-02 05:27:19.407107
22919	3	1-4872	22917	4	76	340	83	2012-08-02 05:27:19.407107
22917	3	1-6672	22917	76	8	3237	43	2012-08-02 05:27:19.407107
22921	4	1-5981	22920	48	36	635	47	2012-08-02 05:27:21.428438
22922	4	1-6505	22920	36	59	948	36	2012-08-02 05:27:21.428438
22923	4	1-6612	22920	59	36	765	76	2012-08-02 05:27:21.428438
22920	4	1-6677	22920	36	48	1514	2	2012-08-02 05:27:21.428438
22925	5	1-5956	22924	59	72	1446	78	2012-08-02 05:27:22.97549
22926	5	1-6164	22924	72	28	1008	52	2012-08-02 05:27:22.97549
22927	5	1-6453	22924	28	1	1472	9	2012-08-02 05:27:22.97549
22928	5	1-6651	22924	1	50	2485	1	2012-08-02 05:27:22.97549
22924	5	1-6681	22924	50	59	5629	15	2012-08-02 05:27:22.97549
22930	4	1-6487	22929	72	97	5967	43	2012-08-02 05:27:23.78199
22931	4	1-6521	22929	97	56	6558	67	2012-08-02 05:27:23.78199
22932	4	1-4500	22929	56	15	3440	60	2012-08-02 05:27:23.78199
22929	4	1-6683	22929	15	72	8260	94	2012-08-02 05:27:23.78199
22934	4	1-6241	22933	27	5	2429	100	2012-08-02 05:27:24.577424
22935	4	1-6523	22933	5	29	1724	41	2012-08-02 05:27:24.577424
22936	4	1-6297	22933	29	68	1072	36	2012-08-02 05:27:24.577424
22933	4	1-6685	22933	68	27	3846	61	2012-08-02 05:27:24.577424
22938	4	1-6309	22937	48	12	122	64	2012-08-02 05:27:24.577424
22939	4	1-6555	22937	12	36	55	47	2012-08-02 05:27:24.577424
22940	4	1-6505	22937	36	68	48	36	2012-08-02 05:27:24.577424
22937	4	1-6685	22937	68	48	179	61	2012-08-02 05:27:24.577424
22942	3	1-5212	22941	41	27	2154	51	2012-08-02 05:27:25.384016
22943	3	1-6227	22941	27	14	2496	29	2012-08-02 05:27:25.384016
22941	3	1-6687	22941	14	41	1638	56	2012-08-02 05:27:25.384016
22945	5	1-6576	22944	58	31	5019	1	2012-08-02 05:27:27.383505
22946	5	1-6394	22944	31	88	3542	97	2012-08-02 05:27:27.383505
22947	5	1-5548	22944	88	51	1659	3	2012-08-02 05:27:27.383505
22948	5	1-6583	22944	51	64	3191	66	2012-08-02 05:27:27.383505
22944	5	1-6692	22944	64	58	4678	42	2012-08-02 05:27:27.383505
22950	4	1-6393	22949	61	79	301	90	2012-08-02 05:27:27.792434
22951	4	1-6549	22949	79	61	272	98	2012-08-02 05:27:27.792434
22952	4	1-2920	22949	61	19	409	31	2012-08-02 05:27:27.792434
22949	4	1-6693	22949	19	61	329	100	2012-08-02 05:27:27.792434
22954	4	1-6523	22953	5	40	1844	41	2012-08-02 05:27:27.792434
22955	4	1-3963	22953	40	42	1124	14	2012-08-02 05:27:27.792434
22956	4	1-6551	22953	42	19	1961	31	2012-08-02 05:27:27.792434
22953	4	1-6693	22953	19	5	1594	100	2012-08-02 05:27:27.792434
22958	3	1-6565	22957	98	82	310	79	2012-08-02 05:27:28.192046
22959	3	1-6174	22957	82	42	304	40	2012-08-02 05:27:28.192046
22957	3	1-6694	22957	42	98	193	64	2012-08-02 05:27:28.192046
22961	5	1-6555	22960	12	61	395	47	2012-08-02 05:27:28.192046
22962	5	1-5383	22960	61	9	555	27	2012-08-02 05:27:28.192046
22963	5	1-6386	22960	9	82	523	79	2012-08-02 05:27:28.192046
22964	5	1-6174	22960	82	42	713	40	2012-08-02 05:27:28.192046
22960	5	1-6694	22960	42	12	629	64	2012-08-02 05:27:28.192046
22966	5	1-6555	22965	12	4	2395	47	2012-08-02 05:27:28.192046
22967	5	1-5088	22965	4	24	1938	65	2012-08-02 05:27:28.192046
22968	5	1-6239	22965	24	82	2357	79	2012-08-02 05:27:28.192046
22969	5	1-6174	22965	82	42	3546	40	2012-08-02 05:27:28.192046
22965	5	1-6694	22965	42	12	3453	64	2012-08-02 05:27:28.192046
22970	1	1-3981	\N	81	81	1921	40	2012-08-02 05:27:28.192046
22972	5	1-6651	22971	1	27	313	1	2012-08-02 05:27:29.40726
22973	5	1-6407	22971	27	82	228	56	2012-08-02 05:27:29.40726
22974	5	1-6005	22971	82	42	206	96	2012-08-02 05:27:29.40726
22975	5	1-5698	22971	42	76	176	50	2012-08-02 05:27:29.40726
22971	5	1-6697	22971	76	1	216	9	2012-08-02 05:27:29.40726
22977	4	1-5925	22976	17	82	3520	56	2012-08-02 05:27:29.40726
22978	4	1-6005	22976	82	42	3604	96	2012-08-02 05:27:29.40726
22979	4	1-5698	22976	42	76	3481	50	2012-08-02 05:27:29.40726
22976	4	1-6697	22976	76	17	4833	9	2012-08-02 05:27:29.40726
22981	2	1-1599	22980	81	76	1466	50	2012-08-02 05:27:29.40726
22980	2	1-6697	22980	76	81	2128	9	2012-08-02 05:27:29.40726
22983	4	1-6340	22982	25	84	1575	86	2012-08-02 05:27:30.202857
22984	4	1-6617	22982	84	91	1976	99	2012-08-02 05:27:30.202857
22985	4	1-6427	22982	91	29	2237	23	2012-08-02 05:27:30.202857
22982	4	1-6699	22982	29	25	2278	26	2012-08-02 05:27:30.202857
22987	5	1-6661	22986	82	75	597	91	2012-08-02 05:27:30.738823
22988	5	1-6676	22986	75	10	1849	93	2012-08-02 05:27:30.738823
22989	5	1-4219	22986	10	85	569	20	2012-08-02 05:27:30.738823
22990	5	1-6337	22986	85	25	496	6	2012-08-02 05:27:30.738823
22986	5	1-6700	22986	25	82	1342	28	2012-08-02 05:27:30.738823
22992	3	1-6448	22991	50	23	1134	8	2012-08-02 05:27:30.738823
22993	3	1-6519	22991	23	25	975	6	2012-08-02 05:27:30.738823
22991	3	1-6700	22991	25	50	2672	28	2012-08-02 05:27:30.738823
22995	4	1-6638	22994	32	13	2150	81	2012-08-02 05:27:31.529051
22996	4	1-6621	22994	13	18	1657	65	2012-08-02 05:27:31.529051
22997	4	1-5668	22994	18	79	974	10	2012-08-02 05:27:31.529051
22994	4	1-6702	22994	79	32	3281	57	2012-08-02 05:27:31.529051
22999	3	1-837	22998	85	66	490	93	2012-08-02 05:27:31.529051
23000	3	1-6588	22998	66	79	493	10	2012-08-02 05:27:31.529051
22998	3	1-6702	22998	79	85	2251	57	2012-08-02 05:27:31.529051
23002	2	1-5555	23001	98	9	1057	53	2012-08-02 05:27:31.915527
23001	2	1-6703	23001	9	98	5217	1	2012-08-02 05:27:31.915527
23004	4	1-5898	23003	100	50	475	28	2012-08-02 05:27:32.291283
23005	4	1-6448	23003	50	72	144	8	2012-08-02 05:27:32.291283
23006	4	1-5798	23003	72	67	229	34	2012-08-02 05:27:32.291283
23003	4	1-6704	23003	67	100	635	23	2012-08-02 05:27:32.291283
23008	5	1-6605	23007	58	25	179	37	2012-08-02 05:27:32.291283
23009	5	1-6539	23007	25	53	87	38	2012-08-02 05:27:32.291283
23010	5	1-6696	23007	53	6	28	55	2012-08-02 05:27:32.291283
23011	5	1-5374	23007	6	67	29	34	2012-08-02 05:27:32.291283
23007	5	1-6704	23007	67	58	95	23	2012-08-02 05:27:32.291283
23013	5	1-6699	23012	29	25	3544	26	2012-08-02 05:27:32.291283
23014	5	1-6340	23012	25	5	2560	86	2012-08-02 05:27:32.291283
23015	5	1-3958	23012	5	6	960	55	2012-08-02 05:27:32.291283
23016	5	1-5374	23012	6	67	994	34	2012-08-02 05:27:32.291283
23012	5	1-6704	23012	67	29	3331	23	2012-08-02 05:27:32.291283
23018	4	1-6699	23017	29	19	27	26	2012-08-02 05:27:32.291283
23019	4	1-5175	23017	19	17	14	74	2012-08-02 05:27:32.291283
23020	4	1-5408	23017	17	67	7	34	2012-08-02 05:27:32.291283
23017	4	1-6704	23017	67	29	24	23	2012-08-02 05:27:32.291283
23022	4	1-6605	23021	58	61	1862	37	2012-08-02 05:27:32.291283
23023	4	1-6054	23021	61	59	625	4	2012-08-02 05:27:32.291283
23024	4	1-5325	23021	59	67	234	34	2012-08-02 05:27:32.291283
23021	4	1-6704	23021	67	58	872	23	2012-08-02 05:27:32.291283
23026	4	1-6098	23025	96	82	2895	54	2012-08-02 05:27:32.291283
23027	4	1-5272	23025	82	17	825	74	2012-08-02 05:27:32.291283
23028	4	1-5408	23025	17	67	442	34	2012-08-02 05:27:32.291283
23025	4	1-6704	23025	67	96	1836	23	2012-08-02 05:27:32.291283
23030	2	1-5944	23029	85	37	751	94	2012-08-02 05:27:32.689436
23029	2	1-6705	23029	37	85	1679	22	2012-08-02 05:27:32.689436
23032	4	1-1004	23031	29	13	2270	81	2012-08-02 05:27:33.462715
23033	4	1-6621	23031	13	24	2256	65	2012-08-02 05:27:33.462715
23034	4	1-6239	23031	24	48	3031	79	2012-08-02 05:27:33.462715
23031	4	1-6707	23031	48	29	5988	33	2012-08-02 05:27:33.462715
23036	4	1-5229	23035	75	98	2670	1	2012-08-02 05:27:34.26933
23037	4	1-5957	23035	98	6	2225	60	2012-08-02 05:27:34.26933
23038	4	1-5290	23035	6	56	1926	73	2012-08-02 05:27:34.26933
23035	4	1-6709	23035	56	75	4783	75	2012-08-02 05:27:34.26933
23040	4	1-2268	23039	96	46	696	28	2012-08-02 05:27:34.26933
23041	4	1-5178	23039	46	40	251	22	2012-08-02 05:27:34.26933
23042	4	1-6391	23039	40	56	336	73	2012-08-02 05:27:34.26933
23039	4	1-6709	23039	56	96	849	75	2012-08-02 05:27:34.26933
23044	5	1-4888	23043	28	42	2004	40	2012-08-02 05:27:35.032042
23045	5	1-6694	23043	42	79	2054	64	2012-08-02 05:27:35.032042
23046	5	1-3161	23043	79	50	776	1	2012-08-02 05:27:35.032042
23047	5	1-6681	23043	50	49	1720	15	2012-08-02 05:27:35.032042
23043	5	1-6711	23043	49	28	3008	5	2012-08-02 05:27:35.032042
23049	5	1-4438	23048	41	76	1350	50	2012-08-02 05:27:35.418346
23050	5	1-6697	23048	76	1	1531	9	2012-08-02 05:27:35.418346
23051	5	1-5611	23048	1	100	549	3	2012-08-02 05:27:35.418346
23052	5	1-6665	23048	100	42	1707	37	2012-08-02 05:27:35.418346
23048	5	1-6712	23048	42	41	1427	18	2012-08-02 05:27:35.418346
23054	5	1-4438	23053	41	91	2699	50	2012-08-02 05:27:35.418346
23055	5	1-6616	23053	91	1	1579	9	2012-08-02 05:27:35.418346
23056	5	1-5611	23053	1	100	668	3	2012-08-02 05:27:35.418346
23057	5	1-6665	23053	100	42	2451	37	2012-08-02 05:27:35.418346
23053	5	1-6712	23053	42	41	2417	18	2012-08-02 05:27:35.418346
23059	4	1-5057	23058	98	9	3280	30	2012-08-02 05:27:35.418346
23060	4	1-4799	23058	9	62	2154	28	2012-08-02 05:27:35.418346
23061	4	1-5652	23058	62	42	3118	37	2012-08-02 05:27:35.418346
23058	4	1-6712	23058	42	98	3338	18	2012-08-02 05:27:35.418346
23063	4	1-5756	23062	67	88	2223	38	2012-08-02 05:27:35.418346
23064	4	1-5260	23062	88	58	846	23	2012-08-02 05:27:35.418346
23065	4	1-6605	23062	58	42	1915	37	2012-08-02 05:27:35.418346
23062	4	1-6712	23062	42	67	2059	18	2012-08-02 05:27:35.418346
23067	4	1-5083	23066	52	93	625	59	2012-08-02 05:27:35.849737
23068	4	1-6061	23066	93	93	247	45	2012-08-02 05:27:35.849737
23069	4	1-5740	23066	93	16	241	80	2012-08-02 05:27:35.849737
23066	4	1-6713	23066	16	52	1282	74	2012-08-02 05:27:35.849737
23071	5	1-4791	23070	100	21	1527	49	2012-08-02 05:27:35.849737
23072	5	1-39	23070	21	50	463	52	2012-08-02 05:27:35.849737
23073	5	1-6558	23070	50	7	571	82	2012-08-02 05:27:35.849737
23074	5	1-3682	23070	7	16	838	80	2012-08-02 05:27:35.849737
23070	5	1-6713	23070	16	100	4655	74	2012-08-02 05:27:35.849737
23076	5	1-5408	23075	17	41	68	34	2012-08-02 05:27:35.849737
23077	5	1-4052	23075	41	48	13	79	2012-08-02 05:27:35.849737
23078	5	1-6707	23075	48	78	21	33	2012-08-02 05:27:35.849737
23079	5	1-5571	23075	78	16	28	80	2012-08-02 05:27:35.849737
23075	5	1-6713	23075	16	17	166	74	2012-08-02 05:27:35.849737
23081	5	1-3764	23080	91	34	120	81	2012-08-02 05:27:35.849737
23082	5	1-2308	23080	34	28	13	32	2012-08-02 05:27:35.849737
23083	5	1-6332	23080	28	78	49	33	2012-08-02 05:27:35.849737
23084	5	1-5571	23080	78	16	65	80	2012-08-02 05:27:35.849737
23080	5	1-6713	23080	16	91	390	74	2012-08-02 05:27:35.849737
23086	4	1-5693	23085	3	40	2012	51	2012-08-02 05:27:35.849737
23087	4	1-6595	23085	40	91	1405	21	2012-08-02 05:27:35.849737
23088	4	1-4610	23085	91	16	504	80	2012-08-02 05:27:35.849737
23085	4	1-6713	23085	16	3	3083	74	2012-08-02 05:27:35.849737
23090	4	1-6558	23089	50	74	2381	82	2012-08-02 05:27:36.623333
23091	4	1-6417	23089	74	17	3505	12	2012-08-02 05:27:36.623333
23092	4	1-5134	23089	17	67	2323	95	2012-08-02 05:27:36.623333
23089	4	1-6715	23089	67	50	1609	52	2012-08-02 05:27:36.623333
23093	1	1-1392	\N	19	19	130	12	2012-08-02 05:27:36.623333
23095	5	1-6573	23094	88	13	5302	84	2012-08-02 05:27:38.203127
23096	5	1-5893	23094	13	78	910	52	2012-08-02 05:27:38.203127
23097	5	1-5680	23094	78	36	746	76	2012-08-02 05:27:38.203127
23098	5	1-6677	23094	36	32	1238	2	2012-08-02 05:27:38.203127
23094	5	1-6719	23094	32	88	2461	24	2012-08-02 05:27:38.203127
23100	5	1-6041	23099	51	26	3286	37	2012-08-02 05:27:38.203127
23101	5	1-6689	23099	26	80	1610	25	2012-08-02 05:27:38.203127
23102	5	1-5679	23099	80	36	1019	76	2012-08-02 05:27:38.203127
23103	5	1-6677	23099	36	32	1887	2	2012-08-02 05:27:38.203127
23099	5	1-6719	23099	32	51	4187	24	2012-08-02 05:27:38.203127
23104	1	1-939	\N	45	45	331	76	2012-08-02 05:27:38.203127
23106	3	1-6346	23105	71	4	1943	47	2012-08-02 05:27:38.998728
23107	3	1-5088	23105	4	61	1776	65	2012-08-02 05:27:38.998728
23105	3	1-6721	23105	61	71	2212	72	2012-08-02 05:27:38.998728
23109	5	1-1791	23108	33	94	304	46	2012-08-02 05:27:39.396415
23110	5	1-6584	23108	94	35	493	7	2012-08-02 05:27:39.396415
23111	5	1-5091	23108	35	63	275	97	2012-08-02 05:27:39.396415
23112	5	1-6047	23108	63	78	250	11	2012-08-02 05:27:39.396415
23108	5	1-6722	23108	78	33	752	21	2012-08-02 05:27:39.396415
23114	4	1-6216	23113	19	31	3794	1	2012-08-02 05:27:39.396415
23115	4	1-6394	23113	31	63	2273	97	2012-08-02 05:27:39.396415
23116	4	1-6047	23113	63	78	2136	11	2012-08-02 05:27:39.396415
23113	4	1-6722	23113	78	19	6638	21	2012-08-02 05:27:39.396415
23118	5	1-6143	23117	67	58	302	42	2012-08-02 05:27:39.396415
23119	5	1-6576	23117	58	31	282	1	2012-08-02 05:27:39.396415
23120	5	1-6394	23117	31	63	174	97	2012-08-02 05:27:39.396415
23121	5	1-6047	23117	63	78	168	11	2012-08-02 05:27:39.396415
23117	5	1-6722	23117	78	67	536	21	2012-08-02 05:27:39.396415
23123	3	1-4610	23122	91	84	140	80	2012-08-02 05:27:39.396415
23124	3	1-6459	23122	84	78	91	11	2012-08-02 05:27:39.396415
23122	3	1-6722	23122	78	91	319	21	2012-08-02 05:27:39.396415
23126	5	1-6610	23125	67	44	1998	9	2012-08-02 05:27:39.805543
23127	5	1-3782	23125	44	70	1466	39	2012-08-02 05:27:39.805543
23128	5	1-5867	23125	70	72	1065	73	2012-08-02 05:27:39.805543
23129	5	1-6623	23125	72	21	2070	5	2012-08-02 05:27:39.805543
23125	5	1-6723	23125	21	67	2128	91	2012-08-02 05:27:39.805543
23131	5	1-6676	23130	75	73	2894	93	2012-08-02 05:27:39.805543
23132	5	1-4283	23130	73	86	761	61	2012-08-02 05:27:39.805543
23133	5	1-6531	23130	86	72	401	73	2012-08-02 05:27:39.805543
23134	5	1-6623	23130	72	21	820	5	2012-08-02 05:27:39.805543
23130	5	1-6723	23130	21	75	888	91	2012-08-02 05:27:39.805543
23136	5	1-6634	23135	14	80	4583	21	2012-08-02 05:27:39.805543
23137	5	1-6640	23135	80	77	5310	74	2012-08-02 05:27:39.805543
23138	5	1-2769	23135	77	72	1221	73	2012-08-02 05:27:39.805543
23139	5	1-6623	23135	72	21	2528	5	2012-08-02 05:27:39.805543
23135	5	1-6723	23135	21	14	2769	91	2012-08-02 05:27:39.805543
23141	3	1-6085	23140	4	35	124	66	2012-08-02 05:27:40.612505
23142	3	1-5964	23140	35	75	135	26	2012-08-02 05:27:40.612505
23140	3	1-6725	23140	75	4	151	68	2012-08-02 05:27:40.612505
23144	5	1-6104	23143	89	18	195	32	2012-08-02 05:27:40.612505
23145	5	1-5869	23143	18	51	98	3	2012-08-02 05:27:40.612505
23146	5	1-6583	23143	51	35	177	66	2012-08-02 05:27:40.612505
23147	5	1-5964	23143	35	75	243	26	2012-08-02 05:27:40.612505
23143	5	1-6725	23143	75	89	340	68	2012-08-02 05:27:40.612505
23149	5	1-4524	23148	94	72	352	73	2012-08-02 05:27:42.191957
23150	5	1-6623	23148	72	21	539	5	2012-08-02 05:27:42.191957
23151	5	1-6723	23148	21	58	436	91	2012-08-02 05:27:42.191957
23152	5	1-6473	23148	58	70	285	34	2012-08-02 05:27:42.191957
23148	5	1-6729	23148	70	94	628	47	2012-08-02 05:27:42.191957
23154	4	1-2216	23153	100	21	909	5	2012-08-02 05:27:42.191957
23155	4	1-6723	23153	21	58	820	91	2012-08-02 05:27:42.191957
23156	4	1-6473	23153	58	70	597	34	2012-08-02 05:27:42.191957
23153	4	1-6729	23153	70	100	1465	47	2012-08-02 05:27:42.191957
23158	4	1-6564	23157	10	54	3473	87	2012-08-02 05:27:43.352192
23159	4	1-6586	23157	54	67	3499	21	2012-08-02 05:27:43.352192
23160	4	1-6143	23157	67	30	1307	42	2012-08-02 05:27:43.352192
23157	4	1-6732	23157	30	10	5816	11	2012-08-02 05:27:43.352192
23162	4	1-2805	23161	3	83	1927	71	2012-08-02 05:27:44.158655
23163	4	1-6381	23161	83	97	1802	43	2012-08-02 05:27:44.158655
23164	4	1-6521	23161	97	73	2602	67	2012-08-02 05:27:44.158655
23161	4	1-6734	23161	73	3	5046	23	2012-08-02 05:27:44.158655
23166	5	1-6631	23165	80	42	305	40	2012-08-02 05:27:44.158655
23167	5	1-6694	23165	42	66	344	64	2012-08-02 05:27:44.158655
23168	5	1-4502	23165	66	76	265	88	2012-08-02 05:27:44.158655
23169	5	1-6489	23165	76	73	131	67	2012-08-02 05:27:44.158655
23165	5	1-6734	23165	73	80	262	23	2012-08-02 05:27:44.158655
23171	3	1-6526	23170	87	48	2346	74	2012-08-02 05:27:44.55662
23172	3	1-4911	23170	48	20	683	88	2012-08-02 05:27:44.55662
23170	3	1-6735	23170	20	87	1340	62	2012-08-02 05:27:44.55662
23174	4	1-6112	23173	30	56	1062	75	2012-08-02 05:27:44.55662
23175	4	1-2507	23173	56	95	606	61	2012-08-02 05:27:44.55662
23176	4	1-6078	23173	95	20	756	88	2012-08-02 05:27:44.55662
23173	4	1-6735	23173	20	30	2155	62	2012-08-02 05:27:44.55662
23178	4	1-6662	23177	59	77	1797	35	2012-08-02 05:27:44.55662
23179	4	1-5097	23177	77	95	1039	61	2012-08-02 05:27:44.55662
23180	4	1-6078	23177	95	20	1428	88	2012-08-02 05:27:44.55662
23177	4	1-6735	23177	20	59	4486	62	2012-08-02 05:27:44.55662
23182	3	1-6583	23181	51	99	1533	66	2012-08-02 05:27:45.341182
23181	3	1-6737	23181	72	51	1886	3	2012-08-02 05:27:45.341182
23183	3	1-6405	23181	99	72	1803	14	2012-08-02 05:27:45.341182
23185	2	1-3986	23184	58	72	2164	14	2012-08-02 05:27:45.341182
23184	2	1-6737	23184	72	58	3255	3	2012-08-02 05:27:45.341182
23187	3	1-6267	23186	99	32	4798	53	2012-08-02 05:27:45.341182
23188	3	1-3598	23186	32	72	2697	14	2012-08-02 05:27:45.341182
23186	3	1-6737	23186	72	99	4751	3	2012-08-02 05:27:45.341182
23190	5	1-6480	23189	23	89	693	68	2012-08-02 05:27:46.921214
23191	5	1-6104	23189	89	68	369	32	2012-08-02 05:27:46.921214
23192	5	1-5078	23189	68	96	325	69	2012-08-02 05:27:46.921214
23193	5	1-6368	23189	96	50	361	39	2012-08-02 05:27:46.921214
23189	5	1-6741	23189	50	23	1218	88	2012-08-02 05:27:46.921214
23195	5	1-6501	23194	77	28	1023	7	2012-08-02 05:27:47.307836
23196	5	1-4319	23194	28	51	410	39	2012-08-02 05:27:47.307836
23197	5	1-5541	23194	51	1	298	44	2012-08-02 05:27:47.307836
23198	5	1-6666	23194	1	63	327	36	2012-08-02 05:27:47.307836
23194	5	1-6742	23194	63	77	555	13	2012-08-02 05:27:47.307836
23200	5	1-6501	23199	77	66	1791	7	2012-08-02 05:27:47.307836
23201	5	1-4830	23199	66	12	478	64	2012-08-02 05:27:47.307836
23202	5	1-6555	23199	12	36	337	47	2012-08-02 05:27:47.307836
23203	5	1-6505	23199	36	63	462	36	2012-08-02 05:27:47.307836
23199	5	1-6742	23199	63	77	873	13	2012-08-02 05:27:47.307836
23205	4	1-6371	23204	53	48	2853	2	2012-08-02 05:27:47.307836
23206	4	1-5981	23204	48	36	1219	47	2012-08-02 05:27:47.307836
23207	4	1-6505	23204	36	63	1756	36	2012-08-02 05:27:47.307836
23204	4	1-6742	23204	63	53	3485	13	2012-08-02 05:27:47.307836
23209	4	1-6563	23208	50	86	2828	86	2012-08-02 05:27:47.307836
23210	4	1-4562	23208	86	1	1483	44	2012-08-02 05:27:47.307836
23211	4	1-6666	23208	1	63	1903	36	2012-08-02 05:27:47.307836
23208	4	1-6742	23208	63	50	3781	13	2012-08-02 05:27:47.307836
23213	5	1-5895	23212	61	25	750	98	2012-08-02 05:27:47.716909
23214	5	1-1221	23212	25	88	202	24	2012-08-02 05:27:47.716909
23215	5	1-6573	23212	88	96	463	84	2012-08-02 05:27:47.716909
23216	5	1-6544	23212	96	44	311	61	2012-08-02 05:27:47.716909
23212	5	1-6743	23212	44	61	1004	70	2012-08-02 05:27:47.716909
23218	4	1-6602	23217	100	73	2545	89	2012-08-02 05:27:48.10344
23219	4	1-6313	23217	73	45	4210	14	2012-08-02 05:27:48.10344
23220	4	1-6701	23217	45	95	1428	81	2012-08-02 05:27:48.10344
23217	4	1-6744	23217	95	100	1827	9	2012-08-02 05:27:48.10344
23222	4	1-4906	23221	9	42	256	40	2012-08-02 05:27:48.10344
23223	4	1-6694	23221	42	33	262	64	2012-08-02 05:27:48.10344
23224	4	1-6196	23221	33	95	145	81	2012-08-02 05:27:48.10344
23221	4	1-6744	23221	95	9	226	9	2012-08-02 05:27:48.10344
23226	4	1-5925	23225	17	73	62	56	2012-08-02 05:27:48.10344
23227	4	1-3811	23225	73	45	113	14	2012-08-02 05:27:48.10344
23228	4	1-6701	23225	45	95	49	81	2012-08-02 05:27:48.10344
23225	4	1-6744	23225	95	17	81	9	2012-08-02 05:27:48.10344
23230	5	1-6360	23229	80	79	631	90	2012-08-02 05:27:49.683686
23231	5	1-6549	23229	79	19	589	98	2012-08-02 05:27:49.683686
23232	5	1-5231	23229	19	44	447	72	2012-08-02 05:27:49.683686
23233	5	1-6305	23229	44	44	578	28	2012-08-02 05:27:49.683686
23229	5	1-6748	23229	44	80	560	8	2012-08-02 05:27:49.683686
23235	5	1-6191	23234	17	40	1585	60	2012-08-02 05:27:49.683686
23236	5	1-6020	23234	40	28	2757	59	2012-08-02 05:27:49.683686
23237	5	1-6238	23234	28	44	1040	72	2012-08-02 05:27:49.683686
23238	5	1-6305	23234	44	44	1383	28	2012-08-02 05:27:49.683686
23234	5	1-6748	23234	44	17	1376	8	2012-08-02 05:27:49.683686
23240	5	1-5088	23239	4	18	562	65	2012-08-02 05:27:50.081364
23241	5	1-5668	23239	18	27	408	10	2012-08-02 05:27:50.081364
23242	5	1-5391	23239	27	50	257	39	2012-08-02 05:27:50.081364
23243	5	1-6741	23239	50	50	971	88	2012-08-02 05:27:50.081364
23239	5	1-6749	23239	50	4	656	47	2012-08-02 05:27:50.081364
23245	4	1-4442	23244	10	17	273	25	2012-08-02 05:27:50.081364
23246	4	1-6268	23244	17	95	179	61	2012-08-02 05:27:50.081364
23247	4	1-6078	23244	95	50	324	88	2012-08-02 05:27:50.081364
23244	4	1-6749	23244	50	10	225	47	2012-08-02 05:27:50.081364
23249	5	1-6695	23248	93	39	3183	83	2012-08-02 05:27:50.879047
23250	5	1-6625	23248	39	14	3061	29	2012-08-02 05:27:50.879047
23251	5	1-6687	23248	14	56	1402	56	2012-08-02 05:27:50.879047
23252	5	1-3563	23248	56	95	972	93	2012-08-02 05:27:50.879047
23248	5	1-6751	23248	95	93	1169	35	2012-08-02 05:27:50.879047
23254	3	1-1904	23253	93	75	1472	91	2012-08-02 05:27:50.879047
23255	3	1-6676	23253	75	95	3298	93	2012-08-02 05:27:50.879047
23253	3	1-6751	23253	95	93	4149	35	2012-08-02 05:27:50.879047
23257	4	1-6395	23256	60	27	709	1	2012-08-02 05:27:50.879047
23258	4	1-6407	23256	27	56	468	56	2012-08-02 05:27:50.879047
23259	4	1-3563	23256	56	95	367	93	2012-08-02 05:27:50.879047
23256	4	1-6751	23256	95	60	500	35	2012-08-02 05:27:50.879047
23261	3	1-3158	23260	68	14	805	83	2012-08-02 05:27:51.276722
23262	3	1-6745	23260	14	42	690	36	2012-08-02 05:27:51.276722
23260	3	1-6752	23260	42	68	1365	24	2012-08-02 05:27:51.276722
23264	5	1-6249	23263	25	62	862	77	2012-08-02 05:27:53.674302
23265	5	1-6646	23263	62	88	1813	61	2012-08-02 05:27:53.674302
23266	5	1-194	23263	88	61	619	65	2012-08-02 05:27:53.674302
23267	5	1-6721	23263	61	16	784	72	2012-08-02 05:27:53.674302
23263	5	1-6758	23263	16	25	751	69	2012-08-02 05:27:53.674302
23269	4	1-2987	23268	38	17	483	8	2012-08-02 05:27:59.960224
23270	4	1-6191	23268	17	40	510	60	2012-08-02 05:27:59.960224
23271	4	1-6020	23268	40	60	813	59	2012-08-02 05:27:59.960224
23268	4	1-6774	23268	60	38	3861	7	2012-08-02 05:27:59.960224
23273	5	1-3233	23272	65	42	302	37	2012-08-02 05:28:00.347029
23274	5	1-6712	23272	42	94	328	18	2012-08-02 05:28:00.347029
23275	5	1-6232	23272	94	5	542	100	2012-08-02 05:28:00.347029
23276	5	1-6523	23272	5	77	683	41	2012-08-02 05:28:00.347029
23272	5	1-6775	23272	77	65	346	8	2012-08-02 05:28:00.347029
23278	5	1-4564	23277	27	60	713	59	2012-08-02 05:28:01.529297
23279	5	1-6774	23277	60	38	3565	7	2012-08-02 05:28:01.529297
23280	5	1-2987	23277	38	17	470	8	2012-08-02 05:28:01.529297
23281	5	1-6191	23277	17	98	521	60	2012-08-02 05:28:01.529297
23277	5	1-6778	23277	98	27	900	25	2012-08-02 05:28:01.529297
23283	5	1-6268	23282	17	48	2625	61	2012-08-02 05:28:01.529297
23284	5	1-6309	23282	48	51	3118	64	2012-08-02 05:28:01.529297
23285	5	1-5708	23282	51	7	2065	39	2012-08-02 05:28:01.529297
23286	5	1-6628	23282	7	98	2211	60	2012-08-02 05:28:01.529297
23282	5	1-6778	23282	98	17	4056	25	2012-08-02 05:28:01.529297
23288	3	1-5408	23287	17	67	1103	34	2012-08-02 05:28:02.700392
23289	3	1-6704	23287	67	17	2969	23	2012-08-02 05:28:02.700392
23287	3	1-6781	23287	17	17	2547	74	2012-08-02 05:28:02.700392
23291	5	1-6295	23290	1	72	1053	94	2012-08-02 05:28:03.485085
23292	5	1-6487	23290	72	3	984	43	2012-08-02 05:28:03.485085
23293	5	1-6347	23290	3	61	937	98	2012-08-02 05:28:03.485085
23294	5	1-2920	23290	61	89	1435	31	2012-08-02 05:28:03.485085
23290	5	1-6783	23290	89	1	1193	17	2012-08-02 05:28:03.485085
23296	3	1-6078	23295	95	50	783	88	2012-08-02 05:28:05.429954
23297	3	1-6749	23295	50	30	539	47	2012-08-02 05:28:05.429954
23295	3	1-6788	23295	30	95	436	61	2012-08-02 05:28:05.429954
23299	3	1-5097	23298	77	65	3284	61	2012-08-02 05:28:07.430377
23300	3	1-5440	23298	65	86	1169	96	2012-08-02 05:28:07.430377
23298	3	1-6793	23298	86	77	4803	35	2012-08-02 05:28:07.430377
23302	5	1-3662	23301	49	98	436	60	2012-08-02 05:28:07.430377
23303	5	1-6778	23301	98	17	734	25	2012-08-02 05:28:07.430377
23304	5	1-6268	23301	17	65	437	61	2012-08-02 05:28:07.430377
23305	5	1-5440	23301	65	86	157	96	2012-08-02 05:28:07.430377
23301	5	1-6793	23301	86	49	649	35	2012-08-02 05:28:07.430377
23307	4	1-6003	23306	70	93	2282	76	2012-08-02 05:28:07.430377
23308	4	1-6787	23306	93	78	4560	42	2012-08-02 05:28:07.430377
23309	4	1-3066	23306	78	86	579	96	2012-08-02 05:28:07.430377
23306	4	1-6793	23306	86	70	2424	35	2012-08-02 05:28:07.430377
23311	4	1-6395	23310	60	31	472	1	2012-08-02 05:28:07.430377
23312	4	1-6394	23310	31	28	324	97	2012-08-02 05:28:07.430377
23313	4	1-3529	23310	28	86	56	96	2012-08-02 05:28:07.430377
23310	4	1-6793	23310	86	60	250	35	2012-08-02 05:28:07.430377
23315	4	1-5325	23314	59	42	650	34	2012-08-02 05:28:07.850021
23316	4	1-6730	23314	42	19	2163	92	2012-08-02 05:28:07.850021
23317	4	1-2828	23314	19	98	1132	51	2012-08-02 05:28:07.850021
23314	4	1-6794	23314	98	59	1715	4	2012-08-02 05:28:07.850021
23319	4	1-6433	23318	35	100	762	9	2012-08-02 05:28:07.850021
23320	4	1-6602	23318	100	95	1304	89	2012-08-02 05:28:07.850021
23321	4	1-3321	23318	95	98	730	51	2012-08-02 05:28:07.850021
23318	4	1-6794	23318	98	35	1115	4	2012-08-02 05:28:07.850021
23323	4	1-6793	23322	86	91	679	35	2012-08-02 05:28:08.247994
23324	4	1-4260	23322	91	17	431	46	2012-08-02 05:28:08.247994
23325	4	1-6325	23322	17	39	248	6	2012-08-02 05:28:08.247994
23322	4	1-6795	23322	39	86	220	96	2012-08-02 05:28:08.247994
23327	4	1-6687	23326	14	91	1464	56	2012-08-02 05:28:08.668104
23328	4	1-6524	23326	91	46	543	55	2012-08-02 05:28:08.668104
23329	4	1-6278	23326	46	25	509	78	2012-08-02 05:28:08.668104
23326	4	1-6796	23326	25	14	2412	29	2012-08-02 05:28:08.668104
23331	5	1-4716	23330	36	45	259	4	2012-08-02 05:28:08.668104
23332	5	1-6133	23330	45	67	240	95	2012-08-02 05:28:08.668104
23333	5	1-6715	23330	67	31	185	52	2012-08-02 05:28:08.668104
23334	5	1-5421	23330	31	25	120	78	2012-08-02 05:28:08.668104
23330	5	1-6796	23330	25	36	637	29	2012-08-02 05:28:08.668104
23336	5	1-4105	23335	81	52	4342	87	2012-08-02 05:28:09.076678
23337	5	1-6777	23335	52	50	5957	88	2012-08-02 05:28:09.076678
23338	5	1-6749	23335	50	7	4133	47	2012-08-02 05:28:09.076678
23339	5	1-6159	23335	7	78	4245	66	2012-08-02 05:28:09.076678
23335	5	1-6797	23335	78	81	5777	26	2012-08-02 05:28:09.076678
23341	5	1-6268	23340	17	65	10	61	2012-08-02 05:28:09.894263
23342	5	1-5440	23340	65	86	4	96	2012-08-02 05:28:09.894263
23343	5	1-6793	23340	86	70	18	35	2012-08-02 05:28:09.894263
23344	5	1-6003	23340	70	41	18	76	2012-08-02 05:28:09.894263
23340	5	1-6799	23340	41	17	15	25	2012-08-02 05:28:09.894263
23346	4	1-5999	23345	77	60	1843	31	2012-08-02 05:28:10.30332
23347	4	1-6791	23345	60	48	2724	2	2012-08-02 05:28:10.30332
23348	4	1-5981	23345	48	22	818	47	2012-08-02 05:28:10.30332
23345	4	1-6800	23345	22	77	3189	87	2012-08-02 05:28:10.30332
23350	4	1-6586	23349	54	53	2049	21	2012-08-02 05:28:10.30332
23351	4	1-2653	23349	53	12	350	64	2012-08-02 05:28:10.30332
23352	4	1-6555	23349	12	22	248	47	2012-08-02 05:28:10.30332
23349	4	1-6800	23349	22	54	1319	87	2012-08-02 05:28:10.30332
23354	5	1-6777	23353	52	76	1124	88	2012-08-02 05:28:10.30332
23355	5	1-6489	23353	76	63	509	67	2012-08-02 05:28:10.30332
23356	5	1-5666	23353	63	90	269	25	2012-08-02 05:28:10.30332
23357	5	1-6792	23353	90	22	162	47	2012-08-02 05:28:10.30332
23353	5	1-6800	23353	22	52	886	87	2012-08-02 05:28:10.30332
23359	5	1-6586	23358	54	33	817	21	2012-08-02 05:28:10.30332
23360	5	1-1791	23358	33	30	397	46	2012-08-02 05:28:10.30332
23361	5	1-6007	23358	30	90	126	25	2012-08-02 05:28:10.30332
23362	5	1-6792	23358	90	22	81	47	2012-08-02 05:28:10.30332
23358	5	1-6800	23358	22	54	476	87	2012-08-02 05:28:10.30332
23364	3	1-6158	23363	67	47	1215	62	2012-08-02 05:28:11.949295
23365	3	1-5567	23363	47	6	1116	54	2012-08-02 05:28:11.949295
23363	3	1-6804	23363	6	67	1890	60	2012-08-02 05:28:11.949295
23367	4	1-6060	23366	4	91	2215	70	2012-08-02 05:28:11.949295
23368	4	1-3216	23366	91	62	1839	14	2012-08-02 05:28:11.949295
23369	4	1-6477	23366	62	6	1567	54	2012-08-02 05:28:11.949295
23366	4	1-6804	23366	6	4	2725	60	2012-08-02 05:28:11.949295
23371	3	1-6158	23370	67	46	1534	62	2012-08-02 05:28:11.949295
23372	3	1-5444	23370	46	6	1320	54	2012-08-02 05:28:11.949295
23370	3	1-6804	23370	6	67	2309	60	2012-08-02 05:28:11.949295
23374	3	1-5359	23373	35	3	2231	90	2012-08-02 05:28:11.949295
23375	3	1-6652	23373	3	6	1297	54	2012-08-02 05:28:11.949295
23373	3	1-6804	23373	6	35	2352	60	2012-08-02 05:28:11.949295
23377	3	1-6791	23376	60	9	6902	2	2012-08-02 05:28:12.756067
23378	3	1-4983	23376	9	66	1721	50	2012-08-02 05:28:12.756067
23376	3	1-6806	23376	66	60	3789	31	2012-08-02 05:28:12.756067
23380	5	1-6479	23379	57	73	1210	1	2012-08-02 05:28:13.164874
23381	5	1-6596	23379	73	42	332	40	2012-08-02 05:28:13.164874
23382	5	1-6694	23379	42	12	373	64	2012-08-02 05:28:13.164874
23383	5	1-6555	23379	12	51	299	47	2012-08-02 05:28:13.164874
23379	5	1-6807	23379	51	57	4092	92	2012-08-02 05:28:13.164874
23385	5	1-6262	23384	73	68	1546	32	2012-08-02 05:28:13.993641
23386	5	1-5078	23384	68	25	1245	69	2012-08-02 05:28:13.993641
23387	5	1-6249	23384	25	100	1130	77	2012-08-02 05:28:13.993641
23388	5	1-5352	23384	100	41	913	67	2012-08-02 05:28:13.993641
23384	5	1-6809	23384	41	73	5175	2	2012-08-02 05:28:13.993641
23390	4	1-3476	23389	74	73	8	67	2012-08-02 05:28:16.447961
23391	4	1-6734	23389	73	96	15	23	2012-08-02 05:28:16.447961
23389	4	1-6815	23389	60	74	21	38	2012-08-02 05:28:16.447961
23392	4	1-6098	23389	96	60	21	54	2012-08-02 05:28:16.447961
23394	5	1-6047	23393	63	74	347	11	2012-08-02 05:28:17.265827
23395	5	1-5806	23393	74	96	219	84	2012-08-02 05:28:17.265827
23396	5	1-6544	23393	96	44	164	61	2012-08-02 05:28:17.265827
23397	5	1-6743	23393	44	28	592	70	2012-08-02 05:28:17.265827
23393	5	1-6817	23393	28	63	325	97	2012-08-02 05:28:17.265827
23399	5	1-6815	23398	60	88	1453	38	2012-08-02 05:28:17.685292
23400	5	1-5260	23398	88	3	436	23	2012-08-02 05:28:17.685292
23401	5	1-2805	23398	3	9	128	71	2012-08-02 05:28:17.685292
23402	5	1-6801	23398	9	38	581	10	2012-08-02 05:28:17.685292
23398	5	1-6818	23398	38	60	1725	54	2012-08-02 05:28:17.685292
23404	5	1-6815	23403	60	74	2260	38	2012-08-02 05:28:17.685292
23405	5	1-3476	23403	74	17	830	67	2012-08-02 05:28:17.685292
23406	5	1-5122	23403	17	9	145	71	2012-08-02 05:28:17.685292
23407	5	1-6801	23403	9	38	730	10	2012-08-02 05:28:17.685292
23403	5	1-6818	23403	38	60	2411	54	2012-08-02 05:28:17.685292
23409	3	1-6018	23408	25	23	582	8	2012-08-02 05:28:18.094234
23410	3	1-6519	23408	23	79	428	6	2012-08-02 05:28:18.094234
23408	3	1-6819	23408	79	25	3754	87	2012-08-02 05:28:18.094234
23412	5	1-5870	23411	94	91	2453	50	2012-08-02 05:28:18.094234
23413	5	1-6616	23411	91	31	1358	9	2012-08-02 05:28:18.094234
23414	5	1-4280	23411	31	85	573	20	2012-08-02 05:28:18.094234
23415	5	1-6337	23411	85	79	446	6	2012-08-02 05:28:18.094234
23411	5	1-6819	23411	79	94	4043	87	2012-08-02 05:28:18.094234
23417	5	1-3782	23416	44	41	1185	39	2012-08-02 05:28:18.514158
23418	5	1-6528	23416	41	85	2053	58	2012-08-02 05:28:18.514158
23419	5	1-6767	23416	85	94	2029	46	2012-08-02 05:28:18.514158
23420	5	1-6584	23416	94	32	3248	7	2012-08-02 05:28:18.514158
23416	5	1-6820	23416	32	44	1852	9	2012-08-02 05:28:18.514158
23422	5	1-6488	23421	77	53	1563	59	2012-08-02 05:28:18.514158
23423	5	1-4705	23421	53	85	1357	58	2012-08-02 05:28:18.514158
23424	5	1-6767	23421	85	94	1424	46	2012-08-02 05:28:18.514158
23425	5	1-6584	23421	94	32	2420	7	2012-08-02 05:28:18.514158
23421	5	1-6820	23421	32	77	1465	9	2012-08-02 05:28:18.514158
23427	5	1-2754	23426	97	50	366	52	2012-08-02 05:28:18.514158
23428	5	1-6558	23426	50	54	640	82	2012-08-02 05:28:18.514158
23429	5	1-5753	23426	54	77	576	13	2012-08-02 05:28:18.514158
23430	5	1-6501	23426	77	32	1348	7	2012-08-02 05:28:18.514158
23426	5	1-6820	23426	32	97	958	9	2012-08-02 05:28:18.514158
23432	4	1-2216	23431	100	21	254	5	2012-08-02 05:28:19.751657
23433	4	1-6723	23431	21	75	163	91	2012-08-02 05:28:19.751657
23434	4	1-6676	23431	75	79	315	93	2012-08-02 05:28:19.751657
23431	4	1-6823	23431	79	100	575	47	2012-08-02 05:28:19.751657
23436	5	1-6807	23435	51	57	247	92	2012-08-02 05:28:19.751657
23437	5	1-6479	23435	57	27	44	1	2012-08-02 05:28:19.751657
23438	5	1-6407	23435	27	56	24	56	2012-08-02 05:28:19.751657
23439	5	1-3563	23435	56	79	16	93	2012-08-02 05:28:19.751657
23435	5	1-6823	23435	79	51	30	47	2012-08-02 05:28:19.751657
23441	4	1-6788	23440	30	44	2123	61	2012-08-02 05:28:19.751657
23442	4	1-6743	23440	44	42	5558	70	2012-08-02 05:28:19.751657
23443	4	1-4241	23440	42	79	1692	93	2012-08-02 05:28:19.751657
23440	4	1-6823	23440	79	30	3713	47	2012-08-02 05:28:19.751657
23445	4	1-4954	23444	30	50	943	52	2012-08-02 05:28:20.558299
23446	4	1-6558	23444	50	55	1566	82	2012-08-02 05:28:20.558299
23447	4	1-4966	23444	55	82	1425	11	2012-08-02 05:28:20.558299
23444	4	1-6825	23444	82	30	1950	16	2012-08-02 05:28:20.558299
23449	4	1-6545	23448	74	50	185	52	2012-08-02 05:28:20.558299
23450	4	1-6558	23448	50	55	311	82	2012-08-02 05:28:20.558299
23451	4	1-4966	23448	55	82	286	11	2012-08-02 05:28:20.558299
23448	4	1-6825	23448	82	74	395	16	2012-08-02 05:28:20.558299
23453	3	1-6702	23452	79	19	3825	57	2012-08-02 05:28:21.76258
23454	3	1-5069	23452	19	92	881	92	2012-08-02 05:28:21.76258
23452	3	1-6828	23452	92	79	1310	10	2012-08-02 05:28:21.76258
23456	4	1-5178	23455	46	81	639	22	2012-08-02 05:28:22.580193
23457	4	1-6575	23455	81	9	498	71	2012-08-02 05:28:22.580193
23458	4	1-6801	23455	9	1	2369	10	2012-08-02 05:28:22.580193
23455	4	1-6830	23455	1	46	1859	28	2012-08-02 05:28:22.580193
23460	4	1-6748	23459	44	82	2643	8	2012-08-02 05:28:22.580193
23461	4	1-1806	23459	82	92	1182	92	2012-08-02 05:28:22.580193
23462	4	1-6828	23459	92	1	2696	10	2012-08-02 05:28:22.580193
23459	4	1-6830	23459	1	44	2647	28	2012-08-02 05:28:22.580193
23464	5	1-6331	23463	90	32	469	53	2012-08-02 05:28:22.989211
23465	5	1-3598	23463	32	45	329	14	2012-08-02 05:28:22.989211
23466	5	1-6701	23463	45	95	131	81	2012-08-02 05:28:22.989211
23467	5	1-6744	23463	95	36	198	9	2012-08-02 05:28:22.989211
23463	5	1-6831	23463	36	90	281	29	2012-08-02 05:28:22.989211
23469	4	1-6687	23468	14	40	1433	56	2012-08-02 05:28:22.989211
23470	4	1-4065	23468	40	95	910	81	2012-08-02 05:28:22.989211
23471	4	1-6744	23468	95	36	1444	9	2012-08-02 05:28:22.989211
23468	4	1-6831	23468	36	14	2162	29	2012-08-02 05:28:22.989211
23473	3	1-4386	23472	51	40	890	51	2012-08-02 05:28:24.646396
23474	3	1-6595	23472	40	49	437	21	2012-08-02 05:28:24.646396
23472	3	1-6835	23472	49	51	1925	27	2012-08-02 05:28:24.646396
23476	5	1-6718	23475	94	84	2279	80	2012-08-02 05:28:24.646396
23477	5	1-6459	23475	84	10	960	11	2012-08-02 05:28:24.646396
23478	5	1-6564	23475	10	54	617	87	2012-08-02 05:28:24.646396
23479	5	1-6586	23475	54	49	669	21	2012-08-02 05:28:24.646396
23475	5	1-6835	23475	49	94	3322	27	2012-08-02 05:28:24.646396
23481	5	1-5736	23480	66	78	62	39	2012-08-02 05:28:24.646396
23482	5	1-4868	23480	78	87	42	40	2012-08-02 05:28:24.646396
23483	5	1-5060	23480	87	54	18	87	2012-08-02 05:28:24.646396
23484	5	1-6586	23480	54	49	23	21	2012-08-02 05:28:24.646396
23480	5	1-6835	23480	49	66	138	27	2012-08-02 05:28:24.646396
23486	5	1-6386	23485	9	99	1143	79	2012-08-02 05:28:24.646396
23487	5	1-3570	23485	99	21	132	5	2012-08-02 05:28:24.646396
23488	5	1-6723	23485	21	14	123	91	2012-08-02 05:28:24.646396
23489	5	1-6634	23485	14	49	171	21	2012-08-02 05:28:24.646396
23485	5	1-6835	23485	49	9	1143	27	2012-08-02 05:28:24.646396
23491	5	1-3064	23490	44	16	449	72	2012-08-02 05:28:24.646396
23492	5	1-2526	23490	16	21	264	5	2012-08-02 05:28:24.646396
23493	5	1-6723	23490	21	14	253	91	2012-08-02 05:28:24.646396
23494	5	1-6634	23490	14	49	365	21	2012-08-02 05:28:24.646396
23490	5	1-6835	23490	49	44	2518	27	2012-08-02 05:28:24.646396
23496	5	1-6356	23495	5	61	1951	65	2012-08-02 05:28:25.055625
23497	5	1-6721	23495	61	43	1198	72	2012-08-02 05:28:25.055625
23498	5	1-1718	23495	43	39	141	6	2012-08-02 05:28:25.055625
23499	5	1-6795	23495	39	89	90	96	2012-08-02 05:28:25.055625
23495	5	1-6836	23495	89	5	3532	34	2012-08-02 05:28:25.055625
23501	4	1-6356	23500	5	44	676	65	2012-08-02 05:28:25.055625
23502	4	1-2982	23500	44	39	33	6	2012-08-02 05:28:25.055625
23503	4	1-6795	23500	39	89	24	96	2012-08-02 05:28:25.055625
23500	4	1-6836	23500	89	5	1077	34	2012-08-02 05:28:25.055625
23505	4	1-6780	23504	56	66	286	77	2012-08-02 05:28:25.464564
23506	4	1-5073	23504	66	50	214	39	2012-08-02 05:28:25.464564
23507	4	1-6741	23504	50	100	846	88	2012-08-02 05:28:25.464564
23504	4	1-6837	23504	100	56	340	75	2012-08-02 05:28:25.464564
23508	1	1-2913	\N	6	6	2316	39	2012-08-02 05:28:25.464564
23510	5	1-6739	23509	98	31	1300	86	2012-08-02 05:28:27.088405
23511	5	1-6808	23509	31	5	1012	34	2012-08-02 05:28:27.088405
23512	5	1-6356	23509	5	24	1047	65	2012-08-02 05:28:27.088405
23513	5	1-6239	23509	24	71	1327	79	2012-08-02 05:28:27.088405
23509	5	1-6841	23509	71	98	1981	62	2012-08-02 05:28:27.088405
23515	5	1-6662	23514	59	9	1668	35	2012-08-02 05:28:27.088405
23516	5	1-6449	23514	9	9	3236	70	2012-08-02 05:28:27.088405
23517	5	1-6422	23514	9	100	1736	45	2012-08-02 05:28:27.088405
23518	5	1-6412	23514	100	71	2267	79	2012-08-02 05:28:27.088405
23514	5	1-6841	23514	71	59	3428	62	2012-08-02 05:28:27.088405
23520	5	1-6268	23519	17	44	257	61	2012-08-02 05:28:28.348161
23521	5	1-6743	23519	44	28	886	70	2012-08-02 05:28:28.348161
23522	5	1-6817	23519	28	79	465	97	2012-08-02 05:28:28.348161
23523	5	1-5845	23519	79	17	302	30	2012-08-02 05:28:28.348161
23519	5	1-6844	23519	17	17	402	25	2012-08-02 05:28:28.348161
23525	5	1-6050	23524	100	36	1382	14	2012-08-02 05:28:29.618384
23526	5	1-6063	23524	36	84	1078	80	2012-08-02 05:28:29.618384
23527	5	1-6459	23524	84	82	709	11	2012-08-02 05:28:29.618384
23528	5	1-6825	23524	82	26	981	16	2012-08-02 05:28:29.618384
23524	5	1-6847	23524	26	100	4262	51	2012-08-02 05:28:29.618384
23530	4	1-6795	23529	39	57	893	96	2012-08-02 05:28:31.662109
23531	4	1-4318	23529	57	56	502	56	2012-08-02 05:28:31.662109
23532	4	1-3563	23529	56	25	475	93	2012-08-02 05:28:31.662109
23529	4	1-6852	23529	25	39	771	6	2012-08-02 05:28:31.662109
23534	4	1-4325	23533	89	66	444	50	2012-08-02 05:28:31.662109
23535	4	1-6806	23533	66	70	1081	31	2012-08-02 05:28:31.662109
23536	4	1-6607	23533	70	25	797	93	2012-08-02 05:28:31.662109
23533	4	1-6852	23533	25	89	1317	6	2012-08-02 05:28:31.662109
23538	4	1-6795	23537	39	57	4692	96	2012-08-02 05:28:31.662109
23539	4	1-4318	23537	57	7	2720	56	2012-08-02 05:28:31.662109
23540	4	1-1006	23537	7	25	2347	93	2012-08-02 05:28:31.662109
23537	4	1-6852	23537	25	39	3931	6	2012-08-02 05:28:31.662109
23541	1	1-1006	\N	7	7	6030	93	2012-08-02 05:28:31.662109
23543	4	1-6613	23542	63	46	2601	28	2012-08-02 05:28:32.08232
23544	4	1-5178	23542	46	81	1147	22	2012-08-02 05:28:32.08232
23545	4	1-6575	23542	81	81	1148	71	2012-08-02 05:28:32.08232
23542	4	1-6853	23542	81	63	2036	55	2012-08-02 05:28:32.08232
23547	5	1-6391	23546	40	41	2617	73	2012-08-02 05:28:33.275318
23548	5	1-6399	23546	41	25	2050	93	2012-08-02 05:28:33.275318
23549	5	1-6852	23546	25	26	2920	6	2012-08-02 05:28:33.275318
23550	5	1-6821	23546	26	79	2150	77	2012-08-02 05:28:33.275318
23546	5	1-6856	23546	79	40	2113	22	2012-08-02 05:28:33.275318
23552	5	1-2227	23551	71	67	462	38	2012-08-02 05:28:33.695249
23553	5	1-6339	23551	67	93	300	59	2012-08-02 05:28:33.695249
23554	5	1-6061	23551	93	100	92	45	2012-08-02 05:28:33.695249
23555	5	1-6412	23551	100	89	85	79	2012-08-02 05:28:33.695249
23551	5	1-6857	23551	89	71	6485	92	2012-08-02 05:28:33.695249
23557	3	1-6257	23556	20	62	1957	14	2012-08-02 05:28:35.341652
23558	3	1-6477	23556	62	90	1894	54	2012-08-02 05:28:35.341652
23556	3	1-6861	23556	90	20	1303	22	2012-08-02 05:28:35.341652
23560	4	1-6727	23559	44	90	933	86	2012-08-02 05:28:35.341652
23561	4	1-6736	23559	90	96	1073	23	2012-08-02 05:28:35.341652
23562	4	1-6098	23559	96	90	1577	54	2012-08-02 05:28:35.341652
23559	4	1-6861	23559	90	44	1130	22	2012-08-02 05:28:35.341652
23564	5	1-5076	23563	58	16	1151	72	2012-08-02 05:28:35.750175
23565	5	1-6758	23563	16	25	997	69	2012-08-02 05:28:35.750175
23566	5	1-6249	23563	25	79	1035	77	2012-08-02 05:28:35.750175
23567	5	1-6856	23563	79	52	1195	22	2012-08-02 05:28:35.750175
23563	5	1-6862	23563	52	58	2395	12	2012-08-02 05:28:35.750175
23569	5	1-6109	23568	4	91	1180	99	2012-08-02 05:28:35.750175
23570	5	1-6427	23568	91	96	1656	23	2012-08-02 05:28:35.750175
23571	5	1-6098	23568	96	90	2499	54	2012-08-02 05:28:35.750175
23572	5	1-6861	23568	90	52	1765	22	2012-08-02 05:28:35.750175
23568	5	1-6862	23568	52	4	3774	12	2012-08-02 05:28:35.750175
23574	5	1-686	23573	36	25	96	93	2012-08-02 05:28:35.750175
23575	5	1-6852	23573	25	26	173	6	2012-08-02 05:28:35.750175
23576	5	1-6821	23573	26	79	161	77	2012-08-02 05:28:35.750175
23577	5	1-6856	23573	79	52	199	22	2012-08-02 05:28:35.750175
23573	5	1-6862	23573	52	36	429	12	2012-08-02 05:28:35.750175
23579	3	1-6708	23578	65	37	2876	64	2012-08-02 05:28:36.159226
23580	3	1-5690	23578	37	75	1366	70	2012-08-02 05:28:36.159226
23578	3	1-6863	23578	75	65	7411	98	2012-08-02 05:28:36.159226
23582	2	1-6674	23581	28	49	319	42	2012-08-02 05:28:36.568033
23581	2	1-6864	23581	49	28	2888	67	2012-08-02 05:28:36.568033
23584	4	1-3500	23583	44	21	1231	5	2012-08-02 05:28:36.965902
23585	4	1-6723	23583	21	70	1191	91	2012-08-02 05:28:36.965902
23586	4	1-6635	23583	70	99	1115	41	2012-08-02 05:28:36.965902
23583	4	1-6865	23583	99	44	1216	75	2012-08-02 05:28:36.965902
23588	4	1-2507	23587	56	27	427	61	2012-08-02 05:28:36.965902
23589	4	1-6241	23587	27	5	443	100	2012-08-02 05:28:36.965902
23590	4	1-6523	23587	5	99	516	41	2012-08-02 05:28:36.965902
23587	4	1-6865	23587	99	56	575	75	2012-08-02 05:28:36.965902
23592	5	1-4095	23591	5	17	39	25	2012-08-02 05:28:36.965902
23593	5	1-6268	23591	17	27	23	61	2012-08-02 05:28:36.965902
23594	5	1-6241	23591	27	5	24	100	2012-08-02 05:28:36.965902
23595	5	1-6523	23591	5	99	30	41	2012-08-02 05:28:36.965902
23591	5	1-6865	23591	99	5	34	75	2012-08-02 05:28:36.965902
23596	1	1-2282	\N	21	21	2130	61	2012-08-02 05:28:36.965902
23598	4	1-6744	23597	95	100	653	9	2012-08-02 05:28:39.374277
23599	4	1-6602	23597	100	11	1007	89	2012-08-02 05:28:39.374277
23600	4	1-6512	23597	11	45	659	76	2012-08-02 05:28:39.374277
23597	4	1-6871	23597	45	95	461	81	2012-08-02 05:28:39.374277
23602	5	1-3032	23601	62	73	785	2	2012-08-02 05:28:39.771967
23603	5	1-6262	23601	73	68	275	32	2012-08-02 05:28:39.771967
23604	5	1-5078	23601	68	80	260	69	2012-08-02 05:28:39.771967
23605	5	1-6786	23601	80	76	231	5	2012-08-02 05:28:39.771967
23601	5	1-6872	23601	76	62	1286	79	2012-08-02 05:28:39.771967
23607	4	1-3351	23606	30	46	201	55	2012-08-02 05:28:39.771967
23608	4	1-6278	23606	46	56	205	78	2012-08-02 05:28:39.771967
23609	4	1-6281	23606	56	76	118	5	2012-08-02 05:28:39.771967
23606	4	1-6872	23606	76	30	659	79	2012-08-02 05:28:39.771967
23611	5	1-6322	23610	16	61	3071	10	2012-08-02 05:28:39.771967
23612	5	1-4221	23610	61	99	417	41	2012-08-02 05:28:39.771967
23613	5	1-6865	23610	99	44	521	75	2012-08-02 05:28:39.771967
23614	5	1-3500	23610	44	76	604	5	2012-08-02 05:28:39.771967
23610	5	1-6872	23610	76	16	3632	79	2012-08-02 05:28:39.771967
23616	5	1-4194	23615	26	41	919	76	2012-08-02 05:28:40.192064
23617	5	1-6799	23615	41	17	727	25	2012-08-02 05:28:40.192064
23618	5	1-6268	23615	17	44	461	61	2012-08-02 05:28:40.192064
23619	5	1-6743	23615	44	8	1663	70	2012-08-02 05:28:40.192064
23615	5	1-6873	23615	8	26	2230	42	2012-08-02 05:28:40.192064
23621	5	1-6743	23620	44	8	256	70	2012-08-02 05:28:41.572939
23622	5	1-6873	23620	8	30	308	42	2012-08-02 05:28:41.572939
23623	5	1-4766	23620	30	44	71	28	2012-08-02 05:28:41.572939
23624	5	1-6748	23620	44	1	67	8	2012-08-02 05:28:41.572939
23620	5	1-6876	23620	1	44	73	61	2012-08-02 05:28:41.572939
23626	3	1-4705	23625	53	99	3805	58	2012-08-02 05:28:42.403864
23627	3	1-6843	23625	99	38	3177	17	2012-08-02 05:28:42.403864
23625	3	1-6878	23625	38	53	5765	59	2012-08-02 05:28:42.403864
23629	4	1-6061	23628	93	84	1496	45	2012-08-02 05:28:42.403864
23630	4	1-5118	23628	84	40	1611	14	2012-08-02 05:28:42.403864
23631	4	1-6854	23628	40	38	1495	17	2012-08-02 05:28:42.403864
23628	4	1-6878	23628	38	93	3615	59	2012-08-02 05:28:42.403864
23633	2	1-6630	23632	74	42	1912	58	2012-08-02 05:28:43.618762
23632	2	1-6881	23632	42	74	2875	5	2012-08-02 05:28:43.618762
23635	4	1-6007	23634	30	90	482	25	2012-08-02 05:28:44.458432
23636	4	1-6792	23634	90	31	277	47	2012-08-02 05:28:44.458432
23637	4	1-6772	23634	31	76	63	24	2012-08-02 05:28:44.458432
23634	4	1-6883	23634	76	30	1688	46	2012-08-02 05:28:44.458432
23639	5	1-5961	23638	76	92	1305	74	2012-08-02 05:28:44.458432
23640	5	1-6875	23638	92	98	742	12	2012-08-02 05:28:44.458432
23641	5	1-4987	23638	98	93	341	30	2012-08-02 05:28:44.458432
23642	5	1-6016	23638	93	76	145	24	2012-08-02 05:28:44.458432
23638	5	1-6883	23638	76	76	4082	46	2012-08-02 05:28:44.458432
23644	3	1-6796	23643	25	9	5904	29	2012-08-02 05:28:44.878404
23645	3	1-6171	23643	9	76	3200	85	2012-08-02 05:28:44.878404
23643	3	1-6884	23643	76	25	1921	78	2012-08-02 05:28:44.878404
23647	4	1-6472	23646	42	17	2477	30	2012-08-02 05:28:44.878404
23648	4	1-6844	23646	17	12	3115	25	2012-08-02 05:28:44.878404
23649	4	1-4759	23646	12	76	3153	85	2012-08-02 05:28:44.878404
23646	4	1-6884	23646	76	42	3132	78	2012-08-02 05:28:44.878404
23651	4	1-5680	23650	78	13	1475	76	2012-08-02 05:28:46.51375
23652	4	1-6441	23650	13	94	1387	38	2012-08-02 05:28:46.51375
23653	4	1-6669	23650	94	1	2270	48	2012-08-02 05:28:46.51375
23650	4	1-6888	23650	1	78	2078	52	2012-08-02 05:28:46.51375
23655	5	1-5648	23654	7	30	1366	20	2012-08-02 05:28:46.51375
23656	5	1-5714	23654	30	60	2169	54	2012-08-02 05:28:46.51375
23657	5	1-6815	23654	60	94	1736	38	2012-08-02 05:28:46.51375
23658	5	1-6669	23654	94	1	2911	48	2012-08-02 05:28:46.51375
23654	5	1-6888	23654	1	7	2729	52	2012-08-02 05:28:46.51375
23660	4	1-5680	23659	78	57	424	76	2012-08-02 05:28:46.51375
23661	4	1-6228	23659	57	28	338	58	2012-08-02 05:28:46.51375
23662	4	1-6845	23659	28	1	404	48	2012-08-02 05:28:46.51375
23659	4	1-6888	23659	1	78	470	52	2012-08-02 05:28:46.51375
23664	5	1-6545	23663	74	28	1471	52	2012-08-02 05:28:47.331307
23665	5	1-6453	23663	28	17	1872	9	2012-08-02 05:28:47.331307
23666	5	1-5925	23663	17	44	1221	56	2012-08-02 05:28:47.331307
23667	5	1-6880	23663	44	96	1031	6	2012-08-02 05:28:47.331307
23663	5	1-6890	23663	96	74	3732	16	2012-08-02 05:28:47.331307
23669	4	1-6545	23668	74	7	443	52	2012-08-02 05:28:47.331307
23670	4	1-5648	23668	7	85	284	20	2012-08-02 05:28:47.331307
23671	4	1-6337	23668	85	96	245	6	2012-08-02 05:28:47.331307
23668	4	1-6890	23668	96	74	999	16	2012-08-02 05:28:47.331307
23673	4	1-6865	23672	99	5	2722	75	2012-08-02 05:28:47.740198
23674	4	1-4095	23672	5	80	2933	25	2012-08-02 05:28:47.740198
23675	4	1-5679	23672	80	62	1847	76	2012-08-02 05:28:47.740198
23672	4	1-6891	23672	62	99	2426	41	2012-08-02 05:28:47.740198
23677	5	1-5760	23676	40	83	1394	88	2012-08-02 05:28:48.137936
23678	5	1-6755	23676	83	68	247	8	2012-08-02 05:28:48.137936
23679	5	1-3090	23676	68	93	232	76	2012-08-02 05:28:48.137936
23680	5	1-6787	23676	93	57	488	42	2012-08-02 05:28:48.137936
23676	5	1-6892	23676	57	40	3469	53	2012-08-02 05:28:48.137936
23682	5	1-5465	23681	23	36	2965	14	2012-08-02 05:28:48.54716
23683	5	1-6063	23681	36	84	2271	80	2012-08-02 05:28:48.54716
23684	5	1-6459	23681	84	82	1467	11	2012-08-02 05:28:48.54716
23685	5	1-6825	23681	82	94	1993	16	2012-08-02 05:28:48.54716
23681	5	1-6893	23681	94	23	2527	68	2012-08-02 05:28:48.54716
23687	5	1-6462	23686	64	25	1204	89	2012-08-02 05:28:50.195319
23688	5	1-6811	23686	25	63	556	67	2012-08-02 05:28:50.195319
23689	5	1-6866	23686	63	90	417	86	2012-08-02 05:28:50.195319
23690	5	1-6736	23686	90	5	475	23	2012-08-02 05:28:50.195319
23686	5	1-6897	23686	5	64	1216	17	2012-08-02 05:28:50.195319
23692	5	1-6462	23691	64	24	3258	89	2012-08-02 05:28:50.195319
23693	5	1-5043	23691	24	63	1388	67	2012-08-02 05:28:50.195319
23694	5	1-6866	23691	63	90	1061	86	2012-08-02 05:28:50.195319
23695	5	1-6736	23691	90	5	1235	23	2012-08-02 05:28:50.195319
23691	5	1-6897	23691	5	64	3223	17	2012-08-02 05:28:50.195319
23697	5	1-5979	23696	33	25	385	93	2012-08-02 05:28:51.435829
23698	5	1-6852	23696	25	39	564	6	2012-08-02 05:28:51.435829
23699	5	1-6795	23696	39	22	588	96	2012-08-02 05:28:51.435829
23696	5	1-6900	23696	54	33	552	18	2012-08-02 05:28:51.435829
23700	5	1-4673	23696	22	54	415	49	2012-08-02 05:28:51.435829
23702	5	1-4620	23701	9	85	912	20	2012-08-02 05:28:51.435829
23703	5	1-6337	23701	85	39	783	6	2012-08-02 05:28:51.435829
23704	5	1-6795	23701	39	22	945	96	2012-08-02 05:28:51.435829
23705	5	1-4673	23701	22	54	772	49	2012-08-02 05:28:51.435829
23701	5	1-6900	23701	54	9	1190	18	2012-08-02 05:28:51.435829
23707	4	1-6881	23706	42	28	1370	5	2012-08-02 05:28:52.280186
23708	4	1-4888	23706	28	47	998	40	2012-08-02 05:28:52.280186
23709	4	1-6759	23706	47	25	384	9	2012-08-02 05:28:52.280186
23706	4	1-6902	23706	25	42	754	58	2012-08-02 05:28:52.280186
23711	4	1-5795	23710	31	90	1361	54	2012-08-02 05:28:52.713097
23712	4	1-6861	23710	90	20	974	22	2012-08-02 05:28:52.713097
23713	4	1-6257	23710	20	56	1522	14	2012-08-02 05:28:52.713097
23710	4	1-6903	23710	56	31	4422	36	2012-08-02 05:28:52.713097
23715	4	1-4479	23714	34	75	16	26	2012-08-02 05:28:52.713097
23716	4	1-6725	23714	75	23	24	68	2012-08-02 05:28:52.713097
23717	4	1-5465	23714	23	56	29	14	2012-08-02 05:28:52.713097
23714	4	1-6903	23714	56	34	87	36	2012-08-02 05:28:52.713097
23719	4	1-6261	23718	75	62	1389	76	2012-08-02 05:28:53.533357
23720	4	1-6891	23718	62	99	2015	41	2012-08-02 05:28:53.533357
23721	4	1-6865	23718	99	49	2497	75	2012-08-02 05:28:53.533357
23718	4	1-6905	23718	49	75	1904	90	2012-08-02 05:28:53.533357
23723	4	1-6563	23722	50	41	2132	86	2012-08-02 05:28:54.771556
23724	4	1-6766	23722	41	14	1450	29	2012-08-02 05:28:54.771556
23725	4	1-6687	23722	14	52	885	56	2012-08-02 05:28:54.771556
23722	4	1-6908	23722	52	50	3095	13	2012-08-02 05:28:54.771556
23727	4	1-6891	23726	62	77	2811	41	2012-08-02 05:28:57.247656
23728	4	1-6775	23726	77	24	924	8	2012-08-02 05:28:57.247656
23729	4	1-5397	23726	24	13	348	35	2012-08-02 05:28:57.247656
23726	4	1-6914	23726	13	62	3069	76	2012-08-02 05:28:57.247656
23731	5	1-6787	23730	93	30	596	42	2012-08-02 05:28:57.247656
23732	5	1-4766	23730	30	44	116	28	2012-08-02 05:28:57.247656
23733	5	1-6748	23730	44	24	85	8	2012-08-02 05:28:57.247656
23734	5	1-5397	23730	24	13	37	35	2012-08-02 05:28:57.247656
23730	5	1-6914	23730	13	93	375	76	2012-08-02 05:28:57.247656
23736	4	1-6758	23735	16	80	1749	69	2012-08-02 05:28:57.679565
23737	4	1-6786	23735	80	3	1304	5	2012-08-02 05:28:57.679565
23738	4	1-5738	23735	3	92	973	54	2012-08-02 05:28:57.679565
23735	4	1-6915	23735	92	16	2351	72	2012-08-02 05:28:57.679565
23740	3	1-6361	23739	87	75	439	26	2012-08-02 05:28:59.307302
23741	3	1-6725	23739	75	8	563	68	2012-08-02 05:28:59.307302
23739	3	1-6919	23739	8	87	498	55	2012-08-02 05:28:59.307302
23743	5	1-6407	23742	27	44	333	56	2012-08-02 05:28:59.716283
23744	5	1-6880	23742	44	71	288	6	2012-08-02 05:28:59.716283
23745	5	1-6529	23742	71	92	170	12	2012-08-02 05:28:59.716283
23746	5	1-6907	23742	92	38	308	81	2012-08-02 05:28:59.716283
23742	5	1-6920	23742	38	27	440	1	2012-08-02 05:28:59.716283
23748	5	1-5957	23747	98	83	360	60	2012-08-02 05:28:59.716283
23749	5	1-5669	23747	83	92	211	74	2012-08-02 05:28:59.716283
23750	5	1-6875	23747	92	92	134	12	2012-08-02 05:28:59.716283
23751	5	1-6907	23747	92	38	274	81	2012-08-02 05:28:59.716283
23747	5	1-6920	23747	38	98	407	1	2012-08-02 05:28:59.716283
23753	5	1-6394	23752	31	63	83	97	2012-08-02 05:28:59.716283
23754	5	1-6047	23752	63	74	92	11	2012-08-02 05:28:59.716283
23755	5	1-5806	23752	74	89	60	84	2012-08-02 05:28:59.716283
23756	5	1-6850	23752	89	38	69	81	2012-08-02 05:28:59.716283
23752	5	1-6920	23752	38	31	118	1	2012-08-02 05:28:59.716283
23757	1	1-4376	\N	92	92	2386	84	2012-08-02 05:28:59.716283
23759	4	1-5717	23758	8	12	592	64	2012-08-02 05:29:00.138425
23760	4	1-6555	23758	12	88	359	47	2012-08-02 05:29:00.138425
23761	4	1-6760	23758	88	9	512	63	2012-08-02 05:29:00.138425
23758	4	1-6921	23758	9	8	653	82	2012-08-02 05:29:00.138425
23763	3	1-4087	23762	27	52	6526	100	2012-08-02 05:29:00.138425
23764	3	1-5653	23762	52	9	5338	63	2012-08-02 05:29:00.138425
23762	3	1-6921	23762	9	27	7087	82	2012-08-02 05:29:00.138425
23765	1	1-156	\N	51	51	645	47	2012-08-02 05:29:00.138425
23767	2	1-5657	23766	79	65	1270	53	2012-08-02 05:29:00.968231
23766	2	1-6923	23766	65	79	4127	90	2012-08-02 05:29:00.968231
23769	5	1-5536	23768	69	54	381	49	2012-08-02 05:29:01.774837
23770	5	1-6900	23768	54	9	508	18	2012-08-02 05:29:01.774837
23771	5	1-4620	23768	9	71	337	20	2012-08-02 05:29:01.774837
23772	5	1-6509	23768	71	80	451	7	2012-08-02 05:29:01.774837
23768	5	1-6925	23768	80	69	1798	32	2012-08-02 05:29:01.774837
23774	4	1-2766	23773	59	74	1964	16	2012-08-02 05:29:01.774837
23775	4	1-6545	23773	74	2	854	52	2012-08-02 05:29:01.774837
23776	4	1-5423	23773	2	80	609	7	2012-08-02 05:29:01.774837
23773	4	1-6925	23773	80	59	2766	32	2012-08-02 05:29:01.774837
23778	5	1-4604	23777	32	41	1015	5	2012-08-02 05:29:01.774837
23779	5	1-5788	23777	41	67	493	95	2012-08-02 05:29:01.774837
23780	5	1-6715	23777	67	2	380	52	2012-08-02 05:29:01.774837
23781	5	1-5423	23777	2	80	287	7	2012-08-02 05:29:01.774837
23777	5	1-6925	23777	80	32	1376	32	2012-08-02 05:29:01.774837
23783	4	1-5869	23782	18	31	1322	3	2012-08-02 05:29:01.774837
23784	4	1-5565	23782	31	64	1142	37	2012-08-02 05:29:01.774837
23785	4	1-6762	23782	64	80	507	7	2012-08-02 05:29:01.774837
23782	4	1-6925	23782	80	18	2501	32	2012-08-02 05:29:01.774837
23787	2	1-5749	23786	63	75	4009	92	2012-08-02 05:29:02.208404
23786	2	1-6926	23786	75	63	5183	36	2012-08-02 05:29:02.208404
23789	2	1-6328	23788	100	75	1418	92	2012-08-02 05:29:02.208404
23788	2	1-6926	23788	75	100	2374	36	2012-08-02 05:29:02.208404
23791	5	1-6285	23790	83	27	1576	63	2012-08-02 05:29:03.014937
23792	5	1-6909	23790	27	91	1075	50	2012-08-02 05:29:03.014937
23793	5	1-6616	23790	91	36	588	9	2012-08-02 05:29:03.014937
23794	5	1-6831	23790	36	49	778	29	2012-08-02 05:29:03.014937
23790	5	1-6928	23790	49	83	2820	36	2012-08-02 05:29:03.014937
23796	4	1-2226	23795	6	84	2003	80	2012-08-02 05:29:03.014937
23797	4	1-6459	23795	84	12	1230	11	2012-08-02 05:29:03.014937
23798	4	1-5547	23795	12	49	765	29	2012-08-02 05:29:03.014937
23795	4	1-6928	23795	49	6	3059	36	2012-08-02 05:29:03.014937
23800	5	1-2226	23799	6	64	363	80	2012-08-02 05:29:03.014937
23801	5	1-6771	23799	64	91	157	50	2012-08-02 05:29:03.014937
23802	5	1-6616	23799	91	36	95	9	2012-08-02 05:29:03.014937
23803	5	1-6831	23799	36	49	138	29	2012-08-02 05:29:03.014937
23799	5	1-6928	23799	49	6	554	36	2012-08-02 05:29:03.014937
23805	5	1-4057	23804	45	90	1014	51	2012-08-02 05:29:03.014937
23806	5	1-6522	23804	90	91	230	50	2012-08-02 05:29:03.014937
23807	5	1-6616	23804	91	36	152	9	2012-08-02 05:29:03.014937
23808	5	1-6831	23804	36	49	242	29	2012-08-02 05:29:03.014937
23804	5	1-6928	23804	49	45	1060	36	2012-08-02 05:29:03.014937
23810	5	1-5185	23809	39	44	3000	28	2012-08-02 05:29:04.266931
23811	5	1-6748	23809	44	24	2858	8	2012-08-02 05:29:04.266931
23812	5	1-5397	23809	24	93	1621	35	2012-08-02 05:29:04.266931
23813	5	1-6695	23809	93	74	6491	83	2012-08-02 05:29:04.266931
23809	5	1-6931	23809	74	39	5155	86	2012-08-02 05:29:04.266931
23815	5	1-6491	23814	23	94	1899	16	2012-08-02 05:29:05.117609
23816	5	1-6893	23814	94	71	2438	68	2012-08-02 05:29:05.117609
23817	5	1-6002	23814	71	74	2722	83	2012-08-02 05:29:05.117609
23818	5	1-6931	23814	74	76	2206	86	2012-08-02 05:29:05.117609
23814	5	1-6933	23814	76	23	2052	43	2012-08-02 05:29:05.117609
23820	4	1-6639	23819	90	49	186	29	2012-08-02 05:29:06.747953
23821	4	1-6928	23819	49	59	737	36	2012-08-02 05:29:06.747953
23822	4	1-1879	23819	59	54	286	44	2012-08-02 05:29:06.747953
23819	4	1-6937	23819	54	90	156	34	2012-08-02 05:29:06.747953
23824	4	1-6176	23823	89	52	212	22	2012-08-02 05:29:08.32781
23825	4	1-6862	23823	52	98	444	12	2012-08-02 05:29:08.32781
23826	4	1-4987	23823	98	34	229	30	2012-08-02 05:29:08.32781
23823	4	1-6941	23823	34	89	780	25	2012-08-02 05:29:08.32781
23828	3	1-5608	23827	85	42	1509	78	2012-08-02 05:29:08.32781
23829	3	1-6472	23827	42	34	1294	30	2012-08-02 05:29:08.32781
23827	3	1-6941	23827	34	85	4634	25	2012-08-02 05:29:08.32781
23831	4	1-4895	23830	9	77	1647	50	2012-08-02 05:29:11.521102
23832	4	1-6643	23830	77	9	1505	73	2012-08-02 05:29:11.521102
23833	4	1-6408	23830	9	86	1071	70	2012-08-02 05:29:11.521102
23830	4	1-6949	23830	86	9	1831	65	2012-08-02 05:29:11.521102
23835	5	1-5706	23834	89	16	1890	72	2012-08-02 05:29:11.521102
23836	5	1-6758	23834	16	96	1628	69	2012-08-02 05:29:11.521102
23837	5	1-6368	23834	96	77	1872	39	2012-08-02 05:29:11.521102
23838	5	1-6879	23834	77	86	2076	70	2012-08-02 05:29:11.521102
23834	5	1-6949	23834	86	89	3816	65	2012-08-02 05:29:11.521102
23840	5	1-4031	23839	12	50	567	39	2012-08-02 05:29:11.521102
23841	5	1-6741	23839	50	100	2145	88	2012-08-02 05:29:11.521102
23842	5	1-6837	23839	100	57	825	75	2012-08-02 05:29:11.521102
23843	5	1-3640	23839	57	86	1024	70	2012-08-02 05:29:11.521102
23839	5	1-6949	23839	86	12	2030	65	2012-08-02 05:29:11.521102
23845	4	1-6666	23844	1	35	2511	36	2012-08-02 05:29:13.479262
23846	4	1-4032	23844	35	76	1680	88	2012-08-02 05:29:13.479262
23847	4	1-6489	23844	76	70	805	67	2012-08-02 05:29:13.479262
23844	4	1-6954	23844	70	1	1869	44	2012-08-02 05:29:13.479262
23849	5	1-6555	23848	12	88	351	47	2012-08-02 05:29:13.877246
23850	5	1-6760	23848	88	96	666	63	2012-08-02 05:29:13.877246
23851	5	1-2148	23848	96	6	141	55	2012-08-02 05:29:13.877246
23852	5	1-5374	23848	6	47	181	34	2012-08-02 05:29:13.877246
23848	5	1-6955	23848	47	12	436	64	2012-08-02 05:29:13.877246
23853	1	1-2148	\N	96	96	1573	55	2012-08-02 05:29:13.877246
23855	4	1-6337	23854	85	39	1071	6	2012-08-02 05:29:15.048262
23856	4	1-6795	23854	39	10	1147	96	2012-08-02 05:29:15.048262
23857	4	1-1141	23854	10	52	1152	14	2012-08-02 05:29:15.048262
23854	4	1-6958	23854	52	85	1407	20	2012-08-02 05:29:15.048262
23859	3	1-6337	23858	85	49	4	6	2012-08-02 05:29:15.048262
23860	3	1-2339	23858	49	52	4	14	2012-08-02 05:29:15.048262
23858	3	1-6958	23858	52	85	5	20	2012-08-02 05:29:15.048262
23862	4	1-4683	23861	35	94	810	16	2012-08-02 05:29:15.048262
23863	4	1-6893	23861	94	23	1030	68	2012-08-02 05:29:15.048262
23864	4	1-5465	23861	23	52	1212	14	2012-08-02 05:29:15.048262
23861	4	1-6958	23861	52	35	1743	20	2012-08-02 05:29:15.048262
23865	1	1-1141	\N	10	10	2351	14	2012-08-02 05:29:15.048262
23867	4	1-5151	23866	50	36	1503	14	2012-08-02 05:29:15.866469
23868	4	1-6063	23866	36	27	1105	80	2012-08-02 05:29:15.866469
23869	4	1-6839	23866	27	82	609	67	2012-08-02 05:29:15.866469
23866	4	1-6960	23866	82	50	542	19	2012-08-02 05:29:15.866469
23871	5	1-5008	23870	40	44	746	56	2012-08-02 05:29:15.866469
23872	5	1-6880	23870	44	26	763	6	2012-08-02 05:29:15.866469
23873	5	1-6821	23870	26	100	723	77	2012-08-02 05:29:15.866469
23874	5	1-5352	23870	100	82	731	67	2012-08-02 05:29:15.866469
23870	5	1-6960	23870	82	40	704	19	2012-08-02 05:29:15.866469
23876	4	1-6959	23875	4	96	2875	69	2012-08-02 05:29:16.673434
23877	4	1-6368	23875	96	77	3229	39	2012-08-02 05:29:16.673434
23878	4	1-6879	23875	77	13	3496	70	2012-08-02 05:29:16.673434
23875	4	1-6962	23875	13	4	5651	71	2012-08-02 05:29:16.673434
23880	4	1-6853	23879	81	40	71	55	2012-08-02 05:29:16.673434
23881	4	1-2432	23879	40	9	16	35	2012-08-02 05:29:16.673434
23882	4	1-6449	23879	9	13	28	70	2012-08-02 05:29:16.673434
23879	4	1-6962	23879	13	81	46	71	2012-08-02 05:29:16.673434
23884	4	1-6381	23883	83	8	2857	43	2012-08-02 05:29:16.673434
23885	4	1-5928	23883	8	96	2124	75	2012-08-02 05:29:16.673434
23886	4	1-3889	23883	96	13	1895	70	2012-08-02 05:29:16.673434
23883	4	1-6962	23883	13	83	3230	71	2012-08-02 05:29:16.673434
23887	1	1-5928	\N	8	8	400	75	2012-08-02 05:29:16.673434
23889	4	1-6026	23888	1	40	3465	14	2012-08-02 05:29:17.083013
23890	4	1-6854	23888	40	64	2698	17	2012-08-02 05:29:17.083013
23891	4	1-6462	23888	64	39	2015	89	2012-08-02 05:29:17.083013
23888	4	1-6963	23888	39	1	2624	97	2012-08-02 05:29:17.083013
23893	5	1-6784	23892	94	38	1229	19	2012-08-02 05:29:17.083013
23894	5	1-6046	23892	38	10	1134	8	2012-08-02 05:29:17.083013
23895	5	1-5963	23892	10	100	975	9	2012-08-02 05:29:17.083013
23896	5	1-6602	23892	100	39	1752	89	2012-08-02 05:29:17.083013
23892	5	1-6963	23892	39	94	3018	97	2012-08-02 05:29:17.083013
23898	5	1-6885	23897	70	82	305	67	2012-08-02 05:29:17.083013
23899	5	1-6960	23897	82	76	292	19	2012-08-02 05:29:17.083013
23900	5	1-5517	23897	76	100	140	9	2012-08-02 05:29:17.083013
23901	5	1-6602	23897	100	39	256	89	2012-08-02 05:29:17.083013
23897	5	1-6963	23897	39	70	448	97	2012-08-02 05:29:17.083013
23903	2	1-1879	23902	59	92	738	44	2012-08-02 05:29:18.684963
23902	2	1-6967	23902	92	59	4443	36	2012-08-02 05:29:18.684963
23905	4	1-6500	23904	7	42	2450	80	2012-08-02 05:29:19.126962
23906	4	1-1598	23904	42	57	260	42	2012-08-02 05:29:19.126962
23907	4	1-6892	23904	57	47	1118	53	2012-08-02 05:29:19.126962
23904	4	1-6968	23904	47	7	6314	48	2012-08-02 05:29:19.126962
23909	4	1-3866	23908	18	79	736	77	2012-08-02 05:29:20.308992
23910	4	1-6856	23908	79	65	626	22	2012-08-02 05:29:20.308992
23911	4	1-6947	23908	65	81	2716	40	2012-08-02 05:29:20.308992
23908	4	1-6971	23908	81	18	2633	35	2012-08-02 05:29:20.308992
23913	4	1-6449	23912	9	86	1221	70	2012-08-02 05:29:20.308992
23914	4	1-6949	23912	86	74	1769	65	2012-08-02 05:29:20.308992
23915	4	1-6525	23912	74	81	833	40	2012-08-02 05:29:20.308992
23912	4	1-6971	23912	81	9	859	35	2012-08-02 05:29:20.308992
23917	5	1-5333	23916	47	54	1597	49	2012-08-02 05:29:20.308992
23918	5	1-6900	23916	54	60	2003	18	2012-08-02 05:29:20.308992
23919	5	1-4537	23916	60	65	805	22	2012-08-02 05:29:20.308992
23920	5	1-6947	23916	65	81	3904	40	2012-08-02 05:29:20.308992
23916	5	1-6971	23916	81	47	4231	35	2012-08-02 05:29:20.308992
23922	4	1-5527	23921	81	60	1530	54	2012-08-02 05:29:20.71803
23923	4	1-6815	23921	60	24	999	38	2012-08-02 05:29:20.71803
23924	4	1-2979	23921	24	47	412	89	2012-08-02 05:29:20.71803
23921	4	1-6972	23921	47	81	3831	16	2012-08-02 05:29:20.71803
23926	5	1-1785	23925	96	92	1519	74	2012-08-02 05:29:23.844492
23927	5	1-6875	23925	92	92	792	12	2012-08-02 05:29:23.844492
23928	5	1-6907	23925	92	38	1323	81	2012-08-02 05:29:23.844492
23929	5	1-6920	23925	38	12	1320	1	2012-08-02 05:29:23.844492
23925	5	1-6980	23925	12	96	3224	35	2012-08-02 05:29:23.844492
23931	4	1-6003	23930	70	45	3510	76	2012-08-02 05:29:23.844492
23932	4	1-6871	23930	45	38	1870	81	2012-08-02 05:29:23.844492
23933	4	1-6920	23930	38	12	2025	1	2012-08-02 05:29:23.844492
23930	4	1-6980	23930	12	70	5367	35	2012-08-02 05:29:23.844492
23935	5	1-3538	23934	72	47	114	34	2012-08-02 05:29:23.844492
23936	5	1-6955	23934	47	33	172	64	2012-08-02 05:29:23.844492
23937	5	1-6196	23934	33	38	66	81	2012-08-02 05:29:23.844492
23938	5	1-6920	23934	38	12	71	1	2012-08-02 05:29:23.844492
23934	5	1-6980	23934	12	72	189	35	2012-08-02 05:29:23.844492
23939	1	1-1785	\N	96	96	1897	74	2012-08-02 05:29:23.844492
23941	4	1-6954	23940	70	99	6657	44	2012-08-02 05:29:25.845038
23942	4	1-5854	23940	99	36	2669	76	2012-08-02 05:29:25.845038
23943	4	1-6677	23940	36	10	4847	2	2012-08-02 05:29:25.845038
23940	4	1-6985	23940	10	70	3157	67	2012-08-02 05:29:25.845038
23945	4	1-6954	23944	70	1	130	44	2012-08-02 05:29:25.845038
23946	4	1-6666	23944	1	70	158	36	2012-08-02 05:29:25.845038
23947	4	1-4825	23944	70	10	89	2	2012-08-02 05:29:25.845038
23944	4	1-6985	23944	10	70	58	67	2012-08-02 05:29:25.845038
23949	4	1-6951	23948	46	14	1047	55	2012-08-02 05:29:27.391463
23950	4	1-6750	23948	14	98	3696	53	2012-08-02 05:29:27.391463
23951	4	1-1901	23948	98	62	1055	2	2012-08-02 05:29:27.391463
23948	4	1-6989	23948	62	46	1460	54	2012-08-02 05:29:27.391463
23953	5	1-6951	23952	46	46	4524	55	2012-08-02 05:29:27.391463
23954	5	1-6278	23952	46	72	4503	78	2012-08-02 05:29:27.391463
23955	5	1-6164	23952	72	29	2703	52	2012-08-02 05:29:27.391463
23956	5	1-6167	23952	29	62	3644	2	2012-08-02 05:29:27.391463
23952	5	1-6989	23952	62	46	5318	54	2012-08-02 05:29:27.391463
23958	5	1-6370	23957	82	35	5508	95	2012-08-02 05:29:28.573731
23959	5	1-6259	23957	35	1	2598	17	2012-08-02 05:29:28.573731
23960	5	1-6295	23957	1	39	1787	94	2012-08-02 05:29:28.573731
23961	5	1-6943	23957	39	100	2878	5	2012-08-02 05:29:28.573731
23957	5	1-6992	23957	100	82	4492	71	2012-08-02 05:29:28.573731
23963	5	1-6348	23962	56	35	1727	95	2012-08-02 05:29:28.573731
23964	5	1-6259	23962	35	1	819	17	2012-08-02 05:29:28.573731
23965	5	1-6295	23962	1	39	566	94	2012-08-02 05:29:28.573731
23966	5	1-6943	23962	39	100	915	5	2012-08-02 05:29:28.573731
23962	5	1-6992	23962	100	56	1435	71	2012-08-02 05:29:28.573731
23968	4	1-6820	23967	32	17	485	9	2012-08-02 05:29:29.347239
23969	4	1-5925	23967	17	52	310	56	2012-08-02 05:29:29.347239
23970	4	1-6908	23967	52	16	1013	13	2012-08-02 05:29:29.347239
23967	4	1-6994	23967	16	32	861	7	2012-08-02 05:29:29.347239
23972	5	1-610	23971	49	52	921	14	2012-08-02 05:29:29.745263
23973	5	1-6958	23971	52	10	1097	20	2012-08-02 05:29:29.745263
23974	5	1-6936	23971	10	63	1249	55	2012-08-02 05:29:29.745263
23975	5	1-6613	23971	63	71	1242	28	2012-08-02 05:29:29.745263
23971	5	1-6995	23971	71	49	757	30	2012-08-02 05:29:29.745263
23977	5	1-6941	23976	34	11	609	25	2012-08-02 05:29:29.745263
23978	5	1-2380	23976	11	46	210	54	2012-08-02 05:29:29.745263
23979	5	1-6951	23976	46	63	191	55	2012-08-02 05:29:29.745263
23980	5	1-6613	23976	63	71	233	28	2012-08-02 05:29:29.745263
23976	5	1-6995	23976	71	34	174	30	2012-08-02 05:29:29.745263
23982	4	1-5118	23981	84	52	304	14	2012-08-02 05:29:30.551949
23983	4	1-6958	23981	52	30	463	20	2012-08-02 05:29:30.551949
23984	4	1-5714	23981	30	3	1034	54	2012-08-02 05:29:30.551949
23981	4	1-6997	23981	3	84	244	45	2012-08-02 05:29:30.551949
23986	4	1-6164	23985	72	50	85	52	2012-08-02 05:29:30.939009
23987	4	1-6558	23985	50	27	114	82	2012-08-02 05:29:30.939009
23988	4	1-4087	23985	27	1	102	100	2012-08-02 05:29:30.939009
23985	4	1-6998	23985	1	72	148	78	2012-08-02 05:29:30.939009
23990	3	1-6164	23989	72	91	764	52	2012-08-02 05:29:30.939009
23991	3	1-6547	23989	91	1	679	100	2012-08-02 05:29:30.939009
23989	3	1-6998	23989	1	72	1143	78	2012-08-02 05:29:30.939009
23993	5	1-6528	23992	41	99	2420	58	2012-08-02 05:29:31.347676
23994	5	1-6843	23992	99	1	2355	17	2012-08-02 05:29:31.347676
23995	5	1-6295	23992	1	39	1643	94	2012-08-02 05:29:31.347676
23996	5	1-6943	23992	39	70	2683	5	2012-08-02 05:29:31.347676
23992	5	1-6999	23992	70	41	1486	39	2012-08-02 05:29:31.347676
23998	4	1-6741	23997	50	100	2036	88	2012-08-02 05:29:31.347676
23999	4	1-6837	23997	100	44	756	75	2012-08-02 05:29:31.347676
24000	4	1-3500	23997	44	70	820	5	2012-08-02 05:29:31.347676
23997	4	1-6999	23997	70	50	558	39	2012-08-02 05:29:31.347676
24002	5	1-6736	24001	90	96	398	23	2012-08-02 05:29:32.134027
24003	5	1-6098	24001	96	46	621	54	2012-08-02 05:29:32.134027
24004	5	1-6951	24001	46	6	582	55	2012-08-02 05:29:32.134027
24005	5	1-5374	24001	6	45	741	34	2012-08-02 05:29:32.134027
24001	5	1-7001	24001	45	90	339	86	2012-08-02 05:29:32.134027
24007	4	1-6446	24006	4	16	2111	79	2012-08-02 05:29:32.520845
24008	4	1-6322	24006	16	1	1183	10	2012-08-02 05:29:32.520845
24009	4	1-6830	24006	1	83	778	28	2012-08-02 05:29:32.520845
24006	4	1-7002	24006	83	4	3880	7	2012-08-02 05:29:32.520845
24011	5	1-6820	24010	32	31	615	9	2012-08-02 05:29:32.520845
24012	5	1-4280	24010	31	10	227	20	2012-08-02 05:29:32.520845
24013	5	1-6936	24010	10	63	239	55	2012-08-02 05:29:32.520845
24014	5	1-6613	24010	63	83	219	28	2012-08-02 05:29:32.520845
24010	5	1-7002	24010	83	32	1199	7	2012-08-02 05:29:32.520845
24016	3	1-6446	24015	4	79	1464	79	2012-08-02 05:29:32.520845
24017	3	1-4819	24015	79	83	430	28	2012-08-02 05:29:32.520845
24015	3	1-7002	24015	83	4	2403	7	2012-08-02 05:29:32.520845
24019	5	1-6895	24018	45	9	90	18	2012-08-02 05:29:32.520845
24020	5	1-4620	24018	9	10	60	20	2012-08-02 05:29:32.520845
24021	5	1-6936	24018	10	63	69	55	2012-08-02 05:29:32.520845
24022	5	1-6613	24018	63	83	70	28	2012-08-02 05:29:32.520845
24018	5	1-7002	24018	83	45	421	7	2012-08-02 05:29:32.520845
24024	5	1-6693	24023	19	22	965	100	2012-08-02 05:29:32.918892
24025	5	1-3959	24023	22	78	374	13	2012-08-02 05:29:32.918892
24026	5	1-6740	24023	78	35	145	60	2012-08-02 05:29:32.918892
24027	5	1-5359	24023	35	48	123	90	2012-08-02 05:29:32.918892
24023	5	1-7003	24023	48	19	1369	31	2012-08-02 05:29:32.918892
24029	4	1-6607	24028	70	25	1375	93	2012-08-02 05:29:32.918892
24030	4	1-4812	24028	25	52	380	30	2012-08-02 05:29:32.918892
24031	4	1-1308	24028	52	48	182	90	2012-08-02 05:29:32.918892
24028	4	1-7003	24028	48	70	2095	31	2012-08-02 05:29:32.918892
24033	5	1-6607	24032	70	8	63	93	2012-08-02 05:29:32.918892
24034	5	1-3282	24032	8	100	20	88	2012-08-02 05:29:32.918892
24035	5	1-6837	24032	100	49	8	75	2012-08-02 05:29:32.918892
24036	5	1-6905	24032	49	48	6	90	2012-08-02 05:29:32.918892
24032	5	1-7003	24032	48	70	79	31	2012-08-02 05:29:32.918892
24038	4	1-6698	24037	23	1	1825	15	2012-08-02 05:29:32.918892
24039	4	1-6761	24037	1	42	311	38	2012-08-02 05:29:32.918892
24040	4	1-4896	24037	42	48	153	90	2012-08-02 05:29:32.918892
24037	4	1-7003	24037	48	23	2192	31	2012-08-02 05:29:32.918892
24042	4	1-6528	24041	41	28	1473	58	2012-08-02 05:29:34.156867
24043	4	1-6845	24041	28	7	1731	48	2012-08-02 05:29:34.156867
24044	4	1-6500	24041	7	24	1018	80	2012-08-02 05:29:34.156867
24041	4	1-7006	24041	24	41	786	39	2012-08-02 05:29:34.156867
24046	5	1-6347	24045	3	57	3859	98	2012-08-02 05:29:34.532292
24047	5	1-6911	24045	57	27	3629	80	2012-08-02 05:29:34.532292
24048	5	1-6839	24045	27	82	2140	67	2012-08-02 05:29:34.532292
24049	5	1-6960	24045	82	99	2039	19	2012-08-02 05:29:34.532292
24045	5	1-7007	24045	99	3	3990	43	2012-08-02 05:29:34.532292
24051	5	1-5443	24050	98	60	763	49	2012-08-02 05:29:34.532292
24052	5	1-6303	24050	60	76	1330	88	2012-08-02 05:29:34.532292
24053	5	1-6489	24050	76	82	662	67	2012-08-02 05:29:34.532292
24054	5	1-6960	24050	82	99	653	19	2012-08-02 05:29:34.532292
24050	5	1-7007	24050	99	98	1321	43	2012-08-02 05:29:34.532292
24056	5	1-6958	24055	52	10	2225	20	2012-08-02 05:29:34.94122
24057	5	1-6936	24055	10	63	2496	55	2012-08-02 05:29:34.94122
24058	5	1-6613	24055	63	71	2444	28	2012-08-02 05:29:34.94122
24059	5	1-6995	24055	71	76	1468	30	2012-08-02 05:29:34.94122
24055	5	1-7008	24055	76	52	1898	14	2012-08-02 05:29:34.94122
24061	3	1-5846	24060	94	49	637	75	2012-08-02 05:29:35.759093
24062	3	1-6905	24060	49	29	452	90	2012-08-02 05:29:35.759093
24060	3	1-7010	24060	29	94	1003	61	2012-08-02 05:29:35.759093
24064	2	1-6440	24063	36	40	625	70	2012-08-02 05:29:38.521063
24063	2	1-7017	24063	40	36	8703	78	2012-08-02 05:29:38.521063
24066	4	1-6046	24065	38	12	4397	8	2012-08-02 05:29:40.089552
24067	4	1-5933	24065	12	70	3844	91	2012-08-02 05:29:40.089552
24068	4	1-6635	24065	70	53	3629	41	2012-08-02 05:29:40.089552
24065	4	1-7021	24065	53	38	5229	19	2012-08-02 05:29:40.089552
24070	5	1-2489	24069	91	1	133	100	2012-08-02 05:29:40.089552
24071	5	1-6998	24069	1	76	248	78	2012-08-02 05:29:40.089552
24072	5	1-6140	24069	76	70	127	91	2012-08-02 05:29:40.089552
24073	5	1-6635	24069	70	53	136	41	2012-08-02 05:29:40.089552
24069	5	1-7021	24069	53	91	221	19	2012-08-02 05:29:40.089552
24075	4	1-6758	24074	16	77	2218	69	2012-08-02 05:29:40.487537
24076	4	1-6287	24074	77	38	803	81	2012-08-02 05:29:40.487537
24077	4	1-6920	24074	38	12	872	1	2012-08-02 05:29:40.487537
24074	4	1-7022	24074	12	16	3651	72	2012-08-02 05:29:40.487537
24079	5	1-4955	24078	67	50	911	19	2012-08-02 05:29:40.487537
24080	5	1-5151	24078	50	45	2036	14	2012-08-02 05:29:40.487537
24081	5	1-6701	24078	45	38	680	81	2012-08-02 05:29:40.487537
24082	5	1-6920	24078	38	12	859	1	2012-08-02 05:29:40.487537
24078	5	1-7022	24078	12	67	4188	72	2012-08-02 05:29:40.487537
24084	5	1-6966	24083	18	68	366	8	2012-08-02 05:29:40.487537
24085	5	1-3090	24083	68	45	266	76	2012-08-02 05:29:40.487537
24086	5	1-6871	24083	45	38	166	81	2012-08-02 05:29:40.487537
24087	5	1-6920	24083	38	12	210	1	2012-08-02 05:29:40.487537
24083	5	1-7022	24083	12	18	1024	72	2012-08-02 05:29:40.487537
24089	5	1-6960	24088	82	99	2514	19	2012-08-02 05:29:44.143935
24090	5	1-7007	24088	99	23	3606	43	2012-08-02 05:29:44.143935
24091	5	1-6491	24088	23	94	2480	16	2012-08-02 05:29:44.143935
24092	5	1-6893	24088	94	71	2366	68	2012-08-02 05:29:44.143935
24088	5	1-7031	24088	71	82	3600	67	2012-08-02 05:29:44.143935
24094	4	1-7006	24093	24	77	3730	39	2012-08-02 05:29:44.54131
24095	4	1-6879	24093	77	9	3938	70	2012-08-02 05:29:44.54131
24096	4	1-6422	24093	9	35	1870	45	2012-08-02 05:29:44.54131
24093	4	1-7032	24093	35	24	4947	80	2012-08-02 05:29:44.54131
24098	5	1-6990	24097	38	35	1298	45	2012-08-02 05:29:45.347781
24099	5	1-7032	24097	35	24	2903	80	2012-08-02 05:29:45.347781
24100	5	1-7006	24097	24	82	1850	39	2012-08-02 05:29:45.347781
24101	5	1-6785	24097	82	3	5172	10	2012-08-02 05:29:45.347781
24097	5	1-7034	24097	3	38	1437	98	2012-08-02 05:29:45.347781
24103	4	1-6611	24102	95	57	391	42	2012-08-02 05:29:45.756622
24104	4	1-6892	24102	57	20	2532	53	2012-08-02 05:29:45.756622
24105	4	1-4921	24102	20	13	616	74	2012-08-02 05:29:45.756622
24102	4	1-7035	24102	13	95	445	39	2012-08-02 05:29:45.756622
24107	3	1-6876	24106	1	27	2123	61	2012-08-02 05:29:46.916454
24108	3	1-6241	24106	27	17	1979	100	2012-08-02 05:29:46.916454
24106	3	1-7038	24106	17	1	2100	8	2012-08-02 05:29:46.916454
24110	4	1-6876	24109	1	79	5458	61	2012-08-02 05:29:46.916454
24111	4	1-6424	24109	79	19	4208	31	2012-08-02 05:29:46.916454
24112	4	1-6693	24109	19	17	3654	100	2012-08-02 05:29:46.916454
24109	4	1-7038	24109	17	1	4575	8	2012-08-02 05:29:46.916454
24114	5	1-6876	24113	1	5	218	61	2012-08-02 05:29:47.325487
24115	5	1-4498	24113	5	10	101	20	2012-08-02 05:29:47.325487
24113	5	1-7039	24113	47	1	179	8	2012-08-02 05:29:47.325487
24116	5	1-6936	24113	10	6	141	55	2012-08-02 05:29:47.325487
24117	5	1-5374	24113	6	47	174	34	2012-08-02 05:29:47.325487
24119	2	1-5751	24118	48	42	1818	66	2012-08-02 05:29:47.712102
24118	2	1-7040	24118	42	48	7708	87	2012-08-02 05:29:47.712102
24121	5	1-3762	24120	50	52	638	56	2012-08-02 05:29:48.099244
24122	5	1-6908	24120	52	50	2163	13	2012-08-02 05:29:48.099244
24123	5	1-6563	24120	50	90	1599	86	2012-08-02 05:29:48.099244
24124	5	1-6736	24120	90	33	1564	23	2012-08-02 05:29:48.099244
24120	5	1-7041	24120	33	50	2152	8	2012-08-02 05:29:48.099244
24126	4	1-6368	24125	96	95	1705	39	2012-08-02 05:29:50.076402
24127	4	1-6611	24125	95	18	1566	42	2012-08-02 05:29:50.076402
24128	4	1-5929	24125	18	54	1278	31	2012-08-02 05:29:50.076402
24125	4	1-7046	24125	54	96	1462	69	2012-08-02 05:29:50.076402
24130	4	1-6435	24129	22	95	3431	39	2012-08-02 05:29:50.076402
24131	4	1-6611	24129	95	18	3458	42	2012-08-02 05:29:50.076402
24132	4	1-5929	24129	18	54	3099	31	2012-08-02 05:29:50.076402
24129	4	1-7046	24129	54	22	3891	69	2012-08-02 05:29:50.076402
24134	4	1-6249	24133	25	37	1575	77	2012-08-02 05:29:50.076402
24135	4	1-6516	24133	37	48	1285	28	2012-08-02 05:29:50.076402
24136	4	1-6942	24133	48	54	1069	31	2012-08-02 05:29:50.076402
24133	4	1-7046	24133	54	25	1356	69	2012-08-02 05:29:50.076402
24138	3	1-6720	24137	63	65	1881	42	2012-08-02 05:29:51.270069
24139	3	1-6418	24137	65	36	808	36	2012-08-02 05:29:51.270069
24137	3	1-7049	24137	36	63	5215	64	2012-08-02 05:29:51.270069
24141	3	1-5919	24140	55	83	912	86	2012-08-02 05:29:52.065187
24142	3	1-1896	24140	83	35	305	40	2012-08-02 05:29:52.065187
24140	3	1-7051	24140	35	55	3961	93	2012-08-02 05:29:52.065187
24144	4	1-3329	24143	89	89	742	25	2012-08-02 05:29:52.065187
24145	4	1-6176	24143	89	65	132	22	2012-08-02 05:29:52.065187
24146	4	1-6947	24143	65	35	322	40	2012-08-02 05:29:52.065187
24143	4	1-7051	24143	35	89	4874	93	2012-08-02 05:29:52.065187
24148	5	1-6071	24147	98	1	1664	100	2012-08-02 05:29:53.247563
24149	5	1-6998	24147	1	72	2700	78	2012-08-02 05:29:53.247563
24150	5	1-5722	24147	72	93	1844	76	2012-08-02 05:29:53.247563
24151	5	1-6787	24147	93	87	3497	42	2012-08-02 05:29:53.247563
24147	5	1-7054	24147	87	98	4922	7	2012-08-02 05:29:53.247563
24153	5	1-6947	24152	65	47	405	40	2012-08-02 05:29:54.031862
24154	5	1-6759	24152	47	29	154	9	2012-08-02 05:29:54.031862
24155	5	1-4607	24152	29	87	71	25	2012-08-02 05:29:54.031862
24156	5	1-7015	24152	87	34	57	75	2012-08-02 05:29:54.031862
24152	5	1-7056	24152	34	65	63	22	2012-08-02 05:29:54.031862
24158	4	1-6662	24157	59	9	580	35	2012-08-02 05:29:54.440796
24159	4	1-6449	24157	9	28	1083	70	2012-08-02 05:29:54.440796
24160	4	1-6817	24157	28	32	575	97	2012-08-02 05:29:54.440796
24157	4	1-7057	24157	32	59	1177	62	2012-08-02 05:29:54.440796
24162	5	1-6958	24161	52	85	2572	20	2012-08-02 05:29:55.634323
24163	5	1-6337	24161	85	26	2072	6	2012-08-02 05:29:55.634323
24164	5	1-6821	24161	26	41	1705	77	2012-08-02 05:29:55.634323
24165	5	1-6956	24161	41	54	3138	59	2012-08-02 05:29:55.634323
24161	5	1-7060	24161	54	52	1988	14	2012-08-02 05:29:55.634323
24167	5	1-6929	24166	40	48	7813	85	2012-08-02 05:29:56.429727
24168	5	1-6515	24166	48	100	1842	36	2012-08-02 05:29:56.429727
24169	5	1-5084	24166	100	44	486	56	2012-08-02 05:29:56.429727
24170	5	1-6880	24166	44	91	438	6	2012-08-02 05:29:56.429727
24166	5	1-7062	24166	91	40	5571	17	2012-08-02 05:29:56.429727
24172	4	1-5936	24171	21	10	437	93	2012-08-02 05:29:56.429727
24173	4	1-4219	24171	10	85	147	20	2012-08-02 05:29:56.429727
24174	4	1-6337	24171	85	91	139	6	2012-08-02 05:29:56.429727
24171	4	1-7062	24171	91	21	2063	17	2012-08-02 05:29:56.429727
24176	3	1-7041	24175	33	68	1078	8	2012-08-02 05:29:57.998355
24177	3	1-3090	24175	68	65	213	76	2012-08-02 05:29:57.998355
24175	3	1-7066	24175	65	33	3238	23	2012-08-02 05:29:57.998355
24179	5	1-6726	24178	75	94	548	16	2012-08-02 05:29:58.374313
24180	5	1-6893	24178	94	89	462	68	2012-08-02 05:29:58.374313
24181	5	1-6104	24178	89	56	180	32	2012-08-02 05:29:58.374313
24182	5	1-5899	24178	56	51	134	2	2012-08-02 05:29:58.374313
24178	5	1-7067	24178	51	75	1883	15	2012-08-02 05:29:58.374313
24184	5	1-6362	24183	86	47	611	34	2012-08-02 05:29:58.374313
24185	5	1-7039	24183	47	39	531	8	2012-08-02 05:29:58.374313
24186	5	1-791	24183	39	29	173	52	2012-08-02 05:29:58.374313
24187	5	1-6167	24183	29	51	238	2	2012-08-02 05:29:58.374313
24183	5	1-7067	24183	51	86	4335	15	2012-08-02 05:29:58.374313
24189	3	1-6563	24188	50	83	1361	86	2012-08-02 05:29:59.214171
24190	3	1-1896	24188	83	50	1427	40	2012-08-02 05:29:59.214171
24188	3	1-7069	24188	50	50	1676	13	2012-08-02 05:29:59.214171
24192	5	1-6880	24191	44	20	1074	6	2012-08-02 05:29:59.611968
24193	5	1-3393	24191	20	31	409	86	2012-08-02 05:29:59.611968
24194	5	1-6808	24191	31	47	311	34	2012-08-02 05:29:59.611968
24195	5	1-6955	24191	47	61	656	64	2012-08-02 05:29:59.611968
24191	5	1-7070	24191	61	44	1173	56	2012-08-02 05:29:59.611968
24197	5	1-2199	24196	77	8	3856	15	2012-08-02 05:29:59.611968
24198	5	1-5003	24196	8	54	1216	44	2012-08-02 05:29:59.611968
24199	5	1-6937	24196	54	47	707	34	2012-08-02 05:29:59.611968
24200	5	1-6955	24196	47	61	1616	64	2012-08-02 05:29:59.611968
24196	5	1-7070	24196	61	77	3133	56	2012-08-02 05:29:59.611968
24202	5	1-6880	24201	44	42	432	6	2012-08-02 05:29:59.611968
24203	5	1-2544	24201	42	46	138	69	2012-08-02 05:29:59.611968
24204	5	1-7027	24201	46	47	94	34	2012-08-02 05:29:59.611968
24205	5	1-6955	24201	47	61	218	64	2012-08-02 05:29:59.611968
24201	5	1-7070	24201	61	44	429	56	2012-08-02 05:29:59.611968
24206	1	1-2199	\N	77	77	109	15	2012-08-02 05:29:59.611968
24208	5	1-6060	24207	4	37	3240	70	2012-08-02 05:30:00.407485
24209	5	1-4715	24207	37	19	2079	31	2012-08-02 05:30:00.407485
24210	5	1-6693	24207	19	1	1638	100	2012-08-02 05:30:00.407485
24211	5	1-6998	24207	1	99	2644	78	2012-08-02 05:30:00.407485
24207	5	1-7072	24207	99	4	3837	60	2012-08-02 05:30:00.407485
24213	5	1-5596	24212	93	66	2236	85	2012-08-02 05:30:00.407485
24214	5	1-3517	24212	66	30	613	73	2012-08-02 05:30:00.407485
24215	5	1-6904	24212	30	1	522	100	2012-08-02 05:30:00.407485
24216	5	1-6998	24212	1	99	921	78	2012-08-02 05:30:00.407485
24212	5	1-7072	24212	99	93	1462	60	2012-08-02 05:30:00.407485
24218	5	1-4552	24217	38	9	222	30	2012-08-02 05:30:00.407485
24219	5	1-7020	24217	9	27	278	82	2012-08-02 05:30:00.407485
24220	5	1-4087	24217	27	1	315	100	2012-08-02 05:30:00.407485
24221	5	1-6998	24217	1	99	576	78	2012-08-02 05:30:00.407485
24217	5	1-7072	24217	99	38	947	60	2012-08-02 05:30:00.407485
24223	3	1-6900	24222	54	75	3867	18	2012-08-02 05:30:01.214283
24224	3	1-7030	24222	75	56	2986	71	2012-08-02 05:30:01.214283
24222	3	1-7074	24222	56	54	3395	49	2012-08-02 05:30:01.214283
24226	5	1-6303	24225	60	100	2326	88	2012-08-02 05:30:01.214283
24227	5	1-6837	24225	100	34	880	75	2012-08-02 05:30:01.214283
24228	5	1-7056	24225	34	81	942	22	2012-08-02 05:30:01.214283
24229	5	1-6575	24225	81	56	883	71	2012-08-02 05:30:01.214283
24225	5	1-7074	24225	56	60	1416	49	2012-08-02 05:30:01.214283
24231	4	1-6987	24230	32	40	3500	19	2012-08-02 05:30:02.396483
24232	4	1-5008	24230	40	44	3275	56	2012-08-02 05:30:02.396483
24233	4	1-6880	24230	44	3	2959	6	2012-08-02 05:30:02.396483
24230	4	1-7077	24230	3	32	4005	12	2012-08-02 05:30:02.396483
24235	3	1-4509	24234	66	52	311	100	2012-08-02 05:30:02.79416
24236	3	1-5653	24234	52	6	177	63	2012-08-02 05:30:02.79416
24234	3	1-7078	24234	6	66	641	22	2012-08-02 05:30:02.79416
24238	4	1-6391	24237	40	30	3404	73	2012-08-02 05:30:02.79416
24239	4	1-6904	24237	30	52	1740	100	2012-08-02 05:30:02.79416
24240	4	1-5653	24237	52	6	1014	63	2012-08-02 05:30:02.79416
24237	4	1-7078	24237	6	40	3750	22	2012-08-02 05:30:02.79416
24242	3	1-6808	24241	31	5	1168	34	2012-08-02 05:30:03.181102
24243	3	1-6356	24241	5	10	1123	65	2012-08-02 05:30:03.181102
24241	3	1-7079	24241	10	31	1613	86	2012-08-02 05:30:03.181102
24245	4	1-6736	24244	90	3	418	23	2012-08-02 05:30:03.181102
24246	4	1-2805	24244	3	82	160	71	2012-08-02 05:30:03.181102
24247	4	1-6934	24244	82	10	215	65	2012-08-02 05:30:03.181102
24244	4	1-7079	24244	10	90	361	86	2012-08-02 05:30:03.181102
24248	1	1-2805	\N	3	3	996	71	2012-08-02 05:30:03.181102
24250	5	1-7010	24249	29	79	680	61	2012-08-02 05:30:05.14959
24251	5	1-6424	24249	79	54	541	31	2012-08-02 05:30:05.14959
24252	5	1-7046	24249	54	22	664	69	2012-08-02 05:30:05.14959
24253	5	1-6435	24249	22	8	572	39	2012-08-02 05:30:05.14959
24249	5	1-7084	24249	8	29	289	90	2012-08-02 05:30:05.14959
24255	4	1-7041	24254	33	39	1846	8	2012-08-02 05:30:05.968545
24256	4	1-791	24254	39	39	655	52	2012-08-02 05:30:05.968545
24257	4	1-6964	24254	39	45	3960	83	2012-08-02 05:30:05.968545
24254	4	1-7086	24254	45	33	1425	23	2012-08-02 05:30:05.968545
24259	5	1-4122	24258	32	66	338	27	2012-08-02 05:30:06.377106
24260	5	1-5736	24258	66	24	167	39	2012-08-02 05:30:06.377106
24261	5	1-7085	24258	24	17	130	96	2012-08-02 05:30:06.377106
24262	5	1-6358	24258	17	26	109	41	2012-08-02 05:30:06.377106
24258	5	1-7087	24258	26	32	3633	83	2012-08-02 05:30:06.377106
24264	4	1-7013	24263	27	56	1638	37	2012-08-02 05:30:06.377106
24265	4	1-6043	24263	56	61	628	10	2012-08-02 05:30:06.377106
24266	4	1-4221	24263	61	26	70	41	2012-08-02 05:30:06.377106
24263	4	1-7087	24263	26	27	2346	83	2012-08-02 05:30:06.377106
24268	5	1-6948	24267	92	89	1778	84	2012-08-02 05:30:06.78614
24269	5	1-6850	24267	89	38	1920	81	2012-08-02 05:30:06.78614
24270	5	1-6920	24267	38	83	3104	1	2012-08-02 05:30:06.78614
24271	5	1-6389	24267	83	77	1868	94	2012-08-02 05:30:06.78614
24267	5	1-7088	24267	77	92	3389	62	2012-08-02 05:30:06.78614
24273	5	1-3538	24272	72	47	983	34	2012-08-02 05:30:08.01239
24274	5	1-6955	24272	47	12	1796	64	2012-08-02 05:30:08.01239
24275	5	1-6555	24272	12	7	1099	47	2012-08-02 05:30:08.01239
24276	5	1-6159	24272	7	48	871	66	2012-08-02 05:30:08.01239
24272	5	1-7091	24272	48	72	1343	35	2012-08-02 05:30:08.01239
24278	5	1-7046	24277	54	95	2408	69	2012-08-02 05:30:08.829956
24279	5	1-6312	24277	95	68	1720	26	2012-08-02 05:30:08.829956
24280	5	1-4009	24277	68	44	603	94	2012-08-02 05:30:08.829956
24281	5	1-5783	24277	44	41	870	48	2012-08-02 05:30:08.829956
24277	5	1-7093	24277	41	54	2169	31	2012-08-02 05:30:08.829956
24283	5	1-6693	24282	19	1	435	100	2012-08-02 05:30:08.829956
24284	5	1-6998	24282	1	82	743	78	2012-08-02 05:30:08.829956
24285	5	1-4671	24282	82	44	137	94	2012-08-02 05:30:08.829956
24286	5	1-5783	24282	44	41	203	48	2012-08-02 05:30:08.829956
24282	5	1-7093	24282	41	19	521	31	2012-08-02 05:30:08.829956
24288	4	1-6550	24287	10	100	772	9	2012-08-02 05:30:10.487041
24289	4	1-6602	24287	100	6	1436	89	2012-08-02 05:30:10.487041
24290	4	1-2361	24287	6	21	415	91	2012-08-02 05:30:10.487041
24287	4	1-7097	24287	21	10	637	71	2012-08-02 05:30:10.487041
24292	5	1-7074	24291	56	92	1683	49	2012-08-02 05:30:10.487041
24293	5	1-6842	24291	92	24	1140	8	2012-08-02 05:30:10.487041
24294	5	1-5397	24291	24	12	693	35	2012-08-02 05:30:10.487041
24295	5	1-7053	24291	12	21	636	91	2012-08-02 05:30:10.487041
24291	5	1-7097	24291	21	56	988	71	2012-08-02 05:30:10.487041
24297	4	1-2420	24296	47	33	437	23	2012-08-02 05:30:11.315671
24298	4	1-7041	24296	33	12	647	8	2012-08-02 05:30:11.315671
24299	4	1-6125	24296	12	22	969	58	2012-08-02 05:30:11.315671
24296	4	1-7099	24296	22	47	1303	72	2012-08-02 05:30:11.315671
24301	5	1-6575	24300	81	83	1023	71	2012-08-02 05:30:12.133489
24302	5	1-6381	24300	83	3	849	43	2012-08-02 05:30:12.133489
24303	5	1-6347	24300	3	25	729	98	2012-08-02 05:30:12.133489
24304	5	1-1221	24300	25	56	199	24	2012-08-02 05:30:12.133489
24300	5	1-7101	24300	56	81	1198	22	2012-08-02 05:30:12.133489
24306	5	1-6575	24305	81	10	974	71	2012-08-02 05:30:12.133489
24307	5	1-6550	24305	10	6	1193	9	2012-08-02 05:30:12.133489
24308	5	1-6816	24305	6	25	441	98	2012-08-02 05:30:12.133489
24309	5	1-1221	24305	25	56	140	24	2012-08-02 05:30:12.133489
24305	5	1-7101	24305	56	81	980	22	2012-08-02 05:30:12.133489
24311	4	1-6391	24310	40	53	749	73	2012-08-02 05:30:12.133489
24312	4	1-5856	24310	53	25	210	98	2012-08-02 05:30:12.133489
24313	4	1-1221	24310	25	56	66	24	2012-08-02 05:30:12.133489
24310	4	1-7101	24310	56	40	463	22	2012-08-02 05:30:12.133489
24314	1	1-1221	\N	25	25	176	24	2012-08-02 05:30:12.133489
24316	5	1-6808	24315	31	47	179	34	2012-08-02 05:30:13.359889
24317	5	1-6955	24315	47	79	356	64	2012-08-02 05:30:13.359889
24318	5	1-3161	24315	79	50	123	1	2012-08-02 05:30:13.359889
24319	5	1-6681	24315	50	15	248	15	2012-08-02 05:30:13.359889
24315	5	1-7104	24315	15	31	250	86	2012-08-02 05:30:13.359889
24321	4	1-6808	24320	31	47	1345	34	2012-08-02 05:30:13.359889
24322	4	1-6955	24320	47	81	2728	64	2012-08-02 05:30:13.359889
24323	4	1-1501	24320	81	15	1788	15	2012-08-02 05:30:13.359889
24320	4	1-7104	24320	15	31	1839	86	2012-08-02 05:30:13.359889
24325	4	1-6736	24324	90	100	1629	23	2012-08-02 05:30:13.359889
24326	4	1-5898	24324	100	47	1783	28	2012-08-02 05:30:13.359889
24327	4	1-5770	24324	47	15	1187	15	2012-08-02 05:30:13.359889
24324	4	1-7104	24324	15	90	1409	86	2012-08-02 05:30:13.359889
24329	5	1-6643	24328	77	30	4646	73	2012-08-02 05:30:13.768662
24330	5	1-6904	24328	30	48	3458	100	2012-08-02 05:30:13.768662
24331	5	1-5099	24328	48	63	1353	97	2012-08-02 05:30:13.768662
24332	5	1-6047	24328	63	19	1233	11	2012-08-02 05:30:13.768662
24328	5	1-7105	24328	19	77	5174	50	2012-08-02 05:30:13.768662
24334	5	1-6093	24333	23	71	1060	4	2012-08-02 05:30:13.768662
24335	5	1-5910	24333	71	71	415	28	2012-08-02 05:30:13.768662
24336	5	1-6995	24333	71	64	308	30	2012-08-02 05:30:13.768662
24337	5	1-6212	24333	64	19	309	11	2012-08-02 05:30:13.768662
24333	5	1-7105	24333	19	23	1438	50	2012-08-02 05:30:13.768662
24339	5	1-7018	24338	14	98	8754	48	2012-08-02 05:30:14.233013
24340	5	1-5912	24338	98	90	2335	25	2012-08-02 05:30:14.233013
24341	5	1-6792	24338	90	12	1369	47	2012-08-02 05:30:14.233013
24342	5	1-6981	24338	12	96	1911	75	2012-08-02 05:30:14.233013
24338	5	1-7106	24338	96	14	2973	49	2012-08-02 05:30:14.233013
24344	5	1-5078	24343	68	95	10	69	2012-08-02 05:30:14.652631
24345	5	1-6312	24343	95	25	7	26	2012-08-02 05:30:14.652631
24346	5	1-6340	24343	25	90	6	86	2012-08-02 05:30:14.652631
24347	5	1-6736	24343	90	65	7	23	2012-08-02 05:30:14.652631
24343	5	1-7107	24343	65	68	10	32	2012-08-02 05:30:14.652631
24349	5	1-5899	24348	56	10	2866	2	2012-08-02 05:30:14.652631
24350	5	1-6985	24348	10	63	2038	67	2012-08-02 05:30:14.652631
24351	5	1-6866	24348	63	90	1522	86	2012-08-02 05:30:14.652631
24352	5	1-6736	24348	90	65	1730	23	2012-08-02 05:30:14.652631
24348	5	1-7107	24348	65	56	2535	32	2012-08-02 05:30:14.652631
24354	5	1-6880	24353	44	26	2735	6	2012-08-02 05:30:15.072482
24355	5	1-6821	24353	26	50	1950	77	2012-08-02 05:30:15.072482
24356	5	1-6682	24353	50	35	1200	4	2012-08-02 05:30:15.072482
24357	5	1-6433	24353	35	13	663	9	2012-08-02 05:30:15.072482
24353	5	1-7108	24353	13	44	3550	56	2012-08-02 05:30:15.072482
24359	3	1-2583	24358	9	91	1201	50	2012-08-02 05:30:15.072482
24360	3	1-6616	24358	91	13	628	9	2012-08-02 05:30:15.072482
24358	3	1-7108	24358	13	9	3540	56	2012-08-02 05:30:15.072482
24362	3	1-6689	24361	26	17	2774	25	2012-08-02 05:30:15.503484
24363	3	1-5744	24361	17	87	1793	14	2012-08-02 05:30:15.503484
24361	3	1-7109	24361	87	26	5305	37	2012-08-02 05:30:15.503484
24365	5	1-4031	24364	12	82	267	39	2012-08-02 05:30:15.901086
24366	5	1-6785	24364	82	90	910	10	2012-08-02 05:30:15.901086
24367	5	1-7083	24364	90	50	1381	40	2012-08-02 05:30:15.901086
24368	5	1-7069	24364	50	35	1417	13	2012-08-02 05:30:15.901086
24364	5	1-7110	24364	35	12	1047	65	2012-08-02 05:30:15.901086
24370	4	1-6989	24369	62	46	2107	54	2012-08-02 05:30:17.149747
24371	4	1-6951	24369	46	10	1443	55	2012-08-02 05:30:17.149747
24372	4	1-2794	24369	10	41	1070	17	2012-08-02 05:30:17.149747
24369	4	1-7113	24369	41	62	1595	2	2012-08-02 05:30:17.149747
24374	2	1-6693	24373	19	93	625	100	2012-08-02 05:30:17.547301
24373	2	1-7114	24373	93	19	740	31	2012-08-02 05:30:17.547301
24376	5	1-6902	24375	25	22	577	58	2012-08-02 05:30:18.375944
24377	5	1-7099	24375	22	18	712	72	2012-08-02 05:30:18.375944
24378	5	1-6966	24375	18	23	282	8	2012-08-02 05:30:18.375944
24379	5	1-6519	24375	23	86	212	6	2012-08-02 05:30:18.375944
24375	5	1-7116	24375	86	25	358	9	2012-08-02 05:30:18.375944
24381	3	1-4280	24380	31	85	1684	20	2012-08-02 05:30:18.375944
24382	3	1-6337	24380	85	86	1590	6	2012-08-02 05:30:18.375944
24380	3	1-7116	24380	86	31	3298	9	2012-08-02 05:30:18.375944
24383	1	1-6337	\N	85	85	568	6	2012-08-02 05:30:18.375944
24385	4	1-4662	24384	93	93	249	76	2012-08-02 05:30:20.02189
24386	4	1-6787	24384	93	14	398	42	2012-08-02 05:30:20.02189
24387	4	1-6874	24384	14	7	813	53	2012-08-02 05:30:20.02189
24384	4	1-7120	24384	7	93	1240	97	2012-08-02 05:30:20.02189
24389	4	1-6885	24388	70	28	1355	67	2012-08-02 05:30:20.02189
24390	4	1-6674	24388	28	14	430	42	2012-08-02 05:30:20.02189
24391	4	1-6874	24388	14	7	1237	53	2012-08-02 05:30:20.02189
24388	4	1-7120	24388	7	70	2657	97	2012-08-02 05:30:20.02189
24393	4	1-6243	24392	12	27	1399	36	2012-08-02 05:30:20.02189
24394	4	1-29	24392	27	14	394	42	2012-08-02 05:30:20.02189
24395	4	1-6874	24392	14	7	1134	53	2012-08-02 05:30:20.02189
24392	4	1-7120	24392	7	12	2437	97	2012-08-02 05:30:20.02189
24397	4	1-6784	24396	94	45	858	19	2012-08-02 05:30:20.02189
24398	4	1-6115	24396	45	90	899	29	2012-08-02 05:30:20.02189
24399	4	1-6331	24396	90	7	1262	53	2012-08-02 05:30:20.02189
24396	4	1-7120	24396	7	94	2742	97	2012-08-02 05:30:20.02189
24401	4	1-6026	24400	1	1	1431	14	2012-08-02 05:30:20.02189
24402	4	1-5675	24400	1	14	133	42	2012-08-02 05:30:20.02189
24403	4	1-6874	24400	14	7	400	53	2012-08-02 05:30:20.02189
24400	4	1-7120	24400	7	1	896	97	2012-08-02 05:30:20.02189
24405	5	1-5863	24404	72	29	2927	90	2012-08-02 05:30:20.850931
24406	5	1-7010	24404	29	19	5482	61	2012-08-02 05:30:20.850931
24407	5	1-6317	24404	19	70	3231	5	2012-08-02 05:30:20.850931
24408	5	1-6999	24404	70	8	1830	39	2012-08-02 05:30:20.850931
24404	5	1-7122	24404	8	72	5952	99	2012-08-02 05:30:20.850931
24410	5	1-6728	24409	55	75	972	97	2012-08-02 05:30:20.850931
24411	5	1-3396	24409	75	23	435	43	2012-08-02 05:30:20.850931
24412	5	1-4162	24409	23	13	184	74	2012-08-02 05:30:20.850931
24413	5	1-7035	24409	13	8	134	39	2012-08-02 05:30:20.850931
24409	5	1-7122	24409	8	55	489	99	2012-08-02 05:30:20.850931
24415	4	1-7042	24414	71	23	1856	43	2012-08-02 05:30:20.850931
24416	4	1-4162	24414	23	13	793	74	2012-08-02 05:30:20.850931
24417	4	1-7035	24414	13	8	582	39	2012-08-02 05:30:20.850931
24414	4	1-7122	24414	8	71	2153	99	2012-08-02 05:30:20.850931
24419	4	1-5945	24418	72	20	1198	53	2012-08-02 05:30:20.850931
24420	4	1-4921	24418	20	13	307	74	2012-08-02 05:30:20.850931
24421	4	1-7035	24418	13	8	233	39	2012-08-02 05:30:20.850931
24418	4	1-7122	24418	8	72	895	99	2012-08-02 05:30:20.850931
24423	5	1-6030	24422	7	41	530	17	2012-08-02 05:30:22.530398
24424	5	1-7113	24422	41	10	808	2	2012-08-02 05:30:22.530398
24422	5	1-7126	24422	99	7	1322	74	2012-08-02 05:30:22.530398
24425	5	1-6985	24422	10	70	443	67	2012-08-02 05:30:22.530398
24426	5	1-6954	24422	70	99	787	44	2012-08-02 05:30:22.530398
24428	3	1-7035	24427	13	51	3049	39	2012-08-02 05:30:22.530398
24429	3	1-5541	24427	51	99	2333	44	2012-08-02 05:30:22.530398
24427	3	1-7126	24427	99	13	4402	74	2012-08-02 05:30:22.530398
24431	3	1-6875	24430	92	42	1920	12	2012-08-02 05:30:22.530398
24432	3	1-6855	24430	42	99	1764	44	2012-08-02 05:30:22.530398
24430	3	1-7126	24430	99	92	3359	74	2012-08-02 05:30:22.530398
24434	3	1-6689	24433	26	12	282	25	2012-08-02 05:30:24.695435
24435	3	1-4759	24433	12	28	224	85	2012-08-02 05:30:24.695435
24433	3	1-7131	24433	28	26	723	37	2012-08-02 05:30:24.695435
24437	4	1-6043	24436	56	3	1168	10	2012-08-02 05:30:24.695435
24438	4	1-7034	24436	3	44	397	98	2012-08-02 05:30:24.695435
24439	4	1-5848	24436	44	28	713	85	2012-08-02 05:30:24.695435
24436	4	1-7131	24436	28	56	2830	37	2012-08-02 05:30:24.695435
24441	4	1-6588	24440	66	3	3850	10	2012-08-02 05:30:25.932963
24442	4	1-7034	24440	3	65	1395	98	2012-08-02 05:30:25.932963
24443	4	1-6708	24440	65	49	897	64	2012-08-02 05:30:25.932963
24440	4	1-7134	24440	49	66	4098	93	2012-08-02 05:30:25.932963
24445	5	1-7010	24444	29	86	1831	61	2012-08-02 05:30:26.761315
24446	5	1-5542	24444	86	1	476	10	2012-08-02 05:30:26.761315
24447	5	1-6830	24444	1	71	232	28	2012-08-02 05:30:26.761315
24448	5	1-6995	24444	71	72	88	30	2012-08-02 05:30:26.761315
24444	5	1-7136	24444	72	29	1550	90	2012-08-02 05:30:26.761315
24450	5	1-6549	24449	79	32	1279	98	2012-08-02 05:30:26.761315
24451	5	1-7055	24449	32	37	677	77	2012-08-02 05:30:26.761315
24452	5	1-6516	24449	37	71	293	28	2012-08-02 05:30:26.761315
24453	5	1-6995	24449	71	72	121	30	2012-08-02 05:30:26.761315
24449	5	1-7136	24449	72	79	2321	90	2012-08-02 05:30:26.761315
24455	4	1-6824	24454	17	12	1513	80	2012-08-02 05:30:26.761315
24456	4	1-6035	24454	12	71	289	28	2012-08-02 05:30:26.761315
24457	4	1-6995	24454	71	72	123	30	2012-08-02 05:30:26.761315
24454	4	1-7136	24454	72	17	2450	90	2012-08-02 05:30:26.761315
24459	4	1-6232	24458	94	48	541	100	2012-08-02 05:30:27.181526
24460	4	1-5099	24458	48	1	179	97	2012-08-02 05:30:27.181526
24461	4	1-6026	24458	1	90	225	14	2012-08-02 05:30:27.181526
24458	4	1-7137	24458	90	94	455	18	2012-08-02 05:30:27.181526
24463	3	1-4537	24462	60	20	705	22	2012-08-02 05:30:27.181526
24464	3	1-6257	24462	20	90	829	14	2012-08-02 05:30:27.181526
24462	3	1-7137	24462	90	60	1778	18	2012-08-02 05:30:27.181526
24466	4	1-7110	24465	35	12	1388	65	2012-08-02 05:30:27.590249
24467	4	1-4031	24465	12	6	261	39	2012-08-02 05:30:27.590249
24468	4	1-7121	24465	6	69	183	35	2012-08-02 05:30:27.590249
24465	4	1-7138	24465	69	35	2542	13	2012-08-02 05:30:27.590249
24470	5	1-6740	24469	78	26	1274	60	2012-08-02 05:30:27.590249
24471	5	1-2053	24469	26	33	293	23	2012-08-02 05:30:27.590249
24472	5	1-7041	24469	33	24	388	8	2012-08-02 05:30:27.590249
24473	5	1-5397	24469	24	69	186	35	2012-08-02 05:30:27.590249
24469	5	1-7138	24469	69	78	3156	13	2012-08-02 05:30:27.590249
24475	4	1-6994	24474	16	38	1084	7	2012-08-02 05:30:27.590249
24476	4	1-2987	24474	38	24	131	8	2012-08-02 05:30:27.590249
24477	4	1-5397	24474	24	69	67	35	2012-08-02 05:30:27.590249
24474	4	1-7138	24474	69	16	1204	13	2012-08-02 05:30:27.590249
24479	5	1-5913	24478	1	21	949	16	2012-08-02 05:30:27.590249
24480	5	1-5775	24478	21	66	144	27	2012-08-02 05:30:27.590249
24481	5	1-5736	24478	66	6	85	39	2012-08-02 05:30:27.590249
24482	5	1-7121	24478	6	69	89	35	2012-08-02 05:30:27.590249
24478	5	1-7138	24478	69	1	1866	13	2012-08-02 05:30:27.590249
24484	4	1-6588	24483	66	3	828	10	2012-08-02 05:30:31.734848
24485	4	1-7034	24483	3	38	306	98	2012-08-02 05:30:31.734848
24486	4	1-6990	24483	38	62	368	45	2012-08-02 05:30:31.734848
24483	4	1-7148	24483	62	66	864	93	2012-08-02 05:30:31.734848
24488	5	1-6588	24487	66	90	1540	10	2012-08-02 05:30:31.734848
24489	5	1-7083	24487	90	23	2590	40	2012-08-02 05:30:31.734848
24490	5	1-6753	24487	23	38	540	98	2012-08-02 05:30:31.734848
24491	5	1-6990	24487	38	62	661	45	2012-08-02 05:30:31.734848
24487	5	1-7148	24487	62	66	1579	93	2012-08-02 05:30:31.734848
24493	5	1-7106	24492	96	48	2437	49	2012-08-02 05:30:32.563513
24494	5	1-5239	24492	48	6	914	39	2012-08-02 05:30:32.563513
24495	5	1-7121	24492	6	12	816	35	2012-08-02 05:30:32.563513
24496	5	1-7053	24492	12	45	615	91	2012-08-02 05:30:32.563513
24492	5	1-7150	24492	45	96	1690	75	2012-08-02 05:30:32.563513
24498	5	1-2268	24497	96	44	200	28	2012-08-02 05:30:32.563513
24499	5	1-6748	24497	44	24	180	8	2012-08-02 05:30:32.563513
24500	5	1-5397	24497	24	12	96	35	2012-08-02 05:30:32.563513
24501	5	1-7053	24497	12	45	78	91	2012-08-02 05:30:32.563513
24497	5	1-7150	24497	45	96	228	75	2012-08-02 05:30:32.563513
24503	5	1-7056	24502	34	85	1924	22	2012-08-02 05:30:32.563513
24504	5	1-5944	24502	85	77	1296	94	2012-08-02 05:30:32.563513
24505	5	1-7088	24502	77	69	2289	62	2012-08-02 05:30:32.563513
24506	5	1-5514	24502	69	45	616	91	2012-08-02 05:30:32.563513
24502	5	1-7150	24502	45	34	1867	75	2012-08-02 05:30:32.563513
24508	5	1-3889	24507	96	28	1594	70	2012-08-02 05:30:32.563513
24509	5	1-6817	24507	28	32	882	97	2012-08-02 05:30:32.563513
24510	5	1-7057	24507	32	69	1881	62	2012-08-02 05:30:32.563513
24511	5	1-5514	24507	69	45	534	91	2012-08-02 05:30:32.563513
24507	5	1-7150	24507	45	96	1707	75	2012-08-02 05:30:32.563513
24513	5	1-6519	24512	23	86	1082	6	2012-08-02 05:30:32.983663
24514	5	1-7116	24512	86	8	2196	9	2012-08-02 05:30:32.983663
24515	5	1-6380	24512	8	75	1579	100	2012-08-02 05:30:32.983663
24516	5	1-7100	24512	75	90	1699	33	2012-08-02 05:30:32.983663
24512	5	1-7151	24512	90	23	1201	8	2012-08-02 05:30:32.983663
24518	3	1-2710	24517	94	9	472	35	2012-08-02 05:30:33.414506
24519	3	1-6449	24517	9	35	915	70	2012-08-02 05:30:33.414506
24517	3	1-7152	24517	35	94	2121	9	2012-08-02 05:30:33.414506
24521	5	1-3175	24520	85	51	679	64	2012-08-02 05:30:33.414506
24522	5	1-5708	24520	51	6	465	39	2012-08-02 05:30:33.414506
24523	5	1-7121	24520	6	9	506	35	2012-08-02 05:30:33.414506
24524	5	1-6449	24520	9	35	1031	70	2012-08-02 05:30:33.414506
24520	5	1-7152	24520	35	85	2509	9	2012-08-02 05:30:33.414506
24526	4	1-3667	24525	92	93	515	30	2012-08-02 05:30:33.834506
24527	4	1-6016	24525	93	56	224	24	2012-08-02 05:30:33.834506
24528	4	1-7101	24525	56	3	1339	22	2012-08-02 05:30:33.834506
24525	4	1-7153	24525	3	92	846	99	2012-08-02 05:30:33.834506
24529	2	1-7160	24529	60	90	1210	84	2012-08-02 05:30:36.783923
24530	2	1-1533	24529	90	60	630	46	2012-08-02 05:30:36.783923
24532	4	1-6645	24531	48	10	998	20	2012-08-02 05:30:37.988222
24533	4	1-6936	24531	10	14	1038	55	2012-08-02 05:30:37.988222
24534	4	1-6750	24531	14	75	3468	53	2012-08-02 05:30:37.988222
24531	4	1-7163	24531	75	48	2640	71	2012-08-02 05:30:37.988222
24536	4	1-7041	24535	33	1	1942	8	2012-08-02 05:30:38.385849
24537	4	1-6876	24535	1	35	1287	61	2012-08-02 05:30:38.385849
24538	4	1-2632	24535	35	30	444	17	2012-08-02 05:30:38.385849
24535	4	1-7164	24535	30	33	2230	23	2012-08-02 05:30:38.385849
24540	3	1-6781	24539	17	7	675	74	2012-08-02 05:30:38.385849
24541	3	1-6030	24539	7	30	207	17	2012-08-02 05:30:38.385849
24539	3	1-7164	24539	30	17	1125	23	2012-08-02 05:30:38.385849
24543	3	1-7068	24542	1	57	1786	100	2012-08-02 05:30:38.385849
24544	3	1-6877	24542	57	30	749	17	2012-08-02 05:30:38.385849
24542	3	1-7164	24542	30	1	4118	23	2012-08-02 05:30:38.385849
24546	4	1-4972	24545	93	8	400	70	2012-08-02 05:30:39.214775
24547	4	1-6873	24545	8	65	459	42	2012-08-02 05:30:39.214775
24548	4	1-6418	24545	65	12	448	36	2012-08-02 05:30:39.214775
24545	4	1-7166	24545	12	93	440	35	2012-08-02 05:30:39.214775
24550	4	1-3538	24549	72	47	4303	34	2012-08-02 05:30:39.214775
24551	4	1-7039	24549	47	78	4529	8	2012-08-02 05:30:39.214775
24552	4	1-6840	24549	78	12	3828	36	2012-08-02 05:30:39.214775
24549	4	1-7166	24549	12	72	4543	35	2012-08-02 05:30:39.214775
24553	1	1-4972	\N	93	93	4187	70	2012-08-02 05:30:39.214775
24555	2	1-6635	24554	70	54	73	41	2012-08-02 05:30:39.645469
24554	2	1-7167	24554	54	70	88	91	2012-08-02 05:30:39.645469
24557	5	1-6433	24556	35	25	3075	9	2012-08-02 05:30:40.882657
24558	5	1-6902	24556	25	85	4623	58	2012-08-02 05:30:40.882657
24559	5	1-6767	24556	85	60	4317	46	2012-08-02 05:30:40.882657
24560	5	1-7160	24556	60	95	6419	84	2012-08-02 05:30:40.882657
24556	5	1-7170	24556	95	35	5420	4	2012-08-02 05:30:40.882657
24562	4	1-7144	24561	55	17	3617	96	2012-08-02 05:30:41.29164
24563	4	1-6358	24561	17	79	2766	41	2012-08-02 05:30:41.29164
24564	4	1-7139	24561	79	77	2479	83	2012-08-02 05:30:41.29164
24561	4	1-7171	24561	77	55	5821	3	2012-08-02 05:30:41.29164
24566	4	1-6474	24565	65	56	1807	1	2012-08-02 05:30:41.29164
24567	4	1-5947	24565	56	79	774	41	2012-08-02 05:30:41.29164
24568	4	1-7139	24565	79	77	845	83	2012-08-02 05:30:41.29164
24565	4	1-7171	24565	77	65	2419	3	2012-08-02 05:30:41.29164
24570	4	1-7141	24569	100	20	1026	32	2012-08-02 05:30:41.29164
24571	4	1-5852	24569	20	71	470	68	2012-08-02 05:30:41.29164
24572	4	1-6002	24569	71	77	497	83	2012-08-02 05:30:41.29164
24569	4	1-7171	24569	77	100	1427	3	2012-08-02 05:30:41.29164
24574	2	1-6934	24573	82	92	6533	65	2012-08-02 05:30:43.777481
24573	2	1-7177	24573	92	82	5632	71	2012-08-02 05:30:43.777481
24576	3	1-7064	24575	79	93	2766	4	2012-08-02 05:30:43.777481
24577	3	1-6984	24575	93	92	2291	65	2012-08-02 05:30:43.777481
24575	3	1-7177	24575	92	79	2001	71	2012-08-02 05:30:43.777481
24579	4	1-6696	24578	53	14	304	55	2012-08-02 05:30:44.65002
24580	4	1-6750	24578	14	2	1287	53	2012-08-02 05:30:44.65002
24581	4	1-5244	24578	2	29	164	100	2012-08-02 05:30:44.65002
24578	4	1-7179	24578	29	53	825	38	2012-08-02 05:30:44.65002
24583	4	1-3476	24582	74	82	581	67	2012-08-02 05:30:44.65002
24584	4	1-6960	24582	82	91	519	19	2012-08-02 05:30:44.65002
24585	4	1-2489	24582	91	29	288	100	2012-08-02 05:30:44.65002
24582	4	1-7179	24582	29	74	1450	38	2012-08-02 05:30:44.65002
24587	5	1-7173	24586	50	50	3702	25	2012-08-02 05:30:45.069905
24588	5	1-7175	24586	50	39	5246	81	2012-08-02 05:30:45.069905
24589	5	1-6944	24586	39	100	3431	3	2012-08-02 05:30:45.069905
24590	5	1-7141	24586	100	83	2358	32	2012-08-02 05:30:45.069905
24586	5	1-7180	24586	83	50	3060	71	2012-08-02 05:30:45.069905
24592	3	1-3333	24591	44	89	961	100	2012-08-02 05:30:47.301708
24593	3	1-7111	24591	89	35	573	27	2012-08-02 05:30:47.301708
24591	3	1-7185	24591	35	44	2774	35	2012-08-02 05:30:47.301708
24595	4	1-2876	24594	57	8	170	70	2012-08-02 05:30:48.119035
24596	4	1-6873	24594	8	14	166	42	2012-08-02 05:30:48.119035
24597	4	1-6874	24594	14	76	458	53	2012-08-02 05:30:48.119035
24594	4	1-7187	24594	76	57	345	8	2012-08-02 05:30:48.119035
24599	4	1-6519	24598	23	86	323	6	2012-08-02 05:30:48.119035
24600	4	1-7116	24598	86	89	557	9	2012-08-02 05:30:48.119035
24601	4	1-3149	24598	89	76	468	53	2012-08-02 05:30:48.119035
24598	4	1-7187	24598	76	23	421	8	2012-08-02 05:30:48.119035
24603	4	1-2876	24602	57	35	344	70	2012-08-02 05:30:48.119035
24604	4	1-7152	24602	35	89	712	9	2012-08-02 05:30:48.119035
24605	4	1-3149	24602	89	76	616	53	2012-08-02 05:30:48.119035
24602	4	1-7187	24602	76	57	570	8	2012-08-02 05:30:48.119035
24607	3	1-791	24606	39	68	816	52	2012-08-02 05:30:48.119035
24608	3	1-5921	24606	68	76	1984	53	2012-08-02 05:30:48.119035
24606	3	1-7187	24606	76	39	2091	8	2012-08-02 05:30:48.119035
24610	4	1-6519	24609	23	3	1474	6	2012-08-02 05:30:48.119035
24611	4	1-7077	24609	3	9	2320	12	2012-08-02 05:30:48.119035
24612	4	1-3202	24609	9	76	1456	53	2012-08-02 05:30:48.119035
24609	4	1-7187	24609	76	23	1587	8	2012-08-02 05:30:48.119035
24614	2	1-4257	24613	67	23	734	37	2012-08-02 05:30:50.572156
24613	2	1-7193	24613	23	67	3376	35	2012-08-02 05:30:50.572156
24616	4	1-6815	24615	60	39	2576	38	2012-08-02 05:30:52.184729
24617	4	1-6579	24615	39	77	2474	87	2012-08-02 05:30:52.184729
24618	4	1-5999	24615	77	73	1497	31	2012-08-02 05:30:52.184729
24615	4	1-7197	24615	73	60	3373	54	2012-08-02 05:30:52.184729
24620	5	1-6951	24619	46	40	1431	55	2012-08-02 05:30:52.184729
24621	5	1-2432	24619	40	34	286	35	2012-08-02 05:30:52.184729
24622	5	1-6684	24619	34	41	331	48	2012-08-02 05:30:52.184729
24623	5	1-7093	24619	41	73	738	31	2012-08-02 05:30:52.184729
24619	5	1-7197	24619	73	46	1924	54	2012-08-02 05:30:52.184729
24625	5	1-6951	24624	46	82	172	55	2012-08-02 05:30:52.184729
24626	5	1-4767	24624	82	28	64	61	2012-08-02 05:30:52.184729
24627	5	1-5483	24624	28	41	34	48	2012-08-02 05:30:52.184729
24628	5	1-7093	24624	41	73	80	31	2012-08-02 05:30:52.184729
24624	5	1-7197	24624	73	46	219	54	2012-08-02 05:30:52.184729
24630	5	1-2738	24629	1	7	600	4	2012-08-02 05:30:52.184729
24631	5	1-6848	24629	7	28	334	58	2012-08-02 05:30:52.184729
24632	5	1-6845	24629	28	41	394	48	2012-08-02 05:30:52.184729
24633	5	1-7093	24629	41	73	979	31	2012-08-02 05:30:52.184729
24629	5	1-7197	24629	73	1	2839	54	2012-08-02 05:30:52.184729
24635	4	1-6433	24634	35	25	645	9	2012-08-02 05:30:52.605005
24636	4	1-6902	24634	25	99	813	58	2012-08-02 05:30:52.605005
24637	4	1-6843	24634	99	21	666	17	2012-08-02 05:30:52.605005
24634	4	1-7198	24634	21	35	1356	4	2012-08-02 05:30:52.605005
24639	3	1-6133	24638	45	35	627	95	2012-08-02 05:30:52.605005
24640	3	1-6259	24638	35	21	316	17	2012-08-02 05:30:52.605005
24638	3	1-7198	24638	21	45	805	4	2012-08-02 05:30:52.605005
24642	4	1-6533	24641	25	54	4048	59	2012-08-02 05:30:52.605005
24643	4	1-7060	24641	54	40	2564	14	2012-08-02 05:30:52.605005
24644	4	1-6854	24641	40	21	2347	17	2012-08-02 05:30:52.605005
24641	4	1-7198	24641	21	25	6403	4	2012-08-02 05:30:52.605005
24646	5	1-7195	24645	24	67	7394	21	2012-08-02 05:30:55.079722
24647	5	1-4626	24645	67	95	2191	84	2012-08-02 05:30:55.079722
24648	5	1-7170	24645	95	93	1525	4	2012-08-02 05:30:55.079722
24649	5	1-6984	24645	93	38	924	65	2012-08-02 05:30:55.079722
24645	5	1-7204	24645	38	24	2250	22	2012-08-02 05:30:55.079722
24651	4	1-6575	24650	81	79	398	71	2012-08-02 05:30:55.079722
24652	4	1-7064	24650	79	93	410	4	2012-08-02 05:30:55.079722
24653	4	1-6984	24650	93	38	253	65	2012-08-02 05:30:55.079722
24650	4	1-7204	24650	38	81	629	22	2012-08-02 05:30:55.079722
24655	5	1-6257	24654	20	45	1441	14	2012-08-02 05:30:55.079722
24656	5	1-6701	24654	45	17	592	81	2012-08-02 05:30:55.079722
24657	5	1-5887	24654	17	93	325	4	2012-08-02 05:30:55.079722
24658	5	1-6984	24654	93	38	283	65	2012-08-02 05:30:55.079722
24654	5	1-7204	24654	38	20	992	22	2012-08-02 05:30:55.079722
24660	5	1-5944	24659	85	33	112	94	2012-08-02 05:30:55.079722
24661	5	1-6222	24659	33	95	48	84	2012-08-02 05:30:55.079722
24662	5	1-7170	24659	95	93	50	4	2012-08-02 05:30:55.079722
24663	5	1-6984	24659	93	38	45	65	2012-08-02 05:30:55.079722
24659	5	1-7204	24659	38	85	163	22	2012-08-02 05:30:55.079722
24665	4	1-5200	24664	20	23	296	50	2012-08-02 05:30:55.079722
24666	4	1-6093	24664	23	93	231	4	2012-08-02 05:30:55.079722
24667	4	1-6984	24664	93	38	214	65	2012-08-02 05:30:55.079722
24664	4	1-7204	24664	38	20	797	22	2012-08-02 05:30:55.079722
24669	4	1-5664	24668	75	15	786	86	2012-08-02 05:30:55.511062
24670	4	1-7145	24668	15	23	789	43	2012-08-02 05:30:55.511062
24671	4	1-6491	24668	23	84	476	16	2012-08-02 05:30:55.511062
24668	4	1-7205	24668	84	75	5271	55	2012-08-02 05:30:55.511062
24673	3	1-6956	24672	41	95	6401	59	2012-08-02 05:30:57.963396
24674	3	1-5711	24672	95	87	1808	63	2012-08-02 05:30:57.963396
24672	3	1-7211	24672	87	41	3721	77	2012-08-02 05:30:57.963396
24676	5	1-5637	24675	65	100	776	3	2012-08-02 05:30:57.963396
24677	5	1-7141	24675	100	68	533	32	2012-08-02 05:30:57.963396
24678	5	1-5078	24675	68	96	474	69	2012-08-02 05:30:57.963396
24679	5	1-6805	24675	96	87	613	63	2012-08-02 05:30:57.963396
24675	5	1-7211	24675	87	65	1369	77	2012-08-02 05:30:57.963396
24681	2	1-5557	24680	77	87	1144	63	2012-08-02 05:30:57.963396
24680	2	1-7211	24680	87	77	2761	77	2012-08-02 05:30:57.963396
24682	1	1-2716	\N	84	84	248	69	2012-08-02 05:30:57.963396
24684	4	1-6682	24683	50	35	857	4	2012-08-02 05:30:58.372281
24685	4	1-6433	24683	35	25	554	9	2012-08-02 05:30:58.372281
24686	4	1-6902	24683	25	40	949	58	2012-08-02 05:30:58.372281
24683	4	1-7212	24683	40	50	1189	77	2012-08-02 05:30:58.372281
24688	4	1-6759	24687	47	2	2743	9	2012-08-02 05:30:58.781154
24689	4	1-7045	24687	2	40	562	19	2012-08-02 05:30:58.781154
24690	4	1-5008	24687	40	34	471	56	2012-08-02 05:30:58.781154
24687	4	1-7213	24687	34	47	9134	40	2012-08-02 05:30:58.781154
24692	4	1-2761	24691	97	42	50	14	2012-08-02 05:30:58.781154
24693	4	1-7129	24691	42	40	10	19	2012-08-02 05:30:58.781154
24694	4	1-5008	24691	40	34	9	56	2012-08-02 05:30:58.781154
24691	4	1-7213	24691	34	97	196	40	2012-08-02 05:30:58.781154
24696	5	1-7133	24695	82	75	2908	49	2012-08-02 05:30:59.202979
24697	5	1-7186	24695	75	45	2531	75	2012-08-02 05:30:59.202979
24698	5	1-1892	24695	45	89	1613	100	2012-08-02 05:30:59.202979
24699	5	1-7111	24695	89	89	1292	27	2012-08-02 05:30:59.202979
24695	5	1-7214	24695	89	82	5057	34	2012-08-02 05:30:59.202979
24701	5	1-7039	24700	47	79	2206	8	2012-08-02 05:30:59.202979
24702	5	1-7143	24700	79	94	459	18	2012-08-02 05:30:59.202979
24703	5	1-6232	24700	94	89	700	100	2012-08-02 05:30:59.202979
24704	5	1-7111	24700	89	89	578	27	2012-08-02 05:30:59.202979
24700	5	1-7214	24700	89	47	2335	34	2012-08-02 05:30:59.202979
24706	5	1-7039	24705	47	24	485	8	2012-08-02 05:30:59.202979
24707	5	1-5397	24705	24	44	278	35	2012-08-02 05:30:59.202979
24708	5	1-3333	24705	44	89	134	100	2012-08-02 05:30:59.202979
24709	5	1-7111	24705	89	89	115	27	2012-08-02 05:30:59.202979
24705	5	1-7214	24705	89	47	479	34	2012-08-02 05:30:59.202979
24711	4	1-5876	24710	79	38	590	19	2012-08-02 05:31:00.915835
24712	4	1-6046	24710	38	68	284	8	2012-08-02 05:31:00.915835
24713	4	1-3090	24710	68	94	141	76	2012-08-02 05:31:00.915835
24710	4	1-7218	24710	94	79	5504	21	2012-08-02 05:31:00.915835
24714	1	1-3090	\N	68	68	6591	76	2012-08-02 05:31:00.915835
24716	3	1-6456	24715	9	28	1445	85	2012-08-02 05:31:01.31334
24717	3	1-7131	24715	28	44	3057	37	2012-08-02 05:31:01.31334
24715	3	1-7219	24715	44	9	3988	43	2012-08-02 05:31:01.31334
24719	3	1-5720	24718	95	4	728	79	2012-08-02 05:31:01.31334
24720	3	1-7080	24718	4	44	597	37	2012-08-02 05:31:01.31334
24718	3	1-7219	24718	44	95	1009	43	2012-08-02 05:31:01.31334
24722	5	1-6725	24721	75	89	2855	68	2012-08-02 05:31:03.77677
24723	5	1-6104	24721	89	83	1559	32	2012-08-02 05:31:03.77677
24724	5	1-7180	24721	83	50	2061	71	2012-08-02 05:31:03.77677
24725	5	1-7173	24721	50	33	2350	25	2012-08-02 05:31:03.77677
24721	5	1-7225	24721	33	75	2145	26	2012-08-02 05:31:03.77677
24727	4	1-7114	24726	93	73	164	31	2012-08-02 05:31:04.605383
24728	4	1-7197	24726	73	100	341	54	2012-08-02 05:31:04.605383
24729	4	1-3777	24726	100	89	18	89	2012-08-02 05:31:04.605383
24726	4	1-7227	24726	89	93	201	100	2012-08-02 05:31:04.605383
24731	4	1-7111	24730	89	58	4513	27	2012-08-02 05:31:04.605383
24732	4	1-7182	24730	58	77	1936	30	2012-08-02 05:31:04.605383
24733	4	1-6973	24730	77	89	707	89	2012-08-02 05:31:04.605383
24730	4	1-7227	24730	89	89	7989	100	2012-08-02 05:31:04.605383
24735	5	1-6518	24734	19	35	266	45	2012-08-02 05:31:05.014306
24734	5	1-7228	24734	25	19	478	96	2012-08-02 05:31:05.014306
24736	5	1-7032	24734	35	84	641	80	2012-08-02 05:31:05.014306
24737	5	1-6459	24734	84	21	338	11	2012-08-02 05:31:05.014306
24738	5	1-6644	24734	21	25	1067	57	2012-08-02 05:31:05.014306
24740	4	1-6365	24739	67	3	1257	6	2012-08-02 05:31:05.014306
24741	4	1-7077	24739	3	40	1923	12	2012-08-02 05:31:05.014306
24742	4	1-4403	24739	40	25	3239	57	2012-08-02 05:31:05.014306
24739	4	1-7228	24739	25	67	1841	96	2012-08-02 05:31:05.014306
24744	4	1-5912	24743	98	33	261	25	2012-08-02 05:31:05.855237
24745	4	1-7225	24743	33	68	206	26	2012-08-02 05:31:05.855237
24746	4	1-4009	24743	68	10	63	94	2012-08-02 05:31:05.855237
24743	4	1-7230	24743	10	98	1144	48	2012-08-02 05:31:05.855237
24748	5	1-6500	24747	7	12	495	80	2012-08-02 05:31:05.855237
24749	5	1-6035	24747	12	71	151	28	2012-08-02 05:31:05.855237
24750	5	1-6995	24747	71	12	90	30	2012-08-02 05:31:05.855237
24751	5	1-5828	24747	12	10	52	94	2012-08-02 05:31:05.855237
24747	5	1-7230	24747	10	7	957	48	2012-08-02 05:31:05.855237
24753	5	1-7112	24752	10	10	2237	62	2012-08-02 05:31:05.855237
24754	5	1-7037	24752	10	95	2274	16	2012-08-02 05:31:05.855237
24755	5	1-7178	24752	95	12	588	30	2012-08-02 05:31:05.855237
24756	5	1-5828	24752	12	10	202	94	2012-08-02 05:31:05.855237
24752	5	1-7230	24752	10	10	3922	48	2012-08-02 05:31:05.855237
24757	1	1-5783	\N	44	44	997	48	2012-08-02 05:31:05.855237
24759	5	1-7111	24758	89	89	267	27	2012-08-02 05:31:06.264442
24760	5	1-7214	24758	89	49	1131	34	2012-08-02 05:31:06.264442
24761	5	1-5778	24758	49	56	162	71	2012-08-02 05:31:06.264442
24762	5	1-6348	24758	56	13	254	95	2012-08-02 05:31:06.264442
24758	5	1-7231	24758	13	89	308	100	2012-08-02 05:31:06.264442
24764	3	1-7091	24763	48	48	801	35	2012-08-02 05:31:06.662317
24765	3	1-3252	24763	48	32	693	40	2012-08-02 05:31:06.662317
24763	3	1-7232	24763	32	48	462	66	2012-08-02 05:31:06.662317
24767	5	1-7091	24766	48	12	4655	35	2012-08-02 05:31:06.662317
24768	5	1-7053	24766	12	5	3861	91	2012-08-02 05:31:06.662317
24769	5	1-6680	24766	5	90	1924	10	2012-08-02 05:31:06.662317
24770	5	1-7083	24766	90	32	3026	40	2012-08-02 05:31:06.662317
24766	5	1-7232	24766	32	48	2529	66	2012-08-02 05:31:06.662317
24772	4	1-7091	24771	48	72	236	35	2012-08-02 05:31:06.662317
24773	4	1-3538	24771	72	1	213	34	2012-08-02 05:31:06.662317
24774	4	1-6869	24771	1	32	144	40	2012-08-02 05:31:06.662317
24771	4	1-7232	24771	32	48	124	66	2012-08-02 05:31:06.662317
24776	4	1-6797	24775	78	51	1148	26	2012-08-02 05:31:06.662317
24777	4	1-7061	24775	51	90	579	10	2012-08-02 05:31:06.662317
24778	4	1-7083	24775	90	32	969	40	2012-08-02 05:31:06.662317
24775	4	1-7232	24775	32	78	861	66	2012-08-02 05:31:06.662317
24780	5	1-7091	24779	48	60	1988	35	2012-08-02 05:31:06.662317
24781	5	1-1317	24779	60	26	983	43	2012-08-02 05:31:06.662317
24782	5	1-6918	24779	26	90	670	10	2012-08-02 05:31:06.662317
24783	5	1-7083	24779	90	32	1128	40	2012-08-02 05:31:06.662317
24779	5	1-7232	24779	32	48	1009	66	2012-08-02 05:31:06.662317
24785	4	1-5842	24784	18	73	61	31	2012-08-02 05:31:10.97137
24786	4	1-7197	24784	73	76	146	54	2012-08-02 05:31:10.97137
24787	4	1-4602	24784	76	6	67	92	2012-08-02 05:31:10.97137
24784	4	1-7243	24784	6	18	88	13	2012-08-02 05:31:10.97137
24789	4	1-5842	24788	18	89	4507	31	2012-08-02 05:31:10.97137
24790	4	1-6783	24788	89	87	3217	17	2012-08-02 05:31:10.97137
24791	4	1-7102	24788	87	6	3834	92	2012-08-02 05:31:10.97137
24788	4	1-7243	24788	6	18	5742	13	2012-08-02 05:31:10.97137
24793	4	1-6464	24792	10	50	379	32	2012-08-02 05:31:10.97137
24794	4	1-4143	24792	50	17	416	81	2012-08-02 05:31:10.97137
24795	4	1-6372	24792	17	6	495	92	2012-08-02 05:31:10.97137
24792	4	1-7243	24792	6	10	883	13	2012-08-02 05:31:10.97137
24797	5	1-5618	24796	78	46	1827	62	2012-08-02 05:31:11.755453
24798	5	1-5444	24796	46	3	1687	54	2012-08-02 05:31:11.755453
24799	5	1-6997	24796	3	62	352	45	2012-08-02 05:31:11.755453
24800	5	1-7148	24796	62	50	766	93	2012-08-02 05:31:11.755453
24796	5	1-7245	24796	50	78	6596	79	2012-08-02 05:31:11.755453
24802	4	1-3823	24801	46	89	365	100	2012-08-02 05:31:11.755453
24803	4	1-7111	24801	89	8	286	27	2012-08-02 05:31:11.755453
24804	4	1-6141	24801	8	50	194	93	2012-08-02 05:31:11.755453
24801	4	1-7245	24801	50	46	1693	79	2012-08-02 05:31:11.755453
24806	3	1-6023	24805	13	39	1170	38	2012-08-02 05:31:12.915478
24807	3	1-6579	24805	39	56	917	87	2012-08-02 05:31:12.915478
24805	3	1-7248	24805	56	13	5573	16	2012-08-02 05:31:12.915478
24809	4	1-7178	24808	95	64	991	30	2012-08-02 05:31:12.915478
24810	4	1-6212	24808	64	10	696	11	2012-08-02 05:31:12.915478
24811	4	1-6564	24808	10	56	457	87	2012-08-02 05:31:12.915478
24808	4	1-7248	24808	56	95	3160	16	2012-08-02 05:31:12.915478
24813	2	1-6381	24812	83	82	2021	43	2012-08-02 05:31:14.826831
24812	2	1-7253	24812	82	83	4205	71	2012-08-02 05:31:14.826831
24815	5	1-7064	24814	79	35	16	4	2012-08-02 05:31:14.826831
24816	5	1-6433	24814	35	2	10	9	2012-08-02 05:31:14.826831
24817	5	1-7045	24814	2	99	2	19	2012-08-02 05:31:14.826831
24818	5	1-7007	24814	99	82	4	43	2012-08-02 05:31:14.826831
24814	5	1-7253	24814	82	79	12	71	2012-08-02 05:31:14.826831
24820	5	1-6452	24819	30	24	1900	89	2012-08-02 05:31:14.826831
24821	5	1-5043	24819	24	63	653	67	2012-08-02 05:31:14.826831
24822	5	1-6866	24819	63	15	402	86	2012-08-02 05:31:14.826831
24823	5	1-7145	24819	15	82	512	43	2012-08-02 05:31:14.826831
24819	5	1-7253	24819	82	30	1690	71	2012-08-02 05:31:14.826831
24825	4	1-7232	24824	32	99	2790	66	2012-08-02 05:31:16.727064
24826	4	1-6405	24824	99	14	3487	14	2012-08-02 05:31:16.727064
24827	4	1-5651	24824	14	95	545	59	2012-08-02 05:31:16.727064
24824	4	1-7258	24824	95	32	6833	40	2012-08-02 05:31:16.727064
24829	3	1-7232	24828	32	2	661	66	2012-08-02 05:31:16.727064
24830	3	1-5864	24828	2	95	127	59	2012-08-02 05:31:16.727064
24828	3	1-7258	24828	95	32	1605	40	2012-08-02 05:31:16.727064
24832	5	1-6753	24831	23	65	186	98	2012-08-02 05:31:16.727064
24833	5	1-6708	24831	65	37	89	64	2012-08-02 05:31:16.727064
24834	5	1-5690	24831	37	93	52	70	2012-08-02 05:31:16.727064
24835	5	1-7095	24831	93	95	63	59	2012-08-02 05:31:16.727064
24831	5	1-7258	24831	95	23	1239	40	2012-08-02 05:31:16.727064
24837	3	1-2761	24836	97	14	18	14	2012-08-02 05:31:16.727064
24838	3	1-5651	24836	14	95	4	59	2012-08-02 05:31:16.727064
24836	3	1-7258	24836	95	97	86	40	2012-08-02 05:31:16.727064
24840	5	1-6807	24839	51	12	1778	92	2012-08-02 05:31:17.533987
24841	5	1-4665	24839	12	95	251	16	2012-08-02 05:31:17.533987
24842	5	1-7178	24839	95	17	125	30	2012-08-02 05:31:17.533987
24843	5	1-6844	24839	17	4	173	25	2012-08-02 05:31:17.533987
24839	5	1-7260	24839	4	51	130	47	2012-08-02 05:31:17.533987
24845	2	1-3149	24844	89	18	34	53	2012-08-02 05:31:18.704954
24844	2	1-7263	24844	18	89	51	9	2012-08-02 05:31:18.704954
24847	2	1-6664	24846	89	18	5251	53	2012-08-02 05:31:18.704954
24846	2	1-7263	24846	18	89	8662	9	2012-08-02 05:31:18.704954
24849	5	1-7077	24848	3	13	845	12	2012-08-02 05:31:19.898305
24850	5	1-6252	24848	13	32	496	97	2012-08-02 05:31:19.898305
24851	5	1-7057	24848	32	10	757	62	2012-08-02 05:31:19.898305
24852	5	1-7037	24848	10	1	654	16	2012-08-02 05:31:19.898305
24848	5	1-7266	24848	1	3	786	6	2012-08-02 05:31:19.898305
24854	2	1-5777	24853	23	1	3048	16	2012-08-02 05:31:19.898305
24853	2	1-7266	24853	1	23	4586	6	2012-08-02 05:31:19.898305
24856	4	1-6830	24855	1	100	2153	28	2012-08-02 05:31:20.295902
24857	4	1-7259	24855	100	17	1360	12	2012-08-02 05:31:20.295902
24858	4	1-5134	24855	17	58	501	95	2012-08-02 05:31:20.295902
24855	4	1-7267	24855	58	1	4568	10	2012-08-02 05:31:20.295902
24860	4	1-7083	24859	90	97	413	40	2012-08-02 05:31:20.295902
24861	4	1-2761	24859	97	53	75	14	2012-08-02 05:31:20.295902
24862	4	1-7254	24859	53	58	32	95	2012-08-02 05:31:20.295902
24859	4	1-7267	24859	58	90	384	10	2012-08-02 05:31:20.295902
24864	5	1-4221	24863	61	54	234	41	2012-08-02 05:31:20.295902
24865	5	1-7167	24863	54	21	231	91	2012-08-02 05:31:20.295902
24866	5	1-7097	24863	21	56	224	71	2012-08-02 05:31:20.295902
24867	5	1-6348	24863	56	58	226	95	2012-08-02 05:31:20.295902
24863	5	1-7267	24863	58	61	2731	10	2012-08-02 05:31:20.295902
24869	4	1-6144	24868	71	32	463	51	2012-08-02 05:31:20.693825
24870	4	1-7161	24868	32	8	231	68	2012-08-02 05:31:20.693825
24871	4	1-6919	24868	8	35	175	55	2012-08-02 05:31:20.693825
24868	4	1-7268	24868	35	71	1158	31	2012-08-02 05:31:20.693825
24873	5	1-6984	24872	93	9	367	65	2012-08-02 05:31:21.080457
24874	5	1-4895	24872	9	91	300	50	2012-08-02 05:31:21.080457
24875	5	1-6616	24872	91	25	168	9	2012-08-02 05:31:21.080457
24876	5	1-6902	24872	25	73	277	58	2012-08-02 05:31:21.080457
24872	5	1-7269	24872	73	93	455	4	2012-08-02 05:31:21.080457
24878	2	1-6848	24877	7	73	2803	58	2012-08-02 05:31:21.080457
24877	2	1-7269	24877	73	7	4934	4	2012-08-02 05:31:21.080457
24879	1	1-6616	\N	91	91	1941	9	2012-08-02 05:31:21.080457
24881	5	1-7005	24880	34	28	759	62	2012-08-02 05:31:23.05801
24882	5	1-7169	24880	28	8	545	68	2012-08-02 05:31:23.05801
24883	5	1-6919	24880	8	35	434	55	2012-08-02 05:31:23.05801
24884	5	1-7268	24880	35	77	3013	31	2012-08-02 05:31:23.05801
24880	5	1-7274	24880	77	34	1317	5	2012-08-02 05:31:23.05801
24886	5	1-7238	24885	54	72	3996	78	2012-08-02 05:31:23.842295
24887	5	1-5722	24885	72	41	2513	76	2012-08-02 05:31:23.842295
24888	5	1-6799	24885	41	50	1646	25	2012-08-02 05:31:23.842295
24889	5	1-7175	24885	50	20	1969	81	2012-08-02 05:31:23.842295
24885	5	1-7276	24885	20	54	5799	46	2012-08-02 05:31:23.842295
24891	4	1-7238	24890	54	72	181	78	2012-08-02 05:31:23.842295
24892	4	1-5722	24890	72	45	118	76	2012-08-02 05:31:23.842295
24893	4	1-6871	24890	45	20	82	81	2012-08-02 05:31:23.842295
24890	4	1-7276	24890	20	54	252	46	2012-08-02 05:31:23.842295
24895	4	1-7167	24894	54	58	386	91	2012-08-02 05:31:25.390895
24896	4	1-6473	24894	58	90	246	34	2012-08-02 05:31:25.390895
24897	4	1-6639	24894	90	8	234	29	2012-08-02 05:31:25.390895
24894	4	1-7280	24894	8	54	345	41	2012-08-02 05:31:25.390895
24899	5	1-7167	24898	54	31	482	91	2012-08-02 05:31:25.390895
24900	5	1-1972	24898	31	94	288	97	2012-08-02 05:31:25.390895
24901	5	1-6784	24898	94	45	117	19	2012-08-02 05:31:25.390895
24902	5	1-6115	24898	45	8	160	29	2012-08-02 05:31:25.390895
24898	5	1-7280	24898	8	54	319	41	2012-08-02 05:31:25.390895
24904	5	1-6927	24903	81	69	1082	61	2012-08-02 05:31:25.788852
24905	5	1-7208	24903	69	55	711	99	2012-08-02 05:31:25.788852
24906	5	1-6728	24903	55	63	1477	97	2012-08-02 05:31:25.788852
24907	5	1-6047	24903	63	43	1505	11	2012-08-02 05:31:25.788852
24903	5	1-7281	24903	43	81	1211	39	2012-08-02 05:31:25.788852
24909	5	1-5622	24908	99	45	1415	83	2012-08-02 05:31:26.562267
24910	5	1-7086	24908	45	65	469	23	2012-08-02 05:31:26.562267
24911	5	1-7107	24908	65	83	512	32	2012-08-02 05:31:26.562267
24912	5	1-7180	24908	83	75	548	71	2012-08-02 05:31:26.562267
24908	5	1-7283	24908	75	99	3034	65	2012-08-02 05:31:26.562267
24914	5	1-3221	24913	8	54	1565	46	2012-08-02 05:31:26.562267
24915	5	1-7238	24913	54	76	1020	78	2012-08-02 05:31:26.562267
24916	5	1-6140	24913	76	21	396	91	2012-08-02 05:31:26.562267
24917	5	1-7097	24913	21	75	456	71	2012-08-02 05:31:26.562267
24913	5	1-7283	24913	75	8	2638	65	2012-08-02 05:31:26.562267
24919	5	1-6493	24918	22	32	1314	51	2012-08-02 05:31:26.562267
24920	5	1-7161	24918	32	89	737	68	2012-08-02 05:31:26.562267
24921	5	1-6104	24918	89	83	372	32	2012-08-02 05:31:26.562267
24922	5	1-7180	24918	83	75	456	71	2012-08-02 05:31:26.562267
24918	5	1-7283	24918	75	22	2880	65	2012-08-02 05:31:26.562267
24924	4	1-7174	24923	8	5	1084	44	2012-08-02 05:31:26.960187
24925	4	1-3642	24923	5	84	586	80	2012-08-02 05:31:26.960187
24926	4	1-6459	24923	84	96	338	11	2012-08-02 05:31:26.960187
24923	4	1-7284	24923	96	8	962	79	2012-08-02 05:31:26.960187
24928	4	1-7080	24927	4	58	4838	37	2012-08-02 05:31:26.960187
24929	4	1-7256	24927	58	47	2066	28	2012-08-02 05:31:26.960187
24930	4	1-6087	24927	47	96	1435	11	2012-08-02 05:31:26.960187
24927	4	1-7284	24927	96	4	4165	79	2012-08-02 05:31:26.960187
24932	5	1-7260	24931	4	36	1592	47	2012-08-02 05:31:27.369492
24933	5	1-6505	24931	36	37	2076	36	2012-08-02 05:31:27.369492
24934	5	1-7184	24931	37	60	1071	67	2012-08-02 05:31:27.369492
24935	5	1-4417	24931	60	29	1107	15	2012-08-02 05:31:27.369492
24931	5	1-7285	24931	29	4	2533	25	2012-08-02 05:31:27.369492
24937	3	1-5268	24936	80	23	1276	31	2012-08-02 05:31:27.369492
24938	3	1-6698	24936	23	29	903	15	2012-08-02 05:31:27.369492
24936	3	1-7285	24936	29	80	2068	25	2012-08-02 05:31:27.369492
24940	3	1-1281	24939	86	40	2726	14	2012-08-02 05:31:29.302338
24941	3	1-6854	24939	40	96	2084	17	2012-08-02 05:31:29.302338
24939	3	1-7290	24939	96	86	4028	69	2012-08-02 05:31:29.302338
24943	4	1-7103	24942	90	91	3042	52	2012-08-02 05:31:29.302338
24944	4	1-6547	24942	91	57	2382	100	2012-08-02 05:31:29.302338
24945	4	1-6877	24942	57	96	1371	17	2012-08-02 05:31:29.302338
24942	4	1-7290	24942	96	90	2920	69	2012-08-02 05:31:29.302338
24947	3	1-5945	24946	72	76	3522	53	2012-08-02 05:31:29.689535
24948	3	1-7187	24946	76	24	977	8	2012-08-02 05:31:29.689535
24946	3	1-7291	24946	24	72	9349	99	2012-08-02 05:31:29.689535
24950	5	1-6829	24949	16	77	2842	31	2012-08-02 05:31:30.087054
24951	5	1-7274	24949	77	34	1618	5	2012-08-02 05:31:30.087054
24952	5	1-7005	24949	34	10	1214	62	2012-08-02 05:31:30.087054
24953	5	1-7037	24949	10	4	1518	16	2012-08-02 05:31:30.087054
24949	5	1-7292	24949	4	16	3423	36	2012-08-02 05:31:30.087054
24955	4	1-4374	24954	52	4	640	16	2012-08-02 05:31:30.473817
24956	4	1-7292	24954	4	36	1319	36	2012-08-02 05:31:30.473817
24957	4	1-6686	24954	36	97	1726	57	2012-08-02 05:31:30.473817
24954	4	1-7293	24954	97	52	1316	86	2012-08-02 05:31:30.473817
24959	5	1-4374	24958	52	4	1144	16	2012-08-02 05:31:30.473817
24960	5	1-7292	24958	4	37	2374	36	2012-08-02 05:31:30.473817
24961	5	1-7184	24958	37	7	1344	67	2012-08-02 05:31:30.473817
24962	5	1-7012	24958	7	97	3042	57	2012-08-02 05:31:30.473817
24958	5	1-7293	24958	97	52	2335	86	2012-08-02 05:31:30.473817
24964	3	1-6799	24963	41	11	520	25	2012-08-02 05:31:32.053642
24965	3	1-6932	24963	11	8	525	73	2012-08-02 05:31:32.053642
24963	3	1-7297	24963	8	41	629	76	2012-08-02 05:31:32.053642
24967	5	1-7231	24966	13	89	1195	100	2012-08-02 05:31:32.815957
24968	5	1-7111	24966	89	66	889	27	2012-08-02 05:31:32.815957
24969	5	1-5736	24966	66	24	460	39	2012-08-02 05:31:32.815957
24970	5	1-7085	24966	24	55	375	96	2012-08-02 05:31:32.815957
24966	5	1-7299	24966	55	13	1127	95	2012-08-02 05:31:32.815957
24972	5	1-7231	24971	13	72	796	100	2012-08-02 05:31:32.815957
24973	5	1-5246	24971	72	43	314	11	2012-08-02 05:31:32.815957
24974	5	1-7281	24971	43	24	252	39	2012-08-02 05:31:32.815957
24975	5	1-7085	24971	24	55	219	96	2012-08-02 05:31:32.815957
24971	5	1-7299	24971	55	13	703	95	2012-08-02 05:31:32.815957
24977	4	1-5302	24976	51	73	3155	58	2012-08-02 05:31:33.213673
24978	4	1-7271	24976	73	92	1398	74	2012-08-02 05:31:33.213673
24979	4	1-6875	24976	92	13	396	12	2012-08-02 05:31:33.213673
24976	4	1-7300	24976	13	51	8319	41	2012-08-02 05:31:33.213673
24981	3	1-7068	24980	1	75	1478	100	2012-08-02 05:31:34.771471
24982	3	1-7100	24980	75	26	1034	33	2012-08-02 05:31:34.771471
24980	3	1-7304	24980	26	1	3065	23	2012-08-02 05:31:34.771471
24984	5	1-6487	24983	72	23	417	43	2012-08-02 05:31:35.566866
24985	5	1-6491	24983	23	4	399	16	2012-08-02 05:31:35.566866
24986	5	1-7292	24983	4	22	893	36	2012-08-02 05:31:35.566866
24987	5	1-6982	24983	22	62	254	18	2012-08-02 05:31:35.566866
24983	5	1-7306	24983	62	72	432	94	2012-08-02 05:31:35.566866
24989	4	1-7212	24988	40	30	2000	77	2012-08-02 05:31:35.975855
24990	4	1-6492	24988	30	76	3007	88	2012-08-02 05:31:35.975855
24991	4	1-6489	24988	76	14	1464	67	2012-08-02 05:31:35.975855
24988	4	1-7307	24988	14	40	1407	58	2012-08-02 05:31:35.975855
24993	3	1-7229	24992	34	45	1663	75	2012-08-02 05:31:37.953234
24994	3	1-1892	24992	45	14	1154	100	2012-08-02 05:31:37.953234
24992	3	1-7312	24992	14	34	3969	44	2012-08-02 05:31:37.953234
24996	4	1-5777	24995	23	4	140	16	2012-08-02 05:31:38.351022
24997	4	1-7292	24995	4	22	255	36	2012-08-02 05:31:38.351022
24998	4	1-6982	24995	22	87	59	18	2012-08-02 05:31:38.351022
24995	4	1-7313	24995	87	23	225	6	2012-08-02 05:31:38.351022
25000	2	1-7167	24999	54	97	5690	91	2012-08-02 05:31:40.726436
24999	2	1-7319	24999	97	54	5830	41	2012-08-02 05:31:40.726436
25002	2	1-5790	25001	30	72	5995	37	2012-08-02 05:31:41.499403
25001	2	1-7321	25001	72	30	4516	75	2012-08-02 05:31:41.499403
25004	3	1-7016	25003	16	68	3051	61	2012-08-02 05:31:41.886097
25005	3	1-6276	25003	68	82	2349	42	2012-08-02 05:31:41.886097
25003	3	1-7322	25003	82	16	6392	16	2012-08-02 05:31:41.886097
25007	5	1-7182	25006	58	9	3599	30	2012-08-02 05:31:42.670479
25008	5	1-7020	25006	9	59	3356	82	2012-08-02 05:31:42.670479
25009	5	1-5319	25006	59	93	2125	70	2012-08-02 05:31:42.670479
25010	5	1-7095	25006	93	69	2611	59	2012-08-02 05:31:42.670479
25006	5	1-7324	25006	69	58	4835	27	2012-08-02 05:31:42.670479
25012	4	1-7182	25011	58	17	844	30	2012-08-02 05:31:42.670479
25013	4	1-6844	25011	17	27	874	25	2012-08-02 05:31:42.670479
25014	4	1-4564	25011	27	69	565	59	2012-08-02 05:31:42.670479
25011	4	1-7324	25011	69	58	1089	27	2012-08-02 05:31:42.670479
25016	5	1-5736	25015	66	7	895	39	2012-08-02 05:31:42.670479
25017	5	1-6628	25015	7	4	795	60	2012-08-02 05:31:42.670479
25018	5	1-6060	25015	4	93	635	70	2012-08-02 05:31:42.670479
25019	5	1-7095	25015	93	69	877	59	2012-08-02 05:31:42.670479
25015	5	1-7324	25015	69	66	1826	27	2012-08-02 05:31:42.670479
25021	4	1-1209	25020	56	84	386	80	2012-08-02 05:31:44.217285
25022	4	1-6459	25020	84	96	95	11	2012-08-02 05:31:44.217285
25023	4	1-7284	25020	96	4	116	79	2012-08-02 05:31:44.217285
25020	4	1-7328	25020	4	56	1736	60	2012-08-02 05:31:44.217285
25025	4	1-4552	25024	38	64	79	30	2012-08-02 05:31:44.217285
25026	4	1-6212	25024	64	96	35	11	2012-08-02 05:31:44.217285
25027	4	1-7284	25024	96	4	48	79	2012-08-02 05:31:44.217285
25024	4	1-7328	25024	4	38	803	60	2012-08-02 05:31:44.217285
25029	4	1-6261	25028	75	81	1729	76	2012-08-02 05:31:44.604425
25030	4	1-7024	25028	81	13	1083	95	2012-08-02 05:31:44.604425
25031	4	1-7231	25028	13	25	1252	100	2012-08-02 05:31:44.604425
25028	4	1-7329	25028	25	75	2513	90	2012-08-02 05:31:44.604425
25033	3	1-7167	25032	54	82	1797	91	2012-08-02 05:31:45.001852
25034	3	1-7216	25032	82	10	1948	26	2012-08-02 05:31:45.001852
25032	3	1-7330	25032	10	54	1240	41	2012-08-02 05:31:45.001852
25036	3	1-7085	25035	24	42	777	96	2012-08-02 05:31:46.990396
25037	3	1-5698	25035	42	73	833	50	2012-08-02 05:31:46.990396
25035	3	1-7335	25035	73	24	808	39	2012-08-02 05:31:46.990396
25039	4	1-7216	25038	82	75	4523	26	2012-08-02 05:31:48.172439
25040	4	1-6725	25038	75	8	4690	68	2012-08-02 05:31:48.172439
25041	4	1-6919	25038	8	81	3354	55	2012-08-02 05:31:48.172439
25038	4	1-7338	25038	81	82	5583	91	2012-08-02 05:31:48.172439
25043	4	1-7097	25042	21	48	3578	71	2012-08-02 05:31:48.172439
25044	4	1-6645	25042	48	10	1450	20	2012-08-02 05:31:48.172439
25045	4	1-6936	25042	10	81	1617	55	2012-08-02 05:31:48.172439
25042	4	1-7338	25042	81	21	3013	91	2012-08-02 05:31:48.172439
25047	4	1-5834	25046	11	32	207	51	2012-08-02 05:31:48.172439
25048	4	1-7161	25046	32	8	116	68	2012-08-02 05:31:48.172439
25049	4	1-6919	25046	8	81	99	55	2012-08-02 05:31:48.172439
25046	4	1-7338	25046	81	11	196	91	2012-08-02 05:31:48.172439
25051	3	1-6670	25050	63	32	22	51	2012-08-02 05:31:48.581492
25052	3	1-7161	25050	32	17	10	68	2012-08-02 05:31:48.581492
25050	3	1-7339	25050	17	63	28	8	2012-08-02 05:31:48.581492
25054	5	1-2876	25053	57	28	1996	70	2012-08-02 05:31:48.581492
25055	5	1-6817	25053	28	32	990	97	2012-08-02 05:31:48.581492
25056	5	1-7057	25053	32	28	1612	62	2012-08-02 05:31:48.581492
25057	5	1-7169	25053	28	17	1304	68	2012-08-02 05:31:48.581492
25053	5	1-7339	25053	17	57	3856	8	2012-08-02 05:31:48.581492
25059	5	1-7155	25058	87	96	1012	11	2012-08-02 05:31:48.581492
25060	5	1-7284	25058	96	78	2737	79	2012-08-02 05:31:48.581492
25061	5	1-5618	25058	78	28	686	62	2012-08-02 05:31:48.581492
25062	5	1-7169	25058	28	17	516	68	2012-08-02 05:31:48.581492
25058	5	1-7339	25058	17	87	1665	8	2012-08-02 05:31:48.581492
25064	5	1-5397	25063	24	34	892	35	2012-08-02 05:31:48.581492
25065	5	1-6684	25063	34	10	1105	48	2012-08-02 05:31:48.581492
25066	5	1-7112	25063	10	28	655	62	2012-08-02 05:31:48.581492
25067	5	1-7169	25063	28	17	518	68	2012-08-02 05:31:48.581492
25063	5	1-7339	25063	17	24	1753	8	2012-08-02 05:31:48.581492
25069	4	1-3233	25068	65	72	771	37	2012-08-02 05:31:48.581492
25070	4	1-7321	25068	72	91	520	75	2012-08-02 05:31:48.581492
25071	4	1-6924	25068	91	17	262	68	2012-08-02 05:31:48.581492
25068	4	1-7339	25068	17	65	945	8	2012-08-02 05:31:48.581492
25073	4	1-5870	25072	94	23	521	50	2012-08-02 05:31:49.001455
25074	4	1-6093	25072	23	45	230	4	2012-08-02 05:31:49.001455
25075	4	1-6133	25072	45	28	122	95	2012-08-02 05:31:49.001455
25072	4	1-7340	25072	28	94	1308	87	2012-08-02 05:31:49.001455
25077	4	1-5999	25076	77	7	3031	31	2012-08-02 05:31:49.001455
25078	4	1-7326	25076	7	36	3019	72	2012-08-02 05:31:49.001455
25079	4	1-6756	25076	36	28	541	95	2012-08-02 05:31:49.001455
25076	4	1-7340	25076	28	77	6039	87	2012-08-02 05:31:49.001455
25081	2	1-6232	25080	94	68	679	100	2012-08-02 05:31:49.377264
25080	2	1-7341	25080	68	94	593	18	2012-08-02 05:31:49.377264
25083	5	1-7306	25082	62	92	439	94	2012-08-02 05:31:49.377264
25084	5	1-7318	25082	92	56	325	71	2012-08-02 05:31:49.377264
25085	5	1-6348	25082	56	13	386	95	2012-08-02 05:31:49.377264
25086	5	1-7231	25082	13	68	356	100	2012-08-02 05:31:49.377264
25082	5	1-7341	25082	68	62	346	18	2012-08-02 05:31:49.377264
25088	4	1-7030	25087	75	56	1961	71	2012-08-02 05:31:49.377264
25089	4	1-6348	25087	56	13	2389	95	2012-08-02 05:31:49.377264
25090	4	1-7231	25087	13	68	2256	100	2012-08-02 05:31:49.377264
25087	4	1-7341	25087	68	75	2248	18	2012-08-02 05:31:49.377264
25092	4	1-7313	25091	87	36	2355	6	2012-08-02 05:31:49.377264
25093	4	1-6678	25091	36	45	602	75	2012-08-02 05:31:49.377264
25094	4	1-1892	25091	45	68	432	100	2012-08-02 05:31:49.377264
25091	4	1-7341	25091	68	87	533	18	2012-08-02 05:31:49.377264
25096	5	1-6425	25095	25	54	133	46	2012-08-02 05:31:50.581363
25097	5	1-7238	25095	54	99	90	78	2012-08-02 05:31:50.581363
25098	5	1-7072	25095	99	4	118	60	2012-08-02 05:31:50.581363
25099	5	1-6060	25095	4	21	90	70	2012-08-02 05:31:50.581363
25095	5	1-7344	25095	21	25	219	43	2012-08-02 05:31:50.581363
25101	5	1-5595	25100	79	8	792	29	2012-08-02 05:31:50.581363
25102	5	1-7280	25100	8	46	1355	41	2012-08-02 05:31:50.581363
25103	5	1-5741	25100	46	56	493	20	2012-08-02 05:31:50.581363
25104	5	1-7247	25100	56	21	402	70	2012-08-02 05:31:50.581363
25100	5	1-7344	25100	21	79	1040	43	2012-08-02 05:31:50.581363
25106	4	1-7231	25105	13	25	2496	100	2012-08-02 05:31:50.979393
25107	4	1-7329	25105	25	4	4453	90	2012-08-02 05:31:50.979393
25108	4	1-5165	25105	4	27	2513	31	2012-08-02 05:31:50.979393
25105	4	1-7345	25105	27	13	2429	95	2012-08-02 05:31:50.979393
25110	4	1-7231	25109	13	51	955	100	2012-08-02 05:31:50.979393
25111	4	1-6988	25109	51	48	602	7	2012-08-02 05:31:50.979393
25112	4	1-7239	25109	48	27	742	31	2012-08-02 05:31:50.979393
25109	4	1-7345	25109	27	13	816	95	2012-08-02 05:31:50.979393
25114	4	1-6985	25113	10	7	634	67	2012-08-02 05:31:51.366117
25115	4	1-7012	25113	7	90	1533	57	2012-08-02 05:31:51.366117
25116	4	1-7278	25113	90	39	926	7	2012-08-02 05:31:51.366117
25113	4	1-7346	25113	39	10	873	2	2012-08-02 05:31:51.366117
25118	4	1-6414	25117	43	91	1920	74	2012-08-02 05:31:52.13936
25119	4	1-6946	25117	91	95	654	69	2012-08-02 05:31:52.13936
25120	4	1-6312	25117	95	41	394	26	2012-08-02 05:31:52.13936
25117	4	1-7348	25117	41	43	6032	95	2012-08-02 05:31:52.13936
25122	4	1-4009	25121	68	13	1923	94	2012-08-02 05:31:52.948118
25123	4	1-6738	25121	13	28	1766	82	2012-08-02 05:31:52.948118
25124	4	1-6950	25121	28	86	3852	57	2012-08-02 05:31:52.948118
25121	4	1-7350	25121	86	68	4883	26	2012-08-02 05:31:52.948118
25126	4	1-7321	25125	72	75	415	75	2012-08-02 05:31:56.836763
25127	4	1-5229	25125	75	57	170	1	2012-08-02 05:31:56.836763
25128	4	1-3689	25125	57	14	80	58	2012-08-02 05:31:56.836763
25125	4	1-7360	25125	14	72	922	37	2012-08-02 05:31:56.836763
25130	2	1-1723	25129	30	14	543	58	2012-08-02 05:31:56.836763
25129	2	1-7360	25129	14	30	7042	37	2012-08-02 05:31:56.836763
25132	3	1-5231	25131	19	47	139	72	2012-08-02 05:31:57.234403
25133	3	1-2420	25131	47	4	40	23	2012-08-02 05:31:57.234403
25131	3	1-7361	25131	4	19	217	98	2012-08-02 05:31:57.234403
25135	5	1-7117	25134	47	76	339	53	2012-08-02 05:31:57.234403
25136	5	1-7187	25134	76	95	285	8	2012-08-02 05:31:57.234403
25137	5	1-2501	25134	95	45	181	83	2012-08-02 05:31:57.234403
25138	5	1-7086	25134	45	4	65	23	2012-08-02 05:31:57.234403
25134	5	1-7361	25134	4	47	361	98	2012-08-02 05:31:57.234403
25140	2	1-3773	25139	88	4	1634	23	2012-08-02 05:31:57.234403
25139	2	1-7361	25139	4	88	9370	98	2012-08-02 05:31:57.234403
25142	5	1-5708	25141	51	7	539	39	2012-08-02 05:31:59.211883
25143	5	1-6810	25141	7	90	608	57	2012-08-02 05:31:59.211883
25144	5	1-7278	25141	90	48	333	7	2012-08-02 05:31:59.211883
25145	5	1-7239	25141	48	13	388	31	2012-08-02 05:31:59.211883
25141	5	1-7366	25141	13	51	891	64	2012-08-02 05:31:59.211883
25147	4	1-3246	25146	24	10	158	11	2012-08-02 05:31:59.211883
25148	4	1-6564	25146	10	77	147	87	2012-08-02 05:31:59.211883
25149	4	1-5999	25146	77	13	117	31	2012-08-02 05:31:59.211883
25146	4	1-7366	25146	13	24	269	64	2012-08-02 05:31:59.211883
25151	4	1-1658	25150	53	64	4885	37	2012-08-02 05:31:59.211883
25152	4	1-6762	25150	64	48	2044	7	2012-08-02 05:31:59.211883
25153	4	1-7239	25150	48	13	2455	31	2012-08-02 05:31:59.211883
25150	4	1-7366	25150	13	53	5800	64	2012-08-02 05:31:59.211883
25155	4	1-5708	25154	51	81	460	39	2012-08-02 05:31:59.211883
25156	4	1-6927	25154	81	79	414	61	2012-08-02 05:31:59.211883
25157	4	1-6424	25154	79	13	311	31	2012-08-02 05:31:59.211883
25154	4	1-7366	25154	13	51	737	64	2012-08-02 05:31:59.211883
25159	2	1-6560	25158	89	27	2565	100	2012-08-02 05:32:00.007501
25158	2	1-7368	25158	27	89	9714	47	2012-08-02 05:32:00.007501
25161	4	1-7186	25160	75	96	2459	75	2012-08-02 05:32:00.405367
25162	4	1-3889	25160	96	21	2060	70	2012-08-02 05:32:00.405367
25163	4	1-7344	25160	21	70	5584	43	2012-08-02 05:32:00.405367
25160	4	1-7369	25160	70	75	2724	49	2012-08-02 05:32:00.405367
25165	3	1-6745	25164	14	17	2154	36	2012-08-02 05:32:00.814102
25166	3	1-306	25164	17	31	485	23	2012-08-02 05:32:00.814102
25164	3	1-7370	25164	31	14	1952	83	2012-08-02 05:32:00.814102
25168	3	1-5876	25167	79	70	73	19	2012-08-02 05:32:01.598436
25169	3	1-3428	25167	70	62	40	70	2012-08-02 05:32:01.598436
25167	3	1-7372	25167	62	79	450	21	2012-08-02 05:32:01.598436
25171	5	1-4610	25170	91	47	1094	80	2012-08-02 05:32:01.598436
25172	5	1-6069	25170	47	34	826	44	2012-08-02 05:32:01.598436
25173	5	1-7229	25170	34	96	310	75	2012-08-02 05:32:01.598436
25174	5	1-3889	25170	96	62	244	70	2012-08-02 05:32:01.598436
25170	5	1-7372	25170	62	91	2937	21	2012-08-02 05:32:01.598436
25176	2	1-5962	25175	2	12	4257	10	2012-08-02 05:32:01.985282
25175	2	1-7373	25175	12	2	8939	96	2012-08-02 05:32:01.985282
25178	5	1-409	25177	67	7	2433	31	2012-08-02 05:32:03.189366
25179	5	1-7326	25177	7	67	3847	72	2012-08-02 05:32:03.189366
25180	5	1-4955	25177	67	64	1086	19	2012-08-02 05:32:03.189366
25181	5	1-6442	25177	64	2	1409	21	2012-08-02 05:32:03.189366
25177	5	1-7376	25177	2	67	3063	82	2012-08-02 05:32:03.189366
25183	5	1-6766	25182	41	8	1782	29	2012-08-02 05:32:04.36045
25184	5	1-7280	25182	8	60	2876	41	2012-08-02 05:32:04.36045
25185	5	1-6952	25182	60	54	2393	46	2012-08-02 05:32:04.36045
25186	5	1-7238	25182	54	78	1621	78	2012-08-02 05:32:04.36045
25182	5	1-7379	25182	78	41	2871	86	2012-08-02 05:32:04.36045
25188	4	1-7145	25187	15	25	3666	43	2012-08-02 05:32:04.36045
25189	4	1-6425	25187	25	54	2285	46	2012-08-02 05:32:04.36045
25190	4	1-7238	25187	54	78	1584	78	2012-08-02 05:32:04.36045
25187	4	1-7379	25187	78	15	2872	86	2012-08-02 05:32:04.36045
25192	2	1-6328	25191	100	35	713	92	2012-08-02 05:32:05.16733
25191	2	1-7381	25191	35	100	2368	36	2012-08-02 05:32:05.16733
25194	3	1-5084	25193	100	6	282	56	2012-08-02 05:32:05.16733
25195	3	1-7275	25193	6	35	675	92	2012-08-02 05:32:05.16733
25193	3	1-7381	25193	35	100	2461	36	2012-08-02 05:32:05.16733
25197	4	1-5229	25196	75	58	2414	1	2012-08-02 05:32:05.554516
25198	4	1-7156	25196	58	3	3357	38	2012-08-02 05:32:05.554516
25199	4	1-3487	25196	3	67	3049	46	2012-08-02 05:32:05.554516
25196	4	1-7382	25196	67	75	3690	75	2012-08-02 05:32:05.554516
25201	5	1-6905	25200	49	79	892	90	2012-08-02 05:32:05.554516
25202	5	1-6549	25200	79	74	885	98	2012-08-02 05:32:05.554516
25203	5	1-4383	25200	74	85	840	58	2012-08-02 05:32:05.554516
25204	5	1-6767	25200	85	67	988	46	2012-08-02 05:32:05.554516
25200	5	1-7382	25200	67	49	1204	75	2012-08-02 05:32:05.554516
25206	4	1-3665	25205	77	35	240	45	2012-08-02 05:32:06.726205
25207	4	1-7032	25205	35	47	675	80	2012-08-02 05:32:06.726205
25208	4	1-6069	25205	47	55	567	44	2012-08-02 05:32:06.726205
25205	4	1-7385	25205	55	77	553	6	2012-08-02 05:32:06.726205
25210	5	1-3665	25209	77	21	2617	45	2012-08-02 05:32:06.726205
25211	5	1-7294	25209	21	31	5268	97	2012-08-02 05:32:06.726205
25212	5	1-5389	25209	31	8	4261	79	2012-08-02 05:32:06.726205
25213	5	1-7174	25209	8	55	5454	44	2012-08-02 05:32:06.726205
25209	5	1-7385	25209	55	77	5656	6	2012-08-02 05:32:06.726205
25215	4	1-3665	25214	77	100	767	45	2012-08-02 05:32:06.726205
25216	4	1-6412	25214	100	8	1100	79	2012-08-02 05:32:06.726205
25217	4	1-7174	25214	8	55	1469	44	2012-08-02 05:32:06.726205
25214	4	1-7385	25214	55	77	1589	6	2012-08-02 05:32:06.726205
25219	3	1-7308	25218	57	82	261	39	2012-08-02 05:32:07.576776
25220	3	1-6785	25218	82	59	586	10	2012-08-02 05:32:07.576776
25218	3	1-7387	25218	59	57	4032	34	2012-08-02 05:32:07.576776
25222	3	1-5999	25221	77	7	669	31	2012-08-02 05:32:08.79185
25223	3	1-7326	25221	7	97	433	72	2012-08-02 05:32:08.79185
25221	3	1-7390	25221	97	77	2052	87	2012-08-02 05:32:08.79185
25225	3	1-2490	25224	87	7	170	31	2012-08-02 05:32:08.79185
25226	3	1-7326	25224	7	97	137	72	2012-08-02 05:32:08.79185
25224	3	1-7390	25224	97	87	808	87	2012-08-02 05:32:08.79185
25228	4	1-2490	25227	87	13	164	31	2012-08-02 05:32:08.79185
25229	4	1-7366	25227	13	48	235	64	2012-08-02 05:32:08.79185
25230	4	1-6562	25227	48	97	89	72	2012-08-02 05:32:08.79185
25227	4	1-7390	25227	97	87	640	87	2012-08-02 05:32:08.79185
25232	5	1-5185	25231	39	100	1312	28	2012-08-02 05:32:09.211983
25233	5	1-7259	25231	100	13	1696	12	2012-08-02 05:32:09.211983
25234	5	1-6252	25231	13	87	1396	97	2012-08-02 05:32:09.211983
25235	5	1-7316	25231	87	47	3059	48	2012-08-02 05:32:09.211983
25231	5	1-7391	25231	47	39	2190	86	2012-08-02 05:32:09.211983
25237	3	1-7362	25236	75	80	4306	35	2012-08-02 05:32:11.686392
25238	3	1-7380	25236	80	82	2028	99	2012-08-02 05:32:11.686392
25236	3	1-7397	25236	82	75	2597	23	2012-08-02 05:32:11.686392
25240	4	1-5898	25239	100	100	2576	28	2012-08-02 05:32:11.686392
25241	4	1-7259	25239	100	4	3479	12	2012-08-02 05:32:11.686392
25242	4	1-6109	25239	4	82	1146	99	2012-08-02 05:32:11.686392
25239	4	1-7397	25239	82	100	2273	23	2012-08-02 05:32:11.686392
25244	4	1-7391	25243	47	15	1888	86	2012-08-02 05:32:12.11733
25245	4	1-7145	25243	15	95	2740	43	2012-08-02 05:32:12.11733
25246	4	1-5720	25243	95	80	2941	79	2012-08-02 05:32:12.11733
25243	4	1-7398	25243	80	47	2822	48	2012-08-02 05:32:12.11733
25248	4	1-7259	25247	100	13	2038	12	2012-08-02 05:32:12.548499
25249	4	1-6252	25247	13	32	1115	97	2012-08-02 05:32:12.548499
25250	4	1-7057	25247	32	44	1588	62	2012-08-02 05:32:12.548499
25247	4	1-7399	25247	44	100	2370	28	2012-08-02 05:32:12.548499
25252	5	1-7223	25251	95	93	1711	59	2012-08-02 05:32:12.548499
25253	5	1-6061	25251	93	21	697	45	2012-08-02 05:32:12.548499
25254	5	1-7294	25251	21	32	1246	97	2012-08-02 05:32:12.548499
25255	5	1-7057	25251	32	44	2351	62	2012-08-02 05:32:12.548499
25251	5	1-7399	25251	44	95	4650	28	2012-08-02 05:32:12.548499
25257	5	1-6204	25256	22	86	167	57	2012-08-02 05:32:12.548499
25258	5	1-7350	25256	86	68	190	26	2012-08-02 05:32:12.548499
25259	5	1-4009	25256	68	77	67	94	2012-08-02 05:32:12.548499
25260	5	1-7088	25256	77	44	115	62	2012-08-02 05:32:12.548499
25256	5	1-7399	25256	44	22	237	28	2012-08-02 05:32:12.548499
25262	4	1-7112	25261	10	98	185	62	2012-08-02 05:32:13.376932
25263	4	1-6739	25261	98	78	54	86	2012-08-02 05:32:13.376932
25264	4	1-7287	25261	78	29	35	34	2012-08-02 05:32:13.376932
25261	4	1-7401	25261	29	10	656	48	2012-08-02 05:32:13.376932
25266	3	1-2737	25265	55	78	250	86	2012-08-02 05:32:13.376932
25267	3	1-7287	25265	78	29	185	34	2012-08-02 05:32:13.376932
25265	3	1-7401	25265	29	55	4019	48	2012-08-02 05:32:13.376932
25269	4	1-7353	25268	90	90	2442	33	2012-08-02 05:32:13.785844
25270	4	1-7151	25268	90	63	1251	8	2012-08-02 05:32:13.785844
25271	4	1-6670	25268	63	93	734	51	2012-08-02 05:32:13.785844
25268	4	1-7402	25268	93	90	3716	70	2012-08-02 05:32:13.785844
25273	4	1-6817	25272	28	94	1987	97	2012-08-02 05:32:13.785844
25274	4	1-6784	25272	94	95	549	19	2012-08-02 05:32:13.785844
25275	4	1-5626	25272	95	93	799	51	2012-08-02 05:32:13.785844
25272	4	1-7402	25272	93	28	5301	70	2012-08-02 05:32:13.785844
25277	5	1-7294	25276	21	88	972	97	2012-08-02 05:32:15.023199
25278	5	1-5548	25276	88	67	318	3	2012-08-02 05:32:15.023199
25279	5	1-7395	25276	67	33	312	25	2012-08-02 05:32:15.023199
25280	5	1-7225	25276	33	36	223	26	2012-08-02 05:32:15.023199
25276	5	1-7405	25276	36	21	670	45	2012-08-02 05:32:15.023199
25282	4	1-7148	25281	62	89	1042	93	2012-08-02 05:32:15.023199
25283	4	1-3329	25281	89	33	201	25	2012-08-02 05:32:15.023199
25284	4	1-7225	25281	33	36	162	26	2012-08-02 05:32:15.023199
25281	4	1-7405	25281	36	62	546	45	2012-08-02 05:32:15.023199
25286	5	1-6765	25285	99	23	2713	50	2012-08-02 05:32:15.023199
25287	5	1-6093	25285	23	58	1903	4	2012-08-02 05:32:15.023199
25288	5	1-7234	25285	58	33	674	25	2012-08-02 05:32:15.023199
25289	5	1-7225	25285	33	36	598	26	2012-08-02 05:32:15.023199
25285	5	1-7405	25285	36	99	2235	45	2012-08-02 05:32:15.023199
25291	5	1-6765	25290	99	77	2754	50	2012-08-02 05:32:15.023199
25292	5	1-6643	25290	77	6	2670	73	2012-08-02 05:32:15.023199
25293	5	1-6910	25290	6	82	558	91	2012-08-02 05:32:15.023199
25294	5	1-7216	25290	82	36	578	26	2012-08-02 05:32:15.023199
25290	5	1-7405	25290	36	99	2214	45	2012-08-02 05:32:15.023199
25296	4	1-6765	25295	99	25	213	50	2012-08-02 05:32:15.023199
25297	4	1-3651	25295	25	33	46	25	2012-08-02 05:32:15.023199
25298	4	1-7225	25295	33	36	43	26	2012-08-02 05:32:15.023199
25295	4	1-7405	25295	36	99	167	45	2012-08-02 05:32:15.023199
25300	4	1-5118	25299	84	42	849	14	2012-08-02 05:32:15.023199
25301	4	1-7129	25299	42	90	188	19	2012-08-02 05:32:15.023199
25302	4	1-4187	25299	90	36	156	26	2012-08-02 05:32:15.023199
25299	4	1-7405	25299	36	84	680	45	2012-08-02 05:32:15.023199
25304	5	1-6790	25303	34	84	161	80	2012-08-02 05:32:15.874383
25305	5	1-6459	25303	84	43	106	11	2012-08-02 05:32:15.874383
25306	5	1-7281	25303	43	82	91	39	2012-08-02 05:32:15.874383
25307	5	1-6785	25303	82	3	342	10	2012-08-02 05:32:15.874383
25303	5	1-7407	25303	3	34	142	26	2012-08-02 05:32:15.874383
25309	3	1-7158	25308	83	26	3810	21	2012-08-02 05:32:16.282869
25310	3	1-5886	25308	26	93	775	27	2012-08-02 05:32:16.282869
25308	3	1-7408	25308	93	83	7708	88	2012-08-02 05:32:16.282869
25312	5	1-7237	25311	36	57	4623	57	2012-08-02 05:32:17.122418
25313	5	1-7388	25311	57	8	4372	73	2012-08-02 05:32:17.122418
25314	5	1-7297	25311	8	1	4464	76	2012-08-02 05:32:17.122418
25315	5	1-7162	25311	1	75	1774	39	2012-08-02 05:32:17.122418
25311	5	1-7410	25311	75	36	4852	24	2012-08-02 05:32:17.122418
25317	3	1-6939	25316	64	1	1075	76	2012-08-02 05:32:17.122418
25318	3	1-7162	25316	1	75	457	39	2012-08-02 05:32:17.122418
25316	3	1-7410	25316	75	64	1338	24	2012-08-02 05:32:17.122418
25320	4	1-6573	25319	88	5	1927	84	2012-08-02 05:32:17.122418
25321	4	1-2625	25319	5	43	316	11	2012-08-02 05:32:17.122418
25322	4	1-7281	25319	43	75	259	39	2012-08-02 05:32:17.122418
25319	4	1-7410	25319	75	88	775	24	2012-08-02 05:32:17.122418
25324	5	1-6574	25323	85	92	79	56	2012-08-02 05:32:17.122418
25325	5	1-4773	25323	92	64	31	30	2012-08-02 05:32:17.122418
25326	5	1-6212	25323	64	43	34	11	2012-08-02 05:32:17.122418
25327	5	1-7281	25323	43	75	30	39	2012-08-02 05:32:17.122418
25323	5	1-7410	25323	75	85	95	24	2012-08-02 05:32:17.122418
25329	5	1-6323	25328	22	80	28	35	2012-08-02 05:32:17.973282
25330	5	1-7380	25328	80	82	20	99	2012-08-02 05:32:17.973282
25331	5	1-7397	25328	82	33	38	23	2012-08-02 05:32:17.973282
25332	5	1-7041	25328	33	27	62	8	2012-08-02 05:32:17.973282
25328	5	1-7412	25328	27	22	84	12	2012-08-02 05:32:17.973282
25334	5	1-7208	25333	69	82	514	99	2012-08-02 05:32:18.382144
25335	5	1-7397	25333	82	69	1007	23	2012-08-02 05:32:18.382144
25336	5	1-3745	25333	69	67	830	46	2012-08-02 05:32:18.382144
25337	5	1-7382	25333	67	64	1051	75	2012-08-02 05:32:18.382144
25333	5	1-7413	25333	64	69	710	61	2012-08-02 05:32:18.382144
25339	4	1-7158	25338	83	2	2205	21	2012-08-02 05:32:19.232845
25340	4	1-4540	25338	2	17	873	2	2012-08-02 05:32:19.232845
25341	4	1-7389	25338	17	42	4530	40	2012-08-02 05:32:19.232845
25338	4	1-7415	25338	42	83	3179	88	2012-08-02 05:32:19.232845
25343	5	1-7041	25342	33	27	356	8	2012-08-02 05:32:19.654471
25344	5	1-7412	25342	27	98	440	12	2012-08-02 05:32:19.654471
25345	5	1-4987	25342	98	71	217	30	2012-08-02 05:32:19.654471
25346	5	1-4497	25342	71	45	139	27	2012-08-02 05:32:19.654471
25342	5	1-7416	25342	45	33	238	23	2012-08-02 05:32:19.654471
25348	3	1-5517	25347	76	25	681	9	2012-08-02 05:32:21.68727
25349	3	1-6902	25347	25	26	352	58	2012-08-02 05:32:21.68727
25347	3	1-7421	25347	26	76	5289	19	2012-08-02 05:32:21.68727
25351	3	1-7418	25350	43	74	336	98	2012-08-02 05:32:21.68727
25352	3	1-4383	25350	74	26	94	58	2012-08-02 05:32:21.68727
25350	3	1-7421	25350	26	43	1519	19	2012-08-02 05:32:21.68727
25354	4	1-6745	25353	14	100	629	36	2012-08-02 05:32:22.50512
25355	4	1-5084	25353	100	6	176	56	2012-08-02 05:32:22.50512
25356	4	1-7275	25353	6	6	1084	92	2012-08-02 05:32:22.50512
25353	4	1-7423	25353	6	14	601	83	2012-08-02 05:32:22.50512
25358	2	1-5651	25357	14	33	610	59	2012-08-02 05:32:29.044778
25357	2	1-7439	25357	33	14	2203	14	2012-08-02 05:32:29.044778
25360	4	1-7129	25359	42	99	68	19	2012-08-02 05:32:29.044778
25361	4	1-7007	25359	99	1	134	43	2012-08-02 05:32:29.044778
25362	4	1-6318	25359	1	33	70	59	2012-08-02 05:32:29.044778
25359	4	1-7439	25359	33	42	314	14	2012-08-02 05:32:29.044778
25364	4	1-7178	25363	95	70	1605	30	2012-08-02 05:32:30.702118
25365	4	1-7434	25363	70	95	2686	84	2012-08-02 05:32:30.702118
25366	4	1-7170	25363	95	15	2878	4	2012-08-02 05:32:30.702118
25363	4	1-7443	25363	15	95	3350	16	2012-08-02 05:32:30.702118
25368	4	1-7016	25367	16	71	780	61	2012-08-02 05:32:30.702118
25369	4	1-7327	25367	71	79	773	71	2012-08-02 05:32:30.702118
25370	4	1-7064	25367	79	15	1168	4	2012-08-02 05:32:30.702118
25367	4	1-7443	25367	15	16	1386	16	2012-08-02 05:32:30.702118
25372	5	1-4552	25371	38	17	1540	30	2012-08-02 05:32:31.519347
25373	5	1-6844	25371	17	33	1772	25	2012-08-02 05:32:31.519347
25374	5	1-7225	25371	33	52	1522	26	2012-08-02 05:32:31.519347
25375	5	1-1931	25371	52	75	1183	93	2012-08-02 05:32:31.519347
25371	5	1-7445	25371	75	38	7663	60	2012-08-02 05:32:31.519347
25377	5	1-2053	25376	26	1	379	23	2012-08-02 05:32:31.519347
25378	5	1-7068	25376	1	68	250	100	2012-08-02 05:32:31.519347
25379	5	1-7341	25376	68	33	286	18	2012-08-02 05:32:31.519347
25380	5	1-5979	25376	33	75	221	93	2012-08-02 05:32:31.519347
25376	5	1-7445	25376	75	26	1499	60	2012-08-02 05:32:31.519347
25382	5	1-6562	25381	48	97	258	72	2012-08-02 05:32:31.939309
25383	5	1-7390	25381	97	26	2899	87	2012-08-02 05:32:31.939309
25384	5	1-6769	25381	26	82	383	91	2012-08-02 05:32:31.939309
25385	5	1-7216	25381	82	33	391	26	2012-08-02 05:32:31.939309
25381	5	1-7446	25381	33	48	437	64	2012-08-02 05:32:31.939309
25387	4	1-7351	25386	56	46	907	58	2012-08-02 05:32:32.403289
25388	4	1-4452	25386	46	55	419	44	2012-08-02 05:32:32.403289
25389	4	1-7385	25386	55	93	413	6	2012-08-02 05:32:32.403289
25386	4	1-7447	25386	93	56	425	49	2012-08-02 05:32:32.403289
25391	5	1-6842	25390	92	27	194	8	2012-08-02 05:32:32.403289
25392	5	1-7412	25390	27	42	256	12	2012-08-02 05:32:32.403289
25393	5	1-6855	25390	42	55	269	44	2012-08-02 05:32:32.403289
25394	5	1-7385	25390	55	93	278	6	2012-08-02 05:32:32.403289
25390	5	1-7447	25390	93	92	301	49	2012-08-02 05:32:32.403289
25396	2	1-7130	25395	16	64	6560	50	2012-08-02 05:32:32.823161
25395	2	1-7448	25395	64	16	4743	24	2012-08-02 05:32:32.823161
25398	2	1-4105	25397	81	54	1460	87	2012-08-02 05:32:33.618614
25397	2	1-7450	25397	54	81	3127	26	2012-08-02 05:32:33.618614
25400	2	1-2842	25399	25	54	2005	87	2012-08-02 05:32:33.618614
25399	2	1-7450	25399	54	25	4631	26	2012-08-02 05:32:33.618614
25402	5	1-7446	25401	33	79	1610	64	2012-08-02 05:32:33.618614
25403	5	1-3161	25401	79	58	511	1	2012-08-02 05:32:33.618614
25404	5	1-7156	25401	58	39	570	38	2012-08-02 05:32:33.618614
25405	5	1-6579	25401	39	54	615	87	2012-08-02 05:32:33.618614
25401	5	1-7450	25401	54	33	1639	26	2012-08-02 05:32:33.618614
25407	5	1-7415	25406	42	83	3028	88	2012-08-02 05:32:34.833853
25408	5	1-7158	25406	83	26	2117	21	2012-08-02 05:32:34.833853
25409	5	1-5886	25406	26	66	609	27	2012-08-02 05:32:34.833853
25410	5	1-5736	25406	66	14	293	39	2012-08-02 05:32:34.833853
25406	5	1-7453	25406	14	42	4281	40	2012-08-02 05:32:34.833853
25412	2	1-5991	25411	39	29	2628	38	2012-08-02 05:32:35.253765
25411	2	1-7454	25411	29	39	6488	12	2012-08-02 05:32:35.253765
25414	5	1-6855	25413	42	47	442	44	2012-08-02 05:32:35.253765
25415	5	1-7358	25413	47	96	235	49	2012-08-02 05:32:35.253765
25416	5	1-6024	25413	96	58	125	1	2012-08-02 05:32:35.253765
25417	5	1-7156	25413	58	29	159	38	2012-08-02 05:32:35.253765
25413	5	1-7454	25413	29	42	464	12	2012-08-02 05:32:35.253765
25419	4	1-6262	25418	73	68	1324	32	2012-08-02 05:32:35.651589
25420	4	1-5078	25418	68	90	1342	69	2012-08-02 05:32:35.651589
25421	4	1-7103	25418	90	33	1759	52	2012-08-02 05:32:35.651589
25418	4	1-7455	25418	33	73	3517	2	2012-08-02 05:32:35.651589
25422	1	1-3558	\N	88	88	4902	2	2012-08-02 05:32:35.651589
25424	4	1-5905	25423	93	94	1234	97	2012-08-02 05:32:36.480263
25425	4	1-6784	25423	94	64	525	19	2012-08-02 05:32:36.480263
25426	4	1-6442	25423	64	97	763	21	2012-08-02 05:32:36.480263
25423	4	1-7457	25423	97	93	1593	76	2012-08-02 05:32:36.480263
25428	4	1-7393	25427	38	59	5001	71	2012-08-02 05:32:37.706616
25429	4	1-4450	25427	59	78	1164	78	2012-08-02 05:32:37.706616
25430	4	1-7379	25427	78	28	1190	86	2012-08-02 05:32:37.706616
25427	4	1-7460	25427	28	38	8719	35	2012-08-02 05:32:37.706616
25432	2	1-2712	25431	37	37	323	8	2012-08-02 05:32:38.535287
25431	2	1-7462	25431	37	37	9891	18	2012-08-02 05:32:38.535287
25434	4	1-7343	25433	8	45	996	27	2012-08-02 05:32:40.147852
25435	4	1-7416	25433	45	17	1168	23	2012-08-02 05:32:40.147852
25436	4	1-6781	25433	17	97	761	74	2012-08-02 05:32:40.147852
25433	4	1-7466	25433	97	8	5162	64	2012-08-02 05:32:40.147852
25438	4	1-7416	25437	45	13	3902	23	2012-08-02 05:32:42.644457
25439	4	1-7428	25437	13	80	3934	87	2012-08-02 05:32:42.644457
25440	4	1-5896	25437	80	30	1784	16	2012-08-02 05:32:42.644457
25437	4	1-7472	25437	30	45	2086	27	2012-08-02 05:32:42.644457
25442	4	1-5898	25441	100	47	2560	28	2012-08-02 05:32:43.053495
25443	4	1-6087	25441	47	54	1352	11	2012-08-02 05:32:43.053495
25444	4	1-7352	25441	54	8	2024	17	2012-08-02 05:32:43.053495
25441	4	1-7473	25441	8	100	3486	23	2012-08-02 05:32:43.053495
25446	4	1-7342	25445	38	47	2492	28	2012-08-02 05:32:43.053495
25447	4	1-6087	25445	47	54	1338	11	2012-08-02 05:32:43.053495
25448	4	1-7352	25445	54	8	2035	17	2012-08-02 05:32:43.053495
25445	4	1-7473	25445	8	38	3562	23	2012-08-02 05:32:43.053495
25450	3	1-6981	25449	12	96	1010	75	2012-08-02 05:32:44.677367
25451	3	1-3889	25449	96	49	642	70	2012-08-02 05:32:44.677367
25449	3	1-7477	25449	49	12	985	47	2012-08-02 05:32:44.677367
25453	4	1-5869	25452	18	67	1173	3	2012-08-02 05:32:45.903526
25454	4	1-7395	25452	67	11	1342	25	2012-08-02 05:32:45.903526
25455	4	1-6932	25452	11	81	1094	73	2012-08-02 05:32:45.903526
25452	4	1-7480	25452	81	18	2676	32	2012-08-02 05:32:45.903526
25457	5	1-4068	25456	60	9	359	30	2012-08-02 05:32:45.903526
25458	5	1-7020	25456	9	5	409	82	2012-08-02 05:32:45.903526
25459	5	1-2893	25456	5	57	443	57	2012-08-02 05:32:45.903526
25460	5	1-7388	25456	57	81	434	73	2012-08-02 05:32:45.903526
25456	5	1-7480	25456	81	60	1158	32	2012-08-02 05:32:45.903526
25462	5	1-5686	25461	97	74	896	38	2012-08-02 05:32:45.903526
25463	5	1-3476	25461	74	7	353	67	2012-08-02 05:32:45.903526
25464	5	1-7012	25461	7	57	780	57	2012-08-02 05:32:45.903526
25465	5	1-7388	25461	57	81	778	73	2012-08-02 05:32:45.903526
25461	5	1-7480	25461	81	97	2115	32	2012-08-02 05:32:45.903526
25467	5	1-6434	25466	36	51	279	33	2012-08-02 05:32:45.903526
25468	5	1-3524	25466	51	33	109	23	2012-08-02 05:32:45.903526
25469	5	1-5012	25466	33	57	115	57	2012-08-02 05:32:45.903526
25470	5	1-7388	25466	57	81	118	73	2012-08-02 05:32:45.903526
25466	5	1-7480	25466	81	36	332	32	2012-08-02 05:32:45.903526
25471	1	1-3524	\N	51	51	1768	23	2012-08-02 05:32:45.903526
25472	1	1-4108	\N	7	7	2321	23	2012-08-02 05:32:45.903526
25474	3	1-5756	25473	67	39	1725	38	2012-08-02 05:32:46.71036
25475	3	1-6579	25473	39	13	1637	87	2012-08-02 05:32:46.71036
25473	3	1-7482	25473	13	67	2261	18	2012-08-02 05:32:46.71036
25477	5	1-4620	25476	9	23	784	20	2012-08-02 05:32:46.71036
25478	5	1-7149	25476	23	58	692	1	2012-08-02 05:32:46.71036
25479	5	1-7156	25476	58	39	758	38	2012-08-02 05:32:46.71036
25480	5	1-6579	25476	39	13	804	87	2012-08-02 05:32:46.71036
25476	5	1-7482	25476	13	9	1240	18	2012-08-02 05:32:46.71036
25482	4	1-3677	25481	8	53	1038	25	2012-08-02 05:32:46.71036
25483	4	1-3386	25481	53	19	372	97	2012-08-02 05:32:46.71036
25484	4	1-7262	25481	19	13	485	87	2012-08-02 05:32:46.71036
25481	4	1-7482	25481	13	8	936	18	2012-08-02 05:32:46.71036
25486	4	1-5808	25485	40	65	1133	72	2012-08-02 05:32:46.71036
25487	4	1-7203	25485	65	48	2567	14	2012-08-02 05:32:46.71036
25488	4	1-7422	25485	48	13	1765	87	2012-08-02 05:32:46.71036
25485	4	1-7482	25485	13	40	3657	18	2012-08-02 05:32:46.71036
25490	5	1-5611	25489	1	55	109	3	2012-08-02 05:32:47.108283
25491	5	1-7144	25489	55	47	90	96	2012-08-02 05:32:47.108283
25492	5	1-5353	25489	47	78	68	78	2012-08-02 05:32:47.108283
25493	5	1-7379	25489	78	90	153	86	2012-08-02 05:32:47.108283
25489	5	1-7483	25489	90	1	229	9	2012-08-02 05:32:47.108283
25495	4	1-4149	25494	73	95	217	19	2012-08-02 05:32:47.539143
25496	4	1-5626	25494	95	20	305	51	2012-08-02 05:32:47.539143
25497	4	1-7463	25494	20	72	402	70	2012-08-02 05:32:47.539143
25494	4	1-7484	25494	72	73	811	18	2012-08-02 05:32:47.539143
25499	5	1-2746	25498	26	30	1079	16	2012-08-02 05:32:47.539143
25500	5	1-7472	25498	30	51	908	27	2012-08-02 05:32:47.539143
25501	5	1-4386	25498	51	20	539	51	2012-08-02 05:32:47.539143
25502	5	1-7463	25498	20	72	793	70	2012-08-02 05:32:47.539143
25498	5	1-7484	25498	72	26	1786	18	2012-08-02 05:32:47.539143
25503	1	1-2241	\N	72	72	3390	16	2012-08-02 05:32:47.539143
25505	3	1-7095	25504	93	34	201	59	2012-08-02 05:32:47.948154
25506	3	1-2123	25504	34	99	113	57	2012-08-02 05:32:47.948154
25504	3	1-7485	25504	99	93	121	70	2012-08-02 05:32:47.948154
25508	3	1-6873	25507	8	33	5423	42	2012-08-02 05:32:47.948154
25509	3	1-7201	25507	33	99	3674	57	2012-08-02 05:32:47.948154
25507	3	1-7485	25507	99	8	3954	70	2012-08-02 05:32:47.948154
25511	5	1-7463	25510	20	90	376	70	2012-08-02 05:32:48.35685
25512	5	1-7353	25510	90	77	333	33	2012-08-02 05:32:48.35685
25513	5	1-6582	25510	77	73	177	58	2012-08-02 05:32:48.35685
25514	5	1-7271	25510	73	49	180	74	2012-08-02 05:32:48.35685
25510	5	1-7486	25510	49	20	188	51	2012-08-02 05:32:48.35685
25516	5	1-6748	25515	44	23	1069	8	2012-08-02 05:32:48.765669
25517	5	1-6519	25515	23	23	964	6	2012-08-02 05:32:48.765669
25518	5	1-5777	25515	23	30	726	16	2012-08-02 05:32:48.765669
25519	5	1-7472	25515	30	73	822	27	2012-08-02 05:32:48.765669
25515	5	1-7487	25515	73	44	1112	28	2012-08-02 05:32:48.765669
25521	5	1-7178	25520	95	84	457	30	2012-08-02 05:32:49.980879
25522	5	1-6733	25520	84	83	543	88	2012-08-02 05:32:49.980879
25523	5	1-7158	25520	83	97	443	21	2012-08-02 05:32:49.980879
25524	5	1-7457	25520	97	56	854	76	2012-08-02 05:32:49.980879
25520	5	1-7490	25520	56	95	993	16	2012-08-02 05:32:49.980879
25526	5	1-7085	25525	24	47	3768	96	2012-08-02 05:32:51.218323
25527	5	1-5353	25525	47	78	1543	78	2012-08-02 05:32:51.218323
25528	5	1-7379	25525	78	90	1871	86	2012-08-02 05:32:51.218323
25529	5	1-7483	25525	90	35	1510	9	2012-08-02 05:32:51.218323
25525	5	1-7493	25525	35	24	7410	39	2012-08-02 05:32:51.218323
25531	5	1-7121	25530	6	9	1314	35	2012-08-02 05:32:51.218323
25532	5	1-3800	25530	9	63	428	67	2012-08-02 05:32:51.218323
25533	5	1-6866	25530	63	90	238	86	2012-08-02 05:32:51.218323
25534	5	1-7483	25530	90	35	258	9	2012-08-02 05:32:51.218323
25530	5	1-7493	25530	35	6	1704	39	2012-08-02 05:32:51.218323
25536	4	1-6927	25535	81	16	132	61	2012-08-02 05:32:51.218323
25537	4	1-5096	25535	16	90	26	86	2012-08-02 05:32:51.218323
25538	4	1-7483	25535	90	35	28	9	2012-08-02 05:32:51.218323
25535	4	1-7493	25535	35	81	187	39	2012-08-02 05:32:51.218323
25540	5	1-7472	25539	30	45	1322	27	2012-08-02 05:32:52.03821
25541	5	1-7416	25539	45	1	2208	23	2012-08-02 05:32:52.03821
25542	5	1-7068	25539	1	75	1464	100	2012-08-02 05:32:52.03821
25543	5	1-7100	25539	75	11	1409	33	2012-08-02 05:32:52.03821
25539	5	1-7495	25539	11	30	1265	16	2012-08-02 05:32:52.03821
25545	5	1-7488	25544	18	67	54	46	2012-08-02 05:32:52.03821
25546	5	1-7382	25544	67	45	68	75	2012-08-02 05:32:52.03821
25547	5	1-1892	25544	45	75	51	100	2012-08-02 05:32:52.03821
25548	5	1-7100	25544	75	11	55	33	2012-08-02 05:32:52.03821
25544	5	1-7495	25544	11	18	56	16	2012-08-02 05:32:52.03821
25550	4	1-6464	25549	10	83	163	32	2012-08-02 05:32:54.535048
25551	4	1-7180	25549	83	28	243	71	2012-08-02 05:32:54.535048
25552	4	1-3943	25549	28	7	211	46	2012-08-02 05:32:54.535048
25549	4	1-7501	25549	7	10	372	13	2012-08-02 05:32:54.535048
25554	5	1-5680	25553	78	56	2467	76	2012-08-02 05:32:56.589398
25555	5	1-7490	25553	56	18	2930	16	2012-08-02 05:32:56.589398
25556	5	1-7488	25553	18	67	2720	46	2012-08-02 05:32:56.589398
25557	5	1-7382	25553	67	92	3273	75	2012-08-02 05:32:56.589398
25553	5	1-7506	25553	92	78	2555	52	2012-08-02 05:32:56.589398
25559	5	1-2526	25558	16	34	1662	5	2012-08-02 05:32:57.020447
25560	5	1-7005	25558	34	98	1210	62	2012-08-02 05:32:57.020447
25561	5	1-6739	25558	98	15	854	86	2012-08-02 05:32:57.020447
25562	5	1-7145	25558	15	60	1328	43	2012-08-02 05:32:57.020447
25558	5	1-7507	25558	60	16	2502	72	2012-08-02 05:32:57.020447
25564	4	1-7203	25563	65	42	1878	14	2012-08-02 05:32:57.020447
25565	4	1-7129	25563	42	89	416	19	2012-08-02 05:32:57.020447
25566	4	1-5909	25563	89	60	425	43	2012-08-02 05:32:57.020447
25563	4	1-7507	25563	60	65	831	72	2012-08-02 05:32:57.020447
25568	2	1-6365	25567	67	46	2256	6	2012-08-02 05:32:57.451255
25567	2	1-7508	25567	46	67	8331	96	2012-08-02 05:32:57.451255
25570	2	1-2979	25569	24	19	279	89	2012-08-02 05:32:57.860067
25569	2	1-7509	25569	19	24	9837	38	2012-08-02 05:32:57.860067
25572	4	1-7082	25571	3	98	2415	68	2012-08-02 05:32:58.677573
25573	4	1-7272	25571	98	87	4883	40	2012-08-02 05:32:58.677573
25574	4	1-5060	25571	87	70	2136	87	2012-08-02 05:32:58.677573
25571	4	1-7511	25571	70	3	4146	27	2012-08-02 05:32:58.677573
25576	5	1-4386	25575	51	20	214	51	2012-08-02 05:32:58.677573
25577	5	1-7463	25575	20	28	347	70	2012-08-02 05:32:58.677573
25578	5	1-6817	25575	28	19	155	97	2012-08-02 05:32:58.677573
25579	5	1-7262	25575	19	70	167	87	2012-08-02 05:32:58.677573
25575	5	1-7511	25575	70	51	328	27	2012-08-02 05:32:58.677573
25581	4	1-6805	25580	96	81	2134	63	2012-08-02 05:32:59.086407
25582	4	1-7432	25580	81	31	1124	23	2012-08-02 05:32:59.086407
25583	4	1-7370	25580	31	15	3503	83	2012-08-02 05:32:59.086407
25580	4	1-7512	25580	15	96	1904	69	2012-08-02 05:32:59.086407
25585	5	1-2766	25584	59	18	1435	16	2012-08-02 05:32:59.495215
25586	5	1-7488	25584	18	7	1228	46	2012-08-02 05:32:59.495215
25587	5	1-7501	25584	7	16	1895	13	2012-08-02 05:32:59.495215
25588	5	1-6994	25584	16	55	1768	7	2012-08-02 05:32:59.495215
25584	5	1-7513	25584	55	59	2072	32	2012-08-02 05:32:59.495215
25590	4	1-5078	25589	68	90	539	69	2012-08-02 05:32:59.495215
25591	4	1-7103	25589	90	2	651	52	2012-08-02 05:32:59.495215
25592	4	1-5423	25589	2	55	472	7	2012-08-02 05:32:59.495215
25589	4	1-7513	25589	55	68	577	32	2012-08-02 05:32:59.495215
25594	5	1-2766	25593	59	30	694	16	2012-08-02 05:32:59.495215
25595	5	1-7472	25593	30	100	798	27	2012-08-02 05:32:59.495215
25596	5	1-7452	25593	100	16	681	13	2012-08-02 05:32:59.495215
25597	5	1-6994	25593	16	55	702	7	2012-08-02 05:32:59.495215
25593	5	1-7513	25593	55	59	908	32	2012-08-02 05:32:59.495215
25599	5	1-6378	25598	34	43	1548	17	2012-08-02 05:33:00.324014
25600	5	1-7424	25598	43	14	1098	83	2012-08-02 05:33:00.324014
25601	5	1-6745	25598	14	37	949	36	2012-08-02 05:33:00.324014
25602	5	1-7184	25598	37	53	444	67	2012-08-02 05:33:00.324014
25598	5	1-7515	25598	53	34	2084	18	2012-08-02 05:33:00.324014
25604	4	1-4620	25603	9	56	1227	20	2012-08-02 05:33:00.324014
25605	4	1-4886	25603	56	37	495	36	2012-08-02 05:33:00.324014
25606	4	1-7184	25603	37	53	280	67	2012-08-02 05:33:00.324014
25603	4	1-7515	25603	53	9	1587	18	2012-08-02 05:33:00.324014
25608	5	1-5979	25607	33	25	1853	93	2012-08-02 05:33:00.324014
25609	5	1-7363	25607	25	51	918	12	2012-08-02 05:33:00.324014
25610	5	1-5206	25607	51	24	849	89	2012-08-02 05:33:00.324014
25611	5	1-5043	25607	24	53	358	67	2012-08-02 05:33:00.324014
25607	5	1-7515	25607	53	33	2146	18	2012-08-02 05:33:00.324014
25613	5	1-6116	25612	24	44	516	62	2012-08-02 05:33:01.152719
25614	5	1-7399	25612	44	77	852	28	2012-08-02 05:33:01.152719
25615	5	1-2118	25612	77	79	165	90	2012-08-02 05:33:01.152719
25616	5	1-6549	25612	79	57	122	98	2012-08-02 05:33:01.152719
25612	5	1-7517	25612	57	24	995	82	2012-08-02 05:33:01.152719
25618	5	1-3682	25617	7	12	43	80	2012-08-02 05:33:01.152719
25619	5	1-6035	25617	12	77	13	28	2012-08-02 05:33:01.152719
25620	5	1-2118	25617	77	79	3	90	2012-08-02 05:33:01.152719
25621	5	1-6549	25617	79	57	3	98	2012-08-02 05:33:01.152719
25617	5	1-7517	25617	57	7	25	82	2012-08-02 05:33:01.152719
25623	5	1-2893	25622	5	99	6886	57	2012-08-02 05:33:01.152719
25624	5	1-7485	25622	99	17	5605	70	2012-08-02 05:33:01.152719
25625	5	1-5137	25622	17	79	1166	90	2012-08-02 05:33:01.152719
25626	5	1-6549	25622	79	57	891	98	2012-08-02 05:33:01.152719
25622	5	1-7517	25622	57	5	7575	82	2012-08-02 05:33:01.152719
25628	4	1-5324	25627	46	77	741	28	2012-08-02 05:33:01.152719
25629	4	1-2118	25627	77	79	160	90	2012-08-02 05:33:01.152719
25630	4	1-6549	25627	79	57	132	98	2012-08-02 05:33:01.152719
25627	4	1-7517	25627	57	46	1203	82	2012-08-02 05:33:01.152719
25632	3	1-7225	25631	33	33	2095	26	2012-08-02 05:33:01.981709
25633	3	1-7446	25631	33	55	2127	64	2012-08-02 05:33:01.981709
25631	3	1-7519	25631	55	33	2568	25	2012-08-02 05:33:01.981709
25635	5	1-7124	25634	11	76	3385	53	2012-08-02 05:33:02.467513
25636	5	1-7187	25634	76	27	3319	8	2012-08-02 05:33:02.467513
25637	5	1-7412	25634	27	13	4151	12	2012-08-02 05:33:02.467513
25638	5	1-6252	25634	13	51	3206	97	2012-08-02 05:33:02.467513
25634	5	1-7520	25634	51	11	5891	72	2012-08-02 05:33:02.467513
25640	5	1-4955	25639	67	64	516	19	2012-08-02 05:33:02.467513
25641	5	1-6442	25639	64	97	699	21	2012-08-02 05:33:02.467513
25642	5	1-7457	25639	97	93	1361	76	2012-08-02 05:33:02.467513
25643	5	1-5905	25639	93	51	983	97	2012-08-02 05:33:02.467513
25639	5	1-7520	25639	51	67	1861	72	2012-08-02 05:33:02.467513
25645	4	1-6966	25644	18	27	465	8	2012-08-02 05:33:02.467513
25646	4	1-7412	25644	27	13	618	12	2012-08-02 05:33:02.467513
25647	4	1-6252	25644	13	51	507	97	2012-08-02 05:33:02.467513
25644	4	1-7520	25644	51	18	991	72	2012-08-02 05:33:02.467513
25649	2	1-6414	25648	43	26	433	74	2012-08-02 05:33:03.285094
25648	2	1-7522	25648	26	43	5336	95	2012-08-02 05:33:03.285094
25651	2	1-5160	25650	50	86	518	54	2012-08-02 05:33:04.102534
25650	2	1-7524	25650	86	50	7776	28	2012-08-02 05:33:04.102534
25653	5	1-5802	25652	25	97	169	21	2012-08-02 05:33:04.500537
25654	5	1-7457	25652	97	93	278	76	2012-08-02 05:33:04.500537
25655	5	1-5905	25652	93	88	170	97	2012-08-02 05:33:04.500537
25656	5	1-5548	25652	88	18	63	3	2012-08-02 05:33:04.500537
25652	5	1-7525	25652	18	25	777	2	2012-08-02 05:33:04.500537
25658	5	1-7351	25657	56	73	5346	58	2012-08-02 05:33:06.20278
25659	5	1-7271	25657	73	91	5009	74	2012-08-02 05:33:06.20278
25660	5	1-6946	25657	91	96	2023	69	2012-08-02 05:33:06.20278
25661	5	1-6805	25657	96	31	2659	63	2012-08-02 05:33:06.20278
25657	5	1-7529	25657	31	56	2598	49	2012-08-02 05:33:06.20278
25663	4	1-7351	25662	56	40	2189	58	2012-08-02 05:33:06.20278
25664	4	1-7212	25662	40	77	2788	77	2012-08-02 05:33:06.20278
25665	4	1-5557	25662	77	31	1086	63	2012-08-02 05:33:06.20278
25662	4	1-7529	25662	31	56	1062	49	2012-08-02 05:33:06.20278
25667	4	1-7385	25666	55	23	1290	6	2012-08-02 05:33:07.009623
25668	4	1-5777	25666	23	30	667	16	2012-08-02 05:33:07.009623
25669	4	1-7472	25666	30	50	535	27	2012-08-02 05:33:07.009623
25666	4	1-7531	25666	50	55	1776	44	2012-08-02 05:33:07.009623
25671	4	1-5854	25670	99	56	265	76	2012-08-02 05:33:07.009623
25672	4	1-7490	25670	56	30	245	16	2012-08-02 05:33:07.009623
25673	4	1-7472	25670	30	50	216	27	2012-08-02 05:33:07.009623
25670	4	1-7531	25670	50	99	787	44	2012-08-02 05:33:07.009623
25675	5	1-6937	25674	54	42	225	34	2012-08-02 05:33:07.009623
25676	5	1-6730	25674	42	12	729	92	2012-08-02 05:33:07.009623
25677	5	1-4665	25674	12	30	91	16	2012-08-02 05:33:07.009623
25678	5	1-7472	25674	30	50	96	27	2012-08-02 05:33:07.009623
25674	5	1-7531	25674	50	54	418	44	2012-08-02 05:33:07.009623
25680	5	1-7229	25679	34	48	1461	75	2012-08-02 05:33:07.009623
25681	5	1-6868	25679	48	15	563	4	2012-08-02 05:33:07.009623
25682	5	1-7443	25679	15	30	664	16	2012-08-02 05:33:07.009623
25683	5	1-7472	25679	30	50	738	27	2012-08-02 05:33:07.009623
25679	5	1-7531	25679	50	34	3391	44	2012-08-02 05:33:07.009623
25685	5	1-6648	25684	95	96	514	79	2012-08-02 05:33:07.009623
25686	5	1-7349	25684	96	75	1542	15	2012-08-02 05:33:07.009623
25687	5	1-6726	25684	75	30	670	16	2012-08-02 05:33:07.009623
25688	5	1-7472	25684	30	50	753	27	2012-08-02 05:33:07.009623
25684	5	1-7531	25684	50	95	3499	44	2012-08-02 05:33:07.009623
25689	1	1-5775	\N	21	21	944	27	2012-08-02 05:33:07.009623
25691	5	1-7117	25690	47	76	384	53	2012-08-02 05:33:07.860403
25692	5	1-7187	25690	76	27	371	8	2012-08-02 05:33:07.860403
25693	5	1-7412	25690	27	98	457	12	2012-08-02 05:33:07.860403
25694	5	1-4987	25690	98	14	225	30	2012-08-02 05:33:07.860403
25690	5	1-7533	25690	14	47	355	98	2012-08-02 05:33:07.860403
25696	4	1-7529	25695	31	60	389	49	2012-08-02 05:33:08.688974
25697	4	1-6679	25695	60	87	240	17	2012-08-02 05:33:08.688974
25698	4	1-7102	25695	87	97	150	92	2012-08-02 05:33:08.688974
25695	4	1-7535	25695	97	31	811	63	2012-08-02 05:33:08.688974
25700	4	1-7529	25699	31	31	5884	49	2012-08-02 05:33:08.688974
25701	4	1-5613	25699	31	82	2351	95	2012-08-02 05:33:08.688974
25702	4	1-6105	25699	82	97	1133	92	2012-08-02 05:33:08.688974
25699	4	1-7535	25699	97	31	7328	63	2012-08-02 05:33:08.688974
25704	4	1-4600	25703	63	86	63	15	2012-08-02 05:33:08.688974
25705	4	1-6362	25703	86	42	7	34	2012-08-02 05:33:08.688974
25706	4	1-6730	25703	42	97	15	92	2012-08-02 05:33:08.688974
25703	4	1-7535	25703	97	63	115	63	2012-08-02 05:33:08.688974
25708	5	1-6960	25707	82	45	1176	19	2012-08-02 05:33:09.893064
25709	5	1-7309	25707	45	71	992	60	2012-08-02 05:33:09.893064
25710	5	1-6965	25707	71	48	408	7	2012-08-02 05:33:09.893064
25711	5	1-7239	25707	48	97	394	31	2012-08-02 05:33:09.893064
25707	5	1-7538	25707	97	82	1628	67	2012-08-02 05:33:09.893064
25713	4	1-7012	25712	7	98	272	57	2012-08-02 05:33:09.893064
25714	4	1-7491	25712	98	95	101	93	2012-08-02 05:33:09.893064
25715	4	1-3725	25712	95	97	36	31	2012-08-02 05:33:09.893064
25712	4	1-7538	25712	97	7	149	67	2012-08-02 05:33:09.893064
25717	3	1-5666	25716	63	80	2157	25	2012-08-02 05:33:09.893064
25718	3	1-5268	25716	80	97	1190	31	2012-08-02 05:33:09.893064
25716	3	1-7538	25716	97	63	4989	67	2012-08-02 05:33:09.893064
25720	3	1-5940	25719	8	87	784	87	2012-08-02 05:33:09.893064
25721	3	1-2490	25719	87	97	278	31	2012-08-02 05:33:09.893064
25719	3	1-7538	25719	97	8	1193	67	2012-08-02 05:33:09.893064
25723	5	1-6984	25722	93	12	3426	65	2012-08-02 05:33:11.130555
25724	5	1-4031	25722	12	40	995	39	2012-08-02 05:33:11.130555
25725	5	1-7527	25722	40	27	986	1	2012-08-02 05:33:11.130555
25726	5	1-7431	25722	27	29	1260	24	2012-08-02 05:33:11.130555
25722	5	1-7541	25722	29	93	3541	4	2012-08-02 05:33:11.130555
25728	3	1-7487	25727	73	50	5238	28	2012-08-02 05:33:13.185325
25729	3	1-6448	25727	50	81	981	8	2012-08-02 05:33:13.185325
25727	3	1-7546	25727	81	73	8899	27	2012-08-02 05:33:13.185325
25731	3	1-4386	25730	51	29	419	51	2012-08-02 05:33:13.185325
25732	3	1-3925	25730	29	81	41	8	2012-08-02 05:33:13.185325
25730	3	1-7546	25730	81	51	671	27	2012-08-02 05:33:13.185325
25733	1	1-6448	\N	50	50	906	8	2012-08-02 05:33:13.185325
25735	3	1-7095	25734	93	93	1479	59	2012-08-02 05:33:13.583326
25736	3	1-6061	25734	93	62	640	45	2012-08-02 05:33:13.583326
25734	3	1-7547	25734	62	93	956	70	2012-08-02 05:33:13.583326
25738	2	1-6422	25737	9	62	369	45	2012-08-02 05:33:13.583326
25737	2	1-7547	25737	62	9	671	70	2012-08-02 05:33:13.583326
25740	3	1-6250	25739	51	63	2334	63	2012-08-02 05:33:14.798336
25741	3	1-4600	25739	63	86	1312	15	2012-08-02 05:33:14.798336
25739	3	1-7550	25739	86	51	3447	34	2012-08-02 05:33:14.798336
25743	3	1-4052	25742	41	96	986	79	2012-08-02 05:33:14.798336
25744	3	1-7349	25742	96	86	2091	15	2012-08-02 05:33:14.798336
25742	3	1-7550	25742	86	41	5855	34	2012-08-02 05:33:14.798336
25746	2	1-4604	25745	32	21	1058	5	2012-08-02 05:33:16.455413
25745	2	1-7554	25745	21	32	6199	32	2012-08-02 05:33:16.455413
25748	4	1-7174	25747	8	47	625	44	2012-08-02 05:33:18.068329
25749	4	1-7358	25747	47	90	325	49	2012-08-02 05:33:18.068329
25750	4	1-5721	25747	90	7	221	95	2012-08-02 05:33:18.068329
25747	4	1-7558	25747	7	8	553	79	2012-08-02 05:33:18.068329
25752	4	1-2946	25751	53	48	837	7	2012-08-02 05:33:18.068329
25753	4	1-7239	25751	48	27	946	31	2012-08-02 05:33:18.068329
25754	4	1-7345	25751	27	7	955	95	2012-08-02 05:33:18.068329
25751	4	1-7558	25751	7	53	2425	79	2012-08-02 05:33:18.068329
25756	5	1-7474	25755	31	92	6398	75	2012-08-02 05:33:19.261753
25757	5	1-7506	25755	92	7	2426	52	2012-08-02 05:33:19.261753
25758	5	1-5648	25755	7	23	784	20	2012-08-02 05:33:19.261753
25759	5	1-7149	25755	23	58	424	1	2012-08-02 05:33:19.261753
25755	5	1-7561	25755	58	31	5933	78	2012-08-02 05:33:19.261753
25761	4	1-7474	25760	31	64	2400	75	2012-08-02 05:33:19.261753
25762	4	1-7413	25760	64	22	769	61	2012-08-02 05:33:19.261753
25763	4	1-5228	25760	22	58	151	1	2012-08-02 05:33:19.261753
25760	4	1-7561	25760	58	31	2168	78	2012-08-02 05:33:19.261753
25765	5	1-7271	25764	73	17	852	74	2012-08-02 05:33:22.920275
25766	5	1-5408	25764	17	45	443	34	2012-08-02 05:33:22.920275
25767	5	1-7001	25764	45	90	199	86	2012-08-02 05:33:22.920275
25768	5	1-7483	25764	90	6	294	9	2012-08-02 05:33:22.920275
25764	5	1-7570	25764	6	73	823	58	2012-08-02 05:33:22.920275
25770	4	1-6412	25769	100	96	1107	79	2012-08-02 05:33:23.340265
25771	4	1-7349	25769	96	86	3415	15	2012-08-02 05:33:23.340265
25772	4	1-6362	25769	86	34	571	34	2012-08-02 05:33:23.340265
25769	4	1-7571	25769	34	100	798	45	2012-08-02 05:33:23.340265
25774	4	1-5740	25773	93	47	867	80	2012-08-02 05:33:23.340265
25775	4	1-6069	25773	47	54	814	44	2012-08-02 05:33:23.340265
25776	4	1-6937	25773	54	34	500	34	2012-08-02 05:33:23.340265
25773	4	1-7571	25773	34	93	729	45	2012-08-02 05:33:23.340265
25778	3	1-7365	25777	40	69	1148	69	2012-08-02 05:33:26.157384
25779	3	1-2875	25777	69	10	302	9	2012-08-02 05:33:26.157384
25777	3	1-7578	25777	10	40	1982	26	2012-08-02 05:33:26.157384
25781	5	1-4009	25780	68	77	551	94	2012-08-02 05:33:26.157384
25782	5	1-7088	25780	77	98	675	62	2012-08-02 05:33:26.157384
25783	5	1-6739	25780	98	90	311	86	2012-08-02 05:33:26.157384
25784	5	1-7483	25780	90	10	299	9	2012-08-02 05:33:26.157384
25780	5	1-7578	25780	10	68	2183	26	2012-08-02 05:33:26.157384
25786	5	1-5264	25785	66	73	1124	58	2012-08-02 05:33:27.008108
25787	5	1-7271	25785	73	49	1056	74	2012-08-02 05:33:27.008108
25788	5	1-7486	25785	49	20	1015	51	2012-08-02 05:33:27.008108
25789	5	1-7463	25785	20	3	1868	70	2012-08-02 05:33:27.008108
25785	5	1-7580	25785	3	66	997	96	2012-08-02 05:33:27.008108
25791	4	1-5353	25790	47	31	276	78	2012-08-02 05:33:27.008108
25792	4	1-7474	25790	31	96	638	75	2012-08-02 05:33:27.008108
25793	4	1-3889	25790	96	3	610	70	2012-08-02 05:33:27.008108
25790	4	1-7580	25790	3	47	364	96	2012-08-02 05:33:27.008108
25795	2	1-6706	25794	62	25	1435	85	2012-08-02 05:33:27.836435
25794	2	1-7582	25794	25	62	666	39	2012-08-02 05:33:27.836435
25797	4	1-6628	25796	7	71	3211	60	2012-08-02 05:33:27.836435
25798	4	1-6965	25796	71	26	1352	7	2012-08-02 05:33:27.836435
25799	4	1-7532	25796	26	25	8207	85	2012-08-02 05:33:27.836435
25796	4	1-7582	25796	25	7	3870	39	2012-08-02 05:33:27.836435
25801	2	1-5355	25800	39	35	3094	54	2012-08-02 05:33:28.234389
25800	2	1-7583	25800	35	39	7044	33	2012-08-02 05:33:28.234389
25803	4	1-7507	25802	60	18	2241	72	2012-08-02 05:33:29.427788
25804	4	1-6966	25802	18	55	877	8	2012-08-02 05:33:29.427788
25805	4	1-4240	25802	55	26	842	80	2012-08-02 05:33:29.427788
25802	4	1-7586	25802	26	60	1433	43	2012-08-02 05:33:29.427788
25807	4	1-5720	25806	95	80	250	79	2012-08-02 05:33:29.427788
25808	4	1-7398	25806	80	7	229	48	2012-08-02 05:33:29.427788
25809	4	1-6500	25806	7	26	134	80	2012-08-02 05:33:29.427788
25806	4	1-7586	25806	26	95	244	43	2012-08-02 05:33:29.427788
25811	5	1-7369	25810	70	96	408	49	2012-08-02 05:33:29.427788
25812	5	1-6024	25810	96	83	247	1	2012-08-02 05:33:29.427788
25813	5	1-7530	25810	83	78	204	33	2012-08-02 05:33:29.427788
25814	5	1-5571	25810	78	26	349	80	2012-08-02 05:33:29.427788
25810	5	1-7586	25810	26	70	730	43	2012-08-02 05:33:29.427788
25816	4	1-7469	25815	91	13	1564	94	2012-08-02 05:33:29.427788
25817	4	1-6738	25815	13	20	1465	82	2012-08-02 05:33:29.427788
25818	4	1-5830	25815	20	26	1821	80	2012-08-02 05:33:29.427788
25815	4	1-7586	25815	26	91	3882	43	2012-08-02 05:33:29.427788
25820	3	1-7580	25819	3	19	1855	96	2012-08-02 05:33:30.23434
25821	3	1-6518	25819	19	32	1034	45	2012-08-02 05:33:30.23434
25819	3	1-7588	25819	32	3	3964	70	2012-08-02 05:33:30.23434
25823	4	1-5137	25822	17	79	198	90	2012-08-02 05:33:30.23434
25824	4	1-6549	25822	79	38	176	98	2012-08-02 05:33:30.23434
25825	4	1-6990	25822	38	32	191	45	2012-08-02 05:33:30.23434
25822	4	1-7588	25822	32	17	814	70	2012-08-02 05:33:30.23434
25827	5	1-1478	25826	13	71	52	66	2012-08-02 05:33:30.23434
25828	5	1-7587	25826	71	88	98	67	2012-08-02 05:33:30.23434
25829	5	1-5493	25826	88	34	61	34	2012-08-02 05:33:30.23434
25830	5	1-7571	25826	34	32	82	45	2012-08-02 05:33:30.23434
25826	5	1-7588	25826	32	13	380	70	2012-08-02 05:33:30.23434
25832	5	1-7396	25831	84	71	467	66	2012-08-02 05:33:30.23434
25833	5	1-7587	25831	71	88	905	67	2012-08-02 05:33:30.23434
25834	5	1-5493	25831	88	34	576	34	2012-08-02 05:33:30.23434
25835	5	1-7571	25831	34	32	790	45	2012-08-02 05:33:30.23434
25831	5	1-7588	25831	32	84	3718	70	2012-08-02 05:33:30.23434
25836	1	1-6154	\N	3	3	3045	98	2012-08-02 05:33:30.23434
25838	3	1-7549	25837	86	9	1227	29	2012-08-02 05:33:31.471669
25839	3	1-6171	25837	9	51	796	85	2012-08-02 05:33:31.471669
25837	3	1-7591	25837	51	86	629	53	2012-08-02 05:33:31.471669
25841	3	1-7549	25840	86	78	2602	29	2012-08-02 05:33:31.471669
25842	3	1-4783	25840	78	51	1532	85	2012-08-02 05:33:31.471669
25840	3	1-7591	25840	51	86	1271	53	2012-08-02 05:33:31.471669
25844	4	1-7549	25843	86	36	4987	29	2012-08-02 05:33:31.471669
25845	4	1-4716	25843	36	53	2150	4	2012-08-02 05:33:31.471669
25846	4	1-3636	25843	53	51	1383	85	2012-08-02 05:33:31.471669
25843	4	1-7591	25843	51	86	1672	53	2012-08-02 05:33:31.471669
25848	4	1-4240	25847	55	43	2206	80	2012-08-02 05:33:31.880539
25849	4	1-7392	25847	43	93	4261	81	2012-08-02 05:33:31.880539
25850	4	1-2728	25847	93	79	2971	51	2012-08-02 05:33:31.880539
25847	4	1-7592	25847	79	55	2316	8	2012-08-02 05:33:31.880539
25852	4	1-1131	25851	78	43	15	80	2012-08-02 05:33:31.880539
25853	4	1-7392	25851	43	93	33	81	2012-08-02 05:33:31.880539
25854	4	1-2728	25851	93	79	26	51	2012-08-02 05:33:31.880539
25851	4	1-7592	25851	79	78	23	8	2012-08-02 05:33:31.880539
25856	4	1-1131	25855	78	43	400	80	2012-08-02 05:33:31.880539
25857	4	1-7392	25855	43	35	895	81	2012-08-02 05:33:31.880539
25858	4	1-5851	25855	35	79	693	51	2012-08-02 05:33:31.880539
25855	4	1-7592	25855	79	78	626	8	2012-08-02 05:33:31.880539
25860	5	1-7443	25859	15	30	1009	16	2012-08-02 05:33:33.504556
25861	5	1-7472	25859	30	3	683	27	2012-08-02 05:33:33.504556
25862	5	1-7082	25859	3	98	294	68	2012-08-02 05:33:33.504556
25863	5	1-7272	25859	98	26	439	40	2012-08-02 05:33:33.504556
25859	5	1-7596	25859	26	15	1405	4	2012-08-02 05:33:33.504556
25865	5	1-7443	25864	15	48	3093	16	2012-08-02 05:33:33.504556
25866	5	1-4445	25864	48	28	751	62	2012-08-02 05:33:33.504556
25867	5	1-7169	25864	28	98	490	68	2012-08-02 05:33:33.504556
25868	5	1-7272	25864	98	26	896	40	2012-08-02 05:33:33.504556
25864	5	1-7596	25864	26	15	3515	4	2012-08-02 05:33:33.504556
25870	5	1-7234	25869	58	87	1192	25	2012-08-02 05:33:33.504556
25871	5	1-7015	25869	87	91	787	75	2012-08-02 05:33:33.504556
25872	5	1-6924	25869	91	98	361	68	2012-08-02 05:33:33.504556
25873	5	1-7272	25869	98	26	773	40	2012-08-02 05:33:33.504556
25869	5	1-7596	25869	26	58	3552	4	2012-08-02 05:33:33.504556
25875	4	1-5736	25874	66	82	461	39	2012-08-02 05:33:36.388437
25876	4	1-6785	25874	82	24	1162	10	2012-08-02 05:33:36.388437
25874	4	1-7603	25874	37	66	1171	27	2012-08-02 05:33:36.388437
25877	4	1-7567	25874	24	37	387	59	2012-08-02 05:33:36.388437
25879	4	1-7523	25878	62	48	952	85	2012-08-02 05:33:36.388437
25880	4	1-6515	25878	48	22	184	36	2012-08-02 05:33:36.388437
25881	4	1-7516	25878	22	37	255	59	2012-08-02 05:33:36.388437
25878	4	1-7603	25878	37	62	846	27	2012-08-02 05:33:36.388437
25883	4	1-7082	25882	3	75	3043	68	2012-08-02 05:33:36.388437
25884	4	1-5655	25882	75	27	1547	25	2012-08-02 05:33:36.388437
25885	4	1-4564	25882	27	37	1154	59	2012-08-02 05:33:36.388437
25882	4	1-7603	25882	37	3	4651	27	2012-08-02 05:33:36.388437
25887	4	1-7594	25886	12	30	2719	42	2012-08-02 05:33:36.388437
25888	4	1-4766	25886	30	95	657	28	2012-08-02 05:33:36.388437
25889	4	1-7223	25886	95	37	254	59	2012-08-02 05:33:36.388437
25886	4	1-7603	25886	37	12	1061	27	2012-08-02 05:33:36.388437
25891	4	1-7523	25890	62	13	99	85	2012-08-02 05:33:36.388437
25892	4	1-4782	25890	13	67	14	38	2012-08-02 05:33:36.388437
25893	4	1-6339	25890	67	37	14	59	2012-08-02 05:33:36.388437
25890	4	1-7603	25890	37	62	64	27	2012-08-02 05:33:36.388437
25895	2	1-6918	25894	26	46	516	10	2012-08-02 05:33:36.797158
25894	2	1-7604	25894	46	26	955	43	2012-08-02 05:33:36.797158
25897	5	1-7369	25896	70	96	1846	49	2012-08-02 05:33:36.797158
25898	5	1-6024	25896	96	27	1092	1	2012-08-02 05:33:36.797158
25899	5	1-7431	25896	27	67	1349	24	2012-08-02 05:33:36.797158
25900	5	1-4885	25896	67	46	1456	10	2012-08-02 05:33:36.797158
25896	5	1-7604	25896	46	70	3378	43	2012-08-02 05:33:36.797158
25902	4	1-6887	25901	93	28	887	5	2012-08-02 05:33:36.797158
25903	4	1-4888	25901	28	54	633	40	2012-08-02 05:33:36.797158
25904	4	1-7536	25901	54	46	573	10	2012-08-02 05:33:36.797158
25901	4	1-7604	25901	46	93	1339	43	2012-08-02 05:33:36.797158
25906	4	1-7525	25905	18	23	6615	2	2012-08-02 05:33:39.216741
25907	4	1-1819	25905	23	46	1086	81	2012-08-02 05:33:39.216741
25908	4	1-4184	25905	46	11	412	44	2012-08-02 05:33:39.216741
25905	4	1-7610	25905	11	18	473	3	2012-08-02 05:33:39.216741
25910	5	1-7395	25909	67	17	1001	25	2012-08-02 05:33:39.216741
25911	5	1-6268	25909	17	71	609	61	2012-08-02 05:33:39.216741
25912	5	1-7327	25909	71	57	589	71	2012-08-02 05:33:39.216741
25913	5	1-6953	25909	57	11	663	44	2012-08-02 05:33:39.216741
25909	5	1-7610	25909	11	67	770	3	2012-08-02 05:33:39.216741
25915	5	1-6474	25914	65	90	2185	1	2012-08-02 05:33:39.216741
25916	5	1-7435	25914	90	92	2277	94	2012-08-02 05:33:39.216741
25917	5	1-7318	25914	92	57	2112	71	2012-08-02 05:33:39.216741
25918	5	1-6953	25914	57	11	2411	44	2012-08-02 05:33:39.216741
25914	5	1-7610	25914	11	65	2846	3	2012-08-02 05:33:39.216741
25920	4	1-7604	25919	46	94	147	43	2012-08-02 05:33:40.454398
25921	4	1-3003	25919	94	6	26	8	2012-08-02 05:33:40.454398
25922	4	1-3486	25919	6	11	25	57	2012-08-02 05:33:40.454398
25919	4	1-7613	25919	11	46	84	10	2012-08-02 05:33:40.454398
25924	4	1-7604	25923	46	88	3359	43	2012-08-02 05:33:40.454398
25925	4	1-5351	25923	88	13	774	90	2012-08-02 05:33:40.454398
25926	4	1-7196	25923	13	11	564	57	2012-08-02 05:33:40.454398
25923	4	1-7613	25923	11	46	1899	10	2012-08-02 05:33:40.454398
25928	4	1-7394	25927	63	49	1610	75	2012-08-02 05:33:40.454398
25929	4	1-6905	25927	49	13	1133	90	2012-08-02 05:33:40.454398
25930	4	1-7196	25927	13	11	1019	57	2012-08-02 05:33:40.454398
25927	4	1-7613	25927	11	63	4235	10	2012-08-02 05:33:40.454398
25932	5	1-7483	25931	90	6	1313	9	2012-08-02 05:33:41.68057
25933	5	1-7570	25931	6	26	3588	58	2012-08-02 05:33:41.68057
25934	5	1-6691	25931	26	30	736	71	2012-08-02 05:33:41.68057
25935	5	1-6452	25931	30	7	991	89	2012-08-02 05:33:41.68057
25931	5	1-7616	25931	7	90	910	86	2012-08-02 05:33:41.68057
25937	4	1-6728	25936	55	19	5605	97	2012-08-02 05:33:42.089303
25938	4	1-7262	25936	19	26	5097	87	2012-08-02 05:33:42.089303
25939	4	1-6769	25936	26	6	507	91	2012-08-02 05:33:42.089303
25936	4	1-7617	25936	6	55	3763	99	2012-08-02 05:33:42.089303
25941	3	1-7286	25940	67	68	2364	68	2012-08-02 05:33:42.089303
25942	3	1-7202	25940	68	6	446	91	2012-08-02 05:33:42.089303
25940	3	1-7617	25940	6	67	3359	99	2012-08-02 05:33:42.089303
25944	2	1-6981	25943	12	76	2080	75	2012-08-02 05:33:42.509327
25943	2	1-7618	25943	76	12	1933	47	2012-08-02 05:33:42.509327
25946	5	1-5653	25945	52	81	2860	63	2012-08-02 05:33:42.928988
25947	5	1-7432	25945	81	97	1496	23	2012-08-02 05:33:42.928988
25948	5	1-7557	25945	97	64	754	19	2012-08-02 05:33:42.928988
25949	5	1-6442	25945	64	68	814	21	2012-08-02 05:33:42.928988
25945	5	1-7619	25945	68	52	3689	100	2012-08-02 05:33:42.928988
25951	4	1-7341	25950	68	73	3794	18	2012-08-02 05:33:42.928988
25952	4	1-4149	25950	73	64	942	19	2012-08-02 05:33:42.928988
25953	4	1-6442	25950	64	68	801	21	2012-08-02 05:33:42.928988
25950	4	1-7619	25950	68	68	3752	100	2012-08-02 05:33:42.928988
25955	5	1-7341	25954	68	41	1043	18	2012-08-02 05:33:42.928988
25956	5	1-4438	25954	41	53	1046	50	2012-08-02 05:33:42.928988
25957	5	1-7471	25954	53	64	145	19	2012-08-02 05:33:42.928988
25958	5	1-6442	25954	64	68	165	21	2012-08-02 05:33:42.928988
25954	5	1-7619	25954	68	68	893	100	2012-08-02 05:33:42.928988
25960	4	1-3136	25959	100	76	64	78	2012-08-02 05:33:42.928988
25961	4	1-6140	25959	76	14	31	91	2012-08-02 05:33:42.928988
25962	4	1-6634	25959	14	68	48	21	2012-08-02 05:33:42.928988
25959	4	1-7619	25959	68	100	267	100	2012-08-02 05:33:42.928988
25963	1	1-3136	\N	100	100	1494	78	2012-08-02 05:33:42.928988
25965	4	1-6212	25964	64	54	154	11	2012-08-02 05:33:43.758012
25966	4	1-7352	25964	54	64	186	17	2012-08-02 05:33:43.758012
25967	4	1-7569	25964	64	30	179	15	2012-08-02 05:33:43.758012
25964	4	1-7621	25964	30	64	189	30	2012-08-02 05:33:43.758012
25969	3	1-7199	25968	39	96	1178	79	2012-08-02 05:33:43.758012
25970	3	1-7349	25968	96	30	2128	15	2012-08-02 05:33:43.758012
25968	3	1-7621	25968	30	39	2409	30	2012-08-02 05:33:43.758012
25972	4	1-3723	25971	55	54	1546	11	2012-08-02 05:33:43.758012
25973	4	1-7352	25971	54	64	2274	17	2012-08-02 05:33:43.758012
25974	4	1-7569	25971	64	30	1924	15	2012-08-02 05:33:43.758012
25971	4	1-7621	25971	30	55	2472	30	2012-08-02 05:33:43.758012
25976	3	1-7020	25975	9	73	1703	82	2012-08-02 05:33:43.758012
25977	3	1-7555	25975	73	30	1213	15	2012-08-02 05:33:43.758012
25975	3	1-7621	25975	30	9	1784	30	2012-08-02 05:33:43.758012
25979	3	1-7434	25978	70	71	851	84	2012-08-02 05:33:43.758012
25980	3	1-7502	25978	71	30	387	15	2012-08-02 05:33:43.758012
25978	3	1-7621	25978	30	70	601	30	2012-08-02 05:33:43.758012
25982	4	1-7251	25981	59	60	207	49	2012-08-02 05:33:43.758012
25983	4	1-6679	25981	60	64	234	17	2012-08-02 05:33:43.758012
25984	4	1-7569	25981	64	30	243	15	2012-08-02 05:33:43.758012
25981	4	1-7621	25981	30	59	384	30	2012-08-02 05:33:43.758012
25985	1	1-3723	\N	55	55	3071	11	2012-08-02 05:33:43.758012
25987	5	1-7582	25986	25	40	818	39	2012-08-02 05:33:44.564377
25988	5	1-7527	25986	40	27	806	1	2012-08-02 05:33:44.564377
25989	5	1-7431	25986	27	16	1024	24	2012-08-02 05:33:44.564377
25990	5	1-7130	25986	16	22	1520	50	2012-08-02 05:33:44.564377
25986	5	1-7623	25986	22	25	1317	85	2012-08-02 05:33:44.564377
25992	5	1-7249	25991	84	43	1747	80	2012-08-02 05:33:44.973108
25993	5	1-7392	25991	43	33	3442	81	2012-08-02 05:33:44.973108
25994	5	1-7510	25991	33	47	2005	28	2012-08-02 05:33:44.973108
25995	5	1-6087	25991	47	48	1316	11	2012-08-02 05:33:44.973108
25991	5	1-7624	25991	48	84	2434	24	2012-08-02 05:33:44.973108
25997	4	1-7541	25996	29	71	3067	4	2012-08-02 05:33:44.973108
25998	4	1-5910	25996	71	47	1043	28	2012-08-02 05:33:44.973108
25999	4	1-6087	25996	47	48	698	11	2012-08-02 05:33:44.973108
25996	4	1-7624	25996	48	29	1317	24	2012-08-02 05:33:44.973108
26001	4	1-7456	26000	5	33	2230	26	2012-08-02 05:33:44.973108
26002	4	1-7446	26000	33	24	2470	64	2012-08-02 05:33:44.973108
26003	4	1-3246	26000	24	48	1398	11	2012-08-02 05:33:44.973108
26000	4	1-7624	26000	48	5	2738	24	2012-08-02 05:33:44.973108
26004	1	1-2060	\N	88	88	1813	64	2012-08-02 05:33:44.973108
26006	4	1-6906	26005	5	97	1627	37	2012-08-02 05:33:48.254508
26007	4	1-7607	26005	97	74	528	98	2012-08-02 05:33:48.254508
26008	4	1-4383	26005	74	9	380	58	2012-08-02 05:33:48.254508
26005	4	1-7632	26005	9	5	1249	65	2012-08-02 05:33:48.254508
26010	3	1-7442	26009	78	7	2863	4	2012-08-02 05:33:48.254508
26011	3	1-6848	26009	7	9	1548	58	2012-08-02 05:33:48.254508
26009	3	1-7632	26009	9	78	5961	65	2012-08-02 05:33:48.254508
26013	4	1-7566	26012	57	78	1632	38	2012-08-02 05:33:48.663316
26014	4	1-3825	26012	78	33	285	26	2012-08-02 05:33:48.663316
26015	4	1-7446	26012	33	95	186	64	2012-08-02 05:33:48.663316
26012	4	1-7633	26012	95	57	3303	80	2012-08-02 05:33:48.663316
26017	5	1-6035	26016	12	22	243	28	2012-08-02 05:33:48.663316
26018	5	1-6204	26016	22	25	139	57	2012-08-02 05:33:48.663316
26019	5	1-7228	26016	25	7	58	96	2012-08-02 05:33:48.663316
26020	5	1-7514	26016	7	95	39	64	2012-08-02 05:33:48.663316
26016	5	1-7633	26016	95	12	973	80	2012-08-02 05:33:48.663316
26022	4	1-7186	26021	75	44	669	75	2012-08-02 05:33:49.074004
26023	4	1-3500	26021	44	82	592	5	2012-08-02 05:33:49.074004
26024	4	1-7573	26021	82	9	241	45	2012-08-02 05:33:49.074004
26021	4	1-7634	26021	9	75	844	49	2012-08-02 05:33:49.074004
26026	3	1-6632	26025	16	38	945	98	2012-08-02 05:33:49.074004
26027	3	1-6990	26025	38	9	991	45	2012-08-02 05:33:49.074004
26025	3	1-7634	26025	9	16	3825	49	2012-08-02 05:33:49.074004
26029	4	1-6024	26028	96	58	117	1	2012-08-02 05:33:49.074004
26030	4	1-7156	26028	58	32	145	38	2012-08-02 05:33:49.074004
26031	4	1-7279	26028	32	9	57	45	2012-08-02 05:33:49.074004
26028	4	1-7634	26028	9	96	224	49	2012-08-02 05:33:49.074004
26032	1	1-3500	\N	44	44	1588	5	2012-08-02 05:33:49.074004
26034	5	1-7088	26033	77	98	1053	62	2012-08-02 05:33:50.715549
26035	5	1-6739	26033	98	78	735	86	2012-08-02 05:33:50.715549
26036	5	1-7287	26033	78	34	1130	34	2012-08-02 05:33:50.715549
26037	5	1-7571	26033	34	70	1562	45	2012-08-02 05:33:50.715549
26033	5	1-7638	26033	70	77	569	94	2012-08-02 05:33:50.715549
26039	3	1-6312	26038	95	19	1686	26	2012-08-02 05:33:52.750185
26040	3	1-5175	26038	19	93	688	74	2012-08-02 05:33:52.750185
26038	3	1-7643	26038	93	95	3018	69	2012-08-02 05:33:52.750185
26042	5	1-7027	26041	46	51	501	34	2012-08-02 05:33:52.750185
26043	5	1-6250	26041	51	81	420	63	2012-08-02 05:33:52.750185
26044	5	1-7432	26041	81	17	228	23	2012-08-02 05:33:52.750185
26045	5	1-6781	26041	17	93	186	74	2012-08-02 05:33:52.750185
26041	5	1-7643	26041	93	46	919	69	2012-08-02 05:33:52.750185
26047	4	1-1281	26046	86	53	2059	14	2012-08-02 05:33:52.750185
26048	4	1-7254	26046	53	43	1219	95	2012-08-02 05:33:52.750185
26049	4	1-6414	26046	43	93	453	74	2012-08-02 05:33:52.750185
26046	4	1-7643	26046	93	86	2505	69	2012-08-02 05:33:52.750185
26051	4	1-6786	26050	80	41	336	5	2012-08-02 05:33:52.750185
26052	4	1-5788	26050	41	43	160	95	2012-08-02 05:33:52.750185
26053	4	1-6414	26050	43	93	63	74	2012-08-02 05:33:52.750185
26050	4	1-7643	26050	93	80	375	69	2012-08-02 05:33:52.750185
26055	3	1-4680	26054	80	47	2765	44	2012-08-02 05:33:53.195608
26056	3	1-7358	26054	47	68	597	49	2012-08-02 05:33:53.195608
26054	3	1-7644	26054	68	80	9568	68	2012-08-02 05:33:53.195608
26058	5	1-7463	26057	20	28	239	70	2012-08-02 05:33:53.607468
26059	5	1-6817	26057	28	12	125	97	2012-08-02 05:33:53.607468
26060	5	1-6243	26057	12	22	89	36	2012-08-02 05:33:53.607468
26061	5	1-7516	26057	22	59	158	59	2012-08-02 05:33:53.607468
26057	5	1-7645	26057	59	20	126	51	2012-08-02 05:33:53.607468
26063	4	1-7149	26062	23	90	4403	1	2012-08-02 05:33:54.433387
26064	4	1-7435	26062	90	48	4744	94	2012-08-02 05:33:54.433387
26065	4	1-6578	26062	48	49	5634	53	2012-08-02 05:33:54.433387
26062	4	1-7647	26062	49	23	3897	20	2012-08-02 05:33:54.433387
26067	4	1-7247	26066	56	7	1894	70	2012-08-02 05:33:54.433387
26068	4	1-7250	26066	7	51	2227	85	2012-08-02 05:33:54.433387
26069	4	1-7591	26066	51	49	2672	53	2012-08-02 05:33:54.433387
26066	4	1-7647	26066	49	56	1917	20	2012-08-02 05:33:54.433387
26071	5	1-7412	26070	27	42	931	12	2012-08-02 05:33:54.832478
26072	5	1-6855	26070	42	54	862	44	2012-08-02 05:33:54.832478
26073	5	1-6937	26070	54	34	444	34	2012-08-02 05:33:54.832478
26074	5	1-7571	26070	34	13	542	45	2012-08-02 05:33:54.832478
26070	5	1-7648	26070	13	27	799	8	2012-08-02 05:33:54.832478
26076	4	1-6840	26075	78	22	42	36	2012-08-02 05:33:54.832478
26077	4	1-7516	26075	22	93	75	59	2012-08-02 05:33:54.832478
26078	4	1-6061	26075	93	13	33	45	2012-08-02 05:33:54.832478
26075	4	1-7648	26075	13	78	53	8	2012-08-02 05:33:54.832478
26080	5	1-2501	26079	95	15	5109	83	2012-08-02 05:33:54.832478
26081	5	1-7512	26079	15	46	3663	69	2012-08-02 05:33:54.832478
26082	5	1-7027	26079	46	34	2559	34	2012-08-02 05:33:54.832478
26083	5	1-7571	26079	34	13	3660	45	2012-08-02 05:33:54.832478
26079	5	1-7648	26079	13	95	6318	8	2012-08-02 05:33:54.832478
26085	4	1-3486	26084	6	25	1901	57	2012-08-02 05:33:54.832478
26086	4	1-7228	26084	25	19	1107	96	2012-08-02 05:33:54.832478
26087	4	1-6518	26084	19	13	801	45	2012-08-02 05:33:54.832478
26084	4	1-7648	26084	13	6	1404	8	2012-08-02 05:33:54.832478
26089	4	1-6937	26088	54	34	775	34	2012-08-02 05:33:56.968395
26090	4	1-7571	26088	34	100	990	45	2012-08-02 05:33:56.968395
26091	4	1-6412	26088	100	94	1255	79	2012-08-02 05:33:56.968395
26088	4	1-7653	26088	94	54	1438	44	2012-08-02 05:33:56.968395
26093	2	1-4759	26092	12	83	430	85	2012-08-02 05:33:57.77591
26092	2	1-7655	26092	83	12	386	25	2012-08-02 05:33:57.77591
26095	3	1-7525	26094	18	69	689	2	2012-08-02 05:33:58.173483
26096	3	1-7465	26094	69	98	151	92	2012-08-02 05:33:58.173483
26094	3	1-7656	26094	98	18	67	3	2012-08-02 05:33:58.173483
26098	4	1-6474	26097	65	27	2346	1	2012-08-02 05:33:58.173483
26099	4	1-6407	26097	27	6	1594	56	2012-08-02 05:33:58.173483
26100	4	1-7275	26097	6	98	7573	92	2012-08-02 05:33:58.173483
26097	4	1-7656	26097	98	65	3811	3	2012-08-02 05:33:58.173483
26102	4	1-7568	26101	77	27	1224	31	2012-08-02 05:33:58.173483
26103	4	1-7345	26101	27	82	1267	95	2012-08-02 05:33:58.173483
26104	4	1-6105	26101	82	98	1044	92	2012-08-02 05:33:58.173483
26101	4	1-7656	26101	98	77	623	3	2012-08-02 05:33:58.173483
26106	4	1-7395	26105	67	4	2500	25	2012-08-02 05:33:58.173483
26107	4	1-7260	26105	4	8	1667	47	2012-08-02 05:33:58.173483
26108	4	1-5586	26105	8	98	3308	92	2012-08-02 05:33:58.173483
26105	4	1-7656	26105	98	67	1983	3	2012-08-02 05:33:58.173483
26110	5	1-6578	26109	48	49	1042	53	2012-08-02 05:33:59.788172
26111	5	1-7647	26109	49	23	592	20	2012-08-02 05:33:59.788172
26112	5	1-7149	26109	23	82	622	1	2012-08-02 05:33:59.788172
26113	5	1-7595	26109	82	23	335	27	2012-08-02 05:33:59.788172
26109	5	1-7660	26109	23	48	1068	94	2012-08-02 05:33:59.788172
26115	5	1-7318	26114	92	53	3287	71	2012-08-02 05:33:59.788172
26116	5	1-7440	26114	53	22	4158	65	2012-08-02 05:33:59.788172
26117	5	1-7600	26114	22	82	1730	1	2012-08-02 05:33:59.788172
26118	5	1-7595	26114	82	23	1089	27	2012-08-02 05:33:59.788172
26114	5	1-7660	26114	23	92	3576	94	2012-08-02 05:33:59.788172
26120	5	1-7318	26119	92	48	166	71	2012-08-02 05:33:59.788172
26121	5	1-6645	26119	48	23	79	20	2012-08-02 05:33:59.788172
26122	5	1-7149	26119	23	82	88	1	2012-08-02 05:33:59.788172
26123	5	1-7595	26119	82	23	54	27	2012-08-02 05:33:59.788172
26119	5	1-7660	26119	23	92	185	94	2012-08-02 05:33:59.788172
26125	4	1-7158	26124	83	97	116	21	2012-08-02 05:34:01.424054
26126	4	1-7457	26124	97	45	209	76	2012-08-02 05:34:01.424054
26127	4	1-6871	26124	45	55	153	81	2012-08-02 05:34:01.424054
26124	4	1-7664	26124	55	83	152	88	2012-08-02 05:34:01.424054
26129	5	1-7404	26128	57	26	4462	79	2012-08-02 05:34:01.822611
26130	5	1-7627	26128	26	46	3126	82	2012-08-02 05:34:01.822611
26131	5	1-5324	26128	46	16	2338	28	2012-08-02 05:34:01.822611
26132	5	1-7581	26128	16	30	4495	29	2012-08-02 05:34:01.822611
26128	5	1-7665	26128	30	57	5335	65	2012-08-02 05:34:01.822611
26134	2	1-3286	26133	38	52	2350	45	2012-08-02 05:34:02.221051
26133	2	1-7666	26133	52	38	4399	19	2012-08-02 05:34:02.221051
26136	4	1-4187	26135	90	89	2414	26	2012-08-02 05:34:02.221051
26137	4	1-7288	26135	89	19	1922	96	2012-08-02 05:34:02.221051
26138	4	1-6518	26135	19	52	1252	45	2012-08-02 05:34:02.221051
26135	4	1-7666	26135	52	90	3228	19	2012-08-02 05:34:02.221051
26139	1	1-4187	\N	90	90	916	26	2012-08-02 05:34:02.221051
26141	4	1-7631	26140	44	21	2598	77	2012-08-02 05:34:03.050224
26142	4	1-7577	26140	21	41	3255	87	2012-08-02 05:34:03.050224
26143	4	1-4960	26140	41	75	2160	15	2012-08-02 05:34:03.050224
26140	4	1-7668	26140	75	44	5565	63	2012-08-02 05:34:03.050224
26145	4	1-7432	26144	81	13	2491	23	2012-08-02 05:34:03.050224
26146	4	1-7428	26144	13	41	2126	87	2012-08-02 05:34:03.050224
26147	4	1-4960	26144	41	75	1531	15	2012-08-02 05:34:03.050224
26144	4	1-7668	26144	75	81	4281	63	2012-08-02 05:34:03.050224
26149	2	1-6492	26148	30	90	2252	88	2012-08-02 05:34:03.856992
26148	2	1-7670	26148	90	30	4802	77	2012-08-02 05:34:03.856992
26151	5	1-7584	26150	59	80	353	35	2012-08-02 05:34:06.747453
26152	5	1-7380	26150	80	55	168	99	2012-08-02 05:34:06.747453
26153	5	1-6728	26150	55	87	253	97	2012-08-02 05:34:06.747453
26154	5	1-7316	26150	87	89	378	48	2012-08-02 05:34:06.747453
26150	5	1-7677	26150	89	59	1078	62	2012-08-02 05:34:06.747453
26156	4	1-6116	26155	24	64	827	62	2012-08-02 05:34:07.609369
26157	4	1-7615	26155	64	82	2021	95	2012-08-02 05:34:07.609369
26158	4	1-6105	26155	82	31	1670	92	2012-08-02 05:34:07.609369
26155	4	1-7679	26155	31	24	1279	82	2012-08-02 05:34:07.609369
26160	4	1-2893	26159	5	25	123	57	2012-08-02 05:34:07.609369
26161	4	1-7228	26159	25	4	70	96	2012-08-02 05:34:07.609369
26162	4	1-5545	26159	4	31	121	92	2012-08-02 05:34:07.609369
26159	4	1-7679	26159	31	5	102	82	2012-08-02 05:34:07.609369
26164	4	1-3667	26163	92	14	1010	30	2012-08-02 05:34:09.708848
26165	4	1-7533	26163	14	45	1643	98	2012-08-02 05:34:09.708848
26166	4	1-7658	26163	45	39	4191	4	2012-08-02 05:34:09.708848
26163	4	1-7684	26163	39	92	1543	99	2012-08-02 05:34:09.708848
26168	4	1-7618	26167	76	11	2502	47	2012-08-02 05:34:10.118093
26169	4	1-6798	26167	11	57	1748	57	2012-08-02 05:34:10.118093
26170	4	1-7388	26167	57	3	1857	73	2012-08-02 05:34:10.118093
26167	4	1-7685	26167	3	76	1939	75	2012-08-02 05:34:10.118093
26172	3	1-7618	26171	76	94	3985	47	2012-08-02 05:34:10.118093
26173	3	1-4524	26171	94	3	2942	73	2012-08-02 05:34:10.118093
26171	3	1-7685	26171	3	76	3080	75	2012-08-02 05:34:10.118093
26175	3	1-7216	26174	82	10	43	26	2012-08-02 05:34:10.958727
26176	3	1-7436	26174	10	69	40	66	2012-08-02 05:34:10.958727
26174	3	1-7687	26174	69	82	40	91	2012-08-02 05:34:10.958727
26178	5	1-7431	26177	27	85	1753	24	2012-08-02 05:34:12.615534
26179	5	1-6574	26177	85	26	1446	56	2012-08-02 05:34:12.615534
26180	5	1-7639	26177	26	66	1325	33	2012-08-02 05:34:12.615534
26181	5	1-7543	26177	66	77	876	35	2012-08-02 05:34:12.615534
26177	5	1-7691	26177	77	27	1386	1	2012-08-02 05:34:12.615534
26183	4	1-6389	26182	83	77	287	94	2012-08-02 05:34:12.615534
26184	4	1-7088	26182	77	59	558	62	2012-08-02 05:34:12.615534
26185	4	1-7584	26182	59	77	278	35	2012-08-02 05:34:12.615534
26182	4	1-7691	26182	77	83	447	1	2012-08-02 05:34:12.615534
26187	5	1-4872	26186	4	59	998	83	2012-08-02 05:34:13.036109
26188	5	1-7599	26186	59	43	1930	95	2012-08-02 05:34:13.036109
26189	5	1-6414	26186	43	82	661	74	2012-08-02 05:34:13.036109
26190	5	1-7689	26186	82	44	1663	69	2012-08-02 05:34:13.036109
26186	5	1-7692	26186	44	4	956	75	2012-08-02 05:34:13.036109
26192	5	1-2406	26191	19	5	331	9	2012-08-02 05:34:13.036109
26193	5	1-7667	26191	5	17	858	23	2012-08-02 05:34:13.036109
26194	5	1-6781	26191	17	82	830	74	2012-08-02 05:34:13.036109
26195	5	1-7689	26191	82	44	2388	69	2012-08-02 05:34:13.036109
26191	5	1-7692	26191	44	19	1570	75	2012-08-02 05:34:13.036109
26197	4	1-3317	26196	57	66	1173	33	2012-08-02 05:34:13.456347
26198	4	1-7543	26196	66	77	740	35	2012-08-02 05:34:13.456347
26199	4	1-7691	26196	77	13	1118	1	2012-08-02 05:34:13.456347
26196	4	1-7693	26196	13	57	2200	55	2012-08-02 05:34:13.456347
26201	4	1-4133	26200	96	43	3858	11	2012-08-02 05:34:13.456347
26202	4	1-7281	26200	43	40	3317	39	2012-08-02 05:34:13.456347
26203	4	1-7527	26200	40	13	3197	1	2012-08-02 05:34:13.456347
26200	4	1-7693	26200	13	96	6483	55	2012-08-02 05:34:13.456347
26205	4	1-7581	26204	16	30	920	29	2012-08-02 05:34:14.318626
26206	4	1-7665	26204	30	74	501	65	2012-08-02 05:34:14.318626
26207	4	1-6525	26204	74	38	149	40	2012-08-02 05:34:14.318626
26204	4	1-7695	26204	38	16	1043	28	2012-08-02 05:34:14.318626
26209	4	1-7274	26208	77	82	1334	5	2012-08-02 05:34:15.169192
26210	4	1-7573	26208	82	83	633	45	2012-08-02 05:34:15.169192
26211	4	1-4729	26208	83	45	458	53	2012-08-02 05:34:15.169192
26208	4	1-7697	26208	45	77	2640	31	2012-08-02 05:34:15.169192
26213	4	1-6783	26212	89	54	4339	17	2012-08-02 05:34:15.169192
26214	4	1-7622	26212	54	83	984	45	2012-08-02 05:34:15.169192
26215	4	1-4729	26212	83	45	792	53	2012-08-02 05:34:15.169192
26212	4	1-7697	26212	45	89	5073	31	2012-08-02 05:34:15.169192
26217	3	1-6780	26216	56	50	1030	77	2012-08-02 05:34:15.590977
26218	3	1-7377	26216	50	1	386	22	2012-08-02 05:34:15.590977
26216	3	1-7698	26216	1	56	2401	75	2012-08-02 05:34:15.590977
26220	4	1-7162	26219	1	40	1048	39	2012-08-02 05:34:16.011264
26221	4	1-7527	26219	40	27	649	1	2012-08-02 05:34:16.011264
26222	4	1-7431	26219	27	28	518	24	2012-08-02 05:34:16.011264
26219	4	1-7699	26219	28	1	3581	76	2012-08-02 05:34:16.011264
26224	5	1-5905	26223	93	88	2031	97	2012-08-02 05:34:16.011264
26225	5	1-5548	26223	88	65	719	3	2012-08-02 05:34:16.011264
26226	5	1-6474	26223	65	27	444	1	2012-08-02 05:34:16.011264
26227	5	1-7431	26223	27	28	425	24	2012-08-02 05:34:16.011264
26223	5	1-7699	26223	28	93	3513	76	2012-08-02 05:34:16.011264
26229	5	1-5751	26228	48	69	155	66	2012-08-02 05:34:16.442487
26230	5	1-7687	26228	69	14	158	91	2012-08-02 05:34:16.442487
26231	5	1-6634	26228	14	68	254	21	2012-08-02 05:34:16.442487
26232	5	1-7589	26228	68	85	473	37	2012-08-02 05:34:16.442487
26228	5	1-7700	26228	85	48	628	87	2012-08-02 05:34:16.442487
26234	5	1-7597	26233	88	44	2346	69	2012-08-02 05:34:18.563477
26235	5	1-7692	26233	44	4	1531	75	2012-08-02 05:34:18.563477
26236	5	1-4872	26233	4	34	1813	83	2012-08-02 05:34:18.563477
26237	5	1-7628	26233	34	39	753	66	2012-08-02 05:34:18.563477
26233	5	1-7705	26233	39	88	1464	20	2012-08-02 05:34:18.563477
26239	5	1-7149	26238	23	27	1186	1	2012-08-02 05:34:18.563477
26240	5	1-7431	26238	27	68	1471	24	2012-08-02 05:34:18.563477
26241	5	1-3158	26238	68	34	1119	83	2012-08-02 05:34:18.563477
26242	5	1-7628	26238	34	39	498	66	2012-08-02 05:34:18.563477
26238	5	1-7705	26238	39	23	1039	20	2012-08-02 05:34:18.563477
26244	5	1-7354	26243	46	55	1042	64	2012-08-02 05:34:18.563477
26245	5	1-7519	26243	55	4	1990	25	2012-08-02 05:34:18.563477
26246	5	1-7260	26243	4	7	1456	47	2012-08-02 05:34:18.563477
26247	5	1-6159	26243	7	39	1475	66	2012-08-02 05:34:18.563477
26243	5	1-7705	26243	39	46	3087	20	2012-08-02 05:34:18.563477
26249	3	1-7684	26248	39	71	1245	99	2012-08-02 05:34:18.994624
26250	3	1-7042	26248	71	57	463	43	2012-08-02 05:34:18.994624
26248	3	1-7706	26248	57	39	8180	4	2012-08-02 05:34:18.994624
26252	4	1-5325	26251	59	45	294	34	2012-08-02 05:34:18.994624
26253	4	1-7001	26251	45	76	79	86	2012-08-02 05:34:18.994624
26254	4	1-6933	26251	76	57	45	43	2012-08-02 05:34:18.994624
26251	4	1-7706	26251	57	59	1207	4	2012-08-02 05:34:18.994624
26256	2	1-6104	26255	89	52	629	32	2012-08-02 05:34:19.425814
26255	2	1-7707	26255	52	89	1263	68	2012-08-02 05:34:19.425814
26258	5	1-6790	26257	34	57	2649	80	2012-08-02 05:34:22.375332
26259	5	1-7566	26257	57	3	2171	38	2012-08-02 05:34:22.375332
26260	5	1-3487	26257	3	7	1745	46	2012-08-02 05:34:22.375332
26261	5	1-7501	26257	7	99	2599	13	2012-08-02 05:34:22.375332
26257	5	1-7714	26257	99	34	2693	26	2012-08-02 05:34:22.375332
26263	4	1-7330	26262	10	60	1293	41	2012-08-02 05:34:22.375332
26264	4	1-6952	26262	60	7	1230	46	2012-08-02 05:34:22.375332
26265	4	1-7501	26262	7	99	1929	13	2012-08-02 05:34:22.375332
26262	4	1-7714	26262	99	10	2106	26	2012-08-02 05:34:22.375332
26267	5	1-7446	26266	33	52	439	64	2012-08-02 05:34:22.375332
26268	5	1-7220	26266	52	60	153	41	2012-08-02 05:34:22.375332
26269	5	1-6952	26266	60	7	162	46	2012-08-02 05:34:22.375332
26270	5	1-7501	26266	7	99	283	13	2012-08-02 05:34:22.375332
26266	5	1-7714	26266	99	33	345	26	2012-08-02 05:34:22.375332
26272	5	1-6519	26271	23	18	53	6	2012-08-02 05:34:22.795829
26273	5	1-4270	26271	18	83	56	85	2012-08-02 05:34:22.795829
26274	5	1-7655	26271	83	11	47	25	2012-08-02 05:34:22.795829
26275	5	1-6932	26271	11	75	45	73	2012-08-02 05:34:22.795829
26271	5	1-7715	26271	75	23	61	8	2012-08-02 05:34:22.795829
26277	5	1-2207	26276	66	62	4847	46	2012-08-02 05:34:22.795829
26278	5	1-6710	26276	62	83	4544	85	2012-08-02 05:34:22.795829
26279	5	1-7655	26276	83	11	4094	25	2012-08-02 05:34:22.795829
26280	5	1-6932	26276	11	75	4152	73	2012-08-02 05:34:22.795829
26276	5	1-7715	26276	75	66	6078	8	2012-08-02 05:34:22.795829
26282	3	1-7574	26281	29	88	1952	20	2012-08-02 05:34:24.077737
26283	3	1-7597	26281	88	24	2660	69	2012-08-02 05:34:24.077737
26281	3	1-7718	26281	24	29	4055	96	2012-08-02 05:34:24.077737
26285	4	1-7708	26284	46	60	1322	42	2012-08-02 05:34:24.077737
26286	4	1-7154	26284	60	55	1849	5	2012-08-02 05:34:24.077737
26287	4	1-7181	26284	55	24	1469	69	2012-08-02 05:34:24.077737
26284	4	1-7718	26284	24	46	2543	96	2012-08-02 05:34:24.077737
26288	4	1-7719	26288	6	40	1452	58	2012-08-02 05:34:24.519791
26289	4	1-7212	26288	40	81	2039	77	2012-08-02 05:34:24.519791
26290	4	1-6263	26288	81	51	1628	89	2012-08-02 05:34:24.519791
26291	4	1-7559	26288	51	6	610	7	2012-08-02 05:34:24.519791
26293	4	1-6788	26292	30	71	4252	61	2012-08-02 05:34:24.950776
26294	4	1-7327	26292	71	30	2986	71	2012-08-02 05:34:24.950776
26295	4	1-6452	26292	30	79	3472	89	2012-08-02 05:34:24.950776
26292	4	1-7720	26292	79	30	6285	47	2012-08-02 05:34:24.950776
26297	4	1-4919	26296	77	66	513	84	2012-08-02 05:34:24.950776
26298	4	1-7680	26296	66	30	365	71	2012-08-02 05:34:24.950776
26299	4	1-6452	26296	30	79	381	89	2012-08-02 05:34:24.950776
26296	4	1-7720	26296	79	77	763	47	2012-08-02 05:34:24.950776
26301	4	1-7707	26300	52	8	3242	68	2012-08-02 05:34:26.232535
26302	4	1-6919	26300	8	75	2793	55	2012-08-02 05:34:26.232535
26303	4	1-5664	26300	75	85	552	86	2012-08-02 05:34:26.232535
26300	4	1-7723	26300	85	52	1568	32	2012-08-02 05:34:26.232535
26305	5	1-7707	26304	52	80	1694	68	2012-08-02 05:34:26.232535
26306	5	1-4680	26304	80	17	1281	44	2012-08-02 05:34:26.232535
26307	5	1-5004	26304	17	7	260	89	2012-08-02 05:34:26.232535
26308	5	1-7616	26304	7	85	233	86	2012-08-02 05:34:26.232535
26304	5	1-7723	26304	85	52	736	32	2012-08-02 05:34:26.232535
26310	3	1-7637	26309	20	97	5617	69	2012-08-02 05:34:26.232535
26311	3	1-6916	26309	97	85	1784	86	2012-08-02 05:34:26.232535
26309	3	1-7723	26309	85	20	5740	32	2012-08-02 05:34:26.232535
26313	5	1-7503	26312	98	53	5253	14	2012-08-02 05:34:27.513884
26314	5	1-7254	26312	53	73	3507	95	2012-08-02 05:34:27.513884
26315	5	1-7476	26312	73	97	392	23	2012-08-02 05:34:27.513884
26316	5	1-7557	26312	97	96	259	19	2012-08-02 05:34:27.513884
26312	5	1-7726	26312	96	98	2604	93	2012-08-02 05:34:27.513884
26318	5	1-7687	26317	69	21	362	91	2012-08-02 05:34:28.375899
26319	5	1-7097	26317	21	26	491	71	2012-08-02 05:34:28.375899
26320	5	1-7661	26317	26	30	404	30	2012-08-02 05:34:28.375899
26321	5	1-6119	26317	30	94	388	11	2012-08-02 05:34:28.375899
26317	5	1-7728	26317	94	69	385	66	2012-08-02 05:34:28.375899
26323	3	1-5517	26322	76	5	728	9	2012-08-02 05:34:28.807255
26324	3	1-7667	26322	5	91	1868	23	2012-08-02 05:34:28.807255
26322	3	1-7729	26322	91	76	1634	19	2012-08-02 05:34:28.807255
26326	5	1-4744	26325	63	66	849	84	2012-08-02 05:34:29.702152
26327	5	1-7680	26325	66	26	768	71	2012-08-02 05:34:29.702152
26328	5	1-7661	26325	26	71	672	30	2012-08-02 05:34:29.702152
26329	5	1-4497	26325	71	48	449	27	2012-08-02 05:34:29.702152
26325	5	1-7731	26325	48	63	2208	63	2012-08-02 05:34:29.702152
26331	4	1-7692	26330	44	96	109	75	2012-08-02 05:34:30.132875
26332	4	1-3889	26330	96	7	87	70	2012-08-02 05:34:30.132875
26333	4	1-7250	26330	7	12	85	85	2012-08-02 05:34:30.132875
26330	4	1-7732	26330	12	44	184	69	2012-08-02 05:34:30.132875
26335	4	1-7692	26334	44	56	1107	75	2012-08-02 05:34:30.132875
26336	4	1-6780	26334	56	39	797	77	2012-08-02 05:34:30.132875
26337	4	1-5471	26334	39	12	783	85	2012-08-02 05:34:30.132875
26334	4	1-7732	26334	12	44	1781	69	2012-08-02 05:34:30.132875
26339	4	1-6786	26338	80	82	209	5	2012-08-02 05:34:30.132875
26340	4	1-7573	26338	82	89	102	45	2012-08-02 05:34:30.132875
26341	4	1-7025	26338	89	12	100	85	2012-08-02 05:34:30.132875
26338	4	1-7732	26338	12	80	239	69	2012-08-02 05:34:30.132875
26343	5	1-7261	26342	89	81	2047	63	2012-08-02 05:34:31.005942
26344	5	1-7432	26342	81	91	993	23	2012-08-02 05:34:31.005942
26345	5	1-7729	26342	91	78	665	19	2012-08-02 05:34:31.005942
26346	5	1-7733	26342	78	99	575	39	2012-08-02 05:34:31.005942
26342	5	1-7734	26342	99	89	2161	3	2012-08-02 05:34:31.005942
26348	4	1-6474	26347	65	82	632	1	2012-08-02 05:34:31.005942
26349	4	1-7595	26347	82	66	370	27	2012-08-02 05:34:31.005942
26350	4	1-5736	26347	66	99	192	39	2012-08-02 05:34:31.005942
26347	4	1-7734	26347	99	65	894	3	2012-08-02 05:34:31.005942
26352	5	1-7261	26351	89	27	3847	63	2012-08-02 05:34:31.005942
26353	5	1-6909	26351	27	53	3012	50	2012-08-02 05:34:31.005942
26354	5	1-7471	26351	53	78	501	19	2012-08-02 05:34:31.005942
26355	5	1-7733	26351	78	99	587	39	2012-08-02 05:34:31.005942
26351	5	1-7734	26351	99	89	2995	3	2012-08-02 05:34:31.005942
26357	5	1-7395	26356	67	53	827	25	2012-08-02 05:34:31.005942
26358	5	1-3386	26356	53	76	302	97	2012-08-02 05:34:31.005942
26359	5	1-3910	26356	76	43	138	11	2012-08-02 05:34:31.005942
26360	5	1-7281	26356	43	99	117	39	2012-08-02 05:34:31.005942
26356	5	1-7734	26356	99	67	609	3	2012-08-02 05:34:31.005942
26361	1	1-3910	\N	76	76	6	11	2012-08-02 05:34:31.005942
26363	5	1-7523	26362	62	12	751	85	2012-08-02 05:34:32.2984
26364	5	1-7732	26362	12	22	1957	69	2012-08-02 05:34:32.2984
26365	5	1-1100	26362	22	3	575	22	2012-08-02 05:34:32.2984
26366	5	1-7153	26362	3	33	420	99	2012-08-02 05:34:32.2984
26362	5	1-7737	26362	33	62	479	27	2012-08-02 05:34:32.2984
26368	4	1-6141	26367	8	98	1983	93	2012-08-02 05:34:32.718288
26369	4	1-7503	26367	98	48	2606	14	2012-08-02 05:34:32.718288
26370	4	1-7422	26367	48	68	1132	87	2012-08-02 05:34:32.718288
26367	4	1-7738	26367	68	8	4130	27	2012-08-02 05:34:32.718288
26372	5	1-6930	26371	22	3	1296	73	2012-08-02 05:34:32.718288
26373	5	1-7685	26371	3	56	1129	75	2012-08-02 05:34:32.718288
26374	5	1-6780	26371	56	21	751	77	2012-08-02 05:34:32.718288
26375	5	1-7577	26371	21	68	982	87	2012-08-02 05:34:32.718288
26371	5	1-7738	26371	68	22	4483	27	2012-08-02 05:34:32.718288
26377	4	1-7153	26376	3	33	1432	99	2012-08-02 05:34:33.127291
26378	4	1-7737	26376	33	12	1432	27	2012-08-02 05:34:33.127291
26379	4	1-7594	26376	12	41	3560	42	2012-08-02 05:34:33.127291
26376	4	1-7739	26376	41	3	2233	22	2012-08-02 05:34:33.127291
26381	5	1-2974	26380	35	24	821	51	2012-08-02 05:34:34.795356
26382	5	1-7625	26380	24	68	352	52	2012-08-02 05:34:34.795356
26383	5	1-5921	26380	68	45	548	53	2012-08-02 05:34:34.795356
26384	5	1-7697	26380	45	77	2226	31	2012-08-02 05:34:34.795356
26380	5	1-7743	26380	77	35	2052	96	2012-08-02 05:34:34.795356
26386	4	1-7514	26385	7	55	1274	64	2012-08-02 05:34:34.795356
26387	4	1-7519	26385	55	80	2101	25	2012-08-02 05:34:34.795356
26388	4	1-5268	26385	80	77	1304	31	2012-08-02 05:34:34.795356
26385	4	1-7743	26385	77	7	1621	96	2012-08-02 05:34:34.795356
26390	4	1-5545	26389	4	31	660	92	2012-08-02 05:34:34.795356
26391	4	1-7679	26389	31	67	487	82	2012-08-02 05:34:34.795356
26392	4	1-409	26389	67	77	347	31	2012-08-02 05:34:34.795356
26389	4	1-7743	26389	77	4	438	96	2012-08-02 05:34:34.795356
26394	3	1-7409	26393	49	56	5946	6	2012-08-02 05:34:35.215414
26395	3	1-7142	26393	56	45	3322	92	2012-08-02 05:34:35.215414
26393	3	1-7744	26393	45	49	6174	72	2012-08-02 05:34:35.215414
26397	5	1-4700	26396	59	26	292	71	2012-08-02 05:34:36.055199
26398	5	1-7661	26396	26	17	244	30	2012-08-02 05:34:36.055199
26399	5	1-6844	26396	17	4	301	25	2012-08-02 05:34:36.055199
26400	5	1-7260	26396	4	94	202	47	2012-08-02 05:34:36.055199
26396	5	1-7746	26396	94	59	1305	78	2012-08-02 05:34:36.055199
26402	5	1-5918	26401	27	29	847	24	2012-08-02 05:34:36.055199
26403	5	1-7541	26401	29	58	2149	4	2012-08-02 05:34:36.055199
26404	5	1-7234	26401	58	4	800	25	2012-08-02 05:34:36.055199
26405	5	1-7260	26401	4	94	543	47	2012-08-02 05:34:36.055199
26401	5	1-7746	26401	94	27	3558	78	2012-08-02 05:34:36.055199
26407	5	1-6440	26406	36	7	897	70	2012-08-02 05:34:36.055199
26408	5	1-7250	26406	7	83	972	85	2012-08-02 05:34:36.055199
26409	5	1-7655	26406	83	4	803	25	2012-08-02 05:34:36.055199
26410	5	1-7260	26406	4	94	554	47	2012-08-02 05:34:36.055199
26406	5	1-7746	26406	94	36	3691	78	2012-08-02 05:34:36.055199
26411	1	1-6792	\N	90	90	1777	47	2012-08-02 05:34:36.055199
26413	3	1-6927	26412	81	79	1002	61	2012-08-02 05:34:36.894738
26414	3	1-6424	26412	79	82	451	31	2012-08-02 05:34:36.894738
26412	3	1-7748	26412	82	81	1860	39	2012-08-02 05:34:36.894738
26416	4	1-5794	26415	93	2	3184	46	2012-08-02 05:34:39.868618
26417	4	1-7518	26415	2	28	1016	62	2012-08-02 05:34:39.868618
26418	4	1-7169	26415	28	21	899	68	2012-08-02 05:34:39.868618
26415	4	1-7755	26415	21	93	2890	55	2012-08-02 05:34:39.868618
26420	2	1-6452	26419	30	57	1641	89	2012-08-02 05:34:40.299307
26419	2	1-7756	26419	57	30	1207	71	2012-08-02 05:34:40.299307
26422	4	1-7729	26421	91	78	1464	19	2012-08-02 05:34:44.530275
26423	4	1-7716	26421	78	98	1066	68	2012-08-02 05:34:44.530275
26424	4	1-7272	26421	98	46	2040	40	2012-08-02 05:34:44.530275
26421	4	1-7766	26421	46	91	2078	23	2012-08-02 05:34:44.530275
26426	4	1-7557	26425	97	78	113	19	2012-08-02 05:34:44.530275
26427	4	1-7716	26425	78	98	90	68	2012-08-02 05:34:44.530275
26428	4	1-7272	26425	98	46	188	40	2012-08-02 05:34:44.530275
26425	4	1-7766	26425	46	97	210	23	2012-08-02 05:34:44.530275
26430	2	1-3732	26429	39	46	5322	40	2012-08-02 05:34:44.530275
26429	2	1-7766	26429	46	39	7080	23	2012-08-02 05:34:44.530275
26432	5	1-7465	26431	69	31	2282	92	2012-08-02 05:34:44.961413
26433	5	1-7679	26431	31	54	1692	82	2012-08-02 05:34:44.961413
26434	5	1-5753	26431	54	99	1305	13	2012-08-02 05:34:44.961413
26435	5	1-7714	26431	99	76	1378	26	2012-08-02 05:34:44.961413
26431	5	1-7767	26431	76	69	7974	2	2012-08-02 05:34:44.961413
26437	5	1-4983	26436	9	52	516	50	2012-08-02 05:34:44.961413
26438	5	1-7736	26436	52	22	556	100	2012-08-02 05:34:44.961413
26439	5	1-3959	26436	22	99	259	13	2012-08-02 05:34:44.961413
26440	5	1-7714	26436	99	76	292	26	2012-08-02 05:34:44.961413
26436	5	1-7767	26436	76	9	1805	2	2012-08-02 05:34:44.961413
26442	5	1-7465	26441	69	78	59	92	2012-08-02 05:34:44.961413
26443	5	1-7711	26441	78	30	37	1	2012-08-02 05:34:44.961413
26444	5	1-4742	26441	30	99	22	13	2012-08-02 05:34:44.961413
26445	5	1-7714	26441	99	76	27	26	2012-08-02 05:34:44.961413
26441	5	1-7767	26441	76	69	178	2	2012-08-02 05:34:44.961413
26447	5	1-7341	26446	68	12	395	18	2012-08-02 05:34:45.812198
26448	5	1-7123	26446	12	58	227	37	2012-08-02 05:34:45.812198
26449	5	1-7256	26446	58	16	85	28	2012-08-02 05:34:45.812198
26450	5	1-7581	26446	16	78	127	29	2012-08-02 05:34:45.812198
26446	5	1-7769	26446	78	68	399	100	2012-08-02 05:34:45.812198
26452	4	1-7100	26451	75	66	1758	33	2012-08-02 05:34:45.812198
26453	4	1-7543	26451	66	42	898	35	2012-08-02 05:34:45.812198
26454	4	1-6590	26451	42	78	657	29	2012-08-02 05:34:45.812198
26451	4	1-7769	26451	78	75	2092	100	2012-08-02 05:34:45.812198
26456	3	1-5246	26455	72	56	933	11	2012-08-02 05:34:45.812198
26457	3	1-7724	26455	56	78	796	29	2012-08-02 05:34:45.812198
26455	3	1-7769	26455	78	72	2670	100	2012-08-02 05:34:45.812198
26459	5	1-7393	26458	38	26	481	71	2012-08-02 05:34:46.243462
26460	5	1-7661	26458	26	30	340	30	2012-08-02 05:34:46.243462
26461	5	1-6119	26458	30	43	281	11	2012-08-02 05:34:46.243462
26462	5	1-7281	26458	43	9	188	39	2012-08-02 05:34:46.243462
26458	5	1-7770	26458	9	38	496	35	2012-08-02 05:34:46.243462
26464	4	1-7380	26463	80	33	1796	99	2012-08-02 05:34:46.243462
26465	4	1-7737	26463	33	66	1797	27	2012-08-02 05:34:46.243462
26466	4	1-5736	26463	66	9	946	39	2012-08-02 05:34:46.243462
26463	4	1-7770	26463	9	80	2862	35	2012-08-02 05:34:46.243462
26468	5	1-7521	26467	74	18	474	82	2012-08-02 05:34:46.243462
26469	5	1-7702	26467	18	97	388	23	2012-08-02 05:34:46.243462
26470	5	1-7557	26467	97	78	233	19	2012-08-02 05:34:46.243462
26471	5	1-7733	26467	78	9	258	39	2012-08-02 05:34:46.243462
26467	5	1-7770	26467	9	74	796	35	2012-08-02 05:34:46.243462
26472	1	1-5736	\N	66	66	1091	39	2012-08-02 05:34:46.243462
26474	3	1-7256	26473	58	16	1833	28	2012-08-02 05:34:46.685556
26475	3	1-7581	26473	16	86	2259	29	2012-08-02 05:34:46.685556
26473	3	1-7771	26473	86	58	5976	37	2012-08-02 05:34:46.685556
26477	5	1-1723	26476	30	40	46	58	2012-08-02 05:34:46.685556
26478	5	1-7212	26476	40	37	65	77	2012-08-02 05:34:46.685556
26479	5	1-6516	26476	37	16	51	28	2012-08-02 05:34:46.685556
26480	5	1-7581	26476	16	86	99	29	2012-08-02 05:34:46.685556
26476	5	1-7771	26476	86	30	413	37	2012-08-02 05:34:46.685556
26482	4	1-7424	26481	43	59	1310	83	2012-08-02 05:34:47.922975
26483	4	1-7599	26481	59	14	2798	95	2012-08-02 05:34:47.922975
26484	4	1-5125	26481	14	15	1041	80	2012-08-02 05:34:47.922975
26481	4	1-7774	26481	15	43	1591	17	2012-08-02 05:34:47.922975
26486	5	1-7113	26485	41	69	485	2	2012-08-02 05:34:47.922975
26487	5	1-7465	26485	69	31	154	92	2012-08-02 05:34:47.922975
26488	5	1-7679	26485	31	20	127	82	2012-08-02 05:34:47.922975
26489	5	1-5830	26485	20	15	149	80	2012-08-02 05:34:47.922975
26485	5	1-7774	26485	15	41	247	17	2012-08-02 05:34:47.922975
26491	4	1-5606	26490	1	63	2402	67	2012-08-02 05:34:48.740675
26492	4	1-6866	26490	63	76	1564	86	2012-08-02 05:34:48.740675
26493	4	1-6933	26490	76	81	1270	43	2012-08-02 05:34:48.740675
26490	4	1-7776	26490	81	1	8137	100	2012-08-02 05:34:48.740675
26495	5	1-6988	26494	51	60	470	7	2012-08-02 05:34:48.740675
26496	5	1-7683	26494	60	92	280	21	2012-08-02 05:34:48.740675
26497	5	1-7437	26494	92	71	137	99	2012-08-02 05:34:48.740675
26498	5	1-7042	26494	71	81	118	43	2012-08-02 05:34:48.740675
26494	5	1-7776	26494	81	51	796	100	2012-08-02 05:34:48.740675
26500	4	1-5373	26499	84	33	715	59	2012-08-02 05:34:50.431142
26501	4	1-7439	26499	33	45	2710	14	2012-08-02 05:34:50.431142
26502	4	1-6701	26499	45	95	1021	81	2012-08-02 05:34:50.431142
26499	4	1-7780	26499	95	84	1471	64	2012-08-02 05:34:50.431142
26504	4	1-5120	26503	31	12	1766	85	2012-08-02 05:34:50.431142
26505	4	1-7732	26503	12	77	4079	69	2012-08-02 05:34:50.431142
26506	4	1-6287	26503	77	95	2053	81	2012-08-02 05:34:50.431142
26503	4	1-7780	26503	95	31	3134	64	2012-08-02 05:34:50.431142
26508	2	1-6196	26507	33	95	1207	81	2012-08-02 05:34:50.431142
26507	2	1-7780	26507	95	33	2038	64	2012-08-02 05:34:50.431142
26510	4	1-7743	26509	77	29	2364	96	2012-08-02 05:34:52.552502
26511	4	1-7574	26509	29	23	977	20	2012-08-02 05:34:52.552502
26512	4	1-7149	26509	23	56	760	1	2012-08-02 05:34:52.552502
26509	4	1-7785	26509	56	77	2420	31	2012-08-02 05:34:52.552502
26514	4	1-7743	26513	77	7	1067	96	2012-08-02 05:34:52.552502
26515	4	1-7514	26513	7	79	773	64	2012-08-02 05:34:52.552502
26516	4	1-3161	26513	79	56	249	1	2012-08-02 05:34:52.552502
26513	4	1-7785	26513	56	77	931	31	2012-08-02 05:34:52.552502
26518	5	1-7307	26517	14	26	3868	58	2012-08-02 05:34:52.983544
26519	5	1-6691	26517	26	26	819	71	2012-08-02 05:34:52.983544
26520	5	1-7661	26517	26	14	751	30	2012-08-02 05:34:52.983544
26521	5	1-7533	26517	14	34	1302	98	2012-08-02 05:34:52.983544
26517	5	1-7786	26517	34	14	4038	67	2012-08-02 05:34:52.983544
26523	4	1-5122	26522	17	26	511	71	2012-08-02 05:34:52.983544
26524	4	1-7661	26522	26	14	464	30	2012-08-02 05:34:52.983544
26525	4	1-7533	26522	14	34	796	98	2012-08-02 05:34:52.983544
26522	4	1-7786	26522	34	17	2521	67	2012-08-02 05:34:52.983544
26527	4	1-4500	26526	56	4	447	60	2012-08-02 05:34:52.983544
26528	4	1-6011	26526	4	97	460	37	2012-08-02 05:34:52.983544
26529	4	1-7607	26526	97	34	201	98	2012-08-02 05:34:52.983544
26526	4	1-7786	26526	34	56	644	67	2012-08-02 05:34:52.983544
26531	5	1-7541	26530	29	25	118	4	2012-08-02 05:34:54.673898
26532	5	1-6533	26530	25	28	84	59	2012-08-02 05:34:54.673898
26533	5	1-6238	26530	28	49	32	72	2012-08-02 05:34:54.673898
26534	5	1-7409	26530	49	7	42	6	2012-08-02 05:34:54.673898
26530	5	1-7790	26530	7	29	43	24	2012-08-02 05:34:54.673898
26536	4	1-7581	26535	16	30	1858	29	2012-08-02 05:34:55.104943
26537	4	1-7665	26535	30	74	1435	65	2012-08-02 05:34:55.104943
26538	4	1-6525	26535	74	82	603	40	2012-08-02 05:34:55.104943
26535	4	1-7791	26535	82	16	1486	28	2012-08-02 05:34:55.104943
26540	4	1-7758	26539	46	53	6432	50	2012-08-02 05:34:55.104943
26541	4	1-7471	26539	53	41	938	19	2012-08-02 05:34:55.104943
26542	4	1-6214	26539	41	82	649	40	2012-08-02 05:34:55.104943
26539	4	1-7791	26539	82	46	2096	28	2012-08-02 05:34:55.104943
26544	4	1-7758	26543	46	78	2606	50	2012-08-02 05:34:55.104943
26545	4	1-4559	26543	78	78	307	39	2012-08-02 05:34:55.104943
26546	4	1-4868	26543	78	82	254	40	2012-08-02 05:34:55.104943
26543	4	1-7791	26543	82	46	834	28	2012-08-02 05:34:55.104943
26548	5	1-6942	26547	48	77	817	31	2012-08-02 05:34:55.104943
26549	5	1-7743	26547	77	56	1041	96	2012-08-02 05:34:55.104943
26550	5	1-7323	26547	56	23	301	99	2012-08-02 05:34:55.104943
26551	5	1-5772	26547	23	82	337	40	2012-08-02 05:34:55.104943
26547	5	1-7791	26547	82	48	1142	28	2012-08-02 05:34:55.104943
26553	5	1-7221	26552	67	77	302	31	2012-08-02 05:34:55.944778
26554	5	1-7743	26552	77	60	369	96	2012-08-02 05:34:55.944778
26555	5	1-7563	26552	60	71	269	68	2012-08-02 05:34:55.944778
26556	5	1-6002	26552	71	83	259	83	2012-08-02 05:34:55.944778
26552	5	1-7793	26552	83	67	1285	88	2012-08-02 05:34:55.944778
26558	4	1-6837	26557	100	91	263	75	2012-08-02 05:34:55.944778
26559	4	1-6924	26557	91	71	132	68	2012-08-02 05:34:55.944778
26560	4	1-6002	26557	71	83	138	83	2012-08-02 05:34:55.944778
26557	4	1-7793	26557	83	100	737	88	2012-08-02 05:34:55.944778
26562	4	1-7355	26561	8	48	2502	14	2012-08-02 05:34:55.944778
26563	4	1-7422	26561	48	16	1624	87	2012-08-02 05:34:55.944778
26564	4	1-4332	26561	16	83	431	83	2012-08-02 05:34:55.944778
26561	4	1-7793	26561	83	8	2454	88	2012-08-02 05:34:55.944778
26566	4	1-6845	26565	28	7	818	48	2012-08-02 05:34:58.883271
26567	4	1-6500	26565	7	15	497	80	2012-08-02 05:34:58.883271
26568	4	1-7774	26565	15	61	775	17	2012-08-02 05:34:58.883271
26565	4	1-7800	26565	61	28	674	58	2012-08-02 05:34:58.883271
26570	4	1-6988	26569	51	55	4102	7	2012-08-02 05:34:59.78922
26571	4	1-7513	26569	55	36	3754	32	2012-08-02 05:34:59.78922
26572	4	1-6434	26569	36	30	2322	33	2012-08-02 05:34:59.78922
26569	4	1-7802	26569	30	51	8978	100	2012-08-02 05:34:59.78922
26574	5	1-6032	26573	78	31	1543	92	2012-08-02 05:35:00.209115
26575	5	1-7679	26573	31	18	1303	82	2012-08-02 05:35:00.209115
26576	5	1-7702	26573	18	17	1176	23	2012-08-02 05:35:00.209115
26577	5	1-6781	26573	17	40	1218	74	2012-08-02 05:35:00.209115
26573	5	1-7803	26573	40	78	1605	33	2012-08-02 05:35:00.209115
26579	4	1-6905	26578	49	61	216	90	2012-08-02 05:35:01.114906
26580	4	1-6572	26578	61	85	209	63	2012-08-02 05:35:01.114906
26581	4	1-7787	26578	85	64	196	40	2012-08-02 05:35:01.114906
26578	4	1-7805	26578	64	49	283	75	2012-08-02 05:35:01.114906
26583	4	1-7742	26582	33	20	5140	46	2012-08-02 05:35:01.114906
26584	4	1-7357	26582	20	85	4335	63	2012-08-02 05:35:01.114906
26585	4	1-7787	26582	85	64	4108	40	2012-08-02 05:35:01.114906
26582	4	1-7805	26582	64	33	5989	75	2012-08-02 05:35:01.114906
26587	5	1-7627	26586	26	18	877	82	2012-08-02 05:35:01.954714
26588	5	1-7702	26586	18	97	804	23	2012-08-02 05:35:01.954714
26589	5	1-7557	26586	97	40	540	19	2012-08-02 05:35:01.954714
26590	5	1-5008	26586	40	63	579	56	2012-08-02 05:35:01.954714
26586	5	1-7807	26586	63	26	1216	79	2012-08-02 05:35:01.954714
26592	5	1-6824	26591	17	57	754	80	2012-08-02 05:35:02.794395
26593	5	1-7566	26591	57	16	701	38	2012-08-02 05:35:02.794395
26594	5	1-3275	26591	16	92	250	74	2012-08-02 05:35:02.794395
26595	5	1-6875	26591	92	36	163	12	2012-08-02 05:35:02.794395
26591	5	1-7809	26591	36	17	703	90	2012-08-02 05:35:02.794395
26597	4	1-7196	26596	13	98	6152	57	2012-08-02 05:35:02.794395
26598	4	1-7491	26596	98	25	2975	93	2012-08-02 05:35:02.794395
26599	4	1-7363	26596	25	36	1468	12	2012-08-02 05:35:02.794395
26596	4	1-7809	26596	36	13	6424	90	2012-08-02 05:35:02.794395
26601	4	1-5657	26600	79	20	2786	53	2012-08-02 05:35:02.794395
26602	4	1-4921	26600	20	92	775	74	2012-08-02 05:35:02.794395
26603	4	1-6875	26600	92	36	523	12	2012-08-02 05:35:02.794395
26600	4	1-7809	26600	36	79	2342	90	2012-08-02 05:35:02.794395
26605	5	1-7721	26604	30	88	1727	67	2012-08-02 05:35:04.06514
26606	5	1-5493	26604	88	42	1059	34	2012-08-02 05:35:04.06514
26607	5	1-6730	26604	42	31	3559	92	2012-08-02 05:35:04.06514
26608	5	1-7679	26604	31	25	2814	82	2012-08-02 05:35:04.06514
26604	5	1-7812	26604	25	30	3683	57	2012-08-02 05:35:04.06514
26610	3	1-6425	26609	25	92	3595	46	2012-08-02 05:35:05.813788
26611	3	1-7789	26609	92	17	1553	83	2012-08-02 05:35:05.813788
26609	3	1-7816	26609	17	25	6089	43	2012-08-02 05:35:05.813788
26613	2	1-6973	26612	77	42	986	89	2012-08-02 05:35:08.310743
26612	2	1-7822	26612	42	77	7679	30	2012-08-02 05:35:08.310743
26615	3	1-6965	26614	71	4	482	7	2012-08-02 05:35:10.454298
26616	3	1-6446	26614	4	75	352	79	2012-08-02 05:35:10.454298
26614	3	1-7827	26614	75	71	976	60	2012-08-02 05:35:10.454298
26618	5	1-7762	26617	84	84	1417	77	2012-08-02 05:35:10.454298
26619	5	1-2410	26617	84	20	333	41	2012-08-02 05:35:10.454298
26620	5	1-7414	26617	20	63	286	56	2012-08-02 05:35:10.454298
26621	5	1-7807	26617	63	75	546	79	2012-08-02 05:35:10.454298
26617	5	1-7827	26617	75	84	1541	60	2012-08-02 05:35:10.454298
26623	5	1-2053	26622	26	97	511	23	2012-08-02 05:35:10.454298
26624	5	1-7557	26622	97	40	323	19	2012-08-02 05:35:10.454298
26625	5	1-5008	26622	40	63	325	56	2012-08-02 05:35:10.454298
26626	5	1-7807	26622	63	75	643	79	2012-08-02 05:35:10.454298
26622	5	1-7827	26622	75	26	1882	60	2012-08-02 05:35:10.454298
26628	4	1-7414	26627	20	63	629	56	2012-08-02 05:35:11.305134
26629	4	1-7807	26627	63	94	542	79	2012-08-02 05:35:11.305134
26630	4	1-7653	26627	94	87	286	44	2012-08-02 05:35:11.305134
26627	4	1-7829	26627	87	20	1621	41	2012-08-02 05:35:11.305134
26632	5	1-7685	26631	3	64	124	75	2012-08-02 05:35:11.735741
26633	5	1-7413	26631	64	19	80	61	2012-08-02 05:35:11.735741
26634	5	1-7824	26631	19	78	119	97	2012-08-02 05:35:11.735741
26635	5	1-6471	26631	78	31	79	72	2012-08-02 05:35:11.735741
26631	5	1-7830	26631	31	3	119	73	2012-08-02 05:35:11.735741
26637	4	1-7505	26636	14	56	5042	76	2012-08-02 05:35:12.166707
26638	4	1-7490	26636	56	81	5378	16	2012-08-02 05:35:12.166707
26639	4	1-5527	26636	81	76	3142	54	2012-08-02 05:35:12.166707
26636	4	1-7831	26636	76	14	3656	91	2012-08-02 05:35:12.166707
26641	3	1-7216	26640	82	43	28	26	2012-08-02 05:35:12.166707
26642	3	1-5263	26640	43	76	23	54	2012-08-02 05:35:12.166707
26640	3	1-7831	26640	76	82	27	91	2012-08-02 05:35:12.166707
26644	4	1-5266	26643	84	60	129	42	2012-08-02 05:35:12.166707
26645	4	1-7154	26643	60	3	189	5	2012-08-02 05:35:12.166707
26646	4	1-5738	26643	3	76	168	54	2012-08-02 05:35:12.166707
26643	4	1-7831	26643	76	84	211	91	2012-08-02 05:35:12.166707
26648	4	1-6033	26647	8	60	430	42	2012-08-02 05:35:12.166707
26649	4	1-7154	26647	60	3	686	5	2012-08-02 05:35:12.166707
26650	4	1-5738	26647	3	76	661	54	2012-08-02 05:35:12.166707
26647	4	1-7831	26647	76	8	902	91	2012-08-02 05:35:12.166707
26652	2	1-6817	26651	28	41	132	97	2012-08-02 05:35:12.597715
26651	2	1-7832	26651	41	28	310	70	2012-08-02 05:35:12.597715
26654	4	1-7250	26653	7	83	832	85	2012-08-02 05:35:12.597715
26655	4	1-7655	26653	83	53	702	25	2012-08-02 05:35:12.597715
26656	4	1-3386	26653	53	41	252	97	2012-08-02 05:35:12.597715
26653	4	1-7832	26653	41	7	751	70	2012-08-02 05:35:12.597715
26658	5	1-7700	26657	85	94	969	87	2012-08-02 05:35:13.86821
26659	5	1-5870	26657	94	8	713	50	2012-08-02 05:35:13.86821
26660	5	1-7200	26657	8	76	390	54	2012-08-02 05:35:13.86821
26661	5	1-7831	26657	76	65	534	91	2012-08-02 05:35:13.86821
26657	5	1-7835	26657	65	85	696	37	2012-08-02 05:35:13.86821
26663	4	1-6720	26662	63	41	1727	42	2012-08-02 05:35:14.741288
26664	4	1-7739	26662	41	94	1259	22	2012-08-02 05:35:14.741288
26665	4	1-7548	26662	94	65	2177	15	2012-08-02 05:35:14.741288
26662	4	1-7837	26662	65	63	1701	64	2012-08-02 05:35:14.741288
26667	4	1-7700	26666	85	41	2045	87	2012-08-02 05:35:16.000659
26668	4	1-4960	26666	41	75	1332	15	2012-08-02 05:35:16.000659
26669	4	1-6726	26666	75	72	461	16	2012-08-02 05:35:16.000659
26666	4	1-7840	26666	72	85	1960	37	2012-08-02 05:35:16.000659
26671	4	1-7072	26670	99	77	2300	60	2012-08-02 05:35:16.431355
26672	4	1-7564	26670	77	30	806	29	2012-08-02 05:35:16.431355
26673	4	1-7665	26670	30	99	974	65	2012-08-02 05:35:16.431355
26670	4	1-7841	26670	99	99	1358	78	2012-08-02 05:35:16.431355
26675	4	1-5635	26674	44	78	1431	92	2012-08-02 05:35:17.701653
26676	4	1-7711	26674	78	56	851	1	2012-08-02 05:35:17.701653
26677	4	1-5947	26674	56	71	384	41	2012-08-02 05:35:17.701653
26674	4	1-7844	26674	71	44	3611	5	2012-08-02 05:35:17.701653
26679	3	1-3275	26678	16	8	2263	74	2012-08-02 05:35:18.541312
26680	3	1-7811	26678	8	55	5377	31	2012-08-02 05:35:18.541312
26678	3	1-7846	26678	55	16	6827	38	2012-08-02 05:35:18.541312
26682	2	1-7817	26681	2	55	966	31	2012-08-02 05:35:18.541312
26681	2	1-7846	26681	55	2	1361	38	2012-08-02 05:35:18.541312
26684	4	1-7830	26683	31	10	1696	73	2012-08-02 05:35:20.220562
26685	4	1-5108	26683	10	8	100	18	2012-08-02 05:35:20.220562
26686	4	1-7778	26683	8	89	1302	92	2012-08-02 05:35:20.220562
26683	4	1-7850	26683	89	31	1539	72	2012-08-02 05:35:20.220562
26688	4	1-6838	26687	79	97	752	66	2012-08-02 05:35:20.220562
26689	4	1-5631	26687	97	8	232	18	2012-08-02 05:35:20.220562
26690	4	1-7778	26687	8	89	3405	92	2012-08-02 05:35:20.220562
26687	4	1-7850	26687	89	79	4542	72	2012-08-02 05:35:20.220562
26692	4	1-7181	26691	55	77	585	69	2012-08-02 05:35:21.049183
26693	4	1-6287	26691	77	33	185	81	2012-08-02 05:35:21.049183
26694	4	1-7510	26691	33	1	73	28	2012-08-02 05:35:21.049183
26691	4	1-7852	26691	1	55	1168	5	2012-08-02 05:35:21.049183
26696	4	1-5635	26695	44	90	1643	92	2012-08-02 05:35:21.049183
26697	4	1-7252	26695	90	33	427	81	2012-08-02 05:35:21.049183
26698	4	1-7510	26695	33	1	233	28	2012-08-02 05:35:21.049183
26695	4	1-7852	26695	1	44	5174	5	2012-08-02 05:35:21.049183
26700	2	1-7363	26699	25	78	1125	12	2012-08-02 05:35:22.330755
26699	2	1-7855	26699	78	25	2348	93	2012-08-02 05:35:22.330755
26702	2	1-7705	26701	39	80	679	20	2012-08-02 05:35:22.750549
26701	2	1-7856	26701	80	39	676	66	2012-08-02 05:35:22.750549
26704	4	1-7587	26703	71	14	1985	67	2012-08-02 05:35:22.750549
26705	4	1-7307	26703	14	90	1863	58	2012-08-02 05:35:22.750549
26706	4	1-7383	26703	90	80	496	20	2012-08-02 05:35:22.750549
26703	4	1-7856	26703	80	71	1014	66	2012-08-02 05:35:22.750549
26708	4	1-7536	26707	54	24	769	10	2012-08-02 05:35:23.60119
26709	4	1-7567	26707	24	33	337	59	2012-08-02 05:35:23.60119
26710	4	1-7439	26707	33	73	1319	14	2012-08-02 05:35:23.60119
26707	4	1-7858	26707	73	54	968	40	2012-08-02 05:35:23.60119
26712	4	1-7288	26711	89	67	809	96	2012-08-02 05:35:24.010078
26713	4	1-6365	26711	67	93	491	6	2012-08-02 05:35:24.010078
26714	4	1-7447	26711	93	57	484	49	2012-08-02 05:35:24.010078
26711	4	1-7859	26711	57	89	1054	26	2012-08-02 05:35:24.010078
26716	5	1-5292	26715	15	78	1827	97	2012-08-02 05:35:24.010078
26717	5	1-6471	26715	78	49	1258	72	2012-08-02 05:35:24.010078
26718	5	1-7409	26715	49	93	1681	6	2012-08-02 05:35:24.010078
26719	5	1-7447	26715	93	57	1874	49	2012-08-02 05:35:24.010078
26715	5	1-7859	26715	57	15	4622	26	2012-08-02 05:35:24.010078
26720	1	1-1718	\N	43	43	651	6	2012-08-02 05:35:24.010078
26722	5	1-7833	26721	61	87	363	55	2012-08-02 05:35:24.452113
26723	5	1-7626	26721	87	92	264	74	2012-08-02 05:35:24.452113
26724	5	1-6875	26721	92	36	145	12	2012-08-02 05:35:24.452113
26725	5	1-7809	26721	36	32	527	90	2012-08-02 05:35:24.452113
26721	5	1-7860	26721	32	61	565	16	2012-08-02 05:35:24.452113
26727	2	1-6798	26726	11	47	2139	57	2012-08-02 05:35:25.280799
26726	2	1-7862	26726	47	11	4890	47	2012-08-02 05:35:25.280799
26729	4	1-7773	26728	65	7	3519	6	2012-08-02 05:35:26.551229
26730	4	1-7790	26728	7	31	2535	24	2012-08-02 05:35:26.551229
26731	4	1-4102	26728	31	13	1473	21	2012-08-02 05:35:26.551229
26728	4	1-7865	26728	13	65	1116	18	2012-08-02 05:35:26.551229
26733	4	1-7441	26732	36	76	2167	54	2012-08-02 05:35:26.551229
26734	4	1-7831	26732	76	14	2085	91	2012-08-02 05:35:26.551229
26735	4	1-6634	26732	14	13	2468	21	2012-08-02 05:35:26.551229
26732	4	1-7865	26732	13	36	1939	18	2012-08-02 05:35:26.551229
26737	5	1-4620	26736	9	80	79	20	2012-08-02 05:35:26.551229
26738	5	1-7856	26736	80	69	130	66	2012-08-02 05:35:26.551229
26739	5	1-7687	26736	69	14	108	91	2012-08-02 05:35:26.551229
26740	5	1-6634	26736	14	13	140	21	2012-08-02 05:35:26.551229
26736	5	1-7865	26736	13	9	121	18	2012-08-02 05:35:26.551229
26742	5	1-7778	26741	8	69	3945	92	2012-08-02 05:35:26.551229
26743	5	1-7741	26741	69	27	1763	70	2012-08-02 05:35:26.551229
26744	5	1-5615	26741	27	64	134	19	2012-08-02 05:35:26.551229
26745	5	1-6442	26741	64	13	192	21	2012-08-02 05:35:26.551229
26741	5	1-7865	26741	13	8	212	18	2012-08-02 05:35:26.551229
26747	4	1-7286	26746	67	71	336	68	2012-08-02 05:35:31.622336
26748	4	1-6002	26746	71	83	199	83	2012-08-02 05:35:31.622336
26749	4	1-7793	26746	83	15	607	88	2012-08-02 05:35:31.622336
26746	4	1-7877	26746	15	67	622	99	2012-08-02 05:35:31.622336
26751	3	1-5772	26750	23	42	3638	40	2012-08-02 05:35:31.622336
26752	3	1-7415	26750	42	15	2446	88	2012-08-02 05:35:31.622336
26750	3	1-7877	26750	15	23	3718	99	2012-08-02 05:35:31.622336
26754	4	1-5937	26753	60	8	2828	36	2012-08-02 05:35:31.622336
26755	4	1-7753	26753	8	52	1867	49	2012-08-02 05:35:31.622336
26756	4	1-7301	26753	52	15	3456	88	2012-08-02 05:35:31.622336
26753	4	1-7877	26753	15	60	5632	99	2012-08-02 05:35:31.622336
26758	4	1-3341	26757	15	99	1113	13	2012-08-02 05:35:32.041899
26759	4	1-7714	26757	99	40	1379	26	2012-08-02 05:35:32.041899
26760	4	1-7365	26757	40	48	1414	69	2012-08-02 05:35:32.041899
26757	4	1-7878	26757	48	15	1528	98	2012-08-02 05:35:32.041899
26762	4	1-7832	26761	41	17	1640	70	2012-08-02 05:35:32.881434
26763	4	1-5137	26761	17	17	467	90	2012-08-02 05:35:32.881434
26764	4	1-6824	26761	17	62	530	80	2012-08-02 05:35:32.881434
26761	4	1-7880	26761	62	41	511	97	2012-08-02 05:35:32.881434
26766	5	1-7774	26765	15	61	4279	17	2012-08-02 05:35:34.982685
26767	5	1-7498	26765	61	12	1483	76	2012-08-02 05:35:34.982685
26768	5	1-7642	26765	12	7	1736	46	2012-08-02 05:35:34.982685
26769	5	1-7501	26765	7	54	2849	13	2012-08-02 05:35:34.982685
26765	5	1-7885	26765	54	15	2659	80	2012-08-02 05:35:34.982685
26771	4	1-7156	26770	58	42	773	38	2012-08-02 05:35:35.402103
26772	4	1-2798	26770	42	65	664	15	2012-08-02 05:35:35.402103
26773	4	1-7837	26770	65	25	379	64	2012-08-02 05:35:35.402103
26770	4	1-7886	26770	25	58	719	1	2012-08-02 05:35:35.402103
26775	4	1-6389	26774	83	16	2641	94	2012-08-02 05:35:35.402103
26776	4	1-7867	26774	16	7	3282	96	2012-08-02 05:35:35.402103
26777	4	1-7514	26774	7	25	2468	64	2012-08-02 05:35:35.402103
26774	4	1-7886	26774	25	83	5114	1	2012-08-02 05:35:35.402103
26779	4	1-5631	26778	97	65	449	18	2012-08-02 05:35:36.650589
26780	4	1-7773	26778	65	68	2083	6	2012-08-02 05:35:36.650589
26781	4	1-7825	26778	68	1	1715	44	2012-08-02 05:35:36.650589
26778	4	1-7889	26778	1	97	1140	66	2012-08-02 05:35:36.650589
26783	4	1-7762	26782	84	45	6043	77	2012-08-02 05:35:37.501574
26784	4	1-6383	26782	45	27	6674	37	2012-08-02 05:35:37.501574
26785	4	1-7828	26782	27	86	6270	31	2012-08-02 05:35:37.501574
26782	4	1-7891	26782	86	84	7608	60	2012-08-02 05:35:37.501574
26787	3	1-6011	26786	4	27	281	37	2012-08-02 05:35:37.501574
26788	3	1-7828	26786	27	86	265	31	2012-08-02 05:35:37.501574
26786	3	1-7891	26786	86	4	323	60	2012-08-02 05:35:37.501574
26790	5	1-7790	26789	7	5	280	24	2012-08-02 05:35:39.169675
26791	5	1-7456	26789	5	34	248	26	2012-08-02 05:35:39.169675
26792	5	1-6790	26789	34	15	271	80	2012-08-02 05:35:39.169675
26793	5	1-7774	26789	15	82	439	17	2012-08-02 05:35:39.169675
26789	5	1-7895	26789	82	7	282	6	2012-08-02 05:35:39.169675
26795	3	1-7447	26794	93	60	5482	49	2012-08-02 05:35:39.169675
26796	3	1-6679	26794	60	82	7565	17	2012-08-02 05:35:39.169675
26794	3	1-7895	26794	82	93	5017	6	2012-08-02 05:35:39.169675
26798	4	1-7878	26797	48	14	2030	98	2012-08-02 05:35:39.987159
26799	4	1-7712	26797	14	41	1698	97	2012-08-02 05:35:39.987159
26800	4	1-7832	26797	41	45	3748	70	2012-08-02 05:35:39.987159
26797	4	1-7897	26797	45	48	2724	69	2012-08-02 05:35:39.987159
26802	4	1-7878	26801	48	46	2445	98	2012-08-02 05:35:39.987159
26803	4	1-7796	26801	46	81	2296	73	2012-08-02 05:35:39.987159
26804	4	1-7875	26801	81	45	2457	70	2012-08-02 05:35:39.987159
26801	4	1-7897	26801	45	48	2421	69	2012-08-02 05:35:39.987159
26806	4	1-6076	26805	30	59	2574	59	2012-08-02 05:35:39.987159
26807	4	1-7645	26805	59	20	2140	51	2012-08-02 05:35:39.987159
26808	4	1-7463	26805	20	45	4230	70	2012-08-02 05:35:39.987159
26805	4	1-7897	26805	45	30	4194	69	2012-08-02 05:35:39.987159
26810	4	1-7027	26809	46	42	92	34	2012-08-02 05:35:39.987159
26811	4	1-6730	26809	42	69	321	92	2012-08-02 05:35:39.987159
26812	4	1-7741	26809	69	45	138	70	2012-08-02 05:35:39.987159
26809	4	1-7897	26809	45	46	137	69	2012-08-02 05:35:39.987159
26814	4	1-7819	26813	40	73	4810	14	2012-08-02 05:35:40.849404
26815	4	1-7858	26813	73	5	3813	40	2012-08-02 05:35:40.849404
26816	4	1-7896	26813	5	73	2068	99	2012-08-02 05:35:40.849404
26813	4	1-7899	26813	73	40	2573	63	2012-08-02 05:35:40.849404
26818	4	1-7655	26817	83	4	1180	25	2012-08-02 05:35:41.280115
26819	4	1-7260	26817	4	100	469	47	2012-08-02 05:35:41.280115
26820	4	1-2216	26817	100	67	190	5	2012-08-02 05:35:41.280115
26817	4	1-7900	26817	67	83	2482	85	2012-08-02 05:35:41.280115
26822	4	1-7655	26821	83	84	285	25	2012-08-02 05:35:41.280115
26823	4	1-7760	26821	84	93	91	43	2012-08-02 05:35:41.280115
26824	4	1-6887	26821	93	67	37	5	2012-08-02 05:35:41.280115
26821	4	1-7900	26821	67	83	535	85	2012-08-02 05:35:41.280115
26826	4	1-7316	26825	87	86	4543	48	2012-08-02 05:35:42.959198
26827	4	1-1668	26825	86	80	398	20	2012-08-02 05:35:42.959198
26828	4	1-7856	26825	80	57	593	66	2012-08-02 05:35:42.959198
26825	4	1-7904	26825	57	87	2862	97	2012-08-02 05:35:42.959198
26830	4	1-7845	26829	57	73	1425	93	2012-08-02 05:35:42.959198
26831	4	1-7836	26829	73	80	285	20	2012-08-02 05:35:42.959198
26832	4	1-7856	26829	80	57	352	66	2012-08-02 05:35:42.959198
26829	4	1-7904	26829	57	57	1856	97	2012-08-02 05:35:42.959198
26834	5	1-4865	26833	49	65	2711	15	2012-08-02 05:35:42.959198
26835	5	1-7837	26833	65	24	1611	64	2012-08-02 05:35:42.959198
26836	5	1-3246	26833	24	94	813	11	2012-08-02 05:35:42.959198
26837	5	1-7728	26833	94	57	704	66	2012-08-02 05:35:42.959198
26833	5	1-7904	26833	57	49	3719	97	2012-08-02 05:35:42.959198
26839	5	1-7678	26838	66	61	3631	90	2012-08-02 05:35:43.39033
26840	5	1-6572	26838	61	27	2924	63	2012-08-02 05:35:43.39033
26841	5	1-6909	26838	27	52	2005	50	2012-08-02 05:35:43.39033
26842	5	1-7736	26838	52	31	1935	100	2012-08-02 05:35:43.39033
26838	5	1-7905	26838	31	66	2558	1	2012-08-02 05:35:43.39033
26844	5	1-6389	26843	83	16	519	94	2012-08-02 05:35:43.39033
26845	5	1-7867	26843	16	42	698	96	2012-08-02 05:35:43.39033
26846	5	1-5698	26843	42	52	648	50	2012-08-02 05:35:43.39033
26847	5	1-7736	26843	52	31	663	100	2012-08-02 05:35:43.39033
26843	5	1-7905	26843	31	83	930	1	2012-08-02 05:35:43.39033
26849	5	1-7451	26848	80	16	95	94	2012-08-02 05:35:43.39033
26850	5	1-7867	26848	16	42	136	96	2012-08-02 05:35:43.39033
26851	5	1-5698	26848	42	52	135	50	2012-08-02 05:35:43.39033
26852	5	1-7736	26848	52	31	148	100	2012-08-02 05:35:43.39033
26848	5	1-7905	26848	31	80	221	1	2012-08-02 05:35:43.39033
26854	3	1-6407	26853	27	43	2925	56	2012-08-02 05:35:43.39033
26855	3	1-4370	26853	43	31	2184	100	2012-08-02 05:35:43.39033
26853	3	1-7905	26853	31	27	3361	1	2012-08-02 05:35:43.39033
26857	4	1-7790	26856	7	16	1394	24	2012-08-02 05:35:43.843375
26858	4	1-7130	26856	16	52	1728	50	2012-08-02 05:35:43.843375
26859	4	1-7736	26856	52	14	1688	100	2012-08-02 05:35:43.843375
26856	4	1-7906	26856	14	7	1579	6	2012-08-02 05:35:43.843375
26861	4	1-7790	26860	7	18	2077	24	2012-08-02 05:35:43.843375
26862	4	1-5591	26860	18	52	2448	50	2012-08-02 05:35:43.843375
26863	4	1-7736	26860	52	14	2431	100	2012-08-02 05:35:43.843375
26860	4	1-7906	26860	14	7	2313	6	2012-08-02 05:35:43.843375
26865	5	1-6730	26864	42	78	5	92	2012-08-02 05:35:46.837333
26866	5	1-7711	26864	78	29	2	1	2012-08-02 05:35:46.837333
26867	5	1-7640	26864	29	50	1	5	2012-08-02 05:35:46.837333
26868	5	1-7910	26864	50	99	1	55	2012-08-02 05:35:46.837333
26864	5	1-7913	26864	99	42	2	34	2012-08-02 05:35:46.837333
26870	5	1-6763	26869	54	67	496	79	2012-08-02 05:35:48.096511
26871	5	1-7873	26869	67	3	1291	10	2012-08-02 05:35:48.096511
26872	5	1-7407	26869	3	52	495	26	2012-08-02 05:35:48.096511
26873	5	1-1931	26869	52	39	413	93	2012-08-02 05:35:48.096511
26869	5	1-7916	26869	39	54	600	47	2012-08-02 05:35:48.096511
26875	3	1-6763	26874	54	28	1386	79	2012-08-02 05:35:48.096511
26876	3	1-7652	26874	28	39	1025	93	2012-08-02 05:35:48.096511
26874	3	1-7916	26874	39	54	1581	47	2012-08-02 05:35:48.096511
26878	5	1-5062	26877	49	17	1090	23	2012-08-02 05:35:48.096511
26879	5	1-6781	26877	17	92	1109	74	2012-08-02 05:35:48.096511
26880	5	1-6875	26877	92	78	725	12	2012-08-02 05:35:48.096511
26881	5	1-7855	26877	78	39	1545	93	2012-08-02 05:35:48.096511
26877	5	1-7916	26877	39	49	2416	47	2012-08-02 05:35:48.096511
26883	4	1-5119	26882	96	63	852	56	2012-08-02 05:35:48.516574
26884	4	1-7807	26882	63	67	1553	79	2012-08-02 05:35:48.516574
26885	4	1-7873	26882	67	71	3901	10	2012-08-02 05:35:48.516574
26882	4	1-7917	26882	71	96	3433	86	2012-08-02 05:35:48.516574
26887	4	1-7287	26886	78	1	2411	34	2012-08-02 05:35:48.516574
26888	4	1-6869	26886	1	54	1697	40	2012-08-02 05:35:48.516574
26889	4	1-7536	26886	54	71	1540	10	2012-08-02 05:35:48.516574
26886	4	1-7917	26886	71	78	1537	86	2012-08-02 05:35:48.516574
26891	4	1-7422	26890	48	9	875	87	2012-08-02 05:35:50.664973
26892	4	1-7764	26890	9	94	279	47	2012-08-02 05:35:50.664973
26893	4	1-4524	26890	94	88	200	73	2012-08-02 05:35:50.664973
26890	4	1-7922	26890	88	48	1384	14	2012-08-02 05:35:50.664973
26895	3	1-2184	26894	98	31	223	100	2012-08-02 05:35:51.499589
26896	3	1-7905	26894	31	63	275	1	2012-08-02 05:35:51.499589
26894	3	1-7924	26894	63	98	1402	70	2012-08-02 05:35:51.499589
26898	4	1-5137	26897	17	73	346	90	2012-08-02 05:35:51.499589
26899	4	1-7914	26897	73	78	334	92	2012-08-02 05:35:51.499589
26900	4	1-7711	26897	78	63	200	1	2012-08-02 05:35:51.499589
26897	4	1-7924	26897	63	17	1280	70	2012-08-02 05:35:51.499589
26902	5	1-5302	26901	51	40	442	58	2012-08-02 05:35:52.350027
26903	5	1-7212	26901	40	39	641	77	2012-08-02 05:35:52.350027
26904	5	1-5471	26901	39	83	732	85	2012-08-02 05:35:52.350027
26905	5	1-7655	26901	83	42	659	25	2012-08-02 05:35:52.350027
26901	5	1-7926	26901	42	51	484	41	2012-08-02 05:35:52.350027
26907	4	1-4685	26906	89	86	672	29	2012-08-02 05:35:54.515644
26908	4	1-7771	26906	86	27	2055	37	2012-08-02 05:35:54.515644
26909	4	1-7828	26906	27	39	1673	31	2012-08-02 05:35:54.515644
26906	4	1-7931	26906	39	89	2186	45	2012-08-02 05:35:54.515644
26911	3	1-6013	26910	51	27	211	37	2012-08-02 05:35:54.515644
26912	3	1-7828	26910	27	39	211	31	2012-08-02 05:35:54.515644
26910	3	1-7931	26910	39	51	340	45	2012-08-02 05:35:54.515644
26914	4	1-6412	26913	100	53	228	79	2012-08-02 05:35:54.515644
26915	4	1-2946	26913	53	48	84	7	2012-08-02 05:35:54.515644
26916	4	1-7239	26913	48	39	102	31	2012-08-02 05:35:54.515644
26913	4	1-7931	26913	39	100	172	45	2012-08-02 05:35:54.515644
26918	4	1-7025	26917	89	83	78	85	2012-08-02 05:35:54.515644
26919	4	1-7655	26917	83	80	65	25	2012-08-02 05:35:54.515644
26920	4	1-5268	26917	80	39	45	31	2012-08-02 05:35:54.515644
26917	4	1-7931	26917	39	89	77	45	2012-08-02 05:35:54.515644
26922	4	1-6412	26921	100	28	329	79	2012-08-02 05:35:54.515644
26923	4	1-7652	26921	28	95	259	93	2012-08-02 05:35:54.515644
26924	4	1-3725	26921	95	39	124	31	2012-08-02 05:35:54.515644
26921	4	1-7931	26921	39	100	228	45	2012-08-02 05:35:54.515644
26926	3	1-7395	26925	67	11	3527	25	2012-08-02 05:35:56.680726
26927	3	1-2380	26925	11	83	598	54	2012-08-02 05:35:56.680726
26925	3	1-7936	26925	83	67	5229	3	2012-08-02 05:35:56.680726
26929	4	1-7141	26928	100	32	615	32	2012-08-02 05:35:56.680726
26930	4	1-4604	26928	32	3	264	5	2012-08-02 05:35:56.680726
26931	4	1-5738	26928	3	83	140	54	2012-08-02 05:35:56.680726
26928	4	1-7936	26928	83	100	1414	3	2012-08-02 05:35:56.680726
26933	5	1-5962	26932	2	71	39	10	2012-08-02 05:35:57.498273
26934	5	1-7917	26932	71	63	34	86	2012-08-02 05:35:57.498273
26935	5	1-7765	26932	63	77	14	3	2012-08-02 05:35:57.498273
26936	5	1-7568	26932	77	29	27	31	2012-08-02 05:35:57.498273
26932	5	1-7938	26932	29	2	42	96	2012-08-02 05:35:57.498273
26938	3	1-7932	26937	24	2	3911	38	2012-08-02 05:35:57.498273
26939	3	1-7817	26937	2	29	2736	31	2012-08-02 05:35:57.498273
26937	3	1-7938	26937	29	24	4841	96	2012-08-02 05:35:57.498273
26941	5	1-7917	26940	71	14	3216	86	2012-08-02 05:35:57.918319
26942	5	1-7869	26940	14	10	1113	47	2012-08-02 05:35:57.918319
26943	5	1-7430	26940	10	80	733	20	2012-08-02 05:35:57.918319
26944	5	1-7856	26940	80	70	1411	66	2012-08-02 05:35:57.918319
26940	5	1-7939	26940	70	71	3482	10	2012-08-02 05:35:57.918319
26946	2	1-809	26945	73	22	1348	7	2012-08-02 05:35:59.177638
26945	2	1-7942	26945	22	73	7847	35	2012-08-02 05:35:59.177638
26948	5	1-7533	26947	14	19	1745	98	2012-08-02 05:35:59.597693
26949	5	1-5231	26947	19	49	1079	72	2012-08-02 05:35:59.597693
26950	5	1-7409	26947	49	7	1073	6	2012-08-02 05:35:59.597693
26951	5	1-7790	26947	7	12	838	24	2012-08-02 05:35:59.597693
26947	5	1-7943	26947	12	14	1344	30	2012-08-02 05:35:59.597693
26953	4	1-7426	26952	97	14	997	100	2012-08-02 05:35:59.597693
26954	4	1-7906	26952	14	7	890	6	2012-08-02 05:35:59.597693
26955	4	1-7790	26952	7	12	750	24	2012-08-02 05:35:59.597693
26952	4	1-7943	26952	12	97	1298	30	2012-08-02 05:35:59.597693
26957	5	1-7251	26956	59	60	17	49	2012-08-02 05:35:59.597693
26958	5	1-6679	26956	60	82	22	17	2012-08-02 05:35:59.597693
26959	5	1-7895	26956	82	7	14	6	2012-08-02 05:35:59.597693
26960	5	1-7790	26956	7	12	14	24	2012-08-02 05:35:59.597693
26956	5	1-7943	26956	12	59	27	30	2012-08-02 05:35:59.597693
26962	4	1-7533	26961	14	41	128	98	2012-08-02 05:35:59.597693
26963	4	1-4114	26961	41	7	42	6	2012-08-02 05:35:59.597693
26964	4	1-7790	26961	7	12	40	24	2012-08-02 05:35:59.597693
26961	4	1-7943	26961	12	14	80	30	2012-08-02 05:35:59.597693
26966	5	1-5828	26965	12	16	198	94	2012-08-02 05:35:59.597693
26967	5	1-7867	26965	16	67	269	96	2012-08-02 05:35:59.597693
26968	5	1-6365	26965	67	7	165	6	2012-08-02 05:35:59.597693
26969	5	1-7790	26965	7	12	154	24	2012-08-02 05:35:59.597693
26965	5	1-7943	26965	12	12	318	30	2012-08-02 05:35:59.597693
26971	2	1-6016	26970	93	12	882	24	2012-08-02 05:35:59.597693
26970	2	1-7943	26970	12	93	1826	30	2012-08-02 05:35:59.597693
26973	5	1-7543	26972	66	91	609	35	2012-08-02 05:36:02.624683
26974	5	1-4260	26972	91	82	547	46	2012-08-02 05:36:02.624683
26975	5	1-7606	26972	82	7	344	6	2012-08-02 05:36:02.624683
26976	5	1-7790	26972	7	50	345	24	2012-08-02 05:36:02.624683
26972	5	1-7950	26972	50	66	966	33	2012-08-02 05:36:02.624683
26978	4	1-7933	26977	73	42	2615	63	2012-08-02 05:36:02.624683
26979	4	1-7636	26977	42	7	815	6	2012-08-02 05:36:02.624683
26980	4	1-7790	26977	7	50	836	24	2012-08-02 05:36:02.624683
26977	4	1-7950	26977	50	73	2391	33	2012-08-02 05:36:02.624683
26982	3	1-7807	26981	63	75	1329	79	2012-08-02 05:36:03.475958
26983	3	1-7827	26981	75	9	2428	60	2012-08-02 05:36:03.475958
26981	3	1-7952	26981	9	63	1077	56	2012-08-02 05:36:03.475958
26985	3	1-7637	26984	20	30	1047	69	2012-08-02 05:36:06.417947
26986	3	1-6076	26984	30	51	565	59	2012-08-02 05:36:06.417947
26984	3	1-7959	26984	51	20	1205	32	2012-08-02 05:36:06.417947
26988	3	1-5078	26987	68	30	3616	69	2012-08-02 05:36:06.417947
26989	3	1-6076	26987	30	51	1963	59	2012-08-02 05:36:06.417947
26987	3	1-7959	26987	51	68	4208	32	2012-08-02 05:36:06.417947
26990	1	1-5078	\N	68	68	65	69	2012-08-02 05:36:06.417947
26991	1	1-6086	\N	37	37	1228	69	2012-08-02 05:36:06.417947
26993	3	1-7342	26992	38	48	2269	28	2012-08-02 05:36:06.815354
26994	3	1-6942	26992	48	57	789	31	2012-08-02 05:36:06.815354
26992	3	1-7960	26992	57	38	5141	23	2012-08-02 05:36:06.815354
26996	4	1-7935	26995	96	21	663	77	2012-08-02 05:36:06.815354
26997	4	1-7577	26995	21	87	533	87	2012-08-02 05:36:06.815354
26998	4	1-2490	26995	87	57	121	31	2012-08-02 05:36:06.815354
26995	4	1-7960	26995	57	96	917	23	2012-08-02 05:36:06.815354
26999	1	1-2490	\N	87	87	2838	31	2012-08-02 05:36:06.815354
27001	2	1-7201	27000	33	14	3210	57	2012-08-02 05:36:07.213461
27000	2	1-7961	27000	14	33	5005	42	2012-08-02 05:36:07.213461
27003	5	1-6348	27002	56	82	1243	95	2012-08-02 05:36:07.622032
27004	5	1-6105	27002	82	78	936	92	2012-08-02 05:36:07.622032
27005	5	1-7711	27002	78	82	377	1	2012-08-02 05:36:07.622032
27006	5	1-7595	27002	82	44	208	27	2012-08-02 05:36:07.622032
27002	5	1-7962	27002	44	56	1199	71	2012-08-02 05:36:07.622032
27008	5	1-7440	27007	53	74	3595	65	2012-08-02 05:36:07.622032
27009	5	1-6525	27007	74	5	1678	40	2012-08-02 05:36:07.622032
27010	5	1-7896	27007	5	33	754	99	2012-08-02 05:36:07.622032
27011	5	1-7737	27007	33	44	605	27	2012-08-02 05:36:07.622032
27007	5	1-7962	27007	44	53	3757	71	2012-08-02 05:36:07.622032
27013	5	1-7556	27012	28	82	1091	95	2012-08-02 05:36:07.622032
27014	5	1-6105	27012	82	78	861	92	2012-08-02 05:36:07.622032
27015	5	1-7711	27012	78	82	382	1	2012-08-02 05:36:07.622032
27016	5	1-7595	27012	82	44	221	27	2012-08-02 05:36:07.622032
27012	5	1-7962	27012	44	28	1397	71	2012-08-02 05:36:07.622032
27018	4	1-2542	27017	67	5	1212	40	2012-08-02 05:36:07.622032
27019	4	1-7896	27017	5	33	568	99	2012-08-02 05:36:07.622032
27020	4	1-7737	27017	33	44	476	27	2012-08-02 05:36:07.622032
27017	4	1-7962	27017	44	67	3083	71	2012-08-02 05:36:07.622032
27022	4	1-5019	27021	76	24	1611	51	2012-08-02 05:36:10.030419
27023	4	1-7625	27021	24	91	1092	52	2012-08-02 05:36:10.030419
27024	4	1-6547	27021	91	34	1074	100	2012-08-02 05:36:10.030419
27021	4	1-7968	27021	34	76	3184	44	2012-08-02 05:36:10.030419
27026	4	1-7795	27025	52	78	3148	19	2012-08-02 05:36:10.439492
27027	4	1-7733	27025	78	81	3923	39	2012-08-02 05:36:10.439492
27028	4	1-6927	27025	81	98	3831	61	2012-08-02 05:36:10.439492
27025	4	1-7969	27025	98	52	6110	100	2012-08-02 05:36:10.439492
27030	3	1-7557	27029	97	64	1400	19	2012-08-02 05:36:10.85909
27031	3	1-6442	27029	64	22	1505	21	2012-08-02 05:36:10.85909
27029	3	1-7970	27029	22	97	2790	23	2012-08-02 05:36:10.85909
27033	4	1-7606	27032	82	42	505	6	2012-08-02 05:36:11.267954
27034	4	1-7946	27032	42	25	430	57	2012-08-02 05:36:11.267954
27035	4	1-7228	27032	25	21	209	96	2012-08-02 05:36:11.267954
27032	4	1-7971	27032	21	82	904	46	2012-08-02 05:36:11.267954
27037	3	1-7919	27036	74	36	5701	18	2012-08-02 05:36:12.483119
27038	3	1-7441	27036	36	23	7639	54	2012-08-02 05:36:12.483119
27036	3	1-7974	27036	23	74	5292	6	2012-08-02 05:36:12.483119
27040	5	1-7658	27039	45	81	2155	4	2012-08-02 05:36:12.880912
27041	5	1-6814	27039	81	78	993	92	2012-08-02 05:36:12.880912
27042	5	1-7711	27039	78	63	386	1	2012-08-02 05:36:12.880912
27043	5	1-7924	27039	63	61	1598	70	2012-08-02 05:36:12.880912
27039	5	1-7975	27039	61	45	1250	98	2012-08-02 05:36:12.880912
27045	5	1-6970	27044	46	78	394	33	2012-08-02 05:36:12.880912
27046	5	1-6032	27044	78	78	347	92	2012-08-02 05:36:12.880912
27047	5	1-7711	27044	78	63	194	1	2012-08-02 05:36:12.880912
27048	5	1-7924	27044	63	61	1026	70	2012-08-02 05:36:12.880912
27044	5	1-7975	27044	61	46	1024	98	2012-08-02 05:36:12.880912
27050	4	1-7921	27049	32	57	935	76	2012-08-02 05:36:13.278647
27051	4	1-6228	27049	57	98	843	58	2012-08-02 05:36:13.278647
27052	4	1-7694	27049	98	18	802	81	2012-08-02 05:36:13.278647
27049	4	1-7976	27049	18	32	954	54	2012-08-02 05:36:13.278647
27054	3	1-7921	27053	32	45	5893	76	2012-08-02 05:36:13.278647
27055	3	1-6871	27053	45	18	5015	81	2012-08-02 05:36:13.278647
27053	3	1-7976	27053	18	32	5991	54	2012-08-02 05:36:13.278647
27057	4	1-7896	27056	5	33	615	99	2012-08-02 05:36:13.687587
27058	4	1-7737	27056	33	62	594	27	2012-08-02 05:36:13.687587
27059	4	1-7523	27056	62	92	789	85	2012-08-02 05:36:13.687587
27056	4	1-7977	27056	92	5	1139	40	2012-08-02 05:36:13.687587
27061	4	1-7896	27060	5	88	1591	99	2012-08-02 05:36:13.687587
27062	4	1-6979	27060	88	44	779	98	2012-08-02 05:36:13.687587
27063	4	1-5848	27060	44	92	1538	85	2012-08-02 05:36:13.687587
27060	4	1-7977	27060	92	5	2558	40	2012-08-02 05:36:13.687587
27065	4	1-7457	27064	97	12	5137	76	2012-08-02 05:36:14.085304
27066	4	1-7642	27064	12	92	4156	46	2012-08-02 05:36:14.085304
27067	4	1-7789	27064	92	35	1530	83	2012-08-02 05:36:14.085304
27064	4	1-7978	27064	35	97	3847	21	2012-08-02 05:36:14.085304
27069	4	1-7865	27068	13	34	4029	18	2012-08-02 05:36:14.085304
27070	4	1-6378	27068	34	43	2786	17	2012-08-02 05:36:14.085304
27071	4	1-7424	27068	43	35	1840	83	2012-08-02 05:36:14.085304
27068	4	1-7978	27068	35	13	5097	21	2012-08-02 05:36:14.085304
27073	4	1-7413	27072	64	65	1537	61	2012-08-02 05:36:14.881021
27074	4	1-5440	27072	65	21	590	96	2012-08-02 05:36:14.881021
27075	4	1-7971	27072	21	18	2915	46	2012-08-02 05:36:14.881021
27072	4	1-7980	27072	18	64	2365	75	2012-08-02 05:36:14.881021
27077	5	1-5722	27076	72	12	1910	76	2012-08-02 05:36:15.843101
27078	5	1-7642	27076	12	82	2247	46	2012-08-02 05:36:15.843101
27079	5	1-7606	27076	82	74	1396	6	2012-08-02 05:36:15.843101
27080	5	1-7919	27076	74	61	1666	18	2012-08-02 05:36:15.843101
27076	5	1-7982	27076	61	72	2573	78	2012-08-02 05:36:15.843101
27082	4	1-7680	27081	66	79	2262	71	2012-08-02 05:36:16.239726
27083	4	1-7064	27081	79	58	3280	4	2012-08-02 05:36:16.239726
27084	4	1-7234	27081	58	85	1215	25	2012-08-02 05:36:16.239726
27081	4	1-7983	27081	85	66	2606	84	2012-08-02 05:36:16.239726
27086	4	1-7680	27085	66	84	23	71	2012-08-02 05:36:16.239726
27087	4	1-7585	27085	84	58	27	4	2012-08-02 05:36:16.239726
27088	4	1-7234	27085	58	85	11	25	2012-08-02 05:36:16.239726
27085	4	1-7983	27085	85	66	25	84	2012-08-02 05:36:16.239726
27090	3	1-4620	27089	9	52	476	20	2012-08-02 05:36:16.648615
27091	3	1-7818	27089	52	94	561	4	2012-08-02 05:36:16.648615
27089	3	1-7984	27089	94	9	701	18	2012-08-02 05:36:16.648615
27093	3	1-7974	27092	23	68	318	6	2012-08-02 05:36:17.443953
27094	3	1-7825	27092	68	94	130	44	2012-08-02 05:36:17.443953
27092	3	1-7986	27092	94	23	783	54	2012-08-02 05:36:17.443953
27096	3	1-7690	27095	84	94	72	79	2012-08-02 05:36:17.443953
27097	3	1-7653	27095	94	94	57	44	2012-08-02 05:36:17.443953
27095	3	1-7986	27095	94	84	428	54	2012-08-02 05:36:17.443953
27099	2	1-5268	27098	80	34	717	31	2012-08-02 05:36:19.874672
27098	2	1-7992	27098	34	80	1546	25	2012-08-02 05:36:19.874672
27101	5	1-5679	27100	80	93	554	76	2012-08-02 05:36:19.874672
27102	5	1-5905	27100	93	88	374	97	2012-08-02 05:36:19.874672
27103	5	1-5548	27100	88	77	155	3	2012-08-02 05:36:19.874672
27104	5	1-7568	27100	77	34	300	31	2012-08-02 05:36:19.874672
27100	5	1-7992	27100	34	80	899	25	2012-08-02 05:36:19.874672
27106	4	1-7015	27105	87	64	3640	75	2012-08-02 05:36:19.874672
27107	4	1-7413	27105	64	79	2214	61	2012-08-02 05:36:19.874672
27108	4	1-6424	27105	79	34	1628	31	2012-08-02 05:36:19.874672
27105	4	1-7992	27105	34	87	5004	25	2012-08-02 05:36:19.874672
27110	4	1-3386	27109	53	88	360	97	2012-08-02 05:36:19.874672
27111	4	1-5548	27109	88	77	157	3	2012-08-02 05:36:19.874672
27112	4	1-7568	27109	77	34	322	31	2012-08-02 05:36:19.874672
27109	4	1-7992	27109	34	53	1020	25	2012-08-02 05:36:19.874672
27114	4	1-4609	27113	5	20	592	41	2012-08-02 05:36:20.703377
27115	4	1-7414	27113	20	72	553	56	2012-08-02 05:36:20.703377
27113	4	1-7994	27113	57	5	2906	48	2012-08-02 05:36:20.703377
27116	4	1-7909	27113	72	57	1487	36	2012-08-02 05:36:20.703377
27118	4	1-5770	27117	47	65	2322	15	2012-08-02 05:36:21.111992
27119	4	1-7837	27117	65	33	1270	64	2012-08-02 05:36:21.111992
27120	4	1-6196	27117	33	38	547	81	2012-08-02 05:36:21.111992
27117	4	1-7995	27117	38	47	4822	28	2012-08-02 05:36:21.111992
27122	4	1-5160	27121	50	82	742	54	2012-08-02 05:36:21.111992
27123	4	1-7920	27121	82	98	209	58	2012-08-02 05:36:21.111992
27124	4	1-7694	27121	98	38	151	81	2012-08-02 05:36:21.111992
27121	4	1-7995	27121	38	50	1430	28	2012-08-02 05:36:21.111992
27126	5	1-7206	27125	51	79	726	42	2012-08-02 05:36:21.907731
27127	5	1-5974	27125	79	7	243	17	2012-08-02 05:36:21.907731
27128	5	1-7985	27125	7	19	260	11	2012-08-02 05:36:21.907731
27129	5	1-7881	27125	19	83	307	82	2012-08-02 05:36:21.907731
27125	5	1-7997	27125	83	51	733	9	2012-08-02 05:36:21.907731
27131	2	1-6011	27130	4	37	5217	37	2012-08-02 05:36:24.315686
27130	2	1-8003	27130	37	4	8843	60	2012-08-02 05:36:24.315686
27133	3	1-7967	27132	92	5	497	40	2012-08-02 05:36:24.713472
27134	3	1-7896	27132	5	8	270	99	2012-08-02 05:36:24.713472
27132	3	1-8004	27132	8	92	381	18	2012-08-02 05:36:24.713472
27136	4	1-4620	27135	9	52	855	20	2012-08-02 05:36:24.713472
27137	4	1-7818	27135	52	97	1087	4	2012-08-02 05:36:24.713472
27138	4	1-7996	27135	97	8	804	99	2012-08-02 05:36:24.713472
27135	4	1-8004	27135	8	9	1170	18	2012-08-02 05:36:24.713472
27140	5	1-6525	27139	74	61	1014	40	2012-08-02 05:36:25.133379
27141	5	1-7649	27139	61	83	157	82	2012-08-02 05:36:25.133379
27142	5	1-7997	27139	83	6	229	9	2012-08-02 05:36:25.133379
27143	5	1-7570	27139	6	10	399	58	2012-08-02 05:36:25.133379
27139	5	1-8005	27139	10	74	2502	65	2012-08-02 05:36:25.133379
27145	3	1-7442	27144	78	7	740	4	2012-08-02 05:36:25.133379
27146	3	1-6848	27144	7	10	304	58	2012-08-02 05:36:25.133379
27144	3	1-8005	27144	10	78	2032	65	2012-08-02 05:36:25.133379
27148	3	1-7751	27147	14	22	1660	6	2012-08-02 05:36:25.133379
27149	3	1-7364	27147	22	10	660	58	2012-08-02 05:36:25.133379
27147	3	1-8005	27147	10	14	4460	65	2012-08-02 05:36:25.133379
27151	4	1-7753	27150	8	48	1111	49	2012-08-02 05:36:25.542369
27152	4	1-5239	27150	48	24	493	39	2012-08-02 05:36:25.542369
27153	4	1-7085	27150	24	100	462	96	2012-08-02 05:36:25.542369
27150	4	1-8006	27150	100	8	1391	36	2012-08-02 05:36:25.542369
27155	5	1-6982	27154	22	9	847	18	2012-08-02 05:36:28.348211
27156	5	1-4620	27154	9	80	566	20	2012-08-02 05:36:28.348211
27157	5	1-7856	27154	80	70	955	66	2012-08-02 05:36:28.348211
27158	5	1-7939	27154	70	92	2066	10	2012-08-02 05:36:28.348211
27154	5	1-8013	27154	92	22	3715	36	2012-08-02 05:36:28.348211
27160	4	1-1879	27159	59	1	1044	44	2012-08-02 05:36:28.348211
27161	4	1-7889	27159	1	70	602	66	2012-08-02 05:36:28.348211
27162	4	1-7939	27159	70	92	1420	10	2012-08-02 05:36:28.348211
27159	4	1-8013	27159	92	59	2785	36	2012-08-02 05:36:28.348211
27164	4	1-6196	27163	33	19	395	81	2012-08-02 05:36:30.325945
27165	4	1-7956	27163	19	10	431	13	2012-08-02 05:36:30.325945
27166	4	1-6464	27163	10	73	190	32	2012-08-02 05:36:30.325945
27163	4	1-8018	27163	73	33	645	64	2012-08-02 05:36:30.325945
27168	5	1-1511	27167	74	24	1930	51	2012-08-02 05:36:30.325945
27169	5	1-7625	27167	24	2	1336	52	2012-08-02 05:36:30.325945
27170	5	1-5423	27167	2	55	1072	7	2012-08-02 05:36:30.325945
27171	5	1-7513	27167	55	73	1450	32	2012-08-02 05:36:30.325945
27167	5	1-8018	27167	73	74	4950	64	2012-08-02 05:36:30.325945
27173	5	1-6679	27172	60	98	372	17	2012-08-02 05:36:31.530066
27174	5	1-7930	27172	98	7	272	26	2012-08-02 05:36:31.530066
27175	5	1-3974	27172	7	99	93	65	2012-08-02 05:36:31.530066
27176	5	1-7841	27172	99	61	117	78	2012-08-02 05:36:31.530066
27172	5	1-8021	27172	61	60	291	49	2012-08-02 05:36:31.530066
27178	4	1-6211	27177	75	93	1714	73	2012-08-02 05:36:31.530066
27179	4	1-6193	27177	93	99	878	65	2012-08-02 05:36:31.530066
27180	4	1-7841	27177	99	61	1199	78	2012-08-02 05:36:31.530066
27177	4	1-8021	27177	61	75	3236	49	2012-08-02 05:36:31.530066
27182	2	1-6011	27181	4	80	423	37	2012-08-02 05:36:31.949891
27181	2	1-8022	27181	80	4	784	60	2012-08-02 05:36:31.949891
27184	3	1-7564	27183	77	86	81	29	2012-08-02 05:36:31.949891
27185	3	1-7771	27183	86	80	196	37	2012-08-02 05:36:31.949891
27183	3	1-8022	27183	80	77	400	60	2012-08-02 05:36:31.949891
27187	4	1-7762	27186	84	37	2161	77	2012-08-02 05:36:31.949891
27188	4	1-6516	27186	37	62	1218	28	2012-08-02 05:36:31.949891
27189	4	1-5652	27186	62	80	1293	37	2012-08-02 05:36:31.949891
27186	4	1-8022	27186	80	84	3233	60	2012-08-02 05:36:31.949891
27191	4	1-7331	27190	40	65	298	15	2012-08-02 05:36:31.949891
27192	4	1-7837	27190	65	53	210	64	2012-08-02 05:36:31.949891
27193	4	1-1658	27190	53	80	175	37	2012-08-02 05:36:31.949891
27190	4	1-8022	27190	80	40	574	60	2012-08-02 05:36:31.949891
27195	4	1-7564	27194	77	30	226	29	2012-08-02 05:36:31.949891
27196	4	1-7665	27194	30	13	274	65	2012-08-02 05:36:31.949891
27197	4	1-7127	27194	13	80	181	37	2012-08-02 05:36:31.949891
27194	4	1-8022	27194	80	77	642	60	2012-08-02 05:36:31.949891
27199	4	1-7574	27198	29	80	35	20	2012-08-02 05:36:33.995764
27200	4	1-7856	27198	80	69	51	66	2012-08-02 05:36:33.995764
27201	4	1-7687	27198	69	61	37	91	2012-08-02 05:36:33.995764
27198	4	1-8027	27198	61	29	82	96	2012-08-02 05:36:33.995764
27203	5	1-8006	27202	100	45	5340	36	2012-08-02 05:36:33.995764
27204	5	1-4057	27202	45	24	3571	51	2012-08-02 05:36:33.995764
27205	5	1-7625	27202	24	82	1690	52	2012-08-02 05:36:33.995764
27206	5	1-7987	27202	82	61	1132	91	2012-08-02 05:36:33.995764
27202	5	1-8027	27202	61	100	2522	96	2012-08-02 05:36:33.995764
27208	5	1-7971	27207	21	92	1983	46	2012-08-02 05:36:33.995764
27209	5	1-7789	27207	92	34	835	83	2012-08-02 05:36:33.995764
27210	5	1-7628	27207	34	69	281	66	2012-08-02 05:36:33.995764
27211	5	1-7687	27207	69	61	221	91	2012-08-02 05:36:33.995764
27207	5	1-8027	27207	61	21	524	96	2012-08-02 05:36:33.995764
27213	4	1-7971	27212	21	40	3819	46	2012-08-02 05:36:33.995764
27214	4	1-1589	27212	40	82	560	52	2012-08-02 05:36:33.995764
27215	4	1-7987	27212	82	61	408	91	2012-08-02 05:36:33.995764
27212	4	1-8027	27212	61	21	988	96	2012-08-02 05:36:33.995764
27217	5	1-8006	27216	100	8	704	36	2012-08-02 05:36:33.995764
27218	5	1-7753	27216	8	21	450	49	2012-08-02 05:36:33.995764
27219	5	1-39	27216	21	82	151	52	2012-08-02 05:36:33.995764
27220	5	1-7987	27216	82	61	115	91	2012-08-02 05:36:33.995764
27216	5	1-8027	27216	61	100	292	96	2012-08-02 05:36:33.995764
27222	5	1-8006	27221	100	59	1410	36	2012-08-02 05:36:33.995764
27223	5	1-1879	27221	59	1	488	44	2012-08-02 05:36:33.995764
27224	5	1-7889	27221	1	69	260	66	2012-08-02 05:36:33.995764
27225	5	1-7687	27221	69	61	222	91	2012-08-02 05:36:33.995764
27221	5	1-8027	27221	61	100	574	96	2012-08-02 05:36:33.995764
27227	5	1-5264	27226	66	90	243	58	2012-08-02 05:36:33.995764
27228	5	1-7383	27226	90	80	56	20	2012-08-02 05:36:33.995764
27229	5	1-7856	27226	80	69	98	66	2012-08-02 05:36:33.995764
27230	5	1-7687	27226	69	61	86	91	2012-08-02 05:36:33.995764
27226	5	1-8027	27226	61	66	230	96	2012-08-02 05:36:33.995764
27232	5	1-7514	27231	7	24	1457	64	2012-08-02 05:36:33.995764
27233	5	1-3246	27231	24	94	804	11	2012-08-02 05:36:33.995764
27234	5	1-7728	27231	94	69	762	66	2012-08-02 05:36:33.995764
27235	5	1-7687	27231	69	61	685	91	2012-08-02 05:36:33.995764
27231	5	1-8027	27231	61	7	1858	96	2012-08-02 05:36:33.995764
27237	4	1-6829	27236	16	39	2384	31	2012-08-02 05:36:34.813814
27238	4	1-7931	27236	39	61	4144	45	2012-08-02 05:36:34.813814
27239	4	1-7779	27236	61	91	5920	2	2012-08-02 05:36:34.813814
27236	4	1-8029	27236	91	16	3012	36	2012-08-02 05:36:34.813814
27241	4	1-7753	27240	8	75	750	49	2012-08-02 05:36:34.813814
27242	4	1-6211	27240	75	62	402	73	2012-08-02 05:36:34.813814
27243	4	1-7419	27240	62	91	1765	2	2012-08-02 05:36:34.813814
27240	4	1-8029	27240	91	8	932	36	2012-08-02 05:36:34.813814
27245	2	1-7950	27244	50	60	2676	33	2012-08-02 05:36:36.029099
27244	2	1-8032	27244	60	50	924	24	2012-08-02 05:36:36.029099
27247	5	1-7239	27246	48	39	909	31	2012-08-02 05:36:36.824441
27248	5	1-7931	27246	39	53	1402	45	2012-08-02 05:36:36.824441
27249	5	1-7839	27246	53	74	707	6	2012-08-02 05:36:36.824441
27250	5	1-7919	27246	74	99	762	18	2012-08-02 05:36:36.824441
27246	5	1-8034	27246	99	48	822	7	2012-08-02 05:36:36.824441
27252	3	1-7683	27251	60	13	1388	21	2012-08-02 05:36:36.824441
27253	3	1-7865	27251	13	99	1569	18	2012-08-02 05:36:36.824441
27251	3	1-8034	27251	99	60	2029	7	2012-08-02 05:36:36.824441
27255	5	1-7813	27254	68	14	2871	57	2012-08-02 05:36:38.835253
27256	5	1-7961	27254	14	32	4386	42	2012-08-02 05:36:38.835253
27257	5	1-7575	27254	32	18	1379	82	2012-08-02 05:36:38.835253
27258	5	1-7702	27254	18	73	1152	23	2012-08-02 05:36:38.835253
27254	5	1-8039	27254	73	68	1225	56	2012-08-02 05:36:38.835253
27260	3	1-7424	27259	43	27	3433	83	2012-08-02 05:36:39.232867
27261	3	1-7013	27259	27	88	2663	37	2012-08-02 05:36:39.232867
27259	3	1-8040	27259	88	43	4054	17	2012-08-02 05:36:39.232867
27263	5	1-6547	27262	91	72	4092	100	2012-08-02 05:36:40.415022
27264	5	1-5246	27262	72	19	1735	11	2012-08-02 05:36:40.415022
27265	5	1-7881	27262	19	18	2026	82	2012-08-02 05:36:40.415022
27266	5	1-7702	27262	18	87	1829	23	2012-08-02 05:36:40.415022
27262	5	1-8043	27262	87	91	4197	52	2012-08-02 05:36:40.415022
27268	4	1-7716	27267	78	71	4605	68	2012-08-02 05:36:40.823859
27269	4	1-6002	27267	71	59	2977	83	2012-08-02 05:36:40.823859
27270	4	1-7599	27267	59	19	4005	95	2012-08-02 05:36:40.823859
27267	4	1-8044	27267	19	78	8285	19	2012-08-02 05:36:40.823859
27272	4	1-7207	27271	92	99	756	65	2012-08-02 05:36:40.823859
27273	4	1-5622	27271	99	59	299	83	2012-08-02 05:36:40.823859
27274	4	1-7599	27271	59	19	439	95	2012-08-02 05:36:40.823859
27271	4	1-8044	27271	19	92	993	19	2012-08-02 05:36:40.823859
27276	5	1-7990	27275	48	7	2688	89	2012-08-02 05:36:41.232845
27277	5	1-7616	27275	7	63	2524	86	2012-08-02 05:36:41.232845
27278	5	1-7765	27275	63	77	1206	3	2012-08-02 05:36:41.232845
27279	5	1-7568	27275	77	15	2599	31	2012-08-02 05:36:41.232845
27275	5	1-8045	27275	15	48	4427	83	2012-08-02 05:36:41.232845
27281	3	1-7700	27280	85	62	2880	87	2012-08-02 05:36:41.630396
27282	3	1-7277	27280	62	81	579	31	2012-08-02 05:36:41.630396
27280	3	1-8046	27280	81	85	4295	37	2012-08-02 05:36:41.630396
27284	4	1-7919	27283	74	75	19	18	2012-08-02 05:36:43.607933
27285	4	1-6894	27283	75	63	6	56	2012-08-02 05:36:43.607933
27286	4	1-7807	27283	63	44	11	79	2012-08-02 05:36:43.607933
27283	4	1-8051	27283	44	74	17	6	2012-08-02 05:36:43.607933
27288	4	1-7919	27287	74	99	8	18	2012-08-02 05:36:43.607933
27289	4	1-8034	27287	99	4	8	7	2012-08-02 05:36:43.607933
27290	4	1-6446	27287	4	44	5	79	2012-08-02 05:36:43.607933
27287	4	1-8051	27287	44	74	8	6	2012-08-02 05:36:43.607933
27291	1	1-2344	\N	58	58	1438	79	2012-08-02 05:36:43.607933
27293	3	1-4100	27292	91	62	795	45	2012-08-02 05:36:47.2535
27294	3	1-7547	27292	62	32	1062	70	2012-08-02 05:36:47.2535
27292	3	1-8060	27292	32	91	6194	40	2012-08-02 05:36:47.2535
27296	4	1-7649	27295	61	81	810	82	2012-08-02 05:36:47.2535
27297	4	1-7834	27295	81	69	975	92	2012-08-02 05:36:47.2535
27298	4	1-7741	27295	69	32	424	70	2012-08-02 05:36:47.2535
27295	4	1-8060	27295	32	61	3288	40	2012-08-02 05:36:47.2535
27300	3	1-6930	27299	22	62	934	73	2012-08-02 05:36:47.64057
27301	3	1-7419	27299	62	10	3269	2	2012-08-02 05:36:47.64057
27299	3	1-8061	27299	10	22	3272	27	2012-08-02 05:36:47.64057
27303	4	1-7756	27302	57	57	5365	71	2012-08-02 05:36:48.424849
27304	4	1-6953	27302	57	1	5429	44	2012-08-02 05:36:48.424849
27305	4	1-7889	27302	1	36	2427	66	2012-08-02 05:36:48.424849
27302	4	1-8063	27302	36	57	8486	89	2012-08-02 05:36:48.424849
27307	2	1-2501	27306	95	33	520	83	2012-08-02 05:36:49.231189
27306	2	1-8065	27306	33	95	2173	8	2012-08-02 05:36:49.231189
27309	4	1-3233	27308	65	88	1762	37	2012-08-02 05:36:49.231189
27310	4	1-8040	27308	88	43	1392	17	2012-08-02 05:36:49.231189
27311	4	1-7424	27308	43	33	612	83	2012-08-02 05:36:49.231189
27308	4	1-8065	27308	33	65	4109	8	2012-08-02 05:36:49.231189
27313	3	1-2207	27312	66	92	9	46	2012-08-02 05:36:49.231189
27314	3	1-7789	27312	92	33	3	83	2012-08-02 05:36:49.231189
27312	3	1-8065	27312	33	66	22	8	2012-08-02 05:36:49.231189
27316	2	1-7921	27315	32	98	428	76	2012-08-02 05:36:50.822423
27315	2	1-8069	27315	98	32	551	54	2012-08-02 05:36:50.822423
27318	4	1-8022	27317	80	9	4080	60	2012-08-02 05:36:51.617508
27319	4	1-7952	27317	9	68	2430	56	2012-08-02 05:36:51.617508
27320	4	1-7813	27317	68	24	4932	57	2012-08-02 05:36:51.617508
27317	4	1-8071	27317	24	80	1445	37	2012-08-02 05:36:51.617508
27321	4	1-8074	27321	12	19	2399	98	2012-08-02 05:36:52.821791
27322	4	1-5231	27321	19	31	1984	72	2012-08-02 05:36:52.821791
27323	4	1-7830	27321	31	2	3099	73	2012-08-02 05:36:52.821791
27324	4	1-7019	27321	2	12	1881	33	2012-08-02 05:36:52.821791
27326	4	1-7611	27325	14	61	4848	16	2012-08-02 05:36:56.036532
27327	4	1-7833	27325	61	45	3286	55	2012-08-02 05:36:56.036532
27328	4	1-4730	27325	45	41	4447	37	2012-08-02 05:36:56.036532
27325	4	1-8082	27325	41	14	3834	66	2012-08-02 05:36:56.036532
27330	3	1-7153	27329	3	8	3133	99	2012-08-02 05:36:56.434306
27331	3	1-8004	27329	8	83	4701	18	2012-08-02 05:36:56.434306
27329	3	1-8083	27329	83	3	4738	22	2012-08-02 05:36:56.434306
27333	4	1-6119	27332	30	19	1027	11	2012-08-02 05:36:56.832124
27334	4	1-7881	27332	19	25	1130	82	2012-08-02 05:36:56.832124
27335	4	1-7812	27332	25	25	1560	57	2012-08-02 05:36:56.832124
27332	4	1-8084	27332	25	30	1019	30	2012-08-02 05:36:56.832124
27337	4	1-7199	27336	39	44	2185	79	2012-08-02 05:36:56.832124
27338	4	1-8051	27336	44	42	3953	6	2012-08-02 05:36:56.832124
27339	4	1-7946	27336	42	25	3921	57	2012-08-02 05:36:56.832124
27336	4	1-8084	27336	25	39	2592	30	2012-08-02 05:36:56.832124
27341	5	1-6738	27340	13	18	1515	82	2012-08-02 05:36:58.036673
27342	5	1-7702	27340	18	87	1080	23	2012-08-02 05:36:58.036673
27343	5	1-8043	27340	87	7	1958	52	2012-08-02 05:36:58.036673
27344	5	1-5648	27340	7	97	1063	20	2012-08-02 05:36:58.036673
27340	5	1-8087	27340	97	13	2108	94	2012-08-02 05:36:58.036673
27346	3	1-7318	27345	92	48	1031	71	2012-08-02 05:36:58.036673
27347	3	1-6645	27345	48	97	491	20	2012-08-02 05:36:58.036673
27345	3	1-8087	27345	97	92	1141	94	2012-08-02 05:36:58.036673
27349	5	1-7813	27348	68	74	252	57	2012-08-02 05:36:58.854214
27350	5	1-7821	27348	74	53	286	43	2012-08-02 05:36:58.854214
27351	5	1-5515	27348	53	59	119	62	2012-08-02 05:36:58.854214
27352	5	1-7584	27348	59	39	54	35	2012-08-02 05:36:58.854214
27348	5	1-8089	27348	39	68	107	56	2012-08-02 05:36:58.854214
27354	5	1-7909	27353	72	34	2772	36	2012-08-02 05:36:58.854214
27355	5	1-2343	27353	34	34	1180	5	2012-08-02 05:36:58.854214
27356	5	1-7005	27353	34	59	897	62	2012-08-02 05:36:58.854214
27357	5	1-7584	27353	59	39	451	35	2012-08-02 05:36:58.854214
27353	5	1-8089	27353	39	72	998	56	2012-08-02 05:36:58.854214
27358	1	1-3757	\N	12	12	6373	57	2012-08-02 05:36:58.854214
27360	4	1-7837	27359	65	32	2036	64	2012-08-02 05:36:59.671529
27361	4	1-7315	27359	32	29	2332	88	2012-08-02 05:36:59.671529
27362	4	1-6834	27359	29	40	1667	95	2012-08-02 05:36:59.671529
27359	4	1-8091	27359	40	65	2848	15	2012-08-02 05:36:59.671529
27364	3	1-8048	27363	38	83	3807	60	2012-08-02 05:36:59.671529
27365	3	1-8055	27363	83	40	2735	95	2012-08-02 05:36:59.671529
27363	3	1-8091	27363	40	38	5001	15	2012-08-02 05:36:59.671529
27367	4	1-6786	27366	80	34	5004	5	2012-08-02 05:37:00.069326
27368	4	1-7005	27366	34	3	2328	62	2012-08-02 05:37:00.069326
27369	4	1-7993	27366	3	33	1662	83	2012-08-02 05:37:00.069326
27366	4	1-8092	27366	33	80	8374	69	2012-08-02 05:37:00.069326
27370	1	1-2885	\N	3	3	2234	5	2012-08-02 05:37:00.069326
27372	4	1-7914	27371	73	78	4557	92	2012-08-02 05:37:00.511334
27373	4	1-7711	27371	78	58	1392	1	2012-08-02 05:37:00.511334
27374	4	1-7156	27371	58	89	995	38	2012-08-02 05:37:00.511334
27371	4	1-8093	27371	89	73	9239	90	2012-08-02 05:37:00.511334
27376	3	1-7714	27375	99	43	1119	26	2012-08-02 05:37:00.909034
27377	3	1-5263	27375	43	61	813	54	2012-08-02 05:37:00.909034
27375	3	1-8094	27375	61	99	1213	13	2012-08-02 05:37:00.909034
27379	4	1-7714	27378	99	76	53	26	2012-08-02 05:37:00.909034
27380	4	1-7244	27378	76	18	27	81	2012-08-02 05:37:00.909034
27381	4	1-7976	27378	18	61	28	54	2012-08-02 05:37:00.909034
27378	4	1-8094	27378	61	99	49	13	2012-08-02 05:37:00.909034
27383	4	1-6464	27382	10	50	215	32	2012-08-02 05:37:00.909034
27384	4	1-4143	27382	50	18	232	81	2012-08-02 05:37:00.909034
27385	4	1-7976	27382	18	61	266	54	2012-08-02 05:37:00.909034
27382	4	1-8094	27382	61	10	509	13	2012-08-02 05:37:00.909034
27387	3	1-6848	27386	7	4	1174	58	2012-08-02 05:37:02.102582
27388	3	1-7911	27386	4	6	1068	53	2012-08-02 05:37:02.102582
27386	3	1-8097	27386	6	7	2229	4	2012-08-02 05:37:02.102582
27390	4	1-6814	27389	81	78	2883	92	2012-08-02 05:37:02.102582
27391	4	1-7711	27389	78	98	1727	1	2012-08-02 05:37:02.102582
27392	4	1-5555	27389	98	6	1679	53	2012-08-02 05:37:02.102582
27389	4	1-8097	27389	6	81	4059	4	2012-08-02 05:37:02.102582
27394	3	1-7856	27393	80	14	469	66	2012-08-02 05:37:02.888895
27395	3	1-7611	27393	14	63	509	16	2012-08-02 05:37:02.888895
27393	3	1-8099	27393	63	80	303	20	2012-08-02 05:37:02.888895
27397	4	1-7437	27396	92	55	149	99	2012-08-02 05:37:04.114703
27398	4	1-6728	27396	55	88	310	97	2012-08-02 05:37:04.114703
27399	4	1-5548	27396	88	32	134	3	2012-08-02 05:37:04.114703
27396	4	1-8102	27396	32	92	292	21	2012-08-02 05:37:04.114703
27401	4	1-7980	27400	18	62	5098	75	2012-08-02 05:37:04.523585
27402	4	1-5712	27400	62	1	3456	84	2012-08-02 05:37:04.523585
27403	4	1-7657	27400	1	68	1984	67	2012-08-02 05:37:04.523585
27400	4	1-8103	27400	68	18	7451	46	2012-08-02 05:37:04.523585
27405	5	1-8009	27404	81	22	5760	28	2012-08-02 05:37:06.148944
27406	5	1-6204	27404	22	25	4313	57	2012-08-02 05:37:06.148944
27407	5	1-8084	27404	25	30	2741	30	2012-08-02 05:37:06.148944
27408	5	1-6119	27404	30	8	2812	11	2012-08-02 05:37:06.148944
27404	5	1-8107	27404	8	81	2723	81	2012-08-02 05:37:06.148944
27410	5	1-7589	27409	68	85	1647	37	2012-08-02 05:37:07.761867
27411	5	1-7700	27409	85	9	1948	87	2012-08-02 05:37:07.761867
27412	5	1-7764	27409	9	4	571	47	2012-08-02 05:37:07.761867
27413	5	1-5088	27409	4	78	454	65	2012-08-02 05:37:07.761867
27409	5	1-8111	27409	78	68	990	21	2012-08-02 05:37:07.761867
27415	5	1-7878	27414	48	46	1724	98	2012-08-02 05:37:08.977078
27416	5	1-7796	27414	46	6	1724	73	2012-08-02 05:37:08.977078
27417	5	1-6910	27414	6	14	410	91	2012-08-02 05:37:08.977078
27418	5	1-7505	27414	14	70	666	76	2012-08-02 05:37:08.977078
27414	5	1-8114	27414	70	48	1604	69	2012-08-02 05:37:08.977078
27420	4	1-7279	27419	32	62	1631	45	2012-08-02 05:37:10.988066
27421	4	1-7547	27419	62	61	2411	70	2012-08-02 05:37:10.988066
27422	4	1-7975	27419	61	73	2426	98	2012-08-02 05:37:10.988066
27419	4	1-8119	27419	73	32	4405	38	2012-08-02 05:37:10.988066
27424	3	1-7887	27423	96	12	151	33	2012-08-02 05:37:10.988066
27425	3	1-8074	27423	12	73	167	98	2012-08-02 05:37:10.988066
27423	3	1-8119	27423	73	96	325	38	2012-08-02 05:37:10.988066
27427	3	1-7636	27426	42	74	1340	6	2012-08-02 05:37:11.396571
27428	3	1-7919	27426	74	77	1385	18	2012-08-02 05:37:11.396571
27426	3	1-8120	27426	77	42	5136	63	2012-08-02 05:37:11.396571
27430	5	1-8110	27429	12	50	1190	77	2012-08-02 05:37:11.396571
27431	5	1-7377	27429	50	3	724	22	2012-08-02 05:37:11.396571
27432	5	1-7153	27429	3	8	446	99	2012-08-02 05:37:11.396571
27433	5	1-8004	27429	8	77	624	18	2012-08-02 05:37:11.396571
27429	5	1-8120	27429	77	12	2389	63	2012-08-02 05:37:11.396571
27435	4	1-2794	27434	10	64	1802	17	2012-08-02 05:37:11.794453
27436	4	1-7569	27434	64	38	2121	15	2012-08-02 05:37:11.794453
27437	4	1-8048	27434	38	25	1463	60	2012-08-02 05:37:11.794453
27434	4	1-8121	27434	25	10	1967	55	2012-08-02 05:37:11.794453
27439	4	1-1374	27438	90	6	214	53	2012-08-02 05:37:11.794453
27440	4	1-8097	27438	6	69	483	4	2012-08-02 05:37:11.794453
27441	4	1-8001	27438	69	25	508	60	2012-08-02 05:37:11.794453
27438	4	1-8121	27438	25	90	684	55	2012-08-02 05:37:11.794453
27443	4	1-7281	27442	43	81	1631	39	2012-08-02 05:37:12.61224
27444	4	1-6927	27442	81	98	1051	61	2012-08-02 05:37:12.61224
27445	4	1-7969	27442	98	70	1105	100	2012-08-02 05:37:12.61224
27442	4	1-8123	27442	70	43	2810	11	2012-08-02 05:37:12.61224
27447	4	1-7907	27446	78	64	597	75	2012-08-02 05:37:12.61224
27448	4	1-7413	27446	64	98	299	61	2012-08-02 05:37:12.61224
27449	4	1-7969	27446	98	70	353	100	2012-08-02 05:37:12.61224
27446	4	1-8123	27446	70	78	1008	11	2012-08-02 05:37:12.61224
27451	4	1-7728	27450	94	41	1453	66	2012-08-02 05:37:12.61224
27452	4	1-7671	27450	41	56	898	71	2012-08-02 05:37:12.61224
27453	4	1-7140	27450	56	70	410	100	2012-08-02 05:37:12.61224
27450	4	1-8123	27450	70	94	1428	11	2012-08-02 05:37:12.61224
27455	4	1-4716	27454	36	25	609	4	2012-08-02 05:37:13.02102
27456	4	1-6533	27454	25	95	406	59	2012-08-02 05:37:13.02102
27457	4	1-5711	27454	95	5	129	63	2012-08-02 05:37:13.02102
27454	4	1-8124	27454	5	36	1576	29	2012-08-02 05:37:13.02102
27459	4	1-4716	27458	36	7	253	4	2012-08-02 05:37:13.02102
27460	4	1-7735	27458	7	95	155	59	2012-08-02 05:37:13.02102
27461	4	1-5711	27458	95	5	51	63	2012-08-02 05:37:13.02102
27458	4	1-8124	27458	5	36	637	29	2012-08-02 05:37:13.02102
27463	4	1-7209	27462	84	94	4301	79	2012-08-02 05:37:13.418798
27464	4	1-7653	27462	94	47	4759	44	2012-08-02 05:37:13.418798
27465	4	1-7358	27462	47	88	2491	49	2012-08-02 05:37:13.418798
27462	4	1-8125	27462	88	84	6631	66	2012-08-02 05:37:13.418798
27467	5	1-7611	27466	14	16	262	16	2012-08-02 05:37:13.418798
27468	5	1-4323	27466	16	72	42	56	2012-08-02 05:37:13.418798
27469	5	1-7909	27466	72	8	102	36	2012-08-02 05:37:13.418798
27470	5	1-7753	27466	8	88	73	49	2012-08-02 05:37:13.418798
27466	5	1-8125	27466	88	14	201	66	2012-08-02 05:37:13.418798
27472	4	1-7649	27471	61	83	1113	82	2012-08-02 05:37:13.838654
27473	4	1-7997	27471	83	32	1575	9	2012-08-02 05:37:13.838654
27474	4	1-8033	27471	32	23	630	81	2012-08-02 05:37:13.838654
27471	4	1-8126	27471	23	61	7414	40	2012-08-02 05:37:13.838654
27476	3	1-6513	27475	5	86	87	65	2012-08-02 05:37:13.838654
27477	3	1-4699	27475	86	23	29	81	2012-08-02 05:37:13.838654
27475	3	1-8126	27475	23	5	494	40	2012-08-02 05:37:13.838654
27479	4	1-7505	27478	14	70	3271	76	2012-08-02 05:37:16.346497
27480	4	1-8114	27478	70	46	6025	69	2012-08-02 05:37:16.346497
27481	4	1-7027	27478	46	10	3257	34	2012-08-02 05:37:16.346497
27478	4	1-8132	27478	10	14	2631	91	2012-08-02 05:37:16.346497
27483	5	1-7909	27482	72	22	1176	36	2012-08-02 05:37:19.185489
27484	5	1-6982	27482	22	77	216	18	2012-08-02 05:37:19.185489
27485	5	1-8120	27482	77	5	631	63	2012-08-02 05:37:19.185489
27486	5	1-8124	27482	5	95	5446	29	2012-08-02 05:37:19.185489
27482	5	1-8139	27482	95	72	671	56	2012-08-02 05:37:19.185489
27488	4	1-7513	27487	55	32	376	32	2012-08-02 05:37:19.981044
27489	4	1-4604	27487	32	55	153	5	2012-08-02 05:37:19.981044
27490	4	1-7181	27487	55	51	101	69	2012-08-02 05:37:19.981044
27487	4	1-8141	27487	51	55	386	7	2012-08-02 05:37:19.981044
27492	4	1-7879	27491	71	32	2986	32	2012-08-02 05:37:19.981044
27493	4	1-4604	27491	32	55	1629	5	2012-08-02 05:37:19.981044
27494	4	1-7181	27491	55	51	1036	69	2012-08-02 05:37:19.981044
27491	4	1-8141	27491	51	71	5335	7	2012-08-02 05:37:19.981044
27496	4	1-6770	27495	34	39	669	94	2012-08-02 05:37:19.981044
27497	4	1-6943	27495	39	55	1050	5	2012-08-02 05:37:19.981044
27498	4	1-7181	27495	55	51	693	69	2012-08-02 05:37:19.981044
27495	4	1-8141	27495	51	34	3712	7	2012-08-02 05:37:19.981044
27500	4	1-7958	27499	2	53	1107	43	2012-08-02 05:37:21.185242
27501	4	1-5515	27499	53	74	196	62	2012-08-02 05:37:21.185242
27502	4	1-7973	27499	74	62	97	44	2012-08-02 05:37:21.185242
27499	4	1-8144	27499	62	2	3201	31	2012-08-02 05:37:21.185242
27504	4	1-8094	27503	61	54	1790	13	2012-08-02 05:37:21.992091
27505	4	1-7885	27503	54	62	1548	80	2012-08-02 05:37:21.992091
27506	4	1-7880	27503	62	97	1273	97	2012-08-02 05:37:21.992091
27503	4	1-8146	27503	97	61	1046	54	2012-08-02 05:37:21.992091
27508	5	1-6547	27507	91	52	536	100	2012-08-02 05:37:22.411927
27509	5	1-7795	27507	52	64	269	19	2012-08-02 05:37:22.411927
27510	5	1-6442	27507	64	92	379	21	2012-08-02 05:37:22.411927
27511	5	1-7437	27507	92	8	205	99	2012-08-02 05:37:22.411927
27507	5	1-8147	27507	8	91	554	52	2012-08-02 05:37:22.411927
27513	3	1-7953	27512	91	80	1936	69	2012-08-02 05:37:23.615718
27514	3	1-6786	27512	80	61	1376	5	2012-08-02 05:37:23.615718
27512	3	1-8150	27512	61	91	2541	42	2012-08-02 05:37:23.615718
27516	4	1-7953	27515	91	67	57	69	2012-08-02 05:37:23.615718
27517	4	1-2285	27515	67	39	13	94	2012-08-02 05:37:23.615718
27518	4	1-6943	27515	39	61	26	5	2012-08-02 05:37:23.615718
27515	4	1-8150	27515	61	91	60	42	2012-08-02 05:37:23.615718
27520	3	1-7857	27519	22	30	2104	36	2012-08-02 05:37:24.809003
27521	3	1-4660	27519	30	52	1640	57	2012-08-02 05:37:24.809003
27519	3	1-8153	27519	52	22	3391	62	2012-08-02 05:37:24.809003
27523	4	1-8120	27522	77	42	930	63	2012-08-02 05:37:25.789034
27524	4	1-7636	27522	42	42	264	6	2012-08-02 05:37:25.789034
27525	4	1-7946	27522	42	55	235	57	2012-08-02 05:37:25.789034
27522	4	1-8155	27522	55	77	266	18	2012-08-02 05:37:25.789034
27527	4	1-6894	27526	75	72	836	56	2012-08-02 05:37:25.789034
27528	4	1-7909	27526	72	30	2150	36	2012-08-02 05:37:25.789034
27529	4	1-4660	27526	30	55	1677	57	2012-08-02 05:37:25.789034
27526	4	1-8155	27526	55	75	2321	18	2012-08-02 05:37:25.789034
27531	4	1-7855	27530	78	89	1936	93	2012-08-02 05:37:26.201113
27532	4	1-3329	27530	89	85	341	25	2012-08-02 05:37:26.201113
27533	4	1-7983	27530	85	50	577	84	2012-08-02 05:37:26.201113
27530	4	1-8156	27530	50	78	1230	12	2012-08-02 05:37:26.201113
27535	4	1-7949	27534	46	88	117	95	2012-08-02 05:37:26.201113
27536	4	1-7356	27534	88	63	42	63	2012-08-02 05:37:26.201113
27537	4	1-4744	27534	63	50	15	84	2012-08-02 05:37:26.201113
27534	4	1-8156	27534	50	46	39	12	2012-08-02 05:37:26.201113
27539	4	1-7955	27538	91	10	5470	51	2012-08-02 05:37:26.201113
27540	4	1-8028	27538	10	49	2239	45	2012-08-02 05:37:26.201113
27541	4	1-4598	27538	49	50	792	84	2012-08-02 05:37:26.201113
27538	4	1-8156	27538	50	91	2104	12	2012-08-02 05:37:26.201113
27543	4	1-4987	27542	98	30	256	30	2012-08-02 05:37:26.201113
27544	4	1-6119	27542	30	74	268	11	2012-08-02 05:37:26.201113
27545	4	1-5806	27542	74	50	169	84	2012-08-02 05:37:26.201113
27542	4	1-8156	27542	50	98	487	12	2012-08-02 05:37:26.201113
27547	5	1-7203	27546	65	73	630	14	2012-08-02 05:37:27.007673
27548	5	1-7858	27546	73	5	368	40	2012-08-02 05:37:27.007673
27549	5	1-6513	27546	5	78	52	65	2012-08-02 05:37:27.007673
27550	5	1-8111	27546	78	45	91	21	2012-08-02 05:37:27.007673
27546	5	1-8158	27546	45	65	412	72	2012-08-02 05:37:27.007673
27552	5	1-2526	27551	16	61	1322	5	2012-08-02 05:37:27.007673
27553	5	1-8150	27551	61	69	2323	42	2012-08-02 05:37:27.007673
27554	5	1-6111	27551	69	78	332	65	2012-08-02 05:37:27.007673
27555	5	1-8111	27551	78	45	595	21	2012-08-02 05:37:27.007673
27551	5	1-8158	27551	45	16	2755	72	2012-08-02 05:37:27.007673
27557	5	1-7868	27556	93	89	2092	55	2012-08-02 05:37:27.007673
27558	5	1-3643	27556	89	93	892	73	2012-08-02 05:37:27.007673
27559	5	1-6193	27556	93	78	408	65	2012-08-02 05:37:27.007673
27560	5	1-8111	27556	78	45	782	21	2012-08-02 05:37:27.007673
27556	5	1-8158	27556	45	93	3874	72	2012-08-02 05:37:27.007673
27562	4	1-6834	27561	29	61	3973	95	2012-08-02 05:37:27.82533
27563	4	1-2319	27561	61	61	1589	54	2012-08-02 05:37:27.82533
27564	4	1-8094	27561	61	69	3056	13	2012-08-02 05:37:27.82533
27561	4	1-8160	27561	69	29	5550	88	2012-08-02 05:37:27.82533
27566	3	1-7365	27565	40	30	1901	69	2012-08-02 05:37:28.598722
27567	3	1-6076	27565	30	21	1058	59	2012-08-02 05:37:28.598722
27565	3	1-8162	27565	21	40	2173	26	2012-08-02 05:37:28.598722
27569	4	1-7244	27568	76	81	949	81	2012-08-02 05:37:28.598722
27570	4	1-8009	27568	81	95	1993	28	2012-08-02 05:37:28.598722
27571	4	1-7223	27568	95	21	802	59	2012-08-02 05:37:28.598722
27568	4	1-8162	27568	21	76	1789	26	2012-08-02 05:37:28.598722
27573	5	1-8121	27572	25	99	554	55	2012-08-02 05:37:29.416694
27574	5	1-7913	27572	99	10	662	34	2012-08-02 05:37:29.416694
27575	5	1-8132	27572	10	14	418	91	2012-08-02 05:37:29.416694
27576	5	1-7505	27572	14	88	406	76	2012-08-02 05:37:29.416694
27572	5	1-8164	27572	88	25	616	60	2012-08-02 05:37:29.416694
27578	4	1-8121	27577	25	10	9	55	2012-08-02 05:37:29.416694
27579	4	1-2794	27577	10	61	9	17	2012-08-02 05:37:29.416694
27580	4	1-7498	27577	61	88	3	76	2012-08-02 05:37:29.416694
27577	4	1-8164	27577	88	25	7	60	2012-08-02 05:37:29.416694
27582	5	1-8099	27581	63	97	1894	20	2012-08-02 05:37:29.825212
27583	5	1-8087	27581	97	7	4400	94	2012-08-02 05:37:29.825212
27584	5	1-6993	27581	7	50	2986	5	2012-08-02 05:37:29.825212
27585	5	1-7910	27581	50	82	2923	55	2012-08-02 05:37:29.825212
27581	5	1-8165	27581	82	63	2562	16	2012-08-02 05:37:29.825212
27587	5	1-6667	27586	16	57	2735	61	2012-08-02 05:37:30.246944
27588	5	1-5116	27586	57	50	1577	84	2012-08-02 05:37:30.246944
27589	5	1-8156	27586	50	78	2816	12	2012-08-02 05:37:30.246944
27590	5	1-7855	27586	78	46	3713	93	2012-08-02 05:37:30.246944
27586	5	1-8166	27586	46	16	6968	20	2012-08-02 05:37:30.246944
27592	3	1-7247	27591	56	42	1261	70	2012-08-02 05:37:30.246944
27593	3	1-4241	27591	42	46	498	93	2012-08-02 05:37:30.246944
27591	3	1-8166	27591	46	56	1415	20	2012-08-02 05:37:30.246944
27595	3	1-7799	27594	36	81	2712	16	2012-08-02 05:37:30.655721
27596	3	1-5527	27594	81	80	971	54	2012-08-02 05:37:30.655721
27594	3	1-8167	27594	80	36	6743	6	2012-08-02 05:37:30.655721
27598	4	1-5740	27597	93	62	1577	80	2012-08-02 05:37:31.053609
27599	4	1-7880	27597	62	96	1390	97	2012-08-02 05:37:31.053609
27600	4	1-4356	27597	96	88	663	13	2012-08-02 05:37:31.053609
27597	4	1-8168	27597	88	93	1451	45	2012-08-02 05:37:31.053609
27602	4	1-4598	27601	49	89	1663	84	2012-08-02 05:37:31.053609
27603	4	1-6850	27601	89	19	1835	81	2012-08-02 05:37:31.053609
27604	4	1-7956	27601	19	88	1909	13	2012-08-02 05:37:31.053609
27601	4	1-8168	27601	88	49	4342	45	2012-08-02 05:37:31.053609
27606	3	1-8156	27605	50	51	356	12	2012-08-02 05:37:31.849236
27607	3	1-5206	27605	51	65	289	89	2012-08-02 05:37:31.849236
27605	3	1-8170	27605	65	50	139	84	2012-08-02 05:37:31.849236
27609	4	1-7795	27608	52	92	618	19	2012-08-02 05:37:32.246722
27610	4	1-7207	27608	92	99	644	65	2012-08-02 05:37:32.246722
27611	4	1-5622	27608	99	75	348	83	2012-08-02 05:37:32.246722
27608	4	1-8171	27608	75	52	1440	100	2012-08-02 05:37:32.246722
27613	4	1-3959	27612	22	88	1138	13	2012-08-02 05:37:32.246722
27614	4	1-8168	27612	88	99	2686	45	2012-08-02 05:37:32.246722
27615	4	1-5149	27612	99	75	455	83	2012-08-02 05:37:32.246722
27612	4	1-8171	27612	75	22	2256	100	2012-08-02 05:37:32.246722
27617	5	1-6738	27616	13	83	509	82	2012-08-02 05:37:33.064634
27618	5	1-7997	27616	83	46	1091	9	2012-08-02 05:37:33.064634
27619	5	1-7157	27616	46	84	323	24	2012-08-02 05:37:33.064634
27620	5	1-8113	27616	84	67	380	68	2012-08-02 05:37:33.064634
27616	5	1-8173	27616	67	13	616	94	2012-08-02 05:37:33.064634
27622	4	1-6738	27621	13	24	2813	82	2012-08-02 05:37:33.064634
27623	4	1-6116	27621	24	28	1977	62	2012-08-02 05:37:33.064634
27624	4	1-7169	27621	28	67	1785	68	2012-08-02 05:37:33.064634
27621	4	1-8173	27621	67	13	3142	94	2012-08-02 05:37:33.064634
27626	3	1-7608	27625	90	82	1670	52	2012-08-02 05:37:33.893283
27627	3	1-7987	27625	82	42	1338	91	2012-08-02 05:37:33.893283
27625	3	1-8175	27625	42	90	1805	12	2012-08-02 05:37:33.893283
27629	5	1-5206	27628	51	7	847	89	2012-08-02 05:37:33.893283
27630	5	1-7616	27628	7	78	679	86	2012-08-02 05:37:33.893283
27631	5	1-7287	27628	78	10	904	34	2012-08-02 05:37:33.893283
27632	5	1-8132	27628	10	42	790	91	2012-08-02 05:37:33.893283
27628	5	1-8175	27628	42	51	1073	12	2012-08-02 05:37:33.893283
27634	5	1-8073	27633	62	8	1848	68	2012-08-02 05:37:33.893283
27635	5	1-6919	27633	8	99	1603	55	2012-08-02 05:37:33.893283
27636	5	1-7913	27633	99	10	2669	34	2012-08-02 05:37:33.893283
27637	5	1-8132	27633	10	42	2347	91	2012-08-02 05:37:33.893283
27633	5	1-8175	27633	42	62	3209	12	2012-08-02 05:37:33.893283
27638	1	1-2515	\N	20	20	1188	55	2012-08-02 05:37:33.893283
27640	3	1-7005	27639	34	74	751	62	2012-08-02 05:37:35.561491
27641	3	1-7973	27639	74	1	658	44	2012-08-02 05:37:35.561491
27639	3	1-8179	27639	1	34	1456	5	2012-08-02 05:37:35.561491
27643	4	1-7910	27642	50	99	269	55	2012-08-02 05:37:35.561491
27644	4	1-7913	27642	99	60	401	34	2012-08-02 05:37:35.561491
27645	4	1-5953	27642	60	1	138	44	2012-08-02 05:37:35.561491
27642	4	1-8179	27642	1	50	333	5	2012-08-02 05:37:35.561491
27647	4	1-5738	27646	3	84	568	54	2012-08-02 05:37:35.561491
27648	4	1-7690	27646	84	94	221	79	2012-08-02 05:37:35.561491
27649	4	1-7653	27646	94	1	246	44	2012-08-02 05:37:35.561491
27646	4	1-8179	27646	1	3	683	5	2012-08-02 05:37:35.561491
27651	2	1-8094	27650	61	47	268	13	2012-08-02 05:37:36.378924
27650	2	1-8181	27650	47	61	235	54	2012-08-02 05:37:36.378924
27653	5	1-8147	27652	8	46	3808	52	2012-08-02 05:37:36.765697
27654	5	1-8025	27652	46	100	1505	45	2012-08-02 05:37:36.765697
27655	5	1-6412	27652	100	94	1736	79	2012-08-02 05:37:36.765697
27656	5	1-7653	27652	94	95	1809	44	2012-08-02 05:37:36.765697
27652	5	1-8182	27652	95	8	1699	99	2012-08-02 05:37:36.765697
27658	4	1-8147	27657	8	7	1864	52	2012-08-02 05:37:36.765697
27659	4	1-7590	27657	7	80	963	68	2012-08-02 05:37:36.765697
27660	4	1-4680	27657	80	95	710	44	2012-08-02 05:37:36.765697
27657	4	1-8182	27657	95	8	745	99	2012-08-02 05:37:36.765697
27662	5	1-6728	27661	55	97	42	97	2012-08-02 05:37:36.765697
27663	5	1-8146	27661	97	84	37	54	2012-08-02 05:37:36.765697
27664	5	1-7690	27661	84	94	16	79	2012-08-02 05:37:36.765697
27665	5	1-7653	27661	94	95	19	44	2012-08-02 05:37:36.765697
27661	5	1-8182	27661	95	55	20	99	2012-08-02 05:37:36.765697
27667	4	1-8004	27666	8	31	1560	18	2012-08-02 05:37:36.765697
27668	4	1-7998	27666	31	73	1542	1	2012-08-02 05:37:36.765697
27669	4	1-5784	27666	73	95	853	44	2012-08-02 05:37:36.765697
27666	4	1-8182	27666	95	8	962	99	2012-08-02 05:37:36.765697
27671	3	1-7286	27670	67	80	647	68	2012-08-02 05:37:36.765697
27672	3	1-4680	27670	80	95	528	44	2012-08-02 05:37:36.765697
27670	3	1-8182	27670	95	67	613	99	2012-08-02 05:37:36.765697
27674	4	1-8161	27673	66	27	7168	29	2012-08-02 05:37:37.1857
27675	4	1-3177	27673	27	24	264	82	2012-08-02 05:37:37.1857
27676	4	1-6116	27673	24	21	186	62	2012-08-02 05:37:37.1857
27673	4	1-8183	27673	21	66	271	38	2012-08-02 05:37:37.1857
27678	5	1-8105	27677	81	69	2240	97	2012-08-02 05:37:38.003146
27679	5	1-5578	27677	69	22	979	36	2012-08-02 05:37:38.003146
27680	5	1-7516	27677	22	21	1677	59	2012-08-02 05:37:38.003146
27681	5	1-8162	27677	21	28	3516	26	2012-08-02 05:37:38.003146
27677	5	1-8185	27677	28	81	1337	76	2012-08-02 05:37:38.003146
27683	5	1-7680	27682	66	57	1114	71	2012-08-02 05:37:38.423009
27684	5	1-6953	27682	57	95	1245	44	2012-08-02 05:37:38.423009
27685	5	1-8182	27682	95	52	1321	99	2012-08-02 05:37:38.423009
27686	5	1-4709	27682	52	22	1150	28	2012-08-02 05:37:38.423009
27682	5	1-8186	27682	22	66	1268	84	2012-08-02 05:37:38.423009
27688	3	1-6850	27687	89	33	5039	81	2012-08-02 05:37:38.423009
27689	3	1-7510	27687	33	22	3610	28	2012-08-02 05:37:38.423009
27687	3	1-8186	27687	22	89	4375	84	2012-08-02 05:37:38.423009
27691	5	1-6437	27690	7	84	1295	54	2012-08-02 05:37:39.240698
27692	5	1-7690	27690	84	94	593	79	2012-08-02 05:37:39.240698
27693	5	1-7653	27690	94	1	775	44	2012-08-02 05:37:39.240698
27694	5	1-8179	27690	1	100	2523	5	2012-08-02 05:37:39.240698
27690	5	1-8188	27690	100	7	1262	22	2012-08-02 05:37:39.240698
27696	4	1-4562	27695	86	1	1183	44	2012-08-02 05:37:39.638401
27697	4	1-7889	27695	1	71	498	66	2012-08-02 05:37:39.638401
27698	4	1-7587	27695	71	15	642	67	2012-08-02 05:37:39.638401
27695	4	1-8189	27695	15	86	3299	86	2012-08-02 05:37:39.638401
27700	5	1-6808	27699	31	60	1270	34	2012-08-02 05:37:39.638401
27701	5	1-5953	27699	60	1	436	44	2012-08-02 05:37:39.638401
27702	5	1-7889	27699	1	71	216	66	2012-08-02 05:37:39.638401
27703	5	1-7587	27699	71	15	327	67	2012-08-02 05:37:39.638401
27699	5	1-8189	27699	15	31	1976	86	2012-08-02 05:37:39.638401
27705	4	1-7869	27704	14	77	942	47	2012-08-02 05:37:39.638401
27706	4	1-4919	27704	77	1	722	84	2012-08-02 05:37:39.638401
27707	4	1-7657	27704	1	15	431	67	2012-08-02 05:37:39.638401
27704	4	1-8189	27704	15	14	2936	86	2012-08-02 05:37:39.638401
27708	1	1-6808	\N	31	31	1440	34	2012-08-02 05:37:39.638401
27710	4	1-6909	27709	27	25	1155	50	2012-08-02 05:37:42.091071
27711	4	1-3651	27709	25	57	104	25	2012-08-02 05:37:42.091071
27712	4	1-8100	27709	57	97	133	78	2012-08-02 05:37:42.091071
27709	4	1-8195	27709	97	27	3638	63	2012-08-02 05:37:42.091071
27714	4	1-8187	27713	24	42	29	98	2012-08-02 05:37:42.091071
27715	4	1-3794	27713	42	57	3	25	2012-08-02 05:37:42.091071
27716	4	1-8100	27713	57	97	5	78	2012-08-02 05:37:42.091071
27713	4	1-8195	27713	97	24	152	63	2012-08-02 05:37:42.091071
27717	1	1-715	\N	84	84	318	25	2012-08-02 05:37:42.091071
27719	5	1-8161	27718	66	95	1005	29	2012-08-02 05:37:43.295399
27720	5	1-8139	27718	95	92	140	56	2012-08-02 05:37:43.295399
27721	5	1-4773	27718	92	71	40	30	2012-08-02 05:37:43.295399
27722	5	1-4497	27718	71	76	21	27	2012-08-02 05:37:43.295399
27718	5	1-8198	27718	76	66	51	38	2012-08-02 05:37:43.295399
27724	5	1-7337	27723	23	60	5204	14	2012-08-02 05:37:43.295399
27725	5	1-7842	27723	60	95	1207	44	2012-08-02 05:37:43.295399
27726	5	1-8182	27723	95	33	1396	99	2012-08-02 05:37:43.295399
27727	5	1-7737	27723	33	76	1595	27	2012-08-02 05:37:43.295399
27723	5	1-8198	27723	76	23	5451	38	2012-08-02 05:37:43.295399
27728	1	1-4497	\N	71	71	636	27	2012-08-02 05:37:43.295399
27730	4	1-6932	27729	11	35	721	73	2012-08-02 05:37:43.704344
27731	4	1-8118	27729	35	79	715	90	2012-08-02 05:37:43.704344
27732	4	1-5657	27729	79	30	851	53	2012-08-02 05:37:43.704344
27729	4	1-8199	27729	30	11	719	25	2012-08-02 05:37:43.704344
27734	4	1-8100	27733	57	59	5954	78	2012-08-02 05:37:43.704344
27735	4	1-4700	27733	59	38	1488	71	2012-08-02 05:37:43.704344
27736	4	1-7614	27733	38	30	2096	53	2012-08-02 05:37:43.704344
27733	4	1-8199	27733	30	57	1781	25	2012-08-02 05:37:43.704344
27738	4	1-8193	27737	28	45	5822	93	2012-08-02 05:37:44.113272
27739	4	1-5866	27737	45	49	2242	4	2012-08-02 05:37:44.113272
27740	4	1-8133	27737	49	61	3590	40	2012-08-02 05:37:44.113272
27737	4	1-8200	27737	61	28	2871	98	2012-08-02 05:37:44.113272
27742	3	1-7611	27741	14	1	1802	16	2012-08-02 05:37:46.532627
27743	3	1-7768	27741	1	97	788	32	2012-08-02 05:37:46.532627
27741	3	1-8206	27741	97	14	1367	66	2012-08-02 05:37:46.532627
27745	4	1-7939	27744	70	24	1682	10	2012-08-02 05:37:46.532627
27746	4	1-7567	27744	24	16	785	59	2012-08-02 05:37:46.532627
27747	4	1-8067	27744	16	97	380	32	2012-08-02 05:37:46.532627
27744	4	1-8206	27744	97	70	677	66	2012-08-02 05:37:46.532627
27749	4	1-7894	27748	6	8	3219	64	2012-08-02 05:37:46.941394
27750	4	1-7343	27748	8	76	761	27	2012-08-02 05:37:46.941394
27751	4	1-8198	27748	76	20	1975	38	2012-08-02 05:37:46.941394
27748	4	1-8207	27748	20	6	3729	84	2012-08-02 05:37:46.941394
27753	4	1-7894	27752	6	79	1025	64	2012-08-02 05:37:46.941394
27754	4	1-3161	27752	79	58	371	1	2012-08-02 05:37:46.941394
27755	4	1-7156	27752	58	20	471	38	2012-08-02 05:37:46.941394
27752	4	1-8207	27752	20	6	1027	84	2012-08-02 05:37:46.941394
27757	5	1-7587	27756	71	56	1256	67	2012-08-02 05:37:47.339209
27758	5	1-4500	27756	56	83	525	60	2012-08-02 05:37:47.339209
27759	5	1-5669	27756	83	92	211	74	2012-08-02 05:37:47.339209
27760	5	1-6875	27756	92	98	85	12	2012-08-02 05:37:47.339209
27756	5	1-8208	27756	98	71	1044	66	2012-08-02 05:37:47.339209
27762	5	1-2067	27761	64	88	266	76	2012-08-02 05:37:47.339209
27763	5	1-8164	27761	88	83	485	60	2012-08-02 05:37:47.339209
27764	5	1-5669	27761	83	92	238	74	2012-08-02 05:37:47.339209
27765	5	1-6875	27761	92	98	117	12	2012-08-02 05:37:47.339209
27761	5	1-8208	27761	98	64	1763	66	2012-08-02 05:37:47.339209
27767	4	1-7209	27766	84	78	317	79	2012-08-02 05:37:47.339209
27768	4	1-5618	27766	78	14	84	62	2012-08-02 05:37:47.339209
27769	4	1-7629	27766	14	98	28	12	2012-08-02 05:37:47.339209
27766	4	1-8208	27766	98	84	490	66	2012-08-02 05:37:47.339209
27771	5	1-6562	27770	48	93	1666	72	2012-08-02 05:37:48.157087
27772	5	1-7868	27770	93	99	956	55	2012-08-02 05:37:48.157087
27773	5	1-7913	27770	99	70	1765	34	2012-08-02 05:37:48.157087
27774	5	1-8127	27770	70	44	1338	28	2012-08-02 05:37:48.157087
27770	5	1-8210	27770	44	48	2641	64	2012-08-02 05:37:48.157087
27776	4	1-3246	27775	24	8	187	11	2012-08-02 05:37:48.157087
27777	4	1-8107	27775	8	33	191	81	2012-08-02 05:37:48.157087
27778	4	1-7510	27775	33	44	136	28	2012-08-02 05:37:48.157087
27775	4	1-8210	27775	44	24	287	64	2012-08-02 05:37:48.157087
27780	5	1-7686	27779	55	55	996	57	2012-08-02 05:37:49.372078
27781	5	1-8155	27779	55	34	1264	18	2012-08-02 05:37:49.372078
27782	5	1-6378	27779	34	98	898	17	2012-08-02 05:37:49.372078
27783	5	1-7930	27779	98	40	532	26	2012-08-02 05:37:49.372078
27779	5	1-8213	27779	40	55	3798	94	2012-08-02 05:37:49.372078
27785	5	1-5574	27784	80	12	1355	63	2012-08-02 05:37:49.372078
27786	5	1-8110	27784	12	91	596	77	2012-08-02 05:37:49.372078
27787	5	1-8088	27784	91	98	440	17	2012-08-02 05:37:49.372078
27788	5	1-7930	27784	98	40	264	26	2012-08-02 05:37:49.372078
27784	5	1-8213	27784	40	80	1914	94	2012-08-02 05:37:49.372078
27790	5	1-6222	27789	33	6	1336	84	2012-08-02 05:37:49.372078
27791	5	1-7894	27789	6	90	1135	64	2012-08-02 05:37:49.372078
27792	5	1-8149	27789	90	98	892	17	2012-08-02 05:37:49.372078
27793	5	1-7930	27789	98	40	539	26	2012-08-02 05:37:49.372078
27789	5	1-8213	27789	40	33	3926	94	2012-08-02 05:37:49.372078
27795	5	1-8210	27794	44	86	5355	64	2012-08-02 05:37:50.201113
27796	5	1-8014	27794	86	89	1984	93	2012-08-02 05:37:50.201113
27797	5	1-3329	27794	89	57	486	25	2012-08-02 05:37:50.201113
27798	5	1-8100	27794	57	28	1608	78	2012-08-02 05:37:50.201113
27794	5	1-8215	27794	28	44	2545	28	2012-08-02 05:37:50.201113
27800	5	1-5652	27799	62	97	240	37	2012-08-02 05:37:50.201113
27801	5	1-7607	27799	97	42	108	98	2012-08-02 05:37:50.201113
27802	5	1-3794	27799	42	57	29	25	2012-08-02 05:37:50.201113
27803	5	1-8100	27799	57	28	98	78	2012-08-02 05:37:50.201113
27799	5	1-8215	27799	28	62	157	28	2012-08-02 05:37:50.201113
27805	4	1-8186	27804	22	66	171	84	2012-08-02 05:37:50.201113
27806	4	1-7680	27804	66	59	166	71	2012-08-02 05:37:50.201113
27807	4	1-4450	27804	59	28	87	78	2012-08-02 05:37:50.201113
27804	4	1-8215	27804	28	22	140	28	2012-08-02 05:37:50.201113
27808	1	1-1201	\N	61	61	262	28	2012-08-02 05:37:50.201113
27810	4	1-8149	27809	90	98	838	17	2012-08-02 05:37:51.427408
27811	4	1-7930	27809	98	68	636	26	2012-08-02 05:37:51.427408
27812	4	1-4009	27809	68	22	236	94	2012-08-02 05:37:51.427408
27809	4	1-8218	27809	22	90	848	64	2012-08-02 05:37:51.427408
27814	4	1-6720	27813	63	91	234	42	2012-08-02 05:37:51.427408
27815	4	1-7953	27813	91	67	241	69	2012-08-02 05:37:51.427408
27816	4	1-2285	27813	67	22	61	94	2012-08-02 05:37:51.427408
27813	4	1-8218	27813	22	63	234	64	2012-08-02 05:37:51.427408
27817	1	1-2641	\N	58	58	741	94	2012-08-02 05:37:51.427408
27818	1	1-2285	\N	67	67	413	94	2012-08-02 05:37:51.427408
27820	3	1-6512	27819	11	59	4784	76	2012-08-02 05:37:51.847264
27821	3	1-8134	27819	59	56	2983	48	2012-08-02 05:37:51.847264
27819	3	1-8219	27819	56	11	9513	89	2012-08-02 05:37:51.847264
27823	4	1-8111	27822	78	92	1813	21	2012-08-02 05:37:53.846989
27824	4	1-7437	27822	92	33	850	99	2012-08-02 05:37:53.846989
27825	4	1-7737	27822	33	3	824	27	2012-08-02 05:37:53.846989
27822	4	1-8224	27822	3	78	841	65	2012-08-02 05:37:53.846989
27827	3	1-8111	27826	78	26	4514	21	2012-08-02 05:37:53.846989
27828	3	1-5886	27826	26	3	1572	27	2012-08-02 05:37:53.846989
27826	3	1-8224	27826	3	78	1833	65	2012-08-02 05:37:53.846989
27830	4	1-6519	27829	23	26	1136	6	2012-08-02 05:37:55.051162
27831	4	1-6821	27829	26	81	787	77	2012-08-02 05:37:55.051162
27832	4	1-6263	27829	81	79	465	89	2012-08-02 05:37:55.051162
27829	4	1-8227	27829	79	23	1717	8	2012-08-02 05:37:55.051162
27834	4	1-6670	27833	63	24	3400	51	2012-08-02 05:37:55.051162
27835	4	1-7625	27833	24	43	1747	52	2012-08-02 05:37:55.051162
27836	4	1-6659	27833	43	79	1004	89	2012-08-02 05:37:55.051162
27833	4	1-8227	27833	79	63	3858	8	2012-08-02 05:37:55.051162
27838	4	1-8108	27837	74	92	374	56	2012-08-02 05:37:55.051162
27839	4	1-4773	27837	92	77	123	30	2012-08-02 05:37:55.051162
27840	4	1-6973	27837	77	79	90	89	2012-08-02 05:37:55.051162
27837	4	1-8227	27837	79	74	383	8	2012-08-02 05:37:55.051162
27842	4	1-5353	27841	47	28	860	78	2012-08-02 05:37:55.449095
27843	4	1-8215	27841	28	22	1037	28	2012-08-02 05:37:55.449095
27844	4	1-8186	27841	22	59	944	84	2012-08-02 05:37:55.449095
27841	4	1-8228	27841	59	47	1486	96	2012-08-02 05:37:55.449095
27846	5	1-8023	27845	84	74	2257	35	2012-08-02 05:37:55.449095
27847	5	1-7521	27845	74	46	1391	82	2012-08-02 05:37:55.449095
27848	5	1-5324	27845	46	22	989	28	2012-08-02 05:37:55.449095
27849	5	1-8186	27845	22	59	1099	84	2012-08-02 05:37:55.449095
27845	5	1-8228	27845	59	84	2111	96	2012-08-02 05:37:55.449095
27851	3	1-5818	27850	74	57	2436	61	2012-08-02 05:37:55.449095
27852	3	1-5116	27850	57	59	2311	84	2012-08-02 05:37:55.449095
27850	3	1-8228	27850	59	74	4736	96	2012-08-02 05:37:55.449095
27854	3	1-5828	27853	12	13	1442	94	2012-08-02 05:37:57.483965
27855	3	1-6738	27853	13	89	1025	82	2012-08-02 05:37:57.483965
27853	3	1-8233	27853	89	12	2855	30	2012-08-02 05:37:57.483965
27857	3	1-7737	27856	33	16	1185	27	2012-08-02 05:37:58.290181
27858	3	1-7635	27856	16	18	1005	51	2012-08-02 05:37:58.290181
27856	3	1-8235	27856	18	33	1139	99	2012-08-02 05:37:58.290181
27860	4	1-5937	27859	60	22	2814	36	2012-08-02 05:37:58.290181
27861	4	1-7516	27859	22	59	5110	59	2012-08-02 05:37:58.290181
27862	4	1-7645	27859	59	18	4172	51	2012-08-02 05:37:58.290181
27859	4	1-8235	27859	18	60	4874	99	2012-08-02 05:37:58.290181
27864	4	1-4709	27863	52	44	305	28	2012-08-02 05:37:58.290181
27865	4	1-8210	27863	44	74	652	64	2012-08-02 05:37:58.290181
27866	4	1-1511	27863	74	18	253	51	2012-08-02 05:37:58.290181
27863	4	1-8235	27863	18	52	318	99	2012-08-02 05:37:58.290181
27868	5	1-6211	27867	75	88	684	73	2012-08-02 05:37:59.097031
27869	5	1-7922	27867	88	73	5028	14	2012-08-02 05:37:59.097031
27870	5	1-8217	27867	73	18	358	47	2012-08-02 05:37:59.097031
27871	5	1-8129	27867	18	32	2359	50	2012-08-02 05:37:59.097031
27867	5	1-8237	27867	32	75	1281	49	2012-08-02 05:37:59.097031
27873	5	1-7946	27872	42	55	1932	57	2012-08-02 05:38:00.29399
27874	5	1-8155	27872	55	12	2170	18	2012-08-02 05:38:00.29399
27875	5	1-3418	27872	12	18	393	47	2012-08-02 05:38:00.29399
27876	5	1-8129	27872	18	85	1998	50	2012-08-02 05:38:00.29399
27872	5	1-8240	27872	85	42	2509	6	2012-08-02 05:38:00.29399
27878	5	1-6355	27877	63	87	633	55	2012-08-02 05:38:00.29399
27879	5	1-7626	27877	87	95	542	74	2012-08-02 05:38:00.29399
27880	5	1-8184	27877	95	18	250	47	2012-08-02 05:38:00.29399
27881	5	1-8129	27877	18	85	1596	50	2012-08-02 05:38:00.29399
27877	5	1-8240	27877	85	63	2513	6	2012-08-02 05:38:00.29399
27883	2	1-8073	27882	62	73	1512	68	2012-08-02 05:38:00.746989
27882	2	1-8241	27882	73	62	5234	12	2012-08-02 05:38:00.746989
27885	4	1-5680	27884	78	12	671	76	2012-08-02 05:38:01.144878
27886	4	1-7642	27884	12	18	469	46	2012-08-02 05:38:01.144878
27887	4	1-7980	27884	18	86	220	75	2012-08-02 05:38:01.144878
27884	4	1-8242	27884	86	78	1187	52	2012-08-02 05:38:01.144878
27889	3	1-7941	27888	53	63	4412	10	2012-08-02 05:38:01.144878
27890	3	1-7394	27888	63	86	1130	75	2012-08-02 05:38:01.144878
27888	3	1-8242	27888	86	53	6746	52	2012-08-02 05:38:01.144878
27892	4	1-7207	27891	92	99	2833	65	2012-08-02 05:38:02.73564
27893	4	1-7841	27891	99	52	2856	78	2012-08-02 05:38:02.73564
27894	4	1-6589	27891	52	7	3796	37	2012-08-02 05:38:02.73564
27891	4	1-8246	27891	7	92	3130	19	2012-08-02 05:38:02.73564
27896	3	1-7459	27895	46	45	1084	55	2012-08-02 05:38:02.73564
27897	3	1-4730	27895	45	7	1360	37	2012-08-02 05:38:02.73564
27895	3	1-8246	27895	7	46	1226	19	2012-08-02 05:38:02.73564
27899	2	1-3513	27898	75	7	75	37	2012-08-02 05:38:02.73564
27898	2	1-8246	27898	7	75	73	19	2012-08-02 05:38:02.73564
27901	2	1-8074	27900	12	29	322	98	2012-08-02 05:38:04.35968
27900	2	1-8250	27900	29	12	2389	33	2012-08-02 05:38:04.35968
27903	4	1-2979	27902	24	7	1105	89	2012-08-02 05:38:05.185079
27904	4	1-7616	27902	7	76	1065	86	2012-08-02 05:38:05.185079
27905	4	1-6933	27902	76	21	1030	43	2012-08-02 05:38:05.185079
27902	4	1-8252	27902	21	24	1562	38	2012-08-02 05:38:05.185079
27907	3	1-7817	27906	2	2	2219	31	2012-08-02 05:38:05.185079
27908	3	1-7958	27906	2	21	2000	43	2012-08-02 05:38:05.185079
27906	3	1-8252	27906	21	2	3060	38	2012-08-02 05:38:05.185079
27910	5	1-7019	27909	2	12	1078	33	2012-08-02 05:38:05.678853
27911	5	1-8074	27909	12	97	1348	98	2012-08-02 05:38:05.678853
27912	5	1-6336	27909	97	18	781	46	2012-08-02 05:38:05.678853
27913	5	1-7980	27909	18	27	637	75	2012-08-02 05:38:05.678853
27909	5	1-8253	27909	27	2	1811	73	2012-08-02 05:38:05.678853
27915	3	1-7510	27914	33	22	447	28	2012-08-02 05:38:06.084245
27916	3	1-6204	27914	22	43	299	57	2012-08-02 05:38:06.084245
27914	3	1-8254	27914	43	33	745	81	2012-08-02 05:38:06.084245
27918	4	1-6372	27917	17	82	2898	92	2012-08-02 05:38:06.084245
27919	4	1-8247	27917	82	22	1038	28	2012-08-02 05:38:06.084245
27920	4	1-6204	27917	22	43	818	57	2012-08-02 05:38:06.084245
27917	4	1-8254	27917	43	17	2407	81	2012-08-02 05:38:06.084245
27922	4	1-7334	27921	92	46	4136	81	2012-08-02 05:38:06.879798
27923	4	1-4184	27921	46	1	1268	44	2012-08-02 05:38:06.879798
27924	4	1-8179	27921	1	22	2991	5	2012-08-02 05:38:06.879798
27921	4	1-8256	27921	22	92	4847	47	2012-08-02 05:38:06.879798
27926	2	1-2216	27925	100	22	1526	5	2012-08-02 05:38:06.879798
27925	2	1-8256	27925	22	100	2594	47	2012-08-02 05:38:06.879798
27928	4	1-5116	27927	57	59	337	84	2012-08-02 05:38:07.686533
27929	4	1-8228	27927	59	2	555	96	2012-08-02 05:38:07.686533
27930	4	1-7823	27927	2	97	378	87	2012-08-02 05:38:07.686533
27927	4	1-8258	27927	97	57	442	61	2012-08-02 05:38:07.686533
27932	5	1-6276	27931	68	32	2739	42	2012-08-02 05:38:07.686533
27933	5	1-7575	27931	32	18	926	82	2012-08-02 05:38:07.686533
27934	5	1-7702	27931	18	95	832	23	2012-08-02 05:38:07.686533
27935	5	1-8024	27931	95	97	2004	87	2012-08-02 05:38:07.686533
27931	5	1-8258	27931	97	68	2905	61	2012-08-02 05:38:07.686533
27937	3	1-717	27936	1	47	565	13	2012-08-02 05:38:08.492792
27938	3	1-8181	27936	47	43	391	54	2012-08-02 05:38:08.492792
27936	3	1-8260	27936	43	1	1624	86	2012-08-02 05:38:08.492792
27940	4	1-4233	27939	74	74	2754	63	2012-08-02 05:38:08.492792
27941	4	1-6802	27939	74	47	930	13	2012-08-02 05:38:08.492792
27942	4	1-8181	27939	47	43	801	54	2012-08-02 05:38:08.492792
27939	4	1-8260	27939	43	74	4143	86	2012-08-02 05:38:08.492792
27944	4	1-6933	27943	76	33	1162	43	2012-08-02 05:38:08.492792
27945	4	1-3657	27943	33	47	395	13	2012-08-02 05:38:08.492792
27946	4	1-8181	27943	47	43	363	54	2012-08-02 05:38:08.492792
27943	4	1-8260	27943	43	76	2001	86	2012-08-02 05:38:08.492792
27948	3	1-4374	27947	52	81	500	16	2012-08-02 05:38:08.492792
27949	3	1-5527	27947	81	43	229	54	2012-08-02 05:38:08.492792
27947	3	1-8260	27947	43	52	1407	86	2012-08-02 05:38:08.492792
27951	3	1-6351	27950	89	47	1	13	2012-08-02 05:38:08.492792
27952	3	1-8181	27950	47	43	1	54	2012-08-02 05:38:08.492792
27950	3	1-8260	27950	43	89	6	86	2012-08-02 05:38:08.492792
27954	4	1-7589	27953	68	88	316	37	2012-08-02 05:38:10.901615
27955	4	1-8040	27953	88	64	473	17	2012-08-02 05:38:10.901615
27956	4	1-7569	27953	64	74	555	15	2012-08-02 05:38:10.901615
27953	4	1-8266	27953	74	68	181	21	2012-08-02 05:38:10.901615
27958	4	1-7469	27957	91	22	1123	94	2012-08-02 05:38:12.09431
27959	4	1-8218	27957	22	32	3698	64	2012-08-02 05:38:12.09431
27960	4	1-7315	27957	32	73	3917	88	2012-08-02 05:38:12.09431
27957	4	1-8269	27957	73	91	3291	43	2012-08-02 05:38:12.09431
27962	3	1-8137	27961	11	52	1099	49	2012-08-02 05:38:12.09431
27963	3	1-7301	27961	52	73	2255	88	2012-08-02 05:38:12.09431
27961	3	1-8269	27961	73	11	1992	43	2012-08-02 05:38:12.09431
27965	4	1-8221	27964	32	27	789	75	2012-08-02 05:38:14.083078
27966	4	1-8253	27964	27	93	1915	73	2012-08-02 05:38:14.083078
27967	4	1-6193	27964	93	31	827	65	2012-08-02 05:38:14.083078
27964	4	1-8274	27964	31	32	1797	15	2012-08-02 05:38:14.083078
27969	4	1-7414	27968	20	100	1306	56	2012-08-02 05:38:14.491732
27970	4	1-2978	27968	100	88	419	73	2012-08-02 05:38:14.491732
27971	4	1-7922	27968	88	99	2210	14	2012-08-02 05:38:14.491732
27968	4	1-8275	27968	99	20	1945	41	2012-08-02 05:38:14.491732
27973	3	1-7403	27972	9	40	2759	16	2012-08-02 05:38:14.491732
27974	3	1-7849	27972	40	99	3521	14	2012-08-02 05:38:14.491732
27972	3	1-8275	27972	99	9	3619	41	2012-08-02 05:38:14.491732
27976	4	1-8079	27975	7	99	4485	33	2012-08-02 05:38:14.878638
27977	4	1-8220	27975	99	33	4719	22	2012-08-02 05:38:14.878638
27978	4	1-6329	27975	33	59	2508	1	2012-08-02 05:38:14.878638
27975	4	1-8276	27975	59	7	3823	19	2012-08-02 05:38:14.878638
27980	4	1-8079	27979	7	47	2351	33	2012-08-02 05:38:14.878638
27981	4	1-7892	27979	47	96	1832	49	2012-08-02 05:38:14.878638
27982	4	1-6024	27979	96	59	1112	1	2012-08-02 05:38:14.878638
27979	4	1-8276	27979	59	7	1843	19	2012-08-02 05:38:14.878638
27984	4	1-7687	27983	69	8	3032	91	2012-08-02 05:38:16.126915
27985	4	1-6033	27983	8	81	1324	42	2012-08-02 05:38:16.126915
27986	4	1-8244	27983	81	11	971	94	2012-08-02 05:38:16.126915
27983	4	1-8279	27983	11	69	3105	66	2012-08-02 05:38:16.126915
27988	4	1-7553	27987	52	57	359	25	2012-08-02 05:38:16.126915
27989	4	1-8100	27987	57	82	1199	78	2012-08-02 05:38:16.126915
27990	4	1-4671	27987	82	11	242	94	2012-08-02 05:38:16.126915
27987	4	1-8279	27987	11	52	846	66	2012-08-02 05:38:16.126915
27992	4	1-7958	27991	2	21	1234	43	2012-08-02 05:38:16.95555
27993	4	1-8252	27991	21	96	1326	38	2012-08-02 05:38:16.95555
27994	4	1-7887	27991	96	55	507	33	2012-08-02 05:38:16.95555
27991	4	1-8281	27991	55	2	1956	31	2012-08-02 05:38:16.95555
27996	5	1-7958	27995	2	23	583	43	2012-08-02 05:38:16.95555
27997	5	1-6491	27995	23	1	485	16	2012-08-02 05:38:16.95555
27998	5	1-7768	27995	1	36	206	32	2012-08-02 05:38:16.95555
27999	5	1-6434	27995	36	55	161	33	2012-08-02 05:38:16.95555
27995	5	1-8281	27995	55	2	758	31	2012-08-02 05:38:16.95555
28001	5	1-7958	28000	2	1	1935	43	2012-08-02 05:38:16.95555
28002	5	1-6318	28000	1	16	983	59	2012-08-02 05:38:16.95555
28003	5	1-8067	28000	16	36	499	32	2012-08-02 05:38:16.95555
28004	5	1-6434	28000	36	55	433	33	2012-08-02 05:38:16.95555
28000	5	1-8281	28000	55	2	2264	31	2012-08-02 05:38:16.95555
28006	5	1-7887	28005	96	12	2514	33	2012-08-02 05:38:17.762139
28007	5	1-8074	28005	12	88	1593	98	2012-08-02 05:38:17.762139
28008	5	1-3773	28005	88	87	168	23	2012-08-02 05:38:17.762139
28009	5	1-8043	28005	87	28	193	52	2012-08-02 05:38:17.762139
28005	5	1-8283	28005	28	96	9519	38	2012-08-02 05:38:17.762139
28011	5	1-8034	28010	99	60	972	7	2012-08-02 05:38:18.579956
28012	5	1-7683	28010	60	68	602	21	2012-08-02 05:38:18.579956
28013	5	1-7589	28010	68	88	1079	37	2012-08-02 05:38:18.579956
28014	5	1-8040	28010	88	42	1660	17	2012-08-02 05:38:18.579956
28010	5	1-8285	28010	42	99	830	18	2012-08-02 05:38:18.579956
28016	5	1-8016	28015	10	39	878	35	2012-08-02 05:38:19.784264
28017	5	1-8089	28015	39	97	1927	56	2012-08-02 05:38:19.784264
28018	5	1-8234	28015	97	55	3186	6	2012-08-02 05:38:19.784264
28019	5	1-7999	28015	55	59	2071	75	2012-08-02 05:38:19.784264
28015	5	1-8288	28015	59	10	2798	85	2012-08-02 05:38:19.784264
28021	3	1-8171	28020	75	48	3183	100	2012-08-02 05:38:20.645622
28022	3	1-5099	28020	48	15	1254	97	2012-08-02 05:38:20.645622
28020	3	1-8290	28020	15	75	777	83	2012-08-02 05:38:20.645622
28024	5	1-8171	28023	75	57	1225	100	2012-08-02 05:38:20.645622
28025	5	1-6877	28023	57	61	784	17	2012-08-02 05:38:20.645622
28026	5	1-7498	28023	61	81	256	76	2012-08-02 05:38:20.645622
28027	5	1-8105	28023	81	15	427	97	2012-08-02 05:38:20.645622
28023	5	1-8290	28023	15	75	281	83	2012-08-02 05:38:20.645622
28029	3	1-8109	28028	77	8	878	36	2012-08-02 05:38:23.487117
28030	3	1-7753	28028	8	31	561	49	2012-08-02 05:38:23.487117
28028	3	1-8297	28028	31	77	1616	21	2012-08-02 05:38:23.487117
28032	2	1-8028	28031	10	77	893	45	2012-08-02 05:38:24.66889
28031	2	1-8300	28031	77	10	1995	51	2012-08-02 05:38:24.66889
28034	5	1-7606	28033	82	62	2606	6	2012-08-02 05:38:25.066624
28035	5	1-3245	28033	62	81	1008	42	2012-08-02 05:38:25.066624
28036	5	1-8244	28033	81	11	583	94	2012-08-02 05:38:25.066624
28037	5	1-8279	28033	11	50	1472	66	2012-08-02 05:38:25.066624
28033	5	1-8301	28033	50	82	5435	46	2012-08-02 05:38:25.066624
28039	2	1-5961	28038	76	50	1268	74	2012-08-02 05:38:27.066168
28038	2	1-8306	28038	50	76	5335	46	2012-08-02 05:38:27.066168
28041	4	1-8299	28040	75	83	1344	1	2012-08-02 05:38:27.066168
28042	4	1-7530	28040	83	5	985	33	2012-08-02 05:38:27.066168
28043	4	1-8174	28040	5	50	656	74	2012-08-02 05:38:27.066168
28040	4	1-8306	28040	50	75	3540	46	2012-08-02 05:38:27.066168
28045	4	1-5722	28044	72	12	990	76	2012-08-02 05:38:27.486033
28046	4	1-7642	28044	12	62	381	46	2012-08-02 05:38:27.486033
28044	4	1-8307	28044	29	72	4078	78	2012-08-02 05:38:27.486033
28047	4	1-6710	28044	62	29	109	85	2012-08-02 05:38:27.486033
28049	3	1-3978	28048	1	39	41	77	2012-08-02 05:38:27.486033
28050	3	1-5471	28048	39	29	16	85	2012-08-02 05:38:27.486033
28048	3	1-8307	28048	29	1	650	78	2012-08-02 05:38:27.486033
28052	3	1-8102	28051	32	67	5695	21	2012-08-02 05:38:27.884026
28053	3	1-4626	28051	67	90	1106	84	2012-08-02 05:38:27.884026
28051	3	1-8308	28051	90	32	5798	3	2012-08-02 05:38:27.884026
28055	5	1-7141	28054	100	36	1466	32	2012-08-02 05:38:27.884026
28056	5	1-6434	28054	36	99	801	33	2012-08-02 05:38:27.884026
28057	5	1-8220	28054	99	26	556	22	2012-08-02 05:38:27.884026
28058	5	1-7210	28054	26	90	400	84	2012-08-02 05:38:27.884026
28054	5	1-8308	28054	90	100	3073	3	2012-08-02 05:38:27.884026
28060	4	1-116	28059	64	80	1766	83	2012-08-02 05:38:28.701418
28061	4	1-8194	28059	80	46	1780	9	2012-08-02 05:38:28.701418
28062	4	1-7157	28059	46	30	564	24	2012-08-02 05:38:28.701418
28059	4	1-8310	28059	30	64	3886	75	2012-08-02 05:38:28.701418
28064	4	1-3501	28063	86	96	815	38	2012-08-02 05:38:28.701418
28065	4	1-7887	28063	96	60	442	33	2012-08-02 05:38:28.701418
28066	4	1-8032	28063	60	30	157	24	2012-08-02 05:38:28.701418
28063	4	1-8310	28063	30	86	1140	75	2012-08-02 05:38:28.701418
28067	1	1-116	\N	64	64	82	83	2012-08-02 05:38:28.701418
28069	4	1-8137	28068	11	31	8	49	2012-08-02 05:38:29.496924
28070	4	1-8297	28068	31	33	26	21	2012-08-02 05:38:29.496924
28071	4	1-1791	28068	33	11	12	46	2012-08-02 05:38:29.496924
28068	4	1-8312	28068	11	11	13	43	2012-08-02 05:38:29.496924
28073	4	1-8023	28072	84	34	3238	35	2012-08-02 05:38:30.292428
28074	4	1-6684	28072	34	71	3620	48	2012-08-02 05:38:30.292428
28075	4	1-7954	28072	71	9	2490	59	2012-08-02 05:38:30.292428
28072	4	1-8314	28072	9	84	3652	96	2012-08-02 05:38:30.292428
28077	4	1-7563	28076	60	61	2757	68	2012-08-02 05:38:30.292428
28078	4	1-7749	28076	61	22	1531	36	2012-08-02 05:38:30.292428
28079	4	1-7516	28076	22	9	2439	59	2012-08-02 05:38:30.292428
28076	4	1-8314	28076	9	60	3839	96	2012-08-02 05:38:30.292428
28081	3	1-8023	28080	84	34	1662	35	2012-08-02 05:38:30.292428
28082	3	1-7788	28080	34	9	1019	59	2012-08-02 05:38:30.292428
28080	3	1-8314	28080	9	84	1674	96	2012-08-02 05:38:30.292428
28084	4	1-7775	28083	34	18	586	51	2012-08-02 05:38:31.496507
28085	4	1-8235	28083	18	51	685	99	2012-08-02 05:38:31.496507
28086	4	1-8042	28083	51	48	1571	85	2012-08-02 05:38:31.496507
28083	4	1-8317	28083	48	34	5353	29	2012-08-02 05:38:31.496507
28088	4	1-8240	28087	85	55	1320	6	2012-08-02 05:38:31.883347
28089	4	1-7999	28087	55	33	600	75	2012-08-02 05:38:31.883347
28090	4	1-7742	28087	33	87	363	46	2012-08-02 05:38:31.883347
28087	4	1-8318	28087	87	85	1137	50	2012-08-02 05:38:31.883347
28092	3	1-914	28091	78	12	361	76	2012-08-02 05:38:31.883347
28093	3	1-7642	28091	12	87	436	46	2012-08-02 05:38:31.883347
28091	3	1-8318	28091	87	78	1870	50	2012-08-02 05:38:31.883347
28095	5	1-7085	28094	24	7	554	96	2012-08-02 05:38:33.485609
28096	5	1-7514	28094	7	90	474	64	2012-08-02 05:38:33.485609
28097	5	1-8149	28094	90	98	461	17	2012-08-02 05:38:33.485609
28098	5	1-7930	28094	98	40	344	26	2012-08-02 05:38:33.485609
28094	5	1-8322	28094	40	24	633	39	2012-08-02 05:38:33.485609
28100	5	1-7085	28099	24	84	131	96	2012-08-02 05:38:33.485609
28101	5	1-8023	28099	84	71	144	35	2012-08-02 05:38:33.485609
28102	5	1-1454	28099	71	98	100	17	2012-08-02 05:38:33.485609
28103	5	1-7930	28099	98	40	77	26	2012-08-02 05:38:33.485609
28099	5	1-8322	28099	40	24	145	39	2012-08-02 05:38:33.485609
28105	5	1-7527	28104	40	59	2581	1	2012-08-02 05:38:33.485609
28106	5	1-8276	28104	59	93	3904	19	2012-08-02 05:38:33.485609
28107	5	1-6139	28104	93	98	1921	17	2012-08-02 05:38:33.485609
28108	5	1-7930	28104	98	40	1418	26	2012-08-02 05:38:33.485609
28104	5	1-8322	28104	40	40	2725	39	2012-08-02 05:38:33.485609
28110	4	1-6927	28109	81	35	768	61	2012-08-02 05:38:33.485609
28111	4	1-2632	28109	35	98	500	17	2012-08-02 05:38:33.485609
28112	4	1-7930	28109	98	40	402	26	2012-08-02 05:38:33.485609
28109	4	1-8322	28109	40	81	795	39	2012-08-02 05:38:33.485609
28114	5	1-6810	28113	7	98	349	57	2012-08-02 05:38:33.485609
28115	5	1-7491	28113	98	76	175	93	2012-08-02 05:38:33.485609
28116	5	1-7089	28113	76	98	168	17	2012-08-02 05:38:33.485609
28117	5	1-7930	28113	98	40	138	26	2012-08-02 05:38:33.485609
28113	5	1-8322	28113	40	7	276	39	2012-08-02 05:38:33.485609
28119	4	1-8034	28118	99	60	2387	7	2012-08-02 05:38:35.926903
28120	4	1-7683	28118	60	68	1152	21	2012-08-02 05:38:35.926903
28121	4	1-7589	28118	68	7	1609	37	2012-08-02 05:38:35.926903
28118	4	1-8328	28118	7	99	2616	18	2012-08-02 05:38:35.926903
28123	2	1-7123	28122	12	7	2176	37	2012-08-02 05:38:35.926903
28122	2	1-8328	28122	7	12	3761	18	2012-08-02 05:38:35.926903
28125	3	1-7998	28124	31	75	2641	1	2012-08-02 05:38:35.926903
28126	3	1-7159	28124	75	7	1743	37	2012-08-02 05:38:35.926903
28124	3	1-8328	28124	7	31	3202	18	2012-08-02 05:38:35.926903
28128	4	1-8035	28127	37	4	1631	68	2012-08-02 05:38:36.324727
28129	4	1-7988	28127	4	8	1579	64	2012-08-02 05:38:36.324727
28130	4	1-7343	28127	8	42	412	27	2012-08-02 05:38:36.324727
28127	4	1-8329	28127	42	37	2185	41	2012-08-02 05:38:36.324727
28132	4	1-6952	28131	60	75	1964	46	2012-08-02 05:38:36.324727
28133	4	1-8299	28131	75	82	713	1	2012-08-02 05:38:36.324727
28134	4	1-7595	28131	82	42	409	27	2012-08-02 05:38:36.324727
28131	4	1-8329	28131	42	60	2186	41	2012-08-02 05:38:36.324727
28136	5	1-4513	28135	14	11	194	43	2012-08-02 05:38:36.324727
28137	5	1-8137	28135	11	31	101	49	2012-08-02 05:38:36.324727
28138	5	1-8297	28135	31	26	313	21	2012-08-02 05:38:36.324727
28139	5	1-5886	28135	26	42	96	27	2012-08-02 05:38:36.324727
28135	5	1-8329	28135	42	14	517	41	2012-08-02 05:38:36.324727
28141	5	1-8035	28140	37	4	1199	68	2012-08-02 05:38:36.324727
28142	5	1-7988	28140	4	79	1229	64	2012-08-02 05:38:36.324727
28143	5	1-3161	28140	79	82	449	1	2012-08-02 05:38:36.324727
28144	5	1-7595	28140	82	42	270	27	2012-08-02 05:38:36.324727
28140	5	1-8329	28140	42	37	1517	41	2012-08-02 05:38:36.324727
28146	3	1-2840	28145	46	75	236	83	2012-08-02 05:38:36.744904
28147	3	1-8171	28145	75	42	874	100	2012-08-02 05:38:36.744904
28145	3	1-8330	28145	42	46	1211	57	2012-08-02 05:38:36.744904
28149	4	1-8254	28148	43	25	5905	81	2012-08-02 05:38:36.744904
28150	4	1-7727	28148	25	16	1120	52	2012-08-02 05:38:36.744904
28151	4	1-8257	28148	16	42	1762	100	2012-08-02 05:38:36.744904
28148	4	1-8330	28148	42	43	2555	57	2012-08-02 05:38:36.744904
28153	3	1-6816	28152	6	28	157	98	2012-08-02 05:38:37.948959
28154	3	1-8193	28152	28	84	84	93	2012-08-02 05:38:37.948959
28152	3	1-8333	28152	84	6	1969	9	2012-08-02 05:38:37.948959
28156	5	1-8206	28155	97	97	433	66	2012-08-02 05:38:38.744386
28157	5	1-5631	28155	97	31	168	18	2012-08-02 05:38:38.744386
28158	5	1-7998	28155	31	82	164	1	2012-08-02 05:38:38.744386
28159	5	1-7595	28155	82	95	107	27	2012-08-02 05:38:38.744386
28155	5	1-8335	28155	95	97	225	32	2012-08-02 05:38:38.744386
28161	5	1-7548	28160	94	32	5225	15	2012-08-02 05:38:39.15324
28162	5	1-8221	28160	32	91	2292	75	2012-08-02 05:38:39.15324
28163	5	1-6924	28160	91	4	1060	68	2012-08-02 05:38:39.15324
28164	5	1-8010	28160	4	28	719	76	2012-08-02 05:38:39.15324
28160	5	1-8336	28160	28	94	3710	22	2012-08-02 05:38:39.15324
28166	4	1-4910	28165	59	30	157	53	2012-08-02 05:38:39.15324
28167	4	1-8199	28165	30	80	116	25	2012-08-02 05:38:39.15324
28168	4	1-5679	28165	80	28	71	76	2012-08-02 05:38:39.15324
28165	4	1-8336	28165	28	59	388	22	2012-08-02 05:38:39.15324
28170	5	1-1851	28169	31	48	737	16	2012-08-02 05:38:39.15324
28171	5	1-4445	28169	48	28	240	62	2012-08-02 05:38:39.15324
28172	5	1-7169	28169	28	4	218	68	2012-08-02 05:38:39.15324
28173	5	1-8010	28169	4	28	168	76	2012-08-02 05:38:39.15324
28169	5	1-8336	28169	28	31	1022	22	2012-08-02 05:38:39.15324
28175	4	1-4910	28174	59	30	1960	53	2012-08-02 05:38:39.15324
28176	4	1-8199	28174	30	57	1622	25	2012-08-02 05:38:39.15324
28177	4	1-2976	28174	57	28	712	76	2012-08-02 05:38:39.15324
28174	4	1-8336	28174	28	59	4340	22	2012-08-02 05:38:39.15324
28178	1	1-2976	\N	57	57	275	76	2012-08-02 05:38:39.15324
28180	3	1-8199	28179	30	11	979	25	2012-08-02 05:38:39.5623
28181	3	1-6932	28179	11	66	963	73	2012-08-02 05:38:39.5623
28179	3	1-8337	28179	66	30	1181	53	2012-08-02 05:38:39.5623
28183	2	1-5256	28182	84	20	54	34	2012-08-02 05:38:39.960031
28182	2	1-8338	28182	20	84	4324	58	2012-08-02 05:38:39.960031
28185	5	1-7765	28184	63	32	1596	3	2012-08-02 05:38:40.358037
28186	5	1-8102	28184	32	58	3758	21	2012-08-02 05:38:40.358037
28187	5	1-7777	28184	58	55	3136	6	2012-08-02 05:38:40.358037
28188	5	1-7999	28184	55	81	2003	75	2012-08-02 05:38:40.358037
28184	5	1-8339	28184	81	63	3293	86	2012-08-02 05:38:40.358037
28190	5	1-7869	28189	14	100	848	47	2012-08-02 05:38:40.77764
28191	5	1-8289	28189	100	76	1754	93	2012-08-02 05:38:40.77764
28192	5	1-7089	28189	76	42	1632	17	2012-08-02 05:38:40.77764
28193	5	1-8285	28189	42	99	861	18	2012-08-02 05:38:40.77764
28189	5	1-8340	28189	99	14	2294	86	2012-08-02 05:38:40.77764
28195	4	1-3517	28194	66	35	1636	73	2012-08-02 05:38:41.175359
28196	4	1-8118	28194	35	73	1486	90	2012-08-02 05:38:41.175359
28197	4	1-7914	28194	73	32	1361	92	2012-08-02 05:38:41.175359
28194	4	1-8341	28194	32	66	6137	85	2012-08-02 05:38:41.175359
28199	4	1-5226	28198	41	35	145	73	2012-08-02 05:38:41.175359
28200	4	1-8118	28198	35	73	140	90	2012-08-02 05:38:41.175359
28201	4	1-7914	28198	73	32	136	92	2012-08-02 05:38:41.175359
28198	4	1-8341	28198	32	41	650	85	2012-08-02 05:38:41.175359
28203	3	1-8043	28202	87	26	422	52	2012-08-02 05:38:45.163423
28204	3	1-5880	28202	26	33	81	86	2012-08-02 05:38:45.163423
28202	3	1-8351	28202	33	87	389	23	2012-08-02 05:38:45.163423
28206	4	1-7935	28205	96	97	4766	77	2012-08-02 05:38:45.163423
28207	4	1-8248	28205	97	7	2814	89	2012-08-02 05:38:45.163423
28208	4	1-7616	28205	7	33	1320	86	2012-08-02 05:38:45.163423
28205	4	1-8351	28205	33	96	6666	23	2012-08-02 05:38:45.163423
28210	3	1-7999	28209	55	81	151	75	2012-08-02 05:38:46.765589
28211	3	1-8339	28209	81	41	245	86	2012-08-02 05:38:46.765589
28209	3	1-8355	28209	41	55	240	6	2012-08-02 05:38:46.765589
28213	4	1-4310	28212	83	21	74	62	2012-08-02 05:38:47.163196
28214	4	1-8183	28212	21	66	43	38	2012-08-02 05:38:47.163196
28215	4	1-8161	28212	66	50	448	29	2012-08-02 05:38:47.163196
28212	4	1-8356	28212	50	83	680	81	2012-08-02 05:38:47.163196
28217	4	1-6372	28216	17	32	413	92	2012-08-02 05:38:47.163196
28218	4	1-8341	28216	32	4	1433	85	2012-08-02 05:38:47.163196
28219	4	1-7265	28216	4	50	176	29	2012-08-02 05:38:47.163196
28216	4	1-8356	28216	50	17	485	81	2012-08-02 05:38:47.163196
28221	4	1-6279	28220	1	83	1194	1	2012-08-02 05:38:47.163196
28222	4	1-7530	28220	83	11	791	33	2012-08-02 05:38:47.163196
28223	4	1-8259	28220	11	50	2016	29	2012-08-02 05:38:47.163196
28220	4	1-8356	28220	50	1	6295	81	2012-08-02 05:38:47.163196
28224	1	1-6279	\N	1	1	486	1	2012-08-02 05:38:47.163196
28226	3	1-1582	28225	100	97	1759	77	2012-08-02 05:38:48.356361
28227	3	1-8248	28225	97	13	2155	89	2012-08-02 05:38:48.356361
28225	3	1-8359	28225	13	100	4812	86	2012-08-02 05:38:48.356361
28229	4	1-8211	28228	42	42	3727	25	2012-08-02 05:38:49.597704
28230	4	1-7926	28228	42	77	2542	41	2012-08-02 05:38:49.597704
28231	4	1-6775	28228	77	65	1151	8	2012-08-02 05:38:49.597704
28228	4	1-8362	28228	65	42	4589	62	2012-08-02 05:38:49.597704
28233	4	1-7169	28232	28	4	407	68	2012-08-02 05:38:49.597704
28234	4	1-7988	28232	4	51	433	64	2012-08-02 05:38:49.597704
28235	4	1-5233	28232	51	65	112	8	2012-08-02 05:38:49.597704
28232	4	1-8362	28232	65	28	476	62	2012-08-02 05:38:49.597704
28237	5	1-6673	28236	13	82	2130	47	2012-08-02 05:38:50.009831
28238	5	1-8303	28236	82	80	1731	10	2012-08-02 05:38:50.009831
28239	5	1-8292	28236	80	46	1424	91	2012-08-02 05:38:50.009831
28240	5	1-8282	28236	46	87	450	18	2012-08-02 05:38:50.009831
28236	5	1-8363	28236	87	13	7753	66	2012-08-02 05:38:50.009831
28242	5	1-7747	28241	49	13	3610	2	2012-08-02 05:38:50.817859
28243	5	1-4801	28241	13	81	435	42	2012-08-02 05:38:50.817859
28244	5	1-8244	28241	81	22	352	94	2012-08-02 05:38:50.817859
28245	5	1-8218	28241	22	57	1369	64	2012-08-02 05:38:50.817859
28241	5	1-8365	28241	57	49	1664	15	2012-08-02 05:38:50.817859
28247	4	1-5834	28246	11	18	2541	51	2012-08-02 05:38:51.637219
28248	4	1-8235	28246	18	96	2174	99	2012-08-02 05:38:51.637219
28249	4	1-3433	28246	96	36	885	83	2012-08-02 05:38:51.637219
28246	4	1-8367	28246	36	11	2892	91	2012-08-02 05:38:51.637219
28251	4	1-4066	28250	9	18	271	51	2012-08-02 05:38:51.637219
28252	4	1-8235	28250	18	96	323	99	2012-08-02 05:38:51.637219
28253	4	1-3433	28250	96	36	183	83	2012-08-02 05:38:51.637219
28250	4	1-8367	28250	36	9	832	91	2012-08-02 05:38:51.637219
28255	5	1-8282	28254	46	31	801	18	2012-08-02 05:38:51.637219
28256	5	1-7998	28254	31	77	789	1	2012-08-02 05:38:51.637219
28257	5	1-7417	28254	77	3	336	62	2012-08-02 05:38:51.637219
28258	5	1-7993	28254	3	36	374	83	2012-08-02 05:38:51.637219
28254	5	1-8367	28254	36	46	1725	91	2012-08-02 05:38:51.637219
28260	5	1-7702	28259	18	5	398	23	2012-08-02 05:38:56.082803
28261	5	1-8360	28259	5	96	953	55	2012-08-02 05:38:56.082803
28262	5	1-4133	28259	96	26	553	11	2012-08-02 05:38:56.082803
28263	5	1-8131	28259	26	23	504	36	2012-08-02 05:38:56.082803
28259	5	1-8378	28259	23	18	454	82	2012-08-02 05:38:56.082803
28265	3	1-8086	28264	11	30	6337	58	2012-08-02 05:38:56.492617
28266	3	1-5481	28264	30	70	1691	63	2012-08-02 05:38:56.492617
28264	3	1-8379	28264	70	11	4598	22	2012-08-02 05:38:56.492617
28268	5	1-7557	28267	97	92	2365	19	2012-08-02 05:38:56.891198
28269	5	1-7207	28267	92	86	2948	65	2012-08-02 05:38:56.891198
28270	5	1-8371	28267	86	4	4818	68	2012-08-02 05:38:56.891198
28271	5	1-8010	28267	4	46	3944	76	2012-08-02 05:38:56.891198
28267	5	1-8380	28267	46	97	3522	23	2012-08-02 05:38:56.891198
28273	5	1-7926	28272	42	37	1236	41	2012-08-02 05:38:57.300928
28274	5	1-8035	28272	37	4	1069	68	2012-08-02 05:38:57.300928
28275	5	1-8010	28272	4	81	849	76	2012-08-02 05:38:57.300928
28276	5	1-8105	28272	81	49	1560	97	2012-08-02 05:38:57.300928
28272	5	1-8381	28272	49	42	1753	25	2012-08-02 05:38:57.300928
28278	4	1-8233	28277	89	76	101	30	2012-08-02 05:38:57.710042
28279	4	1-7008	28277	76	53	129	14	2012-08-02 05:38:57.710042
28280	4	1-7254	28277	53	69	68	95	2012-08-02 05:38:57.710042
28277	4	1-8382	28277	69	89	36	82	2012-08-02 05:38:57.710042
28282	4	1-5830	28281	20	74	2415	80	2012-08-02 05:38:57.710042
28283	4	1-8346	28281	74	90	4427	7	2012-08-02 05:38:57.710042
28284	4	1-7662	28281	90	69	4567	95	2012-08-02 05:38:57.710042
28281	4	1-8382	28281	69	20	2491	82	2012-08-02 05:38:57.710042
28286	2	1-7963	28285	30	14	3226	58	2012-08-02 05:38:58.518521
28285	2	1-8384	28285	14	30	4472	66	2012-08-02 05:38:58.518521
28288	4	1-8218	28287	22	79	472	64	2012-08-02 05:38:59.314532
28289	4	1-3161	28287	79	82	162	1	2012-08-02 05:38:59.314532
28290	4	1-7595	28287	82	48	91	27	2012-08-02 05:38:59.314532
28287	4	1-8386	28287	48	22	148	94	2012-08-02 05:38:59.314532
28292	5	1-8218	28291	22	57	218	64	2012-08-02 05:38:59.314532
28293	5	1-8365	28291	57	74	253	15	2012-08-02 05:38:59.314532
28294	5	1-8266	28291	74	26	89	21	2012-08-02 05:38:59.314532
28295	5	1-5886	28291	26	48	31	27	2012-08-02 05:38:59.314532
28291	5	1-8386	28291	48	22	59	94	2012-08-02 05:38:59.314532
28297	4	1-8372	28296	77	16	3244	38	2012-08-02 05:38:59.745635
28298	4	1-3275	28296	16	69	960	74	2012-08-02 05:38:59.745635
28299	4	1-8229	28296	69	53	513	89	2012-08-02 05:38:59.745635
28296	4	1-8387	28296	53	77	405	12	2012-08-02 05:38:59.745635
28301	4	1-8372	28300	77	67	4798	38	2012-08-02 05:38:59.745635
28302	4	1-6339	28300	67	22	4605	59	2012-08-02 05:38:59.745635
28303	4	1-3236	28300	22	53	528	89	2012-08-02 05:38:59.745635
28300	4	1-8387	28300	53	77	500	12	2012-08-02 05:38:59.745635
28305	4	1-7880	28304	62	100	2025	97	2012-08-02 05:39:00.564219
28306	4	1-3918	28304	100	1	686	44	2012-08-02 05:39:00.564219
28307	4	1-8179	28304	1	14	1771	5	2012-08-02 05:39:00.564219
28304	4	1-8389	28304	14	62	2647	80	2012-08-02 05:39:00.564219
28309	3	1-8346	28308	74	63	3785	7	2012-08-02 05:39:00.564219
28310	3	1-3258	28308	63	14	983	5	2012-08-02 05:39:00.564219
28308	3	1-8389	28308	14	74	1726	80	2012-08-02 05:39:00.564219
28312	3	1-7966	28311	67	83	2691	86	2012-08-02 05:39:01.373807
28313	3	1-1896	28311	83	19	1723	40	2012-08-02 05:39:01.373807
28311	3	1-8391	28311	19	67	3157	62	2012-08-02 05:39:01.373807
28315	3	1-7993	28314	3	17	1632	83	2012-08-02 05:39:01.373807
28316	3	1-6889	28314	17	19	859	40	2012-08-02 05:39:01.373807
28314	3	1-8391	28314	19	3	1912	62	2012-08-02 05:39:01.373807
28318	4	1-8330	28317	42	74	965	57	2012-08-02 05:39:02.556681
28319	4	1-7821	28317	74	61	788	43	2012-08-02 05:39:02.556681
28320	4	1-3242	28317	61	35	276	51	2012-08-02 05:39:02.556681
28317	4	1-8394	28317	35	42	788	100	2012-08-02 05:39:02.556681
28322	4	1-5269	28321	22	39	1229	35	2012-08-02 05:39:03.363788
28323	4	1-8089	28321	39	97	1789	56	2012-08-02 05:39:03.363788
28324	4	1-8234	28321	97	50	1960	6	2012-08-02 05:39:03.363788
28321	4	1-8396	28321	50	22	7987	93	2012-08-02 05:39:03.363788
28326	4	1-8380	28325	46	5	1978	23	2012-08-02 05:39:04.634976
28327	4	1-8360	28325	5	89	4454	55	2012-08-02 05:39:04.634976
28328	4	1-3643	28325	89	68	2263	73	2012-08-02 05:39:04.634976
28325	4	1-8399	28325	68	46	2465	76	2012-08-02 05:39:04.634976
28330	4	1-8164	28329	88	9	976	60	2012-08-02 05:39:04.634976
28331	4	1-7952	28329	9	100	730	56	2012-08-02 05:39:04.634976
28332	4	1-2978	28329	100	68	327	73	2012-08-02 05:39:04.634976
28329	4	1-8399	28329	68	88	392	76	2012-08-02 05:39:04.634976
28334	5	1-7642	28333	12	82	571	46	2012-08-02 05:39:04.634976
28335	5	1-7606	28333	82	61	384	6	2012-08-02 05:39:04.634976
28336	5	1-8280	28333	61	46	360	98	2012-08-02 05:39:04.634976
28337	5	1-7796	28333	46	68	363	73	2012-08-02 05:39:04.634976
28333	5	1-8399	28333	68	12	448	76	2012-08-02 05:39:04.634976
28339	4	1-5669	28338	83	85	1982	74	2012-08-02 05:39:05.456052
28340	4	1-7864	28338	85	93	846	72	2012-08-02 05:39:05.456052
28341	4	1-7868	28338	93	88	399	55	2012-08-02 05:39:05.456052
28338	4	1-8401	28338	88	83	3844	60	2012-08-02 05:39:05.456052
28343	4	1-5762	28342	74	85	2395	74	2012-08-02 05:39:05.456052
28344	4	1-7864	28342	85	93	1051	72	2012-08-02 05:39:05.456052
28345	4	1-7868	28342	93	88	510	55	2012-08-02 05:39:05.456052
28342	4	1-8401	28342	88	74	5057	60	2012-08-02 05:39:05.456052
28347	4	1-7331	28346	40	75	335	15	2012-08-02 05:39:05.456052
28348	4	1-6726	28346	75	61	121	16	2012-08-02 05:39:05.456052
28349	4	1-7833	28346	61	88	75	55	2012-08-02 05:39:05.456052
28346	4	1-8401	28346	88	40	749	60	2012-08-02 05:39:05.456052
28351	4	1-8340	28350	99	41	2794	86	2012-08-02 05:39:06.666062
28352	4	1-8355	28350	41	42	2483	6	2012-08-02 05:39:06.666062
28353	4	1-7946	28350	42	8	2217	57	2012-08-02 05:39:06.666062
28350	4	1-8404	28350	8	99	1152	18	2012-08-02 05:39:06.666062
28355	4	1-5767	28354	51	35	2617	51	2012-08-02 05:39:07.484099
28356	4	1-8394	28354	35	44	7096	100	2012-08-02 05:39:07.484099
28357	4	1-4686	28354	44	97	1044	31	2012-08-02 05:39:07.484099
28354	4	1-8406	28354	97	51	4666	71	2012-08-02 05:39:07.484099
28359	4	1-7614	28358	38	11	2575	53	2012-08-02 05:39:07.484099
28360	4	1-7289	28358	11	46	1218	2	2012-08-02 05:39:07.484099
28361	4	1-8334	28358	46	97	469	31	2012-08-02 05:39:07.484099
28358	4	1-8406	28358	97	38	2474	71	2012-08-02 05:39:07.484099
28363	5	1-8386	28362	48	19	1881	94	2012-08-02 05:39:09.053477
28364	5	1-6150	28362	19	19	774	59	2012-08-02 05:39:09.053477
28365	5	1-8238	28362	19	14	766	5	2012-08-02 05:39:09.053477
28366	5	1-8389	28362	14	100	656	80	2012-08-02 05:39:09.053477
28362	5	1-8410	28362	100	48	2086	27	2012-08-02 05:39:09.053477
28368	4	1-8335	28367	95	32	5511	32	2012-08-02 05:39:09.053477
28369	4	1-4604	28367	32	14	1979	5	2012-08-02 05:39:09.053477
28370	4	1-8389	28367	14	100	1701	80	2012-08-02 05:39:09.053477
28367	4	1-8410	28367	100	95	5433	27	2012-08-02 05:39:09.053477
28372	4	1-8122	28371	14	29	390	1	2012-08-02 05:39:09.053477
28373	4	1-7640	28371	29	14	178	5	2012-08-02 05:39:09.053477
28374	4	1-8389	28371	14	100	240	80	2012-08-02 05:39:09.053477
28371	4	1-8410	28371	100	14	863	27	2012-08-02 05:39:09.053477
28376	3	1-777	28375	52	19	1492	59	2012-08-02 05:39:09.474121
28377	3	1-8238	28375	19	23	2093	5	2012-08-02 05:39:09.474121
28375	3	1-8411	28375	23	52	2074	75	2012-08-02 05:39:09.474121
28379	5	1-8366	28378	42	22	568	94	2012-08-02 05:39:09.861087
28380	5	1-8218	28378	22	51	2200	64	2012-08-02 05:39:09.861087
28381	5	1-5233	28378	51	87	628	8	2012-08-02 05:39:09.861087
28382	5	1-7155	28378	87	53	478	11	2012-08-02 05:39:09.861087
28378	5	1-8412	28378	53	42	446	82	2012-08-02 05:39:09.861087
28384	3	1-7768	28383	1	63	330	32	2012-08-02 05:39:10.258739
28385	3	1-7951	28383	63	56	487	88	2012-08-02 05:39:10.258739
28383	3	1-8413	28383	56	1	1330	16	2012-08-02 05:39:10.258739
28387	5	1-7988	28386	4	84	1213	64	2012-08-02 05:39:11.165145
28388	5	1-5373	28386	84	19	671	59	2012-08-02 05:39:11.165145
28389	5	1-8238	28386	19	14	919	5	2012-08-02 05:39:11.165145
28390	5	1-8389	28386	14	28	1616	80	2012-08-02 05:39:11.165145
28386	5	1-8415	28386	28	4	1106	68	2012-08-02 05:39:11.165145
28392	4	1-6203	28391	28	97	848	87	2012-08-02 05:39:11.960578
28393	4	1-8258	28391	97	57	957	61	2012-08-02 05:39:11.960578
28394	4	1-5116	28391	57	88	704	84	2012-08-02 05:39:11.960578
28391	4	1-8417	28391	88	28	1537	62	2012-08-02 05:39:11.960578
28396	2	1-6948	28395	92	88	956	84	2012-08-02 05:39:11.960578
28395	2	1-8417	28395	88	92	2170	62	2012-08-02 05:39:11.960578
28398	5	1-8183	28397	21	96	2383	38	2012-08-02 05:39:11.960578
28399	5	1-7887	28397	96	99	1020	33	2012-08-02 05:39:11.960578
28400	5	1-8220	28397	99	26	930	22	2012-08-02 05:39:11.960578
28401	5	1-7210	28397	26	88	879	84	2012-08-02 05:39:11.960578
28397	5	1-8417	28397	88	21	2003	62	2012-08-02 05:39:11.960578
28403	5	1-8183	28402	21	85	3003	38	2012-08-02 05:39:11.960578
28404	5	1-2873	28402	85	100	1979	5	2012-08-02 05:39:11.960578
28405	5	1-8188	28402	100	26	860	22	2012-08-02 05:39:11.960578
28406	5	1-7210	28402	26	88	901	84	2012-08-02 05:39:11.960578
28402	5	1-8417	28402	88	21	2276	62	2012-08-02 05:39:11.960578
28408	4	1-7993	28407	3	48	1245	83	2012-08-02 05:39:11.960578
28409	4	1-7990	28407	48	65	749	89	2012-08-02 05:39:11.960578
28410	4	1-8170	28407	65	88	405	84	2012-08-02 05:39:11.960578
28407	4	1-8417	28407	88	3	1122	62	2012-08-02 05:39:11.960578
28412	5	1-5676	28411	84	83	185	81	2012-08-02 05:39:14.745423
28413	5	1-4310	28411	83	3	50	62	2012-08-02 05:39:14.745423
28414	5	1-7993	28411	3	36	54	83	2012-08-02 05:39:14.745423
28415	5	1-8367	28411	36	57	248	91	2012-08-02 05:39:14.745423
28411	5	1-8424	28411	57	84	232	34	2012-08-02 05:39:14.745423
28417	2	1-8132	28416	10	57	3527	91	2012-08-02 05:39:14.745423
28416	2	1-8424	28416	57	10	3404	34	2012-08-02 05:39:14.745423
28419	5	1-7001	28418	45	14	551	86	2012-08-02 05:39:14.745423
28420	5	1-7869	28418	14	48	213	47	2012-08-02 05:39:14.745423
28421	5	1-7305	28418	48	36	246	83	2012-08-02 05:39:14.745423
28422	5	1-8367	28418	36	57	1191	91	2012-08-02 05:39:14.745423
28418	5	1-8424	28418	57	45	1182	34	2012-08-02 05:39:14.745423
28424	5	1-8294	28423	63	37	2187	77	2012-08-02 05:39:16.336759
28425	5	1-6516	28423	37	22	1114	28	2012-08-02 05:39:16.336759
28426	5	1-6204	28423	22	8	557	57	2012-08-02 05:39:16.336759
28427	5	1-8404	28423	8	86	206	18	2012-08-02 05:39:16.336759
28423	5	1-8428	28423	86	63	2184	49	2012-08-02 05:39:16.336759
28429	5	1-7301	28428	52	83	828	88	2012-08-02 05:39:16.336759
28430	5	1-6755	28428	83	6	109	8	2012-08-02 05:39:16.336759
28431	5	1-3486	28428	6	8	103	57	2012-08-02 05:39:16.336759
28432	5	1-8404	28428	8	86	43	18	2012-08-02 05:39:16.336759
28428	5	1-8428	28428	86	52	513	49	2012-08-02 05:39:16.336759
28434	4	1-6211	28433	75	6	661	73	2012-08-02 05:39:16.336759
28435	4	1-6910	28433	6	46	155	91	2012-08-02 05:39:16.336759
28436	4	1-8282	28433	46	86	74	18	2012-08-02 05:39:16.336759
28433	4	1-8428	28433	86	75	1232	49	2012-08-02 05:39:16.336759
28438	4	1-8372	28437	77	32	1727	38	2012-08-02 05:39:17.541672
28439	4	1-7279	28437	32	53	800	45	2012-08-02 05:39:17.541672
28440	4	1-7839	28437	53	79	478	6	2012-08-02 05:39:17.541672
28437	4	1-8431	28437	79	77	171	12	2012-08-02 05:39:17.541672
28442	2	1-8233	28441	89	51	4834	30	2012-08-02 05:39:17.939805
28441	2	1-8432	28441	51	89	1608	82	2012-08-02 05:39:17.939805
28444	4	1-7712	28443	14	49	2312	97	2012-08-02 05:39:19.133087
28445	4	1-8381	28443	49	84	2539	25	2012-08-02 05:39:19.133087
28446	4	1-7760	28443	84	70	1281	43	2012-08-02 05:39:19.133087
28443	4	1-8435	28443	70	14	2040	98	2012-08-02 05:39:19.133087
28448	3	1-6336	28447	97	11	2425	46	2012-08-02 05:39:19.133087
28449	3	1-8312	28447	11	70	2644	43	2012-08-02 05:39:19.133087
28447	3	1-8435	28447	70	97	4246	98	2012-08-02 05:39:19.133087
28451	4	1-1308	28450	52	32	2433	90	2012-08-02 05:39:20.359011
28452	4	1-7860	28450	32	63	2874	16	2012-08-02 05:39:20.359011
28453	4	1-8099	28450	63	19	2085	20	2012-08-02 05:39:20.359011
28450	4	1-8438	28450	19	52	4504	30	2012-08-02 05:39:20.359011
28455	3	1-7913	28454	99	60	3497	34	2012-08-02 05:39:21.950517
28456	3	1-5953	28454	60	15	1545	44	2012-08-02 05:39:21.950517
28454	3	1-8442	28454	15	99	1831	55	2012-08-02 05:39:21.950517
28458	5	1-5794	28457	93	11	1359	46	2012-08-02 05:39:21.950517
28459	5	1-8312	28457	11	53	1572	43	2012-08-02 05:39:21.950517
28460	5	1-5515	28457	53	74	722	62	2012-08-02 05:39:21.950517
28461	5	1-7973	28457	74	15	928	44	2012-08-02 05:39:21.950517
28457	5	1-8442	28457	15	93	1156	55	2012-08-02 05:39:21.950517
28463	3	1-8221	28462	32	28	1323	75	2012-08-02 05:39:22.348601
28464	3	1-8400	28462	28	32	3708	29	2012-08-02 05:39:22.348601
28462	3	1-8443	28462	32	32	3441	15	2012-08-02 05:39:22.348601
28466	3	1-8429	28465	55	97	1893	54	2012-08-02 05:39:24.328025
28467	3	1-4463	28465	97	74	265	17	2012-08-02 05:39:24.328025
28465	3	1-8448	28465	74	55	7075	31	2012-08-02 05:39:24.328025
28469	2	1-2894	28468	6	41	1266	42	2012-08-02 05:39:25.12333
28468	2	1-8450	28468	41	6	7480	47	2012-08-02 05:39:25.12333
28471	4	1-8320	28470	12	71	2024	67	2012-08-02 05:39:25.521251
28472	4	1-1625	28470	71	98	566	10	2012-08-02 05:39:25.521251
28473	4	1-8342	28470	98	82	797	13	2012-08-02 05:39:25.521251
28470	4	1-8451	28470	82	12	1070	76	2012-08-02 05:39:25.521251
28475	3	1-8105	28474	81	96	1327	97	2012-08-02 05:39:25.521251
28476	3	1-4356	28474	96	82	542	13	2012-08-02 05:39:25.521251
28474	3	1-8451	28474	82	81	880	76	2012-08-02 05:39:25.521251
28478	4	1-8323	28477	45	51	1681	26	2012-08-02 05:39:25.521251
28479	4	1-7061	28477	51	98	706	10	2012-08-02 05:39:25.521251
28480	4	1-8342	28477	98	82	1226	13	2012-08-02 05:39:25.521251
28477	4	1-8451	28477	82	45	2031	76	2012-08-02 05:39:25.521251
28482	3	1-8164	28481	88	25	721	60	2012-08-02 05:39:25.521251
28483	3	1-1375	28481	25	82	180	13	2012-08-02 05:39:25.521251
28481	3	1-8451	28481	82	88	325	76	2012-08-02 05:39:25.521251
28485	3	1-8206	28484	97	81	3853	66	2012-08-02 05:39:26.737139
28486	3	1-3734	28484	81	80	2175	37	2012-08-02 05:39:26.737139
28484	3	1-8454	28484	80	97	2311	32	2012-08-02 05:39:26.737139
28488	4	1-6434	28487	36	12	459	33	2012-08-02 05:39:26.737139
28489	4	1-8074	28487	12	99	548	98	2012-08-02 05:39:26.737139
28490	4	1-7989	28487	99	80	479	37	2012-08-02 05:39:26.737139
28487	4	1-8454	28487	80	36	550	32	2012-08-02 05:39:26.737139
28492	4	1-4239	28491	69	80	631	10	2012-08-02 05:39:26.737139
28493	4	1-8292	28491	80	65	744	91	2012-08-02 05:39:26.737139
28494	4	1-7835	28491	65	80	847	37	2012-08-02 05:39:26.737139
28491	4	1-8454	28491	80	69	994	32	2012-08-02 05:39:26.737139
28495	1	1-6434	\N	36	36	4219	33	2012-08-02 05:39:26.737139
28497	5	1-6790	28496	34	28	5479	80	2012-08-02 05:39:27.146224
28498	5	1-8415	28496	28	61	2823	68	2012-08-02 05:39:27.146224
28499	5	1-7749	28496	61	23	1362	36	2012-08-02 05:39:27.146224
28500	5	1-8378	28496	23	92	918	82	2012-08-02 05:39:27.146224
28496	5	1-8455	28496	92	34	6616	26	2012-08-02 05:39:27.146224
28502	4	1-8367	28501	36	5	1524	91	2012-08-02 05:39:28.350353
28503	4	1-6680	28501	5	63	668	10	2012-08-02 05:39:28.350353
28504	4	1-7394	28501	63	3	223	75	2012-08-02 05:39:28.350353
28501	4	1-8458	28501	3	36	397	83	2012-08-02 05:39:28.350353
28506	4	1-8262	28505	20	49	1153	97	2012-08-02 05:39:28.350353
28507	4	1-4865	28505	49	32	1071	15	2012-08-02 05:39:28.350353
28508	4	1-8221	28505	32	3	559	75	2012-08-02 05:39:28.350353
28505	4	1-8458	28505	3	20	1218	83	2012-08-02 05:39:28.350353
28510	4	1-6116	28509	24	3	408	62	2012-08-02 05:39:31.543138
28511	4	1-7993	28509	3	48	447	83	2012-08-02 05:39:31.543138
28512	4	1-7990	28509	48	88	265	89	2012-08-02 05:39:31.543138
28509	4	1-8466	28509	88	24	590	82	2012-08-02 05:39:31.543138
28514	4	1-8366	28513	42	53	6352	94	2012-08-02 05:39:31.543138
28515	4	1-7044	28513	53	69	3557	74	2012-08-02 05:39:31.543138
28516	4	1-8229	28513	69	88	2306	89	2012-08-02 05:39:31.543138
28513	4	1-8466	28513	88	42	5208	82	2012-08-02 05:39:31.543138
28518	5	1-3233	28517	65	80	302	37	2012-08-02 05:39:32.72572
28519	5	1-8454	28517	80	66	319	32	2012-08-02 05:39:32.72572
28520	5	1-8076	28517	66	84	157	96	2012-08-02 05:39:32.72572
28521	5	1-8023	28517	84	32	155	35	2012-08-02 05:39:32.72572
28517	5	1-8469	28517	32	65	391	8	2012-08-02 05:39:32.72572
28523	5	1-3486	28522	6	87	953	57	2012-08-02 05:39:32.72572
28524	5	1-8422	28522	87	24	329	3	2012-08-02 05:39:32.72572
28525	5	1-8421	28522	24	84	283	96	2012-08-02 05:39:32.72572
28526	5	1-8023	28522	84	32	296	35	2012-08-02 05:39:32.72572
28522	5	1-8469	28522	32	6	789	8	2012-08-02 05:39:32.72572
28528	5	1-6519	28527	23	61	570	6	2012-08-02 05:39:32.72572
28529	5	1-8280	28527	61	28	492	98	2012-08-02 05:39:32.72572
28530	5	1-8193	28527	28	22	1118	93	2012-08-02 05:39:32.72572
28531	5	1-5269	28527	22	32	241	35	2012-08-02 05:39:32.72572
28527	5	1-8469	28527	32	23	666	8	2012-08-02 05:39:32.72572
28532	1	1-6519	\N	23	23	1476	6	2012-08-02 05:39:32.72572
28534	5	1-7833	28533	61	96	738	55	2012-08-02 05:39:34.339286
28535	5	1-4133	28533	96	78	349	11	2012-08-02 05:39:34.339286
28536	5	1-7907	28533	78	28	217	75	2012-08-02 05:39:34.339286
28537	5	1-8400	28533	28	15	915	29	2012-08-02 05:39:34.339286
28533	5	1-8473	28533	15	61	1197	16	2012-08-02 05:39:34.339286
28539	4	1-7833	28538	61	57	3018	55	2012-08-02 05:39:34.339286
28540	4	1-3317	28538	57	11	1362	33	2012-08-02 05:39:34.339286
28541	4	1-8259	28538	11	15	3500	29	2012-08-02 05:39:34.339286
28538	4	1-8473	28538	15	61	4733	16	2012-08-02 05:39:34.339286
28543	4	1-8342	28542	98	47	6144	13	2012-08-02 05:39:35.14557
28544	4	1-8181	28542	47	51	7045	54	2012-08-02 05:39:35.14557
28545	4	1-7194	28542	51	52	2004	51	2012-08-02 05:39:35.14557
28542	4	1-8475	28542	52	98	3813	10	2012-08-02 05:39:35.14557
28547	3	1-8292	28546	80	9	6028	91	2012-08-02 05:39:35.14557
28548	3	1-4066	28546	9	52	2009	51	2012-08-02 05:39:35.14557
28546	3	1-8475	28546	52	80	4959	10	2012-08-02 05:39:35.14557
28550	3	1-8258	28549	97	76	2468	61	2012-08-02 05:39:37.896496
28551	3	1-5340	28549	76	76	1857	50	2012-08-02 05:39:37.896496
28549	3	1-8482	28549	76	97	2316	87	2012-08-02 05:39:37.896496
28553	4	1-8258	28552	97	19	2047	61	2012-08-02 05:39:37.896496
28554	4	1-7824	28552	19	58	3230	97	2012-08-02 05:39:37.896496
28555	4	1-4223	28552	58	76	918	50	2012-08-02 05:39:37.896496
28552	4	1-8482	28552	76	97	1360	87	2012-08-02 05:39:37.896496
28557	3	1-6460	28556	57	49	3777	4	2012-08-02 05:39:38.305593
28558	3	1-8133	28556	49	19	6355	40	2012-08-02 05:39:38.305593
28556	3	1-8483	28556	19	57	5986	76	2012-08-02 05:39:38.305593
28560	3	1-8114	28559	70	93	17	69	2012-08-02 05:39:38.305593
28561	3	1-5954	28559	93	19	7	40	2012-08-02 05:39:38.305593
28559	3	1-8483	28559	19	70	7	76	2012-08-02 05:39:38.305593
28563	4	1-1668	28562	86	19	109	20	2012-08-02 05:39:39.090216
28564	4	1-8438	28562	19	69	248	30	2012-08-02 05:39:39.090216
28565	4	1-8209	28562	69	98	184	41	2012-08-02 05:39:39.090216
28562	4	1-8485	28562	98	86	931	48	2012-08-02 05:39:39.090216
28567	2	1-4609	28566	5	98	349	41	2012-08-02 05:39:39.090216
28566	2	1-8485	28566	98	5	1779	48	2012-08-02 05:39:39.090216
28569	5	1-8380	28568	46	96	33	23	2012-08-02 05:39:41.520659
28570	5	1-7935	28568	96	50	32	77	2012-08-02 05:39:41.520659
28571	5	1-7377	28568	50	9	15	22	2012-08-02 05:39:41.520659
28572	5	1-7944	28568	9	88	11	84	2012-08-02 05:39:41.520659
28568	5	1-8491	28568	88	46	57	76	2012-08-02 05:39:41.520659
28574	4	1-8440	28573	100	77	2289	30	2012-08-02 05:39:41.520659
28575	4	1-6973	28573	77	65	1374	89	2012-08-02 05:39:41.520659
28576	4	1-8170	28573	65	88	526	84	2012-08-02 05:39:41.520659
28573	4	1-8491	28573	88	100	2874	76	2012-08-02 05:39:41.520659
28578	4	1-8440	28577	100	12	968	30	2012-08-02 05:39:41.520659
28579	4	1-5828	28577	12	33	492	94	2012-08-02 05:39:41.520659
28580	4	1-6222	28577	33	88	174	84	2012-08-02 05:39:41.520659
28577	4	1-8491	28577	88	100	1074	76	2012-08-02 05:39:41.520659
28582	3	1-8225	28581	74	40	1505	19	2012-08-02 05:39:42.327512
28583	3	1-5360	28581	40	24	150	39	2012-08-02 05:39:42.327512
28581	3	1-8493	28581	24	74	8933	96	2012-08-02 05:39:42.327512
28585	4	1-6961	28584	59	97	817	87	2012-08-02 05:39:45.531611
28586	4	1-8258	28584	97	51	1110	61	2012-08-02 05:39:45.531611
28587	4	1-7890	28584	51	84	847	79	2012-08-02 05:39:45.531611
28584	4	1-8501	28584	84	59	1028	19	2012-08-02 05:39:45.531611
28589	4	1-8292	28588	80	27	429	91	2012-08-02 05:39:47.133346
28590	4	1-595	28588	27	4	230	58	2012-08-02 05:39:47.133346
28591	4	1-7911	28588	4	46	238	53	2012-08-02 05:39:47.133346
28588	4	1-8505	28588	46	80	364	10	2012-08-02 05:39:47.133346
28593	3	1-7475	28592	54	9	991	56	2012-08-02 05:39:48.339886
28594	3	1-2583	28592	9	85	195	50	2012-08-02 05:39:48.339886
28592	3	1-8508	28592	85	54	4343	55	2012-08-02 05:39:48.339886
28596	5	1-8224	28595	3	86	1422	65	2012-08-02 05:39:48.748607
28597	5	1-8371	28595	86	4	1741	68	2012-08-02 05:39:48.748607
28598	5	1-8010	28595	4	88	1068	76	2012-08-02 05:39:48.748607
28599	5	1-8164	28595	88	88	2346	60	2012-08-02 05:39:48.748607
28595	5	1-8509	28595	88	3	1576	27	2012-08-02 05:39:48.748607
28601	5	1-8335	28600	95	20	2880	32	2012-08-02 05:39:48.748607
28602	5	1-5852	28600	20	4	1252	68	2012-08-02 05:39:48.748607
28603	5	1-8010	28600	4	88	889	76	2012-08-02 05:39:48.748607
28604	5	1-8164	28600	88	88	2100	60	2012-08-02 05:39:48.748607
28600	5	1-8509	28600	88	95	1518	27	2012-08-02 05:39:48.748607
28606	5	1-8224	28605	3	99	2221	65	2012-08-02 05:39:48.748607
28607	5	1-7841	28605	99	22	2931	78	2012-08-02 05:39:48.748607
28608	5	1-6059	28605	22	88	1037	76	2012-08-02 05:39:48.748607
28609	5	1-8164	28605	88	88	2566	60	2012-08-02 05:39:48.748607
28605	5	1-8509	28605	88	3	1941	27	2012-08-02 05:39:48.748607
28611	4	1-8234	28610	97	61	1639	6	2012-08-02 05:39:51.134954
28612	4	1-8280	28610	61	99	1418	98	2012-08-02 05:39:51.134954
28613	4	1-7989	28610	99	59	1254	37	2012-08-02 05:39:51.134954
28610	4	1-8515	28610	59	97	1066	56	2012-08-02 05:39:51.134954
28615	4	1-3959	28614	22	47	7	13	2012-08-02 05:39:52.692544
28616	4	1-8181	28614	47	67	8	54	2012-08-02 05:39:52.692544
28617	4	1-8205	28614	67	58	3	91	2012-08-02 05:39:52.692544
28614	4	1-8519	28614	58	22	17	100	2012-08-02 05:39:52.692544
28619	5	1-6306	28618	100	55	1677	85	2012-08-02 05:39:52.692544
28620	5	1-8478	28618	55	67	2163	21	2012-08-02 05:39:52.692544
28621	5	1-6143	28618	67	70	1165	42	2012-08-02 05:39:52.692544
28622	5	1-8368	28618	70	58	972	91	2012-08-02 05:39:52.692544
28618	5	1-8519	28618	58	100	6550	100	2012-08-02 05:39:52.692544
28624	4	1-8280	28623	61	28	427	98	2012-08-02 05:39:53.090473
28625	4	1-8193	28623	28	39	971	93	2012-08-02 05:39:53.090473
28626	4	1-8231	28623	39	83	529	97	2012-08-02 05:39:53.090473
28623	4	1-8520	28623	83	61	494	6	2012-08-02 05:39:53.090473
28628	3	1-8280	28627	61	14	2940	98	2012-08-02 05:39:53.090473
28629	3	1-7712	28627	14	83	3407	97	2012-08-02 05:39:53.090473
28627	3	1-8520	28627	83	61	3290	6	2012-08-02 05:39:53.090473
28631	3	1-7796	28630	46	62	378	73	2012-08-02 05:39:53.499308
28632	3	1-7419	28630	62	6	154	2	2012-08-02 05:39:53.499308
28630	3	1-8521	28630	6	46	4129	98	2012-08-02 05:39:53.499308
28634	3	1-6267	28633	99	46	3319	53	2012-08-02 05:39:54.33926
28635	3	1-8505	28633	46	18	3720	10	2012-08-02 05:39:54.33926
28633	3	1-8523	28633	18	99	3353	3	2012-08-02 05:39:54.33926
28637	4	1-8421	28636	24	35	266	96	2012-08-02 05:39:54.33926
28638	4	1-2974	28636	35	52	143	51	2012-08-02 05:39:54.33926
28639	4	1-8475	28636	52	18	304	10	2012-08-02 05:39:54.33926
28636	4	1-8523	28636	18	24	332	3	2012-08-02 05:39:54.33926
28641	4	1-7568	28640	77	55	2600	31	2012-08-02 05:39:54.33926
28642	4	1-8429	28640	55	90	1133	54	2012-08-02 05:39:54.33926
28643	4	1-8490	28640	90	18	1132	10	2012-08-02 05:39:54.33926
28640	4	1-8523	28640	18	77	1315	3	2012-08-02 05:39:54.33926
28645	4	1-1508	28644	31	66	810	73	2012-08-02 05:39:54.33926
28646	4	1-8337	28644	66	46	1024	53	2012-08-02 05:39:54.33926
28647	4	1-8505	28644	46	18	1667	10	2012-08-02 05:39:54.33926
28644	4	1-8523	28644	18	31	2182	3	2012-08-02 05:39:54.33926
28649	4	1-8454	28648	80	97	66	32	2012-08-02 05:39:55.1457
28650	4	1-5686	28648	97	38	28	38	2012-08-02 05:39:55.1457
28651	4	1-4144	28648	38	82	23	80	2012-08-02 05:39:55.1457
28648	4	1-8525	28648	82	80	59	37	2012-08-02 05:39:55.1457
28653	3	1-7607	28652	97	57	758	98	2012-08-02 05:39:55.1457
28654	3	1-6911	28652	57	82	686	80	2012-08-02 05:39:55.1457
28652	3	1-8525	28652	82	97	1812	37	2012-08-02 05:39:55.1457
28656	4	1-8515	28655	59	41	10	56	2012-08-02 05:39:55.1457
28657	4	1-5212	28655	41	93	13	51	2012-08-02 05:39:55.1457
28658	4	1-7537	28655	93	82	4	80	2012-08-02 05:39:55.1457
28655	4	1-8525	28655	82	59	11	37	2012-08-02 05:39:55.1457
28660	5	1-8274	28659	31	8	7486	15	2012-08-02 05:39:59.122697
28661	5	1-8434	28659	8	35	4764	83	2012-08-02 05:39:59.122697
28662	5	1-7682	28659	35	18	1260	24	2012-08-02 05:39:59.122697
28663	5	1-5591	28659	18	88	1668	50	2012-08-02 05:39:59.122697
28659	5	1-8535	28659	88	31	2989	65	2012-08-02 05:39:59.122697
28665	4	1-7841	28664	99	28	173	78	2012-08-02 05:39:59.122697
28666	4	1-8376	28664	28	31	149	70	2012-08-02 05:39:59.122697
28667	4	1-3661	28664	31	88	66	50	2012-08-02 05:39:59.122697
28664	4	1-8535	28664	88	99	124	65	2012-08-02 05:39:59.122697
28669	5	1-8038	28668	77	39	601	73	2012-08-02 05:39:59.54285
28668	5	1-8536	28668	48	77	2947	48	2012-08-02 05:39:59.54285
28670	5	1-8502	28668	39	81	1119	42	2012-08-02 05:39:59.54285
28671	5	1-8244	28668	81	11	889	94	2012-08-02 05:39:59.54285
28672	5	1-8279	28668	11	48	3083	66	2012-08-02 05:39:59.54285
28674	4	1-8509	28673	88	20	1390	27	2012-08-02 05:40:01.951158
28675	4	1-8068	28673	20	83	1880	97	2012-08-02 05:40:01.951158
28676	4	1-8520	28673	83	41	1756	6	2012-08-02 05:40:01.951158
28673	4	1-8542	28673	41	88	1929	60	2012-08-02 05:40:01.951158
28678	4	1-8423	28677	52	86	2096	25	2012-08-02 05:40:02.349042
28679	4	1-6031	28677	86	45	1305	93	2012-08-02 05:40:02.349042
28680	4	1-5866	28677	45	27	547	4	2012-08-02 05:40:02.349042
28677	4	1-8543	28677	27	52	2856	92	2012-08-02 05:40:02.349042
28682	3	1-7252	28681	90	17	1572	81	2012-08-02 05:40:02.349042
28683	3	1-5887	28681	17	27	900	4	2012-08-02 05:40:02.349042
28681	3	1-8543	28681	27	90	4907	92	2012-08-02 05:40:02.349042
28685	4	1-8017	28684	32	82	2360	46	2012-08-02 05:40:03.917903
28686	4	1-7606	28684	82	89	1058	6	2012-08-02 05:40:03.917903
28687	4	1-4325	28684	89	91	271	50	2012-08-02 05:40:03.917903
28684	4	1-8547	28684	91	32	8419	60	2012-08-02 05:40:03.917903
28689	3	1-6930	28688	22	39	1016	73	2012-08-02 05:40:05.133191
28690	3	1-8502	28688	39	7	1318	42	2012-08-02 05:40:05.133191
28688	3	1-8550	28688	7	22	4070	27	2012-08-02 05:40:05.133191
28692	4	1-8335	28691	95	71	237	32	2012-08-02 05:40:05.133191
28693	4	1-7794	28691	71	68	72	61	2012-08-02 05:40:05.133191
28694	4	1-6276	28691	68	7	49	42	2012-08-02 05:40:05.133191
28691	4	1-8550	28691	7	95	153	27	2012-08-02 05:40:05.133191
28696	4	1-8068	28695	20	83	4178	97	2012-08-02 05:40:05.133191
28697	4	1-8520	28695	83	62	3062	6	2012-08-02 05:40:05.133191
28698	4	1-3245	28695	62	7	1201	42	2012-08-02 05:40:05.133191
28695	4	1-8550	28695	7	20	3936	27	2012-08-02 05:40:05.133191
28700	4	1-7115	28699	36	84	3310	79	2012-08-02 05:40:08.371906
28701	4	1-8501	28699	84	8	3759	19	2012-08-02 05:40:08.371906
28702	4	1-6991	28699	8	86	2776	44	2012-08-02 05:40:08.371906
28699	4	1-8558	28699	86	36	3988	23	2012-08-02 05:40:08.371906
28704	3	1-5012	28703	33	94	1372	57	2012-08-02 05:40:08.371906
28705	3	1-7918	28703	94	86	741	44	2012-08-02 05:40:08.371906
28703	3	1-8558	28703	86	33	1226	23	2012-08-02 05:40:08.371906
28707	5	1-8271	28706	25	48	2989	71	2012-08-02 05:40:08.371906
28708	5	1-6645	28706	48	49	1583	20	2012-08-02 05:40:08.371906
28709	5	1-4049	28706	49	66	311	27	2012-08-02 05:40:08.371906
28710	5	1-8532	28706	66	86	671	44	2012-08-02 05:40:08.371906
28706	5	1-8558	28706	86	25	1138	23	2012-08-02 05:40:08.371906
28712	4	1-7901	28711	73	46	7424	2	2012-08-02 05:40:08.78121
28713	4	1-8334	28711	46	32	3699	31	2012-08-02 05:40:08.78121
28714	4	1-8527	28711	32	12	1765	68	2012-08-02 05:40:08.78121
28711	4	1-8559	28711	12	73	1416	12	2012-08-02 05:40:08.78121
28716	4	1-7570	28715	6	28	2628	58	2012-08-02 05:40:11.244956
28717	4	1-6845	28715	28	8	2933	48	2012-08-02 05:40:11.244956
28718	4	1-8373	28715	8	67	1803	100	2012-08-02 05:40:11.244956
28715	4	1-8565	28715	67	6	1103	9	2012-08-02 05:40:11.244956
28720	5	1-7599	28719	59	25	688	95	2012-08-02 05:40:11.644543
28721	5	1-8479	28719	25	47	822	69	2012-08-02 05:40:11.644543
28722	5	1-8495	28719	47	98	238	30	2012-08-02 05:40:11.644543
28723	5	1-8101	28719	98	31	68	27	2012-08-02 05:40:11.644543
28719	5	1-8566	28719	31	59	320	83	2012-08-02 05:40:11.644543
28725	4	1-6996	28724	59	68	3200	73	2012-08-02 05:40:12.053327
28726	4	1-8399	28724	68	59	2586	76	2012-08-02 05:40:12.053327
28727	4	1-8134	28724	59	40	2082	48	2012-08-02 05:40:12.053327
28724	4	1-8567	28724	40	59	5178	9	2012-08-02 05:40:12.053327
28729	3	1-6036	28728	71	80	1469	79	2012-08-02 05:40:12.053327
28730	3	1-7398	28728	80	40	1132	48	2012-08-02 05:40:12.053327
28728	3	1-8567	28728	40	71	3087	9	2012-08-02 05:40:12.053327
28732	4	1-5611	28731	1	32	73	3	2012-08-02 05:40:12.053327
28733	4	1-8102	28731	32	61	129	21	2012-08-02 05:40:12.053327
28734	4	1-8054	28731	61	40	75	48	2012-08-02 05:40:12.053327
28731	4	1-8567	28731	40	1	206	9	2012-08-02 05:40:12.053327
28736	3	1-5497	28735	90	46	917	53	2012-08-02 05:40:12.451128
28737	3	1-8505	28735	46	15	676	10	2012-08-02 05:40:12.451128
28735	3	1-8568	28735	15	90	5210	3	2012-08-02 05:40:12.451128
28739	3	1-8520	28738	83	62	719	6	2012-08-02 05:40:13.633446
28740	3	1-3245	28738	62	54	263	42	2012-08-02 05:40:13.633446
28738	3	1-8571	28738	54	83	1051	97	2012-08-02 05:40:13.633446
28742	4	1-1879	28741	59	97	345	44	2012-08-02 05:40:16.063536
28743	4	1-8530	28741	97	75	274	55	2012-08-02 05:40:16.063536
28744	4	1-8557	28741	75	57	512	51	2012-08-02 05:40:16.063536
28741	4	1-8577	28741	57	59	1071	36	2012-08-02 05:40:16.063536
28746	3	1-8378	28745	23	32	4524	82	2012-08-02 05:40:16.063536
28747	3	1-8370	28745	32	57	2602	51	2012-08-02 05:40:16.063536
28745	3	1-8577	28745	57	23	5905	36	2012-08-02 05:40:16.063536
28749	3	1-7949	28748	46	25	9413	95	2012-08-02 05:40:17.290369
28750	3	1-8479	28748	25	9	9138	69	2012-08-02 05:40:17.290369
28748	3	1-8580	28748	9	46	3797	12	2012-08-02 05:40:17.290369
28752	4	1-5206	28751	51	11	289	89	2012-08-02 05:40:17.290369
28753	4	1-6512	28751	11	70	227	76	2012-08-02 05:40:17.290369
28754	4	1-8114	28751	70	9	539	69	2012-08-02 05:40:17.290369
28751	4	1-8580	28751	9	51	307	12	2012-08-02 05:40:17.290369
28756	4	1-7608	28755	90	16	692	52	2012-08-02 05:40:17.290369
28757	4	1-8257	28755	16	26	1399	100	2012-08-02 05:40:17.290369
28758	4	1-8575	28755	26	9	1075	69	2012-08-02 05:40:17.290369
28755	4	1-8580	28755	9	90	617	12	2012-08-02 05:40:17.290369
28760	4	1-8530	28759	97	57	3491	55	2012-08-02 05:40:18.085757
28761	4	1-3317	28759	57	47	1464	33	2012-08-02 05:40:18.085757
28762	4	1-7892	28759	47	51	855	49	2012-08-02 05:40:18.085757
28759	4	1-8582	28759	51	97	4404	44	2012-08-02 05:40:18.085757
28764	4	1-8558	28763	86	73	1818	23	2012-08-02 05:40:18.085757
28765	4	1-8039	28763	73	96	1763	56	2012-08-02 05:40:18.085757
28766	4	1-2806	28763	96	51	229	49	2012-08-02 05:40:18.085757
28763	4	1-8582	28763	51	86	1308	44	2012-08-02 05:40:18.085757
28768	5	1-8530	28767	97	89	2232	55	2012-08-02 05:40:18.085757
28769	5	1-3643	28767	89	2	1102	73	2012-08-02 05:40:18.085757
28770	5	1-7019	28767	2	47	590	33	2012-08-02 05:40:18.085757
28771	5	1-7892	28767	47	51	402	49	2012-08-02 05:40:18.085757
28767	5	1-8582	28767	51	97	2415	44	2012-08-02 05:40:18.085757
28773	4	1-8530	28772	97	4	1370	55	2012-08-02 05:40:18.085757
28774	4	1-8325	28772	4	69	919	32	2012-08-02 05:40:18.085757
28775	4	1-5536	28772	69	51	225	49	2012-08-02 05:40:18.085757
28772	4	1-8582	28772	51	97	1416	44	2012-08-02 05:40:18.085757
28777	4	1-8539	28776	12	28	603	78	2012-08-02 05:40:19.743317
28778	4	1-8215	28776	28	22	928	28	2012-08-02 05:40:19.743317
28779	4	1-6204	28776	22	81	710	57	2012-08-02 05:40:19.743317
28776	4	1-8586	28776	81	12	504	96	2012-08-02 05:40:19.743317
28781	4	1-8438	28780	19	39	849	30	2012-08-02 05:40:20.152002
28782	4	1-7199	28780	39	28	625	79	2012-08-02 05:40:20.152002
28783	4	1-7652	28780	28	51	417	93	2012-08-02 05:40:20.152002
28780	4	1-8587	28780	51	19	409	20	2012-08-02 05:40:20.152002
28785	4	1-8438	28784	19	48	2276	30	2012-08-02 05:40:20.152002
28786	4	1-4941	28784	48	98	1984	57	2012-08-02 05:40:20.152002
28787	4	1-7491	28784	98	51	928	93	2012-08-02 05:40:20.152002
28784	4	1-8587	28784	51	19	999	20	2012-08-02 05:40:20.152002
28789	4	1-8348	28788	39	68	325	52	2012-08-02 05:40:20.152002
28790	4	1-5921	28788	68	39	761	53	2012-08-02 05:40:20.152002
28791	4	1-8439	28788	39	51	457	93	2012-08-02 05:40:20.152002
28788	4	1-8587	28788	51	39	494	20	2012-08-02 05:40:20.152002
28793	3	1-8511	28792	9	42	1085	82	2012-08-02 05:40:20.549757
28794	3	1-5519	28792	42	21	97	89	2012-08-02 05:40:20.549757
28792	3	1-8588	28792	21	9	3600	84	2012-08-02 05:40:20.549757
28796	4	1-8505	28795	46	63	1084	10	2012-08-02 05:40:22.560516
28797	4	1-7394	28795	63	33	266	75	2012-08-02 05:40:22.560516
28798	4	1-7742	28795	33	36	135	46	2012-08-02 05:40:22.560516
28795	4	1-8593	28795	36	46	1138	53	2012-08-02 05:40:22.560516
28800	4	1-8345	28799	60	67	1880	100	2012-08-02 05:40:22.560516
28801	4	1-8565	28799	67	12	875	9	2012-08-02 05:40:22.560516
28802	4	1-7843	28799	12	36	523	46	2012-08-02 05:40:22.560516
28799	4	1-8593	28799	36	60	4734	53	2012-08-02 05:40:22.560516
28804	5	1-8563	28803	29	27	233	4	2012-08-02 05:40:23.367235
28805	5	1-8543	28803	27	76	1052	92	2012-08-02 05:40:23.367235
28806	5	1-7948	28803	76	19	342	98	2012-08-02 05:40:23.367235
28807	5	1-5231	28803	19	42	225	72	2012-08-02 05:40:23.367235
28803	5	1-8595	28803	42	29	111	1	2012-08-02 05:40:23.367235
28809	3	1-8498	28808	82	36	2644	79	2012-08-02 05:40:24.173794
28810	3	1-8049	28808	36	46	1880	64	2012-08-02 05:40:24.173794
28808	3	1-8597	28808	46	82	6393	26	2012-08-02 05:40:24.173794
28812	2	1-5945	28811	72	35	765	53	2012-08-02 05:40:25.388569
28811	2	1-8600	28811	35	72	601	99	2012-08-02 05:40:25.388569
28814	2	1-7884	28813	55	35	5121	53	2012-08-02 05:40:25.388569
28813	2	1-8600	28813	35	55	4061	99	2012-08-02 05:40:25.388569
28816	4	1-7899	28815	73	26	4035	63	2012-08-02 05:40:25.388569
28817	4	1-8538	28815	26	47	2637	98	2012-08-02 05:40:25.388569
28818	4	1-7117	28815	47	35	3226	53	2012-08-02 05:40:25.388569
28815	4	1-8600	28815	35	73	2959	99	2012-08-02 05:40:25.388569
28820	5	1-7843	28819	12	11	583	46	2012-08-02 05:40:26.195277
28821	5	1-8312	28819	11	11	673	43	2012-08-02 05:40:26.195277
28822	5	1-8137	28819	11	63	416	49	2012-08-02 05:40:26.195277
28823	5	1-8294	28819	63	97	664	77	2012-08-02 05:40:26.195277
28819	5	1-8602	28819	97	12	614	9	2012-08-02 05:40:26.195277
28825	4	1-8477	28824	85	74	399	80	2012-08-02 05:40:27.001819
28826	4	1-8346	28824	74	98	878	7	2012-08-02 05:40:27.001819
28827	4	1-6071	28824	98	64	326	100	2012-08-02 05:40:27.001819
28824	4	1-8604	28824	64	85	139	45	2012-08-02 05:40:27.001819
28829	3	1-7008	28828	76	13	3205	14	2012-08-02 05:40:28.592764
28830	3	1-8499	28828	13	45	2112	100	2012-08-02 05:40:28.592764
28828	3	1-8608	28828	45	76	1905	30	2012-08-02 05:40:28.592764
28832	3	1-7595	28831	82	66	755	27	2012-08-02 05:40:30.570252
28833	3	1-8532	28831	66	6	1634	44	2012-08-02 05:40:30.570252
28831	3	1-8613	28831	6	82	1096	1	2012-08-02 05:40:30.570252
28834	1	1-7595	\N	82	82	2674	27	2012-08-02 05:40:30.570252
28836	4	1-4660	28835	30	74	230	57	2012-08-02 05:40:31.354772
28837	4	1-7821	28835	74	9	240	43	2012-08-02 05:40:31.354772
28838	4	1-6456	28835	9	21	155	85	2012-08-02 05:40:31.354772
28835	4	1-8615	28835	21	30	329	36	2012-08-02 05:40:31.354772
28840	4	1-8573	28839	18	77	2465	47	2012-08-02 05:40:31.354772
28841	4	1-8327	28839	77	9	4141	43	2012-08-02 05:40:31.354772
28842	4	1-6456	28839	9	21	2838	85	2012-08-02 05:40:31.354772
28839	4	1-8615	28839	21	18	6376	36	2012-08-02 05:40:31.354772
28844	4	1-6829	28843	16	2	306	31	2012-08-02 05:40:31.354772
28845	4	1-7958	28843	2	9	249	43	2012-08-02 05:40:31.354772
28846	4	1-6456	28843	9	21	175	85	2012-08-02 05:40:31.354772
28843	4	1-8615	28843	21	16	404	36	2012-08-02 05:40:31.354772
28848	4	1-8378	28847	23	20	1018	82	2012-08-02 05:40:31.354772
28849	4	1-5830	28847	20	38	1175	80	2012-08-02 05:40:31.354772
28850	4	1-6079	28847	38	21	483	85	2012-08-02 05:40:31.354772
28847	4	1-8615	28847	21	23	1145	36	2012-08-02 05:40:31.354772
28852	3	1-7115	28851	36	36	2889	79	2012-08-02 05:40:31.763548
28853	3	1-8049	28851	36	78	2935	64	2012-08-02 05:40:31.763548
28851	3	1-8616	28851	78	36	3201	23	2012-08-02 05:40:31.763548
28855	4	1-7577	28854	21	33	3110	87	2012-08-02 05:40:32.570378
28856	4	1-8319	28854	33	32	2320	31	2012-08-02 05:40:32.570378
28857	4	1-8527	28854	32	22	1096	68	2012-08-02 05:40:32.570378
28854	4	1-8618	28854	22	21	2009	77	2012-08-02 05:40:32.570378
28859	4	1-6263	28858	81	53	4047	89	2012-08-02 05:40:32.570378
28860	4	1-8387	28858	53	62	3933	12	2012-08-02 05:40:32.570378
28861	4	1-8073	28858	62	22	2654	68	2012-08-02 05:40:32.570378
28858	4	1-8618	28858	22	81	5038	77	2012-08-02 05:40:32.570378
28863	4	1-8340	28862	99	55	4286	86	2012-08-02 05:40:33.78574
28864	4	1-1235	28862	55	49	1999	15	2012-08-02 05:40:33.78574
28865	4	1-7747	28862	49	54	3093	2	2012-08-02 05:40:33.78574
28862	4	1-8621	28862	54	99	2163	18	2012-08-02 05:40:33.78574
28867	4	1-8034	28866	99	48	1018	7	2012-08-02 05:40:33.78574
28868	4	1-7239	28866	48	7	1127	31	2012-08-02 05:40:33.78574
28869	4	1-8468	28866	7	54	1148	2	2012-08-02 05:40:33.78574
28866	4	1-8621	28866	54	99	943	18	2012-08-02 05:40:33.78574
28871	4	1-5979	28870	33	76	3753	93	2012-08-02 05:40:33.78574
28872	4	1-7089	28870	76	41	3159	17	2012-08-02 05:40:33.78574
28873	4	1-7113	28870	41	54	5684	2	2012-08-02 05:40:33.78574
28870	4	1-8621	28870	54	33	4833	18	2012-08-02 05:40:33.78574
28875	3	1-8034	28874	99	39	266	7	2012-08-02 05:40:33.78574
28876	3	1-7346	28874	39	54	241	2	2012-08-02 05:40:33.78574
28874	3	1-8621	28874	54	99	221	18	2012-08-02 05:40:33.78574
28878	5	1-8490	28877	90	18	947	10	2012-08-02 05:40:34.592473
28879	5	1-8523	28877	18	24	1205	3	2012-08-02 05:40:34.592473
28880	5	1-8421	28877	24	12	1128	96	2012-08-02 05:40:34.592473
28881	5	1-8539	28877	12	98	1372	78	2012-08-02 05:40:34.592473
28877	5	1-8623	28877	98	90	866	54	2012-08-02 05:40:34.592473
28883	4	1-7084	28882	8	66	38	90	2012-08-02 05:40:35.409584
28884	4	1-8324	28882	66	26	18	87	2012-08-02 05:40:35.409584
28885	4	1-8620	28882	26	2	129	40	2012-08-02 05:40:35.409584
28882	4	1-8625	28882	2	8	104	39	2012-08-02 05:40:35.409584
28887	4	1-5541	28886	51	97	549	44	2012-08-02 05:40:35.409584
28888	4	1-8530	28886	97	73	561	55	2012-08-02 05:40:35.409584
28889	4	1-8427	28886	73	2	569	40	2012-08-02 05:40:35.409584
28886	4	1-8625	28886	2	51	622	39	2012-08-02 05:40:35.409584
28891	4	1-7527	28890	40	29	1243	1	2012-08-02 05:40:35.409584
28892	4	1-8563	28890	29	68	3253	4	2012-08-02 05:40:35.409584
28893	4	1-8284	28890	68	2	1155	40	2012-08-02 05:40:35.409584
28890	4	1-8625	28890	2	40	1286	39	2012-08-02 05:40:35.409584
28895	4	1-7912	28894	15	83	159	97	2012-08-02 05:40:35.409584
28896	4	1-8520	28894	83	91	156	6	2012-08-02 05:40:35.409584
28897	4	1-7374	28894	91	2	147	40	2012-08-02 05:40:35.409584
28894	4	1-8625	28894	2	15	164	39	2012-08-02 05:40:35.409584
28899	2	1-4868	28898	78	2	2369	40	2012-08-02 05:40:35.409584
28898	2	1-8625	28898	2	78	2670	39	2012-08-02 05:40:35.409584
28900	1	1-3429	\N	37	37	967	44	2012-08-02 05:40:35.409584
28902	5	1-7407	28901	3	82	1596	26	2012-08-02 05:40:37.000549
28903	5	1-8498	28901	82	28	1014	79	2012-08-02 05:40:37.000549
28904	5	1-7652	28901	28	51	791	93	2012-08-02 05:40:37.000549
28905	5	1-8587	28901	51	62	907	20	2012-08-02 05:40:37.000549
28901	5	1-8629	28901	62	3	3722	10	2012-08-02 05:40:37.000549
28907	4	1-7751	28906	14	61	1701	6	2012-08-02 05:40:38.193692
28908	4	1-8280	28906	61	17	1267	98	2012-08-02 05:40:38.193692
28909	4	1-8265	28906	17	69	2951	95	2012-08-02 05:40:38.193692
28906	4	1-8632	28906	69	14	3711	65	2012-08-02 05:40:38.193692
28911	4	1-7751	28910	14	87	1425	6	2012-08-02 05:40:38.193692
28912	4	1-8564	28910	87	95	1009	99	2012-08-02 05:40:38.193692
28913	4	1-8607	28910	95	69	1908	95	2012-08-02 05:40:38.193692
28910	4	1-8632	28910	69	14	2732	65	2012-08-02 05:40:38.193692
28915	4	1-7567	28914	24	66	747	59	2012-08-02 05:40:38.602625
28916	4	1-8512	28914	66	65	403	91	2012-08-02 05:40:38.602625
28917	4	1-7835	28914	65	75	463	37	2012-08-02 05:40:38.602625
28914	4	1-8633	28914	75	24	1650	10	2012-08-02 05:40:38.602625
28919	4	1-7394	28918	63	3	571	75	2012-08-02 05:40:38.602625
28920	4	1-8458	28918	3	31	1134	83	2012-08-02 05:40:38.602625
28921	4	1-8085	28918	31	75	426	37	2012-08-02 05:40:38.602625
28918	4	1-8633	28918	75	63	1533	10	2012-08-02 05:40:38.602625
28923	5	1-7498	28922	61	100	1037	76	2012-08-02 05:40:39.0004
28924	5	1-8440	28922	100	98	1170	30	2012-08-02 05:40:39.0004
28925	5	1-8101	28922	98	31	363	27	2012-08-02 05:40:39.0004
28926	5	1-8566	28922	31	95	1851	83	2012-08-02 05:40:39.0004
28922	5	1-8634	28922	95	61	2881	17	2012-08-02 05:40:39.0004
28928	3	1-5940	28927	8	52	2401	87	2012-08-02 05:40:40.648658
28929	3	1-8270	28927	52	86	523	13	2012-08-02 05:40:40.648658
28927	3	1-8638	28927	86	8	3142	67	2012-08-02 05:40:40.648658
28931	4	1-6224	28930	21	23	650	38	2012-08-02 05:40:40.648658
28932	4	1-7337	28930	23	23	595	14	2012-08-02 05:40:40.648658
28933	4	1-8098	28930	23	86	178	13	2012-08-02 05:40:40.648658
28930	4	1-8638	28930	86	21	1070	67	2012-08-02 05:40:40.648658
28935	4	1-8017	28934	32	75	309	46	2012-08-02 05:40:41.035243
28936	4	1-8299	28934	75	13	88	1	2012-08-02 05:40:41.035243
28937	4	1-8562	28934	13	89	219	42	2012-08-02 05:40:41.035243
28934	4	1-8639	28934	89	32	1121	60	2012-08-02 05:40:41.035243
28939	4	1-7146	28938	85	14	210	27	2012-08-02 05:40:41.035243
28940	4	1-8122	28938	14	13	118	1	2012-08-02 05:40:41.035243
28941	4	1-8562	28938	13	89	399	42	2012-08-02 05:40:41.035243
28938	4	1-8639	28938	89	85	2778	60	2012-08-02 05:40:41.035243
28943	3	1-5357	28942	83	51	464	9	2012-08-02 05:40:41.035243
28944	3	1-7206	28942	51	89	421	42	2012-08-02 05:40:41.035243
28942	3	1-8639	28942	89	83	2953	60	2012-08-02 05:40:41.035243
28946	5	1-8580	28945	9	51	171	12	2012-08-02 05:40:41.444129
28947	5	1-5206	28945	51	11	122	89	2012-08-02 05:40:41.444129
28948	5	1-6512	28945	11	88	73	76	2012-08-02 05:40:41.444129
28949	5	1-8164	28945	88	44	138	60	2012-08-02 05:40:41.444129
28945	5	1-8640	28945	44	9	397	69	2012-08-02 05:40:41.444129
28951	4	1-5954	28950	93	2	2227	40	2012-08-02 05:40:41.444129
28952	4	1-8625	28950	2	7	2133	39	2012-08-02 05:40:41.444129
28953	4	1-6628	28950	7	44	1969	60	2012-08-02 05:40:41.444129
28950	4	1-8640	28950	44	93	6359	69	2012-08-02 05:40:41.444129
28955	5	1-8415	28954	28	22	590	68	2012-08-02 05:40:43.045957
28956	5	1-8618	28954	22	81	1116	77	2012-08-02 05:40:43.045957
28957	5	1-6263	28954	81	53	892	89	2012-08-02 05:40:43.045957
28958	5	1-8387	28954	53	56	863	12	2012-08-02 05:40:43.045957
28954	5	1-8644	28954	56	28	835	80	2012-08-02 05:40:43.045957
28960	4	1-8265	28959	17	69	1540	95	2012-08-02 05:40:44.636934
28961	4	1-8632	28959	69	22	1998	65	2012-08-02 05:40:44.636934
28962	4	1-6493	28959	22	25	922	51	2012-08-02 05:40:44.636934
28959	4	1-8648	28959	25	17	641	98	2012-08-02 05:40:44.636934
28964	4	1-5231	28963	19	93	106	72	2012-08-02 05:40:44.636934
28965	4	1-7868	28963	93	75	64	55	2012-08-02 05:40:44.636934
28966	4	1-8557	28963	75	25	157	51	2012-08-02 05:40:44.636934
28963	4	1-8648	28963	25	19	129	98	2012-08-02 05:40:44.636934
28968	4	1-8122	28967	14	99	2925	1	2012-08-02 05:40:45.045782
28969	4	1-8636	28967	99	61	5896	90	2012-08-02 05:40:45.045782
28970	4	1-6572	28967	61	85	5022	63	2012-08-02 05:40:45.045782
28967	4	1-8649	28967	85	14	5391	27	2012-08-02 05:40:45.045782
28972	4	1-8566	28971	31	38	2084	83	2012-08-02 05:40:45.045782
28973	4	1-2325	28971	38	99	778	80	2012-08-02 05:40:45.045782
28974	4	1-5966	28971	99	85	327	63	2012-08-02 05:40:45.045782
28971	4	1-8649	28971	85	31	399	27	2012-08-02 05:40:45.045782
28976	4	1-8566	28975	31	90	1388	83	2012-08-02 05:40:45.045782
28977	4	1-8377	28975	90	2	700	7	2012-08-02 05:40:45.045782
28978	4	1-8152	28975	2	85	217	63	2012-08-02 05:40:45.045782
28975	4	1-8649	28975	85	31	265	27	2012-08-02 05:40:45.045782
28980	4	1-7452	28979	100	39	376	13	2012-08-02 05:40:45.045782
28981	4	1-8465	28979	39	20	411	46	2012-08-02 05:40:45.045782
28982	4	1-7357	28979	20	85	346	63	2012-08-02 05:40:45.045782
28979	4	1-8649	28979	85	100	426	27	2012-08-02 05:40:45.045782
28984	4	1-8320	28983	12	60	3196	67	2012-08-02 05:40:45.454809
28985	4	1-4417	28983	60	76	2428	15	2012-08-02 05:40:45.454809
28986	4	1-8598	28983	76	82	1324	56	2012-08-02 05:40:45.454809
28983	4	1-8650	28983	82	12	1794	76	2012-08-02 05:40:45.454809
28988	3	1-8380	28987	46	73	770	23	2012-08-02 05:40:45.454809
28989	3	1-8039	28987	73	82	683	56	2012-08-02 05:40:45.454809
28987	3	1-8650	28987	82	46	1135	76	2012-08-02 05:40:45.454809
28991	4	1-8440	28990	100	69	381	30	2012-08-02 05:40:45.454809
28992	4	1-8209	28990	69	20	255	41	2012-08-02 05:40:45.454809
28993	4	1-7414	28990	20	82	206	56	2012-08-02 05:40:45.454809
28990	4	1-8650	28990	82	100	384	76	2012-08-02 05:40:45.454809
28995	2	1-8390	28994	36	57	5766	28	2012-08-02 05:40:45.874827
28994	2	1-8651	28994	57	36	3402	96	2012-08-02 05:40:45.874827
28997	3	1-8539	28996	12	28	1111	78	2012-08-02 05:40:45.874827
28998	3	1-8215	28996	28	57	1640	28	2012-08-02 05:40:45.874827
28996	3	1-8651	28996	57	12	969	96	2012-08-02 05:40:45.874827
29000	2	1-8486	28999	18	62	4784	29	2012-08-02 05:40:46.272495
28999	2	1-8652	28999	62	18	6987	15	2012-08-02 05:40:46.272495
29002	4	1-8024	29001	95	9	2683	87	2012-08-02 05:40:47.068242
29003	4	1-7764	29001	9	99	854	47	2012-08-02 05:40:47.068242
29004	4	1-8019	29001	99	79	587	64	2012-08-02 05:40:47.068242
29001	4	1-8654	29001	79	95	1173	23	2012-08-02 05:40:47.068242
29006	4	1-8271	29005	25	28	3104	71	2012-08-02 05:40:47.068242
29007	4	1-7556	29005	28	13	3424	95	2012-08-02 05:40:47.068242
29008	4	1-3019	29005	13	79	548	64	2012-08-02 05:40:47.068242
29005	4	1-8654	29005	79	25	1186	23	2012-08-02 05:40:47.068242
29010	5	1-6141	29009	8	51	525	93	2012-08-02 05:40:47.476959
29011	5	1-8587	29009	51	19	565	20	2012-08-02 05:40:47.476959
29012	5	1-8438	29009	19	93	1286	30	2012-08-02 05:40:47.476959
29013	5	1-6016	29009	93	17	613	24	2012-08-02 05:40:47.476959
29009	5	1-8655	29009	17	8	744	27	2012-08-02 05:40:47.476959
29014	1	1-4219	\N	10	10	998	20	2012-08-02 05:40:47.476959
29016	3	1-6059	29015	22	57	1074	76	2012-08-02 05:40:47.874553
29017	3	1-6460	29015	57	67	484	4	2012-08-02 05:40:47.874553
29015	3	1-8656	29015	67	22	4600	78	2012-08-02 05:40:47.874553
29019	4	1-8138	29018	48	82	251	56	2012-08-02 05:40:47.874553
29020	4	1-8650	29018	82	57	363	76	2012-08-02 05:40:47.874553
29021	4	1-6460	29018	57	67	173	4	2012-08-02 05:40:47.874553
29018	4	1-8656	29018	67	48	1739	78	2012-08-02 05:40:47.874553
29023	5	1-8321	29022	31	35	896	83	2012-08-02 05:40:47.874553
29024	5	1-7682	29022	35	70	183	24	2012-08-02 05:40:47.874553
29025	5	1-7696	29022	70	29	102	1	2012-08-02 05:40:47.874553
29026	5	1-8563	29022	29	67	200	4	2012-08-02 05:40:47.874553
29022	5	1-8656	29022	67	31	2239	78	2012-08-02 05:40:47.874553
29028	5	1-5206	29027	51	88	917	89	2012-08-02 05:40:49.078942
29029	5	1-8466	29027	88	25	1384	82	2012-08-02 05:40:49.078942
29030	5	1-7812	29027	25	74	1272	57	2012-08-02 05:40:49.078942
29031	5	1-7821	29027	74	90	1019	43	2012-08-02 05:40:49.078942
29027	5	1-8659	29027	90	51	1498	12	2012-08-02 05:40:49.078942
29033	5	1-5206	29032	51	13	1128	89	2012-08-02 05:40:49.078942
29034	5	1-8359	29032	13	14	1949	86	2012-08-02 05:40:49.078942
29035	5	1-7869	29032	14	77	580	47	2012-08-02 05:40:49.078942
29036	5	1-8327	29032	77	90	856	43	2012-08-02 05:40:49.078942
29032	5	1-8659	29032	90	51	1523	12	2012-08-02 05:40:49.078942
29038	4	1-4987	29037	98	48	618	30	2012-08-02 05:40:49.078942
29039	4	1-4941	29037	48	74	523	57	2012-08-02 05:40:49.078942
29040	4	1-7821	29037	74	90	593	43	2012-08-02 05:40:49.078942
29037	4	1-8659	29037	90	98	1234	12	2012-08-02 05:40:49.078942
29042	3	1-8321	29041	31	95	794	83	2012-08-02 05:40:50.283218
29043	3	1-8634	29041	95	72	236	17	2012-08-02 05:40:50.283218
29041	3	1-8662	29041	72	31	7896	78	2012-08-02 05:40:50.283218
29045	4	1-5324	29044	46	76	1779	28	2012-08-02 05:40:51.111975
29046	4	1-8471	29044	76	70	1529	42	2012-08-02 05:40:51.111975
29047	4	1-8368	29044	70	28	1407	91	2012-08-02 05:40:51.111975
29044	4	1-8664	29044	28	46	2516	82	2012-08-02 05:40:51.111975
29049	5	1-7458	29048	49	14	663	27	2012-08-02 05:40:51.111975
29050	5	1-8122	29048	14	13	411	1	2012-08-02 05:40:51.111975
29051	5	1-8562	29048	13	70	1530	42	2012-08-02 05:40:51.111975
29052	5	1-8368	29048	70	28	1532	91	2012-08-02 05:40:51.111975
29048	5	1-8664	29048	28	49	2981	82	2012-08-02 05:40:51.111975
29054	4	1-8626	29053	57	1	687	16	2012-08-02 05:40:51.907274
29055	4	1-7768	29053	1	63	199	32	2012-08-02 05:40:51.907274
29056	4	1-7951	29053	63	52	344	88	2012-08-02 05:40:51.907274
29053	4	1-8666	29053	52	57	1085	54	2012-08-02 05:40:51.907274
29058	4	1-3777	29057	100	13	61	89	2012-08-02 05:40:51.907274
29059	4	1-8359	29057	13	83	105	86	2012-08-02 05:40:51.907274
29060	4	1-8388	29057	83	52	217	88	2012-08-02 05:40:51.907274
29057	4	1-8666	29057	52	100	911	54	2012-08-02 05:40:51.907274
29062	4	1-8343	29061	63	67	2116	38	2012-08-02 05:40:52.294194
29063	4	1-8627	29061	67	62	1144	45	2012-08-02 05:40:52.294194
29064	4	1-7547	29061	62	30	1922	70	2012-08-02 05:40:52.294194
29061	4	1-8667	29061	30	63	2263	55	2012-08-02 05:40:52.294194
29066	4	1-8636	29065	99	34	1806	90	2012-08-02 05:40:52.691863
29067	4	1-8541	29065	34	71	752	61	2012-08-02 05:40:52.691863
29068	4	1-7327	29065	71	97	475	71	2012-08-02 05:40:52.691863
29065	4	1-8668	29065	97	99	1303	1	2012-08-02 05:40:52.691863
29070	4	1-7451	29069	80	11	491	94	2012-08-02 05:40:52.691863
29071	4	1-8279	29069	11	41	1145	66	2012-08-02 05:40:52.691863
29072	4	1-7671	29069	41	97	523	71	2012-08-02 05:40:52.691863
29069	4	1-8668	29069	97	80	1600	1	2012-08-02 05:40:52.691863
29074	2	1-6746	29073	27	65	2548	5	2012-08-02 05:40:53.100812
29073	2	1-8669	29073	65	27	8512	41	2012-08-02 05:40:53.100812
29076	3	1-7823	29075	2	43	4202	87	2012-08-02 05:40:53.89614
29077	3	1-8418	29075	43	29	6300	92	2012-08-02 05:40:53.89614
29075	3	1-8671	29075	29	2	6344	96	2012-08-02 05:40:53.89614
29079	4	1-8325	29078	4	66	3022	32	2012-08-02 05:40:54.680854
29080	4	1-8076	29078	66	12	1456	96	2012-08-02 05:40:54.680854
29081	4	1-8539	29078	12	96	1510	78	2012-08-02 05:40:54.680854
29078	4	1-8673	29078	96	4	4918	55	2012-08-02 05:40:54.680854
29083	5	1-7475	29082	54	82	2159	56	2012-08-02 05:40:54.680854
29084	5	1-8650	29082	82	95	4203	76	2012-08-02 05:40:54.680854
29085	5	1-7759	29082	95	12	1267	96	2012-08-02 05:40:54.680854
29086	5	1-8539	29082	12	96	1396	78	2012-08-02 05:40:54.680854
29082	5	1-8673	29082	96	54	4833	55	2012-08-02 05:40:54.680854
29088	4	1-8595	29087	42	29	1992	1	2012-08-02 05:40:56.326727
29089	4	1-8563	29087	29	69	4756	4	2012-08-02 05:40:56.326727
29090	4	1-8001	29087	69	62	4971	60	2012-08-02 05:40:56.326727
29087	4	1-8677	29087	62	42	3525	72	2012-08-02 05:40:56.326727
29092	4	1-8652	29091	62	74	590	15	2012-08-02 05:40:57.917519
29093	4	1-8266	29091	74	77	211	21	2012-08-02 05:40:57.917519
29094	4	1-8109	29091	77	83	143	36	2012-08-02 05:40:57.917519
29091	4	1-8681	29091	83	62	394	29	2012-08-02 05:40:57.917519
29096	4	1-7708	29095	46	70	3383	42	2012-08-02 05:40:58.712942
29097	4	1-8368	29095	70	46	2483	91	2012-08-02 05:40:58.712942
29098	4	1-8282	29095	46	75	1027	18	2012-08-02 05:40:58.712942
29095	4	1-8683	29095	75	46	6635	96	2012-08-02 05:40:58.712942
29100	4	1-8539	29099	12	76	1336	78	2012-08-02 05:40:58.712942
29101	4	1-6140	29099	76	46	538	91	2012-08-02 05:40:58.712942
29102	4	1-8282	29099	46	75	201	18	2012-08-02 05:40:58.712942
29099	4	1-8683	29099	75	12	1383	96	2012-08-02 05:40:58.712942
29104	3	1-8264	29103	79	77	2682	45	2012-08-02 05:41:00.315031
29105	3	1-8300	29103	77	86	6003	51	2012-08-02 05:41:00.315031
29103	3	1-8687	29103	86	79	4374	84	2012-08-02 05:41:00.315031
29107	2	1-5712	29106	62	98	1481	84	2012-08-02 05:41:01.618852
29106	2	1-8690	29106	98	62	5033	75	2012-08-02 05:41:01.618852
29109	4	1-8477	29108	85	28	3307	80	2012-08-02 05:41:02.118379
29110	4	1-8415	29108	28	12	1730	68	2012-08-02 05:41:02.118379
29111	4	1-8559	29108	12	51	1049	12	2012-08-02 05:41:02.118379
29108	4	1-8691	29108	51	85	1504	45	2012-08-02 05:41:02.118379
29113	3	1-8320	29112	12	96	4397	67	2012-08-02 05:41:02.524399
29114	3	1-1677	29112	96	45	365	24	2012-08-02 05:41:02.524399
29112	3	1-8692	29112	45	12	2118	76	2012-08-02 05:41:02.524399
29116	5	1-8323	29115	45	19	4202	26	2012-08-02 05:41:02.524399
29117	5	1-5175	29115	19	69	2087	74	2012-08-02 05:41:02.524399
29118	5	1-8229	29115	69	17	1177	89	2012-08-02 05:41:02.524399
29119	5	1-8663	29115	17	45	596	24	2012-08-02 05:41:02.524399
29115	5	1-8692	29115	45	45	4443	76	2012-08-02 05:41:02.524399
29121	4	1-8439	29120	39	34	407	93	2012-08-02 05:41:03.750739
29122	4	1-8685	29120	34	15	1074	29	2012-08-02 05:41:03.750739
29123	4	1-8473	29120	15	97	1528	16	2012-08-02 05:41:03.750739
29120	4	1-8695	29120	97	39	786	53	2012-08-02 05:41:03.750739
29125	4	1-5767	29124	51	25	338	51	2012-08-02 05:41:05.352941
29126	4	1-8648	29124	25	74	281	98	2012-08-02 05:41:05.352941
29127	4	1-4383	29124	74	36	276	58	2012-08-02 05:41:05.352941
29124	4	1-8699	29124	36	51	379	71	2012-08-02 05:41:05.352941
29129	3	1-7426	29128	97	51	1071	100	2012-08-02 05:41:05.751063
29130	3	1-6988	29128	51	34	508	7	2012-08-02 05:41:05.751063
29128	3	1-8700	29128	34	97	1564	30	2012-08-02 05:41:05.751063
29132	5	1-7426	29131	97	63	2318	100	2012-08-02 05:41:05.751063
29133	5	1-8606	29131	63	46	1818	91	2012-08-02 05:41:05.751063
29134	5	1-8282	29131	46	99	734	18	2012-08-02 05:41:05.751063
29135	5	1-8034	29131	99	34	785	7	2012-08-02 05:41:05.751063
29131	5	1-8700	29131	34	97	2860	30	2012-08-02 05:41:05.751063
29137	3	1-7701	29136	95	22	1375	40	2012-08-02 05:41:05.751063
29138	3	1-8302	29136	22	34	467	7	2012-08-02 05:41:05.751063
29136	3	1-8700	29136	34	95	1890	30	2012-08-02 05:41:05.751063
29140	2	1-8482	29139	76	78	2938	87	2012-08-02 05:41:06.957549
29139	2	1-8703	29139	78	76	3323	50	2012-08-02 05:41:06.957549
29142	5	1-8191	29141	50	4	294	68	2012-08-02 05:41:06.957549
29143	5	1-7988	29141	4	79	288	64	2012-08-02 05:41:06.957549
29144	5	1-8654	29141	79	95	524	23	2012-08-02 05:41:06.957549
29145	5	1-8024	29141	95	78	1090	87	2012-08-02 05:41:06.957549
29141	5	1-8703	29141	78	50	1718	50	2012-08-02 05:41:06.957549
29147	4	1-8535	29146	88	74	5158	65	2012-08-02 05:41:06.957549
29148	4	1-4844	29146	74	66	2208	90	2012-08-02 05:41:06.957549
29149	4	1-8324	29146	66	78	1456	87	2012-08-02 05:41:06.957549
29146	4	1-8703	29146	78	88	2731	50	2012-08-02 05:41:06.957549
29151	3	1-7993	29150	3	35	1832	83	2012-08-02 05:41:07.377522
29152	3	1-7682	29150	35	64	361	24	2012-08-02 05:41:07.377522
29150	3	1-8704	29150	64	3	2248	62	2012-08-02 05:41:07.377522
29154	3	1-7169	29153	28	46	287	68	2012-08-02 05:41:07.377522
29155	3	1-4003	29153	46	64	53	24	2012-08-02 05:41:07.377522
29153	3	1-8704	29153	64	28	378	62	2012-08-02 05:41:07.377522
29157	4	1-8403	29156	1	28	4258	61	2012-08-02 05:41:08.195081
29158	4	1-8589	29156	28	17	3009	8	2012-08-02 05:41:08.195081
29159	4	1-6191	29156	17	62	2725	60	2012-08-02 05:41:08.195081
29156	4	1-8706	29156	62	1	5961	33	2012-08-02 05:41:08.195081
29161	4	1-3035	29160	65	25	2754	51	2012-08-02 05:41:09.399261
29162	4	1-8648	29160	25	74	1783	98	2012-08-02 05:41:09.399261
29163	4	1-4383	29160	74	87	1359	58	2012-08-02 05:41:09.399261
29160	4	1-8709	29160	87	65	3623	9	2012-08-02 05:41:09.399261
29165	4	1-8537	29164	55	47	388	13	2012-08-02 05:41:09.399261
29166	4	1-8181	29164	47	82	504	54	2012-08-02 05:41:09.399261
29167	4	1-7920	29164	82	87	158	58	2012-08-02 05:41:09.399261
29164	4	1-8709	29164	87	55	465	9	2012-08-02 05:41:09.399261
29169	5	1-6029	29168	8	6	1831	44	2012-08-02 05:41:11.785989
29170	5	1-8613	29168	6	13	1024	1	2012-08-02 05:41:11.785989
29171	5	1-8562	29168	13	81	3229	42	2012-08-02 05:41:11.785989
29172	5	1-8244	29168	81	84	2181	94	2012-08-02 05:41:11.785989
29168	5	1-8715	29168	84	8	2926	9	2012-08-02 05:41:11.785989
29174	5	1-7331	29173	40	8	263	15	2012-08-02 05:41:12.978788
29175	5	1-8434	29173	8	35	172	83	2012-08-02 05:41:12.978788
29176	5	1-7682	29173	35	6	47	24	2012-08-02 05:41:12.978788
29177	5	1-8026	29173	6	98	76	67	2012-08-02 05:41:12.978788
29173	5	1-8718	29173	98	40	471	60	2012-08-02 05:41:12.978788
29179	5	1-8677	29178	62	18	3460	72	2012-08-02 05:41:12.978788
29180	5	1-6966	29178	18	96	1655	8	2012-08-02 05:41:12.978788
29181	5	1-1169	29178	96	6	435	24	2012-08-02 05:41:12.978788
29182	5	1-8026	29178	6	98	711	67	2012-08-02 05:41:12.978788
29178	5	1-8718	29178	98	62	4423	60	2012-08-02 05:41:12.978788
29183	1	1-1169	\N	96	96	1119	24	2012-08-02 05:41:12.978788
29185	4	1-8564	29184	87	51	833	99	2012-08-02 05:41:14.580949
29184	4	1-8722	29184	36	87	2778	6	2012-08-02 05:41:14.580949
29186	4	1-8042	29184	51	55	794	85	2012-08-02 05:41:14.580949
29187	4	1-8478	29184	55	36	479	21	2012-08-02 05:41:14.580949
29189	4	1-8692	29188	45	81	3258	76	2012-08-02 05:41:14.978735
29190	4	1-7024	29188	81	88	1742	95	2012-08-02 05:41:14.978735
29191	4	1-7356	29188	88	61	572	63	2012-08-02 05:41:14.978735
29188	4	1-8723	29188	61	45	508	24	2012-08-02 05:41:14.978735
29193	4	1-8375	29192	71	87	672	6	2012-08-02 05:41:14.978735
29194	4	1-8564	29192	87	73	438	99	2012-08-02 05:41:14.978735
29195	4	1-7899	29192	73	61	496	63	2012-08-02 05:41:14.978735
29192	4	1-8723	29192	61	71	466	24	2012-08-02 05:41:14.978735
29197	5	1-8345	29196	60	45	405	100	2012-08-02 05:41:16.205241
29198	5	1-8608	29196	45	93	358	30	2012-08-02 05:41:16.205241
29199	5	1-6016	29196	93	6	180	24	2012-08-02 05:41:16.205241
29200	5	1-8026	29196	6	100	296	67	2012-08-02 05:41:16.205241
29196	5	1-8726	29196	100	60	649	53	2012-08-02 05:41:16.205241
29202	4	1-8674	29201	39	11	137	94	2012-08-02 05:41:17.42052
29203	4	1-8279	29201	11	92	321	66	2012-08-02 05:41:17.42052
29204	4	1-3260	29201	92	61	181	88	2012-08-02 05:41:17.42052
29201	4	1-8729	29201	61	39	390	36	2012-08-02 05:41:17.42052
29206	3	1-8708	29205	25	83	1475	86	2012-08-02 05:41:17.42052
29207	3	1-8388	29205	83	61	3040	88	2012-08-02 05:41:17.42052
29205	3	1-8729	29205	61	25	7624	36	2012-08-02 05:41:17.42052
29209	3	1-8635	29208	40	100	4722	67	2012-08-02 05:41:18.624375
29210	3	1-8726	29208	100	72	5563	53	2012-08-02 05:41:18.624375
29208	3	1-8732	29208	72	40	8949	3	2012-08-02 05:41:18.624375
29212	4	1-8458	29211	3	31	2153	83	2012-08-02 05:41:19.81781
29213	4	1-8085	29211	31	75	443	37	2012-08-02 05:41:19.81781
29214	4	1-8633	29211	75	22	871	10	2012-08-02 05:41:19.81781
29211	4	1-8735	29211	22	3	1982	75	2012-08-02 05:41:19.81781
29216	4	1-8617	29215	62	61	2124	30	2012-08-02 05:41:20.635177
29217	4	1-4022	29215	61	88	485	50	2012-08-02 05:41:20.635177
29218	4	1-8535	29215	88	8	330	65	2012-08-02 05:41:20.635177
29215	4	1-8737	29215	8	62	3640	52	2012-08-02 05:41:20.635177
29220	4	1-8727	29219	7	13	704	1	2012-08-02 05:41:20.635177
29221	4	1-8562	29219	13	65	1029	42	2012-08-02 05:41:20.635177
29222	4	1-8596	29219	65	8	209	65	2012-08-02 05:41:20.635177
29219	4	1-8737	29219	8	7	2497	52	2012-08-02 05:41:20.635177
29224	4	1-8617	29223	62	93	1289	30	2012-08-02 05:41:20.635177
29225	4	1-6016	29223	93	50	269	24	2012-08-02 05:41:20.635177
29226	4	1-8574	29223	50	8	155	65	2012-08-02 05:41:20.635177
29223	4	1-8737	29223	8	62	1942	52	2012-08-02 05:41:20.635177
29227	1	1-6016	\N	93	93	798	24	2012-08-02 05:41:20.635177
29229	4	1-8478	29228	55	48	6155	21	2012-08-02 05:41:21.055129
29230	4	1-8349	29228	48	42	3048	93	2012-08-02 05:41:21.055129
29231	4	1-8176	29228	42	1	1949	16	2012-08-02 05:41:21.055129
29228	4	1-8738	29228	1	55	4064	85	2012-08-02 05:41:21.055129
29233	4	1-6017	29232	59	82	376	56	2012-08-02 05:41:21.44602
29234	4	1-8650	29232	82	59	345	76	2012-08-02 05:41:21.44602
29235	4	1-8134	29232	59	32	221	48	2012-08-02 05:41:21.44602
29232	4	1-8739	29232	32	59	2220	79	2012-08-02 05:41:21.44602
29237	4	1-7911	29236	4	39	1046	53	2012-08-02 05:41:21.832655
29238	4	1-8439	29236	39	34	606	93	2012-08-02 05:41:21.832655
29239	4	1-8685	29236	34	48	1785	29	2012-08-02 05:41:21.832655
29236	4	1-8740	29236	48	4	1014	58	2012-08-02 05:41:21.832655
29241	4	1-8324	29240	66	26	537	87	2012-08-02 05:41:22.230455
29242	4	1-8620	29240	26	5	2806	40	2012-08-02 05:41:22.230455
29243	4	1-6513	29240	5	88	301	65	2012-08-02 05:41:22.230455
29240	4	1-8741	29240	88	66	1590	90	2012-08-02 05:41:22.230455
29245	5	1-8561	29244	17	63	3494	46	2012-08-02 05:41:22.230455
29246	5	1-8392	29244	63	61	1749	63	2012-08-02 05:41:22.230455
29247	5	1-8723	29244	61	50	1243	24	2012-08-02 05:41:22.230455
29248	5	1-8574	29244	50	88	1107	65	2012-08-02 05:41:22.230455
29244	5	1-8741	29244	88	17	7262	90	2012-08-02 05:41:22.230455
29250	4	1-8541	29249	34	68	82	61	2012-08-02 05:41:22.230455
29251	4	1-6276	29249	68	65	57	42	2012-08-02 05:41:22.230455
29252	4	1-8596	29249	65	88	22	65	2012-08-02 05:41:22.230455
29249	4	1-8741	29249	88	34	162	90	2012-08-02 05:41:22.230455
29254	4	1-7539	29253	36	68	843	52	2012-08-02 05:41:23.083455
29255	4	1-5921	29253	68	30	1832	53	2012-08-02 05:41:23.083455
29256	4	1-8199	29253	30	3	1362	25	2012-08-02 05:41:23.083455
29253	4	1-8743	29253	3	36	1176	19	2012-08-02 05:41:23.083455
29258	4	1-6961	29257	59	43	1272	87	2012-08-02 05:41:23.083455
29259	4	1-8418	29257	43	52	2304	92	2012-08-02 05:41:23.083455
29260	4	1-8423	29257	52	3	1730	25	2012-08-02 05:41:23.083455
29257	4	1-8743	29257	3	59	1581	19	2012-08-02 05:41:23.083455
29262	4	1-3513	29261	75	80	393	37	2012-08-02 05:41:23.083455
29263	4	1-8454	29261	80	48	467	32	2012-08-02 05:41:23.083455
29264	4	1-8243	29261	48	3	358	25	2012-08-02 05:41:23.083455
29261	4	1-8743	29261	3	75	338	19	2012-08-02 05:41:23.083455
29266	3	1-8052	29265	3	30	972	53	2012-08-02 05:41:23.083455
29267	3	1-8199	29265	30	3	770	25	2012-08-02 05:41:23.083455
29265	3	1-8743	29265	3	3	739	19	2012-08-02 05:41:23.083455
29269	5	1-6214	29268	41	2	944	40	2012-08-02 05:41:23.083455
29270	5	1-8625	29268	2	15	1056	39	2012-08-02 05:41:23.083455
29271	5	1-7912	29268	15	49	1026	97	2012-08-02 05:41:23.083455
29272	5	1-8381	29268	49	3	1175	25	2012-08-02 05:41:23.083455
29268	5	1-8743	29268	3	41	1142	19	2012-08-02 05:41:23.083455
29274	5	1-7945	29273	87	57	3714	98	2012-08-02 05:41:23.083455
29275	5	1-6911	29273	57	62	3537	80	2012-08-02 05:41:23.083455
29276	5	1-7880	29273	62	49	3329	97	2012-08-02 05:41:23.083455
29277	5	1-8381	29273	49	3	3826	25	2012-08-02 05:41:23.083455
29273	5	1-8743	29273	3	87	3730	19	2012-08-02 05:41:23.083455
29279	2	1-8512	29278	66	61	144	91	2012-08-02 05:41:23.923007
29278	2	1-8745	29278	61	66	1679	59	2012-08-02 05:41:23.923007
29281	2	1-6046	29280	38	46	719	8	2012-08-02 05:41:24.320693
29280	2	1-8746	29280	46	38	2475	19	2012-08-02 05:41:24.320693
29283	4	1-7207	29282	92	67	2225	65	2012-08-02 05:41:24.320693
29284	4	1-3999	29282	67	28	479	61	2012-08-02 05:41:24.320693
29285	4	1-8589	29282	28	46	317	8	2012-08-02 05:41:24.320693
29282	4	1-8746	29282	46	92	2528	19	2012-08-02 05:41:24.320693
29287	4	1-7716	29286	78	4	3142	68	2012-08-02 05:41:24.320693
29288	4	1-7988	29286	4	51	2626	64	2012-08-02 05:41:24.320693
29289	4	1-5233	29286	51	46	535	8	2012-08-02 05:41:24.320693
29286	4	1-8746	29286	46	78	4374	19	2012-08-02 05:41:24.320693
29291	3	1-6991	29290	8	34	78	44	2012-08-02 05:41:24.320693
29292	3	1-7023	29290	34	46	13	8	2012-08-02 05:41:24.320693
29290	3	1-8746	29290	46	8	115	19	2012-08-02 05:41:24.320693
29294	5	1-8743	29293	3	46	95	19	2012-08-02 05:41:24.729739
29295	5	1-7459	29293	46	10	104	55	2012-08-02 05:41:24.729739
29296	5	1-2794	29293	10	98	104	17	2012-08-02 05:41:24.729739
29297	5	1-7930	29293	98	20	82	26	2012-08-02 05:41:24.729739
29293	5	1-8747	29293	20	3	99	25	2012-08-02 05:41:24.729739
29298	1	1-2794	\N	10	10	977	17	2012-08-02 05:41:24.729739
29299	1	1-7930	\N	98	98	2446	26	2012-08-02 05:41:24.729739
29301	4	1-7419	29300	62	97	1860	2	2012-08-02 05:41:25.558295
29302	4	1-8296	29300	97	35	423	81	2012-08-02 05:41:25.558295
29303	4	1-5851	29300	35	70	195	51	2012-08-02 05:41:25.558295
29300	4	1-8749	29300	70	62	762	73	2012-08-02 05:41:25.558295
29305	4	1-8399	29304	68	98	2713	76	2012-08-02 05:41:25.558295
29306	4	1-8069	29304	98	51	2896	54	2012-08-02 05:41:25.558295
29307	4	1-7194	29304	51	70	723	51	2012-08-02 05:41:25.558295
29304	4	1-8749	29304	70	68	3386	73	2012-08-02 05:41:25.558295
29309	4	1-8502	29308	39	65	4686	42	2012-08-02 05:41:25.558295
29310	4	1-8596	29308	65	22	1739	65	2012-08-02 05:41:25.558295
29311	4	1-6493	29308	22	70	690	51	2012-08-02 05:41:25.558295
29308	4	1-8749	29308	70	39	3499	73	2012-08-02 05:41:25.558295
29313	4	1-7019	29312	2	5	161	33	2012-08-02 05:41:25.558295
29314	4	1-8174	29312	5	49	88	74	2012-08-02 05:41:25.558295
29315	4	1-7486	29312	49	70	70	51	2012-08-02 05:41:25.558295
29312	4	1-8749	29312	70	2	360	73	2012-08-02 05:41:25.558295
29317	3	1-6444	29316	81	48	1967	29	2012-08-02 05:41:25.978331
29318	3	1-8740	29316	48	12	971	58	2012-08-02 05:41:25.978331
29316	3	1-8750	29316	12	81	5719	5	2012-08-02 05:41:25.978331
29320	2	1-4960	29319	41	68	636	15	2012-08-02 05:41:26.37614
29319	2	1-8751	29319	68	41	1100	87	2012-08-02 05:41:26.37614
29322	3	1-8319	29321	33	23	1331	31	2012-08-02 05:41:26.37614
29323	3	1-6698	29321	23	68	936	15	2012-08-02 05:41:26.37614
29321	3	1-8751	29321	68	33	2024	87	2012-08-02 05:41:26.37614
29325	4	1-8142	29324	42	52	1729	38	2012-08-02 05:41:26.37614
29326	4	1-217	29324	52	5	963	59	2012-08-02 05:41:26.37614
29327	4	1-8590	29324	5	68	775	15	2012-08-02 05:41:26.37614
29324	4	1-8751	29324	68	42	2023	87	2012-08-02 05:41:26.37614
29329	3	1-5228	29328	22	13	397	1	2012-08-02 05:41:26.784947
29330	3	1-8562	29328	13	10	971	42	2012-08-02 05:41:26.784947
29328	3	1-8752	29328	10	22	1469	61	2012-08-02 05:41:26.784947
29332	2	1-6276	29331	68	10	3363	42	2012-08-02 05:41:26.784947
29331	2	1-8752	29331	10	68	5207	61	2012-08-02 05:41:26.784947
29334	2	1-8680	29333	94	10	1002	42	2012-08-02 05:41:26.784947
29333	2	1-8752	29333	10	94	1588	61	2012-08-02 05:41:26.784947
29335	1	1-1115	\N	68	68	3201	42	2012-08-02 05:41:26.784947
29337	4	1-8751	29336	68	79	3375	87	2012-08-02 05:41:27.613705
29338	4	1-8239	29336	79	75	1421	37	2012-08-02 05:41:27.613705
29339	4	1-8633	29336	75	64	5383	10	2012-08-02 05:41:27.613705
29336	4	1-8754	29336	64	68	1384	15	2012-08-02 05:41:27.613705
29341	4	1-8590	29340	5	76	3838	15	2012-08-02 05:41:28.408974
29342	4	1-8598	29340	76	98	2757	56	2012-08-02 05:41:28.408974
29343	4	1-7602	29340	98	44	1985	51	2012-08-02 05:41:28.408974
29340	4	1-8756	29340	44	5	5913	59	2012-08-02 05:41:28.408974
29345	4	1-6238	29344	28	93	18	72	2012-08-02 05:41:28.408974
29346	4	1-7868	29344	93	75	9	55	2012-08-02 05:41:28.408974
29347	4	1-8557	29344	75	44	19	51	2012-08-02 05:41:28.408974
29344	4	1-8756	29344	44	28	56	59	2012-08-02 05:41:28.408974
29349	4	1-8118	29348	35	34	252	90	2012-08-02 05:41:28.817895
29350	4	1-8541	29348	34	69	170	61	2012-08-02 05:41:28.817895
29351	4	1-7208	29348	69	35	119	99	2012-08-02 05:41:28.817895
29348	4	1-8757	29348	35	35	255	73	2012-08-02 05:41:28.817895
29353	5	1-7155	29352	87	15	725	11	2012-08-02 05:41:29.646372
29354	5	1-8544	29352	15	12	1167	68	2012-08-02 05:41:29.646372
29355	5	1-8559	29352	12	51	924	12	2012-08-02 05:41:29.646372
29356	5	1-8691	29352	51	90	1730	45	2012-08-02 05:41:29.646372
29352	5	1-8759	29352	90	87	1018	8	2012-08-02 05:41:29.646372
29358	4	1-8088	29357	91	28	2945	17	2012-08-02 05:41:30.03304
29359	4	1-8641	29357	28	18	7313	48	2012-08-02 05:41:30.03304
29360	4	1-8020	29357	18	36	4259	90	2012-08-02 05:41:30.03304
29357	4	1-8760	29357	36	91	3494	77	2012-08-02 05:41:30.03304
29362	3	1-8110	29361	12	97	1093	77	2012-08-02 05:41:32.077116
29363	3	1-8602	29361	97	44	541	9	2012-08-02 05:41:32.077116
29361	3	1-8765	29361	44	12	3420	63	2012-08-02 05:41:32.077116
29365	4	1-7787	29364	85	2	1500	40	2012-08-02 05:41:32.077116
29366	4	1-8625	29364	2	55	1122	39	2012-08-02 05:41:32.077116
29367	4	1-8350	29364	55	44	310	9	2012-08-02 05:41:32.077116
29364	4	1-8765	29364	44	85	2402	63	2012-08-02 05:41:32.077116
29369	4	1-8359	29368	13	4	1293	86	2012-08-02 05:41:32.507891
29370	4	1-8480	29368	4	71	1708	48	2012-08-02 05:41:32.507891
29371	4	1-7954	29368	71	37	980	59	2012-08-02 05:41:32.507891
29368	4	1-8766	29368	37	13	918	89	2012-08-02 05:41:32.507891
29373	4	1-7616	29372	7	4	797	86	2012-08-02 05:41:32.507891
29374	4	1-8480	29372	4	71	1296	48	2012-08-02 05:41:32.507891
29375	4	1-7954	29372	71	37	916	59	2012-08-02 05:41:32.507891
29372	4	1-8766	29372	37	7	1055	89	2012-08-02 05:41:32.507891
29377	4	1-8387	29376	53	21	436	12	2012-08-02 05:41:32.507891
29378	4	1-4739	29376	21	71	367	48	2012-08-02 05:41:32.507891
29379	4	1-7954	29376	71	37	320	59	2012-08-02 05:41:32.507891
29376	4	1-8766	29376	37	53	455	89	2012-08-02 05:41:32.507891
29381	4	1-6512	29380	11	59	244	76	2012-08-02 05:41:32.507891
29382	4	1-8134	29380	59	71	238	48	2012-08-02 05:41:32.507891
29383	4	1-7954	29380	71	37	213	59	2012-08-02 05:41:32.507891
29380	4	1-8766	29380	37	11	311	89	2012-08-02 05:41:32.507891
29385	4	1-8707	29384	74	23	5209	31	2012-08-02 05:41:33.756512
29386	4	1-6698	29384	23	75	3852	15	2012-08-02 05:41:33.756512
29387	4	1-6726	29384	75	78	1556	16	2012-08-02 05:41:33.756512
29384	4	1-8769	29384	78	74	4011	7	2012-08-02 05:41:33.756512
29389	2	1-6238	29388	28	80	1967	72	2012-08-02 05:41:34.154091
29388	2	1-8770	29388	80	28	7585	59	2012-08-02 05:41:34.154091
29391	4	1-8766	29390	37	86	1998	89	2012-08-02 05:41:34.154091
29392	4	1-1588	29390	86	62	513	60	2012-08-02 05:41:34.154091
29393	4	1-8677	29390	62	80	337	72	2012-08-02 05:41:34.154091
29390	4	1-8770	29390	80	37	1640	59	2012-08-02 05:41:34.154091
29394	1	1-3874	\N	12	12	993	59	2012-08-02 05:41:34.154091
29396	4	1-6810	29395	7	81	6732	57	2012-08-02 05:41:34.961066
29397	4	1-8586	29395	81	64	4139	96	2012-08-02 05:41:34.961066
29398	4	1-8728	29395	64	82	3559	79	2012-08-02 05:41:34.961066
29395	4	1-8772	29395	82	7	6401	39	2012-08-02 05:41:34.961066
29400	5	1-8459	29399	29	63	5842	46	2012-08-02 05:41:35.369431
29401	5	1-8392	29399	63	61	4305	63	2012-08-02 05:41:35.369431
29402	5	1-8723	29399	61	70	4507	24	2012-08-02 05:41:35.369431
29403	5	1-7696	29399	70	98	3188	1	2012-08-02 05:41:35.369431
29399	5	1-8773	29399	98	29	3743	35	2012-08-02 05:41:35.369431
29405	4	1-8469	29404	32	66	3902	8	2012-08-02 05:41:35.369431
29406	4	1-2207	29404	66	75	2914	46	2012-08-02 05:41:35.369431
29407	4	1-8299	29404	75	98	1185	1	2012-08-02 05:41:35.369431
29404	4	1-8773	29404	98	32	1396	35	2012-08-02 05:41:35.369431
29409	4	1-8257	29408	16	67	5364	100	2012-08-02 05:41:36.187319
29410	4	1-8565	29408	67	55	2887	9	2012-08-02 05:41:36.187319
29411	4	1-8537	29408	55	21	2065	13	2012-08-02 05:41:36.187319
29408	4	1-8775	29408	21	16	3640	52	2012-08-02 05:41:36.187319
29413	4	1-7455	29412	33	97	1239	2	2012-08-02 05:41:36.187319
29414	4	1-8296	29412	97	19	422	81	2012-08-02 05:41:36.187319
29415	4	1-7956	29412	19	21	376	13	2012-08-02 05:41:36.187319
29412	4	1-8775	29412	21	33	747	52	2012-08-02 05:41:36.187319
29417	3	1-3486	29416	6	17	621	57	2012-08-02 05:41:37.402603
29418	3	1-8758	29416	17	16	493	69	2012-08-02 05:41:37.402603
29416	3	1-8778	29416	16	6	613	8	2012-08-02 05:41:37.402603
29420	3	1-7782	29419	28	55	2650	5	2012-08-02 05:41:37.402603
29421	3	1-7181	29419	55	16	1867	69	2012-08-02 05:41:37.402603
29419	3	1-8778	29419	16	28	2392	8	2012-08-02 05:41:37.402603
29423	2	1-8761	29422	86	16	4722	69	2012-08-02 05:41:37.402603
29422	2	1-8778	29422	16	86	6894	8	2012-08-02 05:41:37.402603
29425	5	1-7912	29424	15	60	2141	97	2012-08-02 05:41:37.800244
29426	5	1-8513	29424	60	11	1314	22	2012-08-02 05:41:37.800244
29427	5	1-8086	29424	11	36	2090	58	2012-08-02 05:41:37.800244
29428	5	1-8699	29424	36	7	1858	71	2012-08-02 05:41:37.800244
29424	5	1-8779	29424	7	15	3376	39	2012-08-02 05:41:37.800244
29430	4	1-7332	29429	17	48	1186	29	2012-08-02 05:41:37.800244
29431	4	1-8740	29429	48	36	567	58	2012-08-02 05:41:37.800244
29432	4	1-8699	29429	36	7	624	71	2012-08-02 05:41:37.800244
29429	4	1-8779	29429	7	17	1400	39	2012-08-02 05:41:37.800244
29434	4	1-8625	29433	2	95	262	39	2012-08-02 05:41:38.606973
29435	4	1-6611	29433	95	91	209	42	2012-08-02 05:41:38.606973
29436	4	1-7953	29433	91	32	171	69	2012-08-02 05:41:38.606973
29433	4	1-8781	29433	32	2	291	40	2012-08-02 05:41:38.606973
29438	4	1-8327	29437	77	25	2195	43	2012-08-02 05:41:38.993303
29439	4	1-6425	29437	25	26	1436	46	2012-08-02 05:41:38.993303
29440	4	1-8734	29437	26	74	1703	28	2012-08-02 05:41:38.993303
29437	4	1-8782	29437	74	77	1388	47	2012-08-02 05:41:38.993303
29442	5	1-7334	29441	92	19	2560	81	2012-08-02 05:41:38.993303
29443	5	1-7956	29441	19	39	2437	13	2012-08-02 05:41:38.993303
29444	5	1-8465	29441	39	26	2346	46	2012-08-02 05:41:38.993303
29445	5	1-8734	29441	26	74	2922	28	2012-08-02 05:41:38.993303
29441	5	1-8782	29441	74	92	2502	47	2012-08-02 05:41:38.993303
29447	5	1-8670	29446	62	80	2714	32	2012-08-02 05:41:38.993303
29448	5	1-4638	29446	80	39	958	13	2012-08-02 05:41:38.993303
29449	5	1-8465	29446	39	26	1030	46	2012-08-02 05:41:38.993303
29450	5	1-8734	29446	26	74	1433	28	2012-08-02 05:41:38.993303
29446	5	1-8782	29446	74	62	1370	47	2012-08-02 05:41:38.993303
29452	3	1-8772	29451	82	15	3368	39	2012-08-02 05:41:41.402094
29453	3	1-7912	29451	15	77	2673	97	2012-08-02 05:41:41.402094
29451	3	1-8788	29451	77	82	1938	79	2012-08-02 05:41:41.402094
29455	4	1-6976	29454	66	76	4312	92	2012-08-02 05:41:41.799626
29456	4	1-7948	29454	76	66	644	98	2012-08-02 05:41:41.799626
29457	4	1-8555	29454	66	35	330	93	2012-08-02 05:41:41.799626
29454	4	1-8789	29454	35	66	6724	82	2012-08-02 05:41:41.799626
29459	4	1-8731	29458	39	37	336	59	2012-08-02 05:41:43.821424
29460	4	1-8766	29458	37	57	409	89	2012-08-02 05:41:43.821424
29461	4	1-7756	29458	57	60	259	71	2012-08-02 05:41:43.821424
29458	4	1-8794	29458	60	39	485	99	2012-08-02 05:41:43.821424
29463	3	1-7042	29462	71	24	4964	43	2012-08-02 05:41:43.821424
29464	3	1-8272	29462	24	60	2726	71	2012-08-02 05:41:43.821424
29462	3	1-8794	29462	60	71	5597	99	2012-08-02 05:41:43.821424
29466	3	1-8731	29465	39	53	2130	59	2012-08-02 05:41:43.821424
29467	3	1-8236	29465	53	60	1155	71	2012-08-02 05:41:43.821424
29465	3	1-8794	29465	60	39	2578	99	2012-08-02 05:41:43.821424
29469	4	1-8654	29468	79	36	3544	23	2012-08-02 05:41:44.241105
29470	4	1-7115	29468	36	57	2556	79	2012-08-02 05:41:44.241105
29471	4	1-8784	29468	57	28	4130	88	2012-08-02 05:41:44.241105
29468	4	1-8795	29468	28	79	2213	64	2012-08-02 05:41:44.241105
29473	4	1-8654	29472	79	31	694	23	2012-08-02 05:41:44.241105
29474	4	1-7928	29472	31	55	443	81	2012-08-02 05:41:44.241105
29475	4	1-7664	29472	55	28	485	88	2012-08-02 05:41:44.241105
29472	4	1-8795	29472	28	79	336	64	2012-08-02 05:41:44.241105
29477	3	1-8580	29476	9	46	135	12	2012-08-02 05:41:44.638875
29478	3	1-7949	29476	46	92	222	95	2012-08-02 05:41:44.638875
29476	3	1-8796	29476	92	9	491	69	2012-08-02 05:41:44.638875
29480	4	1-8781	29479	32	22	8949	40	2012-08-02 05:41:44.638875
29481	4	1-8302	29479	22	90	2207	7	2012-08-02 05:41:44.638875
29482	4	1-7662	29479	90	92	1961	95	2012-08-02 05:41:44.638875
29479	4	1-8796	29479	92	32	6135	69	2012-08-02 05:41:44.638875
29484	4	1-6916	29483	97	83	22	86	2012-08-02 05:41:44.638875
29485	4	1-8388	29483	83	29	43	88	2012-08-02 05:41:44.638875
29486	4	1-6834	29483	29	92	25	95	2012-08-02 05:41:44.638875
29483	4	1-8796	29483	92	97	86	69	2012-08-02 05:41:44.638875
29488	4	1-6916	29487	97	83	223	86	2012-08-02 05:41:44.638875
29489	4	1-8388	29487	83	34	451	88	2012-08-02 05:41:44.638875
29490	4	1-6019	29487	34	92	245	95	2012-08-02 05:41:44.638875
29487	4	1-8796	29487	92	97	864	69	2012-08-02 05:41:44.638875
29492	5	1-7864	29491	85	42	1516	72	2012-08-02 05:41:45.445676
29493	5	1-8595	29491	42	98	834	1	2012-08-02 05:41:45.445676
29494	5	1-8773	29491	98	32	907	35	2012-08-02 05:41:45.445676
29491	5	1-8798	29491	93	85	3110	74	2012-08-02 05:41:45.445676
29495	5	1-8469	29491	32	93	2342	8	2012-08-02 05:41:45.445676
29497	4	1-8733	29496	17	30	3691	53	2012-08-02 05:41:46.240961
29498	4	1-8154	29496	30	4	1139	86	2012-08-02 05:41:46.240961
29499	4	1-8480	29496	4	32	2240	48	2012-08-02 05:41:46.240961
29496	4	1-8800	29496	32	17	3649	25	2012-08-02 05:41:46.240961
29501	3	1-8445	29500	6	76	898	70	2012-08-02 05:41:46.240961
29502	3	1-7076	29500	76	32	459	48	2012-08-02 05:41:46.240961
29500	3	1-8800	29500	32	6	762	25	2012-08-02 05:41:46.240961
29504	5	1-8786	29503	23	81	589	75	2012-08-02 05:41:46.638894
29505	5	1-8339	29503	81	41	833	86	2012-08-02 05:41:46.638894
29506	5	1-8355	29503	41	87	713	6	2012-08-02 05:41:46.638894
29507	5	1-8564	29503	87	21	468	99	2012-08-02 05:41:46.638894
29503	5	1-8801	29503	21	23	1221	51	2012-08-02 05:41:46.638894
29509	2	1-4537	29508	60	51	2465	22	2012-08-02 05:41:47.467692
29508	2	1-8803	29508	51	60	6929	18	2012-08-02 05:41:47.467692
29511	3	1-8145	29510	1	99	331	33	2012-08-02 05:41:47.467692
29512	3	1-8220	29510	99	51	281	22	2012-08-02 05:41:47.467692
29510	3	1-8803	29510	51	1	888	18	2012-08-02 05:41:47.467692
29514	4	1-5979	29513	33	51	1943	93	2012-08-02 05:41:50.262525
29515	4	1-8587	29513	51	52	2117	20	2012-08-02 05:41:50.262525
29516	4	1-7818	29513	52	48	2941	4	2012-08-02 05:41:50.262525
29513	4	1-8810	29513	48	33	2307	18	2012-08-02 05:41:50.262525
29518	3	1-2766	29517	59	1	48	16	2012-08-02 05:41:51.07983
29519	3	1-8738	29517	1	53	33	85	2012-08-02 05:41:51.07983
29517	3	1-8812	29517	53	59	188	32	2012-08-02 05:41:51.07983
29521	4	1-8076	29520	66	64	1712	96	2012-08-02 05:41:51.07983
29522	4	1-8728	29520	64	68	736	79	2012-08-02 05:41:51.07983
29523	4	1-8689	29520	68	53	974	85	2012-08-02 05:41:51.07983
29520	4	1-8812	29520	53	66	7122	32	2012-08-02 05:41:51.07983
29525	5	1-7019	29524	2	99	594	33	2012-08-02 05:41:53.103861
29526	5	1-8220	29524	99	26	618	22	2012-08-02 05:41:53.103861
29527	5	1-7210	29524	26	6	666	84	2012-08-02 05:41:53.103861
29528	5	1-7894	29524	6	77	688	64	2012-08-02 05:41:53.103861
29524	5	1-8817	29524	77	2	1067	73	2012-08-02 05:41:53.103861
29530	4	1-4865	29529	49	76	1922	15	2012-08-02 05:41:54.32969
29531	4	1-8598	29529	76	85	1476	56	2012-08-02 05:41:54.32969
29532	4	1-8441	29529	85	28	1528	16	2012-08-02 05:41:54.32969
29529	4	1-8820	29529	28	49	2336	97	2012-08-02 05:41:54.32969
29534	3	1-8513	29533	60	31	3135	22	2012-08-02 05:41:54.32969
29535	3	1-1851	29533	31	28	2219	16	2012-08-02 05:41:54.32969
29533	3	1-8820	29533	28	60	3580	97	2012-08-02 05:41:54.32969
29537	4	1-8513	29536	60	8	2802	22	2012-08-02 05:41:54.32969
29538	4	1-8665	29536	8	42	3221	93	2012-08-02 05:41:54.32969
29539	4	1-8176	29536	42	28	1953	16	2012-08-02 05:41:54.32969
29536	4	1-8820	29536	28	60	3175	97	2012-08-02 05:41:54.32969
29541	4	1-6781	29540	17	69	611	74	2012-08-02 05:41:54.738828
29542	4	1-8229	29540	69	88	365	89	2012-08-02 05:41:54.738828
29543	4	1-8466	29540	88	15	759	82	2012-08-02 05:41:54.738828
29540	4	1-8821	29540	15	17	651	23	2012-08-02 05:41:54.738828
29545	5	1-6578	29544	48	39	854	53	2012-08-02 05:41:55.147501
29546	5	1-8439	29544	39	33	444	93	2012-08-02 05:41:55.147501
29547	5	1-8548	29544	33	46	259	19	2012-08-02 05:41:55.147501
29548	5	1-7459	29544	46	59	245	55	2012-08-02 05:41:55.147501
29544	5	1-8822	29544	59	48	820	94	2012-08-02 05:41:55.147501
29550	4	1-3451	29549	45	37	447	41	2012-08-02 05:41:55.147501
29551	4	1-8035	29549	37	8	348	68	2012-08-02 05:41:55.147501
29552	4	1-6919	29549	8	59	313	55	2012-08-02 05:41:55.147501
29549	4	1-8822	29549	59	45	1077	94	2012-08-02 05:41:55.147501
29554	4	1-8462	29553	23	78	519	16	2012-08-02 05:41:55.147501
29555	4	1-8769	29553	78	82	1381	7	2012-08-02 05:41:55.147501
29556	4	1-7544	29553	82	59	287	55	2012-08-02 05:41:55.147501
29553	4	1-8822	29553	59	23	1027	94	2012-08-02 05:41:55.147501
29558	4	1-6451	29557	87	42	3588	87	2012-08-02 05:41:55.147501
29559	4	1-8142	29557	42	53	3033	38	2012-08-02 05:41:55.147501
29560	4	1-6696	29557	53	59	1222	55	2012-08-02 05:41:55.147501
29557	4	1-8822	29557	59	87	4820	94	2012-08-02 05:41:55.147501
29562	4	1-8544	29561	15	23	4038	68	2012-08-02 05:41:56.35202
29563	4	1-5465	29561	23	13	5006	14	2012-08-02 05:41:56.35202
29564	4	1-8499	29561	13	78	3277	100	2012-08-02 05:41:56.35202
29561	4	1-8825	29561	78	15	2357	11	2012-08-02 05:41:56.35202
29566	4	1-8077	29565	77	62	1108	60	2012-08-02 05:41:59.124768
29567	4	1-8677	29565	62	65	531	72	2012-08-02 05:41:59.124768
29568	4	1-7203	29565	65	14	719	14	2012-08-02 05:41:59.124768
29565	4	1-8832	29565	14	77	1426	3	2012-08-02 05:41:59.124768
29570	3	1-8777	29569	50	40	1009	16	2012-08-02 05:41:59.124768
29571	3	1-7849	29569	40	14	958	14	2012-08-02 05:41:59.124768
29569	3	1-8832	29569	14	50	1932	3	2012-08-02 05:41:59.124768
29573	3	1-7694	29572	98	18	2444	81	2012-08-02 05:42:01.146484
29574	3	1-7976	29572	18	4	2074	54	2012-08-02 05:42:01.146484
29572	3	1-8837	29572	4	98	3603	58	2012-08-02 05:42:01.146484
29576	3	1-8116	29575	88	47	123	13	2012-08-02 05:42:01.146484
29577	3	1-8181	29575	47	4	143	54	2012-08-02 05:42:01.146484
29575	3	1-8837	29575	4	88	263	58	2012-08-02 05:42:01.146484
29579	4	1-8750	29578	12	39	1073	5	2012-08-02 05:42:01.146484
29580	4	1-5838	29578	39	98	81	76	2012-08-02 05:42:01.146484
29581	4	1-8069	29578	98	4	102	54	2012-08-02 05:42:01.146484
29578	4	1-8837	29578	4	12	192	58	2012-08-02 05:42:01.146484
29583	4	1-6845	29582	28	18	454	48	2012-08-02 05:42:01.146484
29584	4	1-8020	29582	18	3	282	90	2012-08-02 05:42:01.146484
29585	4	1-6652	29582	3	4	172	54	2012-08-02 05:42:01.146484
29582	4	1-8837	29582	4	28	375	58	2012-08-02 05:42:01.146484
29587	4	1-8699	29586	36	28	668	71	2012-08-02 05:42:01.146484
29588	4	1-7556	29586	28	61	660	95	2012-08-02 05:42:01.146484
29589	4	1-2319	29586	61	4	245	54	2012-08-02 05:42:01.146484
29586	4	1-8837	29586	4	36	536	58	2012-08-02 05:42:01.146484
29591	4	1-5481	29590	30	61	940	63	2012-08-02 05:42:01.146484
29592	4	1-8723	29590	61	99	969	24	2012-08-02 05:42:01.146484
29593	4	1-8675	29590	99	4	944	54	2012-08-02 05:42:01.146484
29590	4	1-8837	29590	4	30	2120	58	2012-08-02 05:42:01.146484
29594	4	1-8840	29594	32	52	632	20	2012-08-02 05:42:02.350931
29595	4	1-7818	29594	52	62	696	4	2012-08-02 05:42:02.350931
29596	4	1-8232	29594	62	63	219	32	2012-08-02 05:42:02.350931
29597	4	1-7951	29594	63	32	483	88	2012-08-02 05:42:02.350931
29599	2	1-7188	29598	82	32	4858	88	2012-08-02 05:42:02.350931
29598	2	1-8840	29598	32	82	6647	20	2012-08-02 05:42:02.350931
29601	4	1-8289	29600	100	33	5932	93	2012-08-02 05:42:02.748915
29602	4	1-8548	29600	33	8	3111	19	2012-08-02 05:42:02.748915
29603	4	1-6991	29600	8	27	2046	44	2012-08-02 05:42:02.748915
29600	4	1-8841	29600	27	100	3650	47	2012-08-02 05:42:02.748915
29605	5	1-4919	29604	77	9	217	84	2012-08-02 05:42:02.748915
29606	5	1-8511	29604	9	24	219	82	2012-08-02 05:42:02.748915
29607	5	1-6116	29604	24	74	134	62	2012-08-02 05:42:02.748915
29608	5	1-7973	29604	74	27	144	44	2012-08-02 05:42:02.748915
29604	5	1-8841	29604	27	77	283	47	2012-08-02 05:42:02.748915
29610	4	1-5586	29609	8	8	483	92	2012-08-02 05:42:02.748915
29611	4	1-8698	29609	8	100	302	97	2012-08-02 05:42:02.748915
29612	4	1-3918	29609	100	27	112	44	2012-08-02 05:42:02.748915
29609	4	1-8841	29609	27	8	230	47	2012-08-02 05:42:02.748915
29614	5	1-8670	29613	62	50	2214	32	2012-08-02 05:42:02.748915
29615	5	1-4143	29613	50	83	2181	81	2012-08-02 05:42:02.748915
29616	5	1-4310	29613	83	74	539	62	2012-08-02 05:42:02.748915
29617	5	1-7973	29613	74	27	604	44	2012-08-02 05:42:02.748915
29613	5	1-8841	29613	27	62	1246	47	2012-08-02 05:42:02.748915
29619	4	1-8289	29618	100	39	1586	93	2012-08-02 05:42:02.748915
29620	4	1-8231	29618	39	100	861	97	2012-08-02 05:42:02.748915
29621	4	1-3918	29618	100	27	352	44	2012-08-02 05:42:02.748915
29618	4	1-8841	29618	27	100	766	47	2012-08-02 05:42:02.748915
29623	5	1-6798	29622	11	74	1215	57	2012-08-02 05:42:02.748915
29624	5	1-7821	29622	74	53	1479	43	2012-08-02 05:42:02.748915
29625	5	1-5515	29622	53	74	637	62	2012-08-02 05:42:02.748915
29626	5	1-7973	29622	74	27	794	44	2012-08-02 05:42:02.748915
29622	5	1-8841	29622	27	11	1760	47	2012-08-02 05:42:02.748915
29628	4	1-5088	29627	4	26	135	65	2012-08-02 05:42:02.748915
29629	4	1-4530	29627	26	74	51	62	2012-08-02 05:42:02.748915
29630	4	1-7973	29627	74	27	64	44	2012-08-02 05:42:02.748915
29627	4	1-8841	29627	27	4	147	47	2012-08-02 05:42:02.748915
29631	1	1-1721	\N	71	71	457	57	2012-08-02 05:42:02.748915
29632	1	1-3977	\N	12	12	290	62	2012-08-02 05:42:02.748915
29633	1	1-4143	\N	50	50	18	81	2012-08-02 05:42:02.748915
29635	4	1-7251	29634	59	52	711	49	2012-08-02 05:42:03.168998
29636	4	1-7301	29634	52	28	1355	88	2012-08-02 05:42:03.168998
29637	4	1-8795	29634	28	91	817	64	2012-08-02 05:42:03.168998
29634	4	1-8842	29634	91	59	1249	30	2012-08-02 05:42:03.168998
29639	3	1-6733	29638	84	28	3611	88	2012-08-02 05:42:03.168998
29640	3	1-8795	29638	28	91	2187	64	2012-08-02 05:42:03.168998
29638	3	1-8842	29638	91	84	3360	30	2012-08-02 05:42:03.168998
29642	3	1-8635	29641	40	98	799	67	2012-08-02 05:42:03.975621
29643	3	1-8718	29641	98	81	3918	60	2012-08-02 05:42:03.975621
29641	3	1-8844	29641	81	40	1037	3	2012-08-02 05:42:03.975621
29645	3	1-8544	29644	15	12	3287	68	2012-08-02 05:42:04.781867
29646	3	1-8559	29644	12	28	951	12	2012-08-02 05:42:04.781867
29644	3	1-8846	29644	28	15	5585	11	2012-08-02 05:42:04.781867
29648	4	1-8412	29647	53	42	62	82	2012-08-02 05:42:04.781867
29649	4	1-5519	29647	42	53	12	89	2012-08-02 05:42:04.781867
29650	4	1-8387	29647	53	28	9	12	2012-08-02 05:42:04.781867
29647	4	1-8846	29647	28	53	89	11	2012-08-02 05:42:04.781867
29652	2	1-8127	29651	70	31	3952	28	2012-08-02 05:42:05.17983
29651	2	1-8847	29651	31	70	6720	34	2012-08-02 05:42:05.17983
29654	5	1-8698	29653	8	60	2378	97	2012-08-02 05:42:05.589161
29655	5	1-8513	29653	60	29	1909	22	2012-08-02 05:42:05.589161
29656	5	1-8361	29653	29	62	1035	52	2012-08-02 05:42:05.589161
29657	5	1-8617	29653	62	86	1404	30	2012-08-02 05:42:05.589161
29653	5	1-8848	29653	86	8	4272	92	2012-08-02 05:42:05.589161
29659	4	1-6492	29658	30	28	1338	88	2012-08-02 05:42:07.202739
29660	4	1-8795	29658	28	90	942	64	2012-08-02 05:42:07.202739
29661	4	1-8149	29658	90	29	974	17	2012-08-02 05:42:07.202739
29658	4	1-8852	29658	29	30	898	77	2012-08-02 05:42:07.202739
29663	4	1-7033	29662	18	59	429	55	2012-08-02 05:42:07.202739
29664	4	1-8822	29662	59	90	1697	94	2012-08-02 05:42:07.202739
29665	4	1-8836	29662	90	29	930	17	2012-08-02 05:42:07.202739
29662	4	1-8852	29662	29	18	876	77	2012-08-02 05:42:07.202739
29667	4	1-6263	29666	81	13	51	89	2012-08-02 05:42:07.202739
29668	4	1-7433	29666	13	95	40	83	2012-08-02 05:42:07.202739
29669	4	1-8634	29666	95	29	65	17	2012-08-02 05:42:07.202739
29666	4	1-8852	29666	29	81	62	77	2012-08-02 05:42:07.202739
29671	5	1-8039	29670	73	82	159	56	2012-08-02 05:42:08.009346
29672	5	1-8650	29670	82	95	312	76	2012-08-02 05:42:08.009346
29673	5	1-7759	29670	95	12	95	96	2012-08-02 05:42:08.009346
29674	5	1-8539	29670	12	75	106	78	2012-08-02 05:42:08.009346
29670	5	1-8854	29670	75	73	151	23	2012-08-02 05:42:08.009346
29676	4	1-8699	29675	36	84	2362	71	2012-08-02 05:42:10.450766
29677	4	1-7585	29675	84	48	2270	4	2012-08-02 05:42:10.450766
29678	4	1-8810	29675	48	24	1462	18	2012-08-02 05:42:10.450766
29675	4	1-8860	29675	24	36	2186	58	2012-08-02 05:42:10.450766
29680	4	1-8039	29679	73	82	64	56	2012-08-02 05:42:11.26841
29681	4	1-8650	29679	82	81	124	76	2012-08-02 05:42:11.26841
29682	4	1-7024	29679	81	92	73	95	2012-08-02 05:42:11.26841
29679	4	1-8862	29679	92	73	62	23	2012-08-02 05:42:11.26841
29684	4	1-3745	29683	69	34	1243	46	2012-08-02 05:42:11.26841
29685	4	1-8492	29683	34	41	3138	5	2012-08-02 05:42:11.26841
29686	4	1-5788	29683	41	92	1596	95	2012-08-02 05:42:11.26841
29683	4	1-8862	29683	92	69	1528	23	2012-08-02 05:42:11.26841
29688	2	1-7585	29687	84	49	4585	4	2012-08-02 05:42:12.914584
29687	2	1-8866	29687	49	84	3761	71	2012-08-02 05:42:12.914584
29690	4	1-7923	29689	49	49	1011	15	2012-08-02 05:42:13.323242
29691	4	1-7747	29689	49	46	2009	2	2012-08-02 05:42:13.323242
29692	4	1-8334	29689	46	59	885	31	2012-08-02 05:42:13.323242
29689	4	1-8867	29689	59	49	1275	62	2012-08-02 05:42:13.323242
29694	4	1-7169	29693	28	85	618	68	2012-08-02 05:42:13.323242
29695	4	1-8719	29693	85	74	387	7	2012-08-02 05:42:13.323242
29696	4	1-8707	29693	74	59	503	31	2012-08-02 05:42:13.323242
29693	4	1-8867	29693	59	28	753	62	2012-08-02 05:42:13.323242
29698	4	1-7923	29697	49	68	75	15	2012-08-02 05:42:13.323242
29699	4	1-8751	29697	68	62	173	87	2012-08-02 05:42:13.323242
29700	4	1-7277	29697	62	59	64	31	2012-08-02 05:42:13.323242
29697	4	1-8867	29697	59	49	98	62	2012-08-02 05:42:13.323242
29701	1	1-3779	\N	56	56	4186	31	2012-08-02 05:42:13.323242
29703	4	1-3661	29702	31	81	2270	50	2012-08-02 05:42:14.118748
29704	4	1-8688	29702	81	83	3140	86	2012-08-02 05:42:14.118748
29705	4	1-8388	29702	83	9	5830	88	2012-08-02 05:42:14.118748
29702	4	1-8869	29702	9	31	7350	70	2012-08-02 05:42:14.118748
29707	4	1-8854	29706	75	5	972	23	2012-08-02 05:42:14.505469
29708	4	1-5904	29706	5	4	492	54	2012-08-02 05:42:14.505469
29709	4	1-8837	29706	4	42	736	58	2012-08-02 05:42:14.505469
29706	4	1-8870	29706	42	75	991	78	2012-08-02 05:42:14.505469
29711	3	1-8854	29710	75	55	3568	23	2012-08-02 05:42:14.505469
29712	3	1-8705	29710	55	42	1541	58	2012-08-02 05:42:14.505469
29710	3	1-8870	29710	42	75	2747	78	2012-08-02 05:42:14.505469
29714	4	1-8321	29713	31	95	1431	83	2012-08-02 05:42:14.505469
29715	4	1-8634	29713	95	61	1976	17	2012-08-02 05:42:14.505469
29716	4	1-7800	29713	61	42	1635	58	2012-08-02 05:42:14.505469
29713	4	1-8870	29713	42	31	3066	78	2012-08-02 05:42:14.505469
29718	4	1-8676	29717	52	14	3102	14	2012-08-02 05:42:14.505469
29719	4	1-8853	29717	14	93	3569	2	2012-08-02 05:42:14.505469
29720	4	1-8304	29717	93	42	1034	58	2012-08-02 05:42:14.505469
29717	4	1-8870	29717	42	52	2208	78	2012-08-02 05:42:14.505469
29722	5	1-8089	29721	39	43	613	56	2012-08-02 05:42:16.549488
29723	5	1-4370	29721	43	78	442	100	2012-08-02 05:42:16.549488
29724	5	1-8825	29721	78	94	293	11	2012-08-02 05:42:16.549488
29725	5	1-7728	29721	94	85	306	66	2012-08-02 05:42:16.549488
29721	5	1-8875	29721	85	39	303	35	2012-08-02 05:42:16.549488
29727	5	1-8089	29726	39	51	1942	56	2012-08-02 05:42:16.549488
29728	5	1-5400	29726	51	78	1368	100	2012-08-02 05:42:16.549488
29729	5	1-8825	29726	78	94	913	11	2012-08-02 05:42:16.549488
29730	5	1-7728	29726	94	85	957	66	2012-08-02 05:42:16.549488
29726	5	1-8875	29726	85	39	955	35	2012-08-02 05:42:16.549488
29731	1	1-724	\N	34	34	1091	100	2012-08-02 05:42:16.549488
29733	3	1-8875	29732	85	42	1770	35	2012-08-02 05:42:16.969229
29734	3	1-6590	29732	42	30	519	29	2012-08-02 05:42:16.969229
29732	3	1-8876	29732	30	85	5373	66	2012-08-02 05:42:16.969229
29736	3	1-8477	29735	85	62	1052	80	2012-08-02 05:42:18.617401
29737	3	1-7880	29735	62	80	732	97	2012-08-02 05:42:18.617401
29735	3	1-8880	29735	80	85	474	45	2012-08-02 05:42:18.617401
29739	5	1-8477	29738	85	99	1926	80	2012-08-02 05:42:18.617401
29740	5	1-5966	29738	99	85	682	63	2012-08-02 05:42:18.617401
29741	5	1-8649	29738	85	20	773	27	2012-08-02 05:42:18.617401
29742	5	1-8068	29738	20	80	933	97	2012-08-02 05:42:18.617401
29738	5	1-8880	29738	80	85	688	45	2012-08-02 05:42:18.617401
29744	3	1-8725	29743	25	20	4349	83	2012-08-02 05:42:18.617401
29745	3	1-8262	29743	20	80	4073	97	2012-08-02 05:42:18.617401
29743	3	1-8880	29743	80	25	3504	45	2012-08-02 05:42:18.617401
29747	2	1-8599	29746	66	19	695	74	2012-08-02 05:42:20.230211
29746	2	1-8884	29746	19	66	902	24	2012-08-02 05:42:20.230211
29749	5	1-4885	29748	67	8	4941	10	2012-08-02 05:42:20.230211
29750	5	1-8879	29748	8	90	4901	45	2012-08-02 05:42:20.230211
29751	5	1-8759	29748	90	93	2726	8	2012-08-02 05:42:20.230211
29752	5	1-8798	29748	93	19	3712	74	2012-08-02 05:42:20.230211
29748	5	1-8884	29748	19	67	4979	24	2012-08-02 05:42:20.230211
29754	5	1-7565	29753	60	57	846	79	2012-08-02 05:42:20.230211
29755	5	1-8784	29753	57	83	1744	88	2012-08-02 05:42:20.230211
29756	5	1-6755	29753	83	93	306	8	2012-08-02 05:42:20.230211
29757	5	1-8798	29753	93	19	444	74	2012-08-02 05:42:20.230211
29753	5	1-8884	29753	19	60	633	24	2012-08-02 05:42:20.230211
29759	4	1-5429	29758	85	78	1057	64	2012-08-02 05:42:20.230211
29760	4	1-8616	29758	78	17	1359	23	2012-08-02 05:42:20.230211
29761	4	1-6781	29758	17	19	1443	74	2012-08-02 05:42:20.230211
29758	4	1-8884	29758	19	85	2181	24	2012-08-02 05:42:20.230211
29763	4	1-8511	29762	9	73	863	82	2012-08-02 05:42:21.478721
29764	4	1-7555	29762	73	68	572	15	2012-08-02 05:42:21.478721
29765	4	1-8751	29762	68	2	1018	87	2012-08-02 05:42:21.478721
29762	4	1-8887	29762	2	9	1047	84	2012-08-02 05:42:21.478721
29767	4	1-8511	29766	9	15	4770	82	2012-08-02 05:42:21.478721
29768	4	1-8821	29766	15	13	3989	23	2012-08-02 05:42:21.478721
29769	4	1-7428	29766	13	2	3544	87	2012-08-02 05:42:21.478721
29766	4	1-8887	29766	2	9	4592	84	2012-08-02 05:42:21.478721
29771	4	1-8511	29770	9	15	12	82	2012-08-02 05:42:21.478721
29772	4	1-8821	29770	15	98	11	23	2012-08-02 05:42:21.478721
29773	4	1-8660	29770	98	2	7	87	2012-08-02 05:42:21.478721
29770	4	1-8887	29770	2	9	10	84	2012-08-02 05:42:21.478721
29775	4	1-7459	29774	46	79	2835	55	2012-08-02 05:42:21.876587
29776	4	1-8791	29774	79	57	2970	91	2012-08-02 05:42:21.876587
29777	4	1-8424	29774	57	81	2776	34	2012-08-02 05:42:21.876587
29774	4	1-8888	29774	81	46	2654	19	2012-08-02 05:42:21.876587
29779	3	1-6390	29778	60	68	631	79	2012-08-02 05:42:22.694059
29780	3	1-8689	29778	68	40	640	85	2012-08-02 05:42:22.694059
29778	3	1-8890	29778	40	60	8213	65	2012-08-02 05:42:22.694059
29782	5	1-8557	29781	75	81	71	51	2012-08-02 05:42:23.898301
29783	5	1-3265	29781	81	32	8	3	2012-08-02 05:42:23.898301
29784	5	1-8102	29781	32	25	18	21	2012-08-02 05:42:23.898301
29785	5	1-8785	29781	25	94	22	36	2012-08-02 05:42:23.898301
29781	5	1-8893	29781	94	75	31	55	2012-08-02 05:42:23.898301
29787	4	1-8716	29786	18	85	433	45	2012-08-02 05:42:24.307075
29788	4	1-8477	29786	85	82	943	80	2012-08-02 05:42:24.307075
29789	4	1-8525	29786	82	64	1903	37	2012-08-02 05:42:24.307075
29786	4	1-8894	29786	64	18	2678	69	2012-08-02 05:42:24.307075
29791	4	1-8838	29790	61	20	1500	82	2012-08-02 05:42:24.307075
29792	4	1-5830	29790	20	82	1347	80	2012-08-02 05:42:24.307075
29793	4	1-8525	29790	82	64	2767	37	2012-08-02 05:42:24.307075
29790	4	1-8894	29790	64	61	3967	69	2012-08-02 05:42:24.307075
29795	2	1-8383	29794	82	61	5749	4	2012-08-02 05:42:25.124672
29794	2	1-8896	29794	61	82	9685	12	2012-08-02 05:42:25.124672
29797	4	1-7868	29796	93	93	487	55	2012-08-02 05:42:25.533526
29798	4	1-5794	29796	93	34	415	46	2012-08-02 05:42:25.533526
29799	4	1-8492	29796	34	35	555	5	2012-08-02 05:42:25.533526
29796	4	1-8897	29796	35	93	1076	72	2012-08-02 05:42:25.533526
29801	2	1-2526	29800	16	35	1277	5	2012-08-02 05:42:25.533526
29800	2	1-8897	29800	35	16	2961	72	2012-08-02 05:42:25.533526
29803	2	1-3727	29802	45	35	1863	5	2012-08-02 05:42:25.533526
29802	2	1-8897	29802	35	45	5321	72	2012-08-02 05:42:25.533526
29805	3	1-8787	29804	39	67	3175	100	2012-08-02 05:42:26.72665
29806	3	1-8565	29804	67	92	2035	9	2012-08-02 05:42:26.72665
29804	3	1-8900	29804	92	39	6669	17	2012-08-02 05:42:26.72665
29808	3	1-8699	29807	36	28	1900	71	2012-08-02 05:42:27.1465
29809	3	1-3943	29807	28	28	1461	46	2012-08-02 05:42:27.1465
29807	3	1-8901	29807	28	36	1776	58	2012-08-02 05:42:27.1465
29811	4	1-8757	29810	35	66	555	73	2012-08-02 05:42:29.145976
29812	4	1-8337	29810	66	39	406	53	2012-08-02 05:42:29.145976
29813	4	1-8439	29810	39	89	145	93	2012-08-02 05:42:29.145976
29810	4	1-8906	29810	89	35	442	99	2012-08-02 05:42:29.145976
29815	3	1-8748	29814	27	66	3301	98	2012-08-02 05:42:29.145976
29816	3	1-8555	29814	66	89	1653	93	2012-08-02 05:42:29.145976
29814	3	1-8906	29814	89	27	5096	99	2012-08-02 05:42:29.145976
29818	4	1-8785	29817	25	23	2197	36	2012-08-02 05:42:29.565999
29819	4	1-8378	29817	23	15	1796	82	2012-08-02 05:42:29.565999
29820	4	1-8821	29817	15	76	1497	23	2012-08-02 05:42:29.565999
29817	4	1-8907	29817	76	25	1879	21	2012-08-02 05:42:29.565999
29822	5	1-8882	29821	53	30	175	77	2012-08-02 05:42:29.565999
29823	5	1-6492	29821	30	28	260	88	2012-08-02 05:42:29.565999
29824	5	1-8795	29821	28	78	182	64	2012-08-02 05:42:29.565999
29825	5	1-8616	29821	78	76	226	23	2012-08-02 05:42:29.565999
29821	5	1-8907	29821	76	53	320	21	2012-08-02 05:42:29.565999
29827	4	1-6970	29826	46	1	814	33	2012-08-02 05:42:30.383674
29828	4	1-8403	29826	1	94	469	61	2012-08-02 05:42:30.383674
29829	4	1-8827	29826	94	45	455	84	2012-08-02 05:42:30.383674
29826	4	1-8909	29826	45	46	2785	98	2012-08-02 05:42:30.383674
29831	3	1-6924	29830	91	12	1686	68	2012-08-02 05:42:31.201373
29832	3	1-8559	29830	12	96	1054	12	2012-08-02 05:42:31.201373
29830	3	1-8911	29830	96	91	4031	75	2012-08-02 05:42:31.201373
29834	4	1-8762	29833	78	19	2742	38	2012-08-02 05:42:31.598878
29835	4	1-8050	29833	19	8	2656	48	2012-08-02 05:42:31.598878
29836	4	1-8373	29833	8	79	1780	100	2012-08-02 05:42:31.598878
29833	4	1-8912	29833	79	78	1902	36	2012-08-02 05:42:31.598878
29838	4	1-8893	29837	94	82	767	55	2012-08-02 05:42:31.598878
29839	4	1-4767	29837	82	98	319	61	2012-08-02 05:42:31.598878
29840	4	1-7969	29837	98	79	477	100	2012-08-02 05:42:31.598878
29837	4	1-8912	29837	79	94	527	36	2012-08-02 05:42:31.598878
29842	4	1-8848	29841	86	52	3467	92	2012-08-02 05:42:32.836549
29843	4	1-8423	29841	52	83	2612	25	2012-08-02 05:42:32.836549
29844	4	1-8610	29841	83	27	1258	51	2012-08-02 05:42:32.836549
29841	4	1-8915	29841	27	86	1031	30	2012-08-02 05:42:32.836549
29846	4	1-7682	29845	35	22	294	24	2012-08-02 05:42:33.642632
29847	4	1-8694	29845	22	30	224	76	2012-08-02 05:42:33.642632
29848	4	1-7540	29845	30	29	106	92	2012-08-02 05:42:33.642632
29845	4	1-8917	29845	29	35	2391	83	2012-08-02 05:42:33.642632
29850	3	1-7849	29849	40	23	240	14	2012-08-02 05:42:34.040536
29851	3	1-8098	29849	23	81	71	13	2012-08-02 05:42:34.040536
29849	3	1-8918	29849	81	40	165	16	2012-08-02 05:42:34.040536
29853	4	1-7933	29852	73	61	1085	63	2012-08-02 05:42:37.29934
29854	4	1-8723	29852	61	32	967	24	2012-08-02 05:42:37.29934
29855	4	1-4664	29852	32	22	833	59	2012-08-02 05:42:37.29934
29852	4	1-8926	29852	22	73	1201	33	2012-08-02 05:42:37.29934
29857	4	1-8174	29856	5	19	753	74	2012-08-02 05:42:37.29934
29858	4	1-8884	29856	19	32	928	24	2012-08-02 05:42:37.29934
29859	4	1-4664	29856	32	22	820	59	2012-08-02 05:42:37.29934
29856	4	1-8926	29856	22	5	1213	33	2012-08-02 05:42:37.29934
29861	4	1-8403	29860	1	31	2394	61	2012-08-02 05:42:37.29934
29862	4	1-7310	29860	31	27	1694	25	2012-08-02 05:42:37.29934
29863	4	1-4564	29860	27	22	1437	59	2012-08-02 05:42:37.29934
29860	4	1-8926	29860	22	1	2554	33	2012-08-02 05:42:37.29934
29865	4	1-8403	29864	1	69	653	61	2012-08-02 05:42:37.29934
29866	4	1-7208	29864	69	39	466	99	2012-08-02 05:42:37.29934
29867	4	1-8731	29864	39	22	386	59	2012-08-02 05:42:37.29934
29864	4	1-8926	29864	22	1	691	33	2012-08-02 05:42:37.29934
29869	4	1-8403	29868	1	28	865	61	2012-08-02 05:42:37.29934
29870	4	1-5483	29868	28	71	554	48	2012-08-02 05:42:37.29934
29871	4	1-7954	29868	71	22	501	59	2012-08-02 05:42:37.29934
29868	4	1-8926	29868	22	1	906	33	2012-08-02 05:42:37.29934
29872	1	1-538	\N	34	34	682	59	2012-08-02 05:42:37.29934
29874	3	1-8127	29873	70	53	974	28	2012-08-02 05:42:37.741642
29875	3	1-8920	29873	53	59	995	3	2012-08-02 05:42:37.741642
29873	3	1-8927	29873	59	70	1479	34	2012-08-02 05:42:37.741642
29877	5	1-8888	29876	81	89	3425	19	2012-08-02 05:42:37.741642
29878	5	1-5909	29876	89	19	3108	43	2012-08-02 05:42:37.741642
29879	5	1-8374	29876	19	53	2032	28	2012-08-02 05:42:37.741642
29880	5	1-8920	29876	53	59	2309	3	2012-08-02 05:42:37.741642
29876	5	1-8927	29876	59	81	3816	34	2012-08-02 05:42:37.741642
29881	1	1-5909	\N	89	89	2699	43	2012-08-02 05:42:37.741642
29883	4	1-7948	29882	76	97	1692	98	2012-08-02 05:42:39.763243
29884	4	1-6336	29882	97	28	895	46	2012-08-02 05:42:39.763243
29885	4	1-8901	29882	28	84	1083	58	2012-08-02 05:42:39.763243
29882	4	1-8932	29882	84	76	4615	92	2012-08-02 05:42:39.763243
29887	4	1-8698	29886	8	97	1291	97	2012-08-02 05:42:39.763243
29888	4	1-8146	29886	97	82	1194	54	2012-08-02 05:42:39.763243
29889	4	1-7920	29886	82	84	428	58	2012-08-02 05:42:39.763243
29886	4	1-8932	29886	84	8	2007	92	2012-08-02 05:42:39.763243
29891	2	1-8557	29890	75	65	8300	51	2012-08-02 05:42:40.569807
29890	2	1-8934	29890	65	75	4998	55	2012-08-02 05:42:40.569807
29893	3	1-8589	29892	28	17	996	8	2012-08-02 05:42:42.204683
29894	3	1-6191	29892	17	14	650	60	2012-08-02 05:42:42.204683
29892	3	1-8938	29892	14	28	1957	61	2012-08-02 05:42:42.204683
29896	4	1-6424	29895	79	59	666	31	2012-08-02 05:42:42.204683
29897	4	1-8867	29895	59	9	688	62	2012-08-02 05:42:42.204683
29898	4	1-7449	29895	9	14	398	60	2012-08-02 05:42:42.204683
29895	4	1-8938	29895	14	79	1344	61	2012-08-02 05:42:42.204683
29900	3	1-8867	29899	59	14	6289	62	2012-08-02 05:42:43.044368
29901	3	1-7629	29899	14	43	2217	12	2012-08-02 05:42:43.044368
29899	3	1-8940	29899	43	59	4016	31	2012-08-02 05:42:43.044368
29903	4	1-8810	29902	48	1	495	18	2012-08-02 05:42:45.088513
29904	4	1-8145	29902	1	47	227	33	2012-08-02 05:42:45.088513
29905	4	1-7892	29902	47	7	161	49	2012-08-02 05:42:45.088513
29902	4	1-8945	29902	7	48	662	4	2012-08-02 05:42:45.088513
29907	4	1-5910	29906	71	53	2180	28	2012-08-02 05:42:45.088513
29908	4	1-8920	29906	53	100	2783	3	2012-08-02 05:42:45.088513
29909	4	1-7314	29906	100	7	1162	49	2012-08-02 05:42:45.088513
29906	4	1-8945	29906	7	71	5296	4	2012-08-02 05:42:45.088513
29911	3	1-8183	29910	21	38	3591	38	2012-08-02 05:42:45.485942
29912	3	1-4144	29910	38	87	2270	80	2012-08-02 05:42:45.485942
29910	3	1-8946	29910	87	21	3415	62	2012-08-02 05:42:45.485942
29914	4	1-7857	29913	22	78	2049	36	2012-08-02 05:42:45.485942
29915	4	1-8762	29913	78	38	2731	38	2012-08-02 05:42:45.485942
29916	4	1-4144	29913	38	87	2061	80	2012-08-02 05:42:45.485942
29913	4	1-8946	29913	87	22	3701	62	2012-08-02 05:42:45.485942
29918	4	1-6203	29917	28	42	1793	87	2012-08-02 05:42:45.485942
29919	4	1-8142	29917	42	38	1455	38	2012-08-02 05:42:45.485942
29920	4	1-4144	29917	38	87	1256	80	2012-08-02 05:42:45.485942
29917	4	1-8946	29917	87	28	2580	62	2012-08-02 05:42:45.485942
29922	4	1-7562	29921	38	14	1565	14	2012-08-02 05:42:46.701345
29923	4	1-8853	29921	14	93	1686	2	2012-08-02 05:42:46.701345
29924	4	1-8304	29921	93	32	457	58	2012-08-02 05:42:46.701345
29921	4	1-8949	29921	32	38	641	12	2012-08-02 05:42:46.701345
29926	3	1-7848	29925	100	61	4082	17	2012-08-02 05:42:46.701345
29927	3	1-7800	29925	61	32	3784	58	2012-08-02 05:42:46.701345
29925	3	1-8949	29925	32	100	5577	12	2012-08-02 05:42:46.701345
29929	5	1-8365	29928	57	76	556	15	2012-08-02 05:42:47.530224
29930	5	1-8598	29928	76	98	464	56	2012-08-02 05:42:47.530224
29931	5	1-7602	29928	98	27	388	51	2012-08-02 05:42:47.530224
29932	5	1-8915	29928	27	70	323	30	2012-08-02 05:42:47.530224
29928	5	1-8951	29928	70	57	490	64	2012-08-02 05:42:47.530224
29934	5	1-8616	29933	78	31	1229	23	2012-08-02 05:42:47.530224
29935	5	1-7928	29933	31	35	815	81	2012-08-02 05:42:47.530224
29936	5	1-5851	29933	35	27	681	51	2012-08-02 05:42:47.530224
29937	5	1-8915	29933	27	70	600	30	2012-08-02 05:42:47.530224
29933	5	1-8951	29933	70	78	961	64	2012-08-02 05:42:47.530224
29939	4	1-8940	29938	43	7	3075	31	2012-08-02 05:42:49.574119
29940	4	1-8468	29938	7	97	3438	2	2012-08-02 05:42:49.574119
29941	4	1-8296	29938	97	29	1322	81	2012-08-02 05:42:49.574119
29938	4	1-8956	29938	29	43	1678	12	2012-08-02 05:42:49.574119
29943	4	1-7848	29942	100	61	2205	17	2012-08-02 05:42:49.574119
29944	4	1-7800	29942	61	98	2153	58	2012-08-02 05:42:49.574119
29945	4	1-7694	29942	98	29	2068	81	2012-08-02 05:42:49.574119
29942	4	1-8956	29942	29	100	2860	12	2012-08-02 05:42:49.574119
29947	4	1-4987	29946	98	70	1328	30	2012-08-02 05:42:52.028554
29948	4	1-8951	29946	70	57	1195	64	2012-08-02 05:42:52.028554
29949	4	1-8365	29946	57	25	806	15	2012-08-02 05:42:52.028554
29946	4	1-8962	29946	25	98	4316	12	2012-08-02 05:42:52.028554
29951	4	1-6252	29950	13	77	1889	97	2012-08-02 05:42:52.028554
29952	4	1-8788	29950	77	60	1108	79	2012-08-02 05:42:52.028554
29953	4	1-7264	29950	60	25	551	15	2012-08-02 05:42:52.028554
29950	4	1-8962	29950	25	13	3422	12	2012-08-02 05:42:52.028554
29955	4	1-6209	29954	61	54	754	59	2012-08-02 05:42:52.028554
29956	4	1-7060	29954	54	66	381	14	2012-08-02 05:42:52.028554
29957	4	1-8935	29954	66	25	310	15	2012-08-02 05:42:52.028554
29954	4	1-8962	29954	25	61	2018	12	2012-08-02 05:42:52.028554
29959	4	1-8131	29958	26	94	726	36	2012-08-02 05:42:52.494951
29960	4	1-8893	29958	94	51	961	55	2012-08-02 05:42:52.494951
29961	4	1-1428	29958	51	22	414	90	2012-08-02 05:42:52.494951
29958	4	1-8963	29958	22	26	891	11	2012-08-02 05:42:52.494951
29963	4	1-8059	29962	84	19	340	38	2012-08-02 05:42:52.494951
29964	4	1-8050	29962	19	18	333	48	2012-08-02 05:42:52.494951
29965	4	1-8020	29962	18	22	211	90	2012-08-02 05:42:52.494951
29962	4	1-8963	29962	22	84	489	11	2012-08-02 05:42:52.494951
29967	5	1-8107	29966	8	11	1223	81	2012-08-02 05:42:52.494951
29968	5	1-8790	29966	11	35	272	99	2012-08-02 05:42:52.494951
29969	5	1-8757	29966	35	35	569	73	2012-08-02 05:42:52.494951
29970	5	1-8118	29966	35	22	549	90	2012-08-02 05:42:52.494951
29966	5	1-8963	29966	22	8	1286	11	2012-08-02 05:42:52.494951
29972	3	1-5405	29971	89	35	1582	73	2012-08-02 05:42:52.494951
29973	3	1-8118	29971	35	22	1487	90	2012-08-02 05:42:52.494951
29971	3	1-8963	29971	22	89	3548	11	2012-08-02 05:42:52.494951
29975	5	1-434	29974	86	25	1279	51	2012-08-02 05:42:52.494951
29976	5	1-8648	29974	25	46	1042	98	2012-08-02 05:42:52.494951
29977	5	1-7796	29974	46	35	1007	73	2012-08-02 05:42:52.494951
29978	5	1-8118	29974	35	22	975	90	2012-08-02 05:42:52.494951
29974	5	1-8963	29974	22	86	2397	11	2012-08-02 05:42:52.494951
29980	4	1-7912	29979	15	60	414	97	2012-08-02 05:42:53.332331
29981	4	1-8513	29979	60	30	338	22	2012-08-02 05:42:53.332331
29982	4	1-8948	29979	30	23	270	59	2012-08-02 05:42:53.332331
29979	4	1-8965	29979	23	15	490	39	2012-08-02 05:42:53.332331
29984	3	1-8350	29983	55	77	1641	9	2012-08-02 05:42:53.332331
29985	3	1-6488	29983	77	23	1975	59	2012-08-02 05:42:53.332331
29983	3	1-8965	29983	23	55	4047	39	2012-08-02 05:42:53.332331
29987	4	1-6611	29986	95	81	2238	42	2012-08-02 05:42:53.332331
29988	4	1-8244	29986	81	19	1777	94	2012-08-02 05:42:53.332331
29989	4	1-6150	29986	19	23	1062	59	2012-08-02 05:42:53.332331
29986	4	1-8965	29986	23	95	2232	39	2012-08-02 05:42:53.332331
29991	4	1-5655	29990	75	86	2921	25	2012-08-02 05:42:54.149933
29992	4	1-6031	29990	86	6	1921	93	2012-08-02 05:42:54.149933
29993	4	1-8883	29990	6	86	5607	40	2012-08-02 05:42:54.149933
29990	4	1-8967	29990	86	75	5526	68	2012-08-02 05:42:54.149933
29995	4	1-7908	29994	92	42	2556	70	2012-08-02 05:42:54.149933
29996	4	1-4241	29994	42	6	1076	93	2012-08-02 05:42:54.149933
29997	4	1-8883	29994	6	86	3355	40	2012-08-02 05:42:54.149933
29994	4	1-8967	29994	86	92	3325	68	2012-08-02 05:42:54.149933
29999	3	1-7657	29998	1	28	1911	67	2012-08-02 05:42:55.342948
30000	3	1-6674	29998	28	91	832	42	2012-08-02 05:42:55.342948
29998	3	1-8970	29998	91	1	2685	84	2012-08-02 05:42:55.342948
30002	4	1-8529	30001	7	84	3721	70	2012-08-02 05:42:56.171453
30003	4	1-8891	30001	84	48	1504	18	2012-08-02 05:42:56.171453
30004	4	1-8572	30001	48	27	2076	42	2012-08-02 05:42:56.171453
30001	4	1-8972	30001	27	7	4874	67	2012-08-02 05:42:56.171453
30006	2	1-6674	30005	28	27	679	42	2012-08-02 05:42:56.171453
30005	2	1-8972	30005	27	28	1737	67	2012-08-02 05:42:56.171453
30008	3	1-8528	30007	100	51	584	31	2012-08-02 05:42:56.171453
30009	3	1-8637	30007	51	27	329	42	2012-08-02 05:42:56.171453
30007	3	1-8972	30007	27	100	940	67	2012-08-02 05:42:56.171453
30011	4	1-8644	30010	56	82	986	80	2012-08-02 05:42:56.60261
30012	4	1-8525	30010	82	59	2061	37	2012-08-02 05:42:56.60261
30013	4	1-8515	30010	59	4	1392	56	2012-08-02 05:42:56.60261
30010	4	1-8973	30010	4	56	1340	12	2012-08-02 05:42:56.60261
30015	4	1-7209	30014	84	60	1164	79	2012-08-02 05:42:57.000279
30016	4	1-7264	30014	60	20	765	15	2012-08-02 05:42:57.000279
30017	4	1-8930	30014	20	29	4096	53	2012-08-02 05:42:57.000279
30014	4	1-8974	30014	29	84	1787	66	2012-08-02 05:42:57.000279
30019	4	1-4022	30018	61	81	1005	50	2012-08-02 05:42:57.409131
30020	4	1-8688	30018	81	4	1657	86	2012-08-02 05:42:57.409131
30021	4	1-8480	30018	4	44	2885	48	2012-08-02 05:42:57.409131
30018	4	1-8975	30018	44	61	1900	30	2012-08-02 05:42:57.409131
30023	4	1-8951	30022	70	78	2571	64	2012-08-02 05:42:57.409131
30024	4	1-8616	30022	78	73	3144	23	2012-08-02 05:42:57.409131
30025	4	1-8075	30022	73	44	2233	48	2012-08-02 05:42:57.409131
30022	4	1-8975	30022	44	70	1677	30	2012-08-02 05:42:57.409131
30027	5	1-434	30026	86	27	258	51	2012-08-02 05:42:58.23806
30028	5	1-8915	30026	27	59	228	30	2012-08-02 05:42:58.23806
30029	5	1-7251	30026	59	7	156	49	2012-08-02 05:42:58.23806
30030	5	1-8945	30026	7	44	707	4	2012-08-02 05:42:58.23806
30026	5	1-8977	30026	44	86	469	11	2012-08-02 05:42:58.23806
30032	4	1-8480	30031	4	100	1857	48	2012-08-02 05:42:59.861566
30033	4	1-8385	30031	100	7	320	49	2012-08-02 05:42:59.861566
30034	4	1-8945	30031	7	27	1287	4	2012-08-02 05:42:59.861566
30031	4	1-8981	30031	27	4	1006	86	2012-08-02 05:42:59.861566
30036	4	1-8646	30035	15	47	628	33	2012-08-02 05:42:59.861566
30037	4	1-7892	30035	47	7	457	49	2012-08-02 05:42:59.861566
30038	4	1-8945	30035	7	27	1922	4	2012-08-02 05:42:59.861566
30035	4	1-8981	30035	27	15	1569	86	2012-08-02 05:42:59.861566
30040	4	1-8776	30039	47	91	1393	42	2012-08-02 05:42:59.861566
30041	4	1-8970	30039	91	79	4453	84	2012-08-02 05:42:59.861566
30042	4	1-7915	30039	79	27	2791	4	2012-08-02 05:42:59.861566
30039	4	1-8981	30039	27	47	2442	86	2012-08-02 05:42:59.861566
30044	5	1-8131	30043	26	94	1924	36	2012-08-02 05:43:00.27057
30045	5	1-8893	30043	94	75	2848	55	2012-08-02 05:43:00.27057
30046	5	1-8925	30043	75	34	2611	43	2012-08-02 05:43:00.27057
30047	5	1-8950	30043	34	58	4404	57	2012-08-02 05:43:00.27057
30043	5	1-8982	30043	58	26	2110	11	2012-08-02 05:43:00.27057
30049	5	1-3501	30048	86	33	222	38	2012-08-02 05:43:00.70149
30050	5	1-8452	30048	33	20	172	41	2012-08-02 05:43:00.70149
30051	5	1-7414	30048	20	82	163	56	2012-08-02 05:43:00.70149
30052	5	1-8650	30048	82	5	357	76	2012-08-02 05:43:00.70149
30048	5	1-8983	30048	5	86	314	75	2012-08-02 05:43:00.70149
30054	5	1-8786	30053	23	28	705	75	2012-08-02 05:43:01.519339
30055	5	1-8400	30053	28	76	3621	29	2012-08-02 05:43:01.519339
30056	5	1-8556	30053	76	53	756	28	2012-08-02 05:43:01.519339
30057	5	1-8920	30053	53	51	903	3	2012-08-02 05:43:01.519339
30053	5	1-8985	30053	51	23	1319	51	2012-08-02 05:43:01.519339
30059	5	1-8934	30058	65	79	1504	55	2012-08-02 05:43:01.519339
30060	5	1-8791	30058	79	86	1558	91	2012-08-02 05:43:01.519339
30061	5	1-8190	30058	86	53	1001	28	2012-08-02 05:43:01.519339
30062	5	1-8920	30058	53	51	1197	3	2012-08-02 05:43:01.519339
30058	5	1-8985	30058	51	65	1752	51	2012-08-02 05:43:01.519339
30064	5	1-8699	30063	36	38	834	71	2012-08-02 05:43:03.165069
30065	5	1-7614	30063	38	60	768	53	2012-08-02 05:43:03.165069
30066	5	1-8345	30063	60	67	316	100	2012-08-02 05:43:03.165069
30067	5	1-8565	30063	67	13	152	9	2012-08-02 05:43:03.165069
30063	5	1-8989	30063	13	36	922	58	2012-08-02 05:43:03.165069
30069	4	1-8699	30068	36	56	129	71	2012-08-02 05:43:03.165069
30070	4	1-7140	30068	56	67	44	100	2012-08-02 05:43:03.165069
30071	4	1-8565	30068	67	13	22	9	2012-08-02 05:43:03.165069
30068	4	1-8989	30068	13	36	138	58	2012-08-02 05:43:03.165069
30073	2	1-8908	30072	22	13	34	9	2012-08-02 05:43:03.165069
30072	2	1-8989	30072	13	22	303	58	2012-08-02 05:43:03.165069
30075	3	1-8879	30074	8	29	4620	45	2012-08-02 05:43:03.574051
30076	3	1-7478	30074	29	60	1683	69	2012-08-02 05:43:03.574051
30074	3	1-8990	30074	60	8	6430	10	2012-08-02 05:43:03.574051
30078	3	1-8165	30077	82	1	2599	16	2012-08-02 05:43:03.97187
30079	3	1-8738	30077	1	80	3683	85	2012-08-02 05:43:03.97187
30077	3	1-8991	30077	80	82	4065	55	2012-08-02 05:43:03.97187
30081	3	1-5794	30080	93	62	247	46	2012-08-02 05:43:03.97187
30082	3	1-6710	30080	62	80	189	85	2012-08-02 05:43:03.97187
30080	3	1-8991	30080	80	93	258	55	2012-08-02 05:43:03.97187
30084	4	1-2733	30083	1	11	499	33	2012-08-02 05:43:03.97187
30085	4	1-8259	30083	11	15	1343	29	2012-08-02 05:43:03.97187
30086	4	1-8811	30083	15	80	2076	85	2012-08-02 05:43:03.97187
30083	4	1-8991	30083	80	1	2907	55	2012-08-02 05:43:03.97187
30088	4	1-4767	30087	82	51	175	61	2012-08-02 05:43:03.97187
30089	4	1-7890	30087	51	68	123	79	2012-08-02 05:43:03.97187
30090	4	1-8689	30087	68	80	331	85	2012-08-02 05:43:03.97187
30087	4	1-8991	30087	80	82	466	55	2012-08-02 05:43:03.97187
30092	4	1-8000	30091	96	64	74	96	2012-08-02 05:43:03.97187
30093	4	1-8728	30091	64	68	65	79	2012-08-02 05:43:03.97187
30094	4	1-8689	30091	68	80	178	85	2012-08-02 05:43:03.97187
30091	4	1-8991	30091	80	96	255	55	2012-08-02 05:43:03.97187
30096	3	1-7735	30095	7	19	2616	59	2012-08-02 05:43:04.789503
30097	3	1-8238	30095	19	82	2758	5	2012-08-02 05:43:04.789503
30095	3	1-8993	30095	82	7	5429	4	2012-08-02 05:43:04.789503
30099	4	1-6659	30098	43	17	1125	89	2012-08-02 05:43:06.02793
30100	4	1-8663	30098	17	18	597	24	2012-08-02 05:43:06.02793
30101	4	1-5591	30098	18	35	730	50	2012-08-02 05:43:06.02793
30098	4	1-8996	30098	35	43	1670	52	2012-08-02 05:43:06.02793
30103	3	1-8714	30102	8	98	3744	76	2012-08-02 05:43:07.67428
30104	3	1-8069	30102	98	2	5504	54	2012-08-02 05:43:07.67428
30102	3	1-9000	30102	2	8	3298	77	2012-08-02 05:43:07.67428
30106	4	1-8869	30105	9	76	305	70	2012-08-02 05:43:08.105018
30107	4	1-7076	30105	76	44	119	48	2012-08-02 05:43:08.105018
30108	4	1-8975	30105	44	21	69	30	2012-08-02 05:43:08.105018
30105	4	1-9001	30105	21	9	230	88	2012-08-02 05:43:08.105018
30110	5	1-6755	30109	83	93	949	8	2012-08-02 05:43:08.105018
30111	5	1-8798	30109	93	49	1391	74	2012-08-02 05:43:08.105018
30112	5	1-7486	30109	49	27	1453	51	2012-08-02 05:43:08.105018
30113	5	1-8915	30109	27	21	1230	30	2012-08-02 05:43:08.105018
30109	5	1-9001	30109	21	83	5329	88	2012-08-02 05:43:08.105018
30115	5	1-6019	30114	34	88	348	95	2012-08-02 05:43:08.105018
30116	5	1-7356	30114	88	91	143	63	2012-08-02 05:43:08.105018
30117	5	1-8815	30114	91	27	125	51	2012-08-02 05:43:08.105018
30118	5	1-8915	30114	27	21	109	30	2012-08-02 05:43:08.105018
30114	5	1-9001	30114	21	34	492	88	2012-08-02 05:43:08.105018
30119	1	1-6019	\N	34	34	1736	95	2012-08-02 05:43:08.105018
30121	4	1-8607	30120	95	92	7574	95	2012-08-02 05:43:08.524867
30122	4	1-8862	30120	92	5	6930	23	2012-08-02 05:43:08.524867
30123	4	1-5904	30120	5	27	5407	54	2012-08-02 05:43:08.524867
30120	4	1-9002	30120	27	95	3859	99	2012-08-02 05:43:08.524867
30125	4	1-8757	30124	35	35	2718	73	2012-08-02 05:43:08.524867
30126	4	1-8118	30124	35	3	2683	90	2012-08-02 05:43:08.524867
30127	4	1-6652	30124	3	27	1753	54	2012-08-02 05:43:08.524867
30124	4	1-9002	30124	27	35	1272	99	2012-08-02 05:43:08.524867
30128	1	1-1805	\N	67	67	1573	90	2012-08-02 05:43:08.524867
30129	1	1-2185	\N	51	51	276	90	2012-08-02 05:43:08.524867
30131	3	1-8585	30130	61	22	1336	40	2012-08-02 05:43:08.944936
30132	3	1-8302	30130	22	100	259	7	2012-08-02 05:43:08.944936
30130	3	1-9003	30130	100	61	262	75	2012-08-02 05:43:08.944936
30134	3	1-8458	30133	3	90	2650	83	2012-08-02 05:43:08.944936
30135	3	1-8377	30133	90	100	1046	7	2012-08-02 05:43:08.944936
30133	3	1-9003	30133	100	3	1537	75	2012-08-02 05:43:08.944936
30137	4	1-8702	30136	49	25	2415	45	2012-08-02 05:43:08.944936
30138	4	1-8725	30136	25	90	2965	83	2012-08-02 05:43:08.944936
30139	4	1-8377	30136	90	100	1449	7	2012-08-02 05:43:08.944936
30136	4	1-9003	30136	100	49	2637	75	2012-08-02 05:43:08.944936
30141	4	1-7461	30140	88	51	5102	42	2012-08-02 05:43:10.14897
30142	4	1-8470	30140	51	79	1720	26	2012-08-02 05:43:10.14897
30143	4	1-8453	30140	79	27	1889	100	2012-08-02 05:43:10.14897
30140	4	1-9006	30140	27	88	8213	73	2012-08-02 05:43:10.14897
30145	4	1-7750	30144	9	36	3016	90	2012-08-02 05:43:10.955547
30146	4	1-8760	30144	36	50	2491	77	2012-08-02 05:43:10.955547
30147	4	1-7377	30144	50	56	1537	22	2012-08-02 05:43:10.955547
30144	4	1-9008	30144	56	9	7360	57	2012-08-02 05:43:10.955547
30149	5	1-7918	30148	94	6	35	44	2012-08-02 05:43:10.955547
30150	5	1-8613	30148	6	83	20	1	2012-08-02 05:43:10.955547
30151	5	1-7530	30148	83	99	15	33	2012-08-02 05:43:10.955547
30152	5	1-8220	30148	99	56	15	22	2012-08-02 05:43:10.955547
30148	5	1-9008	30148	56	94	74	57	2012-08-02 05:43:10.955547
30154	5	1-6975	30153	64	8	754	92	2012-08-02 05:43:11.784202
30155	5	1-8698	30153	8	69	417	97	2012-08-02 05:43:11.784202
30156	5	1-5578	30153	69	94	173	36	2012-08-02 05:43:11.784202
30157	5	1-8893	30153	94	27	224	55	2012-08-02 05:43:11.784202
30153	5	1-9010	30153	27	64	658	91	2012-08-02 05:43:11.784202
30159	5	1-7710	30158	81	55	4259	50	2012-08-02 05:43:11.784202
30160	5	1-8997	30158	55	40	3846	6	2012-08-02 05:43:11.784202
30161	5	1-3754	30158	40	94	1927	36	2012-08-02 05:43:11.784202
30162	5	1-8893	30158	94	27	2574	55	2012-08-02 05:43:11.784202
30158	5	1-9010	30158	27	81	7767	91	2012-08-02 05:43:11.784202
30164	4	1-8504	30163	32	61	358	68	2012-08-02 05:43:11.784202
30165	4	1-7749	30163	61	94	211	36	2012-08-02 05:43:11.784202
30166	4	1-8893	30163	94	27	284	55	2012-08-02 05:43:11.784202
30163	4	1-9010	30163	27	32	865	91	2012-08-02 05:43:11.784202
30168	2	1-8758	30167	17	16	2929	69	2012-08-02 05:43:12.602174
30167	2	1-9012	30167	16	17	2961	57	2012-08-02 05:43:12.602174
30170	4	1-9014	30169	48	36	2888	76	2012-08-02 05:43:13.795384
30171	4	1-7709	30169	36	17	757	98	2012-08-02 05:43:13.795384
30172	4	1-8265	30169	17	34	2224	95	2012-08-02 05:43:13.795384
30169	4	1-9015	30169	34	48	6372	37	2012-08-02 05:43:13.795384
30174	4	1-8719	30173	85	100	1217	7	2012-08-02 05:43:15.894242
30175	4	1-9003	30173	100	61	1255	75	2012-08-02 05:43:15.894242
30176	4	1-8585	30173	61	79	6533	40	2012-08-02 05:43:15.894242
30173	4	1-9020	30173	79	85	3135	68	2012-08-02 05:43:15.894242
30178	4	1-7749	30177	61	94	470	36	2012-08-02 05:43:15.894242
30179	4	1-8893	30177	94	73	734	55	2012-08-02 05:43:15.894242
30180	4	1-8427	30177	73	79	778	40	2012-08-02 05:43:15.894242
30177	4	1-9020	30177	79	61	689	68	2012-08-02 05:43:15.894242
30182	4	1-7908	30181	92	42	32	70	2012-08-02 05:43:15.894242
30183	4	1-4241	30181	42	6	14	93	2012-08-02 05:43:15.894242
30184	4	1-8883	30181	6	79	45	40	2012-08-02 05:43:15.894242
30181	4	1-9020	30181	79	92	40	68	2012-08-02 05:43:15.894242
30186	3	1-7951	30185	63	69	2819	88	2012-08-02 05:43:16.713451
30187	3	1-8823	30185	69	9	409	23	2012-08-02 05:43:16.713451
30185	3	1-9022	30185	9	63	2230	32	2012-08-02 05:43:16.713451
30189	4	1-8843	30188	89	74	158	28	2012-08-02 05:43:16.713451
30190	4	1-8782	30188	74	49	105	47	2012-08-02 05:43:16.713451
30191	4	1-5062	30188	49	9	34	23	2012-08-02 05:43:16.713451
30188	4	1-9022	30188	9	89	290	32	2012-08-02 05:43:16.713451
30193	4	1-8076	30192	66	95	1003	96	2012-08-02 05:43:16.713451
30194	4	1-8081	30192	95	69	1306	88	2012-08-02 05:43:16.713451
30195	4	1-8823	30192	69	9	298	23	2012-08-02 05:43:16.713451
30192	4	1-9022	30192	9	66	2554	32	2012-08-02 05:43:16.713451
30197	4	1-8843	30196	89	50	1482	28	2012-08-02 05:43:16.713451
30198	4	1-5160	30196	50	1	729	54	2012-08-02 05:43:16.713451
30199	4	1-9016	30196	1	9	283	23	2012-08-02 05:43:16.713451
30196	4	1-9022	30196	9	89	2574	32	2012-08-02 05:43:16.713451
30201	3	1-8990	30200	60	20	1622	10	2012-08-02 05:43:17.553539
30202	3	1-8546	30200	20	34	470	78	2012-08-02 05:43:17.553539
30200	3	1-9024	30200	34	60	317	69	2012-08-02 05:43:17.553539
30204	5	1-7076	30203	76	44	8	48	2012-08-02 05:43:17.984428
30205	5	1-8975	30203	44	69	6	30	2012-08-02 05:43:17.984428
30206	5	1-8209	30203	69	77	4	41	2012-08-02 05:43:17.984428
30207	5	1-6775	30203	77	9	2	8	2012-08-02 05:43:17.984428
30203	5	1-9025	30203	9	76	17	70	2012-08-02 05:43:17.984428
30209	4	1-6945	30208	64	35	1956	81	2012-08-02 05:43:18.392919
30210	4	1-5851	30208	35	27	1646	51	2012-08-02 05:43:18.392919
30211	4	1-8915	30208	27	98	1462	30	2012-08-02 05:43:18.392919
30208	4	1-9026	30208	98	64	1804	41	2012-08-02 05:43:18.392919
30213	5	1-7602	30212	98	65	2762	51	2012-08-02 05:43:21.663219
30214	5	1-8934	30212	65	75	2370	55	2012-08-02 05:43:21.663219
30215	5	1-8925	30212	75	34	2153	43	2012-08-02 05:43:21.663219
30216	5	1-8950	30212	34	21	3598	57	2012-08-02 05:43:21.663219
30212	5	1-9034	30212	21	98	3307	56	2012-08-02 05:43:21.663219
30217	4	1-9035	30217	95	72	7300	1	2012-08-02 05:43:22.071836
30218	4	1-8885	30217	72	99	6754	42	2012-08-02 05:43:22.071836
30219	4	1-9019	30217	99	51	5405	56	2012-08-02 05:43:22.071836
30220	4	1-5400	30217	51	95	3278	100	2012-08-02 05:43:22.071836
30222	5	1-7912	30221	15	80	293	97	2012-08-02 05:43:23.740096
30223	5	1-8880	30221	80	53	252	45	2012-08-02 05:43:23.740096
30224	5	1-7839	30221	53	87	145	6	2012-08-02 05:43:23.740096
30225	5	1-8564	30221	87	73	109	99	2012-08-02 05:43:23.740096
30221	5	1-9039	30221	73	15	307	39	2012-08-02 05:43:23.740096
30227	5	1-8439	30226	39	18	315	93	2012-08-02 05:43:24.557814
30228	5	1-8658	30226	18	81	147	34	2012-08-02 05:43:24.557814
30229	5	1-8888	30226	81	46	149	19	2012-08-02 05:43:24.557814
30230	5	1-7459	30226	46	8	169	55	2012-08-02 05:43:24.557814
30226	5	1-9041	30226	8	39	506	53	2012-08-02 05:43:24.557814
30232	4	1-5876	30231	79	46	499	19	2012-08-02 05:43:24.977435
30233	4	1-7459	30231	46	63	235	55	2012-08-02 05:43:24.977435
30234	4	1-8343	30231	63	53	101	38	2012-08-02 05:43:24.977435
30231	4	1-9042	30231	53	79	5569	21	2012-08-02 05:43:24.977435
30236	3	1-8042	30235	51	14	2798	85	2012-08-02 05:43:26.203666
30237	3	1-8197	30235	14	48	762	88	2012-08-02 05:43:26.203666
30235	3	1-9045	30235	48	51	2587	99	2012-08-02 05:43:26.203666
30239	4	1-3433	30238	96	95	2136	83	2012-08-02 05:43:26.203666
30240	4	1-8634	30238	95	17	1699	17	2012-08-02 05:43:26.203666
30241	4	1-8899	30238	17	48	1904	88	2012-08-02 05:43:26.203666
30238	4	1-9045	30238	48	96	7252	99	2012-08-02 05:43:26.203666
30243	2	1-3745	30242	69	51	3150	46	2012-08-02 05:43:27.010249
30242	2	1-9047	30242	51	69	8070	23	2012-08-02 05:43:27.010249
30245	3	1-7792	30244	5	39	596	13	2012-08-02 05:43:27.010249
30246	3	1-8465	30244	39	51	474	46	2012-08-02 05:43:27.010249
30244	3	1-9047	30244	51	5	1846	23	2012-08-02 05:43:27.010249
30248	5	1-6492	30247	30	69	2483	88	2012-08-02 05:43:28.225547
30249	5	1-8823	30247	69	76	817	23	2012-08-02 05:43:28.225547
30250	5	1-8907	30247	76	25	1165	21	2012-08-02 05:43:28.225547
30251	5	1-8785	30247	25	26	1547	36	2012-08-02 05:43:28.225547
30247	5	1-9050	30247	26	30	1662	77	2012-08-02 05:43:28.225547
30253	4	1-8613	30252	6	21	2685	1	2012-08-02 05:43:31.108796
30254	4	1-7669	30252	21	97	1861	97	2012-08-02 05:43:31.108796
30255	4	1-8146	30252	97	17	1706	54	2012-08-02 05:43:31.108796
30252	4	1-9057	30252	17	6	4209	44	2012-08-02 05:43:31.108796
30257	4	1-7735	30256	7	23	176	59	2012-08-02 05:43:31.517441
30258	4	1-8965	30256	23	15	376	39	2012-08-02 05:43:31.517441
30259	4	1-7912	30256	15	45	373	97	2012-08-02 05:43:31.517441
30256	4	1-9058	30256	45	7	265	4	2012-08-02 05:43:31.517441
30261	4	1-9054	30260	78	81	2574	76	2012-08-02 05:43:31.9375
30262	4	1-7024	30260	81	34	1480	95	2012-08-02 05:43:31.9375
30263	4	1-9015	30260	34	82	3578	37	2012-08-02 05:43:31.9375
30260	4	1-9059	30260	82	78	4138	16	2012-08-02 05:43:31.9375
30265	4	1-7384	30264	36	23	494	59	2012-08-02 05:43:31.9375
30266	4	1-8965	30264	23	24	1048	39	2012-08-02 05:43:31.9375
30267	4	1-8802	30264	24	82	1945	37	2012-08-02 05:43:31.9375
30264	4	1-9059	30264	82	36	2630	16	2012-08-02 05:43:31.9375
30269	4	1-7564	30268	77	48	383	29	2012-08-02 05:43:32.335162
30270	4	1-8740	30268	48	36	147	58	2012-08-02 05:43:32.335162
30271	4	1-8699	30268	36	89	129	71	2012-08-02 05:43:32.335162
30268	4	1-9060	30268	89	77	1690	60	2012-08-02 05:43:32.335162
30273	4	1-6513	30272	5	35	902	65	2012-08-02 05:43:33.1195
30274	4	1-8859	30272	35	62	311	20	2012-08-02 05:43:33.1195
30275	4	1-8629	30272	62	33	781	10	2012-08-02 05:43:33.1195
30272	4	1-9062	30272	33	5	7042	40	2012-08-02 05:43:33.1195
30277	5	1-8375	30276	71	87	1223	6	2012-08-02 05:43:33.528448
30278	5	1-8564	30276	87	35	837	99	2012-08-02 05:43:33.528448
30279	5	1-8757	30276	35	2	1598	73	2012-08-02 05:43:33.528448
30280	5	1-7019	30276	2	23	871	33	2012-08-02 05:43:33.528448
30276	5	1-9063	30276	23	71	808	24	2012-08-02 05:43:33.528448
30282	3	1-8629	30281	62	64	4666	10	2012-08-02 05:43:35.980907
30283	3	1-8754	30281	64	66	1122	15	2012-08-02 05:43:35.980907
30281	3	1-9069	30281	66	62	1292	20	2012-08-02 05:43:35.980907
30285	4	1-7188	30284	82	36	786	88	2012-08-02 05:43:35.980907
30286	4	1-9056	30284	36	66	526	14	2012-08-02 05:43:35.980907
30287	4	1-8935	30284	66	66	634	15	2012-08-02 05:43:35.980907
30284	4	1-9069	30284	66	82	839	20	2012-08-02 05:43:35.980907
30289	4	1-7023	30288	34	74	294	8	2012-08-02 05:43:37.185122
30290	4	1-8108	30288	74	4	237	56	2012-08-02 05:43:37.185122
30291	4	1-8973	30288	4	56	211	12	2012-08-02 05:43:37.185122
30288	4	1-9072	30288	56	34	2022	44	2012-08-02 05:43:37.185122
30293	4	1-5966	30292	99	14	1143	63	2012-08-02 05:43:37.991532
30294	4	1-9055	30292	14	89	1051	23	2012-08-02 05:43:37.991532
30295	4	1-8524	30292	89	88	1174	40	2012-08-02 05:43:37.991532
30292	4	1-9074	30292	88	99	3186	80	2012-08-02 05:43:37.991532
30297	4	1-6771	30296	64	55	517	50	2012-08-02 05:43:37.991532
30298	4	1-8997	30296	55	91	488	6	2012-08-02 05:43:37.991532
30299	4	1-7374	30296	91	88	419	40	2012-08-02 05:43:37.991532
30296	4	1-9074	30296	88	64	1208	80	2012-08-02 05:43:37.991532
30301	4	1-6035	30300	12	45	124	28	2012-08-02 05:43:37.991532
30302	4	1-9038	30300	45	95	146	30	2012-08-02 05:43:37.991532
30303	4	1-7701	30300	95	88	113	40	2012-08-02 05:43:37.991532
30300	4	1-9074	30300	88	12	360	80	2012-08-02 05:43:37.991532
30305	4	1-3035	30304	65	25	1550	51	2012-08-02 05:43:39.218273
30306	4	1-8648	30304	25	66	1130	98	2012-08-02 05:43:39.218273
30307	4	1-8555	30304	66	100	835	93	2012-08-02 05:43:39.218273
30304	4	1-9077	30304	100	65	1810	9	2012-08-02 05:43:39.218273
30309	4	1-8537	30308	55	39	1017	13	2012-08-02 05:43:39.218273
30310	4	1-8465	30308	39	64	994	46	2012-08-02 05:43:39.218273
30311	4	1-8036	30308	64	100	523	93	2012-08-02 05:43:39.218273
30308	4	1-9077	30308	100	55	1163	9	2012-08-02 05:43:39.218273
30313	4	1-8660	30312	98	1	3936	87	2012-08-02 05:43:39.637871
30314	4	1-8941	30312	1	16	3269	20	2012-08-02 05:43:39.637871
30315	4	1-6667	30312	16	40	2012	61	2012-08-02 05:43:39.637871
30312	4	1-9078	30312	40	98	6330	23	2012-08-02 05:43:39.637871
30317	3	1-6798	30316	11	87	934	57	2012-08-02 05:43:40.046745
30318	3	1-8422	30316	87	55	125	3	2012-08-02 05:43:40.046745
30316	3	1-9079	30316	55	11	3669	47	2012-08-02 05:43:40.046745
30320	3	1-6532	30319	53	1	612	9	2012-08-02 05:43:40.046745
30321	3	1-5611	30319	1	55	128	3	2012-08-02 05:43:40.046745
30319	3	1-9079	30319	55	53	4678	47	2012-08-02 05:43:40.046745
30323	5	1-7915	30322	79	79	2668	4	2012-08-02 05:43:40.455554
30324	5	1-9073	30322	79	15	2305	86	2012-08-02 05:43:40.455554
30325	5	1-8646	30322	15	23	853	33	2012-08-02 05:43:40.455554
30326	5	1-9063	30322	23	8	769	24	2012-08-02 05:43:40.455554
30322	5	1-9080	30322	8	79	4536	84	2012-08-02 05:43:40.455554
30328	5	1-7502	30327	71	87	2087	15	2012-08-02 05:43:40.455554
30329	5	1-9071	30327	87	43	1360	52	2012-08-02 05:43:40.455554
30330	5	1-6659	30327	43	17	995	89	2012-08-02 05:43:40.455554
30331	5	1-8663	30327	17	8	573	24	2012-08-02 05:43:40.455554
30327	5	1-9080	30327	8	71	3752	84	2012-08-02 05:43:40.455554
30333	5	1-7851	30332	57	67	656	2	2012-08-02 05:43:40.455554
30334	5	1-8609	30332	67	46	134	98	2012-08-02 05:43:40.455554
30335	5	1-6970	30332	46	23	65	33	2012-08-02 05:43:40.455554
30336	5	1-9063	30332	23	8	69	24	2012-08-02 05:43:40.455554
30332	5	1-9080	30332	8	57	474	84	2012-08-02 05:43:40.455554
30338	4	1-8875	30337	85	49	61	35	2012-08-02 05:43:42.135115
30339	4	1-3662	30337	49	81	33	60	2012-08-02 05:43:42.135115
30340	4	1-8844	30337	81	67	8	3	2012-08-02 05:43:42.135115
30337	4	1-9084	30337	67	85	79	66	2012-08-02 05:43:42.135115
30342	5	1-7209	30341	84	28	1594	79	2012-08-02 05:43:42.135115
30343	5	1-7652	30341	28	39	1036	93	2012-08-02 05:43:42.135115
30344	5	1-8231	30341	39	4	505	97	2012-08-02 05:43:42.135115
30345	5	1-8874	30341	4	67	220	3	2012-08-02 05:43:42.135115
30341	5	1-9084	30341	67	84	2525	66	2012-08-02 05:43:42.135115
30347	4	1-6978	30346	82	34	931	63	2012-08-02 05:43:42.135115
30348	4	1-8978	30346	34	7	1410	91	2012-08-02 05:43:42.135115
30349	4	1-1301	30346	7	67	249	3	2012-08-02 05:43:42.135115
30346	4	1-9084	30346	67	82	3014	66	2012-08-02 05:43:42.135115
30351	4	1-5864	30350	2	23	69	59	2012-08-02 05:43:42.135115
30352	4	1-8965	30350	23	8	136	39	2012-08-02 05:43:42.135115
30353	4	1-3269	30350	8	67	14	3	2012-08-02 05:43:42.135115
30350	4	1-9084	30350	67	2	175	66	2012-08-02 05:43:42.135115
30355	3	1-7325	30354	21	34	331	26	2012-08-02 05:43:42.135115
30356	3	1-195	30354	34	67	106	3	2012-08-02 05:43:42.135115
30354	3	1-9084	30354	67	21	1405	66	2012-08-02 05:43:42.135115
30358	4	1-6922	30357	5	89	806	32	2012-08-02 05:43:43.383398
30359	4	1-8843	30357	89	76	575	28	2012-08-02 05:43:43.383398
30360	4	1-8471	30357	76	59	483	42	2012-08-02 05:43:43.383398
30357	4	1-9087	30357	59	5	2084	71	2012-08-02 05:43:43.383398
30362	2	1-8880	30361	80	75	1526	45	2012-08-02 05:43:43.792344
30361	2	1-9088	30361	75	80	1915	97	2012-08-02 05:43:43.792344
30364	4	1-8760	30363	36	89	1837	77	2012-08-02 05:43:44.201053
30365	4	1-8986	30363	89	98	1117	58	2012-08-02 05:43:44.201053
30366	4	1-7694	30363	98	9	757	81	2012-08-02 05:43:44.201053
30363	4	1-9089	30363	9	36	2642	90	2012-08-02 05:43:44.201053
30368	4	1-5165	30367	4	7	2755	31	2012-08-02 05:43:44.201053
30369	4	1-8468	30367	7	97	2923	2	2012-08-02 05:43:44.201053
30370	4	1-8296	30367	97	9	1066	81	2012-08-02 05:43:44.201053
30367	4	1-9089	30367	9	4	4589	90	2012-08-02 05:43:44.201053
30372	4	1-8081	30371	95	83	952	88	2012-08-02 05:43:44.610137
30373	4	1-6755	30371	83	9	104	8	2012-08-02 05:43:44.610137
30374	4	1-9025	30371	9	89	552	70	2012-08-02 05:43:44.610137
30371	4	1-9090	30371	89	95	844	96	2012-08-02 05:43:44.610137
30376	3	1-6518	30375	19	62	156	45	2012-08-02 05:43:44.610137
30377	3	1-7547	30375	62	89	190	70	2012-08-02 05:43:44.610137
30375	3	1-9090	30375	89	19	327	96	2012-08-02 05:43:44.610137
30379	4	1-8835	30378	85	51	2046	64	2012-08-02 05:43:44.610137
30380	4	1-5233	30378	51	9	453	8	2012-08-02 05:43:44.610137
30381	4	1-9025	30378	9	89	3194	70	2012-08-02 05:43:44.610137
30378	4	1-9090	30378	89	85	6477	96	2012-08-02 05:43:44.610137
30383	3	1-9078	30382	40	40	3040	23	2012-08-02 05:43:46.24489
30384	3	1-2657	30382	40	100	623	15	2012-08-02 05:43:46.24489
30382	3	1-9094	30382	100	40	1144	61	2012-08-02 05:43:46.24489
30386	4	1-8678	30385	53	52	2727	78	2012-08-02 05:43:46.24489
30387	4	1-8676	30385	52	66	2991	14	2012-08-02 05:43:46.24489
30388	4	1-8935	30385	66	100	2719	15	2012-08-02 05:43:46.24489
30385	4	1-9094	30385	100	53	6215	61	2012-08-02 05:43:46.24489
30390	4	1-8785	30389	25	83	1300	36	2012-08-02 05:43:50.391549
30391	4	1-8681	30389	83	76	3459	29	2012-08-02 05:43:50.391549
30392	4	1-8556	30389	76	56	727	28	2012-08-02 05:43:50.391549
30389	4	1-9104	30389	56	25	1011	21	2012-08-02 05:43:50.391549
30394	4	1-9019	30393	99	4	1768	56	2012-08-02 05:43:50.82465
30395	4	1-8973	30393	4	61	2161	12	2012-08-02 05:43:50.82465
30396	4	1-6209	30393	61	53	1114	59	2012-08-02 05:43:50.82465
30393	4	1-9105	30393	53	99	1879	42	2012-08-02 05:43:50.82465
30398	4	1-9028	30397	25	51	572	31	2012-08-02 05:43:51.234849
30399	4	1-8637	30397	51	51	304	42	2012-08-02 05:43:51.234849
30400	4	1-8470	30397	51	47	121	26	2012-08-02 05:43:51.234849
30397	4	1-9106	30397	47	25	205	43	2012-08-02 05:43:51.234849
30402	4	1-8689	30401	68	41	3836	85	2012-08-02 05:43:51.666789
30403	4	1-5226	30401	41	38	688	73	2012-08-02 05:43:51.666789
30404	4	1-9067	30401	38	96	216	26	2012-08-02 05:43:51.666789
30401	4	1-9107	30401	96	68	1557	79	2012-08-02 05:43:51.666789
30406	5	1-6756	30405	36	92	1153	95	2012-08-02 05:43:53.325755
30407	5	1-8862	30405	92	76	839	23	2012-08-02 05:43:53.325755
30408	5	1-8907	30405	76	25	921	21	2012-08-02 05:43:53.325755
30409	5	1-8785	30405	25	54	942	36	2012-08-02 05:43:53.325755
30405	5	1-9111	30405	54	36	4822	72	2012-08-02 05:43:53.325755
30411	4	1-9012	30410	16	81	1259	57	2012-08-02 05:43:53.726128
30412	4	1-8586	30410	81	60	853	96	2012-08-02 05:43:53.726128
30413	4	1-7563	30410	60	69	686	68	2012-08-02 05:43:53.726128
30410	4	1-9112	30410	69	16	1277	69	2012-08-02 05:43:53.726128
30415	4	1-8609	30414	67	46	1215	98	2012-08-02 05:43:54.545349
30416	4	1-6970	30414	46	23	422	33	2012-08-02 05:43:54.545349
30417	4	1-9063	30414	23	79	317	24	2012-08-02 05:43:54.545349
30414	4	1-9114	30414	79	67	8352	2	2012-08-02 05:43:54.545349
30419	4	1-8875	30418	85	18	2126	35	2012-08-02 05:43:54.966471
30420	4	1-3866	30418	18	18	583	77	2012-08-02 05:43:54.966471
30421	4	1-7033	30418	18	91	186	55	2012-08-02 05:43:54.966471
30418	4	1-9115	30418	91	85	4750	66	2012-08-02 05:43:54.966471
30423	3	1-7325	30422	21	95	369	26	2012-08-02 05:43:54.966471
30424	3	1-8457	30422	95	91	110	55	2012-08-02 05:43:54.966471
30422	3	1-9115	30422	91	21	3180	66	2012-08-02 05:43:54.966471
30425	4	1-9117	30425	46	1	173	87	2012-08-02 05:43:55.819374
30426	4	1-8941	30425	1	16	136	20	2012-08-02 05:43:55.819374
30427	4	1-6667	30425	16	79	79	61	2012-08-02 05:43:55.819374
30428	4	1-6424	30425	79	46	57	31	2012-08-02 05:43:55.819374
30430	4	1-8941	30429	1	56	1216	20	2012-08-02 05:43:55.819374
30431	4	1-4886	30429	56	16	520	36	2012-08-02 05:43:55.819374
30432	4	1-6829	30429	16	46	421	31	2012-08-02 05:43:55.819374
30429	4	1-9117	30429	46	1	1403	87	2012-08-02 05:43:55.819374
30434	4	1-8142	30433	42	42	2059	38	2012-08-02 05:43:55.819374
30435	4	1-4896	30433	42	4	1031	90	2012-08-02 05:43:55.819374
30436	4	1-5165	30433	4	46	703	31	2012-08-02 05:43:55.819374
30433	4	1-9117	30433	46	42	2407	87	2012-08-02 05:43:55.819374
30438	4	1-9088	30437	75	4	5060	97	2012-08-02 05:43:57.434257
30439	4	1-8874	30437	4	91	2405	3	2012-08-02 05:43:57.434257
30440	4	1-8984	30437	91	34	1683	68	2012-08-02 05:43:57.434257
30437	4	1-9121	30437	34	75	3976	45	2012-08-02 05:43:57.434257
30442	5	1-8595	30441	42	80	2813	1	2012-08-02 05:43:57.854521
30443	5	1-7451	30441	80	2	816	94	2012-08-02 05:43:57.854521
30444	5	1-8964	30441	2	56	665	65	2012-08-02 05:43:57.854521
30445	5	1-9097	30441	56	92	1283	57	2012-08-02 05:43:57.854521
30441	5	1-9122	30441	92	42	7047	72	2012-08-02 05:43:57.854521
30447	4	1-8002	30446	33	44	816	48	2012-08-02 05:43:57.854521
30448	4	1-8975	30446	44	48	524	30	2012-08-02 05:43:57.854521
30449	4	1-4941	30446	48	92	397	57	2012-08-02 05:43:57.854521
30446	4	1-9122	30446	92	33	2825	72	2012-08-02 05:43:57.854521
30451	4	1-7933	30450	73	82	754	63	2012-08-02 05:43:59.491054
30452	4	1-9053	30450	82	45	610	19	2012-08-02 05:43:59.491054
30453	4	1-6115	30450	45	83	600	29	2012-08-02 05:43:59.491054
30450	4	1-9126	30450	83	73	962	33	2012-08-02 05:43:59.491054
30455	4	1-9063	30454	23	84	2662	24	2012-08-02 05:43:59.491054
30456	4	1-8113	30454	84	31	2476	68	2012-08-02 05:43:59.491054
30457	4	1-8531	30454	31	83	2209	29	2012-08-02 05:43:59.491054
30454	4	1-9126	30454	83	23	3602	33	2012-08-02 05:43:59.491054
30459	3	1-8496	30458	73	31	2849	68	2012-08-02 05:43:59.491054
30460	3	1-8531	30458	31	83	2618	29	2012-08-02 05:43:59.491054
30458	3	1-9126	30458	83	73	4397	33	2012-08-02 05:43:59.491054
30462	5	1-8564	30461	87	35	1373	99	2012-08-02 05:43:59.914139
30463	5	1-8757	30461	35	38	2808	73	2012-08-02 05:43:59.914139
30464	5	1-9067	30461	38	47	1082	26	2012-08-02 05:43:59.914139
30465	5	1-9106	30461	47	66	1954	43	2012-08-02 05:43:59.914139
30461	5	1-9127	30461	66	87	1874	6	2012-08-02 05:43:59.914139
30467	4	1-8830	30466	10	29	1704	81	2012-08-02 05:44:00.334898
30468	4	1-8956	30466	29	42	2248	12	2012-08-02 05:44:00.334898
30469	4	1-6855	30466	42	71	2372	44	2012-08-02 05:44:00.334898
30466	4	1-9128	30466	71	10	5427	60	2012-08-02 05:44:00.334898
30471	4	1-8912	30470	79	93	2363	36	2012-08-02 05:44:01.573448
30472	4	1-8992	30470	93	19	2687	94	2012-08-02 05:44:01.573448
30473	4	1-6150	30470	19	35	1025	59	2012-08-02 05:44:01.573448
30470	4	1-9131	30470	35	79	3164	100	2012-08-02 05:44:01.573448
30475	3	1-9036	30474	51	2	228	66	2012-08-02 05:44:01.573448
30476	3	1-5864	30474	2	35	67	59	2012-08-02 05:44:01.573448
30474	3	1-9131	30474	35	51	224	100	2012-08-02 05:44:01.573448
30478	4	1-8887	30477	2	5	3508	84	2012-08-02 05:44:01.993675
30479	4	1-8460	30477	5	5	1955	78	2012-08-02 05:44:01.993675
30480	4	1-8994	30477	5	24	2183	28	2012-08-02 05:44:01.993675
30477	4	1-9132	30477	24	2	2316	87	2012-08-02 05:44:01.993675
30482	3	1-8686	30481	89	77	2390	64	2012-08-02 05:44:03.198289
30483	3	1-8817	30481	77	32	1856	73	2012-08-02 05:44:03.198289
30481	3	1-9135	30481	32	89	7288	45	2012-08-02 05:44:03.198289
30485	4	1-9028	30484	25	46	2360	31	2012-08-02 05:44:04.027398
30486	4	1-9117	30484	46	71	5742	87	2012-08-02 05:44:04.027398
30487	4	1-4784	30484	71	51	1784	46	2012-08-02 05:44:04.027398
30484	4	1-9137	30484	51	25	1052	43	2012-08-02 05:44:04.027398
30489	4	1-9125	30488	60	19	4639	86	2012-08-02 05:44:05.265217
30490	4	1-9021	30488	19	55	4256	81	2012-08-02 05:44:05.265217
30491	4	1-7664	30488	55	45	4138	88	2012-08-02 05:44:05.265217
30488	4	1-9140	30488	45	60	5279	66	2012-08-02 05:44:05.265217
30493	4	1-9125	30492	60	21	225	86	2012-08-02 05:44:05.265217
30494	4	1-8643	30492	21	17	101	17	2012-08-02 05:44:05.265217
30495	4	1-8899	30492	17	45	197	88	2012-08-02 05:44:05.265217
30492	4	1-9140	30492	45	60	254	66	2012-08-02 05:44:05.265217
30497	5	1-5557	30496	77	14	594	63	2012-08-02 05:44:06.078784
30498	5	1-9055	30496	14	76	437	23	2012-08-02 05:44:06.078784
30499	5	1-8907	30496	76	25	429	21	2012-08-02 05:44:06.078784
30500	5	1-8785	30496	25	11	393	36	2012-08-02 05:44:06.078784
30496	5	1-9142	30496	11	77	1996	77	2012-08-02 05:44:06.078784
30502	4	1-8714	30501	8	5	1405	76	2012-08-02 05:44:06.078784
30503	4	1-8983	30501	5	70	950	75	2012-08-02 05:44:06.078784
30504	4	1-8226	30501	70	11	254	36	2012-08-02 05:44:06.078784
30501	4	1-9142	30501	11	8	1465	77	2012-08-02 05:44:06.078784
30506	4	1-6284	30505	20	44	81	4	2012-08-02 05:44:06.078784
30507	4	1-8977	30505	44	26	44	11	2012-08-02 05:44:06.078784
30508	4	1-8131	30505	26	11	34	36	2012-08-02 05:44:06.078784
30505	4	1-9142	30505	11	20	210	77	2012-08-02 05:44:06.078784
30510	3	1-2744	30509	80	78	734	8	2012-08-02 05:44:06.078784
30511	3	1-6840	30509	78	11	550	36	2012-08-02 05:44:06.078784
30509	3	1-9142	30509	11	80	3629	77	2012-08-02 05:44:06.078784
30513	4	1-8717	30512	5	66	313	15	2012-08-02 05:44:06.078784
30514	4	1-9069	30512	66	56	374	20	2012-08-02 05:44:06.078784
30515	4	1-4886	30512	56	11	150	36	2012-08-02 05:44:06.078784
30512	4	1-9142	30512	11	5	1031	77	2012-08-02 05:44:06.078784
30517	4	1-7377	30516	50	26	1080	22	2012-08-02 05:44:06.078784
30518	4	1-7210	30516	26	32	1244	84	2012-08-02 05:44:06.078784
30519	4	1-5873	30516	32	11	212	36	2012-08-02 05:44:06.078784
30516	4	1-9142	30516	11	50	1539	77	2012-08-02 05:44:06.078784
30521	3	1-6580	30520	30	61	1691	17	2012-08-02 05:44:06.500589
30522	3	1-7800	30520	61	22	855	58	2012-08-02 05:44:06.500589
30520	3	1-9143	30520	22	30	3320	20	2012-08-02 05:44:06.500589
30524	3	1-6419	30523	77	28	1260	46	2012-08-02 05:44:06.500589
30525	3	1-8901	30523	28	22	962	58	2012-08-02 05:44:06.500589
30523	3	1-9143	30523	22	77	3963	20	2012-08-02 05:44:06.500589
30527	4	1-8345	30526	60	51	2848	100	2012-08-02 05:44:06.899012
30528	4	1-9036	30526	51	60	4215	66	2012-08-02 05:44:06.899012
30529	4	1-9125	30526	60	32	4381	86	2012-08-02 05:44:06.899012
30526	4	1-9144	30526	32	60	4484	53	2012-08-02 05:44:06.899012
30530	4	1-9147	30530	97	48	844	18	2012-08-02 05:44:08.114683
30531	4	1-8572	30530	48	99	1082	42	2012-08-02 05:44:08.114683
30532	4	1-9019	30530	99	4	829	56	2012-08-02 05:44:08.114683
30533	4	1-8973	30530	4	97	826	12	2012-08-02 05:44:08.114683
30535	4	1-7967	30534	92	79	4237	40	2012-08-02 05:44:08.114683
30536	4	1-9020	30534	79	12	3365	68	2012-08-02 05:44:08.114683
30537	4	1-8559	30534	12	97	2543	12	2012-08-02 05:44:08.114683
30534	4	1-9147	30534	97	92	3048	18	2012-08-02 05:44:08.114683
30539	4	1-3537	30538	88	56	30	65	2012-08-02 05:44:08.114683
30540	4	1-9097	30538	56	4	85	57	2012-08-02 05:44:08.114683
30541	4	1-8881	30538	4	97	92	12	2012-08-02 05:44:08.114683
30538	4	1-9147	30538	97	88	113	18	2012-08-02 05:44:08.114683
30543	3	1-8874	30542	4	52	1953	3	2012-08-02 05:44:09.319845
30544	3	1-9149	30542	52	24	3670	29	2012-08-02 05:44:09.319845
30542	3	1-9150	30542	24	4	5437	97	2012-08-02 05:44:09.319845
30546	4	1-8606	30545	63	86	5703	91	2012-08-02 05:44:09.717484
30547	4	1-8190	30545	86	56	2526	28	2012-08-02 05:44:09.717484
30548	4	1-9104	30545	56	99	2407	21	2012-08-02 05:44:09.717484
30545	4	1-9151	30545	99	63	9383	100	2012-08-02 05:44:09.717484
30550	4	1-8145	30549	1	1	419	33	2012-08-02 05:44:11.717651
30551	4	1-8403	30549	1	94	308	61	2012-08-02 05:44:11.717651
30552	4	1-8827	30549	94	4	290	84	2012-08-02 05:44:11.717651
30549	4	1-9156	30549	4	1	1077	18	2012-08-02 05:44:11.717651
30554	2	1-5837	30553	63	4	1935	84	2012-08-02 05:44:11.717651
30553	2	1-9156	30553	4	63	8155	18	2012-08-02 05:44:11.717651
30556	2	1-8531	30555	31	47	1502	29	2012-08-02 05:44:13.728247
30555	2	1-9161	30555	47	31	3618	68	2012-08-02 05:44:13.728247
30558	3	1-7459	30557	46	96	2440	55	2012-08-02 05:44:14.523637
30559	3	1-8000	30557	96	72	478	96	2012-08-02 05:44:14.523637
30557	3	1-9163	30557	72	46	3742	19	2012-08-02 05:44:14.523637
30561	4	1-5517	30560	76	1	526	9	2012-08-02 05:44:14.523637
30562	4	1-5611	30560	1	24	192	3	2012-08-02 05:44:14.523637
30563	4	1-8421	30560	24	72	139	96	2012-08-02 05:44:14.523637
30560	4	1-9163	30560	72	76	1415	19	2012-08-02 05:44:14.523637
30565	4	1-6776	30564	58	4	89	97	2012-08-02 05:44:14.523637
30566	4	1-8874	30564	4	24	36	3	2012-08-02 05:44:14.523637
30567	4	1-8421	30564	24	72	26	96	2012-08-02 05:44:14.523637
30564	4	1-9163	30564	72	58	271	19	2012-08-02 05:44:14.523637
30569	4	1-9081	30568	38	79	6351	4	2012-08-02 05:44:16.545989
30570	4	1-9073	30568	79	32	5091	86	2012-08-02 05:44:16.545989
30571	4	1-9144	30568	32	52	4505	53	2012-08-02 05:44:16.545989
30568	4	1-9168	30568	52	38	7766	82	2012-08-02 05:44:16.545989
30573	3	1-8821	30572	15	82	1742	23	2012-08-02 05:44:16.545989
30574	3	1-2629	30572	82	52	1109	53	2012-08-02 05:44:16.545989
30572	3	1-9168	30572	52	15	1995	82	2012-08-02 05:44:16.545989
30576	5	1-7953	30575	91	9	1542	69	2012-08-02 05:44:17.352979
30577	5	1-8580	30575	9	86	582	12	2012-08-02 05:44:17.352979
30578	5	1-8015	30575	86	34	243	52	2012-08-02 05:44:17.352979
30579	5	1-8799	30575	34	45	214	35	2012-08-02 05:44:17.352979
30575	5	1-9170	30575	45	91	2269	42	2012-08-02 05:44:17.352979
30581	4	1-7953	30580	91	56	1176	69	2012-08-02 05:44:17.352979
30582	4	1-6097	30580	56	28	215	26	2012-08-02 05:44:17.352979
30583	4	1-9145	30580	28	45	139	35	2012-08-02 05:44:17.352979
30580	4	1-9170	30580	45	91	1599	42	2012-08-02 05:44:17.352979
30585	5	1-5974	30584	79	12	1589	17	2012-08-02 05:44:17.352979
30586	5	1-5439	30584	12	87	758	15	2012-08-02 05:44:17.352979
30587	5	1-9071	30584	87	34	417	52	2012-08-02 05:44:17.352979
30588	5	1-8799	30584	34	45	449	35	2012-08-02 05:44:17.352979
30584	5	1-9170	30584	45	79	5841	42	2012-08-02 05:44:17.352979
30590	3	1-8358	30589	46	92	1143	56	2012-08-02 05:44:18.581929
30591	3	1-4773	30589	92	46	221	30	2012-08-02 05:44:18.581929
30589	3	1-9173	30589	46	46	4615	53	2012-08-02 05:44:18.581929
30593	3	1-8531	30592	31	15	996	29	2012-08-02 05:44:20.581414
30594	3	1-8811	30592	15	3	843	85	2012-08-02 05:44:20.581414
30592	3	1-9178	30592	3	31	1710	68	2012-08-02 05:44:20.581414
30596	4	1-7749	30595	61	83	1669	36	2012-08-02 05:44:20.581414
30597	4	1-8681	30595	83	15	2361	29	2012-08-02 05:44:20.581414
30598	4	1-8811	30595	15	3	2193	85	2012-08-02 05:44:20.581414
30595	4	1-9178	30595	3	61	4875	68	2012-08-02 05:44:20.581414
30600	5	1-8452	30599	33	37	718	41	2012-08-02 05:44:20.990529
30601	5	1-8035	30599	37	61	648	68	2012-08-02 05:44:20.990529
30602	5	1-7749	30599	61	93	443	36	2012-08-02 05:44:20.990529
30603	5	1-8992	30599	93	18	802	94	2012-08-02 05:44:20.990529
30599	5	1-9179	30599	18	33	918	38	2012-08-02 05:44:20.990529
30605	3	1-8681	30604	83	48	1363	29	2012-08-02 05:44:21.797201
30606	3	1-8740	30604	48	37	575	58	2012-08-02 05:44:21.797201
30604	3	1-9181	30604	37	83	694	36	2012-08-02 05:44:21.797201
30608	4	1-8992	30607	93	22	4328	94	2012-08-02 05:44:21.797201
30609	4	1-8914	30607	22	89	3022	77	2012-08-02 05:44:21.797201
30610	4	1-8986	30607	89	37	2099	58	2012-08-02 05:44:21.797201
30607	4	1-9181	30607	37	93	2949	36	2012-08-02 05:44:21.797201
30612	3	1-9050	30611	26	89	615	77	2012-08-02 05:44:21.797201
30613	3	1-8986	30611	89	37	447	58	2012-08-02 05:44:21.797201
30611	3	1-9181	30611	37	26	657	36	2012-08-02 05:44:21.797201
30615	5	1-9025	30614	9	42	2895	70	2012-08-02 05:44:22.581626
30616	5	1-4241	30614	42	94	1015	93	2012-08-02 05:44:22.581626
30617	5	1-7406	30614	94	100	242	7	2012-08-02 05:44:22.581626
30618	5	1-9003	30614	100	2	367	75	2012-08-02 05:44:22.581626
30614	5	1-9183	30614	2	9	402	8	2012-08-02 05:44:22.581626
30620	4	1-8268	30619	41	78	3527	16	2012-08-02 05:44:22.581626
30621	4	1-9054	30619	78	5	2541	76	2012-08-02 05:44:22.581626
30622	4	1-8983	30619	5	2	2223	75	2012-08-02 05:44:22.581626
30619	4	1-9183	30619	2	41	3017	8	2012-08-02 05:44:22.581626
30623	1	1-2778	\N	29	29	1917	70	2012-08-02 05:44:22.581626
30625	4	1-8026	30624	6	6	4515	67	2012-08-02 05:44:24.250627
30626	4	1-9083	30624	6	32	5858	82	2012-08-02 05:44:24.250627
30627	4	1-8370	30624	32	31	3005	51	2012-08-02 05:44:24.250627
30624	4	1-9187	30624	31	6	3181	24	2012-08-02 05:44:24.250627
30629	3	1-8655	30628	17	16	844	27	2012-08-02 05:44:24.250627
30630	3	1-7635	30628	16	31	653	51	2012-08-02 05:44:24.250627
30628	3	1-9187	30628	31	17	791	24	2012-08-02 05:44:24.250627
30632	4	1-8675	30631	99	17	1380	54	2012-08-02 05:44:24.250627
30633	4	1-9057	30631	17	76	2966	44	2012-08-02 05:44:24.250627
30634	4	1-5019	30631	76	31	1271	51	2012-08-02 05:44:24.250627
30631	4	1-9187	30631	31	99	1563	24	2012-08-02 05:44:24.250627
30636	3	1-8655	30635	17	51	93	27	2012-08-02 05:44:24.250627
30637	3	1-4386	30635	51	31	67	51	2012-08-02 05:44:24.250627
30635	3	1-9187	30635	31	17	84	24	2012-08-02 05:44:24.250627
30639	4	1-8113	30638	84	69	1720	68	2012-08-02 05:44:24.250627
30640	4	1-9112	30638	69	6	3020	69	2012-08-02 05:44:24.250627
30641	4	1-8806	30638	6	31	1190	51	2012-08-02 05:44:24.250627
30638	4	1-9187	30638	31	84	1509	24	2012-08-02 05:44:24.250627
30643	4	1-7237	30642	36	21	153	57	2012-08-02 05:44:24.250627
30644	4	1-9034	30642	21	98	137	56	2012-08-02 05:44:24.250627
30645	4	1-7602	30642	98	31	111	51	2012-08-02 05:44:24.250627
30642	4	1-9187	30642	31	36	149	24	2012-08-02 05:44:24.250627
30647	4	1-7456	30646	5	47	1396	26	2012-08-02 05:44:24.250627
30648	4	1-9106	30646	47	61	2443	43	2012-08-02 05:44:24.250627
30649	4	1-3242	30646	61	31	1196	51	2012-08-02 05:44:24.250627
30646	4	1-9187	30646	31	5	1612	24	2012-08-02 05:44:24.250627
30651	3	1-8942	30650	34	94	879	61	2012-08-02 05:44:25.488885
30652	3	1-8827	30650	94	23	553	84	2012-08-02 05:44:25.488885
30650	3	1-9190	30650	23	34	4744	6	2012-08-02 05:44:25.488885
30654	4	1-8190	30653	86	56	2619	28	2012-08-02 05:44:25.897869
30655	4	1-9104	30653	56	48	3227	21	2012-08-02 05:44:25.897869
30656	4	1-8349	30653	48	52	1379	93	2012-08-02 05:44:25.897869
30653	4	1-9191	30653	52	86	4574	91	2012-08-02 05:44:25.897869
30658	4	1-8504	30657	32	69	1318	68	2012-08-02 05:44:25.897869
30659	4	1-9112	30657	69	86	2302	69	2012-08-02 05:44:25.897869
30660	4	1-4235	30657	86	52	947	93	2012-08-02 05:44:25.897869
30657	4	1-9191	30657	52	32	3212	91	2012-08-02 05:44:25.897869
30662	4	1-9056	30661	36	66	1429	14	2012-08-02 05:44:27.124367
30663	4	1-8935	30661	66	66	1514	15	2012-08-02 05:44:27.124367
30664	4	1-9069	30661	66	80	1763	20	2012-08-02 05:44:27.124367
30661	4	1-9194	30661	80	36	2765	88	2012-08-02 05:44:27.124367
30666	4	1-7332	30665	17	48	8	29	2012-08-02 05:44:27.51093
30667	4	1-8740	30665	48	36	4	58	2012-08-02 05:44:27.51093
30668	4	1-8699	30665	36	87	5	71	2012-08-02 05:44:27.51093
30665	4	1-9195	30665	87	17	8	39	2012-08-02 05:44:27.51093
30670	3	1-8578	30669	17	66	148	84	2012-08-02 05:44:27.51093
30671	3	1-7680	30669	66	87	144	71	2012-08-02 05:44:27.51093
30669	3	1-9195	30669	87	17	280	39	2012-08-02 05:44:27.51093
30673	4	1-8294	30672	63	91	3755	77	2012-08-02 05:44:29.886294
30674	4	1-8088	30672	91	39	2207	17	2012-08-02 05:44:29.886294
30675	4	1-8787	30672	39	19	711	100	2012-08-02 05:44:29.886294
30672	4	1-9201	30672	19	63	4011	49	2012-08-02 05:44:29.886294
30677	3	1-9120	30676	59	1	1435	23	2012-08-02 05:44:29.886294
30678	3	1-7068	30676	1	19	663	100	2012-08-02 05:44:29.886294
30676	3	1-9201	30676	19	59	3886	49	2012-08-02 05:44:29.886294
30680	5	1-6696	30679	53	80	481	55	2012-08-02 05:44:30.295626
30681	5	1-9157	30679	80	23	587	51	2012-08-02 05:44:30.295626
30682	5	1-8786	30679	23	2	306	75	2012-08-02 05:44:30.295626
30683	5	1-9183	30679	2	42	385	8	2012-08-02 05:44:30.295626
30679	5	1-9202	30679	42	53	1295	38	2012-08-02 05:44:30.295626
30685	5	1-8579	30684	65	15	3110	72	2012-08-02 05:44:30.295626
30686	5	1-3771	30684	15	100	522	7	2012-08-02 05:44:30.295626
30687	5	1-9003	30684	100	2	931	75	2012-08-02 05:44:30.295626
30688	5	1-9183	30684	2	42	1201	8	2012-08-02 05:44:30.295626
30684	5	1-9202	30684	42	65	4139	38	2012-08-02 05:44:30.295626
30690	5	1-8627	30689	67	84	1433	45	2012-08-02 05:44:30.295626
30691	5	1-8961	30689	84	17	595	27	2012-08-02 05:44:30.295626
30692	5	1-9165	30689	17	2	557	75	2012-08-02 05:44:30.295626
30693	5	1-9183	30689	2	42	727	8	2012-08-02 05:44:30.295626
30689	5	1-9202	30689	42	67	2535	38	2012-08-02 05:44:30.295626
30695	5	1-8118	30694	35	3	527	90	2012-08-02 05:44:31.124409
30696	5	1-6652	30694	3	17	228	54	2012-08-02 05:44:31.124409
30697	5	1-9057	30694	17	71	373	44	2012-08-02 05:44:31.124409
30698	5	1-9128	30694	71	29	564	60	2012-08-02 05:44:31.124409
30694	5	1-9204	30694	29	35	828	73	2012-08-02 05:44:31.124409
30700	5	1-5856	30699	53	49	463	98	2012-08-02 05:44:31.124409
30701	5	1-3674	30699	49	17	262	54	2012-08-02 05:44:31.124409
30702	5	1-9057	30699	17	71	552	44	2012-08-02 05:44:31.124409
30703	5	1-9128	30699	71	29	1070	60	2012-08-02 05:44:31.124409
30699	5	1-9204	30699	29	53	2016	73	2012-08-02 05:44:31.124409
30705	3	1-9211	30704	9	4	3061	58	2012-08-02 05:44:34.737121
30706	3	1-7911	30704	4	37	2164	53	2012-08-02 05:44:34.737121
30704	3	1-9213	30704	37	9	2827	11	2012-08-02 05:44:34.737121
30708	4	1-7001	30707	45	63	279	86	2012-08-02 05:44:35.134786
30709	4	1-7765	30707	63	40	85	3	2012-08-02 05:44:35.134786
30710	4	1-8635	30707	40	6	52	67	2012-08-02 05:44:35.134786
30707	4	1-9214	30707	6	45	980	34	2012-08-02 05:44:35.134786
30712	4	1-8957	30711	52	1	605	23	2012-08-02 05:44:35.134786
30713	4	1-7068	30711	1	1	368	100	2012-08-02 05:44:35.134786
30714	4	1-5606	30711	1	6	104	67	2012-08-02 05:44:35.134786
30711	4	1-9214	30711	6	52	2022	34	2012-08-02 05:44:35.134786
30716	4	1-8958	30715	99	85	2039	72	2012-08-02 05:44:35.134786
30717	4	1-9164	30715	85	9	562	35	2012-08-02 05:44:35.134786
30718	4	1-3800	30715	9	6	168	67	2012-08-02 05:44:35.134786
30715	4	1-9214	30715	6	99	3351	34	2012-08-02 05:44:35.134786
30720	5	1-5953	30719	60	17	1173	44	2012-08-02 05:44:35.134786
30721	5	1-5004	30719	17	17	225	89	2012-08-02 05:44:35.134786
30722	5	1-8663	30719	17	6	119	24	2012-08-02 05:44:35.134786
30723	5	1-8026	30719	6	6	148	67	2012-08-02 05:44:35.134786
30719	5	1-9214	30719	6	60	3391	34	2012-08-02 05:44:35.134786
30725	4	1-8676	30724	52	66	1320	14	2012-08-02 05:44:35.555338
30726	4	1-8935	30724	66	66	871	15	2012-08-02 05:44:35.555338
30727	4	1-9069	30724	66	35	631	20	2012-08-02 05:44:35.555338
30724	4	1-9215	30724	35	52	3039	78	2012-08-02 05:44:35.555338
30729	3	1-8405	30728	49	96	2341	63	2012-08-02 05:44:35.953025
30730	3	1-8828	30728	96	43	1383	41	2012-08-02 05:44:35.953025
30728	3	1-9216	30728	43	49	5113	9	2012-08-02 05:44:35.953025
30732	4	1-6678	30731	36	2	68	75	2012-08-02 05:44:36.361923
30733	4	1-9183	30731	2	87	90	8	2012-08-02 05:44:36.361923
30734	4	1-7155	30731	87	95	66	11	2012-08-02 05:44:36.361923
30731	4	1-9217	30731	95	36	258	6	2012-08-02 05:44:36.361923
30736	3	1-8942	30735	34	68	1506	61	2012-08-02 05:44:36.361923
30737	3	1-4420	30735	68	95	776	11	2012-08-02 05:44:36.361923
30735	3	1-9217	30735	95	34	3164	6	2012-08-02 05:44:36.361923
30739	4	1-7937	30738	94	17	1969	63	2012-08-02 05:44:36.770653
30740	4	1-9212	30738	17	3	7398	5	2012-08-02 05:44:36.770653
30741	4	1-7071	30738	3	58	3582	88	2012-08-02 05:44:36.770653
30738	4	1-9218	30738	58	94	3850	13	2012-08-02 05:44:36.770653
30743	4	1-8465	30742	39	34	1480	46	2012-08-02 05:44:36.770653
30744	4	1-8492	30742	34	3	2981	5	2012-08-02 05:44:36.770653
30745	4	1-7071	30742	3	58	1513	88	2012-08-02 05:44:36.770653
30742	4	1-9218	30742	58	39	1704	13	2012-08-02 05:44:36.770653
30747	5	1-8951	30746	70	57	3140	64	2012-08-02 05:44:37.96589
30748	5	1-8365	30746	57	87	2026	15	2012-08-02 05:44:37.96589
30749	5	1-9071	30746	87	34	743	52	2012-08-02 05:44:37.96589
30750	5	1-8799	30746	34	6	533	35	2012-08-02 05:44:37.96589
30746	5	1-9221	30746	6	70	3647	30	2012-08-02 05:44:37.96589
30752	3	1-6373	30751	40	34	894	52	2012-08-02 05:44:37.96589
30753	3	1-8799	30751	34	6	671	35	2012-08-02 05:44:37.96589
30751	3	1-9221	30751	6	40	4792	30	2012-08-02 05:44:37.96589
30755	4	1-8975	30754	44	96	1005	30	2012-08-02 05:44:38.385974
30756	4	1-9166	30754	96	22	744	4	2012-08-02 05:44:38.385974
30757	4	1-9007	30754	22	55	157	98	2012-08-02 05:44:38.385974
30754	4	1-9222	30754	55	44	2275	48	2012-08-02 05:44:38.385974
30759	4	1-7954	30758	71	95	1268	59	2012-08-02 05:44:38.385974
30760	4	1-5711	30758	95	26	295	63	2012-08-02 05:44:38.385974
30761	4	1-8538	30758	26	55	125	98	2012-08-02 05:44:38.385974
30758	4	1-9222	30758	55	71	2132	48	2012-08-02 05:44:38.385974
30763	5	1-8385	30762	100	53	326	49	2012-08-02 05:44:38.385974
30764	5	1-9033	30762	53	51	264	13	2012-08-02 05:44:38.385974
30765	5	1-8814	30762	51	27	113	99	2012-08-02 05:44:38.385974
30766	5	1-8748	30762	27	55	100	98	2012-08-02 05:44:38.385974
30762	5	1-9222	30762	55	100	2087	48	2012-08-02 05:44:38.385974
30768	5	1-9071	30767	87	53	1572	52	2012-08-02 05:44:38.795089
30769	5	1-7941	30767	53	90	1523	10	2012-08-02 05:44:38.795089
30770	5	1-9141	30767	90	46	1902	39	2012-08-02 05:44:38.795089
30771	5	1-8251	30767	46	74	1745	83	2012-08-02 05:44:38.795089
30767	5	1-9223	30767	74	87	2528	15	2012-08-02 05:44:38.795089
30773	5	1-9110	30772	24	46	3209	9	2012-08-02 05:44:40.485556
30774	5	1-7157	30772	46	17	951	24	2012-08-02 05:44:40.485556
30775	5	1-8655	30772	17	20	1096	27	2012-08-02 05:44:40.485556
30776	5	1-8068	30772	20	73	1416	97	2012-08-02 05:44:40.485556
30772	5	1-9227	30772	73	24	1919	62	2012-08-02 05:44:40.485556
30778	4	1-6193	30777	93	56	503	65	2012-08-02 05:44:43.319427
30779	4	1-9097	30777	56	72	1384	57	2012-08-02 05:44:43.319427
30780	4	1-9162	30777	72	8	840	72	2012-08-02 05:44:43.319427
30777	4	1-9234	30777	8	93	1068	73	2012-08-02 05:44:43.319427
30782	4	1-8813	30781	65	34	1378	52	2012-08-02 05:44:43.717399
30783	4	1-8799	30781	34	29	1323	35	2012-08-02 05:44:43.717399
30784	4	1-8459	30781	29	73	1579	46	2012-08-02 05:44:43.717399
30781	4	1-9235	30781	73	65	1810	63	2012-08-02 05:44:43.717399
30786	4	1-8813	30785	65	34	347	52	2012-08-02 05:44:43.717399
30787	4	1-8799	30785	34	91	383	35	2012-08-02 05:44:43.717399
30788	4	1-4260	30785	91	73	300	46	2012-08-02 05:44:43.717399
30785	4	1-9235	30785	73	65	396	63	2012-08-02 05:44:43.717399
30790	4	1-9053	30789	82	51	2227	19	2012-08-02 05:44:43.717399
30791	4	1-9129	30789	51	71	3982	87	2012-08-02 05:44:43.717399
30792	4	1-4784	30789	71	73	1534	46	2012-08-02 05:44:43.717399
30789	4	1-9235	30789	73	82	2162	63	2012-08-02 05:44:43.717399
30794	2	1-6670	30793	63	59	1617	51	2012-08-02 05:44:46.965478
30793	2	1-9243	30793	59	63	2673	8	2012-08-02 05:44:46.965478
30796	2	1-6513	30795	5	7	367	65	2012-08-02 05:44:47.374285
30795	2	1-9244	30795	7	5	1917	40	2012-08-02 05:44:47.374285
30798	4	1-7944	30797	9	66	358	84	2012-08-02 05:44:48.180965
30799	4	1-7680	30797	66	9	229	71	2012-08-02 05:44:48.180965
30800	4	1-8172	30797	9	93	77	51	2012-08-02 05:44:48.180965
30797	4	1-9246	30797	93	9	385	22	2012-08-02 05:44:48.180965
30802	3	1-6437	30801	7	51	1915	54	2012-08-02 05:44:48.180965
30803	3	1-7194	30801	51	93	510	51	2012-08-02 05:44:48.180965
30801	3	1-9246	30801	93	7	2690	22	2012-08-02 05:44:48.180965
30805	4	1-8047	30804	56	90	1114	92	2012-08-02 05:44:48.180965
30806	4	1-7252	30804	90	35	283	81	2012-08-02 05:44:48.180965
30807	4	1-5851	30804	35	93	177	51	2012-08-02 05:44:48.180965
30804	4	1-9246	30804	93	56	995	22	2012-08-02 05:44:48.180965
30809	4	1-7721	30808	30	69	2036	67	2012-08-02 05:44:48.999008
30810	4	1-5544	30808	69	96	526	26	2012-08-02 05:44:48.999008
30811	4	1-9107	30808	96	26	4222	79	2012-08-02 05:44:48.999008
30808	4	1-9248	30808	26	30	4618	57	2012-08-02 05:44:48.999008
30813	3	1-8586	30812	81	64	634	96	2012-08-02 05:44:48.999008
30814	3	1-8728	30812	64	26	662	79	2012-08-02 05:44:48.999008
30812	3	1-9248	30812	26	81	848	57	2012-08-02 05:44:48.999008
30816	4	1-7825	30815	68	71	15	44	2012-08-02 05:44:49.419237
30817	4	1-9128	30815	71	40	28	60	2012-08-02 05:44:49.419237
30818	4	1-7331	30815	40	91	13	15	2012-08-02 05:44:49.419237
30815	4	1-9249	30815	91	68	23	6	2012-08-02 05:44:49.419237
30820	3	1-8140	30819	48	57	934	64	2012-08-02 05:44:49.419237
30821	3	1-8365	30819	57	91	924	15	2012-08-02 05:44:49.419237
30819	3	1-9249	30819	91	48	1682	6	2012-08-02 05:44:49.419237
30822	1	1-1501	\N	81	81	457	15	2012-08-02 05:44:49.419237
30824	4	1-9048	30823	67	9	1305	30	2012-08-02 05:44:49.816723
30825	4	1-4799	30823	9	76	674	28	2012-08-02 05:44:49.816723
30826	4	1-8471	30823	76	26	474	42	2012-08-02 05:44:49.816723
30823	4	1-9250	30823	26	67	9898	5	2012-08-02 05:44:49.816723
30828	4	1-9211	30827	9	84	82	58	2012-08-02 05:44:50.714468
30829	4	1-8932	30827	84	69	379	92	2012-08-02 05:44:50.714468
30830	4	1-7741	30827	69	38	162	70	2012-08-02 05:44:50.714468
30827	4	1-9252	30827	38	9	52	11	2012-08-02 05:44:50.714468
30832	4	1-7742	30831	33	73	2	46	2012-08-02 05:44:51.109599
30833	4	1-9235	30831	73	82	3	63	2012-08-02 05:44:51.109599
30834	4	1-9138	30831	82	41	3	55	2012-08-02 05:44:51.109599
30831	4	1-9253	30831	41	33	3	75	2012-08-02 05:44:51.109599
30836	5	1-9146	30835	80	8	3000	72	2012-08-02 05:44:51.109599
30837	5	1-9234	30835	8	38	4262	73	2012-08-02 05:44:51.109599
30838	5	1-9067	30835	38	95	1741	26	2012-08-02 05:44:51.109599
30839	5	1-8457	30835	95	41	1108	55	2012-08-02 05:44:51.109599
30835	5	1-9253	30835	41	80	1574	75	2012-08-02 05:44:51.109599
30841	4	1-8753	30840	78	66	5292	41	2012-08-02 05:44:51.52946
30840	4	1-9254	30840	85	78	5989	35	2012-08-02 05:44:51.52946
30842	4	1-8053	30840	66	85	4038	10	2012-08-02 05:44:51.52946
30843	4	1-9233	30840	85	85	4849	34	2012-08-02 05:44:51.52946
30845	3	1-8545	30844	46	65	2129	38	2012-08-02 05:44:52.336263
30846	3	1-8579	30844	65	25	628	72	2012-08-02 05:44:52.336263
30844	3	1-9256	30844	25	46	8288	87	2012-08-02 05:44:52.336263
30848	3	1-9218	30847	58	20	1790	13	2012-08-02 05:44:57.231855
30849	3	1-9051	30847	20	3	753	82	2012-08-02 05:44:57.231855
30847	3	1-9268	30847	3	58	2259	88	2012-08-02 05:44:57.231855
30851	5	1-9240	30850	37	56	2729	71	2012-08-02 05:44:57.231855
30852	5	1-7140	30850	56	78	1208	100	2012-08-02 05:44:57.231855
30853	5	1-8825	30850	78	53	760	11	2012-08-02 05:44:57.231855
30854	5	1-8412	30850	53	3	617	82	2012-08-02 05:44:57.231855
30850	5	1-9268	30850	3	37	2907	88	2012-08-02 05:44:57.231855
30856	3	1-7796	30855	46	67	2219	73	2012-08-02 05:44:58.02742
30857	3	1-9255	30855	67	92	6232	5	2012-08-02 05:44:58.02742
30855	3	1-9270	30855	92	46	2692	98	2012-08-02 05:44:58.02742
30859	4	1-6708	30858	65	77	338	64	2012-08-02 05:44:58.02742
30860	4	1-8817	30858	77	67	484	73	2012-08-02 05:44:58.02742
30861	4	1-9255	30858	67	92	1366	5	2012-08-02 05:44:58.02742
30858	4	1-9270	30858	92	65	593	98	2012-08-02 05:44:58.02742
30863	4	1-8190	30862	86	45	520	28	2012-08-02 05:44:58.43658
30864	4	1-9038	30862	45	97	449	30	2012-08-02 05:44:58.43658
30865	4	1-7426	30862	97	23	316	100	2012-08-02 05:44:58.43658
30862	4	1-9271	30862	23	86	1046	91	2012-08-02 05:44:58.43658
30867	4	1-9171	30866	81	92	373	95	2012-08-02 05:44:58.43658
30868	4	1-8862	30866	92	1	282	23	2012-08-02 05:44:58.43658
30869	4	1-7068	30866	1	23	167	100	2012-08-02 05:44:58.43658
30866	4	1-9271	30866	23	81	594	91	2012-08-02 05:44:58.43658
30871	4	1-1301	30870	7	10	81	3	2012-08-02 05:44:58.43658
30872	4	1-9133	30870	10	91	123	52	2012-08-02 05:44:58.43658
30873	4	1-6547	30870	91	23	110	100	2012-08-02 05:44:58.43658
30870	4	1-9271	30870	23	7	448	91	2012-08-02 05:44:58.43658
30875	4	1-8858	30874	28	45	1277	28	2012-08-02 05:44:58.43658
30876	4	1-9038	30874	45	97	1390	30	2012-08-02 05:44:58.43658
30877	4	1-7426	30874	97	23	1233	100	2012-08-02 05:44:58.43658
30874	4	1-9271	30874	23	28	5140	91	2012-08-02 05:44:58.43658
30878	1	1-2932	\N	5	5	3119	100	2012-08-02 05:44:58.43658
30880	4	1-9253	30879	41	39	3309	75	2012-08-02 05:44:59.287494
30881	4	1-1952	30879	39	17	989	54	2012-08-02 05:44:59.287494
30882	4	1-9057	30879	17	80	1725	44	2012-08-02 05:44:59.287494
30879	4	1-9273	30879	80	41	3440	55	2012-08-02 05:44:59.287494
30884	4	1-8925	30883	75	34	329	43	2012-08-02 05:44:59.287494
30885	4	1-8950	30883	34	94	453	57	2012-08-02 05:44:59.287494
30886	4	1-7918	30883	94	80	193	44	2012-08-02 05:44:59.287494
30883	4	1-9273	30883	80	75	440	55	2012-08-02 05:44:59.287494
30888	2	1-4482	30887	51	80	525	44	2012-08-02 05:44:59.287494
30887	2	1-9273	30887	80	51	1258	55	2012-08-02 05:44:59.287494
30890	5	1-6708	30889	65	65	1519	64	2012-08-02 05:44:59.685131
30891	5	1-9009	30889	65	68	911	41	2012-08-02 05:44:59.685131
30892	5	1-9261	30889	68	46	1092	39	2012-08-02 05:44:59.685131
30893	5	1-8251	30889	46	3	1029	83	2012-08-02 05:44:59.685131
30889	5	1-9274	30889	3	65	2279	98	2012-08-02 05:44:59.685131
30895	4	1-7954	30894	71	15	1929	59	2012-08-02 05:45:00.082907
30896	4	1-8661	30894	15	61	1677	69	2012-08-02 05:45:00.082907
30897	4	1-8838	30894	61	17	753	82	2012-08-02 05:45:00.082907
30894	4	1-9275	30894	17	71	2469	48	2012-08-02 05:45:00.082907
30899	4	1-9237	30898	32	79	4591	34	2012-08-02 05:45:00.911853
30900	4	1-9239	30898	79	95	6859	71	2012-08-02 05:45:00.911853
30901	4	1-7255	30898	95	63	808	8	2012-08-02 05:45:00.911853
30898	4	1-9277	30898	63	32	1382	3	2012-08-02 05:45:00.911853
30903	4	1-9025	30902	9	38	341	70	2012-08-02 05:45:03.783966
30904	4	1-9252	30902	38	95	63	11	2012-08-02 05:45:03.783966
30905	4	1-9217	30902	95	72	140	6	2012-08-02 05:45:03.783966
30902	4	1-9284	30902	72	9	69	8	2012-08-02 05:45:03.783966
30907	3	1-7155	30906	87	95	793	11	2012-08-02 05:45:03.783966
30908	3	1-9217	30906	95	72	2302	6	2012-08-02 05:45:03.783966
30906	3	1-9284	30906	72	87	1463	8	2012-08-02 05:45:03.783966
30910	4	1-9277	30909	63	24	2032	3	2012-08-02 05:45:03.783966
30911	4	1-8421	30909	24	23	1472	96	2012-08-02 05:45:03.783966
30912	4	1-9269	30909	23	72	1861	6	2012-08-02 05:45:03.783966
30909	4	1-9284	30909	72	63	1248	8	2012-08-02 05:45:03.783966
30914	5	1-8365	30913	57	87	2186	15	2012-08-02 05:45:04.192847
30915	5	1-9071	30913	87	2	1173	52	2012-08-02 05:45:04.192847
30916	5	1-5423	30913	2	100	733	7	2012-08-02 05:45:04.192847
30917	5	1-9003	30913	100	27	1091	75	2012-08-02 05:45:04.192847
30913	5	1-9285	30913	27	57	2315	64	2012-08-02 05:45:04.192847
30919	4	1-9110	30918	24	55	703	9	2012-08-02 05:45:06.238627
30920	4	1-8537	30918	55	81	698	13	2012-08-02 05:45:06.238627
30921	4	1-8918	30918	81	40	1745	16	2012-08-02 05:45:06.238627
30918	4	1-9290	30918	40	24	370	62	2012-08-02 05:45:06.238627
30923	3	1-9206	30922	30	89	4600	77	2012-08-02 05:45:09.033763
30924	3	1-8986	30922	89	8	3030	58	2012-08-02 05:45:09.033763
30922	3	1-9297	30922	8	30	5141	7	2012-08-02 05:45:09.033763
30926	4	1-9206	30925	30	91	219	77	2012-08-02 05:45:09.033763
30927	4	1-8088	30925	91	61	175	17	2012-08-02 05:45:09.033763
30928	4	1-7800	30925	61	8	135	58	2012-08-02 05:45:09.033763
30925	4	1-9297	30925	8	30	237	7	2012-08-02 05:45:09.033763
30930	4	1-7662	30929	90	92	99	95	2012-08-02 05:45:09.033763
30931	4	1-8862	30929	92	55	87	23	2012-08-02 05:45:09.033763
30932	4	1-8705	30929	55	8	42	58	2012-08-02 05:45:09.033763
30929	4	1-9297	30929	8	90	83	7	2012-08-02 05:45:09.033763
30934	5	1-8251	30933	46	35	762	83	2012-08-02 05:45:09.442621
30935	5	1-7682	30933	35	5	194	24	2012-08-02 05:45:09.442621
30936	5	1-7456	30933	5	47	176	26	2012-08-02 05:45:09.442621
30937	5	1-9106	30933	47	5	308	43	2012-08-02 05:45:09.442621
30933	5	1-9298	30933	5	46	826	39	2012-08-02 05:45:09.442621
30939	5	1-8721	30938	98	23	755	14	2012-08-02 05:45:09.442621
30940	5	1-8098	30938	23	39	233	13	2012-08-02 05:45:09.442621
30941	5	1-8465	30938	39	51	252	46	2012-08-02 05:45:09.442621
30942	5	1-9137	30938	51	5	206	43	2012-08-02 05:45:09.442621
30938	5	1-9298	30938	5	98	565	39	2012-08-02 05:45:09.442621
30944	4	1-8564	30943	87	67	1111	99	2012-08-02 05:45:09.851576
30945	4	1-7286	30943	67	69	1076	68	2012-08-02 05:45:09.851576
30943	4	1-9299	30943	71	87	1570	6	2012-08-02 05:45:09.851576
30946	4	1-9112	30943	69	71	2000	69	2012-08-02 05:45:09.851576
30948	4	1-5225	30947	61	53	873	59	2012-08-02 05:45:09.851576
30949	4	1-9105	30947	53	91	1580	42	2012-08-02 05:45:09.851576
30950	4	1-7953	30947	91	71	1650	69	2012-08-02 05:45:09.851576
30947	4	1-9299	30947	71	61	1431	6	2012-08-02 05:45:09.851576
30952	3	1-8855	30951	43	31	3957	87	2012-08-02 05:45:11.862225
30953	3	1-9288	30951	31	13	2525	75	2012-08-02 05:45:11.862225
30951	3	1-9304	30951	13	43	4583	79	2012-08-02 05:45:11.862225
30955	4	1-9088	30954	75	4	1050	97	2012-08-02 05:45:12.271379
30956	4	1-8874	30954	4	38	512	3	2012-08-02 05:45:12.271379
30957	4	1-8944	30954	38	28	375	38	2012-08-02 05:45:12.271379
30954	4	1-9305	30954	28	75	805	45	2012-08-02 05:45:12.271379
30959	4	1-9088	30958	75	73	754	97	2012-08-02 05:45:12.271379
30960	4	1-9227	30958	73	34	1076	62	2012-08-02 05:45:12.271379
30961	4	1-8939	30958	34	28	267	38	2012-08-02 05:45:12.271379
30958	4	1-9305	30958	28	75	575	45	2012-08-02 05:45:12.271379
30963	3	1-7638	30962	70	18	2605	94	2012-08-02 05:45:12.271379
30964	3	1-9179	30962	18	28	2970	38	2012-08-02 05:45:12.271379
30962	3	1-9305	30962	28	70	6846	45	2012-08-02 05:45:12.271379
30966	4	1-8961	30965	84	15	740	27	2012-08-02 05:45:12.271379
30967	4	1-9205	30965	15	46	1026	87	2012-08-02 05:45:12.271379
30968	4	1-8545	30965	46	28	722	38	2012-08-02 05:45:12.271379
30965	4	1-9305	30965	28	84	1678	45	2012-08-02 05:45:12.271379
30970	3	1-5423	30969	2	90	1838	7	2012-08-02 05:45:13.8628
30971	3	1-7662	30969	90	41	1615	95	2012-08-02 05:45:13.8628
30969	3	1-9309	30969	41	2	3445	52	2012-08-02 05:45:13.8628
30973	4	1-8799	30972	34	73	1732	35	2012-08-02 05:45:13.8628
30974	4	1-809	30972	73	90	690	7	2012-08-02 05:45:13.8628
30975	4	1-7662	30972	90	41	693	95	2012-08-02 05:45:13.8628
30972	4	1-9309	30972	41	34	1689	52	2012-08-02 05:45:13.8628
30977	4	1-8799	30976	34	49	1660	35	2012-08-02 05:45:13.8628
30978	4	1-3662	30976	49	83	1008	60	2012-08-02 05:45:13.8628
30979	4	1-8055	30976	83	41	593	95	2012-08-02 05:45:13.8628
30976	4	1-9309	30976	41	34	1530	52	2012-08-02 05:45:13.8628
30981	4	1-9052	30980	30	49	875	90	2012-08-02 05:45:13.8628
30982	4	1-8072	30980	49	83	495	60	2012-08-02 05:45:13.8628
30983	4	1-8055	30980	83	41	294	95	2012-08-02 05:45:13.8628
30980	4	1-9309	30980	41	30	766	52	2012-08-02 05:45:13.8628
30985	5	1-6659	30984	43	17	855	89	2012-08-02 05:45:17.110829
30986	5	1-8663	30984	17	5	385	24	2012-08-02 05:45:17.110829
30987	5	1-7456	30984	5	95	272	26	2012-08-02 05:45:17.110829
30988	5	1-8457	30984	95	65	128	55	2012-08-02 05:45:17.110829
30984	5	1-9317	30984	65	43	1493	52	2012-08-02 05:45:17.110829
30990	5	1-588	30989	47	100	1112	27	2012-08-02 05:45:17.110829
30991	5	1-7452	30989	100	94	792	13	2012-08-02 05:45:17.110829
30992	5	1-7937	30989	94	82	432	63	2012-08-02 05:45:17.110829
30993	5	1-9138	30989	82	65	444	55	2012-08-02 05:45:17.110829
30989	5	1-9317	30989	65	47	5565	52	2012-08-02 05:45:17.110829
30995	4	1-8977	30994	44	9	2371	11	2012-08-02 05:45:17.54183
30996	4	1-9211	30994	9	98	3175	58	2012-08-02 05:45:17.54183
30997	4	1-7694	30994	98	15	2404	81	2012-08-02 05:45:17.54183
30994	4	1-9318	30994	15	44	4490	4	2012-08-02 05:45:17.54183
30999	3	1-8977	30998	44	8	2849	11	2012-08-02 05:45:17.54183
31000	3	1-8107	30998	8	15	2508	81	2012-08-02 05:45:17.54183
30998	3	1-9318	30998	15	44	5027	4	2012-08-02 05:45:17.54183
31002	4	1-9038	31001	45	98	4071	30	2012-08-02 05:45:18.770261
31003	4	1-9026	31001	98	78	3537	41	2012-08-02 05:45:18.770261
31004	4	1-8487	31001	78	41	894	67	2012-08-02 05:45:18.770261
31001	4	1-9321	31001	41	45	4849	28	2012-08-02 05:45:18.770261
31006	5	1-7356	31005	88	96	1500	63	2012-08-02 05:45:19.189829
31007	5	1-8828	31005	96	68	1073	41	2012-08-02 05:45:19.189829
31008	5	1-9261	31005	68	87	1039	39	2012-08-02 05:45:19.189829
31009	5	1-9283	31005	87	10	799	61	2012-08-02 05:45:19.189829
31005	5	1-9322	31005	10	88	4811	95	2012-08-02 05:45:19.189829
31011	5	1-9033	31010	53	81	1635	13	2012-08-02 05:45:22.018036
31012	5	1-8918	31010	81	1	3672	16	2012-08-02 05:45:22.018036
31013	5	1-7768	31010	1	61	1648	32	2012-08-02 05:45:22.018036
31014	5	1-9158	31010	61	2	3050	68	2012-08-02 05:45:22.018036
31010	5	1-9329	31010	2	53	1783	49	2012-08-02 05:45:22.018036
31016	4	1-9110	31015	24	12	2558	9	2012-08-02 05:45:22.42701
31017	4	1-7843	31015	12	92	1628	46	2012-08-02 05:45:22.42701
31018	4	1-7789	31015	92	65	626	83	2012-08-02 05:45:22.42701
31015	4	1-9330	31015	65	24	2031	62	2012-08-02 05:45:22.42701
31020	4	1-9224	31019	95	2	319	94	2012-08-02 05:45:22.42701
31021	4	1-8964	31019	2	99	312	65	2012-08-02 05:45:22.42701
31022	4	1-5622	31019	99	65	154	83	2012-08-02 05:45:22.42701
31019	4	1-9330	31019	65	95	563	62	2012-08-02 05:45:22.42701
31024	3	1-8176	31023	42	1	249	16	2012-08-02 05:45:22.835957
31025	3	1-7768	31023	1	100	88	32	2012-08-02 05:45:22.835957
31023	3	1-9331	31023	100	42	529	93	2012-08-02 05:45:22.835957
31027	4	1-8658	31026	18	79	37	34	2012-08-02 05:45:22.835957
31028	4	1-9239	31026	79	5	52	71	2012-08-02 05:45:22.835957
31029	4	1-6922	31026	5	100	17	32	2012-08-02 05:45:22.835957
31026	4	1-9331	31026	100	18	107	93	2012-08-02 05:45:22.835957
31031	4	1-8878	31030	77	19	3913	94	2012-08-02 05:45:22.835957
31032	4	1-6150	31030	19	16	1884	59	2012-08-02 05:45:22.835957
31033	4	1-8067	31030	16	100	797	32	2012-08-02 05:45:22.835957
31030	4	1-9331	31030	100	77	5340	93	2012-08-02 05:45:22.835957
31035	4	1-6031	31034	86	81	1300	93	2012-08-02 05:45:23.642808
31036	4	1-8287	31034	81	3	470	83	2012-08-02 05:45:23.642808
31037	4	1-9274	31034	3	13	896	98	2012-08-02 05:45:23.642808
31034	4	1-9333	31034	13	86	2348	25	2012-08-02 05:45:23.642808
31039	4	1-6772	31038	31	6	1346	24	2012-08-02 05:45:25.745438
31040	4	1-8026	31038	6	24	1745	67	2012-08-02 05:45:25.745438
31041	4	1-9176	31038	24	15	1127	18	2012-08-02 05:45:25.745438
31038	4	1-9338	31038	15	31	6680	47	2012-08-02 05:45:25.745438
31043	3	1-5615	31042	27	51	85	19	2012-08-02 05:45:26.934841
31044	3	1-9129	31042	51	60	169	87	2012-08-02 05:45:26.934841
31042	3	1-9341	31042	60	27	1133	70	2012-08-02 05:45:26.934841
31046	2	1-5739	31045	94	14	265	58	2012-08-02 05:45:27.332705
31045	2	1-9342	31045	14	94	3738	81	2012-08-02 05:45:27.332705
31048	5	1-5465	31047	23	66	274	14	2012-08-02 05:45:28.504053
31049	5	1-8935	31047	66	87	258	15	2012-08-02 05:45:28.504053
31050	5	1-9071	31047	87	30	140	52	2012-08-02 05:45:28.504053
31047	5	1-9345	31047	73	23	280	68	2012-08-02 05:45:28.504053
31051	5	1-9052	31047	30	73	154	90	2012-08-02 05:45:28.504053
31053	4	1-8559	31052	12	86	308	12	2012-08-02 05:45:28.504053
31054	4	1-8015	31052	86	30	171	52	2012-08-02 05:45:28.504053
31055	4	1-9052	31052	30	73	208	90	2012-08-02 05:45:28.504053
31052	4	1-9345	31052	73	12	419	68	2012-08-02 05:45:28.504053
31057	5	1-9102	31056	41	40	2336	60	2012-08-02 05:45:28.504053
31058	5	1-7331	31056	40	87	1299	15	2012-08-02 05:45:28.504053
31059	5	1-9071	31056	87	30	866	52	2012-08-02 05:45:28.504053
31060	5	1-9052	31056	30	73	1178	90	2012-08-02 05:45:28.504053
31056	5	1-9345	31056	73	41	2644	68	2012-08-02 05:45:28.504053
31062	4	1-6667	31061	16	70	713	61	2012-08-02 05:45:29.41004
31063	4	1-7704	31061	70	4	513	57	2012-08-02 05:45:29.41004
31064	4	1-8881	31061	4	30	534	12	2012-08-02 05:45:29.41004
31061	4	1-9347	31061	30	16	1221	20	2012-08-02 05:45:29.41004
31066	4	1-7843	31065	12	34	2	46	2012-08-02 05:45:30.69134
31067	4	1-8492	31065	34	38	3	5	2012-08-02 05:45:30.69134
31068	4	1-7797	31065	38	92	2	14	2012-08-02 05:45:30.69134
31065	4	1-9350	31065	92	12	3	9	2012-08-02 05:45:30.69134
31070	4	1-8436	31069	51	19	2066	59	2012-08-02 05:45:30.69134
31071	4	1-8238	31069	19	38	2439	5	2012-08-02 05:45:30.69134
31072	4	1-7797	31069	38	92	1478	14	2012-08-02 05:45:30.69134
31069	4	1-9350	31069	92	51	3016	9	2012-08-02 05:45:30.69134
31074	4	1-9315	31073	27	62	5341	86	2012-08-02 05:45:30.69134
31075	4	1-9130	31073	62	46	1658	75	2012-08-02 05:45:30.69134
31076	4	1-8526	31073	46	92	1759	14	2012-08-02 05:45:30.69134
31073	4	1-9350	31073	92	27	3667	9	2012-08-02 05:45:30.69134
31078	3	1-8436	31077	51	54	722	59	2012-08-02 05:45:30.69134
31079	3	1-7060	31077	54	92	461	14	2012-08-02 05:45:30.69134
31077	3	1-9350	31077	92	51	996	9	2012-08-02 05:45:30.69134
31081	4	1-7206	31080	51	86	432	42	2012-08-02 05:45:30.69134
31082	4	1-8192	31080	86	65	108	72	2012-08-02 05:45:30.69134
31083	4	1-7203	31080	65	92	216	14	2012-08-02 05:45:30.69134
31080	4	1-9350	31080	92	51	484	9	2012-08-02 05:45:30.69134
31085	3	1-8107	31084	8	17	450	81	2012-08-02 05:45:31.498135
31086	3	1-6372	31084	17	30	439	92	2012-08-02 05:45:31.498135
31084	3	1-9352	31084	30	8	544	11	2012-08-02 05:45:31.498135
31088	4	1-9211	31087	9	84	964	58	2012-08-02 05:45:31.498135
31089	4	1-5256	31087	84	92	131	34	2012-08-02 05:45:31.498135
31090	4	1-9346	31087	92	30	574	92	2012-08-02 05:45:31.498135
31087	4	1-9352	31087	30	9	713	11	2012-08-02 05:45:31.498135
31092	3	1-8412	31091	53	81	2637	82	2012-08-02 05:45:31.498135
31093	3	1-7834	31091	81	30	2713	92	2012-08-02 05:45:31.498135
31091	3	1-9352	31091	30	53	3450	11	2012-08-02 05:45:31.498135
31095	3	1-9341	31094	60	93	5890	70	2012-08-02 05:45:31.896255
31096	3	1-5594	31094	93	27	772	29	2012-08-02 05:45:31.896255
31094	3	1-9353	31094	27	60	2927	87	2012-08-02 05:45:31.896255
31098	4	1-9288	31097	31	81	81	75	2012-08-02 05:45:31.896255
31099	4	1-8339	31097	81	41	73	86	2012-08-02 05:45:31.896255
31100	4	1-6766	31097	41	27	32	29	2012-08-02 05:45:31.896255
31097	4	1-9353	31097	27	31	223	87	2012-08-02 05:45:31.896255
31102	4	1-9288	31101	31	19	952	75	2012-08-02 05:45:31.896255
31103	4	1-2406	31101	19	36	164	9	2012-08-02 05:45:31.896255
31104	4	1-6831	31101	36	27	199	29	2012-08-02 05:45:31.896255
31101	4	1-9353	31101	27	31	1915	87	2012-08-02 05:45:31.896255
31106	4	1-8115	31105	48	22	650	97	2012-08-02 05:45:31.896255
31107	4	1-9302	31105	22	41	670	86	2012-08-02 05:45:31.896255
31108	4	1-6766	31105	41	27	410	29	2012-08-02 05:45:31.896255
31105	4	1-9353	31105	27	48	4053	87	2012-08-02 05:45:31.896255
31110	5	1-9291	31109	67	17	6187	25	2012-08-02 05:45:32.305078
31111	5	1-8733	31109	17	91	5188	53	2012-08-02 05:45:32.305078
31112	5	1-9339	31109	91	47	1647	26	2012-08-02 05:45:32.305078
31113	5	1-9106	31109	47	11	2464	43	2012-08-02 05:45:32.305078
31109	5	1-9354	31109	11	67	4170	82	2012-08-02 05:45:32.305078
31115	5	1-8821	31114	15	82	782	23	2012-08-02 05:45:32.305078
31116	5	1-2629	31114	82	91	537	53	2012-08-02 05:45:32.305078
31117	5	1-9339	31114	91	47	212	26	2012-08-02 05:45:32.305078
31118	5	1-9106	31114	47	11	394	43	2012-08-02 05:45:32.305078
31114	5	1-9354	31114	11	15	829	82	2012-08-02 05:45:32.305078
31120	5	1-7464	31119	97	47	1023	98	2012-08-02 05:45:33.520489
31121	5	1-7117	31119	47	91	1116	53	2012-08-02 05:45:33.520489
31122	5	1-9339	31119	91	47	364	26	2012-08-02 05:45:33.520489
31123	5	1-9106	31119	47	29	627	43	2012-08-02 05:45:33.520489
31119	5	1-9357	31119	29	97	2220	52	2012-08-02 05:45:33.520489
31125	4	1-7464	31124	97	97	2340	98	2012-08-02 05:45:33.520489
31126	4	1-6336	31124	97	51	1359	46	2012-08-02 05:45:33.520489
31127	4	1-9137	31124	51	29	1040	43	2012-08-02 05:45:33.520489
31124	4	1-9357	31124	29	97	4224	52	2012-08-02 05:45:33.520489
31129	5	1-8025	31128	46	83	120	45	2012-08-02 05:45:33.520489
31130	5	1-4729	31128	83	91	92	53	2012-08-02 05:45:33.520489
31131	5	1-9339	31128	91	47	35	26	2012-08-02 05:45:33.520489
31132	5	1-9106	31128	47	29	63	43	2012-08-02 05:45:33.520489
31128	5	1-9357	31128	29	46	260	52	2012-08-02 05:45:33.520489
31134	5	1-8882	31133	53	18	439	77	2012-08-02 05:45:35.553347
31135	5	1-7033	31133	18	41	194	55	2012-08-02 05:45:35.553347
31136	5	1-9253	31133	41	27	247	75	2012-08-02 05:45:35.553347
31137	5	1-9285	31133	27	35	600	64	2012-08-02 05:45:35.553347
31133	5	1-9362	31133	35	53	867	21	2012-08-02 05:45:35.553347
31139	5	1-9274	31138	3	72	3801	98	2012-08-02 05:45:38.768535
31140	5	1-8352	31138	72	51	1477	64	2012-08-02 05:45:38.768535
31141	5	1-5233	31138	51	63	412	8	2012-08-02 05:45:38.768535
31142	5	1-9277	31138	63	95	874	3	2012-08-02 05:45:38.768535
31138	5	1-9370	31138	95	3	1627	83	2012-08-02 05:45:38.768535
31144	4	1-9227	31143	73	95	3101	62	2012-08-02 05:45:39.177103
31145	4	1-9224	31143	95	2	2285	94	2012-08-02 05:45:39.177103
31146	4	1-8964	31143	2	20	2902	65	2012-08-02 05:45:39.177103
31143	4	1-9371	31143	20	73	2072	97	2012-08-02 05:45:39.177103
31148	4	1-9052	31147	30	73	2183	90	2012-08-02 05:45:40.403421
31149	4	1-9345	31147	73	68	4884	68	2012-08-02 05:45:40.403421
31150	4	1-7202	31147	68	59	1349	91	2012-08-02 05:45:40.403421
31147	4	1-9374	31147	59	30	1611	52	2012-08-02 05:45:40.403421
31152	4	1-8237	31151	32	53	2074	49	2012-08-02 05:45:42.425497
31153	4	1-9033	31151	53	95	1970	13	2012-08-02 05:45:42.425497
31154	4	1-9292	31151	95	39	3695	94	2012-08-02 05:45:42.425497
31151	4	1-9379	31151	39	32	3990	50	2012-08-02 05:45:42.425497
31155	3	1-9380	31155	66	74	2576	8	2012-08-02 05:45:42.822866
31156	3	1-8108	31155	74	51	1052	56	2012-08-02 05:45:42.822866
31157	3	1-5400	31155	51	66	278	100	2012-08-02 05:45:42.822866
31159	4	1-9277	31158	63	24	4910	3	2012-08-02 05:45:42.822866
31160	4	1-8421	31158	24	14	2172	96	2012-08-02 05:45:42.822866
31161	4	1-8873	31158	14	66	402	100	2012-08-02 05:45:42.822866
31158	4	1-9380	31158	66	63	4937	8	2012-08-02 05:45:42.822866
31163	3	1-3762	31162	50	51	189	56	2012-08-02 05:45:42.822866
31164	3	1-5400	31162	51	66	76	100	2012-08-02 05:45:42.822866
31162	3	1-9380	31162	66	50	1089	8	2012-08-02 05:45:42.822866
31166	3	1-7024	31165	81	88	633	95	2012-08-02 05:45:44.447245
31167	3	1-7356	31165	88	26	197	63	2012-08-02 05:45:44.447245
31165	3	1-9384	31165	26	81	1247	76	2012-08-02 05:45:44.447245
31169	4	1-8983	31168	5	56	1625	75	2012-08-02 05:45:44.447245
31170	4	1-6780	31168	56	77	1054	77	2012-08-02 05:45:44.447245
31171	4	1-5557	31168	77	26	362	63	2012-08-02 05:45:44.447245
31168	4	1-9384	31168	26	5	2362	76	2012-08-02 05:45:44.447245
31173	5	1-7043	31172	97	5	512	43	2012-08-02 05:45:44.447245
31174	5	1-9298	31172	5	55	1191	39	2012-08-02 05:45:44.447245
31175	5	1-8350	31172	55	49	417	9	2012-08-02 05:45:44.447245
31176	5	1-8405	31172	49	26	256	63	2012-08-02 05:45:44.447245
31172	5	1-9384	31172	26	97	1790	76	2012-08-02 05:45:44.447245
31178	3	1-4977	31177	37	88	691	95	2012-08-02 05:45:44.447245
31179	3	1-7356	31177	88	26	248	63	2012-08-02 05:45:44.447245
31177	3	1-9384	31177	26	37	1808	76	2012-08-02 05:45:44.447245
31180	1	1-4977	\N	37	37	1305	95	2012-08-02 05:45:44.447245
31182	4	1-9053	31181	82	40	1473	19	2012-08-02 05:45:45.667157
31183	4	1-5008	31181	40	51	1555	56	2012-08-02 05:45:45.667157
31184	4	1-5400	31181	51	37	1155	100	2012-08-02 05:45:45.667157
31181	4	1-9387	31181	37	82	1286	63	2012-08-02 05:45:45.667157
31186	4	1-7155	31185	87	36	2871	11	2012-08-02 05:45:46.868462
31187	4	1-8177	31185	36	88	2441	58	2012-08-02 05:45:46.868462
31188	4	1-8116	31185	88	55	1290	13	2012-08-02 05:45:46.868462
31185	4	1-9390	31185	55	87	4439	8	2012-08-02 05:45:46.868462
31190	4	1-2255	31189	74	32	2351	50	2012-08-02 05:45:46.868462
31191	4	1-8237	31189	32	53	1144	49	2012-08-02 05:45:46.868462
31192	4	1-9033	31189	53	55	1017	13	2012-08-02 05:45:46.868462
31189	4	1-9390	31189	55	74	3607	8	2012-08-02 05:45:46.868462
31194	4	1-7325	31193	21	96	103	26	2012-08-02 05:45:49.265963
31195	4	1-9107	31193	96	54	935	79	2012-08-02 05:45:49.265963
31196	4	1-6353	31193	54	11	160	4	2012-08-02 05:45:49.265963
31193	4	1-9396	31193	11	21	426	66	2012-08-02 05:45:49.265963
31198	4	1-9125	31197	60	15	623	86	2012-08-02 05:45:49.265963
31199	4	1-8646	31197	15	86	272	33	2012-08-02 05:45:49.265963
31200	4	1-9391	31197	86	11	217	4	2012-08-02 05:45:49.265963
31197	4	1-9396	31197	11	60	597	66	2012-08-02 05:45:49.265963
31202	4	1-9325	31201	70	65	454	72	2012-08-02 05:45:50.47021
31203	4	1-7203	31201	65	66	813	14	2012-08-02 05:45:50.47021
31204	4	1-8935	31201	66	78	775	15	2012-08-02 05:45:50.47021
31201	4	1-9399	31201	78	70	1051	54	2012-08-02 05:45:50.47021
31206	3	1-8665	31205	8	77	4172	93	2012-08-02 05:45:51.674757
31207	3	1-8878	31205	77	37	2757	94	2012-08-02 05:45:51.674757
31205	3	1-9402	31205	37	8	4657	22	2012-08-02 05:45:51.674757
31209	5	1-5619	31208	83	11	727	43	2012-08-02 05:45:51.674757
31210	5	1-8137	31208	11	53	425	49	2012-08-02 05:45:51.674757
31211	5	1-9033	31208	53	95	408	13	2012-08-02 05:45:51.674757
31212	5	1-9292	31208	95	37	773	94	2012-08-02 05:45:51.674757
31208	5	1-9402	31208	37	83	1730	22	2012-08-02 05:45:51.674757
31214	4	1-8863	31213	49	82	4332	47	2012-08-02 05:45:52.083327
31215	4	1-8303	31213	82	85	4270	10	2012-08-02 05:45:52.083327
31216	4	1-9233	31213	85	94	4598	34	2012-08-02 05:45:52.083327
31213	4	1-9403	31213	94	49	6111	33	2012-08-02 05:45:52.083327
31218	5	1-9052	31217	30	33	1967	90	2012-08-02 05:45:53.309948
31219	5	1-7964	31217	33	73	1181	97	2012-08-02 05:45:53.309948
31220	5	1-9227	31217	73	74	1737	62	2012-08-02 05:45:53.309948
31221	5	1-7973	31217	74	57	2141	44	2012-08-02 05:45:53.309948
31217	5	1-9406	31217	57	30	1460	52	2012-08-02 05:45:53.309948
31223	4	1-8907	31222	76	68	4143	21	2012-08-02 05:45:54.116261
31224	4	1-9312	31222	68	9	2744	41	2012-08-02 05:45:54.116261
31225	4	1-7403	31222	9	91	2061	16	2012-08-02 05:45:54.116261
31222	4	1-9408	31222	91	76	3504	23	2012-08-02 05:45:54.116261
31227	3	1-9228	31226	32	11	1002	33	2012-08-02 05:45:54.116261
31228	3	1-7495	31226	11	91	934	16	2012-08-02 05:45:54.116261
31226	3	1-9408	31226	91	32	1780	23	2012-08-02 05:45:54.116261
31230	4	1-7792	31229	5	95	48	13	2012-08-02 05:45:54.116261
31231	4	1-9292	31229	95	23	95	94	2012-08-02 05:45:54.116261
31232	4	1-8462	31229	23	91	53	16	2012-08-02 05:45:54.116261
31229	4	1-9408	31229	91	5	109	23	2012-08-02 05:45:54.116261
31234	4	1-8948	31233	30	15	434	59	2012-08-02 05:45:54.536268
31235	4	1-8661	31233	15	71	358	69	2012-08-02 05:45:54.536268
31236	4	1-9299	31233	71	49	253	6	2012-08-02 05:45:54.536268
31233	4	1-9409	31233	49	30	565	22	2012-08-02 05:45:54.536268
31238	3	1-7494	31237	65	15	2255	29	2012-08-02 05:45:55.310003
31239	3	1-8811	31237	15	11	1390	85	2012-08-02 05:45:55.310003
31237	3	1-9411	31237	11	65	9892	21	2012-08-02 05:45:55.310003
31241	3	1-9396	31240	11	92	1499	66	2012-08-02 05:45:56.127461
31242	3	1-3260	31240	92	70	969	88	2012-08-02 05:45:56.127461
31240	3	1-9413	31240	70	11	715	4	2012-08-02 05:45:56.127461
31244	3	1-9396	31243	11	91	2660	66	2012-08-02 05:45:56.127461
31245	3	1-8180	31243	91	70	1472	88	2012-08-02 05:45:56.127461
31243	3	1-9413	31243	70	11	1173	4	2012-08-02 05:45:56.127461
31247	4	1-9073	31246	79	21	311	86	2012-08-02 05:45:56.127461
31248	4	1-8643	31246	21	17	160	17	2012-08-02 05:45:56.127461
31249	4	1-8899	31246	17	70	360	88	2012-08-02 05:45:56.127461
31246	4	1-9413	31246	70	79	340	4	2012-08-02 05:45:56.127461
31251	4	1-8981	31250	27	21	3258	86	2012-08-02 05:45:56.127461
31252	4	1-8643	31250	21	17	1701	17	2012-08-02 05:45:56.127461
31253	4	1-8899	31250	17	70	3866	88	2012-08-02 05:45:56.127461
31250	4	1-9413	31250	70	27	3698	4	2012-08-02 05:45:56.127461
31255	4	1-8825	31254	78	53	1418	11	2012-08-02 05:45:56.525308
31256	4	1-8412	31254	53	67	662	82	2012-08-02 05:45:56.525308
31257	4	1-9291	31254	67	39	630	25	2012-08-02 05:45:56.525308
31254	4	1-9414	31254	39	78	3926	100	2012-08-02 05:45:56.525308
31259	3	1-9036	31258	51	52	1040	66	2012-08-02 05:45:56.525308
31260	3	1-7553	31258	52	39	223	25	2012-08-02 05:45:56.525308
31258	3	1-9414	31258	39	51	1395	100	2012-08-02 05:45:56.525308
31262	5	1-9387	31261	37	35	813	63	2012-08-02 05:45:56.525308
31263	5	1-9065	31261	35	61	602	69	2012-08-02 05:45:56.525308
31264	5	1-8838	31261	61	67	175	82	2012-08-02 05:45:56.525308
31265	5	1-9291	31261	67	39	185	25	2012-08-02 05:45:56.525308
31261	5	1-9414	31261	39	37	1280	100	2012-08-02 05:45:56.525308
31267	4	1-8575	31266	26	61	518	69	2012-08-02 05:45:56.525308
31268	4	1-8838	31266	61	67	154	82	2012-08-02 05:45:56.525308
31269	4	1-9291	31266	67	39	166	25	2012-08-02 05:45:56.525308
31266	4	1-9414	31266	39	26	1179	100	2012-08-02 05:45:56.525308
31271	4	1-8648	31270	25	57	594	98	2012-08-02 05:45:56.945369
31272	4	1-6911	31270	57	82	351	80	2012-08-02 05:45:56.945369
31273	4	1-8525	31270	82	100	606	37	2012-08-02 05:45:56.945369
31270	4	1-9415	31270	100	25	1155	51	2012-08-02 05:45:56.945369
31275	3	1-8648	31274	25	99	414	98	2012-08-02 05:45:56.945369
31276	3	1-7989	31274	99	100	290	37	2012-08-02 05:45:56.945369
31274	3	1-9415	31274	100	25	668	51	2012-08-02 05:45:56.945369
31278	3	1-9206	31277	30	65	1471	77	2012-08-02 05:45:57.752016
31279	3	1-9286	31277	65	52	1450	42	2012-08-02 05:45:57.752016
31277	3	1-9417	31277	52	30	1455	7	2012-08-02 05:45:57.752016
31281	3	1-6696	31280	53	8	458	55	2012-08-02 05:45:58.149764
31282	3	1-9041	31280	8	15	404	53	2012-08-02 05:45:58.149764
31280	3	1-9418	31280	15	53	3788	38	2012-08-02 05:45:58.149764
31284	4	1-3023	31283	29	21	884	43	2012-08-02 05:45:58.547606
31285	4	1-9310	31283	21	24	586	3	2012-08-02 05:45:58.547606
31286	4	1-8421	31283	24	50	219	96	2012-08-02 05:45:58.547606
31283	4	1-9419	31283	50	29	7302	6	2012-08-02 05:45:58.547606
31288	4	1-969	31287	75	48	205	51	2012-08-02 05:45:59.398575
31289	4	1-9388	31287	48	33	151	75	2012-08-02 05:45:59.398575
31290	4	1-7742	31287	33	51	125	46	2012-08-02 05:45:59.398575
31287	4	1-9421	31287	51	75	537	66	2012-08-02 05:45:59.398575
31292	4	1-7797	31291	38	66	1007	14	2012-08-02 05:46:01.784474
31293	4	1-8935	31291	66	78	1008	15	2012-08-02 05:46:01.784474
31294	4	1-9399	31291	78	52	1437	54	2012-08-02 05:46:01.784474
31291	4	1-9427	31291	52	38	1611	5	2012-08-02 05:46:01.784474
31296	3	1-5493	31295	88	85	1054	34	2012-08-02 05:46:02.182277
31297	3	1-9254	31295	85	5	862	35	2012-08-02 05:46:02.182277
31295	3	1-9428	31295	5	88	2678	67	2012-08-02 05:46:02.182277
31299	4	1-8726	31298	100	91	1087	53	2012-08-02 05:46:02.182277
31300	4	1-9339	31298	91	28	305	26	2012-08-02 05:46:02.182277
31301	4	1-9145	31298	28	5	193	35	2012-08-02 05:46:02.182277
31298	4	1-9428	31298	5	100	707	67	2012-08-02 05:46:02.182277
31303	3	1-6273	31302	77	17	1044	82	2012-08-02 05:46:04.601967
31304	3	1-9275	31302	17	81	3191	48	2012-08-02 05:46:04.601967
31302	3	1-9434	31302	81	77	3116	20	2012-08-02 05:46:04.601967
31306	4	1-9186	31305	65	65	2874	91	2012-08-02 05:46:06.215873
31307	4	1-7835	31305	65	48	2966	37	2012-08-02 05:46:06.215873
31308	4	1-9014	31305	48	5	826	76	2012-08-02 05:46:06.215873
31305	4	1-9438	31305	5	65	8530	94	2012-08-02 05:46:06.215873
31310	4	1-8574	31309	50	56	2350	65	2012-08-02 05:46:07.089519
31311	4	1-9097	31309	56	6	3818	57	2012-08-02 05:46:07.089519
31312	4	1-9226	31309	6	40	4469	16	2012-08-02 05:46:07.089519
31309	4	1-9440	31309	40	50	3169	24	2012-08-02 05:46:07.089519
31314	4	1-8694	31313	22	95	2330	76	2012-08-02 05:46:07.089519
31315	4	1-7759	31313	95	98	642	96	2012-08-02 05:46:07.089519
31316	4	1-9232	31313	98	40	1545	16	2012-08-02 05:46:07.089519
31313	4	1-9440	31313	40	22	1671	24	2012-08-02 05:46:07.089519
31318	3	1-8375	31317	71	36	4513	6	2012-08-02 05:46:07.089519
31319	3	1-7799	31317	36	40	2878	16	2012-08-02 05:46:07.089519
31317	3	1-9440	31317	40	71	3183	24	2012-08-02 05:46:07.089519
31321	5	1-9275	31320	17	81	1613	48	2012-08-02 05:46:09.629741
31322	5	1-9434	31320	81	37	1453	20	2012-08-02 05:46:09.629741
31323	5	1-7934	31320	37	62	725	100	2012-08-02 05:46:09.629741
31324	5	1-9264	31320	62	86	415	62	2012-08-02 05:46:09.629741
31320	5	1-9446	31320	86	17	572	82	2012-08-02 05:46:09.629741
31326	5	1-7555	31325	73	66	2109	15	2012-08-02 05:46:09.629741
31327	5	1-9069	31325	66	37	2657	20	2012-08-02 05:46:09.629741
31328	5	1-7934	31325	37	62	1716	100	2012-08-02 05:46:09.629741
31329	5	1-9264	31325	62	86	1272	62	2012-08-02 05:46:09.629741
31325	5	1-9446	31325	86	73	2269	82	2012-08-02 05:46:09.629741
31331	4	1-9362	31330	35	6	546	21	2012-08-02 05:46:10.438541
31332	4	1-8968	31330	6	16	385	36	2012-08-02 05:46:10.438541
31333	4	1-6829	31330	16	29	302	31	2012-08-02 05:46:10.438541
31330	4	1-9448	31330	29	35	359	64	2012-08-02 05:46:10.438541
31335	4	1-9009	31334	65	14	3223	41	2012-08-02 05:46:10.438541
31336	4	1-4513	31334	14	25	1401	43	2012-08-02 05:46:10.438541
31337	4	1-9028	31334	25	29	4330	31	2012-08-02 05:46:10.438541
31334	4	1-9448	31334	29	65	5331	64	2012-08-02 05:46:10.438541
31339	3	1-9069	31338	66	77	526	20	2012-08-02 05:46:11.277968
31340	3	1-6419	31338	77	82	227	46	2012-08-02 05:46:11.277968
31338	3	1-9450	31338	82	66	532	15	2012-08-02 05:46:11.277968
31342	4	1-9185	31341	61	44	295	98	2012-08-02 05:46:12.084552
31343	4	1-5848	31341	44	28	575	85	2012-08-02 05:46:12.084552
31344	4	1-9433	31341	28	19	520	80	2012-08-02 05:46:12.084552
31341	4	1-9452	31341	19	61	396	74	2012-08-02 05:46:12.084552
31346	2	1-8512	31345	66	9	750	91	2012-08-02 05:46:12.493422
31345	2	1-9453	31345	9	66	2696	59	2012-08-02 05:46:12.493422
31348	4	1-9358	31347	48	8	3026	72	2012-08-02 05:46:12.493422
31349	4	1-9234	31347	8	6	3271	73	2012-08-02 05:46:12.493422
31350	4	1-6910	31347	6	9	591	91	2012-08-02 05:46:12.493422
31347	4	1-9453	31347	9	48	3557	59	2012-08-02 05:46:12.493422
31352	2	1-7556	31351	28	82	2166	95	2012-08-02 05:46:13.71963
31351	2	1-9456	31351	82	28	2322	71	2012-08-02 05:46:13.71963
31354	2	1-9132	31353	24	26	537	87	2012-08-02 05:46:14.90182
31353	2	1-9459	31353	26	24	507	28	2012-08-02 05:46:14.90182
31356	4	1-9232	31355	98	47	806	16	2012-08-02 05:46:15.31099
31357	4	1-9360	31355	47	7	445	31	2012-08-02 05:46:15.31099
31358	4	1-8468	31355	7	10	444	2	2012-08-02 05:46:15.31099
31355	4	1-9460	31355	10	98	327	96	2012-08-02 05:46:15.31099
31360	4	1-7823	31359	2	26	2761	87	2012-08-02 05:46:15.31099
31361	4	1-9459	31359	26	27	2176	28	2012-08-02 05:46:15.31099
31362	4	1-9375	31359	27	10	5063	2	2012-08-02 05:46:15.31099
31359	4	1-9460	31359	10	2	3791	96	2012-08-02 05:46:15.31099
31363	4	1-9461	31363	48	94	489	61	2012-08-02 05:46:15.708524
31364	4	1-8827	31363	94	92	751	84	2012-08-02 05:46:15.708524
31365	4	1-9365	31363	92	65	578	77	2012-08-02 05:46:15.708524
31366	4	1-9286	31363	65	48	623	42	2012-08-02 05:46:15.708524
31368	4	1-7768	31367	1	61	169	32	2012-08-02 05:46:16.526328
31369	4	1-9158	31367	61	69	301	68	2012-08-02 05:46:16.526328
31370	4	1-9112	31367	69	62	537	69	2012-08-02 05:46:16.526328
31367	4	1-9463	31367	62	1	390	16	2012-08-02 05:46:16.526328
31372	3	1-9408	31371	91	57	3379	23	2012-08-02 05:46:16.526328
31373	3	1-7882	31371	57	62	1973	69	2012-08-02 05:46:16.526328
31371	3	1-9463	31371	62	91	1625	16	2012-08-02 05:46:16.526328
31375	4	1-9337	31374	82	26	4082	87	2012-08-02 05:46:18.558769
31376	4	1-9459	31374	26	56	2826	28	2012-08-02 05:46:18.558769
31377	4	1-9104	31374	56	55	3066	21	2012-08-02 05:46:18.558769
31374	4	1-9468	31374	55	82	2316	97	2012-08-02 05:46:18.558769
31379	5	1-9302	31378	22	41	2157	86	2012-08-02 05:46:18.558769
31380	5	1-8355	31378	41	48	2004	6	2012-08-02 05:46:18.558769
31381	5	1-8140	31378	48	35	1248	64	2012-08-02 05:46:18.558769
31382	5	1-9362	31378	35	55	1848	21	2012-08-02 05:46:18.558769
31378	5	1-9468	31378	55	22	1740	97	2012-08-02 05:46:18.558769
31384	5	1-9337	31383	82	52	3662	87	2012-08-02 05:46:18.558769
31385	5	1-8270	31383	52	25	872	13	2012-08-02 05:46:18.558769
31386	5	1-9467	31383	25	35	1007	64	2012-08-02 05:46:18.558769
31387	5	1-9362	31383	35	55	1579	21	2012-08-02 05:46:18.558769
31383	5	1-9468	31383	55	82	1574	97	2012-08-02 05:46:18.558769
31389	4	1-4356	31388	96	25	751	13	2012-08-02 05:46:18.558769
31390	4	1-9467	31388	25	35	888	64	2012-08-02 05:46:18.558769
31391	4	1-9362	31388	35	55	1426	21	2012-08-02 05:46:18.558769
31388	4	1-9468	31388	55	96	1456	97	2012-08-02 05:46:18.558769
31393	5	1-9369	31392	90	85	1509	25	2012-08-02 05:46:19.365429
31394	5	1-5608	31392	85	5	322	78	2012-08-02 05:46:19.365429
31395	5	1-8994	31392	5	45	230	28	2012-08-02 05:46:19.365429
31396	5	1-9038	31392	45	100	176	30	2012-08-02 05:46:19.365429
31392	5	1-9470	31392	100	90	3907	44	2012-08-02 05:46:19.365429
31398	5	1-9406	31397	57	77	390	52	2012-08-02 05:46:19.365429
31399	5	1-8056	31397	77	18	174	31	2012-08-02 05:46:19.365429
31400	5	1-7888	31397	18	45	45	28	2012-08-02 05:46:19.365429
31401	5	1-9038	31397	45	100	36	30	2012-08-02 05:46:19.365429
31397	5	1-9470	31397	100	57	827	44	2012-08-02 05:46:19.365429
31403	4	1-9369	31402	90	92	177	25	2012-08-02 05:46:19.365429
31404	4	1-8697	31402	92	92	42	56	2012-08-02 05:46:19.365429
31405	4	1-4773	31402	92	100	15	30	2012-08-02 05:46:19.365429
31402	4	1-9470	31402	100	90	384	44	2012-08-02 05:46:19.365429
31407	3	1-9340	31406	86	86	1380	77	2012-08-02 05:46:21.420603
31408	3	1-9367	31406	86	15	5245	40	2012-08-02 05:46:21.420603
31406	3	1-9475	31406	15	86	4087	91	2012-08-02 05:46:21.420603
31410	3	1-8583	31409	44	7	871	65	2012-08-02 05:46:21.420603
31411	3	1-9244	31409	7	15	2661	40	2012-08-02 05:46:21.420603
31409	3	1-9475	31409	15	44	3025	91	2012-08-02 05:46:21.420603
31413	4	1-5493	31412	88	92	1813	34	2012-08-02 05:46:23.000474
31414	4	1-9346	31412	92	38	6044	92	2012-08-02 05:46:23.000474
31415	4	1-9457	31412	38	11	2395	84	2012-08-02 05:46:23.000474
31412	4	1-9479	31412	11	88	4438	67	2012-08-02 05:46:23.000474
31417	4	1-8230	31416	90	92	519	34	2012-08-02 05:46:23.000474
31418	4	1-9346	31416	92	38	2240	92	2012-08-02 05:46:23.000474
31419	4	1-9457	31416	38	11	1149	84	2012-08-02 05:46:23.000474
31416	4	1-9479	31416	11	90	2757	67	2012-08-02 05:46:23.000474
31421	4	1-6224	31420	21	35	1554	38	2012-08-02 05:46:23.000474
31422	4	1-9455	31420	35	26	1084	22	2012-08-02 05:46:23.000474
31423	4	1-7210	31420	26	11	1081	84	2012-08-02 05:46:23.000474
31420	4	1-9479	31420	11	21	2738	67	2012-08-02 05:46:23.000474
31425	4	1-2829	31424	96	82	614	46	2012-08-02 05:46:23.807004
31426	4	1-9450	31424	82	18	1846	15	2012-08-02 05:46:23.807004
31427	4	1-9410	31424	18	26	1173	63	2012-08-02 05:46:23.807004
31424	4	1-9481	31424	26	96	2181	35	2012-08-02 05:46:23.807004
31429	3	1-7937	31428	94	77	2853	63	2012-08-02 05:46:25.431323
31430	3	1-9431	31428	77	89	6251	54	2012-08-02 05:46:25.431323
31428	3	1-9485	31428	89	94	6393	13	2012-08-02 05:46:25.431323
31432	4	1-9403	31431	94	49	1958	33	2012-08-02 05:46:25.839997
31433	4	1-8863	31431	49	82	1603	47	2012-08-02 05:46:25.839997
31434	4	1-8303	31431	82	89	1826	10	2012-08-02 05:46:25.839997
31431	4	1-9486	31431	89	94	1275	34	2012-08-02 05:46:25.839997
31436	4	1-9403	31435	94	39	1383	33	2012-08-02 05:46:25.839997
31437	4	1-5355	31435	39	90	1025	54	2012-08-02 05:46:25.839997
31438	4	1-8490	31435	90	89	1129	10	2012-08-02 05:46:25.839997
31435	4	1-9486	31435	89	94	843	34	2012-08-02 05:46:25.839997
31440	3	1-8663	31439	17	84	1499	24	2012-08-02 05:46:27.033076
31441	3	1-8113	31439	84	39	1353	68	2012-08-02 05:46:27.033076
31439	3	1-9489	31439	39	17	3629	89	2012-08-02 05:46:27.033076
31443	4	1-3444	31442	36	4	2036	57	2012-08-02 05:46:27.441921
31444	4	1-8881	31442	4	42	2308	12	2012-08-02 05:46:27.441921
31445	4	1-6855	31442	42	67	2434	44	2012-08-02 05:46:27.441921
31442	4	1-9490	31442	67	36	3996	90	2012-08-02 05:46:27.441921
31447	4	1-9052	31446	30	11	1503	90	2012-08-02 05:46:27.839817
31448	4	1-8030	31446	11	26	231	22	2012-08-02 05:46:27.839817
31449	4	1-7210	31446	26	18	147	84	2012-08-02 05:46:27.839817
31446	4	1-9491	31446	18	30	2011	52	2012-08-02 05:46:27.839817
31451	3	1-6503	31450	82	74	593	11	2012-08-02 05:46:27.839817
31452	3	1-5806	31450	74	18	255	84	2012-08-02 05:46:27.839817
31450	3	1-9491	31450	18	82	4230	52	2012-08-02 05:46:27.839817
31454	4	1-9423	31453	1	89	3875	10	2012-08-02 05:46:28.248583
31455	4	1-9486	31453	89	61	1897	34	2012-08-02 05:46:28.248583
31456	4	1-8426	31453	61	92	1811	70	2012-08-02 05:46:28.248583
31453	4	1-9492	31453	92	1	3066	45	2012-08-02 05:46:28.248583
31458	2	1-8095	31457	39	92	1803	70	2012-08-02 05:46:28.248583
31457	2	1-9492	31457	92	39	3343	45	2012-08-02 05:46:28.248583
31460	4	1-9324	31459	64	68	3549	86	2012-08-02 05:46:28.248583
31461	4	1-9398	31459	68	58	2360	58	2012-08-02 05:46:28.248583
31462	4	1-7438	31459	58	92	1350	70	2012-08-02 05:46:28.248583
31459	4	1-9492	31459	92	64	2705	45	2012-08-02 05:46:28.248583
31464	4	1-7452	31463	100	25	305	13	2012-08-02 05:46:29.05505
31465	4	1-9467	31463	25	35	287	64	2012-08-02 05:46:29.05505
31466	4	1-9362	31463	35	38	367	21	2012-08-02 05:46:29.05505
31463	4	1-9494	31463	38	100	434	27	2012-08-02 05:46:29.05505
31468	5	1-9165	31467	17	49	1210	75	2012-08-02 05:46:29.05505
31469	5	1-8702	31467	49	89	1048	45	2012-08-02 05:46:29.05505
31470	5	1-8686	31467	89	35	689	64	2012-08-02 05:46:29.05505
31471	5	1-9362	31467	35	38	1005	21	2012-08-02 05:46:29.05505
31467	5	1-9494	31467	38	17	1356	27	2012-08-02 05:46:29.05505
31473	5	1-8612	31472	43	63	1201	50	2012-08-02 05:46:29.05505
31474	5	1-9070	31472	63	48	765	6	2012-08-02 05:46:29.05505
31475	5	1-8140	31472	48	35	509	64	2012-08-02 05:46:29.05505
31476	5	1-9362	31472	35	38	807	21	2012-08-02 05:46:29.05505
31472	5	1-9494	31472	38	43	1183	27	2012-08-02 05:46:29.05505
31478	4	1-9357	31477	29	7	2565	52	2012-08-02 05:46:29.475141
31479	4	1-7590	31477	7	37	1345	68	2012-08-02 05:46:29.475141
31480	4	1-9394	31477	37	76	1000	78	2012-08-02 05:46:29.475141
31477	4	1-9495	31477	76	29	637	43	2012-08-02 05:46:29.475141
31482	5	1-9310	31481	21	100	1196	3	2012-08-02 05:46:29.475141
31483	5	1-7141	31481	100	61	867	32	2012-08-02 05:46:29.475141
31484	5	1-9158	31481	61	37	1614	68	2012-08-02 05:46:29.475141
31485	5	1-9394	31481	37	76	1202	78	2012-08-02 05:46:29.475141
31481	5	1-9495	31481	76	21	767	43	2012-08-02 05:46:29.475141
31487	4	1-9298	31486	5	87	3440	39	2012-08-02 05:46:29.475141
31488	4	1-9283	31486	87	53	3395	61	2012-08-02 05:46:29.475141
31489	4	1-8678	31486	53	76	1891	78	2012-08-02 05:46:29.475141
31486	4	1-9495	31486	76	5	1275	43	2012-08-02 05:46:29.475141
31491	4	1-8988	31490	81	100	4294	27	2012-08-02 05:46:30.679765
31492	4	1-7452	31490	100	45	3221	13	2012-08-02 05:46:30.679765
31493	4	1-9335	31490	45	79	1439	43	2012-08-02 05:46:30.679765
31490	4	1-9498	31490	79	81	4533	18	2012-08-02 05:46:30.679765
31495	4	1-9091	31494	94	53	1742	49	2012-08-02 05:46:30.679765
31496	4	1-9033	31494	53	45	1581	13	2012-08-02 05:46:30.679765
31497	4	1-9335	31494	45	79	750	43	2012-08-02 05:46:30.679765
31494	4	1-9498	31494	79	94	2508	18	2012-08-02 05:46:30.679765
31499	4	1-9469	31498	84	20	3059	10	2012-08-02 05:46:30.679765
31500	4	1-8546	31498	20	76	980	78	2012-08-02 05:46:30.679765
31501	4	1-9495	31498	76	79	651	43	2012-08-02 05:46:30.679765
31498	4	1-9498	31498	79	84	2306	18	2012-08-02 05:46:30.679765
31503	3	1-9180	31502	78	20	1125	82	2012-08-02 05:46:31.508141
31504	3	1-5830	31502	20	79	1198	80	2012-08-02 05:46:31.508141
31502	3	1-9500	31502	79	78	1670	27	2012-08-02 05:46:31.508141
31506	4	1-9205	31505	15	46	1188	87	2012-08-02 05:46:31.508141
31507	4	1-8545	31505	46	38	761	38	2012-08-02 05:46:31.508141
31508	4	1-4144	31505	38	79	633	80	2012-08-02 05:46:31.508141
31505	4	1-9500	31505	79	15	940	27	2012-08-02 05:46:31.508141
31510	4	1-8612	31509	43	22	907	50	2012-08-02 05:46:31.508141
31511	4	1-7623	31509	22	28	729	85	2012-08-02 05:46:31.508141
31512	4	1-9433	31509	28	79	639	80	2012-08-02 05:46:31.508141
31509	4	1-9500	31509	79	43	953	27	2012-08-02 05:46:31.508141
31514	5	1-8353	31513	90	82	463	46	2012-08-02 05:46:32.31477
31515	5	1-9450	31513	82	18	1350	15	2012-08-02 05:46:32.31477
31516	5	1-9410	31513	18	67	832	63	2012-08-02 05:46:32.31477
31517	5	1-8856	31513	67	43	833	89	2012-08-02 05:46:32.31477
31513	5	1-9502	31513	43	90	2194	84	2012-08-02 05:46:32.31477
31519	3	1-6991	31518	8	67	1787	44	2012-08-02 05:46:32.701571
31520	3	1-9490	31518	67	74	2969	90	2012-08-02 05:46:32.701571
31518	3	1-9503	31518	74	8	2118	19	2012-08-02 05:46:32.701571
31522	3	1-6224	31521	21	65	836	38	2012-08-02 05:46:33.961183
31523	3	1-8579	31521	65	35	218	72	2012-08-02 05:46:33.961183
31521	3	1-9506	31521	35	21	3776	67	2012-08-02 05:46:33.961183
31525	3	1-9494	31524	38	43	2180	27	2012-08-02 05:46:35.584983
31526	3	1-8612	31524	43	5	1561	50	2012-08-02 05:46:35.584983
31524	3	1-9510	31524	5	38	2106	21	2012-08-02 05:46:35.584983
31528	4	1-6143	31527	67	48	2353	42	2012-08-02 05:46:35.584983
31529	4	1-9461	31527	48	75	1754	61	2012-08-02 05:46:35.584983
31530	4	1-9351	31527	75	5	2269	50	2012-08-02 05:46:35.584983
31527	4	1-9510	31527	5	67	3991	21	2012-08-02 05:46:35.584983
31532	4	1-9205	31531	15	26	611	87	2012-08-02 05:46:36.767224
31533	4	1-9459	31531	26	66	410	28	2012-08-02 05:46:36.767224
31534	4	1-9303	31531	66	49	158	11	2012-08-02 05:46:36.767224
31531	4	1-9513	31531	49	15	621	27	2012-08-02 05:46:36.767224
31536	4	1-8452	31535	33	20	1469	41	2012-08-02 05:46:37.176028
31537	4	1-7414	31535	20	4	1185	56	2012-08-02 05:46:37.176028
31538	4	1-8973	31535	4	64	1305	12	2012-08-02 05:46:37.176028
31535	4	1-9514	31535	64	33	2231	38	2012-08-02 05:46:37.176028
31540	5	1-6696	31539	53	8	1117	55	2012-08-02 05:46:37.176028
31541	5	1-9041	31539	8	46	3244	53	2012-08-02 05:46:37.176028
31542	5	1-8358	31539	46	4	1118	56	2012-08-02 05:46:37.176028
31543	5	1-8973	31539	4	64	1423	12	2012-08-02 05:46:37.176028
31539	5	1-9514	31539	64	53	2814	38	2012-08-02 05:46:37.176028
31545	5	1-4896	31544	42	74	1357	90	2012-08-02 05:46:37.176028
31546	5	1-9503	31544	74	40	987	19	2012-08-02 05:46:37.176028
31547	5	1-5008	31544	40	4	1057	56	2012-08-02 05:46:37.176028
31548	5	1-8973	31544	4	64	1366	12	2012-08-02 05:46:37.176028
31544	5	1-9514	31544	64	42	2741	38	2012-08-02 05:46:37.176028
31550	3	1-9148	31549	21	100	646	67	2012-08-02 05:46:37.609017
31551	3	1-8726	31549	100	51	991	53	2012-08-02 05:46:37.609017
31549	3	1-9515	31549	51	21	1350	78	2012-08-02 05:46:37.609017
31553	2	1-3007	31552	13	19	1482	16	2012-08-02 05:46:38.006682
31552	2	1-9516	31552	19	13	8001	73	2012-08-02 05:46:38.006682
31555	4	1-9177	31554	75	28	1449	79	2012-08-02 05:46:39.200116
31556	4	1-7652	31554	28	39	1124	93	2012-08-02 05:46:39.200116
31557	4	1-8231	31554	39	11	654	97	2012-08-02 05:46:39.200116
31554	4	1-9519	31554	11	75	1240	78	2012-08-02 05:46:39.200116
31559	4	1-7156	31558	58	65	359	38	2012-08-02 05:46:39.608678
31560	4	1-8579	31558	65	8	245	72	2012-08-02 05:46:39.608678
31561	4	1-9234	31558	8	98	296	73	2012-08-02 05:46:39.608678
31558	4	1-9520	31558	98	58	288	1	2012-08-02 05:46:39.608678
31563	3	1-9298	31562	5	17	962	39	2012-08-02 05:46:40.017559
31564	3	1-8578	31562	17	43	130	84	2012-08-02 05:46:40.017559
31562	3	1-9521	31562	43	5	1336	43	2012-08-02 05:46:40.017559
31566	4	1-9310	31565	21	51	826	3	2012-08-02 05:46:40.017559
31567	4	1-8985	31565	51	86	430	51	2012-08-02 05:46:40.017559
31568	4	1-8687	31565	86	43	108	84	2012-08-02 05:46:40.017559
31565	4	1-9521	31565	43	21	1448	43	2012-08-02 05:46:40.017559
31570	5	1-7907	31569	78	80	78	75	2012-08-02 05:46:40.835476
31571	5	1-9146	31569	80	8	134	72	2012-08-02 05:46:40.835476
31569	5	1-9523	31569	20	78	107	11	2012-08-02 05:46:40.835476
31572	5	1-9234	31569	8	98	172	73	2012-08-02 05:46:40.835476
31573	5	1-9520	31569	98	20	179	1	2012-08-02 05:46:40.835476
31575	4	1-9511	31574	72	48	5722	42	2012-08-02 05:46:41.254949
31576	4	1-9461	31574	48	75	3573	61	2012-08-02 05:46:41.254949
31577	4	1-9351	31574	75	53	3871	50	2012-08-02 05:46:41.254949
31574	4	1-9524	31574	53	72	6340	58	2012-08-02 05:46:41.254949
31579	4	1-9421	31578	51	21	5199	66	2012-08-02 05:46:43.2656
31580	4	1-7325	31578	21	19	1247	26	2012-08-02 05:46:43.2656
31581	4	1-5175	31578	19	66	720	74	2012-08-02 05:46:43.2656
31578	4	1-9529	31578	66	51	1201	46	2012-08-02 05:46:43.2656
31582	1	1-3858	\N	73	73	2471	46	2012-08-02 05:46:43.2656
31584	3	1-5226	31583	41	98	347	73	2012-08-02 05:46:44.480894
31585	3	1-9520	31583	98	27	194	1	2012-08-02 05:46:44.480894
31583	3	1-9532	31583	27	41	3053	85	2012-08-02 05:46:44.480894
31586	1	1-5226	\N	41	41	113	73	2012-08-02 05:46:44.480894
31587	1	1-3094	\N	3	3	478	73	2012-08-02 05:46:44.480894
31589	5	1-8053	31588	66	89	561	10	2012-08-02 05:46:44.900839
31590	5	1-9486	31588	89	94	369	34	2012-08-02 05:46:44.900839
31591	5	1-9403	31588	94	49	533	33	2012-08-02 05:46:44.900839
31592	5	1-8863	31588	49	52	411	47	2012-08-02 05:46:44.900839
31588	5	1-9533	31588	52	66	661	41	2012-08-02 05:46:44.900839
31594	4	1-7044	31593	53	71	992	74	2012-08-02 05:46:46.193565
31595	4	1-9095	31593	71	70	1140	88	2012-08-02 05:46:46.193565
31596	4	1-9413	31593	70	23	1094	4	2012-08-02 05:46:46.193565
31593	4	1-9536	31593	23	53	1700	94	2012-08-02 05:46:46.193565
31598	3	1-385	31597	72	27	1116	9	2012-08-02 05:46:47.010931
31599	3	1-9315	31597	27	91	1343	86	2012-08-02 05:46:47.010931
31597	3	1-9538	31597	91	72	4323	19	2012-08-02 05:46:47.010931
31601	4	1-5439	31600	12	18	2030	15	2012-08-02 05:46:48.22651
31602	4	1-9410	31600	18	34	1187	63	2012-08-02 05:46:48.22651
31603	4	1-8978	31600	34	58	1832	91	2012-08-02 05:46:48.22651
31600	4	1-9541	31600	58	12	3831	17	2012-08-02 05:46:48.22651
31605	4	1-8852	31604	29	18	1087	77	2012-08-02 05:46:48.22651
31606	4	1-7033	31604	18	79	511	55	2012-08-02 05:46:48.22651
31607	4	1-8791	31604	79	58	537	91	2012-08-02 05:46:48.22651
31604	4	1-9541	31604	58	29	1202	17	2012-08-02 05:46:48.22651
31609	4	1-9028	31608	25	86	1020	31	2012-08-02 05:46:49.022068
31610	4	1-8829	31608	86	28	1035	85	2012-08-02 05:46:49.022068
31611	4	1-9433	31608	28	72	693	80	2012-08-02 05:46:49.022068
31608	4	1-9543	31608	72	25	458	43	2012-08-02 05:46:49.022068
31613	4	1-8950	31612	34	4	461	57	2012-08-02 05:46:49.022068
31614	4	1-8881	31612	4	56	459	12	2012-08-02 05:46:49.022068
31615	4	1-8644	31612	56	72	386	80	2012-08-02 05:46:49.022068
31612	4	1-9543	31612	72	34	307	43	2012-08-02 05:46:49.022068
31617	4	1-9518	31616	8	85	1593	68	2012-08-02 05:46:50.635152
31618	4	1-8719	31616	85	74	1105	7	2012-08-02 05:46:50.635152
31619	4	1-8707	31616	74	32	1591	31	2012-08-02 05:46:50.635152
31616	4	1-9547	31616	32	8	1846	49	2012-08-02 05:46:50.635152
31621	2	1-8919	31620	3	32	2510	31	2012-08-02 05:46:50.635152
31620	2	1-9547	31620	32	3	2986	49	2012-08-02 05:46:50.635152
31623	4	1-9081	31622	38	69	2483	4	2012-08-02 05:46:51.430417
31624	4	1-8001	31622	69	40	1802	60	2012-08-02 05:46:51.430417
31625	4	1-7331	31622	40	90	637	15	2012-08-02 05:46:51.430417
31622	4	1-9549	31622	90	38	4240	82	2012-08-02 05:46:51.430417
31627	4	1-9291	31626	67	29	1812	25	2012-08-02 05:46:51.430417
31628	4	1-9064	31626	29	56	444	47	2012-08-02 05:46:51.430417
31629	4	1-3011	31626	56	90	212	15	2012-08-02 05:46:51.430417
31626	4	1-9549	31626	90	67	1477	82	2012-08-02 05:46:51.430417
31631	3	1-409	31630	67	99	1257	31	2012-08-02 05:46:51.430417
31632	3	1-8096	31630	99	90	332	15	2012-08-02 05:46:51.430417
31630	3	1-9549	31630	90	67	2319	82	2012-08-02 05:46:51.430417
31634	4	1-5519	31633	42	13	351	89	2012-08-02 05:46:51.430417
31635	4	1-7433	31633	13	74	186	83	2012-08-02 05:46:51.430417
31636	4	1-9223	31633	74	90	200	15	2012-08-02 05:46:51.430417
31633	4	1-9549	31633	90	42	1456	82	2012-08-02 05:46:51.430417
31638	4	1-9223	31637	74	18	3084	15	2012-08-02 05:46:53.032769
31639	4	1-9410	31637	18	91	1406	63	2012-08-02 05:46:53.032769
31640	4	1-8815	31637	91	20	855	51	2012-08-02 05:46:53.032769
31637	4	1-9553	31637	20	74	2779	83	2012-08-02 05:46:53.032769
31642	4	1-8807	31641	6	82	873	95	2012-08-02 05:46:53.032769
31643	4	1-9456	31641	82	9	878	71	2012-08-02 05:46:53.032769
31644	4	1-8172	31641	9	20	287	51	2012-08-02 05:46:53.032769
31641	4	1-9553	31641	20	6	1066	83	2012-08-02 05:46:53.032769
31646	4	1-9311	31645	13	82	503	59	2012-08-02 05:46:53.032769
31647	4	1-9314	31645	82	35	233	81	2012-08-02 05:46:53.032769
31648	4	1-5851	31645	35	20	158	51	2012-08-02 05:46:53.032769
31645	4	1-9553	31645	20	13	599	83	2012-08-02 05:46:53.032769
31650	4	1-8807	31649	6	88	1121	95	2012-08-02 05:46:53.032769
31651	4	1-7356	31649	88	91	404	63	2012-08-02 05:46:53.032769
31652	4	1-8815	31649	91	20	308	51	2012-08-02 05:46:53.032769
31649	4	1-9553	31649	20	6	1251	83	2012-08-02 05:46:53.032769
31654	4	1-9113	31653	69	63	2722	6	2012-08-02 05:46:53.032769
31655	4	1-6355	31653	63	80	680	55	2012-08-02 05:46:53.032769
31656	4	1-9157	31653	80	20	855	51	2012-08-02 05:46:53.032769
31653	4	1-9553	31653	20	69	3772	83	2012-08-02 05:46:53.032769
31658	3	1-8610	31657	83	83	3021	51	2012-08-02 05:46:53.452684
31659	3	1-9085	31657	83	5	2050	50	2012-08-02 05:46:53.452684
31657	3	1-9554	31657	5	83	9370	25	2012-08-02 05:46:53.452684
31661	3	1-8445	31660	6	31	113	70	2012-08-02 05:46:53.452684
31662	3	1-3661	31660	31	5	27	50	2012-08-02 05:46:53.452684
31660	3	1-9554	31660	5	6	173	25	2012-08-02 05:46:53.452684
31664	2	1-8742	31663	72	50	6960	70	2012-08-02 05:46:53.850429
31663	2	1-9555	31663	50	72	7849	52	2012-08-02 05:46:53.850429
31666	4	1-7941	31665	53	89	1398	10	2012-08-02 05:46:53.850429
31667	4	1-9486	31665	89	61	958	34	2012-08-02 05:46:53.850429
31668	4	1-8426	31665	61	50	1281	70	2012-08-02 05:46:53.850429
31665	4	1-9555	31665	50	53	1452	52	2012-08-02 05:46:53.850429
31670	4	1-9034	31669	21	98	3333	56	2012-08-02 05:46:54.248119
31671	4	1-7602	31669	98	83	1862	51	2012-08-02 05:46:54.248119
31672	4	1-9085	31669	83	95	1282	50	2012-08-02 05:46:54.248119
31669	4	1-9556	31669	95	21	5421	57	2012-08-02 05:46:54.248119
31674	3	1-9034	31673	21	9	41	56	2012-08-02 05:46:54.248119
31675	3	1-2583	31673	9	95	13	50	2012-08-02 05:46:54.248119
31673	3	1-9556	31673	95	21	60	57	2012-08-02 05:46:54.248119
31677	2	1-5808	31676	40	79	98	72	2012-08-02 05:46:56.270341
31676	2	1-9561	31676	79	40	1916	18	2012-08-02 05:46:56.270341
31679	3	1-9262	31678	41	80	298	75	2012-08-02 05:46:56.270341
31680	3	1-9146	31678	80	79	129	72	2012-08-02 05:46:56.270341
31678	3	1-9561	31678	79	41	3439	18	2012-08-02 05:46:56.270341
31682	4	1-8294	31681	63	89	634	77	2012-08-02 05:46:57.076851
31683	4	1-6860	31681	89	81	283	48	2012-08-02 05:46:57.076851
31684	4	1-9434	31681	81	99	294	20	2012-08-02 05:46:57.076851
31681	4	1-9563	31681	99	63	464	49	2012-08-02 05:46:57.076851
31686	4	1-5721	31685	90	19	737	95	2012-08-02 05:46:57.076851
31687	4	1-9465	31685	19	51	549	93	2012-08-02 05:46:57.076851
31688	4	1-8587	31685	51	99	583	20	2012-08-02 05:46:57.076851
31685	4	1-9563	31685	99	90	992	49	2012-08-02 05:46:57.076851
31690	4	1-6211	31689	75	75	779	73	2012-08-02 05:46:57.076851
31691	4	1-9349	31689	75	30	299	12	2012-08-02 05:46:57.076851
31692	4	1-9347	31689	30	99	774	20	2012-08-02 05:46:57.076851
31689	4	1-9563	31689	99	75	1426	49	2012-08-02 05:46:57.076851
31694	3	1-9486	31693	89	96	36	34	2012-08-02 05:46:59.087226
31695	3	1-9416	31693	96	88	44	42	2012-08-02 05:46:59.087226
31693	3	1-9568	31693	88	89	90	10	2012-08-02 05:46:59.087226
31697	4	1-8546	31696	20	5	146	78	2012-08-02 05:46:59.087226
31698	4	1-8994	31696	5	76	150	28	2012-08-02 05:46:59.087226
31699	4	1-8471	31696	76	88	131	42	2012-08-02 05:46:59.087226
31696	4	1-9568	31696	88	20	469	10	2012-08-02 05:46:59.087226
31701	4	1-5088	31700	4	56	1316	65	2012-08-02 05:47:00.314026
31702	4	1-9097	31700	56	21	3071	57	2012-08-02 05:47:00.314026
31703	4	1-9034	31700	21	8	2285	56	2012-08-02 05:47:00.314026
31700	4	1-9571	31700	8	4	1846	47	2012-08-02 05:47:00.314026
31705	4	1-8783	31704	83	65	4178	77	2012-08-02 05:47:00.314026
31706	4	1-9286	31704	65	7	4751	42	2012-08-02 05:47:00.314026
31707	4	1-9566	31704	7	8	4386	56	2012-08-02 05:47:00.314026
31704	4	1-9571	31704	8	83	4681	47	2012-08-02 05:47:00.314026
31709	4	1-9264	31708	62	86	3737	62	2012-08-02 05:47:01.131639
31710	4	1-9446	31708	86	42	4791	82	2012-08-02 05:47:01.131639
31711	4	1-5519	31708	42	62	1179	89	2012-08-02 05:47:01.131639
31708	4	1-9573	31708	62	62	5794	100	2012-08-02 05:47:01.131639
31713	4	1-9270	31712	92	44	1558	98	2012-08-02 05:47:01.926869
31714	4	1-5848	31712	44	14	2160	85	2012-08-02 05:47:01.926869
31715	4	1-8197	31712	14	74	908	88	2012-08-02 05:47:01.926869
31712	4	1-9575	31712	74	92	4399	5	2012-08-02 05:47:01.926869
31717	3	1-9048	31716	67	84	145	30	2012-08-02 05:47:01.926869
31718	3	1-6733	31716	84	74	158	88	2012-08-02 05:47:01.926869
31716	3	1-9575	31716	74	67	963	5	2012-08-02 05:47:01.926869
31720	4	1-5635	31719	44	90	378	92	2012-08-02 05:47:01.926869
31721	4	1-7252	31719	90	55	124	81	2012-08-02 05:47:01.926869
31722	4	1-7664	31719	55	74	137	88	2012-08-02 05:47:01.926869
31719	4	1-9575	31719	74	44	943	5	2012-08-02 05:47:01.926869
31724	3	1-4564	31723	27	53	385	59	2012-08-02 05:47:02.324687
31725	3	1-9105	31723	53	67	696	42	2012-08-02 05:47:02.324687
31723	3	1-9576	31723	67	27	442	25	2012-08-02 05:47:02.324687
31727	4	1-7934	31726	37	62	1890	100	2012-08-02 05:47:04.821364
31728	4	1-9264	31726	62	58	1180	62	2012-08-02 05:47:04.821364
31729	4	1-9557	31726	58	60	1446	32	2012-08-02 05:47:04.821364
31726	4	1-9582	31726	60	37	3473	20	2012-08-02 05:47:04.821364
31731	5	1-9563	31730	99	85	4053	49	2012-08-02 05:47:04.821364
31732	5	1-4080	31730	85	62	719	100	2012-08-02 05:47:04.821364
31733	5	1-9264	31730	62	58	538	62	2012-08-02 05:47:04.821364
31734	5	1-9557	31730	58	60	789	32	2012-08-02 05:47:04.821364
31730	5	1-9582	31730	60	99	2269	20	2012-08-02 05:47:04.821364
31736	4	1-7359	31735	36	27	1073	34	2012-08-02 05:47:04.821364
31737	4	1-8971	31735	27	1	1149	16	2012-08-02 05:47:04.821364
31738	4	1-7768	31735	1	60	574	32	2012-08-02 05:47:04.821364
31735	4	1-9582	31735	60	36	1721	20	2012-08-02 05:47:04.821364
31740	3	1-9583	31739	36	92	224	34	2012-08-02 05:47:05.652785
31741	3	1-9346	31739	92	79	549	92	2012-08-02 05:47:05.652785
31739	3	1-9584	31739	79	36	924	49	2012-08-02 05:47:05.652785
31743	2	1-7675	31742	5	79	2760	92	2012-08-02 05:47:05.652785
31742	2	1-9584	31742	79	5	5678	49	2012-08-02 05:47:05.652785
31745	4	1-9531	31744	5	60	4194	79	2012-08-02 05:47:06.846517
31746	4	1-7264	31744	60	78	2733	15	2012-08-02 05:47:06.846517
31747	4	1-9399	31744	78	91	3951	54	2012-08-02 05:47:06.846517
31744	4	1-9587	31744	91	5	3161	65	2012-08-02 05:47:06.846517
31749	4	1-5622	31748	99	74	574	83	2012-08-02 05:47:06.846517
31750	4	1-9223	31748	74	78	806	15	2012-08-02 05:47:06.846517
31751	4	1-9399	31748	78	91	1207	54	2012-08-02 05:47:06.846517
31748	4	1-9587	31748	91	99	1000	65	2012-08-02 05:47:06.846517
31753	5	1-8694	31752	22	5	513	76	2012-08-02 05:47:09.729894
31754	5	1-8983	31752	5	33	337	75	2012-08-02 05:47:09.729894
31755	5	1-7742	31752	33	51	216	46	2012-08-02 05:47:09.729894
31756	5	1-9421	31752	51	81	713	66	2012-08-02 05:47:09.729894
31752	5	1-9594	31752	81	22	400	24	2012-08-02 05:47:09.729894
31758	4	1-8026	31757	6	100	428	67	2012-08-02 05:47:09.729894
31759	4	1-8726	31757	100	29	871	53	2012-08-02 05:47:09.729894
31760	4	1-8974	31757	29	81	406	66	2012-08-02 05:47:09.729894
31757	4	1-9594	31757	81	6	281	24	2012-08-02 05:47:09.729894
31762	3	1-8983	31761	5	80	1755	75	2012-08-02 05:47:10.127949
31763	3	1-9146	31761	80	58	2155	72	2012-08-02 05:47:10.127949
31761	3	1-9595	31761	58	5	3069	76	2012-08-02 05:47:10.127949
31765	4	1-7540	31764	30	38	3987	92	2012-08-02 05:47:10.127949
31766	4	1-9457	31764	38	47	2337	84	2012-08-02 05:47:10.127949
31767	4	1-9245	31764	47	58	2080	72	2012-08-02 05:47:10.127949
31764	4	1-9595	31764	58	30	4140	76	2012-08-02 05:47:10.127949
31769	4	1-9445	31768	50	95	1178	28	2012-08-02 05:47:10.127949
31770	4	1-7223	31768	95	48	498	59	2012-08-02 05:47:10.127949
31771	4	1-9358	31768	48	58	544	72	2012-08-02 05:47:10.127949
31768	4	1-9595	31768	58	50	1176	76	2012-08-02 05:47:10.127949
31773	4	1-8081	31772	95	70	1573	88	2012-08-02 05:47:11.751938
31774	4	1-9413	31772	70	44	1147	4	2012-08-02 05:47:11.751938
31775	4	1-8977	31772	44	13	580	11	2012-08-02 05:47:11.751938
31772	4	1-9599	31772	13	95	1075	96	2012-08-02 05:47:11.751938
31777	4	1-7823	31776	2	26	981	87	2012-08-02 05:47:11.751938
31778	4	1-9459	31776	26	66	906	28	2012-08-02 05:47:11.751938
31779	4	1-9303	31776	66	13	481	11	2012-08-02 05:47:11.751938
31776	4	1-9599	31776	13	2	1150	96	2012-08-02 05:47:11.751938
31781	5	1-9551	31780	91	81	6	66	2012-08-02 05:47:12.171871
31782	5	1-9594	31780	81	50	4	24	2012-08-02 05:47:12.171871
31783	5	1-8574	31780	50	56	5	65	2012-08-02 05:47:12.171871
31784	5	1-9097	31780	56	84	14	57	2012-08-02 05:47:12.171871
31780	5	1-9600	31780	84	91	8	51	2012-08-02 05:47:12.171871
31786	4	1-9388	31785	48	13	140	75	2012-08-02 05:47:12.171871
31787	4	1-9304	31785	13	26	252	79	2012-08-02 05:47:12.171871
31788	4	1-9248	31785	26	84	304	57	2012-08-02 05:47:12.171871
31785	4	1-9600	31785	84	48	193	51	2012-08-02 05:47:12.171871
31790	4	1-1131	31789	78	72	2703	80	2012-08-02 05:47:12.569746
31791	4	1-9543	31789	72	21	2343	43	2012-08-02 05:47:12.569746
31792	4	1-9310	31789	21	84	3689	3	2012-08-02 05:47:12.569746
31789	4	1-9601	31789	84	78	4232	8	2012-08-02 05:47:12.569746
31794	4	1-9568	31793	88	42	4295	10	2012-08-02 05:47:13.818537
31795	4	1-9155	31793	42	35	3865	40	2012-08-02 05:47:13.818537
31796	4	1-9231	31793	35	89	378	64	2012-08-02 05:47:13.818537
31793	4	1-9604	31793	89	88	1354	42	2012-08-02 05:47:13.818537
31798	3	1-9461	31797	48	5	808	61	2012-08-02 05:47:13.818537
31799	3	1-8245	31797	5	89	297	64	2012-08-02 05:47:13.818537
31797	3	1-9604	31797	89	48	1129	42	2012-08-02 05:47:13.818537
31801	4	1-8596	31800	65	7	605	65	2012-08-02 05:47:13.818537
31802	4	1-9244	31800	7	35	3094	40	2012-08-02 05:47:13.818537
31803	4	1-9231	31800	35	89	331	64	2012-08-02 05:47:13.818537
31800	4	1-9604	31800	89	65	1294	42	2012-08-02 05:47:13.818537
31805	4	1-9566	31804	7	85	4250	56	2012-08-02 05:47:13.818537
31806	4	1-8441	31804	85	12	4872	16	2012-08-02 05:47:13.818537
31807	4	1-9487	31804	12	89	1150	64	2012-08-02 05:47:13.818537
31804	4	1-9604	31804	89	7	4827	42	2012-08-02 05:47:13.818537
31809	4	1-3762	31808	50	93	349	56	2012-08-02 05:47:14.238393
31810	4	1-9512	31808	93	62	565	100	2012-08-02 05:47:14.238393
31811	4	1-9264	31808	62	72	359	62	2012-08-02 05:47:14.238393
31808	4	1-9605	31808	72	50	1308	8	2012-08-02 05:47:14.238393
31813	4	1-6840	31812	78	17	1542	36	2012-08-02 05:47:14.238393
31814	4	1-9426	31812	17	54	1320	18	2012-08-02 05:47:14.238393
31815	4	1-8943	31812	54	72	574	62	2012-08-02 05:47:14.238393
31812	4	1-9605	31812	72	78	2139	8	2012-08-02 05:47:14.238393
31817	4	1-7075	31816	54	91	2215	54	2012-08-02 05:47:14.238393
31818	4	1-9587	31816	91	26	1879	65	2012-08-02 05:47:14.238393
31819	4	1-4530	31816	26	72	651	62	2012-08-02 05:47:14.238393
31816	4	1-9605	31816	72	54	2598	8	2012-08-02 05:47:14.238393
31821	4	1-6840	31820	78	17	76	36	2012-08-02 05:47:14.238393
31822	4	1-9426	31820	17	6	72	18	2012-08-02 05:47:14.238393
31823	4	1-9420	31820	6	72	23	62	2012-08-02 05:47:14.238393
31820	4	1-9605	31820	72	78	96	8	2012-08-02 05:47:14.238393
31825	5	1-9105	31824	53	65	4735	42	2012-08-02 05:47:16.315358
31826	5	1-8596	31824	65	26	1682	65	2012-08-02 05:47:16.315358
31827	5	1-4530	31824	26	72	439	62	2012-08-02 05:47:16.315358
31828	5	1-9605	31824	72	54	1324	8	2012-08-02 05:47:16.315358
31824	5	1-9610	31824	54	53	3878	59	2012-08-02 05:47:16.315358
31830	4	1-9314	31829	82	83	1776	81	2012-08-02 05:47:16.315358
31831	4	1-4310	31829	83	72	378	62	2012-08-02 05:47:16.315358
31832	4	1-9605	31829	72	54	1265	8	2012-08-02 05:47:16.315358
31829	4	1-9610	31829	54	82	4114	59	2012-08-02 05:47:16.315358
31834	4	1-9347	31833	30	56	2430	20	2012-08-02 05:47:17.53076
31835	4	1-7247	31833	56	76	2317	70	2012-08-02 05:47:17.53076
31836	4	1-7076	31833	76	87	1182	48	2012-08-02 05:47:17.53076
31833	4	1-9613	31833	87	30	975	12	2012-08-02 05:47:17.53076
31838	4	1-9529	31837	66	51	2995	46	2012-08-02 05:47:17.939705
31839	4	1-9137	31837	51	66	1407	43	2012-08-02 05:47:17.939705
31840	4	1-9127	31837	66	28	1080	6	2012-08-02 05:47:17.939705
31837	4	1-9614	31837	28	66	2308	74	2012-08-02 05:47:17.939705
31842	5	1-7486	31841	49	91	36	51	2012-08-02 05:47:17.939705
31843	5	1-9551	31841	91	81	23	66	2012-08-02 05:47:17.939705
31844	5	1-9594	31841	81	71	13	24	2012-08-02 05:47:17.939705
31845	5	1-8375	31841	71	28	16	6	2012-08-02 05:47:17.939705
31841	5	1-9614	31841	28	49	45	74	2012-08-02 05:47:17.939705
31847	4	1-8229	31846	69	13	1085	89	2012-08-02 05:47:17.939705
31848	4	1-7433	31846	13	69	735	83	2012-08-02 05:47:17.939705
31849	4	1-9113	31846	69	28	526	6	2012-08-02 05:47:17.939705
31846	4	1-9614	31846	28	69	1721	74	2012-08-02 05:47:17.939705
31851	3	1-6946	31850	91	71	449	69	2012-08-02 05:47:17.939705
31852	3	1-9299	31850	71	28	338	6	2012-08-02 05:47:17.939705
31850	3	1-9614	31850	28	91	1117	74	2012-08-02 05:47:17.939705
31854	3	1-7892	31853	47	53	541	49	2012-08-02 05:47:18.359673
31855	3	1-9033	31853	53	15	452	13	2012-08-02 05:47:18.359673
31853	3	1-9615	31853	15	47	832	33	2012-08-02 05:47:18.359673
31857	4	1-8496	31856	73	2	1933	68	2012-08-02 05:47:18.359673
31858	4	1-9329	31856	2	53	1167	49	2012-08-02 05:47:18.359673
31859	4	1-9033	31856	53	15	1105	13	2012-08-02 05:47:18.359673
31856	4	1-9615	31856	15	73	2303	33	2012-08-02 05:47:18.359673
31861	5	1-9068	31860	42	28	1975	85	2012-08-02 05:47:19.16651
31862	5	1-9433	31860	28	72	1817	80	2012-08-02 05:47:19.16651
31863	5	1-9543	31860	72	21	1652	43	2012-08-02 05:47:19.16651
31864	5	1-9310	31860	21	34	2727	3	2012-08-02 05:47:19.16651
31860	5	1-9617	31860	34	42	1494	59	2012-08-02 05:47:19.16651
31866	4	1-9085	31865	83	5	1788	50	2012-08-02 05:47:19.585978
31867	4	1-9510	31865	5	65	3118	21	2012-08-02 05:47:19.585978
31868	4	1-7494	31865	65	21	1910	29	2012-08-02 05:47:19.585978
31865	4	1-9618	31865	21	83	1819	51	2012-08-02 05:47:19.585978
31870	4	1-1317	31869	60	11	1738	43	2012-08-02 05:47:19.983907
31871	4	1-9354	31869	11	99	3569	82	2012-08-02 05:47:19.983907
31872	4	1-9392	31869	99	67	3182	58	2012-08-02 05:47:19.983907
31869	4	1-9619	31869	67	60	3603	35	2012-08-02 05:47:19.983907
31874	4	1-8644	31873	56	64	5470	80	2012-08-02 05:47:21.210083
31875	4	1-6771	31873	64	88	2155	50	2012-08-02 05:47:21.210083
31876	4	1-9497	31873	88	60	5233	17	2012-08-02 05:47:21.210083
31873	4	1-9622	31873	60	56	6681	12	2012-08-02 05:47:21.210083
31878	4	1-9347	31877	30	2	216	20	2012-08-02 05:47:21.210083
31879	4	1-1193	31877	2	88	29	50	2012-08-02 05:47:21.210083
31880	4	1-9497	31877	88	60	73	17	2012-08-02 05:47:21.210083
31877	4	1-9622	31877	60	30	97	12	2012-08-02 05:47:21.210083
31882	4	1-9413	31881	70	62	560	4	2012-08-02 05:47:22.005712
31883	4	1-8232	31881	62	60	208	32	2012-08-02 05:47:22.005712
31881	4	1-9624	31881	57	70	659	88	2012-08-02 05:47:22.005712
31884	4	1-9582	31881	60	57	549	20	2012-08-02 05:47:22.005712
31886	4	1-9552	31885	5	57	675	35	2012-08-02 05:47:22.005712
31887	4	1-1568	31885	57	60	271	32	2012-08-02 05:47:22.005712
31888	4	1-9582	31885	60	57	774	20	2012-08-02 05:47:22.005712
31885	4	1-9624	31885	57	5	1033	88	2012-08-02 05:47:22.005712
31890	2	1-9016	31889	1	99	1426	23	2012-08-02 05:47:22.414736
31889	2	1-9625	31889	99	1	5607	54	2012-08-02 05:47:22.414736
31892	4	1-9253	31891	41	52	2958	75	2012-08-02 05:47:22.834437
31893	4	1-777	31891	52	97	1954	59	2012-08-02 05:47:22.834437
31894	4	1-9462	31891	97	65	1015	90	2012-08-02 05:47:22.834437
31891	4	1-9626	31891	65	41	2353	55	2012-08-02 05:47:22.834437
31896	3	1-9472	31895	78	97	5222	59	2012-08-02 05:47:22.834437
31897	3	1-9462	31895	97	65	2734	90	2012-08-02 05:47:22.834437
31895	3	1-9626	31895	65	78	6392	55	2012-08-02 05:47:22.834437
31899	4	1-9605	31898	72	63	782	8	2012-08-02 05:47:24.071957
31900	4	1-6670	31898	63	83	748	51	2012-08-02 05:47:24.071957
31901	4	1-9085	31898	83	16	664	50	2012-08-02 05:47:24.071957
31898	4	1-9629	31898	16	72	218	62	2012-08-02 05:47:24.071957
31903	4	1-8807	31902	6	78	2404	95	2012-08-02 05:47:27.750849
31904	4	1-9630	31902	78	53	899	43	2012-08-02 05:47:27.750849
31905	4	1-5515	31902	53	19	144	62	2012-08-02 05:47:27.750849
31902	4	1-9638	31902	19	6	6653	83	2012-08-02 05:47:27.750849
31907	4	1-4022	31906	61	88	1464	50	2012-08-02 05:47:28.181735
31908	4	1-9497	31906	88	60	2429	17	2012-08-02 05:47:28.181735
31909	4	1-9622	31906	60	71	2119	12	2012-08-02 05:47:28.181735
31906	4	1-9639	31906	71	61	4111	30	2012-08-02 05:47:28.181735
31911	4	1-9137	31910	51	66	2689	43	2012-08-02 05:47:29.419164
31912	4	1-9127	31910	66	48	1816	6	2012-08-02 05:47:29.419164
31913	4	1-8140	31910	48	82	818	64	2012-08-02 05:47:29.419164
31910	4	1-9642	31910	82	51	4863	46	2012-08-02 05:47:29.419164
31915	5	1-9355	31914	87	48	295	86	2012-08-02 05:47:31.475678
31916	5	1-9496	31914	48	81	417	66	2012-08-02 05:47:31.475678
31917	5	1-9594	31914	81	84	306	24	2012-08-02 05:47:31.475678
31918	5	1-8113	31914	84	66	393	68	2012-08-02 05:47:31.475678
31914	5	1-9647	31914	66	87	327	30	2012-08-02 05:47:31.475678
31920	4	1-9633	31919	39	40	557	60	2012-08-02 05:47:32.724175
31921	4	1-7331	31919	40	78	301	15	2012-08-02 05:47:32.724175
31922	4	1-9399	31919	78	53	484	54	2012-08-02 05:47:32.724175
31919	4	1-9650	31919	53	39	542	10	2012-08-02 05:47:32.724175
31924	4	1-7742	31923	33	28	473	46	2012-08-02 05:47:33.110919
31925	4	1-8901	31923	28	67	620	58	2012-08-02 05:47:33.110919
31926	4	1-9619	31923	67	89	704	35	2012-08-02 05:47:33.110919
31923	4	1-9651	31923	89	33	576	75	2012-08-02 05:47:33.110919
31928	3	1-9289	31927	49	12	5916	80	2012-08-02 05:47:33.530537
31929	3	1-6035	31927	12	31	757	28	2012-08-02 05:47:33.530537
31927	3	1-9652	31927	31	49	5290	9	2012-08-02 05:47:33.530537
31931	3	1-7843	31930	12	92	1241	46	2012-08-02 05:47:33.530537
31932	3	1-9591	31930	92	31	307	28	2012-08-02 05:47:33.530537
31930	3	1-9652	31930	31	12	2760	9	2012-08-02 05:47:33.530537
31934	5	1-9242	31933	82	72	2037	98	2012-08-02 05:47:34.33729
31935	5	1-8352	31933	72	77	723	64	2012-08-02 05:47:34.33729
31936	5	1-8817	31933	77	98	1116	73	2012-08-02 05:47:34.33729
31937	5	1-9520	31933	98	37	1144	1	2012-08-02 05:47:34.33729
31933	5	1-9654	31933	37	82	967	24	2012-08-02 05:47:34.33729
31939	4	1-7456	31938	5	42	527	26	2012-08-02 05:47:34.33729
31940	4	1-9300	31938	42	96	1060	49	2012-08-02 05:47:34.33729
31941	4	1-6024	31938	96	37	629	1	2012-08-02 05:47:34.33729
31938	4	1-9654	31938	37	5	575	24	2012-08-02 05:47:34.33729
31943	4	1-9242	31942	82	66	2306	98	2012-08-02 05:47:34.33729
31944	4	1-8555	31942	66	10	1932	93	2012-08-02 05:47:34.33729
31945	4	1-9546	31942	10	37	1078	1	2012-08-02 05:47:34.33729
31942	4	1-9654	31942	37	82	999	24	2012-08-02 05:47:34.33729
31946	1	1-6024	\N	96	96	23	1	2012-08-02 05:47:34.33729
31948	4	1-9630	31947	78	53	516	43	2012-08-02 05:47:34.768383
31949	4	1-5515	31947	53	86	230	62	2012-08-02 05:47:34.768383
31950	4	1-9446	31947	86	89	417	82	2012-08-02 05:47:34.768383
31947	4	1-9655	31947	89	78	496	95	2012-08-02 05:47:34.768383
31952	2	1-9157	31951	80	50	877	51	2012-08-02 05:47:35.188098
31951	2	1-9656	31951	50	80	1430	55	2012-08-02 05:47:35.188098
31954	3	1-7626	31953	87	49	3962	74	2012-08-02 05:47:35.188098
31955	3	1-7486	31953	49	50	2852	51	2012-08-02 05:47:35.188098
31953	3	1-9656	31953	50	87	6665	55	2012-08-02 05:47:35.188098
31957	3	1-8430	31956	34	21	3686	78	2012-08-02 05:47:35.597525
31958	3	1-9148	31956	21	12	2085	67	2012-08-02 05:47:35.597525
31956	3	1-9657	31956	12	34	6000	75	2012-08-02 05:47:35.597525
31960	5	1-9146	31959	80	79	1298	72	2012-08-02 05:47:35.597525
31961	5	1-6838	31959	79	81	245	66	2012-08-02 05:47:35.597525
31962	5	1-9594	31959	81	6	165	24	2012-08-02 05:47:35.597525
31963	5	1-8026	31959	6	12	244	67	2012-08-02 05:47:35.597525
31959	5	1-9657	31959	12	80	766	75	2012-08-02 05:47:35.597525
31965	4	1-9497	31964	88	41	605	17	2012-08-02 05:47:37.232167
31966	4	1-7113	31964	41	97	985	2	2012-08-02 05:47:37.232167
31967	4	1-8296	31964	97	18	323	81	2012-08-02 05:47:37.232167
31964	4	1-9661	31964	18	88	260	50	2012-08-02 05:47:37.232167
31969	4	1-8836	31968	90	60	553	17	2012-08-02 05:47:37.662999
31970	4	1-9622	31968	60	88	488	12	2012-08-02 05:47:37.662999
31971	4	1-9602	31968	88	61	238	15	2012-08-02 05:47:37.662999
31968	4	1-9662	31968	61	90	1763	94	2012-08-02 05:47:37.662999
31973	4	1-9359	31972	85	15	753	97	2012-08-02 05:47:37.662999
31974	4	1-8290	31972	15	74	366	83	2012-08-02 05:47:37.662999
31975	4	1-9223	31972	74	61	376	15	2012-08-02 05:47:37.662999
31972	4	1-9662	31972	61	85	3149	94	2012-08-02 05:47:37.662999
31977	4	1-9661	31976	18	22	4151	50	2012-08-02 05:47:39.298403
31978	4	1-7623	31976	22	28	2872	85	2012-08-02 05:47:39.298403
31979	4	1-9433	31976	28	13	2168	80	2012-08-02 05:47:39.298403
31976	4	1-9666	31976	13	18	5140	81	2012-08-02 05:47:39.298403
31981	3	1-9444	31980	37	61	1420	69	2012-08-02 05:47:39.707015
31982	3	1-8838	31980	61	47	549	82	2012-08-02 05:47:39.707015
31980	3	1-9667	31980	47	37	2565	84	2012-08-02 05:47:39.707015
31984	5	1-8460	31983	5	5	2102	78	2012-08-02 05:47:39.707015
31985	5	1-8994	31983	5	66	2151	28	2012-08-02 05:47:39.707015
31986	5	1-9303	31983	66	53	983	11	2012-08-02 05:47:39.707015
31987	5	1-8412	31983	53	47	773	82	2012-08-02 05:47:39.707015
31983	5	1-9667	31983	47	5	4116	84	2012-08-02 05:47:39.707015
31989	4	1-8609	31988	67	44	525	98	2012-08-02 05:47:40.126866
31990	4	1-5848	31988	44	28	897	85	2012-08-02 05:47:40.126866
31991	4	1-9433	31988	28	13	712	80	2012-08-02 05:47:40.126866
31988	4	1-9668	31988	13	67	3090	2	2012-08-02 05:47:40.126866
31993	5	1-8922	31992	60	43	7702	2	2012-08-02 05:47:40.535865
31994	5	1-9086	31992	43	13	571	11	2012-08-02 05:47:40.535865
31995	5	1-9599	31992	13	23	848	96	2012-08-02 05:47:40.535865
31996	5	1-9269	31992	23	87	869	6	2012-08-02 05:47:40.535865
31992	5	1-9669	31992	87	60	2780	34	2012-08-02 05:47:40.535865
31998	5	1-8426	31997	61	38	4416	70	2012-08-02 05:47:40.535865
31999	5	1-9252	31997	38	13	923	11	2012-08-02 05:47:40.535865
32000	5	1-9599	31997	13	23	1399	96	2012-08-02 05:47:40.535865
32001	5	1-9269	31997	23	87	1462	6	2012-08-02 05:47:40.535865
31997	5	1-9669	31997	87	61	4776	34	2012-08-02 05:47:40.535865
32003	4	1-8555	32002	66	81	632	93	2012-08-02 05:47:40.955724
32004	4	1-8287	32002	81	74	275	83	2012-08-02 05:47:40.955724
32005	4	1-9223	32002	74	29	423	15	2012-08-02 05:47:40.955724
32002	4	1-9670	32002	29	66	769	98	2012-08-02 05:47:40.955724
32007	4	1-7797	32006	38	72	2763	14	2012-08-02 05:47:41.386597
32008	4	1-9586	32006	72	100	1014	42	2012-08-02 05:47:41.386597
32009	4	1-4678	32006	100	45	137	83	2012-08-02 05:47:41.386597
32006	4	1-9671	32006	45	38	9643	5	2012-08-02 05:47:41.386597
32011	4	1-9205	32010	15	46	593	87	2012-08-02 05:47:41.79532
32012	4	1-8545	32010	46	19	408	38	2012-08-02 05:47:41.79532
32013	4	1-8050	32010	19	87	432	48	2012-08-02 05:47:41.79532
32010	4	1-9672	32010	87	15	437	27	2012-08-02 05:47:41.79532
32015	4	1-8122	32014	14	58	1017	1	2012-08-02 05:47:41.79532
32016	4	1-7156	32014	58	19	1484	38	2012-08-02 05:47:41.79532
32017	4	1-8050	32014	19	87	1591	48	2012-08-02 05:47:41.79532
32014	4	1-9672	32014	87	14	1628	27	2012-08-02 05:47:41.79532
32019	3	1-9634	32018	51	11	3212	25	2012-08-02 05:47:42.612908
32020	3	1-9663	32018	11	61	1594	56	2012-08-02 05:47:42.612908
32018	3	1-9674	32018	61	51	3000	28	2012-08-02 05:47:42.612908
32022	4	1-8264	32021	79	98	1324	45	2012-08-02 05:47:44.281097
32023	4	1-8780	32021	98	13	1251	80	2012-08-02 05:47:44.281097
32024	4	1-9668	32021	13	3	3164	2	2012-08-02 05:47:44.281097
32021	4	1-9678	32021	3	79	4296	84	2012-08-02 05:47:44.281097
32026	4	1-9292	32025	95	19	2695	94	2012-08-02 05:47:45.087852
32027	4	1-6150	32025	19	53	843	59	2012-08-02 05:47:45.087852
32028	4	1-9105	32025	53	99	783	42	2012-08-02 05:47:45.087852
32025	4	1-9680	32025	99	95	2612	13	2012-08-02 05:47:45.087852
32030	4	1-9615	32029	15	77	1366	33	2012-08-02 05:47:45.087852
32031	4	1-6582	32029	77	72	433	58	2012-08-02 05:47:45.087852
32032	4	1-9511	32029	72	99	294	42	2012-08-02 05:47:45.087852
32029	4	1-9680	32029	99	15	1078	13	2012-08-02 05:47:45.087852
32034	4	1-8814	32033	51	37	515	99	2012-08-02 05:47:45.087852
32035	4	1-6803	32033	37	89	141	64	2012-08-02 05:47:45.087852
32036	4	1-9604	32033	89	99	385	42	2012-08-02 05:47:45.087852
32033	4	1-9680	32033	99	51	1547	13	2012-08-02 05:47:45.087852
32038	4	1-7937	32037	94	87	1022	63	2012-08-02 05:47:45.087852
32039	4	1-9189	32037	87	88	1176	73	2012-08-02 05:47:45.087852
32040	4	1-7461	32037	88	99	580	42	2012-08-02 05:47:45.087852
32037	4	1-9680	32037	99	94	2374	13	2012-08-02 05:47:45.087852
32042	4	1-8814	32041	51	39	243	99	2012-08-02 05:47:45.087852
32043	4	1-8731	32041	39	53	135	59	2012-08-02 05:47:45.087852
32044	4	1-9105	32041	53	99	161	42	2012-08-02 05:47:45.087852
32041	4	1-9680	32041	99	51	687	13	2012-08-02 05:47:45.087852
32046	4	1-9335	32045	45	1	186	43	2012-08-02 05:47:45.087852
32047	4	1-6318	32045	1	53	74	59	2012-08-02 05:47:45.087852
32048	4	1-9105	32045	53	99	99	42	2012-08-02 05:47:45.087852
32045	4	1-9680	32045	99	45	473	13	2012-08-02 05:47:45.087852
32050	4	1-9051	32049	20	99	329	82	2012-08-02 05:47:45.087852
32051	4	1-9392	32049	99	72	234	58	2012-08-02 05:47:45.087852
32052	4	1-9511	32049	72	99	158	42	2012-08-02 05:47:45.087852
32049	4	1-9680	32049	99	20	768	13	2012-08-02 05:47:45.087852
32053	1	1-2741	\N	71	71	1232	58	2012-08-02 05:47:45.087852
32055	3	1-8773	32054	98	34	3100	35	2012-08-02 05:47:46.745132
32056	3	1-7788	32054	34	41	1542	59	2012-08-02 05:47:46.745132
32054	3	1-9684	32054	41	98	3580	1	2012-08-02 05:47:46.745132
32058	4	1-7156	32057	58	35	564	38	2012-08-02 05:47:46.745132
32059	4	1-9455	32057	35	30	337	22	2012-08-02 05:47:46.745132
32060	4	1-8948	32057	30	41	225	59	2012-08-02 05:47:46.745132
32057	4	1-9684	32057	41	58	541	1	2012-08-02 05:47:46.745132
32062	4	1-7530	32061	83	99	1848	33	2012-08-02 05:47:46.745132
32063	4	1-8220	32061	99	30	1615	22	2012-08-02 05:47:46.745132
32064	4	1-8948	32061	30	41	1145	59	2012-08-02 05:47:46.745132
32061	4	1-9684	32061	41	83	2918	1	2012-08-02 05:47:46.745132
32066	4	1-7159	32065	75	48	467	37	2012-08-02 05:47:46.745132
32067	4	1-9611	32065	48	30	396	22	2012-08-02 05:47:46.745132
32068	4	1-8948	32065	30	41	286	59	2012-08-02 05:47:46.745132
32065	4	1-9684	32065	41	75	740	1	2012-08-02 05:47:46.745132
32070	2	1-9627	32069	25	59	5597	91	2012-08-02 05:47:47.574191
32069	2	1-9686	32069	59	25	8967	60	2012-08-02 05:47:47.574191
32072	3	1-8828	32071	96	72	4248	41	2012-08-02 05:47:47.982886
32073	3	1-9681	32071	72	39	5143	56	2012-08-02 05:47:47.982886
32071	3	1-9687	32071	39	96	4440	63	2012-08-02 05:47:47.982886
32075	5	1-8145	32074	1	60	1871	33	2012-08-02 05:47:49.639851
32076	5	1-8032	32074	60	6	621	24	2012-08-02 05:47:49.639851
32077	5	1-8026	32074	6	12	970	67	2012-08-02 05:47:49.639851
32078	5	1-9657	32074	12	41	3209	75	2012-08-02 05:47:49.639851
32074	5	1-9691	32074	41	1	3936	18	2012-08-02 05:47:49.639851
32080	3	1-8862	32079	92	5	16	23	2012-08-02 05:47:52.512375
32081	3	1-5904	32079	5	81	9	54	2012-08-02 05:47:52.512375
32079	3	1-9698	32079	81	92	23	95	2012-08-02 05:47:52.512375
32083	4	1-9630	32082	78	19	618	43	2012-08-02 05:47:52.512375
32084	4	1-8374	32082	19	50	386	28	2012-08-02 05:47:52.512375
32085	4	1-5160	32082	50	81	225	54	2012-08-02 05:47:52.512375
32082	4	1-9698	32082	81	78	679	95	2012-08-02 05:47:52.512375
32087	4	1-9465	32086	19	39	945	93	2012-08-02 05:47:52.512375
32088	4	1-8231	32086	39	97	493	97	2012-08-02 05:47:52.512375
32089	4	1-8146	32086	97	81	419	54	2012-08-02 05:47:52.512375
32086	4	1-9698	32086	81	19	1317	95	2012-08-02 05:47:52.512375
32090	1	1-5904	\N	5	5	1859	54	2012-08-02 05:47:52.512375
32092	5	1-8370	32091	32	91	78	51	2012-08-02 05:47:53.34105
32093	5	1-9551	32091	91	81	48	66	2012-08-02 05:47:53.34105
32094	5	1-9594	32091	81	6	26	24	2012-08-02 05:47:53.34105
32095	5	1-8026	32091	6	18	30	67	2012-08-02 05:47:53.34105
32091	5	1-9700	32091	18	32	155	82	2012-08-02 05:47:53.34105
32097	4	1-9392	32096	99	67	3896	58	2012-08-02 05:47:53.34105
32098	4	1-9619	32096	67	9	3341	35	2012-08-02 05:47:53.34105
32099	4	1-3800	32096	9	18	1095	67	2012-08-02 05:47:53.34105
32096	4	1-9700	32096	18	99	5769	82	2012-08-02 05:47:53.34105
32101	3	1-5519	32100	42	24	421	89	2012-08-02 05:47:53.34105
32102	3	1-5043	32100	24	18	175	67	2012-08-02 05:47:53.34105
32100	3	1-9700	32100	18	42	1231	82	2012-08-02 05:47:53.34105
32104	3	1-4827	32103	42	48	2599	37	2012-08-02 05:47:54.567492
32105	3	1-9611	32103	48	85	1492	22	2012-08-02 05:47:54.567492
32103	3	1-9703	32103	85	42	5688	68	2012-08-02 05:47:54.567492
32107	3	1-7908	32106	92	65	457	70	2012-08-02 05:47:54.567492
32108	3	1-9621	32106	65	85	204	22	2012-08-02 05:47:54.567492
32106	3	1-9703	32106	85	92	921	68	2012-08-02 05:47:54.567492
32110	5	1-9633	32109	39	83	3240	60	2012-08-02 05:47:54.976469
32111	5	1-5357	32109	83	27	540	9	2012-08-02 05:47:54.976469
32112	5	1-9315	32109	27	30	911	86	2012-08-02 05:47:54.976469
32113	5	1-9646	32109	30	62	1289	36	2012-08-02 05:47:54.976469
32109	5	1-9704	32109	62	39	3102	10	2012-08-02 05:47:54.976469
32115	4	1-9681	32114	72	61	187	56	2012-08-02 05:47:55.805281
32116	4	1-9674	32114	61	27	356	28	2012-08-02 05:47:55.805281
32117	4	1-9375	32114	27	88	989	2	2012-08-02 05:47:55.805281
32114	4	1-9706	32114	88	72	155	41	2012-08-02 05:47:55.805281
32119	4	1-4773	32118	92	98	1900	30	2012-08-02 05:47:56.600435
32120	4	1-9026	32118	98	78	1787	41	2012-08-02 05:47:56.600435
32121	4	1-8487	32118	78	31	489	67	2012-08-02 05:47:56.600435
32118	4	1-9708	32118	31	92	6278	56	2012-08-02 05:47:56.600435
32123	4	1-8973	32122	4	29	1042	12	2012-08-02 05:47:56.600435
32124	4	1-4959	32122	29	24	210	89	2012-08-02 05:47:56.600435
32125	4	1-5043	32122	24	31	73	67	2012-08-02 05:47:56.600435
32122	4	1-9708	32122	31	4	993	56	2012-08-02 05:47:56.600435
32127	5	1-2806	32126	96	21	39	49	2012-08-02 05:47:56.600435
32128	5	1-5862	32126	21	67	35	58	2012-08-02 05:47:56.600435
32129	5	1-9619	32126	67	9	39	35	2012-08-02 05:47:56.600435
32130	5	1-3800	32126	9	31	17	67	2012-08-02 05:47:56.600435
32126	5	1-9708	32126	31	96	264	56	2012-08-02 05:47:56.600435
32131	1	1-4959	\N	29	29	1722	89	2012-08-02 05:47:56.600435
32133	4	1-9041	32132	8	91	2429	53	2012-08-02 05:47:57.01149
32134	4	1-9339	32132	91	20	899	26	2012-08-02 05:47:57.01149
32135	4	1-8747	32132	20	85	1012	25	2012-08-02 05:47:57.01149
32132	4	1-9709	32132	85	8	897	55	2012-08-02 05:47:57.01149
32137	4	1-1568	32136	57	89	937	32	2012-08-02 05:47:57.420308
32138	4	1-8843	32136	89	66	736	28	2012-08-02 05:47:57.420308
32139	4	1-9303	32136	66	98	389	11	2012-08-02 05:47:57.420308
32136	4	1-9710	32136	98	57	2349	35	2012-08-02 05:47:57.420308
32141	4	1-1454	32140	71	41	73	17	2012-08-02 05:47:57.420308
32142	4	1-7113	32140	41	43	147	2	2012-08-02 05:47:57.420308
32143	4	1-9086	32140	43	98	17	11	2012-08-02 05:47:57.420308
32140	4	1-9710	32140	98	71	103	35	2012-08-02 05:47:57.420308
32145	5	1-6382	32144	73	38	406	21	2012-08-02 05:47:57.420308
32146	5	1-9494	32144	38	14	596	27	2012-08-02 05:47:57.420308
32147	5	1-8122	32144	14	20	367	1	2012-08-02 05:47:57.420308
32148	5	1-9523	32144	20	98	238	11	2012-08-02 05:47:57.420308
32144	5	1-9710	32144	98	73	1442	35	2012-08-02 05:47:57.420308
32150	5	1-5333	32149	47	7	1664	49	2012-08-02 05:47:57.420308
32151	5	1-4510	32149	7	98	750	73	2012-08-02 05:47:57.420308
32152	5	1-9520	32149	98	20	851	1	2012-08-02 05:47:57.420308
32153	5	1-9523	32149	20	98	547	11	2012-08-02 05:47:57.420308
32149	5	1-9710	32149	98	47	3342	35	2012-08-02 05:47:57.420308
32155	4	1-7704	32154	70	4	1667	57	2012-08-02 05:47:57.851532
32156	4	1-8881	32154	4	49	1510	12	2012-08-02 05:47:57.851532
32157	4	1-6302	32154	49	17	1261	92	2012-08-02 05:47:57.851532
32154	4	1-9711	32154	17	70	2657	61	2012-08-02 05:47:57.851532
32159	4	1-8816	32158	66	18	84	67	2012-08-02 05:47:57.851532
32160	4	1-9700	32158	18	66	488	82	2012-08-02 05:47:57.851532
32161	4	1-6976	32158	66	17	465	92	2012-08-02 05:47:57.851532
32158	4	1-9711	32158	17	66	1005	61	2012-08-02 05:47:57.851532
32163	3	1-9543	32162	72	1	1462	43	2012-08-02 05:47:58.282609
32164	3	1-6318	32162	1	68	522	59	2012-08-02 05:47:58.282609
32162	3	1-9712	32162	68	72	2319	80	2012-08-02 05:47:58.282609
32166	4	1-9668	32165	13	43	1391	2	2012-08-02 05:47:58.282609
32167	4	1-9086	32165	43	41	131	11	2012-08-02 05:47:58.282609
32168	4	1-8845	32165	41	68	67	59	2012-08-02 05:47:58.282609
32165	4	1-9712	32165	68	13	344	80	2012-08-02 05:47:58.282609
32170	4	1-8995	32169	73	50	2600	77	2012-08-02 05:47:58.282609
32171	4	1-7377	32169	50	30	1548	22	2012-08-02 05:47:58.282609
32172	4	1-8948	32169	30	68	1184	59	2012-08-02 05:47:58.282609
32169	4	1-9712	32169	68	73	6397	80	2012-08-02 05:47:58.282609
32174	5	1-8994	32173	5	66	246	28	2012-08-02 05:47:59.519802
32175	5	1-9303	32173	66	13	125	11	2012-08-02 05:47:59.519802
32176	5	1-9599	32173	13	23	287	96	2012-08-02 05:47:59.519802
32177	5	1-9269	32173	23	65	454	6	2012-08-02 05:47:59.519802
32173	5	1-9715	32173	65	5	236	78	2012-08-02 05:47:59.519802
32179	5	1-9406	32178	57	46	3190	52	2012-08-02 05:48:01.574606
32180	5	1-8025	32178	46	64	1438	45	2012-08-02 05:48:01.574606
32181	5	1-9324	32178	64	30	2271	86	2012-08-02 05:48:01.574606
32182	5	1-9646	32178	30	70	3069	36	2012-08-02 05:48:01.574606
32178	5	1-9720	32178	70	57	4918	44	2012-08-02 05:48:01.574606
32184	4	1-9085	32183	83	8	2454	50	2012-08-02 05:48:02.381372
32185	4	1-7200	32183	8	81	607	54	2012-08-02 05:48:02.381372
32186	4	1-9698	32183	81	29	969	95	2012-08-02 05:48:02.381372
32183	4	1-9722	32183	29	83	4958	51	2012-08-02 05:48:02.381372
32188	4	1-8648	32187	25	49	437	98	2012-08-02 05:48:02.381372
32189	4	1-3674	32187	49	81	138	54	2012-08-02 05:48:02.381372
32190	4	1-9698	32187	81	29	222	95	2012-08-02 05:48:02.381372
32187	4	1-9722	32187	29	25	1143	51	2012-08-02 05:48:02.381372
32192	4	1-9691	32191	41	34	1955	18	2012-08-02 05:48:03.651948
32193	4	1-7014	32191	34	74	1580	90	2012-08-02 05:48:03.651948
32194	4	1-9503	32191	74	92	1038	19	2012-08-02 05:48:03.651948
32191	4	1-9725	32191	92	41	1669	75	2012-08-02 05:48:03.651948
32195	4	1-9728	32195	51	64	3418	96	2012-08-02 05:48:04.900447
32196	4	1-8728	32195	64	60	2684	79	2012-08-02 05:48:04.900447
32197	4	1-7264	32195	60	29	1559	15	2012-08-02 05:48:04.900447
32198	4	1-9670	32195	29	51	2228	98	2012-08-02 05:48:04.900447
32200	4	1-9269	32199	23	5	2983	6	2012-08-02 05:48:04.900447
32201	4	1-1742	32199	5	29	891	15	2012-08-02 05:48:04.900447
32202	4	1-9670	32199	29	51	1354	98	2012-08-02 05:48:04.900447
32199	4	1-9728	32199	51	23	2209	96	2012-08-02 05:48:04.900447
32204	3	1-5095	32203	93	29	495	15	2012-08-02 05:48:04.900447
32205	3	1-9670	32203	29	51	787	98	2012-08-02 05:48:04.900447
32203	3	1-9728	32203	51	93	1344	96	2012-08-02 05:48:04.900447
32206	1	1-6365	\N	67	67	2230	6	2012-08-02 05:48:04.900447
32208	3	1-9627	32207	25	7	2746	91	2012-08-02 05:48:05.76293
32209	3	1-1301	32207	7	62	377	3	2012-08-02 05:48:05.76293
32207	3	1-9730	32207	62	25	4844	60	2012-08-02 05:48:05.76293
32211	4	1-9665	32210	74	63	11	100	2012-08-02 05:48:05.76293
32212	4	1-8606	32210	63	7	7	91	2012-08-02 05:48:05.76293
32213	4	1-1301	32210	7	62	1	3	2012-08-02 05:48:05.76293
32210	4	1-9730	32210	62	74	13	60	2012-08-02 05:48:05.76293
32215	5	1-9522	32214	4	51	134	44	2012-08-02 05:48:05.76293
32216	5	1-7672	32214	51	38	78	70	2012-08-02 05:48:05.76293
32217	5	1-9252	32214	38	85	20	11	2012-08-02 05:48:05.76293
32218	5	1-9332	32214	85	62	14	3	2012-08-02 05:48:05.76293
32214	5	1-9730	32214	62	4	207	60	2012-08-02 05:48:05.76293
32220	3	1-7672	32219	51	50	770	70	2012-08-02 05:48:06.956922
32221	3	1-9555	32219	50	23	542	52	2012-08-02 05:48:06.956922
32219	3	1-9733	32219	23	51	1817	44	2012-08-02 05:48:06.956922
32223	4	1-7672	32222	51	65	2122	70	2012-08-02 05:48:06.956922
32224	4	1-9621	32222	65	29	1268	22	2012-08-02 05:48:06.956922
32225	4	1-8361	32222	29	23	681	52	2012-08-02 05:48:06.956922
32222	4	1-9733	32222	23	51	3380	44	2012-08-02 05:48:06.956922
32227	4	1-8648	32226	25	66	48	98	2012-08-02 05:48:07.752523
32228	4	1-8555	32226	66	51	27	93	2012-08-02 05:48:07.752523
32229	4	1-8587	32226	51	43	20	20	2012-08-02 05:48:07.752523
32226	4	1-9735	32226	43	25	87	51	2012-08-02 05:48:07.752523
32231	4	1-8687	32230	86	5	1632	84	2012-08-02 05:48:07.752523
32232	4	1-2625	32230	5	100	250	11	2012-08-02 05:48:07.752523
32233	4	1-9632	32230	100	43	442	20	2012-08-02 05:48:07.752523
32230	4	1-9735	32230	43	86	2508	51	2012-08-02 05:48:07.752523
32235	4	1-9551	32234	91	13	1832	66	2012-08-02 05:48:07.752523
32236	4	1-6673	32234	13	10	667	47	2012-08-02 05:48:07.752523
32237	4	1-7430	32234	10	43	422	20	2012-08-02 05:48:07.752523
32234	4	1-9735	32234	43	91	2407	51	2012-08-02 05:48:07.752523
32239	3	1-9159	32238	3	1	177	56	2012-08-02 05:48:07.752523
32240	3	1-9718	32238	1	43	117	20	2012-08-02 05:48:07.752523
32238	3	1-9735	32238	43	3	739	51	2012-08-02 05:48:07.752523
32241	1	1-7836	\N	73	73	1950	20	2012-08-02 05:48:07.752523
32242	1	1-5950	\N	42	42	2601	93	2012-08-02 05:48:07.752523
32244	3	1-9621	32243	65	26	1122	22	2012-08-02 05:48:08.194292
32245	3	1-7210	32243	26	87	790	84	2012-08-02 05:48:08.194292
32243	3	1-9736	32243	87	65	2648	70	2012-08-02 05:48:08.194292
32247	3	1-9621	32246	65	9	3064	22	2012-08-02 05:48:08.194292
32248	3	1-7944	32246	9	87	2084	84	2012-08-02 05:48:08.194292
32246	3	1-9736	32246	87	65	7108	70	2012-08-02 05:48:08.194292
32249	1	1-7944	\N	9	9	2922	84	2012-08-02 05:48:08.194292
32251	4	1-9594	32250	81	17	1343	24	2012-08-02 05:48:08.603814
32252	4	1-8655	32250	17	15	1739	27	2012-08-02 05:48:08.603814
32253	4	1-9205	32250	15	36	2388	87	2012-08-02 05:48:08.603814
32250	4	1-9737	32250	36	81	1779	66	2012-08-02 05:48:08.603814
32255	4	1-9635	32254	9	45	8556	94	2012-08-02 05:48:10.569865
32256	4	1-3451	32254	45	72	2759	41	2012-08-02 05:48:10.569865
32257	4	1-9681	32254	72	44	2233	56	2012-08-02 05:48:10.569865
32254	4	1-9742	32254	44	9	9551	69	2012-08-02 05:48:10.569865
32259	4	1-8843	32258	89	50	1809	28	2012-08-02 05:48:11.365608
32260	4	1-5160	32258	50	81	658	54	2012-08-02 05:48:11.365608
32261	4	1-9698	32258	81	40	1238	95	2012-08-02 05:48:11.365608
32258	4	1-9744	32258	40	89	4247	32	2012-08-02 05:48:11.365608
32263	4	1-9209	32262	78	49	1852	46	2012-08-02 05:48:11.365608
32264	4	1-9644	32262	49	6	1324	83	2012-08-02 05:48:11.365608
32265	4	1-8807	32262	6	40	1039	95	2012-08-02 05:48:11.365608
32262	4	1-9744	32262	40	78	5108	32	2012-08-02 05:48:11.365608
32267	3	1-7869	32266	14	52	43	47	2012-08-02 05:48:11.774426
32268	3	1-9533	32266	52	10	47	41	2012-08-02 05:48:11.774426
32266	3	1-9745	32266	10	14	191	86	2012-08-02 05:48:11.774426
32270	4	1-9496	32269	48	13	5615	66	2012-08-02 05:48:11.774426
32271	4	1-6673	32269	13	52	1405	47	2012-08-02 05:48:11.774426
32272	4	1-9533	32269	52	10	1560	41	2012-08-02 05:48:11.774426
32269	4	1-9745	32269	10	48	6460	86	2012-08-02 05:48:11.774426
32274	4	1-8646	32273	15	49	213	33	2012-08-02 05:48:11.774426
32275	4	1-8863	32273	49	52	127	47	2012-08-02 05:48:11.774426
32276	4	1-9533	32273	52	10	159	41	2012-08-02 05:48:11.774426
32273	4	1-9745	32273	10	15	737	86	2012-08-02 05:48:11.774426
32278	4	1-8863	32277	49	52	1494	47	2012-08-02 05:48:12.978594
32279	4	1-9533	32277	52	14	2574	41	2012-08-02 05:48:12.978594
32280	4	1-4513	32277	14	70	1060	43	2012-08-02 05:48:12.978594
32277	4	1-9748	32277	70	49	1810	33	2012-08-02 05:48:12.978594
32282	5	1-9391	32281	86	91	2120	4	2012-08-02 05:48:12.978594
32283	5	1-9606	32281	91	89	1258	82	2012-08-02 05:48:12.978594
32284	5	1-9655	32281	89	78	1488	95	2012-08-02 05:48:12.978594
32285	5	1-9630	32281	78	70	1540	43	2012-08-02 05:48:12.978594
32281	5	1-9748	32281	70	86	2757	33	2012-08-02 05:48:12.978594
32287	5	1-9731	32286	9	74	792	90	2012-08-02 05:48:13.398628
32288	5	1-9503	32286	74	3	467	19	2012-08-02 05:48:13.398628
32289	5	1-8052	32286	3	91	509	53	2012-08-02 05:48:13.398628
32290	5	1-9339	32286	91	96	166	26	2012-08-02 05:48:13.398628
32286	5	1-9749	32286	96	9	389	74	2012-08-02 05:48:13.398628
32292	4	1-7534	32291	50	71	2041	34	2012-08-02 05:48:13.796391
32293	4	1-8514	32291	71	63	3662	37	2012-08-02 05:48:13.796391
32294	4	1-9548	32291	63	95	2386	50	2012-08-02 05:48:13.796391
32291	4	1-9750	32291	95	50	3320	9	2012-08-02 05:48:13.796391
32296	3	1-9609	32295	33	20	7057	22	2012-08-02 05:48:13.796391
32297	3	1-5200	32295	20	95	2544	50	2012-08-02 05:48:13.796391
32295	3	1-9750	32295	95	33	3892	9	2012-08-02 05:48:13.796391
32299	4	1-7534	32298	50	7	857	34	2012-08-02 05:48:13.796391
32300	4	1-9631	32298	7	91	1847	66	2012-08-02 05:48:13.796391
32301	4	1-9098	32298	91	95	773	50	2012-08-02 05:48:13.796391
32298	4	1-9750	32298	95	50	1224	9	2012-08-02 05:48:13.796391
32303	4	1-9269	32302	23	29	2176	6	2012-08-02 05:48:16.205043
32304	4	1-3023	32302	29	70	520	43	2012-08-02 05:48:16.205043
32305	4	1-9748	32302	70	39	734	33	2012-08-02 05:48:16.205043
32302	4	1-9756	32302	39	23	1694	96	2012-08-02 05:48:16.205043
32307	5	1-8081	32306	95	5	445	88	2012-08-02 05:48:16.205043
32308	5	1-9552	32306	5	60	233	35	2012-08-02 05:48:16.205043
32309	5	1-1317	32306	60	70	91	43	2012-08-02 05:48:16.205043
32310	5	1-9748	32306	70	39	130	33	2012-08-02 05:48:16.205043
32306	5	1-9756	32306	39	95	301	96	2012-08-02 05:48:16.205043
32312	5	1-7563	32311	60	37	1489	68	2012-08-02 05:48:16.205043
32313	5	1-9394	32311	37	76	989	78	2012-08-02 05:48:16.205043
32314	5	1-9495	32311	76	70	563	43	2012-08-02 05:48:16.205043
32315	5	1-9748	32311	70	39	846	33	2012-08-02 05:48:16.205043
32311	5	1-9756	32311	39	60	2076	96	2012-08-02 05:48:16.205043
32317	4	1-6518	32316	19	64	303	45	2012-08-02 05:48:16.205043
32318	4	1-9324	32316	64	15	463	86	2012-08-02 05:48:16.205043
32319	4	1-8646	32316	15	39	177	33	2012-08-02 05:48:16.205043
32316	4	1-9756	32316	39	19	474	96	2012-08-02 05:48:16.205043
32321	4	1-5095	32320	93	29	692	15	2012-08-02 05:48:16.205043
32322	4	1-9670	32320	29	46	1246	98	2012-08-02 05:48:16.205043
32323	4	1-6970	32320	46	39	577	33	2012-08-02 05:48:16.205043
32320	4	1-9756	32320	39	93	1659	96	2012-08-02 05:48:16.205043
32325	5	1-9631	32324	7	81	2083	66	2012-08-02 05:48:16.625104
32326	5	1-9594	32324	81	85	1366	24	2012-08-02 05:48:16.625104
32327	5	1-5429	32324	85	35	573	64	2012-08-02 05:48:16.625104
32328	5	1-9362	32324	35	66	801	21	2012-08-02 05:48:16.625104
32324	5	1-9757	32324	66	7	1075	34	2012-08-02 05:48:16.625104
32330	2	1-8522	32329	82	66	2395	21	2012-08-02 05:48:16.625104
32329	2	1-9757	32329	66	82	3704	34	2012-08-02 05:48:16.625104
32332	4	1-9371	32331	20	22	2834	97	2012-08-02 05:48:17.454
32333	4	1-9302	32331	22	15	3485	86	2012-08-02 05:48:17.454
32334	4	1-8646	32331	15	67	1374	33	2012-08-02 05:48:17.454
32331	4	1-9759	32331	67	20	4269	65	2012-08-02 05:48:17.454
32336	4	1-9759	32335	67	26	4250	65	2012-08-02 05:48:18.271508
32337	4	1-4530	32335	26	74	810	62	2012-08-02 05:48:18.271508
32338	4	1-7973	32335	74	87	513	44	2012-08-02 05:48:18.271508
32335	4	1-9761	32335	87	67	2517	33	2012-08-02 05:48:18.271508
32340	4	1-9756	32339	39	74	2769	96	2012-08-02 05:48:18.271508
32341	4	1-8225	32339	74	8	689	19	2012-08-02 05:48:18.271508
32342	4	1-6991	32339	8	87	321	44	2012-08-02 05:48:18.271508
32339	4	1-9761	32339	87	39	1709	33	2012-08-02 05:48:18.271508
32344	4	1-8174	32343	5	3	1367	74	2012-08-02 05:48:18.271508
32345	4	1-9754	32343	3	91	726	27	2012-08-02 05:48:18.271508
32346	4	1-8332	32343	91	87	524	44	2012-08-02 05:48:18.271508
32343	4	1-9761	32343	87	5	3053	33	2012-08-02 05:48:18.271508
32348	3	1-9224	32347	95	18	4536	94	2012-08-02 05:48:20.261708
32349	3	1-9179	32347	18	30	4579	38	2012-08-02 05:48:20.261708
32347	3	1-9766	32347	30	95	6818	62	2012-08-02 05:48:20.261708
32351	4	1-9616	32350	94	18	665	77	2012-08-02 05:48:20.261708
32352	4	1-7033	32350	18	63	328	55	2012-08-02 05:48:20.261708
32353	4	1-8343	32350	63	30	338	38	2012-08-02 05:48:20.261708
32350	4	1-9766	32350	30	94	569	62	2012-08-02 05:48:20.261708
32355	4	1-8686	32354	89	65	1767	64	2012-08-02 05:48:22.239629
32356	4	1-9009	32354	65	68	457	41	2012-08-02 05:48:22.239629
32357	4	1-9261	32354	68	52	246	39	2012-08-02 05:48:22.239629
32354	4	1-9771	32354	52	89	5809	45	2012-08-02 05:48:22.239629
32359	2	1-6771	32358	64	98	432	50	2012-08-02 05:48:23.852354
32358	2	1-9775	32358	98	64	926	80	2012-08-02 05:48:23.852354
32361	3	1-9368	32360	7	75	776	61	2012-08-02 05:48:23.852354
32362	3	1-9351	32360	75	98	1102	50	2012-08-02 05:48:23.852354
32360	3	1-9775	32360	98	7	2409	80	2012-08-02 05:48:23.852354
32364	3	1-5471	32363	39	28	2337	85	2012-08-02 05:48:24.250183
32365	3	1-9433	32363	28	29	1830	80	2012-08-02 05:48:24.250183
32363	3	1-9776	32363	29	39	2490	77	2012-08-02 05:48:24.250183
32367	4	1-3587	32366	21	63	2098	37	2012-08-02 05:48:24.250183
32368	4	1-9548	32366	63	98	1492	50	2012-08-02 05:48:24.250183
32369	4	1-9775	32366	98	29	3030	80	2012-08-02 05:48:24.250183
32366	4	1-9776	32366	29	21	4624	77	2012-08-02 05:48:24.250183
32371	4	1-8717	32370	5	29	84	15	2012-08-02 05:48:24.250183
32372	4	1-9670	32370	29	57	158	98	2012-08-02 05:48:24.250183
32373	4	1-6911	32370	57	29	152	80	2012-08-02 05:48:24.250183
32370	4	1-9776	32370	29	5	252	77	2012-08-02 05:48:24.250183
32375	3	1-7251	32374	59	53	206	49	2012-08-02 05:48:24.670206
32376	3	1-9033	32374	53	30	173	13	2012-08-02 05:48:24.670206
32374	3	1-9777	32374	30	59	362	30	2012-08-02 05:48:24.670206
32378	4	1-9026	32377	98	51	163	41	2012-08-02 05:48:24.670206
32379	4	1-5302	32377	51	88	129	58	2012-08-02 05:48:24.670206
32380	4	1-8116	32377	88	30	70	13	2012-08-02 05:48:24.670206
32377	4	1-9777	32377	30	98	152	30	2012-08-02 05:48:24.670206
32382	4	1-9355	32381	87	68	4608	86	2012-08-02 05:48:24.670206
32383	4	1-9398	32381	68	88	3798	58	2012-08-02 05:48:24.670206
32384	4	1-8116	32381	88	30	2219	13	2012-08-02 05:48:24.670206
32381	4	1-9777	32381	30	87	5251	30	2012-08-02 05:48:24.670206
32386	4	1-666	32385	13	28	498	46	2012-08-02 05:48:24.670206
32387	4	1-8901	32385	28	88	662	58	2012-08-02 05:48:24.670206
32388	4	1-8116	32385	88	30	396	13	2012-08-02 05:48:24.670206
32385	4	1-9777	32385	30	13	959	30	2012-08-02 05:48:24.670206
32390	3	1-9166	32389	96	23	2029	4	2012-08-02 05:48:25.079237
32391	3	1-9536	32389	23	81	1544	94	2012-08-02 05:48:25.079237
32389	3	1-9778	32389	81	96	3128	30	2012-08-02 05:48:25.079237
32393	5	1-9631	32392	7	81	120	66	2012-08-02 05:48:25.89676
32394	5	1-9594	32392	81	80	44	24	2012-08-02 05:48:25.89676
32395	5	1-9363	32392	80	9	12	74	2012-08-02 05:48:25.89676
32396	5	1-9731	32392	9	37	14	90	2012-08-02 05:48:25.89676
32392	5	1-9780	32392	37	7	111	34	2012-08-02 05:48:25.89676
32398	4	1-9239	32397	79	80	1570	71	2012-08-02 05:48:25.89676
32399	4	1-6392	32397	80	42	540	38	2012-08-02 05:48:25.89676
32400	4	1-4896	32397	42	37	156	90	2012-08-02 05:48:25.89676
32397	4	1-9780	32397	37	79	1441	34	2012-08-02 05:48:25.89676
32402	4	1-8426	32401	61	36	64	70	2012-08-02 05:48:25.89676
32403	4	1-4711	32401	36	9	5	74	2012-08-02 05:48:25.89676
32404	4	1-9731	32401	9	37	7	90	2012-08-02 05:48:25.89676
32401	4	1-9780	32401	37	61	71	34	2012-08-02 05:48:25.89676
32406	4	1-9416	32405	96	67	9391	42	2012-08-02 05:48:25.89676
32407	4	1-9576	32405	67	2	4221	25	2012-08-02 05:48:25.89676
32408	4	1-5494	32405	2	37	499	90	2012-08-02 05:48:25.89676
32405	4	1-9780	32405	37	96	5699	34	2012-08-02 05:48:25.89676
32410	5	1-6827	32409	57	85	176	25	2012-08-02 05:48:25.89676
32411	5	1-9709	32409	85	87	130	55	2012-08-02 05:48:25.89676
32412	5	1-7626	32409	87	9	89	74	2012-08-02 05:48:25.89676
32413	5	1-9731	32409	9	37	171	90	2012-08-02 05:48:25.89676
32409	5	1-9780	32409	37	57	2081	34	2012-08-02 05:48:25.89676
32414	1	1-9594	\N	81	81	325	24	2012-08-02 05:48:25.89676
32416	4	1-9618	32415	21	48	3799	51	2012-08-02 05:48:27.145569
32417	4	1-9388	32415	48	70	2586	75	2012-08-02 05:48:27.145569
32418	4	1-8226	32415	70	40	807	36	2012-08-02 05:48:27.145569
32415	4	1-9783	32415	40	21	4018	29	2012-08-02 05:48:27.145569
32420	4	1-3386	32419	53	22	455	97	2012-08-02 05:48:27.543163
32421	4	1-9302	32419	22	30	539	86	2012-08-02 05:48:27.543163
32422	4	1-9646	32419	30	73	699	36	2012-08-02 05:48:27.543163
32419	4	1-9784	32419	73	53	1352	25	2012-08-02 05:48:27.543163
32424	4	1-9709	32423	85	41	365	55	2012-08-02 05:48:27.543163
32425	4	1-9253	32423	41	70	510	75	2012-08-02 05:48:27.543163
32426	4	1-8226	32423	70	73	176	36	2012-08-02 05:48:27.543163
32423	4	1-9784	32423	73	85	382	25	2012-08-02 05:48:27.543163
32428	4	1-9311	32427	13	15	382	59	2012-08-02 05:48:27.952212
32429	4	1-8661	32427	15	9	296	69	2012-08-02 05:48:27.952212
32430	4	1-9635	32427	9	57	303	94	2012-08-02 05:48:27.952212
32427	4	1-9785	32427	57	13	479	83	2012-08-02 05:48:27.952212
32432	4	1-9562	32431	47	39	3656	56	2012-08-02 05:48:27.952212
32433	4	1-8518	32431	39	23	2177	4	2012-08-02 05:48:27.952212
32434	4	1-9536	32431	23	57	3269	94	2012-08-02 05:48:27.952212
32431	4	1-9785	32431	57	47	6503	83	2012-08-02 05:48:27.952212
32436	4	1-8740	32435	48	67	600	58	2012-08-02 05:48:28.371925
32437	4	1-9619	32435	67	71	660	35	2012-08-02 05:48:28.371925
32438	4	1-1454	32435	71	93	447	17	2012-08-02 05:48:28.371925
32435	4	1-9786	32435	93	48	1076	29	2012-08-02 05:48:28.371925
32440	5	1-9487	32439	12	52	665	64	2012-08-02 05:48:28.758694
32441	5	1-8507	32439	52	73	75	36	2012-08-02 05:48:28.758694
32442	5	1-9784	32439	73	85	116	25	2012-08-02 05:48:28.758694
32443	5	1-9709	32439	85	28	79	55	2012-08-02 05:48:28.758694
32439	5	1-9787	32439	28	12	3855	16	2012-08-02 05:48:28.758694
32445	5	1-2129	32444	53	70	71	75	2012-08-02 05:48:28.758694
32446	5	1-8226	32444	70	73	20	36	2012-08-02 05:48:28.758694
32447	5	1-9784	32444	73	85	37	25	2012-08-02 05:48:28.758694
32448	5	1-9709	32444	85	28	30	55	2012-08-02 05:48:28.758694
32444	5	1-9787	32444	28	53	1715	16	2012-08-02 05:48:28.758694
32450	3	1-7963	32449	30	4	1291	58	2012-08-02 05:48:29.189664
32451	3	1-7911	32449	4	11	878	53	2012-08-02 05:48:29.189664
32449	3	1-9788	32449	11	30	2919	66	2012-08-02 05:48:29.189664
32453	2	1-7860	32452	32	18	3364	16	2012-08-02 05:48:29.576444
32452	2	1-9789	32452	18	32	3489	90	2012-08-02 05:48:29.576444
32455	4	1-9503	32454	74	40	386	19	2012-08-02 05:48:29.576444
32456	4	1-5008	32454	40	85	386	56	2012-08-02 05:48:29.576444
32457	4	1-8441	32454	85	18	428	16	2012-08-02 05:48:29.576444
32454	4	1-9789	32454	18	74	568	90	2012-08-02 05:48:29.576444
32459	3	1-8793	32458	55	36	1611	39	2012-08-02 05:48:29.576444
32460	3	1-8937	32458	36	18	3193	16	2012-08-02 05:48:29.576444
32458	3	1-9789	32458	18	55	4421	90	2012-08-02 05:48:29.576444
32462	4	1-9503	32461	74	97	456	19	2012-08-02 05:48:29.576444
32463	4	1-5513	32461	97	85	356	56	2012-08-02 05:48:29.576444
32464	4	1-8441	32461	85	18	429	16	2012-08-02 05:48:29.576444
32461	4	1-9789	32461	18	74	618	90	2012-08-02 05:48:29.576444
32465	1	1-5513	\N	97	97	5406	56	2012-08-02 05:48:29.576444
32467	5	1-9616	32466	94	18	2037	77	2012-08-02 05:48:31.244712
32468	5	1-7033	32466	18	87	940	55	2012-08-02 05:48:31.244712
32469	5	1-7626	32466	87	9	793	74	2012-08-02 05:48:31.244712
32470	5	1-9731	32466	9	4	1880	90	2012-08-02 05:48:31.244712
32466	5	1-9793	32466	4	94	1863	62	2012-08-02 05:48:31.244712
32472	2	1-8134	32471	59	38	1434	48	2012-08-02 05:48:32.448977
32471	2	1-9796	32471	38	59	9884	76	2012-08-02 05:48:32.448977
32474	5	1-7799	32473	36	18	224	16	2012-08-02 05:48:32.835784
32475	5	1-9789	32473	18	4	314	90	2012-08-02 05:48:32.835784
32476	5	1-9793	32473	4	58	326	62	2012-08-02 05:48:32.835784
32477	5	1-9557	32473	58	21	484	32	2012-08-02 05:48:32.835784
32473	5	1-9797	32473	21	36	296	6	2012-08-02 05:48:32.835784
32479	5	1-5777	32478	23	18	293	16	2012-08-02 05:48:32.835784
32480	5	1-9789	32478	18	4	411	90	2012-08-02 05:48:32.835784
32481	5	1-9793	32478	4	58	427	62	2012-08-02 05:48:32.835784
32482	5	1-9557	32478	58	21	634	32	2012-08-02 05:48:32.835784
32478	5	1-9797	32478	21	23	388	6	2012-08-02 05:48:32.835784
32483	1	1-5777	\N	23	23	2163	16	2012-08-02 05:48:32.835784
32485	4	1-9632	32484	100	77	5535	20	2012-08-02 05:48:34.481883
32486	4	1-6419	32484	77	51	2494	46	2012-08-02 05:48:34.481883
32487	4	1-9137	32484	51	37	1629	43	2012-08-02 05:48:34.481883
32484	4	1-9801	32484	37	100	3443	11	2012-08-02 05:48:34.481883
32489	4	1-9528	32488	38	12	225	9	2012-08-02 05:48:34.481883
32490	4	1-7843	32488	12	51	178	46	2012-08-02 05:48:34.481883
32491	4	1-9137	32488	51	37	123	43	2012-08-02 05:48:34.481883
32488	4	1-9801	32488	37	38	275	11	2012-08-02 05:48:34.481883
32493	4	1-8131	32492	26	18	841	36	2012-08-02 05:48:34.481883
32494	4	1-9525	32492	18	76	579	78	2012-08-02 05:48:34.481883
32495	4	1-9495	32492	76	37	376	43	2012-08-02 05:48:34.481883
32492	4	1-9801	32492	37	26	941	11	2012-08-02 05:48:34.481883
32497	4	1-7907	32496	78	34	769	75	2012-08-02 05:48:34.481883
32498	4	1-8430	32496	34	76	560	78	2012-08-02 05:48:34.481883
32499	4	1-9495	32496	76	37	379	43	2012-08-02 05:48:34.481883
32496	4	1-9801	32496	37	78	989	11	2012-08-02 05:48:34.481883
32501	3	1-7741	32500	69	32	1391	70	2012-08-02 05:48:34.890991
32502	3	1-9741	32500	32	38	2605	40	2012-08-02 05:48:34.890991
32500	3	1-9802	32500	38	69	3637	92	2012-08-02 05:48:34.890991
32504	4	1-7252	32503	90	55	663	81	2012-08-02 05:48:34.890991
32505	4	1-7664	32503	55	17	695	88	2012-08-02 05:48:34.890991
32506	4	1-9385	32503	17	38	1406	40	2012-08-02 05:48:34.890991
32503	4	1-9802	32503	38	90	2126	92	2012-08-02 05:48:34.890991
32508	4	1-7264	32507	60	29	323	15	2012-08-02 05:48:35.697427
32509	4	1-9670	32507	29	99	545	98	2012-08-02 05:48:35.697427
32510	4	1-7989	32507	99	16	463	37	2012-08-02 05:48:35.697427
32507	4	1-9804	32507	16	60	471	79	2012-08-02 05:48:35.697427
32512	3	1-6139	32511	93	93	1847	17	2012-08-02 05:48:36.095293
32513	3	1-9786	32511	93	24	4126	29	2012-08-02 05:48:36.095293
32511	3	1-9805	32511	24	93	3966	19	2012-08-02 05:48:36.095293
32515	2	1-6115	32514	45	24	499	29	2012-08-02 05:48:36.095293
32514	2	1-9805	32514	24	45	482	19	2012-08-02 05:48:36.095293
32517	4	1-9324	32516	64	63	348	86	2012-08-02 05:48:36.504015
32518	4	1-7765	32516	63	91	159	3	2012-08-02 05:48:36.504015
32519	4	1-8984	32516	91	28	114	68	2012-08-02 05:48:36.504015
32516	4	1-9806	32516	28	64	218	45	2012-08-02 05:48:36.504015
32521	4	1-8686	32520	89	77	562	64	2012-08-02 05:48:36.504015
32522	4	1-8817	32520	77	98	907	73	2012-08-02 05:48:36.504015
32523	4	1-9569	32520	98	28	427	68	2012-08-02 05:48:36.504015
32520	4	1-9806	32520	28	89	825	45	2012-08-02 05:48:36.504015
32525	4	1-9439	32524	52	65	1579	70	2012-08-02 05:48:37.29974
32526	4	1-9621	32524	65	29	1049	22	2012-08-02 05:48:37.29974
32527	4	1-8361	32524	29	84	626	52	2012-08-02 05:48:37.29974
32524	4	1-9808	32524	84	52	1407	18	2012-08-02 05:48:37.29974
32529	4	1-7014	32528	34	74	1244	90	2012-08-02 05:48:37.29974
32530	4	1-9503	32528	74	36	850	19	2012-08-02 05:48:37.29974
32531	4	1-7539	32528	36	84	651	52	2012-08-02 05:48:37.29974
32528	4	1-9808	32528	84	34	1479	18	2012-08-02 05:48:37.29974
32533	5	1-8802	32532	24	48	2544	37	2012-08-02 05:48:38.095017
32534	5	1-9611	32532	48	29	2011	22	2012-08-02 05:48:38.095017
32535	5	1-8361	32532	29	84	935	52	2012-08-02 05:48:38.095017
32536	5	1-9808	32532	84	46	1635	18	2012-08-02 05:48:38.095017
32532	5	1-9810	32532	46	24	1903	39	2012-08-02 05:48:38.095017
32538	5	1-7332	32537	17	24	218	29	2012-08-02 05:48:38.095017
32539	5	1-9805	32537	24	36	213	19	2012-08-02 05:48:38.095017
32540	5	1-7539	32537	36	84	129	52	2012-08-02 05:48:38.095017
32541	5	1-9808	32537	84	46	232	18	2012-08-02 05:48:38.095017
32537	5	1-9810	32537	46	17	277	39	2012-08-02 05:48:38.095017
32543	4	1-9337	32542	82	26	275	87	2012-08-02 05:48:38.51513
32544	4	1-9459	32542	26	48	225	28	2012-08-02 05:48:38.51513
32545	4	1-6942	32542	48	68	162	31	2012-08-02 05:48:38.51513
32542	4	1-9811	32542	68	82	132	97	2012-08-02 05:48:38.51513
32547	3	1-8517	32546	15	34	694	6	2012-08-02 05:48:39.321516
32548	3	1-8942	32546	34	54	161	61	2012-08-02 05:48:39.321516
32546	3	1-9813	32546	54	15	6386	17	2012-08-02 05:48:39.321516
32550	5	1-9347	32549	30	36	535	20	2012-08-02 05:48:39.741457
32551	5	1-7359	32549	36	82	292	34	2012-08-02 05:48:39.741457
32552	5	1-8522	32549	82	38	167	21	2012-08-02 05:48:39.741457
32553	5	1-9494	32549	38	98	221	27	2012-08-02 05:48:39.741457
32549	5	1-9814	32549	98	30	233	12	2012-08-02 05:48:39.741457
32555	5	1-9347	32554	30	7	1982	20	2012-08-02 05:48:39.741457
32556	5	1-9029	32554	7	82	1038	34	2012-08-02 05:48:39.741457
32557	5	1-8522	32554	82	38	601	21	2012-08-02 05:48:39.741457
32558	5	1-9494	32554	38	98	802	27	2012-08-02 05:48:39.741457
32554	5	1-9814	32554	98	30	855	12	2012-08-02 05:48:39.741457
32560	4	1-9574	32559	68	85	1733	25	2012-08-02 05:48:40.150335
32561	4	1-9709	32559	85	56	1289	55	2012-08-02 05:48:40.150335
32562	4	1-8980	32559	56	47	1972	5	2012-08-02 05:48:40.150335
32559	4	1-9815	32559	47	68	1993	15	2012-08-02 05:48:40.150335
32564	4	1-9670	32563	29	49	81	98	2012-08-02 05:48:40.150335
32565	4	1-3674	32563	49	52	46	54	2012-08-02 05:48:40.150335
32566	4	1-9427	32563	52	47	50	5	2012-08-02 05:48:40.150335
32563	4	1-9815	32563	47	29	53	15	2012-08-02 05:48:40.150335
32568	3	1-9205	32567	15	94	695	87	2012-08-02 05:48:41.763592
32569	3	1-9779	32567	94	53	891	37	2012-08-02 05:48:41.763592
32567	3	1-9819	32567	53	15	553	27	2012-08-02 05:48:41.763592
32571	4	1-8439	32570	39	51	2310	93	2012-08-02 05:48:42.17222
32572	4	1-8587	32570	51	57	1605	20	2012-08-02 05:48:42.17222
32573	4	1-9624	32570	57	28	1310	88	2012-08-02 05:48:42.17222
32570	4	1-9820	32570	28	39	6199	53	2012-08-02 05:48:42.17222
32575	4	1-9397	32574	91	60	2520	7	2012-08-02 05:48:43.38771
32576	4	1-7683	32574	60	65	1696	21	2012-08-02 05:48:43.38771
32577	4	1-7494	32574	65	93	1151	29	2012-08-02 05:48:43.38771
32574	4	1-9823	32574	93	91	2398	82	2012-08-02 05:48:43.38771
32579	2	1-9665	32578	74	3	4716	100	2012-08-02 05:48:45.013475
32578	2	1-9827	32578	3	74	4056	60	2012-08-02 05:48:45.013475
32581	5	1-2406	32580	19	36	247	9	2012-08-02 05:48:45.809003
32582	5	1-6831	32580	36	93	308	29	2012-08-02 05:48:45.809003
32583	5	1-9823	32580	93	91	497	82	2012-08-02 05:48:45.809003
32584	5	1-9397	32580	91	68	405	7	2012-08-02 05:48:45.809003
32580	5	1-9829	32580	68	19	1387	75	2012-08-02 05:48:45.809003
32586	4	1-6289	32585	8	93	297	29	2012-08-02 05:48:45.809003
32587	4	1-9823	32585	93	91	504	82	2012-08-02 05:48:45.809003
32588	4	1-9397	32585	91	68	431	7	2012-08-02 05:48:45.809003
32585	4	1-9829	32585	68	8	1555	75	2012-08-02 05:48:45.809003
32590	3	1-6924	32589	91	85	880	68	2012-08-02 05:48:45.809003
32591	3	1-8719	32589	85	68	518	7	2012-08-02 05:48:45.809003
32589	3	1-9829	32589	68	91	1898	75	2012-08-02 05:48:45.809003
32593	4	1-9257	32592	64	79	1125	15	2012-08-02 05:48:45.809003
32594	4	1-9812	32592	79	73	1093	35	2012-08-02 05:48:45.809003
32595	4	1-809	32592	73	68	517	7	2012-08-02 05:48:45.809003
32592	4	1-9829	32592	68	64	2090	75	2012-08-02 05:48:45.809003
32597	4	1-9304	32596	13	25	908	79	2012-08-02 05:48:45.809003
32598	4	1-9027	32596	25	73	220	35	2012-08-02 05:48:45.809003
32599	4	1-809	32596	73	68	112	7	2012-08-02 05:48:45.809003
32596	4	1-9829	32596	68	13	489	75	2012-08-02 05:48:45.809003
32601	5	1-9512	32600	93	63	2009	100	2012-08-02 05:48:46.228904
32602	5	1-8606	32600	63	46	1214	91	2012-08-02 05:48:46.228904
32603	5	1-8282	32600	46	46	473	18	2012-08-02 05:48:46.228904
32604	5	1-9810	32600	46	26	619	39	2012-08-02 05:48:46.228904
32600	5	1-9830	32600	26	93	1586	56	2012-08-02 05:48:46.228904
32606	5	1-3014	32605	4	67	3110	54	2012-08-02 05:48:46.228904
32607	5	1-8205	32605	67	46	1316	91	2012-08-02 05:48:46.228904
32608	5	1-8282	32605	46	46	637	18	2012-08-02 05:48:46.228904
32609	5	1-9810	32605	46	26	1035	39	2012-08-02 05:48:46.228904
32605	5	1-9830	32605	26	4	4090	56	2012-08-02 05:48:46.228904
32610	1	1-5400	\N	51	51	620	100	2012-08-02 05:48:46.228904
32612	4	1-7590	32611	7	61	1314	68	2012-08-02 05:48:47.841851
32613	4	1-7749	32611	61	62	895	36	2012-08-02 05:48:47.841851
32614	4	1-9704	32611	62	32	2242	10	2012-08-02 05:48:47.841851
32611	4	1-9834	32611	32	7	2285	52	2012-08-02 05:48:47.841851
32616	4	1-9261	32615	68	26	526	39	2012-08-02 05:48:48.2622
32617	4	1-9830	32615	26	9	1569	56	2012-08-02 05:48:48.2622
32618	4	1-2583	32615	9	50	513	50	2012-08-02 05:48:48.2622
32615	4	1-9835	32615	50	68	544	41	2012-08-02 05:48:48.2622
32620	4	1-9261	32619	68	24	378	39	2012-08-02 05:48:48.2622
32621	4	1-8802	32619	24	63	554	37	2012-08-02 05:48:48.2622
32622	4	1-9548	32619	63	50	336	50	2012-08-02 05:48:48.2622
32619	4	1-9835	32619	50	68	373	41	2012-08-02 05:48:48.2622
32624	3	1-6945	32623	64	18	39	81	2012-08-02 05:48:48.2622
32625	3	1-9661	32623	18	50	34	50	2012-08-02 05:48:48.2622
32623	3	1-9835	32623	50	64	42	41	2012-08-02 05:48:48.2622
32627	4	1-9502	32626	43	92	6904	84	2012-08-02 05:48:49.07957
32628	4	1-9365	32626	92	34	4930	77	2012-08-02 05:48:49.07957
32629	4	1-5508	32626	34	16	1429	73	2012-08-02 05:48:49.07957
32626	4	1-9837	32626	16	43	2783	89	2012-08-02 05:48:49.07957
32631	4	1-8663	32630	17	85	1227	24	2012-08-02 05:48:49.07957
32632	4	1-5429	32630	85	77	578	64	2012-08-02 05:48:49.07957
32633	4	1-8817	32630	77	16	970	73	2012-08-02 05:48:49.07957
32630	4	1-9837	32630	16	17	2081	89	2012-08-02 05:48:49.07957
32635	4	1-9577	32634	13	4	1711	60	2012-08-02 05:48:50.289463
32636	4	1-9522	32634	4	54	1367	44	2012-08-02 05:48:50.289463
32637	4	1-6937	32634	54	66	786	34	2012-08-02 05:48:50.289463
32634	4	1-9840	32634	66	13	780	24	2012-08-02 05:48:50.289463
32639	4	1-9242	32638	82	66	2185	98	2012-08-02 05:48:50.289463
32640	4	1-8555	32638	66	18	1879	93	2012-08-02 05:48:50.289463
32641	4	1-8658	32638	18	66	872	34	2012-08-02 05:48:50.289463
32638	4	1-9840	32638	66	82	925	24	2012-08-02 05:48:50.289463
32643	4	1-7354	32642	46	77	1527	64	2012-08-02 05:48:51.14535
32644	4	1-8817	32642	77	98	2573	73	2012-08-02 05:48:51.14535
32645	4	1-9520	32642	98	83	2878	1	2012-08-02 05:48:51.14535
32642	4	1-9842	32642	83	46	4503	20	2012-08-02 05:48:51.14535
32647	4	1-7247	32646	56	38	1036	70	2012-08-02 05:48:51.963868
32648	4	1-9252	32646	38	13	308	11	2012-08-02 05:48:51.963868
32649	4	1-9599	32646	13	64	667	96	2012-08-02 05:48:51.963868
32646	4	1-9844	32646	64	56	1175	20	2012-08-02 05:48:51.963868
32651	4	1-9835	32650	50	68	945	41	2012-08-02 05:48:53.578593
32652	4	1-9261	32650	68	24	1054	39	2012-08-02 05:48:53.578593
32653	4	1-8802	32650	24	94	1701	37	2012-08-02 05:48:53.578593
32650	4	1-9848	32650	94	50	772	50	2012-08-02 05:48:53.578593
32655	4	1-8452	32654	33	72	717	41	2012-08-02 05:48:53.97753
32656	4	1-9681	32654	72	93	731	56	2012-08-02 05:48:53.97753
32657	4	1-9512	32654	93	36	1217	100	2012-08-02 05:48:53.97753
32654	4	1-9849	32654	36	33	1088	38	2012-08-02 05:48:53.97753
32659	4	1-8661	32658	15	59	3768	69	2012-08-02 05:48:54.398738
32660	4	1-9833	32658	59	33	2955	52	2012-08-02 05:48:54.398738
32661	4	1-7455	32658	33	75	5103	2	2012-08-02 05:48:54.398738
32658	4	1-9850	32658	75	15	4382	59	2012-08-02 05:48:54.398738
32663	4	1-8661	32662	15	59	1347	69	2012-08-02 05:48:54.398738
32664	4	1-9833	32662	59	29	1106	52	2012-08-02 05:48:54.398738
32665	4	1-6167	32662	29	75	1665	2	2012-08-02 05:48:54.398738
32662	4	1-9850	32662	75	15	1497	59	2012-08-02 05:48:54.398738
32667	5	1-9835	32666	50	9	5	41	2012-08-02 05:48:55.593233
32668	5	1-7403	32666	9	18	4	16	2012-08-02 05:48:55.593233
32669	5	1-9789	32666	18	4	4	90	2012-08-02 05:48:55.593233
32670	5	1-9793	32666	4	99	4	62	2012-08-02 05:48:55.593233
32666	5	1-9853	32666	99	50	4	50	2012-08-02 05:48:55.593233
32672	4	1-9494	32671	38	52	2928	27	2012-08-02 05:48:56.003094
32673	4	1-9236	32671	52	71	1885	69	2012-08-02 05:48:56.003094
32674	4	1-9299	32671	71	4	1584	6	2012-08-02 05:48:56.003094
32671	4	1-9854	32671	4	38	2012	21	2012-08-02 05:48:56.003094
32675	1	1-8151	\N	77	77	416	27	2012-08-02 05:48:56.003094
32677	5	1-9483	32676	71	47	857	23	2012-08-02 05:48:56.414794
32678	5	1-9608	32676	47	85	588	68	2012-08-02 05:48:56.414794
32679	5	1-8719	32676	85	68	357	7	2012-08-02 05:48:56.414794
32680	5	1-9829	32676	68	20	1350	75	2012-08-02 05:48:56.414794
32676	5	1-9855	32676	20	71	699	96	2012-08-02 05:48:56.414794
32682	3	1-9651	32681	89	20	2972	75	2012-08-02 05:48:57.642823
32683	3	1-9855	32681	20	34	1571	96	2012-08-02 05:48:57.642823
32681	3	1-9858	32681	34	89	3988	35	2012-08-02 05:48:57.642823
32685	5	1-9723	32684	85	51	1276	44	2012-08-02 05:48:57.642823
32686	5	1-7672	32684	51	38	880	70	2012-08-02 05:48:57.642823
32687	5	1-9252	32684	38	13	269	11	2012-08-02 05:48:57.642823
32688	5	1-9599	32684	13	34	596	96	2012-08-02 05:48:57.642823
32684	5	1-9858	32684	34	85	1590	35	2012-08-02 05:48:57.642823
32690	4	1-7014	32689	34	4	2620	90	2012-08-02 05:48:58.43927
32691	4	1-9793	32689	4	99	1372	62	2012-08-02 05:48:58.43927
32692	4	1-9853	32689	99	5	820	50	2012-08-02 05:48:58.43927
32689	4	1-9860	32689	5	34	5874	18	2012-08-02 05:48:58.43927
32694	5	1-7794	32693	71	54	158	61	2012-08-02 05:48:58.837135
32695	5	1-7579	32693	54	24	132	29	2012-08-02 05:48:58.837135
32696	5	1-9805	32693	24	92	139	19	2012-08-02 05:48:58.837135
32697	5	1-9725	32693	92	12	197	75	2012-08-02 05:48:58.837135
32693	5	1-9861	32693	12	71	462	32	2012-08-02 05:48:58.837135
32699	4	1-5852	32698	20	66	174	68	2012-08-02 05:48:58.837135
32700	4	1-9647	32698	66	34	123	30	2012-08-02 05:48:58.837135
32701	4	1-9585	32698	34	12	172	75	2012-08-02 05:48:58.837135
32698	4	1-9861	32698	12	20	419	32	2012-08-02 05:48:58.837135
32703	5	1-8843	32702	89	57	1411	28	2012-08-02 05:48:58.837135
32704	5	1-8651	32702	57	74	865	96	2012-08-02 05:48:58.837135
32705	5	1-8225	32702	74	92	382	19	2012-08-02 05:48:58.837135
32706	5	1-9725	32702	92	12	654	75	2012-08-02 05:48:58.837135
32702	5	1-9861	32702	12	89	1847	32	2012-08-02 05:48:58.837135
32708	3	1-9816	32707	91	34	648	30	2012-08-02 05:48:58.837135
32709	3	1-9585	32707	34	12	1081	75	2012-08-02 05:48:58.837135
32707	3	1-9861	32707	12	91	3138	32	2012-08-02 05:48:58.837135
32711	3	1-9252	32710	38	38	320	11	2012-08-02 05:49:00.088168
32710	3	1-9864	32710	59	38	1194	70	2012-08-02 05:49:00.088168
32712	3	1-9528	32710	38	59	253	9	2012-08-02 05:49:00.088168
32714	5	1-3661	32713	31	50	533	50	2012-08-02 05:49:00.088168
32715	5	1-9835	32713	50	68	685	41	2012-08-02 05:49:00.088168
32716	5	1-7740	32713	68	38	219	11	2012-08-02 05:49:00.088168
32717	5	1-9528	32713	38	59	196	9	2012-08-02 05:49:00.088168
32713	5	1-9864	32713	59	31	1330	70	2012-08-02 05:49:00.088168
32719	4	1-9530	32718	86	93	608	2	2012-08-02 05:49:00.088168
32720	4	1-8304	32718	93	41	182	58	2012-08-02 05:49:00.088168
32721	4	1-9807	32718	41	59	74	9	2012-08-02 05:49:00.088168
32718	4	1-9864	32718	59	86	556	70	2012-08-02 05:49:00.088168
32723	4	1-9005	32722	64	16	1672	91	2012-08-02 05:49:00.896122
32724	4	1-9826	32722	16	70	1244	36	2012-08-02 05:49:00.896122
32725	4	1-9720	32722	70	89	1607	44	2012-08-02 05:49:00.896122
32722	4	1-9866	32722	89	64	2207	38	2012-08-02 05:49:00.896122
32727	4	1-7337	32726	23	50	1671	14	2012-08-02 05:49:00.896122
32728	4	1-9327	32726	50	4	1873	60	2012-08-02 05:49:00.896122
32729	4	1-9522	32726	4	89	1329	44	2012-08-02 05:49:00.896122
32726	4	1-9866	32726	89	23	2069	38	2012-08-02 05:49:00.896122
32731	4	1-8452	32730	33	68	114	41	2012-08-02 05:49:00.896122
32732	4	1-9261	32730	68	51	128	39	2012-08-02 05:49:00.896122
32733	4	1-5541	32730	51	89	103	44	2012-08-02 05:49:00.896122
32730	4	1-9866	32730	89	33	167	38	2012-08-02 05:49:00.896122
32735	4	1-9705	32734	30	34	155	96	2012-08-02 05:49:00.896122
32736	4	1-9858	32734	34	85	400	35	2012-08-02 05:49:00.896122
32737	4	1-9723	32734	85	89	310	44	2012-08-02 05:49:00.896122
32734	4	1-9866	32734	89	30	506	38	2012-08-02 05:49:00.896122
32739	2	1-6123	32738	89	11	562	22	2012-08-02 05:49:02.543868
32738	2	1-9870	32738	11	89	5676	23	2012-08-02 05:49:02.543868
32741	4	1-9668	32740	13	88	1503	2	2012-08-02 05:49:02.955284
32742	4	1-9706	32740	88	68	184	41	2012-08-02 05:49:02.955284
32743	4	1-9261	32740	68	39	185	39	2012-08-02 05:49:02.955284
32740	4	1-9871	32740	39	13	369	80	2012-08-02 05:49:02.955284
32745	3	1-7315	32744	32	51	2313	88	2012-08-02 05:49:04.159812
32746	3	1-9689	32744	51	4	1616	23	2012-08-02 05:49:04.159812
32744	3	1-9874	32744	4	32	2061	64	2012-08-02 05:49:04.159812
32748	3	1-9741	32747	32	38	4142	40	2012-08-02 05:49:04.977757
32749	3	1-9802	32747	38	52	3007	92	2012-08-02 05:49:04.977757
32747	3	1-9876	32747	52	32	4255	70	2012-08-02 05:49:04.977757
32751	4	1-5594	32750	93	93	848	29	2012-08-02 05:49:04.977757
32752	4	1-9823	32750	93	81	1430	82	2012-08-02 05:49:04.977757
32753	4	1-7834	32750	81	52	1146	92	2012-08-02 05:49:04.977757
32750	4	1-9876	32750	52	93	2360	70	2012-08-02 05:49:04.977757
32755	4	1-5594	32754	93	93	720	29	2012-08-02 05:49:04.977757
32756	4	1-9823	32754	93	66	1247	82	2012-08-02 05:49:04.977757
32757	4	1-6976	32754	66	52	898	92	2012-08-02 05:49:04.977757
32754	4	1-9876	32754	52	93	1951	70	2012-08-02 05:49:04.977757
32759	3	1-9388	32758	48	34	2769	75	2012-08-02 05:49:07.82463
32760	3	1-8430	32758	34	86	1933	78	2012-08-02 05:49:07.82463
32758	3	1-9883	32758	86	48	3901	51	2012-08-02 05:49:07.82463
32762	3	1-9388	32761	48	22	255	75	2012-08-02 05:49:07.82463
32763	3	1-8603	32761	22	86	175	78	2012-08-02 05:49:07.82463
32761	3	1-9883	32761	86	48	356	51	2012-08-02 05:49:07.82463
32765	4	1-7384	32764	36	70	1117	59	2012-08-02 05:49:08.655485
32766	4	1-9477	32764	70	4	795	23	2012-08-02 05:49:08.655485
32767	4	1-9874	32764	4	34	816	64	2012-08-02 05:49:08.655485
32764	4	1-9885	32764	34	36	8144	16	2012-08-02 05:49:08.655485
32769	4	1-7190	32768	31	51	36	88	2012-08-02 05:49:08.655485
32770	4	1-9689	32768	51	4	20	23	2012-08-02 05:49:08.655485
32771	4	1-9874	32768	4	34	21	64	2012-08-02 05:49:08.655485
32768	4	1-9885	32768	34	31	208	16	2012-08-02 05:49:08.655485
32773	4	1-8634	32772	95	21	3028	17	2012-08-02 05:49:09.075438
32774	4	1-5936	32772	21	84	634	93	2012-08-02 05:49:09.075438
32775	4	1-9441	32772	84	12	457	96	2012-08-02 05:49:09.075438
32772	4	1-9886	32772	12	95	1886	83	2012-08-02 05:49:09.075438
32777	3	1-9527	32776	74	55	770	81	2012-08-02 05:49:09.474205
32778	3	1-7664	32776	55	98	862	88	2012-08-02 05:49:09.474205
32776	3	1-9887	32776	98	74	687	36	2012-08-02 05:49:09.474205
32780	5	1-9685	32779	3	77	5193	52	2012-08-02 05:49:10.302964
32781	5	1-8056	32779	77	68	3376	31	2012-08-02 05:49:10.302964
32782	5	1-9811	32779	68	11	3064	97	2012-08-02 05:49:10.302964
32783	5	1-9519	32779	11	41	5685	78	2012-08-02 05:49:10.302964
32779	5	1-9889	32779	41	3	6237	46	2012-08-02 05:49:10.302964
32785	4	1-9886	32784	12	72	3055	83	2012-08-02 05:49:11.132101
32786	4	1-8437	32784	72	50	1941	14	2012-08-02 05:49:11.132101
32787	4	1-9327	32784	50	40	1579	60	2012-08-02 05:49:11.132101
32784	4	1-9891	32784	40	12	1209	96	2012-08-02 05:49:11.132101
32789	4	1-9886	32788	12	69	171	83	2012-08-02 05:49:11.132101
32790	4	1-9113	32788	69	41	89	6	2012-08-02 05:49:11.132101
32791	4	1-8542	32788	41	40	71	60	2012-08-02 05:49:11.132101
32788	4	1-9891	32788	40	12	61	96	2012-08-02 05:49:11.132101
32793	4	1-9886	32792	12	6	4816	83	2012-08-02 05:49:11.132101
32794	4	1-8807	32792	6	46	3485	95	2012-08-02 05:49:11.132101
32795	4	1-9308	32792	46	40	1864	60	2012-08-02 05:49:11.132101
32792	4	1-9891	32792	40	12	1649	96	2012-08-02 05:49:11.132101
32797	3	1-9858	32796	34	49	2058	35	2012-08-02 05:49:11.132101
32798	3	1-3662	32796	49	40	1101	60	2012-08-02 05:49:11.132101
32796	3	1-9891	32796	40	34	987	96	2012-08-02 05:49:11.132101
32800	4	1-9483	32799	71	47	1787	23	2012-08-02 05:49:11.132101
32801	4	1-9608	32799	47	41	1343	68	2012-08-02 05:49:11.132101
32802	4	1-9102	32799	41	40	1135	60	2012-08-02 05:49:11.132101
32799	4	1-9891	32799	40	71	1330	96	2012-08-02 05:49:11.132101
32804	4	1-9672	32803	87	98	1326	27	2012-08-02 05:49:11.564364
32805	4	1-9814	32803	98	86	1173	12	2012-08-02 05:49:11.564364
32806	4	1-8015	32803	86	38	554	52	2012-08-02 05:49:11.564364
32803	4	1-9892	32803	38	87	1744	48	2012-08-02 05:49:11.564364
32808	4	1-8614	32807	81	15	1115	59	2012-08-02 05:49:11.564364
32809	4	1-8661	32807	15	59	1123	69	2012-08-02 05:49:11.564364
32810	4	1-9833	32807	59	38	1032	52	2012-08-02 05:49:11.564364
32807	4	1-9892	32807	38	81	4373	48	2012-08-02 05:49:11.564364
32812	4	1-9864	32811	59	93	410	70	2012-08-02 05:49:12.371041
32813	4	1-5594	32811	93	21	105	29	2012-08-02 05:49:12.371041
32814	4	1-9618	32811	21	53	64	51	2012-08-02 05:49:12.371041
32811	4	1-9894	32811	53	59	96	9	2012-08-02 05:49:12.371041
32816	3	1-6831	32815	36	21	2916	29	2012-08-02 05:49:12.371041
32817	3	1-9618	32815	21	53	1875	51	2012-08-02 05:49:12.371041
32815	3	1-9894	32815	53	36	2973	9	2012-08-02 05:49:12.371041
32819	4	1-6831	32818	36	93	2338	29	2012-08-02 05:49:12.371041
32820	4	1-9823	32818	93	32	3086	82	2012-08-02 05:49:12.371041
32821	4	1-8370	32818	32	53	1388	51	2012-08-02 05:49:12.371041
32818	4	1-9894	32818	53	36	2290	9	2012-08-02 05:49:12.371041
32823	4	1-8902	32822	36	37	2565	68	2012-08-02 05:49:12.371041
32824	4	1-9394	32822	37	86	1390	78	2012-08-02 05:49:12.371041
32825	4	1-9883	32822	86	53	2009	51	2012-08-02 05:49:12.371041
32822	4	1-9894	32822	53	36	3498	9	2012-08-02 05:49:12.371041
32827	5	1-8152	32826	2	26	617	63	2012-08-02 05:49:13.18945
32828	5	1-9481	32826	26	9	925	35	2012-08-02 05:49:13.18945
32829	5	1-3800	32826	9	18	330	67	2012-08-02 05:49:13.18945
32830	5	1-9700	32826	18	14	1890	82	2012-08-02 05:49:13.18945
32826	5	1-9896	32826	14	2	2509	7	2012-08-02 05:49:13.18945
32832	3	1-9648	32831	22	52	3281	54	2012-08-02 05:49:14.416483
32833	3	1-9427	32831	52	90	2215	5	2012-08-02 05:49:14.416483
32831	3	1-9899	32831	90	22	2947	8	2012-08-02 05:49:14.416483
32835	2	1-7782	32834	28	90	3420	5	2012-08-02 05:49:14.416483
32834	2	1-9899	32834	90	28	4646	8	2012-08-02 05:49:14.416483
32837	4	1-9891	32836	40	98	1737	96	2012-08-02 05:49:15.234663
32838	4	1-9232	32836	98	97	3891	16	2012-08-02 05:49:15.234663
32839	4	1-8695	32836	97	86	1788	53	2012-08-02 05:49:15.234663
32836	4	1-9901	32836	86	40	1836	60	2012-08-02 05:49:15.234663
32841	2	1-8826	32840	69	49	809	69	2012-08-02 05:49:16.030517
32840	2	1-9903	32840	49	69	4368	14	2012-08-02 05:49:16.030517
32843	3	1-9518	32842	8	28	2333	68	2012-08-02 05:49:17.223699
32844	3	1-9806	32842	28	75	2845	45	2012-08-02 05:49:17.223699
32842	3	1-9906	32842	75	8	4433	49	2012-08-02 05:49:17.223699
32846	4	1-9444	32845	37	59	2319	69	2012-08-02 05:49:18.428365
32847	4	1-9833	32845	59	98	1484	52	2012-08-02 05:49:18.428365
32848	4	1-9869	32845	98	19	1014	13	2012-08-02 05:49:18.428365
32845	4	1-9909	32845	19	37	4473	84	2012-08-02 05:49:18.428365
32850	4	1-8460	32849	5	52	474	78	2012-08-02 05:49:18.428365
32851	4	1-8676	32849	52	23	609	14	2012-08-02 05:49:18.428365
32852	4	1-8098	32849	23	19	171	13	2012-08-02 05:49:18.428365
32849	4	1-9909	32849	19	5	964	84	2012-08-02 05:49:18.428365
32854	4	1-8902	32853	36	28	861	68	2012-08-02 05:49:18.826401
32855	4	1-9806	32853	28	84	1648	45	2012-08-02 05:49:18.826401
32856	4	1-5118	32853	84	48	1903	14	2012-08-02 05:49:18.826401
32853	4	1-9910	32853	48	36	841	9	2012-08-02 05:49:18.826401
32858	5	1-8717	32857	5	79	292	15	2012-08-02 05:49:19.635148
32859	5	1-9812	32857	79	42	272	35	2012-08-02 05:49:19.635148
32860	5	1-6590	32857	42	93	227	29	2012-08-02 05:49:19.635148
32861	5	1-9823	32857	93	29	415	82	2012-08-02 05:49:19.635148
32857	5	1-9912	32857	29	5	1004	77	2012-08-02 05:49:19.635148
32863	4	1-8605	32862	16	38	1591	49	2012-08-02 05:49:19.635148
32864	4	1-9795	32862	38	78	2046	27	2012-08-02 05:49:19.635148
32865	4	1-9180	32862	78	29	1465	82	2012-08-02 05:49:19.635148
32862	4	1-9912	32862	29	16	3727	77	2012-08-02 05:49:19.635148
32867	4	1-9840	32866	66	72	916	24	2012-08-02 05:49:20.851453
32868	4	1-9696	32866	72	84	582	93	2012-08-02 05:49:20.851453
32869	4	1-9441	32866	84	68	360	96	2012-08-02 05:49:20.851453
32866	4	1-9915	32866	68	66	1022	34	2012-08-02 05:49:20.851453
32871	4	1-9593	32870	27	39	1674	53	2012-08-02 05:49:20.851453
32872	4	1-8439	32870	39	84	1035	93	2012-08-02 05:49:20.851453
32873	4	1-9441	32870	84	68	753	96	2012-08-02 05:49:20.851453
32870	4	1-9915	32870	68	27	2514	34	2012-08-02 05:49:20.851453
32875	4	1-3265	32874	81	24	146	3	2012-08-02 05:49:24.530652
32876	4	1-8421	32874	24	68	99	96	2012-08-02 05:49:24.530652
32877	4	1-9915	32874	68	75	234	34	2012-08-02 05:49:24.530652
32874	4	1-9924	32874	75	81	1627	51	2012-08-02 05:49:24.530652
32879	4	1-9519	32878	11	98	296	78	2012-08-02 05:49:25.768195
32880	4	1-6553	32878	98	37	263	88	2012-08-02 05:49:25.768195
32881	4	1-9240	32878	37	10	259	71	2012-08-02 05:49:25.768195
32878	4	1-9927	32878	10	11	169	97	2012-08-02 05:49:25.768195
32883	4	1-9302	32882	22	48	1555	86	2012-08-02 05:49:25.768195
32884	4	1-9496	32882	48	41	2287	66	2012-08-02 05:49:25.768195
32885	4	1-7671	32882	41	10	1585	71	2012-08-02 05:49:25.768195
32882	4	1-9927	32882	10	22	1142	97	2012-08-02 05:49:25.768195
32887	5	1-8122	32886	14	7	994	1	2012-08-02 05:49:26.60852
32888	5	1-9118	32886	7	52	442	47	2012-08-02 05:49:26.60852
32889	5	1-9533	32886	52	71	625	41	2012-08-02 05:49:26.60852
32890	5	1-9792	32886	71	51	694	22	2012-08-02 05:49:26.60852
32886	5	1-9929	32886	51	14	2090	27	2012-08-02 05:49:26.60852
32892	5	1-9564	32891	10	27	1637	99	2012-08-02 05:49:26.60852
32893	5	1-5587	32891	27	68	797	21	2012-08-02 05:49:26.60852
32894	5	1-9312	32891	68	71	638	41	2012-08-02 05:49:26.60852
32895	5	1-9792	32891	71	51	911	22	2012-08-02 05:49:26.60852
32891	5	1-9929	32891	51	10	3533	27	2012-08-02 05:49:26.60852
32897	5	1-9236	32896	52	47	1470	69	2012-08-02 05:49:26.60852
32898	5	1-8495	32896	47	69	481	30	2012-08-02 05:49:26.60852
32899	5	1-8209	32896	69	71	382	41	2012-08-02 05:49:26.60852
32900	5	1-9792	32896	71	51	559	22	2012-08-02 05:49:26.60852
32896	5	1-9929	32896	51	52	2221	27	2012-08-02 05:49:26.60852
32902	2	1-7728	32901	94	47	1507	66	2012-08-02 05:49:27.437766
32901	2	1-9931	32901	47	94	1944	11	2012-08-02 05:49:27.437766
32904	4	1-9447	32903	87	4	2769	23	2012-08-02 05:49:27.437766
32905	4	1-9874	32903	4	100	2846	64	2012-08-02 05:49:27.437766
32906	4	1-9923	32903	100	47	1516	66	2012-08-02 05:49:27.437766
32903	4	1-9931	32903	47	87	2078	11	2012-08-02 05:49:27.437766
32908	5	1-7907	32907	78	49	409	75	2012-08-02 05:49:27.437766
32909	5	1-8702	32907	49	64	302	45	2012-08-02 05:49:27.437766
32910	5	1-9324	32907	64	48	402	86	2012-08-02 05:49:27.437766
32911	5	1-9496	32907	48	47	453	66	2012-08-02 05:49:27.437766
32907	5	1-9931	32907	47	78	659	11	2012-08-02 05:49:27.437766
32913	4	1-9689	32912	51	4	2165	23	2012-08-02 05:49:28.255409
32914	4	1-9874	32912	4	64	2455	64	2012-08-02 05:49:28.255409
32915	4	1-8768	32912	64	41	2529	53	2012-08-02 05:49:28.255409
32912	4	1-9933	32912	41	51	3486	88	2012-08-02 05:49:28.255409
32917	4	1-9056	32916	36	98	1747	14	2012-08-02 05:49:28.255409
32918	4	1-9326	32916	98	4	1391	58	2012-08-02 05:49:28.255409
32919	4	1-7911	32916	4	41	1530	53	2012-08-02 05:49:28.255409
32916	4	1-9933	32916	41	36	2627	88	2012-08-02 05:49:28.255409
32921	3	1-8881	32920	4	17	1176	12	2012-08-02 05:49:28.664271
32922	3	1-5134	32920	17	16	814	95	2012-08-02 05:49:28.664271
32920	3	1-9934	32920	16	4	1125	57	2012-08-02 05:49:28.664271
32924	4	1-8422	32923	87	95	127	3	2012-08-02 05:49:28.664271
32925	4	1-9370	32923	95	6	232	83	2012-08-02 05:49:28.664271
32926	4	1-8807	32923	6	16	229	95	2012-08-02 05:49:28.664271
32923	4	1-9934	32923	16	87	343	57	2012-08-02 05:49:28.664271
32928	4	1-9644	32927	49	13	304	83	2012-08-02 05:49:29.482193
32929	4	1-9311	32927	13	97	295	59	2012-08-02 05:49:29.482193
32930	4	1-9462	32927	97	69	162	90	2012-08-02 05:49:29.482193
32927	4	1-9936	32927	69	49	346	46	2012-08-02 05:49:29.482193
32932	4	1-9857	32931	79	77	3001	6	2012-08-02 05:49:29.482193
32933	4	1-8936	32931	77	34	1367	18	2012-08-02 05:49:29.482193
32934	4	1-7014	32931	34	69	1233	90	2012-08-02 05:49:29.482193
32931	4	1-9936	32931	69	79	2811	46	2012-08-02 05:49:29.482193
32936	4	1-7475	32935	54	62	209	56	2012-08-02 05:49:31.912435
32937	4	1-9879	32935	62	88	334	50	2012-08-02 05:49:31.912435
32938	4	1-9497	32935	88	85	887	17	2012-08-02 05:49:31.912435
32935	4	1-9942	32935	85	54	459	55	2012-08-02 05:49:31.912435
32940	4	1-8343	32939	63	64	916	38	2012-08-02 05:49:31.912435
32941	4	1-9005	32939	64	58	859	91	2012-08-02 05:49:31.912435
32942	4	1-9541	32939	58	85	1846	17	2012-08-02 05:49:31.912435
32939	4	1-9942	32939	85	63	972	55	2012-08-02 05:49:31.912435
32944	4	1-8902	32943	36	66	1620	68	2012-08-02 05:49:32.322021
32945	4	1-9647	32943	66	34	1281	30	2012-08-02 05:49:32.322021
32946	4	1-9585	32943	34	40	2015	75	2012-08-02 05:49:32.322021
32943	4	1-9943	32943	40	36	1589	9	2012-08-02 05:49:32.322021
32948	2	1-9176	32947	24	52	539	18	2012-08-02 05:49:33.128531
32947	2	1-9945	32947	52	24	4675	67	2012-08-02 05:49:33.128531
32950	5	1-9005	32949	64	32	321	91	2012-08-02 05:49:33.559525
32951	5	1-8504	32949	32	66	148	68	2012-08-02 05:49:33.559525
32952	5	1-9647	32949	66	34	121	30	2012-08-02 05:49:33.559525
32953	5	1-9585	32949	34	86	198	75	2012-08-02 05:49:33.559525
32949	5	1-9946	32949	86	64	326	38	2012-08-02 05:49:33.559525
32955	5	1-8587	32954	51	57	36	20	2012-08-02 05:49:35.592647
32956	5	1-9624	32954	57	5	45	88	2012-08-02 05:49:35.592647
32957	5	1-9552	32954	5	9	28	35	2012-08-02 05:49:35.592647
32958	5	1-3800	32954	9	35	12	67	2012-08-02 05:49:35.592647
32954	5	1-9951	32954	35	51	34	93	2012-08-02 05:49:35.592647
32960	5	1-8878	32959	77	8	200	94	2012-08-02 05:49:35.592647
32961	5	1-9938	32959	8	79	159	15	2012-08-02 05:49:35.592647
32962	5	1-9812	32959	79	9	167	35	2012-08-02 05:49:35.592647
32963	5	1-3800	32959	9	35	75	67	2012-08-02 05:49:35.592647
32959	5	1-9951	32959	35	77	222	93	2012-08-02 05:49:35.592647
32965	5	1-9381	32964	4	57	3185	20	2012-08-02 05:49:35.592647
32966	5	1-9624	32964	57	5	4338	88	2012-08-02 05:49:35.592647
32967	5	1-9552	32964	5	9	2966	35	2012-08-02 05:49:35.592647
32968	5	1-3800	32964	9	35	1363	67	2012-08-02 05:49:35.592647
32964	5	1-9951	32964	35	4	4167	93	2012-08-02 05:49:35.592647
32969	1	1-3800	\N	9	9	94	67	2012-08-02 05:49:35.592647
32971	4	1-9608	32970	47	28	789	68	2012-08-02 05:49:37.217105
32972	4	1-9806	32970	28	98	1192	45	2012-08-02 05:49:37.217105
32973	4	1-8780	32970	98	94	1694	80	2012-08-02 05:49:37.217105
32970	4	1-9955	32970	94	47	1334	23	2012-08-02 05:49:37.217105
32975	4	1-8152	32974	2	34	1819	63	2012-08-02 05:49:38.854063
32976	4	1-8978	32974	34	81	2521	91	2012-08-02 05:49:38.854063
32977	4	1-9171	32974	81	74	1621	95	2012-08-02 05:49:38.854063
32974	4	1-9959	32974	74	2	7229	7	2012-08-02 05:49:38.854063
32979	3	1-9480	32978	5	61	912	68	2012-08-02 05:49:38.854063
32980	3	1-9192	32978	61	74	421	95	2012-08-02 05:49:38.854063
32978	3	1-9959	32978	74	5	1948	7	2012-08-02 05:49:38.854063
32982	4	1-7478	32981	29	100	2587	69	2012-08-02 05:49:39.727034
32983	4	1-9911	32981	100	47	2852	84	2012-08-02 05:49:39.727034
32984	4	1-9245	32981	47	70	2551	72	2012-08-02 05:49:39.727034
32981	4	1-9961	32981	70	29	5045	45	2012-08-02 05:49:39.727034
32986	4	1-9382	32985	93	61	2238	6	2012-08-02 05:49:39.727034
32987	4	1-5225	32985	61	48	1314	59	2012-08-02 05:49:39.727034
32988	4	1-9358	32985	48	70	1424	72	2012-08-02 05:49:39.727034
32985	4	1-9961	32985	70	93	3013	45	2012-08-02 05:49:39.727034
32990	4	1-6916	32989	97	1	554	86	2012-08-02 05:49:42.974963
32991	4	1-9838	32989	1	90	636	34	2012-08-02 05:49:42.974963
32992	4	1-6639	32989	90	65	831	29	2012-08-02 05:49:42.974963
32989	4	1-9969	32989	65	97	1674	69	2012-08-02 05:49:42.974963
32994	3	1-9648	32993	22	1	4569	54	2012-08-02 05:49:43.837396
32995	3	1-9016	32993	1	58	2361	23	2012-08-02 05:49:43.837396
32993	3	1-9971	32993	58	22	2163	8	2012-08-02 05:49:43.837396
32997	4	1-7782	32996	28	47	50	5	2012-08-02 05:49:44.283805
32998	4	1-9815	32996	47	68	43	15	2012-08-02 05:49:44.283805
32999	4	1-9574	32996	68	64	32	25	2012-08-02 05:49:44.283805
32996	4	1-9972	32996	64	28	53	8	2012-08-02 05:49:44.283805
33001	5	1-9648	33000	22	67	324	54	2012-08-02 05:49:44.283805
33002	5	1-8205	33000	67	16	99	91	2012-08-02 05:49:44.283805
33003	5	1-9826	33000	16	73	72	36	2012-08-02 05:49:44.283805
33004	5	1-9784	33000	73	64	113	25	2012-08-02 05:49:44.283805
33000	5	1-9972	33000	64	22	204	8	2012-08-02 05:49:44.283805
33006	4	1-9636	33005	93	69	2315	97	2012-08-02 05:49:44.283805
33007	4	1-5578	33005	69	73	931	36	2012-08-02 05:49:44.283805
33008	4	1-9784	33005	73	64	1650	25	2012-08-02 05:49:44.283805
33005	4	1-9972	33005	64	93	3351	8	2012-08-02 05:49:44.283805
33010	5	1-1506	33009	5	57	1091	33	2012-08-02 05:49:44.283805
33011	5	1-9841	33009	57	16	1381	91	2012-08-02 05:49:44.283805
33012	5	1-9826	33009	16	73	1126	36	2012-08-02 05:49:44.283805
33013	5	1-9784	33009	73	64	2006	25	2012-08-02 05:49:44.283805
33009	5	1-9972	33009	64	5	4096	8	2012-08-02 05:49:44.283805
33015	4	1-6353	33014	54	27	122	4	2012-08-02 05:49:46.31251
33016	4	1-8981	33014	27	1	71	86	2012-08-02 05:49:46.31251
33017	4	1-9838	33014	1	79	55	34	2012-08-02 05:49:46.31251
33014	4	1-9977	33014	79	54	1046	79	2012-08-02 05:49:46.31251
33019	4	1-5618	33018	78	68	131	62	2012-08-02 05:49:46.31251
33020	4	1-9729	33018	68	36	92	49	2012-08-02 05:49:46.31251
33021	4	1-9583	33018	36	79	33	34	2012-08-02 05:49:46.31251
33018	4	1-9977	33018	79	78	629	79	2012-08-02 05:49:46.31251
33023	4	1-9920	33022	36	8	623	27	2012-08-02 05:49:46.31251
33024	4	1-6141	33022	8	18	331	93	2012-08-02 05:49:46.31251
33025	4	1-8658	33022	18	79	108	34	2012-08-02 05:49:46.31251
33022	4	1-9977	33022	79	36	2168	79	2012-08-02 05:49:46.31251
33026	1	1-4052	\N	41	41	695	79	2012-08-02 05:49:46.31251
33028	4	1-8919	33027	3	68	486	31	2012-08-02 05:49:46.721681
33029	4	1-9811	33027	68	40	362	97	2012-08-02 05:49:46.721681
33030	4	1-3759	33027	40	16	84	32	2012-08-02 05:49:46.721681
33027	4	1-9978	33027	16	3	720	49	2012-08-02 05:49:46.721681
33032	5	1-9560	33031	95	77	1328	80	2012-08-02 05:49:46.721681
33033	5	1-9928	33031	77	87	925	57	2012-08-02 05:49:46.721681
33034	5	1-8422	33031	87	100	310	3	2012-08-02 05:49:46.721681
33035	5	1-7141	33031	100	16	213	32	2012-08-02 05:49:46.721681
33031	5	1-9978	33031	16	95	1982	49	2012-08-02 05:49:46.721681
33037	4	1-2869	33036	32	96	27	63	2012-08-02 05:49:46.721681
33038	4	1-8828	33036	96	6	23	41	2012-08-02 05:49:46.721681
33039	4	1-3861	33036	6	16	9	32	2012-08-02 05:49:46.721681
33036	4	1-9978	33036	16	32	85	49	2012-08-02 05:49:46.721681
33041	4	1-9518	33040	8	40	715	68	2012-08-02 05:49:46.721681
33042	4	1-5251	33040	40	62	244	4	2012-08-02 05:49:46.721681
33043	4	1-8232	33040	62	16	92	32	2012-08-02 05:49:46.721681
33040	4	1-9978	33040	16	8	891	49	2012-08-02 05:49:46.721681
33044	1	1-2869	\N	32	32	493	63	2012-08-02 05:49:46.721681
33046	4	1-9857	33045	79	65	2603	6	2012-08-02 05:49:48.787994
33047	4	1-9715	33045	65	86	1189	78	2012-08-02 05:49:48.787994
33048	4	1-9883	33045	86	87	2143	51	2012-08-02 05:49:48.787994
33045	4	1-9983	33045	87	79	2943	46	2012-08-02 05:49:48.787994
33050	5	1-9245	33049	47	58	783	72	2012-08-02 05:49:50.434186
33051	5	1-9595	33049	58	20	1287	76	2012-08-02 05:49:50.434186
33052	5	1-5024	33049	20	49	232	33	2012-08-02 05:49:50.434186
33053	5	1-8863	33049	49	51	154	47	2012-08-02 05:49:50.434186
33049	5	1-9987	33049	51	47	1066	84	2012-08-02 05:49:50.434186
33055	3	1-8460	33054	5	7	872	78	2012-08-02 05:49:50.434186
33056	3	1-8929	33054	7	51	272	47	2012-08-02 05:49:50.434186
33054	3	1-9987	33054	51	5	2002	84	2012-08-02 05:49:50.434186
33058	4	1-5873	33057	32	73	218	36	2012-08-02 05:49:50.434186
33059	4	1-9784	33057	73	29	442	25	2012-08-02 05:49:50.434186
33060	4	1-9064	33057	29	51	155	47	2012-08-02 05:49:50.434186
33057	4	1-9987	33057	51	32	1330	84	2012-08-02 05:49:50.434186
33062	3	1-7564	33061	77	65	1253	29	2012-08-02 05:49:51.2851
33063	3	1-9969	33061	65	87	2398	69	2012-08-02 05:49:51.2851
33061	3	1-9989	33061	87	77	3782	60	2012-08-02 05:49:51.2851
33065	4	1-7579	33064	54	65	3608	29	2012-08-02 05:49:51.671711
33066	4	1-9969	33064	65	19	4236	69	2012-08-02 05:49:51.671711
33067	4	1-8831	33064	19	88	2008	70	2012-08-02 05:49:51.671711
33064	4	1-9990	33064	88	54	5973	61	2012-08-02 05:49:51.671711
33069	4	1-7327	33068	71	14	819	71	2012-08-02 05:49:51.671711
33070	4	1-9930	33068	14	51	836	44	2012-08-02 05:49:51.671711
33071	4	1-7672	33068	51	88	390	70	2012-08-02 05:49:51.671711
33068	4	1-9990	33068	88	71	1256	61	2012-08-02 05:49:51.671711
33073	4	1-9973	33072	26	98	3262	14	2012-08-02 05:49:51.671711
33074	4	1-9326	33072	98	58	1600	58	2012-08-02 05:49:51.671711
33075	4	1-7438	33072	58	88	741	70	2012-08-02 05:49:51.671711
33072	4	1-9990	33072	88	26	2388	61	2012-08-02 05:49:51.671711
33077	2	1-8056	33076	77	8	1217	31	2012-08-02 05:49:52.500517
33076	2	1-9992	33076	8	77	3146	52	2012-08-02 05:49:52.500517
33079	3	1-9869	33078	98	39	3912	13	2012-08-02 05:49:52.500517
33080	3	1-9336	33078	39	8	1747	31	2012-08-02 05:49:52.500517
33078	3	1-9992	33078	8	98	5580	52	2012-08-02 05:49:52.500517
33082	3	1-7590	33081	7	47	271	68	2012-08-02 05:49:52.500517
33083	3	1-7215	33081	47	8	141	31	2012-08-02 05:49:52.500517
33081	3	1-9992	33081	8	7	545	52	2012-08-02 05:49:52.500517
33085	4	1-9892	33084	38	5	443	48	2012-08-02 05:49:52.500517
33086	4	1-4609	33084	5	68	81	41	2012-08-02 05:49:52.500517
33087	4	1-8601	33084	68	8	30	31	2012-08-02 05:49:52.500517
33084	4	1-9992	33084	8	38	119	52	2012-08-02 05:49:52.500517
33089	3	1-9826	33088	16	74	2167	36	2012-08-02 05:49:52.909691
33090	3	1-9527	33088	74	88	1130	81	2012-08-02 05:49:52.909691
33088	3	1-9993	33088	88	16	4729	91	2012-08-02 05:49:52.909691
33092	3	1-8694	33091	22	50	2066	76	2012-08-02 05:49:54.986724
33093	3	1-9445	33091	50	76	1843	28	2012-08-02 05:49:54.986724
33091	3	1-9998	33091	76	22	1373	24	2012-08-02 05:49:54.986724
33095	4	1-9095	33094	71	74	303	88	2012-08-02 09:10:31.355931
33096	4	1-9575	33094	74	81	1214	5	2012-08-02 09:10:31.355931
33097	4	1-6090	33094	81	78	83	86	2012-08-02 09:10:31.355931
33094	4	1-10005	33094	78	71	472	74	2012-08-02 09:10:31.355931
33099	4	1-7853	33098	25	64	1072	100	2012-08-02 09:10:31.355931
33100	4	1-8604	33098	64	64	414	45	2012-08-02 09:10:31.355931
33101	4	1-9324	33098	64	78	609	86	2012-08-02 09:10:31.355931
33098	4	1-10005	33098	78	25	4462	74	2012-08-02 09:10:31.355931
33103	4	1-8184	33102	95	52	1441	47	2012-08-02 09:10:31.355931
33104	4	1-9533	33102	52	79	2024	41	2012-08-02 09:10:31.355931
33105	4	1-9260	33102	79	78	513	86	2012-08-02 09:10:31.355931
33102	4	1-10005	33102	78	95	3930	74	2012-08-02 09:10:31.355931
33107	3	1-9932	33106	65	82	4139	89	2012-08-02 09:10:31.811847
33108	3	1-9692	33106	82	1	3719	34	2012-08-02 09:10:31.811847
33106	3	1-10006	33106	1	65	7939	61	2012-08-02 09:10:31.811847
33110	4	1-9973	33109	26	48	681	14	2012-08-02 09:10:31.811847
33111	4	1-9910	33109	48	50	267	9	2012-08-02 09:10:31.811847
33112	4	1-7534	33109	50	1	159	34	2012-08-02 09:10:31.811847
33109	4	1-10006	33109	1	26	374	61	2012-08-02 09:10:31.811847
33114	5	1-9792	33113	71	28	335	22	2012-08-02 10:16:55.717318
33115	5	1-9979	33113	28	66	184	43	2012-08-02 10:16:55.717318
33116	5	1-9127	33113	66	4	177	6	2012-08-02 10:16:55.717318
33117	5	1-9854	33113	4	60	219	21	2012-08-02 10:16:55.717318
33113	5	1-10011	33113	60	71	242	41	2012-08-02 10:16:55.717318
33119	5	1-9875	33118	75	26	1483	100	2012-08-02 10:16:55.717318
33120	5	1-8575	33118	26	71	1108	69	2012-08-02 10:16:55.717318
33121	5	1-9299	33118	71	4	925	6	2012-08-02 10:16:55.717318
33122	5	1-9854	33118	4	60	1168	21	2012-08-02 10:16:55.717318
33118	5	1-10011	33118	60	75	1318	41	2012-08-02 10:16:55.717318
33124	5	1-9792	33123	71	83	5141	22	2012-08-02 10:16:55.717318
33125	5	1-5619	33123	83	66	2276	43	2012-08-02 10:16:55.717318
33126	5	1-9127	33123	66	4	2315	6	2012-08-02 10:16:55.717318
33127	5	1-9854	33123	4	60	3019	21	2012-08-02 10:16:55.717318
33123	5	1-10011	33123	60	71	3517	41	2012-08-02 10:16:55.717318
33129	5	1-9936	33128	69	92	1318	46	2012-08-02 10:16:56.176296
33130	5	1-7789	33128	92	13	724	83	2012-08-02 10:16:56.176296
33131	5	1-9311	33128	13	16	718	59	2012-08-02 10:16:56.176296
33132	5	1-8067	33128	16	23	365	32	2012-08-02 10:16:56.176296
33128	5	1-10012	33128	23	69	604	90	2012-08-02 10:16:56.176296
33134	4	1-9527	33133	74	80	1993	81	2012-08-02 10:16:56.587857
33135	4	1-9581	33133	80	4	2241	58	2012-08-02 10:16:56.587857
33136	4	1-7911	33133	4	88	2136	53	2012-08-02 10:16:56.587857
33133	4	1-10013	33133	88	74	2006	36	2012-08-02 10:16:56.587857
33138	4	1-9527	33137	74	45	3073	81	2012-08-02 10:16:56.587857
33139	4	1-9999	33137	45	97	5066	16	2012-08-02 10:16:56.587857
33140	4	1-8695	33137	97	88	2876	53	2012-08-02 10:16:56.587857
33137	4	1-10013	33137	88	74	2890	36	2012-08-02 10:16:56.587857
33142	3	1-9206	33141	30	91	2839	77	2012-08-02 10:16:56.987407
33143	3	1-8088	33141	91	72	2030	17	2012-08-02 10:16:56.987407
33141	3	1-10014	33141	72	30	3435	7	2012-08-02 10:16:56.987407
33145	4	1-9116	33144	11	14	466	71	2012-08-02 10:16:56.987407
33146	4	1-9930	33144	14	89	674	44	2012-08-02 10:16:56.987407
33147	4	1-9965	33144	89	72	448	17	2012-08-02 10:16:56.987407
33144	4	1-10014	33144	72	11	930	7	2012-08-02 10:16:56.987407
33149	4	1-9915	33148	68	27	699	34	2012-08-02 10:16:57.420724
33150	4	1-9593	33148	27	91	361	53	2012-08-02 10:16:57.420724
33151	4	1-9660	33148	91	5	293	87	2012-08-02 10:16:57.420724
33148	4	1-10015	33148	5	68	270	96	2012-08-02 10:16:57.420724
33153	4	1-9232	33152	98	47	2065	16	2012-08-02 10:16:57.420724
33154	4	1-9360	33152	47	60	1144	31	2012-08-02 10:16:57.420724
33155	4	1-9905	33152	60	5	849	87	2012-08-02 10:16:57.420724
33152	4	1-10015	33152	5	98	834	96	2012-08-02 10:16:57.420724
33157	4	1-9483	33156	71	82	4963	23	2012-08-02 10:16:57.420724
33158	4	1-2629	33156	82	91	3306	53	2012-08-02 10:16:57.420724
33159	4	1-9660	33156	91	5	3264	87	2012-08-02 10:16:57.420724
33156	4	1-10015	33156	5	71	3645	96	2012-08-02 10:16:57.420724
33161	3	1-8225	33160	74	51	1133	19	2012-08-02 10:16:57.420724
33162	3	1-9129	33160	51	5	2221	87	2012-08-02 10:16:57.420724
33160	3	1-10015	33160	5	74	2538	96	2012-08-02 10:16:57.420724
33164	4	1-8679	33163	63	69	6	90	2012-08-02 10:16:58.663543
33165	4	1-9936	33163	69	28	12	46	2012-08-02 10:16:58.663543
33166	4	1-8901	33163	28	19	14	58	2012-08-02 10:16:58.663543
33163	4	1-10018	33163	19	63	12	13	2012-08-02 10:16:58.663543
33168	3	1-8761	33167	86	100	664	69	2012-08-02 10:16:59.437265
33169	3	1-9911	33167	100	98	559	84	2012-08-02 10:16:59.437265
33167	3	1-10020	33167	98	86	1239	8	2012-08-02 10:16:59.437265
33171	3	1-6670	33170	63	18	1656	51	2012-08-02 10:16:59.437265
33172	3	1-9925	33170	18	98	908	84	2012-08-02 10:16:59.437265
33170	3	1-10020	33170	98	63	2022	8	2012-08-02 10:16:59.437265
33174	4	1-9093	33173	95	62	2862	86	2012-08-02 10:16:59.437265
33175	4	1-9130	33173	62	8	1021	75	2012-08-02 10:16:59.437265
33176	4	1-9935	33173	8	98	2073	84	2012-08-02 10:16:59.437265
33173	4	1-10020	33173	98	95	6336	8	2012-08-02 10:16:59.437265
33178	4	1-9533	33177	52	68	727	41	2012-08-02 10:19:30.60809
33179	4	1-9261	33177	68	98	795	39	2012-08-02 10:19:30.60809
33180	4	1-8721	33177	98	89	919	14	2012-08-02 10:19:30.60809
33177	4	1-10021	33177	89	52	460	47	2012-08-02 10:19:30.60809
33182	4	1-8783	33181	83	63	232	77	2012-08-02 10:19:30.60809
33183	4	1-9995	33181	63	38	662	5	2012-08-02 10:19:30.60809
33184	4	1-7797	33181	38	89	477	14	2012-08-02 10:19:30.60809
33181	4	1-10021	33181	89	83	271	47	2012-08-02 10:19:30.60809
33186	4	1-9533	33185	52	64	753	41	2012-08-02 10:19:30.60809
33187	4	1-6945	33185	64	42	793	81	2012-08-02 10:19:30.60809
33188	4	1-9976	33185	42	89	723	14	2012-08-02 10:19:30.60809
33185	4	1-10021	33185	89	52	415	47	2012-08-02 10:19:30.60809
33190	4	1-8140	33189	48	56	993	64	2012-08-02 10:19:48.585218
33191	4	1-8657	33189	56	99	964	73	2012-08-02 10:19:48.585218
33192	4	1-9393	33189	99	34	779	98	2012-08-02 10:19:48.585218
33189	4	1-10031	33189	34	48	1631	6	2012-08-02 10:19:48.585218
33194	2	1-9287	33193	53	28	1331	100	2012-08-02 10:19:49.445183
33193	2	1-10033	33193	28	53	3298	43	2012-08-02 10:19:49.445183
33196	5	1-7377	33195	50	30	1176	22	2012-08-02 10:19:50.650094
33197	5	1-8948	33195	30	16	1020	59	2012-08-02 10:19:50.650094
33198	5	1-8067	33195	16	23	502	32	2012-08-02 10:19:50.650094
33199	5	1-10012	33195	23	80	804	90	2012-08-02 10:19:50.650094
33195	5	1-10036	33195	80	50	1741	77	2012-08-02 10:19:50.650094
33201	5	1-7033	33200	18	78	951	55	2012-08-02 10:19:50.650094
33202	5	1-9472	33200	78	16	875	59	2012-08-02 10:19:50.650094
33203	5	1-8067	33200	16	23	468	32	2012-08-02 10:19:50.650094
33204	5	1-10012	33200	23	80	815	90	2012-08-02 10:19:50.650094
33200	5	1-10036	33200	80	18	1917	77	2012-08-02 10:19:50.650094
33206	3	1-8814	33205	51	67	2187	99	2012-08-02 10:19:51.063009
33207	3	1-7286	33205	67	12	1593	68	2012-08-02 10:19:51.063009
33205	3	1-10037	33205	12	51	5972	13	2012-08-02 10:19:51.063009
33209	4	1-9051	33208	20	14	700	82	2012-08-02 10:19:51.063009
33210	4	1-9896	33208	14	5	842	7	2012-08-02 10:19:51.063009
33211	4	1-9480	33208	5	12	337	68	2012-08-02 10:19:51.063009
33208	4	1-10037	33208	12	20	1304	13	2012-08-02 10:19:51.063009
33213	3	1-5165	33212	4	94	305	31	2012-08-02 10:19:51.473529
33214	3	1-9716	33212	94	51	267	80	2012-08-02 10:19:51.473529
33212	3	1-10038	33212	51	4	737	90	2012-08-02 10:19:51.473529
33216	4	1-9793	33215	4	68	4215	62	2012-08-02 10:19:51.473529
33217	4	1-9729	33215	68	95	3153	49	2012-08-02 10:19:51.473529
33218	4	1-9560	33215	95	51	1731	80	2012-08-02 10:19:51.473529
33215	4	1-10038	33215	51	4	5602	90	2012-08-02 10:19:51.473529
33220	4	1-9602	33219	88	75	706	15	2012-08-02 10:19:51.883805
33221	4	1-6726	33219	75	47	213	16	2012-08-02 10:19:51.883805
33222	4	1-9360	33219	47	65	93	31	2012-08-02 10:19:51.883805
33219	4	1-10039	33219	65	88	1265	12	2012-08-02 10:19:51.883805
33224	4	1-4431	33223	24	45	301	81	2012-08-02 10:19:51.883805
33225	4	1-9999	33223	45	47	406	16	2012-08-02 10:19:51.883805
33226	4	1-9360	33223	47	65	206	31	2012-08-02 10:19:51.883805
33223	4	1-10039	33223	65	24	3279	12	2012-08-02 10:19:51.883805
33228	4	1-9842	33227	83	77	1893	20	2012-08-02 10:20:57.02679
33229	4	1-6419	33227	77	92	800	46	2012-08-02 10:20:57.02679
33230	4	1-9591	33227	92	22	309	28	2012-08-02 10:20:57.02679
33227	4	1-10042	33227	22	83	1607	1	2012-08-02 10:20:57.02679
33232	4	1-9523	33231	20	12	1168	11	2012-08-02 10:20:57.02679
33233	4	1-9919	33231	12	9	795	30	2012-08-02 10:20:57.02679
33234	4	1-4799	33231	9	22	424	28	2012-08-02 10:20:57.02679
33231	4	1-10042	33231	22	20	2303	1	2012-08-02 10:20:57.02679
33236	4	1-7159	33235	75	48	114	37	2012-08-02 10:20:57.02679
33237	4	1-9014	33235	48	50	40	76	2012-08-02 10:20:57.02679
33238	4	1-9445	33235	50	22	32	28	2012-08-02 10:20:57.02679
33235	4	1-10042	33235	22	75	178	1	2012-08-02 10:20:57.02679
33240	4	1-7451	33239	80	89	683	94	2012-08-02 10:20:57.02679
33241	4	1-9713	33239	89	50	413	76	2012-08-02 10:20:57.02679
33242	4	1-9445	33239	50	22	335	28	2012-08-02 10:20:57.02679
33239	4	1-10042	33239	22	80	1870	1	2012-08-02 10:20:57.02679
33244	4	1-945	33243	80	88	612	53	2012-08-02 10:21:01.968377
33245	4	1-10013	33243	88	62	667	36	2012-08-02 10:21:01.968377
33246	4	1-10027	33243	62	69	627	57	2012-08-02 10:21:01.968377
33243	4	1-10054	33243	69	80	2185	70	2012-08-02 10:21:01.968377
33248	5	1-9951	33247	35	39	1024	93	2012-08-02 10:21:02.36707
33249	5	1-8231	33247	39	77	455	97	2012-08-02 10:21:02.36707
33250	5	1-8788	33247	77	69	311	79	2012-08-02 10:21:02.36707
33251	5	1-10028	33247	69	40	345	22	2012-08-02 10:21:02.36707
33247	5	1-10055	33247	40	35	446	67	2012-08-02 10:21:02.36707
33253	5	1-9939	33252	40	21	313	77	2012-08-02 10:21:02.36707
33254	5	1-3587	33252	21	16	101	37	2012-08-02 10:21:02.36707
33255	5	1-9804	33252	16	69	75	79	2012-08-02 10:21:02.36707
33256	5	1-10028	33252	69	40	72	22	2012-08-02 10:21:02.36707
33252	5	1-10055	33252	40	40	100	67	2012-08-02 10:21:02.36707
33258	4	1-8710	33257	74	16	1944	37	2012-08-02 10:21:02.36707
33259	4	1-9804	33257	16	69	1900	79	2012-08-02 10:21:02.36707
33260	4	1-10028	33257	69	40	2390	22	2012-08-02 10:21:02.36707
33257	4	1-10055	33257	40	74	3507	67	2012-08-02 10:21:02.36707
33262	4	1-6224	33261	21	33	527	38	2012-08-02 10:21:02.36707
33263	4	1-8452	33261	33	71	383	41	2012-08-02 10:21:02.36707
33264	4	1-9792	33261	71	40	521	22	2012-08-02 10:21:02.36707
33261	4	1-10055	33261	40	21	830	67	2012-08-02 10:21:02.36707
33266	3	1-6224	33265	21	35	363	38	2012-08-02 10:21:02.36707
33267	3	1-9455	33265	35	40	306	22	2012-08-02 10:21:02.36707
33265	3	1-10055	33265	40	21	528	67	2012-08-02 10:21:02.36707
33269	4	1-9741	33268	32	97	1660	40	2012-08-02 10:21:02.767505
33270	4	1-2761	33268	97	13	454	14	2012-08-02 10:21:02.767505
33271	4	1-9040	33268	13	35	508	92	2012-08-02 10:21:02.767505
33268	4	1-10056	33268	35	32	800	70	2012-08-02 10:21:02.767505
33273	4	1-5873	33272	32	62	478	36	2012-08-02 10:21:03.541599
33274	4	1-9704	33272	62	32	1075	10	2012-08-02 10:21:03.541599
33275	4	1-9834	33272	32	28	1044	52	2012-08-02 10:21:03.541599
33272	4	1-10058	33272	28	32	2832	84	2012-08-02 10:21:03.541599
33277	3	1-9810	33276	46	90	2651	39	2012-08-02 10:21:05.486703
33278	3	1-9991	33276	90	49	723	31	2012-08-02 10:21:05.486703
33276	3	1-10063	33276	49	46	5219	18	2012-08-02 10:21:05.486703
33280	3	1-9863	33279	30	48	1668	28	2012-08-02 10:21:05.486703
33281	3	1-6942	33279	48	49	470	31	2012-08-02 10:21:05.486703
33279	3	1-10063	33279	49	30	3681	18	2012-08-02 10:21:05.486703
33283	2	1-9845	33282	36	43	2891	83	2012-08-02 10:21:06.624907
33282	2	1-10066	33282	43	36	6562	7	2012-08-02 10:21:06.624907
33285	3	1-6770	33284	34	57	63	94	2012-08-02 10:21:06.624907
33286	3	1-9785	33284	57	43	102	83	2012-08-02 10:21:06.624907
33284	3	1-10066	33284	43	34	321	7	2012-08-02 10:21:06.624907
33288	2	1-9863	33287	30	17	4175	28	2012-08-02 10:21:08.183235
33287	2	1-10070	33287	17	30	4478	18	2012-08-02 10:21:08.183235
33290	4	1-9912	33289	29	63	1851	77	2012-08-02 10:21:08.570323
33291	4	1-9995	33289	63	75	2595	5	2012-08-02 10:21:08.570323
33292	4	1-9198	33289	75	19	1018	40	2012-08-02 10:21:08.570323
33289	4	1-10071	33289	19	29	1408	82	2012-08-02 10:21:08.570323
33294	4	1-2571	33293	98	41	2062	53	2012-08-02 10:21:08.570323
33295	4	1-9933	33293	41	17	2274	88	2012-08-02 10:21:08.570323
33296	4	1-9385	33293	17	19	3226	40	2012-08-02 10:21:08.570323
33293	4	1-10071	33293	19	98	6054	82	2012-08-02 10:21:08.570323
33298	4	1-9655	33297	89	21	1155	95	2012-08-02 10:21:08.570323
33299	4	1-9740	33297	21	84	618	48	2012-08-02 10:21:08.570323
33300	4	1-9743	33297	84	19	674	40	2012-08-02 10:21:08.570323
33297	4	1-10071	33297	19	89	1366	82	2012-08-02 10:21:08.570323
33302	2	1-9484	33301	78	51	2317	71	2012-08-02 10:21:08.960071
33301	2	1-10072	33301	51	78	2635	91	2012-08-02 10:21:08.960071
33304	4	1-7835	33303	65	16	856	37	2012-08-02 10:21:08.960071
33305	4	1-9804	33303	16	57	940	79	2012-08-02 10:21:08.960071
33306	4	1-8540	33303	57	51	586	71	2012-08-02 10:21:08.960071
33303	4	1-10072	33303	51	65	681	91	2012-08-02 10:21:08.960071
33308	4	1-9445	33307	50	76	3290	28	2012-08-02 10:21:10.540401
33309	4	1-9998	33307	76	99	2720	24	2012-08-02 10:21:10.540401
33310	4	1-8675	33307	99	2	2781	54	2012-08-02 10:21:10.540401
33307	4	1-10076	33307	2	50	3321	76	2012-08-02 10:21:10.540401
33312	2	1-7747	33311	49	67	1065	2	2012-08-02 10:21:11.711576
33311	2	1-10079	33311	67	49	628	15	2012-08-02 10:21:11.711576
33314	3	1-9574	33313	68	35	3738	25	2012-08-02 10:21:11.711576
33315	3	1-9278	33313	35	67	5771	2	2012-08-02 10:21:11.711576
33313	3	1-10079	33313	67	68	3776	15	2012-08-02 10:21:11.711576
33317	3	1-8266	33316	74	42	298	21	2012-08-02 10:21:11.711576
33318	3	1-9781	33316	42	67	1194	2	2012-08-02 10:21:11.711576
33316	3	1-10079	33316	67	74	854	15	2012-08-02 10:21:11.711576
33320	4	1-9574	33319	68	6	1397	25	2012-08-02 10:21:11.711576
33321	4	1-8445	33319	6	86	1656	70	2012-08-02 10:21:11.711576
33322	4	1-9530	33319	86	67	1747	2	2012-08-02 10:21:11.711576
33319	4	1-10079	33319	67	68	1270	15	2012-08-02 10:21:11.711576
33324	5	1-9749	33323	96	49	1316	74	2012-08-02 10:21:13.248202
33325	5	1-7486	33323	49	87	1100	51	2012-08-02 10:21:13.248202
33326	5	1-9983	33323	87	85	1397	46	2012-08-02 10:21:13.248202
33327	5	1-9746	33323	85	98	543	99	2012-08-02 10:21:13.248202
33323	5	1-10083	33323	98	96	589	26	2012-08-02 10:21:13.248202
33329	5	1-9749	33328	96	3	2437	74	2012-08-02 10:21:13.248202
33330	5	1-5693	33328	3	87	1648	51	2012-08-02 10:21:13.248202
33331	5	1-9983	33328	87	85	2206	46	2012-08-02 10:21:13.248202
33332	5	1-9746	33328	85	98	905	99	2012-08-02 10:21:13.248202
33328	5	1-10083	33328	98	96	1035	26	2012-08-02 10:21:13.248202
33334	4	1-9300	33333	42	38	2896	49	2012-08-02 10:21:13.248202
33335	4	1-9795	33333	38	10	3437	27	2012-08-02 10:21:13.248202
33336	4	1-9564	33333	10	98	1377	99	2012-08-02 10:21:13.248202
33333	4	1-10083	33333	98	42	1649	26	2012-08-02 10:21:13.248202
33338	4	1-9749	33337	96	3	1745	74	2012-08-02 10:21:13.248202
33339	4	1-9754	33337	3	10	1366	27	2012-08-02 10:21:13.248202
33340	4	1-9564	33337	10	98	561	99	2012-08-02 10:21:13.248202
33337	4	1-10083	33337	98	96	690	26	2012-08-02 10:21:13.248202
33342	5	1-9300	33341	42	75	3611	49	2012-08-02 10:21:13.248202
33343	5	1-8774	33341	75	69	1572	90	2012-08-02 10:21:13.248202
33344	5	1-9936	33341	69	85	3248	46	2012-08-02 10:21:13.248202
33345	5	1-9746	33341	85	98	1495	99	2012-08-02 10:21:13.248202
33341	5	1-10083	33341	98	42	1920	26	2012-08-02 10:21:13.248202
33347	5	1-8663	33346	17	17	352	24	2012-08-02 10:21:14.409656
33348	5	1-8655	33346	17	10	376	27	2012-08-02 10:21:14.409656
33349	5	1-9564	33346	10	98	119	99	2012-08-02 10:21:14.409656
33350	5	1-10083	33346	98	39	113	26	2012-08-02 10:21:14.409656
33346	5	1-10086	33346	39	17	707	89	2012-08-02 10:21:14.409656
33352	5	1-3321	33351	95	87	1425	51	2012-08-02 10:21:14.409656
33353	5	1-9983	33351	87	85	1610	46	2012-08-02 10:21:14.409656
33354	5	1-9746	33351	85	98	557	99	2012-08-02 10:21:14.409656
33355	5	1-10083	33351	98	39	537	26	2012-08-02 10:21:14.409656
33351	5	1-10086	33351	39	95	3419	89	2012-08-02 10:21:14.409656
33357	5	1-8408	33356	53	53	1225	37	2012-08-02 10:21:14.409656
33358	5	1-9819	33356	53	10	738	27	2012-08-02 10:21:14.409656
33359	5	1-9564	33356	10	98	272	99	2012-08-02 10:21:14.409656
33360	5	1-10083	33356	98	39	299	26	2012-08-02 10:21:14.409656
33356	5	1-10086	33356	39	53	2172	89	2012-08-02 10:21:14.409656
33362	4	1-9636	33361	93	82	78	97	2012-08-02 10:21:15.194431
33363	4	1-9337	33361	82	5	168	87	2012-08-02 10:21:15.194431
33364	4	1-10015	33361	5	74	179	96	2012-08-02 10:21:15.194431
33361	4	1-10088	33361	74	93	102	8	2012-08-02 10:21:15.194431
33366	4	1-9758	33365	46	97	2432	52	2012-08-02 10:21:16.355077
33367	4	1-7464	33365	97	46	1241	98	2012-08-02 10:21:16.355077
33368	4	1-10023	33365	46	18	1494	62	2012-08-02 10:21:16.355077
33365	4	1-10091	33365	18	46	4102	81	2012-08-02 10:21:16.355077
33370	5	1-9825	33369	31	60	1002	33	2012-08-02 10:21:16.355077
33371	5	1-8032	33369	60	82	336	24	2012-08-02 10:21:16.355077
33372	5	1-9242	33369	82	46	748	98	2012-08-02 10:21:16.355077
33373	5	1-10023	33369	46	18	901	62	2012-08-02 10:21:16.355077
33369	5	1-10091	33369	18	31	2639	81	2012-08-02 10:21:16.355077
33375	5	1-9046	33374	82	87	698	51	2012-08-02 10:21:17.538227
33376	5	1-9983	33374	87	92	989	46	2012-08-02 10:21:17.538227
33377	5	1-9591	33374	92	76	443	28	2012-08-02 10:21:17.538227
33378	5	1-9998	33374	76	90	323	24	2012-08-02 10:21:17.538227
33374	5	1-10094	33374	90	82	771	19	2012-08-02 10:21:17.538227
33380	5	1-9770	33379	50	16	3137	59	2012-08-02 10:21:17.538227
33381	5	1-8067	33379	16	89	1616	32	2012-08-02 10:21:17.538227
33382	5	1-8843	33379	89	76	1254	28	2012-08-02 10:21:17.538227
33383	5	1-9998	33379	76	90	1038	24	2012-08-02 10:21:17.538227
33379	5	1-10094	33379	90	50	2805	19	2012-08-02 10:21:17.538227
33385	4	1-9877	33384	63	16	3942	1	2012-08-02 10:21:19.107086
33386	4	1-9799	33384	16	36	3613	54	2012-08-02 10:21:19.107086
33387	4	1-9727	33384	36	12	1930	94	2012-08-02 10:21:19.107086
33384	4	1-10098	33384	12	63	8149	53	2012-08-02 10:21:19.107086
33389	4	1-8295	33388	17	62	4260	4	2012-08-02 10:21:19.86925
33390	4	1-10097	33388	62	99	2960	73	2012-08-02 10:21:19.86925
33391	4	1-9393	33388	99	25	2122	98	2012-08-02 10:21:19.86925
33388	4	1-10100	33388	25	17	5160	43	2012-08-02 10:21:19.86925
33393	2	1-4019	33392	54	78	276	69	2012-08-02 10:21:20.423155
33392	2	1-10101	33392	78	54	1823	86	2012-08-02 10:21:20.423155
33395	4	1-9985	33394	22	67	1540	42	2012-08-02 10:21:23.162067
33396	4	1-9576	33394	67	11	921	25	2012-08-02 10:21:23.162067
33397	4	1-9663	33394	11	83	437	56	2012-08-02 10:21:23.162067
33394	4	1-10108	33394	83	22	1017	75	2012-08-02 10:21:23.162067
33399	4	1-8702	33398	49	64	226	45	2012-08-02 10:21:23.162067
33400	4	1-9324	33398	64	96	375	86	2012-08-02 10:21:23.162067
33401	4	1-5119	33398	96	83	104	56	2012-08-02 10:21:23.162067
33398	4	1-10108	33398	83	49	246	75	2012-08-02 10:21:23.162067
33403	3	1-7334	33402	92	80	1017	81	2012-08-02 10:21:24.357937
33404	3	1-9581	33402	80	7	900	58	2012-08-02 10:21:24.357937
33402	3	1-10111	33402	7	92	1275	47	2012-08-02 10:21:24.357937
33406	3	1-5088	33405	4	65	1922	65	2012-08-02 10:21:24.357937
33407	3	1-9688	33405	65	7	2024	58	2012-08-02 10:21:24.357937
33405	3	1-10111	33405	7	4	2934	47	2012-08-02 10:21:24.357937
33409	4	1-4524	33408	94	98	1122	73	2012-08-02 10:21:24.357937
33410	4	1-9520	33408	98	57	1170	1	2012-08-02 10:21:24.357937
33411	4	1-3689	33408	57	7	843	58	2012-08-02 10:21:24.357937
33408	4	1-10111	33408	7	94	1594	47	2012-08-02 10:21:24.357937
33413	4	1-4524	33412	94	93	1127	73	2012-08-02 10:21:24.357937
33414	4	1-6193	33412	93	65	558	65	2012-08-02 10:21:24.357937
33415	4	1-9688	33412	65	7	793	58	2012-08-02 10:21:24.357937
33412	4	1-10111	33412	7	94	1549	47	2012-08-02 10:21:24.357937
33417	4	1-4919	33416	77	24	1762	84	2012-08-02 10:21:24.357937
33418	4	1-10045	33416	24	22	1710	6	2012-08-02 10:21:24.357937
33419	4	1-7364	33416	22	7	1005	58	2012-08-02 10:21:24.357937
33416	4	1-10111	33416	7	77	2004	47	2012-08-02 10:21:24.357937
33421	3	1-9853	33420	99	8	1190	50	2012-08-02 10:21:24.756573
33422	3	1-7200	33420	8	50	563	54	2012-08-02 10:21:24.756573
33420	3	1-10112	33420	50	99	1140	62	2012-08-02 10:21:24.756573
33424	4	1-9729	33423	68	57	682	49	2012-08-02 10:21:24.756573
33425	4	1-4883	33423	57	77	120	63	2012-08-02 10:21:24.756573
33426	4	1-9431	33423	77	50	354	54	2012-08-02 10:21:24.756573
33423	4	1-10112	33423	50	68	733	62	2012-08-02 10:21:24.756573
33428	3	1-9853	33427	99	39	3819	50	2012-08-02 10:21:24.756573
33429	3	1-8767	33427	39	50	1610	54	2012-08-02 10:21:24.756573
33427	3	1-10112	33427	50	99	3454	62	2012-08-02 10:21:24.756573
33431	3	1-9984	33430	32	90	4359	35	2012-08-02 10:21:24.756573
33432	3	1-1386	33430	90	50	1828	54	2012-08-02 10:21:24.756573
33430	3	1-10112	33430	50	32	4211	62	2012-08-02 10:21:24.756573
33434	4	1-10091	33433	18	34	1430	81	2012-08-02 10:21:25.165939
33435	4	1-2308	33433	34	23	168	32	2012-08-02 10:21:25.165939
33433	4	1-10113	33433	89	18	538	62	2012-08-02 10:21:25.165939
33436	4	1-10012	33433	23	89	251	90	2012-08-02 10:21:25.165939
33438	5	1-9984	33437	32	34	1152	35	2012-08-02 10:21:25.165939
33439	5	1-7788	33437	34	16	743	59	2012-08-02 10:21:25.165939
33440	5	1-8067	33437	16	23	354	32	2012-08-02 10:21:25.165939
33441	5	1-10012	33437	23	89	548	90	2012-08-02 10:21:25.165939
33437	5	1-10113	33437	89	32	1222	62	2012-08-02 10:21:25.165939
33443	5	1-9902	33442	64	61	602	6	2012-08-02 10:21:25.165939
33444	5	1-5225	33442	61	16	336	59	2012-08-02 10:21:25.165939
33445	5	1-8067	33442	16	23	165	32	2012-08-02 10:21:25.165939
33446	5	1-10012	33442	23	89	263	90	2012-08-02 10:21:25.165939
33442	5	1-10113	33442	89	64	604	62	2012-08-02 10:21:25.165939
33448	4	1-9509	33447	25	52	113	27	2012-08-02 10:21:25.563768
33449	4	1-9236	33447	52	100	56	69	2012-08-02 10:21:25.563768
33450	4	1-9911	33447	100	18	51	84	2012-08-02 10:21:25.563768
33447	4	1-10114	33447	18	25	202	64	2012-08-02 10:21:25.563768
33452	4	1-4502	33451	66	89	3864	88	2012-08-02 10:21:25.563768
33453	4	1-9817	33451	89	100	1396	69	2012-08-02 10:21:25.563768
33454	4	1-9911	33451	100	18	1402	84	2012-08-02 10:21:25.563768
33451	4	1-10114	33451	18	66	6086	64	2012-08-02 10:21:25.563768
33456	3	1-1511	33455	74	18	164	51	2012-08-02 10:21:25.563768
33457	3	1-9925	33455	18	18	116	84	2012-08-02 10:21:25.563768
33455	3	1-10114	33455	18	74	541	64	2012-08-02 10:21:25.563768
33459	3	1-9127	33458	66	91	120	6	2012-08-02 10:21:25.962018
33460	3	1-7374	33458	91	47	103	40	2012-08-02 10:21:25.962018
33458	3	1-10115	33458	47	66	133	43	2012-08-02 10:21:25.962018
33462	4	1-6456	33461	9	14	94	85	2012-08-02 10:21:25.962018
33463	4	1-8197	33461	14	17	52	88	2012-08-02 10:21:25.962018
33464	4	1-9385	33461	17	47	104	40	2012-08-02 10:21:25.962018
33461	4	1-10115	33461	47	9	135	43	2012-08-02 10:21:25.962018
33466	4	1-9964	33465	80	31	6610	16	2012-08-02 10:21:25.962018
33467	4	1-7190	33465	31	17	1397	88	2012-08-02 10:21:25.962018
33468	4	1-9385	33465	17	47	2786	40	2012-08-02 10:21:25.962018
33465	4	1-10115	33465	47	80	3643	43	2012-08-02 10:21:25.962018
33470	4	1-9293	33469	80	40	4171	87	2012-08-02 10:21:25.962018
33471	4	1-6599	33469	40	59	763	32	2012-08-02 10:21:25.962018
33472	4	1-8865	33469	59	47	3326	40	2012-08-02 10:21:25.962018
33469	4	1-10115	33469	47	80	4385	43	2012-08-02 10:21:25.962018
33474	4	1-6318	33473	1	99	381	59	2012-08-02 10:21:25.962018
33475	4	1-9092	33473	99	59	128	32	2012-08-02 10:21:25.962018
33476	4	1-8865	33473	59	47	570	40	2012-08-02 10:21:25.962018
33473	4	1-10115	33473	47	1	768	43	2012-08-02 10:21:25.962018
33478	4	1-10024	33477	14	21	6833	86	2012-08-02 10:21:26.747339
33479	4	1-9940	33477	21	62	1834	36	2012-08-02 10:21:26.747339
33480	4	1-10027	33477	62	100	1419	57	2012-08-02 10:21:26.747339
33477	4	1-10117	33477	100	14	3791	54	2012-08-02 10:21:26.747339
33482	4	1-10073	33481	3	30	1164	96	2012-08-02 10:21:26.747339
33483	4	1-10084	33481	30	70	1684	61	2012-08-02 10:21:26.747339
33484	4	1-7704	33481	70	100	1175	57	2012-08-02 10:21:26.747339
33481	4	1-10117	33481	100	3	3239	54	2012-08-02 10:21:26.747339
33486	4	1-10024	33485	14	92	364	86	2012-08-02 10:21:26.747339
33487	4	1-9798	33485	92	77	100	80	2012-08-02 10:21:26.747339
33488	4	1-9928	33485	77	100	69	57	2012-08-02 10:21:26.747339
33485	4	1-10117	33485	100	14	193	54	2012-08-02 10:21:26.747339
33490	4	1-8205	33489	67	81	814	91	2012-08-02 10:21:26.747339
33491	4	1-9171	33489	81	16	572	95	2012-08-02 10:21:26.747339
33492	4	1-9934	33489	16	100	778	57	2012-08-02 10:21:26.747339
33489	4	1-10117	33489	100	67	2189	54	2012-08-02 10:21:26.747339
33494	5	1-9560	33493	95	12	4430	80	2012-08-02 10:21:27.919177
33495	5	1-6035	33493	12	17	1474	28	2012-08-02 10:21:27.919177
33496	5	1-10070	33493	17	70	2162	18	2012-08-02 10:21:27.919177
33497	5	1-10009	33493	70	55	2697	47	2012-08-02 10:21:27.919177
33493	5	1-10120	33493	55	95	5998	49	2012-08-02 10:21:27.919177
33499	5	1-5862	33498	21	81	60	58	2012-08-02 10:21:27.919177
33500	5	1-8905	33498	81	17	16	28	2012-08-02 10:21:27.919177
33501	5	1-10070	33498	17	70	23	18	2012-08-02 10:21:27.919177
33502	5	1-10009	33498	70	55	29	47	2012-08-02 10:21:27.919177
33498	5	1-10120	33498	55	21	66	49	2012-08-02 10:21:27.919177
33504	3	1-10100	33503	25	1	4628	43	2012-08-02 10:21:28.69255
33505	3	1-6318	33503	1	50	1531	59	2012-08-02 10:21:28.69255
33503	3	1-10122	33503	50	25	2437	98	2012-08-02 10:21:28.69255
33507	3	1-10031	33506	34	61	2626	6	2012-08-02 10:21:28.69255
33508	3	1-5225	33506	61	50	1037	59	2012-08-02 10:21:28.69255
33506	3	1-10122	33506	50	34	1730	98	2012-08-02 10:21:28.69255
33510	3	1-10031	33509	34	64	1903	6	2012-08-02 10:21:28.69255
33511	3	1-8286	33509	64	50	418	59	2012-08-02 10:21:28.69255
33509	3	1-10122	33509	50	34	935	98	2012-08-02 10:21:28.69255
33513	4	1-10023	33512	46	95	590	62	2012-08-02 10:21:28.69255
33514	4	1-9224	33512	95	19	402	94	2012-08-02 10:21:28.69255
33515	4	1-6150	33512	19	50	221	59	2012-08-02 10:21:28.69255
33512	4	1-10122	33512	50	46	512	98	2012-08-02 10:21:28.69255
33517	4	1-10023	33516	46	22	2131	62	2012-08-02 10:21:28.69255
33518	4	1-7857	33516	22	83	1363	36	2012-08-02 10:21:28.69255
33519	4	1-7135	33516	83	50	707	59	2012-08-02 10:21:28.69255
33516	4	1-10122	33516	50	46	1742	98	2012-08-02 10:21:28.69255
33521	4	1-9240	33520	37	68	1972	71	2012-08-02 10:21:29.112493
33522	4	1-10089	33520	68	23	1344	4	2012-08-02 10:21:29.112493
33523	4	1-9536	33520	23	92	1376	94	2012-08-02 10:21:29.112493
33520	4	1-10123	33520	92	37	2779	88	2012-08-02 10:21:29.112493
33525	3	1-9887	33524	98	39	4808	36	2012-08-02 10:21:29.112493
33526	3	1-8674	33524	39	92	2301	94	2012-08-02 10:21:29.112493
33524	3	1-10123	33524	92	98	6447	88	2012-08-02 10:21:29.112493
33528	4	1-9326	33527	98	19	928	58	2012-08-02 10:21:29.49951
33529	4	1-10018	33527	19	20	788	13	2012-08-02 10:21:29.49951
33530	4	1-9051	33527	20	38	506	82	2012-08-02 10:21:29.49951
33527	4	1-10124	33527	38	98	1375	14	2012-08-02 10:21:29.49951
33532	4	1-9910	33531	48	55	426	9	2012-08-02 10:21:29.49951
33533	4	1-8537	33531	55	20	406	13	2012-08-02 10:21:29.49951
33534	4	1-9051	33531	20	38	298	82	2012-08-02 10:21:29.49951
33531	4	1-10124	33531	38	48	923	14	2012-08-02 10:21:29.49951
33536	4	1-9882	33535	94	64	444	41	2012-08-02 10:21:33.092667
33537	4	1-6945	33535	64	80	440	81	2012-08-02 10:21:33.092667
33538	4	1-9581	33535	80	12	523	58	2012-08-02 10:21:33.092667
33535	4	1-10133	33535	12	94	1293	25	2012-08-02 10:21:33.092667
33540	4	1-6031	33539	86	40	639	93	2012-08-02 10:21:33.092667
33541	4	1-8824	33539	40	65	295	65	2012-08-02 10:21:33.092667
33542	4	1-9688	33539	65	12	408	58	2012-08-02 10:21:33.092667
33539	4	1-10133	33539	12	86	1009	25	2012-08-02 10:21:33.092667
33544	4	1-2541	33543	87	40	1106	93	2012-08-02 10:21:33.092667
33545	4	1-8824	33543	40	65	533	65	2012-08-02 10:21:33.092667
33546	4	1-9688	33543	65	12	767	58	2012-08-02 10:21:33.092667
33543	4	1-10133	33543	12	87	1976	25	2012-08-02 10:21:33.092667
33548	4	1-9426	33547	17	70	1295	18	2012-08-02 10:21:34.663624
33549	4	1-10009	33547	70	99	1537	47	2012-08-02 10:21:34.663624
33550	4	1-8019	33547	99	85	1041	64	2012-08-02 10:21:34.663624
33547	4	1-10137	33547	85	17	1398	36	2012-08-02 10:21:34.663624
33552	5	1-9784	33551	73	87	2347	25	2012-08-02 10:21:34.663624
33553	5	1-2541	33551	87	10	1380	93	2012-08-02 10:21:34.663624
33554	5	1-9546	33551	10	66	786	1	2012-08-02 10:21:34.663624
33555	5	1-10119	33551	66	85	718	64	2012-08-02 10:21:34.663624
33551	5	1-10137	33551	85	73	1064	36	2012-08-02 10:21:34.663624
33557	4	1-10027	33556	62	69	934	57	2012-08-02 10:21:34.663624
33558	4	1-10054	33556	69	83	3254	70	2012-08-02 10:21:34.663624
33559	4	1-3602	33556	83	85	669	64	2012-08-02 10:21:34.663624
33556	4	1-10137	33556	85	62	994	36	2012-08-02 10:21:34.663624
33561	5	1-8383	33560	82	82	208	4	2012-08-02 10:21:35.459608
33562	5	1-9881	33560	82	50	195	76	2012-08-02 10:21:35.459608
33563	5	1-9445	33560	50	3	173	28	2012-08-02 10:21:35.459608
33564	5	1-10085	33560	3	95	171	33	2012-08-02 10:21:35.459608
33560	5	1-10139	33560	95	82	326	12	2012-08-02 10:21:35.459608
33566	3	1-3386	33565	53	82	150	97	2012-08-02 10:22:00.297096
33567	3	1-9337	33565	82	15	350	87	2012-08-02 10:22:00.297096
33565	3	1-10141	33565	15	53	402	25	2012-08-02 10:22:00.297096
33569	4	1-8172	33568	9	81	789	51	2012-08-02 10:22:01.175856
33570	4	1-10000	33568	81	69	622	90	2012-08-02 10:22:01.175856
33571	4	1-9936	33568	69	90	762	46	2012-08-02 10:22:01.175856
33568	4	1-10143	33568	90	9	3569	71	2012-08-02 10:22:01.175856
33573	4	1-9508	33572	7	46	3544	98	2012-08-02 10:22:01.959208
33574	4	1-10023	33572	46	94	2161	62	2012-08-02 10:22:01.959208
33575	4	1-9616	33572	94	48	1211	77	2012-08-02 10:22:01.959208
33572	4	1-10145	33572	48	7	4591	26	2012-08-02 10:22:01.959208
33577	4	1-9347	33576	30	46	1304	20	2012-08-02 10:22:03.89283
33578	4	1-7354	33576	46	56	339	64	2012-08-02 10:22:03.89283
33579	4	1-8657	33576	56	10	275	73	2012-08-02 10:22:03.89283
33576	4	1-10150	33576	10	30	671	12	2012-08-02 10:22:03.89283
33581	4	1-8015	33580	86	38	311	52	2012-08-02 10:22:03.89283
33582	4	1-9892	33580	38	77	1137	48	2012-08-02 10:22:03.89283
33583	4	1-8038	33580	77	10	202	73	2012-08-02 10:22:03.89283
33580	4	1-10150	33580	10	86	567	12	2012-08-02 10:22:03.89283
33585	3	1-4257	33584	67	53	5239	37	2012-08-02 10:22:04.301357
33586	3	1-9819	33584	53	16	3522	27	2012-08-02 10:22:04.301357
33584	3	1-10151	33584	16	67	4729	35	2012-08-02 10:22:04.301357
33588	4	1-9935	33587	8	24	5997	84	2012-08-02 10:22:05.075001
33589	4	1-10045	33587	24	41	4706	6	2012-08-02 10:22:05.075001
33590	4	1-8542	33587	41	8	4323	60	2012-08-02 10:22:05.075001
33587	4	1-10153	33587	8	8	3173	75	2012-08-02 10:22:05.075001
33592	4	1-9944	33591	66	96	1066	80	2012-08-02 10:22:05.483862
33593	4	1-10051	33591	96	84	1227	48	2012-08-02 10:22:05.483862
33594	4	1-9743	33591	84	91	1417	40	2012-08-02 10:22:05.483862
33591	4	1-10154	33591	91	66	2426	47	2012-08-02 10:22:05.483862
33596	3	1-9854	33595	4	42	795	21	2012-08-02 10:22:05.870393
33597	3	1-9781	33595	42	38	2437	2	2012-08-02 10:22:05.870393
33595	3	1-10155	33595	38	4	837	6	2012-08-02 10:22:05.870393
33599	4	1-9715	33598	65	95	528	78	2012-08-02 10:22:05.870393
33600	4	1-9960	33598	95	95	1006	97	2012-08-02 10:22:05.870393
33601	4	1-9747	33598	95	38	2475	2	2012-08-02 10:22:05.870393
33598	4	1-10155	33598	38	65	1056	6	2012-08-02 10:22:05.870393
33603	5	1-10133	33602	12	87	2251	25	2012-08-02 10:22:06.63264
33604	5	1-2541	33602	87	84	1063	93	2012-08-02 10:22:06.63264
33605	5	1-9441	33602	84	30	621	96	2012-08-02 10:22:06.63264
33606	5	1-10084	33602	30	32	846	61	2012-08-02 10:22:06.63264
33602	5	1-10157	33602	32	12	1036	58	2012-08-02 10:22:06.63264
33608	4	1-8865	33607	59	97	3452	40	2012-08-02 10:22:07.804143
33609	4	1-2761	33607	97	13	881	14	2012-08-02 10:22:07.804143
33610	4	1-9948	33607	13	72	504	72	2012-08-02 10:22:07.804143
33607	4	1-10160	33607	72	59	826	32	2012-08-02 10:22:07.804143
33612	4	1-10135	33611	85	83	6979	36	2012-08-02 10:22:07.804143
33613	4	1-7135	33611	83	48	3291	59	2012-08-02 10:22:07.804143
33614	4	1-9358	33611	48	72	3216	72	2012-08-02 10:22:07.804143
33611	4	1-10160	33611	72	85	5275	32	2012-08-02 10:22:07.804143
33616	2	1-3576	33615	22	67	657	53	2012-08-02 10:22:08.964194
33615	2	1-10163	33615	67	22	9695	25	2012-08-02 10:22:08.964194
33618	4	1-10130	33617	27	36	3062	100	2012-08-02 10:22:09.351004
33619	4	1-9849	33617	36	34	2886	38	2012-08-02 10:22:09.351004
33620	4	1-9294	33617	34	77	2753	77	2012-08-02 10:22:09.351004
33617	4	1-10164	33617	77	27	4170	14	2012-08-02 10:22:09.351004
33622	2	1-6938	33621	90	75	2538	88	2012-08-02 10:22:09.737764
33621	2	1-10165	33621	75	90	8340	61	2012-08-02 10:22:09.737764
33624	3	1-8467	33623	36	14	377	85	2012-08-02 10:22:09.737764
33625	3	1-8197	33623	14	75	125	88	2012-08-02 10:22:09.737764
33623	3	1-10165	33623	75	36	576	61	2012-08-02 10:22:09.737764
33627	4	1-7756	33626	57	68	232	71	2012-08-02 10:22:10.147072
33628	4	1-10089	33626	68	22	180	4	2012-08-02 10:22:10.147072
33629	4	1-9007	33626	22	29	51	98	2012-08-02 10:22:10.147072
33626	4	1-10166	33626	29	57	402	89	2012-08-02 10:22:10.147072
33631	4	1-5043	33630	24	69	899	67	2012-08-02 10:22:10.147072
33632	4	1-5544	33630	69	7	222	26	2012-08-02 10:22:10.147072
33633	4	1-9508	33630	7	29	292	98	2012-08-02 10:22:10.147072
33630	4	1-10166	33630	29	24	2527	89	2012-08-02 10:22:10.147072
33635	4	1-8232	33634	62	23	197	32	2012-08-02 10:22:11.737977
33636	4	1-10012	33634	23	89	260	90	2012-08-02 10:22:11.737977
33637	4	1-10113	33634	89	24	493	62	2012-08-02 10:22:11.737977
33634	4	1-10170	33634	24	62	618	4	2012-08-02 10:22:11.737977
33639	5	1-9606	33638	91	32	155	82	2012-08-02 10:22:11.737977
33640	5	1-8370	33638	32	81	91	51	2012-08-02 10:22:11.737977
33641	5	1-10000	33638	81	89	110	90	2012-08-02 10:22:11.737977
33642	5	1-10113	33638	89	24	226	62	2012-08-02 10:22:11.737977
33638	5	1-10170	33638	24	91	307	4	2012-08-02 10:22:11.737977
33644	4	1-10097	33643	62	99	1617	73	2012-08-02 10:22:11.737977
33645	4	1-9393	33643	99	46	1286	98	2012-08-02 10:22:11.737977
33646	4	1-10023	33643	46	24	1444	62	2012-08-02 10:22:11.737977
33643	4	1-10170	33643	24	62	2098	4	2012-08-02 10:22:11.737977
33648	4	1-7735	33647	7	97	1731	59	2012-08-02 10:22:11.737977
33649	4	1-9462	33647	97	89	903	90	2012-08-02 10:22:11.737977
33650	4	1-10113	33647	89	24	2002	62	2012-08-02 10:22:11.737977
33647	4	1-10170	33647	24	7	2928	4	2012-08-02 10:22:11.737977
33651	1	1-9793	\N	4	4	1408	62	2012-08-02 10:22:11.737977
33653	4	1-10104	33652	25	86	2503	61	2012-08-02 10:22:13.306613
33654	4	1-6531	33652	86	10	1270	73	2012-08-02 10:22:13.306613
33655	4	1-10150	33652	10	78	3869	12	2012-08-02 10:22:13.306613
33652	4	1-10174	33652	78	25	4007	77	2012-08-02 10:22:13.306613
33657	4	1-8933	33656	21	38	2926	5	2012-08-02 10:22:14.080368
33658	4	1-7797	33656	38	13	1754	14	2012-08-02 10:22:14.080368
33659	4	1-9948	33656	13	61	922	72	2012-08-02 10:22:14.080368
33656	4	1-10176	33656	61	21	2397	23	2012-08-02 10:22:14.080368
33661	4	1-7882	33660	57	100	1182	69	2012-08-02 10:22:14.080368
33662	4	1-9911	33660	100	47	1169	84	2012-08-02 10:22:14.080368
33663	4	1-9245	33660	47	61	938	72	2012-08-02 10:22:14.080368
33660	4	1-10176	33660	61	57	2467	23	2012-08-02 10:22:14.080368
33665	4	1-9971	33664	58	93	4208	8	2012-08-02 10:22:14.080368
33666	4	1-9636	33664	93	78	3090	97	2012-08-02 10:22:14.080368
33667	4	1-6471	33664	78	61	1849	72	2012-08-02 10:22:14.080368
33664	4	1-10176	33664	61	58	5130	23	2012-08-02 10:22:14.080368
33669	4	1-10153	33668	8	8	693	75	2012-08-02 10:22:14.87607
33670	4	1-9935	33668	8	57	882	84	2012-08-02 10:22:14.87607
33671	4	1-7851	33668	57	14	429	2	2012-08-02 10:22:14.87607
33668	4	1-10178	33668	14	8	1402	60	2012-08-02 10:22:14.87607
33673	2	1-6833	33672	58	14	2047	2	2012-08-02 10:22:14.87607
33672	2	1-10178	33672	14	58	7918	60	2012-08-02 10:22:14.87607
33675	3	1-2821	33674	49	90	452	44	2012-08-02 10:22:19.140153
33676	3	1-9369	33674	90	81	137	25	2012-08-02 10:22:19.140153
33674	3	1-10189	33674	81	49	5932	5	2012-08-02 10:22:19.140153
33678	4	1-10147	33677	39	20	2389	26	2012-08-02 10:22:19.526977
33679	4	1-8747	33677	20	11	1914	25	2012-08-02 10:22:19.526977
33680	4	1-9663	33677	11	56	623	56	2012-08-02 10:22:19.526977
33677	4	1-10190	33677	56	39	4931	92	2012-08-02 10:22:19.526977
33682	2	1-10137	33681	85	28	4055	36	2012-08-02 10:22:21.449132
33681	2	1-10195	33681	28	85	7185	64	2012-08-02 10:22:21.449132
33684	4	1-9473	33683	92	81	3626	91	2012-08-02 10:22:21.835923
33685	4	1-7710	33683	81	29	1888	50	2012-08-02 10:22:21.835923
33686	4	1-10180	33683	29	86	1292	9	2012-08-02 10:22:21.835923
33683	4	1-10196	33683	86	92	6905	73	2012-08-02 10:22:21.835923
33688	4	1-9717	33687	49	2	356	46	2012-08-02 10:22:21.835923
33689	4	1-9175	33687	2	38	64	11	2012-08-02 10:22:21.835923
33690	4	1-9528	33687	38	86	59	9	2012-08-02 10:22:21.835923
33687	4	1-10196	33687	86	49	356	73	2012-08-02 10:22:21.835923
33692	4	1-9827	33691	3	8	1533	60	2012-08-02 10:22:22.2226
33693	4	1-10153	33691	8	49	1243	75	2012-08-02 10:22:22.2226
33694	4	1-8702	33691	49	29	1165	45	2012-08-02 10:22:22.2226
33691	4	1-10197	33691	29	3	1814	100	2012-08-02 10:22:22.2226
33696	3	1-9849	33695	36	67	119	38	2012-08-02 10:22:22.2226
33697	3	1-8627	33695	67	29	71	45	2012-08-02 10:22:22.2226
33695	3	1-10197	33695	29	36	113	100	2012-08-02 10:22:22.2226
33699	3	1-7806	33698	88	37	368	68	2012-08-02 10:22:23.382628
33700	3	1-9394	33698	37	37	199	78	2012-08-02 10:22:23.382628
33698	3	1-10200	33698	37	88	4844	12	2012-08-02 10:22:23.382628
33702	4	1-10201	33701	82	12	3041	11	2012-08-02 10:22:25.339901
33703	4	1-9919	33701	12	96	2388	30	2012-08-02 10:22:25.339901
33704	4	1-9166	33701	96	38	2798	4	2012-08-02 10:22:25.339901
33701	4	1-10205	33701	38	82	2270	74	2012-08-02 10:22:25.339901
33706	4	1-9185	33705	61	46	335	98	2012-08-02 10:22:25.339901
33707	4	1-10023	33705	46	24	385	62	2012-08-02 10:22:25.339901
33708	4	1-10170	33705	24	38	572	4	2012-08-02 10:22:25.339901
33705	4	1-10205	33705	38	61	473	74	2012-08-02 10:22:25.339901
33710	4	1-9095	33709	71	37	615	88	2012-08-02 10:22:25.339901
33711	4	1-9240	33709	37	68	636	71	2012-08-02 10:22:25.339901
33712	4	1-10089	33709	68	38	632	4	2012-08-02 10:22:25.339901
33709	4	1-10205	33709	38	71	557	74	2012-08-02 10:22:25.339901
33714	4	1-8865	33713	59	35	2205	40	2012-08-02 10:22:26.533431
33715	4	1-9231	33713	35	64	211	64	2012-08-02 10:22:26.533431
33716	4	1-8768	33713	64	4	217	53	2012-08-02 10:22:26.533431
33713	4	1-10208	33713	4	59	574	32	2012-08-02 10:22:26.533431
33718	4	1-9496	33717	48	41	925	66	2012-08-02 10:22:28.842044
33719	4	1-7671	33717	41	58	460	71	2012-08-02 10:22:28.842044
33720	4	1-10127	33717	58	25	379	7	2012-08-02 10:22:28.842044
33717	4	1-10214	33717	25	48	877	86	2012-08-02 10:22:28.842044
33722	5	1-1235	33721	55	79	2501	15	2012-08-02 10:22:28.842044
33723	5	1-9812	33721	79	74	2222	35	2012-08-02 10:22:28.842044
33724	5	1-7521	33721	74	14	1241	82	2012-08-02 10:22:28.842044
33725	5	1-9896	33721	14	25	1748	7	2012-08-02 10:22:28.842044
33721	5	1-10214	33721	25	55	4642	86	2012-08-02 10:22:28.842044
33727	4	1-10193	33726	81	79	5972	15	2012-08-02 10:22:29.637928
33728	4	1-9812	33726	79	12	5163	35	2012-08-02 10:22:29.637928
33729	4	1-7052	33726	12	94	2446	38	2012-08-02 10:22:29.637928
33726	4	1-10216	33726	94	81	5460	22	2012-08-02 10:22:29.637928
33731	4	1-6437	33730	7	84	11	54	2012-08-02 10:22:29.637928
33732	4	1-7690	33730	84	100	5	79	2012-08-02 10:22:29.637928
33733	4	1-10046	33730	100	94	4	38	2012-08-02 10:22:29.637928
33730	4	1-10216	33730	94	7	11	22	2012-08-02 10:22:29.637928
33735	4	1-9347	33734	30	7	1098	20	2012-08-02 10:22:30.841996
33736	4	1-9029	33734	7	43	231	34	2012-08-02 10:22:30.841996
33737	4	1-10080	33734	43	9	70	41	2012-08-02 10:22:30.841996
33734	4	1-10219	33734	9	30	1180	12	2012-08-02 10:22:30.841996
33739	4	1-8015	33738	86	29	273	52	2012-08-02 10:22:30.841996
33740	4	1-6167	33738	29	88	275	2	2012-08-02 10:22:30.841996
33741	4	1-9706	33738	88	9	26	41	2012-08-02 10:22:30.841996
33738	4	1-10219	33738	9	86	717	12	2012-08-02 10:22:30.841996
33743	2	1-5690	33742	37	58	3705	70	2012-08-02 10:22:32.799622
33742	2	1-10224	33742	58	37	7569	64	2012-08-02 10:22:32.799622
33745	4	1-8768	33744	64	38	1774	53	2012-08-02 10:22:32.799622
33746	4	1-10095	33744	38	23	1011	37	2012-08-02 10:22:32.799622
33747	4	1-10029	33744	23	58	1062	70	2012-08-02 10:22:32.799622
33744	4	1-10224	33744	58	64	2271	64	2012-08-02 10:22:32.799622
33749	3	1-9887	33748	98	26	2289	36	2012-08-02 10:22:33.583952
33750	3	1-9050	33748	26	28	1980	77	2012-08-02 10:22:33.583952
33748	3	1-10226	33748	28	98	3576	88	2012-08-02 10:22:33.583952
33752	4	1-9240	33751	37	80	505	71	2012-08-02 10:22:33.583952
33753	4	1-6392	33751	80	34	268	38	2012-08-02 10:22:33.583952
33754	4	1-9294	33751	34	28	256	77	2012-08-02 10:22:33.583952
33751	4	1-10226	33751	28	37	526	88	2012-08-02 10:22:33.583952
33756	4	1-9878	33755	16	100	2207	79	2012-08-02 10:22:33.583952
33757	4	1-10046	33755	100	34	1840	38	2012-08-02 10:22:33.583952
33758	4	1-9294	33755	34	28	1792	77	2012-08-02 10:22:33.583952
33755	4	1-10226	33755	28	16	3750	88	2012-08-02 10:22:33.583952
33760	3	1-9834	33759	32	29	70	52	2012-08-02 10:22:33.970848
33761	3	1-6167	33759	29	42	98	2	2012-08-02 10:22:33.970848
33759	3	1-10227	33759	42	32	82	10	2012-08-02 10:22:33.970848
33763	4	1-9998	33762	76	90	163	24	2012-08-02 10:22:34.346475
33764	4	1-10094	33762	90	78	373	19	2012-08-02 10:22:34.346475
33765	4	1-7716	33762	78	38	307	68	2012-08-02 10:22:34.346475
33762	4	1-10228	33762	38	76	232	28	2012-08-02 10:22:34.346475
33767	4	1-9998	33766	76	90	848	24	2012-08-02 10:22:34.346475
33768	4	1-10094	33766	90	22	2291	19	2012-08-02 10:22:34.346475
33769	4	1-295	33766	22	38	1149	68	2012-08-02 10:22:34.346475
33766	4	1-10228	33766	38	76	1025	28	2012-08-02 10:22:34.346475
33771	4	1-10103	33770	77	53	1085	83	2012-08-02 10:22:35.937344
33772	4	1-9824	33770	53	84	990	48	2012-08-02 10:22:35.937344
33773	4	1-9743	33770	84	18	1025	40	2012-08-02 10:22:35.937344
33770	4	1-10232	33770	18	77	1441	29	2012-08-02 10:22:35.937344
33775	4	1-4919	33774	77	79	1467	84	2012-08-02 10:22:36.700018
33776	4	1-8264	33774	79	47	773	45	2012-08-02 10:22:36.700018
33777	4	1-10233	33774	47	10	913	66	2012-08-02 10:22:36.700018
33774	4	1-10234	33774	10	77	1929	47	2012-08-02 10:22:36.700018
33779	5	1-6560	33778	89	93	644	100	2012-08-02 10:22:36.700018
33780	5	1-10159	33778	93	46	660	52	2012-08-02 10:22:36.700018
33781	5	1-8025	33778	46	47	314	45	2012-08-02 10:22:36.700018
33782	5	1-10233	33778	47	10	431	66	2012-08-02 10:22:36.700018
33778	5	1-10234	33778	10	89	1058	47	2012-08-02 10:22:36.700018
33784	4	1-6560	33783	89	64	1669	100	2012-08-02 10:22:36.700018
33785	4	1-8604	33783	64	47	749	45	2012-08-02 10:22:36.700018
33786	4	1-10233	33783	47	10	1057	66	2012-08-02 10:22:36.700018
33783	4	1-10234	33783	10	89	2667	47	2012-08-02 10:22:36.700018
33788	4	1-9699	33787	56	4	2656	53	2012-08-02 10:22:37.495499
33789	4	1-10208	33787	4	48	6629	32	2012-08-02 10:22:37.495499
33790	4	1-8243	33787	48	88	4052	25	2012-08-02 10:22:37.495499
33787	4	1-10236	33787	88	56	6558	31	2012-08-02 10:22:37.495499
33792	4	1-9965	33791	89	10	3300	17	2012-08-02 10:22:37.881983
33793	4	1-10186	33791	10	94	1044	80	2012-08-02 10:22:37.881983
33794	4	1-9955	33791	94	6	772	23	2012-08-02 10:22:37.881983
33791	4	1-10237	33791	6	89	6319	44	2012-08-02 10:22:37.881983
33796	4	1-6574	33795	85	83	3117	56	2012-08-02 10:22:38.666503
33797	4	1-10108	33795	83	22	7048	75	2012-08-02 10:22:38.666503
33798	4	1-8603	33795	22	17	4770	78	2012-08-02 10:22:38.666503
33795	4	1-10239	33795	17	85	4066	24	2012-08-02 10:22:38.666503
33800	2	1-9469	33799	84	43	5340	10	2012-08-02 10:25:16.187794
33799	2	1-10246	33799	43	84	3992	18	2012-08-02 10:25:16.187794
33802	4	1-8628	33801	54	14	1018	14	2012-08-02 10:25:16.187794
33803	4	1-8853	33801	14	42	1156	2	2012-08-02 10:25:16.187794
33804	4	1-10227	33801	42	43	1113	10	2012-08-02 10:25:16.187794
33801	4	1-10246	33801	43	54	834	18	2012-08-02 10:25:16.187794
33806	3	1-9827	33805	3	83	371	60	2012-08-02 10:25:17.385257
33807	3	1-8055	33805	83	32	159	95	2012-08-02 10:25:17.385257
33805	3	1-10249	33805	32	3	729	100	2012-08-02 10:25:17.385257
33809	4	1-8604	33808	64	47	1772	45	2012-08-02 10:25:17.385257
33810	4	1-10233	33808	47	38	1880	66	2012-08-02 10:25:17.385257
33811	4	1-8249	33808	38	32	905	95	2012-08-02 10:25:17.385257
33808	4	1-10249	33808	32	64	5256	100	2012-08-02 10:25:17.385257
33813	4	1-5020	33812	95	39	416	10	2012-08-02 10:25:17.385257
33814	4	1-9633	33812	39	83	371	60	2012-08-02 10:25:17.385257
33815	4	1-8055	33812	83	32	221	95	2012-08-02 10:25:17.385257
33812	4	1-10249	33812	32	95	1411	100	2012-08-02 10:25:17.385257
33817	4	1-10151	33816	16	47	2970	35	2012-08-02 10:25:18.173078
33818	4	1-5188	33816	47	39	2222	92	2012-08-02 10:25:18.173078
33819	4	1-10147	33816	39	18	1409	26	2012-08-02 10:25:18.173078
33816	4	1-10251	33816	18	16	2574	27	2012-08-02 10:25:18.173078
33821	3	1-9564	33820	10	98	613	99	2012-08-02 10:25:18.173078
33822	3	1-10083	33820	98	18	769	26	2012-08-02 10:25:18.173078
33820	3	1-10251	33820	18	10	1463	27	2012-08-02 10:25:18.173078
33824	4	1-10243	33823	5	88	1265	12	2012-08-02 10:25:18.572659
33825	4	1-9602	33823	88	68	767	15	2012-08-02 10:25:18.572659
33826	4	1-9574	33823	68	60	616	25	2012-08-02 10:25:18.572659
33823	4	1-10252	33823	60	5	166	63	2012-08-02 10:25:18.572659
33828	4	1-9431	33827	77	24	389	54	2012-08-02 10:25:18.960819
33829	4	1-9726	33827	24	83	82	1	2012-08-02 10:25:18.960819
33830	4	1-9842	33827	83	94	95	20	2012-08-02 10:25:18.960819
33827	4	1-10253	33827	94	77	162	63	2012-08-02 10:25:18.960819
33832	4	1-9138	33831	82	82	4048	55	2012-08-02 10:25:18.960819
33833	4	1-8165	33831	82	86	3422	16	2012-08-02 10:25:18.960819
33834	4	1-9921	33831	86	94	2029	20	2012-08-02 10:25:18.960819
33831	4	1-10253	33831	94	82	3628	63	2012-08-02 10:25:18.960819
33836	4	1-9880	33835	38	62	4337	18	2012-08-02 10:25:18.960819
33837	4	1-10199	33835	62	86	4559	16	2012-08-02 10:25:18.960819
33838	4	1-9921	33835	86	94	2713	20	2012-08-02 10:25:18.960819
33835	4	1-10253	33835	94	38	4868	63	2012-08-02 10:25:18.960819
33840	4	1-9880	33839	38	70	259	18	2012-08-02 10:25:18.960819
33841	4	1-10009	33839	70	10	264	47	2012-08-02 10:25:18.960819
33842	4	1-7430	33839	10	94	151	20	2012-08-02 10:25:18.960819
33839	4	1-10253	33839	94	38	280	63	2012-08-02 10:25:18.960819
33844	3	1-10217	33843	49	89	3666	88	2012-08-02 10:25:19.375065
33845	3	1-9817	33843	89	17	787	69	2012-08-02 10:25:19.375065
33843	3	1-10254	33843	17	49	4675	6	2012-08-02 10:25:19.375065
33847	3	1-10173	33846	27	56	1680	55	2012-08-02 10:25:20.13785
33848	3	1-8980	33846	56	53	2362	5	2012-08-02 10:25:20.13785
33846	3	1-10256	33846	53	27	3535	10	2012-08-02 10:25:20.13785
33850	4	1-9426	33849	17	1	3149	18	2012-08-02 10:25:21.309698
33851	4	1-10128	33849	1	43	2979	34	2012-08-02 10:25:21.309698
33849	4	1-10259	33849	95	17	3892	36	2012-08-02 10:25:21.309698
33852	4	1-10080	33849	43	95	2002	41	2012-08-02 10:25:21.309698
33854	4	1-5084	33853	100	62	670	56	2012-08-02 10:25:21.309698
33855	4	1-9879	33853	62	50	1015	50	2012-08-02 10:25:21.309698
33856	4	1-9835	33853	50	95	1226	41	2012-08-02 10:25:21.309698
33853	4	1-10259	33853	95	100	2569	36	2012-08-02 10:25:21.309698
33858	2	1-9242	33857	82	37	368	98	2012-08-02 10:25:22.106318
33857	2	1-10261	33857	37	82	233	24	2012-08-02 10:25:22.106318
33860	3	1-10094	33859	90	87	3033	19	2012-08-02 10:25:22.106318
33861	3	1-7945	33859	87	37	2214	98	2012-08-02 10:25:22.106318
33859	3	1-10261	33859	37	90	1507	24	2012-08-02 10:25:22.106318
33863	4	1-10147	33862	39	18	535	26	2012-08-02 10:25:24.06309
33864	4	1-10251	33862	18	54	1135	27	2012-08-02 10:25:24.06309
33865	4	1-8449	33862	54	98	415	76	2012-08-02 10:25:24.06309
33862	4	1-10266	33862	98	39	726	92	2012-08-02 10:25:24.06309
33867	3	1-10139	33866	95	88	5168	12	2012-08-02 10:25:24.858696
33868	3	1-9602	33866	88	32	3434	15	2012-08-02 10:25:24.858696
33866	3	1-10268	33866	32	95	3033	33	2012-08-02 10:25:24.858696
33870	3	1-8220	33869	99	81	393	22	2012-08-02 10:25:24.858696
33871	3	1-10193	33869	81	32	439	15	2012-08-02 10:25:24.858696
33869	3	1-10268	33869	32	99	412	33	2012-08-02 10:25:24.858696
33873	4	1-8174	33872	5	12	4215	74	2012-08-02 10:25:24.858696
33874	4	1-9259	33872	12	66	5137	53	2012-08-02 10:25:24.858696
33875	4	1-10074	33872	66	32	5172	15	2012-08-02 10:25:24.858696
33872	4	1-10268	33872	32	5	5712	33	2012-08-02 10:25:24.858696
33877	4	1-9841	33876	57	16	365	91	2012-08-02 10:25:24.858696
33878	4	1-9826	33876	16	22	358	36	2012-08-02 10:25:24.858696
33879	4	1-9572	33876	22	32	217	15	2012-08-02 10:25:24.858696
33876	4	1-10268	33876	32	57	240	33	2012-08-02 10:25:24.858696
33881	3	1-10257	33880	12	30	1269	22	2012-08-02 10:25:26.053458
33882	3	1-8948	33880	30	52	929	59	2012-08-02 10:25:26.053458
33880	3	1-10271	33880	52	12	2202	41	2012-08-02 10:25:26.053458
33884	4	1-8053	33883	66	27	1662	10	2012-08-02 10:25:26.053458
33885	4	1-10173	33883	27	78	963	55	2012-08-02 10:25:26.053458
33886	4	1-9472	33883	78	52	757	59	2012-08-02 10:25:26.053458
33883	4	1-10271	33883	52	66	1973	41	2012-08-02 10:25:26.053458
33888	4	1-4364	33887	33	4	884	53	2012-08-02 10:25:26.053458
33889	4	1-10208	33887	4	22	2589	32	2012-08-02 10:25:26.053458
33890	4	1-8724	33887	22	52	875	59	2012-08-02 10:25:26.053458
33887	4	1-10271	33887	52	33	2364	41	2012-08-02 10:25:26.053458
33892	4	1-9875	33891	75	71	1242	100	2012-08-02 10:25:26.053458
33893	4	1-9764	33891	71	29	1108	20	2012-08-02 10:25:26.053458
33894	4	1-8931	33891	29	52	401	59	2012-08-02 10:25:26.053458
33891	4	1-10271	33891	52	75	1141	41	2012-08-02 10:25:26.053458
33896	3	1-9779	33895	94	52	3385	37	2012-08-02 10:25:26.466128
33897	3	1-10269	33895	52	31	3290	95	2012-08-02 10:25:26.466128
33895	3	1-10272	33895	31	94	3240	87	2012-08-02 10:25:26.466128
33899	4	1-5668	33898	18	43	4	10	2012-08-02 10:25:27.22918
33900	4	1-10246	33898	43	70	2	18	2012-08-02 10:25:27.22918
33901	4	1-10009	33898	70	81	2	47	2012-08-02 10:25:27.22918
33898	4	1-10274	33898	81	18	8	65	2012-08-02 10:25:27.22918
33903	4	1-3999	33902	67	26	289	61	2012-08-02 10:25:27.22918
33904	4	1-9973	33902	26	89	479	14	2012-08-02 10:25:27.22918
33905	4	1-10021	33902	89	81	211	47	2012-08-02 10:25:27.22918
33902	4	1-10274	33902	81	67	1260	65	2012-08-02 10:25:27.22918
33907	5	1-4530	33906	26	22	270	62	2012-08-02 10:25:27.22918
33908	5	1-7857	33906	22	17	147	36	2012-08-02 10:25:27.22918
33909	5	1-9426	33906	17	70	123	18	2012-08-02 10:25:27.22918
33910	5	1-10009	33906	70	81	131	47	2012-08-02 10:25:27.22918
33906	5	1-10274	33906	81	26	858	65	2012-08-02 10:25:27.22918
33912	5	1-1986	33911	84	18	1125	31	2012-08-02 10:25:27.22918
33913	5	1-7888	33911	18	17	378	28	2012-08-02 10:25:27.22918
33914	5	1-10070	33911	17	70	497	18	2012-08-02 10:25:27.22918
33915	5	1-10009	33911	70	81	555	47	2012-08-02 10:25:27.22918
33911	5	1-10274	33911	81	84	3805	65	2012-08-02 10:25:27.22918
33917	5	1-9523	33916	20	85	1177	11	2012-08-02 10:25:27.628097
33918	5	1-9332	33916	85	79	830	3	2012-08-02 10:25:27.628097
33919	5	1-10132	33916	79	89	753	90	2012-08-02 10:25:27.628097
33920	5	1-10113	33916	89	67	1412	62	2012-08-02 10:25:27.628097
33916	5	1-10275	33916	67	20	2372	1	2012-08-02 10:25:27.628097
33922	4	1-9826	33921	16	18	1989	36	2012-08-02 10:25:28.799998
33923	4	1-9525	33921	18	95	1420	78	2012-08-02 10:25:28.799998
33924	4	1-9960	33921	95	61	2725	97	2012-08-02 10:25:28.799998
33921	4	1-10278	33921	61	16	2051	91	2012-08-02 10:25:28.799998
33926	4	1-7835	33925	65	53	251	37	2012-08-02 10:25:28.799998
33927	4	1-9819	33925	53	68	172	27	2012-08-02 10:25:28.799998
33928	4	1-10277	33925	68	61	243	97	2012-08-02 10:25:28.799998
33925	4	1-10278	33925	61	65	191	91	2012-08-02 10:25:28.799998
33930	5	1-9871	33929	39	87	941	80	2012-08-02 10:25:29.198472
33931	5	1-10183	33929	87	92	584	46	2012-08-02 10:25:29.198472
33932	5	1-9591	33929	92	3	279	28	2012-08-02 10:25:29.198472
33933	5	1-10085	33929	3	78	291	33	2012-08-02 10:25:29.198472
33929	5	1-10279	33929	78	39	405	39	2012-08-02 10:25:29.198472
33935	4	1-9871	33934	39	12	32	80	2012-08-02 10:25:29.198472
33936	4	1-6035	33934	12	3	10	28	2012-08-02 10:25:29.198472
33937	4	1-10085	33934	3	78	10	33	2012-08-02 10:25:29.198472
33934	4	1-10279	33934	78	39	14	39	2012-08-02 10:25:29.198472
33939	4	1-9871	33938	39	96	3502	80	2012-08-02 10:25:29.198472
33940	4	1-10051	33938	96	84	5201	48	2012-08-02 10:25:29.198472
33941	4	1-10265	33938	84	78	1007	33	2012-08-02 10:25:29.198472
33938	4	1-10279	33938	78	39	1454	39	2012-08-02 10:25:29.198472
33943	4	1-9508	33942	7	57	306	98	2012-08-02 10:25:30.382102
33944	4	1-6911	33942	57	96	240	80	2012-08-02 10:25:30.382102
33945	4	1-10051	33942	96	93	306	48	2012-08-02 10:25:30.382102
33942	4	1-10282	33942	93	7	234	26	2012-08-02 10:25:30.382102
33947	3	1-6790	33946	34	96	334	80	2012-08-02 10:25:30.382102
33948	3	1-10051	33946	96	93	437	48	2012-08-02 10:25:30.382102
33946	3	1-10282	33946	93	34	343	26	2012-08-02 10:25:30.382102
33950	3	1-10213	33949	20	55	3782	56	2012-08-02 10:25:31.167479
33951	3	1-10110	33949	55	1	6954	17	2012-08-02 10:25:31.167479
33949	3	1-10284	33949	1	20	7928	99	2012-08-02 10:25:31.167479
33953	4	1-5863	33952	72	80	194	90	2012-08-02 10:25:31.167479
33954	4	1-10036	33952	80	91	358	77	2012-08-02 10:25:31.167479
33955	4	1-8088	33952	91	1	284	17	2012-08-02 10:25:31.167479
33952	4	1-10284	33952	1	72	386	99	2012-08-02 10:25:31.167479
33957	4	1-9445	33956	50	97	613	28	2012-08-02 10:25:31.970379
33958	4	1-9401	33956	97	37	291	71	2012-08-02 10:25:31.970379
33959	4	1-10258	33956	37	62	242	51	2012-08-02 10:25:31.970379
33956	4	1-10286	33956	62	50	905	76	2012-08-02 10:25:31.970379
33961	4	1-10266	33960	98	33	2048	92	2012-08-02 10:25:31.970379
33962	4	1-9952	33960	33	95	971	89	2012-08-02 10:25:31.970379
33963	4	1-3321	33960	95	62	416	51	2012-08-02 10:25:31.970379
33960	4	1-10286	33960	62	98	1650	76	2012-08-02 10:25:31.970379
33965	4	1-10136	33964	97	21	2679	26	2012-08-02 10:25:31.970379
33966	4	1-8624	33964	21	37	1292	71	2012-08-02 10:25:31.970379
33967	4	1-10258	33964	37	62	1193	51	2012-08-02 10:25:31.970379
33964	4	1-10286	33964	62	97	4977	76	2012-08-02 10:25:31.970379
33969	4	1-8540	33968	57	68	1751	71	2012-08-02 10:25:32.368874
33970	4	1-10089	33968	68	62	1407	4	2012-08-02 10:25:32.368874
33971	4	1-10097	33968	62	81	955	73	2012-08-02 10:25:32.368874
33968	4	1-10287	33968	81	57	3478	79	2012-08-02 10:25:32.368874
33973	4	1-10089	33972	68	100	3336	4	2012-08-02 10:25:32.766788
33974	4	1-10057	33972	100	24	2791	100	2012-08-02 10:25:32.766788
33975	4	1-10240	33972	24	22	5890	42	2012-08-02 10:25:32.766788
33972	4	1-10288	33972	22	68	4222	71	2012-08-02 10:25:32.766788
33977	3	1-7140	33976	56	24	230	100	2012-08-02 10:25:32.766788
33978	3	1-10240	33976	24	22	579	42	2012-08-02 10:25:32.766788
33976	3	1-10288	33976	22	56	496	71	2012-08-02 10:25:32.766788
33980	4	1-9360	33979	47	94	1417	31	2012-08-02 10:25:34.402254
33981	4	1-9716	33979	94	96	1738	80	2012-08-02 10:25:34.402254
33982	4	1-10051	33979	96	85	2310	48	2012-08-02 10:25:34.402254
33979	4	1-10292	33979	85	47	2503	16	2012-08-02 10:25:34.402254
33984	2	1-7365	33983	40	22	4461	69	2012-08-02 10:25:35.5733
33983	2	1-10295	33983	22	40	6600	26	2012-08-02 10:25:35.5733
33986	4	1-10000	33985	81	80	444	90	2012-08-02 10:25:35.960041
33987	4	1-10036	33985	80	65	502	77	2012-08-02 10:25:35.960041
33988	4	1-9286	33985	65	97	275	42	2012-08-02 10:25:35.960041
33985	4	1-10296	33985	97	81	626	51	2012-08-02 10:25:35.960041
33990	4	1-9159	33989	3	93	729	56	2012-08-02 10:25:35.960041
33991	4	1-9512	33989	93	24	902	100	2012-08-02 10:25:35.960041
33992	4	1-10240	33989	24	97	1563	42	2012-08-02 10:25:35.960041
33989	4	1-10296	33989	97	3	4627	51	2012-08-02 10:25:35.960041
33994	4	1-10000	33993	81	35	1138	90	2012-08-02 10:25:35.960041
33995	4	1-9957	33993	35	24	217	100	2012-08-02 10:25:35.960041
33996	4	1-10240	33993	24	97	390	42	2012-08-02 10:25:35.960041
33993	4	1-10296	33993	97	81	1193	51	2012-08-02 10:25:35.960041
33998	4	1-9455	33997	35	56	3970	22	2012-08-02 10:25:37.087292
33999	4	1-8047	33997	56	5	5552	92	2012-08-02 10:25:37.087292
34000	4	1-10291	33997	5	80	2737	43	2012-08-02 10:25:37.087292
33997	4	1-10299	33997	80	35	5091	38	2012-08-02 10:25:37.087292
34002	4	1-9483	34001	71	47	1140	23	2012-08-02 10:25:38.667486
34003	4	1-9608	34001	47	37	863	68	2012-08-02 10:25:38.667486
34004	4	1-9394	34001	37	81	659	78	2012-08-02 10:25:38.667486
34001	4	1-10303	34001	81	71	842	96	2012-08-02 10:25:38.667486
34006	3	1-7563	34005	60	37	655	68	2012-08-02 10:25:38.667486
34007	3	1-9394	34005	37	81	538	78	2012-08-02 10:25:38.667486
34005	3	1-10303	34005	81	60	739	96	2012-08-02 10:25:38.667486
34009	3	1-9528	34008	38	40	1097	9	2012-08-02 10:25:39.452582
34010	3	1-9904	34008	40	5	233	16	2012-08-02 10:25:39.452582
34008	3	1-10305	34008	5	38	4924	11	2012-08-02 10:25:39.452582
34012	4	1-10141	34011	15	17	6666	25	2012-08-02 10:25:39.850209
34013	4	1-5744	34011	17	72	3950	14	2012-08-02 10:25:39.850209
34014	4	1-9586	34011	72	64	3315	42	2012-08-02 10:25:39.850209
34011	4	1-10306	34011	64	15	6459	87	2012-08-02 10:25:39.850209
34016	4	1-10136	34015	97	76	2758	26	2012-08-02 10:25:40.236947
34017	4	1-7244	34015	76	80	1322	81	2012-08-02 10:25:40.236947
34018	4	1-9581	34015	80	44	1444	58	2012-08-02 10:25:40.236947
34015	4	1-10307	34015	44	97	4483	76	2012-08-02 10:25:40.236947
34020	2	1-8819	34019	24	53	387	39	2012-08-02 10:25:40.657368
34019	2	1-10308	34019	53	24	8794	18	2012-08-02 10:25:40.657368
34022	5	1-9480	34021	5	38	830	68	2012-08-02 10:25:41.452853
34023	5	1-10228	34021	38	17	654	28	2012-08-02 10:25:41.452853
34024	5	1-10070	34021	17	70	858	18	2012-08-02 10:25:41.452853
34025	5	1-10009	34021	70	27	958	47	2012-08-02 10:25:41.452853
34021	5	1-10310	34021	27	5	1721	7	2012-08-02 10:25:41.452853
34027	5	1-7683	34026	60	6	306	21	2012-08-02 10:25:41.452853
34028	5	1-8968	34026	6	17	213	36	2012-08-02 10:25:41.452853
34029	5	1-9426	34026	17	70	203	18	2012-08-02 10:25:41.452853
34030	5	1-10009	34026	70	27	247	47	2012-08-02 10:25:41.452853
34026	5	1-10310	34026	27	60	484	7	2012-08-02 10:25:41.452853
34031	1	1-8109	\N	77	77	826	36	2012-08-02 10:25:41.452853
34032	1	1-3418	\N	12	12	1901	47	2012-08-02 10:25:41.452853
34034	4	1-9932	34033	65	24	409	89	2012-08-02 10:25:41.851076
34035	4	1-5043	34033	24	40	116	67	2012-08-02 10:25:41.851076
34036	4	1-9939	34033	40	93	291	77	2012-08-02 10:25:41.851076
34033	4	1-10311	34033	93	65	897	61	2012-08-02 10:25:41.851076
34038	3	1-5228	34037	22	22	1253	1	2012-08-02 10:25:41.851076
34039	3	1-10221	34037	22	93	1113	77	2012-08-02 10:25:41.851076
34037	3	1-10311	34037	93	22	3612	61	2012-08-02 10:25:41.851076
34041	4	1-8134	34040	59	85	1724	48	2012-08-02 10:25:43.801197
34042	4	1-10292	34040	85	51	1546	16	2012-08-02 10:25:43.801197
34043	4	1-10314	34040	51	60	947	84	2012-08-02 10:25:43.801197
34040	4	1-10316	34040	60	59	2495	76	2012-08-02 10:25:43.801197
34045	4	1-10136	34044	97	96	1185	26	2012-08-02 10:25:43.801197
34046	4	1-9377	34044	96	51	1139	16	2012-08-02 10:25:43.801197
34047	4	1-10314	34044	51	60	753	84	2012-08-02 10:25:43.801197
34044	4	1-10316	34044	60	97	2142	76	2012-08-02 10:25:43.801197
34049	4	1-5429	34048	85	85	788	64	2012-08-02 10:25:44.188443
34050	4	1-10137	34048	85	62	1151	36	2012-08-02 10:25:44.188443
34051	4	1-9704	34048	62	4	2792	10	2012-08-02 10:25:44.188443
34048	4	1-10317	34048	4	85	1645	24	2012-08-02 10:25:44.188443
34053	2	1-10000	34052	81	19	5451	90	2012-08-02 10:25:44.961748
34052	2	1-10319	34052	19	81	5777	51	2012-08-02 10:25:44.961748
34055	4	1-10084	34054	30	42	4385	61	2012-08-02 10:25:45.746324
34056	4	1-10198	34054	42	60	3205	35	2012-08-02 10:25:45.746324
34057	4	1-1317	34054	60	77	1489	43	2012-08-02 10:25:45.746324
34054	4	1-10321	34054	77	30	2837	96	2012-08-02 10:25:45.746324
34059	4	1-10306	34058	64	15	260	87	2012-08-02 10:25:46.917434
34060	4	1-10141	34058	15	87	215	25	2012-08-02 10:25:46.917434
34058	4	1-10324	34058	21	64	167	42	2012-08-02 10:25:46.917434
34061	4	1-2541	34058	87	21	88	93	2012-08-02 10:25:46.917434
34063	4	1-10315	34062	16	22	3458	61	2012-08-02 10:25:47.691152
34064	4	1-5228	34062	22	16	424	1	2012-08-02 10:25:47.691152
34065	4	1-9799	34062	16	17	229	54	2012-08-02 10:25:47.691152
34062	4	1-10326	34062	17	16	5229	76	2012-08-02 10:25:47.691152
34067	4	1-10124	34066	38	13	1371	14	2012-08-02 10:25:48.088715
34068	4	1-9040	34066	13	33	1217	92	2012-08-02 10:25:48.088715
34069	4	1-9952	34066	33	95	613	89	2012-08-02 10:25:48.088715
34066	4	1-10327	34066	95	38	575	82	2012-08-02 10:25:48.088715
34071	4	1-9369	34070	90	60	2255	25	2012-08-02 10:25:51.603785
34072	4	1-10252	34070	60	34	811	63	2012-08-02 10:25:51.603785
34073	4	1-8978	34070	34	12	1332	91	2012-08-02 10:25:51.603785
34070	4	1-10336	34070	12	90	3971	44	2012-08-02 10:25:51.603785
34075	4	1-9965	34074	89	29	1188	17	2012-08-02 10:25:51.603785
34076	4	1-8852	34074	29	99	1091	77	2012-08-02 10:25:51.603785
34077	4	1-9966	34074	99	12	530	91	2012-08-02 10:25:51.603785
34074	4	1-10336	34074	12	89	1614	44	2012-08-02 10:25:51.603785
34079	2	1-9753	34078	100	98	7097	71	2012-08-02 10:27:23.001032
34078	2	1-10346	34078	98	100	9835	50	2012-08-02 10:27:23.001032
34081	4	1-10319	34080	19	69	709	51	2012-08-02 10:27:23.432272
34082	4	1-8549	34080	69	52	255	37	2012-08-02 10:27:23.432272
34083	4	1-10269	34080	52	10	244	95	2012-08-02 10:27:23.432272
34080	4	1-10347	34080	10	19	589	90	2012-08-02 10:27:23.432272
34085	4	1-10036	34084	80	99	4481	77	2012-08-02 10:27:23.432272
34086	4	1-9966	34084	99	81	1691	91	2012-08-02 10:27:23.432272
34087	4	1-9171	34084	81	10	1023	95	2012-08-02 10:27:23.432272
34084	4	1-10347	34084	10	80	2545	90	2012-08-02 10:27:23.432272
34089	4	1-9936	34088	69	75	177	46	2012-08-02 10:27:23.432272
34090	4	1-8299	34088	75	6	58	1	2012-08-02 10:27:23.432272
34091	4	1-7347	34088	6	10	41	95	2012-08-02 10:27:23.432272
34088	4	1-10347	34088	10	69	102	90	2012-08-02 10:27:23.432272
34093	4	1-7964	34092	33	41	39	97	2012-08-02 10:27:23.432272
34094	4	1-9153	34092	41	52	26	37	2012-08-02 10:27:23.432272
34095	4	1-10269	34092	52	10	28	95	2012-08-02 10:27:23.432272
34092	4	1-10347	34092	10	33	75	90	2012-08-02 10:27:23.432272
34097	5	1-9570	34096	66	17	4239	81	2012-08-02 10:27:25.06777
34098	5	1-6372	34096	17	33	4959	92	2012-08-02 10:27:25.06777
34099	5	1-9952	34096	33	95	3223	89	2012-08-02 10:27:25.06777
34100	5	1-10327	34096	95	26	3899	82	2012-08-02 10:27:25.06777
34096	5	1-10351	34096	26	66	4667	43	2012-08-02 10:27:25.06777
34102	5	1-10321	34101	77	55	1385	96	2012-08-02 10:27:25.06777
34103	5	1-9412	34101	55	84	592	79	2012-08-02 10:27:25.06777
34104	5	1-10169	34101	84	95	428	89	2012-08-02 10:27:25.06777
34105	5	1-10327	34101	95	26	537	82	2012-08-02 10:27:25.06777
34101	5	1-10351	34101	26	77	667	43	2012-08-02 10:27:25.06777
34107	4	1-5848	34106	44	41	1347	85	2012-08-02 10:27:25.896347
34108	4	1-10337	34106	41	92	426	84	2012-08-02 10:27:25.896347
34109	4	1-9365	34106	92	62	147	77	2012-08-02 10:27:25.896347
34106	4	1-10353	34106	62	44	1544	98	2012-08-02 10:27:25.896347
34111	3	1-3674	34110	49	2	678	54	2012-08-02 10:27:25.896347
34112	3	1-9000	34110	2	62	200	77	2012-08-02 10:27:25.896347
34110	3	1-10353	34110	62	49	2196	98	2012-08-02 10:27:25.896347
34114	4	1-10261	34113	37	70	1373	24	2012-08-02 10:27:25.896347
34115	4	1-7696	34113	70	22	499	1	2012-08-02 10:27:25.896347
34116	4	1-10221	34113	22	62	255	77	2012-08-02 10:27:25.896347
34113	4	1-10353	34113	62	37	3031	98	2012-08-02 10:27:25.896347
34118	5	1-10340	34117	74	85	1370	32	2012-08-02 10:27:26.691551
34119	5	1-10135	34117	85	17	1918	36	2012-08-02 10:27:26.691551
34120	5	1-9426	34117	17	70	1808	18	2012-08-02 10:27:26.691551
34121	5	1-10009	34117	70	51	2184	47	2012-08-02 10:27:26.691551
34117	5	1-10355	34117	51	74	1241	6	2012-08-02 10:27:26.691551
34122	1	1-6982	\N	22	22	841	18	2012-08-02 10:27:26.691551
34124	5	1-9955	34123	94	58	1387	23	2012-08-02 10:27:28.657956
34125	5	1-9971	34123	58	5	1230	8	2012-08-02 10:27:28.657956
34126	5	1-1506	34123	5	78	375	33	2012-08-02 10:27:28.657956
34127	5	1-10279	34123	78	72	529	39	2012-08-02 10:27:28.657956
34123	5	1-10360	34123	72	94	1392	80	2012-08-02 10:27:28.657956
34129	5	1-9918	34128	20	92	557	46	2012-08-02 10:27:28.657956
34130	5	1-9591	34128	92	3	273	28	2012-08-02 10:27:28.657956
34131	5	1-10085	34128	3	78	292	33	2012-08-02 10:27:28.657956
34132	5	1-10279	34128	78	72	417	39	2012-08-02 10:27:28.657956
34128	5	1-10360	34128	72	20	1109	80	2012-08-02 10:27:28.657956
34134	2	1-9954	34133	36	36	4223	70	2012-08-02 10:27:29.066977
34133	2	1-10361	34133	36	36	5906	51	2012-08-02 10:27:29.066977
34136	3	1-8549	34135	69	23	410	37	2012-08-02 10:27:29.066977
34137	3	1-10029	34135	23	36	473	70	2012-08-02 10:27:29.066977
34135	3	1-10361	34135	36	69	1242	51	2012-08-02 10:27:29.066977
34139	3	1-6574	34138	85	55	342	56	2012-08-02 10:27:29.465499
34140	3	1-10110	34138	55	22	603	17	2012-08-02 10:27:29.465499
34138	3	1-10362	34138	22	85	646	24	2012-08-02 10:27:29.465499
34142	4	1-9577	34141	13	4	2840	60	2012-08-02 10:27:29.465499
34143	4	1-9522	34141	4	89	1909	44	2012-08-02 10:27:29.465499
34144	4	1-9965	34141	89	22	1151	17	2012-08-02 10:27:29.465499
34141	4	1-10362	34141	22	13	1539	24	2012-08-02 10:27:29.465499
34146	4	1-8675	34145	99	36	1310	54	2012-08-02 10:27:29.465499
34147	4	1-9727	34145	36	12	622	94	2012-08-02 10:27:29.465499
34148	4	1-9846	34145	12	22	1088	17	2012-08-02 10:27:29.465499
34145	4	1-10362	34145	22	99	1507	24	2012-08-02 10:27:29.465499
34150	4	1-8655	34149	17	16	68	27	2012-08-02 10:27:29.465499
34151	4	1-10151	34149	16	71	76	35	2012-08-02 10:27:29.465499
34152	4	1-1454	34149	71	22	46	17	2012-08-02 10:27:29.465499
34149	4	1-10362	34149	22	17	64	24	2012-08-02 10:27:29.465499
34154	4	1-7565	34153	60	100	4088	79	2012-08-02 10:27:29.465499
34155	4	1-10046	34153	100	95	3293	38	2012-08-02 10:27:29.465499
34156	4	1-9941	34153	95	22	2208	17	2012-08-02 10:27:29.465499
34153	4	1-10362	34153	22	60	3277	24	2012-08-02 10:27:29.465499
34158	4	1-6499	34157	73	48	498	29	2012-08-02 10:27:30.659208
34159	4	1-8740	34157	48	90	193	58	2012-08-02 10:27:30.659208
34160	4	1-7383	34157	90	38	34	20	2012-08-02 10:27:30.659208
34157	4	1-10365	34157	38	73	1059	12	2012-08-02 10:27:30.659208
34162	4	1-7806	34161	88	83	511	68	2012-08-02 10:27:30.659208
34163	4	1-10301	34161	83	42	190	75	2012-08-02 10:27:30.659208
34164	4	1-10106	34161	42	38	142	20	2012-08-02 10:27:30.659208
34161	4	1-10365	34161	38	88	4486	12	2012-08-02 10:27:30.659208
34166	3	1-10321	34165	77	24	426	96	2012-08-02 10:27:31.057118
34167	3	1-7932	34165	24	65	263	38	2012-08-02 10:27:31.057118
34165	3	1-10366	34165	65	77	275	43	2012-08-02 10:27:31.057118
34169	4	1-3657	34168	33	50	345	13	2012-08-02 10:27:31.057118
34170	4	1-9975	34168	50	86	341	75	2012-08-02 10:27:31.057118
34171	4	1-9946	34168	86	65	529	38	2012-08-02 10:27:31.057118
34168	4	1-10366	34168	65	33	672	43	2012-08-02 10:27:31.057118
34173	4	1-10321	34172	77	55	382	96	2012-08-02 10:27:31.057118
34174	4	1-9412	34172	55	100	158	79	2012-08-02 10:27:31.057118
34175	4	1-10046	34172	100	65	140	38	2012-08-02 10:27:31.057118
34172	4	1-10366	34172	65	77	190	43	2012-08-02 10:27:31.057118
34177	4	1-9964	34176	80	41	2340	16	2012-08-02 10:27:31.057118
34178	4	1-9997	34176	41	12	1385	35	2012-08-02 10:27:31.057118
34179	4	1-7052	34176	12	65	818	38	2012-08-02 10:27:31.057118
34176	4	1-10366	34176	65	80	1157	43	2012-08-02 10:27:31.057118
34181	3	1-7754	34180	37	47	5173	16	2012-08-02 10:27:31.875572
34182	3	1-9360	34180	47	60	2899	31	2012-08-02 10:27:31.875572
34180	3	1-10368	34180	60	37	8537	6	2012-08-02 10:27:31.875572
34184	4	1-9715	34183	65	17	138	78	2012-08-02 10:27:31.875572
34185	4	1-10239	34183	17	73	110	24	2012-08-02 10:27:31.875572
34186	4	1-9265	34183	73	60	98	31	2012-08-02 10:27:31.875572
34183	4	1-10368	34183	60	65	293	6	2012-08-02 10:27:31.875572
34188	4	1-10188	34187	67	85	576	32	2012-08-02 10:27:32.27355
34189	4	1-10135	34187	85	44	740	36	2012-08-02 10:27:32.27355
34190	4	1-10192	34187	44	95	389	94	2012-08-02 10:27:32.27355
34187	4	1-10369	34187	95	67	1347	78	2012-08-02 10:27:32.27355
34192	4	1-9177	34191	75	84	1348	79	2012-08-02 10:27:32.27355
34193	4	1-10169	34191	84	57	893	89	2012-08-02 10:27:32.27355
34194	4	1-6859	34191	57	95	338	94	2012-08-02 10:27:32.27355
34191	4	1-10369	34191	95	75	1255	78	2012-08-02 10:27:32.27355
34196	4	1-10239	34195	17	70	2098	24	2012-08-02 10:27:32.27355
34197	4	1-7696	34195	70	80	1481	1	2012-08-02 10:27:32.27355
34198	4	1-7451	34195	80	95	638	94	2012-08-02 10:27:32.27355
34195	4	1-10369	34195	95	17	2425	78	2012-08-02 10:27:32.27355
34200	3	1-9880	34199	38	54	1950	18	2012-08-02 10:27:32.682288
34201	3	1-8628	34199	54	90	579	14	2012-08-02 10:27:32.682288
34199	3	1-10370	34199	90	38	7065	63	2012-08-02 10:27:32.682288
34203	4	1-9574	34202	68	6	690	25	2012-08-02 10:27:33.466623
34204	4	1-8445	34202	6	76	761	70	2012-08-02 10:27:33.466623
34205	4	1-7076	34202	76	53	364	48	2012-08-02 10:27:33.466623
34202	4	1-10372	34202	53	68	675	15	2012-08-02 10:27:33.466623
34207	4	1-10369	34206	95	28	1220	78	2012-08-02 10:27:35.035427
34208	4	1-10342	34206	28	10	1151	73	2012-08-02 10:27:35.035427
34209	4	1-10150	34206	10	83	2387	12	2012-08-02 10:27:35.035427
34206	4	1-10376	34206	83	95	471	94	2012-08-02 10:27:35.035427
34211	3	1-8054	34210	61	53	2771	48	2012-08-02 10:27:35.444547
34212	3	1-10372	34210	53	17	4850	15	2012-08-02 10:27:35.444547
34210	3	1-10377	34210	17	61	4144	21	2012-08-02 10:27:35.444547
34214	3	1-9975	34213	50	40	1752	75	2012-08-02 10:27:35.842107
34215	3	1-9943	34213	40	86	1073	9	2012-08-02 10:27:35.842107
34213	3	1-10378	34213	86	50	2233	13	2012-08-02 10:27:35.842107
34217	2	1-10170	34216	24	77	1012	4	2012-08-02 10:27:36.628636
34216	2	1-10380	34216	77	24	1163	62	2012-08-02 10:27:36.628636
34219	4	1-10275	34218	67	21	51	1	2012-08-02 10:27:36.628636
34220	4	1-7669	34218	21	45	30	97	2012-08-02 10:27:36.628636
34221	4	1-9058	34218	45	77	17	4	2012-08-02 10:27:36.628636
34218	4	1-10380	34218	77	67	29	62	2012-08-02 10:27:36.628636
34223	4	1-10293	34222	64	1	889	18	2012-08-02 10:27:36.628636
34224	4	1-8145	34222	1	86	424	33	2012-08-02 10:27:36.628636
34225	4	1-9391	34222	86	77	316	4	2012-08-02 10:27:36.628636
34222	4	1-10380	34222	77	64	646	62	2012-08-02 10:27:36.628636
34227	4	1-9984	34226	32	56	108	35	2012-08-02 10:27:36.628636
34228	4	1-10345	34226	56	19	102	66	2012-08-02 10:27:36.628636
34229	4	1-10140	34226	19	77	50	4	2012-08-02 10:27:36.628636
34226	4	1-10380	34226	77	32	105	62	2012-08-02 10:27:36.628636
34231	4	1-9991	34230	90	35	3447	31	2012-08-02 10:27:37.424233
34232	4	1-10090	34230	35	81	3262	34	2012-08-02 10:27:37.424233
34233	4	1-9400	34230	81	50	6144	85	2012-08-02 10:27:37.424233
34230	4	1-10382	34230	50	90	5154	39	2012-08-02 10:27:37.424233
34235	3	1-7332	34234	17	15	617	29	2012-08-02 10:27:37.424233
34236	3	1-8811	34234	15	50	881	85	2012-08-02 10:27:37.424233
34234	3	1-10382	34234	50	17	745	39	2012-08-02 10:27:37.424233
34238	4	1-7964	34237	33	87	2758	97	2012-08-02 10:27:37.822039
34239	4	1-10062	34237	87	65	3371	38	2012-08-02 10:27:37.822039
34240	4	1-10366	34237	65	59	4418	43	2012-08-02 10:27:37.822039
34237	4	1-10383	34237	59	33	4749	90	2012-08-02 10:27:37.822039
34242	4	1-8937	34241	36	16	3494	16	2012-08-02 10:27:42.174516
34243	4	1-4323	34241	16	62	491	56	2012-08-02 10:27:42.174516
34244	4	1-9879	34241	62	79	665	50	2012-08-02 10:27:42.174516
34241	4	1-10394	34241	79	36	2228	39	2012-08-02 10:27:42.174516
34246	4	1-9283	34245	87	1	3280	61	2012-08-02 10:27:42.174516
34247	4	1-2136	34245	1	62	703	56	2012-08-02 10:27:42.174516
34248	4	1-9879	34245	62	79	1043	50	2012-08-02 10:27:42.174516
34245	4	1-10394	34245	79	87	3828	39	2012-08-02 10:27:42.174516
34250	4	1-7332	34249	17	77	396	29	2012-08-02 10:27:42.174516
34251	4	1-10103	34249	77	32	391	83	2012-08-02 10:27:42.174516
34252	4	1-9822	34249	32	79	113	50	2012-08-02 10:27:42.174516
34249	4	1-10394	34249	79	17	428	39	2012-08-02 10:27:42.174516
34254	5	1-10359	34253	52	63	101	53	2012-08-02 10:27:42.572796
34255	5	1-9877	34253	63	7	38	1	2012-08-02 10:27:42.572796
34256	5	1-9118	34253	7	27	16	47	2012-08-02 10:27:42.572796
34257	5	1-10310	34253	27	56	25	7	2012-08-02 10:27:42.572796
34253	5	1-10395	34253	56	52	50	98	2012-08-02 10:27:42.572796
34259	4	1-10261	34258	37	18	2181	24	2012-08-02 10:27:42.572796
34260	4	1-5591	34258	18	39	2467	50	2012-08-02 10:27:42.572796
34261	4	1-10041	34258	39	56	1262	7	2012-08-02 10:27:42.572796
34258	4	1-10395	34258	56	37	2855	98	2012-08-02 10:27:42.572796
34263	4	1-10392	34262	88	51	859	16	2012-08-02 10:27:43.777161
34264	4	1-10314	34262	51	92	528	84	2012-08-02 10:27:43.777161
34265	4	1-9365	34262	92	86	300	77	2012-08-02 10:27:43.777161
34262	4	1-10398	34262	86	88	1123	87	2012-08-02 10:27:43.777161
34267	4	1-10141	34266	15	31	1055	25	2012-08-02 10:27:43.777161
34268	4	1-9389	34266	31	94	172	23	2012-08-02 10:27:43.777161
34269	4	1-10283	34266	94	86	238	77	2012-08-02 10:27:43.777161
34266	4	1-10398	34266	86	15	1065	87	2012-08-02 10:27:43.777161
34271	3	1-7222	34270	77	40	191	67	2012-08-02 10:27:43.777161
34272	3	1-9939	34270	40	86	654	77	2012-08-02 10:27:43.777161
34270	3	1-10398	34270	86	77	3106	87	2012-08-02 10:27:43.777161
34274	4	1-8627	34273	67	47	2334	45	2012-08-02 10:27:44.572949
34275	4	1-10233	34273	47	19	2115	66	2012-08-02 10:27:44.572949
34276	4	1-10140	34273	19	98	675	4	2012-08-02 10:27:44.572949
34273	4	1-10400	34273	98	67	6145	38	2012-08-02 10:27:44.572949
34278	4	1-7279	34277	32	47	340	45	2012-08-02 10:27:44.572949
34279	4	1-10233	34277	47	19	328	66	2012-08-02 10:27:44.572949
34280	4	1-10140	34277	19	98	112	4	2012-08-02 10:27:44.572949
34277	4	1-10400	34277	98	32	1081	38	2012-08-02 10:27:44.572949
34282	3	1-8452	34281	33	33	856	41	2012-08-02 10:27:44.572949
34283	3	1-10222	34281	33	98	132	4	2012-08-02 10:27:44.572949
34281	3	1-10400	34281	98	33	1309	38	2012-08-02 10:27:44.572949
34285	3	1-10090	34284	35	45	3184	34	2012-08-02 10:27:44.981802
34286	3	1-7001	34284	45	57	654	86	2012-08-02 10:27:44.981802
34284	3	1-10401	34284	57	35	5804	31	2012-08-02 10:27:44.981802
34288	4	1-10381	34287	63	83	2833	60	2012-08-02 10:27:47.699369
34289	4	1-8055	34287	83	1	1607	95	2012-08-02 10:27:47.699369
34290	4	1-10393	34287	1	94	985	69	2012-08-02 10:27:47.699369
34287	4	1-10408	34287	94	63	1266	41	2012-08-02 10:27:47.699369
34292	4	1-9875	34291	75	24	350	100	2012-08-02 10:27:47.699369
34293	4	1-10240	34291	24	66	800	42	2012-08-02 10:27:47.699369
34294	4	1-9839	34291	66	94	266	69	2012-08-02 10:27:47.699369
34291	4	1-10408	34291	94	75	364	41	2012-08-02 10:27:47.699369
34296	4	1-5302	34295	51	5	1104	58	2012-08-02 10:27:47.699369
34297	4	1-10276	34295	5	66	3096	42	2012-08-02 10:27:47.699369
34298	4	1-9839	34295	66	94	1043	69	2012-08-02 10:27:47.699369
34295	4	1-10408	34295	94	51	1443	41	2012-08-02 10:27:47.699369
34300	3	1-9875	34299	75	26	2434	100	2012-08-02 10:27:47.699369
34301	3	1-8575	34299	26	94	1647	69	2012-08-02 10:27:47.699369
34299	3	1-10408	34299	94	75	2388	41	2012-08-02 10:27:47.699369
34303	4	1-7888	34302	18	83	1876	28	2012-08-02 10:27:48.097498
34304	4	1-10116	34302	83	16	4670	37	2012-08-02 10:27:48.097498
34305	4	1-10374	34302	16	30	4242	32	2012-08-02 10:27:48.097498
34302	4	1-10409	34302	30	18	5645	31	2012-08-02 10:27:48.097498
34307	3	1-10202	34306	4	45	1701	81	2012-08-02 10:27:48.484011
34308	3	1-9999	34306	45	6	1933	16	2012-08-02 10:27:48.484011
34306	3	1-10410	34306	6	4	2313	32	2012-08-02 10:27:48.484011
34310	4	1-10409	34309	30	35	1854	31	2012-08-02 10:27:48.484011
34311	4	1-10090	34309	35	27	1633	34	2012-08-02 10:27:48.484011
34312	4	1-8971	34309	27	6	1237	16	2012-08-02 10:27:48.484011
34309	4	1-10410	34309	6	30	1654	32	2012-08-02 10:27:48.484011
34314	4	1-10158	34313	95	27	50	10	2012-08-02 10:27:48.484011
34315	4	1-10173	34313	27	48	31	55	2012-08-02 10:27:48.484011
34316	4	1-9499	34313	48	6	29	16	2012-08-02 10:27:48.484011
34313	4	1-10410	34313	6	95	50	32	2012-08-02 10:27:48.484011
34318	4	1-10291	34317	5	40	179	43	2012-08-02 10:27:48.8819
34319	4	1-10300	34317	40	85	129	64	2012-08-02 10:27:48.8819
34320	4	1-10137	34317	85	59	180	36	2012-08-02 10:27:48.8819
34317	4	1-10411	34317	59	5	357	92	2012-08-02 10:27:48.8819
34322	4	1-8225	34321	74	79	218	19	2012-08-02 10:27:49.26852
34323	4	1-10177	34321	79	21	508	86	2012-08-02 10:27:49.26852
34324	4	1-8643	34321	21	44	228	17	2012-08-02 10:27:49.26852
34321	4	1-10412	34321	44	74	549	96	2012-08-02 10:27:49.26852
34326	4	1-9612	34325	75	79	1049	19	2012-08-02 10:27:49.26852
34327	4	1-10177	34325	79	21	2739	86	2012-08-02 10:27:49.26852
34328	4	1-8643	34325	21	44	1377	17	2012-08-02 10:27:49.26852
34325	4	1-10412	34325	44	75	3718	96	2012-08-02 10:27:49.26852
34330	4	1-10252	34329	60	38	206	63	2012-08-02 10:27:49.677574
34331	4	1-9880	34329	38	54	239	18	2012-08-02 10:27:49.677574
34332	4	1-8628	34329	54	15	297	14	2012-08-02 10:27:49.677574
34329	4	1-10413	34329	15	60	555	25	2012-08-02 10:27:49.677574
34334	2	1-10050	34333	47	56	339	51	2012-08-02 10:27:50.473143
34333	2	1-10415	34333	56	47	1977	87	2012-08-02 10:27:50.473143
34336	4	1-10211	34335	63	45	6026	44	2012-08-02 10:27:50.473143
34337	4	1-10294	34335	45	22	2117	65	2012-08-02 10:27:50.473143
34338	4	1-6493	34335	22	56	1079	51	2012-08-02 10:27:50.473143
34335	4	1-10415	34335	56	63	6335	87	2012-08-02 10:27:50.473143
34340	4	1-9939	34339	40	25	1171	77	2012-08-02 10:27:51.268635
34341	4	1-10104	34339	25	67	208	61	2012-08-02 10:27:51.268635
34342	4	1-7870	34339	67	81	39	7	2012-08-02 10:27:51.268635
34339	4	1-10417	34339	81	40	1152	67	2012-08-02 10:27:51.268635
34344	3	1-5122	34343	17	58	307	71	2012-08-02 10:27:51.268635
34345	3	1-10127	34343	58	81	117	7	2012-08-02 10:27:51.268635
34343	3	1-10417	34343	81	17	4387	67	2012-08-02 10:27:51.268635
34347	4	1-10203	34346	20	77	1592	29	2012-08-02 10:27:51.666423
34348	4	1-10103	34346	77	13	1400	83	2012-08-02 10:27:51.666423
34349	4	1-9311	34346	13	4	1117	59	2012-08-02 10:27:51.666423
34346	4	1-10418	34346	4	20	2822	5	2012-08-02 10:27:51.666423
34351	2	1-9375	34350	27	27	626	2	2012-08-02 10:27:52.064185
34350	2	1-10419	34350	27	27	366	28	2012-08-02 10:27:52.064185
34353	4	1-10116	34352	83	23	2931	37	2012-08-02 10:27:52.064185
34354	4	1-10029	34352	23	86	3343	70	2012-08-02 10:27:52.064185
34355	4	1-9530	34352	86	27	2408	2	2012-08-02 10:27:52.064185
34352	4	1-10419	34352	27	83	1509	28	2012-08-02 10:27:52.064185
34357	4	1-6942	34356	48	86	2361	31	2012-08-02 10:27:52.064185
34358	4	1-8829	34356	86	80	2815	85	2012-08-02 10:27:52.064185
34359	4	1-10343	34356	80	27	4400	2	2012-08-02 10:27:52.064185
34356	4	1-10419	34356	27	48	3443	28	2012-08-02 10:27:52.064185
34361	4	1-3861	34360	6	95	1503	32	2012-08-02 10:27:52.837679
34362	4	1-10158	34360	95	4	1653	10	2012-08-02 10:27:52.837679
34363	4	1-10317	34360	4	46	1002	24	2012-08-02 10:27:52.837679
34360	4	1-10421	34360	46	6	3637	41	2012-08-02 10:27:52.837679
34365	4	1-10381	34364	63	37	5936	60	2012-08-02 10:27:55.213041
34366	4	1-10126	34364	37	64	1813	46	2012-08-02 10:27:55.213041
34367	4	1-8036	34364	64	40	767	93	2012-08-02 10:27:55.213041
34364	4	1-10427	34364	40	63	2891	41	2012-08-02 10:27:55.213041
34369	4	1-3963	34368	40	15	1282	14	2012-08-02 10:27:55.213041
34370	4	1-10413	34368	15	87	1524	25	2012-08-02 10:27:55.213041
34371	4	1-2541	34368	87	40	557	93	2012-08-02 10:27:55.213041
34368	4	1-10427	34368	40	40	2330	41	2012-08-02 10:27:55.213041
34373	4	1-1063	34372	59	15	668	14	2012-08-02 10:27:55.213041
34374	4	1-10413	34372	15	87	1022	25	2012-08-02 10:27:55.213041
34375	4	1-2541	34372	87	40	481	93	2012-08-02 10:27:55.213041
34372	4	1-10427	34372	40	59	2036	41	2012-08-02 10:27:55.213041
34377	4	1-7740	34376	68	29	48	11	2012-08-02 10:27:55.213041
34378	4	1-10093	34376	29	8	63	27	2012-08-02 10:27:55.213041
34379	4	1-6141	34376	8	40	39	93	2012-08-02 10:27:55.213041
34376	4	1-10427	34376	40	68	167	41	2012-08-02 10:27:55.213041
34381	4	1-8768	34380	64	76	368	53	2012-08-02 10:27:55.611142
34382	4	1-10354	34380	76	44	207	17	2012-08-02 10:27:55.611142
34383	4	1-10412	34380	44	82	532	96	2012-08-02 10:27:55.611142
34380	4	1-10428	34380	82	64	313	64	2012-08-02 10:27:55.611142
34385	4	1-8052	34384	3	76	1439	53	2012-08-02 10:27:56.384395
34386	4	1-10354	34384	76	44	837	17	2012-08-02 10:27:56.384395
34387	4	1-10412	34384	44	8	2238	96	2012-08-02 10:27:56.384395
34384	4	1-10430	34384	8	3	1114	19	2012-08-02 10:27:56.384395
34389	5	1-9881	34388	82	16	3078	76	2012-08-02 10:27:57.555811
34390	5	1-10315	34388	16	92	3326	61	2012-08-02 10:27:57.555811
34391	5	1-8955	34388	92	82	1165	63	2012-08-02 10:27:57.555811
34392	5	1-9138	34388	82	4	1416	55	2012-08-02 10:27:57.555811
34388	5	1-10433	34388	4	82	3245	4	2012-08-02 10:27:57.555811
34394	4	1-9881	34393	82	97	602	76	2012-08-02 10:27:57.555811
34395	4	1-10136	34393	97	95	421	26	2012-08-02 10:27:57.555811
34396	4	1-8457	34393	95	4	256	55	2012-08-02 10:27:57.555811
34393	4	1-10433	34393	4	82	630	4	2012-08-02 10:27:57.555811
34398	5	1-10097	34397	62	83	1499	73	2012-08-02 10:27:57.555811
34399	5	1-10260	34397	83	60	1399	25	2012-08-02 10:27:57.555811
34400	5	1-10252	34397	60	82	529	63	2012-08-02 10:27:57.555811
34401	5	1-9138	34397	82	4	674	55	2012-08-02 10:27:57.555811
34397	5	1-10433	34397	4	62	1725	4	2012-08-02 10:27:57.555811
34403	5	1-10420	34402	19	87	2689	55	2012-08-02 10:27:57.953321
34404	5	1-7626	34402	87	82	1646	74	2012-08-02 10:27:57.953321
34405	5	1-10201	34402	82	29	1689	11	2012-08-02 10:27:57.953321
34406	5	1-10093	34402	29	34	1849	27	2012-08-02 10:27:57.953321
34402	5	1-10434	34402	34	19	6156	37	2012-08-02 10:27:57.953321
34408	3	1-9804	34407	16	36	2156	79	2012-08-02 10:27:57.953321
34409	3	1-9920	34407	36	34	668	27	2012-08-02 10:27:57.953321
34407	3	1-10434	34407	34	16	2491	37	2012-08-02 10:27:57.953321
34411	5	1-10029	34410	23	36	782	70	2012-08-02 10:27:57.953321
34412	5	1-4711	34410	36	82	79	74	2012-08-02 10:27:57.953321
34413	5	1-10201	34410	82	29	99	11	2012-08-02 10:27:57.953321
34414	5	1-10093	34410	29	34	133	27	2012-08-02 10:27:57.953321
34410	5	1-10434	34410	34	23	545	37	2012-08-02 10:27:57.953321
34416	3	1-10385	34415	43	19	4286	9	2012-08-02 10:27:59.544539
34417	3	1-10238	34415	19	24	3464	42	2012-08-02 10:27:59.544539
34415	3	1-10438	34415	24	43	7041	73	2012-08-02 10:27:59.544539
34419	2	1-7461	34418	88	24	194	42	2012-08-02 10:27:59.544539
34418	2	1-10438	34418	24	88	494	73	2012-08-02 10:27:59.544539
34421	4	1-9655	34420	89	16	2005	95	2012-08-02 10:28:00.328852
34422	4	1-9934	34420	16	62	2090	57	2012-08-02 10:28:00.328852
34423	4	1-10048	34420	62	76	967	55	2012-08-02 10:28:00.328852
34420	4	1-10440	34420	76	89	2458	82	2012-08-02 10:28:00.328852
34425	4	1-6976	34424	66	39	967	92	2012-08-02 10:28:00.328852
34426	4	1-10147	34424	39	95	613	26	2012-08-02 10:28:00.328852
34427	4	1-8457	34424	95	76	331	55	2012-08-02 10:28:00.328852
34424	4	1-10440	34424	76	66	1068	82	2012-08-02 10:28:00.328852
34429	3	1-9896	34428	14	82	1965	7	2012-08-02 10:28:00.328852
34430	3	1-7544	34428	82	76	397	55	2012-08-02 10:28:00.328852
34428	3	1-10440	34428	76	14	1320	82	2012-08-02 10:28:00.328852
34432	3	1-3656	34431	24	27	1689	10	2012-08-02 10:28:00.328852
34433	3	1-10173	34431	27	76	1037	55	2012-08-02 10:28:00.328852
34431	3	1-10440	34431	76	24	3586	82	2012-08-02 10:28:00.328852
34435	3	1-8852	34434	29	39	1362	77	2012-08-02 10:31:34.56874
34436	3	1-5471	34434	39	11	1481	85	2012-08-02 10:31:34.56874
34434	3	1-10444	34434	11	29	1506	17	2012-08-02 10:31:34.56874
34438	4	1-9980	34437	68	79	2256	46	2012-08-02 10:31:34.56874
34439	4	1-9857	34437	79	14	2319	6	2012-08-02 10:31:34.56874
34440	4	1-9031	34437	14	11	2148	85	2012-08-02 10:31:34.56874
34437	4	1-10444	34437	11	68	2206	17	2012-08-02 10:31:34.56874
34442	3	1-9980	34441	68	62	368	46	2012-08-02 10:31:34.56874
34443	3	1-6710	34441	62	11	338	85	2012-08-02 10:31:34.56874
34441	3	1-10444	34441	11	68	353	17	2012-08-02 10:31:34.56874
34445	3	1-10283	34444	94	99	7138	77	2012-08-02 10:31:37.047096
34446	3	1-9966	34444	99	77	2612	91	2012-08-02 10:31:37.047096
34444	3	1-10450	34444	77	94	5948	23	2012-08-02 10:31:37.047096
34448	3	1-10283	34447	94	82	928	77	2012-08-02 10:31:37.047096
34449	3	1-8877	34447	82	77	265	91	2012-08-02 10:31:37.047096
34447	3	1-10450	34447	77	94	683	23	2012-08-02 10:31:37.047096
34451	4	1-8933	34450	21	20	3453	5	2012-08-02 10:31:37.047096
34452	4	1-10203	34450	20	82	2394	29	2012-08-02 10:31:37.047096
34453	4	1-10423	34450	82	77	813	91	2012-08-02 10:31:37.047096
34450	4	1-10450	34450	77	21	2399	23	2012-08-02 10:31:37.047096
34455	4	1-10369	34454	95	34	1690	78	2012-08-02 10:31:38.254753
34456	4	1-9024	34454	34	47	1253	69	2012-08-02 10:31:38.254753
34457	4	1-8495	34454	47	61	387	30	2012-08-02 10:31:38.254753
34454	4	1-10453	34454	61	95	438	94	2012-08-02 10:31:38.254753
34459	3	1-10313	34458	58	66	269	68	2012-08-02 10:31:38.254753
34460	3	1-9647	34458	66	61	225	30	2012-08-02 10:31:38.254753
34458	3	1-10453	34458	61	58	263	94	2012-08-02 10:31:38.254753
34462	5	1-10282	34461	93	19	588	26	2012-08-02 10:31:39.825454
34463	5	1-5175	34461	19	82	283	74	2012-08-02 10:31:39.825454
34464	5	1-10201	34461	82	29	346	11	2012-08-02 10:31:39.825454
34465	5	1-10093	34461	29	2	451	27	2012-08-02 10:31:39.825454
34461	5	1-10457	34461	2	93	772	48	2012-08-02 10:31:39.825454
34467	4	1-9878	34466	16	100	166	79	2012-08-02 10:31:41.395624
34468	4	1-10046	34466	100	35	100	38	2012-08-02 10:31:41.395624
34469	4	1-9455	34466	35	84	56	22	2012-08-02 10:31:41.395624
34466	4	1-10461	34466	84	16	389	88	2012-08-02 10:31:41.395624
34471	3	1-9689	34470	51	89	2091	23	2012-08-02 10:31:41.395624
34472	3	1-6123	34470	89	84	488	22	2012-08-02 10:31:41.395624
34470	3	1-10461	34470	84	51	3735	88	2012-08-02 10:31:41.395624
34474	4	1-9552	34473	5	12	2169	35	2012-08-02 10:31:41.395624
34475	4	1-7052	34473	12	35	929	38	2012-08-02 10:31:41.395624
34476	4	1-9455	34473	35	84	567	22	2012-08-02 10:31:41.395624
34473	4	1-10461	34473	84	5	4378	88	2012-08-02 10:31:41.395624
34478	5	1-7221	34477	67	18	68	31	2012-08-02 10:31:41.395624
34479	5	1-7888	34477	18	3	24	28	2012-08-02 10:31:41.395624
34480	5	1-10085	34477	3	99	25	33	2012-08-02 10:31:41.395624
34481	5	1-8220	34477	99	84	27	22	2012-08-02 10:31:41.395624
34477	5	1-10461	34477	84	67	261	88	2012-08-02 10:31:41.395624
34483	4	1-8495	34482	47	34	446	30	2012-08-02 10:31:41.819701
34484	4	1-9585	34482	34	3	705	75	2012-08-02 10:31:41.819701
34485	4	1-10017	34482	3	31	469	49	2012-08-02 10:31:41.819701
34482	4	1-10462	34482	31	47	1471	69	2012-08-02 10:31:41.819701
34487	4	1-6916	34486	97	18	637	86	2012-08-02 10:31:41.819701
34488	4	1-10263	34486	18	47	1135	35	2012-08-02 10:31:41.819701
34489	4	1-5333	34486	47	31	572	49	2012-08-02 10:31:41.819701
34486	4	1-10462	34486	31	97	1906	69	2012-08-02 10:31:41.819701
34491	4	1-10399	34490	96	41	3828	18	2012-08-02 10:31:41.819701
34492	4	1-9262	34490	41	3	1477	75	2012-08-02 10:31:41.819701
34493	4	1-10017	34490	3	31	1067	49	2012-08-02 10:31:41.819701
34490	4	1-10462	34490	31	96	3639	69	2012-08-02 10:31:41.819701
34495	4	1-10292	34494	85	51	2425	16	2012-08-02 10:31:42.618897
34496	4	1-10314	34494	51	57	1619	84	2012-08-02 10:31:42.618897
34497	4	1-7851	34494	57	81	1711	2	2012-08-02 10:31:42.618897
34494	4	1-10464	34494	81	85	2481	48	2012-08-02 10:31:42.618897
34499	4	1-10282	34498	93	15	584	26	2012-08-02 10:31:42.618897
34500	4	1-5292	34498	15	95	201	97	2012-08-02 10:31:42.618897
34501	4	1-9747	34498	95	81	444	2	2012-08-02 10:31:42.618897
34498	4	1-10464	34498	81	93	723	48	2012-08-02 10:31:42.618897
34503	4	1-5015	34502	59	35	1260	92	2012-08-02 10:31:42.618897
34504	4	1-10056	34502	35	86	2009	70	2012-08-02 10:31:42.618897
34505	4	1-9530	34502	86	81	2075	2	2012-08-02 10:31:42.618897
34502	4	1-10464	34502	81	59	3735	48	2012-08-02 10:31:42.618897
34507	4	1-8614	34506	81	42	222	59	2012-08-02 10:31:42.618897
34508	4	1-9068	34506	42	80	280	85	2012-08-02 10:31:42.618897
34509	4	1-10343	34506	80	81	490	2	2012-08-02 10:31:42.618897
34506	4	1-10464	34506	81	81	898	48	2012-08-02 10:31:42.618897
34510	1	1-5157	\N	64	64	738	59	2012-08-02 10:31:42.618897
34512	4	1-9841	34511	57	65	612	91	2012-08-02 10:31:43.803666
34513	4	1-7835	34511	65	52	791	37	2012-08-02 10:31:43.803666
34514	4	1-10269	34511	52	3	1025	95	2012-08-02 10:31:43.803666
34511	4	1-10467	34511	3	57	395	33	2012-08-02 10:31:43.803666
34516	4	1-10453	34515	61	92	6478	94	2012-08-02 10:31:46.621876
34517	4	1-8730	34515	92	82	2315	74	2012-08-02 10:31:46.621876
34518	4	1-10201	34515	82	85	2960	11	2012-08-02 10:31:46.621876
34515	4	1-10474	34515	85	61	6402	30	2012-08-02 10:31:46.621876
34520	3	1-10277	34519	68	31	2300	97	2012-08-02 10:31:47.40618
34521	3	1-8576	34519	31	95	648	78	2012-08-02 10:31:47.40618
34519	3	1-10476	34519	95	68	2306	27	2012-08-02 10:31:47.40618
34523	4	1-10252	34522	60	30	1087	63	2012-08-02 10:31:48.986021
34524	4	1-10391	34522	30	62	3553	91	2012-08-02 10:31:48.986021
34525	4	1-9323	34522	62	68	1739	28	2012-08-02 10:31:48.986021
34522	4	1-10480	34522	68	60	3116	25	2012-08-02 10:31:48.986021
34527	5	1-9934	34526	16	95	299	57	2012-08-02 10:31:50.577476
34528	5	1-10468	34526	95	4	184	6	2012-08-02 10:31:50.577476
34529	5	1-9854	34526	4	6	238	21	2012-08-02 10:31:50.577476
34530	5	1-8968	34526	6	38	176	36	2012-08-02 10:31:50.577476
34526	5	1-10484	34526	38	16	194	95	2012-08-02 10:31:50.577476
34532	2	1-10328	34531	23	69	3365	40	2012-08-02 10:31:51.385102
34531	2	1-10486	34531	69	23	5773	53	2012-08-02 10:31:51.385102
34534	4	1-10201	34533	82	85	1373	11	2012-08-02 10:31:53.341238
34535	4	1-9332	34533	85	52	805	3	2012-08-02 10:31:53.341238
34536	4	1-9149	34533	52	71	1377	29	2012-08-02 10:31:53.341238
34533	4	1-10491	34533	71	82	1454	74	2012-08-02 10:31:53.341238
34538	4	1-9754	34537	3	16	142	27	2012-08-02 10:31:53.341238
34539	4	1-10151	34537	16	42	163	35	2012-08-02 10:31:53.341238
34540	4	1-6590	34537	42	71	131	29	2012-08-02 10:31:53.341238
34537	4	1-10491	34537	71	3	187	74	2012-08-02 10:31:53.341238
34542	3	1-8655	34541	17	2	697	27	2012-08-02 10:31:53.750492
34543	3	1-10457	34541	2	89	1073	48	2012-08-02 10:31:53.750492
34541	3	1-10492	34541	89	17	738	24	2012-08-02 10:31:53.750492
34545	4	1-7565	34544	60	36	1812	79	2012-08-02 10:31:53.750492
34546	4	1-9920	34544	36	2	713	27	2012-08-02 10:31:53.750492
34547	4	1-10457	34544	2	89	1454	48	2012-08-02 10:31:53.750492
34544	4	1-10492	34544	89	60	1326	24	2012-08-02 10:31:53.750492
34549	4	1-10421	34548	46	91	2884	41	2012-08-02 10:31:53.750492
34550	4	1-5037	34548	91	21	1163	95	2012-08-02 10:31:53.750492
34551	4	1-9740	34548	21	89	877	48	2012-08-02 10:31:53.750492
34548	4	1-10492	34548	89	46	810	24	2012-08-02 10:31:53.750492
34553	4	1-6574	34552	85	83	492	56	2012-08-02 10:31:53.750492
34554	4	1-10108	34552	83	66	1193	75	2012-08-02 10:31:53.750492
34555	4	1-7411	34552	66	89	642	48	2012-08-02 10:31:53.750492
34552	4	1-10492	34552	89	85	598	24	2012-08-02 10:31:53.750492
34557	3	1-3158	34556	68	53	365	83	2012-08-02 10:31:53.750492
34558	3	1-9824	34556	53	89	502	48	2012-08-02 10:31:53.750492
34556	3	1-10492	34556	89	68	470	24	2012-08-02 10:31:53.750492
34560	4	1-8574	34559	50	20	2302	65	2012-08-02 10:31:53.750492
34561	4	1-9371	34559	20	10	1678	97	2012-08-02 10:31:53.750492
34562	4	1-9818	34559	10	89	1735	48	2012-08-02 10:31:53.750492
34559	4	1-10492	34559	89	50	1641	24	2012-08-02 10:31:53.750492
34564	4	1-9363	34563	80	3	227	74	2012-08-02 10:31:53.750492
34565	4	1-9754	34563	3	2	206	27	2012-08-02 10:31:53.750492
34566	4	1-10457	34563	2	89	436	48	2012-08-02 10:31:53.750492
34563	4	1-10492	34563	89	80	413	24	2012-08-02 10:31:53.750492
34568	2	1-10330	34567	1	92	1923	41	2012-08-02 10:31:54.948025
34567	2	1-10495	34567	92	1	6199	19	2012-08-02 10:31:54.948025
34570	4	1-8047	34569	56	39	528	92	2012-08-02 10:31:56.130676
34571	4	1-10147	34569	39	21	368	26	2012-08-02 10:31:56.130676
34572	4	1-8624	34569	21	32	225	71	2012-08-02 10:31:56.130676
34569	4	1-10498	34569	32	56	376	22	2012-08-02 10:31:56.130676
34574	4	1-10463	34573	93	57	842	28	2012-08-02 10:32:00.096984
34575	4	1-8651	34573	57	64	429	96	2012-08-02 10:32:00.096984
34576	4	1-9844	34573	64	23	675	20	2012-08-02 10:32:00.096984
34573	4	1-10508	34573	23	93	1064	90	2012-08-02 10:32:00.096984
34578	4	1-9936	34577	69	64	551	46	2012-08-02 10:32:00.096984
34579	4	1-8036	34577	64	4	273	93	2012-08-02 10:32:00.096984
34580	4	1-9381	34577	4	23	173	20	2012-08-02 10:32:00.096984
34577	4	1-10508	34577	23	69	289	90	2012-08-02 10:32:00.096984
34582	3	1-10391	34581	30	63	1290	91	2012-08-02 10:32:00.494995
34583	3	1-587	34581	63	9	754	48	2012-08-02 10:32:00.494995
34581	3	1-10509	34581	9	30	421	63	2012-08-02 10:32:00.494995
34585	4	1-9570	34584	66	17	437	81	2012-08-02 10:32:02.462459
34586	4	1-6372	34584	17	39	535	92	2012-08-02 10:32:02.462459
34587	4	1-10147	34584	39	98	402	26	2012-08-02 10:32:02.462459
34584	4	1-10514	34584	98	66	459	43	2012-08-02 10:32:02.462459
34589	3	1-9955	34588	94	57	2969	23	2012-08-02 10:32:04.417905
34590	3	1-7882	34588	57	74	699	69	2012-08-02 10:32:04.417905
34588	3	1-10519	34588	74	94	6912	80	2012-08-02 10:32:04.417905
34592	4	1-9370	34591	95	53	766	83	2012-08-02 10:32:04.805169
34593	4	1-9824	34591	53	84	713	48	2012-08-02 10:32:04.805169
34594	4	1-9743	34591	84	17	753	40	2012-08-02 10:32:04.805169
34591	4	1-10520	34591	17	95	601	3	2012-08-02 10:32:04.805169
34596	3	1-8889	34595	90	59	2180	93	2012-08-02 10:32:04.805169
34597	3	1-10212	34595	59	17	2735	40	2012-08-02 10:32:04.805169
34595	3	1-10520	34595	17	90	2827	3	2012-08-02 10:32:04.805169
34599	4	1-10002	34598	80	59	97	49	2012-08-02 10:32:04.805169
34600	4	1-9120	34598	59	78	52	23	2012-08-02 10:32:04.805169
34601	4	1-10357	34598	78	17	100	40	2012-08-02 10:32:04.805169
34598	4	1-10520	34598	17	80	105	3	2012-08-02 10:32:04.805169
34603	2	1-8985	34602	51	32	2682	51	2012-08-02 10:32:05.601007
34602	2	1-10522	34602	32	51	1827	3	2012-08-02 10:32:05.601007
34605	5	1-8829	34604	86	11	4098	85	2012-08-02 10:32:05.998759
34606	5	1-10444	34604	11	85	3937	17	2012-08-02 10:32:05.998759
34607	5	1-9942	34604	85	75	2041	55	2012-08-02 10:32:05.998759
34608	5	1-10373	34604	75	6	1464	99	2012-08-02 10:32:05.998759
34604	5	1-10523	34604	6	86	3152	31	2012-08-02 10:32:05.998759
34610	5	1-9716	34609	94	77	1783	80	2012-08-02 10:32:05.998759
34611	5	1-9928	34609	77	62	1316	57	2012-08-02 10:32:05.998759
34612	5	1-10048	34609	62	75	836	55	2012-08-02 10:32:05.998759
34613	5	1-10373	34609	75	6	611	99	2012-08-02 10:32:05.998759
34609	5	1-10523	34609	6	94	1342	31	2012-08-02 10:32:05.998759
34615	4	1-10367	34614	9	83	6796	73	2012-08-02 10:32:06.408228
34616	4	1-10260	34614	83	29	6265	25	2012-08-02 10:32:06.408228
34617	4	1-9064	34614	29	25	2335	47	2012-08-02 10:32:06.408228
34614	4	1-10524	34614	25	9	3681	19	2012-08-02 10:32:06.408228
34619	3	1-9995	34618	63	75	3945	5	2012-08-02 10:32:07.181979
34620	3	1-9198	34618	75	84	2365	40	2012-08-02 10:32:07.181979
34618	3	1-10526	34618	84	63	1842	77	2012-08-02 10:32:07.181979
34622	4	1-10104	34621	25	65	376	61	2012-08-02 10:32:07.181979
34623	4	1-9932	34621	65	55	197	89	2012-08-02 10:32:07.181979
34624	4	1-10470	34621	55	84	929	40	2012-08-02 10:32:07.181979
34621	4	1-10526	34621	84	25	750	77	2012-08-02 10:32:07.181979
34626	3	1-6833	34625	58	42	1020	2	2012-08-02 10:32:07.571197
34627	3	1-10227	34625	42	91	934	10	2012-08-02 10:32:07.571197
34625	3	1-10527	34625	91	58	1802	60	2012-08-02 10:32:07.571197
34629	4	1-9522	34628	4	37	1873	44	2012-08-02 10:32:07.571197
34630	4	1-9908	34628	37	42	1120	2	2012-08-02 10:32:07.571197
34631	4	1-10227	34628	42	91	1097	10	2012-08-02 10:32:07.571197
34628	4	1-10527	34628	91	4	2263	60	2012-08-02 10:32:07.571197
34633	3	1-10303	34632	81	77	2939	96	2012-08-02 10:32:10.367016
34634	3	1-10425	34632	77	28	3134	87	2012-08-02 10:32:10.367016
34632	3	1-10534	34632	28	81	2914	78	2012-08-02 10:32:10.367016
34636	4	1-9960	34635	95	3	4834	97	2012-08-02 10:32:10.367016
34637	4	1-9076	34635	3	31	2424	95	2012-08-02 10:32:10.367016
34638	4	1-10272	34635	31	28	2708	87	2012-08-02 10:32:10.367016
34635	4	1-10534	34635	28	95	2875	78	2012-08-02 10:32:10.367016
34640	3	1-10303	34639	81	51	1263	96	2012-08-02 10:32:10.367016
34641	3	1-9544	34639	51	28	889	87	2012-08-02 10:32:10.367016
34639	3	1-10534	34639	28	81	1017	78	2012-08-02 10:32:10.367016
34643	3	1-8780	34642	98	94	2852	80	2012-08-02 10:32:11.571656
34644	3	1-9955	34642	94	28	2611	23	2012-08-02 10:32:11.571656
34642	3	1-10537	34642	28	98	1725	45	2012-08-02 10:32:11.571656
34646	4	1-10297	34645	73	6	1612	99	2012-08-02 10:32:11.571656
34647	4	1-10523	34645	6	53	3499	31	2012-08-02 10:32:11.571656
34648	4	1-10144	34645	53	28	3290	23	2012-08-02 10:32:11.571656
34645	4	1-10537	34645	28	73	2321	45	2012-08-02 10:32:11.571656
34650	4	1-10439	34649	16	56	4607	96	2012-08-02 10:32:12.367358
34651	4	1-7323	34649	56	98	1288	99	2012-08-02 10:32:12.367358
34652	4	1-10083	34649	98	30	1520	26	2012-08-02 10:32:12.367358
34649	4	1-10539	34649	30	16	4819	64	2012-08-02 10:32:12.367358
34653	1	1-10083	\N	98	98	104	26	2012-08-02 10:32:12.367358
34655	4	1-9370	34654	95	53	6006	83	2012-08-02 10:35:32.861798
34656	4	1-9824	34654	53	9	5865	48	2012-08-02 10:35:32.861798
34657	4	1-10509	34654	9	97	2676	63	2012-08-02 10:35:32.861798
34654	4	1-10549	34654	97	95	4490	3	2012-08-02 10:35:32.861798
34659	4	1-9641	34658	17	76	4062	70	2012-08-02 10:35:32.861798
34660	4	1-7076	34658	76	9	1813	48	2012-08-02 10:35:32.861798
34661	4	1-10509	34658	9	97	987	63	2012-08-02 10:35:32.861798
34658	4	1-10549	34658	97	17	1976	3	2012-08-02 10:35:32.861798
34663	4	1-9650	34662	53	27	963	10	2012-08-02 10:35:33.108873
34664	4	1-10173	34662	27	4	378	55	2012-08-02 10:35:33.108873
34665	4	1-10433	34662	4	31	565	4	2012-08-02 10:35:33.108873
34662	4	1-10552	34662	31	53	1397	54	2012-08-02 10:35:33.108873
34667	4	1-4463	34666	97	85	737	17	2012-08-02 10:35:33.108873
34668	4	1-9942	34666	85	4	376	55	2012-08-02 10:35:33.108873
34669	4	1-10433	34666	4	31	859	4	2012-08-02 10:35:33.108873
34666	4	1-10552	34666	31	97	3241	54	2012-08-02 10:35:33.108873
34671	2	1-2738	34670	1	31	309	4	2012-08-02 10:35:33.108873
34670	2	1-10552	34670	31	1	1297	54	2012-08-02 10:35:33.108873
34673	4	1-8569	34672	68	86	376	30	2012-08-02 10:35:33.266173
34674	4	1-8848	34672	86	39	1321	92	2012-08-02 10:35:33.266173
34675	4	1-10147	34672	39	11	952	26	2012-08-02 10:35:33.266173
34672	4	1-10554	34672	11	68	1613	18	2012-08-02 10:35:33.266173
34677	4	1-3386	34676	53	10	1092	97	2012-08-02 10:35:33.345098
34678	4	1-9818	34676	10	93	1053	48	2012-08-02 10:35:33.345098
34679	4	1-10282	34676	93	51	925	26	2012-08-02 10:35:33.345098
34676	4	1-10555	34676	51	53	3060	25	2012-08-02 10:35:33.345098
34681	3	1-7882	34680	57	62	792	69	2012-08-02 10:35:33.49017
34682	3	1-9463	34680	62	79	161	16	2012-08-02 10:35:33.49017
34680	3	1-10557	34680	79	57	5490	23	2012-08-02 10:35:33.49017
34683	2	1-10563	34683	1	9	2363	41	2012-08-02 10:35:33.946754
34684	2	1-7403	34683	9	1	851	16	2012-08-02 10:35:33.946754
34686	3	1-10493	34685	2	56	1608	22	2012-08-02 10:35:34.256855
34687	3	1-8047	34685	56	95	1598	92	2012-08-02 10:35:34.256855
34685	3	1-10567	34685	95	2	1438	95	2012-08-02 10:35:34.256855
34689	3	1-9628	34688	57	14	181	71	2012-08-02 10:35:34.612016
34690	3	1-9930	34688	14	39	222	44	2012-08-02 10:35:34.612016
34688	3	1-10572	34688	39	57	1115	25	2012-08-02 10:35:34.612016
34692	3	1-7451	34691	80	84	1389	94	2012-08-02 10:35:34.689888
34693	3	1-10363	34691	84	56	1078	46	2012-08-02 10:35:34.689888
34691	3	1-10573	34691	56	80	4264	1	2012-08-02 10:35:34.689888
34695	4	1-7669	34694	21	31	1008	97	2012-08-02 10:35:34.689888
34696	4	1-8576	34694	31	41	344	78	2012-08-02 10:35:34.689888
34697	4	1-9889	34694	41	56	335	46	2012-08-02 10:35:34.689888
34694	4	1-10573	34694	56	21	1617	1	2012-08-02 10:35:34.689888
34699	3	1-9482	34698	26	25	425	43	2012-08-02 10:35:34.689888
34700	3	1-6425	34698	25	56	318	46	2012-08-02 10:35:34.689888
34698	3	1-10573	34698	56	26	1702	1	2012-08-02 10:35:34.689888
34702	4	1-8294	34701	63	89	638	77	2012-08-02 10:35:34.770229
34703	4	1-6860	34701	89	9	210	48	2012-08-02 10:35:34.770229
34704	4	1-10509	34701	9	4	86	63	2012-08-02 10:35:34.770229
34701	4	1-10574	34701	4	63	632	49	2012-08-02 10:35:34.770229
34706	3	1-7838	34705	20	9	1909	48	2012-08-02 10:35:34.770229
34707	3	1-10509	34705	9	4	794	63	2012-08-02 10:35:34.770229
34705	3	1-10574	34705	4	20	5954	49	2012-08-02 10:35:34.770229
34709	3	1-6560	34708	89	93	274	100	2012-08-02 10:35:35.253698
34710	3	1-10159	34708	93	33	271	52	2012-08-02 10:35:35.253698
34708	3	1-10580	34708	33	89	467	47	2012-08-02 10:35:35.253698
34712	3	1-7669	34711	21	20	648	97	2012-08-02 10:35:35.546002
34713	3	1-9160	34711	20	42	34	19	2012-08-02 10:35:35.546002
34711	3	1-10584	34711	42	21	4347	1	2012-08-02 10:35:35.546002
34715	4	1-9130	34714	62	49	581	75	2012-08-02 10:35:35.623817
34716	4	1-8702	34714	49	93	509	45	2012-08-02 10:35:35.623817
34717	4	1-9382	34714	93	68	357	6	2012-08-02 10:35:35.623817
34714	4	1-10585	34714	68	62	1695	86	2012-08-02 10:35:35.623817
34719	3	1-10265	34718	84	95	566	33	2012-08-02 10:35:35.715244
34720	3	1-10139	34718	95	2	1175	12	2012-08-02 10:35:35.715244
34718	3	1-10586	34718	2	84	2928	48	2012-08-02 10:35:35.715244
34722	4	1-10282	34721	93	5	4003	26	2012-08-02 10:35:35.715244
34723	4	1-3559	34721	5	95	780	33	2012-08-02 10:35:35.715244
34724	4	1-10139	34721	95	2	1679	12	2012-08-02 10:35:35.715244
34721	4	1-10586	34721	2	93	4337	48	2012-08-02 10:35:35.715244
34726	2	1-8944	34725	38	74	2774	38	2012-08-02 10:35:35.793414
34725	2	1-10587	34725	74	38	9816	3	2012-08-02 10:35:35.793414
34728	3	1-4678	34727	100	72	1843	83	2012-08-02 10:35:36.015305
34729	3	1-8437	34727	72	25	1386	14	2012-08-02 10:35:36.015305
34727	3	1-10590	34727	25	100	7274	42	2012-08-02 10:35:36.015305
34731	2	1-9463	34730	62	68	4127	16	2012-08-02 10:35:36.093351
34730	2	1-10591	34730	68	62	9036	69	2012-08-02 10:35:36.093351
34733	4	1-9998	34732	76	5	1489	24	2012-08-02 10:35:36.668705
34734	4	1-9578	34732	5	75	396	55	2012-08-02 10:35:36.668705
34735	4	1-10373	34732	75	10	296	99	2012-08-02 10:35:36.668705
34732	4	1-10599	34732	10	76	1854	28	2012-08-02 10:35:36.668705
34737	3	1-10367	34736	9	83	915	73	2012-08-02 10:35:36.746639
34738	3	1-10260	34736	83	2	615	25	2012-08-02 10:35:36.746639
34736	3	1-10600	34736	2	9	680	19	2012-08-02 10:35:36.746639
34740	3	1-7945	34739	87	54	3236	98	2012-08-02 10:35:36.746639
34741	3	1-10403	34739	54	2	2629	25	2012-08-02 10:35:36.746639
34739	3	1-10600	34739	2	87	3596	19	2012-08-02 10:35:36.746639
34743	3	1-7004	34742	36	15	1887	14	2012-08-02 10:35:36.746639
34744	3	1-10413	34742	15	2	3470	25	2012-08-02 10:35:36.746639
34742	3	1-10600	34742	2	36	5149	19	2012-08-02 10:35:36.746639
34746	3	1-10198	34745	42	56	1788	35	2012-08-02 10:35:37.024396
34747	3	1-10345	34745	56	15	1627	66	2012-08-02 10:35:37.024396
34745	3	1-10604	34745	15	42	2389	61	2012-08-02 10:35:37.024396
34749	3	1-9962	34748	81	61	3584	21	2012-08-02 10:35:37.102316
34750	3	1-8054	34748	61	67	2598	48	2012-08-02 10:35:37.102316
34748	3	1-10605	34748	67	81	3089	86	2012-08-02 10:35:37.102316
34752	4	1-9975	34751	50	62	3468	75	2012-08-02 10:35:37.324285
34753	4	1-10431	34751	62	89	3074	71	2012-08-02 10:35:37.324285
34754	4	1-10167	34751	89	88	4230	12	2012-08-02 10:35:37.324285
34751	4	1-10608	34751	88	50	3257	13	2012-08-02 10:35:37.324285
34756	2	1-10090	34755	35	12	661	34	2012-08-02 10:35:37.601238
34755	2	1-10612	34755	12	35	1150	31	2012-08-02 10:35:37.601238
34758	3	1-8829	34757	86	50	1747	85	2012-08-02 10:35:37.601238
34759	3	1-10161	34757	50	12	404	34	2012-08-02 10:35:37.601238
34757	3	1-10612	34757	12	86	1352	31	2012-08-02 10:35:37.601238
34761	4	1-10177	34760	79	1	1247	86	2012-08-02 10:35:37.757001
34762	4	1-9838	34760	1	74	873	34	2012-08-02 10:35:37.757001
34763	4	1-10171	34760	74	46	901	72	2012-08-02 10:35:37.757001
34760	4	1-10614	34760	46	79	772	19	2012-08-02 10:35:37.757001
34765	3	1-9770	34764	50	48	398	59	2012-08-02 10:35:37.757001
34766	3	1-9358	34764	48	46	362	72	2012-08-02 10:35:37.757001
34764	3	1-10614	34764	46	50	424	19	2012-08-02 10:35:37.757001
34768	4	1-7309	34767	45	29	1353	60	2012-08-02 10:35:37.757001
34769	4	1-10435	34767	29	74	699	34	2012-08-02 10:35:37.757001
34770	4	1-10171	34767	74	46	1064	72	2012-08-02 10:35:37.757001
34767	4	1-10614	34767	46	45	1345	19	2012-08-02 10:35:37.757001
34772	4	1-9902	34771	64	68	304	6	2012-08-02 10:35:37.938866
34773	4	1-10585	34771	68	26	764	86	2012-08-02 10:35:37.938866
34774	4	1-10607	34771	26	96	2398	70	2012-08-02 10:35:37.938866
34771	4	1-10616	34771	96	64	581	62	2012-08-02 10:35:37.938866
34776	2	1-9782	34775	81	96	3522	70	2012-08-02 10:35:37.938866
34775	2	1-10616	34775	96	81	1530	62	2012-08-02 10:35:37.938866
34778	3	1-10617	34777	2	43	2225	47	2012-08-02 10:35:38.381654
34779	3	1-7717	34777	43	32	663	76	2012-08-02 10:35:38.381654
34777	3	1-10622	34777	32	2	1607	91	2012-08-02 10:35:38.381654
34781	4	1-10197	34780	29	71	2066	100	2012-08-02 10:35:39.223099
34782	4	1-9764	34780	71	97	1319	20	2012-08-02 10:35:39.223099
34783	4	1-10560	34780	97	31	1975	90	2012-08-02 10:35:39.223099
34780	4	1-10633	34780	31	29	1935	45	2012-08-02 10:35:39.223099
34785	4	1-10194	34784	51	5	663	61	2012-08-02 10:35:39.223099
34786	4	1-4498	34784	5	97	287	20	2012-08-02 10:35:39.223099
34787	4	1-10560	34784	97	31	573	90	2012-08-02 10:35:39.223099
34784	4	1-10633	34784	31	51	749	45	2012-08-02 10:35:39.223099
34789	3	1-10544	34788	6	39	1270	93	2012-08-02 10:35:39.455595
34790	3	1-8231	34788	39	38	633	97	2012-08-02 10:35:39.455595
34788	3	1-10636	34788	38	6	1003	51	2012-08-02 10:35:39.455595
34792	3	1-8968	34791	6	83	1043	36	2012-08-02 10:35:39.677649
34793	3	1-7135	34791	83	64	513	59	2012-08-02 10:35:39.677649
34791	3	1-10639	34791	64	6	1529	21	2012-08-02 10:35:39.677649
34795	4	1-1506	34794	5	13	439	33	2012-08-02 10:35:39.75508
34796	4	1-10628	34794	13	74	610	6	2012-08-02 10:35:39.75508
34797	4	1-10340	34794	74	64	715	32	2012-08-02 10:35:39.75508
34794	4	1-10640	34794	64	5	1358	8	2012-08-02 10:35:39.75508
34799	3	1-10446	34798	91	74	1655	6	2012-08-02 10:35:39.75508
34800	3	1-10340	34798	74	64	1987	32	2012-08-02 10:35:39.75508
34798	3	1-10640	34798	64	91	3865	8	2012-08-02 10:35:39.75508
34802	4	1-10472	34801	16	98	1072	27	2012-08-02 10:35:58.701003
34803	4	1-9814	34801	98	83	1142	12	2012-08-02 10:35:58.701003
34804	4	1-10376	34801	83	67	317	94	2012-08-02 10:35:58.701003
34801	4	1-10643	34801	67	16	945	32	2012-08-02 10:35:58.701003
34806	4	1-10640	34805	64	93	4	8	2012-08-02 10:35:58.701003
34807	4	1-9636	34805	93	36	2	97	2012-08-02 10:35:58.701003
34808	4	1-10582	34805	36	67	1	94	2012-08-02 10:35:58.701003
34805	4	1-10643	34805	67	64	3	32	2012-08-02 10:35:58.701003
34810	4	1-8088	34809	91	68	120	17	2012-08-02 10:35:59.134291
34811	4	1-9980	34809	68	82	101	46	2012-08-02 10:35:59.134291
34812	4	1-7606	34809	82	69	53	6	2012-08-02 10:35:59.134291
34809	4	1-10644	34809	69	91	152	77	2012-08-02 10:35:59.134291
34814	4	1-8877	34813	82	2	1423	91	2012-08-02 10:35:59.134291
34815	4	1-10617	34813	2	51	2445	47	2012-08-02 10:35:59.134291
34816	4	1-10355	34813	51	69	1366	6	2012-08-02 10:35:59.134291
34813	4	1-10644	34813	69	82	4537	77	2012-08-02 10:35:59.134291
34818	4	1-10104	34817	25	70	2694	61	2012-08-02 10:35:59.134291
34819	4	1-7704	34817	70	95	2086	57	2012-08-02 10:35:59.134291
34820	4	1-10468	34817	95	69	1228	6	2012-08-02 10:35:59.134291
34817	4	1-10644	34817	69	25	4252	77	2012-08-02 10:35:59.134291
34822	2	1-8251	34821	46	32	1604	83	2012-08-02 10:35:59.532936
34821	2	1-10645	34821	32	46	2809	39	2012-08-02 10:35:59.532936
34824	4	1-8988	34823	81	16	2180	27	2012-08-02 10:36:00.328563
34825	4	1-10151	34823	16	42	2014	35	2012-08-02 10:36:00.328563
34826	4	1-6590	34823	42	89	1312	29	2012-08-02 10:36:00.328563
34823	4	1-10647	34823	89	81	2861	18	2012-08-02 10:36:00.328563
34828	4	1-10199	34827	62	72	1365	16	2012-08-02 10:36:00.328563
34829	4	1-10477	34827	72	72	1379	60	2012-08-02 10:36:00.328563
34830	4	1-10407	34827	72	89	585	29	2012-08-02 10:36:00.328563
34827	4	1-10647	34827	89	62	1356	18	2012-08-02 10:36:00.328563
34832	4	1-10285	34831	43	28	879	8	2012-08-02 10:36:00.328563
34833	4	1-7782	34831	28	20	944	5	2012-08-02 10:36:00.328563
34834	4	1-10203	34831	20	89	519	29	2012-08-02 10:36:00.328563
34831	4	1-10647	34831	89	43	1250	18	2012-08-02 10:36:00.328563
34836	4	1-10199	34835	62	41	1332	16	2012-08-02 10:36:00.328563
34837	4	1-9997	34835	41	42	622	35	2012-08-02 10:36:00.328563
34838	4	1-6590	34835	42	89	473	29	2012-08-02 10:36:00.328563
34835	4	1-10647	34835	89	62	1204	18	2012-08-02 10:36:00.328563
34840	3	1-9439	34839	52	93	1907	70	2012-08-02 10:36:00.328563
34841	3	1-5594	34839	93	89	708	29	2012-08-02 10:36:00.328563
34839	3	1-10647	34839	89	52	1886	18	2012-08-02 10:36:00.328563
34843	4	1-10500	34842	5	91	239	10	2012-08-02 10:36:00.328563
34844	4	1-10527	34842	91	72	420	60	2012-08-02 10:36:00.328563
34845	4	1-10407	34842	72	89	173	29	2012-08-02 10:36:00.328563
34842	4	1-10647	34842	89	5	463	18	2012-08-02 10:36:00.328563
34846	1	1-3353	\N	63	63	3714	16	2012-08-02 10:36:00.328563
34847	1	1-2746	\N	26	26	7122	16	2012-08-02 10:36:00.328563
34849	4	1-9718	34848	1	97	2286	20	2012-08-02 10:36:00.738538
34850	4	1-10560	34848	97	93	2948	90	2012-08-02 10:36:00.738538
34851	4	1-10463	34848	93	8	1701	28	2012-08-02 10:36:00.738538
34848	4	1-10648	34848	8	1	5739	56	2012-08-02 10:36:00.738538
34853	4	1-9718	34852	1	23	637	20	2012-08-02 10:36:00.738538
34854	4	1-10508	34852	23	93	753	90	2012-08-02 10:36:00.738538
34855	4	1-10463	34852	93	8	447	28	2012-08-02 10:36:00.738538
34852	4	1-10648	34852	8	1	1553	56	2012-08-02 10:36:00.738538
34857	4	1-9879	34856	62	50	1146	50	2012-08-02 10:36:00.738538
34858	4	1-9835	34856	50	44	1168	41	2012-08-02 10:36:00.738538
34859	4	1-9343	34856	44	8	213	28	2012-08-02 10:36:00.738538
34856	4	1-10648	34856	8	62	897	56	2012-08-02 10:36:00.738538
34861	4	1-10282	34860	93	7	2018	26	2012-08-02 10:36:01.899372
34862	4	1-9508	34860	7	37	3012	98	2012-08-02 10:36:01.899372
34863	4	1-10261	34860	37	19	2636	24	2012-08-02 10:36:01.899372
34860	4	1-10651	34860	19	93	2311	48	2012-08-02 10:36:01.899372
34865	4	1-10509	34864	9	87	590	63	2012-08-02 10:36:02.297522
34866	4	1-9189	34864	87	49	675	73	2012-08-02 10:36:02.297522
34867	4	1-9717	34864	49	42	460	46	2012-08-02 10:36:02.297522
34864	4	1-10652	34864	42	9	1447	48	2012-08-02 10:36:02.297522
34869	4	1-10605	34868	67	39	4124	86	2012-08-02 10:36:02.297522
34870	4	1-10331	34868	39	37	5803	60	2012-08-02 10:36:02.297522
34871	4	1-10126	34868	37	42	1579	46	2012-08-02 10:36:02.297522
34868	4	1-10652	34868	42	67	5020	48	2012-08-02 10:36:02.297522
34873	4	1-10492	34872	89	38	2387	24	2012-08-02 10:36:02.297522
34874	4	1-8619	34872	38	49	1141	73	2012-08-02 10:36:02.297522
34875	4	1-9717	34872	49	42	917	46	2012-08-02 10:36:02.297522
34872	4	1-10652	34872	42	89	3400	48	2012-08-02 10:36:02.297522
34877	4	1-10533	34876	100	94	1668	37	2012-08-02 10:36:03.093151
34878	4	1-9848	34876	94	36	854	50	2012-08-02 10:36:03.093151
34879	4	1-10225	34876	36	1	310	89	2012-08-02 10:36:03.093151
34876	4	1-10654	34876	1	100	1509	15	2012-08-02 10:36:03.093151
34881	4	1-10627	34880	35	49	5160	4	2012-08-02 10:36:03.480574
34882	4	1-8866	34880	49	37	3191	71	2012-08-02 10:36:03.480574
34883	4	1-10258	34880	37	14	2974	51	2012-08-02 10:36:03.480574
34880	4	1-10655	34880	14	35	4675	35	2012-08-02 10:36:03.480574
34885	4	1-10345	34884	56	41	1667	66	2012-08-02 10:36:03.480574
34886	4	1-7671	34884	41	37	993	71	2012-08-02 10:36:03.480574
34887	4	1-10258	34884	37	14	1079	51	2012-08-02 10:36:03.480574
34884	4	1-10655	34884	14	56	1977	35	2012-08-02 10:36:03.480574
34889	4	1-8036	34888	64	5	2881	93	2012-08-02 10:36:03.878436
34890	4	1-10402	34888	5	57	3493	79	2012-08-02 10:36:03.878436
34888	4	1-10656	34888	84	64	7107	46	2012-08-02 10:36:03.878436
34891	4	1-8540	34888	57	84	1548	71	2012-08-02 10:36:03.878436
34893	3	1-10125	34892	17	71	112	61	2012-08-02 10:36:03.878436
34894	3	1-7327	34892	71	84	87	71	2012-08-02 10:36:03.878436
34892	3	1-10656	34892	84	17	429	46	2012-08-02 10:36:03.878436
34896	4	1-10481	34895	48	29	1589	53	2012-08-02 10:36:04.276369
34897	4	1-8974	34895	29	15	352	66	2012-08-02 10:36:04.276369
34898	4	1-10604	34895	15	14	241	61	2012-08-02 10:36:04.276369
34895	4	1-10657	34895	14	48	5455	87	2012-08-02 10:36:04.276369
34900	4	1-10303	34899	81	42	1499	96	2012-08-02 10:36:04.663382
34901	4	1-8163	34899	42	51	1854	16	2012-08-02 10:36:04.663382
34902	4	1-10314	34899	51	76	1618	84	2012-08-02 10:36:04.663382
34899	4	1-10658	34899	76	81	1100	78	2012-08-02 10:36:04.663382
34904	4	1-10075	34903	59	45	1039	81	2012-08-02 10:36:05.049976
34905	4	1-9999	34903	45	67	1030	16	2012-08-02 10:36:05.049976
34906	4	1-10576	34903	67	26	1277	15	2012-08-02 10:36:05.049976
34903	4	1-10659	34903	26	59	3012	90	2012-08-02 10:36:05.049976
34908	4	1-10459	34907	14	52	1058	98	2012-08-02 10:36:05.049976
34909	4	1-10359	34907	52	66	1694	53	2012-08-02 10:36:05.049976
34910	4	1-10074	34907	66	26	1008	15	2012-08-02 10:36:05.049976
34907	4	1-10659	34907	26	14	2446	90	2012-08-02 10:36:05.049976
34912	4	1-10435	34911	29	12	774	34	2012-08-02 10:36:05.437352
34913	4	1-10612	34911	12	38	2004	31	2012-08-02 10:36:05.437352
34914	4	1-8672	34911	38	56	288	52	2012-08-02 10:36:05.437352
34911	4	1-10660	34911	56	29	1895	60	2012-08-02 10:36:05.437352
34916	4	1-10435	34915	29	37	2453	34	2012-08-02 10:36:05.437352
34917	4	1-10595	34915	37	29	1203	22	2012-08-02 10:36:05.437352
34918	4	1-8361	34915	29	56	677	52	2012-08-02 10:36:05.437352
34915	4	1-10660	34915	56	29	4795	60	2012-08-02 10:36:05.437352
34920	3	1-8830	34919	10	46	212	81	2012-08-02 10:36:05.437352
34921	3	1-9758	34919	46	56	106	52	2012-08-02 10:36:05.437352
34919	3	1-10660	34919	56	10	809	60	2012-08-02 10:36:05.437352
34923	4	1-10314	34922	51	76	1018	84	2012-08-02 10:36:06.995377
34924	4	1-10658	34922	76	95	479	78	2012-08-02 10:36:06.995377
34925	4	1-10476	34922	95	90	1655	27	2012-08-02 10:36:06.995377
34922	4	1-10664	34922	90	51	1686	16	2012-08-02 10:36:06.995377
34927	4	1-10159	34926	93	33	1806	52	2012-08-02 10:36:07.382007
34928	4	1-10580	34926	33	27	2540	47	2012-08-02 10:36:07.382007
34929	4	1-10310	34926	27	77	4058	7	2012-08-02 10:36:07.382007
34926	4	1-10665	34926	77	93	2238	100	2012-08-02 10:36:07.382007
34931	4	1-8575	34930	26	96	1575	69	2012-08-02 10:36:07.382007
34932	4	1-10399	34930	96	67	1468	18	2012-08-02 10:36:07.382007
34933	4	1-10447	34930	67	77	3517	7	2012-08-02 10:36:07.382007
34930	4	1-10665	34930	77	26	2283	100	2012-08-02 10:36:07.382007
34935	5	1-9282	34934	11	14	2624	53	2012-08-02 10:36:08.166636
34936	5	1-10182	34934	14	67	622	94	2012-08-02 10:36:08.166636
34937	5	1-10643	34934	67	64	1988	32	2012-08-02 10:36:08.166636
34938	5	1-10640	34934	64	97	3675	8	2012-08-02 10:36:08.166636
34934	5	1-10667	34934	97	11	5367	54	2012-08-02 10:36:08.166636
34940	4	1-10639	34939	64	6	124	21	2012-08-02 10:36:08.564494
34941	4	1-8968	34939	6	74	75	36	2012-08-02 10:36:08.564494
34942	4	1-9527	34939	74	57	70	81	2012-08-02 10:36:08.564494
34939	4	1-10668	34939	57	64	47	59	2012-08-02 10:36:08.564494
34944	4	1-10639	34943	64	53	6197	21	2012-08-02 10:36:08.564494
34945	4	1-8882	34943	53	93	3368	77	2012-08-02 10:36:08.564494
34946	4	1-10312	34943	93	57	2461	81	2012-08-02 10:36:08.564494
34943	4	1-10668	34943	57	64	1973	59	2012-08-02 10:36:08.564494
34947	1	1-185	\N	69	69	558	81	2012-08-02 10:36:08.564494
34949	4	1-4678	34948	100	56	318	83	2012-08-02 10:36:08.962545
34950	4	1-10611	34948	56	50	566	50	2012-08-02 10:36:08.962545
34951	4	1-9835	34948	50	41	760	41	2012-08-02 10:36:08.962545
34948	4	1-10669	34948	41	100	949	42	2012-08-02 10:36:08.962545
34953	3	1-10599	34952	10	48	1207	28	2012-08-02 10:36:11.326725
34954	3	1-6942	34952	48	16	635	31	2012-08-02 10:36:11.326725
34952	3	1-10675	34952	16	10	286	99	2012-08-02 10:36:11.326725
34956	4	1-8748	34955	27	52	2375	98	2012-08-02 10:36:11.326725
34957	4	1-10359	34955	52	61	5630	53	2012-08-02 10:36:11.326725
34958	4	1-10619	34955	61	16	4206	31	2012-08-02 10:36:11.326725
34955	4	1-10675	34955	16	27	2533	99	2012-08-02 10:36:11.326725
34960	4	1-7851	34959	57	42	1819	2	2012-08-02 10:36:12.146332
34961	4	1-10227	34959	42	42	1624	10	2012-08-02 10:36:12.146332
34962	4	1-9155	34959	42	29	1591	40	2012-08-02 10:36:12.146332
34959	4	1-10677	34959	29	57	1676	84	2012-08-02 10:36:12.146332
34964	4	1-10568	34963	32	36	2973	50	2012-08-02 10:36:12.146332
34965	4	1-10225	34963	36	55	888	89	2012-08-02 10:36:12.146332
34966	4	1-10470	34963	55	29	4488	40	2012-08-02 10:36:12.146332
34963	4	1-10677	34963	29	32	4835	84	2012-08-02 10:36:12.146332
34968	4	1-9444	34967	37	96	1197	69	2012-08-02 10:36:12.146332
34969	4	1-10399	34967	96	92	1106	18	2012-08-02 10:36:12.146332
34970	4	1-7967	34967	92	29	1521	40	2012-08-02 10:36:12.146332
34967	4	1-10677	34967	29	37	1810	84	2012-08-02 10:36:12.146332
34972	4	1-10549	34971	97	84	2690	3	2012-08-02 10:36:12.544092
34973	4	1-9601	34971	84	80	1979	8	2012-08-02 10:36:12.544092
34974	4	1-8484	34971	80	66	719	64	2012-08-02 10:36:12.544092
34971	4	1-10678	34971	66	97	1909	63	2012-08-02 10:36:12.544092
34976	4	1-8815	34975	91	6	2146	51	2012-08-02 10:36:12.544092
34977	4	1-10544	34975	6	31	2236	93	2012-08-02 10:36:12.544092
34978	4	1-10516	34975	31	66	1117	64	2012-08-02 10:36:12.544092
34975	4	1-10678	34975	66	91	3505	63	2012-08-02 10:36:12.544092
34980	4	1-9481	34979	26	60	1220	35	2012-08-02 10:36:12.544092
34981	4	1-1317	34979	60	40	459	43	2012-08-02 10:36:12.544092
34982	4	1-10300	34979	40	66	258	64	2012-08-02 10:36:12.544092
34979	4	1-10678	34979	66	26	858	63	2012-08-02 10:36:12.544092
34984	3	1-9695	34983	10	14	1354	85	2012-08-02 10:36:12.932043
34985	3	1-8197	34983	14	71	412	88	2012-08-02 10:36:12.932043
34983	3	1-10679	34983	71	10	2434	79	2012-08-02 10:36:12.932043
34987	4	1-10568	34986	32	32	88	50	2012-08-02 10:36:13.727423
34988	4	1-8237	34986	32	31	39	49	2012-08-02 10:36:13.727423
34989	4	1-10462	34986	31	94	85	69	2012-08-02 10:36:13.727423
34986	4	1-10681	34986	94	32	143	84	2012-08-02 10:36:13.727423
34991	2	1-9444	34990	37	94	760	69	2012-08-02 10:36:13.727423
34990	2	1-10681	34990	94	37	1428	84	2012-08-02 10:36:13.727423
34993	4	1-9365	34992	92	16	742	77	2012-08-02 10:36:13.727423
34994	4	1-8605	34992	16	31	256	49	2012-08-02 10:36:13.727423
34995	4	1-10462	34992	31	94	639	69	2012-08-02 10:36:13.727423
34992	4	1-10681	34992	94	92	1242	84	2012-08-02 10:36:13.727423
34997	3	1-10658	34996	76	34	2929	78	2012-08-02 10:36:13.727423
34998	3	1-9024	34996	34	94	2051	69	2012-08-02 10:36:13.727423
34996	3	1-10681	34996	94	76	4817	84	2012-08-02 10:36:13.727423
35000	4	1-3671	34999	33	19	4811	37	2012-08-02 10:36:14.490202
35001	4	1-10420	34999	19	15	2532	55	2012-08-02 10:36:14.490202
35002	4	1-10356	34999	15	91	6593	88	2012-08-02 10:36:14.490202
34999	4	1-10683	34999	91	33	8829	71	2012-08-02 10:36:14.490202
35004	4	1-2262	35003	3	31	15	49	2012-08-02 10:36:15.661498
35005	4	1-10462	35003	31	95	42	69	2012-08-02 10:36:15.661498
35006	4	1-9542	35003	95	65	34	71	2012-08-02 10:36:15.661498
35003	4	1-10686	35003	65	3	58	91	2012-08-02 10:36:15.661498
35008	4	1-8282	35007	46	67	142	18	2012-08-02 10:36:15.661498
35009	4	1-10447	35007	67	11	342	7	2012-08-02 10:36:15.661498
35010	4	1-9116	35007	11	65	177	71	2012-08-02 10:36:15.661498
35007	4	1-10686	35007	65	46	326	91	2012-08-02 10:36:15.661498
35011	1	1-7028	\N	91	91	663	7	2012-08-02 10:36:15.661498
35013	3	1-10596	35012	16	89	1462	48	2012-08-02 10:36:16.07037
35014	3	1-10492	35012	89	46	723	24	2012-08-02 10:36:16.07037
35012	3	1-10687	35012	46	16	1783	67	2012-08-02 10:36:16.07037
35016	4	1-5544	35015	69	7	1116	26	2012-08-02 10:36:16.07037
35017	4	1-9508	35015	7	37	1483	98	2012-08-02 10:36:16.07037
35018	4	1-10261	35015	37	46	1156	24	2012-08-02 10:36:16.07037
35015	4	1-10687	35015	46	69	4496	67	2012-08-02 10:36:16.07037
35020	4	1-9176	35019	24	5	2491	18	2012-08-02 10:36:16.07037
35021	4	1-10500	35019	5	4	1403	10	2012-08-02 10:36:16.07037
35022	4	1-10317	35019	4	46	775	24	2012-08-02 10:36:16.07037
35019	4	1-10687	35019	46	24	3309	67	2012-08-02 10:36:16.07037
35023	1	1-10	\N	10	10	1692	10	2012-08-02 10:36:16.07037
35025	2	1-4742	35024	30	3	1086	13	2012-08-02 10:36:16.479713
35024	2	1-10688	35024	3	30	3432	1	2012-08-02 10:36:16.479713
35027	4	1-9407	35026	8	4	836	55	2012-08-02 10:36:17.253241
35028	4	1-10433	35026	4	100	1106	4	2012-08-02 10:36:17.253241
35029	4	1-10057	35026	100	98	622	100	2012-08-02 10:36:17.253241
35026	4	1-10690	35026	98	8	3566	67	2012-08-02 10:36:17.253241
35031	4	1-10266	35030	98	5	794	92	2012-08-02 10:36:18.424417
35032	4	1-10291	35030	5	19	228	43	2012-08-02 10:36:18.424417
35033	4	1-10629	35030	19	93	193	31	2012-08-02 10:36:18.424417
35030	4	1-10693	35030	93	98	826	76	2012-08-02 10:36:18.424417
35035	4	1-10136	35034	97	98	151	26	2012-08-02 10:36:18.424417
35036	4	1-10514	35034	98	19	96	43	2012-08-02 10:36:18.424417
35037	4	1-10629	35034	19	93	84	31	2012-08-02 10:36:18.424417
35034	4	1-10693	35034	93	97	370	76	2012-08-02 10:36:18.424417
35039	4	1-7540	35038	30	5	623	92	2012-08-02 10:36:18.424417
35040	4	1-10291	35038	5	19	203	43	2012-08-02 10:36:18.424417
35041	4	1-10629	35038	19	93	195	31	2012-08-02 10:36:18.424417
35038	4	1-10693	35038	93	30	948	76	2012-08-02 10:36:18.424417
35043	3	1-7043	35042	97	19	882	43	2012-08-02 10:36:18.424417
35044	3	1-10629	35042	19	93	851	31	2012-08-02 10:36:18.424417
35042	3	1-10693	35042	93	97	4152	76	2012-08-02 10:36:18.424417
35046	4	1-6121	35045	58	5	1427	92	2012-08-02 10:36:18.424417
35047	4	1-10291	35045	5	19	477	43	2012-08-02 10:36:18.424417
35048	4	1-10629	35045	19	93	471	31	2012-08-02 10:36:18.424417
35045	4	1-10693	35045	93	58	2350	76	2012-08-02 10:36:18.424417
35050	3	1-9695	35049	10	80	286	85	2012-08-02 10:36:19.209058
35051	3	1-10343	35049	80	28	481	2	2012-08-02 10:36:19.209058
35049	3	1-10695	35049	28	10	283	79	2012-08-02 10:36:19.209058
35053	4	1-10456	35052	1	29	389	40	2012-08-02 10:36:19.209058
35054	4	1-10677	35052	29	57	513	84	2012-08-02 10:36:19.209058
35055	4	1-7851	35052	57	28	698	2	2012-08-02 10:36:19.209058
35052	4	1-10695	35052	28	1	455	79	2012-08-02 10:36:19.209058
35057	4	1-10597	35056	80	63	569	18	2012-08-02 10:36:19.607106
35058	4	1-5837	35056	63	76	191	84	2012-08-02 10:36:19.607106
35059	4	1-10658	35056	76	100	122	78	2012-08-02 10:36:19.607106
35056	4	1-10696	35056	100	80	450	1	2012-08-02 10:36:19.607106
35061	3	1-7669	35060	21	31	1639	97	2012-08-02 10:36:19.607106
35062	3	1-8576	35060	31	100	624	78	2012-08-02 10:36:19.607106
35060	3	1-10696	35060	100	21	2357	1	2012-08-02 10:36:19.607106
35064	4	1-9523	35063	20	74	1303	11	2012-08-02 10:36:19.607106
35065	4	1-5806	35063	74	76	829	84	2012-08-02 10:36:19.607106
35066	4	1-10658	35063	76	100	543	78	2012-08-02 10:36:19.607106
35063	4	1-10696	35063	100	20	2052	1	2012-08-02 10:36:19.607106
35068	4	1-9523	35067	20	87	1300	11	2012-08-02 10:36:19.607106
35069	4	1-9447	35067	87	24	2352	23	2012-08-02 10:36:19.607106
35070	4	1-879	35067	24	100	513	78	2012-08-02 10:36:19.607106
35067	4	1-10696	35067	100	20	1993	1	2012-08-02 10:36:19.607106
35072	4	1-7417	35071	77	64	772	62	2012-08-02 10:36:19.607106
35073	4	1-9902	35071	64	65	826	6	2012-08-02 10:36:19.607106
35074	4	1-9715	35071	65	100	451	78	2012-08-02 10:36:19.607106
35071	4	1-10696	35071	100	77	1758	1	2012-08-02 10:36:19.607106
35076	4	1-10466	35075	79	64	2290	62	2012-08-02 10:36:20.016257
35077	4	1-10293	35075	64	43	2162	18	2012-08-02 10:36:20.016257
35078	4	1-10285	35075	43	81	1309	8	2012-08-02 10:36:20.016257
35075	4	1-10697	35075	81	79	4207	10	2012-08-02 10:36:20.016257
35080	3	1-10601	35079	76	15	1936	19	2012-08-02 10:36:20.436271
35081	3	1-10504	35079	15	95	2935	10	2012-08-02 10:36:20.436271
35079	3	1-10698	35079	95	76	2585	14	2012-08-02 10:36:20.436271
35083	4	1-10413	35082	15	60	2709	25	2012-08-02 10:36:20.436271
35084	4	1-10252	35082	60	28	949	63	2012-08-02 10:36:20.436271
35085	4	1-9682	35082	28	95	1737	10	2012-08-02 10:36:20.436271
35082	4	1-10698	35082	95	15	1539	14	2012-08-02 10:36:20.436271
35087	4	1-10601	35086	76	91	822	19	2012-08-02 10:36:20.436271
35088	4	1-3605	35086	91	53	1007	54	2012-08-02 10:36:20.436271
35089	4	1-9650	35086	53	95	1143	10	2012-08-02 10:36:20.436271
35086	4	1-10698	35086	95	76	1051	14	2012-08-02 10:36:20.436271
35091	4	1-9031	35090	14	41	3846	85	2012-08-02 10:36:20.834054
35092	4	1-10337	35090	41	79	2617	84	2012-08-02 10:36:20.834054
35093	4	1-8264	35090	79	13	1520	45	2012-08-02 10:36:20.834054
35090	4	1-10699	35090	13	14	4363	6	2012-08-02 10:36:20.834054
35095	4	1-10217	35094	49	69	4714	88	2012-08-02 10:36:20.834054
35096	4	1-8823	35094	69	28	1485	23	2012-08-02 10:36:20.834054
35097	4	1-10537	35094	28	13	1077	45	2012-08-02 10:36:20.834054
35094	4	1-10699	35094	13	49	3154	6	2012-08-02 10:36:20.834054
35099	4	1-8542	35098	41	10	1622	60	2012-08-02 10:36:20.834054
35100	4	1-8830	35098	10	22	517	81	2012-08-02 10:36:20.834054
35101	4	1-10646	35098	22	13	458	45	2012-08-02 10:36:20.834054
35098	4	1-10699	35098	13	41	1404	6	2012-08-02 10:36:20.834054
35103	2	1-10002	35102	80	30	6654	49	2012-08-02 10:36:21.221225
35102	2	1-10700	35102	30	80	6783	3	2012-08-02 10:36:21.221225
35105	3	1-8856	35104	67	1	1925	89	2012-08-02 10:36:23.154366
35106	3	1-10654	35104	1	69	5838	15	2012-08-02 10:36:23.154366
35104	3	1-10705	35104	69	67	2973	63	2012-08-02 10:36:23.154366
35108	4	1-9189	35107	87	49	1120	73	2012-08-02 10:36:23.154366
35109	4	1-9717	35107	49	38	913	46	2012-08-02 10:36:23.154366
35110	4	1-10583	35107	38	69	1298	15	2012-08-02 10:36:23.154366
35107	4	1-10705	35107	69	87	820	63	2012-08-02 10:36:23.154366
35112	4	1-9138	35111	82	48	1383	55	2012-08-02 10:36:23.154366
35113	4	1-9499	35111	48	67	1154	16	2012-08-02 10:36:23.154366
35114	4	1-10576	35111	67	69	2040	15	2012-08-02 10:36:23.154366
35111	4	1-10705	35111	69	82	1359	63	2012-08-02 10:36:23.154366
35116	4	1-9682	35115	28	91	545	10	2012-08-02 10:36:23.154366
35117	4	1-10527	35115	91	40	968	60	2012-08-02 10:36:23.154366
35118	4	1-7331	35115	40	69	465	15	2012-08-02 10:36:23.154366
35115	4	1-10705	35115	69	28	327	63	2012-08-02 10:36:23.154366
35120	3	1-9481	35119	26	37	1263	35	2012-08-02 10:36:23.154366
35121	3	1-10661	35119	37	69	1026	15	2012-08-02 10:36:23.154366
35119	3	1-10705	35119	69	26	752	63	2012-08-02 10:36:23.154366
35123	3	1-10555	35122	51	93	4444	25	2012-08-02 10:36:23.938959
35124	3	1-10386	35122	93	37	1365	96	2012-08-02 10:36:23.938959
35122	3	1-10707	35122	37	51	1795	26	2012-08-02 10:36:23.938959
35126	4	1-10514	35125	98	40	1725	43	2012-08-02 10:36:23.938959
35127	4	1-10300	35125	40	16	1120	64	2012-08-02 10:36:23.938959
35128	4	1-10439	35125	16	37	1100	96	2012-08-02 10:36:23.938959
35125	4	1-10707	35125	37	98	1773	26	2012-08-02 10:36:23.938959
35130	4	1-9717	35129	49	98	3860	46	2012-08-02 10:36:26.270339
35131	4	1-10662	35129	98	74	3200	34	2012-08-02 10:36:26.270339
35132	4	1-10171	35129	74	53	3969	72	2012-08-02 10:36:26.270339
35129	4	1-10713	35129	53	49	4988	73	2012-08-02 10:36:26.270339
35134	4	1-10544	35133	6	14	4054	93	2012-08-02 10:36:27.021387
35135	4	1-10692	35133	14	59	2741	11	2012-08-02 10:36:27.021387
35136	4	1-10187	35133	59	23	2617	10	2012-08-02 10:36:27.021387
35133	4	1-10715	35133	23	6	2922	51	2012-08-02 10:36:27.021387
35138	3	1-8811	35137	15	41	1220	85	2012-08-02 10:36:27.408411
35139	3	1-10337	35137	41	42	795	84	2012-08-02 10:36:27.408411
35137	3	1-10716	35137	42	15	753	29	2012-08-02 10:36:27.408411
35141	4	1-10646	35140	22	51	540	45	2012-08-02 10:36:28.170561
35142	4	1-10194	35140	51	71	389	61	2012-08-02 10:36:28.170561
35143	4	1-7327	35140	71	36	300	71	2012-08-02 10:36:28.170561
35140	4	1-10718	35140	36	22	818	81	2012-08-02 10:36:28.170561
35145	4	1-10646	35144	22	47	852	45	2012-08-02 10:36:28.170561
35146	4	1-10233	35144	47	41	897	66	2012-08-02 10:36:28.170561
35147	4	1-7671	35144	41	36	457	71	2012-08-02 10:36:28.170561
35144	4	1-10718	35144	36	22	1267	81	2012-08-02 10:36:28.170561
35149	3	1-10668	35148	57	53	3355	59	2012-08-02 10:36:28.170561
35150	3	1-8236	35148	53	36	1543	71	2012-08-02 10:36:28.170561
35148	3	1-10718	35148	36	57	4818	81	2012-08-02 10:36:28.170561
35152	4	1-8015	35151	86	33	1247	52	2012-08-02 10:36:30.106164
35153	4	1-10580	35151	33	36	2128	47	2012-08-02 10:36:30.106164
35154	4	1-5805	35151	36	71	939	79	2012-08-02 10:36:30.106164
35151	4	1-10723	35151	71	86	2123	12	2012-08-02 10:36:30.106164
35156	3	1-9846	35155	12	10	3541	17	2012-08-02 10:36:30.515001
35157	3	1-10186	35155	10	90	976	80	2012-08-02 10:36:30.515001
35155	3	1-10724	35155	90	12	2778	94	2012-08-02 10:36:30.515001
35159	4	1-10363	35158	84	98	4600	46	2012-08-02 10:36:30.515001
35160	4	1-10662	35158	98	60	3738	34	2012-08-02 10:36:30.515001
35161	4	1-10722	35158	60	90	1789	80	2012-08-02 10:36:30.515001
35158	4	1-10724	35158	90	84	5963	94	2012-08-02 10:36:30.515001
35163	4	1-10606	35162	26	84	553	79	2012-08-02 10:36:30.515001
35164	4	1-10169	35162	84	67	286	89	2012-08-02 10:36:30.515001
35165	4	1-10505	35162	67	90	252	80	2012-08-02 10:36:30.515001
35162	4	1-10724	35162	90	26	851	94	2012-08-02 10:36:30.515001
35167	4	1-10119	35166	66	22	559	64	2012-08-02 10:36:30.923975
35168	4	1-10016	35166	22	31	324	49	2012-08-02 10:36:30.923975
35169	4	1-10462	35166	31	94	896	69	2012-08-02 10:36:30.923975
35166	4	1-10725	35166	94	66	750	1	2012-08-02 10:36:30.923975
35171	4	1-10221	35170	22	16	406	77	2012-08-02 10:36:30.923975
35172	4	1-8605	35170	16	31	165	49	2012-08-02 10:36:30.923975
35173	4	1-10462	35170	31	94	488	69	2012-08-02 10:36:30.923975
35170	4	1-10725	35170	94	22	436	1	2012-08-02 10:36:30.923975
35175	3	1-10263	35174	18	37	1469	35	2012-08-02 10:36:33.612967
35176	3	1-10661	35174	37	7	548	15	2012-08-02 10:36:33.612967
35174	3	1-10732	35174	7	18	2022	86	2012-08-02 10:36:33.612967
35178	4	1-10342	35177	28	34	2338	73	2012-08-02 10:36:34.794975
35179	4	1-10709	35177	34	71	3723	58	2012-08-02 10:36:34.794975
35180	4	1-8255	35177	71	76	1070	10	2012-08-02 10:36:34.794975
35177	4	1-10735	35177	76	28	1991	78	2012-08-02 10:36:34.794975
35182	4	1-10431	35181	62	4	1118	71	2012-08-02 10:36:35.568496
35183	4	1-10429	35181	4	4	1069	59	2012-08-02 10:36:35.568496
35184	4	1-10418	35181	4	61	3037	5	2012-08-02 10:36:35.568496
35181	4	1-10737	35181	61	62	1658	75	2012-08-02 10:36:35.568496
35186	4	1-9943	35185	40	34	298	9	2012-08-02 10:36:35.568496
35187	4	1-8987	35185	34	4	188	59	2012-08-02 10:36:35.568496
35188	4	1-10418	35185	4	61	564	5	2012-08-02 10:36:35.568496
35185	4	1-10737	35185	61	40	381	75	2012-08-02 10:36:35.568496
35190	4	1-9943	35189	40	45	976	9	2012-08-02 10:36:35.568496
35191	4	1-10714	35189	45	26	1777	15	2012-08-02 10:36:35.568496
35192	4	1-10247	35189	26	61	1840	5	2012-08-02 10:36:35.568496
35189	4	1-10737	35189	61	40	1246	75	2012-08-02 10:36:35.568496
35194	4	1-10506	35193	37	12	3243	11	2012-08-02 10:36:36.341861
35195	4	1-9919	35193	12	61	2209	30	2012-08-02 10:36:36.341861
35196	4	1-10453	35193	61	62	2032	94	2012-08-02 10:36:36.341861
35193	4	1-10739	35193	62	37	4896	39	2012-08-02 10:36:36.341861
35198	4	1-10506	35197	37	85	3772	11	2012-08-02 10:36:36.341861
35199	4	1-9332	35197	85	88	3085	3	2012-08-02 10:36:36.341861
35200	4	1-10109	35197	88	62	1838	94	2012-08-02 10:36:36.341861
35197	4	1-10739	35197	62	37	5022	39	2012-08-02 10:36:36.341861
35201	1	1-5828	\N	12	12	3515	94	2012-08-02 10:36:36.341861
35203	3	1-10673	35202	98	31	4528	95	2012-08-02 10:36:36.742018
35204	3	1-10272	35202	31	82	3660	87	2012-08-02 10:36:36.742018
35202	3	1-10740	35202	82	98	3023	98	2012-08-02 10:36:36.742018
35206	4	1-10162	35205	42	20	226	5	2012-08-02 10:37:41.025759
35207	4	1-10203	35205	20	15	143	29	2012-08-02 10:37:41.025759
35208	4	1-8811	35205	15	39	228	85	2012-08-02 10:37:41.025759
35205	4	1-10741	35205	39	42	144	4	2012-08-02 10:37:41.025759
35210	4	1-10618	35209	59	53	1445	38	2012-08-02 10:37:41.855418
35211	4	1-6696	35209	53	15	238	55	2012-08-02 10:37:41.855418
35212	4	1-10356	35209	15	93	307	88	2012-08-02 10:37:41.855418
35209	4	1-10743	35209	93	59	2891	51	2012-08-02 10:37:41.855418
35214	4	1-10522	35213	32	52	1710	3	2012-08-02 10:37:41.855418
35215	4	1-9149	35213	52	44	2059	29	2012-08-02 10:37:41.855418
35216	4	1-9862	35213	44	93	528	88	2012-08-02 10:37:41.855418
35213	4	1-10743	35213	93	32	5387	51	2012-08-02 10:37:41.855418
35218	4	1-10293	35217	64	67	2397	18	2012-08-02 10:37:42.253155
35219	4	1-10447	35217	67	11	5725	7	2012-08-02 10:37:42.253155
35220	4	1-9116	35217	11	14	2932	71	2012-08-02 10:37:42.253155
35217	4	1-10744	35217	14	64	1837	62	2012-08-02 10:37:42.253155
35222	3	1-9585	35221	34	62	1742	75	2012-08-02 10:37:42.662035
35223	3	1-5712	35221	62	47	697	84	2012-08-02 10:37:42.662035
35221	3	1-10745	35221	47	34	2151	30	2012-08-02 10:37:42.662035
35225	3	1-10453	35224	61	33	681	94	2012-08-02 10:37:42.662035
35226	3	1-6222	35224	33	47	205	84	2012-08-02 10:37:42.662035
35224	3	1-10745	35224	47	61	870	30	2012-08-02 10:37:42.662035
35228	4	1-4941	35227	48	84	2949	57	2012-08-02 10:37:42.662035
35229	4	1-9600	35227	84	18	1474	51	2012-08-02 10:37:42.662035
35230	4	1-9925	35227	18	47	871	84	2012-08-02 10:37:42.662035
35227	4	1-10745	35227	47	48	4229	30	2012-08-02 10:37:42.662035
35232	4	1-9196	35231	75	21	886	93	2012-08-02 10:37:42.662035
35233	4	1-10324	35231	21	20	1797	42	2012-08-02 10:37:42.662035
35234	4	1-4070	35231	20	47	319	84	2012-08-02 10:37:42.662035
35231	4	1-10745	35231	47	75	1561	30	2012-08-02 10:37:42.662035
35236	4	1-10604	35235	15	51	4112	61	2012-08-02 10:37:43.844368
35237	4	1-7890	35235	51	31	2550	79	2012-08-02 10:37:43.844368
35238	4	1-10218	35235	31	13	800	75	2012-08-02 10:37:43.844368
35235	4	1-10748	35235	13	15	3538	66	2012-08-02 10:37:43.844368
35240	4	1-3734	35239	81	69	1795	37	2012-08-02 10:37:43.844368
35241	4	1-10575	35239	69	83	1201	68	2012-08-02 10:37:43.844368
35242	4	1-10301	35239	83	13	594	75	2012-08-02 10:37:43.844368
35239	4	1-10748	35239	13	81	3097	66	2012-08-02 10:37:43.844368
35244	4	1-7963	35243	30	19	804	58	2012-08-02 10:37:43.844368
35245	4	1-10018	35243	19	25	732	13	2012-08-02 10:37:43.844368
35246	4	1-9208	35243	25	13	228	75	2012-08-02 10:37:43.844368
35243	4	1-10748	35243	13	30	1237	66	2012-08-02 10:37:43.844368
35248	5	1-9098	35247	91	50	215	50	2012-08-02 10:37:43.844368
35249	5	1-9835	35247	50	68	302	41	2012-08-02 10:37:43.844368
35250	5	1-7740	35247	68	78	105	11	2012-08-02 10:37:43.844368
35251	5	1-7907	35247	78	13	84	75	2012-08-02 10:37:43.844368
35247	5	1-10748	35247	13	91	499	66	2012-08-02 10:37:43.844368
35252	1	1-9835	\N	50	50	1442	41	2012-08-02 10:37:43.844368
35254	4	1-9844	35253	64	23	2729	20	2012-08-02 10:37:45.081324
35255	4	1-10508	35253	23	66	3033	90	2012-08-02 10:37:45.081324
35256	4	1-8324	35253	66	58	1120	87	2012-08-02 10:37:45.081324
35253	4	1-10751	35253	58	64	2462	96	2012-08-02 10:37:45.081324
35258	2	1-9544	35257	51	58	2058	87	2012-08-02 10:37:45.081324
35257	2	1-10751	35257	58	51	4643	96	2012-08-02 10:37:45.081324
35260	2	1-9695	35259	10	21	4026	85	2012-08-02 10:37:45.468287
35259	2	1-10752	35259	21	10	9599	79	2012-08-02 10:37:45.468287
35262	4	1-7579	35261	54	77	1733	29	2012-08-02 10:37:47.13649
35263	4	1-10103	35261	77	56	1518	83	2012-08-02 10:37:47.13649
35264	4	1-10611	35261	56	21	2182	50	2012-08-02 10:37:47.13649
35261	4	1-10756	35261	21	54	2132	61	2012-08-02 10:37:47.13649
35266	4	1-8452	35265	33	75	22	41	2012-08-02 10:37:48.331666
35267	4	1-9875	35265	75	26	19	100	2012-08-02 10:37:48.331666
35268	4	1-8575	35265	26	73	11	69	2012-08-02 10:37:48.331666
35265	4	1-10760	35265	73	33	38	38	2012-08-02 10:37:48.331666
35270	4	1-5071	35269	14	95	1553	92	2012-08-02 10:37:48.331666
35271	4	1-10567	35269	95	1	1631	95	2012-08-02 10:37:48.331666
35272	4	1-10393	35269	1	73	991	69	2012-08-02 10:37:48.331666
35269	4	1-10760	35269	73	14	3612	38	2012-08-02 10:37:48.331666
35274	4	1-5071	35273	14	39	119	92	2012-08-02 10:37:48.331666
35275	4	1-10147	35273	39	40	75	26	2012-08-02 10:37:48.331666
35276	4	1-7365	35273	40	73	64	69	2012-08-02 10:37:48.331666
35273	4	1-10760	35273	73	14	255	38	2012-08-02 10:37:48.331666
35278	4	1-9455	35277	35	15	349	22	2012-08-02 10:37:48.331666
35279	4	1-10053	35277	15	57	194	23	2012-08-02 10:37:48.331666
35280	4	1-7882	35277	57	73	105	69	2012-08-02 10:37:48.331666
35277	4	1-10760	35277	73	35	454	38	2012-08-02 10:37:48.331666
35282	4	1-9210	35281	29	3	1220	28	2012-08-02 10:37:48.331666
35283	4	1-10085	35281	3	14	1376	33	2012-08-02 10:37:48.331666
35284	4	1-8354	35281	14	73	1207	69	2012-08-02 10:37:48.331666
35281	4	1-10760	35281	73	29	5623	38	2012-08-02 10:37:48.331666
35286	4	1-9447	35285	87	9	2598	23	2012-08-02 10:37:50.751828
35287	4	1-10613	35285	9	48	3222	87	2012-08-02 10:37:50.751828
35288	4	1-10481	35285	48	24	1874	53	2012-08-02 10:37:50.751828
35285	4	1-10767	35285	24	87	1631	11	2012-08-02 10:37:50.751828
35290	4	1-9447	35289	87	28	1665	23	2012-08-02 10:37:50.751828
35291	4	1-10537	35289	28	83	1245	45	2012-08-02 10:37:50.751828
35292	4	1-4729	35289	83	24	981	53	2012-08-02 10:37:50.751828
35289	4	1-10767	35289	24	87	944	11	2012-08-02 10:37:50.751828
35294	4	1-9528	35293	38	6	1272	9	2012-08-02 10:37:50.751828
35295	4	1-6816	35293	6	52	467	98	2012-08-02 10:37:50.751828
35296	4	1-10359	35293	52	24	1299	53	2012-08-02 10:37:50.751828
35293	4	1-10767	35293	24	38	1294	11	2012-08-02 10:37:50.751828
35298	3	1-9715	35297	65	50	2141	78	2012-08-02 10:37:51.524753
35299	3	1-10670	35297	50	35	4120	41	2012-08-02 10:37:51.524753
35297	3	1-10769	35297	35	65	4619	6	2012-08-02 10:37:51.524753
35301	2	1-8575	35300	26	44	1093	69	2012-08-02 10:37:51.911514
35300	2	1-10770	35300	44	26	1679	100	2012-08-02 10:37:51.911514
35303	4	1-10640	35302	64	54	127	8	2012-08-02 10:37:53.877914
35304	4	1-7075	35302	54	53	103	54	2012-08-02 10:37:53.877914
35305	4	1-9650	35302	53	68	105	10	2012-08-02 10:37:53.877914
35302	4	1-10775	35302	68	64	77	32	2012-08-02 10:37:53.877914
35307	4	1-9856	35306	63	53	2953	54	2012-08-02 10:37:54.286695
35308	4	1-9650	35306	53	42	3145	10	2012-08-02 10:37:54.286695
35309	4	1-9155	35306	42	33	3110	40	2012-08-02 10:37:54.286695
35306	4	1-10776	35306	33	63	2082	72	2012-08-02 10:37:54.286695
35314	4	1-9853	35313	99	21	1542	50	2012-08-02 10:37:55.502015
35315	4	1-10756	35313	21	22	1637	61	2012-08-02 10:37:55.502015
35316	4	1-5228	35313	22	40	567	1	2012-08-02 10:37:55.502015
35313	4	1-10779	35313	40	99	1533	62	2012-08-02 10:37:55.502015
35318	4	1-7629	35317	14	88	560	12	2012-08-02 10:37:55.502015
35319	4	1-10608	35317	88	36	426	13	2012-08-02 10:37:55.502015
35320	4	1-10496	35317	36	40	485	1	2012-08-02 10:37:55.502015
35317	4	1-10779	35317	40	14	1513	62	2012-08-02 10:37:55.502015
35322	4	1-9902	35321	64	42	78	6	2012-08-02 10:37:55.502015
35323	4	1-2544	35321	42	94	25	69	2012-08-02 10:37:55.502015
35324	4	1-10725	35321	94	40	24	1	2012-08-02 10:37:55.502015
35321	4	1-10779	35321	40	64	75	62	2012-08-02 10:37:55.502015
35326	4	1-7449	35325	9	25	52	60	2012-08-02 10:37:55.502015
35327	4	1-1375	35325	25	36	15	13	2012-08-02 10:37:55.502015
35328	4	1-10496	35325	36	40	17	1	2012-08-02 10:37:55.502015
35325	4	1-10779	35325	40	9	54	62	2012-08-02 10:37:55.502015
35333	2	1-10324	35332	21	25	3759	42	2012-08-02 10:37:57.855669
35332	2	1-10784	35332	25	21	2150	93	2012-08-02 10:37:57.855669
35335	4	1-9763	35334	11	34	856	73	2012-08-02 10:37:57.855669
35336	4	1-10709	35334	34	72	1590	58	2012-08-02 10:37:57.855669
35337	4	1-9511	35334	72	25	1749	42	2012-08-02 10:37:57.855669
35334	4	1-10784	35334	25	11	1419	93	2012-08-02 10:37:57.855669
35347	3	1-9962	35346	81	32	4518	21	2012-08-02 10:37:59.811557
35348	3	1-10262	35346	32	80	2779	95	2012-08-02 10:37:59.811557
35346	3	1-10790	35346	80	81	4430	86	2012-08-02 10:37:59.811557
35350	4	1-10263	35349	18	56	1557	35	2012-08-02 10:37:59.811557
35351	4	1-10345	35349	56	38	1239	66	2012-08-02 10:37:59.811557
35352	4	1-8249	35349	38	80	654	95	2012-08-02 10:37:59.811557
35349	4	1-10790	35349	80	18	1053	86	2012-08-02 10:37:59.811557
35381	4	1-10367	35380	9	45	775	73	2012-08-02 10:38:04.507152
35382	4	1-10536	35380	45	28	668	2	2012-08-02 10:38:04.507152
35383	4	1-10695	35380	28	94	351	79	2012-08-02 10:38:04.507152
35380	4	1-10802	35380	94	9	520	19	2012-08-02 10:38:04.507152
35385	4	1-3513	35384	75	23	2541	37	2012-08-02 10:38:04.507152
35386	4	1-10029	35384	23	96	3552	70	2012-08-02 10:38:04.507152
35387	4	1-10451	35384	96	94	1698	79	2012-08-02 10:38:04.507152
35384	4	1-10802	35384	94	75	2567	19	2012-08-02 10:38:04.507152
35389	4	1-7418	35388	43	37	64	98	2012-08-02 10:38:04.507152
35390	4	1-10261	35388	37	60	51	24	2012-08-02 10:38:04.507152
35391	4	1-7565	35388	60	94	61	79	2012-08-02 10:38:04.507152
35388	4	1-10802	35388	94	43	96	19	2012-08-02 10:38:04.507152
35393	4	1-9037	35392	37	4	4116	10	2012-08-02 10:38:04.507152
35394	4	1-10317	35392	4	60	2180	24	2012-08-02 10:38:04.507152
35395	4	1-7565	35392	60	94	2686	79	2012-08-02 10:38:04.507152
35392	4	1-10802	35392	94	37	4357	19	2012-08-02 10:38:04.507152
35397	4	1-7539	35396	36	33	1525	52	2012-08-02 10:38:04.507152
35398	4	1-10580	35396	33	36	2539	47	2012-08-02 10:38:04.507152
35399	4	1-5805	35396	36	94	1158	79	2012-08-02 10:38:04.507152
35396	4	1-10802	35396	94	36	1944	19	2012-08-02 10:38:04.507152
35451	4	1-9769	35450	14	65	924	6	2012-08-02 10:38:09.567524
35452	4	1-9715	35450	65	50	502	78	2012-08-02 10:38:09.567524
35453	4	1-10670	35450	50	38	1133	41	2012-08-02 10:38:09.567524
35450	4	1-10815	35450	38	14	3415	44	2012-08-02 10:38:09.567524
35477	4	1-8721	35476	98	13	390	14	2012-08-02 10:38:11.90992
35478	4	1-9040	35476	13	5	376	92	2012-08-02 10:38:11.90992
35479	4	1-10291	35476	5	23	161	43	2012-08-02 10:38:11.90992
35476	4	1-10821	35476	23	98	354	39	2012-08-02 10:38:11.90992
35481	4	1-8937	35480	36	41	412	16	2012-08-02 10:38:11.90992
35482	4	1-9997	35480	41	60	213	35	2012-08-02 10:38:11.90992
35483	4	1-1317	35480	60	23	95	43	2012-08-02 10:38:11.90992
35480	4	1-10821	35480	23	36	228	39	2012-08-02 10:38:11.90992
35485	4	1-10506	35484	37	85	1240	11	2012-08-02 10:38:11.90992
35486	4	1-9332	35484	85	13	1067	3	2012-08-02 10:38:11.90992
35487	4	1-10121	35484	13	23	629	43	2012-08-02 10:38:11.90992
35484	4	1-10821	35484	23	37	1570	39	2012-08-02 10:38:11.90992
35493	4	1-10583	35492	38	26	5562	15	2012-08-02 10:38:13.114412
35494	4	1-10247	35492	26	84	4551	5	2012-08-02 10:38:13.114412
35495	4	1-10753	35492	84	52	441	63	2012-08-02 10:38:13.114412
35492	4	1-10824	35492	52	38	4066	46	2012-08-02 10:38:13.114412
35497	4	1-8036	35496	64	5	780	93	2012-08-02 10:38:13.114412
35498	4	1-10402	35496	5	91	1050	79	2012-08-02 10:38:13.114412
35499	4	1-10754	35496	91	52	179	63	2012-08-02 10:38:13.114412
35496	4	1-10824	35496	52	64	1733	46	2012-08-02 10:38:13.114412
35501	4	1-9591	35500	92	15	1013	28	2012-08-02 10:38:13.114412
35502	4	1-10635	35500	15	92	580	61	2012-08-02 10:38:13.114412
35503	4	1-8955	35500	92	52	206	63	2012-08-02 10:38:13.114412
35500	4	1-10824	35500	52	92	2133	46	2012-08-02 10:38:13.114412
35509	3	1-10085	35508	3	86	832	33	2012-08-02 10:38:16.66256
35510	3	1-9391	35508	86	66	403	4	2012-08-02 10:38:16.66256
35508	3	1-10833	35508	66	3	1192	28	2012-08-02 10:38:16.66256
35512	4	1-10635	35511	15	36	2784	61	2012-08-02 10:38:16.66256
35513	4	1-8467	35511	36	39	2579	85	2012-08-02 10:38:16.66256
35514	4	1-10741	35511	39	66	1430	4	2012-08-02 10:38:16.66256
35511	4	1-10833	35511	66	15	5226	28	2012-08-02 10:38:16.66256
35519	2	1-9464	35518	16	51	171	68	2012-08-02 10:38:19.005135
35518	2	1-10840	35518	51	16	8806	41	2012-08-02 10:38:19.005135
35521	4	1-4111	35520	81	81	34	78	2012-08-02 10:38:19.005135
35522	4	1-10303	35520	81	60	19	96	2012-08-02 10:38:19.005135
35523	4	1-7563	35520	60	51	3	68	2012-08-02 10:38:19.005135
35520	4	1-10840	35520	51	81	243	41	2012-08-02 10:38:19.005135
35575	4	1-10670	35574	50	97	3064	41	2012-08-02 10:38:26.871685
35576	4	1-10792	35574	97	99	1766	30	2012-08-02 10:38:26.871685
35577	4	1-9953	35574	99	68	1394	79	2012-08-02 10:38:26.871685
35574	4	1-10859	35574	68	50	1611	78	2012-08-02 10:38:26.871685
35579	4	1-10342	35578	28	45	2302	73	2012-08-02 10:38:26.871685
35580	4	1-10536	35578	45	28	2183	2	2012-08-02 10:38:26.871685
35581	4	1-10695	35578	28	68	1358	79	2012-08-02 10:38:26.871685
35578	4	1-10859	35578	68	28	1651	78	2012-08-02 10:38:26.871685
35591	3	1-10691	35590	56	2	786	12	2012-08-02 10:38:28.053742
35590	3	1-10862	35590	32	56	1109	52	2012-08-02 10:38:28.053742
35311	3	1-10500	35310	5	4	1099	10	2012-08-02 10:37:55.10422
35312	3	1-10317	35310	4	63	400	24	2012-08-02 10:37:55.10422
35310	3	1-10778	35310	63	5	2964	18	2012-08-02 10:37:55.10422
35330	3	1-10481	35329	48	38	1119	53	2012-08-02 10:37:56.297508
35331	3	1-10095	35329	38	94	1045	37	2012-08-02 10:37:56.297508
35329	3	1-10781	35329	94	48	1677	87	2012-08-02 10:37:56.297508
35339	4	1-10476	35338	95	8	3372	27	2012-08-02 10:37:57.469118
35340	4	1-6141	35338	8	10	2394	93	2012-08-02 10:37:57.469118
35341	4	1-9546	35338	10	55	1278	1	2012-08-02 10:37:57.469118
35338	4	1-10785	35338	55	95	717	78	2012-08-02 10:37:57.469118
35354	4	1-10544	35353	6	29	1174	93	2012-08-02 10:38:02.915583
35355	4	1-10684	35353	29	41	1071	8	2012-08-02 10:38:02.915583
35356	4	1-8268	35353	41	42	866	16	2012-08-02 10:38:02.915583
35353	4	1-10798	35353	42	6	1154	51	2012-08-02 10:38:02.915583
35358	4	1-10497	35357	95	6	1412	21	2012-08-02 10:38:02.915583
35359	4	1-8968	35357	6	88	901	36	2012-08-02 10:38:02.915583
35360	4	1-10543	35357	88	42	1037	16	2012-08-02 10:38:02.915583
35357	4	1-10798	35357	42	95	1728	51	2012-08-02 10:38:02.915583
35362	4	1-10544	35361	6	18	397	93	2012-08-02 10:38:02.915583
35363	4	1-8658	35361	18	27	168	34	2012-08-02 10:38:02.915583
35364	4	1-8971	35361	27	42	163	16	2012-08-02 10:38:02.915583
35361	4	1-10798	35361	42	6	291	51	2012-08-02 10:38:02.915583
35366	3	1-10522	35365	32	50	680	3	2012-08-02 10:38:02.915583
35367	3	1-8777	35365	50	42	546	16	2012-08-02 10:38:02.915583
35365	3	1-10798	35365	42	32	1001	51	2012-08-02 10:38:02.915583
35368	1	1-3238	\N	40	40	1667	51	2012-08-02 10:38:02.915583
35405	3	1-10756	35404	21	38	2266	61	2012-08-02 10:38:06.440517
35406	3	1-10458	35404	38	25	2403	2	2012-08-02 10:38:06.440517
35404	3	1-10807	35404	25	21	3217	50	2012-08-02 10:38:06.440517
35408	4	1-9753	35407	100	10	888	71	2012-08-02 10:38:06.440517
35409	4	1-9927	35407	10	95	444	97	2012-08-02 10:38:06.440517
35410	4	1-9747	35407	95	25	798	2	2012-08-02 10:38:06.440517
35407	4	1-10807	35407	25	100	1354	50	2012-08-02 10:38:06.440517
35412	3	1-10041	35411	39	39	2427	7	2012-08-02 10:38:06.440517
35413	3	1-7346	35411	39	25	2125	2	2012-08-02 10:38:06.440517
35411	3	1-10807	35411	25	39	4220	50	2012-08-02 10:38:06.440517
35415	2	1-10169	35414	84	99	2382	89	2012-08-02 10:38:07.611944
35414	2	1-10810	35414	99	84	4385	79	2012-08-02 10:38:07.611944
35417	4	1-10754	35416	91	38	966	63	2012-08-02 10:38:07.611944
35418	4	1-9880	35416	38	52	1144	18	2012-08-02 10:38:07.611944
35419	4	1-10783	35416	52	99	1738	89	2012-08-02 10:38:07.611944
35416	4	1-10810	35416	99	91	4294	79	2012-08-02 10:38:07.611944
35429	3	1-6036	35428	71	71	3507	79	2012-08-02 10:38:08.771816
35430	3	1-10723	35428	71	96	6329	12	2012-08-02 10:38:08.771816
35428	3	1-10813	35428	96	71	7258	9	2012-08-02 10:38:08.771816
35432	3	1-8476	35431	36	89	898	71	2012-08-02 10:38:08.771816
35433	3	1-10167	35431	89	96	909	12	2012-08-02 10:38:08.771816
35431	3	1-10813	35431	96	36	1393	9	2012-08-02 10:38:08.771816
35435	4	1-5611	35434	1	10	39	3	2012-08-02 10:38:08.771816
35436	4	1-9133	35434	10	56	57	52	2012-08-02 10:38:08.771816
35437	4	1-10691	35434	56	96	51	12	2012-08-02 10:38:08.771816
35434	4	1-10813	35434	96	1	94	9	2012-08-02 10:38:08.771816
35505	4	1-10662	35504	98	52	92	34	2012-08-02 10:38:15.88932
35506	4	1-8957	35504	52	38	43	23	2012-08-02 10:38:15.88932
35507	4	1-10731	35504	38	9	32	41	2012-08-02 10:38:15.88932
35504	4	1-10831	35504	9	98	80	46	2012-08-02 10:38:15.88932
35525	3	1-9942	35524	85	15	431	55	2012-08-02 10:38:19.778219
35526	3	1-10356	35524	15	75	952	88	2012-08-02 10:38:19.778219
35524	3	1-10841	35524	75	85	1080	17	2012-08-02 10:38:19.778219
35528	5	1-6574	35527	85	62	324	56	2012-08-02 10:38:22.131703
35529	5	1-9879	35527	62	21	556	50	2012-08-02 10:38:22.131703
35530	5	1-10756	35527	21	92	688	61	2012-08-02 10:38:22.131703
35531	5	1-8955	35527	92	88	262	63	2012-08-02 10:38:22.131703
35527	5	1-10847	35527	88	85	400	24	2012-08-02 10:38:22.131703
35533	4	1-5591	35532	18	21	368	50	2012-08-02 10:38:22.131703
35534	4	1-10756	35532	21	92	458	61	2012-08-02 10:38:22.131703
35535	4	1-8955	35532	92	88	176	63	2012-08-02 10:38:22.131703
35532	4	1-10847	35532	88	18	270	24	2012-08-02 10:38:22.131703
35537	5	1-8104	35536	26	62	551	56	2012-08-02 10:38:22.131703
35538	5	1-9879	35536	62	21	957	50	2012-08-02 10:38:22.131703
35539	5	1-10756	35536	21	92	1196	61	2012-08-02 10:38:22.131703
35540	5	1-8955	35536	92	88	460	63	2012-08-02 10:38:22.131703
35536	5	1-10847	35536	88	26	710	24	2012-08-02 10:38:22.131703
35541	1	1-2583	\N	9	9	225	50	2012-08-02 10:38:22.131703
35543	4	1-10431	35542	62	10	773	71	2012-08-02 10:38:23.335866
35544	4	1-9927	35542	10	38	541	97	2012-08-02 10:38:23.335866
35545	4	1-10636	35542	38	27	989	51	2012-08-02 10:38:23.335866
35542	4	1-10850	35542	27	62	866	75	2012-08-02 10:38:23.335866
35547	3	1-10431	35546	62	37	748	71	2012-08-02 10:38:23.335866
35548	3	1-10258	35546	37	27	932	51	2012-08-02 10:38:23.335866
35546	3	1-10850	35546	27	62	827	75	2012-08-02 10:38:23.335866
35572	3	1-10781	35571	94	83	1192	87	2012-08-02 10:38:26.473899
35573	3	1-10502	35571	83	48	428	1	2012-08-02 10:38:26.473899
35571	3	1-10858	35571	48	94	804	37	2012-08-02 10:38:26.473899
35594	4	1-9400	35593	81	4	2662	85	2012-08-02 10:38:28.82744
35595	4	1-7265	35593	4	89	466	29	2012-08-02 10:38:28.82744
35596	4	1-10488	35593	89	53	354	67	2012-08-02 10:38:28.82744
35593	4	1-10864	35593	53	81	1090	34	2012-08-02 10:38:28.82744
35614	2	1-9641	35613	17	53	5147	70	2012-08-02 10:38:30.009498
35613	2	1-10867	35613	53	17	3387	3	2012-08-02 10:38:30.009498
35616	4	1-8864	35615	21	64	1566	91	2012-08-02 10:38:30.009498
35617	4	1-6975	35615	64	35	1917	92	2012-08-02 10:38:30.009498
35618	4	1-10056	35615	35	53	2860	70	2012-08-02 10:38:30.009498
35615	4	1-10867	35615	53	21	2689	3	2012-08-02 10:38:30.009498
35620	3	1-9007	35619	22	98	818	98	2012-08-02 10:38:31.213882
35621	3	1-10673	35619	98	43	1736	95	2012-08-02 10:38:31.213882
35619	3	1-10869	35619	43	22	2513	4	2012-08-02 10:38:31.213882
35623	3	1-10180	35622	29	32	3519	9	2012-08-02 10:38:32.013925
35624	3	1-8033	35622	32	94	1102	81	2012-08-02 10:38:32.013925
35622	3	1-10873	35622	94	29	9110	50	2012-08-02 10:38:32.013925
35629	4	1-7346	35628	39	88	5593	2	2012-08-02 10:38:34.74023
35630	4	1-9706	35628	88	9	751	41	2012-08-02 10:38:34.74023
35631	4	1-10831	35628	9	78	1627	46	2012-08-02 10:38:34.74023
35628	4	1-10879	35628	78	39	6760	7	2012-08-02 10:38:34.74023
35343	4	1-10451	35342	96	1	2740	79	2012-08-02 10:37:58.231186
35344	4	1-10456	35342	1	33	1653	40	2012-08-02 10:37:58.231186
35345	4	1-10776	35342	33	84	856	72	2012-08-02 10:37:58.231186
35342	4	1-10786	35342	84	96	6682	70	2012-08-02 10:37:58.231186
35370	4	1-10607	35369	26	45	5599	70	2012-08-02 10:38:03.7003
35371	4	1-1039	35369	45	72	1495	14	2012-08-02 10:38:03.7003
35372	4	1-9586	35369	72	29	1152	42	2012-08-02 10:38:03.7003
35369	4	1-10800	35369	29	26	1073	86	2012-08-02 10:38:03.7003
35374	4	1-6766	35373	41	48	271	29	2012-08-02 10:38:03.7003
35375	4	1-10712	35373	48	72	358	58	2012-08-02 10:38:03.7003
35376	4	1-9511	35373	72	29	376	42	2012-08-02 10:38:03.7003
35373	4	1-10800	35373	29	41	384	86	2012-08-02 10:38:03.7003
35378	3	1-10263	35377	18	80	2321	35	2012-08-02 10:38:03.7003
35379	3	1-8910	35377	80	29	1279	42	2012-08-02 10:38:03.7003
35377	3	1-10800	35377	29	18	1378	86	2012-08-02 10:38:03.7003
35401	4	1-10262	35400	32	1	1952	95	2012-08-02 10:38:06.064779
35402	4	1-10393	35400	1	44	1384	69	2012-08-02 10:38:06.064779
35403	4	1-10770	35400	44	94	2267	100	2012-08-02 10:38:06.064779
35400	4	1-10806	35400	94	32	2853	21	2012-08-02 10:38:06.064779
35421	3	1-9282	35420	11	24	2045	53	2012-08-02 10:38:07.987362
35422	3	1-10767	35420	24	22	1535	11	2012-08-02 10:38:07.987362
35420	3	1-10811	35420	22	11	5325	54	2012-08-02 10:38:07.987362
35424	5	1-9884	35423	51	76	602	14	2012-08-02 10:38:07.987362
35425	5	1-10601	35423	76	3	425	19	2012-08-02 10:38:07.987362
35426	5	1-8052	35423	3	24	500	53	2012-08-02 10:38:07.987362
35427	5	1-10767	35423	24	22	432	11	2012-08-02 10:38:07.987362
35423	5	1-10811	35423	22	51	1727	54	2012-08-02 10:38:07.987362
35439	4	1-10301	35438	83	22	1143	75	2012-08-02 10:38:09.170128
35440	4	1-9985	35438	22	29	1728	42	2012-08-02 10:38:09.170128
35441	4	1-10800	35438	29	38	1889	86	2012-08-02 10:38:09.170128
35438	4	1-10814	35438	38	83	2133	68	2012-08-02 10:38:09.170128
35443	4	1-10228	35442	38	76	3333	28	2012-08-02 10:38:09.170128
35444	4	1-8471	35442	76	29	2948	42	2012-08-02 10:38:09.170128
35445	4	1-10800	35442	29	38	3196	86	2012-08-02 10:38:09.170128
35442	4	1-10814	35442	38	38	3716	68	2012-08-02 10:38:09.170128
35447	4	1-9900	35446	92	53	395	83	2012-08-02 10:38:09.170128
35448	4	1-9824	35446	53	67	545	48	2012-08-02 10:38:09.170128
35449	4	1-10605	35446	67	38	700	86	2012-08-02 10:38:09.170128
35446	4	1-10814	35446	38	92	833	68	2012-08-02 10:38:09.170128
35455	4	1-4978	35454	57	67	676	48	2012-08-02 10:38:10.340891
35456	4	1-10605	35454	67	1	656	86	2012-08-02 10:38:10.340891
35457	4	1-9838	35454	1	53	581	34	2012-08-02 10:38:10.340891
35454	4	1-10817	35454	53	57	2285	22	2012-08-02 10:38:10.340891
35459	4	1-5619	35458	83	25	215	43	2012-08-02 10:38:10.340891
35460	4	1-6425	35458	25	98	140	46	2012-08-02 10:38:10.340891
35461	4	1-10662	35458	98	53	135	34	2012-08-02 10:38:10.340891
35458	4	1-10817	35458	53	83	585	22	2012-08-02 10:38:10.340891
35463	4	1-8361	35462	29	26	1153	52	2012-08-02 10:38:10.340891
35464	4	1-5880	35462	26	1	421	86	2012-08-02 10:38:10.340891
35465	4	1-9838	35462	1	53	437	34	2012-08-02 10:38:10.340891
35462	4	1-10817	35462	53	29	2017	22	2012-08-02 10:38:10.340891
35467	4	1-5200	35466	20	29	634	50	2012-08-02 10:38:10.340891
35468	4	1-10180	35466	29	50	506	9	2012-08-02 10:38:10.340891
35469	4	1-7534	35466	50	53	349	34	2012-08-02 10:38:10.340891
35466	4	1-10817	35466	53	20	1727	22	2012-08-02 10:38:10.340891
35470	1	1-2737	\N	55	55	381	86	2012-08-02 10:38:10.340891
35472	5	1-10297	35471	73	10	1359	99	2012-08-02 10:38:11.523412
35473	5	1-10599	35471	10	3	6455	28	2012-08-02 10:38:11.523412
35474	5	1-10085	35471	3	78	5256	33	2012-08-02 10:38:11.523412
35475	5	1-10279	35471	78	38	5715	39	2012-08-02 10:38:11.523412
35471	5	1-10820	35471	38	73	2493	45	2012-08-02 10:38:11.523412
35489	4	1-9984	35488	32	56	734	35	2012-08-02 10:38:12.705559
35490	4	1-10345	35488	56	63	714	66	2012-08-02 10:38:12.705559
35491	4	1-6484	35488	63	50	233	98	2012-08-02 10:38:12.705559
35488	4	1-10823	35488	50	32	697	62	2012-08-02 10:38:12.705559
35516	3	1-9811	35515	68	41	4045	97	2012-08-02 10:38:19.391454
35517	3	1-9153	35515	41	3	2258	37	2012-08-02 10:38:19.391454
35515	3	1-10839	35515	3	68	6270	31	2012-08-02 10:38:19.391454
35550	4	1-10643	35549	67	16	2944	32	2012-08-02 10:38:23.722511
35551	4	1-10472	35549	16	10	2476	27	2012-08-02 10:38:23.722511
35552	4	1-9356	35549	10	19	248	78	2012-08-02 10:38:23.722511
35549	4	1-10851	35549	19	67	1330	94	2012-08-02 10:38:23.722511
35554	3	1-6222	35553	33	76	1441	84	2012-08-02 10:38:23.722511
35555	3	1-10658	35553	76	19	719	78	2012-08-02 10:38:23.722511
35553	3	1-10851	35553	19	33	4275	94	2012-08-02 10:38:23.722511
35557	3	1-3451	35556	45	81	324	41	2012-08-02 10:38:23.722511
35558	3	1-4111	35556	81	19	102	78	2012-08-02 10:38:23.722511
35556	3	1-10851	35556	19	45	747	94	2012-08-02 10:38:23.722511
35560	5	1-8730	35559	92	93	702	74	2012-08-02 10:38:23.722511
35561	5	1-10768	35559	93	70	1131	75	2012-08-02 10:38:23.722511
35562	5	1-8226	35559	70	18	363	36	2012-08-02 10:38:23.722511
35563	5	1-9525	35559	18	19	246	78	2012-08-02 10:38:23.722511
35559	5	1-10851	35559	19	92	1819	94	2012-08-02 10:38:23.722511
35565	3	1-10617	35564	2	100	4621	47	2012-08-02 10:38:24.905197
35566	3	1-10822	35564	100	71	1966	8	2012-08-02 10:38:24.905197
35564	3	1-10854	35564	71	2	4239	91	2012-08-02 10:38:24.905197
35568	4	1-9374	35567	59	33	591	52	2012-08-02 10:38:24.905197
35569	4	1-10580	35567	33	100	691	47	2012-08-02 10:38:24.905197
35570	4	1-10822	35567	100	71	322	8	2012-08-02 10:38:24.905197
35567	4	1-10854	35567	71	59	760	91	2012-08-02 10:38:24.905197
35583	4	1-9149	35582	52	89	183	29	2012-08-02 10:38:27.258545
35584	4	1-10488	35582	89	99	116	67	2012-08-02 10:38:27.258545
35585	4	1-10267	35582	99	57	96	45	2012-08-02 10:38:27.258545
35582	4	1-10860	35582	57	52	80	3	2012-08-02 10:38:27.258545
35587	4	1-10121	35586	13	23	905	43	2012-08-02 10:38:27.258545
35588	4	1-10821	35586	23	38	2418	39	2012-08-02 10:38:27.258545
35589	4	1-10820	35586	38	57	1452	45	2012-08-02 10:38:27.258545
35586	4	1-10860	35586	57	13	1433	3	2012-08-02 10:38:27.258545
35626	3	1-9068	35625	42	32	3334	85	2012-08-02 10:38:33.193019
35627	3	1-10620	35625	32	27	1738	65	2012-08-02 10:38:33.193019
35625	3	1-10875	35625	27	42	3149	59	2012-08-02 10:38:33.193019
35645	4	1-6499	35644	73	89	485	29	2012-08-02 10:38:37.833705
35646	4	1-10488	35644	89	53	300	67	2012-08-02 10:38:37.833705
35647	4	1-10864	35644	53	54	749	34	2012-08-02 10:38:37.833705
35644	4	1-10887	35644	54	73	817	12	2012-08-02 10:38:37.833705
35592	3	1-10586	35590	2	32	1420	48	2012-08-02 10:38:28.053742
35598	4	1-10420	35597	19	78	1080	55	2012-08-02 10:38:29.225009
35599	4	1-9472	35597	78	8	889	59	2012-08-02 10:38:29.225009
35600	4	1-10569	35597	8	65	723	58	2012-08-02 10:38:29.225009
35597	4	1-10865	35597	65	19	1867	37	2012-08-02 10:38:29.225009
35602	4	1-10842	35601	31	19	2263	27	2012-08-02 10:38:29.225009
35603	4	1-10405	35601	19	80	1744	81	2012-08-02 10:38:29.225009
35604	4	1-9581	35601	80	65	2066	58	2012-08-02 10:38:29.225009
35601	4	1-10865	35601	65	31	5420	37	2012-08-02 10:38:29.225009
35606	4	1-10781	35605	94	83	3517	87	2012-08-02 10:38:29.225009
35607	4	1-10502	35605	83	57	1253	1	2012-08-02 10:38:29.225009
35608	4	1-3689	35605	57	65	908	58	2012-08-02 10:38:29.225009
35605	4	1-10865	35605	65	94	2389	37	2012-08-02 10:38:29.225009
35610	4	1-10575	35609	69	2	91	68	2012-08-02 10:38:29.225009
35611	4	1-9329	35609	2	21	55	49	2012-08-02 10:38:29.225009
35612	4	1-5862	35609	21	65	48	58	2012-08-02 10:38:29.225009
35609	4	1-10865	35609	65	69	128	37	2012-08-02 10:38:29.225009
35643	2	1-8032	35642	60	71	624	24	2012-08-02 10:38:36.286525
35642	2	1-10883	35642	71	60	4837	33	2012-08-02 10:38:36.286525
35649	4	1-10171	35648	74	18	1925	72	2012-08-02 10:38:38.61951
35650	4	1-10273	35648	18	22	1357	65	2012-08-02 10:38:38.61951
35651	4	1-6493	35648	22	17	650	51	2012-08-02 10:38:38.61951
35648	4	1-10889	35648	17	74	1297	34	2012-08-02 10:38:38.61951
35653	4	1-8818	35652	70	76	2610	14	2012-08-02 10:38:38.61951
35654	4	1-10601	35652	76	82	1951	19	2012-08-02 10:38:38.61951
35655	4	1-9046	35652	82	17	1893	51	2012-08-02 10:38:38.61951
35652	4	1-10889	35652	17	70	4040	34	2012-08-02 10:38:38.61951
35657	4	1-9942	35656	85	39	194	55	2012-08-02 10:38:39.801776
35658	4	1-9676	35656	39	70	106	39	2012-08-02 10:38:39.801776
35659	4	1-5867	35656	70	32	83	73	2012-08-02 10:38:39.801776
35656	4	1-10892	35656	32	85	338	17	2012-08-02 10:38:39.801776
35666	3	1-10413	35665	15	36	81	25	2012-08-02 10:38:40.984077
35667	3	1-5123	35665	36	27	37	2	2012-08-02 10:38:40.984077
35665	3	1-10896	35665	27	15	45	14	2012-08-02 10:38:40.984077
35669	4	1-10601	35668	76	58	2490	19	2012-08-02 10:38:40.984077
35670	4	1-6776	35668	58	95	1028	97	2012-08-02 10:38:40.984077
35671	4	1-9747	35668	95	27	2546	2	2012-08-02 10:38:40.984077
35668	4	1-10896	35668	27	76	3166	14	2012-08-02 10:38:40.984077
35673	4	1-10130	35672	27	93	102	100	2012-08-02 10:38:40.984077
35674	4	1-10159	35672	93	27	105	52	2012-08-02 10:38:40.984077
35675	4	1-7650	35672	27	27	97	2	2012-08-02 10:38:40.984077
35672	4	1-10896	35672	27	27	125	14	2012-08-02 10:38:40.984077
35732	4	1-4941	35731	48	46	344	57	2012-08-02 10:38:50.391199
35733	4	1-10512	35731	46	58	596	58	2012-08-02 10:38:50.391199
35734	4	1-7438	35731	58	21	415	70	2012-08-02 10:38:50.391199
35731	4	1-10919	35731	21	48	398	30	2012-08-02 10:38:50.391199
35736	4	1-9953	35735	99	68	947	79	2012-08-02 10:38:50.391199
35737	4	1-10859	35735	68	28	1264	78	2012-08-02 10:38:50.391199
35738	4	1-8376	35735	28	21	1044	70	2012-08-02 10:38:50.391199
35735	4	1-10919	35735	21	99	1038	30	2012-08-02 10:38:50.391199
35740	4	1-7226	35739	73	46	2051	57	2012-08-02 10:38:50.391199
35741	4	1-10512	35739	46	58	3731	58	2012-08-02 10:38:50.391199
35742	4	1-7438	35739	58	21	2729	70	2012-08-02 10:38:50.391199
35739	4	1-10919	35739	21	73	2752	30	2012-08-02 10:38:50.391199
35749	4	1-8780	35748	98	35	3682	80	2012-08-02 10:38:51.581956
35750	4	1-10773	35748	35	42	1441	84	2012-08-02 10:38:51.581956
35751	4	1-10716	35748	42	57	1433	29	2012-08-02 10:38:51.581956
35748	4	1-10922	35748	57	98	2056	45	2012-08-02 10:38:51.581956
35753	4	1-10233	35752	47	93	1551	66	2012-08-02 10:38:51.581956
35754	4	1-9320	35752	93	20	1203	5	2012-08-02 10:38:51.581956
35755	4	1-10203	35752	20	57	817	29	2012-08-02 10:38:51.581956
35752	4	1-10922	35752	57	47	1178	45	2012-08-02 10:38:51.581956
35757	4	1-4729	35756	83	24	849	53	2012-08-02 10:38:51.581956
35758	4	1-10767	35756	24	56	787	11	2012-08-02 10:38:51.581956
35759	4	1-7724	35756	56	57	774	29	2012-08-02 10:38:51.581956
35756	4	1-10922	35756	57	83	1118	45	2012-08-02 10:38:51.581956
35761	4	1-10297	35760	73	48	1333	99	2012-08-02 10:38:51.581956
35762	4	1-8305	35760	48	20	1634	5	2012-08-02 10:38:51.581956
35763	4	1-10203	35760	20	57	1172	29	2012-08-02 10:38:51.581956
35760	4	1-10922	35760	57	73	1783	45	2012-08-02 10:38:51.581956
35770	4	1-5619	35769	83	59	1073	43	2012-08-02 10:38:53.601128
35771	4	1-10383	35769	59	1	1134	90	2012-08-02 10:38:53.601128
35772	4	1-9603	35769	1	47	821	17	2012-08-02 10:38:53.601128
35769	4	1-10927	35769	47	83	2646	22	2012-08-02 10:38:53.601128
35774	4	1-10053	35773	15	28	1734	23	2012-08-02 10:38:53.601128
35775	4	1-10777	35773	28	1	1232	90	2012-08-02 10:38:53.601128
35776	4	1-9603	35773	1	47	917	17	2012-08-02 10:38:53.601128
35773	4	1-10927	35773	47	15	3042	22	2012-08-02 10:38:53.601128
35778	4	1-9811	35777	68	87	785	97	2012-08-02 10:38:54.786186
35779	4	1-10062	35777	87	97	858	38	2012-08-02 10:38:54.786186
35780	4	1-10880	35777	97	56	650	59	2012-08-02 10:38:54.786186
35777	4	1-10931	35777	56	68	1008	31	2012-08-02 10:38:54.786186
35782	4	1-10148	35781	53	59	2094	51	2012-08-02 10:38:54.786186
35783	4	1-10618	35781	59	97	2563	38	2012-08-02 10:38:54.786186
35784	4	1-10880	35781	97	56	2314	59	2012-08-02 10:38:54.786186
35781	4	1-10931	35781	56	53	4278	31	2012-08-02 10:38:54.786186
35807	4	1-10194	35806	51	32	2212	61	2012-08-02 10:38:58.75591
35808	4	1-10157	35806	32	22	3179	58	2012-08-02 10:38:58.75591
35809	4	1-10901	35806	22	19	1518	73	2012-08-02 10:38:58.75591
35806	4	1-10940	35806	19	51	2364	45	2012-08-02 10:38:58.75591
35852	3	1-6191	35851	17	72	2607	60	2012-08-02 10:39:07.479241
35853	3	1-10407	35851	72	98	440	29	2012-08-02 10:39:07.479241
35851	3	1-10962	35851	98	17	6326	8	2012-08-02 10:39:07.479241
35881	4	1-10899	35880	57	26	2360	94	2012-08-02 10:39:13.34021
35882	4	1-10606	35880	26	100	2027	79	2012-08-02 10:39:13.34021
35883	4	1-10046	35880	100	65	1744	38	2012-08-02 10:39:13.34021
35880	4	1-10977	35880	65	57	1685	28	2012-08-02 10:39:13.34021
35908	2	1-10148	35907	53	34	1390	51	2012-08-02 10:39:15.683106
35907	2	1-10983	35907	34	53	3843	31	2012-08-02 10:39:15.683106
35926	4	1-8249	35925	38	38	3584	95	2012-08-02 10:39:18.027505
35927	4	1-10102	35925	38	23	2186	10	2012-08-02 10:39:18.027505
35928	4	1-10715	35925	23	99	1500	51	2012-08-02 10:39:18.027505
35925	4	1-10990	35925	99	38	7309	66	2012-08-02 10:39:18.027505
35930	3	1-5939	35929	19	37	166	71	2012-08-02 10:39:18.027505
35633	3	1-6770	35632	34	84	675	94	2012-08-02 10:38:34.74023
35634	3	1-10363	35632	84	78	677	46	2012-08-02 10:38:34.74023
35632	3	1-10879	35632	78	34	2983	7	2012-08-02 10:38:34.74023
35636	4	1-10293	35635	64	75	3131	18	2012-08-02 10:38:35.899927
35637	4	1-6894	35635	75	93	602	56	2012-08-02 10:38:35.899927
35638	4	1-9512	35635	93	62	590	100	2012-08-02 10:38:35.899927
35635	4	1-10882	35635	62	64	4299	62	2012-08-02 10:38:35.899927
35640	3	1-7481	35639	68	27	303	14	2012-08-02 10:38:35.899927
35641	3	1-10130	35639	27	62	144	100	2012-08-02 10:38:35.899927
35639	3	1-10882	35639	62	68	1214	62	2012-08-02 10:38:35.899927
35689	2	1-10709	35688	34	22	4439	58	2012-08-02 10:38:42.940084
35688	2	1-10901	35688	22	34	2224	73	2012-08-02 10:38:42.940084
35698	5	1-9953	35697	99	68	128	79	2012-08-02 10:38:45.669057
35699	5	1-10859	35697	68	76	143	78	2012-08-02 10:38:45.669057
35700	5	1-9495	35697	76	23	80	43	2012-08-02 10:38:45.669057
35701	5	1-10821	35697	23	74	174	39	2012-08-02 10:38:45.669057
35697	5	1-10907	35697	74	99	169	30	2012-08-02 10:38:45.669057
35706	4	1-10554	35705	11	1	5889	18	2012-08-02 10:38:46.850819
35707	4	1-10128	35705	1	54	4406	34	2012-08-02 10:38:46.850819
35708	4	1-10887	35705	54	100	3727	12	2012-08-02 10:38:46.850819
35705	4	1-10911	35705	100	11	5346	26	2012-08-02 10:38:46.850819
35720	4	1-10496	35719	36	22	593	1	2012-08-02 10:38:49.60414
35721	4	1-10221	35719	22	39	587	77	2012-08-02 10:38:49.60414
35722	4	1-5471	35719	39	85	622	85	2012-08-02 10:38:49.60414
35719	4	1-10917	35719	85	36	538	13	2012-08-02 10:38:49.60414
35724	4	1-10496	35723	36	22	4207	1	2012-08-02 10:38:49.60414
35725	4	1-10221	35723	22	23	4399	77	2012-08-02 10:38:49.60414
35726	4	1-7526	35723	23	85	3962	85	2012-08-02 10:38:49.60414
35723	4	1-10917	35723	85	36	3615	13	2012-08-02 10:38:49.60414
35744	3	1-10540	35743	53	16	2380	64	2012-08-02 10:38:50.79204
35745	3	1-10439	35743	16	44	2581	96	2012-08-02 10:38:50.79204
35743	3	1-10920	35743	44	53	3068	39	2012-08-02 10:38:50.79204
35765	5	1-9794	35764	12	5	1272	93	2012-08-02 10:38:53.203026
35766	5	1-10402	35764	5	91	2264	79	2012-08-02 10:38:53.203026
35767	5	1-10754	35764	91	26	511	63	2012-08-02 10:38:53.203026
35768	5	1-9481	35764	26	70	987	35	2012-08-02 10:38:53.203026
35764	5	1-10926	35764	70	12	1770	8	2012-08-02 10:38:53.203026
35815	4	1-8387	35814	53	60	3290	12	2012-08-02 10:38:59.529301
35816	4	1-10856	35814	60	21	1829	23	2012-08-02 10:38:59.529301
35817	4	1-8933	35814	21	32	1573	5	2012-08-02 10:38:59.529301
35814	4	1-10942	35814	32	53	5848	89	2012-08-02 10:38:59.529301
35819	4	1-10545	35818	83	57	1500	28	2012-08-02 10:38:59.529301
35820	4	1-10899	35818	57	7	1460	94	2012-08-02 10:38:59.529301
35821	4	1-6993	35818	7	32	716	5	2012-08-02 10:38:59.529301
35818	4	1-10942	35818	32	83	3078	89	2012-08-02 10:38:59.529301
35823	4	1-10892	35822	32	68	647	17	2012-08-02 10:39:03.101981
35824	4	1-9980	35822	68	2	495	46	2012-08-02 10:39:03.101981
35825	4	1-9175	35822	2	4	70	11	2012-08-02 10:39:03.101981
35822	4	1-10951	35822	4	32	221	73	2012-08-02 10:39:03.101981
35827	5	1-10940	35826	19	51	3266	45	2012-08-02 10:39:03.101981
35828	5	1-10194	35826	51	92	2631	61	2012-08-02 10:39:03.101981
35829	5	1-8955	35826	92	43	841	63	2012-08-02 10:39:03.101981
35830	5	1-10638	35826	43	4	681	11	2012-08-02 10:39:03.101981
35826	5	1-10951	35826	4	19	2436	73	2012-08-02 10:39:03.101981
35832	5	1-10892	35831	32	26	3358	17	2012-08-02 10:39:03.101981
35833	5	1-9376	35831	26	91	1588	79	2012-08-02 10:39:03.101981
35834	5	1-10754	35831	91	43	304	63	2012-08-02 10:39:03.101981
35835	5	1-10638	35831	43	4	258	11	2012-08-02 10:39:03.101981
35831	5	1-10951	35831	4	32	967	73	2012-08-02 10:39:03.101981
35837	4	1-10845	35836	89	1	4226	19	2012-08-02 10:39:03.101981
35838	4	1-10330	35836	1	68	3549	41	2012-08-02 10:39:03.101981
35839	4	1-7740	35836	68	4	1155	11	2012-08-02 10:39:03.101981
35836	4	1-10951	35836	4	89	4704	73	2012-08-02 10:39:03.101981
35848	4	1-7371	35847	22	32	401	48	2012-08-02 10:39:06.694644
35849	4	1-10862	35847	32	33	450	52	2012-08-02 10:39:06.694644
35850	4	1-10580	35847	33	49	819	47	2012-08-02 10:39:06.694644
35847	4	1-10960	35847	49	22	858	1	2012-08-02 10:39:06.694644
35855	2	1-10304	35854	69	92	527	80	2012-08-02 10:39:07.866336
35854	2	1-10963	35854	92	69	1357	96	2012-08-02 10:39:07.866336
35865	4	1-10870	35864	63	17	229	83	2012-08-02 10:39:09.04928
35866	4	1-10138	35864	17	72	266	60	2012-08-02 10:39:09.04928
35867	4	1-10407	35864	72	45	117	29	2012-08-02 10:39:09.04928
35864	4	1-10967	35864	45	63	138	57	2012-08-02 10:39:09.04928
35869	4	1-10048	35868	62	56	74	55	2012-08-02 10:39:09.04928
35870	4	1-8980	35868	56	20	137	5	2012-08-02 10:39:09.04928
35871	4	1-10203	35868	20	45	93	29	2012-08-02 10:39:09.04928
35868	4	1-10967	35868	45	62	115	57	2012-08-02 10:39:09.04928
35873	3	1-8881	35872	4	73	133	12	2012-08-02 10:39:09.04928
35874	3	1-6499	35872	73	45	93	29	2012-08-02 10:39:09.04928
35872	3	1-10967	35872	45	4	118	57	2012-08-02 10:39:09.04928
35876	4	1-10870	35875	63	17	1789	83	2012-08-02 10:39:09.04928
35877	4	1-10138	35875	17	77	2254	60	2012-08-02 10:39:09.04928
35878	4	1-7564	35875	77	45	775	29	2012-08-02 10:39:09.04928
35875	4	1-10967	35875	45	63	994	57	2012-08-02 10:39:09.04928
35879	1	1-2840	\N	46	46	153	83	2012-08-02 10:39:09.04928
35885	3	1-10978	35884	93	83	675	70	2012-08-02 10:39:14.112605
35886	3	1-3602	35884	83	9	81	64	2012-08-02 10:39:14.112605
35884	3	1-10979	35884	9	93	879	7	2012-08-02 10:39:14.112605
35888	4	1-10874	35887	61	34	8627	44	2012-08-02 10:39:14.112605
35889	4	1-7023	35887	34	80	1076	8	2012-08-02 10:39:14.112605
35890	4	1-8484	35887	80	9	388	64	2012-08-02 10:39:14.112605
35887	4	1-10979	35887	9	61	4212	7	2012-08-02 10:39:14.112605
35892	4	1-10685	35891	16	49	87	91	2012-08-02 10:39:14.112605
35893	4	1-10465	35891	49	32	79	58	2012-08-02 10:39:14.112605
35894	4	1-10473	35891	32	9	18	64	2012-08-02 10:39:14.112605
35891	4	1-10979	35891	9	16	222	7	2012-08-02 10:39:14.112605
35953	4	1-10561	35952	57	66	1626	28	2012-08-02 10:39:20.791409
35954	4	1-9303	35952	66	4	723	11	2012-08-02 10:39:20.791409
35955	4	1-10986	35952	4	49	3292	62	2012-08-02 10:39:20.791409
35952	4	1-10995	35952	49	57	3485	27	2012-08-02 10:39:20.791409
35971	4	1-10611	35970	56	29	281	50	2012-08-02 10:39:22.338804
35972	4	1-10180	35970	29	12	227	9	2012-08-02 10:39:22.338804
35973	4	1-7843	35970	12	28	208	46	2012-08-02 10:39:22.338804
35970	4	1-10999	35970	28	56	156	83	2012-08-02 10:39:22.338804
35661	5	1-10330	35660	1	9	1355	41	2012-08-02 10:38:41.38223
35662	5	1-10831	35660	9	17	3389	46	2012-08-02 10:38:41.38223
35663	5	1-10125	35660	17	92	1197	61	2012-08-02 10:38:41.38223
35664	5	1-8955	35660	92	100	467	63	2012-08-02 10:38:41.38223
35660	5	1-10895	35660	100	1	1507	19	2012-08-02 10:38:41.38223
35677	4	1-10577	35676	76	36	1024	13	2012-08-02 10:38:42.166168
35678	4	1-10496	35676	36	21	785	1	2012-08-02 10:38:42.166168
35679	4	1-7669	35676	21	33	367	97	2012-08-02 10:38:42.166168
35676	4	1-10898	35676	33	76	1463	33	2012-08-02 10:38:42.166168
35681	4	1-10708	35680	64	61	2445	53	2012-08-02 10:38:42.166168
35682	4	1-10619	35680	61	68	1403	31	2012-08-02 10:38:42.166168
35683	4	1-9811	35680	68	33	856	97	2012-08-02 10:38:42.166168
35680	4	1-10898	35680	33	64	3441	33	2012-08-02 10:38:42.166168
35685	4	1-8354	35684	14	94	1178	69	2012-08-02 10:38:42.166168
35686	4	1-10725	35684	94	21	849	1	2012-08-02 10:38:42.166168
35687	4	1-7669	35684	21	33	433	97	2012-08-02 10:38:42.166168
35684	4	1-10898	35684	33	14	1880	33	2012-08-02 10:38:42.166168
35691	3	1-10794	35690	89	61	1254	43	2012-08-02 10:38:44.110778
35692	3	1-3242	35690	61	68	459	51	2012-08-02 10:38:44.110778
35690	3	1-10903	35690	68	89	3243	12	2012-08-02 10:38:44.110778
35694	4	1-10558	35693	88	21	1782	72	2012-08-02 10:38:45.281993
35695	4	1-7295	35693	21	45	820	2	2012-08-02 10:38:45.281993
35696	4	1-10717	35693	45	78	334	44	2012-08-02 10:38:45.281993
35693	4	1-10906	35693	78	88	9035	92	2012-08-02 10:38:45.281993
35703	3	1-9376	35702	26	1	2129	79	2012-08-02 10:38:46.442004
35704	3	1-10900	35702	1	87	465	26	2012-08-02 10:38:46.442004
35702	3	1-10909	35702	87	26	4896	17	2012-08-02 10:38:46.442004
35710	4	1-9997	35709	41	60	633	35	2012-08-02 10:38:47.624215
35711	4	1-1317	35709	60	66	169	43	2012-08-02 10:38:47.624215
35712	4	1-9570	35709	66	89	84	81	2012-08-02 10:38:47.624215
35709	4	1-10912	35709	89	41	2050	16	2012-08-02 10:38:47.624215
35714	4	1-10674	35713	31	73	799	80	2012-08-02 10:38:47.624215
35715	4	1-8995	35713	73	93	225	77	2012-08-02 10:38:47.624215
35716	4	1-10312	35713	93	89	97	81	2012-08-02 10:38:47.624215
35713	4	1-10912	35713	89	31	2552	16	2012-08-02 10:38:47.624215
35718	2	1-7890	35717	51	48	801	79	2012-08-02 10:38:48.795357
35717	2	1-10915	35717	48	51	1422	61	2012-08-02 10:38:48.795357
35728	4	1-10773	35727	35	87	1873	84	2012-08-02 10:38:50.004189
35729	4	1-10510	35727	87	32	2341	85	2012-08-02 10:38:50.004189
35730	4	1-10620	35727	32	90	1506	65	2012-08-02 10:38:50.004189
35727	4	1-10918	35727	90	35	4616	80	2012-08-02 10:38:50.004189
35747	2	1-10356	35746	15	72	774	88	2012-08-02 10:38:51.181209
35746	2	1-10921	35746	72	15	313	55	2012-08-02 10:38:51.181209
35786	4	1-10597	35785	80	48	3969	18	2012-08-02 10:38:55.979849
35787	4	1-8572	35785	48	28	5635	42	2012-08-02 10:38:55.979849
35788	4	1-10877	35785	28	24	5462	16	2012-08-02 10:38:55.979849
35785	4	1-10932	35785	24	80	3476	1	2012-08-02 10:38:55.979849
35790	4	1-10597	35789	80	81	1384	18	2012-08-02 10:38:55.979849
35791	4	1-8988	35789	81	90	1364	27	2012-08-02 10:38:55.979849
35792	4	1-10664	35789	90	24	1768	16	2012-08-02 10:38:55.979849
35789	4	1-10932	35789	24	80	1168	1	2012-08-02 10:38:55.979849
35794	4	1-10597	35793	80	1	79	18	2012-08-02 10:38:55.979849
35795	4	1-10128	35793	1	27	88	34	2012-08-02 10:38:55.979849
35796	4	1-8971	35793	27	24	88	16	2012-08-02 10:38:55.979849
35793	4	1-10932	35793	24	80	62	1	2012-08-02 10:38:55.979849
35798	4	1-9523	35797	20	29	826	11	2012-08-02 10:38:55.979849
35799	4	1-10093	35797	29	90	1278	27	2012-08-02 10:38:55.979849
35800	4	1-10664	35797	90	24	1808	16	2012-08-02 10:38:55.979849
35797	4	1-10932	35797	24	20	1303	1	2012-08-02 10:38:55.979849
35802	3	1-8885	35801	72	28	1178	42	2012-08-02 10:38:55.979849
35803	3	1-10877	35801	28	24	1349	16	2012-08-02 10:38:55.979849
35801	3	1-10932	35801	24	72	1015	1	2012-08-02 10:38:55.979849
35805	2	1-10535	35804	4	40	8831	70	2012-08-02 10:38:57.97117
35804	2	1-10938	35804	40	4	7962	45	2012-08-02 10:38:57.97117
35811	4	1-9838	35810	1	54	1419	34	2012-08-02 10:38:59.142666
35812	4	1-10887	35810	54	60	1719	12	2012-08-02 10:38:59.142666
35813	4	1-10856	35810	60	6	1521	23	2012-08-02 10:38:59.142666
35810	4	1-10941	35810	6	1	1342	86	2012-08-02 10:38:59.142666
35841	3	1-9608	35840	47	92	683	68	2012-08-02 10:39:04.749561
35842	3	1-9386	35840	92	82	612	85	2012-08-02 10:39:04.749561
35840	3	1-10955	35840	82	47	1510	23	2012-08-02 10:39:04.749561
35844	4	1-10357	35843	78	84	377	40	2012-08-02 10:39:04.749561
35845	4	1-10526	35843	84	23	231	77	2012-08-02 10:39:04.749561
35846	4	1-7526	35843	23	82	121	85	2012-08-02 10:39:04.749561
35843	4	1-10955	35843	82	78	304	23	2012-08-02 10:39:04.749561
35857	4	1-9747	35856	95	28	3328	2	2012-08-02 10:39:08.241849
35858	4	1-10695	35856	28	34	1456	79	2012-08-02 10:39:08.241849
35859	4	1-10897	35856	34	37	1781	60	2012-08-02 10:39:08.241849
35856	4	1-10964	35856	37	95	1959	97	2012-08-02 10:39:08.241849
35861	4	1-10582	35860	36	57	1126	94	2012-08-02 10:39:08.241849
35862	4	1-9785	35860	57	17	1556	83	2012-08-02 10:39:08.241849
35863	4	1-10138	35860	17	37	1368	60	2012-08-02 10:39:08.241849
35860	4	1-10964	35860	37	36	1531	97	2012-08-02 10:39:08.241849
35896	4	1-8445	35895	6	53	1327	70	2012-08-02 10:39:14.908614
35897	4	1-10867	35895	53	21	1012	3	2012-08-02 10:39:14.908614
35898	4	1-8864	35895	21	95	478	91	2012-08-02 10:39:14.908614
35895	4	1-10981	35895	95	6	1505	25	2012-08-02 10:39:14.908614
35900	4	1-9663	35899	11	84	1566	56	2012-08-02 10:39:14.908614
35901	4	1-10529	35899	84	61	1459	97	2012-08-02 10:39:14.908614
35902	4	1-10278	35899	61	95	976	91	2012-08-02 10:39:14.908614
35899	4	1-10981	35899	95	11	3648	25	2012-08-02 10:39:14.908614
35904	4	1-8697	35903	92	84	946	56	2012-08-02 10:39:14.908614
35905	4	1-10529	35903	84	61	1033	97	2012-08-02 10:39:14.908614
35906	4	1-10278	35903	61	95	809	91	2012-08-02 10:39:14.908614
35903	4	1-10981	35903	95	92	3543	25	2012-08-02 10:39:14.908614
35910	4	1-10684	35909	29	54	2464	8	2012-08-02 10:39:16.081614
35911	4	1-7075	35909	54	76	2011	54	2012-08-02 10:39:16.081614
35912	4	1-4602	35909	76	83	1106	92	2012-08-02 10:39:16.081614
35909	4	1-10984	35909	83	29	2139	93	2012-08-02 10:39:16.081614
35914	4	1-9546	35913	10	45	377	1	2012-08-02 10:39:16.081614
35915	4	1-10738	35913	45	79	332	14	2012-08-02 10:39:16.081614
35916	4	1-10876	35913	79	83	369	92	2012-08-02 10:39:16.081614
35913	4	1-10984	35913	83	10	741	93	2012-08-02 10:39:16.081614
35918	4	1-8824	35917	40	85	1842	65	2012-08-02 10:39:16.081614
35919	4	1-10387	35917	85	79	1795	14	2012-08-02 10:39:16.081614
35920	4	1-10876	35917	79	83	2007	92	2012-08-02 10:39:16.081614
35917	4	1-10984	35917	83	40	4056	93	2012-08-02 10:39:16.081614
35922	4	1-10856	35921	60	6	425	23	2012-08-02 10:39:18.424847
35923	4	1-10941	35921	6	39	305	86	2012-08-02 10:39:18.424847
35924	4	1-10331	35921	39	5	492	60	2012-08-02 10:39:18.424847
35921	4	1-10989	35921	5	60	591	12	2012-08-02 10:39:18.424847
35942	4	1-4003	35941	46	70	125	24	2012-08-02 10:39:19.198296
35943	4	1-7696	35941	70	80	80	1	2012-08-02 10:39:19.198296
35944	4	1-10597	35941	80	3	92	18	2012-08-02 10:39:19.198296
35941	4	1-10991	35941	3	46	645	68	2012-08-02 10:39:19.198296
35946	3	1-8719	35945	85	45	1679	7	2012-08-02 10:39:19.198296
35947	3	1-6895	35945	45	3	385	18	2012-08-02 10:39:19.198296
35945	3	1-10991	35945	3	85	2735	68	2012-08-02 10:39:19.198296
35949	4	1-9888	35948	98	57	2430	27	2012-08-02 10:39:19.198296
35950	4	1-10561	35948	57	26	1249	28	2012-08-02 10:39:19.198296
35951	4	1-10175	35948	26	3	693	18	2012-08-02 10:39:19.198296
35948	4	1-10991	35948	3	98	5213	68	2012-08-02 10:39:19.198296
35931	3	1-10258	35929	37	99	130	51	2012-08-02 10:39:18.027505
35929	3	1-10990	35929	99	19	694	66	2012-08-02 10:39:18.027505
35933	4	1-10448	35932	1	95	9	32	2012-08-02 10:39:18.027505
35934	4	1-10158	35932	95	23	7	10	2012-08-02 10:39:18.027505
35935	4	1-10715	35932	23	99	6	51	2012-08-02 10:39:18.027505
35932	4	1-10990	35932	99	1	33	66	2012-08-02 10:39:18.027505
35937	3	1-10320	35936	52	32	542	88	2012-08-02 10:39:18.027505
35938	3	1-10485	35936	32	99	114	51	2012-08-02 10:39:18.027505
35936	3	1-10990	35936	99	52	639	66	2012-08-02 10:39:18.027505
35939	1	1-6974	\N	42	42	48	51	2012-08-02 10:39:18.027505
35940	1	1-8172	\N	9	9	1991	51	2012-08-02 10:39:18.027505
35957	4	1-10548	35956	59	49	346	47	2012-08-02 10:39:21.178299
35958	4	1-10960	35956	49	80	324	1	2012-08-02 10:39:21.178299
35959	4	1-10597	35956	80	54	385	18	2012-08-02 10:39:21.178299
35956	4	1-10996	35956	54	59	648	71	2012-08-02 10:39:21.178299
35961	3	1-9082	35960	91	43	1196	95	2012-08-02 10:39:21.951995
35962	3	1-10869	35960	43	23	1834	4	2012-08-02 10:39:21.951995
35960	3	1-10998	35960	23	91	2337	11	2012-08-02 10:39:21.951995
35964	3	1-8804	35963	47	39	299	85	2012-08-02 10:39:21.951995
35965	3	1-10741	35963	39	23	209	4	2012-08-02 10:39:21.951995
35963	3	1-10998	35963	23	47	277	11	2012-08-02 10:39:21.951995
35967	4	1-10501	35966	66	38	1537	66	2012-08-02 10:39:21.951995
35968	4	1-8249	35966	38	43	1000	95	2012-08-02 10:39:21.951995
35969	4	1-10869	35966	43	23	1688	4	2012-08-02 10:39:21.951995
35966	4	1-10998	35966	23	66	2369	11	2012-08-02 10:39:21.951995
35991	4	1-8702	35990	49	57	831	45	2012-08-02 10:39:24.309233
35992	4	1-10860	35990	57	84	812	3	2012-08-02 10:39:24.309233
35993	4	1-9601	35990	84	37	982	8	2012-08-02 10:39:24.309233
35990	4	1-11004	35990	37	49	893	75	2012-08-02 10:39:24.309233
35995	4	1-8957	35994	52	57	2137	23	2012-08-02 10:39:25.873206
35996	4	1-7882	35994	57	44	988	69	2012-08-02 10:39:25.873206
35997	4	1-10770	35994	44	78	1408	100	2012-08-02 10:39:25.873206
35994	4	1-11008	35994	78	52	5788	34	2012-08-02 10:39:25.873206
35999	4	1-10755	35998	56	66	736	68	2012-08-02 10:39:25.873206
36000	4	1-9647	35998	66	97	502	30	2012-08-02 10:39:25.873206
36001	4	1-7426	35998	97	78	388	100	2012-08-02 10:39:25.873206
35998	4	1-11008	35998	78	56	1622	34	2012-08-02 10:39:25.873206
36003	4	1-10940	36002	19	57	1154	45	2012-08-02 10:39:26.673253
36004	4	1-10860	36002	57	100	989	3	2012-08-02 10:39:26.673253
36005	4	1-7314	36002	100	10	351	49	2012-08-02 10:39:26.673253
36002	4	1-11010	36002	10	19	821	73	2012-08-02 10:39:26.673253
36011	3	1-9942	36010	85	4	339	55	2012-08-02 10:39:28.229163
36012	3	1-10433	36010	4	36	342	4	2012-08-02 10:39:28.229163
36010	3	1-11014	36010	36	85	1499	17	2012-08-02 10:39:28.229163
36014	4	1-10844	36013	50	74	214	99	2012-08-02 10:39:28.229163
36015	4	1-9437	36013	74	4	72	55	2012-08-02 10:39:28.229163
36016	4	1-10433	36013	4	36	158	4	2012-08-02 10:39:28.229163
36013	4	1-11014	36013	36	50	1513	17	2012-08-02 10:39:28.229163
36018	4	1-9376	36017	26	93	1103	79	2012-08-02 10:39:28.229163
36019	4	1-5952	36017	93	43	117	95	2012-08-02 10:39:28.229163
36020	4	1-10869	36017	43	36	190	4	2012-08-02 10:39:28.229163
36017	4	1-11014	36017	36	26	2051	17	2012-08-02 10:39:28.229163
36025	4	1-10522	36024	32	61	474	3	2012-08-02 10:39:29.809232
36026	4	1-10593	36024	61	36	387	13	2012-08-02 10:39:29.809232
36027	4	1-10496	36024	36	44	395	1	2012-08-02 10:39:29.809232
36024	4	1-11018	36024	44	32	763	51	2012-08-02 10:39:29.809232
35975	4	1-10611	35974	56	36	826	50	2012-08-02 10:39:22.338804
35976	4	1-10225	35974	36	53	310	89	2012-08-02 10:39:22.338804
35977	4	1-10975	35974	53	28	553	46	2012-08-02 10:39:22.338804
35974	4	1-10999	35974	28	56	436	83	2012-08-02 10:39:22.338804
35978	1	1-3130	\N	74	74	793	50	2012-08-02 10:39:22.338804
35980	4	1-10327	35979	95	98	3095	82	2012-08-02 10:39:23.148279
35981	4	1-2571	35979	98	91	1398	53	2012-08-02 10:39:23.148279
35982	4	1-9660	35979	91	33	1248	87	2012-08-02 10:39:23.148279
35979	4	1-11001	35979	33	95	2881	89	2012-08-02 10:39:23.148279
35984	3	1-1644	35983	63	91	1645	53	2012-08-02 10:39:23.148279
35985	3	1-9660	35983	91	33	1650	87	2012-08-02 10:39:23.148279
35983	3	1-11001	35983	33	63	4280	89	2012-08-02 10:39:23.148279
35987	4	1-10852	35986	98	83	3238	86	2012-08-02 10:39:23.921828
35988	4	1-1896	35986	83	99	2788	40	2012-08-02 10:39:23.921828
35989	4	1-10579	35986	99	45	2382	20	2012-08-02 10:39:23.921828
35986	4	1-11003	35986	45	98	2857	50	2012-08-02 10:39:23.921828
36007	4	1-10516	36006	31	57	259	64	2012-08-02 10:39:27.841229
36008	4	1-8365	36006	57	17	314	15	2012-08-02 10:39:27.841229
36009	4	1-10377	36006	17	64	316	21	2012-08-02 10:39:27.841229
36006	4	1-11013	36006	64	31	362	93	2012-08-02 10:39:27.841229
36022	3	1-8818	36021	70	79	2071	14	2012-08-02 10:39:29.424811
36023	3	1-10876	36021	79	87	1151	92	2012-08-02 10:39:29.424811
36021	3	1-11017	36021	87	70	6557	34	2012-08-02 10:39:29.424811
36029	4	1-10895	36028	100	82	885	19	2012-08-02 10:39:31.770731
36030	4	1-9046	36028	82	27	742	51	2012-08-02 10:39:31.770731
36031	4	1-9762	36028	27	20	483	48	2012-08-02 10:39:31.770731
36028	4	1-11023	36028	20	100	344	63	2012-08-02 10:39:31.770731
36033	4	1-9682	36032	28	23	3956	10	2012-08-02 10:39:31.770731
36034	4	1-10715	36032	23	27	4117	51	2012-08-02 10:39:31.770731
36035	4	1-9762	36032	27	20	2942	48	2012-08-02 10:39:31.770731
36032	4	1-11023	36032	20	28	2300	63	2012-08-02 10:39:31.770731
36037	3	1-10847	36036	88	19	621	24	2012-08-02 10:39:31.770731
36038	3	1-10651	36036	19	20	532	48	2012-08-02 10:39:31.770731
36036	3	1-11023	36036	20	88	435	63	2012-08-02 10:39:31.770731
36040	4	1-10895	36039	100	58	2844	19	2012-08-02 10:39:31.770731
36041	4	1-6776	36039	58	10	1141	97	2012-08-02 10:39:31.770731
36042	4	1-9818	36039	10	20	1105	48	2012-08-02 10:39:31.770731
36039	4	1-11023	36039	20	100	933	63	2012-08-02 10:39:31.770731
36044	4	1-8856	36043	67	13	882	89	2012-08-02 10:39:31.770731
36045	4	1-7433	36043	13	53	683	83	2012-08-02 10:39:31.770731
36046	4	1-9824	36043	53	20	937	48	2012-08-02 10:39:31.770731
36043	4	1-11023	36043	20	67	836	63	2012-08-02 10:39:31.770731
36048	4	1-11023	36047	20	88	1885	63	2012-08-02 10:39:34.945883
36049	4	1-10847	36047	88	73	2960	24	2012-08-02 10:39:34.945883
36050	4	1-9265	36047	73	93	3077	31	2012-08-02 10:39:34.945883
36047	4	1-11031	36047	93	20	2096	48	2012-08-02 10:39:34.945883
36052	3	1-9511	36051	72	28	1253	42	2012-08-02 10:39:37.23284
36053	3	1-10877	36051	28	53	645	16	2012-08-02 10:39:37.23284
36051	3	1-11037	36051	53	72	2335	58	2012-08-02 10:39:37.23284
36055	4	1-10018	36054	19	91	3733	13	2012-08-02 10:39:37.23284
36056	4	1-10515	36054	91	28	1947	42	2012-08-02 10:39:37.23284
36057	4	1-10877	36054	28	53	1320	16	2012-08-02 10:39:37.23284
36054	4	1-11037	36054	53	19	6284	58	2012-08-02 10:39:37.23284
36059	3	1-10018	36058	19	1	670	13	2012-08-02 10:39:37.23284
36060	3	1-5913	36058	1	53	217	16	2012-08-02 10:39:37.23284
36058	3	1-11037	36058	53	19	1080	58	2012-08-02 10:39:37.23284
36062	4	1-10987	36061	64	26	3075	28	2012-08-02 10:39:38.127857
36063	4	1-10175	36061	26	52	1543	18	2012-08-02 10:39:38.127857
36064	4	1-10783	36061	52	44	1917	89	2012-08-02 10:39:38.127857
36061	4	1-11038	36061	44	64	8134	66	2012-08-02 10:39:38.127857
36066	3	1-10847	36065	88	60	287	24	2012-08-02 11:11:08.047947
36067	3	1-7565	36065	60	49	229	79	2012-08-02 11:11:08.047947
36065	3	1-11044	36065	49	88	325	63	2012-08-02 11:11:08.047947
36069	3	1-10375	36068	20	60	158	24	2012-08-02 11:11:08.047947
36070	3	1-7565	36068	60	49	159	79	2012-08-02 11:11:08.047947
36068	3	1-11044	36068	49	20	286	63	2012-08-02 11:11:08.047947
36072	4	1-9880	36071	38	99	1833	18	2012-08-02 11:11:08.047947
36073	4	1-10891	36071	99	77	1253	97	2012-08-02 11:11:08.047947
36074	4	1-8788	36071	77	49	925	79	2012-08-02 11:11:08.047947
36071	4	1-11044	36071	49	38	1905	63	2012-08-02 11:11:08.047947
36076	4	1-8538	36075	26	50	141	98	2012-08-02 11:11:08.047947
36077	4	1-10823	36075	50	18	349	62	2012-08-02 11:11:08.047947
36078	4	1-4861	36075	18	49	127	79	2012-08-02 11:11:08.047947
36075	4	1-11044	36075	49	26	264	63	2012-08-02 11:11:08.047947
36080	4	1-9481	36079	26	31	2813	35	2012-08-02 11:11:08.047947
36081	4	1-10358	36079	31	28	1524	2	2012-08-02 11:11:08.047947
36082	4	1-10695	36079	28	49	839	79	2012-08-02 11:11:08.047947
36079	4	1-11044	36079	49	26	1762	63	2012-08-02 11:11:08.047947
36084	4	1-10929	36083	80	87	929	88	2012-08-02 11:11:08.047947
36085	4	1-10592	36083	87	84	644	54	2012-08-02 11:11:08.047947
36086	4	1-7690	36083	84	49	249	79	2012-08-02 11:11:08.047947
36083	4	1-11044	36083	49	80	536	63	2012-08-02 11:11:08.047947
36088	4	1-10526	36087	84	34	5452	77	2012-08-02 11:11:08.487871
36089	4	1-5508	36087	34	32	1346	73	2012-08-02 11:11:08.487871
36090	4	1-10892	36087	32	89	4124	17	2012-08-02 11:11:08.487871
36087	4	1-11045	36087	89	84	6730	40	2012-08-02 11:11:08.487871
36092	4	1-10526	36091	84	23	718	77	2012-08-02 11:11:08.487871
36093	4	1-7526	36091	23	42	560	85	2012-08-02 11:11:08.487871
36094	4	1-10961	36091	42	89	427	17	2012-08-02 11:11:08.487871
36091	4	1-11045	36091	89	84	786	40	2012-08-02 11:11:08.487871
36096	4	1-8748	36095	27	50	2816	98	2012-08-02 11:11:08.876498
36097	4	1-10823	36095	50	35	7944	62	2012-08-02 11:11:08.876498
36098	4	1-4572	36095	35	22	3025	93	2012-08-02 11:11:08.876498
36095	4	1-11046	36095	22	27	2708	99	2012-08-02 11:11:08.876498
36100	4	1-4709	36099	52	12	4114	28	2012-08-02 11:11:08.876498
36101	4	1-10578	36099	12	64	4426	21	2012-08-02 11:11:08.876498
36102	4	1-11013	36099	64	22	4866	93	2012-08-02 11:11:08.876498
36099	4	1-11046	36099	22	52	4456	99	2012-08-02 11:11:08.876498
36104	3	1-9879	36103	62	29	2822	50	2012-08-02 11:11:09.687486
36105	3	1-10180	36103	29	11	1983	9	2012-08-02 11:11:09.687486
36103	3	1-11049	36103	11	62	1910	56	2012-08-02 11:11:09.687486
36107	4	1-10936	36106	66	6	1025	25	2012-08-02 11:11:09.687486
36108	4	1-8445	36106	6	51	1256	70	2012-08-02 11:11:09.687486
36109	4	1-10910	36106	51	11	707	9	2012-08-02 11:11:09.687486
36106	4	1-11049	36106	11	66	817	56	2012-08-02 11:11:09.687486
36111	4	1-10823	36110	50	49	671	62	2012-08-02 11:11:10.096452
36112	4	1-7923	36110	49	26	335	15	2012-08-02 11:11:10.096452
36113	4	1-10247	36110	26	31	219	5	2012-08-02 11:11:10.096452
36110	4	1-11050	36110	31	50	386	98	2012-08-02 11:11:10.096452
36115	4	1-10935	36114	13	59	3639	71	2012-08-02 11:11:10.096452
36116	4	1-10548	36114	59	100	1859	47	2012-08-02 11:11:10.096452
36117	4	1-2216	36114	100	31	1203	5	2012-08-02 11:11:10.096452
36114	4	1-11050	36114	31	13	3073	98	2012-08-02 11:11:10.096452
36119	4	1-10935	36118	13	56	449	71	2012-08-02 11:11:10.096452
36120	4	1-7140	36118	56	34	192	100	2012-08-02 11:11:10.096452
36121	4	1-10092	36118	34	31	148	5	2012-08-02 11:11:10.096452
36118	4	1-11050	36118	31	13	378	98	2012-08-02 11:11:10.096452
36123	4	1-10403	36122	54	93	2635	25	2012-08-02 11:11:10.096452
36124	4	1-10386	36122	93	74	1020	96	2012-08-02 11:11:10.096452
36125	4	1-8851	36122	74	31	1229	5	2012-08-02 11:11:10.096452
36122	4	1-11050	36122	31	54	3277	98	2012-08-02 11:11:10.096452
36127	4	1-10669	36126	41	42	4003	42	2012-08-02 11:11:10.484443
36128	4	1-9649	36126	42	33	1766	40	2012-08-02 11:11:10.484443
36129	4	1-10776	36126	33	3	818	72	2012-08-02 11:11:10.484443
36126	4	1-11051	36126	3	41	4891	41	2012-08-02 11:11:10.484443
36131	3	1-10257	36130	12	77	1579	22	2012-08-02 11:11:10.484443
36132	3	1-6886	36130	77	3	569	72	2012-08-02 11:11:10.484443
36130	3	1-11051	36130	3	12	3418	41	2012-08-02 11:11:10.484443
36134	2	1-6141	36133	8	5	172	93	2012-08-02 11:11:10.898383
36133	2	1-11052	36133	5	8	329	27	2012-08-02 11:11:10.898383
36136	4	1-10561	36135	57	12	313	28	2012-08-02 11:11:10.898383
36137	4	1-10578	36135	12	64	294	21	2012-08-02 11:11:10.898383
36138	4	1-11013	36135	64	5	283	93	2012-08-02 11:11:10.898383
36135	4	1-11052	36135	5	57	657	27	2012-08-02 11:11:10.898383
36140	2	1-10004	36139	88	5	1123	93	2012-08-02 11:11:10.898383
36139	2	1-11052	36139	5	88	2764	27	2012-08-02 11:11:10.898383
36142	4	1-9844	36141	64	7	1405	20	2012-08-02 11:11:11.672045
36143	4	1-9029	36141	7	27	766	34	2012-08-02 11:11:11.672045
36144	4	1-8971	36141	27	97	757	16	2012-08-02 11:11:11.672045
36141	4	1-11054	36141	97	64	760	96	2012-08-02 11:11:11.672045
36146	4	1-5037	36145	91	43	651	95	2012-08-02 11:11:13.288184
36147	4	1-10869	36145	43	36	680	4	2012-08-02 11:11:13.288184
36148	4	1-11014	36145	36	39	4728	17	2012-08-02 11:11:13.288184
36145	4	1-11059	36145	39	91	2527	41	2012-08-02 11:11:13.288184
36150	4	1-10230	36149	95	18	715	12	2012-08-02 11:11:13.288184
36151	4	1-11036	36149	18	32	497	73	2012-08-02 11:11:13.288184
36152	4	1-10892	36149	32	39	1527	17	2012-08-02 11:11:13.288184
36149	4	1-11059	36149	39	95	988	41	2012-08-02 11:11:13.288184
36154	4	1-10300	36153	40	16	1344	64	2012-08-02 11:11:14.072843
36155	4	1-10439	36153	16	74	1549	96	2012-08-02 11:11:14.072843
36156	4	1-10088	36153	74	52	983	8	2012-08-02 11:11:14.072843
36153	4	1-11061	36153	52	40	1764	43	2012-08-02 11:11:14.072843
36158	4	1-9764	36157	71	45	2456	20	2012-08-02 11:11:14.847341
36159	4	1-11003	36157	45	39	3179	50	2012-08-02 11:11:14.847341
36160	4	1-10041	36157	39	29	1772	7	2012-08-02 11:11:14.847341
36157	4	1-11063	36157	29	71	2927	100	2012-08-02 11:11:14.847341
36162	3	1-11063	36161	29	71	262	100	2012-08-02 11:11:15.632359
36163	3	1-9764	36161	71	85	138	20	2012-08-02 11:11:15.632359
36161	3	1-11065	36161	85	29	252	7	2012-08-02 11:11:15.632359
36165	4	1-11063	36164	29	5	3749	100	2012-08-02 11:11:15.632359
36166	4	1-10742	36164	5	10	1645	47	2012-08-02 11:11:15.632359
36167	4	1-7430	36164	10	85	956	20	2012-08-02 11:11:15.632359
36164	4	1-11065	36164	85	29	2513	7	2012-08-02 11:11:15.632359
36169	4	1-8427	36168	73	84	311	40	2012-08-02 11:11:16.030855
36170	4	1-10526	36168	84	93	277	77	2012-08-02 11:11:16.030855
36171	4	1-10312	36168	93	3	175	81	2012-08-02 11:11:16.030855
36168	4	1-11066	36168	3	73	354	55	2012-08-02 11:11:16.030855
36173	4	1-8343	36172	63	34	1761	38	2012-08-02 11:11:16.030855
36174	4	1-9294	36172	34	93	1575	77	2012-08-02 11:11:16.030855
36175	4	1-10312	36172	93	3	1001	81	2012-08-02 11:11:16.030855
36172	4	1-11066	36172	3	63	2044	55	2012-08-02 11:11:16.030855
36177	2	1-5987	36176	44	3	3091	81	2012-08-02 11:11:16.030855
36176	2	1-11066	36176	3	44	6448	55	2012-08-02 11:11:16.030855
36179	4	1-9176	36178	24	92	1355	18	2012-08-02 11:11:16.450807
36180	4	1-7967	36178	92	99	1900	40	2012-08-02 11:11:16.450807
36181	4	1-10579	36178	99	18	1796	20	2012-08-02 11:11:16.450807
36178	4	1-11067	36178	18	24	1804	67	2012-08-02 11:11:16.450807
36183	3	1-4934	36182	6	98	722	57	2012-08-02 11:11:16.450807
36184	3	1-11028	36182	98	18	2129	20	2012-08-02 11:11:16.450807
36182	3	1-11067	36182	18	6	2348	67	2012-08-02 11:11:16.450807
36186	4	1-9854	36185	4	64	62	21	2012-08-02 11:11:18.022283
36187	4	1-11013	36185	64	14	50	93	2012-08-02 11:11:18.022283
36188	4	1-10809	36185	14	93	23	74	2012-08-02 11:11:18.022283
36185	4	1-11071	36185	93	4	68	6	2012-08-02 11:11:18.022283
36190	5	1-10837	36189	96	12	3290	28	2012-08-02 11:11:18.022283
36191	5	1-10578	36189	12	64	2989	21	2012-08-02 11:11:18.022283
36192	5	1-11013	36189	64	14	2774	93	2012-08-02 11:11:18.022283
36193	5	1-10809	36189	14	93	1454	74	2012-08-02 11:11:18.022283
36189	5	1-11071	36189	93	96	5075	6	2012-08-02 11:11:18.022283
36195	4	1-10837	36194	96	57	1887	28	2012-08-02 11:11:18.022283
36196	4	1-10899	36194	57	92	2320	94	2012-08-02 11:11:18.022283
36197	4	1-8730	36194	92	93	810	74	2012-08-02 11:11:18.022283
36194	4	1-11071	36194	93	96	2868	6	2012-08-02 11:11:18.022283
36199	4	1-11043	36198	29	92	4916	46	2012-08-02 11:11:18.795729
36200	4	1-9591	36198	92	68	2202	28	2012-08-02 11:11:18.795729
36201	4	1-10480	36198	68	87	3668	25	2012-08-02 11:11:18.795729
36198	4	1-11072	36198	87	29	4003	49	2012-08-02 11:11:18.795729
36203	4	1-8774	36202	75	93	413	90	2012-08-02 11:11:18.795729
36204	4	1-10463	36202	93	68	394	28	2012-08-02 11:11:18.795729
36205	4	1-10480	36202	68	87	735	25	2012-08-02 11:11:18.795729
36202	4	1-11072	36202	87	75	898	49	2012-08-02 11:11:18.795729
36207	4	1-10629	36206	19	53	5230	31	2012-08-02 11:11:20.766065
36208	4	1-10144	36206	53	38	5150	23	2012-08-02 11:11:20.766065
36209	4	1-10731	36206	38	41	3753	41	2012-08-02 11:11:20.766065
36206	4	1-11078	36206	41	19	3516	43	2012-08-02 11:11:20.766065
36211	2	1-11018	36210	44	46	1811	51	2012-08-02 11:11:21.153363
36210	2	1-11079	36210	46	44	4031	1	2012-08-02 11:11:21.153363
36212	4	1-11081	36212	8	29	142	7	2012-08-02 11:11:21.937739
36213	4	1-11063	36212	29	5	157	100	2012-08-02 11:11:21.937739
36214	4	1-10742	36212	5	10	51	47	2012-08-02 11:11:21.937739
36215	4	1-7430	36212	10	8	22	20	2012-08-02 11:11:21.937739
36217	4	1-8012	36216	99	83	1384	28	2012-08-02 11:11:21.937739
36218	4	1-8809	36216	83	99	758	40	2012-08-02 11:11:21.937739
36219	4	1-10579	36216	99	8	636	20	2012-08-02 11:11:21.937739
36216	4	1-11081	36216	8	99	4431	7	2012-08-02 11:11:21.937739
36221	4	1-6895	36220	45	92	317	18	2012-08-02 11:11:21.937739
36222	4	1-7967	36220	92	99	333	40	2012-08-02 11:11:21.937739
36223	4	1-10579	36220	99	8	236	20	2012-08-02 11:11:21.937739
36220	4	1-11081	36220	8	45	1745	7	2012-08-02 11:11:21.937739
36251	3	1-9473	36250	92	63	2067	91	2012-08-02 11:11:25.387474
36252	3	1-587	36250	63	90	1267	48	2012-08-02 11:11:25.387474
36250	3	1-11090	36250	90	92	3541	73	2012-08-02 11:11:25.387474
36254	4	1-10845	36253	89	58	2609	19	2012-08-02 11:11:25.387474
36255	4	1-6776	36253	58	10	1041	97	2012-08-02 11:11:25.387474
36256	4	1-9818	36253	10	90	1002	48	2012-08-02 11:11:25.387474
36253	4	1-11090	36253	90	89	2892	73	2012-08-02 11:11:25.387474
36257	1	1-587	\N	63	63	1292	48	2012-08-02 11:11:25.387474
36275	4	1-11048	36274	66	51	1145	70	2012-08-02 11:11:26.592527
36276	4	1-10910	36274	51	89	594	9	2012-08-02 11:11:26.592527
36277	4	1-6664	36274	89	55	459	53	2012-08-02 11:11:26.592527
36274	4	1-11093	36274	55	66	1385	46	2012-08-02 11:11:26.592527
36279	3	1-10957	36278	37	48	1860	29	2012-08-02 11:11:27.753399
36280	3	1-10712	36278	48	50	1436	58	2012-08-02 11:11:27.753399
36278	3	1-11096	36278	50	37	2383	88	2012-08-02 11:11:27.753399
36282	4	1-10592	36281	87	67	3448	54	2012-08-02 11:11:27.753399
36283	4	1-8205	36281	67	49	1324	91	2012-08-02 11:11:27.753399
36284	4	1-10465	36281	49	50	1622	58	2012-08-02 11:11:27.753399
36281	4	1-11096	36281	50	87	4699	88	2012-08-02 11:11:27.753399
36286	4	1-10921	36285	72	78	1303	55	2012-08-02 11:11:27.753399
36287	4	1-9472	36285	78	8	1100	59	2012-08-02 11:11:27.753399
36288	4	1-10569	36285	8	50	917	58	2012-08-02 11:11:27.753399
36285	4	1-11096	36285	50	72	2715	88	2012-08-02 11:11:27.753399
36301	4	1-9426	36300	17	99	470	18	2012-08-02 11:11:31.223945
36302	4	1-10891	36300	99	38	320	97	2012-08-02 11:11:31.223945
36303	4	1-10636	36300	38	67	484	51	2012-08-02 11:11:31.223945
36300	4	1-11105	36300	67	17	569	36	2012-08-02 11:11:31.223945
36305	4	1-10543	36304	88	97	1979	16	2012-08-02 11:11:31.223945
36306	4	1-11054	36304	97	35	1979	96	2012-08-02 11:11:31.223945
36307	4	1-2974	36304	35	67	1178	51	2012-08-02 11:11:31.223945
36304	4	1-11105	36304	67	88	1583	36	2012-08-02 11:11:31.223945
36342	4	1-10403	36341	54	87	1471	25	2012-08-02 11:11:37.083093
36343	4	1-11072	36341	87	21	1282	49	2012-08-02 11:11:37.083093
36344	4	1-5862	36341	21	42	821	58	2012-08-02 11:11:37.083093
36341	4	1-11120	36341	42	54	2358	98	2012-08-02 11:11:37.083093
36346	4	1-7528	36345	74	37	1578	100	2012-08-02 11:11:37.083093
36347	4	1-10894	36345	37	8	1075	59	2012-08-02 11:11:37.083093
36348	4	1-10569	36345	8	42	930	58	2012-08-02 11:11:37.083093
36345	4	1-11120	36345	42	74	3744	98	2012-08-02 11:11:37.083093
36349	1	1-5862	\N	21	21	3761	58	2012-08-02 11:11:37.083093
36370	4	1-10960	36369	49	83	1161	1	2012-08-02 11:11:42.254873
36371	4	1-9842	36369	83	52	1580	20	2012-08-02 11:11:42.254873
36372	4	1-10746	36369	52	15	466	56	2012-08-02 11:11:42.254873
36369	4	1-11133	36369	15	49	1283	47	2012-08-02 11:11:42.254873
36374	4	1-10960	36373	49	6	1279	1	2012-08-02 11:11:42.254873
36375	4	1-7347	36373	6	63	1007	95	2012-08-02 11:11:42.254873
36376	4	1-10827	36373	63	15	507	56	2012-08-02 11:11:42.254873
36373	4	1-11133	36373	15	49	1404	47	2012-08-02 11:11:42.254873
36408	4	1-10349	36407	74	63	826	72	2012-08-02 11:11:46.957424
36409	4	1-9856	36407	63	2	918	54	2012-08-02 11:11:46.957424
36410	4	1-9000	36407	2	2	514	77	2012-08-02 11:11:46.957424
36407	4	1-11145	36407	2	74	1163	89	2012-08-02 11:11:46.957424
36412	3	1-10505	36411	67	73	3730	80	2012-08-02 11:11:46.957424
36413	3	1-8995	36411	73	2	1571	77	2012-08-02 11:11:46.957424
36411	3	1-11145	36411	2	67	3557	89	2012-08-02 11:11:46.957424
36415	4	1-10975	36414	53	2	2818	46	2012-08-02 11:11:46.957424
36416	4	1-7518	36414	2	94	862	62	2012-08-02 11:11:46.957424
36417	4	1-9616	36414	94	2	814	77	2012-08-02 11:11:46.957424
36414	4	1-11145	36414	2	53	1972	89	2012-08-02 11:11:46.957424
36430	4	1-9946	36429	86	65	538	38	2012-08-02 11:11:48.140241
36431	4	1-10977	36429	65	12	500	28	2012-08-02 11:11:48.140241
36432	4	1-10578	36429	12	92	505	21	2012-08-02 11:11:48.140241
36429	4	1-11148	36429	92	86	346	75	2012-08-02 11:11:48.140241
36438	2	1-8467	36437	36	13	2120	85	2012-08-02 11:11:49.322907
36437	2	1-11151	36437	13	36	2667	61	2012-08-02 11:11:49.322907
36440	3	1-7890	36439	51	10	1677	79	2012-08-02 11:11:49.322907
36441	3	1-9695	36439	10	13	1552	85	2012-08-02 11:11:49.322907
36439	3	1-11151	36439	13	51	2454	61	2012-08-02 11:11:49.322907
36443	2	1-3495	36442	51	13	1473	85	2012-08-02 11:11:49.322907
36442	2	1-11151	36442	13	51	2542	61	2012-08-02 11:11:49.322907
36445	3	1-10157	36444	32	18	680	58	2012-08-02 11:11:49.322907
36446	3	1-6494	36444	18	13	271	85	2012-08-02 11:11:49.322907
36444	3	1-11151	36444	13	32	485	61	2012-08-02 11:11:49.322907
36447	1	1-2973	\N	100	100	2667	58	2012-08-02 11:11:49.322907
36451	3	1-10257	36450	12	7	4146	22	2012-08-02 11:11:50.505526
36452	3	1-6437	36450	7	90	3747	54	2012-08-02 11:11:50.505526
36450	3	1-11154	36450	90	12	6324	41	2012-08-02 11:11:50.505526
36454	4	1-11078	36453	41	59	573	43	2012-08-02 11:11:50.505526
36455	4	1-10383	36453	59	3	600	90	2012-08-02 11:11:50.505526
36456	4	1-6652	36453	3	90	370	54	2012-08-02 11:11:50.505526
36453	4	1-11154	36453	90	41	640	41	2012-08-02 11:11:50.505526
36487	3	1-10921	36486	72	56	141	55	2012-08-02 11:11:54.052869
36488	3	1-8980	36486	56	57	253	5	2012-08-02 11:11:54.052869
36486	3	1-11163	36486	57	72	300	88	2012-08-02 11:11:54.052869
36490	4	1-10087	36489	27	45	820	9	2012-08-02 11:11:54.052869
36491	4	1-10714	36489	45	26	1551	15	2012-08-02 11:11:54.052869
36492	4	1-10247	36489	26	57	1669	5	2012-08-02 11:11:54.052869
36489	4	1-11163	36489	57	27	2101	88	2012-08-02 11:11:54.052869
36517	4	1-9650	36516	53	83	477	10	2012-08-02 11:11:57.577745
36518	4	1-10884	36516	83	24	407	43	2012-08-02 11:11:57.577745
36519	4	1-8272	36516	24	93	224	71	2012-08-02 11:11:57.577745
36516	4	1-11172	36516	93	53	443	54	2012-08-02 11:11:57.577745
36524	4	1-7606	36523	82	14	39	6	2012-08-02 11:11:58.737639
36523	4	1-11175	36523	3	82	70	46	2012-08-02 11:11:58.737639
36225	4	1-10579	36224	99	45	2540	20	2012-08-02 11:11:22.324789
36226	4	1-11003	36224	45	40	1498	50	2012-08-02 11:11:22.324789
36227	4	1-9734	36224	40	57	480	42	2012-08-02 11:11:22.324789
36224	4	1-11083	36224	57	99	6047	40	2012-08-02 11:11:22.324789
36229	4	1-7096	36228	66	93	124	74	2012-08-02 11:11:22.324789
36230	4	1-10768	36228	93	22	144	75	2012-08-02 11:11:22.324789
36231	4	1-9985	36228	22	57	152	42	2012-08-02 11:11:22.324789
36228	4	1-11083	36228	57	66	3136	40	2012-08-02 11:11:22.324789
36243	2	1-8818	36242	70	13	2003	14	2012-08-02 11:11:23.87241
36242	2	1-11086	36242	13	70	3398	34	2012-08-02 11:11:23.87241
36245	2	1-9640	36244	83	13	3065	14	2012-08-02 11:11:23.87241
36244	2	1-11086	36244	13	83	5230	34	2012-08-02 11:11:23.87241
36290	3	1-8365	36289	57	26	208	15	2012-08-02 11:11:29.688061
36291	3	1-10247	36289	26	70	191	5	2012-08-02 11:11:29.688061
36289	3	1-11101	36289	70	57	212	64	2012-08-02 11:11:29.688061
36293	4	1-10631	36292	10	8	3246	67	2012-08-02 11:11:29.688061
36294	4	1-9407	36292	8	56	1339	55	2012-08-02 11:11:29.688061
36295	4	1-8980	36292	56	70	2421	5	2012-08-02 11:11:29.688061
36292	4	1-11101	36292	70	10	2987	64	2012-08-02 11:11:29.688061
36297	4	1-9973	36296	26	13	186	14	2012-08-02 11:11:30.837301
36298	4	1-9948	36296	13	18	122	72	2012-08-02 11:11:30.837301
36299	4	1-10273	36296	18	13	99	65	2012-08-02 11:11:30.837301
36296	4	1-11103	36296	13	26	85	61	2012-08-02 11:11:30.837301
36309	4	1-8892	36308	81	53	442	31	2012-08-02 11:11:32.795243
36310	4	1-10144	36308	53	28	404	23	2012-08-02 11:11:32.795243
36311	4	1-10777	36308	28	1	269	90	2012-08-02 11:11:32.795243
36308	4	1-11109	36308	1	81	1520	21	2012-08-02 11:11:32.795243
36313	2	1-10215	36312	32	1	753	90	2012-08-02 11:11:32.795243
36312	2	1-11109	36312	1	32	4278	21	2012-08-02 11:11:32.795243
36318	2	1-9818	36317	10	41	797	48	2012-08-02 11:11:33.933733
36317	2	1-11112	36317	41	10	1214	97	2012-08-02 11:11:33.933733
36320	3	1-10636	36319	38	27	2577	51	2012-08-02 11:11:33.933733
36321	3	1-9762	36319	27	41	1374	48	2012-08-02 11:11:33.933733
36319	3	1-11112	36319	41	38	2133	97	2012-08-02 11:11:33.933733
36323	4	1-10766	36322	35	10	244	64	2012-08-02 11:11:33.933733
36324	4	1-10631	36322	10	16	256	67	2012-08-02 11:11:33.933733
36325	4	1-10596	36322	16	41	351	48	2012-08-02 11:11:33.933733
36322	4	1-11112	36322	41	35	736	97	2012-08-02 11:11:33.933733
36327	4	1-10582	36326	36	22	1971	94	2012-08-02 11:11:33.933733
36328	4	1-8914	36326	22	89	1624	77	2012-08-02 11:11:33.933733
36329	4	1-6860	36326	89	41	815	48	2012-08-02 11:11:33.933733
36326	4	1-11112	36326	41	36	1876	97	2012-08-02 11:11:33.933733
36331	4	1-10278	36330	61	81	1529	91	2012-08-02 11:11:33.933733
36332	4	1-9171	36330	81	21	1184	95	2012-08-02 11:11:33.933733
36333	4	1-9740	36330	21	41	880	48	2012-08-02 11:11:33.933733
36330	4	1-11112	36330	41	61	2032	97	2012-08-02 11:11:33.933733
36335	3	1-9076	36334	3	21	482	95	2012-08-02 11:11:33.933733
36336	3	1-9740	36334	21	41	360	48	2012-08-02 11:11:33.933733
36334	3	1-11112	36334	41	3	837	97	2012-08-02 11:11:33.933733
36351	4	1-11057	36350	66	49	3322	62	2012-08-02 11:11:37.867504
36352	4	1-10995	36350	49	19	3672	27	2012-08-02 11:11:37.867504
36353	4	1-10405	36350	19	41	2666	81	2012-08-02 11:11:37.867504
36350	4	1-11122	36350	41	66	5487	97	2012-08-02 11:11:37.867504
36355	4	1-10582	36354	36	69	2805	94	2012-08-02 11:11:37.867504
36356	4	1-11029	36354	69	19	1857	27	2012-08-02 11:11:37.867504
36357	4	1-10405	36354	19	41	1380	81	2012-08-02 11:11:37.867504
36354	4	1-11122	36354	41	36	2908	97	2012-08-02 11:11:37.867504
36378	3	1-9577	36377	13	51	5357	60	2012-08-02 11:11:43.016532
36379	3	1-11118	36377	51	96	3631	87	2012-08-02 11:11:43.016532
36377	3	1-11134	36377	96	13	5201	24	2012-08-02 11:11:43.016532
36381	3	1-10861	36380	80	6	1791	54	2012-08-02 11:11:43.016532
36382	3	1-11068	36380	6	96	1332	87	2012-08-02 11:11:43.016532
36380	3	1-11134	36380	96	80	2201	24	2012-08-02 11:11:43.016532
36384	4	1-10611	36383	56	56	5266	50	2012-08-02 11:11:43.819221
36385	4	1-7400	36383	56	44	690	1	2012-08-02 11:11:43.819221
36386	4	1-11018	36383	44	77	832	51	2012-08-02 11:11:43.819221
36383	4	1-11137	36383	77	56	3814	83	2012-08-02 11:11:43.819221
36388	4	1-10630	36387	72	18	2618	72	2012-08-02 11:11:43.819221
36389	4	1-10273	36387	18	22	1512	65	2012-08-02 11:11:43.819221
36390	4	1-6493	36387	22	77	593	51	2012-08-02 11:11:43.819221
36387	4	1-11137	36387	77	72	3556	83	2012-08-02 11:11:43.819221
36392	4	1-10749	36391	49	16	2045	7	2012-08-02 11:11:44.979358
36393	4	1-10685	36391	16	22	1024	91	2012-08-02 11:11:44.979358
36394	4	1-11035	36391	22	86	1413	21	2012-08-02 11:11:44.979358
36391	4	1-11140	36391	86	49	2534	81	2012-08-02 11:11:44.979358
36434	4	1-8788	36433	77	10	1222	79	2012-08-02 11:11:48.526807
36435	4	1-9695	36433	10	32	1394	85	2012-08-02 11:11:48.526807
36436	4	1-10620	36433	32	5	943	65	2012-08-02 11:11:48.526807
36433	4	1-11149	36433	5	77	1342	97	2012-08-02 11:11:48.526807
36449	2	1-10761	36448	64	76	2753	42	2012-08-02 11:11:49.720618
36448	2	1-11152	36448	76	64	7925	66	2012-08-02 11:11:49.720618
36458	4	1-10725	36457	94	72	1345	1	2012-08-02 11:11:50.892453
36459	4	1-8885	36457	72	29	1278	42	2012-08-02 11:11:50.892453
36460	4	1-10800	36457	29	50	1211	86	2012-08-02 11:11:50.892453
36457	4	1-11155	36457	50	94	1606	69	2012-08-02 11:11:50.892453
36462	4	1-10725	36461	94	45	4776	1	2012-08-02 11:11:50.892453
36463	4	1-10738	36461	45	73	4385	14	2012-08-02 11:11:50.892453
36464	4	1-11097	36461	73	50	3325	86	2012-08-02 11:11:50.892453
36461	4	1-11155	36461	50	94	5014	69	2012-08-02 11:11:50.892453
36466	4	1-8831	36465	19	21	1734	70	2012-08-02 11:11:50.892453
36467	4	1-10919	36465	21	87	1683	30	2012-08-02 11:11:50.892453
36468	4	1-9355	36465	87	50	1466	86	2012-08-02 11:11:50.892453
36465	4	1-11155	36465	50	19	2222	69	2012-08-02 11:11:50.892453
36470	4	1-10770	36469	44	93	28	100	2012-08-02 11:11:50.892453
36471	4	1-10159	36469	93	26	28	52	2012-08-02 11:11:50.892453
36472	4	1-5880	36469	26	50	11	86	2012-08-02 11:11:50.892453
36469	4	1-11155	36469	50	44	16	69	2012-08-02 11:11:50.892453
36521	3	1-10132	36520	79	93	5289	90	2012-08-02 11:11:58.362033
36522	3	1-10463	36520	93	40	3332	28	2012-08-02 11:11:58.362033
36520	3	1-11174	36520	40	79	6942	3	2012-08-02 11:11:58.362033
36537	3	1-7157	36536	46	26	486	24	2012-08-02 11:11:59.522129
36538	3	1-8104	36536	26	56	250	56	2012-08-02 11:11:59.522129
36536	3	1-11177	36536	56	46	2245	9	2012-08-02 11:11:59.522129
36233	3	1-10104	36232	25	26	789	61	2012-08-02 11:11:23.485277
36234	3	1-9973	36232	26	84	800	14	2012-08-02 11:11:23.485277
36232	3	1-11085	36232	84	25	2578	77	2012-08-02 11:11:23.485277
36236	3	1-10312	36235	93	42	769	81	2012-08-02 11:11:23.485277
36237	3	1-9976	36235	42	84	421	14	2012-08-02 11:11:23.485277
36235	3	1-11085	36235	84	93	1731	77	2012-08-02 11:11:23.485277
36239	4	1-7526	36238	23	32	563	85	2012-08-02 11:11:23.485277
36240	4	1-10620	36238	32	85	257	65	2012-08-02 11:11:23.485277
36241	4	1-10387	36238	85	84	189	14	2012-08-02 11:11:23.485277
36238	4	1-11085	36238	84	23	903	77	2012-08-02 11:11:23.485277
36247	4	1-11059	36246	39	38	532	41	2012-08-02 11:11:24.613506
36248	4	1-10815	36246	38	45	1546	44	2012-08-02 11:11:24.613506
36249	4	1-10294	36246	45	10	566	65	2012-08-02 11:11:24.613506
36246	4	1-11088	36246	10	39	655	17	2012-08-02 11:11:24.613506
36259	4	1-9076	36258	3	43	1387	95	2012-08-02 11:11:25.785107
36260	4	1-10869	36258	43	97	2035	4	2012-08-02 11:11:25.785107
36261	4	1-10096	36258	97	6	1627	41	2012-08-02 11:11:25.785107
36258	4	1-11091	36258	6	3	2663	97	2012-08-02 11:11:25.785107
36263	4	1-10062	36262	87	95	709	38	2012-08-02 11:11:25.785107
36264	4	1-9941	36262	95	39	480	17	2012-08-02 11:11:25.785107
36265	4	1-11059	36262	39	6	366	41	2012-08-02 11:11:25.785107
36262	4	1-11091	36262	6	87	608	97	2012-08-02 11:11:25.785107
36267	4	1-9058	36266	45	49	576	4	2012-08-02 11:11:25.785107
36268	4	1-8866	36266	49	61	475	71	2012-08-02 11:11:25.785107
36269	4	1-11005	36266	61	6	443	41	2012-08-02 11:11:25.785107
36266	4	1-11091	36266	6	45	820	97	2012-08-02 11:11:25.785107
36271	4	1-10582	36270	36	90	1397	94	2012-08-02 11:11:25.785107
36272	4	1-8836	36270	90	39	777	17	2012-08-02 11:11:25.785107
36273	4	1-11059	36270	39	6	671	41	2012-08-02 11:11:25.785107
36270	4	1-11091	36270	6	36	1265	97	2012-08-02 11:11:25.785107
36315	3	1-10646	36314	22	77	6083	45	2012-08-02 11:11:33.170861
36316	3	1-10711	36314	77	89	748	63	2012-08-02 11:11:33.170861
36314	3	1-11110	36314	89	22	7328	81	2012-08-02 11:11:33.170861
36338	4	1-11054	36337	97	60	535	96	2012-08-02 11:11:35.525612
36339	4	1-7563	36337	60	66	441	68	2012-08-02 11:11:35.525612
36340	4	1-9647	36337	66	43	353	30	2012-08-02 11:11:35.525612
36337	4	1-11115	36337	43	97	531	16	2012-08-02 11:11:35.525612
36359	3	1-10498	36358	32	59	1121	22	2012-08-02 11:11:40.607873
36360	3	1-4910	36358	59	96	414	53	2012-08-02 11:11:40.607873
36358	3	1-11128	36358	96	32	791	71	2012-08-02 11:11:40.607873
36362	4	1-10068	36361	50	56	1857	59	2012-08-02 11:11:40.607873
36363	4	1-10931	36361	56	56	3251	31	2012-08-02 11:11:40.607873
36364	4	1-9699	36361	56	96	1633	53	2012-08-02 11:11:40.607873
36361	4	1-11128	36361	96	50	3446	71	2012-08-02 11:11:40.607873
36366	4	1-10498	36365	32	97	4086	22	2012-08-02 11:11:40.607873
36367	4	1-10952	36365	97	90	3537	3	2012-08-02 11:11:40.607873
36368	4	1-5497	36365	90	96	1214	53	2012-08-02 11:11:40.607873
36365	4	1-11128	36365	96	32	2586	71	2012-08-02 11:11:40.607873
36396	4	1-11047	36395	88	52	2963	34	2012-08-02 11:11:45.377269
36397	4	1-8957	36395	52	47	954	23	2012-08-02 11:11:45.377269
36398	4	1-11138	36395	47	20	1384	98	2012-08-02 11:11:45.377269
36395	4	1-11141	36395	20	88	5506	27	2012-08-02 11:11:45.377269
36399	1	1-6333	\N	3	3	221	23	2012-08-02 11:11:45.377269
36401	4	1-10899	36400	57	57	526	94	2012-08-02 11:11:46.570756
36402	4	1-9785	36400	57	90	1021	83	2012-08-02 11:11:46.570756
36403	4	1-8377	36400	90	5	460	7	2012-08-02 11:11:46.570756
36400	4	1-11144	36400	5	57	374	28	2012-08-02 11:11:46.570756
36405	3	1-10720	36404	59	58	3611	71	2012-08-02 11:11:46.570756
36406	3	1-10127	36404	58	5	3766	7	2012-08-02 11:11:46.570756
36404	3	1-11144	36404	5	59	3177	28	2012-08-02 11:11:46.570756
36419	4	1-10340	36418	74	16	628	32	2012-08-02 11:11:47.753155
36420	4	1-10472	36418	16	19	654	27	2012-08-02 11:11:47.753155
36421	4	1-10405	36418	19	60	452	81	2012-08-02 11:11:47.753155
36418	4	1-11147	36418	60	74	642	6	2012-08-02 11:11:47.753155
36423	3	1-10834	36422	34	59	7766	90	2012-08-02 11:11:47.753155
36424	3	1-10075	36422	59	60	4557	81	2012-08-02 11:11:47.753155
36422	3	1-11147	36422	60	34	7491	6	2012-08-02 11:11:47.753155
36426	4	1-10834	36425	34	60	766	90	2012-08-02 11:11:47.753155
36427	4	1-10244	36425	60	66	438	43	2012-08-02 11:11:47.753155
36428	4	1-9570	36425	66	60	409	81	2012-08-02 11:11:47.753155
36425	4	1-11147	36425	60	34	709	6	2012-08-02 11:11:47.753155
36474	4	1-10606	36473	26	79	4175	79	2012-08-02 11:11:51.267894
36475	4	1-10954	36473	79	42	3014	35	2012-08-02 11:11:51.267894
36476	4	1-6590	36473	42	90	2806	29	2012-08-02 11:11:51.267894
36473	4	1-11156	36473	90	26	4705	94	2012-08-02 11:11:51.267894
36478	2	1-9145	36477	28	66	2776	35	2012-08-02 11:11:53.66609
36477	2	1-11162	36477	66	28	3833	26	2012-08-02 11:11:53.66609
36480	4	1-10653	36479	62	28	2516	2	2012-08-02 11:11:53.66609
36481	4	1-10695	36479	28	79	1470	79	2012-08-02 11:11:53.66609
36482	4	1-10954	36479	79	66	964	35	2012-08-02 11:11:53.66609
36479	4	1-11162	36479	66	62	1482	26	2012-08-02 11:11:53.66609
36484	3	1-9377	36483	96	41	1270	16	2012-08-02 11:11:53.66609
36485	3	1-9997	36483	41	66	684	35	2012-08-02 11:11:53.66609
36483	3	1-11162	36483	66	96	1092	26	2012-08-02 11:11:53.66609
36494	4	1-11149	36493	5	77	329	97	2012-08-02 11:11:54.815166
36495	4	1-8788	36493	77	10	284	79	2012-08-02 11:11:54.815166
36496	4	1-9695	36493	10	100	308	85	2012-08-02 11:11:54.815166
36493	4	1-11164	36493	100	5	244	65	2012-08-02 11:11:54.815166
36498	3	1-6437	36497	7	84	628	54	2012-08-02 11:11:55.599876
36499	3	1-7690	36497	84	76	213	79	2012-08-02 11:11:55.599876
36497	3	1-11167	36497	76	7	826	22	2012-08-02 11:11:55.599876
36501	4	1-10952	36500	97	88	98	3	2012-08-02 11:11:55.599876
36502	4	1-10109	36500	88	26	50	94	2012-08-02 11:11:55.599876
36503	4	1-10606	36500	26	76	34	79	2012-08-02 11:11:55.599876
36500	4	1-11167	36500	76	97	133	22	2012-08-02 11:11:55.599876
36505	4	1-6886	36504	77	63	2078	72	2012-08-02 11:11:55.599876
36506	4	1-9856	36504	63	84	2709	54	2012-08-02 11:11:55.599876
36507	4	1-7690	36504	84	76	1018	79	2012-08-02 11:11:55.599876
36504	4	1-11167	36504	76	77	4365	22	2012-08-02 11:11:55.599876
36509	4	1-7612	36508	65	73	823	5	2012-08-02 11:11:55.599876
36510	4	1-10949	36508	73	26	732	94	2012-08-02 11:11:55.599876
36511	4	1-10606	36508	26	76	582	79	2012-08-02 11:11:55.599876
36508	4	1-11167	36508	76	65	2631	22	2012-08-02 11:11:55.599876
36513	4	1-10828	36512	86	66	1282	76	2012-08-02 11:11:55.599876
36514	4	1-10945	36512	66	77	511	97	2012-08-02 11:11:55.599876
36515	4	1-8788	36512	77	76	416	79	2012-08-02 11:11:55.599876
36512	4	1-11167	36512	76	86	1942	22	2012-08-02 11:11:55.599876
36559	4	1-7023	36558	34	37	660	8	2012-08-02 11:12:01.509541
36560	4	1-11004	36558	37	86	361	75	2012-08-02 11:12:01.509541
36561	4	1-9946	36558	86	95	362	38	2012-08-02 11:12:01.509541
36558	4	1-11182	36558	95	34	5324	44	2012-08-02 11:12:01.509541
36563	4	1-10309	36562	61	73	1056	5	2012-08-02 11:12:01.509541
36564	4	1-10949	36562	73	18	712	94	2012-08-02 11:12:01.509541
36565	4	1-9241	36562	18	95	178	38	2012-08-02 11:12:01.509541
36562	4	1-11182	36562	95	61	2967	44	2012-08-02 11:12:01.509541
36607	4	1-10822	36606	100	52	372	8	2012-08-02 11:12:07.380029
36608	4	1-11061	36606	52	11	616	43	2012-08-02 11:12:07.380029
36609	4	1-8137	36606	11	87	351	49	2012-08-02 11:12:07.380029
36606	4	1-11199	36606	87	100	544	47	2012-08-02 11:12:07.380029
36621	4	1-9172	36620	39	22	3688	91	2012-08-02 11:12:15.157959
36622	4	1-11035	36620	22	92	4714	21	2012-08-02 11:12:15.157959
36623	4	1-11148	36620	92	32	2840	75	2012-08-02 11:12:15.157959
36620	4	1-11217	36620	32	39	5387	36	2012-08-02 11:12:15.157959
36625	4	1-10538	36624	4	42	3423	84	2012-08-02 11:12:16.340182
36626	4	1-10716	36624	42	45	3213	29	2012-08-02 11:12:16.340182
36627	4	1-10967	36624	45	13	3723	57	2012-08-02 11:12:16.340182
36624	4	1-11220	36624	13	4	7285	72	2012-08-02 11:12:16.340182
36629	4	1-9946	36628	86	97	2102	38	2012-08-02 11:12:17.632767
36630	4	1-10880	36628	97	42	1721	59	2012-08-02 11:12:17.632767
36631	4	1-9068	36628	42	81	2117	85	2012-08-02 11:12:17.632767
36628	4	1-11223	36628	81	86	1360	75	2012-08-02 11:12:17.632767
36633	4	1-10106	36632	42	45	540	20	2012-08-02 11:12:17.632767
36634	4	1-11003	36632	45	22	785	50	2012-08-02 11:12:17.632767
36635	4	1-7623	36632	22	81	678	85	2012-08-02 11:12:17.632767
36632	4	1-11223	36632	81	42	479	75	2012-08-02 11:12:17.632767
36637	4	1-9031	36636	14	81	243	85	2012-08-02 11:12:18.781677
36638	4	1-11223	36636	81	42	116	75	2012-08-02 11:12:18.781677
36639	4	1-10106	36636	42	20	89	20	2012-08-02 11:12:18.781677
36636	4	1-11225	36636	20	14	377	6	2012-08-02 11:12:18.781677
36641	4	1-10803	36640	35	27	2364	51	2012-08-02 11:12:18.781677
36642	4	1-10850	36640	27	42	1604	75	2012-08-02 11:12:18.781677
36643	4	1-10106	36640	42	20	1388	20	2012-08-02 11:12:18.781677
36640	4	1-11225	36640	20	35	6672	6	2012-08-02 11:12:18.781677
36645	3	1-6678	36644	36	42	367	75	2012-08-02 11:12:18.781677
36646	3	1-10106	36644	42	20	334	20	2012-08-02 11:12:18.781677
36644	3	1-11225	36644	20	36	1686	6	2012-08-02 11:12:18.781677
36655	4	1-10514	36654	98	11	1564	43	2012-08-02 11:12:19.577338
36656	4	1-8137	36654	11	10	733	49	2012-08-02 11:12:19.577338
36657	4	1-11010	36654	10	18	1529	73	2012-08-02 11:12:19.577338
36654	4	1-11228	36654	18	98	1808	26	2012-08-02 11:12:19.577338
36659	4	1-10514	36658	98	40	2410	43	2012-08-02 11:12:19.577338
36660	4	1-10300	36658	40	56	1667	64	2012-08-02 11:12:19.577338
36661	4	1-8657	36658	56	18	1642	73	2012-08-02 11:12:19.577338
36658	4	1-11228	36658	18	98	2325	26	2012-08-02 11:12:19.577338
36662	1	1-8657	\N	56	56	3624	73	2012-08-02 11:12:19.577338
36682	4	1-9325	36681	70	4	1654	72	2012-08-02 11:12:21.908586
36683	4	1-10538	36681	4	64	690	84	2012-08-02 11:12:21.908586
36684	4	1-11170	36681	64	65	1568	42	2012-08-02 11:12:21.908586
36681	4	1-11234	36681	65	70	3952	54	2012-08-02 11:12:21.908586
36686	3	1-10946	36685	89	33	1629	4	2012-08-02 11:12:21.908586
36687	3	1-9843	36685	33	65	1448	42	2012-08-02 11:12:21.908586
36685	3	1-11234	36685	65	89	3823	54	2012-08-02 11:12:21.908586
36744	4	1-10668	36743	57	89	1550	59	2012-08-02 11:12:29.379563
36745	4	1-11209	36743	89	84	998	30	2012-08-02 11:12:29.379563
36746	4	1-6733	36743	84	56	1035	88	2012-08-02 11:12:29.379563
36743	4	1-11253	36743	56	57	2302	81	2012-08-02 11:12:29.379563
36748	4	1-10646	36747	22	89	796	45	2012-08-02 11:12:29.379563
36749	4	1-7025	36747	89	14	758	85	2012-08-02 11:12:29.379563
36750	4	1-8197	36747	14	56	413	88	2012-08-02 11:12:29.379563
36747	4	1-11253	36747	56	22	1000	81	2012-08-02 11:12:29.379563
36752	3	1-10764	36751	83	4	2808	65	2012-08-02 11:12:29.379563
36753	3	1-5601	36751	4	56	1677	88	2012-08-02 11:12:29.379563
36751	3	1-11253	36751	56	83	4461	81	2012-08-02 11:12:29.379563
36803	3	1-10717	36802	45	95	1728	44	2012-08-02 11:12:35.247051
36804	3	1-6648	36802	95	90	218	79	2012-08-02 11:12:35.247051
36802	3	1-11268	36802	90	45	2281	2	2012-08-02 11:12:35.247051
36806	3	1-11161	36805	60	17	158	81	2012-08-02 11:12:35.247051
36807	3	1-8136	36805	17	90	57	79	2012-08-02 11:12:35.247051
36805	3	1-11268	36805	90	60	646	2	2012-08-02 11:12:35.247051
36836	2	1-10105	36835	8	78	6851	25	2012-08-02 11:12:37.56716
36835	2	1-11274	36835	78	8	7407	29	2012-08-02 11:12:37.56716
36854	3	1-11199	36853	87	34	3274	47	2012-08-02 11:12:38.727431
36855	3	1-5628	36853	34	60	937	37	2012-08-02 11:12:38.727431
36853	3	1-11277	36853	60	87	5437	49	2012-08-02 11:12:38.727431
36871	3	1-10770	36870	44	94	3284	100	2012-08-02 11:12:43.81278
36872	3	1-10806	36870	94	72	3446	21	2012-08-02 11:12:43.81278
36870	3	1-11290	36870	72	44	2404	69	2012-08-02 11:12:43.81278
36874	4	1-10954	36873	79	56	627	35	2012-08-02 11:12:44.972704
36875	4	1-10345	36873	56	52	461	66	2012-08-02 11:12:44.972704
36876	4	1-10320	36873	52	74	448	88	2012-08-02 11:12:44.972704
36873	4	1-11293	36873	74	79	1118	79	2012-08-02 11:12:44.972704
36878	4	1-10007	36877	24	74	1338	96	2012-08-02 11:12:44.972704
36879	4	1-8851	36877	74	57	1675	5	2012-08-02 11:12:44.972704
36880	4	1-11163	36877	57	74	1828	88	2012-08-02 11:12:44.972704
36877	4	1-11293	36877	74	24	5526	79	2012-08-02 11:12:44.972704
36882	4	1-9027	36881	25	56	329	35	2012-08-02 11:12:44.972704
36883	4	1-10345	36881	56	52	318	66	2012-08-02 11:12:44.972704
36884	4	1-10320	36881	52	74	408	88	2012-08-02 11:12:44.972704
36881	4	1-11293	36881	74	25	1339	79	2012-08-02 11:12:44.972704
36890	4	1-10303	36889	81	78	1176	96	2012-08-02 11:12:47.725661
36891	4	1-11206	36889	78	67	877	99	2012-08-02 11:12:47.725661
36892	4	1-7286	36889	67	47	866	68	2012-08-02 11:12:47.725661
36889	4	1-11300	36889	47	81	927	78	2012-08-02 11:12:47.725661
36913	4	1-10630	36912	72	79	4039	72	2012-08-02 11:12:52.388119
36914	4	1-6137	36912	79	59	766	84	2012-08-02 11:12:52.388119
36915	4	1-11255	36912	59	43	810	22	2012-08-02 11:12:52.388119
36912	4	1-11312	36912	43	72	4596	83	2012-08-02 11:12:52.388119
36525	4	1-9031	36523	14	85	31	85	2012-08-02 11:11:58.737639
36526	4	1-10917	36523	85	3	24	13	2012-08-02 11:11:58.737639
36528	4	1-10125	36527	17	32	70	61	2012-08-02 11:11:58.737639
36529	4	1-10157	36527	32	19	90	58	2012-08-02 11:11:58.737639
36530	4	1-10018	36527	19	3	77	13	2012-08-02 11:11:58.737639
36527	4	1-11175	36527	3	17	231	46	2012-08-02 11:11:58.737639
36532	4	1-7518	36531	2	49	410	62	2012-08-02 11:11:58.737639
36533	4	1-10995	36531	49	25	482	27	2012-08-02 11:11:58.737639
36534	4	1-11159	36531	25	3	407	13	2012-08-02 11:11:58.737639
36531	4	1-11175	36531	3	2	1321	46	2012-08-02 11:11:58.737639
36535	1	1-8116	\N	88	88	344	13	2012-08-02 11:11:58.737639
36551	4	1-11082	36550	5	93	1980	7	2012-08-02 11:11:59.897958
36552	4	1-10978	36550	93	96	828	70	2012-08-02 11:11:59.897958
36553	4	1-10616	36550	96	61	133	62	2012-08-02 11:11:59.897958
36550	4	1-11178	36550	61	5	5947	13	2012-08-02 11:11:59.897958
36555	4	1-9512	36554	93	31	2837	100	2012-08-02 11:12:01.093497
36556	4	1-11135	36554	31	72	2332	83	2012-08-02 11:12:01.093497
36557	4	1-10630	36554	72	10	2432	72	2012-08-02 11:12:01.093497
36554	4	1-11181	36554	10	93	1451	56	2012-08-02 11:12:01.093497
36571	3	1-11006	36570	20	94	3229	87	2012-08-02 11:12:02.297576
36572	3	1-10970	36570	94	84	1895	77	2012-08-02 11:12:02.297576
36570	3	1-11184	36570	84	20	5662	19	2012-08-02 11:12:02.297576
36581	3	1-11039	36580	73	5	1293	93	2012-08-02 11:12:06.970827
36582	3	1-11052	36580	5	63	2483	27	2012-08-02 11:12:06.970827
36580	3	1-11195	36580	63	73	3918	99	2012-08-02 11:12:06.970827
36584	4	1-4709	36583	52	57	1722	28	2012-08-02 11:12:06.970827
36585	4	1-10899	36583	57	69	2053	94	2012-08-02 11:12:06.970827
36586	4	1-11029	36583	69	63	1235	27	2012-08-02 11:12:06.970827
36583	4	1-11195	36583	63	52	2244	99	2012-08-02 11:12:06.970827
36588	4	1-8305	36587	48	73	1630	5	2012-08-02 11:12:06.970827
36589	4	1-10949	36587	73	69	1369	94	2012-08-02 11:12:06.970827
36590	4	1-11029	36587	69	63	846	27	2012-08-02 11:12:06.970827
36587	4	1-11195	36587	63	48	1580	99	2012-08-02 11:12:06.970827
36611	3	1-9901	36610	86	31	3211	60	2012-08-02 11:12:13.224419
36612	3	1-5502	36610	31	35	120	82	2012-08-02 11:12:13.224419
36610	3	1-11212	36610	35	86	7744	53	2012-08-02 11:12:13.224419
36648	4	1-10996	36647	54	61	2030	71	2012-08-02 11:12:19.168347
36649	4	1-11005	36647	61	41	1852	41	2012-08-02 11:12:19.168347
36650	4	1-11078	36647	41	21	1770	43	2012-08-02 11:12:19.168347
36647	4	1-11227	36647	21	54	1095	18	2012-08-02 11:12:19.168347
36652	3	1-10285	36651	43	52	2131	8	2012-08-02 11:12:19.168347
36653	3	1-11061	36651	52	21	3727	43	2012-08-02 11:12:19.168347
36651	3	1-11227	36651	21	43	2324	18	2012-08-02 11:12:19.168347
36668	3	1-11077	36667	16	91	6140	82	2012-08-02 11:12:21.529328
36669	3	1-10290	36667	91	23	1750	32	2012-08-02 11:12:21.529328
36667	3	1-11233	36667	23	16	2211	55	2012-08-02 11:12:21.529328
36671	3	1-9472	36670	78	78	328	59	2012-08-02 11:12:21.529328
36672	3	1-10953	36670	78	23	281	32	2012-08-02 11:12:21.529328
36670	3	1-11233	36670	23	78	421	55	2012-08-02 11:12:21.529328
36674	4	1-8343	36673	63	97	820	38	2012-08-02 11:12:21.529328
36675	4	1-10880	36673	97	78	630	59	2012-08-02 11:12:21.529328
36676	4	1-10953	36673	78	23	535	32	2012-08-02 11:12:21.529328
36673	4	1-11233	36673	23	63	943	55	2012-08-02 11:12:21.529328
36678	4	1-9225	36677	4	35	1131	80	2012-08-02 11:12:21.529328
36679	4	1-10773	36677	35	13	458	84	2012-08-02 11:12:21.529328
36680	4	1-11211	36677	13	23	583	32	2012-08-02 11:12:21.529328
36677	4	1-11233	36677	23	4	1161	55	2012-08-02 11:12:21.529328
36697	4	1-7623	36696	22	81	867	85	2012-08-02 11:12:22.704417
36698	4	1-11223	36696	81	42	431	75	2012-08-02 11:12:22.704417
36699	4	1-10106	36696	42	63	343	20	2012-08-02 11:12:22.704417
36696	4	1-11236	36696	63	22	1426	50	2012-08-02 11:12:22.704417
36701	4	1-10180	36700	29	50	499	9	2012-08-02 11:12:22.704417
36702	4	1-7534	36700	50	2	276	34	2012-08-02 11:12:22.704417
36703	4	1-11106	36700	2	63	171	20	2012-08-02 11:12:22.704417
36700	4	1-11236	36700	63	29	778	50	2012-08-02 11:12:22.704417
36705	4	1-10829	36704	73	78	1697	11	2012-08-02 11:12:22.704417
36706	4	1-7907	36704	78	42	1160	75	2012-08-02 11:12:22.704417
36707	4	1-10106	36704	42	63	1124	20	2012-08-02 11:12:22.704417
36704	4	1-11236	36704	63	73	5700	50	2012-08-02 11:12:22.704417
36729	4	1-10138	36728	17	51	360	60	2012-08-02 11:12:25.058232
36730	4	1-11118	36728	51	94	454	87	2012-08-02 11:12:25.058232
36731	4	1-10970	36728	94	45	414	77	2012-08-02 11:12:25.058232
36728	4	1-11242	36728	45	17	334	83	2012-08-02 11:12:25.058232
36740	4	1-10815	36739	38	79	2276	44	2012-08-02 11:12:28.595194
36741	4	1-11210	36739	79	19	780	9	2012-08-02 11:12:28.595194
36742	4	1-10238	36739	19	35	1492	42	2012-08-02 11:12:28.595194
36739	4	1-11251	36739	35	38	778	41	2012-08-02 11:12:28.595194
36755	2	1-9508	36754	7	21	721	98	2012-08-02 11:12:30.893329
36754	2	1-11256	36754	21	7	514	26	2012-08-02 11:12:30.893329
36757	4	1-8453	36756	79	85	2244	100	2012-08-02 11:12:30.893329
36758	4	1-9755	36756	85	47	925	23	2012-08-02 11:12:30.893329
36759	4	1-11138	36756	47	21	1973	98	2012-08-02 11:12:30.893329
36756	4	1-11256	36756	21	79	1622	26	2012-08-02 11:12:30.893329
36765	4	1-10998	36764	23	96	4581	11	2012-08-02 11:12:31.699949
36766	4	1-10849	36764	96	54	2131	56	2012-08-02 11:12:31.699949
36767	4	1-11213	36764	54	76	717	24	2012-08-02 11:12:31.699949
36764	4	1-11259	36764	76	23	5131	4	2012-08-02 11:12:31.699949
36769	4	1-10162	36768	42	84	3327	5	2012-08-02 11:12:31.699949
36770	4	1-10753	36768	84	20	369	63	2012-08-02 11:12:31.699949
36771	4	1-10375	36768	20	76	238	24	2012-08-02 11:12:31.699949
36768	4	1-11259	36768	76	42	2206	4	2012-08-02 11:12:31.699949
36773	4	1-10747	36772	6	5	2106	100	2012-08-02 11:12:31.699949
36774	4	1-10742	36772	5	31	1002	47	2012-08-02 11:12:31.699949
36775	4	1-6772	36772	31	76	224	24	2012-08-02 11:12:31.699949
36772	4	1-11259	36772	76	6	2161	4	2012-08-02 11:12:31.699949
36777	2	1-10860	36776	57	97	582	3	2012-08-02 11:12:32.882289
36776	2	1-11263	36776	97	57	1279	45	2012-08-02 11:12:32.882289
36779	4	1-10535	36778	4	96	811	70	2012-08-02 11:12:32.882289
36780	4	1-11183	36778	96	1	500	9	2012-08-02 11:12:32.882289
36781	4	1-5611	36778	1	97	187	3	2012-08-02 11:12:32.882289
36778	4	1-11263	36778	97	4	700	45	2012-08-02 11:12:32.882289
36783	5	1-7025	36782	89	81	2951	85	2012-08-02 11:12:32.882289
36784	5	1-11223	36782	81	3	1699	75	2012-08-02 11:12:32.882289
36540	3	1-6036	36539	71	59	413	79	2012-08-02 11:11:59.522129
36541	3	1-6017	36539	59	56	81	56	2012-08-02 11:11:59.522129
36539	3	1-11177	36539	56	71	839	9	2012-08-02 11:11:59.522129
36543	4	1-8476	36542	36	32	1327	71	2012-08-02 11:11:59.522129
36544	4	1-10498	36542	32	76	1783	22	2012-08-02 11:11:59.522129
36545	4	1-2964	36542	76	56	189	56	2012-08-02 11:11:59.522129
36542	4	1-11177	36542	56	36	1962	9	2012-08-02 11:11:59.522129
36547	4	1-7534	36546	50	2	1434	34	2012-08-02 11:11:59.522129
36548	4	1-11106	36546	2	52	899	20	2012-08-02 11:11:59.522129
36549	4	1-10746	36546	52	56	240	56	2012-08-02 11:11:59.522129
36546	4	1-11177	36546	56	50	2563	9	2012-08-02 11:11:59.522129
36567	4	1-8476	36566	36	58	1728	71	2012-08-02 11:12:01.911108
36568	4	1-10127	36566	58	93	1934	7	2012-08-02 11:12:01.911108
36569	4	1-10978	36566	93	96	2484	70	2012-08-02 11:12:01.911108
36566	4	1-11183	36566	96	36	1953	9	2012-08-02 11:12:01.911108
36574	3	1-7553	36573	52	6	1723	25	2012-08-02 11:12:04.253594
36575	3	1-8445	36573	6	10	967	70	2012-08-02 11:12:04.253594
36573	3	1-11189	36573	10	52	8794	66	2012-08-02 11:12:04.253594
36577	4	1-10755	36576	56	2	470	68	2012-08-02 11:12:05.424338
36578	4	1-9329	36576	2	10	280	49	2012-08-02 11:12:05.424338
36579	4	1-11010	36576	10	89	708	73	2012-08-02 11:12:05.424338
36576	4	1-11192	36576	89	56	895	34	2012-08-02 11:12:05.424338
36592	3	1-7544	36591	82	16	436	55	2012-08-02 11:12:06.595309
36593	3	1-11077	36591	16	95	1704	82	2012-08-02 11:12:06.595309
36591	3	1-11196	36591	95	82	2222	7	2012-08-02 11:12:06.595309
36595	4	1-10978	36594	93	96	4978	70	2012-08-02 11:12:06.595309
36596	4	1-10616	36594	96	38	2309	62	2012-08-02 11:12:06.595309
36597	4	1-11146	36594	38	95	2901	82	2012-08-02 11:12:06.595309
36594	4	1-11196	36594	95	93	4112	7	2012-08-02 11:12:06.595309
36599	4	1-10685	36598	16	81	241	91	2012-08-02 11:12:06.595309
36600	4	1-9171	36598	81	63	193	95	2012-08-02 11:12:06.595309
36601	4	1-10904	36598	63	95	274	82	2012-08-02 11:12:06.595309
36598	4	1-11196	36598	95	16	416	7	2012-08-02 11:12:06.595309
36603	4	1-10850	36602	27	59	2375	75	2012-08-02 11:12:08.164732
36604	4	1-8288	36602	59	85	3228	85	2012-08-02 11:12:08.164732
36605	4	1-10917	36602	85	30	3023	13	2012-08-02 11:12:08.164732
36602	4	1-11198	36602	30	27	2640	51	2012-08-02 11:12:08.164732
36614	3	1-9904	36613	40	97	1762	16	2012-08-02 11:12:14.771122
36615	3	1-11054	36613	97	20	1299	96	2012-08-02 11:12:14.771122
36613	3	1-11216	36613	20	40	2746	9	2012-08-02 11:12:14.771122
36617	4	1-7534	36616	50	27	54	34	2012-08-02 11:12:14.771122
36618	4	1-8971	36616	27	97	45	16	2012-08-02 11:12:14.771122
36619	4	1-11054	36616	97	20	38	96	2012-08-02 11:12:14.771122
36616	4	1-11216	36616	20	50	94	9	2012-08-02 11:12:14.771122
36664	4	1-11146	36663	38	24	2222	82	2012-08-02 11:12:21.124269
36665	4	1-3656	36663	24	43	912	10	2012-08-02 11:12:21.124269
36666	4	1-10499	36663	43	63	599	38	2012-08-02 11:12:21.124269
36663	4	1-11232	36663	63	38	2066	62	2012-08-02 11:12:21.124269
36689	4	1-9865	36688	11	89	1854	29	2012-08-02 11:12:22.317417
36690	4	1-10488	36688	89	16	1045	67	2012-08-02 11:12:22.317417
36691	4	1-10596	36688	16	62	1208	48	2012-08-02 11:12:22.317417
36688	4	1-11235	36688	62	11	4485	9	2012-08-02 11:12:22.317417
36693	4	1-8476	36692	36	14	3539	71	2012-08-02 11:12:22.317417
36694	4	1-9930	36692	14	100	4465	44	2012-08-02 11:12:22.317417
36695	4	1-11030	36692	100	62	1345	48	2012-08-02 11:12:22.317417
36692	4	1-11235	36692	62	36	5159	9	2012-08-02 11:12:22.317417
36709	4	1-9386	36708	92	81	1380	85	2012-08-02 11:12:23.48894
36710	4	1-11223	36708	81	62	657	75	2012-08-02 11:12:23.48894
36711	4	1-5712	36708	62	39	368	84	2012-08-02 11:12:23.48894
36708	4	1-11238	36708	39	92	1293	68	2012-08-02 11:12:23.48894
36713	4	1-9386	36712	92	81	4403	85	2012-08-02 11:12:23.48894
36714	4	1-11223	36712	81	66	2106	75	2012-08-02 11:12:23.48894
36715	4	1-6882	36712	66	39	1163	84	2012-08-02 11:12:23.48894
36712	4	1-11238	36712	39	92	4104	68	2012-08-02 11:12:23.48894
36717	4	1-9192	36716	61	2	1803	95	2012-08-02 11:12:23.48894
36718	4	1-10493	36716	2	20	2264	22	2012-08-02 11:12:23.48894
36719	4	1-9950	36716	20	39	1115	84	2012-08-02 11:12:23.48894
36716	4	1-11238	36716	39	61	4361	68	2012-08-02 11:12:23.48894
36721	4	1-8421	36720	24	74	3059	96	2012-08-02 11:12:24.693493
36722	4	1-8851	36720	74	57	2290	5	2012-08-02 11:12:24.693493
36723	4	1-11163	36720	57	60	1652	88	2012-08-02 11:12:24.693493
36720	4	1-11241	36720	60	24	5823	3	2012-08-02 11:12:24.693493
36725	4	1-10132	36724	79	89	1971	90	2012-08-02 11:12:24.693493
36726	4	1-10992	36724	89	84	1096	30	2012-08-02 11:12:24.693493
36727	4	1-6733	36724	84	60	795	88	2012-08-02 11:12:24.693493
36724	4	1-11241	36724	60	79	2893	3	2012-08-02 11:12:24.693493
36733	3	1-10907	36732	74	97	2903	30	2012-08-02 11:12:25.864716
36734	3	1-7426	36732	97	39	2583	100	2012-08-02 11:12:25.864716
36732	3	1-11244	36732	39	74	2603	39	2012-08-02 11:12:25.864716
36736	4	1-10907	36735	74	69	2165	30	2012-08-02 11:12:25.864716
36737	4	1-8209	36735	69	75	1631	41	2012-08-02 11:12:25.864716
36738	4	1-9875	36735	75	39	1798	100	2012-08-02 11:12:25.864716
36735	4	1-11244	36735	39	74	1876	39	2012-08-02 11:12:25.864716
36761	4	1-10860	36760	57	90	4507	3	2012-08-02 11:12:31.291449
36762	4	1-8889	36760	90	31	3276	93	2012-08-02 11:12:31.291449
36763	4	1-10516	36760	31	11	1919	64	2012-08-02 11:12:31.291449
36760	4	1-11258	36760	11	57	5470	45	2012-08-02 11:12:31.291449
36820	3	1-6882	36819	66	87	1227	84	2012-08-02 11:12:36.022845
36821	3	1-10510	36819	87	32	980	85	2012-08-02 11:12:36.022845
36819	3	1-11270	36819	32	66	2440	75	2012-08-02 11:12:36.022845
36823	3	1-8702	36822	49	89	89	45	2012-08-02 11:12:36.022845
36824	3	1-7025	36822	89	32	60	85	2012-08-02 11:12:36.022845
36822	3	1-11270	36822	32	49	151	75	2012-08-02 11:12:36.022845
36833	3	1-11243	36832	81	47	1793	84	2012-08-02 11:12:37.191805
36834	3	1-9245	36832	47	78	1151	72	2012-08-02 11:12:37.191805
36832	3	1-11273	36832	78	81	2895	11	2012-08-02 11:12:37.191805
36861	2	1-10629	36860	19	82	1255	31	2012-08-02 11:12:40.649859
36860	2	1-11282	36860	82	19	1245	43	2012-08-02 11:12:40.649859
36863	4	1-9422	36862	91	39	510	100	2012-08-02 11:12:41.821404
36864	4	1-11244	36862	39	72	488	39	2012-08-02 11:12:41.821404
36865	4	1-10360	36862	72	68	1208	80	2012-08-02 11:12:41.821404
36862	4	1-11285	36862	68	91	1298	45	2012-08-02 11:12:41.821404
36886	4	1-214	36885	19	64	375	21	2012-08-02 11:12:46.540917
36887	4	1-11013	36885	64	11	329	93	2012-08-02 11:12:46.540917
36885	4	1-11296	36885	31	19	1726	82	2012-08-02 11:12:46.540917
36785	5	1-10017	36782	3	30	981	49	2012-08-02 11:12:32.882289
36786	5	1-10700	36782	30	97	861	3	2012-08-02 11:12:32.882289
36782	5	1-11263	36782	97	89	3371	45	2012-08-02 11:12:32.882289
36788	4	1-10194	36787	51	70	719	61	2012-08-02 11:12:32.882289
36789	4	1-7704	36787	70	87	529	57	2012-08-02 11:12:32.882289
36790	4	1-8422	36787	87	97	184	3	2012-08-02 11:12:32.882289
36787	4	1-11263	36787	97	51	806	45	2012-08-02 11:12:32.882289
36792	5	1-10297	36791	73	37	203	99	2012-08-02 11:12:32.882289
36793	5	1-6803	36791	37	22	85	64	2012-08-02 11:12:32.882289
36794	5	1-10016	36791	22	30	58	49	2012-08-02 11:12:32.882289
36795	5	1-10700	36791	30	97	60	3	2012-08-02 11:12:32.882289
36791	5	1-11263	36791	97	73	278	45	2012-08-02 11:12:32.882289
36797	4	1-11009	36796	77	49	651	91	2012-08-02 11:12:32.882289
36798	4	1-10465	36796	49	40	863	58	2012-08-02 11:12:32.882289
36799	4	1-11194	36796	40	97	349	3	2012-08-02 11:12:32.882289
36796	4	1-11263	36796	97	77	1649	45	2012-08-02 11:12:32.882289
36801	2	1-7989	36800	99	27	919	37	2012-08-02 11:12:34.47351
36800	2	1-11266	36800	27	99	7331	98	2012-08-02 11:12:34.47351
36809	3	1-10493	36808	2	77	1442	22	2012-08-02 11:12:35.633704
36810	3	1-6886	36808	77	67	719	72	2012-08-02 11:12:35.633704
36808	3	1-11269	36808	67	2	986	95	2012-08-02 11:12:35.633704
36812	4	1-10493	36811	2	20	1321	22	2012-08-02 11:12:35.633704
36813	4	1-9950	36811	20	47	761	84	2012-08-02 11:12:35.633704
36814	4	1-9245	36811	47	67	653	72	2012-08-02 11:12:35.633704
36811	4	1-11269	36811	67	2	900	95	2012-08-02 11:12:35.633704
36816	4	1-11264	36815	34	35	1157	80	2012-08-02 11:12:35.633704
36817	4	1-10773	36815	35	47	479	84	2012-08-02 11:12:35.633704
36818	4	1-9245	36815	47	67	462	72	2012-08-02 11:12:35.633704
36815	4	1-11269	36815	67	34	714	95	2012-08-02 11:12:35.633704
36826	3	1-11200	36825	57	91	1588	27	2012-08-02 11:12:36.80497
36827	3	1-8332	36825	91	84	1927	44	2012-08-02 11:12:36.80497
36825	3	1-11272	36825	84	57	1855	13	2012-08-02 11:12:36.80497
36829	4	1-11082	36828	5	11	2642	7	2012-08-02 11:12:36.80497
36830	4	1-9116	36828	11	14	1531	71	2012-08-02 11:12:36.80497
36831	4	1-9930	36828	14	84	2560	44	2012-08-02 11:12:36.80497
36828	4	1-11272	36828	84	5	2513	13	2012-08-02 11:12:36.80497
36838	4	1-9660	36837	91	63	662	87	2012-08-02 11:12:37.942975
36839	4	1-10211	36837	63	90	470	44	2012-08-02 11:12:37.942975
36840	4	1-9369	36837	90	94	191	25	2012-08-02 11:12:37.942975
36837	4	1-11275	36837	94	91	929	53	2012-08-02 11:12:37.942975
36842	4	1-8974	36841	29	64	433	66	2012-08-02 11:12:37.942975
36843	4	1-10987	36841	64	68	148	28	2012-08-02 11:12:37.942975
36844	4	1-10480	36841	68	94	215	25	2012-08-02 11:12:37.942975
36841	4	1-11275	36841	94	29	1143	53	2012-08-02 11:12:37.942975
36846	4	1-10973	36845	62	34	421	100	2012-08-02 11:12:37.942975
36847	4	1-10092	36845	34	8	372	5	2012-08-02 11:12:37.942975
36848	4	1-11186	36845	8	94	274	25	2012-08-02 11:12:37.942975
36845	4	1-11275	36845	94	62	1902	53	2012-08-02 11:12:37.942975
36850	4	1-10182	36849	14	7	206	94	2012-08-02 11:12:37.942975
36851	4	1-6993	36849	7	8	154	5	2012-08-02 11:12:37.942975
36852	4	1-11186	36849	8	94	117	25	2012-08-02 11:12:37.942975
36849	4	1-11275	36849	94	14	833	53	2012-08-02 11:12:37.942975
36857	4	1-11054	36856	97	74	842	96	2012-08-02 11:12:40.274544
36858	4	1-8851	36856	74	8	1028	5	2012-08-02 11:12:40.274544
36859	4	1-11186	36856	8	25	695	25	2012-08-02 11:12:40.274544
36856	4	1-11281	36856	25	97	876	16	2012-08-02 11:12:40.274544
36867	4	1-10694	36866	36	60	4268	12	2012-08-02 11:12:42.219001
36868	4	1-10856	36866	60	73	3123	23	2012-08-02 11:12:42.219001
36869	4	1-8075	36866	73	56	1733	48	2012-08-02 11:12:42.219001
36866	4	1-11286	36866	56	36	6122	6	2012-08-02 11:12:42.219001
36894	4	1-6733	36893	84	27	3112	88	2012-08-02 11:12:49.283536
36895	4	1-10087	36893	27	19	1242	9	2012-08-02 11:12:49.283536
36896	4	1-10238	36893	19	83	2423	42	2012-08-02 11:12:49.283536
36893	4	1-11304	36893	83	84	2483	30	2012-08-02 11:12:49.283536
36903	4	1-9432	36902	34	48	2326	29	2012-08-02 11:12:50.432315
36904	4	1-10712	36902	48	19	3407	58	2012-08-02 11:12:50.432315
36905	4	1-10018	36902	19	92	3348	13	2012-08-02 11:12:50.432315
36902	4	1-11307	36902	92	34	4256	65	2012-08-02 11:12:50.432315
36907	4	1-11088	36906	10	14	1045	17	2012-08-02 11:12:50.432315
36908	4	1-11190	36906	14	85	693	85	2012-08-02 11:12:50.432315
36909	4	1-10917	36906	85	92	650	13	2012-08-02 11:12:50.432315
36906	4	1-11307	36906	92	10	850	65	2012-08-02 11:12:50.432315
36921	4	1-9973	36920	26	48	1594	14	2012-08-02 11:12:52.785597
36922	4	1-9910	36920	48	36	762	9	2012-08-02 11:12:52.785597
36923	4	1-8476	36920	36	30	688	71	2012-08-02 11:12:52.785597
36920	4	1-11313	36920	30	26	717	61	2012-08-02 11:12:52.785597
36933	2	1-6121	36932	58	58	1324	92	2012-08-02 11:12:55.845883
36932	2	1-11321	36932	58	58	2984	76	2012-08-02 11:12:55.845883
36935	4	1-10934	36934	29	5	70	65	2012-08-02 11:12:55.845883
36936	4	1-11149	36934	5	88	65	97	2012-08-02 11:12:55.845883
36937	4	1-8904	36934	88	58	49	92	2012-08-02 11:12:55.845883
36934	4	1-11321	36934	58	29	155	76	2012-08-02 11:12:55.845883
36939	4	1-6531	36938	86	18	265	73	2012-08-02 11:12:59.271544
36940	4	1-11228	36938	18	20	413	26	2012-08-02 11:12:59.271544
36941	4	1-8747	36938	20	31	511	25	2012-08-02 11:12:59.271544
36938	4	1-11331	36938	31	86	488	61	2012-08-02 11:12:59.271544
36954	4	1-8584	36953	69	53	60	31	2012-08-02 11:13:00.806721
36955	4	1-10144	36953	53	47	49	23	2012-08-02 11:13:00.806721
36956	4	1-11138	36953	47	81	81	98	2012-08-02 11:13:00.806721
36953	4	1-11334	36953	81	69	170	39	2012-08-02 11:13:00.806721
36966	4	1-11107	36965	78	78	2881	59	2012-08-02 11:13:03.182251
36967	4	1-10953	36965	78	78	2448	32	2012-08-02 11:13:03.182251
36968	4	1-9209	36965	78	43	984	46	2012-08-02 11:13:03.182251
36965	4	1-11340	36965	43	78	2138	74	2012-08-02 11:13:03.182251
36970	4	1-10768	36969	93	3	3777	75	2012-08-02 11:13:03.182251
36971	4	1-10017	36969	3	29	1681	49	2012-08-02 11:13:03.182251
36972	4	1-11043	36969	29	43	1503	46	2012-08-02 11:13:03.182251
36969	4	1-11340	36969	43	93	3440	74	2012-08-02 11:13:03.182251
36973	1	1-9092	\N	99	99	761	32	2012-08-02 11:13:03.182251
36978	4	1-10484	36977	38	63	5438	95	2012-08-02 11:13:04.331216
36979	4	1-10827	36977	63	54	2877	56	2012-08-02 11:13:04.331216
36980	4	1-11213	36977	54	25	1357	24	2012-08-02 11:13:04.331216
36977	4	1-11343	36977	25	38	5445	36	2012-08-02 11:13:04.331216
36993	4	1-10575	36992	69	9	1059	68	2012-08-02 11:13:05.480265
36992	4	1-11346	36992	73	69	1418	37	2012-08-02 11:13:05.480265
36888	4	1-9763	36885	11	31	165	73	2012-08-02 11:12:46.540917
36898	5	1-10159	36897	93	56	4973	52	2012-08-02 11:12:49.681048
36899	5	1-10691	36897	56	18	3696	12	2012-08-02 11:12:49.681048
36900	5	1-11036	36897	18	18	2898	73	2012-08-02 11:12:49.681048
36901	5	1-11228	36897	18	86	3838	26	2012-08-02 11:12:49.681048
36897	5	1-11305	36897	86	93	6515	100	2012-08-02 11:12:49.681048
36911	2	1-10845	36910	89	80	766	19	2012-08-02 11:12:50.819099
36910	2	1-11308	36910	80	89	8808	73	2012-08-02 11:12:50.819099
36925	4	1-10463	36924	93	57	14	28	2012-08-02 11:12:53.161303
36926	4	1-10899	36924	57	19	20	94	2012-08-02 11:12:53.161303
36927	4	1-6150	36924	19	54	12	59	2012-08-02 11:12:53.161303
36924	4	1-11314	36924	54	93	14	90	2012-08-02 11:12:53.161303
36958	4	1-11143	36957	93	63	1123	95	2012-08-02 11:13:02.386499
36959	4	1-10904	36957	63	76	1573	82	2012-08-02 11:13:02.386499
36960	4	1-11024	36957	76	3	1012	34	2012-08-02 11:13:02.386499
36957	4	1-11338	36957	3	93	2584	50	2012-08-02 11:13:02.386499
36962	4	1-9629	36961	16	49	598	62	2012-08-02 11:13:02.386499
36963	4	1-10995	36961	49	88	756	27	2012-08-02 11:13:02.386499
36964	4	1-11047	36961	88	3	585	34	2012-08-02 11:13:02.386499
36961	4	1-11338	36961	3	16	1501	50	2012-08-02 11:13:02.386499
36975	3	1-11290	36974	72	95	2983	69	2012-08-02 11:13:03.579887
36976	3	1-9542	36974	95	76	1996	71	2012-08-02 11:13:03.579887
36974	3	1-11341	36974	76	72	4748	21	2012-08-02 11:13:03.579887
36982	4	1-10862	36981	32	56	2011	52	2012-08-02 11:13:04.706838
36983	4	1-10691	36981	56	18	2034	12	2012-08-02 11:13:04.706838
36984	4	1-11036	36981	18	65	1831	73	2012-08-02 11:13:04.706838
36981	4	1-11344	36981	65	32	1803	48	2012-08-02 11:13:04.706838
37002	2	1-10285	37001	43	20	2292	8	2012-08-02 11:13:05.877969
37001	2	1-11347	37001	20	43	2587	18	2012-08-02 11:13:05.877969
37059	4	1-8730	37058	92	78	225	74	2012-08-02 11:13:18.950085
37060	4	1-11107	37058	78	42	307	59	2012-08-02 11:13:18.950085
37061	4	1-9068	37058	42	80	358	85	2012-08-02 11:13:18.950085
37058	4	1-11380	37058	80	92	624	94	2012-08-02 11:13:18.950085
37063	3	1-6222	37062	33	87	108	84	2012-08-02 11:13:18.950085
37064	3	1-10510	37062	87	80	132	85	2012-08-02 11:13:18.950085
37062	3	1-11380	37062	80	33	252	94	2012-08-02 11:13:18.950085
37066	4	1-9938	37065	8	100	231	15	2012-08-02 11:13:18.950085
37067	4	1-11379	37065	100	91	279	57	2012-08-02 11:13:18.950085
37068	4	1-10134	37065	91	80	150	85	2012-08-02 11:13:18.950085
37065	4	1-11380	37065	80	8	295	94	2012-08-02 11:13:18.950085
37097	4	1-11262	37096	86	79	3791	57	2012-08-02 11:13:19.734778
37098	4	1-11354	37096	79	24	2717	39	2012-08-02 11:13:19.734778
37099	4	1-11310	37096	24	89	1155	13	2012-08-02 11:13:19.734778
37096	4	1-11383	37096	89	86	4029	29	2012-08-02 11:13:19.734778
37101	3	1-11179	37100	64	25	1659	27	2012-08-02 11:13:19.734778
37102	3	1-11159	37100	25	89	1057	13	2012-08-02 11:13:19.734778
37100	3	1-11383	37100	89	64	3694	29	2012-08-02 11:13:19.734778
37162	4	1-10730	37161	92	68	1170	80	2012-08-02 11:13:25.555476
37163	4	1-11285	37161	68	89	697	45	2012-08-02 11:13:25.555476
37164	4	1-7025	37161	89	42	365	85	2012-08-02 11:13:25.555476
37161	4	1-11398	37161	42	92	2567	69	2012-08-02 11:13:25.555476
37166	3	1-10730	37165	92	38	2677	80	2012-08-02 11:13:25.555476
37167	3	1-6079	37165	38	42	647	85	2012-08-02 11:13:25.555476
37165	3	1-11398	37165	42	92	5169	69	2012-08-02 11:13:25.555476
37169	4	1-11359	37168	77	84	2868	94	2012-08-02 11:13:26.705297
37170	4	1-10363	37168	84	34	2046	46	2012-08-02 11:13:26.705297
37171	4	1-11326	37168	34	41	1238	22	2012-08-02 11:13:26.705297
37168	4	1-11401	37168	41	77	4554	6	2012-08-02 11:13:26.705297
37173	4	1-10340	37172	74	82	210	32	2012-08-02 11:13:26.705297
37174	4	1-11345	37172	82	99	95	33	2012-08-02 11:13:26.705297
37175	4	1-8220	37172	99	41	72	22	2012-08-02 11:13:26.705297
37172	4	1-11401	37172	41	74	268	6	2012-08-02 11:13:26.705297
37177	4	1-10340	37176	74	66	1329	32	2012-08-02 11:13:26.705297
37178	4	1-10449	37176	66	89	1837	25	2012-08-02 11:13:26.705297
37179	4	1-6176	37176	89	41	383	22	2012-08-02 11:13:26.705297
37176	4	1-11401	37176	41	74	1556	6	2012-08-02 11:13:26.705297
37181	4	1-10837	37180	96	68	318	28	2012-08-02 11:13:26.705297
37182	4	1-10480	37180	68	89	490	25	2012-08-02 11:13:26.705297
37183	4	1-6176	37180	89	41	113	22	2012-08-02 11:13:26.705297
37180	4	1-11401	37180	41	96	505	6	2012-08-02 11:13:26.705297
37185	3	1-10834	37184	34	11	1006	90	2012-08-02 11:13:26.705297
37186	3	1-8030	37184	11	41	238	22	2012-08-02 11:13:26.705297
37184	3	1-11401	37184	41	34	1115	6	2012-08-02 11:13:26.705297
37188	4	1-10694	37187	36	69	36	12	2012-08-02 11:13:26.705297
37189	4	1-11064	37187	69	85	76	50	2012-08-02 11:13:26.705297
37190	4	1-5743	37187	85	41	9	22	2012-08-02 11:13:26.705297
37187	4	1-11401	37187	41	36	44	6	2012-08-02 11:13:26.705297
37192	4	1-10694	37191	36	60	741	12	2012-08-02 11:13:26.705297
37193	4	1-10856	37191	60	89	640	23	2012-08-02 11:13:26.705297
37194	4	1-6123	37191	89	41	181	22	2012-08-02 11:13:26.705297
37191	4	1-11401	37191	41	36	901	6	2012-08-02 11:13:26.705297
37196	4	1-10689	37195	48	66	275	32	2012-08-02 11:13:26.705297
37197	4	1-10449	37195	66	89	520	25	2012-08-02 11:13:26.705297
37198	4	1-6176	37195	89	41	148	22	2012-08-02 11:13:26.705297
37195	4	1-11401	37195	41	48	823	6	2012-08-02 11:13:26.705297
37208	4	1-8605	37207	16	53	182	49	2012-08-02 11:13:28.01982
37209	4	1-11303	37207	53	60	256	12	2012-08-02 11:13:28.01982
37210	4	1-10856	37207	60	100	219	23	2012-08-02 11:13:28.01982
37207	4	1-11404	37207	100	16	449	77	2012-08-02 11:13:28.01982
37218	2	1-6281	37217	56	92	187	5	2012-08-02 11:13:30.374755
37217	2	1-11411	37217	92	56	2930	78	2012-08-02 11:13:30.374755
37222	4	1-10967	37221	45	91	492	57	2012-08-02 11:13:31.932553
37223	4	1-10134	37221	91	42	117	85	2012-08-02 11:13:31.932553
37224	4	1-10961	37221	42	7	45	17	2012-08-02 11:13:31.932553
37221	4	1-11414	37221	7	45	865	29	2012-08-02 11:13:31.932553
37234	2	1-5186	37233	11	21	964	81	2012-08-02 11:13:35.33513
37233	2	1-11423	37233	21	11	4584	61	2012-08-02 11:13:35.33513
37259	4	1-10977	37258	65	57	296	28	2012-08-02 11:13:39.975895
37260	4	1-10899	37258	57	74	421	94	2012-08-02 11:13:39.975895
37261	4	1-10040	37258	74	21	311	58	2012-08-02 11:13:39.975895
37258	4	1-11435	37258	21	65	302	38	2012-08-02 11:13:39.975895
37291	4	1-11309	37290	59	53	4789	59	2012-08-02 11:13:51.886711
37292	4	1-8236	37290	53	72	1871	71	2012-08-02 11:13:51.886711
37293	4	1-11454	37290	72	90	769	78	2012-08-02 11:13:51.886711
37290	4	1-11466	37290	90	59	6561	80	2012-08-02 11:13:51.886711
36917	4	1-10982	36916	53	50	1075	9	2012-08-02 11:12:52.388119
36918	4	1-7534	36916	50	37	692	34	2012-08-02 11:12:52.388119
36919	4	1-10595	36916	37	43	403	22	2012-08-02 11:12:52.388119
36916	4	1-11312	36916	43	53	2422	83	2012-08-02 11:12:52.388119
36929	4	1-10473	36928	32	33	1012	64	2012-08-02 11:12:54.686138
36930	4	1-10838	36928	33	63	1049	87	2012-08-02 11:12:54.686138
36931	4	1-10211	36928	63	79	543	44	2012-08-02 11:12:54.686138
36928	4	1-11319	36928	79	32	6013	58	2012-08-02 11:12:54.686138
36943	4	1-10982	36942	53	1	351	9	2012-08-02 11:13:00.408955
36944	4	1-5611	36942	1	88	141	3	2012-08-02 11:13:00.408955
36945	4	1-10109	36942	88	29	80	94	2012-08-02 11:13:00.408955
36942	4	1-11333	36942	29	53	849	83	2012-08-02 11:13:00.408955
36947	3	1-8377	36946	90	34	328	7	2012-08-02 11:13:00.408955
36948	3	1-6770	36946	34	29	69	94	2012-08-02 11:13:00.408955
36946	3	1-11333	36946	29	90	756	83	2012-08-02 11:13:00.408955
36950	4	1-8771	36949	18	67	223	51	2012-08-02 11:13:00.408955
36951	4	1-11105	36949	67	44	277	36	2012-08-02 11:13:00.408955
36952	4	1-10192	36949	44	29	147	94	2012-08-02 11:13:00.408955
36949	4	1-11333	36949	29	18	1627	83	2012-08-02 11:13:00.408955
36986	4	1-11283	36985	22	1	1804	86	2012-08-02 11:13:05.082423
36987	4	1-9838	36985	1	54	1845	34	2012-08-02 11:13:05.082423
36988	4	1-11246	36985	54	82	3362	32	2012-08-02 11:13:05.082423
36985	4	1-11345	36985	82	22	2015	33	2012-08-02 11:13:05.082423
36990	3	1-10628	36989	13	74	2635	6	2012-08-02 11:13:05.082423
36991	3	1-10340	36989	74	82	2990	32	2012-08-02 11:13:05.082423
36989	3	1-11345	36989	82	13	1961	33	2012-08-02 11:13:05.082423
37004	4	1-10845	37003	89	50	1220	19	2012-08-02 11:13:07.46978
37005	4	1-9770	37003	50	78	258	59	2012-08-02 11:13:07.46978
37006	4	1-10762	37003	78	12	33	57	2012-08-02 11:13:07.46978
37003	4	1-11351	37003	12	89	6895	73	2012-08-02 11:13:07.46978
37012	3	1-7332	37011	17	86	4148	29	2012-08-02 11:13:08.629456
37013	3	1-11262	37011	86	79	4920	57	2012-08-02 11:13:08.629456
37011	3	1-11354	37011	79	17	4446	39	2012-08-02 11:13:08.629456
37019	2	1-11304	37018	83	37	2549	30	2012-08-02 11:13:09.78955
37018	2	1-11357	37018	37	83	2728	42	2012-08-02 11:13:09.78955
37044	3	1-11029	37043	69	90	189	27	2012-08-02 11:13:17.41469
37045	3	1-10664	37043	90	46	262	16	2012-08-02 11:13:17.41469
37043	3	1-11377	37043	46	69	266	94	2012-08-02 11:13:17.41469
37070	4	1-10975	37069	53	61	2425	46	2012-08-02 11:13:18.574619
37071	4	1-11355	37069	61	9	2580	62	2012-08-02 11:13:18.574619
37072	4	1-7449	37069	9	68	1902	60	2012-08-02 11:13:18.574619
37069	4	1-11381	37069	68	53	1785	89	2012-08-02 11:13:18.574619
37074	4	1-10975	37073	53	28	47	46	2012-08-02 11:13:18.574619
37075	4	1-10999	37073	28	17	30	83	2012-08-02 11:13:18.574619
37076	4	1-10138	37073	17	68	32	60	2012-08-02 11:13:18.574619
37073	4	1-11381	37073	68	53	32	89	2012-08-02 11:13:18.574619
37078	3	1-7433	37077	13	17	576	83	2012-08-02 11:13:18.574619
37079	3	1-10138	37077	17	68	699	60	2012-08-02 11:13:18.574619
37077	3	1-11381	37077	68	13	792	89	2012-08-02 11:13:18.574619
37104	4	1-10691	37103	56	69	1827	12	2012-08-02 11:13:20.121622
37105	4	1-11064	37103	69	93	3317	50	2012-08-02 11:13:20.121622
37106	4	1-11143	37103	93	19	1097	95	2012-08-02 11:13:20.121622
37103	4	1-11384	37103	19	56	2371	52	2012-08-02 11:13:20.121622
37108	4	1-10691	37107	56	88	1078	12	2012-08-02 11:13:20.121622
37109	4	1-10608	37107	88	79	780	13	2012-08-02 11:13:20.121622
37110	4	1-7296	37107	79	19	431	95	2012-08-02 11:13:20.121622
37107	4	1-11384	37107	19	56	1141	52	2012-08-02 11:13:20.121622
37116	5	1-10886	37115	67	47	866	23	2012-08-02 11:13:21.30417
37117	5	1-11138	37115	47	21	1158	98	2012-08-02 11:13:21.30417
37118	5	1-11256	37115	21	11	597	26	2012-08-02 11:13:21.30417
37119	5	1-10554	37115	11	84	660	18	2012-08-02 11:13:21.30417
37115	5	1-11387	37115	84	67	892	33	2012-08-02 11:13:21.30417
37121	4	1-10708	37120	64	1	4311	53	2012-08-02 11:13:21.30417
37122	4	1-11034	37120	1	82	4722	17	2012-08-02 11:13:21.30417
37123	4	1-11130	37120	82	84	2864	18	2012-08-02 11:13:21.30417
37120	4	1-11387	37120	84	64	4932	33	2012-08-02 11:13:21.30417
37125	5	1-9841	37124	57	99	1971	91	2012-08-02 11:13:21.30417
37126	5	1-10146	37124	99	21	632	98	2012-08-02 11:13:21.30417
37127	5	1-11256	37124	21	11	464	26	2012-08-02 11:13:21.30417
37128	5	1-10554	37124	11	84	730	18	2012-08-02 11:13:21.30417
37124	5	1-11387	37124	84	57	1406	33	2012-08-02 11:13:21.30417
37130	4	1-10708	37129	64	76	451	53	2012-08-02 11:13:21.30417
37131	4	1-10354	37129	76	82	271	17	2012-08-02 11:13:21.30417
37132	4	1-11130	37129	82	84	201	18	2012-08-02 11:13:21.30417
37129	4	1-11387	37129	84	64	422	33	2012-08-02 11:13:21.30417
37134	3	1-11107	37133	78	70	4666	59	2012-08-02 11:13:22.484345
37135	3	1-9477	37133	70	33	2848	23	2012-08-02 11:13:22.484345
37133	3	1-11390	37133	33	78	4685	74	2012-08-02 11:13:22.484345
37137	4	1-11108	37136	7	59	1194	80	2012-08-02 11:13:22.484345
37138	4	1-11309	37136	59	70	919	59	2012-08-02 11:13:22.484345
37139	4	1-9477	37136	70	33	671	23	2012-08-02 11:13:22.484345
37136	4	1-11390	37136	33	7	1322	74	2012-08-02 11:13:22.484345
37141	4	1-10768	37140	93	3	1909	75	2012-08-02 11:13:22.484345
37142	4	1-10017	37140	3	59	1146	49	2012-08-02 11:13:22.484345
37143	4	1-9120	37140	59	33	583	23	2012-08-02 11:13:22.484345
37140	4	1-11390	37140	33	93	1289	74	2012-08-02 11:13:22.484345
37158	4	1-9682	37157	28	32	2449	10	2012-08-02 11:13:24.793267
37159	4	1-9834	37157	32	24	2480	52	2012-08-02 11:13:24.793267
37160	4	1-11279	37157	24	100	1717	35	2012-08-02 11:13:24.793267
37157	4	1-11396	37157	100	28	1243	63	2012-08-02 11:13:24.793267
37212	3	1-11274	37211	78	67	287	29	2012-08-02 11:13:30.75082
37213	3	1-11094	37211	67	59	73	72	2012-08-02 11:13:30.75082
37211	3	1-11410	37211	59	78	401	25	2012-08-02 11:13:30.75082
37215	3	1-8117	37214	99	33	882	40	2012-08-02 11:13:30.75082
37216	3	1-10776	37214	33	59	486	72	2012-08-02 11:13:30.75082
37214	3	1-11410	37214	59	99	2954	25	2012-08-02 11:13:30.75082
37220	2	1-10603	37219	13	82	314	91	2012-08-02 11:13:31.551961
37219	2	1-11412	37219	82	13	4808	21	2012-08-02 11:13:31.551961
37226	4	1-9542	37225	95	93	956	71	2012-08-02 11:13:34.572958
37227	4	1-11172	37225	93	24	1412	54	2012-08-02 11:13:34.572958
37228	4	1-9726	37225	24	78	281	1	2012-08-02 11:13:34.572958
37225	4	1-11421	37225	78	95	1426	69	2012-08-02 11:13:34.572958
37230	4	1-8831	37229	19	96	86	70	2012-08-02 11:13:34.572958
37231	4	1-11183	37229	96	43	49	9	2012-08-02 11:13:34.572958
36994	4	1-11292	36992	9	80	998	80	2012-08-02 11:13:05.480265
36995	4	1-9917	36992	80	73	180	73	2012-08-02 11:13:05.480265
36997	4	1-10958	36996	20	11	1858	43	2012-08-02 11:13:05.480265
36998	4	1-8137	36996	11	7	1115	49	2012-08-02 11:13:05.480265
36999	4	1-4510	36996	7	73	503	73	2012-08-02 11:13:05.480265
36996	4	1-11346	36996	73	20	3965	37	2012-08-02 11:13:05.480265
37000	1	1-4510	\N	7	7	1814	73	2012-08-02 11:13:05.480265
37008	4	1-11018	37007	44	27	1404	51	2012-08-02 11:13:07.84488
37009	4	1-10850	37007	27	40	1165	75	2012-08-02 11:13:07.84488
37010	4	1-9943	37007	40	43	924	9	2012-08-02 11:13:07.84488
37007	4	1-11352	37007	43	44	672	1	2012-08-02 11:13:07.84488
37015	4	1-9616	37014	94	16	1829	77	2012-08-02 11:13:09.01619
37016	4	1-8605	37014	16	29	835	49	2012-08-02 11:13:09.01619
37017	4	1-11043	37014	29	61	1176	46	2012-08-02 11:13:09.01619
37014	4	1-11355	37014	61	94	1596	62	2012-08-02 11:13:09.01619
37021	2	1-9320	37020	93	19	6322	5	2012-08-02 11:13:11.303056
37020	2	1-11361	37020	19	93	8285	66	2012-08-02 11:13:11.303056
37023	4	1-10238	37022	19	83	64	42	2012-08-02 11:13:12.452102
37024	4	1-11304	37022	83	47	64	30	2012-08-02 11:13:12.452102
37025	4	1-4684	37022	47	63	12	19	2012-08-02 11:13:12.452102
37022	4	1-11364	37022	63	19	34	9	2012-08-02 11:13:12.452102
37027	4	1-8476	37026	36	93	354	71	2012-08-02 11:13:12.452102
37028	4	1-11172	37026	93	7	777	54	2012-08-02 11:13:12.452102
37029	4	1-8291	37026	7	63	129	19	2012-08-02 11:13:12.452102
37026	4	1-11364	37026	63	36	390	9	2012-08-02 11:13:12.452102
37030	1	1-2077	\N	62	62	269	9	2012-08-02 11:13:12.452102
37031	1	1-385	\N	72	72	159	9	2012-08-02 11:13:12.452102
37033	4	1-11169	37032	90	67	793	72	2012-08-02 11:13:13.590162
37034	4	1-11269	37032	67	34	1121	95	2012-08-02 11:13:13.590162
37035	4	1-11264	37032	34	43	1660	80	2012-08-02 11:13:13.590162
37032	4	1-11367	37032	43	90	1410	76	2012-08-02 11:13:13.590162
37037	3	1-10926	37036	70	37	7081	8	2012-08-02 11:13:14.750192
37038	3	1-11004	37036	37	53	5444	75	2012-08-02 11:13:14.750192
37036	3	1-11370	37036	53	70	4833	35	2012-08-02 11:13:14.750192
37040	4	1-10186	37039	10	59	788	80	2012-08-02 11:13:17.027691
37041	4	1-11309	37039	59	78	794	59	2012-08-02 11:13:17.027691
37042	4	1-10953	37039	78	35	782	32	2012-08-02 11:13:17.027691
37039	4	1-11376	37039	35	10	1749	17	2012-08-02 11:13:17.027691
37047	4	1-10870	37046	63	53	2046	83	2012-08-02 11:13:18.187985
37048	4	1-10982	37046	53	45	829	9	2012-08-02 11:13:18.187985
37049	4	1-10714	37046	45	100	1331	15	2012-08-02 11:13:18.187985
37046	4	1-11379	37046	100	63	1352	57	2012-08-02 11:13:18.187985
37051	4	1-10512	37050	46	41	574	58	2012-08-02 11:13:18.187985
37052	4	1-9807	37050	41	45	199	9	2012-08-02 11:13:18.187985
37053	4	1-10714	37050	45	100	337	15	2012-08-02 11:13:18.187985
37050	4	1-11379	37050	100	46	360	57	2012-08-02 11:13:18.187985
37055	4	1-10870	37054	63	17	2334	83	2012-08-02 11:13:18.187985
37056	4	1-10138	37054	17	40	2654	60	2012-08-02 11:13:18.187985
37057	4	1-7331	37054	40	100	1319	15	2012-08-02 11:13:18.187985
37054	4	1-11379	37054	100	63	1437	57	2012-08-02 11:13:18.187985
37081	4	1-9374	37080	59	96	5249	52	2012-08-02 11:13:19.348013
37082	4	1-10442	37080	96	43	3081	80	2012-08-02 11:13:19.348013
37083	4	1-11367	37080	43	41	2292	76	2012-08-02 11:13:19.348013
37080	4	1-11382	37080	41	59	5454	91	2012-08-02 11:13:19.348013
37085	4	1-8282	37084	46	52	225	18	2012-08-02 11:13:19.348013
37086	4	1-10783	37084	52	51	275	89	2012-08-02 11:13:19.348013
37087	4	1-11291	37084	51	41	236	76	2012-08-02 11:13:19.348013
37084	4	1-11382	37084	41	46	579	91	2012-08-02 11:13:19.348013
37089	4	1-11035	37088	22	91	1112	21	2012-08-02 11:13:19.348013
37090	4	1-4610	37088	91	43	423	80	2012-08-02 11:13:19.348013
37091	4	1-11367	37088	43	41	333	76	2012-08-02 11:13:19.348013
37088	4	1-11382	37088	41	22	837	91	2012-08-02 11:13:19.348013
37093	4	1-9171	37092	81	43	137	95	2012-08-02 11:13:19.348013
37094	4	1-10869	37092	43	37	204	4	2012-08-02 11:13:19.348013
37095	4	1-9592	37092	37	41	71	76	2012-08-02 11:13:19.348013
37092	4	1-11382	37092	41	81	192	91	2012-08-02 11:13:19.348013
37112	4	1-8098	37111	23	92	1206	13	2012-08-02 11:13:20.53064
37113	4	1-11307	37111	92	83	1500	65	2012-08-02 11:13:20.53064
37114	4	1-11336	37111	83	73	970	35	2012-08-02 11:13:20.53064
37111	4	1-11385	37111	73	23	3993	14	2012-08-02 11:13:20.53064
37145	2	1-7967	37144	92	46	996	40	2012-08-02 11:13:22.859494
37144	2	1-11391	37144	46	92	2567	18	2012-08-02 11:13:22.859494
37147	4	1-11387	37146	84	67	1134	33	2012-08-02 11:13:22.859494
37148	4	1-10886	37146	67	78	596	23	2012-08-02 11:13:22.859494
37149	4	1-10357	37146	78	46	442	40	2012-08-02 11:13:22.859494
37146	4	1-11391	37146	46	84	1551	18	2012-08-02 11:13:22.859494
37151	4	1-10783	37150	52	83	1995	89	2012-08-02 11:13:22.859494
37152	4	1-10545	37150	83	83	1017	28	2012-08-02 11:13:22.859494
37153	4	1-8809	37150	83	46	598	40	2012-08-02 11:13:22.859494
37150	4	1-11391	37150	46	52	2847	18	2012-08-02 11:13:22.859494
37155	3	1-7375	37154	32	42	1355	42	2012-08-02 11:13:22.859494
37156	3	1-9649	37154	42	46	523	40	2012-08-02 11:13:22.859494
37154	3	1-11391	37154	46	32	2928	18	2012-08-02 11:13:22.859494
37200	4	1-11379	37199	100	79	3002	57	2012-08-02 11:13:27.599334
37201	4	1-11354	37199	79	17	2407	39	2012-08-02 11:13:27.599334
37202	4	1-8578	37199	17	70	975	84	2012-08-02 11:13:27.599334
37199	4	1-11403	37199	70	100	3097	15	2012-08-02 11:13:27.599334
37204	4	1-11379	37203	100	32	1823	57	2012-08-02 11:13:27.599334
37205	4	1-10532	37203	32	20	1916	42	2012-08-02 11:13:27.599334
37206	4	1-4070	37203	20	70	415	84	2012-08-02 11:13:27.599334
37203	4	1-11403	37203	70	100	1575	15	2012-08-02 11:13:27.599334
37248	3	1-10934	37247	29	85	1934	65	2012-08-02 11:13:37.29101
37249	3	1-10387	37247	85	19	1268	14	2012-08-02 11:13:37.29101
37247	3	1-11428	37247	19	29	4574	76	2012-08-02 11:13:37.29101
37251	4	1-10845	37250	89	63	650	19	2012-08-02 11:13:39.588927
37252	4	1-11364	37250	63	85	1641	9	2012-08-02 11:13:39.588927
37253	4	1-3175	37250	85	93	371	64	2012-08-02 11:13:39.588927
37250	4	1-11434	37250	93	89	799	73	2012-08-02 11:13:39.588927
37255	4	1-5856	37254	53	81	158	98	2012-08-02 11:13:39.588927
37256	4	1-11334	37254	81	53	362	39	2012-08-02 11:13:39.588927
37257	4	1-10540	37254	53	93	276	64	2012-08-02 11:13:39.588927
37254	4	1-11434	37254	93	53	605	73	2012-08-02 11:13:39.588927
37263	4	1-11345	37262	82	78	228	33	2012-08-02 11:13:40.749313
37262	4	1-11437	37262	22	82	357	32	2012-08-02 11:13:40.749313
37232	4	1-11352	37229	43	78	27	1	2012-08-02 11:13:34.572958
37229	4	1-11421	37229	78	19	146	69	2012-08-02 11:13:34.572958
37236	4	1-11376	37235	35	39	2705	17	2012-08-02 11:13:35.722017
37237	4	1-11059	37235	39	51	2003	41	2012-08-02 11:13:35.722017
37238	4	1-5302	37235	51	81	1582	58	2012-08-02 11:13:35.722017
37235	4	1-11424	37235	81	35	1368	32	2012-08-02 11:13:35.722017
37240	4	1-11345	37239	82	57	841	33	2012-08-02 11:13:35.722017
37241	4	1-9841	37239	57	49	1195	91	2012-08-02 11:13:35.722017
37242	4	1-10465	37239	49	81	1476	58	2012-08-02 11:13:35.722017
37239	4	1-11424	37239	81	82	1342	32	2012-08-02 11:13:35.722017
37244	4	1-10588	37243	60	45	1596	29	2012-08-02 11:13:35.722017
37245	4	1-10967	37243	45	46	2051	57	2012-08-02 11:13:35.722017
37246	4	1-10512	37243	46	81	3696	58	2012-08-02 11:13:35.722017
37243	4	1-11424	37243	81	60	3565	32	2012-08-02 11:13:35.722017
37271	4	1-11338	37270	3	56	1274	50	2012-08-02 11:13:45.798923
37272	4	1-7400	37270	56	44	220	1	2012-08-02 11:13:45.798923
37273	4	1-11018	37270	44	48	479	51	2012-08-02 11:13:45.798923
37270	4	1-11450	37270	48	3	505	34	2012-08-02 11:13:45.798923
37275	4	1-11106	37274	2	18	330	20	2012-08-02 11:13:45.798923
37276	4	1-11067	37274	18	85	366	67	2012-08-02 11:13:45.798923
37277	4	1-11402	37274	85	48	373	51	2012-08-02 11:13:45.798923
37274	4	1-11450	37274	48	2	406	34	2012-08-02 11:13:45.798923
37283	4	1-11236	37282	63	21	1293	50	2012-08-02 11:13:49.190803
37284	4	1-10235	37282	21	95	316	65	2012-08-02 11:13:49.190803
37285	4	1-4771	37282	95	39	56	49	2012-08-02 11:13:49.190803
37282	4	1-11459	37282	39	63	225	20	2012-08-02 11:13:49.190803
37287	4	1-11130	37286	82	24	1302	18	2012-08-02 11:13:50.361925
37288	4	1-8819	37286	24	89	136	39	2012-08-02 11:13:50.361925
37289	4	1-10207	37286	89	42	49	87	2012-08-02 11:13:50.361925
37286	4	1-11462	37286	42	82	7533	17	2012-08-02 11:13:50.361925
37303	4	1-8001	37302	69	68	2467	60	2012-08-02 11:13:52.637974
37304	4	1-11381	37302	68	67	2580	89	2012-08-02 11:13:52.637974
37305	4	1-10505	37302	67	33	2742	80	2012-08-02 11:13:52.637974
37302	4	1-11468	37302	33	69	2453	4	2012-08-02 11:13:52.637974
37307	4	1-10747	37306	6	39	1086	100	2012-08-02 11:13:52.637974
37308	4	1-11244	37306	39	46	1162	39	2012-08-02 11:13:52.637974
37309	4	1-8444	37306	46	33	972	80	2012-08-02 11:13:52.637974
37306	4	1-11468	37306	33	6	989	4	2012-08-02 11:13:52.637974
37322	2	1-11021	37321	19	40	6427	69	2012-08-02 11:13:53.797887
37321	2	1-11471	37321	40	19	5730	19	2012-08-02 11:13:53.797887
37324	3	1-11364	37323	63	88	3731	9	2012-08-02 11:13:53.797887
37325	3	1-10204	37323	88	40	1609	69	2012-08-02 11:13:53.797887
37323	3	1-11471	37323	40	63	1465	19	2012-08-02 11:13:53.797887
37351	4	1-10569	37350	8	45	4310	58	2012-08-02 11:13:56.871467
37352	4	1-11328	37350	45	92	4071	21	2012-08-02 11:13:56.871467
37353	4	1-11148	37350	92	92	2897	75	2012-08-02 11:13:56.871467
37350	4	1-11479	37350	92	8	5471	59	2012-08-02 11:13:56.871467
37355	3	1-8170	37354	65	70	680	84	2012-08-02 11:13:58.03155
37356	3	1-11403	37354	70	17	2433	15	2012-08-02 11:13:58.03155
37354	3	1-11482	37354	17	65	1386	89	2012-08-02 11:13:58.03155
37369	4	1-9432	37368	34	25	870	29	2012-08-02 11:14:01.500582
37370	4	1-11393	37368	25	39	768	86	2012-08-02 11:14:01.500582
37371	4	1-10331	37368	39	35	730	60	2012-08-02 11:14:01.500582
37368	4	1-11491	37368	35	34	3627	65	2012-08-02 11:14:01.500582
37405	4	1-10134	37404	91	42	1037	85	2012-08-02 11:14:07.334501
37406	4	1-10961	37404	42	70	898	17	2012-08-02 11:14:07.334501
37407	4	1-11372	37404	70	88	1212	48	2012-08-02 11:14:07.334501
37404	4	1-11506	37404	88	91	1928	57	2012-08-02 11:14:07.334501
37409	4	1-10532	37408	32	40	3525	42	2012-08-02 11:14:07.334501
37410	4	1-10494	37408	40	70	1319	17	2012-08-02 11:14:07.334501
37411	4	1-11372	37408	70	88	1856	48	2012-08-02 11:14:07.334501
37408	4	1-11506	37408	88	32	3077	57	2012-08-02 11:14:07.334501
37436	5	1-9985	37435	22	83	3031	42	2012-08-02 11:14:11.212304
37437	5	1-11304	37435	83	59	3027	30	2012-08-02 11:14:11.212304
37438	5	1-7251	37435	59	30	1968	49	2012-08-02 11:14:11.212304
37439	5	1-10700	37435	30	29	2035	3	2012-08-02 11:14:11.212304
37435	5	1-11516	37435	29	22	1993	75	2012-08-02 11:14:11.212304
37441	5	1-11479	37440	92	89	54	59	2012-08-02 11:14:11.212304
37442	5	1-11209	37440	89	59	41	30	2012-08-02 11:14:11.212304
37443	5	1-7251	37440	59	30	27	49	2012-08-02 11:14:11.212304
37444	5	1-10700	37440	30	29	28	3	2012-08-02 11:14:11.212304
37440	5	1-11516	37440	29	92	28	75	2012-08-02 11:14:11.212304
37456	4	1-11336	37455	83	70	237	35	2012-08-02 11:14:12.394694
37457	4	1-10926	37455	70	50	416	8	2012-08-02 11:14:12.394694
37458	4	1-11512	37455	50	92	489	24	2012-08-02 11:14:12.394694
37455	4	1-11519	37455	92	83	357	65	2012-08-02 11:14:12.394694
37518	4	1-11502	37517	98	15	668	56	2012-08-02 11:14:19.278342
37519	4	1-11133	37517	15	36	1932	47	2012-08-02 11:14:19.278342
37520	4	1-5805	37517	36	19	826	79	2012-08-02 11:14:19.278342
37517	4	1-11537	37517	19	98	1121	14	2012-08-02 11:14:19.278342
37522	4	1-9586	37521	72	83	422	42	2012-08-02 11:14:19.278342
37523	4	1-11304	37521	83	99	404	30	2012-08-02 11:14:19.278342
37524	4	1-9953	37521	99	19	350	79	2012-08-02 11:14:19.278342
37521	4	1-11537	37521	19	72	484	14	2012-08-02 11:14:19.278342
37526	4	1-11502	37525	98	57	3107	56	2012-08-02 11:14:19.278342
37527	4	1-11230	37525	57	32	3060	58	2012-08-02 11:14:19.278342
37528	4	1-11394	37525	32	19	3092	79	2012-08-02 11:14:19.278342
37525	4	1-11537	37525	19	98	4678	14	2012-08-02 11:14:19.278342
37530	4	1-11105	37529	67	59	1679	36	2012-08-02 11:14:20.438204
37531	4	1-10411	37529	59	80	3266	92	2012-08-02 11:14:20.438204
37532	4	1-11330	37529	80	22	1290	38	2012-08-02 11:14:20.438204
37529	4	1-11540	37529	22	67	1248	51	2012-08-02 11:14:20.438204
37552	3	1-11502	37551	98	42	856	56	2012-08-02 11:14:22.747415
37553	3	1-9873	37551	42	76	700	87	2012-08-02 11:14:22.747415
37551	3	1-11546	37551	76	98	1728	14	2012-08-02 11:14:22.747415
37555	4	1-8826	37554	69	99	71	69	2012-08-02 11:14:22.747415
37556	4	1-10676	37554	99	6	60	54	2012-08-02 11:14:22.747415
37557	4	1-11068	37554	6	76	81	87	2012-08-02 11:14:22.747415
37554	4	1-11546	37554	76	69	251	14	2012-08-02 11:14:22.747415
37559	4	1-9586	37558	72	25	1622	42	2012-08-02 11:14:22.747415
37560	4	1-9701	37558	25	6	431	54	2012-08-02 11:14:22.747415
37561	4	1-11068	37558	6	76	586	87	2012-08-02 11:14:22.747415
37558	4	1-11546	37558	76	72	1816	14	2012-08-02 11:14:22.747415
37264	4	1-10279	37262	78	94	322	39	2012-08-02 11:13:40.749313
37265	4	1-11055	37262	94	22	248	75	2012-08-02 11:13:40.749313
37267	4	1-6499	37266	73	48	2656	29	2012-08-02 11:13:43.14689
37268	4	1-10712	37266	48	36	3382	58	2012-08-02 11:13:43.14689
37269	4	1-10881	37266	36	32	2894	92	2012-08-02 11:13:43.14689
37266	4	1-11443	37266	32	73	4250	12	2012-08-02 11:13:43.14689
37279	4	1-10459	37278	14	81	2169	98	2012-08-02 11:13:48.814959
37280	4	1-11334	37278	81	94	5815	39	2012-08-02 11:13:48.814959
37281	4	1-11055	37278	94	50	4822	75	2012-08-02 11:13:48.814959
37278	4	1-11458	37278	50	14	2877	90	2012-08-02 11:13:48.814959
37330	2	1-7964	37329	33	10	1139	97	2012-08-02 11:13:54.582459
37329	2	1-11474	37329	10	33	2835	90	2012-08-02 11:13:54.582459
37347	4	1-11307	37346	92	5	2085	65	2012-08-02 11:13:56.109039
37348	4	1-11149	37346	5	15	2545	97	2012-08-02 11:13:56.109039
37349	4	1-8290	37346	15	100	1659	83	2012-08-02 11:13:56.109039
37346	4	1-11477	37346	100	92	1864	13	2012-08-02 11:13:56.109039
37358	4	1-11186	37357	8	31	4321	25	2012-08-02 11:13:58.407149
37359	4	1-11331	37357	31	90	3771	61	2012-08-02 11:13:58.407149
37360	4	1-11288	37357	90	87	3487	62	2012-08-02 11:13:58.407149
37357	4	1-11483	37357	87	8	6214	5	2012-08-02 11:13:58.407149
37362	3	1-10300	37361	40	25	190	64	2012-08-02 11:14:01.11395
37363	3	1-9509	37361	25	84	81	27	2012-08-02 11:14:01.11395
37361	3	1-11490	37361	84	40	438	43	2012-08-02 11:14:01.11395
37365	4	1-6887	37364	93	73	1146	5	2012-08-02 11:14:01.11395
37366	4	1-10949	37364	73	69	785	94	2012-08-02 11:14:01.11395
37367	4	1-11029	37364	69	84	396	27	2012-08-02 11:14:01.11395
37364	4	1-11490	37364	84	93	2506	43	2012-08-02 11:14:01.11395
37386	4	1-11242	37385	45	100	5150	83	2012-08-02 11:14:04.649856
37387	4	1-11477	37385	100	20	6252	13	2012-08-02 11:14:04.649856
37388	4	1-9051	37385	20	25	4413	82	2012-08-02 11:14:04.649856
37385	4	1-11499	37385	25	45	5641	77	2012-08-02 11:14:04.649856
37390	4	1-11242	37389	45	100	282	83	2012-08-02 11:14:04.649856
37391	4	1-11477	37389	100	33	352	13	2012-08-02 11:14:04.649856
37392	4	1-10452	37389	33	25	228	82	2012-08-02 11:14:04.649856
37389	4	1-11499	37389	25	45	300	77	2012-08-02 11:14:04.649856
37394	4	1-8445	37393	6	96	1911	70	2012-08-02 11:14:06.95889
37395	4	1-11183	37393	96	32	1234	9	2012-08-02 11:14:06.95889
37396	4	1-8033	37393	32	47	680	81	2012-08-02 11:14:06.95889
37393	4	1-11504	37393	47	6	1935	25	2012-08-02 11:14:06.95889
37398	4	1-8697	37397	92	15	673	56	2012-08-02 11:14:06.95889
37399	4	1-11133	37397	15	69	2024	47	2012-08-02 11:14:06.95889
37400	4	1-10721	37397	69	47	829	81	2012-08-02 11:14:06.95889
37397	4	1-11504	37397	47	92	2719	25	2012-08-02 11:14:06.95889
37402	3	1-11331	37401	31	11	1591	61	2012-08-02 11:14:06.95889
37403	3	1-5186	37401	11	47	490	81	2012-08-02 11:14:06.95889
37401	3	1-11504	37401	47	31	1699	25	2012-08-02 11:14:06.95889
37421	4	1-6103	37420	70	16	627	67	2012-08-02 11:14:08.881126
37422	4	1-10596	37420	16	40	965	48	2012-08-02 11:14:08.881126
37423	4	1-11113	37420	40	35	1067	70	2012-08-02 11:14:08.881126
37420	4	1-11510	37420	35	70	4083	52	2012-08-02 11:14:08.881126
37425	3	1-9696	37424	72	29	993	93	2012-08-02 11:14:10.041316
37426	3	1-10684	37424	29	50	1252	8	2012-08-02 11:14:10.041316
37424	3	1-11512	37424	50	72	1412	24	2012-08-02 11:14:10.041316
37428	4	1-11374	37427	78	13	532	33	2012-08-02 11:14:10.825701
37429	4	1-11129	37427	13	39	385	49	2012-08-02 11:14:10.825701
37430	4	1-11459	37427	39	33	991	20	2012-08-02 11:14:10.825701
37427	4	1-11515	37427	33	78	2712	97	2012-08-02 11:14:10.825701
37432	4	1-9281	37431	91	16	1784	77	2012-08-02 11:14:10.825701
37433	4	1-8605	37431	16	39	592	49	2012-08-02 11:14:10.825701
37434	4	1-11459	37431	39	33	1760	20	2012-08-02 11:14:10.825701
37431	4	1-11515	37431	33	91	5567	97	2012-08-02 11:14:10.825701
37460	4	1-11384	37459	19	29	3315	52	2012-08-02 11:14:13.157551
37461	4	1-8062	37459	29	39	244	49	2012-08-02 11:14:13.157551
37462	4	1-11459	37459	39	10	640	20	2012-08-02 11:14:13.157551
37459	4	1-11521	37459	10	19	1824	95	2012-08-02 11:14:13.157551
37464	4	1-10827	37463	63	15	109	56	2012-08-02 11:14:13.157551
37465	4	1-11133	37463	15	31	226	47	2012-08-02 11:14:13.157551
37466	4	1-7902	37463	31	10	99	20	2012-08-02 11:14:13.157551
37463	4	1-11521	37463	10	63	289	95	2012-08-02 11:14:13.157551
37468	4	1-11117	37467	8	3	949	75	2012-08-02 11:14:13.157551
37469	4	1-10017	37467	3	39	505	49	2012-08-02 11:14:13.157551
37470	4	1-11459	37467	39	10	1570	20	2012-08-02 11:14:13.157551
37467	4	1-11521	37467	10	8	5311	95	2012-08-02 11:14:13.157551
37472	4	1-10869	37471	43	59	592	4	2012-08-02 11:14:13.157551
37473	4	1-5325	37471	59	2	195	34	2012-08-02 11:14:13.157551
37474	4	1-11106	37471	2	10	125	20	2012-08-02 11:14:13.157551
37471	4	1-11521	37471	10	43	449	95	2012-08-02 11:14:13.157551
37476	4	1-10948	37475	25	31	265	25	2012-08-02 11:14:13.157551
37477	4	1-11331	37475	31	5	210	61	2012-08-02 11:14:13.157551
37478	4	1-4498	37475	5	10	85	20	2012-08-02 11:14:13.157551
37475	4	1-11521	37475	10	25	320	95	2012-08-02 11:14:13.157551
37485	3	1-10290	37484	91	66	1933	32	2012-08-02 11:14:16.615163
37486	3	1-10449	37484	66	79	3160	25	2012-08-02 11:14:16.615163
37484	3	1-11530	37484	79	91	4847	82	2012-08-02 11:14:16.615163
37488	4	1-3656	37487	24	68	469	10	2012-08-02 11:14:16.615163
37489	4	1-10775	37487	68	66	361	32	2012-08-02 11:14:16.615163
37490	4	1-10449	37487	66	79	624	25	2012-08-02 11:14:16.615163
37487	4	1-11530	37487	79	24	1012	82	2012-08-02 11:14:16.615163
37492	3	1-10430	37491	8	50	607	19	2012-08-02 11:14:17.764358
37493	3	1-9770	37491	50	57	372	59	2012-08-02 11:14:17.764358
37491	3	1-11533	37491	57	8	2190	96	2012-08-02 11:14:17.764358
37495	4	1-11429	37494	60	100	2513	57	2012-08-02 11:14:17.764358
37496	4	1-10010	37494	100	92	536	75	2012-08-02 11:14:17.764358
37497	4	1-11479	37494	92	57	586	59	2012-08-02 11:14:17.764358
37494	4	1-11533	37494	57	60	3538	96	2012-08-02 11:14:17.764358
37499	4	1-10088	37498	74	50	676	8	2012-08-02 11:14:17.764358
37500	4	1-11512	37498	50	32	465	24	2012-08-02 11:14:17.764358
37501	4	1-4664	37498	32	57	290	59	2012-08-02 11:14:17.764358
37498	4	1-11533	37498	57	74	1856	96	2012-08-02 11:14:17.764358
37534	4	1-10162	37533	42	73	826	5	2012-08-02 11:14:20.825378
37535	4	1-10949	37533	73	65	754	94	2012-08-02 11:14:20.825378
37536	4	1-9186	37533	65	80	287	91	2012-08-02 11:14:20.825378
37533	4	1-11541	37533	80	42	509	4	2012-08-02 11:14:20.825378
37538	4	1-9843	37537	33	40	2518	42	2012-08-02 11:14:20.825378
37295	4	1-11320	37294	62	95	850	69	2012-08-02 11:13:51.886711
37296	4	1-9542	37294	95	72	643	71	2012-08-02 11:13:51.886711
37297	4	1-11454	37294	72	90	285	78	2012-08-02 11:13:51.886711
37294	4	1-11466	37294	90	62	2623	80	2012-08-02 11:13:51.886711
37299	4	1-9917	37298	80	98	29	73	2012-08-02 11:13:51.886711
37300	4	1-9569	37298	98	47	14	68	2012-08-02 11:13:51.886711
37301	4	1-11300	37298	47	90	15	78	2012-08-02 11:13:51.886711
37298	4	1-11466	37298	90	80	166	80	2012-08-02 11:13:51.886711
37311	4	1-9841	37310	57	52	220	91	2012-08-02 11:13:53.035415
37312	4	1-11323	37310	52	27	90	99	2012-08-02 11:13:53.035415
37313	4	1-8748	37310	27	40	45	98	2012-08-02 11:13:53.035415
37310	4	1-11469	37310	40	57	308	33	2012-08-02 11:13:53.035415
37315	3	1-11129	37314	13	16	4363	49	2012-08-02 11:13:53.035415
37316	3	1-6632	37314	16	40	749	98	2012-08-02 11:13:53.035415
37314	3	1-11469	37314	40	13	6424	33	2012-08-02 11:13:53.035415
37318	4	1-10886	37317	67	10	2594	23	2012-08-02 11:13:53.035415
37319	4	1-10479	37317	10	63	1690	66	2012-08-02 11:13:53.035415
37320	4	1-6484	37317	63	40	326	98	2012-08-02 11:13:53.035415
37317	4	1-11469	37317	40	67	2851	33	2012-08-02 11:13:53.035415
37327	3	1-11166	37326	22	42	5161	42	2012-08-02 11:13:54.195816
37328	3	1-9649	37326	42	5	1363	40	2012-08-02 11:13:54.195816
37326	3	1-11472	37326	5	22	3771	63	2012-08-02 11:13:54.195816
37332	4	1-11347	37331	20	24	1035	18	2012-08-02 11:13:55.356002
37333	4	1-11427	37331	24	27	1671	2	2012-08-02 11:13:55.356002
37334	4	1-11448	37331	27	57	804	12	2012-08-02 11:13:55.356002
37331	4	1-11475	37331	57	20	1744	8	2012-08-02 11:13:55.356002
37336	4	1-11061	37335	52	59	60	43	2012-08-02 11:13:55.356002
37337	4	1-10383	37335	59	78	40	90	2012-08-02 11:13:55.356002
37338	4	1-10564	37335	78	57	24	12	2012-08-02 11:13:55.356002
37335	4	1-11475	37335	57	52	59	8	2012-08-02 11:13:55.356002
37340	4	1-11004	37339	37	33	1553	75	2012-08-02 11:13:55.356002
37341	4	1-7646	37339	33	27	1522	2	2012-08-02 11:13:55.356002
37342	4	1-11448	37339	27	57	938	12	2012-08-02 11:13:55.356002
37339	4	1-11475	37339	57	37	2607	8	2012-08-02 11:13:55.356002
37344	3	1-1806	37343	82	32	1147	92	2012-08-02 11:13:55.356002
37345	3	1-11443	37343	32	57	1312	12	2012-08-02 11:13:55.356002
37343	3	1-11475	37343	57	82	3800	8	2012-08-02 11:13:55.356002
37373	4	1-11070	37372	53	86	3772	39	2012-08-02 11:14:01.887571
37374	4	1-10944	37372	86	22	2515	1	2012-08-02 11:14:01.887571
37375	4	1-10763	37372	22	88	1141	24	2012-08-02 11:14:01.887571
37372	4	1-11492	37372	88	53	6169	65	2012-08-02 11:14:01.887571
37377	4	1-11287	37376	13	54	1058	55	2012-08-02 11:14:03.080588
37378	4	1-7475	37376	54	54	513	56	2012-08-02 11:14:03.080588
37379	4	1-11213	37376	54	69	257	24	2012-08-02 11:14:03.080588
37376	4	1-11495	37376	69	13	1865	98	2012-08-02 11:14:03.080588
37381	5	1-761	37380	30	47	563	68	2012-08-02 11:14:03.080588
37382	5	1-11300	37380	47	56	595	78	2012-08-02 11:14:03.080588
37383	5	1-11436	37380	56	20	500	63	2012-08-02 11:14:03.080588
37384	5	1-10375	37380	20	69	357	24	2012-08-02 11:14:03.080588
37380	5	1-11495	37380	69	30	2616	98	2012-08-02 11:14:03.080588
37413	4	1-11404	37412	100	79	2423	77	2012-08-02 11:14:08.107531
37414	4	1-10758	37412	79	42	787	84	2012-08-02 11:14:08.107531
37415	4	1-10716	37412	42	93	766	29	2012-08-02 11:14:08.107531
37412	4	1-11508	37412	93	100	1141	23	2012-08-02 11:14:08.107531
37417	4	1-10357	37416	78	48	3176	40	2012-08-02 11:14:08.107531
37418	4	1-11497	37416	48	34	2116	65	2012-08-02 11:14:08.107531
37419	4	1-9432	37416	34	93	1071	29	2012-08-02 11:14:08.107531
37416	4	1-11508	37416	93	78	1607	23	2012-08-02 11:14:08.107531
37446	3	1-10758	37445	79	64	1232	84	2012-08-02 11:14:11.587941
37447	3	1-11170	37445	64	48	2530	42	2012-08-02 11:14:11.587941
37445	3	1-11517	37445	48	79	4900	77	2012-08-02 11:14:11.587941
37449	4	1-11242	37448	45	72	569	83	2012-08-02 11:14:11.587941
37450	4	1-8437	37448	72	72	505	14	2012-08-02 11:14:11.587941
37451	4	1-9586	37448	72	48	411	42	2012-08-02 11:14:11.587941
37448	4	1-11517	37448	48	45	814	77	2012-08-02 11:14:11.587941
37453	3	1-6714	37452	8	64	357	84	2012-08-02 11:14:11.587941
37454	3	1-11170	37452	64	48	843	42	2012-08-02 11:14:11.587941
37452	3	1-11517	37452	48	8	1876	77	2012-08-02 11:14:11.587941
37480	5	1-11326	37479	34	86	1677	22	2012-08-02 11:14:13.930346
37481	5	1-10828	37479	86	20	1225	76	2012-08-02 11:14:13.930346
37482	5	1-5024	37479	20	78	296	33	2012-08-02 11:14:13.930346
37483	5	1-10279	37479	78	94	449	39	2012-08-02 11:14:13.930346
37479	5	1-11523	37479	94	34	1804	46	2012-08-02 11:14:13.930346
37503	4	1-11005	37502	61	41	2750	41	2012-08-02 11:14:18.515468
37504	4	1-11078	37502	41	59	1920	43	2012-08-02 11:14:18.515468
37505	4	1-10383	37502	59	13	1570	90	2012-08-02 11:14:18.515468
37502	4	1-11536	37502	13	61	4125	71	2012-08-02 11:14:18.515468
37507	3	1-10068	37506	50	54	1335	59	2012-08-02 11:14:18.515468
37508	3	1-11314	37506	54	13	1157	90	2012-08-02 11:14:18.515468
37506	3	1-11536	37506	13	50	3049	71	2012-08-02 11:14:18.515468
37510	4	1-9930	37509	14	84	220	44	2012-08-02 11:14:18.515468
37511	4	1-11272	37509	84	63	157	13	2012-08-02 11:14:18.515468
37512	4	1-8679	37509	63	13	67	90	2012-08-02 11:14:18.515468
37509	4	1-11536	37509	13	14	181	71	2012-08-02 11:14:18.515468
37514	4	1-7140	37513	56	37	438	100	2012-08-02 11:14:18.515468
37515	4	1-10894	37513	37	54	276	59	2012-08-02 11:14:18.515468
37516	4	1-11314	37513	54	13	298	90	2012-08-02 11:14:18.515468
37513	4	1-11536	37513	13	56	976	71	2012-08-02 11:14:18.515468
37571	4	1-11209	37570	89	71	1652	30	2012-08-02 11:14:23.511993
37572	4	1-11158	37570	71	6	1823	54	2012-08-02 11:14:23.511993
37573	4	1-11068	37570	6	64	2642	87	2012-08-02 11:14:23.511993
37570	4	1-11548	37570	64	89	2081	59	2012-08-02 11:14:23.511993
37575	4	1-11386	37574	36	16	1415	7	2012-08-02 11:14:24.672122
37576	4	1-10685	37574	16	59	834	91	2012-08-02 11:14:24.672122
37577	4	1-9374	37574	59	22	1038	52	2012-08-02 11:14:24.672122
37574	4	1-11551	37574	22	36	730	48	2012-08-02 11:14:24.672122
37595	4	1-5302	37594	51	21	1508	58	2012-08-02 11:14:29.246334
37596	4	1-11435	37594	21	22	1308	38	2012-08-02 11:14:29.246334
37597	4	1-11540	37594	22	43	1158	51	2012-08-02 11:14:29.246334
37594	4	1-11563	37594	43	51	1944	41	2012-08-02 11:14:29.246334
37599	4	1-10230	37598	95	88	4030	12	2012-08-02 11:14:29.246334
37600	4	1-10608	37598	88	30	2987	13	2012-08-02 11:14:29.246334
37601	4	1-11198	37598	30	43	2426	51	2012-08-02 11:14:29.246334
37598	4	1-11563	37598	43	95	4490	41	2012-08-02 11:14:29.246334
37539	4	1-10494	37537	40	7	909	17	2012-08-02 11:14:20.825378
37540	4	1-11481	37537	7	80	1225	91	2012-08-02 11:14:20.825378
37537	4	1-11541	37537	80	33	2359	4	2012-08-02 11:14:20.825378
37542	4	1-9843	37541	33	79	2448	42	2012-08-02 11:14:20.825378
37543	4	1-5974	37541	79	7	824	17	2012-08-02 11:14:20.825378
37544	4	1-11481	37541	7	80	1137	91	2012-08-02 11:14:20.825378
37541	4	1-11541	37541	80	33	2241	4	2012-08-02 11:14:20.825378
37546	2	1-5228	37545	22	55	1258	1	2012-08-02 11:14:21.985338
37545	2	1-11544	37545	55	22	5178	61	2012-08-02 11:14:21.985338
37548	4	1-6531	37547	86	6	1324	73	2012-08-02 11:14:21.985338
37549	4	1-11203	37547	6	43	1069	9	2012-08-02 11:14:21.985338
37550	4	1-11352	37547	43	55	624	1	2012-08-02 11:14:21.985338
37547	4	1-11544	37547	55	86	3274	61	2012-08-02 11:14:21.985338
37563	4	1-10985	37562	65	5	461	65	2012-08-02 11:14:23.136105
37564	4	1-11149	37562	5	78	584	97	2012-08-02 11:14:23.136105
37565	4	1-11374	37562	78	84	167	33	2012-08-02 11:14:23.136105
37562	4	1-11547	37562	84	65	1227	20	2012-08-02 11:14:23.136105
37567	4	1-10746	37566	52	84	1841	56	2012-08-02 11:14:23.136105
37568	4	1-10529	37566	84	78	2015	97	2012-08-02 11:14:23.136105
37569	4	1-11374	37566	78	84	642	33	2012-08-02 11:14:23.136105
37566	4	1-11547	37566	84	52	5304	20	2012-08-02 11:14:23.136105
37583	4	1-10948	37582	25	25	4102	25	2012-08-02 11:14:26.561211
37584	4	1-11281	37582	25	37	5379	16	2012-08-02 11:14:26.561211
37585	4	1-11444	37582	37	60	3969	88	2012-08-02 11:14:26.561211
37582	4	1-11556	37582	60	25	4452	95	2012-08-02 11:14:26.561211
37587	4	1-11006	37586	20	76	2369	87	2012-08-02 11:14:27.743563
37588	4	1-11546	37586	76	79	3412	14	2012-08-02 11:14:27.743563
37589	4	1-10876	37586	79	97	1836	92	2012-08-02 11:14:27.743563
37586	4	1-11559	37586	97	20	5077	19	2012-08-02 11:14:27.743563
37591	4	1-11006	37590	20	76	151	87	2012-08-02 11:14:27.743563
37592	4	1-11546	37590	76	13	220	14	2012-08-02 11:14:27.743563
37593	4	1-9040	37590	13	97	115	92	2012-08-02 11:14:27.743563
37590	4	1-11559	37590	97	20	321	19	2012-08-02 11:14:27.743563
37603	4	1-11013	37602	64	31	207	93	2012-08-02 11:14:30.008411
37604	4	1-10516	37602	31	10	140	64	2012-08-02 11:14:30.008411
37605	4	1-10631	37602	10	48	159	67	2012-08-02 11:14:30.008411
37602	4	1-11565	37602	48	64	191	21	2012-08-02 11:14:30.008411
37614	4	1-6889	37613	17	42	2164	40	2012-08-02 11:14:31.158049
37615	4	1-11534	37613	42	20	1603	22	2012-08-02 11:14:31.158049
37616	4	1-9950	37613	20	68	919	84	2012-08-02 11:14:31.158049
37613	4	1-11568	37613	68	17	3510	83	2012-08-02 11:14:31.158049
37618	4	1-11292	37617	9	68	1572	80	2012-08-02 11:14:34.626495
37619	4	1-11285	37617	68	93	1934	45	2012-08-02 11:14:34.626495
37620	4	1-9382	37617	93	35	1479	6	2012-08-02 11:14:34.626495
37617	4	1-11577	37617	35	9	1631	68	2012-08-02 11:14:34.626495
37630	4	1-11313	37629	30	69	1440	61	2012-08-02 11:14:35.77558
37631	4	1-4196	37629	69	29	400	93	2012-08-02 11:14:35.77558
37632	4	1-10684	37629	29	1	498	8	2012-08-02 11:14:35.77558
37629	4	1-11580	37629	1	30	1486	71	2012-08-02 11:14:35.77558
37634	4	1-10548	37633	59	36	306	47	2012-08-02 11:14:35.77558
37635	4	1-10490	37633	36	29	133	93	2012-08-02 11:14:35.77558
37636	4	1-10684	37633	29	1	171	8	2012-08-02 11:14:35.77558
37633	4	1-11580	37633	1	59	525	71	2012-08-02 11:14:35.77558
37638	3	1-11005	37637	61	86	2873	41	2012-08-02 11:14:35.77558
37639	3	1-5916	37637	86	1	1012	8	2012-08-02 11:14:35.77558
37637	3	1-11580	37637	1	61	3145	71	2012-08-02 11:14:35.77558
37641	4	1-11199	37640	87	34	2966	47	2012-08-02 11:14:39.222523
37642	4	1-5628	37640	34	23	2362	37	2012-08-02 11:14:39.222523
37643	4	1-11329	37640	23	48	2260	92	2012-08-02 11:14:39.222523
37640	4	1-11589	37640	48	87	1771	49	2012-08-02 11:14:39.222523
37660	4	1-11115	37659	43	46	3411	16	2012-08-02 11:14:41.498714
37661	4	1-11377	37659	46	69	3673	94	2012-08-02 11:14:41.498714
37662	4	1-11029	37659	69	83	2758	27	2012-08-02 11:14:41.498714
37659	4	1-11595	37659	83	43	2120	30	2012-08-02 11:14:41.498714
37705	2	1-6628	37704	7	25	1333	60	2012-08-02 11:14:46.891545
37704	2	1-11609	37704	25	7	2132	39	2012-08-02 11:14:46.891545
37707	4	1-11523	37706	94	34	282	46	2012-08-02 11:14:46.891545
37708	4	1-11326	37706	34	22	156	22	2012-08-02 11:14:46.891545
37709	4	1-9664	37706	22	25	71	60	2012-08-02 11:14:46.891545
37706	4	1-11609	37706	25	94	118	39	2012-08-02 11:14:46.891545
37711	4	1-8721	37710	98	73	598	14	2012-08-02 11:14:46.891545
37712	4	1-11097	37710	73	39	300	86	2012-08-02 11:14:46.891545
37713	4	1-10331	37710	39	25	410	60	2012-08-02 11:14:46.891545
37710	4	1-11609	37710	25	98	712	39	2012-08-02 11:14:46.891545
37740	4	1-10211	37739	63	90	871	44	2012-08-02 11:14:51.556199
37741	4	1-9369	37739	90	93	521	25	2012-08-02 11:14:51.556199
37742	4	1-10386	37739	93	96	229	96	2012-08-02 11:14:51.556199
37739	4	1-11621	37739	96	63	836	87	2012-08-02 11:14:51.556199
37778	4	1-10956	37777	32	67	140	72	2012-08-02 11:14:55.047636
37779	4	1-11269	37777	67	31	205	95	2012-08-02 11:14:55.047636
37780	4	1-10272	37777	31	51	252	87	2012-08-02 11:14:55.047636
37777	4	1-11629	37777	51	32	253	71	2012-08-02 11:14:55.047636
37782	4	1-11313	37781	30	5	2857	61	2012-08-02 11:14:55.047636
37783	4	1-8245	37781	5	33	1208	64	2012-08-02 11:14:55.047636
37784	4	1-10838	37781	33	51	2529	87	2012-08-02 11:14:55.047636
37781	4	1-11629	37781	51	30	2747	71	2012-08-02 11:14:55.047636
37802	3	1-6975	37801	64	32	1528	92	2012-08-02 11:14:55.820897
37803	3	1-11443	37801	32	64	1281	12	2012-08-02 11:14:55.820897
37801	3	1-11632	37801	64	64	1581	91	2012-08-02 11:14:55.820897
37805	4	1-11265	37804	10	23	2357	37	2012-08-02 11:14:55.820897
37806	4	1-11329	37804	23	32	1602	92	2012-08-02 11:14:55.820897
37807	4	1-11443	37804	32	64	1963	12	2012-08-02 11:14:55.820897
37804	4	1-11632	37804	64	10	2456	91	2012-08-02 11:14:55.820897
37809	4	1-11323	37808	52	99	421	99	2012-08-02 11:14:55.820897
37810	4	1-6272	37808	99	32	421	92	2012-08-02 11:14:55.820897
37811	4	1-11443	37808	32	64	520	12	2012-08-02 11:14:55.820897
37808	4	1-11632	37808	64	52	657	91	2012-08-02 11:14:55.820897
37825	4	1-11632	37824	64	32	689	91	2012-08-02 11:14:59.665625
37826	4	1-8504	37824	32	47	318	68	2012-08-02 11:14:59.665625
37827	4	1-7215	37824	47	36	185	31	2012-08-02 11:14:59.665625
37824	4	1-11643	37824	36	64	406	12	2012-08-02 11:14:59.665625
37829	4	1-11024	37828	76	27	909	34	2012-08-02 11:15:01.245814
37830	4	1-8971	37828	27	47	682	16	2012-08-02 11:15:01.245814
37831	4	1-9360	37828	47	96	322	31	2012-08-02 11:15:01.245814
37579	4	1-11493	37578	77	83	3268	10	2012-08-02 11:14:26.174917
37580	4	1-10884	37578	83	19	2838	43	2012-08-02 11:14:26.174917
37581	4	1-10503	37578	19	88	2429	50	2012-08-02 11:14:26.174917
37578	4	1-11555	37578	88	77	3302	55	2012-08-02 11:14:26.174917
37607	3	1-6224	37606	21	27	1935	38	2012-08-02 11:14:30.77091
37608	3	1-11464	37606	27	95	233	56	2012-08-02 11:14:30.77091
37606	3	1-11566	37606	95	21	7598	67	2012-08-02 11:14:30.77091
37610	4	1-11402	37609	85	67	212	51	2012-08-02 11:14:30.77091
37611	4	1-11105	37609	67	100	123	36	2012-08-02 11:14:30.77091
37612	4	1-5084	37609	100	95	15	56	2012-08-02 11:14:30.77091
37609	4	1-11566	37609	95	85	522	67	2012-08-02 11:14:30.77091
37645	4	1-11430	37644	34	83	3048	45	2012-08-02 11:14:39.598138
37646	4	1-4729	37644	83	38	2338	53	2012-08-02 11:14:39.598138
37647	4	1-10095	37644	38	33	2051	37	2012-08-02 11:14:39.598138
37644	4	1-11590	37644	33	34	2729	10	2012-08-02 11:14:39.598138
37649	3	1-11430	37648	34	51	201	45	2012-08-02 11:14:39.598138
37650	3	1-6013	37648	51	33	134	37	2012-08-02 11:14:39.598138
37648	3	1-11590	37648	33	34	179	10	2012-08-02 11:14:39.598138
37652	3	1-11297	37651	52	41	3220	16	2012-08-02 11:14:40.747269
37653	3	1-9997	37651	41	95	1460	35	2012-08-02 11:14:40.747269
37651	3	1-11593	37651	95	52	3812	31	2012-08-02 11:14:40.747269
37655	4	1-11031	37654	93	32	1078	48	2012-08-02 11:14:40.747269
37656	4	1-10862	37654	32	24	1055	52	2012-08-02 11:14:40.747269
37657	4	1-11279	37654	24	95	632	35	2012-08-02 11:14:40.747269
37654	4	1-11593	37654	95	93	1838	31	2012-08-02 11:14:40.747269
37658	1	1-2950	\N	47	47	410	35	2012-08-02 11:14:40.747269
37664	4	1-11589	37663	48	87	1609	49	2012-08-02 11:14:41.8964
37665	4	1-11199	37663	87	51	2641	47	2012-08-02 11:14:41.8964
37666	4	1-10355	37663	51	72	1592	6	2012-08-02 11:14:41.8964
37663	4	1-11596	37663	72	48	2094	92	2012-08-02 11:14:41.8964
37672	3	1-11062	37671	55	40	666	48	2012-08-02 11:14:43.034382
37673	3	1-11113	37671	40	6	430	70	2012-08-02 11:14:43.034382
37671	3	1-11598	37671	6	55	1564	93	2012-08-02 11:14:43.034382
37675	4	1-10516	37674	31	10	2577	64	2012-08-02 11:14:43.034382
37676	4	1-10631	37674	10	7	2062	67	2012-08-02 11:14:43.034382
37677	4	1-8529	37674	7	6	1275	70	2012-08-02 11:14:43.034382
37674	4	1-11598	37674	6	31	5407	93	2012-08-02 11:14:43.034382
37679	4	1-9763	37678	11	65	440	73	2012-08-02 11:14:43.034382
37680	4	1-11344	37678	65	40	300	48	2012-08-02 11:14:43.034382
37681	4	1-11113	37678	40	6	230	70	2012-08-02 11:14:43.034382
37678	4	1-11598	37678	6	11	992	93	2012-08-02 11:14:43.034382
37682	1	1-1180	\N	9	9	1919	48	2012-08-02 11:14:43.034382
37694	3	1-10596	37693	16	74	4804	48	2012-08-02 11:14:44.957993
37695	3	1-11581	37693	74	47	3278	42	2012-08-02 11:14:44.957993
37693	3	1-11604	37693	47	16	3479	67	2012-08-02 11:14:44.957993
37715	4	1-5772	37714	23	48	1396	40	2012-08-02 11:14:47.289697
37716	4	1-11497	37714	48	10	766	65	2012-08-02 11:14:47.289697
37717	4	1-11088	37714	10	47	699	17	2012-08-02 11:14:47.289697
37714	4	1-11610	37714	47	23	1437	99	2012-08-02 11:14:47.289697
37719	4	1-8305	37718	48	73	308	5	2012-08-02 11:14:47.289697
37720	4	1-10949	37718	73	90	267	94	2012-08-02 11:14:47.289697
37721	4	1-8836	37718	90	47	125	17	2012-08-02 11:14:47.289697
37718	4	1-11610	37718	47	48	290	99	2012-08-02 11:14:47.289697
37723	4	1-5863	37722	72	93	3053	90	2012-08-02 11:14:47.289697
37724	4	1-10463	37722	93	87	2706	28	2012-08-02 11:14:47.289697
37725	4	1-10930	37722	87	47	2218	17	2012-08-02 11:14:47.289697
37722	4	1-11610	37722	47	72	5406	99	2012-08-02 11:14:47.289697
37727	4	1-10416	37726	83	30	1883	71	2012-08-02 11:14:47.289697
37728	4	1-11313	37726	30	35	1779	61	2012-08-02 11:14:47.289697
37729	4	1-2632	37726	35	47	1070	17	2012-08-02 11:14:47.289697
37726	4	1-11610	37726	47	83	2669	99	2012-08-02 11:14:47.289697
37730	1	1-2632	\N	35	35	3891	17	2012-08-02 11:14:47.289697
37737	3	1-7286	37736	67	9	919	68	2012-08-02 11:14:49.633653
37738	3	1-11292	37736	9	37	520	80	2012-08-02 11:14:49.633653
37736	3	1-11617	37736	37	67	1479	99	2012-08-02 11:14:49.633653
37759	4	1-11364	37758	63	32	267	9	2012-08-02 11:14:53.478813
37760	4	1-8033	37758	32	80	137	81	2012-08-02 11:14:53.478813
37761	4	1-9581	37758	80	47	134	58	2012-08-02 11:14:53.478813
37758	4	1-11626	37758	47	63	118	19	2012-08-02 11:14:53.478813
37763	4	1-10330	37762	1	64	401	41	2012-08-02 11:14:53.478813
37764	4	1-6945	37762	64	80	392	81	2012-08-02 11:14:53.478813
37765	4	1-9581	37762	80	47	460	58	2012-08-02 11:14:53.478813
37762	4	1-11626	37762	47	1	490	19	2012-08-02 11:14:53.478813
37767	3	1-9770	37766	50	8	1773	59	2012-08-02 11:14:53.478813
37768	3	1-10569	37766	8	47	1496	58	2012-08-02 11:14:53.478813
37766	3	1-11626	37766	47	50	1645	19	2012-08-02 11:14:53.478813
37770	4	1-11549	37769	62	47	4016	42	2012-08-02 11:14:53.478813
37771	4	1-11604	37769	47	6	4564	67	2012-08-02 11:14:53.478813
37772	4	1-8921	37769	6	47	4515	58	2012-08-02 11:14:53.478813
37769	4	1-11626	37769	47	62	5018	19	2012-08-02 11:14:53.478813
37813	4	1-11269	37812	67	25	2466	95	2012-08-02 11:14:59.290105
37814	4	1-10948	37812	25	89	2127	25	2012-08-02 11:14:59.290105
37815	4	1-6176	37812	89	51	537	22	2012-08-02 11:14:59.290105
37812	4	1-11641	37812	51	67	1793	72	2012-08-02 11:14:59.290105
37817	4	1-9306	37816	94	91	2256	6	2012-08-02 11:14:59.290105
37818	4	1-7374	37816	91	42	1960	40	2012-08-02 11:14:59.290105
37819	4	1-11534	37816	42	51	1521	22	2012-08-02 11:14:59.290105
37816	4	1-11641	37816	51	94	5299	72	2012-08-02 11:14:59.290105
37821	4	1-3727	37820	45	75	1174	5	2012-08-02 11:14:59.290105
37822	4	1-9198	37820	75	42	923	40	2012-08-02 11:14:59.290105
37823	4	1-11534	37820	42	51	750	22	2012-08-02 11:14:59.290105
37820	4	1-11641	37820	51	45	2736	72	2012-08-02 11:14:59.290105
37863	4	1-8748	37862	27	54	707	98	2012-08-02 11:15:02.79246
37864	4	1-10403	37862	54	31	644	25	2012-08-02 11:15:02.79246
37865	4	1-11331	37862	31	100	611	61	2012-08-02 11:15:02.79246
37862	4	1-11650	37862	100	27	643	99	2012-08-02 11:15:02.79246
37890	3	1-10992	37889	89	43	4296	30	2012-08-02 11:15:05.113014
37891	3	1-11115	37889	43	88	4946	16	2012-08-02 11:15:05.113014
37889	3	1-11656	37889	88	89	6097	90	2012-08-02 11:15:05.113014
37897	3	1-11300	37896	47	27	7021	78	2012-08-02 11:15:06.297318
37898	3	1-10550	37896	27	15	5521	13	2012-08-02 11:15:06.297318
37896	3	1-11659	37896	15	47	7822	68	2012-08-02 11:15:06.297318
37900	4	1-9616	37899	94	89	1907	77	2012-08-02 11:15:08.648528
37901	4	1-6860	37899	89	88	834	48	2012-08-02 11:15:08.648528
37899	4	1-11665	37899	61	94	1952	62	2012-08-02 11:15:08.648528
37622	4	1-9055	37621	14	21	1074	23	2012-08-02 11:14:35.399738
37623	4	1-8933	37621	21	73	1562	5	2012-08-02 11:14:35.399738
37624	4	1-10949	37621	73	58	1523	94	2012-08-02 11:14:35.399738
37621	4	1-11579	37621	58	14	1028	63	2012-08-02 11:14:35.399738
37626	4	1-9138	37625	82	48	1142	55	2012-08-02 11:14:35.399738
37627	4	1-9499	37625	48	46	1192	16	2012-08-02 11:14:35.399738
37628	4	1-11377	37625	46	58	1277	94	2012-08-02 11:14:35.399738
37625	4	1-11579	37625	58	82	897	63	2012-08-02 11:14:35.399738
37668	4	1-11284	37667	73	89	5093	77	2012-08-02 11:14:42.271923
37669	4	1-6860	37667	89	32	1415	48	2012-08-02 11:14:42.271923
37670	4	1-10862	37667	32	36	856	52	2012-08-02 11:14:42.271923
37667	4	1-11597	37667	36	73	5080	24	2012-08-02 11:14:42.271923
37684	2	1-9068	37683	42	94	1710	85	2012-08-02 11:14:44.218345
37683	2	1-11602	37683	94	42	2037	59	2012-08-02 11:14:44.218345
37686	4	1-10762	37685	78	63	338	57	2012-08-02 11:14:44.218345
37687	4	1-10870	37685	63	51	592	83	2012-08-02 11:14:44.218345
37688	4	1-7713	37685	51	94	283	85	2012-08-02 11:14:44.218345
37685	4	1-11602	37685	94	78	513	59	2012-08-02 11:14:44.218345
37690	4	1-11314	37689	54	1	4980	90	2012-08-02 11:14:44.218345
37691	4	1-9603	37689	1	14	3735	17	2012-08-02 11:14:44.218345
37692	4	1-11190	37689	14	94	2335	85	2012-08-02 11:14:44.218345
37689	4	1-11602	37689	94	54	4298	59	2012-08-02 11:14:44.218345
37697	4	1-11402	37696	85	67	4357	51	2012-08-02 11:14:45.344938
37698	4	1-11105	37696	67	26	4459	36	2012-08-02 11:14:45.344938
37699	4	1-9050	37696	26	73	3465	77	2012-08-02 11:14:45.344938
37696	4	1-11605	37696	73	85	6064	67	2012-08-02 11:14:45.344938
37701	4	1-11402	37700	85	28	198	51	2012-08-02 11:14:45.344938
37702	4	1-11573	37700	28	86	278	91	2012-08-02 11:14:45.344938
37703	4	1-9340	37700	86	73	123	77	2012-08-02 11:14:45.344938
37700	4	1-11605	37700	73	85	244	67	2012-08-02 11:14:45.344938
37732	5	1-10505	37731	67	33	1849	80	2012-08-02 11:14:48.849192
37733	5	1-11468	37731	33	56	1867	4	2012-08-02 11:14:48.849192
37734	5	1-11608	37731	56	53	1681	65	2012-08-02 11:14:48.849192
37735	5	1-11070	37731	53	41	1425	39	2012-08-02 11:14:48.849192
37731	5	1-11614	37731	41	67	1541	89	2012-08-02 11:14:48.849192
37744	3	1-11237	37743	13	5	2297	61	2012-08-02 11:14:51.954176
37745	3	1-8245	37743	5	32	672	64	2012-08-02 11:14:51.954176
37743	3	1-11622	37743	32	13	2385	3	2012-08-02 11:14:51.954176
37747	4	1-11516	37746	29	22	2277	75	2012-08-02 11:14:51.954176
37748	4	1-11437	37746	22	51	2518	32	2012-08-02 11:14:51.954176
37749	4	1-10706	37746	51	32	855	64	2012-08-02 11:14:51.954176
37746	4	1-11622	37746	32	29	3111	3	2012-08-02 11:14:51.954176
37751	4	1-8889	37750	90	29	970	93	2012-08-02 11:14:51.954176
37752	4	1-10684	37750	29	80	933	8	2012-08-02 11:14:51.954176
37753	4	1-8484	37750	80	32	412	64	2012-08-02 11:14:51.954176
37750	4	1-11622	37750	32	90	1520	3	2012-08-02 11:14:51.954176
37755	4	1-8969	37754	5	93	463	90	2012-08-02 11:14:53.091797
37756	4	1-10463	37754	93	68	336	28	2012-08-02 11:14:53.091797
37757	4	1-10480	37754	68	71	476	25	2012-08-02 11:14:53.091797
37754	4	1-11625	37754	71	5	2325	45	2012-08-02 11:14:53.091797
37774	4	1-11367	37773	43	94	2448	76	2012-08-02 11:14:54.262813
37775	4	1-10064	37773	94	1	852	8	2012-08-02 11:14:54.262813
37776	4	1-11580	37773	1	94	2667	71	2012-08-02 11:14:54.262813
37773	4	1-11628	37773	94	43	2626	80	2012-08-02 11:14:54.262813
37786	4	1-9245	37785	47	10	437	72	2012-08-02 11:14:55.42302
37787	4	1-11181	37785	10	15	171	56	2012-08-02 11:14:55.42302
37788	4	1-11133	37785	15	31	362	47	2012-08-02 11:14:55.42302
37785	4	1-11631	37785	31	47	678	84	2012-08-02 11:14:55.42302
37790	4	1-11211	37789	13	66	1492	32	2012-08-02 11:14:55.42302
37791	4	1-10449	37789	66	29	2226	25	2012-08-02 11:14:55.42302
37792	4	1-9064	37789	29	31	652	47	2012-08-02 11:14:55.42302
37789	4	1-11631	37789	31	13	1447	84	2012-08-02 11:14:55.42302
37794	4	1-11255	37793	59	76	4868	22	2012-08-02 11:14:55.42302
37795	4	1-2964	37793	76	15	605	56	2012-08-02 11:14:55.42302
37796	4	1-11133	37793	15	31	1726	47	2012-08-02 11:14:55.42302
37793	4	1-11631	37793	31	59	4367	84	2012-08-02 11:14:55.42302
37798	4	1-2625	37797	5	96	18	11	2012-08-02 11:14:55.42302
37799	4	1-10849	37797	96	15	13	56	2012-08-02 11:14:55.42302
37800	4	1-11133	37797	15	31	40	47	2012-08-02 11:14:55.42302
37797	4	1-11631	37797	31	5	109	84	2012-08-02 11:14:55.42302
37835	3	1-11269	37834	67	46	918	95	2012-08-02 11:15:00.836972
37836	3	1-9308	37834	46	5	550	60	2012-08-02 11:15:00.836972
37834	3	1-11646	37834	5	67	736	72	2012-08-02 11:15:00.836972
37838	4	1-10273	37837	18	18	3435	65	2012-08-02 11:15:00.836972
37839	4	1-5668	37837	18	39	2655	10	2012-08-02 11:15:00.836972
37840	4	1-9633	37837	39	5	2701	60	2012-08-02 11:15:00.836972
37837	4	1-11646	37837	5	18	4292	72	2012-08-02 11:15:00.836972
37858	4	1-11595	37857	83	18	1145	30	2012-08-02 11:15:02.428143
37859	4	1-4265	37857	18	59	668	100	2012-08-02 11:15:02.428143
37860	4	1-11453	37857	59	14	743	50	2012-08-02 11:15:02.428143
37857	4	1-11649	37857	14	83	1747	27	2012-08-02 11:15:02.428143
37861	1	1-4265	\N	18	18	3350	100	2012-08-02 11:15:02.428143
37867	4	1-7365	37866	40	69	224	69	2012-08-02 11:15:03.566223
37868	4	1-10805	37866	69	47	116	58	2012-08-02 11:15:03.566223
37869	4	1-11626	37866	47	19	131	19	2012-08-02 11:15:03.566223
37866	4	1-11652	37866	19	40	230	26	2012-08-02 11:15:03.566223
37875	4	1-6191	37874	17	68	135	60	2012-08-02 11:15:04.737407
37876	4	1-11381	37874	68	51	124	89	2012-08-02 11:15:04.737407
37877	4	1-11291	37874	51	90	100	76	2012-08-02 11:15:04.737407
37874	4	1-11655	37874	90	17	148	8	2012-08-02 11:15:04.737407
37879	4	1-11654	37878	66	67	3164	63	2012-08-02 11:15:04.737407
37880	4	1-8856	37878	67	51	2642	89	2012-08-02 11:15:04.737407
37881	4	1-11291	37878	51	90	2202	76	2012-08-02 11:15:04.737407
37878	4	1-11655	37878	90	66	3397	8	2012-08-02 11:15:04.737407
37883	3	1-11229	37882	67	86	3759	22	2012-08-02 11:15:04.737407
37884	3	1-10828	37882	86	90	2265	76	2012-08-02 11:15:04.737407
37882	3	1-11655	37882	90	67	3664	8	2012-08-02 11:15:04.737407
37886	4	1-11654	37885	66	14	2155	63	2012-08-02 11:15:04.737407
37887	4	1-9055	37885	14	10	1961	23	2012-08-02 11:15:04.737407
37888	4	1-9868	37885	10	90	1330	76	2012-08-02 11:15:04.737407
37885	4	1-11655	37885	90	66	2179	8	2012-08-02 11:15:04.737407
37907	3	1-11667	37907	96	80	3319	24	2012-08-02 11:15:09.432839
37828	4	1-11645	37828	96	76	1958	82	2012-08-02 11:15:01.245814
37833	2	1-6173	37832	70	96	980	31	2012-08-02 11:15:01.245814
37832	2	1-11645	37832	96	70	6045	82	2012-08-02 11:15:01.245814
37842	4	1-10930	37841	87	82	2923	17	2012-08-02 11:15:02.041455
37843	4	1-11130	37841	82	86	2084	18	2012-08-02 11:15:02.041455
37844	4	1-11489	37841	86	34	1831	90	2012-08-02 11:15:02.041455
37841	4	1-11648	37841	34	87	3356	28	2012-08-02 11:15:02.041455
37846	4	1-9303	37845	66	85	225	11	2012-08-02 11:15:02.041455
37847	4	1-9332	37845	85	79	202	3	2012-08-02 11:15:02.041455
37848	4	1-10132	37845	79	34	234	90	2012-08-02 11:15:02.041455
37845	4	1-11648	37845	34	66	435	28	2012-08-02 11:15:02.041455
37850	3	1-10175	37849	26	86	2073	18	2012-08-02 11:15:02.041455
37851	3	1-11489	37849	86	34	1867	90	2012-08-02 11:15:02.041455
37849	3	1-11648	37849	34	26	3507	28	2012-08-02 11:15:02.041455
37853	4	1-10774	37852	92	34	1157	9	2012-08-02 11:15:02.041455
37854	4	1-8987	37852	34	54	774	59	2012-08-02 11:15:02.041455
37855	4	1-11314	37852	54	34	881	90	2012-08-02 11:15:02.041455
37852	4	1-11648	37852	34	92	1663	28	2012-08-02 11:15:02.041455
37856	1	1-10463	\N	93	93	23	28	2012-08-02 11:15:02.041455
37871	4	1-9055	37870	14	27	2735	23	2012-08-02 11:15:04.35059
37872	4	1-10059	37870	27	29	1741	93	2012-08-02 11:15:04.35059
37873	4	1-10684	37870	29	66	2250	8	2012-08-02 11:15:04.35059
37870	4	1-11654	37870	66	14	2586	63	2012-08-02 11:15:04.35059
37893	4	1-7075	37892	54	30	1321	54	2012-08-02 11:15:05.510735
37894	4	1-11572	37892	30	73	1813	5	2012-08-02 11:15:05.510735
37895	4	1-10949	37892	73	70	1851	94	2012-08-02 11:15:05.510735
37892	4	1-11657	37892	70	54	1395	8	2012-08-02 11:15:05.510735
37945	4	1-11424	37944	81	80	3267	32	2012-08-02 11:15:11.399837
37946	4	1-11612	37944	80	68	2177	80	2012-08-02 11:15:11.399837
37947	4	1-11285	37944	68	95	1082	45	2012-08-02 11:15:11.399837
37944	4	1-11673	37944	95	81	8103	58	2012-08-02 11:15:11.399837
37949	3	1-8255	37948	71	34	10	10	2012-08-02 11:15:11.399837
37950	3	1-11430	37948	34	95	6	45	2012-08-02 11:15:11.399837
37948	3	1-11673	37948	95	71	57	58	2012-08-02 11:15:11.399837
37999	3	1-8542	37998	41	68	1011	60	2012-08-02 11:15:16.6464
38000	3	1-11381	37998	68	27	803	89	2012-08-02 11:15:16.6464
37998	3	1-11685	37998	27	41	1314	6	2012-08-02 11:15:16.6464
38002	3	1-8936	38001	77	52	1831	18	2012-08-02 11:15:16.6464
38003	3	1-10783	38001	52	27	2308	89	2012-08-02 11:15:16.6464
38001	3	1-11685	38001	27	77	4828	6	2012-08-02 11:15:16.6464
38009	4	1-11671	38008	80	20	1963	37	2012-08-02 11:15:17.845583
38010	4	1-10958	38008	20	66	682	43	2012-08-02 11:15:17.845583
38011	4	1-9570	38008	66	56	467	81	2012-08-02 11:15:17.845583
38008	4	1-11688	38008	56	80	638	62	2012-08-02 11:15:17.845583
38013	4	1-7449	38012	9	68	1390	60	2012-08-02 11:15:17.845583
38014	4	1-11381	38012	68	11	1390	89	2012-08-02 11:15:17.845583
38015	4	1-11339	38012	11	56	1142	81	2012-08-02 11:15:17.845583
38012	4	1-11688	38012	56	9	1769	62	2012-08-02 11:15:17.845583
38017	4	1-7923	38016	49	17	2439	15	2012-08-02 11:15:17.845583
38018	4	1-11482	38016	17	11	1570	89	2012-08-02 11:15:17.845583
38019	4	1-11339	38016	11	56	1550	81	2012-08-02 11:15:17.845583
38016	4	1-11688	38016	56	49	2884	62	2012-08-02 11:15:17.845583
38045	4	1-11576	38044	52	60	1614	88	2012-08-02 11:15:22.172465
38046	4	1-11556	38044	60	3	1789	95	2012-08-02 11:15:22.172465
38047	4	1-10467	38044	3	13	585	33	2012-08-02 11:15:22.172465
38044	4	1-11699	38044	13	52	1546	21	2012-08-02 11:15:22.172465
38049	4	1-10861	38048	80	30	3905	54	2012-08-02 11:15:23.356133
38050	4	1-11572	38048	30	70	4669	5	2012-08-02 11:15:23.356133
38051	4	1-10888	38048	70	16	3689	94	2012-08-02 11:15:23.356133
38048	4	1-11702	38048	16	80	2813	24	2012-08-02 11:15:23.356133
38053	4	1-9265	38052	73	52	1219	31	2012-08-02 11:15:23.356133
38054	4	1-11297	38052	52	46	1311	16	2012-08-02 11:15:23.356133
38055	4	1-11377	38052	46	16	1387	94	2012-08-02 11:15:23.356133
38052	4	1-11702	38052	16	73	1192	24	2012-08-02 11:15:23.356133
38063	4	1-11546	38062	76	73	214	14	2012-08-02 11:15:24.518882
38064	4	1-11097	38062	73	1	113	86	2012-08-02 11:15:24.518882
38065	4	1-9838	38062	1	77	86	34	2012-08-02 11:15:24.518882
38062	4	1-11705	38062	77	76	99	87	2012-08-02 11:15:24.518882
38067	4	1-11546	38066	76	48	598	14	2012-08-02 11:15:24.518882
38068	4	1-9910	38066	48	50	239	9	2012-08-02 11:15:24.518882
38069	4	1-7534	38066	50	77	145	34	2012-08-02 11:15:24.518882
38066	4	1-11705	38066	77	76	215	87	2012-08-02 11:15:24.518882
38071	4	1-10502	38070	83	44	338	1	2012-08-02 11:15:24.518882
38072	4	1-11018	38070	44	48	667	51	2012-08-02 11:15:24.518882
38073	4	1-11450	38070	48	77	638	34	2012-08-02 11:15:24.518882
38070	4	1-11705	38070	77	83	990	87	2012-08-02 11:15:24.518882
38107	4	1-11448	38106	27	89	2124	12	2012-08-02 11:15:29.184735
38108	4	1-10794	38106	89	1	1140	43	2012-08-02 11:15:29.184735
38109	4	1-6318	38106	1	8	575	59	2012-08-02 11:15:29.184735
38106	4	1-11717	38106	8	27	2313	2	2012-08-02 11:15:29.184735
38131	4	1-10230	38130	95	73	4443	12	2012-08-02 11:15:32.700252
38132	4	1-6499	38130	73	25	2569	29	2012-08-02 11:15:32.700252
38133	4	1-11393	38130	25	49	4159	86	2012-08-02 11:15:32.700252
38130	4	1-11726	38130	49	95	5882	41	2012-08-02 11:15:32.700252
38135	4	1-6945	38134	64	42	1416	81	2012-08-02 11:15:32.700252
38136	4	1-9976	38134	42	73	1203	14	2012-08-02 11:15:32.700252
38137	4	1-11097	38134	73	49	888	86	2012-08-02 11:15:32.700252
38134	4	1-11726	38134	49	64	1445	41	2012-08-02 11:15:32.700252
38139	4	1-11554	38138	55	51	1064	41	2012-08-02 11:15:33.874328
38140	4	1-5302	38138	51	32	897	58	2012-08-02 11:15:33.874328
38141	4	1-11394	38138	32	99	840	79	2012-08-02 11:15:33.874328
38138	4	1-11729	38138	99	55	860	33	2012-08-02 11:15:33.874328
38143	4	1-11547	38142	84	76	1935	20	2012-08-02 11:15:33.874328
38144	4	1-8202	38142	76	96	456	70	2012-08-02 11:15:33.874328
38145	4	1-10451	38142	96	99	249	79	2012-08-02 11:15:33.874328
38142	4	1-11729	38142	99	84	255	33	2012-08-02 11:15:33.874328
38147	4	1-11711	38146	67	90	3093	61	2012-08-02 11:15:33.874328
38148	4	1-11288	38146	90	18	2966	62	2012-08-02 11:15:33.874328
38149	4	1-4861	38146	18	99	1241	79	2012-08-02 11:15:33.874328
38146	4	1-11729	38146	99	67	1308	33	2012-08-02 11:15:33.874328
38200	3	1-8484	38199	80	33	2306	64	2012-08-02 11:15:44.03822
38201	3	1-10838	38199	33	71	3319	87	2012-08-02 11:15:44.03822
38199	3	1-11755	38199	71	80	5459	8	2012-08-02 11:15:44.03822
38203	4	1-11512	38202	50	85	2199	24	2012-08-02 11:15:44.03822
37902	4	1-11506	37899	88	61	1150	57	2012-08-02 11:15:08.648528
37904	4	1-849	37903	22	70	729	17	2012-08-02 11:15:08.648528
37905	4	1-11372	37903	70	88	862	48	2012-08-02 11:15:08.648528
37906	4	1-11506	37903	88	61	1203	57	2012-08-02 11:15:08.648528
37903	4	1-11665	37903	61	22	2066	62	2012-08-02 11:15:08.648528
37923	3	1-9185	37922	61	54	209	98	2012-08-02 11:15:09.819771
37924	3	1-10403	37922	54	7	144	25	2012-08-02 11:15:09.819771
37922	3	1-11668	37922	7	61	352	74	2012-08-02 11:15:09.819771
37926	4	1-9185	37925	61	21	3934	98	2012-08-02 11:15:09.819771
37927	4	1-11256	37925	21	20	2540	26	2012-08-02 11:15:09.819771
37928	4	1-8747	37925	20	7	2477	25	2012-08-02 11:15:09.819771
37925	4	1-11668	37925	7	61	6340	74	2012-08-02 11:15:09.819771
37930	4	1-9220	37929	47	28	594	51	2012-08-02 11:15:09.819771
37931	4	1-11573	37929	28	53	920	91	2012-08-02 11:15:09.819771
37932	4	1-11417	37929	53	7	618	25	2012-08-02 11:15:09.819771
37929	4	1-11668	37929	7	47	1770	74	2012-08-02 11:15:09.819771
37933	1	1-3794	\N	42	42	736	25	2012-08-02 11:15:09.819771
37941	4	1-10853	37940	36	62	884	36	2012-08-02 11:15:10.991124
37942	4	1-10027	37940	62	61	735	57	2012-08-02 11:15:10.991124
37943	4	1-11665	37940	61	80	1329	62	2012-08-02 11:15:10.991124
37940	4	1-11671	37940	80	36	5035	37	2012-08-02 11:15:10.991124
37952	4	1-11404	37951	100	23	712	77	2012-08-02 11:15:12.322084
37953	4	1-7526	37951	23	100	477	85	2012-08-02 11:15:12.322084
37954	4	1-11164	37951	100	35	340	65	2012-08-02 11:15:12.322084
37951	4	1-11674	37951	35	100	352	23	2012-08-02 11:15:12.322084
37956	3	1-8524	37955	89	48	3640	40	2012-08-02 11:15:12.322084
37957	3	1-11497	37955	48	35	2396	65	2012-08-02 11:15:12.322084
37955	3	1-11674	37955	35	89	3050	23	2012-08-02 11:15:12.322084
37963	3	1-11024	37962	76	66	2748	34	2012-08-02 11:15:13.522864
37964	3	1-4375	37962	66	84	473	85	2012-08-02 11:15:13.522864
37962	3	1-11677	37962	84	76	8359	82	2012-08-02 11:15:13.522864
37966	4	1-11499	37965	25	8	773	77	2012-08-02 11:15:13.522864
37967	4	1-6714	37965	8	87	93	84	2012-08-02 11:15:13.522864
37968	4	1-10510	37965	87	84	62	85	2012-08-02 11:15:13.522864
37965	4	1-11677	37965	84	25	1110	82	2012-08-02 11:15:13.522864
37970	3	1-5313	37969	78	53	78	4	2012-08-02 11:15:13.522864
37971	3	1-3636	37969	53	84	26	85	2012-08-02 11:15:13.522864
37969	3	1-11677	37969	84	78	483	82	2012-08-02 11:15:13.522864
37973	4	1-11553	37972	63	22	4721	55	2012-08-02 11:15:14.700276
37974	4	1-10067	37972	22	19	3560	54	2012-08-02 11:15:14.700276
37975	4	1-11669	37972	19	85	2759	32	2012-08-02 11:15:14.700276
37972	4	1-11680	37972	85	63	4121	69	2012-08-02 11:15:14.700276
38005	4	1-11580	38004	1	30	162	71	2012-08-02 11:15:17.02225
38006	4	1-11313	38004	30	5	167	61	2012-08-02 11:15:17.02225
38007	4	1-4498	38004	5	53	80	20	2012-08-02 11:15:17.02225
38004	4	1-11686	38004	53	1	51	8	2012-08-02 11:15:17.02225
38021	4	1-11674	38020	35	71	348	23	2012-08-02 11:15:18.234471
38022	4	1-11240	38020	71	22	130	63	2012-08-02 11:15:18.234471
38023	4	1-11166	38020	22	41	373	42	2012-08-02 11:15:18.234471
38020	4	1-11689	38020	41	35	314	65	2012-08-02 11:15:18.234471
38025	4	1-11336	38024	83	100	254	35	2012-08-02 11:15:18.234471
38026	4	1-11396	38024	100	22	157	63	2012-08-02 11:15:18.234471
38027	4	1-11166	38024	22	41	488	42	2012-08-02 11:15:18.234471
38024	4	1-11689	38024	41	83	444	65	2012-08-02 11:15:18.234471
38029	4	1-11088	38028	10	70	685	17	2012-08-02 11:15:18.234471
38030	4	1-11372	38028	70	74	878	48	2012-08-02 11:15:18.234471
38031	4	1-11581	38028	74	41	623	42	2012-08-02 11:15:18.234471
38028	4	1-11689	38028	41	10	613	65	2012-08-02 11:15:18.234471
38033	4	1-11070	38032	53	86	155	39	2012-08-02 11:15:18.234471
38034	4	1-10944	38032	86	72	147	1	2012-08-02 11:15:18.234471
38035	4	1-8885	38032	72	41	168	42	2012-08-02 11:15:18.234471
38032	4	1-11689	38032	41	53	179	65	2012-08-02 11:15:18.234471
38036	1	1-6111	\N	69	69	1720	65	2012-08-02 11:15:18.234471
38038	3	1-11665	38037	61	28	2280	62	2012-08-02 11:15:20.569822
38039	3	1-6203	38037	28	69	1226	87	2012-08-02 11:15:20.569822
38037	3	1-11695	38037	69	61	1522	57	2012-08-02 11:15:20.569822
38041	4	1-10048	38040	62	96	3909	55	2012-08-02 11:15:20.569822
38042	4	1-8000	38040	96	96	1246	96	2012-08-02 11:15:20.569822
38043	4	1-11621	38040	96	69	4312	87	2012-08-02 11:15:20.569822
38040	4	1-11695	38040	69	62	6406	57	2012-08-02 11:15:20.569822
38059	4	1-8464	38058	79	39	791	86	2012-08-02 11:15:24.120713
38060	4	1-10331	38058	39	5	1709	60	2012-08-02 11:15:24.120713
38061	4	1-11646	38058	5	41	2840	72	2012-08-02 11:15:24.120713
38058	4	1-11704	38058	41	79	1957	55	2012-08-02 11:15:24.120713
38075	4	1-10497	38074	95	72	1923	21	2012-08-02 11:15:25.281863
38076	4	1-11290	38074	72	63	1755	69	2012-08-02 11:15:25.281863
38077	4	1-11553	38074	63	16	2370	55	2012-08-02 11:15:25.281863
38074	4	1-11707	38074	16	95	2044	51	2012-08-02 11:15:25.281863
38079	4	1-11444	38078	37	61	1210	88	2012-08-02 11:15:26.476247
38080	4	1-11389	38078	61	57	783	46	2012-08-02 11:15:26.476247
38081	4	1-10835	38078	57	72	446	44	2012-08-02 11:15:26.476247
38078	4	1-11710	38078	72	37	1530	16	2012-08-02 11:15:26.476247
38083	4	1-11377	38082	46	74	1569	94	2012-08-02 11:15:26.476247
38084	4	1-11020	38082	74	8	446	19	2012-08-02 11:15:26.476247
38085	4	1-6991	38082	8	72	387	44	2012-08-02 11:15:26.476247
38082	4	1-11710	38082	72	46	1456	16	2012-08-02 11:15:26.476247
38087	4	1-9493	38086	24	19	747	19	2012-08-02 11:15:27.637065
38088	4	1-11652	38086	19	20	1349	26	2012-08-02 11:15:27.637065
38089	4	1-8747	38086	20	26	1636	25	2012-08-02 11:15:27.637065
38086	4	1-11713	38086	26	24	2486	76	2012-08-02 11:15:27.637065
38091	4	1-9493	38090	24	6	227	19	2012-08-02 11:15:27.637065
38092	4	1-11463	38090	6	68	248	28	2012-08-02 11:15:27.637065
38093	4	1-10480	38090	68	26	483	25	2012-08-02 11:15:27.637065
38090	4	1-11713	38090	26	24	745	76	2012-08-02 11:15:27.637065
38111	4	1-10783	38110	52	74	1468	89	2012-08-02 11:15:31.131114
38112	4	1-10349	38110	74	41	947	72	2012-08-02 11:15:31.131114
38113	4	1-11704	38110	41	66	427	55	2012-08-02 11:15:31.131114
38110	4	1-11721	38110	66	52	1514	18	2012-08-02 11:15:31.131114
38115	4	1-7375	38114	32	47	1387	42	2012-08-02 11:15:31.131114
38116	4	1-11604	38114	47	8	1317	67	2012-08-02 11:15:31.131114
38117	4	1-9407	38114	8	66	473	55	2012-08-02 11:15:31.131114
38114	4	1-11721	38114	66	32	2066	18	2012-08-02 11:15:31.131114
38119	4	1-11489	38118	86	14	741	90	2012-08-02 11:15:31.131114
37908	3	1-9363	37907	80	93	1386	74	2012-08-02 11:15:09.432839
37909	3	1-10768	37907	93	96	1860	75	2012-08-02 11:15:09.432839
37911	4	1-5283	37910	80	57	839	4	2012-08-02 11:15:09.432839
37912	4	1-10885	37910	57	93	405	74	2012-08-02 11:15:09.432839
37913	4	1-10768	37910	93	96	668	75	2012-08-02 11:15:09.432839
37910	4	1-11667	37910	96	80	1465	24	2012-08-02 11:15:09.432839
37915	4	1-9265	37914	73	82	3798	31	2012-08-02 11:15:09.432839
37916	4	1-11282	37914	82	37	5476	43	2012-08-02 11:15:09.432839
37917	4	1-8159	37914	37	96	1770	75	2012-08-02 11:15:09.432839
37914	4	1-11667	37914	96	73	3891	24	2012-08-02 11:15:09.432839
37919	4	1-11519	37918	92	53	458	65	2012-08-02 11:15:09.432839
37920	4	1-11070	37918	53	94	378	39	2012-08-02 11:15:09.432839
37921	4	1-11055	37918	94	96	296	75	2012-08-02 11:15:09.432839
37918	4	1-11667	37918	96	92	654	24	2012-08-02 11:15:09.432839
37935	3	1-9495	37934	76	53	1930	43	2012-08-02 11:15:10.604364
37936	3	1-9287	37934	53	72	634	100	2012-08-02 11:15:10.604364
37934	3	1-11670	37934	72	76	5213	78	2012-08-02 11:15:10.604364
37938	3	1-8138	37937	48	93	454	56	2012-08-02 11:15:10.604364
37939	3	1-9512	37937	93	72	474	100	2012-08-02 11:15:10.604364
37937	3	1-11670	37937	72	48	3901	78	2012-08-02 11:15:10.604364
37959	4	1-10896	37958	27	73	1815	14	2012-08-02 11:15:12.709591
37960	4	1-11097	37958	73	1	1067	86	2012-08-02 11:15:12.709591
37961	4	1-9838	37958	1	45	907	34	2012-08-02 11:15:12.709591
37958	4	1-11675	37958	45	27	1946	2	2012-08-02 11:15:12.709591
37977	3	1-11055	37976	94	86	828	75	2012-08-02 11:15:16.259253
37978	3	1-3501	37976	86	59	288	38	2012-08-02 11:15:16.259253
37976	3	1-11684	37976	59	94	2046	39	2012-08-02 11:15:16.259253
37980	3	1-6611	37979	95	27	709	42	2012-08-02 11:15:16.259253
37981	3	1-11439	37979	27	59	175	38	2012-08-02 11:15:16.259253
37979	3	1-11684	37979	59	95	1333	39	2012-08-02 11:15:16.259253
37983	3	1-8251	37982	46	12	327	83	2012-08-02 11:15:16.259253
37984	3	1-11501	37982	12	59	80	38	2012-08-02 11:15:16.259253
37982	3	1-11684	37982	59	46	614	39	2012-08-02 11:15:16.259253
37986	4	1-10944	37985	86	57	2556	1	2012-08-02 11:15:16.259253
37987	4	1-3689	37985	57	21	1104	58	2012-08-02 11:15:16.259253
37988	4	1-11435	37985	21	59	615	38	2012-08-02 11:15:16.259253
37985	4	1-11684	37985	59	86	4895	39	2012-08-02 11:15:16.259253
37990	4	1-8578	37989	17	64	22	84	2012-08-02 11:15:16.259253
37991	4	1-11170	37989	64	27	36	42	2012-08-02 11:15:16.259253
37992	4	1-11439	37989	27	59	9	38	2012-08-02 11:15:16.259253
37989	4	1-11684	37989	59	17	76	39	2012-08-02 11:15:16.259253
37994	4	1-10207	37993	89	64	872	87	2012-08-02 11:15:16.259253
37995	4	1-11548	37993	64	87	408	59	2012-08-02 11:15:16.259253
37996	4	1-11016	37993	87	59	109	38	2012-08-02 11:15:16.259253
37993	4	1-11684	37993	59	89	942	39	2012-08-02 11:15:16.259253
37997	1	1-3501	\N	86	86	4755	38	2012-08-02 11:15:16.259253
38057	2	1-11531	38056	79	8	4158	81	2012-08-02 11:15:23.743163
38056	2	1-11703	38056	8	79	4590	18	2012-08-02 11:15:23.743163
38095	3	1-11621	38094	96	32	208	87	2012-08-02 11:15:28.432689
38096	3	1-11460	38094	32	85	54	70	2012-08-02 11:15:28.432689
38094	3	1-11715	38094	85	96	64	96	2012-08-02 11:15:28.432689
38098	4	1-11206	38097	78	96	5532	99	2012-08-02 11:15:28.432689
38099	4	1-9598	38097	96	66	5844	46	2012-08-02 11:15:28.432689
38100	4	1-11048	38097	66	85	5195	70	2012-08-02 11:15:28.432689
38097	4	1-11715	38097	85	78	6961	96	2012-08-02 11:15:28.432689
38102	4	1-10428	38101	82	10	1845	64	2012-08-02 11:15:28.432689
38103	4	1-10631	38101	10	7	2236	67	2012-08-02 11:15:28.432689
38104	4	1-8529	38101	7	85	2092	70	2012-08-02 11:15:28.432689
38101	4	1-11715	38101	85	82	2854	96	2012-08-02 11:15:28.432689
38105	1	1-4574	\N	28	28	984	87	2012-08-02 11:15:28.432689
38123	4	1-11713	38122	26	94	5238	76	2012-08-02 11:15:32.313624
38124	4	1-10064	38122	94	50	1600	8	2012-08-02 11:15:32.313624
38125	4	1-11512	38122	50	55	1668	24	2012-08-02 11:15:32.313624
38122	4	1-11725	38122	55	26	3881	25	2012-08-02 11:15:32.313624
38127	4	1-5364	38126	41	46	7	16	2012-08-02 11:15:32.313624
38128	4	1-11377	38126	46	16	7	94	2012-08-02 11:15:32.313624
38129	4	1-11702	38126	16	55	6	24	2012-08-02 11:15:32.313624
38126	4	1-11725	38126	55	41	16	25	2012-08-02 11:15:32.313624
38151	4	1-11659	38150	15	47	2081	68	2012-08-02 11:15:34.62543
38152	4	1-11300	38150	47	34	1142	78	2012-08-02 11:15:34.62543
38153	4	1-9024	38150	34	78	427	69	2012-08-02 11:15:34.62543
38150	4	1-11731	38150	78	15	2403	13	2012-08-02 11:15:34.62543
38155	4	1-11221	38154	64	66	2635	8	2012-08-02 11:15:34.62543
38156	4	1-11654	38154	66	35	1568	63	2012-08-02 11:15:34.62543
38157	4	1-9065	38154	35	78	1042	69	2012-08-02 11:15:34.62543
38154	4	1-11731	38154	78	64	6160	13	2012-08-02 11:15:34.62543
38159	3	1-7296	38158	79	1	222	95	2012-08-02 11:15:34.62543
38160	3	1-10393	38158	1	78	98	69	2012-08-02 11:15:34.62543
38158	3	1-11731	38158	78	79	653	13	2012-08-02 11:15:34.62543
38162	5	1-11359	38161	77	58	1216	94	2012-08-02 11:15:38.947247
38163	5	1-11579	38161	58	43	812	63	2012-08-02 11:15:38.947247
38164	5	1-10638	38161	43	5	764	11	2012-08-02 11:15:38.947247
38165	5	1-11378	38161	5	21	738	35	2012-08-02 11:15:38.947247
38161	5	1-11742	38161	21	77	1325	6	2012-08-02 11:15:38.947247
38179	4	1-11377	38178	46	74	926	94	2012-08-02 11:15:41.278827
38180	4	1-11020	38178	74	19	250	19	2012-08-02 11:15:41.278827
38181	4	1-11652	38178	19	18	437	26	2012-08-02 11:15:41.278827
38178	4	1-11748	38178	18	46	904	16	2012-08-02 11:15:41.278827
38266	3	1-10411	38265	59	48	17	92	2012-08-02 11:15:54.281626
38267	3	1-11589	38265	48	31	5	49	2012-08-02 11:15:54.281626
38265	3	1-11781	38265	31	59	22	36	2012-08-02 11:15:54.281626
38269	3	1-9050	38268	26	16	1749	77	2012-08-02 11:15:54.281626
38270	3	1-8605	38268	16	31	448	49	2012-08-02 11:15:54.281626
38268	3	1-11781	38268	31	26	2887	36	2012-08-02 11:15:54.281626
38272	4	1-11738	38271	55	1	1971	19	2012-08-02 11:15:54.281626
38273	4	1-10330	38271	1	65	1092	41	2012-08-02 11:15:54.281626
38274	4	1-11639	38271	65	31	213	49	2012-08-02 11:15:54.281626
38271	4	1-11781	38271	31	55	1525	36	2012-08-02 11:15:54.281626
38293	4	1-11756	38292	28	48	3923	87	2012-08-02 11:15:56.66897
38294	4	1-8115	38292	48	43	756	97	2012-08-02 11:15:56.66897
38295	4	1-10959	38292	43	90	430	66	2012-08-02 11:15:56.66897
38292	4	1-11787	38292	90	28	1348	67	2012-08-02 11:15:56.66897
38340	4	1-11791	38339	70	55	2186	1	2012-08-02 11:16:04.869133
38341	4	1-10785	38339	55	56	708	78	2012-08-02 11:16:04.869133
38339	4	1-11809	38339	27	70	1527	39	2012-08-02 11:16:04.869133
38120	4	1-10459	38118	14	13	448	98	2012-08-02 11:15:31.131114
38121	4	1-11287	38118	13	66	224	55	2012-08-02 11:15:31.131114
38118	4	1-11721	38118	66	86	994	18	2012-08-02 11:15:31.131114
38167	4	1-4500	38166	56	5	1096	60	2012-08-02 11:15:39.333955
38168	4	1-11646	38166	5	10	987	72	2012-08-02 11:15:39.333955
38169	4	1-11181	38166	10	67	314	56	2012-08-02 11:15:39.333955
38166	4	1-11743	38166	67	56	2914	67	2012-08-02 11:15:39.333955
38171	4	1-10547	38170	70	5	828	60	2012-08-02 11:15:39.333955
38172	4	1-11646	38170	5	10	783	72	2012-08-02 11:15:39.333955
38173	4	1-11181	38170	10	67	262	56	2012-08-02 11:15:39.333955
38170	4	1-11743	38170	67	70	2552	67	2012-08-02 11:15:39.333955
38175	4	1-8528	38174	100	29	1093	31	2012-08-02 11:15:39.333955
38176	4	1-11660	38174	29	1	830	61	2012-08-02 11:15:39.333955
38177	4	1-2136	38174	1	67	166	56	2012-08-02 11:15:39.333955
38174	4	1-11743	38174	67	100	2263	67	2012-08-02 11:15:39.333955
38183	5	1-11133	38182	15	6	197	47	2012-08-02 11:15:42.847989
38184	5	1-2894	38182	6	41	89	42	2012-08-02 11:15:42.847989
38185	5	1-11689	38182	41	53	62	65	2012-08-02 11:15:42.847989
38186	5	1-11070	38182	53	45	35	39	2012-08-02 11:15:42.847989
38182	5	1-11751	38182	45	15	94	56	2012-08-02 11:15:42.847989
38188	2	1-11368	38187	34	45	952	39	2012-08-02 11:15:42.847989
38187	2	1-11751	38187	45	34	2968	56	2012-08-02 11:15:42.847989
38190	5	1-8518	38189	39	33	368	4	2012-08-02 11:15:42.847989
38191	5	1-9843	38189	33	41	334	42	2012-08-02 11:15:42.847989
38192	5	1-11689	38189	41	53	296	65	2012-08-02 11:15:42.847989
38193	5	1-11070	38189	53	45	214	39	2012-08-02 11:15:42.847989
38189	5	1-11751	38189	45	39	724	56	2012-08-02 11:15:42.847989
38195	4	1-9512	38194	93	59	1378	100	2012-08-02 11:15:42.847989
38196	4	1-11453	38194	59	78	1668	50	2012-08-02 11:15:42.847989
38197	4	1-4559	38194	78	45	199	39	2012-08-02 11:15:42.847989
38194	4	1-11751	38194	45	93	755	56	2012-08-02 11:15:42.847989
38198	1	1-4031	\N	12	12	75	39	2012-08-02 11:15:42.847989
38215	4	1-11694	38214	45	69	646	96	2012-08-02 11:15:45.950096
38216	4	1-10304	38214	69	68	405	80	2012-08-02 11:15:45.950096
38217	4	1-11285	38214	68	58	460	45	2012-08-02 11:15:45.950096
38214	4	1-11760	38214	58	45	1109	60	2012-08-02 11:15:45.950096
38219	3	1-7286	38218	67	66	1106	68	2012-08-02 11:15:47.245425
38220	3	1-9647	38218	66	39	701	30	2012-08-02 11:15:47.245425
38218	3	1-11763	38218	39	67	1406	99	2012-08-02 11:15:47.245425
38222	4	1-6272	38221	99	61	3879	92	2012-08-02 11:15:47.245425
38223	4	1-11369	38221	61	83	2658	27	2012-08-02 11:15:47.245425
38224	4	1-11595	38221	83	39	1625	30	2012-08-02 11:15:47.245425
38221	4	1-11763	38221	39	99	3501	99	2012-08-02 11:15:47.245425
38226	4	1-11658	38225	18	64	3158	29	2012-08-02 11:15:47.245425
38227	4	1-11179	38225	64	83	1855	27	2012-08-02 11:15:47.245425
38228	4	1-11595	38225	83	39	1291	30	2012-08-02 11:15:47.245425
38225	4	1-11763	38225	39	18	3166	99	2012-08-02 11:15:47.245425
38247	4	1-6590	38246	42	93	1938	29	2012-08-02 11:15:49.574199
38248	4	1-11508	38246	93	27	3187	23	2012-08-02 11:15:49.574199
38249	4	1-10059	38246	27	77	2089	93	2012-08-02 11:15:49.574199
38246	4	1-11769	38246	77	42	2032	35	2012-08-02 11:15:49.574199
38254	4	1-11627	38253	17	28	768	67	2012-08-02 11:15:52.734572
38255	4	1-11756	38253	28	83	1669	87	2012-08-02 11:15:52.734572
38256	4	1-10502	38253	83	18	446	1	2012-08-02 11:15:52.734572
38253	4	1-11777	38253	18	17	744	74	2012-08-02 11:15:52.734572
38258	4	1-11040	38257	50	66	5741	40	2012-08-02 11:15:52.734572
38259	4	1-11527	38257	66	86	1595	39	2012-08-02 11:15:52.734572
38260	4	1-10944	38257	86	18	1216	1	2012-08-02 11:15:52.734572
38257	4	1-11777	38257	18	50	2357	74	2012-08-02 11:15:52.734572
38276	2	1-11115	38275	43	10	762	16	2012-08-02 11:15:55.065674
38275	2	1-11782	38275	10	43	802	30	2012-08-02 11:15:55.065674
38278	4	1-8209	38277	69	75	2587	41	2012-08-02 11:15:55.065674
38279	4	1-9875	38277	75	84	2610	100	2012-08-02 11:15:55.065674
38280	4	1-11176	38277	84	10	2417	16	2012-08-02 11:15:55.065674
38277	4	1-11782	38277	10	69	3752	30	2012-08-02 11:15:55.065674
38282	4	1-4022	38281	61	88	1544	50	2012-08-02 11:15:55.065674
38283	4	1-11555	38281	88	48	1977	55	2012-08-02 11:15:55.065674
38284	4	1-9499	38281	48	10	1804	16	2012-08-02 11:15:55.065674
38281	4	1-11782	38281	10	61	2802	30	2012-08-02 11:15:55.065674
38286	3	1-8209	38285	69	9	1034	41	2012-08-02 11:15:55.065674
38287	3	1-7403	38285	9	10	876	16	2012-08-02 11:15:55.065674
38285	3	1-11782	38285	10	69	1428	30	2012-08-02 11:15:55.065674
38289	4	1-4684	38288	47	19	23	19	2012-08-02 11:15:55.065674
38290	4	1-11652	38288	19	18	38	26	2012-08-02 11:15:55.065674
38291	4	1-11748	38288	18	10	76	16	2012-08-02 11:15:55.065674
38288	4	1-11782	38288	10	47	125	30	2012-08-02 11:15:55.065674
38300	4	1-11777	38299	18	50	993	74	2012-08-02 11:15:58.227009
38301	4	1-11040	38299	50	66	1877	40	2012-08-02 11:15:58.227009
38302	4	1-11527	38299	66	70	405	39	2012-08-02 11:15:58.227009
38299	4	1-11791	38299	70	18	660	1	2012-08-02 11:15:58.227009
38304	4	1-11419	38303	81	88	2570	50	2012-08-02 11:15:58.227009
38305	4	1-11555	38303	88	39	3120	55	2012-08-02 11:15:58.227009
38306	4	1-9676	38303	39	70	1401	39	2012-08-02 11:15:58.227009
38303	4	1-11791	38303	70	81	3049	1	2012-08-02 11:15:58.227009
38308	5	1-11777	38307	18	91	1081	74	2012-08-02 11:15:58.227009
38309	5	1-6946	38307	91	63	424	69	2012-08-02 11:15:58.227009
38310	5	1-11553	38307	63	39	497	55	2012-08-02 11:15:58.227009
38311	5	1-9676	38307	39	70	231	39	2012-08-02 11:15:58.227009
38307	5	1-11791	38307	70	18	521	1	2012-08-02 11:15:58.227009
38313	5	1-11777	38312	18	61	1886	74	2012-08-02 11:15:58.227009
38314	5	1-9185	38312	61	13	1330	98	2012-08-02 11:15:58.227009
38315	5	1-11287	38312	13	39	747	55	2012-08-02 11:15:58.227009
38316	5	1-9676	38312	39	70	365	39	2012-08-02 11:15:58.227009
38312	5	1-11791	38312	70	18	864	1	2012-08-02 11:15:58.227009
38318	5	1-7347	38317	6	1	593	95	2012-08-02 11:15:58.227009
38319	5	1-10393	38317	1	63	434	69	2012-08-02 11:15:58.227009
38320	5	1-11553	38317	63	39	556	55	2012-08-02 11:15:58.227009
38321	5	1-9676	38317	39	70	282	39	2012-08-02 11:15:58.227009
38317	5	1-11791	38317	70	6	693	1	2012-08-02 11:15:58.227009
38323	5	1-11777	38322	18	85	2035	74	2012-08-02 11:15:58.227009
38204	4	1-5429	38202	85	33	817	64	2012-08-02 11:15:44.03822
38205	4	1-10838	38202	33	71	1310	87	2012-08-02 11:15:44.03822
38202	4	1-11755	38202	71	50	2401	8	2012-08-02 11:15:44.03822
38207	4	1-11295	38206	26	56	465	81	2012-08-02 11:15:44.03822
38208	4	1-11688	38206	56	28	707	62	2012-08-02 11:15:44.03822
38209	4	1-6203	38206	28	71	414	87	2012-08-02 11:15:44.03822
38206	4	1-11755	38206	71	26	796	8	2012-08-02 11:15:44.03822
38211	4	1-11512	38210	50	99	215	24	2012-08-02 11:15:44.03822
38212	4	1-8135	38210	99	48	168	66	2012-08-02 11:15:44.03822
38213	4	1-10878	38210	48	71	110	87	2012-08-02 11:15:44.03822
38210	4	1-11755	38210	71	50	218	8	2012-08-02 11:15:44.03822
38230	4	1-11658	38229	18	45	467	29	2012-08-02 11:15:48.789614
38231	4	1-10967	38229	45	100	456	57	2012-08-02 11:15:48.789614
38232	4	1-10010	38229	100	52	131	75	2012-08-02 11:15:48.789614
38229	4	1-11768	38229	52	18	579	99	2012-08-02 11:15:48.789614
38234	5	1-5937	38233	60	55	140	36	2012-08-02 11:15:48.789614
38235	5	1-11738	38233	55	27	221	19	2012-08-02 11:15:48.789614
38236	5	1-11456	38233	27	78	112	11	2012-08-02 11:15:48.789614
38237	5	1-7907	38233	78	52	67	75	2012-08-02 11:15:48.789614
38233	5	1-11768	38233	52	60	302	99	2012-08-02 11:15:48.789614
38239	4	1-7273	38238	72	28	896	19	2012-08-02 11:15:48.789614
38240	4	1-11722	38238	28	8	2223	95	2012-08-02 11:15:48.789614
38241	4	1-11117	38238	8	52	451	75	2012-08-02 11:15:48.789614
38238	4	1-11768	38238	52	72	2297	99	2012-08-02 11:15:48.789614
38243	4	1-7273	38242	72	27	698	19	2012-08-02 11:15:48.789614
38244	4	1-11456	38242	27	78	426	11	2012-08-02 11:15:48.789614
38245	4	1-7907	38242	78	52	308	75	2012-08-02 11:15:48.789614
38242	4	1-11768	38242	52	72	1677	99	2012-08-02 11:15:48.789614
38251	3	1-11501	38250	12	85	697	38	2012-08-02 11:15:50.35883
38252	3	1-2873	38250	85	76	326	5	2012-08-02 11:15:50.35883
38250	3	1-11770	38250	76	12	2404	83	2012-08-02 11:15:50.35883
38262	4	1-11693	38261	70	34	5198	46	2012-08-02 11:15:53.518543
38263	4	1-11326	38261	34	16	2911	22	2012-08-02 11:15:53.518543
38264	4	1-11709	38261	16	11	3274	85	2012-08-02 11:15:53.518543
38261	4	1-11779	38261	11	70	6899	98	2012-08-02 11:15:53.518543
38297	3	1-10485	38296	32	43	1527	51	2012-08-02 11:15:57.442773
38298	3	1-11563	38296	43	52	1556	41	2012-08-02 11:15:57.442773
38296	3	1-11789	38296	52	32	9264	88	2012-08-02 11:15:57.442773
38356	4	1-10255	38355	44	42	2081	40	2012-08-02 11:16:06.857917
38357	4	1-11534	38355	42	20	1753	22	2012-08-02 11:16:06.857917
38358	4	1-9950	38355	20	37	1142	84	2012-08-02 11:16:06.857917
38355	4	1-11812	38355	37	44	3960	82	2012-08-02 11:16:06.857917
38383	4	1-11392	38382	29	83	302	27	2012-08-02 11:16:09.277745
38384	4	1-11595	38382	83	75	113	30	2012-08-02 11:16:09.277745
38385	4	1-9196	38382	75	38	41	93	2012-08-02 11:16:09.277745
38382	4	1-11819	38382	38	29	847	53	2012-08-02 11:16:09.277745
38387	3	1-10354	38386	76	21	734	17	2012-08-02 11:16:09.277745
38388	3	1-5936	38386	21	38	88	93	2012-08-02 11:16:09.277745
38386	3	1-11819	38386	38	76	2109	53	2012-08-02 11:16:09.277745
38390	4	1-9901	38389	86	29	2644	60	2012-08-02 11:16:09.277745
38391	4	1-11545	38389	29	41	640	18	2012-08-02 11:16:09.277745
38392	4	1-2916	38389	41	38	114	93	2012-08-02 11:16:09.277745
38389	4	1-11819	38389	38	86	3039	53	2012-08-02 11:16:09.277745
38394	4	1-10182	38393	14	70	172	94	2012-08-02 11:16:09.277745
38395	4	1-11657	38393	70	12	84	8	2012-08-02 11:16:09.277745
38396	4	1-9794	38393	12	38	39	93	2012-08-02 11:16:09.277745
38393	4	1-11819	38393	38	14	1066	53	2012-08-02 11:16:09.277745
38398	2	1-8216	38397	99	45	771	78	2012-08-02 11:16:11.609493
38397	2	1-11825	38397	45	99	5217	97	2012-08-02 11:16:11.609493
38400	4	1-10582	38399	36	16	2189	94	2012-08-02 11:16:11.609493
38401	4	1-11702	38399	16	90	1159	24	2012-08-02 11:16:11.609493
38402	4	1-10816	38399	90	45	406	78	2012-08-02 11:16:11.609493
38399	4	1-11825	38399	45	36	3305	97	2012-08-02 11:16:11.609493
38412	4	1-11664	38411	76	8	1253	5	2012-08-02 11:16:13.90738
38413	4	1-11186	38411	8	79	898	25	2012-08-02 11:16:13.90738
38414	4	1-11530	38411	79	61	1536	82	2012-08-02 11:16:13.90738
38411	4	1-11831	38411	61	76	1124	29	2012-08-02 11:16:13.90738
38416	4	1-11179	38415	64	50	1372	27	2012-08-02 11:16:13.90738
38417	4	1-11700	38415	50	37	758	84	2012-08-02 11:16:13.90738
38418	4	1-11812	38415	37	61	2692	82	2012-08-02 11:16:13.90738
38415	4	1-11831	38415	61	64	2100	29	2012-08-02 11:16:13.90738
38420	3	1-10884	38419	83	59	602	43	2012-08-02 11:16:15.045628
38421	3	1-10383	38419	59	16	275	90	2012-08-02 11:16:15.045628
38419	3	1-11834	38419	16	83	1614	10	2012-08-02 11:16:15.045628
38423	3	1-10499	38422	43	42	792	38	2012-08-02 11:16:15.045628
38424	3	1-4896	38422	42	16	216	90	2012-08-02 11:16:15.045628
38422	3	1-11834	38422	16	43	1749	10	2012-08-02 11:16:15.045628
38432	4	1-10501	38431	66	90	2013	66	2012-08-02 11:16:17.321403
38433	4	1-11787	38431	90	48	3790	67	2012-08-02 11:16:17.321403
38434	4	1-11565	38431	48	1	2646	21	2012-08-02 11:16:17.321403
38431	4	1-11840	38431	1	66	5671	11	2012-08-02 11:16:17.321403
38436	3	1-8177	38435	36	45	2488	58	2012-08-02 11:16:17.321403
38437	3	1-11328	38435	45	1	1672	21	2012-08-02 11:16:17.321403
38435	3	1-11840	38435	1	36	4027	11	2012-08-02 11:16:17.321403
38447	4	1-11221	38446	64	67	3230	8	2012-08-02 11:16:18.481743
38448	4	1-11229	38446	67	32	3048	22	2012-08-02 11:16:18.481743
38449	4	1-11790	38446	32	78	2702	68	2012-08-02 11:16:18.481743
38446	4	1-11844	38446	78	64	5001	13	2012-08-02 11:16:18.481743
38451	3	1-11082	38450	5	5	354	7	2012-08-02 11:16:18.481743
38452	3	1-9480	38450	5	78	183	68	2012-08-02 11:16:18.481743
38450	3	1-11844	38450	78	5	371	13	2012-08-02 11:16:18.481743
38477	2	1-11088	38476	10	60	3123	17	2012-08-02 11:16:22.338424
38476	2	1-11853	38476	60	10	3680	65	2012-08-02 11:16:22.338424
38485	4	1-11285	38484	68	93	2341	45	2012-08-02 11:16:25.80742
38486	4	1-9382	38484	93	22	1788	6	2012-08-02 11:16:25.80742
38487	4	1-7364	38484	22	93	1075	58	2012-08-02 11:16:25.80742
38484	4	1-11861	38484	93	68	1932	80	2012-08-02 11:16:25.80742
38488	1	1-7839	\N	53	53	3444	6	2012-08-02 11:16:25.80742
38490	4	1-3242	38489	61	28	1356	51	2012-08-02 11:16:26.593607
38491	4	1-11573	38489	28	81	2073	91	2012-08-02 11:16:26.593607
38492	4	1-9171	38489	81	70	1454	95	2012-08-02 11:16:26.593607
38489	4	1-11864	38489	70	61	2901	43	2012-08-02 11:16:26.593607
38494	2	1-7302	38493	63	70	2784	95	2012-08-02 11:16:26.593607
38493	2	1-11864	38493	70	63	6194	43	2012-08-02 11:16:26.593607
38324	5	1-7864	38322	85	41	1073	72	2012-08-02 11:15:58.227009
38325	5	1-11704	38322	41	39	708	55	2012-08-02 11:15:58.227009
38326	5	1-9676	38322	39	70	361	39	2012-08-02 11:15:58.227009
38322	5	1-11791	38322	70	18	893	1	2012-08-02 11:15:58.227009
38328	2	1-4664	38327	32	54	492	59	2012-08-02 11:16:00.603075
38327	2	1-11797	38327	54	32	7638	24	2012-08-02 11:16:00.603075
38330	4	1-9542	38329	95	30	2464	71	2012-08-02 11:16:04.085157
38331	4	1-11313	38329	30	70	2278	61	2012-08-02 11:16:04.085157
38332	4	1-7704	38329	70	15	1654	57	2012-08-02 11:16:04.085157
38329	4	1-11806	38329	15	95	2801	69	2012-08-02 11:16:04.085157
38334	4	1-9542	38333	95	47	1306	71	2012-08-02 11:16:04.085157
38335	4	1-11773	38333	47	78	1059	59	2012-08-02 11:16:04.085157
38336	4	1-10762	38333	78	15	728	57	2012-08-02 11:16:04.085157
38333	4	1-11806	38333	15	95	1353	69	2012-08-02 11:16:04.085157
38338	2	1-3918	38337	100	41	592	44	2012-08-02 11:16:05.269263
38337	2	1-11808	38337	41	100	2112	97	2012-08-02 11:16:05.269263
38349	3	1-11482	38348	17	100	1753	89	2012-08-02 11:16:06.051463
38350	3	1-11485	38348	100	43	1325	73	2012-08-02 11:16:06.051463
38348	3	1-11811	38348	43	17	3670	15	2012-08-02 11:16:06.051463
38352	4	1-11482	38351	17	51	867	89	2012-08-02 11:16:06.051463
38353	4	1-11291	38351	51	8	842	76	2012-08-02 11:16:06.051463
38354	4	1-11794	38351	8	43	419	73	2012-08-02 11:16:06.051463
38351	4	1-11811	38351	43	17	1452	15	2012-08-02 11:16:06.051463
38360	3	1-11676	38359	35	62	1763	69	2012-08-02 11:16:07.256137
38361	3	1-9463	38359	62	41	836	16	2012-08-02 11:16:07.256137
38359	3	1-11814	38359	41	35	2812	59	2012-08-02 11:16:07.256137
38363	4	1-11717	38362	8	88	2820	2	2012-08-02 11:16:07.256137
38364	4	1-9706	38362	88	9	319	41	2012-08-02 11:16:07.256137
38365	4	1-7403	38362	9	41	215	16	2012-08-02 11:16:07.256137
38362	4	1-11814	38362	41	8	917	59	2012-08-02 11:16:07.256137
38367	4	1-10762	38366	78	17	506	57	2012-08-02 11:16:07.256137
38368	4	1-11505	38366	17	94	293	27	2012-08-02 11:16:07.256137
38369	4	1-10729	38366	94	41	218	16	2012-08-02 11:16:07.256137
38366	4	1-11814	38366	41	78	962	59	2012-08-02 11:16:07.256137
38379	4	1-10977	38378	65	72	7242	28	2012-08-02 11:16:08.471378
38380	4	1-9136	38378	72	94	4400	52	2012-08-02 11:16:08.471378
38381	4	1-11723	38378	94	67	5045	98	2012-08-02 11:16:08.471378
38378	4	1-11818	38378	67	65	7154	38	2012-08-02 11:16:08.471378
38408	4	1-11546	38407	76	23	1448	14	2012-08-02 11:16:13.520711
38409	4	1-8098	38407	23	25	434	13	2012-08-02 11:16:13.520711
38410	4	1-11514	38407	25	17	417	32	2012-08-02 11:16:13.520711
38407	4	1-11830	38407	17	76	458	87	2012-08-02 11:16:13.520711
38454	3	1-7861	38453	46	67	153	98	2012-08-02 11:16:20.393545
38455	3	1-11818	38453	67	75	83	38	2012-08-02 11:16:20.393545
38453	3	1-11848	38453	75	46	2225	80	2012-08-02 11:16:20.393545
38457	4	1-9682	38456	28	79	35	10	2012-08-02 11:16:21.553426
38458	4	1-10466	38456	79	74	24	62	2012-08-02 11:16:21.553426
38459	4	1-7973	38456	74	65	25	44	2012-08-02 11:16:21.553426
38456	4	1-11851	38456	65	28	22	63	2012-08-02 11:16:21.553426
38461	4	1-9055	38460	14	10	178	23	2012-08-02 11:16:21.553426
38462	4	1-10479	38460	10	45	184	66	2012-08-02 11:16:21.553426
38463	4	1-9947	38460	45	65	167	44	2012-08-02 11:16:21.553426
38460	4	1-11851	38460	65	14	173	63	2012-08-02 11:16:21.553426
38465	4	1-11136	38464	72	10	1478	23	2012-08-02 11:16:21.553426
38466	4	1-10479	38464	10	45	1564	66	2012-08-02 11:16:21.553426
38467	4	1-9947	38464	45	65	1448	44	2012-08-02 11:16:21.553426
38464	4	1-11851	38464	65	72	1534	63	2012-08-02 11:16:21.553426
38483	2	1-11576	38482	52	92	4080	88	2012-08-02 11:16:25.033725
38482	2	1-11860	38482	92	52	3582	21	2012-08-02 11:16:25.033725
38521	4	1-11678	38520	74	18	2457	99	2012-08-02 11:16:30.836393
38522	4	1-11658	38520	18	25	2487	29	2012-08-02 11:16:30.836393
38523	4	1-11393	38520	25	71	4689	86	2012-08-02 11:16:30.836393
38520	4	1-11874	38520	71	74	4689	46	2012-08-02 11:16:30.836393
38532	4	1-9040	38531	13	61	1228	92	2012-08-02 11:16:34.30558
38533	4	1-11369	38531	61	88	732	27	2012-08-02 11:16:34.30558
38534	4	1-11047	38531	88	80	399	34	2012-08-02 11:16:34.30558
38531	4	1-11884	38531	80	13	1487	14	2012-08-02 11:16:34.30558
38536	4	1-11816	38535	99	56	3763	4	2012-08-02 11:16:35.45468
38537	4	1-11608	38535	56	85	2606	65	2012-08-02 11:16:35.45468
38538	4	1-10387	38535	85	92	2098	14	2012-08-02 11:16:35.45468
38535	4	1-11886	38535	92	99	4890	55	2012-08-02 11:16:35.45468
38540	3	1-9676	38539	39	98	1578	39	2012-08-02 11:16:35.45468
38541	3	1-8721	38539	98	92	1622	14	2012-08-02 11:16:35.45468
38539	3	1-11886	38539	92	39	3837	55	2012-08-02 11:16:35.45468
38547	4	1-8518	38546	39	6	1901	4	2012-08-02 11:16:38.548193
38548	4	1-10747	38546	6	59	2039	100	2012-08-02 11:16:38.548193
38549	4	1-11453	38546	59	4	2530	50	2012-08-02 11:16:38.548193
38546	4	1-11895	38546	4	39	3248	56	2012-08-02 11:16:38.548193
38551	4	1-8518	38550	39	100	227	4	2012-08-02 11:16:38.548193
38552	4	1-10057	38550	100	59	238	100	2012-08-02 11:16:38.548193
38553	4	1-11453	38550	59	4	298	50	2012-08-02 11:16:38.548193
38550	4	1-11895	38550	4	39	385	56	2012-08-02 11:16:38.548193
38585	4	1-10093	38584	29	20	259	27	2012-08-02 11:16:42.724536
38586	4	1-9752	38584	20	29	221	60	2012-08-02 11:16:42.724536
38587	4	1-11545	38584	29	20	68	18	2012-08-02 11:16:42.724536
38584	4	1-11907	38584	20	29	199	11	2012-08-02 11:16:42.724536
38589	4	1-9332	38588	85	77	2742	3	2012-08-02 11:16:42.724536
38590	4	1-8077	38588	77	29	3115	60	2012-08-02 11:16:42.724536
38591	4	1-11545	38588	29	20	1043	18	2012-08-02 11:16:42.724536
38588	4	1-11907	38588	20	85	3320	11	2012-08-02 11:16:42.724536
38593	4	1-10187	38592	59	39	618	10	2012-08-02 11:16:42.724536
38594	4	1-9633	38592	39	29	604	60	2012-08-02 11:16:42.724536
38595	4	1-11545	38592	29	20	206	18	2012-08-02 11:16:42.724536
38592	4	1-11907	38592	20	59	671	11	2012-08-02 11:16:42.724536
38597	4	1-11378	38596	5	96	1230	35	2012-08-02 11:16:42.724536
38598	4	1-11294	38596	96	45	1680	7	2012-08-02 11:16:42.724536
38599	4	1-6895	38596	45	20	407	18	2012-08-02 11:16:42.724536
38596	4	1-11907	38596	20	5	1331	11	2012-08-02 11:16:42.724536
38601	4	1-11243	38600	81	13	2061	84	2012-08-02 11:16:42.724536
38602	4	1-11211	38600	13	63	2541	32	2012-08-02 11:16:42.724536
38603	4	1-5018	38600	63	20	699	18	2012-08-02 11:16:42.724536
38342	4	1-11436	38339	56	27	352	63	2012-08-02 11:16:04.869133
38344	5	1-11614	38343	41	100	6342	89	2012-08-02 11:16:04.869133
38345	5	1-11485	38343	100	13	4582	73	2012-08-02 11:16:04.869133
38346	5	1-11804	38343	13	56	2277	78	2012-08-02 11:16:04.869133
38347	5	1-11436	38343	56	27	1454	63	2012-08-02 11:16:04.869133
38343	5	1-11809	38343	27	41	8126	39	2012-08-02 11:16:04.869133
38371	4	1-9164	38370	85	21	3395	35	2012-08-02 11:16:08.858103
38372	4	1-11742	38370	21	89	6355	6	2012-08-02 11:16:08.858103
38373	4	1-4325	38370	89	33	2384	50	2012-08-02 11:16:08.858103
38370	4	1-11817	38370	33	85	8167	72	2012-08-02 11:16:08.858103
38375	4	1-10030	38374	54	10	607	91	2012-08-02 11:16:08.858103
38376	4	1-11265	38374	10	94	824	37	2012-08-02 11:16:08.858103
38377	4	1-9848	38374	94	33	435	50	2012-08-02 11:16:08.858103
38374	4	1-11817	38374	33	54	1520	72	2012-08-02 11:16:08.858103
38404	4	1-10936	38403	66	31	689	25	2012-08-02 11:16:13.156317
38405	4	1-11331	38403	31	100	658	61	2012-08-02 11:16:13.156317
38406	4	1-11650	38403	100	43	698	99	2012-08-02 11:16:13.156317
38403	4	1-11829	38403	43	66	548	56	2012-08-02 11:16:13.156317
38426	4	1-11612	38425	80	33	1007	80	2012-08-02 11:16:15.432227
38427	4	1-11468	38425	33	33	912	4	2012-08-02 11:16:15.432227
38428	4	1-9843	38425	33	67	870	42	2012-08-02 11:16:15.432227
38425	4	1-11835	38425	67	80	819	32	2012-08-02 11:16:15.432227
38430	2	1-5273	38429	87	75	1390	43	2012-08-02 11:16:16.582227
38429	2	1-11838	38429	75	87	9189	30	2012-08-02 11:16:16.582227
38439	4	1-11769	38438	77	21	1005	35	2012-08-02 11:16:17.708311
38440	4	1-11742	38438	21	22	1626	6	2012-08-02 11:16:17.708311
38441	4	1-7364	38438	22	3	847	58	2012-08-02 11:16:17.708311
38438	4	1-11841	38438	3	77	1208	93	2012-08-02 11:16:17.708311
38443	4	1-11745	38442	83	28	1956	51	2012-08-02 11:16:17.708311
38444	4	1-11573	38442	28	27	3172	91	2012-08-02 11:16:17.708311
38445	4	1-595	38442	27	3	1676	58	2012-08-02 11:16:17.708311
38442	4	1-11841	38442	3	83	2599	93	2012-08-02 11:16:17.708311
38469	3	1-11097	38468	73	80	627	86	2012-08-02 11:16:21.940122
38470	3	1-10443	38468	80	83	38	21	2012-08-02 11:16:21.940122
38468	3	1-11852	38468	83	73	5069	14	2012-08-02 11:16:21.940122
38472	4	1-9910	38471	48	36	432	9	2012-08-02 11:16:21.940122
38473	4	1-8476	38471	36	90	96	71	2012-08-02 11:16:21.940122
38474	4	1-10022	38471	90	83	17	21	2012-08-02 11:16:21.940122
38471	4	1-11852	38471	83	48	3691	14	2012-08-02 11:16:21.940122
38475	1	1-4082	\N	74	74	841	71	2012-08-02 11:16:21.940122
38479	4	1-11494	38478	11	63	4425	40	2012-08-02 11:16:24.271335
38480	4	1-11822	38478	63	64	3211	49	2012-08-02 11:16:24.271335
38481	4	1-9373	38478	64	35	1076	74	2012-08-02 11:16:24.271335
38478	4	1-11858	38478	35	11	5456	93	2012-08-02 11:16:24.271335
38501	4	1-9600	38500	84	28	289	51	2012-08-02 11:16:28.151504
38502	4	1-11573	38500	28	10	398	91	2012-08-02 11:16:28.151504
38503	4	1-11265	38500	10	23	422	37	2012-08-02 11:16:28.151504
38500	4	1-11868	38500	23	84	550	57	2012-08-02 11:16:28.151504
38543	4	1-10836	38542	93	52	1275	66	2012-08-02 11:16:36.990477
38544	4	1-10320	38542	52	60	1410	88	2012-08-02 11:16:36.990477
38545	4	1-11556	38542	60	5	1597	95	2012-08-02 11:16:36.990477
38542	4	1-11891	38542	5	93	1580	47	2012-08-02 11:16:36.990477
38571	2	1-8939	38570	34	51	1337	38	2012-08-02 11:16:41.575506
38570	2	1-11904	38570	51	34	7262	62	2012-08-02 11:16:41.575506
38573	4	1-9616	38572	94	36	1207	77	2012-08-02 11:16:41.575506
38574	4	1-11261	38572	36	63	350	55	2012-08-02 11:16:41.575506
38575	4	1-8343	38572	63	51	258	38	2012-08-02 11:16:41.575506
38572	4	1-11904	38572	51	94	1443	62	2012-08-02 11:16:41.575506
38577	4	1-11719	38576	96	72	435	88	2012-08-02 11:16:41.575506
38578	4	1-10921	38576	72	63	191	55	2012-08-02 11:16:41.575506
38579	4	1-8343	38576	63	51	166	38	2012-08-02 11:16:41.575506
38576	4	1-11904	38576	51	96	1095	62	2012-08-02 11:16:41.575506
38581	4	1-11183	38580	96	36	314	9	2012-08-02 11:16:43.111969
38582	4	1-8476	38580	36	94	252	71	2012-08-02 11:16:43.111969
38583	4	1-11628	38580	94	100	223	80	2012-08-02 11:16:43.111969
38580	4	1-11906	38580	100	96	440	70	2012-08-02 11:16:43.111969
38496	4	1-7752	38495	77	36	388	70	2012-08-02 11:16:26.593607
38497	4	1-11087	38495	36	60	158	88	2012-08-02 11:16:26.593607
38498	4	1-11556	38495	60	70	206	95	2012-08-02 11:16:26.593607
38495	4	1-11864	38495	70	77	464	43	2012-08-02 11:16:26.593607
38499	1	1-3242	\N	61	61	696	51	2012-08-02 11:16:26.593607
38505	4	1-11146	38504	38	61	381	82	2012-08-02 11:16:28.924881
38506	4	1-11831	38504	61	89	190	29	2012-08-02 11:16:28.924881
38507	4	1-10488	38504	89	88	93	67	2012-08-02 11:16:28.924881
38504	4	1-11870	38504	88	38	437	62	2012-08-02 11:16:28.924881
38509	4	1-11146	38508	38	61	345	82	2012-08-02 11:16:28.924881
38510	4	1-11831	38508	61	72	172	29	2012-08-02 11:16:28.924881
38511	4	1-11205	38508	72	88	84	67	2012-08-02 11:16:28.924881
38508	4	1-11870	38508	88	38	396	62	2012-08-02 11:16:28.924881
38513	4	1-9154	38512	89	89	1305	42	2012-08-02 11:16:28.924881
38514	4	1-11837	38512	89	10	1038	64	2012-08-02 11:16:28.924881
38515	4	1-10631	38512	10	88	1119	67	2012-08-02 11:16:28.924881
38512	4	1-11870	38512	88	89	7348	62	2012-08-02 11:16:28.924881
38517	4	1-11479	38516	92	53	477	59	2012-08-02 11:16:30.07419
38518	4	1-8236	38516	53	56	265	71	2012-08-02 11:16:30.07419
38519	4	1-7140	38516	56	84	135	100	2012-08-02 11:16:30.07419
38516	4	1-11873	38516	84	92	235	75	2012-08-02 11:16:30.07419
38525	3	1-10522	38524	32	61	712	3	2012-08-02 11:16:33.134849
38526	3	1-10593	38524	61	65	657	13	2012-08-02 11:16:33.134849
38524	3	1-11880	38524	65	32	1016	51	2012-08-02 11:16:33.134849
38528	4	1-11806	38527	15	86	2456	69	2012-08-02 11:16:33.918993
38529	4	1-4235	38527	86	83	1118	93	2012-08-02 11:16:33.918993
38530	4	1-11740	38527	83	48	1165	44	2012-08-02 11:16:33.918993
38527	4	1-11883	38527	48	15	1361	57	2012-08-02 11:16:33.918993
38555	2	1-11658	38554	18	66	3767	29	2012-08-02 11:16:40.051362
38554	2	1-11899	38554	66	18	4513	99	2012-08-02 11:16:40.051362
38557	3	1-10341	38556	67	61	353	82	2012-08-02 11:16:40.051362
38558	3	1-11831	38556	61	66	274	29	2012-08-02 11:16:40.051362
38556	3	1-11899	38556	66	67	435	99	2012-08-02 11:16:40.051362
38560	3	1-11592	38559	81	83	1226	81	2012-08-02 11:16:41.199927
38561	3	1-10764	38559	83	5	603	65	2012-08-02 11:16:41.199927
38559	3	1-11902	38559	5	81	1003	3	2012-08-02 11:16:41.199927
38563	4	1-11592	38562	81	86	6033	81	2012-08-02 11:16:41.199927
38564	4	1-11857	38562	86	100	4711	85	2012-08-02 11:16:41.199927
38565	4	1-11164	38562	100	5	2950	65	2012-08-02 11:16:41.199927
38562	4	1-11902	38562	5	81	4922	3	2012-08-02 11:16:41.199927
38567	4	1-9133	38566	10	90	4260	52	2012-08-02 11:16:41.199927
38568	4	1-10172	38566	90	41	1909	42	2012-08-02 11:16:41.199927
38569	4	1-11689	38566	41	5	1689	65	2012-08-02 11:16:41.199927
38566	4	1-11902	38566	5	10	3068	3	2012-08-02 11:16:41.199927
38600	4	1-11907	38600	20	81	2339	11	2012-08-02 11:16:42.724536
38605	4	1-11738	38604	55	70	7561	19	2012-08-02 11:16:43.499661
38606	4	1-8223	38604	70	89	2061	73	2012-08-02 11:16:43.499661
38607	4	1-11192	38604	89	30	2091	34	2012-08-02 11:16:43.499661
38604	4	1-11908	38604	30	55	4836	36	2012-08-02 11:16:43.499661
38609	4	1-11714	38608	87	26	128	25	2012-08-02 11:16:43.499661
38610	4	1-11713	38608	26	64	151	76	2012-08-02 11:16:43.499661
38611	4	1-10843	38608	64	30	64	34	2012-08-02 11:16:43.499661
38608	4	1-11908	38608	30	87	152	36	2012-08-02 11:16:43.499661
38613	4	1-10027	38612	62	84	1061	57	2012-08-02 11:16:43.499661
38614	4	1-9600	38612	84	48	587	51	2012-08-02 11:16:43.499661
38615	4	1-11450	38612	48	30	530	34	2012-08-02 11:16:43.499661
38612	4	1-11908	38612	30	62	1371	36	2012-08-02 11:16:43.499661
38617	4	1-10027	38616	62	17	754	57	2012-08-02 11:16:43.499661
38618	4	1-11505	38616	17	88	506	27	2012-08-02 11:16:43.499661
38619	4	1-11047	38616	88	30	345	34	2012-08-02 11:16:43.499661
38616	4	1-11908	38616	30	62	933	36	2012-08-02 11:16:43.499661
38621	4	1-10484	38620	38	5	246	95	2012-08-02 11:16:43.499661
38622	4	1-11891	38620	5	84	258	47	2012-08-02 11:16:43.499661
38623	4	1-11042	38620	84	30	86	34	2012-08-02 11:16:43.499661
38620	4	1-11908	38620	30	38	244	36	2012-08-02 11:16:43.499661
38625	4	1-6392	38624	80	85	1698	38	2012-08-02 11:16:43.897895
38626	4	1-2873	38624	85	70	617	5	2012-08-02 11:16:43.897895
38627	4	1-10888	38624	70	18	270	94	2012-08-02 11:16:43.897895
38624	4	1-11909	38624	18	80	5909	71	2012-08-02 11:16:43.897895
38629	4	1-10744	38628	14	69	824	62	2012-08-02 11:16:43.897895
38630	4	1-3106	38628	69	70	211	5	2012-08-02 11:16:43.897895
38631	4	1-10888	38628	70	18	98	94	2012-08-02 11:16:43.897895
38628	4	1-11909	38628	18	14	2287	71	2012-08-02 11:16:43.897895
38632	1	1-10888	\N	70	70	4386	94	2012-08-02 11:16:43.897895
38634	3	1-10631	38633	10	28	79	67	2012-08-02 11:16:45.897988
38635	3	1-11756	38633	28	55	189	87	2012-08-02 11:16:45.897988
38633	3	1-11914	38633	55	10	88	64	2012-08-02 11:16:45.897988
38637	2	1-10838	38636	33	55	978	87	2012-08-02 11:16:45.897988
38636	2	1-11914	38636	55	33	532	64	2012-08-02 11:16:45.897988
38639	4	1-11291	38638	51	29	467	76	2012-08-02 11:16:47.046968
38640	4	1-10934	38638	29	18	180	65	2012-08-02 11:16:47.046968
38641	4	1-5668	38638	18	81	79	10	2012-08-02 11:16:47.046968
38638	4	1-11917	38638	81	51	794	89	2012-08-02 11:16:47.046968
38643	4	1-6859	38642	57	58	1279	94	2012-08-02 11:16:47.046968
38644	4	1-11579	38642	58	28	502	63	2012-08-02 11:16:47.046968
38645	4	1-9682	38642	28	81	552	10	2012-08-02 11:16:47.046968
38642	4	1-11917	38642	81	57	5553	89	2012-08-02 11:16:47.046968
38647	3	1-11672	38646	84	84	277	100	2012-08-02 11:16:47.489209
38648	3	1-11873	38646	84	90	379	75	2012-08-02 11:16:47.489209
38646	3	1-11918	38646	90	84	1849	8	2012-08-02 11:16:47.489209
38650	4	1-11815	38649	44	100	511	4	2012-08-02 11:16:47.489209
38651	4	1-10057	38649	100	84	388	100	2012-08-02 11:16:47.489209
38652	4	1-11873	38649	84	90	465	75	2012-08-02 11:16:47.489209
38649	4	1-11918	38649	90	44	2528	8	2012-08-02 11:16:47.489209
38654	4	1-9133	38653	10	22	2721	52	2012-08-02 11:16:47.875544
38655	4	1-11551	38653	22	36	1677	48	2012-08-02 11:16:47.875544
38656	4	1-11386	38653	36	34	2850	7	2012-08-02 11:16:47.875544
38653	4	1-11919	38653	34	10	1823	3	2012-08-02 11:16:47.875544
38658	3	1-4664	38657	32	89	2142	59	2012-08-02 11:16:48.264351
38659	3	1-11209	38657	89	77	1070	30	2012-08-02 11:16:48.264351
38657	3	1-11920	38657	77	32	3161	24	2012-08-02 11:16:48.264351
38661	4	1-11725	38660	55	79	563	25	2012-08-02 11:16:48.264351
38662	4	1-11530	38660	79	99	714	82	2012-08-02 11:16:48.264351
38663	4	1-3930	38660	99	77	91	30	2012-08-02 11:16:48.264351
38660	4	1-11920	38660	77	55	302	24	2012-08-02 11:16:48.264351
38665	4	1-3888	38664	41	43	1152	51	2012-08-02 11:16:48.264351
38666	4	1-11563	38664	43	97	1649	41	2012-08-02 11:16:48.264351
38667	4	1-10792	38664	97	77	829	30	2012-08-02 11:16:48.264351
38664	4	1-11920	38664	77	41	2834	24	2012-08-02 11:16:48.264351
38669	4	1-11725	38668	55	2	685	25	2012-08-02 11:16:48.264351
38670	4	1-5494	38668	2	89	95	90	2012-08-02 11:16:48.264351
38671	4	1-10992	38668	89	77	78	30	2012-08-02 11:16:48.264351
38668	4	1-11920	38668	77	55	308	24	2012-08-02 11:16:48.264351
38673	4	1-11725	38672	55	50	2552	25	2012-08-02 11:16:48.264351
38674	4	1-11564	38672	50	19	1644	98	2012-08-02 11:16:48.264351
38675	4	1-8642	38672	19	77	270	30	2012-08-02 11:16:48.264351
38672	4	1-11920	38672	77	55	1105	24	2012-08-02 11:16:48.264351
38677	4	1-7183	38676	87	32	263	22	2012-08-02 11:16:48.264351
38678	4	1-11790	38676	32	66	270	68	2012-08-02 11:16:48.264351
38679	4	1-9647	38676	66	77	203	30	2012-08-02 11:16:48.264351
38676	4	1-11920	38676	77	87	838	24	2012-08-02 11:16:48.264351
38681	4	1-9676	38680	39	45	401	39	2012-08-02 11:16:48.67307
38682	4	1-11751	38680	45	42	1405	56	2012-08-02 11:16:48.67307
38683	4	1-10871	38680	42	54	973	44	2012-08-02 11:16:48.67307
38680	4	1-11921	38680	54	39	862	55	2012-08-02 11:16:48.67307
38685	4	1-11230	38684	57	21	5312	58	2012-08-02 11:16:49.059592
38686	4	1-11435	38684	21	95	4270	38	2012-08-02 11:16:49.059592
38687	4	1-9941	38684	95	36	2561	17	2012-08-02 11:16:49.059592
38684	4	1-11922	38684	36	57	6821	56	2012-08-02 11:16:49.059592
38689	4	1-10168	38688	94	8	1751	80	2012-08-02 11:16:49.059592
38690	4	1-11772	38688	8	76	1911	53	2012-08-02 11:16:49.059592
38691	4	1-10354	38688	76	36	1024	17	2012-08-02 11:16:49.059592
38688	4	1-11922	38688	36	94	3035	56	2012-08-02 11:16:49.059592
38693	3	1-5313	38692	78	100	441	4	2012-08-02 11:16:52.97174
38694	3	1-10057	38692	100	21	216	100	2012-08-02 11:16:52.97174
38692	3	1-11932	38692	21	78	3186	82	2012-08-02 11:16:52.97174
38696	3	1-11301	38695	88	91	930	45	2012-08-02 11:16:52.97174
38697	3	1-9422	38695	91	21	209	100	2012-08-02 11:16:52.97174
38695	3	1-11932	38695	21	88	3424	82	2012-08-02 11:16:52.97174
38699	4	1-6945	38698	64	56	1440	81	2012-08-02 11:16:54.120655
38700	4	1-11688	38698	56	96	2364	62	2012-08-02 11:16:54.120655
38701	4	1-11719	38698	96	74	967	88	2012-08-02 11:16:54.120655
38698	4	1-11935	38698	74	64	1535	41	2012-08-02 11:16:54.120655
38703	4	1-5916	38702	86	50	991	8	2012-08-02 11:16:54.120655
38704	4	1-11512	38702	50	91	1109	24	2012-08-02 11:16:54.120655
38705	4	1-11322	38702	91	74	1706	88	2012-08-02 11:16:54.120655
38702	4	1-11935	38702	74	86	2920	41	2012-08-02 11:16:54.120655
38707	2	1-5276	38706	51	74	504	88	2012-08-02 11:16:54.120655
38706	2	1-11935	38706	74	51	865	41	2012-08-02 11:16:54.120655
38736	4	1-11813	38735	9	30	3946	1	2012-08-02 11:16:55.302803
38737	4	1-4742	38735	30	65	1768	13	2012-08-02 11:16:55.302803
38738	4	1-11880	38735	65	85	2119	51	2012-08-02 11:16:55.302803
38735	4	1-11938	38735	85	9	4953	18	2012-08-02 11:16:55.302803
38740	4	1-10891	38739	99	96	657	97	2012-08-02 11:16:55.302803
38741	4	1-4356	38739	96	65	281	13	2012-08-02 11:16:55.302803
38742	4	1-11880	38739	65	85	370	51	2012-08-02 11:16:55.302803
38739	4	1-11938	38739	85	99	950	18	2012-08-02 11:16:55.302803
38744	4	1-10891	38743	99	15	508	97	2012-08-02 11:16:55.302803
38745	4	1-4563	38743	15	65	178	13	2012-08-02 11:16:55.302803
38746	4	1-11880	38743	65	85	250	51	2012-08-02 11:16:55.302803
38743	4	1-11938	38743	85	99	686	18	2012-08-02 11:16:55.302803
38781	4	1-11614	38780	41	39	1631	89	2012-08-02 11:17:01.225118
38782	4	1-8364	38780	39	39	864	17	2012-08-02 11:17:01.225118
38783	4	1-11059	38780	39	78	566	41	2012-08-02 11:17:01.225118
38780	4	1-11953	38780	78	41	2517	39	2012-08-02 11:17:01.225118
38785	3	1-10540	38784	53	65	3549	64	2012-08-02 11:17:01.225118
38786	3	1-9009	38784	65	78	1395	41	2012-08-02 11:17:01.225118
38784	3	1-11953	38784	78	53	6803	39	2012-08-02 11:17:01.225118
38818	4	1-11831	38817	61	93	2262	29	2012-08-02 11:17:08.417874
38819	4	1-11508	38817	93	3	2857	23	2012-08-02 11:17:08.417874
38820	4	1-11696	38817	3	38	2419	50	2012-08-02 11:17:08.417874
38817	4	1-11971	38817	38	61	3820	82	2012-08-02 11:17:08.417874
38822	4	1-11831	38821	61	64	380	29	2012-08-02 11:17:08.417874
38823	4	1-11179	38821	64	59	236	27	2012-08-02 11:17:08.417874
38824	4	1-11905	38821	59	38	259	50	2012-08-02 11:17:08.417874
38821	4	1-11971	38821	38	61	512	82	2012-08-02 11:17:08.417874
38826	2	1-7851	38825	57	23	4634	2	2012-08-02 11:17:09.59992
38825	2	1-11974	38825	23	57	3850	84	2012-08-02 11:17:09.59992
38828	3	1-11211	38827	13	56	1772	32	2012-08-02 11:17:09.59992
38829	3	1-5899	38827	56	23	1826	2	2012-08-02 11:17:09.59992
38827	3	1-11974	38827	23	13	1522	84	2012-08-02 11:17:09.59992
38831	3	1-8135	38830	99	70	953	66	2012-08-02 11:17:11.886768
38832	3	1-11407	38830	70	12	759	69	2012-08-02 11:17:11.886768
38830	3	1-11980	38830	12	99	1079	24	2012-08-02 11:17:11.886768
38852	4	1-11905	38851	59	39	3614	50	2012-08-02 11:17:14.297599
38853	4	1-5316	38851	39	7	1367	17	2012-08-02 11:17:14.297599
38854	4	1-11481	38851	7	88	1827	91	2012-08-02 11:17:14.297599
38851	4	1-11986	38851	88	59	3275	27	2012-08-02 11:17:14.297599
38856	4	1-10405	38855	19	34	962	81	2012-08-02 11:17:14.297599
38857	4	1-11574	38855	34	16	1131	7	2012-08-02 11:17:14.297599
38858	4	1-10685	38855	16	88	644	91	2012-08-02 11:17:14.297599
38855	4	1-11986	38855	88	19	1170	27	2012-08-02 11:17:14.297599
38886	5	1-11711	38885	67	100	1734	61	2012-08-02 11:17:20.959129
38887	5	1-11650	38885	100	72	1614	99	2012-08-02 11:17:20.959129
38888	5	1-7273	38885	72	19	703	19	2012-08-02 11:17:20.959129
38889	5	1-11652	38885	19	72	1138	26	2012-08-02 11:17:20.959129
38885	5	1-12003	38885	72	67	792	33	2012-08-02 11:17:20.959129
38981	3	1-11158	38980	71	74	3630	54	2012-08-02 11:17:35.819311
38982	3	1-11099	38980	74	50	2318	7	2012-08-02 11:17:35.819311
38980	3	1-12040	38980	50	71	3792	30	2012-08-02 11:17:35.819311
39003	4	1-11969	39002	92	22	1442	70	2012-08-02 11:17:38.97931
39004	4	1-11998	39002	22	32	685	84	2012-08-02 11:17:38.97931
39005	4	1-10568	39002	32	20	337	50	2012-08-02 11:17:38.97931
39002	4	1-12049	39002	20	92	1187	66	2012-08-02 11:17:38.97931
39007	4	1-11954	39006	72	91	6696	3	2012-08-02 11:17:40.198824
39008	4	1-11716	39006	91	25	4155	64	2012-08-02 11:17:40.198824
39009	4	1-9509	39006	25	35	2184	27	2012-08-02 11:17:40.198824
39006	4	1-12052	39006	35	72	6435	51	2012-08-02 11:17:40.198824
39011	4	1-8549	39010	69	23	982	37	2012-08-02 11:17:40.198824
39012	4	1-11868	39010	23	17	1278	57	2012-08-02 11:17:40.198824
39013	4	1-11505	39010	17	35	776	27	2012-08-02 11:17:40.198824
39010	4	1-12052	39010	35	69	2545	51	2012-08-02 11:17:40.198824
39030	4	1-8453	39029	79	84	2319	100	2012-08-02 11:17:41.409799
39031	4	1-11873	39029	84	34	3577	75	2012-08-02 11:17:41.409799
39032	4	1-5253	39029	34	23	762	8	2012-08-02 11:17:41.409799
39029	4	1-12055	39029	23	79	1884	26	2012-08-02 11:17:41.409799
39034	4	1-11187	39033	16	38	168	95	2012-08-02 11:17:41.409799
39035	4	1-12053	39033	38	51	192	64	2012-08-02 11:17:41.409799
39036	4	1-5233	39033	51	23	50	8	2012-08-02 11:17:41.409799
39033	4	1-12055	39033	23	16	126	26	2012-08-02 11:17:41.409799
39038	4	1-12003	39037	72	55	435	33	2012-08-02 11:17:41.409799
39039	4	1-11554	39037	55	86	584	41	2012-08-02 11:17:41.409799
39040	4	1-5916	39037	86	23	212	8	2012-08-02 11:17:41.409799
39037	4	1-12055	39037	23	72	588	26	2012-08-02 11:17:41.409799
39054	4	1-11957	39053	81	86	5429	22	2012-08-02 11:17:44.152642
39055	4	1-11588	39053	86	15	1974	31	2012-08-02 11:17:44.152642
39056	4	1-11095	39053	15	28	1369	21	2012-08-02 11:17:44.152642
39053	4	1-12062	39053	28	81	4418	95	2012-08-02 11:17:44.152642
39058	4	1-12022	39057	19	7	1304	79	2012-08-02 11:17:44.152642
39059	4	1-11075	39057	7	94	557	100	2012-08-02 11:17:44.152642
39060	4	1-10806	39057	94	28	550	21	2012-08-02 11:17:44.152642
39057	4	1-12062	39057	28	19	1796	95	2012-08-02 11:17:44.152642
39061	1	1-3823	\N	46	46	1922	100	2012-08-02 11:17:44.152642
39067	2	1-11820	39066	42	64	1068	4	2012-08-02 11:17:45.313023
39066	2	1-12066	39066	64	42	4935	74	2012-08-02 11:17:45.313023
39073	4	1-9225	39072	4	100	1159	80	2012-08-02 11:17:49.124408
39074	4	1-11906	39072	100	97	2260	70	2012-08-02 11:17:49.124408
39075	4	1-11734	39072	97	10	1598	20	2012-08-02 11:17:49.124408
39072	4	1-12075	39072	10	4	1286	55	2012-08-02 11:17:49.124408
39080	4	1-12056	39079	46	72	1627	6	2012-08-02 11:17:50.284477
39081	4	1-11596	39079	72	48	2001	92	2012-08-02 11:17:50.284477
39082	4	1-11589	39079	48	28	1437	49	2012-08-02 11:17:50.284477
39079	4	1-12078	39079	28	46	1512	38	2012-08-02 11:17:50.284477
39084	4	1-11389	39083	61	46	2725	46	2012-08-02 11:17:52.549216
39085	4	1-11866	39083	46	47	2601	20	2012-08-02 11:17:52.549216
39086	4	1-11582	39083	47	1	2977	40	2012-08-02 11:17:52.549216
39083	4	1-12084	39083	1	61	4264	88	2012-08-02 11:17:52.549216
39088	4	1-11860	39087	92	13	374	21	2012-08-02 11:17:52.549216
39089	4	1-11222	39087	13	91	334	6	2012-08-02 11:17:52.549216
39090	4	1-7374	39087	91	1	293	40	2012-08-02 11:17:52.549216
39087	4	1-12084	39087	1	92	427	88	2012-08-02 11:17:52.549216
38709	4	1-10641	38708	100	4	385	50	2012-08-02 11:16:54.883231
38710	4	1-11895	38708	4	84	438	56	2012-08-02 11:16:54.883231
38711	4	1-10529	38708	84	2	401	97	2012-08-02 11:16:54.883231
38708	4	1-11936	38708	2	100	1641	69	2012-08-02 11:16:54.883231
38713	4	1-8831	38712	19	40	2662	70	2012-08-02 11:16:54.883231
38714	4	1-11636	38712	40	33	1347	90	2012-08-02 11:16:54.883231
38715	4	1-7964	38712	33	2	774	97	2012-08-02 11:16:54.883231
38712	4	1-11936	38712	2	19	3467	69	2012-08-02 11:16:54.883231
38748	3	1-8910	38747	80	56	86	42	2012-08-02 11:16:56.916231
38749	3	1-10598	38747	56	21	15	37	2012-08-02 11:16:56.916231
38747	3	1-11942	38747	21	80	709	35	2012-08-02 11:16:56.916231
38769	4	1-10992	38768	89	71	1649	30	2012-08-02 11:17:00.418404
38770	4	1-11158	38768	71	19	1612	54	2012-08-02 11:17:00.418404
38771	4	1-11669	38768	19	48	1322	32	2012-08-02 11:17:00.418404
38768	4	1-11951	38768	48	89	1897	90	2012-08-02 11:17:00.418404
38773	4	1-9957	38772	35	84	1125	100	2012-08-02 11:17:00.418404
38774	4	1-11873	38772	84	22	1823	75	2012-08-02 11:17:00.418404
38775	4	1-11437	38772	22	48	2666	32	2012-08-02 11:17:00.418404
38772	4	1-11951	38772	48	35	4075	90	2012-08-02 11:17:00.418404
38777	3	1-6652	38776	3	19	1185	54	2012-08-02 11:17:00.418404
38778	3	1-11669	38776	19	48	1091	32	2012-08-02 11:17:00.418404
38776	3	1-11951	38776	48	3	1760	90	2012-08-02 11:17:00.418404
38779	1	1-9623	\N	52	52	366	90	2012-08-02 11:17:00.418404
38788	3	1-10593	38787	61	65	791	13	2012-08-02 11:17:01.624068
38789	3	1-11880	38787	65	72	964	51	2012-08-02 11:17:01.624068
38787	3	1-11954	38787	72	61	1089	3	2012-08-02 11:17:01.624068
38791	3	1-8889	38790	90	83	1091	93	2012-08-02 11:17:01.624068
38792	3	1-11745	38790	83	72	869	51	2012-08-02 11:17:01.624068
38790	3	1-11954	38790	72	90	1257	3	2012-08-02 11:17:01.624068
38793	1	1-3265	\N	81	81	587	3	2012-08-02 11:17:01.624068
38795	3	1-11790	38794	32	61	2215	68	2012-08-02 11:17:02.816395
38796	3	1-9192	38794	61	81	1178	95	2012-08-02 11:17:02.816395
38794	3	1-11957	38794	81	32	1977	22	2012-08-02 11:17:02.816395
38798	4	1-11709	38797	16	100	4408	85	2012-08-02 11:17:02.816395
38799	4	1-11164	38797	100	24	3518	65	2012-08-02 11:17:02.816395
38800	4	1-10615	38797	24	81	1442	95	2012-08-02 11:17:02.816395
38797	4	1-11957	38797	81	16	2442	22	2012-08-02 11:17:02.816395
38806	4	1-11601	38805	72	60	238	10	2012-08-02 11:17:03.998477
38807	4	1-11762	38805	60	41	165	96	2012-08-02 11:17:03.998477
38808	4	1-11958	38805	41	53	1508	5	2012-08-02 11:17:03.998477
38805	4	1-11960	38805	53	72	890	28	2012-08-02 11:17:03.998477
38834	4	1-11209	38833	89	21	168	30	2012-08-02 11:17:12.67162
38835	4	1-11150	38833	21	33	157	40	2012-08-02 11:17:12.67162
38836	4	1-10776	38833	33	58	103	72	2012-08-02 11:17:12.67162
38833	4	1-11982	38833	58	89	240	59	2012-08-02 11:17:12.67162
38838	4	1-11972	38837	88	22	60	1	2012-08-02 11:17:12.67162
38839	4	1-10763	38837	22	98	35	24	2012-08-02 11:17:12.67162
38840	4	1-11839	38837	98	58	40	72	2012-08-02 11:17:12.67162
38837	4	1-11982	38837	58	88	94	59	2012-08-02 11:17:12.67162
38842	3	1-9637	38841	52	3	1470	8	2012-08-02 11:17:12.67162
38843	3	1-11776	38841	3	58	2303	72	2012-08-02 11:17:12.67162
38841	3	1-11982	38841	58	52	5495	59	2012-08-02 11:17:12.67162
38873	4	1-11769	38872	77	47	534	35	2012-08-02 11:17:19.39034
38874	4	1-5333	38872	47	64	188	49	2012-08-02 11:17:19.39034
38875	4	1-9373	38872	64	52	48	74	2012-08-02 11:17:19.39034
38872	4	1-11999	38872	52	77	795	93	2012-08-02 11:17:19.39034
38877	3	1-9441	38876	84	12	3660	96	2012-08-02 11:17:19.39034
38878	3	1-6620	38876	12	52	422	74	2012-08-02 11:17:19.39034
38876	3	1-11999	38876	52	84	7156	93	2012-08-02 11:17:19.39034
38880	2	1-11303	38879	53	98	2030	12	2012-08-02 11:17:20.561531
38879	2	1-12002	38879	98	53	1470	49	2012-08-02 11:17:20.561531
38882	4	1-9518	38881	8	47	4243	68	2012-08-02 11:17:20.561531
38883	4	1-7215	38881	47	36	2532	31	2012-08-02 11:17:20.561531
38884	4	1-11643	38881	36	98	5732	12	2012-08-02 11:17:20.561531
38881	4	1-12002	38881	98	8	4806	49	2012-08-02 11:17:20.561531
38907	4	1-10897	38906	34	29	3111	60	2012-08-02 11:17:24.129968
38908	4	1-11545	38906	29	5	1187	18	2012-08-02 11:17:24.129968
38909	4	1-10500	38906	5	68	740	10	2012-08-02 11:17:24.129968
38906	4	1-12011	38906	68	34	1676	79	2012-08-02 11:17:24.129968
38915	4	1-10032	38914	44	14	4539	90	2012-08-02 11:17:25.334523
38916	4	1-10459	38914	14	70	3053	98	2012-08-02 11:17:25.334523
38917	4	1-11693	38914	70	21	3416	46	2012-08-02 11:17:25.334523
38914	4	1-12014	38914	21	44	5620	3	2012-08-02 11:17:25.334523
38919	4	1-8889	38918	90	11	1178	93	2012-08-02 11:17:25.334523
38920	4	1-9763	38918	11	49	742	73	2012-08-02 11:17:25.334523
38921	4	1-9717	38918	49	21	774	46	2012-08-02 11:17:25.334523
38918	4	1-12014	38918	21	90	1377	3	2012-08-02 11:17:25.334523
38923	4	1-8864	38922	21	52	999	91	2012-08-02 11:17:25.334523
38924	4	1-11323	38922	52	96	848	99	2012-08-02 11:17:25.334523
38925	4	1-9598	38922	96	21	859	46	2012-08-02 11:17:25.334523
38922	4	1-12014	38922	21	21	1553	3	2012-08-02 11:17:25.334523
38927	4	1-9791	38926	81	6	386	23	2012-08-02 11:17:25.334523
38928	4	1-10941	38926	6	71	378	86	2012-08-02 11:17:25.334523
38929	4	1-11874	38926	71	21	411	46	2012-08-02 11:17:25.334523
38926	4	1-12014	38926	21	81	758	3	2012-08-02 11:17:25.334523
38935	3	1-11705	38934	77	64	1635	87	2012-08-02 11:17:26.505718
38936	3	1-11548	38934	64	53	777	59	2012-08-02 11:17:26.505718
38934	3	1-12017	38934	53	77	1538	34	2012-08-02 11:17:26.505718
38938	4	1-11991	38937	46	91	1169	45	2012-08-02 11:17:26.505718
38939	4	1-9422	38937	91	37	497	100	2012-08-02 11:17:26.505718
38940	4	1-10894	38937	37	53	333	59	2012-08-02 11:17:26.505718
38937	4	1-12017	38937	53	46	1034	34	2012-08-02 11:17:26.505718
38942	3	1-11106	38941	2	29	1204	20	2012-08-02 11:17:26.505718
38943	3	1-8931	38941	29	53	459	59	2012-08-02 11:17:26.505718
38941	3	1-12017	38941	53	2	1496	34	2012-08-02 11:17:26.505718
38945	4	1-7364	38944	22	3	1715	58	2012-08-02 11:17:28.826119
38946	4	1-11841	38944	3	83	2697	93	2012-08-02 11:17:28.826119
38947	4	1-11740	38944	83	46	2784	44	2012-08-02 11:17:28.826119
38944	4	1-12023	38944	46	22	2989	6	2012-08-02 11:17:28.826119
38949	4	1-11746	38948	45	15	5490	22	2012-08-02 11:17:29.997158
38950	4	1-10053	38948	15	71	2801	23	2012-08-02 11:17:29.997158
38951	4	1-11240	38948	71	59	1127	63	2012-08-02 11:17:29.997158
38948	4	1-12028	38948	59	45	5490	28	2012-08-02 11:17:29.997158
38969	3	1-11756	38968	28	97	1169	87	2012-08-02 11:17:34.261328
38968	3	1-12037	38968	72	28	527	67	2012-08-02 11:17:34.261328
38717	4	1-11579	38716	58	95	1906	63	2012-08-02 11:16:54.507499
38718	4	1-11828	38716	95	94	6170	87	2012-08-02 11:16:54.507499
38719	4	1-10970	38716	94	70	4968	77	2012-08-02 11:16:54.507499
38716	4	1-11937	38716	70	58	3709	94	2012-08-02 11:16:54.507499
38721	3	1-11702	38720	16	73	478	24	2012-08-02 11:16:54.507499
38722	3	1-11284	38720	73	70	756	77	2012-08-02 11:16:54.507499
38720	3	1-11937	38720	70	16	651	94	2012-08-02 11:16:54.507499
38724	4	1-11579	38723	58	72	1711	63	2012-08-02 11:16:54.507499
38725	4	1-11136	38723	72	100	1502	23	2012-08-02 11:16:54.507499
38726	4	1-11404	38723	100	70	3091	77	2012-08-02 11:16:54.507499
38723	4	1-11937	38723	70	58	2772	94	2012-08-02 11:16:54.507499
38728	4	1-11657	38727	70	50	60	8	2012-08-02 11:16:54.507499
38729	4	1-11512	38727	50	73	59	24	2012-08-02 11:16:54.507499
38730	4	1-11284	38727	73	70	93	77	2012-08-02 11:16:54.507499
38727	4	1-11937	38727	70	70	87	94	2012-08-02 11:16:54.507499
38732	4	1-11579	38731	58	20	92	63	2012-08-02 11:16:54.507499
38733	4	1-10375	38731	20	73	70	24	2012-08-02 11:16:54.507499
38734	4	1-11284	38731	73	70	131	77	2012-08-02 11:16:54.507499
38731	4	1-11937	38731	70	58	132	94	2012-08-02 11:16:54.507499
38751	2	1-9772	38750	74	10	722	10	2012-08-02 11:16:57.313937
38750	2	1-11943	38750	10	74	7017	94	2012-08-02 11:16:57.313937
38753	3	1-11020	38752	74	37	322	19	2012-08-02 11:16:57.313937
38754	3	1-9037	38752	37	10	175	10	2012-08-02 11:16:57.313937
38752	3	1-11943	38752	10	74	2265	94	2012-08-02 11:16:57.313937
38756	3	1-11940	38755	86	34	697	26	2012-08-02 11:16:59.645095
38757	3	1-195	38755	34	88	154	3	2012-08-02 11:16:59.645095
38755	3	1-11949	38755	88	86	914	78	2012-08-02 11:16:59.645095
38759	5	1-11569	38758	70	22	517	52	2012-08-02 11:16:59.645095
38760	5	1-11551	38758	22	36	237	48	2012-08-02 11:16:59.645095
38761	5	1-11386	38758	36	34	300	7	2012-08-02 11:16:59.645095
38762	5	1-11919	38758	34	88	143	3	2012-08-02 11:16:59.645095
38758	5	1-11949	38758	88	70	849	78	2012-08-02 11:16:59.645095
38764	5	1-11569	38763	70	24	2978	52	2012-08-02 11:16:59.645095
38765	5	1-11279	38763	24	96	1433	35	2012-08-02 11:16:59.645095
38766	5	1-11294	38763	96	34	1495	7	2012-08-02 11:16:59.645095
38767	5	1-11919	38763	34	88	746	3	2012-08-02 11:16:59.645095
38763	5	1-11949	38763	88	70	4659	78	2012-08-02 11:16:59.645095
38802	4	1-9198	38801	75	63	532	40	2012-08-02 11:17:03.202829
38803	4	1-11822	38801	63	73	364	49	2012-08-02 11:17:03.202829
38804	4	1-9328	38801	73	41	82	96	2012-08-02 11:17:03.202829
38801	4	1-11958	38801	41	75	756	5	2012-08-02 11:17:03.202829
38810	3	1-8845	38809	41	8	1268	59	2012-08-02 11:17:06.739018
38811	3	1-11717	38809	8	62	3995	2	2012-08-02 11:17:06.739018
38809	3	1-11966	38809	62	41	2592	11	2012-08-02 11:17:06.739018
38813	3	1-7907	38812	78	65	640	75	2012-08-02 11:17:06.739018
38814	3	1-11679	38812	65	62	1557	2	2012-08-02 11:17:06.739018
38812	3	1-11966	38812	62	78	1041	11	2012-08-02 11:17:06.739018
38815	1	1-2046	\N	22	22	1009	2	2012-08-02 11:17:06.739018
38816	1	1-7907	\N	78	78	3104	75	2012-08-02 11:17:06.739018
38845	4	1-11804	38844	13	56	3236	78	2012-08-02 11:17:13.466838
38846	4	1-11436	38844	56	35	2432	63	2012-08-02 11:17:13.466838
38847	4	1-9065	38844	35	38	2650	69	2012-08-02 11:17:13.466838
38844	4	1-11984	38844	38	13	5535	73	2012-08-02 11:17:13.466838
38849	3	1-11879	38848	66	66	1604	42	2012-08-02 11:17:13.466838
38850	3	1-9839	38848	66	38	589	69	2012-08-02 11:17:13.466838
38848	3	1-11984	38848	38	66	1239	73	2012-08-02 11:17:13.466838
38860	3	1-9029	38859	7	45	549	34	2012-08-02 11:17:15.866266
38861	3	1-11675	38859	45	43	965	2	2012-08-02 11:17:15.866266
38859	3	1-11990	38859	43	7	1581	20	2012-08-02 11:17:15.866266
38863	2	1-9295	38862	96	43	3544	2	2012-08-02 11:17:15.866266
38862	2	1-11990	38862	43	96	6252	20	2012-08-02 11:17:15.866266
38865	4	1-8984	38864	91	47	251	68	2012-08-02 11:17:17.01574
38866	4	1-11300	38864	47	98	250	78	2012-08-02 11:17:17.01574
38867	4	1-6553	38864	98	43	210	88	2012-08-02 11:17:17.01574
38864	4	1-11993	38864	43	91	376	3	2012-08-02 11:17:17.01574
38869	4	1-9791	38868	81	10	1581	23	2012-08-02 11:17:17.01574
38870	4	1-10479	38868	10	52	1578	66	2012-08-02 11:17:17.01574
38871	4	1-10320	38868	52	43	1839	88	2012-08-02 11:17:17.01574
38868	4	1-11993	38868	43	81	3432	3	2012-08-02 11:17:17.01574
38891	4	1-7206	38890	51	41	320	42	2012-08-02 11:17:21.346268
38892	4	1-11689	38890	41	18	333	65	2012-08-02 11:17:21.346268
38893	4	1-5668	38890	18	31	256	10	2012-08-02 11:17:21.346268
38890	4	1-12004	38890	31	51	330	9	2012-08-02 11:17:21.346268
38895	4	1-11897	38894	28	50	892	8	2012-08-02 11:17:21.346268
38896	4	1-11512	38894	50	67	1037	24	2012-08-02 11:17:21.346268
38897	4	1-4885	38894	67	31	1127	10	2012-08-02 11:17:21.346268
38894	4	1-12004	38894	31	28	1459	9	2012-08-02 11:17:21.346268
38899	4	1-11762	38898	60	41	1653	96	2012-08-02 11:17:23.754384
38900	4	1-11958	38898	41	48	6118	5	2012-08-02 11:17:23.754384
38901	4	1-11624	38898	48	2	829	68	2012-08-02 11:17:23.754384
38898	4	1-12010	38898	2	60	5866	10	2012-08-02 11:17:23.754384
38903	4	1-12004	38902	31	46	143	9	2012-08-02 11:17:23.754384
38904	4	1-7157	38902	46	11	21	24	2012-08-02 11:17:23.754384
38905	4	1-12007	38902	11	2	28	68	2012-08-02 11:17:23.754384
38902	4	1-12010	38902	2	31	249	10	2012-08-02 11:17:23.754384
38911	4	1-9941	38910	95	7	1599	17	2012-08-02 11:17:24.936746
38912	4	1-11481	38910	7	55	2116	91	2012-08-02 11:17:24.936746
38913	4	1-11311	38910	55	58	2021	14	2012-08-02 11:17:24.936746
38910	4	1-12013	38910	58	95	2204	38	2012-08-02 11:17:24.936746
38931	4	1-11164	38930	100	13	954	65	2012-08-02 11:17:26.118942
38932	4	1-11103	38930	13	48	745	61	2012-08-02 11:17:26.118942
38933	4	1-9724	38930	48	26	708	42	2012-08-02 11:17:26.118942
38930	4	1-12016	38930	26	100	1261	85	2012-08-02 11:17:26.118942
38953	4	1-11937	38952	70	26	2122	94	2012-08-02 11:17:32.759248
38954	4	1-10606	38952	26	78	1597	79	2012-08-02 11:17:32.759248
38955	4	1-11528	38952	78	1	917	6	2012-08-02 11:17:32.759248
38952	4	1-12033	38952	1	70	2507	77	2012-08-02 11:17:32.759248
38957	4	1-10339	38956	62	55	925	87	2012-08-02 11:17:32.759248
38958	4	1-11914	38956	55	52	547	64	2012-08-02 11:17:32.759248
38959	4	1-11218	38956	52	1	534	6	2012-08-02 11:17:32.759248
38956	4	1-12033	38956	1	62	1694	77	2012-08-02 11:17:32.759248
38961	4	1-11548	38960	64	53	802	59	2012-08-02 11:17:33.896897
38962	4	1-12017	38960	53	85	2253	34	2012-08-02 11:17:33.896897
38960	4	1-12036	38960	38	64	1187	87	2012-08-02 11:17:33.896897
38963	4	1-11455	38960	85	38	927	60	2012-08-02 11:17:33.896897
38965	4	1-11548	38964	64	54	245	59	2012-08-02 11:17:33.896897
38966	4	1-11314	38964	54	49	299	90	2012-08-02 11:17:33.896897
38967	4	1-8072	38964	49	38	207	60	2012-08-02 11:17:33.896897
38964	4	1-12036	38964	38	64	310	87	2012-08-02 11:17:33.896897
39019	3	1-11311	39018	55	47	1128	14	2012-08-02 11:17:41.012387
39020	3	1-12032	39018	47	85	301	18	2012-08-02 11:17:41.012387
39018	3	1-12054	39018	85	55	1549	91	2012-08-02 11:17:41.012387
39022	4	1-8583	39021	44	85	635	65	2012-08-02 11:17:41.012387
39023	4	1-10387	39021	85	47	643	14	2012-08-02 11:17:41.012387
39024	4	1-12032	39021	47	85	205	18	2012-08-02 11:17:41.012387
39021	4	1-12054	39021	85	44	1358	91	2012-08-02 11:17:41.012387
39026	4	1-6033	39025	8	67	1840	42	2012-08-02 11:17:41.012387
39027	4	1-11835	39025	67	63	2190	32	2012-08-02 11:17:41.012387
39028	4	1-5018	39025	63	85	613	18	2012-08-02 11:17:41.012387
39025	4	1-12054	39025	85	8	4084	91	2012-08-02 11:17:41.012387
39069	4	1-8148	39068	88	47	1750	14	2012-08-02 11:17:46.494947
39070	4	1-12032	39068	47	79	584	18	2012-08-02 11:17:46.494947
39071	4	1-11531	39068	79	18	692	81	2012-08-02 11:17:46.494947
39068	4	1-12068	39068	18	88	4801	15	2012-08-02 11:17:46.494947
39077	3	1-12031	39076	96	94	1342	52	2012-08-02 11:17:49.897874
39078	3	1-11723	39076	94	97	1421	98	2012-08-02 11:17:49.897874
39076	3	1-12077	39076	97	96	2104	5	2012-08-02 11:17:49.897874
39127	3	1-10529	39126	84	97	2180	97	2012-08-02 11:17:56.405367
39128	3	1-8146	39126	97	60	1890	54	2012-08-02 11:17:56.405367
39126	3	1-12094	39126	60	84	2210	56	2012-08-02 11:17:56.405367
39130	3	1-11213	39129	54	80	1124	24	2012-08-02 11:17:56.405367
39131	3	1-10861	39129	80	60	1746	54	2012-08-02 11:17:56.405367
39129	3	1-12094	39129	60	54	2199	56	2012-08-02 11:17:56.405367
39133	4	1-10871	39132	42	54	927	44	2012-08-02 11:17:56.405367
39134	4	1-11921	39132	54	22	966	55	2012-08-02 11:17:56.405367
39135	4	1-10067	39132	22	60	878	54	2012-08-02 11:17:56.405367
39132	4	1-12094	39132	60	42	1139	56	2012-08-02 11:17:56.405367
39141	3	1-9682	39140	28	68	571	10	2012-08-02 11:17:57.543694
39142	3	1-12011	39140	68	72	453	79	2012-08-02 11:17:57.543694
39140	3	1-12097	39140	72	28	821	63	2012-08-02 11:17:57.543694
39144	4	1-10638	39143	43	59	3660	11	2012-08-02 11:17:57.543694
39145	4	1-10187	39143	59	68	2021	10	2012-08-02 11:17:57.543694
39146	4	1-12011	39143	68	72	2456	79	2012-08-02 11:17:57.543694
39143	4	1-12097	39143	72	43	6835	63	2012-08-02 11:17:57.543694
39163	3	1-11961	39162	26	18	340	81	2012-08-02 11:18:01.002038
39164	3	1-12068	39162	18	48	1186	15	2012-08-02 11:18:01.002038
39162	3	1-12106	39162	48	26	592	52	2012-08-02 11:18:01.002038
39166	4	1-7650	39165	27	23	1132	2	2012-08-02 11:18:01.002038
39167	4	1-11974	39165	23	70	765	84	2012-08-02 11:18:01.002038
39168	4	1-11403	39165	70	48	2227	15	2012-08-02 11:18:01.002038
39165	4	1-12106	39165	48	27	1722	52	2012-08-02 11:18:01.002038
39178	4	1-11531	39177	79	42	1115	81	2012-08-02 11:18:02.139928
39179	4	1-9976	39177	42	99	998	14	2012-08-02 11:18:02.139928
39180	4	1-12071	39177	99	93	496	13	2012-08-02 11:18:02.139928
39177	4	1-12109	39177	93	79	896	18	2012-08-02 11:18:02.139928
39182	4	1-11805	39181	90	96	126	35	2012-08-02 11:18:03.288998
39183	4	1-11294	39181	96	83	167	7	2012-08-02 11:18:03.288998
39184	4	1-11603	39181	83	93	87	20	2012-08-02 11:18:03.288998
39181	4	1-12112	39181	93	90	69	74	2012-08-02 11:18:03.288998
39185	1	1-809	\N	73	73	1254	7	2012-08-02 11:18:03.288998
39194	4	1-11880	39193	65	6	1624	51	2012-08-02 11:18:06.724886
39195	4	1-12021	39193	6	36	1088	11	2012-08-02 11:18:06.724886
39196	4	1-8177	39193	36	70	955	58	2012-08-02 11:18:06.724886
39193	4	1-12121	39193	70	65	1156	13	2012-08-02 11:18:06.724886
39197	1	1-808	\N	36	36	1334	51	2012-08-02 11:18:06.724886
39199	4	1-11453	39198	59	71	3106	50	2012-08-02 11:18:08.028376
39200	4	1-11415	39198	71	89	2454	54	2012-08-02 11:18:08.028376
39201	4	1-10946	39198	89	71	1296	4	2012-08-02 11:18:08.028376
39198	4	1-12123	39198	71	59	2375	100	2012-08-02 11:18:08.028376
39208	3	1-11283	39207	22	71	2126	86	2012-08-02 11:18:13.355285
39209	3	1-11874	39207	71	70	1345	46	2012-08-02 11:18:13.355285
39207	3	1-12138	39207	70	22	3553	33	2012-08-02 11:18:13.355285
39211	4	1-10886	39210	67	8	2154	23	2012-08-02 11:18:13.355285
39212	4	1-12130	39210	8	84	1581	89	2012-08-02 11:18:13.355285
39213	4	1-11433	39210	84	70	895	46	2012-08-02 11:18:13.355285
39210	4	1-12138	39210	70	67	2384	33	2012-08-02 11:18:13.355285
39215	4	1-11711	39214	67	86	1117	61	2012-08-02 11:18:13.355285
39216	4	1-6531	39214	86	49	378	73	2012-08-02 11:18:13.355285
39217	4	1-9717	39214	49	70	255	46	2012-08-02 11:18:13.355285
39214	4	1-12138	39214	70	67	721	33	2012-08-02 11:18:13.355285
39219	4	1-11699	39218	13	91	259	21	2012-08-02 11:18:13.355285
39220	4	1-4610	39218	91	16	89	80	2012-08-02 11:18:13.355285
39221	4	1-11365	39218	16	70	32	46	2012-08-02 11:18:13.355285
39218	4	1-12138	39218	70	13	110	33	2012-08-02 11:18:13.355285
39258	3	1-12130	39257	8	74	3684	89	2012-08-02 11:18:22.160998
39259	3	1-10349	39257	74	68	1197	72	2012-08-02 11:18:22.160998
39257	3	1-12161	39257	68	8	9125	23	2012-08-02 11:18:22.160998
39260	1	1-803	\N	93	93	189	89	2012-08-02 11:18:22.160998
39281	4	1-10087	39280	27	28	514	9	2012-08-02 11:18:28.590617
39282	4	1-11897	39280	28	71	234	8	2012-08-02 11:18:28.590617
39283	4	1-12165	39280	71	94	200	20	2012-08-02 11:18:28.590617
39280	4	1-12178	39280	94	27	1738	88	2012-08-02 11:18:28.590617
39285	4	1-10485	39284	32	6	697	51	2012-08-02 11:18:28.590617
39286	4	1-12021	39284	6	74	409	11	2012-08-02 11:18:28.590617
39287	4	1-11449	39284	74	94	305	20	2012-08-02 11:18:28.590617
39284	4	1-12178	39284	94	32	2824	88	2012-08-02 11:18:28.590617
39301	4	1-11694	39300	45	59	1146	96	2012-08-02 11:18:31.651266
39302	4	1-12173	39300	59	39	572	55	2012-08-02 11:18:31.651266
39303	4	1-9676	39300	39	5	195	39	2012-08-02 11:18:31.651266
39300	4	1-12186	39300	5	45	2900	60	2012-08-02 11:18:31.651266
39305	3	1-11757	39304	6	24	401	18	2012-08-02 11:18:31.651266
39306	3	1-8819	39304	24	5	140	39	2012-08-02 11:18:31.651266
39304	3	1-12186	39304	5	6	2567	60	2012-08-02 11:18:31.651266
39308	4	1-8830	39307	10	68	117	81	2012-08-02 11:18:31.651266
39309	4	1-11594	39307	68	66	95	40	2012-08-02 11:18:31.651266
39310	4	1-11527	39307	66	5	25	39	2012-08-02 11:18:31.651266
39307	4	1-12186	39307	5	10	466	60	2012-08-02 11:18:31.651266
38970	3	1-11930	38968	97	72	371	26	2012-08-02 11:17:34.261328
38972	5	1-11402	38971	85	48	133	51	2012-08-02 11:17:34.261328
38973	5	1-11796	38971	48	24	178	76	2012-08-02 11:17:34.261328
38974	5	1-9493	38971	24	19	49	19	2012-08-02 11:17:34.261328
38975	5	1-11652	38971	19	72	80	26	2012-08-02 11:17:34.261328
38971	5	1-12037	38971	72	85	146	67	2012-08-02 11:17:34.261328
38977	4	1-11565	38976	48	79	3887	21	2012-08-02 11:17:34.261328
38978	4	1-5876	38976	79	19	830	19	2012-08-02 11:17:34.261328
38979	4	1-11652	38976	19	72	1523	26	2012-08-02 11:17:34.261328
38976	4	1-12037	38976	72	48	3079	67	2012-08-02 11:17:34.261328
38984	4	1-12013	38983	58	21	4311	38	2012-08-02 11:17:36.990463
38985	4	1-10863	38983	21	64	2851	87	2012-08-02 11:17:36.990463
38986	4	1-11548	38983	64	15	1911	59	2012-08-02 11:17:36.990463
38983	4	1-12044	38983	15	58	4446	14	2012-08-02 11:17:36.990463
38988	4	1-11898	38987	62	48	1103	66	2012-08-02 11:17:36.990463
38989	4	1-10878	38987	48	64	803	87	2012-08-02 11:17:36.990463
38990	4	1-11548	38987	64	15	581	59	2012-08-02 11:17:36.990463
38987	4	1-12044	38987	15	62	1459	14	2012-08-02 11:17:36.990463
38991	1	1-7060	\N	54	54	409	14	2012-08-02 11:17:36.990463
38993	3	1-8601	38992	68	36	922	31	2012-08-02 11:17:38.17261
38994	3	1-11643	38992	36	14	1388	12	2012-08-02 11:17:38.17261
38992	3	1-12048	38992	14	68	3307	41	2012-08-02 11:17:38.17261
38996	4	1-5276	38995	51	19	1681	88	2012-08-02 11:17:38.17261
38997	4	1-11948	38995	19	39	1206	53	2012-08-02 11:17:38.17261
38998	4	1-11798	38995	39	14	1280	12	2012-08-02 11:17:38.17261
38995	4	1-12048	38995	14	51	3540	41	2012-08-02 11:17:38.17261
39000	3	1-4364	38999	33	39	489	53	2012-08-02 11:17:38.17261
39001	3	1-11798	38999	39	14	529	12	2012-08-02 11:17:38.17261
38999	3	1-12048	38999	14	33	1492	41	2012-08-02 11:17:38.17261
39015	4	1-9509	39014	25	98	4410	27	2012-08-02 11:17:40.616093
39016	4	1-11934	39014	98	28	1648	19	2012-08-02 11:17:40.616093
39017	4	1-11722	39014	28	38	4808	95	2012-08-02 11:17:40.616093
39014	4	1-12053	39014	38	25	5949	64	2012-08-02 11:17:40.616093
39042	4	1-7825	39041	68	16	1633	44	2012-08-02 11:17:41.796561
39043	4	1-12030	39041	16	67	1346	98	2012-08-02 11:17:41.796561
39044	4	1-11818	39041	67	46	1849	38	2012-08-02 11:17:41.796561
39041	4	1-12056	39041	46	68	2071	6	2012-08-02 11:17:41.796561
39046	4	1-7825	39045	68	54	1339	44	2012-08-02 11:17:41.796561
39047	4	1-11921	39045	54	63	1376	55	2012-08-02 11:17:41.796561
39048	4	1-8343	39045	63	46	1406	38	2012-08-02 11:17:41.796561
39045	4	1-12056	39045	46	68	1635	6	2012-08-02 11:17:41.796561
39049	1	1-4301	\N	70	70	973	38	2012-08-02 11:17:41.796561
39051	3	1-11362	39050	72	39	2253	45	2012-08-02 11:17:43.777476
39052	3	1-4579	39050	39	1	970	53	2012-08-02 11:17:43.777476
39050	3	1-12061	39050	1	72	5018	85	2012-08-02 11:17:43.777476
39063	4	1-11796	39062	48	24	965	76	2012-08-02 11:17:44.926177
39064	4	1-9493	39062	24	6	284	19	2012-08-02 11:17:44.926177
39065	4	1-11463	39062	6	71	299	28	2012-08-02 11:17:44.926177
39062	4	1-12064	39062	71	48	673	51	2012-08-02 11:17:44.926177
39095	4	1-12007	39094	11	66	7649	68	2012-08-02 11:17:54.085311
39096	4	1-9647	39094	66	21	3996	30	2012-08-02 11:17:54.085311
39097	4	1-11845	39094	21	87	2678	99	2012-08-02 11:17:54.085311
39094	4	1-12090	39094	87	11	3984	24	2012-08-02 11:17:54.085311
39099	4	1-11725	39098	55	93	834	25	2012-08-02 11:17:54.085311
39100	4	1-10386	39098	93	78	298	96	2012-08-02 11:17:54.085311
39101	4	1-11206	39098	78	87	194	99	2012-08-02 11:17:54.085311
39098	4	1-12090	39098	87	55	385	24	2012-08-02 11:17:54.085311
39103	4	1-12007	39102	11	92	289	68	2012-08-02 11:17:54.085311
39104	4	1-9900	39102	92	16	116	83	2012-08-02 11:17:54.085311
39105	4	1-11683	39102	16	87	54	99	2012-08-02 11:17:54.085311
39102	4	1-12090	39102	87	11	110	24	2012-08-02 11:17:54.085311
39107	4	1-11839	39106	98	41	3402	72	2012-08-02 11:17:54.085311
39108	4	1-11704	39106	41	75	2088	55	2012-08-02 11:17:54.085311
39109	4	1-10373	39106	75	87	1449	99	2012-08-02 11:17:54.085311
39106	4	1-12090	39106	87	98	3077	24	2012-08-02 11:17:54.085311
39111	4	1-4885	39110	67	60	17	10	2012-08-02 11:17:54.085311
39112	4	1-11762	39110	60	78	12	96	2012-08-02 11:17:54.085311
39113	4	1-11206	39110	78	87	8	99	2012-08-02 11:17:54.085311
39110	4	1-12090	39110	87	67	17	24	2012-08-02 11:17:54.085311
39115	4	1-11519	39114	92	13	116	65	2012-08-02 11:17:54.085311
39116	4	1-11103	39114	13	100	88	61	2012-08-02 11:17:54.085311
39117	4	1-11650	39114	100	87	82	99	2012-08-02 11:17:54.085311
39114	4	1-12090	39114	87	92	177	24	2012-08-02 11:17:54.085311
39123	4	1-11088	39122	10	7	426	17	2012-08-02 11:17:56.029889
39124	4	1-11481	39122	7	28	583	91	2012-08-02 11:17:56.029889
39125	4	1-8858	39122	28	71	156	28	2012-08-02 11:17:56.029889
39122	4	1-12093	39122	71	10	354	65	2012-08-02 11:17:56.029889
39152	4	1-6678	39151	36	89	196	75	2012-08-02 11:17:58.316953
39153	4	1-12095	39151	89	16	1161	77	2012-08-02 11:17:58.316953
39154	4	1-8605	39151	16	75	544	49	2012-08-02 11:17:58.316953
39151	4	1-12099	39151	75	36	716	6	2012-08-02 11:17:58.316953
39156	3	1-11881	39155	27	32	4036	81	2012-08-02 11:18:00.604149
39157	3	1-11550	39155	32	25	724	56	2012-08-02 11:18:00.604149
39155	3	1-12105	39155	25	27	5860	68	2012-08-02 11:18:00.604149
39159	4	1-9900	39158	92	16	1322	83	2012-08-02 11:18:00.604149
39160	4	1-11683	39158	16	43	584	99	2012-08-02 11:18:00.604149
39161	4	1-11829	39158	43	25	365	56	2012-08-02 11:18:00.604149
39158	4	1-12105	39158	25	92	3465	68	2012-08-02 11:18:00.604149
39170	4	1-11647	39169	78	71	2997	28	2012-08-02 11:18:01.753028
39171	4	1-12064	39169	71	48	5498	51	2012-08-02 11:18:01.753028
39172	4	1-11796	39169	48	99	6430	76	2012-08-02 11:18:01.753028
39169	4	1-12108	39169	99	78	6050	100	2012-08-02 11:18:01.753028
39174	4	1-10806	39173	94	72	1915	21	2012-08-02 11:18:01.753028
39175	4	1-11290	39173	72	99	1774	69	2012-08-02 11:18:01.753028
39176	4	1-12074	39173	99	99	1141	76	2012-08-02 11:18:01.753028
39173	4	1-12108	39173	99	94	1374	100	2012-08-02 11:18:01.753028
39223	3	1-11652	39222	19	18	4174	26	2012-08-02 11:18:13.719985
39224	3	1-11748	39222	18	30	4161	16	2012-08-02 11:18:13.719985
39222	3	1-12139	39222	30	19	4960	19	2012-08-02 11:18:13.719985
39226	2	1-12140	39225	35	68	3906	76	2012-08-02 11:18:14.869097
39225	2	1-12142	39225	68	35	3608	97	2012-08-02 11:18:14.869097
39250	4	1-11919	39249	34	13	3880	3	2012-08-02 11:18:20.60287
39251	4	1-10121	39249	13	4	2511	43	2012-08-02 11:18:20.60287
39252	4	1-11989	39249	4	68	6402	69	2012-08-02 11:18:20.60287
39249	4	1-12157	39249	68	34	5340	7	2012-08-02 11:18:20.60287
39092	3	1-11948	39091	19	23	1950	53	2012-08-02 11:17:52.549216
39093	3	1-10328	39091	23	1	1499	40	2012-08-02 11:17:52.549216
39091	3	1-12084	39091	1	19	2228	88	2012-08-02 11:17:52.549216
39119	4	1-10589	39118	99	40	3181	71	2012-08-02 11:17:55.631513
39120	4	1-11976	39118	40	34	4471	9	2012-08-02 11:17:55.631513
39121	4	1-8987	39118	34	82	2421	59	2012-08-02 11:17:55.631513
39118	4	1-12092	39118	82	99	3097	74	2012-08-02 11:17:55.631513
39137	4	1-11261	39136	36	98	1035	55	2012-08-02 11:17:56.781219
39138	4	1-11498	39136	98	30	762	13	2012-08-02 11:17:56.781219
39139	4	1-11786	39136	30	89	490	75	2012-08-02 11:17:56.781219
39136	4	1-12095	39136	89	36	2714	77	2012-08-02 11:17:56.781219
39148	4	1-11881	39147	27	68	3607	81	2012-08-02 11:17:57.930402
39149	4	1-11594	39147	68	1	1968	40	2012-08-02 11:17:57.930402
39150	4	1-12084	39147	1	64	1620	88	2012-08-02 11:17:57.930402
39147	4	1-12098	39147	64	27	6919	68	2012-08-02 11:17:57.930402
39187	4	1-9949	39186	7	23	5213	37	2012-08-02 11:18:04.813504
39188	4	1-11868	39186	23	63	6624	57	2012-08-02 11:18:04.813504
39189	4	1-11862	39186	63	34	2728	7	2012-08-02 11:18:04.813504
39186	4	1-12116	39186	34	7	6718	9	2012-08-02 11:18:04.813504
39191	3	1-8033	39190	32	49	877	81	2012-08-02 11:18:04.813504
39192	3	1-10749	39190	49	34	648	7	2012-08-02 11:18:04.813504
39190	3	1-12116	39190	34	32	1635	9	2012-08-02 11:18:04.813504
39203	5	1-10897	39202	34	45	926	60	2012-08-02 11:18:11.830589
39204	5	1-11694	39202	45	74	528	96	2012-08-02 11:18:11.830589
39205	5	1-10088	39202	74	23	301	8	2012-08-02 11:18:11.830589
39206	5	1-12055	39202	23	56	749	26	2012-08-02 11:18:11.830589
39202	5	1-12134	39202	56	34	555	79	2012-08-02 11:18:11.830589
39228	4	1-11628	39227	94	87	4329	80	2012-08-02 11:18:15.244833
39229	4	1-12122	39227	87	48	4293	92	2012-08-02 11:18:15.244833
39230	4	1-11589	39227	48	32	2378	49	2012-08-02 11:18:15.244833
39227	4	1-12143	39227	32	94	6178	71	2012-08-02 11:18:15.244833
39232	4	1-11976	39231	40	96	1330	9	2012-08-02 11:18:15.244833
39233	4	1-7820	39231	96	55	242	36	2012-08-02 11:18:15.244833
39234	4	1-12015	39231	55	32	373	49	2012-08-02 11:18:15.244833
39231	4	1-12143	39231	32	40	1010	71	2012-08-02 11:18:15.244833
39236	4	1-11773	39235	47	89	1645	59	2012-08-02 11:18:15.244833
39237	4	1-11209	39235	89	59	1155	30	2012-08-02 11:18:15.244833
39238	4	1-7251	39235	59	32	696	49	2012-08-02 11:18:15.244833
39235	4	1-12143	39235	32	47	2249	71	2012-08-02 11:18:15.244833
39240	3	1-11386	39239	36	13	873	7	2012-08-02 11:18:16.404878
39241	3	1-11985	39239	13	71	723	32	2012-08-02 11:18:16.404878
39239	3	1-12146	39239	71	36	545	48	2012-08-02 11:18:16.404878
39243	4	1-11794	39242	8	43	1035	73	2012-08-02 11:18:20.238255
39244	4	1-11811	39242	43	48	3602	15	2012-08-02 11:18:20.238255
39245	4	1-12106	39242	48	91	3609	52	2012-08-02 11:18:20.238255
39242	4	1-12156	39242	91	8	2069	76	2012-08-02 11:18:20.238255
39247	3	1-12108	39246	99	65	1966	100	2012-08-02 11:18:20.238255
39248	3	1-12125	39246	65	91	2935	52	2012-08-02 11:18:20.238255
39246	3	1-12156	39246	91	99	1734	76	2012-08-02 11:18:20.238255
39265	4	1-9259	39264	12	39	870	53	2012-08-02 11:18:27.850404
39266	4	1-11798	39264	39	1	1127	12	2012-08-02 11:18:27.850404
39267	4	1-12025	39264	1	28	598	10	2012-08-02 11:18:27.850404
39264	4	1-12176	39264	28	12	742	74	2012-08-02 11:18:27.850404
39269	4	1-11764	39268	23	100	447	85	2012-08-02 11:18:27.850404
39270	4	1-11164	39268	100	18	364	65	2012-08-02 11:18:27.850404
39271	4	1-5668	39268	18	28	284	10	2012-08-02 11:18:27.850404
39268	4	1-12176	39268	28	23	371	74	2012-08-02 11:18:27.850404
39289	4	1-10503	39288	19	38	4728	50	2012-08-02 11:18:28.977507
39290	4	1-11532	39288	38	72	4431	83	2012-08-02 11:18:28.977507
39291	4	1-8437	39288	72	93	3270	14	2012-08-02 11:18:28.977507
39288	4	1-12179	39288	93	19	7225	43	2012-08-02 11:18:28.977507
39326	2	1-12026	39325	13	49	2318	15	2012-08-02 11:18:33.574221
39325	2	1-12191	39325	49	13	8650	50	2012-08-02 11:18:33.574221
39328	4	1-10451	39327	96	31	377	79	2012-08-02 11:18:35.816606
39329	4	1-10218	39327	31	89	111	75	2012-08-02 11:18:35.816606
39330	4	1-12095	39327	89	35	451	77	2012-08-02 11:18:35.816606
39327	4	1-12197	39327	35	96	924	70	2012-08-02 11:18:35.816606
39332	4	1-11183	39331	96	46	659	9	2012-08-02 11:18:35.816606
39333	4	1-7157	39331	46	73	199	24	2012-08-02 11:18:35.816606
39334	4	1-11284	39331	73	35	340	77	2012-08-02 11:18:35.816606
39331	4	1-12197	39331	35	96	910	70	2012-08-02 11:18:35.816606
39336	4	1-10451	39335	96	78	700	79	2012-08-02 11:18:35.816606
39337	4	1-11528	39335	78	80	451	6	2012-08-02 11:18:35.816606
39338	4	1-11941	39335	80	35	468	77	2012-08-02 11:18:35.816606
39335	4	1-12197	39335	35	96	1282	70	2012-08-02 11:18:35.816606
39340	4	1-11998	39339	22	14	2315	84	2012-08-02 11:18:35.816606
39341	4	1-10323	39339	14	19	1543	16	2012-08-02 11:18:35.816606
39342	4	1-11928	39339	19	35	1167	77	2012-08-02 11:18:35.816606
39339	4	1-12197	39339	35	22	3289	70	2012-08-02 11:18:35.816606
39347	4	1-11796	39346	48	68	1176	76	2012-08-02 11:18:36.987947
39348	4	1-12142	39346	68	8	909	97	2012-08-02 11:18:36.987947
39349	4	1-11850	39346	8	83	950	68	2012-08-02 11:18:36.987947
39346	4	1-12200	39346	83	48	1130	51	2012-08-02 11:18:36.987947
39351	4	1-11450	39350	48	37	1360	34	2012-08-02 11:18:36.987947
39352	4	1-10595	39350	37	32	817	22	2012-08-02 11:18:36.987947
39353	4	1-11790	39350	32	83	877	68	2012-08-02 11:18:36.987947
39350	4	1-12200	39350	83	48	1366	51	2012-08-02 11:18:36.987947
39355	3	1-11450	39354	48	56	572	34	2012-08-02 11:18:36.987947
39356	3	1-10755	39354	56	83	320	68	2012-08-02 11:18:36.987947
39354	3	1-12200	39354	83	48	535	51	2012-08-02 11:18:36.987947
39358	4	1-11450	39357	48	82	1733	34	2012-08-02 11:18:36.987947
39359	4	1-11127	39357	82	97	549	24	2012-08-02 11:18:36.987947
39360	4	1-12057	39357	97	83	919	68	2012-08-02 11:18:36.987947
39357	4	1-12200	39357	83	48	1579	51	2012-08-02 11:18:36.987947
39418	4	1-11794	39417	8	89	1140	73	2012-08-02 11:18:44.372263
39419	4	1-11192	39417	89	30	1461	34	2012-08-02 11:18:44.372263
39420	4	1-12185	39417	30	53	1474	38	2012-08-02 11:18:44.372263
39417	4	1-12219	39417	53	8	2237	76	2012-08-02 11:18:44.372263
39456	2	1-3693	39455	33	83	385	14	2012-08-02 11:18:46.658793
39455	2	1-12225	39455	83	33	8154	15	2012-08-02 11:18:46.658793
39472	4	1-11396	39471	100	95	256	63	2012-08-02 11:18:49.730361
39473	4	1-11828	39471	95	70	1098	87	2012-08-02 11:18:49.730361
39474	4	1-12193	39471	70	67	869	7	2012-08-02 11:18:49.730361
39471	4	1-12234	39471	67	100	364	35	2012-08-02 11:18:49.730361
39499	3	1-12240	39499	69	33	381	90	2012-08-02 11:18:52.415547
39254	4	1-9381	39253	4	93	2946	20	2012-08-02 11:18:21.774066
39255	4	1-12112	39253	93	23	2320	74	2012-08-02 11:18:21.774066
39256	4	1-11764	39253	23	21	2495	85	2012-08-02 11:18:21.774066
39253	4	1-12160	39253	21	4	4424	93	2012-08-02 11:18:21.774066
39262	3	1-12160	39261	21	46	1602	93	2012-08-02 11:18:27.121335
39263	3	1-12086	39261	46	20	1234	80	2012-08-02 11:18:27.121335
39261	3	1-12174	39261	20	21	816	85	2012-08-02 11:18:27.121335
39273	4	1-9528	39272	38	43	3014	9	2012-08-02 11:18:28.226095
39274	4	1-11352	39272	43	22	1464	1	2012-08-02 11:18:28.226095
39275	4	1-11995	39272	22	14	5263	15	2012-08-02 11:18:28.226095
39272	4	1-12177	39272	14	38	4925	11	2012-08-02 11:18:28.226095
39277	4	1-8845	39276	41	88	375	59	2012-08-02 11:18:28.226095
39278	4	1-11972	39276	88	22	188	1	2012-08-02 11:18:28.226095
39279	4	1-11995	39276	22	14	765	15	2012-08-02 11:18:28.226095
39276	4	1-12177	39276	14	41	810	11	2012-08-02 11:18:28.226095
39293	4	1-11921	39292	54	4	818	55	2012-08-02 11:18:29.364252
39294	4	1-9225	39292	4	87	771	80	2012-08-02 11:18:29.364252
39295	4	1-12122	39292	87	84	997	92	2012-08-02 11:18:29.364252
39292	4	1-12180	39292	84	54	856	44	2012-08-02 11:18:29.364252
39297	4	1-11733	39296	69	51	2655	9	2012-08-02 11:18:30.502373
39298	4	1-7206	39296	51	67	2134	42	2012-08-02 11:18:30.502373
39299	4	1-11835	39296	67	15	2167	32	2012-08-02 11:18:30.502373
39296	4	1-12184	39296	15	69	5921	99	2012-08-02 11:18:30.502373
39316	3	1-8437	39315	72	62	205	14	2012-08-02 11:18:32.038113
39317	3	1-11898	39315	62	38	91	66	2012-08-02 11:18:32.038113
39315	3	1-12187	39315	38	72	364	83	2012-08-02 11:18:32.038113
39319	3	1-5792	39318	19	62	355	14	2012-08-02 11:18:32.038113
39320	3	1-11898	39318	62	38	182	66	2012-08-02 11:18:32.038113
39318	3	1-12187	39318	38	19	840	83	2012-08-02 11:18:32.038113
39322	4	1-9763	39321	11	66	2388	73	2012-08-02 11:18:33.209406
39323	4	1-11879	39321	66	35	3220	42	2012-08-02 11:18:33.209406
39324	4	1-11251	39321	35	74	1671	41	2012-08-02 11:18:33.209406
39321	4	1-12190	39321	74	11	3839	93	2012-08-02 11:18:33.209406
39369	3	1-11158	39368	71	74	2928	54	2012-08-02 11:18:37.739071
39370	3	1-11099	39368	74	23	2118	7	2012-08-02 11:18:37.739071
39368	3	1-12202	39368	23	71	2700	30	2012-08-02 11:18:37.739071
39372	4	1-11059	39371	39	48	554	41	2012-08-02 11:18:40.083726
39373	4	1-11827	39371	48	12	900	83	2012-08-02 11:18:40.083726
39374	4	1-11501	39371	12	31	396	38	2012-08-02 11:18:40.083726
39371	4	1-12208	39371	31	39	685	17	2012-08-02 11:18:40.083726
39376	4	1-11481	39375	7	5	151	91	2012-08-02 11:18:40.083726
39377	4	1-6680	39375	5	43	80	10	2012-08-02 11:18:40.083726
39378	4	1-10499	39375	43	31	64	38	2012-08-02 11:18:40.083726
39375	4	1-12208	39375	31	7	113	17	2012-08-02 11:18:40.083726
39380	4	1-11481	39379	7	53	2066	91	2012-08-02 11:18:40.083726
39381	4	1-11417	39379	53	85	1584	25	2012-08-02 11:18:40.083726
39382	4	1-12113	39379	85	31	797	38	2012-08-02 11:18:40.083726
39379	4	1-12208	39379	31	7	1474	17	2012-08-02 11:18:40.083726
39396	5	1-11660	39395	29	5	2730	61	2012-08-02 11:18:41.232527
39397	5	1-4498	39395	5	53	1197	20	2012-08-02 11:18:41.232527
39398	5	1-11686	39395	53	23	701	8	2012-08-02 11:18:41.232527
39399	5	1-12055	39395	23	62	1754	26	2012-08-02 11:18:41.232527
39395	5	1-12211	39395	62	29	3092	31	2012-08-02 11:18:41.232527
39406	4	1-11204	39405	82	84	4086	92	2012-08-02 11:18:42.392774
39407	4	1-12180	39405	84	84	3486	44	2012-08-02 11:18:42.392774
39408	4	1-11272	39405	84	7	3105	13	2012-08-02 11:18:42.392774
39405	4	1-12215	39405	7	82	3717	1	2012-08-02 11:18:42.392774
39410	4	1-11204	39409	82	61	1540	92	2012-08-02 11:18:42.392774
39411	4	1-11369	39409	61	25	1197	27	2012-08-02 11:18:42.392774
39412	4	1-11159	39409	25	7	994	13	2012-08-02 11:18:42.392774
39409	4	1-12215	39409	7	82	1291	1	2012-08-02 11:18:42.392774
39414	3	1-7640	39413	29	13	3702	5	2012-08-02 11:18:42.392774
39415	3	1-12171	39413	13	7	3045	13	2012-08-02 11:18:42.392774
39413	3	1-12215	39413	7	29	4288	1	2012-08-02 11:18:42.392774
39416	1	1-1675	\N	2	2	905	5	2012-08-02 11:18:42.392774
39426	4	1-11433	39425	84	48	3663	46	2012-08-02 11:18:45.145114
39427	4	1-11408	39425	48	61	6092	92	2012-08-02 11:18:45.145114
39428	4	1-11369	39425	61	27	4030	27	2012-08-02 11:18:45.145114
39425	4	1-12221	39425	27	84	4941	89	2012-08-02 11:18:45.145114
39430	3	1-12079	39429	35	29	549	53	2012-08-02 11:18:45.145114
39431	3	1-11392	39429	29	27	357	27	2012-08-02 11:18:45.145114
39429	3	1-12221	39429	27	35	505	89	2012-08-02 11:18:45.145114
39436	4	1-4579	39435	39	39	1441	53	2012-08-02 11:18:46.294302
39437	4	1-11798	39435	39	73	1387	12	2012-08-02 11:18:46.294302
39438	4	1-6499	39435	73	31	485	29	2012-08-02 11:18:46.294302
39435	4	1-12224	39435	31	39	2554	45	2012-08-02 11:18:46.294302
39440	4	1-9382	39439	93	55	1269	6	2012-08-02 11:18:46.294302
39441	4	1-12172	39439	55	61	733	82	2012-08-02 11:18:46.294302
39442	4	1-11831	39439	61	31	364	29	2012-08-02 11:18:46.294302
39439	4	1-12224	39439	31	93	2581	45	2012-08-02 11:18:46.294302
39444	3	1-8969	39443	5	15	504	90	2012-08-02 11:18:46.294302
39445	3	1-11924	39443	15	31	280	29	2012-08-02 11:18:46.294302
39443	3	1-12224	39443	31	5	2384	45	2012-08-02 11:18:46.294302
39447	4	1-10711	39446	77	43	40	63	2012-08-02 11:18:46.294302
39448	4	1-10638	39446	43	56	36	11	2012-08-02 11:18:46.294302
39449	4	1-7724	39446	56	31	34	29	2012-08-02 11:18:46.294302
39446	4	1-12224	39446	31	77	334	45	2012-08-02 11:18:46.294302
39451	4	1-11009	39450	77	44	78	91	2012-08-02 11:18:46.294302
39452	4	1-8583	39450	44	34	39	65	2012-08-02 11:18:46.294302
39453	4	1-9432	39450	34	31	20	29	2012-08-02 11:18:46.294302
39450	4	1-12224	39450	31	77	207	45	2012-08-02 11:18:46.294302
39454	1	1-4579	\N	39	39	3283	53	2012-08-02 11:18:46.294302
39458	4	1-11218	39457	52	31	5557	6	2012-08-02 11:18:47.433288
39459	4	1-12168	39457	31	47	6193	71	2012-08-02 11:18:47.433288
39460	4	1-11773	39457	47	92	4448	59	2012-08-02 11:18:47.433288
39457	4	1-12228	39457	92	52	6289	64	2012-08-02 11:18:47.433288
39462	3	1-11265	39461	10	41	2583	37	2012-08-02 11:18:50.095922
39463	3	1-11978	39461	41	89	2082	45	2012-08-02 11:18:50.095922
39461	3	1-12233	39461	89	10	2239	91	2012-08-02 11:18:50.095922
39465	3	1-7835	39464	65	41	2576	37	2012-08-02 11:18:50.095922
39466	3	1-11978	39464	41	89	2101	45	2012-08-02 11:18:50.095922
39464	3	1-12233	39464	89	65	2286	91	2012-08-02 11:18:50.095922
39468	4	1-8858	39467	28	71	776	28	2012-08-02 11:18:50.095922
39312	4	1-11306	39311	50	1	464	79	2012-08-02 11:18:31.651266
39313	4	1-10456	39311	1	66	347	40	2012-08-02 11:18:31.651266
39314	4	1-11527	39311	66	5	104	39	2012-08-02 11:18:31.651266
39311	4	1-12186	39311	5	50	2138	60	2012-08-02 11:18:31.651266
39344	3	1-9373	39343	64	90	1473	74	2012-08-02 11:18:36.214759
39345	3	1-11805	39343	90	87	1486	35	2012-08-02 11:18:36.214759
39343	3	1-12198	39343	87	64	8192	49	2012-08-02 11:18:36.214759
39362	3	1-12151	39361	82	13	2782	65	2012-08-02 11:18:37.363591
39363	3	1-11103	39361	13	89	2281	61	2012-08-02 11:18:37.363591
39361	3	1-12201	39361	89	82	2628	58	2012-08-02 11:18:37.363591
39365	4	1-12121	39364	70	39	3392	13	2012-08-02 11:18:37.363591
39366	4	1-9336	39364	39	29	2119	31	2012-08-02 11:18:37.363591
39367	4	1-11660	39364	29	89	2061	61	2012-08-02 11:18:37.363591
39364	4	1-12201	39364	89	70	2473	58	2012-08-02 11:18:37.363591
39384	4	1-12053	39383	38	52	2595	64	2012-08-02 11:18:40.845599
39385	4	1-11218	39383	52	68	2573	6	2012-08-02 11:18:40.845599
39386	4	1-7825	39383	68	60	2062	44	2012-08-02 11:18:40.845599
39383	4	1-12210	39383	60	38	2147	95	2012-08-02 11:18:40.845599
39388	4	1-11891	39387	5	93	4542	47	2012-08-02 11:18:40.845599
39389	4	1-10836	39387	93	45	4147	66	2012-08-02 11:18:40.845599
39390	4	1-9947	39387	45	60	3890	44	2012-08-02 11:18:40.845599
39387	4	1-12210	39387	60	5	4056	95	2012-08-02 11:18:40.845599
39392	4	1-9852	39391	12	21	799	85	2012-08-02 11:18:40.845599
39393	4	1-12160	39391	21	83	1602	93	2012-08-02 11:18:40.845599
39394	4	1-11740	39391	83	60	1748	44	2012-08-02 11:18:40.845599
39391	4	1-12210	39391	60	12	1860	95	2012-08-02 11:18:40.845599
39401	2	1-10606	39400	26	79	196	79	2012-08-02 11:18:42.017055
39400	2	1-12212	39400	79	26	213	94	2012-08-02 11:18:42.017055
39403	3	1-12022	39402	19	79	4203	79	2012-08-02 11:18:42.790525
39404	3	1-12212	39402	79	55	2125	94	2012-08-02 11:18:42.790525
39402	3	1-12214	39402	55	19	8746	95	2012-08-02 11:18:42.790525
39422	4	1-10863	39421	21	97	1626	87	2012-08-02 11:18:44.758408
39423	4	1-11930	39421	97	72	742	26	2012-08-02 11:18:44.758408
39424	4	1-12037	39421	72	13	1516	67	2012-08-02 11:18:44.758408
39421	4	1-12220	39421	13	21	2065	38	2012-08-02 11:18:44.758408
39433	3	1-11548	39432	64	34	910	59	2012-08-02 11:18:45.918585
39434	3	1-12183	39432	34	88	735	11	2012-08-02 11:18:45.918585
39432	3	1-12223	39432	88	64	1276	87	2012-08-02 11:18:45.918585
39476	4	1-11363	39475	85	92	1169	66	2012-08-02 11:18:50.492944
39477	4	1-11969	39475	92	96	1332	70	2012-08-02 11:18:50.492944
39478	4	1-11183	39475	96	31	642	9	2012-08-02 11:18:50.492944
39475	4	1-12235	39475	31	85	3129	84	2012-08-02 11:18:50.492944
39480	4	1-10788	39479	12	40	2570	48	2012-08-02 11:18:50.492944
39481	4	1-11113	39479	40	96	1923	70	2012-08-02 11:18:50.492944
39482	4	1-11183	39479	96	31	1022	9	2012-08-02 11:18:50.492944
39479	4	1-12235	39479	31	12	5489	84	2012-08-02 11:18:50.492944
39512	4	1-11679	39511	65	84	1689	2	2012-08-02 11:18:55.16701
39513	4	1-12199	39511	84	84	659	92	2012-08-02 11:18:55.16701
39514	4	1-12180	39511	84	10	334	44	2012-08-02 11:18:55.16701
39511	4	1-12247	39511	10	65	1806	75	2012-08-02 11:18:55.16701
39516	4	1-7646	39515	33	84	2030	2	2012-08-02 11:18:55.16701
39517	4	1-12199	39515	84	84	894	92	2012-08-02 11:18:55.16701
39518	4	1-12180	39515	84	10	512	44	2012-08-02 11:18:55.16701
39515	4	1-12247	39515	10	33	3522	75	2012-08-02 11:18:55.16701
39532	4	1-12142	39531	68	48	14	97	2012-08-02 11:18:58.625043
39533	4	1-11865	39531	48	71	10	86	2012-08-02 11:18:58.625043
39534	4	1-11874	39531	71	44	10	46	2012-08-02 11:18:58.625043
39531	4	1-12256	39531	44	68	14	76	2012-08-02 11:18:58.625043
39572	4	1-11997	39571	90	24	4675	32	2012-08-02 11:19:00.934049
39573	4	1-12252	39571	24	40	3966	75	2012-08-02 11:19:00.934049
39574	4	1-9943	39571	40	99	2981	9	2012-08-02 11:19:00.934049
39571	4	1-12262	39571	99	90	2803	63	2012-08-02 11:19:00.934049
39576	4	1-10638	39575	43	59	3443	11	2012-08-02 11:19:00.934049
39577	4	1-10187	39575	59	31	3219	10	2012-08-02 11:19:00.934049
39578	4	1-12004	39575	31	99	3903	9	2012-08-02 11:19:00.934049
39575	4	1-12262	39575	99	43	3797	63	2012-08-02 11:19:00.934049
39590	4	1-12055	39589	23	62	195	26	2012-08-02 11:19:04.007617
39591	4	1-12211	39589	62	29	360	31	2012-08-02 11:19:04.007617
39592	4	1-11660	39589	29	75	332	61	2012-08-02 11:19:04.007617
39589	4	1-12270	39589	75	23	75	8	2012-08-02 11:19:04.007617
39594	4	1-12176	39593	28	23	248	74	2012-08-02 11:19:05.134505
39595	4	1-11764	39593	23	21	217	85	2012-08-02 11:19:05.134505
39596	4	1-12160	39593	21	13	314	93	2012-08-02 11:19:05.134505
39593	4	1-12273	39593	13	28	261	10	2012-08-02 11:19:05.134505
39598	4	1-10884	39597	83	3	4401	43	2012-08-02 11:19:05.134505
39599	4	1-12118	39597	3	30	5778	30	2012-08-02 11:19:05.134505
39600	4	1-12107	39597	30	13	6927	93	2012-08-02 11:19:05.134505
39597	4	1-12273	39597	13	83	6134	10	2012-08-02 11:19:05.134505
39602	4	1-5613	39601	31	54	2286	95	2012-08-02 11:19:06.27261
39603	4	1-12101	39601	54	97	2334	98	2012-08-02 11:19:06.27261
39604	4	1-12077	39601	97	91	3567	5	2012-08-02 11:19:06.27261
39601	4	1-12276	39601	91	31	4398	49	2012-08-02 11:19:06.27261
39614	4	1-12036	39613	38	70	6005	87	2012-08-02 11:19:09.752469
39615	4	1-12193	39613	70	5	4681	7	2012-08-02 11:19:09.752469
39616	4	1-9480	39613	5	80	2507	68	2012-08-02 11:19:09.752469
39613	4	1-12285	39613	80	38	4217	60	2012-08-02 11:19:09.752469
39622	3	1-10734	39621	59	36	443	31	2012-08-02 11:19:10.890569
39623	3	1-11643	39621	36	29	708	12	2012-08-02 11:19:10.890569
39621	3	1-12288	39621	29	59	2119	16	2012-08-02 11:19:10.890569
39632	4	1-11997	39631	90	66	1212	32	2012-08-02 11:19:12.072644
39633	4	1-10449	39631	66	85	1225	25	2012-08-02 11:19:12.072644
39634	4	1-12113	39631	85	45	320	38	2012-08-02 11:19:12.072644
39631	4	1-12291	39631	45	90	1221	63	2012-08-02 11:19:12.072644
39636	3	1-10638	39635	43	84	932	11	2012-08-02 11:19:12.072644
39637	3	1-8059	39635	84	45	401	38	2012-08-02 11:19:12.072644
39635	3	1-12291	39635	45	43	1655	63	2012-08-02 11:19:12.072644
39639	4	1-10375	39638	20	55	686	24	2012-08-02 11:19:12.072644
39640	4	1-11725	39638	55	85	1105	25	2012-08-02 11:19:12.072644
39641	4	1-12113	39638	85	45	332	38	2012-08-02 11:19:12.072644
39638	4	1-12291	39638	45	20	1463	63	2012-08-02 11:19:12.072644
39642	1	1-10375	\N	20	20	4262	24	2012-08-02 11:19:12.072644
39648	3	1-12083	39647	66	53	446	43	2012-08-02 11:19:16.72384
39647	3	1-12302	39647	14	66	684	29	2012-08-02 11:19:16.72384
39469	4	1-12093	39467	71	4	1773	65	2012-08-02 11:18:50.095922
39470	4	1-11975	39467	4	89	2308	45	2012-08-02 11:18:50.095922
39467	4	1-12233	39467	89	28	2881	91	2012-08-02 11:18:50.095922
39484	4	1-11082	39483	5	67	2015	7	2012-08-02 11:18:52.028869
39485	4	1-12234	39483	67	56	731	35	2012-08-02 11:18:52.028869
39486	4	1-10345	39483	56	91	593	66	2012-08-02 11:18:52.028869
39483	4	1-12239	39483	91	5	2321	13	2012-08-02 11:18:52.028869
39488	4	1-11198	39487	30	6	580	51	2012-08-02 11:18:52.028869
39489	4	1-12021	39487	6	66	371	11	2012-08-02 11:18:52.028869
39490	4	1-10501	39487	66	91	201	66	2012-08-02 11:18:52.028869
39487	4	1-12239	39487	91	30	794	13	2012-08-02 11:18:52.028869
39492	4	1-11198	39491	30	100	1556	51	2012-08-02 11:18:52.028869
39493	4	1-6050	39491	100	62	508	14	2012-08-02 11:18:52.028869
39494	4	1-11898	39491	62	91	407	66	2012-08-02 11:18:52.028869
39491	4	1-12239	39491	91	30	1851	13	2012-08-02 11:18:52.028869
39496	4	1-10515	39495	91	1	848	42	2012-08-02 11:18:52.028869
39497	4	1-11780	39495	1	66	365	11	2012-08-02 11:18:52.028869
39498	4	1-10501	39495	66	91	219	66	2012-08-02 11:18:52.028869
39495	4	1-12239	39495	91	91	994	13	2012-08-02 11:18:52.028869
39536	2	1-9065	39535	35	19	5295	69	2012-08-02 11:18:59.000354
39535	2	1-12257	39535	19	35	5488	63	2012-08-02 11:18:59.000354
39538	4	1-11766	39537	97	60	1100	90	2012-08-02 11:18:59.000354
39539	4	1-10244	39537	60	4	484	43	2012-08-02 11:18:59.000354
39540	4	1-11989	39537	4	19	945	69	2012-08-02 11:18:59.000354
39537	4	1-12257	39537	19	97	1036	63	2012-08-02 11:18:59.000354
39542	4	1-11997	39541	90	80	956	32	2012-08-02 11:18:59.000354
39543	4	1-11612	39541	80	62	1347	80	2012-08-02 11:18:59.000354
39544	4	1-11320	39541	62	19	487	69	2012-08-02 11:18:59.000354
39541	4	1-12257	39541	19	90	592	63	2012-08-02 11:18:59.000354
39546	3	1-10929	39545	80	89	2239	88	2012-08-02 11:18:59.000354
39547	3	1-9817	39545	89	19	905	69	2012-08-02 11:18:59.000354
39545	3	1-12257	39545	19	80	1191	63	2012-08-02 11:18:59.000354
39556	4	1-11574	39555	34	23	4205	7	2012-08-02 11:19:00.160668
39557	4	1-12202	39555	23	17	3823	30	2012-08-02 11:19:00.160668
39558	4	1-10019	39555	17	39	1939	21	2012-08-02 11:19:00.160668
39555	4	1-12260	39555	39	34	4998	81	2012-08-02 11:19:00.160668
39560	4	1-11688	39559	56	9	809	62	2012-08-02 11:19:00.160668
39561	4	1-7449	39559	9	62	545	60	2012-08-02 11:19:00.160668
39562	4	1-12103	39559	62	39	235	21	2012-08-02 11:19:00.160668
39559	4	1-12260	39559	39	56	609	81	2012-08-02 11:19:00.160668
39564	4	1-10749	39563	49	23	561	7	2012-08-02 11:19:00.160668
39565	4	1-12202	39563	23	17	543	30	2012-08-02 11:19:00.160668
39566	4	1-10019	39563	17	39	293	21	2012-08-02 11:19:00.160668
39563	4	1-12260	39563	39	49	803	81	2012-08-02 11:19:00.160668
39568	3	1-11947	39567	28	48	86	67	2012-08-02 11:19:00.160668
39569	3	1-11565	39567	48	39	92	21	2012-08-02 11:19:00.160668
39567	3	1-12260	39567	39	28	288	81	2012-08-02 11:19:00.160668
39570	1	1-4310	\N	83	83	1652	62	2012-08-02 11:19:00.160668
39588	2	1-12001	39587	39	69	3318	91	2012-08-02 11:19:02.471897
39587	2	1-12266	39587	69	39	4894	21	2012-08-02 11:19:02.471897
39625	3	1-10010	39624	100	33	1365	75	2012-08-02 11:19:11.67487
39626	3	1-7646	39624	33	73	1190	2	2012-08-02 11:19:11.67487
39624	3	1-12290	39624	73	100	6140	57	2012-08-02 11:19:11.67487
39628	4	1-10468	39627	95	77	1364	6	2012-08-02 11:19:11.67487
39629	4	1-8936	39627	77	24	365	18	2012-08-02 11:19:11.67487
39630	4	1-11427	39627	24	73	698	2	2012-08-02 11:19:11.67487
39627	4	1-12290	39627	73	95	3737	57	2012-08-02 11:19:11.67487
39644	4	1-11915	39643	2	44	2032	46	2012-08-02 11:19:13.995089
39645	4	1-12256	39643	44	29	3246	76	2012-08-02 11:19:13.995089
39646	4	1-10934	39643	29	36	2263	65	2012-08-02 11:19:13.995089
39643	4	1-12296	39643	36	2	1829	8	2012-08-02 11:19:13.995089
39655	5	1-11931	39654	13	22	3300	64	2012-08-02 11:19:18.027812
39656	5	1-10016	39654	22	60	2318	49	2012-08-02 11:19:18.027812
39657	5	1-10997	39654	60	50	991	78	2012-08-02 11:19:18.027812
39658	5	1-12275	39654	50	8	2794	39	2012-08-02 11:19:18.027812
39654	5	1-12306	39654	8	13	5817	80	2012-08-02 11:19:18.027812
39666	3	1-11724	39665	37	75	5043	54	2012-08-02 11:19:20.336446
39667	3	1-12128	39665	75	2	2332	84	2012-08-02 11:19:20.336446
39665	3	1-12312	39665	2	37	9683	17	2012-08-02 11:19:20.336446
39687	4	1-2380	39686	11	19	728	54	2012-08-02 11:19:22.656582
39688	4	1-11669	39686	19	80	671	32	2012-08-02 11:19:22.656582
39689	4	1-11612	39686	80	31	1100	80	2012-08-02 11:19:22.656582
39686	4	1-12318	39686	31	11	2051	25	2012-08-02 11:19:22.656582
39695	3	1-12029	39694	50	18	3404	94	2012-08-02 11:19:23.827988
39696	3	1-9241	39694	18	52	538	38	2012-08-02 11:19:23.827988
39694	3	1-12321	39694	52	50	7713	89	2012-08-02 11:19:23.827988
39698	4	1-11557	39697	75	28	592	10	2012-08-02 11:19:24.955259
39699	4	1-12176	39697	28	99	658	74	2012-08-02 11:19:24.955259
39700	4	1-10589	39697	99	32	720	71	2012-08-02 11:19:24.955259
39697	4	1-12324	39697	32	75	2866	48	2012-08-02 11:19:24.955259
39718	4	1-9120	39717	59	10	460	23	2012-08-02 11:19:26.126022
39719	4	1-10479	39717	10	92	421	66	2012-08-02 11:19:26.126022
39720	4	1-11969	39717	92	96	661	70	2012-08-02 11:19:26.126022
39717	4	1-12327	39717	96	59	914	49	2012-08-02 11:19:26.126022
39722	4	1-9373	39721	64	7	686	74	2012-08-02 11:19:26.126022
39723	4	1-11108	39721	7	100	731	80	2012-08-02 11:19:26.126022
39724	4	1-11906	39721	100	96	1423	70	2012-08-02 11:19:26.126022
39721	4	1-12327	39721	96	64	2090	49	2012-08-02 11:19:26.126022
39730	4	1-10320	39729	52	19	4193	88	2012-08-02 11:19:28.468416
39731	4	1-11948	39729	19	76	3922	53	2012-08-02 11:19:28.468416
39732	4	1-10354	39729	76	37	2392	17	2012-08-02 11:19:28.468416
39729	4	1-12333	39729	37	52	3264	66	2012-08-02 11:19:28.468416
39752	3	1-10884	39751	83	3	486	43	2012-08-02 11:19:30.722095
39753	3	1-12118	39751	3	93	764	30	2012-08-02 11:19:30.722095
39751	3	1-12340	39751	93	83	566	10	2012-08-02 11:19:30.722095
39829	4	1-12212	39828	79	42	2422	94	2012-08-02 11:19:47.352458
39830	4	1-12227	39828	42	22	1419	22	2012-08-02 11:19:47.352458
39831	4	1-9664	39828	22	23	791	60	2012-08-02 11:19:47.352458
39828	4	1-12382	39828	23	79	3074	79	2012-08-02 11:19:47.352458
39833	4	1-6353	39832	54	69	33	4	2012-08-02 11:19:47.352458
39834	4	1-12240	39832	69	49	130	90	2012-08-02 11:19:47.352458
39835	4	1-8072	39832	49	23	66	60	2012-08-02 11:19:47.352458
39832	4	1-12382	39832	23	54	257	79	2012-08-02 11:19:47.352458
39500	3	1-7964	39499	33	45	178	97	2012-08-02 11:18:52.415547
39501	3	1-9058	39499	45	69	95	4	2012-08-02 11:18:52.415547
39503	3	1-6652	39502	3	89	807	54	2012-08-02 11:18:52.415547
39504	3	1-10946	39502	89	69	342	4	2012-08-02 11:18:52.415547
39502	3	1-12240	39502	69	3	1476	90	2012-08-02 11:18:52.415547
39506	4	1-9603	39505	1	37	1236	17	2012-08-02 11:18:52.415547
39507	4	1-11724	39505	37	89	847	54	2012-08-02 11:18:52.415547
39508	4	1-10946	39505	89	69	388	4	2012-08-02 11:18:52.415547
39505	4	1-12240	39505	69	1	1803	90	2012-08-02 11:18:52.415547
39510	2	1-12093	39509	71	10	6019	65	2012-08-02 11:18:53.620239
39509	2	1-12243	39509	10	71	3274	28	2012-08-02 11:18:53.620239
39520	3	1-12201	39519	89	45	104	58	2012-08-02 11:18:55.940371
39521	3	1-11328	39519	45	57	113	21	2012-08-02 11:18:55.940371
39519	3	1-12249	39519	57	89	86	61	2012-08-02 11:18:55.940371
39523	3	1-11180	39522	25	67	3249	42	2012-08-02 11:18:57.100269
39524	3	1-11835	39522	67	24	3533	32	2012-08-02 11:18:57.100269
39522	3	1-12252	39522	24	25	2923	75	2012-08-02 11:18:57.100269
39526	3	1-11676	39525	35	50	7029	69	2012-08-02 11:18:58.238171
39527	3	1-12245	39525	50	93	5448	81	2012-08-02 11:18:58.238171
39525	3	1-12255	39525	93	35	8141	59	2012-08-02 11:18:58.238171
39529	3	1-12228	39528	92	70	2324	64	2012-08-02 11:18:58.238171
39530	3	1-6465	39528	70	93	830	81	2012-08-02 11:18:58.238171
39528	3	1-12255	39528	93	92	1499	59	2012-08-02 11:18:58.238171
39549	3	1-11218	39548	52	12	655	6	2012-08-02 11:18:59.376264
39550	3	1-11586	39548	12	20	391	23	2012-08-02 11:18:59.376264
39548	3	1-12258	39548	20	52	772	64	2012-08-02 11:18:59.376264
39552	4	1-10016	39551	22	75	2956	49	2012-08-02 11:18:59.376264
39553	4	1-12099	39551	75	12	3450	6	2012-08-02 11:18:59.376264
39554	4	1-11586	39551	12	20	2210	23	2012-08-02 11:18:59.376264
39551	4	1-12258	39551	20	22	4681	64	2012-08-02 11:18:59.376264
39580	4	1-3333	39579	44	65	697	100	2012-08-02 11:19:01.720992
39581	4	1-12125	39579	65	26	1068	52	2012-08-02 11:19:01.720992
39582	4	1-11961	39579	26	53	1305	81	2012-08-02 11:19:01.720992
39579	4	1-12265	39579	53	44	1414	35	2012-08-02 11:19:01.720992
39584	4	1-11396	39583	100	65	1413	63	2012-08-02 11:19:01.720992
39585	4	1-8813	39583	65	26	1473	52	2012-08-02 11:19:01.720992
39586	4	1-11961	39583	26	53	1814	81	2012-08-02 11:19:01.720992
39583	4	1-12265	39583	53	100	1981	35	2012-08-02 11:19:01.720992
39606	4	1-11638	39605	47	100	651	80	2012-08-02 11:19:07.819161
39607	4	1-11906	39605	100	96	1085	70	2012-08-02 11:19:07.819161
39608	4	1-10616	39605	96	97	410	62	2012-08-02 11:19:07.819161
39605	4	1-12280	39605	97	47	1030	83	2012-08-02 11:19:07.819161
39610	4	1-11683	39609	16	83	3576	99	2012-08-02 11:19:07.819161
39611	4	1-10416	39609	83	14	2775	71	2012-08-02 11:19:07.819161
39612	4	1-10744	39609	14	97	1948	62	2012-08-02 11:19:07.819161
39609	4	1-12280	39609	97	16	6482	83	2012-08-02 11:19:07.819161
39618	4	1-12112	39617	93	90	1322	74	2012-08-02 11:19:10.128115
39619	4	1-11805	39617	90	85	2545	35	2012-08-02 11:19:10.128115
39620	4	1-9723	39617	85	48	2067	44	2012-08-02 11:19:10.128115
39617	4	1-12286	39617	48	93	1598	20	2012-08-02 11:19:10.128115
39660	3	1-11850	39659	8	15	2652	68	2012-08-02 11:19:19.94982
39661	3	1-10352	39659	15	97	723	59	2012-08-02 11:19:19.94982
39659	3	1-12311	39659	97	8	2345	97	2012-08-02 11:19:19.94982
39663	3	1-8904	39662	88	35	5958	92	2012-08-02 11:19:19.94982
39664	3	1-10680	39662	35	97	1719	59	2012-08-02 11:19:19.94982
39662	3	1-12311	39662	97	88	6160	97	2012-08-02 11:19:19.94982
39691	4	1-7257	39690	97	28	1309	10	2012-08-02 11:19:23.441636
39692	4	1-12176	39690	28	3	1215	74	2012-08-02 11:19:23.441636
39693	4	1-9754	39690	3	63	766	27	2012-08-02 11:19:23.441636
39690	4	1-12320	39690	63	97	5013	17	2012-08-02 11:19:23.441636
39714	4	1-12011	39713	68	1	893	79	2012-08-02 11:19:25.739336
39715	4	1-10456	39713	1	1	775	40	2012-08-02 11:19:25.739336
39716	4	1-12084	39713	1	7	1236	88	2012-08-02 11:19:25.739336
39713	4	1-12326	39713	7	68	399	10	2012-08-02 11:19:25.739336
39734	4	1-11560	39733	37	2	2682	63	2012-08-02 11:19:29.21965
39735	4	1-12232	39733	2	28	4207	41	2012-08-02 11:19:29.21965
39736	4	1-12196	39733	28	77	9455	86	2012-08-02 11:19:29.21965
39733	4	1-12335	39733	77	37	6423	66	2012-08-02 11:19:29.21965
39755	4	1-12029	39754	50	8	4763	94	2012-08-02 11:19:31.50836
39756	4	1-12231	39754	8	49	4324	13	2012-08-02 11:19:31.50836
39757	4	1-12332	39754	49	38	3311	11	2012-08-02 11:19:31.50836
39754	4	1-12341	39754	38	50	5098	89	2012-08-02 11:19:31.50836
39759	3	1-11173	39758	72	34	1502	59	2012-08-02 11:19:31.50836
39760	3	1-12183	39758	34	38	1327	11	2012-08-02 11:19:31.50836
39758	3	1-12341	39758	38	72	2282	89	2012-08-02 11:19:31.50836
39762	4	1-11874	39761	71	44	2604	46	2012-08-02 11:19:33.839735
39763	4	1-12256	39761	44	94	3332	76	2012-08-02 11:19:33.839735
39764	4	1-12144	39761	94	15	2209	79	2012-08-02 11:19:33.839735
39761	4	1-12347	39761	15	71	2988	86	2012-08-02 11:19:33.839735
39773	3	1-11879	39772	66	56	1409	42	2012-08-02 11:19:34.999848
39774	3	1-10598	39772	56	48	397	37	2012-08-02 11:19:34.999848
39772	3	1-12350	39772	48	66	2917	73	2012-08-02 11:19:34.999848
39776	4	1-11991	39775	46	81	3539	45	2012-08-02 11:19:37.286736
39777	4	1-5461	39775	81	79	431	26	2012-08-02 11:19:37.286736
39778	4	1-8453	39775	79	12	418	100	2012-08-02 11:19:37.286736
39775	4	1-12356	39775	12	46	4182	34	2012-08-02 11:19:37.286736
39779	1	1-6615	\N	84	84	574	45	2012-08-02 11:19:37.286736
39787	4	1-11897	39786	28	71	1880	8	2012-08-02 11:19:38.435714
39788	4	1-12165	39786	71	80	1656	20	2012-08-02 11:19:38.435714
39789	4	1-12292	39786	80	99	1043	36	2012-08-02 11:19:38.435714
39786	4	1-12359	39786	99	28	4021	9	2012-08-02 11:19:38.435714
39791	4	1-4234	39790	87	46	1872	46	2012-08-02 11:19:38.435714
39792	4	1-11866	39790	46	80	1522	20	2012-08-02 11:19:38.435714
39793	4	1-12292	39790	80	99	987	36	2012-08-02 11:19:38.435714
39790	4	1-12359	39790	99	87	3920	9	2012-08-02 11:19:38.435714
39795	4	1-12282	39794	84	38	3385	81	2012-08-02 11:19:39.584951
39796	4	1-9967	39794	38	38	647	36	2012-08-02 11:19:39.584951
39797	4	1-10484	39794	38	12	551	95	2012-08-02 11:19:39.584951
39794	4	1-12362	39794	12	84	2399	56	2012-08-02 11:19:39.584951
39802	4	1-10106	39801	42	53	1316	20	2012-08-02 11:19:40.745033
39803	4	1-11686	39801	53	23	731	8	2012-08-02 11:19:40.745033
39804	4	1-12055	39801	23	45	1736	26	2012-08-02 11:19:40.745033
39801	4	1-12365	39801	45	42	1343	75	2012-08-02 11:19:40.745033
39809	3	1-11766	39808	97	14	899	90	2012-08-02 11:19:43.031706
39808	3	1-12370	39808	64	97	862	63	2012-08-02 11:19:43.031706
39649	3	1-5515	39647	53	14	128	62	2012-08-02 11:19:16.72384
39651	4	1-10825	39650	23	57	2987	21	2012-08-02 11:19:16.72384
39652	4	1-12249	39650	57	90	1633	61	2012-08-02 11:19:16.72384
39653	4	1-11288	39650	90	14	1181	62	2012-08-02 11:19:16.72384
39650	4	1-12302	39650	14	23	7202	29	2012-08-02 11:19:16.72384
39669	4	1-8583	39668	44	10	1745	65	2012-08-02 11:19:20.723102
39670	4	1-12243	39668	10	71	1197	28	2012-08-02 11:19:20.723102
39671	4	1-12064	39668	71	91	2816	51	2012-08-02 11:19:20.723102
39668	4	1-12313	39668	91	44	3214	91	2012-08-02 11:19:20.723102
39673	4	1-10749	39672	49	16	593	7	2012-08-02 11:19:21.872433
39674	4	1-10685	39672	16	32	310	91	2012-08-02 11:19:21.872433
39675	4	1-8504	39672	32	32	141	68	2012-08-02 11:19:21.872433
39672	4	1-12316	39672	32	49	704	81	2012-08-02 11:19:21.872433
39677	4	1-2745	39676	77	92	248	8	2012-08-02 11:19:21.872433
39678	4	1-12169	39676	92	97	255	24	2012-08-02 11:19:21.872433
39679	4	1-12057	39676	97	32	423	68	2012-08-02 11:19:21.872433
39676	4	1-12316	39676	32	77	2220	81	2012-08-02 11:19:21.872433
39681	5	1-7981	39680	99	86	2069	41	2012-08-02 11:19:21.872433
39682	5	1-5916	39680	86	92	756	8	2012-08-02 11:19:21.872433
39683	5	1-12169	39680	92	97	784	24	2012-08-02 11:19:21.872433
39684	5	1-12057	39680	97	32	1313	68	2012-08-02 11:19:21.872433
39680	5	1-12316	39680	32	99	6943	81	2012-08-02 11:19:21.872433
39685	1	1-5916	\N	86	86	57	8	2012-08-02 11:19:21.872433
39702	4	1-11020	39701	74	19	326	19	2012-08-02 11:19:25.352516
39703	4	1-11652	39701	19	72	248	26	2012-08-02 11:19:25.352516
39704	4	1-12037	39701	72	20	208	67	2012-08-02 11:19:25.352516
39701	4	1-12325	39701	20	74	2778	94	2012-08-02 11:19:25.352516
39706	4	1-11226	39705	72	61	2038	82	2012-08-02 11:19:25.352516
39707	4	1-11831	39705	61	72	766	29	2012-08-02 11:19:25.352516
39708	4	1-11205	39705	72	20	408	67	2012-08-02 11:19:25.352516
39705	4	1-12325	39705	20	72	6405	94	2012-08-02 11:19:25.352516
39710	4	1-6993	39709	7	91	77	5	2012-08-02 11:19:25.352516
39711	4	1-12276	39709	91	51	61	49	2012-08-02 11:19:25.352516
39712	4	1-7659	39709	51	20	9	67	2012-08-02 11:19:25.352516
39709	4	1-12325	39709	20	7	170	94	2012-08-02 11:19:25.352516
39726	3	1-11183	39725	96	2	288	9	2012-08-02 11:19:26.512961
39727	3	1-7045	39725	2	91	74	19	2012-08-02 11:19:26.512961
39725	3	1-12328	39725	91	96	371	70	2012-08-02 11:19:26.512961
39728	1	1-10910	\N	51	51	3088	9	2012-08-02 11:19:26.512961
39738	4	1-11612	39737	80	7	1607	80	2012-08-02 11:19:31.131315
39739	4	1-9368	39737	7	89	435	61	2012-08-02 11:19:31.131315
39740	4	1-12201	39737	89	48	436	58	2012-08-02 11:19:31.131315
39737	4	1-12339	39737	48	80	1165	32	2012-08-02 11:19:31.131315
39742	3	1-7794	39741	71	89	1881	61	2012-08-02 11:19:31.131315
39743	3	1-12201	39741	89	48	1909	58	2012-08-02 11:19:31.131315
39741	3	1-12339	39741	48	71	5155	32	2012-08-02 11:19:31.131315
39745	3	1-9797	39744	21	22	959	6	2012-08-02 11:19:31.131315
39746	3	1-7364	39744	22	48	549	58	2012-08-02 11:19:31.131315
39744	3	1-12339	39744	48	21	1641	32	2012-08-02 11:19:31.131315
39748	4	1-8724	39747	22	34	364	59	2012-08-02 11:19:31.131315
39749	4	1-12183	39747	34	36	320	11	2012-08-02 11:19:31.131315
39750	4	1-8177	39747	36	48	312	58	2012-08-02 11:19:31.131315
39747	4	1-12339	39747	48	22	979	32	2012-08-02 11:19:31.131315
39766	3	1-11519	39765	92	13	186	65	2012-08-02 11:19:34.22667
39767	3	1-11103	39765	13	77	108	61	2012-08-02 11:19:34.22667
39765	3	1-12348	39765	77	92	368	24	2012-08-02 11:19:34.22667
39769	4	1-10816	39768	90	50	1106	78	2012-08-02 11:19:34.22667
39770	4	1-12275	39768	50	2	2157	39	2012-08-02 11:19:34.22667
39771	4	1-12117	39768	2	77	812	61	2012-08-02 11:19:34.22667
39768	4	1-12348	39768	77	90	2781	24	2012-08-02 11:19:34.22667
39781	2	1-11995	39780	22	49	1557	15	2012-08-02 11:19:37.662313
39780	2	1-12357	39780	49	22	607	1	2012-08-02 11:19:37.662313
39783	4	1-11587	39782	52	7	4653	79	2012-08-02 11:19:37.662313
39784	4	1-11075	39782	7	21	2698	100	2012-08-02 11:19:37.662313
39785	4	1-12314	39782	21	49	5759	15	2012-08-02 11:19:37.662313
39782	4	1-12357	39782	49	52	4860	1	2012-08-02 11:19:37.662313
39799	3	1-12357	39798	49	52	2808	1	2012-08-02 11:19:39.971519
39800	3	1-4139	39798	52	59	2407	2	2012-08-02 11:19:39.971519
39798	3	1-12363	39798	59	49	3249	15	2012-08-02 11:19:39.971519
39806	3	1-11627	39805	17	10	3241	67	2012-08-02 11:19:42.280544
39807	3	1-11988	39805	10	38	6445	16	2012-08-02 11:19:42.280544
39805	3	1-12369	39805	38	17	2372	74	2012-08-02 11:19:42.280544
39816	2	1-11310	39815	24	36	3128	13	2012-08-02 11:19:43.441353
39815	2	1-12372	39815	36	24	6791	39	2012-08-02 11:19:43.441353
39818	4	1-7347	39817	6	19	1524	95	2012-08-02 11:19:44.623779
39819	4	1-12022	39817	19	15	1327	79	2012-08-02 11:19:44.623779
39820	4	1-12347	39817	15	54	1903	86	2012-08-02 11:19:44.623779
39817	4	1-12375	39817	54	6	1955	1	2012-08-02 11:19:44.623779
39825	4	1-11222	39824	13	55	1358	6	2012-08-02 11:19:46.954926
39826	4	1-12172	39824	55	100	957	82	2012-08-02 11:19:46.954926
39827	4	1-11842	39824	100	86	1088	80	2012-08-02 11:19:46.954926
39824	4	1-12381	39824	86	13	1778	21	2012-08-02 11:19:46.954926
39860	2	1-5543	39859	66	51	323	11	2012-08-02 11:19:50.435086
39859	2	1-12390	39859	51	66	7248	24	2012-08-02 11:19:50.435086
39887	3	1-12011	39886	68	79	2131	79	2012-08-02 11:19:55.848666
39888	3	1-12212	39886	79	99	1706	94	2012-08-02 11:19:55.848666
39886	3	1-12404	39886	99	68	1272	10	2012-08-02 11:19:55.848666
39890	4	1-12176	39889	28	78	94	74	2012-08-02 11:19:55.848666
39891	4	1-10808	39889	78	36	65	97	2012-08-02 11:19:55.848666
39892	4	1-10582	39889	36	99	71	94	2012-08-02 11:19:55.848666
39889	4	1-12404	39889	99	28	71	10	2012-08-02 11:19:55.848666
39897	5	1-6887	39896	93	91	4308	5	2012-08-02 11:19:57.030953
39898	5	1-12276	39896	91	60	4214	49	2012-08-02 11:19:57.030953
39899	5	1-10997	39896	60	50	1355	78	2012-08-02 11:19:57.030953
39900	5	1-12275	39896	50	16	2870	39	2012-08-02 11:19:57.030953
39896	5	1-12407	39896	16	93	8487	43	2012-08-02 11:19:57.030953
39902	3	1-6318	39901	1	47	234	59	2012-08-02 11:19:57.030953
39903	3	1-12222	39901	47	16	151	39	2012-08-02 11:19:57.030953
39901	3	1-12407	39901	16	1	514	43	2012-08-02 11:19:57.030953
39905	4	1-9570	39904	66	68	372	81	2012-08-02 11:19:57.030953
39906	4	1-11594	39904	68	66	351	40	2012-08-02 11:19:57.030953
39907	4	1-11527	39904	66	16	116	39	2012-08-02 11:19:57.030953
39904	4	1-12407	39904	16	66	413	43	2012-08-02 11:19:57.030953
39810	3	1-10459	39808	14	64	513	98	2012-08-02 11:19:43.031706
39812	4	1-12232	39811	2	91	118	41	2012-08-02 11:19:43.031706
39813	4	1-5037	39811	91	54	43	95	2012-08-02 11:19:43.031706
39814	4	1-12101	39811	54	64	41	98	2012-08-02 11:19:43.031706
39811	4	1-12370	39811	64	2	80	63	2012-08-02 11:19:43.031706
39822	3	1-12249	39821	57	64	1507	61	2012-08-02 11:19:45.021359
39823	3	1-12308	39821	64	73	1941	69	2012-08-02 11:19:45.021359
39821	3	1-12376	39821	73	57	2165	21	2012-08-02 11:19:45.021359
39842	4	1-11603	39841	83	80	2890	20	2012-08-02 11:19:47.728103
39843	4	1-12292	39841	80	18	2164	36	2012-08-02 11:19:47.728103
39844	4	1-9525	39841	18	32	1419	78	2012-08-02 11:19:47.728103
39841	4	1-12383	39841	32	83	5495	7	2012-08-02 11:19:47.728103
39849	3	1-12333	39848	37	93	999	66	2012-08-02 11:19:50.059928
39850	3	1-9320	39848	93	78	612	5	2012-08-02 11:19:50.059928
39848	3	1-12389	39848	78	37	998	17	2012-08-02 11:19:50.059928
39852	4	1-12333	39851	37	63	2251	66	2012-08-02 11:19:50.059928
39853	4	1-6484	39851	63	97	652	98	2012-08-02 11:19:50.059928
39854	4	1-12077	39851	97	78	940	5	2012-08-02 11:19:50.059928
39851	4	1-12389	39851	78	37	1856	17	2012-08-02 11:19:50.059928
39856	4	1-11724	39855	37	75	1535	54	2012-08-02 11:19:50.059928
39857	4	1-12128	39855	75	83	1016	84	2012-08-02 11:19:50.059928
39858	4	1-11732	39855	83	78	979	5	2012-08-02 11:19:50.059928
39855	4	1-12389	39855	78	37	2060	17	2012-08-02 11:19:50.059928
39894	3	1-11176	39893	84	75	4127	16	2012-08-02 11:19:56.268626
39895	3	1-11916	39893	75	5	3861	42	2012-08-02 11:19:56.268626
39893	3	1-12405	39893	5	84	4375	100	2012-08-02 11:19:56.268626
39909	4	1-10862	39908	32	26	157	52	2012-08-02 11:19:57.417779
39910	4	1-11961	39908	26	53	183	81	2012-08-02 11:19:57.417779
39911	4	1-12265	39908	53	11	189	35	2012-08-02 11:19:57.417779
39908	4	1-12408	39908	11	32	149	48	2012-08-02 11:19:57.417779
39913	3	1-10878	39912	48	97	752	87	2012-08-02 11:19:58.568741
39914	3	1-11930	39912	97	22	250	26	2012-08-02 11:19:58.568741
39912	3	1-12411	39912	22	48	1284	66	2012-08-02 11:19:58.568741
39947	4	1-12402	39946	12	92	3399	8	2012-08-02 11:20:02.026832
39948	4	1-12169	39946	92	99	2917	24	2012-08-02 11:20:02.026832
39949	4	1-8135	39946	99	10	2307	66	2012-08-02 11:20:02.026832
39946	4	1-12420	39946	10	12	4383	23	2012-08-02 11:20:02.026832
39951	3	1-12130	39950	8	5	3338	89	2012-08-02 11:20:02.026832
39952	3	1-12334	39950	5	10	1189	66	2012-08-02 11:20:02.026832
39950	3	1-12420	39950	10	8	2686	23	2012-08-02 11:20:02.026832
39969	4	1-12389	39968	78	65	3156	17	2012-08-02 11:20:03.165061
39970	4	1-12384	39968	65	13	660	33	2012-08-02 11:20:03.165061
39971	4	1-11699	39968	13	38	1987	21	2012-08-02 11:20:03.165061
39968	4	1-12423	39968	38	78	1480	5	2012-08-02 11:20:03.165061
39977	3	1-7375	39976	32	89	871	42	2012-08-02 11:20:04.336464
39978	3	1-11837	39976	89	94	558	64	2012-08-02 11:20:04.336464
39976	3	1-12426	39976	94	32	1335	18	2012-08-02 11:20:04.336464
39980	4	1-8819	39979	24	8	1084	39	2012-08-02 11:20:04.336464
39981	4	1-12306	39979	8	13	1992	80	2012-08-02 11:20:04.336464
39982	4	1-11931	39979	13	94	997	64	2012-08-02 11:20:04.336464
39979	4	1-12426	39979	94	24	2723	18	2012-08-02 11:20:04.336464
39984	4	1-9505	39983	8	75	217	16	2012-08-02 11:20:04.336464
39985	4	1-11916	39983	75	89	209	42	2012-08-02 11:20:04.336464
39986	4	1-11837	39983	89	94	159	64	2012-08-02 11:20:04.336464
39983	4	1-12426	39983	94	8	453	18	2012-08-02 11:20:04.336464
39987	1	1-9658	\N	65	65	1964	80	2012-08-02 11:20:04.336464
39988	1	1-5339	\N	9	9	824	80	2012-08-02 11:20:04.336464
39989	1	1-8444	\N	46	46	2598	80	2012-08-02 11:20:04.336464
40028	4	1-11498	40027	98	30	1856	13	2012-08-02 11:20:07.430521
40029	4	1-11198	40027	30	91	1203	51	2012-08-02 11:20:07.430521
40030	4	1-12313	40027	91	96	1020	91	2012-08-02 11:20:07.430521
40027	4	1-12434	40027	96	98	3174	55	2012-08-02 11:20:07.430521
40032	4	1-10067	40031	22	15	124	54	2012-08-02 11:20:07.430521
40033	4	1-12409	40031	15	39	84	21	2012-08-02 11:20:07.430521
40034	4	1-12001	40031	39	96	55	91	2012-08-02 11:20:07.430521
40031	4	1-12434	40031	96	22	177	55	2012-08-02 11:20:07.430521
40036	4	1-12204	40035	47	98	3588	1	2012-08-02 11:20:07.430521
40037	4	1-5555	40035	98	72	2938	53	2012-08-02 11:20:07.430521
40038	4	1-11977	40035	72	96	1908	91	2012-08-02 11:20:07.430521
40035	4	1-12434	40035	96	47	6411	55	2012-08-02 11:20:07.430521
40039	1	1-5735	\N	79	79	1167	51	2012-08-02 11:20:07.430521
40045	3	1-12419	40044	68	54	1913	95	2012-08-02 11:20:08.601905
40046	3	1-12101	40044	54	51	799	98	2012-08-02 11:20:08.601905
40044	3	1-12437	40044	51	68	1437	57	2012-08-02 11:20:08.601905
40048	3	1-7918	40047	94	16	407	44	2012-08-02 11:20:08.601905
40049	3	1-12030	40047	16	51	269	98	2012-08-02 11:20:08.601905
40047	3	1-12437	40047	51	94	965	57	2012-08-02 11:20:08.601905
40051	4	1-10468	40050	95	68	462	6	2012-08-02 11:20:08.601905
40052	4	1-7825	40050	68	16	318	44	2012-08-02 11:20:08.601905
40053	4	1-12030	40050	16	51	229	98	2012-08-02 11:20:08.601905
40050	4	1-12437	40050	51	95	892	57	2012-08-02 11:20:08.601905
40058	4	1-9763	40057	11	69	185	73	2012-08-02 11:20:09.761988
40059	4	1-12438	40057	69	80	118	63	2012-08-02 11:20:09.761988
40060	4	1-10929	40057	80	12	177	88	2012-08-02 11:20:09.761988
40057	4	1-12440	40057	12	11	384	93	2012-08-02 11:20:09.761988
40069	2	1-666	40068	13	28	606	46	2012-08-02 11:20:14.432725
40068	2	1-12452	40068	28	13	1310	30	2012-08-02 11:20:14.432725
40071	4	1-12340	40070	93	31	1378	10	2012-08-02 11:20:14.432725
40072	4	1-12004	40070	31	87	1609	9	2012-08-02 11:20:14.432725
40073	4	1-4234	40070	87	28	880	46	2012-08-02 11:20:14.432725
40070	4	1-12452	40070	28	93	1937	30	2012-08-02 11:20:14.432725
40075	4	1-11845	40074	21	69	3271	99	2012-08-02 11:20:14.432725
40076	4	1-11733	40074	69	87	1847	9	2012-08-02 11:20:14.432725
40077	4	1-4234	40074	87	28	1163	46	2012-08-02 11:20:14.432725
40074	4	1-12452	40074	28	21	2950	30	2012-08-02 11:20:14.432725
40078	1	1-4234	\N	87	87	366	46	2012-08-02 11:20:14.432725
40079	1	1-5232	\N	46	46	231	46	2012-08-02 11:20:14.432725
40080	1	1-4329	\N	78	78	663	46	2012-08-02 11:20:14.432725
40106	3	1-12306	40105	8	16	1549	80	2012-08-02 11:20:16.785814
40107	3	1-11365	40105	16	34	456	46	2012-08-02 11:20:16.785814
40105	3	1-12458	40105	34	8	1211	39	2012-08-02 11:20:16.785814
40109	4	1-12403	40108	33	10	1635	65	2012-08-02 11:20:16.785814
40110	4	1-12243	40108	10	14	767	28	2012-08-02 11:20:16.785814
40111	4	1-12355	40108	14	34	642	46	2012-08-02 11:20:16.785814
39837	4	1-12347	39836	15	91	1120	86	2012-08-02 11:19:47.352458
39838	4	1-12132	39836	91	77	214	3	2012-08-02 11:19:47.352458
39839	4	1-8077	39836	77	23	212	60	2012-08-02 11:19:47.352458
39836	4	1-12382	39836	23	15	865	79	2012-08-02 11:19:47.352458
39840	1	1-4258	\N	70	70	4318	60	2012-08-02 11:19:47.352458
39846	3	1-10218	39845	31	42	808	75	2012-08-02 11:19:48.501798
39847	3	1-10106	39845	42	88	266	20	2012-08-02 11:19:48.501798
39845	3	1-12385	39845	88	31	6596	79	2012-08-02 11:19:48.501798
39862	4	1-11449	39861	74	93	2941	20	2012-08-02 11:19:51.970866
39863	4	1-12112	39861	93	17	2124	74	2012-08-02 11:19:51.970866
39864	4	1-11627	39861	17	72	2569	67	2012-08-02 11:19:51.970866
39861	4	1-12394	39861	72	74	3793	11	2012-08-02 11:19:51.970866
39866	4	1-12223	39865	88	97	6116	87	2012-08-02 11:19:51.970866
39867	4	1-11930	39865	97	72	1863	26	2012-08-02 11:19:51.970866
39868	4	1-12037	39865	72	72	3113	67	2012-08-02 11:19:51.970866
39865	4	1-12394	39865	72	88	4713	11	2012-08-02 11:19:51.970866
39870	3	1-7724	39869	56	72	397	29	2012-08-02 11:19:51.970866
39871	3	1-11205	39869	72	72	274	67	2012-08-02 11:19:51.970866
39869	3	1-12394	39869	72	56	458	11	2012-08-02 11:19:51.970866
39873	4	1-6176	39872	89	9	559	22	2012-08-02 11:19:53.164283
39874	4	1-8273	39872	9	28	148	10	2012-08-02 11:19:53.164283
39875	4	1-12176	39872	28	48	112	74	2012-08-02 11:19:53.164283
39872	4	1-12397	39872	48	89	3374	25	2012-08-02 11:19:53.164283
39877	5	1-12296	39876	36	92	1453	8	2012-08-02 11:19:55.462086
39878	5	1-12169	39876	92	90	1345	24	2012-08-02 11:19:55.462086
39879	5	1-10816	39876	90	50	702	78	2012-08-02 11:19:55.462086
39880	5	1-12275	39876	50	33	1796	39	2012-08-02 11:19:55.462086
39876	5	1-12403	39876	33	36	1995	65	2012-08-02 11:19:55.462086
39882	5	1-9371	39881	20	10	256	97	2012-08-02 11:19:55.462086
39883	5	1-12388	39881	10	60	250	49	2012-08-02 11:19:55.462086
39884	5	1-10997	39881	60	50	106	78	2012-08-02 11:19:55.462086
39885	5	1-12275	39881	50	33	296	39	2012-08-02 11:19:55.462086
39881	5	1-12403	39881	33	20	358	65	2012-08-02 11:19:55.462086
39954	4	1-9040	39953	13	84	3301	92	2012-08-02 11:20:02.402723
39955	4	1-12180	39953	84	59	3055	44	2012-08-02 11:20:02.402723
39956	4	1-11315	39953	59	83	2291	45	2012-08-02 11:20:02.402723
39953	4	1-12421	39953	83	13	2782	14	2012-08-02 11:20:02.402723
39958	4	1-11898	39957	62	46	1625	66	2012-08-02 11:20:02.402723
39959	4	1-12417	39957	46	22	1595	13	2012-08-02 11:20:02.402723
39960	4	1-12319	39957	22	83	1600	45	2012-08-02 11:20:02.402723
39957	4	1-12421	39957	83	62	1960	14	2012-08-02 11:20:02.402723
39973	4	1-12249	39972	57	64	1415	61	2012-08-02 11:20:03.551811
39974	4	1-12308	39972	64	68	1862	69	2012-08-02 11:20:03.551811
39975	4	1-12157	39972	68	81	1440	7	2012-08-02 11:20:03.551811
39972	4	1-12424	39972	81	57	1991	21	2012-08-02 11:20:03.551811
39991	3	1-10674	39990	31	8	531	80	2012-08-02 11:20:04.723067
39992	3	1-11772	39990	8	29	567	53	2012-08-02 11:20:04.723067
39990	3	1-12427	39990	29	31	1119	16	2012-08-02 11:20:04.723067
40009	3	1-11794	40008	8	43	294	73	2012-08-02 11:20:05.872235
40010	3	1-11811	40008	43	75	587	15	2012-08-02 11:20:05.872235
40008	3	1-12430	40008	75	8	1028	76	2012-08-02 11:20:05.872235
40012	3	1-12331	40011	58	13	560	50	2012-08-02 11:20:05.872235
40013	3	1-12026	40011	13	75	315	15	2012-08-02 11:20:05.872235
40011	3	1-12430	40011	75	58	707	76	2012-08-02 11:20:05.872235
40015	4	1-9076	40014	3	54	752	95	2012-08-02 11:20:07.043702
40016	4	1-12101	40014	54	64	644	98	2012-08-02 11:20:07.043702
40017	4	1-12370	40014	64	78	1128	63	2012-08-02 11:20:07.043702
40014	4	1-12432	40014	78	3	1607	97	2012-08-02 11:20:07.043702
40019	5	1-9160	40018	20	28	952	19	2012-08-02 11:20:07.043702
40020	5	1-11722	40018	28	54	2203	95	2012-08-02 11:20:07.043702
40021	5	1-12101	40018	54	64	1887	98	2012-08-02 11:20:07.043702
40022	5	1-12370	40018	64	78	3305	63	2012-08-02 11:20:07.043702
40018	5	1-12432	40018	78	20	4710	97	2012-08-02 11:20:07.043702
40024	4	1-3918	40023	100	16	287	44	2012-08-02 11:20:07.043702
40025	4	1-12030	40023	16	64	225	98	2012-08-02 11:20:07.043702
40026	4	1-12370	40023	64	78	454	63	2012-08-02 11:20:07.043702
40023	4	1-12432	40023	78	100	746	97	2012-08-02 11:20:07.043702
40041	4	1-10637	40040	8	68	1911	57	2012-08-02 11:20:07.83926
40042	4	1-12419	40040	68	21	3450	95	2012-08-02 11:20:07.83926
40043	4	1-9740	40040	21	81	1387	48	2012-08-02 11:20:07.83926
40040	4	1-12435	40040	81	8	4253	33	2012-08-02 11:20:07.83926
40062	4	1-8506	40061	53	53	4102	10	2012-08-02 11:20:12.47991
40063	4	1-12127	40061	53	30	2638	30	2012-08-02 11:20:12.47991
40064	4	1-12107	40061	30	12	2192	93	2012-08-02 11:20:12.47991
40061	4	1-12447	40061	12	53	5573	19	2012-08-02 11:20:12.47991
40082	4	1-7418	40081	43	64	1001	98	2012-08-02 11:20:14.819348
40083	4	1-12370	40081	64	72	1416	63	2012-08-02 11:20:14.819348
40084	4	1-11136	40081	72	36	906	23	2012-08-02 11:20:14.819348
40081	4	1-12453	40081	36	43	2000	19	2012-08-02 11:20:14.819348
40086	3	1-11535	40085	74	15	3335	22	2012-08-02 11:20:14.819348
40087	3	1-10053	40085	15	36	1309	23	2012-08-02 11:20:14.819348
40085	3	1-12453	40085	36	74	2949	19	2012-08-02 11:20:14.819348
40089	4	1-12205	40088	93	98	1470	12	2012-08-02 11:20:14.819348
40090	4	1-12002	40088	98	59	913	49	2012-08-02 11:20:14.819348
40091	4	1-9120	40088	59	36	406	23	2012-08-02 11:20:14.819348
40088	4	1-12453	40088	36	93	1035	19	2012-08-02 11:20:14.819348
40093	4	1-6991	40092	8	46	111	44	2012-08-02 11:20:14.819348
40094	4	1-12023	40092	46	12	101	6	2012-08-02 11:20:14.819348
40095	4	1-11586	40092	12	36	58	23	2012-08-02 11:20:14.819348
40092	4	1-12453	40092	36	8	160	19	2012-08-02 11:20:14.819348
40097	4	1-11916	40096	75	27	2768	42	2012-08-02 11:20:16.022967
40098	4	1-12396	40096	27	78	2892	90	2012-08-02 11:20:16.022967
40099	4	1-10564	40096	78	28	2417	12	2012-08-02 11:20:16.022967
40096	4	1-12456	40096	28	75	3244	16	2012-08-02 11:20:16.022967
40101	3	1-10734	40100	59	36	23	31	2012-08-02 11:20:16.022967
40102	3	1-11643	40100	36	28	51	12	2012-08-02 11:20:16.022967
40100	3	1-12456	40100	28	59	82	16	2012-08-02 11:20:16.022967
40121	4	1-12211	40120	62	36	973	31	2012-08-02 11:20:17.19583
40122	4	1-11643	40120	36	40	1032	12	2012-08-02 11:20:17.19583
40123	4	1-12414	40120	40	8	323	66	2012-08-02 11:20:17.19583
40120	4	1-12459	40120	8	62	1077	26	2012-08-02 11:20:17.19583
40125	4	1-7244	40124	76	53	2038	81	2012-08-02 11:20:17.19583
40126	4	1-12265	40124	53	56	1567	35	2012-08-02 11:20:17.19583
40127	4	1-10345	40124	56	8	1051	66	2012-08-02 11:20:17.19583
39916	5	1-11249	39915	35	68	1538	36	2012-08-02 11:19:59.331014
39917	5	1-11944	39915	68	62	1551	26	2012-08-02 11:19:59.331014
39918	5	1-12211	39915	62	77	2806	31	2012-08-02 11:19:59.331014
39919	5	1-12323	39915	77	25	2264	33	2012-08-02 11:19:59.331014
39915	5	1-12413	39915	25	35	5374	79	2012-08-02 11:19:59.331014
39921	4	1-11433	39920	84	44	1049	46	2012-08-02 11:20:00.480137
39922	4	1-12256	39920	44	58	970	76	2012-08-02 11:20:00.480137
39923	4	1-12331	39920	58	40	660	50	2012-08-02 11:20:00.480137
39920	4	1-12416	39920	40	84	1876	89	2012-08-02 11:20:00.480137
39925	4	1-12079	39924	35	39	2774	53	2012-08-02 11:20:00.480137
39926	4	1-11798	39924	39	38	2479	12	2012-08-02 11:20:00.480137
39927	4	1-11964	39924	38	40	1109	50	2012-08-02 11:20:00.480137
39924	4	1-12416	39924	40	35	3510	89	2012-08-02 11:20:00.480137
39929	4	1-11173	39928	72	88	1309	59	2012-08-02 11:20:00.480137
39930	4	1-11972	39928	88	81	752	1	2012-08-02 11:20:00.480137
39931	4	1-11419	39928	81	40	618	50	2012-08-02 11:20:00.480137
39928	4	1-12416	39928	40	72	2438	89	2012-08-02 11:20:00.480137
39933	4	1-8364	39932	39	94	4	17	2012-08-02 11:20:00.480137
39934	4	1-12059	39932	94	3	2	23	2012-08-02 11:20:00.480137
39935	4	1-11696	39932	3	40	2	50	2012-08-02 11:20:00.480137
39932	4	1-12416	39932	40	39	6	89	2012-08-02 11:20:00.480137
39937	4	1-11891	39936	5	35	211	47	2012-08-02 11:20:01.662484
39938	4	1-11461	39936	35	70	114	61	2012-08-02 11:20:01.662484
39939	4	1-7704	39936	70	68	77	57	2012-08-02 11:20:01.662484
39936	4	1-12419	39936	68	5	221	95	2012-08-02 11:20:01.662484
39941	3	1-11891	39940	5	58	2776	47	2012-08-02 11:20:01.662484
39942	3	1-4963	39940	58	68	869	57	2012-08-02 11:20:01.662484
39940	3	1-12419	39940	68	5	2692	95	2012-08-02 11:20:01.662484
39944	3	1-12022	39943	19	68	1216	79	2012-08-02 11:20:01.662484
39945	3	1-8293	39943	68	68	447	57	2012-08-02 11:20:01.662484
39943	3	1-12419	39943	68	19	1421	95	2012-08-02 11:20:01.662484
39962	3	1-12078	39961	28	31	4214	38	2012-08-02 11:20:02.789408
39963	3	1-12208	39961	31	67	5634	17	2012-08-02 11:20:02.789408
39961	3	1-12422	39961	67	28	5037	49	2012-08-02 11:20:02.789408
39965	4	1-9328	39964	73	74	401	96	2012-08-02 11:20:02.789408
39966	4	1-8851	39964	74	78	550	5	2012-08-02 11:20:02.789408
39967	4	1-12389	39964	78	67	1228	17	2012-08-02 11:20:02.789408
39964	4	1-12422	39964	67	73	1508	49	2012-08-02 11:20:02.789408
39994	2	1-10861	39993	80	46	103	54	2012-08-02 11:20:05.098618
39993	2	1-12428	39993	46	80	109	24	2012-08-02 11:20:05.098618
39996	2	1-8675	39995	99	46	3318	54	2012-08-02 11:20:05.098618
39995	2	1-12428	39995	46	99	4319	24	2012-08-02 11:20:05.098618
39998	3	1-11725	39997	55	11	333	25	2012-08-02 11:20:05.098618
39999	3	1-2380	39997	11	46	100	54	2012-08-02 11:20:05.098618
39997	3	1-12428	39997	46	55	149	24	2012-08-02 11:20:05.098618
40001	4	1-3270	40000	49	35	48	42	2012-08-02 11:20:05.098618
40002	4	1-11251	40000	35	52	23	41	2012-08-02 11:20:05.098618
40003	4	1-12358	40000	52	46	56	54	2012-08-02 11:20:05.098618
40000	4	1-12428	40000	46	49	88	24	2012-08-02 11:20:05.098618
40005	4	1-2569	40004	34	35	1131	42	2012-08-02 11:20:05.098618
40006	4	1-11251	40004	35	52	551	41	2012-08-02 11:20:05.098618
40007	4	1-12358	40004	52	46	1360	54	2012-08-02 11:20:05.098618
40004	4	1-12428	40004	46	34	2176	24	2012-08-02 11:20:05.098618
40055	3	1-11914	40054	55	13	1484	64	2012-08-02 11:20:09.364057
40056	3	1-12076	40054	13	73	607	62	2012-08-02 11:20:09.364057
40054	3	1-12439	40054	73	55	3581	87	2012-08-02 11:20:09.364057
40066	3	1-12005	40065	51	12	706	88	2012-08-02 11:20:12.869762
40067	3	1-12440	40065	12	4	709	93	2012-08-02 11:20:12.869762
40065	3	1-12448	40065	4	51	6274	96	2012-08-02 11:20:12.869762
40104	2	1-12262	40103	99	78	978	63	2012-08-02 11:20:16.398432
40103	2	1-12457	40103	78	99	1029	9	2012-08-02 11:20:16.398432
40129	4	1-5373	40128	84	50	1651	59	2012-08-02 11:20:18.808868
40130	4	1-12299	40128	50	39	1790	53	2012-08-02 11:20:18.808868
40131	4	1-11798	40128	39	1	2447	12	2012-08-02 11:20:18.808868
40128	4	1-12462	40128	1	84	2848	64	2012-08-02 11:20:18.808868
40149	3	1-12289	40148	63	21	198	1	2012-08-02 11:20:20.732595
40150	3	1-7669	40148	21	63	111	97	2012-08-02 11:20:20.732595
40148	3	1-12468	40148	63	63	228	33	2012-08-02 11:20:20.732595
40199	4	1-10863	40198	21	97	198	87	2012-08-02 11:20:25.436242
40200	4	1-11930	40198	97	72	76	26	2012-08-02 11:20:25.436242
40201	4	1-12037	40198	72	11	130	67	2012-08-02 11:20:25.436242
40198	4	1-12480	40198	11	21	300	38	2012-08-02 11:20:25.436242
40251	3	1-10179	40250	68	75	1781	30	2012-08-02 11:20:29.320022
40252	3	1-9196	40250	75	4	577	93	2012-08-02 11:20:29.320022
40250	3	1-12490	40250	4	68	6431	28	2012-08-02 11:20:29.320022
40258	4	1-11181	40257	10	84	1489	56	2012-08-02 11:20:30.504167
40259	4	1-12282	40257	84	5	3397	81	2012-08-02 11:20:30.504167
40260	4	1-11478	40257	5	85	986	27	2012-08-02 11:20:30.504167
40257	4	1-12494	40257	85	10	2637	72	2012-08-02 11:20:30.504167
40266	4	1-12489	40265	67	60	515	10	2012-08-02 11:20:32.051483
40267	4	1-11762	40265	60	74	361	96	2012-08-02 11:20:32.051483
40268	4	1-8851	40265	74	86	419	5	2012-08-02 11:20:32.051483
40265	4	1-12498	40265	86	67	223	84	2012-08-02 11:20:32.051483
40295	2	1-1308	40294	52	38	692	90	2012-08-02 11:20:34.782688
40294	2	1-12504	40294	38	52	1754	30	2012-08-02 11:20:34.782688
40297	4	1-10242	40296	55	100	1381	80	2012-08-02 11:20:34.782688
40298	4	1-11906	40296	100	40	2207	70	2012-08-02 11:20:34.782688
40299	4	1-11636	40296	40	38	877	90	2012-08-02 11:20:34.782688
40296	4	1-12504	40296	38	55	2482	30	2012-08-02 11:20:34.782688
40301	4	1-9355	40300	87	47	801	86	2012-08-02 11:20:34.782688
40302	4	1-8776	40300	47	27	359	42	2012-08-02 11:20:34.782688
40303	4	1-12396	40300	27	38	362	90	2012-08-02 11:20:34.782688
40300	4	1-12504	40300	38	87	1105	30	2012-08-02 11:20:34.782688
40305	4	1-1091	40304	82	75	688	16	2012-08-02 11:20:34.782688
40306	4	1-11916	40304	75	27	575	42	2012-08-02 11:20:34.782688
40307	4	1-12396	40304	27	38	589	90	2012-08-02 11:20:34.782688
40304	4	1-12504	40304	38	82	1824	30	2012-08-02 11:20:34.782688
40309	4	1-11581	40308	74	55	231	42	2012-08-02 11:20:35.943322
40310	4	1-12429	40308	55	96	155	11	2012-08-02 11:20:35.943322
40311	4	1-10849	40308	96	81	110	56	2012-08-02 11:20:35.943322
40308	4	1-12507	40308	81	74	305	48	2012-08-02 11:20:35.943322
40313	2	1-10610	40312	20	56	319	32	2012-08-02 11:20:38.263355
40312	2	1-12513	40312	56	20	9916	24	2012-08-02 11:20:38.263355
40320	2	1-12527	40320	48	59	7190	4	2012-08-02 11:20:43.717401
40108	4	1-12458	40108	34	33	1927	39	2012-08-02 11:20:16.785814
40113	4	1-10540	40112	53	11	2075	64	2012-08-02 11:20:16.785814
40114	4	1-11349	40112	11	16	1904	80	2012-08-02 11:20:16.785814
40115	4	1-11365	40112	16	34	789	46	2012-08-02 11:20:16.785814
40112	4	1-12458	40112	34	53	2949	39	2012-08-02 11:20:16.785814
40117	4	1-12117	40116	2	86	540	61	2012-08-02 11:20:16.785814
40118	4	1-6531	40116	86	49	269	73	2012-08-02 11:20:16.785814
40119	4	1-9717	40116	49	34	268	46	2012-08-02 11:20:16.785814
40116	4	1-12458	40116	34	2	1073	39	2012-08-02 11:20:16.785814
40135	3	1-12107	40134	30	100	439	93	2012-08-02 11:20:20.346337
40136	3	1-10065	40134	100	92	290	60	2012-08-02 11:20:20.346337
40134	3	1-12467	40134	92	30	356	30	2012-08-02 11:20:20.346337
40138	4	1-12340	40137	93	4	3512	10	2012-08-02 11:20:20.346337
40139	4	1-12330	40137	4	62	5611	82	2012-08-02 11:20:20.346337
40140	4	1-12181	40137	62	92	3173	60	2012-08-02 11:20:20.346337
40137	4	1-12467	40137	92	93	4633	30	2012-08-02 11:20:20.346337
40142	4	1-12340	40141	93	79	416	10	2012-08-02 11:20:20.346337
40143	4	1-10466	40141	79	9	348	62	2012-08-02 11:20:20.346337
40144	4	1-7449	40141	9	92	332	60	2012-08-02 11:20:20.346337
40141	4	1-12467	40141	92	93	516	30	2012-08-02 11:20:20.346337
40146	3	1-9196	40145	75	100	70	93	2012-08-02 11:20:20.346337
40147	3	1-10065	40145	100	92	59	60	2012-08-02 11:20:20.346337
40145	3	1-12467	40145	92	75	93	30	2012-08-02 11:20:20.346337
40160	4	1-12232	40159	2	28	71	41	2012-08-02 11:20:21.542331
40161	4	1-12196	40159	28	63	91	86	2012-08-02 11:20:21.542331
40162	4	1-12269	40159	63	49	93	92	2012-08-02 11:20:21.542331
40159	4	1-12470	40159	49	2	80	63	2012-08-02 11:20:21.542331
40164	3	1-11157	40163	27	63	2043	86	2012-08-02 11:20:21.542331
40165	3	1-12269	40163	63	49	2853	92	2012-08-02 11:20:21.542331
40163	3	1-12470	40163	49	27	3349	63	2012-08-02 11:20:21.542331
40171	4	1-12402	40170	12	3	1551	8	2012-08-02 11:20:23.865399
40172	4	1-11776	40170	3	21	2543	72	2012-08-02 11:20:23.865399
40173	4	1-7295	40170	21	3	2383	2	2012-08-02 11:20:23.865399
40170	4	1-12476	40170	3	12	1893	23	2012-08-02 11:20:23.865399
40175	4	1-8524	40174	89	72	1233	40	2012-08-02 11:20:23.865399
40176	4	1-12361	40174	72	24	334	18	2012-08-02 11:20:23.865399
40177	4	1-11427	40174	24	3	1091	2	2012-08-02 11:20:23.865399
40174	4	1-12476	40174	3	89	923	23	2012-08-02 11:20:23.865399
40178	1	1-3732	\N	39	39	1758	40	2012-08-02 11:20:23.865399
40190	4	1-9910	40189	48	51	894	9	2012-08-02 11:20:25.038434
40191	4	1-7206	40189	51	35	813	42	2012-08-02 11:20:25.038434
40192	4	1-11251	40189	35	71	402	41	2012-08-02 11:20:25.038434
40189	4	1-12479	40189	71	48	2065	14	2012-08-02 11:20:25.038434
40194	4	1-12071	40193	99	91	792	13	2012-08-02 11:20:25.038434
40195	4	1-10515	40193	91	35	640	42	2012-08-02 11:20:25.038434
40196	4	1-11251	40193	35	71	320	41	2012-08-02 11:20:25.038434
40193	4	1-12479	40193	71	99	1661	14	2012-08-02 11:20:25.038434
40197	1	1-7206	\N	51	51	4980	42	2012-08-02 11:20:25.038434
40206	3	1-12452	40205	28	55	736	30	2012-08-02 11:20:27.394391
40207	3	1-10242	40205	55	84	329	80	2012-08-02 11:20:27.394391
40205	3	1-12485	40205	84	28	504	46	2012-08-02 11:20:27.394391
40209	4	1-11866	40208	46	10	2411	20	2012-08-02 11:20:27.394391
40210	4	1-12192	40208	10	94	1890	31	2012-08-02 11:20:27.394391
40211	4	1-9716	40208	94	84	1860	80	2012-08-02 11:20:27.394391
40208	4	1-12485	40208	84	46	3340	46	2012-08-02 11:20:27.394391
40213	4	1-11866	40212	46	66	563	20	2012-08-02 11:20:27.394391
40214	4	1-12254	40212	66	94	551	71	2012-08-02 11:20:27.394391
40215	4	1-11628	40212	94	84	391	80	2012-08-02 11:20:27.394391
40212	4	1-12485	40212	84	46	740	46	2012-08-02 11:20:27.394391
40217	3	1-12256	40216	44	36	317	76	2012-08-02 11:20:27.394391
40218	3	1-9558	40216	36	84	143	80	2012-08-02 11:20:27.394391
40216	3	1-12485	40216	84	44	274	46	2012-08-02 11:20:27.394391
40220	3	1-10999	40219	28	47	737	83	2012-08-02 11:20:27.394391
40221	3	1-11638	40219	47	84	514	80	2012-08-02 11:20:27.394391
40219	3	1-12485	40219	84	28	1131	46	2012-08-02 11:20:27.394391
40223	4	1-9196	40222	75	83	2925	93	2012-08-02 11:20:28.558352
40224	4	1-11740	40222	83	53	2819	44	2012-08-02 11:20:28.558352
40225	4	1-11973	40222	53	28	4103	25	2012-08-02 11:20:28.558352
40222	4	1-12488	40222	28	75	4440	30	2012-08-02 11:20:28.558352
40227	4	1-11845	40226	21	43	791	99	2012-08-02 11:20:28.558352
40228	4	1-11829	40226	43	66	576	56	2012-08-02 11:20:28.558352
40229	4	1-10936	40226	66	28	672	25	2012-08-02 11:20:28.558352
40226	4	1-12488	40226	28	21	773	30	2012-08-02 11:20:28.558352
40254	4	1-12292	40253	80	68	1266	36	2012-08-02 11:20:30.879963
40255	4	1-11944	40253	68	62	1277	26	2012-08-02 11:20:30.879963
40256	4	1-10653	40253	62	12	2285	2	2012-08-02 11:20:30.879963
40253	4	1-12493	40253	12	80	1607	20	2012-08-02 11:20:30.879963
40262	4	1-11352	40261	43	77	1846	1	2012-08-02 11:20:31.675507
40263	4	1-7417	40261	77	73	746	62	2012-08-02 11:20:31.675507
40264	4	1-12439	40261	73	54	5981	87	2012-08-02 11:20:31.675507
40261	4	1-12496	40261	54	43	2587	9	2012-08-02 11:20:31.675507
40315	3	1-12468	40314	63	59	7858	33	2012-08-02 11:20:39.789928
40316	3	1-6195	40314	59	36	1849	73	2012-08-02 11:20:39.789928
40314	3	1-12517	40314	36	63	5275	97	2012-08-02 11:20:39.789928
40318	3	1-10916	40317	97	71	913	41	2012-08-02 11:20:40.961064
40319	3	1-12479	40317	71	65	3065	14	2012-08-02 11:20:40.961064
40317	3	1-12520	40317	65	97	7872	70	2012-08-02 11:20:40.961064
40337	4	1-12269	40336	63	84	351	92	2012-08-02 11:20:48.789151
40338	4	1-12180	40336	84	46	280	44	2012-08-02 11:20:48.789151
40339	4	1-12023	40336	46	4	274	6	2012-08-02 11:20:48.789151
40336	4	1-12540	40336	4	63	213	86	2012-08-02 11:20:48.789151
40341	4	1-11794	40340	8	69	2599	73	2012-08-02 11:20:51.098418
40342	4	1-12438	40340	69	2	1443	63	2012-08-02 11:20:51.098418
40343	4	1-12232	40340	2	87	1550	41	2012-08-02 11:20:51.098418
40340	4	1-12546	40340	87	8	7457	76	2012-08-02 11:20:51.098418
40345	4	1-10945	40344	66	63	327	97	2012-08-02 11:20:51.098418
40346	4	1-12468	40344	63	60	634	33	2012-08-02 11:20:51.098418
40347	4	1-2683	40344	60	87	141	41	2012-08-02 11:20:51.098418
40344	4	1-12546	40344	87	66	885	76	2012-08-02 11:20:51.098418
40349	4	1-9558	40348	36	79	341	80	2012-08-02 11:20:51.098418
40350	4	1-11730	40348	79	65	151	64	2012-08-02 11:20:51.098418
40351	4	1-9009	40348	65	87	86	41	2012-08-02 11:20:51.098418
40348	4	1-12546	40348	87	36	595	76	2012-08-02 11:20:51.098418
40124	4	1-12459	40124	8	76	5194	26	2012-08-02 11:20:17.19583
40133	2	1-11187	40132	16	15	3106	95	2012-08-02 11:20:19.970268
40132	2	1-12466	40132	15	16	9999	26	2012-08-02 11:20:19.970268
40152	4	1-11883	40151	48	87	6514	57	2012-08-02 11:20:21.119568
40153	4	1-8422	40151	87	82	2292	3	2012-08-02 11:20:21.119568
40154	4	1-12277	40151	82	89	3123	91	2012-08-02 11:20:21.119568
40151	4	1-12469	40151	89	48	5796	44	2012-08-02 11:20:21.119568
40156	4	1-11921	40155	54	49	552	55	2012-08-02 11:20:21.119568
40157	4	1-12450	40155	49	82	184	3	2012-08-02 11:20:21.119568
40158	4	1-12277	40155	82	89	271	91	2012-08-02 11:20:21.119568
40155	4	1-12469	40155	89	54	542	44	2012-08-02 11:20:21.119568
40167	4	1-11944	40166	68	79	1342	26	2012-08-02 11:20:23.49027
40168	4	1-8453	40166	79	46	1864	100	2012-08-02 11:20:23.49027
40169	4	1-12425	40166	46	37	1471	68	2012-08-02 11:20:23.49027
40166	4	1-12475	40166	37	68	1227	36	2012-08-02 11:20:23.49027
40180	4	1-7690	40179	84	79	1109	79	2012-08-02 11:20:24.650818
40181	4	1-12212	40179	79	99	963	94	2012-08-02 11:20:24.650818
40182	4	1-12404	40179	99	78	779	10	2012-08-02 11:20:24.650818
40179	4	1-12478	40179	78	84	3033	54	2012-08-02 11:20:24.650818
40184	4	1-10946	40183	89	56	217	4	2012-08-02 11:20:24.650818
40185	4	1-11608	40183	56	18	172	65	2012-08-02 11:20:24.650818
40186	4	1-5668	40183	18	78	117	10	2012-08-02 11:20:24.650818
40183	4	1-12478	40183	78	89	482	54	2012-08-02 11:20:24.650818
40188	2	1-8736	40187	35	78	1222	10	2012-08-02 11:20:24.650818
40187	2	1-12478	40187	78	35	5836	54	2012-08-02 11:20:24.650818
40203	3	1-12170	40202	22	98	870	24	2012-08-02 11:20:25.834925
40204	3	1-11839	40202	98	83	192	72	2012-08-02 11:20:25.834925
40202	3	1-12481	40202	83	22	9463	69	2012-08-02 11:20:25.834925
40231	4	1-12176	40230	28	18	4095	74	2012-08-02 11:20:28.944385
40232	4	1-10950	40230	18	23	3152	2	2012-08-02 11:20:28.944385
40233	4	1-11974	40230	23	67	2172	84	2012-08-02 11:20:28.944385
40230	4	1-12489	40230	67	28	4272	10	2012-08-02 11:20:28.944385
40235	4	1-12176	40234	28	3	354	74	2012-08-02 11:20:28.944385
40236	4	1-9754	40234	3	50	276	27	2012-08-02 11:20:28.944385
40237	4	1-11700	40234	50	67	131	84	2012-08-02 11:20:28.944385
40234	4	1-12489	40234	67	28	308	10	2012-08-02 11:20:28.944385
40239	4	1-10884	40238	83	77	14	43	2012-08-02 11:20:28.944385
40240	4	1-7752	40238	77	22	11	70	2012-08-02 11:20:28.944385
40241	4	1-11998	40238	22	67	7	84	2012-08-02 11:20:28.944385
40238	4	1-12489	40238	67	83	17	10	2012-08-02 11:20:28.944385
40243	4	1-10775	40242	68	24	1734	32	2012-08-02 11:20:28.944385
40244	4	1-12252	40242	24	88	1638	75	2012-08-02 11:20:28.944385
40245	4	1-11395	40242	88	67	772	84	2012-08-02 11:20:28.944385
40242	4	1-12489	40242	67	68	2061	10	2012-08-02 11:20:28.944385
40247	4	1-10466	40246	79	62	2078	62	2012-08-02 11:20:28.944385
40248	4	1-11939	40246	62	75	1326	54	2012-08-02 11:20:28.944385
40249	4	1-12128	40246	75	67	921	84	2012-08-02 11:20:28.944385
40246	4	1-12489	40246	67	79	2487	10	2012-08-02 11:20:28.944385
40270	4	1-11371	40269	90	95	587	89	2012-08-02 11:20:33.212148
40271	4	1-10327	40269	95	100	511	82	2012-08-02 11:20:33.212148
40272	4	1-11842	40269	100	36	513	80	2012-08-02 11:20:33.212148
40269	4	1-12500	40269	36	90	1283	43	2012-08-02 11:20:33.212148
40274	4	1-10503	40273	19	38	673	50	2012-08-02 11:20:33.212148
40275	4	1-11532	40273	38	47	643	83	2012-08-02 11:20:33.212148
40276	4	1-11638	40273	47	36	387	80	2012-08-02 11:20:33.212148
40273	4	1-12500	40273	36	19	1009	43	2012-08-02 11:20:33.212148
40278	3	1-8272	40277	24	94	924	71	2012-08-02 11:20:33.212148
40279	3	1-11628	40277	94	36	706	80	2012-08-02 11:20:33.212148
40277	3	1-12500	40277	36	24	1975	43	2012-08-02 11:20:33.212148
40281	4	1-9570	40280	66	28	2353	81	2012-08-02 11:20:33.212148
40282	4	1-11947	40280	28	46	706	67	2012-08-02 11:20:33.212148
40283	4	1-12354	40280	46	36	946	80	2012-08-02 11:20:33.212148
40280	4	1-12500	40280	36	66	2920	43	2012-08-02 11:20:33.212148
40285	3	1-11869	40284	36	36	114	76	2012-08-02 11:20:33.212148
40286	3	1-9558	40284	36	36	62	80	2012-08-02 11:20:33.212148
40284	3	1-12500	40284	36	36	195	43	2012-08-02 11:20:33.212148
40288	4	1-12424	40287	81	13	3717	21	2012-08-02 11:20:34.395872
40289	4	1-11222	40287	13	31	2019	6	2012-08-02 11:20:34.395872
40290	4	1-12168	40287	31	65	1448	71	2012-08-02 11:20:34.395872
40287	4	1-12503	40287	65	81	4504	7	2012-08-02 11:20:34.395872
40292	3	1-11603	40291	83	66	706	20	2012-08-02 11:20:34.395872
40293	3	1-12254	40291	66	65	582	71	2012-08-02 11:20:34.395872
40291	3	1-12503	40291	65	83	1966	7	2012-08-02 11:20:34.395872
40323	3	1-10744	40322	14	22	1347	62	2012-08-02 11:20:44.491589
40324	3	1-849	40322	22	75	369	17	2012-08-02 11:20:44.491589
40322	3	1-12529	40322	75	14	2927	71	2012-08-02 11:20:44.491589
40326	3	1-10022	40325	90	26	3018	21	2012-08-02 11:20:44.491589
40327	3	1-728	40325	26	75	636	17	2012-08-02 11:20:44.491589
40325	3	1-12529	40325	75	90	5616	71	2012-08-02 11:20:44.491589
40357	4	1-11696	40356	3	40	2167	50	2012-08-02 11:20:51.473831
40358	4	1-9734	40356	40	27	1564	42	2012-08-02 11:20:51.473831
40359	4	1-12396	40356	27	78	1830	90	2012-08-02 11:20:51.473831
40356	4	1-12547	40356	78	3	2163	23	2012-08-02 11:20:51.473831
40383	3	1-12205	40382	93	40	4394	12	2012-08-02 11:20:53.851525
40384	3	1-12414	40382	40	9	2442	66	2012-08-02 11:20:53.851525
40382	3	1-12553	40382	9	93	2756	19	2012-08-02 11:20:53.851525
40386	4	1-8052	40385	3	39	413	53	2012-08-02 11:20:53.851525
40387	4	1-11798	40385	39	40	498	12	2012-08-02 11:20:53.851525
40388	4	1-12414	40385	40	9	293	66	2012-08-02 11:20:53.851525
40385	4	1-12553	40385	9	3	349	19	2012-08-02 11:20:53.851525
40390	4	1-12073	40389	5	36	681	83	2012-08-02 11:20:53.851525
40391	4	1-8999	40389	36	77	513	37	2012-08-02 11:20:53.851525
40392	4	1-12274	40389	77	9	139	66	2012-08-02 11:20:53.851525
40389	4	1-12553	40389	9	5	191	19	2012-08-02 11:20:53.851525
40402	4	1-9612	40401	75	91	1142	19	2012-08-02 11:20:55.056277
40403	4	1-12328	40401	91	74	4835	70	2012-08-02 11:20:55.056277
40404	4	1-12537	40401	74	71	2350	32	2012-08-02 11:20:55.056277
40401	4	1-12556	40401	71	75	4741	96	2012-08-02 11:20:55.056277
40406	4	1-11599	40405	66	52	404	41	2012-08-02 11:20:55.056277
40407	4	1-12358	40405	52	19	932	54	2012-08-02 11:20:55.056277
40408	4	1-11669	40405	19	71	722	32	2012-08-02 11:20:55.056277
40405	4	1-12556	40405	71	66	1478	96	2012-08-02 11:20:55.056277
40410	4	1-12173	40409	59	98	1892	55	2012-08-02 11:20:55.056277
40321	2	1-5325	40320	59	48	856	34	2012-08-02 11:20:43.717401
40329	4	1-12489	40328	67	93	34	10	2012-08-02 11:20:46.071419
40330	4	1-12164	40328	93	64	19	49	2012-08-02 11:20:46.071419
40331	4	1-9373	40328	64	14	6	74	2012-08-02 11:20:46.071419
40328	4	1-12533	40328	14	67	15	84	2012-08-02 11:20:46.071419
40333	4	1-11732	40332	83	91	140	5	2012-08-02 11:20:46.071419
40334	4	1-12276	40332	91	64	164	49	2012-08-02 11:20:46.071419
40335	4	1-9373	40332	64	14	54	74	2012-08-02 11:20:46.071419
40332	4	1-12533	40332	14	83	154	84	2012-08-02 11:20:46.071419
40365	4	1-12451	40364	24	41	334	26	2012-08-02 11:20:53.075923
40366	4	1-12203	40364	41	75	767	16	2012-08-02 11:20:53.075923
40367	4	1-11916	40364	75	60	730	42	2012-08-02 11:20:53.075923
40364	4	1-12551	40364	60	24	787	57	2012-08-02 11:20:53.075923
40369	4	1-12451	40368	24	18	22	26	2012-08-02 11:20:53.075923
40370	4	1-11748	40368	18	75	44	16	2012-08-02 11:20:53.075923
40371	4	1-11916	40368	75	60	44	42	2012-08-02 11:20:53.075923
40368	4	1-12551	40368	60	24	49	57	2012-08-02 11:20:53.075923
40394	4	1-11956	40393	93	34	4081	9	2012-08-02 11:20:54.260199
40395	4	1-8987	40393	34	42	2713	59	2012-08-02 11:20:54.260199
40396	4	1-11692	40393	42	66	3995	85	2012-08-02 11:20:54.260199
40393	4	1-12554	40393	66	93	5998	44	2012-08-02 11:20:54.260199
40414	4	1-12491	40413	18	38	2495	60	2012-08-02 11:20:55.454389
40415	4	1-12036	40413	38	70	1595	87	2012-08-02 11:20:55.454389
40416	4	1-12193	40413	70	81	558	7	2012-08-02 11:20:55.454389
40413	4	1-12557	40413	81	18	1608	51	2012-08-02 11:20:55.454389
40418	3	1-12515	40417	31	74	2330	54	2012-08-02 11:20:55.454389
40419	3	1-11099	40417	74	81	1105	7	2012-08-02 11:20:55.454389
40417	3	1-12557	40417	81	31	4791	51	2012-08-02 11:20:55.454389
40421	4	1-12479	40420	71	99	885	14	2012-08-02 11:20:56.648337
40422	4	1-12071	40420	99	30	400	13	2012-08-02 11:20:56.648337
40423	4	1-11786	40420	30	54	238	75	2012-08-02 11:20:56.648337
40420	4	1-12560	40420	54	71	180	41	2012-08-02 11:20:56.648337
40424	1	1-1063	\N	59	59	2064	14	2012-08-02 11:20:56.648337
40439	4	1-9303	40438	66	47	621	11	2012-08-02 11:20:57.830405
40440	4	1-8804	40438	47	66	437	85	2012-08-02 11:20:57.830405
40441	4	1-12554	40438	66	100	538	44	2012-08-02 11:20:57.830405
40438	4	1-12563	40438	100	66	1452	28	2012-08-02 11:20:57.830405
40443	4	1-9732	40442	27	92	4778	2	2012-08-02 11:20:57.830405
40444	4	1-11778	40442	92	66	1896	85	2012-08-02 11:20:57.830405
40445	4	1-12554	40442	66	100	1885	44	2012-08-02 11:20:57.830405
40442	4	1-12563	40442	100	27	5224	28	2012-08-02 11:20:57.830405
40454	4	1-12491	40453	18	25	6832	60	2012-08-02 11:20:58.97991
40455	4	1-1375	40453	25	79	1112	13	2012-08-02 11:20:58.97991
40456	4	1-7296	40453	79	27	382	95	2012-08-02 11:20:58.97991
40453	4	1-12566	40453	27	18	3299	51	2012-08-02 11:20:58.97991
40458	3	1-10497	40457	95	32	3753	21	2012-08-02 11:20:58.97991
40459	3	1-11160	40457	32	27	539	95	2012-08-02 11:20:58.97991
40457	3	1-12566	40457	27	95	5663	51	2012-08-02 11:20:58.97991
40498	4	1-12480	40497	11	27	3134	38	2012-08-02 11:21:04.00732
40499	4	1-11464	40497	27	81	737	56	2012-08-02 11:21:04.00732
40500	4	1-12507	40497	81	92	1527	48	2012-08-02 11:21:04.00732
40497	4	1-12579	40497	92	11	1571	67	2012-08-02 11:21:04.00732
40502	3	1-12480	40501	11	63	2244	38	2012-08-02 11:21:04.00732
40503	3	1-4414	40501	63	92	624	48	2012-08-02 11:21:04.00732
40501	3	1-12579	40501	92	11	850	67	2012-08-02 11:21:04.00732
40505	4	1-11988	40504	10	98	2279	16	2012-08-02 11:21:04.00732
40506	4	1-12279	40504	98	65	796	73	2012-08-02 11:21:04.00732
40507	4	1-11344	40504	65	92	780	48	2012-08-02 11:21:04.00732
40504	4	1-12579	40504	92	10	1069	67	2012-08-02 11:21:04.00732
40509	4	1-12473	40508	62	90	114	63	2012-08-02 11:21:04.00732
40510	4	1-11997	40508	90	71	216	32	2012-08-02 11:21:04.00732
40511	4	1-12146	40508	71	92	195	48	2012-08-02 11:21:04.00732
40508	4	1-12579	40508	92	62	273	67	2012-08-02 11:21:04.00732
40548	3	1-11425	40547	45	68	3919	69	2012-08-02 11:21:07.587592
40549	3	1-11927	40547	68	59	1460	3	2012-08-02 11:21:07.587592
40547	3	1-12588	40547	59	45	4820	23	2012-08-02 11:21:07.587592
40563	4	1-11961	40562	26	28	298	81	2012-08-02 11:21:09.963143
40564	4	1-11947	40562	28	7	95	67	2012-08-02 11:21:09.963143
40565	4	1-8529	40562	7	50	79	70	2012-08-02 11:21:09.963143
40562	4	1-12594	40562	50	26	262	52	2012-08-02 11:21:09.963143
40567	4	1-12156	40566	91	4	323	76	2012-08-02 11:21:09.963143
40568	4	1-11302	40566	4	77	170	43	2012-08-02 11:21:09.963143
40569	4	1-7752	40566	77	50	143	70	2012-08-02 11:21:09.963143
40566	4	1-12594	40566	50	91	520	52	2012-08-02 11:21:09.963143
40582	4	1-6071	40581	98	59	1215	100	2012-08-02 11:21:12.327546
40583	4	1-11453	40581	59	38	1521	50	2012-08-02 11:21:12.327546
40584	4	1-11532	40581	38	95	1936	83	2012-08-02 11:21:12.327546
40581	4	1-12600	40581	95	98	3214	7	2012-08-02 11:21:12.327546
40586	4	1-10685	40585	16	52	1736	91	2012-08-02 11:21:12.327546
40587	4	1-11323	40585	52	19	1505	99	2012-08-02 11:21:12.327546
40588	4	1-12539	40585	19	95	1823	83	2012-08-02 11:21:12.327546
40585	4	1-12600	40585	95	16	3062	7	2012-08-02 11:21:12.327546
40596	4	1-12575	40595	84	58	1739	98	2012-08-02 11:21:17.003001
40597	4	1-11637	40595	58	37	703	100	2012-08-02 11:21:17.003001
40598	4	1-11633	40595	37	42	272	76	2012-08-02 11:21:17.003001
40595	4	1-12612	40595	42	84	819	71	2012-08-02 11:21:17.003001
40600	3	1-11824	40599	71	91	1645	52	2012-08-02 11:21:17.003001
40601	3	1-12156	40599	91	42	800	76	2012-08-02 11:21:17.003001
40599	3	1-12612	40599	42	71	2452	71	2012-08-02 11:21:17.003001
40603	4	1-5902	40602	10	79	1821	79	2012-08-02 11:21:17.003001
40604	4	1-12212	40602	79	89	1855	94	2012-08-02 11:21:17.003001
40605	4	1-9713	40602	89	42	1339	76	2012-08-02 11:21:17.003001
40602	4	1-12612	40602	42	10	4977	71	2012-08-02 11:21:17.003001
40607	3	1-7140	40606	56	37	649	100	2012-08-02 11:21:17.003001
40608	3	1-11633	40606	37	42	327	76	2012-08-02 11:21:17.003001
40606	3	1-12612	40606	42	56	1284	71	2012-08-02 11:21:17.003001
40609	1	1-6404	\N	29	29	250	100	2012-08-02 11:21:17.003001
40627	4	1-11636	40626	40	15	2220	90	2012-08-02 11:21:19.356773
40628	4	1-11924	40626	15	72	1408	29	2012-08-02 11:21:19.356773
40629	4	1-11205	40626	72	31	942	67	2012-08-02 11:21:19.356773
40626	4	1-12618	40626	31	40	4597	70	2012-08-02 11:21:19.356773
40631	4	1-2184	40630	98	46	810	100	2012-08-02 11:21:19.356773
40632	4	1-12425	40630	46	33	568	68	2012-08-02 11:21:19.356773
40633	4	1-12583	40630	33	31	873	67	2012-08-02 11:21:19.356773
40353	4	1-11524	40352	52	69	63	73	2012-08-02 11:20:51.098418
40354	4	1-12438	40352	69	2	51	63	2012-08-02 11:20:51.098418
40355	4	1-12232	40352	2	87	81	41	2012-08-02 11:20:51.098418
40352	4	1-12546	40352	87	52	573	76	2012-08-02 11:20:51.098418
40361	2	1-8863	40360	49	81	83	47	2012-08-02 11:20:52.292349
40360	2	1-12549	40360	81	49	301	33	2012-08-02 11:20:52.292349
40363	2	1-10511	40362	54	81	1329	47	2012-08-02 11:20:52.292349
40362	2	1-12549	40362	81	54	6553	33	2012-08-02 11:20:52.292349
40373	3	1-11181	40372	10	42	469	56	2012-08-02 11:20:53.46267
40374	3	1-10871	40372	42	88	272	44	2012-08-02 11:20:53.46267
40372	3	1-12552	40372	88	10	1091	72	2012-08-02 11:20:53.46267
40376	4	1-11823	40375	87	24	1377	52	2012-08-02 11:20:53.46267
40377	4	1-11279	40375	24	85	906	35	2012-08-02 11:20:53.46267
40378	4	1-9723	40375	85	88	754	44	2012-08-02 11:20:53.46267
40375	4	1-12552	40375	88	87	3988	72	2012-08-02 11:20:53.46267
40380	3	1-6838	40379	79	45	316	66	2012-08-02 11:20:53.46267
40381	3	1-9947	40379	45	88	292	44	2012-08-02 11:20:53.46267
40379	3	1-12552	40379	88	79	1561	72	2012-08-02 11:20:53.46267
40398	4	1-12404	40397	99	60	3546	10	2012-08-02 11:20:54.669447
40399	4	1-11762	40397	60	59	2864	96	2012-08-02 11:20:54.669447
40400	4	1-12173	40397	59	31	2230	55	2012-08-02 11:20:54.669447
40397	4	1-12555	40397	31	99	3615	94	2012-08-02 11:20:54.669447
40426	3	1-12119	40425	17	79	1674	6	2012-08-02 11:20:57.045909
40427	3	1-8431	40425	79	100	486	12	2012-08-02 11:20:57.045909
40425	3	1-12561	40425	100	17	2991	99	2012-08-02 11:20:57.045909
40429	4	1-11829	40428	43	42	1534	56	2012-08-02 11:20:57.045909
40430	4	1-10871	40428	42	95	1091	44	2012-08-02 11:20:57.045909
40431	4	1-9424	40428	95	100	335	12	2012-08-02 11:20:57.045909
40428	4	1-12561	40428	100	43	2232	99	2012-08-02 11:20:57.045909
40433	4	1-12539	40432	19	32	3618	83	2012-08-02 11:20:57.045909
40434	4	1-9822	40432	32	11	1083	50	2012-08-02 11:20:57.045909
40435	4	1-11562	40432	11	100	479	12	2012-08-02 11:20:57.045909
40432	4	1-12561	40432	100	19	3247	99	2012-08-02 11:20:57.045909
40476	4	1-12077	40475	97	20	172	5	2012-08-02 11:21:02.460876
40477	4	1-12304	40475	20	66	55	20	2012-08-02 11:21:02.460876
40478	4	1-12254	40475	66	84	60	71	2012-08-02 11:21:02.460876
40475	4	1-12575	40475	84	97	133	98	2012-08-02 11:21:02.460876
40480	3	1-4114	40479	41	31	1565	6	2012-08-02 11:21:02.460876
40481	3	1-12168	40479	31	84	1855	71	2012-08-02 11:21:02.460876
40479	3	1-12575	40479	84	41	4734	98	2012-08-02 11:21:02.460876
40483	4	1-11287	40482	13	75	726	55	2012-08-02 11:21:02.460876
40484	4	1-10373	40482	75	83	568	99	2012-08-02 11:21:02.460876
40485	4	1-10416	40482	83	84	433	71	2012-08-02 11:21:02.460876
40482	4	1-12575	40482	84	13	1181	98	2012-08-02 11:21:02.460876
40493	2	1-5120	40492	31	73	2818	85	2012-08-02 11:21:03.609721
40492	2	1-12578	40492	73	31	5076	64	2012-08-02 11:21:03.609721
40495	3	1-11349	40494	11	20	1058	80	2012-08-02 11:21:03.609721
40496	3	1-12174	40494	20	73	626	85	2012-08-02 11:21:03.609721
40494	3	1-12578	40494	73	11	1137	64	2012-08-02 11:21:03.609721
40513	4	1-12232	40512	2	27	3589	41	2012-08-02 11:21:04.780881
40514	4	1-6746	40512	27	48	1488	5	2012-08-02 11:21:04.780881
40515	4	1-11624	40512	48	94	395	68	2012-08-02 11:21:04.780881
40512	4	1-12581	40512	94	2	3207	63	2012-08-02 11:21:04.780881
40517	4	1-11997	40516	90	21	287	32	2012-08-02 11:21:04.780881
40518	4	1-9797	40516	21	67	121	6	2012-08-02 11:21:04.780881
40519	4	1-6206	40516	67	94	27	68	2012-08-02 11:21:04.780881
40516	4	1-12581	40516	94	90	224	63	2012-08-02 11:21:04.780881
40521	4	1-8488	40520	48	84	236	71	2012-08-02 11:21:04.780881
40522	4	1-12575	40520	84	30	464	98	2012-08-02 11:21:04.780881
40523	4	1-761	40520	30	94	77	68	2012-08-02 11:21:04.780881
40520	4	1-12581	40520	94	48	665	63	2012-08-02 11:21:04.780881
40525	4	1-11136	40524	72	91	22	23	2012-08-02 11:21:04.780881
40526	4	1-4379	40524	91	97	3	24	2012-08-02 11:21:04.780881
40527	4	1-12057	40524	97	94	4	68	2012-08-02 11:21:04.780881
40524	4	1-12581	40524	94	72	31	63	2012-08-02 11:21:04.780881
40533	3	1-12185	40532	30	42	1060	38	2012-08-02 11:21:05.963458
40534	3	1-4896	40532	42	29	406	90	2012-08-02 11:21:05.963458
40532	3	1-12584	40532	29	30	1280	34	2012-08-02 11:21:05.963458
40536	4	1-8957	40535	52	71	21	23	2012-08-02 11:21:05.963458
40537	4	1-11240	40535	71	97	9	63	2012-08-02 11:21:05.963458
40538	4	1-11766	40535	97	29	12	90	2012-08-02 11:21:05.963458
40535	4	1-12584	40535	29	52	47	34	2012-08-02 11:21:05.963458
40540	4	1-12295	40539	94	19	605	54	2012-08-02 11:21:05.963458
40541	4	1-11669	40539	19	61	537	32	2012-08-02 11:21:05.963458
40542	4	1-11890	40539	61	29	323	90	2012-08-02 11:21:05.963458
40539	4	1-12584	40539	29	94	1269	34	2012-08-02 11:21:05.963458
40544	4	1-9640	40543	83	47	475	14	2012-08-02 11:21:05.963458
40545	4	1-12032	40543	47	86	175	18	2012-08-02 11:21:05.963458
40546	4	1-11489	40543	86	29	164	90	2012-08-02 11:21:05.963458
40543	4	1-12584	40543	29	83	677	34	2012-08-02 11:21:05.963458
40551	4	1-12292	40550	80	55	770	36	2012-08-02 11:21:08.372282
40552	4	1-12015	40550	55	45	1343	49	2012-08-02 11:21:08.372282
40553	4	1-11752	40550	45	53	731	3	2012-08-02 11:21:08.372282
40550	4	1-12590	40550	53	80	1089	20	2012-08-02 11:21:08.372282
40575	3	1-11890	40574	61	29	1134	90	2012-08-02 11:21:10.736867
40576	3	1-12584	40574	29	24	3007	34	2012-08-02 11:21:10.736867
40574	3	1-12596	40574	24	61	2788	32	2012-08-02 11:21:10.736867
40578	4	1-8243	40577	48	44	1569	25	2012-08-02 11:21:10.736867
40579	4	1-12400	40577	44	28	1184	47	2012-08-02 11:21:10.736867
40580	4	1-12406	40577	28	24	1467	34	2012-08-02 11:21:10.736867
40577	4	1-12596	40577	24	48	2025	32	2012-08-02 11:21:10.736867
40592	4	1-11565	40591	48	39	1350	21	2012-08-02 11:21:14.274664
40593	4	1-12001	40591	39	26	1128	91	2012-08-02 11:21:14.274664
40594	4	1-12519	40591	26	27	544	96	2012-08-02 11:21:14.274664
40591	4	1-12605	40591	27	48	1075	67	2012-08-02 11:21:14.274664
40644	4	1-11827	40643	48	95	961	83	2012-08-02 11:21:19.765527
40645	4	1-12600	40643	95	13	1077	7	2012-08-02 11:21:19.765527
40646	4	1-11985	40643	13	78	694	32	2012-08-02 11:21:19.765527
40643	4	1-12619	40643	78	48	863	41	2012-08-02 11:21:19.765527
40648	4	1-4111	40647	81	32	711	78	2012-08-02 11:21:19.765527
40649	4	1-12383	40647	32	13	2384	7	2012-08-02 11:21:19.765527
40650	4	1-11985	40647	13	78	1817	32	2012-08-02 11:21:19.765527
40647	4	1-12619	40647	78	81	2672	41	2012-08-02 11:21:19.765527
40655	2	1-12622	40655	64	14	3487	33	2012-08-02 11:21:20.937522
40411	4	1-11498	40409	98	25	1327	13	2012-08-02 11:20:55.056277
40412	4	1-11514	40409	25	71	1201	32	2012-08-02 11:20:55.056277
40409	4	1-12556	40409	71	59	2645	96	2012-08-02 11:20:55.056277
40437	2	1-11923	40436	10	78	8896	87	2012-08-02 11:20:57.432968
40436	2	1-12562	40436	78	10	6532	4	2012-08-02 11:20:57.432968
40447	3	1-12231	40446	8	63	3360	13	2012-08-02 11:20:58.582017
40448	3	1-8679	40446	63	62	1519	90	2012-08-02 11:20:58.582017
40446	3	1-12565	40446	62	8	4176	94	2012-08-02 11:20:58.582017
40450	4	1-12404	40449	99	68	803	10	2012-08-02 11:20:58.582017
40451	4	1-10775	40449	68	61	586	32	2012-08-02 11:20:58.582017
40452	4	1-11890	40449	61	62	314	90	2012-08-02 11:20:58.582017
40449	4	1-12565	40449	62	99	939	94	2012-08-02 11:20:58.582017
40461	3	1-9391	40460	86	91	1483	4	2012-08-02 11:20:59.742471
40462	3	1-9606	40460	91	13	577	82	2012-08-02 11:20:59.742471
40460	3	1-12568	40460	13	86	2941	33	2012-08-02 11:20:59.742471
40464	4	1-11699	40463	13	24	5867	21	2012-08-02 11:20:59.742471
40465	4	1-11849	40463	24	55	1479	6	2012-08-02 11:20:59.742471
40466	4	1-12172	40463	55	13	554	82	2012-08-02 11:20:59.742471
40463	4	1-12568	40463	13	13	2877	33	2012-08-02 11:20:59.742471
40468	4	1-2683	40467	60	71	395	41	2012-08-02 11:20:59.742471
40469	4	1-12479	40467	71	56	1473	14	2012-08-02 11:20:59.742471
40470	4	1-12366	40467	56	13	427	82	2012-08-02 11:20:59.742471
40467	4	1-12568	40467	13	60	2256	33	2012-08-02 11:20:59.742471
40472	4	1-12162	40471	1	55	606	42	2012-08-02 11:20:59.742471
40473	4	1-12429	40471	55	53	286	11	2012-08-02 11:20:59.742471
40474	4	1-8412	40471	53	13	182	82	2012-08-02 11:20:59.742471
40471	4	1-12568	40471	13	1	1004	33	2012-08-02 11:20:59.742471
40487	3	1-12073	40486	5	19	8472	83	2012-08-02 11:21:03.222696
40488	3	1-5792	40486	19	47	2390	14	2012-08-02 11:21:03.222696
40486	3	1-12577	40486	47	5	5713	19	2012-08-02 11:21:03.222696
40490	3	1-12536	40489	4	11	1505	26	2012-08-02 11:21:03.222696
40491	3	1-5210	40489	11	47	712	14	2012-08-02 11:21:03.222696
40489	3	1-12577	40489	47	4	2227	19	2012-08-02 11:21:03.222696
40529	4	1-12480	40528	11	46	1809	38	2012-08-02 11:21:05.565369
40530	4	1-12056	40528	46	67	1823	6	2012-08-02 11:21:05.565369
40531	4	1-6206	40528	67	33	524	68	2012-08-02 11:21:05.565369
40528	4	1-12583	40528	33	11	771	67	2012-08-02 11:21:05.565369
40555	4	1-8052	40554	3	29	598	53	2012-08-02 11:21:09.156381
40556	4	1-12427	40554	29	98	1207	16	2012-08-02 11:21:09.156381
40557	4	1-12279	40554	98	24	380	73	2012-08-02 11:21:09.156381
40554	4	1-12592	40554	24	3	502	19	2012-08-02 11:21:09.156381
40559	4	1-1915	40558	60	29	1873	53	2012-08-02 11:21:09.156381
40560	4	1-12427	40558	29	98	3891	16	2012-08-02 11:21:09.156381
40561	4	1-12279	40558	98	24	1262	73	2012-08-02 11:21:09.156381
40558	4	1-12592	40558	24	60	1715	19	2012-08-02 11:21:09.156381
40571	4	1-12584	40570	29	27	3622	34	2012-08-02 11:21:10.360732
40572	4	1-8971	40570	27	19	3359	16	2012-08-02 11:21:10.360732
40573	4	1-11928	40570	19	33	2333	77	2012-08-02 11:21:10.360732
40570	4	1-12595	40570	33	29	1009	90	2012-08-02 11:21:10.360732
40590	2	1-8913	40589	24	35	1375	39	2012-08-02 11:21:13.887775
40589	2	1-12604	40589	35	24	8236	3	2012-08-02 11:21:13.887775
40611	4	1-12599	40610	39	36	3522	11	2012-08-02 11:21:17.40059
40612	4	1-8177	40610	36	93	2034	58	2012-08-02 11:21:17.40059
40613	4	1-11861	40610	93	30	2126	80	2012-08-02 11:21:17.40059
40610	4	1-12613	40610	30	39	7781	17	2012-08-02 11:21:17.40059
40615	4	1-8804	40614	47	66	825	85	2012-08-02 11:21:18.572577
40616	4	1-12554	40614	66	46	989	44	2012-08-02 11:21:18.572577
40617	4	1-12023	40614	46	32	873	6	2012-08-02 11:21:18.572577
40614	4	1-12617	40614	32	47	947	11	2012-08-02 11:21:18.572577
40619	4	1-8177	40618	36	21	207	58	2012-08-02 11:21:18.572577
40620	4	1-11435	40618	21	46	185	38	2012-08-02 11:21:18.572577
40621	4	1-12056	40618	46	32	191	6	2012-08-02 11:21:18.572577
40618	4	1-12617	40618	32	36	236	11	2012-08-02 11:21:18.572577
40623	4	1-10501	40622	66	93	825	66	2012-08-02 11:21:18.572577
40624	4	1-12463	40622	93	23	1671	50	2012-08-02 11:21:18.572577
40625	4	1-12449	40622	23	32	1087	6	2012-08-02 11:21:18.572577
40622	4	1-12617	40622	32	66	1401	11	2012-08-02 11:21:18.572577
40652	4	1-11686	40651	53	87	22	8	2012-08-02 11:21:20.141439
40653	4	1-12616	40651	87	75	49	55	2012-08-02 11:21:20.141439
40654	4	1-10373	40651	75	20	38	99	2012-08-02 11:21:20.141439
40651	4	1-12620	40651	20	53	35	20	2012-08-02 11:21:20.141439
40665	3	1-12414	40664	40	93	3264	66	2012-08-02 11:21:23.655203
40666	3	1-12463	40664	93	85	4057	50	2012-08-02 11:21:23.655203
40664	3	1-12629	40664	85	40	8617	12	2012-08-02 11:21:23.655203
40668	2	1-11964	40667	38	85	103	50	2012-08-02 11:21:23.655203
40667	2	1-12629	40667	85	38	241	12	2012-08-02 11:21:23.655203
40677	2	1-10661	40676	37	83	3084	15	2012-08-02 11:21:27.146426
40676	2	1-12638	40676	83	37	9374	35	2012-08-02 11:21:27.146426
40679	2	1-5849	40678	67	69	73	29	2012-08-02 11:21:29.466507
40678	2	1-12644	40678	69	67	9127	39	2012-08-02 11:21:29.466507
40695	3	1-12403	40694	33	83	3601	65	2012-08-02 11:21:30.659848
40696	3	1-11336	40694	83	56	2063	35	2012-08-02 11:21:30.659848
40694	3	1-12648	40694	56	33	3437	39	2012-08-02 11:21:30.659848
40698	4	1-11996	40697	53	57	1581	4	2012-08-02 11:21:30.659848
40699	4	1-10885	40697	57	90	742	74	2012-08-02 11:21:30.659848
40700	4	1-11805	40697	90	56	1410	35	2012-08-02 11:21:30.659848
40697	4	1-12648	40697	56	53	2515	39	2012-08-02 11:21:30.659848
40702	4	1-12559	40701	88	70	1493	64	2012-08-02 11:21:30.659848
40703	4	1-6465	40701	70	53	545	81	2012-08-02 11:21:30.659848
40704	4	1-12265	40701	53	56	589	35	2012-08-02 11:21:30.659848
40701	4	1-12648	40701	56	88	1118	39	2012-08-02 11:21:30.659848
40706	4	1-10207	40705	89	70	1716	87	2012-08-02 11:21:30.659848
40707	4	1-12193	40705	70	67	1377	7	2012-08-02 11:21:30.659848
40708	4	1-12234	40705	67	56	585	35	2012-08-02 11:21:30.659848
40705	4	1-12648	40705	56	89	1121	39	2012-08-02 11:21:30.659848
40710	4	1-12403	40709	33	86	1816	65	2012-08-02 11:21:30.659848
40711	4	1-4699	40709	86	53	706	81	2012-08-02 11:21:30.659848
40712	4	1-12265	40709	53	56	775	35	2012-08-02 11:21:30.659848
40709	4	1-12648	40709	56	33	1496	39	2012-08-02 11:21:30.659848
40729	4	1-10207	40728	89	97	2566	87	2012-08-02 11:21:32.229149
40728	4	1-12651	40728	96	89	1676	39	2012-08-02 11:21:32.229149
40630	4	1-12618	40630	31	98	4342	70	2012-08-02 11:21:19.356773
40635	4	1-6870	40634	27	46	96	100	2012-08-02 11:21:19.356773
40636	4	1-12425	40634	46	33	68	68	2012-08-02 11:21:19.356773
40637	4	1-12583	40634	33	31	107	67	2012-08-02 11:21:19.356773
40634	4	1-12618	40634	31	27	540	70	2012-08-02 11:21:19.356773
40639	4	1-12537	40638	74	12	14	32	2012-08-02 11:21:19.356773
40640	4	1-3542	40638	12	27	3	96	2012-08-02 11:21:19.356773
40641	4	1-12605	40638	27	31	5	67	2012-08-02 11:21:19.356773
40638	4	1-12618	40638	31	74	26	70	2012-08-02 11:21:19.356773
40642	1	1-3542	\N	12	12	720	96	2012-08-02 11:21:19.356773
40658	3	1-11505	40657	17	98	4379	27	2012-08-02 11:21:22.892901
40659	3	1-11934	40657	98	46	1313	19	2012-08-02 11:21:22.892901
40657	3	1-12627	40657	46	17	7095	57	2012-08-02 11:21:22.892901
40661	4	1-12451	40660	24	37	568	26	2012-08-02 11:21:22.892901
40662	4	1-3980	40660	37	24	149	73	2012-08-02 11:21:22.892901
40663	4	1-12592	40660	24	46	212	19	2012-08-02 11:21:22.892901
40660	4	1-12627	40660	46	24	1274	57	2012-08-02 11:21:22.892901
40670	3	1-11394	40669	32	48	1636	79	2012-08-02 11:21:24.042018
40671	3	1-11681	40669	48	93	315	66	2012-08-02 11:21:24.042018
40669	3	1-12630	40669	93	32	2872	58	2012-08-02 11:21:24.042018
40673	4	1-11985	40672	13	80	835	32	2012-08-02 11:21:25.213371
40674	4	1-11612	40672	80	16	1252	80	2012-08-02 11:21:25.213371
40675	4	1-11365	40672	16	77	550	46	2012-08-02 11:21:25.213371
40672	4	1-12633	40672	77	13	924	7	2012-08-02 11:21:25.213371
40688	3	1-11829	40687	43	66	1067	56	2012-08-02 11:21:31.046356
40689	3	1-10936	40687	66	14	659	25	2012-08-02 11:21:31.046356
40687	3	1-12647	40687	14	43	2768	99	2012-08-02 11:21:31.046356
40691	4	1-10341	40690	67	100	2353	82	2012-08-02 11:21:31.046356
40692	4	1-11842	40690	100	80	1892	80	2012-08-02 11:21:31.046356
40693	4	1-12000	40690	80	14	1114	25	2012-08-02 11:21:31.046356
40690	4	1-12647	40690	14	67	5251	99	2012-08-02 11:21:31.046356
40737	3	1-10959	40736	43	96	1711	66	2012-08-02 11:21:35.764971
40738	3	1-11925	40736	96	39	1393	76	2012-08-02 11:21:35.764971
40736	3	1-12660	40736	39	43	2931	97	2012-08-02 11:21:35.764971
40740	4	1-12388	40739	10	59	4507	49	2012-08-02 11:21:35.764971
40741	4	1-9120	40739	59	10	2640	23	2012-08-02 11:21:35.764971
40742	4	1-9868	40739	10	39	2126	76	2012-08-02 11:21:35.764971
40739	4	1-12660	40739	39	10	4584	97	2012-08-02 11:21:35.764971
40744	4	1-12046	40743	57	20	1917	37	2012-08-02 11:21:35.764971
40745	4	1-10958	40743	20	36	922	43	2012-08-02 11:21:35.764971
40746	4	1-11869	40743	36	39	621	76	2012-08-02 11:21:35.764971
40743	4	1-12660	40743	39	57	1387	97	2012-08-02 11:21:35.764971
40786	4	1-11915	40785	2	77	2925	46	2012-08-02 11:21:42.053335
40787	4	1-12633	40785	77	80	3761	7	2012-08-02 11:21:42.053335
40788	4	1-12241	40785	80	48	3326	16	2012-08-02 11:21:42.053335
40785	4	1-12676	40785	48	2	3785	8	2012-08-02 11:21:42.053335
40790	3	1-12616	40789	87	48	666	55	2012-08-02 11:21:42.053335
40791	3	1-9499	40789	48	48	498	16	2012-08-02 11:21:42.053335
40789	3	1-12676	40789	48	87	582	8	2012-08-02 11:21:42.053335
40793	4	1-12169	40792	92	97	2202	24	2012-08-02 11:21:42.053335
40794	4	1-12057	40792	97	41	2860	68	2012-08-02 11:21:42.053335
40795	4	1-11640	40792	41	48	2139	16	2012-08-02 11:21:42.053335
40792	4	1-12676	40792	48	92	2740	8	2012-08-02 11:21:42.053335
40797	4	1-11887	40796	15	57	728	58	2012-08-02 11:21:42.053335
40798	4	1-11854	40796	57	80	705	7	2012-08-02 11:21:42.053335
40799	4	1-12241	40796	80	48	775	16	2012-08-02 11:21:42.053335
40796	4	1-12676	40796	48	15	1095	8	2012-08-02 11:21:42.053335
40823	3	1-12183	40822	34	66	1014	11	2012-08-02 11:21:46.429608
40824	3	1-10501	40822	66	28	561	66	2012-08-02 11:21:46.429608
40822	3	1-12687	40822	28	34	1315	59	2012-08-02 11:21:46.429608
40826	4	1-9314	40825	82	53	1430	81	2012-08-02 11:21:46.429608
40827	4	1-12265	40825	53	56	1399	35	2012-08-02 11:21:46.429608
40828	4	1-10345	40825	56	28	1193	66	2012-08-02 11:21:46.429608
40825	4	1-12687	40825	28	82	2852	59	2012-08-02 11:21:46.429608
40879	3	1-4638	40878	80	39	476	13	2012-08-02 11:21:53.412361
40880	3	1-9336	40878	39	95	148	31	2012-08-02 11:21:53.412361
40878	3	1-12705	40878	95	80	2648	32	2012-08-02 11:21:53.412361
40882	3	1-9797	40881	21	5	456	6	2012-08-02 11:21:53.412361
40883	3	1-5941	40881	5	95	61	31	2012-08-02 11:21:53.412361
40881	3	1-12705	40881	95	21	1267	32	2012-08-02 11:21:53.412361
40885	4	1-10588	40884	60	72	383	29	2012-08-02 11:21:53.412361
40886	4	1-11205	40884	72	100	174	67	2012-08-02 11:21:53.412361
40887	4	1-8528	40884	100	95	64	31	2012-08-02 11:21:53.412361
40884	4	1-12705	40884	95	60	1381	32	2012-08-02 11:21:53.412361
40889	3	1-10588	40888	60	7	539	29	2012-08-02 11:21:53.412361
40890	3	1-12661	40888	7	95	79	31	2012-08-02 11:21:53.412361
40888	3	1-12705	40888	95	60	1821	32	2012-08-02 11:21:53.412361
40892	3	1-5612	40891	88	43	131	36	2012-08-02 11:21:53.412361
40893	3	1-12024	40891	43	95	60	31	2012-08-02 11:21:53.412361
40891	3	1-12705	40891	95	88	1410	32	2012-08-02 11:21:53.412361
40911	3	1-12713	40910	39	95	1929	74	2012-08-02 11:21:57.347689
40912	3	1-12415	40910	95	8	2372	6	2012-08-02 11:21:57.347689
40910	3	1-12716	40910	8	39	1799	97	2012-08-02 11:21:57.347689
40914	4	1-12576	40913	55	72	1665	48	2012-08-02 11:21:57.347689
40915	4	1-12701	40913	72	46	3031	44	2012-08-02 11:21:57.347689
40916	4	1-12023	40913	46	8	3042	6	2012-08-02 11:21:57.347689
40913	4	1-12716	40913	8	55	2350	97	2012-08-02 11:21:57.347689
40918	4	1-12388	40917	10	28	2015	49	2012-08-02 11:21:57.347689
40919	4	1-12078	40917	28	46	2217	38	2012-08-02 11:21:57.347689
40920	4	1-12056	40917	46	8	2495	6	2012-08-02 11:21:57.347689
40917	4	1-12716	40917	8	10	2109	97	2012-08-02 11:21:57.347689
40930	4	1-10414	40929	22	32	568	78	2012-08-02 11:22:04.739061
40931	4	1-12383	40929	32	38	2038	7	2012-08-02 11:22:04.739061
40932	4	1-12518	40929	38	6	3179	2	2012-08-02 11:22:04.739061
40929	4	1-12734	40929	6	22	3070	71	2012-08-02 11:22:04.739061
40933	1	1-3649	\N	10	10	1429	7	2012-08-02 11:22:04.739061
40939	4	1-11664	40938	76	13	3590	5	2012-08-02 11:22:05.876967
40940	4	1-12171	40938	13	33	2897	13	2012-08-02 11:22:05.876967
40941	4	1-10452	40938	33	83	1872	82	2012-08-02 11:22:05.876967
40938	4	1-12737	40938	83	76	3170	29	2012-08-02 11:22:05.876967
40943	2	1-12228	40942	92	32	1313	64	2012-08-02 11:22:07.015042
40942	2	1-12742	40942	32	92	1505	59	2012-08-02 11:22:07.015042
40945	4	1-12299	40944	50	80	1683	53	2012-08-02 11:22:07.015042
40946	4	1-12722	40944	80	79	1805	80	2012-08-02 11:22:07.015042
40656	2	1-8354	40655	14	64	968	69	2012-08-02 11:21:20.937522
40681	3	1-11229	40680	67	61	2206	22	2012-08-02 11:21:30.273011
40682	3	1-12460	40680	61	75	819	89	2012-08-02 11:21:30.273011
40680	3	1-12646	40680	75	67	2393	8	2012-08-02 11:21:30.273011
40684	4	1-11662	40683	77	94	587	17	2012-08-02 11:21:30.273011
40685	4	1-12059	40683	94	8	246	23	2012-08-02 11:21:30.273011
40686	4	1-12130	40683	8	75	238	89	2012-08-02 11:21:30.273011
40683	4	1-12646	40683	75	77	724	8	2012-08-02 11:21:30.273011
40714	4	1-12533	40713	14	83	3911	84	2012-08-02 11:21:31.433183
40715	4	1-11732	40713	83	96	3260	5	2012-08-02 11:21:31.433183
40716	4	1-12031	40713	96	30	1866	52	2012-08-02 11:21:31.433183
40713	4	1-12649	40713	30	14	1509	74	2012-08-02 11:21:31.433183
40718	4	1-11627	40717	17	46	369	67	2012-08-02 11:21:31.433183
40719	4	1-12354	40717	46	6	506	80	2012-08-02 11:21:31.433183
40720	4	1-11979	40717	6	30	327	52	2012-08-02 11:21:31.433183
40717	4	1-12649	40717	30	17	282	74	2012-08-02 11:21:31.433183
40722	4	1-11888	40721	34	89	4243	73	2012-08-02 11:21:31.433183
40723	4	1-11192	40721	89	49	5382	34	2012-08-02 11:21:31.433183
40724	4	1-11872	40721	49	30	6926	52	2012-08-02 11:21:31.433183
40721	4	1-12649	40721	30	34	6345	74	2012-08-02 11:21:31.433183
40726	3	1-10589	40725	99	71	1139	71	2012-08-02 11:21:31.433183
40727	3	1-11824	40725	71	30	947	52	2012-08-02 11:21:31.433183
40725	3	1-12649	40725	30	99	904	74	2012-08-02 11:21:31.433183
40733	4	1-12347	40732	15	47	971	86	2012-08-02 11:21:34.980127
40734	4	1-12251	40732	47	81	545	7	2012-08-02 11:21:34.980127
40735	4	1-12424	40732	81	34	797	21	2012-08-02 11:21:34.980127
40732	4	1-12658	40732	34	15	582	79	2012-08-02 11:21:34.980127
40757	4	1-11985	40756	13	71	4564	32	2012-08-02 11:21:38.117688
40758	4	1-7794	40756	71	64	1396	61	2012-08-02 11:21:38.117688
40759	4	1-12308	40756	64	28	1395	69	2012-08-02 11:21:38.117688
40756	4	1-12666	40756	28	13	6481	7	2012-08-02 11:21:38.117688
40761	4	1-12241	40760	80	24	1865	16	2012-08-02 11:21:38.117688
40762	4	1-11750	40760	24	14	506	33	2012-08-02 11:21:38.117688
40763	4	1-8354	40760	14	28	356	69	2012-08-02 11:21:38.117688
40760	4	1-12666	40760	28	80	1850	7	2012-08-02 11:21:38.117688
40765	4	1-10674	40764	31	16	798	80	2012-08-02 11:21:39.302038
40766	4	1-11365	40764	16	86	173	46	2012-08-02 11:21:39.302038
40767	4	1-11525	40764	86	62	38	94	2012-08-02 11:21:39.302038
40764	4	1-12669	40764	62	31	3211	16	2012-08-02 11:21:39.302038
40772	4	1-8136	40771	17	18	1827	79	2012-08-02 11:21:40.473127
40773	4	1-12523	40771	18	76	980	29	2012-08-02 11:21:40.473127
40774	4	1-11664	40771	76	53	638	5	2012-08-02 11:21:40.473127
40771	4	1-12672	40771	53	17	8042	81	2012-08-02 11:21:40.473127
40776	4	1-9770	40775	50	50	997	59	2012-08-02 11:21:41.655283
40777	4	1-12299	40775	50	29	826	53	2012-08-02 11:21:41.655283
40778	4	1-12427	40775	29	77	1067	16	2012-08-02 11:21:41.655283
40775	4	1-12675	40775	77	50	1146	19	2012-08-02 11:21:41.655283
40780	3	1-12536	40779	4	18	934	26	2012-08-02 11:21:41.655283
40781	3	1-11748	40779	18	77	1177	16	2012-08-02 11:21:41.655283
40779	3	1-12675	40779	77	4	1301	19	2012-08-02 11:21:41.655283
40783	3	1-12536	40782	4	96	3485	26	2012-08-02 11:21:41.655283
40784	3	1-9377	40782	96	77	3064	16	2012-08-02 11:21:41.655283
40782	3	1-12675	40782	77	4	4054	19	2012-08-02 11:21:41.655283
40808	4	1-11586	40807	12	3	1538	23	2012-08-02 11:21:42.881726
40809	4	1-11696	40807	3	11	1532	50	2012-08-02 11:21:42.881726
40810	4	1-11562	40807	11	98	683	12	2012-08-02 11:21:42.881726
40807	4	1-12679	40807	98	12	2379	6	2012-08-02 11:21:42.881726
40812	4	1-11586	40811	12	21	2116	23	2012-08-02 11:21:42.881726
40813	4	1-10541	40811	21	11	1891	50	2012-08-02 11:21:42.881726
40814	4	1-11562	40811	11	98	874	12	2012-08-02 11:21:42.881726
40811	4	1-12679	40811	98	12	3156	6	2012-08-02 11:21:42.881726
40816	3	1-7825	40815	68	95	29	44	2012-08-02 11:21:42.881726
40817	3	1-9424	40815	95	98	10	12	2012-08-02 11:21:42.881726
40815	3	1-12679	40815	98	68	37	6	2012-08-02 11:21:42.881726
40819	4	1-11586	40818	12	8	22	23	2012-08-02 11:21:42.881726
40820	4	1-12130	40818	8	37	27	89	2012-08-02 11:21:42.881726
40821	4	1-11758	40818	37	98	8	12	2012-08-02 11:21:42.881726
40818	4	1-12679	40818	98	12	31	6	2012-08-02 11:21:42.881726
40832	3	1-9717	40831	49	48	1319	46	2012-08-02 11:21:47.236116
40833	3	1-11408	40831	48	73	1899	92	2012-08-02 11:21:47.236116
40831	3	1-12689	40831	73	49	1836	73	2012-08-02 11:21:47.236116
40835	4	1-12592	40834	24	18	1279	19	2012-08-02 11:21:47.236116
40836	4	1-12398	40834	18	35	1460	79	2012-08-02 11:21:47.236116
40837	4	1-10855	40834	35	73	975	92	2012-08-02 11:21:47.236116
40834	4	1-12689	40834	73	24	1099	73	2012-08-02 11:21:47.236116
40839	4	1-12649	40838	30	17	570	74	2012-08-02 11:21:48.407505
40840	4	1-11627	40838	17	48	405	67	2012-08-02 11:21:48.407505
40841	4	1-11565	40838	48	40	242	21	2012-08-02 11:21:48.407505
40838	4	1-12692	40838	40	30	1218	52	2012-08-02 11:21:48.407505
40843	4	1-10172	40842	90	55	323	42	2012-08-02 11:21:48.407505
40844	4	1-12429	40842	55	90	160	11	2012-08-02 11:21:48.407505
40845	4	1-12593	40842	90	40	90	21	2012-08-02 11:21:48.407505
40842	4	1-12692	40842	40	90	693	52	2012-08-02 11:21:48.407505
40847	3	1-6103	40846	70	48	430	67	2012-08-02 11:21:48.407505
40848	3	1-11565	40846	48	40	425	21	2012-08-02 11:21:48.407505
40846	3	1-12692	40846	40	70	3524	52	2012-08-02 11:21:48.407505
40850	3	1-5880	40849	26	80	88	86	2012-08-02 11:21:48.407505
40851	3	1-10443	40849	80	40	29	21	2012-08-02 11:21:48.407505
40849	3	1-12692	40849	40	26	256	52	2012-08-02 11:21:48.407505
40869	4	1-12023	40868	46	12	897	6	2012-08-02 11:21:51.843443
40870	4	1-11586	40868	12	73	607	23	2012-08-02 11:21:51.843443
40871	4	1-8075	40868	73	72	426	48	2012-08-02 11:21:51.843443
40868	4	1-12701	40868	72	46	833	44	2012-08-02 11:21:51.843443
40873	3	1-12099	40872	75	62	149	6	2012-08-02 11:21:53.025745
40874	3	1-3245	40872	62	67	69	42	2012-08-02 11:21:53.025745
40872	3	1-12704	40872	67	75	133	49	2012-08-02 11:21:53.025745
40876	3	1-12378	40875	62	74	3925	48	2012-08-02 11:21:53.025745
40877	3	1-11581	40875	74	67	2617	42	2012-08-02 11:21:53.025745
40875	3	1-12704	40875	67	62	5060	49	2012-08-02 11:21:53.025745
40904	3	1-11905	40903	59	71	1094	50	2012-08-02 11:21:54.197115
40905	3	1-11415	40903	71	87	724	54	2012-08-02 11:21:54.197115
40903	3	1-12707	40903	87	59	1132	27	2012-08-02 11:21:54.197115
40907	4	1-12307	40906	4	46	333	100	2012-08-02 11:21:54.197115
40908	4	1-12425	40906	46	23	243	68	2012-08-02 11:21:54.197115
40730	4	1-11930	40728	97	16	1134	26	2012-08-02 11:21:32.229149
40731	4	1-11187	40728	16	96	1610	95	2012-08-02 11:21:32.229149
40748	5	1-10785	40747	55	86	1187	78	2012-08-02 11:21:37.75325
40749	5	1-11940	40747	86	62	1263	26	2012-08-02 11:21:37.75325
40750	5	1-12211	40747	62	77	2250	31	2012-08-02 11:21:37.75325
40751	5	1-12323	40747	77	31	1788	33	2012-08-02 11:21:37.75325
40747	5	1-12665	40747	31	55	2169	1	2012-08-02 11:21:37.75325
40753	4	1-5555	40752	98	29	906	53	2012-08-02 11:21:37.75325
40754	4	1-12427	40752	29	24	2054	16	2012-08-02 11:21:37.75325
40755	4	1-11750	40752	24	31	694	33	2012-08-02 11:21:37.75325
40752	4	1-12665	40752	31	98	906	1	2012-08-02 11:21:37.75325
40769	3	1-12463	40768	93	23	1884	50	2012-08-02 11:21:40.075695
40770	3	1-12449	40768	23	62	1260	6	2012-08-02 11:21:40.075695
40768	3	1-12671	40768	62	93	905	66	2012-08-02 11:21:40.075695
40801	3	1-12289	40800	63	98	4051	1	2012-08-02 11:21:42.45101
40802	3	1-5555	40800	98	12	2408	53	2012-08-02 11:21:42.45101
40800	3	1-12677	40800	12	63	7055	33	2012-08-02 11:21:42.45101
40804	4	1-6195	40803	59	24	561	73	2012-08-02 11:21:42.45101
40805	4	1-12592	40803	24	60	625	19	2012-08-02 11:21:42.45101
40806	4	1-1915	40803	60	12	559	53	2012-08-02 11:21:42.45101
40803	4	1-12677	40803	12	59	2070	33	2012-08-02 11:21:42.45101
40830	2	1-12415	40829	95	7	6731	6	2012-08-02 11:21:46.816578
40829	2	1-12688	40829	7	95	5669	74	2012-08-02 11:21:46.816578
40853	4	1-12336	40852	44	53	74	81	2012-08-02 11:21:49.169599
40854	4	1-12265	40852	53	12	67	35	2012-08-02 11:21:49.169599
40855	4	1-7052	40852	12	18	32	38	2012-08-02 11:21:49.169599
40852	4	1-12694	40852	18	44	38	100	2012-08-02 11:21:49.169599
40857	4	1-11785	40856	7	28	3989	47	2012-08-02 11:21:49.169599
40858	4	1-12406	40856	28	30	4373	34	2012-08-02 11:21:49.169599
40859	4	1-12185	40856	30	18	4013	38	2012-08-02 11:21:49.169599
40856	4	1-12694	40856	18	7	4956	100	2012-08-02 11:21:49.169599
40861	4	1-11588	40860	86	94	1575	31	2012-08-02 11:21:50.318842
40862	4	1-9716	40860	94	80	2299	80	2012-08-02 11:21:50.318842
40863	4	1-12000	40860	80	39	2470	25	2012-08-02 11:21:50.318842
40860	4	1-12697	40860	39	86	3003	22	2012-08-02 11:21:50.318842
40865	4	1-12585	40864	59	80	3414	38	2012-08-02 11:21:51.468072
40866	4	1-12236	40864	80	41	2945	37	2012-08-02 11:21:51.468072
40867	4	1-11978	40864	41	64	2788	45	2012-08-02 11:21:51.468072
40864	4	1-12700	40864	64	59	2478	61	2012-08-02 11:21:51.468072
40895	2	1-7052	40894	12	46	917	38	2012-08-02 11:21:53.798983
40894	2	1-12706	40894	46	12	1887	35	2012-08-02 11:21:53.798983
40897	3	1-5333	40896	47	28	513	49	2012-08-02 11:21:53.798983
40898	3	1-12078	40896	28	46	519	38	2012-08-02 11:21:53.798983
40896	3	1-12706	40896	46	47	1138	35	2012-08-02 11:21:53.798983
40900	4	1-12410	40899	21	45	2240	18	2012-08-02 11:21:53.798983
40901	4	1-12608	40899	45	65	2815	8	2012-08-02 11:21:53.798983
40902	4	1-11859	40899	65	46	1948	38	2012-08-02 11:21:53.798983
40899	4	1-12706	40899	46	21	4742	35	2012-08-02 11:21:53.798983
40935	4	1-12269	40934	63	12	4174	92	2012-08-02 11:22:05.125694
40936	4	1-12681	40934	12	63	2709	16	2012-08-02 11:22:05.125694
40937	4	1-12674	40934	63	58	2397	45	2012-08-02 11:22:05.125694
40934	4	1-12735	40934	58	63	2417	86	2012-08-02 11:22:05.125694
40961	4	1-12605	40960	27	99	6674	67	2012-08-02 11:22:11.633168
40962	4	1-10267	40960	99	84	6383	45	2012-08-02 11:22:11.633168
40963	4	1-8961	40960	84	61	2688	27	2012-08-02 11:22:11.633168
40960	4	1-12752	40960	61	27	3478	96	2012-08-02 11:22:11.633168
40965	4	1-12605	40964	27	47	1452	67	2012-08-02 11:22:11.633168
40966	4	1-12482	40964	47	98	1151	42	2012-08-02 11:22:11.633168
40967	4	1-12343	40964	98	61	571	27	2012-08-02 11:22:11.633168
40964	4	1-12752	40964	61	27	748	96	2012-08-02 11:22:11.633168
40969	4	1-3762	40968	50	81	690	56	2012-08-02 11:22:12.815234
40970	4	1-12507	40968	81	72	1673	48	2012-08-02 11:22:12.815234
40971	4	1-12701	40968	72	32	2970	44	2012-08-02 11:22:12.815234
40968	4	1-12755	40968	32	50	2486	8	2012-08-02 11:22:12.815234
40973	4	1-9123	40972	61	81	29	56	2012-08-02 11:22:12.815234
40974	4	1-12507	40972	81	72	77	48	2012-08-02 11:22:12.815234
40975	4	1-12701	40972	72	32	149	44	2012-08-02 11:22:12.815234
40972	4	1-12755	40972	32	61	136	8	2012-08-02 11:22:12.815234
40977	4	1-12550	40976	85	6	3967	2	2012-08-02 11:22:15.135514
40978	4	1-12734	40976	6	32	3432	71	2012-08-02 11:22:15.135514
40979	4	1-10956	40976	32	86	1506	72	2012-08-02 11:22:15.135514
40976	4	1-12761	40976	86	85	1572	30	2012-08-02 11:22:15.135514
40981	3	1-11845	40980	21	51	1943	99	2012-08-02 11:22:15.135514
40982	3	1-12521	40980	51	86	1987	72	2012-08-02 11:22:15.135514
40980	3	1-12761	40980	86	21	2235	30	2012-08-02 11:22:15.135514
40984	4	1-12757	40983	11	4	96	55	2012-08-02 11:22:15.135514
40985	4	1-9225	40983	4	59	88	80	2012-08-02 11:22:15.135514
40986	4	1-12759	40983	59	86	92	72	2012-08-02 11:22:15.135514
40983	4	1-12761	40983	86	11	117	30	2012-08-02 11:22:15.135514
40988	3	1-10242	40987	55	59	2039	80	2012-08-02 11:22:15.135514
40989	3	1-12759	40987	59	86	2196	72	2012-08-02 11:22:15.135514
40987	3	1-12761	40987	86	55	2878	30	2012-08-02 11:22:15.135514
41003	4	1-11179	41002	64	25	585	27	2012-08-02 11:22:18.75946
41004	4	1-11159	41002	25	39	472	13	2012-08-02 11:22:18.75946
41005	4	1-9336	41002	39	87	259	31	2012-08-02 11:22:18.75946
41002	4	1-12770	41002	87	64	1027	29	2012-08-02 11:22:18.75946
41007	3	1-12386	41006	55	69	5249	39	2012-08-02 11:22:18.75946
41008	3	1-8584	41006	69	87	2112	31	2012-08-02 11:22:18.75946
41006	3	1-12770	41006	87	55	8453	29	2012-08-02 11:22:18.75946
41010	4	1-12656	41009	35	12	466	92	2012-08-02 11:22:18.75946
41011	4	1-12681	41009	12	59	352	16	2012-08-02 11:22:18.75946
41012	4	1-10734	41009	59	87	97	31	2012-08-02 11:22:18.75946
41009	4	1-12770	41009	87	35	408	29	2012-08-02 11:22:18.75946
41026	3	1-10497	41025	95	38	1464	21	2012-08-02 11:22:19.974844
41027	3	1-12423	41025	38	65	739	5	2012-08-02 11:22:19.974844
41025	3	1-12773	41025	65	95	2362	51	2012-08-02 11:22:19.974844
41029	4	1-12631	41028	74	86	1368	46	2012-08-02 11:22:19.974844
41030	4	1-11525	41028	86	7	505	94	2012-08-02 11:22:19.974844
41031	4	1-6993	41028	7	65	285	5	2012-08-02 11:22:19.974844
41028	4	1-12773	41028	65	74	1058	51	2012-08-02 11:22:19.974844
41032	1	1-2846	\N	9	9	141	5	2012-08-02 11:22:19.974844
41038	4	1-12595	41037	33	78	2183	90	2012-08-02 11:22:22.351083
41039	4	1-12547	41037	78	21	2585	23	2012-08-02 11:22:22.351083
41037	4	1-12779	41037	65	33	4865	77	2012-08-02 11:22:22.351083
40909	4	1-12654	40906	23	87	399	54	2012-08-02 11:21:54.197115
40906	4	1-12707	40906	87	4	691	27	2012-08-02 11:21:54.197115
40922	4	1-12634	40921	35	56	425	4	2012-08-02 11:21:59.666365
40923	4	1-11608	40921	56	64	374	65	2012-08-02 11:21:59.666365
40924	4	1-12642	40921	64	80	649	53	2012-08-02 11:21:59.666365
40921	4	1-12722	40921	80	35	703	80	2012-08-02 11:21:59.666365
40926	4	1-12253	40925	11	54	2478	95	2012-08-02 11:22:01.235112
40927	4	1-12101	40925	54	58	2694	98	2012-08-02 11:22:01.235112
40928	4	1-12653	40925	58	25	7030	60	2012-08-02 11:22:01.235112
40925	4	1-12725	40925	25	11	1858	19	2012-08-02 11:22:01.235112
40991	4	1-12402	40990	12	2	2274	8	2012-08-02 11:22:16.693227
40992	4	1-11915	40990	2	7	2316	46	2012-08-02 11:22:16.693227
40993	4	1-12069	40990	7	3	989	75	2012-08-02 11:22:16.693227
40990	4	1-12765	40990	3	12	2674	23	2012-08-02 11:22:16.693227
40995	4	1-7140	40994	56	19	1761	100	2012-08-02 11:22:17.996906
40996	4	1-12739	40994	19	38	1768	24	2012-08-02 11:22:17.996906
40997	4	1-8619	40994	38	36	1075	73	2012-08-02 11:22:17.996906
40994	4	1-12768	40994	36	56	3674	71	2012-08-02 11:22:17.996906
41083	4	1-12015	41082	55	62	1408	49	2012-08-02 11:22:35.896828
41084	4	1-12378	41082	62	40	1012	48	2012-08-02 11:22:35.896828
41085	4	1-11113	41082	40	66	910	70	2012-08-02 11:22:35.896828
41082	4	1-12814	41082	66	55	844	36	2012-08-02 11:22:35.896828
41087	4	1-12024	41086	43	72	2733	31	2012-08-02 11:22:35.896828
41088	4	1-12812	41086	72	40	4119	48	2012-08-02 11:22:35.896828
41089	4	1-11113	41086	40	66	4153	70	2012-08-02 11:22:35.896828
41086	4	1-12814	41086	66	43	4318	36	2012-08-02 11:22:35.896828
41091	4	1-12015	41090	55	8	555	49	2012-08-02 11:22:35.896828
41092	4	1-12659	41090	8	97	577	83	2012-08-02 11:22:35.896828
41093	4	1-9276	41090	97	66	279	70	2012-08-02 11:22:35.896828
41090	4	1-12814	41090	66	55	294	36	2012-08-02 11:22:35.896828
41095	4	1-12728	41094	93	62	935	6	2012-08-02 11:22:35.896828
41096	4	1-12671	41094	62	92	672	66	2012-08-02 11:22:35.896828
41097	4	1-11969	41094	92	66	1197	70	2012-08-02 11:22:35.896828
41094	4	1-12814	41094	66	93	1310	36	2012-08-02 11:22:35.896828
41119	4	1-12603	41118	21	56	1064	99	2012-08-02 11:22:39.388621
41120	4	1-12801	41118	56	5	2251	48	2012-08-02 11:22:39.388621
41121	4	1-4609	41118	5	35	372	41	2012-08-02 11:22:39.388621
41118	4	1-12823	41118	35	21	1458	11	2012-08-02 11:22:39.388621
41123	4	1-8412	41122	53	72	1439	82	2012-08-02 11:22:39.388621
41124	4	1-6161	41122	72	14	586	70	2012-08-02 11:22:39.388621
41125	4	1-12807	41122	14	35	470	41	2012-08-02 11:22:39.388621
41122	4	1-12823	41122	35	53	1898	11	2012-08-02 11:22:39.388621
41127	4	1-12506	41126	12	7	5104	46	2012-08-02 11:22:39.388621
41128	4	1-12069	41126	7	54	1947	75	2012-08-02 11:22:39.388621
41129	4	1-12560	41126	54	35	1384	41	2012-08-02 11:22:39.388621
41126	4	1-12823	41126	35	12	5611	11	2012-08-02 11:22:39.388621
41131	4	1-11449	41130	74	76	169	20	2012-08-02 11:22:39.388621
41132	4	1-8202	41130	76	14	41	70	2012-08-02 11:22:39.388621
41133	4	1-12807	41130	14	35	39	41	2012-08-02 11:22:39.388621
41130	4	1-12823	41130	35	74	183	11	2012-08-02 11:22:39.388621
41150	3	1-11842	41149	100	20	1660	80	2012-08-02 11:22:40.604474
41151	3	1-12174	41149	20	68	1088	85	2012-08-02 11:22:40.604474
41149	3	1-12826	41149	68	100	1186	82	2012-08-02 11:22:40.604474
41153	3	1-12059	41152	94	8	298	23	2012-08-02 11:22:41.775281
41154	3	1-12130	41152	8	83	228	89	2012-08-02 11:22:41.775281
41152	3	1-12829	41152	83	94	897	17	2012-08-02 11:22:41.775281
41156	4	1-12599	41155	39	26	1594	11	2012-08-02 11:22:41.775281
41157	4	1-9702	41155	26	61	967	22	2012-08-02 11:22:41.775281
41158	4	1-12460	41155	61	83	422	89	2012-08-02 11:22:41.775281
41155	4	1-12829	41155	83	39	2356	17	2012-08-02 11:22:41.775281
41160	3	1-12059	41159	94	24	1371	23	2012-08-02 11:22:41.775281
41161	3	1-4558	41159	24	83	398	89	2012-08-02 11:22:41.775281
41159	3	1-12829	41159	83	94	2544	17	2012-08-02 11:22:41.775281
41178	2	1-8272	41177	24	74	1107	71	2012-08-02 11:22:43.743182
41177	2	1-12834	41177	74	24	4019	43	2012-08-02 11:22:43.743182
41180	4	1-7009	41179	63	68	1017	10	2012-08-02 11:22:43.743182
41181	4	1-10775	41179	68	18	479	32	2012-08-02 11:22:43.743182
41182	4	1-12831	41179	18	74	1226	71	2012-08-02 11:22:43.743182
41179	4	1-12834	41179	74	63	5385	43	2012-08-02 11:22:43.743182
41184	3	1-10339	41183	62	78	526	87	2012-08-02 11:22:44.903286
41185	3	1-12562	41183	78	15	603	4	2012-08-02 11:22:44.903286
41183	3	1-12837	41183	15	62	919	77	2012-08-02 11:22:44.903286
41191	4	1-3888	41190	41	74	600	51	2012-08-02 11:22:46.074313
41192	4	1-12631	41190	74	48	792	46	2012-08-02 11:22:46.074313
41193	4	1-11408	41190	48	94	1319	92	2012-08-02 11:22:46.074313
41190	4	1-12840	41190	94	41	1393	24	2012-08-02 11:22:46.074313
41195	4	1-12754	41194	27	83	1365	82	2012-08-02 11:22:46.074313
41196	4	1-12737	41194	83	35	1889	29	2012-08-02 11:22:46.074313
41197	4	1-12656	41194	35	94	1829	92	2012-08-02 11:22:46.074313
41194	4	1-12840	41194	94	27	1972	24	2012-08-02 11:22:46.074313
41199	3	1-12839	41198	64	20	1609	26	2012-08-02 11:22:46.074313
41200	3	1-11011	41198	20	94	1220	92	2012-08-02 11:22:46.074313
41198	3	1-12840	41198	94	64	1410	24	2012-08-02 11:22:46.074313
41202	3	1-12815	41201	65	93	3281	24	2012-08-02 11:22:47.278678
41203	3	1-12182	41201	93	41	1410	18	2012-08-02 11:22:47.278678
41201	3	1-12843	41201	41	65	5088	54	2012-08-02 11:22:47.278678
41205	4	1-12815	41204	65	64	107	24	2012-08-02 11:22:47.278678
41206	4	1-12839	41204	64	100	131	26	2012-08-02 11:22:47.278678
41207	4	1-8057	41204	100	41	43	18	2012-08-02 11:22:47.278678
41204	4	1-12843	41204	41	65	161	54	2012-08-02 11:22:47.278678
41209	4	1-12820	41208	69	51	442	47	2012-08-02 11:22:47.278678
41210	4	1-10355	41208	51	77	251	6	2012-08-02 11:22:47.278678
41211	4	1-8936	41208	77	41	106	18	2012-08-02 11:22:47.278678
41208	4	1-12843	41208	41	69	403	54	2012-08-02 11:22:47.278678
41225	4	1-12832	41224	38	23	357	68	2012-08-02 11:22:48.450151
41226	4	1-12654	41224	23	69	544	54	2012-08-02 11:22:48.450151
41227	4	1-12820	41224	69	43	560	47	2012-08-02 11:22:48.450151
41224	4	1-12846	41224	43	38	403	16	2012-08-02 11:22:48.450151
41233	4	1-12831	41232	18	10	4821	71	2012-08-02 11:22:53.156836
41234	4	1-5902	41232	10	31	1631	79	2012-08-02 11:22:53.156836
41235	4	1-10218	41232	31	92	603	75	2012-08-02 11:22:53.156836
41232	4	1-12858	41232	92	18	1186	32	2012-08-02 11:22:53.156836
41237	4	1-12831	41236	18	22	1410	71	2012-08-02 11:22:53.156836
41238	4	1-10414	41236	22	94	292	78	2012-08-02 11:22:53.156836
40947	4	1-11730	40944	79	32	813	64	2012-08-02 11:22:07.015042
40944	4	1-12742	40944	32	50	1634	59	2012-08-02 11:22:07.015042
40949	4	1-11940	40948	86	16	2721	26	2012-08-02 11:22:11.235432
40950	4	1-11187	40948	16	96	3123	95	2012-08-02 11:22:11.235432
40951	4	1-12651	40948	96	63	2627	39	2012-08-02 11:22:11.235432
40948	4	1-12751	40948	63	86	2967	78	2012-08-02 11:22:11.235432
40953	4	1-11436	40952	56	91	517	63	2012-08-02 11:22:11.235432
40954	4	1-12501	40952	91	96	542	95	2012-08-02 11:22:11.235432
40955	4	1-12651	40952	96	63	509	39	2012-08-02 11:22:11.235432
40952	4	1-12751	40952	63	56	641	78	2012-08-02 11:22:11.235432
40957	4	1-11940	40956	86	72	2401	26	2012-08-02 11:22:11.235432
40958	4	1-11962	40956	72	96	1724	95	2012-08-02 11:22:11.235432
40959	4	1-12651	40956	96	63	1696	39	2012-08-02 11:22:11.235432
40956	4	1-12751	40956	63	86	2239	78	2012-08-02 11:22:11.235432
40999	4	1-6206	40998	67	83	323	68	2012-08-02 11:22:18.372422
41000	4	1-12200	40998	83	48	523	51	2012-08-02 11:22:18.372422
41001	4	1-11450	40998	48	87	541	34	2012-08-02 11:22:18.372422
40998	4	1-12769	40998	87	67	1018	6	2012-08-02 11:22:18.372422
41014	4	1-12336	41013	44	32	4530	81	2012-08-02 11:22:19.576837
41015	4	1-11550	41013	32	81	910	56	2012-08-02 11:22:19.576837
41016	4	1-12507	41013	81	46	1974	48	2012-08-02 11:22:19.576837
41013	4	1-12772	41013	46	44	2501	100	2012-08-02 11:22:19.576837
41018	4	1-12425	41017	46	83	1232	68	2012-08-02 11:22:19.576837
41019	4	1-12200	41017	83	27	1714	51	2012-08-02 11:22:19.576837
41020	4	1-9762	41017	27	46	1153	48	2012-08-02 11:22:19.576837
41017	4	1-12772	41017	46	46	1735	100	2012-08-02 11:22:19.576837
41022	4	1-12034	41021	31	55	869	36	2012-08-02 11:22:19.576837
41023	4	1-12015	41021	55	62	1721	49	2012-08-02 11:22:19.576837
41024	4	1-12378	41021	62	46	1469	48	2012-08-02 11:22:19.576837
41021	4	1-12772	41021	46	31	2306	100	2012-08-02 11:22:19.576837
41034	4	1-6553	41033	98	19	1050	88	2012-08-02 11:22:20.770288
41035	4	1-11948	41033	19	44	956	53	2012-08-02 11:22:20.770288
41036	4	1-12720	41033	44	78	1913	7	2012-08-02 11:22:20.770288
41033	4	1-12775	41033	78	98	1116	78	2012-08-02 11:22:20.770288
41056	2	1-12595	41055	33	54	23	90	2012-08-02 11:22:28.150301
41055	2	1-12794	41055	54	33	63	77	2012-08-02 11:22:28.150301
41058	4	1-10908	41057	50	24	75	34	2012-08-02 11:22:28.150301
41059	4	1-12596	41057	24	61	105	32	2012-08-02 11:22:28.150301
41060	4	1-11890	41057	61	54	65	90	2012-08-02 11:22:28.150301
41057	4	1-12794	41057	54	50	237	77	2012-08-02 11:22:28.150301
41062	4	1-8377	41061	90	78	636	7	2012-08-02 11:22:29.332704
41063	4	1-12775	41061	78	91	370	78	2012-08-02 11:22:29.332704
41064	4	1-12673	41061	91	76	691	5	2012-08-02 11:22:29.332704
41061	4	1-12797	41061	76	90	1286	83	2012-08-02 11:22:29.332704
41071	4	1-12778	41070	6	10	211	20	2012-08-02 11:22:30.503719
41072	4	1-12075	41070	10	75	155	55	2012-08-02 11:22:30.503719
41073	4	1-10373	41070	75	56	99	99	2012-08-02 11:22:30.503719
41070	4	1-12801	41070	56	6	213	48	2012-08-02 11:22:30.503719
41075	4	1-12436	41074	2	39	414	17	2012-08-02 11:22:30.503719
41076	4	1-12599	41074	39	21	293	11	2012-08-02 11:22:30.503719
41077	4	1-12603	41074	21	56	245	99	2012-08-02 11:22:30.503719
41074	4	1-12801	41074	56	2	594	48	2012-08-02 11:22:30.503719
41079	4	1-11581	41078	74	55	298	42	2012-08-02 11:22:30.503719
41080	4	1-12429	41078	55	21	191	11	2012-08-02 11:22:30.503719
41081	4	1-12603	41078	21	56	165	99	2012-08-02 11:22:30.503719
41078	4	1-12801	41078	56	74	412	48	2012-08-02 11:22:30.503719
41099	4	1-12281	41098	92	14	619	85	2012-08-02 11:22:36.670154
41100	4	1-8197	41098	14	12	215	88	2012-08-02 11:22:36.670154
41101	4	1-12440	41098	12	19	355	93	2012-08-02 11:22:36.670154
41098	4	1-12816	41098	19	92	1306	54	2012-08-02 11:22:36.670154
41103	4	1-2738	41102	1	93	781	4	2012-08-02 11:22:36.670154
41104	4	1-12486	41102	93	12	245	88	2012-08-02 11:22:36.670154
41105	4	1-12440	41102	12	19	635	93	2012-08-02 11:22:36.670154
41102	4	1-12816	41102	19	1	3675	54	2012-08-02 11:22:36.670154
41107	4	1-11099	41106	74	13	365	7	2012-08-02 11:22:36.670154
41108	4	1-11985	41106	13	93	330	32	2012-08-02 11:22:36.670154
41109	4	1-6587	41106	93	19	91	93	2012-08-02 11:22:36.670154
41106	4	1-12816	41106	19	74	543	54	2012-08-02 11:22:36.670154
41111	4	1-7929	41110	45	98	255	78	2012-08-02 11:22:36.670154
41112	4	1-6553	41110	98	12	226	88	2012-08-02 11:22:36.670154
41113	4	1-12440	41110	12	19	611	93	2012-08-02 11:22:36.670154
41110	4	1-12816	41110	19	45	3677	54	2012-08-02 11:22:36.670154
41143	4	1-9699	41142	56	44	186	53	2012-08-02 11:22:40.161601
41144	4	1-12720	41142	44	80	357	7	2012-08-02 11:22:40.161601
41145	4	1-12241	41142	80	50	423	16	2012-08-02 11:22:40.161601
41142	4	1-12825	41142	50	56	373	31	2012-08-02 11:22:40.161601
41147	2	1-11297	41146	52	50	4772	16	2012-08-02 11:22:40.161601
41146	2	1-12825	41146	50	52	4437	31	2012-08-02 11:22:40.161601
41148	1	1-533	\N	18	18	410	16	2012-08-02 11:22:40.161601
41163	4	1-12654	41162	23	74	856	54	2012-08-02 11:22:42.548342
41164	4	1-11099	41162	74	80	555	7	2012-08-02 11:22:42.548342
41165	4	1-12241	41162	80	38	620	16	2012-08-02 11:22:42.548342
41162	4	1-12832	41162	38	23	556	68	2012-08-02 11:22:42.548342
41167	4	1-12475	41166	37	81	3249	36	2012-08-02 11:22:42.548342
41168	4	1-12632	41166	81	12	6184	92	2012-08-02 11:22:42.548342
41169	4	1-12681	41166	12	38	4603	16	2012-08-02 11:22:42.548342
41166	4	1-12832	41166	38	37	4303	68	2012-08-02 11:22:42.548342
41171	4	1-12756	41170	21	62	4528	14	2012-08-02 11:22:42.548342
41172	4	1-12625	41170	62	14	2684	84	2012-08-02 11:22:42.548342
41173	4	1-10323	41170	14	38	1777	16	2012-08-02 11:22:42.548342
41170	4	1-12832	41170	38	21	1719	68	2012-08-02 11:22:42.548342
41175	3	1-12475	41174	37	88	1037	36	2012-08-02 11:22:42.548342
41176	3	1-10543	41174	88	38	1338	16	2012-08-02 11:22:42.548342
41174	3	1-12832	41174	38	37	1311	68	2012-08-02 11:22:42.548342
41213	4	1-12545	41212	21	15	855	79	2012-08-02 11:22:47.665344
41214	4	1-12347	41212	15	37	1260	86	2012-08-02 11:22:47.665344
41215	4	1-9996	41212	37	82	607	44	2012-08-02 11:22:47.665344
41212	4	1-12844	41212	82	21	635	47	2012-08-02 11:22:47.665344
41217	4	1-12545	41216	21	59	4965	79	2012-08-02 11:22:47.665344
41218	4	1-6017	41216	59	59	1235	56	2012-08-02 11:22:47.665344
41219	4	1-12702	41216	59	82	3326	44	2012-08-02 11:22:47.665344
41216	4	1-12844	41216	82	21	3582	47	2012-08-02 11:22:47.665344
41221	4	1-8895	41220	70	92	139	12	2012-08-02 11:22:47.665344
41040	4	1-10541	41037	21	65	2245	50	2012-08-02 11:22:22.351083
41042	2	1-9439	41041	52	41	5365	70	2012-08-02 11:22:24.670211
41041	2	1-12785	41041	41	52	6353	18	2012-08-02 11:22:24.670211
41044	4	1-12608	41043	45	77	1715	8	2012-08-02 11:22:24.670211
41045	4	1-12740	41043	77	74	1956	91	2012-08-02 11:22:24.670211
41046	4	1-8080	41043	74	41	1150	70	2012-08-02 11:22:24.670211
41043	4	1-12785	41043	41	45	1620	18	2012-08-02 11:22:24.670211
41048	2	1-3727	41047	45	74	1022	5	2012-08-02 11:22:26.990612
41047	2	1-12790	41047	74	45	2763	72	2012-08-02 11:22:26.990612
41050	3	1-11969	41049	92	74	3415	70	2012-08-02 11:22:27.774733
41051	3	1-12537	41049	74	84	1781	32	2012-08-02 11:22:27.774733
41049	3	1-12793	41049	84	92	2052	66	2012-08-02 11:22:27.774733
41053	3	1-12417	41052	46	25	1149	13	2012-08-02 11:22:27.774733
41054	3	1-11514	41052	25	84	1065	32	2012-08-02 11:22:27.774733
41052	3	1-12793	41052	84	46	1279	66	2012-08-02 11:22:27.774733
41066	2	1-12515	41065	31	21	876	54	2012-08-02 11:22:30.105975
41065	2	1-12799	41065	21	31	1800	51	2012-08-02 11:22:30.105975
41068	3	1-12631	41067	74	73	1353	46	2012-08-02 11:22:30.105975
41069	3	1-6388	41067	73	21	278	54	2012-08-02 11:22:30.105975
41067	3	1-12799	41067	21	74	837	51	2012-08-02 11:22:30.105975
41115	4	1-9734	41114	40	60	786	42	2012-08-02 11:22:38.626019
41116	4	1-12551	41114	60	86	774	57	2012-08-02 11:22:38.626019
41117	4	1-9119	41114	86	67	247	12	2012-08-02 11:22:38.626019
41114	4	1-12821	41114	67	40	1198	50	2012-08-02 11:22:38.626019
41135	4	1-9645	41134	73	86	839	57	2012-08-02 11:22:39.775033
41136	4	1-9119	41134	86	67	238	12	2012-08-02 11:22:39.775033
41137	4	1-12821	41134	67	24	1030	50	2012-08-02 11:22:39.775033
41134	4	1-12824	41134	24	73	3810	77	2012-08-02 11:22:39.775033
41139	4	1-5872	41138	93	94	1404	17	2012-08-02 11:22:39.775033
41140	4	1-12059	41138	94	21	771	23	2012-08-02 11:22:39.775033
41141	4	1-10541	41138	21	24	745	50	2012-08-02 11:22:39.775033
41138	4	1-12824	41138	24	93	3794	77	2012-08-02 11:22:39.775033
41187	4	1-2178	41186	2	74	308	31	2012-08-02 11:22:45.676422
41188	4	1-12726	41186	74	92	681	8	2012-08-02 11:22:45.676422
41189	4	1-12169	41186	92	64	638	24	2012-08-02 11:22:45.676422
41186	4	1-12839	41186	64	2	789	26	2012-08-02 11:22:45.676422
41229	4	1-12241	41228	80	38	391	16	2012-08-02 11:22:49.245482
41230	4	1-12832	41228	38	33	319	68	2012-08-02 11:22:49.245482
41231	4	1-12583	41228	33	28	435	67	2012-08-02 11:22:49.245482
41228	4	1-12848	41228	28	80	386	7	2012-08-02 11:22:49.245482
41302	4	1-12825	41301	50	74	1479	31	2012-08-02 11:23:02.582878
41303	4	1-12726	41301	74	65	2285	8	2012-08-02 11:23:02.582878
41304	4	1-11859	41301	65	94	1027	38	2012-08-02 11:23:02.582878
41301	4	1-12882	41301	94	50	2454	16	2012-08-02 11:23:02.582878
41306	4	1-12674	41305	63	64	354	45	2012-08-02 11:23:02.582878
41307	4	1-12700	41305	64	59	215	61	2012-08-02 11:23:02.582878
41308	4	1-12585	41305	59	94	203	38	2012-08-02 11:23:02.582878
41305	4	1-12882	41305	94	63	523	16	2012-08-02 11:23:02.582878
41310	4	1-11750	41309	24	70	407	33	2012-08-02 11:23:02.582878
41311	4	1-12850	41309	70	28	601	49	2012-08-02 11:23:02.582878
41312	4	1-12078	41309	28	94	532	38	2012-08-02 11:23:02.582878
41309	4	1-12882	41309	94	24	1540	16	2012-08-02 11:23:02.582878
41318	4	1-12686	41317	42	85	329	35	2012-08-02 11:23:03.742743
41319	4	1-9723	41317	85	82	268	44	2012-08-02 11:23:03.742743
41320	4	1-12844	41317	82	71	296	47	2012-08-02 11:23:03.742743
41317	4	1-12885	41317	71	42	190	33	2012-08-02 11:23:03.742743
41335	4	1-12817	41334	84	50	1860	16	2012-08-02 11:23:09.156232
41336	4	1-12825	41334	50	99	1355	31	2012-08-02 11:23:09.156232
41337	4	1-12873	41334	99	75	2664	12	2012-08-02 11:23:09.156232
41334	4	1-12899	41334	75	84	5469	88	2012-08-02 11:23:09.156232
41339	3	1-7221	41338	67	99	113	31	2012-08-02 11:23:09.156232
41340	3	1-12873	41338	99	75	231	12	2012-08-02 11:23:09.156232
41338	3	1-12899	41338	75	67	495	88	2012-08-02 11:23:09.156232
41359	3	1-12769	41358	87	36	3905	6	2012-08-02 11:23:12.636473
41360	3	1-6678	41358	36	56	834	75	2012-08-02 11:23:12.636473
41358	3	1-12908	41358	56	87	2517	34	2012-08-02 11:23:12.636473
41362	4	1-9593	41361	27	12	1835	53	2012-08-02 11:23:12.636473
41363	4	1-12787	41361	12	7	1899	46	2012-08-02 11:23:12.636473
41364	4	1-12069	41361	7	56	825	75	2012-08-02 11:23:12.636473
41361	4	1-12908	41361	56	27	2963	34	2012-08-02 11:23:12.636473
41373	3	1-12128	41372	75	87	711	84	2012-08-02 11:23:13.80786
41374	3	1-10510	41372	87	32	917	85	2012-08-02 11:23:13.80786
41372	3	1-12911	41372	32	75	1030	54	2012-08-02 11:23:13.80786
41381	4	1-12876	41380	66	9	2332	41	2012-08-02 11:23:16.569875
41382	4	1-7403	41380	9	63	2136	16	2012-08-02 11:23:16.569875
41383	4	1-12674	41380	63	32	2080	45	2012-08-02 11:23:16.569875
41380	4	1-12918	41380	32	66	2236	46	2012-08-02 11:23:16.569875
41399	3	1-11911	41398	58	40	321	90	2012-08-02 11:23:17.763025
41400	3	1-12892	41398	40	78	657	22	2012-08-02 11:23:17.763025
41398	3	1-12921	41398	78	58	4062	40	2012-08-02 11:23:17.763025
41402	4	1-12582	41401	69	32	890	45	2012-08-02 11:23:17.763025
41403	4	1-12918	41401	32	34	499	46	2012-08-02 11:23:17.763025
41404	4	1-11326	41401	34	78	239	22	2012-08-02 11:23:17.763025
41401	4	1-12921	41401	78	69	1585	40	2012-08-02 11:23:17.763025
41416	4	1-12686	41415	42	100	5140	35	2012-08-02 11:23:19.74081
41417	4	1-12855	41415	100	49	5701	76	2012-08-02 11:23:19.74081
41418	4	1-9364	41415	49	55	2723	88	2012-08-02 11:23:19.74081
41415	4	1-12926	41415	55	42	3104	33	2012-08-02 11:23:19.74081
41420	4	1-12850	41419	70	64	607	49	2012-08-02 11:23:19.74081
41421	4	1-9373	41419	64	71	223	74	2012-08-02 11:23:19.74081
41422	4	1-9095	41419	71	55	252	88	2012-08-02 11:23:19.74081
41419	4	1-12926	41419	55	70	321	33	2012-08-02 11:23:19.74081
41424	4	1-12850	41423	70	60	3369	49	2012-08-02 11:23:19.74081
41425	4	1-10997	41423	60	98	1442	78	2012-08-02 11:23:19.74081
41426	4	1-6553	41423	98	55	1386	88	2012-08-02 11:23:19.74081
41423	4	1-12926	41423	55	70	1774	33	2012-08-02 11:23:19.74081
41428	4	1-12850	41427	70	57	4897	49	2012-08-02 11:23:19.74081
41429	4	1-4883	41427	57	80	978	63	2012-08-02 11:23:19.74081
41430	4	1-10929	41427	80	55	1988	88	2012-08-02 11:23:19.74081
41427	4	1-12926	41427	55	70	2562	33	2012-08-02 11:23:19.74081
41434	2	1-11180	41433	25	76	3815	42	2012-08-02 11:23:20.912193
41433	2	1-12929	41433	76	25	5670	75	2012-08-02 11:23:20.912193
41436	4	1-12858	41435	92	71	2247	32	2012-08-02 11:23:20.912193
41437	4	1-7794	41435	71	48	744	61	2012-08-02 11:23:20.912193
41222	4	1-12572	41220	92	83	237	93	2012-08-02 11:22:47.665344
41223	4	1-11740	41220	83	82	250	44	2012-08-02 11:22:47.665344
41220	4	1-12844	41220	82	70	287	47	2012-08-02 11:22:47.665344
41241	2	1-11972	41240	88	81	1744	1	2012-08-02 11:22:53.543014
41240	2	1-12859	41240	81	88	9618	59	2012-08-02 11:22:53.543014
41254	3	1-12765	41253	3	28	2076	23	2012-08-02 11:22:57.509065
41255	3	1-10537	41253	28	70	813	45	2012-08-02 11:22:57.509065
41253	3	1-12870	41253	70	3	1393	75	2012-08-02 11:22:57.509065
41257	2	1-8702	41256	49	70	2426	45	2012-08-02 11:22:57.509065
41256	2	1-12870	41256	70	49	4572	75	2012-08-02 11:22:57.509065
41271	4	1-12572	41270	92	76	6843	93	2012-08-02 11:22:59.10205
41272	4	1-11892	41270	76	48	5329	34	2012-08-02 11:22:59.10205
41273	4	1-12802	41270	48	99	2143	31	2012-08-02 11:22:59.10205
41270	4	1-12873	41270	99	92	4599	12	2012-08-02 11:22:59.10205
41275	4	1-12821	41274	67	50	4806	50	2012-08-02 11:22:59.10205
41276	4	1-8191	41274	50	47	841	68	2012-08-02 11:22:59.10205
41277	4	1-7215	41274	47	99	434	31	2012-08-02 11:22:59.10205
41274	4	1-12873	41274	99	67	951	12	2012-08-02 11:22:59.10205
41295	3	1-11388	41294	11	44	701	53	2012-08-02 11:23:01.420239
41296	3	1-12720	41294	44	83	1049	7	2012-08-02 11:23:01.420239
41294	3	1-12880	41294	83	11	2298	18	2012-08-02 11:23:01.420239
41298	4	1-12608	41297	45	15	2680	8	2012-08-02 11:23:01.420239
41299	4	1-11887	41297	15	57	1526	58	2012-08-02 11:23:01.420239
41300	4	1-11854	41297	57	83	1267	7	2012-08-02 11:23:01.420239
41297	4	1-12880	41297	83	45	2827	18	2012-08-02 11:23:01.420239
41355	4	1-12778	41354	6	7	4055	20	2012-08-02 11:23:12.249847
41356	4	1-9029	41354	7	87	2216	34	2012-08-02 11:23:12.249847
41357	4	1-12769	41354	87	62	4096	6	2012-08-02 11:23:12.249847
41354	4	1-12907	41354	62	6	3528	48	2012-08-02 11:23:12.249847
41369	4	1-12332	41368	49	39	3285	11	2012-08-02 11:23:13.410418
41370	4	1-12830	41368	39	20	3997	30	2012-08-02 11:23:13.410418
41371	4	1-12781	41368	20	61	3799	53	2012-08-02 11:23:13.410418
41368	4	1-12910	41368	61	49	3843	13	2012-08-02 11:23:13.410418
41376	5	1-10529	41375	84	39	97	97	2012-08-02 11:23:14.61459
41377	5	1-12713	41375	39	34	121	74	2012-08-02 11:23:14.61459
41378	5	1-11888	41375	34	69	88	73	2012-08-02 11:23:14.61459
41379	5	1-12438	41375	69	61	76	63	2012-08-02 11:23:14.61459
41375	5	1-12913	41375	61	84	89	56	2012-08-02 11:23:14.61459
41412	4	1-11619	41411	42	74	1529	51	2012-08-02 11:23:19.343074
41413	4	1-12631	41411	74	66	1571	46	2012-08-02 11:23:19.343074
41414	4	1-12876	41411	66	17	992	41	2012-08-02 11:23:19.343074
41411	4	1-12925	41411	17	42	6149	79	2012-08-02 11:23:19.343074
41432	2	1-12803	41431	27	35	3033	4	2012-08-02 11:23:20.514438
41431	2	1-12928	41431	35	27	4663	80	2012-08-02 11:23:20.514438
41446	4	1-8232	41445	62	18	541	32	2012-08-02 11:23:24.171569
41447	4	1-12831	41445	18	71	1185	71	2012-08-02 11:23:24.171569
41448	4	1-11824	41445	71	61	480	52	2012-08-02 11:23:24.171569
41445	4	1-12937	41445	61	62	2745	4	2012-08-02 11:23:24.171569
41450	4	1-10542	41449	53	90	1487	43	2012-08-02 11:23:24.171569
41451	4	1-11371	41449	90	49	511	89	2012-08-02 11:23:24.171569
41452	4	1-12614	41449	49	61	836	52	2012-08-02 11:23:24.171569
41449	4	1-12937	41449	61	53	5260	4	2012-08-02 11:23:24.171569
41487	4	1-10929	41486	80	55	1259	88	2012-08-02 11:23:28.902649
41488	4	1-12926	41486	55	32	1349	33	2012-08-02 11:23:28.902649
41489	4	1-8311	41486	32	77	624	91	2012-08-02 11:23:28.902649
41486	4	1-12949	41486	77	80	745	63	2012-08-02 11:23:28.902649
41491	3	1-11766	41490	97	71	7989	90	2012-08-02 11:23:28.902649
41492	3	1-12724	41490	71	77	5495	91	2012-08-02 11:23:28.902649
41490	3	1-12949	41490	77	97	6766	63	2012-08-02 11:23:28.902649
41494	3	1-11630	41493	68	72	1700	53	2012-08-02 11:23:28.902649
41495	3	1-11977	41493	72	77	1237	91	2012-08-02 11:23:28.902649
41493	3	1-12949	41493	77	68	1609	63	2012-08-02 11:23:28.902649
41554	3	1-12958	41553	51	40	6079	58	2012-08-02 11:23:33.576217
41555	3	1-11194	41553	40	62	2345	3	2012-08-02 11:23:33.576217
41553	3	1-12961	41553	62	51	3851	66	2012-08-02 11:23:33.576217
41557	4	1-12800	41556	77	57	4341	13	2012-08-02 11:23:33.576217
41558	4	1-11200	41556	57	15	3516	27	2012-08-02 11:23:33.576217
41559	4	1-12351	41556	15	62	1626	3	2012-08-02 11:23:33.576217
41556	4	1-12961	41556	62	77	2691	66	2012-08-02 11:23:33.576217
41561	3	1-8583	41560	44	18	1109	65	2012-08-02 11:23:34.747401
41562	3	1-5668	41560	18	69	737	10	2012-08-02 11:23:34.747401
41560	3	1-12964	41560	69	44	2459	91	2012-08-02 11:23:34.747401
41564	4	1-11323	41563	52	9	4193	99	2012-08-02 11:23:34.747401
41565	4	1-12636	41563	9	9	3565	47	2012-08-02 11:23:34.747401
41566	4	1-11634	41563	9	69	1660	10	2012-08-02 11:23:34.747401
41563	4	1-12964	41563	69	52	5579	91	2012-08-02 11:23:34.747401
41567	1	1-294	\N	90	90	52	47	2012-08-02 11:23:34.747401
41569	4	1-12954	41568	37	52	1106	78	2012-08-02 11:23:37.112039
41570	4	1-8676	41568	52	56	1437	14	2012-08-02 11:23:37.112039
41571	4	1-12366	41568	56	6	569	82	2012-08-02 11:23:37.112039
41568	4	1-12970	41568	6	37	1174	38	2012-08-02 11:23:37.112039
41619	4	1-12895	41618	14	26	276	52	2012-08-02 11:23:43.77499
41620	4	1-5880	41618	26	63	113	86	2012-08-02 11:23:43.77499
41621	4	1-12269	41618	63	10	216	92	2012-08-02 11:23:43.77499
41618	4	1-12987	41618	10	14	695	15	2012-08-02 11:23:43.77499
41632	4	1-11473	41631	8	6	481	82	2012-08-02 11:23:44.946237
41633	4	1-12970	41631	6	37	965	38	2012-08-02 11:23:44.946237
41634	4	1-12954	41631	37	88	885	78	2012-08-02 11:23:44.946237
41631	4	1-12990	41631	88	8	1627	96	2012-08-02 11:23:44.946237
41648	3	1-12918	41647	32	34	2585	46	2012-08-02 11:23:46.117465
41649	3	1-12950	41647	34	39	5235	56	2012-08-02 11:23:46.117465
41647	3	1-12993	41647	39	32	2475	45	2012-08-02 11:23:46.117465
41673	4	1-12624	41672	17	40	58	26	2012-08-02 11:23:53.941308
41674	4	1-7365	41672	40	50	56	69	2012-08-02 11:23:53.941308
41675	4	1-12245	41672	50	47	53	81	2012-08-02 11:23:53.941308
41672	4	1-13013	41672	47	17	42	94	2012-08-02 11:23:53.941308
41677	4	1-12624	41676	17	45	190	26	2012-08-02 11:23:53.941308
41678	4	1-12365	41676	45	9	167	75	2012-08-02 11:23:53.941308
41679	4	1-12598	41676	9	47	165	81	2012-08-02 11:23:53.941308
41676	4	1-13013	41676	47	17	135	94	2012-08-02 11:23:53.941308
41708	3	1-12957	41707	31	83	1570	7	2012-08-02 11:23:58.228472
41709	3	1-11603	41707	83	71	388	20	2012-08-02 11:23:58.228472
41707	3	1-13024	41707	71	31	978	81	2012-08-02 11:23:58.228472
41711	4	1-9967	41710	38	62	30	36	2012-08-02 11:23:58.228472
41239	4	1-9454	41236	94	92	159	75	2012-08-02 11:22:53.156836
41236	4	1-12858	41236	92	18	329	32	2012-08-02 11:22:53.156836
41243	4	1-12103	41242	62	34	1992	21	2012-08-02 11:22:56.724639
41244	4	1-12658	41242	34	18	1442	79	2012-08-02 11:22:56.724639
41245	4	1-12523	41242	18	87	1371	29	2012-08-02 11:22:56.724639
41242	4	1-12867	41242	87	62	3328	60	2012-08-02 11:22:56.724639
41247	3	1-12272	41246	26	64	3412	45	2012-08-02 11:22:57.949342
41248	3	1-12700	41246	64	31	1872	61	2012-08-02 11:22:57.949342
41246	3	1-12869	41246	31	26	7610	60	2012-08-02 11:22:57.949342
41250	4	1-11694	41249	45	76	705	96	2012-08-02 11:22:57.949342
41251	4	1-12131	41249	76	64	555	45	2012-08-02 11:22:57.949342
41252	4	1-12700	41249	64	31	345	61	2012-08-02 11:22:57.949342
41249	4	1-12869	41249	31	45	1589	60	2012-08-02 11:22:57.949342
41259	4	1-12608	41258	45	2	517	8	2012-08-02 11:22:58.72632
41260	4	1-11915	41258	2	86	382	46	2012-08-02 11:22:58.72632
41261	4	1-11525	41258	86	26	124	94	2012-08-02 11:22:58.72632
41258	4	1-12872	41258	26	45	607	18	2012-08-02 11:22:58.72632
41263	4	1-11531	41262	79	41	2296	81	2012-08-02 11:22:58.72632
41264	4	1-11445	41262	41	86	854	46	2012-08-02 11:22:58.72632
41265	4	1-11525	41262	86	26	348	94	2012-08-02 11:22:58.72632
41262	4	1-12872	41262	26	79	2122	18	2012-08-02 11:22:58.72632
41267	4	1-11388	41266	11	12	1420	53	2012-08-02 11:22:58.72632
41268	4	1-12787	41266	12	86	1385	46	2012-08-02 11:22:58.72632
41269	4	1-11525	41266	86	26	596	94	2012-08-02 11:22:58.72632
41266	4	1-12872	41266	26	11	3845	18	2012-08-02 11:22:58.72632
41279	4	1-11778	41278	92	68	1621	85	2012-08-02 11:22:59.871787
41280	4	1-12826	41278	68	100	1259	82	2012-08-02 11:22:59.871787
41281	4	1-11842	41278	100	51	1254	80	2012-08-02 11:22:59.871787
41278	4	1-12875	41278	51	92	3814	2	2012-08-02 11:22:59.871787
41283	4	1-12244	41282	67	54	702	95	2012-08-02 11:22:59.871787
41284	4	1-12101	41282	54	25	708	98	2012-08-02 11:22:59.871787
41285	4	1-12719	41282	25	51	982	80	2012-08-02 11:22:59.871787
41282	4	1-12875	41282	51	67	4047	2	2012-08-02 11:22:59.871787
41287	4	1-10885	41286	57	17	1076	74	2012-08-02 11:23:01.036268
41288	4	1-11627	41286	17	28	1503	67	2012-08-02 11:23:01.036268
41289	4	1-12848	41286	28	83	1556	7	2012-08-02 11:23:01.036268
41286	4	1-12878	41286	83	57	2246	4	2012-08-02 11:23:01.036268
41291	4	1-10885	41290	57	79	1580	74	2012-08-02 11:23:01.036268
41292	4	1-9693	41290	79	44	1147	53	2012-08-02 11:23:01.036268
41293	4	1-12720	41290	44	83	2226	7	2012-08-02 11:23:01.036268
41290	4	1-12878	41290	83	57	3255	4	2012-08-02 11:23:01.036268
41314	4	1-12831	41313	18	33	889	71	2012-08-02 11:23:03.356209
41315	4	1-3671	41313	33	94	551	37	2012-08-02 11:23:03.356209
41316	4	1-9848	41313	94	9	269	50	2012-08-02 11:23:03.356209
41313	4	1-12884	41313	9	18	205	32	2012-08-02 11:23:03.356209
41322	2	1-11778	41321	92	66	2039	85	2012-08-02 11:23:05.676201
41321	2	1-12890	41321	66	92	3586	2	2012-08-02 11:23:05.676201
41324	4	1-8273	41323	9	68	2559	10	2012-08-02 11:23:06.836243
41325	4	1-10775	41323	68	12	2149	32	2012-08-02 11:23:06.836243
41326	4	1-12856	41323	12	40	1348	90	2012-08-02 11:23:06.836243
41323	4	1-12892	41323	40	9	5622	22	2012-08-02 11:23:06.836243
41328	3	1-11550	41327	32	59	104	56	2012-08-02 11:23:07.609567
41329	3	1-12702	41327	59	10	244	44	2012-08-02 11:23:07.609567
41327	3	1-12894	41327	10	32	477	81	2012-08-02 11:23:07.609567
41331	4	1-12538	41330	34	10	1846	44	2012-08-02 11:23:08.394079
41332	4	1-12894	41330	10	46	3922	81	2012-08-02 11:23:08.394079
41333	4	1-12294	41330	46	69	2656	25	2012-08-02 11:23:08.394079
41330	4	1-12897	41330	69	34	2630	23	2012-08-02 11:23:08.394079
41342	4	1-12702	41341	59	16	2733	44	2012-08-02 11:23:09.565198
41343	4	1-12030	41341	16	58	2030	98	2012-08-02 11:23:09.565198
41344	4	1-11637	41341	58	6	923	100	2012-08-02 11:23:09.565198
41341	4	1-12900	41341	6	59	1124	56	2012-08-02 11:23:09.565198
41346	4	1-12282	41345	84	49	1676	81	2012-08-02 11:23:09.565198
41347	4	1-10749	41345	49	98	1485	7	2012-08-02 11:23:09.565198
41348	4	1-6071	41345	98	6	551	100	2012-08-02 11:23:09.565198
41345	4	1-12900	41345	6	84	734	56	2012-08-02 11:23:09.565198
41350	3	1-7449	41349	9	62	2249	60	2012-08-02 11:23:10.725197
41351	3	1-12103	41349	62	6	647	21	2012-08-02 11:23:10.725197
41349	3	1-12903	41349	6	9	4990	62	2012-08-02 11:23:10.725197
41353	2	1-12654	41352	23	45	4832	54	2012-08-02 11:23:11.863359
41352	2	1-12906	41352	45	23	8972	68	2012-08-02 11:23:11.863359
41366	3	1-9037	41365	37	83	3686	10	2012-08-02 11:23:13.02337
41367	3	1-12657	41365	83	88	4117	20	2012-08-02 11:23:13.02337
41365	3	1-12909	41365	88	37	7475	19	2012-08-02 11:23:13.02337
41385	3	1-12253	41384	11	96	1504	95	2012-08-02 11:23:17.376362
41386	3	1-12651	41384	96	86	1356	39	2012-08-02 11:23:17.376362
41384	3	1-12920	41384	86	11	1333	19	2012-08-02 11:23:17.376362
41388	4	1-12398	41387	18	18	282	79	2012-08-02 11:23:17.376362
41389	4	1-12523	41387	18	55	269	29	2012-08-02 11:23:17.376362
41390	4	1-12386	41387	55	86	178	39	2012-08-02 11:23:17.376362
41387	4	1-12920	41387	86	18	196	19	2012-08-02 11:23:17.376362
41392	4	1-12536	41391	4	98	410	26	2012-08-02 11:23:17.376362
41393	4	1-12433	41391	98	66	853	40	2012-08-02 11:23:17.376362
41394	4	1-11527	41391	66	86	293	39	2012-08-02 11:23:17.376362
41391	4	1-12920	41391	86	4	335	19	2012-08-02 11:23:17.376362
41396	3	1-9770	41395	50	47	536	59	2012-08-02 11:23:17.376362
41397	3	1-12222	41395	47	86	403	39	2012-08-02 11:23:17.376362
41395	3	1-12920	41395	86	50	466	19	2012-08-02 11:23:17.376362
41406	2	1-12030	41405	16	44	3592	98	2012-08-02 11:23:18.547636
41405	2	1-12924	41405	44	16	4986	44	2012-08-02 11:23:18.547636
41408	4	1-12894	41407	10	99	574	81	2012-08-02 11:23:18.547636
41409	4	1-7981	41407	99	81	162	41	2012-08-02 11:23:18.547636
41410	4	1-12883	41407	81	44	161	98	2012-08-02 11:23:18.547636
41407	4	1-12924	41407	44	10	255	44	2012-08-02 11:23:18.547636
41444	2	1-10774	41443	92	73	706	9	2012-08-02 11:23:22.625081
41443	2	1-12933	41443	73	92	981	28	2012-08-02 11:23:22.625081
41475	4	1-12914	41474	43	61	2979	26	2012-08-02 11:23:27.333831
41476	4	1-12008	41474	61	74	2395	5	2012-08-02 11:23:27.333831
41477	4	1-12790	41474	74	28	5621	72	2012-08-02 11:23:27.333831
41474	4	1-12945	41474	28	43	5460	16	2012-08-02 11:23:27.333831
41479	4	1-12914	41478	43	45	1226	26	2012-08-02 11:23:27.333831
41480	4	1-12365	41478	45	85	1081	75	2012-08-02 11:23:27.333831
41481	4	1-11207	41478	85	28	1210	72	2012-08-02 11:23:27.333831
41478	4	1-12945	41478	28	43	1625	16	2012-08-02 11:23:27.333831
41438	4	1-9724	41435	48	76	593	42	2012-08-02 11:23:20.912193
41435	4	1-12929	41435	76	92	1300	75	2012-08-02 11:23:20.912193
41440	4	1-12606	41439	6	55	155	88	2012-08-02 11:23:20.912193
41441	4	1-12926	41439	55	1	168	33	2012-08-02 11:23:20.912193
41442	4	1-12162	41439	1	76	125	42	2012-08-02 11:23:20.912193
41439	4	1-12929	41439	76	6	300	75	2012-08-02 11:23:20.912193
41454	4	1-11087	41453	36	12	175	88	2012-08-02 11:23:24.580265
41455	4	1-12440	41453	12	84	397	93	2012-08-02 11:23:24.580265
41456	4	1-9441	41453	84	3	221	96	2012-08-02 11:23:24.580265
41453	4	1-12938	41453	3	36	553	70	2012-08-02 11:23:24.580265
41458	4	1-12729	41457	23	27	1662	28	2012-08-02 11:23:24.580265
41459	4	1-9732	41457	27	91	2376	2	2012-08-02 11:23:24.580265
41460	4	1-12573	41457	91	3	780	96	2012-08-02 11:23:24.580265
41457	4	1-12938	41457	3	23	2542	70	2012-08-02 11:23:24.580265
41462	4	1-12908	41461	56	87	343	34	2012-08-02 11:23:25.731152
41463	4	1-12769	41461	87	80	309	6	2012-08-02 11:23:25.731152
41464	4	1-11941	41461	80	64	157	77	2012-08-02 11:23:25.731152
41461	4	1-12941	41461	64	56	196	75	2012-08-02 11:23:25.731152
41466	2	1-6780	41465	56	64	3164	77	2012-08-02 11:23:25.731152
41465	2	1-12941	41465	64	56	5734	75	2012-08-02 11:23:25.731152
41468	4	1-12303	41467	69	21	1071	30	2012-08-02 11:23:25.731152
41469	4	1-11845	41467	21	21	1052	99	2012-08-02 11:23:25.731152
41470	4	1-12207	41467	21	64	868	77	2012-08-02 11:23:25.731152
41467	4	1-12941	41467	64	69	1886	75	2012-08-02 11:23:25.731152
41472	3	1-12858	41471	92	12	1539	32	2012-08-02 11:23:25.731152
41473	3	1-10129	41471	12	64	392	77	2012-08-02 11:23:25.731152
41471	3	1-12941	41471	64	92	857	75	2012-08-02 11:23:25.731152
41497	3	1-12282	41496	84	41	38	81	2012-08-02 11:23:29.300201
41498	3	1-11445	41496	41	34	13	46	2012-08-02 11:23:29.300201
41496	3	1-12950	41496	34	84	21	56	2012-08-02 11:23:29.300201
41500	3	1-10529	41499	84	47	1416	97	2012-08-02 11:23:29.300201
41501	3	1-11635	41499	47	34	720	46	2012-08-02 11:23:29.300201
41499	3	1-12950	41499	34	84	1406	56	2012-08-02 11:23:29.300201
41519	4	1-9454	41518	94	56	400	75	2012-08-02 11:23:30.449552
41520	4	1-12908	41518	56	30	1218	34	2012-08-02 11:23:30.449552
41521	4	1-12185	41518	30	37	1037	38	2012-08-02 11:23:30.449552
41518	4	1-12954	41518	37	94	857	78	2012-08-02 11:23:30.449552
41523	4	1-12530	41522	99	50	6834	77	2012-08-02 11:23:30.449552
41524	4	1-10908	41522	50	30	2207	34	2012-08-02 11:23:30.449552
41525	4	1-12185	41522	30	37	2383	38	2012-08-02 11:23:30.449552
41522	4	1-12954	41522	37	99	2496	78	2012-08-02 11:23:30.449552
41527	4	1-12234	41526	67	85	627	35	2012-08-02 11:23:32.018196
41528	4	1-9723	41526	85	10	404	44	2012-08-02 11:23:32.018196
41529	4	1-12894	41526	10	31	697	81	2012-08-02 11:23:32.018196
41526	4	1-12957	41526	31	67	1970	7	2012-08-02 11:23:32.018196
41531	3	1-12241	41530	80	6	1562	16	2012-08-02 11:23:32.018196
41532	3	1-12727	41530	6	31	518	81	2012-08-02 11:23:32.018196
41530	3	1-12957	41530	31	80	1567	7	2012-08-02 11:23:32.018196
41534	3	1-12518	41533	38	60	308	2	2012-08-02 11:23:32.018196
41535	3	1-11161	41533	60	31	68	81	2012-08-02 11:23:32.018196
41533	3	1-12957	41533	31	38	206	7	2012-08-02 11:23:32.018196
41537	4	1-12775	41536	78	34	1014	78	2012-08-02 11:23:32.018196
41538	4	1-9024	41536	34	50	686	69	2012-08-02 11:23:32.018196
41539	4	1-12245	41536	50	31	585	81	2012-08-02 11:23:32.018196
41536	4	1-12957	41536	31	78	1952	7	2012-08-02 11:23:32.018196
41541	4	1-11847	41540	74	53	1025	45	2012-08-02 11:23:32.018196
41542	4	1-11041	41540	53	10	205	44	2012-08-02 11:23:32.018196
41543	4	1-12894	41540	10	31	450	81	2012-08-02 11:23:32.018196
41540	4	1-12957	41540	31	74	1623	7	2012-08-02 11:23:32.018196
41544	1	1-10749	\N	49	49	1921	7	2012-08-02 11:23:32.018196
41573	4	1-12936	41572	75	30	3362	4	2012-08-02 11:23:37.874152
41574	4	1-12896	41572	30	6	3113	1	2012-08-02 11:23:37.874152
41575	4	1-7347	41572	6	77	2568	95	2012-08-02 11:23:37.874152
41572	4	1-12972	41572	77	75	5432	5	2012-08-02 11:23:37.874152
41577	4	1-9198	41576	75	69	1246	40	2012-08-02 11:23:37.874152
41578	4	1-12582	41576	69	74	1322	45	2012-08-02 11:23:37.874152
41579	4	1-12838	41576	74	77	675	95	2012-08-02 11:23:37.874152
41576	4	1-12972	41576	77	75	1546	5	2012-08-02 11:23:37.874152
41593	4	1-10734	41592	59	99	416	31	2012-08-02 11:23:40.238195
41594	4	1-12873	41592	99	12	908	12	2012-08-02 11:23:40.238195
41595	4	1-12080	41592	12	11	530	1	2012-08-02 11:23:40.238195
41592	4	1-12978	41592	11	59	1629	16	2012-08-02 11:23:40.238195
41597	4	1-10674	41596	31	27	1439	80	2012-08-02 11:23:40.238195
41598	4	1-12803	41596	27	30	1101	4	2012-08-02 11:23:40.238195
41599	4	1-12896	41596	30	11	977	1	2012-08-02 11:23:40.238195
41596	4	1-12978	41596	11	31	3025	16	2012-08-02 11:23:40.238195
41601	3	1-11750	41600	24	63	508	33	2012-08-02 11:23:40.238195
41602	3	1-12289	41600	63	11	468	1	2012-08-02 11:23:40.238195
41600	3	1-12978	41600	11	24	1579	16	2012-08-02 11:23:40.238195
41603	1	1-10131	\N	20	20	741	31	2012-08-02 11:23:40.238195
41623	3	1-11578	41622	6	71	3287	90	2012-08-02 11:23:44.161638
41624	3	1-12724	41622	71	18	1266	91	2012-08-02 11:23:44.161638
41622	3	1-12988	41622	18	6	5644	75	2012-08-02 11:23:44.161638
41626	3	1-12858	41625	92	53	3315	32	2012-08-02 11:23:44.161638
41627	3	1-9926	41625	53	18	605	91	2012-08-02 11:23:44.161638
41625	3	1-12988	41625	18	92	2875	75	2012-08-02 11:23:44.161638
41629	3	1-11458	41628	50	71	515	90	2012-08-02 11:23:44.161638
41630	3	1-12724	41628	71	18	251	91	2012-08-02 11:23:44.161638
41628	3	1-12988	41628	18	50	1420	75	2012-08-02 11:23:44.161638
41636	4	1-3671	41635	33	20	553	37	2012-08-02 11:23:45.333175
41637	4	1-10958	41635	20	19	200	43	2012-08-02 11:23:45.333175
41638	4	1-10503	41635	19	90	140	50	2012-08-02 11:23:45.333175
41635	4	1-12991	41635	90	33	1103	71	2012-08-02 11:23:45.333175
41640	4	1-11824	41639	71	47	3922	52	2012-08-02 11:23:45.333175
41641	4	1-588	41639	47	59	770	27	2012-08-02 11:23:45.333175
41642	4	1-11905	41639	59	90	696	50	2012-08-02 11:23:45.333175
41639	4	1-12991	41639	90	71	5767	71	2012-08-02 11:23:45.333175
41644	4	1-10988	41643	38	22	504	70	2012-08-02 11:23:45.333175
41645	4	1-11998	41643	22	32	319	84	2012-08-02 11:23:45.333175
41646	4	1-10568	41643	32	90	208	50	2012-08-02 11:23:45.333175
41643	4	1-12991	41643	90	38	1870	71	2012-08-02 11:23:45.333175
41651	4	1-11827	41650	48	97	2235	83	2012-08-02 11:23:46.504397
41650	4	1-12995	41650	18	48	1615	41	2012-08-02 11:23:46.504397
41483	4	1-11714	41482	87	44	1653	25	2012-08-02 11:23:28.515498
41484	4	1-12400	41482	44	56	808	47	2012-08-02 11:23:28.515498
41485	4	1-9540	41482	56	81	159	35	2012-08-02 11:23:28.515498
41482	4	1-12948	41482	81	87	2394	36	2012-08-02 11:23:28.515498
41503	4	1-12298	41502	1	39	400	11	2012-08-02 11:23:29.686653
41504	4	1-12830	41502	39	85	472	30	2012-08-02 11:23:29.686653
41505	4	1-12550	41502	85	51	1548	2	2012-08-02 11:23:29.686653
41502	4	1-12951	41502	51	1	612	84	2012-08-02 11:23:29.686653
41507	3	1-10510	41506	87	66	1884	85	2012-08-02 11:23:29.686653
41508	3	1-12890	41506	66	51	3568	2	2012-08-02 11:23:29.686653
41506	3	1-12951	41506	51	87	1459	84	2012-08-02 11:23:29.686653
41510	4	1-12332	41509	49	36	621	11	2012-08-02 11:23:30.835937
41511	4	1-8177	41509	36	36	437	58	2012-08-02 11:23:30.835937
41512	4	1-10881	41509	36	79	310	92	2012-08-02 11:23:30.835937
41509	4	1-12953	41509	79	49	1451	13	2012-08-02 11:23:30.835937
41514	3	1-11200	41513	57	4	2393	27	2012-08-02 11:23:30.835937
41515	3	1-12137	41513	4	79	1079	92	2012-08-02 11:23:30.835937
41513	3	1-12953	41513	79	57	5327	13	2012-08-02 11:23:30.835937
41516	1	1-8177	\N	36	36	686	58	2012-08-02 11:23:30.835937
41517	1	1-2307	\N	34	34	484	58	2012-08-02 11:23:30.835937
41546	4	1-12755	41545	32	15	6626	8	2012-08-02 11:23:32.79149
41547	4	1-11887	41545	15	35	4749	58	2012-08-02 11:23:32.79149
41548	4	1-11792	41545	35	14	6278	25	2012-08-02 11:23:32.79149
41545	4	1-12959	41545	14	32	7274	44	2012-08-02 11:23:32.79149
41550	4	1-12755	41549	32	77	370	8	2012-08-02 11:23:32.79149
41551	4	1-12740	41549	77	53	481	91	2012-08-02 11:23:32.79149
41552	4	1-11417	41549	53	14	343	25	2012-08-02 11:23:32.79149
41549	4	1-12959	41549	14	32	402	44	2012-08-02 11:23:32.79149
41581	4	1-9429	41580	5	60	976	10	2012-08-02 11:23:39.862557
41582	4	1-11762	41580	60	3	718	96	2012-08-02 11:23:39.862557
41583	4	1-12938	41580	3	10	2078	70	2012-08-02 11:23:39.862557
41580	4	1-12977	41580	10	5	2816	99	2012-08-02 11:23:39.862557
41585	4	1-12119	41584	17	62	1248	6	2012-08-02 11:23:39.862557
41586	4	1-12671	41584	62	92	858	66	2012-08-02 11:23:39.862557
41587	4	1-11969	41584	92	10	1461	70	2012-08-02 11:23:39.862557
41584	4	1-12977	41584	10	17	2008	99	2012-08-02 11:23:39.862557
41589	4	1-12746	41588	38	20	1842	37	2012-08-02 11:23:39.862557
41590	4	1-10958	41588	20	77	863	43	2012-08-02 11:23:39.862557
41591	4	1-7752	41588	77	10	712	70	2012-08-02 11:23:39.862557
41588	4	1-12977	41588	10	38	1059	99	2012-08-02 11:23:39.862557
41605	2	1-12952	41604	65	41	4290	70	2012-08-02 11:23:41.000653
41604	2	1-12980	41604	41	65	6033	93	2012-08-02 11:23:41.000653
41607	4	1-12433	41606	98	66	2547	40	2012-08-02 11:23:42.194146
41608	4	1-11527	41606	66	63	730	39	2012-08-02 11:23:42.194146
41609	4	1-12751	41606	63	57	859	78	2012-08-02 11:23:42.194146
41606	4	1-12983	41606	57	98	1466	26	2012-08-02 11:23:42.194146
41611	4	1-3980	41610	37	83	1009	73	2012-08-02 11:23:42.194146
41612	4	1-12944	41610	83	37	1773	38	2012-08-02 11:23:42.194146
41613	4	1-12954	41610	37	57	1806	78	2012-08-02 11:23:42.194146
41610	4	1-12983	41610	57	37	3509	26	2012-08-02 11:23:42.194146
41615	4	1-12580	41614	99	91	1919	66	2012-08-02 11:23:43.365977
41616	4	1-8180	41614	91	12	1257	88	2012-08-02 11:23:43.365977
41617	4	1-12440	41614	12	18	3649	93	2012-08-02 11:23:43.365977
41614	4	1-12986	41614	18	99	2005	24	2012-08-02 11:23:43.365977
41669	4	1-10022	41668	90	38	1435	21	2012-08-02 11:23:53.543515
41670	4	1-12423	41668	38	76	287	5	2012-08-02 11:23:53.543515
41671	4	1-12797	41668	76	51	142	83	2012-08-02 11:23:53.543515
41668	4	1-13012	41668	51	90	7584	71	2012-08-02 11:23:53.543515
41681	4	1-8136	41680	17	68	1879	79	2012-08-02 11:23:54.703668
41682	4	1-12985	41680	68	50	1312	63	2012-08-02 11:23:54.703668
41683	4	1-12349	41680	50	80	1543	70	2012-08-02 11:23:54.703668
41680	4	1-13016	41680	80	17	6375	81	2012-08-02 11:23:54.703668
41685	3	1-13013	41684	47	18	2231	94	2012-08-02 11:23:54.703668
41686	3	1-6027	41684	18	80	823	70	2012-08-02 11:23:54.703668
41684	3	1-13016	41684	80	47	3557	81	2012-08-02 11:23:54.703668
41719	3	1-12213	41718	31	41	4278	4	2012-08-02 11:23:59.785955
41720	3	1-11119	41718	41	44	1250	19	2012-08-02 11:23:59.785955
41718	3	1-13028	41718	44	31	6394	42	2012-08-02 11:23:59.785955
41722	4	1-12889	41721	62	77	75	74	2012-08-02 11:23:59.785955
41723	4	1-2769	41721	77	24	17	73	2012-08-02 11:23:59.785955
41724	4	1-12592	41721	24	44	24	19	2012-08-02 11:23:59.785955
41721	4	1-13028	41721	44	62	123	42	2012-08-02 11:23:59.785955
41730	2	1-12025	41729	1	90	4841	10	2012-08-02 11:24:03.210707
41729	2	1-13037	41729	90	1	8518	12	2012-08-02 11:24:03.210707
41732	3	1-12462	41731	1	45	126	64	2012-08-02 11:24:03.210707
41733	3	1-10972	41731	45	90	60	10	2012-08-02 11:24:03.210707
41731	3	1-13037	41731	90	1	106	12	2012-08-02 11:24:03.210707
41755	3	1-5668	41754	18	20	2403	10	2012-08-02 11:24:09.430813
41756	3	1-12237	41754	20	68	593	58	2012-08-02 11:24:09.430813
41754	3	1-13053	41754	68	18	7454	65	2012-08-02 11:24:09.430813
41758	3	1-11975	41757	4	88	584	45	2012-08-02 11:24:09.430813
41759	3	1-4157	41757	88	68	47	58	2012-08-02 11:24:09.430813
41757	3	1-13053	41757	68	4	809	65	2012-08-02 11:24:09.430813
41775	3	1-12367	41774	14	14	1871	84	2012-08-02 11:24:12.922075
41776	3	1-10323	41774	14	1	1089	16	2012-08-02 11:24:12.922075
41774	3	1-13062	41774	1	14	3344	50	2012-08-02 11:24:12.922075
41778	3	1-7400	41777	56	11	104	1	2012-08-02 11:24:12.922075
41779	3	1-12978	41777	11	1	263	16	2012-08-02 11:24:12.922075
41777	3	1-13062	41777	1	56	830	50	2012-08-02 11:24:12.922075
41781	4	1-12849	41780	51	9	331	19	2012-08-02 11:24:12.922075
41782	4	1-12763	41780	9	28	385	72	2012-08-02 11:24:12.922075
41783	4	1-12945	41780	28	1	377	16	2012-08-02 11:24:12.922075
41780	4	1-13062	41780	1	51	1210	50	2012-08-02 11:24:12.922075
41785	4	1-12849	41784	51	10	220	19	2012-08-02 11:24:12.922075
41786	4	1-12685	41784	10	96	174	26	2012-08-02 11:24:12.922075
41787	4	1-9377	41784	96	1	184	16	2012-08-02 11:24:12.922075
41784	4	1-13062	41784	1	51	689	50	2012-08-02 11:24:12.922075
41789	4	1-914	41788	78	10	191	76	2012-08-02 11:24:12.922075
41790	4	1-12652	41788	10	48	299	55	2012-08-02 11:24:12.922075
41791	4	1-9499	41788	48	1	275	16	2012-08-02 11:24:12.922075
41788	4	1-13062	41788	1	78	1081	50	2012-08-02 11:24:12.922075
41652	4	1-9276	41650	97	23	963	70	2012-08-02 11:23:46.504397
41653	4	1-12729	41650	23	18	507	28	2012-08-02 11:23:46.504397
41655	4	1-12717	41654	75	97	1328	83	2012-08-02 11:23:46.504397
41656	4	1-9276	41654	97	23	667	70	2012-08-02 11:23:46.504397
41657	4	1-12729	41654	23	18	409	28	2012-08-02 11:23:46.504397
41654	4	1-12995	41654	18	75	1520	41	2012-08-02 11:23:46.504397
41659	4	1-12018	41658	38	53	1460	47	2012-08-02 11:23:46.504397
41660	4	1-6532	41658	53	73	416	9	2012-08-02 11:23:46.504397
41661	4	1-12933	41658	73	18	564	28	2012-08-02 11:23:46.504397
41658	4	1-12995	41658	18	38	2102	41	2012-08-02 11:23:46.504397
41663	2	1-9343	41662	44	18	377	28	2012-08-02 11:23:46.504397
41662	2	1-12995	41662	18	44	1492	41	2012-08-02 11:23:46.504397
41665	4	1-12883	41664	81	72	2982	98	2012-08-02 11:23:46.504397
41666	4	1-12586	41664	72	23	1107	70	2012-08-02 11:23:46.504397
41667	4	1-12729	41664	23	18	724	28	2012-08-02 11:23:46.504397
41664	4	1-12995	41664	18	81	2866	41	2012-08-02 11:23:46.504397
41688	4	1-12336	41687	44	47	3279	81	2012-08-02 11:23:55.476965
41689	4	1-13013	41687	47	90	2146	94	2012-08-02 11:23:55.476965
41690	4	1-8836	41687	90	56	923	17	2012-08-02 11:23:55.476965
41687	4	1-13017	41687	56	44	1774	100	2012-08-02 11:23:55.476965
41692	4	1-12425	41691	46	83	1101	68	2012-08-02 11:23:55.476965
41693	4	1-12200	41691	83	30	1472	51	2012-08-02 11:23:55.476965
41694	4	1-12905	41691	30	56	929	17	2012-08-02 11:23:55.476965
41691	4	1-13017	41691	56	46	1794	100	2012-08-02 11:23:55.476965
41696	4	1-10894	41695	37	71	2641	59	2012-08-02 11:23:55.476965
41697	4	1-13014	41695	71	90	2214	22	2012-08-02 11:23:55.476965
41698	4	1-12940	41695	90	56	2223	17	2012-08-02 11:23:55.476965
41695	4	1-13017	41695	56	37	4505	100	2012-08-02 11:23:55.476965
41700	4	1-12739	41699	19	73	503	24	2012-08-02 11:23:55.476965
41701	4	1-11284	41699	73	93	819	77	2012-08-02 11:23:55.476965
41702	4	1-5872	41699	93	56	259	17	2012-08-02 11:23:55.476965
41699	4	1-13017	41699	56	19	550	100	2012-08-02 11:23:55.476965
41704	4	1-12653	41703	58	50	816	60	2012-08-02 11:23:56.637412
41705	4	1-11306	41703	50	68	187	79	2012-08-02 11:23:56.637412
41706	4	1-12985	41703	68	11	166	63	2012-08-02 11:23:56.637412
41703	4	1-13020	41703	11	58	344	98	2012-08-02 11:23:56.637412
41739	4	1-12729	41738	23	22	764	28	2012-08-02 11:24:04.779544
41740	4	1-12342	41738	22	85	666	53	2012-08-02 11:24:04.779544
41741	4	1-12877	41738	85	32	1460	72	2012-08-02 11:24:04.779544
41738	4	1-13041	41738	32	23	1279	70	2012-08-02 11:24:04.779544
41743	4	1-12980	41742	41	3	778	93	2012-08-02 11:24:04.779544
41744	4	1-9200	41742	3	85	245	53	2012-08-02 11:24:04.779544
41745	4	1-12877	41742	85	32	577	72	2012-08-02 11:24:04.779544
41742	4	1-13041	41742	32	41	543	70	2012-08-02 11:24:04.779544
41747	3	1-13043	41746	80	83	4987	12	2012-08-02 11:24:07.111196
41748	3	1-10376	41746	83	7	1243	94	2012-08-02 11:24:07.111196
41746	3	1-13046	41746	7	80	4761	98	2012-08-02 11:24:07.111196
41750	4	1-12719	41749	25	16	3444	80	2012-08-02 11:24:07.111196
41751	4	1-11365	41749	16	86	1461	46	2012-08-02 11:24:07.111196
41752	4	1-11525	41749	86	7	628	94	2012-08-02 11:24:07.111196
41749	4	1-13046	41749	7	25	2630	98	2012-08-02 11:24:07.111196
41753	1	1-11365	\N	16	16	927	46	2012-08-02 11:24:07.111196
41761	4	1-12074	41760	99	49	1354	76	2012-08-02 11:24:10.204408
41762	4	1-9364	41760	49	12	552	88	2012-08-02 11:24:10.204408
41763	4	1-12440	41760	12	56	1227	93	2012-08-02 11:24:10.204408
41760	4	1-13055	41760	56	99	2807	69	2012-08-02 11:24:10.204408
41765	4	1-8831	41764	19	36	858	70	2012-08-02 11:24:10.204408
41766	4	1-11087	41764	36	12	267	88	2012-08-02 11:24:10.204408
41767	4	1-12440	41764	12	56	595	93	2012-08-02 11:24:10.204408
41764	4	1-13055	41764	56	19	1364	69	2012-08-02 11:24:10.204408
41769	4	1-10765	41768	24	36	1396	70	2012-08-02 11:24:10.204408
41770	4	1-11087	41768	36	12	440	88	2012-08-02 11:24:10.204408
41771	4	1-12440	41768	12	56	991	93	2012-08-02 11:24:10.204408
41768	4	1-13055	41768	56	24	2297	69	2012-08-02 11:24:10.204408
41773	2	1-4235	41772	86	56	684	93	2012-08-02 11:24:10.204408
41772	2	1-13055	41772	56	86	1720	69	2012-08-02 11:24:10.204408
41800	4	1-12830	41799	39	85	894	30	2012-08-02 11:24:17.175544
41801	4	1-12550	41799	85	91	2475	2	2012-08-02 11:24:17.175544
41802	4	1-12573	41799	91	66	655	96	2012-08-02 11:24:17.175544
41799	4	1-13073	41799	66	39	898	11	2012-08-02 11:24:17.175544
41804	4	1-7724	41803	56	64	1185	29	2012-08-02 11:24:17.175544
41805	4	1-11179	41803	64	61	687	27	2012-08-02 11:24:17.175544
41806	4	1-12752	41803	61	66	825	96	2012-08-02 11:24:17.175544
41803	4	1-13073	41803	66	56	1252	11	2012-08-02 11:24:17.175544
41808	4	1-12830	41807	39	55	47	30	2012-08-02 11:24:17.175544
41809	4	1-10242	41807	55	28	32	80	2012-08-02 11:24:17.175544
41810	4	1-13061	41807	28	66	28	96	2012-08-02 11:24:17.175544
41807	4	1-13073	41807	66	39	43	11	2012-08-02 11:24:17.175544
41812	3	1-7593	41811	16	28	1772	80	2012-08-02 11:24:17.175544
41813	3	1-13061	41811	28	66	1700	96	2012-08-02 11:24:17.175544
41811	3	1-13073	41811	66	16	2831	11	2012-08-02 11:24:17.175544
41854	2	1-12163	41853	11	99	572	59	2012-08-02 11:24:25.320192
41853	2	1-13094	41853	99	11	7584	16	2012-08-02 11:24:25.320192
41868	3	1-12990	41867	88	66	5757	96	2012-08-02 11:24:28.767381
41869	3	1-13073	41867	66	44	3976	11	2012-08-02 11:24:28.767381
41867	3	1-13103	41867	44	88	6746	78	2012-08-02 11:24:28.767381
41871	4	1-7418	41870	43	44	1335	98	2012-08-02 11:24:29.949566
41872	4	1-12924	41870	44	72	1848	44	2012-08-02 11:24:29.949566
41873	4	1-13080	41870	72	64	3061	45	2012-08-02 11:24:29.949566
41870	4	1-13106	41870	64	43	2039	19	2012-08-02 11:24:29.949566
41875	4	1-12685	41874	10	98	1195	26	2012-08-02 11:24:29.949566
41876	4	1-12433	41874	98	69	2195	40	2012-08-02 11:24:29.949566
41877	4	1-12582	41874	69	64	2084	45	2012-08-02 11:24:29.949566
41874	4	1-13106	41874	64	10	1451	19	2012-08-02 11:24:29.949566
41879	4	1-12398	41878	18	1	439	79	2012-08-02 11:24:29.949566
41880	4	1-10456	41878	1	69	370	40	2012-08-02 11:24:29.949566
41881	4	1-12582	41878	69	64	393	45	2012-08-02 11:24:29.949566
41878	4	1-13106	41878	64	18	306	19	2012-08-02 11:24:29.949566
41919	3	1-12796	41918	81	19	3061	92	2012-08-02 11:24:37.075581
41920	3	1-12916	41918	19	8	4612	37	2012-08-02 11:24:37.075581
41918	3	1-13124	41918	8	81	4100	60	2012-08-02 11:24:37.075581
41922	4	1-12103	41921	62	58	557	21	2012-08-02 11:24:37.075581
41923	4	1-13029	41921	58	94	521	48	2012-08-02 11:24:37.075581
41924	4	1-12783	41921	94	8	1098	37	2012-08-02 11:24:37.075581
41712	4	1-10027	41710	62	69	21	57	2012-08-02 11:23:58.228472
41713	4	1-12780	41710	69	71	35	20	2012-08-02 11:23:58.228472
41710	4	1-13024	41710	71	38	159	81	2012-08-02 11:23:58.228472
41715	4	1-9758	41714	46	27	2693	52	2012-08-02 11:23:58.228472
41716	4	1-7650	41714	27	12	1956	2	2012-08-02 11:23:58.228472
41717	4	1-12493	41714	12	71	1157	20	2012-08-02 11:23:58.228472
41714	4	1-13024	41714	71	46	5499	81	2012-08-02 11:23:58.228472
41726	4	1-11967	41725	7	38	244	21	2012-08-02 11:24:00.537144
41727	4	1-12423	41725	38	76	186	5	2012-08-02 11:24:00.537144
41728	4	1-12797	41725	76	95	350	83	2012-08-02 11:24:00.537144
41725	4	1-13030	41725	95	7	163	28	2012-08-02 11:24:00.537144
41735	4	1-11362	41734	72	83	2597	45	2012-08-02 11:24:03.98406
41736	4	1-12421	41734	83	99	2908	14	2012-08-02 11:24:03.98406
41737	4	1-12071	41734	99	70	1386	13	2012-08-02 11:24:03.98406
41734	4	1-13039	41734	70	72	3339	85	2012-08-02 11:24:03.98406
41796	4	1-11962	41795	72	77	872	95	2012-08-02 11:24:16.777883
41797	4	1-12972	41795	77	48	1945	5	2012-08-02 11:24:16.777883
41798	4	1-11624	41795	48	58	733	68	2012-08-02 11:24:16.777883
41795	4	1-13071	41795	58	72	1184	26	2012-08-02 11:24:16.777883
41815	3	1-11792	41814	35	14	2019	25	2012-08-02 11:24:17.573325
41816	3	1-12959	41814	14	71	737	44	2012-08-02 11:24:17.573325
41814	3	1-13074	41814	71	35	4850	58	2012-08-02 11:24:17.573325
41818	4	1-11861	41817	93	80	513	80	2012-08-02 11:24:17.573325
41819	4	1-12000	41817	80	14	202	25	2012-08-02 11:24:17.573325
41820	4	1-12959	41817	14	71	93	44	2012-08-02 11:24:17.573325
41817	4	1-13074	41817	71	93	770	58	2012-08-02 11:24:17.573325
41822	3	1-9807	41821	41	8	260	9	2012-08-02 11:24:17.573325
41823	3	1-6029	41821	8	71	102	44	2012-08-02 11:24:17.573325
41821	3	1-13074	41821	71	41	1212	58	2012-08-02 11:24:17.573325
41836	4	1-7521	41835	74	100	492	82	2012-08-02 11:24:20.71302
41837	4	1-11842	41835	100	28	502	80	2012-08-02 11:24:20.71302
41838	4	1-13061	41835	28	25	345	96	2012-08-02 11:24:20.71302
41835	4	1-13082	41835	25	74	1041	35	2012-08-02 11:24:20.71302
41840	3	1-11875	41839	79	61	2333	27	2012-08-02 11:24:20.71302
41841	3	1-12752	41839	61	25	2285	96	2012-08-02 11:24:20.71302
41839	3	1-13082	41839	25	79	7162	35	2012-08-02 11:24:20.71302
41843	4	1-12855	41842	100	36	1333	76	2012-08-02 11:24:20.71302
41844	4	1-9558	41842	36	28	617	80	2012-08-02 11:24:20.71302
41845	4	1-13061	41842	28	25	448	96	2012-08-02 11:24:20.71302
41842	4	1-13082	41842	25	100	1430	35	2012-08-02 11:24:20.71302
41847	3	1-3662	41846	49	45	13	60	2012-08-02 11:24:20.71302
41848	3	1-11694	41846	45	25	6	96	2012-08-02 11:24:20.71302
41846	3	1-13082	41846	25	49	22	35	2012-08-02 11:24:20.71302
41850	4	1-12807	41849	14	78	2472	41	2012-08-02 11:24:24.160096
41851	4	1-8487	41849	78	100	892	67	2012-08-02 11:24:24.160096
41852	4	1-8528	41849	100	30	556	31	2012-08-02 11:24:24.160096
41849	4	1-13090	41849	30	14	2495	70	2012-08-02 11:24:24.160096
41856	3	1-12791	41855	86	32	74	85	2012-08-02 11:24:26.082758
41857	3	1-12911	41855	32	35	65	54	2012-08-02 11:24:26.082758
41855	3	1-13096	41855	35	86	41	58	2012-08-02 11:24:26.082758
41895	3	1-12443	41894	1	86	3388	58	2012-08-02 11:24:30.74489
41896	3	1-12791	41894	86	73	4356	85	2012-08-02 11:24:30.74489
41894	3	1-13109	41894	73	1	8167	75	2012-08-02 11:24:30.74489
41902	3	1-11892	41901	76	37	1987	34	2012-08-02 11:24:33.496367
41903	3	1-10595	41901	37	4	987	22	2012-08-02 11:24:33.496367
41901	3	1-13115	41901	4	76	2832	93	2012-08-02 11:24:33.496367
41905	3	1-12917	41904	35	34	2057	46	2012-08-02 11:24:33.496367
41906	3	1-11326	41904	34	4	1620	22	2012-08-02 11:24:33.496367
41904	3	1-13115	41904	4	35	5187	93	2012-08-02 11:24:33.496367
41908	4	1-8824	41907	40	10	653	65	2012-08-02 11:24:33.496367
41909	4	1-12243	41907	10	45	400	28	2012-08-02 11:24:33.496367
41910	4	1-11746	41907	45	4	425	22	2012-08-02 11:24:33.496367
41907	4	1-13115	41907	4	40	1444	93	2012-08-02 11:24:33.496367
41942	3	1-12277	41941	82	53	6140	91	2012-08-02 11:24:39.384613
41943	3	1-11417	41941	53	38	4603	25	2012-08-02 11:24:39.384613
41941	3	1-13129	41941	38	82	4161	3	2012-08-02 11:24:39.384613
41945	4	1-9843	41944	33	60	541	42	2012-08-02 11:24:40.555909
41946	4	1-12551	41944	60	69	571	57	2012-08-02 11:24:40.555909
41947	4	1-12780	41944	69	70	1105	20	2012-08-02 11:24:40.555909
41944	4	1-13132	41944	70	33	558	4	2012-08-02 11:24:40.555909
41949	4	1-10344	41948	12	60	1379	42	2012-08-02 11:24:40.555909
41950	4	1-12551	41948	60	69	1585	57	2012-08-02 11:24:40.555909
41951	4	1-12780	41948	69	70	3339	20	2012-08-02 11:24:40.555909
41948	4	1-13132	41948	70	12	1836	4	2012-08-02 11:24:40.555909
41953	4	1-10344	41952	12	30	97	42	2012-08-02 11:24:40.555909
41954	4	1-13004	41952	30	69	110	57	2012-08-02 11:24:40.555909
41955	4	1-12780	41952	69	70	233	20	2012-08-02 11:24:40.555909
41952	4	1-13132	41952	70	12	129	4	2012-08-02 11:24:40.555909
41969	4	1-10730	41968	92	27	501	80	2012-08-02 11:24:41.705139
41970	4	1-12803	41968	27	30	439	4	2012-08-02 11:24:41.705139
41971	4	1-12896	41968	30	84	447	1	2012-08-02 11:24:41.705139
41968	4	1-13135	41968	84	92	538	69	2012-08-02 11:24:41.705139
41973	4	1-12505	41972	95	8	1296	92	2012-08-02 11:24:42.909484
41974	4	1-13119	41972	8	72	1302	88	2012-08-02 11:24:42.909484
41975	4	1-10921	41972	72	16	587	55	2012-08-02 11:24:42.909484
41972	4	1-13137	41972	16	95	969	17	2012-08-02 11:24:42.909484
41977	4	1-12059	41976	94	10	1246	23	2012-08-02 11:24:42.909484
41978	4	1-9868	41976	10	10	951	76	2012-08-02 11:24:42.909484
41979	4	1-12652	41976	10	16	1596	55	2012-08-02 11:24:42.909484
41976	4	1-13137	41976	16	94	2640	17	2012-08-02 11:24:42.909484
41793	3	1-9693	41792	79	76	1874	53	2012-08-02 11:24:14.10453
41794	3	1-10354	41792	76	90	321	17	2012-08-02 11:24:14.10453
41792	3	1-13065	41792	90	79	8668	74	2012-08-02 11:24:14.10453
41825	4	1-12377	41824	38	95	3862	83	2012-08-02 11:24:19.144836
41826	4	1-13030	41824	95	7	1817	28	2012-08-02 11:24:19.144836
41827	4	1-11967	41824	7	24	2747	21	2012-08-02 11:24:19.144836
41824	4	1-13077	41824	24	38	3603	13	2012-08-02 11:24:19.144836
41829	4	1-12838	41828	74	54	855	95	2012-08-02 11:24:19.939707
41830	4	1-12101	41828	54	44	793	98	2012-08-02 11:24:19.939707
41831	4	1-12924	41828	44	72	1125	44	2012-08-02 11:24:19.939707
41828	4	1-13080	41828	72	74	1909	45	2012-08-02 11:24:19.939707
41833	3	1-12918	41832	32	57	1622	46	2012-08-02 11:24:19.939707
41834	3	1-10835	41832	57	72	916	44	2012-08-02 11:24:19.939707
41832	3	1-13080	41832	72	32	1647	45	2012-08-02 11:24:19.939707
41859	3	1-12236	41858	80	20	962	37	2012-08-02 11:24:26.844739
41860	3	1-10958	41858	20	85	320	43	2012-08-02 11:24:26.844739
41858	3	1-13098	41858	85	80	1619	38	2012-08-02 11:24:26.844739
41862	3	1-6986	41861	27	82	426	31	2012-08-02 11:24:26.844739
41863	3	1-11282	41861	82	85	525	43	2012-08-02 11:24:26.844739
41861	3	1-13098	41861	85	27	3070	38	2012-08-02 11:24:26.844739
41865	3	1-6889	41864	17	19	4054	40	2012-08-02 11:24:28.004933
41866	3	1-13047	41864	19	68	2638	7	2012-08-02 11:24:28.004933
41864	3	1-13101	41864	68	17	7483	83	2012-08-02 11:24:28.004933
41883	4	1-13073	41882	66	41	770	11	2012-08-02 11:24:30.358203
41884	4	1-8845	41882	41	88	476	59	2012-08-02 11:24:30.358203
41885	4	1-11972	41882	88	6	320	1	2012-08-02 11:24:30.358203
41882	4	1-13107	41882	6	66	483	96	2012-08-02 11:24:30.358203
41887	4	1-10428	41886	82	82	2810	64	2012-08-02 11:24:30.358203
41888	4	1-12663	41886	82	47	4227	55	2012-08-02 11:24:30.358203
41889	4	1-12204	41886	47	6	2843	1	2012-08-02 11:24:30.358203
41886	4	1-13107	41886	6	82	4433	96	2012-08-02 11:24:30.358203
41891	4	1-10304	41890	69	27	3180	80	2012-08-02 11:24:30.358203
41892	4	1-12803	41890	27	30	2824	4	2012-08-02 11:24:30.358203
41893	4	1-12896	41890	30	6	2910	1	2012-08-02 11:24:30.358203
41890	4	1-13107	41890	6	69	4677	96	2012-08-02 11:24:30.358203
41898	4	1-11810	41897	13	24	1222	21	2012-08-02 11:24:31.529657
41899	4	1-11849	41897	24	62	719	6	2012-08-02 11:24:31.529657
41900	4	1-12671	41897	62	56	532	66	2012-08-02 11:24:31.529657
41897	4	1-13110	41897	56	13	1044	18	2012-08-02 11:24:31.529657
41912	4	1-11190	41911	14	32	1765	85	2012-08-02 11:24:33.89365
41913	4	1-12911	41911	32	95	2022	54	2012-08-02 11:24:33.89365
41914	4	1-12979	41911	95	38	1034	29	2012-08-02 11:24:33.89365
41911	4	1-13116	41911	38	14	2669	17	2012-08-02 11:24:33.89365
41916	3	1-12969	41915	91	81	5868	60	2012-08-02 11:24:35.065115
41917	3	1-12796	41915	81	8	4937	92	2012-08-02 11:24:35.065115
41915	3	1-13119	41915	8	91	5478	88	2012-08-02 11:24:35.065115
41938	4	1-13104	41937	66	77	1389	66	2012-08-02 11:24:38.986892
41939	4	1-12800	41937	77	25	2221	13	2012-08-02 11:24:38.986892
41940	4	1-9208	41937	25	76	708	75	2012-08-02 11:24:38.986892
41937	4	1-13128	41937	76	66	4016	16	2012-08-02 11:24:38.986892
41961	4	1-10792	41960	97	97	38	30	2012-08-02 11:24:41.318316
41962	4	1-13019	41960	97	39	35	75	2012-08-02 11:24:41.318316
41963	4	1-11736	41960	39	40	26	22	2012-08-02 11:24:41.318316
41960	4	1-13134	41960	40	97	64	41	2012-08-02 11:24:41.318316
41965	4	1-12018	41964	38	28	834	47	2012-08-02 11:24:41.318316
41966	4	1-12406	41964	28	37	894	34	2012-08-02 11:24:41.318316
41967	4	1-10595	41964	37	40	488	22	2012-08-02 11:24:41.318316
41964	4	1-13134	41964	40	38	1360	41	2012-08-02 11:24:41.318316
41921	4	1-13124	41921	8	62	1023	60	2012-08-02 11:24:37.075581
41926	4	1-7215	41925	47	30	429	31	2012-08-02 11:24:38.611274
41927	4	1-13090	41925	30	22	1683	70	2012-08-02 11:24:38.611274
41928	4	1-11998	41925	22	45	1097	84	2012-08-02 11:24:38.611274
41925	4	1-13127	41925	45	47	808	68	2012-08-02 11:24:38.611274
41930	4	1-12583	41929	33	20	3391	67	2012-08-02 11:24:38.611274
41931	4	1-11893	41929	20	62	4059	14	2012-08-02 11:24:38.611274
41932	4	1-12625	41929	62	45	2484	84	2012-08-02 11:24:38.611274
41929	4	1-13127	41929	45	33	2028	68	2012-08-02 11:24:38.611274
41934	4	1-13071	41933	58	61	5607	26	2012-08-02 11:24:38.611274
41935	4	1-12008	41933	61	86	6330	5	2012-08-02 11:24:38.611274
41936	4	1-12498	41933	86	45	3966	84	2012-08-02 11:24:38.611274
41933	4	1-13127	41933	45	58	3303	68	2012-08-02 11:24:38.611274
41957	4	1-12075	41956	10	47	54	55	2012-08-02 11:24:40.931417
41958	4	1-12204	41956	47	81	34	1	2012-08-02 11:24:40.931417
41959	4	1-11419	41956	81	31	32	50	2012-08-02 11:24:40.931417
41956	4	1-13133	41956	31	10	65	20	2012-08-02 11:24:40.931417
41981	4	1-1896	41980	83	94	897	40	2012-08-02 11:24:44.12585
41982	4	1-12968	41980	94	91	564	32	2012-08-02 11:24:44.12585
41983	4	1-9816	41980	91	67	106	30	2012-08-02 11:24:44.12585
41980	4	1-13140	41980	67	83	951	86	2012-08-02 11:24:44.12585
41985	4	1-12132	41984	91	34	671	3	2012-08-02 11:24:44.12585
41986	4	1-9617	41984	34	89	369	59	2012-08-02 11:24:44.12585
41987	4	1-11209	41984	89	67	286	30	2012-08-02 11:24:44.12585
41984	4	1-13140	41984	67	91	2768	86	2012-08-02 11:24:44.12585
41989	4	1-6724	41988	16	71	1629	58	2012-08-02 11:24:44.12585
41990	4	1-8255	41988	71	53	597	10	2012-08-02 11:24:44.12585
41991	4	1-12127	41988	53	67	530	30	2012-08-02 11:24:44.12585
41988	4	1-13140	41988	67	16	5334	86	2012-08-02 11:24:44.12585
41993	3	1-10936	41992	66	14	1345	25	2012-08-02 11:24:44.94585
41994	3	1-12959	41992	14	63	604	44	2012-08-02 11:24:44.94585
41992	3	1-13142	41992	63	66	2979	56	2012-08-02 11:24:44.94585
41996	3	1-2978	41995	100	6	1580	73	2012-08-02 11:24:44.94585
41997	3	1-13023	41995	6	63	927	44	2012-08-02 11:24:44.94585
41995	3	1-13142	41995	63	100	6632	56	2012-08-02 11:24:44.94585
41999	3	1-12547	41998	78	6	1200	23	2012-08-02 11:24:45.708282
42000	3	1-10941	41998	6	6	1088	86	2012-08-02 11:24:45.708282
41998	3	1-13145	41998	6	78	1088	90	2012-08-02 11:24:45.708282
42002	3	1-11503	42001	54	9	4492	60	2012-08-02 11:24:46.868161
42003	3	1-13068	42001	9	49	3820	87	2012-08-02 11:24:46.868161
42001	3	1-13148	42001	49	54	6013	53	2012-08-02 11:24:46.868161
42005	4	1-12970	42004	6	71	1208	38	2012-08-02 11:24:48.415244
42006	4	1-12038	42004	71	69	589	40	2012-08-02 11:24:48.415244
42007	4	1-12582	42004	69	97	516	45	2012-08-02 11:24:48.415244
42004	4	1-13152	42004	97	6	657	82	2012-08-02 11:24:48.415244
42009	4	1-12970	42008	6	30	3005	38	2012-08-02 11:24:48.415244
42010	4	1-9705	42008	30	76	961	96	2012-08-02 11:24:48.415244
42011	4	1-12131	42008	76	97	987	45	2012-08-02 11:24:48.415244
42008	4	1-13152	42008	97	6	1434	82	2012-08-02 11:24:48.415244
42013	4	1-12471	42012	71	59	1580	4	2012-08-02 11:24:48.415244
42014	4	1-5325	42012	59	46	657	34	2012-08-02 11:24:48.415244
42015	4	1-11991	42012	46	97	785	45	2012-08-02 11:24:48.415244
42012	4	1-13152	42012	97	71	1240	82	2012-08-02 11:24:48.415244
42017	3	1-11203	42016	6	49	4221	9	2012-08-02 11:24:48.823952
42018	3	1-13083	42016	49	53	2832	94	2012-08-02 11:24:48.823952
42016	3	1-13153	42016	53	6	4612	73	2012-08-02 11:24:48.823952
42020	4	1-12438	42019	69	47	880	63	2012-08-02 11:24:48.823952
42021	4	1-13150	42019	47	88	977	3	2012-08-02 11:24:48.823952
42022	4	1-10109	42019	88	53	617	94	2012-08-02 11:24:48.823952
42019	4	1-13153	42019	53	69	1098	73	2012-08-02 11:24:48.823952
42024	4	1-12924	42023	44	95	904	44	2012-08-02 11:24:49.983851
42025	4	1-6648	42023	95	36	130	79	2012-08-02 11:24:49.983851
42026	4	1-13112	42023	36	67	334	7	2012-08-02 11:24:49.983851
42023	4	1-13156	42023	67	44	589	98	2012-08-02 11:24:49.983851
42028	4	1-12379	42027	9	20	1372	1	2012-08-02 11:24:49.983851
42029	4	1-12226	42027	20	11	2143	38	2012-08-02 11:24:49.983851
42030	4	1-11480	42027	11	67	1312	7	2012-08-02 11:24:49.983851
42027	4	1-13156	42027	67	9	2423	98	2012-08-02 11:24:49.983851
42032	4	1-12943	42031	69	78	3426	90	2012-08-02 11:24:50.381471
42033	4	1-12547	42031	78	45	3691	23	2012-08-02 11:24:50.381471
42034	4	1-11425	42031	45	66	2589	69	2012-08-02 11:24:50.381471
42031	4	1-13157	42031	66	69	2846	24	2012-08-02 11:24:50.381471
42036	4	1-12943	42035	69	15	2857	90	2012-08-02 11:24:50.381471
42037	4	1-11924	42035	15	17	1864	29	2012-08-02 11:24:50.381471
42038	4	1-12847	42035	17	66	1792	69	2012-08-02 11:24:50.381471
42035	4	1-13157	42035	66	69	2162	24	2012-08-02 11:24:50.381471
42040	4	1-12839	42039	64	72	1785	26	2012-08-02 11:24:50.381471
42041	4	1-12003	42039	72	14	1261	33	2012-08-02 11:24:50.381471
42042	4	1-8354	42039	14	66	1074	69	2012-08-02 11:24:50.381471
42039	4	1-13157	42039	66	64	1372	24	2012-08-02 11:24:50.381471
42044	3	1-12580	42043	99	70	318	66	2012-08-02 11:24:50.381471
42045	3	1-11407	42043	70	66	262	69	2012-08-02 11:24:50.381471
42043	3	1-13157	42043	66	99	338	24	2012-08-02 11:24:50.381471
42047	4	1-12839	42046	64	72	899	26	2012-08-02 11:24:50.381471
42048	4	1-11962	42046	72	1	682	95	2012-08-02 11:24:50.381471
42049	4	1-10393	42046	1	66	523	69	2012-08-02 11:24:50.381471
42046	4	1-13157	42046	66	64	679	24	2012-08-02 11:24:50.381471
42051	3	1-10972	42050	45	83	2208	10	2012-08-02 11:24:51.575071
42052	3	1-12657	42050	83	68	3399	20	2012-08-02 11:24:51.575071
42050	3	1-13160	42050	68	45	7423	64	2012-08-02 11:24:51.575071
42054	4	1-12750	42053	37	34	2488	21	2012-08-02 11:24:51.961695
42055	4	1-12658	42053	34	35	1765	79	2012-08-02 11:24:51.961695
42056	4	1-10855	42053	35	15	1458	92	2012-08-02 11:24:51.961695
42053	4	1-13161	42053	15	37	1309	96	2012-08-02 11:24:51.961695
42058	4	1-10088	42057	74	77	1579	8	2012-08-02 11:24:51.961695
42059	4	1-11662	42057	77	95	1676	17	2012-08-02 11:24:51.961695
42060	4	1-12505	42057	95	15	2615	92	2012-08-02 11:24:51.961695
42057	4	1-13161	42057	15	74	2475	96	2012-08-02 11:24:51.961695
42062	3	1-12889	42061	62	82	3726	74	2012-08-02 11:24:54.347969
42063	3	1-12195	42061	82	11	2239	76	2012-08-02 11:24:54.347969
42061	3	1-13166	42061	11	62	7230	42	2012-08-02 11:24:54.347969
42065	4	1-12956	42064	64	66	241	29	2012-08-02 11:24:54.347969
42066	4	1-12083	42064	66	36	220	43	2012-08-02 11:24:54.347969
42067	4	1-11869	42064	36	11	130	76	2012-08-02 11:24:54.347969
42064	4	1-13166	42064	11	64	465	42	2012-08-02 11:24:54.347969
42069	4	1-10022	42068	90	80	983	21	2012-08-02 11:24:55.132544
42070	4	1-13143	42068	80	16	649	83	2012-08-02 11:24:55.132544
42071	4	1-11683	42068	16	8	307	99	2012-08-02 11:24:55.132544
42068	4	1-13169	42068	8	90	1555	71	2012-08-02 11:24:55.132544
42073	4	1-6392	42072	80	80	1108	38	2012-08-02 11:24:55.132544
42074	4	1-12236	42072	80	9	903	37	2012-08-02 11:24:55.132544
42075	4	1-12975	42072	9	8	374	99	2012-08-02 11:24:55.132544
42072	4	1-13169	42072	8	80	1963	71	2012-08-02 11:24:55.132544
42092	3	1-6195	42091	59	69	447	73	2012-08-02 11:24:56.292589
42093	3	1-12438	42091	69	54	115	63	2012-08-02 11:24:56.292589
42091	3	1-13172	42091	54	59	4103	33	2012-08-02 11:24:56.292589
42121	4	1-12657	42120	83	29	38	20	2012-08-02 11:25:01.441006
42122	4	1-8931	42120	29	88	13	59	2012-08-02 11:25:01.441006
42123	4	1-11972	42120	88	9	9	1	2012-08-02 11:25:01.441006
42120	4	1-13185	42120	9	83	17	10	2012-08-02 11:25:01.441006
42159	2	1-12304	42158	20	75	507	20	2012-08-02 11:25:06.92148
42158	2	1-13198	42158	75	20	4443	5	2012-08-02 11:25:06.92148
42161	4	1-12797	42160	76	32	2083	83	2012-08-02 11:25:06.92148
42162	4	1-9822	42160	32	31	271	50	2012-08-02 11:25:06.92148
42163	4	1-13133	42160	31	75	233	20	2012-08-02 11:25:06.92148
42160	4	1-13198	42160	75	76	2787	5	2012-08-02 11:25:06.92148
42173	3	1-12213	42172	31	30	1179	4	2012-08-02 11:25:07.717116
42174	3	1-12896	42172	30	22	1121	1	2012-08-02 11:25:07.717116
42172	3	1-13202	42172	22	31	1777	42	2012-08-02 11:25:07.717116
42253	4	1-13229	42252	16	44	5820	68	2012-08-02 11:25:19.091902
42254	4	1-11867	42252	44	97	4706	5	2012-08-02 11:25:19.091902
42255	4	1-13136	42252	97	36	3998	96	2012-08-02 11:25:19.091902
42252	4	1-13230	42252	36	16	5140	35	2012-08-02 11:25:19.091902
42257	4	1-13229	42256	16	80	830	68	2012-08-02 11:25:19.091902
42258	4	1-12285	42256	80	45	1119	60	2012-08-02 11:25:19.091902
42259	4	1-11694	42256	45	36	538	96	2012-08-02 11:25:19.091902
42256	4	1-13230	42256	36	16	712	35	2012-08-02 11:25:19.091902
42261	4	1-12990	42260	88	42	10	96	2012-08-02 11:25:20.254567
42262	4	1-12271	42260	42	7	5	8	2012-08-02 11:25:20.254567
42263	4	1-13092	42260	7	86	4	90	2012-08-02 11:25:20.254567
42260	4	1-13233	42260	86	88	6	78	2012-08-02 11:25:20.254567
42265	4	1-12673	42264	91	86	1900	5	2012-08-02 11:25:20.254567
42266	4	1-12498	42264	86	71	1138	84	2012-08-02 11:25:20.254567
42267	4	1-12999	42264	71	86	726	90	2012-08-02 11:25:20.254567
42264	4	1-13233	42264	86	91	1098	78	2012-08-02 11:25:20.254567
42273	4	1-9868	42272	10	94	263	76	2012-08-02 11:25:22.619491
42274	4	1-12144	42272	94	68	209	79	2012-08-02 11:25:22.619491
42275	4	1-12985	42272	68	10	196	63	2012-08-02 11:25:22.619491
42272	4	1-13239	42272	10	10	327	23	2012-08-02 11:25:22.619491
42304	4	1-11406	42303	73	46	1718	81	2012-08-02 11:25:26.164453
42305	4	1-12294	42303	46	39	765	25	2012-08-02 11:25:26.164453
42306	4	1-12697	42303	39	34	547	22	2012-08-02 11:25:26.164453
42303	4	1-13248	42303	34	73	4360	43	2012-08-02 11:25:26.164453
42308	3	1-10518	42307	96	39	294	25	2012-08-02 11:25:26.164453
42309	3	1-12697	42307	39	34	222	22	2012-08-02 11:25:26.164453
42307	3	1-13248	42307	34	96	1873	43	2012-08-02 11:25:26.164453
42310	1	1-6176	\N	89	89	819	22	2012-08-02 11:25:26.164453
42339	2	1-13252	42338	34	6	3543	29	2012-08-02 11:25:28.933653
42338	2	1-13255	42338	6	34	4643	47	2012-08-02 11:25:28.933653
42341	3	1-12545	42340	21	18	1476	79	2012-08-02 11:25:28.933653
42342	3	1-12523	42340	18	6	969	29	2012-08-02 11:25:28.933653
42340	3	1-13255	42340	6	21	1419	47	2012-08-02 11:25:28.933653
42344	4	1-11634	42343	9	25	325	10	2012-08-02 11:25:28.933653
42345	4	1-13000	42343	25	15	439	90	2012-08-02 11:25:28.933653
42346	4	1-11924	42343	15	6	309	29	2012-08-02 11:25:28.933653
42343	4	1-13255	42343	6	9	657	47	2012-08-02 11:25:28.933653
42355	4	1-12839	42354	64	45	974	26	2012-08-02 11:25:31.323161
42356	4	1-12365	42354	45	1	791	75	2012-08-02 11:25:31.323161
42357	4	1-12443	42354	1	11	544	58	2012-08-02 11:25:31.323161
42354	4	1-13261	42354	11	64	795	24	2012-08-02 11:25:31.323161
42381	3	1-6013	42380	51	77	2701	37	2012-08-02 11:25:38.40866
42382	3	1-12274	42380	77	47	503	66	2012-08-02 11:25:38.40866
42380	3	1-13279	42380	47	51	5563	45	2012-08-02 11:25:38.40866
42384	4	1-13022	42383	61	93	751	36	2012-08-02 11:25:38.40866
42385	4	1-12728	42383	93	62	430	6	2012-08-02 11:25:38.40866
42386	4	1-12671	42383	62	47	248	66	2012-08-02 11:25:38.40866
42383	4	1-13279	42383	47	61	3022	45	2012-08-02 11:25:38.40866
42388	4	1-13152	42387	97	99	1269	82	2012-08-02 11:25:38.40866
42389	4	1-2174	42387	99	66	303	16	2012-08-02 11:25:38.40866
42390	4	1-13104	42387	66	47	86	66	2012-08-02 11:25:38.40866
42387	4	1-13279	42387	47	97	1050	45	2012-08-02 11:25:38.40866
42396	4	1-13138	42395	63	3	5911	9	2012-08-02 11:25:41.128704
42397	4	1-13235	42395	3	89	7155	28	2012-08-02 11:25:41.128704
42398	4	1-11701	42395	89	28	5677	6	2012-08-02 11:25:41.128704
42395	4	1-13286	42395	28	63	5981	65	2012-08-02 11:25:41.128704
42469	3	1-13018	42468	11	1	1940	12	2012-08-02 11:25:51.610151
42470	3	1-13007	42468	1	45	4484	14	2012-08-02 11:25:51.610151
42468	3	1-13312	42468	45	11	3515	30	2012-08-02 11:25:51.610151
42472	4	1-12753	42471	60	38	1347	99	2012-08-02 11:25:51.610151
42473	4	1-12746	42471	38	68	2313	37	2012-08-02 11:25:51.610151
42474	4	1-12609	42471	68	45	2031	14	2012-08-02 11:25:51.610151
42471	4	1-13312	42471	45	60	1599	30	2012-08-02 11:25:51.610151
42480	2	1-9540	42479	56	60	755	35	2012-08-02 11:25:52.803265
42479	2	1-13314	42479	60	56	2415	47	2012-08-02 11:25:52.803265
42482	4	1-13090	42481	30	9	3124	70	2012-08-02 11:25:53.587928
42483	4	1-12747	42481	9	19	1204	16	2012-08-02 11:25:53.587928
42484	4	1-11928	42481	19	96	618	77	2012-08-02 11:25:53.587928
42481	4	1-13317	42481	96	30	1086	31	2012-08-02 11:25:53.587928
42486	4	1-13291	42485	78	50	593	97	2012-08-02 11:25:53.587928
42487	4	1-13285	42485	50	99	235	78	2012-08-02 11:25:53.587928
42488	4	1-12530	42485	99	96	474	77	2012-08-02 11:25:53.587928
42485	4	1-13317	42485	96	78	949	31	2012-08-02 11:25:53.587928
42490	4	1-13291	42489	78	45	4072	97	2012-08-02 11:25:53.587928
42491	4	1-9058	42489	45	15	2297	4	2012-08-02 11:25:53.587928
42492	4	1-12837	42489	15	96	2783	77	2012-08-02 11:25:53.587928
42489	4	1-13317	42489	96	78	6028	31	2012-08-02 11:25:53.587928
42077	3	1-2582	42076	18	23	1287	53	2012-08-02 11:24:55.920634
42078	3	1-12283	42076	23	9	876	51	2012-08-02 11:24:55.920634
42076	3	1-13170	42076	9	18	8548	85	2012-08-02 11:24:55.920634
42106	4	1-12522	42105	48	98	333	24	2012-08-02 11:25:00.269988
42107	4	1-11839	42105	98	82	365	72	2012-08-02 11:25:00.269988
42108	4	1-13171	42105	82	86	356	9	2012-08-02 11:25:00.269988
42105	4	1-13183	42105	86	48	337	3	2012-08-02 11:25:00.269988
42110	4	1-12961	42109	62	51	2207	66	2012-08-02 11:25:00.269988
42111	4	1-12958	42109	51	41	3485	58	2012-08-02 11:25:00.269988
42112	4	1-9807	42109	41	86	1318	9	2012-08-02 11:25:00.269988
42109	4	1-13183	42109	86	62	1344	3	2012-08-02 11:25:00.269988
42114	4	1-12961	42113	62	45	679	66	2012-08-02 11:25:00.269988
42115	4	1-9947	42113	45	93	612	44	2012-08-02 11:25:00.269988
42116	4	1-11956	42113	93	86	403	9	2012-08-02 11:25:00.269988
42113	4	1-13183	42113	86	62	412	3	2012-08-02 11:25:00.269988
42118	3	1-12955	42117	10	54	2347	87	2012-08-02 11:25:00.269988
42119	3	1-12496	42117	54	86	1123	9	2012-08-02 11:25:00.269988
42117	3	1-13183	42117	86	10	1240	3	2012-08-02 11:25:00.269988
42125	4	1-12573	42124	91	37	595	96	2012-08-02 11:25:01.838976
42126	4	1-12750	42124	37	24	749	21	2012-08-02 11:25:01.838976
42127	4	1-13077	42124	24	75	631	13	2012-08-02 11:25:01.838976
42124	4	1-13186	42124	75	91	2855	2	2012-08-02 11:25:01.838976
42129	2	1-9740	42128	21	71	1468	48	2012-08-02 11:25:03.021187
42128	2	1-13189	42128	71	21	2162	95	2012-08-02 11:25:03.021187
42131	4	1-10393	42130	1	46	498	69	2012-08-02 11:25:03.021187
42132	4	1-13038	42130	46	71	424	32	2012-08-02 11:25:03.021187
42133	4	1-12146	42130	71	71	383	48	2012-08-02 11:25:03.021187
42130	4	1-13189	42130	71	1	641	95	2012-08-02 11:25:03.021187
42135	4	1-11788	42134	30	46	2919	69	2012-08-02 11:25:03.021187
42136	4	1-13038	42134	46	71	2516	32	2012-08-02 11:25:03.021187
42137	4	1-12146	42134	71	71	2290	48	2012-08-02 11:25:03.021187
42134	4	1-13189	42134	71	30	3856	95	2012-08-02 11:25:03.021187
42165	4	1-12843	42164	41	24	1169	54	2012-08-02 11:25:07.330226
42166	4	1-9726	42164	24	63	317	1	2012-08-02 11:25:07.330226
42167	4	1-13146	42164	63	62	314	26	2012-08-02 11:25:07.330226
42164	4	1-13200	42164	62	41	309	18	2012-08-02 11:25:07.330226
42169	4	1-13154	42168	96	19	671	100	2012-08-02 11:25:07.330226
42170	4	1-12739	42168	19	64	692	24	2012-08-02 11:25:07.330226
42171	4	1-12839	42168	64	62	913	26	2012-08-02 11:25:07.330226
42168	4	1-13200	42168	62	96	944	18	2012-08-02 11:25:07.330226
42176	4	1-12471	42175	71	30	460	4	2012-08-02 11:25:08.490508
42177	4	1-12896	42175	30	30	390	1	2012-08-02 11:25:08.490508
42178	4	1-4742	42175	30	45	191	13	2012-08-02 11:25:08.490508
42175	4	1-13203	42175	45	71	537	82	2012-08-02 11:25:08.490508
42180	4	1-12471	42179	71	35	4768	4	2012-08-02 11:25:08.490508
42181	4	1-11513	42179	35	77	1151	66	2012-08-02 11:25:08.490508
42182	4	1-12800	42179	77	45	1514	13	2012-08-02 11:25:08.490508
42179	4	1-13203	42179	45	71	4870	82	2012-08-02 11:25:08.490508
42184	4	1-12181	42183	62	62	1048	60	2012-08-02 11:25:08.490508
42185	4	1-12103	42183	62	24	571	21	2012-08-02 11:25:08.490508
42186	4	1-13077	42183	24	45	593	13	2012-08-02 11:25:08.490508
42183	4	1-13203	42183	45	62	1949	82	2012-08-02 11:25:08.490508
42223	4	1-13089	42222	5	22	3582	28	2012-08-02 11:25:11.652325
42224	4	1-12342	42222	22	12	3261	53	2012-08-02 11:25:11.652325
42225	4	1-12787	42222	12	29	3456	46	2012-08-02 11:25:11.652325
42222	4	1-13211	42222	29	5	4528	33	2012-08-02 11:25:11.652325
42227	4	1-13089	42226	5	22	213	28	2012-08-02 11:25:11.652325
42228	4	1-12342	42226	22	19	203	53	2012-08-02 11:25:11.652325
42229	4	1-13181	42226	19	29	186	46	2012-08-02 11:25:11.652325
42226	4	1-13211	42226	29	5	256	33	2012-08-02 11:25:11.652325
42240	2	1-12409	42239	15	68	863	21	2012-08-02 11:25:16.724159
42239	2	1-13225	42239	68	15	4217	54	2012-08-02 11:25:16.724159
42269	4	1-13185	42268	9	25	2511	10	2012-08-02 11:25:21.813307
42270	4	1-13000	42268	25	90	1802	90	2012-08-02 11:25:21.813307
42271	4	1-12776	42268	90	15	617	43	2012-08-02 11:25:21.813307
42268	4	1-13237	42268	15	9	2238	1	2012-08-02 11:25:21.813307
42277	3	1-9716	42276	94	28	1251	80	2012-08-02 11:25:23.031427
42278	3	1-13061	42276	28	53	1158	96	2012-08-02 11:25:23.031427
42276	3	1-13240	42276	53	94	914	31	2012-08-02 11:25:23.031427
42280	4	1-12379	42279	9	20	3217	1	2012-08-02 11:25:24.204162
42281	4	1-12226	42279	20	11	3803	38	2012-08-02 11:25:24.204162
42282	4	1-11480	42279	11	67	1762	7	2012-08-02 11:25:24.204162
42279	4	1-13243	42279	67	9	7511	98	2012-08-02 11:25:24.204162
42284	4	1-12379	42283	9	84	598	1	2012-08-02 11:25:24.204162
42285	4	1-13135	42283	84	68	535	69	2012-08-02 11:25:24.204162
42286	4	1-12157	42283	68	67	325	7	2012-08-02 11:25:24.204162
42283	4	1-13243	42283	67	9	1391	98	2012-08-02 11:25:24.204162
42300	4	1-12797	42299	76	51	298	83	2012-08-02 11:25:25.377138
42301	4	1-7713	42299	51	45	133	85	2012-08-02 11:25:25.377138
42302	4	1-13178	42299	45	59	325	60	2012-08-02 11:25:25.377138
42299	4	1-13246	42299	59	76	177	5	2012-08-02 11:25:25.377138
42312	4	1-8845	42311	41	50	685	59	2012-08-02 11:25:26.565595
42313	4	1-12299	42311	50	29	669	53	2012-08-02 11:25:26.565595
42314	4	1-11392	42311	29	63	435	27	2012-08-02 11:25:26.565595
42311	4	1-13249	42311	63	41	1171	11	2012-08-02 11:25:26.565595
42316	4	1-11243	42315	81	45	1150	84	2012-08-02 11:25:26.565595
42317	4	1-13127	42315	45	98	911	68	2012-08-02 11:25:26.565595
42318	4	1-9888	42315	98	63	443	27	2012-08-02 11:25:26.565595
42315	4	1-13249	42315	63	81	1273	11	2012-08-02 11:25:26.565595
42320	4	1-7724	42319	56	36	196	29	2012-08-02 11:25:26.565595
42321	4	1-12528	42319	36	88	57	97	2012-08-02 11:25:26.565595
42322	4	1-12621	42319	88	63	65	27	2012-08-02 11:25:26.565595
42319	4	1-13249	42319	63	56	191	11	2012-08-02 11:25:26.565595
42328	4	1-12018	42327	38	9	2574	47	2012-08-02 11:25:28.546934
42329	4	1-11634	42327	9	25	1268	10	2012-08-02 11:25:28.546934
42330	4	1-13000	42327	25	34	1709	90	2012-08-02 11:25:28.546934
42327	4	1-13253	42327	34	38	3582	41	2012-08-02 11:25:28.546934
42359	4	1-9713	42358	89	58	716	76	2012-08-02 11:25:32.107626
42360	4	1-12331	42358	58	74	777	50	2012-08-02 11:25:32.107626
42361	4	1-12888	42358	74	95	1455	38	2012-08-02 11:25:32.107626
42358	4	1-13263	42358	95	89	1015	94	2012-08-02 11:25:32.107626
42363	4	1-12833	42362	53	11	3826	89	2012-08-02 11:25:32.107626
42364	4	1-11339	42362	11	70	3837	81	2012-08-02 11:25:32.107626
42365	4	1-13242	42362	70	95	6021	38	2012-08-02 11:25:32.107626
42080	4	1-12933	42079	73	22	1607	28	2012-08-02 11:24:55.53012
42081	4	1-12342	42079	22	85	1104	53	2012-08-02 11:24:55.53012
42082	4	1-12877	42079	85	82	1908	72	2012-08-02 11:24:55.53012
42079	4	1-13171	42079	82	73	1550	9	2012-08-02 11:24:55.53012
42084	4	1-12933	42083	73	37	4793	28	2012-08-02 11:24:55.53012
42085	4	1-13064	42083	37	33	6037	62	2012-08-02 11:24:55.53012
42086	4	1-12777	42083	33	82	5466	72	2012-08-02 11:24:55.53012
42083	4	1-13171	42083	82	73	4531	9	2012-08-02 11:24:55.53012
42088	4	1-6664	42087	89	80	2064	53	2012-08-02 11:24:55.53012
42089	4	1-12722	42087	80	59	2229	80	2012-08-02 11:24:55.53012
42090	4	1-12759	42087	59	82	2456	72	2012-08-02 11:24:55.53012
42087	4	1-13171	42087	82	89	2612	9	2012-08-02 11:24:55.53012
42095	4	1-13032	42094	55	19	3	43	2012-08-02 11:24:58.314903
42096	4	1-10503	42094	19	91	2	50	2012-08-02 11:24:58.314903
42097	4	1-12418	42094	91	45	2	85	2012-08-02 11:24:58.314903
42094	4	1-13178	42094	45	55	5	60	2012-08-02 11:24:58.314903
42099	4	1-13032	42098	55	36	2326	43	2012-08-02 11:24:58.314903
42100	4	1-11869	42098	36	46	1460	76	2012-08-02 11:24:58.314903
42101	4	1-11735	42098	46	45	1599	85	2012-08-02 11:24:58.314903
42098	4	1-13178	42098	45	55	4154	60	2012-08-02 11:24:58.314903
42103	3	1-12989	42102	67	33	1393	90	2012-08-02 11:24:58.314903
42104	3	1-7676	42102	33	45	620	85	2012-08-02 11:24:58.314903
42102	3	1-13178	42102	45	67	1636	60	2012-08-02 11:24:58.314903
42139	4	1-11861	42138	93	59	4727	80	2012-08-02 11:25:04.93251
42140	4	1-12759	42138	59	77	4345	72	2012-08-02 11:25:04.93251
42141	4	1-11728	42138	77	53	1404	22	2012-08-02 11:25:04.93251
42138	4	1-13194	42138	53	93	3331	58	2012-08-02 11:25:04.93251
42143	4	1-12791	42142	86	29	1480	85	2012-08-02 11:25:04.93251
42144	4	1-10206	42142	29	39	235	25	2012-08-02 11:25:04.93251
42145	4	1-12697	42142	39	53	255	22	2012-08-02 11:25:04.93251
42142	4	1-13194	42142	53	86	694	58	2012-08-02 11:25:04.93251
42147	4	1-11394	42146	32	31	605	79	2012-08-02 11:25:04.93251
42148	4	1-10218	42146	31	39	239	75	2012-08-02 11:25:04.93251
42149	4	1-11736	42146	39	53	228	22	2012-08-02 11:25:04.93251
42146	4	1-13194	42146	53	32	641	58	2012-08-02 11:25:04.93251
42151	4	1-8905	42150	81	37	1021	28	2012-08-02 11:25:04.93251
42152	4	1-13064	42150	37	25	1648	62	2012-08-02 11:25:04.93251
42153	4	1-12730	42150	25	53	1423	22	2012-08-02 11:25:04.93251
42150	4	1-13194	42150	53	81	4042	58	2012-08-02 11:25:04.93251
42155	4	1-12696	42154	26	84	245	35	2012-08-02 11:25:04.93251
42156	4	1-13182	42154	84	36	187	39	2012-08-02 11:25:04.93251
42157	4	1-13097	42154	36	53	301	22	2012-08-02 11:25:04.93251
42154	4	1-13194	42154	53	26	866	58	2012-08-02 11:25:04.93251
42188	4	1-12222	42187	47	63	567	39	2012-08-02 11:25:08.866093
42189	4	1-12751	42187	63	99	801	78	2012-08-02 11:25:08.866093
42190	4	1-12530	42187	99	22	2164	77	2012-08-02 11:25:08.866093
42187	4	1-13204	42187	22	47	761	59	2012-08-02 11:25:08.866093
42192	4	1-12837	42191	15	22	5368	77	2012-08-02 11:25:10.048654
42193	4	1-13204	42191	22	22	1764	59	2012-08-02 11:25:10.048654
42194	4	1-13006	42191	22	22	2309	89	2012-08-02 11:25:10.048654
42191	4	1-13207	42191	22	15	4586	4	2012-08-02 11:25:10.048654
42196	4	1-11513	42195	35	77	110	66	2012-08-02 11:25:10.048654
42197	4	1-12800	42195	77	92	187	13	2012-08-02 11:25:10.048654
42198	4	1-13120	42195	92	22	166	89	2012-08-02 11:25:10.048654
42195	4	1-13207	42195	22	35	353	4	2012-08-02 11:25:10.048654
42200	4	1-12923	42199	67	82	54	31	2012-08-02 11:25:10.048654
42201	4	1-11282	42199	82	90	83	43	2012-08-02 11:25:10.048654
42202	4	1-11371	42199	90	22	54	89	2012-08-02 11:25:10.048654
42199	4	1-13207	42199	22	67	115	4	2012-08-02 11:25:10.048654
42204	4	1-13133	42203	31	7	1758	20	2012-08-02 11:25:11.265631
42205	4	1-9029	42203	7	46	746	34	2012-08-02 11:25:11.265631
42206	4	1-11991	42203	46	80	648	45	2012-08-02 11:25:11.265631
42203	4	1-13209	42203	80	31	1099	50	2012-08-02 11:25:11.265631
42208	3	1-12418	42207	91	72	3300	85	2012-08-02 11:25:11.265631
42209	3	1-11362	42207	72	80	2089	45	2012-08-02 11:25:11.265631
42207	3	1-13209	42207	80	91	3653	50	2012-08-02 11:25:11.265631
42211	4	1-13133	42210	31	65	4095	20	2012-08-02 11:25:11.265631
42212	4	1-10985	42210	65	4	1322	65	2012-08-02 11:25:11.265631
42213	4	1-11975	42210	4	80	1336	45	2012-08-02 11:25:11.265631
42210	4	1-13209	42210	80	31	2409	50	2012-08-02 11:25:11.265631
42215	4	1-12888	42214	74	30	4175	38	2012-08-02 11:25:11.265631
42216	4	1-9705	42214	30	76	1249	96	2012-08-02 11:25:11.265631
42217	4	1-12131	42214	76	80	1199	45	2012-08-02 11:25:11.265631
42214	4	1-13209	42214	80	74	2403	50	2012-08-02 11:25:11.265631
42219	4	1-12888	42218	74	41	125	38	2012-08-02 11:25:11.265631
42220	4	1-10759	42218	41	76	31	96	2012-08-02 11:25:11.265631
42221	4	1-12131	42218	76	80	32	45	2012-08-02 11:25:11.265631
42218	4	1-13209	42218	80	74	68	50	2012-08-02 11:25:11.265631
42231	4	1-13182	42230	84	86	2681	39	2012-08-02 11:25:12.050696
42232	4	1-12920	42230	86	9	1972	19	2012-08-02 11:25:12.050696
42233	4	1-12763	42230	9	37	2033	72	2012-08-02 11:25:12.050696
42230	4	1-13213	42230	37	84	5300	35	2012-08-02 11:25:12.050696
42235	4	1-12410	42234	21	45	797	18	2012-08-02 11:25:12.050696
42236	4	1-12608	42234	45	3	705	8	2012-08-02 11:25:12.050696
42237	4	1-11776	42234	3	37	845	72	2012-08-02 11:25:12.050696
42234	4	1-13213	42234	37	21	2398	35	2012-08-02 11:25:12.050696
42238	1	1-2712	\N	37	37	1171	8	2012-08-02 11:25:12.050696
42242	3	1-11394	42241	32	37	3132	79	2012-08-02 11:25:17.553175
42243	3	1-13005	42241	37	35	1336	57	2012-08-02 11:25:17.553175
42241	3	1-13226	42241	35	32	6147	58	2012-08-02 11:25:17.553175
42245	4	1-11861	42244	93	42	67	80	2012-08-02 11:25:17.553175
42246	4	1-1598	42244	42	30	9	42	2012-08-02 11:25:17.553175
42247	4	1-13004	42244	30	35	8	57	2012-08-02 11:25:17.553175
42244	4	1-13226	42244	35	93	51	58	2012-08-02 11:25:17.553175
42249	4	1-8711	42248	69	21	764	47	2012-08-02 11:25:17.553175
42250	4	1-12545	42248	21	37	858	79	2012-08-02 11:25:17.553175
42251	4	1-13005	42248	37	35	530	57	2012-08-02 11:25:17.553175
42248	4	1-13226	42248	35	69	3537	58	2012-08-02 11:25:17.553175
42288	4	1-13008	42287	76	26	3868	87	2012-08-02 11:25:24.602023
42289	4	1-6769	42287	26	46	368	91	2012-08-02 11:25:24.602023
42290	4	1-13113	42287	46	13	223	11	2012-08-02 11:25:24.602023
42287	4	1-13244	42287	13	76	884	63	2012-08-02 11:25:24.602023
42292	4	1-13150	42291	47	34	1700	3	2012-08-02 11:25:24.602023
42293	4	1-9617	42291	34	34	771	59	2012-08-02 11:25:24.602023
42294	4	1-12183	42291	34	13	553	11	2012-08-02 11:25:24.602023
42291	4	1-13244	42291	13	47	2229	63	2012-08-02 11:25:24.602023
42296	4	1-13009	42295	4	10	3486	76	2012-08-02 11:25:24.602023
42297	4	1-12652	42295	10	87	3974	55	2012-08-02 11:25:24.602023
42298	4	1-13158	42295	87	13	1245	11	2012-08-02 11:25:24.602023
42295	4	1-13244	42295	13	4	5033	63	2012-08-02 11:25:24.602023
42324	4	1-12979	42323	95	17	562	29	2012-08-02 11:25:26.950328
42325	4	1-12847	42323	17	69	601	69	2012-08-02 11:25:26.950328
42326	4	1-10805	42323	69	97	329	58	2012-08-02 11:25:26.950328
42323	4	1-13250	42323	97	95	1095	54	2012-08-02 11:25:26.950328
42332	3	1-11849	42331	24	8	283	6	2012-08-02 11:25:28.160499
42333	3	1-12716	42331	8	37	231	97	2012-08-02 11:25:28.160499
42331	3	1-13254	42331	37	24	504	21	2012-08-02 11:25:28.160499
42335	4	1-13077	42334	24	92	3270	13	2012-08-02 11:25:28.160499
42336	4	1-13120	42334	92	67	2779	89	2012-08-02 11:25:28.160499
42337	4	1-11518	42334	67	37	1162	97	2012-08-02 11:25:28.160499
42334	4	1-13254	42334	37	24	2595	21	2012-08-02 11:25:28.160499
42348	4	1-10772	42347	61	77	1545	9	2012-08-02 11:25:29.719316
42349	4	1-2602	42347	77	24	322	73	2012-08-02 11:25:29.719316
42350	4	1-12592	42347	24	39	342	19	2012-08-02 11:25:29.719316
42347	4	1-13257	42347	39	61	3956	99	2012-08-02 11:25:29.719316
42352	3	1-10772	42351	61	2	1856	9	2012-08-02 11:25:29.719316
42353	3	1-7045	42351	2	39	367	19	2012-08-02 11:25:29.719316
42351	3	1-13257	42351	39	61	4491	99	2012-08-02 11:25:29.719316
42367	3	1-12080	42366	12	9	2374	1	2012-08-02 11:25:34.086727
42368	3	1-13185	42366	9	93	4334	10	2012-08-02 11:25:34.086727
42366	3	1-13268	42366	93	12	4117	12	2012-08-02 11:25:34.086727
42370	4	1-12614	42369	49	4	4124	52	2012-08-02 11:25:36.461597
42371	4	1-13260	42369	4	66	1865	81	2012-08-02 11:25:36.461597
42372	4	1-13167	42369	66	47	892	67	2012-08-02 11:25:36.461597
42369	4	1-13274	42369	47	49	1630	89	2012-08-02 11:25:36.461597
42374	3	1-11339	42373	11	66	1823	81	2012-08-02 11:25:36.461597
42375	3	1-13167	42373	66	47	981	67	2012-08-02 11:25:36.461597
42373	3	1-13274	42373	47	11	2018	89	2012-08-02 11:25:36.461597
42377	4	1-11518	42376	67	37	972	97	2012-08-02 11:25:36.461597
42378	4	1-13254	42376	37	19	2281	21	2012-08-02 11:25:36.461597
42379	4	1-12927	42376	19	47	971	67	2012-08-02 11:25:36.461597
42376	4	1-13274	42376	47	67	2210	89	2012-08-02 11:25:36.461597
42392	4	1-12636	42391	9	38	5095	47	2012-08-02 11:25:39.159937
42393	4	1-9344	42391	38	97	909	58	2012-08-02 11:25:39.159937
42394	4	1-13250	42391	97	64	1919	54	2012-08-02 11:25:39.159937
42391	4	1-13280	42391	64	9	8655	99	2012-08-02 11:25:39.159937
42412	4	1-12080	42411	12	9	1019	1	2012-08-02 11:25:41.957032
42413	4	1-13185	42411	9	34	1541	10	2012-08-02 11:25:41.957032
42414	4	1-11430	42411	34	26	1285	45	2012-08-02 11:25:41.957032
42411	4	1-13288	42411	26	12	2133	12	2012-08-02 11:25:41.957032
42416	4	1-13007	42415	1	99	505	14	2012-08-02 11:25:41.957032
42417	4	1-12071	42415	99	22	200	13	2012-08-02 11:25:41.957032
42418	4	1-12319	42415	22	26	152	45	2012-08-02 11:25:41.957032
42415	4	1-13288	42415	26	1	275	12	2012-08-02 11:25:41.957032
42420	4	1-12462	42419	1	69	321	64	2012-08-02 11:25:41.957032
42421	4	1-13258	42419	69	49	249	75	2012-08-02 11:25:41.957032
42422	4	1-8702	42419	49	26	187	45	2012-08-02 11:25:41.957032
42419	4	1-13288	42419	26	1	347	12	2012-08-02 11:25:41.957032
42424	4	1-12462	42423	1	11	2554	64	2012-08-02 11:25:41.957032
42425	4	1-11349	42423	11	45	2252	80	2012-08-02 11:25:41.957032
42426	4	1-13025	42423	45	26	1346	45	2012-08-02 11:25:41.957032
42423	4	1-13288	42423	26	1	2628	12	2012-08-02 11:25:41.957032
42428	4	1-12922	42427	32	11	2034	33	2012-08-02 11:25:41.957032
42429	4	1-7495	42427	11	63	1784	16	2012-08-02 11:25:41.957032
42430	4	1-12674	42427	63	26	1503	45	2012-08-02 11:25:41.957032
42427	4	1-13288	42427	26	32	3043	12	2012-08-02 11:25:41.957032
42431	1	1-8098	\N	23	23	115	13	2012-08-02 11:25:41.957032
42444	4	1-6531	42443	86	24	1	73	2012-08-02 11:25:43.994978
42445	4	1-12592	42443	24	39	1	19	2012-08-02 11:25:43.994978
42446	4	1-13257	42443	39	57	4	99	2012-08-02 11:25:43.994978
42443	4	1-13293	42443	57	86	3	61	2012-08-02 11:25:43.994978
42448	2	1-11650	42447	100	57	4876	99	2012-08-02 11:25:43.994978
42447	2	1-13293	42447	57	100	5151	61	2012-08-02 11:25:43.994978
42458	3	1-11901	42457	64	97	2406	58	2012-08-02 11:25:47.632607
42459	3	1-13250	42457	97	81	3646	54	2012-08-02 11:25:47.632607
42457	3	1-13302	42457	81	64	7520	74	2012-08-02 11:25:47.632607
42461	4	1-11694	42460	45	42	965	96	2012-08-02 11:25:50.836867
42462	4	1-12271	42460	42	7	527	8	2012-08-02 11:25:50.836867
42463	4	1-13092	42460	7	67	481	90	2012-08-02 11:25:50.836867
42460	4	1-13311	42460	67	45	1658	60	2012-08-02 11:25:50.836867
42465	4	1-13246	42464	59	86	1385	5	2012-08-02 11:25:50.836867
42466	4	1-12498	42464	86	71	863	84	2012-08-02 11:25:50.836867
42467	4	1-12999	42464	71	67	609	90	2012-08-02 11:25:50.836867
42464	4	1-13311	42464	67	59	2256	60	2012-08-02 11:25:50.836867
42476	4	1-8705	42475	55	97	1102	58	2012-08-02 11:25:52.405473
42477	4	1-13250	42475	97	15	3106	54	2012-08-02 11:25:52.405473
42478	4	1-12409	42475	15	94	2316	21	2012-08-02 11:25:52.405473
42475	4	1-13313	42475	94	55	2434	23	2012-08-02 11:25:52.405473
42502	4	1-9593	42501	27	61	1297	53	2012-08-02 11:25:56.417239
42503	4	1-12910	42501	61	65	1197	13	2012-08-02 11:25:56.417239
42504	4	1-12267	42501	65	4	1386	77	2012-08-02 11:25:56.417239
42501	4	1-13323	42501	4	27	2170	34	2012-08-02 11:25:56.417239
42556	4	1-13158	42555	87	56	1710	11	2012-08-02 11:26:05.444433
42557	4	1-7724	42555	56	38	1503	29	2012-08-02 11:26:05.444433
42558	4	1-13116	42555	38	84	3200	17	2012-08-02 11:26:05.444433
42555	4	1-13346	42555	84	87	4184	55	2012-08-02 11:26:05.444433
42562	3	1-13346	42561	84	48	2400	55	2012-08-02 11:26:06.593594
42563	3	1-9499	42561	48	70	1408	16	2012-08-02 11:26:06.593594
42561	3	1-13350	42561	70	84	2723	17	2012-08-02 11:26:06.593594
42565	4	1-13346	42564	84	87	943	55	2012-08-02 11:26:06.593594
42566	4	1-13158	42564	87	39	334	11	2012-08-02 11:26:06.593594
42567	4	1-12836	42564	39	70	335	16	2012-08-02 11:26:06.593594
42564	4	1-13350	42564	70	84	832	17	2012-08-02 11:26:06.593594
42577	4	1-12299	42576	50	23	554	53	2012-08-02 11:26:08.173921
42578	4	1-12283	42576	23	94	413	51	2012-08-02 11:26:08.173921
42579	4	1-13347	42576	94	65	440	100	2012-08-02 11:26:08.173921
42576	4	1-13353	42576	65	50	503	59	2012-08-02 11:26:08.173921
42620	4	1-12592	42619	24	8	1656	19	2012-08-02 11:26:11.325063
42619	4	1-13361	42619	51	24	2520	73	2012-08-02 11:26:11.325063
42362	4	1-13263	42362	95	53	4552	94	2012-08-02 11:25:32.107626
42400	4	1-13270	42399	56	62	975	70	2012-08-02 11:25:41.526325
42401	4	1-13052	42399	62	55	384	42	2012-08-02 11:25:41.526325
42402	4	1-12429	42399	55	30	200	11	2012-08-02 11:25:41.526325
42399	4	1-13287	42399	30	56	281	23	2012-08-02 11:25:41.526325
42404	4	1-13270	42403	56	62	4742	70	2012-08-02 11:25:41.526325
42405	4	1-13052	42403	62	1	2061	42	2012-08-02 11:25:41.526325
42406	4	1-11780	42403	1	30	800	11	2012-08-02 11:25:41.526325
42403	4	1-13287	42403	30	56	1239	23	2012-08-02 11:25:41.526325
42408	4	1-8075	42407	73	17	1200	48	2012-08-02 11:25:41.526325
42409	4	1-13088	42407	17	46	1064	91	2012-08-02 11:25:41.526325
42410	4	1-13113	42407	46	30	975	11	2012-08-02 11:25:41.526325
42407	4	1-13287	42407	30	73	1704	23	2012-08-02 11:25:41.526325
42433	3	1-12922	42432	32	78	302	33	2012-08-02 11:25:42.77523
42434	3	1-6032	42432	78	11	131	92	2012-08-02 11:25:42.77523
42432	3	1-13292	42432	11	32	877	12	2012-08-02 11:25:42.77523
42436	4	1-12462	42435	1	82	1785	64	2012-08-02 11:25:42.77523
42437	4	1-12663	42435	82	83	1230	55	2012-08-02 11:25:42.77523
42438	4	1-11784	42435	83	11	483	92	2012-08-02 11:25:42.77523
42435	4	1-13292	42435	11	1	3335	12	2012-08-02 11:25:42.77523
42440	4	1-12080	42439	12	63	841	1	2012-08-02 11:25:42.77523
42441	4	1-13146	42439	63	20	507	26	2012-08-02 11:25:42.77523
42442	4	1-11011	42439	20	11	257	92	2012-08-02 11:25:42.77523
42439	4	1-13292	42439	11	12	2195	12	2012-08-02 11:25:42.77523
42450	4	1-9807	42449	41	22	541	9	2012-08-02 11:25:44.780263
42451	4	1-13002	42449	22	95	864	38	2012-08-02 11:25:44.780263
42452	4	1-13263	42449	95	26	582	94	2012-08-02 11:25:44.780263
42449	4	1-13295	42449	26	41	1495	58	2012-08-02 11:25:44.780263
42454	4	1-11861	42453	93	81	436	80	2012-08-02 11:25:44.780263
42455	4	1-13271	42453	81	57	248	89	2012-08-02 11:25:44.780263
42456	4	1-6859	42453	57	26	96	94	2012-08-02 11:25:44.780263
42453	4	1-13295	42453	26	93	259	58	2012-08-02 11:25:44.780263
42498	4	1-12620	42497	20	10	5800	20	2012-08-02 11:25:53.985582
42499	4	1-12075	42497	10	83	5189	55	2012-08-02 11:25:53.985582
42500	4	1-11784	42497	83	45	4407	92	2012-08-02 11:25:53.985582
42497	4	1-13319	42497	45	20	6144	99	2012-08-02 11:25:53.985582
42510	4	1-10542	42509	53	53	2855	43	2012-08-02 11:25:56.803419
42511	4	1-9287	42509	53	6	1200	100	2012-08-02 11:25:56.803419
42512	4	1-12900	42509	6	42	772	56	2012-08-02 11:25:56.803419
42509	4	1-13325	42509	42	53	7716	4	2012-08-02 11:25:56.803419
42514	4	1-12923	42513	67	26	124	31	2012-08-02 11:25:56.803419
42515	4	1-13199	42513	26	26	96	10	2012-08-02 11:25:56.803419
42516	4	1-11426	42513	26	42	34	56	2012-08-02 11:25:56.803419
42513	4	1-13325	42513	42	67	439	4	2012-08-02 11:25:56.803419
42534	4	1-10956	42533	32	52	2586	72	2012-08-02 11:26:01.079355
42535	4	1-13193	42533	52	39	1155	25	2012-08-02 11:26:01.079355
42536	4	1-12380	42533	39	40	477	62	2012-08-02 11:26:01.079355
42533	4	1-13335	42533	40	32	5839	71	2012-08-02 11:26:01.079355
42538	4	1-9018	42537	80	43	489	81	2012-08-02 11:26:01.079355
42539	4	1-13216	42537	43	32	847	93	2012-08-02 11:26:01.079355
42540	4	1-13118	42537	32	40	242	62	2012-08-02 11:26:01.079355
42537	4	1-13335	42537	40	80	2983	71	2012-08-02 11:26:01.079355
42560	2	1-11623	42559	13	33	1247	98	2012-08-02 11:26:06.969062
42559	2	1-13349	42559	33	13	9137	78	2012-08-02 11:26:06.969062
42569	4	1-12861	42568	31	56	1148	26	2012-08-02 11:26:07.754192
42570	4	1-12134	42568	56	36	771	79	2012-08-02 11:26:07.754192
42571	4	1-13112	42568	36	59	1746	7	2012-08-02 11:26:07.754192
42568	4	1-13352	42568	59	31	5382	37	2012-08-02 11:26:07.754192
42573	4	1-11832	42572	60	68	1754	81	2012-08-02 11:26:07.754192
42574	4	1-11594	42572	68	19	1586	40	2012-08-02 11:26:07.754192
42575	4	1-13047	42572	19	59	1164	7	2012-08-02 11:26:07.754192
42572	4	1-13352	42572	59	60	3762	37	2012-08-02 11:26:07.754192
42595	4	1-13286	42594	28	64	2455	65	2012-08-02 11:26:08.958394
42596	4	1-12642	42594	64	19	1479	53	2012-08-02 11:26:08.958394
42597	4	1-13181	42594	19	41	444	46	2012-08-02 11:26:08.958394
42594	4	1-13355	42594	41	28	6707	6	2012-08-02 11:26:08.958394
42643	3	1-6887	42642	93	76	69	5	2012-08-02 11:26:12.950056
42644	3	1-12797	42642	76	64	82	83	2012-08-02 11:26:12.950056
42642	3	1-13365	42642	64	93	165	43	2012-08-02 11:26:12.950056
42646	4	1-13340	42645	24	68	2663	63	2012-08-02 11:26:12.950056
42647	4	1-11630	42645	68	3	2045	53	2012-08-02 11:26:12.950056
42648	4	1-12643	42645	3	64	2183	83	2012-08-02 11:26:12.950056
42645	4	1-13365	42645	64	24	4693	43	2012-08-02 11:26:12.950056
42670	4	1-12631	42669	74	86	3740	46	2012-08-02 11:26:14.165465
42671	4	1-11525	42669	86	26	1418	94	2012-08-02 11:26:14.165465
42672	4	1-13295	42669	26	2	3164	58	2012-08-02 11:26:14.165465
42669	4	1-13369	42669	2	74	2813	51	2012-08-02 11:26:14.165465
42704	2	1-5542	42703	86	56	333	10	2012-08-02 11:26:19.468651
42703	2	1-13381	42703	56	86	849	61	2012-08-02 11:26:19.468651
42734	2	1-11714	42733	87	80	4326	25	2012-08-02 11:26:21.446797
42733	2	1-13387	42733	80	87	5659	36	2012-08-02 11:26:21.446797
42736	4	1-12632	42735	81	15	343	92	2012-08-02 11:26:21.446797
42737	4	1-13161	42735	15	88	253	96	2012-08-02 11:26:21.446797
42738	4	1-9476	42735	88	80	139	25	2012-08-02 11:26:21.446797
42735	4	1-13387	42735	80	81	209	36	2012-08-02 11:26:21.446797
42740	4	1-10484	42739	38	30	1446	95	2012-08-02 11:26:21.446797
42741	4	1-13326	42739	30	79	3253	32	2012-08-02 11:26:21.446797
42742	4	1-11793	42739	79	80	897	25	2012-08-02 11:26:21.446797
42739	4	1-13387	42739	80	38	1490	36	2012-08-02 11:26:21.446797
42791	3	1-3888	42790	41	37	1252	51	2012-08-02 11:26:26.596176
42792	3	1-12760	42790	37	89	2980	83	2012-08-02 11:26:26.596176
42790	3	1-13399	42790	89	41	2606	24	2012-08-02 11:26:26.596176
42794	4	1-11839	42793	98	32	2510	72	2012-08-02 11:26:26.596176
42795	4	1-13041	42793	32	45	2321	70	2012-08-02 11:26:26.596176
42796	4	1-13341	42793	45	89	2083	83	2012-08-02 11:26:26.596176
42793	4	1-13399	42793	89	98	2050	24	2012-08-02 11:26:26.596176
42798	4	1-7249	42797	84	27	1585	80	2012-08-02 11:26:26.596176
42799	4	1-12803	42797	27	39	1392	4	2012-08-02 11:26:26.596176
42800	4	1-12238	42797	39	89	1800	83	2012-08-02 11:26:26.596176
42797	4	1-13399	42797	89	84	1815	24	2012-08-02 11:26:26.596176
42806	4	1-12740	42805	77	40	4864	91	2012-08-02 11:26:27.778526
42807	4	1-13236	42805	40	94	5161	76	2012-08-02 11:26:27.778526
42805	4	1-13402	42805	35	77	4143	8	2012-08-02 11:26:27.778526
42494	4	1-12323	42493	77	11	1051	33	2012-08-02 11:25:53.587928
42495	4	1-7495	42493	11	19	917	16	2012-08-02 11:25:53.587928
42496	4	1-11928	42493	19	96	618	77	2012-08-02 11:25:53.587928
42493	4	1-13317	42493	96	77	1428	31	2012-08-02 11:25:53.587928
42506	4	1-13293	42505	57	28	1712	61	2012-08-02 11:25:55.986268
42507	4	1-5483	42505	28	71	1103	48	2012-08-02 11:25:55.986268
42508	4	1-13189	42505	71	31	1857	95	2012-08-02 11:25:55.986268
42505	4	1-13324	42505	31	57	1444	99	2012-08-02 11:25:55.986268
42518	4	1-9287	42517	53	25	2181	100	2012-08-02 11:26:00.681608
42519	4	1-12454	42517	25	25	1534	27	2012-08-02 11:26:00.681608
42520	4	1-11159	42517	25	98	1066	13	2012-08-02 11:26:00.681608
42517	4	1-13334	42517	98	53	6233	43	2012-08-02 11:26:00.681608
42522	4	1-11869	42521	36	10	522	76	2012-08-02 11:26:00.681608
42523	4	1-12652	42521	10	98	436	55	2012-08-02 11:26:00.681608
42524	4	1-11498	42521	98	98	234	13	2012-08-02 11:26:00.681608
42521	4	1-13334	42521	98	36	1664	43	2012-08-02 11:26:00.681608
42526	4	1-11371	42525	90	5	407	89	2012-08-02 11:26:00.681608
42527	4	1-12334	42525	5	77	102	66	2012-08-02 11:26:00.681608
42528	4	1-12800	42525	77	98	122	13	2012-08-02 11:26:00.681608
42525	4	1-13334	42525	98	90	881	43	2012-08-02 11:26:00.681608
42530	4	1-11869	42529	36	58	29	76	2012-08-02 11:26:00.681608
42531	4	1-12331	42529	58	53	25	50	2012-08-02 11:26:00.681608
42532	4	1-13026	42529	53	98	8	13	2012-08-02 11:26:00.681608
42529	4	1-13334	42529	98	36	58	43	2012-08-02 11:26:00.681608
42542	4	1-13129	42541	38	88	1286	3	2012-08-02 11:26:04.262141
42543	4	1-10109	42541	88	26	393	94	2012-08-02 11:26:04.262141
42544	4	1-13295	42541	26	46	507	58	2012-08-02 11:26:04.262141
42541	4	1-13342	42541	46	38	3142	25	2012-08-02 11:26:04.262141
42546	4	1-13129	42545	38	49	1497	3	2012-08-02 11:26:04.262141
42547	4	1-8393	42545	49	26	280	94	2012-08-02 11:26:04.262141
42548	4	1-13295	42545	26	46	425	58	2012-08-02 11:26:04.262141
42545	4	1-13342	42545	46	38	3105	25	2012-08-02 11:26:04.262141
42550	4	1-12697	42549	39	9	1243	22	2012-08-02 11:26:04.262141
42551	4	1-8273	42549	9	20	392	10	2012-08-02 11:26:04.262141
42552	4	1-12237	42549	20	46	162	58	2012-08-02 11:26:04.262141
42549	4	1-13342	42549	46	39	1522	25	2012-08-02 11:26:04.262141
42554	2	1-10330	42553	1	11	331	41	2012-08-02 11:26:05.046846
42553	2	1-13345	42553	11	1	2778	19	2012-08-02 11:26:05.046846
42581	3	1-13241	42580	73	72	5319	82	2012-08-02 11:26:08.571845
42582	3	1-6161	42580	72	5	2110	70	2012-08-02 11:26:08.571845
42580	3	1-13354	42580	5	73	1645	26	2012-08-02 11:26:08.571845
42584	3	1-13188	42583	80	32	4592	87	2012-08-02 11:26:08.571845
42585	3	1-11460	42583	32	5	1253	70	2012-08-02 11:26:08.571845
42583	3	1-13354	42583	5	80	1149	26	2012-08-02 11:26:08.571845
42587	4	1-12134	42586	56	68	2723	79	2012-08-02 11:26:08.571845
42588	4	1-12985	42586	68	50	2492	63	2012-08-02 11:26:08.571845
42589	4	1-12349	42586	50	5	3843	70	2012-08-02 11:26:08.571845
42586	4	1-13354	42586	5	56	3533	26	2012-08-02 11:26:08.571845
42591	4	1-13200	42590	62	80	3465	18	2012-08-02 11:26:08.571845
42592	4	1-13091	42590	80	77	4508	43	2012-08-02 11:26:08.571845
42593	4	1-7752	42590	77	5	3643	70	2012-08-02 11:26:08.571845
42590	4	1-13354	42590	5	62	3409	26	2012-08-02 11:26:08.571845
42599	4	1-12818	42598	43	97	3079	17	2012-08-02 11:26:10.938883
42600	4	1-7257	42598	97	53	943	10	2012-08-02 11:26:10.938883
42601	4	1-12127	42598	53	81	680	30	2012-08-02 11:26:10.938883
42598	4	1-13360	42598	81	43	2966	49	2012-08-02 11:26:10.938883
42603	4	1-12818	42602	43	94	1057	17	2012-08-02 11:26:10.938883
42604	4	1-12059	42602	94	23	473	23	2012-08-02 11:26:10.938883
42605	4	1-13351	42602	23	81	232	30	2012-08-02 11:26:10.938883
42602	4	1-13360	42602	81	43	1014	49	2012-08-02 11:26:10.938883
42607	4	1-7675	42606	5	35	1726	92	2012-08-02 11:26:10.938883
42608	4	1-10680	42606	35	89	559	59	2012-08-02 11:26:10.938883
42609	4	1-11209	42606	89	81	417	30	2012-08-02 11:26:10.938883
42606	4	1-13360	42606	81	5	2080	49	2012-08-02 11:26:10.938883
42611	4	1-4883	42610	57	10	160	63	2012-08-02 11:26:10.938883
42612	4	1-13239	42610	10	23	267	23	2012-08-02 11:26:10.938883
42613	4	1-13351	42610	23	81	156	30	2012-08-02 11:26:10.938883
42610	4	1-13360	42610	81	57	815	49	2012-08-02 11:26:10.938883
42615	4	1-11752	42614	45	81	1890	3	2012-08-02 11:26:10.938883
42616	4	1-9791	42614	81	23	958	23	2012-08-02 11:26:10.938883
42617	4	1-13351	42614	23	81	558	30	2012-08-02 11:26:10.938883
42614	4	1-13360	42614	81	45	2958	49	2012-08-02 11:26:10.938883
42618	1	1-7251	\N	59	59	1778	49	2012-08-02 11:26:10.938883
42635	4	1-12074	42634	99	10	277	76	2012-08-02 11:26:12.132263
42636	4	1-12652	42634	10	87	475	55	2012-08-02 11:26:12.132263
42637	4	1-13158	42634	87	58	224	11	2012-08-02 11:26:12.132263
42634	4	1-13363	42634	58	99	453	69	2012-08-02 11:26:12.132263
42650	4	1-11716	42649	91	69	3744	64	2012-08-02 11:26:13.358762
42651	4	1-13258	42649	69	39	2972	75	2012-08-02 11:26:13.358762
42652	4	1-11736	42649	39	73	2439	22	2012-08-02 11:26:13.358762
42649	4	1-13366	42649	73	91	5304	3	2012-08-02 11:26:13.358762
42654	4	1-12590	42653	53	27	1743	20	2012-08-02 11:26:13.358762
42655	4	1-10726	42653	27	45	464	28	2012-08-02 11:26:13.358762
42656	4	1-11746	42653	45	73	478	22	2012-08-02 11:26:13.358762
42653	4	1-13366	42653	73	53	1125	3	2012-08-02 11:26:13.358762
42658	4	1-8864	42657	21	53	772	91	2012-08-02 11:26:13.358762
42659	4	1-11417	42657	53	39	519	25	2012-08-02 11:26:13.358762
42660	4	1-12697	42657	39	73	553	22	2012-08-02 11:26:13.358762
42657	4	1-13366	42657	73	21	1329	3	2012-08-02 11:26:13.358762
42662	4	1-12522	42661	48	98	479	24	2012-08-02 11:26:13.358762
42663	4	1-11839	42661	98	77	539	72	2012-08-02 11:26:13.358762
42664	4	1-11728	42661	77	73	196	22	2012-08-02 11:26:13.358762
42661	4	1-13366	42661	73	48	472	3	2012-08-02 11:26:13.358762
42666	4	1-8913	42665	24	86	423	39	2012-08-02 11:26:13.358762
42667	4	1-12920	42665	86	74	438	19	2012-08-02 11:26:13.358762
42668	4	1-11535	42665	74	73	680	22	2012-08-02 11:26:13.358762
42665	4	1-13366	42665	73	24	1647	3	2012-08-02 11:26:13.358762
42674	2	1-9702	42673	26	38	2280	22	2012-08-02 11:26:15.336549
42673	2	1-13371	42673	38	26	4032	11	2012-08-02 11:26:15.336549
42676	4	1-10322	42675	60	4	249	77	2012-08-02 11:26:15.336549
42677	4	1-13323	42675	4	37	351	34	2012-08-02 11:26:15.336549
42678	4	1-10595	42675	37	38	187	22	2012-08-02 11:26:15.336549
42675	4	1-13371	42675	38	60	332	11	2012-08-02 11:26:15.336549
42621	4	1-11317	42619	8	8	712	6	2012-08-02 11:26:11.325063
42622	4	1-12716	42619	8	51	417	97	2012-08-02 11:26:11.325063
42624	3	1-12592	42623	24	58	1286	19	2012-08-02 11:26:11.325063
42625	3	1-6776	42623	58	51	266	97	2012-08-02 11:26:11.325063
42623	3	1-13361	42623	51	24	1773	73	2012-08-02 11:26:11.325063
42627	4	1-12438	42626	69	4	1376	63	2012-08-02 11:26:11.325063
42628	4	1-13009	42626	4	66	966	76	2012-08-02 11:26:11.325063
42629	4	1-10945	42626	66	51	276	97	2012-08-02 11:26:11.325063
42626	4	1-13361	42626	51	69	2459	73	2012-08-02 11:26:11.325063
42631	4	1-9569	42630	98	47	25	68	2012-08-02 11:26:11.325063
42632	4	1-7215	42630	47	78	11	31	2012-08-02 11:26:11.325063
42633	4	1-13291	42630	78	51	7	97	2012-08-02 11:26:11.325063
42630	4	1-13361	42630	51	98	68	73	2012-08-02 11:26:11.325063
42639	2	1-13311	42638	67	83	2331	60	2012-08-02 11:26:12.563318
42638	2	1-13364	42638	83	67	5897	90	2012-08-02 11:26:12.563318
42641	2	1-8072	42640	49	83	508	60	2012-08-02 11:26:12.563318
42640	2	1-13364	42640	83	49	2987	90	2012-08-02 11:26:12.563318
42692	4	1-12671	42691	62	77	541	66	2012-08-02 11:26:17.480661
42693	4	1-12800	42691	77	65	936	13	2012-08-02 11:26:17.480661
42694	4	1-12267	42691	65	49	1213	77	2012-08-02 11:26:17.480661
42691	4	1-13376	42691	49	62	705	6	2012-08-02 11:26:17.480661
42696	4	1-12031	42695	96	27	23	52	2012-08-02 11:26:18.69543
42697	4	1-7650	42695	27	59	18	2	2012-08-02 11:26:18.69543
42698	4	1-12363	42695	59	40	20	15	2012-08-02 11:26:18.69543
42695	4	1-13380	42695	40	96	41	5	2012-08-02 11:26:18.69543
42700	4	1-12808	42699	35	14	1608	38	2012-08-02 11:26:18.69543
42701	4	1-5071	42699	14	10	752	92	2012-08-02 11:26:18.69543
42702	4	1-12987	42699	10	40	2052	15	2012-08-02 11:26:18.69543
42699	4	1-13380	42699	40	35	4359	5	2012-08-02 11:26:18.69543
42713	3	1-12585	42712	59	67	3577	38	2012-08-02 11:26:20.242499
42714	3	1-9296	42712	67	8	2126	2	2012-08-02 11:26:20.242499
42712	3	1-13384	42712	8	59	3540	61	2012-08-02 11:26:20.242499
42716	3	1-11570	42715	75	24	627	18	2012-08-02 11:26:20.242499
42717	3	1-11427	42715	24	8	1641	2	2012-08-02 11:26:20.242499
42715	3	1-13384	42715	8	75	2988	61	2012-08-02 11:26:20.242499
42719	3	1-8298	42718	56	9	9	100	2012-08-02 11:26:20.242499
42720	3	1-12441	42718	9	8	12	2	2012-08-02 11:26:20.242499
42718	3	1-13384	42718	8	56	23	61	2012-08-02 11:26:20.242499
42722	4	1-13200	42721	62	80	579	18	2012-08-02 11:26:21.833694
42723	4	1-13091	42721	80	77	619	43	2012-08-02 11:26:21.833694
42724	4	1-7752	42721	77	88	411	70	2012-08-02 11:26:21.833694
42721	4	1-13386	42721	88	62	693	26	2012-08-02 11:26:21.833694
42726	4	1-12003	42725	72	35	2665	33	2012-08-02 11:26:21.833694
42727	4	1-13294	42725	35	50	1923	63	2012-08-02 11:26:21.833694
42728	4	1-12349	42725	50	88	2562	70	2012-08-02 11:26:21.833694
42725	4	1-13386	42725	88	72	4460	26	2012-08-02 11:26:21.833694
42730	4	1-12365	42729	45	85	2055	75	2012-08-02 11:26:21.833694
42731	4	1-11207	42729	85	32	1945	72	2012-08-02 11:26:21.833694
42732	4	1-13041	42729	32	88	1547	70	2012-08-02 11:26:21.833694
42729	4	1-13386	42729	88	45	2756	26	2012-08-02 11:26:21.833694
42752	3	1-12400	42751	44	51	1785	47	2012-08-02 11:26:22.651417
42753	3	1-10355	42751	51	46	599	6	2012-08-02 11:26:22.651417
42751	3	1-13389	42751	46	44	4165	25	2012-08-02 11:26:22.651417
42755	4	1-12697	42754	39	20	1361	22	2012-08-02 11:26:22.651417
42756	4	1-5200	42754	20	23	368	50	2012-08-02 11:26:22.651417
42757	4	1-12449	42754	23	46	182	6	2012-08-02 11:26:22.651417
42754	4	1-13389	42754	46	39	1620	25	2012-08-02 11:26:22.651417
42759	4	1-12380	42758	39	61	172	62	2012-08-02 11:26:22.651417
42760	4	1-12798	42758	61	95	99	57	2012-08-02 11:26:22.651417
42761	4	1-10468	42758	95	46	45	6	2012-08-02 11:26:22.651417
42758	4	1-13389	42758	46	39	421	25	2012-08-02 11:26:22.651417
42763	2	1-1867	42762	98	46	106	6	2012-08-02 11:26:22.651417
42762	2	1-13389	42762	46	98	1017	25	2012-08-02 11:26:22.651417
42765	4	1-13240	42764	53	77	1678	31	2012-08-02 11:26:23.856512
42766	4	1-12323	42764	77	63	1022	33	2012-08-02 11:26:23.856512
42767	4	1-12289	42764	63	82	704	1	2012-08-02 11:26:23.856512
42764	4	1-13392	42764	82	53	2863	96	2012-08-02 11:26:23.856512
42776	4	1-10793	42775	100	24	1355	43	2012-08-02 11:26:25.822572
42777	4	1-13340	42775	24	80	1116	63	2012-08-02 11:26:25.822572
42778	4	1-10929	42775	80	14	2167	88	2012-08-02 11:26:25.822572
42775	4	1-13397	42775	14	100	2420	75	2012-08-02 11:26:25.822572
42780	4	1-11458	42779	50	86	2051	90	2012-08-02 11:26:25.822572
42781	4	1-13233	42779	86	98	3200	78	2012-08-02 11:26:25.822572
42782	4	1-6553	42779	98	14	3045	88	2012-08-02 11:26:25.822572
42779	4	1-13397	42779	14	50	3500	75	2012-08-02 11:26:25.822572
42810	4	1-12271	42809	42	7	845	8	2012-08-02 11:26:28.176291
42811	4	1-13092	42809	7	86	797	90	2012-08-02 11:26:28.176291
42812	4	1-13233	42809	86	41	1206	78	2012-08-02 11:26:28.176291
42809	4	1-13403	42809	41	42	1497	96	2012-08-02 11:26:28.176291
42831	3	1-13326	42830	30	22	214	32	2012-08-02 11:26:32.564429
42832	3	1-8724	42830	22	2	74	59	2012-08-02 11:26:32.564429
42830	3	1-13414	42830	2	30	92	95	2012-08-02 11:26:32.564429
42845	4	1-13061	42844	28	37	2504	96	2012-08-02 11:26:37.271162
42846	4	1-12750	42844	37	53	2486	21	2012-08-02 11:26:37.271162
42847	4	1-8882	42844	53	84	703	77	2012-08-02 11:26:37.271162
42844	4	1-13426	42844	84	28	5116	80	2012-08-02 11:26:37.271162
42876	4	1-11956	42875	93	87	2394	9	2012-08-02 11:26:43.204062
42877	4	1-12465	42875	87	80	1434	6	2012-08-02 11:26:43.204062
42878	4	1-11941	42875	80	10	1598	77	2012-08-02 11:26:43.204062
42875	4	1-13440	42875	10	93	3423	44	2012-08-02 11:26:43.204062
42880	4	1-11921	42879	54	87	624	55	2012-08-02 11:26:43.204062
42881	4	1-13158	42879	87	60	306	11	2012-08-02 11:26:43.204062
42882	4	1-10322	42879	60	10	282	77	2012-08-02 11:26:43.204062
42879	4	1-13440	42879	10	54	605	44	2012-08-02 11:26:43.204062
42883	1	1-5105	\N	91	91	629	11	2012-08-02 11:26:43.204062
42891	4	1-12474	42890	56	22	235	1	2012-08-02 11:26:44.023142
42892	4	1-13202	42890	22	70	314	42	2012-08-02 11:26:44.023142
42893	4	1-12912	42890	70	20	306	72	2012-08-02 11:26:44.023142
42890	4	1-13443	42890	20	56	938	22	2012-08-02 11:26:44.023142
42895	4	1-13374	42894	64	35	505	79	2012-08-02 11:26:44.023142
42896	4	1-13402	42894	35	3	569	8	2012-08-02 11:26:44.023142
42897	4	1-11776	42894	3	20	831	72	2012-08-02 11:26:44.023142
42894	4	1-13443	42894	20	64	2755	22	2012-08-02 11:26:44.023142
42680	4	1-7724	42679	56	58	860	29	2012-08-02 11:26:15.336549
42681	4	1-13067	42679	58	25	608	62	2012-08-02 11:26:15.336549
42682	4	1-12730	42679	25	38	487	22	2012-08-02 11:26:15.336549
42679	4	1-13371	42679	38	56	930	11	2012-08-02 11:26:15.336549
42684	3	1-13287	42683	30	89	2675	23	2012-08-02 11:26:15.336549
42685	3	1-6123	42683	89	38	789	22	2012-08-02 11:26:15.336549
42683	3	1-13371	42683	38	30	1580	11	2012-08-02 11:26:15.336549
42687	4	1-13321	42686	48	3	647	9	2012-08-02 11:26:15.336549
42688	4	1-13235	42686	3	45	760	28	2012-08-02 11:26:15.336549
42689	4	1-11746	42686	45	38	825	22	2012-08-02 11:26:15.336549
42686	4	1-13371	42686	38	48	1655	11	2012-08-02 11:26:15.336549
42690	1	1-3105	\N	38	38	225	23	2012-08-02 11:26:15.336549
42706	4	1-9953	42705	99	18	2337	79	2012-08-02 11:26:19.855415
42707	4	1-12523	42705	18	66	1878	29	2012-08-02 11:26:19.855415
42708	4	1-12083	42705	66	95	1609	43	2012-08-02 11:26:19.855415
42705	4	1-13382	42705	95	99	2980	30	2012-08-02 11:26:19.855415
42710	3	1-13019	42709	97	100	4573	75	2012-08-02 11:26:19.855415
42711	3	1-10793	42709	100	95	2387	43	2012-08-02 11:26:19.855415
42709	3	1-13382	42709	95	97	4763	30	2012-08-02 11:26:19.855415
42744	4	1-10885	42743	57	3	1071	74	2012-08-02 11:26:22.231701
42745	4	1-9754	42743	3	25	846	27	2012-08-02 11:26:22.231701
42746	4	1-11159	42743	25	6	681	13	2012-08-02 11:26:22.231701
42743	4	1-13388	42743	6	57	2371	4	2012-08-02 11:26:22.231701
42748	4	1-10344	42747	12	98	2448	42	2012-08-02 11:26:22.231701
42749	4	1-12343	42747	98	25	1134	27	2012-08-02 11:26:22.231701
42750	4	1-11159	42747	25	6	950	13	2012-08-02 11:26:22.231701
42747	4	1-13388	42747	6	12	3440	4	2012-08-02 11:26:22.231701
42769	3	1-9119	42768	86	4	1321	12	2012-08-02 11:26:24.64072
42770	3	1-13175	42768	4	82	2980	25	2012-08-02 11:26:24.64072
42768	3	1-13395	42768	82	86	3476	57	2012-08-02 11:26:24.64072
42772	4	1-13306	42771	15	15	742	92	2012-08-02 11:26:24.64072
42773	4	1-13161	42771	15	88	696	96	2012-08-02 11:26:24.64072
42774	4	1-9476	42771	88	82	483	25	2012-08-02 11:26:24.64072
42771	4	1-13395	42771	82	15	569	57	2012-08-02 11:26:24.64072
42784	3	1-10806	42783	94	94	927	21	2012-08-02 11:26:26.20944
42785	3	1-13313	42783	94	79	998	23	2012-08-02 11:26:26.20944
42783	3	1-13398	42783	79	94	761	100	2012-08-02 11:26:26.20944
42787	4	1-12336	42786	44	43	271	81	2012-08-02 11:26:26.20944
42788	4	1-13216	42786	43	9	603	93	2012-08-02 11:26:26.20944
42789	4	1-10826	42786	9	79	122	23	2012-08-02 11:26:26.20944
42786	4	1-13398	42786	79	44	119	100	2012-08-02 11:26:26.20944
42802	4	1-13071	42801	58	20	1294	26	2012-08-02 11:26:27.391834
42803	4	1-11011	42801	20	10	858	92	2012-08-02 11:26:27.391834
42804	4	1-12987	42801	10	91	2044	15	2012-08-02 11:26:27.391834
42801	4	1-13401	42801	91	58	1033	68	2012-08-02 11:26:27.391834
42814	3	1-12289	42813	63	53	259	1	2012-08-02 11:26:28.56418
42815	3	1-12841	42813	53	38	56	65	2012-08-02 11:26:28.56418
42813	3	1-13404	42813	38	63	557	33	2012-08-02 11:26:28.56418
42817	4	1-12715	42816	15	72	2014	49	2012-08-02 11:26:28.56418
42818	4	1-12393	42816	72	28	819	6	2012-08-02 11:26:28.56418
42819	4	1-13286	42816	28	38	445	65	2012-08-02 11:26:28.56418
42816	4	1-13404	42816	38	15	4581	33	2012-08-02 11:26:28.56418
42849	3	1-8527	42848	32	58	1462	68	2012-08-02 11:26:38.055142
42850	3	1-13071	42848	58	19	1759	26	2012-08-02 11:26:38.055142
42848	3	1-13428	42848	19	32	4131	31	2012-08-02 11:26:38.055142
42852	4	1-13199	42851	26	20	4000	10	2012-08-02 11:26:38.055142
42853	4	1-12237	42851	20	80	2163	58	2012-08-02 11:26:38.055142
42854	4	1-13227	42851	80	19	1532	26	2012-08-02 11:26:38.055142
42851	4	1-13428	42851	19	26	4526	31	2012-08-02 11:26:38.055142
42855	1	1-2178	\N	2	2	815	31	2012-08-02 11:26:38.055142
42857	4	1-12900	42856	6	66	1347	56	2012-08-02 11:26:40.03285
42858	4	1-10936	42856	66	39	1348	25	2012-08-02 11:26:40.03285
42859	4	1-12697	42856	39	38	1297	22	2012-08-02 11:26:40.03285
42856	4	1-13433	42856	38	6	1192	100	2012-08-02 11:26:40.03285
42861	4	1-13353	42860	65	47	4416	59	2012-08-02 11:26:40.03285
42862	4	1-12222	42860	47	36	2929	39	2012-08-02 11:26:40.03285
42863	4	1-13097	42860	36	38	4343	22	2012-08-02 11:26:40.03285
42860	4	1-13433	42860	38	65	4394	100	2012-08-02 11:26:40.03285
42869	3	1-13353	42868	65	34	2528	59	2012-08-02 11:26:41.226091
42870	3	1-12183	42868	34	22	2241	11	2012-08-02 11:26:41.226091
42868	3	1-13436	42868	22	65	2238	100	2012-08-02 11:26:41.226091
42872	4	1-11785	42871	7	67	1977	47	2012-08-02 11:26:41.226091
42873	4	1-13331	42871	67	46	2109	91	2012-08-02 11:26:41.226091
42874	4	1-13113	42871	46	22	2055	11	2012-08-02 11:26:41.226091
42871	4	1-13436	42871	22	7	2082	100	2012-08-02 11:26:41.226091
42887	4	1-12449	42886	23	28	499	6	2012-08-02 11:26:43.601757
42888	4	1-13286	42886	28	63	500	65	2012-08-02 11:26:43.601757
42889	4	1-13138	42886	63	93	470	9	2012-08-02 11:26:43.601757
42886	4	1-13442	42886	93	23	781	50	2012-08-02 11:26:43.601757
42906	2	1-12349	42905	50	25	1035	70	2012-08-02 11:26:47.220183
42905	2	1-13450	42905	25	50	1324	63	2012-08-02 11:26:47.220183
42908	3	1-13008	42907	76	32	2827	87	2012-08-02 11:26:47.220183
42909	3	1-11460	42907	32	25	467	70	2012-08-02 11:26:47.220183
42907	3	1-13450	42907	25	76	711	63	2012-08-02 11:26:47.220183
42911	4	1-8813	42910	65	72	2191	52	2012-08-02 11:26:47.220183
42912	4	1-13221	42910	72	32	1819	72	2012-08-02 11:26:47.220183
42913	4	1-13041	42910	32	25	1325	70	2012-08-02 11:26:47.220183
42910	4	1-13450	42910	25	65	2709	63	2012-08-02 11:26:47.220183
42915	3	1-13020	42914	11	72	993	98	2012-08-02 11:26:47.220183
42916	3	1-12586	42914	72	25	280	70	2012-08-02 11:26:47.220183
42914	3	1-13450	42914	25	11	573	63	2012-08-02 11:26:47.220183
42918	4	1-11630	42917	68	73	2921	53	2012-08-02 11:26:47.220183
42919	4	1-13427	42917	73	31	3122	89	2012-08-02 11:26:47.220183
42920	4	1-8402	42917	31	25	1537	70	2012-08-02 11:26:47.220183
42917	4	1-13450	42917	25	68	3214	63	2012-08-02 11:26:47.220183
42922	4	1-13253	42921	34	33	704	41	2012-08-02 11:26:48.013555
42923	4	1-4364	42921	33	19	284	53	2012-08-02 11:26:48.013555
42924	4	1-13181	42921	19	68	249	46	2012-08-02 11:26:48.013555
42921	4	1-13453	42921	68	34	342	90	2012-08-02 11:26:48.013555
42951	4	1-13307	42950	21	18	276	19	2012-08-02 11:26:54.810076
42952	4	1-12398	42950	18	44	353	79	2012-08-02 11:26:54.810076
42953	4	1-13451	42950	44	63	239	34	2012-08-02 11:26:54.810076
42950	4	1-13470	42950	63	21	419	65	2012-08-02 11:26:54.810076
42955	4	1-13307	42954	21	85	3816	19	2012-08-02 11:26:54.810076
42808	4	1-12144	42805	94	35	3633	79	2012-08-02 11:26:27.778526
42821	4	1-9095	42820	71	14	1854	88	2012-08-02 11:26:28.960842
42822	4	1-13397	42820	14	33	1674	75	2012-08-02 11:26:28.960842
42823	4	1-7646	42820	33	42	1986	2	2012-08-02 11:26:28.960842
42820	4	1-13405	42820	42	71	2098	74	2012-08-02 11:26:28.960842
42825	2	1-10950	42824	18	42	2397	2	2012-08-02 11:26:28.960842
42824	2	1-13405	42824	42	18	2718	74	2012-08-02 11:26:28.960842
42827	4	1-13214	42826	10	71	1393	71	2012-08-02 11:26:28.960842
42828	4	1-11824	42826	71	27	1160	52	2012-08-02 11:26:28.960842
42829	4	1-7650	42826	27	42	1040	2	2012-08-02 11:26:28.960842
42826	4	1-13405	42826	42	10	1377	74	2012-08-02 11:26:28.960842
42834	3	1-13274	42833	47	74	2024	89	2012-08-02 11:26:34.100074
42835	3	1-10349	42833	74	50	1858	72	2012-08-02 11:26:34.100074
42833	3	1-13418	42833	50	47	973	67	2012-08-02 11:26:34.100074
42837	4	1-11480	42836	11	83	718	7	2012-08-02 11:26:36.46422
42838	4	1-11603	42836	83	10	419	20	2012-08-02 11:26:36.46422
42839	4	1-12075	42836	10	9	374	55	2012-08-02 11:26:36.46422
42836	4	1-13424	42836	9	11	1156	38	2012-08-02 11:26:36.46422
42841	4	1-9210	42840	29	87	125	28	2012-08-02 11:26:36.46422
42842	4	1-10930	42840	87	84	115	17	2012-08-02 11:26:36.46422
42843	4	1-13346	42840	84	9	181	55	2012-08-02 11:26:36.46422
42840	4	1-13424	42840	9	29	570	38	2012-08-02 11:26:36.46422
42865	4	1-9558	42864	36	28	3520	80	2012-08-02 11:26:40.430765
42866	4	1-13061	42864	28	53	2550	96	2012-08-02 11:26:40.430765
42867	4	1-13240	42864	53	78	1575	31	2012-08-02 11:26:40.430765
42864	4	1-13434	42864	78	36	7633	76	2012-08-02 11:26:40.430765
42885	2	1-10322	42884	60	98	5211	77	2012-08-02 11:26:42.816986
42884	2	1-13441	42884	98	60	8927	11	2012-08-02 11:26:42.816986
42903	3	1-13235	42902	3	27	1928	28	2012-08-02 11:26:45.648723
42904	3	1-11987	42902	27	61	1819	100	2012-08-02 11:26:45.648723
42902	3	1-13447	42902	61	3	1504	9	2012-08-02 11:26:45.648723
42947	4	1-12678	42946	85	60	6566	37	2012-08-02 11:26:52.469722
42948	4	1-11832	42946	60	68	3376	81	2012-08-02 11:26:52.469722
42949	4	1-11594	42946	68	44	3367	40	2012-08-02 11:26:52.469722
42946	4	1-13465	42946	44	85	7619	15	2012-08-02 11:26:52.469722
42959	3	1-13424	42958	9	80	3080	38	2012-08-02 11:26:55.196257
42960	3	1-12236	42958	80	48	1576	37	2012-08-02 11:26:55.196257
42958	3	1-13471	42958	48	9	1641	55	2012-08-02 11:26:55.196257
42962	4	1-13424	42961	9	14	3289	38	2012-08-02 11:26:55.196257
42963	4	1-5071	42961	14	19	1198	92	2012-08-02 11:26:55.196257
42964	4	1-12916	42961	19	48	1392	37	2012-08-02 11:26:55.196257
42961	4	1-13471	42961	48	9	1594	55	2012-08-02 11:26:55.196257
42966	3	1-11784	42965	83	19	289	92	2012-08-02 11:26:55.196257
42967	3	1-12916	42965	19	48	369	37	2012-08-02 11:26:55.196257
42965	3	1-13471	42965	48	83	464	55	2012-08-02 11:26:55.196257
42969	4	1-12827	42968	84	58	2008	21	2012-08-02 11:26:55.196257
42970	4	1-13029	42968	58	94	1854	48	2012-08-02 11:26:55.196257
42971	4	1-12783	42968	94	48	3850	37	2012-08-02 11:26:55.196257
42968	4	1-13471	42968	48	84	5905	55	2012-08-02 11:26:55.196257
42972	1	1-9272	\N	48	48	887	38	2012-08-02 11:26:55.196257
42999	4	1-11203	42998	6	86	542	9	2012-08-02 11:26:57.671379
43000	4	1-13183	42998	86	10	491	3	2012-08-02 11:26:57.671379
43001	4	1-12955	42998	10	63	762	87	2012-08-02 11:26:57.671379
42998	4	1-13477	42998	63	6	605	73	2012-08-02 11:26:57.671379
43003	4	1-12438	43002	69	47	1765	63	2012-08-02 11:26:57.671379
43004	4	1-13150	43002	47	10	1814	3	2012-08-02 11:26:57.671379
43005	4	1-12955	43002	10	63	2906	87	2012-08-02 11:26:57.671379
43002	4	1-13477	43002	63	69	2382	73	2012-08-02 11:26:57.671379
43007	4	1-12768	43006	36	80	968	71	2012-08-02 11:26:57.671379
43008	4	1-6392	43006	80	21	511	38	2012-08-02 11:26:57.671379
43009	4	1-10863	43006	21	63	354	87	2012-08-02 11:26:57.671379
43006	4	1-13477	43006	63	36	302	73	2012-08-02 11:26:57.671379
43011	3	1-12944	43010	83	21	3398	38	2012-08-02 11:26:57.671379
43012	3	1-10863	43010	21	63	2368	87	2012-08-02 11:26:57.671379
43010	3	1-13477	43010	63	83	2030	73	2012-08-02 11:26:57.671379
43014	4	1-12438	43013	69	61	392	63	2012-08-02 11:26:57.671379
43015	4	1-12913	43013	61	42	448	56	2012-08-02 11:26:57.671379
43016	4	1-9873	43013	42	63	485	87	2012-08-02 11:26:57.671379
43013	4	1-13477	43013	63	69	459	73	2012-08-02 11:26:57.671379
43022	4	1-10030	43021	54	40	578	91	2012-08-02 11:26:58.843133
43023	4	1-13236	43021	40	10	663	76	2012-08-02 11:26:58.843133
43024	4	1-12652	43021	10	80	1091	55	2012-08-02 11:26:58.843133
43021	4	1-13480	43021	80	54	1587	72	2012-08-02 11:26:58.843133
43026	4	1-11759	43025	52	54	1854	2	2012-08-02 11:26:58.843133
43027	4	1-13144	43025	54	82	983	64	2012-08-02 11:26:58.843133
43028	4	1-12663	43025	82	80	1419	55	2012-08-02 11:26:58.843133
43025	4	1-13480	43025	80	52	2135	72	2012-08-02 11:26:58.843133
43054	3	1-11824	43053	71	4	68	52	2012-08-02 11:27:02.776472
43055	3	1-13260	43053	4	97	26	81	2012-08-02 11:27:02.776472
43053	3	1-13490	43053	97	71	118	71	2012-08-02 11:27:02.776472
43057	4	1-10022	43056	90	94	1557	21	2012-08-02 11:27:02.776472
43058	4	1-13313	43056	94	97	1105	23	2012-08-02 11:27:02.776472
43059	4	1-12749	43056	97	97	709	81	2012-08-02 11:27:02.776472
43056	4	1-13490	43056	97	90	3669	71	2012-08-02 11:27:02.776472
43061	2	1-9018	43060	80	97	582	81	2012-08-02 11:27:02.776472
43060	2	1-13490	43060	97	80	3236	71	2012-08-02 11:27:02.776472
43075	4	1-12749	43074	97	66	629	81	2012-08-02 11:27:05.118969
43076	4	1-13167	43074	66	47	370	67	2012-08-02 11:27:05.118969
43077	4	1-13274	43074	47	14	831	89	2012-08-02 11:27:05.118969
43074	4	1-13496	43074	14	97	761	23	2012-08-02 11:27:05.118969
43087	3	1-13236	43086	40	36	2720	76	2012-08-02 11:27:06.323253
43088	3	1-9558	43086	36	22	1601	80	2012-08-02 11:27:06.323253
43086	3	1-13499	43086	22	40	2306	91	2012-08-02 11:27:06.323253
43090	4	1-13236	43089	40	66	454	76	2012-08-02 11:27:06.323253
43091	4	1-10945	43089	66	49	194	97	2012-08-02 11:27:06.323253
43092	4	1-13021	43089	49	22	258	80	2012-08-02 11:27:06.323253
43089	4	1-13499	43089	22	40	378	91	2012-08-02 11:27:06.323253
43094	4	1-11323	43093	52	88	1410	99	2012-08-02 11:27:06.323253
43095	4	1-6979	43093	88	25	710	98	2012-08-02 11:27:06.323253
43096	4	1-12719	43093	25	22	1052	80	2012-08-02 11:27:06.323253
43093	4	1-13499	43093	22	52	1591	91	2012-08-02 11:27:06.323253
43116	2	1-4994	43115	25	58	1113	85	2012-08-02 11:27:08.65467
43115	2	1-13505	43115	58	25	3694	52	2012-08-02 11:27:08.65467
43118	3	1-13260	43117	4	86	682	81	2012-08-02 11:27:08.65467
42899	4	1-12087	42898	16	6	852	80	2012-08-02 11:26:44.023142
42900	4	1-11979	42898	6	72	535	52	2012-08-02 11:26:44.023142
42901	4	1-13221	42898	72	20	493	72	2012-08-02 11:26:44.023142
42898	4	1-13443	42898	20	16	1644	22	2012-08-02 11:26:44.023142
42926	4	1-13254	42925	37	81	1534	21	2012-08-02 11:26:48.433006
42927	4	1-8892	42925	81	42	433	31	2012-08-02 11:26:48.433006
42928	4	1-13344	42925	42	6	945	87	2012-08-02 11:26:48.433006
42925	4	1-13454	42925	6	37	759	97	2012-08-02 11:26:48.433006
42930	4	1-12621	42929	88	12	1953	27	2012-08-02 11:26:48.433006
42931	4	1-13300	42929	12	9	2185	60	2012-08-02 11:26:48.433006
42932	4	1-13068	42929	9	6	2310	87	2012-08-02 11:26:48.433006
42929	4	1-13454	42929	6	88	1904	97	2012-08-02 11:26:48.433006
42934	4	1-12388	42933	10	45	1894	49	2012-08-02 11:26:48.433006
42935	4	1-11752	42933	45	10	1178	3	2012-08-02 11:26:48.433006
42936	4	1-12955	42933	10	6	2171	87	2012-08-02 11:26:48.433006
42933	4	1-13454	42933	6	10	1938	97	2012-08-02 11:26:48.433006
42938	4	1-12621	42937	88	15	375	27	2012-08-02 11:26:48.433006
42939	4	1-12351	42937	15	10	185	3	2012-08-02 11:26:48.433006
42940	4	1-12955	42937	10	6	346	87	2012-08-02 11:26:48.433006
42937	4	1-13454	42937	6	88	323	97	2012-08-02 11:26:48.433006
42942	2	1-9693	42941	79	68	4109	53	2012-08-02 11:26:49.593691
42941	2	1-13457	42941	68	79	7672	74	2012-08-02 11:26:49.593691
42944	3	1-12387	42943	87	43	936	61	2012-08-02 11:26:49.593691
42945	3	1-13419	42943	43	68	628	53	2012-08-02 11:26:49.593691
42943	3	1-13457	42943	68	87	1365	74	2012-08-02 11:26:49.593691
42974	4	1-13367	42973	43	91	34	2	2012-08-02 11:26:55.649625
42975	4	1-12573	42973	91	37	9	96	2012-08-02 11:26:55.649625
42976	4	1-12750	42973	37	19	13	21	2012-08-02 11:26:55.649625
42973	4	1-13472	42973	19	43	41	25	2012-08-02 11:26:55.649625
42978	4	1-13395	42977	82	62	2778	57	2012-08-02 11:26:55.649625
42979	4	1-7703	42977	62	15	863	54	2012-08-02 11:26:55.649625
42980	4	1-12409	42977	15	19	691	21	2012-08-02 11:26:55.649625
42977	4	1-13472	42977	19	82	2556	25	2012-08-02 11:26:55.649625
42982	4	1-12438	42981	69	68	2452	63	2012-08-02 11:26:56.88707
42983	4	1-11630	42981	68	19	2567	53	2012-08-02 11:26:56.88707
42984	4	1-13181	42981	19	84	2101	46	2012-08-02 11:26:56.88707
42981	4	1-13475	42981	84	69	3155	73	2012-08-02 11:26:56.88707
42986	4	1-6193	42985	93	64	196	65	2012-08-02 11:26:56.88707
42987	4	1-12642	42985	64	19	326	53	2012-08-02 11:26:56.88707
42988	4	1-13181	42985	19	84	270	46	2012-08-02 11:26:56.88707
42985	4	1-13475	42985	84	93	411	73	2012-08-02 11:26:56.88707
42990	4	1-13435	42989	51	43	1603	61	2012-08-02 11:26:56.88707
42991	4	1-13419	42989	43	19	1252	53	2012-08-02 11:26:56.88707
42992	4	1-13181	42989	19	84	1078	46	2012-08-02 11:26:56.88707
42989	4	1-13475	42989	84	51	1702	73	2012-08-02 11:26:56.88707
42994	4	1-12944	42993	83	87	1452	38	2012-08-02 11:26:56.88707
42995	4	1-1219	42993	87	74	322	51	2012-08-02 11:26:56.88707
42996	4	1-12631	42993	74	84	521	46	2012-08-02 11:26:56.88707
42993	4	1-13475	42993	84	83	823	73	2012-08-02 11:26:56.88707
42997	1	1-3196	\N	99	99	1096	46	2012-08-02 11:26:56.88707
43030	4	1-11810	43029	13	58	303	21	2012-08-02 11:26:59.229694
43031	4	1-13029	43029	58	94	234	48	2012-08-02 11:26:59.229694
43032	4	1-12783	43029	94	34	406	37	2012-08-02 11:26:59.229694
43029	4	1-13481	43029	34	13	342	18	2012-08-02 11:26:59.229694
43034	4	1-13472	43033	19	22	2502	25	2012-08-02 11:27:01.561032
43035	4	1-3576	43033	22	44	482	53	2012-08-02 11:27:01.561032
43036	4	1-12720	43033	44	7	764	7	2012-08-02 11:27:01.561032
43033	4	1-13486	43033	7	19	804	21	2012-08-02 11:27:01.561032
43038	4	1-13472	43037	19	99	1013	25	2012-08-02 11:27:01.561032
43039	4	1-8117	43037	99	19	328	40	2012-08-02 11:27:01.561032
43040	4	1-13047	43037	19	7	253	7	2012-08-02 11:27:01.561032
43037	4	1-13486	43037	7	19	278	21	2012-08-02 11:27:01.561032
43042	4	1-13313	43041	94	89	428	23	2012-08-02 11:27:01.561032
43043	4	1-8524	43041	89	19	484	40	2012-08-02 11:27:01.561032
43044	4	1-13047	43041	19	7	352	7	2012-08-02 11:27:01.561032
43041	4	1-13486	43041	7	94	409	21	2012-08-02 11:27:01.561032
43046	4	1-13029	43045	58	69	1998	48	2012-08-02 11:27:01.561032
43047	4	1-12133	43045	69	19	1906	40	2012-08-02 11:27:01.561032
43048	4	1-13047	43045	19	7	1546	7	2012-08-02 11:27:01.561032
43045	4	1-13486	43045	7	58	2004	21	2012-08-02 11:27:01.561032
43063	4	1-7650	43062	27	84	2413	2	2012-08-02 11:27:03.56104
43064	4	1-12199	43062	84	10	1403	92	2012-08-02 11:27:03.56104
43065	4	1-12987	43062	10	42	3696	15	2012-08-02 11:27:03.56104
43062	4	1-13492	43062	42	27	3200	52	2012-08-02 11:27:03.56104
43067	4	1-13260	43066	4	68	590	81	2012-08-02 11:27:03.56104
43068	4	1-11594	43066	68	44	561	40	2012-08-02 11:27:03.56104
43069	4	1-13465	43066	44	42	1209	15	2012-08-02 11:27:03.56104
43066	4	1-13492	43066	42	4	1161	52	2012-08-02 11:27:03.56104
43071	4	1-13221	43070	72	52	3055	72	2012-08-02 11:27:03.56104
43072	4	1-11759	43070	52	59	2540	2	2012-08-02 11:27:03.56104
43073	4	1-12363	43070	59	42	3171	15	2012-08-02 11:27:03.56104
43070	4	1-13492	43070	42	72	3075	52	2012-08-02 11:27:03.56104
43079	4	1-13314	43078	60	28	195	47	2012-08-02 11:27:05.925457
43080	4	1-12406	43078	28	48	214	34	2012-08-02 11:27:05.925457
43081	4	1-12802	43078	48	92	87	31	2012-08-02 11:27:05.925457
43078	4	1-13498	43078	92	60	71	35	2012-08-02 11:27:05.925457
43083	4	1-13229	43082	16	33	873	68	2012-08-02 11:27:05.925457
43084	4	1-12583	43082	33	100	1341	67	2012-08-02 11:27:05.925457
43085	4	1-8528	43082	100	92	744	31	2012-08-02 11:27:05.925457
43082	4	1-13498	43082	92	16	636	35	2012-08-02 11:27:05.925457
43106	4	1-10322	43105	60	4	415	77	2012-08-02 11:27:07.097589
43107	4	1-13323	43105	4	24	589	34	2012-08-02 11:27:07.097589
43108	4	1-12596	43105	24	49	688	32	2012-08-02 11:27:07.097589
43105	4	1-13501	43105	49	60	551	11	2012-08-02 11:27:07.097589
43110	3	1-13363	43109	58	46	2294	69	2012-08-02 11:27:07.097589
43111	3	1-13038	43109	46	49	1626	32	2012-08-02 11:27:07.097589
43109	3	1-13501	43109	49	58	1316	11	2012-08-02 11:27:07.097589
43113	3	1-5004	43112	17	22	521	89	2012-08-02 11:27:08.268094
43114	3	1-13207	43112	22	45	706	4	2012-08-02 11:27:08.268094
43112	3	1-13504	43112	45	17	3758	44	2012-08-02 11:27:08.268094
43220	4	1-13089	43219	5	7	1841	28	2012-08-02 11:27:21.750267
43221	4	1-11967	43219	7	58	1044	21	2012-08-02 11:27:21.750267
43222	4	1-13029	43219	58	92	407	48	2012-08-02 11:27:21.750267
43219	4	1-13538	43219	92	5	5999	33	2012-08-02 11:27:21.750267
42956	4	1-10513	42954	85	4	1743	77	2012-08-02 11:26:54.810076
42957	4	1-13323	42954	4	63	2843	34	2012-08-02 11:26:54.810076
42954	4	1-13470	42954	63	21	5378	65	2012-08-02 11:26:54.810076
43018	4	1-12119	43017	17	28	386	6	2012-08-02 11:26:58.444928
43019	4	1-13286	43017	28	64	417	65	2012-08-02 11:26:58.444928
43020	4	1-12642	43017	64	30	742	53	2012-08-02 11:26:58.444928
43017	4	1-13479	43017	30	17	577	99	2012-08-02 11:26:58.444928
43050	4	1-8205	43049	67	40	330	91	2012-08-02 11:27:01.95929
43051	4	1-13236	43049	40	94	383	76	2012-08-02 11:27:01.95929
43052	4	1-12144	43049	94	13	295	79	2012-08-02 11:27:01.95929
43049	4	1-13488	43049	13	67	834	54	2012-08-02 11:27:01.95929
43098	4	1-13440	43097	10	79	2177	44	2012-08-02 11:27:06.710134
43099	4	1-11210	43097	79	93	397	9	2012-08-02 11:27:06.710134
43100	4	1-13442	43097	93	28	374	50	2012-08-02 11:27:06.710134
43097	4	1-13500	43097	28	10	1983	77	2012-08-02 11:27:06.710134
43102	4	1-13376	43101	49	62	1913	6	2012-08-02 11:27:06.710134
43103	4	1-12671	43101	62	82	972	66	2012-08-02 11:27:06.710134
43104	4	1-13318	43101	82	28	716	50	2012-08-02 11:27:06.710134
43101	4	1-13500	43101	28	49	4971	77	2012-08-02 11:27:06.710134
43136	4	1-7347	43135	6	78	887	95	2012-08-02 11:27:12.599056
43137	4	1-13408	43135	78	6	561	86	2012-08-02 11:27:12.599056
43138	4	1-13145	43135	6	30	470	90	2012-08-02 11:27:12.599056
43135	4	1-13515	43135	30	6	1248	1	2012-08-02 11:27:12.599056
43140	4	1-13146	43139	63	62	1012	26	2012-08-02 11:27:12.599056
43141	4	1-13200	43139	62	86	729	18	2012-08-02 11:27:12.599056
43142	4	1-11489	43139	86	30	459	90	2012-08-02 11:27:12.599056
43139	4	1-13515	43139	30	63	1392	1	2012-08-02 11:27:12.599056
43144	4	1-12866	43143	60	86	2131	61	2012-08-02 11:27:12.599056
43145	4	1-5542	43143	86	25	810	10	2012-08-02 11:27:12.599056
43146	4	1-13000	43143	25	30	798	90	2012-08-02 11:27:12.599056
43143	4	1-13515	43143	30	60	2554	1	2012-08-02 11:27:12.599056
43148	4	1-13202	43147	22	90	4817	42	2012-08-02 11:27:12.599056
43149	4	1-12868	43147	90	68	1156	46	2012-08-02 11:27:12.599056
43150	4	1-13453	43147	68	30	1197	90	2012-08-02 11:27:12.599056
43147	4	1-13515	43147	30	22	3879	1	2012-08-02 11:27:12.599056
43165	4	1-13375	43164	34	4	3988	17	2012-08-02 11:27:15.009877
43166	4	1-13464	43164	4	21	3784	99	2012-08-02 11:27:15.009877
43167	4	1-12207	43164	21	58	3493	77	2012-08-02 11:27:15.009877
43164	4	1-13520	43164	58	34	5844	60	2012-08-02 11:27:15.009877
43180	4	1-11863	43179	5	24	3087	57	2012-08-02 11:27:16.943419
43181	4	1-12451	43179	24	73	1005	26	2012-08-02 11:27:16.943419
43182	4	1-13241	43179	73	43	2824	82	2012-08-02 11:27:16.943419
43179	4	1-13525	43179	43	5	4592	28	2012-08-02 11:27:16.943419
43184	4	1-13357	43183	14	99	238	67	2012-08-02 11:27:16.943419
43185	4	1-10267	43183	99	97	164	45	2012-08-02 11:27:16.943419
43186	4	1-13152	43183	97	43	181	82	2012-08-02 11:27:16.943419
43183	4	1-13525	43183	43	14	295	28	2012-08-02 11:27:16.943419
43205	4	1-8984	43204	91	33	724	68	2012-08-02 11:27:17.739084
43206	4	1-12583	43204	33	7	839	67	2012-08-02 11:27:17.739084
43207	4	1-8529	43204	7	56	526	70	2012-08-02 11:27:17.739084
43204	4	1-13528	43204	56	91	1387	3	2012-08-02 11:27:17.739084
43209	3	1-8984	43208	91	92	832	68	2012-08-02 11:27:17.739084
43210	3	1-7908	43208	92	56	498	70	2012-08-02 11:27:17.739084
43208	3	1-13528	43208	56	91	1448	3	2012-08-02 11:27:17.739084
43212	4	1-13191	43211	44	50	3202	72	2012-08-02 11:27:20.169976
43213	4	1-13418	43211	50	77	1852	67	2012-08-02 11:27:20.169976
43214	4	1-12628	43211	77	58	803	53	2012-08-02 11:27:20.169976
43211	4	1-13534	43211	58	44	2417	86	2012-08-02 11:27:20.169976
43230	3	1-9873	43229	42	63	2955	87	2012-08-02 11:27:22.677663
43231	3	1-13477	43229	63	65	2727	73	2012-08-02 11:27:22.677663
43229	3	1-13540	43229	65	42	2795	56	2012-08-02 11:27:22.677663
43239	4	1-12313	43238	91	46	1344	91	2012-08-02 11:27:27.04239
43240	4	1-13113	43238	46	22	1304	11	2012-08-02 11:27:27.04239
43241	4	1-13436	43238	22	1	1316	100	2012-08-02 11:27:27.04239
43238	4	1-13550	43238	1	91	1185	51	2012-08-02 11:27:27.04239
43246	4	1-13206	43245	65	79	2249	44	2012-08-02 11:27:27.848631
43247	4	1-11210	43245	79	22	683	9	2012-08-02 11:27:27.848631
43248	4	1-13002	43245	22	74	1038	38	2012-08-02 11:27:27.848631
43245	4	1-13553	43245	74	65	1410	56	2012-08-02 11:27:27.848631
43250	3	1-13219	43249	79	83	1890	73	2012-08-02 11:27:27.848631
43251	3	1-12944	43249	83	74	3098	38	2012-08-02 11:27:27.848631
43249	3	1-13553	43249	74	79	4340	56	2012-08-02 11:27:27.848631
43269	3	1-9872	43268	4	9	497	37	2012-08-02 11:27:31.75991
43270	3	1-12975	43268	9	26	26	99	2012-08-02 11:27:31.75991
43268	3	1-13563	43268	26	4	8555	14	2012-08-02 11:27:31.75991
43280	4	1-13196	43279	6	21	4098	46	2012-08-02 11:27:32.974984
43281	4	1-13309	43279	21	90	2376	22	2012-08-02 11:27:32.974984
43282	4	1-12940	43279	90	25	2376	17	2012-08-02 11:27:32.974984
43279	4	1-13567	43279	25	6	7374	43	2012-08-02 11:27:32.974984
43301	4	1-13290	43300	11	9	3475	37	2012-08-02 11:27:38.068659
43302	4	1-12975	43300	9	78	1391	99	2012-08-02 11:27:38.068659
43303	4	1-13555	43300	78	44	3922	5	2012-08-02 11:27:38.068659
43300	4	1-13579	43300	44	11	2498	29	2012-08-02 11:27:38.068659
43312	4	1-12403	43311	33	13	358	65	2012-08-02 11:27:39.261479
43313	4	1-11103	43311	13	59	261	61	2012-08-02 11:27:39.261479
43314	4	1-12585	43311	59	64	305	38	2012-08-02 11:27:39.261479
43311	4	1-13582	43311	64	33	340	39	2012-08-02 11:27:39.261479
43316	4	1-12920	43315	86	37	706	19	2012-08-02 11:27:39.261479
43317	4	1-9037	43315	37	43	710	10	2012-08-02 11:27:39.261479
43318	4	1-10499	43315	43	64	544	38	2012-08-02 11:27:39.261479
43315	4	1-13582	43315	64	86	657	39	2012-08-02 11:27:39.261479
43320	4	1-12920	43319	86	53	1894	19	2012-08-02 11:27:39.261479
43321	4	1-8506	43319	53	43	1799	10	2012-08-02 11:27:39.261479
43322	4	1-10499	43319	43	64	1404	38	2012-08-02 11:27:39.261479
43319	4	1-13582	43319	64	86	1730	39	2012-08-02 11:27:39.261479
43324	3	1-8350	43323	55	22	922	9	2012-08-02 11:27:39.261479
43325	3	1-13002	43323	22	64	1660	38	2012-08-02 11:27:39.261479
43323	3	1-13582	43323	64	55	2178	39	2012-08-02 11:27:39.261479
43327	3	1-12117	43326	2	59	196	61	2012-08-02 11:27:39.261479
43328	3	1-12585	43326	59	64	271	38	2012-08-02 11:27:39.261479
43326	3	1-13582	43326	64	2	356	39	2012-08-02 11:27:39.261479
43119	3	1-11857	43117	86	58	488	85	2012-08-02 11:27:08.65467
43117	3	1-13505	43117	58	4	1744	52	2012-08-02 11:27:08.65467
43121	4	1-13221	43120	72	79	2301	72	2012-08-02 11:27:08.65467
43122	4	1-6137	43120	79	87	458	84	2012-08-02 11:27:08.65467
43123	4	1-10510	43120	87	58	527	85	2012-08-02 11:27:08.65467
43120	4	1-13505	43120	58	72	2394	52	2012-08-02 11:27:08.65467
43125	3	1-6503	43124	82	47	203	11	2012-08-02 11:27:08.65467
43126	3	1-8804	43124	47	58	212	85	2012-08-02 11:27:08.65467
43124	3	1-13505	43124	58	82	1015	52	2012-08-02 11:27:08.65467
43128	4	1-13239	43127	10	79	3011	23	2012-08-02 11:27:11.019619
43129	4	1-13398	43127	79	46	2939	100	2012-08-02 11:27:11.019619
43130	4	1-12425	43127	46	18	2219	68	2012-08-02 11:27:11.019619
43127	4	1-13511	43127	18	10	1839	63	2012-08-02 11:27:11.019619
43131	1	1-11136	\N	72	72	5629	23	2012-08-02 11:27:11.019619
43133	3	1-13151	43132	53	50	2379	72	2012-08-02 11:27:12.224339
43134	3	1-13418	43132	50	55	1351	67	2012-08-02 11:27:12.224339
43132	3	1-13514	43132	55	53	1823	59	2012-08-02 11:27:12.224339
43152	4	1-12917	43151	35	91	665	46	2012-08-02 11:27:13.362106
43153	4	1-13506	43151	91	35	417	13	2012-08-02 11:27:13.362106
43154	4	1-13192	43151	35	20	438	88	2012-08-02 11:27:13.362106
43151	4	1-13517	43151	20	35	1509	93	2012-08-02 11:27:13.362106
43156	2	1-3282	43155	8	20	645	88	2012-08-02 11:27:13.362106
43155	2	1-13517	43155	20	8	2271	93	2012-08-02 11:27:13.362106
43158	3	1-8824	43157	40	4	1262	65	2012-08-02 11:27:13.362106
43159	3	1-5601	43157	4	20	728	88	2012-08-02 11:27:13.362106
43157	3	1-13517	43157	20	40	2684	93	2012-08-02 11:27:13.362106
43161	4	1-11100	43160	35	25	336	27	2012-08-02 11:27:13.362106
43162	4	1-11159	43160	25	35	292	13	2012-08-02 11:27:13.362106
43163	4	1-13192	43160	35	20	319	88	2012-08-02 11:27:13.362106
43160	4	1-13517	43160	20	35	1191	93	2012-08-02 11:27:13.362106
43169	4	1-12905	43168	30	84	136	17	2012-08-02 11:27:15.385591
43170	4	1-13346	43168	84	9	160	55	2012-08-02 11:27:15.385591
43171	4	1-13424	43168	9	98	378	38	2012-08-02 11:27:15.385591
43168	4	1-13522	43168	98	30	226	51	2012-08-02 11:27:15.385591
43173	3	1-13347	43172	94	63	2188	100	2012-08-02 11:27:15.385591
43174	3	1-13385	43172	63	98	3037	38	2012-08-02 11:27:15.385591
43172	3	1-13522	43172	98	94	2239	51	2012-08-02 11:27:15.385591
43176	4	1-13347	43175	94	61	417	100	2012-08-02 11:27:15.385591
43177	4	1-13447	43175	61	22	325	9	2012-08-02 11:27:15.385591
43178	4	1-13002	43175	22	98	550	38	2012-08-02 11:27:15.385591
43175	4	1-13522	43175	98	94	416	51	2012-08-02 11:27:15.385591
43188	4	1-12724	43187	71	26	733	91	2012-08-02 11:27:16.568012
43189	4	1-12519	43187	26	53	314	96	2012-08-02 11:27:16.568012
43190	4	1-13240	43187	53	92	231	31	2012-08-02 11:27:16.568012
43187	4	1-13526	43187	92	71	1033	90	2012-08-02 11:27:16.568012
43192	4	1-8072	43191	49	70	4319	60	2012-08-02 11:27:16.568012
43193	4	1-13308	43191	70	53	1768	96	2012-08-02 11:27:16.568012
43194	4	1-13240	43191	53	92	1389	31	2012-08-02 11:27:16.568012
43191	4	1-13526	43191	92	49	6618	90	2012-08-02 11:27:16.568012
43196	4	1-9537	43195	84	28	133	80	2012-08-02 11:27:16.568012
43197	4	1-13061	43195	28	53	128	96	2012-08-02 11:27:16.568012
43198	4	1-13240	43195	53	92	105	31	2012-08-02 11:27:16.568012
43195	4	1-13526	43195	92	84	523	90	2012-08-02 11:27:16.568012
43200	4	1-12547	43199	78	5	128	23	2012-08-02 11:27:16.568012
43201	4	1-7792	43199	5	5	57	13	2012-08-02 11:27:16.568012
43202	4	1-5267	43199	5	92	20	31	2012-08-02 11:27:16.568012
43199	4	1-13526	43199	92	78	100	90	2012-08-02 11:27:16.568012
43203	1	1-4387	\N	28	28	1700	96	2012-08-02 11:27:16.568012
43216	4	1-13258	43215	69	34	2417	75	2012-08-02 11:27:20.943618
43217	4	1-12683	43215	34	4	1410	47	2012-08-02 11:27:20.943618
43218	4	1-12871	43215	4	7	914	8	2012-08-02 11:27:20.943618
43215	4	1-13536	43215	7	69	2529	64	2012-08-02 11:27:20.943618
43233	4	1-13152	43232	97	68	4984	82	2012-08-02 11:27:23.488864
43234	4	1-8011	43232	68	81	1768	80	2012-08-02 11:27:23.488864
43235	4	1-13271	43232	81	13	1077	89	2012-08-02 11:27:23.488864
43232	4	1-13542	43232	13	97	3132	45	2012-08-02 11:27:23.488864
43236	1	1-8011	\N	68	68	1486	80	2012-08-02 11:27:23.488864
43237	1	1-4267	\N	50	50	117	82	2012-08-02 11:27:23.488864
43243	3	1-13315	43242	28	46	89	69	2012-08-02 11:27:26.633186
43244	3	1-13038	43242	46	30	75	32	2012-08-02 11:27:26.633186
43242	3	1-13551	43242	30	28	64	88	2012-08-02 11:27:26.633186
43257	4	1-13511	43256	18	61	2801	63	2012-08-02 11:27:29.384122
43258	4	1-12913	43256	61	79	3087	56	2012-08-02 11:27:29.384122
43259	4	1-13219	43256	79	45	1460	73	2012-08-02 11:27:29.384122
43256	4	1-13557	43256	45	18	3438	68	2012-08-02 11:27:29.384122
43261	3	1-12574	43260	98	84	847	46	2012-08-02 11:27:29.384122
43262	3	1-13475	43260	84	45	1362	73	2012-08-02 11:27:29.384122
43260	3	1-13557	43260	45	98	3240	68	2012-08-02 11:27:29.384122
43264	4	1-13448	43263	6	13	860	65	2012-08-02 11:27:29.384122
43265	4	1-11103	43263	13	86	727	61	2012-08-02 11:27:29.384122
43266	4	1-6531	43263	86	45	390	73	2012-08-02 11:27:29.384122
43263	4	1-13557	43263	45	6	956	68	2012-08-02 11:27:29.384122
43267	1	1-3999	\N	67	67	82	61	2012-08-02 11:27:29.384122
43272	4	1-13499	43271	22	8	2604	91	2012-08-02 11:27:32.588199
43273	4	1-6033	43271	8	98	888	42	2012-08-02 11:27:32.588199
43274	4	1-12343	43271	98	91	322	27	2012-08-02 11:27:32.588199
43271	4	1-13565	43271	91	22	2400	80	2012-08-02 11:27:32.588199
43276	4	1-13061	43275	28	14	491	96	2012-08-02 11:27:32.588199
43277	4	1-8873	43275	14	25	141	100	2012-08-02 11:27:32.588199
43278	4	1-12454	43275	25	91	94	27	2012-08-02 11:27:32.588199
43275	4	1-13565	43275	91	28	703	80	2012-08-02 11:27:32.588199
43284	4	1-11662	43283	77	95	1087	17	2012-08-02 11:27:33.74851
43285	4	1-12505	43283	95	10	793	92	2012-08-02 11:27:33.74851
43286	4	1-12987	43283	10	85	1215	15	2012-08-02 11:27:33.74851
43283	4	1-13568	43283	85	77	2188	8	2012-08-02 11:27:33.74851
43288	3	1-43	43287	35	44	285	40	2012-08-02 11:27:33.74851
43289	3	1-13465	43287	44	85	336	15	2012-08-02 11:27:33.74851
43287	3	1-13568	43287	85	35	632	8	2012-08-02 11:27:33.74851
43291	4	1-12165	43290	71	96	1415	20	2012-08-02 11:27:33.74851
43292	4	1-9295	43290	96	59	821	2	2012-08-02 11:27:33.74851
43293	4	1-12363	43290	59	85	736	15	2012-08-02 11:27:33.74851
43290	4	1-13568	43290	85	71	1831	8	2012-08-02 11:27:33.74851
43224	3	1-12715	43223	15	20	212	49	2012-08-02 11:27:21.750267
43225	3	1-7838	43223	20	92	40	48	2012-08-02 11:27:21.750267
43223	3	1-13538	43223	92	15	615	33	2012-08-02 11:27:21.750267
43227	3	1-6195	43226	59	65	33	73	2012-08-02 11:27:21.750267
43228	3	1-11344	43226	65	92	14	48	2012-08-02 11:27:21.750267
43226	3	1-13538	43226	92	59	223	33	2012-08-02 11:27:21.750267
43253	4	1-12797	43252	76	36	5330	83	2012-08-02 11:27:28.61077
43254	4	1-8999	43252	36	9	3293	37	2012-08-02 11:27:28.61077
43255	4	1-12975	43252	9	78	1256	99	2012-08-02 11:27:28.61077
43252	4	1-13555	43252	78	76	3375	5	2012-08-02 11:27:28.61077
43298	3	1-12596	43297	24	56	2881	32	2012-08-02 11:27:34.135171
43299	3	1-5899	43297	56	48	2596	2	2012-08-02 11:27:34.135171
43297	3	1-13569	43297	48	24	2627	34	2012-08-02 11:27:34.135171
43305	4	1-5872	43304	93	4	513	17	2012-08-02 11:27:38.853085
43306	4	1-13464	43304	4	78	448	99	2012-08-02 11:27:38.853085
43307	4	1-13555	43304	78	68	1349	5	2012-08-02 11:27:38.853085
43304	4	1-13581	43304	68	93	1520	77	2012-08-02 11:27:38.853085
43309	3	1-13520	43308	58	59	1091	60	2012-08-02 11:27:38.853085
43310	3	1-13246	43308	59	68	625	5	2012-08-02 11:27:38.853085
43308	3	1-13581	43308	68	58	707	77	2012-08-02 11:27:38.853085
43446	2	1-6317	43445	19	100	189	5	2012-08-02 11:27:56.667174
43445	2	1-13625	43445	100	19	372	61	2012-08-02 11:27:56.667174
43448	3	1-9724	43447	48	12	3231	42	2012-08-02 11:27:56.667174
43449	3	1-12374	43447	12	100	1760	5	2012-08-02 11:27:56.667174
43447	3	1-13625	43447	100	48	4023	61	2012-08-02 11:27:56.667174
43451	4	1-11288	43450	90	55	712	62	2012-08-02 11:27:56.667174
43452	4	1-13332	43450	55	40	285	68	2012-08-02 11:27:56.667174
43453	4	1-13247	43450	40	100	397	5	2012-08-02 11:27:56.667174
43450	4	1-13625	43450	100	90	910	61	2012-08-02 11:27:56.667174
43455	4	1-9724	43454	48	64	1797	42	2012-08-02 11:27:56.667174
43456	4	1-12956	43454	64	76	874	29	2012-08-02 11:27:56.667174
43457	4	1-11664	43454	76	100	853	5	2012-08-02 11:27:56.667174
43454	4	1-13625	43454	100	48	2088	61	2012-08-02 11:27:56.667174
43479	4	1-13520	43478	58	34	2361	60	2012-08-02 11:27:57.794761
43480	4	1-13375	43478	34	27	1604	17	2012-08-02 11:27:57.794761
43481	4	1-13078	43478	27	16	2424	87	2012-08-02 11:27:57.794761
43478	4	1-13629	43478	16	58	1417	77	2012-08-02 11:27:57.794761
43499	3	1-13549	43498	40	7	1371	79	2012-08-02 11:28:00.831823
43500	3	1-11075	43498	7	10	775	100	2012-08-02 11:28:00.831823
43498	3	1-13634	43498	10	40	1298	96	2012-08-02 11:28:00.831823
43506	4	1-11287	43505	13	80	795	55	2012-08-02 11:28:02.777661
43507	4	1-13480	43505	80	79	1226	72	2012-08-02 11:28:02.777661
43508	4	1-13165	43505	79	32	835	10	2012-08-02 11:28:02.777661
43505	4	1-13640	43505	32	13	1306	98	2012-08-02 11:28:02.777661
43518	4	1-12657	43517	83	37	1199	20	2012-08-02 11:28:07.419662
43519	4	1-13072	43517	37	68	1070	37	2012-08-02 11:28:07.419662
43520	4	1-12609	43517	68	28	842	14	2012-08-02 11:28:07.419662
43517	4	1-13653	43517	28	83	565	10	2012-08-02 11:28:07.419662
43522	3	1-11430	43521	34	83	918	45	2012-08-02 11:28:07.419662
43523	3	1-12421	43521	83	28	1075	14	2012-08-02 11:28:07.419662
43521	3	1-13653	43521	28	34	803	10	2012-08-02 11:28:07.419662
43525	4	1-10499	43524	43	64	1392	38	2012-08-02 11:28:07.419662
43526	4	1-13582	43524	64	98	1757	39	2012-08-02 11:28:07.419662
43527	4	1-8721	43524	98	28	2311	14	2012-08-02 11:28:07.419662
43524	4	1-13653	43524	28	43	1740	10	2012-08-02 11:28:07.419662
43529	4	1-13640	43528	32	99	2210	98	2012-08-02 11:28:07.419662
43530	4	1-7989	43528	99	68	2058	37	2012-08-02 11:28:07.419662
43531	4	1-12609	43528	68	28	1840	14	2012-08-02 11:28:07.419662
43528	4	1-13653	43528	28	32	1404	10	2012-08-02 11:28:07.419662
43563	2	1-13609	43562	33	53	2779	59	2012-08-02 11:28:10.148295
43562	2	1-13659	43562	53	33	5375	77	2012-08-02 11:28:10.148295
43608	2	1-9308	43607	46	28	3944	60	2012-08-02 11:28:14.866431
43607	2	1-13671	43607	28	46	7467	95	2012-08-02 11:28:14.866431
43610	3	1-13264	43609	1	77	115	3	2012-08-02 11:28:14.866431
43611	3	1-8077	43609	77	28	129	60	2012-08-02 11:28:14.866431
43609	3	1-13671	43609	28	1	306	95	2012-08-02 11:28:14.866431
43638	3	1-13609	43637	33	2	1028	59	2012-08-02 11:28:20.733231
43639	3	1-13414	43637	2	55	959	95	2012-08-02 11:28:20.733231
43637	3	1-13686	43637	55	33	2675	77	2012-08-02 11:28:20.733231
43641	4	1-13602	43640	47	34	819	10	2012-08-02 11:28:20.733231
43642	4	1-11430	43640	34	79	836	45	2012-08-02 11:28:20.733231
43643	4	1-13462	43640	79	55	467	95	2012-08-02 11:28:20.733231
43640	4	1-13686	43640	55	47	1641	77	2012-08-02 11:28:20.733231
43645	3	1-11856	43644	19	67	2084	2	2012-08-02 11:28:20.733231
43646	3	1-12244	43644	67	55	387	95	2012-08-02 11:28:20.733231
43644	3	1-13686	43644	55	19	1583	77	2012-08-02 11:28:20.733231
43670	4	1-11344	43669	65	94	127	48	2012-08-02 11:28:25.053385
43671	4	1-12783	43669	94	68	287	37	2012-08-02 11:28:25.053385
43672	4	1-12609	43669	68	58	250	14	2012-08-02 11:28:25.053385
43669	4	1-13697	43669	58	65	132	73	2012-08-02 11:28:25.053385
43680	3	1-13658	43679	70	67	5368	39	2012-08-02 11:28:27.428723
43681	3	1-5849	43679	67	38	1636	29	2012-08-02 11:28:27.428723
43679	3	1-13705	43679	38	70	7267	58	2012-08-02 11:28:27.428723
43683	4	1-13227	43682	80	64	385	26	2012-08-02 11:28:27.428723
43684	4	1-9968	43682	64	64	230	42	2012-08-02 11:28:27.428723
43685	4	1-12956	43682	64	38	114	29	2012-08-02 11:28:27.428723
43682	4	1-13705	43682	38	80	683	58	2012-08-02 11:28:27.428723
43709	4	1-8693	43708	21	44	6846	5	2012-08-02 11:28:30.535322
43710	4	1-13579	43708	44	44	4580	29	2012-08-02 11:28:30.535322
43711	4	1-9862	43708	44	92	2418	88	2012-08-02 11:28:30.535322
43708	4	1-13712	43708	92	21	6549	11	2012-08-02 11:28:30.535322
43713	4	1-12714	43712	49	42	607	49	2012-08-02 11:28:30.535322
43714	4	1-13675	43712	42	71	931	74	2012-08-02 11:28:30.535322
43715	4	1-9095	43712	71	92	876	88	2012-08-02 11:28:30.535322
43712	4	1-13712	43712	92	49	2406	11	2012-08-02 11:28:30.535322
43737	2	1-11438	43736	57	6	842	51	2012-08-02 11:28:34.42507
43736	2	1-13721	43736	6	57	2283	66	2012-08-02 11:28:34.42507
43739	4	1-11560	43738	37	10	509	63	2012-08-02 11:28:34.42507
43740	4	1-13239	43738	10	45	615	23	2012-08-02 11:28:34.42507
43741	4	1-13630	43738	45	6	571	51	2012-08-02 11:28:34.42507
43738	4	1-13721	43738	6	37	1596	66	2012-08-02 11:28:34.42507
43743	4	1-13526	43742	92	15	588	90	2012-08-02 11:28:35.596534
43744	4	1-11924	43742	15	58	169	29	2012-08-02 11:28:35.596534
43742	4	1-13724	43742	62	92	291	31	2012-08-02 11:28:35.596534
43295	3	1-2255	43294	74	13	1349	50	2012-08-02 11:27:33.74851
43296	3	1-12026	43294	13	85	847	15	2012-08-02 11:27:33.74851
43294	3	1-13568	43294	85	74	2424	8	2012-08-02 11:27:33.74851
43330	4	1-13112	43329	36	71	3780	7	2012-08-02 11:27:39.648159
43331	4	1-11749	43329	71	48	2863	2	2012-08-02 11:27:39.648159
43332	4	1-13569	43329	48	2	3804	34	2012-08-02 11:27:39.648159
43329	4	1-13583	43329	2	36	1361	79	2012-08-02 11:27:39.648159
43351	4	1-13402	43350	35	3	424	8	2012-08-02 11:27:42.932116
43352	4	1-11776	43350	3	52	323	72	2012-08-02 11:27:42.932116
43353	4	1-13193	43350	52	31	85	25	2012-08-02 11:27:42.932116
43350	4	1-13590	43350	31	35	722	79	2012-08-02 11:27:42.932116
43400	4	1-12237	43399	20	11	33	58	2012-08-02 11:27:47.287245
43401	4	1-13261	43399	11	73	52	24	2012-08-02 11:27:47.287245
43402	4	1-11284	43399	73	47	96	77	2012-08-02 11:27:47.287245
43399	4	1-13602	43399	47	20	55	10	2012-08-02 11:27:47.287245
43404	2	1-12245	43403	50	22	439	81	2012-08-02 11:27:48.491294
43403	2	1-13605	43403	22	50	2509	69	2012-08-02 11:27:48.491294
43406	4	1-12331	43405	58	23	1633	50	2012-08-02 11:27:49.662717
43407	4	1-12449	43405	23	80	368	6	2012-08-02 11:27:49.662717
43408	4	1-11941	43405	80	2	131	77	2012-08-02 11:27:49.662717
43405	4	1-13608	43405	2	58	4382	76	2012-08-02 11:27:49.662717
43410	4	1-12144	43409	94	78	191	79	2012-08-02 11:27:49.662717
43411	4	1-11528	43409	78	80	46	6	2012-08-02 11:27:49.662717
43412	4	1-11941	43409	80	2	18	77	2012-08-02 11:27:49.662717
43409	4	1-13608	43409	2	94	663	76	2012-08-02 11:27:49.662717
43414	3	1-10945	43413	66	91	766	97	2012-08-02 11:27:49.662717
43415	3	1-9281	43413	91	2	125	77	2012-08-02 11:27:49.662717
43413	3	1-13608	43413	2	66	4773	76	2012-08-02 11:27:49.662717
43416	1	1-10945	\N	66	66	1156	97	2012-08-02 11:27:49.662717
43418	3	1-13544	43417	85	22	1276	59	2012-08-02 11:27:50.847531
43419	3	1-13006	43417	22	69	746	89	2012-08-02 11:27:50.847531
43417	3	1-13610	43417	69	85	5479	64	2012-08-02 11:27:50.847531
43420	1	1-3236	\N	22	22	98	89	2012-08-02 11:27:50.847531
43438	4	1-13019	43437	97	1	645	75	2012-08-02 11:27:55.871376
43439	4	1-12443	43437	1	80	467	58	2012-08-02 11:27:55.871376
43440	4	1-13227	43437	80	43	354	26	2012-08-02 11:27:55.871376
43437	4	1-13624	43437	43	97	629	30	2012-08-02 11:27:55.871376
43442	4	1-13487	43441	63	65	3072	20	2012-08-02 11:27:55.871376
43443	4	1-10985	43441	65	78	1291	65	2012-08-02 11:27:55.871376
43444	4	1-13123	43441	78	43	1253	26	2012-08-02 11:27:55.871376
43441	4	1-13624	43441	43	63	2350	30	2012-08-02 11:27:55.871376
43459	3	1-13313	43458	94	2	1101	23	2012-08-02 11:27:57.176007
43460	3	1-12982	43458	2	95	565	53	2012-08-02 11:27:57.176007
43458	3	1-13626	43458	95	94	900	21	2012-08-02 11:27:57.176007
43471	4	1-11708	43470	71	48	2347	2	2012-08-02 11:27:58.263015
43472	4	1-13569	43470	48	49	2559	34	2012-08-02 11:27:58.263015
43473	4	1-11872	43470	49	51	2927	52	2012-08-02 11:27:58.263015
43470	4	1-13628	43470	51	71	2884	46	2012-08-02 11:27:58.263015
43475	4	1-11708	43474	71	59	648	2	2012-08-02 11:27:58.263015
43476	4	1-12363	43474	59	72	734	15	2012-08-02 11:27:58.263015
43477	4	1-13519	43474	72	51	805	52	2012-08-02 11:27:58.263015
43474	4	1-13628	43474	51	71	795	46	2012-08-02 11:27:58.263015
43487	4	1-11987	43486	27	25	2779	100	2012-08-02 11:27:59.982053
43488	4	1-12454	43486	25	59	2229	27	2012-08-02 11:27:59.982053
43489	4	1-11905	43486	59	71	2200	50	2012-08-02 11:27:59.982053
43486	4	1-13633	43486	71	27	3421	28	2012-08-02 11:27:59.982053
43491	4	1-13533	43490	10	94	1620	4	2012-08-02 11:27:59.982053
43492	4	1-12862	43490	94	92	1065	48	2012-08-02 11:27:59.982053
43493	4	1-13368	43490	92	71	582	50	2012-08-02 11:27:59.982053
43490	4	1-13633	43490	71	10	937	28	2012-08-02 11:27:59.982053
43495	4	1-10179	43494	68	97	1513	30	2012-08-02 11:27:59.982053
43496	4	1-13019	43494	97	29	1480	75	2012-08-02 11:27:59.982053
43497	4	1-13482	43494	29	71	1587	50	2012-08-02 11:27:59.982053
43494	4	1-13633	43494	71	68	2597	28	2012-08-02 11:27:59.982053
43569	3	1-11903	43568	99	19	1017	92	2012-08-02 11:28:13.275174
43570	3	1-12916	43568	19	48	1561	37	2012-08-02 11:28:13.275174
43568	3	1-13666	43568	48	99	2838	93	2012-08-02 11:28:13.275174
43572	4	1-13118	43571	32	38	2091	62	2012-08-02 11:28:13.275174
43573	4	1-12455	43571	38	94	1407	48	2012-08-02 11:28:13.275174
43574	4	1-12783	43571	94	48	3071	37	2012-08-02 11:28:13.275174
43571	4	1-13666	43571	48	32	5947	93	2012-08-02 11:28:13.275174
43576	4	1-4226	43575	57	58	35	21	2012-08-02 11:28:13.275174
43577	4	1-13029	43575	58	94	35	48	2012-08-02 11:28:13.275174
43578	4	1-12783	43575	94	48	80	37	2012-08-02 11:28:13.275174
43575	4	1-13666	43575	48	57	160	93	2012-08-02 11:28:13.275174
43613	4	1-12533	43612	14	27	3170	84	2012-08-02 11:28:16.424944
43614	4	1-13394	43612	27	11	1645	43	2012-08-02 11:28:16.424944
43615	4	1-8137	43612	11	42	819	49	2012-08-02 11:28:16.424944
43612	4	1-13675	43612	42	14	1237	74	2012-08-02 11:28:16.424944
43617	3	1-10808	43616	78	10	747	97	2012-08-02 11:28:16.424944
43618	3	1-12388	43616	10	42	684	49	2012-08-02 11:28:16.424944
43616	3	1-13675	43616	42	78	1176	74	2012-08-02 11:28:16.424944
43624	2	1-13199	43623	26	21	2375	10	2012-08-02 11:28:18.391519
43623	2	1-13679	43623	21	26	7458	31	2012-08-02 11:28:18.391519
43626	4	1-13588	43625	45	20	2720	95	2012-08-02 11:28:19.186746
43627	4	1-11963	43625	20	38	1619	16	2012-08-02 11:28:19.186746
43628	4	1-13437	43625	38	39	823	38	2012-08-02 11:28:19.186746
43625	4	1-13681	43625	39	45	2142	45	2012-08-02 11:28:19.186746
43648	3	1-13414	43647	2	31	7844	95	2012-08-02 11:28:22.314052
43649	3	1-13324	43647	31	18	4697	99	2012-08-02 11:28:22.314052
43647	3	1-13690	43647	18	2	7389	59	2012-08-02 11:28:22.314052
43651	4	1-13014	43650	71	90	725	22	2012-08-02 11:28:22.314052
43652	4	1-12940	43650	90	4	711	17	2012-08-02 11:28:22.314052
43653	4	1-13464	43650	4	18	542	99	2012-08-02 11:28:22.314052
43650	4	1-13690	43650	18	71	886	59	2012-08-02 11:28:22.314052
43655	4	1-13006	43654	22	22	687	89	2012-08-02 11:28:22.314052
43656	4	1-13207	43654	22	2	1379	4	2012-08-02 11:28:22.314052
43657	4	1-12668	43654	2	18	297	99	2012-08-02 11:28:22.314052
43654	4	1-13690	43654	18	22	519	59	2012-08-02 11:28:22.314052
43659	3	1-11209	43658	89	60	458	30	2012-08-02 11:28:22.314052
43660	3	1-12753	43658	60	18	357	99	2012-08-02 11:28:22.314052
43658	3	1-13690	43658	18	89	645	59	2012-08-02 11:28:22.314052
43334	4	1-11987	43333	27	6	2692	100	2012-08-02 11:27:41.683599
43335	4	1-12900	43333	6	65	2589	56	2012-08-02 11:27:41.683599
43336	4	1-13206	43333	65	4	3312	44	2012-08-02 11:27:41.683599
43333	4	1-13589	43333	4	27	4196	28	2012-08-02 11:27:41.683599
43338	4	1-13272	43337	42	84	1814	46	2012-08-02 11:27:41.683599
43339	4	1-13475	43337	84	6	2098	73	2012-08-02 11:27:41.683599
43340	4	1-13023	43337	6	4	1625	44	2012-08-02 11:27:41.683599
43337	4	1-13589	43337	4	42	2088	28	2012-08-02 11:27:41.683599
43342	4	1-11967	43341	7	94	505	21	2012-08-02 11:27:41.683599
43343	4	1-13313	43341	94	34	484	23	2012-08-02 11:27:41.683599
43344	4	1-12538	43341	34	4	293	44	2012-08-02 11:27:41.683599
43341	4	1-13589	43341	4	7	424	28	2012-08-02 11:27:41.683599
43346	4	1-13533	43345	10	12	2315	4	2012-08-02 11:27:41.683599
43347	4	1-10344	43345	12	28	1476	42	2012-08-02 11:27:41.683599
43348	4	1-13177	43345	28	4	964	44	2012-08-02 11:27:41.683599
43345	4	1-13589	43345	4	10	1465	28	2012-08-02 11:27:41.683599
43349	1	1-10344	\N	12	12	2357	42	2012-08-02 11:27:41.683599
43355	4	1-12795	43354	38	19	795	15	2012-08-02 11:27:43.340954
43356	4	1-3489	43354	19	24	83	43	2012-08-02 11:27:43.340954
43357	4	1-13340	43354	24	59	60	63	2012-08-02 11:27:43.340954
43354	4	1-13592	43354	59	38	206	97	2012-08-02 11:27:43.340954
43359	4	1-12576	43358	55	56	828	48	2012-08-02 11:27:43.340954
43360	4	1-13343	43358	56	24	392	43	2012-08-02 11:27:43.340954
43361	4	1-13340	43358	24	59	306	63	2012-08-02 11:27:43.340954
43358	4	1-13592	43358	59	55	1136	97	2012-08-02 11:27:43.340954
43363	3	1-11374	43362	78	35	1321	33	2012-08-02 11:27:43.340954
43364	3	1-13294	43362	35	59	1117	63	2012-08-02 11:27:43.340954
43362	3	1-13592	43362	59	78	4370	97	2012-08-02 11:27:43.340954
43366	4	1-13299	43365	4	84	1154	17	2012-08-02 11:27:44.513646
43367	4	1-13346	43365	84	80	1607	55	2012-08-02 11:27:44.513646
43368	4	1-13480	43365	80	82	2250	72	2012-08-02 11:27:44.513646
43365	4	1-13595	43365	82	4	739	35	2012-08-02 11:27:44.513646
43370	4	1-13314	43369	60	4	758	47	2012-08-02 11:27:44.513646
43371	4	1-12871	43369	4	3	468	8	2012-08-02 11:27:44.513646
43372	4	1-11776	43369	3	82	747	72	2012-08-02 11:27:44.513646
43369	4	1-13595	43369	82	60	255	35	2012-08-02 11:27:44.513646
43374	3	1-13183	43373	86	79	1600	3	2012-08-02 11:27:45.695868
43375	3	1-6155	43373	79	32	787	2	2012-08-02 11:27:45.695868
43373	3	1-13597	43373	32	86	3142	9	2012-08-02 11:27:45.695868
43377	4	1-13442	43376	93	22	3565	50	2012-08-02 11:27:45.695868
43378	4	1-13410	43376	22	43	1823	25	2012-08-02 11:27:45.695868
43379	4	1-13367	43376	43	32	953	2	2012-08-02 11:27:45.695868
43376	4	1-13597	43376	32	93	3970	9	2012-08-02 11:27:45.695868
43381	3	1-12879	43380	17	43	584	25	2012-08-02 11:27:45.695868
43382	3	1-13367	43380	43	32	337	2	2012-08-02 11:27:45.695868
43380	3	1-13597	43380	32	17	1551	9	2012-08-02 11:27:45.695868
43384	4	1-13202	43383	22	30	170	42	2012-08-02 11:27:46.491424
43385	4	1-13004	43383	30	15	146	57	2012-08-02 11:27:46.491424
43386	4	1-13306	43383	15	13	140	92	2012-08-02 11:27:46.491424
43383	4	1-13600	43383	13	22	134	1	2012-08-02 11:27:46.491424
43388	4	1-5551	43387	70	36	1219	31	2012-08-02 11:27:46.491424
43389	4	1-13529	43387	36	15	2803	57	2012-08-02 11:27:46.491424
43390	4	1-13306	43387	15	13	3222	92	2012-08-02 11:27:46.491424
43387	4	1-13600	43387	13	70	3708	1	2012-08-02 11:27:46.491424
43392	3	1-13146	43391	63	20	2296	26	2012-08-02 11:27:46.491424
43393	3	1-11011	43391	20	13	1915	92	2012-08-02 11:27:46.491424
43391	3	1-13600	43391	13	63	2308	1	2012-08-02 11:27:46.491424
43395	4	1-7640	43394	29	44	467	5	2012-08-02 11:27:46.491424
43396	4	1-13579	43394	44	35	341	29	2012-08-02 11:27:46.491424
43397	4	1-12656	43394	35	13	416	92	2012-08-02 11:27:46.491424
43394	4	1-13600	43394	13	29	535	1	2012-08-02 11:27:46.491424
43398	1	1-1518	\N	24	24	34	57	2012-08-02 11:27:46.491424
43422	4	1-12245	43421	50	43	1589	81	2012-08-02 11:27:52.435896
43423	4	1-13216	43421	43	89	2953	93	2012-08-02 11:27:52.435896
43424	4	1-13054	43421	89	51	916	95	2012-08-02 11:27:52.435896
43421	4	1-13615	43421	51	50	2007	69	2012-08-02 11:27:52.435896
43426	4	1-10765	43425	24	40	4420	70	2012-08-02 11:27:52.435896
43427	4	1-13570	43425	40	45	2806	45	2012-08-02 11:27:52.435896
43428	4	1-13588	43425	45	51	3100	95	2012-08-02 11:27:52.435896
43425	4	1-13615	43425	51	24	6857	69	2012-08-02 11:27:52.435896
43430	4	1-13581	43429	68	50	283	77	2012-08-02 11:27:54.733372
43431	4	1-10908	43429	50	24	85	34	2012-08-02 11:27:54.733372
43432	4	1-12596	43429	24	64	114	32	2012-08-02 11:27:54.733372
43429	4	1-13622	43429	64	68	248	5	2012-08-02 11:27:54.733372
43434	4	1-11624	43433	48	59	8	68	2012-08-02 11:27:54.733372
43435	4	1-12297	43433	59	24	8	34	2012-08-02 11:27:54.733372
43436	4	1-12596	43433	24	64	10	32	2012-08-02 11:27:54.733372
43433	4	1-13622	43433	64	48	22	5	2012-08-02 11:27:54.733372
43462	4	1-12596	43461	24	49	909	32	2012-08-02 11:27:56.258298
43463	4	1-13501	43461	49	74	797	11	2012-08-02 11:27:56.258298
43464	4	1-11449	43461	74	53	685	20	2012-08-02 11:27:56.258298
43461	4	1-13627	43461	53	24	710	34	2012-08-02 11:27:56.258298
43466	4	1-12150	43465	41	53	1507	86	2012-08-02 11:27:56.258298
43467	4	1-12710	43465	53	69	1321	57	2012-08-02 11:27:56.258298
43468	4	1-12780	43465	69	53	2509	20	2012-08-02 11:27:56.258298
43465	4	1-13627	43465	53	41	2726	34	2012-08-02 11:27:56.258298
43469	1	1-4905	\N	46	46	2602	34	2012-08-02 11:27:56.258298
43483	4	1-10805	43482	69	80	1809	58	2012-08-02 11:27:58.733665
43484	4	1-13227	43482	80	45	1174	26	2012-08-02 11:27:58.733665
43485	4	1-12365	43482	45	94	859	75	2012-08-02 11:27:58.733665
43482	4	1-13631	43482	94	69	4099	69	2012-08-02 11:27:58.733665
43502	4	1-13588	43501	45	67	3250	95	2012-08-02 11:28:00.396833
43503	4	1-7814	43501	67	9	721	100	2012-08-02 11:28:00.396833
43504	4	1-12441	43501	9	23	929	2	2012-08-02 11:28:00.396833
43501	4	1-13635	43501	23	45	2818	45	2012-08-02 11:28:00.396833
43510	4	1-12719	43509	25	6	2282	80	2012-08-02 11:28:06.634738
43511	4	1-11979	43509	6	72	1372	52	2012-08-02 11:28:06.634738
43512	4	1-13221	43509	72	54	1213	72	2012-08-02 11:28:06.634738
43509	4	1-13651	43509	54	25	1889	98	2012-08-02 11:28:06.634738
43514	4	1-9032	43513	87	50	2	66	2012-08-02 11:28:06.634738
43515	4	1-13449	43513	50	72	4	52	2012-08-02 11:28:06.634738
43516	4	1-13221	43513	72	54	4	72	2012-08-02 11:28:06.634738
43513	4	1-13651	43513	54	87	7	98	2012-08-02 11:28:06.634738
43532	4	1-13654	43532	1	78	767	65	2012-08-02 11:28:08.203628
43533	4	1-13123	43532	78	37	669	26	2012-08-02 11:28:08.203628
43534	4	1-3980	43532	37	45	178	73	2012-08-02 11:28:08.203628
43535	4	1-13557	43532	45	1	398	68	2012-08-02 11:28:08.203628
43537	3	1-12441	43536	9	12	2682	2	2012-08-02 11:28:09.375063
43538	3	1-12493	43536	12	75	1080	20	2012-08-02 11:28:09.375063
43536	3	1-13657	43536	75	9	3301	100	2012-08-02 11:28:09.375063
43540	4	1-13634	43539	10	78	383	96	2012-08-02 11:28:09.375063
43541	4	1-11206	43539	78	20	183	99	2012-08-02 11:28:09.375063
43542	4	1-12620	43539	20	75	105	20	2012-08-02 11:28:09.375063
43539	4	1-13657	43539	75	10	359	100	2012-08-02 11:28:09.375063
43544	4	1-13634	43543	10	42	435	96	2012-08-02 11:28:09.375063
43545	4	1-12271	43543	42	71	157	8	2012-08-02 11:28:09.375063
43546	4	1-12165	43543	71	75	114	20	2012-08-02 11:28:09.375063
43543	4	1-13657	43543	75	10	398	100	2012-08-02 11:28:09.375063
43548	4	1-11826	43547	45	66	464	74	2012-08-02 11:28:09.375063
43549	4	1-13598	43547	66	83	1232	7	2012-08-02 11:28:09.375063
43550	4	1-11603	43547	83	75	512	20	2012-08-02 11:28:09.375063
43547	4	1-13657	43547	75	45	2051	100	2012-08-02 11:28:09.375063
43552	4	1-12314	43551	21	60	593	15	2012-08-02 11:28:09.375063
43553	4	1-12602	43551	60	20	126	99	2012-08-02 11:28:09.375063
43554	4	1-12620	43551	20	75	88	20	2012-08-02 11:28:09.375063
43551	4	1-13657	43551	75	21	367	100	2012-08-02 11:28:09.375063
43556	3	1-10742	43555	5	10	273	47	2012-08-02 11:28:09.375063
43557	3	1-7430	43555	10	75	151	20	2012-08-02 11:28:09.375063
43555	3	1-13657	43555	75	5	655	100	2012-08-02 11:28:09.375063
43559	4	1-10049	43558	54	19	1004	77	2012-08-02 11:28:09.375063
43560	4	1-11856	43558	19	12	1033	2	2012-08-02 11:28:09.375063
43561	4	1-12493	43558	12	75	608	20	2012-08-02 11:28:09.375063
43558	4	1-13657	43558	75	54	2714	100	2012-08-02 11:28:09.375063
43565	4	1-13014	43564	71	90	2007	22	2012-08-02 11:28:10.568682
43566	4	1-12940	43564	90	95	2144	17	2012-08-02 11:28:10.568682
43567	4	1-12505	43564	95	12	2900	92	2012-08-02 11:28:10.568682
43564	4	1-13660	43564	12	71	2251	59	2012-08-02 11:28:10.568682
43580	4	1-11284	43579	73	47	1939	77	2012-08-02 11:28:12.888559
43581	4	1-13602	43579	47	92	815	10	2012-08-02 11:28:12.888559
43582	4	1-12919	43579	92	37	1124	84	2012-08-02 11:28:12.888559
43579	4	1-13667	43579	37	73	1423	24	2012-08-02 11:28:12.888559
43584	4	1-12835	43583	41	28	1789	14	2012-08-02 11:28:12.888559
43585	4	1-13653	43583	28	92	1063	10	2012-08-02 11:28:12.888559
43586	4	1-12919	43583	92	37	1548	84	2012-08-02 11:28:12.888559
43583	4	1-13667	43583	37	41	2072	24	2012-08-02 11:28:12.888559
43588	2	1-11737	43587	80	37	3199	84	2012-08-02 11:28:12.888559
43587	2	1-13667	43587	37	80	4503	24	2012-08-02 11:28:12.888559
43590	4	1-13514	43589	55	34	2577	59	2012-08-02 11:28:14.457753
43591	4	1-12183	43589	34	84	1827	11	2012-08-02 11:28:14.457753
43592	4	1-8059	43589	84	59	1101	38	2012-08-02 11:28:14.457753
43589	4	1-13670	43589	59	55	2383	67	2012-08-02 11:28:14.457753
43594	4	1-12473	43593	62	59	817	63	2012-08-02 11:28:14.457753
43595	4	1-13592	43593	59	74	2683	97	2012-08-02 11:28:14.457753
43596	4	1-13612	43593	74	59	1416	38	2012-08-02 11:28:14.457753
43593	4	1-13670	43593	59	62	3091	67	2012-08-02 11:28:14.457753
43598	3	1-13514	43597	55	87	1416	59	2012-08-02 11:28:14.457753
43599	3	1-11016	43597	87	59	539	38	2012-08-02 11:28:14.457753
43597	3	1-13670	43597	59	55	1236	67	2012-08-02 11:28:14.457753
43601	3	1-974	43600	54	83	160	73	2012-08-02 11:28:14.457753
43602	3	1-12944	43600	83	59	278	38	2012-08-02 11:28:14.457753
43600	3	1-13670	43600	59	54	705	67	2012-08-02 11:28:14.457753
43604	4	1-8528	43603	100	78	1281	31	2012-08-02 11:28:14.457753
43605	4	1-13291	43603	78	74	1037	97	2012-08-02 11:28:14.457753
43606	4	1-13612	43603	74	59	830	38	2012-08-02 11:28:14.457753
43603	4	1-13670	43603	59	100	2168	67	2012-08-02 11:28:14.457753
43620	4	1-9058	43619	45	39	65	4	2012-08-02 11:28:16.811248
43621	4	1-12238	43619	39	89	84	83	2012-08-02 11:28:16.811248
43622	4	1-13399	43619	89	59	84	24	2012-08-02 11:28:16.811248
43619	4	1-13678	43619	59	45	94	97	2012-08-02 11:28:16.811248
43630	4	1-13175	43629	4	39	5331	25	2012-08-02 11:28:20.347033
43631	4	1-12380	43629	39	6	2036	62	2012-08-02 11:28:20.347033
43632	4	1-13573	43629	6	2	3525	40	2012-08-02 11:28:20.347033
43629	4	1-13685	43629	2	4	3381	12	2012-08-02 11:28:20.347033
43634	4	1-13661	43633	88	58	878	21	2012-08-02 11:28:20.347033
43635	4	1-13029	43633	58	69	818	48	2012-08-02 11:28:20.347033
43636	4	1-12133	43633	69	2	729	40	2012-08-02 11:28:20.347033
43633	4	1-13685	43633	2	88	902	12	2012-08-02 11:28:20.347033
43662	4	1-13310	43661	55	80	1812	68	2012-08-02 11:28:23.097742
43663	4	1-12285	43661	80	62	3154	60	2012-08-02 11:28:23.097742
43664	4	1-12103	43661	62	90	1898	21	2012-08-02 11:28:23.097742
43661	4	1-13692	43661	90	55	2388	35	2012-08-02 11:28:23.097742
43666	4	1-13429	43665	30	72	7640	40	2012-08-02 11:28:23.097742
43667	4	1-12361	43665	72	13	2029	18	2012-08-02 11:28:23.097742
43668	4	1-11810	43665	13	90	2414	21	2012-08-02 11:28:23.097742
43665	4	1-13692	43665	90	30	3046	35	2012-08-02 11:28:23.097742
43678	2	1-12059	43677	94	37	1068	23	2012-08-02 11:28:26.600446
43677	2	1-13701	43677	37	94	1979	17	2012-08-02 11:28:26.600446
43694	4	1-11854	43693	57	83	524	7	2012-08-02 11:28:30.922001
43695	4	1-11603	43693	83	66	238	20	2012-08-02 11:28:30.922001
43696	4	1-12254	43693	66	35	248	71	2012-08-02 11:28:30.922001
43693	4	1-13711	43693	35	57	611	58	2012-08-02 11:28:30.922001
43698	4	1-13227	43697	80	37	837	26	2012-08-02 11:28:30.922001
43699	4	1-3980	43697	37	36	192	73	2012-08-02 11:28:30.922001
43700	4	1-12768	43697	36	35	540	71	2012-08-02 11:28:30.922001
43697	4	1-13711	43697	35	80	1351	58	2012-08-02 11:28:30.922001
43702	4	1-13369	43701	2	69	4552	51	2012-08-02 11:28:30.922001
43703	4	1-10890	43701	69	8	2040	48	2012-08-02 11:28:30.922001
43704	4	1-13377	43701	8	35	1696	71	2012-08-02 11:28:30.922001
43701	4	1-13711	43701	35	2	4653	58	2012-08-02 11:28:30.922001
43706	3	1-7383	43705	90	66	3	20	2012-08-02 11:28:30.922001
43707	3	1-12254	43705	66	35	4	71	2012-08-02 11:28:30.922001
43705	3	1-13711	43705	35	90	13	58	2012-08-02 11:28:30.922001
43717	4	1-12885	43716	71	5	2628	33	2012-08-02 11:28:31.707131
43718	4	1-12748	43716	5	58	8203	14	2012-08-02 11:28:31.707131
43719	4	1-13697	43716	58	62	4424	73	2012-08-02 11:28:31.707131
43674	4	1-12189	43673	40	80	1318	58	2012-08-02 11:28:25.826838
43675	4	1-13227	43673	80	43	973	26	2012-08-02 11:28:25.826838
43676	4	1-13624	43673	43	23	1685	30	2012-08-02 11:28:25.826838
43673	4	1-13699	43673	23	40	6305	21	2012-08-02 11:28:25.826838
43687	4	1-10047	43686	29	72	990	15	2012-08-02 11:28:30.157943
43688	4	1-13519	43686	72	51	1095	52	2012-08-02 11:28:30.157943
43689	4	1-13628	43686	51	75	1090	46	2012-08-02 11:28:30.157943
43686	4	1-13710	43686	75	29	2122	24	2012-08-02 11:28:30.157943
43691	3	1-13678	43690	59	47	5799	97	2012-08-02 11:28:30.157943
43692	3	1-11635	43690	47	75	2810	46	2012-08-02 11:28:30.157943
43690	3	1-13710	43690	75	59	5802	24	2012-08-02 11:28:30.157943
43727	4	1-13147	43726	28	33	1085	62	2012-08-02 11:28:32.889244
43728	4	1-12777	43726	33	54	1248	72	2012-08-02 11:28:32.889244
43729	4	1-13651	43726	54	66	2242	98	2012-08-02 11:28:32.889244
43726	4	1-13717	43726	66	28	1659	6	2012-08-02 11:28:32.889244
43731	2	1-10536	43730	45	78	4697	2	2012-08-02 11:28:34.027524
43730	2	1-13720	43730	78	45	6776	73	2012-08-02 11:28:34.027524
43733	4	1-13540	43732	65	66	984	56	2012-08-02 11:28:34.027524
43734	4	1-10936	43732	66	43	918	25	2012-08-02 11:28:34.027524
43735	4	1-13367	43732	43	78	737	2	2012-08-02 11:28:34.027524
43732	4	1-13720	43732	78	65	1239	73	2012-08-02 11:28:34.027524
43764	3	1-12410	43763	21	41	810	18	2012-08-02 11:28:41.71681
43765	3	1-12843	43763	41	89	1040	54	2012-08-02 11:28:41.71681
43763	3	1-13739	43763	89	21	5373	35	2012-08-02 11:28:41.71681
43770	4	1-13614	43769	69	35	2031	33	2012-08-02 11:28:42.910342
43771	4	1-13294	43769	35	61	1051	63	2012-08-02 11:28:42.910342
43772	4	1-12913	43769	61	72	709	56	2012-08-02 11:28:42.910342
43769	4	1-13741	43769	72	69	8064	50	2012-08-02 11:28:42.910342
43774	4	1-12032	43773	47	99	1106	18	2012-08-02 11:28:43.735007
43775	4	1-10891	43773	99	38	910	97	2012-08-02 11:28:43.735007
43776	4	1-12795	43773	38	58	4144	15	2012-08-02 11:28:43.735007
43773	4	1-13744	43773	58	47	3065	14	2012-08-02 11:28:43.735007
43802	4	1-13408	43801	78	53	892	86	2012-08-02 11:28:44.878308
43803	4	1-12710	43801	53	24	672	57	2012-08-02 11:28:44.878308
43804	4	1-12451	43801	24	27	255	26	2012-08-02 11:28:44.878308
43801	4	1-13747	43801	27	78	1039	95	2012-08-02 11:28:44.878308
43806	4	1-13408	43805	78	56	2580	86	2012-08-02 11:28:44.878308
43807	4	1-13406	43805	56	68	828	36	2012-08-02 11:28:44.878308
43808	4	1-11944	43805	68	27	727	26	2012-08-02 11:28:44.878308
43805	4	1-13747	43805	27	78	2984	95	2012-08-02 11:28:44.878308
43813	2	1-12406	43812	28	2	1454	34	2012-08-02 11:28:47.243134
43812	2	1-13753	43812	2	28	1465	47	2012-08-02 11:28:47.243134
43815	4	1-12871	43814	4	71	2262	8	2012-08-02 11:28:47.243134
43816	4	1-12165	43814	71	53	2555	20	2012-08-02 11:28:47.243134
43817	4	1-13627	43814	53	2	2829	34	2012-08-02 11:28:47.243134
43814	4	1-13753	43814	2	4	3530	47	2012-08-02 11:28:47.243134
43819	3	1-7430	43818	10	53	1612	20	2012-08-02 11:28:47.243134
43820	3	1-13627	43818	53	2	1809	34	2012-08-02 11:28:47.243134
43818	3	1-13753	43818	2	10	2287	47	2012-08-02 11:28:47.243134
43829	4	1-10478	43828	60	80	3259	40	2012-08-02 11:28:49.552546
43830	4	1-13723	43828	80	32	3396	84	2012-08-02 11:28:49.552546
43831	4	1-10568	43828	32	6	2400	50	2012-08-02 11:28:49.552546
43828	4	1-13759	43828	6	60	3757	32	2012-08-02 11:28:49.552546
43863	4	1-13708	43862	79	26	451	73	2012-08-02 11:28:53.156625
43864	4	1-13599	43862	26	46	525	91	2012-08-02 11:28:53.156625
43865	4	1-13113	43862	46	66	356	11	2012-08-02 11:28:53.156625
43862	4	1-13768	43862	66	79	1030	94	2012-08-02 11:28:53.156625
43867	4	1-12786	43866	3	94	152	29	2012-08-02 11:28:53.156625
43868	4	1-13327	43866	94	68	198	41	2012-08-02 11:28:53.156625
43869	4	1-7740	43866	68	66	56	11	2012-08-02 11:28:53.156625
43866	4	1-13768	43866	66	3	186	94	2012-08-02 11:28:53.156625
43871	4	1-12404	43870	99	50	459	10	2012-08-02 11:28:53.156625
43872	4	1-13058	43870	50	46	200	91	2012-08-02 11:28:53.156625
43873	4	1-13113	43870	46	66	161	11	2012-08-02 11:28:53.156625
43870	4	1-13768	43870	66	99	552	94	2012-08-02 11:28:53.156625
43875	3	1-9186	43874	65	46	1012	91	2012-08-02 11:28:53.156625
43876	3	1-13113	43874	46	66	821	11	2012-08-02 11:28:53.156625
43874	3	1-13768	43874	66	65	2845	94	2012-08-02 11:28:53.156625
43878	4	1-12786	43877	3	82	1407	29	2012-08-02 11:28:53.156625
43879	4	1-10423	43877	82	46	459	91	2012-08-02 11:28:53.156625
43880	4	1-13113	43877	46	66	405	11	2012-08-02 11:28:53.156625
43877	4	1-13768	43877	66	3	1525	94	2012-08-02 11:28:53.156625
43909	3	1-8892	43908	81	92	241	31	2012-08-02 11:28:55.123393
43910	3	1-13526	43908	92	86	590	90	2012-08-02 11:28:55.123393
43908	3	1-13773	43908	86	81	1511	21	2012-08-02 11:28:55.123393
43912	3	1-13034	43911	71	68	518	46	2012-08-02 11:28:55.123393
43913	3	1-13453	43911	68	86	554	90	2012-08-02 11:28:55.123393
43911	3	1-13773	43911	86	71	2185	21	2012-08-02 11:28:55.123393
43955	4	1-13091	43954	80	24	1919	43	2012-08-02 11:29:03.420622
43956	4	1-13340	43954	24	47	1477	63	2012-08-02 11:29:03.420622
43957	4	1-13150	43954	47	42	1583	3	2012-08-02 11:29:03.420622
43954	4	1-13795	43954	42	80	1581	18	2012-08-02 11:29:03.420622
43962	4	1-12621	43961	88	25	3306	27	2012-08-02 11:29:06.182748
43963	4	1-11159	43961	25	78	2924	13	2012-08-02 11:29:06.182748
43964	4	1-12902	43961	78	38	1811	21	2012-08-02 11:29:06.182748
43961	4	1-13800	43961	38	88	2955	97	2012-08-02 11:29:06.182748
43968	2	1-12361	43967	72	49	201	18	2012-08-02 11:29:06.945209
43967	2	1-13802	43967	49	72	3237	40	2012-08-02 11:29:06.945209
43970	3	1-12968	43969	94	63	1485	32	2012-08-02 11:29:06.945209
43971	3	1-5018	43969	63	49	181	18	2012-08-02 11:29:06.945209
43969	3	1-13802	43969	49	94	5168	40	2012-08-02 11:29:06.945209
43985	2	1-5601	43984	4	6	1965	88	2012-08-02 11:29:11.276055
43984	2	1-13814	43984	6	4	3816	65	2012-08-02 11:29:11.276055
43987	4	1-10866	43986	30	100	385	9	2012-08-02 11:29:11.276055
43988	4	1-13644	43986	100	71	969	92	2012-08-02 11:29:11.276055
43989	4	1-13785	43986	71	6	475	88	2012-08-02 11:29:11.276055
43986	4	1-13814	43986	6	30	1008	65	2012-08-02 11:29:11.276055
43999	4	1-13777	43998	34	46	6870	76	2012-08-02 11:29:12.470004
44000	4	1-11735	43998	46	6	7087	85	2012-08-02 11:29:12.470004
44001	4	1-13617	43998	6	15	3697	25	2012-08-02 11:29:12.470004
43998	4	1-13817	43998	15	34	5762	48	2012-08-02 11:29:12.470004
44003	4	1-13088	44002	17	72	221	91	2012-08-02 11:29:12.470004
44004	4	1-13396	44002	72	90	259	44	2012-08-02 11:29:12.470004
44005	4	1-9369	44002	90	15	147	25	2012-08-02 11:29:12.470004
43716	4	1-13713	43716	62	71	3877	47	2012-08-02 11:28:31.707131
43721	3	1-13498	43720	92	30	1448	35	2012-08-02 11:28:32.513694
43722	3	1-13429	43720	30	51	2292	40	2012-08-02 11:28:32.513694
43720	3	1-13716	43720	51	92	2443	31	2012-08-02 11:28:32.513694
43724	3	1-12812	43723	72	69	2200	48	2012-08-02 11:28:32.513694
43725	3	1-12133	43723	69	51	1552	40	2012-08-02 11:28:32.513694
43723	3	1-13716	43723	51	72	1864	31	2012-08-02 11:28:32.513694
43755	2	1-10405	43754	19	58	939	81	2012-08-02 11:28:35.983455
43754	2	1-13725	43754	58	19	1283	27	2012-08-02 11:28:35.983455
43757	3	1-13108	43756	31	16	6694	69	2012-08-02 11:28:35.983455
43758	3	1-12250	43756	16	58	4466	81	2012-08-02 11:28:35.983455
43756	3	1-13725	43756	58	31	6344	27	2012-08-02 11:28:35.983455
43760	4	1-13347	43759	94	61	5260	100	2012-08-02 11:28:40.943643
43761	4	1-13447	43759	61	100	2208	9	2012-08-02 11:28:40.943643
43762	4	1-13644	43759	100	76	3103	92	2012-08-02 11:28:40.943643
43759	4	1-13737	43759	76	94	9750	51	2012-08-02 11:28:40.943643
43767	3	1-13298	43766	48	64	651	66	2012-08-02 11:28:42.103973
43768	3	1-10987	43766	64	34	301	28	2012-08-02 11:28:42.103973
43766	3	1-13740	43766	34	48	782	51	2012-08-02 11:28:42.103973
43778	4	1-11619	43777	42	94	41	51	2012-08-02 11:28:44.502577
43779	4	1-13347	43777	94	6	36	100	2012-08-02 11:28:44.502577
43780	4	1-12900	43777	6	79	42	56	2012-08-02 11:28:44.502577
43777	4	1-13746	43777	79	42	119	79	2012-08-02 11:28:44.502577
43782	4	1-13402	43781	35	98	1263	8	2012-08-02 11:28:44.502577
43783	4	1-6566	43781	98	65	324	73	2012-08-02 11:28:44.502577
43784	4	1-13540	43781	65	79	320	56	2012-08-02 11:28:44.502577
43781	4	1-13746	43781	79	35	1016	79	2012-08-02 11:28:44.502577
43786	4	1-12542	43785	23	43	1528	81	2012-08-02 11:28:44.502577
43787	4	1-13216	43785	43	20	3330	93	2012-08-02 11:28:44.502577
43788	4	1-9974	43785	20	79	835	56	2012-08-02 11:28:44.502577
43785	4	1-13746	43785	79	23	2697	79	2012-08-02 11:28:44.502577
43790	4	1-13488	43789	13	35	606	54	2012-08-02 11:28:44.502577
43791	4	1-8736	43789	35	26	128	10	2012-08-02 11:28:44.502577
43792	4	1-11426	43789	26	79	58	56	2012-08-02 11:28:44.502577
43789	4	1-13746	43789	79	13	200	79	2012-08-02 11:28:44.502577
43794	4	1-12994	43793	20	79	391	23	2012-08-02 11:28:44.502577
43795	4	1-13398	43793	79	6	397	100	2012-08-02 11:28:44.502577
43796	4	1-12900	43793	6	79	561	56	2012-08-02 11:28:44.502577
43793	4	1-13746	43793	79	20	1935	79	2012-08-02 11:28:44.502577
43798	4	1-13112	43797	36	45	1060	7	2012-08-02 11:28:44.502577
43799	4	1-6895	43797	45	75	285	18	2012-08-02 11:28:44.502577
43800	4	1-6894	43797	75	79	110	56	2012-08-02 11:28:44.502577
43797	4	1-13746	43797	79	36	381	79	2012-08-02 11:28:44.502577
43810	3	1-13088	43809	17	67	2693	91	2012-08-02 11:28:45.674267
43811	3	1-13604	43809	67	58	2394	55	2012-08-02 11:28:45.674267
43809	3	1-13749	43809	58	17	3246	48	2012-08-02 11:28:45.674267
43837	4	1-11030	43836	100	69	43	48	2012-08-02 11:28:50.33668
43838	4	1-12133	43836	69	94	26	40	2012-08-02 11:28:50.33668
43839	4	1-12968	43836	94	3	11	32	2012-08-02 11:28:50.33668
43836	4	1-13761	43836	3	100	184	44	2012-08-02 11:28:50.33668
43841	4	1-9424	43840	95	63	1912	12	2012-08-02 11:28:50.33668
43842	4	1-13607	43840	63	9	953	50	2012-08-02 11:28:50.33668
43843	4	1-12884	43840	9	3	488	32	2012-08-02 11:28:50.33668
43840	4	1-13761	43840	3	95	8762	44	2012-08-02 11:28:50.33668
43927	4	1-11150	43926	21	80	4092	40	2012-08-02 11:28:59.111716
43928	4	1-13723	43926	80	27	4140	84	2012-08-02 11:28:59.111716
43929	4	1-13394	43926	27	94	2381	43	2012-08-02 11:28:59.111716
43926	4	1-13783	43926	94	21	4315	30	2012-08-02 11:28:59.111716
43931	3	1-13019	43930	97	100	2161	75	2012-08-02 11:28:59.111716
43932	3	1-10793	43930	100	94	1168	43	2012-08-02 11:28:59.111716
43930	3	1-13783	43930	94	97	2174	30	2012-08-02 11:28:59.111716
43942	3	1-13487	43941	63	27	2996	20	2012-08-02 11:29:02.658287
43943	3	1-10726	43941	27	47	833	28	2012-08-02 11:29:02.658287
43941	3	1-13792	43941	47	63	2534	30	2012-08-02 11:29:02.658287
43945	4	1-13019	43944	97	42	703	75	2012-08-02 11:29:02.658287
43946	4	1-10106	43944	42	27	755	20	2012-08-02 11:29:02.658287
43947	4	1-10726	43944	27	47	218	28	2012-08-02 11:29:02.658287
43944	4	1-13792	43944	47	97	691	30	2012-08-02 11:29:02.658287
43949	4	1-10242	43948	55	22	519	80	2012-08-02 11:29:02.658287
43950	4	1-13499	43948	22	28	778	91	2012-08-02 11:29:02.658287
43951	4	1-8858	43948	28	47	208	28	2012-08-02 11:29:02.658287
43948	4	1-13792	43948	47	55	686	30	2012-08-02 11:29:02.658287
43952	1	1-10106	\N	42	42	3630	20	2012-08-02 11:29:02.658287
43953	1	1-8858	\N	28	28	12	28	2012-08-02 11:29:02.658287
43966	2	1-10610	43965	20	53	1189	32	2012-08-02 11:29:05.796114
43965	2	1-13801	43965	53	20	8520	24	2012-08-02 11:29:05.796114
43995	4	1-12250	43994	16	95	1837	81	2012-08-02 11:29:12.082503
43996	4	1-13336	43994	95	9	1620	50	2012-08-02 11:29:12.082503
43997	4	1-12884	43994	9	29	1280	32	2012-08-02 11:29:12.082503
43994	4	1-13816	43994	29	16	2575	69	2012-08-02 11:29:12.082503
44022	4	1-13758	44021	95	12	8045	92	2012-08-02 11:29:15.596151
44023	4	1-13660	44021	12	82	5181	59	2012-08-02 11:29:15.596151
44024	4	1-9314	44021	82	23	2154	81	2012-08-02 11:29:15.596151
44021	4	1-13825	44021	23	95	2983	18	2012-08-02 11:29:15.596151
44026	3	1-12608	44025	45	26	1252	8	2012-08-02 11:29:15.596151
44027	3	1-11295	44025	26	23	751	81	2012-08-02 11:29:15.596151
44025	3	1-13825	44025	23	45	1182	18	2012-08-02 11:29:15.596151
44029	4	1-12843	44028	41	79	1478	54	2012-08-02 11:29:15.596151
44030	4	1-12115	44028	79	73	390	43	2012-08-02 11:29:15.596151
44031	4	1-11406	44028	73	23	233	81	2012-08-02 11:29:15.596151
44028	4	1-13825	44028	23	41	400	18	2012-08-02 11:29:15.596151
44033	4	1-13262	44032	24	14	2640	89	2012-08-02 11:29:15.596151
44034	4	1-13496	44032	14	97	2446	23	2012-08-02 11:29:15.596151
44035	4	1-12749	44032	97	23	2046	81	2012-08-02 11:29:15.596151
44032	4	1-13825	44032	23	24	3880	18	2012-08-02 11:29:15.596151
44037	4	1-10500	44036	5	9	97	10	2012-08-02 11:29:15.596151
44038	4	1-12960	44036	9	97	97	23	2012-08-02 11:29:15.596151
44039	4	1-12749	44036	97	23	82	81	2012-08-02 11:29:15.596151
44036	4	1-13825	44036	23	5	156	18	2012-08-02 11:29:15.596151
44076	4	1-13758	44075	95	89	233	92	2012-08-02 11:29:19.178344
44075	4	1-13833	44075	54	95	86	18	2012-08-02 11:29:19.178344
43745	4	1-13067	43742	58	62	55	62	2012-08-02 11:28:35.596534
43747	4	1-13291	43746	78	39	12	97	2012-08-02 11:28:35.596534
43748	4	1-12713	43746	39	33	8	74	2012-08-02 11:28:35.596534
43749	4	1-13473	43746	33	62	4	62	2012-08-02 11:28:35.596534
43746	4	1-13724	43746	62	78	27	31	2012-08-02 11:28:35.596534
43751	4	1-13498	43750	92	72	2826	35	2012-08-02 11:28:35.596534
43752	4	1-8979	43750	72	58	1340	29	2012-08-02 11:28:35.596534
43753	4	1-13067	43750	58	62	639	62	2012-08-02 11:28:35.596534
43750	4	1-13724	43750	62	92	5036	31	2012-08-02 11:28:35.596534
43822	3	1-13343	43821	56	6	2641	43	2012-08-02 11:28:48.812571
43823	3	1-13196	43821	6	40	1699	46	2012-08-02 11:28:48.812571
43821	3	1-13757	43821	40	56	5304	48	2012-08-02 11:28:48.812571
43825	4	1-13088	43824	17	28	709	91	2012-08-02 11:28:48.812571
43826	4	1-8858	43824	28	42	190	28	2012-08-02 11:28:48.812571
43827	4	1-13272	43824	42	40	236	46	2012-08-02 11:28:48.812571
43824	4	1-13757	43824	40	17	766	48	2012-08-02 11:28:48.812571
43833	4	1-13332	43832	55	1	1446	68	2012-08-02 11:28:49.950697
43834	4	1-13654	43832	1	64	2888	65	2012-08-02 11:28:49.950697
43835	4	1-12642	43832	64	3	4953	53	2012-08-02 11:28:49.950697
43832	4	1-13760	43832	3	55	3002	62	2012-08-02 11:28:49.950697
43845	2	1-12482	43844	47	48	3735	42	2012-08-02 11:28:51.13227
43844	2	1-13763	43844	48	47	9812	67	2012-08-02 11:28:51.13227
43847	4	1-8528	43846	100	42	5	31	2012-08-02 11:28:51.13227
43848	4	1-13344	43846	42	40	9	87	2012-08-02 11:28:51.13227
43849	4	1-13422	43846	40	48	4	42	2012-08-02 11:28:51.13227
43846	4	1-13763	43846	48	100	14	67	2012-08-02 11:28:51.13227
43851	4	1-3662	43850	49	9	1595	60	2012-08-02 11:28:52.382745
43852	4	1-13703	43850	9	9	2305	83	2012-08-02 11:28:52.382745
43853	4	1-13339	43850	9	17	2559	71	2012-08-02 11:28:52.382745
43850	4	1-13766	43850	17	49	2307	35	2012-08-02 11:28:52.382745
43855	4	1-8979	43854	72	85	287	29	2012-08-02 11:28:52.382745
43856	4	1-13431	43854	85	66	329	20	2012-08-02 11:28:52.382745
43857	4	1-12254	43854	66	17	418	71	2012-08-02 11:28:52.382745
43854	4	1-13766	43854	17	72	382	35	2012-08-02 11:28:52.382745
43859	4	1-13507	43858	17	12	1856	84	2012-08-02 11:28:52.382745
43860	4	1-10788	43858	12	8	1215	48	2012-08-02 11:28:52.382745
43861	4	1-13377	43858	8	17	1103	71	2012-08-02 11:28:52.382745
43858	4	1-13766	43858	17	17	1048	35	2012-08-02 11:28:52.382745
43882	4	1-7449	43881	9	9	2303	60	2012-08-02 11:28:53.543122
43883	4	1-13703	43881	9	9	2940	83	2012-08-02 11:28:53.543122
43884	4	1-13339	43881	9	89	2882	71	2012-08-02 11:28:53.543122
43881	4	1-13769	43881	89	9	2820	62	2012-08-02 11:28:53.543122
43886	4	1-13573	43885	6	21	2216	40	2012-08-02 11:28:53.543122
43887	4	1-11981	43885	21	66	574	20	2012-08-02 11:28:53.543122
43888	4	1-12254	43885	66	89	693	71	2012-08-02 11:28:53.543122
43885	4	1-13769	43885	89	6	993	62	2012-08-02 11:28:53.543122
43890	3	1-12455	43889	38	8	1221	48	2012-08-02 11:28:53.543122
43891	3	1-13377	43889	8	89	1137	71	2012-08-02 11:28:53.543122
43889	3	1-13769	43889	89	38	1755	62	2012-08-02 11:28:53.543122
43893	4	1-13259	43892	91	38	518	24	2012-08-02 11:28:53.543122
43894	4	1-8619	43892	38	36	329	73	2012-08-02 11:28:53.543122
43895	4	1-12768	43892	36	89	1175	71	2012-08-02 11:28:53.543122
43892	4	1-13769	43892	89	91	1879	62	2012-08-02 11:28:53.543122
43897	4	1-12642	43896	64	73	213	53	2012-08-02 11:28:54.736808
43898	4	1-13427	43896	73	22	183	89	2012-08-02 11:28:54.736808
43899	4	1-13207	43896	22	37	246	4	2012-08-02 11:28:54.736808
43896	4	1-13772	43896	37	64	186	65	2012-08-02 11:28:54.736808
43901	4	1-12642	43900	64	73	1245	53	2012-08-02 11:28:54.736808
43902	4	1-13427	43900	73	43	1336	89	2012-08-02 11:28:54.736808
43903	4	1-13474	43900	43	37	920	4	2012-08-02 11:28:54.736808
43900	4	1-13772	43900	37	64	869	65	2012-08-02 11:28:54.736808
43905	4	1-12642	43904	64	80	228	53	2012-08-02 11:28:54.736808
43906	4	1-12722	43904	80	35	231	80	2012-08-02 11:28:54.736808
43907	4	1-12634	43904	35	37	131	4	2012-08-02 11:28:54.736808
43904	4	1-13772	43904	37	64	140	65	2012-08-02 11:28:54.736808
43915	3	1-11582	43914	47	94	1965	40	2012-08-02 11:28:55.930098
43916	3	1-12968	43914	94	78	1303	32	2012-08-02 11:28:55.930098
43914	3	1-13775	43914	78	47	1635	20	2012-08-02 11:28:55.930098
43918	2	1-12746	43917	38	97	2021	37	2012-08-02 11:28:58.338996
43917	2	1-13781	43917	97	38	1757	99	2012-08-02 11:28:58.338996
43920	4	1-13293	43919	57	28	1039	61	2012-08-02 11:28:58.338996
43921	4	1-5483	43919	28	94	534	48	2012-08-02 11:28:58.338996
43922	4	1-12783	43919	94	97	1015	37	2012-08-02 11:28:58.338996
43919	4	1-13781	43919	97	57	1098	99	2012-08-02 11:28:58.338996
43924	3	1-12620	43923	20	37	104	20	2012-08-02 11:28:58.338996
43925	3	1-13072	43923	37	97	97	37	2012-08-02 11:28:58.338996
43923	3	1-13781	43923	97	20	119	99	2012-08-02 11:28:58.338996
43934	4	1-13369	43933	2	48	2288	51	2012-08-02 11:29:01.863128
43935	4	1-13298	43933	48	82	1028	66	2012-08-02 11:29:01.863128
43936	4	1-13318	43933	82	61	617	50	2012-08-02 11:29:01.863128
43933	4	1-13790	43933	61	2	3674	58	2012-08-02 11:29:01.863128
43938	4	1-13734	43937	58	83	2092	38	2012-08-02 11:29:01.863128
43939	4	1-9673	43937	83	95	575	81	2012-08-02 11:29:01.863128
43940	4	1-13336	43937	95	61	305	50	2012-08-02 11:29:01.863128
43937	4	1-13790	43937	61	58	1946	58	2012-08-02 11:29:01.863128
43959	3	1-13665	43958	16	38	7426	16	2012-08-02 11:29:05.409457
43960	3	1-13437	43958	38	51	2587	38	2012-08-02 11:29:05.409457
43958	3	1-13799	43958	51	16	7442	75	2012-08-02 11:29:05.409457
43973	4	1-12968	43972	94	64	1107	32	2012-08-02 11:29:09.320977
43974	4	1-13622	43972	64	68	1383	5	2012-08-02 11:29:09.320977
43975	4	1-13581	43972	68	57	903	77	2012-08-02 11:29:09.320977
43972	4	1-13809	43972	57	94	2956	40	2012-08-02 11:29:09.320977
43977	4	1-13723	43976	80	69	1098	84	2012-08-02 11:29:09.320977
43978	4	1-9124	43976	69	68	250	5	2012-08-02 11:29:09.320977
43979	4	1-13581	43976	68	57	246	77	2012-08-02 11:29:09.320977
43976	4	1-13809	43976	57	80	1211	40	2012-08-02 11:29:09.320977
43981	4	1-12736	43980	97	55	519	95	2012-08-02 11:29:10.513766
43982	4	1-13686	43980	55	49	2055	77	2012-08-02 11:29:10.513766
43983	4	1-13376	43980	49	41	1147	6	2012-08-02 11:29:10.513766
43980	4	1-13812	43980	41	97	2620	71	2012-08-02 11:29:10.513766
43991	4	1-11528	43990	78	41	2272	6	2012-08-02 11:29:11.695894
43992	4	1-13812	43990	41	17	4484	71	2012-08-02 11:29:11.695894
43993	4	1-13766	43990	17	37	3668	35	2012-08-02 11:29:11.695894
43990	4	1-13815	43990	37	78	3909	79	2012-08-02 11:29:11.695894
44006	4	1-13818	44006	53	66	2369	29	2012-08-02 11:29:12.856069
44002	4	1-13817	44002	15	17	248	48	2012-08-02 11:29:12.470004
44041	4	1-13146	44040	63	43	2593	26	2012-08-02 11:29:15.971766
44042	4	1-13624	44040	43	21	3882	30	2012-08-02 11:29:15.971766
44043	4	1-11150	44040	21	60	3285	40	2012-08-02 11:29:15.971766
44040	4	1-13826	44040	60	63	3015	1	2012-08-02 11:29:15.971766
44045	4	1-13421	44044	37	43	373	81	2012-08-02 11:29:15.971766
44046	4	1-13216	44044	43	11	715	93	2012-08-02 11:29:15.971766
44047	4	1-11494	44044	11	60	533	40	2012-08-02 11:29:15.971766
44044	4	1-13826	44044	60	37	507	1	2012-08-02 11:29:15.971766
44049	4	1-4139	44048	52	48	788	2	2012-08-02 11:29:15.971766
44050	4	1-13569	44048	48	70	869	34	2012-08-02 11:29:15.971766
44051	4	1-13195	44048	70	60	1135	40	2012-08-02 11:29:15.971766
44048	4	1-13826	44048	60	52	1085	1	2012-08-02 11:29:15.971766
44053	4	1-13786	44052	48	50	772	85	2012-08-02 11:29:15.971766
44054	4	1-10161	44052	50	70	200	34	2012-08-02 11:29:15.971766
44055	4	1-13195	44052	70	60	314	40	2012-08-02 11:29:15.971766
44052	4	1-13826	44052	60	48	360	1	2012-08-02 11:29:15.971766
44057	4	1-10048	44056	62	80	177	55	2012-08-02 11:29:17.18967
44058	4	1-13480	44056	80	52	167	72	2012-08-02 11:29:17.18967
44059	4	1-13193	44056	52	69	60	25	2012-08-02 11:29:17.18967
44056	4	1-13829	44056	69	62	428	57	2012-08-02 11:29:17.18967
44061	3	1-10134	44060	91	6	6	85	2012-08-02 11:29:17.18967
44062	3	1-13617	44060	6	69	2	25	2012-08-02 11:29:17.18967
44060	3	1-13829	44060	69	91	18	57	2012-08-02 11:29:17.18967
44068	4	1-13792	44067	47	11	371	30	2012-08-02 11:29:18.371887
44069	4	1-13018	44067	11	48	182	12	2012-08-02 11:29:18.371887
44070	4	1-13822	44067	48	30	187	10	2012-08-02 11:29:18.371887
44067	4	1-13832	44067	30	47	130	28	2012-08-02 11:29:18.371887
44072	4	1-10720	44071	59	17	1330	71	2012-08-02 11:29:18.371887
44073	4	1-13766	44071	17	96	1149	35	2012-08-02 11:29:18.371887
44074	4	1-13807	44071	96	30	1690	10	2012-08-02 11:29:18.371887
44071	4	1-13832	44071	30	59	1217	28	2012-08-02 11:29:18.371887
44131	2	1-11102	44130	99	71	703	65	2012-08-02 11:29:22.394404
44130	2	1-13842	44130	71	99	6952	49	2012-08-02 11:29:22.394404
44136	4	1-9974	44135	20	79	174	56	2012-08-02 11:29:24.79138
44137	4	1-13746	44135	79	36	489	79	2012-08-02 11:29:24.79138
44138	4	1-13112	44135	36	70	1105	7	2012-08-02 11:29:24.79138
44135	4	1-13848	44135	70	20	801	93	2012-08-02 11:29:24.79138
44159	4	1-12243	44158	10	34	566	28	2012-08-02 11:29:27.940484
44160	4	1-13740	44158	34	91	1242	51	2012-08-02 11:29:27.940484
44161	4	1-12313	44158	91	99	1196	91	2012-08-02 11:29:27.940484
44158	4	1-13856	44158	99	10	978	65	2012-08-02 11:29:27.940484
44163	3	1-13784	44162	68	91	399	51	2012-08-02 11:29:27.940484
44164	3	1-12313	44162	91	99	398	91	2012-08-02 11:29:27.940484
44162	3	1-13856	44162	99	68	337	65	2012-08-02 11:29:27.940484
44194	3	1-13534	44193	58	37	4259	86	2012-08-02 11:29:34.635924
44195	3	1-9996	44193	37	23	2103	44	2012-08-02 11:29:34.635924
44193	3	1-13873	44193	23	58	1593	53	2012-08-02 11:29:34.635924
44221	3	1-12173	44220	59	87	495	55	2012-08-02 11:29:35.795893
44222	3	1-13158	44220	87	65	149	11	2012-08-02 11:29:35.795893
44220	3	1-13876	44220	65	59	1015	96	2012-08-02 11:29:35.795893
44224	3	1-7708	44223	46	1	793	42	2012-08-02 11:29:35.795893
44225	3	1-11780	44223	1	65	253	11	2012-08-02 11:29:35.795893
44223	3	1-13876	44223	65	46	1938	96	2012-08-02 11:29:35.795893
44227	4	1-13791	44226	69	62	1091	70	2012-08-02 11:29:35.795893
44228	4	1-13052	44226	62	1	394	42	2012-08-02 11:29:35.795893
44229	4	1-11780	44226	1	65	127	11	2012-08-02 11:29:35.795893
44226	4	1-13876	44226	65	69	984	96	2012-08-02 11:29:35.795893
44231	4	1-9544	44230	51	40	1004	87	2012-08-02 11:29:35.795893
44232	4	1-13422	44230	40	1	569	42	2012-08-02 11:29:35.795893
44233	4	1-11780	44230	1	65	201	11	2012-08-02 11:29:35.795893
44230	4	1-13876	44230	65	51	1705	96	2012-08-02 11:29:35.795893
44235	4	1-13278	44234	88	18	1804	93	2012-08-02 11:29:35.795893
44236	4	1-12986	44234	18	66	768	24	2012-08-02 11:29:35.795893
44237	4	1-5543	44234	66	65	178	11	2012-08-02 11:29:35.795893
44234	4	1-13876	44234	65	88	1513	96	2012-08-02 11:29:35.795893
44248	4	1-11560	44247	37	50	1311	63	2012-08-02 11:29:39.376229
44249	4	1-13836	44247	50	41	1892	38	2012-08-02 11:29:39.376229
44250	4	1-13756	44247	41	63	3679	70	2012-08-02 11:29:39.376229
44247	4	1-13884	44247	63	37	2962	66	2012-08-02 11:29:39.376229
44301	4	1-11927	44300	68	43	1607	3	2012-08-02 11:29:46.549736
44302	4	1-13736	44300	43	48	566	12	2012-08-02 11:29:46.549736
44303	4	1-13822	44300	48	10	315	10	2012-08-02 11:29:46.549736
44300	4	1-13903	44300	10	68	8990	69	2012-08-02 11:29:46.549736
44332	4	1-6206	44331	67	1	1459	68	2012-08-02 11:29:54.404978
44333	4	1-13654	44331	1	68	2476	65	2012-08-02 11:29:54.404978
44334	4	1-13784	44331	68	54	2642	51	2012-08-02 11:29:54.404978
44331	4	1-13923	44331	54	67	5504	6	2012-08-02 11:29:54.404978
44340	2	1-13848	44339	70	69	5075	93	2012-08-02 11:29:55.59852
44339	2	1-13926	44339	69	70	8190	7	2012-08-02 11:29:55.59852
44346	2	1-13042	44345	34	44	1787	35	2012-08-02 11:29:56.769677
44345	2	1-13929	44345	44	34	8055	73	2012-08-02 11:29:56.769677
44348	4	1-11879	44347	66	70	120	42	2012-08-02 11:29:56.769677
44349	4	1-12912	44347	70	82	104	72	2012-08-02 11:29:56.769677
44350	4	1-13595	44347	82	44	27	35	2012-08-02 11:29:56.769677
44347	4	1-13929	44347	44	66	122	73	2012-08-02 11:29:56.769677
44352	4	1-9569	44351	98	47	493	68	2012-08-02 11:29:56.769677
44353	4	1-7215	44351	47	92	254	31	2012-08-02 11:29:56.769677
44354	4	1-13498	44351	92	44	206	35	2012-08-02 11:29:56.769677
44351	4	1-13929	44351	44	98	1150	73	2012-08-02 11:29:56.769677
44355	1	1-7461	\N	88	88	325	42	2012-08-02 11:29:56.769677
44357	4	1-13815	44356	37	1	2096	79	2012-08-02 11:29:58.029416
44358	4	1-10900	44356	1	27	430	26	2012-08-02 11:29:58.029416
44359	4	1-13747	44356	27	56	1609	95	2012-08-02 11:29:58.029416
44356	4	1-13932	44356	56	37	2224	35	2012-08-02 11:29:58.029416
44361	3	1-13467	44360	81	91	3088	63	2012-08-02 11:29:58.029416
44362	3	1-12501	44360	91	56	3387	95	2012-08-02 11:29:58.029416
44360	3	1-13932	44360	56	81	5859	35	2012-08-02 11:29:58.029416
44374	4	1-13784	44373	68	94	1117	51	2012-08-02 11:30:00.36048
44375	4	1-13347	44373	94	84	1184	100	2012-08-02 11:30:00.36048
44376	4	1-13275	44373	84	51	1131	56	2012-08-02 11:30:00.36048
44373	4	1-13938	44373	51	68	832	65	2012-08-02 11:30:00.36048
44378	3	1-13717	44377	66	13	2322	6	2012-08-02 11:30:01.554338
44379	3	1-13916	44377	13	70	1462	42	2012-08-02 11:30:01.554338
44377	3	1-13941	44377	70	66	3832	98	2012-08-02 11:30:01.554338
44007	4	1-12083	44006	66	24	1691	43	2012-08-02 11:29:12.856069
44008	4	1-13340	44006	24	25	1005	63	2012-08-02 11:29:12.856069
44009	4	1-13764	44006	25	53	576	68	2012-08-02 11:29:12.856069
44011	4	1-12078	44010	28	64	563	38	2012-08-02 11:29:15.220572
44012	4	1-13582	44010	64	55	557	39	2012-08-02 11:29:15.220572
44013	4	1-8350	44010	55	15	178	9	2012-08-02 11:29:15.220572
44010	4	1-13824	44010	15	28	648	49	2012-08-02 11:29:15.220572
44015	4	1-13675	44014	42	64	4529	74	2012-08-02 11:29:15.220572
44016	4	1-11901	44014	64	41	2501	58	2012-08-02 11:29:15.220572
44017	4	1-9807	44014	41	15	804	9	2012-08-02 11:29:15.220572
44014	4	1-13824	44014	15	42	3062	49	2012-08-02 11:29:15.220572
44019	3	1-13180	44018	49	79	2980	46	2012-08-02 11:29:15.220572
44020	3	1-12301	44018	79	15	1227	9	2012-08-02 11:29:15.220572
44018	3	1-13824	44018	15	49	5891	49	2012-08-02 11:29:15.220572
44064	4	1-13758	44063	95	76	1503	92	2012-08-02 11:29:17.587342
44065	4	1-13537	44063	76	14	450	89	2012-08-02 11:29:17.587342
44066	4	1-13496	44063	14	11	388	23	2012-08-02 11:29:17.587342
44063	4	1-13830	44063	11	95	438	18	2012-08-02 11:29:17.587342
44080	3	1-13529	44079	36	10	6843	57	2012-08-02 11:29:18.780703
44081	3	1-11416	44079	10	82	2897	16	2012-08-02 11:29:18.780703
44079	3	1-13834	44079	82	36	3118	31	2012-08-02 11:29:18.780703
44083	4	1-13199	44082	26	32	1951	10	2012-08-02 11:29:18.780703
44084	4	1-13640	44082	32	36	2818	98	2012-08-02 11:29:18.780703
44085	4	1-13694	44082	36	82	1898	16	2012-08-02 11:29:18.780703
44082	4	1-13834	44082	82	26	2180	31	2012-08-02 11:29:18.780703
44087	4	1-13794	44086	53	10	3892	17	2012-08-02 11:29:18.780703
44088	4	1-10186	44086	10	80	1676	80	2012-08-02 11:29:18.780703
44089	4	1-13593	44086	80	82	1372	16	2012-08-02 11:29:18.780703
44086	4	1-13834	44086	82	53	1639	31	2012-08-02 11:29:18.780703
44091	4	1-13291	44090	78	88	23	97	2012-08-02 11:29:18.780703
44092	4	1-12621	44090	88	94	25	27	2012-08-02 11:29:18.780703
44093	4	1-10729	44090	94	82	24	16	2012-08-02 11:29:18.780703
44090	4	1-13834	44090	82	78	29	31	2012-08-02 11:29:18.780703
44095	3	1-12323	44094	77	11	261	33	2012-08-02 11:29:18.780703
44096	3	1-7495	44094	11	82	257	16	2012-08-02 11:29:18.780703
44094	3	1-13834	44094	82	77	314	31	2012-08-02 11:29:18.780703
44097	1	1-10734	\N	59	59	1312	31	2012-08-02 11:29:18.780703
44106	2	1-12749	44105	97	86	1539	81	2012-08-02 11:29:20.4272
44105	2	1-13837	44105	86	97	3060	23	2012-08-02 11:29:20.4272
44108	4	1-8705	44107	55	58	555	58	2012-08-02 11:29:20.4272
44109	4	1-13734	44107	58	83	922	38	2012-08-02 11:29:20.4272
44110	4	1-9673	44107	83	86	392	81	2012-08-02 11:29:20.4272
44107	4	1-13837	44107	86	55	1167	23	2012-08-02 11:29:20.4272
44112	3	1-13461	44111	79	69	2243	47	2012-08-02 11:29:20.4272
44113	3	1-10721	44111	69	86	929	81	2012-08-02 11:29:20.4272
44111	3	1-13837	44111	86	79	2900	23	2012-08-02 11:29:20.4272
44133	3	1-11925	44132	96	53	1262	76	2012-08-02 11:29:23.984961
44134	3	1-13745	44132	53	5	1001	15	2012-08-02 11:29:23.984961
44132	3	1-13846	44132	5	96	2190	66	2012-08-02 11:29:23.984961
44140	4	1-13580	44139	26	69	4765	69	2012-08-02 11:29:25.178304
44141	4	1-10805	44139	69	41	2092	58	2012-08-02 11:29:25.178304
44142	4	1-9807	44139	41	15	688	9	2012-08-02 11:29:25.178304
44139	4	1-13849	44139	15	26	2089	89	2012-08-02 11:29:25.178304
44144	4	1-13496	44143	14	71	1950	23	2012-08-02 11:29:25.178304
44145	4	1-11240	44143	71	78	785	63	2012-08-02 11:29:25.178304
44146	4	1-12457	44143	78	15	783	9	2012-08-02 11:29:25.178304
44143	4	1-13849	44143	15	14	2481	89	2012-08-02 11:29:25.178304
44148	4	1-6859	44147	57	26	236	94	2012-08-02 11:29:25.178304
44149	4	1-13295	44147	26	41	586	58	2012-08-02 11:29:25.178304
44150	4	1-9807	44147	41	15	205	9	2012-08-02 11:29:25.178304
44147	4	1-13849	44147	15	57	661	89	2012-08-02 11:29:25.178304
44152	3	1-13748	44151	65	61	735	100	2012-08-02 11:29:25.178304
44153	3	1-13447	44151	61	15	533	9	2012-08-02 11:29:25.178304
44151	3	1-13849	44151	15	65	1757	89	2012-08-02 11:29:25.178304
44168	3	1-12753	44167	60	61	1373	99	2012-08-02 11:29:29.884913
44169	3	1-10772	44167	61	53	521	9	2012-08-02 11:29:29.884913
44167	3	1-13861	44167	53	60	2283	30	2012-08-02 11:29:29.884913
44171	4	1-13731	44170	38	35	2016	33	2012-08-02 11:29:29.884913
44172	4	1-13294	44170	35	78	1295	63	2012-08-02 11:29:29.884913
44173	4	1-12457	44170	78	53	1105	9	2012-08-02 11:29:29.884913
44170	4	1-13861	44170	53	38	5087	30	2012-08-02 11:29:29.884913
44175	4	1-12748	44174	5	88	20	14	2012-08-02 11:29:31.862809
44176	4	1-12217	44174	88	46	6	91	2012-08-02 11:29:31.862809
44177	4	1-13113	44174	46	80	5	11	2012-08-02 11:29:31.862809
44174	4	1-13865	44174	80	5	8	33	2012-08-02 11:29:31.862809
44179	4	1-9372	44178	22	22	1446	80	2012-08-02 11:29:31.862809
44180	4	1-13499	44178	22	46	1901	91	2012-08-02 11:29:31.862809
44181	4	1-13113	44178	46	80	1359	11	2012-08-02 11:29:31.862809
44178	4	1-13865	44178	80	22	2297	33	2012-08-02 11:29:31.862809
44183	4	1-13294	44182	35	94	746	63	2012-08-02 11:29:31.862809
44184	4	1-13282	44182	94	46	502	91	2012-08-02 11:29:31.862809
44185	4	1-13113	44182	46	80	443	11	2012-08-02 11:29:31.862809
44182	4	1-13865	44182	80	35	923	33	2012-08-02 11:29:31.862809
44187	4	1-6195	44186	59	26	304	73	2012-08-02 11:29:31.862809
44188	4	1-13599	44186	26	46	469	91	2012-08-02 11:29:31.862809
44189	4	1-13113	44186	46	80	422	11	2012-08-02 11:29:31.862809
44186	4	1-13865	44186	80	59	897	33	2012-08-02 11:29:31.862809
44191	3	1-8311	44190	32	46	474	91	2012-08-02 11:29:31.862809
44192	3	1-13113	44190	46	80	428	11	2012-08-02 11:29:31.862809
44190	3	1-13865	44190	80	32	914	33	2012-08-02 11:29:31.862809
44197	4	1-13179	44196	86	1	57	68	2012-08-02 11:29:35.033431
44198	4	1-13654	44196	1	4	82	65	2012-08-02 11:29:35.033431
44199	4	1-11975	44196	4	60	72	45	2012-08-02 11:29:35.033431
44196	4	1-13874	44196	60	86	113	64	2012-08-02 11:29:35.033431
44201	4	1-13258	44200	69	16	263	75	2012-08-02 11:29:35.033431
44202	4	1-13665	44200	16	63	299	16	2012-08-02 11:29:35.033431
44203	4	1-12674	44200	63	60	215	45	2012-08-02 11:29:35.033431
44200	4	1-13874	44200	60	69	365	64	2012-08-02 11:29:35.033431
44205	4	1-13179	44204	86	59	762	68	2012-08-02 11:29:35.033431
44206	4	1-12297	44204	59	46	685	34	2012-08-02 11:29:35.033431
44207	4	1-11991	44204	46	60	667	45	2012-08-02 11:29:35.033431
44204	4	1-13874	44204	60	86	1261	64	2012-08-02 11:29:35.033431
44077	4	1-13722	44075	89	60	123	44	2012-08-02 11:29:19.178344
44078	4	1-13821	44075	60	54	74	67	2012-08-02 11:29:19.178344
44099	4	1-12960	44098	9	79	4544	23	2012-08-02 11:29:19.996112
44100	4	1-13398	44098	79	61	3159	100	2012-08-02 11:29:19.996112
44101	4	1-13447	44098	61	69	1797	9	2012-08-02 11:29:19.996112
44098	4	1-13835	44098	69	9	6659	10	2012-08-02 11:29:19.996112
44103	3	1-13832	44102	30	39	1853	28	2012-08-02 11:29:19.996112
44104	3	1-12767	44102	39	69	836	9	2012-08-02 11:29:19.996112
44102	3	1-13835	44102	69	30	3196	10	2012-08-02 11:29:19.996112
44115	4	1-1251	44114	99	1	1160	68	2012-08-02 11:29:20.814024
44116	4	1-13654	44114	1	21	1696	65	2012-08-02 11:29:20.814024
44117	4	1-13307	44114	21	100	878	19	2012-08-02 11:29:20.814024
44114	4	1-13838	44114	100	99	6984	47	2012-08-02 11:29:20.814024
44119	4	1-10972	44118	45	9	862	10	2012-08-02 11:29:21.974274
44120	4	1-12960	44118	9	11	624	23	2012-08-02 11:29:21.974274
44121	4	1-13779	44118	11	48	602	53	2012-08-02 11:29:21.974274
44118	4	1-13841	44118	48	45	2486	64	2012-08-02 11:29:21.974274
44123	4	1-13179	44122	86	59	2071	68	2012-08-02 11:29:21.974274
44124	4	1-12297	44122	59	27	1728	34	2012-08-02 11:29:21.974274
44125	4	1-9593	44122	27	48	865	53	2012-08-02 11:29:21.974274
44122	4	1-13841	44122	48	86	3690	64	2012-08-02 11:29:21.974274
44127	4	1-13258	44126	69	54	529	75	2012-08-02 11:29:21.974274
44128	4	1-12560	44126	54	33	385	41	2012-08-02 11:29:21.974274
44129	4	1-4364	44126	33	48	135	53	2012-08-02 11:29:21.974274
44126	4	1-13841	44126	48	69	639	64	2012-08-02 11:29:21.974274
44155	4	1-13674	44154	70	63	431	1	2012-08-02 11:29:25.576146
44156	4	1-13146	44154	63	45	456	26	2012-08-02 11:29:25.576146
44157	4	1-12365	44154	45	24	404	75	2012-08-02 11:29:25.576146
44154	4	1-13850	44154	24	70	337	74	2012-08-02 11:29:25.576146
44166	2	1-10350	44165	33	82	677	9	2012-08-02 11:29:28.31607
44165	2	1-13857	44165	82	33	3533	14	2012-08-02 11:29:28.31607
44243	2	1-9754	44242	3	58	611	27	2012-08-02 11:29:36.591718
44242	2	1-13878	44242	58	3	784	74	2012-08-02 11:29:36.591718
44245	3	1-10808	44244	78	88	1962	97	2012-08-02 11:29:36.591718
44246	3	1-12621	44244	88	58	2152	27	2012-08-02 11:29:36.591718
44244	3	1-13878	44244	58	78	3013	74	2012-08-02 11:29:36.591718
44252	4	1-9454	44251	94	24	4901	75	2012-08-02 11:29:38.989309
44253	4	1-13850	44251	24	26	4149	74	2012-08-02 11:29:38.989309
44254	4	1-12699	44251	26	90	2712	29	2012-08-02 11:29:38.989309
44251	4	1-13885	44251	90	94	8275	78	2012-08-02 11:29:38.989309
44256	3	1-12966	44255	53	64	775	42	2012-08-02 11:29:38.989309
44257	3	1-12956	44255	64	90	457	29	2012-08-02 11:29:38.989309
44255	3	1-13885	44255	90	53	1396	78	2012-08-02 11:29:38.989309
44270	4	1-13746	44269	79	25	588	79	2012-08-02 11:29:44.549699
44271	4	1-9027	44269	25	81	117	35	2012-08-02 11:29:44.549699
44272	4	1-13467	44269	81	56	54	63	2012-08-02 11:29:44.549699
44269	4	1-13898	44269	56	79	213	56	2012-08-02 11:29:44.549699
44285	4	1-11810	44284	13	58	3080	21	2012-08-02 11:29:45.743019
44286	4	1-13029	44284	58	17	2406	48	2012-08-02 11:29:45.743019
44287	4	1-13088	44284	17	81	1676	91	2012-08-02 11:29:45.743019
44284	4	1-13901	44284	81	13	3438	18	2012-08-02 11:29:45.743019
44289	3	1-8872	44288	45	26	1586	73	2012-08-02 11:29:45.743019
44290	3	1-13599	44288	26	81	2268	91	2012-08-02 11:29:45.743019
44288	3	1-13901	44288	81	45	5381	18	2012-08-02 11:29:45.743019
44291	1	1-4483	\N	94	94	76	48	2012-08-02 11:29:45.743019
44305	4	1-13755	44304	28	53	3109	31	2012-08-02 11:29:48.152049
44306	4	1-13794	44304	53	27	5713	17	2012-08-02 11:29:48.152049
44307	4	1-13078	44304	27	25	6265	87	2012-08-02 11:29:48.152049
44304	4	1-13907	44304	25	28	6221	74	2012-08-02 11:29:48.152049
44309	3	1-13755	44308	28	42	616	31	2012-08-02 11:29:48.152049
44310	3	1-13344	44308	42	25	1167	87	2012-08-02 11:29:48.152049
44308	3	1-13907	44308	25	28	1195	74	2012-08-02 11:29:48.152049
44324	4	1-12912	44323	70	79	3742	72	2012-08-02 11:29:54.029273
44325	4	1-13165	44323	79	30	2341	10	2012-08-02 11:29:54.029273
44326	4	1-13832	44323	30	42	1711	28	2012-08-02 11:29:54.029273
44323	4	1-13922	44323	42	70	3336	42	2012-08-02 11:29:54.029273
44328	4	1-11439	44327	27	64	221	38	2012-08-02 11:29:54.029273
44329	4	1-13582	44327	64	50	274	39	2012-08-02 11:29:54.029273
44330	4	1-13869	44327	50	42	238	28	2012-08-02 11:29:54.029273
44327	4	1-13922	44327	42	27	494	42	2012-08-02 11:29:54.029273
44336	4	1-7713	44335	51	9	2440	85	2012-08-02 11:29:55.200678
44337	4	1-13896	44335	9	11	1680	33	2012-08-02 11:29:55.200678
44338	4	1-7495	44335	11	91	923	16	2012-08-02 11:29:55.200678
44335	4	1-13924	44335	91	51	8948	83	2012-08-02 11:29:55.200678
44342	4	1-13674	44341	70	100	5181	1	2012-08-02 11:29:55.996314
44343	4	1-13586	44341	100	54	3331	67	2012-08-02 11:29:55.996314
44344	4	1-13833	44341	54	49	5181	18	2012-08-02 11:29:55.996314
44341	4	1-13927	44341	49	70	4119	74	2012-08-02 11:29:55.996314
44366	3	1-9432	44365	34	58	1280	29	2012-08-02 11:29:59.576353
44367	3	1-13067	44365	58	14	931	62	2012-08-02 11:29:59.576353
44365	3	1-13936	44365	14	34	2556	65	2012-08-02 11:29:59.576353
44369	2	1-4530	44368	26	14	529	62	2012-08-02 11:29:59.576353
44368	2	1-13936	44368	14	26	1491	65	2012-08-02 11:29:59.576353
44371	3	1-10594	44370	1	58	746	29	2012-08-02 11:29:59.576353
44372	3	1-13067	44370	58	14	583	62	2012-08-02 11:29:59.576353
44370	3	1-13936	44370	14	1	1715	65	2012-08-02 11:29:59.576353
44411	3	1-12363	44410	59	85	1280	15	2012-08-02 11:30:06.614302
44412	3	1-12678	44410	85	46	1093	37	2012-08-02 11:30:06.614302
44410	3	1-13955	44410	46	59	995	2	2012-08-02 11:30:06.614302
44414	4	1-13783	44413	94	60	91	30	2012-08-02 11:30:08.195798
44415	4	1-12753	44413	60	9	78	99	2012-08-02 11:30:08.195798
44416	4	1-12636	44413	9	43	71	47	2012-08-02 11:30:08.195798
44413	4	1-13958	44413	43	94	46	43	2012-08-02 11:30:08.195798
44422	3	1-9172	44421	39	86	2178	91	2012-08-02 11:30:10.516962
44423	3	1-13828	44421	86	89	738	47	2012-08-02 11:30:10.516962
44421	3	1-13964	44421	89	39	6836	36	2012-08-02 11:30:10.516962
44425	4	1-12473	44424	62	56	719	63	2012-08-02 11:30:11.709306
44426	4	1-13898	44424	56	68	3269	56	2012-08-02 11:30:11.709306
44427	4	1-13664	44424	68	42	2465	77	2012-08-02 11:30:11.709306
44424	4	1-13967	44424	42	62	1860	67	2012-08-02 11:30:11.709306
44429	4	1-13827	44428	73	71	5892	82	2012-08-02 11:30:12.885015
44430	4	1-12329	44428	71	38	1646	21	2012-08-02 11:30:12.885015
44431	4	1-13800	44428	38	27	2126	97	2012-08-02 11:30:12.885015
44428	4	1-13970	44428	27	73	1409	68	2012-08-02 11:30:12.885015
44209	4	1-13855	44208	4	38	522	3	2012-08-02 11:29:35.033431
44210	4	1-8944	44208	38	39	350	38	2012-08-02 11:29:35.033431
44211	4	1-13681	44208	39	60	835	45	2012-08-02 11:29:35.033431
44208	4	1-13874	44208	60	4	1659	64	2012-08-02 11:29:35.033431
44213	4	1-13179	44212	86	76	3615	68	2012-08-02 11:29:35.033431
44214	4	1-13320	44212	76	66	2780	78	2012-08-02 11:29:35.033431
44215	4	1-13531	44212	66	60	2821	45	2012-08-02 11:29:35.033431
44212	4	1-13874	44212	60	86	5648	64	2012-08-02 11:29:35.033431
44217	4	1-578	44216	29	53	44	94	2012-08-02 11:29:35.033431
44218	4	1-12833	44216	53	13	35	89	2012-08-02 11:29:35.033431
44219	4	1-13542	44216	13	60	94	45	2012-08-02 11:29:35.033431
44216	4	1-13874	44216	60	29	206	64	2012-08-02 11:29:35.033431
44239	4	1-13772	44238	37	78	2362	65	2012-08-02 11:29:36.193916
44240	4	1-13123	44238	78	45	1761	26	2012-08-02 11:29:36.193916
44241	4	1-12365	44238	45	21	1218	75	2012-08-02 11:29:36.193916
44238	4	1-13877	44238	21	37	2538	4	2012-08-02 11:29:36.193916
44259	2	1-13344	44258	42	65	2207	87	2012-08-02 11:29:39.762563
44258	2	1-13886	44258	65	42	2758	31	2012-08-02 11:29:39.762563
44261	3	1-13879	44260	38	29	4267	32	2012-08-02 11:29:42.160175
44262	3	1-13816	44260	29	74	5812	69	2012-08-02 11:29:42.160175
44260	3	1-13893	44260	74	38	7499	95	2012-08-02 11:29:42.160175
44264	2	1-11788	44263	30	74	257	69	2012-08-02 11:29:42.160175
44263	2	1-13893	44263	74	30	416	95	2012-08-02 11:29:42.160175
44266	4	1-6032	44265	78	12	1375	92	2012-08-02 11:29:43.786822
44267	4	1-13660	44265	12	42	1065	59	2012-08-02 11:29:43.786822
44268	4	1-11692	44265	42	9	1419	85	2012-08-02 11:29:43.786822
44265	4	1-13896	44265	9	78	1598	33	2012-08-02 11:29:43.786822
44274	4	1-11463	44273	6	47	573	28	2012-08-02 11:29:44.95862
44275	4	1-13792	44273	47	79	1630	30	2012-08-02 11:29:44.95862
44276	4	1-11201	44273	79	83	623	41	2012-08-02 11:29:44.95862
44273	4	1-13899	44273	83	6	618	19	2012-08-02 11:29:44.95862
44278	4	1-11463	44277	6	42	1577	28	2012-08-02 11:29:44.95862
44279	4	1-13272	44277	42	66	1739	46	2012-08-02 11:29:44.95862
44280	4	1-12876	44277	66	83	1608	41	2012-08-02 11:29:44.95862
44277	4	1-13899	44277	83	6	1647	19	2012-08-02 11:29:44.95862
44282	3	1-11006	44281	20	67	1213	87	2012-08-02 11:29:44.95862
44283	3	1-13680	44281	67	83	1118	41	2012-08-02 11:29:44.95862
44281	3	1-13899	44281	83	20	1215	19	2012-08-02 11:29:44.95862
44293	4	1-13888	44292	39	84	843	100	2012-08-02 11:29:46.141041
44294	4	1-11176	44292	84	19	893	16	2012-08-02 11:29:46.141041
44295	4	1-11928	44292	19	36	706	77	2012-08-02 11:29:46.141041
44292	4	1-13902	44292	36	39	932	20	2012-08-02 11:29:46.141041
44297	4	1-12050	44296	1	64	402	32	2012-08-02 11:29:46.141041
44298	4	1-13622	44296	64	68	940	5	2012-08-02 11:29:46.141041
44299	4	1-13581	44296	68	36	1150	77	2012-08-02 11:29:46.141041
44296	4	1-13902	44296	36	1	1525	20	2012-08-02 11:29:46.141041
44312	4	1-12776	44311	90	24	3830	43	2012-08-02 11:29:53.255963
44313	4	1-13340	44311	24	10	2072	63	2012-08-02 11:29:53.255963
44314	4	1-13239	44311	10	80	2200	23	2012-08-02 11:29:53.255963
44311	4	1-13921	44311	80	90	9270	90	2012-08-02 11:29:53.255963
44316	4	1-12776	44315	90	63	183	43	2012-08-02 11:29:53.255963
44317	4	1-7009	44315	63	9	57	10	2012-08-02 11:29:53.255963
44318	4	1-12960	44315	9	80	51	23	2012-08-02 11:29:53.255963
44315	4	1-13921	44315	80	90	308	90	2012-08-02 11:29:53.255963
44320	4	1-10244	44319	60	63	28	43	2012-08-02 11:29:53.255963
44321	4	1-7009	44319	63	9	9	10	2012-08-02 11:29:53.255963
44322	4	1-12960	44319	9	80	8	23	2012-08-02 11:29:53.255963
44319	4	1-13921	44319	80	60	52	90	2012-08-02 11:29:53.255963
44364	2	1-13530	44363	11	2	1665	59	2012-08-02 11:29:58.802918
44363	2	1-13934	44363	2	11	2400	96	2012-08-02 11:29:58.802918
44385	3	1-13219	44384	79	36	414	73	2012-08-02 11:30:03.509658
44386	3	1-12768	44384	36	97	1087	71	2012-08-02 11:30:03.509658
44384	3	1-13946	44384	97	79	1144	56	2012-08-02 11:30:03.509658
44388	3	1-2978	44387	100	36	236	73	2012-08-02 11:30:03.509658
44389	3	1-12768	44387	36	97	639	71	2012-08-02 11:30:03.509658
44387	3	1-13946	44387	97	100	693	56	2012-08-02 11:30:03.509658
44395	3	1-2542	44394	67	19	111	40	2012-08-02 11:30:04.669386
44396	3	1-13047	44394	19	65	49	7	2012-08-02 11:30:04.669386
44394	3	1-13949	44394	65	67	400	71	2012-08-02 11:30:04.669386
44398	3	1-12852	44397	3	19	1172	40	2012-08-02 11:30:04.669386
44399	3	1-13047	44397	19	65	593	7	2012-08-02 11:30:04.669386
44397	3	1-13949	44397	65	3	5585	71	2012-08-02 11:30:04.669386
44401	4	1-13575	44400	67	71	129	74	2012-08-02 11:30:04.669386
44402	4	1-9095	44400	71	23	114	88	2012-08-02 11:30:04.669386
44403	4	1-12893	44400	23	65	64	7	2012-08-02 11:30:04.669386
44400	4	1-13949	44400	65	67	787	71	2012-08-02 11:30:04.669386
44409	2	1-9765	44408	89	37	599	40	2012-08-02 11:30:05.840752
44408	2	1-13952	44408	37	89	6670	58	2012-08-02 11:30:05.840752
44418	4	1-11941	44417	80	36	2052	77	2012-08-02 11:30:09.753486
44419	4	1-13902	44417	36	37	2626	20	2012-08-02 11:30:09.753486
44420	4	1-13072	44417	37	9	2622	37	2012-08-02 11:30:09.753486
44417	4	1-13962	44417	9	80	1897	6	2012-08-02 11:30:09.753486
44438	2	1-13968	44437	33	61	1306	58	2012-08-02 11:30:13.647977
44437	2	1-13972	44437	61	33	2360	75	2012-08-02 11:30:13.647977
44464	3	1-13396	44463	72	60	1884	44	2012-08-02 11:30:17.213447
44465	3	1-13821	44463	60	1	594	67	2012-08-02 11:30:17.213447
44463	3	1-13982	44463	1	72	3990	91	2012-08-02 11:30:17.213447
44467	3	1-9340	44466	86	42	546	77	2012-08-02 11:30:17.213447
44468	3	1-13967	44466	42	1	224	67	2012-08-02 11:30:17.213447
44466	3	1-13982	44466	1	86	1967	91	2012-08-02 11:30:17.213447
44513	4	1-13878	44512	58	70	1427	74	2012-08-02 11:30:23.924613
44514	4	1-13674	44512	70	48	1765	1	2012-08-02 11:30:23.924613
44515	4	1-13786	44512	48	47	3603	85	2012-08-02 11:30:23.924613
44512	4	1-13998	44512	47	58	1008	27	2012-08-02 11:30:23.924613
44542	3	1-7467	44541	77	53	483	75	2012-08-02 11:30:31.832093
44543	3	1-13917	44541	53	19	973	45	2012-08-02 11:30:31.832093
44541	3	1-14018	44541	19	77	4381	33	2012-08-02 11:30:31.832093
44556	3	1-6494	44555	18	9	1308	85	2012-08-02 11:30:32.970668
44557	3	1-13896	44555	9	35	710	33	2012-08-02 11:30:32.970668
44555	3	1-14021	44555	35	18	7179	58	2012-08-02 11:30:32.970668
44381	4	1-12610	44380	87	79	2008	72	2012-08-02 11:30:02.736062
44382	4	1-13165	44380	79	53	1304	10	2012-08-02 11:30:02.736062
44383	4	1-12127	44380	53	78	1065	30	2012-08-02 11:30:02.736062
44380	4	1-13944	44380	78	87	2311	97	2012-08-02 11:30:02.736062
44391	2	1-9208	44390	25	47	1486	75	2012-08-02 11:30:03.906984
44390	2	1-13947	44390	47	25	5503	13	2012-08-02 11:30:03.906984
44393	2	1-12670	44392	63	47	813	75	2012-08-02 11:30:03.906984
44392	2	1-13947	44392	47	63	3460	13	2012-08-02 11:30:03.906984
44405	4	1-13733	44404	91	16	2920	87	2012-08-02 11:30:05.056251
44406	4	1-13629	44404	16	47	1639	77	2012-08-02 11:30:05.056251
44407	4	1-13602	44404	47	66	913	10	2012-08-02 11:30:05.056251
44404	4	1-13950	44404	66	91	1233	49	2012-08-02 11:30:05.056251
44448	4	1-12540	44447	4	71	1147	86	2012-08-02 11:30:14.435146
44449	4	1-13963	44447	71	6	1455	2	2012-08-02 11:30:14.435146
44450	4	1-13852	44447	6	3	1644	46	2012-08-02 11:30:14.435146
44447	4	1-13974	44447	3	4	1372	6	2012-08-02 11:30:14.435146
44452	4	1-13791	44451	69	63	567	70	2012-08-02 11:30:16.762619
44453	4	1-13884	44451	63	46	310	66	2012-08-02 11:30:16.762619
44454	4	1-12417	44451	46	24	204	13	2012-08-02 11:30:16.762619
44451	4	1-13980	44451	24	69	529	96	2012-08-02 11:30:16.762619
44456	4	1-13530	44455	11	39	4759	59	2012-08-02 11:30:16.762619
44457	4	1-12695	44455	39	93	3761	37	2012-08-02 11:30:16.762619
44458	4	1-13420	44455	93	24	1678	13	2012-08-02 11:30:16.762619
44455	4	1-13980	44455	24	11	5408	96	2012-08-02 11:30:16.762619
44460	4	1-1089	44459	99	6	858	2	2012-08-02 11:30:16.762619
44461	4	1-13852	44459	6	91	944	46	2012-08-02 11:30:16.762619
44462	4	1-13506	44459	91	24	611	13	2012-08-02 11:30:16.762619
44459	4	1-13980	44459	24	99	2120	96	2012-08-02 11:30:16.762619
44470	4	1-6979	44469	88	14	2278	98	2012-08-02 11:30:18.003509
44471	4	1-13979	44469	14	81	5456	80	2012-08-02 11:30:18.003509
44472	4	1-13271	44469	81	74	3226	89	2012-08-02 11:30:18.003509
44469	4	1-13983	44469	74	88	4614	99	2012-08-02 11:30:18.003509
44474	4	1-13462	44473	79	3	1763	95	2012-08-02 11:30:19.190513
44475	4	1-12543	44473	3	13	605	79	2012-08-02 11:30:19.190513
44476	4	1-13488	44473	13	65	1349	54	2012-08-02 11:30:19.190513
44473	4	1-13985	44473	65	79	3697	45	2012-08-02 11:30:19.190513
44478	2	1-13403	44477	41	13	2509	96	2012-08-02 11:30:20.7384
44477	2	1-13990	44477	13	41	5582	78	2012-08-02 11:30:20.7384
44480	3	1-9897	44479	45	28	300	80	2012-08-02 11:30:20.7384
44481	3	1-13061	44479	28	13	185	96	2012-08-02 11:30:20.7384
44479	3	1-13990	44479	13	45	752	78	2012-08-02 11:30:20.7384
44483	4	1-6553	44482	98	22	1525	88	2012-08-02 11:30:20.7384
44484	4	1-13778	44482	22	70	1161	60	2012-08-02 11:30:20.7384
44485	4	1-13308	44482	70	13	396	96	2012-08-02 11:30:20.7384
44482	4	1-13990	44482	13	98	2005	78	2012-08-02 11:30:20.7384
44487	3	1-6140	44486	76	26	223	91	2012-08-02 11:30:20.7384
44488	3	1-12519	44486	26	13	92	96	2012-08-02 11:30:20.7384
44486	3	1-13990	44486	13	76	504	78	2012-08-02 11:30:20.7384
44501	4	1-12893	44500	23	71	689	7	2012-08-02 11:30:21.901036
44502	4	1-11749	44500	71	6	491	2	2012-08-02 11:30:21.901036
44503	4	1-13852	44500	6	7	567	46	2012-08-02 11:30:21.901036
44500	4	1-13994	44500	7	23	1005	88	2012-08-02 11:30:21.901036
44534	4	1-13816	44533	29	22	1182	69	2012-08-02 11:30:29.07869
44535	4	1-12170	44533	22	90	614	24	2012-08-02 11:30:29.07869
44536	4	1-10816	44533	90	99	349	78	2012-08-02 11:30:29.07869
44533	4	1-14011	44533	99	29	578	32	2012-08-02 11:30:29.07869
44538	4	1-13501	44537	49	80	444	11	2012-08-02 11:30:29.07869
44539	4	1-13865	44537	80	61	1026	33	2012-08-02 11:30:29.07869
44540	4	1-5464	44537	61	99	263	78	2012-08-02 11:30:29.07869
44537	4	1-14011	44537	99	49	448	32	2012-08-02 11:30:29.07869
44433	4	1-13827	44432	73	19	1806	82	2012-08-02 11:30:12.885015
44434	4	1-214	44432	19	38	416	21	2012-08-02 11:30:12.885015
44435	4	1-13800	44432	38	27	573	97	2012-08-02 11:30:12.885015
44432	4	1-13970	44432	27	73	405	68	2012-08-02 11:30:12.885015
44436	1	1-214	\N	19	19	1208	21	2012-08-02 11:30:12.885015
44440	4	1-13878	44439	58	79	2308	74	2012-08-02 11:30:14.048472
44441	4	1-13163	44439	79	73	1583	55	2012-08-02 11:30:14.048472
44442	4	1-8427	44439	73	43	1361	40	2012-08-02 11:30:14.048472
44439	4	1-13973	44439	43	58	1913	27	2012-08-02 11:30:14.048472
44444	3	1-13059	44443	4	11	3709	93	2012-08-02 11:30:14.048472
44445	3	1-11494	44443	11	43	2726	40	2012-08-02 11:30:14.048472
44443	3	1-13973	44443	43	4	3849	27	2012-08-02 11:30:14.048472
44446	1	1-1385	\N	61	61	1541	40	2012-08-02 11:30:14.048472
44490	3	1-13307	44489	21	45	2142	19	2012-08-02 11:30:21.125048
44491	3	1-7309	44489	45	33	2068	60	2012-08-02 11:30:21.125048
44489	3	1-13991	44489	33	21	3357	65	2012-08-02 11:30:21.125048
44493	4	1-11103	44492	13	43	68	61	2012-08-02 11:30:21.125048
44494	4	1-13419	44492	43	54	51	53	2012-08-02 11:30:21.125048
44495	4	1-11503	44492	54	33	49	60	2012-08-02 11:30:21.125048
44492	4	1-13991	44492	33	13	87	65	2012-08-02 11:30:21.125048
44497	4	1-7808	44496	85	6	753	2	2012-08-02 11:30:22.366312
44498	4	1-13852	44496	6	68	855	46	2012-08-02 11:30:22.366312
44499	4	1-13453	44496	68	19	1135	90	2012-08-02 11:30:22.366312
44496	4	1-13993	44496	19	85	1363	80	2012-08-02 11:30:22.366312
44505	4	1-13271	44504	81	13	895	89	2012-08-02 11:30:23.150851
44506	4	1-13542	44504	13	72	2134	45	2012-08-02 11:30:23.150851
44507	4	1-12544	44504	72	45	1002	72	2012-08-02 11:30:23.150851
44504	4	1-13996	44504	45	81	1792	80	2012-08-02 11:30:23.150851
44509	4	1-13593	44508	80	63	1468	16	2012-08-02 11:30:23.150851
44510	4	1-12674	44508	63	72	1422	45	2012-08-02 11:30:23.150851
44511	4	1-12544	44508	72	45	799	72	2012-08-02 11:30:23.150851
44508	4	1-13996	44508	45	80	1707	80	2012-08-02 11:30:23.150851
44517	3	1-13593	44516	80	66	3266	16	2012-08-02 11:30:24.323159
44518	3	1-13104	44516	66	58	1191	66	2012-08-02 11:30:24.323159
44516	3	1-14000	44516	58	80	3808	80	2012-08-02 11:30:24.323159
44520	3	1-12136	44519	15	37	811	4	2012-08-02 11:30:25.883989
44521	3	1-13772	44519	37	55	901	65	2012-08-02 11:30:25.883989
44519	3	1-14003	44519	55	15	1227	99	2012-08-02 11:30:25.883989
44523	4	1-13994	44522	7	7	6576	88	2012-08-02 11:30:27.067488
44524	4	1-12326	44522	7	66	1873	10	2012-08-02 11:30:27.067488
44525	4	1-13950	44522	66	40	2004	49	2012-08-02 11:30:27.067488
44522	4	1-14006	44522	40	7	3999	46	2012-08-02 11:30:27.067488
44527	4	1-13309	44526	21	9	104	22	2012-08-02 11:30:27.067488
44528	4	1-8273	44526	9	66	46	10	2012-08-02 11:30:27.067488
44529	4	1-13950	44526	66	40	62	49	2012-08-02 11:30:27.067488
44526	4	1-14006	44526	40	21	155	46	2012-08-02 11:30:27.067488
44531	3	1-13709	44530	54	9	566	9	2012-08-02 11:30:28.283346
44532	3	1-13954	44530	9	44	254	35	2012-08-02 11:30:28.283346
44530	3	1-14009	44530	44	54	4641	53	2012-08-02 11:30:28.283346
44545	4	1-13474	44544	43	37	2738	4	2012-08-02 11:30:32.230146
44546	4	1-13772	44544	37	78	2242	65	2012-08-02 11:30:32.230146
44547	4	1-13123	44544	78	4	1470	26	2012-08-02 11:30:32.230146
44544	4	1-14019	44544	4	43	4586	89	2012-08-02 11:30:32.230146
44549	3	1-13542	44548	13	81	307	45	2012-08-02 11:30:32.230146
44550	3	1-5461	44548	81	4	40	26	2012-08-02 11:30:32.230146
44548	3	1-14019	44548	4	13	140	89	2012-08-02 11:30:32.230146
44552	4	1-7433	44551	13	12	680	83	2012-08-02 11:30:32.230146
44553	4	1-11501	44551	12	89	284	38	2012-08-02 11:30:32.230146
44554	4	1-13986	44551	89	4	235	26	2012-08-02 11:30:32.230146
44551	4	1-14019	44551	4	13	970	89	2012-08-02 11:30:32.230146
44559	3	1-12919	44558	92	32	1227	84	2012-08-02 11:30:33.358312
44560	3	1-10568	44558	32	79	525	50	2012-08-02 11:30:33.358312
44558	3	1-14022	44558	79	92	1150	10	2012-08-02 11:30:33.358312
44562	3	1-12960	44561	9	21	110	23	2012-08-02 11:30:33.358312
44563	3	1-10541	44561	21	79	68	50	2012-08-02 11:30:33.358312
44561	3	1-14022	44561	79	9	172	10	2012-08-02 11:30:33.358312
44565	4	1-13950	44564	66	20	2069	49	2012-08-02 11:30:33.358312
44566	4	1-7838	44564	20	92	905	48	2012-08-02 11:30:33.358312
44567	4	1-13368	44564	92	79	486	50	2012-08-02 11:30:33.358312
44564	4	1-14022	44564	79	66	1672	10	2012-08-02 11:30:33.358312
44569	4	1-13447	44568	61	43	519	9	2012-08-02 11:30:34.543923
44570	4	1-11352	44568	43	48	340	1	2012-08-02 11:30:34.543923
44571	4	1-13786	44568	48	25	609	85	2012-08-02 11:30:34.543923
44568	4	1-14026	44568	25	61	751	100	2012-08-02 11:30:34.543923
44573	4	1-11826	44572	45	70	1209	74	2012-08-02 11:30:34.543923
44574	4	1-13674	44572	70	48	1461	1	2012-08-02 11:30:34.543923
44575	4	1-13786	44572	48	25	2917	85	2012-08-02 11:30:34.543923
44572	4	1-14026	44572	25	45	4007	100	2012-08-02 11:30:34.543923
44577	3	1-13894	44576	68	14	1813	17	2012-08-02 11:30:34.543923
44578	3	1-11190	44576	14	25	1156	85	2012-08-02 11:30:34.543923
44576	3	1-14026	44576	25	68	1651	100	2012-08-02 11:30:34.543923
44580	3	1-13195	44579	70	81	7471	40	2012-08-02 11:30:36.876255
44581	3	1-14017	44579	81	56	3937	11	2012-08-02 11:30:36.876255
44579	3	1-14030	44579	56	70	6546	34	2012-08-02 11:30:36.876255
44583	4	1-14010	44582	78	25	326	13	2012-08-02 11:30:36.876255
44584	4	1-11514	44582	25	49	278	32	2012-08-02 11:30:36.876255
44585	4	1-13501	44582	49	56	231	11	2012-08-02 11:30:36.876255
44582	4	1-14030	44582	56	78	445	34	2012-08-02 11:30:36.876255
44587	4	1-13581	44586	68	42	1533	77	2012-08-02 11:30:38.533307
44588	4	1-13967	44586	42	77	1234	67	2012-08-02 11:30:38.533307
44589	4	1-12628	44586	77	43	522	53	2012-08-02 11:30:38.533307
44586	4	1-14033	44586	43	68	1275	5	2012-08-02 11:30:38.533307
44591	3	1-8369	44590	69	89	956	38	2012-08-02 11:30:39.30669
44592	3	1-13986	44590	89	86	709	26	2012-08-02 11:30:39.30669
44590	3	1-14035	44590	86	69	4058	76	2012-08-02 11:30:39.30669
44594	4	1-13745	44593	53	1	1638	15	2012-08-02 11:30:39.30669
44595	4	1-6761	44593	1	89	263	38	2012-08-02 11:30:39.30669
44596	4	1-13986	44593	89	86	225	26	2012-08-02 11:30:39.30669
44593	4	1-14035	44593	86	53	1487	76	2012-08-02 11:30:39.30669
44598	4	1-13772	44597	37	30	893	65	2012-08-02 11:30:39.704933
44599	4	1-10866	44597	30	82	372	9	2012-08-02 11:30:39.704933
44600	4	1-13857	44597	82	85	1962	14	2012-08-02 11:30:39.704933
44597	4	1-14036	44597	85	37	746	4	2012-08-02 11:30:39.704933
44602	2	1-11854	44601	57	93	5047	7	2012-08-02 11:30:40.103243
44601	2	1-14037	44601	93	57	6141	58	2012-08-02 11:30:40.103243
44604	4	1-10881	44603	36	89	2501	92	2012-08-02 11:30:40.103243
44605	4	1-13722	44603	89	56	1715	44	2012-08-02 11:30:40.103243
44606	4	1-13637	44603	56	93	1694	7	2012-08-02 11:30:40.103243
44603	4	1-14037	44603	93	36	2648	58	2012-08-02 11:30:40.103243
44608	4	1-10484	44607	38	31	1059	95	2012-08-02 11:30:40.501224
44609	4	1-13324	44607	31	57	775	99	2012-08-02 11:30:40.501224
44610	4	1-13293	44607	57	5	865	61	2012-08-02 11:30:40.501224
44607	4	1-14038	44607	5	38	1009	36	2012-08-02 11:30:40.501224
\.


--
-- Data for Name: tmvtremoved; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tmvtremoved (id, nb, oruuid, grp, own_src, own_dst, qtt, nat, created, deleted) FROM stdin;
\.


--
-- Data for Name: torder; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY torder (id, uuid, own, nr, qtt_requ, np, qtt_prov, qtt, start, created, updated) FROM stdin;
13895	1-13895	14	24	6065	44	2998	2998	0	2012-08-02 11:29:42.966515	\N
731	1-731	90	2	9251	59	414	414	0	2012-08-02 04:45:07.701836	\N
6343	1-6343	63	76	5112	4	522	522	0	2012-08-02 05:25:05.780439	\N
6344	1-6344	66	8	7530	61	1602	1602	0	2012-08-02 05:25:06.167078	\N
2775	1-2775	30	55	6977	29	1344	79	0	2012-08-02 05:01:05.927263	2012-08-02 05:17:20.288562
3513	1-3513	75	19	3844	37	4743	1471	2	2012-08-02 05:06:07.022056	2012-08-02 10:38:04.507152
8106	1-8106	4	43	9704	76	653	653	0	2012-08-02 05:37:05.739741	\N
12525	1-12525	78	64	8249	92	3175	3175	3	2012-08-02 11:20:43.338024	\N
12216	1-12216	46	55	4621	62	1404	1404	5	2012-08-02 11:18:43.1905	\N
5464	1-5464	61	33	7454	78	1920	1657	0	2012-08-02 05:19:06.793759	2012-08-02 11:30:29.07869
728	1-728	26	21	5134	17	1499	863	0	2012-08-02 04:45:06.895302	2012-08-02 11:20:44.491589
4563	1-4563	15	97	9471	13	3791	3613	5	2012-08-02 05:13:06.776998	2012-08-02 11:16:55.302803
8908	1-8908	22	58	4458	9	525	491	3	2012-08-02 05:42:29.974839	2012-08-02 05:43:03.165069
4263	1-4263	31	62	3630	33	624	624	0	2012-08-02 05:11:05.906178	\N
6935	1-6935	65	47	7686	10	1923	1923	0	2012-08-02 05:29:05.95254	\N
1677	1-1677	96	67	8340	24	978	613	0	2012-08-02 04:53:07.159155	2012-08-02 05:41:02.524399
8096	1-8096	99	31	3410	15	1402	1070	7	2012-08-02 05:37:01.70442	2012-08-02 05:46:51.430417
7528	1-7528	74	98	4281	100	1906	328	0	2012-08-02 05:33:05.793976	2012-08-02 11:11:37.083093
8999	1-8999	36	83	9876	37	7511	3705	5	2012-08-02 05:43:07.26559	2012-08-02 11:27:28.61077
8913	1-8913	24	3	9621	39	2800	1002	0	2012-08-02 05:42:32.01877	2012-08-02 11:26:13.358762
10872	1-10872	87	95	9364	71	3979	3979	3	2012-08-02 10:38:32.409091	\N
10223	1-10223	3	5	2228	49	48	48	11	2012-08-02 10:22:32.412644	\N
13904	1-13904	7	91	4318	15	3624	3624	10	2012-08-02 11:29:46.947711	\N
8090	1-8090	8	3	3420	81	1647	1647	14	2012-08-02 05:36:59.262734	\N
43	1-43	35	8	8751	40	8091	795	0	2012-08-02 04:43:53.922299	2012-08-02 11:27:33.74851
3064	1-3064	44	27	7704	72	1618	1112	0	2012-08-02 05:03:05.206284	2012-08-02 05:28:24.646396
4114	1-4114	41	98	9162	6	3327	1459	0	2012-08-02 05:10:07.286456	2012-08-02 11:21:02.460876
2221	1-2221	2	78	6720	8	67	67	0	2012-08-02 04:57:06.588005	\N
1444	1-1444	60	100	8152	83	55	55	0	2012-08-02 04:51:08.445381	\N
12222	1-12222	47	59	8934	39	6809	2759	22	2012-08-02 11:18:45.522058	2012-08-02 11:26:40.03285
10541	1-10541	21	23	9182	50	8891	3942	9	2012-08-02 10:35:32.237399	2012-08-02 11:30:33.358312
2360	1-2360	7	40	8966	58	407	407	0	2012-08-02 04:58:07.060654	\N
9550	1-9550	39	32	8223	40	608	608	8	2012-08-02 05:46:51.839477	\N
8998	1-8998	24	72	9340	51	144	144	0	2012-08-02 05:43:06.845374	\N
738	1-738	48	99	1434	95	242	27	0	2012-08-02 04:45:09.43705	2012-08-02 04:46:45.023656
13445	1-13445	68	85	2900	2	160	160	6	2012-08-02 11:26:44.808817	\N
1676	1-1676	52	97	9278	1	278	278	4	2012-08-02 04:53:06.706319	\N
2208	1-2208	42	8	6240	10	1084	1084	0	2012-08-02 04:57:00.930461	\N
3068	1-3068	78	33	8937	76	2021	2021	3	2012-08-02 05:03:06.819791	\N
3214	1-3214	18	22	9185	41	935	935	0	2012-08-02 05:04:06.215481	\N
8101	1-8101	98	30	9901	27	3222	2791	10	2012-08-02 05:37:03.706107	2012-08-02 05:40:39.0004
10222	1-10222	33	41	4179	4	773	641	0	2012-08-02 10:22:32.02597	2012-08-02 10:27:44.572949
10881	1-10881	36	58	8198	92	8322	2617	4	2012-08-02 10:38:35.513636	2012-08-02 11:30:40.103243
4861	1-4861	18	62	8713	79	3875	1115	0	2012-08-02 05:15:07.904922	2012-08-02 11:15:33.874328
2359	1-2359	67	50	9753	42	2458	1294	0	2012-08-02 04:58:06.6299	2012-08-02 05:03:23.514099
2779	1-2779	83	7	8382	35	296	296	0	2012-08-02 05:01:07.562728	\N
1806	1-1806	82	8	9779	92	4492	2107	5	2012-08-02 04:54:05.198928	2012-08-02 11:13:55.356002
3067	1-3067	10	86	8928	45	1113	923	0	2012-08-02 05:03:06.422202	2012-08-02 05:03:15.360152
2211	1-2211	97	53	9780	40	4088	3812	0	2012-08-02 04:57:02.25657	2012-08-02 05:01:02.964724
38	1-38	33	54	1924	18	16	16	0	2012-08-02 04:43:53.866974	\N
4111	1-4111	81	41	8990	78	3152	2305	0	2012-08-02 05:10:06.126037	2012-08-02 11:21:19.765527
12088	1-12088	4	63	7322	34	490	490	0	2012-08-02 11:17:54.483025	\N
13910	1-13910	47	70	5334	93	8072	8072	11	2012-08-02 11:29:49.345428	\N
6349	1-6349	100	12	1237	29	4	4	4	2012-08-02 05:25:08.16661	\N
2362	1-2362	5	21	4923	54	949	949	0	2012-08-02 04:58:07.922473	\N
10546	1-10546	96	35	6898	92	8723	8723	4	2012-08-02 10:35:32.616371	\N
6057	1-6057	11	18	4669	12	383	383	0	2012-08-02 05:23:08.160733	\N
2365	1-2365	82	54	6798	8	563	268	0	2012-08-02 04:58:09.225888	2012-08-02 05:03:31.295058
13448	1-13448	6	68	5551	65	5103	4243	8	2012-08-02 11:26:46.046558	2012-08-02 11:27:29.384122
10878	1-10878	48	66	4373	87	3516	1851	2	2012-08-02 10:38:34.353258	2012-08-02 11:19:58.568741
13913	1-13913	59	99	1383	36	39	39	4	2012-08-02 11:29:50.494159	\N
12096	1-12096	94	34	7454	52	3638	3638	0	2012-08-02 11:17:57.156808	\N
6940	1-6940	49	12	6822	83	1436	1436	0	2012-08-02 05:29:07.941372	\N
1560	1-1560	95	66	3005	14	835	835	0	2012-08-02 04:52:09.518007	\N
5467	1-5467	5	97	3947	80	1632	1632	0	2012-08-02 05:19:07.953741	\N
991	1-991	96	34	5417	41	330	200	0	2012-08-02 04:47:06.604057	2012-08-02 04:47:13.045032
2361	1-2361	6	89	9946	91	2928	1092	0	2012-08-02 04:58:07.491522	2012-08-02 05:30:10.487041
13919	1-13919	40	64	9237	37	7994	7994	7	2012-08-02 11:29:52.847302	\N
5469	1-5469	90	30	9766	82	1831	1831	0	2012-08-02 05:19:08.727013	\N
13916	1-13916	13	6	1908	42	1589	127	11	2012-08-02 11:29:51.654221	2012-08-02 11:30:01.554338
13898	1-13898	56	63	1085	56	5415	1933	5	2012-08-02 11:29:44.549699	2012-08-02 11:30:11.709306
5461	1-5461	81	45	7054	26	1231	602	2	2012-08-02 05:19:05.611075	2012-08-02 11:30:32.230146
4116	1-4116	30	87	7643	37	1201	1201	0	2012-08-02 05:10:08.059684	\N
1562	1-1562	65	89	4843	64	718	718	0	2012-08-02 04:52:11.727054	\N
1329	1-1329	90	35	2765	59	213	213	0	2012-08-02 04:50:11.19104	\N
10547	1-10547	70	67	9037	60	5320	4492	17	2012-08-02 10:35:32.694003	2012-08-02 11:15:39.333955
9007	1-9007	22	4	9559	98	3633	2607	0	2012-08-02 05:43:10.5579	2012-08-02 10:38:31.213882
3514	1-3514	26	57	8214	51	379	379	0	2012-08-02 05:06:07.419739	\N
2227	1-2227	71	92	5186	38	582	8	0	2012-08-02 04:57:09.17343	2012-08-02 05:28:33.695249
11289	1-11289	34	57	9052	66	894	894	2	2012-08-02 11:12:43.413024	\N
8701	1-8701	82	31	5800	92	786	786	6	2012-08-02 05:41:06.151317	\N
9588	1-9588	75	53	3851	83	1702	1702	3	2012-08-02 05:47:07.266441	\N
6064	1-6064	65	19	9490	79	517	517	0	2012-08-02 05:23:10.967451	\N
11753	1-11753	37	76	4509	61	1133	1133	12	2012-08-02 11:15:42.461057	\N
11762	1-11762	60	10	7171	96	6000	227	2	2012-08-02 11:15:46.723045	2012-08-02 11:23:39.862557
8408	1-8408	53	89	8732	37	5613	4388	5	2012-08-02 05:39:08.280314	2012-08-02 10:21:14.409656
451	1-451	48	91	9832	5	588	588	6	2012-08-02 04:44:06.954221	\N
5160	1-5160	50	28	8372	54	5785	2905	0	2012-08-02 05:17:07.359361	2012-08-02 05:48:11.365608
2080	1-2080	96	67	8381	92	2918	1265	0	2012-08-02 04:56:05.630812	2012-08-02 05:06:39.650504
13896	1-13896	9	85	5558	33	7234	3246	0	2012-08-02 11:29:43.786822	2012-08-02 11:30:32.970668
6351	1-6351	89	86	8801	13	1828	1827	4	2012-08-02 05:25:08.961886	2012-08-02 05:38:08.492792
6065	1-6065	34	20	7768	92	417	417	0	2012-08-02 05:23:11.364989	\N
2215	1-2215	84	26	9324	61	1826	1200	0	2012-08-02 04:57:03.968706	2012-08-02 04:57:07.018686
1682	1-1682	90	67	6404	46	1455	1455	0	2012-08-02 04:53:09.424151	\N
12217	1-12217	88	14	7377	91	3147	3141	0	2012-08-02 11:18:43.575894	2012-08-02 11:29:31.862809
5475	1-5475	33	24	5708	46	1932	1932	0	2012-08-02 05:19:11.092446	\N
890	1-890	92	19	7234	88	4445	3023	0	2012-08-02 04:46:10.479626	2012-08-02 05:17:20.288562
12994	1-12994	20	79	9894	23	2020	1629	0	2012-08-02 11:23:46.891322	2012-08-02 11:28:44.502577
94	1-94	45	30	9343	7	447	447	0	2012-08-02 04:43:54.526838	\N
1335	1-1335	2	9	6701	13	602	249	0	2012-08-02 04:50:13.147154	2012-08-02 04:54:39.493989
1002	1-1002	3	69	9862	87	955	955	0	2012-08-02 04:47:12.558689	\N
3364	1-3364	78	56	3011	77	190	190	5	2012-08-02 05:05:06.567533	\N
8110	1-8110	12	63	6235	77	3768	889	15	2012-08-02 05:37:07.364052	2012-08-02 05:41:32.077116
2369	1-2369	47	21	5544	28	169	169	0	2012-08-02 04:58:10.993642	\N
8407	1-8407	43	57	1746	70	20	20	4	2012-08-02 05:39:07.882642	\N
5164	1-5164	70	18	2170	69	33	33	3	2012-08-02 05:17:08.928589	\N
6352	1-6352	56	75	6181	24	746	746	3	2012-08-02 05:25:09.404119	\N
9589	1-9589	67	26	5353	42	1977	1977	4	2012-08-02 05:47:07.664194	\N
766	1-766	66	6	9248	81	3029	340	0	2012-08-02 04:45:20.761661	2012-08-02 04:58:37.753995
8705	1-8705	55	23	7284	58	3910	670	4	2012-08-02 05:41:07.786438	2012-08-02 11:29:20.4272
2640	1-2640	82	58	6095	6	134	134	0	2012-08-02 05:00:08.846371	\N
12526	1-12526	50	76	4311	2	1580	1580	5	2012-08-02 11:20:42.54078	\N
5483	1-5483	28	61	8475	48	5490	2370	0	2012-08-02 05:19:14.274574	2012-08-02 11:28:58.338996
11295	1-11295	26	8	7639	81	5610	4394	5	2012-08-02 11:12:45.746122	2012-08-02 11:29:15.596151
464	1-464	83	65	1435	23	41	41	0	2012-08-02 04:44:08.049662	\N
12986	1-12986	18	93	6030	24	3402	629	3	2012-08-02 11:23:43.365977	2012-08-02 11:29:35.795893
6068	1-6068	37	15	9933	80	918	918	0	2012-08-02 05:23:12.569211	\N
9590	1-9590	70	18	5853	54	322	322	7	2012-08-02 05:47:08.094949	\N
894	1-894	73	33	1666	52	104	104	0	2012-08-02 04:46:12.965148	\N
9884	1-9884	51	54	8369	14	3408	2806	4	2012-08-02 05:49:08.257768	2012-08-02 10:38:07.987362
11298	1-11298	14	60	9040	75	1524	1524	10	2012-08-02 11:12:46.939702	\N
9004	1-9004	55	10	1024	80	112	112	0	2012-08-02 05:43:09.353697	\N
6353	1-6353	54	79	3911	4	695	380	8	2012-08-02 05:25:09.801456	2012-08-02 11:19:47.352458
8409	1-8409	18	33	8644	24	102	102	12	2012-08-02 05:39:08.666783	\N
5768	1-5768	98	36	9296	71	416	416	3	2012-08-02 05:21:09.745101	\N
1687	1-1687	15	100	4408	25	316	316	0	2012-08-02 04:53:11.721427	\N
12992	1-12992	52	58	1706	92	1636	1636	6	2012-08-02 11:23:45.719613	\N
11292	1-11292	9	68	6378	80	6228	3138	0	2012-08-02 11:12:44.585865	2012-08-02 11:14:49.633653
5769	1-5769	13	89	5086	80	503	503	0	2012-08-02 05:21:10.143028	\N
12997	1-12997	71	66	7152	13	3395	3395	8	2012-08-02 11:23:47.69745	\N
3519	1-3519	11	39	9928	23	164	164	0	2012-08-02 05:06:09.375484	\N
11765	1-11765	90	78	9314	58	3350	3350	0	2012-08-02 11:15:48.027105	\N
6070	1-6070	84	98	4770	39	706	706	0	2012-08-02 05:23:13.33145	\N
105	1-105	32	73	9666	4	1302	1242	0	2012-08-02 04:43:54.660695	2012-08-02 05:15:00.289234
2498	1-2498	44	13	7586	48	1273	345	0	2012-08-02 04:59:07.797585	2012-08-02 05:18:01.11501
7820	1-7820	96	9	7942	36	1978	1736	0	2012-08-02 05:35:07.449118	2012-08-02 11:18:15.244833
1004	1-1004	29	33	9247	81	3717	279	0	2012-08-02 04:47:13.045032	2012-08-02 05:27:33.462715
10229	1-10229	87	17	8634	76	1735	1735	5	2012-08-02 10:22:34.755293	\N
474	1-474	50	1	6168	23	348	348	3	2012-08-02 04:44:09.246091	\N
5479	1-5479	60	19	7802	5	6392	6392	0	2012-08-02 05:19:12.672369	\N
11301	1-11301	88	82	5336	45	2936	2006	0	2012-08-02 11:12:48.112644	2012-08-02 11:16:52.97174
780	1-780	22	91	8583	51	55	55	0	2012-08-02 04:45:26.287608	\N
2229	1-2229	61	19	3473	81	1890	1033	0	2012-08-02 04:57:10.046391	2012-08-02 05:14:37.040092
10231	1-10231	9	6	8619	29	3261	3261	0	2012-08-02 10:22:35.550872	\N
132	1-132	74	71	5269	94	102	102	0	2012-08-02 04:43:54.962075	\N
5171	1-5171	46	39	8835	69	675	675	9	2012-08-02 05:17:11.690846	\N
8708	1-8708	25	36	6575	86	1652	177	14	2012-08-02 05:41:09.001503	2012-08-02 05:41:17.42052
2233	1-2233	56	54	8346	29	852	852	0	2012-08-02 04:57:11.792033	\N
4873	1-4873	56	96	9894	7	3970	3970	0	2012-08-02 05:15:12.555996	\N
1696	1-1696	55	51	4728	37	1020	708	0	2012-08-02 04:53:15.323407	2012-08-02 04:58:41.709167
5778	1-5778	49	34	6805	71	1013	851	0	2012-08-02 05:21:13.733135	2012-08-02 05:31:06.264442
11754	1-11754	33	13	2049	43	927	927	8	2012-08-02 11:15:43.661661	\N
8710	1-8710	74	67	3727	37	2432	488	0	2012-08-02 05:41:09.786104	2012-08-02 10:21:02.36707
2378	1-2378	27	87	6276	32	1216	899	0	2012-08-02 04:58:14.904415	2012-08-02 04:59:28.89117
10891	1-10891	99	18	7772	97	6580	2932	0	2012-08-02 10:38:39.393169	2012-08-02 11:28:43.735007
12218	1-12218	98	76	2265	40	55	55	0	2012-08-02 11:18:43.974211	\N
13897	1-13897	60	29	8948	72	6735	6735	9	2012-08-02 11:29:44.163115	\N
9592	1-9592	37	4	6613	76	2603	2532	0	2012-08-02 05:47:08.923464	2012-08-02 11:13:19.348013
8112	1-8112	56	22	7774	79	1207	1207	10	2012-08-02 05:37:08.170641	\N
13446	1-13446	80	66	9755	39	1287	1287	0	2012-08-02 11:26:45.207042	\N
10551	1-10551	63	69	4309	79	1318	1318	2	2012-08-02 10:35:33.03142	\N
13900	1-13900	89	74	8128	97	8570	8570	4	2012-08-02 11:29:45.356304	\N
6355	1-6355	63	6	6825	55	1828	143	0	2012-08-02 05:25:10.734895	2012-08-02 05:46:53.032769
3075	1-3075	3	77	4219	81	62	62	9	2012-08-02 05:03:09.681617	\N
1345	1-1345	100	89	8961	84	690	690	0	2012-08-02 04:50:18.936041	\N
7233	1-7233	29	43	4423	79	799	799	12	2012-08-02 05:31:07.071497	\N
799	1-799	57	67	9048	47	178	178	0	2012-08-02 04:45:32.675137	\N
3226	1-3226	63	19	5937	17	2705	2705	0	2012-08-02 05:04:11.198757	\N
481	1-481	60	91	3454	65	289	289	0	2012-08-02 04:44:09.997679	\N
4571	1-4571	79	9	7472	99	482	482	5	2012-08-02 05:13:10.069658	\N
6649	1-6649	53	1	9080	69	1888	1888	7	2012-08-02 05:27:10.20236	\N
1009	1-1009	75	33	8673	72	3486	10	0	2012-08-02 04:47:15.563661	2012-08-02 05:26:21.537143
4877	1-4877	42	98	3175	47	238	238	13	2012-08-02 05:15:14.124972	\N
8711	1-8711	69	58	5492	47	1645	881	9	2012-08-02 05:41:10.183791	2012-08-02 11:25:17.553175
5773	1-5773	77	29	6397	60	551	551	10	2012-08-02 05:21:11.744709	\N
6648	1-6648	95	44	8820	79	1392	530	0	2012-08-02 05:27:09.815832	2012-08-02 11:24:49.983851
7235	1-7235	35	84	3791	51	386	386	0	2012-08-02 05:31:07.855719	\N
13906	1-13906	96	23	4329	100	1339	1339	13	2012-08-02 11:29:47.765896	\N
194	1-194	88	61	9785	65	3403	1411	0	2012-08-02 04:43:55.85513	2012-08-02 05:27:53.674302
9888	1-9888	98	68	9241	27	4795	1922	0	2012-08-02 05:49:09.883134	2012-08-02 11:25:26.565595
7236	1-7236	97	90	8175	27	1624	1624	0	2012-08-02 05:31:08.242315	\N
12251	1-12251	47	86	5930	7	3408	2863	3	2012-08-02 11:18:56.713633	2012-08-02 11:21:34.980127
1692	1-1692	58	38	5155	14	1186	1186	0	2012-08-02 04:53:13.997759	\N
1219	1-1219	87	38	5635	51	1342	1020	0	2012-08-02 04:49:09.111449	2012-08-02 11:26:56.88707
220	1-220	66	25	9282	51	432	316	0	2012-08-02 04:43:56.356183	2012-08-02 04:44:23.850508
2937	1-2937	1	86	1453	90	34	34	0	2012-08-02 05:02:13.594486	\N
7826	1-7826	44	49	943	71	292	292	7	2012-08-02 05:35:10.023254	\N
2509	1-2509	43	48	9747	71	357	357	0	2012-08-02 04:59:12.559288	\N
181	1-181	35	74	9216	27	420	420	0	2012-08-02 04:43:55.687989	\N
7537	1-7537	93	51	6174	80	2192	2188	0	2012-08-02 05:33:09.495737	2012-08-02 05:39:55.1457
12254	1-12254	66	20	6742	71	9303	6747	0	2012-08-02 11:18:57.851469	2012-08-02 11:28:53.543122
1952	1-1952	39	75	4769	54	2126	223	0	2012-08-02 04:55:10.353306	2012-08-02 05:44:59.287494
12993	1-12993	39	56	7678	45	3825	1350	0	2012-08-02 11:23:46.117465	2012-08-02 11:23:46.117465
3523	1-3523	59	23	8457	29	145	145	0	2012-08-02 05:06:11.03274	\N
13451	1-13451	44	79	5234	34	4131	3892	6	2012-08-02 11:26:46.421913	2012-08-02 11:26:54.810076
13908	1-13908	43	49	8823	39	7343	7343	4	2012-08-02 11:29:48.947498	\N
4273	1-4273	39	6	748	89	53	53	0	2012-08-02 05:11:10.027492	\N
187	1-187	83	35	9325	76	273	273	0	2012-08-02 04:43:55.754917	\N
1582	1-1582	100	86	7672	77	2822	1063	0	2012-08-02 04:52:22.354728	2012-08-02 05:38:48.356361
12996	1-12996	54	59	4889	76	3906	3906	0	2012-08-02 11:23:47.288768	\N
13912	1-13912	24	87	5350	66	2046	2046	17	2012-08-02 11:29:49.720915	\N
4576	1-4576	63	78	4978	30	63	63	7	2012-08-02 05:13:12.113655	\N
8712	1-8712	36	30	9384	70	1423	1423	0	2012-08-02 05:41:10.581388	\N
2793	1-2793	81	23	9249	97	638	638	0	2012-08-02 05:01:13.431285	\N
192	1-192	51	35	7400	9	207	207	0	2012-08-02 04:43:55.822685	\N
2511	1-2511	99	16	9343	14	1013	1013	5	2012-08-02 04:59:13.409856	\N
4577	1-4577	50	30	5161	61	1371	1371	0	2012-08-02 05:13:12.522161	\N
13915	1-13915	16	36	3582	21	5053	5053	0	2012-08-02 11:29:51.256776	\N
8713	1-8713	95	69	7620	9	85	85	8	2012-08-02 05:41:10.990199	\N
12242	1-12242	92	74	7612	82	5187	5187	3	2012-08-02 11:18:53.200205	\N
7240	1-7240	25	68	5405	2	4131	4131	6	2012-08-02 05:31:09.777787	\N
13918	1-13918	4	53	9803	87	3892	3892	6	2012-08-02 11:29:52.450102	\N
7241	1-7241	21	9	8462	31	1272	1272	0	2012-08-02 05:31:10.175607	\N
13920	1-13920	17	68	6677	76	2997	2997	7	2012-08-02 11:29:53.653874	\N
7242	1-7242	94	86	2324	40	1256	1256	6	2012-08-02 05:31:10.573473	\N
12248	1-12248	43	35	9061	19	2621	2621	0	2012-08-02 11:18:55.553375	\N
1019	1-1019	56	99	9230	93	3497	2228	0	2012-08-02 04:47:20.448657	2012-08-02 05:04:39.373687
6957	1-6957	44	67	9460	29	657	657	8	2012-08-02 05:29:14.661489	\N
5497	1-5497	90	3	9700	53	3808	1677	0	2012-08-02 05:19:19.845712	2012-08-02 11:11:40.607873
498	1-498	63	24	9547	81	5178	4374	0	2012-08-02 04:44:11.977739	2012-08-02 05:14:21.714887
2380	1-2380	11	25	7689	54	2777	22	0	2012-08-02 04:58:15.810556	2012-08-02 11:20:05.098618
5498	1-5498	94	98	8755	85	1285	1285	0	2012-08-02 05:19:20.243357	\N
4282	1-4282	73	19	4441	45	1373	1373	5	2012-08-02 05:11:13.672901	\N
1959	1-1959	94	86	9439	40	701	701	0	2012-08-02 04:55:13.071341	\N
209	1-209	22	51	5291	9	157	157	0	2012-08-02 04:43:56.15573	\N
2657	1-2657	40	23	3754	15	989	366	0	2012-08-02 05:00:15.994153	2012-08-02 05:43:46.24489
259	1-259	82	93	3216	10	404	262	0	2012-08-02 04:43:57.347046	2012-08-02 04:50:08.870915
1589	1-1589	40	46	9617	52	1903	326	0	2012-08-02 04:52:25.251577	2012-08-02 05:36:33.995764
3235	1-3235	88	7	9901	37	1691	1691	0	2012-08-02 05:04:14.778242	\N
8716	1-8716	18	69	5368	45	1229	796	9	2012-08-02 05:41:12.194452	2012-08-02 05:42:24.307075
5505	1-5505	96	38	3297	32	609	578	0	2012-08-02 05:19:23.204216	2012-08-02 05:25:32.563533
8414	1-8414	83	61	5436	44	494	494	11	2012-08-02 05:39:10.767764	\N
2656	1-2656	28	74	7471	94	424	424	0	2012-08-02 05:00:15.562918	\N
5784	1-5784	73	1	7482	44	4335	3482	5	2012-08-02 05:21:16.141526	2012-08-02 05:37:36.765697
12528	1-12528	36	29	2387	97	725	668	16	2012-08-02 11:20:44.104773	2012-08-02 11:25:26.565595
7244	1-7244	76	26	8242	81	4744	408	0	2012-08-02 05:31:11.368763	2012-08-02 11:20:17.19583
7249	1-7249	84	24	9885	80	8825	5493	0	2012-08-02 05:31:13.291334	2012-08-02 11:26:26.596176
4582	1-4582	45	50	8782	11	400	400	0	2012-08-02 05:13:14.554881	\N
820	1-820	13	99	9760	13	5164	1104	0	2012-08-02 04:45:41.082638	2012-08-02 05:15:14.898584
13444	1-13444	58	46	4798	85	3108	3108	5	2012-08-02 11:26:44.41117	\N
5501	1-5501	96	49	7351	79	253	253	11	2012-08-02 05:19:21.60244	\N
819	1-819	50	73	3865	87	93	93	2	2012-08-02 04:45:40.994006	\N
9890	1-9890	62	68	9768	73	762	762	3	2012-08-02 05:49:10.723646	\N
13449	1-13449	50	66	1668	52	3483	3479	3	2012-08-02 11:26:46.808786	2012-08-02 11:28:06.634738
255	1-255	24	24	4078	10	731	731	0	2012-08-02 04:43:57.258386	\N
849	1-849	22	62	9343	17	3939	1759	0	2012-08-02 04:45:52.565107	2012-08-02 11:20:44.491589
9011	1-9011	67	53	9432	78	27	27	2	2012-08-02 05:43:12.20431	\N
3693	1-3693	33	15	7129	14	1211	826	6	2012-08-02 05:07:19.439193	2012-08-02 11:18:46.658793
815	1-815	46	89	5548	95	1077	1077	0	2012-08-02 04:45:39.292404	\N
5781	1-5781	61	65	7909	14	313	313	0	2012-08-02 05:21:14.926403	\N
7246	1-7246	4	35	8892	54	1547	1547	2	2012-08-02 05:31:12.142179	\N
5780	1-5780	48	95	5953	94	409	343	2	2012-08-02 05:21:14.5286	2012-08-02 05:24:13.133182
4287	1-4287	43	90	7862	13	291	291	0	2012-08-02 05:11:15.738863	\N
8717	1-8717	5	77	3843	15	1295	606	8	2012-08-02 05:41:12.581129	2012-08-02 05:49:19.635148
13905	1-13905	44	15	9869	2	102	102	8	2012-08-02 11:29:47.357006	\N
3532	1-3532	80	89	9014	85	3752	3752	7	2012-08-02 05:06:14.546209	\N
1964	1-1964	75	73	9761	51	32	32	0	2012-08-02 04:55:15.314094	\N
2665	1-2665	50	39	9507	47	2796	1104	0	2012-08-02 05:00:19.343181	2012-08-02 05:18:53.72762
13909	1-13909	79	7	2834	54	900	900	0	2012-08-02 11:29:48.527692	\N
2797	1-2797	61	50	6791	23	550	550	5	2012-08-02 05:01:15.077444	\N
1588	1-1588	86	89	9493	60	2986	1901	0	2012-08-02 04:52:24.776178	2012-08-02 05:41:34.154091
8416	1-8416	65	97	7601	42	752	752	0	2012-08-02 05:39:11.562977	\N
12531	1-12531	94	54	6769	75	349	349	2	2012-08-02 11:20:45.297555	\N
3086	1-3086	90	19	8417	46	381	381	2	2012-08-02 05:03:14.144804	\N
6364	1-6364	10	18	4218	5	670	670	0	2012-08-02 05:25:14.343202	\N
13452	1-13452	50	75	9248	81	649	649	2	2012-08-02 11:26:47.604013	\N
10553	1-10553	59	32	2901	86	331	331	9	2012-08-02 10:35:33.199942	\N
12229	1-12229	82	23	9139	87	5209	5209	10	2012-08-02 11:18:48.205818	\N
825	1-825	93	61	6573	37	2357	2357	0	2012-08-02 04:45:44.309005	\N
3529	1-3529	28	97	8236	96	1490	588	0	2012-08-02 05:06:13.385865	2012-08-02 05:28:07.430377
1590	1-1590	48	60	1305	67	77	77	0	2012-08-02 04:52:25.748438	\N
1039	1-1039	45	70	5767	14	1933	438	0	2012-08-02 04:47:31.938973	2012-08-02 10:38:03.7003
2943	1-2943	49	50	3531	59	601	105	0	2012-08-02 05:02:16.069121	2012-08-02 05:17:21.869166
7539	1-7539	36	19	9752	52	8105	4957	0	2012-08-02 05:33:10.313094	2012-08-02 10:38:04.507152
524	1-524	6	47	7248	31	1072	858	0	2012-08-02 04:44:16.333235	2012-08-02 04:44:52.347639
5502	1-5502	31	60	8935	82	1113	993	0	2012-08-02 05:19:22.000117	2012-08-02 11:12:13.224419
1044	1-1044	18	85	3910	17	332	332	0	2012-08-02 04:47:34.755951	\N
6366	1-6366	91	8	4296	40	3290	3290	6	2012-08-02 05:25:15.149336	\N
3537	1-3537	88	18	7447	65	2189	1961	0	2012-08-02 05:06:16.534515	2012-08-02 05:44:08.114683
6656	1-6656	10	52	5620	99	1213	1213	0	2012-08-02 05:27:12.998626	\N
12534	1-12534	97	30	5425	70	2081	2081	0	2012-08-02 11:20:46.446918	\N
6079	1-6079	38	80	5728	85	2547	1417	0	2012-08-02 05:23:16.946298	2012-08-02 11:13:25.555476
13455	1-13455	84	44	8006	39	723	723	4	2012-08-02 11:26:48.81964	\N
6961	1-6961	59	19	6235	87	5497	3408	0	2012-08-02 05:29:16.275878	2012-08-02 05:41:23.083455
6657	1-6657	48	89	5549	87	342	342	0	2012-08-02 05:27:13.3851	\N
1464	1-1464	88	57	6737	17	217	217	2	2012-08-02 04:51:20.765918	\N
13902	1-13902	36	77	5902	20	7916	2833	0	2012-08-02 11:29:46.141041	2012-08-02 11:30:09.753486
6082	1-6082	40	60	5799	19	228	228	0	2012-08-02 05:23:18.128553	\N
1968	1-1968	100	100	5412	88	863	863	0	2012-08-02 04:55:17.070639	\N
9013	1-9013	22	86	9722	53	2900	2900	0	2012-08-02 05:43:12.999721	\N
13458	1-13458	94	53	7614	85	9206	9206	4	2012-08-02 11:26:49.980035	\N
10235	1-10235	21	50	9420	65	2415	2099	5	2012-08-02 10:22:37.108612	2012-08-02 11:13:49.190803
1969	1-1969	42	45	6585	31	260	260	0	2012-08-02 04:55:17.512431	\N
13461	1-13461	79	23	9808	47	8162	5919	11	2012-08-02 11:26:51.162057	2012-08-02 11:29:20.4272
8117	1-8117	99	25	4675	40	1907	697	2	2012-08-02 05:37:10.148176	2012-08-02 11:27:01.561032
246	1-246	23	79	9209	17	1747	474	0	2012-08-02 04:43:57.028272	2012-08-02 04:48:03.546719
279	1-279	39	8	8925	30	105	105	0	2012-08-02 04:43:58.245477	\N
1462	1-1462	83	53	7524	79	470	332	0	2012-08-02 04:51:18.708512	2012-08-02 05:11:51.129193
276	1-276	50	20	5955	64	50	50	0	2012-08-02 04:43:58.123075	\N
520	1-520	90	21	8410	69	573	197	0	2012-08-02 04:44:15.537337	2012-08-02 04:45:53.979782
13911	1-13911	73	69	8654	58	4380	4380	5	2012-08-02 11:29:50.118549	\N
3980	1-3980	37	26	7842	73	2342	685	0	2012-08-02 05:09:13.683265	2012-08-02 11:28:30.922001
358	1-358	67	69	8194	20	833	652	0	2012-08-02 04:44:01.187165	2012-08-02 04:52:05.927995
870	1-870	47	83	6780	19	262	108	0	2012-08-02 04:46:01.625883	2012-08-02 05:12:19.19552
12532	1-12532	96	100	8117	85	712	712	2	2012-08-02 11:20:45.684813	\N
7255	1-7255	95	71	9240	8	1388	580	2	2012-08-02 05:31:15.589332	2012-08-02 05:45:00.911853
3969	1-3969	55	38	6629	21	579	579	6	2012-08-02 05:09:09.319491	\N
11764	1-11764	23	74	6840	85	8471	5312	5	2012-08-02 11:15:47.640262	2012-08-02 11:19:05.134505
9018	1-9018	80	71	5469	81	1212	141	0	2012-08-02 05:43:15.054661	2012-08-02 11:27:02.776472
1374	1-1374	90	55	5440	53	1921	1707	3	2012-08-02 04:50:33.309641	2012-08-02 05:37:11.794453
7542	1-7542	90	19	8431	94	675	675	2	2012-08-02 05:33:11.53928	\N
1235	1-1235	55	86	8753	15	5746	291	0	2012-08-02 04:49:16.690888	2012-08-02 10:22:28.842044
309	1-309	9	64	6047	42	1165	1137	0	2012-08-02 04:43:59.223999	2012-08-02 04:48:06.299242
10241	1-10241	19	23	9122	29	4220	4220	0	2012-08-02 10:25:14.221798	\N
12998	1-12998	52	7	7213	2	3507	3507	7	2012-08-02 11:23:48.108398	\N
11299	1-11299	35	96	8257	50	795	795	10	2012-08-02 11:12:47.33703	\N
3978	1-3978	1	78	5844	77	1122	1081	0	2012-08-02 05:09:12.887715	2012-08-02 05:38:27.486033
13914	1-13914	65	98	6071	22	7888	7888	12	2012-08-02 11:29:50.870115	\N
5509	1-5509	37	89	6387	88	1382	1382	0	2012-08-02 05:19:24.806365	\N
12538	1-12538	34	23	6196	44	4872	2733	2	2012-08-02 11:20:48.016157	2012-08-02 11:27:41.683599
4298	1-4298	74	75	6112	46	124	124	2	2012-08-02 05:11:20.257754	\N
6663	1-6663	44	1	9565	83	692	692	0	2012-08-02 05:27:15.782984	\N
12230	1-12230	26	94	9878	66	1316	1316	10	2012-08-02 11:18:48.5818	\N
11761	1-11761	6	3	7367	9	1258	1258	0	2012-08-02 11:15:46.336916	\N
1706	1-1706	1	55	8687	87	1498	1498	0	2012-08-02 04:53:19.886646	\N
8402	1-8402	31	89	5684	70	3638	2101	3	2012-08-02 05:39:05.870769	2012-08-02 11:26:47.220183
8720	1-8720	15	96	8290	35	139	139	13	2012-08-02 05:41:13.785427	\N
535	1-535	26	8	7293	40	1020	1020	3	2012-08-02 04:44:19.69275	\N
13001	1-13001	38	65	7351	43	1189	1189	8	2012-08-02 11:23:49.246116	\N
12535	1-12535	75	36	4931	14	411	411	0	2012-08-02 11:20:46.822549	\N
13004	1-13004	30	42	3980	57	4714	4450	6	2012-08-02 11:23:50.417571	2012-08-02 11:27:46.491424
12237	1-12237	20	10	9357	58	5757	2806	0	2012-08-02 11:18:51.266583	2012-08-02 11:27:47.287245
5514	1-5514	69	62	9533	91	2843	174	9	2012-08-02 05:19:26.872357	2012-08-02 05:30:32.563513
10556	1-10556	15	40	3302	83	200	200	0	2012-08-02 10:35:33.423684	\N
320	1-320	12	92	7235	74	240	240	0	2012-08-02 04:43:59.678825	\N
10893	1-10893	74	29	5121	40	2960	2960	0	2012-08-02 10:38:40.188396	\N
2095	1-2095	17	27	5157	48	390	390	0	2012-08-02 04:56:11.796198	\N
13917	1-13917	53	75	1869	45	8384	7411	11	2012-08-02 11:29:52.052866	2012-08-02 11:30:31.832093
10245	1-10245	40	79	9305	34	2379	2379	6	2012-08-02 10:25:15.800947	\N
8419	1-8419	3	91	8913	40	5847	5847	0	2012-08-02 05:39:12.7571	\N
2400	1-2400	64	97	9298	42	1116	1116	0	2012-08-02 04:58:24.51593	\N
9596	1-9596	87	61	5643	31	164	164	8	2012-08-02 05:47:10.547791	\N
10558	1-10558	88	92	5879	72	2706	924	4	2012-08-02 10:35:33.57064	2012-08-02 10:38:45.281993
7545	1-7545	39	85	8171	36	511	511	5	2012-08-02 05:33:12.788027	\N
8420	1-8420	72	84	9342	61	1546	1546	9	2012-08-02 05:39:13.143663	\N
2519	1-2519	14	9	9943	81	3989	1472	0	2012-08-02 04:59:16.901879	2012-08-02 05:20:02.726631
900	1-900	21	59	6147	28	276	171	0	2012-08-02 04:46:16.555965	2012-08-02 04:56:11.354042
3701	1-3701	45	19	8799	40	3611	3611	0	2012-08-02 05:07:22.621384	\N
9017	1-9017	81	1	7321	95	1455	1455	5	2012-08-02 05:43:14.635095	\N
11758	1-11758	37	89	8687	12	2691	2683	12	2012-08-02 11:15:45.187687	2012-08-02 11:21:42.881726
9893	1-9893	1	79	6031	48	1451	1451	7	2012-08-02 05:49:11.97332	\N
2806	1-2806	96	56	6253	49	995	727	7	2012-08-02 05:01:18.878365	2012-08-02 05:47:56.600435
9301	1-9301	8	19	8681	10	5554	5554	0	2012-08-02 05:45:10.658443	\N
6090	1-6090	81	5	9798	86	1210	1127	7	2012-08-02 05:23:21.343702	2012-08-02 09:10:31.355931
3974	1-3974	7	26	4136	65	1599	348	0	2012-08-02 05:09:11.308228	2012-08-02 05:36:31.530066
11302	1-11302	4	76	8111	43	4344	4174	3	2012-08-02 11:12:48.499044	2012-08-02 11:21:09.963143
12227	1-12227	42	94	2967	22	2407	988	4	2012-08-02 11:18:47.808215	2012-08-02 11:19:47.352458
8422	1-8422	87	57	9859	3	3845	478	8	2012-08-02 05:39:13.939346	2012-08-02 11:20:21.119568
4902	1-4902	87	72	1195	98	115	115	0	2012-08-02 05:15:23.947163	\N
867	1-867	64	46	7171	69	133	133	0	2012-08-02 04:46:00.166985	\N
13011	1-13011	71	96	6220	73	1408	1408	0	2012-08-02 11:23:52.759618	\N
1375	1-1375	25	60	8509	13	2459	1152	0	2012-08-02 04:50:33.773346	2012-08-02 11:20:58.97991
6969	1-6969	9	1	6850	64	3220	3220	9	2012-08-02 05:29:19.50289	\N
10559	1-10559	82	1	8200	11	2332	2332	6	2012-08-02 10:35:33.648041	\N
906	1-906	7	56	9892	86	1611	1611	0	2012-08-02 04:46:20.102083	\N
10248	1-10248	39	11	7427	30	1146	1146	3	2012-08-02 10:25:16.987796	\N
10250	1-10250	31	86	7109	48	299	299	10	2012-08-02 10:25:17.786492	\N
5186	1-5186	11	61	7664	81	2425	971	0	2012-08-02 05:17:17.558162	2012-08-02 11:14:06.95889
1238	1-1238	60	14	7366	76	327	327	0	2012-08-02 04:49:18.546447	\N
8128	1-8128	38	19	6687	59	7056	7056	0	2012-08-02 05:37:14.64509	\N
5521	1-5521	42	27	4592	54	375	375	0	2012-08-02 05:19:29.678168	\N
8130	1-8130	45	24	5490	15	489	489	11	2012-08-02 05:37:15.529106	\N
1241	1-1241	11	83	9504	80	2231	2231	0	2012-08-02 04:49:20.435525	\N
2249	1-2249	75	38	9056	4	1197	1197	3	2012-08-02 04:57:18.74135	\N
4603	1-4603	19	82	6397	25	1123	1123	0	2012-08-02 05:13:23.139977	\N
359	1-359	32	22	6118	31	352	352	0	2012-08-02 04:44:01.242403	\N
7553	1-7553	52	66	9597	25	4144	1839	3	2012-08-02 05:33:16.046838	2012-08-02 11:12:04.253594
914	1-914	78	50	6653	76	1355	252	0	2012-08-02 04:46:24.951516	2012-08-02 11:24:12.922075
1251	1-1251	99	47	8397	68	2063	903	0	2012-08-02 04:49:27.637724	2012-08-02 11:29:20.814024
379	1-379	16	76	8251	10	117	117	0	2012-08-02 04:44:02.418685	\N
9897	1-9897	45	78	7457	80	4770	4470	0	2012-08-02 05:49:13.598917	2012-08-02 11:30:20.7384
5795	1-5795	31	36	4360	54	1398	37	0	2012-08-02 05:21:20.494567	2012-08-02 05:28:52.713097
7551	1-7551	79	66	8229	50	2500	2500	0	2012-08-02 05:33:15.196309	\N
10562	1-10562	94	44	5586	89	198	198	21	2012-08-02 10:35:33.86949	\N
13003	1-13003	15	80	8129	77	1152	1152	11	2012-08-02 11:23:50.030461	\N
9597	1-9597	32	31	8497	13	1801	1801	0	2012-08-02 05:47:10.945423	\N
1984	1-1984	71	70	393	51	16	16	0	2012-08-02 04:55:24.097028	\N
1075	1-1075	46	22	3659	28	336	205	0	2012-08-02 04:47:55.424511	2012-08-02 04:48:06.299242
1377	1-1377	99	48	7539	2	1621	937	0	2012-08-02 04:50:34.736706	2012-08-02 04:50:55.284861
13925	1-13925	30	5	4997	63	45	45	3	2012-08-02 11:29:54.802891	\N
2255	1-2255	74	8	6921	50	5181	1439	0	2012-08-02 04:57:21.403956	2012-08-02 11:27:33.74851
7552	1-7552	49	20	9377	87	321	321	0	2012-08-02 05:33:15.628065	\N
9895	1-9895	78	32	8258	65	423	423	0	2012-08-02 05:49:12.802871	\N
10244	1-10244	60	90	8202	43	4740	3790	0	2012-08-02 10:25:15.414494	2012-08-02 11:29:53.255963
3706	1-3706	86	68	7508	34	1312	1312	0	2012-08-02 05:07:24.598766	\N
12244	1-12244	67	2	8569	95	1616	527	4	2012-08-02 11:18:54.007244	2012-08-02 11:28:20.733231
3984	1-3984	72	96	5064	22	588	588	0	2012-08-02 05:09:15.263237	\N
13456	1-13456	34	78	5493	53	2272	2272	0	2012-08-02 11:26:49.195381	\N
5191	1-5191	47	32	8490	94	555	555	7	2012-08-02 05:17:19.493378	\N
13928	1-13928	90	83	4114	80	841	841	9	2012-08-02 11:29:56.382949	\N
2531	1-2531	29	59	8762	69	378	378	0	2012-08-02 04:59:22.027454	\N
554	1-554	6	50	8148	93	1168	404	0	2012-08-02 04:44:24.259937	2012-08-02 05:21:32.59489
13459	1-13459	61	62	9400	82	9728	9728	9	2012-08-02 11:26:50.38875	\N
1986	1-1986	84	65	6906	31	2409	462	5	2012-08-02 04:55:24.96985	2012-08-02 10:25:27.22918
12238	1-12238	39	4	6967	83	9209	7325	0	2012-08-02 11:18:51.664515	2012-08-02 11:28:16.811248
6373	1-6373	40	30	7041	52	2376	1482	0	2012-08-02 05:25:17.988981	2012-08-02 05:44:37.96589
6671	1-6671	40	71	7502	15	2029	2029	0	2012-08-02 05:27:19.020066	\N
13931	1-13931	98	99	9695	2	3364	3364	0	2012-08-02 11:29:57.60342	\N
4305	1-4305	10	25	7668	87	2104	2104	0	2012-08-02 05:11:23.152911	\N
11306	1-11306	50	60	6540	79	1662	1011	0	2012-08-02 11:12:50.056511	2012-08-02 11:23:56.637412
5792	1-5792	19	83	8967	14	6147	3402	0	2012-08-02 05:21:19.323853	2012-08-02 11:21:03.222696
3708	1-3708	83	64	6654	95	520	520	0	2012-08-02 05:07:25.360982	\N
12548	1-12548	17	10	7764	38	879	879	9	2012-08-02 11:20:51.860446	\N
9307	1-9307	28	48	9092	17	718	718	0	2012-08-02 05:45:13.078098	\N
6977	1-6977	19	3	9495	36	3899	3899	2	2012-08-02 05:29:22.67403	\N
578	1-578	29	64	9885	94	2268	321	0	2012-08-02 04:44:30.71472	2012-08-02 11:29:35.033431
10565	1-10565	20	77	9004	55	95	95	8	2012-08-02 10:35:34.101965	\N
11774	1-11774	97	13	5448	28	2343	2343	3	2012-08-02 11:15:51.916615	\N
2254	1-2254	52	9	5604	97	494	494	0	2012-08-02 04:57:20.961995	\N
3247	1-3247	73	44	5819	93	135	135	0	2012-08-02 05:04:19.243651	\N
403	1-403	17	64	6477	19	314	314	0	2012-08-02 04:44:03.89155	\N
10566	1-10566	98	1	9633	43	1997	1997	0	2012-08-02 10:35:34.179659	\N
1066	1-1066	16	35	5881	25	2240	2240	0	2012-08-02 04:47:51.204313	\N
2673	1-2673	56	1	8122	59	737	737	0	2012-08-02 05:00:22.702106	\N
8721	1-8721	98	39	6611	14	9089	2494	0	2012-08-02 05:41:14.183196	2012-08-02 11:28:07.419662
1987	1-1987	65	98	4643	27	113	113	0	2012-08-02 04:55:25.422375	\N
3245	1-3245	62	6	9378	42	5096	2555	7	2012-08-02 05:04:18.413391	2012-08-02 11:21:53.025745
1060	1-1060	51	36	7857	26	732	452	0	2012-08-02 04:47:46.752427	2012-08-02 04:51:36.726515
6374	1-6374	8	49	6329	60	2684	2684	9	2012-08-02 05:25:18.397727	\N
1859	1-1859	79	91	7727	20	213	213	0	2012-08-02 04:54:28.5012	\N
12539	1-12539	19	99	4717	83	5945	504	4	2012-08-02 11:20:48.40243	2012-08-02 11:21:12.327546
4609	1-4609	5	48	9822	41	2069	675	0	2012-08-02 05:13:25.658872	2012-08-02 11:22:39.388621
13937	1-13937	23	94	6401	16	9962	9962	4	2012-08-02 11:29:59.973878	\N
3559	1-3559	5	26	9767	33	1962	1182	0	2012-08-02 05:06:25.417364	2012-08-02 10:35:35.715244
6375	1-6375	24	20	8011	64	325	325	0	2012-08-02 05:25:18.795433	\N
1497	1-1497	23	23	8428	10	945	945	0	2012-08-02 04:51:37.79807	\N
929	1-929	88	38	9573	10	1109	1109	2	2012-08-02 04:46:33.192792	\N
6675	1-6675	6	37	4001	93	97	97	2	2012-08-02 05:27:20.639861	\N
4911	1-4911	48	74	2446	88	1537	854	0	2012-08-02 05:15:27.427242	2012-08-02 05:27:44.55662
930	1-930	15	35	9409	63	493	493	0	2012-08-02 04:46:33.336203	\N
8724	1-8724	22	32	5340	59	2049	736	0	2012-08-02 05:41:15.398496	2012-08-02 11:26:32.564429
560	1-560	16	49	9476	93	1393	1393	0	2012-08-02 04:44:26.194194	\N
4607	1-4607	29	9	4625	25	2193	2122	8	2012-08-02 05:13:24.797464	2012-08-02 05:29:54.031862
9898	1-9898	10	38	8900	65	1511	1511	0	2012-08-02 05:49:14.018816	\N
13940	1-13940	8	59	7058	88	1102	1102	0	2012-08-02 11:30:01.133974	\N
12259	1-12259	1	81	7676	2	2012	2012	0	2012-08-02 11:18:59.762829	\N
12253	1-12253	11	19	4318	95	5817	1835	3	2012-08-02 11:18:57.486811	2012-08-02 11:23:17.376362
13943	1-13943	23	27	5879	59	765	765	0	2012-08-02 11:30:02.327152	\N
1863	1-1863	98	27	8101	100	282	282	6	2012-08-02 04:54:30.323701	\N
4315	1-4315	47	42	4277	85	1009	1009	2	2012-08-02 05:11:27.30758	\N
1857	1-1857	78	64	8528	92	2087	1433	2	2012-08-02 04:54:27.584069	2012-08-02 05:08:21.292559
7265	1-7265	4	85	5051	29	897	255	3	2012-08-02 05:31:19.511353	2012-08-02 10:38:28.82744
13922	1-13922	42	28	4138	42	9114	5284	6	2012-08-02 11:29:54.029273	2012-08-02 11:29:54.029273
661	1-661	10	61	3316	49	591	239	0	2012-08-02 04:44:51.18718	2012-08-02 04:46:06.885666
1095	1-1095	16	3	9293	15	577	526	0	2012-08-02 04:48:04.441389	2012-08-02 04:48:43.598762
11767	1-11767	32	70	4358	15	3762	3762	0	2012-08-02 11:15:49.187315	\N
4325	1-4325	89	6	8376	50	3222	123	0	2012-08-02 05:11:31.38412	2012-08-02 11:16:08.858103
1868	1-1868	89	65	7554	70	2914	1593	0	2012-08-02 04:54:32.577468	2012-08-02 05:02:54.243824
2414	1-2414	66	51	8514	72	1056	235	0	2012-08-02 04:58:30.69379	2012-08-02 05:11:09.596145
12540	1-12540	4	6	8392	86	7691	6331	8	2012-08-02 11:20:48.789151	2012-08-02 11:30:14.435146
2106	1-2106	58	72	8776	83	620	620	4	2012-08-02 04:56:16.657735	\N
945	1-945	80	70	8509	53	2391	1582	0	2012-08-02 04:46:40.384589	2012-08-02 10:21:01.968377
4610	1-4610	91	21	7778	80	3617	658	0	2012-08-02 05:13:26.078643	2012-08-02 11:18:13.355285
1091	1-1091	82	30	9524	16	4588	2831	0	2012-08-02 04:48:02.85051	2012-08-02 11:20:34.782688
7559	1-7559	51	89	4380	7	1710	1100	5	2012-08-02 05:33:18.466253	2012-08-02 05:34:24.519791
5536	1-5536	69	32	3809	49	1040	434	8	2012-08-02 05:19:35.589178	2012-08-02 05:40:18.085757
6377	1-6377	48	27	7500	72	323	323	2	2012-08-02 05:25:19.690341	\N
1089	1-1089	99	96	9966	2	4540	938	0	2012-08-02 04:48:02.342262	2012-08-02 11:30:16.762619
4322	1-4322	53	97	5745	14	840	840	0	2012-08-02 05:11:30.147096	\N
7842	1-7842	60	14	7518	44	1780	573	0	2012-08-02 05:35:16.862116	2012-08-02 05:37:43.295399
4918	1-4918	38	16	7850	42	61	61	0	2012-08-02 05:15:30.167131	\N
5196	1-5196	52	94	9272	96	1152	1152	11	2012-08-02 05:17:21.482249	\N
9900	1-9900	92	68	4004	83	1929	96	23	2012-08-02 05:49:14.814915	2012-08-02 11:18:00.604149
9023	1-9023	57	99	7004	36	592	592	3	2012-08-02 05:43:17.144666	\N
5801	1-5801	7	98	9876	92	2274	2274	8	2012-08-02 05:21:22.892042	\N
11310	1-11310	24	39	7768	13	4829	546	0	2012-08-02 11:12:51.592555	2012-08-02 11:19:43.441353
1384	1-1384	6	84	7544	68	2	2	5	2012-08-02 04:50:38.040699	\N
13460	1-13460	12	58	2559	88	1513	1513	4	2012-08-02 11:26:50.775515	\N
3570	1-3570	99	79	3789	5	535	403	0	2012-08-02 05:06:29.848038	2012-08-02 05:28:24.646396
11771	1-11771	52	97	6785	93	608	608	2	2012-08-02 11:15:49.961519	\N
12246	1-12246	46	61	7886	34	197	197	3	2012-08-02 11:18:54.76923	\N
8425	1-8425	2	45	5536	33	456	456	0	2012-08-02 05:39:15.154815	\N
606	1-606	25	49	6289	76	506	506	0	2012-08-02 04:44:37.767471	\N
9908	1-9908	37	44	9662	2	5978	4858	0	2012-08-02 05:49:18.04161	2012-08-02 10:32:07.571197
1994	1-1994	79	76	6865	45	50	50	0	2012-08-02 04:55:28.540189	\N
11775	1-11775	71	79	4838	48	1232	1232	7	2012-08-02 11:15:51.541038	\N
12543	1-12543	3	95	8924	79	4229	3624	7	2012-08-02 11:20:49.93817	2012-08-02 11:30:19.190513
10570	1-10570	39	1	8770	73	1269	1269	0	2012-08-02 10:35:34.467955	\N
3253	1-3253	89	14	7336	54	1047	862	0	2012-08-02 05:04:21.67473	2012-08-02 05:09:28.225583
7838	1-7838	20	49	9367	48	4721	1867	12	2012-08-02 05:35:15.149828	2012-08-02 11:30:33.358312
7560	1-7560	80	34	7501	9	728	728	0	2012-08-02 05:33:18.863803	\N
13010	1-13010	78	33	8057	10	3098	3098	0	2012-08-02 11:23:53.156921	\N
2683	1-2683	60	33	4182	41	1124	588	0	2012-08-02 05:00:27.044505	2012-08-02 11:20:59.742471
13463	1-13463	58	97	7243	93	6598	6598	2	2012-08-02 11:26:51.948313	\N
3716	1-3716	96	57	8917	94	236	236	0	2012-08-02 05:07:28.553652	\N
2109	1-2109	70	12	8659	2	2039	1528	0	2012-08-02 04:56:17.99489	2012-08-02 05:06:57.308487
10902	1-10902	74	32	3178	2	147	147	4	2012-08-02 10:38:43.724572	\N
1871	1-1871	21	98	7731	86	190	190	0	2012-08-02 04:54:33.980835	\N
6983	1-6983	55	73	5590	88	1659	1659	0	2012-08-02 05:29:25.07141	\N
13466	1-13466	85	43	8711	88	4581	4581	0	2012-08-02 11:26:53.285544	\N
10571	1-10571	2	56	8857	28	313	313	3	2012-08-02 10:35:34.534308	\N
13469	1-13469	41	36	9421	2	900	900	3	2012-08-02 11:26:54.422946	\N
13021	1-13021	49	97	2626	80	3665	3407	0	2012-08-02 11:23:57.035059	2012-08-02 11:27:06.323253
10905	1-10905	60	37	4803	50	1178	1178	7	2012-08-02 10:38:44.895671	\N
3717	1-3717	13	99	6033	44	2891	2891	0	2012-08-02 05:07:28.951667	\N
13015	1-13015	35	22	5853	37	3794	3794	0	2012-08-02 11:23:55.101427	\N
5807	1-5807	79	32	3377	2	389	389	0	2012-08-02 05:21:25.278357	\N
1386	1-1386	90	35	8320	54	3589	1741	0	2012-08-02 04:50:39.034566	2012-08-02 10:21:24.756573
12261	1-12261	94	54	7774	12	385	385	4	2012-08-02 11:19:00.558496	\N
5539	1-5539	99	36	3815	85	962	962	10	2012-08-02 05:19:36.782424	\N
13005	1-13005	37	79	7456	57	6393	4527	5	2012-08-02 11:23:50.803732	2012-08-02 11:25:17.553175
9907	1-9907	66	11	2981	85	964	964	6	2012-08-02 05:49:17.632985	\N
3571	1-3571	6	78	9203	2	1017	1017	7	2012-08-02 05:06:30.245389	\N
5540	1-5540	70	78	8631	76	427	427	0	2012-08-02 05:19:37.169004	\N
7270	1-7270	31	39	6046	2	3262	3262	0	2012-08-02 05:31:21.489263	\N
1867	1-1867	98	25	7220	6	1001	653	8	2012-08-02 04:54:32.146614	2012-08-02 11:26:22.651417
2117	1-2117	81	20	6437	95	31	31	0	2012-08-02 04:56:21.529444	\N
9905	1-9905	60	31	3628	87	3279	2430	11	2012-08-02 05:49:16.826358	2012-08-02 10:16:57.420724
641	1-641	85	49	7369	97	397	397	0	2012-08-02 04:44:46.023372	\N
1511	1-1511	74	64	8330	51	3262	915	0	2012-08-02 04:51:44.338138	2012-08-02 10:21:25.563768
1400	1-1400	73	74	4620	67	93	93	0	2012-08-02 04:50:44.811988	\N
662	1-662	19	81	8117	78	63	63	0	2012-08-02 04:44:51.231414	\N
652	1-652	100	76	7373	72	383	383	0	2012-08-02 04:44:49.548499	\N
665	1-665	2	8	9051	88	238	238	0	2012-08-02 04:44:51.662173	\N
1724	1-1724	97	61	9465	32	380	380	0	2012-08-02 04:53:27.74407	\N
2544	1-2544	42	6	6500	69	2167	1375	6	2012-08-02 04:59:27.608973	2012-08-02 10:37:55.502015
6103	1-6103	70	52	8891	67	1406	349	0	2012-08-02 05:23:26.580929	2012-08-02 11:21:48.407505
11776	1-11776	3	8	5340	72	9374	1782	10	2012-08-02 11:15:52.358938	2012-08-02 11:27:44.513646
3269	1-3269	8	39	3469	3	381	29	2	2012-08-02 05:04:28.082239	2012-08-02 05:43:42.135115
9027	1-9027	25	79	9285	35	2320	1654	2	2012-08-02 05:43:18.823996	2012-08-02 11:29:44.549699
2964	1-2964	76	22	9560	56	1361	77	7	2012-08-02 05:02:24.654977	2012-08-02 11:14:55.42302
5210	1-5210	11	26	6682	14	5866	5154	0	2012-08-02 05:17:26.973944	2012-08-02 11:21:03.222696
674	1-674	79	90	8736	25	384	384	0	2012-08-02 04:44:53.353349	\N
13935	1-13935	74	35	5168	28	4960	4960	0	2012-08-02 11:29:59.200555	\N
1999	1-1999	78	33	7688	83	2455	1262	6	2012-08-02 04:55:30.749703	2012-08-02 05:17:12.85116
6384	1-6384	93	21	4956	46	586	586	3	2012-08-02 05:25:22.485428	\N
6390	1-6390	60	65	5744	79	1405	774	9	2012-08-02 05:25:24.871954	2012-08-02 05:42:22.694059
12541	1-12541	74	57	4606	29	817	817	0	2012-08-02 11:20:49.175688	\N
9024	1-9024	34	78	8173	69	6471	1737	0	2012-08-02 05:43:17.553539	2012-08-02 11:23:32.018196
2267	1-2267	15	20	9943	38	1071	856	0	2012-08-02 04:57:26.585465	2012-08-02 04:57:57.138768
1151	1-1151	25	13	5225	2	857	212	0	2012-08-02 04:48:34.980683	2012-08-02 04:53:58.725
6385	1-6385	99	46	8442	15	177	177	0	2012-08-02 05:25:22.894087	\N
561	1-561	45	31	7772	15	654	125	0	2012-08-02 04:44:26.24909	2012-08-02 04:49:54.461358
11316	1-11316	44	30	5368	62	60	60	0	2012-08-02 11:12:53.923739	\N
1158	1-1158	9	69	7413	37	2160	1029	0	2012-08-02 04:48:37.952637	2012-08-02 05:05:31.783059
1174	1-1174	94	73	8684	83	422	422	0	2012-08-02 04:48:45.035371	\N
8138	1-8138	48	78	9472	56	2100	1395	10	2012-08-02 05:37:18.776754	2012-08-02 11:15:10.604364
2821	1-2821	49	5	8119	44	1231	735	0	2012-08-02 05:01:25.19726	2012-08-02 10:22:19.140153
6102	1-6102	32	83	8389	37	43	43	0	2012-08-02 05:23:26.183674	\N
9913	1-9913	93	38	8083	36	457	457	4	2012-08-02 05:49:20.055946	\N
6986	1-6986	27	38	7247	31	1272	846	5	2012-08-02 05:29:26.231526	2012-08-02 11:24:26.844739
1729	1-1729	88	61	7610	34	1348	1228	0	2012-08-02 04:53:30.329706	2012-08-02 05:03:09.250277
688	1-688	42	86	4127	10	122	122	0	2012-08-02 04:44:56.359894	\N
4927	1-4927	98	19	4654	5	3433	3433	0	2012-08-02 05:15:33.68265	\N
6388	1-6388	73	46	6715	54	1483	1205	0	2012-08-02 05:25:24.076582	2012-08-02 11:22:30.105975
11318	1-11318	6	62	2110	65	429	429	0	2012-08-02 11:12:55.072879	\N
4332	1-4332	16	87	9601	83	2708	2277	0	2012-08-02 05:11:34.247618	2012-08-02 05:34:55.944778
3379	1-3379	51	85	6934	81	884	884	0	2012-08-02 05:05:12.879028	\N
9313	1-9313	98	75	8751	70	1196	1196	11	2012-08-02 05:45:15.486817	\N
13488	1-13488	13	79	1635	54	5032	2243	0	2012-08-02 11:27:01.95929	2012-08-02 11:30:19.190513
1876	1-1876	1	19	7391	81	53	53	0	2012-08-02 04:54:36.678444	\N
7847	1-7847	45	90	1467	77	41	41	7	2012-08-02 05:35:18.961223	\N
4934	1-4934	6	67	5435	57	1700	978	7	2012-08-02 05:15:36.422576	2012-08-02 11:11:16.450807
9914	1-9914	100	92	7963	83	704	704	3	2012-08-02 05:49:20.453447	\N
13467	1-13467	81	35	8812	63	5083	1941	5	2012-08-02 11:26:53.67169	2012-08-02 11:29:58.029416
13473	1-13473	33	74	7345	62	6919	6915	2	2012-08-02 11:26:56.080895	2012-08-02 11:28:35.596534
11324	1-11324	91	70	8481	52	4289	4289	7	2012-08-02 11:12:56.995342	\N
4001	1-4001	57	19	8538	85	8803	1701	0	2012-08-02 05:09:22.071995	2012-08-02 05:19:49.158616
4003	1-4003	46	68	6773	24	1557	724	0	2012-08-02 05:09:22.855897	2012-08-02 10:39:19.198296
13476	1-13476	47	31	8226	32	807	807	11	2012-08-02 11:26:57.284685	\N
13482	1-13482	29	75	3059	50	3659	2072	0	2012-08-02 11:26:59.627443	2012-08-02 11:27:59.982053
11327	1-11327	75	46	6692	52	2033	2033	7	2012-08-02 11:12:58.13309	\N
2962	1-2962	100	36	9271	81	535	535	0	2012-08-02 05:02:23.859708	\N
1877	1-1877	9	25	2866	51	91	91	3	2012-08-02 04:54:37.150945	\N
1257	1-1257	49	12	4033	98	38	38	0	2012-08-02 04:49:31.505879	\N
4002	1-4002	95	52	2008	34	122	122	0	2012-08-02 05:09:22.469492	\N
13938	1-13938	51	56	2401	65	1788	956	0	2012-08-02 11:30:00.36048	2012-08-02 11:30:00.36048
9030	1-9030	1	36	7518	27	2487	2487	0	2012-08-02 05:43:20.016981	\N
13479	1-13479	30	53	9851	99	7991	7414	4	2012-08-02 11:26:58.444928	2012-08-02 11:26:58.444928
11783	1-11783	78	93	9459	51	1004	1004	0	2012-08-02 11:15:54.690227	\N
13485	1-13485	96	3	5220	1	1345	1345	6	2012-08-02 11:27:00.810262	\N
2695	1-2695	12	18	7151	53	545	545	0	2012-08-02 05:00:32.048963	\N
3576	1-3576	22	25	9238	53	2336	1197	0	2012-08-02 05:06:32.401991	2012-08-02 11:27:01.561032
8432	1-8432	51	30	9839	82	4012	2404	5	2012-08-02 05:39:17.939805	2012-08-02 05:39:17.939805
11337	1-11337	17	98	6318	49	1225	1225	2	2012-08-02 11:13:02.000188	\N
1411	1-1411	32	4	6582	55	548	548	0	2012-08-02 04:50:50.987359	\N
8433	1-8433	27	75	8632	52	956	956	5	2012-08-02 05:39:18.337451	\N
2963	1-2963	18	45	9997	34	53	53	0	2012-08-02 05:02:24.25699	\N
8427	1-8427	73	55	8748	40	9358	6339	7	2012-08-02 05:39:15.939114	2012-08-02 11:30:14.048472
5543	1-5543	66	24	9534	11	2926	991	0	2012-08-02 05:19:38.351084	2012-08-02 11:29:35.795893
13491	1-13491	18	29	7611	42	3297	3297	0	2012-08-02 11:27:03.174417	\N
1519	1-1519	4	98	7878	69	128	128	0	2012-08-02 04:51:48.780748	\N
13944	1-13944	78	30	4163	97	9829	7518	11	2012-08-02 11:30:02.736062	2012-08-02 11:30:02.736062
13494	1-13494	57	3	7691	85	3750	3750	0	2012-08-02 11:27:04.356466	\N
11795	1-11795	64	59	9873	76	2446	2446	0	2012-08-02 11:15:59.829482	\N
974	1-974	54	67	4222	73	1045	885	0	2012-08-02 04:46:58.603141	2012-08-02 11:28:14.457753
721	1-721	23	71	6900	77	457	457	0	2012-08-02 04:45:04.714097	\N
11801	1-11801	64	50	8060	82	869	869	3	2012-08-02 11:16:02.184533	\N
12544	1-12544	72	45	3192	72	1844	43	0	2012-08-02 11:20:50.313736	2012-08-02 11:30:23.150851
10255	1-10255	44	82	8492	40	4588	2507	0	2012-08-02 10:25:19.75132	2012-08-02 11:16:06.857917
2136	1-2136	1	61	7267	56	1879	808	0	2012-08-02 04:56:29.828003	2012-08-02 11:15:39.333955
9032	1-9032	87	98	6312	66	1928	1926	0	2012-08-02 05:43:20.845853	2012-08-02 11:28:06.634738
9916	1-9916	83	19	9805	23	375	375	11	2012-08-02 05:49:21.271396	\N
1189	1-1189	18	64	3618	81	95	95	0	2012-08-02 04:48:52.710896	\N
5813	1-5813	77	67	8136	16	963	963	0	2012-08-02 05:21:27.754116	\N
12263	1-12263	73	94	4014	96	2895	2895	15	2012-08-02 11:19:01.333945	\N
11788	1-11788	30	95	9772	69	7467	4291	12	2012-08-02 11:15:57.033552	2012-08-02 11:29:42.160175
13468	1-13468	52	87	8816	34	234	234	5	2012-08-02 11:26:54.047377	\N
4005	1-4005	64	19	8545	91	1304	1304	0	2012-08-02 05:09:23.640435	\N
3385	1-3385	57	99	7422	9	1274	709	0	2012-08-02 05:05:15.320623	2012-08-02 05:15:15.296323
13927	1-13927	49	18	8309	74	6885	2766	7	2012-08-02 11:29:55.996314	2012-08-02 11:29:55.996314
6997	1-6997	3	54	7703	45	1831	1235	5	2012-08-02 05:29:30.551949	2012-08-02 05:31:11.755453
12281	1-12281	92	54	5356	85	4589	3970	0	2012-08-02 11:19:08.205874	2012-08-02 11:22:36.670154
13930	1-13930	100	85	7345	39	180	180	22	2012-08-02 11:29:57.167606	\N
2702	1-2702	64	96	3140	79	82	82	0	2012-08-02 05:00:34.998736	\N
13034	1-13034	71	21	9106	46	2923	2405	6	2012-08-02 11:24:02.07335	2012-08-02 11:28:55.123393
6993	1-6993	7	94	6764	5	5131	913	3	2012-08-02 05:29:28.960677	2012-08-02 11:22:19.974844
9606	1-9606	91	4	8846	82	5467	3477	10	2012-08-02 05:47:14.66935	2012-08-02 11:20:59.742471
3273	1-3273	4	57	9159	77	79	79	3	2012-08-02 05:04:29.728714	\N
1278	1-1278	95	41	8655	84	441	397	0	2012-08-02 04:49:42.111972	2012-08-02 05:03:09.250277
13933	1-13933	91	50	3915	72	1526	1526	7	2012-08-02 11:29:58.41599	\N
2129	1-2129	53	16	6660	75	336	116	0	2012-08-02 04:56:26.777255	2012-08-02 05:48:28.758694
2003	1-2003	31	58	8850	97	231	231	0	2012-08-02 04:55:32.517126	\N
11780	1-11780	1	42	7748	11	3627	1881	2	2012-08-02 11:15:53.894555	2012-08-02 11:29:35.795893
11315	1-11315	59	44	6299	45	4801	2510	0	2012-08-02 11:12:53.547983	2012-08-02 11:20:02.402723
4938	1-4938	5	8	9737	44	3188	3188	0	2012-08-02 05:15:37.969332	\N
7282	1-7282	93	46	8219	13	2370	2370	0	2012-08-02 05:31:26.186569	\N
1265	1-1265	26	57	8096	60	620	620	0	2012-08-02 04:49:36.200888	\N
6688	1-6688	48	58	5567	8	79	79	2	2012-08-02 05:27:25.781973	\N
8143	1-8143	59	73	8477	21	3546	3546	7	2012-08-02 05:37:20.788068	\N
12303	1-12303	69	75	7116	30	5157	4086	0	2012-08-02 11:19:16.337632	2012-08-02 11:23:25.731152
2968	1-2968	53	38	5134	4	706	706	3	2012-08-02 05:02:26.289686	\N
9607	1-9607	50	3	1905	11	83	83	9	2012-08-02 05:47:15.067024	\N
13939	1-13939	93	63	9067	21	2210	2210	0	2012-08-02 11:30:00.758394	\N
11325	1-11325	57	94	8905	83	895	895	20	2012-08-02 11:12:57.381871	\N
12278	1-12278	65	68	8014	9	3144	3144	0	2012-08-02 11:19:07.056685	\N
1296	1-1296	29	88	7955	31	719	673	0	2012-08-02 04:49:52.727295	2012-08-02 04:50:42.447879
13031	1-13031	77	26	1536	20	530	530	0	2012-08-02 11:24:00.934991	\N
7854	1-7854	18	34	6461	89	725	725	9	2012-08-02 05:35:21.910984	\N
1268	1-1268	91	82	8976	48	977	977	0	2012-08-02 04:49:37.206328	\N
13942	1-13942	86	64	5034	5	3878	3878	0	2012-08-02 11:30:01.940468	\N
7572	1-7572	12	81	7246	38	758	758	0	2012-08-02 05:33:23.748981	\N
6690	1-6690	57	9	4712	13	2390	2390	3	2012-08-02 05:27:26.577006	\N
12284	1-12284	51	6	7824	16	557	557	14	2012-08-02 11:19:09.365844	\N
3100	1-3100	87	32	6861	37	140	140	0	2012-08-02 05:03:19.835419	\N
11342	1-11342	9	79	9749	7	2432	2432	13	2012-08-02 11:13:03.955396	\N
13945	1-13945	99	88	9814	32	8013	8013	0	2012-08-02 11:30:03.122807	\N
2132	1-2132	58	71	2233	2	347	347	0	2012-08-02 04:56:28.091884	\N
12287	1-12287	16	66	2565	16	1786	1786	5	2012-08-02 11:19:10.492994	\N
2978	1-2978	100	56	7764	73	3578	1016	0	2012-08-02 05:02:30.35535	2012-08-02 11:30:03.509658
2133	1-2133	54	33	5132	81	194	194	3	2012-08-02 04:56:28.522533	\N
9922	1-9922	65	33	5118	61	2274	2274	7	2012-08-02 05:49:23.724042	\N
2423	1-2423	17	19	3970	4	914	914	0	2012-08-02 04:58:34.627408	\N
1083	1-1083	12	19	9409	60	4228	2059	0	2012-08-02 04:47:58.839796	2012-08-02 05:12:16.842046
13040	1-13040	22	44	9208	55	58	58	18	2012-08-02 11:24:04.404066	\N
4350	1-4350	64	67	8746	49	1038	1038	7	2012-08-02 05:11:41.683529	\N
13025	1-13025	45	80	8555	45	6276	4930	14	2012-08-02 11:23:58.60374	2012-08-02 11:25:41.957032
13948	1-13948	72	20	6046	15	2344	2344	6	2012-08-02 11:30:04.293571	\N
10264	1-10264	42	4	4959	70	953	953	0	2012-08-02 10:25:23.289351	\N
12293	1-12293	54	94	7190	33	1406	1406	7	2012-08-02 11:19:12.86827	\N
3731	1-3731	88	23	3907	45	591	591	5	2012-08-02 05:07:34.465035	\N
9917	1-9917	80	80	5166	73	964	755	0	2012-08-02 05:49:21.669183	2012-08-02 11:13:51.886711
2135	1-2135	12	39	6513	34	436	436	0	2012-08-02 04:56:29.408196	\N
1290	1-1290	91	92	5078	41	160	160	0	2012-08-02 04:49:48.407999	\N
6991	1-6991	8	19	9534	44	8352	846	8	2012-08-02 05:29:28.187078	2012-08-02 11:20:14.819348
7576	1-7576	70	21	9845	52	1642	1642	0	2012-08-02 05:33:25.339792	\N
8446	1-8446	54	58	8398	97	1291	1291	0	2012-08-02 05:39:23.532451	\N
2554	1-2554	94	98	8028	99	589	589	0	2012-08-02 04:59:32.072845	\N
1532	1-1532	96	96	5310	98	841	841	0	2012-08-02 04:51:54.790694	\N
8447	1-8447	61	70	7396	21	1003	1003	3	2012-08-02 05:39:23.919374	\N
13050	1-13050	69	73	7822	17	1925	1925	6	2012-08-02 11:24:08.260049	\N
6401	1-6401	28	51	8600	24	757	593	0	2012-08-02 05:25:29.337649	2012-08-02 05:26:14.862037
5824	1-5824	69	10	3367	49	70	70	11	2012-08-02 05:21:32.197106	\N
3735	1-3735	53	97	9312	52	993	993	0	2012-08-02 05:07:36.033448	\N
4947	1-4947	21	55	5678	37	32	32	7	2012-08-02 05:15:41.49377	\N
2557	1-2557	21	12	3962	80	309	309	0	2012-08-02 04:59:33.332321	\N
3283	1-3283	70	24	7053	30	47	47	0	2012-08-02 05:04:33.761501	\N
3106	1-3106	69	62	8844	5	4464	712	0	2012-08-02 05:03:22.288423	2012-08-02 11:16:43.897895
11793	1-11793	79	32	9930	25	3163	2266	8	2012-08-02 11:15:59.000071	2012-08-02 11:26:21.446797
7001	1-7001	45	34	8876	86	4145	2044	0	2012-08-02 05:29:32.134027	2012-08-02 10:27:44.981802
1548	1-1548	30	94	6948	52	845	845	0	2012-08-02 04:52:03.529284	\N
1620	1-1620	91	30	9875	15	1144	1144	0	2012-08-02 04:52:39.77982	\N
13020	1-13020	11	63	2894	98	6668	5331	5	2012-08-02 11:23:56.637412	2012-08-02 11:26:47.220183
9926	1-9926	53	32	6158	91	2219	1614	0	2012-08-02 05:49:25.359285	2012-08-02 11:23:44.161638
13029	1-13029	58	21	8472	48	9012	739	11	2012-08-02 11:24:00.161525	2012-08-02 11:29:45.743019
2991	1-2991	22	55	8289	95	3524	2112	0	2012-08-02 05:02:35.660946	2012-08-02 05:21:27.333921
3739	1-3739	15	24	7865	37	4973	3590	0	2012-08-02 05:07:37.569083	2012-08-02 05:19:55.11399
12264	1-12264	7	65	8985	18	1422	1422	4	2012-08-02 11:19:02.107304	\N
10900	1-10900	1	79	5191	26	1460	565	0	2012-08-02 10:38:43.337584	2012-08-02 11:29:58.029416
10921	1-10921	72	88	7156	55	3765	1230	0	2012-08-02 10:38:51.181209	2012-08-02 11:24:42.909484
7000	1-7000	60	35	5422	10	1812	1812	0	2012-08-02 05:29:31.74733	\N
1746	1-1746	27	91	5111	49	413	413	0	2012-08-02 04:53:38.206563	\N
7289	1-7289	11	53	8874	2	5770	4552	7	2012-08-02 05:31:28.926775	2012-08-02 05:39:07.484099
9612	1-9612	75	96	9123	19	2706	515	5	2012-08-02 05:47:17.133007	2012-08-02 11:20:55.056277
3736	1-3736	37	42	5765	32	320	320	6	2012-08-02 05:07:36.420037	\N
13023	1-13023	6	73	8073	44	9163	6611	0	2012-08-02 11:23:57.83056	2012-08-02 11:27:41.683599
8449	1-8449	54	27	9313	76	3481	3066	4	2012-08-02 05:39:24.714792	2012-08-02 10:25:24.06309
12573	1-12573	91	2	8565	96	2843	804	4	2012-08-02 11:21:01.278361	2012-08-02 11:26:55.649625
8429	1-8429	55	31	7045	54	3489	463	13	2012-08-02 05:39:16.7464	2012-08-02 05:39:54.33926
12270	1-12270	75	61	9805	8	2352	2277	5	2012-08-02 11:19:04.007617	2012-08-02 11:19:04.007617
12560	1-12560	54	75	9147	41	8038	6089	8	2012-08-02 11:20:56.648337	2012-08-02 11:29:21.974274
5551	1-5551	70	1	8617	31	3241	2022	0	2012-08-02 05:19:41.533297	2012-08-02 11:27:46.491424
10267	1-10267	99	67	7352	45	7377	734	0	2012-08-02 10:25:24.471724	2012-08-02 11:27:16.943419
9316	1-9316	57	66	6572	86	1615	1615	10	2012-08-02 05:45:16.713037	\N
13956	1-13956	4	47	7798	23	2000	2000	4	2012-08-02 11:30:07.797946	\N
3585	1-3585	89	23	7239	63	615	615	0	2012-08-02 05:06:36.026438	\N
11332	1-11332	16	62	9964	70	4145	4145	2	2012-08-02 11:13:00.044348	\N
13047	1-13047	19	40	9095	7	7871	1276	8	2012-08-02 11:24:06.701915	2012-08-02 11:30:04.669386
9044	1-9044	30	47	9355	53	2433	2433	3	2012-08-02 05:43:25.795146	\N
4359	1-4359	83	22	7177	28	236	236	4	2012-08-02 05:11:45.362626	\N
10581	1-10581	23	68	8788	2	9511	9511	6	2012-08-02 10:35:35.336127	\N
4953	1-4953	77	82	9533	54	4308	4308	0	2012-08-02 05:15:43.847064	\N
7864	1-7864	85	74	8949	72	5087	601	2	2012-08-02 05:35:26.131353	2012-08-02 11:15:58.227009
9043	1-9043	42	8	4236	47	541	541	0	2012-08-02 05:43:25.386312	\N
2982	1-2982	44	65	9062	6	830	797	0	2012-08-02 05:02:31.97947	2012-08-02 05:28:25.055625
13014	1-13014	71	59	4987	22	5130	184	3	2012-08-02 11:23:54.316654	2012-08-02 11:28:22.314052
12569	1-12569	57	64	6664	55	3292	3292	6	2012-08-02 11:21:00.140348	\N
11335	1-11335	65	72	4899	71	1589	1589	2	2012-08-02 11:13:01.193755	\N
4364	1-4364	33	41	5372	53	2280	488	7	2012-08-02 05:11:47.417183	2012-08-02 11:29:21.974274
7296	1-7296	79	13	9310	95	5685	4650	0	2012-08-02 05:31:31.678024	2012-08-02 11:20:58.97991
1644	1-1644	63	89	6091	53	2470	825	0	2012-08-02 04:52:50.761548	2012-08-02 10:39:23.148279
1755	1-1755	96	66	5076	15	438	438	0	2012-08-02 04:53:41.874425	\N
7863	1-7863	66	62	5750	7	211	211	0	2012-08-02 05:35:25.722761	\N
4019	1-4019	54	86	6641	69	1608	1278	3	2012-08-02 05:09:29.430355	2012-08-02 10:21:20.423155
3740	1-3740	25	37	7932	30	44	44	0	2012-08-02 05:07:37.966763	\N
7861	1-7861	46	80	4557	98	843	690	0	2012-08-02 05:35:24.872174	2012-08-02 11:16:20.393545
4365	1-4365	3	53	5065	56	325	325	0	2012-08-02 05:11:47.814917	\N
9049	1-9049	25	70	8834	54	1074	1074	0	2012-08-02 05:43:27.827789	\N
10925	1-10925	41	98	4865	86	856	856	7	2012-08-02 10:38:52.36845	\N
2025	1-2025	35	30	5506	63	390	390	0	2012-08-02 04:55:41.742719	\N
1897	1-1897	7	32	8772	53	1739	1568	0	2012-08-02 04:54:46.398134	2012-08-02 05:06:53.671591
13959	1-13959	2	90	2426	76	1208	1208	8	2012-08-02 11:30:08.582518	\N
10928	1-10928	75	26	6744	64	945	945	14	2012-08-02 10:38:53.998446	\N
11799	1-11799	7	39	1663	44	1650	1650	12	2012-08-02 11:16:01.411559	\N
13026	1-13026	53	50	4389	13	1797	1789	5	2012-08-02 11:23:58.990402	2012-08-02 11:26:00.681608
11317	1-11317	8	19	5440	6	3557	2845	0	2012-08-02 11:12:54.310367	2012-08-02 11:26:11.325063
5235	1-5235	66	32	6663	46	2567	2567	0	2012-08-02 05:17:36.818466	\N
4958	1-4958	84	100	9122	14	803	803	0	2012-08-02 05:15:45.846908	\N
12267	1-12267	65	13	2459	77	3201	602	0	2012-08-02 11:19:02.869734	2012-08-02 11:26:17.480661
3590	1-3590	77	81	9879	14	5756	5756	0	2012-08-02 05:06:38.026169	\N
10930	1-10930	87	28	8783	17	8195	2939	0	2012-08-02 10:38:55.184209	2012-08-02 11:26:36.46422
13962	1-13962	9	37	9066	6	6874	4977	9	2012-08-02 11:30:09.753486	2012-08-02 11:30:09.753486
2283	1-2283	20	75	2061	91	106	106	6	2012-08-02 04:57:33.647575	\N
13965	1-13965	90	90	5863	58	3396	3396	0	2012-08-02 11:30:10.913726	\N
13969	1-13969	65	86	2848	21	165	165	5	2012-08-02 11:30:12.084664	\N
4369	1-4369	79	64	6407	70	506	506	0	2012-08-02 05:11:49.439071	\N
1770	1-1770	39	83	7904	32	603	603	0	2012-08-02 04:53:48.669072	\N
13954	1-13954	9	9	4706	35	5350	5096	5	2012-08-02 11:30:07.024753	2012-08-02 11:30:28.283346
7298	1-7298	36	52	8773	81	392	392	0	2012-08-02 05:31:32.429541	\N
13474	1-13474	43	89	7212	4	6428	2770	9	2012-08-02 11:26:56.478227	2012-08-02 11:30:32.230146
2287	1-2287	53	6	1401	53	311	311	0	2012-08-02 04:57:35.348873	\N
1615	1-1615	42	64	7872	91	1238	839	0	2012-08-02 04:52:37.404476	2012-08-02 04:57:38.365076
1630	1-1630	34	68	4322	70	1249	572	0	2012-08-02 04:52:44.519811	2012-08-02 05:00:52.037763
1918	1-1918	90	72	5812	43	251	251	0	2012-08-02 04:54:55.303031	\N
10594	1-10594	1	65	4074	29	1858	1112	0	2012-08-02 10:35:36.303628	2012-08-02 11:29:59.576353
1916	1-1916	47	81	6192	3	606	475	0	2012-08-02 04:54:54.330963	2012-08-02 04:57:23.558207
2029	1-2029	51	39	8959	66	431	431	8	2012-08-02 04:55:43.510357	\N
4374	1-4374	52	86	7759	16	4189	80	0	2012-08-02 05:11:51.549623	2012-08-02 05:38:08.492792
4375	1-4375	66	34	5171	85	1797	394	0	2012-08-02 05:11:51.958085	2012-08-02 11:15:13.522864
7870	1-7870	67	61	9623	7	6852	6813	10	2012-08-02 05:35:28.672313	2012-08-02 10:27:51.268635
10916	1-10916	97	70	6386	41	1261	348	5	2012-08-02 10:38:49.215982	2012-08-02 11:20:40.961064
1920	1-1920	96	61	7773	47	1474	1151	0	2012-08-02 04:54:55.778043	2012-08-02 05:18:14.829328
4644	1-4644	12	39	6234	92	3073	2052	0	2012-08-02 05:13:40.056933	2012-08-02 05:19:14.671889
3298	1-3298	72	38	7873	54	3240	489	0	2012-08-02 05:04:39.771562	2012-08-02 05:13:57.585237
8744	1-8744	50	36	6913	76	1066	1066	3	2012-08-02 05:41:23.514259	\N
1922	1-1922	66	80	9989	74	1416	358	0	2012-08-02 04:54:56.705665	2012-08-02 05:02:33.570341
13027	1-13027	11	18	9618	31	750	750	0	2012-08-02 11:23:59.39947	\N
9325	1-9325	70	54	5700	72	3123	1015	2	2012-08-02 05:45:20.427552	2012-08-02 11:12:21.908586
5844	1-5844	47	68	8708	61	806	806	7	2012-08-02 05:21:40.207526	\N
9058	1-9058	45	97	7232	4	5178	1863	0	2012-08-02 05:43:31.517441	2012-08-02 11:28:16.811248
3605	1-3605	91	19	7230	54	9340	2290	0	2012-08-02 05:06:44.125472	2012-08-02 10:36:20.436271
1783	1-1783	24	58	8037	87	1629	79	0	2012-08-02 04:53:54.657025	2012-08-02 05:21:47.233426
7009	1-7009	63	43	5226	10	1816	733	0	2012-08-02 05:29:35.339041	2012-08-02 11:29:53.255963
13054	1-13054	89	93	7317	95	2858	1942	16	2012-08-02 11:24:09.817573	2012-08-02 11:27:52.435896
8457	1-8457	95	26	9918	55	6371	4438	0	2012-08-02 05:39:27.94174	2012-08-02 10:28:00.328852
7308	1-7308	57	34	8201	39	928	667	0	2012-08-02 05:31:36.362468	2012-08-02 05:32:07.576776
11344	1-11344	65	73	7031	48	7148	4124	0	2012-08-02 11:13:04.706838	2012-08-02 11:28:25.053385
12271	1-12271	42	96	9185	8	5503	3969	5	2012-08-02 11:19:04.372514	2012-08-02 11:28:09.375063
5553	1-5553	29	6	6633	79	235	235	7	2012-08-02 05:19:42.317719	\N
9319	1-9319	46	31	7750	70	1510	1510	0	2012-08-02 05:45:17.950544	\N
3294	1-3294	3	26	8888	38	3313	1648	0	2012-08-02 05:04:38.169217	2012-08-02 05:07:22.22319
2567	1-2567	74	11	2081	76	115	115	0	2012-08-02 04:59:37.608062	\N
10913	1-10913	44	71	7508	32	21	21	6	2012-08-02 10:38:48.010705	\N
7011	1-7011	12	50	6436	45	522	522	7	2012-08-02 05:29:36.156805	\N
12268	1-12268	5	36	9204	26	3897	3897	4	2012-08-02 11:19:03.267483	\N
8157	1-8157	11	59	4846	69	1777	1777	0	2012-08-02 05:37:26.61024	\N
13033	1-13033	25	36	6703	46	4244	4244	2	2012-08-02 11:24:01.697348	\N
13036	1-13036	70	48	7943	10	1447	1447	16	2012-08-02 11:24:02.835515	\N
7866	1-7866	35	24	4256	77	1078	1078	0	2012-08-02 05:35:26.982297	\N
3117	1-3117	45	96	9620	70	2630	2197	0	2012-08-02 05:03:26.762511	2012-08-02 05:14:41.638569
10270	1-10270	50	41	6929	81	133	133	5	2012-08-02 10:25:25.666743	\N
8456	1-8456	2	7	3762	62	159	159	0	2012-08-02 05:39:27.555617	\N
6122	1-6122	45	28	3076	87	276	276	0	2012-08-02 05:23:34.126922	\N
13953	1-13953	42	81	5610	18	3315	3315	7	2012-08-02 11:30:06.238441	\N
11800	1-11800	22	86	4320	80	330	330	2	2012-08-02 11:16:01.798069	\N
1777	1-1777	44	57	5013	53	948	948	0	2012-08-02 04:53:51.927277	\N
5838	1-5838	39	5	8689	76	852	771	3	2012-08-02 05:21:37.798839	2012-08-02 05:42:01.146484
13957	1-13957	89	26	5123	20	8369	8369	0	2012-08-02 11:30:07.411288	\N
3296	1-3296	52	86	5899	15	3096	3096	4	2012-08-02 05:04:38.964668	\N
7303	1-7303	23	74	4312	87	3214	3214	0	2012-08-02 05:31:34.385147	\N
2719	1-2719	65	65	9226	17	2515	2515	0	2012-08-02 05:00:42.27995	\N
5839	1-5839	67	62	8587	13	658	658	0	2012-08-02 05:21:38.196815	\N
11803	1-11803	24	14	9498	90	346	346	0	2012-08-02 11:16:02.935834	\N
13960	1-13960	57	48	6362	93	6937	6937	15	2012-08-02 11:30:08.958117	\N
1915	1-1915	60	19	5652	53	6899	4467	0	2012-08-02 04:54:53.827979	2012-08-02 11:21:42.45101
3007	1-3007	13	73	6067	16	2091	609	0	2012-08-02 05:02:42.158053	2012-08-02 05:46:38.006682
4963	1-4963	58	47	4267	57	1508	639	8	2012-08-02 05:15:47.769068	2012-08-02 11:20:01.662484
5841	1-5841	44	96	5795	81	649	649	0	2012-08-02 05:21:39.003127	\N
13045	1-13045	23	43	2795	38	36	36	11	2012-08-02 11:24:06.315512	\N
6124	1-6124	23	70	5384	88	866	866	10	2012-08-02 05:23:34.911704	\N
3299	1-3299	96	67	6263	52	437	437	0	2012-08-02 05:04:40.169306	\N
2569	1-2569	34	24	7273	42	4278	3147	0	2012-08-02 04:59:38.447673	2012-08-02 11:20:05.098618
2147	1-2147	70	100	8382	57	893	893	3	2012-08-02 04:56:34.269882	\N
12274	1-12274	77	37	9197	66	2517	1875	2	2012-08-02 11:19:05.510341	2012-08-02 11:25:38.40866
3113	1-3113	100	94	9556	14	3087	2332	0	2012-08-02 05:03:25.138449	2012-08-02 05:07:52.254268
13048	1-13048	99	20	5951	13	1155	1155	0	2012-08-02 11:24:07.497335	\N
11350	1-11350	72	55	8164	9	89	89	3	2012-08-02 11:13:07.086471	\N
13051	1-13051	2	77	6175	52	1290	1290	6	2012-08-02 11:24:08.657802	\N
11353	1-11353	33	26	8250	40	4401	4401	0	2012-08-02 11:13:08.24257	\N
8461	1-8461	26	26	6253	86	1196	1196	15	2012-08-02 05:39:29.54382	\N
2297	1-2297	43	78	4982	15	131	131	0	2012-08-02 04:57:39.67979	\N
6126	1-6126	31	61	8503	28	365	365	11	2012-08-02 05:23:35.706995	\N
11356	1-11356	32	8	9711	86	7879	7879	8	2012-08-02 11:13:09.403134	\N
4379	1-4379	91	23	8980	24	1740	1610	3	2012-08-02 05:11:53.593188	2012-08-02 11:21:04.780881
5847	1-5847	57	6	2973	20	683	683	0	2012-08-02 05:21:41.577288	\N
13057	1-13057	59	55	6065	52	363	363	12	2012-08-02 11:24:10.977435	\N
6123	1-6123	89	23	7802	22	2540	520	7	2012-08-02 05:23:34.513639	2012-08-02 11:26:15.336549
4971	1-4971	65	33	7473	81	1885	1885	5	2012-08-02 05:15:50.940043	\N
7004	1-7004	36	19	6864	14	2636	749	3	2012-08-02 05:29:33.328172	2012-08-02 10:35:36.746639
2445	1-2445	38	43	9670	89	126	126	5	2012-08-02 04:58:44.3053	\N
1932	1-1932	22	18	9219	57	3487	3487	0	2012-08-02 04:55:01.292261	\N
3006	1-3006	2	38	8090	15	1092	1092	0	2012-08-02 05:02:41.703978	\N
8463	1-8463	59	4	8511	77	1300	1300	5	2012-08-02 05:39:30.350557	\N
9061	1-9061	77	69	8129	87	2644	2644	5	2012-08-02 05:43:32.744	\N
10603	1-10603	13	21	6612	91	1633	1319	10	2012-08-02 10:35:36.957992	2012-08-02 11:13:31.551961
11823	1-11823	87	72	5975	52	2207	830	0	2012-08-02 11:16:10.847227	2012-08-02 11:20:53.46267
2040	1-2040	56	98	5568	86	801	345	0	2012-08-02 04:55:48.315986	2012-08-02 05:18:07.756143
11802	1-11802	79	2	9073	69	318	318	0	2012-08-02 11:16:02.571278	\N
2041	1-2041	34	49	7809	18	273	273	0	2012-08-02 04:55:48.747091	\N
2061	1-2061	65	26	4430	32	1189	1095	3	2012-08-02 04:55:57.430155	2012-08-02 04:59:54.635641
2189	1-2189	45	39	9900	36	2021	1587	0	2012-08-02 04:56:52.664837	2012-08-02 05:26:16.11233
2174	1-2174	99	82	5040	16	1595	1292	0	2012-08-02 04:56:46.080185	2012-08-02 11:25:38.40866
9328	1-9328	73	49	9475	96	2538	2055	0	2012-08-02 05:45:21.620259	2012-08-02 11:20:02.789408
2306	1-2306	35	73	5452	13	109	109	0	2012-08-02 04:57:43.591814	\N
6413	1-6413	92	40	7629	28	123	123	0	2012-08-02 05:25:34.165938	\N
13951	1-13951	81	12	7370	17	1876	1876	8	2012-08-02 11:30:05.442698	\N
10914	1-10914	48	20	7044	41	916	916	0	2012-08-02 10:38:48.397447	\N
8464	1-8464	79	55	6530	86	2726	1935	3	2012-08-02 05:39:30.770245	2012-08-02 11:15:24.120713
2159	1-2159	8	95	9976	34	751	669	0	2012-08-02 04:56:39.562094	2012-08-02 05:03:09.250277
2450	1-2450	57	62	2758	88	764	764	0	2012-08-02 04:58:46.536874	\N
7871	1-7871	17	4	7284	20	271	271	0	2012-08-02 05:35:29.103526	\N
7311	1-7311	93	32	9334	35	498	498	3	2012-08-02 05:31:37.55572	\N
2308	1-2308	34	81	9075	32	1254	1073	0	2012-08-02 04:57:44.453368	2012-08-02 10:21:25.165939
13035	1-13035	89	5	8415	95	1560	1560	11	2012-08-02 11:24:02.459561	\N
2456	1-2456	96	28	7801	16	1036	1036	0	2012-08-02 04:58:49.275338	\N
13478	1-13478	45	42	7731	84	3934	3934	8	2012-08-02 11:26:58.069399	\N
10598	1-10598	56	42	9177	37	7655	7243	0	2012-08-02 10:35:36.591334	2012-08-02 11:19:34.999848
13955	1-13955	46	37	6506	2	6388	5393	0	2012-08-02 11:30:06.614302	2012-08-02 11:30:06.614302
7872	1-7872	73	87	9492	22	290	290	2	2012-08-02 05:35:29.522943	\N
3607	1-3607	19	7	1307	41	70	70	13	2012-08-02 05:06:44.898932	\N
5856	1-5856	53	73	7329	98	2083	585	4	2012-08-02 05:21:45.200518	2012-08-02 11:13:39.588927
2723	1-2723	90	36	6051	51	497	497	0	2012-08-02 05:00:43.992457	\N
11820	1-11820	42	74	8950	4	6755	5687	0	2012-08-02 11:16:09.664644	2012-08-02 11:17:45.313023
12558	1-12558	89	38	9415	69	4511	4511	3	2012-08-02 11:20:55.840835	\N
13967	1-13967	42	77	9207	67	7625	4307	11	2012-08-02 11:30:11.709306	2012-08-02 11:30:38.533307
4976	1-4976	67	49	8149	86	501	501	0	2012-08-02 05:15:52.884552	\N
13958	1-13958	43	47	4048	43	2712	2666	4	2012-08-02 11:30:08.195798	2012-08-02 11:30:08.195798
4654	1-4654	6	81	4842	57	509	509	0	2012-08-02 05:13:44.189261	\N
2452	1-2452	63	36	6621	42	944	944	0	2012-08-02 04:58:47.40943	\N
2168	1-2168	97	67	6436	99	993	597	0	2012-08-02 04:56:43.483956	2012-08-02 05:24:46.926828
2582	1-2582	18	85	9449	53	1573	232	0	2012-08-02 04:59:44.017131	2012-08-02 11:24:55.920634
10923	1-10923	98	93	4955	24	1649	1649	3	2012-08-02 10:38:51.969752	\N
5565	1-5565	31	3	9030	37	8114	6891	4	2012-08-02 05:19:47.147566	2012-08-02 05:29:01.774837
13044	1-13044	22	18	7600	83	2741	2741	0	2012-08-02 11:24:05.91769	\N
13484	1-13484	80	51	6292	79	2555	2555	0	2012-08-02 11:27:00.401165	\N
13961	1-13961	56	67	2698	21	89	89	8	2012-08-02 11:30:09.355935	\N
4655	1-4655	48	78	6400	92	742	742	0	2012-08-02 05:13:44.597786	\N
3011	1-3011	56	47	7363	15	5503	2862	0	2012-08-02 05:02:43.769499	2012-08-02 05:46:51.430417
10602	1-10602	48	79	8821	82	4100	4100	3	2012-08-02 10:35:36.892072	\N
7874	1-7874	31	26	5025	32	819	819	4	2012-08-02 05:35:30.351659	\N
8137	1-8137	11	43	9453	49	5882	815	9	2012-08-02 05:37:18.367959	2012-08-02 11:28:16.424944
5849	1-5849	67	39	7921	29	4644	1107	12	2012-08-02 05:21:42.383441	2012-08-02 11:28:27.428723
2073	1-2073	67	22	8257	29	378	378	6	2012-08-02 04:56:02.613738	\N
7026	1-7026	66	26	4537	98	339	339	13	2012-08-02 05:29:42.08872	\N
9066	1-9066	95	61	8833	8	1193	1193	5	2012-08-02 05:43:34.76586	\N
2067	1-2067	64	66	7071	76	1493	1178	0	2012-08-02 04:56:00.00471	2012-08-02 05:37:47.339209
2075	1-2075	90	19	8594	53	7942	7942	0	2012-08-02 04:56:03.486203	\N
11815	1-11815	44	8	2259	4	665	154	0	2012-08-02 11:16:08.073344	2012-08-02 11:16:47.489209
11352	1-11352	43	9	8251	1	6527	1554	6	2012-08-02 11:13:07.84488	2012-08-02 11:30:34.543923
7876	1-7876	44	97	5405	48	1344	1344	3	2012-08-02 05:35:31.202095	\N
7029	1-7029	69	94	9092	2	8130	8130	0	2012-08-02 05:29:43.281875	\N
10933	1-10933	68	18	8456	93	470	470	6	2012-08-02 10:38:55.582016	\N
2586	1-2586	88	66	9831	31	3595	3595	0	2012-08-02 04:59:45.740543	\N
5251	1-5251	40	68	9447	4	3616	1045	0	2012-08-02 05:17:43.082396	2012-08-02 05:49:46.721681
13970	1-13970	27	97	5287	68	4653	2839	8	2012-08-02 11:30:12.885015	2012-08-02 11:30:12.885015
9620	1-9620	89	64	7719	40	1991	1991	0	2012-08-02 05:47:20.381522	\N
2065	1-2065	30	6	9709	85	2030	980	0	2012-08-02 04:55:59.142552	2012-08-02 05:05:45.938596
11826	1-11826	45	100	9235	74	3021	1348	0	2012-08-02 11:16:11.996189	2012-08-02 11:30:34.543923
2173	1-2173	35	19	8624	23	890	890	11	2012-08-02 04:56:45.638547	\N
3612	1-3612	51	24	7343	83	413	413	6	2012-08-02 05:06:46.909715	\N
4978	1-4978	57	22	2366	48	940	264	0	2012-08-02 05:15:53.647108	2012-08-02 10:38:10.340891
5858	1-5858	55	63	9429	85	947	947	6	2012-08-02 05:21:46.029648	\N
10939	1-10939	91	1	2059	15	1500	1500	4	2012-08-02 10:38:58.36892	\N
5569	1-5569	86	27	7673	100	1319	1319	6	2012-08-02 05:19:48.760546	\N
6423	1-6423	49	66	2067	6	658	658	2	2012-08-02 05:25:38.329473	\N
2578	1-2578	11	86	7822	2	5794	2361	0	2012-08-02 04:59:42.293647	2012-08-02 05:26:03.446696
2318	1-2318	19	38	7001	20	199	199	0	2012-08-02 04:57:48.806373	\N
5244	1-5244	2	53	7468	100	1060	896	0	2012-08-02 05:17:40.331247	2012-08-02 05:30:44.65002
10610	1-10610	20	24	7408	32	2382	874	7	2012-08-02 10:35:37.468643	2012-08-02 11:29:05.796114
2625	1-2625	5	84	7136	11	1277	149	0	2012-08-02 05:00:02.469081	2012-08-02 11:14:55.42302
2610	1-2610	47	49	5591	79	1927	417	7	2012-08-02 04:59:55.90616	2012-08-02 05:24:05.194268
6714	1-6714	8	77	9248	84	2218	1768	5	2012-08-02 05:27:36.236754	2012-08-02 11:15:13.522864
12314	1-12314	21	100	4266	15	9583	3231	0	2012-08-02 11:19:21.110004	2012-08-02 11:28:09.375063
3308	1-3308	58	43	2222	58	363	363	0	2012-08-02 05:04:43.848439	\N
2343	1-2343	34	36	4825	5	2055	23	0	2012-08-02 04:57:59.702083	2012-08-02 05:36:58.854214
13483	1-13483	59	99	3834	74	109	109	0	2012-08-02 11:27:00.014332	\N
12564	1-12564	86	88	7644	17	2047	2047	2	2012-08-02 11:20:58.217383	\N
2623	1-2623	95	31	9477	73	2453	2240	0	2012-08-02 05:00:01.618927	2012-08-02 05:26:59.691364
4399	1-4399	69	94	7835	88	3979	306	2	2012-08-02 05:12:01.768979	2012-08-02 05:24:19.877509
11348	1-11348	45	56	1478	97	318	318	8	2012-08-02 11:13:06.264907	\N
11807	1-11807	4	2	5726	28	757	757	4	2012-08-02 11:16:04.493531	\N
3309	1-3309	24	68	5587	48	1778	1182	0	2012-08-02 05:04:44.257403	2012-08-02 05:13:34.27819
13966	1-13966	79	41	6905	40	7054	7054	4	2012-08-02 11:30:11.311415	\N
1469	1-1469	48	55	9902	80	3535	2683	0	2012-08-02 04:51:23.182133	2012-08-02 04:58:55.503711
3019	1-3019	13	95	8398	64	1356	239	0	2012-08-02 05:02:46.940398	2012-08-02 05:40:47.068242
2352	1-2352	52	32	9494	59	3162	3162	0	2012-08-02 04:58:03.579613	\N
3753	1-3753	97	9	9909	10	4718	1764	8	2012-08-02 05:07:43.161318	2012-08-02 05:26:14.862037
5864	1-5864	2	66	9422	59	4077	3814	6	2012-08-02 05:21:48.426657	2012-08-02 05:44:01.573448
13049	1-13049	93	3	7256	4	3716	3716	0	2012-08-02 11:24:07.884209	\N
2353	1-2353	66	38	6438	12	86	86	0	2012-08-02 04:58:04.010626	\N
12301	1-12301	79	46	9535	9	3980	2753	0	2012-08-02 11:19:15.950784	2012-08-02 11:29:15.220572
9937	1-9937	100	33	5340	65	811	811	3	2012-08-02 05:49:29.902006	\N
12323	1-12323	77	31	8743	33	7663	1277	0	2012-08-02 11:19:24.579289	2012-08-02 11:29:18.780703
5871	1-5871	72	24	8310	6	4932	3051	0	2012-08-02 05:21:51.254845	2012-08-02 05:22:53.66409
2332	1-2332	100	99	2161	95	179	179	2	2012-08-02 04:57:54.962013	\N
2333	1-2333	73	11	5600	78	488	488	0	2012-08-02 04:57:55.392883	\N
5253	1-5253	34	75	7528	8	1811	1049	12	2012-08-02 05:17:43.86657	2012-08-02 11:17:41.409799
12567	1-12567	87	49	9966	93	6029	6029	0	2012-08-02 11:20:59.355374	\N
4038	1-4038	4	3	6688	36	1558	1164	0	2012-08-02 05:09:36.965289	2012-08-02 05:23:54.12682
10280	1-10280	28	41	5876	100	1693	1693	3	2012-08-02 10:25:29.597518	\N
3023	1-3023	29	6	6678	43	2105	29	0	2012-08-02 05:02:48.564504	2012-08-02 05:48:16.205043
3018	1-3018	64	90	9101	15	830	830	0	2012-08-02 05:02:46.55387	\N
3614	1-3614	98	16	5626	48	487	487	0	2012-08-02 05:06:47.705075	\N
12570	1-12570	32	39	7360	57	2312	2312	4	2012-08-02 11:21:00.515888	\N
11360	1-11360	1	67	6341	59	1017	1017	8	2012-08-02 11:13:10.938495	\N
3615	1-3615	81	62	5859	13	1286	1286	0	2012-08-02 05:06:48.091845	\N
12283	1-12283	23	53	6580	51	4951	3662	0	2012-08-02 11:19:08.990239	2012-08-02 11:26:08.173921
10281	1-10281	43	10	8475	47	927	927	0	2012-08-02 10:25:29.984026	\N
13971	1-13971	58	86	2450	33	440	440	7	2012-08-02 11:30:13.272006	\N
6129	1-6129	53	17	9331	14	1262	1262	2	2012-08-02 05:23:36.933385	\N
12295	1-12295	94	34	5171	54	2607	2002	0	2012-08-02 11:19:13.630506	2012-08-02 11:21:05.963458
2489	1-2489	91	19	9047	100	5576	4956	0	2012-08-02 04:59:03.869577	2012-08-02 05:30:44.65002
6716	1-6716	77	20	2319	11	134	134	0	2012-08-02 05:27:37.021067	\N
13968	1-13968	33	75	7325	58	9761	8455	4	2012-08-02 11:30:12.479286	2012-08-02 11:30:13.647977
7036	1-7036	87	73	3673	29	346	346	0	2012-08-02 05:29:46.154355	\N
6717	1-6717	85	62	9848	71	903	903	0	2012-08-02 05:27:37.419067	\N
9334	1-9334	7	29	6046	49	592	592	7	2012-08-02 05:45:24.051436	\N
4395	1-4395	96	69	9914	87	522	522	5	2012-08-02 05:12:00.13407	\N
11366	1-11366	5	76	7528	74	540	540	4	2012-08-02 11:13:13.20327	\N
6132	1-6132	68	60	6450	66	9	9	0	2012-08-02 05:23:38.126733	\N
3022	1-3022	35	84	8405	16	2118	2118	0	2012-08-02 05:02:48.166696	\N
13977	1-13977	66	11	6894	56	4523	4523	4	2012-08-02 11:30:15.624567	\N
13493	1-13493	50	66	9985	22	5008	5008	6	2012-08-02 11:27:03.958939	\N
13963	1-13963	71	86	6943	2	9660	8205	5	2012-08-02 11:30:10.140717	2012-08-02 11:30:14.435146
3024	1-3024	64	10	9395	46	713	713	0	2012-08-02 05:02:48.962133	\N
10609	1-10609	48	17	5984	100	431	431	3	2012-08-02 10:35:37.402193	\N
7314	1-7314	100	3	4232	49	1780	267	0	2012-08-02 05:31:38.748619	2012-08-02 10:39:26.673253
7317	1-7317	28	28	9611	96	1344	1344	0	2012-08-02 05:31:39.942104	\N
13983	1-13983	74	89	6612	99	9827	5213	0	2012-08-02 11:30:18.003509	2012-08-02 11:30:18.003509
4039	1-4039	83	28	5975	38	446	446	0	2012-08-02 05:09:37.362729	\N
5872	1-5872	93	77	8576	17	3178	1002	2	2012-08-02 05:21:51.652542	2012-08-02 11:27:38.853085
4398	1-4398	69	96	8526	85	4590	4590	0	2012-08-02 05:12:01.360141	\N
7883	1-7883	45	69	9976	74	2160	2160	0	2012-08-02 05:35:34.141105	\N
9075	1-9075	79	38	5478	31	144	144	8	2012-08-02 05:43:38.411946	\N
3316	1-3316	88	81	5600	3	252	252	2	2012-08-02 05:04:47.076942	\N
5868	1-5868	34	66	8334	37	625	625	0	2012-08-02 05:21:50.061809	\N
2733	1-2733	1	55	8782	33	1819	1320	0	2012-08-02 05:00:48.236987	2012-08-02 05:43:03.97187
4400	1-4400	5	31	9229	72	115	115	6	2012-08-02 05:12:02.179295	\N
5256	1-5256	84	58	7856	34	1339	706	0	2012-08-02 05:17:45.048667	2012-08-02 05:45:31.498135
5873	1-5873	32	84	9292	36	1663	691	0	2012-08-02 05:21:52.061614	2012-08-02 10:21:03.541599
13462	1-13462	79	45	9535	95	6277	4047	19	2012-08-02 11:26:51.562113	2012-08-02 11:30:19.190513
12317	1-12317	57	28	6643	60	1289	1289	7	2012-08-02 11:19:22.270349	\N
2747	1-2747	63	48	3359	89	156	156	0	2012-08-02 05:00:54.114493	\N
4663	1-4663	73	16	4350	61	246	246	0	2012-08-02 05:13:47.824522	\N
4671	1-4671	82	78	9379	94	1928	634	0	2012-08-02 05:13:51.038154	2012-08-02 05:38:16.126915
12297	1-12297	59	68	7786	34	8715	6294	0	2012-08-02 11:19:14.370704	2012-08-02 11:29:35.033431
8471	1-8471	76	28	6880	42	6561	996	0	2012-08-02 05:39:33.532501	2012-08-02 10:38:09.170128
2754	1-2754	97	9	7410	52	2858	2492	6	2012-08-02 05:00:57.109543	2012-08-02 05:28:18.514158
2894	1-2894	6	47	3062	42	2167	508	0	2012-08-02 05:01:55.715477	2012-08-02 11:15:42.847989
12285	1-12285	80	68	4768	60	8537	47	3	2012-08-02 11:19:09.752469	2012-08-02 11:28:23.097742
2769	1-2769	77	74	6557	73	1552	314	0	2012-08-02 05:01:03.384693	2012-08-02 11:23:59.785955
2750	1-2750	55	31	9366	23	512	512	0	2012-08-02 05:00:55.396511	\N
9628	1-9628	57	25	7805	71	1732	1551	9	2012-08-02 05:47:23.674149	2012-08-02 10:35:34.612016
4677	1-4677	85	68	6176	81	1612	405	9	2012-08-02 05:13:53.490836	2012-08-02 05:14:22.941468
2858	1-2858	82	72	7982	12	1037	1037	0	2012-08-02 05:01:40.75667	\N
9629	1-9629	16	50	6761	62	2768	1952	0	2012-08-02 05:47:24.071957	2012-08-02 11:13:02.386499
3621	1-3621	84	58	6670	1	1093	675	0	2012-08-02 05:06:50.478569	2012-08-02 05:22:20.756743
12571	1-12571	4	50	6744	88	3953	3953	8	2012-08-02 11:21:00.903046	\N
2838	1-2838	37	89	8083	79	3024	2206	0	2012-08-02 05:01:32.346712	2012-08-02 05:15:37.582481
3759	1-3759	40	97	9782	32	2878	2096	10	2012-08-02 05:07:45.548107	2012-08-02 05:49:46.721681
13056	1-13056	38	89	4697	15	2607	2607	3	2012-08-02 11:24:10.590922	\N
8169	1-8169	10	77	7531	46	1179	1179	0	2012-08-02 05:37:31.44034	\N
5004	1-5004	17	44	7492	89	1670	438	18	2012-08-02 05:16:03.967849	2012-08-02 11:27:08.268094
13976	1-13976	94	67	8824	65	4744	4744	12	2012-08-02 11:30:15.214424	\N
13489	1-13489	58	39	2076	72	367	367	6	2012-08-02 11:27:02.390033	\N
10924	1-10924	11	94	7515	24	455	455	11	2012-08-02 10:38:52.8153	\N
3348	1-3348	88	88	7717	37	1999	227	0	2012-08-02 05:05:00.036491	2012-08-02 05:09:19.249594
6137	1-6137	79	72	5412	84	1238	14	0	2012-08-02 05:23:40.115511	2012-08-02 11:27:08.65467
4662	1-4662	93	97	4367	76	1226	977	0	2012-08-02 05:13:47.427155	2012-08-02 05:30:20.02189
13059	1-13059	4	27	5253	93	6269	2560	17	2012-08-02 11:24:11.762308	2012-08-02 11:30:14.048472
2902	1-2902	58	19	8989	45	2357	2287	0	2012-08-02 05:01:59.007534	2012-08-02 05:03:15.360152
4042	1-4042	73	18	9136	60	2656	2656	0	2012-08-02 05:09:38.556028	\N
2849	1-2849	52	78	7662	51	178	178	0	2012-08-02 05:01:36.94247	\N
4406	1-4406	99	59	6083	38	1005	707	0	2012-08-02 05:12:04.620995	2012-08-02 05:12:04.620995
8754	1-8754	64	10	9404	15	2606	100	0	2012-08-02 05:41:27.613705	2012-08-02 05:43:35.980907
3332	1-3332	63	47	9748	90	4225	3289	0	2012-08-02 05:04:53.56232	2012-08-02 05:06:33.197275
2755	1-2755	86	95	6341	10	1214	940	0	2012-08-02 05:00:57.539979	2012-08-02 05:26:14.862037
3630	1-3630	26	11	9146	13	1269	1269	0	2012-08-02 05:06:54.080248	\N
11358	1-11358	38	53	7328	89	652	652	0	2012-08-02 11:13:10.176369	\N
13981	1-13981	82	91	6041	84	6833	6833	0	2012-08-02 11:30:17.613312	\N
7323	1-7323	56	96	6431	99	2181	592	0	2012-08-02 05:31:42.273014	2012-08-02 10:32:12.367358
12300	1-12300	67	58	2250	67	118	118	0	2012-08-02 11:19:15.541963	\N
7320	1-7320	88	26	5751	37	1391	1391	4	2012-08-02 05:31:41.112827	\N
12307	1-12307	4	27	9410	100	4898	4565	6	2012-08-02 11:19:18.789847	2012-08-02 11:21:54.197115
9340	1-9340	86	91	8000	77	4488	2439	0	2012-08-02 05:45:26.537042	2012-08-02 11:30:17.213447
13979	1-13979	14	98	2880	80	7167	1711	19	2012-08-02 11:30:16.37585	2012-08-02 11:30:18.003509
3414	1-3414	88	18	2315	97	97	97	0	2012-08-02 05:05:26.965719	\N
3335	1-3335	17	26	9598	28	1061	1061	0	2012-08-02 05:04:54.777502	\N
13984	1-13984	62	27	8211	86	2829	2829	4	2012-08-02 11:30:18.390142	\N
3624	1-3624	68	18	7650	1	1120	1120	0	2012-08-02 05:06:51.6716	\N
13988	1-13988	15	4	6040	29	291	291	0	2012-08-02 11:30:19.577604	\N
7047	1-7047	99	34	9898	30	587	587	0	2012-08-02 05:29:50.474415	\N
2899	1-2899	24	25	2657	80	594	594	5	2012-08-02 05:01:57.770044	\N
8472	1-8472	18	46	9874	88	1873	1873	0	2012-08-02 05:39:33.919096	\N
2854	1-2854	38	19	5801	24	810	810	0	2012-08-02 05:01:39.088393	\N
12305	1-12305	51	96	6438	14	2017	2017	11	2012-08-02 11:19:17.640995	\N
8755	1-8755	88	29	7590	39	1100	1100	0	2012-08-02 05:41:28.011355	\N
7598	1-7598	80	19	3272	7	1521	1521	0	2012-08-02 05:33:34.333462	\N
5579	1-5579	55	75	5563	7	722	722	0	2012-08-02 05:19:52.705348	\N
5878	1-5878	39	40	2024	71	101	101	0	2012-08-02 05:21:54.094328	\N
6140	1-6140	76	78	3565	91	1874	559	0	2012-08-02 05:23:41.331119	2012-08-02 11:30:20.7384
5580	1-5580	96	4	9245	67	604	604	11	2012-08-02 05:19:53.09241	\N
11821	1-11821	71	98	9353	77	5664	5664	2	2012-08-02 11:16:10.062471	\N
7048	1-7048	91	91	2142	5	215	215	10	2012-08-02 05:29:50.872283	\N
10432	1-10432	42	46	7177	30	1086	1086	3	2012-08-02 10:27:57.168693	\N
6436	1-6436	26	44	4866	80	1132	1132	5	2012-08-02 05:25:43.59086	\N
3631	1-3631	32	96	8539	81	3867	3867	0	2012-08-02 05:06:54.466787	\N
6139	1-6139	93	19	7364	17	4051	283	0	2012-08-02 05:23:40.933377	2012-08-02 05:48:36.095293
10937	1-10937	21	45	5371	71	1501	1501	8	2012-08-02 10:38:57.573509	\N
4672	1-4672	11	31	7300	2	1829	1829	0	2012-08-02 05:13:51.435876	\N
7050	1-7050	76	11	6252	52	742	742	10	2012-08-02 05:29:51.667505	\N
12310	1-12310	22	74	5809	31	576	576	5	2012-08-02 11:19:19.552093	\N
2882	1-2882	89	87	9491	28	669	669	0	2012-08-02 05:01:50.757981	\N
7888	1-7888	18	31	8833	28	3498	1175	8	2012-08-02 05:35:36.242095	2012-08-02 10:31:41.395624
4674	1-4674	8	64	7627	81	399	399	0	2012-08-02 05:13:52.24254	\N
7601	1-7601	19	46	3989	9	579	579	0	2012-08-02 05:33:35.57098	\N
2903	1-2903	78	99	6680	71	119	119	0	2012-08-02 05:01:59.416231	\N
2895	1-2895	14	97	3484	53	364	364	0	2012-08-02 05:01:56.123895	\N
2905	1-2905	62	99	8922	48	3105	3105	0	2012-08-02 05:02:00.222934	\N
2907	1-2907	74	84	5810	45	31	31	0	2012-08-02 05:02:01.095931	\N
6724	1-6724	16	86	8880	58	2717	1088	7	2012-08-02 05:27:40.203306	2012-08-02 11:24:44.12585
13492	1-13492	42	15	8462	52	9118	1682	13	2012-08-02 11:27:03.56104	2012-08-02 11:27:03.56104
3155	1-3155	8	30	9165	44	2955	2744	0	2012-08-02 05:03:42.178767	2012-08-02 05:06:00.963958
13078	1-13078	27	17	6088	87	9325	636	8	2012-08-02 11:24:18.369312	2012-08-02 11:29:48.152049
9344	1-9344	38	47	9472	58	2675	1766	5	2012-08-02 05:45:28.106155	2012-08-02 11:25:39.159937
3777	1-3777	100	54	6410	89	555	476	0	2012-08-02 05:07:52.641103	2012-08-02 05:40:51.907274
4684	1-4684	47	30	4697	19	933	882	7	2012-08-02 05:13:56.392516	2012-08-02 11:15:55.065674
6438	1-6438	51	43	9291	2	1647	1647	0	2012-08-02 05:25:44.385972	\N
5587	1-5587	27	99	7068	21	3553	2756	11	2012-08-02 05:19:55.909751	2012-08-02 05:49:26.60852
9950	1-9950	20	22	6098	84	4086	149	12	2012-08-02 05:49:35.194717	2012-08-02 11:16:06.857917
9349	1-9349	75	73	6905	12	2667	2368	10	2012-08-02 05:45:30.227363	2012-08-02 05:46:57.076851
9949	1-9949	7	9	7155	37	7327	2114	4	2012-08-02 05:49:34.797063	2012-08-02 11:18:04.813504
13495	1-13495	78	11	9704	42	2585	2585	0	2012-08-02 11:27:04.743509	\N
9637	1-9637	52	59	6344	8	1902	432	7	2012-08-02 05:47:27.330813	2012-08-02 11:17:12.67162
13507	1-13507	17	35	5198	84	9602	7746	5	2012-08-02 11:27:09.439302	2012-08-02 11:28:52.382745
11843	1-11843	60	36	7964	71	1075	1075	4	2012-08-02 11:16:18.879488	\N
4686	1-4686	44	100	7765	31	1855	811	0	2012-08-02 05:13:57.187459	2012-08-02 05:39:07.484099
2916	1-2916	41	18	8891	93	2506	2392	0	2012-08-02 05:02:04.852683	2012-08-02 11:16:09.277745
13498	1-13498	92	31	8506	35	8193	3006	2	2012-08-02 11:27:05.925457	2012-08-02 11:29:56.769677
5880	1-5880	26	52	3743	86	1566	852	0	2012-08-02 05:21:54.901364	2012-08-02 11:23:43.77499
3764	1-3764	91	74	7002	81	2844	2150	0	2012-08-02 05:07:47.514209	2012-08-02 05:27:35.849737
8474	1-8474	91	26	8409	29	1210	1210	3	2012-08-02 05:39:34.736848	\N
6440	1-6440	36	78	8118	70	2171	649	0	2012-08-02 05:25:45.181431	2012-08-02 05:34:36.055199
11392	1-11392	29	53	2500	27	1850	756	0	2012-08-02 11:13:23.246462	2012-08-02 11:25:26.565595
10936	1-10936	66	56	6820	25	8632	1976	4	2012-08-02 10:38:57.17581	2012-08-02 11:28:34.027524
13975	1-13975	77	64	7147	54	8339	8339	0	2012-08-02 11:30:14.824793	\N
3047	1-3047	79	30	8616	73	1514	1514	0	2012-08-02 05:02:58.298655	\N
3433	1-3433	96	99	8590	83	5194	1990	0	2012-08-02 05:05:34.623005	2012-08-02 05:43:26.203666
11375	1-11375	23	74	4020	27	1183	1183	5	2012-08-02 11:13:16.641688	\N
12309	1-12309	55	68	7123	24	764	764	4	2012-08-02 11:19:19.18807	\N
13058	1-13058	50	10	3669	91	1956	1756	10	2012-08-02 11:24:11.375504	2012-08-02 11:28:53.156625
12315	1-12315	15	63	4645	54	6970	6970	4	2012-08-02 11:19:21.496666	\N
13978	1-13978	7	50	6718	30	3489	3489	0	2012-08-02 11:30:16.000275	\N
3167	1-3167	62	1	9680	35	837	540	0	2012-08-02 05:03:47.095291	2012-08-02 05:06:35.617497
10289	1-10289	20	6	6441	88	924	924	2	2012-08-02 10:25:33.153734	\N
3637	1-3637	68	91	9072	43	985	985	0	2012-08-02 05:06:56.899467	\N
13070	1-13070	33	73	9340	39	4151	4151	6	2012-08-02 11:24:16.015571	\N
3634	1-3634	15	19	9593	41	1159	1159	3	2012-08-02 05:06:55.704612	\N
4049	1-4049	49	20	8262	27	1631	745	0	2012-08-02 05:09:41.384732	2012-08-02 05:40:08.371906
11395	1-11395	88	75	5061	84	2458	1686	11	2012-08-02 11:13:24.406559	2012-08-02 11:20:28.944385
13510	1-13510	1	13	6140	89	7175	7175	3	2012-08-02 11:27:10.639797	\N
5882	1-5882	23	28	8698	96	402	402	0	2012-08-02 05:21:55.696794	\N
13501	1-13501	49	32	4111	11	4084	745	0	2012-08-02 11:27:07.097589	2012-08-02 11:30:36.876255
8178	1-8178	19	98	8194	40	1529	1529	3	2012-08-02 05:37:35.108457	\N
13081	1-13081	72	25	8242	93	2144	2144	11	2012-08-02 11:24:20.326414	\N
7893	1-7893	34	27	6195	34	563	563	5	2012-08-02 05:35:38.341243	\N
3636	1-3636	53	4	9494	85	6151	1409	0	2012-08-02 05:06:56.510912	2012-08-02 11:15:13.522864
3419	1-3419	81	47	5626	95	1375	1375	0	2012-08-02 05:05:28.976561	\N
13097	1-13097	36	39	3959	22	6745	2101	0	2012-08-02 11:24:26.458117	2012-08-02 11:26:40.03285
13513	1-13513	82	51	9400	42	1975	1975	0	2012-08-02 11:27:11.79299	\N
4415	1-4415	63	54	8945	46	656	656	9	2012-08-02 05:12:08.180003	\N
3766	1-3766	61	64	9664	85	2630	2009	0	2012-08-02 05:07:48.309908	2012-08-02 05:08:59.298633
3063	1-3063	16	32	7443	53	1048	1048	0	2012-08-02 05:03:04.797324	\N
13087	1-13087	25	11	7465	16	5130	5130	17	2012-08-02 11:24:22.635578	\N
11846	1-11846	74	75	8405	52	3179	3179	5	2012-08-02 11:16:19.630619	\N
9348	1-9348	58	59	7002	87	2736	2736	4	2012-08-02 05:45:29.818621	\N
6443	1-6443	81	6	9896	10	1504	1504	0	2012-08-02 05:25:46.407885	\N
5588	1-5588	66	11	7042	76	361	361	0	2012-08-02 05:19:56.307566	\N
9944	1-9944	66	47	9401	80	5687	4621	0	2012-08-02 05:49:32.73087	2012-08-02 10:22:05.483862
4056	1-4056	62	30	9415	47	1025	1025	7	2012-08-02 05:09:44.168661	\N
3773	1-3773	88	98	9076	23	1990	188	4	2012-08-02 05:07:51.049814	2012-08-02 05:38:17.762139
3137	1-3137	23	89	7625	42	3333	1342	9	2012-08-02 05:03:34.930456	2012-08-02 05:26:26.453943
13084	1-13084	94	77	6921	2	3906	3906	4	2012-08-02 11:24:21.475431	\N
3430	1-3430	16	64	7139	59	2412	2412	0	2012-08-02 05:05:33.396247	\N
3148	1-3148	31	21	5623	69	316	316	2	2012-08-02 05:03:39.37246	\N
8763	1-8763	39	33	5457	40	4644	4644	4	2012-08-02 05:41:31.270477	\N
3150	1-3150	59	78	7200	13	219	219	7	2012-08-02 05:03:40.156767	\N
8481	1-8481	68	35	4851	16	3042	3042	0	2012-08-02 05:39:37.498952	\N
8764	1-8764	1	44	4139	45	421	421	0	2012-08-02 05:41:31.679177	\N
5593	1-5593	7	70	6288	4	974	974	0	2012-08-02 05:19:58.306946	\N
4693	1-4693	8	64	8405	5	1926	1926	0	2012-08-02 05:13:59.975296	\N
13100	1-13100	35	3	8596	4	4010	4010	0	2012-08-02 11:24:27.618038	\N
3175	1-3175	85	9	6493	64	1764	714	0	2012-08-02 05:03:50.350655	2012-08-02 11:13:39.588927
3165	1-3165	39	25	7639	91	1668	1668	0	2012-08-02 05:03:46.277813	\N
3184	1-3184	26	47	1752	23	58	58	0	2012-08-02 05:03:53.982058	\N
3457	1-3457	5	99	8680	20	1885	1885	0	2012-08-02 05:05:44.369525	\N
3464	1-3464	36	24	8989	1	2674	1222	0	2012-08-02 05:05:47.153695	2012-08-02 05:25:11.525567
8771	1-8771	18	83	4390	51	711	488	0	2012-08-02 05:41:34.563036	2012-08-02 11:13:00.408955
5273	1-5273	87	30	6711	43	2076	686	0	2012-08-02 05:17:51.72349	2012-08-02 11:16:16.582227
5015	1-5015	59	48	7080	92	2544	1284	0	2012-08-02 05:16:08.400135	2012-08-02 10:31:42.618897
10615	1-10615	24	65	4970	95	2137	695	0	2012-08-02 10:35:37.861431	2012-08-02 11:17:02.816395
11830	1-11830	17	32	7558	87	8811	8353	9	2012-08-02 11:16:13.520711	2012-08-02 11:16:13.520711
9953	1-9953	99	30	8993	79	8633	3477	4	2012-08-02 05:49:36.399655	2012-08-02 11:26:19.855415
9356	1-9356	10	27	6332	78	966	718	0	2012-08-02 05:45:33.111809	2012-08-02 10:38:23.722511
11833	1-11833	28	48	9839	47	239	239	0	2012-08-02 11:16:14.669927	\N
7898	1-7898	26	63	8636	69	2083	2083	4	2012-08-02 05:35:40.407208	\N
5020	1-5020	95	100	6047	10	2200	1456	3	2012-08-02 05:16:10.455189	2012-08-02 10:25:17.385257
13989	1-13989	15	12	6754	27	5219	5219	16	2012-08-02 11:30:20.351337	\N
4070	1-4070	20	42	5787	84	1374	427	0	2012-08-02 05:09:49.761415	2012-08-02 11:13:27.599334
13497	1-13497	55	66	9883	75	5798	5798	3	2012-08-02 11:27:05.505674	\N
3444	1-3444	36	90	8262	57	4422	1882	4	2012-08-02 05:05:39.088511	2012-08-02 05:46:27.441921
7058	1-7058	17	35	7769	26	247	247	0	2012-08-02 05:29:54.827524	\N
9947	1-9947	45	66	6593	44	6433	24	0	2012-08-02 05:49:33.97955	2012-08-02 11:25:00.269988
10943	1-10943	4	79	6359	64	588	588	3	2012-08-02 10:38:59.941021	\N
11863	1-11863	5	28	9658	57	9498	6411	16	2012-08-02 11:16:26.193924	2012-08-02 11:27:16.943419
7605	1-7605	86	98	7542	37	2708	2708	0	2012-08-02 05:33:37.195061	\N
11836	1-11836	74	30	8066	99	831	831	7	2012-08-02 11:16:15.808012	\N
7309	1-7309	45	19	4384	60	5078	665	7	2012-08-02 05:31:36.749433	2012-08-02 11:30:21.125048
13060	1-13060	83	25	9733	86	2325	2325	0	2012-08-02 11:24:12.159741	\N
10947	1-10947	72	73	2484	15	2039	2039	0	2012-08-02 10:39:01.157946	\N
4074	1-4074	19	76	4567	9	810	109	3	2012-08-02 05:09:51.308099	2012-08-02 05:18:18.42019
13992	1-13992	65	23	2947	54	4310	4310	11	2012-08-02 11:30:21.489428	\N
13503	1-13503	25	60	3064	47	2397	2397	10	2012-08-02 11:27:07.881428	\N
11373	1-11373	6	61	6741	37	2881	2881	0	2012-08-02 11:13:15.888097	\N
13995	1-13995	100	48	5464	64	681	681	0	2012-08-02 11:30:22.763602	\N
3205	1-3205	42	11	8776	40	464	464	0	2012-08-02 05:04:02.45095	\N
5018	1-5018	63	32	8669	18	2607	1114	0	2012-08-02 05:16:09.615738	2012-08-02 11:29:06.945209
13063	1-13063	66	13	8988	80	2366	2366	3	2012-08-02 11:24:13.308621	\N
6148	1-6148	36	49	5935	66	118	118	9	2012-08-02 05:23:44.558722	\N
3651	1-3651	25	50	8461	25	2032	188	6	2012-08-02 05:07:02.554458	2012-08-02 05:37:42.091071
11869	1-11869	36	43	8640	76	5874	2998	0	2012-08-02 11:16:28.538447	2012-08-02 11:26:00.681608
7059	1-7059	57	24	6747	78	212	212	0	2012-08-02 05:29:55.214196	\N
13066	1-13066	41	50	9540	6	319	319	9	2012-08-02 11:24:14.480007	\N
8487	1-8487	78	41	7922	67	2866	591	8	2012-08-02 05:39:39.896769	2012-08-02 11:24:24.160096
5013	1-5013	71	94	5477	70	1681	1681	0	2012-08-02 05:16:07.582599	\N
13509	1-13509	81	24	5101	34	2827	2827	0	2012-08-02 11:27:10.223807	\N
9956	1-9956	55	41	3086	81	1358	1358	3	2012-08-02 05:49:37.636745	\N
13998	1-13998	47	85	9212	27	2732	1724	0	2012-08-02 11:30:23.924613	2012-08-02 11:30:23.924613
13512	1-13512	3	51	7127	86	1401	1401	0	2012-08-02 11:27:11.40582	\N
6731	1-6731	21	93	4184	71	600	600	0	2012-08-02 05:27:42.965837	\N
8488	1-8488	48	63	7160	71	3630	3394	0	2012-08-02 05:39:40.294352	2012-08-02 11:21:04.780881
13069	1-13069	18	1	7165	30	52	52	0	2012-08-02 11:24:15.629235	\N
3453	1-3453	94	3	8955	73	758	758	2	2012-08-02 05:05:42.745483	\N
14001	1-14001	55	89	1955	48	31	31	0	2012-08-02 11:30:25.110376	\N
4699	1-4699	86	65	7529	81	3006	2271	0	2012-08-02 05:14:02.273372	2012-08-02 11:21:30.659848
7333	1-7333	87	68	8819	76	698	698	0	2012-08-02 05:31:46.194979	\N
9643	1-9643	46	86	6051	81	1010	1010	2	2012-08-02 05:47:29.840849	\N
14004	1-14004	39	24	7989	10	5499	5499	10	2012-08-02 11:30:26.270836	\N
6450	1-6450	93	15	6519	27	688	688	0	2012-08-02 05:25:49.180926	\N
3455	1-3455	77	64	8653	84	757	757	0	2012-08-02 05:05:43.574259	\N
7902	1-7902	31	47	5878	20	4055	3956	7	2012-08-02 05:35:42.119643	2012-08-02 11:14:13.157551
10621	1-10621	2	4	4666	85	1177	1177	4	2012-08-02 10:35:38.304119	\N
9958	1-9958	63	62	8756	86	820	820	7	2012-08-02 05:49:38.443033	\N
5283	1-5283	80	24	9963	4	6168	5329	4	2012-08-02 05:17:55.590449	2012-08-02 11:15:09.432839
3496	1-3496	80	55	9442	75	1849	1394	0	2012-08-02 05:06:00.168571	2012-08-02 05:14:31.75932
9938	1-9938	8	94	7122	15	5826	5436	5	2012-08-02 05:49:30.310637	2012-08-02 11:13:18.950085
3468	1-3468	56	26	6222	63	495	495	0	2012-08-02 05:05:48.788671	\N
4071	1-4071	60	98	7753	49	843	843	0	2012-08-02 05:09:50.148053	\N
14007	1-14007	70	91	7735	9	1672	1672	9	2012-08-02 11:30:27.497934	\N
7609	1-7609	7	65	9239	54	2697	2697	0	2012-08-02 05:33:38.807903	\N
14013	1-14013	12	95	2206	24	1896	1896	0	2012-08-02 11:30:29.841556	\N
3473	1-3473	36	69	6094	57	1402	1402	2	2012-08-02 05:05:50.799493	\N
4425	1-4425	4	49	9399	41	1813	1645	11	2012-08-02 05:12:12.124941	2012-08-02 05:22:37.331277
7903	1-7903	24	35	7845	37	756	756	3	2012-08-02 05:35:42.539545	\N
14010	1-14010	78	34	2591	13	2269	1943	0	2012-08-02 11:30:28.670191	2012-08-02 11:30:36.876255
8489	1-8489	100	73	7096	95	412	412	6	2012-08-02 05:39:40.714569	\N
3802	1-3802	97	9	5177	25	565	565	8	2012-08-02 05:08:02.407543	\N
3483	1-3483	3	8	9892	15	1801	1801	0	2012-08-02 05:05:54.799713	\N
3490	1-3490	59	18	7876	10	1431	1431	7	2012-08-02 05:05:57.78248	\N
3491	1-3491	57	89	6944	69	1281	1281	0	2012-08-02 05:05:58.191078	\N
3492	1-3492	63	94	6833	18	249	249	11	2012-08-02 05:05:58.577501	\N
9957	1-9957	35	90	9558	100	2839	1497	0	2012-08-02 05:49:38.03433	2012-08-02 11:17:00.418404
6032	1-6032	78	33	5136	92	5110	1714	0	2012-08-02 05:22:57.796202	2012-08-02 11:29:43.786822
9095	1-9095	71	74	5062	88	5864	710	2	2012-08-02 05:43:46.665033	2012-08-02 11:30:04.669386
6027	1-6027	18	94	9035	70	4503	3680	0	2012-08-02 05:22:55.75244	2012-08-02 11:23:54.703668
13072	1-13072	37	20	6802	37	7116	3327	5	2012-08-02 11:24:16.391212	2012-08-02 11:30:09.753486
9359	1-9359	85	94	9092	97	3418	2665	8	2012-08-02 05:45:34.360644	2012-08-02 05:47:37.662999
12334	1-12334	5	89	9330	66	3385	2094	3	2012-08-02 11:19:28.84408	2012-08-02 11:26:00.681608
9093	1-9093	95	8	7933	86	3802	940	7	2012-08-02 05:43:45.858252	2012-08-02 10:16:59.437265
3807	1-3807	76	64	8619	79	1602	1602	0	2012-08-02 05:08:04.385447	\N
13987	1-13987	97	31	1294	8	449	449	3	2012-08-02 11:30:19.975864	\N
12319	1-12319	22	13	8655	45	8749	6997	8	2012-08-02 11:19:23.065801	2012-08-02 11:25:41.957032
13502	1-13502	13	11	9240	37	3411	3411	0	2012-08-02 11:27:07.494527	\N
12322	1-12322	87	32	8761	49	2376	2376	14	2012-08-02 11:19:24.203595	\N
4080	1-4080	85	49	9975	100	1847	1128	0	2012-08-02 05:09:53.717097	2012-08-02 05:47:04.821364
10623	1-10623	89	12	7291	2	721	721	3	2012-08-02 10:35:38.459683	\N
8494	1-8494	42	55	9590	5	325	325	0	2012-08-02 05:39:42.725432	\N
8777	1-8777	50	3	7353	16	6344	4789	0	2012-08-02 05:41:36.982512	2012-08-02 10:38:02.915583
13997	1-13997	7	14	5102	37	8000	8000	3	2012-08-02 11:30:23.538004	\N
13508	1-13508	56	90	3001	84	1012	1012	2	2012-08-02 11:27:09.8373	\N
10624	1-10624	19	31	9299	90	92	92	0	2012-08-02 10:35:38.537314	\N
3875	1-3875	45	3	5398	93	3634	3509	0	2012-08-02 05:08:31.700209	2012-08-02 05:21:54.491972
7063	1-7063	99	72	8649	70	2359	2359	6	2012-08-02 05:29:56.827612	\N
3818	1-3818	47	21	9386	18	548	548	2	2012-08-02 05:08:08.807445	\N
8185	1-8185	28	26	8146	76	3573	2236	0	2012-08-02 05:37:38.003146	2012-08-02 05:37:38.003146
6457	1-6457	65	21	9361	81	544	544	0	2012-08-02 05:25:51.920673	\N
3834	1-3834	100	76	4701	70	7	7	0	2012-08-02 05:08:15.270814	\N
7336	1-7336	30	99	8159	55	3157	3157	3	2012-08-02 05:31:47.399172	\N
13999	1-13999	64	48	5866	86	2273	2273	11	2012-08-02 11:30:24.720723	\N
12591	1-12591	70	24	7701	48	99	99	9	2012-08-02 11:21:08.769754	\N
10625	1-10625	94	68	9830	46	859	859	9	2012-08-02 10:35:38.614427	\N
3821	1-3821	74	67	3753	17	55	55	0	2012-08-02 05:08:10.023271	\N
6156	1-6156	90	8	8984	54	6377	6377	13	2012-08-02 05:23:47.729644	\N
14002	1-14002	47	42	4184	20	798	798	0	2012-08-02 11:30:25.497028	\N
6029	1-6029	8	9	7284	44	5484	3551	2	2012-08-02 05:22:56.580868	2012-08-02 11:24:17.573325
6157	1-6157	19	27	3369	64	326	326	0	2012-08-02 05:23:48.127263	\N
4436	1-4436	70	100	9852	89	732	732	0	2012-08-02 05:12:16.466536	\N
5024	1-5024	20	76	6702	33	1639	1111	0	2012-08-02 05:16:12.112345	2012-08-02 11:14:13.930346
4086	1-4086	50	35	9403	75	569	569	0	2012-08-02 05:09:56.081788	\N
8497	1-8497	58	84	7092	18	661	661	0	2012-08-02 05:39:43.929434	\N
3827	1-3827	95	73	8109	67	1161	1161	5	2012-08-02 05:08:12.453351	\N
11867	1-11867	44	68	7369	5	8140	3434	5	2012-08-02 11:16:27.764895	2012-08-02 11:25:19.091902
7065	1-7065	91	6	3582	86	791	791	5	2012-08-02 05:29:57.61175	\N
4789	1-4789	36	77	7490	100	1079	1079	3	2012-08-02 05:14:38.653468	\N
5027	1-5027	31	32	7579	19	152	152	0	2012-08-02 05:16:13.316899	\N
3836	1-3836	30	15	5955	62	50	50	4	2012-08-02 05:08:16.143844	\N
3864	1-3864	82	39	7672	85	4700	2390	0	2012-08-02 05:08:27.31397	2012-08-02 05:22:12.823799
14005	1-14005	58	55	7668	99	9778	9778	8	2012-08-02 11:30:26.657313	\N
8500	1-8500	51	88	2981	2	626	626	8	2012-08-02 05:39:45.14472	\N
10298	1-10298	47	47	8361	33	703	703	14	2012-08-02 10:25:36.71173	\N
12337	1-12337	65	77	8298	60	1934	1934	4	2012-08-02 11:19:29.98193	\N
8187	1-8187	24	63	7901	98	3510	3481	0	2012-08-02 05:37:38.83188	2012-08-02 05:37:42.091071
4451	1-4451	15	50	4567	2	827	321	2	2012-08-02 05:12:22.344419	2012-08-02 05:26:32.652602
11847	1-11847	74	7	6201	45	4241	3216	0	2012-08-02 11:16:20.017478	2012-08-02 11:23:32.018196
14008	1-14008	36	3	9037	90	7765	7765	8	2012-08-02 11:30:27.884978	\N
7908	1-7908	92	68	5882	70	4774	1231	2	2012-08-02 05:35:44.694344	2012-08-02 11:27:17.739084
8184	1-8184	95	74	6424	47	3154	1463	0	2012-08-02 05:37:37.594526	2012-08-02 09:10:31.355931
14014	1-14014	72	56	4718	13	1139	1139	4	2012-08-02 11:30:30.229797	\N
3888	1-3888	41	24	7757	51	4394	1390	0	2012-08-02 05:08:36.895049	2012-08-02 11:26:26.596176
11859	1-11859	65	8	5391	38	3846	871	4	2012-08-02 11:16:24.658039	2012-08-02 11:23:02.582878
3843	1-3843	53	19	9700	16	2987	2987	0	2012-08-02 05:08:18.916469	\N
9096	1-9096	1	11	8598	72	3607	3607	0	2012-08-02 05:43:47.073453	\N
14011	1-14011	99	78	5153	32	8789	7763	2	2012-08-02 11:30:29.07869	2012-08-02 11:30:29.07869
3873	1-3873	89	75	9736	21	2252	2252	0	2012-08-02 05:08:30.915987	\N
8503	1-8503	11	48	9862	60	746	746	5	2012-08-02 05:39:46.36047	\N
14020	1-14020	30	48	2595	55	1096	1096	16	2012-08-02 11:30:32.594021	\N
10302	1-10302	61	22	4407	44	69	69	7	2012-08-02 10:25:38.281035	\N
3853	1-3853	75	75	6705	95	280	280	0	2012-08-02 05:08:22.905427	\N
8776	1-8776	47	86	3821	42	2222	470	6	2012-08-02 05:41:36.584798	2012-08-02 11:20:34.782688
8498	1-8498	82	26	8591	79	5535	1877	6	2012-08-02 05:39:44.349469	2012-08-02 05:40:37.000549
14023	1-14023	28	34	4844	44	8404	8404	11	2012-08-02 11:30:33.745056	\N
14025	1-14025	58	11	7831	23	8941	8941	11	2012-08-02 11:30:34.919264	\N
14017	1-14017	81	40	8087	11	5920	1983	0	2012-08-02 11:30:31.444941	2012-08-02 11:30:36.876255
3854	1-3854	71	25	7933	41	157	157	0	2012-08-02 05:08:23.314126	\N
3879	1-3879	46	4	8002	22	131	131	8	2012-08-02 05:08:33.281917	\N
3883	1-3883	27	35	7154	44	4551	4551	0	2012-08-02 05:08:34.884352	\N
3896	1-3896	40	25	7755	38	277	277	0	2012-08-02 05:08:40.143199	\N
11879	1-11879	66	73	6056	42	8666	2313	7	2012-08-02 11:16:31.97499	2012-08-02 11:29:56.769677
13088	1-13088	17	48	7424	91	7042	679	7	2012-08-02 11:24:23.022771	2012-08-02 11:29:45.743019
9099	1-9099	92	46	2406	52	257	257	6	2012-08-02 05:43:48.323877	\N
5612	1-5612	88	32	7794	36	1126	995	15	2012-08-02 05:20:05.944	2012-08-02 11:21:53.412361
4479	1-4479	34	36	3614	26	695	679	3	2012-08-02 05:12:33.439093	2012-08-02 05:28:52.713097
9967	1-9967	38	81	4623	36	1165	488	9	2012-08-02 05:49:42.168578	2012-08-02 11:23:58.228472
6746	1-6746	27	41	7012	5	4409	373	0	2012-08-02 05:27:48.899012	2012-08-02 11:21:04.780881
8191	1-8191	50	50	5969	68	1231	96	0	2012-08-02 05:37:40.456238	2012-08-02 11:22:59.10205
13079	1-13079	4	12	8654	63	216	216	6	2012-08-02 11:24:19.53134	\N
7075	1-7075	54	8	9291	54	8887	3237	13	2012-08-02 05:30:01.612196	2012-08-02 11:15:05.510735
13518	1-13518	77	19	7420	79	3449	3449	5	2012-08-02 11:27:13.784085	\N
7073	1-7073	37	66	9313	50	1977	1977	0	2012-08-02 05:30:00.82772	\N
11855	1-11855	29	84	9021	30	2023	2023	5	2012-08-02 11:16:23.122559	\N
11865	1-11865	48	97	8357	86	6937	6927	6	2012-08-02 11:16:26.980562	2012-08-02 11:18:58.625043
10626	1-10626	79	12	9070	47	1714	1714	10	2012-08-02 10:35:38.692097	\N
6471	1-6471	78	97	9172	72	6476	1676	0	2012-08-02 05:25:57.480443	2012-08-02 10:22:14.080368
3925	1-3925	29	51	8366	8	1126	821	0	2012-08-02 05:08:51.696275	2012-08-02 05:33:13.185325
8192	1-8192	86	42	2658	72	751	643	15	2012-08-02 05:37:40.865352	2012-08-02 05:45:30.69134
5040	1-5040	59	14	2009	38	54	54	2	2012-08-02 05:16:18.564896	\N
13076	1-13076	58	100	8199	73	3846	3846	5	2012-08-02 11:24:18.766926	\N
12326	1-12326	7	88	7594	10	2498	226	0	2012-08-02 11:19:25.739336	2012-08-02 11:30:27.067488
4711	1-4711	36	70	8042	74	986	902	0	2012-08-02 05:14:07.094881	2012-08-02 10:27:57.953321
3918	1-3918	100	97	6539	44	2799	770	9	2012-08-02 05:08:48.916325	2012-08-02 11:20:07.043702
4139	1-4139	52	1	9509	2	8377	2026	0	2012-08-02 05:10:17.176736	2012-08-02 11:29:15.971766
3855	1-3855	26	74	8385	54	2118	1549	10	2012-08-02 05:08:23.7011	2012-08-02 05:15:54.420207
4478	1-4478	9	13	5657	79	1262	1007	4	2012-08-02 05:12:33.052325	2012-08-02 05:23:14.126826
5048	1-5048	26	22	8913	63	271	271	3	2012-08-02 05:16:21.802067	\N
6465	1-6465	70	64	3820	81	1454	79	0	2012-08-02 05:25:55.126619	2012-08-02 11:21:30.659848
6163	1-6163	77	9	9590	59	538	538	0	2012-08-02 05:23:50.568996	\N
6469	1-6469	33	15	9347	12	439	439	5	2012-08-02 05:25:56.707005	\N
9361	1-9361	73	90	6483	94	1113	1113	0	2012-08-02 05:45:35.155704	\N
9963	1-9963	24	56	9516	6	283	283	12	2012-08-02 05:49:40.511504	\N
13523	1-13523	80	87	9645	39	993	993	0	2012-08-02 11:27:15.783338	\N
6463	1-6463	92	20	4872	37	1683	1683	5	2012-08-02 05:25:54.320475	\N
9100	1-9100	23	44	4758	63	117	117	0	2012-08-02 05:43:48.721453	\N
12338	1-12338	70	59	9242	74	4184	4184	2	2012-08-02 11:19:30.357527	\N
13085	1-13085	86	35	8394	99	3450	3450	5	2012-08-02 11:24:21.862404	\N
3928	1-3928	30	83	9303	87	449	449	0	2012-08-02 05:08:52.926891	\N
11399	1-11399	67	34	7084	82	1144	1144	0	2012-08-02 11:13:25.93133	\N
11875	1-11875	79	35	6653	27	3003	670	0	2012-08-02 11:16:30.449803	2012-08-02 11:24:20.71302
3930	1-3930	99	82	9336	30	1723	1632	3	2012-08-02 05:08:53.710784	2012-08-02 11:16:48.264351
9101	1-9101	86	39	3199	54	275	275	0	2012-08-02 05:43:49.130432	\N
11871	1-11871	32	49	9103	28	225	225	4	2012-08-02 11:16:29.311817	\N
3938	1-3938	77	17	7623	2	2764	1017	0	2012-08-02 05:08:56.921581	2012-08-02 05:24:47.755777
5045	1-5045	64	56	8747	16	1237	1237	0	2012-08-02 05:16:20.575493	\N
6467	1-6467	26	55	4937	77	991	991	0	2012-08-02 05:25:55.911036	\N
9103	1-9103	53	63	9589	65	1338	1338	0	2012-08-02 05:43:49.983189	\N
11405	1-11405	44	23	5142	63	2015	2015	0	2012-08-02 11:13:28.406552	\N
8507	1-8507	52	64	4007	36	657	582	0	2012-08-02 05:39:47.94184	2012-08-02 05:48:28.758694
11882	1-11882	94	99	9314	51	862	862	2	2012-08-02 11:16:33.532562	\N
5913	1-5913	1	13	8828	16	4665	2270	0	2012-08-02 05:22:08.272472	2012-08-02 10:39:37.23284
6747	1-6747	28	19	8628	69	4059	4059	0	2012-08-02 05:27:49.285714	\N
4720	1-4720	65	13	6689	31	916	916	0	2012-08-02 05:14:10.808064	\N
13093	1-13093	91	46	4269	16	2995	2995	0	2012-08-02 11:24:24.944663	\N
11885	1-11885	87	84	6102	85	2566	2566	5	2012-08-02 11:16:34.681328	\N
8196	1-8196	33	12	7457	22	270	270	8	2012-08-02 05:37:42.511153	\N
8510	1-8510	49	65	7030	20	763	763	6	2012-08-02 05:39:49.157329	\N
13099	1-13099	62	67	7256	84	3725	3725	0	2012-08-02 11:24:27.231377	\N
10632	1-10632	39	97	6462	51	1312	1312	0	2012-08-02 10:35:39.145365	\N
4125	1-4125	70	3	4158	48	292	292	3	2012-08-02 05:10:11.606382	\N
13102	1-13102	95	73	9653	50	2308	2308	7	2012-08-02 11:24:28.380495	\N
6170	1-6170	42	41	9590	13	705	705	0	2012-08-02 05:23:53.342251	\N
13105	1-13105	52	62	8774	29	1281	1281	3	2012-08-02 11:24:29.563217	\N
4482	1-4482	51	55	3574	44	1852	751	0	2012-08-02 05:12:34.61033	2012-08-02 05:44:59.287494
10304	1-10304	69	96	9907	80	6813	2701	0	2012-08-02 10:25:39.065828	2012-08-02 11:24:30.358203
11862	1-11862	63	57	9304	7	5057	2329	3	2012-08-02 11:16:25.420378	2012-08-02 11:18:04.813504
5056	1-5056	16	79	9980	17	1813	1813	3	2012-08-02 05:16:25.017068	\N
11418	1-11418	81	78	9200	57	6565	6565	5	2012-08-02 11:13:33.42402	\N
9108	1-9108	90	63	3436	90	1062	1062	0	2012-08-02 05:43:52.088066	\N
9109	1-9109	8	70	9168	17	349	349	5	2012-08-02 05:43:52.507956	\N
4475	1-4475	24	66	7412	50	1731	1731	0	2012-08-02 05:12:31.859441	\N
4476	1-4476	28	13	7690	54	563	563	0	2012-08-02 05:12:32.25716	\N
4477	1-4477	55	31	7761	54	230	230	7	2012-08-02 05:12:32.64401	\N
3951	1-3951	88	75	8749	40	370	370	0	2012-08-02 05:09:02.128839	\N
10637	1-10637	8	33	7937	57	6918	5007	0	2012-08-02 10:35:39.533819	2012-08-02 11:20:07.83926
4188	1-4188	53	24	8034	47	2602	1406	0	2012-08-02 05:10:36.566891	2012-08-02 05:24:05.703273
13993	1-13993	19	90	7087	80	9293	7930	5	2012-08-02 11:30:22.366312	2012-08-02 11:30:22.366312
10634	1-10634	94	7	7233	96	828	828	6	2012-08-02 10:35:39.300806	\N
4723	1-4723	2	59	3782	53	570	509	8	2012-08-02 05:14:12.034347	2012-08-02 05:23:07.342959
13516	1-13516	65	48	8546	15	8341	8341	14	2012-08-02 11:27:12.974695	\N
8788	1-8788	77	97	7928	79	7251	1047	7	2012-08-02 05:41:41.402094	2012-08-02 11:11:55.599876
10635	1-10635	15	28	9396	61	6689	3325	3	2012-08-02 10:35:39.378406	2012-08-02 10:38:16.66256
4174	1-4174	7	22	7152	90	1567	1567	2	2012-08-02 05:10:31.053615	\N
5062	1-5062	49	47	9641	23	4586	1420	0	2012-08-02 05:16:27.449408	2012-08-02 05:43:16.713451
5313	1-5313	78	82	9686	4	3018	2499	0	2012-08-02 05:18:07.358296	2012-08-02 11:16:52.97174
9653	1-9653	26	52	9574	57	162	162	0	2012-08-02 05:47:33.92861	\N
9366	1-9366	12	66	5799	91	1302	1302	16	2012-08-02 05:45:37.166601	\N
4151	1-4151	18	3	5406	80	2972	2972	0	2012-08-02 05:10:22.038094	\N
4223	1-4223	58	97	6139	50	1746	828	0	2012-08-02 05:10:50.302471	2012-08-02 05:39:37.896496
12351	1-12351	15	27	9808	3	4887	3076	0	2012-08-02 11:19:35.397695	2012-08-02 11:26:48.433006
4189	1-4189	47	4	5448	38	1834	461	0	2012-08-02 05:10:36.953981	2012-08-02 05:25:23.678507
7620	1-7620	29	41	9167	83	1021	1021	5	2012-08-02 05:33:43.348979	\N
4175	1-4175	50	92	5446	95	219	219	0	2012-08-02 05:10:31.440497	\N
4221	1-4221	61	10	7038	41	971	119	2	2012-08-02 05:10:49.506878	2012-08-02 05:31:20.295902
6756	1-6756	36	72	7306	95	2349	655	11	2012-08-02 05:27:52.856504	2012-08-02 05:43:53.325755
4178	1-4178	35	31	6388	91	309	95	0	2012-08-02 05:10:32.6114	2012-08-02 05:19:30.064724
4157	1-4157	88	45	8354	58	1245	1198	0	2012-08-02 05:10:24.391459	2012-08-02 11:24:09.430813
4724	1-4724	22	26	9978	20	944	944	0	2012-08-02 05:14:12.443032	\N
4158	1-4158	82	86	8188	57	976	976	0	2012-08-02 05:10:24.767085	\N
7925	1-7925	47	43	9112	1	1642	1642	6	2012-08-02 05:35:51.930327	\N
12587	1-12587	33	90	8555	43	2868	2868	3	2012-08-02 11:21:07.195441	\N
12345	1-12345	47	75	5497	27	2269	2269	3	2012-08-02 11:19:33.066434	\N
4168	1-4168	33	81	9686	94	1511	1362	0	2012-08-02 05:10:28.667185	2012-08-02 05:21:33.390577
4201	1-4201	15	69	5502	45	917	917	8	2012-08-02 05:10:41.651225	\N
4235	1-4235	86	69	8119	93	3941	1192	0	2012-08-02 05:10:54.97619	2012-08-02 11:24:10.204408
6173	1-6173	70	82	8836	31	2019	1039	7	2012-08-02 05:23:54.548957	2012-08-02 11:15:01.245814
13075	1-13075	86	64	6658	16	945	945	3	2012-08-02 11:24:17.960305	\N
7081	1-7081	3	1	6317	80	556	556	0	2012-08-02 05:30:03.967356	\N
13521	1-13521	96	48	6729	92	1671	1671	2	2012-08-02 11:27:14.20399	\N
6754	1-6754	72	25	1925	59	269	269	5	2012-08-02 05:27:52.061033	\N
6761	1-6761	1	15	6589	38	1154	580	4	2012-08-02 05:27:54.867274	2012-08-02 11:30:39.30669
8516	1-8516	48	28	9543	43	1566	1566	0	2012-08-02 05:39:51.53244	\N
5059	1-5059	29	9	4653	53	592	592	5	2012-08-02 05:16:26.243414	\N
8517	1-8517	15	17	7065	6	1587	893	9	2012-08-02 05:39:51.919155	2012-08-02 05:48:39.321516
11876	1-11876	51	48	9742	98	1684	1684	0	2012-08-02 11:16:31.223452	\N
7927	1-7927	84	37	6794	16	2356	2356	6	2012-08-02 05:35:52.780989	\N
5309	1-5309	82	88	5458	52	367	367	6	2012-08-02 05:18:05.800558	\N
13524	1-13524	92	40	9192	49	1504	1504	3	2012-08-02 11:27:16.18103	\N
11397	1-11397	66	64	8332	86	1223	1223	0	2012-08-02 11:13:25.17982	\N
5061	1-5061	95	46	6671	5	137	137	0	2012-08-02 05:16:27.051895	\N
13527	1-13527	38	4	1416	97	86	86	0	2012-08-02 11:27:17.341108	\N
6757	1-6757	54	50	6729	5	1200	1200	5	2012-08-02 05:27:53.254698	\N
11878	1-11878	72	73	5189	41	752	752	13	2012-08-02 11:16:32.367442	\N
4734	1-4734	32	28	8842	10	690	690	0	2012-08-02 05:14:16.555444	\N
11400	1-11400	24	37	5891	98	101	101	6	2012-08-02 11:13:26.307117	\N
9970	1-9970	6	55	7687	98	2452	2452	0	2012-08-02 05:49:43.40623	\N
10309	1-10309	61	44	8726	5	4753	3697	7	2012-08-02 10:25:41.066434	2012-08-02 11:12:01.509541
3866	1-3866	18	35	7592	77	3246	1927	0	2012-08-02 05:08:28.109128	2012-08-02 05:43:54.966471
12593	1-12593	90	11	9815	21	6524	6434	5	2012-08-02 11:21:09.565352	2012-08-02 11:21:48.407505
12342	1-12342	22	28	9154	53	8822	3588	0	2012-08-02 11:19:31.917465	2012-08-02 11:25:11.652325
9649	1-9649	42	42	8823	40	6323	2671	0	2012-08-02 05:47:32.31544	2012-08-02 11:13:54.195816
4179	1-4179	65	6	9476	40	2940	2940	0	2012-08-02 05:10:33.009222	\N
7929	1-7929	45	54	7344	78	563	308	15	2012-08-02 05:35:53.620496	2012-08-02 11:22:36.670154
4733	1-4733	53	30	4369	92	2511	561	0	2012-08-02 05:14:16.122854	2012-08-02 05:18:52.531287
12360	1-12360	46	78	9377	81	2643	2643	0	2012-08-02 11:19:38.822512	\N
11837	1-11837	89	42	8906	64	7514	5759	0	2012-08-02 11:16:16.194666	2012-08-02 11:20:04.336464
4226	1-4226	57	93	8231	21	1912	1877	2	2012-08-02 05:10:51.45147	2012-08-02 11:28:13.275174
9968	1-9968	64	26	7395	42	5282	5052	4	2012-08-02 05:49:42.577264	2012-08-02 11:28:27.428723
9369	1-9369	90	44	9622	25	5827	890	5	2012-08-02 05:45:38.38155	2012-08-02 11:29:12.470004
7630	1-7630	66	4	5113	54	1040	1040	10	2012-08-02 05:33:47.448081	\N
4192	1-4192	61	31	6274	80	1044	1044	0	2012-08-02 05:10:38.14706	\N
8792	1-8792	70	36	5990	5	5747	5747	4	2012-08-02 05:41:43.014985	\N
6764	1-6764	41	60	6208	34	214	214	0	2012-08-02 05:27:56.049262	\N
5932	1-5932	92	89	7625	54	2017	2017	0	2012-08-02 05:22:16.171969	\N
7622	1-7622	54	17	9554	45	2226	1242	7	2012-08-02 05:33:44.177765	2012-08-02 05:34:15.169192
4204	1-4204	96	78	9328	71	86	86	3	2012-08-02 05:10:42.833372	\N
4209	1-4209	52	24	8261	30	2114	2114	0	2012-08-02 05:10:44.84466	\N
4216	1-4216	71	20	9220	24	411	411	0	2012-08-02 05:10:47.58467	\N
4227	1-4227	16	75	8412	60	2576	2576	0	2012-08-02 05:10:51.827176	\N
4742	1-4742	30	1	6835	13	4117	1050	0	2012-08-02 05:14:19.77049	2012-08-02 11:25:08.490508
9124	1-9124	69	84	7912	5	2270	2020	11	2012-08-02 05:43:58.662433	2012-08-02 11:29:09.320977
5084	1-5084	100	36	6478	56	1994	234	0	2012-08-02 05:16:36.453728	2012-08-02 11:14:30.77091
4559	1-4559	78	50	7639	39	991	485	0	2012-08-02 05:13:05.162751	2012-08-02 11:15:42.847989
4244	1-4244	76	22	9760	16	718	718	3	2012-08-02 05:10:58.478407	\N
4252	1-4252	79	56	5683	5	3501	962	0	2012-08-02 05:11:01.627044	2012-08-02 05:17:12.07713
13506	1-13506	91	46	9947	13	7240	6212	7	2012-08-02 11:27:09.052667	2012-08-02 11:30:16.762619
13996	1-13996	45	72	1829	80	4022	523	0	2012-08-02 11:30:23.150851	2012-08-02 11:30:23.150851
11903	1-11903	99	93	9773	92	4090	3073	0	2012-08-02 11:16:41.951077	2012-08-02 11:28:13.275174
5069	1-5069	19	57	8430	92	3059	400	0	2012-08-02 05:16:30.289299	2012-08-02 05:28:21.76258
4247	1-4247	22	52	7546	69	643	643	0	2012-08-02 05:10:59.649567	\N
4741	1-4741	6	43	9380	37	1910	1910	0	2012-08-02 05:14:19.384162	\N
4771	1-4771	95	65	6806	49	1272	1216	5	2012-08-02 05:14:31.350469	2012-08-02 11:13:49.190803
13092	1-13092	7	8	8356	90	8357	7075	4	2012-08-02 11:24:24.558228	2012-08-02 11:26:28.176291
4540	1-4540	2	21	8699	2	4425	3381	5	2012-08-02 05:12:57.660089	2012-08-02 05:32:19.232845
14000	1-14000	58	66	2377	80	7843	4035	0	2012-08-02 11:30:24.323159	2012-08-02 11:30:24.323159
13519	1-13519	72	15	3475	52	4674	2774	16	2012-08-02 11:27:14.601302	2012-08-02 11:28:30.157943
9123	1-9123	61	8	8016	56	1845	1816	4	2012-08-02 05:43:58.26491	2012-08-02 11:22:12.815234
4254	1-4254	20	75	9879	50	3359	3359	0	2012-08-02 05:11:02.389497	\N
6484	1-6484	63	66	5510	98	1818	607	13	2012-08-02 05:26:02.640558	2012-08-02 11:19:50.059928
4782	1-4782	13	85	9640	38	1459	1445	0	2012-08-02 05:14:35.847015	2012-08-02 05:33:36.388437
12343	1-12343	98	42	5626	27	2894	867	9	2012-08-02 11:19:32.293096	2012-08-02 11:27:32.588199
12589	1-12589	77	20	4525	51	1900	1900	5	2012-08-02 11:21:07.974257	\N
4256	1-4256	38	41	3522	34	1110	1110	0	2012-08-02 05:11:03.173887	\N
7371	1-7371	22	1	7995	48	3837	3436	2	2012-08-02 05:32:01.211956	2012-08-02 10:39:06.694644
13086	1-13086	24	36	6982	15	6294	6294	2	2012-08-02 11:24:22.249019	\N
4751	1-4751	94	49	9516	40	2624	2348	2	2012-08-02 05:14:23.350603	2012-08-02 05:22:24.104661
6768	1-6768	63	99	6167	60	1631	1631	0	2012-08-02 05:27:57.640214	\N
9974	1-9974	20	93	5154	56	1388	379	6	2012-08-02 05:49:45.119216	2012-08-02 11:29:24.79138
4558	1-4558	24	23	9185	89	2723	2325	8	2012-08-02 05:13:04.775578	2012-08-02 11:22:41.775281
11877	1-11877	44	91	4437	52	2447	2447	12	2012-08-02 11:16:31.599037	\N
4261	1-4261	97	77	9052	62	559	559	2	2012-08-02 05:11:05.107765	\N
14006	1-14006	40	49	2055	46	5469	1315	8	2012-08-02 11:30:27.067488	2012-08-02 11:30:27.067488
5935	1-5935	43	96	6387	50	1475	1475	4	2012-08-02 05:22:17.442404	\N
5634	1-5634	12	88	9427	70	998	998	11	2012-08-02 05:20:14.983429	\N
7636	1-7636	42	63	8304	6	2713	294	0	2012-08-02 05:33:49.908658	2012-08-02 05:37:25.789034
10965	1-10965	24	29	9079	92	4176	4176	2	2012-08-02 10:39:08.640138	\N
8202	1-8202	76	20	6597	70	1702	1205	4	2012-08-02 05:37:44.931054	2012-08-02 11:22:39.388621
14012	1-14012	75	83	9197	57	5166	5166	10	2012-08-02 11:30:29.466261	\N
7367	1-7367	20	61	2886	92	1091	1091	9	2012-08-02 05:31:59.609933	\N
4512	1-4512	60	95	6769	58	326	326	0	2012-08-02 05:12:46.498651	\N
5267	1-5267	5	13	8367	31	3040	3020	0	2012-08-02 05:17:49.37013	2012-08-02 11:27:16.568012
13095	1-13095	84	84	8775	11	1429	1429	8	2012-08-02 11:24:25.695848	\N
10968	1-10968	94	21	6472	76	1744	1744	7	2012-08-02 10:39:09.856738	\N
6772	1-6772	31	47	7069	24	1850	217	0	2012-08-02 05:27:59.175838	2012-08-02 11:12:31.699949
11889	1-11889	32	75	6492	70	3338	3338	11	2012-08-02 11:16:36.228318	\N
14016	1-14016	80	55	6448	68	168	168	2	2012-08-02 11:30:30.616645	\N
8201	1-8201	40	30	7097	12	145	145	0	2012-08-02 05:37:44.521935	\N
9372	1-9372	22	33	8743	80	6446	5000	7	2012-08-02 05:45:39.608217	2012-08-02 11:29:31.862809
7640	1-7640	29	1	8132	5	7251	2903	8	2012-08-02 05:33:51.535119	2012-08-02 11:27:46.491424
10971	1-10971	81	51	8799	64	2530	2530	0	2012-08-02 10:39:11.028542	\N
6776	1-6776	58	19	9228	97	3994	429	7	2012-08-02 05:28:00.744855	2012-08-02 11:26:11.325063
5640	1-5640	67	19	9435	85	6596	6596	5	2012-08-02 05:20:17.380259	\N
11901	1-11901	64	74	7370	58	5198	291	0	2012-08-02 11:16:40.813187	2012-08-02 11:29:15.220572
10974	1-10974	36	33	8472	60	9897	9897	6	2012-08-02 10:39:12.190416	\N
6773	1-6773	54	74	5302	2	1794	1794	2	2012-08-02 05:27:59.573965	\N
8203	1-8203	5	34	5082	35	444	444	0	2012-08-02 05:37:45.328586	\N
13950	1-13950	66	10	4378	49	6243	875	12	2012-08-02 11:30:05.056251	2012-08-02 11:30:33.358312
5089	1-5089	52	52	6232	84	376	376	0	2012-08-02 05:16:38.486653	\N
9118	1-9118	7	1	9692	47	5723	5265	5	2012-08-02 05:43:56.218513	2012-08-02 10:27:42.572796
8204	1-8204	84	92	8388	96	14	14	0	2012-08-02 05:37:45.737563	\N
5090	1-5090	79	24	3317	4	1532	1532	4	2012-08-02 05:16:38.895655	\N
4756	1-4756	50	24	7465	79	3315	3315	0	2012-08-02 05:14:25.405731	\N
10980	1-10980	90	18	9153	24	2101	2101	0	2012-08-02 10:39:14.52204	\N
4555	1-4555	98	93	8164	85	994	994	7	2012-08-02 05:13:03.604537	\N
7641	1-7641	14	41	6056	70	1101	1101	3	2012-08-02 05:33:51.943637	\N
4556	1-4556	75	30	8594	13	787	787	0	2012-08-02 05:13:03.991036	\N
4763	1-4763	12	42	9919	20	716	716	0	2012-08-02 05:14:28.212496	\N
12072	1-12072	37	18	9172	16	1321	1321	14	2012-08-02 11:17:48.019614	\N
9976	1-9976	42	81	4050	14	3832	487	5	2012-08-02 05:49:45.90393	2012-08-02 11:18:02.139928
4550	1-4550	13	39	3657	65	921	921	5	2012-08-02 05:13:01.63781	\N
4775	1-4775	11	99	9807	51	1948	1948	0	2012-08-02 05:14:32.974505	\N
4774	1-4774	55	35	9270	66	3418	3418	10	2012-08-02 05:14:32.576719	\N
4778	1-4778	82	99	2122	16	48	48	7	2012-08-02 05:14:34.2119	\N
14015	1-14015	30	74	9554	97	50	50	4	2012-08-02 11:30:31.025132	\N
5123	1-5123	36	25	4587	2	2272	1422	0	2012-08-02 05:16:52.341742	2012-08-02 10:38:40.984077
10641	1-10641	100	69	8365	50	2358	1973	0	2012-08-02 10:35:57.905968	2012-08-02 11:16:54.883231
7374	1-7374	91	6	5795	40	5619	2697	0	2012-08-02 05:32:02.405311	2012-08-02 11:17:52.549216
6587	1-6587	93	32	5457	93	1670	1579	7	2012-08-02 05:26:44.91876	2012-08-02 11:22:36.670154
4844	1-4844	74	65	5922	90	2565	193	10	2012-08-02 05:15:01.140074	2012-08-02 05:41:06.957549
5953	1-5953	60	34	8732	44	4078	786	7	2012-08-02 05:22:24.933286	2012-08-02 05:44:35.134786
4825	1-4825	70	36	6157	2	3745	3353	0	2012-08-02 05:14:53.350895	2012-08-02 05:29:25.845038
10322	1-10322	60	11	6714	77	6278	121	0	2012-08-02 10:25:46.133087	2012-08-02 11:27:07.097589
5947	1-5947	56	1	7326	41	3511	2329	0	2012-08-02 05:22:22.44739	2012-08-02 05:35:17.701653
5101	1-5101	41	19	8981	59	4470	4470	0	2012-08-02 05:16:43.447603	\N
5122	1-5122	17	67	6656	71	1408	445	10	2012-08-02 05:16:51.943982	2012-08-02 10:27:51.268635
12598	1-12598	9	75	9758	81	9993	9828	0	2012-08-02 11:21:11.521369	2012-08-02 11:23:53.941308
13531	1-13531	66	78	8258	45	9857	7036	0	2012-08-02 11:27:18.965869	2012-08-02 11:29:35.033431
8528	1-8528	100	67	9227	31	5772	1445	0	2012-08-02 05:39:56.327823	2012-08-02 11:28:51.13227
5351	1-5351	88	43	5046	90	1596	786	4	2012-08-02 05:18:22.331442	2012-08-02 05:33:40.454398
4798	1-4798	34	97	8571	18	967	967	3	2012-08-02 05:14:42.434557	\N
9136	1-9136	72	28	8761	52	5458	1058	0	2012-08-02 05:44:03.607604	2012-08-02 11:16:08.471378
10642	1-10642	95	67	3785	22	24	24	0	2012-08-02 10:35:58.303502	\N
8527	1-8527	32	31	8536	68	4324	1	4	2012-08-02 05:39:55.930186	2012-08-02 11:26:38.055142
10966	1-10966	23	36	2050	11	1387	1387	0	2012-08-02 10:39:09.459176	\N
10318	1-10318	3	54	4381	2	1590	1590	3	2012-08-02 10:25:44.575159	\N
10969	1-10969	27	31	9554	53	2506	2506	2	2012-08-02 10:39:10.243125	\N
4802	1-4802	58	78	9493	88	2726	2726	0	2012-08-02 05:14:44.036365	\N
12344	1-12344	19	20	9897	8	2939	2939	3	2012-08-02 11:19:32.690885	\N
5119	1-5119	96	86	6140	56	1790	834	0	2012-08-02 05:16:50.718307	2012-08-02 10:21:23.162067
4804	1-4804	62	71	6819	38	542	542	4	2012-08-02 05:14:44.853905	\N
14024	1-14024	1	37	6686	96	1643	1643	7	2012-08-02 11:30:34.154476	\N
6596	1-6596	73	1	5874	40	1631	1299	11	2012-08-02 05:26:48.664565	2012-08-02 05:28:13.164874
8529	1-8529	7	67	9650	70	9038	1345	0	2012-08-02 05:39:56.725514	2012-08-02 11:27:17.739084
11407	1-11407	70	66	3044	69	2626	1605	0	2012-08-02 11:13:29.214814	2012-08-02 11:24:50.381471
8797	1-8797	51	86	7267	85	2139	2139	0	2012-08-02 05:41:45.048057	\N
5954	1-5954	93	69	8430	40	3556	1322	4	2012-08-02 05:22:25.342093	2012-08-02 05:40:41.444129
9378	1-9378	94	51	5588	87	1327	1327	8	2012-08-02 05:45:42.027388	\N
4808	1-4808	14	93	7543	34	538	538	0	2012-08-02 05:14:46.456209	\N
7090	1-7090	6	36	1671	40	682	682	0	2012-08-02 05:30:07.603579	\N
6179	1-6179	62	36	6151	76	160	160	0	2012-08-02 05:23:56.998046	\N
5096	1-5096	16	61	5711	86	1547	1521	0	2012-08-02 05:16:41.414763	2012-08-02 05:32:51.218323
5095	1-5095	93	96	6253	15	2764	1577	4	2012-08-02 05:16:40.999608	2012-08-02 05:48:16.205043
6779	1-6779	59	76	9048	3	933	933	8	2012-08-02 05:28:01.938128	\N
9981	1-9981	74	67	7038	98	2005	2005	7	2012-08-02 05:49:47.959597	\N
9659	1-9659	39	71	9752	63	579	579	5	2012-08-02 05:47:36.458834	\N
5647	1-5647	47	11	9891	57	3322	3322	3	2012-08-02 05:20:20.208976	\N
6181	1-6181	20	2	6415	21	439	439	2	2012-08-02 05:23:57.804431	\N
13540	1-13540	65	73	4656	56	5016	917	0	2012-08-02 11:27:22.677663	2012-08-02 11:28:44.502577
5584	1-5584	48	61	6765	47	1904	1775	2	2012-08-02 05:19:54.716319	2012-08-02 05:20:21.04861
5939	1-5939	19	66	7267	71	2801	1745	0	2012-08-02 05:22:19.099543	2012-08-02 10:39:18.027505
9982	1-9982	59	73	9569	87	335	335	2	2012-08-02 05:49:48.379477	\N
12601	1-12601	85	66	9967	72	359	359	0	2012-08-02 11:21:12.714443	\N
7092	1-7092	86	6	5257	19	628	628	8	2012-08-02 05:30:08.42118	\N
12353	1-12353	66	71	1519	84	20	20	0	2012-08-02 11:19:36.171098	\N
6599	1-6599	40	87	4430	32	899	136	2	2012-08-02 05:26:49.890985	2012-08-02 10:21:25.962018
6601	1-6601	83	10	9689	73	667	667	8	2012-08-02 05:26:50.69755	\N
13543	1-13543	29	85	6029	54	1336	1336	13	2012-08-02 11:27:23.882261	\N
4828	1-4828	79	97	2244	88	424	424	0	2012-08-02 05:14:54.587729	\N
6782	1-6782	21	37	8718	43	932	932	0	2012-08-02 05:28:03.086987	\N
8533	1-8533	81	89	7926	20	3235	3235	0	2012-08-02 05:39:58.327431	\N
5941	1-5941	5	6	5164	31	1215	1154	0	2012-08-02 05:22:19.917039	2012-08-02 11:21:53.412361
7094	1-7094	34	90	8954	41	1646	1646	8	2012-08-02 05:30:09.260901	\N
13546	1-13546	48	24	4529	55	556	556	5	2012-08-02 11:27:25.064862	\N
7378	1-7378	62	59	6151	34	422	422	3	2012-08-02 05:32:03.974017	\N
8534	1-8534	37	19	6510	70	826	826	4	2012-08-02 05:39:58.724978	\N
4831	1-4831	80	47	9464	60	3444	3444	8	2012-08-02 05:14:55.80358	\N
7646	1-7646	33	75	4987	2	7758	1030	11	2012-08-02 05:33:54.0245	2012-08-02 11:26:28.960842
6594	1-6594	62	34	9530	46	1399	1399	4	2012-08-02 05:26:47.813709	\N
9986	1-9986	1	43	7404	72	1253	1253	8	2012-08-02 05:49:50.025393	\N
9134	1-9134	59	88	8248	84	691	691	3	2012-08-02 05:44:02.80044	\N
11420	1-11420	98	54	5968	45	2896	2896	7	2012-08-02 11:13:34.186285	\N
13548	1-13548	32	33	5867	61	2072	2072	10	2012-08-02 11:27:26.246789	\N
4835	1-4835	84	58	9081	76	408	408	8	2012-08-02 05:14:57.405277	\N
13550	1-13550	1	100	9598	51	8737	7552	0	2012-08-02 11:27:27.04239	2012-08-02 11:27:27.04239
13122	1-13122	10	96	9189	24	5257	5257	0	2012-08-02 11:24:36.691046	\N
13125	1-13125	41	96	5366	33	2047	2047	0	2012-08-02 11:24:37.837798	\N
5133	1-5133	99	26	7635	76	2535	2535	0	2012-08-02 05:16:56.398104	\N
4854	1-4854	40	18	9972	97	2112	2112	0	2012-08-02 05:15:05.206224	\N
11857	1-11857	86	81	5963	85	6224	1025	0	2012-08-02 11:16:23.884883	2012-08-02 11:27:08.65467
12606	1-12606	6	75	6481	88	4009	3854	4	2012-08-02 11:21:14.660859	2012-08-02 11:23:20.912193
12346	1-12346	10	40	9415	60	2672	2672	0	2012-08-02 11:19:33.4531	\N
8804	1-8804	47	11	7560	85	8631	6858	6	2012-08-02 05:41:47.876304	2012-08-02 11:27:08.65467
10973	1-10973	62	53	3975	100	915	494	6	2012-08-02 10:39:11.80376	2012-08-02 11:12:37.942975
11406	1-11406	73	43	7745	81	5199	3248	5	2012-08-02 11:13:28.817168	2012-08-02 11:29:15.596151
13532	1-13532	37	1	5742	52	4145	4145	0	2012-08-02 11:27:19.761632	\N
6188	1-6188	46	85	5425	53	401	401	2	2012-08-02 05:24:00.799848	\N
6494	1-6494	18	58	8577	85	3749	2170	0	2012-08-02 05:26:06.630238	2012-08-02 11:30:32.970668
11416	1-11416	10	57	6702	16	3403	506	2	2012-08-02 11:13:32.683864	2012-08-02 11:29:18.780703
5148	1-5148	24	83	2335	85	236	236	0	2012-08-02 05:17:02.541123	\N
5150	1-5150	23	47	8398	24	1879	1879	0	2012-08-02 05:17:03.347432	\N
12367	1-12367	14	50	7138	84	4853	2982	0	2012-08-02 11:19:41.529203	2012-08-02 11:24:12.922075
12352	1-12352	35	38	4174	76	470	470	2	2012-08-02 11:19:35.784246	\N
7098	1-7098	25	67	8510	93	499	499	4	2012-08-02 05:30:10.918239	\N
11890	1-11890	61	32	8939	90	5693	3857	8	2012-08-02 11:16:36.615197	2012-08-02 11:22:28.150301
5958	1-5958	70	8	5013	54	591	591	13	2012-08-02 05:22:27.034206	\N
12597	1-12597	34	82	7451	86	15	15	0	2012-08-02 11:21:11.134525	\N
13535	1-13535	52	26	7607	22	6552	6552	0	2012-08-02 11:27:20.556522	\N
5358	1-5358	92	70	6256	17	1226	1226	5	2012-08-02 05:18:25.060355	\N
9996	1-9996	37	86	7743	44	4318	1608	4	2012-08-02 05:49:54.158111	2012-08-02 11:29:34.635924
5149	1-5149	99	45	6628	83	1143	654	0	2012-08-02 05:17:02.949939	2012-08-02 05:37:32.246722
5361	1-5361	97	4	7187	91	262	262	0	2012-08-02 05:18:26.231416	\N
14028	1-14028	77	75	8356	35	9401	9401	2	2012-08-02 11:30:36.103653	\N
5960	1-5960	86	33	3877	79	210	210	0	2012-08-02 05:22:27.851733	\N
5360	1-5360	40	19	3966	39	1097	947	13	2012-08-02 05:18:25.833663	2012-08-02 05:39:42.327512
11893	1-11893	20	67	4227	14	5238	1179	4	2012-08-02 11:16:37.785926	2012-08-02 11:24:38.611274
9988	1-9988	74	98	9930	54	1784	1784	6	2012-08-02 05:49:50.876405	\N
5420	1-5420	88	21	6148	100	822	822	0	2012-08-02 05:18:49.384012	\N
7940	1-7940	30	75	6169	71	1609	1609	0	2012-08-02 05:35:58.338116	\N
5974	1-5974	79	42	7774	17	2663	7	0	2012-08-02 05:22:33.619244	2012-08-02 11:14:20.825378
11419	1-11419	81	1	8761	50	9048	5828	4	2012-08-02 11:13:33.810674	2012-08-02 11:24:40.931417
10976	1-10976	95	74	5975	10	533	533	0	2012-08-02 10:39:12.942068	\N
6496	1-6496	65	59	7258	11	338	338	7	2012-08-02 05:26:07.458734	\N
14026	1-14026	25	85	4382	100	6527	118	0	2012-08-02 11:30:34.543923	2012-08-02 11:30:34.543923
12364	1-12364	36	42	9013	47	1701	1701	2	2012-08-02 11:19:40.347453	\N
5671	1-5671	32	45	6723	88	227	227	0	2012-08-02 05:20:29.843184	\N
6497	1-6497	70	28	7677	98	771	771	0	2012-08-02 05:26:07.856443	\N
11896	1-11896	16	35	9484	15	8805	8805	3	2012-08-02 11:16:38.912671	\N
9139	1-9139	28	94	8663	17	4374	4374	2	2012-08-02 05:44:04.85613	\N
5355	1-5355	39	33	7149	54	5459	1340	0	2012-08-02 05:18:23.900457	2012-08-02 05:46:25.839997
6493	1-6493	22	65	9997	51	5648	400	0	2012-08-02 05:26:06.221666	2012-08-02 11:11:43.819221
7096	1-7096	66	40	4475	74	271	147	0	2012-08-02 05:30:10.078338	2012-08-02 11:11:22.324789
14032	1-14032	18	98	4173	37	7777	7777	7	2012-08-02 11:30:37.693893	\N
11409	1-11409	99	14	9665	1	1313	1313	0	2012-08-02 11:13:29.977138	\N
8549	1-8549	69	51	4569	37	2272	625	5	2012-08-02 05:40:04.735499	2012-08-02 11:17:40.198824
12603	1-12603	21	11	4403	99	4100	2626	5	2012-08-02 11:21:13.500909	2012-08-02 11:22:39.388621
7651	1-7651	76	82	6618	38	79	79	11	2012-08-02 05:33:56.172873	\N
8212	1-8212	1	59	4271	16	1192	1192	0	2012-08-02 05:37:48.974429	\N
9994	1-9994	60	81	6638	12	82	82	10	2012-08-02 05:49:53.351967	\N
14033	1-14033	43	53	2663	5	6694	5419	3	2012-08-02 11:30:38.533307	2012-08-02 11:30:38.533307
9383	1-9383	4	63	8940	40	774	774	4	2012-08-02 05:45:44.027176	\N
14034	1-14034	91	41	7560	34	8031	8031	4	2012-08-02 11:30:38.920121	\N
5357	1-5357	83	60	9831	9	1724	720	4	2012-08-02 05:18:24.673766	2012-08-02 05:47:54.976469
8805	1-8805	75	69	3421	54	2955	2955	0	2012-08-02 05:41:48.273899	\N
6917	1-6917	74	94	4553	42	1515	1515	0	2012-08-02 05:28:58.47838	\N
5969	1-5969	6	91	9054	48	2062	2062	12	2012-08-02 05:22:31.552902	\N
6194	1-6194	58	43	8402	38	1026	1026	7	2012-08-02 05:24:03.179153	\N
11413	1-11413	4	41	5340	61	2312	2312	2	2012-08-02 11:13:31.137048	\N
5970	1-5970	90	24	9501	43	880	880	7	2012-08-02 05:22:31.961551	\N
7947	1-7947	28	20	4511	62	860	860	0	2012-08-02 05:36:01.376434	\N
5688	1-5688	59	8	5397	84	640	640	0	2012-08-02 05:20:36.785056	\N
7654	1-7654	24	99	7508	59	696	696	0	2012-08-02 05:33:57.366737	\N
6198	1-6198	59	84	3844	98	630	630	0	2012-08-02 05:24:04.769985	\N
5147	1-5147	7	50	6827	42	2259	908	2	2012-08-02 05:17:02.143502	2012-08-02 05:22:32.790361
8808	1-8808	51	18	7180	83	2784	2784	0	2012-08-02 05:41:49.445116	\N
5689	1-5689	66	8	6808	46	1852	1852	0	2012-08-02 05:20:37.182607	\N
5675	1-5675	1	14	7867	42	843	710	0	2012-08-02 05:20:31.522568	2012-08-02 05:30:20.02189
5392	1-5392	7	97	6206	22	687	687	4	2012-08-02 05:18:38.406905	\N
5691	1-5691	90	33	9482	29	367	367	3	2012-08-02 05:20:37.966678	\N
9279	1-9279	10	100	5377	1	598	598	13	2012-08-02 05:45:01.740158	\N
5399	1-5399	29	8	6273	54	3995	3995	13	2012-08-02 05:18:41.170734	\N
5413	1-5413	32	69	5083	92	1724	1724	0	2012-08-02 05:18:46.639623	\N
5410	1-5410	69	42	8519	70	1025	1025	4	2012-08-02 05:18:45.446349	\N
5422	1-5422	77	56	9726	54	3917	3917	5	2012-08-02 05:18:50.160137	\N
5426	1-5426	35	65	5847	40	1198	1198	0	2012-08-02 05:18:51.74554	\N
10142	1-10142	41	38	8416	83	2624	2624	0	2012-08-02 10:22:00.756025	\N
8816	1-8816	66	61	9115	67	867	783	0	2012-08-02 05:41:52.69503	2012-08-02 05:47:57.851532
5987	1-5987	44	55	9241	81	5239	2148	0	2012-08-02 05:22:38.999941	2012-08-02 11:11:16.030855
13549	1-13549	40	96	7576	79	8639	7268	8	2012-08-02 11:27:25.45104	2012-08-02 11:28:00.831823
7467	1-7467	77	33	5485	75	1347	864	0	2012-08-02 05:32:40.556897	2012-08-02 11:30:31.832093
6281	1-6281	56	78	8334	5	5240	4935	0	2012-08-02 05:24:39.844277	2012-08-02 11:13:30.374755
5990	1-5990	94	81	9947	14	6851	5178	4	2012-08-02 05:22:40.248724	2012-08-02 05:23:59.429971
10003	1-10003	15	24	6335	51	1290	1290	0	2012-08-02 09:10:30.51683	\N
7449	1-7449	9	62	9590	60	9294	123	4	2012-08-02 05:32:33.221053	2012-08-02 11:28:53.543122
5741	1-5741	46	41	1888	20	844	351	0	2012-08-02 05:20:58.494938	2012-08-02 05:31:50.581363
5716	1-5716	1	60	2338	58	246	246	9	2012-08-02 05:20:47.987975	\N
9673	1-9673	83	38	5196	81	2490	1523	6	2012-08-02 05:47:42.193177	2012-08-02 11:29:20.4272
10007	1-10007	24	79	6834	96	2020	682	0	2012-08-02 09:10:32.227115	2012-08-02 11:12:44.972704
10325	1-10325	53	62	1743	89	231	231	5	2012-08-02 10:25:47.304199	\N
8551	1-8551	36	24	7048	71	1926	1926	0	2012-08-02 05:40:05.5435	\N
10146	1-10146	99	91	8352	98	3016	2384	11	2012-08-02 10:22:02.346049	2012-08-02 11:13:21.30417
5696	1-5696	41	78	1435	34	72	72	0	2012-08-02 05:20:39.978166	\N
10001	1-10001	12	9	9008	97	1140	1140	0	2012-08-02 09:10:29.666242	\N
10649	1-10649	69	63	9010	90	4129	4129	0	2012-08-02 10:36:01.125962	\N
13544	1-13544	85	64	4014	59	2294	1018	7	2012-08-02 11:27:24.268633	2012-08-02 11:27:50.847531
10650	1-10650	13	90	7528	10	584	584	4	2012-08-02 10:36:01.512492	\N
7659	1-7659	51	49	5648	67	1374	1365	0	2012-08-02 05:33:59.379777	2012-08-02 11:19:25.352516
8809	1-8809	83	28	4702	40	4083	2727	0	2012-08-02 05:41:49.853689	2012-08-02 11:13:22.859494
5739	1-5739	94	81	9249	58	1889	1624	0	2012-08-02 05:20:57.666732	2012-08-02 05:45:27.332705
5743	1-5743	85	50	7220	22	969	960	6	2012-08-02 05:20:59.312416	2012-08-02 11:13:26.705297
9389	1-9389	31	25	9298	23	1818	1646	0	2012-08-02 05:45:46.47083	2012-08-02 10:27:43.777161
11692	1-11692	42	59	5622	85	8658	3244	9	2012-08-02 11:15:19.432231	2012-08-02 11:29:43.786822
7458	1-7458	49	82	4436	27	1007	344	0	2012-08-02 05:32:36.888958	2012-08-02 05:40:51.111975
10330	1-10330	1	19	9872	41	9031	380	3	2012-08-02 10:25:49.238173	2012-08-02 11:26:05.046846
11422	1-11422	91	12	7696	87	2270	2270	3	2012-08-02 11:13:34.948599	\N
8214	1-8214	45	78	8793	99	415	415	0	2012-08-02 05:37:49.792393	\N
11719	1-11719	96	62	5253	88	2501	1099	3	2012-08-02 11:15:29.970833	2012-08-02 11:16:54.120655
5699	1-5699	14	97	3983	7	695	695	6	2012-08-02 05:20:41.248845	\N
5980	1-5980	31	43	9507	51	471	471	7	2012-08-02 05:22:36.105368	\N
10329	1-10329	66	23	8433	21	1543	1543	8	2012-08-02 10:25:48.840049	\N
8813	1-8813	65	63	7972	52	8578	3189	6	2012-08-02 05:41:51.479765	2012-08-02 11:26:47.220183
5701	1-5701	67	51	9084	14	1705	1705	0	2012-08-02 05:20:42.043979	\N
14027	1-14027	16	31	8518	85	9997	9997	5	2012-08-02 11:30:35.713178	\N
6789	1-6789	35	73	5341	33	653	653	6	2012-08-02 05:28:05.828276	\N
10008	1-10008	66	17	9777	47	405	405	8	2012-08-02 09:10:32.636397	\N
11894	1-11894	99	47	6052	10	1757	1757	4	2012-08-02 11:16:38.172599	\N
6503	1-6503	82	52	5517	11	1201	405	14	2012-08-02 05:26:10.331596	2012-08-02 11:27:08.65467
11431	1-11431	25	23	9766	35	2502	2502	2	2012-08-02 11:13:38.440217	\N
10148	1-10148	53	31	7888	51	3929	445	3	2012-08-02 10:22:03.119199	2012-08-02 10:39:15.683106
14029	1-14029	95	49	8027	60	7341	7341	9	2012-08-02 11:30:36.489356	\N
13541	1-13541	83	11	9577	4	49	49	0	2012-08-02 11:27:23.092333	\N
14031	1-14031	53	50	5493	10	6604	6604	4	2012-08-02 11:30:37.29615	\N
6507	1-6507	46	78	6251	70	1003	1003	0	2012-08-02 05:26:11.977724	\N
7957	1-7957	98	41	8304	86	2247	2247	0	2012-08-02 05:36:05.589077	\N
5986	1-5986	50	3	8296	73	318	318	4	2012-08-02 05:22:38.580301	\N
8819	1-8819	24	18	5990	39	2754	1007	0	2012-08-02 05:41:53.921069	2012-08-02 11:20:04.336464
6283	1-6283	89	58	5751	61	476	476	0	2012-08-02 05:24:40.684162	\N
12368	1-12368	21	47	8273	31	977	977	3	2012-08-02 11:19:41.905078	\N
8696	1-8696	41	42	7585	72	402	402	15	2012-08-02 05:41:04.137475	\N
5709	1-5709	7	1	8746	51	929	929	0	2012-08-02 05:20:45.203461	\N
11440	1-11440	35	58	7987	69	1752	1752	13	2012-08-02 11:13:41.953669	\N
13552	1-13552	33	85	6975	81	1710	1710	0	2012-08-02 11:27:27.451162	\N
12611	1-12611	68	12	6660	95	3030	3030	10	2012-08-02 11:21:16.616331	\N
11698	1-11698	29	16	3548	81	259	259	0	2012-08-02 11:15:21.741662	\N
9675	1-9675	47	38	2323	28	414	414	5	2012-08-02 05:47:43.021892	\N
8693	1-8693	21	11	6703	5	8745	1899	0	2012-08-02 05:41:02.944344	2012-08-02 11:28:30.535322
9395	1-9395	98	95	8314	1	1494	1494	0	2012-08-02 05:45:48.868637	\N
5747	1-5747	3	27	3308	19	5	5	8	2012-08-02 05:21:01.014006	\N
9677	1-9677	91	21	7252	40	1981	1981	4	2012-08-02 05:47:43.87252	\N
7118	1-7118	94	1	2806	2	391	391	3	2012-08-02 05:30:19.226854	\N
10149	1-10149	58	22	7313	20	420	420	0	2012-08-02 10:22:03.505813	\N
5997	1-5997	41	11	9386	71	1599	1599	7	2012-08-02 05:22:43.198467	\N
7119	1-7119	100	52	8581	92	1446	1446	8	2012-08-02 05:30:19.635278	\N
10152	1-10152	45	13	4173	73	833	833	4	2012-08-02 10:22:04.688266	\N
5725	1-5725	2	37	9645	54	1778	1778	11	2012-08-02 05:20:51.656391	\N
6004	1-6004	14	75	9676	48	3506	3506	0	2012-08-02 05:22:46.115955	\N
7468	1-7468	78	47	7833	28	398	398	11	2012-08-02 05:32:40.976851	\N
7470	1-7470	17	74	5564	89	1527	1527	0	2012-08-02 05:32:41.827349	\N
5734	1-5734	75	94	9250	72	2559	2559	3	2012-08-02 05:20:55.589937	\N
5748	1-5748	72	8	7482	14	550	550	0	2012-08-02 05:21:01.422506	\N
6219	1-6219	7	68	3370	31	1509	1509	9	2012-08-02 05:24:13.575986	\N
7396	1-7396	84	70	7366	66	986	519	9	2012-08-02 05:32:11.277634	2012-08-02 05:33:30.23434
10159	1-10159	93	100	8422	52	9031	1188	4	2012-08-02 10:22:07.417233	2012-08-02 11:12:49.681048
10988	1-10988	38	71	5971	70	1904	1400	5	2012-08-02 10:39:17.640515	2012-08-02 11:23:45.333175
6802	1-6802	74	63	3516	13	1595	665	0	2012-08-02 05:28:11.143163	2012-08-02 05:38:08.492792
8823	1-8823	69	88	9552	23	3253	244	0	2012-08-02 05:41:55.556674	2012-08-02 10:36:20.834054
6206	1-6206	67	6	9996	68	3375	1042	3	2012-08-02 05:24:08.18232	2012-08-02 11:29:54.404978
10019	1-10019	17	30	5742	21	4218	1986	11	2012-08-02 10:16:59.050702	2012-08-02 11:19:00.160668
13111	1-13111	97	32	9997	84	916	916	0	2012-08-02 11:24:31.905157	\N
13120	1-13120	92	13	8580	89	7774	4829	0	2012-08-02 11:24:35.451938	2012-08-02 11:25:28.160499
11900	1-11900	77	59	4305	85	3200	3200	9	2012-08-02 11:16:40.437603	\N
13539	1-13539	16	100	9602	25	1938	1938	0	2012-08-02 11:27:22.289181	\N
6040	1-6040	68	41	1213	71	508	508	2	2012-08-02 05:23:01.209744	\N
12610	1-12610	87	97	8464	72	8004	5996	8	2012-08-02 11:21:16.208015	2012-08-02 11:30:02.736062
9401	1-9401	97	28	6303	71	4576	4285	0	2012-08-02 05:45:51.277093	2012-08-02 10:25:31.970379
12607	1-12607	27	55	4739	62	2240	2240	5	2012-08-02 11:21:15.047524	\N
8552	1-8552	30	87	4154	86	314	314	4	2012-08-02 05:40:05.952677	\N
10332	1-10332	55	33	8932	87	7077	7077	0	2012-08-02 10:25:50.035108	\N
8223	1-8223	70	19	6187	73	2302	241	0	2012-08-02 05:37:53.438123	2012-08-02 11:16:43.499661
7132	1-7132	73	49	1890	6	533	533	0	2012-08-02 05:30:25.104489	\N
14035	1-14035	86	26	1283	76	9256	3711	2	2012-08-02 11:30:39.30669	2012-08-02 11:30:39.30669
6221	1-6221	44	55	5732	9	77	77	3	2012-08-02 05:24:14.428805	\N
7127	1-7127	13	65	4560	37	3082	2901	0	2012-08-02 05:30:22.950094	2012-08-02 05:36:31.949891
13114	1-13114	8	35	5048	15	5929	5929	4	2012-08-02 11:24:33.098335	\N
10333	1-10333	27	86	7622	16	1415	1415	5	2012-08-02 10:25:50.433009	\N
6803	1-6803	37	99	8147	64	3616	3390	0	2012-08-02 05:28:11.551635	2012-08-02 11:12:32.882289
8553	1-8553	84	1	4169	72	1283	1283	9	2012-08-02 05:40:06.350628	\N
8222	1-8222	42	75	8630	72	6222	6222	0	2012-08-02 05:37:53.029659	\N
9679	1-9679	59	79	4660	17	299	299	3	2012-08-02 05:47:44.668002	\N
6048	1-6048	25	94	8414	73	569	569	3	2012-08-02 05:23:04.524382	\N
6049	1-6049	53	61	6545	48	3836	3836	0	2012-08-02 05:23:04.933229	\N
10334	1-10334	24	39	9977	15	1292	1292	0	2012-08-02 10:25:50.819499	\N
6041	1-6041	51	24	7700	37	6817	3531	2	2012-08-02 05:23:01.629328	2012-08-02 05:27:38.203127
12366	1-12366	56	14	7765	82	3454	2458	5	2012-08-02 11:19:41.142625	2012-08-02 11:23:37.112039
6208	1-6208	87	41	8732	33	2285	2285	0	2012-08-02 05:24:08.999869	\N
8554	1-8554	5	98	6123	66	214	214	0	2012-08-02 05:40:06.759424	\N
7383	1-7383	90	58	9492	20	2664	2075	0	2012-08-02 05:32:05.930687	2012-08-02 11:28:30.922001
10335	1-10335	77	43	9617	91	3722	3722	0	2012-08-02 10:25:51.217289	\N
13117	1-13117	76	25	9695	8	459	459	5	2012-08-02 11:24:34.269314	\N
10168	1-10168	94	56	5826	80	3875	2124	0	2012-08-02 10:22:10.953419	2012-08-02 11:16:49.059592
7386	1-7386	66	93	3633	62	462	462	0	2012-08-02 05:32:07.156983	\N
6213	1-6213	77	61	5544	95	918	918	0	2012-08-02 05:24:11.066425	\N
13108	1-13108	31	27	7078	69	8370	1676	4	2012-08-02 11:24:31.131624	2012-08-02 11:28:35.983455
8226	1-8226	70	75	9926	36	3520	1900	0	2012-08-02 05:37:54.653374	2012-08-02 10:38:23.722511
7125	1-7125	15	98	6304	60	2378	2378	0	2012-08-02 05:30:22.110303	\N
12620	1-12620	20	99	8559	20	8319	2187	0	2012-08-02 11:21:20.141439	2012-08-02 11:28:58.338996
7128	1-7128	13	9	7750	37	4704	4704	4	2012-08-02 05:30:23.358689	\N
13545	1-13545	51	42	9484	8	2684	2684	0	2012-08-02 11:27:24.666922	\N
7663	1-7663	5	88	2702	38	245	245	0	2012-08-02 05:34:01.026671	\N
10993	1-10993	8	67	9931	45	817	817	2	2012-08-02 10:39:20.404593	\N
11432	1-11432	94	90	6125	31	60	60	9	2012-08-02 11:13:38.815558	\N
14038	1-14038	5	61	7742	36	9641	8632	0	2012-08-02 11:30:40.501224	2012-08-02 11:30:40.501224
12623	1-12623	54	72	7737	78	855	855	3	2012-08-02 11:21:21.324235	\N
11426	1-11426	26	10	4265	56	1973	1881	0	2012-08-02 11:13:36.484463	2012-08-02 11:28:44.502577
10161	1-10161	50	85	6702	34	1751	1147	0	2012-08-02 10:22:08.190967	2012-08-02 11:29:15.971766
10017	1-10017	3	75	8575	49	6198	349	3	2012-08-02 10:16:58.254472	2012-08-02 11:14:13.157551
13547	1-13547	45	46	6798	15	1670	1670	6	2012-08-02 11:27:25.860219	\N
12626	1-12626	53	34	9771	35	1284	1284	3	2012-08-02 11:21:22.506703	\N
11442	1-11442	44	22	7165	38	1194	1194	0	2012-08-02 11:13:42.351302	\N
13551	1-13551	30	32	6781	88	6008	5944	6	2012-08-02 11:27:26.633186	2012-08-02 11:27:26.633186
14039	1-14039	58	4	3668	67	6861	6861	11	2012-08-02 11:30:40.910529	\N
14040	1-14040	49	65	7979	99	246	246	9	2012-08-02 11:30:41.308491	\N
12635	1-12635	82	67	2926	5	1370	1370	0	2012-08-02 11:21:25.975391	\N
12641	1-12641	38	11	7436	23	2695	2695	7	2012-08-02 11:21:28.306696	\N
6534	1-6534	84	56	8767	14	24	24	2	2012-08-02 05:26:23.172378	\N
11720	1-11720	18	88	6499	2	514	514	8	2012-08-02 11:15:30.379789	\N
10156	1-10156	76	30	8262	57	5199	5199	6	2012-08-02 10:22:06.246061	\N
7673	1-7673	62	3	4486	88	415	415	0	2012-08-02 05:34:05.083431	\N
7674	1-7674	59	82	4924	55	680	680	0	2012-08-02 05:34:05.509404	\N
6245	1-6245	79	26	9343	4	612	612	0	2012-08-02 05:24:24.452516	\N
6256	1-6256	42	25	9139	78	828	828	0	2012-08-02 05:24:29.060357	\N
6260	1-6260	28	100	2242	97	178	178	4	2012-08-02 05:24:30.728123	\N
6264	1-6264	30	67	6508	55	287	287	0	2012-08-02 05:24:32.385007	\N
6269	1-6269	67	83	8866	46	11	11	15	2012-08-02 05:24:34.462312	\N
6275	1-6275	54	64	8390	65	566	566	11	2012-08-02 05:24:37.281014	\N
6816	1-6816	6	9	8089	98	3010	1945	0	2012-08-02 05:28:16.845963	2012-08-02 10:37:50.751828
6314	1-6314	77	90	6738	6	690	690	0	2012-08-02 05:24:53.567164	\N
6814	1-6814	81	4	7156	92	5361	1485	2	2012-08-02 05:28:16.039007	2012-08-02 05:37:02.102582
6813	1-6813	45	30	8777	56	1871	1871	0	2012-08-02 05:28:15.630432	\N
11439	1-11439	27	42	9364	38	4447	4042	0	2012-08-02 11:13:41.549222	2012-08-02 11:29:54.029273
8835	1-8835	85	96	8664	64	3550	1504	10	2012-08-02 05:42:00.329093	2012-08-02 05:43:44.610137
6832	1-6832	86	73	7994	53	2049	2049	8	2012-08-02 05:28:23.420131	\N
7686	1-7686	55	94	4180	57	1294	298	6	2012-08-02 05:34:10.538636	2012-08-02 05:37:49.372078
6330	1-6330	6	68	8279	65	1557	1557	8	2012-08-02 05:25:00.273878	\N
8836	1-8836	90	94	8369	17	4662	1354	0	2012-08-02 05:42:00.749119	2012-08-02 11:23:55.476965
6566	1-6566	98	8	3173	73	887	563	0	2012-08-02 05:26:36.353957	2012-08-02 11:28:44.502577
6290	1-6290	50	78	7421	54	880	880	4	2012-08-02 05:24:43.623947	\N
9154	1-9154	89	62	8546	42	1610	305	10	2012-08-02 05:44:10.911223	2012-08-02 11:16:28.924881
6291	1-6291	68	43	2726	15	1061	1061	0	2012-08-02 05:24:44.043666	\N
7676	1-7676	33	90	6199	85	2943	2323	2	2012-08-02 05:34:06.316537	2012-08-02 11:24:58.314903
7400	1-7400	56	50	8745	1	1573	559	0	2012-08-02 05:32:12.968258	2012-08-02 11:24:12.922075
6582	1-6582	77	33	5029	58	2826	2216	9	2012-08-02 05:26:42.863455	2012-08-02 05:47:45.087852
8236	1-8236	53	59	9338	71	5198	364	0	2012-08-02 05:37:58.688215	2012-08-02 11:16:30.07419
9153	1-9153	41	97	7028	37	5743	3459	5	2012-08-02 05:44:10.513241	2012-08-02 10:38:19.391454
8239	1-8239	79	87	5481	37	2488	1067	0	2012-08-02 05:37:59.907373	2012-08-02 05:41:27.613705
12377	1-12377	38	13	6366	83	6987	3125	0	2012-08-02 11:19:45.408081	2012-08-02 11:24:19.144836
12371	1-12371	66	33	4808	44	665	665	2	2012-08-02 11:19:42.656213	\N
6296	1-6296	36	98	5401	40	438	438	3	2012-08-02 05:24:46.120857	\N
13556	1-13556	80	56	6060	64	981	981	0	2012-08-02 11:27:28.997491	\N
10339	1-10339	62	77	8454	87	4881	3430	7	2012-08-02 10:25:52.741684	2012-08-02 11:22:44.903286
6537	1-6537	85	3	9172	90	4177	4177	3	2012-08-02 05:26:24.420661	\N
6826	1-6826	27	35	9050	7	454	454	4	2012-08-02 05:28:20.978533	\N
7965	1-7965	9	63	2030	91	73	73	9	2012-08-02 05:36:08.837466	\N
6299	1-6299	7	16	4900	92	2306	2306	2	2012-08-02 05:24:47.347048	\N
6827	1-6827	57	34	8039	25	907	731	4	2012-08-02 05:28:21.375952	2012-08-02 05:48:25.89676
13560	1-13560	77	50	5031	79	994	994	0	2012-08-02 11:27:30.157645	\N
9152	1-9152	89	1	7293	31	1717	1717	3	2012-08-02 05:44:10.104616	\N
6301	1-6301	97	77	6492	26	33	33	10	2012-08-02 05:24:48.164665	\N
6540	1-6540	67	89	8257	65	2515	2515	0	2012-08-02 05:26:25.658197	\N
6306	1-6306	100	100	7495	85	2347	670	0	2012-08-02 05:24:50.219814	2012-08-02 05:39:52.692544
7142	1-7142	56	6	8155	92	6483	3161	0	2012-08-02 05:30:29.282657	2012-08-02 05:34:35.215414
6812	1-6812	73	59	5853	49	386	386	0	2012-08-02 05:28:15.20901	\N
10338	1-10338	11	93	4245	34	1328	1328	5	2012-08-02 10:25:52.366211	\N
7681	1-7681	65	96	6899	50	66	66	8	2012-08-02 05:34:08.427731	\N
6307	1-6307	27	63	8478	100	1190	1190	0	2012-08-02 05:24:50.650724	\N
6308	1-6308	82	68	8781	34	1239	1239	0	2012-08-02 05:24:51.070277	\N
13121	1-13121	21	6	7575	77	2011	2011	21	2012-08-02 11:24:35.849874	\N
9404	1-9404	14	77	9712	62	1798	1798	2	2012-08-02 05:45:52.492168	\N
8833	1-8833	97	72	8806	15	753	753	0	2012-08-02 05:41:59.522782	\N
7147	1-7147	63	94	8725	51	1563	1563	6	2012-08-02 05:30:31.326164	\N
10994	1-10994	4	59	8234	92	1571	1571	0	2012-08-02 10:39:19.608463	\N
9683	1-9683	17	48	6797	15	5136	5136	12	2012-08-02 05:47:46.336717	\N
9405	1-9405	42	55	5920	60	4552	4552	4	2012-08-02 05:45:52.890176	\N
8834	1-8834	94	75	7169	17	2406	2406	4	2012-08-02 05:41:59.920548	\N
8560	1-8560	33	57	7942	14	652	652	7	2012-08-02 05:40:09.190377	\N
10025	1-10025	89	76	7401	86	1715	1715	2	2012-08-02 10:19:32.214408	\N
6822	1-6822	77	42	5665	54	1141	1141	6	2012-08-02 05:28:19.353893	\N
11441	1-11441	46	28	5397	54	4524	4524	7	2012-08-02 11:13:42.72661	\N
6289	1-6289	8	75	5090	29	1212	915	0	2012-08-02 05:24:43.20399	2012-08-02 05:48:45.809003
7972	1-7972	89	23	4542	2	3656	3656	5	2012-08-02 05:36:11.687779	\N
6324	1-6324	7	74	9722	2	1138	1138	2	2012-08-02 05:24:57.776885	\N
7688	1-7688	68	95	6663	58	123	123	0	2012-08-02 05:34:11.389423	\N
13126	1-13126	30	68	9156	78	925	925	26	2012-08-02 11:24:38.224478	\N
12060	1-12060	43	51	9145	35	2396	2396	5	2012-08-02 11:17:43.39032	\N
10026	1-10026	67	8	8376	23	700	700	0	2012-08-02 10:19:32.601135	\N
6559	1-6559	96	80	4652	82	288	288	3	2012-08-02 05:26:33.49248	\N
12063	1-12063	51	29	6744	38	1353	1353	0	2012-08-02 11:17:44.539768	\N
6556	1-6556	37	67	9747	79	3776	3776	0	2012-08-02 05:26:32.254927	\N
9407	1-9407	8	67	9038	55	4158	1510	0	2012-08-02 05:45:53.707618	2012-08-02 11:15:31.131114
12065	1-12065	84	68	9964	85	6072	6072	5	2012-08-02 11:17:45.710845	\N
7675	1-7675	5	49	7738	92	6930	2444	3	2012-08-02 05:34:05.907717	2012-08-02 11:26:10.938883
6561	1-6561	52	91	9656	59	651	651	0	2012-08-02 05:26:34.276872	\N
6570	1-6570	88	28	4803	69	202	202	0	2012-08-02 05:26:37.956537	\N
6868	1-6868	48	75	7847	4	3288	2725	0	2012-08-02 05:28:38.159516	2012-08-02 05:33:07.009623
12076	1-12076	13	64	8253	62	5091	4484	0	2012-08-02 11:17:49.500088	2012-08-02 11:20:09.364057
11461	1-11461	35	47	5990	61	3956	3842	10	2012-08-02 11:13:49.986148	2012-08-02 11:20:01.662484
7713	1-7713	51	83	5877	85	3029	173	0	2012-08-02 05:34:21.966632	2012-08-02 11:29:55.200678
8232	1-8232	62	4	5217	32	2208	951	10	2012-08-02 05:37:57.07331	2012-08-02 11:23:24.171569
13300	1-13300	12	27	5420	60	6954	4769	14	2012-08-02 11:25:46.782297	2012-08-02 11:26:48.433006
9412	1-9412	55	96	7439	79	3226	2476	5	2012-08-02 05:45:55.729539	2012-08-02 10:27:31.057118
7703	1-7703	62	57	6022	54	2066	1203	6	2012-08-02 05:34:17.71295	2012-08-02 11:26:55.649625
6838	1-6838	79	72	6531	66	1400	87	0	2012-08-02 05:28:25.884286	2012-08-02 11:20:53.46267
12615	1-12615	73	16	6521	54	3500	3500	5	2012-08-02 11:21:18.196746	\N
7701	1-7701	95	30	3499	40	2766	1278	0	2012-08-02 05:34:16.884465	2012-08-02 05:43:37.991532
7165	1-7165	79	62	8103	10	1844	1844	0	2012-08-02 05:30:38.783941	\N
12384	1-12384	65	17	9400	33	2076	1416	0	2012-08-02 11:19:48.10381	2012-08-02 11:20:03.165061
7709	1-7709	36	76	5101	98	1338	581	3	2012-08-02 05:34:20.287534	2012-08-02 05:43:13.795384
8245	1-8245	5	61	5311	64	2268	91	3	2012-08-02 05:38:02.337928	2012-08-02 11:14:55.047636
8287	1-8287	81	93	7416	83	3381	2636	0	2012-08-02 05:38:19.375437	2012-08-02 05:47:40.955724
6846	1-6846	100	98	9122	83	2110	2110	0	2012-08-02 05:28:29.20987	\N
6859	1-6859	57	89	8929	94	3748	1799	0	2012-08-02 05:28:34.523902	2012-08-02 11:29:25.178304
6889	1-6889	17	83	9970	40	7185	108	0	2012-08-02 05:28:46.922675	2012-08-02 11:24:28.004933
10030	1-10030	54	72	4459	91	1788	603	0	2012-08-02 10:19:34.136914	2012-08-02 11:26:58.843133
10345	1-10345	56	35	9760	66	9595	630	10	2012-08-02 10:27:22.581639	2012-08-02 11:21:46.429608
12387	1-12387	87	74	1662	61	1424	488	0	2012-08-02 11:19:49.286157	2012-08-02 11:26:49.593691
7159	1-7159	75	1	9233	37	7642	5318	8	2012-08-02 05:30:36.375287	2012-08-02 10:20:57.02679
6910	1-6910	6	73	9317	91	2236	522	5	2012-08-02 05:28:55.634608	2012-08-02 05:46:12.493422
6849	1-6849	63	63	8678	64	213	213	0	2012-08-02 05:28:30.413956	\N
13130	1-13130	8	48	3292	85	308	308	0	2012-08-02 11:24:39.782542	\N
13554	1-13554	98	22	3260	59	1970	1970	16	2012-08-02 11:27:28.227429	\N
8284	1-8284	68	4	8472	40	3116	1961	8	2012-08-02 05:38:18.171167	2012-08-02 05:40:35.409584
6851	1-6851	48	49	6873	80	371	371	0	2012-08-02 05:28:31.253423	\N
6620	1-6620	12	96	4359	74	720	298	4	2012-08-02 05:26:58.442977	2012-08-02 11:17:19.39034
8286	1-8286	64	6	4389	59	1115	697	0	2012-08-02 05:38:18.966476	2012-08-02 10:21:28.69255
7172	1-7172	64	63	9480	87	4645	4645	0	2012-08-02 05:30:41.722496	\N
11455	1-11455	85	34	8133	60	3959	3032	0	2012-08-02 11:13:47.688176	2012-08-02 11:17:33.896897
6869	1-6869	1	34	6859	40	4998	3157	4	2012-08-02 05:28:38.589877	2012-08-02 05:35:48.516574
11446	1-11446	30	27	6161	94	44	44	11	2012-08-02 11:13:44.274487	\N
10341	1-10341	67	99	7844	82	6424	3718	0	2012-08-02 10:27:20.905146	2012-08-02 11:21:31.046356
11449	1-11449	74	11	6907	20	6727	2627	13	2012-08-02 11:13:45.423575	2012-08-02 11:27:56.258298
6618	1-6618	94	58	8278	14	1951	1951	0	2012-08-02 05:26:57.603254	\N
6894	1-6894	75	18	8332	56	3246	1692	0	2012-08-02 05:28:48.966629	2012-08-02 11:28:44.502577
7979	1-7979	19	22	5200	55	521	521	3	2012-08-02 05:36:14.483249	\N
11452	1-11452	90	28	2087	80	752	752	8	2012-08-02 11:13:46.553921	\N
9690	1-9690	88	91	5134	80	1429	1429	0	2012-08-02 05:47:49.242133	\N
6624	1-6624	4	1	9165	2	7921	7921	3	2012-08-02 05:27:00.100372	\N
12067	1-12067	89	21	7562	95	256	256	5	2012-08-02 11:17:46.097405	\N
6626	1-6626	52	99	5918	74	2158	2158	0	2012-08-02 05:27:00.906628	\N
12070	1-12070	61	6	9278	14	5033	5033	5	2012-08-02 11:17:47.26847	\N
7168	1-7168	9	36	7260	10	1970	1970	3	2012-08-02 05:30:40.054292	\N
6629	1-6629	66	49	7395	90	62	62	0	2012-08-02 05:27:02.121833	\N
9167	1-9167	63	49	8052	34	3057	3057	0	2012-08-02 05:44:16.14843	\N
7406	1-7406	94	93	7275	7	2197	1955	0	2012-08-02 05:32:15.454349	2012-08-02 05:44:22.581626
13297	1-13297	90	59	2635	76	3394	3394	0	2012-08-02 11:25:45.56469	\N
6636	1-6636	14	61	4873	50	207	207	4	2012-08-02 05:27:04.994643	\N
9694	1-9694	85	83	4605	52	145	145	0	2012-08-02 05:47:50.888681	\N
11467	1-11467	41	17	2874	60	1271	1271	7	2012-08-02 11:13:52.261878	\N
13303	1-13303	31	19	4240	86	8511	8511	4	2012-08-02 11:25:48.019327	\N
12082	1-12082	8	39	6755	9	1016	1016	4	2012-08-02 11:17:51.79847	\N
12085	1-12085	70	88	3674	65	773	773	0	2012-08-02 11:17:52.92489	\N
6858	1-6858	60	22	8231	56	744	744	9	2012-08-02 05:28:34.115151	\N
7176	1-7176	81	66	7337	41	1718	1718	0	2012-08-02 05:30:43.368447	\N
9160	1-9160	20	97	5377	19	1394	408	6	2012-08-02 05:44:13.330508	2012-08-02 11:20:07.043702
6901	1-6901	96	100	2614	29	224	224	3	2012-08-02 05:28:51.870348	\N
6867	1-6867	91	38	8669	79	2096	2096	0	2012-08-02 05:28:37.772398	\N
6896	1-6896	68	38	9200	66	148	148	0	2012-08-02 05:28:49.808404	\N
6898	1-6898	30	35	4063	26	104	104	0	2012-08-02 05:28:50.629139	\N
6899	1-6899	33	68	7820	36	297	297	7	2012-08-02 05:28:51.038051	\N
6912	1-6912	25	55	9416	34	1762	1762	13	2012-08-02 05:28:56.451835	\N
6913	1-6913	61	54	6183	61	217	217	0	2012-08-02 05:28:56.849546	\N
7420	1-7420	28	55	4488	15	2057	2057	0	2012-08-02 05:32:21.289589	\N
10999	1-10999	28	46	7893	83	6234	4875	9	2012-08-02 10:39:22.338804	2012-08-02 11:20:27.394391
7202	1-7202	68	68	9844	91	2835	1040	0	2012-08-02 05:30:54.251158	2012-08-02 05:45:40.403421
7722	1-7722	35	60	7601	37	559	559	2	2012-08-02 05:34:25.812877	\N
13310	1-13310	55	35	9653	68	7533	5721	5	2012-08-02 11:25:51.223195	2012-08-02 11:28:23.097742
8569	1-8569	68	18	1897	30	463	87	16	2012-08-02 05:40:12.838017	2012-08-02 10:35:33.266173
9697	1-9697	48	92	8977	40	4692	4692	0	2012-08-02 05:47:52.103534	\N
8578	1-8578	17	39	9757	84	5167	3892	0	2012-08-02 05:40:16.472531	2012-08-02 11:15:16.259253
7717	1-7717	43	47	2222	76	913	250	4	2012-08-02 05:34:23.646738	2012-08-02 10:35:38.381654
9424	1-9424	95	44	9765	12	3446	1189	11	2012-08-02 05:46:00.569601	2012-08-02 11:28:50.33668
7215	1-7215	47	68	9358	31	5694	1708	9	2012-08-02 05:30:59.645112	2012-08-02 11:29:56.769677
7221	1-7221	67	88	5936	31	1704	1221	2	2012-08-02 05:31:02.141884	2012-08-02 11:23:09.156232
13104	1-13104	66	16	7472	66	2811	145	6	2012-08-02 11:24:29.165133	2012-08-02 11:30:24.323159
7222	1-7222	77	87	4167	67	290	99	0	2012-08-02 05:31:02.550874	2012-08-02 10:27:43.777161
10032	1-10032	44	3	6378	90	5839	1300	3	2012-08-02 10:19:49.036664	2012-08-02 11:17:25.334523
13315	1-13315	28	88	5604	69	7992	7903	4	2012-08-02 11:25:51.996544	2012-08-02 11:27:26.633186
11473	1-11473	8	96	8808	82	3007	2526	14	2012-08-02 11:13:54.980263	2012-08-02 11:23:44.946237
13308	1-13308	70	60	7267	96	3192	1028	13	2012-08-02 11:25:49.588316	2012-08-02 11:30:20.7384
9420	1-9420	6	18	9283	62	3301	3278	0	2012-08-02 05:45:58.94537	2012-08-02 05:47:14.238393
8002	1-8002	33	72	8443	48	3028	2212	0	2012-08-02 05:36:23.917954	2012-08-02 05:43:57.854521
7418	1-7418	43	19	9451	98	7510	4774	17	2012-08-02 05:32:20.483106	2012-08-02 11:24:29.949566
11456	1-11456	27	19	8404	11	5737	5199	12	2012-08-02 11:13:48.052751	2012-08-02 11:15:48.789614
11925	1-11925	96	66	6532	76	5705	3050	3	2012-08-02 11:16:50.209804	2012-08-02 11:29:23.984961
10040	1-10040	74	94	9391	58	7347	7036	0	2012-08-02 10:19:52.272488	2012-08-02 11:13:39.975895
10666	1-10666	49	96	5151	1	801	801	0	2012-08-02 10:36:07.780129	\N
7220	1-7220	52	64	5830	41	2076	1923	10	2012-08-02 05:31:01.722097	2012-08-02 05:34:22.375332
7226	1-7226	73	30	4664	57	3606	1555	3	2012-08-02 05:31:04.19664	2012-08-02 10:38:50.391199
7190	1-7190	31	16	8007	88	1893	460	0	2012-08-02 05:30:49.345621	2012-08-02 10:21:25.962018
13131	1-13131	43	29	3060	80	3372	3372	13	2012-08-02 11:24:40.169349	\N
7411	1-7411	66	75	8168	48	4509	3867	0	2012-08-02 05:32:17.553697	2012-08-02 10:31:53.750492
8839	1-8839	26	96	8209	10	170	170	5	2012-08-02 05:42:01.953328	\N
7217	1-7217	56	75	8097	58	1173	1173	0	2012-08-02 05:31:00.473701	\N
11910	1-11910	31	46	6386	45	1749	1749	14	2012-08-02 11:16:44.317719	\N
10034	1-10034	64	3	6237	67	2584	2584	4	2012-08-02 10:19:49.854629	\N
7429	1-7429	38	89	9281	9	784	784	8	2012-08-02 05:32:24.957361	\N
12380	1-12380	39	25	7750	62	4359	1674	5	2012-08-02 11:19:46.192567	2012-08-02 11:28:20.347033
11447	1-11447	22	35	4226	92	1246	1246	4	2012-08-02 11:13:44.661029	\N
11913	1-11913	100	11	8324	38	1403	1403	2	2012-08-02 11:16:45.522349	\N
9169	1-9169	42	59	8461	85	3421	3421	3	2012-08-02 05:44:16.944062	\N
7189	1-7189	8	87	5443	15	414	414	2	2012-08-02 05:30:48.947908	\N
13558	1-13558	96	31	6100	48	2768	2768	10	2012-08-02 11:27:29.759945	\N
7191	1-7191	58	51	5846	87	1304	1304	8	2012-08-02 05:30:49.754644	\N
12373	1-12373	80	95	5976	56	1	1	5	2012-08-02 11:19:43.839333	\N
7192	1-7192	57	51	7710	41	1511	1511	3	2012-08-02 05:30:50.163326	\N
13561	1-13561	88	56	9928	94	2941	2941	5	2012-08-02 11:27:30.964739	\N
10035	1-10035	13	59	5239	79	317	317	5	2012-08-02 10:19:50.252213	\N
8255	1-8255	71	58	9468	10	3476	1799	0	2012-08-02 05:38:06.493107	2012-08-02 11:24:44.12585
8570	1-8570	50	12	7053	4	315	315	0	2012-08-02 05:40:13.235557	\N
7991	1-7991	57	63	3053	56	319	319	0	2012-08-02 05:36:19.465985	\N
10661	1-10661	37	35	6275	15	5884	1226	0	2012-08-02 10:36:05.824121	2012-08-02 11:21:27.146426
9701	1-9701	25	42	9980	54	2874	2443	6	2012-08-02 05:47:53.761119	2012-08-02 11:14:22.747415
13295	1-13295	26	94	3002	58	8752	2316	2	2012-08-02 11:25:44.780263	2012-08-02 11:29:25.178304
11465	1-11465	36	35	5437	99	164	164	5	2012-08-02 11:13:51.499798	\N
10663	1-10663	83	78	4295	86	1867	1867	6	2012-08-02 10:36:06.597296	\N
13298	1-13298	48	51	6245	66	5234	3555	11	2012-08-02 11:25:45.952262	2012-08-02 11:29:01.863128
7183	1-7183	87	24	6762	22	2444	2181	5	2012-08-02 05:30:46.385038	2012-08-02 11:16:48.264351
13564	1-13564	74	62	4345	7	2809	2809	0	2012-08-02 11:27:32.201504	\N
7430	1-7430	10	47	6436	20	4743	696	7	2012-08-02 05:32:25.365938	2012-08-02 11:28:47.243134
7425	1-7425	64	79	5900	61	443	443	0	2012-08-02 05:32:23.300301	\N
7427	1-7427	88	99	4245	3	113	113	12	2012-08-02 05:32:24.128855	\N
7725	1-7725	61	100	6602	42	750	750	3	2012-08-02 05:34:27.083137	\N
13301	1-13301	68	7	6436	55	1370	1370	4	2012-08-02 11:25:47.243525	\N
11476	1-11476	34	89	8097	84	1143	1143	7	2012-08-02 11:13:55.731436	\N
9425	1-9425	72	18	8672	67	785	785	0	2012-08-02 05:46:00.978439	\N
7727	1-7727	25	81	6614	52	1635	515	0	2012-08-02 05:34:27.956246	2012-08-02 05:38:36.744904
13304	1-13304	99	90	5717	48	1043	1043	5	2012-08-02 11:25:48.394943	\N
8916	1-8916	41	38	7631	75	255	255	0	2012-08-02 05:42:33.23391	\N
7224	1-7224	97	59	5986	85	2101	2101	3	2012-08-02 05:31:03.368063	\N
4431	1-4431	24	12	4071	81	496	195	0	2012-08-02 05:12:14.489152	2012-08-02 10:19:51.883805
8000	1-8000	96	55	9524	96	3248	1450	0	2012-08-02 05:36:23.111648	2012-08-02 11:15:20.569822
11488	1-11488	63	54	9330	8	641	641	4	2012-08-02 11:14:00.330047	\N
7730	1-7730	53	3	6046	85	3586	3586	0	2012-08-02 05:34:29.270956	\N
10671	1-10671	90	3	8653	69	1763	1763	0	2012-08-02 10:36:09.747067	\N
7444	1-7444	98	16	5424	1	726	726	0	2012-08-02 05:32:31.110611	\N
8923	1-8923	30	96	9334	15	3291	3291	4	2012-08-02 05:42:36.084345	\N
13580	1-13580	26	89	2399	69	6836	2071	6	2012-08-02 11:27:38.443885	2012-08-02 11:29:25.178304
7495	1-7495	11	33	9177	16	9529	3393	0	2012-08-02 05:32:52.03821	2012-08-02 11:29:55.200678
11460	1-11460	32	87	9435	70	2781	1007	3	2012-08-02 11:13:49.588554	2012-08-02 11:26:47.220183
7476	1-7476	73	95	8904	23	1030	638	7	2012-08-02 05:32:44.279593	2012-08-02 05:34:27.513884
7481	1-7481	68	62	4700	14	2059	1756	0	2012-08-02 05:32:46.312776	2012-08-02 10:38:35.899927
11463	1-11463	6	19	4712	28	5212	2515	0	2012-08-02 11:13:50.737392	2012-08-02 11:29:44.95862
10348	1-10348	100	62	9054	82	3126	3126	5	2012-08-02 10:27:23.852607	\N
7521	1-7521	74	35	6979	82	4748	1150	0	2012-08-02 05:33:02.876367	2012-08-02 11:24:20.71302
8008	1-8008	89	59	9926	14	7308	7308	7	2012-08-02 05:36:26.348678	\N
8014	1-8014	86	64	7584	93	2883	899	2	2012-08-02 05:36:28.745978	2012-08-02 05:37:50.201113
8859	1-8859	35	65	9966	20	5686	5375	6	2012-08-02 05:42:10.042103	2012-08-02 05:43:33.1195
7499	1-7499	55	42	840	50	127	127	0	2012-08-02 05:32:53.706107	\N
7745	1-7745	35	36	5052	94	156	156	9	2012-08-02 05:34:35.635179	\N
10676	1-10676	99	69	3398	54	3075	3015	0	2012-08-02 10:36:11.759906	2012-08-02 11:14:22.747415
8850	1-8850	13	23	1319	9	178	178	12	2012-08-02 05:42:06.406954	\N
7479	1-7479	91	26	9868	75	1025	1025	0	2012-08-02 05:32:45.494836	\N
7787	1-7787	85	63	9972	40	9606	3802	0	2012-08-02 05:34:53.425687	2012-08-02 05:41:32.077116
10352	1-10352	15	68	8381	59	3054	2331	0	2012-08-02 10:27:25.476746	2012-08-02 11:19:19.94982
10672	1-10672	61	68	9857	15	347	347	0	2012-08-02 10:36:10.144703	\N
7500	1-7500	20	25	5271	54	1758	1758	3	2012-08-02 05:32:54.114805	\N
8273	1-8273	9	22	8062	10	3788	643	0	2012-08-02 05:38:13.685552	2012-08-02 11:30:27.067488
7775	1-7775	34	29	7478	51	890	304	2	2012-08-02 05:34:48.332088	2012-08-02 05:38:31.496507
10048	1-10048	62	57	9105	55	6362	399	5	2012-08-02 10:20:59.528082	2012-08-02 11:29:17.18967
10047	1-10047	29	24	3857	15	2187	1197	11	2012-08-02 10:20:59.108446	2012-08-02 11:28:30.157943
9172	1-9172	39	36	8919	91	7767	1901	0	2012-08-02 05:44:18.183871	2012-08-02 11:30:10.516962
9175	1-9175	2	46	7124	11	1407	1273	0	2012-08-02 05:44:19.377503	2012-08-02 10:39:03.101981
13559	1-13559	97	34	7060	8	627	627	6	2012-08-02 11:27:30.56652	\N
8261	1-8261	12	81	7669	40	772	772	3	2012-08-02 05:38:08.901926	\N
8263	1-8263	16	6	1029	68	26	26	3	2012-08-02 05:38:09.697206	\N
8581	1-8581	33	66	8741	2	1028	1028	11	2012-08-02 05:40:17.687997	\N
7504	1-7504	1	48	7570	21	623	623	0	2012-08-02 05:32:55.77208	\N
7757	1-7757	83	33	5869	73	2048	2048	2	2012-08-02 05:34:40.719234	\N
8849	1-8849	1	3	6595	31	2535	2535	9	2012-08-02 05:42:06.009414	\N
11000	1-11000	46	99	1801	97	100	100	14	2012-08-02 10:39:22.747944	\N
11451	1-11451	26	13	493	72	145	145	0	2012-08-02 11:13:46.163487	\N
8007	1-8007	74	78	4861	65	963	963	0	2012-08-02 05:36:25.951126	\N
7489	1-7489	11	15	9793	44	889	889	13	2012-08-02 05:32:49.583206	\N
11002	1-11002	54	9	7152	16	2211	2211	2	2012-08-02 10:39:23.534881	\N
8267	1-8267	34	76	9734	23	1765	1765	2	2012-08-02 05:38:11.29908	\N
9707	1-9707	13	11	2276	86	1570	1570	4	2012-08-02 05:47:56.18075	\N
13562	1-13562	45	39	6926	75	5286	5286	5	2012-08-02 11:27:31.362372	\N
10043	1-10043	100	89	9164	95	525	525	0	2012-08-02 10:20:57.494768	\N
7492	1-7492	61	44	8444	1	1327	1327	0	2012-08-02 05:32:50.809678	\N
10044	1-10044	47	30	3914	46	174	174	7	2012-08-02 10:20:57.892524	\N
12391	1-12391	70	14	9052	82	2437	2437	5	2012-08-02 11:19:50.821893	\N
11457	1-11457	50	4	5318	72	1592	1592	0	2012-08-02 11:13:48.417158	\N
13571	1-13571	51	75	5197	4	2219	2219	2	2012-08-02 11:27:34.919845	\N
9174	1-9174	94	17	5214	59	1120	1120	11	2012-08-02 05:44:18.979395	\N
7518	1-7518	2	46	8072	62	2762	474	12	2012-08-02 05:33:01.583839	2012-08-02 11:11:58.737639
7496	1-7496	64	16	3052	52	668	668	11	2012-08-02 05:32:52.480127	\N
9430	1-9430	35	36	3707	2	5442	5442	3	2012-08-02 05:46:02.989046	\N
7497	1-7497	30	74	7501	56	952	952	0	2012-08-02 05:32:52.899877	\N
13574	1-13574	68	18	9649	99	2011	2011	11	2012-08-02 11:27:36.124171	\N
13577	1-13577	8	92	7041	9	192	192	0	2012-08-02 11:27:37.306432	\N
9429	1-9429	5	99	3432	10	1353	377	0	2012-08-02 05:46:02.591304	2012-08-02 11:23:39.862557
2118	1-2118	77	28	9907	90	2731	1654	5	2012-08-02 04:56:21.960386	2012-08-02 05:33:01.152719
8857	1-8857	6	56	4024	37	1837	1837	0	2012-08-02 05:42:09.24667	\N
11911	1-11911	58	40	4580	90	762	441	0	2012-08-02 11:16:44.74921	2012-08-02 11:23:17.763025
12639	1-12639	17	12	8135	80	977	977	5	2012-08-02 11:21:27.555641	\N
8861	1-8861	15	69	6954	62	491	491	0	2012-08-02 05:42:10.881765	\N
10049	1-10049	54	100	8870	77	4257	3253	14	2012-08-02 10:20:59.925968	2012-08-02 11:28:09.375063
7761	1-7761	22	8	7363	37	3822	3822	6	2012-08-02 05:34:42.420514	\N
11445	1-11445	41	81	5513	46	2493	1626	3	2012-08-02 11:13:43.909701	2012-08-02 11:23:29.300201
7763	1-7763	59	33	8463	92	5026	5026	0	2012-08-02 05:34:43.25999	\N
7772	1-7772	38	53	3035	86	478	478	0	2012-08-02 05:34:47.105477	\N
7781	1-7781	38	25	9990	28	1245	1245	0	2012-08-02 05:34:50.862385	\N
7783	1-7783	48	47	9022	31	891	891	3	2012-08-02 05:34:51.702262	\N
7784	1-7784	47	56	7920	5	6095	6095	3	2012-08-02 05:34:52.133012	\N
11489	1-11489	86	18	8462	90	7965	2903	10	2012-08-02 11:14:00.727684	2012-08-02 11:27:12.599056
11927	1-11927	68	69	9869	3	3886	819	0	2012-08-02 11:16:50.994066	2012-08-02 11:29:46.549736
8032	1-8032	60	33	9551	24	3414	752	10	2012-08-02 05:36:36.029099	2012-08-02 10:38:36.286525
10059	1-10059	27	23	8023	93	5337	1507	3	2012-08-02 10:21:03.928791	2012-08-02 11:15:49.574199
8031	1-8031	20	11	9931	4	776	776	0	2012-08-02 05:36:35.63173	\N
8085	1-8085	31	83	8345	37	3561	2692	3	2012-08-02 05:36:57.241102	2012-08-02 05:41:19.81781
10354	1-10354	76	53	9692	17	5988	202	0	2012-08-02 10:27:26.305432	2012-08-02 11:24:14.10453
12634	1-12634	35	80	4950	4	3199	2643	5	2012-08-02 11:21:25.600107	2012-08-02 11:28:54.736808
11501	1-11501	12	83	8361	38	3935	2478	0	2012-08-02 11:14:05.411623	2012-08-02 11:30:32.230146
10052	1-10052	9	49	3077	88	244	244	23	2012-08-02 10:21:01.185204	\N
13135	1-13135	84	1	3800	69	4685	3612	9	2012-08-02 11:24:41.705139	2012-08-02 11:25:24.204162
10355	1-10355	51	47	9055	6	5621	572	21	2012-08-02 10:27:26.691551	2012-08-02 11:26:22.651417
7798	1-7798	51	85	7239	2	1862	1862	0	2012-08-02 05:34:58.054716	\N
11009	1-11009	77	45	9936	91	4066	3337	11	2012-08-02 10:39:26.276801	2012-08-02 11:18:46.294302
7792	1-7792	5	23	9979	13	4506	3805	0	2012-08-02 05:34:55.524994	2012-08-02 11:27:16.568012
8038	1-8038	77	48	9189	73	1916	1113	0	2012-08-02 05:36:38.448448	2012-08-02 10:22:03.89283
7801	1-7801	37	94	7659	84	371	371	7	2012-08-02 05:34:59.359053	\N
8033	1-8033	32	9	7251	81	5003	1577	3	2012-08-02 05:36:36.426847	2012-08-02 11:18:04.813504
10689	1-10689	48	6	9204	32	3170	2895	4	2012-08-02 10:36:16.866759	2012-08-02 11:13:26.705297
7808	1-7808	85	80	7996	2	4825	4072	11	2012-08-02 05:35:02.38565	2012-08-02 11:30:22.366312
12628	1-12628	77	67	7354	53	3202	1877	6	2012-08-02 11:21:23.268529	2012-08-02 11:30:38.533307
13293	1-13293	57	99	7979	61	9510	740	8	2012-08-02 11:25:43.994978	2012-08-02 11:30:40.501224
7804	1-7804	27	83	5270	10	840	840	0	2012-08-02 05:35:00.695121	\N
11912	1-11912	45	59	4406	10	657	657	0	2012-08-02 11:16:45.135897	\N
8062	1-8062	29	52	8719	49	1036	792	0	2012-08-02 05:36:48.026949	2012-08-02 11:14:13.157551
8277	1-8277	21	84	3651	40	2585	2585	0	2012-08-02 05:38:15.298407	\N
13296	1-13296	66	39	5290	22	7118	7118	2	2012-08-02 11:25:45.167229	\N
8278	1-8278	13	39	8350	18	234	234	0	2012-08-02 05:38:15.707544	\N
8864	1-8864	21	3	7244	91	4803	988	0	2012-08-02 05:42:12.08611	2012-08-02 11:26:13.358762
8030	1-8030	11	90	4410	22	1279	810	9	2012-08-02 05:36:35.211563	2012-08-02 11:13:26.705297
7810	1-7810	15	24	4542	46	181	181	3	2012-08-02 05:35:03.214343	\N
13576	1-13576	96	74	7481	40	4532	4532	3	2012-08-02 11:27:36.908253	\N
10682	1-10682	84	85	2771	84	256	256	2	2012-08-02 10:36:14.114684	\N
11007	1-11007	52	78	8034	64	3689	3689	3	2012-08-02 10:39:25.477212	\N
7815	1-7815	77	40	6648	87	279	279	10	2012-08-02 05:35:05.380075	\N
13570	1-13570	40	70	4817	45	3814	1008	8	2012-08-02 11:27:34.532794	2012-08-02 11:27:52.435896
13305	1-13305	63	4	8298	98	1349	1349	4	2012-08-02 11:25:48.815187	\N
10358	1-10358	31	35	7056	2	4645	3121	0	2012-08-02 10:27:27.873401	2012-08-02 11:11:08.047947
8037	1-8037	81	39	8006	29	2358	2358	14	2012-08-02 05:36:38.039824	\N
9182	1-9182	10	31	3016	53	34	34	0	2012-08-02 05:44:22.194929	\N
13299	1-13299	4	35	2277	17	4065	2911	7	2012-08-02 11:25:46.384474	2012-08-02 11:27:44.513646
11478	1-11478	5	81	8661	27	2702	1716	4	2012-08-02 11:13:56.496069	2012-08-02 11:20:30.504167
8868	1-8868	99	93	5091	92	1951	1951	0	2012-08-02 05:42:13.721105	\N
9435	1-9435	4	51	1706	79	479	479	0	2012-08-02 05:46:04.999635	\N
9184	1-9184	26	80	5756	14	258	258	5	2012-08-02 05:44:23.002051	\N
8591	1-8591	37	94	8215	48	2180	2180	16	2012-08-02 05:40:21.765035	\N
9436	1-9436	81	94	7126	99	1954	1954	0	2012-08-02 05:46:05.408565	\N
8592	1-8592	25	2	6620	38	370	370	0	2012-08-02 05:40:22.16257	\N
7806	1-7806	88	12	6165	68	1082	203	4	2012-08-02 05:35:01.5349	2012-08-02 10:27:30.659208
8041	1-8041	55	47	5962	61	1342	1342	10	2012-08-02 05:36:39.619712	\N
7814	1-7814	67	95	6922	100	1834	1113	10	2012-08-02 05:35:04.937894	2012-08-02 11:28:00.396833
11484	1-11484	56	80	8020	52	373	373	2	2012-08-02 11:13:58.771717	\N
10060	1-10060	35	21	9941	85	668	668	7	2012-08-02 10:21:04.31532	\N
8506	1-8506	53	19	9139	10	9299	3398	0	2012-08-02 05:39:47.544297	2012-08-02 11:27:39.261479
10061	1-10061	86	99	8698	75	772	772	7	2012-08-02 10:21:04.690968	\N
11486	1-11486	17	3	3244	2	571	571	8	2012-08-02 11:13:59.943166	\N
8059	1-8059	84	11	8793	38	6766	4924	0	2012-08-02 05:36:46.844787	2012-08-02 11:28:14.457753
12637	1-12637	84	52	8947	66	1508	1508	7	2012-08-02 11:21:26.770901	\N
8077	1-8077	77	3	6138	60	8007	3443	4	2012-08-02 05:36:54.037094	2012-08-02 11:28:14.866431
12640	1-12640	76	24	7587	14	6103	6103	0	2012-08-02 11:21:27.930805	\N
8080	1-8080	74	91	5124	70	3685	2535	0	2012-08-02 05:36:55.241184	2012-08-02 11:22:24.670211
8057	1-8057	100	26	8911	18	3310	3267	6	2012-08-02 05:36:46.049763	2012-08-02 11:22:47.278678
9559	1-9559	79	11	5464	71	2635	2635	7	2012-08-02 05:46:55.452604	\N
8058	1-8058	60	86	8501	74	112	112	7	2012-08-02 05:36:46.447038	\N
8066	1-8066	20	36	2162	11	7	7	0	2012-08-02 05:36:49.629185	\N
8064	1-8064	86	19	9411	70	2307	2307	4	2012-08-02 05:36:48.82248	\N
8071	1-8071	24	57	8644	37	3272	1827	4	2012-08-02 05:36:51.617508	2012-08-02 05:36:51.617508
8070	1-8070	1	17	6455	34	426	426	0	2012-08-02 05:36:51.219893	\N
8078	1-8078	18	13	3907	43	856	856	5	2012-08-02 05:36:54.445924	\N
13144	1-13144	54	2	9079	64	5120	4137	0	2012-08-02 11:24:45.321383	2012-08-02 11:26:58.843133
9200	1-9200	3	93	9291	53	3007	2762	4	2012-08-02 05:44:29.488747	2012-08-02 11:24:04.779544
13566	1-13566	13	85	7718	78	1854	1854	13	2012-08-02 11:27:33.361511	\N
11006	1-11006	20	19	8064	87	8766	1804	6	2012-08-02 10:39:25.089248	2012-08-02 11:29:44.95862
8873	1-8873	14	96	7625	100	3097	2554	0	2012-08-02 05:42:15.742946	2012-08-02 11:27:32.588199
11470	1-11470	11	78	2215	49	1104	1104	3	2012-08-02 11:13:53.399934	\N
8298	1-8298	56	61	6587	100	3101	3092	10	2012-08-02 05:38:23.895646	2012-08-02 11:26:20.242499
11947	1-11947	28	81	8254	67	2940	2053	4	2012-08-02 11:16:58.88281	2012-08-02 11:21:09.963143
9188	1-9188	28	22	9243	57	1975	1975	3	2012-08-02 05:44:24.68251	\N
12661	1-12661	7	29	9653	31	2255	2176	0	2012-08-02 11:21:36.173383	2012-08-02 11:21:53.412361
11926	1-11926	38	99	1457	44	321	321	0	2012-08-02 11:16:50.618444	\N
8872	1-8872	45	18	6923	73	2403	817	0	2012-08-02 05:42:15.323024	2012-08-02 11:29:45.743019
8376	1-8376	28	78	4349	70	3783	2590	3	2012-08-02 05:38:55.253915	2012-08-02 10:38:50.391199
8290	1-8290	15	97	5992	83	4576	1493	0	2012-08-02 05:38:20.645622	2012-08-02 11:13:56.109039
12392	1-12392	20	31	7621	92	142	142	6	2012-08-02 11:19:51.219661	\N
8353	1-8353	90	84	2912	46	661	198	8	2012-08-02 05:38:45.969966	2012-08-02 05:46:32.31477
8350	1-8350	55	39	9748	9	4150	682	0	2012-08-02 05:38:44.754753	2012-08-02 11:29:15.220572
11929	1-11929	58	72	2814	80	434	434	0	2012-08-02 11:16:51.767401	\N
10065	1-10065	100	93	3991	60	3393	3044	3	2012-08-02 10:21:06.249359	2012-08-02 11:20:20.346337
13575	1-13575	67	71	8725	74	1877	1748	0	2012-08-02 11:27:36.511118	2012-08-02 11:30:04.669386
9193	1-9193	49	56	9185	66	586	586	0	2012-08-02 05:44:26.715722	\N
13572	1-13572	15	32	9692	52	242	242	5	2012-08-02 11:27:35.328487	\N
8871	1-8871	24	11	8784	5	4115	4115	0	2012-08-02 05:42:14.925613	\N
8594	1-8594	86	60	7171	36	124	124	0	2012-08-02 05:40:22.958163	\N
12643	1-12643	3	53	4066	83	6680	4497	3	2012-08-02 11:21:29.091169	2012-08-02 11:26:12.950056
12395	1-12395	41	59	1303	93	1136	1136	0	2012-08-02 11:19:52.40213	\N
9713	1-9713	89	94	9157	76	7080	4612	2	2012-08-02 05:47:58.691351	2012-08-02 11:25:32.107626
13587	1-13587	20	20	8609	4	2661	2661	12	2012-08-02 11:27:41.241337	\N
10364	1-10364	7	46	2546	28	633	633	22	2012-08-02 10:27:30.261125	\N
9714	1-9714	67	39	3705	2	1667	1667	0	2012-08-02 05:47:59.110983	\N
13139	1-13139	51	74	3372	41	1790	1790	0	2012-08-02 11:24:43.307103	\N
13141	1-13141	1	8	3976	57	2211	2211	10	2012-08-02 11:24:44.545681	\N
13578	1-13578	87	95	5747	40	1197	1197	0	2012-08-02 11:27:37.681731	\N
8309	1-8309	60	19	7928	14	2029	2029	4	2012-08-02 05:38:28.292641	\N
12655	1-12655	43	84	9184	30	963	963	5	2012-08-02 11:21:33.798059	\N
11487	1-11487	75	59	7355	37	3327	3327	0	2012-08-02 11:13:59.545267	\N
9197	1-9197	40	88	3477	40	774	774	0	2012-08-02 05:44:28.306604	\N
12408	1-12408	11	35	7810	48	6703	6554	4	2012-08-02 11:19:57.417779	2012-08-02 11:19:57.417779
8324	1-8324	66	90	9334	87	6230	3099	10	2012-08-02 05:38:34.314038	2012-08-02 10:37:45.081324
13584	1-13584	50	35	8778	12	4399	4399	3	2012-08-02 11:27:40.046138	\N
8313	1-8313	72	74	9187	59	2292	2292	6	2012-08-02 05:38:29.883632	\N
11496	1-11496	54	9	9440	45	3096	3096	5	2012-08-02 11:14:03.478668	\N
8311	1-8311	32	33	9096	91	5124	4026	0	2012-08-02 05:38:29.110216	2012-08-02 11:29:31.862809
11012	1-11012	24	6	4084	53	805	805	0	2012-08-02 10:39:27.457152	\N
13147	1-13147	28	6	4106	62	2907	1822	8	2012-08-02 11:24:46.481649	2012-08-02 11:28:32.889244
8364	1-8364	39	89	6557	17	4587	3719	10	2012-08-02 05:38:50.420272	2012-08-02 11:20:00.480137
8315	1-8315	50	84	7862	27	218	218	0	2012-08-02 05:38:30.712598	\N
10069	1-10069	51	19	4469	72	1692	1692	0	2012-08-02 10:21:07.796588	\N
8316	1-8316	37	86	1994	40	365	365	6	2012-08-02 05:38:31.110005	\N
9719	1-9719	100	75	6233	87	1937	1937	0	2012-08-02 05:48:01.144158	\N
9196	1-9196	75	30	6702	93	5095	596	0	2012-08-02 05:44:27.908926	2012-08-02 11:20:29.320022
8293	1-8293	68	79	5406	57	2188	1741	0	2012-08-02 05:38:21.874228	2012-08-02 11:20:01.662484
11950	1-11950	40	93	5352	80	1658	1658	7	2012-08-02 11:17:00.042824	\N
11480	1-11480	11	38	8349	7	5347	1555	0	2012-08-02 11:13:57.258277	2012-08-02 11:26:36.46422
9721	1-9721	16	42	390	81	34	34	0	2012-08-02 05:48:01.972474	\N
8326	1-8326	46	8	6477	19	838	838	0	2012-08-02 05:38:35.131985	\N
9199	1-9199	6	62	8318	47	724	724	0	2012-08-02 05:44:29.080278	\N
11507	1-11507	7	37	9547	11	1299	1299	7	2012-08-02 11:14:07.732044	\N
11962	1-11962	72	26	8069	95	6341	3063	8	2012-08-02 11:17:04.794544	2012-08-02 11:24:50.381471
8882	1-8882	53	21	8298	77	4741	56	0	2012-08-02 05:42:19.435097	2012-08-02 11:26:37.271162
9203	1-9203	96	37	9153	42	1349	1349	4	2012-08-02 05:44:30.726799	\N
13591	1-13591	53	57	7391	81	3565	3565	16	2012-08-02 11:27:42.556313	\N
11959	1-11959	47	24	9001	84	4586	4586	19	2012-08-02 11:17:03.611585	\N
11965	1-11965	58	68	2797	49	741	741	5	2012-08-02 11:17:05.96516	\N
13744	1-13744	58	15	8373	14	6369	3304	12	2012-08-02 11:28:43.735007	2012-08-02 11:28:43.735007
11968	1-11968	66	2	9992	34	1083	1083	0	2012-08-02 11:17:07.238386	\N
8331	1-8331	31	54	6268	80	1567	1567	4	2012-08-02 05:38:37.15352	\N
9207	1-9207	3	38	6571	96	121	121	0	2012-08-02 05:44:32.339869	\N
8344	1-8344	35	98	8518	38	1299	1299	3	2012-08-02 05:38:42.379685	\N
8347	1-8347	64	63	8591	70	1808	1808	3	2012-08-02 05:38:43.561577	\N
8357	1-8357	77	97	5996	79	1409	1409	7	2012-08-02 05:38:47.560989	\N
8674	1-8674	39	36	7923	94	4183	1745	9	2012-08-02 05:40:55.089461	2012-08-02 10:21:29.112493
8892	1-8892	81	21	8002	31	2660	1544	0	2012-08-02 05:42:23.489424	2012-08-02 11:28:55.123393
10077	1-10077	24	30	9133	71	701	701	0	2012-08-02 10:21:10.938456	\N
11494	1-11494	11	93	9671	40	8801	1117	15	2012-08-02 11:14:02.682978	2012-08-02 11:30:14.048472
8658	1-8658	18	93	9048	34	4224	2892	0	2012-08-02 05:40:48.681176	2012-08-02 10:38:02.915583
10448	1-10448	1	66	6167	32	2623	2614	6	2012-08-02 10:31:36.240732	2012-08-02 10:39:18.027505
8886	1-8886	54	51	9425	10	2544	2544	4	2012-08-02 05:42:21.069859	\N
13748	1-13748	65	89	6387	100	3079	2344	4	2012-08-02 11:28:45.26498	2012-08-02 11:29:25.178304
9231	1-9231	35	40	9885	64	1194	274	15	2012-08-02 05:44:42.098642	2012-08-02 10:22:26.533431
8895	1-8895	70	47	1405	12	717	578	0	2012-08-02 05:42:24.727114	2012-08-02 11:22:47.665344
11503	1-11503	54	53	4333	60	4615	74	8	2012-08-02 11:14:06.174495	2012-08-02 11:30:21.125048
9723	1-9723	85	35	7283	44	6484	1405	0	2012-08-02 05:48:02.801298	2012-08-02 11:23:32.018196
8904	1-8904	88	97	6410	92	7505	1498	0	2012-08-02 05:42:28.350659	2012-08-02 11:19:19.94982
13745	1-13745	53	76	5241	15	6303	3664	0	2012-08-02 11:28:44.116183	2012-08-02 11:30:39.30669
8680	1-8680	94	61	9356	42	8749	7747	8	2012-08-02 05:40:57.509025	2012-08-02 05:41:26.784947
8642	1-8642	19	98	7401	30	1420	1150	0	2012-08-02 05:40:42.250588	2012-08-02 11:16:48.264351
8611	1-8611	25	49	7002	55	26	26	8	2012-08-02 05:40:29.785996	\N
10074	1-10074	66	53	8639	15	9054	2874	0	2012-08-02 10:21:09.7557	2012-08-02 10:36:05.049976
8393	1-8393	49	3	2871	94	1030	750	0	2012-08-02 05:39:02.158796	2012-08-02 11:26:04.262141
12645	1-12645	34	17	9189	54	3457	3457	4	2012-08-02 11:21:29.864239	\N
8637	1-8637	51	31	1495	42	849	216	5	2012-08-02 05:40:40.228758	2012-08-02 05:43:51.234849
9220	1-9220	47	74	4493	51	1722	1128	7	2012-08-02 05:44:37.568103	2012-08-02 11:15:09.819771
9442	1-9442	89	1	9346	72	211	211	9	2012-08-02 05:46:07.994763	\N
8672	1-8672	38	31	6149	52	1288	1000	0	2012-08-02 05:40:54.282835	2012-08-02 10:36:05.437352
8395	1-8395	42	47	8948	40	574	574	0	2012-08-02 05:39:02.954668	\N
11933	1-11933	73	97	8479	83	3741	3741	5	2012-08-02 11:16:53.34709	\N
8630	1-8630	57	28	9974	54	5237	5237	4	2012-08-02 05:40:37.398418	\N
8679	1-8679	63	13	5571	90	3254	1662	0	2012-08-02 05:40:57.122149	2012-08-02 11:20:58.582017
8397	1-8397	79	98	4710	100	1719	1719	0	2012-08-02 05:39:03.773124	\N
8631	1-8631	100	19	8143	79	1682	1682	0	2012-08-02 05:40:37.796215	\N
8398	1-8398	94	15	7528	57	545	545	0	2012-08-02 05:39:04.248437	\N
8905	1-8905	81	58	8154	28	2209	1172	6	2012-08-02 05:42:28.759988	2012-08-02 11:25:04.93251
9443	1-9443	79	97	6739	54	2576	2576	5	2012-08-02 05:46:08.425956	\N
10436	1-10436	3	37	9764	86	76	76	0	2012-08-02 10:27:58.748773	\N
12393	1-12393	72	49	2446	6	2062	1243	3	2012-08-02 11:19:51.595353	2012-08-02 11:26:28.56418
13585	1-13585	55	62	7352	98	2512	2512	0	2012-08-02 11:27:40.444159	\N
11500	1-11500	13	79	3268	38	307	307	9	2012-08-02 11:14:05.024953	\N
12399	1-12399	23	32	6933	3	656	656	10	2012-08-02 11:19:53.915302	\N
10437	1-10437	54	58	7278	19	368	368	6	2012-08-02 10:27:59.13551	\N
11946	1-11946	33	30	9437	19	432	432	10	2012-08-02 11:16:58.496335	\N
9724	1-9724	48	61	6597	42	6931	602	11	2012-08-02 05:48:03.232291	2012-08-02 11:27:56.667174
12401	1-12401	40	90	5201	14	2341	2341	11	2012-08-02 11:19:55.075938	\N
9219	1-9219	13	97	8117	31	2317	2317	10	2012-08-02 05:44:37.170532	\N
10443	1-10443	80	86	5717	21	2329	2262	0	2012-08-02 10:31:34.170913	2012-08-02 11:21:48.407505
10078	1-10078	67	94	9974	54	4160	4160	0	2012-08-02 10:21:11.324902	\N
11509	1-11509	46	64	7488	84	3505	3505	0	2012-08-02 11:14:08.494483	\N
7417	1-7417	77	1	9044	62	4050	2196	0	2012-08-02 05:32:20.07442	2012-08-02 11:20:31.675507
8622	1-8622	18	31	3845	86	252	252	0	2012-08-02 05:40:34.194504	\N
8623	1-8623	98	78	8835	54	5784	4918	4	2012-08-02 05:40:34.592473	2012-08-02 05:40:34.592473
8385	1-8385	100	48	9962	49	1957	1311	0	2012-08-02 05:38:58.916408	2012-08-02 05:44:38.385974
10701	1-10701	90	67	9275	7	1130	1130	0	2012-08-02 10:36:21.618803	\N
13751	1-13751	26	14	9916	76	7681	7681	2	2012-08-02 11:28:46.447691	\N
10441	1-10441	38	66	9547	20	574	574	8	2012-08-02 10:31:33.341811	\N
13754	1-13754	88	96	2367	43	685	685	4	2012-08-02 11:28:47.640906	\N
10081	1-10081	98	67	7256	69	1398	1398	2	2012-08-02 10:21:12.474892	\N
10702	1-10702	56	100	6529	32	388	388	5	2012-08-02 10:36:22.005471	\N
8898	1-8898	29	41	3721	4	674	674	0	2012-08-02 05:42:25.931248	\N
10082	1-10082	38	82	3875	16	652	652	0	2012-08-02 10:21:12.861367	\N
10703	1-10703	77	42	5894	19	428	428	4	2012-08-02 10:36:22.392084	\N
10704	1-10704	7	8	9532	37	101	101	8	2012-08-02 10:36:22.767763	\N
10445	1-10445	5	33	2033	97	575	575	4	2012-08-02 10:31:34.991915	\N
8619	1-8619	38	24	7121	73	4628	2083	0	2012-08-02 05:40:32.979003	2012-08-02 11:28:53.543122
9229	1-9229	90	91	5715	80	606	606	0	2012-08-02 05:44:41.314232	\N
9230	1-9230	34	46	9307	15	145	145	0	2012-08-02 05:44:41.711937	\N
8903	1-8903	67	76	5788	72	230	230	0	2012-08-02 05:42:27.96411	\N
8645	1-8645	10	3	4202	4	901	901	0	2012-08-02 05:40:43.44391	\N
8647	1-8647	88	75	9184	96	1661	1661	0	2012-08-02 05:40:44.228189	\N
8614	1-8614	81	48	7512	59	1938	601	2	2012-08-02 05:40:30.968016	2012-08-02 10:31:42.618897
8653	1-8653	90	65	9842	54	3553	3553	0	2012-08-02 05:40:46.659301	\N
8682	1-8682	76	35	4117	60	2228	2228	3	2012-08-02 05:40:58.326234	\N
8684	1-8684	36	82	8538	28	1552	1552	5	2012-08-02 05:40:59.121815	\N
13145	1-13145	6	86	8936	90	9778	8220	0	2012-08-02 11:24:45.708282	2012-08-02 11:27:12.599056
8924	1-8924	45	78	1033	5	55	55	2	2012-08-02 05:42:36.48217	\N
13586	1-13586	100	1	8788	67	5889	2558	0	2012-08-02 11:27:40.866207	2012-08-02 11:29:55.996314
9839	1-9839	66	42	7783	69	3161	1263	8	2012-08-02 05:48:49.891771	2012-08-02 11:17:13.466838
12670	1-12670	63	13	4819	75	1264	451	7	2012-08-02 11:21:39.693403	2012-08-02 11:30:03.906984
8952	1-8952	27	11	8029	22	875	875	0	2012-08-02 05:42:47.949856	\N
11514	1-11514	25	13	5949	32	6070	3109	9	2012-08-02 11:14:10.427875	2012-08-02 11:30:36.876255
8953	1-8953	58	96	7003	60	1552	1552	5	2012-08-02 05:42:48.358645	\N
11511	1-11511	95	63	8952	82	9740	9740	3	2012-08-02 11:14:09.278931	\N
11934	1-11934	98	27	8552	19	3252	291	8	2012-08-02 11:16:53.722698	2012-08-02 11:21:22.892901
8954	1-8954	48	24	8442	53	2935	2935	0	2012-08-02 05:42:48.745422	\N
9828	1-9828	63	83	3294	98	111	111	6	2012-08-02 05:48:45.421978	\N
9836	1-9836	41	34	7433	38	722	722	0	2012-08-02 05:48:48.670747	\N
8928	1-8928	44	85	5553	46	95	95	0	2012-08-02 05:42:38.15036	\N
9260	1-9260	79	41	5361	86	1819	1306	0	2012-08-02 05:44:53.949013	2012-08-02 09:10:31.355931
9241	1-9241	18	94	3826	38	1467	751	2	2012-08-02 05:44:46.136908	2012-08-02 11:19:23.827988
8929	1-8929	7	78	1666	47	666	394	3	2012-08-02 05:42:38.559448	2012-08-02 05:49:50.434186
11532	1-11532	38	50	6880	83	9229	2219	4	2012-08-02 11:14:17.388488	2012-08-02 11:21:12.327546
10706	1-10706	51	32	6552	64	3163	2308	3	2012-08-02 10:36:23.552506	2012-08-02 11:14:51.954176
10376	1-10376	83	12	8761	94	2747	716	9	2012-08-02 10:27:35.035427	2012-08-02 11:24:07.111196
9449	1-9449	35	38	8545	2	1212	1212	0	2012-08-02 05:46:10.880392	\N
8984	1-8984	91	3	8085	68	6299	2695	3	2012-08-02 05:43:01.09929	2012-08-02 11:27:17.739084
11557	1-11557	75	48	7338	10	1827	1235	15	2012-08-02 11:14:26.970517	2012-08-02 11:19:24.955259
12654	1-12654	23	68	4387	54	7783	1152	9	2012-08-02 11:21:33.024634	2012-08-02 11:23:11.863359
9451	1-9451	100	43	4907	87	2873	2873	4	2012-08-02 05:46:11.68672	\N
12412	1-12412	16	52	9166	71	502	502	6	2012-08-02 11:19:58.95537	\N
12650	1-12650	42	63	2180	93	532	532	8	2012-08-02 11:21:31.831058	\N
9437	1-9437	74	99	8591	55	3380	3308	3	2012-08-02 05:46:05.817878	2012-08-02 10:39:28.229163
11945	1-11945	70	76	9583	40	5446	5446	0	2012-08-02 11:16:58.09847	\N
10371	1-10371	24	74	9314	59	3878	3878	6	2012-08-02 10:27:33.069223	\N
5952	1-5952	93	79	4326	95	478	361	0	2012-08-02 05:22:24.524614	2012-08-02 10:39:28.229163
9238	1-9238	6	19	6922	80	5649	5649	0	2012-08-02 05:44:44.955067	\N
12418	1-12418	91	50	4430	85	5416	2114	14	2012-08-02 11:20:01.275417	2012-08-02 11:25:11.265631
8942	1-8942	34	6	5840	61	2800	254	0	2012-08-02 05:42:43.839732	2012-08-02 05:48:39.321516
11520	1-11520	78	46	7400	41	731	731	4	2012-08-02 11:14:12.770456	\N
9831	1-9831	49	39	6629	16	1649	1649	2	2012-08-02 05:48:46.64888	\N
8947	1-8947	66	69	1590	79	354	354	2	2012-08-02 05:42:45.894966	\N
10710	1-10710	32	35	5490	57	2572	2572	0	2012-08-02 10:36:25.11021	\N
11016	1-11016	87	59	5880	38	2698	2050	2	2012-08-02 10:39:29.029255	2012-08-02 11:28:14.457753
13750	1-13750	99	43	8515	86	7161	7161	2	2012-08-02 11:28:46.061144	\N
11526	1-11526	52	12	4809	27	1048	1048	16	2012-08-02 11:14:15.101702	\N
9247	1-9247	12	19	9652	82	2186	2186	0	2012-08-02 05:44:48.590224	\N
9458	1-9458	51	34	7093	54	2634	2634	0	2012-08-02 05:46:14.504031	\N
9832	1-9832	17	58	6419	77	1884	1884	7	2012-08-02 05:48:47.06865	\N
12664	1-12664	2	61	6667	29	251	251	4	2012-08-02 11:21:36.957956	\N
11529	1-11529	18	12	6112	43	1281	1281	7	2012-08-02 11:14:16.239862	\N
8959	1-8959	100	74	5228	98	768	768	0	2012-08-02 05:42:50.824202	\N
8960	1-8960	85	27	7849	73	1353	1353	4	2012-08-02 05:42:51.221797	\N
13762	1-13762	47	5	5465	22	2657	2657	6	2012-08-02 11:28:50.734532	\N
13753	1-13753	2	34	7183	47	9496	2214	4	2012-08-02 11:28:47.243134	2012-08-02 11:28:47.243134
9251	1-9251	63	59	8566	12	1078	1078	0	2012-08-02 05:44:50.214679	\N
13765	1-13765	83	51	3161	15	6235	6235	0	2012-08-02 11:28:51.929674	\N
11539	1-11539	34	33	7063	26	1816	1816	0	2012-08-02 11:14:20.051836	\N
12667	1-12667	7	88	5614	7	590	590	2	2012-08-02 11:21:38.504419	\N
13770	1-13770	25	79	8953	40	738	738	16	2012-08-02 11:28:54.349946	\N
11939	1-11939	62	62	3533	54	2296	970	8	2012-08-02 11:16:55.689623	2012-08-02 11:20:28.944385
11542	1-11542	9	36	9684	80	9710	9710	7	2012-08-02 11:14:21.223042	\N
13776	1-13776	75	17	8873	98	1879	1879	0	2012-08-02 11:28:56.327908	\N
9258	1-9258	7	96	2387	75	381	381	10	2012-08-02 05:44:53.153461	\N
8966	1-8966	70	62	6586	34	337	337	0	2012-08-02 05:42:53.741152	\N
10379	1-10379	72	85	4649	52	181	181	2	2012-08-02 10:27:36.228847	\N
13782	1-13782	66	43	2039	70	4572	4572	19	2012-08-02 11:28:58.725178	\N
13788	1-13788	14	51	4165	4	646	646	3	2012-08-02 11:29:01.089492	\N
9263	1-9263	71	67	8776	77	204	204	0	2012-08-02 05:44:55.175206	\N
8976	1-8976	38	25	8877	14	1537	1537	3	2012-08-02 05:42:57.81795	\N
13795	1-13795	42	3	6252	18	7087	5506	5	2012-08-02 11:29:03.420622	2012-08-02 11:29:03.420622
13798	1-13798	54	19	9315	39	2291	2291	18	2012-08-02 11:29:05.023227	\N
9266	1-9266	47	19	7708	72	4617	4617	0	2012-08-02 05:44:56.412444	\N
13785	1-13785	71	92	9992	88	5450	4975	7	2012-08-02 11:28:59.88528	2012-08-02 11:29:11.276055
13779	1-13779	11	23	7117	53	9519	8917	6	2012-08-02 11:28:57.5544	2012-08-02 11:29:21.974274
9267	1-9267	95	11	8073	44	2151	2151	8	2012-08-02 05:44:56.83452	\N
8944	1-8944	38	3	7708	38	6134	2635	0	2012-08-02 05:42:44.668495	2012-08-02 11:29:35.033431
13756	1-13756	41	38	1865	70	3701	22	16	2012-08-02 11:28:48.425457	2012-08-02 11:29:39.376229
9280	1-9280	78	74	5739	71	4756	4756	0	2012-08-02 05:45:02.138033	\N
11955	1-11955	89	31	5919	47	830	830	3	2012-08-02 11:17:02.01007	\N
12460	1-12460	61	22	8510	89	4326	3085	0	2012-08-02 11:20:17.593878	2012-08-02 11:22:41.775281
13595	1-13595	82	72	8419	35	3161	2140	12	2012-08-02 11:27:44.513646	2012-08-02 11:29:56.769677
10095	1-10095	38	53	8364	37	7910	3803	4	2012-08-02 10:21:17.958099	2012-08-02 11:14:39.598138
12465	1-12465	87	9	4913	6	2994	1560	0	2012-08-02 11:20:19.583692	2012-08-02 11:26:43.204062
9482	1-9482	26	1	5671	43	1498	1073	0	2012-08-02 05:46:24.204889	2012-08-02 10:35:34.689888
13603	1-13603	36	34	1997	16	981	981	13	2012-08-02 11:27:47.696205	\N
9537	1-9537	84	90	6517	80	1700	1567	4	2012-08-02 05:46:46.602235	2012-08-02 11:27:16.568012
11981	1-11981	21	40	9666	20	2859	2285	6	2012-08-02 11:17:12.28493	2012-08-02 11:28:53.543122
11964	1-11964	38	12	6310	50	4426	3214	3	2012-08-02 11:17:05.556346	2012-08-02 11:21:23.655203
11522	1-11522	67	43	5459	72	125	125	8	2012-08-02 11:14:13.554787	\N
12659	1-12659	8	49	6353	83	7374	6797	5	2012-08-02 11:21:35.367044	2012-08-02 11:22:35.896828
9476	1-9476	88	96	4207	25	2969	2347	4	2012-08-02 05:46:21.807376	2012-08-02 11:26:24.64072
11015	1-11015	90	49	9759	97	281	281	0	2012-08-02 10:39:28.620817	\N
11952	1-11952	76	12	9535	4	1618	1618	3	2012-08-02 11:17:00.816539	\N
13644	1-13644	100	9	3349	92	9367	5295	3	2012-08-02 11:28:04.359207	2012-08-02 11:29:11.276055
10721	1-10721	69	47	4202	81	1872	114	11	2012-08-02 10:36:29.332679	2012-08-02 11:29:20.4272
9466	1-9466	87	43	640	80	71	71	0	2012-08-02 05:46:17.752397	\N
9505	1-9505	8	18	7911	16	4191	3974	11	2012-08-02 05:46:33.563281	2012-08-02 11:20:04.336464
13638	1-13638	52	27	7568	67	1863	1863	18	2012-08-02 11:28:02.003949	\N
10720	1-10720	59	28	7109	71	8912	3971	2	2012-08-02 10:36:28.95739	2012-08-02 11:29:18.371887
13149	1-13149	43	8	9277	42	2628	2628	0	2012-08-02 11:24:47.265845	\N
10719	1-10719	39	10	6413	40	3532	3532	10	2012-08-02 10:36:28.568674	\N
11019	1-11019	32	91	6401	84	851	851	0	2012-08-02 10:39:30.188802	\N
9738	1-9738	52	58	8215	57	1700	1700	11	2012-08-02 05:48:09.01246	\N
13155	1-13155	54	6	7363	32	2288	2288	9	2012-08-02 11:24:49.597141	\N
9739	1-9739	79	57	9819	12	2507	2507	6	2012-08-02 05:48:09.387971	\N
9471	1-9471	91	35	8631	2	1266	1266	0	2012-08-02 05:46:19.774631	\N
13606	1-13606	45	80	4976	15	6610	6610	0	2012-08-02 11:27:48.877881	\N
12457	1-12457	78	63	5847	9	6982	4065	3	2012-08-02 11:20:16.398432	2012-08-02 11:29:29.884913
9281	1-9281	91	97	7518	77	3417	1508	0	2012-08-02 05:45:02.546892	2012-08-02 11:27:49.662717
9474	1-9474	45	19	2503	56	950	950	9	2012-08-02 05:46:21.011633	\N
10384	1-10384	3	95	7814	7	3995	3995	3	2012-08-02 10:27:38.219795	\N
10099	1-10099	81	50	8589	59	1424	1424	4	2012-08-02 10:21:19.493739	\N
12656	1-12656	35	29	5732	92	7149	4438	2	2012-08-02 11:21:34.184642	2012-08-02 11:27:46.491424
9478	1-9478	37	98	8738	73	251	251	10	2012-08-02 05:46:22.613691	\N
11056	1-11056	84	81	8762	87	1075	1075	10	2012-08-02 11:11:12.503503	\N
9488	1-9488	58	45	5325	10	1488	1488	0	2012-08-02 05:46:26.646311	\N
13164	1-13164	3	29	4492	24	3450	3450	6	2012-08-02 11:24:53.165925	\N
12442	1-12442	38	74	5828	57	4750	4750	0	2012-08-02 11:20:10.546311	\N
10388	1-10388	81	29	4666	46	1247	1247	0	2012-08-02 10:27:39.810538	\N
10389	1-10389	31	78	5750	36	1384	1384	0	2012-08-02 10:27:40.208206	\N
11543	1-11543	95	66	6912	9	2125	2125	0	2012-08-02 11:14:21.609437	\N
10390	1-10390	23	72	9988	27	856	856	0	2012-08-02 10:27:40.605848	\N
13618	1-13618	62	71	9731	100	52	52	12	2012-08-02 11:27:53.584478	\N
13143	1-13143	80	21	8711	83	6771	6122	0	2012-08-02 11:24:43.719406	2012-08-02 11:24:55.132544
11972	1-11972	88	59	5084	1	3672	599	12	2012-08-02 11:17:08.8153	2012-08-02 11:25:01.441006
12445	1-12445	31	43	1077	77	236	236	4	2012-08-02 11:20:11.717372	\N
10107	1-10107	87	53	7114	58	48	48	0	2012-08-02 10:21:22.786312	\N
9462	1-9462	97	59	9392	90	5567	753	8	2012-08-02 05:46:16.117223	2012-08-02 10:22:11.737977
11528	1-11528	78	79	9371	6	6596	2910	4	2012-08-02 11:14:15.863933	2012-08-02 11:29:11.695894
13158	1-13158	87	55	9868	11	4914	946	6	2012-08-02 11:24:50.768568	2012-08-02 11:29:35.795893
9501	1-9501	53	64	9965	30	813	813	3	2012-08-02 05:46:31.917115	\N
11987	1-11987	27	28	8352	100	7964	674	0	2012-08-02 11:17:14.706071	2012-08-02 11:27:59.982053
12454	1-12454	25	100	7710	27	7257	3400	12	2012-08-02 11:20:15.195286	2012-08-02 11:27:59.982053
9504	1-9504	36	31	7018	88	3872	3872	15	2012-08-02 05:46:33.156067	\N
13641	1-13641	25	65	9139	20	6264	6264	7	2012-08-02 11:28:03.177149	\N
13647	1-13647	58	16	6554	44	2179	2179	0	2012-08-02 11:28:05.496852	\N
9507	1-9507	35	23	5960	38	1426	1426	0	2012-08-02 05:46:34.358862	\N
12410	1-12410	21	35	8473	18	4127	280	4	2012-08-02 11:19:58.193368	2012-08-02 11:28:41.71681
9517	1-9517	88	58	7764	53	1490	1490	5	2012-08-02 05:46:38.415577	\N
9526	1-9526	53	44	3810	62	277	277	0	2012-08-02 05:46:42.050739	\N
9534	1-9534	40	67	8653	12	1696	1696	0	2012-08-02 05:46:45.298476	\N
9535	1-9535	1	18	8970	54	2705	2705	7	2012-08-02 05:46:45.772656	\N
9539	1-9539	7	36	7161	67	23	23	5	2012-08-02 05:46:47.419818	\N
9545	1-9545	84	79	8820	87	128	128	2	2012-08-02 05:46:49.828648	\N
10182	1-10182	14	53	6876	94	1722	722	13	2012-08-02 10:22:16.422509	2012-08-02 11:16:09.277745
10762	1-10762	78	59	7999	57	5693	4088	10	2012-08-02 10:37:49.149624	2012-08-02 11:16:07.256137
9752	1-9752	20	27	8257	60	8787	8566	13	2012-08-02 05:48:14.591905	2012-08-02 11:16:42.724536
9583	1-9583	36	49	9903	34	5402	5145	0	2012-08-02 05:47:05.230429	2012-08-02 05:49:46.31251
9569	1-9569	98	73	8228	68	4179	3220	0	2012-08-02 05:46:59.496352	2012-08-02 11:29:56.769677
9760	1-9760	81	23	7195	53	123	123	4	2012-08-02 05:48:17.862444	\N
12668	1-12668	2	4	9358	99	2362	2065	4	2012-08-02 11:21:38.915383	2012-08-02 11:28:22.314052
9821	1-9821	34	62	8872	7	1572	1572	0	2012-08-02 05:48:42.580965	\N
13599	1-13599	26	73	2743	91	4622	1360	11	2012-08-02 11:27:46.104686	2012-08-02 11:29:45.743019
9817	1-9817	89	88	9275	69	4104	1016	0	2012-08-02 05:48:40.957191	2012-08-02 11:18:59.000354
9578	1-9578	5	24	5221	55	1495	1099	4	2012-08-02 05:47:03.197727	2012-08-02 10:35:36.668705
11524	1-11524	52	76	4677	73	549	486	7	2012-08-02 11:14:14.328423	2012-08-02 11:20:51.098418
9794	1-9794	12	8	7301	93	5268	3957	0	2012-08-02 05:48:31.653566	2012-08-02 11:16:09.277745
12838	1-12838	74	45	5449	95	2891	1361	17	2012-08-02 11:22:45.300769	2012-08-02 11:24:19.939707
10121	1-10121	13	3	9436	43	6137	2092	10	2012-08-02 10:21:28.305916	2012-08-02 11:18:20.60287
9565	1-9565	6	24	9983	85	164	164	0	2012-08-02 05:46:57.883073	\N
11020	1-11020	74	94	6825	19	1959	615	5	2012-08-02 10:39:30.583713	2012-08-02 11:19:25.352516
9791	1-9791	81	3	6799	23	3500	575	0	2012-08-02 05:48:30.416137	2012-08-02 11:26:10.938883
9572	1-9572	22	36	3897	15	2454	2237	0	2012-08-02 05:47:00.73371	2012-08-02 10:25:24.858696
9567	1-9567	50	36	3179	40	488	488	0	2012-08-02 05:46:58.700561	\N
12671	1-12671	62	6	9556	66	7371	2643	5	2012-08-02 11:21:40.075695	2012-08-02 11:27:06.710134
9799	1-9799	16	1	5846	54	5852	2010	0	2012-08-02 05:48:33.653308	2012-08-02 10:25:47.691152
10129	1-10129	12	32	9800	77	3164	2772	0	2012-08-02 10:21:31.522888	2012-08-02 11:23:25.731152
13602	1-13602	47	77	7490	10	4408	1806	6	2012-08-02 11:27:47.287245	2012-08-02 11:30:05.056251
9822	1-9822	32	83	6395	50	2165	698	6	2012-08-02 05:48:42.978558	2012-08-02 11:25:06.92148
9758	1-9758	46	81	9327	52	5898	667	2	2012-08-02 05:48:17.033821	2012-08-02 11:23:58.228472
11095	1-11095	15	31	4899	21	4908	3539	2	2012-08-02 11:11:27.377338	2012-08-02 11:17:44.152642
12662	1-12662	96	64	8254	74	2130	2130	4	2012-08-02 11:21:36.548914	\N
9765	1-9765	89	58	8994	40	5501	4902	8	2012-08-02 05:48:19.886302	2012-08-02 11:30:05.840752
13596	1-13596	67	23	972	9	66	66	12	2012-08-02 11:27:44.900256	\N
11538	1-11538	61	14	1674	1	268	268	0	2012-08-02 11:14:19.653499	\N
12417	1-12417	46	66	7811	13	7724	4776	5	2012-08-02 11:20:00.877771	2012-08-02 11:30:16.762619
10727	1-10727	93	88	2436	58	428	428	0	2012-08-02 10:36:31.688452	\N
10396	1-10396	47	55	7197	34	3129	3129	22	2012-08-02 10:27:42.981675	\N
10125	1-10125	17	46	8956	61	3220	1841	0	2012-08-02 10:21:29.897293	2012-08-02 11:11:58.737639
9751	1-9751	45	15	8663	38	1369	1369	5	2012-08-02 05:48:14.194344	\N
10103	1-10103	77	29	6593	83	7616	3222	5	2012-08-02 10:21:21.206778	2012-08-02 10:37:47.13649
9769	1-9769	14	44	4878	6	1353	429	18	2012-08-02 05:48:21.444	2012-08-02 10:38:09.567524
9816	1-9816	91	32	7248	30	1543	789	12	2012-08-02 05:48:40.548251	2012-08-02 11:24:44.12585
9579	1-9579	34	11	7609	51	3117	3117	6	2012-08-02 05:47:03.617629	\N
11022	1-11022	96	26	1724	16	1322	1322	4	2012-08-02 10:39:31.389966	\N
9580	1-9580	5	36	8406	6	4025	4025	0	2012-08-02 05:47:04.014891	\N
10397	1-10397	75	3	7797	76	1356	1356	5	2012-08-02 10:27:43.379425	\N
11060	1-11060	94	41	9456	49	1301	1301	6	2012-08-02 11:11:13.686354	\N
11970	1-11970	82	89	8745	59	3413	3413	0	2012-08-02 11:17:08.009216	\N
10118	1-10118	28	58	3840	92	881	881	4	2012-08-02 10:21:27.156799	\N
10543	1-10543	88	36	5092	16	6937	2583	0	2012-08-02 10:35:32.40585	2012-08-02 11:22:42.548342
11069	1-11069	66	31	7946	70	1191	1191	0	2012-08-02 11:11:17.23675	\N
11535	1-11535	74	19	4726	22	8317	4302	14	2012-08-02 11:14:18.902423	2012-08-02 11:26:13.358762
9798	1-9798	92	86	8954	80	2868	2768	0	2012-08-02 05:48:33.266627	2012-08-02 10:21:26.747339
11963	1-11963	20	95	5957	16	3848	2229	0	2012-08-02 11:17:05.180704	2012-08-02 11:28:19.186746
11073	1-11073	39	55	7199	34	140	140	22	2012-08-02 11:11:18.408994	\N
9767	1-9767	24	28	2350	21	1079	1079	5	2012-08-02 05:48:20.670698	\N
9768	1-9768	4	69	8046	94	2018	2018	17	2012-08-02 05:48:21.057247	\N
11979	1-11979	6	80	7897	52	5930	3696	2	2012-08-02 11:17:11.500124	2012-08-02 11:28:06.634738
10179	1-10179	68	28	9346	30	6072	2778	15	2012-08-02 10:22:15.26259	2012-08-02 11:27:59.982053
12835	1-12835	41	24	4373	14	4999	3210	0	2012-08-02 11:22:44.140924	2012-08-02 11:28:12.888559
9773	1-9773	56	41	5793	7	718	718	7	2012-08-02 05:48:23.045895	\N
11076	1-11076	45	89	9532	87	687	687	0	2012-08-02 11:11:19.99271	\N
9774	1-9774	30	18	8825	88	3420	3420	0	2012-08-02 05:48:23.443536	\N
10726	1-10726	27	20	9756	28	3026	1511	6	2012-08-02 10:36:31.310778	2012-08-02 11:29:02.658287
13150	1-13150	47	63	6376	3	7753	1679	7	2012-08-02 11:24:47.641588	2012-08-02 11:29:03.420622
11089	1-11089	30	66	8217	58	2291	2291	4	2012-08-02 11:11:25.000463	\N
11092	1-11092	93	48	7745	94	2027	2027	0	2012-08-02 11:11:26.182702	\N
11098	1-11098	67	47	6999	7	4949	4949	5	2012-08-02 11:11:28.526611	\N
10181	1-10181	21	78	8069	25	3875	3875	0	2012-08-02 10:22:16.036064	\N
9790	1-9790	47	65	5952	25	1301	1301	9	2012-08-02 05:48:29.996509	\N
9800	1-9800	17	93	6617	63	127	127	0	2012-08-02 05:48:34.084356	\N
9809	1-9809	9	96	8143	1	52	52	0	2012-08-02 05:48:37.708469	\N
9803	1-9803	31	68	9231	15	8476	8476	0	2012-08-02 05:48:35.288656	\N
10736	1-10736	90	64	4073	15	1574	1574	12	2012-08-02 10:36:35.181906	\N
12836	1-12836	39	11	6964	16	9812	9477	17	2012-08-02 11:22:44.516502	2012-08-02 11:26:06.593594
10207	1-10207	89	39	6278	87	9938	4735	0	2012-08-02 10:22:26.13544	2012-08-02 11:21:32.229149
11549	1-11549	62	19	8405	42	7711	3695	0	2012-08-02 11:14:23.898586	2012-08-02 11:14:53.478813
9847	1-9847	100	1	1999	91	788	788	15	2012-08-02 05:48:53.158513	\N
10733	1-10733	29	30	9752	34	3137	3137	6	2012-08-02 10:36:33.999711	\N
9873	1-9873	42	56	8279	87	9200	5060	0	2012-08-02 05:49:03.762156	2012-08-02 11:27:22.677663
10215	1-10215	32	21	4010	90	804	51	9	2012-08-02 10:22:29.240161	2012-08-02 11:11:32.795243
10192	1-10192	44	36	9057	94	5672	5136	14	2012-08-02 10:22:20.300253	2012-08-02 11:13:00.408955
11736	1-11736	39	75	7795	22	8069	5376	4	2012-08-02 11:15:36.571617	2012-08-02 11:26:13.358762
12848	1-12848	28	67	4193	7	4723	2781	0	2012-08-02 11:22:49.245482	2012-08-02 11:23:01.036268
10209	1-10209	75	41	8904	52	4948	4948	10	2012-08-02 10:22:26.919842	\N
10785	1-10785	55	1	9706	78	5859	3247	3	2012-08-02 10:37:57.469118	2012-08-02 11:21:37.75325
10728	1-10728	80	28	8787	83	4552	4552	4	2012-08-02 10:36:32.087845	\N
9872	1-9872	4	14	2676	37	1372	875	2	2012-08-02 05:49:03.375435	2012-08-02 11:27:31.75991
9851	1-9851	90	33	7647	37	637	637	0	2012-08-02 05:48:54.786803	\N
10738	1-10738	45	1	9729	14	9681	4964	0	2012-08-02 10:36:35.955338	2012-08-02 11:11:50.892453
9852	1-9852	12	95	8242	85	3608	2809	0	2012-08-02 05:48:55.195461	2012-08-02 11:18:40.845599
7433	1-7433	13	89	5323	83	4212	1312	6	2012-08-02 05:32:26.592424	2012-08-02 11:30:32.230146
13165	1-13165	79	72	8223	10	5812	1332	0	2012-08-02 11:24:53.564017	2012-08-02 11:30:02.736062
10184	1-10184	68	30	6184	54	1935	1935	0	2012-08-02 10:22:17.195871	\N
12379	1-12379	9	98	9363	1	5543	356	7	2012-08-02 11:19:46.57956	2012-08-02 11:25:24.204162
10185	1-10185	48	59	7606	18	1839	1839	0	2012-08-02 10:22:17.582679	\N
10204	1-10204	88	9	7636	69	3923	2314	2	2012-08-02 10:22:24.942537	2012-08-02 11:13:53.797887
9855	1-9855	20	75	6047	96	3727	1457	0	2012-08-02 05:48:56.414794	2012-08-02 05:48:57.642823
13159	1-13159	23	68	6372	70	4941	4941	5	2012-08-02 11:24:51.177139	\N
10188	1-10188	67	78	8070	32	4106	3530	0	2012-08-02 10:22:18.75347	2012-08-02 10:27:32.27355
9859	1-9859	87	86	7851	98	26	26	0	2012-08-02 05:48:58.0525	\N
10406	1-10406	52	10	8903	1	2603	2603	9	2012-08-02 10:27:46.926067	\N
13162	1-13162	93	78	8982	25	2659	2659	0	2012-08-02 11:24:52.359806	\N
13594	1-13594	80	77	8371	32	2089	2089	0	2012-08-02 11:27:44.125156	\N
10140	1-10140	19	66	9983	4	5051	4214	0	2012-08-02 10:21:35.857279	2012-08-02 10:27:44.572949
11977	1-11977	72	53	4136	91	3365	220	0	2012-08-02 11:17:10.748895	2012-08-02 11:23:28.902649
13168	1-13168	15	46	4792	13	1408	1408	7	2012-08-02 11:24:54.734609	\N
13601	1-13601	71	100	7797	90	2855	2855	6	2012-08-02 11:27:46.878444	\N
11569	1-11569	70	78	8476	52	7927	4432	7	2012-08-02 11:14:31.544414	2012-08-02 11:16:59.645095
11552	1-11552	14	51	7348	64	1650	1650	0	2012-08-02 11:14:25.047982	\N
13173	1-13173	69	10	9455	76	1245	1245	9	2012-08-02 11:24:56.679265	\N
12431	1-12431	56	4	8418	36	2000	2000	0	2012-08-02 11:20:06.656847	\N
9867	1-9867	49	39	9195	41	1513	1513	6	2012-08-02 05:49:01.306571	\N
11558	1-11558	14	18	8153	42	2360	2360	5	2012-08-02 11:14:27.357187	\N
9862	1-9862	44	29	9677	88	5707	2761	2	2012-08-02 05:48:59.270511	2012-08-02 11:28:30.535322
11561	1-11561	85	94	9908	96	892	892	15	2012-08-02 11:14:28.494727	\N
11983	1-11983	43	65	6457	82	839	839	0	2012-08-02 11:17:13.091025	\N
10404	1-10404	8	28	7047	46	2338	2338	0	2012-08-02 10:27:46.152735	\N
13604	1-13604	67	91	4877	55	4958	2564	6	2012-08-02 11:27:48.082313	2012-08-02 11:28:45.674267
13607	1-13607	63	12	2477	50	1997	1044	8	2012-08-02 11:27:49.27552	2012-08-02 11:28:50.33668
13176	1-13176	56	72	3823	25	448	448	5	2012-08-02 11:24:57.905656	\N
10423	1-10423	82	29	8742	91	3171	1899	0	2012-08-02 10:27:53.633159	2012-08-02 11:28:53.156625
11730	1-11730	79	80	7676	64	3728	2764	0	2012-08-02 11:15:34.24986	2012-08-02 11:22:07.015042
10134	1-10134	91	57	6422	85	3619	2309	0	2012-08-02 10:21:33.492641	2012-08-02 11:29:17.18967
10729	1-10729	94	27	3389	16	3412	3170	2	2012-08-02 10:36:32.463271	2012-08-02 11:29:18.780703
10191	1-10191	92	47	9945	87	5840	5840	5	2012-08-02 10:22:19.924652	\N
11575	1-11575	46	89	7153	83	4750	4750	11	2012-08-02 11:14:33.864213	\N
12446	1-12446	59	11	6868	44	2778	2778	8	2012-08-02 11:20:12.104343	\N
12833	1-12833	53	94	6872	89	5838	1977	6	2012-08-02 11:22:43.33328	2012-08-02 11:29:35.033431
10206	1-10206	29	85	3436	25	610	375	10	2012-08-02 10:22:25.748573	2012-08-02 11:25:04.93251
11584	1-11584	57	50	1552	82	567	567	3	2012-08-02 11:14:37.355792	\N
10210	1-10210	8	79	6529	40	5654	5654	11	2012-08-02 10:22:27.30679	\N
12842	1-12842	50	84	8549	12	1377	1377	7	2012-08-02 11:22:46.869729	\N
12845	1-12845	48	84	9713	49	1442	1442	0	2012-08-02 11:22:48.052097	\N
12851	1-12851	40	23	6430	57	3605	3605	5	2012-08-02 11:22:50.405615	\N
12854	1-12854	1	98	8039	41	4012	4012	0	2012-08-02 11:22:51.621132	\N
12857	1-12857	86	63	8099	77	4438	4438	18	2012-08-02 11:22:52.769653	\N
11727	1-11727	71	70	7222	40	4220	4220	6	2012-08-02 11:15:33.065183	\N
10789	1-10789	52	4	9215	48	2310	2310	0	2012-08-02 10:37:59.037943	\N
10795	1-10795	99	74	7705	86	5479	5479	0	2012-08-02 10:38:01.744886	\N
10422	1-10422	42	64	3157	96	717	717	8	2012-08-02 10:27:53.24651	\N
10424	1-10424	76	32	4174	91	385	385	0	2012-08-02 10:27:54.0308	\N
10220	1-10220	19	99	1489	81	122	122	3	2012-08-02 10:22:31.228988	\N
10426	1-10426	38	49	8488	93	443	443	0	2012-08-02 10:27:54.815238	\N
10518	1-10518	96	43	4244	25	1073	779	0	2012-08-02 10:32:04.020375	2012-08-02 11:25:26.164453
10513	1-10513	85	19	7609	77	3751	2008	0	2012-08-02 10:32:02.064377	2012-08-02 11:26:54.810076
10456	1-10456	1	79	5866	40	5142	1608	11	2012-08-02 10:31:39.427726	2012-08-02 11:24:29.949566
11996	1-11996	53	39	7438	4	5193	3612	0	2012-08-02 11:17:18.595408	2012-08-02 11:21:30.659848
12449	1-12449	23	50	9398	6	6751	3355	7	2012-08-02 11:20:13.24606	2012-08-02 11:27:49.662717
12849	1-12849	51	50	2593	19	1002	451	4	2012-08-02 11:22:49.632303	2012-08-02 11:24:12.922075
10454	1-10454	68	56	2459	95	152	152	6	2012-08-02 10:31:38.654517	\N
10455	1-10455	79	70	7127	28	143	143	0	2012-08-02 10:31:39.040989	\N
11741	1-11741	40	36	3578	84	1923	1923	0	2012-08-02 11:15:38.549584	\N
13167	1-13167	66	81	9819	67	5931	3688	7	2012-08-02 11:24:53.961251	2012-08-02 11:27:05.118969
13616	1-13616	55	31	8654	85	2911	2911	5	2012-08-02 11:27:52.810919	\N
13613	1-13613	17	12	9030	70	4760	4760	0	2012-08-02 11:27:51.640212	\N
10473	1-10473	32	58	7363	64	2528	1498	0	2012-08-02 10:31:46.212975	2012-08-02 11:12:54.686138
11994	1-11994	8	81	6384	4	1537	1537	7	2012-08-02 11:17:17.401787	\N
10511	1-10511	54	33	9427	47	4596	3267	7	2012-08-02 10:32:01.269003	2012-08-02 11:20:52.292349
10460	1-10460	100	31	9807	10	3461	3461	0	2012-08-02 10:31:40.986946	\N
10482	1-10482	5	51	7749	96	1565	1565	2	2012-08-02 10:31:49.782117	\N
10452	1-10452	33	13	9761	82	6641	4541	6	2012-08-02 10:31:37.879132	2012-08-02 11:22:05.876967
10483	1-10483	90	28	8647	49	1047	1047	0	2012-08-02 10:31:50.191039	\N
12450	1-12450	49	55	2772	3	950	766	6	2012-08-02 11:20:13.658748	2012-08-02 11:20:21.119568
12034	1-12034	31	100	7260	36	2915	2046	4	2012-08-02 11:17:33.134378	2012-08-02 11:22:19.576837
12846	1-12846	43	47	6536	16	5477	5074	0	2012-08-02 11:22:48.450151	2012-08-02 11:22:48.450151
13619	1-13619	14	59	6012	54	7250	7250	7	2012-08-02 11:27:53.971589	\N
13174	1-13174	80	26	5777	43	458	458	14	2012-08-02 11:24:57.099272	\N
12444	1-12444	5	76	6677	30	1375	1375	10	2012-08-02 11:20:11.342192	\N
10478	1-10478	60	32	6706	40	6369	3110	12	2012-08-02 10:31:48.201667	2012-08-02 11:28:49.552546
13621	1-13621	40	87	9706	15	6175	6175	6	2012-08-02 11:27:55.10907	\N
12005	1-12005	51	96	7014	88	2341	1635	10	2012-08-02 11:17:21.732591	2012-08-02 11:20:12.869762
10469	1-10469	41	57	8520	92	3211	3211	3	2012-08-02 10:31:44.588728	\N
10471	1-10471	45	36	8192	50	8070	8070	0	2012-08-02 10:31:45.428593	\N
11026	1-11026	96	32	2268	55	650	650	0	2012-08-02 10:39:32.990179	\N
13630	1-13630	45	23	7139	51	9390	8819	0	2012-08-02 11:27:59.164746	2012-08-02 11:28:34.42507
11571	1-11571	42	83	6235	88	1891	1891	5	2012-08-02 11:14:32.328691	\N
9098	1-9098	91	66	6770	50	2970	1982	0	2012-08-02 05:43:47.914984	2012-08-02 10:37:43.844368
10475	1-10475	38	62	9070	29	675	675	0	2012-08-02 10:31:47.019472	\N
13626	1-13626	95	53	5420	21	8791	7891	12	2012-08-02 11:27:57.176007	2012-08-02 11:27:57.176007
10500	1-10500	5	18	8090	10	5080	1502	9	2012-08-02 10:31:56.926196	2012-08-02 11:29:15.596151
10485	1-10485	32	88	9641	51	3114	776	8	2012-08-02 10:31:50.976089	2012-08-02 11:18:28.590617
12884	1-12884	9	50	5400	32	4470	2497	0	2012-08-02 11:23:03.356209	2012-08-02 11:29:12.082503
12020	1-12020	72	21	8763	44	368	368	0	2012-08-02 11:17:27.688105	\N
10487	1-10487	57	18	9018	56	495	495	3	2012-08-02 10:31:51.783447	\N
10490	1-10490	36	47	6297	93	2882	2749	7	2012-08-02 10:31:52.965455	2012-08-02 11:14:35.77558
13636	1-13636	48	26	9812	37	2439	2439	4	2012-08-02 11:28:01.23104	\N
10489	1-10489	19	12	8359	67	1858	1858	0	2012-08-02 10:31:52.578757	\N
13639	1-13639	88	93	6594	49	706	706	6	2012-08-02 11:28:02.390516	\N
11583	1-11583	41	28	9389	14	4414	4414	0	2012-08-02 11:14:36.958133	\N
13642	1-13642	68	54	6031	49	1224	1224	4	2012-08-02 11:28:03.563441	\N
12864	1-12864	67	12	9729	87	7235	7235	3	2012-08-02 11:22:55.532096	\N
13645	1-13645	54	21	9285	24	7661	7661	0	2012-08-02 11:28:04.745501	\N
13648	1-13648	10	99	9388	23	99	99	23	2012-08-02 11:28:05.883824	\N
11744	1-11744	32	11	9236	77	5800	5800	0	2012-08-02 11:15:39.742232	\N
13650	1-13650	55	61	5849	47	5755	5755	4	2012-08-02 11:28:07.010335	\N
13652	1-13652	34	3	7410	15	8904	8904	3	2012-08-02 11:28:07.805748	\N
10537	1-10537	28	23	9436	45	7403	222	5	2012-08-02 10:32:11.571656	2012-08-02 11:22:57.509065
13655	1-13655	2	65	574	85	221	221	0	2012-08-02 11:28:08.590351	\N
11747	1-11747	72	26	5272	61	450	450	3	2012-08-02 11:15:40.892139	\N
12881	1-12881	72	43	8129	13	856	856	5	2012-08-02 11:23:02.196063	\N
12887	1-12887	67	17	8966	13	1350	1350	0	2012-08-02 11:23:04.505373	\N
10507	1-10507	50	58	6946	48	2214	2214	7	2012-08-02 10:31:59.698999	\N
10517	1-10517	51	5	5292	66	1387	1387	0	2012-08-02 10:32:03.633473	\N
10521	1-10521	88	51	7821	59	1209	1209	6	2012-08-02 10:32:05.203297	\N
10525	1-10525	69	75	9059	7	4669	4669	0	2012-08-02 10:32:06.795342	\N
10528	1-10528	13	65	8001	88	3747	3747	2	2012-08-02 10:32:07.969557	\N
10530	1-10530	88	62	2879	40	2577	2577	12	2012-08-02 10:32:08.753832	\N
10531	1-10531	73	100	8666	94	1487	1487	0	2012-08-02 10:32:09.140402	\N
12678	1-12678	85	15	8331	37	7677	18	8	2012-08-02 11:21:43.291045	2012-08-02 11:30:06.614302
13614	1-13614	69	50	6167	33	2701	670	0	2012-08-02 11:27:52.026602	2012-08-02 11:28:42.910342
12473	1-12473	62	67	6294	63	2670	1020	3	2012-08-02 11:20:22.716449	2012-08-02 11:30:11.709306
10750	1-10750	44	100	2015	41	313	313	11	2012-08-02 10:37:44.69502	\N
10825	1-10825	23	29	5394	21	3152	165	0	2012-08-02 10:38:13.489964	2012-08-02 11:19:16.72384
11749	1-11749	71	7	4494	2	3430	76	5	2012-08-02 11:15:41.643679	2012-08-02 11:30:21.901036
13180	1-13180	49	49	7705	46	3949	969	7	2012-08-02 11:24:59.485484	2012-08-02 11:29:15.220572
10835	1-10835	57	46	2975	44	1878	516	0	2012-08-02 10:38:17.436004	2012-08-02 11:24:19.939707
11746	1-11746	45	28	6042	22	7231	13	4	2012-08-02 11:15:40.516666	2012-08-02 11:26:15.336549
10816	1-10816	90	24	6706	78	3928	1365	0	2012-08-02 10:38:09.954236	2012-08-02 11:30:29.07869
13611	1-13611	37	11	5129	82	5772	5772	12	2012-08-02 11:27:50.446679	\N
8910	1-8910	80	35	2358	42	1409	44	0	2012-08-02 05:42:30.781398	2012-08-02 11:16:56.916231
11585	1-11585	2	62	7671	87	5041	5041	3	2012-08-02 11:14:37.74243	\N
12032	1-12032	47	14	9345	18	3467	1096	3	2012-08-02 11:17:32.372683	2012-08-02 11:28:43.735007
10844	1-10844	50	17	8919	99	1483	1269	3	2012-08-02 10:38:20.960504	2012-08-02 10:39:28.229163
12461	1-12461	99	50	6058	72	2529	2529	7	2012-08-02 11:20:18.023168	\N
13193	1-13193	52	72	8450	25	5144	3844	5	2012-08-02 11:25:04.556839	2012-08-02 11:29:17.18967
11025	1-11025	4	61	3475	98	508	508	0	2012-08-02 10:39:32.592849	\N
12026	1-12026	13	50	4456	15	3765	285	4	2012-08-02 11:17:30.383606	2012-08-02 11:27:33.74851
10757	1-10757	83	78	4358	11	1133	1133	7	2012-08-02 10:37:47.533939	\N
11027	1-11027	85	80	9095	30	134	134	2	2012-08-02 10:39:33.365902	\N
11567	1-11567	38	89	8206	46	1134	1134	5	2012-08-02 11:14:30.395219	\N
12024	1-12024	43	36	8850	31	6326	3533	6	2012-08-02 11:17:29.223562	2012-08-02 11:22:35.896828
12046	1-12046	57	97	5625	37	7852	5935	7	2012-08-02 11:17:37.797163	2012-08-02 11:21:35.764971
12038	1-12038	71	38	9475	40	5817	5228	3	2012-08-02 11:17:34.648065	2012-08-02 11:24:48.415244
10771	1-10771	50	47	7613	20	1819	1819	10	2012-08-02 10:37:52.298371	\N
13177	1-13177	28	42	7817	44	6329	5365	0	2012-08-02 11:24:58.701109	2012-08-02 11:27:41.683599
10808	1-10808	78	74	5336	97	3725	951	0	2012-08-02 10:38:07.225067	2012-08-02 11:29:36.591718
10761	1-10761	64	66	8921	42	5533	2780	2	2012-08-02 10:37:48.729597	2012-08-02 11:11:49.720618
10843	1-10843	64	76	7968	34	4479	4415	0	2012-08-02 10:38:20.573979	2012-08-02 11:16:43.499661
12464	1-12464	93	18	5409	14	7052	7052	4	2012-08-02 11:20:19.195995	\N
10801	1-10801	89	85	6072	71	1922	1922	0	2012-08-02 10:38:04.109057	\N
10780	1-10780	8	47	8642	91	1148	1148	9	2012-08-02 10:37:55.90014	\N
10765	1-10765	24	69	9368	70	7533	1717	2	2012-08-02 10:37:50.344874	2012-08-02 11:27:52.435896
10782	1-10782	5	48	8806	9	840	840	9	2012-08-02 10:37:56.706645	\N
12006	1-12006	89	33	8638	5	5858	5858	5	2012-08-02 11:17:22.174995	\N
12684	1-12684	68	92	8391	2	1839	1839	0	2012-08-02 11:21:45.258574	\N
10794	1-10794	89	12	8639	43	4942	2548	6	2012-08-02 10:38:01.358202	2012-08-02 11:15:29.184735
12009	1-12009	45	70	5406	56	1014	1014	0	2012-08-02 11:17:23.356894	\N
11591	1-11591	55	91	8116	14	2400	2400	5	2012-08-02 11:14:39.984816	\N
10819	1-10819	4	79	8226	14	5247	5247	0	2012-08-02 10:38:11.125396	\N
11992	1-11992	18	23	5746	79	1383	1383	4	2012-08-02 11:17:16.628269	\N
10787	1-10787	41	23	8881	54	3171	3171	11	2012-08-02 10:37:58.617986	\N
12690	1-12690	2	47	4317	53	1368	1368	3	2012-08-02 11:21:47.633994	\N
12012	1-12012	17	38	5347	93	1071	1071	0	2012-08-02 11:17:24.527722	\N
13620	1-13620	90	16	8069	26	419	419	7	2012-08-02 11:27:54.35783	\N
10791	1-10791	93	88	8152	76	3118	3118	0	2012-08-02 10:38:00.186851	\N
12693	1-12693	76	91	7247	66	323	323	0	2012-08-02 11:21:48.794258	\N
13623	1-13623	56	3	8661	80	4808	4808	0	2012-08-02 11:27:55.496138	\N
10796	1-10796	27	50	4828	89	819	819	3	2012-08-02 10:38:02.153805	\N
10797	1-10797	9	15	6751	48	434	434	0	2012-08-02 10:38:02.528988	\N
10799	1-10799	3	45	7911	14	2097	2097	7	2012-08-02 10:38:03.313386	\N
13617	1-13617	6	85	6754	25	4053	354	11	2012-08-02 11:27:53.19788	2012-08-02 11:29:17.18967
10804	1-10804	54	82	9572	72	2412	2412	3	2012-08-02 10:38:05.258423	\N
13190	1-13190	92	58	7936	96	1704	1704	0	2012-08-02 11:25:03.783691	\N
12035	1-12035	36	64	8954	82	2877	2877	9	2012-08-02 11:17:33.521052	\N
12043	1-12043	100	48	8594	87	4551	4551	4	2012-08-02 11:17:36.615053	\N
10759	1-10759	41	38	1221	96	338	307	2	2012-08-02 10:37:47.931773	2012-08-02 11:25:11.265631
11570	1-11570	75	61	2486	18	655	28	0	2012-08-02 11:14:31.942406	2012-08-02 11:26:20.242499
10826	1-10826	9	93	4990	23	1063	941	7	2012-08-02 10:38:13.889614	2012-08-02 11:26:26.20944
10812	1-10812	48	19	1990	97	519	519	8	2012-08-02 10:38:08.385554	\N
10818	1-10818	93	19	6698	90	969	969	8	2012-08-02 10:38:10.738782	\N
10830	1-10830	62	69	2930	47	950	950	0	2012-08-02 10:38:15.502604	\N
10832	1-10832	3	94	7553	6	1081	1081	6	2012-08-02 10:38:16.275908	\N
10846	1-10846	74	3	5650	80	2918	2918	0	2012-08-02 10:38:21.755866	\N
10848	1-10848	49	66	3206	30	354	354	0	2012-08-02 10:38:22.529135	\N
10857	1-10857	51	27	7846	46	2820	2820	0	2012-08-02 10:38:26.087093	\N
13199	1-13199	26	31	9173	10	9226	804	0	2012-08-02 11:25:06.126214	2012-08-02 11:29:18.780703
12047	1-12047	30	72	8383	34	1458	1458	0	2012-08-02 11:17:38.570573	\N
11160	1-11160	32	21	5436	95	1140	601	3	2012-08-02 11:11:52.882179	2012-08-02 11:20:58.97991
11215	1-11215	83	18	7550	80	2952	2952	0	2012-08-02 11:12:14.384605	\N
13637	1-13637	56	44	6947	7	7373	5679	0	2012-08-02 11:28:01.617159	2012-08-02 11:30:40.103243
11131	1-11131	94	15	3193	60	419	419	3	2012-08-02 11:11:41.381685	\N
12683	1-12683	34	75	8804	47	5383	3973	4	2012-08-02 11:21:44.485469	2012-08-02 11:27:20.943618
12710	1-12710	53	86	7753	57	7337	5344	0	2012-08-02 11:21:55.390969	2012-08-02 11:28:44.878308
11100	1-11100	35	93	8758	27	2551	2215	4	2012-08-02 11:11:29.301381	2012-08-02 11:27:13.362106
11633	1-11633	37	100	9668	76	4936	4337	0	2012-08-02 11:14:56.196427	2012-08-02 11:21:17.003001
13184	1-13184	85	89	5271	82	3605	3605	10	2012-08-02 11:25:01.054339	\N
11169	1-11169	90	76	8011	72	5107	4314	0	2012-08-02 11:11:56.384704	2012-08-02 11:13:13.590162
11601	1-11601	72	28	8257	10	2658	2420	0	2012-08-02 11:14:43.831281	2012-08-02 11:17:03.998477
12695	1-12695	39	59	7502	37	7188	3427	0	2012-08-02 11:21:49.556997	2012-08-02 11:30:16.762619
12492	1-12492	50	39	3783	98	3922	3922	0	2012-08-02 11:20:30.117699	\N
12484	1-12484	25	67	6185	77	1875	1875	5	2012-08-02 11:20:26.985654	\N
11662	1-11662	77	8	4172	17	4431	1081	6	2012-08-02 11:15:07.488583	2012-08-02 11:27:33.74851
11032	1-11032	31	49	9911	70	840	840	0	2012-08-02 10:39:35.321538	\N
10855	1-10855	35	79	4289	92	3738	1305	7	2012-08-02 10:38:25.689373	2012-08-02 11:24:51.961695
11613	1-11613	24	83	3231	47	640	640	0	2012-08-02 11:14:48.462703	\N
11094	1-11094	67	29	1844	72	708	635	8	2012-08-02 11:11:26.990788	2012-08-02 11:13:30.75082
11074	1-11074	66	58	3027	92	2073	2073	4	2012-08-02 11:11:19.194439	\N
13187	1-13187	31	54	8933	16	4255	4255	9	2012-08-02 11:25:02.237074	\N
11080	1-11080	48	39	7280	70	165	165	4	2012-08-02 11:11:21.55088	\N
13632	1-13632	65	61	6518	41	4839	4839	0	2012-08-02 11:27:59.573268	\N
12019	1-12019	47	40	9195	83	1987	1987	0	2012-08-02 11:17:27.29025	\N
11139	1-11139	91	31	7008	93	78	78	10	2012-08-02 11:11:44.592354	\N
11142	1-11142	67	31	3612	28	826	826	12	2012-08-02 11:11:45.763604	\N
13201	1-13201	70	15	2357	81	523	523	3	2012-08-02 11:25:08.092627	\N
11607	1-11607	60	98	2427	37	1790	1790	6	2012-08-02 11:14:46.107133	\N
12027	1-12027	3	95	5740	67	446	446	2	2012-08-02 11:17:30.770553	\N
12696	1-12696	26	58	3046	35	914	669	6	2012-08-02 11:21:49.932152	2012-08-02 11:25:04.93251
11157	1-11157	27	63	8441	86	7174	5131	10	2012-08-02 11:11:51.676941	2012-08-02 11:20:21.542331
12722	1-12722	80	53	4877	80	5643	675	0	2012-08-02 11:21:59.666365	2012-08-02 11:28:54.736808
12039	1-12039	23	1	9723	38	5301	5301	17	2012-08-02 11:17:35.03494	\N
11620	1-11620	67	17	6280	45	678	678	8	2012-08-02 11:14:51.158428	\N
11104	1-11104	18	29	7688	16	5118	5118	6	2012-08-02 11:11:30.450503	\N
12698	1-12698	95	51	8179	82	1369	1369	0	2012-08-02 11:21:50.694851	\N
12472	1-12472	55	32	7223	86	2701	2701	9	2012-08-02 11:20:22.329387	\N
11623	1-11623	13	78	3988	98	2747	1500	9	2012-08-02 11:14:52.31824	2012-08-02 11:26:06.969062
12042	1-12042	61	49	5084	96	1066	1066	2	2012-08-02 11:17:36.206245	\N
11209	1-11209	89	59	8389	30	6759	514	4	2012-08-02 11:12:12.053809	2012-08-02 11:28:22.314052
13643	1-13643	91	55	7154	64	8314	8314	8	2012-08-02 11:28:03.972311	\N
12045	1-12045	74	76	9898	51	18	18	0	2012-08-02 11:17:37.377521	\N
11121	1-11121	39	56	4684	87	2581	2581	0	2012-08-02 11:11:37.470201	\N
11123	1-11123	13	3	8286	91	4060	4060	3	2012-08-02 11:11:38.651966	\N
11653	1-11653	81	30	3972	68	535	535	0	2012-08-02 11:15:03.941854	\N
11126	1-11126	75	5	8383	90	299	299	0	2012-08-02 11:11:39.448144	\N
13646	1-13646	49	86	6487	78	1828	1828	3	2012-08-02 11:28:05.121136	\N
12051	1-12051	100	39	8020	46	327	327	10	2012-08-02 11:17:39.376671	\N
11644	1-11644	31	47	4907	62	1759	1759	0	2012-08-02 11:15:00.450689	\N
13649	1-13649	60	88	9656	54	9135	9135	5	2012-08-02 11:28:06.259109	\N
12495	1-12495	1	68	9081	11	2064	2064	6	2012-08-02 11:20:31.277895	\N
12497	1-12497	55	43	6571	5	3632	3632	5	2012-08-02 11:20:32.449612	\N
12718	1-12718	18	31	8752	28	361	361	12	2012-08-02 11:21:58.506695	\N
12509	1-12509	91	35	7013	29	3567	3567	9	2012-08-02 11:20:36.694365	\N
12486	1-12486	93	4	7915	88	2857	2612	0	2012-08-02 11:20:28.159101	2012-08-02 11:22:36.670154
12512	1-12512	93	36	1279	9	619	619	3	2012-08-02 11:20:37.876875	\N
11193	1-11193	48	74	5660	48	4914	4914	3	2012-08-02 11:12:05.789151	\N
11197	1-11197	29	60	9257	4	1762	1762	0	2012-08-02 11:12:07.766613	\N
11239	1-11239	42	13	7313	22	982	982	0	2012-08-02 11:12:23.886869	\N
11224	1-11224	58	78	8764	37	3125	3125	4	2012-08-02 11:12:17.997366	\N
11245	1-11245	44	3	4888	50	4077	4077	3	2012-08-02 11:12:26.240212	\N
11248	1-11248	79	11	7377	43	229	229	11	2012-08-02 11:12:27.433996	\N
11254	1-11254	14	79	5512	52	4915	4915	0	2012-08-02 11:12:29.744275	\N
11278	1-11278	23	75	8259	59	4998	4998	6	2012-08-02 11:12:39.114332	\N
11201	1-11201	79	30	4816	41	2196	1573	9	2012-08-02 11:12:08.93799	2012-08-02 11:29:44.95862
11678	1-11678	74	46	7662	99	4417	1960	5	2012-08-02 11:15:13.922692	2012-08-02 11:16:30.836393
11599	1-11599	66	96	6326	41	2099	1695	6	2012-08-02 11:14:43.433275	2012-08-02 11:20:55.056277
12050	1-12050	1	20	5814	32	1549	1147	0	2012-08-02 11:17:39.763676	2012-08-02 11:29:46.141041
12483	1-12483	39	3	1618	84	1117	1117	12	2012-08-02 11:20:26.598542	\N
13654	1-13654	1	68	4177	65	9027	1118	9	2012-08-02 11:28:08.203628	2012-08-02 11:29:54.404978
11700	1-11700	50	27	6639	84	3681	2792	7	2012-08-02 11:15:22.548146	2012-08-02 11:20:28.944385
12519	1-12519	26	91	6421	96	3152	2202	9	2012-08-02 11:20:40.563537	2012-08-02 11:30:20.7384
12522	1-12522	48	3	6914	24	7991	7179	6	2012-08-02 11:20:41.734721	2012-08-02 11:26:13.358762
11033	1-11033	31	18	3576	26	952	952	0	2012-08-02 10:39:35.71942	\N
12501	1-12501	91	63	7673	95	9213	5284	2	2012-08-02 11:20:33.610289	2012-08-02 11:29:58.029416
11640	1-11640	41	68	6881	16	6652	4513	0	2012-08-02 11:14:58.914905	2012-08-02 11:21:42.053335
11103	1-11103	13	65	6720	61	5808	1445	5	2012-08-02 11:11:30.837301	2012-08-02 11:30:21.125048
10866	1-10866	30	65	4459	9	1894	1137	0	2012-08-02 10:38:29.611591	2012-08-02 11:30:39.704933
11207	1-11207	85	75	3650	72	4208	1053	6	2012-08-02 11:12:11.258307	2012-08-02 11:26:21.833694
12685	1-12685	10	19	7512	26	7185	5816	32	2012-08-02 11:21:45.656278	2012-08-02 11:24:29.949566
11611	1-11611	87	21	1452	91	310	310	11	2012-08-02 11:14:47.68696	\N
11637	1-11637	58	98	3680	100	1964	338	2	2012-08-02 11:14:57.732412	2012-08-02 11:23:09.565198
11634	1-11634	9	47	7370	10	3761	508	4	2012-08-02 11:14:56.594502	2012-08-02 11:25:28.933653
11681	1-11681	48	79	9950	66	3445	3130	10	2012-08-02 11:15:15.087026	2012-08-02 11:21:24.042018
12041	1-12041	65	74	5742	91	3789	3789	5	2012-08-02 11:17:35.421677	\N
12487	1-12487	23	57	9501	79	2851	2851	0	2012-08-02 11:20:27.781106	\N
11249	1-11249	35	79	9425	36	2929	1391	5	2012-08-02 11:12:27.830987	2012-08-02 11:19:59.331014
11161	1-11161	60	2	5018	81	1414	1188	11	2012-08-02 11:11:53.279455	2012-08-02 11:23:32.018196
12680	1-12680	56	47	6804	21	2527	2527	0	2012-08-02 11:21:43.703997	\N
12083	1-12083	66	29	4125	43	4325	359	4	2012-08-02 11:17:52.184662	2012-08-02 11:29:12.856069
13192	1-13192	35	13	7968	88	9005	8248	2	2012-08-02 11:25:04.170282	2012-08-02 11:27:13.362106
12682	1-12682	82	24	8503	96	2522	2522	3	2012-08-02 11:21:44.872248	\N
11031	1-11031	93	31	9114	48	6285	3111	7	2012-08-02 10:39:34.945883	2012-08-02 11:14:40.747269
11226	1-11226	72	94	8609	82	3952	1914	7	2012-08-02 11:12:18.395002	2012-08-02 11:19:25.352516
12477	1-12477	58	9	7806	14	3331	3331	0	2012-08-02 11:20:24.252067	\N
11616	1-11616	51	53	8765	1	1133	1133	6	2012-08-02 11:14:50.020599	\N
13183	1-13183	86	9	9039	3	10000	4576	11	2012-08-02 11:25:00.269988	2012-08-02 11:27:45.695868
12058	1-12058	79	75	9188	27	838	838	3	2012-08-02 11:17:42.980816	\N
11116	1-11116	41	47	3065	17	601	601	3	2012-08-02 11:11:35.13861	\N
10515	1-10515	91	13	4045	42	3597	162	0	2012-08-02 10:32:02.871413	2012-08-02 11:20:25.038434
11664	1-11664	76	29	8209	5	9786	3452	9	2012-08-02 11:15:08.262308	2012-08-02 11:27:56.667174
12691	1-12691	41	51	3813	7	103	103	0	2012-08-02 11:21:48.009587	\N
13663	1-13663	99	31	5297	5	3478	3478	10	2012-08-02 11:28:11.728742	\N
11125	1-11125	69	82	7035	38	1582	1582	11	2012-08-02 11:11:39.050505	\N
11132	1-11132	100	30	6105	74	997	997	0	2012-08-02 11:11:41.790257	\N
13673	1-13673	58	39	9667	87	2989	2989	5	2012-08-02 11:28:15.640128	\N
11737	1-11737	80	24	7775	84	6953	3754	19	2012-08-02 11:15:36.958268	2012-08-02 11:28:12.888559
11267	1-11267	98	56	7017	69	2189	2189	2	2012-08-02 11:12:34.860041	\N
11165	1-11165	36	96	8903	89	1186	1186	13	2012-08-02 11:11:54.428281	\N
11168	1-11168	36	45	4635	80	348	348	19	2012-08-02 11:11:55.986506	\N
11642	1-11642	42	73	8740	29	1834	1834	0	2012-08-02 11:15:00.052528	\N
11171	1-11171	92	89	8453	99	1768	1768	0	2012-08-02 11:11:57.190964	\N
13197	1-13197	84	11	7085	48	4352	4352	0	2012-08-02 11:25:06.52421	\N
12080	1-12080	12	12	9837	1	6786	2022	0	2012-08-02 11:17:51.047309	2012-08-02 11:25:42.77523
12510	1-12510	93	85	4783	91	256	256	6	2012-08-02 11:20:37.103434	\N
13612	1-13612	74	97	3980	38	3372	1126	4	2012-08-02 11:27:51.242654	2012-08-02 11:28:14.457753
13682	1-13682	34	44	7623	100	2249	2249	11	2012-08-02 11:28:18.788999	\N
12516	1-12516	81	4	4972	72	378	378	0	2012-08-02 11:20:39.412605	\N
11661	1-11661	13	12	9021	67	1414	1414	0	2012-08-02 11:15:07.102498	\N
13689	1-13689	81	98	2134	26	24	24	13	2012-08-02 11:28:21.506791	\N
11039	1-11039	73	99	6078	93	2915	1622	5	2012-08-02 10:39:38.514284	2012-08-02 11:12:06.970827
13695	1-13695	74	21	9516	44	2697	2697	0	2012-08-02 11:28:24.280161	\N
12524	1-12524	26	17	7955	86	737	737	0	2012-08-02 11:20:42.927676	\N
13698	1-13698	23	100	7211	80	3636	3636	0	2012-08-02 11:28:25.440098	\N
13701	1-13701	37	23	4453	17	8403	6424	6	2012-08-02 11:28:26.600446	2012-08-02 11:28:26.600446
11219	1-11219	44	68	5565	11	484	484	6	2012-08-02 11:12:15.93129	\N
13678	1-13678	59	24	7543	97	8640	2747	0	2012-08-02 11:28:16.811248	2012-08-02 11:28:30.157943
11687	1-11687	77	99	5986	86	1507	1507	10	2012-08-02 11:15:17.445721	\N
11691	1-11691	6	68	7402	7	3302	3302	7	2012-08-02 11:15:18.610044	\N
11697	1-11697	69	13	9577	60	4520	4520	3	2012-08-02 11:15:21.365201	\N
11706	1-11706	94	81	5986	44	858	858	13	2012-08-02 11:15:24.894733	\N
11252	1-11252	27	2	8377	82	93	93	0	2012-08-02 11:12:28.992881	\N
11712	1-11712	19	46	7606	92	2478	2478	0	2012-08-02 11:15:27.260953	\N
11718	1-11718	75	28	8982	6	5801	5801	0	2012-08-02 11:15:29.584003	\N
11276	1-11276	33	82	7944	44	2026	2026	0	2012-08-02 11:12:38.329782	\N
11739	1-11739	86	57	8658	15	9330	9330	4	2012-08-02 11:15:38.162686	\N
13680	1-13680	67	87	9621	41	9660	8542	0	2012-08-02 11:28:18.004813	2012-08-02 11:29:44.95862
12708	1-12708	62	24	7420	40	3084	3084	0	2012-08-02 11:21:54.993179	\N
12717	1-12717	75	41	4303	83	4048	2720	7	2012-08-02 11:21:58.108604	2012-08-02 11:23:46.504397
13656	1-13656	31	35	3687	37	2287	2287	4	2012-08-02 11:28:08.977271	\N
13674	1-13674	70	74	6918	1	9069	231	0	2012-08-02 11:28:16.026817	2012-08-02 11:30:34.543923
10868	1-10868	89	35	8079	76	4078	4078	6	2012-08-02 10:38:30.418486	\N
11102	1-11102	99	49	6339	65	1174	471	9	2012-08-02 11:11:30.074836	2012-08-02 11:29:22.394404
13665	1-13665	16	75	5006	16	7909	184	2	2012-08-02 11:28:12.501829	2012-08-02 11:29:35.033431
12087	1-12087	16	22	9216	80	5713	4861	9	2012-08-02 11:17:53.698432	2012-08-02 11:26:44.023142
11153	1-11153	78	2	3567	58	401	401	4	2012-08-02 11:11:50.118265	\N
11639	1-11639	65	41	9880	49	3180	2967	6	2012-08-02 11:14:58.141272	2012-08-02 11:15:54.281626
11108	1-11108	7	74	5079	80	6203	4278	0	2012-08-02 11:11:32.397687	2012-08-02 11:19:26.126022
12493	1-12493	12	2	8232	20	6287	1835	15	2012-08-02 11:20:30.879963	2012-08-02 11:28:09.375063
11240	1-11240	71	23	6464	63	3114	1063	0	2012-08-02 11:12:24.28452	2012-08-02 11:29:25.178304
13662	1-13662	69	28	8437	57	1392	1392	17	2012-08-02 11:28:11.342204	\N
11600	1-11600	37	6	8064	25	1864	1864	0	2012-08-02 11:14:42.658624	\N
11087	1-11087	36	70	6625	88	2762	1722	10	2012-08-02 11:11:24.236972	2012-08-02 11:24:10.204408
11651	1-11651	100	69	8057	100	233	233	12	2012-08-02 11:15:03.190634	\N
11150	1-11150	21	30	7653	40	8185	651	3	2012-08-02 11:11:48.936037	2012-08-02 11:29:15.971766
11615	1-11615	2	1	5156	46	827	827	0	2012-08-02 11:14:49.257982	\N
11618	1-11618	34	36	9785	72	8485	8485	7	2012-08-02 11:14:50.395962	\N
11606	1-11606	38	15	7702	41	942	942	0	2012-08-02 11:14:45.731432	\N
11053	1-11053	8	98	4806	11	227	227	0	2012-08-02 11:11:11.273189	\N
11042	1-11042	84	47	6354	34	2345	2259	9	2012-08-02 11:11:07.231417	2012-08-02 11:16:43.499661
11058	1-11058	59	88	7026	36	263	263	6	2012-08-02 11:11:12.90135	\N
11041	1-11041	53	45	2322	44	502	297	2	2012-08-02 11:11:06.689639	2012-08-02 11:23:32.018196
13668	1-13668	88	9	2154	59	1407	1407	10	2012-08-02 11:28:13.661851	\N
12703	1-12703	93	78	7105	15	3010	3010	0	2012-08-02 11:21:52.639099	\N
12711	1-12711	97	87	9678	53	1315	1315	4	2012-08-02 11:21:55.78858	\N
13210	1-13210	9	48	5212	84	2035	2035	2	2012-08-02 11:25:10.843789	\N
11111	1-11111	11	18	6659	26	1632	1632	0	2012-08-02 11:11:33.546478	\N
13212	1-13212	26	20	9190	44	4568	4568	11	2012-08-02 11:25:12.438005	\N
13676	1-13676	47	81	3367	41	5085	5085	12	2012-08-02 11:28:17.20925	\N
12081	1-12081	66	53	9341	13	521	521	0	2012-08-02 11:17:51.422398	\N
13215	1-13215	88	89	1223	71	224	224	7	2012-08-02 11:25:13.228122	\N
12499	1-12499	94	66	6011	24	1407	1407	11	2012-08-02 11:20:32.825149	\N
12502	1-12502	90	67	2671	12	14	14	0	2012-08-02 11:20:33.996841	\N
11114	1-11114	28	75	8949	1	1511	1511	6	2012-08-02 11:11:34.718495	\N
13217	1-13217	10	28	3516	83	2894	2894	4	2012-08-02 11:25:14.401998	\N
12089	1-12089	57	85	1347	21	285	285	10	2012-08-02 11:17:54.858358	\N
11084	1-11084	61	44	4267	20	741	741	12	2012-08-02 11:11:23.12077	\N
13220	1-13220	15	6	5105	12	2528	2528	12	2012-08-02 11:25:15.199355	\N
12508	1-12508	72	68	8325	82	2400	2400	0	2012-08-02 11:20:36.319248	\N
12511	1-12511	90	10	6428	44	2946	2946	5	2012-08-02 11:20:37.490295	\N
12514	1-12514	26	3	8492	52	3724	3724	5	2012-08-02 11:20:38.650055	\N
13222	1-13222	49	17	9159	53	1080	1080	6	2012-08-02 11:25:16.348609	\N
13683	1-13683	28	13	1470	50	27	27	0	2012-08-02 11:28:19.573712	\N
11663	1-11663	85	11	5840	37	5363	5363	0	2012-08-02 11:15:07.875507	\N
11124	1-11124	70	99	4664	82	3022	3022	0	2012-08-02 11:11:38.26531	\N
12723	1-12723	46	49	5147	40	3456	3456	3	2012-08-02 11:22:00.451268	\N
11635	1-11635	47	97	7322	46	4067	537	0	2012-08-02 11:14:56.99233	2012-08-02 11:28:30.157943
12714	1-12714	49	11	4165	49	1294	687	0	2012-08-02 11:21:56.949009	2012-08-02 11:28:30.535322
11666	1-11666	80	78	9135	18	1829	1829	0	2012-08-02 11:15:09.046225	\N
12732	1-12732	25	44	7484	36	114	114	7	2012-08-02 11:22:03.954684	\N
11260	1-11260	27	39	7769	15	1914	1914	0	2012-08-02 11:12:32.097538	\N
11682	1-11682	35	92	9952	97	2229	2229	5	2012-08-02 11:15:15.484814	\N
11202	1-11202	26	92	9814	82	2220	2220	18	2012-08-02 11:12:09.346744	\N
11690	1-11690	23	6	7587	70	4370	4370	0	2012-08-02 11:15:19.031814	\N
11208	1-11208	57	66	7474	14	3371	3371	4	2012-08-02 11:12:11.667017	\N
11185	1-11185	6	90	5576	72	473	473	0	2012-08-02 11:12:02.717611	\N
11188	1-11188	20	95	9294	57	1010	1010	6	2012-08-02 11:12:03.877512	\N
11191	1-11191	12	93	6083	23	345	345	7	2012-08-02 11:12:05.037524	\N
11214	1-11214	54	19	6458	15	4707	4707	0	2012-08-02 11:12:14.009137	\N
11247	1-11247	37	9	6780	11	613	613	0	2012-08-02 11:12:26.638347	\N
11231	1-11231	21	53	3061	79	613	613	3	2012-08-02 11:12:20.737531	\N
11250	1-11250	70	90	5121	92	2350	2350	5	2012-08-02 11:12:28.218047	\N
11257	1-11257	36	16	5082	67	1159	1159	13	2012-08-02 11:12:30.517715	\N
11271	1-11271	15	53	6847	92	1365	1365	0	2012-08-02 11:12:36.418134	\N
11280	1-11280	85	91	6441	35	178	178	0	2012-08-02 11:12:39.887829	\N
12164	1-12164	93	10	618	49	421	402	11	2012-08-02 11:18:23.29899	2012-08-02 11:20:46.071419
12091	1-12091	35	80	6007	85	3399	3399	3	2012-08-02 11:17:55.256003	\N
12172	1-12172	55	6	9084	82	8265	6021	0	2012-08-02 11:18:26.359454	2012-08-02 11:20:59.742471
12876	1-12876	66	46	7659	41	8178	3246	4	2012-08-02 11:23:00.261991	2012-08-02 11:29:44.95862
12115	1-12115	79	54	6877	43	2039	1649	0	2012-08-02 11:18:04.437867	2012-08-02 11:29:15.596151
12895	1-12895	14	15	2867	52	1159	883	12	2012-08-02 11:23:07.222898	2012-08-02 11:23:43.77499
13717	1-13717	66	98	7082	6	5673	1692	7	2012-08-02 11:28:32.889244	2012-08-02 11:30:01.554338
12883	1-12883	81	41	5802	98	6097	2954	0	2012-08-02 11:23:02.969442	2012-08-02 11:23:46.504397
12100	1-12100	16	53	4640	21	2943	2943	12	2012-08-02 11:17:58.714959	\N
12866	1-12866	60	1	4735	61	5600	3469	16	2012-08-02 11:22:56.333539	2012-08-02 11:27:12.599056
12150	1-12150	41	34	5496	86	3544	2037	7	2012-08-02 11:18:17.973501	2012-08-02 11:27:56.258298
13208	1-13208	65	25	9060	64	2480	2480	0	2012-08-02 11:25:10.446507	\N
13669	1-13669	39	19	4977	69	6319	6319	2	2012-08-02 11:28:14.059905	\N
12195	1-12195	82	74	5780	76	4418	2179	0	2012-08-02 11:18:35.076412	2012-08-02 11:24:54.347969
12709	1-12709	36	39	9685	13	5886	5886	3	2012-08-02 11:21:54.606457	\N
12181	1-12181	62	82	9796	60	5989	1768	7	2012-08-02 11:18:29.751074	2012-08-02 11:25:08.490508
12712	1-12712	32	69	8643	57	5012	5012	8	2012-08-02 11:21:56.186323	\N
12126	1-12126	100	70	9419	85	1657	1657	5	2012-08-02 11:18:08.801869	\N
13672	1-13672	97	10	7787	80	1628	1628	0	2012-08-02 11:28:15.253047	\N
12129	1-12129	1	47	9866	71	5067	5067	0	2012-08-02 11:18:09.930496	\N
13205	1-13205	69	97	8126	60	3343	3343	0	2012-08-02 11:25:09.651134	\N
12135	1-12135	78	91	8377	43	1244	1244	0	2012-08-02 11:18:12.228788	\N
13218	1-13218	43	86	4136	43	751	751	11	2012-08-02 11:25:14.015154	\N
12763	1-12763	9	19	2535	72	4169	1751	0	2012-08-02 11:22:15.897965	2012-08-02 11:25:12.050696
12721	1-12721	5	73	2663	89	2201	2201	7	2012-08-02 11:22:00.068828	\N
12141	1-12141	68	89	9272	87	9938	9938	0	2012-08-02 11:18:14.471342	\N
12147	1-12147	45	97	7757	14	8817	8817	4	2012-08-02 11:18:16.802713	\N
12152	1-12152	89	10	6355	99	2288	2288	0	2012-08-02 11:18:19.111806	\N
13677	1-13677	91	59	9170	22	5600	5600	5	2012-08-02 11:28:17.595678	\N
12155	1-12155	93	4	8026	94	964	964	5	2012-08-02 11:18:19.884663	\N
12731	1-12731	14	75	6250	25	3875	3875	11	2012-08-02 11:22:03.192398	\N
12158	1-12158	86	42	8537	48	3203	3203	6	2012-08-02 11:18:21.011599	\N
12784	1-12784	26	61	2996	11	860	860	12	2012-08-02 11:22:24.272542	\N
13228	1-13228	15	64	4044	52	2192	2192	0	2012-08-02 11:25:18.316032	\N
13231	1-13231	58	33	7439	59	1680	1680	19	2012-08-02 11:25:19.467946	\N
13234	1-13234	49	10	4877	39	570	570	8	2012-08-02 11:25:20.651422	\N
12166	1-12166	87	61	7381	66	1095	1095	6	2012-08-02 11:18:24.470148	\N
12699	1-12699	26	74	6927	29	4575	1863	0	2012-08-02 11:21:51.081516	2012-08-02 11:29:38.989309
12175	1-12175	49	36	6971	3	190	190	0	2012-08-02 11:18:27.486289	\N
13214	1-13214	10	74	8874	71	9394	8001	4	2012-08-02 11:25:12.840189	2012-08-02 11:26:28.960842
12745	1-12745	67	79	7715	3	734	734	5	2012-08-02 11:22:08.926434	\N
13684	1-13684	100	56	8535	5	1136	1136	3	2012-08-02 11:28:19.949095	\N
13661	1-13661	88	12	6227	21	6917	6039	0	2012-08-02 11:28:10.944145	2012-08-02 11:28:20.347033
13687	1-13687	100	36	5733	64	3208	3208	6	2012-08-02 11:28:21.119977	\N
12757	1-12757	11	30	5832	55	5392	5296	0	2012-08-02 11:22:13.58877	2012-08-02 11:22:15.135514
12766	1-12766	28	99	7790	3	775	775	12	2012-08-02 11:22:17.182655	\N
13693	1-13693	69	85	6210	44	1505	1505	10	2012-08-02 11:28:23.484569	\N
13696	1-13696	5	39	5086	93	4777	4777	0	2012-08-02 11:28:24.655687	\N
12727	1-12727	6	16	1608	81	689	171	0	2012-08-02 11:22:02.010298	2012-08-02 11:23:32.018196
12841	1-12841	53	1	4123	65	1888	1832	4	2012-08-02 11:22:46.460796	2012-08-02 11:26:28.56418
13702	1-13702	18	94	9479	98	7731	7731	3	2012-08-02 11:28:26.99806	\N
11243	1-11243	81	11	5945	84	5726	722	25	2012-08-02 11:12:25.466987	2012-08-02 11:25:26.565595
12189	1-12189	40	21	6774	58	1546	228	2	2012-08-02 11:18:32.822666	2012-08-02 11:28:25.826838
12847	1-12847	17	29	6536	69	7011	4618	4	2012-08-02 11:22:48.848011	2012-08-02 11:25:26.950328
13251	1-13251	55	63	4919	52	4606	4606	9	2012-08-02 11:25:27.762834	\N
13256	1-13256	21	67	1980	96	837	837	0	2012-08-02 11:25:29.320956	\N
13704	1-13704	9	65	9163	7	4426	4426	5	2012-08-02 11:28:28.208776	\N
13707	1-13707	91	65	4561	14	2617	2617	6	2012-08-02 11:28:28.975971	\N
13267	1-13267	20	74	9037	14	7723	7723	0	2012-08-02 11:25:33.302173	\N
12810	1-12810	61	69	6330	92	555	555	0	2012-08-02 11:22:34.361742	\N
13714	1-13714	71	63	9911	58	6142	6142	0	2012-08-02 11:28:31.319875	\N
12804	1-12804	61	2	7085	57	1419	1419	3	2012-08-02 11:22:32.064011	\N
12813	1-12813	55	16	4438	77	3375	3375	3	2012-08-02 11:22:35.521328	\N
12819	1-12819	57	59	2748	85	116	116	10	2012-08-02 11:22:37.863964	\N
12822	1-12822	69	5	6860	24	440	440	4	2012-08-02 11:22:39.001727	\N
12828	1-12828	9	23	9109	73	6	6	8	2012-08-02 11:22:41.399478	\N
12754	1-12754	27	24	4124	82	3677	2312	0	2012-08-02 11:22:12.428598	2012-08-02 11:22:46.074313
12853	1-12853	65	35	4151	89	54	54	0	2012-08-02 11:22:51.201023	\N
12863	1-12863	18	76	5190	55	4757	4757	7	2012-08-02 11:22:54.769682	\N
12886	1-12886	66	92	9991	52	1435	1435	0	2012-08-02 11:23:04.118591	\N
12891	1-12891	78	5	7466	78	554	554	2	2012-08-02 11:23:06.461024	\N
12856	1-12856	12	32	8941	90	5791	4443	15	2012-08-02 11:22:52.383294	2012-08-02 11:23:06.836243
12898	1-12898	76	35	9595	88	4818	4818	0	2012-08-02 11:23:08.78063	\N
12879	1-12879	17	9	9105	25	6459	5875	9	2012-08-02 11:23:01.792275	2012-08-02 11:27:45.695868
12125	1-12125	65	100	3953	52	6307	2304	10	2012-08-02 11:18:08.404395	2012-08-02 11:19:01.720992
12771	1-12771	10	63	8318	84	5788	5788	5	2012-08-02 11:22:19.168136	\N
12730	1-12730	25	62	5066	22	4690	2780	0	2012-08-02 11:22:03.568367	2012-08-02 11:26:15.336549
12902	1-12902	78	13	4739	21	3084	1273	11	2012-08-02 11:23:10.349402	2012-08-02 11:29:06.182748
12170	1-12170	22	69	7428	24	3973	2489	2	2012-08-02 11:18:25.607763	2012-08-02 11:30:29.07869
13715	1-13715	86	12	4976	5	1553	1553	0	2012-08-02 11:28:32.105707	\N
12113	1-12113	85	25	6105	38	3072	1623	0	2012-08-02 11:18:03.67574	2012-08-02 11:19:12.072644
12739	1-12739	19	100	6986	24	7495	4532	4	2012-08-02 11:22:06.628353	2012-08-02 11:25:07.330226
13688	1-13688	48	78	9910	75	727	727	6	2012-08-02 11:28:21.915788	\N
12774	1-12774	22	69	3508	98	329	329	7	2012-08-02 11:22:20.372623	\N
12104	1-12104	68	87	3639	40	314	314	8	2012-08-02 11:18:00.217411	\N
12163	1-12163	11	16	8384	59	934	362	6	2012-08-02 11:18:22.536324	2012-08-02 11:24:25.320192
12868	1-12868	90	42	8606	46	2890	1734	0	2012-08-02 11:22:57.111866	2012-08-02 11:27:12.599056
12110	1-12110	90	79	3985	69	1591	1591	0	2012-08-02 11:18:02.504298	\N
12119	1-12119	17	99	5311	6	3702	394	16	2012-08-02 11:18:05.962667	2012-08-02 11:26:58.444928
13067	1-13067	58	29	7148	62	5850	3034	0	2012-08-02 11:24:14.855611	2012-08-02 11:29:59.576353
12808	1-12808	35	5	7364	38	3264	1656	0	2012-08-02 11:22:33.974546	2012-08-02 11:26:18.69543
12188	1-12188	2	6	9533	82	923	923	0	2012-08-02 11:18:32.424614	\N
12753	1-12753	60	30	6589	99	5831	2676	10	2012-08-02 11:22:12.01996	2012-08-02 11:30:08.195798
12905	1-12905	30	51	3901	17	3169	2104	6	2012-08-02 11:23:11.48793	2012-08-02 11:27:15.385591
12919	1-12919	92	10	4618	84	8907	5008	0	2012-08-02 11:23:16.956535	2012-08-02 11:30:33.358312
12768	1-12768	36	73	2276	71	8315	232	10	2012-08-02 11:22:17.996906	2012-08-02 11:30:03.509658
13691	1-13691	49	28	7308	34	9885	9885	2	2012-08-02 11:28:22.700315	\N
12736	1-12736	97	71	4993	95	1035	516	13	2012-08-02 11:22:05.501313	2012-08-02 11:29:10.513766
12733	1-12733	69	59	3561	84	3405	3405	0	2012-08-02 11:22:04.341349	\N
12786	1-12786	3	94	8727	29	8954	7395	0	2012-08-02 11:22:25.05695	2012-08-02 11:28:53.156625
12741	1-12741	16	9	8435	52	2430	2430	7	2012-08-02 11:22:07.788342	\N
12194	1-12194	16	58	7072	15	9909	9909	0	2012-08-02 11:18:34.689834	\N
13700	1-13700	16	54	9677	6	3923	3923	4	2012-08-02 11:28:26.213525	\N
12744	1-12744	34	57	9433	85	3706	3706	4	2012-08-02 11:22:08.550574	\N
12862	1-12862	94	4	9089	48	6771	5706	0	2012-08-02 11:22:55.145283	2012-08-02 11:27:59.982053
13223	1-13223	78	10	7238	81	2234	2234	0	2012-08-02 11:25:15.972719	\N
12149	1-12149	100	3	905	74	511	511	0	2012-08-02 11:18:17.586903	\N
12153	1-12153	39	6	6287	20	645	645	0	2012-08-02 11:18:18.713639	\N
12728	1-12728	93	36	8609	6	6589	5224	0	2012-08-02 11:22:02.40822	2012-08-02 11:25:38.40866
12131	1-12131	76	96	4888	45	5542	2769	8	2012-08-02 11:18:10.681885	2012-08-02 11:25:11.265631
12159	1-12159	59	43	8807	26	1052	1052	0	2012-08-02 11:18:21.387318	\N
12167	1-12167	1	62	7916	77	3100	3100	15	2012-08-02 11:18:24.083185	\N
12777	1-12777	33	62	5508	72	6856	142	0	2012-08-02 11:22:21.555326	2012-08-02 11:28:32.889244
13232	1-13232	45	57	6194	81	386	386	16	2012-08-02 11:25:19.854796	\N
12762	1-12762	20	30	4982	25	122	122	12	2012-08-02 11:22:15.510799	\N
13694	1-13694	36	98	6397	16	4843	2945	3	2012-08-02 11:28:23.893271	2012-08-02 11:29:18.780703
13238	1-13238	47	31	5750	99	2435	2435	3	2012-08-02 11:25:22.233212	\N
12798	1-12798	61	62	8733	57	6920	6821	3	2012-08-02 11:22:29.71971	2012-08-02 11:26:22.651417
12789	1-12789	83	68	8328	85	1514	1514	5	2012-08-02 11:22:26.20601	\N
12747	1-12747	9	70	6014	16	3627	2423	6	2012-08-02 11:22:09.66652	2012-08-02 11:25:53.587928
12792	1-12792	29	49	9665	56	4048	4048	0	2012-08-02 11:22:27.393319	\N
12206	1-12206	34	84	7120	57	402	402	14	2012-08-02 11:18:39.321571	\N
12134	1-12134	56	26	7783	79	6463	2414	11	2012-08-02 11:18:11.830589	2012-08-02 11:26:08.571845
12209	1-12209	44	97	8869	25	4888	4888	6	2012-08-02 11:18:40.470085	\N
13718	1-13718	77	10	5448	94	1430	1430	3	2012-08-02 11:28:33.253874	\N
12885	1-12885	71	47	5817	33	4088	1270	16	2012-08-02 11:23:03.742743	2012-08-02 11:28:31.707131
12806	1-12806	40	22	6207	98	2955	2955	0	2012-08-02 11:22:32.847889	\N
13727	1-13727	47	61	8690	17	3478	3478	11	2012-08-02 11:28:36.756752	\N
12811	1-12811	12	82	8697	88	3845	3845	3	2012-08-02 11:22:34.747913	\N
13730	1-13730	72	69	5602	33	1643	1643	2	2012-08-02 11:28:38.203737	\N
13732	1-13732	66	47	7498	1	2627	2627	5	2012-08-02 11:28:39.389286	\N
13735	1-13735	24	23	9037	43	1747	1747	0	2012-08-02 11:28:40.181821	\N
13738	1-13738	58	9	6709	22	4698	4698	2	2012-08-02 11:28:41.33045	\N
12913	1-12913	61	63	7460	56	8752	4419	0	2012-08-02 11:23:14.61459	2012-08-02 11:28:42.910342
12860	1-12860	58	8	6666	83	1998	1998	5	2012-08-02 11:22:53.940672	\N
12865	1-12865	97	3	4276	84	4595	4595	12	2012-08-02 11:22:55.918265	\N
12874	1-12874	93	44	4873	23	264	264	5	2012-08-02 11:22:59.484248	\N
12817	1-12817	84	88	8052	16	3597	1737	5	2012-08-02 11:22:37.068514	2012-08-02 11:23:09.156232
12901	1-12901	84	100	5427	45	88	88	19	2012-08-02 11:23:09.952014	\N
12904	1-12904	32	98	4090	13	515	515	6	2012-08-02 11:23:11.100735	\N
12907	1-12907	62	6	9598	48	8956	5428	0	2012-08-02 11:23:12.249847	2012-08-02 11:23:12.249847
12137	1-12137	4	27	2612	92	2289	1210	0	2012-08-02 11:18:12.979696	2012-08-02 11:23:30.835937
13224	1-13224	13	25	6563	5	5067	5067	2	2012-08-02 11:25:17.145193	\N
12102	1-12102	28	78	8653	23	3116	3116	12	2012-08-02 11:17:59.455201	\N
12165	1-12165	71	8	7875	20	9422	3482	0	2012-08-02 11:18:23.685467	2012-08-02 11:28:47.243134
13731	1-13731	38	30	7065	33	3920	1904	11	2012-08-02 11:28:38.60191	2012-08-02 11:29:29.884913
13719	1-13719	8	69	8973	92	1004	1004	0	2012-08-02 11:28:33.629805	\N
12136	1-12136	15	99	8525	4	6191	5380	0	2012-08-02 11:18:12.603877	2012-08-02 11:30:25.883989
12111	1-12111	23	59	7279	22	1205	1205	3	2012-08-02 11:18:02.902112	\N
12782	1-12782	20	33	6972	61	2159	2159	10	2012-08-02 11:22:23.510355	\N
12114	1-12114	35	20	2408	50	781	781	13	2012-08-02 11:18:04.051595	\N
12767	1-12767	39	28	9190	9	5906	5070	4	2012-08-02 11:22:17.588318	2012-08-02 11:29:19.996112
12120	1-12120	96	33	8862	47	1247	1247	7	2012-08-02 11:18:06.33809	\N
12827	1-12827	84	55	8608	21	3372	1364	0	2012-08-02 11:22:40.993518	2012-08-02 11:26:55.196257
12124	1-12124	27	36	4991	8	395	395	2	2012-08-02 11:18:07.641586	\N
12199	1-12199	84	2	9369	92	6781	3825	5	2012-08-02 11:18:36.579017	2012-08-02 11:27:03.56104
13709	1-13709	54	53	1890	9	585	19	8	2012-08-02 11:28:29.771161	2012-08-02 11:30:28.283346
13264	1-13264	1	95	4710	3	2051	1936	0	2012-08-02 11:25:32.494635	2012-08-02 11:28:14.866431
12915	1-12915	10	31	5260	57	4714	4714	0	2012-08-02 11:23:15.012185	\N
12940	1-12940	90	22	6997	17	8624	1170	4	2012-08-02 11:23:25.3555	2012-08-02 11:28:22.314052
12738	1-12738	92	27	6024	64	1852	1852	0	2012-08-02 11:22:06.252717	\N
13000	1-13000	25	10	3928	90	5487	739	0	2012-08-02 11:23:48.870435	2012-08-02 11:27:12.599056
12133	1-12133	69	48	4918	40	5004	791	8	2012-08-02 11:18:11.454947	2012-08-02 11:28:50.33668
13706	1-13706	27	44	1149	18	21	21	0	2012-08-02 11:28:28.588815	\N
12743	1-12743	15	50	9107	48	4106	4106	0	2012-08-02 11:22:08.174955	\N
12117	1-12117	2	39	4223	61	2340	792	11	2012-08-02 11:18:05.189292	2012-08-02 11:27:39.261479
13722	1-13722	89	92	4640	44	3419	1581	7	2012-08-02 11:28:34.811821	2012-08-02 11:30:40.103243
12145	1-12145	87	83	6574	27	1023	1023	2	2012-08-02 11:18:16.029003	\N
12148	1-12148	35	26	7698	83	1478	1478	0	2012-08-02 11:18:17.1896	\N
12962	1-12962	61	8	3942	93	4377	4377	3	2012-08-02 11:23:33.963111	\N
12154	1-12154	43	28	6967	51	3393	3393	3	2012-08-02 11:18:19.498575	\N
12788	1-12788	9	24	3898	92	745	745	8	2012-08-02 11:22:25.830353	\N
12157	1-12157	68	69	9701	7	8133	1028	4	2012-08-02 11:18:20.60287	2012-08-02 11:25:24.204162
12812	1-12812	72	31	4743	48	8072	1753	9	2012-08-02 11:22:35.134609	2012-08-02 11:28:32.513694
13245	1-13245	99	28	8522	36	1278	1278	0	2012-08-02 11:25:24.988643	\N
13713	1-13713	62	73	9914	47	9009	5132	0	2012-08-02 11:28:31.707131	2012-08-02 11:28:31.707131
12758	1-12758	75	14	7445	68	564	564	0	2012-08-02 11:22:13.975336	\N
13728	1-13728	13	8	2502	84	2809	2809	0	2012-08-02 11:28:37.431792	\N
13266	1-13266	43	93	2170	60	667	667	5	2012-08-02 11:25:33.688537	\N
12764	1-12764	31	46	8529	48	4297	4297	14	2012-08-02 11:22:16.306657	\N
13269	1-13269	51	63	5300	80	2272	2272	7	2012-08-02 11:25:34.472959	\N
12927	1-12927	19	21	8758	67	3780	2809	7	2012-08-02 11:23:20.127555	2012-08-02 11:25:36.461597
13276	1-13276	18	91	9117	83	4469	4469	0	2012-08-02 11:25:36.848308	\N
13283	1-13283	67	51	3156	36	397	397	9	2012-08-02 11:25:39.55929	\N
12805	1-12805	68	52	8912	45	4932	4932	5	2012-08-02 11:22:32.472694	\N
12809	1-12809	97	97	6356	37	5046	5046	8	2012-08-02 11:22:33.599065	\N
13740	1-13740	34	28	3370	51	8805	6781	4	2012-08-02 11:28:42.103973	2012-08-02 11:29:27.940484
13703	1-13703	9	60	6094	83	9233	3988	0	2012-08-02 11:28:27.826579	2012-08-02 11:28:53.543122
12999	1-12999	71	84	8864	90	6381	5046	6	2012-08-02 11:23:48.494825	2012-08-02 11:25:50.836867
12956	1-12956	64	42	4041	29	2400	714	0	2012-08-02 11:23:31.620437	2012-08-02 11:29:38.989309
12923	1-12923	67	4	615	31	298	120	0	2012-08-02 11:23:18.956989	2012-08-02 11:25:56.803419
12931	1-12931	6	19	2375	87	36	36	9	2012-08-02 11:23:21.674541	\N
12932	1-12932	36	70	4156	99	341	341	5	2012-08-02 11:23:22.216385	\N
12162	1-12162	1	33	9527	42	8467	7736	0	2012-08-02 11:18:22.912149	2012-08-02 11:23:20.912193
12930	1-12930	19	90	9155	94	1837	1837	3	2012-08-02 11:23:21.287725	\N
12934	1-12934	10	63	3715	100	552	552	0	2012-08-02 11:23:23.011352	\N
12935	1-12935	23	77	6411	90	982	982	4	2012-08-02 11:23:23.386972	\N
12939	1-12939	72	32	5018	82	1393	1393	0	2012-08-02 11:23:24.966676	\N
12942	1-12942	36	39	9102	55	1983	1983	0	2012-08-02 11:23:26.106724	\N
12946	1-12946	100	22	8524	14	1583	1583	6	2012-08-02 11:23:27.73107	\N
12947	1-12947	40	13	6827	40	4145	4145	0	2012-08-02 11:23:28.117625	\N
12951	1-12951	51	2	9921	84	4150	2079	6	2012-08-02 11:23:29.686653	2012-08-02 11:23:29.686653
12965	1-12965	20	13	8719	42	3679	3679	4	2012-08-02 11:23:35.122951	\N
12971	1-12971	45	33	4103	30	2971	2971	0	2012-08-02 11:23:37.487259	\N
12974	1-12974	44	94	4189	73	548	548	3	2012-08-02 11:23:38.669573	\N
13290	1-13290	11	29	5629	37	9187	5712	5	2012-08-02 11:25:43.184738	2012-08-02 11:27:38.068659
11928	1-11928	19	16	7561	77	6072	630	0	2012-08-02 11:16:51.380756	2012-08-02 11:29:46.141041
13282	1-13282	94	63	2091	91	1567	1065	15	2012-08-02 11:25:39.968389	2012-08-02 11:29:31.862809
13736	1-13736	43	3	4299	12	3336	2770	0	2012-08-02 11:28:40.556925	2012-08-02 11:29:46.549736
13815	1-13815	37	35	7218	79	9317	3312	2	2012-08-02 11:29:11.695894	2012-08-02 11:29:58.029416
13806	1-13806	92	41	4262	12	3954	3954	6	2012-08-02 11:29:08.149592	\N
13247	1-13247	40	68	5504	5	9983	9586	6	2012-08-02 11:25:25.777634	2012-08-02 11:27:56.667174
12982	1-12982	2	23	3245	53	1696	1131	5	2012-08-02 11:23:41.80718	2012-08-02 11:27:57.176007
13395	1-13395	82	25	8095	57	9718	2895	0	2012-08-02 11:26:24.64072	2012-08-02 11:26:55.649625
13239	1-13239	10	63	4081	23	6994	574	18	2012-08-02 11:25:22.619491	2012-08-02 11:29:53.255963
13285	1-13285	50	97	6484	78	3536	3301	8	2012-08-02 11:25:40.741695	2012-08-02 11:25:53.587928
12967	1-12967	6	51	9807	85	1821	1821	2	2012-08-02 11:23:35.918539	\N
13784	1-13784	68	65	3249	51	4415	257	7	2012-08-02 11:28:59.50984	2012-08-02 11:30:00.36048
13733	1-13733	91	49	3538	87	8845	5925	9	2012-08-02 11:28:38.988139	2012-08-02 11:30:05.056251
13821	1-13821	60	44	5873	67	4906	4238	0	2012-08-02 11:29:14.049526	2012-08-02 11:30:17.213447
13778	1-13778	22	88	2956	60	2898	1737	6	2012-08-02 11:28:57.123439	2012-08-02 11:30:20.7384
12973	1-12973	69	47	5722	69	2835	2835	0	2012-08-02 11:23:38.271707	\N
12976	1-12976	36	26	5098	91	3578	3578	0	2012-08-02 11:23:39.453818	\N
13419	1-13419	43	61	3154	53	2644	713	0	2012-08-02 11:26:34.486761	2012-08-02 11:30:21.125048
13368	1-13368	92	48	7844	50	4857	3789	0	2012-08-02 11:26:14.574397	2012-08-02 11:30:33.358312
13852	1-13852	6	2	6444	46	7983	3973	3	2012-08-02 11:29:26.338552	2012-08-02 11:30:21.901036
12984	1-12984	77	28	7365	69	6630	6630	0	2012-08-02 11:23:42.967297	\N
13772	1-13772	37	4	6362	65	7776	183	13	2012-08-02 11:28:54.736808	2012-08-02 11:30:39.704933
13868	1-13868	49	51	3189	54	1617	1617	9	2012-08-02 11:29:32.658359	\N
13357	1-13357	14	28	4016	67	4713	4475	3	2012-08-02 11:26:09.745443	2012-08-02 11:27:16.943419
13324	1-13324	31	95	9191	99	7186	270	4	2012-08-02 11:25:55.986268	2012-08-02 11:30:40.501224
13377	1-13377	8	48	9832	71	9704	5768	0	2012-08-02 11:26:17.889248	2012-08-02 11:28:53.543122
13726	1-13726	53	14	7576	71	6974	6974	0	2012-08-02 11:28:36.358662	\N
13265	1-13265	26	19	6463	30	668	668	12	2012-08-02 11:25:32.903161	\N
13729	1-13729	5	60	4150	71	1737	1737	2	2012-08-02 11:28:37.817072	\N
13259	1-13259	91	62	7917	24	2234	1716	0	2012-08-02 11:25:30.947394	2012-08-02 11:28:53.543122
13329	1-13329	45	63	2075	44	4731	4731	0	2012-08-02 11:25:58.568789	\N
13787	1-13787	15	48	6220	15	8134	8134	14	2012-08-02 11:29:00.691809	\N
13883	1-13883	66	36	3122	23	1555	1555	9	2012-08-02 11:29:38.580563	\N
13775	1-13775	78	32	3105	20	4170	2535	12	2012-08-02 11:28:55.930098	2012-08-02 11:28:55.930098
13416	1-13416	40	96	3584	37	1366	1366	0	2012-08-02 11:26:32.951163	\N
13351	1-13351	23	23	8539	30	5131	4185	11	2012-08-02 11:26:07.366787	2012-08-02 11:26:10.938883
13337	1-13337	47	77	9317	11	2468	2468	0	2012-08-02 11:26:01.864163	\N
13723	1-13723	80	40	7940	84	9059	425	5	2012-08-02 11:28:35.209789	2012-08-02 11:29:09.320977
13793	1-13793	1	37	4312	70	393	393	6	2012-08-02 11:29:03.045238	\N
13796	1-13796	41	33	9258	99	2528	2528	8	2012-08-02 11:29:04.238652	\N
13348	1-13348	63	43	5867	71	4925	4925	8	2012-08-02 11:26:06.195866	\N
13803	1-13803	75	53	4015	90	9696	9696	12	2012-08-02 11:29:06.547588	\N
13391	1-13391	82	63	4436	5	5403	5403	7	2012-08-02 11:26:23.436255	\N
13830	1-13830	11	23	8107	18	9986	9548	9	2012-08-02 11:29:17.587342	2012-08-02 11:29:17.587342
13379	1-13379	84	8	8440	45	6824	6824	8	2012-08-02 11:26:19.082022	\N
13840	1-13840	76	66	8648	72	6203	6203	0	2012-08-02 11:29:21.576469	\N
13843	1-13843	7	1	8989	42	8192	8192	19	2012-08-02 11:29:22.792138	\N
13425	1-13425	10	30	7506	45	1242	1242	0	2012-08-02 11:26:36.85093	\N
13407	1-13407	50	79	4467	73	286	286	0	2012-08-02 11:26:29.767531	\N
13863	1-13863	12	18	6029	39	1141	1141	10	2012-08-02 11:29:31.078243	\N
13409	1-13409	6	86	5544	24	709	709	12	2012-08-02 11:26:30.974019	\N
13412	1-13412	71	81	7586	29	669	669	0	2012-08-02 11:26:31.780158	\N
13858	1-13858	82	79	6771	8	3951	3951	14	2012-08-02 11:29:28.714148	\N
13430	1-13430	72	82	6434	34	1461	1461	4	2012-08-02 11:26:39.237405	\N
13871	1-13871	60	98	5777	14	9444	9444	0	2012-08-02 11:29:33.863491	\N
13331	1-13331	67	47	8810	91	9465	7356	9	2012-08-02 11:25:59.897308	2012-08-02 11:26:41.226091
13439	1-13439	6	11	9638	61	1427	1427	7	2012-08-02 11:26:42.430595	\N
13877	1-13877	21	75	3362	4	9198	6660	2	2012-08-02 11:29:36.193916	2012-08-02 11:29:36.193916
13880	1-13880	30	36	7229	7	7957	7957	11	2012-08-02 11:29:37.409586	\N
13855	1-13855	4	64	9825	3	3665	3143	4	2012-08-02 11:29:27.565646	2012-08-02 11:29:35.033431
13320	1-13320	76	68	6927	78	6266	3486	26	2012-08-02 11:25:54.814968	2012-08-02 11:29:35.033431
13422	1-13422	40	87	8889	42	6679	6106	9	2012-08-02 11:26:35.690851	2012-08-02 11:29:35.795893
13278	1-13278	88	96	1312	93	2072	268	0	2012-08-02 11:25:37.633451	2012-08-02 11:29:35.795893
13889	1-13889	97	21	1745	52	2018	2018	5	2012-08-02 11:29:40.978213	\N
13742	1-13742	35	39	2515	59	3238	3238	0	2012-08-02 11:28:42.514081	\N
13370	1-13370	4	9	8024	40	2296	2296	6	2012-08-02 11:26:14.949928	\N
13408	1-13408	78	95	6973	86	7520	3487	7	2012-08-02 11:26:30.573915	2012-08-02 11:28:44.878308
13406	1-13406	56	86	4902	36	1962	1134	5	2012-08-02 11:26:29.358487	2012-08-02 11:28:44.878308
13273	1-13273	96	41	9729	78	329	329	4	2012-08-02 11:25:36.063923	\N
13435	1-13435	51	73	8998	61	9097	7494	0	2012-08-02 11:26:40.817119	2012-08-02 11:26:56.88707
13343	1-13343	56	48	8510	43	4513	1480	7	2012-08-02 11:26:03.864168	2012-08-02 11:28:48.812571
12963	1-12963	97	68	8622	59	491	491	2	2012-08-02 11:23:34.360614	\N
13277	1-13277	43	14	4232	1	532	532	0	2012-08-02 11:25:38.020295	\N
13327	1-13327	94	29	2826	41	4624	4426	10	2012-08-02 11:25:57.201399	2012-08-02 11:28:53.156625
13853	1-13853	25	46	2834	54	4437	4437	2	2012-08-02 11:29:27.145031	\N
13894	1-13894	68	100	7994	17	9154	7341	10	2012-08-02 11:29:43.388791	2012-08-02 11:30:34.543923
13281	1-13281	44	94	7047	27	6717	6717	7	2012-08-02 11:25:38.784198	\N
13321	1-13321	48	11	6066	9	2615	1968	9	2012-08-02 11:25:55.201326	2012-08-02 11:26:15.336549
13848	1-13848	70	7	9884	93	8860	2984	5	2012-08-02 11:29:24.79138	2012-08-02 11:29:55.59852
13284	1-13284	17	22	8101	48	920	920	2	2012-08-02 11:25:40.34389	\N
13882	1-13882	24	36	9217	98	8367	8367	0	2012-08-02 11:29:38.182827	\N
13275	1-13275	84	100	5492	56	5309	4178	3	2012-08-02 11:25:37.234852	2012-08-02 11:30:00.36048
13318	1-13318	82	66	8323	50	9310	7977	6	2012-08-02 11:25:54.405335	2012-08-02 11:29:01.863128
13373	1-13373	54	39	3771	84	6941	6941	5	2012-08-02 11:26:16.143469	\N
12981	1-12981	74	9	5266	49	1857	1857	9	2012-08-02 11:23:41.409626	\N
13800	1-13800	38	21	5209	97	8934	3280	5	2012-08-02 11:29:06.182748	2012-08-02 11:30:12.885015
13856	1-13856	99	91	9198	65	8958	7643	6	2012-08-02 11:29:27.940484	2012-08-02 11:29:27.940484
13884	1-13884	63	70	7450	66	6121	2849	12	2012-08-02 11:29:39.376229	2012-08-02 11:30:16.762619
13316	1-13316	76	6	9581	80	9756	9756	0	2012-08-02 11:25:53.190002	\N
13420	1-13420	93	37	6507	13	3519	1841	0	2012-08-02 11:26:34.895537	2012-08-02 11:30:16.762619
13411	1-13411	92	86	9728	87	5620	5620	0	2012-08-02 11:26:31.382732	\N
13330	1-13330	45	53	7506	71	4478	4478	5	2012-08-02 11:25:59.102206	\N
13805	1-13805	64	57	8799	54	5213	5213	11	2012-08-02 11:29:07.740825	\N
13333	1-13333	63	59	8208	84	1202	1202	0	2012-08-02 11:26:00.284318	\N
13808	1-13808	69	38	4566	40	6057	6057	4	2012-08-02 11:29:08.934293	\N
13367	1-13367	43	25	8305	2	9045	6984	4	2012-08-02 11:26:13.778633	2012-08-02 11:28:34.027524
13811	1-13811	48	13	4855	49	255	255	0	2012-08-02 11:29:10.115875	\N
13376	1-13376	49	77	9259	6	5409	1644	0	2012-08-02 11:26:17.480661	2012-08-02 11:29:10.513766
13859	1-13859	68	53	8437	86	410	410	3	2012-08-02 11:29:29.111593	\N
13336	1-13336	95	81	8353	50	7720	5795	5	2012-08-02 11:26:01.466048	2012-08-02 11:29:12.082503
13356	1-13356	35	57	9704	79	1773	1773	0	2012-08-02 11:26:09.345245	\N
13864	1-13864	74	89	7249	91	4148	4148	6	2012-08-02 11:29:30.272035	\N
13358	1-13358	15	3	9977	15	4201	4201	3	2012-08-02 11:26:10.519112	\N
13820	1-13820	41	75	5835	96	1906	1906	3	2012-08-02 11:29:13.662868	\N
13823	1-13823	34	26	5925	10	861	861	3	2012-08-02 11:29:14.834103	\N
13839	1-13839	75	80	4408	9	1109	1109	4	2012-08-02 11:29:21.189416	\N
13417	1-13417	57	79	9059	49	2932	2932	0	2012-08-02 11:26:33.713473	\N
13867	1-13867	23	75	3569	13	3795	3795	0	2012-08-02 11:29:32.249171	\N
13845	1-13845	57	37	2114	84	1418	1418	5	2012-08-02 11:29:23.587338	\N
13423	1-13423	29	88	4467	61	1420	1420	5	2012-08-02 11:26:36.078057	\N
13851	1-13851	64	12	3277	70	9664	9664	0	2012-08-02 11:29:25.951658	\N
13870	1-13870	10	53	5898	69	3413	3413	10	2012-08-02 11:29:33.45364	\N
13873	1-13873	23	44	8457	53	7235	5642	9	2012-08-02 11:29:34.635924	2012-08-02 11:29:34.635924
13400	1-13400	92	58	8882	38	5269	5269	6	2012-08-02 11:26:26.993903	\N
13836	1-13836	50	63	3937	38	5800	3908	4	2012-08-02 11:29:19.609395	2012-08-02 11:29:39.376229
13432	1-13432	1	37	6416	52	2854	2854	6	2012-08-02 11:26:39.646136	\N
12966	1-12966	53	78	3441	42	1927	1152	0	2012-08-02 11:23:35.509715	2012-08-02 11:29:38.989309
13438	1-13438	69	20	4963	65	1997	1997	3	2012-08-02 11:26:42.02141	\N
13887	1-13887	100	47	9004	30	6552	6552	0	2012-08-02 11:29:40.160452	\N
13890	1-13890	87	22	4923	82	7423	7423	0	2012-08-02 11:29:41.364594	\N
13892	1-13892	47	6	7599	95	5579	5579	0	2012-08-02 11:29:42.580265	\N
13341	1-13341	45	70	5933	83	5577	3494	0	2012-08-02 11:26:03.466559	2012-08-02 11:26:26.596176
13755	1-13755	28	74	9963	31	6955	3230	5	2012-08-02 11:28:48.027849	2012-08-02 11:29:48.152049
13344	1-13344	42	31	2556	87	6563	2235	12	2012-08-02 11:26:04.648552	2012-08-02 11:29:48.152049
13888	1-13888	39	20	7233	100	6646	5803	21	2012-08-02 11:29:40.569458	2012-08-02 11:29:46.141041
13810	1-13810	68	94	6010	78	1562	1562	6	2012-08-02 11:29:09.718131	\N
13822	1-13822	48	12	7340	10	8996	8494	3	2012-08-02 11:29:14.425123	2012-08-02 11:29:46.549736
13289	1-13289	77	30	9239	70	4092	4092	0	2012-08-02 11:25:42.365805	\N
13431	1-13431	85	29	7968	20	9907	9578	0	2012-08-02 11:26:38.828538	2012-08-02 11:28:52.382745
13322	1-13322	13	28	4717	78	31	31	0	2012-08-02 11:25:55.58817	\N
13767	1-13767	65	4	6488	85	417	417	5	2012-08-02 11:28:52.76995	\N
13869	1-13869	50	39	5095	28	4707	4469	0	2012-08-02 11:29:33.045202	2012-08-02 11:29:54.029273
13771	1-13771	4	35	8597	97	1197	1197	0	2012-08-02 11:28:53.92988	\N
13418	1-13418	50	72	9929	67	5767	1591	2	2012-08-02 11:26:34.100074	2012-08-02 11:27:20.169976
13328	1-13328	92	11	6855	1	2263	2263	11	2012-08-02 11:25:58.032821	\N
13813	1-13813	72	75	7786	98	7962	7962	3	2012-08-02 11:29:10.900325	\N
13410	1-13410	22	50	9280	25	9882	8059	13	2012-08-02 11:26:30.154719	2012-08-02 11:27:45.695868
13783	1-13783	94	43	3611	30	7380	800	6	2012-08-02 11:28:59.111716	2012-08-02 11:30:08.195798
13828	1-13828	86	91	6334	47	5868	5130	9	2012-08-02 11:29:16.789942	2012-08-02 11:30:10.516962
13375	1-13375	34	60	8790	17	6056	464	0	2012-08-02 11:26:16.939325	2012-08-02 11:27:57.794761
13338	1-13338	100	94	9519	21	5444	5444	0	2012-08-02 11:26:02.262347	\N
13774	1-13774	33	70	5974	65	2256	2256	5	2012-08-02 11:28:55.521109	\N
13743	1-13743	87	11	8834	79	2481	2481	6	2012-08-02 11:28:43.319672	\N
13857	1-13857	82	9	1089	14	5872	377	0	2012-08-02 11:29:28.31607	2012-08-02 11:30:39.704933
13780	1-13780	63	69	3909	85	2002	2002	8	2012-08-02 11:28:57.95214	\N
13359	1-13359	4	45	2213	69	331	331	3	2012-08-02 11:26:10.121113	\N
13362	1-13362	58	55	8359	36	2512	2512	0	2012-08-02 11:26:11.71168	\N
13777	1-13777	34	48	5265	76	7221	351	2	2012-08-02 11:28:56.703618	2012-08-02 11:29:12.470004
13413	1-13413	75	41	9093	98	6017	6017	3	2012-08-02 11:26:32.155634	\N
13875	1-13875	49	98	9968	75	3879	3879	0	2012-08-02 11:29:35.409114	\N
13372	1-13372	52	4	7455	69	5838	5838	5	2012-08-02 11:26:15.723212	\N
13378	1-13378	1	7	9626	91	764	764	19	2012-08-02 11:26:18.275591	\N
13752	1-13752	18	96	6384	44	3059	3059	10	2012-08-02 11:28:46.845514	\N
13415	1-13415	55	31	8679	16	368	368	4	2012-08-02 11:26:33.327006	\N
13383	1-13383	24	45	3975	57	272	272	0	2012-08-02 11:26:20.640147	\N
13332	1-13332	55	62	4867	68	2536	805	18	2012-08-02 11:25:59.499899	2012-08-02 11:28:49.950697
13390	1-13390	84	22	9422	39	1265	1265	0	2012-08-02 11:26:23.038311	\N
13393	1-13393	73	24	8237	19	2866	2866	6	2012-08-02 11:26:24.242671	\N
13789	1-13789	47	8	5553	4	4942	4942	2	2012-08-02 11:29:01.476206	\N
13764	1-13764	25	63	2965	68	2493	1917	0	2012-08-02 11:28:51.543231	2012-08-02 11:29:12.856069
13797	1-13797	81	43	9808	73	3962	3962	0	2012-08-02 11:29:04.625533	\N
13437	1-13437	38	16	6524	38	3599	189	8	2012-08-02 11:26:41.601755	2012-08-02 11:29:05.409457
13819	1-13819	54	53	1321	5	867	867	3	2012-08-02 11:29:13.265052	\N
13804	1-13804	95	3	9387	42	3063	3063	0	2012-08-02 11:29:07.343025	\N
13831	1-13831	47	62	5296	46	9191	9191	0	2012-08-02 11:29:17.985031	\N
13847	1-13847	25	34	4429	84	905	905	0	2012-08-02 11:29:24.382611	\N
13807	1-13807	96	35	4603	10	7765	6075	0	2012-08-02 11:29:08.547438	2012-08-02 11:29:18.371887
13421	1-13421	37	1	5804	81	5215	4842	0	2012-08-02 11:26:35.282213	2012-08-02 11:29:15.971766
13854	1-13854	52	100	9817	38	2108	2108	9	2012-08-02 11:29:26.73639	\N
13844	1-13844	51	97	6834	9	1819	1819	5	2012-08-02 11:29:23.189628	\N
13860	1-13860	47	29	7263	67	2901	2901	17	2012-08-02 11:29:29.498248	\N
13862	1-13862	8	81	911	92	1103	1103	13	2012-08-02 11:29:30.658243	\N
13866	1-13866	16	58	6221	18	9423	9423	4	2012-08-02 11:29:31.4649	\N
13872	1-13872	54	17	4077	68	779	779	0	2012-08-02 11:29:34.249385	\N
13891	1-13891	45	29	3259	61	3736	3736	0	2012-08-02 11:29:41.751628	\N
13881	1-13881	41	32	1177	12	20	20	3	2012-08-02 11:29:37.796298	\N
13850	1-13850	24	75	7445	74	6369	1883	6	2012-08-02 11:29:25.576146	2012-08-02 11:29:38.989309
\.


--
-- Data for Name: torderremoved; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY torderremoved (id, uuid, own, nr, qtt_requ, np, qtt_prov, qtt, start, created, updated) FROM stdin;
11	1-11	11	16	805	19	4480	0	0	2012-08-02 04:43:53.553726	2012-08-02 04:43:54.314959
2	1-2	2	4	3034	3	7838	0	0	2012-08-02 04:43:53.420296	2012-08-02 04:43:55.251645
65	1-65	46	41	4313	71	7772	0	0	2012-08-02 04:43:54.180956	2012-08-02 04:43:55.286995
62	1-62	21	72	847	71	659	0	0	2012-08-02 04:43:54.134344	2012-08-02 04:43:55.398174
44	1-44	36	58	835	36	6331	0	0	2012-08-02 04:43:53.933607	2012-08-02 04:43:55.565418
107	1-107	36	80	5809	28	1026	0	0	2012-08-02 04:43:54.682929	2012-08-02 04:43:55.632322
81	1-81	39	40	3	19	6570	0	0	2012-08-02 04:43:54.381731	2012-08-02 04:43:55.877365
109	1-109	61	88	463	87	4198	0	0	2012-08-02 04:43:54.705389	2012-08-02 04:43:55.877365
23	1-23	23	23	453	11	6977	0	0	2012-08-02 04:43:53.698799	2012-08-02 04:43:55.877365
176	1-176	83	81	220	87	1513	0	0	2012-08-02 04:43:55.610193	2012-08-02 04:43:55.877365
87	1-87	7	5	325	81	1514	0	0	2012-08-02 04:43:54.448768	2012-08-02 04:43:56.066747
3	1-3	3	6	2819	5	5047	0	0	2012-08-02 04:43:53.453622	2012-08-02 04:43:56.066747
103	1-103	12	73	72	87	1232	0	0	2012-08-02 04:43:54.638369	2012-08-02 04:43:56.122278
197	1-197	56	8	91	49	7474	0	0	2012-08-02 04:43:55.92221	2012-08-02 04:43:56.122278
160	1-160	64	24	7361	72	2846	0	0	2012-08-02 04:43:55.398174	2012-08-02 04:43:56.178066
136	1-136	56	28	4218	90	1683	0	0	2012-08-02 04:43:55.006693	2012-08-02 04:43:56.178066
100	1-100	53	86	786	41	7226	0	0	2012-08-02 04:43:54.604966	2012-08-02 04:43:56.467323
111	1-111	23	76	6370	46	4222	0	0	2012-08-02 04:43:54.72775	2012-08-02 04:43:56.467323
196	1-196	63	11	3534	28	6814	0	0	2012-08-02 04:43:55.877365	2012-08-02 04:43:56.467323
53	1-53	3	66	3725	65	6402	0	0	2012-08-02 04:43:54.034095	2012-08-02 04:43:56.556197
102	1-102	45	65	2080	44	8935	0	0	2012-08-02 04:43:54.627293	2012-08-02 04:43:56.600607
7	1-7	7	13	2605	12	8654	0	0	2012-08-02 04:43:53.498021	2012-08-02 04:43:56.656361
234	1-234	48	32	228	48	3169	0	0	2012-08-02 04:43:56.713905	2012-08-02 04:43:56.713905
45	1-45	2	37	3162	59	6786	0	0	2012-08-02 04:43:53.94454	2012-08-02 04:43:56.76951
175	1-175	46	41	3050	66	6716	0	0	2012-08-02 04:43:55.598751	2012-08-02 04:43:56.858115
152	1-152	45	55	7247	13	4610	0	0	2012-08-02 04:43:55.251645	2012-08-02 04:43:56.858115
168	1-168	83	41	7600	77	8162	0	0	2012-08-02 04:43:55.509949	2012-08-02 04:43:56.858115
239	1-239	52	53	1905	41	8025	0	0	2012-08-02 04:43:56.858115	2012-08-02 04:43:56.858115
228	1-228	3	3	387	79	5528	0	0	2012-08-02 04:43:56.556197	2012-08-02 04:43:57.028272
250	1-250	93	64	135	92	5650	0	0	2012-08-02 04:43:57.113934	2012-08-02 04:43:57.113934
98	1-98	12	82	993	32	4157	0	0	2012-08-02 04:43:54.571513	2012-08-02 04:43:57.169427
188	1-188	61	41	2362	32	1509	0	0	2012-08-02 04:43:55.766299	2012-08-02 04:43:57.213833
165	1-165	84	87	5855	38	3279	0	0	2012-08-02 04:43:55.465407	2012-08-02 04:43:57.358117
219	1-219	34	42	45	31	3475	0	0	2012-08-02 04:43:56.345024	2012-08-02 04:43:57.480325
262	1-262	60	52	3405	77	1614	0	0	2012-08-02 04:43:57.402907	2012-08-02 04:43:57.54682
236	1-236	75	59	7676	41	523	0	0	2012-08-02 04:43:56.76951	2012-08-02 04:43:57.735435
54	1-54	42	54	1511	67	89	0	0	2012-08-02 04:43:54.045179	2012-08-02 04:43:57.735435
269	1-269	66	65	1693	100	5540	0	0	2012-08-02 04:43:57.735435	2012-08-02 04:43:57.735435
252	1-252	76	31	7772	5	382	0	0	2012-08-02 04:43:57.169427	2012-08-02 04:43:57.834927
21	1-21	21	32	2431	31	8451	0	0	2012-08-02 04:43:53.676535	2012-08-02 04:43:57.912481
138	1-138	14	13	3889	53	7894	0	0	2012-08-02 04:43:55.029039	2012-08-02 04:43:57.912481
104	1-104	60	31	1870	68	8592	0	0	2012-08-02 04:43:54.649613	2012-08-02 04:43:57.912481
137	1-137	14	38	5242	11	7089	0	0	2012-08-02 04:43:55.017873	2012-08-02 04:43:58.145536
83	1-83	2	79	623	38	8101	0	0	2012-08-02 04:43:54.40405	2012-08-02 04:43:58.145536
193	1-193	42	9	2898	65	7760	0	0	2012-08-02 04:43:55.832963	2012-08-02 04:43:58.145536
157	1-157	81	12	2856	65	5571	0	0	2012-08-02 04:43:55.34261	2012-08-02 04:43:58.145536
48	1-48	38	63	814	62	3967	0	0	2012-08-02 04:43:53.978762	2012-08-02 04:43:58.278954
178	1-178	80	90	2166	41	8925	0	0	2012-08-02 04:43:55.632322	2012-08-02 04:43:58.278954
225	1-225	21	3	3447	76	9497	0	0	2012-08-02 04:43:56.467323	2012-08-02 04:43:58.445391
112	1-112	62	74	579	52	9490	0	0	2012-08-02 04:43:54.738691	2012-08-02 04:43:58.478561
113	1-113	63	22	2040	4	3207	0	0	2012-08-02 04:43:54.749987	2012-08-02 04:43:58.478561
210	1-210	21	32	4970	17	6279	0	0	2012-08-02 04:43:56.178066	2012-08-02 04:43:58.478561
270	1-270	88	97	3570	32	8190	0	0	2012-08-02 04:43:57.834927	2012-08-02 04:43:58.478561
8	1-8	8	15	3989	14	7198	0	0	2012-08-02 04:43:53.520288	2012-08-02 04:43:58.589497
212	1-212	14	19	4497	13	7441	0	0	2012-08-02 04:43:56.234536	2012-08-02 04:43:58.755592
40	1-40	2	14	4183	55	7019	0	0	2012-08-02 04:43:53.888807	2012-08-02 04:43:58.855999
68	1-68	49	70	1980	12	1464	0	0	2012-08-02 04:43:54.214343	2012-08-02 04:43:58.94642
56	1-56	44	12	3948	22	2896	0	0	2012-08-02 04:43:54.067904	2012-08-02 04:43:58.94642
139	1-139	22	3	9401	55	1962	0	0	2012-08-02 04:43:55.039909	2012-08-02 04:43:59.168413
308	1-308	39	55	2338	53	3642	0	0	2012-08-02 04:43:59.168413	2012-08-02 04:43:59.168413
286	1-286	56	4	4556	33	5068	0	0	2012-08-02 04:43:58.478561	2012-08-02 04:43:59.312683
258	1-258	74	5	218	4	3718	0	0	2012-08-02 04:43:57.313962	2012-08-02 04:43:59.312683
135	1-135	22	33	2880	83	7517	0	0	2012-08-02 04:43:54.995531	2012-08-02 04:43:59.312683
231	1-231	33	12	5560	15	4765	0	0	2012-08-02 04:43:56.656361	2012-08-02 04:43:59.312683
31	1-31	29	44	581	25	9666	0	0	2012-08-02 04:43:53.788368	2012-08-02 04:43:59.412629
202	1-202	80	43	1739	24	9498	0	0	2012-08-02 04:43:56.010891	2012-08-02 04:43:59.591311
312	1-312	11	91	3813	5	4359	0	0	2012-08-02 04:43:59.312683	2012-08-02 04:43:59.591311
173	1-173	58	35	592	58	3841	0	0	2012-08-02 04:43:55.565418	2012-08-02 04:43:59.591311
125	1-125	69	1	2897	23	4149	0	0	2012-08-02 04:43:54.88386	2012-08-02 04:43:59.712447
323	1-323	53	22	876	1	5224	0	0	2012-08-02 04:43:59.712447	2012-08-02 04:43:59.712447
126	1-126	70	34	4109	74	5314	0	0	2012-08-02 04:43:54.89511	2012-08-02 04:43:59.867737
101	1-101	26	61	100	8	3479	0	0	2012-08-02 04:43:54.616064	2012-08-02 04:43:59.967494
170	1-170	28	100	127	24	4557	0	0	2012-08-02 04:43:55.531937	2012-08-02 04:43:59.967494
260	1-260	6	59	488	87	8777	0	0	2012-08-02 04:43:57.358117	2012-08-02 04:44:00.144891
335	1-335	1	6	1466	96	4686	0	0	2012-08-02 04:44:00.233744	2012-08-02 04:44:00.233744
24	1-24	24	34	267	21	6806	0	0	2012-08-02 04:43:53.710082	2012-08-02 04:44:00.289158
338	1-338	96	29	42	88	2105	0	0	2012-08-02 04:44:00.344457	2012-08-02 04:44:00.344457
85	1-85	51	44	7172	80	8634	0	0	2012-08-02 04:43:54.426496	2012-08-02 04:44:00.455436
215	1-215	90	2	110	91	302	0	0	2012-08-02 04:43:56.278456	2012-08-02 04:44:00.455436
131	1-131	31	87	6212	4	525	0	0	2012-08-02 04:43:54.950709	2012-08-02 04:44:00.754537
240	1-240	57	5	1771	24	3508	0	0	2012-08-02 04:43:56.902907	2012-08-02 04:44:00.821366
299	1-299	2	24	4710	70	5003	0	0	2012-08-02 04:43:58.855999	2012-08-02 04:44:00.865562
49	1-49	39	56	1335	24	8614	0	0	2012-08-02 04:43:53.989285	2012-08-02 04:44:00.865562
184	1-184	26	67	1689	9	8065	0	0	2012-08-02 04:43:55.721212	2012-08-02 04:44:00.965579
300	1-300	20	74	3593	91	4320	0	0	2012-08-02 04:43:58.902253	2012-08-02 04:44:01.031584
115	1-115	39	91	3988	90	4462	0	0	2012-08-02 04:43:54.772413	2012-08-02 04:44:01.031584
19	1-19	19	14	4506	29	5797	0	0	2012-08-02 04:43:53.654241	2012-08-02 04:44:01.153643
362	1-362	36	90	2411	85	1742	0	0	2012-08-02 04:44:01.386358	2012-08-02 04:44:01.386358
50	1-50	30	48	2785	55	8719	0	0	2012-08-02 04:43:54.000537	2012-08-02 04:44:01.554638
316	1-316	25	19	2017	38	6601	0	0	2012-08-02 04:43:59.523422	2012-08-02 04:44:01.554638
35	1-35	31	50	457	17	3360	0	0	2012-08-02 04:43:53.832979	2012-08-02 04:44:01.698628
140	1-140	75	40	847	95	2341	0	0	2012-08-02 04:43:55.0634	2012-08-02 04:44:01.787236
350	1-350	56	57	1615	87	5523	0	0	2012-08-02 04:44:00.754537	2012-08-02 04:44:01.942108
25	1-25	25	36	3702	35	3913	0	0	2012-08-02 04:43:53.721261	2012-08-02 04:44:02.01968
57	1-57	7	45	483	56	2398	0	0	2012-08-02 04:43:54.078491	2012-08-02 04:44:02.01968
319	1-319	56	17	2336	30	8220	0	0	2012-08-02 04:43:59.645562	2012-08-02 04:44:02.01968
370	1-370	68	17	1246	50	5255	0	0	2012-08-02 04:44:01.698628	2012-08-02 04:44:02.01968
340	1-340	79	52	153	6	4928	0	0	2012-08-02 04:44:00.388985	2012-08-02 04:44:02.285657
318	1-318	87	70	4163	49	3399	0	0	2012-08-02 04:43:59.591311	2012-08-02 04:44:02.285657
374	1-374	99	51	8836	17	4174	0	0	2012-08-02 04:44:02.01968	2012-08-02 04:44:02.285657
206	1-206	15	25	7916	6	7826	0	0	2012-08-02 04:43:56.066747	2012-08-02 04:44:02.285657
292	1-292	9	6	5583	57	8621	0	0	2012-08-02 04:43:58.700383	2012-08-02 04:44:02.285657
41	1-41	14	56	1577	30	2593	0	0	2012-08-02 04:43:53.899939	2012-08-02 04:44:02.485329
163	1-163	22	1	2402	44	8318	0	0	2012-08-02 04:43:55.442814	2012-08-02 04:44:02.485329
382	1-382	64	78	454	1	2099	0	0	2012-08-02 04:44:02.485329	2012-08-02 04:44:02.485329
199	1-199	60	75	11	55	9875	0	0	2012-08-02 04:43:55.955654	2012-08-02 04:44:02.728964
124	1-124	59	17	5302	11	3399	0	0	2012-08-02 04:43:54.87304	2012-08-02 04:44:02.728964
75	1-75	47	55	2251	48	395	0	0	2012-08-02 04:43:54.292656	2012-08-02 04:44:02.728964
129	1-129	72	52	4727	4	3541	0	0	2012-08-02 04:43:54.928662	2012-08-02 04:44:02.884051
388	1-388	11	34	3491	53	9131	0	0	2012-08-02 04:44:03.016855	2012-08-02 04:44:03.016855
290	1-290	46	75	1109	8	2297	0	0	2012-08-02 04:43:58.678415	2012-08-02 04:44:03.42941
166	1-166	85	86	5105	91	9758	0	0	2012-08-02 04:43:55.487364	2012-08-02 04:44:03.636911
405	1-405	43	61	1382	24	3392	0	0	2012-08-02 04:44:03.991125	2012-08-02 04:44:03.991125
243	1-243	46	15	958	54	6480	0	0	2012-08-02 04:43:56.947307	2012-08-02 04:44:04.013213
63	1-63	9	46	5191	30	3939	0	0	2012-08-02 04:43:54.14558	2012-08-02 04:44:04.25691
179	1-179	86	72	9868	53	3289	0	0	2012-08-02 04:43:55.654608	2012-08-02 04:44:04.511423
315	1-315	18	24	6209	70	6924	0	3	2012-08-02 04:43:59.49008	2012-08-02 04:44:04.511423
384	1-384	93	48	5097	75	6269	0	0	2012-08-02 04:44:02.728964	2012-08-02 04:44:04.511423
352	1-352	96	23	677	56	7907	0	0	2012-08-02 04:44:00.865562	2012-08-02 04:44:04.511423
293	1-293	73	84	5596	68	9313	0	0	2012-08-02 04:43:58.755592	2012-08-02 04:44:04.688933
415	1-415	25	52	488	11	9863	0	0	2012-08-02 04:44:04.688933	2012-08-02 04:44:04.688933
218	1-218	91	56	4738	55	244	0	0	2012-08-02 04:43:56.333794	2012-08-02 04:44:04.882974
200	1-200	57	30	9487	51	9294	0	0	2012-08-02 04:43:55.967804	2012-08-02 04:44:04.882974
26	1-26	26	38	1850	37	9611	0	0	2012-08-02 04:43:53.732376	2012-08-02 04:44:04.882974
367	1-367	11	38	4489	70	4895	0	0	2012-08-02 04:44:01.554638	2012-08-02 04:44:04.882974
298	1-298	74	38	3844	82	3592	0	0	2012-08-02 04:43:58.845751	2012-08-02 04:44:04.882974
61	1-61	17	70	6268	61	9197	1286	0	2012-08-02 04:43:54.123479	2012-08-02 04:44:04.882974
420	1-420	4	34	90	95	4032	0	0	2012-08-02 04:44:05.225624	2012-08-02 04:44:05.225624
421	1-421	71	50	103	82	702	0	0	2012-08-02 04:44:05.392042	2012-08-02 04:44:05.392042
329	1-329	65	37	5708	46	8989	0	0	2012-08-02 04:43:59.923106	2012-08-02 04:44:05.480499
336	1-336	85	28	5406	20	4639	0	0	2012-08-02 04:44:00.277881	2012-08-02 04:44:05.480499
280	1-280	6	2	5843	30	4527	0	0	2012-08-02 04:43:58.267585	2012-08-02 04:44:05.480499
406	1-406	18	54	6090	28	9771	0	0	2012-08-02 04:44:04.013213	2012-08-02 04:44:05.480499
407	1-407	45	33	1596	2	8042	0	0	2012-08-02 04:44:04.068682	2012-08-02 04:44:05.735273
268	1-268	85	9	7108	82	762	0	0	2012-08-02 04:43:57.646599	2012-08-02 04:44:05.735273
412	1-412	27	55	748	48	5361	0	3	2012-08-02 04:44:04.511423	2012-08-02 04:44:05.735273
390	1-390	1	1	2164	19	1889	0	0	2012-08-02 04:44:03.171674	2012-08-02 04:44:05.735273
77	1-77	53	19	9011	23	5050	1345	0	2012-08-02 04:43:54.314959	2012-08-02 04:44:05.735273
423	1-423	31	23	1454	71	3234	0	0	2012-08-02 04:44:05.735273	2012-08-02 04:44:05.968063
253	1-253	71	87	3799	90	6736	0	0	2012-08-02 04:43:57.213833	2012-08-02 04:44:06.345364
443	1-443	19	34	3378	16	2382	0	0	2012-08-02 04:44:06.345364	2012-08-02 04:44:06.345364
446	1-446	60	57	888	37	7100	0	0	2012-08-02 04:44:06.522533	2012-08-02 04:44:06.522533
333	1-333	44	9	3382	43	8855	0	0	2012-08-02 04:44:00.056495	2012-08-02 04:44:06.589075
121	1-121	57	2	406	56	8838	0	0	2012-08-02 04:43:54.839286	2012-08-02 04:44:06.589075
20	1-20	20	5	824	30	7934	0	0	2012-08-02 04:43:53.665317	2012-08-02 04:44:06.589075
448	1-448	41	43	6040	89	9566	0	0	2012-08-02 04:44:06.589075	2012-08-02 04:44:06.589075
144	1-144	12	20	4903	37	5810	0	0	2012-08-02 04:43:55.12909	2012-08-02 04:44:07.264086
427	1-427	76	80	705	53	1482	0	0	2012-08-02 04:44:05.968063	2012-08-02 04:44:07.673115
154	1-154	67	46	9193	86	4141	0	0	2012-08-02 04:43:55.286995	2012-08-02 04:44:07.673115
460	1-460	59	86	2714	28	7345	0	0	2012-08-02 04:44:07.673115	2012-08-02 04:44:07.673115
251	1-251	75	48	9999	15	8016	0	0	2012-08-02 04:43:57.14751	2012-08-02 04:44:07.861537
167	1-167	58	85	5695	37	7749	0	0	2012-08-02 04:43:55.498554	2012-08-02 04:44:08.425838
410	1-410	71	30	7468	9	8388	0	0	2012-08-02 04:44:04.25691	2012-08-02 04:44:08.448015
28	1-28	4	41	6470	40	2209	0	0	2012-08-02 04:43:53.754798	2012-08-02 04:44:08.626719
74	1-74	42	71	8071	76	7182	0	0	2012-08-02 04:43:54.281353	2012-08-02 04:44:08.626719
213	1-213	89	26	1645	34	9581	0	0	2012-08-02 04:43:56.244963	2012-08-02 04:44:08.626719
205	1-205	43	89	5379	29	9463	0	0	2012-08-02 04:43:56.055471	2012-08-02 04:44:10.064068
353	1-353	47	56	4853	67	5863	0	0	2012-08-02 04:44:00.965579	2012-08-02 04:44:10.086239
471	1-471	17	45	7614	71	5362	0	0	2012-08-02 04:44:08.626719	2012-08-02 04:44:10.086239
244	1-244	32	7	331	81	4714	0	0	2012-08-02 04:43:56.96952	2012-08-02 04:44:10.639252
354	1-354	30	29	7074	85	5578	0	0	2012-08-02 04:44:01.031584	2012-08-02 04:44:10.827218
490	1-490	63	46	355	50	3789	0	0	2012-08-02 04:44:10.860741	2012-08-02 04:44:10.860741
224	1-224	41	89	435	72	8238	0	0	2012-08-02 04:43:56.456268	2012-08-02 04:44:11.02645
80	1-80	54	3	3886	78	7911	0	0	2012-08-02 04:43:54.371088	2012-08-02 04:44:11.02645
495	1-495	6	94	428	9	2457	0	0	2012-08-02 04:44:11.413861	2012-08-02 04:44:11.413861
172	1-172	1	20	6044	15	2592	0	0	2012-08-02 04:43:55.554258	2012-08-02 04:44:11.590666
314	1-314	37	51	921	59	5960	0	0	2012-08-02 04:43:59.412629	2012-08-02 04:44:12.619404
418	1-418	92	51	1455	38	8784	0	0	2012-08-02 04:44:04.882974	2012-08-02 04:44:12.619404
106	1-106	61	34	9582	40	9372	0	0	2012-08-02 04:43:54.67202	2012-08-02 04:44:13.048831
360	1-360	43	68	5128	46	9791	0	0	2012-08-02 04:44:01.286529	2012-08-02 04:44:13.048831
180	1-180	2	32	5314	9	2638	0	0	2012-08-02 04:43:55.665681	2012-08-02 04:44:13.048831
14	1-14	14	23	3850	22	5963	0	0	2012-08-02 04:43:53.587324	2012-08-02 04:44:13.448081
237	1-237	92	41	541	7	775	0	0	2012-08-02 04:43:56.814064	2012-08-02 04:44:13.448081
88	1-88	6	53	3467	82	5492	0	0	2012-08-02 04:43:54.459934	2012-08-02 04:44:13.647161
494	1-494	61	73	853	26	4702	0	0	2012-08-02 04:44:11.402943	2012-08-02 04:44:13.647161
392	1-392	99	90	3418	62	4759	0	0	2012-08-02 04:44:03.271495	2012-08-02 04:44:13.647161
511	1-511	76	33	76	53	1604	0	0	2012-08-02 04:44:13.890634	2012-08-02 04:44:13.890634
254	1-254	20	88	2480	1	3330	0	0	2012-08-02 04:43:57.247175	2012-08-02 04:44:13.968029
58	1-58	45	52	3286	60	4032	0	0	2012-08-02 04:43:54.089723	2012-08-02 04:44:14.88528
90	1-90	56	81	6826	67	1244	0	0	2012-08-02 04:43:54.482126	2012-08-02 04:44:15.692403
248	1-248	29	50	2403	45	442	0	0	2012-08-02 04:43:57.080587	2012-08-02 04:44:17.504155
387	1-387	85	4	4404	22	4736	0	0	2012-08-02 04:44:02.884051	2012-08-02 04:44:18.664455
492	1-492	23	37	2649	51	5727	0	0	2012-08-02 04:44:11.248034	2012-08-02 04:44:18.664455
496	1-496	42	3	2601	20	4657	0	0	2012-08-02 04:44:11.590666	2012-08-02 04:44:18.664455
95	1-95	19	37	5718	60	6666	0	0	2012-08-02 04:43:54.537872	2012-08-02 04:44:18.664455
1	1-1	1	2	4206	1	7580	0	0	2012-08-02 04:43:53.386942	2012-08-02 04:44:18.664455
17	1-17	17	27	5406	20	9231	0	0	2012-08-02 04:43:53.631876	2012-08-02 04:44:19.294927
324	1-324	29	35	1552	84	9462	0	0	2012-08-02 04:43:59.767811	2012-08-02 04:44:19.705894
147	1-147	77	33	3396	97	4538	0	0	2012-08-02 04:43:55.184511	2012-08-02 04:44:20.015425
447	1-447	75	74	3540	32	5118	0	0	2012-08-02 04:44:06.566921	2012-08-02 04:44:20.015425
433	1-433	76	41	403	51	1117	0	0	2012-08-02 04:44:06.068027	2012-08-02 04:44:20.015425
117	1-117	21	92	2037	85	4833	0	0	2012-08-02 04:43:54.794615	2012-08-02 04:44:20.015425
455	1-455	83	49	958	87	5967	0	0	2012-08-02 04:44:07.253119	2012-08-02 04:44:20.015425
505	1-505	10	84	3197	91	2130	0	0	2012-08-02 04:44:12.420041	2012-08-02 04:44:22.093192
164	1-164	82	45	2869	13	6286	0	0	2012-08-02 04:43:55.453965	2012-08-02 04:44:22.093192
223	1-223	57	79	4809	91	1340	0	0	2012-08-02 04:43:56.444929	2012-08-02 04:44:22.093192
485	1-485	10	57	1245	5	887	0	0	2012-08-02 04:44:10.385274	2012-08-02 04:44:22.093192
507	1-507	44	46	2983	34	9241	0	0	2012-08-02 04:44:13.048831	2012-08-02 04:44:22.56875
548	1-548	37	90	373	46	1700	0	0	2012-08-02 04:44:22.56875	2012-08-02 04:44:22.56875
5	1-5	5	9	8989	4	7299	0	0	2012-08-02 04:43:53.475826	2012-08-02 04:44:23.850508
381	1-381	30	28	7262	80	3963	0	0	2012-08-02 04:44:02.474265	2012-08-02 04:44:23.850508
310	1-310	31	28	15	25	1978	0	0	2012-08-02 04:43:59.290568	2012-08-02 04:44:23.850508
529	1-529	75	83	3839	77	2565	0	0	2012-08-02 04:44:18.38816	2012-08-02 04:44:23.850508
317	1-317	6	28	1246	83	9111	0	0	2012-08-02 04:43:59.578763	2012-08-02 04:44:23.850508
553	1-553	14	4	379	28	9580	0	0	2012-08-02 04:44:23.850508	2012-08-02 04:44:23.850508
342	1-342	26	80	5127	2	8637	2693	0	2012-08-02 04:44:00.455436	2012-08-02 04:44:23.850508
232	1-232	30	39	800	41	8130	0	0	2012-08-02 04:43:56.691428	2012-08-02 04:44:25.331686
238	1-238	68	13	5628	52	6724	0	0	2012-08-02 04:43:56.835865	2012-08-02 04:44:25.331686
525	1-525	71	5	1024	72	8889	0	0	2012-08-02 04:44:17.172739	2012-08-02 04:44:25.331686
426	1-426	80	79	2194	15	741	0	0	2012-08-02 04:44:05.923608	2012-08-02 04:44:25.331686
424	1-424	98	30	3319	66	2539	0	0	2012-08-02 04:44:05.890372	2012-08-02 04:44:25.331686
204	1-204	68	39	2850	29	2658	0	0	2012-08-02 04:43:56.044313	2012-08-02 04:44:25.331686
274	1-274	59	68	9927	51	8805	0	0	2012-08-02 04:43:58.078875	2012-08-02 04:44:25.862386
557	1-557	21	84	2059	39	5482	0	0	2012-08-02 04:44:25.331686	2012-08-02 04:44:26.24909
542	1-542	89	35	643	68	7016	0	0	2012-08-02 04:44:20.956566	2012-08-02 04:44:27.33196
70	1-70	29	11	4955	16	8777	0	0	2012-08-02 04:43:54.23647	2012-08-02 04:44:27.33196
547	1-547	60	52	3697	11	4697	0	3	2012-08-02 04:44:22.093192	2012-08-02 04:44:27.33196
272	1-272	2	68	8993	19	8633	0	0	2012-08-02 04:43:57.912481	2012-08-02 04:44:27.33196
564	1-564	1	19	808	35	5115	0	0	2012-08-02 04:44:27.33196	2012-08-02 04:44:27.33196
383	1-383	94	82	4507	91	4463	0	0	2012-08-02 04:44:02.640447	2012-08-02 04:44:28.536941
568	1-568	62	70	3530	92	2837	0	0	2012-08-02 04:44:28.713823	2012-08-02 04:44:28.713823
462	1-462	24	8	1262	63	4798	0	0	2012-08-02 04:44:07.97212	2012-08-02 04:44:29.266055
297	1-297	68	72	9359	30	2317	0	0	2012-08-02 04:43:58.822645	2012-08-02 04:44:29.266055
509	1-509	14	95	6434	82	3864	0	0	2012-08-02 04:44:13.448081	2012-08-02 04:44:29.266055
567	1-567	64	74	1525	82	6935	0	0	2012-08-02 04:44:28.536941	2012-08-02 04:44:29.520494
76	1-76	1	75	1607	24	3306	0	0	2012-08-02 04:43:54.30382	2012-08-02 04:44:29.520494
572	1-572	91	93	951	75	9854	0	0	2012-08-02 04:44:29.520494	2012-08-02 04:44:29.520494
351	1-351	47	29	3441	5	3316	0	0	2012-08-02 04:44:00.821366	2012-08-02 04:44:30.49312
169	1-169	70	99	1081	56	9009	0	0	2012-08-02 04:43:55.52068	2012-08-02 04:44:30.49312
161	1-161	76	79	1218	31	8591	0	0	2012-08-02 04:43:55.420523	2012-08-02 04:44:31.963077
349	1-349	98	15	1896	12	8031	0	0	2012-08-02 04:44:00.721529	2012-08-02 04:44:31.963077
584	1-584	66	54	6191	15	4497	0	0	2012-08-02 04:44:31.963077	2012-08-02 04:44:31.963077
119	1-119	66	4	7063	44	1563	0	0	2012-08-02 04:43:54.816993	2012-08-02 04:44:32.703615
522	1-522	90	42	4524	70	7531	0	0	2012-08-02 04:44:15.692403	2012-08-02 04:44:32.814162
478	1-478	88	26	909	60	2185	0	0	2012-08-02 04:44:09.677398	2012-08-02 04:44:34.338852
571	1-571	13	24	1722	8	8723	0	0	2012-08-02 04:44:29.266055	2012-08-02 04:44:34.338852
458	1-458	24	22	1979	7	8645	0	0	2012-08-02 04:44:07.518414	2012-08-02 04:44:34.338852
469	1-469	94	7	9542	60	3756	0	0	2012-08-02 04:44:08.448015	2012-08-02 04:44:34.338852
593	1-593	80	47	2509	2	4641	0	0	2012-08-02 04:44:34.893489	2012-08-02 04:44:34.893489
482	1-482	37	1	4406	37	9748	0	0	2012-08-02 04:44:10.064068	2012-08-02 04:44:35.601153
594	1-594	55	64	234	1	6080	0	0	2012-08-02 04:44:35.601153	2012-08-02 04:44:35.601153
394	1-394	94	2	1033	93	2013	0	0	2012-08-02 04:44:03.415316	2012-08-02 04:44:35.821689
596	1-596	53	93	857	59	4145	0	0	2012-08-02 04:44:35.821689	2012-08-02 04:44:35.821689
532	1-532	38	59	8806	2	7375	2789	0	2012-08-02 04:44:18.664455	2012-08-02 04:44:35.821689
599	1-599	15	41	250	69	5872	0	0	2012-08-02 04:44:36.694784	2012-08-02 04:44:36.694784
562	1-562	12	59	1152	46	4329	0	0	2012-08-02 04:44:26.989884	2012-08-02 04:44:37.125811
408	1-408	43	80	4213	26	6081	0	0	2012-08-02 04:44:04.101977	2012-08-02 04:44:37.125811
437	1-437	41	29	5260	43	7388	0	0	2012-08-02 04:44:06.212531	2012-08-02 04:44:37.87784
452	1-452	84	78	8991	30	3089	0	0	2012-08-02 04:44:06.998479	2012-08-02 04:44:38.518643
413	1-413	71	53	1229	61	1385	0	0	2012-08-02 04:44:04.644862	2012-08-02 04:44:39.42509
13	1-13	13	21	9639	3	5403	0	0	2012-08-02 04:43:53.576163	2012-08-02 04:44:39.513477
476	1-476	57	43	7495	65	6735	0	0	2012-08-02 04:44:09.290465	2012-08-02 04:44:39.513477
108	1-108	19	17	3159	6	8044	0	0	2012-08-02 04:43:54.694269	2012-08-02 04:44:39.944705
506	1-506	80	59	9792	24	5193	0	0	2012-08-02 04:44:12.619404	2012-08-02 04:44:40.34241
174	1-174	53	58	4412	24	3432	0	0	2012-08-02 04:43:55.587916	2012-08-02 04:44:40.784449
153	1-153	79	76	1182	60	4606	0	0	2012-08-02 04:43:55.273851	2012-08-02 04:44:40.784449
339	1-339	37	66	3802	86	4898	0	0	2012-08-02 04:44:00.377826	2012-08-02 04:44:40.784449
361	1-361	96	76	4092	61	5320	0	0	2012-08-02 04:44:01.364339	2012-08-02 04:44:40.784449
32	1-32	30	46	5968	45	2507	0	0	2012-08-02 04:43:53.799464	2012-08-02 04:44:40.784449
566	1-566	88	12	6988	7	3242	0	0	2012-08-02 04:44:27.874054	2012-08-02 04:44:42.321069
466	1-466	49	97	3167	18	1745	0	0	2012-08-02 04:44:08.104985	2012-08-02 04:44:42.321069
632	1-632	70	42	1802	12	6218	0	0	2012-08-02 04:44:43.337773	2012-08-02 04:44:43.337773
191	1-191	46	58	4560	74	6869	0	0	2012-08-02 04:43:55.810528	2012-08-02 04:44:43.890326
378	1-378	27	57	5333	38	2125	0	0	2012-08-02 04:44:02.285657	2012-08-02 04:44:44.255143
487	1-487	8	34	5172	32	2830	0	0	2012-08-02 04:44:10.639252	2012-08-02 04:44:44.918197
480	1-480	85	13	2621	79	6411	0	0	2012-08-02 04:44:09.887118	2012-08-02 04:44:44.918197
82	1-82	4	51	8599	69	657	0	0	2012-08-02 04:43:54.392874	2012-08-02 04:44:44.918197
602	1-602	58	20	4913	58	8454	0	0	2012-08-02 04:44:37.125811	2012-08-02 04:44:45.669831
371	1-371	2	95	9368	20	4708	0	0	2012-08-02 04:44:01.787236	2012-08-02 04:44:46.067443
301	1-301	36	22	4798	46	8666	0	0	2012-08-02 04:43:58.94642	2012-08-02 04:44:46.067443
527	1-527	4	20	1679	23	8295	0	0	2012-08-02 04:44:17.504155	2012-08-02 04:44:46.355044
526	1-526	64	46	6841	43	8367	0	0	2012-08-02 04:44:17.382529	2012-08-02 04:44:46.355044
226	1-226	84	98	2017	20	8133	0	0	2012-08-02 04:43:56.522812	2012-08-02 04:44:46.355044
513	1-513	51	48	1720	54	9505	0	0	2012-08-02 04:44:14.01241	2012-08-02 04:44:46.963184
636	1-636	76	32	261	13	7377	0	0	2012-08-02 04:44:44.918197	2012-08-02 04:44:46.963184
425	1-425	16	74	3837	30	3156	0	0	2012-08-02 04:44:05.912435	2012-08-02 04:44:47.173171
395	1-395	9	40	9396	45	8736	0	0	2012-08-02 04:44:03.42941	2012-08-02 04:44:47.173171
582	1-582	91	12	2343	42	8899	0	0	2012-08-02 04:44:31.256024	2012-08-02 04:44:48.035275
577	1-577	44	79	9958	92	7227	0	0	2012-08-02 04:44:30.49312	2012-08-02 04:44:49.647927
586	1-586	3	34	7339	7	7224	0	0	2012-08-02 04:44:32.814162	2012-08-02 04:44:49.647927
510	1-510	55	5	4239	49	6221	1843	0	2012-08-02 04:44:13.647161	2012-08-02 04:44:49.647927
422	1-422	92	31	9383	54	5074	0	0	2012-08-02 04:44:05.480499	2012-08-02 04:44:50.16741
86	1-86	44	53	6118	55	4112	0	0	2012-08-02 04:43:54.437627	2012-08-02 04:44:50.357223
612	1-612	95	16	1369	22	718	0	0	2012-08-02 04:44:39.015954	2012-08-02 04:44:50.456909
472	1-472	14	68	5119	94	3271	0	0	2012-08-02 04:44:08.814362	2012-08-02 04:44:50.456909
644	1-644	74	16	4134	98	3423	0	0	2012-08-02 04:44:46.355044	2012-08-02 04:44:50.456909
411	1-411	6	67	2623	34	7644	0	0	2012-08-02 04:44:04.356452	2012-08-02 04:44:50.456909
198	1-198	32	16	6258	67	3632	316	0	2012-08-02 04:43:55.933626	2012-08-02 04:44:50.456909
660	1-660	42	57	4818	17	3283	0	0	2012-08-02 04:44:50.899685	2012-08-02 04:44:50.899685
585	1-585	62	73	167	4	7714	0	0	2012-08-02 04:44:32.703615	2012-08-02 04:44:52.037977
635	1-635	5	54	2268	57	7961	0	0	2012-08-02 04:44:44.255143	2012-08-02 04:44:52.347639
609	1-609	62	4	6572	74	7706	0	0	2012-08-02 04:44:38.165604	2012-08-02 04:44:52.347639
614	1-614	11	3	969	46	4973	0	0	2012-08-02 04:44:39.513477	2012-08-02 04:44:53.165327
84	1-84	55	35	556	52	8300	0	0	2012-08-02 04:43:54.415108	2012-08-02 04:44:53.839678
128	1-128	50	77	3815	88	6971	0	0	2012-08-02 04:43:54.917182	2012-08-02 04:44:54.128286
281	1-281	71	15	5617	86	9228	0	0	2012-08-02 04:43:58.278954	2012-08-02 04:44:54.128286
145	1-145	46	92	1629	24	439	0	0	2012-08-02 04:43:55.162485	2012-08-02 04:44:54.128286
373	1-373	23	87	2504	18	5517	0	0	2012-08-02 04:44:01.942108	2012-08-02 04:44:54.249124
633	1-633	92	55	1740	66	3118	0	0	2012-08-02 04:44:43.890326	2012-08-02 04:44:55.110977
67	1-67	48	55	2073	51	3980	0	0	2012-08-02 04:43:54.203214	2012-08-02 04:44:55.110977
151	1-151	4	55	7125	64	3440	0	0	2012-08-02 04:43:55.229524	2012-08-02 04:44:55.110977
287	1-287	92	4	6450	40	4903	0	0	2012-08-02 04:43:58.589497	2012-08-02 04:44:55.82976
201	1-201	13	40	7473	12	7787	3355	0	2012-08-02 04:43:55.988967	2012-08-02 04:44:55.82976
678	1-678	11	70	3525	69	9470	0	0	2012-08-02 04:44:54.128286	2012-08-02 04:44:57.144742
563	1-563	27	88	5399	70	8880	3199	0	2012-08-02 04:44:27.29883	2012-08-02 04:44:57.144742
256	1-256	81	100	4712	27	4361	0	0	2012-08-02 04:43:57.291488	2012-08-02 04:44:59.343953
372	1-372	38	100	2493	69	3573	0	0	2012-08-02 04:44:01.897832	2012-08-02 04:44:59.343953
643	1-643	12	50	3165	22	8917	0	0	2012-08-02 04:44:46.067443	2012-08-02 04:45:00.658781
158	1-158	61	66	6047	98	4418	0	0	2012-08-02 04:43:55.364611	2012-08-02 04:45:01.951464
120	1-120	67	50	7620	36	2324	0	0	2012-08-02 04:43:54.828083	2012-08-02 04:45:01.951464
283	1-283	10	65	6083	20	5172	0	0	2012-08-02 04:43:58.367393	2012-08-02 04:45:02.735946
325	1-325	11	44	4687	52	7658	2998	0	2012-08-02 04:43:59.778874	2012-08-02 04:45:02.735946
608	1-608	74	24	1101	29	6243	0	0	2012-08-02 04:44:37.87784	2012-08-02 04:45:02.846581
345	1-345	50	71	5636	29	6270	0	0	2012-08-02 04:44:00.555354	2012-08-02 04:45:04.095419
604	1-604	24	28	962	19	8958	0	0	2012-08-02 04:44:37.435886	2012-08-02 04:45:04.095419
457	1-457	85	51	8921	26	1284	0	0	2012-08-02 04:44:07.385725	2012-08-02 04:45:04.871264
637	1-637	9	88	188	84	5620	0	0	2012-08-02 04:44:45.59197	2012-08-02 04:45:05.635534
519	1-519	88	18	6546	34	7008	0	0	2012-08-02 04:44:15.106573	2012-08-02 04:45:05.635534
479	1-479	93	34	5769	90	1954	0	0	2012-08-02 04:44:09.776487	2012-08-02 04:45:05.635534
148	1-148	39	16	4558	87	3230	0	0	2012-08-02 04:43:55.195832	2012-08-02 04:45:05.635534
645	1-645	10	74	5343	54	4437	0	0	2012-08-02 04:44:46.752987	2012-08-02 04:45:05.635534
346	1-346	7	87	7292	90	3195	0	3	2012-08-02 04:44:00.566751	2012-08-02 04:45:05.635534
341	1-341	5	18	7674	10	7451	0	0	2012-08-02 04:44:00.43317	2012-08-02 04:45:06.541699
569	1-569	23	24	6928	88	7406	0	0	2012-08-02 04:44:28.890319	2012-08-02 04:45:06.541699
735	1-735	30	41	2426	73	3213	0	0	2012-08-02 04:45:08.276604	2012-08-02 04:45:08.276604
369	1-369	59	10	9042	48	8245	0	0	2012-08-02 04:44:01.643257	2012-08-02 04:45:08.43164
736	1-736	28	48	3752	10	9480	0	0	2012-08-02 04:45:08.43164	2012-08-02 04:45:08.43164
704	1-704	63	43	1533	87	5403	0	0	2012-08-02 04:44:59.763752	2012-08-02 04:45:11.40314
726	1-726	23	90	5108	75	3584	0	0	2012-08-02 04:45:05.635534	2012-08-02 04:45:11.40314
267	1-267	30	43	6132	90	1322	0	0	2012-08-02 04:43:57.60267	2012-08-02 04:45:11.40314
540	1-540	26	32	7438	92	4772	0	0	2012-08-02 04:44:20.015425	2012-08-02 04:45:11.911952
459	1-459	97	100	1301	57	2261	0	0	2012-08-02 04:44:07.595802	2012-08-02 04:45:12.530782
693	1-693	90	32	5978	24	3934	0	0	2012-08-02 04:44:57.144742	2012-08-02 04:45:12.530782
670	1-670	80	64	1323	41	9595	0	0	2012-08-02 04:44:52.281269	2012-08-02 04:45:12.850928
747	1-747	82	89	416	3	1405	0	0	2012-08-02 04:45:13.027823	2012-08-02 04:45:13.027823
291	1-291	93	88	8230	89	8641	330	0	2012-08-02 04:43:58.689329	2012-08-02 04:45:13.027823
241	1-241	8	84	5541	57	4565	0	0	2012-08-02 04:43:56.914306	2012-08-02 04:45:14.563837
454	1-454	90	3	8546	54	9900	0	0	2012-08-02 04:44:07.208737	2012-08-02 04:45:14.972351
9	1-9	9	17	8677	16	4936	0	0	2012-08-02 04:43:53.531666	2012-08-02 04:45:16.088607
375	1-375	84	43	5815	17	8811	602	0	2012-08-02 04:44:02.185904	2012-08-02 04:45:16.088607
657	1-657	93	93	2658	16	9247	0	0	2012-08-02 04:44:50.456909	2012-08-02 04:45:17.888937
671	1-671	97	31	930	4	8135	0	0	2012-08-02 04:44:52.347639	2012-08-02 04:45:17.888937
759	1-759	19	4	4469	54	4014	0	0	2012-08-02 04:45:17.888937	2012-08-02 04:45:17.888937
521	1-521	88	92	6778	75	5580	0	0	2012-08-02 04:44:15.681057	2012-08-02 04:45:21.258702
689	1-689	99	49	1005	41	3460	0	0	2012-08-02 04:44:56.459362	2012-08-02 04:45:24.473522
744	1-744	47	93	3121	32	4964	0	0	2012-08-02 04:45:11.911952	2012-08-02 04:45:24.473522
534	1-534	26	20	7884	44	4743	0	0	2012-08-02 04:44:19.294927	2012-08-02 04:45:24.473522
616	1-616	43	6	3789	69	1167	0	0	2012-08-02 04:44:39.944705	2012-08-02 04:45:24.473522
37	1-37	21	53	7157	52	6211	0	0	2012-08-02 04:43:53.855198	2012-08-02 04:45:24.473522
783	1-783	66	50	188	82	4075	0	2	2012-08-02 04:45:26.630582	2012-08-02 04:45:26.630582
400	1-400	24	26	2087	42	3528	0	0	2012-08-02 04:44:03.758953	2012-08-02 04:45:27.370688
628	1-628	51	11	652	90	9295	0	0	2012-08-02 04:44:42.796262	2012-08-02 04:45:27.370688
536	1-536	13	89	6017	35	9513	0	0	2012-08-02 04:44:19.705894	2012-08-02 04:45:27.370688
672	1-672	3	42	6544	93	9703	0	0	2012-08-02 04:44:52.944443	2012-08-02 04:45:27.370688
590	1-590	37	64	632	7	2147	0	0	2012-08-02 04:44:34.173116	2012-08-02 04:45:27.370688
556	1-556	100	37	8290	64	6509	0	0	2012-08-02 04:44:25.320706	2012-08-02 04:45:27.370688
789	1-789	4	52	1850	60	2623	0	6	2012-08-02 04:45:28.210993	2012-08-02 04:45:28.210993
793	1-793	42	88	16	46	2443	0	0	2012-08-02 04:45:29.039958	2012-08-02 04:45:29.039958
786	1-786	15	93	5752	37	9472	0	0	2012-08-02 04:45:27.370688	2012-08-02 04:45:29.40452
663	1-663	83	72	3378	25	9999	0	0	2012-08-02 04:44:51.40797	2012-08-02 04:45:29.40452
794	1-794	53	25	9798	29	5677	0	0	2012-08-02 04:45:29.40452	2012-08-02 04:45:29.40452
619	1-619	17	70	1058	76	8965	0	0	2012-08-02 04:44:40.784449	2012-08-02 04:45:29.912978
207	1-207	50	49	9238	73	6612	0	0	2012-08-02 04:43:56.122278	2012-08-02 04:45:30.785605
776	1-776	41	32	2030	49	9716	0	0	2012-08-02 04:45:24.473522	2012-08-02 04:45:30.785605
742	1-742	69	94	109	43	9127	0	0	2012-08-02 04:45:11.40314	2012-08-02 04:45:30.940765
802	1-802	32	78	67	21	2267	0	0	2012-08-02 04:45:33.945514	2012-08-02 04:45:33.945514
227	1-227	4	29	9356	80	2593	0	0	2012-08-02 04:43:56.534008	2012-08-02 04:45:38.419212
812	1-812	10	33	1592	29	5930	0	0	2012-08-02 04:45:38.419212	2012-08-02 04:45:38.419212
814	1-814	98	77	3344	56	4137	0	0	2012-08-02 04:45:38.949749	2012-08-02 04:45:38.949749
817	1-817	11	14	4219	77	1194	0	0	2012-08-02 04:45:39.955096	2012-08-02 04:45:39.955096
741	1-741	70	34	4609	53	208	0	2	2012-08-02 04:45:10.895289	2012-08-02 04:45:40.452613
696	1-696	13	52	6777	1	5959	0	0	2012-08-02 04:44:57.531727	2012-08-02 04:45:40.452613
818	1-818	10	97	5455	13	8705	0	0	2012-08-02 04:45:40.452613	2012-08-02 04:45:40.452613
684	1-684	46	48	4937	55	5367	0	4	2012-08-02 04:44:55.110977	2012-08-02 04:45:42.585046
461	1-461	72	84	1820	48	9741	0	0	2012-08-02 04:44:07.861537	2012-08-02 04:45:42.585046
727	1-727	97	40	9717	24	8801	0	0	2012-08-02 04:45:06.541699	2012-08-02 04:45:42.585046
133	1-133	40	92	6691	55	7420	0	0	2012-08-02 04:43:54.973207	2012-08-02 04:45:42.585046
769	1-769	50	84	2334	55	3476	0	0	2012-08-02 04:45:21.524089	2012-08-02 04:45:42.585046
302	1-302	30	96	1870	57	2436	0	0	2012-08-02 04:43:59.023967	2012-08-02 04:45:43.734002
501	1-501	94	83	2696	67	2795	0	0	2012-08-02 04:44:12.198687	2012-08-02 04:45:43.734002
680	1-680	32	26	8233	52	8239	0	0	2012-08-02 04:44:54.525607	2012-08-02 04:45:43.734002
27	1-27	11	39	228	33	9370	0	0	2012-08-02 04:43:53.743617	2012-08-02 04:45:44.828001
768	1-768	86	98	2946	92	8990	0	0	2012-08-02 04:45:21.258702	2012-08-02 04:45:45.899983
797	1-797	2	43	3100	99	4051	0	0	2012-08-02 04:45:30.940765	2012-08-02 04:45:45.899983
257	1-257	37	99	5814	29	3971	0	0	2012-08-02 04:43:57.302798	2012-08-02 04:45:45.899983
822	1-822	49	55	4025	28	6425	0	0	2012-08-02 04:45:42.585046	2012-08-02 04:45:46.916236
230	1-230	87	80	4587	54	2388	0	0	2012-08-02 04:43:56.600607	2012-08-02 04:45:46.916236
93	1-93	58	62	7567	67	4446	245	0	2012-08-02 04:43:54.515684	2012-08-02 04:45:46.916236
475	1-475	96	76	3915	66	4688	0	0	2012-08-02 04:44:09.268087	2012-08-02 04:45:48.010156
277	1-277	76	65	6474	79	6969	0	0	2012-08-02 04:43:58.145536	2012-08-02 04:45:48.253392
648	1-648	38	10	1299	4	8126	0	0	2012-08-02 04:44:47.173171	2012-08-02 04:45:48.253392
659	1-659	94	13	8847	26	3435	0	0	2012-08-02 04:44:50.888248	2012-08-02 04:45:48.253392
530	1-530	84	13	4255	3	3681	0	0	2012-08-02 04:44:18.45446	2012-08-02 04:45:48.253392
835	1-835	78	4	2619	13	6494	0	0	2012-08-02 04:45:48.253392	2012-08-02 04:45:48.253392
703	1-703	37	27	2159	100	5249	0	0	2012-08-02 04:44:59.343953	2012-08-02 04:45:49.195487
47	1-47	37	61	2800	33	8737	0	0	2012-08-02 04:43:53.966903	2012-08-02 04:45:49.195487
839	1-839	47	100	2118	23	9800	0	0	2012-08-02 04:45:49.195487	2012-08-02 04:45:49.195487
841	1-841	39	55	220	74	5484	0	0	2012-08-02 04:45:50.697649	2012-08-02 04:45:50.697649
629	1-629	90	88	1211	18	7517	0	0	2012-08-02 04:44:42.873472	2012-08-02 04:45:51.272026
208	1-208	19	5	4117	13	5616	0	0	2012-08-02 04:43:56.144651	2012-08-02 04:45:51.272026
550	1-550	50	71	4972	60	8218	0	0	2012-08-02 04:44:23.233913	2012-08-02 04:45:51.272026
745	1-745	24	57	7695	26	9855	0	0	2012-08-02 04:45:12.530782	2012-08-02 04:45:53.073678
698	1-698	95	6	1192	16	357	0	0	2012-08-02 04:44:58.382656	2012-08-02 04:45:53.073678
442	1-442	49	94	5365	97	9626	0	0	2012-08-02 04:44:06.323195	2012-08-02 04:45:53.073678
806	1-806	43	27	7445	15	1351	0	0	2012-08-02 04:45:35.934024	2012-08-02 04:45:53.073678
264	1-264	70	6	6303	42	7662	0	0	2012-08-02 04:43:57.480325	2012-08-02 04:45:53.073678
337	1-337	26	3	8539	84	2074	587	0	2012-08-02 04:44:00.289158	2012-08-02 04:45:53.073678
468	1-468	18	37	9691	89	5387	0	0	2012-08-02 04:44:08.425838	2012-08-02 04:45:53.979782
639	1-639	68	41	8949	53	7320	0	0	2012-08-02 04:44:45.669831	2012-08-02 04:45:53.979782
677	1-677	34	58	1628	50	2937	0	0	2012-08-02 04:44:54.071592	2012-08-02 04:45:53.979782
718	1-718	28	97	3278	33	2686	0	0	2012-08-02 04:45:03.752729	2012-08-02 04:45:53.979782
343	1-343	97	37	8385	53	2785	0	0	2012-08-02 04:44:00.510896	2012-08-02 04:45:53.979782
851	1-851	4	89	5494	85	8123	0	0	2012-08-02 04:45:53.979782	2012-08-02 04:45:53.979782
450	1-450	29	53	8737	94	9353	5506	0	2012-08-02 04:44:06.909813	2012-08-02 04:45:53.979782
71	1-71	51	65	6701	21	4987	0	0	2012-08-02 04:43:54.24792	2012-08-02 04:45:54.653894
834	1-834	70	22	1747	46	974	0	9	2012-08-02 04:45:48.010156	2012-08-02 04:45:55.043781
558	1-558	41	22	3689	95	9258	0	0	2012-08-02 04:44:25.862386	2012-08-02 04:45:55.043781
697	1-697	21	62	2000	40	8375	0	0	2012-08-02 04:44:57.863133	2012-08-02 04:45:55.043781
751	1-751	68	26	5669	98	4581	0	0	2012-08-02 04:45:14.563837	2012-08-02 04:45:59.824428
708	1-708	71	42	5894	14	7515	0	0	2012-08-02 04:45:00.272448	2012-08-02 04:46:00.321812
575	1-575	84	27	3321	10	320	0	0	2012-08-02 04:44:30.21949	2012-08-02 04:46:00.321812
869	1-869	52	27	93	60	6539	0	0	2012-08-02 04:46:00.830273	2012-08-02 04:46:00.830273
545	1-545	29	90	6746	10	1541	0	0	2012-08-02 04:44:21.562854	2012-08-02 04:46:04.310355
508	1-508	82	69	9033	73	6473	1769	0	2012-08-02 04:44:13.293882	2012-08-02 04:46:04.310355
861	1-861	11	27	2650	42	8115	0	0	2012-08-02 04:45:58.035093	2012-08-02 04:46:05.173031
843	1-843	80	54	437	71	7952	0	0	2012-08-02 04:45:51.272026	2012-08-02 04:46:06.885666
625	1-625	63	7	5202	41	2115	0	0	2012-08-02 04:44:41.624905	2012-08-02 04:46:06.885666
355	1-355	53	72	6088	68	3991	0	0	2012-08-02 04:44:01.087266	2012-08-02 04:46:07.339042
862	1-862	60	54	4160	39	7868	0	0	2012-08-02 04:45:58.697495	2012-08-02 04:46:07.339042
573	1-573	4	44	1158	17	1896	0	0	2012-08-02 04:44:29.841351	2012-08-02 04:46:07.339042
676	1-676	85	54	8548	96	7136	0	0	2012-08-02 04:44:53.839678	2012-08-02 04:46:07.339042
858	1-858	64	37	1639	93	9429	0	0	2012-08-02 04:45:57.548468	2012-08-02 04:46:07.339042
882	1-882	24	80	974	54	8236	0	6	2012-08-02 04:46:06.885666	2012-08-02 04:46:07.339042
650	1-650	18	78	1639	14	8410	0	0	2012-08-02 04:44:48.37744	2012-08-02 04:46:07.339042
885	1-885	72	59	5758	62	9322	0	0	2012-08-02 04:46:07.339042	2012-08-02 04:46:07.339042
823	1-823	20	52	2079	85	6368	0	0	2012-08-02 04:45:43.734002	2012-08-02 04:46:08.389417
709	1-709	73	22	3712	29	2681	0	0	2012-08-02 04:45:00.658781	2012-08-02 04:46:08.389417
229	1-229	52	86	8179	79	7577	0	0	2012-08-02 04:43:56.589523	2012-08-02 04:46:08.389417
627	1-627	39	95	4587	68	9978	0	0	2012-08-02 04:44:42.321069	2012-08-02 04:46:08.389417
467	1-467	37	40	6753	68	7340	0	0	2012-08-02 04:44:08.226896	2012-08-02 04:46:08.389417
716	1-716	83	22	4861	51	2947	0	0	2012-08-02 04:45:03.586909	2012-08-02 04:46:11.219821
811	1-811	18	88	2094	68	1507	0	0	2012-08-02 04:45:38.021854	2012-08-02 04:46:11.219821
896	1-896	90	27	420	6	2150	0	0	2012-08-02 04:46:14.069413	2012-08-02 04:46:14.069413
552	1-552	44	90	4373	50	9183	0	0	2012-08-02 04:44:23.453109	2012-08-02 04:46:14.555671
653	1-653	90	49	1374	79	9397	0	0	2012-08-02 04:44:49.647927	2012-08-02 04:46:14.555671
34	1-34	28	49	2510	48	8377	0	0	2012-08-02 04:43:53.821705	2012-08-02 04:46:14.555671
489	1-489	13	56	9948	39	1917	0	0	2012-08-02 04:44:10.827218	2012-08-02 04:46:14.555671
826	1-826	1	48	8575	56	4353	0	0	2012-08-02 04:45:44.51857	2012-08-02 04:46:14.555671
892	1-892	76	88	2845	42	6952	0	0	2012-08-02 04:46:11.83815	2012-08-02 04:46:14.555671
798	1-798	57	94	1592	100	6229	0	0	2012-08-02 04:45:31.857808	2012-08-02 04:46:14.555671
887	1-887	15	68	9581	63	8679	0	0	2012-08-02 04:46:08.389417	2012-08-02 04:46:14.555671
821	1-821	85	33	7133	89	9435	0	0	2012-08-02 04:45:41.546354	2012-08-02 04:46:14.555671
850	1-850	10	84	1555	57	9982	0	6	2012-08-02 04:45:53.073678	2012-08-02 04:46:23.227736
60	1-60	7	69	4266	32	6996	0	0	2012-08-02 04:43:54.111957	2012-08-02 04:46:23.227736
605	1-605	23	87	1946	4	9339	0	3	2012-08-02 04:44:37.502323	2012-08-02 04:46:23.227736
911	1-911	63	4	3667	71	6841	0	0	2012-08-02 04:46:23.227736	2012-08-02 04:46:23.227736
313	1-313	46	15	7039	44	5829	1110	0	2012-08-02 04:43:59.37945	2012-08-02 04:46:23.227736
897	1-897	18	6	5041	49	9892	288	0	2012-08-02 04:46:14.555671	2012-08-02 04:46:23.227736
757	1-757	5	7	5169	57	4626	0	0	2012-08-02 04:45:17.381013	2012-08-02 04:46:28.497479
864	1-864	8	5	1510	89	863	0	0	2012-08-02 04:45:59.515082	2012-08-02 04:46:28.497479
901	1-901	33	53	2981	5	5674	0	0	2012-08-02 04:46:17.052942	2012-08-02 04:46:28.497479
921	1-921	79	79	2171	7	8920	0	0	2012-08-02 04:46:28.497479	2012-08-02 04:46:34.507368
948	1-948	38	25	631	67	3778	0	0	2012-08-02 04:46:42.229238	2012-08-02 04:46:42.229238
488	1-488	23	58	3202	41	5079	0	0	2012-08-02 04:44:10.793996	2012-08-02 04:46:45.023656
876	1-876	15	41	230	90	4591	0	0	2012-08-02 04:46:04.310355	2012-08-02 04:46:45.023656
483	1-483	9	21	4719	99	2367	1979	0	2012-08-02 04:44:10.086239	2012-08-02 04:46:45.023656
402	1-402	42	7	8613	95	9406	0	0	2012-08-02 04:44:03.802838	2012-08-02 04:46:47.122735
618	1-618	69	17	4753	59	5578	0	0	2012-08-02 04:44:40.34241	2012-08-02 04:46:47.763437
917	1-917	37	90	3520	20	9007	0	0	2012-08-02 04:46:25.979001	2012-08-02 04:46:48.868228
925	1-925	12	58	801	81	5557	0	0	2012-08-02 04:46:30.895049	2012-08-02 04:46:48.868228
122	1-122	42	2	1021	50	3963	0	0	2012-08-02 04:43:54.850468	2012-08-02 04:46:48.868228
445	1-445	48	85	4799	21	9220	0	0	2012-08-02 04:44:06.511448	2012-08-02 04:46:48.868228
695	1-695	75	1	6301	26	6705	0	0	2012-08-02 04:44:57.509382	2012-08-02 04:46:48.868228
886	1-886	45	81	4557	53	997	0	0	2012-08-02 04:46:07.881199	2012-08-02 04:46:48.868228
771	1-771	29	85	2501	83	4593	0	0	2012-08-02 04:45:22.617648	2012-08-02 04:46:48.868228
966	1-966	60	45	994	44	6529	0	0	2012-08-02 04:46:53.013082	2012-08-02 04:46:53.013082
958	1-958	96	12	4336	85	7333	0	0	2012-08-02 04:46:48.868228	2012-08-02 04:46:55.112006
363	1-363	52	98	6385	84	8355	0	0	2012-08-02 04:44:01.485951	2012-08-02 04:46:57.045443
551	1-551	19	45	6490	24	320	0	0	2012-08-02 04:44:23.276103	2012-08-02 04:46:57.045443
634	1-634	88	24	5466	84	6127	0	0	2012-08-02 04:44:44.078048	2012-08-02 04:46:57.045443
972	1-972	36	5	2171	73	2563	0	0	2012-08-02 04:46:57.045443	2012-08-02 04:46:57.045443
473	1-473	83	20	1259	12	3840	0	3	2012-08-02 04:44:09.035928	2012-08-02 04:46:58.106277
271	1-271	43	10	1185	78	3352	0	0	2012-08-02 04:43:57.868371	2012-08-02 04:46:58.106277
658	1-658	14	6	5463	27	7105	0	0	2012-08-02 04:44:50.777777	2012-08-02 04:46:58.106277
304	1-304	79	44	6181	19	5659	0	0	2012-08-02 04:43:59.046518	2012-08-02 04:46:58.835141
477	1-477	64	97	4065	12	5986	0	0	2012-08-02 04:44:09.456092	2012-08-02 04:46:58.835141
722	1-722	94	22	3939	33	1130	0	0	2012-08-02 04:45:04.816282	2012-08-02 04:47:01.718771
651	1-651	57	40	4341	25	9369	0	0	2012-08-02 04:44:49.117929	2012-08-02 04:47:02.260036
982	1-982	54	25	31	58	6214	0	0	2012-08-02 04:47:02.260036	2012-08-02 04:47:02.260036
845	1-845	8	99	6304	6	6303	0	0	2012-08-02 04:45:52.156423	2012-08-02 04:47:02.746225
968	1-968	67	26	313	32	9707	0	0	2012-08-02 04:46:53.99626	2012-08-02 04:47:02.746225
365	1-365	89	75	6924	77	8495	833	0	2012-08-02 04:44:01.519444	2012-08-02 04:47:02.746225
871	1-871	4	62	4330	30	3771	0	0	2012-08-02 04:46:02.431904	2012-08-02 04:47:04.160538
919	1-919	98	51	1171	45	5970	0	0	2012-08-02 04:46:27.260421	2012-08-02 04:47:04.160538
904	1-904	31	44	681	90	7721	0	0	2012-08-02 04:46:18.721041	2012-08-02 04:47:04.657876
647	1-647	3	39	2988	32	4670	0	0	2012-08-02 04:44:46.963184	2012-08-02 04:47:05.149686
807	1-807	30	57	2902	27	7976	0	0	2012-08-02 04:45:36.408698	2012-08-02 04:47:05.149686
668	1-668	32	39	1567	6	8072	0	0	2012-08-02 04:44:52.037977	2012-08-02 04:47:05.149686
973	1-973	85	78	6761	13	7723	0	0	2012-08-02 04:46:58.106277	2012-08-02 04:47:05.149686
796	1-796	2	57	8412	39	9433	0	0	2012-08-02 04:45:30.785605	2012-08-02 04:47:05.149686
439	1-439	89	61	5998	45	9989	0	0	2012-08-02 04:44:06.256475	2012-08-02 04:47:06.117672
987	1-987	17	55	3430	57	8575	0	0	2012-08-02 04:47:05.149686	2012-08-02 04:47:06.117672
12	1-12	12	20	5513	12	1091	0	0	2012-08-02 04:43:53.564997	2012-08-02 04:47:06.117672
617	1-617	38	79	1893	34	8641	0	0	2012-08-02 04:44:40.209841	2012-08-02 04:47:06.604057
278	1-278	44	5	6886	70	1112	0	0	2012-08-02 04:43:58.234199	2012-08-02 04:47:11.630623
711	1-711	92	66	3785	2	1881	0	0	2012-08-02 04:45:01.874371	2012-08-02 04:47:11.630623
401	1-401	15	5	6581	19	1789	0	0	2012-08-02 04:44:03.781112	2012-08-02 04:47:11.630623
1000	1-1000	23	65	1937	5	7884	0	0	2012-08-02 04:47:11.630623	2012-08-02 04:47:11.630623
829	1-829	51	7	512	33	2121	0	0	2012-08-02 04:45:46.220223	2012-08-02 04:47:13.045032
883	1-883	49	41	3521	19	2512	0	0	2012-08-02 04:46:06.974366	2012-08-02 04:47:15.089031
1008	1-1008	41	19	2198	41	8981	0	0	2012-08-02 04:47:15.089031	2012-08-02 04:47:15.089031
656	1-656	23	7	956	51	8399	0	0	2012-08-02 04:44:50.357223	2012-08-02 04:47:18.448843
396	1-396	53	92	8118	29	3604	0	0	2012-08-02 04:44:03.603333	2012-08-02 04:47:18.448843
1018	1-1018	40	13	381	41	1617	0	0	2012-08-02 04:47:19.443255	2012-08-02 04:47:19.443255
990	1-990	33	44	1116	61	4552	0	0	2012-08-02 04:47:06.117672	2012-08-02 04:47:21.906704
923	1-923	94	95	480	100	6590	0	0	2012-08-02 04:46:29.547342	2012-08-02 04:47:22.393403
767	1-767	45	13	8292	99	8181	0	0	2012-08-02 04:45:21.104052	2012-08-02 04:47:24.160261
66	1-66	47	62	9031	44	7021	3013	0	2012-08-02 04:43:54.191985	2012-08-02 04:47:24.160261
967	1-967	63	4	4853	39	406	0	0	2012-08-02 04:46:53.510355	2012-08-02 04:47:24.657558
947	1-947	94	42	2075	54	3767	0	0	2012-08-02 04:46:41.102737	2012-08-02 04:47:24.657558
245	1-245	55	80	3767	12	369	0	0	2012-08-02 04:43:56.991608	2012-08-02 04:47:25.154905
975	1-975	1	6	8180	44	3992	1346	0	2012-08-02 04:46:58.835141	2012-08-02 04:47:25.154905
1027	1-1027	34	39	3164	42	7820	0	0	2012-08-02 04:47:24.657558	2012-08-02 04:47:26.105107
938	1-938	93	39	2045	66	3355	0	0	2012-08-02 04:46:35.424426	2012-08-02 04:47:26.105107
1030	1-1030	15	43	1613	39	9043	0	0	2012-08-02 04:47:26.105107	2012-08-02 04:47:26.105107
149	1-149	41	56	2096	40	3688	0	0	2012-08-02 04:43:55.206913	2012-08-02 04:47:27.519442
130	1-130	73	71	6521	69	3183	0	0	2012-08-02 04:43:54.939623	2012-08-02 04:47:27.519442
92	1-92	40	67	5191	40	8103	0	0	2012-08-02 04:43:54.504442	2012-08-02 04:47:27.519442
675	1-675	10	56	4738	77	7369	0	0	2012-08-02 04:44:53.828776	2012-08-02 04:47:27.519442
705	1-705	48	56	6377	67	6192	0	5	2012-08-02 04:45:00.161244	2012-08-02 04:47:27.519442
1033	1-1033	9	62	2545	56	9361	0	0	2012-08-02 04:47:27.519442	2012-08-02 04:47:27.519442
924	1-924	47	92	2642	98	7680	0	0	2012-08-02 04:46:30.751127	2012-08-02 04:47:28.017119
626	1-626	83	60	6587	1	8691	0	0	2012-08-02 04:44:42.310183	2012-08-02 04:47:28.503375
159	1-159	82	60	7835	32	1188	0	0	2012-08-02 04:43:55.376697	2012-08-02 04:47:28.503375
1024	1-1024	67	40	6629	15	7911	0	0	2012-08-02 04:47:22.393403	2012-08-02 04:47:28.503375
781	1-781	93	49	3302	48	9740	0	7	2012-08-02 04:45:26.310115	2012-08-02 04:47:29.464544
1038	1-1038	90	37	340	11	476	0	0	2012-08-02 04:47:31.032912	2012-08-02 04:47:31.032912
1040	1-1040	78	41	748	39	6894	0	0	2012-08-02 04:47:31.949857	2012-08-02 04:47:31.949857
994	1-994	24	74	1148	12	2175	0	0	2012-08-02 04:47:07.99623	2012-08-02 04:47:34.281021
995	1-995	46	58	4516	75	3398	0	0	2012-08-02 04:47:08.460121	2012-08-02 04:47:34.281021
386	1-386	93	99	7342	42	4795	0	0	2012-08-02 04:44:02.872907	2012-08-02 04:47:36.291462
983	1-983	95	81	9898	99	9464	0	0	2012-08-02 04:47:02.746225	2012-08-02 04:47:36.291462
959	1-959	72	79	5368	48	4758	0	0	2012-08-02 04:46:49.423076	2012-08-02 04:47:38.047949
1003	1-1003	30	57	2903	63	9476	0	0	2012-08-02 04:47:13.033764	2012-08-02 04:47:41.361649
583	1-583	53	28	7531	62	6430	444	0	2012-08-02 04:44:31.896719	2012-08-02 04:47:41.361649
1001	1-1001	78	9	1633	5	3813	0	0	2012-08-02 04:47:12.116873	2012-08-02 04:47:43.228679
1041	1-1041	65	31	322	21	176	0	0	2012-08-02 04:47:32.425286	2012-08-02 04:47:45.93507
951	1-951	87	74	4640	36	1380	0	0	2012-08-02 04:46:44.217651	2012-08-02 04:47:51.878081
1043	1-1043	12	47	7515	58	8327	0	0	2012-08-02 04:47:34.281021	2012-08-02 04:47:51.878081
804	1-804	71	15	3902	94	9013	2354	0	2012-08-02 04:45:34.675408	2012-08-02 04:47:51.878081
1070	1-1070	95	97	293	82	528	0	0	2012-08-02 04:47:52.872378	2012-08-02 04:47:52.872378
774	1-774	94	54	3385	55	4817	0	0	2012-08-02 04:45:23.501453	2012-08-02 04:47:54.253317
795	1-795	55	84	4309	5	6899	958	0	2012-08-02 04:45:29.912978	2012-08-02 04:47:54.253317
908	1-908	67	42	4554	30	6124	0	0	2012-08-02 04:46:21.085147	2012-08-02 04:47:56.054253
952	1-952	28	20	9310	85	1220	0	0	2012-08-02 04:46:45.023656	2012-08-02 04:47:57.25996
753	1-753	53	33	4523	3	5138	0	0	2012-08-02 04:45:14.972351	2012-08-02 04:47:58.298357
1015	1-1015	9	66	4594	95	7587	0	0	2012-08-02 04:47:18.448843	2012-08-02 04:47:59.347978
79	1-79	3	77	5355	15	2062	0	0	2012-08-02 04:43:54.359409	2012-08-02 04:48:00.298187
710	1-710	73	68	1948	17	3592	0	0	2012-08-02 04:45:00.869179	2012-08-02 04:48:00.298187
713	1-713	41	13	1096	65	4338	0	0	2012-08-02 04:45:02.735946	2012-08-02 04:48:00.298187
956	1-956	37	41	5984	92	7951	0	0	2012-08-02 04:46:47.122735	2012-08-02 04:48:00.298187
893	1-893	59	55	2944	90	2071	0	0	2012-08-02 04:46:12.324287	2012-08-02 04:48:02.364285
927	1-927	70	22	5148	17	2115	0	0	2012-08-02 04:46:31.414461	2012-08-02 04:48:02.364285
1012	1-1012	79	66	3370	79	8602	0	0	2012-08-02 04:47:17.342538	2012-08-02 04:48:03.546719
865	1-865	69	88	3322	22	8594	0	0	2012-08-02 04:45:59.824428	2012-08-02 04:48:04.441389
493	1-493	93	64	5363	99	9161	0	0	2012-08-02 04:44:11.269926	2012-08-02 04:48:06.299242
1090	1-1090	57	42	362	22	7064	0	0	2012-08-02 04:48:02.364285	2012-08-02 04:48:06.299242
836	1-836	72	74	8324	90	6028	0	0	2012-08-02 04:45:48.806292	2012-08-02 04:48:06.763128
964	1-964	95	19	454	97	5054	0	0	2012-08-02 04:46:52.129663	2012-08-02 04:48:08.155104
828	1-828	62	29	8640	19	8816	1835	0	2012-08-02 04:45:45.899983	2012-08-02 04:48:08.155104
622	1-622	42	69	7477	15	2170	0	0	2012-08-02 04:44:41.293445	2012-08-02 04:48:08.641428
957	1-957	73	5	97	68	8503	0	0	2012-08-02 04:46:47.763437	2012-08-02 04:48:08.641428
263	1-263	36	64	5036	68	2838	0	0	2012-08-02 04:43:57.436501	2012-08-02 04:48:08.641428
1097	1-1097	19	3	1725	64	9217	0	0	2012-08-02 04:48:06.299242	2012-08-02 04:48:11.657079
303	1-303	6	1	3888	36	3840	0	0	2012-08-02 04:43:59.035199	2012-08-02 04:48:11.657079
673	1-673	98	46	5900	1	2644	0	0	2012-08-02 04:44:53.165327	2012-08-02 04:48:11.657079
1107	1-1107	89	74	1027	54	8474	0	6	2012-08-02 04:48:11.657079	2012-08-02 04:48:11.657079
233	1-233	87	36	5124	94	8336	0	0	2012-08-02 04:43:56.702826	2012-08-02 04:48:12.596482
1112	1-1112	29	71	3354	73	9314	0	0	2012-08-02 04:48:14.794983	2012-08-02 04:48:14.794983
1035	1-1035	22	32	1308	57	2781	0	0	2012-08-02 04:47:28.503375	2012-08-02 04:48:18.40696
1053	1-1053	13	62	2222	57	5051	0	0	2012-08-02 04:47:41.361649	2012-08-02 04:48:18.40696
692	1-692	25	48	6479	55	5380	0	7	2012-08-02 04:44:56.614439	2012-08-02 04:48:19.35699
1120	1-1120	88	40	260	21	4808	0	0	2012-08-02 04:48:20.251683	2012-08-02 04:48:20.251683
712	1-712	69	87	9037	66	9614	1834	0	2012-08-02 04:45:01.951464	2012-08-02 04:48:20.251683
1124	1-1124	31	49	1534	13	6863	0	0	2012-08-02 04:48:22.405934	2012-08-02 04:48:22.405934
988	1-988	37	60	674	72	4388	0	0	2012-08-02 04:47:05.642764	2012-08-02 04:48:24.758874
1046	1-1046	11	27	1062	81	5979	0	0	2012-08-02 04:47:36.291462	2012-08-02 04:48:24.758874
1128	1-1128	52	72	8321	97	8599	0	0	2012-08-02 04:48:24.758874	2012-08-02 04:48:24.758874
1025	1-1025	22	25	6269	66	8289	0	0	2012-08-02 04:47:23.707719	2012-08-02 04:48:26.614867
1028	1-1028	54	25	427	80	7770	0	0	2012-08-02 04:47:25.154905	2012-08-02 04:48:28.782166
1071	1-1071	66	36	1518	58	8448	0	0	2012-08-02 04:47:53.369618	2012-08-02 04:48:28.782166
1135	1-1135	37	88	735	56	1449	0	0	2012-08-02 04:48:28.782166	2012-08-02 04:48:28.782166
1136	1-1136	54	59	4453	66	5850	0	0	2012-08-02 04:48:29.245722	2012-08-02 04:48:29.245722
134	1-134	70	85	8732	58	9997	0	0	2012-08-02 04:43:54.984507	2012-08-02 04:48:29.688199
1137	1-1137	33	25	857	18	8558	0	0	2012-08-02 04:48:29.688199	2012-08-02 04:48:29.688199
69	1-69	50	75	9830	74	9887	0	0	2012-08-02 04:43:54.225441	2012-08-02 04:48:30.610702
1139	1-1139	41	74	2751	75	8925	0	0	2012-08-02 04:48:30.610702	2012-08-02 04:48:30.610702
785	1-785	30	70	6763	71	9864	0	0	2012-08-02 04:45:27.304662	2012-08-02 04:48:33.135428
1020	1-1020	54	97	5784	10	9844	0	0	2012-08-02 04:47:20.945978	2012-08-02 04:48:34.516382
1101	1-1101	30	97	4783	1	1351	0	0	2012-08-02 04:48:08.155104	2012-08-02 04:48:34.516382
1150	1-1150	17	80	175	97	9315	0	0	2012-08-02 04:48:34.516382	2012-08-02 04:48:34.516382
484	1-484	61	14	7207	59	3291	0	0	2012-08-02 04:44:10.219277	2012-08-02 04:48:35.444308
746	1-746	27	38	500	37	8432	0	6	2012-08-02 04:45:12.850928	2012-08-02 04:48:37.477428
683	1-683	50	84	6084	90	6056	0	0	2012-08-02 04:44:54.779636	2012-08-02 04:48:37.477428
444	1-444	23	66	5504	59	7599	0	0	2012-08-02 04:44:06.489193	2012-08-02 04:48:37.477428
347	1-347	72	97	7617	31	9622	0	0	2012-08-02 04:44:00.599733	2012-08-02 04:48:37.477428
1048	1-1048	71	13	6923	99	5485	0	7	2012-08-02 04:47:37.794012	2012-08-02 04:48:37.477428
515	1-515	70	82	3305	38	9205	3711	0	2012-08-02 04:44:14.410235	2012-08-02 04:48:37.477428
1109	1-1109	99	66	8240	32	3848	0	0	2012-08-02 04:48:12.596482	2012-08-02 04:48:38.880866
878	1-878	20	42	8669	21	9253	0	0	2012-08-02 04:46:05.173031	2012-08-02 04:48:41.653454
266	1-266	61	98	6823	58	4313	0	0	2012-08-02 04:43:57.591288	2012-08-02 04:48:41.653454
1086	1-1086	22	17	4495	77	6360	0	0	2012-08-02 04:48:00.298187	2012-08-02 04:48:41.653454
1157	1-1157	65	37	8923	44	7444	0	0	2012-08-02 04:48:37.477428	2012-08-02 04:48:41.653454
1167	1-1167	3	44	2698	89	9237	0	0	2012-08-02 04:48:41.653454	2012-08-02 04:48:43.123105
750	1-750	49	44	5540	77	6107	0	0	2012-08-02 04:45:13.978361	2012-08-02 04:48:43.123105
1163	1-1163	8	90	81	84	2333	0	0	2012-08-02 04:48:40.261902	2012-08-02 04:48:43.598762
955	1-955	82	86	656	65	4550	0	0	2012-08-02 04:46:46.548692	2012-08-02 04:48:43.598762
539	1-539	73	63	4050	86	9410	0	0	2012-08-02 04:44:19.926982	2012-08-02 04:48:44.07351
976	1-976	9	60	3670	48	9258	0	0	2012-08-02 04:46:59.310523	2012-08-02 04:48:44.07351
72	1-72	49	28	9626	8	3403	0	0	2012-08-02 04:43:54.258966	2012-08-02 04:48:44.07351
961	1-961	29	38	995	4	9210	0	0	2012-08-02 04:46:49.920112	2012-08-02 04:48:44.549149
1118	1-1118	39	47	7324	60	6212	0	0	2012-08-02 04:48:19.35699	2012-08-02 04:48:44.549149
926	1-926	1	44	8470	84	9801	0	0	2012-08-02 04:46:31.005348	2012-08-02 04:48:44.549149
1096	1-1096	22	38	5394	8	5604	0	0	2012-08-02 04:48:04.928595	2012-08-02 04:48:44.549149
1173	1-1173	53	80	3829	38	9630	0	0	2012-08-02 04:48:44.549149	2012-08-02 04:48:44.549149
222	1-222	29	69	4043	95	5957	0	0	2012-08-02 04:43:56.411516	2012-08-02 04:48:45.234584
1085	1-1085	66	4	3710	14	8658	0	0	2012-08-02 04:47:59.823355	2012-08-02 04:48:46.537892
1092	1-1092	15	100	34	66	5033	0	0	2012-08-02 04:48:03.546719	2012-08-02 04:48:46.990717
1166	1-1166	8	2	2173	46	339	0	0	2012-08-02 04:48:41.189491	2012-08-02 04:48:46.990717
1072	1-1072	13	5	192	72	2578	0	2	2012-08-02 04:47:53.789484	2012-08-02 04:48:48.382795
1181	1-1181	60	72	975	73	5151	0	0	2012-08-02 04:48:48.382795	2012-08-02 04:48:48.382795
162	1-162	83	20	6749	14	3590	0	0	2012-08-02 04:43:55.432798	2012-08-02 04:48:49.885796
203	1-203	50	93	5040	21	1101	0	0	2012-08-02 04:43:56.03323	2012-08-02 04:48:51.780659
942	1-942	53	94	3516	47	8942	0	0	2012-08-02 04:46:38.484556	2012-08-02 04:48:52.245437
832	1-832	85	88	3247	59	1492	0	0	2012-08-02 04:45:47.170591	2012-08-02 04:48:52.245437
1190	1-1190	19	80	596	30	4513	0	0	2012-08-02 04:48:53.174764	2012-08-02 04:48:53.174764
881	1-881	18	88	3745	79	3311	0	0	2012-08-02 04:46:06.786416	2012-08-02 04:48:55.472532
1187	1-1187	50	33	2208	7	5848	0	0	2012-08-02 04:48:51.780659	2012-08-02 04:48:56.42285
114	1-114	32	89	2770	9	7966	0	0	2012-08-02 04:43:54.761102	2012-08-02 04:48:58.256865
449	1-449	16	52	6887	47	5015	0	0	2012-08-02 04:44:06.777354	2012-08-02 04:48:59.923148
909	1-909	39	22	2504	14	511	0	0	2012-08-02 04:46:21.891474	2012-08-02 04:48:59.923148
667	1-667	60	87	4142	78	9074	0	0	2012-08-02 04:44:51.784082	2012-08-02 04:48:59.923148
1172	1-1172	41	84	4915	63	7246	0	0	2012-08-02 04:48:44.07351	2012-08-02 04:48:59.923148
428	1-428	34	49	5244	41	6643	0	0	2012-08-02 04:44:05.990279	2012-08-02 04:48:59.923148
1170	1-1170	34	48	8602	44	8804	0	0	2012-08-02 04:48:43.123105	2012-08-02 04:49:00.398868
790	1-790	23	46	1794	76	8557	0	0	2012-08-02 04:45:28.48732	2012-08-02 04:49:01.326919
1202	1-1202	29	76	4729	83	8460	0	0	2012-08-02 04:49:01.326919	2012-08-02 04:49:01.326919
888	1-888	23	12	5667	47	9739	0	0	2012-08-02 04:46:08.920087	2012-08-02 04:49:02.222038
1199	1-1199	9	14	4443	49	9552	0	0	2012-08-02 04:48:59.923148	2012-08-02 04:49:03.625221
1162	1-1162	59	58	1907	19	4412	0	0	2012-08-02 04:48:39.808566	2012-08-02 04:49:03.625221
430	1-430	3	82	3453	35	4668	0	0	2012-08-02 04:44:06.012586	2012-08-02 04:49:05.010962
1093	1-1093	47	69	3327	94	3731	0	0	2012-08-02 04:48:03.734568	2012-08-02 04:49:05.443119
368	1-368	88	53	8080	18	2311	0	0	2012-08-02 04:44:01.610116	2012-08-02 04:49:06.371081
1183	1-1183	5	81	587	20	8976	0	0	2012-08-02 04:48:49.885796	2012-08-02 04:49:06.813006
127	1-127	71	70	2923	81	7882	0	0	2012-08-02 04:43:54.906199	2012-08-02 04:49:06.813006
899	1-899	15	18	784	27	2316	0	0	2012-08-02 04:46:16.047942	2012-08-02 04:49:06.813006
1214	1-1214	61	4	2254	18	4491	0	0	2012-08-02 04:49:06.813006	2012-08-02 04:49:06.813006
1215	1-1215	8	22	4237	35	223	0	0	2012-08-02 04:49:07.277447	2012-08-02 04:49:07.277447
694	1-694	6	8	3946	45	8659	0	0	2012-08-02 04:44:57.443068	2012-08-02 04:49:08.16101
1217	1-1217	80	39	2595	8	7669	0	0	2012-08-02 04:49:08.16101	2012-08-02 04:49:08.16101
1207	1-1207	35	45	9254	91	6429	0	0	2012-08-02 04:49:03.625221	2012-08-02 04:49:08.625194
730	1-730	88	50	9335	97	6981	0	0	2012-08-02 04:45:07.447903	2012-08-02 04:49:08.625194
903	1-903	41	78	6590	47	1012	0	0	2012-08-02 04:46:18.245902	2012-08-02 04:49:09.564436
1160	1-1160	57	24	504	66	3310	0	0	2012-08-02 04:48:38.880866	2012-08-02 04:49:10.039834
216	1-216	75	53	6257	97	7755	1905	0	2012-08-02 04:43:56.289414	2012-08-02 04:49:10.039834
1222	1-1222	80	20	2570	18	1202	0	0	2012-08-02 04:49:11.066647	2012-08-02 04:49:11.066647
1223	1-1223	5	95	374	42	1800	0	0	2012-08-02 04:49:11.541759	2012-08-02 04:49:12.050061
1225	1-1225	27	80	6318	82	7159	0	0	2012-08-02 04:49:12.050061	2012-08-02 04:49:12.050061
502	1-502	25	75	7083	61	9920	2121	0	2012-08-02 04:44:12.243002	2012-08-02 04:49:12.050061
1197	1-1197	35	25	694	89	954	0	0	2012-08-02 04:48:58.256865	2012-08-02 04:49:13.762418
1218	1-1218	5	20	4370	45	8039	0	0	2012-08-02 04:49:08.625194	2012-08-02 04:49:13.762418
1203	1-1203	43	36	1279	46	2253	0	0	2012-08-02 04:49:01.791313	2012-08-02 04:49:13.762418
1229	1-1229	27	45	6846	15	8490	0	0	2012-08-02 04:49:13.762418	2012-08-02 04:49:13.762418
1230	1-1230	69	52	1603	15	9427	0	0	2012-08-02 04:49:14.259857	2012-08-02 04:49:14.259857
1103	1-1103	29	74	3768	37	6995	0	0	2012-08-02 04:48:09.12755	2012-08-02 04:49:14.75711
827	1-827	96	33	9612	6	7873	0	0	2012-08-02 04:45:44.828001	2012-08-02 04:49:16.193065
1073	1-1073	77	5	2242	8	1491	0	0	2012-08-02 04:47:54.253317	2012-08-02 04:49:16.193065
831	1-831	55	67	2406	71	9928	0	0	2012-08-02 04:45:46.916236	2012-08-02 04:49:16.193065
928	1-928	15	84	2853	19	674	0	0	2012-08-02 04:46:32.430298	2012-08-02 04:49:16.193065
97	1-97	59	49	4260	26	1135	0	0	2012-08-02 04:43:54.560158	2012-08-02 04:49:16.193065
261	1-261	80	26	3932	59	8703	2936	0	2012-08-02 04:43:57.391538	2012-08-02 04:49:16.193065
970	1-970	79	85	8780	67	7788	4498	0	2012-08-02 04:46:55.112006	2012-08-02 04:49:16.193065
1013	1-1013	51	14	6438	3	1726	0	0	2012-08-02 04:47:17.795079	2012-08-02 04:49:17.629776
285	1-285	34	76	9295	3	9163	0	0	2012-08-02 04:43:58.445391	2012-08-02 04:49:18.071331
1194	1-1194	88	26	1665	52	9791	0	10	2012-08-02 04:48:55.947781	2012-08-02 04:49:20.932381
1245	1-1245	64	88	310	56	5792	0	4	2012-08-02 04:49:22.96518	2012-08-02 04:49:22.96518
854	1-854	56	25	5330	22	6400	0	0	2012-08-02 04:45:55.043781	2012-08-02 04:49:28.11433
1099	1-1099	29	31	6809	37	9375	0	0	2012-08-02 04:48:07.227148	2012-08-02 04:49:28.11433
600	1-600	94	62	7884	38	974	0	0	2012-08-02 04:44:36.761477	2012-08-02 04:49:28.11433
282	1-282	38	15	1681	25	9209	0	0	2012-08-02 04:43:58.334212	2012-08-02 04:49:28.11433
937	1-937	2	63	7992	28	9418	0	0	2012-08-02 04:46:35.3692	2012-08-02 04:49:28.11433
544	1-544	55	81	5538	34	9718	0	0	2012-08-02 04:44:21.220279	2012-08-02 04:49:28.11433
1252	1-1252	72	22	6825	15	8738	0	0	2012-08-02 04:49:28.11433	2012-08-02 04:49:28.11433
1126	1-1126	92	18	7638	85	4773	0	0	2012-08-02 04:48:23.820109	2012-08-02 04:49:33.439393
830	1-830	81	16	375	31	1015	0	3	2012-08-02 04:45:46.408424	2012-08-02 04:49:33.439393
380	1-380	44	63	4504	52	7457	0	0	2012-08-02 04:44:02.429711	2012-08-02 04:49:33.936492
1082	1-1082	20	85	714	67	316	0	4	2012-08-02 04:47:58.76244	2012-08-02 04:49:34.422505
377	1-377	63	71	8475	43	7900	0	0	2012-08-02 04:44:02.274355	2012-08-02 04:49:36.212131
787	1-787	88	19	9390	26	6003	0	0	2012-08-02 04:45:27.912824	2012-08-02 04:49:36.212131
1266	1-1266	78	21	553	19	3213	0	0	2012-08-02 04:49:36.212131	2012-08-02 04:49:36.212131
1244	1-1244	57	49	3861	92	7510	0	0	2012-08-02 04:49:22.003893	2012-08-02 04:49:37.692629
1078	1-1078	37	15	8233	63	6847	0	0	2012-08-02 04:47:56.763267	2012-08-02 04:49:37.692629
1261	1-1261	16	31	2716	18	4139	0	0	2012-08-02 04:49:33.439393	2012-08-02 04:49:37.692629
1269	1-1269	94	92	4033	49	8637	0	0	2012-08-02 04:49:37.692629	2012-08-02 04:49:37.692629
1237	1-1237	90	90	341	76	5713	0	0	2012-08-02 04:49:18.071331	2012-08-02 04:49:38.675835
846	1-846	52	63	4119	33	5921	0	0	2012-08-02 04:45:52.465544	2012-08-02 04:49:38.675835
1081	1-1081	83	3	3099	33	6140	0	0	2012-08-02 04:47:58.298357	2012-08-02 04:49:38.675835
1274	1-1274	89	78	186	11	435	0	0	2012-08-02 04:49:39.957609	2012-08-02 04:49:39.957609
1175	1-1175	81	41	4011	44	8020	0	0	2012-08-02 04:48:45.234584	2012-08-02 04:49:43.039873
96	1-96	13	85	7794	39	8888	0	0	2012-08-02 04:43:54.549154	2012-08-02 04:49:54.461358
1289	1-1289	24	37	833	31	5457	0	0	2012-08-02 04:49:47.955321	2012-08-02 04:49:54.461358
749	1-749	65	10	6258	20	6360	0	0	2012-08-02 04:45:13.249163	2012-08-02 04:49:54.461358
979	1-979	78	64	2141	65	2980	0	0	2012-08-02 04:47:00.603324	2012-08-02 04:49:56.902613
1188	1-1188	46	50	1294	88	6261	0	0	2012-08-02 04:48:52.245437	2012-08-02 04:49:57.367116
64	1-64	42	73	4231	36	6794	0	0	2012-08-02 04:43:54.16982	2012-08-02 04:49:57.367116
1304	1-1304	60	88	4021	73	7817	0	0	2012-08-02 04:49:57.367116	2012-08-02 04:49:57.367116
1248	1-1248	65	99	1576	63	2992	0	0	2012-08-02 04:49:25.218675	2012-08-02 04:50:02.525315
546	1-546	43	99	9911	3	9740	0	0	2012-08-02 04:44:21.927658	2012-08-02 04:50:02.525315
1034	1-1034	79	99	6577	49	3092	0	0	2012-08-02 04:47:28.017119	2012-08-02 04:50:02.525315
456	1-456	58	60	7302	3	8236	0	0	2012-08-02 04:44:07.264086	2012-08-02 04:50:03.011659
1271	1-1271	13	9	8169	99	6151	0	0	2012-08-02 04:49:38.675835	2012-08-02 04:50:03.011659
177	1-177	41	71	6539	61	2258	0	0	2012-08-02 04:43:55.62117	2012-08-02 04:50:03.011659
1146	1-1146	61	43	8020	70	7109	0	0	2012-08-02 04:48:33.135428	2012-08-02 04:50:03.011659
1240	1-1240	78	49	4732	60	9548	0	0	2012-08-02 04:49:19.507994	2012-08-02 04:50:03.011659
1313	1-1313	27	7	2533	49	9313	0	0	2012-08-02 04:50:03.011659	2012-08-02 04:50:03.011659
18	1-18	18	28	8495	3	8117	3006	0	2012-08-02 04:43:53.643058	2012-08-02 04:50:03.011659
758	1-758	82	67	5148	98	8573	0	0	2012-08-02 04:45:17.612758	2012-08-02 04:50:03.520664
398	1-398	99	73	4999	41	3931	0	0	2012-08-02 04:44:03.62594	2012-08-02 04:50:03.520664
517	1-517	40	52	8491	3	3541	0	0	2012-08-02 04:44:14.631443	2012-08-02 04:50:03.520664
1121	1-1121	30	67	3097	49	7765	0	0	2012-08-02 04:48:20.726899	2012-08-02 04:50:03.520664
1314	1-1314	48	100	3144	67	4732	0	0	2012-08-02 04:50:03.520664	2012-08-02 04:50:05.976552
725	1-725	71	51	6518	66	1631	0	0	2012-08-02 04:45:04.982317	2012-08-02 04:50:07.445654
1319	1-1319	97	34	1598	43	3179	0	0	2012-08-02 04:50:05.512878	2012-08-02 04:50:08.870915
1153	1-1153	84	47	4156	20	6880	0	0	2012-08-02 04:48:35.897422	2012-08-02 04:50:08.870915
1324	1-1324	21	10	1064	57	3861	0	0	2012-08-02 04:50:08.870915	2012-08-02 04:50:08.870915
613	1-613	13	61	1623	20	3436	0	0	2012-08-02 04:44:39.42509	2012-08-02 04:50:09.35699
1227	1-1227	75	91	537	5	9641	0	6	2012-08-02 04:49:13.000328	2012-08-02 04:50:09.35699
1147	1-1147	28	6	6567	24	1716	0	0	2012-08-02 04:48:33.621133	2012-08-02 04:50:09.35699
1285	1-1285	88	17	2645	94	2086	0	0	2012-08-02 04:49:46.143961	2012-08-02 04:50:09.35699
1134	1-1134	37	93	8556	97	1949	0	0	2012-08-02 04:48:28.304852	2012-08-02 04:50:09.887379
1303	1-1303	19	34	7062	29	7807	0	0	2012-08-02 04:49:56.902613	2012-08-02 04:50:09.887379
1138	1-1138	53	61	4167	89	4574	0	0	2012-08-02 04:48:30.141153	2012-08-02 04:50:10.384596
765	1-765	100	50	1741	4	6703	0	0	2012-08-02 04:45:20.41903	2012-08-02 04:50:11.202161
875	1-875	17	21	9573	99	8873	0	6	2012-08-02 04:46:03.77973	2012-08-02 04:50:11.202161
1326	1-1326	92	99	1282	1	6280	0	0	2012-08-02 04:50:09.887379	2012-08-02 04:50:11.202161
1242	1-1242	60	30	1132	19	2500	0	0	2012-08-02 04:49:20.932381	2012-08-02 04:50:11.202161
321	1-321	67	2	4295	4	4461	0	0	2012-08-02 04:43:59.690414	2012-08-02 04:50:11.202161
1079	1-1079	98	79	5243	20	9453	0	0	2012-08-02 04:47:57.25996	2012-08-02 04:50:11.202161
762	1-762	70	100	7485	40	6850	0	0	2012-08-02 04:45:18.960938	2012-08-02 04:50:11.710862
813	1-813	4	63	1858	51	2867	0	0	2012-08-02 04:45:38.92767	2012-08-02 04:50:11.710862
1332	1-1332	52	40	877	63	9720	0	0	2012-08-02 04:50:11.710862	2012-08-02 04:50:11.710862
1311	1-1311	99	41	1214	63	3289	0	0	2012-08-02 04:50:02.061433	2012-08-02 04:50:14.616612
1337	1-1337	72	78	3238	64	4875	0	0	2012-08-02 04:50:14.616612	2012-08-02 04:50:14.616612
1185	1-1185	53	54	1469	2	3479	0	0	2012-08-02 04:48:50.830514	2012-08-02 04:50:16.063709
1339	1-1339	31	72	3267	8	3336	0	0	2012-08-02 04:50:16.063709	2012-08-02 04:50:16.063709
1152	1-1152	80	92	3925	50	7378	0	0	2012-08-02 04:48:35.444308	2012-08-02 04:50:18.449453
884	1-884	83	71	8913	24	5198	0	0	2012-08-02 04:46:07.040631	2012-08-02 04:50:18.449453
1171	1-1171	26	65	2010	90	4071	0	0	2012-08-02 04:48:43.598762	2012-08-02 04:50:18.449453
1344	1-1344	76	50	5818	61	8769	0	0	2012-08-02 04:50:18.449453	2012-08-02 04:50:18.449453
950	1-950	59	39	9675	45	7675	0	0	2012-08-02 04:46:43.300455	2012-08-02 04:50:19.908321
1347	1-1347	3	12	61	39	2460	0	0	2012-08-02 04:50:19.908321	2012-08-02 04:50:19.908321
1350	1-1350	3	45	6245	26	1012	0	0	2012-08-02 04:50:21.013014	2012-08-02 04:50:21.013014
1159	1-1159	84	37	5409	22	8876	0	0	2012-08-02 04:48:38.416576	2012-08-02 04:50:21.974134
620	1-620	35	95	7069	31	6632	0	0	2012-08-02 04:44:40.884494	2012-08-02 04:50:23.399057
1276	1-1276	53	70	4017	99	2704	0	0	2012-08-02 04:49:40.907951	2012-08-02 04:50:24.901635
640	1-640	37	73	6786	52	5302	0	0	2012-08-02 04:44:45.946056	2012-08-02 04:50:25.884942
1300	1-1300	45	20	9032	85	9013	0	4	2012-08-02 04:49:54.461358	2012-08-02 04:50:25.884942
1297	1-1297	49	65	1932	66	1743	0	0	2012-08-02 04:49:53.191019	2012-08-02 04:50:25.884942
1287	1-1287	2	85	6472	10	8188	0	0	2012-08-02 04:49:47.104574	2012-08-02 04:50:28.723851
723	1-723	19	36	9627	7	9465	0	0	2012-08-02 04:45:04.871264	2012-08-02 04:50:28.723851
1367	1-1367	35	22	182	99	3044	0	0	2012-08-02 04:50:29.983671	2012-08-02 04:50:29.983671
296	1-296	45	73	2263	53	5737	0	0	2012-08-02 04:43:58.811739	2012-08-02 04:50:30.447714
962	1-962	34	94	3863	66	3039	0	0	2012-08-02 04:46:50.328864	2012-08-02 04:50:30.447714
589	1-589	50	15	9995	32	4289	0	0	2012-08-02 04:44:33.89685	2012-08-02 04:50:30.922791
1234	1-1234	3	59	4838	91	3243	0	0	2012-08-02 04:49:16.193065	2012-08-02 04:50:30.922791
805	1-805	32	15	8927	83	4805	0	0	2012-08-02 04:45:35.282288	2012-08-02 04:50:30.922791
1338	1-1338	64	36	3184	23	3439	0	2	2012-08-02 04:50:15.091351	2012-08-02 04:50:30.922791
1369	1-1369	28	23	3450	75	8091	0	0	2012-08-02 04:50:30.922791	2012-08-02 04:50:30.922791
949	1-949	66	41	1529	52	465	0	0	2012-08-02 04:46:42.726524	2012-08-02 04:50:31.894938
853	1-853	59	77	6663	13	5830	0	0	2012-08-02 04:45:54.653894	2012-08-02 04:50:31.894938
265	1-265	32	22	5837	34	1769	0	0	2012-08-02 04:43:57.54682	2012-08-02 04:50:32.370327
1249	1-1249	89	94	2386	20	3971	0	0	2012-08-02 04:49:25.726429	2012-08-02 04:50:32.834168
1098	1-1098	36	94	8861	84	4969	0	0	2012-08-02 04:48:06.763128	2012-08-02 04:50:32.834168
1373	1-1373	17	34	10	94	5914	0	0	2012-08-02 04:50:32.834168	2012-08-02 04:50:32.834168
874	1-874	18	96	594	36	7375	0	0	2012-08-02 04:46:03.658108	2012-08-02 04:50:34.259509
1283	1-1283	99	64	3271	100	7070	0	0	2012-08-02 04:49:45.337118	2012-08-02 04:50:34.259509
435	1-435	9	24	2362	72	6479	0	3	2012-08-02 04:44:06.101588	2012-08-02 04:50:34.259509
1330	1-1330	7	4	8338	30	8199	0	0	2012-08-02 04:50:11.202161	2012-08-02 04:50:34.259509
1376	1-1376	78	72	6556	1	8342	0	0	2012-08-02 04:50:34.259509	2012-08-02 04:50:34.259509
1273	1-1273	16	92	3453	42	4493	0	0	2012-08-02 04:49:39.482699	2012-08-02 04:50:35.737094
1342	1-1342	9	92	6163	71	9114	0	0	2012-08-02 04:50:17.499505	2012-08-02 04:50:35.737094
1360	1-1360	11	85	2052	65	3777	0	0	2012-08-02 04:50:25.884942	2012-08-02 04:50:36.205914
934	1-934	100	98	4234	7	7621	1301	0	2012-08-02 04:46:34.331011	2012-08-02 04:50:36.205914
1228	1-1228	34	21	1501	51	4929	0	0	2012-08-02 04:49:13.287817	2012-08-02 04:50:36.681123
1149	1-1149	87	45	5250	92	5393	0	0	2012-08-02 04:48:34.074246	2012-08-02 04:50:36.681123
1358	1-1358	13	21	3446	70	6605	0	0	2012-08-02 04:50:24.901635	2012-08-02 04:50:36.681123
624	1-624	37	45	8929	43	4037	0	0	2012-08-02 04:44:41.47031	2012-08-02 04:50:37.145233
1349	1-1349	29	54	4904	56	3590	0	0	2012-08-02 04:50:20.527105	2012-08-02 04:50:37.609539
1355	1-1355	98	31	5288	36	6099	0	0	2012-08-02 04:50:23.399057	2012-08-02 04:50:42.447879
432	1-432	45	5	4353	12	1110	0	0	2012-08-02 04:44:06.05688	2012-08-02 04:50:48.600754
1140	1-1140	36	5	9044	94	6419	0	0	2012-08-02 04:48:31.069266	2012-08-02 04:50:48.600754
1406	1-1406	16	12	93	22	9271	0	5	2012-08-02 04:50:48.600754	2012-08-02 04:50:48.600754
1381	1-1381	6	81	3158	45	6423	0	0	2012-08-02 04:50:36.681123	2012-08-02 04:50:50.512175
1077	1-1077	18	86	1121	42	3779	0	0	2012-08-02 04:47:56.054253	2012-08-02 04:50:51.462164
1379	1-1379	67	42	6197	92	8771	0	0	2012-08-02 04:50:35.737094	2012-08-02 04:50:51.462164
1065	1-1065	16	86	3111	71	9999	0	0	2012-08-02 04:47:50.574646	2012-08-02 04:50:51.462164
1357	1-1357	44	7	1781	22	6166	0	0	2012-08-02 04:50:24.349313	2012-08-02 04:50:52.412428
788	1-788	5	60	6253	43	5828	0	0	2012-08-02 04:45:28.155834	2012-08-02 04:50:54.798713
1418	1-1418	49	21	183	86	8109	0	0	2012-08-02 04:50:54.798713	2012-08-02 04:50:54.798713
1108	1-1108	78	43	7834	21	9212	4962	0	2012-08-02 04:48:12.132532	2012-08-02 04:50:54.798713
748	1-748	63	46	4532	48	6149	0	2	2012-08-02 04:45:13.127452	2012-08-02 04:50:55.284861
1419	1-1419	46	70	304	46	9890	0	0	2012-08-02 04:50:55.284861	2012-08-02 04:50:55.284861
998	1-998	64	32	9416	36	9783	0	0	2012-08-02 04:47:09.719513	2012-08-02 04:50:56.235258
189	1-189	2	27	4924	47	2699	0	0	2012-08-02 04:43:55.77727	2012-08-02 04:50:56.235258
1372	1-1372	97	34	7165	98	9642	1706	0	2012-08-02 04:50:32.370327	2012-08-02 04:50:56.235258
852	1-852	85	13	7416	55	7006	0	0	2012-08-02 04:45:54.444084	2012-08-02 04:50:59.825108
1395	1-1395	51	80	158	79	6364	0	0	2012-08-02 04:50:42.447879	2012-08-02 04:51:00.300434
1168	1-1168	3	77	4442	52	3026	0	0	2012-08-02 04:48:42.117957	2012-08-02 04:51:01.228622
1263	1-1263	54	92	4384	100	8346	0	0	2012-08-02 04:49:34.422505	2012-08-02 04:51:01.71434
752	1-752	72	23	8212	61	7380	0	0	2012-08-02 04:45:14.839867	2012-08-02 04:51:01.71434
1432	1-1432	45	87	685	92	9130	0	0	2012-08-02 04:51:01.71434	2012-08-02 04:51:01.71434
891	1-891	24	89	9274	98	7194	1023	0	2012-08-02 04:46:11.219821	2012-08-02 04:51:01.71434
1284	1-1284	31	1	2535	47	5072	0	0	2012-08-02 04:49:45.801078	2012-08-02 04:51:04.067541
1177	1-1177	4	34	5888	50	9436	1564	0	2012-08-02 04:48:46.537892	2012-08-02 04:51:04.067541
1327	1-1327	42	89	2222	83	5570	0	0	2012-08-02 04:50:10.384596	2012-08-02 04:51:04.542873
143	1-143	21	96	9565	26	5162	0	0	2012-08-02 04:43:55.106895	2012-08-02 04:51:10.831582
284	1-284	2	53	1343	97	212	0	2	2012-08-02 04:43:58.378634	2012-08-02 04:51:10.831582
1220	1-1220	77	92	3168	29	2771	0	6	2012-08-02 04:49:09.564436	2012-08-02 04:51:10.831582
141	1-141	21	5	6108	20	1734	0	0	2012-08-02 04:43:55.08459	2012-08-02 04:51:10.831582
1447	1-1447	3	1	2534	96	8414	0	0	2012-08-02 04:51:10.831582	2012-08-02 04:51:10.831582
1325	1-1325	14	20	7413	61	7624	3684	0	2012-08-02 04:50:09.35699	2012-08-02 04:51:10.831582
770	1-770	14	99	8648	16	3789	0	0	2012-08-02 04:45:22.330786	2012-08-02 04:51:15.338593
332	1-332	40	39	7573	79	8620	0	0	2012-08-02 04:44:00.045311	2012-08-02 04:51:15.338593
1312	1-1312	28	3	8595	99	9768	0	0	2012-08-02 04:50:02.525315	2012-08-02 04:51:15.338593
681	1-681	45	37	5357	72	5695	0	0	2012-08-02 04:44:54.746122	2012-08-02 04:51:15.824937
1076	1-1076	34	56	3071	36	2334	0	0	2012-08-02 04:47:55.899139	2012-08-02 04:51:16.774946
322	1-322	22	77	6533	26	2836	0	0	2012-08-02 04:43:59.701412	2012-08-02 04:51:16.774946
1113	1-1113	8	63	6646	38	1473	0	0	2012-08-02 04:48:15.921298	2012-08-02 04:51:16.774946
1458	1-1458	86	98	77	56	4785	0	0	2012-08-02 04:51:16.774946	2012-08-02 04:51:16.774946
531	1-531	79	1	2833	4	6025	0	0	2012-08-02 04:44:18.620383	2012-08-02 04:51:17.283399
1058	1-1058	41	25	6048	55	4135	0	0	2012-08-02 04:47:45.427007	2012-08-02 04:51:17.283399
1431	1-1431	80	13	1306	57	5064	0	0	2012-08-02 04:51:01.228622	2012-08-02 04:51:22.696022
1356	1-1356	44	52	1703	13	8482	0	0	2012-08-02 04:50:23.885395	2012-08-02 04:51:23.679401
810	1-810	31	94	2925	33	2025	0	0	2012-08-02 04:45:36.828742	2012-08-02 04:51:23.679401
1026	1-1026	8	46	5674	62	9231	0	0	2012-08-02 04:47:24.160261	2012-08-02 04:51:23.679401
1005	1-1005	44	82	5351	75	3692	0	0	2012-08-02 04:47:14.29316	2012-08-02 04:51:24.176605
840	1-840	26	32	4265	87	8151	0	0	2012-08-02 04:45:50.024058	2012-08-02 04:51:25.126746
356	1-356	46	29	3486	78	9729	0	0	2012-08-02 04:44:01.120538	2012-08-02 04:51:25.126746
763	1-763	53	76	4758	9	4665	0	0	2012-08-02 04:45:19.391716	2012-08-02 04:51:25.56895
1473	1-1473	75	78	2681	56	5902	0	0	2012-08-02 04:51:25.126746	2012-08-02 04:51:25.56895
1474	1-1474	33	15	812	98	8547	0	0	2012-08-02 04:51:25.56895	2012-08-02 04:51:25.56895
1459	1-1459	66	4	3927	25	1767	0	0	2012-08-02 04:51:17.283399	2012-08-02 04:51:26.055056
907	1-907	96	32	3201	76	8780	0	0	2012-08-02 04:46:20.14605	2012-08-02 04:51:26.055056
1475	1-1475	78	6	4312	60	6829	0	0	2012-08-02 04:51:26.055056	2012-08-02 04:51:26.055056
1470	1-1470	38	25	1093	94	8546	0	0	2012-08-02 04:51:23.679401	2012-08-02 04:51:26.530111
1007	1-1007	24	86	8103	60	6817	0	0	2012-08-02 04:47:14.447954	2012-08-02 04:51:26.994197
1031	1-1031	56	95	6756	33	7599	0	0	2012-08-02 04:47:26.591294	2012-08-02 04:51:26.994197
1477	1-1477	33	2	1129	95	9941	0	0	2012-08-02 04:51:26.994197	2012-08-02 04:51:26.994197
603	1-603	36	1	3608	27	7753	0	0	2012-08-02 04:44:37.424513	2012-08-02 04:51:31.423723
1343	1-1343	97	84	3315	67	2406	0	0	2012-08-02 04:50:17.974591	2012-08-02 04:51:31.423723
1455	1-1455	78	79	422	3	2069	0	0	2012-08-02 04:51:15.338593	2012-08-02 04:51:31.910011
1368	1-1368	36	53	4825	54	5726	0	0	2012-08-02 04:50:30.447714	2012-08-02 04:51:31.910011
1485	1-1485	49	64	1880	33	5858	0	0	2012-08-02 04:51:31.910011	2012-08-02 04:51:36.726515
1494	1-1494	2	27	625	36	7918	0	0	2012-08-02 04:51:36.726515	2012-08-02 04:51:36.726515
971	1-971	24	26	4370	11	7010	75	0	2012-08-02 04:46:56.393829	2012-08-02 04:51:36.726515
1122	1-1122	91	86	4050	67	3026	0	0	2012-08-02 04:48:20.782192	2012-08-02 04:51:37.212597
1495	1-1495	11	31	1020	87	7990	0	0	2012-08-02 04:51:37.212597	2012-08-02 04:51:37.212597
779	1-779	62	32	8491	81	6755	0	0	2012-08-02 04:45:25.845792	2012-08-02 04:51:42.227933
989	1-989	24	90	2957	68	6497	0	0	2012-08-02 04:47:05.653748	2012-08-02 04:51:42.227933
1437	1-1437	87	55	3767	1	9173	0	0	2012-08-02 04:51:04.067541	2012-08-02 04:51:46.679714
1299	1-1299	18	42	1956	93	9421	0	7	2012-08-02 04:49:53.986388	2012-08-02 04:51:50.194711
1522	1-1522	61	25	153	69	399	0	0	2012-08-02 04:51:50.194711	2012-08-02 04:51:50.194711
1468	1-1468	31	57	2298	52	5769	0	0	2012-08-02 04:51:22.696022	2012-08-02 04:51:52.94539
1116	1-1116	54	57	3947	15	3730	0	0	2012-08-02 04:48:18.40696	2012-08-02 04:51:52.94539
1471	1-1471	54	87	4956	82	1744	0	0	2012-08-02 04:51:24.176605	2012-08-02 04:51:54.315396
1396	1-1396	100	90	1343	93	7088	0	0	2012-08-02 04:50:42.9122	2012-08-02 04:51:57.099444
740	1-740	34	47	6677	34	7640	0	0	2012-08-02 04:45:09.911822	2012-08-02 04:51:58.701101
1421	1-1421	66	22	3392	32	8468	0	0	2012-08-02 04:50:56.235258	2012-08-02 04:51:58.701101
514	1-514	89	35	1093	47	8608	0	0	2012-08-02 04:44:14.366161	2012-08-02 04:51:58.701101
1539	1-1539	77	40	1893	35	4988	0	0	2012-08-02 04:51:58.701101	2012-08-02 04:51:58.701101
734	1-734	81	54	9882	46	7089	0	0	2012-08-02 04:45:08.077755	2012-08-02 04:51:59.165182
601	1-601	12	85	4403	80	3157	0	0	2012-08-02 04:44:37.114648	2012-08-02 04:51:59.165182
1540	1-1540	57	80	1137	54	5928	0	11	2012-08-02 04:51:59.165182	2012-08-02 04:51:59.165182
51	1-51	40	64	9466	7	5584	2100	0	2012-08-02 04:43:54.011694	2012-08-02 04:51:59.165182
417	1-417	2	36	9558	95	8921	0	0	2012-08-02 04:44:04.847835	2012-08-02 04:51:59.6075
1310	1-1310	72	68	8889	77	3700	0	0	2012-08-02 04:50:01.608534	2012-08-02 04:51:59.6075
682	1-682	44	82	7942	8	7163	0	0	2012-08-02 04:44:54.757703	2012-08-02 04:52:00.093292
1543	1-1543	64	2	52	67	2860	0	0	2012-08-02 04:52:00.546379	2012-08-02 04:52:00.546379
416	1-416	26	49	3994	20	5991	0	0	2012-08-02 04:44:04.834065	2012-08-02 04:52:01.02139
1364	1-1364	36	23	5697	19	420	0	0	2012-08-02 04:50:27.784903	2012-08-02 04:52:01.02139
1363	1-1363	6	67	2714	38	1499	0	0	2012-08-02 04:50:27.309863	2012-08-02 04:52:01.496654
778	1-778	25	46	4336	45	5407	0	10	2012-08-02 04:45:25.645354	2012-08-02 04:52:01.496654
1047	1-1047	11	84	4586	3	4384	0	0	2012-08-02 04:47:36.766526	2012-08-02 04:52:03.054506
1547	1-1547	18	59	1960	86	4286	0	0	2012-08-02 04:52:03.054506	2012-08-02 04:52:03.054506
1545	1-1545	26	19	4060	67	4516	0	0	2012-08-02 04:52:01.496654	2012-08-02 04:52:05.419795
1541	1-1541	59	12	1128	68	8924	0	0	2012-08-02 04:51:59.6075	2012-08-02 04:52:05.419795
1331	1-1331	24	96	4760	63	6551	0	0	2012-08-02 04:50:11.699997	2012-08-02 04:52:05.927995
720	1-720	67	15	4822	47	4011	0	0	2012-08-02 04:45:04.39377	2012-08-02 04:52:07.640049
1456	1-1456	15	93	2792	85	5500	0	0	2012-08-02 04:51:15.824937	2012-08-02 04:52:08.578907
1484	1-1484	39	27	4619	84	6897	0	0	2012-08-02 04:51:31.423723	2012-08-02 04:52:12.666085
764	1-764	86	10	2510	54	523	0	0	2012-08-02 04:45:19.564358	2012-08-02 04:52:12.666085
1192	1-1192	31	60	5433	31	9406	0	4	2012-08-02 04:48:54.125122	2012-08-02 04:52:18.122814
1573	1-1573	45	32	4893	60	6611	0	0	2012-08-02 04:52:18.122814	2012-08-02 04:52:18.122814
842	1-842	70	7	6438	53	4531	0	0	2012-08-02 04:45:51.106336	2012-08-02 04:52:20.011937
943	1-943	43	7	6023	93	6972	0	0	2012-08-02 04:46:39.600283	2012-08-02 04:52:20.011937
597	1-597	26	51	8098	68	4929	0	0	2012-08-02 04:44:36.164501	2012-08-02 04:52:20.011937
607	1-607	28	70	4680	16	1306	0	0	2012-08-02 04:44:37.800364	2012-08-02 04:52:20.973416
549	1-549	94	70	3315	81	2268	0	2	2012-08-02 04:44:22.977606	2012-08-02 04:52:20.973416
1403	1-1403	48	11	1820	58	6873	0	0	2012-08-02 04:50:47.165034	2012-08-02 04:52:20.973416
1577	1-1577	44	53	718	7	4854	0	0	2012-08-02 04:52:20.011937	2012-08-02 04:52:20.973416
1579	1-1579	57	72	4245	70	7209	0	0	2012-08-02 04:52:20.973416	2012-08-02 04:52:20.973416
1556	1-1556	22	9	6872	15	8418	0	0	2012-08-02 04:52:07.640049	2012-08-02 04:52:21.448712
1580	1-1580	23	48	749	9	9888	0	0	2012-08-02 04:52:21.448712	2012-08-02 04:52:21.448712
78	1-78	29	29	7579	72	8052	0	0	2012-08-02 04:43:54.348315	2012-08-02 04:52:21.901689
1581	1-1581	22	14	1453	19	850	0	0	2012-08-02 04:52:21.901689	2012-08-02 04:52:21.901689
1553	1-1553	18	67	2234	82	7564	0	0	2012-08-02 04:52:05.419795	2012-08-02 04:52:22.842548
1010	1-1010	27	82	4701	12	5858	0	0	2012-08-02 04:47:16.502891	2012-08-02 04:52:22.842548
978	1-978	70	52	6434	31	7316	0	0	2012-08-02 04:46:59.719276	2012-08-02 04:52:23.317795
1583	1-1583	47	12	5266	8	677	0	0	2012-08-02 04:52:22.842548	2012-08-02 04:52:23.317795
1555	1-1555	14	92	885	69	6661	0	0	2012-08-02 04:52:05.927995	2012-08-02 04:52:23.317795
1584	1-1584	65	71	1149	52	7894	0	0	2012-08-02 04:52:23.317795	2012-08-02 04:52:23.317795
1586	1-1586	42	66	41	70	3903	0	0	2012-08-02 04:52:23.815082	2012-08-02 04:52:23.815082
1476	1-1476	70	87	7410	86	4785	0	0	2012-08-02 04:51:26.530111	2012-08-02 04:52:26.212314
52	1-52	41	24	7014	59	7492	76	0	2012-08-02 04:43:54.022831	2012-08-02 04:52:26.212314
1424	1-1424	85	91	1061	97	2451	0	0	2012-08-02 04:50:57.925292	2012-08-02 04:52:26.687425
1592	1-1592	11	88	660	91	1967	0	0	2012-08-02 04:52:26.687425	2012-08-02 04:52:26.687425
1282	1-1282	8	49	3472	9	9262	0	0	2012-08-02 04:49:44.862276	2012-08-02 04:52:29.74751
687	1-687	74	28	6556	93	6203	0	0	2012-08-02 04:44:56.249237	2012-08-02 04:52:30.211563
1382	1-1382	39	21	5250	90	7047	0	0	2012-08-02 04:50:37.145233	2012-08-02 04:52:30.211563
419	1-419	78	23	8308	50	9556	0	0	2012-08-02 04:44:05.070729	2012-08-02 04:52:30.211563
946	1-946	84	14	8197	93	2822	0	0	2012-08-02 04:46:40.583451	2012-08-02 04:52:31.216982
863	1-863	63	14	9245	23	407	0	0	2012-08-02 04:45:59.028985	2012-08-02 04:52:31.216982
1528	1-1528	74	93	7916	57	3666	0	0	2012-08-02 04:51:52.94539	2012-08-02 04:52:32.156544
453	1-453	22	29	7262	93	6764	0	0	2012-08-02 04:44:07.197755	2012-08-02 04:52:32.156544
1507	1-1507	67	81	5615	36	8546	0	0	2012-08-02 04:51:42.227933	2012-08-02 04:52:32.631773
844	1-844	33	81	6641	72	6743	0	0	2012-08-02 04:45:51.658956	2012-08-02 04:52:32.631773
1016	1-1016	32	29	8267	97	1810	0	0	2012-08-02 04:47:18.923808	2012-08-02 04:52:33.570252
655	1-655	78	46	5972	50	2238	0	2	2012-08-02 04:44:50.16741	2012-08-02 04:52:34.056628
1537	1-1537	81	72	4244	46	5347	0	0	2012-08-02 04:51:57.099444	2012-08-02 04:52:34.056628
1442	1-1442	25	71	2400	32	7598	0	0	2012-08-02 04:51:07.471684	2012-08-02 04:52:34.553938
391	1-391	11	82	9001	19	5174	0	0	2012-08-02 04:44:03.194388	2012-08-02 04:52:35.017945
1262	1-1262	11	18	5475	15	5552	0	0	2012-08-02 04:49:33.936492	2012-08-02 04:52:35.017945
1479	1-1479	98	71	8496	70	2092	0	0	2012-08-02 04:51:27.944713	2012-08-02 04:52:35.017945
91	1-91	57	11	7994	84	9118	669	0	2012-08-02 04:43:54.493319	2012-08-02 04:52:35.017945
190	1-190	3	82	9774	23	5505	2638	0	2012-08-02 04:43:55.7888	2012-08-02 04:52:35.017945
1569	1-1569	36	70	6961	95	5493	0	0	2012-08-02 04:52:15.184602	2012-08-02 04:52:35.493284
1611	1-1611	70	36	20	70	9555	0	0	2012-08-02 04:52:35.493284	2012-08-02 04:52:35.493284
1204	1-1204	37	52	6982	12	9923	390	0	2012-08-02 04:49:02.222038	2012-08-02 04:52:35.493284
1612	1-1612	18	100	2321	20	6854	0	0	2012-08-02 04:52:35.979463	2012-08-02 04:52:35.979463
397	1-397	8	99	9285	47	6076	0	0	2012-08-02 04:44:03.61459	2012-08-02 04:52:36.918414
1544	1-1544	70	20	2071	96	6398	0	0	2012-08-02 04:52:01.02139	2012-08-02 04:52:37.404476
1504	1-1504	58	84	3246	52	4168	0	0	2012-08-02 04:51:40.79218	2012-08-02 04:52:37.902065
1596	1-1596	90	54	676	71	8734	0	7	2012-08-02 04:52:28.80874	2012-08-02 04:52:39.31594
1538	1-1538	24	63	1060	23	6834	0	0	2012-08-02 04:51:57.55202	2012-08-02 04:52:40.26621
1200	1-1200	89	24	3291	48	8908	0	0	2012-08-02 04:49:00.398868	2012-08-02 04:52:40.26621
1542	1-1542	70	54	1454	82	8838	0	0	2012-08-02 04:52:00.093292	2012-08-02 04:52:40.26621
1621	1-1621	1	23	2764	8	5774	0	0	2012-08-02 04:52:40.26621	2012-08-02 04:52:40.26621
1608	1-1608	26	21	4084	99	9613	0	16	2012-08-02 04:52:34.056628	2012-08-02 04:52:42.619103
784	1-784	67	73	9194	91	2373	0	0	2012-08-02 04:45:27.160875	2012-08-02 04:52:42.619103
1614	1-1614	61	73	4057	21	8385	0	0	2012-08-02 04:52:36.918414	2012-08-02 04:52:42.619103
1576	1-1576	80	83	616	29	8984	0	0	2012-08-02 04:52:19.537219	2012-08-02 04:52:42.619103
1626	1-1626	92	29	4624	73	7370	0	0	2012-08-02 04:52:42.619103	2012-08-02 04:52:42.619103
714	1-714	59	19	9683	39	859	513	0	2012-08-02 04:45:02.846581	2012-08-02 04:52:42.619103
963	1-963	25	83	5952	45	9283	0	0	2012-08-02 04:46:50.826047	2012-08-02 04:52:43.117058
1600	1-1600	1	34	9211	23	9293	0	0	2012-08-02 04:52:30.211563	2012-08-02 04:52:43.117058
1629	1-1629	92	14	1531	80	1127	0	0	2012-08-02 04:52:44.055987	2012-08-02 04:52:44.055987
1634	1-1634	27	87	2480	14	2301	0	0	2012-08-02 04:52:45.956287	2012-08-02 04:52:46.40883
916	1-916	24	63	5554	92	4559	0	0	2012-08-02 04:46:25.967903	2012-08-02 04:52:48.342372
838	1-838	60	63	8593	69	3453	0	0	2012-08-02 04:45:49.181841	2012-08-02 04:52:48.342372
857	1-857	44	69	8432	42	8304	5184	0	2012-08-02 04:45:57.040311	2012-08-02 04:52:48.342372
1616	1-1616	56	45	942	84	3345	0	0	2012-08-02 04:52:37.902065	2012-08-02 04:52:49.800296
470	1-470	38	83	9180	39	7158	4658	0	2012-08-02 04:44:08.604506	2012-08-02 04:52:49.800296
1056	1-1056	73	12	7189	94	9235	0	0	2012-08-02 04:47:43.228679	2012-08-02 04:52:51.258471
1302	1-1302	59	66	2107	20	1722	0	0	2012-08-02 04:49:56.428032	2012-08-02 04:52:52.1866
1572	1-1572	67	51	3654	87	5096	0	0	2012-08-02 04:52:17.659213	2012-08-02 04:52:55.071644
691	1-691	82	8	4651	45	4307	0	0	2012-08-02 04:44:56.548293	2012-08-02 04:52:55.071644
1057	1-1057	77	57	2264	34	9266	0	0	2012-08-02 04:47:44.432622	2012-08-02 04:52:56.011076
1514	1-1514	24	31	9436	71	5946	0	0	2012-08-02 04:51:45.752071	2012-08-02 04:52:56.011076
1452	1-1452	42	47	8421	27	8105	0	0	2012-08-02 04:51:14.344314	2012-08-02 04:52:56.011076
1500	1-1500	24	14	3486	5	7042	0	0	2012-08-02 04:51:39.234536	2012-08-02 04:52:56.949971
760	1-760	68	21	6035	22	5763	0	0	2012-08-02 04:45:18.419644	2012-08-02 04:52:56.949971
800	1-800	47	49	2191	87	5081	0	3	2012-08-02 04:45:33.061529	2012-08-02 04:52:58.794864
1383	1-1383	76	63	3346	53	5091	0	0	2012-08-02 04:50:37.609539	2012-08-02 04:52:59.700764
1657	1-1657	37	5	4646	21	8765	0	0	2012-08-02 04:52:56.949971	2012-08-02 04:52:59.700764
1516	1-1516	45	45	5767	91	1405	0	7	2012-08-02 04:51:46.679714	2012-08-02 04:52:59.700764
922	1-922	33	54	4491	79	615	0	0	2012-08-02 04:46:29.08323	2012-08-02 04:52:59.700764
679	1-679	4	94	8278	87	9987	1546	0	2012-08-02 04:44:54.249124	2012-08-02 04:52:59.700764
1208	1-1208	12	54	5611	64	4112	0	0	2012-08-02 04:49:04.078173	2012-08-02 04:53:01.214471
1655	1-1655	94	71	2220	57	8662	0	0	2012-08-02 04:52:56.011076	2012-08-02 04:53:02.562039
646	1-646	2	85	2804	37	8949	0	5	2012-08-02 04:44:46.952244	2012-08-02 04:53:02.562039
1306	1-1306	31	96	1757	64	4172	0	0	2012-08-02 04:49:59.112145	2012-08-02 04:53:03.478976
1645	1-1645	33	94	962	25	3855	0	0	2012-08-02 04:52:51.258471	2012-08-02 04:53:03.478976
877	1-877	86	43	1764	72	7131	0	0	2012-08-02 04:46:04.60876	2012-08-02 04:53:03.932227
1412	1-1412	73	100	6872	86	6173	0	0	2012-08-02 04:50:51.462164	2012-08-02 04:53:04.871429
1132	1-1132	46	58	7248	25	9359	0	0	2012-08-02 04:48:26.614867	2012-08-02 04:53:04.871429
1670	1-1670	75	80	1127	43	2343	0	0	2012-08-02 04:53:03.932227	2012-08-02 04:53:05.336402
1334	1-1334	22	73	229	40	5134	0	0	2012-08-02 04:50:12.672137	2012-08-02 04:53:09.876677
1683	1-1683	39	40	2048	65	6796	0	0	2012-08-02 04:53:09.876677	2012-08-02 04:53:09.876677
1378	1-1378	16	97	2763	55	4521	0	0	2012-08-02 04:50:35.200676	2012-08-02 04:53:13.080371
1155	1-1155	49	100	6281	59	2271	0	0	2012-08-02 04:48:36.814349	2012-08-02 04:53:13.080371
1663	1-1663	33	35	2936	45	9184	0	0	2012-08-02 04:52:59.700764	2012-08-02 04:53:13.080371
1423	1-1423	51	93	3661	80	4799	0	0	2012-08-02 04:50:57.472538	2012-08-02 04:53:13.544515
993	1-993	40	93	3052	42	590	0	0	2012-08-02 04:47:07.532103	2012-08-02 04:53:13.544515
1467	1-1467	8	33	8497	66	2552	0	0	2012-08-02 04:51:22.221065	2012-08-02 04:53:13.544515
1068	1-1068	99	58	7335	74	9752	0	5	2012-08-02 04:47:51.878081	2012-08-02 04:53:16.660226
1179	1-1179	33	69	3110	90	808	0	0	2012-08-02 04:48:47.454948	2012-08-02 04:53:16.660226
1699	1-1699	31	7	793	58	7717	0	0	2012-08-02 04:53:16.660226	2012-08-02 04:53:16.660226
1365	1-1365	69	2	8482	73	4646	0	0	2012-08-02 04:50:28.259954	2012-08-02 04:53:17.577085
1243	1-1243	43	15	2174	62	9004	0	0	2012-08-02 04:49:21.518282	2012-08-02 04:53:17.577085
1660	1-1660	48	53	4675	42	3013	0	0	2012-08-02 04:52:58.3418	2012-08-02 04:53:17.577085
1701	1-1701	54	62	3585	2	9433	0	0	2012-08-02 04:53:17.577085	2012-08-02 04:53:18.494568
1110	1-1110	13	62	7780	25	976	0	0	2012-08-02 04:48:13.04936	2012-08-02 04:53:18.494568
1672	1-1672	26	96	2683	82	4082	0	0	2012-08-02 04:53:04.871429	2012-08-02 04:53:18.958398
440	1-440	14	49	9783	94	5784	0	0	2012-08-02 04:44:06.289758	2012-08-02 04:53:18.958398
36	1-36	32	41	2879	51	3523	0	0	2012-08-02 04:43:53.844114	2012-08-02 04:53:19.433578
1352	1-1352	28	22	9949	5	7905	0	0	2012-08-02 04:50:21.974134	2012-08-02 04:53:19.433578
1461	1-1461	88	50	9466	21	9143	1037	0	2012-08-02 04:51:18.233482	2012-08-02 04:53:19.433578
1102	1-1102	67	68	9914	69	3892	0	0	2012-08-02 04:48:08.641428	2012-08-02 04:53:20.372732
1704	1-1704	16	68	2597	13	5756	0	0	2012-08-02 04:53:18.958398	2012-08-02 04:53:20.372732
242	1-242	16	93	8622	78	518	0	0	2012-08-02 04:43:56.936081	2012-08-02 04:53:20.836771
1450	1-1450	2	97	4179	33	7058	0	5	2012-08-02 04:51:12.278769	2012-08-02 04:53:21.279205
1703	1-1703	8	39	716	32	1330	0	7	2012-08-02 04:53:18.494568	2012-08-02 04:53:22.328328
913	1-913	23	71	8886	79	6895	0	0	2012-08-02 04:46:24.631439	2012-08-02 04:53:22.328328
856	1-856	47	53	2920	72	4344	0	0	2012-08-02 04:45:56.521443	2012-08-02 04:53:22.328328
1361	1-1361	33	100	3373	72	7368	0	0	2012-08-02 04:50:26.371057	2012-08-02 04:53:22.328328
1443	1-1443	54	20	2885	94	6616	0	0	2012-08-02 04:51:07.959583	2012-08-02 04:53:22.328328
754	1-754	5	94	7163	72	6880	0	0	2012-08-02 04:45:15.469642	2012-08-02 04:53:22.328328
992	1-992	17	19	4704	39	9507	1204	2	2012-08-02 04:47:07.06816	2012-08-02 04:53:22.328328
941	1-941	35	19	9629	11	5648	693	0	2012-08-02 04:46:37.247065	2012-08-02 04:53:22.328328
1550	1-1550	61	38	95	78	7239	0	0	2012-08-02 04:52:04.446242	2012-08-02 04:53:23.700336
984	1-984	44	16	1850	10	3050	0	0	2012-08-02 04:47:03.243594	2012-08-02 04:53:23.700336
1631	1-1631	51	38	600	61	9384	0	0	2012-08-02 04:52:44.972691	2012-08-02 04:53:23.700336
1714	1-1714	78	10	4026	38	5572	0	0	2012-08-02 04:53:23.700336	2012-08-02 04:53:23.700336
1558	1-1558	70	45	403	28	5295	0	0	2012-08-02 04:52:08.578907	2012-08-02 04:53:24.164796
1216	1-1216	13	15	3082	32	981	0	2	2012-08-02 04:49:07.719461	2012-08-02 04:53:24.164796
1646	1-1646	55	32	1346	62	5600	0	0	2012-08-02 04:52:51.722545	2012-08-02 04:53:25.070497
1393	1-1393	64	52	6075	46	8683	0	0	2012-08-02 04:50:41.575233	2012-08-02 04:53:25.070497
1233	1-1233	40	16	5265	82	5923	0	0	2012-08-02 04:49:15.707229	2012-08-02 04:53:28.385034
1534	1-1534	8	57	1631	37	7914	0	2	2012-08-02 04:51:55.298818	2012-08-02 04:53:28.385034
1184	1-1184	32	81	4147	68	4173	0	0	2012-08-02 04:48:50.366631	2012-08-02 04:53:28.385034
1340	1-1340	23	69	6518	42	8755	0	6	2012-08-02 04:50:16.527472	2012-08-02 04:53:28.385034
559	1-559	52	94	8750	83	432	366	0	2012-08-02 04:44:25.928786	2012-08-02 04:53:28.385034
1433	1-1433	66	55	655	4	3248	0	0	2012-08-02 04:51:02.178649	2012-08-02 04:53:28.848932
1659	1-1659	53	93	2775	29	739	0	0	2012-08-02 04:52:57.866925	2012-08-02 04:53:28.848932
1726	1-1726	4	4	1248	21	8245	0	0	2012-08-02 04:53:28.848932	2012-08-02 04:53:28.848932
1591	1-1591	75	90	6636	26	8271	0	0	2012-08-02 04:52:26.212314	2012-08-02 04:53:29.887489
1049	1-1049	15	90	7137	69	7047	0	0	2012-08-02 04:47:38.047949	2012-08-02 04:53:29.887489
1732	1-1732	69	50	950	2	2362	0	0	2012-08-02 04:53:31.842909	2012-08-02 04:53:31.842909
1639	1-1639	30	45	443	63	8876	0	0	2012-08-02 04:52:48.342372	2012-08-02 04:53:32.296277
1640	1-1640	32	3	2447	50	8765	0	0	2012-08-02 04:52:48.828259	2012-08-02 04:53:32.296277
1733	1-1733	34	5	2836	16	7868	0	0	2012-08-02 04:53:32.296277	2012-08-02 04:53:32.296277
1211	1-1211	61	3	9179	64	8106	799	6	2012-08-02 04:49:05.443119	2012-08-02 04:53:32.296277
1669	1-1669	13	13	8331	96	9375	388	0	2012-08-02 04:53:03.478976	2012-08-02 04:53:32.296277
1654	1-1654	76	65	1535	50	2661	0	0	2012-08-02 04:52:55.535688	2012-08-02 04:53:34.119066
1750	1-1750	41	70	381	42	2488	0	0	2012-08-02 04:53:39.598482	2012-08-02 04:53:39.598482
150	1-150	78	47	7357	78	6727	0	0	2012-08-02 04:43:55.21805	2012-08-02 04:53:40.051255
1305	1-1305	39	65	4956	73	5082	0	0	2012-08-02 04:49:57.8419	2012-08-02 04:53:40.979233
1673	1-1673	76	41	888	93	5448	0	0	2012-08-02 04:53:05.336402	2012-08-02 04:53:40.979233
1253	1-1253	93	49	7058	65	9427	0	0	2012-08-02 04:49:28.601027	2012-08-02 04:53:40.979233
1570	1-1570	23	33	4563	93	9336	0	0	2012-08-02 04:52:15.659819	2012-08-02 04:53:40.979233
1635	1-1635	67	5	7698	51	9074	0	0	2012-08-02 04:52:46.40883	2012-08-02 04:53:43.243987
326	1-326	95	13	7871	10	3909	0	0	2012-08-02 04:43:59.834488	2012-08-02 04:53:43.243987
1758	1-1758	8	32	790	63	9683	0	0	2012-08-02 04:53:43.243987	2012-08-02 04:53:43.243987
186	1-186	87	2	6020	13	1501	0	0	2012-08-02 04:43:55.743773	2012-08-02 04:53:44.161129
1762	1-1762	95	85	1735	91	2137	0	0	2012-08-02 04:53:45.05637	2012-08-02 04:53:45.05637
598	1-598	7	20	7441	4	4450	0	0	2012-08-02 04:44:36.396422	2012-08-02 04:53:45.951164
441	1-441	70	73	3915	22	5125	0	0	2012-08-02 04:44:06.304594	2012-08-02 04:53:45.951164
1764	1-1764	26	33	924	20	8436	0	0	2012-08-02 04:53:45.951164	2012-08-02 04:53:45.951164
1768	1-1768	30	13	2087	24	4105	0	0	2012-08-02 04:53:47.762743	2012-08-02 04:53:47.762743
1720	1-1720	53	100	2082	73	2672	0	0	2012-08-02 04:53:26.396183	2012-08-02 04:53:49.166402
1627	1-1627	69	94	4620	83	8797	0	2	2012-08-02 04:52:43.117058	2012-08-02 04:53:49.166402
1769	1-1769	60	100	1459	42	3968	0	0	2012-08-02 04:53:48.204735	2012-08-02 04:53:49.166402
555	1-555	59	54	8199	83	324	0	0	2012-08-02 04:44:24.536082	2012-08-02 04:53:49.166402
1771	1-1771	92	40	6198	100	6387	0	0	2012-08-02 04:53:49.166402	2012-08-02 04:53:49.166402
965	1-965	75	8	7516	54	8386	4284	0	2012-08-02 04:46:52.648668	2012-08-02 04:53:49.166402
1061	1-1061	65	23	8211	45	4025	0	0	2012-08-02 04:47:47.591873	2012-08-02 04:53:49.644863
1333	1-1333	1	25	6560	12	3162	0	0	2012-08-02 04:50:12.20827	2012-08-02 04:53:49.644863
1407	1-1407	85	13	4961	4	4417	0	0	2012-08-02 04:50:49.076068	2012-08-02 04:53:49.644863
1760	1-1760	49	13	1859	23	5532	0	0	2012-08-02 04:53:44.161129	2012-08-02 04:53:51.031792
1224	1-1224	90	57	8846	69	6872	0	0	2012-08-02 04:49:12.028082	2012-08-02 04:53:53.27504
1717	1-1717	34	62	2872	52	4084	0	0	2012-08-02 04:53:25.070497	2012-08-02 04:53:53.27504
89	1-89	51	83	6102	34	8813	0	0	2012-08-02 04:43:54.471342	2012-08-02 04:53:53.27504
1694	1-1694	32	11	2810	3	3313	0	0	2012-08-02 04:53:14.440932	2012-08-02 04:53:55.087733
755	1-755	11	16	9959	3	8411	0	0	2012-08-02 04:45:16.088607	2012-08-02 04:53:55.087733
1784	1-1784	54	3	6358	11	7780	0	0	2012-08-02 04:53:55.087733	2012-08-02 04:53:55.087733
1104	1-1104	65	95	5221	30	2556	0	0	2012-08-02 04:48:09.580371	2012-08-02 04:53:55.994277
1619	1-1619	91	16	5721	35	6158	2022	0	2012-08-02 04:52:39.31594	2012-08-02 04:53:55.994277
1735	1-1735	26	96	2207	39	4838	0	4	2012-08-02 04:53:33.22408	2012-08-02 04:53:56.889066
1788	1-1788	48	21	246	96	5832	0	0	2012-08-02 04:53:56.889066	2012-08-02 04:53:56.889066
1647	1-1647	75	25	8276	13	9315	0	0	2012-08-02 04:52:52.1866	2012-08-02 04:53:58.725
1594	1-1594	66	27	9254	2	8757	0	0	2012-08-02 04:52:27.869831	2012-08-02 04:54:01.51251
1536	1-1536	91	8	1555	86	6865	0	0	2012-08-02 04:51:56.259629	2012-08-02 04:54:01.51251
1653	1-1653	73	87	9914	8	2265	25	0	2012-08-02 04:52:55.071644	2012-08-02 04:54:01.51251
1799	1-1799	83	93	828	44	5035	0	0	2012-08-02 04:54:01.976318	2012-08-02 04:54:01.976318
1662	1-1662	40	17	6211	85	8341	0	0	2012-08-02 04:52:59.258824	2012-08-02 04:54:03.346444
1054	1-1054	99	39	8647	45	4957	0	2	2012-08-02 04:47:41.825874	2012-08-02 04:54:03.346444
1503	1-1503	1	26	2672	49	4902	0	0	2012-08-02 04:51:40.195564	2012-08-02 04:54:03.346444
1802	1-1802	3	85	8708	91	4124	0	2	2012-08-02 04:54:03.346444	2012-08-02 04:54:03.346444
1123	1-1123	20	67	5891	60	6400	0	0	2012-08-02 04:48:22.041527	2012-08-02 04:54:04.263953
581	1-581	13	35	2581	21	6784	0	0	2012-08-02 04:44:31.222754	2012-08-02 04:54:04.263953
1649	1-1649	68	17	4712	31	3475	0	0	2012-08-02 04:52:53.147923	2012-08-02 04:54:04.263953
1795	1-1795	14	45	3538	35	7724	0	0	2012-08-02 04:54:00.131377	2012-08-02 04:54:04.263953
46	1-46	9	60	5284	31	8228	1721	0	2012-08-02 04:43:53.955801	2012-08-02 04:54:04.263953
366	1-366	58	2	5825	47	4186	0	0	2012-08-02 04:44:01.532943	2012-08-02 04:54:05.674368
1643	1-1643	67	4	8547	33	7195	0	7	2012-08-02 04:52:50.275498	2012-08-02 04:54:05.674368
1440	1-1440	68	18	4529	4	9010	0	0	2012-08-02 04:51:05.969489	2012-08-02 04:54:05.674368
1052	1-1052	51	76	7066	6	8196	0	0	2012-08-02 04:47:40.887013	2012-08-02 04:54:07.010909
1702	1-1702	68	97	3280	17	9917	0	0	2012-08-02 04:53:18.052371	2012-08-02 04:54:07.010909
1564	1-1564	71	12	5443	77	8074	0	0	2012-08-02 04:52:12.666085	2012-08-02 04:54:07.010909
1751	1-1751	27	21	5344	67	2883	0	0	2012-08-02 04:53:40.051255	2012-08-02 04:54:07.475009
1422	1-1422	9	8	3756	97	7177	0	0	2012-08-02 04:50:56.698944	2012-08-02 04:54:08.425326
1813	1-1813	45	68	5995	8	4397	0	0	2012-08-02 04:54:08.425326	2012-08-02 04:54:08.425326
1446	1-1446	87	44	2749	65	4953	0	0	2012-08-02 04:51:09.85923	2012-08-02 04:54:10.236968
1127	1-1127	79	95	8621	86	4138	0	0	2012-08-02 04:48:24.284042	2012-08-02 04:54:13.45176
654	1-654	62	54	4883	28	959	0	6	2012-08-02 04:44:49.703341	2012-08-02 04:54:16.19132
1775	1-1775	59	2	6332	87	5829	0	0	2012-08-02 04:53:51.031792	2012-08-02 04:54:16.633417
1684	1-1684	46	18	5446	23	2523	0	0	2012-08-02 04:53:10.340951	2012-08-02 04:54:16.633417
1738	1-1738	21	18	3249	2	9908	0	0	2012-08-02 04:53:34.571873	2012-08-02 04:54:16.633417
1832	1-1832	72	41	3306	18	9732	0	0	2012-08-02 04:54:16.633417	2012-08-02 04:54:16.633417
1782	1-1782	7	43	4503	21	2191	0	13	2012-08-02 04:53:54.203947	2012-08-02 04:54:17.09766
1628	1-1628	92	83	514	15	783	0	0	2012-08-02 04:52:43.591665	2012-08-02 04:54:18.036602
1836	1-1836	15	34	196	89	2488	0	0	2012-08-02 04:54:18.478502	2012-08-02 04:54:18.478502
99	1-99	52	35	6865	11	4468	0	0	2012-08-02 04:43:54.593762	2012-08-02 04:54:18.942572
977	1-977	47	36	5865	64	4185	0	0	2012-08-02 04:46:59.321727	2012-08-02 04:54:18.942572
1786	1-1786	69	93	5254	95	9034	0	0	2012-08-02 04:53:55.994277	2012-08-02 04:54:18.942572
1753	1-1753	48	93	5818	49	6105	0	0	2012-08-02 04:53:40.979233	2012-08-02 04:54:18.942572
1808	1-1808	15	15	1702	3	902	0	0	2012-08-02 04:54:06.11622	2012-08-02 04:54:18.942572
1087	1-1087	29	10	3861	68	2836	0	0	2012-08-02 04:48:00.79561	2012-08-02 04:54:19.439924
1798	1-1798	95	86	1726	27	9631	0	0	2012-08-02 04:54:01.51251	2012-08-02 04:54:20.334705
1605	1-1605	44	72	7028	81	7403	0	0	2012-08-02 04:52:32.631773	2012-08-02 04:54:20.79888
1841	1-1841	1	9	2302	55	6045	0	0	2012-08-02 04:54:20.79888	2012-08-02 04:54:20.79888
1800	1-1800	42	25	157	40	1639	0	0	2012-08-02 04:54:02.4407	2012-08-02 04:54:21.682801
1667	1-1667	7	37	7571	52	4102	0	0	2012-08-02 04:53:02.562039	2012-08-02 04:54:21.682801
1840	1-1840	59	32	3742	86	3662	0	0	2012-08-02 04:54:20.334705	2012-08-02 04:54:23.47259
1597	1-1597	96	69	3991	1	2770	0	0	2012-08-02 04:52:28.830711	2012-08-02 04:54:24.378692
1664	1-1664	53	9	4173	54	6356	0	0	2012-08-02 04:53:01.214471	2012-08-02 04:54:25.750015
1059	1-1059	91	16	1632	29	3154	0	0	2012-08-02 04:47:45.93507	2012-08-02 04:54:25.750015
1530	1-1530	42	42	4260	85	2821	0	0	2012-08-02 04:51:53.862318	2012-08-02 04:54:27.109223
1420	1-1420	98	58	7742	98	7087	0	0	2012-08-02 04:50:55.770955	2012-08-02 04:54:27.109223
1847	1-1847	28	86	5125	58	8559	0	0	2012-08-02 04:54:23.47259	2012-08-02 04:54:27.109223
1737	1-1737	38	50	7183	13	4506	0	0	2012-08-02 04:53:34.119066	2012-08-02 04:54:29.882214
1810	1-1810	45	17	3990	73	5743	0	0	2012-08-02 04:54:07.010909	2012-08-02 04:54:30.765957
1725	1-1725	90	83	1921	57	6974	0	0	2012-08-02 04:53:28.385034	2012-08-02 04:54:31.22966
1711	1-1711	93	72	8860	19	9529	0	0	2012-08-02 04:53:22.328328	2012-08-02 04:54:31.22966
565	1-565	10	29	7292	28	2482	0	0	2012-08-02 04:44:27.763408	2012-08-02 04:54:33.052479
1872	1-1872	93	35	1625	53	2694	0	0	2012-08-02 04:54:34.455968	2012-08-02 04:54:35.825224
1875	1-1875	6	53	2080	35	4332	0	0	2012-08-02 04:54:35.825224	2012-08-02 04:54:35.825224
1441	1-1441	79	44	7295	39	8712	0	0	2012-08-02 04:51:06.433318	2012-08-02 04:54:37.592721
1527	1-1527	34	26	1615	60	1915	0	2	2012-08-02 04:51:52.48156	2012-08-02 04:54:37.592721
847	1-847	57	30	2392	21	1555	0	0	2012-08-02 04:45:52.498588	2012-08-02 04:54:37.592721
1416	1-1416	66	44	2454	19	4169	0	10	2012-08-02 04:50:53.351386	2012-08-02 04:54:37.592721
1856	1-1856	66	85	5014	42	9482	0	0	2012-08-02 04:54:27.109223	2012-08-02 04:54:37.592721
638	1-638	100	91	6675	44	9075	0	0	2012-08-02 04:44:45.658223	2012-08-02 04:54:37.592721
1828	1-1828	33	28	2286	60	1735	0	0	2012-08-02 04:54:14.821416	2012-08-02 04:54:37.592721
872	1-872	92	1	8603	56	6309	0	0	2012-08-02 04:46:02.851616	2012-08-02 04:54:37.592721
1837	1-1837	24	59	2739	35	4294	0	0	2012-08-02 04:54:18.942572	2012-08-02 04:54:38.07966
1734	1-1734	6	66	758	36	3982	0	0	2012-08-02 04:53:32.760352	2012-08-02 04:54:38.07966
1831	1-1831	88	81	1811	9	9871	0	0	2012-08-02 04:54:16.19132	2012-08-02 04:54:39.493989
1882	1-1882	59	13	222	73	8799	0	0	2012-08-02 04:54:39.493989	2012-08-02 04:54:39.493989
986	1-986	46	90	9610	71	4195	0	0	2012-08-02 04:47:04.657876	2012-08-02 04:54:39.957878
1212	1-1212	77	48	7128	34	3928	0	0	2012-08-02 04:49:05.918493	2012-08-02 04:54:40.918816
1807	1-1807	15	17	5944	18	6473	0	0	2012-08-02 04:54:05.674368	2012-08-02 04:54:41.382771
1601	1-1601	31	16	2756	43	6948	0	0	2012-08-02 04:52:30.741978	2012-08-02 04:54:42.288671
1905	1-1905	76	18	5112	59	5902	0	0	2012-08-02 04:54:49.183036	2012-08-02 04:54:49.183036
1617	1-1617	95	44	1645	40	4480	0	0	2012-08-02 04:52:38.366256	2012-08-02 04:54:50.099107
1907	1-1907	71	37	1584	44	7178	0	10	2012-08-02 04:54:50.099107	2012-08-02 04:54:50.099107
1870	1-1870	72	67	1460	10	8228	0	0	2012-08-02 04:54:33.505645	2012-08-02 04:54:50.552083
1908	1-1908	81	5	365	67	9931	0	0	2012-08-02 04:54:50.552083	2012-08-02 04:54:50.552083
855	1-855	24	10	9248	69	2374	0	0	2012-08-02 04:45:55.803377	2012-08-02 04:54:52.861141
1913	1-1913	86	27	31	10	8923	0	3	2012-08-02 04:54:52.861141	2012-08-02 04:54:52.861141
1709	1-1709	97	6	2808	97	2300	0	0	2012-08-02 04:53:21.279205	2012-08-02 04:54:53.314071
1707	1-1707	22	97	7092	68	7955	0	0	2012-08-02 04:53:20.372732	2012-08-02 04:54:53.314071
1691	1-1691	88	67	826	97	8945	467	0	2012-08-02 04:53:13.544515	2012-08-02 04:54:53.314071
1817	1-1817	64	65	1722	18	6797	0	0	2012-08-02 04:54:10.236968	2012-08-02 04:54:57.147574
1843	1-1843	44	40	7291	70	4909	0	0	2012-08-02 04:54:21.682801	2012-08-02 04:54:57.147574
330	1-330	64	89	8436	42	9050	0	0	2012-08-02 04:43:59.956301	2012-08-02 04:54:59.456325
1482	1-1482	61	42	8508	86	4157	0	0	2012-08-02 04:51:29.347326	2012-08-02 04:54:59.456325
1928	1-1928	48	93	724	61	6526	0	0	2012-08-02 04:54:59.456325	2012-08-02 04:54:59.456325
1362	1-1362	79	4	8719	17	7538	0	0	2012-08-02 04:50:26.846209	2012-08-02 04:55:01.756367
1933	1-1933	56	50	2838	87	5569	0	0	2012-08-02 04:55:01.756367	2012-08-02 04:55:01.756367
328	1-328	77	90	5945	34	7126	0	0	2012-08-02 04:43:59.867737	2012-08-02 04:55:02.629241
1878	1-1878	59	39	6841	91	9423	0	0	2012-08-02 04:54:37.592721	2012-08-02 04:55:04.430008
288	1-288	66	99	8522	75	1596	0	0	2012-08-02 04:43:58.655742	2012-08-02 04:55:04.893728
1940	1-1940	58	48	164	99	7034	0	0	2012-08-02 04:55:04.893728	2012-08-02 04:55:04.893728
1865	1-1865	42	57	2451	55	4049	0	0	2012-08-02 04:54:31.22966	2012-08-02 04:55:05.359321
1941	1-1941	22	45	1462	30	6186	0	0	2012-08-02 04:55:05.359321	2012-08-02 04:55:05.359321
1453	1-1453	63	69	5260	85	6917	0	0	2012-08-02 04:51:14.819205	2012-08-02 04:55:05.834241
1205	1-1205	51	69	5677	31	4511	0	0	2012-08-02 04:49:02.686118	2012-08-02 04:55:05.834241
1944	1-1944	33	52	1315	8	5115	0	0	2012-08-02 04:55:06.74008	2012-08-02 04:55:06.74008
1557	1-1557	93	39	1717	29	8926	0	7	2012-08-02 04:52:08.104012	2012-08-02 04:55:07.193176
1945	1-1945	85	40	1529	21	6568	0	2	2012-08-02 04:55:07.193176	2012-08-02 04:55:07.193176
1609	1-1609	58	32	1966	52	2134	0	0	2012-08-02 04:52:34.553938	2012-08-02 04:55:08.518804
1270	1-1270	36	32	2443	40	9834	0	0	2012-08-02 04:49:38.189871	2012-08-02 04:55:08.518804
1293	1-1293	52	92	8948	23	5077	0	0	2012-08-02 04:49:50.838502	2012-08-02 04:55:08.518804
1948	1-1948	37	70	427	32	7349	0	0	2012-08-02 04:55:08.518804	2012-08-02 04:55:08.518804
1088	1-1088	97	18	4148	70	4570	217	0	2012-08-02 04:48:01.270605	2012-08-02 04:55:08.518804
1949	1-1949	92	31	122	35	9753	0	0	2012-08-02 04:55:08.972212	2012-08-02 04:55:08.972212
1923	1-1923	48	70	2223	33	3512	0	0	2012-08-02 04:54:57.147574	2012-08-02 04:55:09.42556
1950	1-1950	100	76	2232	98	9520	0	0	2012-08-02 04:55:09.42556	2012-08-02 04:55:09.42556
1559	1-1559	60	91	2792	61	1332	0	0	2012-08-02 04:52:09.042999	2012-08-02 04:55:09.88921
1288	1-1288	1	94	4595	28	3417	0	0	2012-08-02 04:49:47.480226	2012-08-02 04:55:09.88921
1685	1-1685	70	91	6623	30	7975	0	0	2012-08-02 04:53:10.804928	2012-08-02 04:55:09.88921
438	1-438	75	30	8787	11	2878	1136	0	2012-08-02 04:44:06.223472	2012-08-02 04:55:09.88921
1789	1-1789	86	14	7912	25	7599	0	0	2012-08-02 04:53:57.353279	2012-08-02 04:55:10.817253
1525	1-1525	7	78	6939	28	8866	0	0	2012-08-02 04:51:51.597425	2012-08-02 04:55:11.248188
1489	1-1489	49	85	8999	9	6220	0	0	2012-08-02 04:51:34.373368	2012-08-02 04:55:11.723349
1885	1-1885	95	52	1733	48	5627	0	0	2012-08-02 04:54:40.918816	2012-08-02 04:55:13.524003
33	1-33	7	19	5692	47	4100	0	0	2012-08-02 04:43:53.810421	2012-08-02 04:55:13.965933
719	1-719	44	19	6274	71	4636	0	0	2012-08-02 04:45:04.095419	2012-08-02 04:55:13.965933
1961	1-1961	43	81	2536	19	9341	0	0	2012-08-02 04:55:13.965933	2012-08-02 04:55:13.965933
1602	1-1602	80	75	105	14	9293	0	0	2012-08-02 04:52:31.216982	2012-08-02 04:55:14.87216
1963	1-1963	7	14	4959	13	2108	0	0	2012-08-02 04:55:14.87216	2012-08-02 04:55:14.87216
393	1-393	33	10	7296	62	522	156	0	2012-08-02 04:44:03.348881	2012-08-02 04:55:14.87216
1862	1-1862	79	17	3195	50	7767	0	5	2012-08-02 04:54:29.882214	2012-08-02 04:55:15.766864
1387	1-1387	19	94	3170	6	5725	0	0	2012-08-02 04:50:39.09008	2012-08-02 04:55:19.279948
1973	1-1973	32	6	4096	45	3600	0	0	2012-08-02 04:55:19.279948	2012-08-02 04:55:19.279948
1977	1-1977	55	48	3898	91	8169	0	0	2012-08-02 04:55:21.025411	2012-08-02 04:55:21.025411
1563	1-1563	43	23	4381	18	6285	0	0	2012-08-02 04:52:12.202012	2012-08-02 04:55:21.467483
1978	1-1978	51	27	402	23	7667	0	0	2012-08-02 04:55:21.467483	2012-08-02 04:55:21.467483
1414	1-1414	99	61	2141	7	6357	0	0	2012-08-02 04:50:52.412428	2012-08-02 04:55:23.643904
1546	1-1546	63	44	6705	30	9736	0	0	2012-08-02 04:52:01.971665	2012-08-02 04:55:23.643904
1206	1-1206	53	46	6877	28	9038	0	0	2012-08-02 04:49:03.149979	2012-08-02 04:55:23.643904
1210	1-1210	90	35	3321	85	7115	0	0	2012-08-02 04:49:05.010962	2012-08-02 04:55:23.643904
915	1-915	63	91	5844	85	8727	110	0	2012-08-02 04:46:25.945961	2012-08-02 04:55:23.643904
773	1-773	8	15	6606	50	4232	0	0	2012-08-02 04:45:22.883	2012-08-02 04:55:25.87539
1080	1-1080	77	46	7752	68	4102	0	0	2012-08-02 04:47:57.746121	2012-08-02 04:55:25.87539
1852	1-1852	85	90	366	5	4230	0	0	2012-08-02 04:54:25.28644	2012-08-02 04:55:26.328478
1749	1-1749	14	38	133	45	4963	0	0	2012-08-02 04:53:39.575975	2012-08-02 04:55:28.970975
1488	1-1488	47	86	1376	28	1379	0	4	2012-08-02 04:51:33.898817	2012-08-02 04:55:28.970975
1995	1-1995	89	15	461	38	2618	0	0	2012-08-02 04:55:28.970975	2012-08-02 04:55:28.970975
642	1-642	83	71	7755	34	4150	0	0	2012-08-02 04:44:46.056452	2012-08-02 04:55:31.63312
1743	1-1743	56	47	1627	81	8581	0	0	2012-08-02 04:53:36.847455	2012-08-02 04:55:32.075248
1398	1-1398	94	42	7800	67	1542	0	0	2012-08-02 04:50:43.861966	2012-08-02 04:55:32.075248
1246	1-1246	22	43	8927	70	7044	0	2	2012-08-02 04:49:23.429083	2012-08-02 04:55:33.379039
2005	1-2005	34	70	3295	11	5803	0	0	2012-08-02 04:55:33.379039	2012-08-02 04:55:33.379039
30	1-30	28	11	8384	43	7316	332	0	2012-08-02 04:43:53.77712	2012-08-02 04:55:33.379039
1390	1-1390	66	46	3422	28	1366	329	0	2012-08-02 04:50:40.128131	2012-08-02 04:55:33.379039
1988	1-1988	2	66	925	98	4275	0	10	2012-08-02 04:55:25.87539	2012-08-02 04:55:33.823454
1705	1-1705	10	51	7447	49	6741	0	0	2012-08-02 04:53:19.433578	2012-08-02 04:55:33.823454
1833	1-1833	86	21	8256	60	9110	0	0	2012-08-02 04:54:17.09766	2012-08-02 04:55:33.823454
1438	1-1438	12	44	2413	89	5509	0	6	2012-08-02 04:51:04.542873	2012-08-02 04:55:34.251929
910	1-910	77	57	6561	4	5901	0	0	2012-08-02 04:46:22.830374	2012-08-02 04:55:34.251929
1165	1-1165	85	89	1669	4	3155	0	0	2012-08-02 04:48:40.736579	2012-08-02 04:55:34.251929
2008	1-2008	64	4	4670	57	7098	0	0	2012-08-02 04:55:34.251929	2012-08-02 04:55:34.251929
1983	1-1983	28	85	3662	61	4278	0	0	2012-08-02 04:55:23.643904	2012-08-02 04:55:36.439407
1531	1-1531	64	100	9762	87	6465	0	0	2012-08-02 04:51:54.315396	2012-08-02 04:55:36.439407
2007	1-2007	46	49	3208	66	5529	0	0	2012-08-02 04:55:33.823454	2012-08-02 04:55:38.637907
1874	1-1874	46	23	5717	7	8690	0	0	2012-08-02 04:54:35.361461	2012-08-02 04:55:38.637907
1496	1-1496	41	96	9434	49	9625	0	0	2012-08-02 04:51:37.67681	2012-08-02 04:55:38.637907
1728	1-1728	56	52	2203	90	8064	0	0	2012-08-02 04:53:29.887489	2012-08-02 04:55:39.080116
2019	1-2019	96	90	5888	61	4757	0	0	2012-08-02 04:55:39.080116	2012-08-02 04:55:39.080116
1745	1-1745	94	55	2793	12	8542	0	0	2012-08-02 04:53:37.731399	2012-08-02 04:55:39.510942
1119	1-1119	50	45	4950	62	8348	0	4	2012-08-02 04:48:19.80982	2012-08-02 04:55:39.510942
1466	1-1466	20	77	2769	76	4969	0	0	2012-08-02 04:51:21.724135	2012-08-02 04:55:39.510942
2020	1-2020	21	76	1766	55	6112	0	0	2012-08-02 04:55:39.510942	2012-08-02 04:55:39.510942
1954	1-1954	43	48	5093	7	6893	0	0	2012-08-02 04:55:11.248188	2012-08-02 04:55:43.068309
2006	1-2006	34	38	180	48	3017	0	0	2012-08-02 04:55:33.809807	2012-08-02 04:55:47.431847
503	1-503	100	27	6373	87	3329	0	0	2012-08-02 04:44:12.32043	2012-08-02 04:55:47.431847
1850	1-1850	10	38	1769	58	2980	0	4	2012-08-02 04:54:24.820175	2012-08-02 04:55:47.431847
2039	1-2039	10	36	35	8	3911	0	0	2012-08-02 04:55:47.874093	2012-08-02 04:55:47.874093
528	1-528	34	69	1228	34	164	0	0	2012-08-02 04:44:18.145328	2012-08-02 04:55:51.365128
1957	1-1957	14	58	4073	87	3872	0	0	2012-08-02 04:55:12.176387	2012-08-02 04:55:53.077493
1624	1-1624	9	57	3615	89	1760	0	0	2012-08-02 04:52:41.680093	2012-08-02 04:55:53.530161
1535	1-1535	28	57	1482	40	7784	0	0	2012-08-02 04:51:55.817982	2012-08-02 04:55:53.530161
1991	1-1991	94	67	839	57	7351	0	0	2012-08-02 04:55:27.212315	2012-08-02 04:55:53.530161
2018	1-2018	80	7	4154	96	5595	0	0	2012-08-02 04:55:38.637907	2012-08-02 04:55:54.40312
2059	1-2059	25	89	82	13	6925	0	0	2012-08-02 04:55:56.579365	2012-08-02 04:55:56.579365
2063	1-2063	63	74	4025	16	6091	0	0	2012-08-02 04:55:58.291796	2012-08-02 04:55:58.291796
1754	1-1754	90	23	4596	41	8119	0	0	2012-08-02 04:53:41.43251	2012-08-02 04:55:58.722601
1780	1-1780	18	1	1343	10	7110	0	0	2012-08-02 04:53:53.27504	2012-08-02 04:55:58.722601
2064	1-2064	4	41	6774	23	6256	0	0	2012-08-02 04:55:58.722601	2012-08-02 04:55:58.722601
1942	1-1942	30	4	3555	69	6067	0	0	2012-08-02 04:55:05.834241	2012-08-02 04:55:59.5623
1259	1-1259	89	10	9380	62	9853	0	3	2012-08-02 04:49:32.47805	2012-08-02 04:55:59.5623
1967	1-1967	92	62	9103	19	7298	0	0	2012-08-02 04:55:16.639999	2012-08-02 04:55:59.5623
2002	1-2002	85	81	9558	44	8700	0	0	2012-08-02 04:55:32.075248	2012-08-02 04:56:00.435403
1105	1-1105	47	45	5974	48	4126	0	0	2012-08-02 04:48:10.044592	2012-08-02 04:56:01.331903
1648	1-1648	63	11	3194	52	9033	0	0	2012-08-02 04:52:52.650831	2012-08-02 04:56:05.199571
2058	1-2058	63	43	1027	69	2782	0	0	2012-08-02 04:55:56.159627	2012-08-02 04:56:07.354419
1526	1-1526	36	43	4454	21	1050	0	11	2012-08-02 04:51:52.028393	2012-08-02 04:56:07.354419
898	1-898	34	65	7845	43	7816	0	0	2012-08-02 04:46:15.749553	2012-08-02 04:56:07.354419
1448	1-1448	38	70	4080	84	1056	0	0	2012-08-02 04:51:11.339935	2012-08-02 04:56:07.354419
1804	1-1804	23	31	8766	45	6699	0	0	2012-08-02 04:54:04.263953	2012-08-02 04:56:07.354419
1757	1-1757	32	8	1178	87	1503	0	0	2012-08-02 04:53:42.791152	2012-08-02 04:56:08.249672
2088	1-2088	59	74	2622	30	8776	0	4	2012-08-02 04:56:08.702359	2012-08-02 04:56:08.702359
1801	1-1801	15	55	6082	45	5732	0	0	2012-08-02 04:54:02.882565	2012-08-02 04:56:09.133452
1410	1-1410	65	94	9741	57	2045	0	0	2012-08-02 04:50:50.512175	2012-08-02 04:56:09.133452
2035	1-2035	43	49	244	20	8463	0	6	2012-08-02 04:55:46.15048	2012-08-02 04:56:09.133452
733	1-733	3	38	4502	44	8034	0	0	2012-08-02 04:45:07.902225	2012-08-02 04:56:09.575444
1971	1-1971	56	84	415	15	2333	0	0	2012-08-02 04:55:18.396224	2012-08-02 04:56:10.923331
1255	1-1255	54	59	8127	77	6156	0	0	2012-08-02 04:49:30.534183	2012-08-02 04:56:11.354042
1397	1-1397	21	59	5600	65	2046	0	0	2012-08-02 04:50:43.397972	2012-08-02 04:56:11.354042
1993	1-1993	71	80	5924	11	1907	0	0	2012-08-02 04:55:28.096074	2012-08-02 04:56:11.354042
2094	1-2094	77	63	233	59	7725	0	0	2012-08-02 04:56:11.354042	2012-08-02 04:56:11.354042
1951	1-1951	65	28	8839	91	9119	0	0	2012-08-02 04:55:09.88921	2012-08-02 04:56:12.226952
1853	1-1853	20	29	4746	32	5317	0	0	2012-08-02 04:54:25.750015	2012-08-02 04:56:12.668953
2030	1-2030	26	23	3650	1	2567	0	0	2012-08-02 04:55:43.941136	2012-08-02 04:56:12.668953
1370	1-1370	74	2	4575	43	3072	0	0	2012-08-02 04:50:31.420258	2012-08-02 04:56:13.111435
2098	1-2098	15	28	45	2	6166	0	0	2012-08-02 04:56:13.111435	2012-08-02 04:56:13.111435
592	1-592	86	39	9029	66	2567	0	0	2012-08-02 04:44:34.727913	2012-08-02 04:56:13.553003
2001	1-2001	85	56	892	71	4189	0	0	2012-08-02 04:55:31.63312	2012-08-02 04:56:13.553003
1618	1-1618	39	43	3052	66	551	0	0	2012-08-02 04:52:38.852086	2012-08-02 04:56:13.553003
2097	1-2097	17	93	4915	23	9966	0	0	2012-08-02 04:56:12.668953	2012-08-02 04:56:13.553003
463	1-463	17	94	8524	93	6995	1471	0	2012-08-02 04:44:07.983561	2012-08-02 04:56:13.553003
1480	1-1480	58	31	5189	37	5038	0	7	2012-08-02 04:51:28.419336	2012-08-02 04:56:14.006213
1822	1-1822	15	67	1583	86	2678	0	0	2012-08-02 04:54:12.093073	2012-08-02 04:56:14.006213
2099	1-2099	69	23	7734	96	9977	0	0	2012-08-02 04:56:13.553003	2012-08-02 04:56:17.110722
2100	1-2100	47	93	1676	31	7750	0	0	2012-08-02 04:56:14.006213	2012-08-02 04:56:17.541405
1427	1-1427	56	65	5691	31	7889	0	0	2012-08-02 04:50:59.372268	2012-08-02 04:56:17.541405
2108	1-2108	8	88	476	67	4171	0	0	2012-08-02 04:56:17.541405	2012-08-02 04:56:17.541405
2114	1-2114	47	67	184	69	5804	0	0	2012-08-02 04:56:20.193022	2012-08-02 04:56:20.193022
2084	1-2084	24	17	6050	65	7506	0	0	2012-08-02 04:56:07.354419	2012-08-02 04:56:22.379957
1772	1-1772	33	48	5365	25	9408	0	0	2012-08-02 04:53:49.644863	2012-08-02 04:56:22.379957
2037	1-2037	26	48	3030	82	4631	0	0	2012-08-02 04:55:47.00106	2012-08-02 04:56:22.379957
1094	1-1094	76	2	6741	53	8990	0	0	2012-08-02 04:48:04.231835	2012-08-02 04:56:25.009285
2126	1-2126	83	40	881	99	4722	0	0	2012-08-02 04:56:25.440481	2012-08-02 04:56:25.440481
2013	1-2013	99	87	7781	23	8587	0	0	2012-08-02 04:55:36.439407	2012-08-02 04:56:27.20812
2130	1-2130	24	85	2019	87	4557	0	0	2012-08-02 04:56:27.20812	2012-08-02 04:56:27.20812
1606	1-1606	92	70	7316	89	1324	0	0	2012-08-02 04:52:33.095939	2012-08-02 04:56:27.650067
2131	1-2131	50	89	1132	70	9878	0	0	2012-08-02 04:56:27.650067	2012-08-02 04:56:27.650067
1965	1-1965	55	48	1011	40	8191	0	0	2012-08-02 04:55:15.766864	2012-08-02 04:56:30.281322
2138	1-2138	84	50	1969	91	4186	0	0	2012-08-02 04:56:30.281322	2012-08-02 04:56:30.281322
2139	1-2139	45	96	186	95	3886	0	0	2012-08-02 04:56:30.734157	2012-08-02 04:56:30.734157
1186	1-1186	36	63	7341	50	8883	0	2	2012-08-02 04:48:51.294384	2012-08-02 04:56:31.617899
2107	1-2107	7	50	3433	23	7805	0	0	2012-08-02 04:56:17.110722	2012-08-02 04:56:31.617899
2028	1-2028	59	95	6103	63	9446	0	0	2012-08-02 04:55:43.068309	2012-08-02 04:56:31.617899
1565	1-1565	92	74	6036	7	563	0	0	2012-08-02 04:52:13.152341	2012-08-02 04:56:31.617899
1042	1-1042	81	92	4476	17	6933	0	0	2012-08-02 04:47:32.91116	2012-08-02 04:56:32.059989
2142	1-2142	11	17	3048	32	8251	0	0	2012-08-02 04:56:32.059989	2012-08-02 04:56:32.059989
1886	1-1886	37	18	2437	4	4729	0	4	2012-08-02 04:54:41.382771	2012-08-02 04:56:32.502071
1690	1-1690	66	79	7152	8	5113	0	0	2012-08-02 04:53:13.080371	2012-08-02 04:56:32.502071
2143	1-2143	28	4	6188	79	8937	0	0	2012-08-02 04:56:32.502071	2012-08-02 04:56:32.502071
2042	1-2042	14	98	489	70	532	0	0	2012-08-02 04:55:49.188632	2012-08-02 04:56:32.944098
1449	1-1449	53	98	6427	46	7719	0	0	2012-08-02 04:51:11.803651	2012-08-02 04:56:32.944098
631	1-631	54	98	6630	22	5044	0	0	2012-08-02 04:44:43.293337	2012-08-02 04:56:32.944098
1835	1-1835	96	72	2884	83	2085	0	0	2012-08-02 04:54:18.036602	2012-08-02 04:56:35.584555
1793	1-1793	90	20	8971	94	8463	0	3	2012-08-02 04:53:59.214824	2012-08-02 04:56:36.490255
2140	1-2140	8	22	6257	55	8532	0	0	2012-08-02 04:56:31.176005	2012-08-02 04:56:37.352265
1250	1-1250	96	61	1368	6	6271	0	0	2012-08-02 04:49:26.201558	2012-08-02 04:56:37.352265
1509	1-1509	77	61	5910	11	3365	0	0	2012-08-02 04:51:43.167211	2012-08-02 04:56:37.352265
1017	1-1017	21	44	5925	26	1523	0	0	2012-08-02 04:47:18.968651	2012-08-02 04:56:37.352265
2154	1-2154	54	34	2401	1	5523	0	0	2012-08-02 04:56:37.352265	2012-08-02 04:56:37.352265
2072	1-2072	43	66	1649	86	5985	0	3	2012-08-02 04:56:02.182649	2012-08-02 04:56:39.098049
1321	1-1321	99	66	6231	21	8256	0	0	2012-08-02 04:50:07.445654	2012-08-02 04:56:39.098049
2149	1-2149	57	1	1666	61	3587	0	0	2012-08-02 04:56:35.14286	2012-08-02 04:56:39.098049
2150	1-2150	91	56	320	72	9905	0	0	2012-08-02 04:56:35.584555	2012-08-02 04:56:40.876702
2089	1-2089	11	73	4229	49	6027	0	0	2012-08-02 04:56:09.133452	2012-08-02 04:56:41.749285
1143	1-1143	42	42	6491	80	9605	0	0	2012-08-02 04:48:31.997417	2012-08-02 04:56:42.191149
1909	1-1909	59	47	2450	74	9499	0	0	2012-08-02 04:54:51.027276	2012-08-02 04:56:42.191149
2165	1-2165	55	74	2458	42	5240	0	0	2012-08-02 04:56:42.191149	2012-08-02 04:56:42.191149
1990	1-1990	24	97	5605	49	5427	0	0	2012-08-02 04:55:26.770551	2012-08-02 04:56:43.053002
1156	1-1156	27	83	8038	58	4332	0	0	2012-08-02 04:48:37.013151	2012-08-02 04:56:43.903774
2169	1-2169	61	50	1361	31	9899	0	0	2012-08-02 04:56:43.903774	2012-08-02 04:56:43.903774
1776	1-1776	99	31	9865	3	3740	0	0	2012-08-02 04:53:51.485487	2012-08-02 04:56:44.765816
2176	1-2176	83	79	1568	35	6429	0	0	2012-08-02 04:56:46.964093	2012-08-02 04:56:46.964093
1607	1-1607	38	96	6528	39	7474	0	0	2012-08-02 04:52:33.570252	2012-08-02 04:56:48.278514
960	1-960	26	96	7074	56	4656	0	0	2012-08-02 04:46:49.898193	2012-08-02 04:56:48.278514
22	1-22	22	8	3326	33	7946	0	0	2012-08-02 04:43:53.687681	2012-08-02 04:56:48.278514
1578	1-1578	6	22	7633	69	8715	0	0	2012-08-02 04:52:20.465075	2012-08-02 04:56:49.162466
1830	1-1830	92	78	3974	95	1552	0	0	2012-08-02 04:54:15.760769	2012-08-02 04:56:49.162466
2096	1-2096	56	67	3890	63	3498	0	0	2012-08-02 04:56:12.226952	2012-08-02 04:56:49.604397
1890	1-1890	62	58	3173	29	8288	0	0	2012-08-02 04:54:43.183436	2012-08-02 04:56:50.046487
2010	1-2010	56	56	2006	25	3993	0	0	2012-08-02 04:55:35.146862	2012-08-02 04:56:50.046487
2047	1-2047	17	48	3601	18	8654	0	0	2012-08-02 04:55:51.365128	2012-08-02 04:56:51.79211
2120	1-2120	32	13	2105	41	4163	0	2	2012-08-02 04:56:22.822179	2012-08-02 04:56:52.233924
2179	1-2179	53	9	2624	96	6534	0	0	2012-08-02 04:56:48.278514	2012-08-02 04:56:53.095975
1823	1-1823	36	29	2961	41	4663	0	0	2012-08-02 04:54:12.568284	2012-08-02 04:56:53.095975
2190	1-2190	38	43	1738	82	6051	0	0	2012-08-02 04:56:53.095975	2012-08-02 04:56:53.095975
729	1-729	83	33	9701	9	8921	1942	0	2012-08-02 04:45:07.160484	2012-08-02 04:56:53.095975
2092	1-2092	74	46	688	82	1009	0	0	2012-08-02 04:56:10.470148	2012-08-02 04:56:53.957867
2183	1-2183	94	25	3974	60	3646	0	0	2012-08-02 04:56:50.046487	2012-08-02 04:56:53.957867
912	1-912	59	83	9812	46	7526	350	0	2012-08-02 04:46:23.725803	2012-08-02 04:56:53.957867
2197	1-2197	77	88	2399	85	2282	0	0	2012-08-02 04:56:56.134017	2012-08-02 04:56:56.134017
1371	1-1371	12	55	9277	89	7038	0	0	2012-08-02 04:50:31.894938	2012-08-02 04:56:56.564843
2016	1-2016	73	42	6164	33	6524	0	0	2012-08-02 04:55:37.754481	2012-08-02 04:56:57.868862
2186	1-2186	76	44	980	53	1502	0	2	2012-08-02 04:56:51.339281	2012-08-02 04:56:59.594059
1919	1-1919	56	38	4641	67	7761	577	0	2012-08-02 04:54:55.755842	2012-08-02 04:56:59.594059
2209	1-2209	21	17	2050	44	1965	0	0	2012-08-02 04:57:01.372388	2012-08-02 04:57:01.372388
1023	1-1023	3	55	7891	82	1028	277	0	2012-08-02 04:47:21.906704	2012-08-02 04:57:01.372388
2024	1-2024	34	96	2885	43	2626	0	0	2012-08-02 04:55:41.312086	2012-08-02 04:57:01.814577
221	1-221	5	91	9844	63	9053	0	0	2012-08-02 04:43:56.400891	2012-08-02 04:57:01.814577
2210	1-2210	21	70	4592	91	5558	0	0	2012-08-02 04:57:01.814577	2012-08-02 04:57:03.968706
873	1-873	70	89	7347	27	7120	0	0	2012-08-02 04:46:02.951538	2012-08-02 04:57:07.018686
2119	1-2119	75	31	6399	48	5435	0	0	2012-08-02 04:56:22.379957	2012-08-02 04:57:07.018686
1638	1-1638	91	26	5076	79	4780	0	0	2012-08-02 04:52:47.856362	2012-08-02 04:57:07.018686
1050	1-1050	78	26	6858	34	9413	0	3	2012-08-02 04:47:39.428623	2012-08-02 04:57:07.018686
2158	1-2158	16	61	929	66	8589	0	0	2012-08-02 04:56:39.098049	2012-08-02 04:57:07.018686
2052	1-2052	2	34	8077	67	6615	0	0	2012-08-02 04:55:53.530161	2012-08-02 04:57:07.483176
2223	1-2223	4	94	1593	3	3493	0	0	2012-08-02 04:57:07.483176	2012-08-02 04:57:07.483176
2234	1-2234	69	15	449	52	5666	0	0	2012-08-02 04:57:12.211161	2012-08-02 04:57:12.211161
1779	1-1779	85	78	6514	56	2672	0	4	2012-08-02 04:53:52.832768	2012-08-02 04:57:12.6536
2181	1-2181	47	65	1777	22	7204	0	4	2012-08-02 04:56:49.162466	2012-08-02 04:57:12.6536
1575	1-1575	69	65	9136	57	5259	0	0	2012-08-02 04:52:19.062209	2012-08-02 04:57:12.6536
1827	1-1827	63	32	4943	6	6119	0	0	2012-08-02 04:54:14.379668	2012-08-02 04:57:12.6536
2022	1-2022	77	22	6305	78	7048	0	0	2012-08-02 04:55:40.417217	2012-08-02 04:57:12.6536
1260	1-1260	33	32	5882	28	838	0	0	2012-08-02 04:49:32.953233	2012-08-02 04:57:12.6536
1976	1-1976	95	4	3112	32	2530	0	0	2012-08-02 04:55:20.583657	2012-08-02 04:57:12.6536
579	1-579	89	15	9267	34	2534	0	0	2012-08-02 04:44:30.880226	2012-08-02 04:57:13.968161
2235	1-2235	49	15	2613	65	9947	0	0	2012-08-02 04:57:12.6536	2012-08-02 04:57:13.968161
2093	1-2093	78	67	3472	30	4683	0	0	2012-08-02 04:56:10.923331	2012-08-02 04:57:14.399193
307	1-307	41	16	2004	31	935	0	0	2012-08-02 04:43:59.13514	2012-08-02 04:57:14.399193
739	1-739	20	24	1970	17	2919	0	0	2012-08-02 04:45:09.580853	2012-08-02 04:57:14.399193
2066	1-2066	9	19	3724	64	6746	0	0	2012-08-02 04:55:59.5623	2012-08-02 04:57:14.841122
1899	1-1899	45	19	4678	46	2947	0	0	2012-08-02 04:54:47.337115	2012-08-02 04:57:14.841122
1943	1-1943	63	19	253	37	344	0	0	2012-08-02 04:55:06.287298	2012-08-02 04:57:14.841122
2240	1-2240	67	27	4252	19	9979	0	0	2012-08-02 04:57:14.841122	2012-08-02 04:57:14.841122
2069	1-2069	47	63	7693	28	7027	0	5	2012-08-02 04:56:00.888385	2012-08-02 04:57:16.12293
1893	1-1893	27	40	8868	92	2008	0	0	2012-08-02 04:54:44.553308	2012-08-02 04:57:16.996004
2070	1-2070	44	25	8245	31	9110	0	0	2012-08-02 04:56:01.331903	2012-08-02 04:57:17.426569
2056	1-2056	18	49	8158	28	9638	0	0	2012-08-02 04:55:55.309112	2012-08-02 04:57:18.288272
860	1-860	42	36	6797	34	4765	0	0	2012-08-02 04:45:57.913339	2012-08-02 04:57:19.183123
2038	1-2038	85	87	7101	38	5747	0	5	2012-08-02 04:55:47.431847	2012-08-02 04:57:19.183123
2230	1-2230	30	87	879	22	3374	0	0	2012-08-02 04:57:10.477081	2012-08-02 04:57:19.183123
2250	1-2250	2	32	457	36	6858	0	9	2012-08-02 04:57:19.183123	2012-08-02 04:57:19.183123
2156	1-2156	89	66	675	88	5451	0	0	2012-08-02 04:56:38.247105	2012-08-02 04:57:22.696334
2087	1-2087	87	44	3114	60	3652	0	3	2012-08-02 04:56:08.249672	2012-08-02 04:57:23.127743
1834	1-1834	64	22	5265	44	2030	0	0	2012-08-02 04:54:17.583679	2012-08-02 04:57:23.127743
1036	1-1036	91	38	1998	81	9183	0	0	2012-08-02 04:47:29.464544	2012-08-02 04:57:23.558207
954	1-954	90	89	6357	34	4380	0	0	2012-08-02 04:46:46.029021	2012-08-02 04:57:23.558207
2260	1-2260	86	96	1055	38	4504	0	0	2012-08-02 04:57:23.558207	2012-08-02 04:57:23.558207
630	1-630	66	3	8604	34	6604	1481	0	2012-08-02 04:44:42.962218	2012-08-02 04:57:23.558207
2264	1-2264	2	14	2872	32	3927	0	0	2012-08-02 04:57:25.303843	2012-08-02 04:57:25.303843
2265	1-2265	95	48	1645	84	9785	0	0	2012-08-02 04:57:25.723937	2012-08-02 04:57:25.723937
2071	1-2071	31	45	5722	73	9779	0	0	2012-08-02 04:56:01.751885	2012-08-02 04:57:27.027656
2141	1-2141	22	85	8000	95	7241	0	0	2012-08-02 04:56:31.617899	2012-08-02 04:57:28.365897
2125	1-2125	24	88	928	2	7446	0	0	2012-08-02 04:56:25.009285	2012-08-02 04:57:28.365897
1196	1-1196	91	82	984	74	2593	0	0	2012-08-02 04:48:56.864698	2012-08-02 04:57:28.365897
2271	1-2271	37	74	1723	85	4715	0	0	2012-08-02 04:57:28.365897	2012-08-02 04:57:28.365897
801	1-801	73	94	3739	81	2713	0	0	2012-08-02 04:45:33.150063	2012-08-02 04:57:29.238827
1014	1-1014	88	60	9781	97	4690	0	0	2012-08-02 04:47:18.283021	2012-08-02 04:57:29.238827
2222	1-2222	56	60	3155	26	6556	0	0	2012-08-02 04:57:07.018686	2012-08-02 04:57:29.238827
2238	1-2238	67	96	2603	15	9757	0	0	2012-08-02 04:57:13.968161	2012-08-02 04:57:29.238827
792	1-792	78	77	4598	24	1578	0	0	2012-08-02 04:45:28.774636	2012-08-02 04:57:29.238827
866	1-866	68	98	6379	81	4865	0	0	2012-08-02 04:46:00.134191	2012-08-02 04:57:29.238827
1388	1-1388	100	57	9180	81	8833	0	0	2012-08-02 04:50:39.609175	2012-08-02 04:57:29.238827
2051	1-2051	79	44	8591	60	9807	2178	0	2012-08-02 04:55:53.077493	2012-08-02 04:57:29.238827
1939	1-1939	65	84	3383	94	7616	0	0	2012-08-02 04:55:04.430008	2012-08-02 04:57:30.586939
2017	1-2017	82	82	1509	17	9628	0	0	2012-08-02 04:55:38.196241	2012-08-02 04:57:31.006997
2247	1-2247	83	58	4451	75	5397	0	4	2012-08-02 04:57:17.846426	2012-08-02 04:57:31.448747
2278	1-2278	89	57	2733	21	7660	0	0	2012-08-02 04:57:31.448747	2012-08-02 04:57:31.448747
1792	1-1792	84	54	4363	14	8765	0	0	2012-08-02 04:53:58.725	2012-08-02 04:57:31.890956
1715	1-1715	76	28	2867	15	8056	0	0	2012-08-02 04:53:24.164796	2012-08-02 04:57:31.890956
2279	1-2279	34	17	737	59	4998	0	6	2012-08-02 04:57:31.890956	2012-08-02 04:57:31.890956
1761	1-1761	74	10	1481	5	2928	0	0	2012-08-02 04:53:44.614107	2012-08-02 04:57:32.344008
2280	1-2280	54	5	935	30	3473	0	5	2012-08-02 04:57:32.344008	2012-08-02 04:57:32.344008
1731	1-1731	96	79	5012	18	5349	0	0	2012-08-02 04:53:31.412013	2012-08-02 04:57:35.791022
2273	1-2273	84	81	4076	84	5387	0	0	2012-08-02 04:57:29.238827	2012-08-02 04:57:35.791022
2145	1-2145	90	1	3514	22	9432	0	0	2012-08-02 04:56:33.397231	2012-08-02 04:57:36.641375
1966	1-1966	25	50	3580	35	1721	0	0	2012-08-02 04:55:16.208952	2012-08-02 04:57:36.641375
1809	1-1809	77	79	746	93	832	0	0	2012-08-02 04:54:06.569314	2012-08-02 04:57:37.945181
2011	1-2011	65	87	4087	92	5849	0	4	2012-08-02 04:55:35.577724	2012-08-02 04:57:37.945181
920	1-920	60	75	2515	91	671	0	0	2012-08-02 04:46:28.155462	2012-08-02 04:57:38.365076
2294	1-2294	15	26	148	11	7771	0	11	2012-08-02 04:57:38.365076	2012-08-02 04:57:38.365076
2295	1-2295	61	19	220	22	4948	0	0	2012-08-02 04:57:38.818182	2012-08-02 04:57:38.818182
1929	1-1929	40	99	7935	80	9100	0	0	2012-08-02 04:54:59.920778	2012-08-02 04:57:40.110894
1258	1-1258	100	95	4837	88	7709	0	0	2012-08-02 04:49:31.992079	2012-08-02 04:57:40.552744
42	1-42	34	57	8839	47	7555	0	0	2012-08-02 04:43:53.911096	2012-08-02 04:57:40.552744
2300	1-2300	17	40	1664	76	2476	0	0	2012-08-02 04:57:40.994612	2012-08-02 04:57:40.994612
2302	1-2302	67	39	55	49	9319	0	0	2012-08-02 04:57:41.834418	2012-08-02 04:57:41.834418
2293	1-2293	68	93	5425	87	4030	0	0	2012-08-02 04:57:37.945181	2012-08-02 04:57:42.265417
2171	1-2171	10	34	1978	96	6935	0	0	2012-08-02 04:56:44.765816	2012-08-02 04:57:42.265417
980	1-980	82	34	9708	22	6959	0	0	2012-08-02 04:47:01.718771	2012-08-02 04:57:42.265417
2303	1-2303	56	87	600	34	9915	0	0	2012-08-02 04:57:42.265417	2012-08-02 04:57:42.265417
2304	1-2304	57	42	932	15	5736	0	0	2012-08-02 04:57:42.718546	2012-08-02 04:57:42.718546
2187	1-2187	5	18	5076	63	3552	0	0	2012-08-02 04:56:51.79211	2012-08-02 04:57:43.149277
940	1-940	32	95	7748	41	5009	0	0	2012-08-02 04:46:36.750393	2012-08-02 04:57:43.149277
2044	1-2044	35	27	6671	60	4821	0	2	2012-08-02 04:55:50.050529	2012-08-02 04:57:43.149277
2103	1-2103	46	47	4469	78	8322	0	10	2012-08-02 04:56:15.354279	2012-08-02 04:57:43.149277
2305	1-2305	25	63	2451	47	9053	0	0	2012-08-02 04:57:43.149277	2012-08-02 04:57:43.149277
2162	1-2162	72	72	9349	59	1929	0	0	2012-08-02 04:56:40.876702	2012-08-02 04:57:44.884357
1351	1-1351	99	95	9531	21	4051	0	0	2012-08-02 04:50:21.487851	2012-08-02 04:57:44.884357
2309	1-2309	58	18	765	95	9183	0	0	2012-08-02 04:57:44.884357	2012-08-02 04:57:44.884357
2311	1-2311	57	89	5598	66	9741	0	0	2012-08-02 04:57:45.77921	2012-08-02 04:57:45.77921
2299	1-2299	86	88	6257	95	9096	0	0	2012-08-02 04:57:40.552744	2012-08-02 04:57:46.198644
2188	1-2188	53	41	8120	38	3665	0	0	2012-08-02 04:56:52.233924	2012-08-02 04:57:46.198644
2312	1-2312	52	72	256	41	2722	0	0	2012-08-02 04:57:46.198644	2012-08-02 04:57:46.198644
1451	1-1451	23	66	8810	89	7641	0	0	2012-08-02 04:51:12.76469	2012-08-02 04:57:47.922079
2298	1-2298	12	80	4449	33	3436	0	0	2012-08-02 04:57:40.110894	2012-08-02 04:57:47.922079
2258	1-2258	63	92	6900	66	8916	0	0	2012-08-02 04:57:22.696334	2012-08-02 04:57:47.922079
2313	1-2313	3	35	576	45	1267	0	6	2012-08-02 04:57:46.640841	2012-08-02 04:57:50.551534
1637	1-1637	88	64	5795	9	4901	0	0	2012-08-02 04:52:47.336755	2012-08-02 04:57:50.993558
2079	1-2079	61	52	4618	31	5215	0	7	2012-08-02 04:56:05.199571	2012-08-02 04:57:50.993558
2323	1-2323	60	31	954	64	7509	0	0	2012-08-02 04:57:50.993558	2012-08-02 04:57:50.993558
2324	1-2324	28	2	2768	81	1380	0	0	2012-08-02 04:57:51.446642	2012-08-02 04:57:51.446642
702	1-702	34	8	6230	11	4426	0	0	2012-08-02 04:44:59.266568	2012-08-02 04:57:52.761319
2192	1-2192	83	17	7308	25	9524	0	0	2012-08-02 04:56:53.957867	2012-08-02 04:57:53.214602
2328	1-2328	90	47	251	17	8921	0	0	2012-08-02 04:57:53.214602	2012-08-02 04:57:53.214602
2164	1-2164	34	28	9544	45	5998	0	0	2012-08-02 04:56:41.749285	2012-08-02 04:57:53.656472
2257	1-2257	87	95	1471	72	8299	0	0	2012-08-02 04:57:22.265494	2012-08-02 04:57:53.656472
1811	1-1811	6	54	5155	16	6156	0	0	2012-08-02 04:54:07.475009	2012-08-02 04:57:54.087211
2330	1-2330	98	90	2345	57	5222	0	0	2012-08-02 04:57:54.087211	2012-08-02 04:57:54.087211
2288	1-2288	80	33	5121	99	9262	0	0	2012-08-02 04:57:35.791022	2012-08-02 04:57:54.518266
2191	1-2191	64	89	4053	37	4765	0	0	2012-08-02 04:56:53.538271	2012-08-02 04:57:54.518266
1483	1-1483	11	8	1793	100	8625	0	0	2012-08-02 04:51:29.833706	2012-08-02 04:57:55.834741
2239	1-2239	59	31	3120	67	3762	0	0	2012-08-02 04:57:14.399193	2012-08-02 04:57:56.276795
1465	1-1465	24	20	6193	100	6520	0	0	2012-08-02 04:51:21.237863	2012-08-02 04:57:57.138768
1955	1-1955	42	38	3844	35	1896	0	0	2012-08-02 04:55:11.701421	2012-08-02 04:57:57.138768
2337	1-2337	15	2	635	20	2438	0	0	2012-08-02 04:57:57.138768	2012-08-02 04:57:57.138768
2277	1-2277	39	22	2695	43	4209	0	0	2012-08-02 04:57:31.006997	2012-08-02 04:57:58.431066
1947	1-1947	68	50	4086	80	2742	0	0	2012-08-02 04:55:08.088186	2012-08-02 04:57:58.431066
1766	1-1766	35	32	4307	89	1496	0	0	2012-08-02 04:53:46.856946	2012-08-02 04:58:01.866816
1713	1-1713	85	22	8260	33	4579	0	4	2012-08-02 04:53:23.247549	2012-08-02 04:58:01.866816
1935	1-1935	29	74	6319	90	9653	0	4	2012-08-02 04:55:02.629241	2012-08-02 04:58:01.866816
2355	1-2355	32	21	261	18	3806	0	0	2012-08-02 04:58:04.872443	2012-08-02 04:58:04.872443
1956	1-1956	71	9	4626	52	4058	0	0	2012-08-02 04:55:11.723349	2012-08-02 04:58:05.768353
2357	1-2357	48	72	1206	9	6471	0	0	2012-08-02 04:58:05.768353	2012-08-02 04:58:05.768353
1727	1-1727	68	61	5709	71	8606	0	0	2012-08-02 04:53:29.313222	2012-08-02 04:58:06.198925
1472	1-1472	95	24	5898	18	3300	0	0	2012-08-02 04:51:24.640606	2012-08-02 04:58:08.364057
2203	1-2203	88	59	7521	75	4298	0	0	2012-08-02 04:56:58.730294	2012-08-02 04:58:08.364057
1765	1-1765	84	17	4157	60	440	0	0	2012-08-02 04:53:46.415096	2012-08-02 04:58:10.109932
2370	1-2370	36	78	253	55	2723	0	0	2012-08-02 04:58:11.435261	2012-08-02 04:58:11.435261
2310	1-2310	65	30	3142	24	9502	0	0	2012-08-02 04:57:45.325989	2012-08-02 04:58:12.296988
2372	1-2372	2	82	285	30	2592	0	0	2012-08-02 04:58:12.296988	2012-08-02 04:58:12.296988
1610	1-1610	70	32	6766	54	7337	0	0	2012-08-02 04:52:35.017945	2012-08-02 04:58:12.739132
1231	1-1231	62	22	6632	74	6165	0	0	2012-08-02 04:49:14.75711	2012-08-02 04:58:12.739132
2336	1-2336	4	68	178	81	319	0	6	2012-08-02 04:57:56.707555	2012-08-02 04:58:12.739132
2049	1-2049	57	16	7872	74	5984	0	0	2012-08-02 04:55:52.226884	2012-08-02 04:58:12.739132
2327	1-2327	16	19	1377	8	5844	0	0	2012-08-02 04:57:52.761319	2012-08-02 04:58:12.739132
2348	1-2348	84	89	998	1	6983	0	0	2012-08-02 04:58:01.866816	2012-08-02 04:58:13.181202
2102	1-2102	93	37	6853	64	5774	0	7	2012-08-02 04:56:14.912284	2012-08-02 04:58:17.522604
1632	1-1632	87	77	5406	48	4940	0	0	2012-08-02 04:52:44.994959	2012-08-02 04:58:18.837208
2289	1-2289	24	81	2441	97	1154	0	0	2012-08-02 04:57:36.221585	2012-08-02 04:58:18.837208
2387	1-2387	92	12	1215	77	7272	0	10	2012-08-02 04:58:18.837208	2012-08-02 04:58:18.837208
649	1-649	32	42	9410	94	1709	0	0	2012-08-02 04:44:48.035275	2012-08-02 04:58:21.466624
2393	1-2393	50	77	133	87	6118	0	0	2012-08-02 04:58:21.466624	2012-08-02 04:58:21.466624
2195	1-2195	79	55	5940	82	6295	0	10	2012-08-02 04:56:55.261569	2012-08-02 04:58:23.631867
2358	1-2358	69	28	3165	61	2741	0	0	2012-08-02 04:58:06.198925	2012-08-02 04:58:23.631867
2198	1-2198	72	85	7083	55	6636	0	0	2012-08-02 04:56:56.564843	2012-08-02 04:58:23.631867
2399	1-2399	57	33	1133	44	9773	0	0	2012-08-02 04:58:24.074213	2012-08-02 04:58:24.074213
2373	1-2373	55	74	4952	19	4349	0	0	2012-08-02 04:58:12.739132	2012-08-02 04:58:25.377863
1981	1-1981	5	54	5497	80	2275	0	0	2012-08-02 04:55:22.770905	2012-08-02 04:58:25.377863
2350	1-2350	41	54	3280	74	3250	0	0	2012-08-02 04:58:02.728656	2012-08-02 04:58:25.377863
1286	1-1286	85	16	3788	14	8570	0	0	2012-08-02 04:49:46.618614	2012-08-02 04:58:26.285767
2152	1-2152	83	94	3269	77	5387	0	0	2012-08-02 04:56:36.490255	2012-08-02 04:58:26.285767
2404	1-2404	52	4	2121	82	1942	0	0	2012-08-02 04:58:26.285767	2012-08-02 04:58:26.285767
1763	1-1763	61	97	1541	9	2192	0	0	2012-08-02 04:53:45.498063	2012-08-02 04:58:26.727646
2124	1-2124	4	23	3021	85	8973	0	0	2012-08-02 04:56:24.567777	2012-08-02 04:58:28.031146
2054	1-2054	1	10	5981	7	4620	0	0	2012-08-02 04:55:54.40312	2012-08-02 04:58:28.031146
2398	1-2398	11	11	1225	85	9674	0	0	2012-08-02 04:58:23.631867	2012-08-02 04:58:28.031146
2408	1-2408	46	85	3907	67	7998	0	13	2012-08-02 04:58:28.031146	2012-08-02 04:58:28.031146
1960	1-1960	23	16	6724	9	2405	0	0	2012-08-02 04:55:13.524003	2012-08-02 04:58:29.357537
2160	1-2160	52	52	8111	12	9853	0	11	2012-08-02 04:56:40.014993	2012-08-02 04:58:31.124695
2074	1-2074	20	12	5917	6	9951	0	0	2012-08-02 04:56:03.055348	2012-08-02 04:58:31.124695
1681	1-1681	70	28	8069	70	9356	0	0	2012-08-02 04:53:08.992915	2012-08-02 04:58:31.566653
2416	1-2416	93	66	208	82	2460	0	0	2012-08-02 04:58:31.566653	2012-08-02 04:58:31.566653
2245	1-2245	44	14	2778	40	9671	0	0	2012-08-02 04:57:16.996004	2012-08-02 04:58:32.439432
2219	1-2219	1	40	3821	4	6114	0	0	2012-08-02 04:57:05.737582	2012-08-02 04:58:32.439432
1693	1-1693	42	12	6603	72	7040	0	0	2012-08-02 04:53:14.008746	2012-08-02 04:58:32.439432
2334	1-2334	4	100	3332	92	8432	0	0	2012-08-02 04:57:55.834741	2012-08-02 04:58:33.743285
2421	1-2421	73	14	4524	8	2779	0	0	2012-08-02 04:58:33.743285	2012-08-02 04:58:33.743285
2402	1-2402	65	48	1867	54	9260	0	5	2012-08-02 04:58:25.377863	2012-08-02 04:58:34.174235
1980	1-1980	5	87	7725	54	7546	0	0	2012-08-02 04:55:22.351211	2012-08-02 04:58:34.174235
2367	1-2367	71	48	3314	57	7700	0	0	2012-08-02 04:58:10.109932	2012-08-02 04:58:34.174235
2090	1-2090	45	21	8020	38	7724	0	0	2012-08-02 04:56:09.575444	2012-08-02 04:58:35.135645
2157	1-2157	62	88	4953	78	2349	0	0	2012-08-02 04:56:38.666929	2012-08-02 04:58:35.135645
1292	1-1292	99	98	8826	47	5391	0	0	2012-08-02 04:49:49.35835	2012-08-02 04:58:35.135645
491	1-491	72	30	8301	80	5268	1741	0	2012-08-02 04:44:11.02645	2012-08-02 04:58:35.135645
2115	1-2115	75	47	8450	18	9670	4509	0	2012-08-02 04:56:20.656903	2012-08-02 04:58:35.135645
1267	1-1267	71	13	5446	15	4766	0	0	2012-08-02 04:49:36.709495	2012-08-02 04:58:35.588934
344	1-344	27	98	8242	76	3398	0	0	2012-08-02 04:44:00.544119	2012-08-02 04:58:37.323005
2415	1-2415	24	6	7495	52	6077	0	0	2012-08-02 04:58:31.124695	2012-08-02 04:58:37.753995
2329	1-2329	19	72	6815	28	8534	0	0	2012-08-02 04:57:53.656472	2012-08-02 04:58:37.753995
2424	1-2424	25	38	744	98	9248	0	0	2012-08-02 04:58:35.135645	2012-08-02 04:58:38.196056
2076	1-2076	2	37	5238	99	7338	0	0	2012-08-02 04:56:03.91729	2012-08-02 04:58:39.510517
1902	1-1902	2	45	4811	94	5391	0	7	2012-08-02 04:54:48.243293	2012-08-02 04:58:39.510517
2434	1-2434	99	99	2140	21	7354	0	0	2012-08-02 04:58:39.510517	2012-08-02 04:58:39.510517
1401	1-1401	67	45	7300	94	4394	3459	0	2012-08-02 04:50:45.27594	2012-08-02 04:58:39.510517
2122	1-2122	41	37	2229	16	8911	0	0	2012-08-02 04:56:23.683726	2012-08-02 04:58:41.709167
2281	1-2281	93	57	5636	54	5337	0	0	2012-08-02 04:57:32.786067	2012-08-02 04:58:42.150999
2086	1-2086	27	29	7350	78	5525	0	10	2012-08-02 04:56:07.818382	2012-08-02 04:58:45.189421
2435	1-2435	45	39	2323	65	5852	0	0	2012-08-02 04:58:39.963786	2012-08-02 04:58:45.189421
1860	1-1860	46	92	7093	91	1078	0	0	2012-08-02 04:54:28.976348	2012-08-02 04:58:48.713405
1790	1-1790	97	11	3206	59	7175	0	0	2012-08-02 04:53:57.797411	2012-08-02 04:58:48.713405
1739	1-1739	7	33	8180	59	8509	0	0	2012-08-02 04:53:35.02467	2012-08-02 04:58:48.713405
1280	1-1280	1	11	7858	89	9075	0	0	2012-08-02 04:49:43.039873	2012-08-02 04:58:48.713405
2455	1-2455	66	15	1301	33	4707	0	0	2012-08-02 04:58:48.713405	2012-08-02 04:58:48.713405
2272	1-2272	66	33	8231	11	6028	65	0	2012-08-02 04:57:28.808174	2012-08-02 04:58:48.713405
2290	1-2290	95	43	3899	50	9743	0	0	2012-08-02 04:57:36.641375	2012-08-02 04:58:52.33671
2081	1-2081	11	66	4047	79	7820	0	4	2012-08-02 04:56:06.061796	2012-08-02 04:58:53.233223
2465	1-2465	6	83	1315	66	4684	0	0	2012-08-02 04:58:53.233223	2012-08-02 04:58:53.233223
2252	1-2252	2	37	1457	62	2760	0	0	2012-08-02 04:57:20.078329	2012-08-02 04:58:53.689857
2057	1-2057	42	57	4105	73	832	0	0	2012-08-02 04:55:55.739998	2012-08-02 04:58:53.689857
2466	1-2466	10	71	663	28	6440	0	0	2012-08-02 04:58:53.689857	2012-08-02 04:58:53.689857
1730	1-1730	73	68	1970	50	2539	0	0	2012-08-02 04:53:30.970603	2012-08-02 04:58:54.183765
2291	1-2291	73	21	2632	22	4176	0	8	2012-08-02 04:57:37.083549	2012-08-02 04:58:55.061199
2340	1-2340	37	80	7227	21	6228	0	0	2012-08-02 04:57:58.431066	2012-08-02 04:58:55.061199
2218	1-2218	77	97	2805	22	6345	0	0	2012-08-02 04:57:05.306153	2012-08-02 04:58:55.061199
2296	1-2296	66	46	2806	78	5815	0	0	2012-08-02 04:57:39.249047	2012-08-02 04:58:55.503711
2470	1-2470	88	80	84	46	7300	0	0	2012-08-02 04:58:55.503711	2012-08-02 04:58:55.503711
1914	1-1914	12	68	5319	32	8185	0	0	2012-08-02 04:54:53.314071	2012-08-02 04:58:55.968492
2430	1-2430	17	4	1461	6	9812	0	0	2012-08-02 04:58:37.753995	2012-08-02 04:58:56.863457
824	1-824	48	64	906	6	2072	0	0	2012-08-02 04:45:44.043491	2012-08-02 04:58:56.863457
2314	1-2314	34	62	7033	40	5793	0	8	2012-08-02 04:57:47.071473	2012-08-02 04:58:57.305604
2436	1-2436	23	75	1070	61	2746	0	11	2012-08-02 04:58:40.405554	2012-08-02 04:58:58.167363
1814	1-1814	57	96	6755	75	2844	0	0	2012-08-02 04:54:08.87819	2012-08-02 04:58:58.167363
1869	1-1869	53	60	8057	39	8728	3587	0	2012-08-02 04:54:33.052479	2012-08-02 04:58:58.167363
1239	1-1239	93	7	7448	41	7735	0	4	2012-08-02 04:49:18.99931	2012-08-02 04:58:58.598363
1499	1-1499	49	7	7907	78	6548	0	0	2012-08-02 04:51:38.748081	2012-08-02 04:58:58.598363
2447	1-2447	13	78	3360	39	3237	0	0	2012-08-02 04:58:45.189421	2012-08-02 04:58:58.598363
1752	1-1752	65	7	6814	24	4056	0	0	2012-08-02 04:53:40.515218	2012-08-02 04:58:58.598363
2346	1-2346	78	64	4369	45	9557	0	0	2012-08-02 04:58:00.994171	2012-08-02 04:58:59.913747
772	1-772	73	98	5620	88	7380	0	0	2012-08-02 04:45:22.628906	2012-08-02 04:59:01.250663
2460	1-2460	18	22	251	88	528	0	0	2012-08-02 04:58:51.032923	2012-08-02 04:59:01.250663
1198	1-1198	69	98	3602	88	1895	0	5	2012-08-02 04:48:58.719157	2012-08-02 04:59:01.250663
2483	1-2483	91	88	3459	98	9578	0	0	2012-08-02 04:59:01.250663	2012-08-02 04:59:01.250663
935	1-935	97	31	6810	88	9761	2240	0	2012-08-02 04:46:34.507368	2012-08-02 04:59:01.250663
146	1-146	23	61	6806	27	7867	0	0	2012-08-02 04:43:55.173524	2012-08-02 04:59:02.55408
1953	1-1953	81	62	6422	16	6863	0	0	2012-08-02 04:55:10.817253	2012-08-02 04:59:02.98501
1144	1-1144	56	73	4084	76	5278	0	0	2012-08-02 04:48:32.450432	2012-08-02 04:59:03.416276
859	1-859	47	73	5228	94	2448	0	0	2012-08-02 04:45:57.714216	2012-08-02 04:59:03.416276
2488	1-2488	61	82	179	73	7714	0	0	2012-08-02 04:59:03.416276	2012-08-02 04:59:03.416276
2497	1-2497	23	40	477	63	5339	0	4	2012-08-02 04:59:07.355307	2012-08-02 04:59:07.355307
2463	1-2463	68	3	1759	43	6378	0	0	2012-08-02 04:58:52.33671	2012-08-02 04:59:08.228116
2499	1-2499	68	23	829	3	5801	0	0	2012-08-02 04:59:08.228116	2012-08-02 04:59:08.228116
2284	1-2284	91	46	1907	4	2323	0	0	2012-08-02 04:57:34.078357	2012-08-02 04:59:12.116739
364	1-364	72	61	5479	55	3170	0	0	2012-08-02 04:44:01.497319	2012-08-02 04:59:12.116739
2508	1-2508	34	40	668	46	4216	0	0	2012-08-02 04:59:12.116739	2012-08-02 04:59:12.116739
2477	1-2477	65	41	1735	7	5150	0	2	2012-08-02 04:58:58.598363	2012-08-02 04:59:12.978642
2510	1-2510	8	7	2518	67	3700	0	0	2012-08-02 04:59:12.978642	2012-08-02 04:59:12.978642
2364	1-2364	54	77	5196	81	7653	0	3	2012-08-02 04:58:08.806423	2012-08-02 04:59:13.840492
2512	1-2512	14	29	1442	77	3291	0	0	2012-08-02 04:59:13.840492	2012-08-02 04:59:13.840492
1767	1-1767	99	77	3312	40	8419	0	0	2012-08-02 04:53:47.309877	2012-08-02 04:59:14.283096
2351	1-2351	16	89	1639	44	1749	0	0	2012-08-02 04:58:03.148733	2012-08-02 04:59:14.724622
2514	1-2514	91	62	907	89	4609	0	0	2012-08-02 04:59:14.724622	2012-08-02 04:59:14.724622
2068	1-2068	2	97	3983	57	8862	0	0	2012-08-02 04:56:00.435403	2012-08-02 04:59:15.586623
2422	1-2422	74	54	7926	63	9192	0	0	2012-08-02 04:58:34.174235	2012-08-02 04:59:15.586623
2516	1-2516	3	71	324	97	5862	0	0	2012-08-02 04:59:15.586623	2012-08-02 04:59:15.586623
621	1-621	21	60	6076	76	8350	0	0	2012-08-02 04:44:41.282472	2012-08-02 04:59:16.028706
2403	1-2403	77	71	4622	62	6662	0	0	2012-08-02 04:58:25.854778	2012-08-02 04:59:16.028706
2034	1-2034	50	71	7284	92	6247	0	0	2012-08-02 04:55:45.708894	2012-08-02 04:59:16.028706
2517	1-2517	95	76	1266	71	7908	0	0	2012-08-02 04:59:16.028706	2012-08-02 04:59:16.028706
2532	1-2532	12	93	221	99	945	0	0	2012-08-02 04:59:22.447536	2012-08-02 04:59:22.447536
2112	1-2112	24	2	5773	74	4690	0	0	2012-08-02 04:56:19.320005	2012-08-02 04:59:22.867103
1336	1-1336	97	36	1456	85	1234	0	0	2012-08-02 04:50:13.611103	2012-08-02 04:59:23.28683
2390	1-2390	32	11	3313	32	4623	0	0	2012-08-02 04:58:20.141036	2012-08-02 04:59:24.126815
2528	1-2528	53	30	606	88	9224	0	0	2012-08-02 04:59:20.746201	2012-08-02 04:59:24.126815
2536	1-2536	33	88	6441	11	6410	0	0	2012-08-02 04:59:24.126815	2012-08-02 04:59:24.126815
1552	1-1552	56	45	8568	94	8041	1040	2	2012-08-02 04:52:04.9331	2012-08-02 04:59:24.126815
1708	1-1708	17	78	4346	16	8146	0	0	2012-08-02 04:53:20.836771	2012-08-02 04:59:24.55777
2389	1-2389	40	85	3473	6	8000	0	0	2012-08-02 04:58:19.721187	2012-08-02 04:59:24.977451
1394	1-1394	74	44	9898	33	6744	0	0	2012-08-02 04:50:42.039232	2012-08-02 04:59:24.977451
1195	1-1195	74	97	9182	99	3739	0	0	2012-08-02 04:48:56.42285	2012-08-02 04:59:24.977451
2538	1-2538	45	33	2586	66	5610	0	8	2012-08-02 04:59:24.977451	2012-08-02 04:59:24.977451
1069	1-1069	38	15	3560	45	260	0	0	2012-08-02 04:47:52.375105	2012-08-02 04:59:27.178037
2384	1-2384	84	15	9153	37	6716	0	0	2012-08-02 04:58:17.522604	2012-08-02 04:59:28.040138
2170	1-2170	6	45	5776	9	5120	0	0	2012-08-02 04:56:44.33473	2012-08-02 04:59:28.040138
1145	1-1145	37	95	5109	29	3883	0	0	2012-08-02 04:48:32.682089	2012-08-02 04:59:28.040138
2546	1-2546	6	2	338	72	2614	0	0	2012-08-02 04:59:28.470767	2012-08-02 04:59:28.470767
2144	1-2144	34	64	4568	98	8398	0	0	2012-08-02 04:56:32.944098	2012-08-02 04:59:28.89117
2547	1-2547	23	32	291	36	9374	0	13	2012-08-02 04:59:28.89117	2012-08-02 04:59:28.89117
500	1-500	21	15	7633	81	1544	0	0	2012-08-02 04:44:12.110264	2012-08-02 04:59:30.194222
2000	1-2000	78	79	5604	29	6117	0	0	2012-08-02 04:55:31.191601	2012-08-02 04:59:30.763067
2525	1-2525	46	20	1878	25	3542	0	0	2012-08-02 04:59:19.464476	2012-08-02 04:59:30.763067
997	1-997	54	44	5324	22	6135	0	0	2012-08-02 04:47:09.266525	2012-08-02 04:59:30.763067
2182	1-2182	74	36	7858	87	8914	0	0	2012-08-02 04:56:49.604397	2012-08-02 04:59:30.763067
2551	1-2551	11	30	2603	34	8427	0	0	2012-08-02 04:59:30.763067	2012-08-02 04:59:30.763067
2553	1-2553	11	4	407	1	6338	0	0	2012-08-02 04:59:31.663968	2012-08-02 04:59:31.663968
2556	1-2556	88	79	222	81	7696	0	0	2012-08-02 04:59:32.912346	2012-08-02 04:59:32.912346
1247	1-1247	45	75	7124	18	9396	200	0	2012-08-02 04:49:23.893001	2012-08-02 04:59:32.912346
2301	1-2301	94	18	5995	82	5993	0	0	2012-08-02 04:57:41.414465	2012-08-02 04:59:33.751963
2363	1-2363	64	18	7435	30	8419	0	0	2012-08-02 04:58:08.364057	2012-08-02 04:59:33.751963
1818	1-1818	35	69	5353	41	6696	0	0	2012-08-02 04:54:10.712091	2012-08-02 04:59:33.751963
2558	1-2558	49	20	840	18	8942	0	2	2012-08-02 04:59:33.751963	2012-08-02 04:59:33.751963
1678	1-1678	11	45	8347	71	4946	0	5	2012-08-02 04:53:07.612456	2012-08-02 04:59:35.508993
1346	1-1346	39	68	9561	81	7327	0	0	2012-08-02 04:50:19.421916	2012-08-02 04:59:35.508993
2050	1-2050	68	45	7599	68	8242	0	0	2012-08-02 04:55:52.646399	2012-08-02 04:59:35.508993
591	1-591	9	60	7489	80	9111	755	0	2012-08-02 04:44:34.338852	2012-08-02 04:59:35.508993
1587	1-1587	43	92	8705	50	9486	0	6	2012-08-02 04:52:24.290694	2012-08-02 04:59:36.348873
2374	1-2374	66	41	4743	11	3119	0	0	2012-08-02 04:58:13.181202	2012-08-02 04:59:36.348873
895	1-895	10	41	4016	80	2688	0	0	2012-08-02 04:46:13.848875	2012-08-02 04:59:36.348873
1925	1-1925	39	63	8467	77	9083	1961	0	2012-08-02 04:54:58.087034	2012-08-02 04:59:36.348873
2317	1-2317	29	81	7963	87	9291	0	2	2012-08-02 04:57:48.375236	2012-08-02 04:59:38.867502
2570	1-2570	43	5	2097	62	5534	0	0	2012-08-02 04:59:38.867502	2012-08-02 04:59:38.867502
1650	1-1650	91	16	3127	86	537	0	0	2012-08-02 04:52:53.611518	2012-08-02 04:59:39.718407
1911	1-1911	91	24	5334	83	5904	0	0	2012-08-02 04:54:51.943976	2012-08-02 04:59:39.718407
2167	1-2167	32	73	5803	97	6673	0	0	2012-08-02 04:56:43.053002	2012-08-02 04:59:40.138294
235	1-235	74	71	4139	47	1934	0	0	2012-08-02 04:43:56.758116	2012-08-02 04:59:40.138294
1864	1-1864	56	73	3657	11	1209	0	0	2012-08-02 04:54:30.765957	2012-08-02 04:59:40.138294
2395	1-2395	90	73	1649	1	281	0	0	2012-08-02 04:58:22.317355	2012-08-02 04:59:40.138294
2573	1-2573	21	44	1742	73	8377	0	0	2012-08-02 04:59:40.138294	2012-08-02 04:59:40.138294
1391	1-1391	65	18	8047	3	3565	0	0	2012-08-02 04:50:40.60304	2012-08-02 04:59:40.580613
1992	1-1992	69	99	6747	81	3153	0	0	2012-08-02 04:55:27.654245	2012-08-02 04:59:40.580613
2392	1-2392	83	16	4196	82	2000	0	11	2012-08-02 04:58:21.03603	2012-08-02 04:59:40.580613
2574	1-2574	52	93	4270	18	6736	0	0	2012-08-02 04:59:40.580613	2012-08-02 04:59:40.580613
2577	1-2577	71	88	1981	27	8645	0	0	2012-08-02 04:59:41.873675	2012-08-02 04:59:41.873675
1353	1-1353	33	33	2985	51	3128	0	0	2012-08-02 04:50:22.449069	2012-08-02 04:59:43.144142
2491	1-2491	12	81	9957	60	5612	0	0	2012-08-02 04:59:04.775347	2012-08-02 04:59:43.144142
2568	1-2568	87	33	6443	24	9563	0	0	2012-08-02 04:59:38.03895	2012-08-02 04:59:43.144142
2580	1-2580	94	29	786	33	8623	0	0	2012-08-02 04:59:43.144142	2012-08-02 04:59:43.144142
465	1-465	9	55	4850	47	1272	0	0	2012-08-02 04:44:08.07172	2012-08-02 04:59:46.590838
2590	1-2590	96	9	432	56	5623	0	0	2012-08-02 04:59:47.41945	2012-08-02 04:59:47.41945
1858	1-1858	90	1	5861	9	4317	0	0	2012-08-02 04:54:28.048054	2012-08-02 04:59:48.689902
2411	1-2411	31	9	2324	7	4515	0	0	2012-08-02 04:58:29.357537	2012-08-02 04:59:48.689902
2530	1-2530	74	91	2498	8	3247	0	0	2012-08-02 04:59:21.597043	2012-08-02 04:59:48.689902
2595	1-2595	7	17	2839	14	3918	0	0	2012-08-02 04:59:49.574207	2012-08-02 04:59:49.574207
1689	1-1689	1	13	9312	1	2420	0	0	2012-08-02 04:53:12.627658	2012-08-02 04:59:51.264292
1846	1-1846	15	98	7668	32	8791	0	0	2012-08-02 04:54:23.019487	2012-08-02 04:59:51.264292
2448	1-2448	28	94	3406	19	4315	0	0	2012-08-02 04:58:45.664184	2012-08-02 04:59:51.264292
1900	1-1900	78	38	3755	1	4033	0	0	2012-08-02 04:54:47.790004	2012-08-02 04:59:51.264292
2009	1-2009	44	68	2453	100	8225	0	0	2012-08-02 04:55:34.694032	2012-08-02 04:59:53.796257
1316	1-1316	89	26	5581	29	4733	0	0	2012-08-02 04:50:04.515125	2012-08-02 04:59:54.635641
2607	1-2607	31	29	1825	26	7602	0	0	2012-08-02 04:59:54.635641	2012-08-02 04:59:54.635641
2476	1-2476	9	39	721	96	4935	0	0	2012-08-02 04:58:58.167363	2012-08-02 04:59:56.337262
2471	1-2471	69	32	7192	26	5956	0	0	2012-08-02 04:58:55.968492	2012-08-02 04:59:57.165507
389	1-389	25	44	5016	6	6956	0	0	2012-08-02 04:44:03.094347	2012-08-02 04:59:57.596589
2349	1-2349	21	51	3036	13	4218	0	0	2012-08-02 04:58:02.309128	2012-08-02 04:59:59.297857
2376	1-2376	26	51	5354	29	1515	0	0	2012-08-02 04:58:14.054122	2012-08-02 04:59:59.297857
2486	1-2486	42	27	8151	61	8705	0	0	2012-08-02 04:59:02.55408	2012-08-02 05:00:00.137569
756	1-756	84	64	6678	48	5709	0	0	2012-08-02 04:45:17.159833	2012-08-02 05:00:00.137569
2620	1-2620	2	73	428	40	5084	0	4	2012-08-02 05:00:00.137569	2012-08-02 05:00:00.137569
1710	1-1710	9	61	8402	44	8163	2281	0	2012-08-02 04:53:21.753782	2012-08-02 05:00:00.137569
2621	1-2621	72	92	2699	93	6631	0	0	2012-08-02 05:00:00.568722	2012-08-02 05:00:00.568722
2606	1-2606	82	56	2765	44	3367	0	3	2012-08-02 04:59:54.215764	2012-08-02 05:00:01.112078
2622	1-2622	52	84	37	56	9764	0	0	2012-08-02 05:00:01.112078	2012-08-02 05:00:01.112078
2204	1-2204	59	63	6226	46	7246	0	0	2012-08-02 04:56:59.161002	2012-08-02 05:00:02.038419
2316	1-2316	14	11	2039	92	7680	0	0	2012-08-02 04:57:47.922079	2012-08-02 05:00:02.877858
1318	1-1318	89	81	4978	74	3144	0	0	2012-08-02 04:50:05.034766	2012-08-02 05:00:03.319857
996	1-996	42	21	9740	74	5978	0	0	2012-08-02 04:47:08.802505	2012-08-02 05:00:03.319857
1067	1-1067	19	21	8301	87	3337	0	0	2012-08-02 04:47:51.414124	2012-08-02 05:00:03.319857
2576	1-2576	65	87	1616	18	1793	0	4	2012-08-02 04:59:41.442576	2012-08-02 05:00:03.319857
2467	1-2467	5	40	9678	64	6248	0	0	2012-08-02 04:58:54.183765	2012-08-02 05:00:04.568435
2484	1-2484	56	86	1355	47	3234	0	0	2012-08-02 04:59:01.692457	2012-08-02 05:00:05.862802
2545	1-2545	61	50	4899	45	9173	0	3	2012-08-02 04:59:28.040138	2012-08-02 05:00:06.713494
2469	1-2469	61	22	9644	80	9619	0	0	2012-08-02 04:58:55.061199	2012-08-02 05:00:06.713494
2023	1-2023	99	32	7328	93	7671	0	0	2012-08-02 04:55:40.870365	2012-08-02 05:00:06.713494
2624	1-2624	88	53	2858	83	3474	0	0	2012-08-02 05:00:02.038419	2012-08-02 05:00:07.575402
1849	1-1849	14	1	4700	69	7583	0	0	2012-08-02 04:54:24.378692	2012-08-02 05:00:07.575402
4	1-4	4	8	9828	7	9098	0	0	2012-08-02 04:43:53.46468	2012-08-02 05:00:07.99533
2599	1-2599	11	32	2742	94	6528	0	0	2012-08-02 04:59:51.264292	2012-08-02 05:00:11.784526
182	1-182	9	1	8689	32	1383	1195	0	2012-08-02 04:43:55.699276	2012-08-02 05:00:11.784526
2627	1-2627	21	38	1855	21	7887	0	0	2012-08-02 05:00:03.319857	2012-08-02 05:00:12.215857
2648	1-2648	37	41	695	11	2366	0	6	2012-08-02 05:00:12.215857	2012-08-02 05:00:12.215857
2651	1-2651	91	32	607	60	6977	0	3	2012-08-02 05:00:13.486109	2012-08-02 05:00:13.486109
743	1-743	36	91	919	14	9107	0	0	2012-08-02 04:45:11.90051	2012-08-02 05:00:17.264087
1236	1-1236	89	72	8886	14	8273	0	0	2012-08-02 04:49:17.629776	2012-08-02 05:00:17.264087
1671	1-1671	98	91	1047	13	5691	0	0	2012-08-02 04:53:04.385082	2012-08-02 05:00:17.264087
2464	1-2464	13	74	3251	69	2657	0	0	2012-08-02 04:58:52.791318	2012-08-02 05:00:17.695298
2660	1-2660	8	9	2155	7	3790	0	5	2012-08-02 05:00:17.264087	2012-08-02 05:00:19.752018
2666	1-2666	31	7	1528	8	8425	0	0	2012-08-02 05:00:19.752018	2012-08-02 05:00:19.752018
816	1-816	30	47	4840	14	4242	0	0	2012-08-02 04:45:39.446991	2012-08-02 05:00:21.033463
2669	1-2669	5	14	2392	47	9631	0	0	2012-08-02 05:00:21.033463	2012-08-02 05:00:21.033463
2446	1-2446	81	14	9496	74	3616	0	0	2012-08-02 04:58:44.758167	2012-08-02 05:00:21.442311
2670	1-2670	8	15	7140	1	5225	0	0	2012-08-02 05:00:21.442311	2012-08-02 05:00:21.442311
1974	1-1974	78	78	9123	8	2176	0	0	2012-08-02 04:55:19.722003	2012-08-02 05:00:21.851319
2672	1-2672	32	72	1662	38	7133	0	0	2012-08-02 05:00:22.293314	2012-08-02 05:00:22.293314
2562	1-2562	94	80	2631	45	6647	0	0	2012-08-02 04:59:35.508993	2012-08-02 05:00:23.121729
2675	1-2675	6	10	1086	82	5620	0	0	2012-08-02 05:00:23.552833	2012-08-02 05:00:23.552833
2549	1-2549	46	72	1627	13	3873	0	0	2012-08-02 04:59:29.77483	2012-08-02 05:00:24.812471
2635	1-2635	25	59	5740	1	6307	0	0	2012-08-02 05:00:06.713494	2012-08-02 05:00:24.812471
2246	1-2246	29	31	5675	25	5709	0	0	2012-08-02 04:57:17.426569	2012-08-02 05:00:25.254448
2679	1-2679	66	22	1498	12	6940	0	2	2012-08-02 05:00:25.254448	2012-08-02 05:00:25.254448
2487	1-2487	32	28	1033	85	6997	0	0	2012-08-02 04:59:02.98501	2012-08-02 05:00:25.696413
2031	1-2031	29	53	5220	21	1780	0	0	2012-08-02 04:55:44.372078	2012-08-02 05:00:26.105222
2686	1-2686	11	36	352	48	4645	0	0	2012-08-02 05:00:28.292494	2012-08-02 05:00:28.292494
2243	1-2243	83	99	4559	79	6691	0	0	2012-08-02 04:57:16.12293	2012-08-02 05:00:30.369318
2691	1-2691	25	62	3403	99	2774	0	0	2012-08-02 05:00:30.369318	2012-08-02 05:00:30.369318
1747	1-1747	31	32	5673	92	2746	0	10	2012-08-02 04:53:38.68124	2012-08-02 05:00:33.286062
2214	1-2214	59	80	6259	63	3470	0	0	2012-08-02 04:57:03.548854	2012-08-02 05:00:33.727893
1566	1-1566	78	77	4371	36	3225	0	0	2012-08-02 04:52:13.616223	2012-08-02 05:00:34.147947
142	1-142	76	41	8775	62	2845	0	0	2012-08-02 04:43:55.095738	2012-08-02 05:00:34.147947
2605	1-2605	64	100	4312	62	5377	0	0	2012-08-02 04:59:53.796257	2012-08-02 05:00:34.567688
2601	1-2601	63	5	5981	62	4353	0	3	2012-08-02 04:59:52.115018	2012-08-02 05:00:34.567688
2701	1-2701	82	62	4368	74	6246	0	0	2012-08-02 05:00:34.567688	2012-08-02 05:00:34.567688
2643	1-2643	70	46	5062	55	5256	0	0	2012-08-02 05:00:10.083803	2012-08-02 05:00:35.407405
2321	1-2321	56	60	1440	48	2714	0	6	2012-08-02 04:57:50.109992	2012-08-02 05:00:35.407405
2703	1-2703	50	40	4195	46	6543	0	4	2012-08-02 05:00:35.407405	2012-08-02 05:00:35.407405
2440	1-2440	29	77	3212	97	2602	0	0	2012-08-02 04:58:42.150999	2012-08-02 05:00:35.827545
2493	1-2493	92	6	8102	79	5780	0	0	2012-08-02 04:59:05.632237	2012-08-02 05:00:35.827545
2661	1-2661	70	69	3232	94	6270	0	5	2012-08-02 05:00:17.695298	2012-08-02 05:00:36.258436
2342	1-2342	88	94	9716	18	8059	0	0	2012-08-02 04:57:59.281831	2012-08-02 05:00:36.258436
1930	1-1930	22	69	5166	19	1754	0	0	2012-08-02 04:55:00.386644	2012-08-02 05:00:36.258436
2705	1-2705	33	99	357	69	2081	0	0	2012-08-02 05:00:36.258436	2012-08-02 05:00:36.258436
1435	1-1435	32	91	6377	26	9335	0	0	2012-08-02 04:51:03.106492	2012-08-02 05:00:36.711722
2707	1-2707	49	4	244	92	5933	0	0	2012-08-02 05:00:37.153585	2012-08-02 05:00:37.153585
2418	1-2418	41	4	8735	95	9283	0	0	2012-08-02 04:58:32.439432	2012-08-02 05:00:38.004322
1881	1-1881	55	46	6553	27	8435	0	0	2012-08-02 04:54:39.029729	2012-08-02 05:00:38.866315
2714	1-2714	64	12	1761	44	9787	0	0	2012-08-02 05:00:40.147623	2012-08-02 05:00:40.147623
2520	1-2520	10	46	3030	95	3280	0	0	2012-08-02 04:59:17.332684	2012-08-02 05:00:40.567448
1719	1-1719	85	28	5654	4	8979	0	0	2012-08-02 04:53:25.965516	2012-08-02 05:00:43.130348
2478	1-2478	74	28	3436	97	6862	0	0	2012-08-02 04:58:59.040826	2012-08-02 05:00:43.130348
2550	1-2550	18	29	742	15	5626	0	0	2012-08-02 04:59:30.194222	2012-08-02 05:00:43.130348
1741	1-1741	98	3	5847	62	5044	0	0	2012-08-02 04:53:35.952654	2012-08-02 05:00:43.130348
2721	1-2721	56	42	3192	28	5333	0	0	2012-08-02 05:00:43.130348	2012-08-02 05:00:43.130348
2137	1-2137	33	57	9509	56	4991	0	0	2012-08-02 04:56:29.839053	2012-08-02 05:00:43.561703
2726	1-2726	83	55	819	37	7271	0	0	2012-08-02 05:00:45.251866	2012-08-02 05:00:45.251866
1512	1-1512	98	87	6295	29	2514	2340	0	2012-08-02 04:51:44.812945	2012-08-02 05:00:45.251866
2228	1-2228	94	21	9991	45	7919	0	0	2012-08-02 04:57:09.593125	2012-08-02 05:00:45.705208
2021	1-2021	46	77	2925	85	1906	0	0	2012-08-02 04:55:39.986355	2012-08-02 05:00:45.705208
2727	1-2727	95	85	486	69	4204	0	0	2012-08-02 05:00:45.705208	2012-08-02 05:00:45.705208
2631	1-2631	74	37	4042	87	3522	0	0	2012-08-02 05:00:04.999905	2012-08-02 05:00:47.36408
2591	1-2591	22	70	1981	3	4208	0	0	2012-08-02 04:59:47.839356	2012-08-02 05:00:49.496231
1359	1-1359	60	3	7182	59	7829	0	0	2012-08-02 04:50:25.387602	2012-08-02 05:00:49.496231
15	1-15	15	24	6128	4	1868	0	0	2012-08-02 04:43:53.609491	2012-08-02 05:00:50.766846
2739	1-2739	44	23	2506	24	6639	0	0	2012-08-02 05:00:50.766846	2012-08-02 05:00:50.766846
2564	1-2564	98	77	1736	41	6403	0	0	2012-08-02 04:59:36.348873	2012-08-02 05:00:52.037763
2742	1-2742	13	57	1496	77	1629	0	0	2012-08-02 05:00:52.037763	2012-08-02 05:00:52.037763
2368	1-2368	24	19	4253	22	9043	0	2	2012-08-02 04:58:10.551701	2012-08-02 05:00:54.534352
932	1-932	100	15	2335	31	332	0	0	2012-08-02 04:46:33.855789	2012-08-02 05:00:54.534352
2748	1-2748	96	56	158	15	5567	0	5	2012-08-02 05:00:54.534352	2012-08-02 05:00:54.534352
2521	1-2521	38	4	3650	100	9505	0	0	2012-08-02 04:59:17.752554	2012-08-02 05:00:54.965229
1323	1-1323	92	24	4420	13	2925	0	0	2012-08-02 04:50:08.384502	2012-08-02 05:00:54.965229
2611	1-2611	69	90	2789	3	2106	0	0	2012-08-02 04:59:56.337262	2012-08-02 05:00:54.965229
2248	1-2248	99	56	7639	49	7165	0	0	2012-08-02 04:57:18.288272	2012-08-02 05:00:55.81654
2731	1-2731	81	29	100	76	362	0	0	2012-08-02 05:00:47.36408	2012-08-02 05:00:55.81654
2563	1-2563	82	74	9960	38	6193	0	0	2012-08-02 04:59:35.928819	2012-08-02 05:00:55.81654
2751	1-2751	11	38	395	56	9574	0	0	2012-08-02 05:00:55.81654	2012-08-02 05:00:55.81654
1975	1-1975	90	87	9519	29	8865	3580	2	2012-08-02 04:55:20.164086	2012-08-02 05:00:55.81654
2604	1-2604	79	83	5418	67	2000	0	3	2012-08-02 04:59:53.374373	2012-08-02 05:00:56.68911
249	1-249	82	83	8784	84	2261	0	0	2012-08-02 04:43:57.091762	2012-08-02 05:00:56.68911
2753	1-2753	28	17	3790	83	3807	0	0	2012-08-02 05:00:56.68911	2012-08-02 05:00:56.68911
2637	1-2637	96	29	3139	3	8791	0	0	2012-08-02 05:00:07.575402	2012-08-02 05:00:57.959709
2220	1-2220	25	66	1611	75	892	0	0	2012-08-02 04:57:06.157079	2012-08-02 05:00:57.959709
2692	1-2692	6	36	83	10	3356	0	0	2012-08-02 05:00:30.800548	2012-08-02 05:00:58.390777
2639	1-2639	12	60	1743	10	9858	0	0	2012-08-02 05:00:08.426356	2012-08-02 05:00:58.390777
2689	1-2689	35	22	1720	62	2480	0	0	2012-08-02 05:00:29.541083	2012-08-02 05:01:02.544567
1787	1-1787	25	22	4835	12	1695	0	0	2012-08-02 04:53:56.447141	2012-08-02 05:01:02.544567
1114	1-1114	20	19	3412	94	1619	0	0	2012-08-02 04:48:17.147294	2012-08-02 05:01:02.544567
2587	1-2587	87	22	5043	56	4179	0	0	2012-08-02 04:59:46.171288	2012-08-02 05:01:02.544567
2767	1-2767	7	31	5701	19	9375	0	0	2012-08-02 05:01:02.544567	2012-08-02 05:01:02.544567
2473	1-2473	37	6	7002	62	9265	0	0	2012-08-02 04:58:56.863457	2012-08-02 05:01:02.964724
123	1-123	68	35	5955	93	9088	0	0	2012-08-02 04:43:54.861484	2012-08-02 05:01:02.964724
155	1-155	80	35	8294	40	7549	0	0	2012-08-02 04:43:55.309333	2012-08-02 05:01:02.964724
2078	1-2078	88	35	3846	18	6236	0	0	2012-08-02 04:56:04.767737	2012-08-02 05:01:02.964724
2768	1-2768	62	40	5514	35	9067	0	5	2012-08-02 05:01:02.964724	2012-08-02 05:01:02.964724
2534	1-2534	45	79	2647	36	2256	0	0	2012-08-02 04:59:23.28683	2012-08-02 05:01:03.804507
2770	1-2770	43	50	1859	27	4304	0	0	2012-08-02 05:01:03.804507	2012-08-02 05:01:03.804507
2618	1-2618	43	66	177	51	9949	0	0	2012-08-02 04:59:59.297857	2012-08-02 05:01:04.224383
2696	1-2696	77	85	5513	82	3605	0	0	2012-08-02 05:00:32.479547	2012-08-02 05:01:04.655402
2765	1-2765	46	49	3986	85	7466	0	0	2012-08-02 05:01:01.694178	2012-08-02 05:01:04.655402
1487	1-1487	8	49	5282	87	3628	0	5	2012-08-02 04:51:32.838043	2012-08-02 05:01:04.655402
2772	1-2772	75	29	3691	49	7997	0	0	2012-08-02 05:01:04.655402	2012-08-02 05:01:04.655402
2774	1-2774	16	79	3522	63	1390	0	0	2012-08-02 05:01:05.518544	2012-08-02 05:01:05.518544
1348	1-1348	85	56	6761	46	3847	0	0	2012-08-02 04:50:20.394178	2012-08-02 05:01:06.347142
1366	1-1366	59	56	6123	96	3903	0	0	2012-08-02 04:50:28.723851	2012-08-02 05:01:06.347142
1759	1-1759	25	33	1430	5	4778	0	0	2012-08-02 04:53:43.70848	2012-08-02 05:01:06.347142
2274	1-2274	19	70	2079	5	4956	0	0	2012-08-02 04:57:29.692301	2012-08-02 05:01:06.347142
1661	1-1661	64	8	535	54	8109	0	8	2012-08-02 04:52:58.794864	2012-08-02 05:01:06.745092
2345	1-2345	2	54	1444	17	8781	0	0	2012-08-02 04:58:00.574427	2012-08-02 05:01:06.745092
2777	1-2777	66	17	1600	60	8208	0	4	2012-08-02 05:01:06.745092	2012-08-02 05:01:06.745092
2589	1-2589	39	100	5731	11	1904	0	0	2012-08-02 04:59:47.010603	2012-08-02 05:01:08.380508
2322	1-2322	41	65	9389	38	7337	0	0	2012-08-02 04:57:50.551534	2012-08-02 05:01:08.380508
2409	1-2409	10	8	2239	3	5343	0	0	2012-08-02 04:58:28.484575	2012-08-02 05:01:08.380508
2676	1-2676	36	8	2984	65	7657	0	5	2012-08-02 05:00:23.961602	2012-08-02 05:01:08.380508
2407	1-2407	61	8	1195	65	2456	0	0	2012-08-02 04:58:27.600425	2012-08-02 05:01:08.380508
2757	1-2757	42	10	9022	91	3936	0	0	2012-08-02 05:00:58.390777	2012-08-02 05:01:08.380508
2630	1-2630	30	64	2123	82	3522	0	0	2012-08-02 05:00:04.568435	2012-08-02 05:01:08.380508
2781	1-2781	63	11	1518	8	9362	0	0	2012-08-02 05:01:08.380508	2012-08-02 05:01:08.380508
1481	1-1481	83	77	7379	85	3128	0	0	2012-08-02 04:51:28.872678	2012-08-02 05:01:10.923799
2787	1-2787	27	82	378	77	9721	0	0	2012-08-02 05:01:10.923799	2012-08-02 05:01:10.923799
2685	1-2685	43	37	4531	57	8526	0	0	2012-08-02 05:00:27.87266	2012-08-02 05:01:11.785278
2790	1-2790	93	42	245	67	5613	0	12	2012-08-02 05:01:12.216516	2012-08-02 05:01:12.216516
2212	1-2212	3	78	952	2	4077	0	0	2012-08-02 04:57:02.687128	2012-08-02 05:01:16.326107
2613	1-2613	2	59	3944	98	6157	0	0	2012-08-02 04:59:57.165507	2012-08-02 05:01:16.779156
2801	1-2801	62	86	218	67	6944	0	6	2012-08-02 05:01:16.779156	2012-08-02 05:01:16.779156
2709	1-2709	72	71	6825	79	9163	0	10	2012-08-02 05:00:38.004322	2012-08-02 05:01:17.640637
2776	1-2776	18	5	6936	56	8617	0	0	2012-08-02 05:01:06.347142	2012-08-02 05:01:17.640637
2749	1-2749	3	13	619	4	2150	0	5	2012-08-02 05:00:54.965229	2012-08-02 05:01:18.060564
2652	1-2652	85	86	434	78	5188	0	0	2012-08-02 05:00:13.905993	2012-08-02 05:01:20.999047
2600	1-2600	17	19	5545	22	9169	0	5	2012-08-02 04:59:51.706372	2012-08-02 05:01:20.999047
2811	1-2811	24	27	1514	19	6838	0	5	2012-08-02 05:01:20.999047	2012-08-02 05:01:20.999047
2642	1-2642	37	41	4114	68	4659	0	0	2012-08-02 05:00:09.674875	2012-08-02 05:01:23.098351
2555	1-2555	55	82	7508	11	9037	5798	0	2012-08-02 04:59:32.492452	2012-08-02 05:01:23.098351
2458	1-2458	8	70	325	17	7325	0	0	2012-08-02 04:58:50.160336	2012-08-02 05:01:26.003719
2429	1-2429	90	76	1998	38	987	0	0	2012-08-02 04:58:37.323005	2012-08-02 05:01:26.003719
2480	1-2480	50	87	4400	64	2955	0	0	2012-08-02 04:58:59.913747	2012-08-02 05:01:26.003719
183	1-183	23	15	8402	17	3301	0	0	2012-08-02 04:43:55.710347	2012-08-02 05:01:26.003719
2593	1-2593	96	7	4364	91	6706	0	0	2012-08-02 04:59:48.689902	2012-08-02 05:01:26.821427
2816	1-2816	72	11	538	41	5700	0	0	2012-08-02 05:01:23.098351	2012-08-02 05:01:26.821427
1408	1-1408	27	15	3627	69	271	0	0	2012-08-02 04:50:49.551216	2012-08-02 05:01:26.821427
2396	1-2396	55	48	4232	93	6736	0	0	2012-08-02 04:58:22.759178	2012-08-02 05:01:26.821427
2543	1-2543	23	9	603	15	9725	0	4	2012-08-02 04:59:27.178037	2012-08-02 05:01:26.821427
1436	1-1436	30	71	9758	94	6261	0	0	2012-08-02 04:51:03.581419	2012-08-02 05:01:26.821427
2825	1-2825	50	41	5257	9	8887	0	0	2012-08-02 05:01:26.821427	2012-08-02 05:01:26.821427
1457	1-1457	51	21	8822	41	6886	3137	0	2012-08-02 04:51:16.277806	2012-08-02 05:01:26.821427
2472	1-2472	32	49	4033	11	6211	0	0	2012-08-02 04:58:56.410305	2012-08-02 05:01:27.264196
1636	1-1636	62	3	3182	55	3156	0	4	2012-08-02 04:52:46.851119	2012-08-02 05:01:28.932658
2236	1-2236	35	52	3155	55	7571	0	0	2012-08-02 04:57:13.117692	2012-08-02 05:01:28.932658
2756	1-2756	40	3	3039	18	6236	0	0	2012-08-02 05:00:57.959709	2012-08-02 05:01:28.932658
2823	1-2823	64	17	2197	87	3685	0	0	2012-08-02 05:01:26.003719	2012-08-02 05:01:29.771994
2612	1-2612	7	44	9752	89	7597	0	11	2012-08-02 04:59:56.745966	2012-08-02 05:01:29.771994
2834	1-2834	59	77	1855	36	1602	0	8	2012-08-02 05:01:30.633783	2012-08-02 05:01:30.633783
574	1-574	39	82	9575	29	7695	0	0	2012-08-02 04:44:30.206017	2012-08-02 05:01:31.064532
2581	1-2581	22	91	716	72	4538	0	0	2012-08-02 04:59:43.575046	2012-08-02 05:01:31.064532
707	1-707	49	82	6132	94	4448	0	0	2012-08-02 04:45:00.260908	2012-08-02 05:01:31.064532
2502	1-2502	98	97	9156	44	9091	0	0	2012-08-02 04:59:09.5095	2012-08-02 05:01:32.755101
2633	1-2633	71	49	1597	84	1722	0	0	2012-08-02 05:00:05.862802	2012-08-02 05:01:32.755101
2839	1-2839	47	44	3329	94	4633	0	0	2012-08-02 05:01:32.755101	2012-08-02 05:01:32.755101
2237	1-2237	22	24	2476	16	4713	0	0	2012-08-02 04:57:13.548592	2012-08-02 05:01:33.594898
2845	1-2845	80	73	1198	13	5506	0	0	2012-08-02 05:01:35.28515	2012-08-02 05:01:35.28515
2644	1-2644	8	25	7848	71	9621	0	0	2012-08-02 05:00:10.52519	2012-08-02 05:01:36.533922
1142	1-1142	31	61	9426	92	9258	0	0	2012-08-02 04:48:31.535689	2012-08-02 05:01:36.533922
2841	1-2841	36	16	2079	57	5031	0	0	2012-08-02 05:01:33.594898	2012-08-02 05:01:37.362291
2338	1-2338	48	3	5518	24	4971	0	0	2012-08-02 04:57:57.58045	2012-08-02 05:01:37.362291
2699	1-2699	80	63	3999	25	7397	0	0	2012-08-02 05:00:33.727893	2012-08-02 05:01:38.190963
1686	1-1686	26	30	6270	11	3313	0	7	2012-08-02 04:53:11.268718	2012-08-02 05:01:38.190963
2718	1-2718	58	22	2016	80	7290	0	7	2012-08-02 05:00:41.870998	2012-08-02 05:01:38.190963
2205	1-2205	53	84	6035	44	8325	0	0	2012-08-02 04:56:59.594059	2012-08-02 05:01:38.190963
2813	1-2813	5	30	2128	97	7395	0	6	2012-08-02 05:01:21.83889	2012-08-02 05:01:38.190963
2830	1-2830	94	55	3548	3	5825	0	0	2012-08-02 05:01:28.932658	2012-08-02 05:01:38.190963
171	1-171	72	30	9534	47	4947	0	0	2012-08-02 04:43:55.542921	2012-08-02 05:01:38.190963
2852	1-2852	83	25	4539	30	9495	0	0	2012-08-02 05:01:38.190963	2012-08-02 05:01:38.190963
1294	1-1294	64	30	9255	84	6869	934	0	2012-08-02 04:49:51.31324	2012-08-02 05:01:38.190963
2206	1-2206	84	91	3170	46	5059	0	0	2012-08-02 04:57:00.046942	2012-08-02 05:01:38.668588
2638	1-2638	3	91	5988	47	5475	0	0	2012-08-02 05:00:07.99533	2012-08-02 05:01:40.325948
2857	1-2857	8	55	7128	9	6884	0	0	2012-08-02 05:01:40.325948	2012-08-02 05:01:40.325948
2760	1-2760	11	54	6089	77	4815	0	0	2012-08-02 05:00:59.650206	2012-08-02 05:01:43.661842
2865	1-2865	78	76	713	54	6743	0	0	2012-08-02 05:01:43.661842	2012-08-02 05:01:43.661842
2850	1-2850	39	50	5211	55	4964	0	0	2012-08-02 05:01:37.362291	2012-08-02 05:01:44.081891
2671	1-2671	8	3	5786	26	3271	0	0	2012-08-02 05:00:21.851319	2012-08-02 05:01:44.491033
2867	1-2867	100	60	44	3	6155	0	10	2012-08-02 05:01:44.491033	2012-08-02 05:01:44.491033
2706	1-2706	57	89	5286	1	6625	0	5	2012-08-02 05:00:36.711722	2012-08-02 05:01:48.655442
2877	1-2877	6	49	293	89	8222	0	0	2012-08-02 05:01:48.655442	2012-08-02 05:01:48.655442
2879	1-2879	78	5	298	36	68	0	0	2012-08-02 05:01:49.506306	2012-08-02 05:01:49.506306
2859	1-2859	7	22	1447	76	3893	0	0	2012-08-02 05:01:41.176518	2012-08-02 05:01:51.174685
2626	1-2626	72	92	7187	69	6265	0	5	2012-08-02 05:00:02.877858	2012-08-02 05:01:52.831904
2832	1-2832	64	87	3553	44	9287	0	0	2012-08-02 05:01:29.771994	2012-08-02 05:01:53.671652
2743	1-2743	83	77	2852	99	2640	0	0	2012-08-02 05:00:52.44641	2012-08-02 05:01:56.532832
2377	1-2377	63	46	6261	87	3882	80	0	2012-08-02 04:58:14.484922	2012-08-02 05:01:56.532832
247	1-247	10	30	9456	97	6248	0	0	2012-08-02 04:43:57.069691	2012-08-02 05:01:58.598889
2887	1-2887	78	97	1885	32	6192	0	5	2012-08-02 05:01:52.831904	2012-08-02 05:01:59.007534
2901	1-2901	20	32	3309	30	3476	0	0	2012-08-02 05:01:58.598889	2012-08-02 05:01:59.825452
2771	1-2771	42	51	4529	54	8181	0	0	2012-08-02 05:01:04.224383	2012-08-02 05:01:59.825452
2438	1-2438	39	40	3632	27	2063	0	0	2012-08-02 04:58:41.278246	2012-08-02 05:02:00.665143
1309	1-1309	12	67	8907	65	8834	0	0	2012-08-02 04:50:01.2884	2012-08-02 05:02:00.665143
436	1-436	74	67	8473	43	6835	0	0	2012-08-02 04:44:06.123898	2012-08-02 05:02:00.665143
1613	1-1613	42	48	8856	27	6511	0	0	2012-08-02 04:52:36.443557	2012-08-02 05:02:01.913707
1084	1-1084	61	95	8813	12	4955	0	0	2012-08-02 04:47:59.347978	2012-08-02 05:02:02.753314
2912	1-2912	41	14	711	87	5587	0	0	2012-08-02 05:02:03.173376	2012-08-02 05:02:03.173376
2789	1-2789	8	99	2312	37	8069	0	0	2012-08-02 05:01:11.785278	2012-08-02 05:02:04.001512
2503	1-2503	7	57	8462	46	4337	0	0	2012-08-02 04:59:09.929366	2012-08-02 05:02:04.001512
2914	1-2914	99	76	1675	71	7075	0	5	2012-08-02 05:02:04.001512	2012-08-02 05:02:04.001512
2495	1-2495	47	47	6466	55	9712	0	0	2012-08-02 04:59:06.471846	2012-08-02 05:02:06.932077
2426	1-2426	99	55	5509	16	5709	0	0	2012-08-02 04:58:36.030492	2012-08-02 05:02:06.932077
2694	1-2694	90	47	502	78	2862	0	12	2012-08-02 05:00:31.617905	2012-08-02 05:02:06.932077
2921	1-2921	68	15	337	47	2239	0	5	2012-08-02 05:02:06.932077	2012-08-02 05:02:06.932077
848	1-848	4	30	5221	86	2042	0	0	2012-08-02 04:45:52.53211	2012-08-02 05:02:07.783043
2923	1-2923	95	39	8	30	8087	0	0	2012-08-02 05:02:07.783043	2012-08-02 05:02:07.783043
1176	1-1176	75	49	4582	74	6578	0	0	2012-08-02 04:48:45.698536	2012-08-02 05:02:09.031601
2584	1-2584	12	34	1504	47	1969	0	0	2012-08-02 04:59:44.86746	2012-08-02 05:02:10.280083
2062	1-2062	66	34	3038	20	8630	0	0	2012-08-02 04:55:57.860952	2012-08-02 05:02:10.280083
2269	1-2269	97	12	4693	36	2316	0	0	2012-08-02 04:57:27.493244	2012-08-02 05:02:10.280083
944	1-944	13	58	8859	47	4241	0	0	2012-08-02 04:46:40.274044	2012-08-02 05:02:10.280083
275	1-275	76	8	9959	9	9647	0	0	2012-08-02 04:43:58.112077	2012-08-02 05:02:11.926122
2933	1-2933	67	53	1623	8	3257	0	0	2012-08-02 05:02:11.926122	2012-08-02 05:02:11.926122
2809	1-2809	29	91	4769	64	3281	0	0	2012-08-02 05:01:20.148432	2012-08-02 05:02:12.765897
1937	1-1937	46	34	6788	64	8038	0	0	2012-08-02 04:55:03.534891	2012-08-02 05:02:12.765897
2941	1-2941	86	7	1831	43	3838	0	0	2012-08-02 05:02:15.240565	2012-08-02 05:02:15.240565
2848	1-2848	13	45	500	63	5460	0	8	2012-08-02 05:01:36.533922	2012-08-02 05:02:16.478267
1651	1-1651	25	33	6105	17	8413	0	0	2012-08-02 04:52:54.097948	2012-08-02 05:02:16.908756
2945	1-2945	3	2	412	33	5162	0	0	2012-08-02 05:02:16.908756	2012-08-02 05:02:16.908756
2427	1-2427	76	36	583	100	1751	0	0	2012-08-02 04:58:36.472424	2012-08-02 05:02:17.74858
2929	1-2929	57	36	4147	34	9060	0	6	2012-08-02 05:02:10.280083	2012-08-02 05:02:17.74858
2947	1-2947	75	81	2247	36	8750	0	6	2012-08-02 05:02:17.74858	2012-08-02 05:02:17.74858
2954	1-2954	61	95	243	67	6936	0	0	2012-08-02 05:02:20.59891	2012-08-02 05:02:20.59891
1493	1-1493	45	59	2927	90	2396	0	0	2012-08-02 04:51:36.251364	2012-08-02 05:02:21.803018
2957	1-2957	87	90	1723	9	1538	0	0	2012-08-02 05:02:21.803018	2012-08-02 05:02:21.803018
2603	1-2603	18	83	3229	80	6347	0	0	2012-08-02 04:59:52.954905	2012-08-02 05:02:22.212404
2944	1-2944	90	63	2149	83	5712	0	0	2012-08-02 05:02:16.478267	2012-08-02 05:02:22.212404
2263	1-2263	98	63	3317	27	3811	0	0	2012-08-02 04:57:24.873038	2012-08-02 05:02:22.212404
2588	1-2588	37	85	3058	54	8231	0	0	2012-08-02 04:59:46.590838	2012-08-02 05:02:22.212404
2354	1-2354	90	50	6639	72	4884	0	0	2012-08-02 04:58:04.43017	2012-08-02 05:02:22.212404
1567	1-1567	51	63	5714	5	8237	0	0	2012-08-02 04:52:14.091134	2012-08-02 05:02:22.212404
2958	1-2958	84	91	640	63	7509	0	11	2012-08-02 05:02:22.212404	2012-08-02 05:02:22.212404
664	1-664	73	63	8192	50	7913	2078	0	2012-08-02 04:44:51.651282	2012-08-02 05:02:22.212404
2608	1-2608	100	74	7530	9	3561	464	0	2012-08-02 04:59:55.066605	2012-08-02 05:02:22.212404
2594	1-2594	58	16	3184	15	8048	0	0	2012-08-02 04:59:49.132279	2012-08-02 05:02:22.654537
2439	1-2439	4	94	1490	51	6196	0	0	2012-08-02 04:58:41.709167	2012-08-02 05:02:22.654537
2959	1-2959	11	15	2793	92	8218	0	0	2012-08-02 05:02:22.654537	2012-08-02 05:02:22.654537
2763	1-2763	50	59	3127	15	3957	0	0	2012-08-02 05:01:00.876384	2012-08-02 05:02:23.063517
2397	1-2397	16	91	1537	15	3179	0	0	2012-08-02 04:58:23.201497	2012-08-02 05:02:23.063517
1781	1-1781	83	95	9813	15	9291	0	0	2012-08-02 04:53:53.739941	2012-08-02 05:02:23.063517
2961	1-2961	85	88	6283	78	5352	0	4	2012-08-02 05:02:23.472456	2012-08-02 05:02:23.472456
2922	1-2922	85	18	4489	36	3050	0	0	2012-08-02 05:02:07.37447	2012-08-02 05:02:25.858944
1688	1-1688	66	18	6461	74	7156	0	0	2012-08-02 04:53:12.174538	2012-08-02 05:02:25.858944
2967	1-2967	28	63	3120	18	8165	0	0	2012-08-02 05:02:25.858944	2012-08-02 05:02:27.096169
2213	1-2213	21	65	7138	4	7689	0	0	2012-08-02 04:57:03.118031	2012-08-02 05:02:27.493789
1439	1-1439	84	48	2205	10	2390	0	3	2012-08-02 04:51:05.494243	2012-08-02 05:02:27.493789
2417	1-2417	93	1	4328	63	7183	0	0	2012-08-02 04:58:32.008646	2012-08-02 05:02:27.493789
2972	1-2972	15	15	586	91	9876	0	0	2012-08-02 05:02:27.902965	2012-08-02 05:02:27.902965
2202	1-2202	92	43	1055	12	836	0	0	2012-08-02 04:56:58.310458	2012-08-02 05:02:31.18389
2552	1-2552	48	59	3245	53	4052	0	0	2012-08-02 04:59:31.233542	2012-08-02 05:02:31.18389
1430	1-1430	26	97	5047	62	1424	0	0	2012-08-02 04:51:00.75325	2012-08-02 05:02:31.58168
2983	1-2983	89	48	1128	15	7534	0	11	2012-08-02 05:02:32.38847	2012-08-02 05:02:32.38847
523	1-523	72	64	5459	74	4613	0	0	2012-08-02 04:44:16.078865	2012-08-02 05:02:33.570341
2896	1-2896	44	87	8476	77	9988	0	0	2012-08-02 05:01:56.532832	2012-08-02 05:02:33.570341
2986	1-2986	23	74	1759	64	6303	0	0	2012-08-02 05:02:33.570341	2012-08-02 05:02:33.570341
1934	1-1934	16	50	8825	77	2661	0	0	2012-08-02 04:55:02.198193	2012-08-02 05:02:34.411944
1910	1-1910	45	67	6980	76	6467	0	0	2012-08-02 04:54:51.480024	2012-08-02 05:02:34.411944
2906	1-2906	3	27	2307	67	5128	0	0	2012-08-02 05:02:00.665143	2012-08-02 05:02:34.411944
1880	1-1880	28	21	280	98	123	0	0	2012-08-02 04:54:38.565727	2012-08-02 05:02:34.820936
2935	1-2935	100	64	3860	91	8276	0	4	2012-08-02 05:02:12.765897	2012-08-02 05:02:34.820936
2989	1-2989	81	98	850	64	7057	0	0	2012-08-02 05:02:34.820936	2012-08-02 05:02:34.820936
2997	1-2997	57	43	3504	6	8622	0	0	2012-08-02 05:02:38.079925	2012-08-02 05:02:38.079925
2292	1-2292	57	39	3857	3	1696	0	0	2012-08-02 04:57:37.514201	2012-08-02 05:02:38.875532
2688	1-2688	19	73	2225	41	2959	0	6	2012-08-02 05:00:29.121101	2012-08-02 05:02:42.565638
2826	1-2826	65	96	3635	49	9866	0	7	2012-08-02 05:01:27.264196	2012-08-02 05:02:44.167298
2934	1-2934	56	7	2137	23	5954	0	0	2012-08-02 05:02:12.346212	2012-08-02 05:02:44.167298
2166	1-2166	20	43	9032	70	4654	0	7	2012-08-02 04:56:42.633216	2012-08-02 05:02:46.144851
2931	1-2931	78	43	1195	52	1583	0	0	2012-08-02 05:02:11.108629	2012-08-02 05:02:46.144851
3017	1-3017	99	16	1848	43	6733	0	5	2012-08-02 05:02:46.144851	2012-08-02 05:02:46.144851
2320	1-2320	35	18	9783	58	5676	0	0	2012-08-02 04:57:49.678789	2012-08-02 05:02:49.349011
2443	1-2443	16	18	7246	99	7017	0	3	2012-08-02 04:58:43.443971	2012-08-02 05:02:49.349011
3025	1-3025	29	96	970	18	7025	0	0	2012-08-02 05:02:49.349011	2012-08-02 05:02:49.349011
486	1-486	33	26	1361	17	2301	0	0	2012-08-02 04:44:10.495864	2012-08-02 05:02:50.177922
1666	1-1666	7	25	9479	10	6053	0	0	2012-08-02 04:53:02.10922	2012-08-02 05:02:50.177922
2386	1-2386	74	100	5058	17	9885	0	0	2012-08-02 04:58:18.40634	2012-08-02 05:02:50.177922
3027	1-3027	56	17	7390	26	6968	0	0	2012-08-02 05:02:50.177922	2012-08-02 05:02:50.177922
3030	1-3030	19	75	1767	46	5022	0	0	2012-08-02 05:02:51.39313	2012-08-02 05:02:51.39313
2475	1-2475	25	31	2102	72	3323	0	0	2012-08-02 04:58:57.736481	2012-08-02 05:02:51.801859
2780	1-2780	87	76	3959	24	5304	0	0	2012-08-02 05:01:07.971791	2012-08-02 05:02:52.630419
3036	1-3036	26	97	546	89	520	0	0	2012-08-02 05:02:53.835058	2012-08-02 05:02:53.835058
2880	1-2880	80	65	5756	87	6147	0	0	2012-08-02 05:01:49.904224	2012-08-02 05:02:54.243824
3037	1-3037	93	39	560	65	2086	0	3	2012-08-02 05:02:54.243824	2012-08-02 05:02:54.243824
1622	1-1622	64	32	6657	92	2711	0	10	2012-08-02 04:52:40.752372	2012-08-02 05:02:55.834358
2678	1-2678	94	1	6899	91	7751	0	0	2012-08-02 05:00:24.812471	2012-08-02 05:02:56.652062
3043	1-3043	70	7	921	10	4204	0	0	2012-08-02 05:02:56.652062	2012-08-02 05:02:56.652062
1903	1-1903	27	39	4377	69	6730	0	0	2012-08-02 04:54:48.717909	2012-08-02 05:02:57.06104
2971	1-2971	36	10	9816	1	9716	0	0	2012-08-02 05:02:27.493789	2012-08-02 05:02:57.06104
414	1-414	74	76	8198	27	3133	0	0	2012-08-02 04:44:04.677727	2012-08-02 05:02:57.06104
3044	1-3044	77	48	2574	39	7892	0	0	2012-08-02 05:02:57.06104	2012-08-02 05:02:57.06104
2500	1-2500	39	36	3687	58	5366	0	7	2012-08-02 04:59:08.659183	2012-08-02 05:02:57.492152
2382	1-2382	8	89	7889	12	9675	0	0	2012-08-02 04:58:16.672161	2012-08-02 05:02:57.889751
3046	1-3046	57	44	909	89	4299	0	16	2012-08-02 05:02:57.889751	2012-08-02 05:02:57.889751
732	1-732	49	58	3403	72	5883	0	0	2012-08-02 04:45:07.889925	2012-08-02 05:03:00.729277
2925	1-2925	79	11	456	39	8275	0	0	2012-08-02 05:02:08.622954	2012-08-02 05:03:02.739681
2647	1-2647	45	94	7609	30	5060	0	0	2012-08-02 05:00:11.784526	2012-08-02 05:03:02.739681
3058	1-3058	84	30	785	84	4669	0	11	2012-08-02 05:03:02.739681	2012-08-02 05:03:02.739681
2897	1-2897	43	71	5015	66	1596	0	0	2012-08-02 05:01:56.952565	2012-08-02 05:03:04.354577
2116	1-2116	85	71	7531	92	2157	0	2	2012-08-02 04:56:21.098957	2012-08-02 05:03:04.354577
2617	1-2617	63	71	3112	57	5253	0	5	2012-08-02 04:59:58.855954	2012-08-02 05:03:04.354577
2444	1-2444	39	40	4624	100	1898	0	6	2012-08-02 04:58:43.885628	2012-08-02 05:03:04.354577
2936	1-2936	58	53	6480	19	3270	0	0	2012-08-02 05:02:13.185839	2012-08-02 05:03:07.63729
2012	1-2012	24	49	6553	46	6687	0	0	2012-08-02 04:55:36.008434	2012-08-02 05:03:09.250277
1520	1-1520	44	49	3248	94	1405	0	4	2012-08-02 04:51:49.255492	2012-08-02 05:03:09.250277
2926	1-2926	49	41	6976	49	9777	0	0	2012-08-02 05:02:09.031601	2012-08-02 05:03:09.250277
3074	1-3074	60	34	170	41	7014	0	3	2012-08-02 05:03:09.250277	2012-08-02 05:03:09.250277
537	1-537	96	46	9535	48	4670	734	0	2012-08-02 04:44:19.739112	2012-08-02 05:03:09.250277
3077	1-3077	64	59	852	7	690	0	0	2012-08-02 05:03:10.488029	2012-08-02 05:03:10.488029
2960	1-2960	32	15	5145	82	7859	0	0	2012-08-02 05:02:23.063517	2012-08-02 05:03:11.327523
2949	1-2949	84	86	1413	79	2011	0	9	2012-08-02 05:02:18.55511	2012-08-02 05:03:12.929741
3062	1-3062	99	77	2374	71	9933	0	0	2012-08-02 05:03:04.354577	2012-08-02 05:03:14.54245
2381	1-2381	60	100	8779	23	9359	0	4	2012-08-02 04:58:16.241223	2012-08-02 05:03:14.54245
2736	1-2736	8	98	7654	86	4545	0	0	2012-08-02 05:00:49.496231	2012-08-02 05:03:15.360152
1291	1-1291	6	98	6642	92	2432	0	4	2012-08-02 04:49:48.872086	2012-08-02 05:03:15.360152
2677	1-2677	57	31	7013	28	3173	0	0	2012-08-02 05:00:24.392844	2012-08-02 05:03:15.360152
3089	1-3089	29	97	1648	98	9295	0	0	2012-08-02 05:03:15.360152	2012-08-02 05:03:15.360152
1921	1-1921	83	9	5012	4	6611	0	10	2012-08-02 04:54:56.252921	2012-08-02 05:03:16.166885
2844	1-2844	99	86	965	48	3282	0	0	2012-08-02 05:01:34.8657	2012-08-02 05:03:16.166885
3091	1-3091	32	4	2731	84	5621	0	0	2012-08-02 05:03:16.166885	2012-08-02 05:03:16.166885
2717	1-2717	39	94	4351	69	4171	0	0	2012-08-02 05:00:41.440445	2012-08-02 05:03:16.984729
2161	1-2161	46	94	6855	96	4713	0	3	2012-08-02 04:56:40.445914	2012-08-02 05:03:16.984729
2996	1-2996	97	23	6408	12	3332	0	0	2012-08-02 05:02:37.693705	2012-08-02 05:03:16.984729
3012	1-3012	99	12	6317	96	9716	0	0	2012-08-02 05:02:44.167298	2012-08-02 05:03:16.984729
1656	1-1656	93	94	8750	96	3542	0	0	2012-08-02 04:52:56.485977	2012-08-02 05:03:16.984729
3093	1-3093	100	44	2305	94	7931	0	4	2012-08-02 05:03:16.984729	2012-08-02 05:03:16.984729
2827	1-2827	13	54	2104	42	5207	0	0	2012-08-02 05:01:27.683908	2012-08-02 05:03:18.619922
3099	1-3099	95	88	850	24	3436	0	0	2012-08-02 05:03:19.437746	2012-08-02 05:03:19.437746
2713	1-2713	26	69	3330	21	5951	0	0	2012-08-02 05:00:39.72777	2012-08-02 05:03:20.244072
2853	1-2853	73	9	7422	91	9040	0	0	2012-08-02 05:01:38.668588	2012-08-02 05:03:20.653182
2110	1-2110	54	29	6193	24	3490	0	0	2012-08-02 04:56:18.447779	2012-08-02 05:03:23.514099
3076	1-3076	44	90	4687	67	3441	0	0	2012-08-02 05:03:10.101642	2012-08-02 05:03:23.514099
273	1-273	56	36	9122	72	6952	0	0	2012-08-02 04:43:58.043941	2012-08-02 05:03:24.331797
3111	1-3111	79	46	2048	36	8980	0	0	2012-08-02 05:03:24.331797	2012-08-02 05:03:24.331797
1045	1-1045	44	33	6247	76	3159	0	0	2012-08-02 04:47:35.82765	2012-08-02 05:03:24.740824
2533	1-2533	4	14	5221	2	8884	0	0	2012-08-02 04:59:22.867103	2012-08-02 05:03:25.536203
3115	1-3115	7	59	3138	13	4306	0	0	2012-08-02 05:03:25.944938	2012-08-02 05:03:25.944938
3118	1-3118	82	51	836	55	2314	0	0	2012-08-02 05:03:27.171667	2012-08-02 05:03:27.171667
2649	1-2649	10	84	6512	9	5631	0	0	2012-08-02 05:00:12.635607	2012-08-02 05:03:28.01093
3120	1-3120	89	83	213	84	7366	0	9	2012-08-02 05:03:28.01093	2012-08-02 05:03:28.01093
3039	1-3039	66	95	2211	52	4688	0	0	2012-08-02 05:02:55.039289	2012-08-02 05:03:28.852735
2449	1-2449	66	39	5427	76	6161	0	0	2012-08-02 04:58:46.106393	2012-08-02 05:03:28.852735
2980	1-2980	70	2	5037	43	6932	0	4	2012-08-02 05:02:31.18389	2012-08-02 05:03:28.852735
3109	1-3109	10	42	3465	90	5087	0	0	2012-08-02 05:03:23.514099	2012-08-02 05:03:28.852735
833	1-833	100	39	5518	17	4767	0	0	2012-08-02 04:45:47.446833	2012-08-02 05:03:28.852735
334	1-334	84	90	4382	14	3566	10	0	2012-08-02 04:44:00.144891	2012-08-02 05:03:28.852735
2993	1-2993	40	48	3524	61	3380	0	0	2012-08-02 05:02:36.478128	2012-08-02 05:03:29.261695
3122	1-3122	28	43	1287	39	8365	0	4	2012-08-02 05:03:28.852735	2012-08-02 05:03:29.261695
1191	1-1191	22	83	2627	39	917	0	8	2012-08-02 04:48:53.650214	2012-08-02 05:03:29.261695
2083	1-2083	87	48	8123	43	2341	0	0	2012-08-02 04:56:06.912255	2012-08-02 05:03:29.261695
3123	1-3123	72	42	318	48	7991	0	0	2012-08-02 05:03:29.261695	2012-08-02 05:03:29.261695
3124	1-3124	63	51	1092	96	9015	0	0	2012-08-02 05:03:29.681719	2012-08-02 05:03:29.681719
3112	1-3112	88	23	2065	47	5844	0	0	2012-08-02 05:03:24.740824	2012-08-02 05:03:30.079431
3056	1-3056	71	92	957	59	5969	0	0	2012-08-02 05:03:01.944347	2012-08-02 05:03:30.079431
3125	1-3125	84	5	2083	81	7686	0	2	2012-08-02 05:03:30.079431	2012-08-02 05:03:30.079431
2347	1-2347	16	60	724	29	538	0	0	2012-08-02 04:58:01.43604	2012-08-02 05:03:30.477471
3126	1-3126	17	67	939	25	2190	0	4	2012-08-02 05:03:30.477471	2012-08-02 05:03:30.477471
1796	1-1796	82	59	8614	54	7591	0	0	2012-08-02 04:54:00.584378	2012-08-02 05:03:31.295058
1998	1-1998	16	60	7209	75	5634	0	0	2012-08-02 04:55:30.30753	2012-08-02 05:03:31.295058
2868	1-2868	36	96	4525	44	4245	0	0	2012-08-02 05:01:44.910772	2012-08-02 05:03:31.295058
3128	1-3128	97	72	1087	96	5486	0	0	2012-08-02 05:03:31.295058	2012-08-02 05:03:31.295058
2479	1-2479	58	63	7478	25	7243	0	0	2012-08-02 04:58:59.482657	2012-08-02 05:03:31.726448
1695	1-1695	94	83	3795	5	3126	0	0	2012-08-02 04:53:14.881489	2012-08-02 05:03:32.53258
2724	1-2724	53	59	4497	5	5830	0	6	2012-08-02 05:00:44.401245	2012-08-02 05:03:32.53258
2636	1-2636	23	9	2014	5	5560	0	2	2012-08-02 05:00:07.166661	2012-08-02 05:03:32.53258
1055	1-1055	2	11	6096	20	7994	0	0	2012-08-02 04:47:42.301015	2012-08-02 05:03:32.53258
1298	1-1298	47	67	6574	29	5255	0	0	2012-08-02 04:49:53.665973	2012-08-02 05:03:32.930771
3101	1-3101	68	33	2173	58	5023	0	3	2012-08-02 05:03:20.244072	2012-08-02 05:03:33.33962
2975	1-2975	73	67	1846	50	6897	0	0	2012-08-02 05:02:29.140247	2012-08-02 05:03:33.737257
3135	1-3135	29	53	1495	84	3726	0	0	2012-08-02 05:03:34.135026	2012-08-02 05:03:34.135026
1820	1-1820	66	45	9449	95	8637	0	0	2012-08-02 04:54:11.607066	2012-08-02 05:03:35.328242
2419	1-2419	34	39	5366	95	4016	0	0	2012-08-02 04:58:32.881635	2012-08-02 05:03:35.328242
2824	1-2824	86	45	1459	97	6162	0	0	2012-08-02 05:01:26.423961	2012-08-02 05:03:35.328242
699	1-699	59	52	5764	95	1439	0	0	2012-08-02 04:44:58.570105	2012-08-02 05:03:35.328242
3138	1-3138	62	95	2527	45	9987	0	0	2012-08-02 05:03:35.328242	2012-08-02 05:03:35.328242
985	1-985	61	45	8512	62	4510	1817	0	2012-08-02 04:47:04.160538	2012-08-02 05:03:35.328242
2886	1-2886	57	33	6421	58	3876	0	0	2012-08-02 05:01:52.411904	2012-08-02 05:03:35.759462
3054	1-3054	2	87	5271	19	7311	0	0	2012-08-02 05:03:01.137827	2012-08-02 05:03:35.759462
2579	1-2579	26	46	9789	60	4683	0	0	2012-08-02 04:59:42.712993	2012-08-02 05:03:35.759462
1413	1-1413	64	58	6595	69	4241	1264	0	2012-08-02 04:50:51.937606	2012-08-02 05:03:35.759462
2866	1-2866	32	17	2396	90	3877	0	0	2012-08-02 05:01:44.081891	2012-08-02 05:03:37.781213
1744	1-1744	81	79	6670	4	7203	0	0	2012-08-02 04:53:37.28941	2012-08-02 05:03:37.781213
2111	1-2111	96	9	5477	85	5590	0	0	2012-08-02 04:56:18.88922	2012-08-02 05:03:37.781213
1722	1-1722	61	38	4186	72	3970	0	0	2012-08-02 04:53:27.302202	2012-08-02 05:03:37.781213
3139	1-3139	67	68	3690	46	8401	0	5	2012-08-02 05:03:35.759462	2012-08-02 05:03:38.985453
2715	1-2715	59	37	7800	46	9182	0	7	2012-08-02 05:00:40.567448	2012-08-02 05:03:38.985453
2614	1-2614	56	79	5799	97	6583	0	0	2012-08-02 04:59:57.596589	2012-08-02 05:03:40.554418
3005	1-3005	67	33	1183	20	895	0	4	2012-08-02 05:02:41.305966	2012-08-02 05:03:40.974809
3152	1-3152	31	6	2811	34	1572	0	0	2012-08-02 05:03:40.974809	2012-08-02 05:03:40.974809
431	1-431	75	1	3423	17	9772	0	0	2012-08-02 04:44:06.045784	2012-08-02 05:03:41.383264
2796	1-2796	81	94	4703	23	6609	0	0	2012-08-02 05:01:14.668794	2012-08-02 05:03:41.781072
3061	1-3061	70	65	5676	60	7267	0	0	2012-08-02 05:03:03.957277	2012-08-02 05:03:41.781072
2970	1-2970	30	99	4753	63	9425	0	4	2012-08-02 05:02:27.096169	2012-08-02 05:03:43.007286
2704	1-2704	47	20	4783	6	9625	0	0	2012-08-02 05:00:35.827545	2012-08-02 05:03:43.007286
3154	1-3154	91	2	2583	94	8135	0	0	2012-08-02 05:03:41.781072	2012-08-02 05:03:43.83615
2898	1-2898	16	10	856	88	8500	0	0	2012-08-02 05:01:57.372739	2012-08-02 05:03:43.83615
2910	1-2910	50	28	2237	80	8808	0	5	2012-08-02 05:02:02.344343	2012-08-02 05:03:43.83615
3159	1-3159	77	88	2811	75	5158	0	0	2012-08-02 05:03:43.83615	2012-08-02 05:03:43.83615
3164	1-3164	62	29	1009	51	5208	0	0	2012-08-02 05:03:45.858198	2012-08-02 05:03:45.858198
1889	1-1889	63	28	6471	29	4403	0	0	2012-08-02 04:54:42.741782	2012-08-02 05:03:49.1507
3173	1-3173	94	31	935	30	6559	0	0	2012-08-02 05:03:49.55398	2012-08-02 05:03:49.55398
3176	1-3176	64	81	1024	26	9136	0	0	2012-08-02 05:03:50.770866	2012-08-02 05:03:50.770866
2650	1-2650	45	70	6067	48	9533	0	0	2012-08-02 05:00:13.05524	2012-08-02 05:03:51.567005
1845	1-1845	34	91	6946	48	8126	0	3	2012-08-02 04:54:22.566651	2012-08-02 05:03:51.567005
399	1-399	42	91	6995	48	7821	0	0	2012-08-02 04:44:03.636911	2012-08-02 05:03:51.567005
2481	1-2481	2	64	6213	100	5917	0	5	2012-08-02 04:59:00.355622	2012-08-02 05:03:52.750688
2990	1-2990	12	66	1939	100	2953	0	0	2012-08-02 05:02:35.229952	2012-08-02 05:03:52.750688
3053	1-3053	44	72	4002	35	5426	0	0	2012-08-02 05:03:00.729277	2012-08-02 05:03:52.750688
2759	1-2759	3	13	8722	87	5082	0	0	2012-08-02 05:00:59.230468	2012-08-02 05:03:53.559694
3185	1-3185	78	42	419	75	1217	0	0	2012-08-02 05:03:54.39071	2012-08-02 05:03:54.39071
3028	1-3028	85	25	2236	7	1724	0	0	2012-08-02 05:02:50.586846	2012-08-02 05:03:55.208719
3097	1-3097	100	96	1141	59	5958	0	0	2012-08-02 05:03:18.619922	2012-08-02 05:03:56.038449
3189	1-3189	20	59	171	22	2941	0	0	2012-08-02 05:03:56.038449	2012-08-02 05:03:56.038449
1561	1-1561	72	11	5283	16	7009	0	7	2012-08-02 04:52:09.992965	2012-08-02 05:03:58.016933
1505	1-1505	86	11	7438	44	6541	0	0	2012-08-02 04:51:41.266883	2012-08-02 05:03:58.016933
1182	1-1182	75	99	6084	25	3659	0	0	2012-08-02 04:48:48.847009	2012-08-02 05:03:58.016933
2788	1-2788	72	11	2305	57	8347	0	0	2012-08-02 05:01:11.354469	2012-08-02 05:03:58.016933
1571	1-1571	94	11	3806	45	2293	0	0	2012-08-02 04:52:16.134816	2012-08-02 05:03:58.016933
3070	1-3070	28	59	3491	53	6456	0	4	2012-08-02 05:03:07.63729	2012-08-02 05:03:59.234081
2615	1-2615	47	89	3442	52	8251	0	0	2012-08-02 04:59:58.01639	2012-08-02 05:03:59.234081
2425	1-2425	44	82	7824	89	6009	0	0	2012-08-02 04:58:35.588934	2012-08-02 05:03:59.234081
2026	1-2026	52	82	5585	87	3011	0	0	2012-08-02 04:55:42.195558	2012-08-02 05:03:59.234081
3197	1-3197	6	53	4524	82	9790	0	10	2012-08-02 05:03:59.234081	2012-08-02 05:03:59.234081
2985	1-2985	79	23	9729	95	9490	0	0	2012-08-02 05:02:33.183921	2012-08-02 05:03:59.632599
3078	1-3078	10	79	3252	16	8356	0	6	2012-08-02 05:03:10.919257	2012-08-02 05:03:59.632599
1700	1-1700	31	79	1280	10	1907	0	0	2012-08-02 04:53:17.124226	2012-08-02 05:03:59.632599
2645	1-2645	46	57	2804	78	2816	0	0	2012-08-02 05:00:10.93431	2012-08-02 05:04:00.848431
2388	1-2388	76	57	6286	50	5356	0	0	2012-08-02 04:58:19.290432	2012-08-02 05:04:00.848431
2674	1-2674	52	57	2044	80	1762	0	0	2012-08-02 05:00:23.121729	2012-08-02 05:04:00.848431
3201	1-3201	14	78	150	57	2826	0	0	2012-08-02 05:04:00.848431	2012-08-02 05:04:00.848431
1996	1-1996	57	58	9718	69	8846	0	3	2012-08-02 04:55:29.423921	2012-08-02 05:04:02.041841
2930	1-2930	86	51	4202	91	1295	0	0	2012-08-02 05:02:10.711047	2012-08-02 05:04:02.041841
3183	1-3183	61	61	4536	28	6503	0	0	2012-08-02 05:03:53.559694	2012-08-02 05:04:02.041841
2864	1-2864	70	61	1591	16	4997	0	0	2012-08-02 05:01:43.253202	2012-08-02 05:04:02.041841
3204	1-3204	48	27	2942	61	9305	0	11	2012-08-02 05:04:02.041841	2012-08-02 05:04:02.041841
1679	1-1679	66	18	5790	27	5584	135	2	2012-08-02 04:53:08.087384	2012-08-02 05:04:02.041841
2524	1-2524	77	27	5380	61	2291	1647	3	2012-08-02 04:59:19.033744	2012-08-02 05:04:02.041841
2883	1-2883	26	92	5348	22	5402	0	0	2012-08-02 05:01:51.174685	2012-08-02 05:04:02.848538
3206	1-3206	64	22	9958	63	8759	0	0	2012-08-02 05:04:02.848538	2012-08-02 05:04:02.848538
931	1-931	33	47	7326	6	5728	0	0	2012-08-02 04:46:33.369443	2012-08-02 05:04:03.677246
2598	1-2598	75	87	4591	76	5890	0	0	2012-08-02 04:59:50.833674	2012-08-02 05:04:04.164126
1513	1-1513	18	26	2760	25	4399	0	2	2012-08-02 04:51:45.26629	2012-08-02 05:04:04.164126
2836	1-2836	5	17	7246	53	9893	0	0	2012-08-02 05:01:31.495758	2012-08-02 05:04:04.164126
3209	1-3209	64	27	242	95	6398	0	0	2012-08-02 05:04:04.164126	2012-08-02 05:04:04.164126
3045	1-3045	100	97	4730	36	6034	0	0	2012-08-02 05:02:57.492152	2012-08-02 05:04:05.408919
2451	1-2451	53	69	4766	12	5695	0	0	2012-08-02 04:58:46.967726	2012-08-02 05:04:05.806604
2891	1-2891	98	55	4639	28	7451	0	7	2012-08-02 05:01:54.466818	2012-08-02 05:04:06.635161
2394	1-2394	9	18	7695	99	5435	0	0	2012-08-02 04:58:21.897563	2012-08-02 05:04:06.635161
3215	1-3215	32	13	311	18	7143	0	0	2012-08-02 05:04:06.635161	2012-08-02 05:04:06.635161
2697	1-2697	24	14	2687	41	533	0	0	2012-08-02 05:00:32.888208	2012-08-02 05:04:07.044529
3219	1-3219	80	9	317	61	2374	0	0	2012-08-02 05:04:08.248435	2012-08-02 05:04:08.248435
2121	1-2121	58	1	5688	32	9260	0	4	2012-08-02 04:56:23.241878	2012-08-02 05:04:08.657498
2955	1-2955	82	92	6714	94	7140	0	0	2012-08-02 05:02:21.008042	2012-08-02 05:04:08.657498
2659	1-2659	82	1	7215	55	7351	0	0	2012-08-02 05:00:16.822634	2012-08-02 05:04:08.657498
2680	1-2680	21	85	7311	47	5885	0	0	2012-08-02 05:00:25.696413	2012-08-02 05:04:09.066657
3194	1-3194	49	25	5902	11	9826	0	0	2012-08-02 05:03:58.016933	2012-08-02 05:04:10.788593
3033	1-3033	96	29	2831	76	8638	0	3	2012-08-02 05:02:52.630419	2012-08-02 05:04:16.811242
3212	1-3212	98	80	2179	70	9695	0	0	2012-08-02 05:04:05.408919	2012-08-02 05:04:16.811242
2441	1-2441	43	93	8579	20	6136	0	0	2012-08-02 04:58:42.581981	2012-08-02 05:04:16.811242
3073	1-3073	72	100	2970	60	2856	0	0	2012-08-02 05:03:08.852438	2012-08-02 05:04:16.811242
2385	1-2385	85	20	7810	100	8250	0	5	2012-08-02 04:58:17.964558	2012-08-02 05:04:18.811052
3181	1-3181	28	100	3839	64	5385	0	0	2012-08-02 05:03:52.750688	2012-08-02 05:04:18.811052
2596	1-2596	89	21	1581	5	6038	0	0	2012-08-02 04:59:49.971668	2012-08-02 05:04:20.039039
1958	1-1958	22	84	6497	86	6452	0	0	2012-08-02 04:55:12.618596	2012-08-02 05:04:20.039039
2833	1-2833	9	87	2056	40	7552	0	0	2012-08-02 05:01:30.202974	2012-08-02 05:04:20.436835
2819	1-2819	97	87	3891	59	9410	0	0	2012-08-02 05:01:24.357783	2012-08-02 05:04:20.436835
3250	1-3250	44	22	2526	87	6422	0	0	2012-08-02 05:04:20.436835	2012-08-02 05:04:20.436835
2548	1-2548	52	55	4412	6	1073	0	0	2012-08-02 04:59:29.343712	2012-08-02 05:04:22.061062
3254	1-3254	47	99	951	55	9174	0	0	2012-08-02 05:04:22.061062	2012-08-02 05:04:22.061062
1272	1-1272	93	38	1869	100	2703	0	0	2012-08-02 04:49:39.162141	2012-08-02 05:04:22.470067
3098	1-3098	95	71	4553	21	5000	0	0	2012-08-02 05:03:19.039951	2012-08-02 05:04:22.470067
2799	1-2799	61	95	8742	17	9849	0	0	2012-08-02 05:01:15.917294	2012-08-02 05:04:27.286656
3268	1-3268	3	33	206	7	4621	0	3	2012-08-02 05:04:27.684457	2012-08-02 05:04:27.684457
2585	1-2585	20	30	4652	22	8514	0	0	2012-08-02 04:59:45.298244	2012-08-02 05:04:28.89999
3127	1-3127	63	40	708	2	357	0	0	2012-08-02 05:03:30.897327	2012-08-02 05:04:28.89999
3271	1-3271	26	2	1754	30	1650	0	3	2012-08-02 05:04:28.89999	2012-08-02 05:04:28.89999
504	1-504	61	77	7762	75	2419	1057	0	2012-08-02 04:44:12.40894	2012-08-02 05:04:28.89999
3157	1-3157	8	6	3022	99	6019	0	0	2012-08-02 05:03:43.007286	2012-08-02 05:04:29.319558
3071	1-3071	94	45	3844	36	8235	0	0	2012-08-02 05:03:08.046194	2012-08-02 05:04:30.9329
2911	1-2911	20	72	4735	95	3713	0	0	2012-08-02 05:02:02.753314	2012-08-02 05:04:30.9329
3244	1-3244	21	45	3899	97	239	0	8	2012-08-02 05:04:18.026641	2012-08-02 05:04:30.9329
2242	1-2242	89	85	5800	90	3701	0	0	2012-08-02 04:57:15.714322	2012-08-02 05:04:31.739236
3278	1-3278	17	33	236	85	6934	0	0	2012-08-02 05:04:31.739236	2012-08-02 05:04:31.739236
3285	1-3285	91	9	334	70	8467	0	0	2012-08-02 05:04:34.578588	2012-08-02 05:04:34.578588
3289	1-3289	100	88	1798	47	1866	0	0	2012-08-02 05:04:36.180485	2012-08-02 05:04:36.180485
3147	1-3147	12	46	4266	7	7399	0	0	2012-08-02 05:03:38.985453	2012-08-02 05:04:37.771315
2889	1-2889	13	46	9982	87	5891	0	9	2012-08-02 05:01:53.671652	2012-08-02 05:04:37.771315
3114	1-3114	27	2	3594	83	7653	0	0	2012-08-02 05:03:25.536203	2012-08-02 05:04:37.771315
2966	1-2966	87	68	2645	2	4788	0	0	2012-08-02 05:02:25.449876	2012-08-02 05:04:38.566976
3084	1-3084	34	65	1379	80	2687	0	0	2012-08-02 05:03:13.327288	2012-08-02 05:04:39.373687
3297	1-3297	10	59	2579	99	4392	0	0	2012-08-02 05:04:39.373687	2012-08-02 05:04:39.373687
2134	1-2134	42	71	2346	10	1563	0	0	2012-08-02 04:56:28.988806	2012-08-02 05:04:40.589326
3225	1-3225	96	11	3172	6	6157	0	0	2012-08-02 05:04:10.788593	2012-08-02 05:04:41.406215
3302	1-3302	75	64	312	41	8058	0	3	2012-08-02 05:04:41.406215	2012-08-02 05:04:41.406215
2539	1-2539	84	11	3630	12	3496	0	0	2012-08-02 04:59:25.432779	2012-08-02 05:04:42.201903
1829	1-1829	58	96	1680	90	3168	0	0	2012-08-02 04:54:15.263399	2012-08-02 05:04:42.201903
2091	1-2091	46	11	2750	14	7219	0	0	2012-08-02 04:56:10.039281	2012-08-02 05:04:42.201903
1883	1-1883	15	12	3555	90	3349	0	0	2012-08-02 04:54:39.957878	2012-08-02 05:04:42.201903
3304	1-3304	77	3	2543	11	8847	0	4	2012-08-02 05:04:42.201903	2012-08-02 05:04:42.201903
3144	1-3144	38	90	9568	38	6989	1571	0	2012-08-02 05:03:37.781213	2012-08-02 05:04:42.201903
3131	1-3131	13	5	8434	11	6742	0	0	2012-08-02 05:03:32.53258	2012-08-02 05:04:42.633272
1803	1-1803	86	1	5676	22	6425	0	0	2012-08-02 04:54:03.810805	2012-08-02 05:04:44.643719
2815	1-2815	56	9	2421	53	5217	0	0	2012-08-02 05:01:22.678537	2012-08-02 05:04:44.643719
3174	1-3174	96	16	8294	23	5862	0	0	2012-08-02 05:03:49.953122	2012-08-02 05:04:45.019713
3311	1-3311	50	54	641	32	6843	0	0	2012-08-02 05:04:45.019713	2012-08-02 05:04:45.019713
3307	1-3307	9	53	4738	85	7965	0	0	2012-08-02 05:04:43.43964	2012-08-02 05:04:45.428471
1946	1-1946	95	66	8445	48	6221	0	0	2012-08-02 04:55:07.646511	2012-08-02 05:04:45.837282
3169	1-3169	85	18	2786	27	9024	0	9	2012-08-02 05:03:47.923739	2012-08-02 05:04:45.837282
2266	1-2266	75	66	2149	27	1517	0	0	2012-08-02 04:57:26.165714	2012-08-02 05:04:45.837282
1824	1-1824	18	76	7293	18	2223	65	0	2012-08-02 04:54:13.010101	2012-08-02 05:04:45.837282
3178	1-3178	29	48	8301	91	8620	0	2	2012-08-02 05:03:51.567005	2012-08-02 05:04:47.883251
3142	1-3142	76	20	3593	42	1201	0	0	2012-08-02 05:03:36.963694	2012-08-02 05:04:47.883251
3318	1-3318	2	42	1637	92	5322	0	5	2012-08-02 05:04:47.883251	2012-08-02 05:04:47.883251
2405	1-2405	48	9	5097	48	7097	0	0	2012-08-02 04:58:26.727646	2012-08-02 05:04:48.690107
3320	1-3320	10	2	3752	9	5884	0	0	2012-08-02 05:04:48.690107	2012-08-02 05:04:48.690107
1698	1-1698	25	11	8435	98	8962	0	0	2012-08-02 04:53:16.207405	2012-08-02 05:04:50.325221
3276	1-3276	9	87	553	45	8287	0	0	2012-08-02 05:04:30.9329	2012-08-02 05:04:50.325221
3325	1-3325	56	55	4117	4	8766	0	4	2012-08-02 05:04:50.733847	2012-08-02 05:04:50.733847
2940	1-2940	10	76	6282	94	3341	0	0	2012-08-02 05:02:14.831817	2012-08-02 05:04:52.744334
2927	1-2927	18	69	1736	68	2312	0	0	2012-08-02 05:02:09.451423	2012-08-02 05:04:55.174842
2884	1-2884	47	1	2454	65	2864	0	0	2012-08-02 05:01:51.594446	2012-08-02 05:04:55.174842
3264	1-3264	7	27	992	26	424	0	0	2012-08-02 05:04:26.082382	2012-08-02 05:04:55.174842
2988	1-2988	19	69	7311	27	9458	0	0	2012-08-02 05:02:34.411944	2012-08-02 05:04:55.174842
2732	1-2732	80	74	5034	25	7056	0	0	2012-08-02 05:00:47.806019	2012-08-02 05:04:55.174842
3336	1-3336	28	25	1949	69	9641	0	3	2012-08-02 05:04:55.174842	2012-08-02 05:04:55.174842
3293	1-3293	42	83	2796	46	5683	0	11	2012-08-02 05:04:37.771315	2012-08-02 05:04:56.843444
1936	1-1936	97	40	4442	17	593	0	0	2012-08-02 04:55:03.093022	2012-08-02 05:04:56.843444
3340	1-3340	6	33	517	40	7409	0	0	2012-08-02 05:04:56.843444	2012-08-02 05:04:56.843444
2331	1-2331	40	79	6658	33	5736	2249	0	2012-08-02 04:57:54.518266	2012-08-02 05:04:56.843444
1502	1-1502	9	61	4023	88	8015	0	0	2012-08-02 04:51:40.184268	2012-08-02 05:04:58.026139
3343	1-3343	99	88	6716	61	4036	0	0	2012-08-02 05:04:58.026139	2012-08-02 05:04:58.026139
3347	1-3347	25	49	15	40	9242	0	0	2012-08-02 05:04:59.627689	2012-08-02 05:04:59.627689
3350	1-3350	26	92	1541	53	1976	0	0	2012-08-02 05:05:00.854224	2012-08-02 05:05:00.854224
3352	1-3352	1	88	1681	98	9811	0	5	2012-08-02 05:05:01.671839	2012-08-02 05:05:01.671839
2251	1-2251	86	49	5109	43	4188	0	0	2012-08-02 04:57:19.636584	2012-08-02 05:05:02.875754
3191	1-3191	85	12	4663	92	8590	0	0	2012-08-02 05:03:56.835116	2012-08-02 05:05:03.273434
3356	1-3356	36	20	493	73	2489	0	7	2012-08-02 05:05:03.273434	2012-08-02 05:05:03.273434
3220	1-3220	71	94	2260	1	7937	0	0	2012-08-02 05:04:08.657498	2012-08-02 05:05:03.671329
1380	1-1380	16	7	6426	85	8272	157	0	2012-08-02 04:50:36.205914	2012-08-02 05:05:03.671329
2874	1-2874	88	71	5587	82	1872	0	0	2012-08-02 05:01:47.396529	2012-08-02 05:05:04.864936
3360	1-3360	13	82	1432	18	5294	0	0	2012-08-02 05:05:04.864936	2012-08-02 05:05:04.864936
3263	1-3263	81	89	1247	71	9812	0	0	2012-08-02 05:04:25.685038	2012-08-02 05:05:05.750171
3119	1-3119	11	23	3426	33	3477	0	0	2012-08-02 05:03:27.613529	2012-08-02 05:05:05.750171
1603	1-1603	43	73	4273	71	3527	0	0	2012-08-02 04:52:31.681189	2012-08-02 05:05:05.750171
3008	1-3008	39	23	2775	73	2971	0	0	2012-08-02 05:02:42.565638	2012-08-02 05:05:05.750171
3198	1-3198	92	17	4178	79	6588	0	0	2012-08-02 05:03:59.632599	2012-08-02 05:05:06.976718
2881	1-2881	93	84	2425	56	1195	0	3	2012-08-02 05:01:50.312802	2012-08-02 05:05:07.363069
3172	1-3172	8	76	6849	61	7911	0	5	2012-08-02 05:03:49.1507	2012-08-02 05:05:07.760849
2951	1-2951	15	74	7282	9	7946	0	10	2012-08-02 05:02:19.361927	2012-08-02 05:05:07.760849
3367	1-3367	55	61	1138	74	4212	0	0	2012-08-02 05:05:07.760849	2012-08-02 05:05:07.760849
2711	1-2711	42	79	6304	37	8779	0	0	2012-08-02 05:00:38.866315	2012-08-02 05:05:08.578536
3229	1-3229	24	43	3344	22	7657	0	0	2012-08-02 05:04:12.391785	2012-08-02 05:05:08.578536
3365	1-3365	47	37	5345	68	8434	0	0	2012-08-02 05:05:06.976718	2012-08-02 05:05:08.578536
3369	1-3369	73	68	9459	43	7709	0	0	2012-08-02 05:05:08.578536	2012-08-02 05:05:08.578536
3088	1-3088	27	48	3340	10	7048	0	7	2012-08-02 05:03:14.951483	2012-08-02 05:05:08.98762
2401	1-2401	5	84	5647	23	6186	0	0	2012-08-02 04:58:24.935693	2012-08-02 05:05:08.98762
2253	1-2253	51	65	7333	35	2241	0	0	2012-08-02 04:57:20.52036	2012-08-02 05:05:08.98762
3366	1-3366	78	97	2370	84	6830	0	0	2012-08-02 05:05:07.363069	2012-08-02 05:05:08.98762
327	1-327	74	58	5210	40	6821	0	0	2012-08-02 04:43:59.856619	2012-08-02 05:05:08.98762
3251	1-3251	58	97	1576	65	3790	0	0	2012-08-02 05:04:20.856944	2012-08-02 05:05:08.98762
3057	1-3057	69	45	5168	85	3919	0	0	2012-08-02 05:03:02.353265	2012-08-02 05:05:09.816388
3372	1-3372	99	79	1044	45	8061	0	0	2012-08-02 05:05:09.816388	2012-08-02 05:05:10.225225
2276	1-2276	36	77	7004	81	4471	0	2	2012-08-02 04:57:30.586939	2012-08-02 05:05:10.720382
1510	1-1510	32	44	2474	42	3488	0	0	2012-08-02 04:51:43.189515	2012-08-02 05:05:10.720382
3374	1-3374	49	50	2228	44	3755	0	0	2012-08-02 05:05:10.720382	2012-08-02 05:05:10.720382
2870	1-2870	6	34	4795	33	7787	0	0	2012-08-02 05:01:45.750248	2012-08-02 05:05:11.167981
1970	1-1970	98	34	8049	89	8433	0	2	2012-08-02 04:55:17.95432	2012-08-02 05:05:11.167981
3375	1-3375	64	97	1353	34	5031	0	0	2012-08-02 05:05:11.167981	2012-08-02 05:05:11.167981
3319	1-3319	72	53	2063	78	821	0	0	2012-08-02 05:04:48.292268	2012-08-02 05:05:12.47042
3133	1-3133	37	27	6937	56	3431	0	0	2012-08-02 05:03:33.33962	2012-08-02 05:05:12.47042
3334	1-3334	25	58	5917	49	6607	0	0	2012-08-02 05:04:54.36843	2012-08-02 05:05:12.47042
2762	1-2762	100	96	5924	60	3572	0	0	2012-08-02 05:01:00.467671	2012-08-02 05:05:14.889524
3384	1-3384	12	3	2318	35	4247	0	0	2012-08-02 05:05:14.889524	2012-08-02 05:05:14.889524
2939	1-2939	89	30	4542	98	9809	0	0	2012-08-02 05:02:14.412014	2012-08-02 05:05:16.127046
2428	1-2428	66	29	8009	30	5453	1122	0	2012-08-02 04:58:36.903199	2012-08-02 05:05:16.127046
3186	1-3186	53	29	1298	14	5653	0	0	2012-08-02 05:03:54.810512	2012-08-02 05:05:16.535741
2667	1-2667	98	99	4935	65	1972	0	0	2012-08-02 05:00:20.183162	2012-08-02 05:05:16.535741
3162	1-3162	12	36	6782	65	6532	0	2	2012-08-02 05:03:45.051743	2012-08-02 05:05:16.535741
3362	1-3362	73	85	4748	23	8162	0	0	2012-08-02 05:05:05.750171	2012-08-02 05:05:18.955351
3395	1-3395	95	100	219	7	9402	0	0	2012-08-02 05:05:19.353164	2012-08-02 05:05:19.353164
3255	1-3255	29	100	1254	71	5575	0	0	2012-08-02 05:04:22.470067	2012-08-02 05:05:20.137914
3249	1-3249	93	2	3680	71	6341	0	0	2012-08-02 05:04:20.039039	2012-08-02 05:05:20.137914
1551	1-1551	84	74	6027	100	3240	0	0	2012-08-02 04:52:04.457226	2012-08-02 05:05:20.524468
2791	1-2791	25	47	8759	36	9393	0	0	2012-08-02 05:01:12.625085	2012-08-02 05:05:20.524468
3399	1-3399	88	51	1011	54	8982	0	6	2012-08-02 05:05:20.955505	2012-08-02 05:05:20.955505
981	1-981	13	57	6491	6	1564	0	0	2012-08-02 04:47:02.249127	2012-08-02 05:05:22.568362
3015	1-3015	17	95	2499	26	1927	0	0	2012-08-02 05:02:45.360396	2012-08-02 05:05:24.556891
3408	1-3408	2	74	1336	95	2911	0	0	2012-08-02 05:05:24.556891	2012-08-02 05:05:24.556891
3199	1-3199	30	97	1160	32	6801	0	8	2012-08-02 05:04:00.030649	2012-08-02 05:05:24.96593
3079	1-3079	64	89	2704	47	4646	0	0	2012-08-02 05:03:11.327523	2012-08-02 05:05:26.147891
3409	1-3409	17	5	1309	37	6476	0	0	2012-08-02 05:05:24.96593	2012-08-02 05:05:26.147891
3102	1-3102	15	21	8688	84	7018	0	0	2012-08-02 05:03:20.653182	2012-08-02 05:05:26.147891
3256	1-3256	86	89	1674	98	2347	0	11	2012-08-02 05:04:22.878762	2012-08-02 05:05:26.147891
3041	1-3041	27	100	2162	32	6694	0	0	2012-08-02 05:02:55.834358	2012-08-02 05:05:26.54592
3180	1-3180	24	40	7972	52	5229	0	0	2012-08-02 05:03:52.364047	2012-08-02 05:05:27.352488
2872	1-2872	85	100	6903	77	7040	0	0	2012-08-02 05:01:46.556526	2012-08-02 05:05:27.352488
3416	1-3416	92	9	1552	34	4116	0	0	2012-08-02 05:05:27.750218	2012-08-02 05:05:27.750218
3107	1-3107	2	63	6921	22	8142	0	0	2012-08-02 05:03:22.697031	2012-08-02 05:05:28.159395
3398	1-3398	19	86	301	74	6179	0	0	2012-08-02 05:05:20.524468	2012-08-02 05:05:28.159395
3146	1-3146	9	43	7827	38	4688	0	0	2012-08-02 05:03:38.587922	2012-08-02 05:05:29.385295
3141	1-3141	81	16	3584	85	4875	0	0	2012-08-02 05:03:36.56605	2012-08-02 05:05:29.385295
2654	1-2654	79	43	8958	79	5268	0	0	2012-08-02 05:00:14.734405	2012-08-02 05:05:29.385295
3422	1-3422	25	90	1313	11	2163	0	0	2012-08-02 05:05:30.191873	2012-08-02 05:05:30.191873
2036	1-2036	74	3	3960	14	3830	0	0	2012-08-02 04:55:46.581183	2012-08-02 05:05:30.590179
3383	1-3383	100	12	5107	3	6590	0	0	2012-08-02 05:05:14.491915	2012-08-02 05:05:30.590179
3423	1-3423	29	14	3003	12	8290	0	0	2012-08-02 05:05:30.590179	2012-08-02 05:05:30.590179
1748	1-1748	93	58	6706	50	8548	0	2	2012-08-02 04:53:39.134456	2012-08-02 05:05:30.987597
3134	1-3134	22	50	3759	85	4588	0	0	2012-08-02 05:03:33.737257	2012-08-02 05:05:30.987597
3354	1-3354	41	21	2740	54	7133	0	0	2012-08-02 05:05:02.466929	2012-08-02 05:05:31.783059
3358	1-3358	35	28	1795	29	4076	0	8	2012-08-02 05:05:04.069223	2012-08-02 05:05:32.20307
3387	1-3387	44	98	7031	47	8813	0	7	2012-08-02 05:05:16.127046	2012-08-02 05:05:33.81619
3431	1-3431	9	26	171	98	5643	0	7	2012-08-02 05:05:33.81619	2012-08-02 05:05:33.81619
3312	1-3312	27	14	2925	53	9593	0	0	2012-08-02 05:04:45.428471	2012-08-02 05:05:35.02039
3435	1-3435	18	22	1155	90	8123	0	0	2012-08-02 05:05:35.407078	2012-08-02 05:05:35.407078
3314	1-3314	19	90	5820	49	3974	0	0	2012-08-02 05:04:46.257594	2012-08-02 05:05:36.224518
1839	1-1839	63	70	7123	52	2207	0	0	2012-08-02 04:54:19.892951	2012-08-02 05:05:36.224518
3424	1-3424	66	85	1040	97	4593	0	0	2012-08-02 05:05:30.987597	2012-08-02 05:05:36.633398
1517	1-1517	93	63	6533	85	5281	0	0	2012-08-02 04:51:47.143863	2012-08-02 05:05:36.633398
2890	1-2890	99	80	9098	51	4671	0	0	2012-08-02 05:01:54.080347	2012-08-02 05:05:36.633398
3241	1-3241	22	60	3393	29	9779	0	5	2012-08-02 05:04:16.811242	2012-08-02 05:05:37.055386
518	1-518	44	60	4529	51	5577	0	0	2012-08-02 04:44:14.88528	2012-08-02 05:05:37.055386
3160	1-3160	6	3	1694	48	1782	0	0	2012-08-02 05:03:44.245122	2012-08-02 05:05:37.055386
3439	1-3439	4	48	1463	60	4281	0	0	2012-08-02 05:05:37.055386	2012-08-02 05:05:37.055386
1585	1-1585	60	35	7608	74	7380	0	0	2012-08-02 04:52:23.804075	2012-08-02 05:05:37.464449
3330	1-3330	69	86	1337	58	4730	0	4	2012-08-02 05:04:52.744334	2012-08-02 05:05:37.873127
3357	1-3357	20	1	1472	61	1357	0	8	2012-08-02 05:05:03.671329	2012-08-02 05:05:37.873127
3378	1-3378	81	78	4478	27	7597	0	0	2012-08-02 05:05:12.47042	2012-08-02 05:05:38.281941
2341	1-2341	11	7	9056	15	6475	0	0	2012-08-02 04:57:58.851101	2012-08-02 05:05:38.281941
1982	1-1982	68	28	3352	82	1519	0	0	2012-08-02 04:55:23.201807	2012-08-02 05:05:38.281941
3427	1-3427	95	10	5497	33	2553	0	0	2012-08-02 05:05:32.20307	2012-08-02 05:05:39.872961
3446	1-3446	18	100	4604	29	3535	0	0	2012-08-02 05:05:39.872961	2012-08-02 05:05:39.872961
2977	1-2977	85	57	1728	3	880	0	0	2012-08-02 05:02:29.957785	2012-08-02 05:05:42.336437
1029	1-1029	69	9	7408	25	2119	0	0	2012-08-02 04:47:25.641061	2012-08-02 05:05:43.165196
73	1-73	52	67	2569	25	1484	0	0	2012-08-02 04:43:54.270496	2012-08-02 05:05:43.165196
3454	1-3454	19	25	832	82	2247	0	0	2012-08-02 05:05:43.165196	2012-08-02 05:05:43.165196
2878	1-2878	56	6	6004	19	4337	0	0	2012-08-02 05:01:49.086631	2012-08-02 05:05:45.938596
3437	1-3437	67	52	1670	6	7551	0	4	2012-08-02 05:05:36.224518	2012-08-02 05:05:45.938596
3462	1-3462	17	53	2273	9	6514	0	0	2012-08-02 05:05:46.347129	2012-08-02 05:05:46.347129
3412	1-3412	13	37	3879	89	5434	0	7	2012-08-02 05:05:26.147891	2012-08-02 05:05:46.744977
3313	1-3313	77	70	3328	66	5166	0	4	2012-08-02 05:04:45.837282	2012-08-02 05:05:48.390925
1021	1-1021	60	35	9632	66	5931	0	0	2012-08-02 04:47:21.42048	2012-08-02 05:05:48.390925
1434	1-1434	45	69	7424	87	1999	0	0	2012-08-02 04:51:02.642767	2012-08-02 05:05:49.208845
2496	1-2496	4	75	7722	98	7982	0	0	2012-08-02 04:59:06.913422	2012-08-02 05:05:49.208845
3471	1-3471	12	50	1576	5	7242	0	3	2012-08-02 05:05:50.004101	2012-08-02 05:05:50.004101
3442	1-3442	22	82	2578	9	9681	0	0	2012-08-02 05:05:38.281941	2012-08-02 05:05:53.572387
3391	1-3391	86	88	1499	39	537	0	0	2012-08-02 05:05:17.762609	2012-08-02 05:05:53.572387
2835	1-2835	85	62	1734	91	9292	0	0	2012-08-02 05:01:31.064532	2012-08-02 05:05:53.572387
3168	1-3168	60	4	8623	35	9236	0	0	2012-08-02 05:03:47.51533	2012-08-02 05:05:53.572387
2928	1-2928	52	62	122	88	5082	0	0	2012-08-02 05:02:09.860099	2012-08-02 05:05:53.572387
1133	1-1133	29	54	8828	51	2931	0	0	2012-08-02 04:48:27.852161	2012-08-02 05:05:53.572387
543	1-543	37	91	6450	82	3810	1558	0	2012-08-02 04:44:21.021418	2012-08-02 05:05:53.572387
1797	1-1797	38	82	9633	60	4556	0	0	2012-08-02 04:54:01.037579	2012-08-02 05:05:53.992739
3481	1-3481	50	42	941	82	7354	0	0	2012-08-02 05:05:53.992739	2012-08-02 05:05:53.992739
2151	1-2151	12	31	8911	83	6703	0	0	2012-08-02 04:56:36.059526	2012-08-02 05:05:54.390585
3458	1-3458	3	5	8439	35	4711	0	0	2012-08-02 05:05:44.767246	2012-08-02 05:05:55.197061
3484	1-3484	65	51	1397	5	7573	0	0	2012-08-02 05:05:55.197061	2012-08-02 05:05:55.197061
3132	1-3132	27	30	939	67	2199	0	0	2012-08-02 05:03:32.930771	2012-08-02 05:06:00.963958
3485	1-3485	32	67	939	6	2149	0	0	2012-08-02 05:05:55.714082	2012-08-02 05:06:00.963958
2561	1-2561	27	30	4999	19	3599	0	4	2012-08-02 04:59:35.078173	2012-08-02 05:06:00.963958
3498	1-3498	91	6	2099	30	8075	0	0	2012-08-02 05:06:00.963958	2012-08-02 05:06:00.963958
905	1-905	78	44	8216	74	2044	2044	0	2012-08-02 04:46:19.40613	2012-08-02 05:06:00.963958
2803	1-2803	56	44	9912	74	7613	2734	0	2012-08-02 05:01:17.640637	2012-08-02 05:06:00.963958
1022	1-1022	48	49	9357	22	8188	0	0	2012-08-02 04:47:21.453917	2012-08-02 05:06:01.372957
3267	1-3267	57	79	2446	8	2004	0	7	2012-08-02 05:04:27.286656	2012-08-02 05:06:01.372957
3499	1-3499	18	8	1099	49	6366	0	3	2012-08-02 05:06:01.372957	2012-08-02 05:06:01.372957
3440	1-3440	13	74	6577	35	7287	0	0	2012-08-02 05:05:37.464449	2012-08-02 05:06:02.190692
2817	1-2817	9	35	9214	96	9164	0	0	2012-08-02 05:01:23.529171	2012-08-02 05:06:04.16803
3438	1-3438	36	51	2503	63	7027	0	0	2012-08-02 05:05:36.633398	2012-08-02 05:06:04.16803
3338	1-3338	61	63	2298	20	2941	0	0	2012-08-02 05:04:56.025966	2012-08-02 05:06:04.16803
3506	1-3506	27	20	6312	35	8747	0	0	2012-08-02 05:06:04.16803	2012-08-02 05:06:04.16803
3055	1-3055	10	73	2272	21	1229	0	4	2012-08-02 05:03:01.546523	2012-08-02 05:06:04.963941
1529	1-1529	27	90	2774	89	415	0	0	2012-08-02 04:51:53.398276	2012-08-02 05:06:04.963941
3403	1-3403	6	19	3361	18	5997	0	2	2012-08-02 05:05:22.568362	2012-08-02 05:06:05.395208
2981	1-2981	53	62	2637	4	7190	0	0	2012-08-02 05:02:31.58168	2012-08-02 05:06:05.817803
3469	1-3469	76	87	2778	75	6134	0	0	2012-08-02 05:05:49.208845	2012-08-02 05:06:06.613203
3381	1-3381	20	41	1051	90	6250	0	5	2012-08-02 05:05:13.674433	2012-08-02 05:06:08.988473
3301	1-3301	12	41	7712	11	5416	0	8	2012-08-02 05:04:41.008695	2012-08-02 05:06:09.784269
3170	1-3170	63	80	6443	59	5546	0	0	2012-08-02 05:03:48.332504	2012-08-02 05:06:09.784269
3279	1-3279	39	68	2149	78	4162	0	0	2012-08-02 05:04:32.137237	2012-08-02 05:06:10.19277
3525	1-3525	86	57	62	66	40	0	0	2012-08-02 05:06:11.827787	2012-08-02 05:06:11.827787
1842	1-1842	6	70	8033	37	7120	0	0	2012-08-02 04:54:21.252062	2012-08-02 05:06:12.612366
3515	1-3515	89	18	221	15	5911	0	7	2012-08-02 05:06:07.806353	2012-08-02 05:06:12.612366
1926	1-1926	3	97	8618	58	8916	0	0	2012-08-02 04:54:58.550579	2012-08-02 05:06:12.999133
3528	1-3528	77	37	451	97	7824	0	3	2012-08-02 05:06:12.999133	2012-08-02 05:06:12.999133
3482	1-3482	2	83	1974	13	3558	0	0	2012-08-02 05:05:54.390585	2012-08-02 05:06:13.761464
2453	1-2453	64	24	6346	61	3233	0	0	2012-08-02 04:58:47.840343	2012-08-02 05:06:13.761464
3415	1-3415	69	33	2075	100	9658	0	0	2012-08-02 05:05:27.352488	2012-08-02 05:06:15.341311
2572	1-2572	15	47	8732	33	9864	2355	0	2012-08-02 04:59:39.718407	2012-08-02 05:06:15.341311
3539	1-3539	18	52	2266	96	4935	0	0	2012-08-02 05:06:17.396442	2012-08-02 05:06:17.396442
3461	1-3461	13	23	4256	52	6757	0	0	2012-08-02 05:05:45.938596	2012-08-02 05:06:17.793913
55	1-55	43	55	6206	43	3383	0	0	2012-08-02 04:43:54.056205	2012-08-02 05:06:17.793913
3394	1-3394	5	23	3185	47	4957	0	5	2012-08-02 05:05:18.955351	2012-08-02 05:06:17.793913
1979	1-1979	20	23	9336	83	5048	0	0	2012-08-02 04:55:21.909306	2012-08-02 05:06:17.793913
3295	1-3295	77	23	4956	68	5916	0	0	2012-08-02 05:04:38.566976	2012-08-02 05:06:17.793913
3540	1-3540	65	43	1176	23	7968	0	4	2012-08-02 05:06:17.793913	2012-08-02 05:06:17.793913
3370	1-3370	6	83	4441	97	9304	1474	0	2012-08-02 05:05:08.98762	2012-08-02 05:06:17.793913
3072	1-3072	90	24	1218	42	3222	0	0	2012-08-02 05:03:08.443905	2012-08-02 05:06:18.214132
918	1-918	52	24	7705	55	2487	0	0	2012-08-02 04:46:26.697015	2012-08-02 05:06:18.214132
3541	1-3541	79	37	3467	24	7639	0	0	2012-08-02 05:06:18.214132	2012-08-02 05:06:18.214132
3402	1-3402	96	100	3200	22	1324	0	0	2012-08-02 05:05:22.170933	2012-08-02 05:06:19.009614
3426	1-3426	83	54	8595	34	4453	0	0	2012-08-02 05:05:31.783059	2012-08-02 05:06:23.417491
2224	1-2224	23	34	7745	44	8264	0	0	2012-08-02 04:57:07.902914	2012-08-02 05:06:24.213104
690	1-690	63	8	5970	92	8373	1854	0	2012-08-02 04:44:56.492635	2012-08-02 05:06:24.213104
1873	1-1873	94	87	7724	72	3832	0	0	2012-08-02 04:54:34.930559	2012-08-02 05:06:26.234883
3561	1-3561	53	98	231	95	5205	0	0	2012-08-02 05:06:26.234883	2012-08-02 05:06:26.234883
2820	1-2820	2	94	9507	65	3152	0	0	2012-08-02 05:01:24.777474	2012-08-02 05:06:27.836743
2180	1-2180	3	28	8500	57	9431	0	0	2012-08-02 04:56:48.720745	2012-08-02 05:06:28.621286
3344	1-3344	60	28	4293	18	841	0	0	2012-08-02 05:04:58.434575	2012-08-02 05:06:28.621286
3567	1-3567	43	88	5943	28	8016	0	0	2012-08-02 05:06:28.621286	2012-08-02 05:06:28.621286
1838	1-1838	26	9	4772	10	7952	0	0	2012-08-02 04:54:19.439924	2012-08-02 05:06:29.030211
376	1-376	44	93	8678	72	2994	0	0	2012-08-02 04:44:02.263044	2012-08-02 05:06:29.450001
2231	1-2231	47	19	9409	98	8782	980	0	2012-08-02 04:57:10.93024	2012-08-02 05:06:29.450001
2700	1-2700	11	62	3718	77	4570	0	0	2012-08-02 05:00:34.147947	2012-08-02 05:06:30.757343
2433	1-2433	18	62	9954	81	7274	0	0	2012-08-02 04:58:39.068742	2012-08-02 05:06:30.757343
3572	1-3572	43	26	1283	62	4670	0	0	2012-08-02 05:06:30.757343	2012-08-02 05:06:30.757343
3574	1-3574	33	51	1797	58	5815	0	0	2012-08-02 05:06:31.573119	2012-08-02 05:06:31.573119
3355	1-3355	25	38	3047	58	7336	0	7	2012-08-02 05:05:02.875754	2012-08-02 05:06:32.81084
1794	1-1794	67	12	6414	58	1967	0	0	2012-08-02 04:53:59.678834	2012-08-02 05:06:32.81084
1117	1-1117	66	38	9395	58	7028	699	0	2012-08-02 04:48:18.892739	2012-08-02 05:06:32.81084
3577	1-3577	19	58	2742	47	8610	0	6	2012-08-02 05:06:32.81084	2012-08-02 05:06:33.197275
3474	1-3474	65	59	4046	75	1209	0	6	2012-08-02 05:05:51.186097	2012-08-02 05:06:33.197275
1906	1-1906	35	49	5926	55	5467	0	0	2012-08-02 04:54:49.635035	2012-08-02 05:06:33.595514
2956	1-2956	67	94	218	28	52	0	5	2012-08-02 05:02:21.405376	2012-08-02 05:06:33.595514
3579	1-3579	63	37	1229	87	3762	0	3	2012-08-02 05:06:33.595514	2012-08-02 05:06:33.595514
737	1-737	42	87	7938	78	3083	980	0	2012-08-02 04:45:08.994897	2012-08-02 05:06:33.595514
3580	1-3580	52	86	213	38	8326	0	0	2012-08-02 05:06:33.992955	2012-08-02 05:06:33.992955
2259	1-2259	94	25	6859	22	3720	0	7	2012-08-02 04:57:23.127743	2012-08-02 05:06:34.390902
2847	1-2847	65	36	9550	65	7956	0	0	2012-08-02 05:01:36.124875	2012-08-02 05:06:34.390902
3536	1-3536	4	66	1516	25	2945	0	0	2012-08-02 05:06:16.136827	2012-08-02 05:06:34.390902
3581	1-3581	10	22	1458	36	8764	0	0	2012-08-02 05:06:34.390902	2012-08-02 05:06:34.390902
1062	1-1062	72	65	6988	22	4755	1917	0	2012-08-02 04:47:48.066886	2012-08-02 05:06:34.390902
3463	1-3463	25	55	2571	37	7169	0	2	2012-08-02 05:05:46.744977	2012-08-02 05:06:34.799816
3257	1-3257	84	42	6871	88	8951	0	0	2012-08-02 05:04:23.265461	2012-08-02 05:06:35.208731
2437	1-2437	47	12	6926	31	9994	0	0	2012-08-02 04:58:40.836562	2012-08-02 05:06:35.208731
3480	1-3480	55	51	8121	62	8828	0	0	2012-08-02 05:05:53.572387	2012-08-02 05:06:35.208731
3565	1-3565	13	52	1236	94	7195	0	0	2012-08-02 05:06:27.836743	2012-08-02 05:06:35.617497
3562	1-3562	85	53	7406	1	8579	0	0	2012-08-02 05:06:26.643735	2012-08-02 05:06:35.617497
3013	1-3013	74	61	3971	60	1584	0	0	2012-08-02 05:02:44.565148	2012-08-02 05:06:36.413033
3586	1-3586	71	77	1614	100	4763	0	0	2012-08-02 05:06:36.413033	2012-08-02 05:06:36.413033
3584	1-3584	47	94	4299	53	8521	0	0	2012-08-02 05:06:35.617497	2012-08-02 05:06:37.230609
429	1-429	72	8	4022	72	7269	0	0	2012-08-02 04:44:06.001429	2012-08-02 05:06:37.62853
3363	1-3363	31	16	2733	67	1399	0	11	2012-08-02 05:05:06.170017	2012-08-02 05:06:38.424002
2270	1-2270	87	70	2851	82	1399	0	0	2012-08-02 04:57:27.934986	2012-08-02 05:06:38.833046
1716	1-1716	96	67	4282	54	7609	0	0	2012-08-02 04:53:24.606528	2012-08-02 05:06:38.833046
3592	1-3592	65	54	1042	49	2863	0	0	2012-08-02 05:06:38.833046	2012-08-02 05:06:38.833046
3593	1-3593	98	53	3394	55	3595	0	9	2012-08-02 05:06:39.252859	2012-08-02 05:06:39.252859
3546	1-3546	47	75	185	88	9103	0	0	2012-08-02 05:06:20.203151	2012-08-02 05:06:39.650504
3507	1-3507	98	18	3531	92	2379	0	0	2012-08-02 05:06:04.576971	2012-08-02 05:06:39.650504
3530	1-3530	8	94	7434	24	6952	0	0	2012-08-02 05:06:13.761464	2012-08-02 05:06:39.650504
3578	1-3578	88	90	7321	18	5058	0	0	2012-08-02 05:06:33.197275	2012-08-02 05:06:39.650504
1848	1-1848	76	56	3712	95	4261	0	0	2012-08-02 04:54:23.925294	2012-08-02 05:06:39.650504
3262	1-3262	79	99	5431	67	2338	0	0	2012-08-02 05:04:25.276277	2012-08-02 05:06:39.650504
3594	1-3594	64	1	4912	75	8444	0	0	2012-08-02 05:06:39.650504	2012-08-02 05:06:39.650504
3208	1-3208	79	19	3177	90	7646	0	0	2012-08-02 05:04:03.677246	2012-08-02 05:06:40.081959
2856	1-2856	16	90	5245	50	3997	0	11	2012-08-02 05:01:39.917279	2012-08-02 05:06:40.081959
2356	1-2356	18	84	2545	50	5281	0	0	2012-08-02 04:58:05.326368	2012-08-02 05:06:40.081959
2462	1-2462	91	79	2651	50	2023	0	0	2012-08-02 04:58:51.894681	2012-08-02 05:06:40.081959
3156	1-3156	60	34	7273	69	9655	0	0	2012-08-02 05:03:42.58756	2012-08-02 05:06:40.081959
3526	1-3526	94	80	1424	39	439	0	0	2012-08-02 05:06:12.226011	2012-08-02 05:06:43.727664
3359	1-3359	93	5	8032	100	5031	0	0	2012-08-02 05:05:04.456031	2012-08-02 05:06:44.523334
3504	1-3504	2	9	4284	69	5218	0	0	2012-08-02 05:06:03.372694	2012-08-02 05:06:45.705654
3413	1-3413	72	32	2300	69	2815	0	0	2012-08-02 05:05:26.54592	2012-08-02 05:06:45.705654
2505	1-2505	85	35	3892	6	4303	0	0	2012-08-02 04:59:10.791456	2012-08-02 05:06:46.103181
2658	1-2658	6	9	5309	76	5274	0	0	2012-08-02 05:00:16.402526	2012-08-02 05:06:47.307689
2646	1-2646	14	41	5687	80	8078	0	6	2012-08-02 05:00:11.375805	2012-08-02 05:06:48.887145
3081	1-3081	5	25	8737	65	6192	0	0	2012-08-02 05:03:12.134367	2012-08-02 05:06:49.284951
3619	1-3619	51	88	2573	96	9934	0	0	2012-08-02 05:06:49.682692	2012-08-02 05:06:49.682692
3322	1-3322	83	36	5555	66	5832	0	0	2012-08-02 05:04:49.540561	2012-08-02 05:06:50.080579
3280	1-3280	42	44	2803	77	8392	0	10	2012-08-02 05:04:32.535098	2012-08-02 05:06:50.080579
2831	1-2831	31	4	2669	10	5066	0	0	2012-08-02 05:01:29.363281	2012-08-02 05:06:50.080579
3083	1-3083	91	36	6430	86	4887	0	0	2012-08-02 05:03:12.929741	2012-08-02 05:06:50.080579
3595	1-3595	9	50	4777	64	9002	0	0	2012-08-02 05:06:40.081959	2012-08-02 05:06:50.86507
3622	1-3622	85	67	1420	8	4963	0	0	2012-08-02 05:06:50.86507	2012-08-02 05:06:50.86507
3166	1-3166	63	32	6117	73	2517	0	0	2012-08-02 05:03:46.68661	2012-08-02 05:06:53.671591
3002	1-3002	2	32	8753	50	5068	0	0	2012-08-02 05:02:40.090811	2012-08-02 05:06:53.671591
3629	1-3629	30	28	186	32	9985	0	0	2012-08-02 05:06:53.671591	2012-08-02 05:06:53.671591
3534	1-3534	99	100	4881	45	6456	0	0	2012-08-02 05:06:15.341311	2012-08-02 05:06:55.284557
3569	1-3569	10	23	1855	93	8192	0	0	2012-08-02 05:06:29.450001	2012-08-02 05:06:55.284557
3588	1-3588	97	39	4507	45	8832	0	12	2012-08-02 05:06:37.230609	2012-08-02 05:06:55.284557
2948	1-2948	26	100	3677	26	960	0	0	2012-08-02 05:02:18.157555	2012-08-02 05:06:55.284557
3633	1-3633	32	59	2149	39	7698	0	0	2012-08-02 05:06:55.284557	2012-08-02 05:06:55.284557
2442	1-2442	15	27	8439	71	4986	0	0	2012-08-02 04:58:43.012773	2012-08-02 05:06:56.102545
1736	1-1736	97	76	4364	50	3891	0	0	2012-08-02 04:53:33.677355	2012-08-02 05:06:56.102545
2201	1-2201	39	17	6135	42	7311	0	0	2012-08-02 04:56:57.868862	2012-08-02 05:06:57.308487
2113	1-2113	48	83	6474	12	4370	0	0	2012-08-02 04:56:19.761825	2012-08-02 05:06:57.308487
3223	1-3223	74	17	3504	84	1603	0	0	2012-08-02 05:04:09.873027	2012-08-02 05:06:57.308487
3638	1-3638	33	71	648	83	7709	0	0	2012-08-02 05:06:57.308487	2012-08-02 05:06:57.308487
2773	1-2773	37	9	3383	68	5348	0	0	2012-08-02 05:01:05.109476	2012-08-02 05:06:58.501621
3608	1-3608	83	63	6537	95	3914	0	0	2012-08-02 05:06:45.285437	2012-08-02 05:06:58.501621
3511	1-3511	43	92	1197	31	7002	0	0	2012-08-02 05:06:06.215494	2012-08-02 05:06:58.501621
3300	1-3300	70	10	7613	2	7674	0	0	2012-08-02 05:04:40.589326	2012-08-02 05:06:59.695173
2687	1-2687	55	39	6173	10	5113	0	0	2012-08-02 05:00:28.712279	2012-08-02 05:06:59.695173
3417	1-3417	95	74	6849	63	7174	0	0	2012-08-02 05:05:28.159395	2012-08-02 05:07:00.899119
3620	1-3620	34	2	3232	36	7866	0	0	2012-08-02 05:06:50.080579	2012-08-02 05:07:00.899119
3644	1-3644	53	2	3132	39	6229	0	0	2012-08-02 05:06:59.695173	2012-08-02 05:07:00.899119
3380	1-3380	30	89	4452	21	9725	0	0	2012-08-02 05:05:13.276665	2012-08-02 05:07:00.899119
3284	1-3284	28	44	1834	93	3645	0	0	2012-08-02 05:04:34.180919	2012-08-02 05:07:00.899119
3647	1-3647	13	93	8387	89	5848	0	0	2012-08-02 05:07:00.899119	2012-08-02 05:07:00.899119
3641	1-3641	25	31	6774	63	9739	0	0	2012-08-02 05:06:58.501621	2012-08-02 05:07:02.125972
3085	1-3085	58	38	2664	74	9212	0	0	2012-08-02 05:03:13.73627	2012-08-02 05:07:03.747369
3552	1-3552	47	75	4502	90	8377	0	0	2012-08-02 05:06:22.62257	2012-08-02 05:07:03.747369
3069	1-3069	59	44	9263	82	6945	0	0	2012-08-02 05:03:07.239872	2012-08-02 05:07:03.747369
3508	1-3508	81	21	2708	74	6702	0	7	2012-08-02 05:06:04.963941	2012-08-02 05:07:03.747369
3648	1-3648	13	75	2364	44	6794	0	0	2012-08-02 05:07:01.319303	2012-08-02 05:07:03.747369
3512	1-3512	83	75	3737	56	4552	0	0	2012-08-02 05:06:06.613203	2012-08-02 05:07:03.747369
3654	1-3654	83	79	1574	75	9401	0	0	2012-08-02 05:07:03.747369	2012-08-02 05:07:03.747369
2722	1-2722	51	74	8256	79	7164	3169	0	2012-08-02 05:00:43.561703	2012-08-02 05:07:03.747369
3655	1-3655	20	61	825	56	3878	0	2	2012-08-02 05:07:04.178756	2012-08-02 05:07:04.178756
3452	1-3452	21	18	663	57	4039	0	0	2012-08-02 05:05:42.336437	2012-08-02 05:07:05.826012
3518	1-3518	1	90	3405	74	3634	0	0	2012-08-02 05:06:08.988473	2012-08-02 05:07:05.826012
3659	1-3659	81	74	2336	18	5505	0	0	2012-08-02 05:07:05.826012	2012-08-02 05:07:05.826012
3505	1-3505	96	39	2024	100	8940	0	0	2012-08-02 05:06:03.781357	2012-08-02 05:07:06.234873
3231	1-3231	18	52	290	20	210	0	0	2012-08-02 05:04:13.187202	2012-08-02 05:07:09.794264
6	1-6	6	11	6109	10	4342	0	0	2012-08-02 04:43:53.486878	2012-08-02 05:07:09.794264
3349	1-3349	76	35	7086	20	6062	0	0	2012-08-02 05:05:00.434145	2012-08-02 05:07:09.794264
3669	1-3669	88	20	5826	6	8264	0	8	2012-08-02 05:07:09.794264	2012-08-02 05:07:09.794264
3224	1-3224	35	11	7907	81	5251	0	0	2012-08-02 05:04:10.259832	2012-08-02 05:07:11.020452
3527	1-3527	5	37	5128	19	3256	0	0	2012-08-02 05:06:12.612366	2012-08-02 05:07:11.020452
3652	1-3652	69	17	1587	57	2947	0	0	2012-08-02 05:07:02.951781	2012-08-02 05:07:11.429588
2371	1-2371	91	57	6269	14	8941	0	0	2012-08-02 04:58:11.866328	2012-08-02 05:07:11.429588
1816	1-1816	78	84	5766	37	4541	0	0	2012-08-02 04:54:09.795446	2012-08-02 05:07:12.622715
2994	1-2994	27	75	1060	3	961	0	0	2012-08-02 05:02:36.87602	2012-08-02 05:07:16.654937
3434	1-3434	49	76	7136	28	6173	0	0	2012-08-02 05:05:35.02039	2012-08-02 05:07:17.063891
1521	1-1521	6	13	6143	14	3375	0	0	2012-08-02 04:51:49.730679	2012-08-02 05:07:17.461566
3653	1-3653	98	87	1747	76	6319	0	10	2012-08-02 05:07:03.349708	2012-08-02 05:07:18.257344
1633	1-1633	77	64	7260	33	6969	0	4	2012-08-02 04:52:45.470199	2012-08-02 05:07:19.837065
3697	1-3697	56	24	502	50	3342	0	0	2012-08-02 05:07:21.052233	2012-08-02 05:07:21.052233
3700	1-3700	72	59	1307	60	2215	0	0	2012-08-02 05:07:22.22319	2012-08-02 05:07:22.22319
3494	1-3494	58	1	3005	48	4875	0	0	2012-08-02 05:05:59.362006	2012-08-02 05:07:24.200778
3705	1-3705	59	67	365	1	5169	0	0	2012-08-02 05:07:24.200778	2012-08-02 05:07:24.200778
3509	1-3509	17	72	4576	19	6914	0	8	2012-08-02 05:06:05.395208	2012-08-02 05:07:24.9742
3042	1-3042	5	35	6051	55	9916	0	0	2012-08-02 05:02:56.243353	2012-08-02 05:07:26.145305
2681	1-2681	19	62	2965	53	9500	0	0	2012-08-02 05:00:26.105222	2012-08-02 05:07:29.349072
3479	1-3479	70	84	3113	43	6411	0	0	2012-08-02 05:05:53.17473	2012-08-02 05:07:29.349072
3719	1-3719	50	77	3312	62	6410	0	0	2012-08-02 05:07:29.735992	2012-08-02 05:07:29.735992
3080	1-3080	33	63	5520	93	6042	0	0	2012-08-02 05:03:11.736519	2012-08-02 05:07:30.918246
1825	1-1825	75	86	4502	93	8704	0	0	2012-08-02 04:54:13.45176	2012-08-02 05:07:32.486979
3488	1-3488	90	5	8192	86	5207	0	0	2012-08-02 05:05:56.976053	2012-08-02 05:07:32.486979
3726	1-3726	48	1	310	5	8780	0	2	2012-08-02 05:07:32.486979	2012-08-02 05:07:32.486979
3628	1-3628	35	91	2165	89	5694	0	0	2012-08-02 05:06:53.262453	2012-08-02 05:07:33.271549
3728	1-3728	89	51	2088	91	5277	0	7	2012-08-02 05:07:33.271549	2012-08-02 05:07:33.271549
3568	1-3568	62	73	9709	9	9757	0	0	2012-08-02 05:06:29.030211	2012-08-02 05:07:34.066828
3730	1-3730	14	7	1083	73	5693	0	0	2012-08-02 05:07:34.066828	2012-08-02 05:07:34.066828
3718	1-3718	92	76	1707	84	4532	0	0	2012-08-02 05:07:29.349072	2012-08-02 05:07:35.238379
3472	1-3472	59	45	8555	61	9930	0	0	2012-08-02 05:05:50.412865	2012-08-02 05:07:38.366413
3388	1-3388	66	65	6604	75	5273	0	0	2012-08-02 05:05:16.535741	2012-08-02 05:07:38.366413
3243	1-3243	63	14	9958	65	1700	0	0	2012-08-02 05:04:17.628901	2012-08-02 05:07:38.366413
2735	1-2735	66	2	8766	31	3641	0	0	2012-08-02 05:00:49.087404	2012-08-02 05:07:39.548438
3750	1-3750	78	87	286	93	4552	0	0	2012-08-02 05:07:41.990288	2012-08-02 05:07:41.990288
3470	1-3470	41	44	2386	10	1241	0	0	2012-08-02 05:05:49.595284	2012-08-02 05:07:42.376874
331	1-331	41	44	4433	100	1342	0	0	2012-08-02 04:43:59.967494	2012-08-02 05:07:42.376874
3171	1-3171	89	29	3708	10	8666	0	0	2012-08-02 05:03:48.719137	2012-08-02 05:07:42.376874
3520	1-3520	10	54	1000	41	8300	0	0	2012-08-02 05:06:09.784269	2012-08-02 05:07:42.774635
3436	1-3436	57	75	956	15	3587	0	0	2012-08-02 05:05:35.826859	2012-08-02 05:07:43.946041
3756	1-3756	47	92	76	11	1047	0	0	2012-08-02 05:07:44.354554	2012-08-02 05:07:44.354554
3758	1-3758	43	93	235	59	3719	0	5	2012-08-02 05:07:45.149898	2012-08-02 05:07:45.149898
3698	1-3698	3	1	6031	64	2438	0	0	2012-08-02 05:07:21.449892	2012-08-02 05:07:47.12749
3676	1-3676	31	54	2062	84	6193	0	0	2012-08-02 05:07:12.622715	2012-08-02 05:07:49.094187
3277	1-3277	18	22	5404	81	5806	0	0	2012-08-02 05:04:31.341698	2012-08-02 05:07:49.094187
1740	1-1740	69	66	4963	93	6860	0	0	2012-08-02 04:53:35.478002	2012-08-02 05:07:49.094187
3768	1-3768	53	93	6011	66	5114	0	0	2012-08-02 05:07:49.094187	2012-08-02 05:07:49.094187
2909	1-2909	32	15	8053	70	6525	0	0	2012-08-02 05:02:01.913707	2012-08-02 05:07:49.49205
2518	1-2518	25	52	5377	83	4732	0	0	2012-08-02 04:59:16.470759	2012-08-02 05:07:49.49205
2664	1-2664	97	5	3406	88	5022	0	0	2012-08-02 05:00:18.945714	2012-08-02 05:07:49.49205
3502	1-3502	84	50	6624	9	3178	0	0	2012-08-02 05:06:02.588353	2012-08-02 05:07:51.447838
3774	1-3774	63	49	118	50	7974	0	0	2012-08-02 05:07:51.447838	2012-08-02 05:07:51.447838
3775	1-3775	42	29	423	82	4880	0	0	2012-08-02 05:07:51.845501	2012-08-02 05:07:51.845501
3741	1-3741	88	75	462	14	7866	0	5	2012-08-02 05:07:38.366413	2012-08-02 05:07:52.254268
3503	1-3503	69	61	4681	94	9887	0	0	2012-08-02 05:06:02.985876	2012-08-02 05:07:52.254268
3646	1-3646	16	29	622	44	6605	0	0	2012-08-02 05:07:00.490298	2012-08-02 05:07:53.812011
2175	1-2175	77	44	7784	20	7622	0	0	2012-08-02 04:56:46.522006	2012-08-02 05:07:53.812011
2792	1-2792	61	92	7104	20	8629	0	0	2012-08-02 05:01:13.022595	2012-08-02 05:07:53.812011
3290	1-3290	22	10	9565	20	9707	0	3	2012-08-02 05:04:36.578494	2012-08-02 05:07:53.812011
3660	1-3660	71	40	4849	89	2753	0	0	2012-08-02 05:07:06.234873	2012-08-02 05:07:54.972474
2033	1-2033	70	3	7194	81	7435	0	0	2012-08-02 04:55:45.267084	2012-08-02 05:07:58.087669
3228	1-3228	23	38	3656	3	7931	0	0	2012-08-02 05:04:12.005219	2012-08-02 05:07:58.087669
1275	1-1275	52	3	3585	81	2599	2134	0	2012-08-02 04:49:40.421533	2012-08-02 05:07:58.087669
3597	1-3597	3	22	3420	73	3096	0	0	2012-08-02 05:06:40.91085	2012-08-02 05:07:58.496521
3200	1-3200	91	22	5891	85	3200	0	0	2012-08-02 05:04:00.42836	2012-08-02 05:07:58.496521
2155	1-2155	32	97	8532	67	3729	0	0	2012-08-02 04:56:37.8166	2012-08-02 05:07:59.668069
3798	1-3798	83	31	119	19	6073	0	6	2012-08-02 05:08:00.827619	2012-08-02 05:08:00.827619
2764	1-2764	57	11	7965	68	6716	0	0	2012-08-02 05:01:01.285153	2012-08-02 05:08:01.236844
576	1-576	80	11	7998	55	2702	0	0	2012-08-02 04:44:30.360678	2012-08-02 05:08:01.236844
3650	1-3650	54	73	1302	13	2172	0	2	2012-08-02 05:07:02.125972	2012-08-02 05:08:02.805654
3803	1-3803	15	94	372	63	6633	0	0	2012-08-02 05:08:02.805654	2012-08-02 05:08:02.805654
3610	1-3610	94	6	9962	35	9868	0	0	2012-08-02 05:06:46.103181	2012-08-02 05:08:03.192177
3361	1-3361	55	9	5770	7	8774	0	9	2012-08-02 05:05:05.274285	2012-08-02 05:08:03.192177
3609	1-3609	47	69	7129	91	5932	0	0	2012-08-02 05:06:45.705654	2012-08-02 05:08:03.601096
3729	1-3729	83	5	2238	85	639	0	0	2012-08-02 05:07:33.680712	2012-08-02 05:08:07.205386
3673	1-3673	75	14	6166	45	7353	0	0	2012-08-02 05:07:11.429588	2012-08-02 05:08:08.012107
3817	1-3817	68	13	556	73	812	0	3	2012-08-02 05:08:08.409814	2012-08-02 05:08:08.409814
1232	1-1232	42	58	9379	76	5007	0	0	2012-08-02 04:49:15.209925	2012-08-02 05:08:09.216437
2900	1-2900	82	17	8179	88	9252	0	0	2012-08-02 05:01:58.190233	2012-08-02 05:08:09.216437
3696	1-3696	53	71	316	55	259	0	0	2012-08-02 05:07:20.654745	2012-08-02 05:08:09.216437
3822	1-3822	41	63	370	57	2228	0	0	2012-08-02 05:08:10.420925	2012-08-02 05:08:10.420925
3828	1-3828	51	44	2244	56	9238	0	0	2012-08-02 05:08:12.851028	2012-08-02 05:08:12.851028
3829	1-3829	95	51	1693	61	9258	0	0	2012-08-02 05:08:13.248897	2012-08-02 05:08:13.248897
3743	1-3743	51	93	3288	58	9431	0	0	2012-08-02 05:07:39.161912	2012-08-02 05:08:14.055607
2915	1-2915	87	27	8503	70	8987	0	0	2012-08-02 05:02:04.432671	2012-08-02 05:08:14.055607
3715	1-3715	34	6	3218	77	9142	0	0	2012-08-02 05:07:28.167352	2012-08-02 05:08:14.055607
2888	1-2888	55	72	9020	21	7880	0	0	2012-08-02 05:01:53.252006	2012-08-02 05:08:14.055607
3831	1-3831	42	51	6046	6	3702	0	0	2012-08-02 05:08:14.055607	2012-08-02 05:08:14.055607
3835	1-3835	75	60	301	46	2269	0	0	2012-08-02 05:08:15.754458	2012-08-02 05:08:15.754458
3371	1-3371	4	65	2967	8	4306	0	0	2012-08-02 05:05:09.419066	2012-08-02 05:08:18.529785
3769	1-3769	75	88	7450	52	7721	0	0	2012-08-02 05:07:49.49205	2012-08-02 05:08:19.700831
3306	1-3306	48	59	2963	53	1125	0	7	2012-08-02 05:04:43.042026	2012-08-02 05:08:19.700831
3845	1-3845	85	53	4951	67	3992	0	0	2012-08-02 05:08:19.700831	2012-08-02 05:08:19.700831
701	1-701	51	52	7854	53	7924	6816	0	2012-08-02 04:44:58.957177	2012-08-02 05:08:19.700831
3832	1-3832	80	67	6058	99	5704	0	0	2012-08-02 05:08:14.475462	2012-08-02 05:08:20.098783
3846	1-3846	35	80	6280	67	5932	0	0	2012-08-02 05:08:20.098783	2012-08-02 05:08:20.098783
3664	1-3664	5	3	5327	64	7687	0	17	2012-08-02 05:07:07.84975	2012-08-02 05:08:21.292559
3672	1-3672	44	99	2839	37	9275	0	4	2012-08-02 05:07:11.020452	2012-08-02 05:08:21.700964
3685	1-3685	57	78	6620	24	6788	0	0	2012-08-02 05:07:16.257292	2012-08-02 05:08:21.700964
3721	1-3721	71	6	3769	72	8093	0	0	2012-08-02 05:07:30.520579	2012-08-02 05:08:22.120982
3851	1-3851	69	58	900	6	2083	0	0	2012-08-02 05:08:22.120982	2012-08-02 05:08:22.120982
1164	1-1164	74	11	9596	1	6028	1132	0	2012-08-02 04:48:40.283906	2012-08-02 05:08:22.120982
2540	1-2540	11	64	6659	72	6943	0	0	2012-08-02 04:59:25.896758	2012-08-02 05:08:24.485229
2818	1-2818	64	31	2055	47	283	0	0	2012-08-02 05:01:23.949084	2012-08-02 05:08:24.485229
2128	1-2128	75	64	7368	83	5223	0	0	2012-08-02 04:56:26.324101	2012-08-02 05:08:25.280733
3857	1-3857	8	97	3541	64	6794	0	0	2012-08-02 05:08:24.485229	2012-08-02 05:08:25.280733
3783	1-3783	39	32	1146	12	2009	0	0	2012-08-02 05:07:54.972474	2012-08-02 05:08:25.70064
3549	1-3549	31	55	2046	51	4436	0	6	2012-08-02 05:06:21.406948	2012-08-02 05:08:25.70064
1405	1-1405	33	52	9373	32	8245	3779	0	2012-08-02 04:50:48.126158	2012-08-02 05:08:25.70064
669	1-669	47	79	9237	88	8699	0	0	2012-08-02 04:44:52.148517	2012-08-02 05:08:27.711842
3865	1-3865	91	88	7048	79	8568	0	6	2012-08-02 05:08:27.711842	2012-08-02 05:08:27.711842
3554	1-3554	77	82	5556	67	8673	0	0	2012-08-02 05:06:23.417491	2012-08-02 05:08:28.882463
3868	1-3868	51	7	1495	82	5233	0	0	2012-08-02 05:08:28.882463	2012-08-02 05:08:28.882463
3690	1-3690	66	53	7622	69	6148	0	0	2012-08-02 05:07:18.257344	2012-08-02 05:08:29.678128
2244	1-2244	64	91	7285	73	9774	0	0	2012-08-02 04:57:16.553742	2012-08-02 05:08:29.678128
3870	1-3870	19	57	3334	91	6153	0	0	2012-08-02 05:08:29.678128	2012-08-02 05:08:29.678128
3799	1-3799	72	61	1826	11	8128	0	5	2012-08-02 05:08:01.236844	2012-08-02 05:08:30.098002
2855	1-2855	86	91	5965	18	3237	0	0	2012-08-02 05:01:39.496999	2012-08-02 05:08:30.098002
541	1-541	65	91	8670	52	8318	0	0	2012-08-02 04:44:20.46904	2012-08-02 05:08:30.098002
3871	1-3871	71	95	1696	91	3742	0	0	2012-08-02 05:08:30.098002	2012-08-02 05:08:30.098002
3087	1-3087	2	19	5722	100	6912	0	0	2012-08-02 05:03:14.54245	2012-08-02 05:08:30.518041
3116	1-3116	12	98	1366	40	1773	0	0	2012-08-02 05:03:26.364955	2012-08-02 05:08:30.518041
1315	1-1315	69	12	5536	40	6254	0	0	2012-08-02 04:50:04.028955	2012-08-02 05:08:30.518041
3604	1-3604	26	40	7063	41	9785	0	0	2012-08-02 05:06:43.727664	2012-08-02 05:08:30.518041
2860	1-2860	16	60	1834	26	1227	0	9	2012-08-02 05:01:41.596113	2012-08-02 05:08:34.474999
3460	1-3460	3	60	1912	95	2544	0	0	2012-08-02 05:05:45.5406	2012-08-02 05:08:34.474999
3687	1-3687	47	28	9338	26	7571	0	0	2012-08-02 05:07:17.063891	2012-08-02 05:08:34.474999
3744	1-3744	61	20	1413	2	9757	0	2	2012-08-02 05:07:39.548438	2012-08-02 05:08:34.474999
3791	1-3791	61	35	3260	38	6678	0	0	2012-08-02 05:07:58.087669	2012-08-02 05:08:34.474999
3833	1-3833	46	20	4622	28	7061	0	0	2012-08-02 05:08:14.873139	2012-08-02 05:08:34.474999
3882	1-3882	90	54	5114	20	8964	0	0	2012-08-02 05:08:34.474999	2012-08-02 05:08:34.474999
3129	1-3129	72	60	7524	99	6755	0	0	2012-08-02 05:03:31.726448	2012-08-02 05:08:35.271091
2800	1-2800	65	72	7763	61	3755	0	0	2012-08-02 05:01:16.326107	2012-08-02 05:08:36.486166
3887	1-3887	74	57	1629	72	4700	0	0	2012-08-02 05:08:36.486166	2012-08-02 05:08:36.486166
3467	1-3467	21	66	7045	38	9053	0	0	2012-08-02 05:05:48.390925	2012-08-02 05:08:37.712429
3890	1-3890	84	24	9	66	5859	0	5	2012-08-02 05:08:37.712429	2012-08-02 05:08:37.712429
3645	1-3645	11	3	2666	59	7532	0	11	2012-08-02 05:07:00.09283	2012-08-02 05:08:38.541081
3894	1-3894	34	43	174	86	2504	0	0	2012-08-02 05:08:39.347613	2012-08-02 05:08:39.347613
3188	1-3188	23	29	2649	63	5748	0	0	2012-08-02 05:03:55.640701	2012-08-02 05:08:39.734563
3860	1-3860	28	12	4276	55	2935	0	0	2012-08-02 05:08:25.70064	2012-08-02 05:08:40.52989
3281	1-3281	90	39	4059	68	5453	0	0	2012-08-02 05:04:32.95502	2012-08-02 05:08:40.52989
3230	1-3230	86	56	3622	99	8897	0	4	2012-08-02 05:04:12.789475	2012-08-02 05:08:40.52989
3897	1-3897	86	55	1397	26	6531	0	0	2012-08-02 05:08:40.52989	2012-08-02 05:08:40.52989
2634	1-2634	56	47	6953	55	5965	3373	0	2012-08-02 05:00:06.294089	2012-08-02 05:08:40.52989
3819	1-3819	44	53	4404	9	3297	0	2	2012-08-02 05:08:09.216437	2012-08-02 05:08:40.950097
3905	1-3905	31	11	1312	33	1402	0	0	2012-08-02 05:08:43.767476	2012-08-02 05:08:43.767476
3780	1-3780	39	20	2572	29	7884	0	0	2012-08-02 05:07:53.812011	2012-08-02 05:08:44.165059
3755	1-3755	27	15	8966	68	6887	0	0	2012-08-02 05:07:43.946041	2012-08-02 05:08:44.165059
3213	1-3213	5	60	3879	69	4182	0	0	2012-08-02 05:04:05.806604	2012-08-02 05:08:44.97138
3901	1-3901	74	70	2528	60	3174	0	0	2012-08-02 05:08:42.165565	2012-08-02 05:08:44.97138
2043	1-2043	2	69	7872	100	6372	0	0	2012-08-02 04:55:49.608398	2012-08-02 05:08:44.97138
3908	1-3908	56	100	2603	86	9899	0	8	2012-08-02 05:08:44.97138	2012-08-02 05:08:44.97138
3795	1-3795	49	28	2277	97	7416	0	5	2012-08-02 05:07:59.668069	2012-08-02 05:08:46.164989
3776	1-3776	40	45	5925	61	9228	0	11	2012-08-02 05:07:52.254268	2012-08-02 05:08:46.164989
3805	1-3805	69	52	5364	53	9193	0	7	2012-08-02 05:08:03.601096	2012-08-02 05:08:46.562602
3376	1-3376	10	52	5590	53	7472	0	5	2012-08-02 05:05:11.642875	2012-08-02 05:08:46.562602
3911	1-3911	42	97	4546	45	4093	0	0	2012-08-02 05:08:46.164989	2012-08-02 05:08:48.529158
3922	1-3922	68	16	3928	20	3179	0	0	2012-08-02 05:08:50.486898	2012-08-02 05:08:50.486898
16	1-16	16	26	8768	25	7577	893	0	2012-08-02 04:43:53.620584	2012-08-02 05:08:50.486898
3804	1-3804	34	38	7730	6	7853	0	0	2012-08-02 05:08:03.192177	2012-08-02 05:08:50.900001
3733	1-3733	75	46	4066	79	7663	0	0	2012-08-02 05:07:35.238379	2012-08-02 05:08:52.105147
3151	1-3151	46	22	9167	79	7418	0	0	2012-08-02 05:03:40.554418	2012-08-02 05:08:52.504551
2522	1-2522	65	22	9358	35	2129	0	2	2012-08-02 04:59:18.17207	2012-08-02 05:08:52.504551
3913	1-3913	70	47	1040	2	7642	0	2	2012-08-02 05:08:46.960517	2012-08-02 05:08:54.10865
3095	1-3095	60	97	2010	2	1691	0	0	2012-08-02 05:03:17.824414	2012-08-02 05:08:54.905048
3933	1-3933	25	54	2615	34	2539	0	3	2012-08-02 05:08:54.905048	2012-08-02 05:08:54.905048
2892	1-2892	43	36	5324	63	6157	0	0	2012-08-02 05:01:54.886646	2012-08-02 05:08:55.303548
2807	1-2807	60	1	4391	86	1564	0	0	2012-08-02 05:01:19.309069	2012-08-02 05:08:55.701865
3936	1-3936	82	27	663	29	2579	0	0	2012-08-02 05:08:56.111726	2012-08-02 05:08:56.111726
3694	1-3694	69	67	5473	3	8542	0	0	2012-08-02 05:07:19.837065	2012-08-02 05:08:57.308036
3792	1-3792	6	42	5483	22	7833	0	4	2012-08-02 05:07:58.496521	2012-08-02 05:08:57.308036
3939	1-3939	41	3	2732	60	8354	0	0	2012-08-02 05:08:57.308036	2012-08-02 05:08:57.308036
1887	1-1887	60	26	4067	56	1322	0	0	2012-08-02 04:54:41.825065	2012-08-02 05:08:58.491302
3899	1-3899	6	20	4389	26	9353	0	3	2012-08-02 05:08:41.358847	2012-08-02 05:08:58.491302
3826	1-3826	43	20	2125	49	2796	0	0	2012-08-02 05:08:12.055926	2012-08-02 05:08:58.491302
3613	1-3613	42	72	7100	9	5274	0	3	2012-08-02 05:06:47.307689	2012-08-02 05:08:59.298633
3944	1-3944	92	62	331	64	2808	0	0	2012-08-02 05:08:59.298633	2012-08-02 05:08:59.298633
3445	1-3445	60	20	3996	2	2708	0	0	2012-08-02 05:05:39.475351	2012-08-02 05:09:00.514587
3432	1-3432	33	98	3794	23	4690	0	0	2012-08-02 05:05:34.225098	2012-08-02 05:09:00.9126
3942	1-3942	90	23	1795	20	5052	0	0	2012-08-02 05:08:58.491302	2012-08-02 05:09:00.9126
3543	1-3543	74	81	7529	26	5546	0	2	2012-08-02 05:06:19.009614	2012-08-02 05:09:00.9126
3948	1-3948	97	26	1287	98	3135	0	9	2012-08-02 05:09:00.9126	2012-08-02 05:09:00.9126
3837	1-3837	36	99	5144	23	7151	0	0	2012-08-02 05:08:16.530204	2012-08-02 05:09:02.912851
3722	1-3722	17	40	8838	36	7124	0	0	2012-08-02 05:07:30.918246	2012-08-02 05:09:02.912851
3187	1-3187	25	47	7733	26	7480	0	0	2012-08-02 05:03:55.208719	2012-08-02 05:09:03.310996
2814	1-2814	4	78	9674	95	2901	0	4	2012-08-02 05:01:22.27013	2012-08-02 05:09:03.310996
3954	1-3954	13	95	319	60	6690	0	0	2012-08-02 05:09:03.310996	2012-08-02 05:09:03.310996
3927	1-3927	24	2	3483	54	5991	0	0	2012-08-02 05:08:52.504551	2012-08-02 05:09:04.50511
3886	1-3886	42	90	1153	28	386	0	0	2012-08-02 05:08:36.088422	2012-08-02 05:09:04.50511
3917	1-3917	59	45	696	90	4443	0	3	2012-08-02 05:08:48.529158	2012-08-02 05:09:04.50511
3957	1-3957	32	54	649	45	3258	0	0	2012-08-02 05:09:04.50511	2012-08-02 05:09:04.50511
3935	1-3935	81	86	461	69	3855	0	0	2012-08-02 05:08:55.701865	2012-08-02 05:09:06.103116
3443	1-3443	21	29	6100	60	9851	0	0	2012-08-02 05:05:38.679935	2012-08-02 05:09:06.103116
3962	1-3962	24	38	127	95	2321	0	0	2012-08-02 05:09:06.501418	2012-08-02 05:09:06.501418
3872	1-3872	67	41	9086	19	6282	0	3	2012-08-02 05:08:30.518041	2012-08-02 05:09:07.285542
3898	1-3898	21	27	6487	53	2679	0	0	2012-08-02 05:08:40.950097	2012-08-02 05:09:07.285542
2619	1-2619	6	20	6025	73	4013	0	0	2012-08-02 04:59:59.728937	2012-08-02 05:09:07.285542
3785	1-3785	88	75	2921	27	7165	0	0	2012-08-02 05:07:55.745948	2012-08-02 05:09:07.285542
3964	1-3964	20	73	2218	75	7610	0	0	2012-08-02 05:09:07.285542	2012-08-02 05:09:07.285542
3967	1-3967	91	60	647	71	7061	0	0	2012-08-02 05:09:08.512433	2012-08-02 05:09:08.512433
3953	1-3953	59	80	6993	99	6512	0	0	2012-08-02 05:09:02.912851	2012-08-02 05:09:08.91028
3970	1-3970	48	9	1161	44	9665	0	0	2012-08-02 05:09:09.727831	2012-08-02 05:09:09.727831
3389	1-3389	87	50	4996	73	3792	0	0	2012-08-02 05:05:16.956169	2012-08-02 05:09:10.910397
3919	1-3919	4	80	7319	29	4065	0	0	2012-08-02 05:08:49.314268	2012-08-02 05:09:10.910397
3976	1-3976	41	40	865	21	4713	0	4	2012-08-02 05:09:12.103257	2012-08-02 05:09:12.103257
3931	1-3931	98	2	7005	91	2257	0	2	2012-08-02 05:08:54.10865	2012-08-02 05:09:14.467415
3892	1-3892	69	56	1050	3	4871	0	0	2012-08-02 05:08:38.541081	2012-08-02 05:09:14.865564
3985	1-3985	95	14	4644	38	5066	0	0	2012-08-02 05:09:15.66109	2012-08-02 05:09:15.66109
3516	1-3516	12	59	5996	11	7762	0	0	2012-08-02 05:06:08.215493	2012-08-02 05:09:16.44549
2383	1-2383	33	67	5498	59	7345	0	0	2012-08-02 04:58:17.103091	2012-08-02 05:09:16.832209
2782	1-2782	94	24	9833	92	5906	0	0	2012-08-02 05:01:08.82282	2012-08-02 05:09:17.2301
3989	1-3989	37	93	823	24	6468	0	0	2012-08-02 05:09:17.2301	2012-08-02 05:09:17.2301
3990	1-3990	60	20	495	66	7980	0	0	2012-08-02 05:09:17.627959	2012-08-02 05:09:17.627959
3710	1-3710	11	20	4186	35	2020	0	5	2012-08-02 05:07:26.145305	2012-08-02 05:09:18.821021
3993	1-3993	59	27	3008	20	7521	0	6	2012-08-02 05:09:18.821021	2012-08-02 05:09:18.821021
3852	1-3852	87	73	2347	8	1017	0	0	2012-08-02 05:08:22.507671	2012-08-02 05:09:19.249594
3934	1-3934	20	63	6127	26	3890	0	0	2012-08-02 05:08:55.303548	2012-08-02 05:09:19.249594
2851	1-2851	36	63	7596	73	9191	0	0	2012-08-02 05:01:37.782136	2012-08-02 05:09:19.249594
3994	1-3994	54	58	2198	63	9836	0	0	2012-08-02 05:09:19.249594	2012-08-02 05:09:19.249594
3564	1-3564	43	43	6244	62	1551	0	0	2012-08-02 05:06:27.450071	2012-08-02 05:09:20.854661
1037	1-1037	85	21	5892	48	3792	0	0	2012-08-02 04:47:30.569047	2012-08-02 05:09:24.038235
1213	1-1213	53	66	2971	53	5696	0	0	2012-08-02 04:49:06.371081	2012-08-02 05:09:24.457976
3988	1-3988	48	15	2379	73	3610	0	0	2012-08-02 05:09:16.832209	2012-08-02 05:09:24.457976
3982	1-3982	16	30	1210	75	2157	0	0	2012-08-02 05:09:14.467415	2012-08-02 05:09:24.844702
1277	1-1277	95	30	6621	59	9093	0	0	2012-08-02 04:49:41.393492	2012-08-02 05:09:24.844702
3692	1-3692	7	73	8868	67	6937	0	0	2012-08-02 05:07:19.041398	2012-08-02 05:09:24.844702
4008	1-4008	56	22	698	30	6313	0	0	2012-08-02 05:09:24.844702	2012-08-02 05:09:24.844702
2784	1-2784	96	15	2794	2	1690	0	0	2012-08-02 05:01:09.653397	2012-08-02 05:09:27.021231
4013	1-4013	89	2	4244	22	6961	0	0	2012-08-02 05:09:27.021231	2012-08-02 05:09:27.021231
4015	1-4015	56	42	3381	6	4645	0	0	2012-08-02 05:09:27.817116	2012-08-02 05:09:27.817116
1320	1-1320	69	60	7755	32	8499	0	5	2012-08-02 04:50:05.976552	2012-08-02 05:09:28.225583
2904	1-2904	74	42	7444	32	7764	0	0	2012-08-02 05:01:59.825452	2012-08-02 05:09:28.225583
3789	1-3789	66	60	2176	14	9047	0	0	2012-08-02 05:07:57.314326	2012-08-02 05:09:28.225583
2004	1-2004	84	60	7692	44	5834	0	0	2012-08-02 04:55:32.936993	2012-08-02 05:09:28.225583
3914	1-3914	93	71	6246	38	7306	0	0	2012-08-02 05:08:47.35841	2012-08-02 05:09:30.203365
4021	1-4021	89	70	1095	49	3687	0	8	2012-08-02 05:09:30.203365	2012-08-02 05:09:30.203365
2758	1-2758	13	33	9795	67	7874	338	0	2012-08-02 05:00:58.810773	2012-08-02 05:09:30.203365
2908	1-2908	66	32	8621	86	1661	0	3	2012-08-02 05:02:01.504715	2012-08-02 05:09:32.202942
4026	1-4026	70	82	583	32	5488	0	4	2012-08-02 05:09:32.202942	2012-08-02 05:09:32.202942
1712	1-1712	29	76	5052	53	5748	0	0	2012-08-02 04:53:22.795321	2012-08-02 05:09:35.009584
2708	1-2708	84	7	8167	35	2349	0	0	2012-08-02 05:00:37.573583	2012-08-02 05:09:36.191363
2559	1-2559	62	28	8410	19	2926	0	5	2012-08-02 04:59:34.227689	2012-08-02 05:09:36.191363
4036	1-4036	24	100	274	7	6004	0	2	2012-08-02 05:09:36.191363	2012-08-02 05:09:36.191363
3327	1-3327	61	49	5201	78	6352	0	0	2012-08-02 05:04:51.55141	2012-08-02 05:09:37.749313
4040	1-4040	99	29	971	49	7675	0	5	2012-08-02 05:09:37.749313	2012-08-02 05:09:37.749313
3839	1-3839	68	54	3552	78	7385	0	0	2012-08-02 05:08:17.347708	2012-08-02 05:09:38.147286
1130	1-1130	92	28	7995	37	6550	0	0	2012-08-02 04:48:25.720345	2012-08-02 05:09:38.147286
3203	1-3203	85	36	6478	13	4318	0	0	2012-08-02 05:04:01.643927	2012-08-02 05:09:39.362781
3737	1-3737	88	84	3921	93	4610	0	0	2012-08-02 05:07:36.806775	2012-08-02 05:09:40.158205
3000	1-3000	26	84	5160	46	5392	0	0	2012-08-02 05:02:39.284387	2012-08-02 05:09:40.158205
3878	1-3878	24	7	4908	33	7395	0	3	2012-08-02 05:08:32.895209	2012-08-02 05:09:40.158205
3966	1-3966	39	43	5465	84	4950	0	0	2012-08-02 05:09:08.125991	2012-08-02 05:09:40.158205
1888	1-1888	24	43	5059	16	2543	0	0	2012-08-02 04:54:42.288671	2012-08-02 05:09:40.158205
4046	1-4046	42	59	1883	43	9031	0	0	2012-08-02 05:09:40.158205	2012-08-02 05:09:40.158205
1895	1-1895	87	43	8978	23	3260	1552	0	2012-08-02 04:54:45.448128	2012-08-02 05:09:40.158205
4007	1-4007	56	55	3847	29	9562	0	3	2012-08-02 05:09:24.457976	2012-08-02 05:09:41.782328
3814	1-3814	32	62	2759	60	9536	0	0	2012-08-02 05:08:07.205386	2012-08-02 05:09:42.191185
4051	1-4051	66	23	1070	62	7089	0	0	2012-08-02 05:09:42.191185	2012-08-02 05:09:42.191185
1891	1-1891	69	39	4191	40	7019	0	0	2012-08-02 04:54:43.636809	2012-08-02 05:09:42.987019
3688	1-3688	30	28	4749	13	8259	0	5	2012-08-02 05:07:17.461566	2012-08-02 05:09:45.373138
3635	1-3635	40	96	2024	27	9036	0	0	2012-08-02 05:06:56.102545	2012-08-02 05:09:46.146606
3596	1-3596	87	31	5066	1	3691	0	0	2012-08-02 05:06:40.501979	2012-08-02 05:09:46.146606
3763	1-3763	86	34	3668	1	7946	0	6	2012-08-02 05:07:47.12749	2012-08-02 05:09:46.146606
2256	1-2256	82	13	4126	3	963	0	5	2012-08-02 04:57:21.834619	2012-08-02 05:09:46.555255
3411	1-3411	58	13	5685	63	715	0	0	2012-08-02 05:05:25.761235	2012-08-02 05:09:46.555255
4062	1-4062	77	45	262	13	8532	0	8	2012-08-02 05:09:46.555255	2012-08-02 05:09:46.555255
4064	1-4064	97	5	1896	93	2396	0	0	2012-08-02 05:09:47.350952	2012-08-02 05:09:47.350952
3560	1-3560	74	31	4927	88	7238	0	11	2012-08-02 05:06:25.837036	2012-08-02 05:09:49.350954
4069	1-4069	41	12	943	99	4329	0	0	2012-08-02 05:09:49.350954	2012-08-02 05:09:49.350954
3992	1-3992	99	86	1033	27	4209	0	4	2012-08-02 05:09:18.434393	2012-08-02 05:09:50.910689
3844	1-3844	87	48	3388	10	6492	0	12	2012-08-02 05:08:19.303095	2012-08-02 05:09:50.910689
3923	1-3923	82	6	7658	38	9842	0	0	2012-08-02 05:08:50.900001	2012-08-02 05:09:52.081625
2995	1-2995	63	34	3755	18	1483	0	0	2012-08-02 05:02:37.28473	2012-08-02 05:09:52.081625
3551	1-3551	25	6	7459	11	9768	0	0	2012-08-02 05:06:22.224772	2012-08-02 05:09:52.081625
4076	1-4076	46	30	2138	6	9278	0	0	2012-08-02 05:09:52.081625	2012-08-02 05:09:52.081625
1593	1-1593	58	56	8624	99	4247	1132	0	2012-08-02 04:52:27.151699	2012-08-02 05:09:52.081625
3867	1-3867	85	94	4369	50	9440	0	0	2012-08-02 05:08:28.495809	2012-08-02 05:09:52.501621
1161	1-1161	48	94	1469	10	469	0	0	2012-08-02 04:48:39.333655	2012-08-02 05:09:52.501621
2808	1-2808	74	29	9076	66	3441	0	0	2012-08-02 05:01:19.728869	2012-08-02 05:09:53.308598
4079	1-4079	90	79	1752	29	8995	0	4	2012-08-02 05:09:53.308598	2012-08-02 05:09:53.308598
4054	1-4054	86	40	2580	31	2042	0	0	2012-08-02 05:09:43.384701	2012-08-02 05:09:54.10386
4010	1-4010	42	31	3024	87	5749	0	3	2012-08-02 05:09:25.812923	2012-08-02 05:09:54.10386
1328	1-1328	17	56	9831	22	9411	0	0	2012-08-02 04:50:10.892798	2012-08-02 05:09:55.2862
2200	1-2200	87	52	6967	59	8620	0	0	2012-08-02 04:56:57.437632	2012-08-02 05:09:55.2862
1490	1-1490	13	52	6549	59	814	814	0	2012-08-02 04:51:34.826675	2012-08-02 05:09:55.2862
3841	1-3841	20	84	6602	25	3942	0	2	2012-08-02 05:08:18.143141	2012-08-02 05:09:57.661387
4090	1-4090	35	42	1646	84	8813	0	0	2012-08-02 05:09:57.661387	2012-08-02 05:09:57.661387
1322	1-1322	91	46	7897	33	5802	0	0	2012-08-02 04:50:07.909462	2012-08-02 05:09:59.351679
3240	1-3240	12	13	9470	74	4749	0	0	2012-08-02 05:04:16.390961	2012-08-02 05:09:59.351679
4061	1-4061	44	1	5678	71	7029	0	0	2012-08-02 05:09:46.146606	2012-08-02 05:10:01.318292
4077	1-4077	55	99	198	94	6452	0	0	2012-08-02 05:09:52.501621	2012-08-02 05:10:03.273669
3606	1-3606	82	58	3048	5	6209	0	0	2012-08-02 05:06:44.523334	2012-08-02 05:10:04.069331
3082	1-3082	9	95	7254	92	3284	0	0	2012-08-02 05:03:12.531848	2012-08-02 05:10:04.46703
1866	1-1866	33	7	8813	71	3414	0	0	2012-08-02 04:54:31.68288	2012-08-02 05:10:04.46703
3397	1-3397	53	71	9982	96	7463	602	0	2012-08-02 05:05:20.137914	2012-08-02 05:10:04.46703
4000	1-4000	24	60	4144	98	1729	0	0	2012-08-02 05:09:21.674214	2012-08-02 05:10:06.501623
2146	1-2146	98	56	5609	23	6111	0	5	2012-08-02 04:56:33.82797	2012-08-02 05:10:06.888585
3856	1-3856	6	23	2377	2	5939	0	0	2012-08-02 05:08:24.098505	2012-08-02 05:10:06.888585
3545	1-3545	34	21	2166	2	3892	0	0	2012-08-02 05:06:19.805066	2012-08-02 05:10:06.888585
3248	1-3248	9	79	3520	2	3797	0	0	2012-08-02 05:04:19.652324	2012-08-02 05:10:06.888585
685	1-685	15	56	6586	23	4926	1867	0	2012-08-02 04:44:55.802682	2012-08-02 05:10:06.888585
880	1-880	68	96	4549	5	8123	0	0	2012-08-02 04:46:06.66452	2012-08-02 05:10:08.855118
2992	1-2992	84	96	7579	3	6552	0	0	2012-08-02 05:02:36.069466	2012-08-02 05:10:08.855118
4106	1-4106	97	5	6274	95	7422	0	0	2012-08-02 05:10:04.069331	2012-08-02 05:10:08.855118
4118	1-4118	3	95	4310	96	8190	0	0	2012-08-02 05:10:08.855118	2012-08-02 05:10:08.855118
4043	1-4043	43	6	2108	34	6625	0	3	2012-08-02 05:09:38.954021	2012-08-02 05:10:09.650739
4124	1-4124	67	98	756	76	8594	0	3	2012-08-02 05:10:11.230541	2012-08-02 05:10:11.230541
3695	1-3695	79	13	9415	23	1906	0	5	2012-08-02 05:07:20.245787	2012-08-02 05:10:12.015332
2032	1-2032	63	49	3646	15	3418	0	0	2012-08-02 04:55:44.814346	2012-08-02 05:10:14.370014
3291	1-3291	45	33	3239	59	2426	0	11	2012-08-02 05:04:36.976174	2012-08-02 05:10:14.370014
3703	1-3703	61	59	9644	15	9574	0	4	2012-08-02 05:07:23.405479	2012-08-02 05:10:14.370014
2535	1-2535	88	33	5254	59	3922	751	8	2012-08-02 04:59:23.70684	2012-08-02 05:10:14.370014
1997	1-1997	65	34	8135	79	5883	0	0	2012-08-02 04:55:29.876947	2012-08-02 05:10:15.552451
3862	1-3862	75	34	3872	97	3819	0	0	2012-08-02 05:08:26.507187	2012-08-02 05:10:15.552451
1402	1-1402	45	99	3243	53	4644	0	0	2012-08-02 04:50:45.762297	2012-08-02 05:10:15.972473
2027	1-2027	48	93	8031	59	4597	0	3	2012-08-02 04:55:42.637703	2012-08-02 05:10:15.972473
1855	1-1855	15	47	6945	57	7935	0	0	2012-08-02 04:54:26.667372	2012-08-02 05:10:15.972473
1549	1-1549	30	47	4932	88	8025	0	0	2012-08-02 04:52:04.004355	2012-08-02 05:10:15.972473
4136	1-4136	20	51	1463	99	7557	0	0	2012-08-02 05:10:15.972473	2012-08-02 05:10:15.972473
1844	1-1844	100	60	8293	58	6306	0	0	2012-08-02 04:54:22.124706	2012-08-02 05:10:16.381474
4024	1-4024	96	72	6647	12	8672	0	0	2012-08-02 05:09:31.418814	2012-08-02 05:10:17.972547
3940	1-3940	88	21	2694	40	9206	0	0	2012-08-02 05:08:57.717916	2012-08-02 05:10:18.358838
3961	1-3961	15	69	2387	41	6079	0	7	2012-08-02 05:09:06.103116	2012-08-02 05:10:19.676322
4145	1-4145	84	49	1291	69	3734	0	0	2012-08-02 05:10:19.676322	2012-08-02 05:10:19.676322
4073	1-4073	78	27	2793	48	7986	0	0	2012-08-02 05:09:50.910689	2012-08-02 05:10:20.480499
2984	1-2984	89	51	8103	12	5828	0	0	2012-08-02 05:02:32.786391	2012-08-02 05:10:20.877917
3956	1-3956	4	43	6536	28	6993	0	0	2012-08-02 05:09:04.118412	2012-08-02 05:10:22.435986
4041	1-4041	14	20	1772	28	5733	0	3	2012-08-02 05:09:38.147286	2012-08-02 05:10:22.435986
4154	1-4154	18	17	1595	79	2373	0	7	2012-08-02 05:10:23.220218	2012-08-02 05:10:23.220218
3497	1-3497	68	37	9386	11	9538	0	2	2012-08-02 05:06:00.588341	2012-08-02 05:10:24.004685
3859	1-3859	87	55	4299	97	9161	0	0	2012-08-02 05:08:25.280733	2012-08-02 05:10:24.004685
4160	1-4160	10	92	332	61	1433	0	0	2012-08-02 05:10:25.540632	2012-08-02 05:10:25.540632
2326	1-2326	86	39	5025	42	5003	0	4	2012-08-02 04:57:52.330428	2012-08-02 05:10:25.949281
4135	1-4135	1	57	137	34	2597	0	8	2012-08-02 05:10:15.552451	2012-08-02 05:10:26.722573
3420	1-3420	94	85	3880	43	9358	0	0	2012-08-02 05:05:29.385295	2012-08-02 05:10:26.722573
2366	1-2366	12	95	8585	25	6898	0	0	2012-08-02 04:58:09.645646	2012-08-02 05:10:26.722573
3686	1-3686	41	68	3742	75	1570	0	0	2012-08-02 05:07:16.654937	2012-08-02 05:10:27.905031
3179	1-3179	96	13	8142	94	7364	0	0	2012-08-02 05:03:51.977365	2012-08-02 05:10:28.291593
4169	1-4169	17	81	1212	94	8809	0	0	2012-08-02 05:10:29.065002	2012-08-02 05:10:29.065002
2177	1-2177	69	35	7724	1	9833	202	0	2012-08-02 04:56:47.40613	2012-08-02 05:10:29.065002
3658	1-3658	37	49	8121	71	6219	0	0	2012-08-02 05:07:05.405815	2012-08-02 05:10:29.893876
4172	1-4172	26	88	2079	35	997	0	0	2012-08-02 05:10:30.269145	2012-08-02 05:10:30.269145
3702	1-3702	2	100	3978	10	4654	0	0	2012-08-02 05:07:23.018996	2012-08-02 05:10:30.667105
4044	1-4044	12	65	6404	36	8968	0	0	2012-08-02 05:09:39.362781	2012-08-02 05:10:31.827278
4176	1-4176	4	83	5499	43	9981	0	0	2012-08-02 05:10:31.827278	2012-08-02 05:10:31.827278
3681	1-3681	95	55	7713	45	5842	0	6	2012-08-02 05:07:14.622801	2012-08-02 05:10:33.384674
4181	1-4181	78	40	1799	96	2222	0	0	2012-08-02 05:10:33.771495	2012-08-02 05:10:33.771495
3720	1-3720	46	66	1032	22	2093	0	0	2012-08-02 05:07:30.133639	2012-08-02 05:10:34.147189
4094	1-4094	91	59	1217	13	8632	0	2	2012-08-02 05:09:59.351679	2012-08-02 05:10:34.147189
3998	1-3998	2	9	2738	43	8999	0	5	2012-08-02 05:09:20.854661	2012-08-02 05:10:34.147189
4182	1-4182	87	13	5541	18	3342	0	4	2012-08-02 05:10:34.147189	2012-08-02 05:10:34.147189
3955	1-3955	73	74	2705	65	6255	0	0	2012-08-02 05:09:03.720578	2012-08-02 05:10:34.556486
4183	1-4183	97	1	4359	77	8367	0	0	2012-08-02 05:10:34.556486	2012-08-02 05:10:34.556486
1125	1-1125	64	69	1032	53	792	0	0	2012-08-02 04:48:23.345163	2012-08-02 05:10:35.793471
4186	1-4186	12	53	236	69	1190	0	5	2012-08-02 05:10:35.793471	2012-08-02 05:10:35.793471
3449	1-3449	64	96	6520	92	3544	0	0	2012-08-02 05:05:41.066153	2012-08-02 05:10:37.738004
3401	1-3401	71	96	7165	98	7521	0	0	2012-08-02 05:05:21.751207	2012-08-02 05:10:37.738004
2504	1-2504	11	96	4036	88	8238	0	0	2012-08-02 04:59:10.360461	2012-08-02 05:10:37.738004
4191	1-4191	90	82	1016	96	4865	0	0	2012-08-02 05:10:37.738004	2012-08-02 05:10:37.738004
3849	1-3849	46	92	4629	82	7910	2092	0	2012-08-02 05:08:21.292559	2012-08-02 05:10:37.738004
4193	1-4193	30	52	1704	24	4834	0	0	2012-08-02 05:10:38.544947	2012-08-02 05:10:38.544947
2616	1-2616	70	20	7708	26	4005	0	0	2012-08-02 04:59:58.436228	2012-08-02 05:10:39.306953
4195	1-4195	84	59	480	20	3128	0	0	2012-08-02 05:10:39.306953	2012-08-02 05:10:39.306953
3591	1-3591	23	67	4249	90	8121	0	0	2012-08-02 05:06:38.424002	2012-08-02 05:10:40.480207
3493	1-3493	76	44	6916	12	9918	0	0	2012-08-02 05:05:58.964218	2012-08-02 05:10:40.480207
3026	1-3026	94	61	9392	4	2742	0	0	2012-08-02 05:02:49.769188	2012-08-02 05:10:41.253409
4202	1-4202	6	85	196	19	7213	0	0	2012-08-02 05:10:42.026742	2012-08-02 05:10:42.026742
3830	1-3830	100	37	5420	58	7755	0	3	2012-08-02 05:08:13.658009	2012-08-02 05:10:42.435677
2597	1-2597	17	16	8707	66	1813	0	0	2012-08-02 04:59:50.391739	2012-08-02 05:10:43.220047
3712	1-3712	46	15	5071	22	3368	0	0	2012-08-02 05:07:26.976696	2012-08-02 05:10:43.606709
4198	1-4198	98	27	2183	67	8657	0	9	2012-08-02 05:10:40.480207	2012-08-02 05:10:43.606709
4017	1-4017	44	95	3561	61	2609	0	0	2012-08-02 05:09:28.623648	2012-08-02 05:10:44.004535
4208	1-4208	48	32	3326	4	5928	0	0	2012-08-02 05:10:44.402339	2012-08-02 05:10:44.402339
4180	1-4180	55	45	2622	98	8007	0	0	2012-08-02 05:10:33.384674	2012-08-02 05:10:45.230841
3382	1-3382	25	19	7926	98	9382	0	7	2012-08-02 05:05:14.083089	2012-08-02 05:10:45.230841
3639	1-3639	75	83	6960	98	2543	0	0	2012-08-02 05:06:57.728336	2012-08-02 05:10:45.230841
2795	1-2795	18	93	2415	40	4867	0	3	2012-08-02 05:01:14.260114	2012-08-02 05:10:46.40216
3751	1-3751	94	10	9768	29	4608	0	0	2012-08-02 05:07:42.376874	2012-08-02 05:10:46.40216
4213	1-4213	65	67	367	93	9074	0	0	2012-08-02 05:10:46.40216	2012-08-02 05:10:46.40216
4214	1-4214	53	76	1940	86	5249	0	0	2012-08-02 05:10:46.811088	2012-08-02 05:10:46.811088
4215	1-4215	100	42	1238	75	2103	0	2	2012-08-02 05:10:47.18689	2012-08-02 05:10:47.18689
1523	1-1523	72	79	4191	5	4262	0	0	2012-08-02 04:51:50.669749	2012-08-02 05:10:49.893745
4222	1-4222	74	81	4442	79	6408	0	0	2012-08-02 05:10:49.893745	2012-08-02 05:10:49.893745
4224	1-4224	28	92	5103	86	6275	0	0	2012-08-02 05:10:50.689072	2012-08-02 05:10:50.689072
3749	1-3749	90	22	3219	59	3725	0	0	2012-08-02 05:07:41.581319	2012-08-02 05:10:51.075766
3617	1-3617	62	37	9738	54	5399	689	0	2012-08-02 05:06:48.887145	2012-08-02 05:10:51.075766
4165	1-4165	8	70	3421	46	5660	0	3	2012-08-02 05:10:27.518489	2012-08-02 05:10:52.6229
4050	1-4050	84	3	3952	98	6121	0	0	2012-08-02 05:09:41.782328	2012-08-02 05:10:52.6229
4229	1-4229	89	63	5098	3	6309	0	0	2012-08-02 05:10:52.6229	2012-08-02 05:10:52.6229
3531	1-3531	21	27	1882	76	3690	0	0	2012-08-02 05:06:14.148255	2012-08-02 05:10:53.81571
902	1-902	11	73	6659	5	3488	0	0	2012-08-02 04:46:18.235149	2012-08-02 05:10:53.81571
3582	1-3582	27	81	5765	47	3973	0	0	2012-08-02 05:06:34.799816	2012-08-02 05:10:55.749295
4237	1-4237	9	37	2545	81	9047	0	0	2012-08-02 05:10:55.749295	2012-08-02 05:10:55.749295
3450	1-3450	100	83	6538	17	8735	0	0	2012-08-02 05:05:41.486195	2012-08-02 05:10:56.135998
3926	1-3926	99	10	7803	68	7729	0	6	2012-08-02 05:08:52.105147	2012-08-02 05:10:56.135998
3707	1-3707	58	41	8988	89	4283	0	0	2012-08-02 05:07:24.9742	2012-08-02 05:10:56.135998
4238	1-4238	52	17	3045	79	5513	0	8	2012-08-02 05:10:56.135998	2012-08-02 05:10:56.135998
4152	1-4152	60	28	9903	41	3822	0	0	2012-08-02 05:10:22.435986	2012-08-02 05:10:57.704946
4242	1-4242	89	14	166	28	1697	0	0	2012-08-02 05:10:57.704946	2012-08-02 05:10:57.704946
4119	1-4119	51	94	6713	35	5852	0	0	2012-08-02 05:10:09.254692	2012-08-02 05:10:58.091705
3724	1-3724	41	35	2174	12	3066	0	0	2012-08-02 05:07:31.702428	2012-08-02 05:10:58.865119
3983	1-3983	19	3	6483	56	3150	0	0	2012-08-02 05:09:14.865564	2012-08-02 05:10:58.865119
1861	1-1861	47	47	2303	53	1698	0	0	2012-08-02 04:54:29.440095	2012-08-02 05:11:00.44522
4205	1-4205	1	17	2374	16	6536	0	10	2012-08-02 05:10:43.220047	2012-08-02 05:11:02.013767
3533	1-3533	23	79	3723	41	4317	0	0	2012-08-02 05:06:14.943838	2012-08-02 05:11:02.787434
4255	1-4255	46	41	2212	66	5341	0	6	2012-08-02 05:11:02.787434	2012-08-02 05:11:02.787434
4012	1-4012	85	53	2543	37	5401	0	0	2012-08-02 05:09:26.623771	2012-08-02 05:11:04.322863
4033	1-4033	21	53	5711	21	6682	0	7	2012-08-02 05:09:35.009584	2012-08-02 05:11:04.322863
3904	1-3904	49	53	9195	63	7595	0	0	2012-08-02 05:08:43.380441	2012-08-02 05:11:04.322863
3163	1-3163	4	60	5666	42	5723	0	0	2012-08-02 05:03:45.449098	2012-08-02 05:11:07.143361
4097	1-4097	45	76	8085	52	5115	0	0	2012-08-02 05:10:00.533911	2012-08-02 05:11:07.960665
4126	1-4126	47	47	2455	33	6935	0	11	2012-08-02 05:10:12.015332	2012-08-02 05:11:08.358442
4141	1-4141	3	60	2528	72	3958	0	2	2012-08-02 05:10:17.972547	2012-08-02 05:11:08.767462
4120	1-4120	57	5	7057	81	8208	0	5	2012-08-02 05:10:09.650739	2012-08-02 05:11:09.176099
4053	1-4053	11	59	2207	39	3264	0	6	2012-08-02 05:09:42.987019	2012-08-02 05:11:09.596145
4084	1-4084	85	59	1864	51	8725	0	0	2012-08-02 05:09:55.2862	2012-08-02 05:11:09.596145
4206	1-4206	99	59	2417	27	5068	0	0	2012-08-02 05:10:43.606709	2012-08-02 05:11:09.596145
1773	1-1773	20	59	3979	90	1385	0	3	2012-08-02 04:53:50.125979	2012-08-02 05:11:09.596145
2127	1-2127	38	89	2168	2	3310	0	0	2012-08-02 04:56:25.89326	2012-08-02 05:11:09.596145
3623	1-3623	37	25	585	11	452	0	7	2012-08-02 05:06:51.274033	2012-08-02 05:11:10.43567
1826	1-1826	1	34	5503	25	3043	0	0	2012-08-02 04:54:13.926652	2012-08-02 05:11:10.43567
4275	1-4275	67	76	948	8	2976	0	0	2012-08-02 05:11:10.855766	2012-08-02 05:11:10.855766
3310	1-3310	12	85	9341	13	9299	0	0	2012-08-02 05:04:44.643719	2012-08-02 05:11:11.651126
4277	1-4277	99	44	1116	33	8350	0	6	2012-08-02 05:11:11.651126	2012-08-02 05:11:11.651126
3987	1-3987	74	11	7451	55	6624	0	6	2012-08-02 05:09:16.44549	2012-08-02 05:11:14.468292
4248	1-4248	14	32	966	100	3551	0	0	2012-08-02 05:11:00.036254	2012-08-02 05:11:14.468292
4285	1-4285	31	77	460	50	2263	0	0	2012-08-02 05:11:14.89939	2012-08-02 05:11:14.89939
4232	1-4232	27	76	3873	73	8154	0	0	2012-08-02 05:10:53.81571	2012-08-02 05:11:16.976527
1341	1-1341	19	69	4392	14	5218	0	0	2012-08-02 04:50:17.013478	2012-08-02 05:11:16.976527
3840	1-3840	95	49	3871	33	3953	0	0	2012-08-02 05:08:17.745372	2012-08-02 05:11:16.976527
3796	1-3796	71	60	7198	69	6651	0	3	2012-08-02 05:08:00.065479	2012-08-02 05:11:16.976527
4264	1-4264	86	24	2817	27	8429	0	0	2012-08-02 05:11:06.314855	2012-08-02 05:11:17.793868
3031	1-3031	66	61	6246	51	9900	0	0	2012-08-02 05:02:51.801859	2012-08-02 05:11:21.086207
4200	1-4200	93	67	5380	61	9811	0	0	2012-08-02 05:10:41.253409	2012-08-02 05:11:21.086207
3404	1-3404	40	90	9711	70	6686	0	0	2012-08-02 05:05:22.966039	2012-08-02 05:11:21.086207
2454	1-2454	3	66	8077	79	5683	0	6	2012-08-02 04:58:48.271421	2012-08-02 05:11:21.914763
3555	1-3555	23	66	3967	60	7840	0	0	2012-08-02 05:06:23.826719	2012-08-02 05:11:21.914763
2105	1-2105	4	66	9104	28	6651	0	0	2012-08-02 04:56:16.226658	2012-08-02 05:11:21.914763
1178	1-1178	21	66	5814	2	5527	0	7	2012-08-02 04:48:46.990717	2012-08-02 05:11:21.914763
3684	1-3684	71	57	5941	83	7429	0	0	2012-08-02 05:07:15.85983	2012-08-02 05:11:21.914763
4302	1-4302	38	83	5410	66	8920	0	3	2012-08-02 05:11:21.914763	2012-08-02 05:11:21.914763
2668	1-2668	69	58	6023	57	7968	164	0	2012-08-02 05:00:20.614976	2012-08-02 05:11:21.914763
4303	1-4303	47	37	416	82	295	0	0	2012-08-02 05:11:22.335142	2012-08-02 05:11:22.335142
4081	1-4081	49	87	5248	69	5809	0	0	2012-08-02 05:09:54.10386	2012-08-02 05:11:22.733128
4304	1-4304	39	88	2140	87	7465	0	2	2012-08-02 05:11:22.733128	2012-08-02 05:11:22.733128
4055	1-4055	90	69	4573	78	8502	0	0	2012-08-02 05:09:43.782137	2012-08-02 05:11:23.561707
3407	1-3407	39	69	2584	40	3235	0	0	2012-08-02 05:05:24.17059	2012-08-02 05:11:23.561707
4306	1-4306	49	73	1549	69	4812	0	0	2012-08-02 05:11:23.561707	2012-08-02 05:11:23.561707
3266	1-3266	51	18	4379	34	5793	0	6	2012-08-02 05:04:26.888911	2012-08-02 05:11:23.970397
4058	1-4058	35	32	3520	40	3360	0	7	2012-08-02 05:09:44.964195	2012-08-02 05:11:23.970397
3920	1-3920	46	66	9732	91	8964	0	0	2012-08-02 05:08:49.702432	2012-08-02 05:11:23.970397
4307	1-4307	71	40	3550	18	6835	0	0	2012-08-02 05:11:23.970397	2012-08-02 05:11:23.970397
3711	1-3711	35	85	2452	78	6211	0	4	2012-08-02 05:07:26.554164	2012-08-02 05:11:24.401631
4290	1-4290	75	33	3913	60	8644	0	0	2012-08-02 05:11:16.976527	2012-08-02 05:11:24.401631
3034	1-3034	55	33	8946	28	8315	0	0	2012-08-02 05:02:53.039361	2012-08-02 05:11:24.401631
4312	1-4312	72	13	550	99	1978	0	10	2012-08-02 05:11:26.069752	2012-08-02 05:11:26.069752
3891	1-3891	79	56	3083	4	3271	0	0	2012-08-02 05:08:38.132634	2012-08-02 05:11:26.467572
3806	1-3806	18	15	1523	12	1022	0	2	2012-08-02 05:08:03.998764	2012-08-02 05:11:26.467572
3924	1-3924	42	79	3260	25	5105	0	0	2012-08-02 05:08:51.298684	2012-08-02 05:11:26.467572
4099	1-4099	89	71	3435	36	7681	0	0	2012-08-02 05:10:01.318292	2012-08-02 05:11:26.887509
2225	1-2225	96	78	7067	54	1526	0	0	2012-08-02 04:57:08.333808	2012-08-02 05:11:26.887509
4314	1-4314	15	88	958	78	7051	0	5	2012-08-02 05:11:26.887509	2012-08-02 05:11:26.887509
3848	1-3848	21	54	7549	22	8978	4081	0	2012-08-02 05:08:20.894399	2012-08-02 05:11:26.887509
4112	1-4112	32	45	2995	60	8935	0	0	2012-08-02 05:10:06.501623	2012-08-02 05:11:29.340321
4321	1-4321	39	45	186	74	7542	0	0	2012-08-02 05:11:29.737945	2012-08-02 05:11:29.737945
3190	1-3190	86	34	2889	16	5359	0	3	2012-08-02 05:03:56.437245	2012-08-02 05:11:30.544854
3679	1-3679	80	70	4101	51	8030	0	0	2012-08-02 05:07:13.816139	2012-08-02 05:11:31.792834
4245	1-4245	83	12	6409	19	6176	0	0	2012-08-02 05:10:58.865119	2012-08-02 05:11:31.792834
3052	1-3052	59	29	9862	67	9537	0	0	2012-08-02 05:03:00.287145	2012-08-02 05:11:31.792834
3331	1-3331	30	48	8711	68	4982	0	0	2012-08-02 05:04:53.153726	2012-08-02 05:11:32.214797
4159	1-4159	97	50	710	23	482	0	5	2012-08-02 05:10:25.153952	2012-08-02 05:11:33.82785
4331	1-4331	83	23	2005	30	8687	0	0	2012-08-02 05:11:33.82785	2012-08-02 05:11:33.82785
4072	1-4072	30	43	7094	65	8706	0	9	2012-08-02 05:09:50.523828	2012-08-02 05:11:35.076412
4107	1-4107	29	82	2763	70	7349	0	0	2012-08-02 05:10:04.46703	2012-08-02 05:11:35.496064
3972	1-3972	65	58	7774	13	8243	0	0	2012-08-02 05:09:10.523319	2012-08-02 05:11:35.905033
3760	1-3760	8	9	5053	28	9049	0	0	2012-08-02 05:07:45.945376	2012-08-02 05:11:35.905033
4014	1-4014	3	12	5273	56	3320	0	0	2012-08-02 05:09:27.419158	2012-08-02 05:11:36.325114
4337	1-4337	42	68	1005	73	5588	0	0	2012-08-02 05:11:36.325114	2012-08-02 05:11:36.325114
4274	1-4274	46	55	946	34	9931	0	0	2012-08-02 05:11:10.43567	2012-08-02 05:11:36.744591
4308	1-4308	97	78	3750	33	8721	0	0	2012-08-02 05:11:24.401631	2012-08-02 05:11:37.153576
3971	1-3971	29	40	6421	52	3563	0	8	2012-08-02 05:09:10.125728	2012-08-02 05:11:37.153576
4279	1-4279	54	61	291	40	1686	0	0	2012-08-02 05:11:12.458033	2012-08-02 05:11:37.153576
4339	1-4339	99	18	1131	35	7924	0	5	2012-08-02 05:11:37.153576	2012-08-02 05:11:37.153576
4113	1-4113	30	2	5460	72	9045	0	2	2012-08-02 05:10:06.888585	2012-08-02 05:11:37.982681
4163	1-4163	46	25	7318	19	6856	0	0	2012-08-02 05:10:26.722573	2012-08-02 05:11:39.219903
2055	1-2055	8	50	6599	71	1294	0	0	2012-08-02 04:55:54.886664	2012-08-02 05:11:39.639552
2917	1-2917	76	95	9714	87	5677	0	0	2012-08-02 05:02:05.273453	2012-08-02 05:11:39.639552
2804	1-2804	17	4	2250	54	2810	0	4	2012-08-02 05:01:18.060564	2012-08-02 05:11:39.639552
4345	1-4345	15	71	1242	86	6679	0	0	2012-08-02 05:11:39.639552	2012-08-02 05:11:39.639552
1074	1-1074	61	86	6967	4	3914	9	0	2012-08-02 04:47:54.739481	2012-08-02 05:11:39.639552
3096	1-3096	83	63	7231	12	6869	0	0	2012-08-02 05:03:18.222531	2012-08-02 05:11:41.275087
4355	1-4355	70	72	177	39	171	0	0	2012-08-02 05:11:43.716353	2012-08-02 05:11:43.716353
3784	1-3784	75	41	4172	10	3995	0	0	2012-08-02 05:07:55.359034	2012-08-02 05:11:46.599823
4362	1-4362	68	73	2805	41	6075	0	11	2012-08-02 05:11:46.599823	2012-08-02 05:11:46.599823
4020	1-4020	69	83	3858	73	4491	0	0	2012-08-02 05:09:29.816667	2012-08-02 05:11:46.997372
4096	1-4096	30	83	9936	63	5150	0	0	2012-08-02 05:10:00.147587	2012-08-02 05:11:46.997372
4291	1-4291	22	7	2988	5	2146	0	2	2012-08-02 05:11:17.396464	2012-08-02 05:11:48.21263
4259	1-4259	9	95	878	53	6340	0	0	2012-08-02 05:11:04.322863	2012-08-02 05:11:51.129193
4173	1-4173	41	79	877	100	7139	0	2	2012-08-02 05:10:30.667105	2012-08-02 05:11:51.129193
4004	1-4004	11	93	3311	100	2675	0	0	2012-08-02 05:09:23.253496	2012-08-02 05:11:51.129193
4377	1-4377	48	16	1499	92	9049	0	0	2012-08-02 05:11:52.753707	2012-08-02 05:11:52.753707
1425	1-1425	58	77	2753	2	2008	0	0	2012-08-02 04:50:58.400142	2012-08-02 05:11:53.184688
3400	1-3400	50	14	6129	5	5269	0	3	2012-08-02 05:05:21.364193	2012-08-02 05:11:54.001868
4381	1-4381	25	49	171	18	9167	0	0	2012-08-02 05:11:54.39977	2012-08-02 05:11:54.39977
3909	1-3909	13	14	7994	33	5072	0	0	2012-08-02 05:08:45.380564	2012-08-02 05:11:55.649591
4384	1-4384	57	7	9	14	8404	0	0	2012-08-02 05:11:55.649591	2012-08-02 05:11:55.649591
3968	1-3968	25	50	7041	80	9653	0	5	2012-08-02 05:09:08.91028	2012-08-02 05:11:58.067591
782	1-782	39	80	9463	88	9350	289	0	2012-08-02 04:45:26.508676	2012-08-02 05:11:58.067591
3535	1-3535	88	73	3377	12	1363	0	5	2012-08-02 05:06:15.739066	2012-08-02 05:11:58.465321
4391	1-4391	72	54	2248	73	7285	0	0	2012-08-02 05:11:58.465321	2012-08-02 05:11:58.465321
3029	1-3029	70	15	2020	32	1693	0	7	2012-08-02 05:02:50.995688	2012-08-02 05:11:59.283308
4393	1-4393	77	32	964	35	5752	0	0	2012-08-02 05:11:59.283308	2012-08-02 05:11:59.283308
4121	1-4121	57	29	1459	47	1989	0	0	2012-08-02 05:10:10.060098	2012-08-02 05:12:03.007891
4405	1-4405	13	83	266	34	4867	0	8	2012-08-02 05:12:04.22313	2012-08-02 05:12:04.22313
4272	1-4272	30	72	2348	59	9112	0	6	2012-08-02 05:11:09.596145	2012-08-02 05:12:04.620995
3949	1-3949	3	31	7007	88	4491	0	13	2012-08-02 05:09:01.33321	2012-08-02 05:12:05.041225
3810	1-3810	87	88	7073	12	2116	0	0	2012-08-02 05:08:05.581621	2012-08-02 05:12:05.041225
1774	1-1774	35	56	7952	44	8009	0	0	2012-08-02 04:53:50.57891	2012-08-02 05:12:06.621868
4243	1-4243	42	68	6470	94	8704	0	6	2012-08-02 05:10:58.091705	2012-08-02 05:12:06.621868
3996	1-3996	12	56	4210	68	5887	0	0	2012-08-02 05:09:20.047935	2012-08-02 05:12:06.621868
4203	1-4203	73	13	6166	44	5810	0	0	2012-08-02 05:10:42.435677	2012-08-02 05:12:06.621868
4344	1-4344	10	77	2350	50	9285	0	2	2012-08-02 05:11:39.219903	2012-08-02 05:12:08.566287
4416	1-4416	50	90	476	65	3436	0	0	2012-08-02 05:12:08.566287	2012-08-02 05:12:08.566287
3553	1-3553	90	21	4852	51	4408	0	3	2012-08-02 05:06:23.020029	2012-08-02 05:12:08.953247
3566	1-3566	24	90	3522	88	3527	0	0	2012-08-02 05:06:28.234466	2012-08-02 05:12:08.953247
4132	1-4132	81	22	4529	49	9697	0	0	2012-08-02 05:10:14.370014	2012-08-02 05:12:09.339916
305	1-305	54	60	6840	47	3146	0	0	2012-08-02 04:43:59.101896	2012-08-02 05:12:09.339916
3979	1-3979	30	22	5891	28	4231	0	3	2012-08-02 05:09:13.285705	2012-08-02 05:12:09.339916
4037	1-4037	82	15	6737	90	9316	22	0	2012-08-02 05:09:36.57849	2012-08-02 05:12:09.339916
4067	1-4067	11	3	2310	31	9542	0	0	2012-08-02 05:09:48.533119	2012-08-02 05:12:10.909044
4271	1-4271	2	31	1921	5	3047	0	4	2012-08-02 05:11:09.176099	2012-08-02 05:12:10.909044
4035	1-4035	53	21	3015	1	2534	0	0	2012-08-02 05:09:35.805105	2012-08-02 05:12:10.909044
3305	1-3305	28	1	5756	5	5986	0	0	2012-08-02 05:04:42.633272	2012-08-02 05:12:10.909044
4422	1-4422	21	5	6398	4	9873	0	4	2012-08-02 05:12:10.909044	2012-08-02 05:12:10.909044
404	1-404	96	26	9162	5	3318	30	0	2012-08-02 04:44:03.968773	2012-08-02 05:12:10.909044
4326	1-4326	41	51	5653	63	4827	0	7	2012-08-02 05:11:31.792834	2012-08-02 05:12:11.31814
4156	1-4156	63	97	5864	64	5193	0	5	2012-08-02 05:10:24.004685	2012-08-02 05:12:11.31814
4424	1-4424	62	23	1893	21	9071	0	0	2012-08-02 05:12:11.715846	2012-08-02 05:12:11.715846
4299	1-4299	33	30	8324	38	5839	0	0	2012-08-02 05:11:20.677468	2012-08-02 05:12:12.51113
3820	1-3820	63	86	3187	95	6995	0	0	2012-08-02 05:08:09.614186	2012-08-02 05:12:12.51113
4426	1-4426	74	32	2255	30	5548	0	5	2012-08-02 05:12:12.51113	2012-08-02 05:12:12.51113
4142	1-4142	84	90	5862	21	9359	0	0	2012-08-02 05:10:18.358838	2012-08-02 05:12:13.704601
3441	1-3441	78	92	5800	1	9635	0	0	2012-08-02 05:05:37.873127	2012-08-02 05:12:14.897829
4027	1-4027	28	93	9417	1	6519	0	0	2012-08-02 05:09:32.62315	2012-08-02 05:12:14.897829
4432	1-4432	14	88	4394	9	5715	0	0	2012-08-02 05:12:14.897829	2012-08-02 05:12:14.897829
512	1-512	4	1	4709	88	4848	1150	0	2012-08-02 04:44:13.968029	2012-08-02 05:12:14.897829
3315	1-3315	19	9	7460	45	4267	0	0	2012-08-02 05:04:46.654956	2012-08-02 05:12:16.842046
4349	1-4349	71	12	6987	19	8809	0	6	2012-08-02 05:11:41.275087	2012-08-02 05:12:16.842046
4437	1-4437	24	60	1430	9	8347	0	0	2012-08-02 05:12:16.842046	2012-08-02 05:12:16.842046
4134	1-4134	38	62	7822	23	8212	0	0	2012-08-02 05:10:15.165493	2012-08-02 05:12:17.637768
3950	1-3950	70	24	6178	47	4994	0	0	2012-08-02 05:09:01.730692	2012-08-02 05:12:18.797669
4443	1-4443	86	37	630	83	4395	0	0	2012-08-02 05:12:19.19552	2012-08-02 05:12:19.19552
4098	1-4098	47	89	3062	9	8260	0	4	2012-08-02 05:10:00.9318	2012-08-02 05:12:20.377749
4380	1-4380	60	11	7743	96	3037	0	0	2012-08-02 05:11:54.001868	2012-08-02 05:12:20.377749
3816	1-3816	68	45	7275	90	9228	0	0	2012-08-02 05:08:08.012107	2012-08-02 05:12:21.570873
4453	1-4453	10	29	8	18	4191	0	0	2012-08-02 05:12:23.128493	2012-08-02 05:12:23.128493
3752	1-3752	69	89	6841	86	6023	0	0	2012-08-02 05:07:42.774635	2012-08-02 05:12:23.526698
4166	1-4166	25	2	4779	8	3612	0	0	2012-08-02 05:10:27.905031	2012-08-02 05:12:23.526698
4454	1-4454	8	80	1402	89	5002	0	0	2012-08-02 05:12:23.526698	2012-08-02 05:12:23.526698
3991	1-3991	53	86	3804	84	8733	0	0	2012-08-02 05:09:18.03694	2012-08-02 05:12:24.31087
4367	1-4367	66	4	5556	1	5564	0	2	2012-08-02 05:11:48.621514	2012-08-02 05:12:24.31087
3110	1-3110	28	15	6918	84	2798	0	0	2012-08-02 05:03:23.923118	2012-08-02 05:12:24.31087
4423	1-4423	60	29	2654	97	7703	0	4	2012-08-02 05:12:11.31814	2012-08-02 05:12:24.708904
3227	1-3227	23	29	8508	3	6010	0	8	2012-08-02 05:04:11.596345	2012-08-02 05:12:24.708904
4457	1-4457	54	21	721	29	3127	0	0	2012-08-02 05:12:24.708904	2012-08-02 05:12:24.708904
4459	1-4459	48	79	82	65	8000	0	0	2012-08-02 05:12:25.50476	2012-08-02 05:12:25.50476
3548	1-3548	5	64	7339	34	8867	0	0	2012-08-02 05:06:20.998248	2012-08-02 05:12:26.288802
4207	1-4207	1	85	8439	45	7206	0	0	2012-08-02 05:10:44.004535	2012-08-02 05:12:27.450695
4251	1-4251	95	99	3660	23	5444	0	4	2012-08-02 05:11:01.218701	2012-08-02 05:12:28.235096
4129	1-4129	57	23	2770	16	9496	0	0	2012-08-02 05:10:13.221127	2012-08-02 05:12:28.633231
4309	1-4309	55	23	1616	58	1333	0	0	2012-08-02 05:11:24.85082	2012-08-02 05:12:28.633231
1898	1-1898	1	55	6360	65	1537	0	0	2012-08-02 04:54:46.862341	2012-08-02 05:12:28.633231
3004	1-3004	71	62	3306	1	1780	0	0	2012-08-02 05:02:40.908651	2012-08-02 05:12:29.041971
4468	1-4468	8	86	1215	62	5169	0	0	2012-08-02 05:12:29.041971	2012-08-02 05:12:29.041971
4470	1-4470	55	42	643	96	1563	0	2	2012-08-02 05:12:29.881785	2012-08-02 05:12:29.881785
4441	1-4441	89	84	3204	54	3606	0	0	2012-08-02 05:12:18.411056	2012-08-02 05:12:30.279439
4456	1-4456	88	84	8681	86	8301	0	10	2012-08-02 05:12:24.31087	2012-08-02 05:12:30.279439
3965	1-3965	77	48	5640	44	3675	0	8	2012-08-02 05:09:07.717189	2012-08-02 05:12:30.279439
3772	1-3772	87	51	739	25	162	0	0	2012-08-02 05:07:50.663101	2012-08-02 05:12:30.677156
1756	1-1756	62	78	9765	63	1939	0	0	2012-08-02 04:53:42.349348	2012-08-02 05:12:36.963181
4488	1-4488	12	63	162	78	7276	0	0	2012-08-02 05:12:36.963181	2012-08-02 05:12:36.963181
4489	1-4489	31	73	42	98	21	0	0	2012-08-02 05:12:37.361002	2012-08-02 05:12:37.361002
4490	1-4490	85	51	3905	45	7207	0	5	2012-08-02 05:12:37.770056	2012-08-02 05:12:37.770056
4123	1-4123	91	8	3621	17	3982	0	0	2012-08-02 05:10:10.844308	2012-08-02 05:12:38.156484
4358	1-4358	74	66	3837	8	6128	0	0	2012-08-02 05:11:44.942678	2012-08-02 05:12:38.156484
4373	1-4373	22	100	6321	86	4722	0	12	2012-08-02 05:11:51.129193	2012-08-02 05:12:38.156484
4491	1-4491	98	17	4003	66	7300	0	0	2012-08-02 05:12:38.156484	2012-08-02 05:12:38.156484
3997	1-3997	27	45	8332	97	5724	0	9	2012-08-02 05:09:20.445717	2012-08-02 05:12:38.55443
4458	1-4458	91	2	4308	1	3965	0	11	2012-08-02 05:12:25.117812	2012-08-02 05:12:38.55443
3709	1-3709	9	38	1059	94	560	0	0	2012-08-02 05:07:25.75893	2012-08-02 05:12:38.952447
3328	1-3328	64	26	4629	54	4181	0	0	2012-08-02 05:04:51.949095	2012-08-02 05:12:39.338888
4161	1-4161	86	42	3575	72	5519	0	4	2012-08-02 05:10:25.949281	2012-08-02 05:12:39.338888
3153	1-3153	15	42	1615	72	472	50	0	2012-08-02 05:03:41.383264	2012-08-02 05:12:39.338888
4346	1-4346	64	27	4699	26	2671	0	2	2012-08-02 05:11:40.070676	2012-08-02 05:12:39.736796
1680	1-1680	57	79	8475	10	8126	0	0	2012-08-02 04:53:08.540058	2012-08-02 05:12:39.736796
4495	1-4495	19	97	2802	79	8276	0	0	2012-08-02 05:12:39.736796	2012-08-02 05:12:39.736796
4148	1-4148	42	12	2628	92	1748	0	8	2012-08-02 05:10:20.877917	2012-08-02 05:12:41.294667
4499	1-4499	88	34	420	21	9632	0	5	2012-08-02 05:12:41.294667	2012-08-02 05:12:41.294667
4501	1-4501	74	41	1090	61	3632	0	0	2012-08-02 05:12:42.079104	2012-08-02 05:12:42.079104
4503	1-4503	86	6	64	71	1702	0	0	2012-08-02 05:12:42.874592	2012-08-02 05:12:42.874592
1524	1-1524	88	46	8296	69	1134	0	0	2012-08-02 04:51:51.122846	2012-08-02 05:12:44.045482
4467	1-4467	22	99	2291	23	7381	0	7	2012-08-02 05:12:28.633231	2012-08-02 05:12:44.432493
2082	1-2082	94	84	9032	48	7882	0	11	2012-08-02 04:56:06.492414	2012-08-02 05:12:44.432493
4449	1-4449	45	1	1497	45	2149	0	0	2012-08-02 05:12:21.570873	2012-08-02 05:12:44.432493
3625	1-3625	84	48	3433	93	3902	0	8	2012-08-02 05:06:52.080795	2012-08-02 05:12:44.432493
4466	1-4466	60	65	6617	99	8794	0	0	2012-08-02 05:12:28.235096	2012-08-02 05:12:44.432493
4378	1-4378	35	65	1970	77	5654	0	0	2012-08-02 05:11:53.184688	2012-08-02 05:12:44.432493
4351	1-4351	46	69	3829	15	2510	0	7	2012-08-02 05:11:42.092301	2012-08-02 05:12:44.908109
2506	1-2506	11	69	5952	64	2042	0	0	2012-08-02 04:59:11.232984	2012-08-02 05:12:44.908109
1642	1-1642	77	84	2594	25	1515	0	0	2012-08-02 04:52:49.800296	2012-08-02 05:12:44.908109
2048	1-2048	81	46	6691	53	4687	0	0	2012-08-02 04:55:51.795838	2012-08-02 05:12:47.272179
4507	1-4507	17	95	4330	65	7767	0	0	2012-08-02 05:12:44.432493	2012-08-02 05:12:47.658858
4515	1-4515	1	64	555	94	906	0	3	2012-08-02 05:12:47.658858	2012-08-02 05:12:47.658858
4030	1-4030	1	1	4123	88	8417	0	7	2012-08-02 05:09:33.794096	2012-08-02 05:12:48.06776
4368	1-4368	91	16	6310	26	3877	0	0	2012-08-02 05:11:49.03024	2012-08-02 05:12:48.852108
4518	1-4518	37	26	987	16	5113	0	0	2012-08-02 05:12:48.852108	2012-08-02 05:12:48.852108
3787	1-3787	100	86	2937	34	2630	0	0	2012-08-02 05:07:56.518949	2012-08-02 05:12:49.238913
4519	1-4519	65	50	1540	86	5241	0	0	2012-08-02 05:12:49.238913	2012-08-02 05:12:49.238913
4493	1-4493	50	94	2012	75	1982	0	0	2012-08-02 05:12:38.952447	2012-08-02 05:12:49.62569
4250	1-4250	98	40	8181	7	8345	0	0	2012-08-02 05:11:00.832002	2012-08-02 05:12:50.034625
3373	1-3373	77	45	8357	30	3761	0	3	2012-08-02 05:05:10.225225	2012-08-02 05:12:50.034625
4292	1-4292	30	25	6786	24	8672	0	0	2012-08-02 05:11:17.793868	2012-08-02 05:12:50.830074
4526	1-4526	9	92	720	98	610	0	7	2012-08-02 05:12:52.034621	2012-08-02 05:12:52.034621
4327	1-4327	37	94	7264	48	9899	0	0	2012-08-02 05:11:32.214797	2012-08-02 05:12:52.421062
4528	1-4528	10	31	523	93	5166	0	0	2012-08-02 05:12:52.820695	2012-08-02 05:12:52.820695
4506	1-4506	52	57	1966	46	8117	0	7	2012-08-02 05:12:44.045482	2012-08-02 05:12:54.046988
3850	1-3850	9	24	5911	38	5468	0	0	2012-08-02 05:08:21.700964	2012-08-02 05:12:54.444689
1665	1-1665	48	86	9128	70	5959	0	0	2012-08-02 04:53:01.656067	2012-08-02 05:12:54.444689
3815	1-3815	69	70	4953	77	9922	0	0	2012-08-02 05:08:07.614289	2012-08-02 05:12:54.444689
4411	1-4411	82	44	3562	56	7566	0	2	2012-08-02 05:12:06.621868	2012-08-02 05:12:54.887207
4150	1-4150	37	28	5244	93	6236	0	5	2012-08-02 05:10:21.662509	2012-08-02 05:12:55.284483
4534	1-4534	84	91	1538	13	8478	0	5	2012-08-02 05:12:55.284483	2012-08-02 05:12:55.284483
3973	1-3973	100	15	4257	50	7143	0	5	2012-08-02 05:09:10.910397	2012-08-02 05:12:56.113135
4523	1-4523	22	97	2465	50	5725	0	0	2012-08-02 05:12:50.830074	2012-08-02 05:12:56.113135
4536	1-4536	10	50	4891	74	7512	0	0	2012-08-02 05:12:56.113135	2012-08-02 05:12:56.113135
1486	1-1486	12	58	9678	51	8967	0	0	2012-08-02 04:51:32.385193	2012-08-02 05:12:57.273451
4130	1-4130	47	43	6146	36	5296	0	0	2012-08-02 05:10:13.618818	2012-08-02 05:12:58.058132
1148	1-1148	83	57	4152	10	935	0	0	2012-08-02 04:48:34.063067	2012-08-02 05:12:58.058132
4541	1-4541	25	88	1841	56	4662	0	5	2012-08-02 05:12:58.058132	2012-08-02 05:12:58.058132
4210	1-4210	87	98	7639	18	9256	0	0	2012-08-02 05:10:45.230841	2012-08-02 05:12:59.250974
2085	1-2085	61	61	4716	42	6528	0	0	2012-08-02 04:56:07.807547	2012-08-02 05:12:59.637879
4492	1-4492	79	96	268	2	8133	0	2	2012-08-02 05:12:38.55443	2012-08-02 05:12:59.637879
4016	1-4016	9	22	3972	60	7966	0	0	2012-08-02 05:09:28.225583	2012-08-02 05:12:59.637879
3303	1-3303	99	37	4850	48	3955	0	0	2012-08-02 05:04:41.81562	2012-08-02 05:13:00.433604
3704	1-3704	44	27	9865	68	5512	0	0	2012-08-02 05:07:23.803295	2012-08-02 05:13:00.433604
4527	1-4527	92	48	1674	62	4511	0	0	2012-08-02 05:12:52.421062	2012-08-02 05:13:00.433604
3626	1-3626	96	7	6396	73	3258	0	0	2012-08-02 05:06:52.478102	2012-08-02 05:13:00.433604
4547	1-4547	34	42	4571	27	8249	0	0	2012-08-02 05:13:00.433604	2012-08-02 05:13:00.433604
3261	1-3261	60	1	9850	10	5483	0	8	2012-08-02 05:04:24.856664	2012-08-02 05:13:02.024505
4421	1-4421	75	1	6162	84	8306	0	0	2012-08-02 05:12:10.522221	2012-08-02 05:13:02.024505
2802	1-2802	100	71	7750	99	3760	0	0	2012-08-02 05:01:17.22101	2012-08-02 05:13:03.217633
3478	1-3478	31	67	6587	9	3709	0	10	2012-08-02 05:05:52.777004	2012-08-02 05:13:04.377756
4268	1-4268	37	8	1244	76	7741	0	2	2012-08-02 05:11:07.960665	2012-08-02 05:13:04.377756
4521	1-4521	86	73	2489	62	5792	0	0	2012-08-02 05:12:50.034625	2012-08-02 05:13:05.552814
4336	1-4336	82	13	5956	62	6891	0	0	2012-08-02 05:11:35.905033	2012-08-02 05:13:05.552814
4560	1-4560	51	62	2276	64	8191	0	3	2012-08-02 05:13:05.552814	2012-08-02 05:13:05.552814
4566	1-4566	47	57	771	97	9738	0	0	2012-08-02 05:13:08.01419	2012-08-02 05:13:08.01419
4471	1-4471	83	54	2772	84	7188	0	3	2012-08-02 05:12:30.279439	2012-08-02 05:13:08.412048
1674	1-1674	94	19	8025	9	3470	0	0	2012-08-02 04:53:05.800476	2012-08-02 05:13:08.412048
4567	1-4567	80	12	1202	19	8812	0	14	2012-08-02 05:13:08.412048	2012-08-02 05:13:08.412048
4508	1-4508	95	15	3597	82	5161	0	6	2012-08-02 05:12:44.908109	2012-08-02 05:13:09.230075
4514	1-4514	84	53	8362	69	7459	0	7	2012-08-02 05:12:47.272179	2012-08-02 05:13:09.230075
4439	1-4439	93	10	5978	62	8286	0	12	2012-08-02 05:12:17.637768	2012-08-02 05:13:09.638437
2468	1-2468	63	62	6766	68	3277	0	0	2012-08-02 04:58:54.619539	2012-08-02 05:13:09.638437
4361	1-4361	53	74	9355	24	6375	0	0	2012-08-02 05:11:46.191114	2012-08-02 05:13:09.638437
4570	1-4570	14	68	1350	75	4208	0	7	2012-08-02 05:13:09.638437	2012-08-02 05:13:09.638437
580	1-580	86	77	5670	74	1993	114	0	2012-08-02 04:44:31.156337	2012-08-02 05:13:09.638437
4573	1-4573	14	71	1466	56	3133	0	0	2012-08-02 05:13:10.887518	2012-08-02 05:13:10.887518
4127	1-4127	28	18	7724	9	9377	0	0	2012-08-02 05:10:12.436881	2012-08-02 05:13:12.930839
4334	1-4334	82	48	5228	43	8410	0	4	2012-08-02 05:11:35.076412	2012-08-02 05:13:15.372371
4584	1-4584	11	43	5319	48	7503	0	0	2012-08-02 05:13:15.372371	2012-08-02 05:13:15.372371
4487	1-4487	22	29	3404	13	5767	0	0	2012-08-02 05:12:36.576695	2012-08-02 05:13:16.598989
3912	1-3912	44	63	4766	55	4286	0	0	2012-08-02 05:08:46.562602	2012-08-02 05:13:16.598989
615	1-615	5	63	3162	41	875	0	0	2012-08-02 04:44:39.933354	2012-08-02 05:13:16.598989
4288	1-4288	20	63	3638	67	2962	0	0	2012-08-02 05:11:16.147579	2012-08-02 05:13:16.598989
1106	1-1106	4	63	8182	10	1760	0	0	2012-08-02 04:48:10.475183	2012-08-02 05:13:16.598989
4587	1-4587	49	49	494	63	7987	0	0	2012-08-02 05:13:16.598989	2012-08-02 05:13:16.598989
4347	1-4347	41	13	7837	28	8074	0	0	2012-08-02 05:11:40.457244	2012-08-02 05:13:17.029915
4230	1-4230	16	19	5417	3	9920	0	0	2012-08-02 05:10:53.042463	2012-08-02 05:13:18.234135
4591	1-4591	7	3	1805	54	7714	0	6	2012-08-02 05:13:18.234135	2012-08-02 05:13:18.234135
2566	1-2566	19	45	8168	67	3081	0	0	2012-08-02 04:59:37.18859	2012-08-02 05:13:19.449411
4594	1-4594	49	60	78	13	4446	0	0	2012-08-02 05:13:19.449411	2012-08-02 05:13:19.449411
4595	1-4595	72	95	954	70	1220	0	0	2012-08-02 05:13:19.859507	2012-08-02 05:13:19.859507
4412	1-4412	42	43	3345	51	5528	0	2	2012-08-02 05:12:07.019701	2012-08-02 05:13:20.278162
4596	1-4596	2	67	101	66	5195	0	0	2012-08-02 05:13:20.278162	2012-08-02 05:13:20.278162
4300	1-4300	62	51	8900	67	7435	30	0	2012-08-02 05:11:21.086207	2012-08-02 05:13:20.278162
4343	1-4343	64	34	5178	24	8706	0	0	2012-08-02 05:11:38.811175	2012-08-02 05:13:22.730917
4532	1-4532	30	77	5370	74	8638	0	7	2012-08-02 05:12:54.444689	2012-08-02 05:13:23.968085
1307	1-1307	47	86	9868	24	4893	0	0	2012-08-02 04:49:59.134315	2012-08-02 05:13:23.968085
4011	1-4011	30	4	5268	26	7111	0	0	2012-08-02 05:09:26.226116	2012-08-02 05:13:23.968085
4089	1-4089	35	86	7913	4	7586	0	9	2012-08-02 05:09:57.274568	2012-08-02 05:13:23.968085
4109	1-4109	56	12	5749	18	2341	0	0	2012-08-02 05:10:05.34127	2012-08-02 05:13:24.388179
4615	1-4615	74	39	529	85	4382	0	2	2012-08-02 05:13:28.112962	2012-08-02 05:13:28.112962
4464	1-4464	56	97	883	40	3285	0	0	2012-08-02 05:12:27.450695	2012-08-02 05:13:29.759012
4155	1-4155	14	11	6110	86	2315	0	0	2012-08-02 05:10:23.618165	2012-08-02 05:13:29.759012
4296	1-4296	47	56	1958	40	8072	0	5	2012-08-02 05:11:19.417972	2012-08-02 05:13:29.759012
4469	1-4469	74	82	2598	69	8254	0	0	2012-08-02 05:12:29.450909	2012-08-02 05:13:30.576546
3946	1-3946	1	23	8259	10	7730	0	0	2012-08-02 05:09:00.117275	2012-08-02 05:13:31.405362
4408	1-4408	13	73	6158	1	5385	0	2	2012-08-02 05:12:05.473148	2012-08-02 05:13:31.405362
3573	1-3573	6	87	9702	49	3417	0	0	2012-08-02 05:06:31.173621	2012-08-02 05:13:31.8252
4316	1-4316	57	16	8592	29	9255	0	8	2012-08-02 05:11:27.716232	2012-08-02 05:13:31.8252
4625	1-4625	3	45	844	55	8058	0	0	2012-08-02 05:13:32.245191	2012-08-02 05:13:32.245191
3627	1-3627	10	61	2397	28	2253	0	3	2012-08-02 05:06:52.876128	2012-08-02 05:13:33.449333
4628	1-4628	63	66	1853	97	5004	0	0	2012-08-02 05:13:33.449333	2012-08-02 05:13:33.449333
4511	1-4511	42	86	4079	94	7182	0	0	2012-08-02 05:12:46.101062	2012-08-02 05:13:33.869408
4561	1-4561	60	13	2579	45	8007	0	0	2012-08-02 05:13:05.959752	2012-08-02 05:13:33.869408
3884	1-3884	76	99	9396	48	7487	0	0	2012-08-02 05:08:35.271091	2012-08-02 05:13:34.27819
3632	1-3632	2	98	7632	8	9983	0	0	2012-08-02 05:06:54.87551	2012-08-02 05:13:34.27819
3292	1-3292	70	98	5177	87	3244	0	0	2012-08-02 05:04:37.384919	2012-08-02 05:13:34.27819
2261	1-2261	84	8	9216	68	4944	2905	0	2012-08-02 04:57:24.011542	2012-08-02 05:13:34.27819
4631	1-4631	46	39	171	6	9529	0	10	2012-08-02 05:13:34.698182	2012-08-02 05:13:34.698182
3778	1-3778	56	81	7060	64	3921	0	0	2012-08-02 05:07:53.038664	2012-08-02 05:13:35.935381
4357	1-4357	35	28	1228	67	744	0	0	2012-08-02 05:11:44.523013	2012-08-02 05:13:35.935381
4348	1-4348	18	28	4251	39	3583	0	0	2012-08-02 05:11:40.866253	2012-08-02 05:13:35.935381
4634	1-4634	32	64	2349	28	8559	0	0	2012-08-02 05:13:35.935381	2012-08-02 05:13:35.935381
4545	1-4545	53	2	7232	61	3050	0	0	2012-08-02 05:12:59.637879	2012-08-02 05:13:36.764155
4293	1-4293	5	75	1998	56	3877	0	6	2012-08-02 05:11:18.202758	2012-08-02 05:13:36.764155
348	1-348	80	53	2720	68	318	0	0	2012-08-02 04:44:00.677327	2012-08-02 05:13:36.764155
4636	1-4636	96	68	2081	75	6847	0	11	2012-08-02 05:13:36.764155	2012-08-02 05:13:36.764155
4606	1-4606	4	92	3050	66	6604	0	7	2012-08-02 05:13:24.388179	2012-08-02 05:13:37.184443
4578	1-4578	17	9	7610	18	8646	0	0	2012-08-02 05:13:12.930839	2012-08-02 05:13:37.990737
4481	1-4481	54	9	2390	3	1814	0	0	2012-08-02 05:12:34.223425	2012-08-02 05:13:37.990737
3522	1-3522	35	9	3454	5	2930	0	5	2012-08-02 05:06:10.627917	2012-08-02 05:13:37.990737
4619	1-4619	47	40	6681	58	8853	0	0	2012-08-02 05:13:29.759012	2012-08-02 05:13:39.62595
3842	1-3842	45	68	8464	65	7054	0	0	2012-08-02 05:08:18.529785	2012-08-02 05:13:39.62595
4006	1-4006	99	68	2609	21	2240	0	0	2012-08-02 05:09:24.038235	2012-08-02 05:13:39.62595
3880	1-3880	74	26	6355	39	8582	0	2	2012-08-02 05:08:33.679578	2012-08-02 05:13:40.056933
4637	1-4637	87	12	1763	87	3138	0	0	2012-08-02 05:13:37.184443	2012-08-02 05:13:40.465888
4544	1-4544	52	99	3121	98	8078	0	0	2012-08-02 05:12:59.250974	2012-08-02 05:13:40.874746
936	1-936	48	11	7884	95	3619	0	0	2012-08-02 04:46:35.004774	2012-08-02 05:13:42.111828
4557	1-4557	99	12	881	43	4038	0	0	2012-08-02 05:13:04.377756	2012-08-02 05:13:42.531996
4249	1-4249	5	63	6590	29	8051	0	0	2012-08-02 05:11:00.44522	2012-08-02 05:13:42.531996
1912	1-1912	47	55	7776	88	9223	0	0	2012-08-02 04:54:52.396909	2012-08-02 05:13:42.95171
3182	1-3182	89	10	8951	49	4269	0	0	2012-08-02 05:03:53.161751	2012-08-02 05:13:43.360686
4568	1-4568	3	88	6519	60	5799	0	0	2012-08-02 05:13:08.821063	2012-08-02 05:13:45.006756
3059	1-3059	9	38	2533	88	7381	0	0	2012-08-02 05:03:03.148744	2012-08-02 05:13:45.006756
3556	1-3556	31	92	3440	34	3539	0	0	2012-08-02 05:06:24.213104	2012-08-02 05:13:45.415398
4656	1-4656	79	74	4924	38	4261	0	12	2012-08-02 05:13:45.006756	2012-08-02 05:13:45.824255
4435	1-4435	9	34	2952	21	7206	0	3	2012-08-02 05:12:16.079929	2012-08-02 05:13:45.824255
3907	1-3907	34	74	5478	35	2112	0	6	2012-08-02 05:08:44.584807	2012-08-02 05:13:45.824255
953	1-953	91	50	7125	5	2035	0	0	2012-08-02 04:46:45.54304	2012-08-02 05:13:45.824255
4075	1-4075	70	79	6243	27	5333	0	0	2012-08-02 05:09:51.683759	2012-08-02 05:13:45.824255
4658	1-4658	45	37	2790	96	7868	0	0	2012-08-02 05:13:45.824255	2012-08-02 05:13:45.824255
4612	1-4612	67	69	1504	79	5115	0	0	2012-08-02 05:13:26.897676	2012-08-02 05:13:46.244541
4569	1-4569	2	69	7612	67	7286	0	0	2012-08-02 05:13:09.230075	2012-08-02 05:13:46.244541
4269	1-4269	95	33	3269	85	4046	0	3	2012-08-02 05:11:08.358442	2012-08-02 05:13:46.244541
4407	1-4407	56	19	3144	3	5587	0	6	2012-08-02 05:12:05.041225	2012-08-02 05:13:47.029058
4588	1-4588	35	28	5137	92	9037	0	0	2012-08-02 05:13:17.029915	2012-08-02 05:13:47.029058
4668	1-4668	100	97	241	25	1127	0	0	2012-08-02 05:13:49.794874	2012-08-02 05:13:49.794874
3550	1-3550	39	55	7063	52	8371	0	0	2012-08-02 05:06:21.8158	2012-08-02 05:13:50.629918
2153	1-2153	52	52	7849	43	4587	0	0	2012-08-02 04:56:36.921216	2012-08-02 05:13:50.629918
4623	1-4623	67	1	3666	55	8049	0	4	2012-08-02 05:13:31.405362	2012-08-02 05:13:50.629918
1641	1-1641	12	64	7288	13	6234	0	0	2012-08-02 04:52:49.314675	2012-08-02 05:13:50.629918
3786	1-3786	63	100	6116	58	7115	0	0	2012-08-02 05:07:56.13227	2012-08-02 05:13:54.319987
3691	1-3691	8	67	5428	39	7511	0	0	2012-08-02 05:07:18.654848	2012-08-02 05:13:54.319987
4679	1-4679	66	44	936	6	9707	0	9	2012-08-02 05:13:54.319987	2012-08-02 05:13:54.319987
4313	1-4313	23	3	6191	56	9607	0	3	2012-08-02 05:11:26.467572	2012-08-02 05:13:55.150491
4681	1-4681	89	33	100	3	8520	0	6	2012-08-02 05:13:55.150491	2012-08-02 05:13:55.150491
2843	1-2843	22	74	6437	31	8326	0	0	2012-08-02 05:01:34.434766	2012-08-02 05:13:55.572734
4682	1-4682	13	83	5184	74	7103	0	0	2012-08-02 05:13:55.572734	2012-08-02 05:13:55.572734
1821	1-1821	86	38	6438	70	6576	0	7	2012-08-02 04:54:12.070681	2012-08-02 05:13:57.585237
2938	1-2938	72	38	5326	42	7631	0	0	2012-08-02 05:02:14.003344	2012-08-02 05:13:57.585237
570	1-570	48	38	4840	10	4896	0	0	2012-08-02 04:44:29.122438	2012-08-02 05:13:57.585237
4687	1-4687	46	50	6269	38	8633	0	0	2012-08-02 05:13:57.585237	2012-08-02 05:13:57.585237
4624	1-4624	8	1	4832	87	8810	0	0	2012-08-02 05:13:31.8252	2012-08-02 05:13:58.00761
4689	1-4689	42	85	693	52	2715	0	0	2012-08-02 05:13:58.394296	2012-08-02 05:13:58.394296
4225	1-4225	28	27	8065	37	8519	0	0	2012-08-02 05:10:51.075766	2012-08-02 05:13:58.781256
4690	1-4690	100	37	6734	27	7294	0	0	2012-08-02 05:13:58.781256	2012-08-02 05:13:59.599451
3881	1-3881	42	32	9208	66	4149	0	0	2012-08-02 05:08:34.077343	2012-08-02 05:14:03.433183
3239	1-3239	58	62	8787	17	8787	0	4	2012-08-02 05:04:16.368739	2012-08-02 05:14:03.819956
3713	1-3713	12	62	9942	56	5799	0	7	2012-08-02 05:07:27.37163	2012-08-02 05:14:03.819956
4703	1-4703	31	17	5387	62	8052	0	0	2012-08-02 05:14:03.819956	2012-08-02 05:14:03.819956
2575	1-2575	86	84	7579	23	3467	0	0	2012-08-02 04:59:41.022767	2012-08-02 05:14:04.219187
4706	1-4706	24	89	166	33	9258	0	0	2012-08-02 05:14:04.994695	2012-08-02 05:14:04.994695
4670	1-4670	55	43	1614	1	5974	0	0	2012-08-02 05:13:50.629918	2012-08-02 05:14:05.408713
4371	1-4371	51	68	1518	40	5891	0	0	2012-08-02 05:11:50.311914	2012-08-02 05:14:05.408713
4418	1-4418	59	90	2880	22	8734	0	0	2012-08-02 05:12:09.339916	2012-08-02 05:14:05.408713
3020	1-3020	68	100	6864	40	9599	0	9	2012-08-02 05:02:47.360611	2012-08-02 05:14:05.408713
4643	1-4643	11	49	2837	68	8662	0	0	2012-08-02 05:13:39.62595	2012-08-02 05:14:05.85144
2513	1-2513	59	98	7696	32	7748	0	6	2012-08-02 04:59:14.283096	2012-08-02 05:14:06.674676
4613	1-4613	11	74	3346	25	3151	0	10	2012-08-02 05:13:27.306682	2012-08-02 05:14:06.674676
4710	1-4710	85	16	3218	98	8684	0	6	2012-08-02 05:14:06.674676	2012-08-02 05:14:06.674676
4517	1-4517	37	41	2192	70	3343	0	0	2012-08-02 05:12:48.454382	2012-08-02 05:14:07.492441
4520	1-4520	79	41	1840	53	7687	0	7	2012-08-02 05:12:49.62569	2012-08-02 05:14:07.492441
4525	1-4525	27	97	5585	81	7818	0	0	2012-08-02 05:12:51.636875	2012-08-02 05:14:07.492441
2315	1-2315	27	37	5703	97	2043	0	0	2012-08-02 04:57:47.491285	2012-08-02 05:14:07.913196
4538	1-4538	61	68	2179	14	6765	0	0	2012-08-02 05:12:56.886697	2012-08-02 05:14:08.310591
2375	1-2375	86	64	3493	31	6568	0	0	2012-08-02 04:58:13.60143	2012-08-02 05:14:09.990021
4548	1-4548	6	36	5051	53	8625	0	0	2012-08-02 05:13:00.853583	2012-08-02 05:14:09.990021
4718	1-4718	100	31	2949	49	3704	0	0	2012-08-02 05:14:09.990021	2012-08-02 05:14:09.990021
2474	1-2474	1	9	8793	71	7998	0	0	2012-08-02 04:58:57.305604	2012-08-02 05:14:11.205548
3465	1-3465	93	9	6004	43	5712	0	4	2012-08-02 05:05:47.562404	2012-08-02 05:14:11.205548
4721	1-4721	62	74	2106	9	7423	0	24	2012-08-02 05:14:11.205548	2012-08-02 05:14:11.205548
3337	1-3337	76	99	4158	47	3822	0	3	2012-08-02 05:04:55.617363	2012-08-02 05:14:13.260582
4320	1-4320	39	7	5332	85	5481	0	11	2012-08-02 05:11:29.340321	2012-08-02 05:14:13.260582
4726	1-4726	41	47	420	93	6750	0	0	2012-08-02 05:14:13.260582	2012-08-02 05:14:13.260582
4714	1-4714	76	14	3351	92	1999	0	0	2012-08-02 05:14:08.310591	2012-08-02 05:14:14.089556
4652	1-4652	36	14	3950	32	1489	0	2	2012-08-02 05:13:43.360686	2012-08-02 05:14:14.089556
2720	1-2720	68	92	4300	93	6081	0	2	2012-08-02 05:00:42.699572	2012-08-02 05:14:15.304693
4732	1-4732	99	9	178	27	6781	0	0	2012-08-02 05:14:15.713559	2012-08-02 05:14:15.713559
4539	1-4539	37	36	2921	1	7347	0	2	2012-08-02 05:12:57.273451	2012-08-02 05:14:17.770404
4737	1-4737	41	72	990	49	2530	0	0	2012-08-02 05:14:17.770404	2012-08-02 05:14:17.770404
4713	1-4713	66	50	284	37	1990	0	0	2012-08-02 05:14:07.913196	2012-08-02 05:14:18.179222
4649	1-4649	39	95	1556	11	3908	0	0	2012-08-02 05:13:42.111828	2012-08-02 05:14:18.179222
4390	1-4390	53	19	1009	50	8484	0	0	2012-08-02 05:11:58.067591	2012-08-02 05:14:18.179222
1894	1-1894	89	19	3915	90	4191	0	0	2012-08-02 04:54:45.006181	2012-08-02 05:14:18.179222
4738	1-4738	3	11	2068	19	8158	0	0	2012-08-02 05:14:18.179222	2012-08-02 05:14:18.179222
2861	1-2861	34	19	9953	18	3452	2125	0	2012-08-02 05:01:42.005098	2012-08-02 05:14:18.179222
4630	1-4630	58	48	5679	99	9671	0	3	2012-08-02 05:13:34.27819	2012-08-02 05:14:18.599673
3065	1-3065	8	35	5515	47	2180	0	5	2012-08-02 05:03:05.626751	2012-08-02 05:14:20.146163
4597	1-4597	20	24	3162	56	9907	0	0	2012-08-02 05:13:20.698199	2012-08-02 05:14:21.714887
1111	1-1111	93	24	5348	68	3037	0	0	2012-08-02 04:48:13.557877	2012-08-02 05:14:21.714887
3466	1-3466	51	3	8492	70	8434	0	0	2012-08-02 05:05:47.982445	2012-08-02 05:14:21.714887
3051	1-3051	21	21	8085	49	4805	0	0	2012-08-02 05:02:59.878363	2012-08-02 05:14:21.714887
4747	1-4747	77	88	1833	24	6126	0	2	2012-08-02 05:14:21.714887	2012-08-02 05:14:21.714887
4554	1-4554	80	68	5563	71	8527	0	0	2012-08-02 05:13:03.217633	2012-08-02 05:14:22.941468
4750	1-4750	74	40	1590	68	6528	0	8	2012-08-02 05:14:22.941468	2012-08-02 05:14:22.941468
4752	1-4752	91	36	490	77	5323	0	0	2012-08-02 05:14:23.770277	2012-08-02 05:14:23.770277
4605	1-4605	10	18	3387	86	9941	0	0	2012-08-02 05:13:23.968085	2012-08-02 05:14:24.201306
2698	1-2698	88	18	4568	32	7063	0	0	2012-08-02 05:00:33.286062	2012-08-02 05:14:24.201306
1962	1-1962	48	69	2475	40	1248	0	0	2012-08-02 04:55:14.419217	2012-08-02 05:14:24.201306
3049	1-3049	95	18	5792	35	3880	0	0	2012-08-02 05:02:59.083256	2012-08-02 05:14:24.201306
2682	1-2682	16	18	8139	66	5492	0	0	2012-08-02 05:00:26.536018	2012-08-02 05:14:24.201306
1498	1-1498	78	86	7813	24	2363	1291	0	2012-08-02 04:51:38.28418	2012-08-02 05:14:24.201306
4246	1-4246	51	76	5409	73	1406	0	3	2012-08-02 05:10:59.273947	2012-08-02 05:14:24.632523
4461	1-4461	77	1	9686	64	6683	0	2	2012-08-02 05:12:26.288802	2012-08-02 05:14:25.803474
4045	1-4045	55	96	1491	46	4282	0	0	2012-08-02 05:09:39.760645	2012-08-02 05:14:25.803474
4611	1-4611	69	96	3269	74	7111	0	0	2012-08-02 05:13:26.487475	2012-08-02 05:14:25.803474
4616	1-4616	62	39	8062	64	3151	0	0	2012-08-02 05:13:28.521691	2012-08-02 05:14:25.803474
4760	1-4760	22	68	27	85	4438	0	0	2012-08-02 05:14:27.007793	2012-08-02 05:14:27.007793
3103	1-3103	56	36	6111	83	8513	0	0	2012-08-02 05:03:21.061741	2012-08-02 05:14:27.416911
3714	1-3714	73	49	5352	36	7363	0	0	2012-08-02 05:07:27.769489	2012-08-02 05:14:27.416911
4633	1-4633	83	5	9813	3	7414	0	0	2012-08-02 05:13:35.526802	2012-08-02 05:14:27.416911
4608	1-4608	35	33	2937	83	7854	0	8	2012-08-02 05:13:25.216766	2012-08-02 05:14:27.416911
3960	1-3960	59	49	5049	5	9323	0	4	2012-08-02 05:09:05.716848	2012-08-02 05:14:27.416911
4278	1-4278	41	73	4836	39	5231	0	0	2012-08-02 05:11:12.060083	2012-08-02 05:14:28.986154
4768	1-4768	22	83	15	49	9388	0	6	2012-08-02 05:14:30.156955	2012-08-02 05:14:30.156955
4769	1-4769	22	7	704	87	6866	0	0	2012-08-02 05:14:30.544053	2012-08-02 05:14:30.544053
3921	1-3921	15	91	4604	81	5510	0	2	2012-08-02 05:08:50.08911	2012-08-02 05:14:30.941649
4749	1-4749	41	81	3787	1	7435	0	0	2012-08-02 05:14:22.532899	2012-08-02 05:14:30.941649
499	1-499	8	91	6011	75	1436	0	0	2012-08-02 04:44:12.021818	2012-08-02 05:14:30.941649
4770	1-4770	68	1	1950	91	6883	0	8	2012-08-02 05:14:30.941649	2012-08-02 05:14:30.941649
4653	1-4653	46	65	4597	32	7613	0	0	2012-08-02 05:13:43.769347	2012-08-02 05:14:31.75932
4772	1-4772	25	75	349	66	9142	0	0	2012-08-02 05:14:31.75932	2012-08-02 05:14:31.75932
4465	1-4465	41	10	8750	55	9760	0	0	2012-08-02 05:12:27.837457	2012-08-02 05:14:33.383205
3678	1-3678	90	78	8017	71	4103	0	0	2012-08-02 05:07:13.418143	2012-08-02 05:14:33.383205
4776	1-4776	97	35	1377	75	8547	0	0	2012-08-02 05:14:33.383205	2012-08-02 05:14:33.383205
2014	1-2014	68	89	2501	13	2905	0	0	2012-08-02 04:55:36.870431	2012-08-02 05:14:35.018568
4446	1-4446	99	96	5370	89	9986	0	4	2012-08-02 05:12:20.377749	2012-08-02 05:14:35.018568
4128	1-4128	79	66	1234	6	642	0	0	2012-08-02 05:10:12.823397	2012-08-02 05:14:37.040092
4702	1-4702	4	66	3058	65	7079	0	4	2012-08-02 05:14:03.433183	2012-08-02 05:14:37.040092
3544	1-3544	1	66	9624	77	9096	0	0	2012-08-02 05:06:19.40801	2012-08-02 05:14:37.040092
3323	1-3323	35	81	8274	6	4002	169	0	2012-08-02 05:04:49.927235	2012-08-02 05:14:37.040092
4753	1-4753	38	24	556	18	3236	0	3	2012-08-02 05:14:24.201306	2012-08-02 05:14:37.449131
4593	1-4593	40	31	7293	42	8947	0	0	2012-08-02 05:13:19.040517	2012-08-02 05:14:37.846907
4212	1-4212	43	41	4245	34	3048	0	0	2012-08-02 05:10:46.026586	2012-08-02 05:14:37.846907
4787	1-4787	90	76	3480	27	2499	0	9	2012-08-02 05:14:37.846907	2012-08-02 05:14:37.846907
2196	1-2196	14	48	9884	4	2353	0	0	2012-08-02 04:56:55.703557	2012-08-02 05:14:40.815418
4794	1-4794	74	74	1868	48	7802	0	0	2012-08-02 05:14:40.815418	2012-08-02 05:14:40.815418
4727	1-4727	10	42	2329	76	9562	0	0	2012-08-02 05:14:13.68072	2012-08-02 05:14:41.227675
4276	1-4276	52	93	2625	2	2121	0	0	2012-08-02 05:11:11.253459	2012-08-02 05:14:41.227675
4388	1-4388	3	87	8245	52	7932	0	3	2012-08-02 05:11:57.272461	2012-08-02 05:14:41.227675
3406	1-3406	96	61	2419	17	7483	0	2	2012-08-02 05:05:23.761621	2012-08-02 05:14:41.227675
4505	1-4505	55	41	4717	88	5131	0	0	2012-08-02 05:12:43.658947	2012-08-02 05:14:41.638569
4757	1-4757	38	64	5242	96	7341	0	0	2012-08-02 05:14:25.803474	2012-08-02 05:14:41.638569
4796	1-4796	32	7	2702	41	3899	0	13	2012-08-02 05:14:41.638569	2012-08-02 05:14:41.638569
4531	1-4531	20	46	5541	81	3998	0	0	2012-08-02 05:12:54.046988	2012-08-02 05:14:42.025573
3947	1-3947	72	78	3927	20	1825	0	0	2012-08-02 05:09:00.514587	2012-08-02 05:14:42.025573
4430	1-4430	62	36	5400	39	9054	0	0	2012-08-02 05:12:14.10217	2012-08-02 05:14:42.025573
4748	1-4748	41	36	2342	78	9026	0	0	2012-08-02 05:14:22.135269	2012-08-02 05:14:42.025573
4712	1-4712	89	81	8636	41	7464	0	0	2012-08-02 05:14:07.492441	2012-08-02 05:14:42.025573
4659	1-4659	36	85	3064	69	5517	0	4	2012-08-02 05:13:46.244541	2012-08-02 05:14:44.456345
4667	1-4667	55	88	4165	59	894	0	3	2012-08-02 05:13:49.393851	2012-08-02 05:14:45.240543
1445	1-1445	16	57	8932	59	3967	0	0	2012-08-02 04:51:09.395677	2012-08-02 05:14:45.240543
4651	1-4651	56	60	7116	98	5448	0	2	2012-08-02 05:13:42.95171	2012-08-02 05:14:45.638414
4806	1-4806	64	72	378	60	3113	0	0	2012-08-02 05:14:45.638414	2012-08-02 05:14:45.638414
3448	1-3448	49	45	8661	23	5441	0	0	2012-08-02 05:05:40.668426	2012-08-02 05:14:46.047468
4765	1-4765	20	53	3454	88	3610	0	0	2012-08-02 05:14:28.986154	2012-08-02 05:14:46.047468
4284	1-4284	18	58	3670	11	8192	0	0	2012-08-02 05:11:14.468292	2012-08-02 05:14:46.047468
2812	1-2812	22	49	5200	5	5302	0	0	2012-08-02 05:01:21.419265	2012-08-02 05:14:47.262559
4785	1-4785	12	6	867	66	7623	0	0	2012-08-02 05:14:37.040092	2012-08-02 05:14:47.671524
4397	1-4397	45	9	6744	12	6331	0	0	2012-08-02 05:12:00.951576	2012-08-02 05:14:48.47809
4704	1-4704	24	23	1387	17	8557	0	0	2012-08-02 05:14:04.219187	2012-08-02 05:14:48.47809
4813	1-4813	59	99	180	95	1672	0	2	2012-08-02 05:14:48.47809	2012-08-02 05:14:48.47809
4814	1-4814	52	46	469	62	8783	0	11	2012-08-02 05:14:48.898023	2012-08-02 05:14:48.898023
4363	1-4363	24	90	4110	83	8747	0	0	2012-08-02 05:11:46.997372	2012-08-02 05:14:49.317875
4131	1-4131	48	93	8267	95	7197	0	4	2012-08-02 05:10:13.98341	2012-08-02 05:14:49.715583
4807	1-4807	7	11	6586	45	9064	0	10	2012-08-02 05:14:46.047468	2012-08-02 05:14:49.715583
4522	1-4522	85	97	7849	91	4722	0	0	2012-08-02 05:12:50.443382	2012-08-02 05:14:49.715583
4797	1-4797	12	41	3535	36	8061	0	0	2012-08-02 05:14:42.025573	2012-08-02 05:14:50.135498
4472	1-4472	86	24	2503	43	7410	0	0	2012-08-02 05:12:30.677156	2012-08-02 05:14:52.124
4823	1-4823	53	39	1430	24	9390	0	0	2012-08-02 05:14:52.521889	2012-08-02 05:14:52.521889
4826	1-4826	46	5	1562	62	5011	0	6	2012-08-02 05:14:53.74809	2012-08-02 05:14:53.74809
4725	1-4725	54	77	8782	19	8445	0	0	2012-08-02 05:14:12.851862	2012-08-02 05:14:54.996669
4460	1-4460	88	74	1223	95	453	0	3	2012-08-02 05:12:25.902433	2012-08-02 05:14:54.996669
4829	1-4829	85	95	2427	61	9637	0	4	2012-08-02 05:14:54.996669	2012-08-02 05:14:54.996669
868	1-868	88	61	8076	24	3811	1092	2	2012-08-02 04:46:00.321812	2012-08-02 05:14:54.996669
4601	1-4601	74	41	6483	64	7805	0	0	2012-08-02 05:13:22.32224	2012-08-02 05:14:56.200925
2953	1-2953	63	91	4701	42	6003	0	0	2012-08-02 05:02:20.190438	2012-08-02 05:14:56.598715
4691	1-4691	38	71	7306	35	7737	0	0	2012-08-02 05:13:59.223602	2012-08-02 05:14:57.813942
4707	1-4707	79	40	6023	61	4991	0	0	2012-08-02 05:14:05.408713	2012-08-02 05:14:57.813942
4836	1-4836	3	46	1874	71	3073	0	0	2012-08-02 05:14:57.813942	2012-08-02 05:14:57.813942
2924	1-2924	88	87	3793	78	3795	0	10	2012-08-02 05:02:08.203157	2012-08-02 05:14:58.642616
3232	1-3232	49	87	6236	16	5536	0	0	2012-08-02 05:04:13.596031	2012-08-02 05:14:58.642616
4838	1-4838	31	78	1732	87	4155	0	0	2012-08-02 05:14:58.642616	2012-08-02 05:14:58.642616
3477	1-3477	4	87	4697	52	3307	460	0	2012-08-02 05:05:52.379416	2012-08-02 05:14:58.642616
3392	1-3392	83	89	2675	42	3353	0	9	2012-08-02 05:05:18.160133	2012-08-02 05:14:59.051711
2969	1-2969	49	53	6024	76	2741	0	0	2012-08-02 05:02:26.698324	2012-08-02 05:14:59.051711
4795	1-4795	40	76	7150	91	9402	0	0	2012-08-02 05:14:41.227675	2012-08-02 05:14:59.051711
4761	1-4761	30	83	9382	49	9748	0	0	2012-08-02 05:14:27.416911	2012-08-02 05:14:59.47151
1604	1-1604	25	68	9937	83	3503	2132	0	2012-08-02 04:52:32.156544	2012-08-02 05:14:59.47151
1429	1-1429	75	79	8116	16	4637	0	0	2012-08-02 04:51:00.300434	2012-08-02 05:15:00.289234
4516	1-4516	21	88	8256	73	5119	0	4	2012-08-02 05:12:48.06776	2012-08-02 05:15:00.289234
4842	1-4842	84	86	882	79	8628	0	10	2012-08-02 05:15:00.289234	2012-08-02 05:15:00.289234
3663	1-3663	40	20	3328	47	1208	0	0	2012-08-02 05:07:07.452254	2012-08-02 05:15:00.720726
4661	1-4661	12	70	6422	91	9249	0	4	2012-08-02 05:13:47.029058	2012-08-02 05:15:02.355599
4811	1-4811	41	66	3211	52	7761	0	0	2012-08-02 05:14:47.671524	2012-08-02 05:15:02.764098
4816	1-4816	35	95	1745	97	4929	0	0	2012-08-02 05:14:49.715583	2012-08-02 05:15:03.581848
4845	1-4845	76	95	649	80	1892	0	0	2012-08-02 05:15:01.537793	2012-08-02 05:15:03.581848
4850	1-4850	20	43	687	95	4144	0	0	2012-08-02 05:15:03.581848	2012-08-02 05:15:03.581848
2379	1-2379	24	55	4172	59	3063	0	0	2012-08-02 04:58:15.357394	2012-08-02 05:15:04.001638
4852	1-4852	6	30	2739	27	5871	0	2	2012-08-02 05:15:04.399384	2012-08-02 05:15:04.399384
4338	1-4338	1	34	4211	76	7622	0	0	2012-08-02 05:11:36.744591	2012-08-02 05:15:06.016019
1264	1-1264	60	77	4846	27	1669	0	0	2012-08-02 04:49:34.908889	2012-08-02 05:15:10.235872
2461	1-2461	44	61	4312	99	784	0	0	2012-08-02 04:58:51.463839	2012-08-02 05:15:10.235872
4294	1-4294	34	80	3312	83	3764	0	0	2012-08-02 05:11:18.61133	2012-08-02 05:15:11.782617
4746	1-4746	96	17	5449	30	1562	0	2	2012-08-02 05:14:21.328531	2012-08-02 05:15:11.782617
1129	1-1129	59	21	8833	93	7736	0	0	2012-08-02 04:48:25.212032	2012-08-02 05:15:13.727019
2194	1-2194	5	85	3895	72	1578	0	0	2012-08-02 04:56:54.841605	2012-08-02 05:15:13.727019
3885	1-3885	87	64	3149	72	3523	0	8	2012-08-02 05:08:35.690807	2012-08-02 05:15:13.727019
2560	1-2560	52	72	8511	21	1793	1793	0	2012-08-02 04:59:34.658515	2012-08-02 05:15:13.727019
700	1-700	75	61	7568	78	8077	0	0	2012-08-02 04:44:58.901681	2012-08-02 05:15:14.898584
4867	1-4867	69	86	2099	61	8521	0	0	2012-08-02 05:15:10.235872	2012-08-02 05:15:14.898584
4092	1-4092	64	70	4481	7	8585	0	0	2012-08-02 05:09:58.468161	2012-08-02 05:15:14.898584
4853	1-4853	13	78	840	7	2295	0	2	2012-08-02 05:15:04.808197	2012-08-02 05:15:14.898584
4731	1-4731	15	99	4023	86	8981	0	0	2012-08-02 05:14:15.304693	2012-08-02 05:15:14.898584
2492	1-2492	20	27	7287	94	2286	0	0	2012-08-02 04:59:05.211557	2012-08-02 05:15:15.296323
1404	1-1404	83	52	9539	98	3117	0	0	2012-08-02 04:50:47.639694	2012-08-02 05:15:15.296323
4800	1-4800	80	27	8327	99	9139	0	0	2012-08-02 05:14:43.240985	2012-08-02 05:15:15.296323
4535	1-4535	98	88	9192	97	7894	0	0	2012-08-02 05:12:55.715573	2012-08-02 05:15:18.091409
4871	1-4871	70	83	3166	17	8915	0	8	2012-08-02 05:15:11.782617	2012-08-02 05:15:18.875846
4880	1-4880	24	98	1266	27	9074	0	0	2012-08-02 05:15:15.296323	2012-08-02 05:15:19.273615
4650	1-4650	21	81	4992	12	7634	0	0	2012-08-02 05:13:42.531996	2012-08-02 05:15:20.411447
4717	1-4717	3	89	1382	9	1724	0	6	2012-08-02 05:14:09.581728	2012-08-02 05:15:20.411447
4876	1-4876	22	72	2209	21	4721	0	10	2012-08-02 05:15:13.727019	2012-08-02 05:15:23.549453
4840	1-4840	63	49	5286	8	3654	0	0	2012-08-02 05:14:59.47151	2012-08-02 05:15:24.344987
4904	1-4904	12	29	43	92	2496	0	0	2012-08-02 05:15:24.731268	2012-08-02 05:15:24.731268
4913	1-4913	82	60	4640	77	7195	0	0	2012-08-02 05:15:28.189303	2012-08-02 05:15:28.189303
3207	1-3207	17	75	7374	41	9866	0	0	2012-08-02 05:04:03.268458	2012-08-02 05:15:28.575814
3748	1-3748	58	75	2848	77	6646	0	11	2012-08-02 05:07:41.194993	2012-08-02 05:15:28.575814
4914	1-4914	64	41	3202	96	9067	0	0	2012-08-02 05:15:28.575814	2012-08-02 05:15:28.575814
4551	1-4551	73	10	3420	73	8465	0	0	2012-08-02 05:13:02.024505	2012-08-02 05:15:28.97407
2740	1-2740	10	10	1330	41	361	0	0	2012-08-02 05:00:51.186817	2012-08-02 05:15:28.97407
4915	1-4915	65	57	326	10	2248	0	2	2012-08-02 05:15:28.97407	2012-08-02 05:15:28.97407
4266	1-4266	94	42	7983	45	5566	0	0	2012-08-02 05:11:07.143361	2012-08-02 05:15:29.37165
4448	1-4448	73	89	2413	53	7648	0	0	2012-08-02 05:12:21.184154	2012-08-02 05:15:29.7694
4917	1-4917	17	53	3655	94	4675	0	11	2012-08-02 05:15:29.7694	2012-08-02 05:15:29.7694
4093	1-4093	73	64	4908	78	4308	0	0	2012-08-02 05:09:58.86569	2012-08-02 05:15:32.100499
4923	1-4923	31	23	391	64	7230	0	0	2012-08-02 05:15:32.100499	2012-08-02 05:15:32.100499
2999	1-2999	85	81	9350	84	6543	0	10	2012-08-02 05:02:38.875532	2012-08-02 05:15:32.487324
4025	1-4025	52	92	4825	81	6067	0	0	2012-08-02 05:09:31.805332	2012-08-02 05:15:32.487324
4754	1-4754	58	92	3653	42	8495	0	2	2012-08-02 05:14:24.632523	2012-08-02 05:15:32.487324
4924	1-4924	67	41	3270	92	6732	0	10	2012-08-02 05:15:32.487324	2012-08-02 05:15:32.487324
3218	1-3218	99	23	3228	47	1700	0	5	2012-08-02 05:04:07.851134	2012-08-02 05:15:32.898334
4533	1-4533	30	93	9674	97	7775	0	2	2012-08-02 05:12:54.887207	2012-08-02 05:15:36.024888
3668	1-3668	93	75	6371	10	9801	0	0	2012-08-02 05:07:09.407348	2012-08-02 05:15:36.809367
4935	1-4935	12	10	2103	75	6486	0	0	2012-08-02 05:15:36.809367	2012-08-02 05:15:36.809367
4833	1-4833	94	42	2720	34	5175	0	0	2012-08-02 05:14:56.598715	2012-08-02 05:15:37.582481
4937	1-4937	69	79	243	39	9955	0	0	2012-08-02 05:15:37.582481	2012-08-02 05:15:37.582481
4889	1-4889	96	2	2545	68	2087	0	0	2012-08-02 05:15:18.875846	2012-08-02 05:15:38.356255
3259	1-3259	55	33	5806	52	2568	0	0	2012-08-02 05:04:24.06108	2012-08-02 05:15:38.356255
4892	1-4892	95	2	5373	33	6663	0	3	2012-08-02 05:15:20.035865	2012-08-02 05:15:38.356255
3583	1-3583	21	88	9092	51	8001	0	0	2012-08-02 05:06:35.208731	2012-08-02 05:15:38.742695
4793	1-4793	19	11	2810	9	539	0	0	2012-08-02 05:14:40.277411	2012-08-02 05:15:38.742695
4575	1-4575	97	88	3415	95	389	0	5	2012-08-02 05:13:11.704705	2012-08-02 05:15:38.742695
4940	1-4940	35	20	532	88	9488	0	0	2012-08-02 05:15:38.742695	2012-08-02 05:15:38.742695
4385	1-4385	100	95	3918	32	1714	0	0	2012-08-02 05:11:56.057086	2012-08-02 05:15:41.880488
2785	1-2785	53	60	7167	7	5899	0	0	2012-08-02 05:01:10.073006	2012-08-02 05:15:41.880488
4429	1-4429	54	21	7077	85	9892	0	0	2012-08-02 05:12:13.704601	2012-08-02 05:15:42.267241
4950	1-4950	13	72	2850	42	6634	0	0	2012-08-02 05:15:42.665257	2012-08-02 05:15:42.665257
4916	1-4916	26	32	2981	24	5394	0	4	2012-08-02 05:15:29.37165	2012-08-02 05:15:43.074054
3346	1-3346	55	70	5617	97	4096	0	0	2012-08-02 05:04:59.230232	2012-08-02 05:15:45.437797
3601	1-3601	65	98	8357	55	9465	0	0	2012-08-02 05:06:42.534583	2012-08-02 05:15:46.996014
3876	1-3876	60	98	6491	95	7374	0	3	2012-08-02 05:08:32.109189	2012-08-02 05:15:46.996014
3975	1-3975	15	43	7400	97	4133	0	0	2012-08-02 05:09:11.706112	2012-08-02 05:15:48.155647
4964	1-4964	31	61	434	43	4918	0	0	2012-08-02 05:15:48.155647	2012-08-02 05:15:48.155647
3666	1-3666	9	93	9058	13	6613	0	0	2012-08-02 05:07:08.634132	2012-08-02 05:15:49.735428
4968	1-4968	32	94	614	93	6484	0	11	2012-08-02 05:15:49.735428	2012-08-02 05:15:49.735428
4088	1-4088	72	38	1461	4	1338	0	3	2012-08-02 05:09:56.876984	2012-08-02 05:15:50.144595
3761	1-3761	24	70	2522	59	671	0	0	2012-08-02 05:07:46.332132	2012-08-02 05:15:50.531131
4253	1-4253	55	2	1876	17	3947	0	0	2012-08-02 05:11:02.013767	2012-08-02 05:15:52.100088
1595	1-1595	56	86	7231	77	6655	0	0	2012-08-02 04:52:28.333649	2012-08-02 05:15:52.486725
4899	1-4899	60	86	782	7	3999	0	0	2012-08-02 05:15:22.77573	2012-08-02 05:15:52.486725
4863	1-4863	22	60	4696	94	4964	0	0	2012-08-02 05:15:08.678647	2012-08-02 05:15:52.486725
4856	1-4856	74	76	3299	74	4239	0	0	2012-08-02 05:15:06.016019	2012-08-02 05:15:54.420207
4869	1-4869	48	100	751	38	964	0	0	2012-08-02 05:15:11.020348	2012-08-02 05:15:54.818026
4981	1-4981	82	26	103	100	2373	0	0	2012-08-02 05:15:54.818026	2012-08-02 05:15:54.818026
4140	1-4140	17	21	7974	26	4940	424	0	2012-08-02 05:10:17.563345	2012-08-02 05:15:54.818026
4029	1-4029	52	60	3186	35	817	0	0	2012-08-02 05:09:33.407403	2012-08-02 05:15:56.066553
3456	1-3456	54	50	8136	15	3309	0	0	2012-08-02 05:05:43.97208	2012-08-02 05:15:56.066553
4669	1-4669	92	31	2164	61	843	0	0	2012-08-02 05:13:50.205083	2012-08-02 05:15:56.464244
4780	1-4780	77	13	5391	64	5804	0	0	2012-08-02 05:14:35.018568	2012-08-02 05:15:58.000139
4779	1-4779	35	72	2959	43	3070	0	0	2012-08-02 05:14:34.620928	2012-08-02 05:15:58.784306
4177	1-4177	60	77	6933	43	4545	0	0	2012-08-02 05:10:32.22489	2012-08-02 05:15:59.170989
4335	1-4335	97	85	7681	47	2813	0	10	2012-08-02 05:11:35.496064	2012-08-02 05:15:59.170989
3863	1-3863	79	64	7192	29	4608	0	0	2012-08-02 05:08:26.905003	2012-08-02 05:15:59.568852
4993	1-4993	16	74	3298	64	6123	0	4	2012-08-02 05:15:59.568852	2012-08-02 05:15:59.568852
4805	1-4805	72	79	1560	88	7831	0	11	2012-08-02 05:14:45.240543	2012-08-02 05:16:01.225997
4832	1-4832	16	64	5229	41	9751	0	5	2012-08-02 05:14:56.200925	2012-08-02 05:16:02.783704
2725	1-2725	47	69	6682	57	9101	0	0	2012-08-02 05:00:44.820942	2012-08-02 05:16:02.783704
5001	1-5001	32	41	1290	92	9459	0	14	2012-08-02 05:16:02.783704	2012-08-02 05:16:02.783704
59	1-59	3	61	8409	68	4741	0	0	2012-08-02 04:43:54.100888	2012-08-02 05:16:03.194659
4589	1-4589	23	12	3922	63	6917	0	0	2012-08-02 05:13:17.438781	2012-08-02 05:16:04.36569
4755	1-4755	28	93	3596	41	1421	0	0	2012-08-02 05:14:25.03043	2012-08-02 05:16:04.36569
4878	1-4878	70	44	1245	41	1048	0	0	2012-08-02 05:15:14.511738	2012-08-02 05:16:04.36569
5005	1-5005	47	66	1897	12	9952	0	5	2012-08-02 05:16:04.36569	2012-08-02 05:16:04.36569
4989	1-4989	12	52	463	100	4613	0	0	2012-08-02 05:15:58.000139	2012-08-02 05:16:04.763589
4848	1-4848	32	52	3463	66	9002	0	0	2012-08-02 05:15:02.764098	2012-08-02 05:16:05.957841
4975	1-4975	6	15	8966	86	8497	0	11	2012-08-02 05:15:52.486725	2012-08-02 05:16:06.753799
4979	1-4979	45	28	3393	15	5322	0	11	2012-08-02 05:15:54.033475	2012-08-02 05:16:06.753799
4078	1-4078	67	28	8266	58	8354	0	0	2012-08-02 05:09:52.910695	2012-08-02 05:16:06.753799
5011	1-5011	93	61	1477	28	9744	0	7	2012-08-02 05:16:06.753799	2012-08-02 05:16:06.753799
4851	1-4851	90	45	1510	55	3228	0	2	2012-08-02 05:15:04.001638	2012-08-02 05:16:08.797785
3589	1-3589	74	19	7143	44	7350	0	0	2012-08-02 05:06:37.62853	2012-08-02 05:16:09.206626
4891	1-4891	45	36	5071	51	5204	0	2	2012-08-02 05:15:19.649254	2012-08-02 05:16:09.206626
3040	1-3040	48	10	1906	43	107	0	0	2012-08-02 05:02:55.447777	2012-08-02 05:16:11.283852
3847	1-3847	1	20	4335	29	1660	0	5	2012-08-02 05:08:20.496818	2012-08-02 05:16:11.283852
5022	1-5022	70	29	232	64	5166	0	0	2012-08-02 05:16:11.283852	2012-08-02 05:16:11.283852
5017	1-5017	28	51	6946	19	5448	0	0	2012-08-02 05:16:09.206626	2012-08-02 05:16:11.703732
4984	1-4984	32	35	428	50	7313	0	0	2012-08-02 05:15:56.066553	2012-08-02 05:16:12.908004
4592	1-4592	38	88	7196	11	7875	0	4	2012-08-02 05:13:18.642839	2012-08-02 05:16:12.908004
3521	1-3521	23	78	9140	11	5694	0	2	2012-08-02 05:06:10.19277	2012-08-02 05:16:12.908004
5026	1-5026	66	84	702	35	945	0	0	2012-08-02 05:16:12.908004	2012-08-02 05:16:12.908004
3699	1-3699	7	2	8635	76	9401	0	0	2012-08-02 05:07:21.847672	2012-08-02 05:16:13.736439
1460	1-1460	78	73	9194	12	3676	0	0	2012-08-02 04:51:17.769512	2012-08-02 05:16:14.134343
4675	1-4675	25	30	2384	21	2421	0	5	2012-08-02 05:13:52.662599	2012-08-02 05:16:14.134343
3812	1-3812	11	73	5623	33	3392	0	0	2012-08-02 05:08:06.410225	2012-08-02 05:16:14.134343
5029	1-5029	87	23	2085	30	6154	0	3	2012-08-02 05:16:14.134343	2012-08-02 05:16:14.134343
4855	1-4855	13	30	1168	80	7291	0	11	2012-08-02 05:15:05.618416	2012-08-02 05:16:14.543307
5030	1-5030	3	80	1747	28	3566	0	2	2012-08-02 05:16:14.543307	2012-08-02 05:16:14.941123
2965	1-2965	31	4	7908	12	9384	0	0	2012-08-02 05:02:25.063259	2012-08-02 05:16:15.338991
4974	1-4974	67	17	1542	10	6284	0	0	2012-08-02 05:15:52.100088	2012-08-02 05:16:15.338991
4901	1-4901	7	3	1538	72	4887	0	0	2012-08-02 05:15:23.549453	2012-08-02 05:16:18.155912
4047	1-4047	23	54	4060	42	9225	0	4	2012-08-02 05:09:40.578339	2012-08-02 05:16:18.155912
4988	1-4988	88	27	24	14	182	0	0	2012-08-02 05:15:57.613309	2012-08-02 05:16:19.371326
4647	1-4647	48	18	2858	69	5879	0	0	2012-08-02 05:13:41.28347	2012-08-02 05:16:20.984519
4218	1-4218	55	18	5578	87	9429	0	0	2012-08-02 05:10:48.357897	2012-08-02 05:16:20.984519
497	1-497	39	44	6254	3	1625	0	0	2012-08-02 04:44:11.800672	2012-08-02 05:16:21.393145
4621	1-4621	81	14	9036	82	2187	0	0	2012-08-02 05:13:30.576546	2012-08-02 05:16:21.393145
4103	1-4103	35	61	3681	46	7584	0	0	2012-08-02 05:10:02.887103	2012-08-02 05:16:21.393145
4847	1-4847	70	79	3342	70	9367	0	0	2012-08-02 05:15:02.355599	2012-08-02 05:16:22.199817
4059	1-4059	98	13	2258	43	1236	0	0	2012-08-02 05:09:45.373138	2012-08-02 05:16:22.597615
4992	1-4992	93	47	1497	61	7013	0	0	2012-08-02 05:15:59.170989	2012-08-02 05:16:22.597615
5050	1-5050	76	61	435	13	6287	0	0	2012-08-02 05:16:22.597615	2012-08-02 05:16:22.597615
5054	1-5054	70	83	946	17	1589	0	11	2012-08-02 05:16:24.210907	2012-08-02 05:16:24.210907
3287	1-3287	88	90	5036	57	4148	0	3	2012-08-02 05:04:35.385091	2012-08-02 05:16:25.823833
4939	1-4939	83	87	2547	2	7975	0	2	2012-08-02 05:15:38.356255	2012-08-02 05:16:26.654127
4543	1-4543	56	59	5088	79	6817	0	0	2012-08-02 05:12:58.853825	2012-08-02 05:16:27.858154
5063	1-5063	16	79	1931	59	8174	0	0	2012-08-02 05:16:27.858154	2012-08-02 05:16:27.858154
1032	1-1032	88	59	5620	5	1042	645	0	2012-08-02 04:47:27.055205	2012-08-02 05:16:27.858154
5064	1-5064	67	92	1824	85	5904	0	5	2012-08-02 05:16:28.255966	2012-08-02 05:16:28.255966
4990	1-4990	62	77	1362	98	1720	0	0	2012-08-02 05:15:58.397648	2012-08-02 05:16:28.653743
4991	1-4991	38	43	2925	3	3706	0	0	2012-08-02 05:15:58.784306	2012-08-02 05:16:29.062886
3326	1-3326	6	24	9313	3	4924	0	0	2012-08-02 05:04:51.14282	2012-08-02 05:16:29.062886
4970	1-4970	2	59	2810	5	3483	0	9	2012-08-02 05:15:50.531131	2012-08-02 05:16:29.471472
5067	1-5067	72	57	1775	59	3420	0	8	2012-08-02 05:16:29.471472	2012-08-02 05:16:29.471472
4864	1-4864	4	12	4130	93	7836	0	4	2012-08-02 05:15:09.065213	2012-08-02 05:16:31.526488
5072	1-5072	12	31	480	12	7787	0	0	2012-08-02 05:16:31.526488	2012-08-02 05:16:31.526488
5074	1-5074	25	95	3159	94	3910	0	0	2012-08-02 05:16:32.343791	2012-08-02 05:16:32.343791
4217	1-4217	24	74	7431	9	2805	0	21	2012-08-02 05:10:47.98265	2012-08-02 05:16:32.741741
5066	1-5066	25	3	1922	71	6122	0	0	2012-08-02 05:16:29.062886	2012-08-02 05:16:32.741741
999	1-999	36	26	3455	59	2708	0	9	2012-08-02 04:47:10.526147	2012-08-02 05:16:32.741741
5075	1-5075	96	39	6002	26	9603	0	0	2012-08-02 05:16:32.741741	2012-08-02 05:16:32.741741
5049	1-5049	37	70	4414	96	1567	0	0	2012-08-02 05:16:22.199817	2012-08-02 05:16:35.183293
5082	1-5082	93	66	813	55	5386	0	0	2012-08-02 05:16:35.63646	2012-08-02 05:16:35.63646
5053	1-5053	78	32	625	62	1374	0	6	2012-08-02 05:16:23.801884	2012-08-02 05:16:36.862576
3510	1-3510	84	4	6219	67	7490	680	0	2012-08-02 05:06:05.817803	2012-08-02 05:16:36.862576
4404	1-4404	98	61	9788	77	7304	0	0	2012-08-02 05:12:03.825315	2012-08-02 05:16:37.68008
3108	1-3108	25	91	5737	61	3294	2484	3	2012-08-02 05:03:23.105396	2012-08-02 05:16:37.68008
4890	1-4890	7	27	3647	60	7101	0	10	2012-08-02 05:15:19.273615	2012-08-02 05:16:39.70207
3746	1-3746	11	23	7954	80	9877	0	0	2012-08-02 05:07:40.333223	2012-08-02 05:16:39.70207
5033	1-5033	29	71	4167	59	7437	0	0	2012-08-02 05:16:15.747759	2012-08-02 05:16:39.70207
3747	1-3747	98	49	9128	98	3069	0	0	2012-08-02 05:07:40.786308	2012-08-02 05:16:39.70207
2822	1-2822	71	71	4719	16	2970	0	0	2012-08-02 05:01:25.584018	2012-08-02 05:16:39.70207
3915	1-3915	85	44	6289	97	6028	0	0	2012-08-02 05:08:47.744892	2012-08-02 05:16:40.519532
4777	1-4777	94	68	5691	93	8988	0	3	2012-08-02 05:14:33.792333	2012-08-02 05:16:40.519532
4745	1-4745	8	68	2771	83	8260	0	9	2012-08-02 05:14:20.941559	2012-08-02 05:16:40.519532
5094	1-5094	12	40	4813	68	7194	0	10	2012-08-02 05:16:40.519532	2012-08-02 05:16:40.519532
2565	1-2565	98	93	9706	40	7105	933	0	2012-08-02 04:59:36.779572	2012-08-02 05:16:40.519532
4887	1-4887	34	97	3863	76	7119	0	0	2012-08-02 05:15:18.091409	2012-08-02 05:16:42.232304
3104	1-3104	70	42	7025	33	4208	0	6	2012-08-02 05:03:21.448424	2012-08-02 05:16:42.232304
4428	1-4428	87	97	7058	7	6534	0	0	2012-08-02 05:12:13.317715	2012-08-02 05:16:42.232304
4957	1-4957	68	97	8951	36	9715	0	6	2012-08-02 05:15:45.437797	2012-08-02 05:16:42.232304
4834	1-4834	47	26	1890	41	4034	0	0	2012-08-02 05:14:57.007838	2012-08-02 05:16:43.845218
4879	1-4879	39	7	9834	99	9632	0	0	2012-08-02 05:15:14.898584	2012-08-02 05:16:43.845218
5023	1-5023	46	19	2103	81	7066	0	2	2012-08-02 05:16:11.703732	2012-08-02 05:16:43.845218
4862	1-4862	81	52	920	37	6923	0	0	2012-08-02 05:15:08.291582	2012-08-02 05:16:44.276653
5103	1-5103	78	37	4462	29	5275	0	0	2012-08-02 05:16:44.276653	2012-08-02 05:16:44.276653
3140	1-3140	23	89	3396	17	5668	0	2	2012-08-02 05:03:36.168443	2012-08-02 05:16:44.696333
1011	1-1011	96	89	4253	17	4122	0	0	2012-08-02 04:47:16.613804	2012-08-02 05:16:44.696333
5104	1-5104	28	17	5243	89	9162	0	0	2012-08-02 05:16:44.696333	2012-08-02 05:16:44.696333
4311	1-4311	11	27	8973	7	7194	0	0	2012-08-02 05:11:25.672231	2012-08-02 05:16:45.491583
5107	1-5107	10	24	295	64	4669	0	0	2012-08-02 05:16:45.889547	2012-08-02 05:16:45.889547
4860	1-4860	42	78	1544	28	1499	0	9	2012-08-02 05:15:07.529251	2012-08-02 05:16:47.082585
4932	1-4932	41	47	1667	70	5847	0	0	2012-08-02 05:15:35.638245	2012-08-02 05:16:47.491806
5086	1-5086	38	98	797	46	3600	0	9	2012-08-02 05:16:37.271284	2012-08-02 05:16:47.491806
4815	1-4815	93	6	1348	90	7204	0	0	2012-08-02 05:14:49.317875	2012-08-02 05:16:47.491806
4849	1-4849	90	82	2636	74	9589	0	5	2012-08-02 05:15:03.173333	2012-08-02 05:16:47.491806
5111	1-5111	45	46	9294	47	6945	0	0	2012-08-02 05:16:47.491806	2012-08-02 05:16:47.491806
4952	1-4952	76	19	4608	26	5423	0	9	2012-08-02 05:15:43.460458	2012-08-02 05:16:47.911406
5065	1-5065	5	1	2218	20	3655	0	0	2012-08-02 05:16:28.653743	2012-08-02 05:16:47.911406
5058	1-5058	72	26	832	90	5766	0	0	2012-08-02 05:16:25.823833	2012-08-02 05:16:47.911406
4546	1-4546	8	22	369	42	355	0	0	2012-08-02 05:13:00.047006	2012-08-02 05:16:49.922629
4949	1-4949	24	39	3419	72	5818	0	0	2012-08-02 05:15:42.267241	2012-08-02 05:16:49.922629
4645	1-4645	75	16	7443	95	9729	0	0	2012-08-02 05:13:40.465888	2012-08-02 05:16:52.75043
5129	1-5129	75	90	454	9	5557	0	3	2012-08-02 05:16:54.750482	2012-08-02 05:16:54.750482
4048	1-4048	23	92	5997	70	5044	0	0	2012-08-02 05:09:40.976113	2012-08-02 05:16:59.226372
4846	1-4846	89	75	759	1	2422	0	0	2012-08-02 05:15:01.935709	2012-08-02 05:16:59.226372
4866	1-4866	19	46	7667	18	1888	0	0	2012-08-02 05:15:09.849129	2012-08-02 05:16:59.226372
5016	1-5016	13	17	3484	19	2192	0	0	2012-08-02 05:16:08.797785	2012-08-02 05:16:59.226372
5141	1-5141	14	22	809	93	809	0	0	2012-08-02 05:16:59.646337	2012-08-02 05:16:59.646337
4719	1-4719	36	90	5211	74	9124	0	6	2012-08-02 05:14:10.399214	2012-08-02 05:17:00.054913
5112	1-5112	6	20	3829	30	4685	0	0	2012-08-02 05:16:47.911406	2012-08-02 05:17:00.054913
4821	1-4821	43	79	2463	91	1519	0	5	2012-08-02 05:14:51.726394	2012-08-02 05:17:00.883969
3060	1-3060	86	50	4685	47	758	0	0	2012-08-02 05:03:03.559598	2012-08-02 05:17:00.883969
4870	1-4870	65	94	1947	49	2183	0	0	2012-08-02 05:15:11.406971	2012-08-02 05:17:00.883969
4762	1-4762	8	94	6239	89	5526	0	0	2012-08-02 05:14:27.825994	2012-08-02 05:17:00.883969
5144	1-5144	96	98	1071	68	9401	0	0	2012-08-02 05:17:00.883969	2012-08-02 05:17:00.883969
5146	1-5146	12	52	4096	22	5549	0	0	2012-08-02 05:17:01.73457	2012-08-02 05:17:01.73457
5153	1-5153	68	22	1191	6	8627	0	0	2012-08-02 05:17:04.573768	2012-08-02 05:17:04.573768
5154	1-5154	84	80	1222	52	6442	0	0	2012-08-02 05:17:04.982905	2012-08-02 05:17:04.982905
4328	1-4328	15	68	5458	11	9072	0	0	2012-08-02 05:11:32.623943	2012-08-02 05:17:05.779432
4419	1-4419	79	74	7379	46	9162	0	2	2012-08-02 05:12:09.737894	2012-08-02 05:17:06.961744
3941	1-3941	56	74	2740	46	3337	0	0	2012-08-02 05:08:58.104655	2012-08-02 05:17:06.961744
5159	1-5159	90	46	3279	28	7468	0	11	2012-08-02 05:17:06.961744	2012-08-02 05:17:06.961744
4190	1-4190	91	30	6497	10	3379	0	0	2012-08-02 05:10:37.35127	2012-08-02 05:17:07.746004
5161	1-5161	14	4	2488	30	8580	0	10	2012-08-02 05:17:07.746004	2012-08-02 05:17:07.746004
5092	1-5092	26	80	5425	71	9387	0	0	2012-08-02 05:16:39.70207	2012-08-02 05:17:08.541531
5163	1-5163	73	40	98	80	7527	0	0	2012-08-02 05:17:08.541531	2012-08-02 05:17:08.541531
4353	1-4353	82	71	2599	67	897	0	0	2012-08-02 05:11:42.909752	2012-08-02 05:17:09.712619
4171	1-4171	77	71	7774	19	3929	0	0	2012-08-02 05:10:29.893876	2012-08-02 05:17:09.712619
5166	1-5166	59	67	129	71	8097	0	10	2012-08-02 05:17:09.712619	2012-08-02 05:17:09.712619
3447	1-3447	6	61	8865	90	5935	0	0	2012-08-02 05:05:40.270866	2012-08-02 05:17:10.09962
3547	1-3547	48	40	8952	18	785	0	0	2012-08-02 05:06:20.611713	2012-08-02 05:17:10.497344
2015	1-2015	76	99	8840	32	4361	0	0	2012-08-02 04:55:37.312362	2012-08-02 05:17:10.89544
5093	1-5093	73	33	810	99	1430	0	5	2012-08-02 05:16:40.121994	2012-08-02 05:17:10.89544
2837	1-2837	53	99	3729	31	7080	0	2	2012-08-02 05:01:31.904411	2012-08-02 05:17:12.07713
4138	1-4138	52	79	5411	72	3083	0	0	2012-08-02 05:10:16.779018	2012-08-02 05:17:12.07713
4639	1-4639	35	56	8632	9	9007	0	0	2012-08-02 05:13:37.990737	2012-08-02 05:17:12.07713
5052	1-5052	39	56	2084	22	4595	0	8	2012-08-02 05:16:23.404146	2012-08-02 05:17:12.07713
2810	1-2810	33	56	6422	90	3607	0	0	2012-08-02 05:01:20.579301	2012-08-02 05:17:12.07713
4857	1-4857	90	72	1256	31	4614	0	0	2012-08-02 05:15:06.39146	2012-08-02 05:17:12.07713
5087	1-5087	32	77	1146	33	3802	0	0	2012-08-02 05:16:37.68008	2012-08-02 05:17:12.85116
4894	1-4894	87	77	7563	60	9164	0	0	2012-08-02 05:15:20.820567	2012-08-02 05:17:12.85116
5174	1-5174	8	43	230	77	4323	0	0	2012-08-02 05:17:12.85116	2012-08-02 05:17:12.85116
1652	1-1652	61	52	7288	71	3122	0	0	2012-08-02 04:52:54.585774	2012-08-02 05:17:13.635396
2786	1-2786	74	91	9600	68	9681	0	0	2012-08-02 05:01:10.481559	2012-08-02 05:17:13.635396
4743	1-4743	50	47	5619	91	9385	0	0	2012-08-02 05:14:20.146163	2012-08-02 05:17:13.635396
4164	1-4164	13	54	9689	71	3828	0	9	2012-08-02 05:10:27.131443	2012-08-02 05:17:13.635396
5176	1-5176	51	71	1488	47	7504	0	7	2012-08-02 05:17:13.635396	2012-08-02 05:17:13.635396
4961	1-4961	97	91	5716	98	7926	301	0	2012-08-02 05:15:46.996014	2012-08-02 05:17:13.635396
3952	1-3952	93	91	8517	98	128	128	0	2012-08-02 05:09:02.526169	2012-08-02 05:17:13.635396
3211	1-3211	14	75	954	2	4474	0	3	2012-08-02 05:04:04.997011	2012-08-02 05:17:14.806893
4933	1-4933	94	97	3288	1	5218	0	4	2012-08-02 05:15:36.024888	2012-08-02 05:17:14.806893
5140	1-5140	18	18	3157	75	9930	0	0	2012-08-02 05:16:59.226372	2012-08-02 05:17:14.806893
1415	1-1415	39	75	2936	2	5714	4389	0	2012-08-02 04:50:52.887572	2012-08-02 05:17:14.806893
4920	1-4920	7	6	6628	68	7456	0	0	2012-08-02 05:15:30.951802	2012-08-02 05:17:16.386824
5183	1-5183	38	37	3883	6	9753	0	0	2012-08-02 05:17:16.386824	2012-08-02 05:17:16.386824
3274	1-3274	23	18	4016	58	7249	0	7	2012-08-02 05:04:30.126145	2012-08-02 05:17:19.093627
3459	1-3459	81	25	7875	16	8604	0	0	2012-08-02 05:05:45.15392	2012-08-02 05:17:19.093627
5190	1-5190	44	49	1111	18	5213	0	2	2012-08-02 05:17:19.093627	2012-08-02 05:17:19.093627
5046	1-5046	78	82	3654	18	9026	0	3	2012-08-02 05:16:20.984519	2012-08-02 05:17:20.288562
5193	1-5193	54	44	2796	55	5365	0	0	2012-08-02 05:17:20.288562	2012-08-02 05:17:20.288562
4692	1-4692	19	27	5054	41	6024	0	0	2012-08-02 05:13:59.599451	2012-08-02 05:17:20.675574
2918	1-2918	81	27	9991	93	9399	0	0	2012-08-02 05:02:05.68197	2012-08-02 05:17:20.675574
5194	1-5194	73	81	870	27	7650	0	0	2012-08-02 05:17:20.675574	2012-08-02 05:17:20.675574
4220	1-4220	27	19	7252	16	9101	0	0	2012-08-02 05:10:49.120205	2012-08-02 05:17:21.08454
5195	1-5195	4	21	3125	19	4778	0	4	2012-08-02 05:17:21.08454	2012-08-02 05:17:21.08454
4494	1-4494	82	72	9162	21	8973	1270	7	2012-08-02 05:12:39.338888	2012-08-02 05:17:21.08454
4942	1-4942	85	28	6720	74	5457	0	0	2012-08-02 05:15:39.538282	2012-08-02 05:17:21.869166
5197	1-5197	89	59	281	85	3104	0	0	2012-08-02 05:17:21.869166	2012-08-02 05:17:21.869166
1854	1-1854	17	75	8745	8	5188	0	3	2012-08-02 04:54:26.203168	2012-08-02 05:17:22.266787
1927	1-1927	12	68	8611	31	9226	0	0	2012-08-02 04:54:59.003442	2012-08-02 05:17:22.266787
4898	1-4898	100	96	8689	75	9657	0	5	2012-08-02 05:15:22.389101	2012-08-02 05:17:22.266787
5198	1-5198	55	31	1885	96	6180	0	0	2012-08-02 05:17:22.266787	2012-08-02 05:17:22.266787
289	1-289	94	96	3933	89	685	265	0	2012-08-02 04:43:58.666804	2012-08-02 05:17:22.266787
4342	1-4342	61	36	6544	85	8158	0	6	2012-08-02 05:11:38.402429	2012-08-02 05:17:22.664804
5199	1-5199	39	85	1336	36	9730	0	0	2012-08-02 05:17:22.664804	2012-08-02 05:17:22.664804
5201	1-5201	4	72	569	35	2795	0	7	2012-08-02 05:17:23.449374	2012-08-02 05:17:23.449374
889	1-889	11	44	6660	64	1886	0	0	2012-08-02 04:46:09.980272	2012-08-02 05:17:23.857879
5172	1-5172	18	31	6743	56	9534	0	0	2012-08-02 05:17:12.07713	2012-08-02 05:17:24.23376
4884	1-4884	83	58	9351	95	6955	0	0	2012-08-02 05:15:16.942679	2012-08-02 05:17:24.23376
5051	1-5051	22	46	5144	58	7831	0	0	2012-08-02 05:16:23.006624	2012-08-02 05:17:24.23376
5068	1-5068	45	100	5704	52	7542	0	0	2012-08-02 05:16:29.880313	2012-08-02 05:17:24.620499
5204	1-5204	74	87	2322	100	7660	0	0	2012-08-02 05:17:24.620499	2012-08-02 05:17:24.620499
5205	1-5205	10	14	947	96	4462	0	0	2012-08-02 05:17:24.996472	2012-08-02 05:17:24.996472
5167	1-5167	51	90	4738	99	6642	0	0	2012-08-02 05:17:10.09962	2012-08-02 05:17:25.780973
5098	1-5098	31	99	7038	97	8277	0	0	2012-08-02 05:16:42.232304	2012-08-02 05:17:25.780973
4980	1-4980	92	78	1197	2	3969	0	9	2012-08-02 05:15:54.420207	2012-08-02 05:17:26.200539
4822	1-4822	48	8	5783	24	8378	0	3	2012-08-02 05:14:52.124	2012-08-02 05:17:26.200539
3195	1-3195	54	44	3176	51	4718	0	0	2012-08-02 05:03:58.438453	2012-08-02 05:17:27.36063
5032	1-5032	67	10	7856	4	7405	0	5	2012-08-02 05:16:15.338991	2012-08-02 05:17:28.145034
3877	1-3877	24	39	7183	42	6017	0	10	2012-08-02 05:08:32.495669	2012-08-02 05:17:28.907328
4617	1-4617	63	39	6689	89	3437	0	0	2012-08-02 05:13:28.941795	2012-08-02 05:17:28.907328
4197	1-4197	75	47	4141	82	3284	0	0	2012-08-02 05:10:40.082682	2012-08-02 05:17:28.907328
5215	1-5215	93	72	646	39	9997	0	2	2012-08-02 05:17:28.907328	2012-08-02 05:17:28.907328
3893	1-3893	71	42	9063	72	4167	1763	4	2012-08-02 05:08:38.950051	2012-08-02 05:17:28.907328
5218	1-5218	11	27	1503	45	4457	0	0	2012-08-02 05:17:30.089492	2012-08-02 05:17:30.089492
4360	1-4360	87	68	9362	94	6194	0	9	2012-08-02 05:11:45.782322	2012-08-02 05:17:31.647284
4841	1-4841	71	16	7892	38	4545	0	0	2012-08-02 05:14:59.880643	2012-08-02 05:17:31.647284
3900	1-3900	37	60	6490	16	7060	0	0	2012-08-02 05:08:41.767499	2012-08-02 05:17:31.647284
5070	1-5070	27	68	5422	5	6850	0	0	2012-08-02 05:16:30.697647	2012-08-02 05:17:31.647284
3288	1-3288	40	37	852	69	70	0	0	2012-08-02 05:04:35.782761	2012-08-02 05:17:31.647284
5222	1-5222	88	94	353	60	8478	0	0	2012-08-02 05:17:31.647284	2012-08-02 05:17:31.647284
4330	1-4330	65	5	3467	57	9097	0	0	2012-08-02 05:11:33.418942	2012-08-02 05:17:32.045506
4281	1-4281	90	65	4575	84	6757	0	0	2012-08-02 05:11:13.264248	2012-08-02 05:17:34.796151
4790	1-4790	27	71	6611	65	8271	0	0	2012-08-02 05:14:39.062364	2012-08-02 05:17:34.796151
2217	1-2217	72	27	8249	86	1293	0	0	2012-08-02 04:57:04.853112	2012-08-02 05:17:34.796151
5202	1-5202	54	64	1898	88	9965	0	0	2012-08-02 05:17:23.857879	2012-08-02 05:17:34.796151
1426	1-1426	59	77	8015	10	5193	0	0	2012-08-02 04:50:58.886427	2012-08-02 05:17:37.215775
5025	1-5025	35	84	1752	2	1004	0	0	2012-08-02 05:16:12.510407	2012-08-02 05:17:38.773627
2459	1-2459	50	82	7426	10	4681	0	0	2012-08-02 04:58:50.591066	2012-08-02 05:17:43.468646
5252	1-5252	71	67	1307	82	5826	0	4	2012-08-02 05:17:43.468646	2012-08-02 05:17:43.468646
5128	1-5128	35	38	1626	62	3640	0	0	2012-08-02 05:16:54.341332	2012-08-02 05:17:44.25317
4632	1-4632	68	43	1140	57	2790	0	0	2012-08-02 05:13:35.129199	2012-08-02 05:17:44.650985
4986	1-4986	80	71	3024	43	3619	0	3	2012-08-02 05:15:56.850843	2012-08-02 05:17:44.650985
3599	1-3599	67	92	7721	21	4858	0	0	2012-08-02 05:06:41.717373	2012-08-02 05:17:48.585702
4657	1-4657	47	34	4946	31	5360	0	0	2012-08-02 05:13:45.415398	2012-08-02 05:17:48.585702
2655	1-2655	12	14	9876	43	2125	0	0	2012-08-02 05:00:15.154221	2012-08-02 05:17:50.950278
3906	1-3906	94	68	9835	20	9255	0	0	2012-08-02 05:08:44.165059	2012-08-02 05:17:50.950278
5271	1-5271	85	36	1936	68	5706	0	0	2012-08-02 05:17:50.950278	2012-08-02 05:17:50.950278
5110	1-5110	56	81	2729	65	2611	0	0	2012-08-02 05:16:47.082585	2012-08-02 05:17:54.032476
5113	1-5113	45	79	6725	70	8403	0	5	2012-08-02 05:16:48.331548	2012-08-02 05:17:54.806106
5281	1-5281	97	99	1507	79	8785	0	8	2012-08-02 05:17:54.806106	2012-08-02 05:17:54.806106
4997	1-4997	1	74	7942	83	6604	0	0	2012-08-02 05:16:01.225997	2012-08-02 05:17:55.203696
5282	1-5282	19	83	4212	74	8672	0	3	2012-08-02 05:17:55.203696	2012-08-02 05:17:55.203696
5009	1-5009	52	100	8610	60	5576	0	5	2012-08-02 05:16:05.957841	2012-08-02 05:17:56.772515
4629	1-4629	8	95	9796	9	7401	398	0	2012-08-02 05:13:33.869408	2012-08-02 05:17:56.772515
5287	1-5287	27	63	1073	74	4145	0	0	2012-08-02 05:17:57.170312	2012-08-02 05:17:57.170312
5168	1-5168	97	20	3678	40	9413	0	0	2012-08-02 05:17:10.497344	2012-08-02 05:17:57.568102
5221	1-5221	75	62	4207	89	6458	0	3	2012-08-02 05:17:31.271715	2012-08-02 05:17:57.568102
4944	1-4944	30	43	7525	62	8065	0	6	2012-08-02 05:15:40.32266	2012-08-02 05:17:57.568102
3345	1-3345	55	68	2336	93	1700	0	0	2012-08-02 05:04:58.821218	2012-08-02 05:17:57.965978
3192	1-3192	83	9	6167	93	6037	0	0	2012-08-02 05:03:57.221484	2012-08-02 05:17:57.965978
5289	1-5289	81	93	3176	1	6974	0	5	2012-08-02 05:17:57.965978	2012-08-02 05:17:57.965978
5291	1-5291	16	77	3176	68	5378	0	0	2012-08-02 05:17:58.761532	2012-08-02 05:17:58.761532
4542	1-4542	81	16	4034	7	3520	0	0	2012-08-02 05:12:58.4557	2012-08-02 05:18:00.330551
5297	1-5297	34	90	2073	13	6095	0	0	2012-08-02 05:18:01.11501	2012-08-02 05:18:01.11501
4930	1-4930	88	5	6814	79	7743	0	0	2012-08-02 05:15:34.853578	2012-08-02 05:18:02.263588
4688	1-4688	94	2	9218	1	5657	0	14	2012-08-02 05:13:58.00761	2012-08-02 05:18:02.263588
4912	1-4912	4	7	2737	53	4723	0	5	2012-08-02 05:15:27.802441	2012-08-02 05:18:02.661599
5002	1-5002	83	11	5725	50	9949	0	0	2012-08-02 05:16:03.194659	2012-08-02 05:18:03.435294
5114	1-5114	10	80	6100	55	1562	0	0	2012-08-02 05:16:48.718099	2012-08-02 05:18:05.413565
3770	1-3770	31	55	9958	32	7264	0	0	2012-08-02 05:07:49.889768	2012-08-02 05:18:06.187246
5310	1-5310	58	61	782	55	8546	0	13	2012-08-02 05:18:06.187246	2012-08-02 05:18:06.187246
4925	1-4925	97	7	9370	23	7877	0	2	2012-08-02 05:15:32.898334	2012-08-02 05:18:06.58488
4792	1-4792	6	42	9894	39	4957	0	0	2012-08-02 05:14:39.868954	2012-08-02 05:18:06.971459
5312	1-5312	28	92	3613	42	5075	0	6	2012-08-02 05:18:06.971459	2012-08-02 05:18:06.971459
4444	1-4444	71	98	8204	74	8108	0	0	2012-08-02 05:12:19.582216	2012-08-02 05:18:07.756143
2783	1-2783	58	98	6423	96	2270	0	0	2012-08-02 05:01:09.231342	2012-08-02 05:18:07.756143
5211	1-5211	99	49	1299	44	3577	0	0	2012-08-02 05:17:27.36063	2012-08-02 05:18:07.756143
5314	1-5314	64	44	2465	98	8006	0	0	2012-08-02 05:18:07.756143	2012-08-02 05:18:07.756143
4809	1-4809	78	89	4826	97	5570	0	0	2012-08-02 05:14:46.853855	2012-08-02 05:18:08.16499
3048	1-3048	71	10	6072	83	4441	0	0	2012-08-02 05:02:58.696352	2012-08-02 05:18:08.16499
5236	1-5236	12	10	3445	89	5530	0	0	2012-08-02 05:17:37.215775	2012-08-02 05:18:08.16499
4985	1-4985	33	46	5429	31	7724	0	0	2012-08-02 05:15:56.464244	2012-08-02 05:18:08.93837
4586	1-4586	22	90	3610	88	4544	0	4	2012-08-02 05:13:16.190163	2012-08-02 05:18:10.507181
5142	1-5142	16	83	5442	20	9749	0	0	2012-08-02 05:17:00.054913	2012-08-02 05:18:10.905002
5323	1-5323	12	36	921	10	2194	0	2	2012-08-02 05:18:11.291708	2012-08-02 05:18:11.291708
5326	1-5326	42	60	613	98	1027	0	6	2012-08-02 05:18:12.463057	2012-08-02 05:18:12.463057
5208	1-5208	29	1	3680	8	4068	0	0	2012-08-02 05:17:26.200539	2012-08-02 05:18:12.849471
5327	1-5327	80	88	504	60	620	0	3	2012-08-02 05:18:12.849471	2012-08-02 05:18:12.849471
5295	1-5295	28	7	3197	91	4572	0	10	2012-08-02 05:18:00.330551	2012-08-02 05:18:13.238202
1051	1-1051	59	8	8658	12	7823	0	0	2012-08-02 04:47:40.213095	2012-08-02 05:18:13.238202
2862	1-2862	24	8	8781	68	7993	0	4	2012-08-02 05:01:42.413944	2012-08-02 05:18:13.238202
3557	1-3557	97	8	5702	7	7471	0	11	2012-08-02 05:06:24.621811	2012-08-02 05:18:13.238202
5328	1-5328	37	48	2536	8	8088	0	0	2012-08-02 05:18:13.238202	2012-08-02 05:18:14.023034
5330	1-5330	57	25	1506	48	3396	0	0	2012-08-02 05:18:14.023034	2012-08-02 05:18:14.023034
4708	1-4708	62	54	4846	30	2682	0	0	2012-08-02 05:14:05.85144	2012-08-02 05:18:14.431958
5213	1-5213	56	4	3817	77	8083	0	0	2012-08-02 05:17:28.145034	2012-08-02 05:18:14.829328
5055	1-5055	69	32	2833	61	8057	0	3	2012-08-02 05:16:24.608421	2012-08-02 05:18:14.829328
5332	1-5332	86	6	552	4	7195	0	2	2012-08-02 05:18:14.829328	2012-08-02 05:18:14.829328
4394	1-4394	7	81	3524	11	5172	0	0	2012-08-02 05:11:59.702897	2012-08-02 05:18:16.81839
5102	1-5102	60	32	3046	19	6857	0	0	2012-08-02 05:16:43.845218	2012-08-02 05:18:16.81839
5337	1-5337	23	11	1822	32	4176	0	2	2012-08-02 05:18:16.81839	2012-08-02 05:18:16.81839
4583	1-4583	83	33	5601	63	5752	0	0	2012-08-02 05:13:14.963942	2012-08-02 05:18:17.227055
4786	1-4786	22	50	9506	60	6228	0	0	2012-08-02 05:14:37.449131	2012-08-02 05:18:18.022422
5315	1-5315	58	20	2740	10	9796	0	0	2012-08-02 05:18:08.16499	2012-08-02 05:18:18.42019
5341	1-5341	100	100	1345	20	2492	0	2	2012-08-02 05:18:18.42019	2012-08-02 05:18:18.42019
4646	1-4646	61	54	4531	78	5819	0	7	2012-08-02 05:13:40.874746	2012-08-02 05:18:19.580327
4529	1-4529	8	8	1468	31	3836	0	0	2012-08-02 05:12:53.251649	2012-08-02 05:18:19.967122
2172	1-2172	2	62	7219	42	9348	0	0	2012-08-02 04:56:45.218774	2012-08-02 05:18:20.762816
4496	1-4496	6	95	7377	33	4044	0	3	2012-08-02 05:12:40.134686	2012-08-02 05:18:20.762816
4185	1-4185	64	42	6165	28	3064	0	6	2012-08-02 05:10:35.373861	2012-08-02 05:18:20.762816
5081	1-5081	69	75	1236	92	3993	0	0	2012-08-02 05:16:35.183293	2012-08-02 05:18:23.502386
5300	1-5300	15	75	1386	5	6632	0	0	2012-08-02 05:18:02.263588	2012-08-02 05:18:23.502386
5117	1-5117	67	21	2603	34	7429	0	0	2012-08-02 05:16:49.922629	2012-08-02 05:18:23.502386
5356	1-5356	99	76	4929	18	4846	0	11	2012-08-02 05:18:24.287011	2012-08-02 05:18:24.287011
5155	1-5155	77	95	2405	17	1268	0	6	2012-08-02 05:17:05.403581	2012-08-02 05:18:26.629311
2609	1-2609	55	4	8183	90	8930	0	0	2012-08-02 04:59:55.486349	2012-08-02 05:18:28.18679
3916	1-3916	1	25	6529	83	9261	0	0	2012-08-02 05:08:48.131351	2012-08-02 05:18:29.380288
5369	1-5369	51	83	3035	25	3384	0	0	2012-08-02 05:18:29.380288	2012-08-02 05:18:29.380288
5280	1-5280	56	14	3460	85	2329	0	0	2012-08-02 05:17:54.419192	2012-08-02 05:18:29.766702
5010	1-5010	53	76	5841	7	4140	0	0	2012-08-02 05:16:06.367213	2012-08-02 05:18:29.766702
5230	1-5230	87	88	5475	71	9868	0	0	2012-08-02 05:17:34.796151	2012-08-02 05:18:32.097692
1399	1-1399	32	67	8068	32	7066	0	0	2012-08-02 04:50:44.33696	2012-08-02 05:18:32.097692
5301	1-5301	28	60	6404	1	9222	0	8	2012-08-02 05:18:02.661599	2012-08-02 05:18:32.097692
5376	1-5376	41	27	352	67	4503	0	14	2012-08-02 05:18:32.097692	2012-08-02 05:18:32.097692
5378	1-5378	67	71	1743	54	3749	0	0	2012-08-02 05:18:32.893564	2012-08-02 05:18:32.893564
3339	1-3339	26	4	4584	86	3274	0	0	2012-08-02 05:04:56.434795	2012-08-02 05:18:33.28023
5379	1-5379	2	53	949	4	2158	0	0	2012-08-02 05:18:33.28023	2012-08-02 05:18:33.28023
5265	1-5265	50	31	7489	92	7916	0	0	2012-08-02 05:17:48.585702	2012-08-02 05:18:34.451225
5031	1-5031	46	28	4978	92	7222	0	8	2012-08-02 05:16:14.941123	2012-08-02 05:18:34.451225
5342	1-5342	17	31	3177	36	3739	0	5	2012-08-02 05:18:18.818108	2012-08-02 05:18:34.451225
4736	1-4736	30	5	7600	87	8099	0	0	2012-08-02 05:14:17.361485	2012-08-02 05:18:34.451225
4903	1-4903	64	86	3362	49	9252	0	0	2012-08-02 05:15:24.344987	2012-08-02 05:18:35.622476
4480	1-4480	11	35	5588	86	8812	0	0	2012-08-02 05:12:33.82568	2012-08-02 05:18:35.622476
5279	1-5279	63	48	2094	89	3633	0	0	2012-08-02 05:17:54.032476	2012-08-02 05:18:36.049046
5284	1-5284	8	25	998	91	1325	0	0	2012-08-02 05:17:56.010574	2012-08-02 05:18:36.440203
3781	1-3781	49	65	9619	1	7686	0	0	2012-08-02 05:07:54.210087	2012-08-02 05:18:37.622421
4642	1-4642	74	17	7703	11	6934	0	7	2012-08-02 05:13:39.228176	2012-08-02 05:18:39.179944
5317	1-5317	76	92	1746	77	1755	0	0	2012-08-02 05:18:08.93837	2012-08-02 05:18:39.179944
5394	1-5394	56	84	5983	38	8619	0	0	2012-08-02 05:18:39.179944	2012-08-02 05:18:39.179944
5308	1-5308	72	17	1034	80	8255	0	0	2012-08-02 05:18:05.413565	2012-08-02 05:18:39.577832
5395	1-5395	4	80	4880	45	6554	0	6	2012-08-02 05:18:39.577832	2012-08-02 05:18:39.577832
4817	1-4817	88	71	7413	27	9789	0	0	2012-08-02 05:14:50.135498	2012-08-02 05:18:40.760271
5331	1-5331	19	85	3848	54	6739	0	4	2012-08-02 05:18:14.431958	2012-08-02 05:18:41.955176
4781	1-4781	95	85	5918	28	3903	0	0	2012-08-02 05:14:35.427347	2012-08-02 05:18:41.955176
5349	1-5349	41	8	1522	85	5347	0	0	2012-08-02 05:18:21.558354	2012-08-02 05:18:41.955176
5169	1-5169	14	30	5412	33	6664	0	0	2012-08-02 05:17:10.89544	2012-08-02 05:18:44.275269
3193	1-3193	10	49	6622	88	9991	0	0	2012-08-02 05:03:57.619155	2012-08-02 05:18:45.059638
5409	1-5409	41	88	2640	91	4125	0	3	2012-08-02 05:18:45.059638	2012-08-02 05:18:45.059638
4936	1-4936	53	12	4281	55	8779	0	4	2012-08-02 05:15:37.206919	2012-08-02 05:18:45.844248
5347	1-5347	19	28	3258	95	9355	0	2	2012-08-02 05:18:20.762816	2012-08-02 05:18:45.844248
5414	1-5414	97	18	926	82	9543	0	10	2012-08-02 05:18:47.026346	2012-08-02 05:18:47.026346
4696	1-4696	63	84	5634	64	5584	0	0	2012-08-02 05:14:01.135186	2012-08-02 05:18:47.424294
5224	1-5224	85	63	1560	94	1319	0	0	2012-08-02 05:17:32.432018	2012-08-02 05:18:47.810841
5255	1-5255	41	11	6384	71	8413	0	0	2012-08-02 05:17:44.650985	2012-08-02 05:18:48.197908
4893	1-4893	26	63	4814	91	5996	0	0	2012-08-02 05:15:20.411447	2012-08-02 05:18:48.197908
4413	1-4413	75	30	4335	99	2340	0	0	2012-08-02 05:12:07.406313	2012-08-02 05:18:48.960157
5419	1-5419	29	85	2088	30	6754	0	0	2012-08-02 05:18:48.960157	2012-08-02 05:18:48.960157
5182	1-5182	39	29	2858	52	7718	0	0	2012-08-02 05:17:16.000399	2012-08-02 05:18:49.772118
5343	1-5343	56	44	3796	20	7654	0	11	2012-08-02 05:18:19.193729	2012-08-02 05:18:50.946782
4843	1-4843	79	90	7163	20	9438	0	4	2012-08-02 05:15:00.720726	2012-08-02 05:18:50.946782
5424	1-5424	53	20	8413	96	7444	0	7	2012-08-02 05:18:50.946782	2012-08-02 05:18:50.946782
5237	1-5237	56	77	4533	84	7373	0	0	2012-08-02 05:17:37.613677	2012-08-02 05:18:51.357374
4028	1-4028	82	67	9286	36	9409	0	0	2012-08-02 05:09:33.009683	2012-08-02 05:18:51.357374
5407	1-5407	39	31	6080	30	3374	0	2	2012-08-02 05:18:44.275269	2012-08-02 05:18:52.132259
1924	1-1924	89	61	9520	55	3668	0	8	2012-08-02 04:54:57.600613	2012-08-02 05:18:52.132259
4440	1-4440	97	52	8496	92	6845	0	0	2012-08-02 05:12:18.035453	2012-08-02 05:18:52.531287
5028	1-5028	56	76	6287	92	3050	0	0	2012-08-02 05:16:13.736439	2012-08-02 05:18:52.531287
5428	1-5428	89	91	4345	47	9350	0	3	2012-08-02 05:18:52.531287	2012-08-02 05:18:52.531287
4820	1-4820	89	96	4124	88	9954	0	3	2012-08-02 05:14:51.328674	2012-08-02 05:18:53.339494
5417	1-5417	46	71	2555	63	8709	0	0	2012-08-02 05:18:48.197908	2012-08-02 05:18:53.339494
4599	1-4599	62	59	8269	39	9491	0	9	2012-08-02 05:13:21.493622	2012-08-02 05:18:53.72762
5431	1-5431	94	47	1152	59	7084	0	0	2012-08-02 05:18:53.72762	2012-08-02 05:18:53.72762
1574	1-1574	66	73	4807	24	983	0	0	2012-08-02 04:52:18.587002	2012-08-02 05:18:54.13765
2684	1-2684	60	85	9930	7	6564	0	0	2012-08-02 05:00:27.464276	2012-08-02 05:18:54.549444
5433	1-5433	83	64	889	85	6143	0	2	2012-08-02 05:18:54.549444	2012-08-02 05:18:54.549444
2537	1-2537	14	47	3496	28	3397	0	0	2012-08-02 04:59:24.55777	2012-08-02 05:18:54.947725
5106	1-5106	49	12	1749	27	7626	0	6	2012-08-02 05:16:45.491583	2012-08-02 05:18:55.357779
5435	1-5435	93	34	3724	12	5802	0	0	2012-08-02 05:18:55.357779	2012-08-02 05:18:55.357779
5192	1-5192	44	48	3931	22	9847	0	0	2012-08-02 05:17:19.902121	2012-08-02 05:18:55.756153
5286	1-5286	53	9	1198	100	6604	0	0	2012-08-02 05:17:56.772515	2012-08-02 05:18:55.756153
5416	1-5416	62	45	3587	48	9938	0	7	2012-08-02 05:18:47.810841	2012-08-02 05:18:55.756153
4858	1-4858	69	50	2796	100	1256	0	0	2012-08-02 05:15:06.778041	2012-08-02 05:18:55.756153
3813	1-3813	34	57	3159	11	2139	0	0	2012-08-02 05:08:06.808113	2012-08-02 05:18:57.292673
5442	1-5442	15	43	2028	89	7744	0	5	2012-08-02 05:18:58.077291	2012-08-02 05:18:58.077291
4366	1-4366	11	67	2579	40	7008	0	8	2012-08-02 05:11:48.21263	2012-08-02 05:18:58.475296
5288	1-5288	87	40	3968	43	7813	0	0	2012-08-02 05:17:57.568102	2012-08-02 05:18:58.475296
5382	1-5382	71	87	5335	38	5954	0	14	2012-08-02 05:18:34.451225	2012-08-02 05:18:59.668851
5367	1-5367	53	92	6750	91	7453	0	6	2012-08-02 05:18:28.584563	2012-08-02 05:18:59.668851
2592	1-2592	57	65	8093	25	5049	0	0	2012-08-02 04:59:48.259096	2012-08-02 05:18:59.668851
5227	1-5227	29	67	3829	11	4916	0	0	2012-08-02 05:17:33.614555	2012-08-02 05:19:00.07895
3475	1-3475	16	77	4689	17	2374	0	0	2012-08-02 05:05:51.572858	2012-08-02 05:19:00.07895
5447	1-5447	68	100	2005	77	8578	0	11	2012-08-02 05:19:00.07895	2012-08-02 05:19:00.07895
5412	1-5412	75	14	7509	80	2417	0	2	2012-08-02 05:18:46.264188	2012-08-02 05:19:01.249943
4907	1-4907	56	29	7650	82	8075	0	2	2012-08-02 05:15:25.880262	2012-08-02 05:19:01.637224
5262	1-5262	7	11	7287	61	8809	0	0	2012-08-02 05:17:47.415001	2012-08-02 05:19:02.432852
4837	1-4837	10	65	7045	78	8430	0	0	2012-08-02 05:14:58.234081	2012-08-02 05:19:02.432852
5038	1-5038	21	10	5851	41	6295	0	4	2012-08-02 05:16:17.747193	2012-08-02 05:19:02.830427
4549	1-4549	82	62	5482	11	7052	0	0	2012-08-02 05:13:01.251524	2012-08-02 05:19:03.217248
4635	1-4635	28	11	5456	84	4322	0	11	2012-08-02 05:13:36.366656	2012-08-02 05:19:03.217248
3797	1-3797	31	45	5993	95	5027	0	10	2012-08-02 05:08:00.452081	2012-08-02 05:19:03.615025
3234	1-3234	40	94	4439	88	5771	0	0	2012-08-02 05:04:14.391637	2012-08-02 05:19:04.012918
4758	1-4758	73	18	5253	48	6786	0	0	2012-08-02 05:14:26.21252	2012-08-02 05:19:04.012918
2690	1-2690	50	63	7606	80	6879	0	0	2012-08-02 05:00:29.94955	2012-08-02 05:19:05.200332
5035	1-5035	3	63	2322	96	1692	0	0	2012-08-02 05:16:16.554095	2012-08-02 05:19:05.200332
4926	1-4926	65	63	948	39	823	0	0	2012-08-02 05:15:33.296246	2012-08-02 05:19:05.200332
3368	1-3368	62	29	6196	59	6108	0	0	2012-08-02 05:05:08.169862	2012-08-02 05:19:06.39582
5463	1-5463	43	84	564	29	2745	0	0	2012-08-02 05:19:06.39582	2012-08-02 05:19:06.39582
4701	1-4701	23	29	1169	41	1423	0	3	2012-08-02 05:14:03.057436	2012-08-02 05:19:09.876135
4922	1-4922	29	94	6444	24	3407	0	4	2012-08-02 05:15:31.724822	2012-08-02 05:19:09.876135
4427	1-4427	7	47	7262	6	7434	0	11	2012-08-02 05:12:12.920287	2012-08-02 05:19:09.876135
4117	1-4117	4	3	9191	35	6272	0	3	2012-08-02 05:10:08.457467	2012-08-02 05:19:09.876135
5472	1-5472	84	32	1721	3	8401	0	0	2012-08-02 05:19:09.876135	2012-08-02 05:19:09.876135
4434	1-4434	78	3	3719	29	3553	0	0	2012-08-02 05:12:15.693157	2012-08-02 05:19:10.285444
3217	1-3217	36	61	6475	7	4382	0	5	2012-08-02 05:04:07.442345	2012-08-02 05:19:10.694063
4401	1-4401	31	31	7869	97	2349	0	0	2012-08-02 05:12:02.588164	2012-08-02 05:19:10.694063
4262	1-4262	14	31	3006	83	1551	0	10	2012-08-02 05:11:05.508138	2012-08-02 05:19:10.694063
4962	1-4962	51	34	3537	22	7077	0	11	2012-08-02 05:15:47.39341	2012-08-02 05:19:11.48989
5411	1-5411	66	23	4612	30	8023	0	7	2012-08-02 05:18:45.844248	2012-08-02 05:19:11.48989
4085	1-4085	13	5	4767	51	4096	0	5	2012-08-02 05:09:55.683965	2012-08-02 05:19:12.285434
4447	1-4447	28	35	4306	79	5679	0	0	2012-08-02 05:12:20.797927	2012-08-02 05:19:13.058903
5480	1-5480	13	51	2102	35	8357	0	0	2012-08-02 05:19:13.058903	2012-08-02 05:19:13.058903
4396	1-4396	29	85	6909	39	4379	0	11	2012-08-02 05:12:00.542718	2012-08-02 05:19:14.671889
5484	1-5484	71	14	2820	85	4993	0	3	2012-08-02 05:19:14.671889	2012-08-02 05:19:14.671889
5311	1-5311	5	23	3526	25	7705	0	0	2012-08-02 05:18:06.58488	2012-08-02 05:19:15.467543
3929	1-3929	42	55	9830	23	8346	907	0	2012-08-02 05:08:53.313298	2012-08-02 05:19:15.467543
5454	1-5454	58	26	2485	21	7074	0	0	2012-08-02 05:19:02.830427	2012-08-02 05:19:15.865354
5247	1-5247	55	89	7645	81	9608	0	0	2012-08-02 05:17:41.513663	2012-08-02 05:19:15.865354
5372	1-5372	81	8	2058	18	2385	0	0	2012-08-02 05:18:30.529245	2012-08-02 05:19:15.865354
5207	1-5207	5	97	5720	90	9399	0	0	2012-08-02 05:17:25.780973	2012-08-02 05:19:15.865354
5487	1-5487	59	21	3324	91	7694	0	0	2012-08-02 05:19:15.865354	2012-08-02 05:19:15.865354
5223	1-5223	72	26	2886	5	7102	0	11	2012-08-02 05:17:32.045506	2012-08-02 05:19:16.274495
3742	1-3742	78	26	7866	97	8804	0	0	2012-08-02 05:07:38.764368	2012-08-02 05:19:16.274495
311	1-311	94	26	2344	53	1772	0	0	2012-08-02 04:43:59.301591	2012-08-02 05:19:16.274495
5488	1-5488	93	6	2029	26	4198	0	0	2012-08-02 05:19:16.274495	2012-08-02 05:19:16.274495
5139	1-5139	79	90	9831	99	8110	3466	0	2012-08-02 05:16:58.817752	2012-08-02 05:19:16.274495
5489	1-5489	38	7	1459	6	4621	0	0	2012-08-02 05:19:16.685433	2012-08-02 05:19:16.685433
4410	1-4410	87	66	6291	100	8794	0	3	2012-08-02 05:12:06.235149	2012-08-02 05:19:17.083133
5441	1-5441	99	66	1557	64	906	0	0	2012-08-02 05:18:57.690775	2012-08-02 05:19:17.083133
5491	1-5491	29	81	6185	8	7498	0	4	2012-08-02 05:19:17.481055	2012-08-02 05:19:17.481055
5453	1-5453	10	7	7958	75	7802	0	0	2012-08-02 05:19:02.432852	2012-08-02 05:19:17.878589
5245	1-5245	83	68	4494	22	2976	0	0	2012-08-02 05:17:40.717994	2012-08-02 05:19:17.878589
2457	1-2457	92	85	9098	78	2869	0	0	2012-08-02 04:58:49.729528	2012-08-02 05:19:19.049905
3324	1-3324	55	85	7916	11	6193	0	0	2012-08-02 05:04:50.325221	2012-08-02 05:19:19.049905
5436	1-5436	86	100	8890	45	9350	0	6	2012-08-02 05:18:55.756153	2012-08-02 05:19:19.447924
5496	1-5496	7	82	451	21	9118	0	4	2012-08-02 05:19:19.447924	2012-08-02 05:19:19.447924
4484	1-4484	66	68	8791	54	9024	0	0	2012-08-02 05:12:35.416437	2012-08-02 05:19:21.204485
5500	1-5500	91	54	1519	78	8040	0	13	2012-08-02 05:19:21.204485	2012-08-02 05:19:21.204485
5445	1-5445	88	94	1354	73	9674	0	0	2012-08-02 05:18:59.282604	2012-08-02 05:19:22.795272
4995	1-4995	44	48	8796	9	2772	0	3	2012-08-02 05:16:00.408924	2012-08-02 05:19:23.601842
5385	1-5385	93	49	3919	35	3858	0	0	2012-08-02 05:18:35.622476	2012-08-02 05:19:23.999593
5425	1-5425	33	80	5333	77	9372	0	0	2012-08-02 05:18:51.357374	2012-08-02 05:19:23.999593
5511	1-5511	94	34	4416	5	8671	0	0	2012-08-02 05:19:25.689571	2012-08-02 05:19:25.689571
5516	1-5516	47	95	1966	69	5563	0	0	2012-08-02 05:19:27.656399	2012-08-02 05:19:27.656399
5427	1-5427	59	55	802	31	7358	0	0	2012-08-02 05:18:52.132259	2012-08-02 05:19:30.064724
5522	1-5522	94	91	1488	55	7141	0	0	2012-08-02 05:19:30.064724	2012-08-02 05:19:30.064724
5474	1-5474	77	58	1367	31	8205	0	0	2012-08-02 05:19:10.694063	2012-08-02 05:19:30.871616
5503	1-5503	47	84	2040	35	1766	0	5	2012-08-02 05:19:22.397468	2012-08-02 05:19:30.871616
2942	1-2942	46	82	7062	84	3674	1345	0	2012-08-02 05:02:15.660498	2012-08-02 05:19:30.871616
4585	1-4585	81	48	7329	87	6889	0	0	2012-08-02 05:13:15.770478	2012-08-02 05:19:31.258079
5525	1-5525	55	56	260	48	8006	0	0	2012-08-02 05:19:31.258079	2012-08-02 05:19:31.258079
4137	1-4137	48	94	5853	45	4402	0	0	2012-08-02 05:10:16.381474	2012-08-02 05:19:32.429172
5457	1-5457	60	87	3237	94	6590	0	0	2012-08-02 05:19:04.012918	2012-08-02 05:19:32.429172
3618	1-3618	26	75	4873	25	8980	0	0	2012-08-02 05:06:49.284951	2012-08-02 05:19:34.009303
5455	1-5455	41	84	7153	7	5877	0	0	2012-08-02 05:19:03.217248	2012-08-02 05:19:34.815583
4382	1-4382	77	70	8696	99	2601	0	2	2012-08-02 05:11:54.808808	2012-08-02 05:19:34.815583
5504	1-5504	72	73	4228	39	1592	0	4	2012-08-02 05:19:22.795272	2012-08-02 05:19:35.98668
2101	1-2101	55	58	8636	65	2694	0	0	2012-08-02 04:56:14.437102	2012-08-02 05:19:35.98668
5432	1-5432	93	65	1415	73	5308	0	10	2012-08-02 05:18:54.13765	2012-08-02 05:19:36.384778
5506	1-5506	43	90	2920	48	6848	0	0	2012-08-02 05:19:23.601842	2012-08-02 05:19:39.533148
5336	1-5336	46	81	3984	30	2943	0	0	2012-08-02 05:18:16.420436	2012-08-02 05:19:39.533148
5546	1-5546	8	48	2516	69	8528	0	0	2012-08-02 05:19:39.533148	2012-08-02 05:19:39.533148
5126	1-5126	92	93	4761	56	3878	0	0	2012-08-02 05:16:53.546034	2012-08-02 05:19:40.715348
5404	1-5404	98	5	3770	6	2983	0	0	2012-08-02 05:18:43.137267	2012-08-02 05:19:40.715348
5184	1-5184	54	97	830	24	363	0	0	2012-08-02 05:17:16.773537	2012-08-02 05:19:40.715348
5478	1-5478	76	97	2839	5	5751	0	0	2012-08-02 05:19:12.285434	2012-08-02 05:19:40.715348
5549	1-5549	93	56	1937	97	5405	0	0	2012-08-02 05:19:40.715348	2012-08-02 05:19:40.715348
4286	1-4286	35	25	3806	85	4871	0	0	2012-08-02 05:11:15.31908	2012-08-02 05:19:41.135679
5321	1-5321	52	53	5003	90	8620	0	0	2012-08-02 05:18:10.507181	2012-08-02 05:19:41.135679
5007	1-5007	57	46	9709	65	3509	0	0	2012-08-02 05:16:05.16174	2012-08-02 05:19:41.920024
5552	1-5552	94	7	619	46	6225	0	0	2012-08-02 05:19:41.920024	2012-08-02 05:19:41.920024
5558	1-5558	35	11	1052	64	5925	0	0	2012-08-02 05:19:44.350769	2012-08-02 05:19:44.350769
5446	1-5446	35	25	5595	92	9231	0	0	2012-08-02 05:18:59.668851	2012-08-02 05:19:45.543475
3001	1-3001	68	70	4785	88	7397	0	0	2012-08-02 05:02:39.693057	2012-08-02 05:19:46.363286
4839	1-4839	21	91	9751	89	8529	0	2	2012-08-02 05:14:59.051711	2012-08-02 05:19:46.363286
3600	1-3600	3	91	4936	19	2794	0	0	2012-08-02 05:06:42.126084	2012-08-02 05:19:46.363286
5299	1-5299	78	11	3788	88	3400	0	3	2012-08-02 05:18:01.88813	2012-08-02 05:19:46.363286
4063	1-4063	21	24	4882	15	9488	0	0	2012-08-02 05:09:46.953338	2012-08-02 05:19:48.363219
5570	1-5570	8	44	2049	19	9457	0	17	2012-08-02 05:19:49.158616	2012-08-02 05:19:49.158616
5495	1-5495	39	94	2491	85	9362	0	0	2012-08-02 05:19:19.049905	2012-08-02 05:19:50.340398
5573	1-5573	85	85	6581	29	9177	0	0	2012-08-02 05:19:50.340398	2012-08-02 05:19:50.340398
5021	1-5021	100	72	4932	61	3950	0	8	2012-08-02 05:16:10.864141	2012-08-02 05:19:51.522708
4956	1-4956	8	97	6760	61	7005	0	0	2012-08-02 05:15:45.040366	2012-08-02 05:19:51.522708
5248	1-5248	88	60	1251	76	1077	0	8	2012-08-02 05:17:41.899847	2012-08-02 05:19:51.522708
5156	1-5156	11	60	8386	68	4853	0	0	2012-08-02 05:17:05.779432	2012-08-02 05:19:51.522708
4735	1-4735	62	60	1421	97	804	0	8	2012-08-02 05:14:16.964173	2012-08-02 05:19:51.522708
5576	1-5576	31	12	1349	60	7055	0	0	2012-08-02 05:19:51.522708	2012-08-02 05:19:51.522708
4695	1-4695	90	18	6101	11	6849	0	0	2012-08-02 05:14:00.748856	2012-08-02 05:19:53.478609
5124	1-5124	78	95	3191	50	8600	0	0	2012-08-02 05:16:52.75043	2012-08-02 05:19:53.876387
5533	1-5533	55	69	5008	81	4658	0	0	2012-08-02 05:19:34.42916	2012-08-02 05:19:53.876387
5152	1-5152	23	69	1629	36	1570	0	0	2012-08-02 05:17:04.164978	2012-08-02 05:19:53.876387
4859	1-4859	46	73	6584	50	8621	0	3	2012-08-02 05:15:07.153696	2012-08-02 05:19:53.876387
5568	1-5568	49	15	4949	24	3039	0	0	2012-08-02 05:19:48.363219	2012-08-02 05:19:55.11399
5563	1-5563	76	88	3920	91	8243	0	5	2012-08-02 05:19:46.363286	2012-08-02 05:19:56.704835
3932	1-3932	23	92	7997	69	5026	0	13	2012-08-02 05:08:54.50694	2012-08-02 05:19:57.102693
5590	1-5590	88	9	736	92	6630	0	0	2012-08-02 05:19:57.102693	2012-08-02 05:19:57.102693
3824	1-3824	27	92	8569	81	6651	527	0	2012-08-02 05:08:11.227378	2012-08-02 05:19:57.102693
4945	1-4945	54	57	7565	26	1395	0	10	2012-08-02 05:15:40.709727	2012-08-02 05:19:57.909205
4485	1-4485	69	94	843	83	1568	0	13	2012-08-02 05:12:35.803238	2012-08-02 05:19:59.102593
5448	1-5448	24	29	2001	16	4813	0	0	2012-08-02 05:19:00.476681	2012-08-02 05:19:59.102593
4818	1-4818	33	29	6997	94	8149	0	0	2012-08-02 05:14:50.522283	2012-08-02 05:19:59.102593
5456	1-5456	44	62	2148	80	3737	0	0	2012-08-02 05:19:03.615025	2012-08-02 05:20:00.295661
4146	1-4146	48	48	9192	28	5181	0	0	2012-08-02 05:10:20.093853	2012-08-02 05:20:00.295661
5390	1-5390	28	64	2366	65	5370	0	7	2012-08-02 05:18:37.622421	2012-08-02 05:20:00.715708
5599	1-5599	93	50	515	26	2553	0	0	2012-08-02 05:20:00.715708	2012-08-02 05:20:00.715708
3790	1-3790	11	58	6856	18	1029	0	0	2012-08-02 05:07:57.701246	2012-08-02 05:20:01.12449
5550	1-5550	56	90	2382	25	2766	0	0	2012-08-02 05:19:41.135679	2012-08-02 05:20:01.919853
5241	1-5241	93	59	4185	30	3703	0	0	2012-08-02 05:17:39.171376	2012-08-02 05:20:01.919853
5277	1-5277	3	51	3976	100	5226	0	0	2012-08-02 05:17:53.281228	2012-08-02 05:20:02.317621
5132	1-5132	22	51	5719	34	7807	0	0	2012-08-02 05:16:56.00043	2012-08-02 05:20:02.317621
5535	1-5535	41	100	1418	34	2480	0	0	2012-08-02 05:19:35.202303	2012-08-02 05:20:02.317621
5466	1-5466	88	81	1023	14	4829	0	0	2012-08-02 05:19:07.555904	2012-08-02 05:20:02.726631
5560	1-5560	75	35	3658	20	8744	0	3	2012-08-02 05:19:45.146144	2012-08-02 05:20:03.919805
2391	1-2391	12	69	5958	45	1448	0	0	2012-08-02 04:58:20.583218	2012-08-02 05:20:03.919805
5607	1-5607	84	45	679	35	8615	0	7	2012-08-02 05:20:03.919805	2012-08-02 05:20:03.919805
5307	1-5307	24	33	7115	38	5980	0	0	2012-08-02 05:18:05.0148	2012-08-02 05:20:04.748447
5582	1-5582	12	50	4639	69	9520	0	0	2012-08-02 05:19:53.876387	2012-08-02 05:20:05.146721
5520	1-5520	75	11	4491	73	6436	0	0	2012-08-02 05:19:29.280714	2012-08-02 05:20:05.146721
5034	1-5034	48	95	6091	89	3030	0	0	2012-08-02 05:16:16.15648	2012-08-02 05:20:05.146721
5492	1-5492	77	65	6932	7	8615	0	0	2012-08-02 05:19:17.878589	2012-08-02 05:20:06.772059
5415	1-5415	97	65	2187	77	2722	0	4	2012-08-02 05:18:47.424294	2012-08-02 05:20:06.772059
5581	1-5581	26	50	9623	42	9452	0	3	2012-08-02 05:19:53.478609	2012-08-02 05:20:06.772059
5238	1-5238	68	87	5788	76	6273	0	12	2012-08-02 05:17:38.000126	2012-08-02 05:20:06.772059
5614	1-5614	100	76	2764	65	9217	0	0	2012-08-02 05:20:06.772059	2012-08-02 05:20:10.452713
5623	1-5623	39	49	721	66	4965	0	6	2012-08-02 05:20:10.452713	2012-08-02 05:20:10.452713
5402	1-5402	74	42	9924	18	6968	0	0	2012-08-02 05:18:42.352972	2012-08-02 05:20:11.281546
5534	1-5534	25	37	6373	84	5924	0	0	2012-08-02 05:19:34.815583	2012-08-02 05:20:12.938805
5523	1-5523	87	74	1660	72	1801	0	0	2012-08-02 05:19:30.462469	2012-08-02 05:20:13.347451
2952	1-2952	31	51	4994	78	3241	0	0	2012-08-02 05:02:19.781764	2012-08-02 05:20:13.347451
5575	1-5575	83	51	3455	74	4206	0	0	2012-08-02 05:19:51.125241	2012-08-02 05:20:13.347451
5630	1-5630	84	57	1145	51	8001	0	0	2012-08-02 05:20:13.347451	2012-08-02 05:20:13.347451
5490	1-5490	52	17	3244	66	5236	0	5	2012-08-02 05:19:17.083133	2012-08-02 05:20:13.778682
4392	1-4392	2	33	5911	23	3756	0	0	2012-08-02 05:11:58.874463	2012-08-02 05:20:14.176531
4810	1-4810	38	9	4121	49	9710	0	0	2012-08-02 05:14:47.262559	2012-08-02 05:20:14.176531
5633	1-5633	35	54	3942	16	1314	0	7	2012-08-02 05:20:14.58528	2012-08-02 05:20:14.58528
5375	1-5375	69	81	3056	43	5260	0	0	2012-08-02 05:18:31.711209	2012-08-02 05:20:16.574087
5345	1-5345	79	98	3534	8	9049	0	9	2012-08-02 05:18:19.967122	2012-08-02 05:20:18.164651
5642	1-5642	12	15	3165	98	4074	0	4	2012-08-02 05:20:18.164651	2012-08-02 05:20:18.164651
5381	1-5381	36	88	3457	2	4372	0	6	2012-08-02 05:18:34.064505	2012-08-02 05:20:18.562542
5645	1-5645	20	94	1740	63	8941	0	3	2012-08-02 05:20:19.40235	2012-08-02 05:20:19.40235
5610	1-5610	72	69	4441	75	5733	0	0	2012-08-02 05:20:05.146721	2012-08-02 05:20:19.788803
4803	1-4803	76	69	9752	85	9550	0	5	2012-08-02 05:14:44.456345	2012-08-02 05:20:19.788803
4640	1-4640	30	5	7387	53	7010	0	0	2012-08-02 05:13:38.421778	2012-08-02 05:20:19.788803
4943	1-4943	94	85	8408	73	6708	0	0	2012-08-02 05:15:39.925228	2012-08-02 05:20:19.788803
5243	1-5243	34	53	5807	60	3841	0	0	2012-08-02 05:17:39.955695	2012-08-02 05:20:19.788803
2335	1-2335	34	54	8568	52	2572	0	0	2012-08-02 04:57:56.276795	2012-08-02 05:20:20.628695
4998	1-4998	98	1	7887	23	8051	0	3	2012-08-02 05:16:01.61269	2012-08-02 05:20:21.04861
5362	1-5362	92	27	2531	79	5131	0	0	2012-08-02 05:18:26.629311	2012-08-02 05:20:21.04861
4372	1-4372	84	52	9864	7	8347	0	0	2012-08-02 05:11:50.731679	2012-08-02 05:20:21.04861
5303	1-5303	28	1	6095	61	7352	0	10	2012-08-02 05:18:03.435294	2012-08-02 05:20:21.04861
5220	1-5220	22	68	3476	14	9629	0	4	2012-08-02 05:17:30.874287	2012-08-02 05:20:23.015189
4788	1-4788	44	95	4696	80	4675	0	0	2012-08-02 05:14:38.255822	2012-08-02 05:20:23.015189
5476	1-5476	52	64	4932	55	9878	0	0	2012-08-02 05:19:11.48989	2012-08-02 05:20:23.015189
5656	1-5656	80	54	1901	19	263	0	0	2012-08-02 05:20:23.810941	2012-08-02 05:20:23.810941
5240	1-5240	52	1	7428	4	7658	0	5	2012-08-02 05:17:38.773627	2012-08-02 05:20:24.617596
5649	1-5649	82	7	2773	1	7291	0	0	2012-08-02 05:20:21.04861	2012-08-02 05:20:24.617596
5658	1-5658	43	77	1028	7	8624	0	0	2012-08-02 05:20:24.617596	2012-08-02 05:20:24.617596
2104	1-2104	79	69	5497	5	4794	0	0	2012-08-02 04:56:15.796143	2012-08-02 05:20:25.412925
5660	1-5660	55	72	55	69	3675	0	0	2012-08-02 05:20:25.412925	2012-08-02 05:20:25.412925
5667	1-5667	26	83	4253	89	7450	0	0	2012-08-02 05:20:28.208126	2012-08-02 05:20:28.208126
5346	1-5346	24	18	2220	16	3984	0	6	2012-08-02 05:18:20.364852	2012-08-02 05:20:29.434492
5561	1-5561	18	36	5498	18	6156	0	0	2012-08-02 05:19:45.543475	2012-08-02 05:20:29.434492
5670	1-5670	44	68	2984	36	5274	0	0	2012-08-02 05:20:29.434492	2012-08-02 05:20:29.434492
4352	1-4352	60	96	2533	50	6800	0	0	2012-08-02 05:11:42.500999	2012-08-02 05:20:30.726926
3738	1-3738	31	89	5180	50	6776	0	0	2012-08-02 05:07:37.193391	2012-08-02 05:20:30.726926
5654	1-5654	76	14	3791	95	8492	0	0	2012-08-02 05:20:23.015189	2012-08-02 05:20:31.124843
3009	1-3009	25	26	6824	95	7418	0	0	2012-08-02 05:02:42.963092	2012-08-02 05:20:31.124843
5305	1-5305	38	38	4265	31	8594	0	0	2012-08-02 05:18:04.208367	2012-08-02 05:20:32.296147
4091	1-4091	84	38	6337	72	6464	0	4	2012-08-02 05:09:58.070348	2012-08-02 05:20:32.296147
5216	1-5216	31	13	2677	99	1348	0	11	2012-08-02 05:17:29.316538	2012-08-02 05:20:32.296147
4928	1-4928	27	86	4686	9	4630	0	0	2012-08-02 05:15:34.058378	2012-08-02 05:20:32.704855
5170	1-5170	50	28	2017	95	2001	0	0	2012-08-02 05:17:11.303924	2012-08-02 05:20:32.704855
4110	1-4110	54	86	8534	92	7161	0	0	2012-08-02 05:10:05.739274	2012-08-02 05:20:32.704855
4433	1-4433	91	65	7878	49	6035	0	0	2012-08-02 05:12:15.306667	2012-08-02 05:20:33.909122
4973	1-4973	48	2	4558	78	957	0	0	2012-08-02 05:15:51.713298	2012-08-02 05:20:33.909122
5681	1-5681	66	33	287	2	7913	0	5	2012-08-02 05:20:33.909122	2012-08-02 05:20:33.909122
1256	1-1256	2	65	8318	49	621	621	0	2012-08-02 04:49:31.019865	2012-08-02 05:20:33.909122
2693	1-2693	7	48	4006	95	532	0	0	2012-08-02 05:00:31.198002	2012-08-02 05:20:34.308821
5598	1-5598	2	80	3731	48	9698	0	0	2012-08-02 05:20:00.295661	2012-08-02 05:20:34.308821
4153	1-4153	9	80	4621	83	1588	0	0	2012-08-02 05:10:22.833637	2012-08-02 05:20:34.308821
5682	1-5682	3	29	1657	80	9785	0	9	2012-08-02 05:20:34.308821	2012-08-02 05:20:34.308821
5559	1-5559	88	14	9576	23	3990	0	3	2012-08-02 05:19:44.748178	2012-08-02 05:20:34.740556
5683	1-5683	8	27	768	14	9973	0	2	2012-08-02 05:20:34.740556	2012-08-02 05:20:34.740556
2871	1-2871	26	28	9614	13	9035	1146	0	2012-08-02 05:01:46.158824	2012-08-02 05:20:34.740556
4211	1-4211	22	23	8056	9	3967	179	0	2012-08-02 05:10:45.628923	2012-08-02 05:20:34.740556
5274	1-5274	10	9	2509	88	5878	0	0	2012-08-02 05:17:52.099058	2012-08-02 05:20:35.138248
3145	1-3145	31	9	8941	33	8075	0	0	2012-08-02 05:03:38.19022	2012-08-02 05:20:35.138248
1491	1-1491	33	91	5681	7	3140	0	0	2012-08-02 04:51:35.301321	2012-08-02 05:20:35.138248
623	1-623	3	91	7027	81	5311	0	0	2012-08-02 04:44:41.459035	2012-08-02 05:20:35.138248
5684	1-5684	83	49	2275	91	7572	0	0	2012-08-02 05:20:35.138248	2012-08-02 05:20:35.138248
118	1-118	65	91	2986	81	78	78	0	2012-08-02 04:43:54.805703	2012-08-02 05:20:35.138248
611	1-611	74	91	8098	81	1455	1455	0	2012-08-02 04:44:39.005288	2012-08-02 05:20:35.138248
5350	1-5350	28	78	7170	68	6533	0	0	2012-08-02 05:18:21.945044	2012-08-02 05:20:35.547306
3405	1-3405	52	78	6576	7	4772	0	0	2012-08-02 05:05:23.353019	2012-08-02 05:20:35.547306
5344	1-5344	83	78	6826	29	7109	0	0	2012-08-02 05:18:19.580327	2012-08-02 05:20:35.547306
5322	1-5322	53	78	3186	83	7153	0	0	2012-08-02 05:18:10.905002	2012-08-02 05:20:35.547306
5604	1-5604	69	23	5931	9	8156	0	5	2012-08-02 05:20:02.726631	2012-08-02 05:20:35.547306
5677	1-5677	32	30	566	38	1096	0	4	2012-08-02 05:20:32.296147	2012-08-02 05:20:38.353369
4101	1-4101	100	5	4070	76	3809	0	0	2012-08-02 05:10:02.091947	2012-08-02 05:20:38.353369
5692	1-5692	26	28	1612	93	7893	0	7	2012-08-02 05:20:38.353369	2012-08-02 05:20:38.353369
5694	1-5694	85	72	679	94	7641	0	0	2012-08-02 05:20:39.149373	2012-08-02 05:20:39.149373
5209	1-5209	30	88	6605	53	4339	0	0	2012-08-02 05:17:26.587391	2012-08-02 05:20:39.56919
5109	1-5109	20	88	9674	44	9042	0	0	2012-08-02 05:16:46.696024	2012-08-02 05:20:39.56919
5537	1-5537	28	39	1482	58	8083	0	0	2012-08-02 05:19:35.98668	2012-08-02 05:20:39.56919
5047	1-5047	24	18	4051	44	9544	0	0	2012-08-02 05:16:21.393145	2012-08-02 05:20:39.56919
5646	1-5646	97	73	3438	69	8464	0	4	2012-08-02 05:20:19.788803	2012-08-02 05:20:39.56919
5486	1-5486	50	92	7222	55	4798	0	6	2012-08-02 05:19:15.467543	2012-08-02 05:20:39.56919
5695	1-5695	63	34	3773	26	9136	0	0	2012-08-02 05:20:39.56919	2012-08-02 05:20:39.56919
3767	1-3767	47	4	5822	32	3578	0	5	2012-08-02 05:07:48.69649	2012-08-02 05:20:40.398139
5100	1-5100	20	4	8112	75	4507	0	0	2012-08-02 05:16:43.049796	2012-08-02 05:20:40.398139
5135	1-5135	13	82	589	72	492	0	0	2012-08-02 05:16:57.19344	2012-08-02 05:20:40.398139
5697	1-5697	92	68	134	4	7428	0	3	2012-08-02 05:20:40.398139	2012-08-02 05:20:40.398139
5673	1-5673	64	50	6543	96	3176	0	0	2012-08-02 05:20:30.726926	2012-08-02 05:20:41.646374
5380	1-5380	60	76	8600	59	9538	0	0	2012-08-02 05:18:33.678212	2012-08-02 05:20:44.408457
5143	1-5143	47	41	4547	26	4091	0	0	2012-08-02 05:17:00.463977	2012-08-02 05:20:46.772284
4473	1-4473	69	36	5598	70	2718	0	3	2012-08-02 05:12:31.086454	2012-08-02 05:20:46.772284
5713	1-5713	27	31	713	62	7480	0	0	2012-08-02 05:20:46.772284	2012-08-02 05:20:46.772284
5014	1-5014	30	49	6016	66	4976	0	5	2012-08-02 05:16:07.991335	2012-08-02 05:20:47.578829
4931	1-4931	66	83	5671	79	3033	0	0	2012-08-02 05:15:35.240293	2012-08-02 05:20:47.578829
5715	1-5715	1	15	3778	49	9957	0	0	2012-08-02 05:20:47.578829	2012-08-02 05:20:47.578829
4969	1-4969	34	40	8238	100	4496	0	10	2012-08-02 05:15:50.144595	2012-08-02 05:20:48.794687
5254	1-5254	97	1	9709	38	7892	0	0	2012-08-02 05:17:44.25317	2012-08-02 05:20:49.203208
5158	1-5158	3	63	7811	67	2956	0	3	2012-08-02 05:17:06.563774	2012-08-02 05:20:49.203208
1985	1-1985	42	1	8378	72	3833	0	0	2012-08-02 04:55:24.527706	2012-08-02 05:20:49.203208
5459	1-5459	61	93	7531	95	7130	0	14	2012-08-02 05:19:04.809151	2012-08-02 05:20:49.623292
5707	1-5707	78	70	1281	76	6999	0	8	2012-08-02 05:20:44.408457	2012-08-02 05:20:50.816447
5510	1-5510	58	16	5457	76	5980	0	0	2012-08-02 05:19:25.204186	2012-08-02 05:20:50.816447
5602	1-5602	24	30	5785	60	8245	0	0	2012-08-02 05:20:01.919853	2012-08-02 05:20:50.816447
1815	1-1815	15	30	6476	31	5323	0	0	2012-08-02 04:54:09.331336	2012-08-02 05:20:50.816447
5723	1-5723	91	76	8049	30	7410	0	0	2012-08-02 05:20:50.816447	2012-08-02 05:20:50.816447
4580	1-4580	76	60	8358	16	7008	1049	0	2012-08-02 05:13:13.737303	2012-08-02 05:20:50.816447
4740	1-4740	17	11	1569	88	1394	0	0	2012-08-02 05:14:18.986142	2012-08-02 05:20:51.236637
4614	1-4614	41	70	8350	46	9678	0	4	2012-08-02 05:13:27.715178	2012-08-02 05:20:52.473877
5727	1-5727	32	57	2082	84	6258	0	0	2012-08-02 05:20:52.473877	2012-08-02 05:20:52.473877
4908	1-4908	57	62	5656	36	6704	0	0	2012-08-02 05:15:26.266857	2012-08-02 05:20:52.904843
5377	1-5377	55	55	6214	62	6595	0	0	2012-08-02 05:18:32.495887	2012-08-02 05:20:52.904843
1064	1-1064	8	11	4683	55	250	250	4	2012-08-02 04:47:49.591511	2012-08-02 05:20:52.904843
2193	1-2193	50	89	3846	68	3333	0	0	2012-08-02 04:56:54.4112	2012-08-02 05:20:53.336057
4627	1-4627	70	53	7468	31	5967	0	0	2012-08-02 05:13:33.05194	2012-08-02 05:20:54.186713
1279	1-1279	70	69	8827	53	5822	1604	0	2012-08-02 04:49:42.575542	2012-08-02 05:20:54.186713
5663	1-5663	70	31	3654	32	5372	0	0	2012-08-02 05:20:26.628315	2012-08-02 05:20:54.653925
5737	1-5737	99	71	250	57	6367	0	11	2012-08-02 05:20:56.837819	2012-08-02 05:20:56.837819
5746	1-5746	83	60	827	39	4320	0	11	2012-08-02 05:21:00.571757	2012-08-02 05:21:00.571757
5624	1-5624	15	1	1538	85	6030	0	0	2012-08-02 05:20:10.872788	2012-08-02 05:21:03.92126
5754	1-5754	91	73	1306	98	1467	0	2	2012-08-02 05:21:03.92126	2012-08-02 05:21:03.92126
5719	1-5719	84	6	396	28	6199	0	0	2012-08-02 05:20:49.203208	2012-08-02 05:21:05.180877
5320	1-5320	54	82	3898	84	8779	0	5	2012-08-02 05:18:10.120466	2012-08-02 05:21:05.749486
5616	1-5616	27	50	3200	83	9736	0	0	2012-08-02 05:20:07.602742	2012-08-02 05:21:05.749486
5451	1-5451	64	82	3060	46	9552	0	6	2012-08-02 05:19:01.637224	2012-08-02 05:21:05.749486
5219	1-5219	91	49	7143	83	9123	0	0	2012-08-02 05:17:30.465323	2012-08-02 05:21:05.749486
5477	1-5477	14	53	8774	34	9979	0	0	2012-08-02 05:19:11.898845	2012-08-02 05:21:06.176608
4018	1-4018	59	70	472	74	191	0	0	2012-08-02 05:09:29.021536	2012-08-02 05:21:06.960805
5761	1-5761	28	22	1157	86	4245	0	0	2012-08-02 05:21:06.960805	2012-08-02 05:21:06.960805
1884	1-1884	68	82	9266	22	7771	1561	0	2012-08-02 04:54:40.433022	2012-08-02 05:21:06.960805
4764	1-4764	39	94	3099	15	9339	0	0	2012-08-02 05:14:28.5994	2012-08-02 05:21:08.165299
5368	1-5368	77	84	4922	63	4035	0	6	2012-08-02 05:18:28.993554	2012-08-02 05:21:08.165299
5585	1-5585	84	84	7756	94	7639	0	8	2012-08-02 05:19:55.11399	2012-08-02 05:21:08.165299
5538	1-5538	34	73	5266	89	2972	0	0	2012-08-02 05:19:36.384778	2012-08-02 05:21:08.562971
1554	1-1554	72	38	6698	99	4052	0	0	2012-08-02 04:52:05.906083	2012-08-02 05:21:08.949598
5779	1-5779	12	50	182	87	699	0	3	2012-08-02 05:21:14.12016	2012-08-02 05:21:14.12016
5627	1-5627	64	18	1165	54	9580	0	0	2012-08-02 05:20:12.110278	2012-08-02 05:21:16.539619
5758	1-5758	82	83	5153	82	9944	0	0	2012-08-02 05:21:05.749486	2012-08-02 05:21:16.93694
5115	1-5115	3	83	2864	43	3136	0	3	2012-08-02 05:16:49.115869	2012-08-02 05:21:16.93694
2163	1-2163	23	35	3823	31	6479	0	0	2012-08-02 04:56:41.318459	2012-08-02 05:21:18.108254
4676	1-4676	23	22	8680	64	2812	0	0	2012-08-02 05:13:53.082359	2012-08-02 05:21:18.108254
5789	1-5789	97	10	2264	35	4394	0	0	2012-08-02 05:21:18.108254	2012-08-02 05:21:18.108254
5526	1-5526	15	75	3363	93	7406	0	0	2012-08-02 05:19:31.656003	2012-08-02 05:21:18.915135
5793	1-5793	50	14	219	80	5175	0	3	2012-08-02 05:21:19.721592	2012-08-02 05:21:19.721592
5242	1-5242	66	70	8784	98	3578	0	0	2012-08-02 05:17:39.569291	2012-08-02 05:21:22.0965
5799	1-5799	27	63	3933	70	7997	0	0	2012-08-02 05:21:22.0965	2012-08-02 05:21:22.0965
5387	1-5387	33	81	5471	65	8483	0	3	2012-08-02 05:18:36.440203	2012-08-02 05:21:24.07404
5804	1-5804	57	72	2137	3	5627	0	0	2012-08-02 05:21:24.07404	2012-08-02 05:21:24.07404
5700	1-5700	98	96	520	66	2705	0	0	2012-08-02 05:20:41.646374	2012-08-02 05:21:26.09599
3838	1-3838	56	12	5398	17	6364	0	0	2012-08-02 05:08:16.95037	2012-08-02 05:21:26.09599
4897	1-4897	7	96	3193	89	5602	0	10	2012-08-02 05:15:21.991391	2012-08-02 05:21:26.09599
5298	1-5298	39	17	9235	27	9631	0	0	2012-08-02 05:18:01.512489	2012-08-02 05:21:26.09599
5386	1-5386	26	89	5966	47	4023	0	6	2012-08-02 05:18:36.049046	2012-08-02 05:21:26.09599
4147	1-4147	11	12	8988	27	3759	0	0	2012-08-02 05:10:20.480499	2012-08-02 05:21:26.09599
5338	1-5338	94	96	3571	12	7093	0	0	2012-08-02 05:18:17.227055	2012-08-02 05:21:26.09599
5685	1-5685	20	68	2794	78	7587	0	5	2012-08-02 05:20:35.547306	2012-08-02 05:21:26.913775
4295	1-4295	62	98	5328	1	7036	0	0	2012-08-02 05:11:19.009509	2012-08-02 05:21:26.913775
2275	1-2275	12	35	5670	78	9500	0	0	2012-08-02 04:57:30.134491	2012-08-02 05:21:26.913775
5304	1-5304	17	80	7709	100	6089	0	0	2012-08-02 05:18:03.832738	2012-08-02 05:21:26.913775
5811	1-5811	87	78	3727	98	7067	0	0	2012-08-02 05:21:26.913775	2012-08-02 05:21:26.913775
5732	1-5732	84	30	1560	25	4942	0	0	2012-08-02 05:20:54.653925	2012-08-02 05:21:27.333921
5127	1-5127	69	33	5933	7	4564	0	4	2012-08-02 05:16:53.944114	2012-08-02 05:21:27.333921
5809	1-5809	45	27	7081	96	9226	0	0	2012-08-02 05:21:26.09599	2012-08-02 05:21:27.333921
5044	1-5044	34	30	4766	66	5230	0	5	2012-08-02 05:16:20.17808	2012-08-02 05:21:27.333921
5812	1-5812	37	95	1893	30	8423	0	4	2012-08-02 05:21:27.333921	2012-08-02 05:21:27.333921
4083	1-4083	96	66	9122	12	4655	624	0	2012-08-02 05:09:54.899396	2012-08-02 05:21:27.333921
2232	1-2232	14	50	4625	88	1055	0	6	2012-08-02 04:57:11.371973	2012-08-02 05:21:28.562718
5816	1-5816	24	88	971	39	3646	0	2	2012-08-02 05:21:28.959996	2012-08-02 05:21:28.959996
5617	1-5617	63	46	4963	57	6599	0	0	2012-08-02 05:20:08.011339	2012-08-02 05:21:29.369037
5755	1-5755	95	57	3065	40	5706	0	6	2012-08-02 05:21:04.351941	2012-08-02 05:21:30.57304
5512	1-5512	57	38	3093	40	9102	0	0	2012-08-02 05:19:26.076553	2012-08-02 05:21:30.57304
3869	1-3869	90	6	6665	100	5493	0	0	2012-08-02 05:08:29.280494	2012-08-02 05:21:30.57304
5820	1-5820	45	40	5560	6	8163	0	0	2012-08-02 05:21:30.57304	2012-08-02 05:21:30.57304
5825	1-5825	22	37	2033	50	737	0	0	2012-08-02 05:21:32.59489	2012-08-02 05:21:32.59489
4341	1-4341	20	59	8806	8	3870	0	0	2012-08-02 05:11:37.982681	2012-08-02 05:21:32.981537
4946	1-4946	33	59	5162	34	2336	0	0	2012-08-02 05:15:41.096305	2012-08-02 05:21:32.981537
5177	1-5177	33	20	8888	56	7380	136	0	2012-08-02 05:17:14.03358	2012-08-02 05:21:32.981537
4698	1-4698	47	92	7485	87	7333	0	0	2012-08-02 05:14:01.886651	2012-08-02 05:21:33.390577
5827	1-5827	7	82	111	81	1551	0	0	2012-08-02 05:21:33.390577	2012-08-02 05:21:33.390577
5625	1-5625	6	59	3764	42	5525	0	0	2012-08-02 05:20:11.281546	2012-08-02 05:21:34.197076
5829	1-5829	5	7	755	59	5635	0	0	2012-08-02 05:21:34.197076	2012-08-02 05:21:34.197076
4697	1-4697	18	35	3653	14	3593	0	0	2012-08-02 05:14:01.510914	2012-08-02 05:21:36.605814
2752	1-2752	53	35	3113	53	4603	0	2	2012-08-02 05:00:56.258467	2012-08-02 05:21:36.605814
5835	1-5835	1	84	2059	35	9757	0	6	2012-08-02 05:21:36.605814	2012-08-02 05:21:36.605814
5836	1-5836	38	55	190	68	562	0	0	2012-08-02 05:21:37.003449	2012-08-02 05:21:37.003449
5562	1-5562	20	20	9090	18	6901	0	6	2012-08-02 05:19:45.95467	2012-08-02 05:21:39.798349
5730	1-5730	94	76	4799	55	4517	0	5	2012-08-02 05:20:53.767072	2012-08-02 05:21:41.168326
5850	1-5850	25	36	686	87	8585	0	9	2012-08-02 05:21:42.781261	2012-08-02 05:21:42.781261
5853	1-5853	32	52	1711	53	2954	0	14	2012-08-02 05:21:44.007471	2012-08-02 05:21:44.007471
5452	1-5452	82	10	3194	68	3994	0	12	2012-08-02 05:19:02.034801	2012-08-02 05:21:45.609331
4023	1-4023	23	69	6903	95	4670	0	6	2012-08-02 05:09:31.021227	2012-08-02 05:21:46.82457
5860	1-5860	90	13	1166	69	4024	0	0	2012-08-02 05:21:46.82457	2012-08-02 05:21:46.82457
5861	1-5861	64	21	1214	58	5446	0	0	2012-08-02 05:21:47.233426	2012-08-02 05:21:47.233426
5603	1-5603	71	34	8347	20	7303	0	5	2012-08-02 05:20:02.317621	2012-08-02 05:21:48.824493
5865	1-5865	58	57	4017	35	9989	0	0	2012-08-02 05:21:48.824493	2012-08-02 05:21:48.824493
5724	1-5724	89	88	4420	11	9437	0	9	2012-08-02 05:20:51.236637	2012-08-02 05:21:52.448072
5874	1-5874	2	31	4653	64	6228	0	4	2012-08-02 05:21:52.448072	2012-08-02 05:21:52.448072
4666	1-4666	93	21	4802	20	1449	0	0	2012-08-02 05:13:48.99551	2012-08-02 05:21:52.85687
5875	1-5875	16	60	208	93	3567	0	0	2012-08-02 05:21:52.85687	2012-08-02 05:21:52.85687
4170	1-4170	43	92	6069	21	3224	0	0	2012-08-02 05:10:29.473984	2012-08-02 05:21:54.491972
5763	1-5763	59	93	2856	21	9282	0	2	2012-08-02 05:21:07.756375	2012-08-02 05:21:54.491972
5672	1-5672	73	3	2653	81	5554	0	11	2012-08-02 05:20:30.241137	2012-08-02 05:21:54.491972
5710	1-5710	30	95	1002	21	1686	0	4	2012-08-02 05:20:45.590826	2012-08-02 05:21:54.491972
5885	1-5885	60	62	861	76	1970	0	0	2012-08-02 05:21:56.900663	2012-08-02 05:21:56.900663
4618	1-4618	24	5	9739	93	9197	0	2	2012-08-02 05:13:29.350206	2012-08-02 05:21:59.297931
5891	1-5891	21	76	893	5	3054	0	0	2012-08-02 05:21:59.297931	2012-08-02 05:21:59.297931
4874	1-4874	54	32	2561	61	2043	0	0	2012-08-02 05:15:12.942954	2012-08-02 05:21:59.720189
5892	1-5892	8	13	1301	32	2743	0	0	2012-08-02 05:21:59.720189	2012-08-02 05:21:59.720189
5371	1-5371	8	79	7335	93	4678	0	2	2012-08-02 05:18:30.153525	2012-08-02 05:22:00.526392
5894	1-5894	19	30	276	16	6445	0	3	2012-08-02 05:22:00.526392	2012-08-02 05:22:00.526392
5482	1-5482	42	86	4954	70	9694	0	7	2012-08-02 05:19:13.832226	2012-08-02 05:22:00.946358
5901	1-5901	42	34	435	90	8545	0	8	2012-08-02 05:22:03.354865	2012-08-02 05:22:03.354865
4875	1-4875	12	96	2809	25	5002	0	0	2012-08-02 05:15:13.340445	2012-08-02 05:22:04.138898
933	1-933	83	60	9710	2	9425	0	0	2012-08-02 04:46:34.319458	2012-08-02 05:22:04.138898
4409	1-4409	3	96	6848	20	8690	0	0	2012-08-02 05:12:05.8596	2012-08-02 05:22:04.138898
4504	1-4504	89	78	7586	2	6408	0	7	2012-08-02 05:12:43.272466	2012-08-02 05:22:04.138898
5294	1-5294	96	10	6503	76	9046	0	0	2012-08-02 05:17:59.932483	2012-08-02 05:22:04.956708
3937	1-3937	45	23	1534	14	1904	0	0	2012-08-02 05:08:56.512994	2012-08-02 05:22:05.374907
5577	1-5577	5	71	2917	70	6529	0	2	2012-08-02 05:19:51.942855	2012-08-02 05:22:05.374907
5906	1-5906	99	41	3156	23	6633	0	4	2012-08-02 05:22:05.374907	2012-08-02 05:22:05.374907
5458	1-5458	62	20	7562	12	8303	0	11	2012-08-02 05:19:04.411364	2012-08-02 05:22:08.691983
5915	1-5915	84	88	995	25	4938	0	0	2012-08-02 05:22:09.100844	2012-08-02 05:22:09.100844
5922	1-5922	6	58	1941	29	4823	0	4	2012-08-02 05:22:11.995275	2012-08-02 05:22:11.995275
3675	1-3675	20	91	9952	61	6383	0	5	2012-08-02 05:07:12.225156	2012-08-02 05:22:12.823799
5917	1-5917	76	93	2651	91	5773	0	0	2012-08-02 05:22:09.918306	2012-08-02 05:22:12.823799
1812	1-1812	80	86	4667	32	3878	0	0	2012-08-02 04:54:07.950216	2012-08-02 05:22:12.823799
5348	1-5348	33	76	4212	66	1798	0	3	2012-08-02 05:18:21.171809	2012-08-02 05:22:12.823799
5924	1-5924	5	66	2686	93	9480	0	3	2012-08-02 05:22:12.823799	2012-08-02 05:22:12.823799
3425	1-3425	70	54	8926	11	3314	0	0	2012-08-02 05:05:31.396523	2012-08-02 05:22:14.083462
2919	1-2919	91	22	9817	1	2740	0	3	2012-08-02 05:02:06.10366	2012-08-02 05:22:14.083462
5817	1-5817	66	75	1090	76	5603	0	0	2012-08-02 05:21:29.369037	2012-08-02 05:22:14.503512
4333	1-4333	3	75	3965	62	7379	0	0	2012-08-02 05:11:34.645509	2012-08-02 05:22:14.503512
4722	1-4722	92	28	6666	55	2926	0	0	2012-08-02 05:14:11.625622	2012-08-02 05:22:15.321079
5931	1-5931	83	53	421	49	4166	0	0	2012-08-02 05:22:15.751915	2012-08-02 05:22:15.751915
5897	1-5897	78	96	4277	10	3402	0	0	2012-08-02 05:22:01.730576	2012-08-02 05:22:17.022646
5934	1-5934	84	89	2969	96	6052	0	0	2012-08-02 05:22:17.022646	2012-08-02 05:22:17.022646
5396	1-5396	27	22	4720	3	3258	0	0	2012-08-02 05:18:39.964761	2012-08-02 05:22:18.679652
5449	1-5449	59	28	5416	1	3578	0	0	2012-08-02 05:19:00.863294	2012-08-02 05:22:20.756743
4641	1-4641	82	72	6363	1	4375	0	6	2012-08-02 05:13:38.830512	2012-08-02 05:22:20.756743
5636	1-5636	84	24	1601	36	4219	0	0	2012-08-02 05:20:15.778755	2012-08-02 05:22:20.756743
5597	1-5597	76	24	1606	58	5396	0	0	2012-08-02 05:19:59.897839	2012-08-02 05:22:20.756743
5823	1-5823	93	20	2669	35	2234	0	8	2012-08-02 05:21:31.799411	2012-08-02 05:22:20.756743
5943	1-5943	2	1	1730	24	6973	0	0	2012-08-02 05:22:20.756743	2012-08-02 05:22:20.756743
4565	1-4565	55	47	5484	52	9595	0	0	2012-08-02 05:13:07.594464	2012-08-02 05:22:21.18795
5674	1-5674	2	62	8390	41	6843	0	0	2012-08-02 05:20:31.124843	2012-08-02 05:22:22.016445
4474	1-4474	82	62	8042	20	5109	0	0	2012-08-02 05:12:31.472723	2012-08-02 05:22:22.016445
5831	1-5831	21	62	1180	12	543	0	0	2012-08-02 05:21:35.003666	2012-08-02 05:22:22.016445
5946	1-5946	86	25	1412	62	7641	0	0	2012-08-02 05:22:22.016445	2012-08-02 05:22:22.016445
5678	1-5678	84	36	2721	86	3962	0	6	2012-08-02 05:20:32.704855	2012-08-02 05:22:22.856415
2662	1-2662	8	36	5149	71	2298	0	0	2012-08-02 05:00:18.115037	2012-08-02 05:22:22.856415
5609	1-5609	65	30	4176	33	8837	0	7	2012-08-02 05:20:04.748447	2012-08-02 05:22:23.275992
4034	1-4034	85	75	975	17	1956	0	0	2012-08-02 05:09:35.407286	2012-08-02 05:22:24.104661
5927	1-5927	7	40	1206	54	8910	0	0	2012-08-02 05:22:14.083462	2012-08-02 05:22:24.104661
5121	1-5121	70	26	4777	88	6552	0	0	2012-08-02 05:16:51.535705	2012-08-02 05:22:25.764045
5403	1-5403	94	13	687	27	526	0	0	2012-08-02 05:18:42.751057	2012-08-02 05:22:29.497884
4965	1-4965	53	44	3063	14	3282	0	0	2012-08-02 05:15:48.553536	2012-08-02 05:22:30.73537
5948	1-5948	51	68	3481	36	4162	0	4	2012-08-02 05:22:22.856415	2012-08-02 05:22:31.132961
5968	1-5968	37	88	1330	68	8329	0	7	2012-08-02 05:22:31.132961	2012-08-02 05:22:31.132961
5041	1-5041	31	4	5071	90	3149	0	8	2012-08-02 05:16:18.973723	2012-08-02 05:22:32.381622
5370	1-5370	39	7	5292	68	9782	0	0	2012-08-02 05:18:29.766702	2012-08-02 05:22:32.381622
5971	1-5971	13	87	2570	79	6827	0	3	2012-08-02 05:22:32.381622	2012-08-02 05:22:32.381622
1354	1-1354	51	98	7866	42	7015	0	0	2012-08-02 04:50:22.923951	2012-08-02 05:22:32.790361
5883	1-5883	85	23	4334	42	1943	0	0	2012-08-02 05:21:56.094528	2012-08-02 05:22:32.790361
5729	1-5729	27	23	1986	73	7033	0	4	2012-08-02 05:20:53.336057	2012-08-02 05:22:32.790361
5972	1-5972	11	42	2392	23	8444	0	0	2012-08-02 05:22:32.790361	2012-08-02 05:22:32.790361
5973	1-5973	44	13	67	26	8645	0	7	2012-08-02 05:22:33.210406	2012-08-02 05:22:33.210406
3945	1-3945	6	1	5009	40	8617	0	0	2012-08-02 05:08:59.708451	2012-08-02 05:22:34.028107
5203	1-5203	80	33	3867	46	9030	0	0	2012-08-02 05:17:24.23376	2012-08-02 05:22:34.447693
4996	1-4996	14	15	5472	14	6990	0	6	2012-08-02 05:16:00.852653	2012-08-02 05:22:34.845457
5977	1-5977	33	84	4134	39	5296	0	8	2012-08-02 05:22:34.845457	2012-08-02 05:22:35.243312
4881	1-4881	100	64	3905	10	9000	0	0	2012-08-02 05:15:15.773543	2012-08-02 05:22:35.243312
5978	1-5978	40	39	5267	82	7336	0	0	2012-08-02 05:22:35.243312	2012-08-02 05:22:35.243312
5643	1-5643	82	70	6565	41	5891	0	0	2012-08-02 05:20:18.562542	2012-08-02 05:22:37.331277
5843	1-5843	66	87	9951	20	7609	0	0	2012-08-02 05:21:39.798349	2012-08-02 05:22:37.331277
5524	1-5524	55	31	9160	82	4801	0	0	2012-08-02 05:19:30.871616	2012-08-02 05:22:37.331277
1938	1-1938	4	96	4717	14	8370	0	2	2012-08-02 04:55:03.977041	2012-08-02 05:22:37.751465
5984	1-5984	27	14	3208	96	5827	0	2	2012-08-02 05:22:37.751465	2012-08-02 05:22:37.751465
2998	1-2998	1	3	4516	68	2186	0	0	2012-08-02 05:02:38.477955	2012-08-02 05:22:38.160119
4929	1-4929	72	3	7862	20	6982	0	4	2012-08-02 05:15:34.467197	2012-08-02 05:22:38.160119
5985	1-5985	25	58	3230	3	9806	0	0	2012-08-02 05:22:38.160119	2012-08-02 05:22:38.160119
5462	1-5462	51	20	9502	18	6200	1690	6	2012-08-02 05:19:05.986972	2012-08-02 05:22:38.160119
5438	1-5438	45	88	9679	68	8241	0	2	2012-08-02 05:18:56.541436	2012-08-02 05:22:39.430825
5173	1-5173	99	87	2982	79	3356	0	0	2012-08-02 05:17:12.48657	2012-08-02 05:22:39.8396
5782	1-5782	55	26	1472	58	7500	0	0	2012-08-02 05:21:15.335038	2012-08-02 05:22:41.099334
706	1-706	66	73	7656	55	2565	0	0	2012-08-02 04:45:00.172377	2012-08-02 05:22:41.099334
5530	1-5530	83	98	4953	26	9516	0	0	2012-08-02 05:19:33.213736	2012-08-02 05:22:41.099334
5992	1-5992	59	34	449	98	5103	0	11	2012-08-02 05:22:41.099334	2012-08-02 05:22:41.099334
5214	1-5214	2	45	8511	8	6371	0	0	2012-08-02 05:17:28.53173	2012-08-02 05:22:41.938874
5764	1-5764	82	14	8501	84	9472	0	0	2012-08-02 05:21:08.165299	2012-08-02 05:22:41.938874
5994	1-5994	54	28	4246	45	6195	0	6	2012-08-02 05:22:41.938874	2012-08-02 05:22:41.938874
5998	1-5998	13	3	878	34	2231	0	11	2012-08-02 05:22:43.618306	2012-08-02 05:22:43.618306
5179	1-5179	59	1	8942	82	3840	0	0	2012-08-02 05:17:14.806893	2012-08-02 05:22:44.447082
4999	1-4999	62	98	1144	80	3258	0	0	2012-08-02 05:16:01.999676	2012-08-02 05:22:44.447082
4909	1-4909	93	29	2879	10	3040	0	4	2012-08-02 05:15:26.653845	2012-08-02 05:22:44.85593
3903	1-3903	60	21	8070	12	3077	0	0	2012-08-02 05:08:42.971758	2012-08-02 05:22:44.85593
5258	1-5258	48	60	4185	72	3817	0	4	2012-08-02 05:17:45.821974	2012-08-02 05:22:45.275967
3210	1-3210	19	8	9712	81	8247	0	0	2012-08-02 05:04:04.584118	2012-08-02 05:22:47.761713
6008	1-6008	18	92	257	8	8841	0	0	2012-08-02 05:22:47.761713	2012-08-02 05:22:47.761713
6000	1-6000	97	80	6938	1	7218	0	0	2012-08-02 05:22:44.447082	2012-08-02 05:22:48.601506
5949	1-5949	30	80	2813	100	2230	0	6	2012-08-02 05:22:23.275992	2012-08-02 05:22:48.601506
5384	1-5384	10	42	5472	100	3724	0	0	2012-08-02 05:18:35.235685	2012-08-02 05:22:48.601506
1697	1-1697	68	1	5124	40	977	977	0	2012-08-02 04:53:15.754234	2012-08-02 05:22:48.601506
3010	1-3010	27	1	9084	40	9513	6786	0	2012-08-02 05:02:43.371909	2012-08-02 05:22:48.601506
3016	1-3016	71	89	3356	17	2581	0	2	2012-08-02 05:02:45.7475	2012-08-02 05:22:49.452457
5296	1-5296	42	89	5612	2	9196	0	6	2012-08-02 05:18:00.716993	2012-08-02 05:22:49.452457
5335	1-5335	61	89	5748	69	6055	0	0	2012-08-02 05:18:16.03378	2012-08-02 05:22:49.452457
6012	1-6012	41	21	1677	89	9810	0	0	2012-08-02 05:22:49.452457	2012-08-02 05:22:49.452457
5138	1-5138	19	21	1234	37	9707	0	0	2012-08-02 05:16:58.408791	2012-08-02 05:22:50.305201
5518	1-5518	42	32	2902	85	5910	0	0	2012-08-02 05:19:28.462794	2012-08-02 05:22:50.305201
2734	1-2734	6	92	3291	37	3397	0	0	2012-08-02 05:00:48.667658	2012-08-02 05:22:50.305201
6014	1-6014	18	37	2945	32	6849	0	0	2012-08-02 05:22:50.305201	2012-08-02 05:22:50.305201
6001	1-6001	87	12	1028	29	7969	0	0	2012-08-02 05:22:44.85593	2012-08-02 05:22:50.747215
4882	1-4882	59	95	6990	62	9234	0	0	2012-08-02 05:15:16.158395	2012-08-02 05:22:50.747215
6015	1-6015	11	54	3846	12	7052	0	0	2012-08-02 05:22:50.747215	2012-08-02 05:22:50.747215
5181	1-5181	75	97	888	37	5336	0	0	2012-08-02 05:17:15.591696	2012-08-02 05:22:53.255377
6021	1-6021	46	37	2581	5	4653	0	0	2012-08-02 05:22:53.255377	2012-08-02 05:22:53.255377
5731	1-5731	78	38	1793	69	7846	0	0	2012-08-02 05:20:54.186713	2012-08-02 05:22:53.66409
4948	1-4948	17	24	2816	60	4440	0	0	2012-08-02 05:15:41.880488	2012-08-02 05:22:53.66409
6022	1-6022	10	94	2068	24	9611	0	8	2012-08-02 05:22:53.66409	2012-08-02 05:22:53.66409
5765	1-5765	69	98	2706	73	5821	0	0	2012-08-02 05:21:08.562971	2012-08-02 05:22:56.172295
6028	1-6028	95	90	218	98	5330	0	0	2012-08-02 05:22:56.172295	2012-08-02 05:22:56.172295
6010	1-6010	85	79	1779	80	7210	0	0	2012-08-02 05:22:48.601506	2012-08-02 05:22:58.635536
6038	1-6038	27	31	845	42	6962	0	4	2012-08-02 05:23:00.337027	2012-08-02 05:23:00.337027
5983	1-5983	52	41	1409	31	8765	0	0	2012-08-02 05:22:37.331277	2012-08-02 05:23:00.767776
6042	1-6042	34	23	512	8	4880	0	5	2012-08-02 05:23:02.049405	2012-08-02 05:23:02.049405
4694	1-4694	83	3	342	37	1834	0	4	2012-08-02 05:14:00.361849	2012-08-02 05:23:02.46912
5903	1-5903	59	71	3337	96	8744	0	11	2012-08-02 05:22:04.138898	2012-08-02 05:23:02.878025
3222	1-3222	79	30	6806	36	2938	0	0	2012-08-02 05:04:09.464383	2012-08-02 05:23:03.286719
6034	1-6034	29	42	6019	19	5521	0	0	2012-08-02 05:22:58.635536	2012-08-02 05:23:03.286719
6045	1-6045	86	37	4421	42	9494	0	0	2012-08-02 05:23:03.286719	2012-08-02 05:23:03.286719
5803	1-5803	15	79	4130	4	4223	0	6	2012-08-02 05:21:23.676376	2012-08-02 05:23:05.762544
5077	1-5077	45	71	2682	8	625	0	0	2012-08-02 05:16:33.570682	2012-08-02 05:23:06.160818
6052	1-6052	60	11	1072	23	2288	0	0	2012-08-02 05:23:06.160818	2012-08-02 05:23:06.160818
4340	1-4340	45	3	1122	37	3395	0	0	2012-08-02 05:11:37.573575	2012-08-02 05:23:06.956293
5826	1-5826	30	56	855	59	2866	0	0	2012-08-02 05:21:32.981537	2012-08-02 05:23:07.342959
5718	1-5718	58	100	6842	1	8089	0	0	2012-08-02 05:20:48.794687	2012-08-02 05:23:07.342959
6055	1-6055	55	53	360	56	9105	0	0	2012-08-02 05:23:07.342959	2012-08-02 05:23:07.342959
6058	1-6058	91	90	348	78	6724	0	0	2012-08-02 05:23:08.569649	2012-08-02 05:23:08.569649
5786	1-5786	15	29	5915	83	6000	0	0	2012-08-02 05:21:16.93694	2012-08-02 05:23:09.354225
4167	1-4167	100	45	4989	50	8796	0	0	2012-08-02 05:10:28.291593	2012-08-02 05:23:09.763065
3808	1-3808	95	81	9874	21	6561	0	0	2012-08-02 05:08:04.772058	2012-08-02 05:23:10.16052
5600	1-5600	95	25	5158	91	5202	0	3	2012-08-02 05:20:01.12449	2012-08-02 05:23:11.762746
6066	1-6066	19	37	3719	25	5478	0	0	2012-08-02 05:23:11.762746	2012-08-02 05:23:11.762746
6067	1-6067	67	66	712	71	7471	0	0	2012-08-02 05:23:12.171373	2012-08-02 05:23:12.171373
5890	1-5890	9	16	3533	79	5051	0	0	2012-08-02 05:21:58.900295	2012-08-02 05:23:14.126826
5572	1-5572	5	26	1932	13	9180	0	0	2012-08-02 05:19:49.953787	2012-08-02 05:23:14.126826
5531	1-5531	100	95	5312	43	7658	0	0	2012-08-02 05:19:33.622378	2012-08-02 05:23:14.126826
5938	1-5938	84	52	4158	22	6389	0	2	2012-08-02 05:22:18.679652	2012-08-02 05:23:14.535686
5430	1-5430	90	2	9371	71	6430	0	7	2012-08-02 05:18:53.339494	2012-08-02 05:23:14.535686
6073	1-6073	54	51	2153	52	5318	0	0	2012-08-02 05:23:14.535686	2012-08-02 05:23:14.535686
5641	1-5641	32	87	2830	63	5696	0	0	2012-08-02 05:20:17.777973	2012-08-02 05:23:15.344229
5507	1-5507	83	77	6169	49	9654	0	0	2012-08-02 05:19:23.999593	2012-08-02 05:23:16.15105
5955	1-5955	80	25	3813	98	4016	0	0	2012-08-02 05:22:25.764045	2012-08-02 05:23:17.333397
6080	1-6080	23	87	2031	25	4735	0	0	2012-08-02 05:23:17.333397	2012-08-02 05:23:17.333397
5334	1-5334	3	28	4644	63	5481	0	0	2012-08-02 05:18:15.636066	2012-08-02 05:23:18.924053
5930	1-5930	51	81	1544	59	3888	0	0	2012-08-02 05:22:15.321079	2012-08-02 05:23:18.924053
6084	1-6084	34	63	585	97	7080	0	0	2012-08-02 05:23:18.924053	2012-08-02 05:23:18.924053
3683	1-3683	50	26	4834	5	3495	0	10	2012-08-02 05:07:15.439761	2012-08-02 05:23:19.333043
6088	1-6088	50	49	425	58	3340	0	2	2012-08-02 05:23:20.526512	2012-08-02 05:23:20.526512
3603	1-3603	80	78	6348	48	2777	0	0	2012-08-02 05:06:43.33001	2012-08-02 05:23:21.741402
6091	1-6091	85	72	479	78	9328	0	0	2012-08-02 05:23:21.741402	2012-08-02 05:23:21.741402
5815	1-5815	82	83	2618	50	9110	0	2	2012-08-02 05:21:28.562718	2012-08-02 05:23:22.150458
6092	1-6092	96	60	1636	95	3112	0	5	2012-08-02 05:23:22.150458	2012-08-02 05:23:22.150458
5911	1-5911	94	27	5873	68	5738	0	3	2012-08-02 05:22:07.43267	2012-08-02 05:23:22.968002
4462	1-4462	8	26	806	37	1996	0	0	2012-08-02 05:12:26.677413	2012-08-02 05:23:22.968002
5564	1-5564	36	85	8931	66	4282	0	0	2012-08-02 05:19:46.761113	2012-08-02 05:23:24.912574
5704	1-5704	74	66	5177	17	8951	0	0	2012-08-02 05:20:43.204373	2012-08-02 05:23:24.912574
6056	1-6056	10	92	5107	50	4921	0	15	2012-08-02 05:23:07.752164	2012-08-02 05:23:24.912574
5036	1-5036	61	45	8495	82	4143	0	0	2012-08-02 05:16:16.962805	2012-08-02 05:23:25.310338
5632	1-5632	68	49	6990	33	6443	0	5	2012-08-02 05:20:14.176531	2012-08-02 05:23:25.310338
5473	1-5473	9	99	4741	3	9256	0	10	2012-08-02 05:19:10.285444	2012-08-02 05:23:25.310338
5821	1-5821	38	11	5891	52	6308	0	7	2012-08-02 05:21:30.993237	2012-08-02 05:23:29.321029
5920	1-5920	58	78	5838	9	3707	0	0	2012-08-02 05:22:11.177736	2012-08-02 05:23:32.1491
5975	1-5975	93	40	5988	16	4151	0	0	2012-08-02 05:22:34.028107	2012-08-02 05:23:32.1491
6118	1-6118	90	85	853	15	9177	0	0	2012-08-02 05:23:32.546825	2012-08-02 05:23:32.546825
1409	1-1409	10	38	2778	17	4309	0	0	2012-08-02 04:50:50.048366	2012-08-02 05:23:33.331153
5080	1-5080	99	38	2066	28	2246	0	0	2012-08-02 05:16:34.774455	2012-08-02 05:23:33.331153
5261	1-5261	14	38	4629	23	2243	0	0	2012-08-02 05:17:47.004254	2012-08-02 05:23:33.331153
6099	1-6099	57	17	6946	47	8093	0	0	2012-08-02 05:23:24.912574	2012-08-02 05:23:33.331153
6096	1-6096	45	39	2272	29	1360	0	12	2012-08-02 05:23:23.741436	2012-08-02 05:23:36.115655
6106	1-6106	15	95	2818	39	8819	0	0	2012-08-02 05:23:27.763056	2012-08-02 05:23:36.115655
5639	1-5639	27	19	3731	30	6642	0	7	2012-08-02 05:20:16.982944	2012-08-02 05:23:36.524562
5857	1-5857	51	30	1863	10	4859	0	2	2012-08-02 05:21:45.609331	2012-08-02 05:23:36.524562
6128	1-6128	62	10	1837	19	8594	0	0	2012-08-02 05:23:36.524562	2012-08-02 05:23:36.524562
3421	1-3421	25	19	6288	30	5596	65	0	2012-08-02 05:05:29.794324	2012-08-02 05:23:36.524562
6130	1-6130	85	24	177	64	1781	0	2	2012-08-02 05:23:37.320014	2012-08-02 05:23:37.320014
5278	1-5278	8	4	9305	82	4286	0	4	2012-08-02 05:17:53.65677	2012-08-02 05:23:37.718045
6131	1-6131	14	57	777	4	7059	0	3	2012-08-02 05:23:37.718045	2012-08-02 05:23:37.718045
2412	1-2412	99	93	7699	48	3375	0	2	2012-08-02 04:58:29.810383	2012-08-02 05:23:38.922388
5388	1-5388	26	93	6256	5	5186	0	0	2012-08-02 05:18:36.837921	2012-08-02 05:23:38.922388
6134	1-6134	53	31	3279	93	8437	0	2	2012-08-02 05:23:38.922388	2012-08-02 05:23:38.922388
6074	1-6074	63	17	2743	46	4678	0	0	2012-08-02 05:23:14.957706	2012-08-02 05:23:39.717743
5810	1-5810	88	82	8449	8	7327	0	10	2012-08-02 05:21:26.516077	2012-08-02 05:23:39.717743
5366	1-5366	25	99	5656	6	7515	0	8	2012-08-02 05:18:28.18679	2012-08-02 05:23:40.513518
1515	1-1515	64	54	5621	94	560	0	0	2012-08-02 04:51:46.215951	2012-08-02 05:23:42.139499
6146	1-6146	92	25	1617	90	3943	0	0	2012-08-02 05:23:43.741195	2012-08-02 05:23:43.741195
3390	1-3390	80	89	9779	72	5183	0	4	2012-08-02 05:05:17.364486	2012-08-02 05:23:44.150091
3410	1-3410	28	92	1864	61	327	0	0	2012-08-02 05:05:25.352542	2012-08-02 05:23:44.150091
110	1-110	45	89	4203	72	733	733	0	2012-08-02 04:43:54.716626	2012-08-02 05:23:44.150091
6142	1-6142	32	5	1415	54	8519	0	2	2012-08-02 05:23:42.139499	2012-08-02 05:23:45.740812
5995	1-5995	77	5	6211	61	6532	0	0	2012-08-02 05:22:42.358902	2012-08-02 05:23:45.740812
6152	1-6152	79	49	97	43	2997	0	3	2012-08-02 05:23:46.149732	2012-08-02 05:23:46.149732
6151	1-6151	57	74	3668	77	4609	0	0	2012-08-02 05:23:45.740812	2012-08-02 05:23:47.321085
1989	1-1989	20	5	8496	68	933	0	6	2012-08-02 04:55:26.328478	2012-08-02 05:23:49.321171
6160	1-6160	5	26	461	5	8118	0	16	2012-08-02 05:23:49.321171	2012-08-02 05:23:49.321171
5270	1-5270	92	17	3832	21	2501	0	0	2012-08-02 05:17:50.552195	2012-08-02 05:23:50.13312
3272	1-3272	49	99	9304	21	7507	0	4	2012-08-02 05:04:29.319558	2012-08-02 05:23:50.13312
5434	1-5434	1	28	7601	48	6455	0	0	2012-08-02 05:18:54.947725	2012-08-02 05:23:54.12682
6120	1-6120	54	47	3109	38	9263	0	0	2012-08-02 05:23:33.331153	2012-08-02 05:23:54.12682
6117	1-6117	63	87	4602	71	8457	0	0	2012-08-02 05:23:32.1491	2012-08-02 05:23:55.344443
1463	1-1463	85	95	6556	23	1917	0	0	2012-08-02 04:51:19.194828	2012-08-02 05:23:55.344443
6138	1-6138	14	6	1879	41	6617	0	0	2012-08-02 05:23:40.513518	2012-08-02 05:23:56.164426
6094	1-6094	53	37	6087	27	3068	0	3	2012-08-02 05:23:22.968002	2012-08-02 05:23:56.164426
1154	1-1154	79	38	4936	14	8669	0	0	2012-08-02 04:48:36.361318	2012-08-02 05:23:56.164426
5766	1-5766	3	6	4581	38	7309	0	4	2012-08-02 05:21:08.949598	2012-08-02 05:23:56.164426
4115	1-4115	21	3	7837	47	5809	0	0	2012-08-02 05:10:07.662036	2012-08-02 05:23:56.164426
5967	1-5967	73	14	8972	29	5976	0	9	2012-08-02 05:22:30.73537	2012-08-02 05:23:56.164426
6178	1-6178	60	2	1046	96	9443	0	0	2012-08-02 05:23:56.599925	2012-08-02 05:23:56.599925
2729	1-2729	71	68	7292	61	3286	0	0	2012-08-02 05:00:46.555717	2012-08-02 05:23:58.986599
6184	1-6184	70	86	2041	67	7558	0	8	2012-08-02 05:23:58.986599	2012-08-02 05:23:58.986599
6062	1-6062	85	56	909	34	4308	0	0	2012-08-02 05:23:10.16052	2012-08-02 05:23:59.429971
5659	1-5659	69	88	7816	49	4876	0	0	2012-08-02 05:20:25.026321	2012-08-02 05:23:59.817307
6186	1-6186	58	49	3524	88	8866	0	11	2012-08-02 05:23:59.817307	2012-08-02 05:23:59.817307
6100	1-6100	69	8	6857	99	8893	0	0	2012-08-02 05:23:25.310338	2012-08-02 05:24:00.215601
4236	1-4236	86	86	5899	72	5717	0	0	2012-08-02 05:10:55.362492	2012-08-02 05:24:00.215601
6147	1-6147	30	72	1264	95	8384	0	7	2012-08-02 05:23:44.150091	2012-08-02 05:24:00.215601
6189	1-6189	5	74	4656	2	2466	0	0	2012-08-02 05:24:01.188856	2012-08-02 05:24:01.188856
5620	1-5620	69	59	9338	28	9277	0	4	2012-08-02 05:20:09.237561	2012-08-02 05:24:01.586912
6114	1-6114	42	20	2415	74	9387	0	0	2012-08-02 05:23:30.9561	2012-08-02 05:24:01.586912
6190	1-6190	29	46	189	59	5690	0	0	2012-08-02 05:24:01.586912	2012-08-02 05:24:01.586912
5000	1-5000	23	21	8879	68	3040	0	0	2012-08-02 05:16:02.397007	2012-08-02 05:24:02.383118
6025	1-6025	27	96	7709	82	7262	0	4	2012-08-02 05:22:54.912575	2012-08-02 05:24:02.383118
5589	1-5589	28	91	5960	9	6239	0	0	2012-08-02 05:19:56.704835	2012-08-02 05:24:05.194268
6162	1-6162	60	82	9531	8	9113	0	14	2012-08-02 05:23:50.13312	2012-08-02 05:24:05.194268
5814	1-5814	3	82	2525	49	4342	0	0	2012-08-02 05:21:28.164912	2012-08-02 05:24:05.194268
5006	1-5006	59	100	6087	79	2597	0	0	2012-08-02 05:16:04.763589	2012-08-02 05:24:05.194268
3895	1-3895	54	24	4932	20	6767	0	0	2012-08-02 05:08:39.734563	2012-08-02 05:24:05.703273
6200	1-6200	46	97	173	24	9904	0	2	2012-08-02 05:24:05.703273	2012-08-02 05:24:05.703273
6201	1-6201	76	45	4354	97	4897	0	12	2012-08-02 05:24:06.104698	2012-08-02 05:24:06.104698
5450	1-5450	40	67	3786	14	9024	0	0	2012-08-02 05:19:01.249943	2012-08-02 05:24:06.53598
2863	1-2863	50	77	7359	14	7151	0	0	2012-08-02 05:01:42.833567	2012-08-02 05:24:06.53598
5470	1-5470	74	61	5945	68	9102	0	9	2012-08-02 05:19:09.113666	2012-08-02 05:24:09.806612
5884	1-5884	46	36	2509	28	8642	0	0	2012-08-02 05:21:56.503007	2012-08-02 05:24:11.883834
2529	1-2529	40	21	9952	23	4197	0	0	2012-08-02 04:59:21.165804	2012-08-02 05:24:11.883834
5888	1-5888	50	10	5141	65	5357	0	0	2012-08-02 05:21:58.115984	2012-08-02 05:24:12.712933
6217	1-6217	78	65	2078	1	4798	0	7	2012-08-02 05:24:12.712933	2012-08-02 05:24:12.712933
6177	1-6177	38	27	4741	6	6610	0	2	2012-08-02 05:23:56.164426	2012-08-02 05:24:13.133182
4455	1-4455	21	97	3004	57	2550	0	7	2012-08-02 05:12:23.924392	2012-08-02 05:24:13.133182
4553	1-4553	49	29	9083	97	5945	0	7	2012-08-02 05:13:02.820026	2012-08-02 05:24:13.133182
3050	1-3050	6	29	9323	36	2749	0	0	2012-08-02 05:02:59.480683	2012-08-02 05:24:13.133182
6218	1-6218	92	77	1348	29	6655	0	0	2012-08-02 05:24:13.133182	2012-08-02 05:24:13.133182
3611	1-3611	26	81	3935	57	1008	0	0	2012-08-02 05:06:46.512274	2012-08-02 05:24:13.984491
5822	1-5822	7	81	1994	92	7668	0	0	2012-08-02 05:21:31.390735	2012-08-02 05:24:13.984491
5705	1-5705	94	50	7124	60	6665	0	7	2012-08-02 05:20:43.602002	2012-08-02 05:24:15.279394
6168	1-6168	34	93	3342	87	6486	0	4	2012-08-02 05:23:52.546874	2012-08-02 05:24:15.279394
5993	1-5993	47	92	1844	87	526	0	10	2012-08-02 05:22:41.519152	2012-08-02 05:24:16.108728
6165	1-6165	5	80	1731	92	6615	0	2	2012-08-02 05:23:51.364873	2012-08-02 05:24:16.108728
6009	1-6009	96	80	1897	42	2070	0	3	2012-08-02 05:22:48.192672	2012-08-02 05:24:16.108728
6226	1-6226	27	21	393	34	9385	0	7	2012-08-02 05:24:16.517329	2012-08-02 05:24:16.517329
6127	1-6127	35	29	4075	95	9647	0	0	2012-08-02 05:23:36.115655	2012-08-02 05:24:16.926451
5832	1-5832	88	85	2254	15	2159	0	0	2012-08-02 05:21:35.401522	2012-08-02 05:24:16.926451
6215	1-6215	47	28	909	44	3587	0	0	2012-08-02 05:24:11.883834	2012-08-02 05:24:17.335656
3092	1-3092	31	26	7834	39	5646	0	0	2012-08-02 05:03:16.575918	2012-08-02 05:24:17.755998
3788	1-3788	97	26	8613	18	3181	0	0	2012-08-02 05:07:56.916631	2012-08-02 05:24:17.755998
6229	1-6229	72	93	6718	26	9155	0	0	2012-08-02 05:24:17.755998	2012-08-02 05:24:17.755998
6230	1-6230	66	48	516	43	3829	0	6	2012-08-02 05:24:18.187209	2012-08-02 05:24:18.187209
6231	1-6231	24	72	920	68	5669	0	12	2012-08-02 05:24:18.60721	2012-08-02 05:24:18.60721
5418	1-5418	12	95	4106	83	2853	0	0	2012-08-02 05:18:48.58443	2012-08-02 05:24:19.457921
6108	1-6108	48	11	5051	45	5275	0	14	2012-08-02 05:23:28.536511	2012-08-02 05:24:19.457921
5728	1-5728	90	11	5386	55	5129	0	13	2012-08-02 05:20:52.904843	2012-08-02 05:24:19.457921
5529	1-5529	44	85	6335	44	9028	0	0	2012-08-02 05:19:32.826978	2012-08-02 05:24:19.877509
6234	1-6234	80	15	2178	85	6481	0	0	2012-08-02 05:24:19.877509	2012-08-02 05:24:19.877509
5485	1-5485	92	1	3835	8	1368	0	6	2012-08-02 05:19:15.069933	2012-08-02 05:24:20.717637
5180	1-5180	6	41	6646	67	9289	0	0	2012-08-02 05:17:15.204681	2012-08-02 05:24:23.180936
5742	1-5742	21	90	8541	67	6091	0	3	2012-08-02 05:20:58.903764	2012-08-02 05:24:23.180936
6242	1-6242	80	67	5591	62	6371	0	0	2012-08-02 05:24:23.180936	2012-08-02 05:24:24.009581
5318	1-5318	20	99	2204	17	9024	0	0	2012-08-02 05:18:09.325357	2012-08-02 05:24:24.009581
6136	1-6136	87	8	2195	17	6191	0	2	2012-08-02 05:23:39.717743	2012-08-02 05:24:24.009581
4324	1-4324	59	67	6139	68	3040	0	0	2012-08-02 05:11:30.975528	2012-08-02 05:24:24.009581
6244	1-6244	8	17	1011	67	7807	0	0	2012-08-02 05:24:24.009581	2012-08-02 05:24:24.009581
2494	1-2494	44	67	8594	15	7091	182	0	2012-08-02 04:59:06.040666	2012-08-02 05:24:24.009581
5976	1-5976	28	46	9882	1	8930	0	9	2012-08-02 05:22:34.447693	2012-08-02 05:24:24.871937
4824	1-4824	30	20	7071	61	3851	0	11	2012-08-02 05:14:52.942145	2012-08-02 05:24:25.280997
5907	1-5907	65	21	3042	96	3276	0	3	2012-08-02 05:22:05.786634	2012-08-02 05:24:25.280997
6247	1-6247	66	61	411	24	2666	0	13	2012-08-02 05:24:25.280997	2012-08-02 05:24:25.280997
6172	1-6172	6	48	3775	3	7075	0	0	2012-08-02 05:23:54.12682	2012-08-02 05:24:26.938157
3575	1-3575	57	21	7146	2	4320	0	0	2012-08-02 05:06:32.004275	2012-08-02 05:24:26.938157
6251	1-6251	54	10	1050	48	5750	0	7	2012-08-02 05:24:26.938157	2012-08-02 05:24:26.938157
6175	1-6175	66	71	2058	4	6949	0	0	2012-08-02 05:23:55.344443	2012-08-02 05:24:27.766833
6187	1-6187	56	99	2224	61	7088	0	0	2012-08-02 05:24:00.215601	2012-08-02 05:24:28.175566
6253	1-6253	1	49	3586	71	3577	0	5	2012-08-02 05:24:27.766833	2012-08-02 05:24:28.175566
6254	1-6254	66	50	3713	99	6203	0	0	2012-08-02 05:24:28.175566	2012-08-02 05:24:28.175566
6233	1-6233	3	89	796	7	4020	0	0	2012-08-02 05:24:19.457921	2012-08-02 05:24:28.606795
6192	1-6192	39	89	2422	21	8295	0	7	2012-08-02 05:24:02.383118	2012-08-02 05:24:28.606795
5908	1-5908	64	89	3302	92	7364	0	0	2012-08-02 05:22:06.195789	2012-08-02 05:24:28.606795
6255	1-6255	4	95	1865	89	6491	0	3	2012-08-02 05:24:28.606795	2012-08-02 05:24:28.606795
6266	1-6266	27	51	1371	85	5150	0	0	2012-08-02 05:24:33.202506	2012-08-02 05:24:33.202506
6166	1-6166	7	2	5499	10	8675	0	0	2012-08-02 05:23:51.751315	2012-08-02 05:24:35.29057
6185	1-6185	3	14	9579	56	4458	0	0	2012-08-02 05:23:59.429971	2012-08-02 05:24:36.848189
6277	1-6277	55	5	80	36	8622	0	2	2012-08-02 05:24:38.120899	2012-08-02 05:24:38.120899
5855	1-5855	92	77	3829	5	3798	0	0	2012-08-02 05:21:44.802846	2012-08-02 05:24:39.413539
6202	1-6202	85	57	5611	77	9006	0	3	2012-08-02 05:24:06.53598	2012-08-02 05:24:39.413539
5791	1-5791	27	57	7699	75	2341	0	0	2012-08-02 05:21:18.915135	2012-08-02 05:24:39.413539
6280	1-6280	8	81	3338	57	6873	0	2	2012-08-02 05:24:39.413539	2012-08-02 05:24:39.413539
6237	1-6237	38	53	1362	81	2722	0	0	2012-08-02 05:24:21.115125	2012-08-02 05:24:40.253037
2527	1-2527	93	53	7839	39	1306	0	0	2012-08-02 04:59:20.304147	2012-08-02 05:24:40.253037
5259	1-5259	1	36	1608	37	7382	0	0	2012-08-02 05:17:46.220065	2012-08-02 05:24:41.932477
5085	1-5085	13	36	6827	4	4143	0	0	2012-08-02 05:16:36.862576	2012-08-02 05:24:41.932477
5354	1-5354	10	34	9625	4	7801	0	0	2012-08-02 05:18:23.502386	2012-08-02 05:24:41.932477
4289	1-4289	27	88	7925	74	3838	0	0	2012-08-02 05:11:16.556478	2012-08-02 05:24:41.932477
5556	1-5556	14	8	2553	59	3122	0	0	2012-08-02 05:19:43.521881	2012-08-02 05:24:41.932477
6095	1-6095	92	36	2284	93	1606	0	0	2012-08-02 05:23:23.354738	2012-08-02 05:24:41.932477
5774	1-5774	95	98	4394	16	2298	0	0	2012-08-02 05:21:12.13134	2012-08-02 05:24:42.772559
3237	1-3237	14	98	4225	22	2600	0	9	2012-08-02 05:04:15.573354	2012-08-02 05:24:42.772559
6037	1-6037	65	85	9207	22	4789	0	0	2012-08-02 05:22:59.905997	2012-08-02 05:24:42.772559
6288	1-6288	44	43	912	98	7381	0	0	2012-08-02 05:24:42.772559	2012-08-02 05:24:42.772559
6292	1-6292	12	16	1579	86	3906	0	3	2012-08-02 05:24:44.463222	2012-08-02 05:24:44.463222
6293	1-6293	96	58	4440	8	6323	0	0	2012-08-02 05:24:44.872395	2012-08-02 05:24:44.872395
5859	1-5859	96	94	7003	3	9966	0	3	2012-08-02 05:21:46.426817	2012-08-02 05:24:46.926828
6077	1-6077	58	48	3416	77	3932	0	0	2012-08-02 05:23:16.15105	2012-08-02 05:24:46.926828
4104	1-4104	55	94	8401	95	6427	0	0	2012-08-02 05:10:03.273669	2012-08-02 05:24:46.926828
6298	1-6298	16	99	1000	94	9120	0	4	2012-08-02 05:24:46.926828	2012-08-02 05:24:46.926828
5833	1-5833	23	20	5624	95	4087	0	0	2012-08-02 05:21:35.81046	2012-08-02 05:24:47.755777
5785	1-5785	1	54	9455	17	8236	0	7	2012-08-02 05:21:16.539619	2012-08-02 05:24:47.755777
6300	1-6300	92	2	2436	20	5935	0	8	2012-08-02 05:24:47.755777	2012-08-02 05:24:47.755777
3995	1-3995	93	88	3837	7	391	0	0	2012-08-02 05:09:19.650421	2012-08-02 05:24:49.391277
5566	1-5566	31	22	5418	49	4947	0	10	2012-08-02 05:19:47.54557	2012-08-02 05:24:51.898893
6286	1-6286	37	22	3291	33	9656	0	11	2012-08-02 05:24:41.932477	2012-08-02 05:24:51.898893
6153	1-6153	26	2	2029	41	1885	0	0	2012-08-02 05:23:46.54759	2012-08-02 05:24:53.975941
6315	1-6315	51	14	801	2	3096	0	8	2012-08-02 05:24:53.975941	2012-08-02 05:24:53.975941
5989	1-5989	80	16	5080	87	9028	0	7	2012-08-02 05:22:39.8396	2012-08-02 05:24:54.395896
5136	1-5136	30	16	7791	43	4920	0	8	2012-08-02 05:16:57.602174	2012-08-02 05:24:54.395896
6316	1-6316	25	88	321	16	4033	0	0	2012-08-02 05:24:54.395896	2012-08-02 05:24:54.395896
5605	1-5605	96	39	9342	60	9319	0	0	2012-08-02 05:20:03.124792	2012-08-02 05:24:55.677723
3342	1-3342	93	18	6311	41	6004	0	0	2012-08-02 05:04:57.628021	2012-08-02 05:24:56.108512
6320	1-6320	14	36	2215	18	3275	0	7	2012-08-02 05:24:56.108512	2012-08-02 05:24:56.108512
5759	1-5759	78	18	3383	90	6086	0	0	2012-08-02 05:21:06.176608	2012-08-02 05:24:56.528368
5306	1-5306	43	63	2791	9	622	0	0	2012-08-02 05:18:04.605998	2012-08-02 05:24:56.528368
5750	1-5750	33	10	2943	44	4697	0	0	2012-08-02 05:21:02.261992	2012-08-02 05:24:58.583216
6326	1-6326	3	44	1440	93	3566	0	3	2012-08-02 05:24:58.583216	2012-08-02 05:24:58.583216
6199	1-6199	49	40	7372	82	8349	0	0	2012-08-02 05:24:05.194268	2012-08-02 05:24:59.003113
6235	1-6235	98	73	5702	95	6749	0	0	2012-08-02 05:24:20.286595	2012-08-02 05:24:59.85385
5877	1-5877	7	1	2896	50	4491	0	0	2012-08-02 05:21:53.685676	2012-08-02 05:24:59.85385
5982	1-5982	79	15	3814	51	4801	0	0	2012-08-02 05:22:36.922578	2012-08-02 05:25:00.693959
4951	1-4951	72	24	5111	32	4772	0	0	2012-08-02 05:15:43.074054	2012-08-02 05:25:01.113546
6258	1-6258	70	33	4375	36	8190	0	0	2012-08-02 05:24:29.899308	2012-08-02 05:25:01.113546
5039	1-5039	23	72	5903	68	5371	0	5	2012-08-02 05:16:18.155912	2012-08-02 05:25:01.113546
6081	1-6081	61	3	9395	96	6590	0	0	2012-08-02 05:23:17.730982	2012-08-02 05:25:02.373071
1389	1-1389	2	82	6310	46	3425	0	0	2012-08-02 04:50:39.642486	2012-08-02 05:25:02.373071
6327	1-6327	33	82	4425	3	3485	0	0	2012-08-02 05:24:59.003113	2012-08-02 05:25:02.373071
6335	1-6335	77	96	2235	82	7286	0	10	2012-08-02 05:25:02.373071	2012-08-02 05:25:02.373071
4297	1-4297	85	9	3865	95	4835	0	0	2012-08-02 05:11:19.849138	2012-08-02 05:25:03.645736
5629	1-5629	25	94	4706	1	2287	0	9	2012-08-02 05:20:12.938805	2012-08-02 05:25:03.645736
5042	1-5042	80	29	8538	27	6232	0	0	2012-08-02 05:16:19.371326	2012-08-02 05:25:03.645736
5951	1-5951	34	17	4913	49	2962	0	0	2012-08-02 05:22:24.104661	2012-08-02 05:25:03.645736
6270	1-6270	13	86	3462	65	1830	0	5	2012-08-02 05:24:34.870747	2012-08-02 05:25:04.507616
211	1-211	17	26	9450	86	2745	712	0	2012-08-02 04:43:56.222929	2012-08-02 05:25:04.507616
6341	1-6341	86	31	224	84	4109	0	0	2012-08-02 05:25:04.927538	2012-08-02 05:25:04.927538
6075	1-6075	76	9	2889	90	9396	0	0	2012-08-02 05:23:15.344229	2012-08-02 05:25:05.359362
6342	1-6342	99	87	516	9	3766	0	0	2012-08-02 05:25:05.359362	2012-08-02 05:25:05.359362
5703	1-5703	98	17	8567	45	8755	0	0	2012-08-02 05:20:42.817279	2012-08-02 05:25:10.210421
5881	1-5881	11	13	1482	17	1437	0	0	2012-08-02 05:21:55.298627	2012-08-02 05:25:10.210421
5965	1-5965	74	31	7686	87	3391	0	5	2012-08-02 05:22:29.91754	2012-08-02 05:25:10.210421
6354	1-6354	17	44	1931	13	9218	0	0	2012-08-02 05:25:10.210421	2012-08-02 05:25:10.210421
6044	1-6044	73	23	5688	71	9602	0	9	2012-08-02 05:23:02.878025	2012-08-02 05:25:11.525567
3038	1-3038	84	23	7798	12	2988	0	2	2012-08-02 05:02:54.652445	2012-08-02 05:25:11.525567
5365	1-5365	54	59	8700	4	8327	0	0	2012-08-02 05:18:27.789149	2012-08-02 05:25:12.33188
5187	1-5187	29	74	6621	22	7267	0	0	2012-08-02 05:17:17.955619	2012-08-02 05:25:12.33188
6359	1-6359	60	22	2804	60	8987	0	11	2012-08-02 05:25:12.33188	2012-08-02 05:25:12.33188
6145	1-6145	67	88	1518	15	3171	0	0	2012-08-02 05:23:43.343253	2012-08-02 05:25:13.547315
4231	1-4231	4	5	3789	69	1590	0	0	2012-08-02 05:10:53.429121	2012-08-02 05:25:15.547029
6367	1-6367	45	69	375	75	7683	0	5	2012-08-02 05:25:15.547029	2012-08-02 05:25:15.547029
4982	1-4982	60	85	6446	37	7978	0	12	2012-08-02 05:15:55.215825	2012-08-02 05:25:16.353446
1417	1-1417	93	98	3601	51	2750	0	0	2012-08-02 04:50:54.334529	2012-08-02 05:25:16.353446
6345	1-6345	62	98	1691	49	710	0	0	2012-08-02 05:25:06.564908	2012-08-02 05:25:16.353446
5988	1-5988	15	63	2813	88	8780	0	0	2012-08-02 05:22:39.430825	2012-08-02 05:25:19.270626
6350	1-6350	26	92	1056	76	948	0	0	2012-08-02 05:25:08.564242	2012-08-02 05:25:19.270626
6053	1-6053	5	80	1968	92	3300	0	0	2012-08-02 05:23:06.569593	2012-08-02 05:25:19.270626
6363	1-6363	12	44	2382	64	1660	0	7	2012-08-02 05:25:13.93441	2012-08-02 05:25:19.270626
6083	1-6083	65	51	4275	37	9918	0	2	2012-08-02 05:23:18.537325	2012-08-02 05:25:20.507878
6220	1-6220	67	5	2402	81	7398	0	10	2012-08-02 05:24:13.984491	2012-08-02 05:25:23.678507
6369	1-6369	32	95	3828	98	7686	0	0	2012-08-02 05:25:16.353446	2012-08-02 05:25:23.678507
2413	1-2413	50	73	6947	50	3773	0	0	2012-08-02 04:58:30.241319	2012-08-02 05:25:25.269635
6246	1-6246	68	1	4330	27	9224	0	11	2012-08-02 05:24:24.871937	2012-08-02 05:25:26.91776
6310	1-6310	66	95	6440	22	9955	0	0	2012-08-02 05:24:51.898893	2012-08-02 05:25:27.315553
5250	1-5250	48	5	4989	22	2991	0	0	2012-08-02 05:17:42.695287	2012-08-02 05:25:27.315553
5942	1-5942	86	94	1595	76	1611	0	0	2012-08-02 05:22:20.336817	2012-08-02 05:25:27.713289
5592	1-5592	100	6	3267	57	9224	0	9	2012-08-02 05:19:57.909205	2012-08-02 05:25:28.111411
5406	1-5406	15	3	4780	90	2502	0	0	2012-08-02 05:18:43.899597	2012-08-02 05:25:28.111411
6398	1-6398	76	57	1176	3	5724	0	5	2012-08-02 05:25:28.111411	2012-08-02 05:25:28.111411
5285	1-5285	29	1	4939	27	4119	0	7	2012-08-02 05:17:56.396849	2012-08-02 05:25:29.735097
6271	1-6271	74	95	8127	2	8402	0	0	2012-08-02 05:24:35.29057	2012-08-02 05:25:31.337389
6334	1-6334	89	95	1842	49	311	0	0	2012-08-02 05:25:01.95336	2012-08-02 05:25:31.337389
2045	1-2045	31	95	7274	99	625	0	0	2012-08-02 04:55:50.48115	2012-08-02 05:25:31.337389
6406	1-6406	92	1	657	95	8140	0	0	2012-08-02 05:25:31.337389	2012-08-02 05:25:31.337389
5189	1-5189	42	58	8634	28	6013	0	0	2012-08-02 05:17:18.717927	2012-08-02 05:25:32.563533
3902	1-3902	96	56	6038	57	8501	0	0	2012-08-02 05:08:42.562886	2012-08-02 05:25:32.563533
5162	1-5162	89	38	8672	36	8041	0	0	2012-08-02 05:17:08.143815	2012-08-02 05:25:32.563533
1917	1-1917	15	24	4799	57	4259	0	0	2012-08-02 04:54:54.827728	2012-08-02 05:25:32.563533
6409	1-6409	75	57	4713	38	7756	0	7	2012-08-02 05:25:32.563533	2012-08-02 05:25:32.563533
3021	1-3021	39	38	5536	37	7708	0	11	2012-08-02 05:02:47.757932	2012-08-02 05:25:33.370486
4590	1-4590	37	83	3506	7	812	0	0	2012-08-02 05:13:17.836446	2012-08-02 05:25:35.359137
6338	1-6338	92	27	4776	9	8096	0	0	2012-08-02 05:25:03.645736	2012-08-02 05:25:35.779043
6400	1-6400	95	35	1024	61	1463	0	2	2012-08-02 05:25:28.917566	2012-08-02 05:25:35.779043
5745	1-5745	69	36	2541	42	3202	0	0	2012-08-02 05:21:00.151943	2012-08-02 05:25:36.187976
6420	1-6420	92	10	5545	38	1162	0	5	2012-08-02 05:25:36.972537	2012-08-02 05:25:36.972537
6421	1-6421	61	18	74	61	137	0	0	2012-08-02 05:25:37.414738	2012-08-02 05:25:37.414738
6426	1-6426	75	84	3345	91	1949	0	3	2012-08-02 05:25:39.568559	2012-08-02 05:25:39.568559
6428	1-6428	19	37	288	63	2245	0	0	2012-08-02 05:25:40.364011	2012-08-02 05:25:40.364011
6376	1-6376	53	98	4124	80	8371	0	4	2012-08-02 05:25:19.270626	2012-08-02 05:25:40.761748
6429	1-6429	1	80	2111	5	3689	0	0	2012-08-02 05:25:40.761748	2012-08-02 05:25:40.761748
5401	1-5401	97	91	9981	8	8627	0	3	2012-08-02 05:18:41.955176	2012-08-02 05:25:41.159644
3793	1-3793	42	44	4987	21	1677	0	0	2012-08-02 05:07:58.894426	2012-08-02 05:25:41.159644
6430	1-6430	23	1	2326	91	6090	0	12	2012-08-02 05:25:41.159644	2012-08-02 05:25:41.159644
3765	1-3765	4	6	9726	56	9115	0	0	2012-08-02 05:07:47.911969	2012-08-02 05:25:41.568516
6089	1-6089	82	4	4689	90	2416	0	12	2012-08-02 05:23:20.957358	2012-08-02 05:25:41.568516
6183	1-6183	26	91	2985	11	8318	0	3	2012-08-02 05:23:58.600074	2012-08-02 05:25:41.977529
4389	1-4389	60	13	9568	21	5772	0	0	2012-08-02 05:11:57.669899	2012-08-02 05:25:41.977529
6311	1-6311	47	74	2560	50	2824	0	0	2012-08-02 05:24:52.329787	2012-08-02 05:25:42.386378
6357	1-6357	58	27	5636	23	7461	0	3	2012-08-02 05:25:11.525567	2012-08-02 05:25:47.180911
5554	1-5554	36	78	9646	23	7143	0	0	2012-08-02 05:19:42.715469	2012-08-02 05:25:47.180911
5787	1-5787	55	17	3192	87	2103	0	6	2012-08-02 05:21:17.323808	2012-08-02 05:25:47.987605
6447	1-6447	94	87	1188	96	4460	0	0	2012-08-02 05:25:47.987605	2012-08-02 05:25:47.987605
6397	1-6397	37	58	5229	17	7751	0	0	2012-08-02 05:25:27.713289	2012-08-02 05:25:50.760794
6454	1-6454	98	17	1799	38	7766	0	0	2012-08-02 05:25:50.760794	2012-08-02 05:25:50.760794
6458	1-6458	4	7	1444	59	3178	0	3	2012-08-02 05:25:52.329491	2012-08-02 05:25:52.329491
4060	1-4060	80	80	6087	14	4779	0	0	2012-08-02 05:09:45.759571	2012-08-02 05:25:53.547147
6466	1-6466	38	15	2707	59	2872	0	0	2012-08-02 05:25:55.535426	2012-08-02 05:25:55.535426
5889	1-5889	5	41	5100	37	8174	0	0	2012-08-02 05:21:58.513644	2012-08-02 05:25:56.308787
6236	1-6236	84	34	1288	46	1873	0	0	2012-08-02 05:24:20.717637	2012-08-02 05:25:56.308787
6468	1-6468	4	46	1616	41	4262	0	0	2012-08-02 05:25:56.308787	2012-08-02 05:25:56.308787
5130	1-5130	51	55	1838	14	4373	0	0	2012-08-02 05:16:55.158981	2012-08-02 05:25:57.082457
6470	1-6470	89	28	169	55	4603	0	4	2012-08-02 05:25:57.082457	2012-08-02 05:25:57.082457
5234	1-5234	90	38	5339	58	9623	0	11	2012-08-02 05:17:36.420454	2012-08-02 05:25:59.46881
3121	1-3121	67	91	5991	39	7256	0	0	2012-08-02 05:03:28.422134	2012-08-02 05:25:59.46881
516	1-516	37	11	7612	46	2500	0	0	2012-08-02 04:44:14.498811	2012-08-02 05:25:59.46881
6135	1-6135	53	54	1874	70	861	0	0	2012-08-02 05:23:39.331108	2012-08-02 05:26:00.242302
2485	1-2485	59	30	5254	73	6707	0	0	2012-08-02 04:59:02.123301	2012-08-02 05:26:00.242302
4199	1-4199	71	41	9029	52	5589	0	2	2012-08-02 05:10:40.866851	2012-08-02 05:26:01.435501
5776	1-5776	85	75	2902	18	3214	0	11	2012-08-02 05:21:12.915834	2012-08-02 05:26:01.435501
6481	1-6481	27	100	3140	41	6119	0	0	2012-08-02 05:26:01.435501	2012-08-02 05:26:01.435501
5131	1-5131	73	7	5684	58	9189	0	4	2012-08-02 05:16:55.580936	2012-08-02 05:26:01.855646
6304	1-6304	11	7	2898	66	7208	0	0	2012-08-02 05:24:49.391277	2012-08-02 05:26:01.855646
6482	1-6482	35	15	1930	7	4379	0	0	2012-08-02 05:26:01.855646	2012-08-02 05:26:01.855646
6483	1-6483	37	87	92	71	6165	0	4	2012-08-02 05:26:02.242599	2012-08-02 05:26:02.242599
5959	1-5959	26	2	9120	20	7828	0	4	2012-08-02 05:22:27.453947	2012-08-02 05:26:03.446696
6486	1-6486	88	77	1779	86	4999	0	0	2012-08-02 05:26:03.446696	2012-08-02 05:26:03.446696
6478	1-6478	21	98	3133	97	5454	0	0	2012-08-02 05:26:00.242302	2012-08-02 05:26:05.037834
6490	1-6490	91	69	703	98	8777	0	0	2012-08-02 05:26:05.037834	2012-08-02 05:26:05.037834
2431	1-2431	11	98	6051	61	1327	0	0	2012-08-02 04:58:38.196056	2012-08-02 05:26:08.26521
6240	1-6240	23	13	7483	96	3095	0	11	2012-08-02 05:24:22.35241	2012-08-02 05:26:08.26521
6498	1-6498	30	78	661	98	9423	0	6	2012-08-02 05:26:08.26521	2012-08-02 05:26:08.26521
5499	1-5499	80	11	5994	13	3959	0	0	2012-08-02 05:19:20.818857	2012-08-02 05:26:09.49156
6475	1-6475	17	97	1158	5	9742	0	7	2012-08-02 05:25:59.071026	2012-08-02 05:26:09.49156
4900	1-4900	84	7	7956	49	5814	0	7	2012-08-02 05:15:23.162683	2012-08-02 05:26:09.49156
6223	1-6223	38	90	4992	93	9847	0	4	2012-08-02 05:24:15.279394	2012-08-02 05:26:09.911619
6197	1-6197	96	71	7740	24	9710	0	6	2012-08-02 05:24:04.37261	2012-08-02 05:26:09.911619
6502	1-6502	51	24	621	90	3035	0	0	2012-08-02 05:26:09.911619	2012-08-02 05:26:09.911619
6504	1-6504	71	18	471	27	4662	0	19	2012-08-02 05:26:10.740538	2012-08-02 05:26:10.740538
6506	1-6506	64	54	902	75	2287	0	0	2012-08-02 05:26:11.580013	2012-08-02 05:26:11.580013
6051	1-6051	34	4	9969	22	8843	0	8	2012-08-02 05:23:05.762544	2012-08-02 05:26:12.375431
2482	1-2482	63	86	9430	60	5081	0	7	2012-08-02 04:59:00.808575	2012-08-02 05:26:12.375431
6508	1-6508	80	73	2343	96	6052	0	0	2012-08-02 05:26:12.375431	2012-08-02 05:26:12.375431
6210	1-6210	1	32	1775	20	5028	0	0	2012-08-02 05:24:09.806612	2012-08-02 05:26:12.795513
6510	1-6510	67	49	1620	30	2327	0	0	2012-08-02 05:26:13.20429	2012-08-02 05:26:13.20429
5996	1-5996	1	91	2481	11	6350	0	0	2012-08-02 05:22:42.778741	2012-08-02 05:26:13.624157
6110	1-6110	56	7	4572	11	8503	0	0	2012-08-02 05:23:29.321029	2012-08-02 05:26:13.624157
5528	1-5528	30	90	8736	87	8643	0	0	2012-08-02 05:19:32.429172	2012-08-02 05:26:13.624157
6321	1-6321	65	90	5066	56	5323	0	0	2012-08-02 05:24:56.528368	2012-08-02 05:26:13.624157
6511	1-6511	68	11	3003	90	7245	0	6	2012-08-02 05:26:13.624157	2012-08-02 05:26:13.624157
5661	1-5661	56	42	3331	51	6805	0	0	2012-08-02 05:20:25.844256	2012-08-02 05:26:14.862037
6319	1-6319	74	24	1113	10	8596	0	0	2012-08-02 05:24:55.677723	2012-08-02 05:26:14.862037
5583	1-5583	72	16	4891	10	3934	0	4	2012-08-02 05:19:54.329953	2012-08-02 05:26:14.862037
6225	1-6225	68	87	4956	80	9013	0	0	2012-08-02 05:24:16.108728	2012-08-02 05:26:14.862037
6396	1-6396	10	87	3923	5	8637	0	0	2012-08-02 05:25:27.315553	2012-08-02 05:26:14.862037
5257	1-5257	2	44	1217	41	39	0	3	2012-08-02 05:17:45.435585	2012-08-02 05:26:16.11233
5687	1-5687	64	44	5232	42	3330	0	8	2012-08-02 05:20:36.387302	2012-08-02 05:26:16.11233
5145	1-5145	20	39	925	37	1097	0	0	2012-08-02 05:17:01.325959	2012-08-02 05:26:16.11233
6517	1-6517	41	41	440	39	6416	0	2	2012-08-02 05:26:16.11233	2012-08-02 05:26:16.11233
6416	1-6416	20	84	461	83	2462	0	0	2012-08-02 05:25:35.359137	2012-08-02 05:26:17.361149
5923	1-5923	3	74	2999	62	1483	0	0	2012-08-02 05:22:12.415243	2012-08-02 05:26:19.868796
3377	1-3377	38	30	5057	21	3922	0	5	2012-08-02 05:05:12.072796	2012-08-02 05:26:20.277474
6207	1-6207	44	30	3294	24	2741	0	3	2012-08-02 05:24:08.591197	2012-08-02 05:26:20.277474
6530	1-6530	8	28	822	63	7268	0	5	2012-08-02 05:26:21.537143	2012-08-02 05:26:21.537143
5757	1-5757	73	48	2808	89	1828	0	4	2012-08-02 05:21:05.180877	2012-08-02 05:26:23.603351
4728	1-4728	83	63	9468	14	9335	0	6	2012-08-02 05:14:14.089556	2012-08-02 05:26:23.603351
357	1-357	49	63	9837	14	8283	0	0	2012-08-02 04:44:01.153643	2012-08-02 05:26:23.603351
6535	1-6535	99	14	4954	94	9736	0	0	2012-08-02 05:26:23.603351	2012-08-02 05:26:23.603351
6538	1-6538	20	92	103	5	4608	0	0	2012-08-02 05:26:24.829477	2012-08-02 05:26:24.829477
5733	1-5733	13	91	4752	37	9599	0	0	2012-08-02 05:20:55.13698	2012-08-02 05:26:25.238686
6265	1-6265	55	91	614	15	2635	0	6	2012-08-02 05:24:32.793816	2012-08-02 05:26:26.055808
6411	1-6411	71	37	6219	38	7565	0	0	2012-08-02 05:25:33.370486	2012-08-02 05:26:26.055808
5393	1-5393	88	40	5979	15	7021	0	6	2012-08-02 05:18:38.793333	2012-08-02 05:26:26.055808
6541	1-6541	73	15	5135	9	9676	0	0	2012-08-02 05:26:26.055808	2012-08-02 05:26:26.055808
5363	1-5363	8	68	9488	72	7155	0	0	2012-08-02 05:18:27.015849	2012-08-02 05:26:26.453943
5329	1-5329	9	68	8431	87	6260	0	0	2012-08-02 05:18:13.636325	2012-08-02 05:26:26.453943
5797	1-5797	20	83	5267	89	6276	0	2	2012-08-02 05:21:21.312333	2012-08-02 05:26:26.453943
2286	1-2286	16	68	8586	13	7756	0	4	2012-08-02 04:57:34.917902	2012-08-02 05:26:26.453943
6542	1-6542	64	36	3574	68	7291	0	4	2012-08-02 05:26:26.453943	2012-08-02 05:26:26.453943
6455	1-6455	24	46	3997	58	4476	0	3	2012-08-02 05:25:51.147558	2012-08-02 05:26:26.885069
6149	1-6149	55	79	4293	11	7763	0	0	2012-08-02 05:23:44.945218	2012-08-02 05:26:27.293887
5662	1-5662	86	82	5335	84	8813	0	11	2012-08-02 05:20:26.230535	2012-08-02 05:26:27.293887
5460	1-5460	46	45	1901	63	5488	0	13	2012-08-02 05:19:05.200332	2012-08-02 05:26:27.293887
6182	1-6182	66	61	3469	86	8542	0	0	2012-08-02 05:23:58.191503	2012-08-02 05:26:28.111468
5800	1-5800	36	5	7857	86	3633	0	7	2012-08-02 05:21:22.494284	2012-08-02 05:26:28.111468
5468	1-5468	66	6	4784	77	5319	0	4	2012-08-02 05:19:08.340375	2012-08-02 05:26:28.962583
4317	1-4317	97	6	9818	4	8652	0	0	2012-08-02 05:11:28.113968	2012-08-02 05:26:28.962583
5726	1-5726	84	9	4732	37	7544	0	0	2012-08-02 05:20:52.065219	2012-08-02 05:26:28.962583
6039	1-6039	14	31	4159	46	2494	0	0	2012-08-02 05:23:00.767776	2012-08-02 05:26:30.211032
6072	1-6072	55	79	7632	66	8821	0	0	2012-08-02 05:23:14.126826	2012-08-02 05:26:30.608534
6552	1-6552	39	66	639	23	4831	0	0	2012-08-02 05:26:30.608534	2012-08-02 05:26:30.608534
5665	1-5665	19	68	9390	66	708	708	0	2012-08-02 05:20:27.423784	2012-08-02 05:26:30.608534
5752	1-5752	29	68	9616	66	5362	541	0	2012-08-02 05:21:03.081709	2012-08-02 05:26:30.608534
5532	1-5532	13	25	6308	76	8060	0	5	2012-08-02 05:19:34.009303	2012-08-02 05:26:31.437153
3801	1-3801	97	79	8704	15	8268	0	2	2012-08-02 05:08:02.021286	2012-08-02 05:26:31.437153
6107	1-6107	57	38	5840	15	9166	0	2	2012-08-02 05:23:28.149641	2012-08-02 05:26:31.437153
5926	1-5926	25	61	6062	82	7419	0	0	2012-08-02 05:22:13.675105	2012-08-02 05:26:31.83508
6557	1-6557	66	21	38	50	5379	0	0	2012-08-02 05:26:32.652602	2012-08-02 05:26:32.652602
4486	1-4486	64	12	6473	78	8029	0	0	2012-08-02 05:12:36.189871	2012-08-02 05:26:33.072707
4402	1-4402	82	82	3119	21	2134	0	3	2012-08-02 05:12:03.007891	2012-08-02 05:26:33.072707
6514	1-6514	30	10	1978	87	9695	0	0	2012-08-02 05:26:14.862037	2012-08-02 05:26:33.072707
5275	1-5275	54	75	5825	11	9172	0	0	2012-08-02 05:17:52.497138	2012-08-02 05:26:35.503161
5398	1-5398	88	79	9033	100	6752	0	6	2012-08-02 05:18:40.760271	2012-08-02 05:26:35.934367
4967	1-4967	16	47	4159	76	6842	0	0	2012-08-02 05:15:49.33804	2012-08-02 05:26:36.751571
5621	1-5621	73	47	3541	15	5239	0	4	2012-08-02 05:20:09.635287	2012-08-02 05:26:36.751571
6006	1-6006	87	66	4599	63	7553	0	0	2012-08-02 05:22:46.944126	2012-08-02 05:26:36.751571
6113	1-6113	100	59	3502	15	4123	0	16	2012-08-02 05:23:30.502854	2012-08-02 05:26:36.751571
4228	1-4228	81	96	3600	34	1623	0	0	2012-08-02 05:10:52.225032	2012-08-02 05:26:37.547286
6569	1-6569	37	65	1826	86	3883	0	0	2012-08-02 05:26:37.547286	2012-08-02 05:26:37.547286
6520	1-6520	12	83	2643	4	1975	0	0	2012-08-02 05:26:17.361149	2012-08-02 05:26:38.353881
6274	1-6274	89	56	4932	94	7474	0	0	2012-08-02 05:24:36.848189	2012-08-02 05:26:40.397758
6180	1-6180	4	100	4021	87	5057	0	5	2012-08-02 05:23:57.395665	2012-08-02 05:26:40.397758
6476	1-6476	67	58	6378	91	9372	0	0	2012-08-02 05:25:59.46881	2012-08-02 05:26:40.397758
1623	1-1623	70	91	6622	87	4950	1702	0	2012-08-02 04:52:41.205227	2012-08-02 05:26:40.397758
6567	1-6567	78	15	8947	47	7192	0	9	2012-08-02 05:26:36.751571	2012-08-02 05:26:40.819686
4354	1-4354	38	39	7340	46	8547	0	0	2012-08-02 05:11:43.318495	2012-08-02 05:26:43.681353
3670	1-3670	36	55	5040	11	3994	0	0	2012-08-02 05:07:10.225136	2012-08-02 05:26:44.079019
6568	1-6568	40	99	2573	61	7958	0	6	2012-08-02 05:26:37.149672	2012-08-02 05:26:44.079019
6445	1-6445	8	23	2739	7	9172	0	4	2012-08-02 05:25:47.180911	2012-08-02 05:26:44.079019
6585	1-6585	21	11	2518	99	7118	0	0	2012-08-02 05:26:44.079019	2012-08-02 05:26:44.079019
6527	1-6527	4	21	3699	78	9293	0	0	2012-08-02 05:26:20.277474	2012-08-02 05:26:44.509925
3809	1-3809	76	51	9803	87	3521	0	3	2012-08-02 05:08:05.171123	2012-08-02 05:26:44.509925
5879	1-5879	3	21	7139	3	6107	0	8	2012-08-02 05:21:54.491972	2012-08-02 05:26:44.509925
6546	1-6546	99	86	3955	96	6469	0	0	2012-08-02 05:26:28.111468	2012-08-02 05:26:45.747196
5650	1-5650	91	69	4282	17	3060	0	0	2012-08-02 05:20:21.468609	2012-08-02 05:26:46.553776
1226	1-1226	17	3	6944	43	5025	0	0	2012-08-02 04:49:12.536337	2012-08-02 05:26:46.553776
6591	1-6591	18	98	1449	69	8530	0	0	2012-08-02 05:26:46.553776	2012-08-02 05:26:46.553776
6592	1-6592	5	19	44	1	5116	0	0	2012-08-02 05:26:46.974043	2012-08-02 05:26:46.974043
6410	1-6410	87	90	7086	64	6149	0	0	2012-08-02 05:25:32.984011	2012-08-02 05:26:47.393721
6593	1-6593	92	81	1344	90	9253	0	0	2012-08-02 05:26:47.393721	2012-08-02 05:26:47.393721
6432	1-6432	47	21	2093	91	3877	0	6	2012-08-02 05:25:41.977529	2012-08-02 05:26:48.211305
6431	1-6431	93	56	3263	74	7343	0	0	2012-08-02 05:25:41.568516	2012-08-02 05:26:48.211305
6485	1-6485	74	21	3466	56	3166	0	0	2012-08-02 05:26:03.048962	2012-08-02 05:26:48.211305
5900	1-5900	44	16	3808	51	7040	0	0	2012-08-02 05:22:02.956839	2012-08-02 05:26:48.211305
6294	1-6294	17	20	7043	99	2379	0	0	2012-08-02 05:24:45.291969	2012-08-02 05:26:49.073561
6598	1-6598	56	92	546	35	5044	0	0	2012-08-02 05:26:49.482262	2012-08-02 05:26:49.482262
6600	1-6600	21	37	505	6	1606	0	7	2012-08-02 05:26:50.288688	2012-08-02 05:26:50.288688
5249	1-5249	46	38	7905	71	4915	0	0	2012-08-02 05:17:42.297882	2012-08-02 05:26:51.095279
6571	1-6571	82	89	586	83	3785	0	4	2012-08-02 05:26:38.353881	2012-08-02 05:26:51.095279
6603	1-6603	55	44	4177	97	2731	0	8	2012-08-02 05:26:51.504147	2012-08-02 05:26:51.504147
6604	1-6604	4	49	1422	24	9666	0	0	2012-08-02 05:26:51.913099	2012-08-02 05:26:51.913099
6548	1-6548	84	37	6098	6	9028	0	3	2012-08-02 05:26:28.962583	2012-08-02 05:26:52.321906
4622	1-4622	67	75	6596	65	9907	0	0	2012-08-02 05:13:30.996502	2012-08-02 05:26:52.730726
6606	1-6606	78	65	274	57	5019	0	4	2012-08-02 05:26:52.730726	2012-08-02 05:26:52.730726
6282	1-6282	32	93	1373	53	6189	0	0	2012-08-02 05:24:40.253037	2012-08-02 05:26:53.150554
2730	1-2730	45	9	6520	17	5467	0	0	2012-08-02 05:00:46.95348	2012-08-02 05:26:53.5484
6597	1-6597	23	51	4847	20	5560	0	0	2012-08-02 05:26:49.073561	2012-08-02 05:26:53.5484
6608	1-6608	40	17	3735	9	8140	0	0	2012-08-02 05:26:53.5484	2012-08-02 05:26:53.5484
4648	1-4648	61	59	8953	20	3122	0	3	2012-08-02 05:13:41.692322	2012-08-02 05:26:53.946204
5702	1-5702	36	87	7677	42	8143	0	0	2012-08-02 05:20:42.430664	2012-08-02 05:26:55.128483
6614	1-6614	55	9	4017	71	8560	0	11	2012-08-02 05:26:55.968025	2012-08-02 05:26:55.968025
6402	1-6402	13	23	7114	19	5624	0	3	2012-08-02 05:25:29.735097	2012-08-02 05:26:57.18316
5796	1-5796	76	99	4872	49	9254	0	6	2012-08-02 05:21:20.914656	2012-08-02 05:26:57.18316
2628	1-2628	88	99	4807	31	5379	0	0	2012-08-02 05:00:03.740005	2012-08-02 05:26:57.18316
6619	1-6619	15	45	501	54	1138	0	0	2012-08-02 05:26:58.023143	2012-08-02 05:26:58.023143
5638	1-5638	45	97	5365	81	6143	0	6	2012-08-02 05:20:16.574087	2012-08-02 05:26:58.851617
6622	1-6622	65	32	1698	88	5592	0	0	2012-08-02 05:26:59.271525	2012-08-02 05:26:59.271525
6543	1-6543	42	5	1258	31	2598	0	0	2012-08-02 05:26:26.885069	2012-08-02 05:26:59.691364
6627	1-6627	98	65	480	11	4141	0	0	2012-08-02 05:27:01.315608	2012-08-02 05:27:01.315608
6387	1-6387	65	58	6069	5	8055	0	5	2012-08-02 05:25:23.678507	2012-08-02 05:27:02.530692
6415	1-6415	49	9	3015	75	1932	0	0	2012-08-02 05:25:34.961664	2012-08-02 05:27:03.757139
1254	1-1254	35	91	5290	15	4238	0	0	2012-08-02 04:49:29.551148	2012-08-02 05:27:03.757139
2523	1-2523	58	56	4913	43	3350	0	0	2012-08-02 04:59:18.603118	2012-08-02 05:27:03.757139
5293	1-5293	52	42	4078	15	3844	0	2	2012-08-02 05:17:59.54608	2012-08-02 05:27:03.757139
6633	1-6633	9	75	2941	91	9871	0	7	2012-08-02 05:27:03.757139	2012-08-02 05:27:03.757139
5644	1-5644	82	33	8013	8	4373	0	0	2012-08-02 05:20:18.9604	2012-08-02 05:27:04.585958
6554	1-6554	7	15	9891	75	6563	0	0	2012-08-02 05:26:31.437153	2012-08-02 05:27:05.417057
6637	1-6637	16	30	126	63	2495	0	0	2012-08-02 05:27:05.417057	2012-08-02 05:27:05.417057
5771	1-5771	53	85	3980	54	2536	0	7	2012-08-02 05:21:10.949605	2012-08-02 05:27:06.987115
5840	1-5840	66	75	2501	85	8589	0	0	2012-08-02 05:21:38.594377	2012-08-02 05:27:06.987115
4581	1-4581	30	75	7895	97	6568	0	0	2012-08-02 05:13:14.135081	2012-08-02 05:27:06.987115
6641	1-6641	100	15	5433	75	8036	0	9	2012-08-02 05:27:06.987115	2012-08-02 05:27:06.987115
775	1-775	76	44	7075	53	4423	0	0	2012-08-02 04:45:23.910159	2012-08-02 05:27:07.396153
5079	1-5079	18	57	9509	84	5896	0	0	2012-08-02 05:16:34.387839	2012-08-02 05:27:08.180332
5217	1-5217	57	90	7106	77	3741	0	0	2012-08-02 05:17:29.703281	2012-08-02 05:27:09.019983
6205	1-6205	23	61	5414	55	6065	0	17	2012-08-02 05:24:07.773765	2012-08-02 05:27:09.019983
6581	1-6581	66	18	2422	13	2682	0	0	2012-08-02 05:26:42.45464	2012-08-02 05:27:09.428845
2663	1-2663	6	87	4669	2	2783	0	0	2012-08-02 05:00:18.512782	2012-08-02 05:27:09.428845
6647	1-6647	47	2	2480	18	3367	0	0	2012-08-02 05:27:09.428845	2012-08-02 05:27:09.428845
3143	1-3143	57	87	6670	2	2087	2087	0	2012-08-02 05:03:37.383718	2012-08-02 05:27:09.428845
6536	1-6536	52	24	5995	96	4557	0	0	2012-08-02 05:26:24.012083	2012-08-02 05:27:10.600498
5437	1-5437	23	84	9637	70	6468	0	0	2012-08-02 05:18:56.154951	2012-08-02 05:27:10.600498
1295	1-1295	75	24	4227	13	2177	0	0	2012-08-02 04:49:52.241386	2012-08-02 05:27:10.600498
6461	1-6461	25	14	9535	33	8802	0	6	2012-08-02 05:25:53.547147	2012-08-02 05:27:10.600498
6650	1-6650	73	33	4771	24	9185	0	5	2012-08-02 05:27:10.600498	2012-08-02 05:27:10.600498
3680	1-3680	23	56	6227	68	2387	147	0	2012-08-02 05:07:14.213865	2012-08-02 05:27:10.600498
6403	1-6403	89	80	6118	9	1711	0	0	2012-08-02 05:25:30.144082	2012-08-02 05:27:11.009527
6495	1-6495	88	1	8628	16	8058	0	0	2012-08-02 05:26:07.038882	2012-08-02 05:27:11.009527
5819	1-5819	8	48	4658	31	9214	0	0	2012-08-02 05:21:30.175359	2012-08-02 05:27:11.804678
6653	1-6653	40	90	360	48	2981	0	3	2012-08-02 05:27:11.804678	2012-08-02 05:27:11.804678
6577	1-6577	29	92	5131	15	7051	0	0	2012-08-02 05:26:40.819686	2012-08-02 05:27:12.213706
6654	1-6654	87	15	503	65	6724	0	9	2012-08-02 05:27:12.213706	2012-08-02 05:27:12.213706
6609	1-6609	30	12	718	59	4175	0	0	2012-08-02 05:26:53.946204	2012-08-02 05:27:12.60058
5914	1-5914	37	12	2671	27	5536	0	11	2012-08-02 05:22:08.691983	2012-08-02 05:27:12.60058
1778	1-1778	18	92	9569	88	3810	0	0	2012-08-02 04:53:52.390938	2012-08-02 05:27:12.60058
6655	1-6655	48	44	3497	12	8081	0	7	2012-08-02 05:27:12.60058	2012-08-02 05:27:12.60058
6642	1-6642	54	88	1661	44	8728	0	5	2012-08-02 05:27:07.396153	2012-08-02 05:27:13.782862
6658	1-6658	24	51	1579	41	4027	0	0	2012-08-02 05:27:13.782862	2012-08-02 05:27:13.782862
6169	1-6169	7	71	6340	100	5878	0	0	2012-08-02 05:23:52.955525	2012-08-02 05:27:14.578475
6660	1-6660	83	41	2897	71	8182	0	2	2012-08-02 05:27:14.578475	2012-08-02 05:27:14.578475
6379	1-6379	49	37	3912	44	9726	0	12	2012-08-02 05:25:20.507878	2012-08-02 05:27:16.975876
3616	1-3616	44	36	9112	1	4182	469	0	2012-08-02 05:06:48.489446	2012-08-02 05:27:16.975876
6248	1-6248	63	19	2487	35	2634	0	6	2012-08-02 05:24:25.723239	2012-08-02 05:27:17.78253
6439	1-6439	85	1	3234	45	8710	0	4	2012-08-02 05:25:44.783662	2012-08-02 05:27:17.78253
1492	1-1492	74	26	4061	77	1902	0	0	2012-08-02 04:51:35.787678	2012-08-02 05:27:17.78253
6101	1-6101	87	61	2672	45	5122	0	5	2012-08-02 05:23:25.774575	2012-08-02 05:27:17.78253
6668	1-6668	3	45	6863	19	8162	0	0	2012-08-02 05:27:17.78253	2012-08-02 05:27:17.78253
6672	1-6672	76	83	106	43	3237	0	8	2012-08-02 05:27:19.407107	2012-08-02 05:27:19.407107
6612	1-6612	59	36	1040	76	937	0	0	2012-08-02 05:26:55.128483	2012-08-02 05:27:21.428438
5956	1-5956	59	15	5694	78	1592	0	0	2012-08-02 05:22:26.216744	2012-08-02 05:27:22.97549
6683	1-6683	15	60	2517	94	8260	0	5	2012-08-02 05:27:23.78199	2012-08-02 05:27:23.78199
6297	1-6297	29	41	3290	36	3833	0	7	2012-08-02 05:24:46.529025	2012-08-02 05:27:24.577424
6685	1-6685	68	36	599	61	4025	0	0	2012-08-02 05:27:24.577424	2012-08-02 05:27:24.577424
6227	1-6227	27	51	7267	29	9308	0	4	2012-08-02 05:24:16.926451	2012-08-02 05:27:25.384016
6692	1-6692	64	66	3124	42	4678	0	0	2012-08-02 05:27:27.383505	2012-08-02 05:27:27.383505
6393	1-6393	61	100	5833	90	5757	0	0	2012-08-02 05:25:26.100073	2012-08-02 05:27:27.792434
6551	1-6551	42	14	4014	31	8046	0	0	2012-08-02 05:26:30.211032	2012-08-02 05:27:27.792434
6565	1-6565	98	64	2319	79	6668	0	0	2012-08-02 05:26:35.934367	2012-08-02 05:27:28.192046
5383	1-5383	61	47	2599	27	4717	0	6	2012-08-02 05:18:34.849234	2012-08-02 05:27:28.192046
6174	1-6174	82	79	3124	40	5503	0	5	2012-08-02 05:23:54.946656	2012-08-02 05:27:28.192046
3981	1-3981	81	79	6721	40	5233	1921	0	2012-08-02 05:09:14.080794	2012-08-02 05:27:28.192046
6651	1-6651	1	9	5161	1	9485	0	0	2012-08-02 05:27:11.009527	2012-08-02 05:27:29.40726
6005	1-6005	82	56	7827	96	8956	0	10	2012-08-02 05:22:46.535304	2012-08-02 05:27:29.40726
1599	1-1599	81	9	9073	50	6679	0	0	2012-08-02 04:52:29.74751	2012-08-02 05:27:29.40726
6617	1-6617	84	86	3943	99	6481	0	0	2012-08-02 05:26:57.18316	2012-08-02 05:27:30.202857
6661	1-6661	82	28	1225	91	597	0	9	2012-08-02 05:27:14.987351	2012-08-02 05:27:30.738823
6700	1-6700	25	6	1353	28	4014	0	5	2012-08-02 05:27:30.738823	2012-08-02 05:27:30.738823
6638	1-6638	32	57	2401	81	2150	0	12	2012-08-02 05:27:05.837587	2012-08-02 05:27:31.529051
837	1-837	85	57	5966	93	1309	0	0	2012-08-02 04:45:49.115449	2012-08-02 05:27:31.529051
6703	1-6703	9	53	209	1	5217	0	3	2012-08-02 05:27:31.915527	2012-08-02 05:27:31.915527
5798	1-5798	72	8	3716	34	8961	0	0	2012-08-02 05:21:21.709983	2012-08-02 05:27:32.291283
6539	1-6539	25	37	3128	38	1940	0	4	2012-08-02 05:26:25.238686	2012-08-02 05:27:32.291283
3958	1-3958	5	86	9761	55	4588	0	0	2012-08-02 05:09:04.914542	2012-08-02 05:27:32.291283
6699	1-6699	29	23	4385	26	5849	0	0	2012-08-02 05:27:30.202857	2012-08-02 05:27:32.291283
6054	1-6054	61	37	9173	4	3472	0	0	2012-08-02 05:23:06.956293	2012-08-02 05:27:32.291283
5272	1-5272	82	54	8475	74	2441	0	4	2012-08-02 05:17:51.347813	2012-08-02 05:27:32.291283
6705	1-6705	37	94	450	22	1679	0	0	2012-08-02 05:27:32.689436	2012-08-02 05:27:32.689436
6621	1-6621	13	81	4735	65	4988	0	12	2012-08-02 05:26:58.851617	2012-08-02 05:27:33.462715
5290	1-5290	6	60	5843	73	6300	0	0	2012-08-02 05:17:58.374871	2012-08-02 05:27:34.26933
6709	1-6709	56	73	1821	75	5632	0	6	2012-08-02 05:27:34.26933	2012-08-02 05:27:34.26933
6711	1-6711	49	15	1546	5	3008	0	0	2012-08-02 05:27:35.032042	2012-08-02 05:27:35.032042
6697	1-6697	76	50	5613	9	8708	0	5	2012-08-02 05:27:29.40726	2012-08-02 05:27:35.418346
6665	1-6665	100	3	977	37	4158	0	9	2012-08-02 05:27:16.578454	2012-08-02 05:27:35.418346
5057	1-5057	98	18	5680	30	5963	0	11	2012-08-02 05:16:25.403827	2012-08-02 05:27:35.418346
6605	1-6605	58	23	2291	37	5516	0	0	2012-08-02 05:26:52.321906	2012-08-02 05:27:35.418346
5083	1-5083	52	74	5042	59	3672	0	0	2012-08-02 05:16:36.056264	2012-08-02 05:27:35.849737
4791	1-4791	100	74	6025	49	2822	0	0	2012-08-02 05:14:39.460163	2012-08-02 05:27:35.849737
6707	1-6707	48	79	2869	33	6009	0	11	2012-08-02 05:27:33.462715	2012-08-02 05:27:35.849737
6332	1-6332	28	32	1470	33	7549	0	0	2012-08-02 05:25:01.113546	2012-08-02 05:27:35.849737
6713	1-6713	16	80	1207	74	9576	0	4	2012-08-02 05:27:35.849737	2012-08-02 05:27:35.849737
6417	1-6417	74	82	4869	12	8532	0	7	2012-08-02 05:25:35.779043	2012-08-02 05:27:36.623333
1392	1-1392	19	82	5781	12	130	130	2	2012-08-02 04:50:41.089405	2012-08-02 05:27:36.623333
5893	1-5893	13	84	6314	52	1364	0	7	2012-08-02 05:22:00.139735	2012-08-02 05:27:38.203127
6719	1-6719	32	2	2657	24	6648	0	0	2012-08-02 05:27:38.203127	2012-08-02 05:27:38.203127
939	1-939	45	25	6185	76	1760	331	0	2012-08-02 04:46:36.253294	2012-08-02 05:27:38.203127
6346	1-6346	71	72	4197	47	3824	0	0	2012-08-02 05:25:06.962341	2012-08-02 05:27:38.998728
5091	1-5091	35	7	7948	97	5513	0	0	2012-08-02 05:16:39.293448	2012-08-02 05:27:39.396415
6216	1-6216	19	21	8328	1	5727	0	3	2012-08-02 05:24:12.304104	2012-08-02 05:27:39.396415
6576	1-6576	58	42	8332	1	9129	0	0	2012-08-02 05:26:40.397758	2012-08-02 05:27:39.396415
6722	1-6722	78	11	2205	21	8245	0	0	2012-08-02 05:27:39.396415	2012-08-02 05:27:39.396415
6610	1-6610	67	91	5008	9	5158	0	6	2012-08-02 05:26:54.333043	2012-08-02 05:27:39.805543
4283	1-4283	73	93	5478	61	1500	0	4	2012-08-02 05:11:14.070627	2012-08-02 05:27:39.805543
6640	1-6640	80	21	4452	74	5310	0	11	2012-08-02 05:27:06.611317	2012-08-02 05:27:39.805543
6085	1-6085	4	68	2817	66	3157	0	7	2012-08-02 05:23:19.333043	2012-08-02 05:27:40.612505
5964	1-5964	35	66	6345	26	9472	0	0	2012-08-02 05:22:29.497884	2012-08-02 05:27:40.612505
6623	1-6623	72	73	2890	5	6160	0	4	2012-08-02 05:26:59.691364	2012-08-02 05:27:42.191957
6729	1-6729	70	34	682	47	2093	0	2	2012-08-02 05:27:42.191957	2012-08-02 05:27:42.191957
6732	1-6732	30	42	741	11	5816	0	0	2012-08-02 05:27:43.352192	2012-08-02 05:27:43.352192
6521	1-6521	97	43	6098	67	9160	0	0	2012-08-02 05:26:17.780728	2012-08-02 05:27:44.158655
6631	1-6631	80	23	260	40	305	0	2	2012-08-02 05:27:02.928474	2012-08-02 05:27:44.158655
6526	1-6526	87	62	1706	74	6445	0	4	2012-08-02 05:26:19.868796	2012-08-02 05:27:44.55662
6112	1-6112	30	62	5981	75	4380	0	0	2012-08-02 05:23:30.116181	2012-08-02 05:27:44.55662
6735	1-6735	20	88	1884	62	7981	0	0	2012-08-02 05:27:44.55662	2012-08-02 05:27:44.55662
6583	1-6583	51	3	2495	66	4901	0	0	2012-08-02 05:26:43.261235	2012-08-02 05:27:45.341182
3986	1-3986	58	3	5825	14	6510	0	2	2012-08-02 05:09:16.047694	2012-08-02 05:27:45.341182
6737	1-6737	72	14	3911	3	9892	0	4	2012-08-02 05:27:45.341182	2012-08-02 05:27:45.341182
6480	1-6480	23	88	6818	68	4563	0	10	2012-08-02 05:26:01.049028	2012-08-02 05:27:46.921214
4319	1-4319	28	7	6169	39	3169	0	0	2012-08-02 05:11:28.942832	2012-08-02 05:27:47.307836
4830	1-4830	66	7	6512	64	1999	0	0	2012-08-02 05:14:55.405551	2012-08-02 05:27:47.307836
6371	1-6371	53	13	6991	2	6267	0	2	2012-08-02 05:25:17.171511	2012-08-02 05:27:47.307836
6742	1-6742	63	36	4000	13	8694	0	2	2012-08-02 05:27:47.307836	2012-08-02 05:27:47.307836
5895	1-5895	61	70	6614	98	5850	0	5	2012-08-02 05:22:00.946358	2012-08-02 05:27:47.716909
6313	1-6313	73	89	2518	14	5657	0	0	2012-08-02 05:24:53.158632	2012-08-02 05:27:48.10344
4906	1-4906	9	9	7775	40	9833	0	0	2012-08-02 05:15:25.49346	2012-08-02 05:27:48.10344
3811	1-3811	73	56	3022	14	5838	0	0	2012-08-02 05:08:06.001573	2012-08-02 05:27:48.10344
6360	1-6360	80	8	1053	90	1334	0	0	2012-08-02 05:25:12.740978	2012-08-02 05:27:49.683686
6305	1-6305	44	72	4235	28	5806	0	8	2012-08-02 05:24:49.81086	2012-08-02 05:27:49.683686
5391	1-5391	27	10	7412	39	5160	0	0	2012-08-02 05:18:38.008999	2012-08-02 05:27:50.081364
4442	1-4442	10	47	7056	25	8785	0	0	2012-08-02 05:12:18.797669	2012-08-02 05:27:50.081364
6625	1-6625	39	83	2010	29	3061	0	6	2012-08-02 05:27:00.508929	2012-08-02 05:27:50.879047
1904	1-1904	93	35	9972	91	5356	0	0	2012-08-02 04:54:49.170937	2012-08-02 05:27:50.879047
6751	1-6751	95	93	3054	35	5818	0	2	2012-08-02 05:27:50.879047	2012-08-02 05:27:50.879047
6752	1-6752	42	36	514	24	1365	0	0	2012-08-02 05:27:51.276722	2012-08-02 05:27:51.276722
6646	1-6646	62	77	3303	61	7084	0	0	2012-08-02 05:27:09.019983	2012-08-02 05:27:53.674302
6020	1-6020	40	60	5281	59	9773	0	0	2012-08-02 05:22:52.824268	2012-08-02 05:27:59.960224
6712	1-6712	42	37	8366	18	9569	0	0	2012-08-02 05:27:35.418346	2012-08-02 05:28:00.347029
6774	1-6774	60	59	1347	7	7426	0	2	2012-08-02 05:27:59.960224	2012-08-02 05:28:01.529297
6309	1-6309	48	61	2627	64	3240	0	0	2012-08-02 05:24:51.478977	2012-08-02 05:28:01.529297
6704	1-6704	67	34	2324	23	9762	0	7	2012-08-02 05:27:32.291283	2012-08-02 05:28:02.700392
2920	1-2920	61	98	4999	31	8096	0	0	2012-08-02 05:02:06.523764	2012-08-02 05:28:03.485085
6078	1-6078	95	61	1776	88	3291	0	5	2012-08-02 05:23:16.5487	2012-08-02 05:28:05.429954
5097	1-5097	77	35	6642	61	5175	0	0	2012-08-02 05:16:41.823596	2012-08-02 05:28:07.430377
6778	1-6778	98	60	2987	25	5690	0	0	2012-08-02 05:28:01.529297	2012-08-02 05:28:07.430377
3066	1-3066	78	42	8142	96	1156	0	0	2012-08-02 05:03:06.035408	2012-08-02 05:28:07.430377
6395	1-6395	60	35	4542	1	9002	0	11	2012-08-02 05:25:26.91776	2012-08-02 05:28:07.430377
2828	1-2828	19	92	7233	51	4219	0	0	2012-08-02 05:01:28.092859	2012-08-02 05:28:07.850021
6794	1-6794	98	51	1675	4	2830	0	0	2012-08-02 05:28:07.850021	2012-08-02 05:28:07.850021
6325	1-6325	17	46	4725	6	4121	0	0	2012-08-02 05:24:58.185425	2012-08-02 05:28:08.247994
6524	1-6524	91	56	1225	55	543	0	3	2012-08-02 05:26:19.029157	2012-08-02 05:28:08.668104
5421	1-5421	31	52	4311	78	2991	0	0	2012-08-02 05:18:49.772118	2012-08-02 05:28:08.668104
6749	1-6749	50	88	7819	47	5553	0	2	2012-08-02 05:27:50.081364	2012-08-02 05:28:09.076678
6793	1-6793	86	96	1884	35	8823	0	0	2012-08-02 05:28:07.430377	2012-08-02 05:28:09.894263
5981	1-5981	48	2	7274	47	3403	0	4	2012-08-02 05:22:36.524831	2012-08-02 05:28:10.30332
2653	1-2653	53	21	8185	64	1598	0	0	2012-08-02 05:00:14.326101	2012-08-02 05:28:10.30332
6777	1-6777	52	87	5042	88	7081	0	0	2012-08-02 05:28:01.142789	2012-08-02 05:28:10.30332
6800	1-6800	22	47	967	87	5870	0	0	2012-08-02 05:28:10.30332	2012-08-02 05:28:10.30332
5567	1-5567	47	62	6123	54	7108	0	2	2012-08-02 05:19:47.954281	2012-08-02 05:28:11.949295
3216	1-3216	91	70	8754	14	8952	0	0	2012-08-02 05:04:07.044529	2012-08-02 05:28:11.949295
6158	1-6158	67	60	5403	62	4392	0	0	2012-08-02 05:23:48.51403	2012-08-02 05:28:11.949295
6804	1-6804	6	54	4332	60	9276	0	0	2012-08-02 05:28:11.949295	2012-08-02 05:28:11.949295
6791	1-6791	60	31	4182	2	9626	0	0	2012-08-02 05:28:06.60155	2012-08-02 05:28:12.756067
6694	1-6694	42	40	6408	64	7308	0	8	2012-08-02 05:27:28.192046	2012-08-02 05:28:13.164874
6809	1-6809	41	67	708	2	5175	0	0	2012-08-02 05:28:13.993641	2012-08-02 05:28:13.993641
6734	1-6734	73	67	2638	23	5323	0	0	2012-08-02 05:27:44.158655	2012-08-02 05:28:16.447961
6544	1-6544	96	84	5060	61	4017	0	0	2012-08-02 05:26:27.293887	2012-08-02 05:28:17.265827
5260	1-5260	88	38	3668	23	1485	0	0	2012-08-02 05:17:46.606358	2012-08-02 05:28:17.685292
6818	1-6818	38	10	1033	54	4136	0	9	2012-08-02 05:28:17.685292	2012-08-02 05:28:17.685292
6018	1-6018	25	87	9860	8	1936	0	12	2012-08-02 05:22:52.006782	2012-08-02 05:28:18.094234
6819	1-6819	79	6	702	87	7797	0	0	2012-08-02 05:28:18.094234	2012-08-02 05:28:18.094234
3782	1-3782	44	9	6220	39	5005	0	0	2012-08-02 05:07:54.585479	2012-08-02 05:28:18.514158
6584	1-6584	94	46	3980	7	8011	0	5	2012-08-02 05:26:43.681353	2012-08-02 05:28:18.514158
6501	1-6501	77	13	4188	7	9890	0	0	2012-08-02 05:26:09.49156	2012-08-02 05:28:18.514158
6676	1-6676	75	91	2463	93	8356	0	0	2012-08-02 05:27:21.042136	2012-08-02 05:28:19.751657
6479	1-6479	57	92	4859	1	1456	0	11	2012-08-02 05:26:00.662312	2012-08-02 05:28:19.751657
6823	1-6823	79	93	1347	47	4318	0	0	2012-08-02 05:28:19.751657	2012-08-02 05:28:19.751657
4954	1-4954	30	16	7886	52	4047	0	0	2012-08-02 05:15:44.244989	2012-08-02 05:28:20.558299
4966	1-4966	55	82	6970	11	6728	0	11	2012-08-02 05:15:48.94037	2012-08-02 05:28:20.558299
6702	1-6702	79	10	2033	57	9357	0	4	2012-08-02 05:27:31.529051	2012-08-02 05:28:21.76258
6801	1-6801	9	71	602	10	3680	0	3	2012-08-02 05:28:10.734239	2012-08-02 05:28:22.580193
6828	1-6828	92	92	1709	10	4006	0	0	2012-08-02 05:28:21.76258	2012-08-02 05:28:22.580193
3598	1-3598	32	53	8680	14	7007	0	0	2012-08-02 05:06:41.319615	2012-08-02 05:28:22.989211
4065	1-4065	40	56	3749	81	2604	0	0	2012-08-02 05:09:47.737734	2012-08-02 05:28:22.989211
6595	1-6595	40	51	8084	21	7319	0	0	2012-08-02 05:26:48.211305	2012-08-02 05:28:24.646396
6718	1-6718	94	27	2028	80	2279	0	0	2012-08-02 05:27:37.816423	2012-08-02 05:28:24.646396
6586	1-6586	54	87	4407	21	7830	0	0	2012-08-02 05:26:44.509925	2012-08-02 05:28:24.646396
6386	1-6386	9	27	4403	79	5365	0	4	2012-08-02 05:25:23.281066	2012-08-02 05:28:24.646396
6835	1-6835	49	21	1113	27	9046	0	0	2012-08-02 05:28:24.646396	2012-08-02 05:28:24.646396
6721	1-6721	61	65	3246	72	4194	0	4	2012-08-02 05:27:38.998728	2012-08-02 05:28:25.055625
6836	1-6836	89	96	56	34	4609	0	3	2012-08-02 05:28:25.055625	2012-08-02 05:28:25.055625
5073	1-5073	66	77	4022	39	3027	0	6	2012-08-02 05:16:31.946258	2012-08-02 05:28:25.464564
2913	1-2913	6	77	7217	39	4326	2316	0	2012-08-02 05:02:03.581748	2012-08-02 05:28:25.464564
6239	1-6239	24	65	6949	79	9897	0	8	2012-08-02 05:24:21.932535	2012-08-02 05:28:27.088405
6841	1-6841	71	79	3224	62	5409	0	0	2012-08-02 05:28:27.088405	2012-08-02 05:28:27.088405
5845	1-5845	79	97	4868	30	3510	0	0	2012-08-02 05:21:40.759802	2012-08-02 05:28:28.348161
6847	1-6847	26	16	935	51	4262	0	6	2012-08-02 05:28:29.618384	2012-08-02 05:28:29.618384
3563	1-3563	56	56	2828	93	3103	0	2	2012-08-02 05:06:27.041613	2012-08-02 05:28:31.662109
6806	1-6806	66	50	1751	31	4870	0	3	2012-08-02 05:28:12.756067	2012-08-02 05:28:31.662109
4318	1-4318	57	96	7412	56	4835	0	8	2012-08-02 05:11:28.522646	2012-08-02 05:28:31.662109
1006	1-1006	7	56	8803	93	8549	6030	0	2012-08-02 04:47:14.31525	2012-08-02 05:28:31.662109
5178	1-5178	46	28	4611	22	2037	0	0	2012-08-02 05:17:14.420138	2012-08-02 05:28:32.08232
6399	1-6399	41	73	3036	93	3148	0	0	2012-08-02 05:25:28.51994	2012-08-02 05:28:33.275318
6857	1-6857	89	79	54	92	6485	0	6	2012-08-02 05:28:33.695249	2012-08-02 05:28:33.695249
6477	1-6477	62	14	3298	54	3461	0	3	2012-08-02 05:25:59.866621	2012-08-02 05:28:35.341652
6727	1-6727	44	22	1042	86	933	0	3	2012-08-02 05:27:41.408026	2012-08-02 05:28:35.341652
5076	1-5076	58	12	5424	72	2937	0	9	2012-08-02 05:16:33.161617	2012-08-02 05:28:35.750175
6427	1-6427	91	99	6442	23	9549	0	11	2012-08-02 05:25:39.977643	2012-08-02 05:28:35.750175
686	1-686	36	12	3728	93	877	0	0	2012-08-02 04:44:55.82976	2012-08-02 05:28:35.750175
6863	1-6863	75	70	763	98	7411	0	13	2012-08-02 05:28:36.159226	2012-08-02 05:28:36.159226
6864	1-6864	49	42	80	67	2888	0	13	2012-08-02 05:28:36.568033	2012-08-02 05:28:36.568033
6723	1-6723	21	5	7778	91	8771	0	0	2012-08-02 05:27:39.805543	2012-08-02 05:28:36.965902
2507	1-2507	56	75	9796	61	8304	0	13	2012-08-02 04:59:11.674953	2012-08-02 05:28:36.965902
6523	1-6523	5	100	6880	41	9146	0	4	2012-08-02 05:26:18.609645	2012-08-02 05:28:36.965902
2282	1-2282	21	75	4564	61	2130	2130	11	2012-08-02 04:57:33.216669	2012-08-02 05:28:36.965902
6744	1-6744	95	81	2550	9	4429	0	7	2012-08-02 05:27:48.10344	2012-08-02 05:28:39.374277
3032	1-3032	62	79	5178	2	3474	0	0	2012-08-02 05:02:52.232813	2012-08-02 05:28:39.771967
3351	1-3351	30	79	705	55	235	0	0	2012-08-02 05:05:01.263204	2012-08-02 05:28:39.771967
6872	1-6872	76	5	912	79	5577	0	7	2012-08-02 05:28:39.771967	2012-08-02 05:28:39.771967
4194	1-4194	26	42	6545	76	2860	0	0	2012-08-02 05:10:38.931411	2012-08-02 05:28:40.192064
6743	1-6743	44	61	2604	70	9959	0	0	2012-08-02 05:27:47.716909	2012-08-02 05:28:41.572939
4705	1-4705	53	59	8324	58	8558	0	0	2012-08-02 05:14:04.594839	2012-08-02 05:28:42.403864
6878	1-6878	38	17	3318	59	9380	0	9	2012-08-02 05:28:42.403864	2012-08-02 05:28:42.403864
6630	1-6630	74	5	5531	58	4524	0	0	2012-08-02 05:27:02.530692	2012-08-02 05:28:43.618762
6007	1-6007	30	46	7320	25	2413	0	0	2012-08-02 05:22:47.352691	2012-08-02 05:28:44.458432
6883	1-6883	76	24	186	46	5770	0	0	2012-08-02 05:28:44.458432	2012-08-02 05:28:44.458432
6796	1-6796	25	78	1581	29	8953	0	6	2012-08-02 05:28:08.668104	2012-08-02 05:28:44.878404
6884	1-6884	76	85	4567	78	5053	0	9	2012-08-02 05:28:44.878404	2012-08-02 05:28:44.878404
6441	1-6441	13	76	6712	38	9177	0	7	2012-08-02 05:25:45.579282	2012-08-02 05:28:46.51375
6669	1-6669	94	38	2177	48	5181	0	3	2012-08-02 05:27:18.202627	2012-08-02 05:28:46.51375
6888	1-6888	1	48	3965	52	5277	0	0	2012-08-02 05:28:46.51375	2012-08-02 05:28:46.51375
6453	1-6453	28	52	6277	9	9972	0	0	2012-08-02 05:25:50.351965	2012-08-02 05:28:47.331307
6890	1-6890	96	6	1047	16	4731	0	3	2012-08-02 05:28:47.331307	2012-08-02 05:28:47.331307
4095	1-4095	5	75	5946	25	7260	0	2	2012-08-02 05:09:59.749531	2012-08-02 05:28:47.740198
5760	1-5760	40	53	8041	88	3435	0	3	2012-08-02 05:21:06.574163	2012-08-02 05:28:48.137936
6825	1-6825	82	11	3664	16	5319	0	13	2012-08-02 05:28:20.558299	2012-08-02 05:28:48.54716
6811	1-6811	25	89	1146	67	556	0	2	2012-08-02 05:28:14.789345	2012-08-02 05:28:50.195319
6897	1-6897	5	23	1652	17	4439	0	4	2012-08-02 05:28:50.195319	2012-08-02 05:28:50.195319
6852	1-6852	25	93	5133	6	9676	0	0	2012-08-02 05:28:31.662109	2012-08-02 05:28:51.435829
4673	1-4673	22	96	9708	49	8823	0	15	2012-08-02 05:13:51.833761	2012-08-02 05:28:51.435829
6881	1-6881	42	58	2296	5	4245	0	10	2012-08-02 05:28:43.618762	2012-08-02 05:28:52.280186
6861	1-6861	90	54	6932	22	5172	0	11	2012-08-02 05:28:35.341652	2012-08-02 05:28:52.713097
6903	1-6903	56	14	1489	36	4509	0	0	2012-08-02 05:28:52.713097	2012-08-02 05:28:52.713097
6865	1-6865	99	41	5949	75	7565	0	0	2012-08-02 05:28:36.965902	2012-08-02 05:28:53.533357
6687	1-6687	14	29	9409	56	6822	0	0	2012-08-02 05:27:25.384016	2012-08-02 05:28:54.771556
6891	1-6891	62	76	4872	41	7252	0	0	2012-08-02 05:28:47.740198	2012-08-02 05:28:57.247656
6914	1-6914	13	35	240	76	3444	0	3	2012-08-02 05:28:57.247656	2012-08-02 05:28:57.247656
6915	1-6915	92	54	742	72	2351	0	0	2012-08-02 05:28:57.679565	2012-08-02 05:28:57.679565
6361	1-6361	87	55	2001	26	2098	0	5	2012-08-02 05:25:13.138461	2012-08-02 05:28:59.307302
6529	1-6529	71	6	2347	12	1686	0	0	2012-08-02 05:26:21.12836	2012-08-02 05:28:59.716283
5957	1-5957	98	1	6856	60	7115	0	0	2012-08-02 05:22:26.625397	2012-08-02 05:28:59.716283
6394	1-6394	31	1	8874	97	6396	0	0	2012-08-02 05:25:26.50909	2012-08-02 05:28:59.716283
4376	1-4376	92	11	8729	84	2386	2386	11	2012-08-02 05:11:52.355715	2012-08-02 05:28:59.716283
5717	1-5717	8	82	2247	64	2731	0	0	2012-08-02 05:20:48.385807	2012-08-02 05:29:00.138425
6921	1-6921	9	63	4530	82	7740	0	0	2012-08-02 05:29:00.138425	2012-08-02 05:29:00.138425
156	1-156	51	64	8711	47	1281	645	0	2012-08-02 04:43:55.320316	2012-08-02 05:29:00.138425
6923	1-6923	65	53	322	90	4127	0	11	2012-08-02 05:29:00.968231	2012-08-02 05:29:00.968231
6509	1-6509	71	20	1115	7	1922	0	0	2012-08-02 05:26:12.795513	2012-08-02 05:29:01.774837
6545	1-6545	74	16	6005	52	2953	0	2	2012-08-02 05:26:27.713738	2012-08-02 05:29:01.774837
6715	1-6715	67	95	2636	52	2174	0	2	2012-08-02 05:27:36.623333	2012-08-02 05:29:01.774837
6925	1-6925	80	7	1644	32	8441	0	0	2012-08-02 05:29:01.774837	2012-08-02 05:29:01.774837
5749	1-5749	63	36	5493	92	8204	0	0	2012-08-02 05:21:01.842422	2012-08-02 05:29:02.208404
6926	1-6926	75	92	3027	36	7557	0	0	2012-08-02 05:29:02.208404	2012-08-02 05:29:02.208404
6285	1-6285	83	36	2276	63	1576	0	8	2012-08-02 05:24:41.523934	2012-08-02 05:29:03.014937
5547	1-5547	12	11	7867	29	5499	0	0	2012-08-02 05:19:39.931013	2012-08-02 05:29:03.014937
2226	1-2226	6	36	8396	80	6177	0	0	2012-08-02 04:57:08.764627	2012-08-02 05:29:03.014937
6522	1-6522	90	51	988	50	230	0	0	2012-08-02 05:26:18.18948	2012-08-02 05:29:03.014937
6695	1-6695	93	35	2244	83	9674	0	0	2012-08-02 05:27:28.64414	2012-08-02 05:29:04.266931
6931	1-6931	74	83	8607	86	7361	0	0	2012-08-02 05:29:04.266931	2012-08-02 05:29:05.117609
6928	1-6928	49	29	1832	36	8230	0	4	2012-08-02 05:29:03.014937	2012-08-02 05:29:06.747953
6862	1-6862	52	22	3118	12	7042	0	5	2012-08-02 05:28:35.750175	2012-08-02 05:29:08.32781
6472	1-6472	42	78	9668	30	8517	0	7	2012-08-02 05:25:57.878213	2012-08-02 05:29:08.32781
6408	1-6408	9	73	2635	70	2070	0	0	2012-08-02 05:25:32.16589	2012-08-02 05:29:11.521102
5706	1-5706	89	65	9666	72	5424	0	0	2012-08-02 05:20:43.999456	2012-08-02 05:29:11.521102
3640	1-3640	57	75	3126	70	4078	0	0	2012-08-02 05:06:58.115002	2012-08-02 05:29:11.521102
4032	1-4032	35	36	7666	88	5363	0	0	2012-08-02 05:09:34.600806	2012-08-02 05:29:13.479262
6760	1-6760	88	47	616	63	1178	0	0	2012-08-02 05:27:54.46967	2012-08-02 05:29:13.877246
2148	1-2148	96	63	9637	55	2059	1573	0	2012-08-02 04:56:34.701015	2012-08-02 05:29:13.877246
6795	1-6795	39	6	6402	96	8599	0	2	2012-08-02 05:28:08.247994	2012-08-02 05:29:15.048262
2339	1-2339	49	6	6687	14	7373	0	0	2012-08-02 04:57:58.011376	2012-08-02 05:29:15.048262
4683	1-4683	35	20	3252	16	1611	0	0	2012-08-02 05:13:55.983553	2012-08-02 05:29:15.048262
1141	1-1141	10	96	7601	14	9576	2351	0	2012-08-02 04:48:31.522154	2012-08-02 05:29:15.048262
6063	1-6063	36	14	7647	80	6258	0	4	2012-08-02 05:23:10.569385	2012-08-02 05:29:15.866469
5352	1-5352	100	77	9176	67	9555	0	0	2012-08-02 05:18:22.718255	2012-08-02 05:29:15.866469
6959	1-6959	4	71	4871	69	2875	0	6	2012-08-02 05:29:15.446797	2012-08-02 05:29:16.673434
6853	1-6853	81	71	1187	55	2107	0	2	2012-08-02 05:28:32.08232	2012-08-02 05:29:16.673434
6962	1-6962	13	70	4761	71	8927	0	3	2012-08-02 05:29:16.673434	2012-08-02 05:29:16.673434
5928	1-5928	8	43	8678	75	7099	400	0	2012-08-02 05:22:14.503512	2012-08-02 05:29:16.673434
6462	1-6462	64	17	6449	89	6710	0	4	2012-08-02 05:25:53.922536	2012-08-02 05:29:17.083013
5963	1-5963	10	8	9471	9	8574	0	8	2012-08-02 05:22:29.078144	2012-08-02 05:29:17.083013
6963	1-6963	39	89	3357	97	6090	0	5	2012-08-02 05:29:17.083013	2012-08-02 05:29:17.083013
6967	1-6967	92	44	279	36	4443	0	5	2012-08-02 05:29:18.684963	2012-08-02 05:29:18.684963
6968	1-6968	47	53	635	48	6314	0	0	2012-08-02 05:29:19.126962	2012-08-02 05:29:19.126962
6856	1-6856	79	77	3177	22	4133	0	0	2012-08-02 05:28:33.275318	2012-08-02 05:29:20.308992
6949	1-6949	86	70	4535	65	9446	0	0	2012-08-02 05:29:11.521102	2012-08-02 05:29:20.308992
6971	1-6971	81	40	5209	35	7723	0	10	2012-08-02 05:29:20.308992	2012-08-02 05:29:20.308992
6972	1-6972	47	89	237	16	3831	0	0	2012-08-02 05:29:20.71803	2012-08-02 05:29:20.71803
6907	1-6907	92	12	863	81	1905	0	0	2012-08-02 05:28:54.362879	2012-08-02 05:29:23.844492
6003	1-6003	70	35	5517	76	5810	0	0	2012-08-02 05:22:45.695769	2012-08-02 05:29:23.844492
6980	1-6980	12	1	2057	35	8780	0	2	2012-08-02 05:29:23.844492	2012-08-02 05:29:23.844492
1785	1-1785	96	35	8483	74	6985	1897	0	2012-08-02 04:53:55.541025	2012-08-02 05:29:23.844492
6677	1-6677	36	76	4539	2	9486	0	0	2012-08-02 05:27:21.428438	2012-08-02 05:29:25.845038
6666	1-6666	1	44	4080	36	5730	0	0	2012-08-02 05:27:16.975876	2012-08-02 05:29:25.845038
1901	1-1901	98	53	4885	2	1860	0	0	2012-08-02 04:54:47.801091	2012-08-02 05:29:27.391463
6278	1-6278	46	55	5459	78	6109	0	0	2012-08-02 05:24:38.540586	2012-08-02 05:29:27.391463
6370	1-6370	82	71	4747	95	7892	0	0	2012-08-02 05:25:16.77361	2012-08-02 05:29:28.573731
6992	1-6992	100	5	2801	71	5927	0	0	2012-08-02 05:29:28.573731	2012-08-02 05:29:28.573731
5925	1-5925	17	9	8765	56	7133	0	2	2012-08-02 05:22:13.255005	2012-08-02 05:29:29.347239
610	1-610	49	30	2669	14	4178	0	0	2012-08-02 04:44:38.518643	2012-08-02 05:29:29.745263
6941	1-6941	34	30	1637	25	6023	0	5	2012-08-02 05:29:08.32781	2012-08-02 05:29:29.745263
5714	1-5714	30	20	4006	54	9026	0	0	2012-08-02 05:20:47.181353	2012-08-02 05:29:30.551949
6558	1-6558	50	52	5348	82	9424	0	0	2012-08-02 05:26:33.072707	2012-08-02 05:29:30.939009
6164	1-6164	72	78	9306	52	7064	0	0	2012-08-02 05:23:50.967001	2012-08-02 05:29:30.939009
6295	1-6295	1	17	8458	94	7888	0	4	2012-08-02 05:24:45.711682	2012-08-02 05:29:31.347676
6741	1-6741	50	39	1817	88	7216	0	0	2012-08-02 05:27:46.921214	2012-08-02 05:29:31.347676
6098	1-6098	96	23	6217	54	9909	0	0	2012-08-02 05:23:24.514775	2012-08-02 05:29:32.134027
6322	1-6322	16	79	4947	10	4254	0	9	2012-08-02 05:24:56.9484	2012-08-02 05:29:32.520845
6820	1-6820	32	7	7498	9	5375	0	0	2012-08-02 05:28:18.514158	2012-08-02 05:29:32.520845
4819	1-4819	79	79	3372	28	1357	0	0	2012-08-02 05:14:50.931395	2012-08-02 05:29:32.520845
7002	1-7002	83	28	1032	7	7903	0	0	2012-08-02 05:29:32.520845	2012-08-02 05:29:32.520845
5359	1-5359	35	60	6698	90	7502	0	0	2012-08-02 05:18:25.447047	2012-08-02 05:29:32.918892
4812	1-4812	25	93	8032	30	2846	0	0	2012-08-02 05:14:48.080197	2012-08-02 05:29:32.918892
6607	1-6607	70	31	7405	93	6235	0	5	2012-08-02 05:26:53.150554	2012-08-02 05:29:32.918892
7003	1-7003	48	90	389	31	5735	0	0	2012-08-02 05:29:32.918892	2012-08-02 05:29:32.918892
6528	1-6528	41	39	2990	58	6511	0	5	2012-08-02 05:26:20.73077	2012-08-02 05:29:34.156867
6839	1-6839	27	80	4481	67	2749	0	3	2012-08-02 05:28:26.293192	2012-08-02 05:29:34.532292
5443	1-5443	98	43	4899	49	2846	0	0	2012-08-02 05:18:58.475296	2012-08-02 05:29:34.532292
6613	1-6613	63	55	5322	28	6809	0	12	2012-08-02 05:26:55.548294	2012-08-02 05:29:34.94122
5846	1-5846	94	61	4473	75	3131	0	0	2012-08-02 05:21:41.168326	2012-08-02 05:29:35.759093
7017	1-7017	40	70	168	78	8703	0	0	2012-08-02 05:29:38.521063	2012-08-02 05:29:38.521063
5933	1-5933	12	8	9061	91	9157	0	0	2012-08-02 05:22:16.580438	2012-08-02 05:29:40.089552
7021	1-7021	53	41	3272	19	5450	0	8	2012-08-02 05:29:40.089552	2012-08-02 05:29:40.089552
6758	1-6758	16	72	7526	69	7343	0	3	2012-08-02 05:27:53.674302	2012-08-02 05:29:40.487537
5151	1-5151	50	19	1391	14	4291	0	0	2012-08-02 05:17:03.756385	2012-08-02 05:29:40.487537
7022	1-7022	12	1	1318	72	8863	0	6	2012-08-02 05:29:40.487537	2012-08-02 05:29:40.487537
7031	1-7031	71	68	1667	67	3600	0	0	2012-08-02 05:29:44.143935	2012-08-02 05:29:44.143935
6879	1-6879	77	39	7571	70	9510	0	0	2012-08-02 05:28:42.801346	2012-08-02 05:29:44.54131
7006	1-7006	24	80	7095	39	6366	0	2	2012-08-02 05:29:34.156867	2012-08-02 05:29:45.347781
6892	1-6892	57	42	942	53	7119	0	0	2012-08-02 05:28:48.137936	2012-08-02 05:29:45.756622
6241	1-6241	27	61	6001	100	7100	0	0	2012-08-02 05:24:22.772231	2012-08-02 05:29:46.916454
7038	1-7038	17	100	4958	8	6675	0	0	2012-08-02 05:29:46.916454	2012-08-02 05:29:46.916454
5374	1-5374	6	55	6233	34	8092	0	5	2012-08-02 05:18:31.324706	2012-08-02 05:29:47.325487
7040	1-7040	42	66	1638	87	7708	0	0	2012-08-02 05:29:47.712102	2012-08-02 05:29:47.712102
6908	1-6908	52	56	1509	13	6271	0	0	2012-08-02 05:28:54.771556	2012-08-02 05:29:48.099244
6368	1-6368	96	69	5500	39	7167	0	0	2012-08-02 05:25:15.955958	2012-08-02 05:29:50.076402
5929	1-5929	18	42	7338	31	6694	0	2	2012-08-02 05:22:14.923392	2012-08-02 05:29:50.076402
6249	1-6249	25	69	8413	77	9845	0	0	2012-08-02 05:24:26.131664	2012-08-02 05:29:50.076402
7049	1-7049	36	36	286	64	5215	0	4	2012-08-02 05:29:51.270069	2012-08-02 05:29:51.270069
5919	1-5919	55	93	9921	86	7264	0	0	2012-08-02 05:22:10.7582	2012-08-02 05:29:52.065187
7051	1-7051	35	40	214	93	8835	0	0	2012-08-02 05:29:52.065187	2012-08-02 05:29:52.065187
7054	1-7054	87	42	2967	7	4922	0	0	2012-08-02 05:29:53.247563	2012-08-02 05:29:53.247563
6947	1-6947	65	22	1107	40	7347	0	0	2012-08-02 05:29:10.725796	2012-08-02 05:29:54.031862
6662	1-6662	59	62	7492	35	4045	0	0	2012-08-02 05:27:15.374325	2012-08-02 05:29:54.440796
6958	1-6958	52	14	6208	20	9512	0	0	2012-08-02 05:29:15.048262	2012-08-02 05:29:55.634323
6929	1-6929	40	17	4771	85	7813	0	5	2012-08-02 05:29:03.459774	2012-08-02 05:29:56.429727
7062	1-7062	91	6	514	17	7634	0	0	2012-08-02 05:29:56.429727	2012-08-02 05:29:56.429727
7066	1-7066	65	76	42	23	3238	0	0	2012-08-02 05:29:57.998355	2012-08-02 05:29:57.998355
6893	1-6893	94	16	6514	68	8823	0	0	2012-08-02 05:28:48.54716	2012-08-02 05:29:58.374313
7067	1-7067	51	2	275	15	6218	0	0	2012-08-02 05:29:58.374313	2012-08-02 05:29:58.374313
6563	1-6563	50	13	9677	86	7920	0	0	2012-08-02 05:26:35.09463	2012-08-02 05:29:59.214171
3393	1-3393	20	6	4034	86	1769	0	0	2012-08-02 05:05:18.546649	2012-08-02 05:29:59.611968
5003	1-5003	8	15	4678	44	1565	0	11	2012-08-02 05:16:03.592501	2012-08-02 05:29:59.611968
7070	1-7070	61	64	2302	56	4735	0	0	2012-08-02 05:29:59.611968	2012-08-02 05:29:59.611968
2199	1-2199	77	56	5522	15	7208	109	0	2012-08-02 04:56:56.995952	2012-08-02 05:29:59.611968
4715	1-4715	37	70	8432	31	6411	0	0	2012-08-02 05:14:08.719904	2012-08-02 05:30:00.407485
5596	1-5596	93	60	4825	85	7999	0	0	2012-08-02 05:19:59.511272	2012-08-02 05:30:00.407485
4087	1-4087	27	82	7798	100	9241	0	0	2012-08-02 05:09:56.479281	2012-08-02 05:30:00.407485
6900	1-6900	54	49	4734	18	8120	0	5	2012-08-02 05:28:51.435829	2012-08-02 05:30:01.214283
6303	1-6303	60	49	3750	88	6577	0	13	2012-08-02 05:24:48.982133	2012-08-02 05:30:01.214283
6987	1-6987	32	12	3435	19	3500	0	17	2012-08-02 05:29:26.618437	2012-08-02 05:30:02.396483
4509	1-4509	66	22	9290	100	8313	0	0	2012-08-02 05:12:45.305799	2012-08-02 05:30:02.79416
7078	1-7078	6	63	657	22	4391	0	4	2012-08-02 05:30:02.79416	2012-08-02 05:30:02.79416
6356	1-6356	5	34	6517	65	7573	0	0	2012-08-02 05:25:11.127944	2012-08-02 05:30:03.181102
7079	1-7079	10	65	1138	86	1974	0	3	2012-08-02 05:30:03.181102	2012-08-02 05:30:03.181102
2805	1-2805	3	23	9376	71	3725	996	6	2012-08-02 05:01:18.480512	2012-08-02 05:30:03.181102
6435	1-6435	22	69	4459	39	4003	0	0	2012-08-02 05:25:43.181726	2012-08-02 05:30:05.14959
6964	1-6964	39	52	574	83	3960	0	6	2012-08-02 05:29:17.491594	2012-08-02 05:30:05.968545
4122	1-4122	32	83	2921	27	338	0	0	2012-08-02 05:10:10.457377	2012-08-02 05:30:06.377106
7087	1-7087	26	41	144	83	5979	0	0	2012-08-02 05:30:06.377106	2012-08-02 05:30:06.377106
6920	1-6920	38	81	5405	1	9426	0	5	2012-08-02 05:28:59.716283	2012-08-02 05:30:06.78614
6555	1-6555	12	64	8247	47	6697	0	3	2012-08-02 05:26:31.83508	2012-08-02 05:30:08.01239
7046	1-7046	54	31	7653	69	9781	0	0	2012-08-02 05:29:50.076402	2012-08-02 05:30:08.829956
6998	1-6998	1	100	4770	78	9123	0	2	2012-08-02 05:29:30.939009	2012-08-02 05:30:08.829956
6602	1-6602	100	9	5246	89	9927	0	0	2012-08-02 05:26:51.095279	2012-08-02 05:30:10.487041
7074	1-7074	56	71	3793	49	6494	0	0	2012-08-02 05:30:01.214283	2012-08-02 05:30:10.487041
6125	1-6125	12	8	4621	58	7895	0	0	2012-08-02 05:23:35.309213	2012-08-02 05:30:11.315671
6347	1-6347	3	43	6823	98	6867	0	5	2012-08-02 05:25:07.34908	2012-08-02 05:30:12.133489
6550	1-6550	10	71	7988	9	9865	0	0	2012-08-02 05:26:29.790951	2012-08-02 05:30:12.133489
6391	1-6391	40	22	5573	73	9138	0	3	2012-08-02 05:25:25.269635	2012-08-02 05:30:12.133489
1221	1-1221	25	98	9960	24	3180	176	0	2012-08-02 04:49:10.039834	2012-08-02 05:30:12.133489
6681	1-6681	50	1	3081	15	7597	0	0	2012-08-02 05:27:22.97549	2012-08-02 05:30:13.359889
6955	1-6955	47	34	3290	64	7978	0	8	2012-08-02 05:29:13.877246	2012-08-02 05:30:13.359889
7104	1-7104	15	15	2843	86	3498	0	16	2012-08-02 05:30:13.359889	2012-08-02 05:30:13.359889
6904	1-6904	30	73	6194	100	5720	0	0	2012-08-02 05:28:53.135566	2012-08-02 05:30:13.768662
7105	1-7105	19	11	1270	50	6612	0	6	2012-08-02 05:30:13.768662	2012-08-02 05:30:13.768662
7018	1-7018	14	49	2622	48	8754	0	11	2012-08-02 05:29:38.918556	2012-08-02 05:30:14.233013
6340	1-6340	25	26	8801	86	7969	0	10	2012-08-02 05:25:04.507616	2012-08-02 05:30:14.652631
6736	1-6736	90	86	7119	23	8529	0	0	2012-08-02 05:27:44.954777	2012-08-02 05:30:14.652631
6880	1-6880	44	56	9228	6	9720	0	0	2012-08-02 05:28:43.198947	2012-08-02 05:30:15.072482
7108	1-7108	13	9	968	56	7090	0	9	2012-08-02 05:30:15.072482	2012-08-02 05:30:15.072482
7109	1-7109	87	14	1696	37	5305	0	0	2012-08-02 05:30:15.503484	2012-08-02 05:30:15.503484
7069	1-7069	50	40	2613	13	3093	0	0	2012-08-02 05:29:59.214171	2012-08-02 05:30:15.901086
6989	1-6989	62	2	4815	54	8885	0	8	2012-08-02 05:29:27.391463	2012-08-02 05:30:17.149747
6693	1-6693	19	31	9895	100	9240	0	0	2012-08-02 05:27:27.792434	2012-08-02 05:30:17.547301
7099	1-7099	22	58	1314	72	2015	0	5	2012-08-02 05:30:11.315671	2012-08-02 05:30:18.375944
4280	1-4280	31	9	5860	20	3026	0	0	2012-08-02 05:11:12.866417	2012-08-02 05:30:18.375944
6337	1-6337	85	20	8913	6	8506	568	12	2012-08-02 05:25:03.237064	2012-08-02 05:30:18.375944
6787	1-6787	93	76	4269	42	9539	0	0	2012-08-02 05:28:05.042792	2012-08-02 05:30:20.02189
6885	1-6885	70	97	2347	67	1660	0	4	2012-08-02 05:28:45.287356	2012-08-02 05:30:20.02189
29	1-29	27	36	1010	42	394	0	0	2012-08-02 04:43:53.766106	2012-08-02 05:30:20.02189
6331	1-6331	90	29	1968	53	3786	0	0	2012-08-02 05:25:00.693959	2012-08-02 05:30:20.02189
7120	1-7120	7	53	3348	97	9972	0	12	2012-08-02 05:30:20.02189	2012-08-02 05:30:20.02189
6999	1-6999	70	5	5234	39	3874	0	0	2012-08-02 05:29:31.347676	2012-08-02 05:30:20.850931
3396	1-3396	75	97	9463	43	4917	0	0	2012-08-02 05:05:19.750947	2012-08-02 05:30:20.850931
4162	1-4162	23	43	9443	74	4635	0	0	2012-08-02 05:10:26.347232	2012-08-02 05:30:20.850931
7122	1-7122	8	39	2233	99	9489	0	0	2012-08-02 05:30:20.850931	2012-08-02 05:30:20.850931
6954	1-6954	70	67	3891	44	9443	0	0	2012-08-02 05:29:13.479262	2012-08-02 05:30:22.530398
7035	1-7035	13	74	5271	39	4443	0	0	2012-08-02 05:29:45.756622	2012-08-02 05:30:22.530398
7126	1-7126	99	44	3955	74	9083	0	11	2012-08-02 05:30:22.530398	2012-08-02 05:30:22.530398
6689	1-6689	26	37	8444	25	4666	0	3	2012-08-02 05:27:26.17927	2012-08-02 05:30:24.695435
6043	1-6043	56	37	8324	10	3955	0	0	2012-08-02 05:23:02.46912	2012-08-02 05:30:24.695435
7134	1-7134	49	64	831	93	4098	0	0	2012-08-02 05:30:25.932963	2012-08-02 05:30:25.932963
7010	1-7010	29	90	3676	61	8996	0	3	2012-08-02 05:29:35.759093	2012-08-02 05:30:26.761315
7055	1-7055	32	98	673	77	677	0	0	2012-08-02 05:29:53.634213	2012-08-02 05:30:26.761315
7136	1-7136	72	30	173	90	6321	0	3	2012-08-02 05:30:26.761315	2012-08-02 05:30:26.761315
6026	1-6026	1	97	2783	14	5121	0	0	2012-08-02 05:22:55.332315	2012-08-02 05:30:27.181526
7137	1-7137	90	14	752	18	2233	0	0	2012-08-02 05:30:27.181526	2012-08-02 05:30:27.181526
7110	1-7110	35	13	2858	65	2435	0	10	2012-08-02 05:30:15.901086	2012-08-02 05:30:27.590249
6740	1-6740	78	13	2754	60	1419	0	0	2012-08-02 05:27:46.52325	2012-08-02 05:30:27.590249
2987	1-2987	38	7	7700	8	1119	0	3	2012-08-02 05:02:34.003215	2012-08-02 05:30:27.590249
7138	1-7138	69	35	404	13	8768	0	0	2012-08-02 05:30:27.590249	2012-08-02 05:30:27.590249
7034	1-7034	3	10	9038	98	3535	0	0	2012-08-02 05:29:45.347781	2012-08-02 05:30:31.734848
6588	1-6588	66	93	6615	10	6711	0	2	2012-08-02 05:26:45.327434	2012-08-02 05:30:31.734848
7106	1-7106	96	75	3067	49	5410	0	0	2012-08-02 05:30:14.233013	2012-08-02 05:30:32.563513
2268	1-2268	96	75	5939	28	5960	0	0	2012-08-02 04:57:27.027656	2012-08-02 05:30:32.563513
7056	1-7056	34	75	2564	22	2929	0	0	2012-08-02 05:29:54.031862	2012-08-02 05:30:32.563513
7150	1-7150	45	91	1634	75	5492	0	4	2012-08-02 05:30:32.563513	2012-08-02 05:30:32.563513
6380	1-6380	8	9	5942	100	4414	0	4	2012-08-02 05:25:20.905561	2012-08-02 05:30:32.983663
2710	1-2710	94	9	6513	35	1527	0	0	2012-08-02 05:00:38.43526	2012-08-02 05:30:33.414506
6449	1-6449	9	35	3676	70	7514	0	4	2012-08-02 05:25:48.783314	2012-08-02 05:30:33.414506
7101	1-7101	56	24	563	22	3980	0	0	2012-08-02 05:30:12.133489	2012-08-02 05:30:33.834506
1533	1-1533	90	84	7109	46	3815	0	0	2012-08-02 04:51:54.834537	2012-08-02 05:30:36.783923
7163	1-7163	75	53	2461	71	2640	0	0	2012-08-02 05:30:37.988222	2012-08-02 05:30:37.988222
6876	1-6876	1	8	7139	61	9159	0	0	2012-08-02 05:28:41.572939	2012-08-02 05:30:38.385849
6030	1-6030	7	74	1345	17	737	0	0	2012-08-02 05:22:56.978499	2012-08-02 05:30:38.385849
7164	1-7164	30	17	769	23	7473	0	0	2012-08-02 05:30:38.385849	2012-08-02 05:30:38.385849
6418	1-6418	65	42	5033	36	6101	0	0	2012-08-02 05:25:36.187976	2012-08-02 05:30:39.214775
7166	1-7166	12	36	4094	35	4983	0	5	2012-08-02 05:30:39.214775	2012-08-02 05:30:39.214775
4972	1-4972	93	35	6886	70	7765	4187	0	2012-08-02 05:15:51.326603	2012-08-02 05:30:39.214775
6635	1-6635	70	91	5248	41	5728	0	0	2012-08-02 05:27:04.585958	2012-08-02 05:30:39.645469
7160	1-7160	60	46	3855	84	7629	0	0	2012-08-02 05:30:36.783923	2012-08-02 05:30:40.882657
6358	1-6358	17	96	3261	41	3401	0	0	2012-08-02 05:25:11.934271	2012-08-02 05:30:41.29164
7139	1-7139	79	41	2720	83	3324	0	3	2012-08-02 05:30:28.032692	2012-08-02 05:30:41.29164
7171	1-7171	77	83	3019	3	9667	0	0	2012-08-02 05:30:41.29164	2012-08-02 05:30:41.29164
6934	1-6934	82	71	4862	65	6748	0	7	2012-08-02 05:29:05.554747	2012-08-02 05:30:43.777481
7177	1-7177	92	65	7399	71	7633	0	0	2012-08-02 05:30:43.777481	2012-08-02 05:30:43.777481
6750	1-6750	14	55	1795	53	8451	0	3	2012-08-02 05:27:50.48099	2012-08-02 05:30:44.65002
7179	1-7179	29	100	407	38	2275	0	3	2012-08-02 05:30:44.65002	2012-08-02 05:30:44.65002
6944	1-6944	39	81	4496	3	3431	0	2	2012-08-02 05:29:09.532364	2012-08-02 05:30:45.069905
7185	1-7185	35	27	387	35	2774	0	0	2012-08-02 05:30:47.301708	2012-08-02 05:30:47.301708
6874	1-6874	14	42	1013	53	4042	0	0	2012-08-02 05:28:40.759238	2012-08-02 05:30:48.119035
7116	1-7116	86	6	3056	9	6409	0	0	2012-08-02 05:30:18.375944	2012-08-02 05:30:48.119035
7152	1-7152	35	70	2187	9	5342	0	0	2012-08-02 05:30:33.414506	2012-08-02 05:30:48.119035
791	1-791	39	8	6625	52	2681	0	0	2012-08-02 04:45:28.675434	2012-08-02 05:30:48.119035
3202	1-3202	9	12	7524	53	4733	0	0	2012-08-02 05:04:01.257672	2012-08-02 05:30:48.119035
7193	1-7193	23	37	141	35	3376	0	0	2012-08-02 05:30:50.572156	2012-08-02 05:30:50.572156
6815	1-6815	60	54	7959	38	9045	0	0	2012-08-02 05:28:16.447961	2012-08-02 05:30:52.184729
2432	1-2432	40	55	7391	35	1899	0	0	2012-08-02 04:58:38.626771	2012-08-02 05:30:52.184729
6951	1-6951	46	54	9816	55	9390	0	8	2012-08-02 05:29:12.34137	2012-08-02 05:30:52.184729
7093	1-7093	41	48	1563	31	4487	0	7	2012-08-02 05:30:08.829956	2012-08-02 05:30:52.184729
6843	1-6843	99	58	4765	17	6198	0	4	2012-08-02 05:28:27.928155	2012-08-02 05:30:52.605005
6259	1-6259	35	95	8111	17	5188	0	7	2012-08-02 05:24:30.297141	2012-08-02 05:30:52.605005
7198	1-7198	21	17	2650	4	8564	0	0	2012-08-02 05:30:52.605005	2012-08-02 05:30:52.605005
7195	1-7195	24	22	1394	21	7394	0	0	2012-08-02 05:30:51.389626	2012-08-02 05:30:55.079722
6575	1-6575	81	22	4918	71	4924	0	0	2012-08-02 05:26:39.999931	2012-08-02 05:30:55.079722
6257	1-6257	20	22	3531	14	5749	0	2	2012-08-02 05:24:29.490763	2012-08-02 05:30:55.079722
5944	1-5944	85	22	9498	94	7093	0	0	2012-08-02 05:22:21.18795	2012-08-02 05:30:55.079722
7204	1-7204	38	65	1229	22	4831	0	12	2012-08-02 05:30:55.079722	2012-08-02 05:30:55.079722
7205	1-7205	84	16	294	55	5271	0	0	2012-08-02 05:30:55.511062	2012-08-02 05:30:55.511062
6956	1-6956	41	77	4377	59	9539	0	0	2012-08-02 05:29:14.274858	2012-08-02 05:30:57.963396
5637	1-5637	65	77	6606	3	4375	0	0	2012-08-02 05:20:16.176125	2012-08-02 05:30:57.963396
7211	1-7211	87	63	3011	77	7851	0	11	2012-08-02 05:30:57.963396	2012-08-02 05:30:57.963396
2716	1-2716	84	32	7723	69	248	248	0	2012-08-02 05:00:40.998731	2012-08-02 05:30:57.963396
6682	1-6682	50	77	2444	4	2057	0	0	2012-08-02 05:27:23.373034	2012-08-02 05:30:58.372281
6759	1-6759	47	40	8383	9	3281	0	0	2012-08-02 05:27:54.071917	2012-08-02 05:30:58.781154
7213	1-7213	34	56	369	40	9330	0	8	2012-08-02 05:30:58.781154	2012-08-02 05:30:58.781154
7133	1-7133	82	34	4160	49	2908	0	0	2012-08-02 05:30:25.513037	2012-08-02 05:30:59.202979
7143	1-7143	79	8	1931	18	459	0	3	2012-08-02 05:30:29.691173	2012-08-02 05:30:59.202979
7039	1-7039	47	34	7346	8	7930	0	0	2012-08-02 05:29:47.325487	2012-08-02 05:30:59.202979
7218	1-7218	94	76	70	21	5504	0	0	2012-08-02 05:31:00.915835	2012-08-02 05:31:00.915835
3090	1-3090	68	8	8930	76	8939	6591	0	2012-08-02 05:03:15.769342	2012-08-02 05:31:00.915835
7131	1-7131	28	85	1447	37	6610	0	17	2012-08-02 05:30:24.695435	2012-08-02 05:31:01.31334
7219	1-7219	44	37	1774	43	4997	0	0	2012-08-02 05:31:01.31334	2012-08-02 05:31:01.31334
7173	1-7173	50	71	4632	25	6052	0	0	2012-08-02 05:30:42.131276	2012-08-02 05:31:03.77677
7114	1-7114	93	100	690	31	904	0	0	2012-08-02 05:30:17.547301	2012-08-02 05:31:04.605383
7227	1-7227	89	89	464	100	8190	0	0	2012-08-02 05:31:04.605383	2012-08-02 05:31:04.605383
6644	1-6644	21	11	859	57	3550	0	3	2012-08-02 05:27:08.180332	2012-08-02 05:31:05.014306
4403	1-4403	40	12	4002	57	6950	0	0	2012-08-02 05:12:03.427804	2012-08-02 05:31:05.014306
5912	1-5912	98	48	8581	25	2596	0	7	2012-08-02 05:22:07.852851	2012-08-02 05:31:05.855237
6995	1-6995	71	28	3988	30	3129	0	0	2012-08-02 05:29:29.745263	2012-08-02 05:31:05.855237
7230	1-7230	10	94	249	48	6023	0	13	2012-08-02 05:31:05.855237	2012-08-02 05:31:05.855237
5783	1-5783	44	94	3983	48	6621	997	4	2012-08-02 05:21:15.744117	2012-08-02 05:31:05.855237
7214	1-7214	89	27	2086	34	9002	0	0	2012-08-02 05:30:59.202979	2012-08-02 05:31:06.264442
3252	1-3252	48	35	7653	40	7817	0	4	2012-08-02 05:04:21.265625	2012-08-02 05:31:06.662317
7053	1-7053	12	35	5627	91	5190	0	8	2012-08-02 05:29:52.850075	2012-08-02 05:31:06.662317
3538	1-3538	72	35	9729	34	9451	0	0	2012-08-02 05:06:16.998816	2012-08-02 05:31:06.662317
6797	1-6797	78	66	4971	26	6925	0	12	2012-08-02 05:28:09.076678	2012-08-02 05:31:06.662317
7091	1-7091	48	66	4407	35	9023	0	0	2012-08-02 05:30:08.01239	2012-08-02 05:31:06.662317
7197	1-7197	73	31	2637	54	8842	0	7	2012-08-02 05:30:52.184729	2012-08-02 05:31:10.97137
5842	1-5842	18	13	7440	31	7191	0	0	2012-08-02 05:21:39.400499	2012-08-02 05:31:10.97137
7243	1-7243	6	92	3640	13	6713	0	0	2012-08-02 05:31:10.97137	2012-08-02 05:31:10.97137
5444	1-5444	46	62	3109	54	3273	0	0	2012-08-02 05:18:58.873516	2012-08-02 05:31:11.755453
7245	1-7245	50	93	844	79	8289	0	0	2012-08-02 05:31:11.755453	2012-08-02 05:31:11.755453
6023	1-6023	13	16	5998	38	2296	0	5	2012-08-02 05:22:54.095068	2012-08-02 05:31:12.915478
7248	1-7248	56	87	788	16	8733	0	7	2012-08-02 05:31:12.915478	2012-08-02 05:31:12.915478
6381	1-6381	83	71	7737	43	7529	0	6	2012-08-02 05:25:21.303437	2012-08-02 05:31:14.826831
6433	1-6433	35	4	9334	9	7048	0	0	2012-08-02 05:25:42.386378	2012-08-02 05:31:14.826831
7253	1-7253	82	43	1402	71	5907	0	0	2012-08-02 05:31:14.826831	2012-08-02 05:31:14.826831
6405	1-6405	99	66	1953	14	5554	0	0	2012-08-02 05:25:30.928415	2012-08-02 05:31:16.727064
7232	1-7232	32	40	9078	66	8436	0	0	2012-08-02 05:31:06.662317	2012-08-02 05:31:16.727064
6753	1-6753	23	40	3345	98	726	0	0	2012-08-02 05:27:51.674493	2012-08-02 05:31:16.727064
7258	1-7258	95	59	342	40	9763	0	0	2012-08-02 05:31:16.727064	2012-08-02 05:31:16.727064
6807	1-6807	51	47	441	92	6117	0	4	2012-08-02 05:28:13.164874	2012-08-02 05:31:17.533987
3149	1-3149	89	9	5651	53	5772	0	3	2012-08-02 05:03:39.769998	2012-08-02 05:31:18.704954
7263	1-7263	18	53	3782	9	8713	0	6	2012-08-02 05:31:18.704954	2012-08-02 05:31:18.704954
7077	1-7077	3	6	5762	12	9093	0	3	2012-08-02 05:30:02.396483	2012-08-02 05:31:19.898305
7266	1-7266	1	16	3049	6	5372	0	0	2012-08-02 05:31:19.898305	2012-08-02 05:31:19.898305
6830	1-6830	1	10	7603	28	7669	0	0	2012-08-02 05:28:22.580193	2012-08-02 05:31:20.295902
7083	1-7083	90	10	5435	40	9507	0	0	2012-08-02 05:30:04.762712	2012-08-02 05:31:20.295902
7267	1-7267	58	95	394	10	7683	0	2	2012-08-02 05:31:20.295902	2012-08-02 05:31:20.295902
6144	1-6144	71	31	2469	51	1369	0	0	2012-08-02 05:23:42.945537	2012-08-02 05:31:20.693825
4895	1-4895	9	65	8230	50	8173	0	3	2012-08-02 05:15:21.207052	2012-08-02 05:31:21.080457
7269	1-7269	73	58	2703	4	5389	0	0	2012-08-02 05:31:21.080457	2012-08-02 05:31:21.080457
6616	1-6616	91	50	9595	9	6509	1941	5	2012-08-02 05:26:56.774797	2012-08-02 05:31:21.080457
7268	1-7268	35	55	455	31	4171	0	4	2012-08-02 05:31:20.693825	2012-08-02 05:31:23.05801
7175	1-7175	50	25	4713	81	7215	0	0	2012-08-02 05:30:42.948967	2012-08-02 05:31:23.842295
7276	1-7276	20	81	1605	46	6051	0	0	2012-08-02 05:31:23.842295	2012-08-02 05:31:23.842295
6473	1-6473	58	91	1240	34	1128	0	0	2012-08-02 05:25:58.275644	2012-08-02 05:31:25.390895
1972	1-1972	31	91	4728	97	2980	0	3	2012-08-02 04:55:18.838107	2012-08-02 05:31:25.390895
6047	1-6047	63	97	6462	11	7307	0	2	2012-08-02 05:23:04.126582	2012-08-02 05:31:25.788852
7107	1-7107	65	23	1979	32	3057	0	0	2012-08-02 05:30:14.652631	2012-08-02 05:31:26.562267
3221	1-3221	8	65	6044	46	4854	0	0	2012-08-02 05:04:09.066657	2012-08-02 05:31:26.562267
7283	1-7283	75	71	1093	65	8552	0	9	2012-08-02 05:31:26.562267	2012-08-02 05:31:26.562267
3642	1-3642	5	44	9669	80	6262	0	4	2012-08-02 05:06:58.910547	2012-08-02 05:31:26.960187
7080	1-7080	4	79	3981	37	5435	0	9	2012-08-02 05:30:03.580723	2012-08-02 05:31:26.960187
6505	1-6505	36	47	5011	36	7903	0	10	2012-08-02 05:26:11.171188	2012-08-02 05:31:27.369492
7285	1-7285	29	15	1663	25	4601	0	12	2012-08-02 05:31:27.369492	2012-08-02 05:31:27.369492
6854	1-6854	40	14	7950	17	8624	0	0	2012-08-02 05:28:32.491313	2012-08-02 05:31:29.302338
7290	1-7290	96	17	2534	69	6948	0	0	2012-08-02 05:31:29.302338	2012-08-02 05:31:29.302338
7291	1-7291	24	8	248	99	9349	0	7	2012-08-02 05:31:29.689535	2012-08-02 05:31:29.689535
7037	1-7037	10	62	3506	16	4446	0	5	2012-08-02 05:29:46.530005	2012-08-02 05:31:30.087054
6686	1-6686	36	36	1188	57	1726	0	3	2012-08-02 05:27:24.986292	2012-08-02 05:31:30.473817
7293	1-7293	97	57	4316	86	3651	0	0	2012-08-02 05:31:30.473817	2012-08-02 05:31:30.473817
6799	1-6799	41	76	3468	25	2908	0	0	2012-08-02 05:28:09.894263	2012-08-02 05:31:32.053642
7111	1-7111	89	100	9642	27	8513	0	4	2012-08-02 05:30:16.310449	2012-08-02 05:31:32.815957
7299	1-7299	55	96	513	95	1830	0	2	2012-08-02 05:31:32.815957	2012-08-02 05:31:32.815957
7300	1-7300	13	12	163	41	8319	0	0	2012-08-02 05:31:33.213673	2012-08-02 05:31:33.213673
7304	1-7304	26	33	650	23	3065	0	9	2012-08-02 05:31:34.771471	2012-08-02 05:31:34.771471
6487	1-6487	72	94	8333	43	8228	0	6	2012-08-02 05:26:03.844515	2012-08-02 05:31:35.566866
6489	1-6489	76	88	7126	67	3571	0	4	2012-08-02 05:26:04.640165	2012-08-02 05:31:35.975855
7312	1-7312	14	100	1033	44	3969	0	0	2012-08-02 05:31:37.953234	2012-08-02 05:31:37.953234
7292	1-7292	4	16	3613	36	8264	0	0	2012-08-02 05:31:30.087054	2012-08-02 05:31:38.351022
7319	1-7319	97	91	3486	41	5830	0	6	2012-08-02 05:31:40.726436	2012-08-02 05:31:40.726436
5790	1-5790	30	75	4463	37	5995	0	0	2012-08-02 05:21:18.517102	2012-08-02 05:31:41.499403
7322	1-7322	82	42	1845	16	6392	0	0	2012-08-02 05:31:41.886097	2012-08-02 05:31:41.886097
5319	1-5319	59	82	9108	70	8117	0	6	2012-08-02 05:18:09.723228	2012-08-02 05:31:42.670479
7182	1-7182	58	27	6089	30	6379	0	0	2012-08-02 05:30:45.94298	2012-08-02 05:31:42.670479
7324	1-7324	69	59	2974	27	7750	0	7	2012-08-02 05:31:42.670479	2012-08-02 05:31:42.670479
1209	1-1209	56	60	3921	80	2433	0	8	2012-08-02 04:49:04.553473	2012-08-02 05:31:44.217285
7328	1-7328	4	79	61	60	2539	0	0	2012-08-02 05:31:44.217285	2012-08-02 05:31:44.217285
6261	1-6261	75	90	9672	76	7243	0	3	2012-08-02 05:24:31.159135	2012-08-02 05:31:44.604425
7167	1-7167	54	41	5445	91	8674	0	0	2012-08-02 05:30:39.645469	2012-08-02 05:31:45.001852
7335	1-7335	73	50	827	39	808	0	0	2012-08-02 05:31:46.990396	2012-08-02 05:31:46.990396
6725	1-6725	75	26	5654	68	8623	0	0	2012-08-02 05:27:40.612505	2012-08-02 05:31:48.172439
6936	1-6936	10	20	4673	55	6849	0	0	2012-08-02 05:29:06.361239	2012-08-02 05:31:48.172439
7338	1-7338	81	55	3591	91	8792	0	0	2012-08-02 05:31:48.172439	2012-08-02 05:31:48.172439
7161	1-7161	32	51	1576	68	1094	0	7	2012-08-02 05:30:37.182092	2012-08-02 05:31:48.581492
2876	1-2876	57	8	5598	70	3986	0	0	2012-08-02 05:01:48.236088	2012-08-02 05:31:48.581492
7284	1-7284	96	11	2354	79	8028	0	0	2012-08-02 05:31:26.960187	2012-08-02 05:31:48.581492
5397	1-5397	24	8	9811	35	5994	0	0	2012-08-02 05:18:40.351373	2012-08-02 05:31:48.581492
7339	1-7339	17	68	2028	8	8247	0	3	2012-08-02 05:31:48.581492	2012-08-02 05:31:48.581492
6133	1-6133	45	4	6792	95	6717	0	7	2012-08-02 05:23:38.524491	2012-08-02 05:31:49.001455
7340	1-7340	28	95	367	87	7347	0	7	2012-08-02 05:31:49.001455	2012-08-02 05:31:49.001455
6232	1-6232	94	18	1925	100	3354	0	0	2012-08-02 05:24:19.01567	2012-08-02 05:31:49.377264
7306	1-7306	62	18	502	94	871	0	0	2012-08-02 05:31:35.566866	2012-08-02 05:31:49.377264
7030	1-7030	75	18	4255	71	4947	0	0	2012-08-02 05:29:43.679529	2012-08-02 05:31:49.377264
7313	1-7313	87	18	541	6	2580	0	0	2012-08-02 05:31:38.351022	2012-08-02 05:31:49.377264
6060	1-6060	4	60	7343	70	7349	0	0	2012-08-02 05:23:09.354225	2012-08-02 05:31:50.581363
5595	1-5595	79	43	9509	29	8894	0	0	2012-08-02 05:19:59.102593	2012-08-02 05:31:50.581363
7329	1-7329	25	100	3189	90	6966	0	4	2012-08-02 05:31:44.604425	2012-08-02 05:31:50.979393
7231	1-7231	13	95	7639	100	9614	0	2	2012-08-02 05:31:06.264442	2012-08-02 05:31:50.979393
6985	1-6985	10	2	8445	67	6330	0	2	2012-08-02 05:29:25.845038	2012-08-02 05:31:51.366117
7348	1-7348	41	26	289	95	6032	0	7	2012-08-02 05:31:52.13936	2012-08-02 05:31:52.13936
6950	1-6950	28	82	1724	57	3852	0	3	2012-08-02 05:29:11.954918	2012-08-02 05:31:52.948118
7321	1-7321	72	37	7151	75	5451	0	0	2012-08-02 05:31:41.499403	2012-08-02 05:31:56.836763
7360	1-7360	14	58	409	37	7964	0	0	2012-08-02 05:31:56.836763	2012-08-02 05:31:56.836763
2420	1-2420	47	72	2017	23	772	0	0	2012-08-02 04:58:33.323463	2012-08-02 05:31:57.234403
7086	1-7086	45	83	4182	23	1959	0	0	2012-08-02 05:30:05.968545	2012-08-02 05:31:57.234403
7361	1-7361	4	23	1379	98	9948	0	0	2012-08-02 05:31:57.234403	2012-08-02 05:31:57.234403
7278	1-7278	90	57	2021	7	1259	0	0	2012-08-02 05:31:24.62896	2012-08-02 05:31:59.211883
6564	1-6564	10	11	9466	87	9968	0	0	2012-08-02 05:26:35.503161	2012-08-02 05:31:59.211883
6762	1-6762	64	37	5520	7	2551	0	0	2012-08-02 05:27:55.254182	2012-08-02 05:31:59.211883
5708	1-5708	51	64	5131	39	3529	0	0	2012-08-02 05:20:44.817044	2012-08-02 05:31:59.211883
7368	1-7368	27	100	1064	47	9714	0	0	2012-08-02 05:32:00.007501	2012-08-02 05:32:00.007501
7344	1-7344	21	70	2155	43	6843	0	0	2012-08-02 05:31:50.581363	2012-08-02 05:32:00.405367
306	1-306	17	36	9676	23	2274	0	0	2012-08-02 04:43:59.124049	2012-08-02 05:32:00.814102
3428	1-3428	70	19	3562	70	2624	0	0	2012-08-02 05:05:32.600851	2012-08-02 05:32:01.598436
7372	1-7372	62	70	225	21	3387	0	0	2012-08-02 05:32:01.598436	2012-08-02 05:32:01.598436
7373	1-7373	12	10	1866	96	8939	0	0	2012-08-02 05:32:01.985282	2012-08-02 05:32:01.985282
7376	1-7376	2	21	1247	82	3063	0	0	2012-08-02 05:32:03.189366	2012-08-02 05:32:03.189366
7280	1-7280	8	29	2328	41	4895	0	6	2012-08-02 05:31:25.390895	2012-08-02 05:32:04.36045
7238	1-7238	54	46	9627	78	8492	0	5	2012-08-02 05:31:09.015844	2012-08-02 05:32:04.36045
6328	1-6328	100	36	2394	92	2131	0	0	2012-08-02 05:24:59.445195	2012-08-02 05:32:05.16733
7381	1-7381	35	92	492	36	4829	0	3	2012-08-02 05:32:05.16733	2012-08-02 05:32:05.16733
5229	1-5229	75	75	8176	1	5685	0	2	2012-08-02 05:17:34.398482	2012-08-02 05:32:05.554516
6767	1-6767	85	58	7047	46	8758	0	0	2012-08-02 05:27:57.242693	2012-08-02 05:32:05.554516
7032	1-7032	35	45	2912	80	9166	0	4	2012-08-02 05:29:44.54131	2012-08-02 05:32:06.726205
5389	1-5389	31	97	9288	79	7916	0	4	2012-08-02 05:18:37.235698	2012-08-02 05:32:06.726205
3665	1-3665	77	6	7461	45	3638	0	3	2012-08-02 05:07:08.247561	2012-08-02 05:32:06.726205
7387	1-7387	59	10	335	34	4032	0	0	2012-08-02 05:32:07.576776	2012-08-02 05:32:07.576776
5999	1-5999	77	87	7950	31	7157	0	2	2012-08-02 05:22:44.038338	2012-08-02 05:32:08.79185
7326	1-7326	7	31	4162	72	7436	0	6	2012-08-02 05:31:43.455178	2012-08-02 05:32:08.79185
7366	1-7366	13	31	3040	64	7932	0	6	2012-08-02 05:31:59.211883	2012-08-02 05:32:08.79185
5185	1-5185	39	86	7336	28	4597	0	8	2012-08-02 05:17:17.17125	2012-08-02 05:32:09.211983
7362	1-7362	75	23	1672	35	4306	0	0	2012-08-02 05:31:57.643263	2012-08-02 05:32:11.686392
6109	1-6109	4	12	8492	99	2804	0	3	2012-08-02 05:23:28.934009	2012-08-02 05:32:11.686392
7391	1-7391	47	48	5445	86	4078	0	4	2012-08-02 05:32:09.211983	2012-08-02 05:32:12.11733
7259	1-7259	100	28	6339	12	8573	0	0	2012-08-02 05:31:17.136566	2012-08-02 05:32:12.548499
7057	1-7057	32	97	4180	62	9366	0	5	2012-08-02 05:29:54.440796	2012-08-02 05:32:12.548499
7350	1-7350	86	57	3907	26	5073	0	15	2012-08-02 05:31:52.948118	2012-08-02 05:32:12.548499
7112	1-7112	10	48	4323	62	3077	0	4	2012-08-02 05:30:16.730006	2012-08-02 05:32:13.376932
7401	1-7401	29	34	98	48	4675	0	0	2012-08-02 05:32:13.376932	2012-08-02 05:32:13.376932
7151	1-7151	90	33	3357	8	2452	0	7	2012-08-02 05:30:32.983663	2012-08-02 05:32:13.785844
7402	1-7402	93	51	876	70	9017	0	0	2012-08-02 05:32:13.785844	2012-08-02 05:32:13.785844
7294	1-7294	21	45	3529	97	7486	0	13	2012-08-02 05:31:30.882525	2012-08-02 05:32:15.023199
7148	1-7148	62	45	1711	93	4251	0	0	2012-08-02 05:30:31.734848	2012-08-02 05:32:15.023199
6093	1-6093	23	50	8947	4	7389	0	11	2012-08-02 05:23:22.570266	2012-08-02 05:32:15.023199
6643	1-6643	77	50	7919	73	8821	0	6	2012-08-02 05:27:07.782585	2012-08-02 05:32:15.023199
6765	1-6765	99	45	3973	50	5680	0	8	2012-08-02 05:27:56.447172	2012-08-02 05:32:15.023199
7405	1-7405	36	26	1480	45	6512	0	0	2012-08-02 05:32:15.023199	2012-08-02 05:32:15.023199
6459	1-6459	84	80	8426	11	5814	0	5	2012-08-02 05:25:52.738474	2012-08-02 05:32:15.874383
7408	1-7408	93	27	430	88	7708	0	0	2012-08-02 05:32:16.282869	2012-08-02 05:32:16.282869
7297	1-7297	8	73	4189	76	5093	0	4	2012-08-02 05:31:32.053642	2012-08-02 05:32:17.122418
6939	1-6939	64	24	1202	76	1075	0	0	2012-08-02 05:29:07.5324	2012-08-02 05:32:17.122418
6573	1-6573	88	24	3233	84	8767	0	12	2012-08-02 05:26:39.182863	2012-08-02 05:32:17.122418
7410	1-7410	75	39	2168	24	7060	0	2	2012-08-02 05:32:17.122418	2012-08-02 05:32:17.122418
6323	1-6323	22	12	3081	35	1067	0	0	2012-08-02 05:24:57.368092	2012-08-02 05:32:17.973282
7397	1-7397	82	99	2975	23	5915	0	19	2012-08-02 05:32:11.686392	2012-08-02 05:32:18.382144
7389	1-7389	17	2	680	40	4530	0	0	2012-08-02 05:32:08.394237	2012-08-02 05:32:19.232845
7041	1-7041	33	23	5023	8	8471	0	6	2012-08-02 05:29:48.099244	2012-08-02 05:32:19.654471
6902	1-6902	25	9	4170	58	8345	0	0	2012-08-02 05:28:52.280186	2012-08-02 05:32:21.68727
7421	1-7421	26	58	117	19	6808	0	4	2012-08-02 05:32:21.68727	2012-08-02 05:32:21.68727
7423	1-7423	6	92	1034	83	601	0	0	2012-08-02 05:32:22.50512	2012-08-02 05:32:22.50512
5651	1-5651	14	14	4566	59	1623	0	8	2012-08-02 05:20:21.855112	2012-08-02 05:32:29.044778
7007	1-7007	99	19	4447	43	9055	0	4	2012-08-02 05:29:34.532292	2012-08-02 05:32:29.044778
7170	1-7170	95	84	8785	4	9873	0	0	2012-08-02 05:30:40.882657	2012-08-02 05:32:30.702118
7016	1-7016	16	16	6304	61	3831	0	0	2012-08-02 05:29:38.12307	2012-08-02 05:32:30.702118
4552	1-4552	38	60	8099	30	1981	0	0	2012-08-02 05:13:02.4334	2012-08-02 05:32:31.519347
7445	1-7445	75	93	1162	60	9162	0	0	2012-08-02 05:32:31.519347	2012-08-02 05:32:31.519347
7390	1-7390	97	72	489	87	6399	0	0	2012-08-02 05:32:08.79185	2012-08-02 05:32:31.939309
4452	1-4452	46	58	9364	44	4798	0	0	2012-08-02 05:12:22.731051	2012-08-02 05:32:32.403289
6842	1-6842	92	49	1960	8	1334	0	3	2012-08-02 05:28:27.519363	2012-08-02 05:32:32.403289
7448	1-7448	64	50	6011	24	4743	0	0	2012-08-02 05:32:32.823161	2012-08-02 05:32:32.823161
4105	1-4105	81	26	9678	87	7447	0	4	2012-08-02 05:10:03.671465	2012-08-02 05:32:33.618614
2842	1-2842	25	26	7529	87	4982	0	0	2012-08-02 05:01:34.014866	2012-08-02 05:32:33.618614
7450	1-7450	54	87	2662	26	9397	0	0	2012-08-02 05:32:33.618614	2012-08-02 05:32:33.618614
7453	1-7453	14	39	230	40	4281	0	10	2012-08-02 05:32:34.833853	2012-08-02 05:32:34.833853
5991	1-5991	39	12	4719	38	2628	0	0	2012-08-02 05:22:40.66813	2012-08-02 05:32:35.253765
7454	1-7454	29	38	2048	12	6952	0	0	2012-08-02 05:32:35.253765	2012-08-02 05:32:35.253765
6262	1-6262	73	2	9195	32	3542	0	0	2012-08-02 05:24:31.567472	2012-08-02 05:32:35.651589
3558	1-3558	88	52	7521	2	8086	4902	0	2012-08-02 05:06:25.019514	2012-08-02 05:32:35.651589
6784	1-6784	94	97	7643	19	3278	0	0	2012-08-02 05:28:03.882558	2012-08-02 05:32:36.480263
7460	1-7460	28	86	527	35	8719	0	0	2012-08-02 05:32:37.706616	2012-08-02 05:32:37.706616
7462	1-7462	37	8	59	18	9891	0	4	2012-08-02 05:32:38.535287	2012-08-02 05:32:38.535287
7466	1-7466	97	74	462	64	5162	0	6	2012-08-02 05:32:40.147852	2012-08-02 05:32:40.147852
5896	1-5896	80	87	8702	16	4076	0	4	2012-08-02 05:22:01.344124	2012-08-02 05:32:42.644457
5898	1-5898	100	23	6996	28	7938	0	0	2012-08-02 05:22:02.128364	2012-08-02 05:32:43.053495
7473	1-7473	8	17	2648	23	7048	0	4	2012-08-02 05:32:43.053495	2012-08-02 05:32:43.053495
7477	1-7477	49	70	416	47	985	0	0	2012-08-02 05:32:44.677367	2012-08-02 05:32:44.677367
5869	1-5869	18	32	6059	3	3334	0	6	2012-08-02 05:21:50.448512	2012-08-02 05:32:45.903526
4068	1-4068	60	32	6897	30	2460	0	4	2012-08-02 05:09:48.930759	2012-08-02 05:32:45.903526
3476	1-3476	74	38	5688	67	2532	0	11	2012-08-02 05:05:51.981892	2012-08-02 05:32:45.903526
7480	1-7480	81	73	2045	32	6281	0	0	2012-08-02 05:32:45.903526	2012-08-02 05:32:45.903526
3524	1-3524	51	33	8568	23	3667	1768	0	2012-08-02 05:06:11.41903	2012-08-02 05:32:45.903526
4108	1-4108	7	33	6272	23	2321	2321	0	2012-08-02 05:10:04.909599	2012-08-02 05:32:45.903526
5756	1-5756	67	18	8296	38	9529	0	0	2012-08-02 05:21:04.760725	2012-08-02 05:32:46.71036
6579	1-6579	39	38	5442	87	7778	0	0	2012-08-02 05:26:41.637138	2012-08-02 05:32:46.71036
3677	1-3677	8	18	8142	25	9737	0	0	2012-08-02 05:07:13.020694	2012-08-02 05:32:46.71036
7482	1-7482	13	87	3891	18	8094	0	6	2012-08-02 05:32:46.71036	2012-08-02 05:32:46.71036
7144	1-7144	55	3	4375	96	3707	0	4	2012-08-02 05:30:30.099873	2012-08-02 05:32:47.108283
5626	1-5626	95	19	3988	51	9001	0	0	2012-08-02 05:20:11.690623	2012-08-02 05:32:47.539143
7484	1-7484	72	70	803	18	2597	0	0	2012-08-02 05:32:47.539143	2012-08-02 05:32:47.539143
2241	1-2241	72	18	8853	16	9026	3390	0	2012-08-02 04:57:15.283363	2012-08-02 05:32:47.539143
2123	1-2123	34	59	8232	57	4831	0	0	2012-08-02 04:56:24.125766	2012-08-02 05:32:47.948154
6873	1-6873	8	70	6039	42	8586	0	2	2012-08-02 05:28:40.192064	2012-08-02 05:32:47.948154
7353	1-7353	90	70	2962	33	2775	0	4	2012-08-02 05:31:54.118886	2012-08-02 05:32:48.35685
6748	1-6748	44	28	8613	8	8838	0	12	2012-08-02 05:27:49.683686	2012-08-02 05:32:48.765669
7178	1-7178	95	16	7493	30	3766	0	0	2012-08-02 05:30:44.24147	2012-08-02 05:32:49.980879
7379	1-7379	78	78	3882	86	8957	0	0	2012-08-02 05:32:04.36045	2012-08-02 05:32:51.218323
7121	1-7121	6	39	2664	35	2908	0	0	2012-08-02 05:30:20.431006	2012-08-02 05:32:51.218323
7493	1-7493	35	9	996	39	9301	0	7	2012-08-02 05:32:51.218323	2012-08-02 05:32:51.218323
7416	1-7416	45	27	3891	23	7516	0	7	2012-08-02 05:32:19.654471	2012-08-02 05:32:52.03821
1892	1-1892	45	75	4195	100	3250	0	0	2012-08-02 04:54:44.089261	2012-08-02 05:32:52.03821
7180	1-7180	83	32	4205	71	6368	0	0	2012-08-02 05:30:45.069905	2012-08-02 05:32:54.535048
7382	1-7382	67	46	7220	75	9286	0	0	2012-08-02 05:32:05.554516	2012-08-02 05:32:56.589398
7145	1-7145	15	86	5562	43	9035	0	2	2012-08-02 05:30:30.509015	2012-08-02 05:32:57.020447
7129	1-7129	42	14	3058	19	682	0	3	2012-08-02 05:30:23.778555	2012-08-02 05:32:57.020447
7508	1-7508	46	6	868	96	8331	0	10	2012-08-02 05:32:57.451255	2012-08-02 05:32:57.451255
7509	1-7509	19	89	11	38	9837	0	0	2012-08-02 05:32:57.860067	2012-08-02 05:32:57.860067
5060	1-5060	87	40	8704	87	5015	0	0	2012-08-02 05:16:26.654127	2012-08-02 05:32:58.677573
7511	1-7511	70	87	1750	27	4474	0	0	2012-08-02 05:32:58.677573	2012-08-02 05:32:58.677573
7370	1-7370	31	23	1299	83	5455	0	6	2012-08-02 05:32:00.814102	2012-08-02 05:32:59.086407
7488	1-7488	18	16	4034	46	4002	0	0	2012-08-02 05:32:49.174437	2012-08-02 05:32:59.495215
7103	1-7103	90	69	4064	52	5452	0	0	2012-08-02 05:30:12.950979	2012-08-02 05:32:59.495215
6994	1-6994	16	13	4082	7	4415	0	8	2012-08-02 05:29:29.347239	2012-08-02 05:32:59.495215
6745	1-6745	14	83	3841	36	4422	0	0	2012-08-02 05:27:48.501358	2012-08-02 05:33:00.324014
7184	1-7184	37	36	5031	67	3139	0	0	2012-08-02 05:30:46.837877	2012-08-02 05:33:00.324014
7515	1-7515	53	67	931	18	5817	0	0	2012-08-02 05:33:00.324014	2012-08-02 05:33:00.324014
7399	1-7399	44	62	3454	28	8109	0	0	2012-08-02 05:32:12.548499	2012-08-02 05:33:01.152719
3682	1-3682	7	82	3079	80	6449	0	0	2012-08-02 05:07:15.031376	2012-08-02 05:33:01.152719
7485	1-7485	99	57	8677	70	9680	0	0	2012-08-02 05:32:47.948154	2012-08-02 05:33:01.152719
7517	1-7517	57	98	841	82	9798	0	0	2012-08-02 05:33:01.152719	2012-08-02 05:33:01.152719
7225	1-7225	33	25	6690	26	6994	0	0	2012-08-02 05:31:03.77677	2012-08-02 05:33:01.981709
7124	1-7124	11	72	5285	53	3385	0	6	2012-08-02 05:30:21.701477	2012-08-02 05:33:02.467513
4955	1-4955	67	72	9884	19	2965	0	18	2012-08-02 05:15:44.642592	2012-08-02 05:33:02.467513
7520	1-7520	51	97	4269	72	8743	0	2	2012-08-02 05:33:02.467513	2012-08-02 05:33:02.467513
7522	1-7522	26	74	81	95	5336	0	4	2012-08-02 05:33:03.285094	2012-08-02 05:33:03.285094
7524	1-7524	86	54	50	28	7776	0	10	2012-08-02 05:33:04.102534	2012-08-02 05:33:04.102534
5802	1-5802	25	2	7782	21	2166	0	0	2012-08-02 05:21:23.289643	2012-08-02 05:33:04.500537
6805	1-6805	96	69	3578	63	5406	0	0	2012-08-02 05:28:12.347337	2012-08-02 05:33:06.20278
7351	1-7351	56	49	3568	58	8442	0	4	2012-08-02 05:31:53.3456	2012-08-02 05:33:06.20278
7385	1-7385	55	44	8948	6	9779	0	12	2012-08-02 05:32:06.726205	2012-08-02 05:33:07.009623
5854	1-5854	99	44	8702	76	4015	0	0	2012-08-02 05:21:44.405224	2012-08-02 05:33:07.009623
4665	1-4665	12	92	6125	16	875	0	0	2012-08-02 05:13:48.609275	2012-08-02 05:33:07.009623
7229	1-7229	34	44	7336	75	3434	0	0	2012-08-02 05:31:05.435934	2012-08-02 05:33:07.009623
7531	1-7531	50	27	1977	44	9871	0	0	2012-08-02 05:33:07.009623	2012-08-02 05:33:07.009623
5775	1-5775	21	16	6910	27	1088	944	0	2012-08-02 05:21:12.518316	2012-08-02 05:33:07.009623
7187	1-7187	76	53	9120	8	9966	0	3	2012-08-02 05:30:48.119035	2012-08-02 05:33:07.860403
7102	1-7102	87	17	2715	92	3984	0	0	2012-08-02 05:30:12.564333	2012-08-02 05:33:08.688974
7529	1-7529	31	63	8843	49	9933	0	0	2012-08-02 05:33:06.20278	2012-08-02 05:33:08.688974
7535	1-7535	97	92	652	63	8254	0	0	2012-08-02 05:33:08.688974	2012-08-02 05:33:08.688974
6960	1-6960	82	67	8511	19	8439	0	0	2012-08-02 05:29:15.866469	2012-08-02 05:33:09.893064
7012	1-7012	7	67	2256	57	5627	0	7	2012-08-02 05:29:36.554434	2012-08-02 05:33:09.893064
5666	1-5666	63	67	7878	25	4605	0	7	2012-08-02 05:20:27.810232	2012-08-02 05:33:09.893064
7538	1-7538	97	31	1404	67	7959	0	6	2012-08-02 05:33:09.893064	2012-08-02 05:33:09.893064
6984	1-6984	93	4	7978	65	7803	0	0	2012-08-02 05:29:25.435992	2012-08-02 05:33:11.130555
7487	1-7487	73	27	4397	28	6350	0	0	2012-08-02 05:32:48.765669	2012-08-02 05:33:13.185325
7546	1-7546	81	8	430	27	9570	0	0	2012-08-02 05:33:13.185325	2012-08-02 05:33:13.185325
6448	1-6448	50	28	6887	8	3165	906	10	2012-08-02 05:25:48.385564	2012-08-02 05:33:13.185325
7095	1-7095	93	70	3025	59	5231	0	2	2012-08-02 05:30:09.669568	2012-08-02 05:33:13.583326
6422	1-6422	9	70	7035	45	3975	0	0	2012-08-02 05:25:37.867622	2012-08-02 05:33:13.583326
4600	1-4600	63	63	9840	15	8974	0	0	2012-08-02 05:13:21.913296	2012-08-02 05:33:14.798336
7550	1-7550	86	15	2183	34	9302	0	9	2012-08-02 05:33:14.798336	2012-08-02 05:33:14.798336
7554	1-7554	21	5	229	32	6199	0	0	2012-08-02 05:33:16.455413	2012-08-02 05:33:16.455413
7174	1-7174	8	79	6400	44	8632	0	0	2012-08-02 05:30:42.540029	2012-08-02 05:33:18.068329
7558	1-7558	7	95	999	79	2978	0	0	2012-08-02 05:33:18.068329	2012-08-02 05:33:18.068329
7506	1-7506	92	75	5971	52	4981	0	0	2012-08-02 05:32:56.589398	2012-08-02 05:33:19.261753
7561	1-7561	58	1	263	78	8101	0	0	2012-08-02 05:33:19.261753	2012-08-02 05:33:19.261753
5408	1-5408	17	74	5856	34	3169	0	0	2012-08-02 05:18:44.672942	2012-08-02 05:33:22.920275
6362	1-6362	86	15	7008	34	1225	0	6	2012-08-02 05:25:13.547315	2012-08-02 05:33:23.340265
6069	1-6069	47	80	9526	44	8972	0	0	2012-08-02 05:23:12.955737	2012-08-02 05:33:23.340265
2875	1-2875	69	69	8330	9	3895	0	0	2012-08-02 05:01:47.815916	2012-08-02 05:33:26.157384
7578	1-7578	10	9	357	26	4165	0	0	2012-08-02 05:33:26.157384	2012-08-02 05:33:26.157384
7271	1-7271	73	58	7886	74	8495	0	11	2012-08-02 05:31:21.88699	2012-08-02 05:33:27.008108
7474	1-7474	31	78	3978	75	9436	0	0	2012-08-02 05:32:43.45129	2012-08-02 05:33:27.008108
6706	1-6706	62	39	489	85	1435	0	4	2012-08-02 05:27:33.087175	2012-08-02 05:33:27.836435
7532	1-7532	26	7	1009	85	8207	0	16	2012-08-02 05:33:07.451505	2012-08-02 05:33:27.836435
7583	1-7583	35	54	1780	33	7044	0	0	2012-08-02 05:33:28.234389	2012-08-02 05:33:28.234389
7507	1-7507	60	43	2831	72	5574	0	5	2012-08-02 05:32:57.020447	2012-08-02 05:33:29.427788
5720	1-5720	95	43	5992	79	7199	0	7	2012-08-02 05:20:49.623292	2012-08-02 05:33:29.427788
5571	1-5571	78	33	3702	80	6459	0	0	2012-08-02 05:19:49.567193	2012-08-02 05:33:29.427788
7586	1-7586	26	80	2937	43	6289	0	3	2012-08-02 05:33:29.427788	2012-08-02 05:33:29.427788
7580	1-7580	3	70	5254	96	3216	0	3	2012-08-02 05:33:27.008108	2012-08-02 05:33:30.23434
6549	1-6549	79	90	4625	98	4845	0	4	2012-08-02 05:26:29.393221	2012-08-02 05:33:30.23434
1478	1-1478	13	70	7171	66	1065	0	0	2012-08-02 04:51:27.480514	2012-08-02 05:33:30.23434
7588	1-7588	32	45	1770	70	8876	0	3	2012-08-02 05:33:30.23434	2012-08-02 05:33:30.23434
6154	1-6154	3	90	7000	98	4287	3045	3	2012-08-02 05:23:46.934188	2012-08-02 05:33:30.23434
6171	1-6171	9	29	6516	85	6508	0	11	2012-08-02 05:23:53.729311	2012-08-02 05:33:31.471669
4783	1-4783	78	29	6444	85	5569	0	8	2012-08-02 05:14:36.255977	2012-08-02 05:33:31.471669
7549	1-7549	86	53	2934	29	8816	0	0	2012-08-02 05:33:14.411698	2012-08-02 05:33:31.471669
4240	1-4240	55	8	2797	80	3381	0	4	2012-08-02 05:10:56.931477	2012-08-02 05:33:31.880539
2728	1-2728	93	81	7929	51	7016	0	0	2012-08-02 05:00:46.147065	2012-08-02 05:33:31.880539
7592	1-7592	79	51	2998	8	2965	0	5	2012-08-02 05:33:31.880539	2012-08-02 05:33:31.880539
7472	1-7472	30	16	7416	27	8957	0	2	2012-08-02 05:32:42.644457	2012-08-02 05:33:33.504556
7443	1-7443	15	4	7417	16	9502	0	0	2012-08-02 05:32:30.702118	2012-08-02 05:33:33.504556
7596	1-7596	26	40	1483	4	8472	0	6	2012-08-02 05:33:33.504556	2012-08-02 05:33:33.504556
6785	1-6785	82	39	2076	10	8172	0	4	2012-08-02 05:28:04.269261	2012-08-02 05:33:36.388437
6515	1-6515	48	85	9110	36	2508	0	3	2012-08-02 05:26:15.305931	2012-08-02 05:33:36.388437
7082	1-7082	3	27	7498	68	5752	0	8	2012-08-02 05:30:04.376349	2012-08-02 05:33:36.388437
4766	1-4766	30	42	3090	28	844	0	9	2012-08-02 05:14:29.372621	2012-08-02 05:33:36.388437
7603	1-7603	37	59	1649	27	7793	0	10	2012-08-02 05:33:36.388437	2012-08-02 05:33:36.388437
6918	1-6918	26	43	1674	10	1186	0	0	2012-08-02 05:28:58.87638	2012-08-02 05:33:36.797158
7369	1-7369	70	43	8709	49	4978	0	6	2012-08-02 05:32:00.405367	2012-08-02 05:33:36.797158
4888	1-4888	28	5	9317	40	6908	0	0	2012-08-02 05:15:18.489146	2012-08-02 05:33:36.797158
1819	1-1819	23	2	9923	81	1823	0	0	2012-08-02 04:54:11.153721	2012-08-02 05:33:39.216741
6268	1-6268	17	25	9295	61	6248	0	0	2012-08-02 05:24:34.031065	2012-08-02 05:33:39.216741
7610	1-7610	11	44	3184	3	4089	0	6	2012-08-02 05:33:39.216741	2012-08-02 05:33:39.216741
3003	1-3003	94	43	7572	8	1857	0	0	2012-08-02 05:02:40.499618	2012-08-02 05:33:40.454398
7604	1-7604	46	10	3781	43	9178	0	2	2012-08-02 05:33:36.797158	2012-08-02 05:33:40.454398
7613	1-7613	11	57	1345	10	6218	0	5	2012-08-02 05:33:40.454398	2012-08-02 05:33:40.454398
7483	1-7483	90	86	2558	9	3931	0	7	2012-08-02 05:32:47.108283	2012-08-02 05:33:41.68057
7262	1-7262	19	97	4082	87	5749	0	0	2012-08-02 05:31:18.318241	2012-08-02 05:33:42.089303
7617	1-7617	6	91	620	99	7122	0	0	2012-08-02 05:33:42.089303	2012-08-02 05:33:42.089303
6981	1-6981	12	47	3159	75	5001	0	0	2012-08-02 05:29:24.276179	2012-08-02 05:33:42.509327
5653	1-5653	52	100	8919	63	9389	0	0	2012-08-02 05:20:22.6174	2012-08-02 05:33:42.928988
4149	1-4149	73	18	5580	19	2389	0	0	2012-08-02 05:10:21.275706	2012-08-02 05:33:42.928988
4438	1-4438	41	18	6039	50	7821	0	0	2012-08-02 05:12:17.251099	2012-08-02 05:33:42.928988
7619	1-7619	68	21	1398	100	8601	0	0	2012-08-02 05:33:42.928988	2012-08-02 05:33:42.928988
3136	1-3136	100	100	6734	78	1783	1494	0	2012-08-02 05:03:34.532856	2012-08-02 05:33:42.928988
6212	1-6212	64	30	1091	11	1228	0	21	2012-08-02 05:24:10.646506	2012-08-02 05:33:43.758012
7349	1-7349	96	79	2845	15	9176	0	7	2012-08-02 05:31:52.550566	2012-08-02 05:33:43.758012
7352	1-7352	54	11	2818	17	6519	0	0	2012-08-02 05:31:53.732179	2012-08-02 05:33:43.758012
7020	1-7020	9	30	4379	82	5746	0	0	2012-08-02 05:29:39.691797	2012-08-02 05:33:43.758012
7434	1-7434	70	30	1920	84	3537	0	14	2012-08-02 05:32:27.001152	2012-08-02 05:33:43.758012
7621	1-7621	30	15	3879	30	7839	0	0	2012-08-02 05:33:43.758012	2012-08-02 05:33:43.758012
3723	1-3723	55	30	7933	11	7803	3071	16	2012-08-02 05:07:31.304826	2012-08-02 05:33:43.758012
7582	1-7582	25	85	8473	39	5354	0	17	2012-08-02 05:33:27.836435	2012-08-02 05:33:44.564377
7392	1-7392	43	80	3522	81	8631	0	0	2012-08-02 05:32:09.631927	2012-08-02 05:33:44.973108
6087	1-6087	47	28	8139	11	6645	0	0	2012-08-02 05:23:20.128449	2012-08-02 05:33:44.973108
7624	1-7624	48	11	2820	24	6489	0	0	2012-08-02 05:33:44.973108	2012-08-02 05:33:44.973108
2060	1-2060	88	26	9559	64	4254	1813	0	2012-08-02 04:55:56.999361	2012-08-02 05:33:44.973108
6906	1-6906	5	65	896	37	1627	0	0	2012-08-02 05:28:53.953757	2012-08-02 05:33:48.254508
7632	1-7632	9	58	1574	65	7210	0	8	2012-08-02 05:33:48.254508	2012-08-02 05:33:48.254508
3825	1-3825	78	38	7012	26	2440	0	0	2012-08-02 05:08:11.635889	2012-08-02 05:33:48.663316
7633	1-7633	95	64	121	80	4276	0	0	2012-08-02 05:33:48.663316	2012-08-02 05:33:48.663316
7186	1-7186	75	49	5350	75	5659	0	0	2012-08-02 05:30:47.72133	2012-08-02 05:33:49.074004
6990	1-6990	38	98	2758	45	3509	0	0	2012-08-02 05:29:27.789351	2012-08-02 05:33:49.074004
7634	1-7634	9	45	1045	49	4893	0	0	2012-08-02 05:33:49.074004	2012-08-02 05:33:49.074004
3500	1-3500	44	75	7862	5	9278	1588	0	2012-08-02 05:06:01.781783	2012-08-02 05:33:49.074004
6739	1-6739	98	62	4413	86	3254	0	0	2012-08-02 05:27:46.125583	2012-08-02 05:33:50.715549
6312	1-6312	95	69	4629	26	3807	0	0	2012-08-02 05:24:52.738568	2012-08-02 05:33:52.750185
6250	1-6250	51	34	2542	63	2792	0	0	2012-08-02 05:24:26.54039	2012-08-02 05:33:52.750185
1281	1-1281	86	69	8834	14	8483	0	0	2012-08-02 04:49:44.387149	2012-08-02 05:33:52.750185
7643	1-7643	93	74	1055	69	6817	0	7	2012-08-02 05:33:52.750185	2012-08-02 05:33:52.750185
7644	1-7644	68	49	208	68	9568	0	6	2012-08-02 05:33:53.195608	2012-08-02 05:33:53.195608
6243	1-6243	12	97	5365	36	4269	0	11	2012-08-02 05:24:23.58995	2012-08-02 05:33:53.607468
7435	1-7435	90	1	6191	94	7021	0	2	2012-08-02 05:32:27.410067	2012-08-02 05:33:54.433387
7591	1-7591	51	85	5126	53	6244	0	2	2012-08-02 05:33:31.471669	2012-08-02 05:33:54.433387
7412	1-7412	27	8	4976	12	6937	0	8	2012-08-02 05:32:17.973282	2012-08-02 05:33:54.832478
6061	1-6061	93	59	9290	45	4494	0	0	2012-08-02 05:23:09.763065	2012-08-02 05:33:54.832478
7512	1-7512	15	83	7601	69	5567	0	0	2012-08-02 05:32:59.086407	2012-08-02 05:33:54.832478
7648	1-7648	13	45	4862	8	8574	0	4	2012-08-02 05:33:54.832478	2012-08-02 05:33:54.832478
7571	1-7571	34	34	6265	45	9153	0	0	2012-08-02 05:33:23.340265	2012-08-02 05:33:56.968395
4759	1-4759	12	25	7800	85	8794	0	0	2012-08-02 05:14:26.61018	2012-08-02 05:33:57.77591
7525	1-7525	18	3	516	2	8081	0	5	2012-08-02 05:33:04.500537	2012-08-02 05:33:58.173483
7275	1-7275	6	56	1447	92	9332	0	0	2012-08-02 05:31:23.444851	2012-08-02 05:33:58.173483
7345	1-7345	27	31	4619	95	5467	0	0	2012-08-02 05:31:50.979393	2012-08-02 05:33:58.173483
7656	1-7656	98	92	9493	3	6484	0	0	2012-08-02 05:33:58.173483	2012-08-02 05:33:58.173483
7647	1-7647	49	53	8798	20	6406	0	0	2012-08-02 05:33:54.433387	2012-08-02 05:33:59.788172
7600	1-7600	22	65	3787	1	1730	0	9	2012-08-02 05:33:35.162092	2012-08-02 05:33:59.788172
7660	1-7660	23	27	1339	94	4829	0	5	2012-08-02 05:33:59.788172	2012-08-02 05:33:59.788172
7158	1-7158	83	88	9755	21	8691	0	0	2012-08-02 05:30:35.933603	2012-08-02 05:34:01.424054
7404	1-7404	57	65	5087	79	4462	0	13	2012-08-02 05:32:14.614423	2012-08-02 05:34:01.822611
3286	1-3286	38	19	4860	45	4002	0	3	2012-08-02 05:04:34.987419	2012-08-02 05:34:02.221051
7666	1-7666	52	45	2643	19	7627	0	6	2012-08-02 05:34:02.221051	2012-08-02 05:34:02.221051
4187	1-4187	90	19	9314	26	7797	916	9	2012-08-02 05:10:36.180616	2012-08-02 05:34:02.221051
7631	1-7631	44	63	4202	77	2598	0	14	2012-08-02 05:33:47.856782	2012-08-02 05:34:03.050224
7668	1-7668	75	15	2886	63	9846	0	10	2012-08-02 05:34:03.050224	2012-08-02 05:34:03.050224
7670	1-7670	90	88	683	77	4802	0	0	2012-08-02 05:34:03.856992	2012-08-02 05:34:03.856992
7677	1-7677	89	48	246	62	1078	0	7	2012-08-02 05:34:06.747453	2012-08-02 05:34:06.747453
7615	1-7615	64	62	725	95	2021	0	0	2012-08-02 05:33:41.271998	2012-08-02 05:34:07.609369
2893	1-2893	5	82	6777	57	8444	0	0	2012-08-02 05:01:55.28443	2012-08-02 05:34:07.609369
3667	1-3667	92	99	8515	30	6129	0	0	2012-08-02 05:07:09.020871	2012-08-02 05:34:09.708848
7388	1-7388	57	57	6712	73	7559	0	4	2012-08-02 05:32:07.985518	2012-08-02 05:34:10.118093
7618	1-7618	76	75	6156	47	8420	0	0	2012-08-02 05:33:42.509327	2012-08-02 05:34:10.118093
7436	1-7436	10	26	39	66	40	0	0	2012-08-02 05:32:27.829546	2012-08-02 05:34:10.958727
7639	1-7639	26	56	1415	33	1325	0	0	2012-08-02 05:33:51.125988	2012-08-02 05:34:12.615534
7088	1-7088	77	94	4127	62	8079	0	0	2012-08-02 05:30:06.78614	2012-08-02 05:34:12.615534
6414	1-6414	43	95	8134	74	3530	0	0	2012-08-02 05:25:34.563706	2012-08-02 05:34:13.036109
7689	1-7689	82	74	1270	69	4051	0	11	2012-08-02 05:34:11.798529	2012-08-02 05:34:13.036109
7691	1-7691	77	35	1825	1	2951	0	17	2012-08-02 05:34:12.615534	2012-08-02 05:34:13.456347
7693	1-7693	13	1	4120	55	8683	0	9	2012-08-02 05:34:13.456347	2012-08-02 05:34:13.456347
7695	1-7695	38	40	65	28	1043	0	0	2012-08-02 05:34:14.318626	2012-08-02 05:34:14.318626
7274	1-7274	77	31	7394	5	4269	0	7	2012-08-02 05:31:23.05801	2012-08-02 05:34:15.169192
6783	1-6783	89	31	9956	17	8749	0	0	2012-08-02 05:28:03.485085	2012-08-02 05:34:15.169192
7698	1-7698	1	22	196	75	2401	0	0	2012-08-02 05:34:15.590977	2012-08-02 05:34:15.590977
7162	1-7162	1	76	6928	39	3279	0	0	2012-08-02 05:30:37.590679	2012-08-02 05:34:16.011264
7699	1-7699	28	24	635	76	7094	0	2	2012-08-02 05:34:16.011264	2012-08-02 05:34:16.011264
5751	1-5751	48	87	9140	66	2393	0	11	2012-08-02 05:21:02.670694	2012-08-02 05:34:16.442487
4872	1-4872	4	75	3884	83	5139	0	6	2012-08-02 05:15:12.169382	2012-08-02 05:34:18.563477
7431	1-7431	27	1	6037	24	7800	0	7	2012-08-02 05:32:25.786124	2012-08-02 05:34:18.563477
6159	1-6159	7	47	7736	66	8134	0	0	2012-08-02 05:23:48.911636	2012-08-02 05:34:18.563477
7684	1-7684	39	4	6889	99	2788	0	0	2012-08-02 05:34:09.708848	2012-08-02 05:34:18.994624
7706	1-7706	57	43	200	4	9387	0	0	2012-08-02 05:34:18.994624	2012-08-02 05:34:18.994624
6104	1-6104	89	68	9414	32	5890	0	7	2012-08-02 05:23:26.967631	2012-08-02 05:34:19.425814
3487	1-3487	3	38	8030	46	7750	0	0	2012-08-02 05:05:56.555983	2012-08-02 05:34:22.375332
7330	1-7330	10	26	3619	41	2533	0	5	2012-08-02 05:31:45.001852	2012-08-02 05:34:22.375332
7446	1-7446	33	26	5585	64	7269	0	7	2012-08-02 05:32:31.939309	2012-08-02 05:34:22.375332
4270	1-4270	18	6	2043	85	2341	0	3	2012-08-02 05:11:08.767462	2012-08-02 05:34:22.795829
7715	1-7715	75	73	4156	8	6139	0	2	2012-08-02 05:34:22.795829	2012-08-02 05:34:22.795829
7597	1-7597	88	20	2797	69	5006	0	0	2012-08-02 05:33:33.92468	2012-08-02 05:34:24.077737
7718	1-7718	24	69	3295	96	6598	0	0	2012-08-02 05:34:24.077737	2012-08-02 05:34:24.077737
7719	1-7719	6	7	586	58	1452	0	10	2012-08-02 05:34:24.519791	2012-08-02 05:34:24.519791
6788	1-6788	30	47	8157	61	6811	0	7	2012-08-02 05:28:05.429954	2012-08-02 05:34:24.950776
7720	1-7720	79	89	2556	47	7048	0	10	2012-08-02 05:34:24.950776	2012-08-02 05:34:24.950776
5664	1-5664	75	55	5547	86	1338	0	0	2012-08-02 05:20:27.026182	2012-08-02 05:34:26.232535
7707	1-7707	52	32	2456	68	6199	0	2	2012-08-02 05:34:19.425814	2012-08-02 05:34:26.232535
7723	1-7723	85	86	2319	32	8044	0	7	2012-08-02 05:34:26.232535	2012-08-02 05:34:26.232535
7726	1-7726	96	19	250	93	2604	0	0	2012-08-02 05:34:27.513884	2012-08-02 05:34:27.513884
7097	1-7097	21	91	4085	71	6374	0	0	2012-08-02 05:30:10.487041	2012-08-02 05:34:28.375899
7667	1-7667	5	9	948	23	2726	0	0	2012-08-02 05:34:02.630479	2012-08-02 05:34:28.807255
7731	1-7731	48	27	415	63	2208	0	0	2012-08-02 05:34:29.702152	2012-08-02 05:34:29.702152
3889	1-3889	96	75	7267	70	7132	0	0	2012-08-02 05:08:37.315207	2012-08-02 05:34:30.132875
7692	1-7692	44	69	7235	75	5273	0	7	2012-08-02 05:34:13.036109	2012-08-02 05:34:30.132875
7573	1-7573	82	5	1799	45	976	0	0	2012-08-02 05:33:24.146664	2012-08-02 05:34:30.132875
7432	1-7432	81	63	8917	23	6332	0	3	2012-08-02 05:32:26.194495	2012-08-02 05:34:31.005942
6474	1-6474	65	3	8872	1	7414	0	0	2012-08-02 05:25:58.673704	2012-08-02 05:34:31.005942
7261	1-7261	89	3	4251	63	5894	0	0	2012-08-02 05:31:17.931621	2012-08-02 05:34:31.005942
7734	1-7734	99	39	1210	3	6659	0	3	2012-08-02 05:34:31.005942	2012-08-02 05:34:31.005942
3910	1-3910	76	97	1777	11	859	6	0	2012-08-02 05:08:45.767156	2012-08-02 05:34:31.005942
1100	1-1100	22	69	5159	22	1550	0	0	2012-08-02 04:48:07.679965	2012-08-02 05:34:32.2984
7503	1-7503	98	93	3762	14	7859	0	0	2012-08-02 05:32:55.352483	2012-08-02 05:34:32.718288
7738	1-7738	68	87	1485	27	8613	0	3	2012-08-02 05:34:32.718288	2012-08-02 05:34:32.718288
7594	1-7594	12	27	2167	42	6279	0	3	2012-08-02 05:33:32.7094	2012-08-02 05:34:33.127291
7697	1-7697	45	53	1510	31	9939	0	5	2012-08-02 05:34:15.169192	2012-08-02 05:34:34.795356
7519	1-7519	55	64	3358	25	6659	0	0	2012-08-02 05:33:01.981709	2012-08-02 05:34:34.795356
5545	1-5545	4	96	437	92	781	0	6	2012-08-02 05:19:39.135413	2012-08-02 05:34:34.795356
7744	1-7744	45	92	2335	72	6174	0	0	2012-08-02 05:34:35.215414	2012-08-02 05:34:35.215414
6844	1-6844	17	30	4737	25	6637	0	3	2012-08-02 05:28:28.348161	2012-08-02 05:34:36.055199
5918	1-5918	27	78	6703	24	1787	0	5	2012-08-02 05:22:10.338208	2012-08-02 05:34:36.055199
7746	1-7746	94	47	1166	78	8554	0	14	2012-08-02 05:34:36.055199	2012-08-02 05:34:36.055199
6792	1-6792	90	25	5514	47	3666	1777	0	2012-08-02 05:28:07.010655	2012-08-02 05:34:36.055199
7748	1-7748	82	31	245	39	1860	0	0	2012-08-02 05:34:36.894738	2012-08-02 05:34:36.894738
7755	1-7755	21	68	838	55	2890	0	9	2012-08-02 05:34:39.868618	2012-08-02 05:34:39.868618
6452	1-6452	30	71	5842	89	8385	0	0	2012-08-02 05:25:49.954258	2012-08-02 05:34:40.299307
7729	1-7729	91	23	3840	19	3763	0	10	2012-08-02 05:34:28.807255	2012-08-02 05:34:44.530275
7272	1-7272	98	68	3462	40	9219	0	3	2012-08-02 05:31:22.273747	2012-08-02 05:34:44.530275
7766	1-7766	46	40	6610	23	9368	0	0	2012-08-02 05:34:44.530275	2012-08-02 05:34:44.530275
5753	1-5753	54	82	3846	13	3494	0	0	2012-08-02 05:21:03.501196	2012-08-02 05:34:44.961413
4983	1-4983	9	2	7860	50	2477	0	4	2012-08-02 05:15:55.685065	2012-08-02 05:34:44.961413
7767	1-7767	76	26	1461	2	9957	0	0	2012-08-02 05:34:44.961413	2012-08-02 05:34:44.961413
7341	1-7341	68	100	6956	18	9238	0	0	2012-08-02 05:31:49.377264	2012-08-02 05:34:45.812198
7100	1-7100	75	100	5354	33	5955	0	0	2012-08-02 05:30:11.724537	2012-08-02 05:34:45.812198
7769	1-7769	78	29	1225	100	5161	0	0	2012-08-02 05:34:45.812198	2012-08-02 05:34:45.812198
7393	1-7393	38	35	4234	71	5482	0	0	2012-08-02 05:32:10.040672	2012-08-02 05:34:46.243462
7380	1-7380	80	35	5484	99	4012	0	0	2012-08-02 05:32:04.758215	2012-08-02 05:34:46.243462
7770	1-7770	9	39	1178	35	4154	0	6	2012-08-02 05:34:46.243462	2012-08-02 05:34:46.243462
5736	1-5736	66	27	9272	39	5690	1091	0	2012-08-02 05:20:56.429131	2012-08-02 05:34:46.243462
7256	1-7256	58	37	7939	28	3984	0	0	2012-08-02 05:31:15.964777	2012-08-02 05:34:46.685556
1723	1-1723	30	37	8167	58	944	0	0	2012-08-02 04:53:27.324441	2012-08-02 05:34:46.685556
5125	1-5125	14	95	5421	80	2315	0	4	2012-08-02 05:16:53.159417	2012-08-02 05:34:47.922975
7465	1-7465	69	2	7850	92	2646	0	0	2012-08-02 05:32:39.750501	2012-08-02 05:34:47.922975
6866	1-6866	63	67	6611	86	5204	0	4	2012-08-02 05:28:37.385773	2012-08-02 05:34:48.740675
7776	1-7776	81	43	1153	100	8933	0	0	2012-08-02 05:34:48.740675	2012-08-02 05:34:48.740675
6701	1-6701	45	14	8467	81	3901	0	0	2012-08-02 05:27:31.142186	2012-08-02 05:34:50.431142
7732	1-7732	12	85	3090	69	8240	0	8	2012-08-02 05:34:30.132875	2012-08-02 05:34:50.431142
7780	1-7780	95	81	3770	64	6643	0	5	2012-08-02 05:34:50.431142	2012-08-02 05:34:50.431142
7149	1-7149	23	20	6874	1	8175	0	0	2012-08-02 05:30:32.154867	2012-08-02 05:34:52.552502
7785	1-7785	56	1	688	31	3351	0	0	2012-08-02 05:34:52.552502	2012-08-02 05:34:52.552502
6691	1-6691	26	58	7115	71	1555	0	0	2012-08-02 05:27:26.986054	2012-08-02 05:34:52.983544
7661	1-7661	26	71	3038	30	2875	0	0	2012-08-02 05:34:00.209066	2012-08-02 05:34:52.983544
7786	1-7786	34	98	2180	67	7203	0	0	2012-08-02 05:34:52.983544	2012-08-02 05:34:52.983544
7541	1-7541	29	24	3124	4	8875	0	4	2012-08-02 05:33:11.130555	2012-08-02 05:34:54.673898
7581	1-7581	16	28	4840	29	9758	0	11	2012-08-02 05:33:27.43921	2012-08-02 05:34:55.104943
7471	1-7471	53	50	8824	19	1584	0	0	2012-08-02 05:32:42.236106	2012-08-02 05:34:55.104943
7758	1-7758	46	28	2393	50	9038	0	0	2012-08-02 05:34:41.150212	2012-08-02 05:34:55.104943
7791	1-7791	82	40	1399	28	5558	0	2	2012-08-02 05:34:55.104943	2012-08-02 05:34:55.104943
7743	1-7743	77	31	5990	96	8952	0	5	2012-08-02 05:34:34.795356	2012-08-02 05:34:55.944778
6837	1-6837	100	88	7604	75	3072	0	5	2012-08-02 05:28:25.464564	2012-08-02 05:34:55.944778
7355	1-7355	8	88	2308	14	2502	0	0	2012-08-02 05:31:54.903567	2012-08-02 05:34:55.944778
6500	1-6500	7	48	6724	80	4594	0	0	2012-08-02 05:26:09.082951	2012-08-02 05:34:58.883271
7802	1-7802	30	33	1564	100	8978	0	7	2012-08-02 05:34:59.78922	2012-08-02 05:34:59.78922
7803	1-7803	40	74	1177	33	1605	0	0	2012-08-02 05:35:00.209115	2012-08-02 05:35:00.209115
6905	1-6905	49	75	5884	90	4603	0	5	2012-08-02 05:28:53.533357	2012-08-02 05:35:01.114906
7805	1-7805	64	40	4232	75	6272	0	0	2012-08-02 05:35:01.114906	2012-08-02 05:35:01.114906
7627	1-7627	26	79	5448	82	4003	0	0	2012-08-02 05:33:46.210608	2012-08-02 05:35:01.954714
7566	1-7566	57	80	4576	38	4504	0	6	2012-08-02 05:33:21.30729	2012-08-02 05:35:02.794395
7196	1-7196	13	90	7735	57	7735	0	9	2012-08-02 05:30:51.786952	2012-08-02 05:35:02.794395
4921	1-4921	20	53	5978	74	1698	0	0	2012-08-02 05:15:31.338567	2012-08-02 05:35:02.794395
7679	1-7679	31	92	8933	82	7804	0	11	2012-08-02 05:34:07.609369	2012-08-02 05:35:04.06514
7816	1-7816	17	83	1156	43	6089	0	9	2012-08-02 05:35:05.813788	2012-08-02 05:35:05.813788
7822	1-7822	42	89	142	30	7679	0	0	2012-08-02 05:35:08.310743	2012-08-02 05:35:08.310743
6965	1-6965	71	60	3973	7	2242	0	5	2012-08-02 05:29:17.878277	2012-08-02 05:35:10.454298
2410	1-2410	84	77	6227	41	1551	0	0	2012-08-02 04:58:28.926589	2012-08-02 05:35:10.454298
2053	1-2053	26	60	8121	23	2386	0	0	2012-08-02 04:55:53.972339	2012-08-02 05:35:10.454298
7829	1-7829	87	44	115	41	1621	0	4	2012-08-02 05:35:11.305134	2012-08-02 05:35:11.305134
7685	1-7685	3	73	5667	75	6272	0	8	2012-08-02 05:34:10.118093	2012-08-02 05:35:11.735741
7490	1-7490	56	76	7520	16	9546	0	0	2012-08-02 05:32:49.980879	2012-08-02 05:35:12.166707
7216	1-7216	82	91	6304	26	7511	0	11	2012-08-02 05:31:00.065056	2012-08-02 05:35:12.166707
5266	1-5266	84	91	9052	42	6093	0	7	2012-08-02 05:17:48.983466	2012-08-02 05:35:12.166707
7154	1-7154	60	42	1684	5	2724	0	0	2012-08-02 05:30:34.2763	2012-08-02 05:35:12.166707
6817	1-6817	28	70	9693	97	5636	0	5	2012-08-02 05:28:17.265827	2012-08-02 05:35:12.597715
7250	1-7250	7	70	3449	85	4116	0	0	2012-08-02 05:31:13.677899	2012-08-02 05:35:12.597715
5870	1-5870	94	87	5910	50	4394	0	0	2012-08-02 05:21:50.857611	2012-08-02 05:35:13.86821
7739	1-7739	41	42	4776	22	3492	0	7	2012-08-02 05:34:33.127291	2012-08-02 05:35:14.741288
7840	1-7840	72	16	342	37	1960	0	0	2012-08-02 05:35:16.000659	2012-08-02 05:35:16.000659
7072	1-7072	99	78	5038	60	8664	0	0	2012-08-02 05:30:00.407485	2012-08-02 05:35:16.431355
7844	1-7844	71	41	362	5	3611	0	0	2012-08-02 05:35:17.701653	2012-08-02 05:35:17.701653
7811	1-7811	8	74	1988	31	5377	0	4	2012-08-02 05:35:03.634151	2012-08-02 05:35:18.541312
7846	1-7846	55	31	5666	38	8188	0	0	2012-08-02 05:35:18.541312	2012-08-02 05:35:18.541312
5108	1-5108	10	73	6616	18	570	0	0	2012-08-02 05:16:46.298259	2012-08-02 05:35:20.220562
7850	1-7850	89	92	3521	72	6081	0	2	2012-08-02 05:35:20.220562	2012-08-02 05:35:20.220562
6287	1-6287	77	69	5233	81	3041	0	4	2012-08-02 05:24:42.36393	2012-08-02 05:35:21.049183
7852	1-7852	1	28	216	5	6342	0	0	2012-08-02 05:35:21.049183	2012-08-02 05:35:21.049183
7363	1-7363	25	93	6811	12	3511	0	0	2012-08-02 05:31:58.041111	2012-08-02 05:35:22.330755
7705	1-7705	39	66	2885	20	6269	0	4	2012-08-02 05:34:18.563477	2012-08-02 05:35:22.750549
7307	1-7307	14	67	7220	58	7138	0	0	2012-08-02 05:31:35.975855	2012-08-02 05:35:22.750549
7439	1-7439	33	59	1412	14	6546	0	3	2012-08-02 05:32:29.044778	2012-08-02 05:35:23.60119
7288	1-7288	89	26	3065	96	2731	0	0	2012-08-02 05:31:28.540497	2012-08-02 05:35:24.010078
7859	1-7859	57	49	2244	26	5676	0	5	2012-08-02 05:35:24.010078	2012-08-02 05:35:24.010078
1718	1-1718	43	72	9456	6	2347	651	0	2012-08-02 04:53:25.523637	2012-08-02 05:35:24.010078
7809	1-7809	36	12	2188	90	9996	0	6	2012-08-02 05:35:02.794395	2012-08-02 05:35:24.452113
7862	1-7862	47	57	1264	47	4890	0	11	2012-08-02 05:35:25.280799	2012-08-02 05:35:25.280799
4102	1-4102	31	24	5265	21	4568	0	0	2012-08-02 05:10:02.489324	2012-08-02 05:35:26.551229
7831	1-7831	76	54	5355	91	7415	0	0	2012-08-02 05:35:12.166707	2012-08-02 05:35:26.551229
6634	1-6634	14	91	4712	21	8029	0	0	2012-08-02 05:27:04.18822	2012-08-02 05:35:26.551229
7778	1-7778	8	18	455	92	8652	0	3	2012-08-02 05:34:49.602878	2012-08-02 05:35:26.551229
7793	1-7793	83	83	839	88	5083	0	0	2012-08-02 05:34:55.944778	2012-08-02 05:35:31.622336
7415	1-7415	42	40	9599	88	8653	0	0	2012-08-02 05:32:19.232845	2012-08-02 05:35:31.622336
7877	1-7877	15	88	4894	99	9972	0	0	2012-08-02 05:35:31.622336	2012-08-02 05:35:31.622336
3341	1-3341	15	98	7769	13	5684	0	0	2012-08-02 05:04:57.241516	2012-08-02 05:35:32.041899
6824	1-6824	17	90	2462	80	2797	0	0	2012-08-02 05:28:20.160675	2012-08-02 05:35:32.881434
7501	1-7501	7	46	5552	13	9927	0	0	2012-08-02 05:32:54.535048	2012-08-02 05:35:34.982685
2798	1-2798	42	38	2704	15	3189	0	0	2012-08-02 05:01:15.486512	2012-08-02 05:35:35.402103
7886	1-7886	25	64	2239	1	5833	0	7	2012-08-02 05:35:35.402103	2012-08-02 05:35:35.402103
7773	1-7773	65	18	1190	6	5602	0	3	2012-08-02 05:34:47.503275	2012-08-02 05:35:36.650589
6383	1-6383	45	77	7370	37	9983	0	3	2012-08-02 05:25:22.098658	2012-08-02 05:35:37.501574
7891	1-7891	86	31	5329	60	7931	0	0	2012-08-02 05:35:37.501574	2012-08-02 05:35:37.501574
7774	1-7774	15	80	4180	17	7331	0	0	2012-08-02 05:34:47.922975	2012-08-02 05:35:39.169675
7447	1-7447	93	6	7494	49	8566	0	11	2012-08-02 05:32:32.403289	2012-08-02 05:35:39.169675
7832	1-7832	41	97	2006	70	6449	0	0	2012-08-02 05:35:12.597715	2012-08-02 05:35:39.987159
7875	1-7875	81	73	2136	70	2457	0	3	2012-08-02 05:35:30.793737	2012-08-02 05:35:39.987159
7463	1-7463	20	51	3911	70	8255	0	3	2012-08-02 05:32:38.932878	2012-08-02 05:35:39.987159
7897	1-7897	45	70	8950	69	9476	0	5	2012-08-02 05:35:39.987159	2012-08-02 05:35:39.987159
7819	1-7819	40	63	2140	14	4810	0	12	2012-08-02 05:35:07.029689	2012-08-02 05:35:40.849404
7260	1-7260	4	25	8703	47	6613	0	5	2012-08-02 05:31:17.533987	2012-08-02 05:35:41.280115
7900	1-7900	67	5	121	85	3017	0	2	2012-08-02 05:35:41.280115	2012-08-02 05:35:41.280115
7316	1-7316	87	97	3482	48	7980	0	0	2012-08-02 05:31:39.544	2012-08-02 05:35:42.959198
7845	1-7845	57	97	1406	93	1425	0	0	2012-08-02 05:35:18.12168	2012-08-02 05:35:42.959198
7904	1-7904	57	66	1212	97	8437	0	3	2012-08-02 05:35:42.959198	2012-08-02 05:35:42.959198
7678	1-7678	66	1	2075	90	3631	0	0	2012-08-02 05:34:07.178325	2012-08-02 05:35:43.39033
6389	1-6389	83	1	8188	94	5315	0	0	2012-08-02 05:25:24.474169	2012-08-02 05:35:43.39033
5698	1-5698	42	96	7411	50	8000	0	3	2012-08-02 05:20:40.840407	2012-08-02 05:35:43.39033
6407	1-6407	27	1	9393	56	8659	0	6	2012-08-02 05:25:31.757116	2012-08-02 05:35:43.39033
7130	1-7130	16	24	6497	50	9808	0	3	2012-08-02 05:30:24.275864	2012-08-02 05:35:43.843375
7736	1-7736	52	50	6237	100	7421	0	2	2012-08-02 05:34:31.878991	2012-08-02 05:35:43.843375
6730	1-6730	42	34	1830	92	6792	0	0	2012-08-02 05:27:42.578882	2012-08-02 05:35:46.837333
1931	1-1931	52	26	4279	93	4048	0	0	2012-08-02 04:55:00.828322	2012-08-02 05:35:48.096511
6763	1-6763	54	47	2009	79	1882	0	2	2012-08-02 05:27:55.651973	2012-08-02 05:35:48.096511
7916	1-7916	39	93	2789	47	4597	0	4	2012-08-02 05:35:48.096511	2012-08-02 05:35:48.096511
7873	1-7873	67	79	1760	10	5192	0	11	2012-08-02 05:35:29.932103	2012-08-02 05:35:48.516574
7536	1-7536	54	40	3066	10	2882	0	0	2012-08-02 05:33:09.086704	2012-08-02 05:35:48.516574
7422	1-7422	48	14	7815	87	5396	0	2	2012-08-02 05:32:22.096602	2012-08-02 05:35:50.664973
7905	1-7905	31	100	4506	1	7345	0	7	2012-08-02 05:35:43.39033	2012-08-02 05:35:51.499589
5137	1-5137	17	70	7635	90	2177	0	0	2012-08-02 05:16:58.011269	2012-08-02 05:35:51.499589
7212	1-7212	40	58	5962	77	8722	0	0	2012-08-02 05:30:58.372281	2012-08-02 05:35:52.350027
4685	1-4685	89	45	2982	29	1298	0	0	2012-08-02 05:13:56.80085	2012-08-02 05:35:54.515644
7828	1-7828	27	37	7306	31	8419	0	2	2012-08-02 05:35:10.885094	2012-08-02 05:35:54.515644
2946	1-2946	53	79	8503	7	3444	0	5	2012-08-02 05:02:17.339959	2012-08-02 05:35:54.515644
7655	1-7655	83	85	9043	25	8221	0	0	2012-08-02 05:33:57.77591	2012-08-02 05:35:54.515644
3725	1-3725	95	93	5709	31	2751	0	4	2012-08-02 05:07:32.100596	2012-08-02 05:35:54.515644
7395	1-7395	67	3	6621	25	9509	0	0	2012-08-02 05:32:10.858184	2012-08-02 05:35:56.680726
7936	1-7936	83	54	357	3	6643	0	0	2012-08-02 05:35:56.680726	2012-08-02 05:35:56.680726
5962	1-5962	2	96	4536	10	4928	0	2	2012-08-02 05:22:28.669396	2012-08-02 05:35:57.498273
7938	1-7938	29	31	2654	96	4883	0	12	2012-08-02 05:35:57.498273	2012-08-02 05:35:57.498273
7917	1-7917	71	10	7955	86	8220	0	0	2012-08-02 05:35:48.516574	2012-08-02 05:35:57.918319
7942	1-7942	22	7	439	35	7847	0	5	2012-08-02 05:35:59.177638	2012-08-02 05:35:59.177638
7409	1-7409	49	72	6380	6	8742	0	8	2012-08-02 05:32:16.702763	2012-08-02 05:35:59.597693
7906	1-7906	14	100	4196	6	4782	0	0	2012-08-02 05:35:43.843375	2012-08-02 05:35:59.597693
7895	1-7895	82	17	7658	6	5313	0	6	2012-08-02 05:35:39.169675	2012-08-02 05:35:59.597693
7533	1-7533	14	30	3337	98	5969	0	5	2012-08-02 05:33:07.860403	2012-08-02 05:35:59.597693
7867	1-7867	16	94	2807	96	4385	0	11	2012-08-02 05:35:27.40206	2012-08-02 05:35:59.597693
7943	1-7943	12	24	2215	30	4893	0	0	2012-08-02 05:35:59.597693	2012-08-02 05:35:59.597693
7543	1-7543	66	33	4622	35	3123	0	0	2012-08-02 05:33:11.959149	2012-08-02 05:36:02.624683
7790	1-7790	7	6	8653	24	9306	0	6	2012-08-02 05:34:54.673898	2012-08-02 05:36:02.624683
7827	1-7827	75	79	2155	60	6827	0	3	2012-08-02 05:35:10.454298	2012-08-02 05:36:03.475958
7637	1-7637	20	32	6318	69	6664	0	13	2012-08-02 05:33:50.306575	2012-08-02 05:36:06.417947
7959	1-7959	51	59	2091	32	5413	0	0	2012-08-02 05:36:06.417947	2012-08-02 05:36:06.417947
5078	1-5078	68	32	8608	69	8931	65	4	2012-08-02 05:16:33.979313	2012-08-02 05:36:06.417947
6086	1-6086	37	32	1615	69	1228	1228	4	2012-08-02 05:23:19.731142	2012-08-02 05:36:06.417947
7342	1-7342	38	23	4476	28	4761	0	6	2012-08-02 05:31:49.786013	2012-08-02 05:36:06.815354
7960	1-7960	57	31	386	23	6058	0	0	2012-08-02 05:36:06.815354	2012-08-02 05:36:06.815354
2490	1-2490	87	87	7885	31	3681	2838	0	2012-08-02 04:59:04.32225	2012-08-02 05:36:06.815354
7201	1-7201	33	42	9801	57	6884	0	0	2012-08-02 05:30:53.842477	2012-08-02 05:36:07.213461
6348	1-6348	56	71	3833	95	6225	0	0	2012-08-02 05:25:07.758215	2012-08-02 05:36:07.622032
7440	1-7440	53	71	5581	65	7753	0	12	2012-08-02 05:32:29.453496	2012-08-02 05:36:07.622032
6105	1-6105	82	95	9857	92	9293	0	9	2012-08-02 05:23:27.36555	2012-08-02 05:36:07.622032
7962	1-7962	44	27	1047	71	9436	0	9	2012-08-02 05:36:07.622032	2012-08-02 05:36:07.622032
7968	1-7968	34	100	1047	44	3184	0	2	2012-08-02 05:36:10.030419	2012-08-02 05:36:10.030419
7733	1-7733	78	19	4222	39	5343	0	13	2012-08-02 05:34:30.552902	2012-08-02 05:36:10.439492
7970	1-7970	22	21	1103	23	2790	0	5	2012-08-02 05:36:10.85909	2012-08-02 05:36:10.85909
7228	1-7228	25	57	6420	96	3763	0	0	2012-08-02 05:31:05.014306	2012-08-02 05:36:11.267954
7441	1-7441	36	18	6094	54	9806	0	5	2012-08-02 05:32:29.851148	2012-08-02 05:36:12.483119
7658	1-7658	45	98	2263	4	6346	0	0	2012-08-02 05:33:58.959594	2012-08-02 05:36:12.880912
7924	1-7924	63	1	787	70	5306	0	0	2012-08-02 05:35:51.499589	2012-08-02 05:36:12.880912
6228	1-6228	57	76	4399	58	4014	0	0	2012-08-02 05:24:17.335656	2012-08-02 05:36:13.278647
6871	1-6871	45	76	9030	81	7747	0	0	2012-08-02 05:28:39.374277	2012-08-02 05:36:13.278647
7523	1-7523	62	27	1616	85	2591	0	0	2012-08-02 05:33:03.693982	2012-08-02 05:36:13.687587
7977	1-7977	92	85	2120	40	3697	0	0	2012-08-02 05:36:13.687587	2012-08-02 05:36:13.687587
7457	1-7457	97	21	4480	76	9432	0	0	2012-08-02 05:32:36.480263	2012-08-02 05:36:14.085304
7978	1-7978	35	83	2256	21	8944	0	0	2012-08-02 05:36:14.085304	2012-08-02 05:36:14.085304
5440	1-5440	65	61	8683	96	3521	0	0	2012-08-02 05:18:57.292673	2012-08-02 05:36:14.881021
7982	1-7982	61	18	1536	78	2573	0	0	2012-08-02 05:36:15.843101	2012-08-02 05:36:15.843101
7064	1-7064	79	71	4679	4	7640	0	4	2012-08-02 05:29:57.225175	2012-08-02 05:36:16.239726
7234	1-7234	58	4	9334	25	3892	0	5	2012-08-02 05:31:07.469306	2012-08-02 05:36:16.239726
7984	1-7984	94	4	446	18	701	0	5	2012-08-02 05:36:16.648615	2012-08-02 05:36:16.648615
7974	1-7974	23	54	6743	6	5610	0	0	2012-08-02 05:36:12.483119	2012-08-02 05:36:17.443953
7986	1-7986	94	44	98	54	1211	0	0	2012-08-02 05:36:17.443953	2012-08-02 05:36:17.443953
5268	1-5268	80	25	9389	31	7003	0	9	2012-08-02 05:17:49.745712	2012-08-02 05:36:19.874672
5905	1-5905	93	76	9006	97	7037	0	0	2012-08-02 05:22:04.956708	2012-08-02 05:36:19.874672
7015	1-7015	87	25	5465	75	4484	0	0	2012-08-02 05:29:37.725422	2012-08-02 05:36:19.874672
7992	1-7992	34	31	2442	25	8469	0	4	2012-08-02 05:36:19.874672	2012-08-02 05:36:19.874672
7994	1-7994	57	36	1439	48	2906	0	2	2012-08-02 05:36:20.703377	2012-08-02 05:36:20.703377
5770	1-5770	47	28	5627	15	3883	0	13	2012-08-02 05:21:10.540745	2012-08-02 05:36:21.111992
7995	1-7995	38	81	495	28	6252	0	4	2012-08-02 05:36:21.111992	2012-08-02 05:36:21.111992
7985	1-7985	7	17	238	11	260	0	10	2012-08-02 05:36:17.057425	2012-08-02 05:36:21.907731
8003	1-8003	37	37	2898	60	8843	0	6	2012-08-02 05:36:24.315686	2012-08-02 05:36:24.315686
7896	1-7896	5	40	8996	99	5866	0	2	2012-08-02 05:35:39.589437	2012-08-02 05:36:24.713472
7996	1-7996	97	4	932	99	804	0	3	2012-08-02 05:36:21.509715	2012-08-02 05:36:24.713472
6525	1-6525	74	65	6311	40	4277	0	0	2012-08-02 05:26:19.460176	2012-08-02 05:36:25.133379
7442	1-7442	78	65	6306	4	3603	0	9	2012-08-02 05:32:30.271178	2012-08-02 05:36:25.133379
8005	1-8005	10	58	857	65	8994	0	10	2012-08-02 05:36:25.133379	2012-08-02 05:36:25.133379
5239	1-5239	48	49	5281	39	2423	0	0	2012-08-02 05:17:38.375878	2012-08-02 05:36:25.542369
4620	1-4620	9	18	6210	20	5296	0	0	2012-08-02 05:13:30.178984	2012-08-02 05:36:28.348211
8013	1-8013	92	10	2832	36	6500	0	0	2012-08-02 05:36:28.348211	2012-08-02 05:36:28.348211
6196	1-6196	33	64	6059	81	3742	0	0	2012-08-02 05:24:03.963515	2012-08-02 05:36:30.325945
8018	1-8018	73	32	1631	64	5595	0	0	2012-08-02 05:36:30.325945	2012-08-02 05:36:30.325945
6679	1-6679	60	49	5842	17	8433	0	0	2012-08-02 05:27:22.202071	2012-08-02 05:36:31.530066
8021	1-8021	61	78	1260	49	3527	0	0	2012-08-02 05:36:31.530066	2012-08-02 05:36:31.530066
6011	1-6011	4	60	9323	37	9903	0	0	2012-08-02 05:22:49.032473	2012-08-02 05:36:31.949891
7771	1-7771	86	29	1996	37	8640	0	0	2012-08-02 05:34:46.685556	2012-08-02 05:36:31.949891
7762	1-7762	84	60	9875	77	9621	0	2	2012-08-02 05:34:42.84015	2012-08-02 05:36:31.949891
1658	1-1658	53	64	9879	37	9189	0	0	2012-08-02 04:52:57.391922	2012-08-02 05:36:31.949891
7665	1-7665	30	29	6843	65	8519	0	0	2012-08-02 05:34:01.822611	2012-08-02 05:36:31.949891
7574	1-7574	29	96	4688	20	2964	0	6	2012-08-02 05:33:24.533268	2012-08-02 05:36:33.995764
4057	1-4057	45	36	8081	51	7939	0	0	2012-08-02 05:09:44.566422	2012-08-02 05:36:33.995764
7628	1-7628	34	83	3303	66	1532	0	9	2012-08-02 05:33:46.608306	2012-08-02 05:36:33.995764
7971	1-7971	21	96	1844	46	9621	0	3	2012-08-02 05:36:11.267954	2012-08-02 05:36:33.995764
39	1-39	21	49	8755	52	3787	0	0	2012-08-02 04:43:53.877483	2012-08-02 05:36:33.995764
8006	1-8006	100	96	2843	36	8845	0	5	2012-08-02 05:36:25.542369	2012-08-02 05:36:33.995764
5264	1-5264	66	96	4997	58	6465	0	0	2012-08-02 05:17:48.18808	2012-08-02 05:36:33.995764
8027	1-8027	61	91	2161	96	7070	0	0	2012-08-02 05:36:33.995764	2012-08-02 05:36:33.995764
7779	1-7779	61	45	3895	2	5920	0	0	2012-08-02 05:34:50.022331	2012-08-02 05:36:34.813814
8029	1-8029	91	2	7287	36	3944	0	7	2012-08-02 05:36:34.813814	2012-08-02 05:36:34.813814
7950	1-7950	50	24	2012	33	6033	0	0	2012-08-02 05:36:02.624683	2012-08-02 05:36:36.029099
7931	1-7931	39	31	4622	45	8549	0	10	2012-08-02 05:35:54.515644	2012-08-02 05:36:36.824441
7865	1-7865	13	21	7946	18	8986	0	2	2012-08-02 05:35:26.551229	2012-08-02 05:36:36.824441
7961	1-7961	14	57	5499	42	9391	0	0	2012-08-02 05:36:07.213461	2012-08-02 05:36:38.835253
7013	1-7013	27	83	4971	37	4301	0	0	2012-08-02 05:29:36.930082	2012-08-02 05:36:39.232867
5246	1-5246	72	100	9614	11	4216	0	5	2012-08-02 05:17:41.115533	2012-08-02 05:36:40.415022
6002	1-6002	71	68	7754	83	9136	0	14	2012-08-02 05:22:45.275967	2012-08-02 05:36:40.823859
8044	1-8044	19	95	2461	19	9278	0	0	2012-08-02 05:36:40.823859	2012-08-02 05:36:40.823859
8045	1-8045	15	31	2493	83	4427	0	11	2012-08-02 05:36:41.232845	2012-08-02 05:36:41.232845
8046	1-8046	81	31	276	37	4295	0	11	2012-08-02 05:36:41.630396	2012-08-02 05:36:41.630396
7807	1-7807	63	56	2729	79	5840	0	4	2012-08-02 05:35:01.954714	2012-08-02 05:36:43.607933
6446	1-6446	4	7	6572	79	5488	0	0	2012-08-02 05:25:47.589944	2012-08-02 05:36:43.607933
2344	1-2344	58	56	5833	79	2502	1438	0	2012-08-02 04:58:00.132756	2012-08-02 05:36:43.607933
4100	1-4100	91	40	6893	45	1237	0	7	2012-08-02 05:10:01.704959	2012-08-02 05:36:47.2535
8060	1-8060	32	70	1163	40	9482	0	0	2012-08-02 05:36:47.2535	2012-08-02 05:36:47.2535
8061	1-8061	10	2	2542	27	3272	0	0	2012-08-02 05:36:47.64057	2012-08-02 05:36:47.64057
8063	1-8063	36	66	1609	89	8486	0	3	2012-08-02 05:36:48.424849	2012-08-02 05:36:48.424849
2501	1-2501	95	8	9998	83	8258	0	0	2012-08-02 04:59:09.090168	2012-08-02 05:36:49.231189
7424	1-7424	43	17	8778	83	8293	0	4	2012-08-02 05:32:22.902554	2012-08-02 05:36:49.231189
8065	1-8065	33	83	437	8	6304	0	0	2012-08-02 05:36:49.231189	2012-08-02 05:36:49.231189
7921	1-7921	32	54	7318	76	7256	0	0	2012-08-02 05:35:50.206807	2012-08-02 05:36:50.822423
8022	1-8022	80	37	2664	60	9713	0	7	2012-08-02 05:36:31.949891	2012-08-02 05:36:51.617508
7830	1-7830	31	72	3053	73	4914	0	3	2012-08-02 05:35:11.735741	2012-08-02 05:36:52.821791
8082	1-8082	41	37	3735	66	3834	0	0	2012-08-02 05:36:56.036532	2012-08-02 05:36:56.036532
8083	1-8083	83	18	4157	22	4738	0	3	2012-08-02 05:36:56.434306	2012-08-02 05:36:56.434306
7881	1-7881	19	11	2868	82	3463	0	0	2012-08-02 05:35:33.290317	2012-08-02 05:36:56.832124
8051	1-8051	44	79	2124	6	3978	0	0	2012-08-02 05:36:43.607933	2012-08-02 05:36:56.832124
5648	1-5648	7	52	7386	20	5249	0	2	2012-08-02 05:20:20.628695	2012-08-02 05:36:58.036673
7318	1-7318	92	94	6857	71	6921	0	0	2012-08-02 05:31:40.328338	2012-08-02 05:36:58.036673
7813	1-7813	68	56	3073	57	8055	0	8	2012-08-02 05:35:04.50711	2012-08-02 05:36:58.854214
7584	1-7584	59	62	2257	35	1136	0	4	2012-08-02 05:33:28.63211	2012-08-02 05:36:58.854214
3757	1-3757	12	56	7456	57	6373	6373	0	2012-08-02 05:07:44.752637	2012-08-02 05:36:58.854214
7837	1-7837	65	15	9198	64	7207	0	0	2012-08-02 05:35:14.741288	2012-08-02 05:36:59.671529
8091	1-8091	40	95	4191	15	7849	0	10	2012-08-02 05:36:59.671529	2012-08-02 05:36:59.671529
8092	1-8092	33	83	1016	69	8374	0	3	2012-08-02 05:37:00.069326	2012-08-02 05:37:00.069326
2885	1-2885	3	69	8406	5	2234	2234	0	2012-08-02 05:01:52.014134	2012-08-02 05:37:00.069326
8093	1-8093	89	38	481	90	9239	0	3	2012-08-02 05:37:00.511334	2012-08-02 05:37:00.511334
5263	1-5263	43	26	4736	54	4644	0	6	2012-08-02 05:17:47.80148	2012-08-02 05:37:00.909034
7714	1-7714	99	13	7548	26	9392	0	8	2012-08-02 05:34:22.375332	2012-08-02 05:37:00.909034
6464	1-6464	10	13	5442	32	2417	0	3	2012-08-02 05:25:54.718124	2012-08-02 05:37:00.909034
6848	1-6848	7	4	9577	58	6163	0	0	2012-08-02 05:28:30.016118	2012-08-02 05:37:02.102582
7711	1-7711	78	92	8782	1	5548	0	14	2012-08-02 05:34:21.127087	2012-08-02 05:37:02.102582
7856	1-7856	80	20	2670	66	5749	0	2	2012-08-02 05:35:22.750549	2012-08-02 05:37:02.888895
5548	1-5548	88	97	7021	3	3358	0	0	2012-08-02 05:19:40.317705	2012-08-02 05:37:04.114703
8103	1-8103	68	67	1584	46	7451	0	0	2012-08-02 05:37:04.523585	2012-08-02 05:37:04.523585
8084	1-8084	25	57	9282	30	6352	0	0	2012-08-02 05:36:56.832124	2012-08-02 05:37:06.148944
7700	1-7700	85	37	6017	87	8470	0	6	2012-08-02 05:34:16.442487	2012-08-02 05:37:07.761867
7878	1-7878	48	69	7118	98	7727	0	3	2012-08-02 05:35:32.041899	2012-08-02 05:37:08.977078
7975	1-7975	61	70	3694	98	4700	0	14	2012-08-02 05:36:12.880912	2012-08-02 05:37:10.988066
8119	1-8119	73	98	2060	38	4730	0	0	2012-08-02 05:37:10.988066	2012-08-02 05:37:10.988066
7919	1-7919	74	6	7375	18	9541	0	0	2012-08-02 05:35:49.378644	2012-08-02 05:37:11.396571
7153	1-7153	3	22	8393	99	6277	0	2	2012-08-02 05:30:33.834506	2012-08-02 05:37:11.396571
8048	1-8048	38	15	6760	60	5270	0	0	2012-08-02 05:36:42.437075	2012-08-02 05:37:11.794453
8097	1-8097	6	53	2655	4	6771	0	3	2012-08-02 05:37:02.102582	2012-08-02 05:37:11.794453
7281	1-7281	43	11	7941	39	7096	0	4	2012-08-02 05:31:25.788852	2012-08-02 05:37:12.61224
7413	1-7413	64	75	8174	61	5609	0	22	2012-08-02 05:32:18.382144	2012-08-02 05:37:12.61224
8123	1-8123	70	100	1340	11	5246	0	10	2012-08-02 05:37:12.61224	2012-08-02 05:37:12.61224
6533	1-6533	25	4	6058	59	4538	0	0	2012-08-02 05:26:22.774512	2012-08-02 05:37:13.02102
4716	1-4716	36	29	9037	4	3924	0	0	2012-08-02 05:14:09.1839	2012-08-02 05:37:13.02102
7358	1-7358	47	44	5886	49	3648	0	0	2012-08-02 05:31:56.041313	2012-08-02 05:37:13.418798
8125	1-8125	88	49	2168	66	6832	0	9	2012-08-02 05:37:13.418798	2012-08-02 05:37:13.418798
7649	1-7649	61	40	8037	82	2080	0	6	2012-08-02 05:33:55.255115	2012-08-02 05:37:13.838654
8126	1-8126	23	81	390	40	7908	0	0	2012-08-02 05:37:13.838654	2012-08-02 05:37:13.838654
7027	1-7027	46	69	9112	34	6503	0	2	2012-08-02 05:29:42.4865	2012-08-02 05:37:16.346497
8124	1-8124	5	63	559	29	7659	0	8	2012-08-02 05:37:13.02102	2012-08-02 05:37:19.185489
7513	1-7513	55	7	6727	32	9137	0	5	2012-08-02 05:32:59.495215	2012-08-02 05:37:19.981044
7879	1-7879	71	7	3689	32	2986	0	9	2012-08-02 05:35:32.461846	2012-08-02 05:37:19.981044
8141	1-8141	51	69	1266	7	9433	0	0	2012-08-02 05:37:19.981044	2012-08-02 05:37:19.981044
8144	1-8144	62	44	37	31	3201	0	0	2012-08-02 05:37:21.185242	2012-08-02 05:37:21.185242
7885	1-7885	54	13	4137	80	4207	0	0	2012-08-02 05:35:34.982685	2012-08-02 05:37:21.992091
6442	1-6442	64	19	4588	21	6727	0	0	2012-08-02 05:25:45.999281	2012-08-02 05:37:22.411927
6786	1-6786	80	69	8654	5	8460	0	6	2012-08-02 05:28:04.655937	2012-08-02 05:37:23.615718
6943	1-6943	39	94	3459	5	7552	0	0	2012-08-02 05:29:09.146092	2012-08-02 05:37:23.615718
8153	1-8153	52	57	1515	62	3391	0	0	2012-08-02 05:37:24.809003	2012-08-02 05:37:24.809003
8120	1-8120	77	18	1957	63	9086	0	5	2012-08-02 05:37:11.396571	2012-08-02 05:37:25.789034
7909	1-7909	72	56	2764	36	7687	0	6	2012-08-02 05:35:45.125294	2012-08-02 05:37:25.789034
7983	1-7983	85	25	1328	84	3208	0	0	2012-08-02 05:36:16.239726	2012-08-02 05:37:26.201113
4744	1-4744	63	63	9072	84	3773	0	0	2012-08-02 05:14:20.532828	2012-08-02 05:37:26.201113
7955	1-7955	91	12	1838	51	5470	0	0	2012-08-02 05:36:04.734941	2012-08-02 05:37:26.201113
6119	1-6119	30	30	4322	11	4776	0	21	2012-08-02 05:23:32.933457	2012-08-02 05:37:26.201113
7858	1-7858	73	14	5924	40	5149	0	7	2012-08-02 05:35:23.60119	2012-08-02 05:37:27.007673
8150	1-8150	61	5	1938	42	4924	0	0	2012-08-02 05:37:23.615718	2012-08-02 05:37:27.007673
8158	1-8158	45	21	1051	72	7041	0	0	2012-08-02 05:37:27.007673	2012-08-02 05:37:27.007673
8160	1-8160	69	13	2791	88	5550	0	0	2012-08-02 05:37:27.82533	2012-08-02 05:37:27.82533
6076	1-6076	30	69	9401	59	6160	0	0	2012-08-02 05:23:15.742102	2012-08-02 05:37:28.598722
8009	1-8009	81	81	3403	28	7753	0	6	2012-08-02 05:36:26.746333	2012-08-02 05:37:28.598722
7505	1-7505	14	91	5718	76	9385	0	0	2012-08-02 05:32:56.180738	2012-08-02 05:37:29.416694
8121	1-8121	25	60	2115	55	3214	0	0	2012-08-02 05:37:11.794453	2012-08-02 05:37:29.416694
8087	1-8087	97	20	2946	94	7649	0	6	2012-08-02 05:36:58.036673	2012-08-02 05:37:29.825212
7855	1-7855	78	12	4250	93	9542	0	7	2012-08-02 05:35:22.330755	2012-08-02 05:37:30.246944
8166	1-8166	46	93	2623	20	8383	0	11	2012-08-02 05:37:30.246944	2012-08-02 05:37:30.246944
8167	1-8167	80	54	500	6	6743	0	3	2012-08-02 05:37:30.655721	2012-08-02 05:37:30.655721
5740	1-5740	93	45	5957	80	7108	0	0	2012-08-02 05:20:58.086245	2012-08-02 05:37:31.053609
4598	1-4598	49	45	6067	84	2455	0	3	2012-08-02 05:13:21.106772	2012-08-02 05:37:31.053609
8156	1-8156	50	84	2312	12	7032	0	0	2012-08-02 05:37:26.201113	2012-08-02 05:37:31.849236
7795	1-7795	52	100	7712	19	4035	0	0	2012-08-02 05:34:56.795444	2012-08-02 05:37:32.246722
8168	1-8168	88	13	3529	45	8479	0	5	2012-08-02 05:37:31.053609	2012-08-02 05:37:32.246722
7997	1-7997	83	82	1487	9	3628	0	8	2012-08-02 05:36:21.907731	2012-08-02 05:37:33.064634
8173	1-8173	67	68	2033	94	3758	0	16	2012-08-02 05:37:33.064634	2012-08-02 05:37:33.064634
7987	1-7987	82	52	3040	91	2993	0	0	2012-08-02 05:36:17.853077	2012-08-02 05:37:33.893283
7287	1-7287	78	86	2872	34	4665	0	10	2012-08-02 05:31:28.14291	2012-08-02 05:37:33.893283
8175	1-8175	42	91	3671	12	6087	0	0	2012-08-02 05:37:33.893283	2012-08-02 05:37:33.893283
2515	1-2515	20	68	9359	55	2090	1188	0	2012-08-02 04:59:15.166731	2012-08-02 05:37:33.893283
7005	1-7005	34	5	9409	62	7159	0	7	2012-08-02 05:29:33.726099	2012-08-02 05:37:35.561491
7910	1-7910	50	5	2918	55	3193	0	0	2012-08-02 05:35:45.556025	2012-08-02 05:37:35.561491
5738	1-5738	3	5	5966	54	5838	0	0	2012-08-02 05:20:57.246724	2012-08-02 05:37:35.561491
8094	1-8094	61	54	3421	13	6885	0	0	2012-08-02 05:37:00.909034	2012-08-02 05:37:36.378924
6412	1-6412	100	45	6475	79	9384	0	0	2012-08-02 05:25:33.768509	2012-08-02 05:37:36.765697
8147	1-8147	8	99	2210	52	6226	0	0	2012-08-02 05:37:22.411927	2012-08-02 05:37:36.765697
6728	1-6728	55	99	3754	97	8659	0	8	2012-08-02 05:27:41.80523	2012-08-02 05:37:36.765697
8004	1-8004	8	99	4971	18	8436	0	0	2012-08-02 05:36:24.713472	2012-08-02 05:37:36.765697
4680	1-4680	80	68	8162	44	6769	0	0	2012-08-02 05:13:54.741733	2012-08-02 05:37:36.765697
3177	1-3177	27	29	6855	82	264	0	0	2012-08-02 05:03:51.158226	2012-08-02 05:37:37.1857
8162	1-8162	21	59	3092	26	7478	0	0	2012-08-02 05:37:28.598722	2012-08-02 05:37:38.003146
6953	1-6953	57	71	7845	44	9748	0	0	2012-08-02 05:29:13.092549	2012-08-02 05:37:38.423009
6850	1-6850	89	84	7606	81	8863	0	0	2012-08-02 05:28:30.84484	2012-08-02 05:37:38.423009
7653	1-7653	94	79	7168	44	9389	0	5	2012-08-02 05:33:56.968395	2012-08-02 05:37:39.240698
4562	1-4562	86	86	8504	44	4878	0	0	2012-08-02 05:13:06.368301	2012-08-02 05:37:39.638401
7889	1-7889	1	44	7629	66	5143	0	0	2012-08-02 05:35:36.650589	2012-08-02 05:37:39.638401
8189	1-8189	15	67	999	86	8211	0	10	2012-08-02 05:37:39.638401	2012-08-02 05:37:39.638401
6808	1-6808	31	86	7687	34	6725	1440	5	2012-08-02 05:28:13.584685	2012-08-02 05:37:39.638401
6909	1-6909	27	63	8575	50	7247	0	11	2012-08-02 05:28:55.192529	2012-08-02 05:37:42.091071
8195	1-8195	97	78	52	63	3790	0	6	2012-08-02 05:37:42.091071	2012-08-02 05:37:42.091071
715	1-715	84	50	7976	25	1795	318	0	2012-08-02 04:45:03.133912	2012-08-02 05:37:42.091071
8139	1-8139	95	29	4145	56	811	0	6	2012-08-02 05:37:19.185489	2012-08-02 05:37:43.295399
8182	1-8182	95	44	5723	99	6756	0	0	2012-08-02 05:37:36.765697	2012-08-02 05:37:43.295399
4497	1-4497	71	30	9084	27	6563	636	0	2012-08-02 05:12:40.521332	2012-08-02 05:37:43.295399
5657	1-5657	79	90	5523	53	6707	0	0	2012-08-02 05:20:24.21953	2012-08-02 05:37:43.704344
4700	1-4700	59	78	7770	71	1972	0	6	2012-08-02 05:14:02.670993	2012-08-02 05:37:43.704344
8200	1-8200	61	40	2942	98	2871	0	3	2012-08-02 05:37:44.113272	2012-08-02 05:37:44.113272
7611	1-7611	14	66	4928	16	7421	0	0	2012-08-02 05:33:39.636593	2012-08-02 05:37:46.532627
7939	1-7939	70	66	3132	10	8650	0	0	2012-08-02 05:35:57.918319	2012-08-02 05:37:46.532627
8198	1-8198	76	27	2143	38	7477	0	0	2012-08-02 05:37:43.295399	2012-08-02 05:37:46.941394
8207	1-8207	20	38	1874	84	4756	0	0	2012-08-02 05:37:46.941394	2012-08-02 05:37:46.941394
7587	1-7587	71	66	2529	67	5213	0	0	2012-08-02 05:33:29.825563	2012-08-02 05:37:47.339209
6875	1-6875	92	74	8327	12	5742	0	2	2012-08-02 05:28:41.164343	2012-08-02 05:37:47.339209
8208	1-8208	98	12	156	66	3297	0	0	2012-08-02 05:37:47.339209	2012-08-02 05:37:47.339209
6562	1-6562	48	64	2918	72	2013	0	14	2012-08-02 05:26:34.674638	2012-08-02 05:37:48.157087
3246	1-3246	24	64	6213	11	4133	0	0	2012-08-02 05:04:18.811052	2012-08-02 05:37:48.157087
6378	1-6378	34	18	5288	17	5232	0	0	2012-08-02 05:25:20.099104	2012-08-02 05:37:49.372078
5574	1-5574	80	94	8990	63	8740	0	3	2012-08-02 05:19:50.738366	2012-08-02 05:37:49.372078
8213	1-8213	40	26	969	94	9638	0	7	2012-08-02 05:37:49.372078	2012-08-02 05:37:49.372078
3329	1-3329	89	93	7626	25	1916	0	0	2012-08-02 05:04:52.347075	2012-08-02 05:37:50.201113
5652	1-5652	62	28	4557	37	7049	0	12	2012-08-02 05:20:22.241838	2012-08-02 05:37:50.201113
4450	1-4450	59	71	8140	78	4276	0	0	2012-08-02 05:12:21.957491	2012-08-02 05:37:50.201113
1201	1-1201	61	78	4436	28	262	262	0	2012-08-02 04:49:00.874391	2012-08-02 05:37:50.201113
4009	1-4009	68	26	8704	94	3511	0	0	2012-08-02 05:09:25.242798	2012-08-02 05:37:51.427408
6720	1-6720	63	64	3772	42	3842	0	2	2012-08-02 05:27:38.591153	2012-08-02 05:37:51.427408
2641	1-2641	58	69	4852	94	741	741	5	2012-08-02 05:00:09.266259	2012-08-02 05:37:51.427408
2285	1-2285	67	69	9295	94	2385	413	5	2012-08-02 04:57:34.498211	2012-08-02 05:37:51.427408
8219	1-8219	56	48	1869	89	9513	0	9	2012-08-02 05:37:51.847264	2012-08-02 05:37:51.847264
7437	1-7437	92	21	2377	99	1341	0	20	2012-08-02 05:32:28.238266	2012-08-02 05:37:53.846989
8111	1-8111	78	65	3388	21	8785	0	10	2012-08-02 05:37:07.761867	2012-08-02 05:37:53.846989
6821	1-6821	26	6	7670	77	7476	0	8	2012-08-02 05:28:18.9456	2012-08-02 05:37:55.051162
7625	1-7625	24	51	8941	52	6217	0	4	2012-08-02 05:33:45.393165	2012-08-02 05:37:55.051162
8227	1-8227	79	89	1146	8	5958	0	0	2012-08-02 05:37:55.051162	2012-08-02 05:37:55.051162
5353	1-5353	47	96	8810	78	6869	0	0	2012-08-02 05:18:23.105077	2012-08-02 05:37:55.449095
8186	1-8186	22	28	6409	84	7857	0	0	2012-08-02 05:37:38.423009	2012-08-02 05:37:55.449095
5818	1-5818	74	96	9471	61	5041	0	0	2012-08-02 05:21:29.777938	2012-08-02 05:37:55.449095
6738	1-6738	13	94	9668	82	9093	0	0	2012-08-02 05:27:45.738965	2012-08-02 05:37:57.483965
7737	1-7737	33	99	7708	27	8987	0	0	2012-08-02 05:34:32.2984	2012-08-02 05:37:58.290181
7645	1-7645	59	59	7252	51	6438	0	6	2012-08-02 05:33:53.607468	2012-08-02 05:37:58.290181
8210	1-8210	44	28	4139	64	8935	0	0	2012-08-02 05:37:48.157087	2012-08-02 05:37:58.290181
8217	1-8217	73	14	4885	47	358	0	3	2012-08-02 05:37:51.040661	2012-08-02 05:37:59.097031
8155	1-8155	55	57	4021	18	6021	0	0	2012-08-02 05:37:25.789034	2012-08-02 05:38:00.29399
8129	1-8129	18	47	878	50	5953	0	0	2012-08-02 05:37:15.064931	2012-08-02 05:38:00.29399
8241	1-8241	73	68	625	12	5234	0	0	2012-08-02 05:38:00.746989	2012-08-02 05:38:00.746989
5680	1-5680	78	52	5604	76	5783	0	0	2012-08-02 05:20:33.511466	2012-08-02 05:38:01.144878
8242	1-8242	86	75	805	52	7933	0	3	2012-08-02 05:38:01.144878	2012-08-02 05:38:01.144878
6589	1-6589	52	78	2839	37	5299	0	0	2012-08-02 05:26:45.747196	2012-08-02 05:38:02.73564
4730	1-4730	45	55	3852	37	6207	0	7	2012-08-02 05:14:14.896246	2012-08-02 05:38:02.73564
8246	1-8246	7	37	3826	19	4429	0	5	2012-08-02 05:38:02.73564	2012-08-02 05:38:02.73564
8250	1-8250	29	98	33	33	2389	0	2	2012-08-02 05:38:04.35968	2012-08-02 05:38:04.35968
2979	1-2979	24	38	9183	89	6596	0	0	2012-08-02 05:02:30.764231	2012-08-02 05:38:05.185079
7817	1-7817	2	38	8139	31	5921	0	5	2012-08-02 05:35:06.211973	2012-08-02 05:38:05.185079
7980	1-7980	18	46	9709	75	8320	0	6	2012-08-02 05:36:14.881021	2012-08-02 05:38:05.678853
7510	1-7510	33	81	8975	28	6504	0	0	2012-08-02 05:32:58.268884	2012-08-02 05:38:06.084245
8247	1-8247	82	92	2832	28	1038	0	0	2012-08-02 05:38:03.144535	2012-08-02 05:38:06.084245
4184	1-4184	46	81	8022	44	3404	0	8	2012-08-02 05:10:34.976146	2012-08-02 05:38:06.879798
8256	1-8256	22	5	3318	47	7441	0	0	2012-08-02 05:38:06.879798	2012-08-02 05:38:06.879798
8228	1-8228	59	84	4192	96	8888	0	0	2012-08-02 05:37:55.449095	2012-08-02 05:38:07.686533
7575	1-7575	32	42	6557	82	2305	0	3	2012-08-02 05:33:24.942409	2012-08-02 05:38:07.686533
717	1-717	1	86	6898	13	5390	0	0	2012-08-02 04:45:03.631109	2012-08-02 05:38:08.492792
4233	1-4233	74	86	3339	63	2980	0	0	2012-08-02 05:10:54.191602	2012-08-02 05:38:08.492792
6933	1-6933	76	86	5662	43	5559	0	0	2012-08-02 05:29:05.117609	2012-08-02 05:38:08.492792
5527	1-5527	81	16	8445	54	5872	0	0	2012-08-02 05:19:32.042511	2012-08-02 05:38:08.492792
8260	1-8260	43	54	984	86	9181	0	0	2012-08-02 05:38:08.492792	2012-08-02 05:38:08.492792
7569	1-7569	64	17	3775	15	5022	0	2	2012-08-02 05:33:22.523048	2012-08-02 05:38:10.901615
7469	1-7469	91	43	6643	94	2687	0	0	2012-08-02 05:32:41.407585	2012-08-02 05:38:12.09431
8269	1-8269	73	88	5304	43	5283	0	0	2012-08-02 05:38:12.09431	2012-08-02 05:38:12.09431
8253	1-8253	27	75	1248	73	3726	0	0	2012-08-02 05:38:05.678853	2012-08-02 05:38:14.083078
7922	1-7922	88	73	1139	14	8622	0	0	2012-08-02 05:35:50.664973	2012-08-02 05:38:14.491732
8275	1-8275	99	14	4401	41	5564	0	2	2012-08-02 05:38:14.491732	2012-08-02 05:38:14.491732
6329	1-6329	33	22	8632	1	5088	0	7	2012-08-02 05:24:59.85385	2012-08-02 05:38:14.878638
8079	1-8079	7	19	5254	33	6836	0	0	2012-08-02 05:36:54.832475	2012-08-02 05:38:14.878638
7687	1-7687	69	66	4572	91	4951	0	6	2012-08-02 05:34:10.958727	2012-08-02 05:38:16.126915
8100	1-8100	57	25	2650	78	8997	0	2	2012-08-02 05:37:03.29727	2012-08-02 05:38:16.126915
8252	1-8252	21	43	3862	38	5948	0	7	2012-08-02 05:38:05.185079	2012-08-02 05:38:16.95555
6491	1-6491	23	43	5877	16	5739	0	6	2012-08-02 05:26:05.447365	2012-08-02 05:38:16.95555
8281	1-8281	55	33	900	31	4978	0	0	2012-08-02 05:38:16.95555	2012-08-02 05:38:16.95555
8283	1-8283	28	52	93	38	9519	0	0	2012-08-02 05:38:17.762139	2012-08-02 05:38:17.762139
8040	1-8040	88	37	4463	17	7579	0	0	2012-08-02 05:36:39.232867	2012-08-02 05:38:18.579956
8016	1-8016	10	85	2772	35	878	0	0	2012-08-02 05:36:29.541313	2012-08-02 05:38:19.784264
5099	1-5099	48	100	8592	97	4169	0	0	2012-08-02 05:16:42.641144	2012-08-02 05:38:20.645622
6877	1-6877	57	100	3919	17	2904	0	7	2012-08-02 05:28:41.982113	2012-08-02 05:38:20.645622
7753	1-7753	8	36	5829	49	4812	0	0	2012-08-02 05:34:39.028875	2012-08-02 05:38:23.487117
8028	1-8028	10	51	6685	45	3132	0	6	2012-08-02 05:36:34.427195	2012-08-02 05:38:24.66889
8301	1-8301	50	66	1048	46	5435	0	0	2012-08-02 05:38:25.066624	2012-08-02 05:38:25.066624
5961	1-5961	76	46	7309	74	2573	0	0	2012-08-02 05:22:28.260512	2012-08-02 05:38:27.066168
8306	1-8306	50	74	1424	46	8875	0	8	2012-08-02 05:38:27.066168	2012-08-02 05:38:27.066168
5722	1-5722	72	78	9316	76	7498	0	0	2012-08-02 05:20:50.407739	2012-08-02 05:38:27.486033
8307	1-8307	29	85	38	78	4728	0	11	2012-08-02 05:38:27.486033	2012-08-02 05:38:27.486033
4626	1-4626	67	21	8478	84	4055	0	6	2012-08-02 05:13:32.643033	2012-08-02 05:38:27.884026
8308	1-8308	90	84	687	3	8871	0	9	2012-08-02 05:38:27.884026	2012-08-02 05:38:27.884026
8194	1-8194	80	83	1659	9	1780	0	0	2012-08-02 05:37:41.682428	2012-08-02 05:38:28.701418
8310	1-8310	30	24	685	75	5026	0	0	2012-08-02 05:38:28.701418	2012-08-02 05:38:28.701418
116	1-116	64	75	9372	83	4535	82	0	2012-08-02 04:43:54.783379	2012-08-02 05:38:28.701418
1791	1-1791	33	21	6800	46	3419	0	0	2012-08-02 04:53:58.26093	2012-08-02 05:38:29.496924
6684	1-6684	34	35	3399	48	5056	0	0	2012-08-02 05:27:24.179624	2012-08-02 05:38:30.292428
7516	1-7516	22	36	4919	59	9714	0	0	2012-08-02 05:33:00.743922	2012-08-02 05:38:30.292428
8314	1-8314	9	59	4697	96	9165	0	0	2012-08-02 05:38:30.292428	2012-08-02 05:38:30.292428
8317	1-8317	48	85	1445	29	5353	0	3	2012-08-02 05:38:31.496507	2012-08-02 05:38:31.496507
8240	1-8240	85	50	3787	6	6342	0	0	2012-08-02 05:38:00.29399	2012-08-02 05:38:31.883347
8318	1-8318	87	46	665	50	3007	0	6	2012-08-02 05:38:31.883347	2012-08-02 05:38:31.883347
7514	1-7514	7	96	6862	64	6485	0	4	2012-08-02 05:32:59.915281	2012-08-02 05:38:33.485609
7085	1-7085	24	39	6629	96	6416	0	4	2012-08-02 05:30:05.570946	2012-08-02 05:38:33.485609
8276	1-8276	59	1	5661	19	9570	0	3	2012-08-02 05:38:14.878638	2012-08-02 05:38:33.485609
6927	1-6927	81	39	8349	61	8280	0	0	2012-08-02 05:29:02.617343	2012-08-02 05:38:33.485609
8322	1-8322	40	26	2252	39	4574	0	10	2012-08-02 05:38:33.485609	2012-08-02 05:38:33.485609
7589	1-7589	68	21	2589	37	5124	0	5	2012-08-02 05:33:30.665326	2012-08-02 05:38:35.926903
7123	1-7123	12	18	3114	37	2403	0	0	2012-08-02 05:30:21.282172	2012-08-02 05:38:35.926903
8328	1-8328	7	37	4156	18	9579	0	6	2012-08-02 05:38:35.926903	2012-08-02 05:38:35.926903
7343	1-7343	8	64	6827	27	2169	0	0	2012-08-02 05:31:50.183797	2012-08-02 05:38:36.324727
6952	1-6952	60	41	5304	46	5749	0	0	2012-08-02 05:29:12.71695	2012-08-02 05:38:36.324727
8297	1-8297	31	49	525	21	1955	0	0	2012-08-02 05:38:23.487117	2012-08-02 05:38:36.324727
8329	1-8329	42	27	992	41	6405	0	7	2012-08-02 05:38:36.324727	2012-08-02 05:38:36.324727
8171	1-8171	75	83	1780	100	8978	0	0	2012-08-02 05:37:32.246722	2012-08-02 05:38:36.744904
8254	1-8254	43	57	3007	81	9057	0	13	2012-08-02 05:38:06.084245	2012-08-02 05:38:36.744904
8333	1-8333	84	93	18	9	1969	0	0	2012-08-02 05:38:37.948959	2012-08-02 05:38:37.948959
5631	1-5631	97	66	7724	18	3082	0	0	2012-08-02 05:20:13.778682	2012-08-02 05:38:38.744386
7548	1-7548	94	22	4269	15	7402	0	8	2012-08-02 05:33:14.003032	2012-08-02 05:38:39.15324
5679	1-5679	80	25	7905	76	5641	0	6	2012-08-02 05:20:33.113775	2012-08-02 05:38:39.15324
4445	1-4445	48	16	5034	62	1780	0	0	2012-08-02 05:12:19.991299	2012-08-02 05:38:39.15324
8336	1-8336	28	76	1489	22	9460	0	0	2012-08-02 05:38:39.15324	2012-08-02 05:38:39.15324
2976	1-2976	57	25	3034	76	1389	275	3	2012-08-02 05:02:29.548842	2012-08-02 05:38:39.15324
6932	1-6932	11	25	7330	73	7500	0	0	2012-08-02 05:29:04.686663	2012-08-02 05:38:39.5623
8338	1-8338	20	34	4	58	4324	0	0	2012-08-02 05:38:39.960031	2012-08-02 05:38:39.960031
7777	1-7777	58	21	3657	6	3136	0	0	2012-08-02 05:34:49.171917	2012-08-02 05:38:40.358037
8285	1-8285	42	17	3061	18	1691	0	5	2012-08-02 05:38:18.579956	2012-08-02 05:38:40.77764
3517	1-3517	66	85	8142	73	2417	0	0	2012-08-02 05:06:08.601816	2012-08-02 05:38:41.175359
7914	1-7914	73	90	6263	92	6388	0	0	2012-08-02 05:35:47.268185	2012-08-02 05:38:41.175359
8043	1-8043	87	23	2853	52	6770	0	5	2012-08-02 05:36:40.415022	2012-08-02 05:38:45.163423
8351	1-8351	33	86	669	23	7055	0	10	2012-08-02 05:38:45.163423	2012-08-02 05:38:45.163423
7999	1-7999	55	6	7351	75	4825	0	2	2012-08-02 05:36:22.725009	2012-08-02 05:38:46.765589
8161	1-8161	66	38	312	29	8621	0	0	2012-08-02 05:37:28.212134	2012-08-02 05:38:47.163196
8341	1-8341	32	92	1637	85	8220	0	6	2012-08-02 05:38:41.175359	2012-08-02 05:38:47.163196
8356	1-8356	50	29	1871	81	7460	0	0	2012-08-02 05:38:47.163196	2012-08-02 05:38:47.163196
6279	1-6279	1	81	6936	1	1680	486	5	2012-08-02 05:24:38.971223	2012-08-02 05:38:47.163196
8248	1-8248	97	77	4032	89	4969	0	3	2012-08-02 05:38:03.542521	2012-08-02 05:38:48.356361
8211	1-8211	42	62	4225	25	3727	0	5	2012-08-02 05:37:48.576736	2012-08-02 05:38:49.597704
8362	1-8362	65	8	1077	62	5065	0	0	2012-08-02 05:38:49.597704	2012-08-02 05:38:49.597704
8363	1-8363	87	18	292	66	7753	0	11	2012-08-02 05:38:50.009831	2012-08-02 05:38:50.009831
4801	1-4801	13	2	3597	42	435	0	0	2012-08-02 05:14:43.638694	2012-08-02 05:38:50.817859
5834	1-5834	11	91	4662	51	6081	0	0	2012-08-02 05:21:36.20779	2012-08-02 05:38:51.637219
8235	1-8235	18	51	7489	99	9513	0	6	2012-08-02 05:37:58.290181	2012-08-02 05:38:51.637219
7998	1-7998	31	18	4964	1	5136	0	0	2012-08-02 05:36:22.316281	2012-08-02 05:38:51.637219
7702	1-7702	18	82	8204	23	7659	0	11	2012-08-02 05:34:17.30418	2012-08-02 05:38:56.082803
8379	1-8379	70	63	928	22	4598	0	6	2012-08-02 05:38:56.492617	2012-08-02 05:38:56.492617
7557	1-7557	97	23	8747	19	5987	0	8	2012-08-02 05:33:17.659766	2012-08-02 05:38:56.891198
7926	1-7926	42	25	5753	41	4262	0	0	2012-08-02 05:35:52.350027	2012-08-02 05:38:57.300928
7254	1-7254	53	14	6980	95	4826	0	6	2012-08-02 05:31:15.22485	2012-08-02 05:38:57.710042
8382	1-8382	69	95	3598	82	2527	0	8	2012-08-02 05:38:57.710042	2012-08-02 05:38:57.710042
8384	1-8384	14	58	3221	66	4472	0	0	2012-08-02 05:38:58.518521	2012-08-02 05:38:58.518521
3161	1-3161	79	64	6280	1	2641	0	3	2012-08-02 05:03:44.653683	2012-08-02 05:38:59.314532
5886	1-5886	26	21	8409	27	3083	0	0	2012-08-02 05:21:57.309521	2012-08-02 05:38:59.314532
3275	1-3275	16	38	9205	74	3473	0	6	2012-08-02 05:04:30.523943	2012-08-02 05:38:59.745635
6339	1-6339	67	38	6479	59	6616	0	5	2012-08-02 05:25:04.087892	2012-08-02 05:38:59.745635
8179	1-8179	1	44	2989	5	9757	0	0	2012-08-02 05:37:35.561491	2012-08-02 05:39:00.564219
3258	1-3258	63	7	8316	5	2325	0	0	2012-08-02 05:04:23.663097	2012-08-02 05:39:00.564219
7966	1-7966	67	62	1899	86	2691	0	6	2012-08-02 05:36:09.235365	2012-08-02 05:39:01.373807
8391	1-8391	19	40	1664	62	5069	0	0	2012-08-02 05:39:01.373807	2012-08-02 05:39:01.373807
8330	1-8330	42	100	2504	57	4731	0	3	2012-08-02 05:38:36.744904	2012-08-02 05:39:02.556681
8396	1-8396	50	6	1287	93	7987	0	0	2012-08-02 05:39:03.363788	2012-08-02 05:39:03.363788
8360	1-8360	5	23	2117	55	5407	0	0	2012-08-02 05:38:48.765225	2012-08-02 05:39:04.634976
7952	1-7952	9	60	5507	56	4237	0	0	2012-08-02 05:36:03.475958	2012-08-02 05:39:04.634976
7642	1-7642	12	76	7837	46	9996	0	5	2012-08-02 05:33:52.352345	2012-08-02 05:39:04.634976
5669	1-5669	83	60	5251	74	3607	0	0	2012-08-02 05:20:29.025608	2012-08-02 05:39:05.456052
5762	1-5762	74	60	3905	74	2395	0	0	2012-08-02 05:21:07.358577	2012-08-02 05:39:05.456052
8401	1-8401	88	55	751	60	9650	0	0	2012-08-02 05:39:05.456052	2012-08-02 05:39:05.456052
7946	1-7946	42	6	8506	57	8735	0	14	2012-08-02 05:36:00.956829	2012-08-02 05:39:06.666062
8394	1-8394	35	51	1791	100	7884	0	8	2012-08-02 05:39:02.556681	2012-08-02 05:39:07.484099
8406	1-8406	97	31	984	71	7140	0	8	2012-08-02 05:39:07.484099	2012-08-02 05:39:07.484099
8386	1-8386	48	27	1049	94	2088	0	8	2012-08-02 05:38:59.314532	2012-08-02 05:39:09.053477
4604	1-4604	32	32	7727	5	6098	0	0	2012-08-02 05:13:23.559569	2012-08-02 05:39:09.053477
8410	1-8410	100	80	1194	27	8382	0	0	2012-08-02 05:39:09.053477	2012-08-02 05:39:09.053477
8411	1-8411	23	5	1997	75	2074	0	0	2012-08-02 05:39:09.474121	2012-08-02 05:39:09.474121
8218	1-8218	22	94	2316	64	9039	0	0	2012-08-02 05:37:51.427408	2012-08-02 05:39:09.861087
8413	1-8413	56	88	242	16	1330	0	4	2012-08-02 05:39:10.258739	2012-08-02 05:39:10.258739
8389	1-8389	14	5	4545	80	8586	0	0	2012-08-02 05:39:00.564219	2012-08-02 05:39:11.165145
5116	1-5116	57	61	9668	84	9491	0	0	2012-08-02 05:16:49.513427	2012-08-02 05:39:11.960578
6948	1-6948	92	62	4830	84	2734	0	0	2012-08-02 05:29:11.12307	2012-08-02 05:39:11.960578
7887	1-7887	96	38	8461	33	4634	0	0	2012-08-02 05:35:35.833234	2012-08-02 05:39:11.960578
8188	1-8188	100	5	4231	22	2122	0	4	2012-08-02 05:37:39.240698	2012-08-02 05:39:11.960578
8417	1-8417	88	84	3124	62	9108	0	0	2012-08-02 05:39:11.960578	2012-08-02 05:39:11.960578
5676	1-5676	84	34	8524	81	7241	0	0	2012-08-02 05:20:31.909515	2012-08-02 05:39:14.745423
8132	1-8132	10	34	9107	91	9713	0	0	2012-08-02 05:37:16.346497	2012-08-02 05:39:14.745423
7305	1-7305	48	47	213	83	246	0	0	2012-08-02 05:31:35.169313	2012-08-02 05:39:14.745423
6516	1-6516	37	77	4819	28	3961	0	0	2012-08-02 05:26:15.703601	2012-08-02 05:39:16.336759
8404	1-8404	8	57	2345	18	1401	0	5	2012-08-02 05:39:06.666062	2012-08-02 05:39:16.336759
8428	1-8428	86	18	230	49	3929	0	0	2012-08-02 05:39:16.336759	2012-08-02 05:39:16.336759
8372	1-8372	77	12	956	38	9769	0	2	2012-08-02 05:38:53.663025	2012-08-02 05:39:17.541672
8233	1-8233	89	82	2114	30	7790	0	3	2012-08-02 05:37:57.483965	2012-08-02 05:39:17.939805
7760	1-7760	84	25	2529	43	1372	0	0	2012-08-02 05:34:42.000449	2012-08-02 05:39:19.133087
8435	1-8435	70	43	3671	98	6286	0	10	2012-08-02 05:39:19.133087	2012-08-02 05:39:19.133087
8099	1-8099	63	16	5181	20	4282	0	2	2012-08-02 05:37:02.888895	2012-08-02 05:39:20.359011
7913	1-7913	99	55	4456	34	8996	0	16	2012-08-02 05:35:46.837333	2012-08-02 05:39:21.950517
8442	1-8442	15	44	2384	55	2987	0	3	2012-08-02 05:39:21.950517	2012-08-02 05:39:21.950517
8443	1-8443	32	29	2640	15	3441	0	3	2012-08-02 05:39:22.348601	2012-08-02 05:39:22.348601
8448	1-8448	74	17	143	31	7075	0	3	2012-08-02 05:39:24.328025	2012-08-02 05:39:24.328025
8450	1-8450	41	42	303	47	7480	0	0	2012-08-02 05:39:25.12333	2012-08-02 05:39:25.12333
1625	1-1625	71	67	9326	10	4040	0	0	2012-08-02 04:52:42.14426	2012-08-02 05:39:25.521251
8105	1-8105	81	76	2875	97	5554	0	10	2012-08-02 05:37:05.330746	2012-08-02 05:39:25.521251
7061	1-7061	51	26	2437	10	1285	0	0	2012-08-02 05:29:56.031967	2012-08-02 05:39:25.521251
8451	1-8451	82	13	2070	76	4306	0	0	2012-08-02 05:39:25.521251	2012-08-02 05:39:25.521251
8206	1-8206	97	32	3194	66	6330	0	5	2012-08-02 05:37:46.532627	2012-08-02 05:39:26.737139
8074	1-8074	12	33	4854	98	6377	0	0	2012-08-02 05:36:52.821791	2012-08-02 05:39:26.737139
4239	1-4239	69	32	3772	10	2770	0	0	2012-08-02 05:10:56.533924	2012-08-02 05:39:26.737139
6434	1-6434	36	32	9447	33	8674	4219	5	2012-08-02 05:25:42.784275	2012-08-02 05:39:26.737139
8455	1-8455	92	82	644	26	6616	0	0	2012-08-02 05:39:27.146224	2012-08-02 05:39:27.146224
8367	1-8367	36	83	1734	91	8412	0	0	2012-08-02 05:38:51.637219	2012-08-02 05:39:28.350353
8221	1-8221	32	15	9192	75	4963	0	13	2012-08-02 05:37:52.642667	2012-08-02 05:39:28.350353
7990	1-7990	48	83	5848	89	3702	0	9	2012-08-02 05:36:19.057028	2012-08-02 05:39:31.543138
8366	1-8366	42	82	5398	94	6920	0	11	2012-08-02 05:38:51.228265	2012-08-02 05:39:31.543138
3233	1-3233	65	8	8515	37	7846	0	0	2012-08-02 05:04:13.982707	2012-08-02 05:39:32.72572
8023	1-8023	84	96	6570	35	7752	0	7	2012-08-02 05:36:32.372105	2012-08-02 05:39:32.72572
5269	1-5269	22	93	6709	35	1573	0	0	2012-08-02 05:17:50.154882	2012-08-02 05:39:32.72572
6519	1-6519	23	8	9338	6	8693	1476	0	2012-08-02 05:26:16.952067	2012-08-02 05:39:32.72572
4133	1-4133	96	55	7696	11	4760	0	0	2012-08-02 05:10:14.779372	2012-08-02 05:39:34.339286
7833	1-7833	61	16	9268	55	7480	0	2	2012-08-02 05:35:13.028572	2012-08-02 05:39:34.339286
8342	1-8342	98	10	3746	13	8167	0	0	2012-08-02 05:38:41.57323	2012-08-02 05:39:35.14557
4066	1-4066	9	91	6555	51	2280	0	0	2012-08-02 05:09:48.124347	2012-08-02 05:39:35.14557
5340	1-5340	76	61	9481	50	8485	0	0	2012-08-02 05:18:18.022422	2012-08-02 05:39:37.896496
7824	1-7824	19	61	2120	97	3349	0	0	2012-08-02 05:35:09.161707	2012-08-02 05:39:37.896496
8133	1-8133	49	4	5090	40	9945	0	7	2012-08-02 05:37:16.744248	2012-08-02 05:39:38.305593
8483	1-8483	19	40	5478	76	5993	0	2	2012-08-02 05:39:38.305593	2012-08-02 05:39:38.305593
1668	1-1668	86	48	7863	20	994	0	0	2012-08-02 04:53:03.015114	2012-08-02 05:39:39.090216
8485	1-8485	98	41	496	48	2710	0	0	2012-08-02 05:39:39.090216	2012-08-02 05:39:39.090216
7935	1-7935	96	23	3659	77	5461	0	0	2012-08-02 05:35:56.260934	2012-08-02 05:39:41.520659
6973	1-6973	77	30	3538	89	3157	0	0	2012-08-02 05:29:21.115935	2012-08-02 05:39:41.520659
8491	1-8491	88	84	493	76	4005	0	0	2012-08-02 05:39:41.520659	2012-08-02 05:39:41.520659
8493	1-8493	24	39	54	96	8933	0	12	2012-08-02 05:39:42.327512	2012-08-02 05:39:42.327512
8258	1-8258	97	87	6588	61	9929	0	0	2012-08-02 05:38:07.686533	2012-08-02 05:39:45.531611
8292	1-8292	80	10	6800	91	8625	0	4	2012-08-02 05:38:21.474379	2012-08-02 05:39:47.133346
8508	1-8508	85	50	87	55	4343	0	4	2012-08-02 05:39:48.339886	2012-08-02 05:39:48.339886
8371	1-8371	86	65	3937	68	6559	0	0	2012-08-02 05:38:53.265551	2012-08-02 05:39:48.748607
8010	1-8010	4	68	9153	76	7637	0	0	2012-08-02 05:36:27.144075	2012-08-02 05:39:48.748607
8224	1-8224	3	27	5148	65	6317	0	0	2012-08-02 05:37:53.846989	2012-08-02 05:39:48.748607
8234	1-8234	97	56	4065	6	6785	0	9	2012-08-02 05:37:57.892588	2012-08-02 05:39:51.134954
3959	1-3959	22	100	9661	13	4960	0	0	2012-08-02 05:09:05.306692	2012-08-02 05:39:52.692544
8519	1-8519	58	91	797	100	6567	0	0	2012-08-02 05:39:52.692544	2012-08-02 05:39:52.692544
8193	1-8193	28	98	3231	93	7995	0	0	2012-08-02 05:37:41.273869	2012-08-02 05:39:53.090473
7712	1-7712	14	98	6085	97	7417	0	4	2012-08-02 05:34:21.535772	2012-08-02 05:39:53.090473
8521	1-8521	6	2	14	98	4129	0	0	2012-08-02 05:39:53.499308	2012-08-02 05:39:53.499308
6267	1-6267	99	3	5924	53	8591	0	0	2012-08-02 05:24:33.611367	2012-08-02 05:39:54.33926
8475	1-8475	52	51	3524	10	9076	0	0	2012-08-02 05:39:35.14557	2012-08-02 05:39:54.33926
7568	1-7568	77	3	3149	31	7072	0	3	2012-08-02 05:33:22.091726	2012-08-02 05:39:54.33926
1508	1-1508	31	3	8004	73	2997	0	0	2012-08-02 04:51:42.681048	2012-08-02 05:39:54.33926
5686	1-5686	97	32	4507	38	2156	0	0	2012-08-02 05:20:35.978425	2012-08-02 05:39:55.1457
7607	1-7607	97	37	3524	98	1595	0	0	2012-08-02 05:33:38.012477	2012-08-02 05:39:55.1457
5212	1-5212	41	56	4633	51	6735	0	0	2012-08-02 05:17:27.758348	2012-08-02 05:39:55.1457
8274	1-8274	31	65	3475	15	9283	0	0	2012-08-02 05:38:14.083078	2012-08-02 05:39:59.122697
7841	1-7841	99	65	6099	78	8634	0	6	2012-08-02 05:35:16.431355	2012-08-02 05:39:59.122697
8536	1-8536	48	66	3014	48	2947	0	5	2012-08-02 05:39:59.54285	2012-08-02 05:39:59.54285
8509	1-8509	88	60	8199	27	6425	0	0	2012-08-02 05:39:48.748607	2012-08-02 05:40:01.951158
5866	1-5866	45	93	7396	4	3475	0	0	2012-08-02 05:21:49.244346	2012-08-02 05:40:02.349042
5887	1-5887	17	81	3370	4	2072	0	0	2012-08-02 05:21:57.707128	2012-08-02 05:40:02.349042
8547	1-8547	91	50	180	60	8419	0	14	2012-08-02 05:40:03.917903	2012-08-02 05:40:03.917903
6930	1-6930	22	27	8845	73	3246	0	0	2012-08-02 05:29:03.846725	2012-08-02 05:40:05.133191
8335	1-8335	95	27	3971	32	8853	0	0	2012-08-02 05:38:38.744386	2012-08-02 05:40:05.133191
8550	1-8550	7	42	1797	27	8159	0	6	2012-08-02 05:40:05.133191	2012-08-02 05:40:05.133191
8501	1-8501	84	79	3553	19	4787	0	0	2012-08-02 05:39:45.531611	2012-08-02 05:40:08.371906
5012	1-5012	33	23	6183	57	7130	0	0	2012-08-02 05:16:07.184941	2012-08-02 05:40:08.371906
6645	1-6645	48	71	8641	20	4601	0	0	2012-08-02 05:27:08.634932	2012-08-02 05:40:08.371906
7901	1-7901	73	12	1334	2	7424	0	2	2012-08-02 05:35:41.699991	2012-08-02 05:40:08.78121
7570	1-7570	6	9	2554	58	7438	0	8	2012-08-02 05:33:22.920275	2012-08-02 05:40:11.244956
7599	1-7599	59	83	4021	95	9860	0	0	2012-08-02 05:33:34.753315	2012-08-02 05:40:11.644543
6996	1-6996	59	9	4188	73	3200	0	2	2012-08-02 05:29:30.143159	2012-08-02 05:40:12.053327
7398	1-7398	80	79	3894	48	4183	0	4	2012-08-02 05:32:12.11733	2012-08-02 05:40:12.053327
8567	1-8567	40	48	2228	9	8471	0	6	2012-08-02 05:40:12.053327	2012-08-02 05:40:12.053327
8568	1-8568	15	10	303	3	5210	0	5	2012-08-02 05:40:12.451128	2012-08-02 05:40:12.451128
8571	1-8571	54	42	177	97	1051	0	0	2012-08-02 05:40:13.633446	2012-08-02 05:40:13.633446
1879	1-1879	59	36	8374	44	3675	0	0	2012-08-02 04:54:38.07966	2012-08-02 05:40:16.063536
8577	1-8577	57	51	2450	36	6976	0	0	2012-08-02 05:40:16.063536	2012-08-02 05:40:16.063536
8479	1-8479	25	95	7324	69	9960	0	2	2012-08-02 05:39:36.71463	2012-08-02 05:40:17.290369
8114	1-8114	70	76	3366	69	8185	0	0	2012-08-02 05:37:08.977078	2012-08-02 05:40:17.290369
7608	1-7608	90	12	2078	52	2362	0	0	2012-08-02 05:33:38.421114	2012-08-02 05:40:17.290369
3317	1-3317	57	55	9843	33	5620	0	0	2012-08-02 05:04:47.474683	2012-08-02 05:40:18.085757
8558	1-8558	86	44	4793	23	8170	0	0	2012-08-02 05:40:08.371906	2012-08-02 05:40:18.085757
3643	1-3643	89	55	7387	73	4257	0	0	2012-08-02 05:06:59.297275	2012-08-02 05:40:18.085757
8582	1-8582	51	49	1361	44	9543	0	5	2012-08-02 05:40:18.085757	2012-08-02 05:40:18.085757
6204	1-6204	22	28	9400	57	7580	0	8	2012-08-02 05:24:07.364901	2012-08-02 05:40:19.743317
7199	1-7199	39	30	4570	79	3988	0	0	2012-08-02 05:30:53.002866	2012-08-02 05:40:20.152002
7491	1-7491	98	57	8276	93	4179	0	4	2012-08-02 05:32:50.390203	2012-08-02 05:40:20.152002
8348	1-8348	39	20	476	52	325	0	0	2012-08-02 05:38:43.959659	2012-08-02 05:40:20.152002
8588	1-8588	21	89	24	84	3600	0	4	2012-08-02 05:40:20.549757	2012-08-02 05:40:20.549757
8505	1-8505	46	53	4574	10	7511	0	0	2012-08-02 05:39:47.133346	2012-08-02 05:40:22.560516
8593	1-8593	36	46	403	53	5872	0	4	2012-08-02 05:40:22.560516	2012-08-02 05:40:22.560516
8543	1-8543	27	4	1506	92	8815	0	4	2012-08-02 05:40:02.349042	2012-08-02 05:40:23.367235
8597	1-8597	46	64	1207	26	6393	0	0	2012-08-02 05:40:24.173794	2012-08-02 05:40:24.173794
5945	1-5945	72	99	3696	53	5485	0	7	2012-08-02 05:22:21.607945	2012-08-02 05:40:25.388569
7884	1-7884	55	99	3511	53	5121	0	9	2012-08-02 05:35:34.562745	2012-08-02 05:40:25.388569
8600	1-8600	35	53	8307	99	7621	0	0	2012-08-02 05:40:25.388569	2012-08-02 05:40:25.388569
8312	1-8312	11	46	4217	43	4902	0	9	2012-08-02 05:38:29.496924	2012-08-02 05:40:26.195277
8346	1-8346	74	80	3851	7	9090	0	0	2012-08-02 05:38:43.174947	2012-08-02 05:40:27.001819
7008	1-7008	76	30	3097	14	5232	0	5	2012-08-02 05:29:34.94122	2012-08-02 05:40:28.592764
8532	1-8532	66	27	1062	44	2305	0	6	2012-08-02 05:39:57.929655	2012-08-02 05:40:30.570252
7595	1-7595	82	1	9524	27	6583	2674	18	2012-08-02 05:33:33.106858	2012-08-02 05:40:30.570252
4660	1-4660	30	36	9663	57	8151	0	0	2012-08-02 05:13:46.642387	2012-08-02 05:40:31.354772
8573	1-8573	18	36	5586	47	2465	0	0	2012-08-02 05:40:14.440059	2012-08-02 05:40:31.354772
7958	1-7958	2	31	7861	43	7108	0	0	2012-08-02 05:36:06.019009	2012-08-02 05:40:31.354772
8615	1-8615	21	85	3219	36	8254	0	6	2012-08-02 05:40:31.354772	2012-08-02 05:40:31.354772
8049	1-8049	36	79	4345	64	4815	0	0	2012-08-02 05:36:42.823499	2012-08-02 05:40:31.763548
7577	1-7577	21	77	4749	87	7880	0	3	2012-08-02 05:33:25.748511	2012-08-02 05:40:32.570378
8073	1-8073	62	12	8611	68	6014	0	4	2012-08-02 05:36:52.424133	2012-08-02 05:40:32.570378
8340	1-8340	99	18	3361	86	9374	0	9	2012-08-02 05:38:40.77764	2012-08-02 05:40:33.78574
7239	1-7239	48	7	5324	31	7063	0	0	2012-08-02 05:31:09.391228	2012-08-02 05:40:33.78574
7089	1-7089	76	93	5090	17	4959	0	0	2012-08-02 05:30:07.194882	2012-08-02 05:40:33.78574
8621	1-8621	54	2	8292	18	8160	0	4	2012-08-02 05:40:33.78574	2012-08-02 05:40:33.78574
8523	1-8523	18	10	6352	3	8387	0	0	2012-08-02 05:39:54.33926	2012-08-02 05:40:34.592473
7084	1-7084	8	39	621	90	327	0	0	2012-08-02 05:30:05.14959	2012-08-02 05:40:35.409584
8530	1-8530	97	44	7347	55	7928	0	5	2012-08-02 05:39:57.134529	2012-08-02 05:40:35.409584
7527	1-7527	40	39	9448	1	9462	0	0	2012-08-02 05:33:05.367856	2012-08-02 05:40:35.409584
8520	1-8520	83	97	9333	6	9477	0	0	2012-08-02 05:39:53.090473	2012-08-02 05:40:35.409584
4868	1-4868	78	39	6182	40	5614	0	4	2012-08-02 05:15:10.644687	2012-08-02 05:40:35.409584
3429	1-3429	37	39	6825	44	3022	967	0	2012-08-02 05:05:32.998591	2012-08-02 05:40:35.409584
7407	1-7407	3	10	5137	26	2233	0	0	2012-08-02 05:32:15.874383	2012-08-02 05:40:37.000549
8280	1-8280	61	6	7346	98	6904	0	0	2012-08-02 05:38:16.54689	2012-08-02 05:40:38.193692
7751	1-7751	14	65	8275	6	4786	0	2	2012-08-02 05:34:38.178497	2012-08-02 05:40:38.193692
7567	1-7567	24	10	4348	59	2256	0	0	2012-08-02 05:33:21.693993	2012-08-02 05:40:38.602625
7394	1-7394	63	10	8990	75	3800	0	3	2012-08-02 05:32:10.449093	2012-08-02 05:40:38.602625
7498	1-7498	61	17	7357	76	2779	0	0	2012-08-02 05:32:53.308685	2012-08-02 05:40:39.0004
5940	1-5940	8	67	3672	87	3185	0	0	2012-08-02 05:22:19.508341	2012-08-02 05:40:40.648658
8638	1-8638	86	13	618	67	4212	0	0	2012-08-02 05:40:40.648658	2012-08-02 05:40:40.648658
8017	1-8017	32	60	6332	46	2669	0	2	2012-08-02 05:36:29.928076	2012-08-02 05:40:41.035243
7146	1-7146	85	60	2473	27	210	0	0	2012-08-02 05:30:30.917337	2012-08-02 05:40:41.035243
8639	1-8639	89	42	876	60	6852	0	0	2012-08-02 05:40:41.035243	2012-08-02 05:40:41.035243
8164	1-8164	88	76	3886	60	9955	0	3	2012-08-02 05:37:29.416694	2012-08-02 05:40:41.444129
8640	1-8640	44	60	1737	69	6756	0	11	2012-08-02 05:40:41.444129	2012-08-02 05:40:41.444129
8618	1-8618	22	68	4154	77	8163	0	7	2012-08-02 05:40:32.570378	2012-08-02 05:40:43.045957
8632	1-8632	69	95	5319	65	8441	0	8	2012-08-02 05:40:38.193692	2012-08-02 05:40:44.636934
5231	1-5231	19	98	8824	72	7517	0	0	2012-08-02 05:17:35.194309	2012-08-02 05:40:44.636934
6572	1-6572	61	90	8214	63	8155	0	0	2012-08-02 05:26:38.773908	2012-08-02 05:40:45.045782
2325	1-2325	38	83	9930	80	3802	0	0	2012-08-02 04:57:51.888591	2012-08-02 05:40:45.045782
8566	1-8566	31	27	1054	83	5643	0	0	2012-08-02 05:40:11.644543	2012-08-02 05:40:45.045782
7357	1-7357	20	46	5461	63	4681	0	0	2012-08-02 05:31:55.665975	2012-08-02 05:40:45.045782
4417	1-4417	60	67	5868	15	7330	0	6	2012-08-02 05:12:08.953247	2012-08-02 05:40:45.454809
8380	1-8380	46	76	6925	23	6303	0	2	2012-08-02 05:38:56.891198	2012-08-02 05:40:45.454809
8440	1-8440	100	76	4060	30	4808	0	6	2012-08-02 05:39:21.154909	2012-08-02 05:40:45.454809
8390	1-8390	36	96	3089	28	5766	0	0	2012-08-02 05:39:00.976259	2012-08-02 05:40:45.874827
8215	1-8215	28	78	3971	28	6447	0	10	2012-08-02 05:37:50.201113	2012-08-02 05:40:45.874827
8486	1-8486	18	15	6582	29	4784	0	0	2012-08-02 05:39:39.51007	2012-08-02 05:40:46.272495
7764	1-7764	9	87	4889	47	1704	0	0	2012-08-02 05:34:43.70206	2012-08-02 05:40:47.068242
8271	1-8271	25	23	2307	71	6093	0	17	2012-08-02 05:38:12.878834	2012-08-02 05:40:47.068242
8438	1-8438	19	20	3724	30	9163	0	6	2012-08-02 05:39:20.359011	2012-08-02 05:40:47.476959
4219	1-4219	10	93	5110	20	1722	998	7	2012-08-02 05:10:48.744614	2012-08-02 05:40:47.476959
6059	1-6059	22	78	5561	76	2111	0	0	2012-08-02 05:23:08.956263	2012-08-02 05:40:47.874553
6460	1-6460	57	76	6051	4	4434	0	0	2012-08-02 05:25:53.127404	2012-08-02 05:40:47.874553
8656	1-8656	67	4	555	78	8578	0	0	2012-08-02 05:40:47.874553	2012-08-02 05:40:47.874553
7812	1-7812	25	82	4506	57	6515	0	7	2012-08-02 05:35:04.06514	2012-08-02 05:40:49.078942
5206	1-5206	51	12	6182	89	5952	0	0	2012-08-02 05:17:25.382968	2012-08-02 05:40:49.078942
8659	1-8659	90	43	1840	12	4255	0	4	2012-08-02 05:40:49.078942	2012-08-02 05:40:49.078942
8662	1-8662	72	17	43	78	7896	0	0	2012-08-02 05:40:50.283218	2012-08-02 05:40:50.283218
5324	1-5324	46	82	7452	28	5847	0	0	2012-08-02 05:18:11.678403	2012-08-02 05:40:51.111975
8664	1-8664	28	91	2770	82	5497	0	11	2012-08-02 05:40:51.111975	2012-08-02 05:40:51.111975
8626	1-8626	57	54	630	16	687	0	8	2012-08-02 05:40:35.80754	2012-08-02 05:40:51.907274
8666	1-8666	52	88	367	54	1996	0	0	2012-08-02 05:40:51.907274	2012-08-02 05:40:51.907274
8667	1-8667	30	70	1727	55	2263	0	0	2012-08-02 05:40:52.294194	2012-08-02 05:40:52.294194
8636	1-8636	99	1	3278	90	7702	0	2	2012-08-02 05:40:39.809073	2012-08-02 05:40:52.691863
8668	1-8668	97	71	624	1	2903	0	0	2012-08-02 05:40:52.691863	2012-08-02 05:40:52.691863
8669	1-8669	65	5	1213	41	8512	0	0	2012-08-02 05:40:53.100812	2012-08-02 05:40:53.100812
8671	1-8671	29	92	4758	96	6344	0	0	2012-08-02 05:40:53.89614	2012-08-02 05:40:53.89614
8325	1-8325	4	55	5271	32	3941	0	9	2012-08-02 05:38:34.722777	2012-08-02 05:40:54.680854
8673	1-8673	96	78	2460	55	9751	0	2	2012-08-02 05:40:54.680854	2012-08-02 05:40:54.680854
8563	1-8563	29	1	3114	4	8442	0	10	2012-08-02 05:40:10.416179	2012-08-02 05:40:56.326727
8652	1-8652	62	29	4887	15	7577	0	4	2012-08-02 05:40:46.272495	2012-08-02 05:40:57.917519
8368	1-8368	70	42	6261	91	6394	0	0	2012-08-02 05:38:52.048347	2012-08-02 05:40:58.712942
8683	1-8683	75	18	892	96	8018	0	0	2012-08-02 05:40:58.712942	2012-08-02 05:40:58.712942
8300	1-8300	77	45	3420	51	7998	0	0	2012-08-02 05:38:24.66889	2012-08-02 05:41:00.315031
8690	1-8690	98	84	513	75	5033	0	0	2012-08-02 05:41:01.618852	2012-08-02 05:41:01.618852
8415	1-8415	28	80	8504	68	6249	0	0	2012-08-02 05:39:11.165145	2012-08-02 05:41:02.118379
8320	1-8320	12	76	3282	67	9617	0	0	2012-08-02 05:38:32.668111	2012-08-02 05:41:02.524399
8323	1-8323	45	76	5659	26	5883	0	0	2012-08-02 05:38:33.916406	2012-08-02 05:41:02.524399
8473	1-8473	15	29	4358	16	7458	0	3	2012-08-02 05:39:34.339286	2012-08-02 05:41:03.750739
5767	1-5767	51	71	3983	51	3628	0	0	2012-08-02 05:21:09.358464	2012-08-02 05:41:05.352941
6988	1-6988	51	100	8377	7	5682	0	5	2012-08-02 05:29:27.004739	2012-08-02 05:41:05.751063
8034	1-8034	99	18	6407	7	8287	0	0	2012-08-02 05:36:36.824441	2012-08-02 05:41:05.751063
8700	1-8700	34	7	1434	30	6314	0	0	2012-08-02 05:41:05.751063	2012-08-02 05:41:05.751063
8482	1-8482	76	50	4460	87	6614	0	4	2012-08-02 05:39:37.896496	2012-08-02 05:41:06.957549
8024	1-8024	95	23	2307	87	5777	0	3	2012-08-02 05:36:32.780553	2012-08-02 05:41:06.957549
8703	1-8703	78	87	4096	50	7772	0	5	2012-08-02 05:41:06.957549	2012-08-02 05:41:06.957549
7993	1-7993	3	62	6203	83	7246	0	0	2012-08-02 05:36:20.294506	2012-08-02 05:41:07.377522
8704	1-8704	64	24	294	62	2626	0	0	2012-08-02 05:41:07.377522	2012-08-02 05:41:07.377522
8706	1-8706	62	60	2015	33	5961	0	0	2012-08-02 05:41:08.195081	2012-08-02 05:41:08.195081
4383	1-4383	74	98	4031	58	4039	0	5	2012-08-02 05:11:55.217284	2012-08-02 05:41:09.399261
8709	1-8709	87	58	1167	9	4088	0	0	2012-08-02 05:41:09.399261	2012-08-02 05:41:09.399261
8715	1-8715	84	94	1813	9	2926	0	2	2012-08-02 05:41:11.785989	2012-08-02 05:41:11.785989
8434	1-8434	8	15	7273	83	4936	0	9	2012-08-02 05:39:18.746321	2012-08-02 05:41:12.978788
6966	1-6966	18	72	7404	8	3645	0	3	2012-08-02 05:29:18.275978	2012-08-02 05:41:12.978788
1169	1-1169	96	8	7418	24	2006	1119	0	2012-08-02 04:48:42.659222	2012-08-02 05:41:12.978788
8722	1-8722	36	21	183	6	2778	0	2	2012-08-02 05:41:14.580949	2012-08-02 05:41:14.580949
8692	1-8692	45	24	1199	76	9819	0	4	2012-08-02 05:41:02.524399	2012-08-02 05:41:14.978735
7899	1-7899	73	99	5207	63	7104	0	7	2012-08-02 05:35:40.849404	2012-08-02 05:41:14.978735
8608	1-8608	45	100	2498	30	2263	0	0	2012-08-02 05:40:28.592764	2012-08-02 05:41:16.205241
8279	1-8279	11	94	2811	66	9972	0	4	2012-08-02 05:38:16.126915	2012-08-02 05:41:17.42052
8729	1-8729	61	88	2460	36	8014	0	0	2012-08-02 05:41:17.42052	2012-08-02 05:41:17.42052
8732	1-8732	72	53	2920	3	8949	0	0	2012-08-02 05:41:18.624375	2012-08-02 05:41:18.624375
8735	1-8735	22	10	420	75	1982	0	8	2012-08-02 05:41:19.81781	2012-08-02 05:41:19.81781
8535	1-8535	88	50	4501	65	8601	0	0	2012-08-02 05:39:59.122697	2012-08-02 05:41:20.635177
8727	1-8727	7	52	962	1	704	0	10	2012-08-02 05:41:16.613796	2012-08-02 05:41:20.635177
8737	1-8737	8	65	261	52	8079	0	0	2012-08-02 05:41:20.635177	2012-08-02 05:41:20.635177
6016	1-6016	93	30	6037	24	3111	798	3	2012-08-02 05:22:51.167139	2012-08-02 05:41:20.635177
8478	1-8478	55	85	5576	21	8797	0	6	2012-08-02 05:39:36.339039	2012-08-02 05:41:21.055129
8739	1-8739	32	48	91	79	2220	0	0	2012-08-02 05:41:21.44602	2012-08-02 05:41:21.44602
8685	1-8685	34	93	902	29	2859	0	2	2012-08-02 05:40:59.53058	2012-08-02 05:41:21.832655
8620	1-8620	26	87	284	40	2935	0	6	2012-08-02 05:40:33.388017	2012-08-02 05:41:22.230455
8561	1-8561	17	90	4563	46	3494	0	3	2012-08-02 05:40:09.598873	2012-08-02 05:41:22.230455
8741	1-8741	88	65	863	90	9014	0	12	2012-08-02 05:41:22.230455	2012-08-02 05:41:22.230455
5921	1-5921	68	52	2826	53	7128	0	15	2012-08-02 05:22:11.586419	2012-08-02 05:41:23.083455
8418	1-8418	43	87	4334	92	8604	0	9	2012-08-02 05:39:12.370427	2012-08-02 05:41:23.083455
8454	1-8454	80	37	3727	32	4707	0	5	2012-08-02 05:39:26.737139	2012-08-02 05:41:23.083455
8199	1-8199	30	53	8520	25	7349	0	0	2012-08-02 05:37:43.704344	2012-08-02 05:41:23.083455
6214	1-6214	41	19	1870	40	1593	0	0	2012-08-02 05:24:11.474757	2012-08-02 05:41:23.083455
8381	1-8381	49	97	7871	25	9293	0	2	2012-08-02 05:38:57.300928	2012-08-02 05:41:23.083455
8745	1-8745	61	91	20	59	1679	0	0	2012-08-02 05:41:23.923007	2012-08-02 05:41:23.923007
6046	1-6046	38	19	7563	8	7351	0	6	2012-08-02 05:23:03.706958	2012-08-02 05:41:24.320693
7207	1-7207	92	19	7402	65	9406	0	0	2012-08-02 05:30:56.339522	2012-08-02 05:41:24.320693
7988	1-7988	4	68	6252	64	7368	0	0	2012-08-02 05:36:18.250487	2012-08-02 05:41:24.320693
8746	1-8746	46	8	824	19	9492	0	0	2012-08-02 05:41:24.320693	2012-08-02 05:41:24.320693
8743	1-8743	3	25	8786	19	8801	0	9	2012-08-02 05:41:23.083455	2012-08-02 05:41:24.729739
2794	1-2794	10	55	8430	17	8728	977	0	2012-08-02 05:01:13.829424	2012-08-02 05:41:24.729739
7930	1-7930	98	17	8665	26	7150	2446	5	2012-08-02 05:35:54.076116	2012-08-02 05:41:24.729739
7419	1-7419	62	73	1565	2	7048	0	0	2012-08-02 05:32:20.902927	2012-08-02 05:41:25.558295
8399	1-8399	68	73	6965	76	8604	0	6	2012-08-02 05:39:04.634976	2012-08-02 05:41:25.558295
8502	1-8502	39	73	3736	42	7123	0	0	2012-08-02 05:39:45.962414	2012-08-02 05:41:25.558295
8749	1-8749	70	51	1109	73	8007	0	0	2012-08-02 05:41:25.558295	2012-08-02 05:41:25.558295
6444	1-6444	81	5	4621	29	1967	0	0	2012-08-02 05:25:46.794253	2012-08-02 05:41:25.978331
4960	1-4960	41	87	9787	15	8601	0	0	2012-08-02 05:15:46.62001	2012-08-02 05:41:26.37614
8319	1-8319	33	87	4565	31	3651	0	12	2012-08-02 05:38:32.281109	2012-08-02 05:41:26.37614
217	1-217	52	38	8299	59	4654	0	0	2012-08-02 04:43:56.314333	2012-08-02 05:41:26.37614
8562	1-8562	13	1	1944	42	7377	0	5	2012-08-02 05:40:09.996439	2012-08-02 05:41:26.784947
6276	1-6276	68	61	8729	42	8557	0	5	2012-08-02 05:24:37.689803	2012-08-02 05:41:26.784947
8752	1-8752	10	42	3516	61	8264	0	0	2012-08-02 05:41:26.784947	2012-08-02 05:41:26.784947
1115	1-1115	68	1	9787	42	5901	3201	0	2012-08-02 04:48:17.62229	2012-08-02 05:41:26.784947
8633	1-8633	75	37	2311	10	9437	0	6	2012-08-02 05:40:38.602625	2012-08-02 05:41:27.613705
8590	1-8590	5	59	5693	15	4613	0	19	2012-08-02 05:40:21.367553	2012-08-02 05:41:28.408974
8756	1-8756	44	51	1605	59	5969	0	4	2012-08-02 05:41:28.408974	2012-08-02 05:41:28.408974
8541	1-8541	34	90	1423	61	1004	0	7	2012-08-02 05:40:01.542502	2012-08-02 05:41:28.817895
8691	1-8691	51	12	1606	45	3234	0	0	2012-08-02 05:41:02.118379	2012-08-02 05:41:29.646372
8641	1-8641	28	17	2448	48	7313	0	0	2012-08-02 05:40:41.864044	2012-08-02 05:41:30.03304
8602	1-8602	97	77	1234	9	1155	0	0	2012-08-02 05:40:26.195277	2012-08-02 05:41:32.077116
8765	1-8765	44	9	487	63	5822	0	0	2012-08-02 05:41:32.077116	2012-08-02 05:41:32.077116
8359	1-8359	13	89	3632	86	8159	0	12	2012-08-02 05:38:48.356361	2012-08-02 05:41:32.507891
7616	1-7616	7	89	7690	86	7528	0	6	2012-08-02 05:33:41.68057	2012-08-02 05:41:32.507891
4739	1-4739	21	12	4351	48	3844	0	0	2012-08-02 05:14:18.599673	2012-08-02 05:41:32.507891
6512	1-6512	11	89	8550	76	6863	0	0	2012-08-02 05:26:14.044356	2012-08-02 05:41:32.507891
6698	1-6698	23	31	8788	15	7516	0	5	2012-08-02 05:27:29.805254	2012-08-02 05:41:33.756512
6238	1-6238	28	59	7618	72	3057	0	0	2012-08-02 05:24:21.524063	2012-08-02 05:41:34.154091
8770	1-8770	80	72	1546	59	9225	0	13	2012-08-02 05:41:34.154091	2012-08-02 05:41:34.154091
3874	1-3874	12	72	4933	59	2437	993	4	2012-08-02 05:08:31.313765	2012-08-02 05:41:34.154091
6810	1-6810	7	39	5997	57	7689	0	0	2012-08-02 05:28:14.39123	2012-08-02 05:41:34.961066
8392	1-8392	63	46	7600	63	6054	0	2	2012-08-02 05:39:01.761235	2012-08-02 05:41:35.369431
2207	1-2207	66	8	9658	46	7770	0	0	2012-08-02 04:57:00.488554	2012-08-02 05:41:35.369431
8257	1-8257	16	52	4160	100	8525	0	8	2012-08-02 05:38:07.277417	2012-08-02 05:41:36.187319
8775	1-8775	21	13	1790	52	4387	0	9	2012-08-02 05:41:36.187319	2012-08-02 05:41:36.187319
3486	1-3486	6	8	5803	57	7909	0	0	2012-08-02 05:05:56.136595	2012-08-02 05:41:37.402603
7181	1-7181	55	5	6256	69	5751	0	3	2012-08-02 05:30:45.489879	2012-08-02 05:41:37.402603
8778	1-8778	16	69	5923	8	9899	0	0	2012-08-02 05:41:37.402603	2012-08-02 05:41:37.402603
8086	1-8086	11	22	3355	58	8427	0	0	2012-08-02 05:36:57.627584	2012-08-02 05:41:37.800244
8779	1-8779	7	71	1665	39	4776	0	0	2012-08-02 05:41:37.800244	2012-08-02 05:41:37.800244
8625	1-8625	2	40	8166	39	9419	0	0	2012-08-02 05:40:35.409584	2012-08-02 05:41:38.606973
8327	1-8327	77	47	3751	43	7192	0	0	2012-08-02 05:38:35.529165	2012-08-02 05:41:38.993303
7956	1-7956	19	81	4689	13	5153	0	0	2012-08-02 05:36:05.165848	2012-08-02 05:41:38.993303
8734	1-8734	26	46	4214	28	6058	0	14	2012-08-02 05:41:19.419988	2012-08-02 05:41:38.993303
8772	1-8772	82	79	4456	39	9769	0	3	2012-08-02 05:41:34.961066	2012-08-02 05:41:41.402094
8789	1-8789	35	93	117	82	6724	0	0	2012-08-02 05:41:41.799626	2012-08-02 05:41:41.799626
8766	1-8766	37	59	3447	89	5146	0	3	2012-08-02 05:41:32.507891	2012-08-02 05:41:43.821424
7042	1-7042	71	99	7476	43	7401	0	0	2012-08-02 05:29:48.507541	2012-08-02 05:41:43.821424
8794	1-8794	60	71	3778	99	8660	0	8	2012-08-02 05:41:43.821424	2012-08-02 05:41:43.821424
7115	1-7115	36	23	8892	79	8755	0	0	2012-08-02 05:30:17.967523	2012-08-02 05:41:44.241105
8654	1-8654	79	64	3257	23	7121	0	2	2012-08-02 05:40:47.068242	2012-08-02 05:41:44.241105
7949	1-7949	46	12	2808	95	9752	0	6	2012-08-02 05:36:02.204939	2012-08-02 05:41:44.638875
8781	1-8781	32	69	4236	40	9240	0	6	2012-08-02 05:41:38.606973	2012-08-02 05:41:44.638875
6834	1-6834	29	88	7228	95	5665	0	12	2012-08-02 05:28:24.248683	2012-08-02 05:41:44.638875
8796	1-8796	92	95	1619	69	7576	0	4	2012-08-02 05:41:44.638875	2012-08-02 05:41:44.638875
8469	1-8469	32	35	2687	8	8090	0	0	2012-08-02 05:39:32.72572	2012-08-02 05:41:45.445676
8154	1-8154	30	53	3445	86	1139	0	0	2012-08-02 05:37:25.217781	2012-08-02 05:41:46.240961
8800	1-8800	32	48	2528	25	4411	0	9	2012-08-02 05:41:46.240961	2012-08-02 05:41:46.240961
8801	1-8801	21	99	392	51	1221	0	0	2012-08-02 05:41:46.638894	2012-08-02 05:41:46.638894
4537	1-4537	60	18	7232	22	3975	0	0	2012-08-02 05:12:56.500217	2012-08-02 05:41:47.467692
8803	1-8803	51	22	1800	18	7817	0	0	2012-08-02 05:41:47.467692	2012-08-02 05:41:47.467692
5979	1-5979	33	18	9074	93	8155	0	0	2012-08-02 05:22:35.674238	2012-08-02 05:41:50.262525
2766	1-2766	59	32	9336	16	7496	0	0	2012-08-02 05:01:02.124819	2012-08-02 05:41:51.07983
8812	1-8812	53	85	410	32	7310	0	0	2012-08-02 05:41:51.07983	2012-08-02 05:41:51.07983
7894	1-7894	6	84	5229	64	6067	0	9	2012-08-02 05:35:38.760974	2012-08-02 05:41:53.103861
4865	1-4865	49	97	6245	15	6002	0	0	2012-08-02 05:15:09.462856	2012-08-02 05:41:54.32969
1851	1-1851	31	22	9956	16	7797	0	0	2012-08-02 04:54:24.831429	2012-08-02 05:41:54.32969
8820	1-8820	28	16	5093	97	9091	0	0	2012-08-02 05:41:54.32969	2012-08-02 05:41:54.32969
8466	1-8466	88	89	3345	82	7941	0	0	2012-08-02 05:39:31.543138	2012-08-02 05:41:54.738828
6578	1-6578	48	94	6024	53	7530	0	4	2012-08-02 05:26:41.239513	2012-08-02 05:41:55.147501
6919	1-6919	8	68	8812	55	9269	0	3	2012-08-02 05:28:59.307302	2012-08-02 05:41:55.147501
8769	1-8769	78	16	1809	7	5392	0	7	2012-08-02 05:41:33.756512	2012-08-02 05:41:55.147501
6451	1-6451	87	94	7156	87	5426	0	11	2012-08-02 05:25:49.556475	2012-08-02 05:41:55.147501
8499	1-8499	13	14	8141	100	5389	0	0	2012-08-02 05:39:44.75794	2012-08-02 05:41:56.35202
8677	1-8677	62	60	9753	72	7853	0	9	2012-08-02 05:40:56.326727	2012-08-02 05:41:59.124768
8832	1-8832	14	14	1008	3	3358	0	7	2012-08-02 05:41:59.124768	2012-08-02 05:41:59.124768
7976	1-7976	18	81	7734	54	9313	0	0	2012-08-02 05:36:13.278647	2012-08-02 05:42:01.146484
8181	1-8181	47	13	6117	54	9491	0	0	2012-08-02 05:37:36.378924	2012-08-02 05:42:01.146484
8750	1-8750	12	58	932	5	6792	0	12	2012-08-02 05:41:25.978331	2012-08-02 05:42:01.146484
6845	1-6845	28	58	4934	48	6734	0	0	2012-08-02 05:28:28.778875	2012-08-02 05:42:01.146484
2319	1-2319	61	95	8689	54	3636	0	0	2012-08-02 04:57:49.248015	2012-08-02 05:42:01.146484
5481	1-5481	30	58	7660	63	3725	0	5	2012-08-02 05:19:13.45687	2012-08-02 05:42:01.146484
7818	1-7818	52	20	3565	4	5285	0	5	2012-08-02 05:35:06.598435	2012-08-02 05:42:02.350931
8840	1-8840	32	88	4132	20	7279	0	0	2012-08-02 05:42:02.350931	2012-08-02 05:42:02.350931
8548	1-8548	33	93	4825	19	3370	0	0	2012-08-02 05:40:04.337832	2012-08-02 05:42:02.748915
6116	1-6116	24	82	5485	62	4048	0	0	2012-08-02 05:23:31.751228	2012-08-02 05:42:02.748915
5586	1-5586	8	47	4326	92	9781	0	0	2012-08-02 05:19:55.500792	2012-08-02 05:42:02.748915
8670	1-8670	62	47	2406	32	4928	0	0	2012-08-02 05:40:53.498404	2012-08-02 05:42:02.748915
8289	1-8289	100	47	4283	93	9272	0	4	2012-08-02 05:38:20.192767	2012-08-02 05:42:02.748915
7821	1-7821	74	57	3495	43	4405	0	0	2012-08-02 05:35:07.891005	2012-08-02 05:42:02.748915
8841	1-8841	27	44	3402	47	8082	0	0	2012-08-02 05:42:02.748915	2012-08-02 05:42:02.748915
1721	1-1721	71	47	6142	57	457	457	0	2012-08-02 04:53:26.849163	2012-08-02 05:42:02.748915
3977	1-3977	12	43	6039	62	290	290	5	2012-08-02 05:09:12.501251	2012-08-02 05:42:02.748915
4143	1-4143	50	32	8629	81	9797	18	4	2012-08-02 05:10:18.779464	2012-08-02 05:42:02.748915
7301	1-7301	52	49	3412	88	7894	0	19	2012-08-02 05:31:33.600651	2012-08-02 05:42:03.168998
8842	1-8842	91	64	2484	30	4609	0	0	2012-08-02 05:42:03.168998	2012-08-02 05:42:03.168998
8718	1-8718	98	67	1376	60	8812	0	14	2012-08-02 05:41:12.978788	2012-08-02 05:42:03.975621
8544	1-8544	15	11	4901	68	8492	0	8	2012-08-02 05:40:02.746833	2012-08-02 05:42:04.781867
8846	1-8846	28	12	328	11	5674	0	0	2012-08-02 05:42:04.781867	2012-08-02 05:42:04.781867
8847	1-8847	31	28	2804	34	6720	0	0	2012-08-02 05:42:05.17983	2012-08-02 05:42:05.17983
8617	1-8617	62	52	2942	30	4817	0	0	2012-08-02 05:40:32.172521	2012-08-02 05:42:05.589161
8149	1-8149	90	64	2948	17	3165	0	0	2012-08-02 05:37:23.207152	2012-08-02 05:42:07.202739
8822	1-8822	59	55	2350	94	9441	0	0	2012-08-02 05:41:55.147501	2012-08-02 05:42:07.202739
6263	1-6263	81	77	8519	89	7083	0	0	2012-08-02 05:24:31.965252	2012-08-02 05:42:07.202739
8539	1-8539	12	96	5893	78	7434	0	10	2012-08-02 05:40:00.758351	2012-08-02 05:42:08.009346
8860	1-8860	24	18	1125	58	2186	0	14	2012-08-02 05:42:10.450766	2012-08-02 05:42:10.450766
8039	1-8039	73	23	3275	56	3894	0	8	2012-08-02 05:36:38.835253	2012-08-02 05:42:11.26841
5788	1-5788	41	5	9757	95	5067	0	5	2012-08-02 05:21:17.710403	2012-08-02 05:42:11.26841
7585	1-7585	84	71	5511	4	6882	0	8	2012-08-02 05:33:29.030089	2012-08-02 05:42:12.914584
8334	1-8334	46	2	9554	31	5053	0	4	2012-08-02 05:38:38.34687	2012-08-02 05:42:13.323242
7169	1-7169	28	62	7997	68	7587	0	7	2012-08-02 05:30:40.463056	2012-08-02 05:42:13.323242
7277	1-7277	62	87	1523	31	643	0	4	2012-08-02 05:31:24.253377	2012-08-02 05:42:13.323242
3779	1-3779	56	7	9541	31	4186	4186	0	2012-08-02 05:07:53.436411	2012-08-02 05:42:13.323242
8388	1-8388	83	86	3579	88	9581	0	0	2012-08-02 05:39:00.166605	2012-08-02 05:42:14.118748
8837	1-8837	4	54	3175	58	7825	0	3	2012-08-02 05:42:01.146484	2012-08-02 05:42:14.505469
8854	1-8854	75	78	2905	23	4691	0	9	2012-08-02 05:42:08.009346	2012-08-02 05:42:14.505469
8321	1-8321	31	78	5656	83	3121	0	4	2012-08-02 05:38:33.076774	2012-08-02 05:42:14.505469
8870	1-8870	42	58	4065	78	9012	0	0	2012-08-02 05:42:14.505469	2012-08-02 05:42:14.505469
4370	1-4370	43	56	9332	100	7382	0	5	2012-08-02 05:11:49.903222	2012-08-02 05:42:16.549488
8089	1-8089	39	35	3328	56	7376	0	0	2012-08-02 05:36:58.854214	2012-08-02 05:42:16.549488
724	1-724	34	56	7419	100	4043	1091	0	2012-08-02 04:45:04.97111	2012-08-02 05:42:16.549488
8876	1-8876	30	29	157	66	5373	0	4	2012-08-02 05:42:16.969229	2012-08-02 05:42:16.969229
7880	1-7880	62	80	9576	97	9260	0	4	2012-08-02 05:35:32.881434	2012-08-02 05:42:18.617401
8649	1-8649	85	63	5798	27	7254	0	9	2012-08-02 05:40:45.045782	2012-08-02 05:42:18.617401
8262	1-8262	20	83	5340	97	5226	0	11	2012-08-02 05:38:09.299559	2012-08-02 05:42:18.617401
8599	1-8599	66	24	767	74	695	0	0	2012-08-02 05:40:24.979928	2012-08-02 05:42:20.230211
8759	1-8759	90	45	5916	8	3744	0	8	2012-08-02 05:41:29.646372	2012-08-02 05:42:20.230211
8784	1-8784	57	79	2663	88	5874	0	15	2012-08-02 05:41:39.811039	2012-08-02 05:42:20.230211
6781	1-6781	17	23	8750	74	9380	0	0	2012-08-02 05:28:02.700392	2012-08-02 05:42:20.230211
8751	1-8751	68	15	3694	87	9713	0	0	2012-08-02 05:41:26.37614	2012-08-02 05:42:21.478721
7428	1-7428	13	23	9223	87	9604	0	0	2012-08-02 05:32:24.537456	2012-08-02 05:42:21.478721
8511	1-8511	9	84	5708	82	6949	0	0	2012-08-02 05:39:49.554981	2012-08-02 05:42:21.478721
8424	1-8424	57	91	7643	34	7594	0	4	2012-08-02 05:39:14.745423	2012-08-02 05:42:21.876587
8890	1-8890	40	85	201	65	8213	0	6	2012-08-02 05:42:22.694059	2012-08-02 05:42:22.694059
8102	1-8102	32	3	4089	21	9892	0	0	2012-08-02 05:37:04.114703	2012-08-02 05:42:23.898301
8477	1-8477	85	45	2470	80	7627	0	10	2012-08-02 05:39:35.952586	2012-08-02 05:42:24.307075
8894	1-8894	64	37	3333	69	6645	0	2	2012-08-02 05:42:24.307075	2012-08-02 05:42:24.307075
8896	1-8896	61	4	4949	12	9685	0	8	2012-08-02 05:42:25.124672	2012-08-02 05:42:25.124672
7868	1-7868	93	72	7195	55	4517	0	0	2012-08-02 05:35:27.821842	2012-08-02 05:42:25.533526
2526	1-2526	16	72	8274	5	5744	0	0	2012-08-02 04:59:19.895302	2012-08-02 05:42:25.533526
8897	1-8897	35	5	2507	72	9358	0	3	2012-08-02 05:42:25.533526	2012-08-02 05:42:25.533526
8900	1-8900	92	9	1743	17	6669	0	6	2012-08-02 05:42:26.72665	2012-08-02 05:42:26.72665
3943	1-3943	28	71	9142	46	8052	0	0	2012-08-02 05:08:58.901135	2012-08-02 05:42:27.1465
8337	1-8337	66	73	2047	53	2611	0	8	2012-08-02 05:38:39.5623	2012-08-02 05:42:29.145976
8906	1-8906	89	93	1043	99	5538	0	2	2012-08-02 05:42:29.145976	2012-08-02 05:42:29.145976
8378	1-8378	23	36	9061	82	8710	0	0	2012-08-02 05:38:56.082803	2012-08-02 05:42:29.565999
8795	1-8795	28	88	9130	64	6677	0	0	2012-08-02 05:41:44.241105	2012-08-02 05:42:29.565999
8909	1-8909	45	84	271	98	2785	0	0	2012-08-02 05:42:30.383674	2012-08-02 05:42:30.383674
8911	1-8911	96	12	774	75	4031	0	0	2012-08-02 05:42:31.201373	2012-08-02 05:42:31.201373
8373	1-8373	8	48	4768	100	3583	0	0	2012-08-02 05:38:54.060905	2012-08-02 05:42:31.598878
7969	1-7969	98	61	4967	100	8045	0	7	2012-08-02 05:36:10.439492	2012-08-02 05:42:31.598878
8423	1-8423	52	92	7825	25	6438	0	0	2012-08-02 05:39:14.359105	2012-08-02 05:42:32.836549
8917	1-8917	29	92	46	83	2391	0	2	2012-08-02 05:42:33.642632	2012-08-02 05:42:33.642632
7849	1-7849	40	16	3006	14	4719	0	5	2012-08-02 05:35:19.800868	2012-08-02 05:42:34.040536
8723	1-8723	61	63	7653	24	8660	0	0	2012-08-02 05:41:14.978735	2012-08-02 05:42:37.29934
8884	1-8884	19	74	6306	24	9623	0	6	2012-08-02 05:42:20.230211	2012-08-02 05:42:37.29934
7310	1-7310	31	61	2324	25	1694	0	0	2012-08-02 05:31:37.147013	2012-08-02 05:42:37.29934
7208	1-7208	69	61	2481	99	1810	0	5	2012-08-02 05:30:56.74812	2012-08-02 05:42:37.29934
8926	1-8926	22	59	3586	33	6565	0	0	2012-08-02 05:42:37.29934	2012-08-02 05:42:37.29934
538	1-538	34	24	9921	59	682	682	3	2012-08-02 04:44:19.838325	2012-08-02 05:42:37.29934
8127	1-8127	70	34	7558	28	6264	0	0	2012-08-02 05:37:14.247487	2012-08-02 05:42:37.741642
8927	1-8927	59	3	2830	34	5295	0	12	2012-08-02 05:42:37.741642	2012-08-02 05:42:37.741642
5909	1-5909	89	19	7731	43	7942	2699	0	2012-08-02 05:22:06.604155	2012-08-02 05:42:37.741642
7948	1-7948	76	92	6350	98	2678	0	8	2012-08-02 05:36:01.796173	2012-08-02 05:42:39.763243
7920	1-7920	82	54	2123	58	795	0	0	2012-08-02 05:35:49.798043	2012-08-02 05:42:39.763243
8557	1-8557	75	55	3558	51	9059	0	8	2012-08-02 05:40:07.963611	2012-08-02 05:42:40.569807
8589	1-8589	28	61	4523	8	4322	0	2	2012-08-02 05:40:20.958728	2012-08-02 05:42:42.204683
8938	1-8938	14	60	584	61	3301	0	0	2012-08-02 05:42:42.204683	2012-08-02 05:42:42.204683
8867	1-8867	59	31	5262	62	9103	0	4	2012-08-02 05:42:13.323242	2012-08-02 05:42:43.044368
8810	1-8810	48	4	5095	18	4264	0	6	2012-08-02 05:41:50.262525	2012-08-02 05:42:45.088513
5910	1-5910	71	4	8771	28	3638	0	7	2012-08-02 05:22:07.02433	2012-08-02 05:42:45.088513
8183	1-8183	21	62	6102	38	9291	0	5	2012-08-02 05:37:37.1857	2012-08-02 05:42:45.485942
8762	1-8762	78	36	3386	38	5473	0	0	2012-08-02 05:41:30.851008	2012-08-02 05:42:45.485942
8946	1-8946	87	80	4451	62	9696	0	0	2012-08-02 05:42:45.485942	2012-08-02 05:42:45.485942
7562	1-7562	38	12	578	14	1565	0	0	2012-08-02 05:33:19.683567	2012-08-02 05:42:46.701345
8949	1-8949	32	58	3998	12	6218	0	0	2012-08-02 05:42:46.701345	2012-08-02 05:42:46.701345
8598	1-8598	76	15	6716	56	6021	0	0	2012-08-02 05:40:24.593337	2012-08-02 05:42:47.530224
7928	1-7928	31	23	1867	81	1258	0	0	2012-08-02 05:35:53.200937	2012-08-02 05:42:47.530224
8940	1-8940	43	12	3544	31	7091	0	9	2012-08-02 05:42:43.044368	2012-08-02 05:42:49.574119
7848	1-7848	100	12	8141	17	6287	0	6	2012-08-02 05:35:19.381255	2012-08-02 05:42:49.574119
4987	1-4987	98	12	7074	30	3937	0	0	2012-08-02 05:15:57.226488	2012-08-02 05:42:52.028554
6252	1-6252	13	12	10000	97	8609	0	0	2012-08-02 05:24:27.347085	2012-08-02 05:42:52.028554
8962	1-8962	25	15	1007	12	9756	0	5	2012-08-02 05:42:52.028554	2012-08-02 05:42:52.028554
1428	1-1428	51	55	3944	90	2026	0	4	2012-08-02 04:50:59.825108	2012-08-02 05:42:52.494951
8020	1-8020	18	48	6781	90	4752	0	0	2012-08-02 05:36:31.14329	2012-08-02 05:42:52.494951
8790	1-8790	11	81	1115	99	272	0	10	2012-08-02 05:41:42.208555	2012-08-02 05:42:52.494951
5405	1-5405	89	11	4044	73	1940	0	0	2012-08-02 05:18:43.51293	2012-08-02 05:42:52.494951
8963	1-8963	22	90	3354	11	8611	0	2	2012-08-02 05:42:52.494951	2012-08-02 05:42:52.494951
8513	1-8513	60	97	9801	22	9498	0	4	2012-08-02 05:39:50.339398	2012-08-02 05:42:53.332331
6488	1-6488	77	9	3336	59	4214	0	0	2012-08-02 05:26:04.242614	2012-08-02 05:42:53.332331
8244	1-8244	81	42	8310	94	6753	0	9	2012-08-02 05:38:01.94024	2012-08-02 05:42:53.332331
5655	1-5655	75	68	9498	25	5660	0	0	2012-08-02 05:20:23.413045	2012-08-02 05:42:54.149933
8967	1-8967	86	40	8458	68	8851	0	14	2012-08-02 05:42:54.149933	2012-08-02 05:42:54.149933
7657	1-7657	1	84	6017	67	4326	0	4	2012-08-02 05:33:58.573104	2012-08-02 05:42:55.342948
8891	1-8891	84	70	3034	18	1504	0	3	2012-08-02 05:42:23.102958	2012-08-02 05:42:56.171453
6674	1-6674	28	67	5138	42	2260	0	0	2012-08-02 05:27:20.202466	2012-08-02 05:42:56.171453
8972	1-8972	27	42	2622	67	7551	0	14	2012-08-02 05:42:56.171453	2012-08-02 05:42:56.171453
8515	1-8515	59	37	2674	56	2468	0	0	2012-08-02 05:39:51.134954	2012-08-02 05:42:56.60261
8930	1-8930	20	15	649	53	4096	0	0	2012-08-02 05:42:38.967843	2012-08-02 05:42:57.000279
8688	1-8688	81	50	2405	86	4797	0	3	2012-08-02 05:41:00.778937	2012-08-02 05:42:57.409131
8616	1-8616	78	64	7056	23	9159	0	0	2012-08-02 05:40:31.763548	2012-08-02 05:42:57.409131
434	1-434	86	11	4832	51	2693	0	0	2012-08-02 04:44:06.07922	2012-08-02 05:42:58.23806
8480	1-8480	4	86	4740	48	9986	0	5	2012-08-02 05:39:37.101122	2012-08-02 05:42:59.861566
8945	1-8945	7	49	2150	4	9874	0	0	2012-08-02 05:42:45.088513	2012-08-02 05:42:59.861566
8970	1-8970	91	42	2190	84	7138	0	4	2012-08-02 05:42:55.342948	2012-08-02 05:42:59.861566
8982	1-8982	58	57	4130	11	2110	0	4	2012-08-02 05:43:00.27057	2012-08-02 05:43:00.27057
8650	1-8650	82	56	4047	76	9017	0	0	2012-08-02 05:40:45.454809	2012-08-02 05:43:00.70149
8400	1-8400	28	75	1491	29	8244	0	0	2012-08-02 05:39:05.055646	2012-08-02 05:43:01.519339
8920	1-8920	53	28	6365	3	8187	0	11	2012-08-02 05:42:34.835859	2012-08-02 05:43:01.519339
7614	1-7614	38	71	3802	53	5439	0	0	2012-08-02 05:33:40.873921	2012-08-02 05:43:03.165069
8565	1-8565	67	100	9454	9	7074	0	0	2012-08-02 05:40:11.244956	2012-08-02 05:43:03.165069
8989	1-8989	13	9	145	58	1363	0	12	2012-08-02 05:43:03.165069	2012-08-02 05:43:03.165069
8879	1-8879	8	10	8438	45	9521	0	0	2012-08-02 05:42:18.21976	2012-08-02 05:43:03.574051
8738	1-8738	1	16	3582	85	7780	0	7	2012-08-02 05:41:21.055129	2012-08-02 05:43:03.97187
5794	1-5794	93	55	6736	46	7960	0	0	2012-08-02 05:21:20.107937	2012-08-02 05:43:03.97187
8259	1-8259	11	33	2108	29	6859	0	3	2012-08-02 05:38:08.095067	2012-08-02 05:43:03.97187
4767	1-4767	82	55	6515	61	2936	0	0	2012-08-02 05:14:29.770355	2012-08-02 05:43:03.97187
8991	1-8991	80	85	4701	55	7951	0	7	2012-08-02 05:43:03.97187	2012-08-02 05:43:03.97187
8993	1-8993	82	5	1978	4	5429	0	9	2012-08-02 05:43:04.789503	2012-08-02 05:43:04.789503
8996	1-8996	35	50	633	52	1670	0	0	2012-08-02 05:43:06.02793	2012-08-02 05:43:06.02793
8069	1-8069	98	76	5502	54	9053	0	3	2012-08-02 05:36:50.822423	2012-08-02 05:43:07.67428
8869	1-8869	9	88	4211	70	7655	0	11	2012-08-02 05:42:14.118748	2012-08-02 05:43:08.105018
8798	1-8798	93	8	5589	74	8657	0	0	2012-08-02 05:41:45.445676	2012-08-02 05:43:08.105018
9001	1-9001	21	30	1322	88	6051	0	5	2012-08-02 05:43:08.105018	2012-08-02 05:43:08.105018
6019	1-6019	34	88	9866	95	7096	1736	6	2012-08-02 05:22:52.415451	2012-08-02 05:43:08.105018
8607	1-8607	95	99	4523	95	9482	0	5	2012-08-02 05:40:28.184036	2012-08-02 05:43:08.524867
9002	1-9002	27	54	6732	99	5131	0	0	2012-08-02 05:43:08.524867	2012-08-02 05:43:08.524867
1805	1-1805	67	73	8437	90	1573	1573	0	2012-08-02 04:54:04.728167	2012-08-02 05:43:08.524867
2185	1-2185	51	73	3736	90	733	276	0	2012-08-02 04:56:50.908681	2012-08-02 05:43:08.524867
8302	1-8302	22	40	7953	7	2933	0	5	2012-08-02 05:38:25.464549	2012-08-02 05:43:08.944936
8458	1-8458	3	75	3350	83	7552	0	11	2012-08-02 05:39:28.350353	2012-08-02 05:43:08.944936
8725	1-8725	25	45	5640	83	7314	0	3	2012-08-02 05:41:15.818646	2012-08-02 05:43:08.944936
9006	1-9006	27	100	1493	73	8213	0	4	2012-08-02 05:43:10.14897	2012-08-02 05:43:10.14897
7750	1-7750	9	57	6158	90	3016	0	0	2012-08-02 05:34:37.745707	2012-08-02 05:43:10.955547
9008	1-9008	56	22	1299	57	7434	0	0	2012-08-02 05:43:10.955547	2012-08-02 05:43:10.955547
8698	1-8698	8	92	6528	97	4388	0	0	2012-08-02 05:41:04.932989	2012-08-02 05:43:11.784202
3754	1-3754	40	6	8763	36	5187	0	0	2012-08-02 05:07:43.558928	2012-08-02 05:43:11.784202
9010	1-9010	27	55	2606	91	9290	0	6	2012-08-02 05:43:11.784202	2012-08-02 05:43:11.784202
8758	1-8758	17	57	3204	69	3422	0	6	2012-08-02 05:41:29.249098	2012-08-02 05:43:12.602174
8265	1-8265	17	98	2285	95	6715	0	8	2012-08-02 05:38:10.492458	2012-08-02 05:43:13.795384
8585	1-8585	61	75	811	40	7869	0	0	2012-08-02 05:40:19.333939	2012-08-02 05:43:15.894242
8893	1-8893	94	36	5337	55	8423	0	0	2012-08-02 05:42:23.898301	2012-08-02 05:43:15.894242
8883	1-8883	6	93	2737	40	9007	0	11	2012-08-02 05:42:19.843517	2012-08-02 05:43:15.894242
7951	1-7951	63	32	1393	88	4133	0	2	2012-08-02 05:36:03.055922	2012-08-02 05:43:16.713451
8782	1-8782	74	28	5428	47	5365	0	0	2012-08-02 05:41:38.993303	2012-08-02 05:43:16.713451
8076	1-8076	66	32	7384	96	4328	0	3	2012-08-02 05:36:53.628432	2012-08-02 05:43:16.713451
9022	1-9022	9	23	598	32	7648	0	4	2012-08-02 05:43:16.713451	2012-08-02 05:43:16.713451
8990	1-8990	60	69	1342	10	8052	0	0	2012-08-02 05:43:03.574051	2012-08-02 05:43:17.553539
6775	1-6775	77	41	4535	8	2423	0	3	2012-08-02 05:28:00.347029	2012-08-02 05:43:17.984428
8915	1-8915	27	51	5572	30	4983	0	0	2012-08-02 05:42:32.836549	2012-08-02 05:43:18.392919
8934	1-8934	65	51	9611	55	8872	0	2	2012-08-02 05:42:40.569807	2012-08-02 05:43:21.663219
9035	1-9035	95	100	2589	1	7300	0	12	2012-08-02 05:43:22.071836	2012-08-02 05:43:22.071836
9039	1-9039	73	99	104	39	307	0	0	2012-08-02 05:43:23.740096	2012-08-02 05:43:23.740096
8888	1-8888	81	34	6129	19	6228	0	0	2012-08-02 05:42:21.876587	2012-08-02 05:43:24.557814
9042	1-9042	53	38	42	21	5569	0	6	2012-08-02 05:43:24.977435	2012-08-02 05:43:24.977435
8042	1-8042	51	99	2071	85	5163	0	0	2012-08-02 05:36:40.017206	2012-08-02 05:43:26.203666
9045	1-9045	48	88	1258	99	9839	0	4	2012-08-02 05:43:26.203666	2012-08-02 05:43:26.203666
3745	1-3745	69	23	8239	46	6843	0	0	2012-08-02 05:07:39.946403	2012-08-02 05:43:27.010249
9047	1-9047	51	46	1819	23	9916	0	0	2012-08-02 05:43:27.010249	2012-08-02 05:43:27.010249
6492	1-6492	30	77	6038	88	9340	0	10	2012-08-02 05:26:05.834807	2012-08-02 05:43:28.225547
8613	1-8613	6	44	7171	1	4825	0	0	2012-08-02 05:40:30.570252	2012-08-02 05:43:31.108796
7912	1-7912	15	39	7070	97	7079	0	0	2012-08-02 05:35:46.406206	2012-08-02 05:43:31.517441
9015	1-9015	34	95	3471	37	9950	0	0	2012-08-02 05:43:13.795384	2012-08-02 05:43:31.9375
9059	1-9059	82	37	4935	16	6768	0	6	2012-08-02 05:43:31.9375	2012-08-02 05:43:31.9375
9060	1-9060	89	71	81	60	1690	0	9	2012-08-02 05:43:32.335162	2012-08-02 05:43:32.335162
9062	1-9062	33	10	472	40	7042	0	7	2012-08-02 05:43:33.1195	2012-08-02 05:43:33.1195
7019	1-7019	2	73	8278	33	5175	0	6	2012-08-02 05:29:39.31673	2012-08-02 05:43:33.528448
8629	1-8629	62	20	2204	10	9169	0	5	2012-08-02 05:40:37.000549	2012-08-02 05:43:35.980907
7188	1-7188	82	20	5999	88	5644	0	5	2012-08-02 05:30:48.53903	2012-08-02 05:43:35.980907
9072	1-9072	56	12	143	44	2022	0	6	2012-08-02 05:43:37.185122	2012-08-02 05:43:37.185122
5966	1-5966	99	80	7434	63	3207	0	9	2012-08-02 05:22:30.326317	2012-08-02 05:43:37.991532
8997	1-8997	55	50	4062	6	4334	0	4	2012-08-02 05:43:06.436942	2012-08-02 05:43:37.991532
9074	1-9074	88	40	1457	80	4754	0	3	2012-08-02 05:43:37.991532	2012-08-02 05:43:37.991532
3035	1-3035	65	9	9293	51	9284	0	0	2012-08-02 05:02:53.436987	2012-08-02 05:43:39.218273
9077	1-9077	100	93	1175	9	2973	0	2	2012-08-02 05:43:39.218273	2012-08-02 05:43:39.218273
8660	1-8660	98	23	5838	87	3943	0	5	2012-08-02 05:40:49.487798	2012-08-02 05:43:39.637871
6798	1-6798	11	47	8153	57	6036	0	8	2012-08-02 05:28:09.485455	2012-08-02 05:43:40.046745
9079	1-9079	55	3	98	47	8347	0	11	2012-08-02 05:43:40.046745	2012-08-02 05:43:40.046745
7915	1-7915	79	84	8543	4	5459	0	5	2012-08-02 05:35:47.687884	2012-08-02 05:43:40.455554
7502	1-7502	71	84	4184	15	2474	0	2	2012-08-02 05:32:54.943392	2012-08-02 05:43:40.455554
9080	1-9080	8	24	1259	84	8762	0	16	2012-08-02 05:43:40.455554	2012-08-02 05:43:40.455554
8844	1-8844	81	60	3023	3	1045	0	12	2012-08-02 05:42:03.975621	2012-08-02 05:43:42.135115
7209	1-7209	84	66	9605	79	7376	0	12	2012-08-02 05:30:57.157137	2012-08-02 05:43:42.135115
6978	1-6978	82	66	2605	63	931	0	5	2012-08-02 05:29:23.071384	2012-08-02 05:43:42.135115
8965	1-8965	23	59	3871	39	8329	0	16	2012-08-02 05:42:53.332331	2012-08-02 05:43:42.135115
9084	1-9084	67	3	515	66	7198	0	4	2012-08-02 05:43:42.135115	2012-08-02 05:43:42.135115
9087	1-9087	59	42	425	71	2084	0	0	2012-08-02 05:43:43.383398	2012-08-02 05:43:43.383398
8880	1-8880	80	97	7167	45	6444	0	4	2012-08-02 05:42:18.617401	2012-08-02 05:43:43.792344
8760	1-8760	36	90	7925	77	7822	0	7	2012-08-02 05:41:30.03304	2012-08-02 05:43:44.201053
9089	1-9089	9	81	1460	90	7231	0	3	2012-08-02 05:43:44.201053	2012-08-02 05:43:44.201053
6755	1-6755	83	88	9120	8	1715	0	0	2012-08-02 05:27:52.459118	2012-08-02 05:43:44.610137
7547	1-7547	62	45	3859	70	7212	0	0	2012-08-02 05:33:13.583326	2012-08-02 05:43:44.610137
9090	1-9090	89	70	2907	96	7648	0	6	2012-08-02 05:43:44.610137	2012-08-02 05:43:44.610137
9078	1-9078	40	61	2742	23	9370	0	0	2012-08-02 05:43:39.637871	2012-08-02 05:43:46.24489
9094	1-9094	100	15	2422	61	7359	0	0	2012-08-02 05:43:46.24489	2012-08-02 05:43:46.24489
8556	1-8556	76	29	6597	28	1483	0	11	2012-08-02 05:40:07.565764	2012-08-02 05:43:50.391549
6209	1-6209	61	12	3366	59	1868	0	0	2012-08-02 05:24:09.39756	2012-08-02 05:43:50.82465
8470	1-8470	51	42	4316	26	1841	0	0	2012-08-02 05:39:33.145808	2012-08-02 05:43:51.234849
8689	1-8689	68	79	1846	85	5959	0	0	2012-08-02 05:41:01.18766	2012-08-02 05:43:51.666789
9111	1-9111	54	36	701	72	4822	0	6	2012-08-02 05:43:53.325755	2012-08-02 05:43:53.325755
9012	1-9012	16	69	3866	57	4220	0	6	2012-08-02 05:43:12.602174	2012-08-02 05:43:53.726128
9114	1-9114	79	24	224	2	8352	0	0	2012-08-02 05:43:54.545349	2012-08-02 05:43:54.545349
8875	1-8875	85	66	4793	35	5215	0	6	2012-08-02 05:42:16.549488	2012-08-02 05:43:54.966471
9115	1-9115	91	55	128	66	7930	0	4	2012-08-02 05:43:54.966471	2012-08-02 05:43:54.966471
6424	1-6424	79	61	9482	31	7862	0	0	2012-08-02 05:25:38.751401	2012-08-02 05:43:55.819374
8941	1-8941	1	87	5123	20	4621	0	5	2012-08-02 05:42:43.453276	2012-08-02 05:43:55.819374
8142	1-8142	42	87	9614	38	8276	0	20	2012-08-02 05:37:20.389912	2012-08-02 05:43:55.819374
9121	1-9121	34	68	1512	45	3976	0	0	2012-08-02 05:43:57.434257	2012-08-02 05:43:57.434257
8595	1-8595	42	72	8961	1	5750	0	10	2012-08-02 05:40:23.367235	2012-08-02 05:43:57.854521
9122	1-9122	92	57	1118	72	9872	0	3	2012-08-02 05:43:57.854521	2012-08-02 05:43:57.854521
7933	1-7933	73	33	3885	63	4454	0	7	2012-08-02 05:35:55.421741	2012-08-02 05:43:59.491054
9063	1-9063	23	33	4348	24	4625	0	12	2012-08-02 05:43:33.528448	2012-08-02 05:43:59.491054
9126	1-9126	83	29	3819	33	8961	0	5	2012-08-02 05:43:59.491054	2012-08-02 05:43:59.491054
8757	1-8757	35	99	3883	73	8503	0	0	2012-08-02 05:41:28.817895	2012-08-02 05:43:59.914139
8956	1-8956	29	81	4898	12	6786	0	7	2012-08-02 05:42:49.574119	2012-08-02 05:44:00.334898
8912	1-8912	79	100	4000	36	4792	0	0	2012-08-02 05:42:31.598878	2012-08-02 05:44:01.573448
9131	1-9131	35	59	684	100	3388	0	0	2012-08-02 05:44:01.573448	2012-08-02 05:44:01.573448
8887	1-8887	2	87	6031	84	9157	0	0	2012-08-02 05:42:21.478721	2012-08-02 05:44:01.993675
9135	1-9135	32	73	828	45	7288	0	0	2012-08-02 05:44:03.198289	2012-08-02 05:44:03.198289
9117	1-9117	46	31	2805	87	9725	0	7	2012-08-02 05:43:55.819374	2012-08-02 05:44:04.027398
9021	1-9021	19	86	3899	81	4256	0	0	2012-08-02 05:43:16.315996	2012-08-02 05:44:05.265217
9140	1-9140	45	88	3645	66	5533	0	0	2012-08-02 05:44:05.265217	2012-08-02 05:44:05.265217
8785	1-8785	25	21	4655	36	6401	0	3	2012-08-02 05:41:40.20876	2012-08-02 05:44:06.078784
8714	1-8714	8	77	4053	76	5149	0	4	2012-08-02 05:41:11.387886	2012-08-02 05:44:06.078784
6284	1-6284	20	77	169	4	81	0	0	2012-08-02 05:24:41.10428	2012-08-02 05:44:06.078784
2744	1-2744	80	77	7565	8	1770	0	0	2012-08-02 05:00:52.855192	2012-08-02 05:44:06.078784
4886	1-4886	56	20	7557	36	3363	0	0	2012-08-02 05:15:17.704731	2012-08-02 05:44:06.078784
9142	1-9142	11	36	1293	77	9870	0	3	2012-08-02 05:44:06.078784	2012-08-02 05:44:06.078784
6580	1-6580	30	20	3983	17	3926	0	0	2012-08-02 05:26:42.046166	2012-08-02 05:44:06.500589
9143	1-9143	22	58	969	20	7283	0	3	2012-08-02 05:44:06.500589	2012-08-02 05:44:06.500589
8345	1-8345	60	53	8528	100	5449	0	2	2012-08-02 05:38:42.78839	2012-08-02 05:44:06.899012
9019	1-9019	99	42	7899	56	8002	0	0	2012-08-02 05:43:15.474537	2012-08-02 05:44:08.114683
9020	1-9020	79	40	8080	68	7229	0	16	2012-08-02 05:43:15.894242	2012-08-02 05:44:08.114683
9147	1-9147	97	12	2966	18	4005	0	4	2012-08-02 05:44:08.114683	2012-08-02 05:44:08.114683
9150	1-9150	24	29	2492	97	5437	0	15	2012-08-02 05:44:09.319845	2012-08-02 05:44:09.319845
9151	1-9151	99	21	1543	100	9383	0	5	2012-08-02 05:44:09.717484	2012-08-02 05:44:09.717484
8403	1-8403	1	33	9262	61	8947	0	0	2012-08-02 05:39:06.26857	2012-08-02 05:44:11.717651
9156	1-9156	4	84	1437	18	9232	0	0	2012-08-02 05:44:11.717651	2012-08-02 05:44:11.717651
9161	1-9161	47	29	486	68	3618	0	0	2012-08-02 05:44:13.728247	2012-08-02 05:44:13.728247
7459	1-7459	46	19	6262	55	7112	0	0	2012-08-02 05:32:37.286996	2012-08-02 05:44:14.523637
5517	1-5517	76	19	4161	9	2075	0	5	2012-08-02 05:19:28.065188	2012-08-02 05:44:14.523637
9163	1-9163	72	96	398	19	5428	0	2	2012-08-02 05:44:14.523637	2012-08-02 05:44:14.523637
9144	1-9144	32	86	8676	53	8989	0	0	2012-08-02 05:44:06.899012	2012-08-02 05:44:16.545989
9168	1-9168	52	53	4836	82	9761	0	15	2012-08-02 05:44:16.545989	2012-08-02 05:44:16.545989
8580	1-8580	9	69	9635	12	5609	0	8	2012-08-02 05:40:17.290369	2012-08-02 05:44:17.352979
6097	1-6097	56	69	825	26	215	0	0	2012-08-02 05:23:24.128054	2012-08-02 05:44:17.352979
9170	1-9170	45	35	593	42	9709	0	0	2012-08-02 05:44:17.352979	2012-08-02 05:44:17.352979
9173	1-9173	46	30	154	53	4615	0	0	2012-08-02 05:44:18.581929	2012-08-02 05:44:18.581929
8531	1-8531	31	68	5706	29	7325	0	0	2012-08-02 05:39:57.532346	2012-08-02 05:44:20.581414
9178	1-9178	3	85	1473	68	6585	0	0	2012-08-02 05:44:20.581414	2012-08-02 05:44:20.581414
8035	1-8035	37	41	5385	68	4895	0	3	2012-08-02 05:36:37.233445	2012-08-02 05:44:20.990529
8681	1-8681	83	36	2663	29	7577	0	0	2012-08-02 05:40:57.917519	2012-08-02 05:44:21.797201
8992	1-8992	93	36	4285	94	7817	0	10	2012-08-02 05:43:04.380877	2012-08-02 05:44:21.797201
9181	1-9181	37	58	2462	36	4300	0	2	2012-08-02 05:44:21.797201	2012-08-02 05:44:21.797201
4241	1-4241	42	70	9661	93	4295	0	0	2012-08-02 05:10:57.307487	2012-08-02 05:44:22.581626
9054	1-9054	78	16	6934	76	5115	0	4	2012-08-02 05:43:29.882657	2012-08-02 05:44:22.581626
2778	1-2778	29	8	7444	70	1917	1917	0	2012-08-02 05:01:07.154483	2012-08-02 05:44:22.581626
9083	1-9083	6	67	3807	82	5858	0	6	2012-08-02 05:43:41.726416	2012-08-02 05:44:24.250627
7635	1-7635	16	27	1743	51	1658	0	8	2012-08-02 05:33:49.498586	2012-08-02 05:44:24.250627
5019	1-5019	76	44	7469	51	3875	0	5	2012-08-02 05:16:10.035583	2012-08-02 05:44:24.250627
4386	1-4386	51	27	9525	51	8122	0	0	2012-08-02 05:11:56.466008	2012-08-02 05:44:24.250627
8806	1-8806	6	69	2572	51	1190	0	0	2012-08-02 05:41:48.660422	2012-08-02 05:44:24.250627
7237	1-7237	36	24	4209	57	4776	0	5	2012-08-02 05:31:08.628985	2012-08-02 05:44:24.250627
9187	1-9187	31	51	5970	24	8889	0	3	2012-08-02 05:44:24.250627	2012-08-02 05:44:24.250627
9190	1-9190	23	84	214	6	4744	0	0	2012-08-02 05:44:25.488885	2012-08-02 05:44:25.488885
8349	1-8349	48	21	8583	93	4427	0	7	2012-08-02 05:38:44.356883	2012-08-02 05:44:25.897869
9191	1-9191	52	93	1945	91	7786	0	3	2012-08-02 05:44:25.897869	2012-08-02 05:44:25.897869
9194	1-9194	80	20	1356	88	2765	0	7	2012-08-02 05:44:27.124367	2012-08-02 05:44:27.124367
8699	1-8699	36	58	6331	71	8888	0	3	2012-08-02 05:41:05.352941	2012-08-02 05:44:27.51093
9195	1-9195	87	71	148	39	288	0	3	2012-08-02 05:44:27.51093	2012-08-02 05:44:27.51093
8787	1-8787	39	17	6990	100	3886	0	0	2012-08-02 05:41:41.004195	2012-08-02 05:44:29.886294
9201	1-9201	19	100	811	49	7897	0	0	2012-08-02 05:44:29.886294	2012-08-02 05:44:29.886294
8786	1-8786	23	51	2779	75	1600	0	0	2012-08-02 05:41:40.606367	2012-08-02 05:44:30.295626
3771	1-3771	15	72	2885	7	522	0	0	2012-08-02 05:07:50.276466	2012-08-02 05:44:30.295626
9202	1-9202	42	8	2145	38	7969	0	0	2012-08-02 05:44:30.295626	2012-08-02 05:44:30.295626
8118	1-8118	35	73	8713	90	8814	0	3	2012-08-02 05:37:10.56793	2012-08-02 05:44:31.124409
9204	1-9204	29	60	1219	73	2844	0	6	2012-08-02 05:44:31.124409	2012-08-02 05:44:31.124409
9213	1-9213	37	53	1377	11	2827	0	0	2012-08-02 05:44:34.737121	2012-08-02 05:44:34.737121
8635	1-8635	40	3	5544	67	5573	0	0	2012-08-02 05:40:39.409701	2012-08-02 05:44:35.134786
5606	1-5606	1	100	7020	67	2506	0	6	2012-08-02 05:20:03.522142	2012-08-02 05:44:35.134786
8958	1-8958	99	34	2167	72	2039	0	0	2012-08-02 05:42:50.415324	2012-08-02 05:44:35.134786
9214	1-9214	6	67	315	34	9744	0	13	2012-08-02 05:44:35.134786	2012-08-02 05:44:35.134786
9215	1-9215	35	20	188	78	3039	0	0	2012-08-02 05:44:35.555338	2012-08-02 05:44:35.555338
9216	1-9216	43	41	850	9	5113	0	8	2012-08-02 05:44:35.953025	2012-08-02 05:44:35.953025
9183	1-9183	2	75	4189	8	5822	0	12	2012-08-02 05:44:22.581626	2012-08-02 05:44:36.361923
4420	1-4420	68	61	5517	11	2863	0	8	2012-08-02 05:12:10.135832	2012-08-02 05:44:36.361923
9212	1-9212	17	63	1468	5	7398	0	6	2012-08-02 05:44:34.328669	2012-08-02 05:44:36.770653
7071	1-7071	3	5	7846	88	5095	0	4	2012-08-02 05:30:00.020891	2012-08-02 05:44:36.770653
8951	1-8951	70	30	5137	64	8357	0	0	2012-08-02 05:42:47.530224	2012-08-02 05:44:37.96589
9221	1-9221	6	35	653	30	8439	0	0	2012-08-02 05:44:37.96589	2012-08-02 05:44:37.96589
8975	1-8975	44	48	6501	30	5181	0	0	2012-08-02 05:42:57.409131	2012-08-02 05:44:38.385974
5711	1-5711	95	59	6379	63	2283	0	0	2012-08-02 05:20:45.987903	2012-08-02 05:44:38.385974
9222	1-9222	55	98	248	48	6494	0	0	2012-08-02 05:44:38.385974	2012-08-02 05:44:38.385974
9141	1-9141	90	10	1367	39	1902	0	7	2012-08-02 05:44:05.680863	2012-08-02 05:44:38.795089
8068	1-8068	20	27	5717	97	8407	0	0	2012-08-02 05:36:50.413369	2012-08-02 05:44:40.485556
9162	1-9162	72	57	1228	72	840	0	5	2012-08-02 05:44:14.137009	2012-08-02 05:44:43.319427
8459	1-8459	29	35	4399	46	7421	0	3	2012-08-02 05:39:28.759477	2012-08-02 05:44:43.717399
4260	1-4260	91	35	1329	46	1278	0	0	2012-08-02 05:11:04.720694	2012-08-02 05:44:43.717399
4784	1-4784	71	87	8292	46	3672	0	0	2012-08-02 05:14:36.653455	2012-08-02 05:44:43.717399
9243	1-9243	59	51	820	8	2673	0	8	2012-08-02 05:44:46.965478	2012-08-02 05:44:46.965478
6513	1-6513	5	40	8059	65	1709	0	3	2012-08-02 05:26:14.442284	2012-08-02 05:44:47.374285
7680	1-7680	66	84	5188	71	5071	0	0	2012-08-02 05:34:08.007736	2012-08-02 05:44:48.180965
7194	1-7194	51	54	8411	51	3237	0	4	2012-08-02 05:30:50.991696	2012-08-02 05:44:48.180965
9246	1-9246	93	51	534	22	4070	0	0	2012-08-02 05:44:48.180965	2012-08-02 05:44:48.180965
7721	1-7721	30	57	7263	67	3763	0	0	2012-08-02 05:34:25.381836	2012-08-02 05:44:48.999008
8586	1-8586	81	57	8179	96	6130	0	7	2012-08-02 05:40:19.743317	2012-08-02 05:44:48.999008
9128	1-9128	71	44	2950	60	7089	0	12	2012-08-02 05:44:00.334898	2012-08-02 05:44:49.419237
9249	1-9249	91	15	760	6	1705	0	0	2012-08-02 05:44:49.419237	2012-08-02 05:44:49.419237
1501	1-1501	81	64	4836	15	3791	457	0	2012-08-02 04:51:39.698411	2012-08-02 05:44:49.419237
9250	1-9250	26	42	349	5	9898	0	4	2012-08-02 05:44:49.816723	2012-08-02 05:44:49.816723
8932	1-8932	84	58	1428	92	7001	0	0	2012-08-02 05:42:39.763243	2012-08-02 05:44:50.714468
9235	1-9235	73	46	2698	63	4371	0	5	2012-08-02 05:44:43.717399	2012-08-02 05:44:51.109599
9067	1-9067	38	73	7366	26	3039	0	2	2012-08-02 05:43:35.174474	2012-08-02 05:44:51.109599
8753	1-8753	78	35	4608	41	5292	0	0	2012-08-02 05:41:27.193795	2012-08-02 05:44:51.52946
9256	1-9256	25	72	229	87	8288	0	5	2012-08-02 05:44:52.336263	2012-08-02 05:44:52.336263
9218	1-9218	58	88	5096	13	7344	0	0	2012-08-02 05:44:36.770653	2012-08-02 05:44:57.231855
9268	1-9268	3	82	947	88	5166	0	0	2012-08-02 05:44:57.231855	2012-08-02 05:44:57.231855
7796	1-7796	46	98	7914	73	7987	0	4	2012-08-02 05:34:57.215158	2012-08-02 05:44:58.02742
9255	1-9255	67	73	2210	5	7598	0	9	2012-08-02 05:44:51.9274	2012-08-02 05:44:58.02742
8190	1-8190	86	91	9647	28	6666	0	0	2012-08-02 05:37:40.058679	2012-08-02 05:44:58.43658
7068	1-7068	1	23	8055	100	6176	0	0	2012-08-02 05:29:58.760906	2012-08-02 05:44:58.43658
6547	1-6547	91	52	8800	100	8873	0	3	2012-08-02 05:26:28.53152	2012-08-02 05:44:58.43658
9271	1-9271	23	100	1571	91	7228	0	3	2012-08-02 05:44:58.43658	2012-08-02 05:44:58.43658
2932	1-2932	5	30	7734	100	7272	3119	0	2012-08-02 05:02:11.517427	2012-08-02 05:44:58.43658
9057	1-9057	17	54	3775	44	9825	0	0	2012-08-02 05:43:31.108796	2012-08-02 05:44:59.287494
8925	1-8925	75	55	5210	43	5093	0	10	2012-08-02 05:42:36.902068	2012-08-02 05:44:59.287494
9273	1-9273	80	44	1727	55	5138	0	10	2012-08-02 05:44:59.287494	2012-08-02 05:44:59.287494
6708	1-6708	65	98	8233	64	5719	0	4	2012-08-02 05:27:33.871558	2012-08-02 05:44:59.685131
7954	1-7954	71	48	9414	59	8617	0	0	2012-08-02 05:36:04.326425	2012-08-02 05:45:00.082907
9237	1-9237	32	3	1083	34	4591	0	14	2012-08-02 05:44:44.546178	2012-08-02 05:45:00.911853
9025	1-9025	9	8	766	70	6999	0	7	2012-08-02 05:43:17.984428	2012-08-02 05:45:03.783966
9217	1-9217	95	11	1428	6	5864	0	5	2012-08-02 05:44:36.361923	2012-08-02 05:45:03.783966
9284	1-9284	72	6	3093	8	2780	0	0	2012-08-02 05:45:03.783966	2012-08-02 05:45:03.783966
9003	1-9003	100	7	4203	75	8080	0	6	2012-08-02 05:43:08.944936	2012-08-02 05:45:04.192847
9290	1-9290	40	16	1743	62	370	0	5	2012-08-02 05:45:06.238627	2012-08-02 05:45:06.238627
8986	1-8986	89	77	7753	58	6693	0	0	2012-08-02 05:43:01.927893	2012-08-02 05:45:09.033763
7800	1-7800	61	17	9442	58	9236	0	0	2012-08-02 05:34:58.883271	2012-08-02 05:45:09.033763
9297	1-9297	8	58	2456	7	5461	0	0	2012-08-02 05:45:09.033763	2012-08-02 05:45:09.033763
7682	1-7682	35	83	8289	24	2339	0	0	2012-08-02 05:34:08.869348	2012-08-02 05:45:09.442621
8465	1-8465	39	13	6288	46	6987	0	0	2012-08-02 05:39:31.167567	2012-08-02 05:45:09.442621
8564	1-8564	87	6	7874	99	6178	0	8	2012-08-02 05:40:10.835973	2012-08-02 05:45:09.851576
7953	1-7953	91	42	6463	69	6773	0	0	2012-08-02 05:36:03.90642	2012-08-02 05:45:09.851576
8855	1-8855	43	79	4343	87	3957	0	0	2012-08-02 05:42:08.429254	2012-08-02 05:45:11.862225
8874	1-8874	4	97	9691	3	5126	0	9	2012-08-02 05:42:16.140436	2012-08-02 05:45:12.271379
9088	1-9088	75	45	6197	97	8779	0	16	2012-08-02 05:43:43.792344	2012-08-02 05:45:12.271379
7638	1-7638	70	45	8252	94	3174	0	13	2012-08-02 05:33:50.715549	2012-08-02 05:45:12.271379
9305	1-9305	28	38	4251	45	9904	0	7	2012-08-02 05:45:12.271379	2012-08-02 05:45:12.271379
5423	1-5423	2	52	8892	7	7172	0	5	2012-08-02 05:18:50.547206	2012-08-02 05:45:13.8628
7662	1-7662	90	7	6726	95	8935	0	10	2012-08-02 05:34:00.617664	2012-08-02 05:45:13.8628
8799	1-8799	34	52	5134	35	6965	0	0	2012-08-02 05:41:45.843336	2012-08-02 05:45:13.8628
9309	1-9309	41	95	2304	52	7430	0	5	2012-08-02 05:45:13.8628	2012-08-02 05:45:13.8628
6659	1-6659	43	52	5116	89	3979	0	0	2012-08-02 05:27:14.180943	2012-08-02 05:45:17.110829
9317	1-9317	65	55	447	52	7058	0	10	2012-08-02 05:45:17.110829	2012-08-02 05:45:17.110829
7694	1-7694	98	58	8965	81	8626	0	2	2012-08-02 05:34:13.887496	2012-08-02 05:45:17.54183
9318	1-9318	15	81	4009	4	9517	0	5	2012-08-02 05:45:17.54183	2012-08-02 05:45:17.54183
9321	1-9321	41	67	625	28	4849	0	0	2012-08-02 05:45:18.770261	2012-08-02 05:45:18.770261
9322	1-9322	10	61	594	95	4811	0	0	2012-08-02 05:45:19.189829	2012-08-02 05:45:19.189829
8918	1-8918	81	13	2231	16	5582	0	3	2012-08-02 05:42:34.040536	2012-08-02 05:45:22.018036
9110	1-9110	24	62	3401	9	6470	0	0	2012-08-02 05:43:52.894522	2012-08-02 05:45:22.42701
9330	1-9330	65	83	529	62	2594	0	0	2012-08-02 05:45:22.42701	2012-08-02 05:45:22.42701
8176	1-8176	42	93	6234	16	4151	0	9	2012-08-02 05:37:34.302143	2012-08-02 05:45:22.835957
6922	1-6922	5	71	1874	32	823	0	4	2012-08-02 05:29:00.55942	2012-08-02 05:45:22.835957
9331	1-9331	100	32	702	93	5976	0	7	2012-08-02 05:45:22.835957	2012-08-02 05:45:22.835957
9333	1-9333	13	98	710	25	2348	0	5	2012-08-02 05:45:23.642808	2012-08-02 05:45:23.642808
9338	1-9338	15	18	868	47	6680	0	2	2012-08-02 05:45:25.745438	2012-08-02 05:45:25.745438
5615	1-5615	27	70	2825	19	219	0	0	2012-08-02 05:20:07.181071	2012-08-02 05:45:26.934841
9342	1-9342	14	58	92	81	3738	0	8	2012-08-02 05:45:27.332705	2012-08-02 05:45:27.332705
5465	1-5465	23	68	7567	14	9486	0	6	2012-08-02 05:19:07.180451	2012-08-02 05:45:28.504053
8559	1-8559	12	68	9684	12	8245	0	2	2012-08-02 05:40:08.78121	2012-08-02 05:45:28.504053
9071	1-9071	87	15	9050	52	6271	0	2	2012-08-02 05:43:36.798498	2012-08-02 05:45:28.504053
6667	1-6667	16	20	8286	61	5539	0	16	2012-08-02 05:27:17.385066	2012-08-02 05:45:29.41004
8492	1-8492	34	46	2591	5	6677	0	0	2012-08-02 05:39:41.940856	2012-08-02 05:45:30.69134
8238	1-8238	19	59	6105	5	8975	0	12	2012-08-02 05:37:59.494543	2012-08-02 05:45:30.69134
8526	1-8526	46	75	1360	14	1759	0	9	2012-08-02 05:39:55.532393	2012-08-02 05:45:30.69134
8436	1-8436	51	9	3268	59	2788	0	3	2012-08-02 05:39:19.552533	2012-08-02 05:45:30.69134
9350	1-9350	92	14	3213	9	8166	0	0	2012-08-02 05:45:30.69134	2012-08-02 05:45:30.69134
8107	1-8107	8	11	6804	81	7095	0	3	2012-08-02 05:37:06.148944	2012-08-02 05:45:31.498135
9211	1-9211	9	11	4279	58	7282	0	7	2012-08-02 05:44:33.908486	2012-08-02 05:45:31.498135
9352	1-9352	30	92	3011	11	4707	0	2	2012-08-02 05:45:31.498135	2012-08-02 05:45:31.498135
9341	1-9341	60	87	1013	70	7023	0	0	2012-08-02 05:45:26.934841	2012-08-02 05:45:31.896255
8339	1-8339	81	75	2630	86	4444	0	0	2012-08-02 05:38:40.358037	2012-08-02 05:45:31.896255
9288	1-9288	31	87	5284	75	3558	0	5	2012-08-02 05:45:05.441417	2012-08-02 05:45:31.896255
9353	1-9353	27	29	698	87	9118	0	0	2012-08-02 05:45:31.896255	2012-08-02 05:45:31.896255
8733	1-8733	17	25	8193	53	8879	0	0	2012-08-02 05:41:19.022222	2012-08-02 05:45:32.305078
8821	1-8821	15	82	8850	23	8672	0	19	2012-08-02 05:41:54.738828	2012-08-02 05:45:32.305078
7117	1-7117	47	98	4139	53	5065	0	0	2012-08-02 05:30:18.795785	2012-08-02 05:45:33.520489
6336	1-6336	97	98	8969	46	5460	0	10	2012-08-02 05:25:02.817126	2012-08-02 05:45:33.520489
9106	1-9106	47	26	4373	43	8458	0	0	2012-08-02 05:43:51.234849	2012-08-02 05:45:33.520489
9285	1-9285	27	75	1064	64	2915	0	0	2012-08-02 05:45:04.192847	2012-08-02 05:45:35.553347
9274	1-9274	3	83	2899	98	6976	0	3	2012-08-02 05:44:59.685131	2012-08-02 05:45:38.768535
8964	1-8964	2	94	2964	65	3879	0	4	2012-08-02 05:42:52.91248	2012-08-02 05:45:39.177103
9345	1-9345	73	90	3527	68	8227	0	4	2012-08-02 05:45:28.504053	2012-08-02 05:45:40.403421
9379	1-9379	39	94	3437	50	3990	0	4	2012-08-02 05:45:42.425497	2012-08-02 05:45:42.425497
8108	1-8108	74	8	1400	56	1663	0	2	2012-08-02 05:37:06.557746	2012-08-02 05:45:42.822866
9277	1-9277	63	8	4214	3	9198	0	6	2012-08-02 05:45:00.911853	2012-08-02 05:45:42.822866
9380	1-9380	66	100	319	8	8602	0	3	2012-08-02 05:45:42.822866	2012-08-02 05:45:42.822866
7024	1-7024	81	76	7347	95	5011	0	5	2012-08-02 05:29:41.304506	2012-08-02 05:45:44.447245
5557	1-5557	77	77	7122	63	3186	0	0	2012-08-02 05:19:43.941933	2012-08-02 05:45:44.447245
8405	1-8405	49	9	3486	63	2597	0	0	2012-08-02 05:39:07.075481	2012-08-02 05:45:44.447245
9384	1-9384	26	63	848	76	7207	0	0	2012-08-02 05:45:44.447245	2012-08-02 05:45:44.447245
4977	1-4977	37	76	7358	95	3279	1305	0	2012-08-02 05:15:53.26027	2012-08-02 05:45:44.447245
9053	1-9053	82	63	3641	19	4310	0	0	2012-08-02 05:43:29.474091	2012-08-02 05:45:45.667157
7155	1-7155	87	8	7763	11	5945	0	8	2012-08-02 05:30:34.685081	2012-08-02 05:45:46.868462
9390	1-9390	55	13	1975	8	8046	0	0	2012-08-02 05:45:46.868462	2012-08-02 05:45:46.868462
9107	1-9107	96	26	712	79	6714	0	8	2012-08-02 05:43:51.666789	2012-08-02 05:45:49.265963
9125	1-9125	60	66	9438	86	9868	0	7	2012-08-02 05:43:59.07162	2012-08-02 05:45:49.265963
7203	1-7203	65	72	3001	14	6823	0	7	2012-08-02 05:30:54.671015	2012-08-02 05:45:50.47021
8665	1-8665	8	22	5858	93	7393	0	2	2012-08-02 05:40:51.509715	2012-08-02 05:45:51.674757
9402	1-9402	37	94	2684	22	6387	0	2	2012-08-02 05:45:51.674757	2012-08-02 05:45:51.674757
9233	1-9233	85	10	6900	34	9447	0	2	2012-08-02 05:44:42.921818	2012-08-02 05:45:52.083327
9227	1-9227	73	97	5080	62	7833	0	11	2012-08-02 05:44:40.485556	2012-08-02 05:45:53.309948
8907	1-8907	76	23	6000	21	8857	0	2	2012-08-02 05:42:29.565999	2012-08-02 05:45:54.116261
9228	1-9228	32	23	1598	33	1002	0	6	2012-08-02 05:44:40.905493	2012-08-02 05:45:54.116261
8462	1-8462	23	94	1009	16	572	0	0	2012-08-02 05:39:29.952557	2012-08-02 05:45:54.116261
9409	1-9409	49	6	205	22	565	0	0	2012-08-02 05:45:54.536268	2012-08-02 05:45:54.536268
9411	1-9411	11	85	459	21	9892	0	9	2012-08-02 05:45:55.310003	2012-08-02 05:45:55.310003
3260	1-3260	92	66	6730	88	5712	0	2	2012-08-02 05:04:24.458733	2012-08-02 05:45:56.127461
9396	1-9396	11	4	1882	66	5182	0	0	2012-08-02 05:45:49.265963	2012-08-02 05:45:56.127461
9073	1-9073	79	4	8212	86	7707	0	8	2012-08-02 05:43:37.605277	2012-08-02 05:45:56.127461
8899	1-8899	17	17	2751	88	6327	0	4	2012-08-02 05:42:26.340081	2012-08-02 05:45:56.127461
8825	1-8825	78	100	7893	11	5741	0	13	2012-08-02 05:41:56.35202	2012-08-02 05:45:56.525308
9036	1-9036	51	100	3660	66	5483	0	4	2012-08-02 05:43:22.502788	2012-08-02 05:45:56.525308
9387	1-9387	37	100	1824	63	2099	0	5	2012-08-02 05:45:45.667157	2012-08-02 05:45:56.525308
9414	1-9414	39	25	620	100	7780	0	0	2012-08-02 05:45:56.525308	2012-08-02 05:45:56.525308
8525	1-8525	82	80	3226	37	9219	0	0	2012-08-02 05:39:55.1457	2012-08-02 05:45:56.945369
9415	1-9415	100	37	578	51	1823	0	2	2012-08-02 05:45:56.945369	2012-08-02 05:45:56.945369
9417	1-9417	52	42	1250	7	1455	0	2	2012-08-02 05:45:57.752016	2012-08-02 05:45:57.752016
9418	1-9418	15	53	119	38	3788	0	0	2012-08-02 05:45:58.149764	2012-08-02 05:45:58.149764
9419	1-9419	50	96	84	6	7302	0	6	2012-08-02 05:45:58.547606	2012-08-02 05:45:58.547606
969	1-969	75	66	5877	51	2355	0	0	2012-08-02 04:46:54.648026	2012-08-02 05:45:59.398575
8935	1-8935	66	14	6693	15	8089	0	0	2012-08-02 05:42:41.000707	2012-08-02 05:46:01.784474
9254	1-9254	85	34	4866	35	6851	0	0	2012-08-02 05:44:51.52946	2012-08-02 05:46:02.182277
9428	1-9428	5	35	633	67	3385	0	4	2012-08-02 05:46:02.182277	2012-08-02 05:46:02.182277
6273	1-6273	77	20	3328	82	1402	0	0	2012-08-02 05:24:36.428501	2012-08-02 05:46:04.601967
9438	1-9438	5	76	507	94	8530	0	7	2012-08-02 05:46:06.215873	2012-08-02 05:46:06.215873
9226	1-9226	6	57	1999	16	4469	0	0	2012-08-02 05:44:40.087855	2012-08-02 05:46:07.089519
7759	1-7759	95	76	5808	96	2004	0	0	2012-08-02 05:34:41.580802	2012-08-02 05:46:07.089519
9440	1-9440	40	16	5924	24	8023	0	0	2012-08-02 05:46:07.089519	2012-08-02 05:46:07.089519
9275	1-9275	17	82	1892	48	7273	0	0	2012-08-02 05:45:00.082907	2012-08-02 05:46:09.629741
7555	1-7555	73	82	3979	15	3894	0	0	2012-08-02 05:33:16.853441	2012-08-02 05:46:09.629741
6829	1-6829	16	36	7428	31	6255	0	0	2012-08-02 05:28:22.171516	2012-08-02 05:46:10.438541
9448	1-9448	29	31	4469	64	5690	0	11	2012-08-02 05:46:10.438541	2012-08-02 05:46:10.438541
9069	1-9069	66	15	6090	20	8082	0	0	2012-08-02 05:43:35.980907	2012-08-02 05:46:11.277968
9452	1-9452	19	80	489	74	396	0	9	2012-08-02 05:46:12.084552	2012-08-02 05:46:12.084552
8512	1-8512	66	59	2099	91	1297	0	10	2012-08-02 05:39:49.952643	2012-08-02 05:46:12.493422
9453	1-9453	9	91	783	59	6253	0	2	2012-08-02 05:46:12.493422	2012-08-02 05:46:12.493422
7556	1-7556	28	71	6592	95	7341	0	8	2012-08-02 05:33:17.250918	2012-08-02 05:46:13.71963
9132	1-9132	24	28	2684	87	2853	0	0	2012-08-02 05:44:01.993675	2012-08-02 05:46:14.90182
8468	1-8468	7	31	6514	2	7953	0	3	2012-08-02 05:39:32.327721	2012-08-02 05:46:15.31099
9460	1-9460	10	2	4567	96	4118	0	2	2012-08-02 05:46:15.31099	2012-08-02 05:46:15.31099
8827	1-8827	94	61	1259	84	2049	0	7	2012-08-02 05:41:57.125297	2012-08-02 05:46:15.708524
9112	1-9112	69	68	4431	69	9136	0	8	2012-08-02 05:43:53.726128	2012-08-02 05:46:16.526328
9408	1-9408	91	16	4134	23	8772	0	5	2012-08-02 05:45:54.116261	2012-08-02 05:46:16.526328
9104	1-9104	56	28	6531	21	9711	0	0	2012-08-02 05:43:50.391549	2012-08-02 05:46:18.558769
8355	1-8355	41	86	5324	6	5440	0	0	2012-08-02 05:38:46.765589	2012-08-02 05:46:18.558769
8270	1-8270	52	87	5640	13	1395	0	0	2012-08-02 05:38:12.49244	2012-08-02 05:46:18.558769
9468	1-9468	55	21	6843	97	7086	0	0	2012-08-02 05:46:18.558769	2012-08-02 05:46:18.558769
5608	1-5608	85	25	5473	78	1831	0	0	2012-08-02 05:20:04.350922	2012-08-02 05:46:19.365429
9038	1-9038	45	28	5218	30	6268	0	9	2012-08-02 05:43:23.320406	2012-08-02 05:46:19.365429
9470	1-9470	100	30	147	44	5118	0	2	2012-08-02 05:46:19.365429	2012-08-02 05:46:19.365429
9367	1-9367	86	77	830	40	5245	0	8	2012-08-02 05:45:37.575524	2012-08-02 05:46:21.420603
9475	1-9475	15	40	3304	91	7112	0	0	2012-08-02 05:46:21.420603	2012-08-02 05:46:21.420603
5493	1-5493	88	67	8093	34	5484	0	9	2012-08-02 05:19:18.287543	2012-08-02 05:46:23.000474
8230	1-8230	90	67	2152	34	519	0	13	2012-08-02 05:37:56.26672	2012-08-02 05:46:23.000474
9479	1-9479	11	84	3231	67	9933	0	7	2012-08-02 05:46:23.000474	2012-08-02 05:46:23.000474
2829	1-2829	96	35	2091	46	614	0	0	2012-08-02 05:01:28.501387	2012-08-02 05:46:23.807004
9485	1-9485	89	54	4067	13	6393	0	7	2012-08-02 05:46:25.431323	2012-08-02 05:46:25.431323
8303	1-8303	82	47	6244	10	7827	0	0	2012-08-02 05:38:25.862229	2012-08-02 05:46:25.839997
8490	1-8490	90	54	2828	10	3208	0	0	2012-08-02 05:39:41.111852	2012-08-02 05:46:25.839997
9489	1-9489	39	68	913	89	3629	0	0	2012-08-02 05:46:27.033076	2012-08-02 05:46:27.033076
6855	1-6855	42	12	7350	44	8143	0	2	2012-08-02 05:28:32.88874	2012-08-02 05:46:27.441921
9052	1-9052	30	52	5709	90	8068	0	5	2012-08-02 05:43:29.043144	2012-08-02 05:46:27.839817
9491	1-9491	18	84	242	52	6241	0	9	2012-08-02 05:46:27.839817	2012-08-02 05:46:27.839817
9423	1-9423	1	45	1952	10	3875	0	0	2012-08-02 05:46:00.182726	2012-08-02 05:46:28.248583
8095	1-8095	39	45	2332	70	1803	0	8	2012-08-02 05:37:01.30684	2012-08-02 05:46:28.248583
9492	1-9492	92	70	3428	45	9114	0	4	2012-08-02 05:46:28.248583	2012-08-02 05:46:28.248583
9467	1-9467	25	13	1820	64	2182	0	8	2012-08-02 05:46:18.172481	2012-08-02 05:46:29.05505
9165	1-9165	17	27	1772	75	1767	0	0	2012-08-02 05:44:15.341962	2012-08-02 05:46:29.05505
9070	1-9070	63	50	1168	6	765	0	5	2012-08-02 05:43:36.378686	2012-08-02 05:46:29.05505
9357	1-9357	29	43	2078	52	9269	0	3	2012-08-02 05:45:33.520489	2012-08-02 05:46:29.475141
9158	1-9158	61	32	2409	68	4965	0	6	2012-08-02 05:44:12.535071	2012-08-02 05:46:29.475141
8678	1-8678	53	61	7917	78	4618	0	7	2012-08-02 05:40:56.724342	2012-08-02 05:46:29.475141
7452	1-7452	100	27	5996	13	5375	0	0	2012-08-02 05:32:34.425165	2012-08-02 05:46:30.679765
9091	1-9091	94	18	2228	49	1742	0	4	2012-08-02 05:43:45.029908	2012-08-02 05:46:30.679765
9498	1-9498	79	43	2483	18	9347	0	0	2012-08-02 05:46:30.679765	2012-08-02 05:46:30.679765
5830	1-5830	20	82	6492	80	8105	0	2	2012-08-02 05:21:34.605921	2012-08-02 05:46:31.508141
4144	1-4144	38	38	6821	80	6243	0	0	2012-08-02 05:10:19.207131	2012-08-02 05:46:31.508141
9500	1-9500	79	80	2181	27	3563	0	4	2012-08-02 05:46:31.508141	2012-08-02 05:46:31.508141
9450	1-9450	82	46	1188	15	3728	0	0	2012-08-02 05:46:11.277968	2012-08-02 05:46:32.31477
9490	1-9490	67	44	4038	90	6965	0	3	2012-08-02 05:46:27.441921	2012-08-02 05:46:32.701571
9506	1-9506	35	72	70	67	3776	0	0	2012-08-02 05:46:33.961183	2012-08-02 05:46:33.961183
8612	1-8612	43	27	3517	50	3669	0	0	2012-08-02 05:40:30.172448	2012-08-02 05:46:35.584983
6143	1-6143	67	21	7783	42	5127	0	0	2012-08-02 05:23:42.547882	2012-08-02 05:46:35.584983
9513	1-9513	49	11	112	27	621	0	0	2012-08-02 05:46:36.767224	2012-08-02 05:46:36.767224
7414	1-7414	20	41	4488	56	4328	0	0	2012-08-02 05:32:18.813361	2012-08-02 05:46:37.176028
8358	1-8358	46	53	6348	56	2261	0	4	2012-08-02 05:38:47.970038	2012-08-02 05:46:37.176028
9514	1-9514	64	12	3809	38	7786	0	5	2012-08-02 05:46:37.176028	2012-08-02 05:46:37.176028
9515	1-9515	51	53	677	78	1350	0	2	2012-08-02 05:46:37.609017	2012-08-02 05:46:37.609017
9516	1-9516	19	16	796	73	8001	0	0	2012-08-02 05:46:38.006682	2012-08-02 05:46:38.006682
7652	1-7652	28	79	5887	93	4652	0	7	2012-08-02 05:33:56.570811	2012-08-02 05:46:39.200116
8579	1-8579	65	38	5187	72	4201	0	7	2012-08-02 05:40:16.881321	2012-08-02 05:46:39.608678
9298	1-9298	5	43	2472	39	6984	0	10	2012-08-02 05:45:09.442621	2012-08-02 05:46:40.017559
9521	1-9521	43	84	69	43	2784	0	5	2012-08-02 05:46:40.017559	2012-08-02 05:46:40.017559
9234	1-9234	8	72	6321	73	9069	0	6	2012-08-02 05:44:43.319427	2012-08-02 05:46:40.835476
9524	1-9524	53	50	2902	58	6340	0	0	2012-08-02 05:46:41.254949	2012-08-02 05:46:41.254949
7325	1-7325	21	66	8217	26	2050	0	14	2012-08-02 05:31:43.068652	2012-08-02 05:46:43.2656
3858	1-3858	73	74	8471	46	2586	2471	0	2012-08-02 05:08:24.883038	2012-08-02 05:46:43.2656
9532	1-9532	27	1	94	85	3053	0	4	2012-08-02 05:46:44.480894	2012-08-02 05:46:44.480894
5226	1-5226	41	85	5504	73	1293	113	0	2012-08-02 05:17:33.216569	2012-08-02 05:46:44.480894
3094	1-3094	3	85	7434	73	478	478	0	2012-08-02 05:03:17.426994	2012-08-02 05:46:44.480894
9403	1-9403	94	34	5908	33	9985	0	4	2012-08-02 05:45:52.083327	2012-08-02 05:46:44.900839
7044	1-7044	53	94	7728	74	4549	0	0	2012-08-02 05:29:49.281114	2012-08-02 05:46:46.193565
9538	1-9538	91	86	910	19	4323	0	0	2012-08-02 05:46:47.010931	2012-08-02 05:46:47.010931
5439	1-5439	12	17	9012	15	5417	0	0	2012-08-02 05:18:56.917112	2012-08-02 05:46:48.22651
8791	1-8791	79	55	4548	91	5065	0	4	2012-08-02 05:41:42.606001	2012-08-02 05:46:48.22651
9028	1-9028	25	43	2591	31	8282	0	0	2012-08-02 05:43:19.210517	2012-08-02 05:46:49.022068
8950	1-8950	34	43	4958	57	8916	0	4	2012-08-02 05:42:47.099078	2012-08-02 05:46:49.022068
8707	1-8707	74	7	4862	31	7303	0	9	2012-08-02 05:41:08.603947	2012-08-02 05:46:50.635152
9547	1-9547	32	31	3993	49	4832	0	4	2012-08-02 05:46:50.635152	2012-08-02 05:46:50.635152
9081	1-9081	38	82	9226	4	8834	0	0	2012-08-02 05:43:40.88658	2012-08-02 05:46:51.430417
9291	1-9291	67	82	4683	25	8980	0	0	2012-08-02 05:45:06.647507	2012-08-02 05:46:51.430417
409	1-409	67	82	8130	31	6863	0	0	2012-08-02 04:44:04.124444	2012-08-02 05:46:51.430417
9549	1-9549	90	15	872	82	9492	0	11	2012-08-02 05:46:51.430417	2012-08-02 05:46:51.430417
9410	1-9410	18	15	6935	63	4598	0	12	2012-08-02 05:45:54.922919	2012-08-02 05:46:53.032769
9456	1-9456	82	95	2500	71	3200	0	0	2012-08-02 05:46:13.71963	2012-08-02 05:46:53.032769
5851	1-5851	35	81	4187	51	3550	0	5	2012-08-02 05:21:43.178896	2012-08-02 05:46:53.032769
7356	1-7356	88	95	7411	63	3106	0	9	2012-08-02 05:31:55.278802	2012-08-02 05:46:53.032769
9553	1-9553	20	51	2004	83	9467	0	0	2012-08-02 05:46:53.032769	2012-08-02 05:46:53.032769
8610	1-8610	83	25	8133	51	4279	0	3	2012-08-02 05:40:29.388568	2012-08-02 05:46:53.452684
9554	1-9554	5	50	784	25	9543	0	11	2012-08-02 05:46:53.452684	2012-08-02 05:46:53.452684
8742	1-8742	72	52	6962	70	6960	0	0	2012-08-02 05:41:22.67444	2012-08-02 05:46:53.850429
7941	1-7941	53	52	6792	10	7333	0	0	2012-08-02 05:35:58.769314	2012-08-02 05:46:53.850429
7602	1-7602	98	56	7909	51	7108	0	7	2012-08-02 05:33:35.968539	2012-08-02 05:46:54.248119
9556	1-9556	95	50	806	57	5481	0	0	2012-08-02 05:46:54.248119	2012-08-02 05:46:54.248119
5808	1-5808	40	18	3956	72	1231	0	0	2012-08-02 05:21:25.68711	2012-08-02 05:46:56.270341
9561	1-9561	79	72	45	18	5355	0	0	2012-08-02 05:46:56.270341	2012-08-02 05:46:56.270341
9434	1-9434	81	48	3960	20	4863	0	4	2012-08-02 05:46:04.601967	2012-08-02 05:46:57.076851
5721	1-5721	90	49	5231	95	4258	0	0	2012-08-02 05:20:50.021096	2012-08-02 05:46:57.076851
6211	1-6211	75	49	9751	73	5356	0	9	2012-08-02 05:24:10.226864	2012-08-02 05:46:57.076851
9486	1-9486	89	10	6994	34	5378	0	4	2012-08-02 05:46:25.839997	2012-08-02 05:46:59.087226
8546	1-8546	20	10	4688	78	1596	0	4	2012-08-02 05:40:03.531167	2012-08-02 05:46:59.087226
9034	1-9034	21	57	9204	56	9103	0	11	2012-08-02 05:43:21.663219	2012-08-02 05:47:00.314026
9571	1-9571	8	56	6079	47	6527	0	0	2012-08-02 05:47:00.314026	2012-08-02 05:47:00.314026
9573	1-9573	62	89	974	100	5794	0	3	2012-08-02 05:47:01.131639	2012-08-02 05:47:01.131639
9270	1-9270	92	5	9159	98	4843	0	0	2012-08-02 05:44:58.02742	2012-08-02 05:47:01.926869
9048	1-9048	67	5	8103	30	1450	0	7	2012-08-02 05:43:27.418939	2012-08-02 05:47:01.926869
5635	1-5635	44	5	9270	92	3899	0	0	2012-08-02 05:20:15.380699	2012-08-02 05:47:01.926869
4564	1-4564	27	25	9217	59	8059	0	0	2012-08-02 05:13:07.185658	2012-08-02 05:47:02.324687
7934	1-7934	37	20	6369	100	4331	0	11	2012-08-02 05:35:55.841372	2012-08-02 05:47:04.821364
9563	1-9563	99	20	3719	49	6935	0	2	2012-08-02 05:46:57.076851	2012-08-02 05:47:04.821364
7768	1-7768	1	16	8004	32	4002	0	0	2012-08-02 05:34:45.392401	2012-08-02 05:47:04.821364
9346	1-9346	92	34	1701	92	9407	0	7	2012-08-02 05:45:28.923643	2012-08-02 05:47:05.652785
9584	1-9584	79	92	1742	49	6602	0	5	2012-08-02 05:47:05.652785	2012-08-02 05:47:05.652785
9531	1-9531	5	65	2656	79	4194	0	15	2012-08-02 05:46:44.072235	2012-08-02 05:47:06.846517
5622	1-5622	99	65	7396	83	4877	0	0	2012-08-02 05:20:10.044086	2012-08-02 05:47:06.846517
9421	1-9421	51	46	1432	66	6449	0	0	2012-08-02 05:45:59.398575	2012-08-02 05:47:09.729894
8726	1-8726	100	67	4082	53	9161	0	0	2012-08-02 05:41:16.205241	2012-08-02 05:47:09.729894
8983	1-8983	5	76	8040	75	7204	0	0	2012-08-02 05:43:00.70149	2012-08-02 05:47:10.127949
9457	1-9457	38	92	8947	84	5881	0	0	2012-08-02 05:46:14.117597	2012-08-02 05:47:10.127949
7223	1-7223	95	28	7476	59	3265	0	0	2012-08-02 05:31:02.948327	2012-08-02 05:47:10.127949
8977	1-8977	44	4	9405	11	6313	0	0	2012-08-02 05:42:58.23806	2012-08-02 05:47:11.751938
7823	1-7823	2	96	9490	87	8322	0	0	2012-08-02 05:35:08.73073	2012-08-02 05:47:11.751938
9097	1-9097	56	65	3112	57	9655	0	6	2012-08-02 05:43:47.50661	2012-08-02 05:47:12.171871
9248	1-9248	26	79	4488	57	5770	0	0	2012-08-02 05:44:48.999008	2012-08-02 05:47:12.171871
1131	1-1131	78	8	8814	80	6166	0	0	2012-08-02 04:48:26.16228	2012-08-02 05:47:12.569746
9568	1-9568	88	42	1239	10	4854	0	0	2012-08-02 05:46:59.087226	2012-08-02 05:47:13.818537
9461	1-9461	48	42	7952	61	6624	0	4	2012-08-02 05:46:15.708524	2012-08-02 05:47:13.818537
9244	1-9244	7	65	1326	40	7672	0	6	2012-08-02 05:44:47.374285	2012-08-02 05:47:13.818537
9566	1-9566	7	42	9299	56	8636	0	4	2012-08-02 05:46:58.303137	2012-08-02 05:47:13.818537
9264	1-9264	62	100	9609	62	7501	0	6	2012-08-02 05:44:55.594987	2012-08-02 05:47:14.238393
8943	1-8943	54	18	1098	62	574	0	0	2012-08-02 05:42:44.260055	2012-08-02 05:47:14.238393
9587	1-9587	91	54	6343	65	6040	0	0	2012-08-02 05:47:06.846517	2012-08-02 05:47:14.238393
6840	1-6840	78	8	6965	36	6038	0	0	2012-08-02 05:28:26.679563	2012-08-02 05:47:14.238393
8596	1-8596	65	42	8059	65	4257	0	2	2012-08-02 05:40:23.775793	2012-08-02 05:47:16.315358
9610	1-9610	54	8	1835	59	7992	0	3	2012-08-02 05:47:16.315358	2012-08-02 05:47:16.315358
9613	1-9613	87	48	1124	12	975	0	0	2012-08-02 05:47:17.53076	2012-08-02 05:47:17.53076
9529	1-9529	66	74	2418	46	4196	0	10	2012-08-02 05:46:43.2656	2012-08-02 05:47:17.939705
8375	1-8375	71	24	3700	6	6424	0	2	2012-08-02 05:38:54.856046	2012-08-02 05:47:17.939705
8229	1-8229	69	74	7991	89	5446	0	0	2012-08-02 05:37:55.857949	2012-08-02 05:47:17.939705
9614	1-9614	28	6	1358	74	5191	0	0	2012-08-02 05:47:17.939705	2012-08-02 05:47:17.939705
7892	1-7892	47	33	5342	49	4248	0	0	2012-08-02 05:35:37.921269	2012-08-02 05:47:18.359673
8496	1-8496	73	33	5282	68	4782	0	0	2012-08-02 05:39:43.520671	2012-08-02 05:47:18.359673
9310	1-9310	21	43	5232	3	9024	0	9	2012-08-02 05:45:14.282687	2012-08-02 05:47:19.16651
9510	1-9510	5	50	4689	21	9215	0	11	2012-08-02 05:46:35.584983	2012-08-02 05:47:19.585978
9354	1-9354	11	43	3917	82	8568	0	4	2012-08-02 05:45:32.305078	2012-08-02 05:47:19.983907
8644	1-8644	56	12	7632	80	7677	0	0	2012-08-02 05:40:43.045957	2012-08-02 05:47:21.210083
1193	1-1193	2	20	5434	50	859	0	0	2012-08-02 04:48:55.472532	2012-08-02 05:47:21.210083
9413	1-9413	70	88	9014	4	8727	0	0	2012-08-02 05:45:56.127461	2012-08-02 05:47:22.005712
9582	1-9582	60	32	2927	20	8786	0	7	2012-08-02 05:47:04.821364	2012-08-02 05:47:22.005712
9625	1-9625	99	23	664	54	5607	0	9	2012-08-02 05:47:22.414736	2012-08-02 05:47:22.414736
777	1-777	52	75	9725	59	7332	0	0	2012-08-02 04:45:25.435261	2012-08-02 05:47:22.834437
9626	1-9626	65	90	3304	55	8745	0	0	2012-08-02 05:47:22.834437	2012-08-02 05:47:22.834437
9605	1-9605	72	62	2123	8	9512	0	0	2012-08-02 05:47:14.238393	2012-08-02 05:47:24.071957
9638	1-9638	19	62	50	83	6653	0	8	2012-08-02 05:47:27.750849	2012-08-02 05:47:27.750849
9639	1-9639	71	12	1179	30	4111	0	9	2012-08-02 05:47:28.181735	2012-08-02 05:47:28.181735
9642	1-9642	82	64	538	46	4863	0	0	2012-08-02 05:47:29.419164	2012-08-02 05:47:29.419164
8113	1-8113	84	24	4724	68	6322	0	5	2012-08-02 05:37:08.57941	2012-08-02 05:47:31.475678
9399	1-9399	78	15	4724	54	8130	0	3	2012-08-02 05:45:50.47021	2012-08-02 05:47:32.724175
7742	1-7742	33	75	7398	46	6454	0	2	2012-08-02 05:34:34.37579	2012-08-02 05:47:33.110919
9289	1-9289	49	9	1924	80	5916	0	0	2012-08-02 05:45:05.840532	2012-08-02 05:47:33.530537
9652	1-9652	31	28	419	9	8050	0	0	2012-08-02 05:47:33.530537	2012-08-02 05:47:33.530537
8352	1-8352	72	98	5495	64	2200	0	7	2012-08-02 05:38:45.561404	2012-08-02 05:47:34.33729
7456	1-7456	5	24	5066	26	4849	0	0	2012-08-02 05:32:36.060333	2012-08-02 05:47:34.33729
9654	1-9654	37	1	2665	24	2541	0	14	2012-08-02 05:47:34.33729	2012-08-02 05:47:34.33729
6024	1-6024	96	49	9230	1	5715	23	0	2012-08-02 05:22:54.503819	2012-08-02 05:47:34.33729
9446	1-9446	86	62	4285	82	8049	0	0	2012-08-02 05:46:09.629741	2012-08-02 05:47:34.768383
9157	1-9157	80	55	1722	51	2319	0	13	2012-08-02 05:44:12.115471	2012-08-02 05:47:35.188098
9656	1-9656	50	51	2260	55	8095	0	5	2012-08-02 05:47:35.188098	2012-08-02 05:47:35.188098
9148	1-9148	21	78	3901	67	2731	0	0	2012-08-02 05:44:08.524209	2012-08-02 05:47:35.597525
9146	1-9146	80	75	3490	72	6716	0	0	2012-08-02 05:44:07.716496	2012-08-02 05:47:35.597525
8296	1-8296	97	2	8469	81	3556	0	6	2012-08-02 05:38:23.078411	2012-08-02 05:47:37.232167
9622	1-9622	60	17	5984	12	9385	0	0	2012-08-02 05:47:21.210083	2012-08-02 05:47:37.662999
9662	1-9662	61	15	373	94	4912	0	11	2012-08-02 05:47:37.662999	2012-08-02 05:47:37.662999
9666	1-9666	13	80	1702	81	5140	0	5	2012-08-02 05:47:39.298403	2012-08-02 05:47:39.298403
8838	1-8838	61	69	5954	82	3131	0	0	2012-08-02 05:42:01.566658	2012-08-02 05:47:39.707015
9667	1-9667	47	82	1052	84	6681	0	15	2012-08-02 05:47:39.707015	2012-08-02 05:47:39.707015
8609	1-8609	67	2	9103	98	1874	0	2	2012-08-02 05:40:28.979555	2012-08-02 05:47:40.126866
8922	1-8922	60	34	1680	2	7702	0	0	2012-08-02 05:42:35.664683	2012-08-02 05:47:40.535865
9669	1-9669	87	6	1427	34	7556	0	6	2012-08-02 05:47:40.535865	2012-08-02 05:47:40.535865
9223	1-9223	74	83	4595	15	7417	0	3	2012-08-02 05:44:38.795089	2012-08-02 05:47:40.955724
9671	1-9671	45	83	52	5	9643	0	2	2012-08-02 05:47:41.386597	2012-08-02 05:47:41.386597
8545	1-8545	46	87	5704	38	4020	0	20	2012-08-02 05:40:03.144705	2012-08-02 05:47:41.79532
8050	1-8050	19	38	4616	48	5012	0	5	2012-08-02 05:36:43.221162	2012-08-02 05:47:41.79532
9634	1-9634	51	28	2936	25	3212	0	2	2012-08-02 05:47:26.137989	2012-08-02 05:47:42.612908
9678	1-9678	3	2	1522	84	4296	0	0	2012-08-02 05:47:44.281097	2012-08-02 05:47:44.281097
9292	1-9292	95	13	3599	94	7258	0	11	2012-08-02 05:45:07.045141	2012-08-02 05:47:45.087852
9615	1-9615	15	13	2001	33	4501	0	0	2012-08-02 05:47:18.359673	2012-08-02 05:47:45.087852
9604	1-9604	89	64	2031	42	8989	0	5	2012-08-02 05:47:13.818537	2012-08-02 05:47:45.087852
7937	1-7937	94	13	9151	63	6276	0	2	2012-08-02 05:35:57.089663	2012-08-02 05:47:45.087852
8731	1-8731	39	99	3522	59	2987	0	0	2012-08-02 05:41:18.215674	2012-08-02 05:47:45.087852
9335	1-9335	45	13	4448	43	2375	0	5	2012-08-02 05:45:24.46059	2012-08-02 05:47:45.087852
9105	1-9105	53	59	5471	42	9933	0	3	2012-08-02 05:43:50.82465	2012-08-02 05:47:45.087852
9680	1-9680	99	42	1464	13	9539	0	0	2012-08-02 05:47:45.087852	2012-08-02 05:47:45.087852
2741	1-2741	71	33	6420	58	3116	1232	0	2012-08-02 05:00:51.595483	2012-08-02 05:47:45.087852
8773	1-8773	98	1	7207	35	9146	0	5	2012-08-02 05:41:35.369431	2012-08-02 05:47:46.745132
7156	1-7156	58	1	6520	38	9635	0	6	2012-08-02 05:30:35.093749	2012-08-02 05:47:46.745132
7530	1-7530	83	1	4542	33	3843	0	0	2012-08-02 05:33:06.611623	2012-08-02 05:47:46.745132
9684	1-9684	41	59	2286	1	7779	0	8	2012-08-02 05:47:46.745132	2012-08-02 05:47:46.745132
9686	1-9686	59	91	4139	60	8967	0	0	2012-08-02 05:47:47.574191	2012-08-02 05:47:47.574191
9687	1-9687	39	56	5118	63	4440	0	0	2012-08-02 05:47:47.982886	2012-08-02 05:47:47.982886
9657	1-9657	12	67	2800	75	9975	0	0	2012-08-02 05:47:35.597525	2012-08-02 05:47:49.639851
8862	1-8862	92	95	9972	23	9744	0	8	2012-08-02 05:42:11.26841	2012-08-02 05:47:52.512375
8374	1-8374	19	43	3266	28	2418	0	3	2012-08-02 05:38:54.458634	2012-08-02 05:47:52.512375
9465	1-9465	19	95	1830	93	1494	0	0	2012-08-02 05:46:17.355013	2012-08-02 05:47:52.512375
5904	1-5904	5	23	9320	54	7767	1859	0	2012-08-02 05:22:04.548051	2012-08-02 05:47:52.512375
8026	1-8026	6	24	5444	67	9163	0	0	2012-08-02 05:36:33.587063	2012-08-02 05:47:53.34105
9392	1-9392	99	82	7699	58	7312	0	0	2012-08-02 05:45:47.653011	2012-08-02 05:47:53.34105
5519	1-5519	42	82	5824	89	2100	0	8	2012-08-02 05:19:28.860563	2012-08-02 05:47:53.34105
4827	1-4827	42	68	4014	37	3557	0	0	2012-08-02 05:14:54.190459	2012-08-02 05:47:54.567492
9703	1-9703	85	22	894	68	6609	0	0	2012-08-02 05:47:54.567492	2012-08-02 05:47:54.567492
9315	1-9315	27	9	4277	86	7595	0	0	2012-08-02 05:45:16.304664	2012-08-02 05:47:54.976469
9674	1-9674	61	56	1745	28	3356	0	0	2012-08-02 05:47:42.612908	2012-08-02 05:47:55.805281
4773	1-4773	92	56	7310	30	2925	0	4	2012-08-02 05:14:32.168119	2012-08-02 05:47:56.600435
8973	1-8973	4	56	7351	12	9674	0	0	2012-08-02 05:42:56.60261	2012-08-02 05:47:56.600435
9708	1-9708	31	67	444	56	7535	0	0	2012-08-02 05:47:56.600435	2012-08-02 05:47:56.600435
4959	1-4959	29	12	7654	89	1932	1722	0	2012-08-02 05:15:46.233386	2012-08-02 05:47:56.600435
9041	1-9041	8	55	2193	53	6583	0	7	2012-08-02 05:43:24.557814	2012-08-02 05:47:57.01149
1568	1-1568	57	35	6924	32	2845	0	0	2012-08-02 04:52:15.173634	2012-08-02 05:47:57.420308
7113	1-7113	41	17	4660	2	9704	0	4	2012-08-02 05:30:17.149747	2012-08-02 05:47:57.420308
6382	1-6382	73	35	1404	21	406	0	7	2012-08-02 05:25:21.689966	2012-08-02 05:47:57.420308
9710	1-9710	98	11	1164	35	7236	0	0	2012-08-02 05:47:57.420308	2012-08-02 05:47:57.420308
6302	1-6302	49	12	5188	92	5698	0	14	2012-08-02 05:24:48.573439	2012-08-02 05:47:57.851532
9711	1-9711	17	92	1322	61	3662	0	4	2012-08-02 05:47:57.851532	2012-08-02 05:47:57.851532
9543	1-9543	72	80	6553	43	6222	0	7	2012-08-02 05:46:49.022068	2012-08-02 05:47:58.282609
9086	1-9086	43	2	5865	11	719	0	0	2012-08-02 05:43:42.985864	2012-08-02 05:47:58.282609
9712	1-9712	68	59	1353	80	9060	0	0	2012-08-02 05:47:58.282609	2012-08-02 05:47:58.282609
8994	1-8994	5	78	4436	28	4960	0	16	2012-08-02 05:43:05.187159	2012-08-02 05:47:59.519802
9406	1-9406	57	44	7050	52	5040	0	11	2012-08-02 05:45:53.309948	2012-08-02 05:48:01.574606
9085	1-9085	83	51	7438	50	8238	0	2	2012-08-02 05:43:42.588408	2012-08-02 05:48:02.381372
9722	1-9722	29	95	533	51	6101	0	0	2012-08-02 05:48:02.381372	2012-08-02 05:48:02.381372
9691	1-9691	41	75	4460	18	5891	0	15	2012-08-02 05:47:49.639851	2012-08-02 05:48:03.651948
8728	1-8728	64	96	7352	79	7706	0	0	2012-08-02 05:41:17.011346	2012-08-02 05:48:04.900447
1742	1-1742	5	6	7402	15	2776	0	0	2012-08-02 04:53:36.405503	2012-08-02 05:48:04.900447
9728	1-9728	51	98	3404	96	6971	0	9	2012-08-02 05:48:04.900447	2012-08-02 05:48:04.900447
6365	1-6365	67	96	9088	6	6399	2230	0	2012-08-02 05:25:14.751856	2012-08-02 05:48:04.900447
9627	1-9627	25	60	9884	91	8343	0	0	2012-08-02 05:47:23.254391	2012-08-02 05:48:05.76293
1301	1-1301	7	91	3459	3	708	0	0	2012-08-02 04:49:54.936811	2012-08-02 05:48:05.76293
9730	1-9730	62	3	265	60	5064	0	5	2012-08-02 05:48:05.76293	2012-08-02 05:48:05.76293
9555	1-9555	50	70	7742	52	9843	0	4	2012-08-02 05:46:53.850429	2012-08-02 05:48:06.956922
9733	1-9733	23	52	859	44	5197	0	0	2012-08-02 05:48:06.956922	2012-08-02 05:48:06.956922
8648	1-8648	25	51	7639	98	6499	0	0	2012-08-02 05:40:44.636934	2012-08-02 05:48:07.752523
8687	1-8687	86	51	8031	84	6114	0	0	2012-08-02 05:41:00.315031	2012-08-02 05:48:07.752523
9551	1-9551	91	51	2154	66	1909	0	3	2012-08-02 05:46:52.237103	2012-08-02 05:48:07.752523
9735	1-9735	43	20	864	51	5741	0	0	2012-08-02 05:48:07.752523	2012-08-02 05:48:07.752523
7836	1-7836	73	93	6415	20	2235	1950	10	2012-08-02 05:35:14.318784	2012-08-02 05:48:07.752523
5950	1-5950	42	98	5321	93	2601	2601	0	2012-08-02 05:22:23.684959	2012-08-02 05:48:07.752523
7210	1-7210	26	22	5049	84	6108	0	0	2012-08-02 05:30:57.565778	2012-08-02 05:48:08.194292
9736	1-9736	87	84	1693	70	9756	0	8	2012-08-02 05:48:08.194292	2012-08-02 05:48:08.194292
7944	1-7944	9	22	4677	84	5375	2922	0	2012-08-02 05:36:00.039815	2012-08-02 05:48:08.194292
9737	1-9737	36	87	2358	66	1779	0	16	2012-08-02 05:48:08.603814	2012-08-02 05:48:08.603814
9742	1-9742	44	56	1486	69	9551	0	0	2012-08-02 05:48:10.569865	2012-08-02 05:48:10.569865
9698	1-9698	81	54	1244	95	4448	0	0	2012-08-02 05:47:52.512375	2012-08-02 05:48:11.365608
9744	1-9744	40	95	1435	32	9355	0	4	2012-08-02 05:48:11.365608	2012-08-02 05:48:11.365608
7869	1-7869	14	86	9660	47	3739	0	5	2012-08-02 05:35:28.252985	2012-08-02 05:48:11.774426
6673	1-6673	13	66	9913	47	4202	0	0	2012-08-02 05:27:19.804787	2012-08-02 05:48:11.774426
9745	1-9745	10	41	1053	86	7388	0	0	2012-08-02 05:48:11.774426	2012-08-02 05:48:11.774426
4513	1-4513	14	41	7577	43	3406	0	0	2012-08-02 05:12:46.885613	2012-08-02 05:48:12.978594
9630	1-9630	78	95	3315	43	3573	0	9	2012-08-02 05:47:24.480707	2012-08-02 05:48:12.978594
9339	1-9339	91	53	8844	26	3628	0	0	2012-08-02 05:45:26.150593	2012-08-02 05:48:13.398628
8514	1-8514	71	34	1709	37	3662	0	0	2012-08-02 05:39:50.737354	2012-08-02 05:48:13.796391
9609	1-9609	33	9	3583	22	7057	0	0	2012-08-02 05:47:15.896072	2012-08-02 05:48:13.796391
9750	1-9750	95	50	5077	9	8436	0	16	2012-08-02 05:48:13.796391	2012-08-02 05:48:13.796391
9269	1-9269	23	96	5785	6	9805	0	5	2012-08-02 05:44:57.629822	2012-08-02 05:48:16.205043
8081	1-8081	95	96	2201	88	4276	0	6	2012-08-02 05:36:55.638927	2012-08-02 05:48:16.205043
9748	1-9748	70	43	3367	33	6277	0	0	2012-08-02 05:48:12.978594	2012-08-02 05:48:16.205043
6518	1-6518	19	96	6446	45	4698	0	0	2012-08-02 05:26:16.543372	2012-08-02 05:48:16.205043
6970	1-6970	46	98	4630	33	2272	0	0	2012-08-02 05:29:19.911492	2012-08-02 05:48:16.205043
9362	1-9362	35	64	5676	21	9246	0	0	2012-08-02 05:45:35.553347	2012-08-02 05:48:16.625104
9757	1-9757	66	21	3053	34	4779	0	8	2012-08-02 05:48:16.625104	2012-08-02 05:48:16.625104
8646	1-8646	15	86	8048	33	3517	0	0	2012-08-02 05:40:43.841438	2012-08-02 05:48:17.454
9759	1-9759	67	33	2474	65	8519	0	0	2012-08-02 05:48:17.454	2012-08-02 05:48:18.271508
9756	1-9756	39	33	2947	96	8973	0	0	2012-08-02 05:48:16.205043	2012-08-02 05:48:18.271508
9761	1-9761	87	44	728	33	7279	0	8	2012-08-02 05:48:18.271508	2012-08-02 05:48:18.271508
9179	1-9179	18	94	7347	38	8467	0	0	2012-08-02 05:44:20.990529	2012-08-02 05:48:20.261708
9766	1-9766	30	38	4346	62	7387	0	2	2012-08-02 05:48:20.261708	2012-08-02 05:48:20.261708
9771	1-9771	52	39	102	45	5809	0	15	2012-08-02 05:48:22.239629	2012-08-02 05:48:22.239629
6771	1-6771	64	80	6738	50	3261	0	0	2012-08-02 05:27:58.800654	2012-08-02 05:48:23.852354
9351	1-9351	75	61	5010	50	7242	0	5	2012-08-02 05:45:31.100654	2012-08-02 05:48:23.852354
9433	1-9433	28	85	8716	80	8379	0	5	2012-08-02 05:46:04.193326	2012-08-02 05:48:24.250183
9775	1-9775	98	50	2863	80	6365	0	8	2012-08-02 05:48:23.852354	2012-08-02 05:48:24.250183
9776	1-9776	29	80	4410	77	7366	0	6	2012-08-02 05:48:24.250183	2012-08-02 05:48:24.250183
9033	1-9033	53	49	8427	13	8605	0	2	2012-08-02 05:43:21.265582	2012-08-02 05:48:24.670206
9026	1-9026	98	30	5865	41	7291	0	4	2012-08-02 05:43:18.392919	2012-08-02 05:48:24.670206
9398	1-9398	68	86	6978	58	6158	0	7	2012-08-02 05:45:50.072422	2012-08-02 05:48:24.670206
9777	1-9777	30	13	2654	30	6724	0	0	2012-08-02 05:48:24.670206	2012-08-02 05:48:24.670206
9778	1-9778	81	94	750	30	3128	0	8	2012-08-02 05:48:25.079237	2012-08-02 05:48:25.079237
9631	1-9631	7	34	1792	66	4050	0	0	2012-08-02 05:47:24.889673	2012-08-02 05:48:25.89676
9239	1-9239	79	34	4450	71	8481	0	2	2012-08-02 05:44:45.36408	2012-08-02 05:48:25.89676
8426	1-8426	61	34	5050	70	7572	0	0	2012-08-02 05:39:15.552389	2012-08-02 05:48:25.89676
9416	1-9416	96	34	4031	42	9435	0	0	2012-08-02 05:45:57.354261	2012-08-02 05:48:25.89676
9780	1-9780	37	90	580	34	9403	0	7	2012-08-02 05:48:25.89676	2012-08-02 05:48:25.89676
9594	1-9594	81	66	5589	24	4273	325	0	2012-08-02 05:47:09.729894	2012-08-02 05:48:25.89676
9783	1-9783	40	36	711	29	4018	0	5	2012-08-02 05:48:27.145569	2012-08-02 05:48:27.145569
9646	1-9646	30	86	3395	36	5057	0	0	2012-08-02 05:47:31.078081	2012-08-02 05:48:27.543163
9253	1-9253	41	55	5994	75	8601	0	3	2012-08-02 05:44:51.109599	2012-08-02 05:48:27.543163
9635	1-9635	9	69	6580	94	8859	0	16	2012-08-02 05:47:26.524834	2012-08-02 05:48:27.952212
9562	1-9562	47	83	6228	56	3656	0	0	2012-08-02 05:46:56.678857	2012-08-02 05:48:27.952212
9619	1-9619	67	58	6921	35	8347	0	0	2012-08-02 05:47:19.983907	2012-08-02 05:48:28.371925
9487	1-9487	12	16	7287	64	1815	0	0	2012-08-02 05:46:26.248791	2012-08-02 05:48:28.758694
9787	1-9787	28	55	79	16	5570	0	7	2012-08-02 05:48:28.758694	2012-08-02 05:48:28.758694
9788	1-9788	11	53	537	66	2919	0	3	2012-08-02 05:48:29.189664	2012-08-02 05:48:29.189664
7860	1-7860	32	90	5056	16	6803	0	0	2012-08-02 05:35:24.452113	2012-08-02 05:48:29.576444
5008	1-5008	40	19	8715	56	9508	0	0	2012-08-02 05:16:05.560292	2012-08-02 05:48:29.576444
8793	1-8793	55	90	4228	39	1611	0	0	2012-08-02 05:41:43.41249	2012-08-02 05:48:29.576444
8441	1-8441	85	56	6001	16	7257	0	0	2012-08-02 05:39:21.552901	2012-08-02 05:48:29.576444
5513	1-5513	97	19	7335	56	5762	5406	0	2012-08-02 05:19:26.474528	2012-08-02 05:48:29.576444
9731	1-9731	9	74	1118	90	2864	0	10	2012-08-02 05:48:06.172266	2012-08-02 05:48:31.244712
9796	1-9796	38	48	209	76	9884	0	0	2012-08-02 05:48:32.448977	2012-08-02 05:48:32.448977
7799	1-7799	36	6	7444	16	5814	0	9	2012-08-02 05:34:58.463529	2012-08-02 05:48:32.835784
9557	1-9557	58	62	2189	32	3353	0	0	2012-08-02 05:46:54.646099	2012-08-02 05:48:32.835784
5777	1-5777	23	6	9878	16	7688	2163	2	2012-08-02 05:21:13.324824	2012-08-02 05:48:32.835784
9632	1-9632	100	11	2889	20	5977	0	11	2012-08-02 05:47:25.30938	2012-08-02 05:48:34.481883
9137	1-9137	51	46	9691	43	8146	0	13	2012-08-02 05:44:04.027398	2012-08-02 05:48:34.481883
8131	1-8131	26	11	4144	36	4029	0	0	2012-08-02 05:37:15.937812	2012-08-02 05:48:34.481883
9801	1-9801	37	43	2076	11	5648	0	0	2012-08-02 05:48:34.481883	2012-08-02 05:48:34.481883
7741	1-7741	69	92	8487	70	3878	0	8	2012-08-02 05:34:33.955705	2012-08-02 05:48:34.890991
7252	1-7252	90	92	8924	81	3069	0	11	2012-08-02 05:31:14.451181	2012-08-02 05:48:34.890991
7264	1-7264	60	79	7648	15	5931	0	7	2012-08-02 05:31:19.113965	2012-08-02 05:48:35.697427
9786	1-9786	93	17	1971	29	5202	0	0	2012-08-02 05:48:28.371925	2012-08-02 05:48:36.095293
6115	1-6115	45	19	1501	29	2158	0	0	2012-08-02 05:23:31.353435	2012-08-02 05:48:36.095293
7765	1-7765	63	86	6143	3	3060	0	0	2012-08-02 05:34:44.110573	2012-08-02 05:48:36.504015
8686	1-8686	89	45	7357	64	5408	0	0	2012-08-02 05:40:59.928311	2012-08-02 05:48:36.504015
9621	1-9621	65	70	9210	22	6707	0	0	2012-08-02 05:47:20.801679	2012-08-02 05:48:37.29974
9503	1-9503	74	90	8506	19	6302	0	0	2012-08-02 05:46:32.701571	2012-08-02 05:48:37.29974
9611	1-9611	48	37	3502	22	3899	0	7	2012-08-02 05:47:16.724427	2012-08-02 05:48:38.095017
9808	1-9808	84	52	1930	18	4753	0	0	2012-08-02 05:48:37.29974	2012-08-02 05:48:38.095017
9459	1-9459	26	87	7428	28	7050	0	0	2012-08-02 05:46:14.90182	2012-08-02 05:48:38.51513
9813	1-9813	54	61	78	17	6386	0	9	2012-08-02 05:48:39.321516	2012-08-02 05:48:39.321516
7359	1-7359	36	20	2188	34	1365	0	0	2012-08-02 05:31:56.439165	2012-08-02 05:48:39.741457
8522	1-8522	82	34	4832	21	3163	0	10	2012-08-02 05:39:53.897185	2012-08-02 05:48:39.741457
9709	1-9709	85	25	2843	55	2790	0	7	2012-08-02 05:47:57.01149	2012-08-02 05:48:40.150335
9670	1-9670	29	15	3759	98	7168	0	6	2012-08-02 05:47:40.955724	2012-08-02 05:48:40.150335
9205	1-9205	15	27	4675	87	6501	0	8	2012-08-02 05:44:31.53355	2012-08-02 05:48:41.763592
9820	1-9820	28	88	783	53	6199	0	5	2012-08-02 05:48:42.17222	2012-08-02 05:48:42.17222
7494	1-7494	65	21	7702	29	5316	0	4	2012-08-02 05:32:51.638357	2012-08-02 05:48:43.38771
9665	1-9665	74	60	4008	100	4727	0	0	2012-08-02 05:47:38.878367	2012-08-02 05:48:45.013475
2406	1-2406	19	75	3962	9	926	0	0	2012-08-02 04:58:27.169449	2012-08-02 05:48:45.809003
9397	1-9397	91	82	3141	7	3356	0	0	2012-08-02 05:45:49.674953	2012-08-02 05:48:45.809003
6924	1-6924	91	75	7692	68	4381	0	0	2012-08-02 05:29:01.366082	2012-08-02 05:48:45.809003
9257	1-9257	64	75	1877	15	1125	0	3	2012-08-02 05:44:52.744957	2012-08-02 05:48:45.809003
9304	1-9304	13	75	2998	79	5743	0	0	2012-08-02 05:45:11.862225	2012-08-02 05:48:45.809003
8606	1-8606	63	100	9219	91	8742	0	0	2012-08-02 05:40:27.797413	2012-08-02 05:48:46.228904
3014	1-3014	4	56	6898	54	5336	0	0	2012-08-02 05:02:44.962859	2012-08-02 05:48:46.228904
5400	1-5400	51	56	8825	100	6775	620	6	2012-08-02 05:18:41.568417	2012-08-02 05:48:46.228904
7749	1-7749	61	68	9562	36	6581	0	7	2012-08-02 05:34:37.314822	2012-08-02 05:48:47.841851
9830	1-9830	26	39	1802	56	7245	0	0	2012-08-02 05:48:46.228904	2012-08-02 05:48:48.2622
9548	1-9548	63	37	5415	50	4214	0	2	2012-08-02 05:46:51.043657	2012-08-02 05:48:48.2622
9661	1-9661	18	81	4322	50	4445	0	0	2012-08-02 05:47:37.232167	2012-08-02 05:48:48.2622
9502	1-9502	43	89	3209	84	9098	0	5	2012-08-02 05:46:32.31477	2012-08-02 05:48:49.07957
9837	1-9837	16	73	2185	89	4864	0	5	2012-08-02 05:48:49.07957	2012-08-02 05:48:49.07957
6937	1-6937	54	44	5830	34	3593	0	0	2012-08-02 05:29:06.747953	2012-08-02 05:48:50.289463
8555	1-8555	66	98	8454	93	7288	0	5	2012-08-02 05:40:07.15687	2012-08-02 05:48:50.289463
8817	1-8817	77	64	5153	73	8973	0	0	2012-08-02 05:41:53.103861	2012-08-02 05:48:51.14535
7247	1-7247	56	20	6891	70	6910	0	0	2012-08-02 05:31:12.540258	2012-08-02 05:48:51.963868
8802	1-8802	24	39	3582	37	6744	0	3	2012-08-02 05:41:47.047543	2012-08-02 05:48:53.578593
9681	1-9681	72	41	6818	56	8294	0	8	2012-08-02 05:47:45.508082	2012-08-02 05:48:53.97753
7455	1-7455	33	52	4818	2	9859	0	10	2012-08-02 05:32:35.651589	2012-08-02 05:48:54.398738
9850	1-9850	75	2	5777	59	5879	0	0	2012-08-02 05:48:54.398738	2012-08-02 05:48:54.398738
9789	1-9789	18	16	6787	90	9825	0	0	2012-08-02 05:48:29.576444	2012-08-02 05:48:55.593233
9494	1-9494	38	21	6434	27	9700	0	11	2012-08-02 05:46:29.05505	2012-08-02 05:48:56.003094
8151	1-8151	77	21	9720	27	416	416	8	2012-08-02 05:37:24.013643	2012-08-02 05:48:56.003094
9829	1-9829	68	7	1949	75	8769	0	13	2012-08-02 05:48:45.809003	2012-08-02 05:48:56.414794
9651	1-9651	89	35	4083	75	3548	0	4	2012-08-02 05:47:33.110919	2012-08-02 05:48:57.642823
9599	1-9599	13	11	2449	96	6022	0	4	2012-08-02 05:47:11.751938	2012-08-02 05:48:57.642823
9860	1-9860	5	50	401	18	5874	0	0	2012-08-02 05:48:58.43927	2012-08-02 05:48:58.43927
9805	1-9805	24	29	3578	19	4800	0	11	2012-08-02 05:48:36.095293	2012-08-02 05:48:58.837135
5852	1-5852	20	32	3710	68	1896	0	0	2012-08-02 05:21:43.587709	2012-08-02 05:48:58.837135
9725	1-9725	92	19	1389	75	2520	0	0	2012-08-02 05:48:03.651948	2012-08-02 05:48:58.837135
9861	1-9861	12	75	1959	32	5866	0	0	2012-08-02 05:48:58.837135	2012-08-02 05:48:58.837135
9252	1-9252	38	70	5771	11	1955	0	3	2012-08-02 05:44:50.714468	2012-08-02 05:49:00.088168
3661	1-3661	31	70	8001	50	3563	0	0	2012-08-02 05:07:06.645679	2012-08-02 05:49:00.088168
8304	1-8304	93	2	5565	58	1673	0	0	2012-08-02 05:38:26.270964	2012-08-02 05:49:00.088168
9720	1-9720	70	36	3694	44	6525	0	10	2012-08-02 05:48:01.574606	2012-08-02 05:49:00.896122
7337	1-7337	23	38	7665	14	7470	0	3	2012-08-02 05:31:47.786134	2012-08-02 05:49:00.896122
5541	1-5541	51	39	6638	44	6183	0	7	2012-08-02 05:19:37.56669	2012-08-02 05:49:00.896122
9866	1-9866	89	44	2634	38	4949	0	0	2012-08-02 05:49:00.896122	2012-08-02 05:49:00.896122
9870	1-9870	11	22	171	23	5676	0	0	2012-08-02 05:49:02.543868	2012-08-02 05:49:02.543868
9668	1-9668	13	80	1739	2	9148	0	11	2012-08-02 05:47:40.126866	2012-08-02 05:49:02.955284
7315	1-7315	32	64	6820	88	8562	0	6	2012-08-02 05:31:39.146339	2012-08-02 05:49:04.159812
9802	1-9802	38	40	5259	92	8770	0	6	2012-08-02 05:48:34.890991	2012-08-02 05:49:04.977757
7834	1-7834	81	82	3823	92	4834	0	0	2012-08-02 05:35:13.448319	2012-08-02 05:49:04.977757
9876	1-9876	52	92	2635	70	8566	0	13	2012-08-02 05:49:04.977757	2012-08-02 05:49:04.977757
8430	1-8430	34	75	8126	78	6179	0	0	2012-08-02 05:39:17.144139	2012-08-02 05:49:07.82463
9388	1-9388	48	51	7633	75	5901	0	4	2012-08-02 05:45:46.073108	2012-08-02 05:49:07.82463
7384	1-7384	36	16	8460	59	1611	0	0	2012-08-02 05:32:06.31713	2012-08-02 05:49:08.655485
9885	1-9885	34	64	603	16	8352	0	0	2012-08-02 05:49:08.655485	2012-08-02 05:49:08.655485
8634	1-8634	95	83	6053	17	9885	0	12	2012-08-02 05:40:39.0004	2012-08-02 05:49:09.075438
7664	1-7664	55	81	5591	88	6469	0	0	2012-08-02 05:34:01.424054	2012-08-02 05:49:09.474205
9685	1-9685	3	46	5994	52	5193	0	6	2012-08-02 05:47:47.165285	2012-08-02 05:49:10.302964
9327	1-9327	50	14	2552	60	3452	0	0	2012-08-02 05:45:21.233576	2012-08-02 05:49:11.132101
9113	1-9113	69	83	4317	6	3337	0	0	2012-08-02 05:43:54.159077	2012-08-02 05:49:11.132101
9886	1-9886	12	96	2363	83	9928	0	0	2012-08-02 05:49:09.075438	2012-08-02 05:49:11.132101
9858	1-9858	34	96	2715	35	8036	0	13	2012-08-02 05:48:57.642823	2012-08-02 05:49:11.132101
9102	1-9102	41	68	3781	60	3471	0	0	2012-08-02 05:43:49.550604	2012-08-02 05:49:11.132101
9672	1-9672	87	48	3274	27	3391	0	9	2012-08-02 05:47:41.79532	2012-08-02 05:49:11.564364
8661	1-8661	15	59	8410	69	8569	0	0	2012-08-02 05:40:49.896652	2012-08-02 05:49:11.564364
9864	1-9864	59	9	463	70	3490	0	2	2012-08-02 05:49:00.088168	2012-08-02 05:49:12.371041
9618	1-9618	21	29	7038	51	7557	0	3	2012-08-02 05:47:19.585978	2012-08-02 05:49:12.371041
6831	1-6831	36	9	5716	29	9362	0	0	2012-08-02 05:28:22.989211	2012-08-02 05:49:12.371041
9894	1-9894	53	51	3345	9	8857	0	0	2012-08-02 05:49:12.371041	2012-08-02 05:49:12.371041
9700	1-9700	18	67	1287	82	9533	0	7	2012-08-02 05:47:53.34105	2012-08-02 05:49:13.18945
9427	1-9427	52	54	2863	5	3876	0	0	2012-08-02 05:46:01.784474	2012-08-02 05:49:14.416483
9899	1-9899	90	5	2846	8	7593	0	6	2012-08-02 05:49:14.416483	2012-08-02 05:49:14.416483
9891	1-9891	40	60	5476	96	6973	0	0	2012-08-02 05:49:11.132101	2012-08-02 05:49:15.234663
9903	1-9903	49	69	488	14	4368	0	9	2012-08-02 05:49:16.030517	2012-08-02 05:49:16.030517
9906	1-9906	75	45	1663	49	4433	0	4	2012-08-02 05:49:17.223699	2012-08-02 05:49:17.223699
9833	1-9833	59	69	7077	52	6577	0	4	2012-08-02 05:48:47.45524	2012-08-02 05:49:18.428365
9909	1-9909	19	13	849	84	5437	0	0	2012-08-02 05:49:18.428365	2012-08-02 05:49:18.428365
5118	1-5118	84	45	6596	14	8303	0	0	2012-08-02 05:16:50.331117	2012-08-02 05:49:18.826401
9823	1-9823	93	29	4522	82	9577	0	8	2012-08-02 05:48:43.38771	2012-08-02 05:49:19.635148
9180	1-9180	78	27	3280	82	2590	0	0	2012-08-02 05:44:21.399471	2012-08-02 05:49:19.635148
9840	1-9840	66	34	2465	24	2621	0	7	2012-08-02 05:48:50.289463	2012-08-02 05:49:20.851453
8439	1-8439	39	53	9171	93	5719	0	0	2012-08-02 05:39:20.768539	2012-08-02 05:49:20.851453
9924	1-9924	75	34	164	51	1627	0	0	2012-08-02 05:49:24.530652	2012-08-02 05:49:24.530652
9519	1-9519	11	97	3740	78	7221	0	0	2012-08-02 05:46:39.200116	2012-08-02 05:49:25.768195
9302	1-9302	22	97	6167	86	8406	0	0	2012-08-02 05:45:11.0668	2012-08-02 05:49:25.768195
8122	1-8122	14	27	9839	1	6222	0	0	2012-08-02 05:37:12.203338	2012-08-02 05:49:26.60852
9312	1-9312	68	21	4091	41	3382	0	11	2012-08-02 05:45:15.089316	2012-08-02 05:49:26.60852
9929	1-9929	51	22	1959	27	7844	0	0	2012-08-02 05:49:26.60852	2012-08-02 05:49:26.60852
7728	1-7728	94	11	5314	66	6074	0	0	2012-08-02 05:34:28.375899	2012-08-02 05:49:27.437766
9923	1-9923	100	64	2050	66	1516	0	0	2012-08-02 05:49:24.13301	2012-08-02 05:49:27.437766
9931	1-9931	47	66	2460	11	4681	0	0	2012-08-02 05:49:27.437766	2012-08-02 05:49:27.437766
9874	1-9874	4	23	5747	64	8199	0	3	2012-08-02 05:49:04.159812	2012-08-02 05:49:28.255409
9056	1-9056	36	88	5512	14	3702	0	2	2012-08-02 05:43:30.711289	2012-08-02 05:49:28.255409
5134	1-5134	17	12	8443	95	6662	0	0	2012-08-02 05:16:56.795678	2012-08-02 05:49:28.664271
8807	1-8807	6	83	8790	95	9151	0	6	2012-08-02 05:41:49.047542	2012-08-02 05:49:28.664271
9644	1-9644	49	46	1717	83	1628	0	6	2012-08-02 05:47:30.249293	2012-08-02 05:49:29.482193
7014	1-7014	34	18	7323	90	6677	0	6	2012-08-02 05:29:37.327614	2012-08-02 05:49:29.482193
9497	1-9497	88	50	3093	17	9227	0	4	2012-08-02 05:46:30.281797	2012-08-02 05:49:31.912435
9541	1-9541	58	91	2900	17	6879	0	0	2012-08-02 05:46:48.22651	2012-08-02 05:49:31.912435
8902	1-8902	36	9	4525	68	5046	0	7	2012-08-02 05:42:27.544382	2012-08-02 05:49:32.322021
9945	1-9945	52	18	74	67	4675	0	0	2012-08-02 05:49:33.128531	2012-08-02 05:49:33.128531
9005	1-9005	64	38	2752	91	2852	0	0	2012-08-02 05:43:09.762789	2012-08-02 05:49:33.559525
8587	1-8587	51	93	6652	20	7735	0	13	2012-08-02 05:40:20.152002	2012-08-02 05:49:35.592647
8878	1-8878	77	93	7380	94	6870	0	0	2012-08-02 05:42:17.79795	2012-08-02 05:49:35.592647
9624	1-9624	57	20	5404	88	7385	0	8	2012-08-02 05:47:22.005712	2012-08-02 05:49:35.592647
3800	1-3800	9	35	8586	67	3959	94	0	2012-08-02 05:08:01.623253	2012-08-02 05:49:35.592647
9806	1-9806	28	68	3225	45	6728	0	2	2012-08-02 05:48:36.504015	2012-08-02 05:49:37.217105
8152	1-8152	2	7	8343	63	2653	0	0	2012-08-02 05:37:24.411272	2012-08-02 05:49:38.854063
9959	1-9959	74	95	1628	7	9177	0	0	2012-08-02 05:49:38.854063	2012-08-02 05:49:38.854063
7478	1-7478	29	45	7464	69	4270	0	0	2012-08-02 05:32:45.086204	2012-08-02 05:49:39.727034
9961	1-9961	70	72	3652	45	8058	0	0	2012-08-02 05:49:39.727034	2012-08-02 05:49:39.727034
6639	1-6639	90	34	925	29	1251	0	4	2012-08-02 05:27:06.224637	2012-08-02 05:49:42.974963
9016	1-9016	1	54	7455	23	4070	0	0	2012-08-02 05:43:14.214961	2012-08-02 05:49:43.837396
9815	1-9815	47	5	1567	15	2089	0	0	2012-08-02 05:48:40.150335	2012-08-02 05:49:44.283805
9648	1-9648	22	8	3662	54	8174	0	14	2012-08-02 05:47:31.906613	2012-08-02 05:49:44.283805
5578	1-5578	69	97	6280	36	3166	0	9	2012-08-02 05:19:52.318534	2012-08-02 05:49:44.283805
9972	1-9972	64	25	3025	8	7704	0	0	2012-08-02 05:49:44.283805	2012-08-02 05:49:44.283805
8981	1-8981	27	4	9359	86	8346	0	5	2012-08-02 05:42:59.861566	2012-08-02 05:49:46.31251
5618	1-5618	78	79	8635	62	2728	0	0	2012-08-02 05:20:08.420286	2012-08-02 05:49:46.31251
9977	1-9977	79	34	133	79	3843	0	16	2012-08-02 05:49:46.31251	2012-08-02 05:49:46.31251
4052	1-4052	41	34	8269	79	2120	695	8	2012-08-02 05:09:42.589055	2012-08-02 05:49:46.31251
8919	1-8919	3	49	3504	31	2996	0	0	2012-08-02 05:42:34.438237	2012-08-02 05:49:46.721681
7141	1-7141	100	3	8827	32	7078	0	0	2012-08-02 05:30:28.862935	2012-08-02 05:49:46.721681
8828	1-8828	96	63	6996	41	6727	0	5	2012-08-02 05:41:57.534305	2012-08-02 05:49:46.721681
9978	1-9978	16	32	339	49	3678	0	10	2012-08-02 05:49:46.721681	2012-08-02 05:49:46.721681
2869	1-2869	32	49	4370	63	1533	493	0	2012-08-02 05:01:45.330434	2012-08-02 05:49:46.721681
9883	1-9883	86	78	3825	51	8409	0	0	2012-08-02 05:49:07.82463	2012-08-02 05:49:48.787994
9595	1-9595	58	72	4333	76	9672	0	0	2012-08-02 05:47:10.127949	2012-08-02 05:49:50.434186
8460	1-8460	5	84	9686	78	5403	0	0	2012-08-02 05:39:29.145982	2012-08-02 05:49:50.434186
9987	1-9987	51	47	467	84	4398	0	4	2012-08-02 05:49:50.434186	2012-08-02 05:49:50.434186
9989	1-9989	87	69	2202	60	3782	0	0	2012-08-02 05:49:51.2851	2012-08-02 05:49:51.2851
9969	1-9969	65	29	3986	69	8308	0	0	2012-08-02 05:49:42.974963	2012-08-02 05:49:51.671711
7672	1-7672	51	44	5542	70	4240	0	0	2012-08-02 05:34:04.674601	2012-08-02 05:49:51.671711
9990	1-9990	88	70	1821	61	9617	0	12	2012-08-02 05:49:51.671711	2012-08-02 05:49:51.671711
8056	1-8056	77	52	7046	31	4767	0	10	2012-08-02 05:36:45.641039	2012-08-02 05:49:52.500517
9869	1-9869	98	52	4964	13	4926	0	6	2012-08-02 05:49:02.146133	2012-08-02 05:49:52.500517
7590	1-7590	7	52	6697	68	3893	0	0	2012-08-02 05:33:31.062812	2012-08-02 05:49:52.500517
9992	1-9992	8	31	2077	52	9390	0	0	2012-08-02 05:49:52.500517	2012-08-02 05:49:52.500517
9993	1-9993	88	81	509	91	4729	0	0	2012-08-02 05:49:52.909691	2012-08-02 05:49:52.909691
8694	1-8694	22	24	2939	76	5133	0	6	2012-08-02 05:41:03.341912	2012-08-02 05:49:54.986724
9575	1-9575	74	88	1040	5	7519	0	0	2012-08-02 05:47:01.926869	2012-08-02 09:10:31.355931
7853	1-7853	25	74	3188	100	1072	0	7	2012-08-02 05:35:21.480453	2012-08-02 09:10:31.355931
10005	1-10005	78	86	864	74	8864	0	7	2012-08-02 09:10:31.355931	2012-08-02 09:10:31.355931
9692	1-9692	82	89	3060	34	3719	0	8	2012-08-02 05:47:50.059823	2012-08-02 09:10:31.811847
10006	1-10006	1	34	2879	61	8313	0	0	2012-08-02 09:10:31.811847	2012-08-02 09:10:31.811847
9979	1-9979	28	22	314	43	184	0	8	2012-08-02 05:49:47.164054	2012-08-02 10:16:55.717318
9299	1-9299	71	69	7009	6	6101	0	0	2012-08-02 05:45:09.851576	2012-08-02 10:16:55.717318
10011	1-10011	60	21	4317	41	5077	0	13	2012-08-02 10:16:55.717318	2012-08-02 10:16:55.717318
7789	1-7789	92	46	9083	83	5271	0	0	2012-08-02 05:34:54.265529	2012-08-02 10:16:56.176296
7911	1-7911	4	58	8153	53	9060	0	0	2012-08-02 05:35:45.997867	2012-08-02 10:16:56.587857
8695	1-8695	97	16	8812	53	5450	0	0	2012-08-02 05:41:03.750739	2012-08-02 10:16:56.587857
9206	1-9206	30	7	7785	77	9129	0	0	2012-08-02 05:44:31.942113	2012-08-02 10:16:56.987407
10014	1-10014	72	17	1818	7	4365	0	0	2012-08-02 10:16:56.987407	2012-08-02 10:16:56.987407
9915	1-9915	68	96	1327	34	4469	0	5	2012-08-02 05:49:20.851453	2012-08-02 10:16:57.420724
9232	1-9232	98	96	2755	16	8307	0	0	2012-08-02 05:44:42.524501	2012-08-02 10:16:57.420724
2629	1-2629	82	23	8261	53	5903	0	0	2012-08-02 05:00:04.160123	2012-08-02 10:16:57.420724
9129	1-9129	51	19	3100	87	6372	0	3	2012-08-02 05:44:00.755512	2012-08-02 10:16:57.420724
8901	1-8901	28	46	3676	58	5117	0	6	2012-08-02 05:42:27.1465	2012-08-02 10:16:58.663543
8761	1-8761	86	8	6869	69	5386	0	0	2012-08-02 05:41:30.441937	2012-08-02 10:16:59.437265
6670	1-6670	63	8	6858	51	8177	0	0	2012-08-02 05:27:18.624197	2012-08-02 10:16:59.437265
10020	1-10020	98	84	2959	8	9597	0	0	2012-08-02 10:16:59.437265	2012-08-02 10:16:59.437265
9261	1-9261	68	41	4180	39	5443	0	6	2012-08-02 05:44:54.358138	2012-08-02 10:19:30.60809
8783	1-8783	83	47	4911	77	4410	0	0	2012-08-02 05:41:39.402344	2012-08-02 10:19:30.60809
9533	1-9533	52	47	4855	41	9130	0	0	2012-08-02 05:46:44.900839	2012-08-02 10:19:30.60809
8140	1-8140	48	6	6576	64	4502	0	0	2012-08-02 05:37:19.58331	2012-08-02 10:19:48.585218
10033	1-10033	28	100	857	43	3298	0	0	2012-08-02 10:19:49.445183	2012-08-02 10:19:49.445183
7377	1-7377	50	77	8766	22	6466	0	5	2012-08-02 05:32:03.576115	2012-08-02 10:19:50.650094
7033	1-7033	18	77	7101	55	3539	0	0	2012-08-02 05:29:44.950151	2012-08-02 10:19:50.650094
8814	1-8814	51	13	5667	99	3058	0	14	2012-08-02 05:41:51.888302	2012-08-02 10:19:51.063009
10037	1-10037	12	68	1317	13	7276	0	14	2012-08-02 10:19:51.063009	2012-08-02 10:19:51.063009
5165	1-5165	4	90	9082	31	6276	0	0	2012-08-02 05:17:09.326399	2012-08-02 10:19:51.473529
10038	1-10038	51	80	1377	90	6339	0	0	2012-08-02 10:19:51.473529	2012-08-02 10:19:51.473529
6726	1-6726	75	15	7639	16	3569	0	0	2012-08-02 05:27:41.009964	2012-08-02 10:19:51.883805
10039	1-10039	65	31	215	12	4544	0	3	2012-08-02 10:19:51.883805	2012-08-02 10:19:51.883805
6419	1-6419	77	20	9785	46	5673	0	0	2012-08-02 05:25:36.574628	2012-08-02 10:20:57.02679
4799	1-4799	9	30	8642	28	6063	0	0	2012-08-02 05:14:42.843272	2012-08-02 10:20:57.02679
9014	1-9014	48	37	8279	76	3754	0	0	2012-08-02 05:43:13.386251	2012-08-02 10:20:57.02679
10042	1-10042	22	28	834	1	5958	0	4	2012-08-02 10:20:57.02679	2012-08-02 10:20:57.02679
10013	1-10013	88	53	5082	36	5563	0	0	2012-08-02 10:16:56.587857	2012-08-02 10:21:01.968377
9951	1-9951	35	67	1776	93	5447	0	4	2012-08-02 05:49:35.592647	2012-08-02 10:21:02.36707
3587	1-3587	21	77	9722	37	4828	0	0	2012-08-02 05:06:36.821959	2012-08-02 10:21:02.36707
10028	1-10028	69	79	1896	22	2807	0	0	2012-08-02 10:19:33.363565	2012-08-02 10:21:02.36707
9792	1-9792	71	41	5531	22	8161	0	0	2012-08-02 05:48:30.847142	2012-08-02 10:21:02.36707
10055	1-10055	40	22	3132	67	5411	0	0	2012-08-02 10:21:02.36707	2012-08-02 10:21:02.36707
9741	1-9741	32	70	3759	40	8407	0	3	2012-08-02 05:48:10.172632	2012-08-02 10:21:02.767505
10058	1-10058	28	52	927	84	2832	0	0	2012-08-02 10:21:03.541599	2012-08-02 10:21:03.541599
9810	1-9810	46	18	3955	39	6485	0	0	2012-08-02 05:48:38.095017	2012-08-02 10:21:05.486703
10063	1-10063	49	31	382	18	8900	0	6	2012-08-02 10:21:05.486703	2012-08-02 10:21:05.486703
9845	1-9845	36	7	3716	83	2891	0	0	2012-08-02 05:48:52.362998	2012-08-02 10:21:06.624907
10066	1-10066	43	83	1717	7	6883	0	7	2012-08-02 10:21:06.624907	2012-08-02 10:21:06.624907
9863	1-9863	30	18	4334	28	5843	0	0	2012-08-02 05:48:59.679645	2012-08-02 10:21:08.183235
9912	1-9912	29	82	2346	77	6582	0	3	2012-08-02 05:49:19.635148	2012-08-02 10:21:08.570323
9933	1-9933	41	53	4837	88	8387	0	8	2012-08-02 05:49:28.255409	2012-08-02 10:21:08.570323
10071	1-10071	19	40	2991	82	8828	0	9	2012-08-02 10:21:08.570323	2012-08-02 10:21:08.570323
9484	1-9484	78	91	2461	71	2317	0	7	2012-08-02 05:46:25.022385	2012-08-02 10:21:08.960071
10072	1-10072	51	71	2724	91	3316	0	0	2012-08-02 10:21:08.960071	2012-08-02 10:21:08.960071
10076	1-10076	2	54	2661	76	3321	0	4	2012-08-02 10:21:10.540401	2012-08-02 10:21:10.540401
7747	1-7747	49	15	4490	2	9777	0	4	2012-08-02 05:34:36.49725	2012-08-02 10:21:11.711576
9278	1-9278	35	25	3225	2	5771	0	2	2012-08-02 05:45:01.33134	2012-08-02 10:21:11.711576
8266	1-8266	74	15	2105	21	779	0	0	2012-08-02 05:38:10.901615	2012-08-02 10:21:11.711576
10079	1-10079	67	2	8609	15	6528	0	3	2012-08-02 10:21:11.711576	2012-08-02 10:21:11.711576
7486	1-7486	49	74	6087	51	6714	0	3	2012-08-02 05:32:48.35685	2012-08-02 10:21:13.248202
5693	1-5693	3	74	7022	51	5942	0	0	2012-08-02 05:20:38.751366	2012-08-02 10:21:13.248202
9795	1-9795	38	49	3867	27	5483	0	0	2012-08-02 05:48:32.051273	2012-08-02 10:21:13.248202
9749	1-9749	96	26	1998	74	5887	0	6	2012-08-02 05:48:13.398628	2012-08-02 10:21:13.248202
9300	1-9300	42	26	3607	49	7567	0	5	2012-08-02 05:45:10.238287	2012-08-02 10:21:13.248202
8663	1-8663	17	89	8736	24	5348	0	0	2012-08-02 05:40:50.714543	2012-08-02 10:21:14.409656
9746	1-9746	85	46	6817	99	3500	0	0	2012-08-02 05:48:12.183324	2012-08-02 10:21:14.409656
10086	1-10086	39	26	667	89	6298	0	0	2012-08-02 10:21:14.409656	2012-08-02 10:21:14.409656
10015	1-10015	5	87	6233	96	7466	0	3	2012-08-02 10:16:57.420724	2012-08-02 10:21:15.194431
7464	1-7464	97	52	7931	98	4604	0	10	2012-08-02 05:32:39.341423	2012-08-02 10:21:16.355077
9825	1-9825	31	81	2475	33	1002	0	0	2012-08-02 05:48:44.18279	2012-08-02 10:21:16.355077
9983	1-9983	87	51	5457	46	9145	0	0	2012-08-02 05:49:48.787994	2012-08-02 10:21:17.538227
8843	1-8843	89	32	9172	28	7425	0	6	2012-08-02 05:42:03.577844	2012-08-02 10:21:17.538227
10098	1-10098	12	94	1767	53	8149	0	7	2012-08-02 10:21:19.107086	2012-08-02 10:21:19.107086
8295	1-8295	17	43	4068	4	4260	0	3	2012-08-02 05:38:22.669416	2012-08-02 10:21:19.86925
10101	1-10101	78	69	173	86	1823	0	5	2012-08-02 10:21:20.423155	2012-08-02 10:21:20.423155
9576	1-9576	67	42	8745	25	5584	0	0	2012-08-02 05:47:02.324687	2012-08-02 10:21:23.162067
9324	1-9324	64	45	4605	86	8017	0	0	2012-08-02 05:45:20.029574	2012-08-02 10:21:23.162067
7334	1-7334	92	47	6532	81	7713	0	8	2012-08-02 05:31:46.592644	2012-08-02 10:21:24.357937
5088	1-5088	4	47	8549	65	8103	0	0	2012-08-02 05:16:38.08895	2012-08-02 10:21:24.357937
9520	1-9520	98	73	5800	1	6704	0	7	2012-08-02 05:46:39.608678	2012-08-02 10:21:24.357937
4524	1-4524	94	47	7359	73	5743	0	0	2012-08-02 05:12:51.22786	2012-08-02 10:21:24.357937
10111	1-10111	7	58	4462	47	9356	0	9	2012-08-02 10:21:24.357937	2012-08-02 10:21:24.357937
7200	1-7200	8	50	2820	54	1560	0	0	2012-08-02 05:30:53.422604	2012-08-02 10:21:24.756573
9729	1-9729	68	62	3691	49	3927	0	0	2012-08-02 05:48:05.342914	2012-08-02 10:21:24.756573
8767	1-8767	39	50	3459	54	1610	0	2	2012-08-02 05:41:32.939037	2012-08-02 10:21:24.756573
10112	1-10112	50	54	4026	62	9538	0	0	2012-08-02 10:21:24.756573	2012-08-02 10:21:24.756573
10091	1-10091	18	62	2616	81	8171	0	7	2012-08-02 10:21:16.355077	2012-08-02 10:21:25.165939
7788	1-7788	34	35	4533	59	3304	0	0	2012-08-02 05:34:53.845568	2012-08-02 10:21:25.165939
8067	1-8067	16	59	9573	32	5146	0	3	2012-08-02 05:36:50.02674	2012-08-02 10:21:25.165939
9236	1-9236	52	27	5114	69	3411	0	0	2012-08-02 05:44:44.126351	2012-08-02 10:21:25.563768
4502	1-4502	66	64	7557	88	5876	0	0	2012-08-02 05:12:42.476897	2012-08-02 10:21:25.563768
10114	1-10114	18	84	1285	64	6829	0	13	2012-08-02 10:21:25.563768	2012-08-02 10:21:25.563768
9127	1-9127	66	43	7191	6	7382	0	2	2012-08-02 05:43:59.914139	2012-08-02 10:21:25.962018
6456	1-6456	9	43	6018	85	4707	0	0	2012-08-02 05:25:51.523144	2012-08-02 10:21:25.962018
9385	1-9385	17	88	3371	40	7522	0	0	2012-08-02 05:45:44.867095	2012-08-02 10:21:25.962018
9293	1-9293	80	43	3952	87	4171	0	4	2012-08-02 05:45:07.432153	2012-08-02 10:21:25.962018
10115	1-10115	47	40	6197	43	9064	0	6	2012-08-02 10:21:25.962018	2012-08-02 10:21:25.962018
9940	1-9940	21	86	5604	36	1834	0	4	2012-08-02 05:49:31.117226	2012-08-02 10:21:26.747339
10073	1-10073	3	54	2740	96	1164	0	11	2012-08-02 10:21:09.358116	2012-08-02 10:21:26.747339
10024	1-10024	14	54	3275	86	7197	0	6	2012-08-02 10:19:31.827383	2012-08-02 10:21:26.747339
10117	1-10117	100	57	2889	54	9412	0	6	2012-08-02 10:21:26.747339	2012-08-02 10:21:26.747339
9560	1-9560	95	49	9590	80	7489	0	0	2012-08-02 05:46:55.861476	2012-08-02 10:21:27.919177
10120	1-10120	55	47	2579	49	6064	0	0	2012-08-02 10:21:27.919177	2012-08-02 10:21:27.919177
10100	1-10100	25	98	3173	43	9788	0	0	2012-08-02 10:21:19.86925	2012-08-02 10:21:28.69255
5225	1-5225	61	6	5814	59	3560	0	0	2012-08-02 05:17:32.818868	2012-08-02 10:21:28.69255
10031	1-10031	34	98	2618	6	6160	0	3	2012-08-02 10:19:48.585218	2012-08-02 10:21:28.69255
9224	1-9224	95	62	9931	94	7542	0	0	2012-08-02 05:44:39.262982	2012-08-02 10:21:28.69255
10122	1-10122	50	59	2844	98	7356	0	5	2012-08-02 10:21:28.69255	2012-08-02 10:21:28.69255
9536	1-9536	23	4	5033	94	7889	0	0	2012-08-02 05:46:46.193565	2012-08-02 10:21:29.112493
10123	1-10123	92	94	2985	88	9226	0	5	2012-08-02 10:21:29.112493	2012-08-02 10:21:29.112493
9326	1-9326	98	14	4874	58	3919	0	0	2012-08-02 05:45:20.847223	2012-08-02 10:21:29.49951
8537	1-8537	55	9	4598	13	4574	0	3	2012-08-02 05:39:59.940731	2012-08-02 10:21:29.49951
9882	1-9882	94	25	1172	41	444	0	4	2012-08-02 05:49:07.426648	2012-08-02 10:21:33.092667
6031	1-6031	86	25	7398	93	5165	0	0	2012-08-02 05:22:57.376076	2012-08-02 10:21:33.092667
9688	1-9688	65	65	2621	58	3992	0	0	2012-08-02 05:47:48.413632	2012-08-02 10:21:33.092667
8019	1-8019	99	47	2163	64	1628	0	0	2012-08-02 05:36:30.7457	2012-08-02 10:21:34.663624
9784	1-9784	73	36	3799	25	8445	0	0	2012-08-02 05:48:27.543163	2012-08-02 10:21:34.663624
10054	1-10054	69	57	1555	70	5439	0	4	2012-08-02 10:21:01.968377	2012-08-02 10:21:34.663624
8383	1-8383	82	12	8638	4	5957	0	0	2012-08-02 05:38:58.121126	2012-08-02 10:21:35.459608
9337	1-9337	82	97	3534	87	8537	0	4	2012-08-02 05:45:25.233385	2012-08-02 10:22:00.297096
10143	1-10143	90	46	405	71	3569	0	4	2012-08-02 10:22:01.175856	2012-08-02 10:22:01.175856
10145	1-10145	48	77	575	26	4591	0	10	2012-08-02 10:22:01.959208	2012-08-02 10:22:01.959208
7354	1-7354	46	20	8299	64	2908	0	0	2012-08-02 05:31:54.516554	2012-08-02 10:22:03.89283
9892	1-9892	38	52	1795	48	7697	0	6	2012-08-02 05:49:11.564364	2012-08-02 10:22:03.89283
4257	1-4257	67	35	5984	37	6775	0	0	2012-08-02 05:11:03.549496	2012-08-02 10:22:04.301357
10045	1-10045	24	84	6286	6	6416	0	2	2012-08-02 10:20:58.301987	2012-08-02 10:22:05.075001
10154	1-10154	91	40	1029	47	2426	0	0	2012-08-02 10:22:05.483862	2012-08-02 10:22:05.483862
9781	1-9781	42	21	855	2	3631	0	8	2012-08-02 05:48:26.327982	2012-08-02 10:22:05.870393
10155	1-10155	38	2	3979	6	1893	0	0	2012-08-02 10:22:05.870393	2012-08-02 10:22:05.870393
10133	1-10133	12	58	2396	25	6529	0	7	2012-08-02 10:21:33.092667	2012-08-02 10:22:06.63264
2761	1-2761	97	40	5005	14	1478	0	0	2012-08-02 05:01:00.06992	2012-08-02 10:22:07.804143
10160	1-10160	72	72	3218	32	6101	0	0	2012-08-02 10:22:07.804143	2012-08-02 10:22:07.804143
10163	1-10163	67	53	176	25	9695	0	8	2012-08-02 10:22:08.964194	2012-08-02 10:22:08.964194
10164	1-10164	77	77	2431	14	4170	0	4	2012-08-02 10:22:09.351004	2012-08-02 10:22:09.351004
6938	1-6938	90	61	3138	88	2538	0	9	2012-08-02 05:29:07.134707	2012-08-02 10:22:09.737764
10165	1-10165	75	88	1021	61	8916	0	3	2012-08-02 10:22:09.737764	2012-08-02 10:22:09.737764
7756	1-7756	57	89	9097	71	7063	0	3	2012-08-02 05:34:40.299307	2012-08-02 10:22:10.147072
10166	1-10166	29	98	275	89	2929	0	0	2012-08-02 10:22:10.147072	2012-08-02 10:22:10.147072
10012	1-10012	23	32	2028	90	3545	0	0	2012-08-02 10:16:56.176296	2012-08-02 10:22:11.737977
8370	1-8370	32	82	9929	51	7164	0	0	2012-08-02 05:38:52.854657	2012-08-02 10:22:11.737977
9393	1-9393	99	73	4605	98	4187	0	8	2012-08-02 05:45:48.051147	2012-08-02 10:22:11.737977
7735	1-7735	7	4	6962	59	4678	0	2	2012-08-02 05:34:31.45906	2012-08-02 10:22:11.737977
9793	1-9793	4	90	8984	62	9615	1408	5	2012-08-02 05:48:31.244712	2012-08-02 10:22:11.737977
10174	1-10174	78	12	3579	77	4007	0	13	2012-08-02 10:22:13.306613	2012-08-02 10:22:13.306613
7797	1-7797	38	5	9909	14	7481	0	0	2012-08-02 05:34:57.646058	2012-08-02 10:22:14.080368
9911	1-9911	100	69	4904	84	6033	0	0	2012-08-02 05:49:19.24655	2012-08-02 10:22:14.080368
10176	1-10176	61	72	3053	23	9994	0	5	2012-08-02 10:22:14.080368	2012-08-02 10:22:14.080368
9935	1-9935	8	75	4153	84	8952	0	4	2012-08-02 05:49:29.062286	2012-08-02 10:22:14.87607
10178	1-10178	14	2	995	60	9320	0	0	2012-08-02 10:22:14.87607	2012-08-02 10:22:14.87607
10189	1-10189	81	25	69	5	5932	0	0	2012-08-02 10:22:19.140153	2012-08-02 10:22:19.140153
10190	1-10190	56	56	400	92	4931	0	4	2012-08-02 10:22:19.526977	2012-08-02 10:22:19.526977
10195	1-10195	28	36	1533	64	7185	0	5	2012-08-02 10:22:21.449132	2012-08-02 10:22:21.449132
7710	1-7710	81	91	9489	50	6147	0	0	2012-08-02 05:34:20.707502	2012-08-02 10:22:21.835923
10196	1-10196	86	9	1092	73	7261	0	3	2012-08-02 10:22:21.835923	2012-08-02 10:22:21.835923
10153	1-10153	8	60	6103	75	5109	0	7	2012-08-02 10:22:05.075001	2012-08-02 10:22:22.2226
9849	1-9849	36	100	3835	38	4093	0	5	2012-08-02 05:48:53.97753	2012-08-02 10:22:22.2226
10200	1-10200	37	78	131	12	4844	0	0	2012-08-02 10:22:23.382628	2012-08-02 10:22:23.382628
9166	1-9166	96	30	4170	4	5571	0	0	2012-08-02 05:44:15.750607	2012-08-02 10:22:25.339901
10023	1-10023	46	98	7089	62	9106	0	0	2012-08-02 10:19:31.429971	2012-08-02 10:22:25.339901
10205	1-10205	38	4	3567	74	3300	0	6	2012-08-02 10:22:25.339901	2012-08-02 10:22:25.339901
8865	1-8865	59	32	1975	40	9553	0	8	2012-08-02 05:42:12.51706	2012-08-02 10:22:26.533431
9496	1-9496	48	86	6585	66	9697	0	0	2012-08-02 05:46:29.884075	2012-08-02 10:22:28.842044
10214	1-10214	25	7	1706	86	5519	0	0	2012-08-02 10:22:28.842044	2012-08-02 10:22:28.842044
9812	1-9812	79	15	8240	35	8917	0	0	2012-08-02 05:48:38.923915	2012-08-02 10:22:29.637928
10216	1-10216	94	38	1958	22	5471	0	4	2012-08-02 10:22:29.637928	2012-08-02 10:22:29.637928
9347	1-9347	30	12	3649	20	9560	0	0	2012-08-02 05:45:29.41004	2012-08-02 10:22:30.841996
10219	1-10219	9	41	40	12	1897	0	0	2012-08-02 10:22:30.841996	2012-08-02 10:22:30.841996
5690	1-5690	37	64	9115	70	7747	0	0	2012-08-02 05:20:37.569038	2012-08-02 10:22:32.799622
10224	1-10224	58	70	2774	64	9840	0	3	2012-08-02 10:22:32.799622	2012-08-02 10:22:32.799622
9887	1-9887	98	88	9461	36	7784	0	3	2012-08-02 05:49:09.474205	2012-08-02 10:22:33.583952
9240	1-9240	37	88	5617	71	6101	0	6	2012-08-02 05:44:45.761268	2012-08-02 10:22:33.583952
10226	1-10226	28	77	3383	88	7852	0	15	2012-08-02 10:22:33.583952	2012-08-02 10:22:33.583952
6167	1-6167	29	52	3497	2	5960	0	7	2012-08-02 05:23:52.148957	2012-08-02 10:22:33.970848
7716	1-7716	78	19	9092	68	9210	0	6	2012-08-02 05:34:23.215966	2012-08-02 10:22:34.346475
295	1-295	22	19	9470	68	4957	0	0	2012-08-02 04:43:58.80027	2012-08-02 10:22:34.346475
10232	1-10232	18	40	668	29	1441	0	0	2012-08-02 10:22:35.937344	2012-08-02 10:22:35.937344
4919	1-4919	77	47	5065	84	4681	0	0	2012-08-02 05:15:30.564897	2012-08-02 10:22:36.700018
8025	1-8025	46	52	6797	45	3377	0	0	2012-08-02 05:36:33.189453	2012-08-02 10:22:36.700018
10234	1-10234	10	66	2202	47	5654	0	4	2012-08-02 10:22:36.700018	2012-08-02 10:22:36.700018
10236	1-10236	88	25	3040	31	6558	0	15	2012-08-02 10:22:37.495499	2012-08-02 10:22:37.495499
10237	1-10237	6	23	525	44	6319	0	0	2012-08-02 10:22:37.881983	2012-08-02 10:22:37.881983
8603	1-8603	22	75	6647	78	4945	0	0	2012-08-02 05:40:26.604259	2012-08-02 10:22:38.666503
9469	1-9469	84	18	5957	10	8399	0	7	2012-08-02 05:46:18.967887	2012-08-02 10:25:16.187794
8853	1-8853	14	14	5368	2	6411	0	10	2012-08-02 05:42:07.622805	2012-08-02 10:25:16.187794
9827	1-9827	3	100	6832	60	5960	0	8	2012-08-02 05:48:45.013475	2012-08-02 10:25:17.385257
8604	1-8604	64	100	6736	45	3074	0	15	2012-08-02 05:40:27.001819	2012-08-02 10:25:17.385257
10249	1-10249	32	95	941	100	7396	0	10	2012-08-02 10:25:17.385257	2012-08-02 10:25:17.385257
5188	1-5188	47	35	6366	92	5666	0	0	2012-08-02 05:17:18.331205	2012-08-02 10:25:18.173078
9564	1-9564	10	27	9568	99	4579	0	6	2012-08-02 05:46:57.485381	2012-08-02 10:25:18.173078
10243	1-10243	5	63	116	12	1265	0	5	2012-08-02 10:25:15.016497	2012-08-02 10:25:18.572659
9431	1-9431	77	63	2077	54	6994	0	0	2012-08-02 05:46:03.386476	2012-08-02 10:25:18.960819
8165	1-8165	82	55	8761	16	8583	0	5	2012-08-02 05:37:29.825212	2012-08-02 10:25:18.960819
9921	1-9921	86	16	5953	20	4742	0	5	2012-08-02 05:49:23.304297	2012-08-02 10:25:18.960819
10253	1-10253	94	20	3721	63	8938	0	0	2012-08-02 10:25:18.960819	2012-08-02 10:25:18.960819
10254	1-10254	17	69	382	6	4675	0	7	2012-08-02 10:25:19.375065	2012-08-02 10:25:19.375065
10256	1-10256	53	5	1646	10	3535	0	0	2012-08-02 10:25:20.13785	2012-08-02 10:25:20.13785
10080	1-10080	43	34	2423	41	2072	0	7	2012-08-02 10:21:12.088218	2012-08-02 10:25:21.309698
10259	1-10259	95	41	2612	36	6461	0	10	2012-08-02 10:25:21.309698	2012-08-02 10:25:21.309698
9242	1-9242	82	24	3218	98	7644	0	0	2012-08-02 05:44:46.545625	2012-08-02 10:25:22.106318
10094	1-10094	90	24	3289	19	9273	0	0	2012-08-02 10:21:17.538227	2012-08-02 10:25:22.106318
10251	1-10251	18	26	2380	27	5172	0	0	2012-08-02 10:25:18.173078	2012-08-02 10:25:24.06309
9602	1-9602	88	12	5948	15	5145	0	8	2012-08-02 05:47:12.989761	2012-08-02 10:25:24.858696
10193	1-10193	81	22	4682	15	6411	0	10	2012-08-02 10:22:20.686845	2012-08-02 10:25:24.858696
8174	1-8174	5	33	9216	74	7079	0	0	2012-08-02 05:37:33.484472	2012-08-02 10:25:24.858696
10268	1-10268	32	15	8175	33	9397	0	6	2012-08-02 10:25:24.858696	2012-08-02 10:25:24.858696
8948	1-8948	30	22	5798	59	5493	0	8	2012-08-02 05:42:46.28142	2012-08-02 10:25:26.053458
8053	1-8053	66	41	6314	10	6261	0	5	2012-08-02 05:36:44.425646	2012-08-02 10:25:26.053458
10208	1-10208	4	53	2944	32	9792	0	0	2012-08-02 10:22:26.533431	2012-08-02 10:25:26.053458
10271	1-10271	52	59	2503	41	7680	0	3	2012-08-02 10:25:26.053458	2012-08-02 10:25:26.053458
9779	1-9779	94	87	3011	37	4276	0	2	2012-08-02 05:48:25.488059	2012-08-02 10:25:26.466128
10246	1-10246	43	10	6127	18	4828	0	0	2012-08-02 10:25:16.187794	2012-08-02 10:25:27.22918
10021	1-10021	89	14	2278	47	1357	0	4	2012-08-02 10:19:30.60809	2012-08-02 10:25:27.22918
7857	1-7857	22	62	8431	36	5663	0	11	2012-08-02 05:35:23.181614	2012-08-02 10:25:27.22918
10274	1-10274	81	47	734	65	5931	0	8	2012-08-02 10:25:27.22918	2012-08-02 10:25:27.22918
10113	1-10113	89	90	2579	62	6497	0	11	2012-08-02 10:21:25.165939	2012-08-02 10:25:27.628097
9826	1-9826	16	91	6836	36	6956	0	0	2012-08-02 05:48:44.604706	2012-08-02 10:25:28.799998
9819	1-9819	53	37	7257	27	4985	0	6	2012-08-02 05:48:41.763592	2012-08-02 10:25:28.799998
10183	1-10183	87	80	851	46	584	0	3	2012-08-02 10:22:16.79807	2012-08-02 10:25:29.198472
6035	1-6035	12	80	8702	28	3061	0	4	2012-08-02 05:22:59.088444	2012-08-02 10:25:29.198472
9871	1-9871	39	39	1884	80	4844	0	0	2012-08-02 05:49:02.955284	2012-08-02 10:25:29.198472
6911	1-6911	57	98	8784	80	8595	0	9	2012-08-02 05:28:56.032489	2012-08-02 10:25:30.382102
6790	1-6790	34	26	7531	80	8894	0	0	2012-08-02 05:28:06.203845	2012-08-02 10:25:30.382102
10213	1-10213	20	99	5197	56	3782	0	6	2012-08-02 10:22:28.455302	2012-08-02 10:25:31.167479
10284	1-10284	1	17	4780	99	8314	0	0	2012-08-02 10:25:31.167479	2012-08-02 10:25:31.167479
9445	1-9445	50	76	7210	28	7464	0	5	2012-08-02 05:46:09.232107	2012-08-02 10:25:31.970379
3321	1-3321	95	89	7710	51	4769	0	0	2012-08-02 05:04:49.153928	2012-08-02 10:25:31.970379
10286	1-10286	62	51	1316	76	7532	0	0	2012-08-02 10:25:31.970379	2012-08-02 10:25:31.970379
10287	1-10287	81	73	736	79	3478	0	0	2012-08-02 10:25:32.368874	2012-08-02 10:25:32.368874
10089	1-10089	68	71	6615	4	6899	0	12	2012-08-02 10:21:15.581787	2012-08-02 10:25:32.766788
10288	1-10288	22	42	4987	71	4718	0	2	2012-08-02 10:25:32.766788	2012-08-02 10:25:32.766788
10051	1-10051	96	80	5982	48	9481	0	5	2012-08-02 10:21:00.776512	2012-08-02 10:25:34.402254
10295	1-10295	22	69	2928	26	6600	0	6	2012-08-02 10:25:35.5733	2012-08-02 10:25:35.5733
9286	1-9286	65	77	6206	42	7099	0	0	2012-08-02 05:45:04.623723	2012-08-02 10:25:35.960041
9159	1-9159	3	51	3578	56	906	0	0	2012-08-02 05:44:12.932818	2012-08-02 10:25:35.960041
10296	1-10296	97	42	1355	51	6446	0	5	2012-08-02 10:25:35.960041	2012-08-02 10:25:35.960041
10299	1-10299	80	43	2523	38	5091	0	10	2012-08-02 10:25:37.087292	2012-08-02 10:25:37.087292
9483	1-9483	71	96	5990	23	8747	0	0	2012-08-02 05:46:24.624645	2012-08-02 10:25:38.667486
9394	1-9394	37	68	7253	78	5977	0	13	2012-08-02 05:45:48.459662	2012-08-02 10:25:38.667486
10305	1-10305	5	16	52	11	4924	0	0	2012-08-02 10:25:39.452582	2012-08-02 10:25:39.452582
5744	1-5744	17	25	8550	14	5840	0	0	2012-08-02 05:20:59.732619	2012-08-02 10:25:39.850209
10307	1-10307	44	58	1203	76	4483	0	8	2012-08-02 10:25:40.236947	2012-08-02 10:25:40.236947
10308	1-10308	53	39	37	18	8794	0	0	2012-08-02 10:25:40.657368	2012-08-02 10:25:40.657368
10070	1-10070	17	28	5170	18	8018	0	2	2012-08-02 10:21:08.183235	2012-08-02 10:25:41.452853
7683	1-7683	60	7	7925	21	5424	0	0	2012-08-02 05:34:09.289061	2012-08-02 10:25:41.452853
8109	1-8109	77	21	2629	36	1847	826	0	2012-08-02 05:37:06.955361	2012-08-02 10:25:41.452853
3418	1-3418	12	18	9502	47	2294	1901	0	2012-08-02 05:05:28.57897	2012-08-02 10:25:41.452853
5043	1-5043	24	89	9030	67	3960	0	0	2012-08-02 05:16:19.769082	2012-08-02 10:25:41.851076
10311	1-10311	93	77	946	61	4509	0	6	2012-08-02 10:25:41.851076	2012-08-02 10:25:41.851076
8134	1-8134	59	76	8721	48	8682	0	0	2012-08-02 05:37:17.164118	2012-08-02 10:25:43.801197
10316	1-10316	60	84	1222	76	4637	0	4	2012-08-02 10:25:43.801197	2012-08-02 10:25:43.801197
9704	1-9704	62	36	3637	10	9211	0	3	2012-08-02 05:47:54.976469	2012-08-02 10:25:44.188443
10000	1-10000	81	51	5240	90	7765	0	0	2012-08-02 05:49:55.826618	2012-08-02 10:25:44.961748
10084	1-10084	30	96	4045	61	6915	0	8	2012-08-02 10:21:13.658575	2012-08-02 10:25:45.746324
10306	1-10306	64	42	2992	87	6719	0	0	2012-08-02 10:25:39.850209	2012-08-02 10:25:46.917434
10326	1-10326	17	54	67	76	5229	0	5	2012-08-02 10:25:47.691152	2012-08-02 10:25:47.691152
10124	1-10124	38	82	1133	14	3669	0	0	2012-08-02 10:21:29.49951	2012-08-02 10:25:48.088715
8978	1-8978	34	63	4051	91	7095	0	9	2012-08-02 05:42:58.635539	2012-08-02 10:25:51.603785
10336	1-10336	12	91	1757	44	5585	0	14	2012-08-02 10:25:51.603785	2012-08-02 10:25:51.603785
10346	1-10346	98	71	5408	50	9835	0	0	2012-08-02 10:27:23.001032	2012-08-02 10:27:23.001032
10319	1-10319	19	90	3897	51	6486	0	0	2012-08-02 10:25:44.961748	2012-08-02 10:27:23.432272
10036	1-10036	80	90	3807	77	8999	0	10	2012-08-02 10:19:50.650094	2012-08-02 10:27:23.432272
8299	1-8299	75	46	7734	1	3388	0	15	2012-08-02 05:38:24.282155	2012-08-02 10:27:23.432272
10347	1-10347	10	95	992	90	3311	0	0	2012-08-02 10:27:23.432272	2012-08-02 10:27:23.432272
9952	1-9952	33	92	7023	89	4807	0	0	2012-08-02 05:49:36.012534	2012-08-02 10:27:25.06777
10351	1-10351	26	82	4231	43	5334	0	0	2012-08-02 10:27:25.06777	2012-08-02 10:27:25.06777
5848	1-5848	44	98	4619	85	9562	0	0	2012-08-02 05:21:41.985785	2012-08-02 10:27:25.896347
3674	1-3674	49	98	4397	54	3089	0	0	2012-08-02 05:07:11.816341	2012-08-02 10:27:25.896347
10353	1-10353	62	77	271	98	6771	0	2	2012-08-02 10:27:25.896347	2012-08-02 10:27:25.896347
10009	1-10009	70	18	6523	47	8604	0	3	2012-08-02 09:10:33.023101	2012-08-02 10:27:26.691551
6982	1-6982	22	36	7619	18	2217	841	9	2012-08-02 05:29:24.684688	2012-08-02 10:27:26.691551
9971	1-9971	58	23	7854	8	7601	0	14	2012-08-02 05:49:43.837396	2012-08-02 10:27:28.657956
9918	1-9918	20	80	1027	46	557	0	3	2012-08-02 05:49:22.077825	2012-08-02 10:27:28.657956
9954	1-9954	36	51	2088	70	4223	0	10	2012-08-02 05:49:36.808271	2012-08-02 10:27:29.066977
10361	1-10361	36	70	1807	51	7148	0	6	2012-08-02 10:27:29.066977	2012-08-02 10:27:29.066977
10110	1-10110	55	56	2694	17	7557	0	0	2012-08-02 10:21:23.960091	2012-08-02 10:27:29.465499
9965	1-9965	89	44	7925	17	6087	0	3	2012-08-02 05:49:41.373615	2012-08-02 10:27:29.465499
9727	1-9727	36	54	4374	94	2552	0	3	2012-08-02 05:48:04.480438	2012-08-02 10:27:29.465499
1454	1-1454	71	35	9160	17	6794	0	0	2012-08-02 04:51:15.316694	2012-08-02 10:27:29.465499
10362	1-10362	22	17	4129	24	7033	0	0	2012-08-02 10:27:29.465499	2012-08-02 10:27:29.465499
8740	1-8740	48	29	6660	58	4071	0	0	2012-08-02 05:41:21.832655	2012-08-02 10:27:30.659208
10365	1-10365	38	20	114	12	5545	0	18	2012-08-02 10:27:30.659208	2012-08-02 10:27:30.659208
7932	1-7932	24	96	4969	38	4174	0	4	2012-08-02 05:35:54.990884	2012-08-02 10:27:31.057118
3657	1-3657	33	43	3442	13	1979	0	0	2012-08-02 05:07:04.99679	2012-08-02 10:27:31.057118
10321	1-10321	77	43	2386	96	5030	0	0	2012-08-02 10:25:45.746324	2012-08-02 10:27:31.057118
9964	1-9964	80	43	4409	16	8950	0	11	2012-08-02 05:49:40.942579	2012-08-02 10:27:31.057118
7754	1-7754	37	6	7084	16	5173	0	9	2012-08-02 05:34:39.448655	2012-08-02 10:27:31.875572
10368	1-10368	60	31	2488	6	8830	0	0	2012-08-02 10:27:31.875572	2012-08-02 10:27:31.875572
10135	1-10135	85	32	6303	36	9637	0	15	2012-08-02 10:21:33.890543	2012-08-02 10:27:32.27355
9177	1-9177	75	78	2349	79	2797	0	0	2012-08-02 05:44:20.172644	2012-08-02 10:27:32.27355
10239	1-10239	17	78	6695	24	6274	0	7	2012-08-02 10:22:38.666503	2012-08-02 10:27:32.27355
10370	1-10370	90	14	134	63	7065	0	0	2012-08-02 10:27:32.682288	2012-08-02 10:27:32.682288
9574	1-9574	68	15	7153	25	8206	0	14	2012-08-02 05:47:01.529133	2012-08-02 10:27:33.466623
10150	1-10150	10	73	2275	12	7494	0	11	2012-08-02 10:22:03.89283	2012-08-02 10:27:35.035427
10372	1-10372	53	48	2656	15	5525	0	12	2012-08-02 10:27:33.466623	2012-08-02 10:27:35.444547
10378	1-10378	86	9	762	13	2233	0	8	2012-08-02 10:27:35.842107	2012-08-02 10:27:35.842107
10170	1-10170	24	62	4534	4	7535	0	5	2012-08-02 10:22:11.737977	2012-08-02 10:27:36.628636
10275	1-10275	67	62	1073	1	2423	0	2	2012-08-02 10:25:27.628097	2012-08-02 10:27:36.628636
8145	1-8145	1	18	6391	33	3272	0	0	2012-08-02 05:37:21.583203	2012-08-02 10:27:36.628636
10380	1-10380	77	4	886	62	1943	0	0	2012-08-02 10:27:36.628636	2012-08-02 10:27:36.628636
9991	1-9991	90	39	4735	31	4170	0	0	2012-08-02 05:49:52.092027	2012-08-02 10:27:37.424233
10382	1-10382	50	85	5340	39	5899	0	20	2012-08-02 10:27:37.424233	2012-08-02 10:27:37.424233
10366	1-10366	65	38	4725	43	6712	0	0	2012-08-02 10:27:31.057118	2012-08-02 10:27:37.822039
4323	1-4323	16	16	6777	56	1259	0	0	2012-08-02 05:11:30.544854	2012-08-02 10:27:42.174516
9283	1-9283	87	39	7231	61	7474	0	8	2012-08-02 05:45:03.364288	2012-08-02 10:27:42.174516
10394	1-10394	79	50	1464	39	6484	0	3	2012-08-02 10:27:42.174516	2012-08-02 10:27:42.174516
9877	1-9877	63	53	7532	1	3980	0	4	2012-08-02 05:49:05.387578	2012-08-02 10:27:42.572796
10395	1-10395	56	7	1029	98	2905	0	0	2012-08-02 10:27:42.572796	2012-08-02 10:27:42.572796
10392	1-10392	88	87	781	16	859	0	10	2012-08-02 10:27:41.390266	2012-08-02 10:27:43.777161
10141	1-10141	15	87	7009	25	8338	0	3	2012-08-02 10:22:00.297096	2012-08-02 10:27:43.777161
10398	1-10398	86	77	984	87	5294	0	7	2012-08-02 10:27:43.777161	2012-08-02 10:27:43.777161
8627	1-8627	67	38	8278	45	4982	0	5	2012-08-02 05:40:36.205503	2012-08-02 10:27:44.572949
7279	1-7279	32	38	6040	45	2828	0	2	2012-08-02 05:31:25.015339	2012-08-02 10:27:44.572949
10400	1-10400	98	4	592	38	8535	0	5	2012-08-02 10:27:44.572949	2012-08-02 10:27:44.572949
10401	1-10401	57	86	288	31	5804	0	0	2012-08-02 10:27:44.981802	2012-08-02 10:27:44.981802
8055	1-8055	83	60	7625	95	5609	0	7	2012-08-02 05:36:45.232371	2012-08-02 10:27:47.699369
10240	1-10240	24	100	3311	42	9222	0	3	2012-08-02 10:22:39.0642	2012-08-02 10:27:47.699369
10276	1-10276	5	58	916	42	3096	0	4	2012-08-02 10:25:28.026671	2012-08-02 10:27:47.699369
10408	1-10408	94	69	3274	41	5461	0	10	2012-08-02 10:27:47.699369	2012-08-02 10:27:47.699369
10374	1-10374	16	37	3919	32	4242	0	11	2012-08-02 10:27:34.251283	2012-08-02 10:27:48.097498
10202	1-10202	4	32	1464	81	1701	0	15	2012-08-02 10:22:24.156031	2012-08-02 10:27:48.484011
10409	1-10409	30	32	4729	31	7499	0	4	2012-08-02 10:27:48.097498	2012-08-02 10:27:48.484011
10410	1-10410	6	16	2124	32	4017	0	9	2012-08-02 10:27:48.484011	2012-08-02 10:27:48.484011
10137	1-10137	85	64	5924	36	8842	0	2	2012-08-02 10:21:34.663624	2012-08-02 10:27:48.8819
8225	1-8225	74	96	8393	19	3927	0	0	2012-08-02 05:37:54.255636	2012-08-02 10:27:49.26852
8643	1-8643	21	86	6750	17	3567	0	0	2012-08-02 05:40:42.659346	2012-08-02 10:27:49.26852
8628	1-8628	54	18	1476	14	1894	0	0	2012-08-02 05:40:36.603173	2012-08-02 10:27:49.677574
10050	1-10050	47	87	1771	51	339	0	0	2012-08-02 10:21:00.37874	2012-08-02 10:27:50.473143
10415	1-10415	56	51	1277	87	8312	0	8	2012-08-02 10:27:50.473143	2012-08-02 10:27:50.473143
9939	1-9939	40	67	627	77	2429	0	0	2012-08-02 05:49:30.708287	2012-08-02 10:27:51.268635
10417	1-10417	81	7	49	67	5539	0	3	2012-08-02 10:27:51.268635	2012-08-02 10:27:51.268635
9311	1-9311	13	83	2878	59	3015	0	0	2012-08-02 05:45:14.6917	2012-08-02 10:27:51.666423
9375	1-9375	27	28	2384	2	6678	0	3	2012-08-02 05:45:40.812374	2012-08-02 10:27:52.064185
10116	1-10116	83	28	2563	37	7601	0	16	2012-08-02 10:21:26.360847	2012-08-02 10:27:52.064185
10419	1-10419	27	2	5557	28	5318	0	0	2012-08-02 10:27:52.064185	2012-08-02 10:27:52.064185
3861	1-3861	6	41	8353	32	3504	0	3	2012-08-02 05:08:26.120644	2012-08-02 10:27:52.837679
10381	1-10381	63	41	3022	60	8769	0	0	2012-08-02 10:27:37.037658	2012-08-02 10:27:55.213041
3963	1-3963	40	41	5128	14	3591	0	2	2012-08-02 05:09:06.899028	2012-08-02 10:27:55.213041
2541	1-2541	87	25	7893	93	4675	0	0	2012-08-02 04:59:26.327408	2012-08-02 10:27:55.213041
10427	1-10427	40	93	1394	41	7424	0	2	2012-08-02 10:27:55.213041	2012-08-02 10:27:55.213041
8768	1-8768	64	64	3772	53	4888	0	0	2012-08-02 05:41:33.34758	2012-08-02 10:27:55.611142
10412	1-10412	44	17	2480	96	7037	0	0	2012-08-02 10:27:49.26852	2012-08-02 10:27:56.384395
10315	1-10315	16	76	5543	61	6784	0	10	2012-08-02 10:25:43.414472	2012-08-02 10:27:57.555811
9881	1-9881	82	4	3839	76	3875	0	0	2012-08-02 05:49:07.006958	2012-08-02 10:27:57.555811
10097	1-10097	62	4	7977	73	7031	0	0	2012-08-02 10:21:18.720299	2012-08-02 10:27:57.555811
7626	1-7626	87	55	8009	74	7296	0	2	2012-08-02 05:33:45.790776	2012-08-02 10:27:57.953321
9804	1-9804	16	37	4815	79	5542	0	0	2012-08-02 05:48:35.697427	2012-08-02 10:27:57.953321
10434	1-10434	34	27	1855	37	9192	0	5	2012-08-02 10:27:57.953321	2012-08-02 10:27:57.953321
10385	1-10385	43	73	2817	9	4286	0	8	2012-08-02 10:27:38.606704	2012-08-02 10:27:59.544539
10438	1-10438	24	42	1483	73	7535	0	0	2012-08-02 10:27:59.544539	2012-08-02 10:27:59.544539
9655	1-9655	89	82	4177	95	5144	0	0	2012-08-02 05:47:34.768383	2012-08-02 10:28:00.328852
6976	1-6976	66	82	6164	92	6642	0	0	2012-08-02 05:29:22.287007	2012-08-02 10:28:00.328852
9896	1-9896	14	82	4115	7	7064	0	4	2012-08-02 05:49:13.18945	2012-08-02 10:28:00.328852
10440	1-10440	76	55	2197	82	8432	0	18	2012-08-02 10:28:00.328852	2012-08-02 10:28:00.328852
8852	1-8852	29	17	5610	77	5376	0	10	2012-08-02 05:42:07.202739	2012-08-02 10:31:34.56874
9857	1-9857	79	46	7343	6	7923	0	13	2012-08-02 05:48:57.245279	2012-08-02 10:31:34.56874
6710	1-6710	62	46	5476	85	5180	0	0	2012-08-02 05:27:34.656195	2012-08-02 10:31:34.56874
9966	1-9966	99	77	9538	91	4833	0	0	2012-08-02 05:49:41.75982	2012-08-02 10:31:37.047096
10283	1-10283	94	23	4997	77	8304	0	4	2012-08-02 10:25:30.76946	2012-08-02 10:31:37.047096
10450	1-10450	77	91	2864	23	9030	0	0	2012-08-02 10:31:37.047096	2012-08-02 10:31:37.047096
10369	1-10369	95	94	1928	78	7937	0	0	2012-08-02 10:27:32.27355	2012-08-02 10:31:38.254753
10313	1-10313	58	94	254	68	269	0	0	2012-08-02 10:25:42.630086	2012-08-02 10:31:38.254753
5175	1-5175	19	26	7442	74	4467	0	0	2012-08-02 05:17:13.248947	2012-08-02 10:31:39.825454
9878	1-9878	16	88	3636	79	2373	0	8	2012-08-02 05:49:05.800285	2012-08-02 10:31:41.395624
9689	1-9689	51	88	7542	23	5892	0	4	2012-08-02 05:47:48.833482	2012-08-02 10:31:41.395624
9552	1-9552	5	88	8849	35	6071	0	2	2012-08-02 05:46:52.635044	2012-08-02 10:31:41.395624
10461	1-10461	84	22	820	88	8763	0	2	2012-08-02 10:31:41.395624	2012-08-02 10:31:41.395624
8495	1-8495	47	69	4707	30	1552	0	0	2012-08-02 05:39:43.123121	2012-08-02 10:31:41.819701
6916	1-6916	97	69	9408	86	3220	0	0	2012-08-02 05:28:58.080694	2012-08-02 10:31:41.819701
9262	1-9262	41	18	4598	75	1775	0	9	2012-08-02 05:44:54.766495	2012-08-02 10:31:41.819701
10292	1-10292	85	48	5010	16	6474	0	0	2012-08-02 10:25:34.402254	2012-08-02 10:31:42.618897
5292	1-5292	15	26	5751	97	2331	0	5	2012-08-02 05:17:59.159297	2012-08-02 10:31:42.618897
9530	1-9530	86	70	6216	2	6838	0	0	2012-08-02 05:46:43.67451	2012-08-02 10:31:42.618897
10464	1-10464	81	2	4088	48	7837	0	0	2012-08-02 10:31:42.618897	2012-08-02 10:31:42.618897
5157	1-5157	64	48	5369	59	738	738	0	2012-08-02 05:17:06.177224	2012-08-02 10:31:42.618897
10269	1-10269	52	37	3472	95	4587	0	0	2012-08-02 10:25:25.280409	2012-08-02 10:31:43.803666
10474	1-10474	85	11	2477	30	6402	0	3	2012-08-02 10:31:46.621876	2012-08-02 10:31:46.621876
10277	1-10277	68	27	1795	97	2543	0	5	2012-08-02 10:25:28.413183	2012-08-02 10:31:47.40618
9323	1-9323	62	91	3232	28	1739	0	6	2012-08-02 05:45:19.610065	2012-08-02 10:31:48.986021
9934	1-9934	16	95	2944	57	4635	0	0	2012-08-02 05:49:28.664271	2012-08-02 10:31:50.577476
10486	1-10486	69	40	2358	53	5773	0	0	2012-08-02 10:31:51.385102	2012-08-02 10:31:51.385102
10201	1-10201	82	74	6224	11	9508	0	0	2012-08-02 10:22:23.769351	2012-08-02 10:31:53.341238
10491	1-10491	71	29	961	74	1641	0	0	2012-08-02 10:31:53.341238	2012-08-02 10:31:53.341238
8655	1-8655	17	24	4315	27	5657	0	2	2012-08-02 05:40:47.476959	2012-08-02 10:31:53.750492
9920	1-9920	36	79	4861	27	2004	0	3	2012-08-02 05:49:22.906421	2012-08-02 10:31:53.750492
10421	1-10421	46	24	1770	41	6521	0	0	2012-08-02 10:27:52.837679	2012-08-02 10:31:53.750492
10108	1-10108	83	56	3824	75	9504	0	0	2012-08-02 10:21:23.162067	2012-08-02 10:31:53.750492
3158	1-3158	68	24	9531	83	7552	0	5	2012-08-02 05:03:43.416415	2012-08-02 10:31:53.750492
8574	1-8574	50	24	4179	65	5919	0	0	2012-08-02 05:40:14.837721	2012-08-02 10:31:53.750492
10457	1-10457	2	27	1749	48	3735	0	4	2012-08-02 10:31:39.825454	2012-08-02 10:31:53.750492
10495	1-10495	92	41	652	19	6199	0	10	2012-08-02 10:31:54.948025	2012-08-02 10:31:54.948025
8624	1-8624	21	26	2294	71	1517	0	0	2012-08-02 05:40:35.00088	2012-08-02 10:31:56.130676
8651	1-8651	57	28	8717	96	5665	0	0	2012-08-02 05:40:45.874827	2012-08-02 10:32:00.096984
9936	1-9936	69	90	4003	46	9225	0	4	2012-08-02 05:49:29.482193	2012-08-02 10:32:00.096984
10391	1-10391	30	63	1348	91	4843	0	13	2012-08-02 10:27:40.992586	2012-08-02 10:32:00.494995
6372	1-6372	17	81	7904	92	9739	0	7	2012-08-02 05:25:17.580064	2012-08-02 10:32:02.462459
10519	1-10519	74	69	276	80	6912	0	0	2012-08-02 10:32:04.417905	2012-08-02 10:32:04.417905
9743	1-9743	84	48	2434	40	3869	0	4	2012-08-02 05:48:10.978787	2012-08-02 10:32:04.805169
10212	1-10212	59	93	1877	40	2735	0	14	2012-08-02 10:22:28.07982	2012-08-02 10:32:04.805169
10520	1-10520	17	40	2942	3	3533	0	0	2012-08-02 10:32:04.805169	2012-08-02 10:32:04.805169
8985	1-8985	51	3	3930	51	6183	0	0	2012-08-02 05:43:01.519339	2012-08-02 10:32:05.601007
10444	1-10444	11	85	7426	17	8002	0	0	2012-08-02 10:31:34.56874	2012-08-02 10:32:05.998759
9928	1-9928	77	80	2845	57	2310	0	0	2012-08-02 05:49:26.19932	2012-08-02 10:32:05.998759
10524	1-10524	25	47	2275	19	3681	0	0	2012-08-02 10:32:06.408228	2012-08-02 10:32:06.408228
9995	1-9995	63	77	2407	5	7202	0	6	2012-08-02 05:49:53.760415	2012-08-02 10:32:07.181979
9932	1-9932	65	61	6729	89	4745	0	5	2012-08-02 05:49:27.846687	2012-08-02 10:32:07.181979
6833	1-6833	58	60	4900	2	3067	0	11	2012-08-02 05:28:23.829163	2012-08-02 10:32:07.571197
9522	1-9522	4	60	7722	44	6612	0	9	2012-08-02 05:46:40.415414	2012-08-02 10:32:07.571197
10425	1-10425	77	96	2150	87	3134	0	5	2012-08-02 10:27:54.428875	2012-08-02 10:32:10.367016
9960	1-9960	95	78	4253	97	8565	0	0	2012-08-02 05:49:39.31809	2012-08-02 10:32:10.367016
10534	1-10534	28	87	5354	78	6806	0	14	2012-08-02 10:32:10.367016	2012-08-02 10:32:10.367016
9955	1-9955	94	80	8342	23	9073	0	0	2012-08-02 05:49:37.217105	2012-08-02 10:32:11.571656
10523	1-10523	6	99	3311	31	7993	0	8	2012-08-02 10:32:05.998759	2012-08-02 10:32:11.571656
10539	1-10539	30	26	1253	64	4819	0	13	2012-08-02 10:32:12.367358	2012-08-02 10:32:12.367358
10083	1-10083	98	99	6444	26	9225	104	0	2012-08-02 10:21:13.248202	2012-08-02 10:32:12.367358
9370	1-9370	95	3	4499	83	8631	0	0	2012-08-02 05:45:38.768535	2012-08-02 10:35:32.861798
7076	1-7076	76	70	7354	48	3945	0	8	2012-08-02 05:30:02.009683	2012-08-02 10:35:32.861798
10173	1-10173	27	10	5998	55	4089	0	8	2012-08-02 10:22:12.919907	2012-08-02 10:35:33.108873
4463	1-4463	97	54	7731	17	2001	0	6	2012-08-02 05:12:27.064089	2012-08-02 10:35:33.108873
10552	1-10552	31	4	1382	54	5935	0	10	2012-08-02 10:35:33.108873	2012-08-02 10:35:33.108873
8848	1-8848	86	30	2466	92	9060	0	2	2012-08-02 05:42:05.589161	2012-08-02 10:35:33.266173
3386	1-3386	53	25	8919	97	3446	0	0	2012-08-02 05:05:15.729272	2012-08-02 10:35:33.345098
10557	1-10557	79	16	39	23	5490	0	9	2012-08-02 10:35:33.49017	2012-08-02 10:35:33.49017
10563	1-10563	1	16	327	41	2363	0	0	2012-08-02 10:35:33.946754	2012-08-02 10:35:33.946754
8047	1-8047	56	22	5794	92	8792	0	0	2012-08-02 05:36:42.028247	2012-08-02 10:35:34.256855
10572	1-10572	39	44	163	25	1115	0	9	2012-08-02 10:35:34.612016	2012-08-02 10:35:34.612016
7451	1-7451	80	1	8815	94	4112	0	2	2012-08-02 05:32:34.0166	2012-08-02 10:35:34.689888
9889	1-9889	41	78	5756	46	6572	0	0	2012-08-02 05:49:10.302964	2012-08-02 10:35:34.689888
10573	1-10573	56	46	1339	1	7583	0	20	2012-08-02 10:35:34.689888	2012-08-02 10:35:34.689888
8294	1-8294	63	49	4876	77	7878	0	0	2012-08-02 05:38:22.271657	2012-08-02 10:35:34.770229
10574	1-10574	4	63	559	49	6586	0	5	2012-08-02 10:35:34.770229	2012-08-02 10:35:34.770229
6560	1-6560	89	47	8096	100	5152	0	0	2012-08-02 05:26:33.868164	2012-08-02 10:35:35.253698
10584	1-10584	42	19	7	1	4347	0	2	2012-08-02 10:35:35.546002	2012-08-02 10:35:35.546002
9130	1-9130	62	86	8614	75	3260	0	0	2012-08-02 05:44:01.164458	2012-08-02 10:35:35.623817
10265	1-10265	84	48	7613	33	1573	0	0	2012-08-02 10:25:23.675973	2012-08-02 10:35:35.715244
10139	1-10139	95	33	3764	12	8348	0	0	2012-08-02 10:21:35.459608	2012-08-02 10:35:35.715244
10587	1-10587	74	38	985	3	9816	0	9	2012-08-02 10:35:35.793414	2012-08-02 10:35:35.793414
10590	1-10590	25	14	986	42	7274	0	4	2012-08-02 10:35:36.015305	2012-08-02 10:35:36.015305
10591	1-10591	68	16	2242	69	9036	0	0	2012-08-02 10:35:36.093351	2012-08-02 10:35:36.093351
9998	1-9998	76	28	9205	24	7954	0	0	2012-08-02 05:49:54.986724	2012-08-02 10:35:36.668705
10260	1-10260	83	73	8749	25	8279	0	0	2012-08-02 10:25:21.708793	2012-08-02 10:35:36.746639
7945	1-7945	87	19	8959	98	9164	0	18	2012-08-02 05:36:00.503989	2012-08-02 10:35:36.746639
10600	1-10600	2	25	6061	19	9425	0	16	2012-08-02 10:35:36.746639	2012-08-02 10:35:36.746639
10198	1-10198	42	61	6176	35	4993	0	0	2012-08-02 10:22:22.609421	2012-08-02 10:35:37.024396
8054	1-8054	61	21	6851	48	5444	0	8	2012-08-02 05:36:44.834265	2012-08-02 10:35:37.102316
9975	1-9975	50	13	5030	75	5561	0	0	2012-08-02 05:49:45.506	2012-08-02 10:35:37.324285
10090	1-10090	35	31	7013	34	8740	0	0	2012-08-02 10:21:15.968315	2012-08-02 10:35:37.601238
8829	1-8829	86	31	6649	85	9695	0	0	2012-08-02 05:41:57.94309	2012-08-02 10:35:37.601238
10177	1-10177	79	19	1638	86	4494	0	0	2012-08-02 10:22:14.478138	2012-08-02 10:35:37.757001
9358	1-9358	48	59	7590	72	8572	0	5	2012-08-02 05:45:33.929462	2012-08-02 10:35:37.757001
10614	1-10614	46	72	1746	19	2541	0	24	2012-08-02 10:35:37.757001	2012-08-02 10:35:37.757001
10585	1-10585	68	6	469	86	2459	0	6	2012-08-02 10:35:35.623817	2012-08-02 10:35:37.938866
9782	1-9782	81	62	1314	70	3522	0	0	2012-08-02 05:48:26.736738	2012-08-02 10:35:37.938866
10622	1-10622	32	76	481	91	1607	0	9	2012-08-02 10:35:38.381654	2012-08-02 10:35:38.381654
10197	1-10197	29	45	2485	100	3993	0	0	2012-08-02 10:22:22.2226	2012-08-02 10:35:39.223099
10633	1-10633	31	90	1820	45	2684	0	0	2012-08-02 10:35:39.223099	2012-08-02 10:35:39.223099
8231	1-8231	39	93	6963	97	4130	0	6	2012-08-02 05:37:56.675453	2012-08-02 10:35:39.455595
7135	1-7135	83	36	8278	59	4511	0	0	2012-08-02 05:30:26.352578	2012-08-02 10:35:39.677649
1506	1-1506	5	8	5873	33	1952	0	3	2012-08-02 04:51:41.741915	2012-08-02 10:35:39.75508
10446	1-10446	91	8	3850	6	1655	0	11	2012-08-02 10:31:35.42308	2012-08-02 10:35:39.75508
9814	1-9814	98	27	2823	12	3403	0	0	2012-08-02 05:48:39.741457	2012-08-02 10:35:58.701003
9636	1-9636	93	8	6330	97	5485	0	3	2012-08-02 05:47:26.933476	2012-08-02 10:35:58.701003
8088	1-8088	91	77	8085	17	8201	0	2	2012-08-02 05:36:58.445168	2012-08-02 10:35:59.134291
8877	1-8877	82	77	4845	91	1688	0	0	2012-08-02 05:42:17.389287	2012-08-02 10:35:59.134291
10644	1-10644	69	6	2423	77	8941	0	14	2012-08-02 10:35:59.134291	2012-08-02 10:35:59.134291
10645	1-10645	32	83	896	39	2809	0	16	2012-08-02 10:35:59.532936	2012-08-02 10:35:59.532936
10151	1-10151	16	27	7252	35	9952	0	2	2012-08-02 10:22:04.301357	2012-08-02 10:36:00.328563
10477	1-10477	72	16	1155	60	1379	0	0	2012-08-02 10:31:47.803973	2012-08-02 10:36:00.328563
7782	1-7782	28	8	4887	5	7064	0	0	2012-08-02 05:34:51.282113	2012-08-02 10:36:00.328563
10199	1-10199	62	18	5156	16	7256	0	11	2012-08-02 10:22:22.99597	2012-08-02 10:36:00.328563
5594	1-5594	93	70	9899	29	4469	0	0	2012-08-02 05:19:58.704785	2012-08-02 10:36:00.328563
10647	1-10647	89	29	2784	18	9020	0	2	2012-08-02 10:36:00.328563	2012-08-02 10:36:00.328563
3353	1-3353	63	18	4335	16	3714	3714	3	2012-08-02 05:05:02.08044	2012-08-02 10:36:00.328563
2746	1-2746	26	18	9453	16	8201	7122	3	2012-08-02 05:00:53.694747	2012-08-02 10:36:00.328563
10560	1-10560	97	20	2438	90	5496	0	3	2012-08-02 10:35:33.714468	2012-08-02 10:36:00.738538
9718	1-9718	1	56	4365	20	3040	0	0	2012-08-02 05:48:00.73516	2012-08-02 10:36:00.738538
10648	1-10648	8	28	1388	56	8189	0	0	2012-08-02 10:36:00.738538	2012-08-02 10:36:00.738538
10282	1-10282	93	48	9139	26	8695	0	0	2012-08-02 10:25:30.382102	2012-08-02 10:36:01.899372
10509	1-10509	9	48	8489	63	5554	0	0	2012-08-02 10:32:00.494995	2012-08-02 10:36:02.297522
10126	1-10126	37	60	7858	46	3392	0	4	2012-08-02 10:21:30.361544	2012-08-02 10:36:02.297522
10652	1-10652	42	46	1956	48	9867	0	11	2012-08-02 10:36:02.297522	2012-08-02 10:36:02.297522
10533	1-10533	100	15	1458	37	1668	0	7	2012-08-02 10:32:09.935831	2012-08-02 10:36:03.093151
10627	1-10627	35	35	3439	4	5160	0	0	2012-08-02 10:35:38.76933	2012-08-02 10:36:03.480574
10655	1-10655	14	51	3113	35	6652	0	3	2012-08-02 10:36:03.480574	2012-08-02 10:36:03.480574
8540	1-8540	57	79	5943	71	3885	0	0	2012-08-02 05:40:01.144733	2012-08-02 10:36:03.878436
10656	1-10656	84	71	1113	46	7536	0	5	2012-08-02 10:36:03.878436	2012-08-02 10:36:03.878436
10657	1-10657	14	61	104	87	5455	0	0	2012-08-02 10:36:04.276369	2012-08-02 10:36:04.276369
8163	1-8163	42	96	1482	16	1854	0	0	2012-08-02 05:37:29.007818	2012-08-02 10:36:04.663382
9999	1-9999	45	81	4697	16	8435	0	0	2012-08-02 05:49:55.384526	2012-08-02 10:36:05.049976
10659	1-10659	26	15	1277	90	5458	0	12	2012-08-02 10:36:05.049976	2012-08-02 10:36:05.049976
10612	1-10612	12	34	1194	31	4506	0	4	2012-08-02 10:35:37.601238	2012-08-02 10:36:05.437352
10435	1-10435	29	60	6596	34	3926	0	0	2012-08-02 10:27:58.351185	2012-08-02 10:36:05.437352
10660	1-10660	56	52	782	60	7499	0	7	2012-08-02 10:36:05.437352	2012-08-02 10:36:05.437352
10314	1-10314	51	16	7343	84	6483	0	0	2012-08-02 10:25:43.016824	2012-08-02 10:36:06.995377
10310	1-10310	27	47	2962	7	6288	0	0	2012-08-02 10:25:41.452853	2012-08-02 10:36:07.382007
10665	1-10665	77	7	6170	100	4521	0	3	2012-08-02 10:36:07.382007	2012-08-02 10:36:07.382007
10667	1-10667	97	8	3478	54	5367	0	17	2012-08-02 10:36:08.166636	2012-08-02 10:36:08.166636
9527	1-9527	74	36	6074	81	7036	0	0	2012-08-02 05:46:42.448426	2012-08-02 10:36:08.564494
10639	1-10639	64	59	2378	21	7850	0	0	2012-08-02 10:35:39.677649	2012-08-02 10:36:08.564494
185	1-185	69	77	9277	81	785	558	0	2012-08-02 04:43:55.732543	2012-08-02 10:36:08.564494
4678	1-4678	100	42	7133	83	2540	0	0	2012-08-02 05:13:53.899658	2012-08-02 10:36:08.962545
6942	1-6942	48	28	7517	31	6303	0	0	2012-08-02 05:29:08.737027	2012-08-02 10:36:11.326725
10675	1-10675	16	31	3935	99	2819	0	0	2012-08-02 10:36:11.326725	2012-08-02 10:36:11.326725
10227	1-10227	42	2	4788	10	4850	0	6	2012-08-02 10:22:33.970848	2012-08-02 10:36:12.146332
10470	1-10470	55	89	852	40	5417	0	0	2012-08-02 10:31:44.986459	2012-08-02 10:36:12.146332
10399	1-10399	96	69	6083	18	6402	0	0	2012-08-02 10:27:44.175052	2012-08-02 10:36:12.146332
10549	1-10549	97	63	3805	3	9156	0	4	2012-08-02 10:35:32.861798	2012-08-02 10:36:12.544092
8815	1-8815	91	63	3881	51	3434	0	2	2012-08-02 05:41:52.275183	2012-08-02 10:36:12.544092
10678	1-10678	66	64	1383	63	6272	0	0	2012-08-02 10:36:12.544092	2012-08-02 10:36:12.544092
10679	1-10679	71	88	200	79	2434	0	10	2012-08-02 10:36:12.932043	2012-08-02 10:36:12.932043
8237	1-8237	32	50	8119	49	4538	0	0	2012-08-02 05:37:59.097031	2012-08-02 10:36:13.727423
9444	1-9444	37	84	7567	69	5696	0	0	2012-08-02 05:46:08.834316	2012-08-02 10:36:13.727423
9365	1-9365	92	84	8205	77	6697	0	0	2012-08-02 05:45:36.768655	2012-08-02 10:36:13.727423
10681	1-10681	94	69	2873	84	7630	0	6	2012-08-02 10:36:13.727423	2012-08-02 10:36:13.727423
10683	1-10683	91	88	5338	71	8829	0	11	2012-08-02 10:36:14.490202	2012-08-02 10:36:14.490202
2262	1-2262	3	91	2204	49	705	0	0	2012-08-02 04:57:24.431099	2012-08-02 10:36:15.661498
10686	1-10686	65	71	186	91	384	0	3	2012-08-02 10:36:15.661498	2012-08-02 10:36:15.661498
7028	1-7028	91	18	8346	7	663	663	0	2012-08-02 05:29:42.884152	2012-08-02 10:36:15.661498
10492	1-10492	89	48	9537	24	9106	0	0	2012-08-02 10:31:53.750492	2012-08-02 10:36:16.07037
5544	1-5544	69	67	6136	26	1864	0	0	2012-08-02 05:19:38.749012	2012-08-02 10:36:16.07037
10687	1-10687	46	24	2014	67	9588	0	11	2012-08-02 10:36:16.07037	2012-08-02 10:36:16.07037
10	1-10	10	18	5676	10	1911	1692	0	2012-08-02 04:43:53.542641	2012-08-02 10:36:16.07037
10688	1-10688	3	13	571	1	3432	0	8	2012-08-02 10:36:16.479713	2012-08-02 10:36:16.479713
10690	1-10690	98	100	317	67	3566	0	8	2012-08-02 10:36:17.253241	2012-08-02 10:36:17.253241
10266	1-10266	98	76	1991	92	3568	0	6	2012-08-02 10:25:24.06309	2012-08-02 10:36:18.424417
10136	1-10136	97	76	9743	26	7194	0	3	2012-08-02 10:21:34.277001	2012-08-02 10:36:18.424417
7540	1-7540	30	76	4366	92	4716	0	4	2012-08-02 05:33:10.733032	2012-08-02 10:36:18.424417
7043	1-7043	97	76	4010	43	1394	0	0	2012-08-02 05:29:48.883189	2012-08-02 10:36:18.424417
10693	1-10693	93	31	1083	76	8646	0	0	2012-08-02 10:36:18.424417	2012-08-02 10:36:18.424417
10343	1-10343	80	85	2810	2	5371	0	0	2012-08-02 10:27:21.742034	2012-08-02 10:36:19.209058
10677	1-10677	29	40	6519	84	8834	0	0	2012-08-02 10:36:12.146332	2012-08-02 10:36:19.209058
5837	1-5837	63	18	5875	84	2126	0	0	2012-08-02 05:21:37.412304	2012-08-02 10:36:19.607106
8576	1-8576	31	97	4039	78	1616	0	0	2012-08-02 05:40:15.666116	2012-08-02 10:36:19.607106
5806	1-5806	74	11	8850	84	5916	0	17	2012-08-02 05:21:24.891644	2012-08-02 10:36:19.607106
879	1-879	24	23	2301	78	513	0	0	2012-08-02 04:46:06.068154	2012-08-02 10:36:19.607106
10696	1-10696	100	78	2168	1	8610	0	4	2012-08-02 10:36:19.607106	2012-08-02 10:36:19.607106
10697	1-10697	81	8	837	10	4207	0	0	2012-08-02 10:36:20.016257	2012-08-02 10:36:20.016257
10504	1-10504	15	19	1758	10	2935	0	0	2012-08-02 10:31:58.517454	2012-08-02 10:36:20.436271
10252	1-10252	60	25	9776	63	3748	0	0	2012-08-02 10:25:18.572659	2012-08-02 10:36:20.436271
10698	1-10698	95	10	5336	14	5175	0	0	2012-08-02 10:36:20.436271	2012-08-02 10:36:20.436271
8264	1-8264	79	84	9831	45	6299	0	0	2012-08-02 05:38:10.095117	2012-08-02 10:36:20.834054
10217	1-10217	49	6	5185	88	8380	0	0	2012-08-02 10:22:30.068806	2012-08-02 10:36:20.834054
10699	1-10699	13	45	2818	6	8921	0	6	2012-08-02 10:36:20.834054	2012-08-02 10:36:20.834054
10002	1-10002	80	3	6382	49	6751	0	2	2012-08-02 09:10:30.096831	2012-08-02 10:36:21.221225
10654	1-10654	1	89	1457	15	7347	0	0	2012-08-02 10:36:03.093151	2012-08-02 10:36:23.154366
9189	1-9189	87	63	1621	73	2971	0	11	2012-08-02 05:44:25.080292	2012-08-02 10:36:23.154366
10576	1-10576	67	16	1476	15	3317	0	5	2012-08-02 10:35:34.914749	2012-08-02 10:36:23.154366
10527	1-10527	91	10	2555	60	5453	0	4	2012-08-02 10:32:07.571197	2012-08-02 10:36:23.154366
10705	1-10705	69	15	7359	63	6231	0	16	2012-08-02 10:36:23.154366	2012-08-02 10:36:23.154366
10555	1-10555	51	26	2095	25	7504	0	14	2012-08-02 10:35:33.345098	2012-08-02 10:36:23.938959
10707	1-10707	37	96	1875	26	3568	0	6	2012-08-02 10:36:23.938959	2012-08-02 10:36:23.938959
10713	1-10713	53	72	2810	73	4988	0	11	2012-08-02 10:36:26.270339	2012-08-02 10:36:26.270339
10692	1-10692	14	93	3734	11	2741	0	0	2012-08-02 10:36:18.037524	2012-08-02 10:36:27.021387
10337	1-10337	41	85	5113	84	3838	0	0	2012-08-02 10:25:51.990739	2012-08-02 10:36:27.408411
7327	1-7327	71	61	5636	71	6029	0	4	2012-08-02 05:31:43.83061	2012-08-02 10:36:28.170561
7671	1-7671	41	66	7082	71	4916	0	3	2012-08-02 05:34:04.265986	2012-08-02 10:36:28.170561
10718	1-10718	36	71	1826	81	6903	0	0	2012-08-02 10:36:28.170561	2012-08-02 10:36:28.170561
8015	1-8015	86	12	4351	52	2799	0	0	2012-08-02 05:36:29.143703	2012-08-02 10:36:30.106164
9846	1-9846	12	94	2152	17	4629	0	3	2012-08-02 05:48:52.760799	2012-08-02 10:36:30.515001
10722	1-10722	60	34	2594	80	1789	0	0	2012-08-02 10:36:29.719399	2012-08-02 10:36:30.515001
10724	1-10724	90	80	1997	94	9592	0	0	2012-08-02 10:36:30.515001	2012-08-02 10:36:30.515001
10119	1-10119	66	1	1388	64	1277	0	9	2012-08-02 10:21:27.532214	2012-08-02 10:36:30.923975
10462	1-10462	31	49	2687	69	9166	0	3	2012-08-02 10:31:41.819701	2012-08-02 10:36:30.923975
10732	1-10732	7	15	218	86	2022	0	19	2012-08-02 10:36:33.612967	2012-08-02 10:36:33.612967
10735	1-10735	76	10	838	78	1991	0	7	2012-08-02 10:36:34.794975	2012-08-02 10:36:34.794975
10429	1-10429	4	71	957	59	1069	0	4	2012-08-02 10:27:55.997791	2012-08-02 10:36:35.568496
10418	1-10418	4	59	1935	5	6423	0	17	2012-08-02 10:27:51.666423	2012-08-02 10:36:35.568496
10737	1-10737	61	5	4408	75	3285	0	2	2012-08-02 10:36:35.568496	2012-08-02 10:36:35.568496
9919	1-9919	12	11	6021	30	5392	0	0	2012-08-02 05:49:22.497692	2012-08-02 10:36:36.341861
10739	1-10739	62	94	3131	39	9918	0	0	2012-08-02 10:36:36.341861	2012-08-02 10:36:36.341861
5828	1-5828	12	30	9176	94	6130	3515	0	2012-08-02 05:21:33.788303	2012-08-02 10:36:36.341861
10740	1-10740	82	87	2211	98	3023	0	0	2012-08-02 10:36:36.742018	2012-08-02 10:36:36.742018
8811	1-8811	15	29	4730	85	8831	0	15	2012-08-02 05:41:50.682217	2012-08-02 10:37:41.025759
6696	1-6696	53	38	9378	55	3848	0	0	2012-08-02 05:27:29.032009	2012-08-02 10:37:41.855418
10743	1-10743	93	88	353	51	8278	0	8	2012-08-02 10:37:41.855418	2012-08-02 10:37:41.855418
10447	1-10447	67	18	3545	7	9584	0	9	2012-08-02 10:31:35.831847	2012-08-02 10:37:42.253155
9585	1-9585	34	30	3438	75	5913	0	3	2012-08-02 05:47:06.051254	2012-08-02 10:37:42.662035
10453	1-10453	61	30	8181	94	9892	0	7	2012-08-02 10:31:38.254753	2012-08-02 10:37:42.662035
9925	1-9925	18	51	2375	84	1895	0	4	2012-08-02 05:49:24.96187	2012-08-02 10:37:42.662035
10745	1-10745	47	84	1344	30	8811	0	0	2012-08-02 10:37:42.662035	2012-08-02 10:37:42.662035
10604	1-10604	15	66	4249	61	6742	0	0	2012-08-02 10:35:37.024396	2012-08-02 10:37:43.844368
3734	1-3734	81	66	5918	37	3970	0	0	2012-08-02 05:07:35.646875	2012-08-02 10:37:43.844368
7963	1-7963	30	66	7363	58	5321	0	0	2012-08-02 05:36:08.042149	2012-08-02 10:37:43.844368
10748	1-10748	13	75	1387	66	8371	0	3	2012-08-02 10:37:43.844368	2012-08-02 10:37:43.844368
9835	1-9835	50	50	5249	41	7492	1442	0	2012-08-02 05:48:48.2622	2012-08-02 10:37:43.844368
10508	1-10508	23	20	2557	90	5139	0	0	2012-08-02 10:32:00.096984	2012-08-02 10:37:45.081324
10751	1-10751	58	87	1787	96	7105	0	8	2012-08-02 10:37:45.081324	2012-08-02 10:37:45.081324
10752	1-10752	21	85	1473	79	9599	0	0	2012-08-02 10:37:45.468287	2012-08-02 10:37:45.468287
7579	1-7579	54	61	5104	29	5473	0	0	2012-08-02 05:33:26.588223	2012-08-02 10:37:47.13649
8452	1-8452	33	38	5654	41	4451	0	0	2012-08-02 05:39:25.919809	2012-08-02 10:37:48.331666
10567	1-10567	95	92	2234	95	3069	0	0	2012-08-02 10:35:34.256855	2012-08-02 10:37:48.331666
10147	1-10147	39	92	8938	26	6743	0	0	2012-08-02 10:22:02.732402	2012-08-02 10:37:48.331666
9455	1-9455	35	38	7882	22	6669	0	0	2012-08-02 05:46:13.322276	2012-08-02 10:37:48.331666
10760	1-10760	73	69	2094	38	9982	0	0	2012-08-02 10:37:48.331666	2012-08-02 10:37:48.331666
10613	1-10613	9	23	2238	87	3222	0	9	2012-08-02 10:35:37.679361	2012-08-02 10:37:50.751828
9447	1-9447	87	11	5074	23	9384	0	3	2012-08-02 05:46:10.027693	2012-08-02 10:37:50.751828
10359	1-10359	52	98	3094	53	8724	0	5	2012-08-02 10:27:28.260278	2012-08-02 10:37:50.751828
10769	1-10769	35	41	3427	6	4619	0	0	2012-08-02 10:37:51.524753	2012-08-02 10:37:51.524753
8575	1-8575	26	100	9024	69	7027	0	5	2012-08-02 05:40:15.246067	2012-08-02 10:37:51.911514
10640	1-10640	64	32	4623	8	9029	0	0	2012-08-02 10:35:39.75508	2012-08-02 10:37:53.877914
9155	1-9155	42	10	7707	40	8566	0	8	2012-08-02 05:44:11.308635	2012-08-02 10:37:54.286695
10778	1-10778	63	24	236	18	2964	0	9	2012-08-02 10:37:55.10422	2012-08-02 10:37:55.10422
9853	1-9853	99	62	6041	50	7375	0	0	2012-08-02 05:48:55.593233	2012-08-02 10:37:55.502015
7629	1-7629	14	62	7202	12	2805	0	2	2012-08-02 05:33:47.006409	2012-08-02 10:37:55.502015
9902	1-9902	64	62	1658	6	1810	0	0	2012-08-02 05:49:15.621689	2012-08-02 10:37:55.502015
10779	1-10779	40	1	967	62	3175	0	3	2012-08-02 10:37:55.502015	2012-08-02 10:37:55.502015
10481	1-10481	48	87	6785	53	4582	0	0	2012-08-02 10:31:49.384377	2012-08-02 10:37:56.297508
10324	1-10324	21	93	2108	42	5723	0	3	2012-08-02 10:25:46.917434	2012-08-02 10:37:57.855669
10784	1-10784	25	42	4018	93	3569	0	9	2012-08-02 10:37:57.855669	2012-08-02 10:37:57.855669
10476	1-10476	95	78	1451	27	7333	0	5	2012-08-02 10:31:47.40618	2012-08-02 10:37:57.469118
10786	1-10786	84	72	589	70	6682	0	6	2012-08-02 10:37:58.231186	2012-08-02 10:37:58.231186
9962	1-9962	81	86	6370	21	8102	0	0	2012-08-02 05:49:40.124847	2012-08-02 10:37:59.811557
10790	1-10790	80	95	2758	86	5483	0	5	2012-08-02 10:37:59.811557	2012-08-02 10:37:59.811557
8268	1-8268	41	8	3669	16	4393	0	0	2012-08-02 05:38:11.6968	2012-08-02 10:38:02.915583
8968	1-8968	6	21	3698	36	2793	0	5	2012-08-02 05:42:54.54757	2012-08-02 10:38:02.915583
10544	1-10544	6	51	6061	93	9131	0	0	2012-08-02 10:35:32.483148	2012-08-02 10:38:02.915583
10798	1-10798	42	16	2116	51	4174	0	7	2012-08-02 10:38:02.915583	2012-08-02 10:38:02.915583
3238	1-3238	40	16	8842	51	1667	1667	0	2012-08-02 05:04:15.97113	2012-08-02 10:38:02.915583
10607	1-10607	26	86	1221	70	7997	0	12	2012-08-02 10:35:37.246678	2012-08-02 10:38:03.7003
6766	1-6766	41	86	4880	29	3945	0	0	2012-08-02 05:27:56.844772	2012-08-02 10:38:03.7003
10263	1-10263	18	86	3550	35	6482	0	2	2012-08-02 10:25:22.903006	2012-08-02 10:38:03.7003
10367	1-10367	9	19	4478	73	8486	0	0	2012-08-02 10:27:31.4776	2012-08-02 10:38:04.507152
10029	1-10029	23	37	5288	70	9212	0	0	2012-08-02 10:19:33.750417	2012-08-02 10:38:04.507152
10261	1-10261	37	98	9581	24	9137	0	13	2012-08-02 10:25:22.106318	2012-08-02 10:38:04.507152
10317	1-10317	4	10	9756	24	6002	0	0	2012-08-02 10:25:44.188443	2012-08-02 10:38:04.507152
10802	1-10802	94	79	5034	19	9484	0	3	2012-08-02 10:38:04.507152	2012-08-02 10:38:04.507152
10262	1-10262	32	21	6168	95	4731	0	0	2012-08-02 10:25:22.516118	2012-08-02 10:38:06.064779
10458	1-10458	38	61	1239	2	2403	0	0	2012-08-02 10:31:40.213724	2012-08-02 10:38:06.440517
9753	1-9753	100	50	8433	71	7985	0	4	2012-08-02 05:48:15.000731	2012-08-02 10:38:06.440517
10807	1-10807	25	2	3590	50	8791	0	7	2012-08-02 10:38:06.440517	2012-08-02 10:38:06.440517
10169	1-10169	84	79	5430	89	3989	0	0	2012-08-02 10:22:11.35113	2012-08-02 10:38:07.611944
10810	1-10810	99	89	3486	79	8679	0	4	2012-08-02 10:38:07.611944	2012-08-02 10:38:07.611944
9282	1-9282	11	54	9037	53	4669	0	0	2012-08-02 05:45:02.955603	2012-08-02 10:38:07.987362
10811	1-10811	22	11	1511	54	7052	0	0	2012-08-02 10:38:07.987362	2012-08-02 10:38:07.987362
10723	1-10723	71	79	3412	12	8452	0	0	2012-08-02 10:36:30.106164	2012-08-02 10:38:08.771816
10167	1-10167	89	71	3597	12	5139	0	0	2012-08-02 10:22:10.567085	2012-08-02 10:38:08.771816
10813	1-10813	96	12	4047	9	8745	0	0	2012-08-02 10:38:08.771816	2012-08-02 10:38:08.771816
10301	1-10301	83	68	3365	75	1927	0	15	2012-08-02 10:25:37.883035	2012-08-02 10:38:09.170128
10228	1-10228	38	68	5631	28	5244	0	0	2012-08-02 10:22:34.346475	2012-08-02 10:38:09.170128
10814	1-10814	38	86	5535	68	6682	0	0	2012-08-02 10:38:09.170128	2012-08-02 10:38:09.170128
9715	1-9715	65	6	9303	78	5185	0	0	2012-08-02 05:47:59.519802	2012-08-02 10:38:09.567524
10605	1-10605	67	48	6575	86	8569	0	7	2012-08-02 10:35:37.102316	2012-08-02 10:38:10.340891
6425	1-6425	25	43	9967	46	7907	0	0	2012-08-02 05:25:39.171116	2012-08-02 10:38:10.340891
8361	1-8361	29	22	7802	52	5107	0	0	2012-08-02 05:38:49.196647	2012-08-02 10:38:10.340891
10817	1-10817	53	34	1252	22	6614	0	15	2012-08-02 10:38:10.340891	2012-08-02 10:38:10.340891
2737	1-2737	55	48	9114	86	1247	381	0	2012-08-02 05:00:49.938189	2012-08-02 10:38:10.340891
10599	1-10599	10	99	1414	28	9516	0	3	2012-08-02 10:35:36.668705	2012-08-02 10:38:11.523412
10291	1-10291	5	92	7449	43	3985	0	0	2012-08-02 10:25:33.927063	2012-08-02 10:38:11.90992
8937	1-8937	36	39	3425	16	7099	0	0	2012-08-02 05:42:41.807145	2012-08-02 10:38:11.90992
10506	1-10506	37	39	9480	11	8255	0	0	2012-08-02 10:31:59.301389	2012-08-02 10:38:11.90992
9984	1-9984	32	62	5968	35	6353	0	8	2012-08-02 05:49:49.185801	2012-08-02 10:38:12.705559
10583	1-10583	38	46	3599	15	6860	0	4	2012-08-02 10:35:35.479668	2012-08-02 10:38:13.114412
8036	1-8036	64	46	8738	93	5224	0	0	2012-08-02 05:36:37.642064	2012-08-02 10:38:13.114412
10824	1-10824	52	63	617	46	7932	0	5	2012-08-02 10:38:13.114412	2012-08-02 10:38:13.114412
10662	1-10662	98	46	6119	34	7165	0	5	2012-08-02 10:36:06.210893	2012-08-02 10:38:15.88932
10085	1-10085	3	28	7154	33	8253	0	0	2012-08-02 10:21:14.034082	2012-08-02 10:38:16.66256
10833	1-10833	66	4	1314	28	6418	0	10	2012-08-02 10:38:16.66256	2012-08-02 10:38:16.66256
10839	1-10839	3	37	1542	31	6270	0	6	2012-08-02 10:38:19.391454	2012-08-02 10:38:19.391454
9464	1-9464	16	41	852	68	171	0	8	2012-08-02 05:46:16.934858	2012-08-02 10:38:19.005135
10840	1-10840	51	68	17	41	9049	0	0	2012-08-02 10:38:19.005135	2012-08-02 10:38:19.005135
10670	1-10670	50	78	3596	41	8317	0	0	2012-08-02 10:36:09.349328	2012-08-02 10:38:26.871685
10342	1-10342	28	78	3862	73	5791	0	0	2012-08-02 10:27:21.321978	2012-08-02 10:38:26.871685
10586	1-10586	2	12	3262	48	8685	0	4	2012-08-02 10:35:35.715244	2012-08-02 10:38:28.053742
10420	1-10420	19	37	9692	55	6301	0	0	2012-08-02 10:27:52.451	2012-08-02 10:38:29.225009
10842	1-10842	31	37	4899	27	2263	0	10	2012-08-02 10:38:20.165007	2012-08-02 10:38:29.225009
10781	1-10781	94	37	3932	87	6386	0	11	2012-08-02 10:37:56.297508	2012-08-02 10:38:29.225009
10865	1-10865	65	58	3378	37	9804	0	3	2012-08-02 10:38:29.225009	2012-08-02 10:38:29.225009
10883	1-10883	71	24	225	33	4837	0	4	2012-08-02 10:38:36.286525	2012-08-02 10:38:36.286525
10171	1-10171	74	34	4486	72	7859	0	2	2012-08-02 10:22:12.146677	2012-08-02 10:38:38.61951
10889	1-10889	17	51	2266	34	5337	0	5	2012-08-02 10:38:38.61951	2012-08-02 10:38:38.61951
5867	1-5867	70	39	5131	73	4088	0	0	2012-08-02 05:21:49.653015	2012-08-02 10:38:39.801776
10413	1-10413	15	14	4858	25	9361	0	4	2012-08-02 10:27:49.677574	2012-08-02 10:38:40.984077
10601	1-10601	76	14	9244	19	7624	0	7	2012-08-02 10:35:36.825394	2012-08-02 10:38:40.984077
10130	1-10130	27	14	3979	100	3308	0	3	2012-08-02 10:21:31.910672	2012-08-02 10:38:40.984077
4941	1-4941	48	30	9588	57	9027	0	0	2012-08-02 05:15:39.14077	2012-08-02 10:38:50.391199
10859	1-10859	68	79	3322	78	4669	0	0	2012-08-02 10:38:26.871685	2012-08-02 10:38:50.391199
7438	1-7438	58	58	6899	70	5235	0	0	2012-08-02 05:32:28.647244	2012-08-02 10:38:50.391199
8780	1-8780	98	45	4827	80	9479	0	12	2012-08-02 05:41:38.209045	2012-08-02 10:38:51.581956
10233	1-10233	47	45	6389	66	9172	0	0	2012-08-02 10:22:36.324068	2012-08-02 10:38:51.581956
10767	1-10767	24	53	6557	11	6623	0	2	2012-08-02 10:37:50.751828	2012-08-02 10:38:51.581956
10922	1-10922	57	29	3901	45	6135	0	0	2012-08-02 10:38:51.581956	2012-08-02 10:38:51.581956
5619	1-5619	83	22	9754	43	4359	0	6	2012-08-02 05:20:08.828834	2012-08-02 10:38:53.601128
10927	1-10927	47	17	1601	22	5688	0	0	2012-08-02 10:38:53.601128	2012-08-02 10:38:53.601128
9811	1-9811	68	31	9787	97	9244	0	5	2012-08-02 05:48:38.51513	2012-08-02 10:38:54.786186
10618	1-10618	59	51	3219	38	4008	0	7	2012-08-02 10:35:38.094066	2012-08-02 10:38:54.786186
10901	1-10901	22	58	7333	73	3742	0	0	2012-08-02 10:38:42.940084	2012-08-02 10:38:58.75591
10962	1-10962	98	29	148	8	6326	0	0	2012-08-02 10:39:07.479241	2012-08-02 10:39:07.479241
10046	1-10046	100	79	7704	38	7121	0	3	2012-08-02 10:20:58.699636	2012-08-02 10:39:13.34021
10983	1-10983	34	51	1009	31	3843	0	0	2012-08-02 10:39:15.683106	2012-08-02 10:39:15.683106
10102	1-10102	38	95	2697	10	2186	0	5	2012-08-02 10:21:20.820042	2012-08-02 10:39:18.027505
10258	1-10258	37	71	5170	51	6550	0	7	2012-08-02 10:25:20.923103	2012-08-02 10:39:18.027505
10158	1-10158	95	32	1532	10	1710	0	7	2012-08-02 10:22:07.030685	2012-08-02 10:39:18.027505
10990	1-10990	99	51	1008	66	8675	0	7	2012-08-02 10:39:18.027505	2012-08-02 10:39:18.027505
6974	1-6974	42	71	6329	51	48	48	2	2012-08-02 05:29:21.502581	2012-08-02 10:39:18.027505
8172	1-8172	9	71	7565	51	3144	1991	2	2012-08-02 05:37:32.655613	2012-08-02 10:39:18.027505
10597	1-10597	80	1	4757	18	6478	0	0	2012-08-02 10:35:36.513964	2012-08-02 10:39:21.178299
9082	1-9082	91	11	2119	95	1196	0	8	2012-08-02 05:43:41.306554	2012-08-02 10:39:21.951995
10741	1-10741	39	85	2406	4	1783	0	0	2012-08-02 10:37:41.025759	2012-08-02 10:39:21.951995
8249	1-8249	38	66	9426	95	6143	0	7	2012-08-02 05:38:03.962183	2012-08-02 10:39:21.951995
9601	1-9601	84	3	5724	8	7193	0	0	2012-08-02 05:47:12.569746	2012-08-02 10:39:24.309233
7882	1-7882	57	23	9627	69	5739	0	0	2012-08-02 05:35:33.710313	2012-08-02 10:39:25.873206
11008	1-11008	78	100	1398	34	7410	0	3	2012-08-02 10:39:25.873206	2012-08-02 10:39:25.873206
10940	1-10940	19	73	4076	45	6784	0	2	2012-08-02 10:38:58.75591	2012-08-02 10:39:26.673253
9942	1-9942	85	17	8278	55	4812	0	0	2012-08-02 05:49:31.912435	2012-08-02 10:39:28.229163
10433	1-10433	4	55	3324	4	8630	0	5	2012-08-02 10:27:57.555811	2012-08-02 10:39:28.229163
9376	1-9376	26	17	8605	79	4820	0	12	2012-08-02 05:45:41.21014	2012-08-02 10:39:28.229163
10496	1-10496	36	13	5409	1	6482	0	8	2012-08-02 10:31:55.346546	2012-08-02 10:39:29.809232
10841	1-10841	75	88	654	17	1080	0	0	2012-08-02 10:38:19.778219	2012-08-02 10:38:19.778219
6574	1-6574	85	24	9975	56	8407	0	6	2012-08-02 05:26:39.591242	2012-08-02 10:38:22.131703
5591	1-5591	18	24	6867	50	9694	0	2	2012-08-02 05:19:57.511568	2012-08-02 10:38:22.131703
10756	1-10756	21	50	6500	61	8377	0	5	2012-08-02 10:37:47.13649	2012-08-02 10:38:22.131703
2583	1-2583	9	56	4872	50	2147	225	0	2012-08-02 04:59:44.436594	2012-08-02 10:38:22.131703
9927	1-9927	10	71	3183	97	2296	0	4	2012-08-02 05:49:25.768195	2012-08-02 10:38:23.335866
10431	1-10431	62	75	6207	71	5713	0	0	2012-08-02 10:27:56.771142	2012-08-02 10:38:23.335866
10858	1-10858	48	1	391	37	804	0	11	2012-08-02 10:38:26.473899	2012-08-02 10:38:26.473899
9400	1-9400	81	34	3551	85	8806	0	2	2012-08-02 05:45:50.879051	2012-08-02 10:38:28.82744
9641	1-9641	17	3	3727	70	9209	0	8	2012-08-02 05:47:28.999498	2012-08-02 10:38:30.009498
10056	1-10056	35	92	3338	70	5669	0	16	2012-08-02 10:21:02.767505	2012-08-02 10:38:30.009498
10673	1-10673	98	98	2526	95	6264	0	11	2012-08-02 10:36:10.542573	2012-08-02 10:38:31.213882
10873	1-10873	94	81	500	50	9110	0	4	2012-08-02 10:38:32.013925	2012-08-02 10:38:32.013925
7346	1-7346	39	7	9084	2	8832	0	0	2012-08-02 05:31:51.366117	2012-08-02 10:38:34.74023
10879	1-10879	78	46	1995	7	9743	0	10	2012-08-02 10:38:34.74023	2012-08-02 10:38:34.74023
10293	1-10293	64	62	5808	18	8579	0	0	2012-08-02 10:25:34.788692	2012-08-02 10:38:35.899927
10882	1-10882	62	100	373	62	5513	0	12	2012-08-02 10:38:35.899927	2012-08-02 10:38:35.899927
10709	1-10709	34	73	4795	58	9752	0	0	2012-08-02 10:36:24.723366	2012-08-02 10:38:42.940084
10821	1-10821	23	43	1725	39	4744	0	15	2012-08-02 10:38:11.90992	2012-08-02 10:38:45.669057
10911	1-10911	100	12	2316	26	5346	0	0	2012-08-02 10:38:46.850819	2012-08-02 10:38:46.850819
5471	1-5471	39	77	5182	85	5971	0	6	2012-08-02 05:19:09.500351	2012-08-02 10:38:49.60414
10221	1-10221	22	1	6277	77	6760	0	3	2012-08-02 10:22:31.638722	2012-08-02 10:38:49.60414
10920	1-10920	44	96	2414	39	3068	0	10	2012-08-02 10:38:50.79204	2012-08-02 10:38:50.79204
10402	1-10402	5	93	3807	79	6807	0	3	2012-08-02 10:27:45.368507	2012-08-02 10:38:53.203026
8387	1-8387	53	89	9381	12	9436	0	5	2012-08-02 05:38:59.745635	2012-08-02 10:38:59.529301
10942	1-10942	32	5	1343	89	8926	0	2	2012-08-02 10:38:59.529301	2012-08-02 10:38:59.529301
9980	1-9980	68	17	3000	46	3220	0	2	2012-08-02 05:49:47.561891	2012-08-02 10:39:03.101981
8955	1-8955	92	61	9015	63	3577	0	0	2012-08-02 05:42:49.17626	2012-08-02 10:39:03.101981
10754	1-10754	91	79	8648	63	1960	0	2	2012-08-02 10:37:46.362984	2012-08-02 10:39:03.101981
10951	1-10951	4	11	1877	73	8328	0	5	2012-08-02 10:39:03.101981	2012-08-02 10:39:03.101981
10580	1-10580	33	52	4914	47	9184	0	4	2012-08-02 10:35:35.253698	2012-08-02 10:39:06.694644
10963	1-10963	92	80	298	96	1357	0	0	2012-08-02 10:39:07.866336	2012-08-02 10:39:07.866336
10407	1-10407	72	60	2623	29	1315	0	17	2012-08-02 10:27:47.301757	2012-08-02 10:39:09.04928
10203	1-10203	20	5	9084	29	6730	0	2	2012-08-02 10:22:24.555829	2012-08-02 10:39:09.04928
8881	1-8881	4	57	5214	12	6212	0	0	2012-08-02 05:42:19.037189	2012-08-02 10:39:09.04928
7564	1-7564	77	60	9763	29	3524	0	12	2012-08-02 05:33:20.489874	2012-08-02 10:39:09.04928
2840	1-2840	46	57	8003	83	2131	153	0	2012-08-02 05:01:33.175233	2012-08-02 10:39:09.04928
3602	1-3602	83	70	4186	64	864	0	0	2012-08-02 05:06:42.932249	2012-08-02 10:39:14.112605
10874	1-10874	61	7	2451	44	8627	0	0	2012-08-02 10:38:32.80645	2012-08-02 10:39:14.112605
10979	1-10979	9	64	285	7	5313	0	11	2012-08-02 10:39:14.112605	2012-08-02 10:39:14.112605
10986	1-10986	4	11	590	62	3292	0	0	2012-08-02 10:39:16.866067	2012-08-02 10:39:20.791409
7843	1-7843	12	9	4539	46	4363	0	0	2012-08-02 05:35:17.271068	2012-08-02 10:39:22.338804
10225	1-10225	36	50	4014	89	1508	0	0	2012-08-02 10:22:33.208508	2012-08-02 10:39:22.338804
3130	1-3130	74	83	9057	50	1111	793	0	2012-08-02 05:03:32.145954	2012-08-02 10:39:22.338804
2571	1-2571	98	82	6630	53	3552	0	0	2012-08-02 04:59:39.28757	2012-08-02 10:39:23.148279
11001	1-11001	33	87	2616	89	7161	0	0	2012-08-02 10:39:23.148279	2012-08-02 10:39:23.148279
10852	1-10852	98	50	2312	86	3238	0	5	2012-08-02 10:38:24.109144	2012-08-02 10:39:23.921828
10377	1-10377	17	15	4393	21	4460	0	5	2012-08-02 10:27:35.444547	2012-08-02 10:39:27.841229
11017	1-11017	87	92	510	34	6557	0	7	2012-08-02 10:39:29.424811	2012-08-02 10:39:29.424811
9046	1-9046	82	19	3112	51	3333	0	5	2012-08-02 05:43:26.590487	2012-08-02 10:39:31.770731
10715	1-10715	23	10	7050	51	8545	0	0	2012-08-02 10:36:27.021387	2012-08-02 10:39:31.770731
10651	1-10651	19	24	2977	48	2843	0	7	2012-08-02 10:36:01.899372	2012-08-02 10:39:31.770731
10895	1-10895	100	63	1593	19	5236	0	4	2012-08-02 10:38:41.38223	2012-08-02 10:39:31.770731
9824	1-9824	53	83	6821	48	9552	0	0	2012-08-02 05:48:43.796114	2012-08-02 10:39:31.770731
11023	1-11023	20	48	7394	63	6733	0	7	2012-08-02 10:39:31.770731	2012-08-02 10:39:34.945883
9511	1-9511	72	58	7934	42	9552	0	0	2012-08-02 05:46:35.993926	2012-08-02 10:39:37.23284
10877	1-10877	28	42	7593	16	8776	0	2	2012-08-02 10:38:33.955329	2012-08-02 10:39:37.23284
11037	1-11037	53	16	1195	58	9699	0	0	2012-08-02 10:39:37.23284	2012-08-02 10:39:37.23284
11038	1-11038	44	89	1556	66	8134	0	2	2012-08-02 10:39:38.127857	2012-08-02 10:39:38.127857
10643	1-10643	67	94	1741	32	5880	0	0	2012-08-02 10:35:58.701003	2012-08-02 10:38:23.722511
10658	1-10658	76	84	8571	78	5892	0	5	2012-08-02 10:36:04.663382	2012-08-02 10:38:23.722511
3451	1-3451	45	94	8644	41	4189	0	0	2012-08-02 05:05:41.950156	2012-08-02 10:38:23.722511
10851	1-10851	19	78	997	94	8171	0	4	2012-08-02 10:38:23.722511	2012-08-02 10:38:23.722511
10617	1-10617	2	91	4866	47	9291	0	6	2012-08-02 10:35:38.016442	2012-08-02 10:38:24.905197
10854	1-10854	71	8	1324	91	4999	0	6	2012-08-02 10:38:24.905197	2012-08-02 10:38:24.905197
9149	1-9149	52	3	2633	29	7289	0	9	2012-08-02 05:44:08.921905	2012-08-02 10:38:27.258545
10820	1-10820	38	39	6383	45	3945	0	16	2012-08-02 10:38:11.523412	2012-08-02 10:38:27.258545
10875	1-10875	27	65	1333	59	3149	0	0	2012-08-02 10:38:33.193019	2012-08-02 10:38:33.193019
10864	1-10864	53	67	589	34	1839	0	20	2012-08-02 10:38:28.82744	2012-08-02 10:38:37.833705
10831	1-10831	9	41	2002	46	5096	0	5	2012-08-02 10:38:15.88932	2012-08-02 10:38:41.38223
10577	1-10577	76	33	937	13	1024	0	0	2012-08-02 10:35:34.992303	2012-08-02 10:38:42.166168
10619	1-10619	61	53	6312	31	5609	0	8	2012-08-02 10:35:38.171272	2012-08-02 10:38:42.166168
10898	1-10898	33	97	1090	33	6784	0	11	2012-08-02 10:38:42.166168	2012-08-02 10:38:42.166168
10903	1-10903	68	51	310	12	3243	0	10	2012-08-02 10:38:44.110778	2012-08-02 10:38:44.110778
10906	1-10906	78	44	143	92	9035	0	0	2012-08-02 10:38:45.281993	2012-08-02 10:38:45.281993
10909	1-10909	87	26	361	17	4896	0	5	2012-08-02 10:38:46.442004	2012-08-02 10:38:46.442004
1317	1-1317	60	35	9778	43	5024	0	0	2012-08-02 04:50:05.015502	2012-08-02 10:38:47.624215
10912	1-10912	89	81	98	16	4602	0	5	2012-08-02 10:38:47.624215	2012-08-02 10:38:47.624215
10915	1-10915	48	79	533	61	1422	0	0	2012-08-02 10:38:48.795357	2012-08-02 10:38:48.795357
10918	1-10918	90	65	1425	80	4616	0	2	2012-08-02 10:38:50.004189	2012-08-02 10:38:50.004189
10356	1-10356	15	55	2682	88	8626	0	6	2012-08-02 10:27:27.089155	2012-08-02 10:38:51.181209
8572	1-8572	48	18	5193	42	8793	0	0	2012-08-02 05:40:14.030982	2012-08-02 10:38:55.979849
8988	1-8988	81	18	6923	27	7838	0	20	2012-08-02 05:43:02.745249	2012-08-02 10:38:55.979849
10128	1-10128	1	18	6207	34	7473	0	10	2012-08-02 10:21:31.123529	2012-08-02 10:38:55.979849
9523	1-9523	20	1	9993	11	6666	0	0	2012-08-02 05:46:40.835476	2012-08-02 10:38:55.979849
10932	1-10932	24	16	9255	1	7024	0	0	2012-08-02 10:38:55.979849	2012-08-02 10:38:55.979849
10938	1-10938	40	70	6448	45	7962	0	7	2012-08-02 10:38:57.97117	2012-08-02 10:38:57.97117
10887	1-10887	54	34	4600	12	6263	0	3	2012-08-02 10:38:37.833705	2012-08-02 10:38:59.142666
9608	1-9608	47	23	5222	68	4266	0	6	2012-08-02 05:47:15.48706	2012-08-02 10:39:04.749561
10955	1-10955	82	85	407	23	1814	0	8	2012-08-02 10:39:04.749561	2012-08-02 10:39:04.749561
9747	1-9747	95	97	3692	2	9591	0	2	2012-08-02 05:48:12.581309	2012-08-02 10:39:08.241849
10964	1-10964	37	60	2075	97	3490	0	11	2012-08-02 10:39:08.241849	2012-08-02 10:39:08.241849
10867	1-10867	53	70	6623	3	7088	0	3	2012-08-02 10:38:30.009498	2012-08-02 10:39:14.908614
9663	1-9663	11	25	8326	56	4220	0	2	2012-08-02 05:47:38.061032	2012-08-02 10:39:14.908614
10981	1-10981	95	91	1970	25	8696	0	0	2012-08-02 10:39:14.908614	2012-08-02 10:39:14.908614
4602	1-4602	76	54	9936	92	6405	0	0	2012-08-02 05:13:22.730917	2012-08-02 10:39:16.081614
9546	1-9546	10	93	6130	1	3519	0	8	2012-08-02 05:46:50.226303	2012-08-02 10:39:16.081614
10984	1-10984	83	92	3059	93	6936	0	8	2012-08-02 10:39:16.081614	2012-08-02 10:39:16.081614
10989	1-10989	5	60	356	12	591	0	0	2012-08-02 10:39:18.424847	2012-08-02 10:39:18.424847
7696	1-7696	70	24	6996	1	5350	0	2	2012-08-02 05:34:14.738341	2012-08-02 10:39:19.198296
8719	1-8719	85	68	7273	7	5263	0	0	2012-08-02 05:41:13.387932	2012-08-02 10:39:19.198296
10991	1-10991	3	18	1027	68	8593	0	0	2012-08-02 10:39:19.198296	2012-08-02 10:39:19.198296
10847	1-10847	88	63	3301	24	5248	0	7	2012-08-02 10:38:22.131703	2012-08-02 11:11:08.047947
7565	1-7565	60	24	6904	79	9881	0	0	2012-08-02 05:33:20.909625	2012-08-02 11:11:08.047947
9880	1-9880	38	63	8186	18	9762	0	4	2012-08-02 05:49:06.598124	2012-08-02 11:11:08.047947
8538	1-8538	26	63	4441	98	2903	0	2	2012-08-02 05:40:00.338261	2012-08-02 11:11:08.047947
9481	1-9481	26	63	4840	35	9389	0	0	2012-08-02 05:46:23.807004	2012-08-02 11:11:08.047947
11044	1-11044	49	79	1990	63	5078	0	6	2012-08-02 11:11:08.047947	2012-08-02 11:11:08.047947
5508	1-5508	34	77	8374	73	2775	0	0	2012-08-02 05:19:24.408715	2012-08-02 11:11:08.487871
11045	1-11045	89	17	3431	40	7516	0	0	2012-08-02 11:11:08.487871	2012-08-02 11:11:08.487871
4572	1-4572	35	62	7404	93	3025	0	0	2012-08-02 05:13:10.478351	2012-08-02 11:11:08.876498
11046	1-11046	22	93	7460	99	7164	0	4	2012-08-02 11:11:08.876498	2012-08-02 11:11:08.876498
9879	1-9879	62	56	4772	50	8538	0	5	2012-08-02 05:49:06.200502	2012-08-02 11:11:09.687486
11049	1-11049	11	9	2337	56	2727	0	11	2012-08-02 11:11:09.687486	2012-08-02 11:11:09.687486
10823	1-10823	50	98	3192	62	9661	0	8	2012-08-02 10:38:12.705559	2012-08-02 11:11:10.096452
2216	1-2216	100	47	8757	5	6796	0	0	2012-08-02 04:57:04.410884	2012-08-02 11:11:10.096452
10935	1-10935	13	98	2879	71	4088	0	0	2012-08-02 10:38:56.789142	2012-08-02 11:11:10.096452
11050	1-11050	31	5	2323	98	7114	0	4	2012-08-02 11:11:10.096452	2012-08-02 11:11:10.096452
10669	1-10669	41	41	3726	42	4952	0	0	2012-08-02 10:36:08.962545	2012-08-02 11:11:10.484443
11051	1-11051	3	72	856	41	8309	0	4	2012-08-02 11:11:10.484443	2012-08-02 11:11:10.484443
6141	1-6141	8	27	8737	93	6670	0	8	2012-08-02 05:23:41.741711	2012-08-02 11:11:10.898383
10561	1-10561	57	27	5574	28	3188	0	3	2012-08-02 10:35:33.791869	2012-08-02 11:11:10.898383
10004	1-10004	88	27	2439	93	1123	0	13	2012-08-02 09:10:30.946899	2012-08-02 11:11:10.898383
9844	1-9844	64	96	2985	20	5984	0	9	2012-08-02 05:48:51.963868	2012-08-02 11:11:11.672045
11014	1-11014	36	4	870	17	9791	0	5	2012-08-02 10:39:28.229163	2012-08-02 11:11:13.288184
10892	1-10892	32	73	2430	17	9994	0	0	2012-08-02 10:38:39.801776	2012-08-02 11:11:13.288184
10439	1-10439	16	64	8482	96	9837	0	8	2012-08-02 10:27:59.931113	2012-08-02 11:11:14.072843
10041	1-10041	39	50	8553	7	5461	0	0	2012-08-02 10:20:55.569588	2012-08-02 11:11:14.847341
9764	1-9764	71	100	5224	20	5021	0	4	2012-08-02 05:48:19.48847	2012-08-02 11:11:15.632359
11065	1-11065	85	20	830	7	2765	0	3	2012-08-02 11:11:15.632359	2012-08-02 11:11:15.632359
10526	1-10526	84	40	8524	77	9270	0	0	2012-08-02 10:32:07.181979	2012-08-02 11:11:16.030855
9294	1-9294	34	38	5903	77	6376	0	0	2012-08-02 05:45:07.840665	2012-08-02 11:11:16.030855
11066	1-11066	3	81	3586	55	8846	0	0	2012-08-02 11:11:16.030855	2012-08-02 11:11:16.030855
9176	1-9176	24	67	6569	18	5512	0	4	2012-08-02 05:44:19.763736	2012-08-02 11:11:16.450807
11028	1-11028	98	57	710	20	2129	0	0	2012-08-02 10:39:33.752837	2012-08-02 11:11:16.450807
9854	1-9854	4	6	5708	21	7513	0	0	2012-08-02 05:48:56.003094	2012-08-02 11:11:18.022283
10809	1-10809	14	93	2269	74	1477	0	0	2012-08-02 10:38:06.838199	2012-08-02 11:11:18.022283
11071	1-11071	93	74	1848	6	8011	0	0	2012-08-02 11:11:18.022283	2012-08-02 11:11:18.022283
9591	1-9591	92	46	9115	28	4826	0	17	2012-08-02 05:47:08.51507	2012-08-02 11:11:18.795729
8774	1-8774	75	49	4090	90	1985	0	0	2012-08-02 05:41:35.789292	2012-08-02 11:11:18.795729
10731	1-10731	38	23	4891	41	3785	0	4	2012-08-02 10:36:33.214941	2012-08-02 11:11:20.766065
11063	1-11063	29	7	3751	100	7095	0	5	2012-08-02 11:11:14.847341	2012-08-02 11:11:21.937739
8012	1-8012	99	7	3519	28	1384	0	0	2012-08-02 05:36:27.950851	2012-08-02 11:11:21.937739
11081	1-11081	8	20	572	7	6318	0	5	2012-08-02 11:11:21.937739	2012-08-02 11:11:21.937739
9473	1-9473	92	73	8714	91	5693	0	9	2012-08-02 05:46:20.60314	2012-08-02 11:11:25.387474
11090	1-11090	90	48	2056	73	6433	0	3	2012-08-02 11:11:25.387474	2012-08-02 11:11:25.387474
587	1-587	63	91	9209	48	6315	1292	3	2012-08-02 04:44:33.708932	2012-08-02 11:11:25.387474
11093	1-11093	55	53	419	46	1385	0	0	2012-08-02 11:11:26.592527	2012-08-02 11:11:26.592527
10957	1-10957	37	88	1218	29	1860	0	0	2012-08-02 10:39:05.523038	2012-08-02 11:11:27.753399
10592	1-10592	87	88	4977	54	4092	0	3	2012-08-02 10:35:36.170899	2012-08-02 11:11:27.753399
11096	1-11096	50	58	3018	88	9797	0	0	2012-08-02 11:11:27.753399	2012-08-02 11:11:27.753399
9426	1-9426	17	36	8198	18	8440	0	13	2012-08-02 05:46:01.387028	2012-08-02 11:11:31.223945
2974	1-2974	35	96	9990	51	6479	0	0	2012-08-02 05:02:28.731371	2012-08-02 11:11:31.223945
11072	1-11072	87	25	4793	49	6183	0	0	2012-08-02 11:11:18.795729	2012-08-02 11:11:37.083093
11120	1-11120	42	58	1435	98	6102	0	7	2012-08-02 11:11:37.083093	2012-08-02 11:11:37.083093
5862	1-5862	21	49	8662	58	8216	3761	0	2012-08-02 05:21:47.642321	2012-08-02 11:11:37.083093
9842	1-9842	83	1	4992	20	8071	0	3	2012-08-02 05:48:51.14535	2012-08-02 11:11:42.254873
10960	1-10960	49	47	3368	1	3622	0	0	2012-08-02 10:39:06.694644	2012-08-02 11:11:42.254873
9000	1-9000	2	54	5982	77	4012	0	5	2012-08-02 05:43:07.67428	2012-08-02 11:11:46.957424
8995	1-8995	73	80	8732	77	4396	0	5	2012-08-02 05:43:05.631047	2012-08-02 11:11:46.957424
11145	1-11145	2	77	2472	89	6692	0	8	2012-08-02 11:11:46.957424	2012-08-02 11:11:46.957424
10578	1-10578	12	28	7281	21	8214	0	5	2012-08-02 10:35:35.058565	2012-08-02 11:11:48.140241
8467	1-8467	36	61	4099	85	5076	0	5	2012-08-02 05:39:31.941165	2012-08-02 11:11:49.322907
7890	1-7890	51	61	7084	79	5998	0	0	2012-08-02 05:35:37.070411	2012-08-02 11:11:49.322907
3495	1-3495	51	61	5605	85	3687	0	0	2012-08-02 05:05:59.781724	2012-08-02 11:11:49.322907
11151	1-11151	13	85	4159	61	8148	0	6	2012-08-02 11:11:49.322907	2012-08-02 11:11:49.322907
2973	1-2973	100	61	9891	58	8056	2667	0	2012-08-02 05:02:28.322658	2012-08-02 11:11:49.322907
10257	1-10257	12	41	9374	22	6994	0	7	2012-08-02 10:25:20.53652	2012-08-02 11:11:50.505526
11154	1-11154	90	54	3626	41	6964	0	6	2012-08-02 11:11:50.505526	2012-08-02 11:11:50.505526
8980	1-8980	56	55	3541	5	7145	0	0	2012-08-02 05:42:59.453135	2012-08-02 11:11:54.052869
10247	1-10247	26	15	7428	5	8470	0	2	2012-08-02 10:25:16.601051	2012-08-02 11:11:54.052869
9650	1-9650	53	54	5326	10	6375	0	4	2012-08-02 05:47:32.724175	2012-08-02 11:11:57.577745
7606	1-7606	82	46	9479	6	6385	0	7	2012-08-02 05:33:37.614991	2012-08-02 11:11:58.737639
10157	1-10157	32	61	3246	58	4985	0	8	2012-08-02 10:22:06.63264	2012-08-02 11:11:58.737639
11175	1-11175	3	13	453	46	1622	0	8	2012-08-02 11:11:58.737639	2012-08-02 11:11:58.737639
8116	1-8116	88	58	7100	13	4442	344	6	2012-08-02 05:37:09.761406	2012-08-02 11:11:58.737639
11178	1-11178	61	62	42	13	5947	0	0	2012-08-02 11:11:59.897958	2012-08-02 11:11:59.897958
11135	1-11135	31	100	2789	83	2332	0	0	2012-08-02 11:11:42.629985	2012-08-02 11:12:01.093497
11184	1-11184	84	77	994	19	5662	0	4	2012-08-02 11:12:02.297576	2012-08-02 11:12:02.297576
11052	1-11052	5	93	2234	27	6233	0	0	2012-08-02 11:11:10.898383	2012-08-02 11:12:06.970827
4709	1-4709	52	99	7530	28	7291	0	0	2012-08-02 05:14:06.266056	2012-08-02 11:12:06.970827
11195	1-11195	63	27	3376	99	7742	0	12	2012-08-02 11:12:06.970827	2012-08-02 11:12:06.970827
11212	1-11212	35	82	36	53	7744	0	4	2012-08-02 11:12:13.224419	2012-08-02 11:12:13.224419
10996	1-10996	54	18	1388	71	2678	0	3	2012-08-02 10:39:21.178299	2012-08-02 11:12:19.168347
11227	1-11227	21	43	5310	18	3419	0	4	2012-08-02 11:12:19.168347	2012-08-02 11:12:19.168347
11077	1-11077	16	55	1693	82	7844	0	22	2012-08-02 11:11:20.379696	2012-08-02 11:12:21.529328
9472	1-9472	78	55	9915	59	9171	0	5	2012-08-02 05:46:20.183238	2012-08-02 11:12:21.529328
10880	1-10880	97	38	5787	59	5315	0	10	2012-08-02 10:38:35.115721	2012-08-02 11:12:21.529328
11233	1-11233	23	32	2246	55	4736	0	0	2012-08-02 11:12:21.529328	2012-08-02 11:12:21.529328
7623	1-7623	22	50	7336	85	6463	0	9	2012-08-02 05:33:44.564377	2012-08-02 11:12:22.704417
10180	1-10180	29	50	9429	9	8026	0	19	2012-08-02 10:22:15.649472	2012-08-02 11:12:22.704417
10829	1-10829	73	50	4797	11	1697	0	0	2012-08-02 10:38:15.104878	2012-08-02 11:12:22.704417
11118	1-11118	51	60	2642	87	4085	0	0	2012-08-02 11:11:36.309873	2012-08-02 11:12:25.058232
10815	1-10815	38	41	2341	44	7237	0	5	2012-08-02 10:38:09.567524	2012-08-02 11:12:28.595194
9508	1-9508	7	26	5755	98	9358	0	13	2012-08-02 05:46:34.767742	2012-08-02 11:12:30.893329
9755	1-9755	85	100	2234	23	925	0	10	2012-08-02 05:48:15.796082	2012-08-02 11:12:30.893329
10998	1-10998	23	4	6806	11	9564	0	6	2012-08-02 10:39:21.951995	2012-08-02 11:12:31.699949
10753	1-10753	84	5	6003	63	810	0	0	2012-08-02 10:37:45.976929	2012-08-02 11:12:31.699949
11259	1-11259	76	24	843	4	9498	0	8	2012-08-02 11:12:31.699949	2012-08-02 11:12:31.699949
10860	1-10860	57	45	8273	3	8403	0	0	2012-08-02 10:38:27.258545	2012-08-02 11:12:32.882289
10535	1-10535	4	45	6347	70	9642	0	12	2012-08-02 10:32:10.776049	2012-08-02 11:12:32.882289
11223	1-11223	81	85	9499	75	6848	0	0	2012-08-02 11:12:17.632767	2012-08-02 11:12:32.882289
10194	1-10194	51	45	6615	61	6614	0	16	2012-08-02 10:22:21.073862	2012-08-02 11:12:32.882289
10297	1-10297	73	45	5835	99	4507	0	0	2012-08-02 10:25:36.336191	2012-08-02 11:12:32.882289
11263	1-11263	97	3	1649	45	8083	0	0	2012-08-02 11:12:32.882289	2012-08-02 11:12:32.882289
11266	1-11266	27	37	120	98	7331	0	5	2012-08-02 11:12:34.47351	2012-08-02 11:12:34.47351
6886	1-6886	77	22	5785	72	3366	0	0	2012-08-02 05:28:45.69601	2012-08-02 11:12:35.633704
10493	1-10493	2	95	3885	22	6635	0	10	2012-08-02 10:31:54.174811	2012-08-02 11:12:35.633704
10773	1-10773	35	80	9911	84	4251	0	0	2012-08-02 10:37:53.104636	2012-08-02 11:12:35.633704
8332	1-8332	91	27	1980	44	2451	0	6	2012-08-02 05:38:37.540303	2012-08-02 11:12:36.80497
9116	1-9116	11	7	8808	71	5106	0	4	2012-08-02 05:43:55.388288	2012-08-02 11:12:36.80497
9660	1-9660	91	53	6722	87	7117	0	0	2012-08-02 05:47:36.85646	2012-08-02 11:12:37.942975
8974	1-8974	29	53	5789	66	2978	0	0	2012-08-02 05:42:57.000279	2012-08-02 11:12:37.942975
10092	1-10092	34	100	566	5	520	0	0	2012-08-02 10:21:16.753641	2012-08-02 11:12:37.942975
11275	1-11275	94	25	666	53	4807	0	4	2012-08-02 11:12:37.942975	2012-08-02 11:12:37.942975
11054	1-11054	97	16	5003	96	5453	0	0	2012-08-02 11:11:11.672045	2012-08-02 11:12:40.274544
11286	1-11286	56	48	1276	6	6122	0	0	2012-08-02 11:12:42.219001	2012-08-02 11:12:42.219001
6733	1-6733	84	30	7129	88	9254	0	3	2012-08-02 05:27:43.761097	2012-08-02 11:12:49.283536
10018	1-10018	19	58	9241	13	9360	0	11	2012-08-02 10:16:58.663543	2012-08-02 11:12:50.432315
10917	1-10917	85	85	8355	13	7850	0	7	2012-08-02 10:38:49.60414	2012-08-02 11:12:50.432315
9973	1-9973	26	61	3128	14	7002	0	0	2012-08-02 05:49:44.710654	2012-08-02 11:12:52.785597
11079	1-11079	46	51	358	1	4031	0	0	2012-08-02 11:11:21.153363	2012-08-02 11:11:21.153363
10579	1-10579	99	40	7188	20	7590	0	0	2012-08-02 10:35:35.125133	2012-08-02 11:11:22.324789
11083	1-11083	57	42	290	40	9183	0	2	2012-08-02 11:11:22.324789	2012-08-02 11:11:22.324789
8818	1-8818	70	34	9375	14	6684	0	0	2012-08-02 05:41:53.523261	2012-08-02 11:11:23.87241
11086	1-11086	13	14	4204	34	8628	0	0	2012-08-02 11:11:23.87241	2012-08-02 11:11:23.87241
8365	1-8365	57	64	7328	15	8937	0	5	2012-08-02 05:38:50.817859	2012-08-02 11:11:29.688061
11101	1-11101	70	5	2324	64	3199	0	0	2012-08-02 11:11:29.688061	2012-08-02 11:11:29.688061
9948	1-9948	13	14	2338	72	1548	0	0	2012-08-02 05:49:34.388568	2012-08-02 11:11:30.837301
10777	1-10777	28	23	1972	90	1501	0	0	2012-08-02 10:37:54.71802	2012-08-02 11:11:32.795243
11109	1-11109	1	90	897	21	5798	0	6	2012-08-02 11:11:32.795243	2012-08-02 11:11:32.795243
9818	1-9818	10	97	5453	48	5692	0	4	2012-08-02 05:48:41.365616	2012-08-02 11:11:33.933733
10636	1-10636	38	97	2680	51	5053	0	0	2012-08-02 10:35:39.455595	2012-08-02 11:11:33.933733
10766	1-10766	35	97	636	64	244	0	8	2012-08-02 10:37:50.342248	2012-08-02 11:11:33.933733
8914	1-8914	22	94	5354	77	4646	0	7	2012-08-02 05:42:32.416518	2012-08-02 11:11:33.933733
10278	1-10278	61	97	7034	91	5556	0	9	2012-08-02 10:25:28.799998	2012-08-02 11:11:33.933733
11112	1-11112	41	48	3643	97	8828	0	0	2012-08-02 11:11:33.933733	2012-08-02 11:11:33.933733
11057	1-11057	66	97	4675	62	3322	0	15	2012-08-02 11:11:12.470153	2012-08-02 11:11:37.867504
11122	1-11122	41	81	3475	97	8395	0	2	2012-08-02 11:11:37.867504	2012-08-02 11:11:37.867504
9577	1-9577	13	24	4216	60	9908	0	6	2012-08-02 05:47:02.72248	2012-08-02 11:11:43.016532
11134	1-11134	96	87	2265	24	7402	0	0	2012-08-02 11:11:43.016532	2012-08-02 11:11:43.016532
10611	1-10611	56	83	4812	50	9121	0	7	2012-08-02 10:35:37.534904	2012-08-02 11:11:43.819221
11137	1-11137	77	51	853	83	7370	0	6	2012-08-02 11:11:43.819221	2012-08-02 11:11:43.819221
11140	1-11140	86	21	1199	81	2534	0	0	2012-08-02 11:11:44.979358	2012-08-02 11:11:44.979358
10620	1-10620	32	85	6536	65	4444	0	7	2012-08-02 10:35:38.237624	2012-08-02 11:11:48.526807
11152	1-11152	76	42	1542	66	7925	0	0	2012-08-02 11:11:49.720618	2012-08-02 11:11:49.720618
10800	1-10800	29	42	7814	86	9131	0	10	2012-08-02 10:38:03.7003	2012-08-02 11:11:50.892453
10725	1-10725	94	69	7924	1	8180	0	6	2012-08-02 10:36:30.923975	2012-08-02 11:11:50.892453
10919	1-10919	21	70	5611	30	5871	0	0	2012-08-02 10:38:50.391199	2012-08-02 11:11:50.892453
11155	1-11155	50	86	5420	69	8858	0	5	2012-08-02 11:11:50.892453	2012-08-02 11:11:50.892453
11174	1-11174	40	28	2081	3	6942	0	16	2012-08-02 11:11:58.362033	2012-08-02 11:11:58.362033
8104	1-8104	26	24	1001	56	801	0	10	2012-08-02 05:37:04.932427	2012-08-02 11:11:59.522129
6036	1-6036	71	9	8124	79	5389	0	0	2012-08-02 05:22:59.497412	2012-08-02 11:11:59.522129
10498	1-10498	32	71	4076	22	7366	0	0	2012-08-02 10:31:56.130676	2012-08-02 11:11:59.522129
11177	1-11177	56	56	544	9	7609	0	5	2012-08-02 11:11:59.522129	2012-08-02 11:11:59.522129
10127	1-10127	58	71	5387	7	6196	0	0	2012-08-02 10:21:30.747851	2012-08-02 11:12:01.911108
11189	1-11189	10	70	439	66	8794	0	11	2012-08-02 11:12:04.253594	2012-08-02 11:12:04.253594
9329	1-9329	2	68	5043	49	3285	0	0	2012-08-02 05:45:22.018036	2012-08-02 11:12:05.424338
7544	1-7544	82	7	4813	55	1120	0	0	2012-08-02 05:33:12.367858	2012-08-02 11:12:06.595309
10978	1-10978	93	7	6790	70	8965	0	0	2012-08-02 10:39:13.725909	2012-08-02 11:12:06.595309
11196	1-11196	95	82	4367	7	6750	0	8	2012-08-02 11:12:06.595309	2012-08-02 11:12:06.595309
8288	1-8288	59	75	4418	85	6026	0	5	2012-08-02 05:38:19.784264	2012-08-02 11:12:08.164732
9904	1-9904	40	9	2104	16	1995	0	0	2012-08-02 05:49:16.428476	2012-08-02 11:12:14.771122
11216	1-11216	20	96	909	9	2840	0	0	2012-08-02 11:12:14.771122	2012-08-02 11:12:14.771122
11232	1-11232	63	38	470	62	2066	0	4	2012-08-02 11:12:21.124269	2012-08-02 11:12:21.124269
9865	1-9865	11	9	3276	29	1854	0	8	2012-08-02 05:49:00.498533	2012-08-02 11:12:22.317417
11235	1-11235	62	48	1897	9	9644	0	9	2012-08-02 11:12:22.317417	2012-08-02 11:12:22.317417
5712	1-5712	62	75	7525	84	6389	0	0	2012-08-02 05:20:46.374661	2012-08-02 11:12:23.48894
9386	1-9386	92	68	3955	85	6395	0	2	2012-08-02 05:45:45.24256	2012-08-02 11:12:23.48894
11238	1-11238	39	84	1834	68	9758	0	5	2012-08-02 11:12:23.48894	2012-08-02 11:12:23.48894
8421	1-8421	24	3	9130	96	8863	0	6	2012-08-02 05:39:13.541434	2012-08-02 11:12:24.693493
11241	1-11241	60	88	1338	3	8716	0	0	2012-08-02 11:12:24.693493	2012-08-02 11:12:24.693493
7426	1-7426	97	30	9090	100	8906	0	6	2012-08-02 05:32:23.720034	2012-08-02 11:12:25.864716
10907	1-10907	74	39	4265	30	5237	0	3	2012-08-02 10:38:45.669057	2012-08-02 11:12:25.864716
11258	1-11258	11	64	1557	45	5470	0	6	2012-08-02 11:12:31.291449	2012-08-02 11:12:31.291449
6882	1-6882	66	75	2872	84	2390	0	2	2012-08-02 05:28:44.049842	2012-08-02 11:12:36.022845
11270	1-11270	32	85	629	75	2591	0	8	2012-08-02 11:12:36.022845	2012-08-02 11:12:36.022845
11273	1-11273	78	72	740	11	2895	0	0	2012-08-02 11:12:37.191805	2012-08-02 11:12:37.191805
10629	1-10629	19	43	5241	31	8279	0	5	2012-08-02 10:35:38.924027	2012-08-02 11:12:40.649859
10360	1-10360	72	39	1292	80	3709	0	5	2012-08-02 10:27:28.657956	2012-08-02 11:12:41.821404
11296	1-11296	31	73	125	82	1726	0	0	2012-08-02 11:12:46.540917	2012-08-02 11:12:46.540917
11305	1-11305	86	26	2732	100	6515	0	5	2012-08-02 11:12:49.681048	2012-08-02 11:12:49.681048
11308	1-11308	80	19	68	73	8808	0	4	2012-08-02 11:12:50.819099	2012-08-02 11:12:50.819099
6150	1-6150	19	94	9518	59	5821	0	0	2012-08-02 05:23:45.343046	2012-08-02 11:12:53.161303
10904	1-10904	63	95	1279	82	1847	0	10	2012-08-02 10:38:44.497604	2012-08-02 11:13:02.386499
10995	1-10995	49	62	6470	27	8395	0	0	2012-08-02 10:39:20.791409	2012-08-02 11:13:02.386499
11341	1-11341	76	71	1337	21	4748	0	5	2012-08-02 11:13:03.579887	2012-08-02 11:13:03.579887
11036	1-11036	18	12	5623	73	5226	0	0	2012-08-02 10:39:36.857292	2012-08-02 11:13:04.706838
10285	1-10285	43	18	6982	8	6611	0	2	2012-08-02 10:25:31.570398	2012-08-02 11:13:05.877969
8730	1-8730	92	94	9487	74	4052	0	2	2012-08-02 05:41:17.817948	2012-08-02 11:13:18.950085
6222	1-6222	33	94	7129	84	3312	0	6	2012-08-02 05:24:14.870714	2012-08-02 11:13:18.950085
11380	1-11380	80	85	568	94	1171	0	0	2012-08-02 11:13:18.950085	2012-08-02 11:13:18.950085
11262	1-11262	86	29	6332	57	8711	0	6	2012-08-02 11:12:33.291168	2012-08-02 11:13:19.734778
11383	1-11383	89	13	1514	29	7723	0	0	2012-08-02 11:13:19.734778	2012-08-02 11:13:19.734778
7025	1-7025	89	45	3934	85	4312	0	5	2012-08-02 05:29:41.69144	2012-08-02 11:13:25.555476
11398	1-11398	42	85	526	69	7736	0	13	2012-08-02 11:13:25.555476	2012-08-02 11:13:25.555476
10363	1-10363	84	94	7557	46	8401	0	0	2012-08-02 10:27:29.863533	2012-08-02 11:13:26.705297
8220	1-8220	99	33	7905	22	9226	0	0	2012-08-02 05:37:52.245022	2012-08-02 11:13:26.705297
10340	1-10340	74	6	7658	32	9229	0	0	2012-08-02 10:25:53.117451	2012-08-02 11:13:26.705297
10837	1-10837	96	6	6824	28	5495	0	7	2012-08-02 10:38:18.607032	2012-08-02 11:13:26.705297
10834	1-10834	34	6	8636	90	9538	0	5	2012-08-02 10:38:17.049436	2012-08-02 11:13:26.705297
11064	1-11064	69	12	1379	50	3393	0	0	2012-08-02 11:11:15.257231	2012-08-02 11:13:26.705297
10104	1-10104	25	77	9730	61	6570	0	5	2012-08-02 10:21:21.604636	2012-08-02 11:11:23.485277
10312	1-10312	93	77	5864	81	4503	0	10	2012-08-02 10:25:42.243659	2012-08-02 11:11:23.485277
11085	1-11085	84	14	733	77	5212	0	2	2012-08-02 11:11:23.485277	2012-08-02 11:11:23.485277
10294	1-10294	45	44	6888	65	2683	0	6	2012-08-02 10:25:35.175664	2012-08-02 11:11:24.613506
10096	1-10096	97	4	1765	41	1627	0	0	2012-08-02 10:21:18.333496	2012-08-02 11:11:25.785107
10062	1-10062	87	97	3728	38	4938	0	0	2012-08-02 10:21:05.077898	2012-08-02 11:11:25.785107
8866	1-8866	49	4	8837	71	7427	0	5	2012-08-02 05:42:12.914584	2012-08-02 11:11:25.785107
11091	1-11091	6	41	2839	97	5356	0	0	2012-08-02 11:11:25.785107	2012-08-02 11:11:25.785107
11110	1-11110	89	63	678	81	7328	0	0	2012-08-02 11:11:33.170861	2012-08-02 11:11:33.170861
7563	1-7563	60	96	7077	68	6300	0	0	2012-08-02 05:33:20.081503	2012-08-02 11:11:35.525612
4910	1-4910	59	22	5375	53	2531	0	0	2012-08-02 05:15:27.040431	2012-08-02 11:11:40.607873
10931	1-10931	56	59	4538	31	8537	0	0	2012-08-02 10:38:54.786186	2012-08-02 11:11:40.607873
11128	1-11128	96	53	2800	71	6823	0	2	2012-08-02 11:11:40.607873	2012-08-02 11:11:40.607873
11141	1-11141	20	98	936	27	5506	0	5	2012-08-02 11:11:45.377269	2012-08-02 11:11:45.377269
6333	1-6333	3	34	3556	23	221	221	7	2012-08-02 05:25:01.544543	2012-08-02 11:11:45.377269
9785	1-9785	57	94	4651	83	9661	0	15	2012-08-02 05:48:27.952212	2012-08-02 11:11:46.570756
11144	1-11144	5	7	3817	28	3551	0	2	2012-08-02 11:11:46.570756	2012-08-02 11:11:46.570756
10472	1-10472	16	32	3274	27	4202	0	0	2012-08-02 10:31:45.82615	2012-08-02 11:11:47.753155
10075	1-10075	59	90	8952	81	5596	0	0	2012-08-02 10:21:10.131452	2012-08-02 11:11:47.753155
11147	1-11147	60	81	5049	6	8842	0	12	2012-08-02 11:11:47.753155	2012-08-02 11:11:47.753155
11156	1-11156	90	29	2698	94	4705	0	7	2012-08-02 11:11:51.267894	2012-08-02 11:11:51.267894
9145	1-9145	28	26	3367	35	3108	0	0	2012-08-02 05:44:07.296888	2012-08-02 11:11:53.66609
10695	1-10695	28	2	9285	79	6212	0	0	2012-08-02 10:36:19.209058	2012-08-02 11:11:53.66609
11162	1-11162	66	35	3641	26	6407	0	0	2012-08-02 11:11:53.66609	2012-08-02 11:11:53.66609
9695	1-9695	10	79	7781	85	8920	0	6	2012-08-02 05:47:51.286134	2012-08-02 11:11:54.815166
6437	1-6437	7	22	7384	54	7596	0	0	2012-08-02 05:25:43.988213	2012-08-02 11:11:55.599876
10952	1-10952	97	22	3671	3	3635	0	4	2012-08-02 10:39:03.488592	2012-08-02 11:11:55.599876
9856	1-9856	63	72	4129	54	6580	0	0	2012-08-02 05:48:56.836278	2012-08-02 11:11:55.599876
7612	1-7612	65	22	2268	5	823	0	9	2012-08-02 05:33:40.045239	2012-08-02 11:11:55.599876
11167	1-11167	76	79	1888	22	9897	0	4	2012-08-02 11:11:55.599876	2012-08-02 11:11:55.599876
7023	1-7023	34	44	9530	8	2043	0	0	2012-08-02 05:29:40.895886	2012-08-02 11:12:01.509541
11182	1-11182	95	38	326	44	8291	0	10	2012-08-02 11:12:01.509541	2012-08-02 11:12:01.509541
10822	1-10822	100	47	3569	8	2660	0	0	2012-08-02 10:38:12.296582	2012-08-02 11:12:07.380029
11217	1-11217	32	75	2233	36	5387	0	6	2012-08-02 11:12:15.157959	2012-08-02 11:12:15.157959
11220	1-11220	13	57	3204	72	7285	0	6	2012-08-02 11:12:16.340182	2012-08-02 11:12:16.340182
9946	1-9946	86	75	2224	38	3857	0	5	2012-08-02 05:49:33.559525	2012-08-02 11:12:17.632767
11003	1-11003	45	20	5611	50	8319	0	5	2012-08-02 10:39:23.921828	2012-08-02 11:12:17.632767
9031	1-9031	14	6	6447	85	6268	0	6	2012-08-02 05:43:20.425989	2012-08-02 11:12:18.781677
10803	1-10803	35	6	5015	51	2364	0	0	2012-08-02 10:38:04.882636	2012-08-02 11:12:18.781677
11225	1-11225	20	20	1366	6	8735	0	23	2012-08-02 11:12:18.781677	2012-08-02 11:12:18.781677
11010	1-11010	10	49	1104	73	3058	0	16	2012-08-02 10:39:26.673253	2012-08-02 11:12:19.577338
10514	1-10514	98	26	5445	43	6254	0	9	2012-08-02 10:32:02.462459	2012-08-02 11:12:19.577338
8657	1-8657	56	64	5959	73	6505	3624	0	2012-08-02 05:40:48.272597	2012-08-02 11:12:19.577338
10538	1-10538	4	72	7533	84	4113	0	0	2012-08-02 10:32:11.980567	2012-08-02 11:12:21.908586
11234	1-11234	65	42	2356	54	7775	0	6	2012-08-02 11:12:21.908586	2012-08-02 11:12:21.908586
10668	1-10668	57	81	8216	59	6925	0	3	2012-08-02 10:36:08.564494	2012-08-02 11:12:29.379563
10646	1-10646	22	81	9529	45	8729	0	6	2012-08-02 10:35:59.919627	2012-08-02 11:12:29.379563
11253	1-11253	56	88	2788	81	7763	0	0	2012-08-02 11:12:29.379563	2012-08-02 11:12:29.379563
10717	1-10717	45	2	2173	44	2062	0	0	2012-08-02 10:36:27.794865	2012-08-02 11:12:35.247051
11268	1-11268	90	79	223	2	2927	0	10	2012-08-02 11:12:35.247051	2012-08-02 11:12:35.247051
10105	1-10105	8	29	7393	25	6851	0	0	2012-08-02 10:21:21.990925	2012-08-02 11:12:37.56716
11277	1-11277	60	37	334	49	5437	0	0	2012-08-02 11:12:38.727431	2012-08-02 11:12:38.727431
10770	1-10770	44	69	4718	100	8666	0	7	2012-08-02 10:37:51.911514	2012-08-02 11:12:43.81278
10954	1-10954	79	79	6133	35	4605	0	3	2012-08-02 10:39:04.344189	2012-08-02 11:12:44.972704
11163	1-11163	57	5	4412	88	5881	0	6	2012-08-02 11:11:54.052869	2012-08-02 11:12:44.972704
11293	1-11293	74	88	2390	79	7983	0	11	2012-08-02 11:12:44.972704	2012-08-02 11:12:44.972704
10303	1-10303	81	78	6148	96	8477	0	0	2012-08-02 10:25:38.667486	2012-08-02 11:12:47.725661
10630	1-10630	72	83	8569	72	9089	0	0	2012-08-02 10:35:39.001566	2012-08-02 11:12:52.388119
11312	1-11312	43	22	1025	83	7018	0	0	2012-08-02 11:12:52.388119	2012-08-02 11:12:52.388119
11319	1-11319	79	44	266	58	6013	0	0	2012-08-02 11:12:54.686138	2012-08-02 11:12:54.686138
5611	1-5611	1	9	4252	3	2086	0	2	2012-08-02 05:20:05.545876	2012-08-02 11:13:00.408955
6770	1-6770	34	7	5885	94	1476	0	0	2012-08-02 05:27:58.402499	2012-08-02 11:13:00.408955
11333	1-11333	29	94	248	83	3232	0	20	2012-08-02 11:13:00.408955	2012-08-02 11:13:00.408955
11246	1-11246	54	34	1588	32	3362	0	7	2012-08-02 11:12:27.035793	2012-08-02 11:13:05.082423
10628	1-10628	13	33	2273	6	3245	0	10	2012-08-02 10:35:38.847034	2012-08-02 11:13:05.082423
11351	1-11351	12	57	6	73	6895	0	9	2012-08-02 11:13:07.46978	2012-08-02 11:13:07.46978
7332	1-7332	17	39	6075	29	6573	0	14	2012-08-02 05:31:45.797578	2012-08-02 11:13:08.629456
11357	1-11357	37	30	2244	42	2728	0	0	2012-08-02 11:13:09.78955	2012-08-02 11:13:09.78955
10664	1-10664	90	27	3710	16	5524	0	0	2012-08-02 10:36:06.995377	2012-08-02 11:13:17.41469
11355	1-11355	61	46	2985	62	4176	0	15	2012-08-02 11:13:09.01619	2012-08-02 11:13:18.574619
10975	1-10975	53	89	3270	46	5843	0	0	2012-08-02 10:39:12.566042	2012-08-02 11:13:18.574619
10138	1-10138	17	83	5773	60	7633	0	0	2012-08-02 10:21:35.061801	2012-08-02 11:13:18.574619
11143	1-11143	93	50	4953	95	2220	0	0	2012-08-02 11:11:46.161748	2012-08-02 11:13:20.121622
10691	1-10691	56	52	9071	12	9472	0	16	2012-08-02 10:36:17.650939	2012-08-02 11:13:20.121622
11138	1-11138	47	23	2144	98	4596	0	4	2012-08-02 11:11:44.205675	2012-08-02 11:13:21.30417
11034	1-11034	1	53	3425	17	4722	0	3	2012-08-02 10:39:36.10602	2012-08-02 11:13:21.30417
10554	1-10554	11	26	5016	18	8892	0	6	2012-08-02 10:35:33.266173	2012-08-02 11:13:21.30417
10708	1-10708	64	33	6550	53	7207	0	2	2012-08-02 10:36:24.325927	2012-08-02 11:13:21.30417
11107	1-11107	78	74	4873	59	7854	0	6	2012-08-02 11:11:32.010829	2012-08-02 11:13:22.484345
9477	1-9477	70	59	4368	23	4314	0	0	2012-08-02 05:46:22.205315	2012-08-02 11:13:22.484345
11390	1-11390	33	23	2741	74	7296	0	10	2012-08-02 11:13:22.484345	2012-08-02 11:13:22.484345
6121	1-6121	58	76	3723	92	3618	0	3	2012-08-02 05:23:33.74022	2012-08-02 11:12:55.845883
11321	1-11321	58	92	636	76	3139	0	2	2012-08-02 11:12:55.845883	2012-08-02 11:12:55.845883
11228	1-11228	18	73	5342	26	8384	0	5	2012-08-02 11:12:19.577338	2012-08-02 11:12:59.271544
10144	1-10144	53	31	8505	23	8893	0	3	2012-08-02 10:22:01.573447	2012-08-02 11:13:00.806721
9209	1-9209	78	32	5899	46	2836	0	0	2012-08-02 05:44:33.113506	2012-08-02 11:13:03.182251
11340	1-11340	43	46	1501	74	5578	0	3	2012-08-02 11:13:03.182251	2012-08-02 11:13:03.182251
9092	1-9092	99	59	2441	32	889	761	5	2012-08-02 05:43:45.438735	2012-08-02 11:13:03.182251
11343	1-11343	25	24	1209	36	5445	0	5	2012-08-02 11:13:04.331216	2012-08-02 11:13:04.331216
10575	1-10575	69	37	3036	68	2351	0	5	2012-08-02 10:35:34.848406	2012-08-02 11:13:05.480265
11346	1-11346	73	73	658	37	5383	0	0	2012-08-02 11:13:05.480265	2012-08-02 11:13:05.480265
4510	1-4510	7	49	6562	73	3067	1814	9	2012-08-02 05:12:45.703239	2012-08-02 11:13:05.480265
10850	1-10850	27	51	7576	75	6837	0	9	2012-08-02 10:38:23.335866	2012-08-02 11:13:07.84488
11043	1-11043	29	49	5233	46	7595	0	4	2012-08-02 11:11:07.626669	2012-08-02 11:13:09.01619
11361	1-11361	19	5	5699	66	8285	0	0	2012-08-02 11:13:11.303056	2012-08-02 11:13:11.303056
10238	1-10238	19	9	3684	42	7443	0	4	2012-08-02 10:22:38.279867	2012-08-02 11:13:12.452102
8291	1-8291	7	54	775	19	129	0	2	2012-08-02 05:38:21.07662	2012-08-02 11:13:12.452102
2077	1-2077	62	19	6445	9	269	269	0	2012-08-02 04:56:04.347958	2012-08-02 11:13:12.452102
385	1-385	72	19	9637	9	3671	159	0	2012-08-02 04:44:02.806547	2012-08-02 11:13:12.452102
11264	1-11264	34	95	1678	80	2817	0	10	2012-08-02 11:12:33.677949	2012-08-02 11:13:13.590162
11370	1-11370	53	75	4429	35	4833	0	0	2012-08-02 11:13:14.750192	2012-08-02 11:13:14.750192
10953	1-10953	78	59	3981	32	4046	0	13	2012-08-02 10:39:03.875281	2012-08-02 11:13:17.027691
10982	1-10982	53	83	4459	9	2255	0	2	2012-08-02 10:39:15.296654	2012-08-02 11:13:18.187985
10714	1-10714	45	9	2493	15	4996	0	9	2012-08-02 10:36:26.645899	2012-08-02 11:13:18.187985
7331	1-7331	40	60	8530	15	4930	0	0	2012-08-02 05:31:45.388509	2012-08-02 11:13:18.187985
10442	1-10442	96	52	4058	80	3081	0	0	2012-08-02 10:31:33.773226	2012-08-02 11:13:19.348013
8282	1-8282	46	91	9761	18	4764	0	7	2012-08-02 05:38:17.364554	2012-08-02 11:13:19.348013
11035	1-11035	22	91	4453	21	7239	0	8	2012-08-02 10:39:36.481654	2012-08-02 11:13:19.348013
11382	1-11382	41	76	2294	91	7062	0	10	2012-08-02 11:13:19.348013	2012-08-02 11:13:19.348013
11385	1-11385	73	35	921	14	3993	0	2	2012-08-02 11:13:20.53064	2012-08-02 11:13:20.53064
7967	1-7967	92	18	6057	40	9484	0	0	2012-08-02 05:36:09.632736	2012-08-02 11:13:22.859494
11387	1-11387	84	18	4053	33	8786	0	6	2012-08-02 11:13:21.30417	2012-08-02 11:13:22.859494
10545	1-10545	83	89	3341	28	2517	0	0	2012-08-02 10:35:32.549655	2012-08-02 11:13:22.859494
11391	1-11391	46	40	951	18	9893	0	8	2012-08-02 11:13:22.859494	2012-08-02 11:13:22.859494
11354	1-11354	79	57	9131	39	9570	0	12	2012-08-02 11:13:08.629456	2012-08-02 11:13:27.599334
11379	1-11379	100	15	6513	57	8253	0	0	2012-08-02 11:13:18.187985	2012-08-02 11:13:27.599334
11428	1-11428	19	14	760	76	4574	0	0	2012-08-02 11:13:37.29101	2012-08-02 11:13:37.29101
10845	1-10845	89	73	9679	19	9471	0	0	2012-08-02 10:38:21.347076	2012-08-02 11:13:39.588927
11434	1-11434	93	64	542	73	1404	0	11	2012-08-02 11:13:39.588927	2012-08-02 11:13:39.588927
11345	1-11345	82	32	7379	33	5140	0	16	2012-08-02 11:13:05.082423	2012-08-02 11:13:40.749313
10712	1-10712	48	29	5684	58	8583	0	9	2012-08-02 10:36:25.872596	2012-08-02 11:13:43.14689
11334	1-11334	81	98	2341	39	6347	0	0	2012-08-02 11:13:00.806721	2012-08-02 11:13:48.814959
11474	1-11474	10	97	727	90	2835	0	9	2012-08-02 11:13:54.582459	2012-08-02 11:13:54.582459
11307	1-11307	92	13	6634	65	8691	0	12	2012-08-02 11:12:50.432315	2012-08-02 11:13:56.109039
11483	1-11483	87	62	3164	5	6214	0	3	2012-08-02 11:13:58.407149	2012-08-02 11:13:58.407149
10300	1-10300	40	43	6141	64	4708	0	0	2012-08-02 10:25:37.485441	2012-08-02 11:14:01.11395
11490	1-11490	84	27	309	43	2944	0	0	2012-08-02 11:14:01.11395	2012-08-02 11:14:01.11395
9051	1-9051	20	13	9146	82	6999	0	0	2012-08-02 05:43:28.645242	2012-08-02 11:14:04.649856
11477	1-11477	100	83	6434	13	8468	0	3	2012-08-02 11:13:56.109039	2012-08-02 11:14:04.649856
8445	1-8445	6	25	7187	70	8889	0	0	2012-08-02 05:39:23.134662	2012-08-02 11:14:06.95889
8697	1-8697	92	25	6170	56	1661	0	0	2012-08-02 05:41:04.535074	2012-08-02 11:14:06.95889
11504	1-11504	47	81	1782	25	6353	0	0	2012-08-02 11:14:06.95889	2012-08-02 11:14:06.95889
11510	1-11510	35	70	1037	52	4083	0	7	2012-08-02 11:14:08.881126	2012-08-02 11:14:08.881126
9696	1-9696	72	24	2092	93	1575	0	0	2012-08-02 05:47:51.694758	2012-08-02 11:14:10.041316
11129	1-11129	13	33	4005	49	4748	0	7	2012-08-02 11:11:40.232123	2012-08-02 11:14:10.825701
11515	1-11515	33	20	1846	97	8279	0	0	2012-08-02 11:14:10.825701	2012-08-02 11:14:10.825701
11384	1-11384	19	95	2331	52	6827	0	9	2012-08-02 11:13:20.121622	2012-08-02 11:14:13.157551
10827	1-10827	63	95	5880	56	3493	0	0	2012-08-02 10:38:14.298653	2012-08-02 11:14:13.157551
11459	1-11459	39	49	1230	20	5186	0	8	2012-08-02 11:13:49.190803	2012-08-02 11:14:13.157551
10869	1-10869	43	95	5758	4	9736	0	0	2012-08-02 10:38:31.213882	2012-08-02 11:14:13.157551
11521	1-11521	10	20	1784	95	8193	0	3	2012-08-02 11:14:13.157551	2012-08-02 11:14:13.157551
10290	1-10290	91	82	7745	32	3683	0	6	2012-08-02 10:25:33.540521	2012-08-02 11:14:16.615163
3656	1-3656	24	82	7112	10	3720	0	0	2012-08-02 05:07:04.576561	2012-08-02 11:14:16.615163
10430	1-10430	8	96	3258	19	1721	0	10	2012-08-02 10:27:56.384395	2012-08-02 11:14:17.764358
11429	1-11429	60	96	1903	57	2513	0	5	2012-08-02 11:13:37.678086	2012-08-02 11:14:17.764358
11533	1-11533	57	59	676	96	7584	0	4	2012-08-02 11:14:17.764358	2012-08-02 11:14:17.764358
10162	1-10162	42	4	2387	5	4379	0	0	2012-08-02 10:22:08.577513	2012-08-02 11:14:20.825378
10494	1-10494	40	42	5924	17	2228	0	3	2012-08-02 10:31:54.561415	2012-08-02 11:14:20.825378
11541	1-11541	80	91	2547	4	5109	0	0	2012-08-02 11:14:20.825378	2012-08-02 11:14:20.825378
5228	1-5228	22	61	9636	1	4050	0	0	2012-08-02 05:17:34.000707	2012-08-02 11:14:21.985338
11544	1-11544	55	1	1187	61	8452	0	14	2012-08-02 11:14:21.985338	2012-08-02 11:14:21.985338
11149	1-11149	5	65	3403	97	4865	0	4	2012-08-02 11:11:48.526807	2012-08-02 11:14:23.136105
10746	1-10746	52	20	7265	56	2547	0	11	2012-08-02 10:37:43.059846	2012-08-02 11:14:23.136105
11281	1-11281	25	25	4375	16	6255	0	5	2012-08-02 11:12:40.274544	2012-08-02 11:14:26.561211
10876	1-10876	79	14	4277	92	5363	0	6	2012-08-02 10:38:33.579929	2012-08-02 11:14:27.743563
11559	1-11559	97	92	838	19	5398	0	0	2012-08-02 11:14:27.743563	2012-08-02 11:14:27.743563
11013	1-11013	64	21	7695	93	8871	0	9	2012-08-02 10:39:27.841229	2012-08-02 11:14:30.008411
11568	1-11568	68	84	786	83	3510	0	3	2012-08-02 11:14:31.158049	2012-08-02 11:14:31.158049
11577	1-11577	35	6	1460	68	1631	0	3	2012-08-02 11:14:34.626495	2012-08-02 11:14:34.626495
4196	1-4196	69	61	1329	93	400	0	0	2012-08-02 05:10:39.693733	2012-08-02 11:14:35.77558
10548	1-10548	59	71	4099	47	2511	0	12	2012-08-02 10:35:32.784245	2012-08-02 11:14:35.77558
11005	1-11005	61	71	8336	41	7918	0	0	2012-08-02 10:39:24.69726	2012-08-02 11:14:35.77558
9834	1-9834	32	10	5707	52	5879	0	0	2012-08-02 05:48:47.841851	2012-08-02 11:13:24.793267
11274	1-11274	78	25	7103	29	7694	0	4	2012-08-02 11:12:37.56716	2012-08-02 11:13:30.75082
11410	1-11410	59	72	404	25	3355	0	3	2012-08-02 11:13:30.75082	2012-08-02 11:13:30.75082
11412	1-11412	82	91	83	21	4808	0	11	2012-08-02 11:13:31.551961	2012-08-02 11:13:31.551961
11172	1-11172	93	71	1196	54	2632	0	2	2012-08-02 11:11:57.577745	2012-08-02 11:13:34.572958
11421	1-11421	78	1	208	69	1572	0	7	2012-08-02 11:13:34.572958	2012-08-02 11:13:34.572958
11376	1-11376	35	32	1929	17	4454	0	6	2012-08-02 11:13:17.027691	2012-08-02 11:13:35.722017
10465	1-10465	49	91	2943	58	4040	0	4	2012-08-02 10:31:43.01941	2012-08-02 11:13:35.722017
10512	1-10512	46	57	4554	58	8597	0	0	2012-08-02 10:32:01.666719	2012-08-02 11:13:35.722017
11338	1-11338	3	34	2035	50	5359	0	11	2012-08-02 11:13:02.386499	2012-08-02 11:13:45.798923
11067	1-11067	18	20	4028	67	4518	0	0	2012-08-02 11:11:16.450807	2012-08-02 11:13:45.798923
11236	1-11236	63	20	1526	50	9197	0	9	2012-08-02 11:12:22.704417	2012-08-02 11:13:49.190803
11462	1-11462	42	87	11	17	7533	0	0	2012-08-02 11:13:50.361925	2012-08-02 11:13:50.361925
8001	1-8001	69	4	8215	60	9748	0	0	2012-08-02 05:36:23.520617	2012-08-02 11:13:52.637974
11244	1-11244	39	100	5522	39	6129	0	0	2012-08-02 11:12:25.864716	2012-08-02 11:13:52.637974
11021	1-11021	19	19	4710	69	6427	0	0	2012-08-02 10:39:30.996819	2012-08-02 11:13:53.797887
11471	1-11471	40	69	6634	19	7195	0	7	2012-08-02 11:13:53.797887	2012-08-02 11:13:53.797887
11148	1-11148	92	21	7939	75	6083	0	0	2012-08-02 11:11:48.140241	2012-08-02 11:13:56.871467
8170	1-8170	65	89	3075	84	1750	0	0	2012-08-02 05:37:31.849236	2012-08-02 11:13:58.03155
11491	1-11491	35	60	311	65	3627	0	0	2012-08-02 11:14:01.500582	2012-08-02 11:14:01.500582
10961	1-10961	42	85	1510	17	1370	0	5	2012-08-02 10:39:07.081103	2012-08-02 11:14:07.334501
10532	1-10532	32	57	4727	42	5441	0	3	2012-08-02 10:32:09.538171	2012-08-02 11:14:07.334501
9985	1-9985	22	75	3991	42	6451	0	0	2012-08-02 05:49:49.617032	2012-08-02 11:14:11.212304
10700	1-10700	30	49	8886	3	9767	0	0	2012-08-02 10:36:21.221225	2012-08-02 11:14:11.212304
10926	1-10926	70	35	5145	8	9267	0	6	2012-08-02 10:38:53.203026	2012-08-02 11:14:12.394694
5805	1-5805	36	47	6246	79	3018	0	0	2012-08-02 05:21:24.483007	2012-08-02 11:14:19.278342
11304	1-11304	83	42	8034	30	8527	0	5	2012-08-02 11:12:49.283536	2012-08-02 11:14:19.278342
11537	1-11537	19	79	4095	14	6283	0	0	2012-08-02 11:14:19.278342	2012-08-02 11:14:19.278342
11330	1-11330	80	92	3000	38	1290	0	3	2012-08-02 11:12:59.657606	2012-08-02 11:14:20.438204
11502	1-11502	98	14	6875	56	4631	0	5	2012-08-02 11:14:05.787484	2012-08-02 11:14:22.747415
8826	1-8826	69	14	2867	69	880	0	0	2012-08-02 05:41:56.73864	2012-08-02 11:14:22.747415
9586	1-9586	72	14	8205	42	7936	0	2	2012-08-02 05:47:06.460065	2012-08-02 11:14:22.747415
11493	1-11493	77	55	3018	10	3268	0	0	2012-08-02 11:14:02.296175	2012-08-02 11:14:26.174917
6224	1-6224	21	67	8519	38	5865	0	4	2012-08-02 05:24:15.699581	2012-08-02 11:14:30.77091
11566	1-11566	95	56	92	67	8120	0	9	2012-08-02 11:14:30.77091	2012-08-02 11:14:30.77091
4729	1-4729	83	45	9462	53	7822	0	0	2012-08-02 05:14:14.498266	2012-08-02 11:14:39.598138
11590	1-11590	33	37	2028	10	2908	0	10	2012-08-02 11:14:39.598138	2012-08-02 11:14:39.598138
9997	1-9997	41	16	8412	35	4997	0	11	2012-08-02 05:49:54.578018	2012-08-02 11:14:40.747269
11593	1-11593	95	35	1652	31	5650	0	7	2012-08-02 11:14:40.747269	2012-08-02 11:14:40.747269
2950	1-2950	47	52	7781	35	410	410	0	2012-08-02 05:02:18.95289	2012-08-02 11:14:40.747269
11199	1-11199	87	49	5577	47	9425	0	0	2012-08-02 11:12:07.380029	2012-08-02 11:14:41.8964
11062	1-11062	55	93	886	48	666	0	10	2012-08-02 11:11:14.449724	2012-08-02 11:14:43.034382
10516	1-10516	31	93	8325	64	6012	0	0	2012-08-02 10:32:03.247037	2012-08-02 11:14:43.034382
11598	1-11598	6	70	1239	93	7963	0	6	2012-08-02 11:14:43.034382	2012-08-02 11:14:43.034382
1180	1-1180	9	93	8540	48	2047	1919	0	2012-08-02 04:48:47.918974	2012-08-02 11:14:43.034382
10596	1-10596	16	67	5554	48	8790	0	0	2012-08-02 10:35:36.436215	2012-08-02 11:14:44.957993
5772	1-5772	23	99	4095	40	5371	0	0	2012-08-02 05:21:11.346959	2012-08-02 11:14:47.289697
8305	1-8305	48	99	2819	5	3572	0	0	2012-08-02 05:38:26.668708	2012-08-02 11:14:47.289697
5863	1-5863	72	99	9999	90	6426	0	0	2012-08-02 05:21:48.039993	2012-08-02 11:14:47.289697
11610	1-11610	47	17	3534	99	9802	0	4	2012-08-02 11:14:47.289697	2012-08-02 11:14:47.289697
2632	1-2632	35	61	9511	17	6357	3891	0	2012-08-02 05:00:05.432314	2012-08-02 11:14:47.289697
11617	1-11617	37	80	301	99	1479	0	4	2012-08-02 11:14:49.633653	2012-08-02 11:14:49.633653
11364	1-11364	63	19	1998	9	6063	0	10	2012-08-02 11:13:12.452102	2012-08-02 11:14:53.478813
9581	1-9581	80	81	5927	58	7768	0	2	2012-08-02 05:47:04.412682	2012-08-02 11:14:53.478813
10569	1-10569	8	59	9173	58	8376	0	7	2012-08-02 10:35:34.401627	2012-08-02 11:14:53.478813
8921	1-8921	6	67	3981	58	4515	0	4	2012-08-02 05:42:35.244645	2012-08-02 11:14:53.478813
10948	1-10948	25	95	6466	25	6494	0	11	2012-08-02 10:39:01.942098	2012-08-02 11:14:59.290105
9306	1-9306	94	72	4747	6	2256	0	12	2012-08-02 05:45:12.68035	2012-08-02 11:14:59.290105
11641	1-11641	51	22	2527	72	9828	0	4	2012-08-02 11:14:59.290105	2012-08-02 11:14:59.290105
8748	1-8748	27	99	8376	98	9344	0	8	2012-08-02 05:41:25.1495	2012-08-02 11:15:02.79246
11656	1-11656	88	16	3500	90	6097	0	8	2012-08-02 11:15:05.113014	2012-08-02 11:15:05.113014
10550	1-10550	27	78	5421	13	5521	0	7	2012-08-02 10:35:32.953534	2012-08-02 11:15:06.297318
6860	1-6860	89	77	6731	48	3557	0	7	2012-08-02 05:28:34.932678	2012-08-02 11:15:08.648528
11506	1-11506	88	48	4416	57	7358	0	6	2012-08-02 11:14:07.334501	2012-08-02 11:15:08.648528
10403	1-10403	54	98	8146	25	7523	0	6	2012-08-02 10:27:45.755174	2012-08-02 11:15:09.819771
11256	1-11256	21	98	6945	26	5737	0	8	2012-08-02 11:12:30.893329	2012-08-02 11:15:09.819771
11668	1-11668	7	25	2584	74	8462	0	0	2012-08-02 11:15:09.819771	2012-08-02 11:15:09.819771
3794	1-3794	42	98	8144	25	2223	736	0	2012-08-02 05:07:59.280997	2012-08-02 11:15:09.819771
10853	1-10853	36	37	4436	36	884	0	0	2012-08-02 10:38:24.495965	2012-08-02 11:15:10.991124
7526	1-7526	23	77	6127	85	5683	0	6	2012-08-02 05:33:04.93129	2012-08-02 11:15:12.322084
11497	1-11497	48	40	7121	65	5278	0	8	2012-08-02 11:14:03.864958	2012-08-02 11:15:12.322084
11024	1-11024	76	82	7035	34	4669	0	0	2012-08-02 10:39:32.170203	2012-08-02 11:15:13.522864
11499	1-11499	25	82	4845	77	6714	0	6	2012-08-02 11:14:04.649856	2012-08-02 11:15:13.522864
11677	1-11677	84	85	279	82	9952	0	2	2012-08-02 11:15:13.522864	2012-08-02 11:15:13.522864
11680	1-11680	85	32	2277	69	4121	0	16	2012-08-02 11:15:14.700276	2012-08-02 11:15:14.700276
11580	1-11580	1	8	2471	71	7985	0	0	2012-08-02 11:14:35.77558	2012-08-02 11:15:17.02225
11674	1-11674	35	65	2617	23	3750	0	0	2012-08-02 11:15:12.322084	2012-08-02 11:15:18.234471
11166	1-11166	22	63	1621	42	6022	0	0	2012-08-02 11:11:55.202188	2012-08-02 11:15:18.234471
11372	1-11372	70	17	3402	48	4808	0	2	2012-08-02 11:13:15.137154	2012-08-02 11:15:18.234471
8885	1-8885	72	1	8007	42	9378	0	11	2012-08-02 05:42:20.66137	2012-08-02 11:15:18.234471
6111	1-6111	69	42	9935	65	2052	1720	0	2012-08-02 05:23:29.729462	2012-08-02 11:15:18.234471
10694	1-10694	36	6	5327	12	5045	0	8	2012-08-02 10:36:18.822548	2012-08-02 11:13:26.705297
11401	1-11401	41	22	1703	6	9766	0	2	2012-08-02 11:13:26.705297	2012-08-02 11:13:26.705297
10856	1-10856	60	12	7804	23	7757	0	0	2012-08-02 10:38:25.303061	2012-08-02 11:13:28.01982
11411	1-11411	92	5	19	78	2930	0	0	2012-08-02 11:13:30.374755	2012-08-02 11:13:30.374755
11414	1-11414	7	17	19	29	865	0	3	2012-08-02 11:13:31.932553	2012-08-02 11:13:31.932553
11423	1-11423	21	81	641	61	4584	0	0	2012-08-02 11:13:35.33513	2012-08-02 11:13:35.33513
10899	1-10899	57	28	6086	94	9160	0	0	2012-08-02 10:38:42.553019	2012-08-02 11:13:39.975895
11309	1-11309	59	80	6250	59	6502	0	3	2012-08-02 11:12:51.205721	2012-08-02 11:13:51.886711
11454	1-11454	72	71	1799	78	1054	0	6	2012-08-02 11:13:47.323583	2012-08-02 11:13:51.886711
11466	1-11466	90	78	769	80	9350	0	0	2012-08-02 11:13:51.886711	2012-08-02 11:13:51.886711
9841	1-9841	57	33	3639	91	5744	0	0	2012-08-02 05:48:50.724877	2012-08-02 11:13:53.035415
6632	1-6632	16	49	5653	98	1694	0	7	2012-08-02 05:27:03.326424	2012-08-02 11:13:53.035415
11469	1-11469	40	98	640	33	9583	0	7	2012-08-02 11:13:53.035415	2012-08-02 11:13:53.035415
11472	1-11472	5	40	502	63	3771	0	5	2012-08-02 11:13:54.195816	2012-08-02 11:13:54.195816
11347	1-11347	20	8	3002	18	3622	0	5	2012-08-02 11:13:05.877969	2012-08-02 11:13:55.356002
11061	1-11061	52	8	3414	43	6167	0	0	2012-08-02 11:11:14.072843	2012-08-02 11:13:55.356002
11004	1-11004	37	8	8730	75	8251	0	0	2012-08-02 10:39:24.309233	2012-08-02 11:13:55.356002
11475	1-11475	57	12	1862	8	8210	0	4	2012-08-02 11:13:55.356002	2012-08-02 11:13:55.356002
11492	1-11492	88	24	788	65	6169	0	6	2012-08-02 11:14:01.887571	2012-08-02 11:14:01.887571
7475	1-7475	54	55	7570	56	3872	0	0	2012-08-02 05:32:43.859888	2012-08-02 11:14:03.080588
11495	1-11495	69	24	555	98	4481	0	6	2012-08-02 11:14:03.080588	2012-08-02 11:14:03.080588
10716	1-10716	42	84	5653	29	6165	0	3	2012-08-02 10:36:27.408411	2012-08-02 11:14:08.107531
10357	1-10357	78	23	1863	40	4095	0	5	2012-08-02 10:27:27.475836	2012-08-02 11:14:08.107531
10758	1-10758	79	77	5543	84	2019	0	5	2012-08-02 10:37:47.853592	2012-08-02 11:14:11.587941
11242	1-11242	45	77	6401	83	6335	0	0	2012-08-02 11:12:25.058232	2012-08-02 11:14:11.587941
11517	1-11517	48	42	2705	77	7590	0	0	2012-08-02 11:14:11.587941	2012-08-02 11:14:11.587941
10279	1-10279	78	33	6040	39	9305	0	0	2012-08-02 10:25:29.198472	2012-08-02 11:14:13.930346
11078	1-11078	41	41	7819	43	7779	0	7	2012-08-02 11:11:20.766065	2012-08-02 11:14:18.515468
10068	1-10068	50	71	5130	59	3192	0	4	2012-08-02 10:21:07.410159	2012-08-02 11:14:18.515468
9930	1-9930	14	71	5367	44	8977	0	5	2012-08-02 05:49:27.028671	2012-08-02 11:14:18.515468
11536	1-11536	13	90	2225	71	8331	0	6	2012-08-02 11:14:18.515468	2012-08-02 11:14:18.515468
11068	1-11068	6	54	3156	87	4641	0	0	2012-08-02 11:11:16.837882	2012-08-02 11:14:23.511993
9374	1-9374	59	91	6819	52	8489	0	8	2012-08-02 05:45:40.403421	2012-08-02 11:14:24.672122
11540	1-11540	22	38	2285	51	2406	0	0	2012-08-02 11:14:20.438204	2012-08-02 11:14:29.246334
10608	1-10608	88	12	9317	13	7450	0	0	2012-08-02 10:35:37.324285	2012-08-02 11:14:29.246334
8933	1-8933	21	23	6187	5	9514	0	0	2012-08-02 05:42:40.172214	2012-08-02 11:14:35.399738
9138	1-9138	82	63	7045	55	9110	0	11	2012-08-02 05:44:04.436267	2012-08-02 11:14:35.399738
11597	1-11597	36	52	450	24	5080	0	2	2012-08-02 11:14:42.271923	2012-08-02 11:14:42.271923
9068	1-9068	42	59	7080	85	9774	0	3	2012-08-02 05:43:35.572085	2012-08-02 11:14:44.218345
10870	1-10870	63	57	3701	83	6990	0	9	2012-08-02 10:38:30.8271	2012-08-02 11:14:44.218345
11602	1-11602	94	85	3497	59	6848	0	0	2012-08-02 11:14:44.218345	2012-08-02 11:14:44.218345
11105	1-11105	67	51	5933	36	8690	0	3	2012-08-02 11:11:31.223945	2012-08-02 11:14:45.344938
11605	1-11605	73	77	2518	67	6308	0	9	2012-08-02 11:14:45.344938	2012-08-02 11:14:45.344938
10505	1-10505	67	89	6838	80	8573	0	0	2012-08-02 10:31:58.914662	2012-08-02 11:14:48.849192
11237	1-11237	13	3	1635	61	2297	0	0	2012-08-02 11:12:23.091009	2012-08-02 11:14:51.954176
11516	1-11516	29	3	4130	75	4298	0	0	2012-08-02 11:14:11.212304	2012-08-02 11:14:51.954176
11622	1-11622	32	64	1356	3	7016	0	0	2012-08-02 11:14:51.954176	2012-08-02 11:14:51.954176
11625	1-11625	71	25	342	45	2325	0	0	2012-08-02 11:14:53.091797	2012-08-02 11:14:53.091797
11367	1-11367	43	80	6737	76	6483	0	0	2012-08-02 11:13:13.590162	2012-08-02 11:14:54.262813
9245	1-9245	47	84	9072	72	9055	0	0	2012-08-02 05:44:47.783324	2012-08-02 11:14:55.42302
9064	1-9064	29	25	9374	47	3586	0	13	2012-08-02 05:43:33.937189	2012-08-02 11:14:55.42302
11255	1-11255	59	84	4445	22	5678	0	0	2012-08-02 11:12:30.119872	2012-08-02 11:14:55.42302
11631	1-11631	31	47	2277	84	6601	0	9	2012-08-02 11:14:55.42302	2012-08-02 11:14:55.42302
11269	1-11269	67	72	4563	95	7310	0	13	2012-08-02 11:12:35.633704	2012-08-02 11:15:00.836972
10273	1-10273	18	72	7697	65	6403	0	0	2012-08-02 10:25:26.853194	2012-08-02 11:15:00.836972
11649	1-11649	14	50	627	27	1747	0	3	2012-08-02 11:15:02.428143	2012-08-02 11:15:02.428143
4265	1-4265	18	30	5810	100	4018	3350	3	2012-08-02 05:11:06.723644	2012-08-02 11:15:02.428143
11626	1-11626	47	58	6220	19	7402	0	6	2012-08-02 11:14:53.478813	2012-08-02 11:15:03.566223
6191	1-6191	17	8	7130	60	8733	0	0	2012-08-02 05:24:01.985562	2012-08-02 11:15:04.737407
8856	1-8856	67	63	5833	89	6282	0	0	2012-08-02 05:42:08.826816	2012-08-02 11:15:04.737407
10828	1-10828	86	22	6440	76	4772	0	4	2012-08-02 10:38:14.718704	2012-08-02 11:15:04.737407
11655	1-11655	90	76	4718	8	9388	0	3	2012-08-02 11:15:04.737407	2012-08-02 11:15:04.737407
9363	1-9363	80	24	2929	74	1625	0	2	2012-08-02 05:45:35.973259	2012-08-02 11:15:09.432839
10768	1-10768	93	74	5323	75	9489	0	3	2012-08-02 10:37:51.12684	2012-08-02 11:15:09.432839
8159	1-8159	37	43	5078	75	1770	0	11	2012-08-02 05:37:27.416549	2012-08-02 11:15:09.432839
11667	1-11667	96	75	3935	24	9329	0	3	2012-08-02 11:15:09.432839	2012-08-02 11:15:09.432839
9495	1-9495	76	78	9427	43	6658	0	0	2012-08-02 05:46:29.475141	2012-08-02 11:15:10.604364
11670	1-11670	72	100	581	78	9114	0	11	2012-08-02 11:15:10.604364	2012-08-02 11:15:10.604364
10896	1-10896	27	2	3951	14	5151	0	0	2012-08-02 10:38:40.984077	2012-08-02 11:15:12.709591
11055	1-11055	94	39	7388	75	6194	0	0	2012-08-02 11:11:12.104962	2012-08-02 11:15:16.259253
6611	1-6611	95	39	8354	42	8571	0	17	2012-08-02 05:26:54.730671	2012-08-02 11:15:16.259253
8251	1-8251	46	39	5345	83	5467	0	0	2012-08-02 05:38:04.757448	2012-08-02 11:15:16.259253
3689	1-3689	57	1	5498	58	4393	0	0	2012-08-02 05:07:17.859411	2012-08-02 11:15:16.259253
11170	1-11170	64	84	1673	42	4977	0	0	2012-08-02 11:11:56.804494	2012-08-02 11:15:16.259253
11684	1-11684	59	38	673	39	9906	0	0	2012-08-02 11:15:16.259253	2012-08-02 11:15:16.259253
3501	1-3501	86	75	9493	38	6843	4755	0	2012-08-02 05:06:02.190692	2012-08-02 11:15:16.259253
11703	1-11703	8	81	2862	18	4590	0	0	2012-08-02 11:15:23.743163	2012-08-02 11:15:23.743163
11621	1-11621	96	96	1446	87	5356	0	8	2012-08-02 11:14:51.556199	2012-08-02 11:15:28.432689
11048	1-11048	66	46	6999	70	6340	0	0	2012-08-02 11:11:09.445015	2012-08-02 11:15:28.432689
11715	1-11715	85	70	7236	96	9879	0	9	2012-08-02 11:15:28.432689	2012-08-02 11:15:28.432689
4574	1-4574	28	96	6204	87	984	984	0	2012-08-02 05:13:11.296161	2012-08-02 11:15:28.432689
5628	1-5628	34	47	4105	37	3299	0	0	2012-08-02 05:20:12.519164	2012-08-02 11:14:39.222523
11029	1-11029	69	94	9588	27	7281	0	0	2012-08-02 10:39:34.15049	2012-08-02 11:14:41.498714
6628	1-6628	7	39	8560	60	9519	0	10	2012-08-02 05:27:01.724252	2012-08-02 11:14:46.891545
11523	1-11523	94	39	512	46	2086	0	8	2012-08-02 11:14:13.930346	2012-08-02 11:14:46.891545
11609	1-11609	25	60	1041	39	2962	0	12	2012-08-02 11:14:46.891545	2012-08-02 11:14:46.891545
10211	1-10211	63	87	7500	44	7910	0	7	2012-08-02 10:22:27.693123	2012-08-02 11:14:51.556199
10272	1-10272	31	95	7368	87	9860	0	9	2012-08-02 10:25:26.466128	2012-08-02 11:14:55.047636
11629	1-11629	51	87	2735	71	3000	0	5	2012-08-02 11:14:55.047636	2012-08-02 11:14:55.047636
6975	1-6975	64	91	3015	92	4199	0	0	2012-08-02 05:29:21.900522	2012-08-02 11:14:55.820897
11329	1-11329	23	37	3999	92	3862	0	0	2012-08-02 11:12:58.895419	2012-08-02 11:14:55.820897
11443	1-11443	32	92	5355	12	9326	0	0	2012-08-02 11:13:43.14689	2012-08-02 11:14:55.820897
11632	1-11632	64	12	3027	91	5383	0	0	2012-08-02 11:14:55.820897	2012-08-02 11:14:59.665625
9360	1-9360	47	16	9664	31	6526	0	7	2012-08-02 05:45:34.757966	2012-08-02 11:15:01.245814
11645	1-11645	96	31	921	82	8003	0	9	2012-08-02 11:15:01.245814	2012-08-02 11:15:01.245814
11130	1-11130	82	17	8448	18	6451	0	7	2012-08-02 11:11:41.00554	2012-08-02 11:15:02.041455
10132	1-10132	79	3	6761	90	8247	0	3	2012-08-02 10:21:32.695035	2012-08-02 11:15:02.041455
10175	1-10175	26	28	6973	18	4309	0	5	2012-08-02 10:22:13.693367	2012-08-02 11:15:02.041455
11648	1-11648	34	90	4564	28	8961	0	10	2012-08-02 11:15:02.041455	2012-08-02 11:15:02.041455
10463	1-10463	93	90	9711	28	9795	23	2	2012-08-02 10:31:42.232372	2012-08-02 11:15:02.041455
10684	1-10684	29	93	6397	8	8639	0	0	2012-08-02 10:36:14.87698	2012-08-02 11:15:04.35059
10949	1-10949	73	5	7749	94	7993	0	8	2012-08-02 10:39:02.328595	2012-08-02 11:15:05.510735
11424	1-11424	81	58	9445	32	9542	0	0	2012-08-02 11:13:35.722017	2012-08-02 11:15:11.399837
11673	1-11673	95	45	435	58	8160	0	0	2012-08-02 11:15:11.399837	2012-08-02 11:15:11.399837
8542	1-8542	41	6	7497	60	8956	0	4	2012-08-02 05:40:01.951158	2012-08-02 11:15:16.6464
11685	1-11685	27	89	2419	6	6142	0	0	2012-08-02 11:15:16.6464	2012-08-02 11:15:16.6464
11671	1-11671	80	62	1627	37	6998	0	0	2012-08-02 11:15:10.991124	2012-08-02 11:15:17.845583
11381	1-11381	68	60	6083	89	7506	0	0	2012-08-02 11:13:18.574619	2012-08-02 11:15:17.845583
7923	1-7923	49	62	4445	15	3860	0	0	2012-08-02 05:35:51.079608	2012-08-02 11:15:17.845583
10467	1-10467	3	95	2492	33	980	0	6	2012-08-02 10:31:43.803666	2012-08-02 11:15:22.172465
11572	1-11572	30	54	4676	5	6482	0	4	2012-08-02 11:14:32.715491	2012-08-02 11:15:23.356133
9265	1-9265	73	24	7782	31	8192	0	0	2012-08-02 05:44:55.992934	2012-08-02 11:15:23.356133
9838	1-9838	1	86	5755	34	6839	0	16	2012-08-02 05:48:49.503738	2012-08-02 11:15:24.518882
7534	1-7534	50	9	8183	34	6007	0	2	2012-08-02 05:33:08.269065	2012-08-02 11:15:24.518882
11018	1-11018	44	1	2621	51	5956	0	0	2012-08-02 10:39:29.809232	2012-08-02 11:15:24.518882
11448	1-11448	27	2	3951	12	3866	0	0	2012-08-02 11:13:45.047691	2012-08-02 11:15:29.184735
10230	1-10230	95	41	9490	12	9188	0	3	2012-08-02 10:22:35.15301	2012-08-02 11:15:32.700252
11726	1-11726	49	86	4042	41	7327	0	3	2012-08-02 11:15:32.700252	2012-08-02 11:15:32.700252
5302	1-5302	51	41	9560	58	8817	0	0	2012-08-02 05:18:03.04833	2012-08-02 11:15:33.874328
11547	1-11547	84	33	1019	20	8466	0	5	2012-08-02 11:14:23.136105	2012-08-02 11:15:33.874328
11729	1-11729	99	79	2163	33	2423	0	13	2012-08-02 11:15:33.874328	2012-08-02 11:15:33.874328
8484	1-8484	80	8	6167	64	3825	0	0	2012-08-02 05:39:38.70347	2012-08-02 11:15:44.03822
5429	1-5429	85	24	7791	64	3813	0	2	2012-08-02 05:18:52.941988	2012-08-02 11:15:44.03822
6203	1-6203	28	62	5810	87	4281	0	0	2012-08-02 05:24:06.95632	2012-08-02 11:15:44.03822
11755	1-11755	71	87	3674	8	8874	0	15	2012-08-02 11:15:44.03822	2012-08-02 11:15:44.03822
10967	1-10967	45	29	6007	57	8087	0	0	2012-08-02 10:39:09.04928	2012-08-02 11:15:48.789614
5937	1-5937	60	99	9212	36	5782	0	0	2012-08-02 05:22:18.271138	2012-08-02 11:15:48.789614
11117	1-11117	8	95	5769	75	1400	0	0	2012-08-02 11:11:35.923348	2012-08-02 11:15:48.789614
11768	1-11768	52	75	797	99	4855	0	0	2012-08-02 11:15:48.789614	2012-08-02 11:15:48.789614
11770	1-11770	76	5	201	83	2404	0	0	2012-08-02 11:15:50.35883	2012-08-02 11:15:50.35883
11779	1-11779	11	85	1945	98	6899	0	0	2012-08-02 11:15:53.518543	2012-08-02 11:15:53.518543
11789	1-11789	52	41	794	88	9264	0	6	2012-08-02 11:15:57.442773	2012-08-02 11:15:57.442773
11534	1-11534	42	40	6496	22	5627	0	0	2012-08-02 11:14:18.139902	2012-08-02 11:16:06.857917
11595	1-11595	83	27	8098	30	6294	0	4	2012-08-02 11:14:41.498714	2012-08-02 11:16:09.277745
5936	1-5936	21	17	5442	93	1159	0	0	2012-08-02 05:22:17.862128	2012-08-02 11:16:09.277745
9901	1-9901	86	53	5565	60	7691	0	6	2012-08-02 05:49:15.234663	2012-08-02 11:16:09.277745
11819	1-11819	38	93	166	53	7061	0	4	2012-08-02 11:16:09.277745	2012-08-02 11:16:09.277745
8216	1-8216	99	97	2597	78	771	0	0	2012-08-02 05:37:50.632018	2012-08-02 11:16:11.609493
11825	1-11825	45	78	627	97	8522	0	4	2012-08-02 11:16:11.609493	2012-08-02 11:16:11.609493
11186	1-11186	8	5	8226	25	6305	0	0	2012-08-02 11:12:03.115284	2012-08-02 11:16:13.90738
11812	1-11812	37	84	1866	82	6652	0	6	2012-08-02 11:16:06.857917	2012-08-02 11:16:13.90738
10383	1-10383	59	43	7183	90	8368	0	6	2012-08-02 10:27:37.822039	2012-08-02 11:16:15.045628
11834	1-11834	16	90	225	10	3363	0	4	2012-08-02 11:16:15.045628	2012-08-02 11:16:15.045628
11787	1-11787	90	66	1491	67	5138	0	7	2012-08-02 11:15:56.66897	2012-08-02 11:16:17.321403
11840	1-11840	1	21	2472	11	9698	0	0	2012-08-02 11:16:17.321403	2012-08-02 11:16:17.321403
11221	1-11221	64	13	6793	8	5865	0	3	2012-08-02 11:12:16.727211	2012-08-02 11:16:18.481743
11844	1-11844	78	68	2171	13	5372	0	16	2012-08-02 11:16:18.481743	2012-08-02 11:16:18.481743
11853	1-11853	60	17	2151	65	3680	0	7	2012-08-02 11:16:22.338424	2012-08-02 11:16:22.338424
11285	1-11285	68	80	6271	45	7812	0	8	2012-08-02 11:12:41.821404	2012-08-02 11:16:25.80742
7839	1-7839	53	45	7897	6	4774	3444	0	2012-08-02 05:35:15.569535	2012-08-02 11:16:25.80742
9171	1-9171	81	91	8072	95	6557	0	0	2012-08-02 05:44:17.775217	2012-08-02 11:16:26.593607
7302	1-7302	63	43	5965	95	2784	0	4	2012-08-02 05:31:33.987399	2012-08-02 11:16:26.593607
11864	1-11864	70	95	4138	43	9559	0	13	2012-08-02 11:16:26.593607	2012-08-02 11:16:26.593607
3242	1-3242	61	43	8416	51	4555	696	0	2012-08-02 05:04:17.23107	2012-08-02 11:16:26.593607
10488	1-10488	89	29	2472	67	1908	0	7	2012-08-02 10:31:52.181436	2012-08-02 11:16:28.924881
11146	1-11146	38	62	4269	82	5849	0	5	2012-08-02 11:11:47.344274	2012-08-02 11:16:28.924881
11870	1-11870	88	67	1107	62	8181	0	4	2012-08-02 11:16:28.924881	2012-08-02 11:16:28.924881
11479	1-11479	92	75	3240	59	6588	0	6	2012-08-02 11:13:56.871467	2012-08-02 11:16:30.07419
10522	1-10522	32	51	7402	3	5403	0	5	2012-08-02 10:32:05.601007	2012-08-02 11:16:33.134849
11806	1-11806	15	57	3435	69	6610	0	8	2012-08-02 11:16:04.085157	2012-08-02 11:16:33.918993
11658	1-11658	18	99	8870	29	9879	0	5	2012-08-02 11:15:05.908677	2012-08-02 11:16:40.051362
11899	1-11899	66	29	3095	99	4948	0	0	2012-08-02 11:16:40.051362	2012-08-02 11:16:40.051362
11665	1-11665	61	57	3717	62	7627	0	2	2012-08-02 11:15:08.648528	2012-08-02 11:15:20.569822
11695	1-11695	69	87	4660	57	7928	0	0	2012-08-02 11:15:20.569822	2012-08-02 11:15:20.569822
10331	1-10331	39	86	4096	60	9144	0	15	2012-08-02 10:25:49.64866	2012-08-02 11:15:24.120713
11707	1-11707	16	55	2306	51	2044	0	16	2012-08-02 11:15:25.281863	2012-08-02 11:15:25.281863
11444	1-11444	37	16	5906	88	5179	0	3	2012-08-02 11:13:43.522694	2012-08-02 11:15:26.476247
11710	1-11710	72	44	785	16	2986	0	0	2012-08-02 11:15:26.476247	2012-08-02 11:15:26.476247
8747	1-8747	20	26	6131	25	7649	0	11	2012-08-02 05:41:24.729739	2012-08-02 11:15:27.637065
10480	1-10480	68	28	4664	25	9183	0	3	2012-08-02 10:31:48.986021	2012-08-02 11:15:27.637065
10783	1-10783	52	18	6338	89	9701	0	0	2012-08-02 10:37:57.082255	2012-08-02 11:15:31.131114
11604	1-11604	47	42	7693	67	9360	0	17	2012-08-02 11:14:44.957993	2012-08-02 11:15:31.131114
11721	1-11721	66	55	817	18	4574	0	0	2012-08-02 11:15:31.131114	2012-08-02 11:15:31.131114
4500	1-4500	56	67	8966	60	6429	0	8	2012-08-02 05:12:41.692479	2012-08-02 11:15:39.333955
11646	1-11646	5	60	5614	72	9638	0	13	2012-08-02 11:15:00.836972	2012-08-02 11:15:39.333955
11743	1-11743	67	56	437	67	7729	0	11	2012-08-02 11:15:39.333955	2012-08-02 11:15:39.333955
11133	1-11133	15	56	2812	47	9194	0	3	2012-08-02 11:11:42.254873	2012-08-02 11:15:42.847989
11368	1-11368	34	56	2237	39	952	0	6	2012-08-02 11:13:13.976738	2012-08-02 11:15:42.847989
11070	1-11070	53	65	6750	39	5979	0	5	2012-08-02 11:11:17.635695	2012-08-02 11:15:42.847989
9512	1-9512	93	56	5015	100	9972	0	15	2012-08-02 05:46:36.380987	2012-08-02 11:15:42.847989
4031	1-4031	12	65	7372	39	2165	75	0	2012-08-02 05:09:34.192167	2012-08-02 11:15:42.847989
11760	1-11760	58	45	420	60	1109	0	7	2012-08-02 11:15:45.950096	2012-08-02 11:15:45.950096
7286	1-7286	67	99	8296	68	8907	0	0	2012-08-02 05:31:27.755975	2012-08-02 11:15:47.245425
6272	1-6272	99	99	3054	92	4300	0	0	2012-08-02 05:24:35.782821	2012-08-02 11:15:47.245425
11763	1-11763	39	30	2949	99	8073	0	4	2012-08-02 11:15:47.245425	2012-08-02 11:15:47.245425
6590	1-6590	42	35	8329	29	8063	0	0	2012-08-02 05:26:46.156181	2012-08-02 11:15:49.574199
10502	1-10502	83	87	6270	1	2465	0	0	2012-08-02 10:31:57.721761	2012-08-02 11:15:52.734572
10944	1-10944	86	39	6661	1	6434	0	7	2012-08-02 10:39:00.339087	2012-08-02 11:15:52.734572
11115	1-11115	43	30	5932	16	9650	0	5	2012-08-02 11:11:35.525612	2012-08-02 11:15:55.065674
9875	1-9875	75	41	8473	100	9936	0	0	2012-08-02 05:49:04.580358	2012-08-02 11:15:55.065674
4022	1-4022	61	30	7026	50	4498	0	0	2012-08-02 05:09:30.627046	2012-08-02 11:15:55.065674
8209	1-8209	69	30	7585	41	6077	0	0	2012-08-02 05:37:47.748156	2012-08-02 11:15:55.065674
11782	1-11782	10	16	4938	30	8909	0	6	2012-08-02 11:15:55.065674	2012-08-02 11:15:55.065674
11040	1-11040	50	74	2468	40	7618	0	0	2012-08-02 10:39:38.901121	2012-08-02 11:15:58.227009
11555	1-11555	88	50	5648	55	8399	0	5	2012-08-02 11:14:26.174917	2012-08-02 11:15:58.227009
6946	1-6946	91	74	7645	69	3550	0	5	2012-08-02 05:29:10.305502	2012-08-02 11:15:58.227009
9185	1-9185	61	74	7688	98	6103	0	0	2012-08-02 05:44:23.388743	2012-08-02 11:15:58.227009
11553	1-11553	63	69	5867	55	8144	0	0	2012-08-02 11:14:25.423274	2012-08-02 11:15:58.227009
11777	1-11777	18	1	3703	74	9096	0	0	2012-08-02 11:15:52.734572	2012-08-02 11:15:58.227009
11797	1-11797	54	59	29	24	7638	0	10	2012-08-02 11:16:00.603075	2012-08-02 11:16:00.603075
11313	1-11313	30	71	8795	61	9238	0	7	2012-08-02 11:12:52.785597	2012-08-02 11:16:04.085157
9542	1-9542	95	69	7404	71	7399	0	0	2012-08-02 05:46:48.635014	2012-08-02 11:16:04.085157
11808	1-11808	41	44	388	97	2112	0	9	2012-08-02 11:16:05.269263	2012-08-02 11:16:05.269263
11485	1-11485	100	89	5647	73	5907	0	0	2012-08-02 11:13:59.158566	2012-08-02 11:16:06.051463
11482	1-11482	17	15	8437	89	5576	0	0	2012-08-02 11:13:58.03155	2012-08-02 11:16:06.051463
9463	1-9463	62	69	8492	16	7139	0	0	2012-08-02 05:46:16.526328	2012-08-02 11:16:07.256137
9706	1-9706	88	2	9096	41	1435	0	2	2012-08-02 05:47:55.805281	2012-08-02 11:16:07.256137
11814	1-11814	41	16	787	59	4691	0	3	2012-08-02 11:16:07.256137	2012-08-02 11:16:07.256137
10977	1-10977	65	38	9368	28	9723	0	6	2012-08-02 10:39:13.34021	2012-08-02 11:16:08.471378
11546	1-11546	76	87	2887	14	9687	0	2	2012-08-02 11:14:22.747415	2012-08-02 11:16:13.520711
11848	1-11848	75	38	31	80	2225	0	6	2012-08-02 11:16:20.393545	2012-08-02 11:16:20.393545
7973	1-7973	74	62	4618	44	5968	0	11	2012-08-02 05:36:12.085645	2012-08-02 11:16:21.553426
9055	1-9055	14	63	6729	23	7436	0	8	2012-08-02 05:43:30.29156	2012-08-02 11:16:21.553426
11851	1-11851	65	44	1548	63	1729	0	0	2012-08-02 11:16:21.553426	2012-08-02 11:16:21.553426
11576	1-11576	52	21	4534	88	5694	0	0	2012-08-02 11:14:34.251142	2012-08-02 11:16:25.033725
11393	1-11393	25	29	4635	86	9616	0	0	2012-08-02 11:13:23.655328	2012-08-02 11:16:30.836393
11884	1-11884	80	34	273	14	1487	0	5	2012-08-02 11:16:34.30558	2012-08-02 11:16:34.30558
11816	1-11816	99	55	3601	4	3763	0	13	2012-08-02 11:16:07.677517	2012-08-02 11:16:35.45468
11886	1-11886	92	14	2757	55	8727	0	6	2012-08-02 11:16:35.45468	2012-08-02 11:16:35.45468
10747	1-10747	6	4	4591	100	5231	0	11	2012-08-02 10:37:43.446592	2012-08-02 11:16:38.548193
8518	1-8518	39	56	7515	4	4673	0	3	2012-08-02 05:39:52.305852	2012-08-02 11:16:38.548193
10093	1-10093	29	11	2476	27	4033	0	2	2012-08-02 10:21:17.140331	2012-08-02 11:16:42.724536
9332	1-9332	85	11	9221	3	8745	0	2	2012-08-02 05:45:23.244956	2012-08-02 11:16:42.724536
9633	1-9633	39	10	6799	60	7473	0	0	2012-08-02 05:47:25.718343	2012-08-02 11:16:42.724536
11378	1-11378	5	11	1906	35	1968	0	5	2012-08-02 11:13:17.801048	2012-08-02 11:16:42.724536
11907	1-11907	20	18	2150	11	7860	0	3	2012-08-02 11:16:42.724536	2012-08-02 11:16:42.724536
9133	1-9133	10	3	4202	52	7161	0	0	2012-08-02 05:44:02.41397	2012-08-02 11:16:47.875544
11230	1-11230	57	56	8382	58	8372	0	0	2012-08-02 11:12:20.339705	2012-08-02 11:16:49.059592
11922	1-11922	36	17	2885	56	9856	0	0	2012-08-02 11:16:49.059592	2012-08-02 11:16:49.059592
11895	1-11895	4	50	2985	56	4071	0	0	2012-08-02 11:16:38.548193	2012-08-02 11:16:54.883231
11936	1-11936	2	97	1040	69	5108	0	0	2012-08-02 11:16:54.883231	2012-08-02 11:16:54.883231
11942	1-11942	21	37	3	35	709	0	2	2012-08-02 11:16:56.916231	2012-08-02 11:16:56.916231
10992	1-10992	89	90	7151	30	7119	0	0	2012-08-02 10:39:18.811565	2012-08-02 11:17:00.418404
11437	1-11437	22	75	3523	32	5541	0	5	2012-08-02 11:13:40.749313	2012-08-02 11:17:00.418404
11951	1-11951	48	32	4703	90	7732	0	8	2012-08-02 11:17:00.418404	2012-08-02 11:17:00.418404
9623	1-9623	52	32	8336	90	366	366	0	2012-08-02 05:47:21.597112	2012-08-02 11:17:00.418404
10593	1-10593	61	3	1910	13	1835	0	0	2012-08-02 10:35:36.237262	2012-08-02 11:17:01.624068
11745	1-11745	83	93	3438	51	2825	0	0	2012-08-02 11:15:40.129431	2012-08-02 11:17:01.624068
3265	1-3265	81	51	9001	3	1155	587	0	2012-08-02 05:04:26.491535	2012-08-02 11:17:01.624068
9192	1-9192	61	68	6047	95	3402	0	0	2012-08-02 05:44:26.306733	2012-08-02 11:17:02.816395
11709	1-11709	16	22	4058	85	7682	0	4	2012-08-02 11:15:26.089209	2012-08-02 11:17:02.816395
11960	1-11960	53	5	1252	28	890	0	0	2012-08-02 11:17:03.998477	2012-08-02 11:17:03.998477
10776	1-10776	33	40	5774	72	4345	0	0	2012-08-02 10:37:54.286695	2012-08-02 11:17:12.67162
10064	1-10064	94	76	6932	8	2452	0	2	2012-08-02 10:21:05.873799	2012-08-02 11:15:32.313624
5364	1-5364	41	25	1237	16	555	0	4	2012-08-02 05:18:27.402407	2012-08-02 11:15:32.313624
11659	1-11659	15	13	5397	68	9903	0	0	2012-08-02 11:15:06.297318	2012-08-02 11:15:34.62543
11654	1-11654	66	8	7886	63	9473	0	3	2012-08-02 11:15:04.35059	2012-08-02 11:15:34.62543
11731	1-11731	78	69	772	13	9216	0	0	2012-08-02 11:15:34.62543	2012-08-02 11:15:34.62543
11359	1-11359	77	6	4161	94	4084	0	0	2012-08-02 11:13:10.562781	2012-08-02 11:15:38.947247
11377	1-11377	46	16	8363	94	9105	0	0	2012-08-02 11:13:17.41469	2012-08-02 11:15:41.278827
10411	1-10411	59	36	1719	92	3640	0	6	2012-08-02 10:27:48.8819	2012-08-02 11:15:54.281626
9050	1-9050	26	36	8516	77	9471	0	2	2012-08-02 05:43:28.225547	2012-08-02 11:15:54.281626
11781	1-11781	31	49	375	36	4434	0	9	2012-08-02 11:15:54.281626	2012-08-02 11:15:54.281626
8115	1-8115	48	87	6634	97	1406	0	0	2012-08-02 05:37:09.374954	2012-08-02 11:15:56.66897
11791	1-11791	70	39	3325	1	8866	0	11	2012-08-02 11:15:58.227009	2012-08-02 11:16:04.869133
11809	1-11809	27	63	1193	39	9653	0	3	2012-08-02 11:16:04.869133	2012-08-02 11:16:04.869133
9164	1-9164	85	72	9283	35	3957	0	8	2012-08-02 05:44:14.944419	2012-08-02 11:16:08.858103
11817	1-11817	33	50	2758	72	9687	0	5	2012-08-02 11:16:08.858103	2012-08-02 11:16:08.858103
11331	1-11331	31	25	7618	61	7329	0	11	2012-08-02 11:12:59.271544	2012-08-02 11:16:13.156317
11468	1-11468	33	80	5895	4	6221	0	0	2012-08-02 11:13:52.637974	2012-08-02 11:16:15.432227
11838	1-11838	75	43	680	30	9189	0	0	2012-08-02 11:16:16.582227	2012-08-02 11:16:16.582227
11742	1-11742	21	35	4849	6	9306	0	6	2012-08-02 11:15:38.947247	2012-08-02 11:16:17.708311
595	1-595	27	91	7311	58	4219	0	0	2012-08-02 04:44:35.799581	2012-08-02 11:16:17.708311
11097	1-11097	73	14	7690	86	6320	0	0	2012-08-02 11:11:28.140049	2012-08-02 11:16:21.940122
11852	1-11852	83	21	10	14	8760	0	0	2012-08-02 11:16:21.940122	2012-08-02 11:16:21.940122
4082	1-4082	74	9	7816	71	841	841	8	2012-08-02 05:09:54.512932	2012-08-02 11:16:21.940122
11858	1-11858	35	74	959	93	5456	0	0	2012-08-02 11:16:24.271335	2012-08-02 11:16:24.271335
11573	1-11573	28	51	3863	91	6841	0	9	2012-08-02 11:14:33.10235	2012-08-02 11:16:28.151504
11556	1-11556	60	88	6034	95	8044	0	17	2012-08-02 11:14:26.561211	2012-08-02 11:16:36.990477
8939	1-8939	34	62	5988	38	1604	0	9	2012-08-02 05:42:42.635772	2012-08-02 11:16:41.575506
9616	1-9616	94	62	8188	77	9670	0	8	2012-08-02 05:47:18.76843	2012-08-02 11:16:41.575506
11904	1-11904	51	38	1240	62	9800	0	5	2012-08-02 11:16:41.575506	2012-08-02 11:16:41.575506
8476	1-8476	36	9	9765	71	8882	0	12	2012-08-02 05:39:35.543665	2012-08-02 11:16:43.111969
2873	1-2873	85	38	8946	5	6804	0	0	2012-08-02 05:01:46.965257	2012-08-02 11:16:43.897895
11909	1-11909	18	94	179	71	8196	0	6	2012-08-02 11:16:43.897895	2012-08-02 11:16:43.897895
10888	1-10888	70	5	9216	94	8443	4386	8	2012-08-02 10:38:38.221795	2012-08-02 11:16:43.897895
11672	1-11672	84	8	1452	100	277	0	5	2012-08-02 11:15:11.775688	2012-08-02 11:16:47.489209
11918	1-11918	90	75	553	8	4377	0	18	2012-08-02 11:16:47.489209	2012-08-02 11:16:47.489209
11751	1-11751	45	39	1437	56	5946	0	4	2012-08-02 11:15:42.847989	2012-08-02 11:16:48.67307
10970	1-10970	94	87	6504	77	7277	0	5	2012-08-02 10:39:10.641932	2012-08-02 11:16:54.507499
11702	1-11702	16	94	6387	24	5648	0	11	2012-08-02 11:15:23.356133	2012-08-02 11:16:54.507499
11404	1-11404	100	23	2804	77	6675	0	7	2012-08-02 11:13:28.01982	2012-08-02 11:16:54.507499
11657	1-11657	70	94	2021	8	1539	0	0	2012-08-02 11:15:05.510735	2012-08-02 11:16:54.507499
11579	1-11579	58	94	9729	63	6948	0	9	2012-08-02 11:14:35.399738	2012-08-02 11:16:54.507499
9772	1-9772	74	94	2601	10	722	0	3	2012-08-02 05:48:22.648242	2012-08-02 11:16:57.313937
11943	1-11943	10	10	354	94	9282	0	0	2012-08-02 11:16:57.313937	2012-08-02 11:16:57.313937
195	1-195	34	26	9686	3	3295	0	0	2012-08-02 04:43:55.868886	2012-08-02 11:16:59.645095
11551	1-11551	22	52	3758	48	2644	0	10	2012-08-02 11:14:24.672122	2012-08-02 11:16:59.645095
11949	1-11949	88	3	703	78	6422	0	5	2012-08-02 11:16:59.645095	2012-08-02 11:16:59.645095
11822	1-11822	63	40	4390	49	3575	0	0	2012-08-02 11:16:10.471774	2012-08-02 11:17:03.202829
11717	1-11717	8	59	2129	2	9128	0	11	2012-08-02 11:15:29.184735	2012-08-02 11:17:06.739018
11966	1-11966	62	2	4116	11	3633	0	4	2012-08-02 11:17:06.739018	2012-08-02 11:17:06.739018
2046	1-2046	22	75	7837	2	1009	1009	2	2012-08-02 04:55:50.923249	2012-08-02 11:17:06.739018
7907	1-7907	78	11	9156	75	7433	3104	0	2012-08-02 05:35:44.274398	2012-08-02 11:17:06.739018
11804	1-11804	13	73	7664	78	5513	0	0	2012-08-02 11:16:03.333544	2012-08-02 11:17:13.466838
11984	1-11984	38	69	2636	73	6774	0	11	2012-08-02 11:17:13.466838	2012-08-02 11:17:13.466838
11675	1-11675	45	34	971	2	2911	0	2	2012-08-02 11:15:12.709591	2012-08-02 11:17:15.866266
11990	1-11990	43	2	2803	20	7833	0	13	2012-08-02 11:17:15.866266	2012-08-02 11:17:15.866266
11300	1-11300	47	68	8559	78	9950	0	20	2012-08-02 11:12:47.725661	2012-08-02 11:17:17.01574
11993	1-11993	43	88	1826	3	3808	0	3	2012-08-02 11:17:17.01574	2012-08-02 11:17:17.01574
11689	1-11689	41	42	3620	65	3930	0	7	2012-08-02 11:15:18.234471	2012-08-02 11:17:21.346268
11512	1-11512	50	8	7169	24	8653	0	11	2012-08-02 11:14:10.041316	2012-08-02 11:17:21.346268
11958	1-11958	41	96	763	5	8382	0	11	2012-08-02 11:17:03.202829	2012-08-02 11:17:23.754384
12010	1-12010	2	68	291	10	6115	0	0	2012-08-02 11:17:23.754384	2012-08-02 11:17:23.754384
9941	1-9941	95	38	8903	17	6848	0	4	2012-08-02 05:49:31.503736	2012-08-02 11:17:24.936746
12016	1-12016	26	42	640	85	1261	0	2	2012-08-02 11:17:26.118942	2012-08-02 11:17:26.118942
11937	1-11937	70	77	9132	94	9473	0	0	2012-08-02 11:16:54.507499	2012-08-02 11:17:32.759248
12033	1-12033	1	6	1254	77	4201	0	17	2012-08-02 11:17:32.759248	2012-08-02 11:17:32.759248
12017	1-12017	53	59	1900	34	6321	0	3	2012-08-02 11:17:26.505718	2012-08-02 11:17:33.896897
11314	1-11314	54	59	6191	90	7629	0	12	2012-08-02 11:12:53.161303	2012-08-02 11:17:33.896897
11311	1-11311	55	91	3109	14	3149	0	5	2012-08-02 11:12:51.979265	2012-08-02 11:17:41.012387
10387	1-10387	85	65	5481	14	5993	0	0	2012-08-02 10:27:39.413093	2012-08-02 11:17:41.012387
12054	1-12054	85	18	976	91	6991	0	0	2012-08-02 11:17:41.012387	2012-08-02 11:17:41.012387
8148	1-8148	88	15	4322	14	1750	0	10	2012-08-02 05:37:22.820504	2012-08-02 11:17:46.494947
11723	1-11723	94	52	5500	98	6466	0	13	2012-08-02 11:15:31.51801	2012-08-02 11:17:49.897874
8146	1-8146	97	97	6508	54	6292	0	0	2012-08-02 05:37:21.992091	2012-08-02 11:17:56.405367
11213	1-11213	54	56	6524	24	3455	0	5	2012-08-02 11:12:13.611354	2012-08-02 11:17:56.405367
12094	1-12094	60	54	4253	56	5548	0	6	2012-08-02 11:17:56.405367	2012-08-02 11:17:56.405367
9682	1-9682	28	63	4914	10	9845	0	4	2012-08-02 05:47:45.92792	2012-08-02 11:17:57.543694
12097	1-12097	72	79	1465	63	7656	0	12	2012-08-02 11:17:57.543694	2012-08-02 11:17:57.543694
12068	1-12068	18	81	777	15	5987	0	0	2012-08-02 11:17:46.494947	2012-08-02 11:18:01.002038
11403	1-11403	70	84	2247	15	9332	0	4	2012-08-02 11:13:27.599334	2012-08-02 11:18:01.002038
12109	1-12109	93	13	469	18	896	0	8	2012-08-02 11:18:02.139928	2012-08-02 11:18:02.139928
11294	1-11294	96	35	2189	7	3342	0	8	2012-08-02 11:12:45.359318	2012-08-02 11:18:03.288998
10764	1-10764	83	81	5177	65	3411	0	15	2012-08-02 10:37:49.956292	2012-08-02 11:16:41.199927
11592	1-11592	81	3	4430	81	7259	0	14	2012-08-02 11:14:40.360502	2012-08-02 11:16:41.199927
11902	1-11902	5	65	4033	3	8993	0	0	2012-08-02 11:16:41.199927	2012-08-02 11:16:41.199927
11738	1-11738	55	36	4571	19	9753	0	0	2012-08-02 11:15:37.378096	2012-08-02 11:16:43.499661
11713	1-11713	26	25	5516	76	8620	0	15	2012-08-02 11:15:27.637065	2012-08-02 11:16:43.499661
9600	1-9600	84	57	3780	51	2551	0	5	2012-08-02 05:47:12.171871	2012-08-02 11:16:43.499661
11047	1-11047	88	27	5396	34	4292	0	5	2012-08-02 11:11:09.311562	2012-08-02 11:16:43.499661
11908	1-11908	30	34	2388	36	7536	0	0	2012-08-02 11:16:43.499661	2012-08-02 11:16:43.499661
10631	1-10631	10	64	7552	67	9157	0	0	2012-08-02 10:35:39.067919	2012-08-02 11:16:45.897988
10838	1-10838	33	64	4345	87	9185	0	0	2012-08-02 10:38:18.231895	2012-08-02 11:16:45.897988
11291	1-11291	51	89	3580	76	3847	0	7	2012-08-02 11:12:44.199497	2012-08-02 11:16:47.046968
11917	1-11917	81	10	346	89	6347	0	4	2012-08-02 11:16:47.046968	2012-08-02 11:16:47.046968
4664	1-4664	32	24	8835	59	9660	0	11	2012-08-02 05:13:48.211433	2012-08-02 11:16:48.264351
11530	1-11530	79	25	4434	82	8109	0	4	2012-08-02 11:14:16.615163	2012-08-02 11:16:48.264351
11563	1-11563	43	51	4828	41	9639	0	3	2012-08-02 11:14:29.246334	2012-08-02 11:16:48.264351
5494	1-5494	2	25	3535	90	594	0	0	2012-08-02 05:19:18.663178	2012-08-02 11:16:48.264351
11564	1-11564	50	25	2187	98	1644	0	5	2012-08-02 11:14:29.633095	2012-08-02 11:16:48.264351
11920	1-11920	77	30	1793	24	8548	0	14	2012-08-02 11:16:48.264351	2012-08-02 11:16:48.264351
10057	1-10057	100	4	3854	100	4255	0	9	2012-08-02 10:21:03.155043	2012-08-02 11:16:52.97174
11932	1-11932	21	100	199	82	6610	0	0	2012-08-02 11:16:52.97174	2012-08-02 11:16:52.97174
6945	1-6945	64	41	5931	81	6476	0	0	2012-08-02 05:29:09.91929	2012-08-02 11:16:54.120655
11322	1-11322	91	24	1028	88	1706	0	11	2012-08-02 11:12:56.221946	2012-08-02 11:16:54.120655
11935	1-11935	74	88	2880	41	5320	0	0	2012-08-02 11:16:54.120655	2012-08-02 11:16:54.120655
11813	1-11813	9	18	3685	1	3946	0	5	2012-08-02 11:16:06.44965	2012-08-02 11:16:55.302803
4356	1-4356	96	97	8059	13	4217	0	4	2012-08-02 05:11:44.124995	2012-08-02 11:16:55.302803
11938	1-11938	85	51	2097	18	6589	0	0	2012-08-02 11:16:55.302803	2012-08-02 11:16:55.302803
11614	1-11614	41	39	8420	89	9514	0	4	2012-08-02 11:14:48.849192	2012-08-02 11:17:01.225118
11953	1-11953	78	41	1202	39	9320	0	15	2012-08-02 11:17:01.225118	2012-08-02 11:17:01.225118
11508	1-11508	93	29	5267	23	8792	0	0	2012-08-02 11:14:08.107531	2012-08-02 11:17:08.417874
11971	1-11971	38	50	2076	82	4332	0	3	2012-08-02 11:17:08.417874	2012-08-02 11:17:08.417874
7851	1-7851	57	84	7121	2	9947	0	2	2012-08-02 05:35:20.640571	2012-08-02 11:17:09.59992
11211	1-11211	13	84	4742	32	6388	0	0	2012-08-02 11:12:12.848826	2012-08-02 11:17:09.59992
11980	1-11980	12	69	700	24	1079	0	0	2012-08-02 11:17:11.886768	2012-08-02 11:17:11.886768
5316	1-5316	39	50	9010	17	3577	0	2	2012-08-02 05:18:08.562804	2012-08-02 11:17:14.297599
11986	1-11986	88	91	2362	27	4445	0	0	2012-08-02 11:17:14.297599	2012-08-02 11:17:14.297599
7273	1-7273	72	99	4923	19	2297	0	0	2012-08-02 05:31:22.67153	2012-08-02 11:17:20.959129
12040	1-12040	50	7	1981	30	3792	0	4	2012-08-02 11:17:35.819311	2012-08-02 11:17:35.819311
12049	1-12049	20	50	214	66	1187	0	0	2012-08-02 11:17:38.97931	2012-08-02 11:17:38.97931
11954	1-11954	72	51	6056	3	9042	0	10	2012-08-02 11:17:01.624068	2012-08-02 11:17:40.198824
12052	1-12052	35	27	2124	51	8980	0	11	2012-08-02 11:17:40.198824	2012-08-02 11:17:40.198824
11873	1-11873	84	100	3719	75	6479	0	0	2012-08-02 11:16:30.07419	2012-08-02 11:17:41.409799
5233	1-5233	51	64	7618	8	2190	0	0	2012-08-02 05:17:36.034081	2012-08-02 11:17:41.409799
11554	1-11554	55	33	1218	41	1648	0	0	2012-08-02 11:14:25.799046	2012-08-02 11:17:41.409799
11957	1-11957	81	95	5546	22	9848	0	14	2012-08-02 11:17:02.816395	2012-08-02 11:17:44.152642
12062	1-12062	28	21	1332	95	6214	0	3	2012-08-02 11:17:44.152642	2012-08-02 11:17:44.152642
3823	1-3823	46	79	9658	100	2347	1922	0	2012-08-02 05:08:10.829817	2012-08-02 11:17:44.152642
12066	1-12066	64	4	306	74	4935	0	10	2012-08-02 11:17:45.313023	2012-08-02 11:17:45.313023
11734	1-11734	97	70	1975	20	1598	0	0	2012-08-02 11:15:35.80946	2012-08-02 11:17:49.124408
11596	1-11596	72	6	3024	92	4095	0	2	2012-08-02 11:14:41.8964	2012-08-02 11:17:50.284477
11389	1-11389	61	88	4886	46	3508	0	2	2012-08-02 11:13:22.086273	2012-08-02 11:17:52.549216
11860	1-11860	92	88	4086	21	3956	0	6	2012-08-02 11:16:25.033725	2012-08-02 11:17:52.549216
10328	1-10328	23	53	5846	40	4864	0	2	2012-08-02 10:25:48.464567	2012-08-02 11:17:52.549216
12092	1-12092	82	59	1885	74	3097	0	0	2012-08-02 11:17:55.631513	2012-08-02 11:17:55.631513
11261	1-11261	36	77	3384	55	1385	0	8	2012-08-02 11:12:32.484442	2012-08-02 11:17:56.781219
12098	1-12098	64	88	829	68	6919	0	13	2012-08-02 11:17:57.930402	2012-08-02 11:17:57.930402
11868	1-11868	23	37	5041	57	8452	0	5	2012-08-02 11:16:28.151504	2012-08-02 11:18:04.813504
12116	1-12116	34	7	2571	9	8353	0	6	2012-08-02 11:18:04.813504	2012-08-02 11:18:04.813504
10897	1-10897	34	79	3110	60	5818	0	7	2012-08-02 10:38:41.768282	2012-08-02 11:18:11.830589
11589	1-11589	48	92	9103	49	7200	0	7	2012-08-02 11:14:39.222523	2012-08-02 11:18:15.244833
11976	1-11976	40	71	3214	9	5801	0	6	2012-08-02 11:17:10.351025	2012-08-02 11:18:15.244833
12143	1-12143	32	49	2544	71	9437	0	7	2012-08-02 11:18:15.244833	2012-08-02 11:18:15.244833
11386	1-11386	36	48	2802	7	5438	0	8	2012-08-02 11:13:20.917385	2012-08-02 11:18:16.404878
12106	1-12106	48	15	5368	52	5923	0	9	2012-08-02 11:18:01.002038	2012-08-02 11:18:20.238255
12108	1-12108	99	76	7748	100	9390	0	0	2012-08-02 11:18:01.753028	2012-08-02 11:18:20.238255
9259	1-9259	12	74	4735	53	6007	0	0	2012-08-02 05:44:53.551274	2012-08-02 11:18:27.850404
11164	1-11164	100	85	9998	65	8370	0	11	2012-08-02 11:11:54.815166	2012-08-02 11:18:27.850404
12179	1-12179	93	14	2285	43	7225	0	8	2012-08-02 11:18:28.977507	2012-08-02 11:18:28.977507
12191	1-12191	49	15	735	50	8650	0	10	2012-08-02 11:18:33.574221	2012-08-02 11:18:33.574221
12095	1-12095	89	75	728	77	4326	0	18	2012-08-02 11:17:56.781219	2012-08-02 11:18:35.816606
7157	1-7157	46	9	7544	24	2544	0	0	2012-08-02 05:30:35.513519	2012-08-02 11:18:35.816606
10451	1-10451	96	70	9672	79	5764	0	0	2012-08-02 10:31:37.492979	2012-08-02 11:18:35.816606
12197	1-12197	35	77	2142	70	6405	0	0	2012-08-02 11:18:35.816606	2012-08-02 11:18:35.816606
11796	1-11796	48	51	5808	76	8749	0	4	2012-08-02 11:16:00.216129	2012-08-02 11:18:36.987947
11790	1-11790	32	22	5116	68	6064	0	3	2012-08-02 11:15:57.840176	2012-08-02 11:18:36.987947
10755	1-10755	56	34	2654	68	1526	0	0	2012-08-02 10:37:46.749558	2012-08-02 11:18:36.987947
11127	1-11127	82	34	1730	24	549	0	10	2012-08-02 11:11:39.834968	2012-08-02 11:18:36.987947
12219	1-12219	53	38	1364	76	2237	0	0	2012-08-02 11:18:44.372263	2012-08-02 11:18:44.372263
12225	1-12225	83	14	107	15	8154	0	9	2012-08-02 11:18:46.658793	2012-08-02 11:18:46.658793
10763	1-10763	22	1	1789	24	1176	0	17	2012-08-02 10:37:49.536108	2012-08-02 11:17:12.67162
11982	1-11982	58	72	2180	59	5829	0	15	2012-08-02 11:17:12.67162	2012-08-02 11:17:12.67162
11769	1-11769	77	93	3616	35	3571	0	7	2012-08-02 11:15:49.574199	2012-08-02 11:17:19.39034
11999	1-11999	52	74	327	93	7951	0	2	2012-08-02 11:17:19.39034	2012-08-02 11:17:19.39034
11303	1-11303	53	49	1403	12	2286	0	0	2012-08-02 11:12:48.907607	2012-08-02 11:17:20.561531
9518	1-9518	8	49	9871	68	8884	0	7	2012-08-02 05:46:38.791013	2012-08-02 11:17:20.561531
11545	1-11545	29	60	8177	18	3144	0	0	2012-08-02 11:14:22.360639	2012-08-02 11:17:24.129968
11693	1-11693	70	98	6792	46	8614	0	14	2012-08-02 11:15:19.807528	2012-08-02 11:17:25.334523
8889	1-8889	90	3	9704	93	8695	0	2	2012-08-02 05:42:22.296477	2012-08-02 11:17:25.334523
9598	1-9598	96	99	6228	46	6703	0	0	2012-08-02 05:47:11.354168	2012-08-02 11:17:25.334523
12014	1-12014	21	46	4992	3	9308	0	0	2012-08-02 11:17:25.334523	2012-08-02 11:17:25.334523
11705	1-11705	77	34	1644	87	2939	0	0	2012-08-02 11:15:24.518882	2012-08-02 11:17:26.505718
9422	1-9422	91	45	2669	100	1216	0	0	2012-08-02 05:45:59.796037	2012-08-02 11:17:26.505718
11106	1-11106	2	34	3319	20	2729	0	12	2012-08-02 11:11:31.623692	2012-08-02 11:17:26.505718
11841	1-11841	3	58	3841	93	6504	0	0	2012-08-02 11:16:17.708311	2012-08-02 11:17:28.826119
12028	1-12028	59	63	942	28	5490	0	10	2012-08-02 11:17:29.997158	2012-08-02 11:17:29.997158
11756	1-11756	28	67	2172	87	6950	0	3	2012-08-02 11:15:44.436011	2012-08-02 11:17:34.261328
11402	1-11402	85	67	5127	51	5273	0	0	2012-08-02 11:13:27.215297	2012-08-02 11:17:34.261328
5876	1-5876	79	21	9214	19	1992	0	6	2012-08-02 05:21:53.277301	2012-08-02 11:17:34.261328
12013	1-12013	58	14	5634	38	6515	0	2	2012-08-02 11:17:24.936746	2012-08-02 11:17:36.990463
12044	1-12044	15	59	2128	14	5905	0	9	2012-08-02 11:17:36.990463	2012-08-02 11:17:36.990463
7060	1-7060	54	59	7733	14	5803	409	0	2012-08-02 05:29:55.634323	2012-08-02 11:17:36.990463
8601	1-8601	68	41	2228	31	952	0	5	2012-08-02 05:40:25.797682	2012-08-02 11:17:38.17261
5276	1-5276	51	41	9677	88	6068	0	3	2012-08-02 05:17:52.894488	2012-08-02 11:17:38.17261
12048	1-12048	14	12	2283	41	8339	0	2	2012-08-02 11:17:38.17261	2012-08-02 11:17:38.17261
9509	1-9509	25	64	9002	27	6788	0	4	2012-08-02 05:46:35.198326	2012-08-02 11:17:40.616093
11818	1-11818	67	98	6249	38	9086	0	3	2012-08-02 11:16:08.471378	2012-08-02 11:17:41.796561
8343	1-8343	63	55	7574	38	7882	0	2	2012-08-02 05:38:41.970853	2012-08-02 11:17:41.796561
4301	1-4301	70	55	7026	38	1314	973	0	2012-08-02 05:11:21.506364	2012-08-02 11:17:41.796561
12061	1-12061	1	53	508	85	5018	0	3	2012-08-02 11:17:43.777476	2012-08-02 11:17:43.777476
9493	1-9493	24	76	4228	19	1307	0	0	2012-08-02 05:46:28.657533	2012-08-02 11:17:44.926177
9647	1-9647	66	68	9054	30	7832	0	0	2012-08-02 05:47:31.475678	2012-08-02 11:17:54.085311
10386	1-10386	93	25	6551	96	2912	0	0	2012-08-02 10:27:39.015364	2012-08-02 11:17:54.085311
12007	1-12007	11	24	2506	68	7966	0	10	2012-08-02 11:17:22.583377	2012-08-02 11:17:54.085311
11704	1-11704	41	72	7275	55	5180	0	8	2012-08-02 11:15:24.120713	2012-08-02 11:17:54.085311
4885	1-4885	67	24	6680	10	7541	0	0	2012-08-02 05:15:17.329187	2012-08-02 11:17:54.085311
12090	1-12090	87	99	3146	24	7750	0	0	2012-08-02 11:17:54.085311	2012-08-02 11:17:54.085311
11088	1-11088	10	65	5383	17	6633	0	0	2012-08-02 11:11:24.613506	2012-08-02 11:17:56.029889
8605	1-8605	16	77	9800	49	4613	0	4	2012-08-02 05:40:27.388807	2012-08-02 11:17:58.316953
11881	1-11881	27	68	7501	81	7643	0	10	2012-08-02 11:16:32.736869	2012-08-02 11:18:00.604149
12105	1-12105	25	56	779	68	9325	0	11	2012-08-02 11:18:00.604149	2012-08-02 11:18:00.604149
11647	1-11647	78	100	4697	28	2997	0	0	2012-08-02 11:15:01.621514	2012-08-02 11:18:01.753028
11290	1-11290	72	21	9504	69	8916	0	2	2012-08-02 11:12:43.81278	2012-08-02 11:18:01.753028
12139	1-12139	30	16	1885	19	4960	0	2	2012-08-02 11:18:13.719985	2012-08-02 11:18:13.719985
12140	1-12140	35	97	2978	76	3906	0	4	2012-08-02 11:18:14.084245	2012-08-02 11:18:14.869097
11919	1-11919	34	7	9027	3	6592	0	0	2012-08-02 11:16:47.875544	2012-08-02 11:18:20.60287
9381	1-9381	4	93	8221	20	6304	0	19	2012-08-02 05:45:43.23199	2012-08-02 11:18:21.774066
12086	1-12086	46	93	1540	80	1234	0	7	2012-08-02 11:17:53.311438	2012-08-02 11:18:27.121335
9528	1-9528	38	11	6137	9	6116	0	2	2012-08-02 05:46:42.845761	2012-08-02 11:18:28.226095
12177	1-12177	14	15	3764	11	5735	0	0	2012-08-02 11:18:28.226095	2012-08-02 11:18:28.226095
12122	1-12122	87	80	3736	92	5290	0	6	2012-08-02 11:18:07.262885	2012-08-02 11:18:29.364252
12184	1-12184	15	32	1720	99	5921	0	0	2012-08-02 11:18:30.502373	2012-08-02 11:18:30.502373
8437	1-8437	72	83	6918	14	7307	0	0	2012-08-02 05:39:19.961636	2012-08-02 11:18:32.038113
12187	1-12187	38	66	161	83	1204	0	0	2012-08-02 11:18:32.038113	2012-08-02 11:18:32.038113
12190	1-12190	74	41	1575	93	3839	0	5	2012-08-02 11:18:33.209406	2012-08-02 11:18:33.209406
11158	1-11158	71	30	8922	54	9993	0	0	2012-08-02 11:11:52.071629	2012-08-02 11:18:37.739071
11059	1-11059	39	17	9487	41	8207	0	0	2012-08-02 11:11:13.288184	2012-08-02 11:18:40.083726
6680	1-6680	5	91	4821	10	2672	0	0	2012-08-02 05:27:22.588544	2012-08-02 11:18:40.083726
11481	1-11481	7	17	6490	91	9105	0	0	2012-08-02 11:13:57.634289	2012-08-02 11:18:40.083726
4498	1-4498	5	61	5035	20	2462	0	8	2012-08-02 05:12:40.908074	2012-08-02 11:18:41.232527
11272	1-11272	84	44	7771	13	7630	0	5	2012-08-02 11:12:36.80497	2012-08-02 11:18:42.392774
11204	1-11204	82	1	4212	92	5626	0	0	2012-08-02 11:12:10.109028	2012-08-02 11:18:42.392774
12215	1-12215	7	13	6392	1	9296	0	15	2012-08-02 11:18:42.392774	2012-08-02 11:18:42.392774
1675	1-1675	2	1	6543	5	905	905	0	2012-08-02 04:53:06.253335	2012-08-02 11:18:42.392774
11369	1-11369	61	92	9895	27	8617	0	0	2012-08-02 11:13:14.363388	2012-08-02 11:18:45.145114
12221	1-12221	27	27	3374	89	5446	0	0	2012-08-02 11:18:45.145114	2012-08-02 11:18:45.145114
6499	1-6499	73	12	9175	29	6801	0	4	2012-08-02 05:26:08.696287	2012-08-02 11:18:46.294302
9382	1-9382	93	45	9206	6	7131	0	4	2012-08-02 05:45:43.629427	2012-08-02 11:18:46.294302
8969	1-8969	5	45	3488	90	967	0	10	2012-08-02 05:42:54.956355	2012-08-02 11:18:46.294302
10711	1-10711	77	45	5805	63	788	0	19	2012-08-02 10:36:25.496881	2012-08-02 11:18:46.294302
12224	1-12224	31	29	722	45	8060	0	5	2012-08-02 11:18:46.294302	2012-08-02 11:18:46.294302
4579	1-4579	39	45	6978	53	5734	3283	0	2012-08-02 05:13:13.339751	2012-08-02 11:18:46.294302
11773	1-11773	47	71	8521	59	7152	0	7	2012-08-02 11:15:51.14344	2012-08-02 11:18:47.433288
809	1-809	73	35	9418	7	4966	1254	0	2012-08-02 04:45:36.806575	2012-08-02 11:18:03.288998
11880	1-11880	65	13	3938	51	6343	0	5	2012-08-02 11:16:33.134849	2012-08-02 11:18:06.724886
808	1-808	36	13	8421	51	2412	1334	0	2012-08-02 04:45:36.53018	2012-08-02 11:18:06.724886
12123	1-12123	71	4	1291	100	2375	0	17	2012-08-02 11:18:08.028376	2012-08-02 11:18:08.028376
11283	1-11283	22	33	3778	86	3930	0	0	2012-08-02 11:12:41.401337	2012-08-02 11:18:13.355285
10886	1-10886	67	33	3988	23	6210	0	11	2012-08-02 10:38:37.446677	2012-08-02 11:18:13.355285
11711	1-11711	67	33	2365	61	5944	0	7	2012-08-02 11:15:26.862754	2012-08-02 11:18:13.355285
12138	1-12138	70	46	1474	33	6768	0	11	2012-08-02 11:18:13.355285	2012-08-02 11:18:13.355285
12161	1-12161	68	72	382	23	9125	0	8	2012-08-02 11:18:22.160998	2012-08-02 11:18:22.160998
803	1-803	93	23	9471	89	2535	189	0	2012-08-02 04:45:34.453545	2012-08-02 11:18:22.160998
10087	1-10087	27	88	6231	9	2576	0	4	2012-08-02 10:21:14.807921	2012-08-02 11:18:28.590617
12178	1-12178	94	20	376	88	4562	0	13	2012-08-02 11:18:28.590617	2012-08-02 11:18:28.590617
9676	1-9676	39	55	8942	39	4920	0	0	2012-08-02 05:47:43.452844	2012-08-02 11:18:31.651266
11757	1-11757	6	60	1955	18	401	0	0	2012-08-02 11:15:44.811613	2012-08-02 11:18:31.651266
8830	1-8830	10	60	7733	81	2550	0	0	2012-08-02 05:41:58.340565	2012-08-02 11:18:31.651266
12186	1-12186	5	39	336	60	8071	0	15	2012-08-02 11:18:31.651266	2012-08-02 11:18:31.651266
12198	1-12198	87	35	710	49	8192	0	5	2012-08-02 11:18:36.214759	2012-08-02 11:18:36.214759
12151	1-12151	82	58	2493	65	2782	0	13	2012-08-02 11:18:18.338073	2012-08-02 11:18:37.363591
12121	1-12121	70	58	3276	13	4548	0	18	2012-08-02 11:18:06.724886	2012-08-02 11:18:37.363591
12053	1-12053	38	95	6940	64	8736	0	4	2012-08-02 11:17:40.616093	2012-08-02 11:18:40.845599
10836	1-10836	93	47	5710	66	5422	0	5	2012-08-02 10:38:17.845043	2012-08-02 11:18:40.845599
12210	1-12210	60	44	7436	95	8063	0	0	2012-08-02 11:18:40.845599	2012-08-02 11:18:40.845599
10606	1-10606	26	94	9931	79	9164	0	0	2012-08-02 10:35:37.180468	2012-08-02 11:18:42.017055
12214	1-12214	55	94	985	95	8746	0	6	2012-08-02 11:18:42.790525	2012-08-02 11:18:42.790525
12220	1-12220	13	67	1513	38	2065	0	10	2012-08-02 11:18:44.758408	2012-08-02 11:18:44.758408
11548	1-11548	64	87	9651	59	7715	0	7	2012-08-02 11:14:23.511993	2012-08-02 11:18:45.918585
11265	1-11265	10	91	4535	37	6186	0	12	2012-08-02 11:12:34.075261	2012-08-02 11:18:50.095922
7835	1-7835	65	91	7172	37	9446	0	5	2012-08-02 05:35:13.86821	2012-08-02 11:18:50.095922
12233	1-12233	89	45	5824	91	7406	0	13	2012-08-02 11:18:50.095922	2012-08-02 11:18:50.095922
11828	1-11828	95	63	1616	87	7268	0	0	2012-08-02 11:16:12.791778	2012-08-02 11:18:49.730361
11363	1-11363	85	84	1867	66	1169	0	0	2012-08-02 11:13:12.076754	2012-08-02 11:18:50.492944
12235	1-12235	31	9	1055	84	8618	0	0	2012-08-02 11:18:50.492944	2012-08-02 11:18:50.492944
11082	1-11082	5	13	6649	7	6991	0	12	2012-08-02 11:11:22.722922	2012-08-02 11:18:52.028869
12021	1-12021	6	51	2433	11	1868	0	0	2012-08-02 11:17:28.074636	2012-08-02 11:18:52.028869
6050	1-6050	100	51	5556	14	1890	0	0	2012-08-02 05:23:05.364834	2012-08-02 11:18:52.028869
12239	1-12239	91	66	1258	13	5960	0	0	2012-08-02 11:18:52.028869	2012-08-02 11:18:52.028869
7964	1-7964	33	90	9641	97	6069	0	0	2012-08-02 05:36:08.439693	2012-08-02 11:18:52.415547
6652	1-6652	3	90	8470	54	5812	0	0	2012-08-02 05:27:11.407014	2012-08-02 11:18:52.415547
9603	1-9603	1	90	8410	17	6709	0	0	2012-08-02 05:47:13.398816	2012-08-02 11:18:52.415547
12093	1-12093	71	28	3499	65	8146	0	0	2012-08-02 11:17:56.029889	2012-08-02 11:18:53.620239
11679	1-11679	65	75	1010	2	3246	0	12	2012-08-02 11:15:14.320986	2012-08-02 11:18:55.16701
12247	1-12247	10	44	287	75	5328	0	4	2012-08-02 11:18:55.16701	2012-08-02 11:18:55.16701
11328	1-11328	45	58	5348	21	5856	0	0	2012-08-02 11:12:58.520163	2012-08-02 11:18:55.940371
11835	1-11835	67	42	6809	32	8709	0	8	2012-08-02 11:16:15.432227	2012-08-02 11:18:57.100269
11676	1-11676	35	59	7913	69	8792	0	7	2012-08-02 11:15:13.110572	2012-08-02 11:18:58.238171
12255	1-12255	93	81	5013	59	9640	0	7	2012-08-02 11:18:58.238171	2012-08-02 11:18:58.238171
12142	1-12142	68	76	4048	97	4531	0	15	2012-08-02 11:18:14.869097	2012-08-02 11:18:58.625043
9065	1-9065	35	63	7152	69	9589	0	4	2012-08-02 05:43:34.357013	2012-08-02 11:18:59.000354
11989	1-11989	4	43	2867	69	7347	0	2	2012-08-02 11:17:15.490411	2012-08-02 11:18:59.000354
11320	1-11320	62	80	3123	69	1337	0	0	2012-08-02 11:12:55.459204	2012-08-02 11:18:59.000354
12257	1-12257	19	69	5767	63	8307	0	4	2012-08-02 11:18:59.000354	2012-08-02 11:18:59.000354
11218	1-11218	52	64	9025	6	9319	0	2	2012-08-02 11:12:15.533852	2012-08-02 11:18:59.376264
12258	1-12258	20	23	2270	64	5453	0	8	2012-08-02 11:18:59.376264	2012-08-02 11:18:59.376264
11574	1-11574	34	81	4379	7	5336	0	2	2012-08-02 11:14:33.477517	2012-08-02 11:19:00.160668
11688	1-11688	56	81	4800	62	9171	0	5	2012-08-02 11:15:17.845583	2012-08-02 11:19:00.160668
12202	1-12202	23	7	5365	30	7066	0	5	2012-08-02 11:18:37.739071	2012-08-02 11:19:00.160668
12260	1-12260	39	21	1794	81	6698	0	2	2012-08-02 11:19:00.160668	2012-08-02 11:19:00.160668
4310	1-4310	83	81	9454	62	2693	1652	0	2012-08-02 05:11:25.263302	2012-08-02 11:19:00.160668
9943	1-9943	40	75	9089	9	7841	0	5	2012-08-02 05:49:32.322021	2012-08-02 11:19:00.934049
10187	1-10187	59	11	8174	10	8475	0	7	2012-08-02 10:22:18.355632	2012-08-02 11:19:00.934049
3333	1-3333	44	35	4082	100	2094	0	0	2012-08-02 05:04:53.96014	2012-08-02 11:19:01.720992
11396	1-11396	100	35	4168	63	3069	0	0	2012-08-02 11:13:24.793267	2012-08-02 11:19:01.720992
12266	1-12266	69	91	2644	21	4894	0	12	2012-08-02 11:19:02.471897	2012-08-02 11:19:02.471897
11660	1-11660	29	31	6048	61	5953	0	2	2012-08-02 11:15:06.715242	2012-08-02 11:19:04.007617
12160	1-12160	21	85	3889	93	7942	0	0	2012-08-02 11:18:21.774066	2012-08-02 11:19:05.134505
12273	1-12273	13	93	5453	10	6395	0	6	2012-08-02 11:19:05.134505	2012-08-02 11:19:05.134505
5613	1-5613	31	49	8295	95	4637	0	0	2012-08-02 05:20:06.36345	2012-08-02 11:19:06.27261
10616	1-10616	96	70	9814	62	4963	0	0	2012-08-02 10:35:37.938866	2012-08-02 11:19:07.819161
12280	1-12280	97	62	2234	83	7512	0	9	2012-08-02 11:19:07.819161	2012-08-02 11:19:07.819161
9480	1-9480	5	7	8366	68	4769	0	8	2012-08-02 05:46:23.409339	2012-08-02 11:19:09.752469
12286	1-12286	48	44	1885	20	1598	0	12	2012-08-02 11:19:10.128115	2012-08-02 11:19:10.128115
12288	1-12288	29	12	490	16	2119	0	0	2012-08-02 11:19:10.890569	2012-08-02 11:19:10.890569
10010	1-10010	100	57	5123	75	2032	0	5	2012-08-02 09:10:33.431833	2012-08-02 11:19:11.67487
12290	1-12290	73	2	1073	57	9877	0	0	2012-08-02 11:19:11.67487	2012-08-02 11:19:11.67487
10449	1-10449	66	32	4921	25	9592	0	3	2012-08-02 10:31:36.648938	2012-08-02 11:19:12.072644
10638	1-10638	43	63	9718	11	9774	0	0	2012-08-02 10:35:39.599997	2012-08-02 11:19:12.072644
12291	1-12291	45	38	589	63	4339	0	0	2012-08-02 11:19:12.072644	2012-08-02 11:19:12.072644
10375	1-10375	20	63	7353	24	5771	4262	7	2012-08-02 10:27:34.637817	2012-08-02 11:19:12.072644
10934	1-10934	29	76	6305	65	4447	0	6	2012-08-02 10:38:56.378342	2012-08-02 11:19:13.995089
5515	1-5515	53	43	4714	62	2176	0	12	2012-08-02 05:19:27.269583	2012-08-02 11:19:16.72384
12302	1-12302	14	62	918	29	7886	0	0	2012-08-02 11:19:16.72384	2012-08-02 11:19:16.72384
10016	1-10016	22	64	7900	49	5656	0	0	2012-08-02 10:16:57.867636	2012-08-02 11:19:18.027812
12312	1-12312	2	84	1524	17	9683	0	0	2012-08-02 11:19:20.336446	2012-08-02 11:19:20.336446
12318	1-12318	31	80	1080	25	2051	0	0	2012-08-02 11:19:22.656582	2012-08-02 11:19:22.656582
12321	1-12321	52	38	222	89	7713	0	9	2012-08-02 11:19:23.827988	2012-08-02 11:19:23.827988
12324	1-12324	32	71	597	48	2866	0	0	2012-08-02 11:19:24.955259	2012-08-02 11:19:24.955259
10479	1-10479	10	23	4876	66	5437	0	0	2012-08-02 10:31:48.588369	2012-08-02 11:19:26.126022
12327	1-12327	96	70	1785	49	3004	0	10	2012-08-02 11:19:26.126022	2012-08-02 11:19:26.126022
10320	1-10320	52	66	6791	88	8840	0	10	2012-08-02 10:25:45.3601	2012-08-02 11:19:28.468416
12118	1-12118	3	43	3762	30	6542	0	2	2012-08-02 11:18:05.564608	2012-08-02 11:19:30.722095
9664	1-9664	22	22	1117	60	862	0	14	2012-08-02 05:47:38.480915	2012-08-02 11:19:47.352458
12240	1-12240	69	4	700	90	3790	0	15	2012-08-02 11:18:52.415547	2012-08-02 11:19:47.352458
12382	1-12382	23	60	780	79	4196	0	2	2012-08-02 11:19:47.352458	2012-08-02 11:19:47.352458
4258	1-4258	70	22	9183	60	4468	4318	5	2012-08-02 05:11:03.936105	2012-08-02 11:19:47.352458
12385	1-12385	88	20	76	79	6596	0	0	2012-08-02 11:19:48.501798	2012-08-02 11:19:48.501798
12112	1-12112	93	20	6433	74	5835	0	5	2012-08-02 11:18:03.288998	2012-08-02 11:19:51.970866
12223	1-12223	88	11	3797	87	7392	0	7	2012-08-02 11:18:45.918585	2012-08-02 11:19:51.970866
12394	1-12394	72	67	4835	11	8964	0	5	2012-08-02 11:19:51.970866	2012-08-02 11:19:51.970866
12397	1-12397	48	74	63	25	3374	0	12	2012-08-02 11:19:53.164283	2012-08-02 11:19:53.164283
12296	1-12296	36	65	4015	8	3282	0	8	2012-08-02 11:19:13.995089	2012-08-02 11:19:55.462086
9371	1-9371	20	65	9296	97	6840	0	0	2012-08-02 05:45:39.177103	2012-08-02 11:19:55.462086
9040	1-9040	13	14	5593	92	6745	0	3	2012-08-02 05:43:24.148802	2012-08-02 11:20:02.402723
11898	1-11898	62	14	4080	66	3408	0	0	2012-08-02 11:16:39.67535	2012-08-02 11:20:02.402723
12249	1-12249	57	21	6025	61	4641	0	0	2012-08-02 11:18:55.940371	2012-08-02 11:20:03.551811
11772	1-11772	8	80	1970	53	2478	0	3	2012-08-02 11:15:50.745525	2012-08-02 11:20:04.723067
11811	1-11811	43	73	2429	15	9311	0	0	2012-08-02 11:16:06.051463	2012-08-02 11:20:05.872235
12430	1-12430	75	15	515	76	1735	0	0	2012-08-02 11:20:05.872235	2012-08-02 11:20:05.872235
9076	1-9076	3	97	8402	95	5045	0	0	2012-08-02 05:43:38.809257	2012-08-02 11:20:07.043702
11722	1-11722	28	19	3112	95	9234	0	0	2012-08-02 11:15:30.744269	2012-08-02 11:20:07.043702
12432	1-12432	78	63	3865	97	7063	0	4	2012-08-02 11:20:07.043702	2012-08-02 11:20:07.043702
12435	1-12435	81	48	715	33	4253	0	3	2012-08-02 11:20:07.83926	2012-08-02 11:20:07.83926
12447	1-12447	12	93	1147	19	5573	0	3	2012-08-02 11:20:12.47991	2012-08-02 11:20:12.47991
12370	1-12370	64	98	3226	63	7245	0	0	2012-08-02 11:19:43.031706	2012-08-02 11:20:14.819348
10053	1-10053	15	22	9887	23	6038	0	2	2012-08-02 10:21:01.5818	2012-08-02 11:20:14.819348
12002	1-12002	98	12	8409	49	7189	0	0	2012-08-02 11:17:20.561531	2012-08-02 11:20:14.819348
12453	1-12453	36	23	1753	19	6144	0	19	2012-08-02 11:20:14.819348	2012-08-02 11:20:14.819348
10564	1-10564	78	90	2335	12	2441	0	0	2012-08-02 10:35:34.024572	2012-08-02 11:20:16.022967
12456	1-12456	28	12	1981	16	3326	0	2	2012-08-02 11:20:16.022967	2012-08-02 11:20:16.022967
11643	1-11643	36	31	4037	12	9317	0	5	2012-08-02 11:14:59.665625	2012-08-02 11:20:17.19583
12459	1-12459	8	66	865	26	6271	0	18	2012-08-02 11:20:17.19583	2012-08-02 11:20:17.19583
12466	1-12466	15	95	657	26	9999	0	4	2012-08-02 11:20:19.970268	2012-08-02 11:20:19.970268
11883	1-11883	48	44	6321	57	7875	0	0	2012-08-02 11:16:33.918993	2012-08-02 11:20:21.119568
12469	1-12469	89	91	3081	44	6338	0	16	2012-08-02 11:20:21.119568	2012-08-02 11:20:21.119568
8453	1-8453	79	26	6282	100	8734	0	3	2012-08-02 05:39:26.317541	2012-08-02 11:20:23.49027
7690	1-7690	84	54	7615	79	3496	0	3	2012-08-02 05:34:12.217999	2012-08-02 11:20:24.650818
10946	1-10946	89	54	7268	4	3872	0	3	2012-08-02 10:39:01.555341	2012-08-02 11:20:24.650818
12478	1-12478	78	10	1913	54	9351	0	11	2012-08-02 11:20:24.650818	2012-08-02 11:20:24.650818
12481	1-12481	83	72	33	69	9463	0	8	2012-08-02 11:20:25.834925	2012-08-02 11:20:25.834925
11974	1-11974	23	2	8618	84	8309	0	2	2012-08-02 11:17:09.59992	2012-08-02 11:20:28.944385
12176	1-12176	28	10	5882	74	7889	0	0	2012-08-02 11:18:27.850404	2012-08-02 11:20:28.944385
10884	1-10884	83	10	9207	43	8748	0	7	2012-08-02 10:38:36.673227	2012-08-02 11:20:28.944385
12252	1-12252	24	32	8760	75	8527	0	0	2012-08-02 11:18:57.100269	2012-08-02 11:20:28.944385
10466	1-10466	79	10	5568	62	4740	0	15	2012-08-02 10:31:43.417424	2012-08-02 11:20:28.944385
10327	1-10327	95	89	6763	82	8617	0	5	2012-08-02 10:25:48.088715	2012-08-02 11:20:33.212148
11638	1-11638	47	83	1836	80	1552	0	6	2012-08-02 11:14:58.517051	2012-08-02 11:20:33.212148
11628	1-11628	94	71	8271	80	8275	0	0	2012-08-02 11:14:54.262813	2012-08-02 11:20:33.212148
9570	1-9570	66	43	8741	81	8361	0	0	2012-08-02 05:46:59.905048	2012-08-02 11:20:33.212148
12500	1-12500	36	80	2015	43	7382	0	12	2012-08-02 11:20:33.212148	2012-08-02 11:20:33.212148
11222	1-11222	13	21	3761	6	3711	0	3	2012-08-02 11:12:17.21571	2012-08-02 11:20:34.395872
12503	1-12503	65	71	1145	7	6470	0	5	2012-08-02 11:20:34.395872	2012-08-02 11:20:34.395872
10744	1-10744	14	71	8397	62	5956	0	4	2012-08-02 10:37:42.253155	2012-08-02 11:20:44.491589
12529	1-12529	75	17	698	71	8543	0	0	2012-08-02 11:20:44.491589	2012-08-02 11:20:44.491589
12396	1-12396	27	42	4341	90	5673	0	4	2012-08-02 11:19:52.777365	2012-08-02 11:20:51.473831
12205	1-12205	93	19	3000	12	5864	0	0	2012-08-02 11:18:38.945338	2012-08-02 11:20:53.851525
11798	1-11798	39	53	6956	12	9747	0	0	2012-08-02 11:16:01.013952	2012-08-02 11:20:53.851525
12553	1-12553	9	66	2382	19	3296	0	0	2012-08-02 11:20:53.851525	2012-08-02 11:20:53.851525
12328	1-12328	91	19	998	70	5206	0	4	2012-08-02 11:19:26.512961	2012-08-02 11:20:55.056277
12358	1-12358	52	41	840	54	2348	0	0	2012-08-02 11:19:38.037969	2012-08-02 11:20:55.056277
12556	1-12556	71	32	3568	96	8864	0	7	2012-08-02 11:20:55.056277	2012-08-02 11:20:55.056277
11923	1-11923	10	4	4143	87	8896	0	0	2012-08-02 11:16:49.447342	2012-08-02 11:20:57.432968
12231	1-12231	8	94	7392	13	7684	0	0	2012-08-02 11:18:48.957149	2012-08-02 11:20:58.582017
12565	1-12565	62	90	1440	94	5115	0	0	2012-08-02 11:20:58.582017	2012-08-02 11:20:58.582017
9391	1-9391	86	33	5667	4	4539	0	0	2012-08-02 05:45:47.25541	2012-08-02 11:20:59.742471
11699	1-11699	13	33	3038	21	9659	0	0	2012-08-02 11:15:22.172465	2012-08-02 11:20:59.742471
12479	1-12479	71	41	1601	14	9149	0	7	2012-08-02 11:20:25.038434	2012-08-02 11:20:59.742471
12568	1-12568	13	82	1121	33	9078	0	0	2012-08-02 11:20:59.742471	2012-08-02 11:20:59.742471
12073	1-12073	5	19	2540	83	9153	0	0	2012-08-02 11:17:48.373207	2012-08-02 11:21:03.222696
12577	1-12577	47	14	1367	19	7940	0	12	2012-08-02 11:21:03.222696	2012-08-02 11:21:03.222696
12480	1-12480	11	67	2715	38	7487	0	11	2012-08-02 11:20:25.436242	2012-08-02 11:21:05.565369
8052	1-8052	3	19	3230	53	4431	0	0	2012-08-02 05:36:44.016795	2012-08-02 11:21:09.156381
12279	1-12279	98	16	6726	73	2438	0	2	2012-08-02 11:19:07.443604	2012-08-02 11:21:09.156381
8971	1-8971	27	34	6977	16	7480	0	5	2012-08-02 05:42:55.77388	2012-08-02 11:21:10.360732
11850	1-11850	8	97	2382	68	3602	0	6	2012-08-02 11:16:21.17746	2012-08-02 11:19:19.94982
12311	1-12311	97	59	1961	97	8505	0	0	2012-08-02 11:19:19.94982	2012-08-02 11:19:19.94982
12320	1-12320	63	27	530	17	5013	0	0	2012-08-02 11:19:23.441636	2012-08-02 11:19:23.441636
12084	1-12084	1	40	6074	88	9775	0	4	2012-08-02 11:17:52.549216	2012-08-02 11:19:25.739336
12335	1-12335	77	86	8741	66	6423	0	6	2012-08-02 11:19:29.21965	2012-08-02 11:19:29.21965
12029	1-12029	50	89	7633	94	8167	0	3	2012-08-02 11:17:31.179321	2012-08-02 11:19:31.50836
12341	1-12341	38	11	4186	89	7380	0	10	2012-08-02 11:19:31.50836	2012-08-02 11:19:31.50836
11874	1-11874	71	86	8233	46	9059	0	0	2012-08-02 11:16:30.836393	2012-08-02 11:19:33.839735
12350	1-12350	48	37	134	73	2917	0	0	2012-08-02 11:19:34.999848	2012-08-02 11:19:34.999848
12356	1-12356	12	100	292	34	4182	0	5	2012-08-02 11:19:37.286736	2012-08-02 11:19:37.286736
6615	1-6615	84	34	8760	45	574	574	0	2012-08-02 05:26:56.376665	2012-08-02 11:19:37.286736
11897	1-11897	28	9	4732	8	3006	0	0	2012-08-02 11:16:39.288406	2012-08-02 11:19:38.435714
12359	1-12359	99	36	1516	9	7941	0	0	2012-08-02 11:19:38.435714	2012-08-02 11:19:38.435714
12362	1-12362	12	95	317	56	2399	0	4	2012-08-02 11:19:39.584951	2012-08-02 11:19:39.584951
12055	1-12055	23	8	2519	26	7032	0	0	2012-08-02 11:17:41.409799	2012-08-02 11:19:40.745033
10459	1-10459	14	90	9499	98	7241	0	14	2012-08-02 10:31:40.60028	2012-08-02 11:19:43.031706
5037	1-5037	91	41	4454	95	1857	0	0	2012-08-02 05:16:17.349524	2012-08-02 11:19:43.031706
12376	1-12376	73	69	1754	21	2165	0	7	2012-08-02 11:19:45.021359	2012-08-02 11:19:45.021359
9525	1-9525	18	36	4891	78	3664	0	4	2012-08-02 05:46:41.652789	2012-08-02 11:19:47.728103
9320	1-9320	93	66	9613	5	8137	0	0	2012-08-02 05:45:18.359195	2012-08-02 11:19:50.059928
12333	1-12333	37	17	4712	66	6514	0	8	2012-08-02 11:19:28.468416	2012-08-02 11:19:50.059928
11724	1-11724	37	17	9316	54	7425	0	0	2012-08-02 11:15:31.916024	2012-08-02 11:19:50.059928
12405	1-12405	5	42	3384	100	4375	0	4	2012-08-02 11:19:56.268626	2012-08-02 11:19:56.268626
10862	1-10862	32	48	4896	52	5638	0	4	2012-08-02 10:38:28.053742	2012-08-02 11:19:57.417779
12411	1-12411	22	26	182	66	1284	0	2	2012-08-02 11:19:58.568741	2012-08-02 11:19:58.568741
8135	1-8135	99	24	3582	66	3428	0	7	2012-08-02 05:37:17.561546	2012-08-02 11:20:02.026832
12420	1-12420	10	66	3074	23	7069	0	3	2012-08-02 11:20:02.026832	2012-08-02 11:20:02.026832
12389	1-12389	78	5	4130	17	9298	0	0	2012-08-02 11:19:50.059928	2012-08-02 11:20:03.165061
7375	1-7375	32	18	4201	42	3613	0	0	2012-08-02 05:32:02.79172	2012-08-02 11:20:04.336464
11931	1-11931	13	80	7431	64	4297	0	0	2012-08-02 11:16:52.562619	2012-08-02 11:20:04.336464
12426	1-12426	94	64	1430	18	4511	0	0	2012-08-02 11:20:04.336464	2012-08-02 11:20:04.336464
9658	1-9658	65	39	6991	80	1964	1964	0	2012-08-02 05:47:36.050416	2012-08-02 11:20:04.336464
5339	1-5339	9	39	3614	80	824	824	0	2012-08-02 05:18:17.62484	2012-08-02 11:20:04.336464
8444	1-8444	46	39	4117	80	3570	2598	0	2012-08-02 05:39:22.736944	2012-08-02 11:20:04.336464
11198	1-11198	30	13	9592	51	8405	0	3	2012-08-02 11:12:08.164732	2012-08-02 11:20:07.430521
10067	1-10067	22	55	4992	54	4562	0	0	2012-08-02 10:21:07.012533	2012-08-02 11:20:07.430521
12434	1-12434	96	91	2320	55	9762	0	3	2012-08-02 11:20:07.430521	2012-08-02 11:20:07.430521
5735	1-5735	79	13	4872	51	1167	1167	3	2012-08-02 05:20:56.009649	2012-08-02 11:20:07.430521
12419	1-12419	68	57	2770	95	9697	0	0	2012-08-02 11:20:01.662484	2012-08-02 11:20:08.601905
7918	1-7918	94	57	2473	44	1376	0	0	2012-08-02 05:35:48.936465	2012-08-02 11:20:08.601905
12437	1-12437	51	98	696	57	3294	0	0	2012-08-02 11:20:08.601905	2012-08-02 11:20:08.601905
9763	1-9763	11	93	7233	73	4776	0	0	2012-08-02 05:48:19.090843	2012-08-02 11:20:09.761988
666	1-666	13	30	6069	46	3296	0	0	2012-08-02 04:44:51.739434	2012-08-02 11:20:14.432725
12004	1-12004	31	10	5535	9	7444	0	0	2012-08-02 11:17:21.346268	2012-08-02 11:20:14.432725
11733	1-11733	69	99	7971	9	4502	0	2	2012-08-02 11:15:35.411078	2012-08-02 11:20:14.432725
4234	1-4234	87	9	6795	46	4281	366	0	2012-08-02 05:10:54.578342	2012-08-02 11:20:14.432725
5232	1-5232	46	9	9016	46	231	231	0	2012-08-02 05:17:35.592176	2012-08-02 11:20:14.432725
4329	1-4329	78	9	3420	46	663	663	0	2012-08-02 05:11:33.021361	2012-08-02 11:20:14.432725
12306	1-12306	8	39	4409	80	9358	0	9	2012-08-02 11:19:18.027812	2012-08-02 11:20:16.785814
12355	1-12355	14	28	522	46	642	0	0	2012-08-02 11:19:36.911105	2012-08-02 11:20:16.785814
10540	1-10540	53	39	9981	64	8280	0	4	2012-08-02 10:32:12.765024	2012-08-02 11:20:16.785814
12458	1-12458	34	46	1624	39	7160	0	0	2012-08-02 11:20:16.785814	2012-08-02 11:20:16.785814
12107	1-12107	30	30	6020	93	9558	0	3	2012-08-02 11:18:01.37723	2012-08-02 11:20:20.346337
12330	1-12330	4	10	3249	82	5611	0	2	2012-08-02 11:19:27.297519	2012-08-02 11:20:20.346337
12340	1-12340	93	30	7165	10	5872	0	8	2012-08-02 11:19:30.722095	2012-08-02 11:20:20.346337
12467	1-12467	92	60	3547	30	5598	0	5	2012-08-02 11:20:20.346337	2012-08-02 11:20:20.346337
12196	1-12196	28	41	3927	86	9546	0	5	2012-08-02 11:18:35.440957	2012-08-02 11:20:21.542331
12470	1-12470	49	92	2097	63	3429	0	4	2012-08-02 11:20:21.542331	2012-08-02 11:20:21.542331
7295	1-7295	21	72	2983	2	3203	0	0	2012-08-02 05:31:31.280374	2012-08-02 11:20:23.865399
12476	1-12476	3	2	3312	23	2816	0	0	2012-08-02 11:20:23.865399	2012-08-02 11:20:23.865399
3732	1-3732	39	23	9667	40	7741	1758	0	2012-08-02 05:07:34.851637	2012-08-02 11:20:23.865399
9910	1-9910	48	14	8015	9	3861	0	6	2012-08-02 05:49:18.826401	2012-08-02 11:20:25.038434
11251	1-11251	35	42	6798	41	3745	0	0	2012-08-02 11:12:28.595194	2012-08-02 11:20:25.038434
7206	1-7206	51	9	9704	42	9826	4980	0	2012-08-02 05:30:55.919734	2012-08-02 11:20:25.038434
12452	1-12452	28	46	2733	30	6933	0	3	2012-08-02 11:20:14.432725	2012-08-02 11:20:27.394391
12192	1-12192	10	20	1623	31	1890	0	0	2012-08-02 11:18:33.949334	2012-08-02 11:20:27.394391
11866	1-11866	46	46	6618	20	7097	0	0	2012-08-02 11:16:27.373701	2012-08-02 11:20:27.394391
12256	1-12256	44	46	4877	76	7879	0	6	2012-08-02 11:18:58.625043	2012-08-02 11:20:27.394391
12485	1-12485	84	80	2245	46	5989	0	8	2012-08-02 11:20:27.394391	2012-08-02 11:20:27.394391
11973	1-11973	53	44	2443	25	4103	0	12	2012-08-02 11:17:09.190931	2012-08-02 11:20:28.558352
12488	1-12488	28	25	4175	30	5213	0	9	2012-08-02 11:20:28.558352	2012-08-02 11:20:28.558352
10653	1-10653	62	26	2470	2	4801	0	4	2012-08-02 10:36:02.695478	2012-08-02 11:20:30.879963
12439	1-12439	73	62	1075	87	9562	0	5	2012-08-02 11:20:09.364057	2012-08-02 11:20:31.675507
12517	1-12517	36	73	1176	97	5275	0	6	2012-08-02 11:20:39.789928	2012-08-02 11:20:39.789928
12520	1-12520	65	14	1800	70	7872	0	0	2012-08-02 11:20:40.961064	2012-08-02 11:20:40.961064
12180	1-12180	84	92	9063	44	8523	0	0	2012-08-02 11:18:29.364252	2012-08-02 11:20:48.789151
11794	1-11794	8	76	9965	73	5487	0	7	2012-08-02 11:15:59.409779	2012-08-02 11:20:51.098418
12468	1-12468	63	97	3724	33	8720	0	17	2012-08-02 11:20:20.732595	2012-08-02 11:20:51.098418
9009	1-9009	65	64	9712	41	6072	0	11	2012-08-02 05:43:11.375575	2012-08-02 11:20:51.098418
12546	1-12546	87	41	1251	76	9510	0	0	2012-08-02 11:20:51.098418	2012-08-02 11:20:51.098418
8863	1-8863	49	33	9102	47	8204	0	0	2012-08-02 05:42:11.688407	2012-08-02 11:20:52.292349
12064	1-12064	71	28	3803	51	8987	0	0	2012-08-02 11:17:44.926177	2012-08-02 11:19:20.723102
8504	1-8504	32	91	4715	68	2283	0	9	2012-08-02 05:39:46.746721	2012-08-02 11:19:21.872433
2745	1-2745	77	81	7947	8	897	0	0	2012-08-02 05:00:53.274934	2012-08-02 11:19:21.872433
12316	1-12316	32	68	1862	81	9867	0	12	2012-08-02 11:19:21.872433	2012-08-02 11:19:21.872433
5916	1-5916	86	41	9509	8	3483	57	0	2012-08-02 05:22:09.509574	2012-08-02 11:19:21.872433
11652	1-11652	19	19	4962	26	9217	0	20	2012-08-02 11:15:03.566223	2012-08-02 11:19:25.352516
11831	1-11831	61	82	9752	29	7632	0	7	2012-08-02 11:16:13.90738	2012-08-02 11:19:25.352516
12325	1-12325	20	67	286	94	9353	0	0	2012-08-02 11:19:25.352516	2012-08-02 11:19:25.352516
11183	1-11183	96	70	8241	9	6661	0	6	2012-08-02 11:12:01.911108	2012-08-02 11:19:26.512961
10910	1-10910	51	70	7727	9	4389	3088	4	2012-08-02 10:38:47.226393	2012-08-02 11:19:26.512961
9368	1-9368	7	80	3697	61	1211	0	0	2012-08-02 05:45:37.984048	2012-08-02 11:19:31.131315
12201	1-12201	89	61	6218	58	7550	0	12	2012-08-02 11:18:37.363591	2012-08-02 11:19:31.131315
7364	1-7364	22	6	9469	58	5851	0	0	2012-08-02 05:31:58.427569	2012-08-02 11:19:31.131315
12339	1-12339	48	58	2768	32	8940	0	5	2012-08-02 11:19:31.131315	2012-08-02 11:19:31.131315
11519	1-11519	92	24	1489	65	1117	0	7	2012-08-02 11:14:12.394694	2012-08-02 11:19:34.22667
12348	1-12348	77	61	624	24	3149	0	16	2012-08-02 11:19:34.22667	2012-08-02 11:19:34.22667
11995	1-11995	22	1	1296	15	7585	0	4	2012-08-02 11:17:17.832978	2012-08-02 11:19:37.662313
11587	1-11587	52	1	4618	79	4653	0	0	2012-08-02 11:14:38.471269	2012-08-02 11:19:37.662313
12357	1-12357	49	15	9318	1	8275	0	3	2012-08-02 11:19:37.662313	2012-08-02 11:19:39.971519
12369	1-12369	38	16	5797	74	2372	0	4	2012-08-02 11:19:42.280544	2012-08-02 11:19:42.280544
12372	1-12372	36	13	2317	39	6791	0	0	2012-08-02 11:19:43.441353	2012-08-02 11:19:43.441353
12375	1-12375	54	86	1598	1	1955	0	0	2012-08-02 11:19:44.623779	2012-08-02 11:19:44.623779
12381	1-12381	86	80	842	21	1778	0	8	2012-08-02 11:19:46.954926	2012-08-02 11:19:46.954926
12390	1-12390	51	11	47	24	7248	0	4	2012-08-02 11:19:50.435086	2012-08-02 11:19:50.435086
12011	1-12011	68	10	3333	79	7609	0	0	2012-08-02 11:17:24.129968	2012-08-02 11:19:55.848666
10582	1-10582	36	97	8642	94	9560	0	0	2012-08-02 10:35:35.413328	2012-08-02 11:19:55.848666
12275	1-12275	50	78	3451	39	9913	0	5	2012-08-02 11:19:05.885639	2012-08-02 11:19:57.030953
6318	1-6318	1	43	8132	59	4370	0	0	2012-08-02 05:24:55.246735	2012-08-02 11:19:57.030953
12407	1-12407	16	39	2348	43	9414	0	0	2012-08-02 11:19:57.030953	2012-08-02 11:19:57.030953
12413	1-12413	25	33	2084	79	5374	0	0	2012-08-02 11:19:59.331014	2012-08-02 11:19:59.331014
11433	1-11433	84	89	5745	46	5607	0	5	2012-08-02 11:13:39.191212	2012-08-02 11:20:00.480137
12079	1-12079	35	89	2682	53	3323	0	6	2012-08-02 11:17:50.648919	2012-08-02 11:20:00.480137
11173	1-11173	72	89	4165	59	2811	0	0	2012-08-02 11:11:57.975642	2012-08-02 11:20:00.480137
12416	1-12416	40	50	1578	89	7830	0	4	2012-08-02 11:20:00.480137	2012-08-02 11:20:00.480137
7704	1-7704	70	61	9335	57	7701	0	0	2012-08-02 05:34:18.132956	2012-08-02 11:20:01.662484
11891	1-11891	5	95	8043	47	9367	0	0	2012-08-02 11:16:36.990477	2012-08-02 11:20:01.662484
12022	1-12022	19	95	7764	79	8050	0	3	2012-08-02 11:17:28.450234	2012-08-02 11:20:01.662484
12208	1-12208	31	38	4270	17	7906	0	8	2012-08-02 11:18:40.083726	2012-08-02 11:20:02.789408
12422	1-12422	67	17	5287	49	6545	0	5	2012-08-02 11:20:02.789408	2012-08-02 11:20:02.789408
10861	1-10861	80	24	4688	54	7545	0	5	2012-08-02 10:38:27.656004	2012-08-02 11:20:05.098618
8675	1-8675	99	24	9108	54	9733	0	0	2012-08-02 05:40:55.498183	2012-08-02 11:20:05.098618
11725	1-11725	55	24	3700	25	9969	0	0	2012-08-02 11:15:32.313624	2012-08-02 11:20:05.098618
3270	1-3270	49	24	5761	42	3632	0	3	2012-08-02 05:04:28.490902	2012-08-02 11:20:05.098618
12428	1-12428	46	54	3778	24	6841	0	0	2012-08-02 11:20:05.098618	2012-08-02 11:20:05.098618
11914	1-11914	55	87	4242	64	2651	0	6	2012-08-02 11:16:45.897988	2012-08-02 11:20:09.364057
12448	1-12448	4	93	239	96	6274	0	6	2012-08-02 11:20:12.869762	2012-08-02 11:20:12.869762
12262	1-12262	99	9	7024	63	7578	0	5	2012-08-02 11:19:00.934049	2012-08-02 11:20:16.398432
5373	1-5373	84	64	8720	59	5181	0	0	2012-08-02 05:18:30.926878	2012-08-02 11:20:18.808868
7669	1-7669	21	1	8341	97	6097	0	6	2012-08-02 05:34:03.448614	2012-08-02 11:20:20.732595
12037	1-12037	72	26	4260	67	8719	0	0	2012-08-02 11:17:34.261328	2012-08-02 11:20:25.436242
12490	1-12490	4	93	246	28	6431	0	0	2012-08-02 11:20:29.320022	2012-08-02 11:20:29.320022
12494	1-12494	85	27	918	72	2637	0	3	2012-08-02 11:20:30.504167	2012-08-02 11:20:30.504167
8851	1-8851	74	96	5197	5	7191	0	7	2012-08-02 05:42:06.816204	2012-08-02 11:20:32.051483
1308	1-1308	52	30	5599	90	3445	0	0	2012-08-02 04:50:00.316371	2012-08-02 11:20:34.782688
11906	1-11906	100	80	3323	70	7415	0	4	2012-08-02 11:16:43.111969	2012-08-02 11:20:34.782688
9355	1-9355	87	30	7632	86	7170	0	5	2012-08-02 05:45:32.702923	2012-08-02 11:20:34.782688
12504	1-12504	38	90	1813	30	7165	0	5	2012-08-02 11:20:34.782688	2012-08-02 11:20:34.782688
10849	1-10849	96	11	3079	56	2254	0	0	2012-08-02 10:38:22.938372	2012-08-02 11:20:35.943322
12513	1-12513	56	32	32	24	9916	0	9	2012-08-02 11:20:38.263355	2012-08-02 11:20:38.263355
12527	1-12527	48	34	241	4	7190	0	6	2012-08-02 11:20:43.717401	2012-08-02 11:20:43.717401
12489	1-12489	67	84	3523	10	9694	0	0	2012-08-02 11:20:28.944385	2012-08-02 11:20:46.071419
12276	1-12276	91	5	6663	49	8837	0	11	2012-08-02 11:19:06.27261	2012-08-02 11:20:46.071419
12203	1-12203	41	26	298	16	767	0	8	2012-08-02 11:18:38.14812	2012-08-02 11:20:53.075923
11916	1-11916	75	16	7670	42	8187	0	0	2012-08-02 11:16:46.660464	2012-08-02 11:20:53.075923
8987	1-8987	34	9	8766	59	6096	0	4	2012-08-02 05:43:02.336727	2012-08-02 11:20:54.260199
12036	1-12036	38	60	6002	87	9097	0	3	2012-08-02 11:17:33.896897	2012-08-02 11:20:55.454389
12557	1-12557	81	7	937	51	6399	0	8	2012-08-02 11:20:55.454389	2012-08-02 11:20:55.454389
11786	1-11786	30	13	1054	75	728	0	5	2012-08-02 11:15:56.271365	2012-08-02 11:20:56.648337
1063	1-1063	59	41	7035	14	2904	2064	0	2012-08-02 04:47:49.127561	2012-08-02 11:20:56.648337
9303	1-9303	66	28	6797	11	3705	0	8	2012-08-02 05:45:11.464939	2012-08-02 11:20:57.830405
12563	1-12563	100	44	1525	28	6676	0	0	2012-08-02 11:20:57.830405	2012-08-02 11:20:57.830405
12491	1-12491	18	51	2536	60	9327	0	0	2012-08-02 11:20:29.706997	2012-08-02 11:20:58.97991
12566	1-12566	27	95	585	51	8962	0	3	2012-08-02 11:20:58.97991	2012-08-02 11:20:58.97991
11464	1-11464	27	38	2984	56	970	0	5	2012-08-02 11:13:51.124244	2012-08-02 11:21:04.00732
4414	1-4414	63	38	3656	48	1062	0	2	2012-08-02 05:12:07.781887	2012-08-02 11:21:04.00732
11988	1-11988	10	67	3946	16	8724	0	0	2012-08-02 11:17:15.081578	2012-08-02 11:21:04.00732
12579	1-12579	92	48	2646	67	3763	0	0	2012-08-02 11:21:04.00732	2012-08-02 11:21:04.00732
12588	1-12588	59	3	1381	23	4820	0	3	2012-08-02 11:21:07.587592	2012-08-02 11:21:07.587592
11961	1-11961	26	52	3099	81	3940	0	0	2012-08-02 11:17:04.37437	2012-08-02 11:21:09.963143
12594	1-12594	50	70	212	52	782	0	8	2012-08-02 11:21:09.963143	2012-08-02 11:21:09.963143
11453	1-11453	59	100	7482	50	9866	0	0	2012-08-02 11:13:46.948109	2012-08-02 11:21:12.327546
12549	1-12549	81	47	578	33	6854	0	14	2012-08-02 11:20:52.292349	2012-08-02 11:20:52.292349
11181	1-11181	10	72	6849	56	4156	0	0	2012-08-02 11:12:01.093497	2012-08-02 11:20:53.46267
11279	1-11279	24	52	6657	35	4688	0	9	2012-08-02 11:12:39.48975	2012-08-02 11:20:53.46267
12552	1-12552	88	44	1173	72	6640	0	0	2012-08-02 11:20:53.46267	2012-08-02 11:20:53.46267
12555	1-12555	31	55	2153	94	3615	0	6	2012-08-02 11:20:54.669447	2012-08-02 11:20:54.669447
8431	1-8431	79	6	1817	12	657	0	7	2012-08-02 05:39:17.541672	2012-08-02 11:20:57.045909
10871	1-10871	42	56	3989	44	3263	0	8	2012-08-02 10:38:31.600754	2012-08-02 11:20:57.045909
12561	1-12561	100	12	1105	99	8470	0	6	2012-08-02 11:20:57.045909	2012-08-02 11:20:57.045909
12077	1-12077	97	98	4127	5	6783	0	0	2012-08-02 11:17:49.897874	2012-08-02 11:21:02.460876
12168	1-12168	31	6	7292	71	9496	0	2	2012-08-02 11:18:24.845446	2012-08-02 11:21:02.460876
10416	1-10416	83	99	6492	71	5091	0	0	2012-08-02 10:27:50.870709	2012-08-02 11:21:02.460876
5120	1-5120	31	64	7047	85	4584	0	2	2012-08-02 05:16:51.13798	2012-08-02 11:21:03.609721
12578	1-12578	73	85	2944	64	6213	0	0	2012-08-02 11:21:03.609721	2012-08-02 11:21:03.609721
12232	1-12232	2	63	5668	41	9616	0	9	2012-08-02 11:18:49.343681	2012-08-02 11:21:04.780881
11997	1-11997	90	63	3838	32	7346	0	0	2012-08-02 11:17:19.003673	2012-08-02 11:21:04.780881
761	1-761	30	98	3850	68	912	0	0	2012-08-02 04:45:18.750953	2012-08-02 11:21:04.780881
12581	1-12581	94	68	335	63	4127	0	9	2012-08-02 11:21:04.780881	2012-08-02 11:21:04.780881
4896	1-4896	42	38	6985	90	3522	0	0	2012-08-02 05:15:21.593955	2012-08-02 11:21:05.963458
8957	1-8957	52	34	7901	23	3760	0	14	2012-08-02 05:42:50.018083	2012-08-02 11:21:05.963458
11669	1-11669	19	54	7563	32	7102	0	3	2012-08-02 11:15:10.206315	2012-08-02 11:21:05.963458
9640	1-9640	83	34	5023	14	3540	0	0	2012-08-02 05:47:28.590426	2012-08-02 11:21:05.963458
12292	1-12292	80	20	7284	36	6230	0	6	2012-08-02 11:19:12.492452	2012-08-02 11:21:08.372282
12584	1-12584	29	90	2387	34	9902	0	12	2012-08-02 11:21:05.963458	2012-08-02 11:21:10.736867
8243	1-8243	48	32	7340	25	5979	0	0	2012-08-02 05:38:01.542564	2012-08-02 11:21:10.736867
12001	1-12001	39	21	5291	91	4501	0	11	2012-08-02 11:17:20.185731	2012-08-02 11:21:14.274664
12600	1-12600	95	83	4205	7	7353	0	11	2012-08-02 11:21:12.327546	2012-08-02 11:21:19.765527
12619	1-12619	78	32	1823	41	3535	0	0	2012-08-02 11:21:19.765527	2012-08-02 11:21:19.765527
12622	1-12622	64	69	299	33	3487	0	0	2012-08-02 11:21:20.937522	2012-08-02 11:21:20.937522
11229	1-11229	67	8	7143	22	9013	0	0	2012-08-02 11:12:19.963875	2012-08-02 11:21:30.273011
12646	1-12646	75	89	780	8	3117	0	4	2012-08-02 11:21:30.273011	2012-08-02 11:21:30.273011
11732	1-11732	83	84	4248	5	4379	0	11	2012-08-02 11:15:35.035492	2012-08-02 11:21:31.433183
12354	1-12354	46	67	913	80	1452	0	0	2012-08-02 11:19:36.546508	2012-08-02 11:21:31.433183
11192	1-11192	89	73	7100	34	9829	0	0	2012-08-02 11:12:05.424338	2012-08-02 11:21:31.433183
10589	1-10589	99	74	3820	71	5040	0	0	2012-08-02 10:35:35.937998	2012-08-02 11:21:31.433183
12424	1-12424	81	7	4340	21	6505	0	7	2012-08-02 11:20:03.551811	2012-08-02 11:21:34.980127
12308	1-12308	64	61	3645	69	5198	0	0	2012-08-02 11:19:18.41467	2012-08-02 11:21:38.117688
12666	1-12666	28	69	1257	7	8331	0	6	2012-08-02 11:21:38.117688	2012-08-02 11:21:38.117688
12669	1-12669	62	94	17	16	3211	0	3	2012-08-02 11:21:39.302038	2012-08-02 11:21:39.302038
12672	1-12672	53	5	348	81	8042	0	15	2012-08-02 11:21:40.473127	2012-08-02 11:21:40.473127
12427	1-12427	29	53	4021	16	9338	0	0	2012-08-02 11:20:04.723067	2012-08-02 11:21:41.655283
11748	1-11748	18	26	2891	16	6362	0	5	2012-08-02 11:15:41.278827	2012-08-02 11:21:41.655283
12675	1-12675	77	16	3366	19	6501	0	3	2012-08-02 11:21:41.655283	2012-08-02 11:21:41.655283
11696	1-11696	3	23	5469	50	6120	0	9	2012-08-02 11:15:20.978899	2012-08-02 11:21:42.881726
11562	1-11562	11	50	4067	12	2036	0	0	2012-08-02 11:14:28.859575	2012-08-02 11:21:42.881726
7825	1-7825	68	6	8677	44	7241	0	11	2012-08-02 05:35:09.603683	2012-08-02 11:21:42.881726
12679	1-12679	98	12	1432	6	5603	0	4	2012-08-02 11:21:42.881726	2012-08-02 11:21:42.881726
9717	1-9717	49	73	8345	46	9122	0	0	2012-08-02 05:48:00.315286	2012-08-02 11:21:47.236116
12689	1-12689	73	92	1995	73	2935	0	0	2012-08-02 11:21:47.236116	2012-08-02 11:21:47.236116
12649	1-12649	30	52	9612	74	9610	0	0	2012-08-02 11:21:31.433183	2012-08-02 11:21:48.407505
10172	1-10172	90	52	4057	42	2232	0	0	2012-08-02 10:22:12.533639	2012-08-02 11:21:48.407505
11565	1-11565	48	67	6911	21	8833	0	0	2012-08-02 11:14:30.008411	2012-08-02 11:21:48.407505
12692	1-12692	40	21	530	52	5691	0	0	2012-08-02 11:21:48.407505	2012-08-02 11:21:48.407505
11586	1-11586	12	6	9558	23	6942	0	0	2012-08-02 11:14:38.10669	2012-08-02 11:21:51.843443
12099	1-12099	75	49	3260	6	4315	0	0	2012-08-02 11:17:58.316953	2012-08-02 11:21:53.025745
12704	1-12704	67	42	2290	49	5193	0	0	2012-08-02 11:21:53.025745	2012-08-02 11:21:53.025745
11415	1-11415	71	50	4006	54	3178	0	5	2012-08-02 11:13:32.308335	2012-08-02 11:21:54.197115
12707	1-12707	87	54	972	27	1823	0	0	2012-08-02 11:21:54.197115	2012-08-02 11:21:54.197115
11608	1-11608	56	4	5138	65	4833	0	9	2012-08-02 11:14:46.493927	2012-08-02 11:21:59.666365
12725	1-12725	25	60	6958	19	1858	0	0	2012-08-02 11:22:01.235112	2012-08-02 11:22:01.235112
12402	1-12402	12	23	7697	8	7224	0	17	2012-08-02 11:19:54.677644	2012-08-02 11:22:16.693227
7140	1-7140	56	71	9889	100	5067	0	3	2012-08-02 05:30:28.441122	2012-08-02 11:22:17.996906
12378	1-12378	62	49	7040	48	6406	0	13	2012-08-02 11:19:45.7837	2012-08-02 11:22:35.896828
11113	1-11113	40	48	7653	70	8713	0	0	2012-08-02 11:11:34.331609	2012-08-02 11:22:35.896828
12015	1-12015	55	36	2557	49	5400	0	7	2012-08-02 11:17:25.732459	2012-08-02 11:22:35.896828
12814	1-12814	66	70	5763	36	6766	0	0	2012-08-02 11:22:35.896828	2012-08-02 11:22:35.896828
12801	1-12801	56	99	1285	48	3470	0	6	2012-08-02 11:22:30.503719	2012-08-02 11:22:39.388621
8412	1-8412	53	11	7257	82	6818	0	4	2012-08-02 05:39:09.861087	2012-08-02 11:22:39.388621
12506	1-12506	12	11	4541	46	5104	0	5	2012-08-02 11:20:35.545183	2012-08-02 11:22:39.388621
12823	1-12823	35	41	1827	11	9150	0	15	2012-08-02 11:22:39.388621	2012-08-02 11:22:39.388621
12174	1-12174	20	80	3678	85	2530	0	3	2012-08-02 11:18:27.121335	2012-08-02 11:22:40.604474
12130	1-12130	8	23	7189	89	9096	0	9	2012-08-02 11:18:10.306096	2012-08-02 11:22:41.775281
12599	1-12599	39	17	6856	11	5409	0	11	2012-08-02 11:21:11.919133	2012-08-02 11:22:41.775281
12829	1-12829	83	89	890	17	5797	0	12	2012-08-02 11:22:41.775281	2012-08-02 11:22:41.775281
8272	1-8272	24	43	8127	71	4981	0	4	2012-08-02 05:38:13.276538	2012-08-02 11:22:43.743182
12834	1-12834	74	71	1164	43	9404	0	10	2012-08-02 11:22:43.743182	2012-08-02 11:22:43.743182
12562	1-12562	78	87	6163	4	7135	0	8	2012-08-02 11:20:57.432968	2012-08-02 11:22:44.903286
11408	1-11408	48	46	4252	92	9310	0	0	2012-08-02 11:13:29.601391	2012-08-02 11:22:46.074313
12737	1-12737	83	82	2838	29	5059	0	16	2012-08-02 11:22:05.876967	2012-08-02 11:22:46.074313
12840	1-12840	94	92	3437	24	4775	0	3	2012-08-02 11:22:46.074313	2012-08-02 11:22:46.074313
12182	1-12182	93	24	2853	18	1410	0	10	2012-08-02 11:18:30.126763	2012-08-02 11:22:47.278678
12815	1-12815	65	54	4568	24	3388	0	6	2012-08-02 11:22:36.283935	2012-08-02 11:22:47.278678
10685	1-10685	16	7	8264	91	4876	0	12	2012-08-02 10:36:15.263491	2012-08-02 11:21:12.327546
12575	1-12575	84	71	2943	98	8251	0	0	2012-08-02 11:21:02.460876	2012-08-02 11:21:17.003001
12156	1-12156	91	52	7804	76	4926	0	6	2012-08-02 11:18:20.238255	2012-08-02 11:21:17.003001
12212	1-12212	79	79	8513	94	9284	0	0	2012-08-02 11:18:42.017055	2012-08-02 11:21:17.003001
12612	1-12612	42	76	2395	71	9532	0	8	2012-08-02 11:21:17.003001	2012-08-02 11:21:17.003001
6404	1-6404	29	71	5769	100	250	250	0	2012-08-02 05:25:30.531034	2012-08-02 11:21:17.003001
11636	1-11636	40	70	8014	90	4444	0	5	2012-08-02 11:14:57.356667	2012-08-02 11:21:19.356773
2184	1-2184	98	70	9624	100	2023	0	0	2012-08-02 04:56:50.48839	2012-08-02 11:21:19.356773
6870	1-6870	27	70	487	100	96	0	2	2012-08-02 05:28:38.976518	2012-08-02 11:21:19.356773
12618	1-12618	31	67	1696	70	9505	0	5	2012-08-02 11:21:19.356773	2012-08-02 11:21:19.356773
3542	1-3542	12	32	5873	96	1275	720	0	2012-08-02 05:06:18.611973	2012-08-02 11:21:19.356773
11505	1-11505	17	57	7610	27	5954	0	6	2012-08-02 11:14:06.560633	2012-08-02 11:21:22.892901
12627	1-12627	46	19	1222	57	8369	0	0	2012-08-02 11:21:22.892901	2012-08-02 11:21:22.892901
12630	1-12630	93	66	175	58	2872	0	4	2012-08-02 11:21:24.042018	2012-08-02 11:21:24.042018
11612	1-11612	80	32	5086	80	8490	0	0	2012-08-02 11:14:48.086792	2012-08-02 11:21:25.213371
11829	1-11829	43	99	5176	56	4090	0	7	2012-08-02 11:16:13.156317	2012-08-02 11:21:31.046356
12647	1-12647	14	25	931	99	8019	0	0	2012-08-02 11:21:31.046356	2012-08-02 11:21:31.046356
10959	1-10959	43	97	3420	66	2141	0	0	2012-08-02 10:39:06.307853	2012-08-02 11:21:35.764971
9120	1-9120	59	49	9099	23	5576	0	0	2012-08-02 05:43:57.025334	2012-08-02 11:21:35.764971
12660	1-12660	39	76	3946	97	8902	0	19	2012-08-02 11:21:35.764971	2012-08-02 11:21:35.764971
12633	1-12633	77	46	2504	7	4685	0	12	2012-08-02 11:21:25.213371	2012-08-02 11:21:42.053335
12616	1-12616	87	8	311	55	715	0	0	2012-08-02 11:21:18.959069	2012-08-02 11:21:42.053335
12057	1-12057	97	24	3288	68	5519	0	11	2012-08-02 11:17:42.194368	2012-08-02 11:21:42.053335
12676	1-12676	48	16	4953	8	8202	0	0	2012-08-02 11:21:42.053335	2012-08-02 11:21:42.053335
10501	1-10501	66	11	8242	66	5356	0	8	2012-08-02 10:31:57.323873	2012-08-02 11:21:46.429608
12687	1-12687	28	66	1513	59	4167	0	9	2012-08-02 11:21:46.429608	2012-08-02 11:21:46.429608
4638	1-4638	80	32	3929	13	1434	0	6	2012-08-02 05:13:37.5709	2012-08-02 11:21:53.412361
9797	1-9797	21	32	3518	6	2220	0	5	2012-08-02 05:48:32.835784	2012-08-02 11:21:53.412361
11205	1-11205	72	29	2451	67	1882	0	10	2012-08-02 11:12:10.484612	2012-08-02 11:21:53.412361
10588	1-10588	60	32	5370	29	2518	0	0	2012-08-02 10:35:35.860417	2012-08-02 11:21:53.412361
12705	1-12705	95	31	235	32	8527	0	6	2012-08-02 11:21:53.412361	2012-08-02 11:21:53.412361
12415	1-12415	95	74	6306	6	9103	0	4	2012-08-02 11:20:00.082316	2012-08-02 11:21:57.347689
12023	1-12023	46	44	7072	6	8176	0	19	2012-08-02 11:17:28.826119	2012-08-02 11:21:57.347689
12056	1-12056	46	38	8305	6	9842	0	5	2012-08-02 11:17:41.796561	2012-08-02 11:21:57.347689
12383	1-12383	32	78	2242	7	9917	0	8	2012-08-02 11:19:47.728103	2012-08-02 11:22:04.739061
3649	1-3649	10	78	9531	7	1429	1429	0	2012-08-02 05:07:01.717028	2012-08-02 11:22:04.739061
12171	1-12171	13	5	6996	13	5942	0	8	2012-08-02 11:18:25.972634	2012-08-02 11:22:05.876967
12228	1-12228	92	59	6008	64	9926	0	0	2012-08-02 11:18:47.433288	2012-08-02 11:22:07.015042
12742	1-12742	32	64	1447	59	3139	0	5	2012-08-02 11:22:07.015042	2012-08-02 11:22:07.015042
11187	1-11187	16	26	5453	95	8007	0	8	2012-08-02 11:12:03.490988	2012-08-02 11:22:11.235432
11436	1-11436	56	78	5683	63	5255	0	8	2012-08-02 11:13:40.362667	2012-08-02 11:22:11.235432
11940	1-11940	86	78	6036	26	7082	0	0	2012-08-02 11:16:56.109668	2012-08-02 11:22:11.235432
11450	1-11450	48	51	5712	34	6285	0	7	2012-08-02 11:13:45.798923	2012-08-02 11:22:18.372422
12507	1-12507	81	56	1939	48	5556	0	3	2012-08-02 11:20:35.943322	2012-08-02 11:22:19.576837
9762	1-9762	27	51	7151	48	5952	0	0	2012-08-02 05:48:18.693192	2012-08-02 11:22:19.576837
12772	1-12772	46	48	3908	100	6542	0	3	2012-08-02 11:22:19.576837	2012-08-02 11:22:19.576837
11948	1-11948	19	88	8478	53	8034	0	6	2012-08-02 11:16:59.26956	2012-08-02 11:22:20.770288
12595	1-12595	33	77	6428	90	3215	0	0	2012-08-02 11:21:10.360732	2012-08-02 11:22:28.150301
12794	1-12794	54	90	79	77	300	0	15	2012-08-02 11:22:28.150301	2012-08-02 11:22:28.150301
8377	1-8377	90	83	8947	7	4619	0	4	2012-08-02 05:38:55.673998	2012-08-02 11:22:29.332704
10373	1-10373	75	55	5622	99	4525	0	0	2012-08-02 10:27:33.853435	2012-08-02 11:22:30.503719
12436	1-12436	2	48	534	17	414	0	0	2012-08-02 11:20:08.214784	2012-08-02 11:22:30.503719
11581	1-11581	74	48	9011	42	7047	0	0	2012-08-02 11:14:36.162578	2012-08-02 11:22:30.503719
8197	1-8197	14	85	4598	88	2887	0	0	2012-08-02 05:37:42.908714	2012-08-02 11:22:36.670154
2738	1-2738	1	54	6916	4	1690	0	0	2012-08-02 05:00:50.347154	2012-08-02 11:22:36.670154
11985	1-11985	13	7	8921	32	8963	0	12	2012-08-02 11:17:13.864543	2012-08-02 11:22:36.670154
12816	1-12816	19	93	1383	54	9201	0	0	2012-08-02 11:22:36.670154	2012-08-02 11:22:36.670154
9699	1-9699	56	31	8289	53	4475	0	0	2012-08-02 05:47:52.921368	2012-08-02 11:22:40.161601
11297	1-11297	52	31	8407	16	9303	0	0	2012-08-02 11:12:46.143973	2012-08-02 11:22:40.161601
533	1-533	18	7	8791	16	1639	410	0	2012-08-02 04:44:19.051821	2012-08-02 11:22:40.161601
11099	1-11099	74	54	8647	7	6461	0	0	2012-08-02 11:11:28.925859	2012-08-02 11:22:42.548342
12681	1-12681	12	92	9310	16	7664	0	4	2012-08-02 11:21:44.087377	2012-08-02 11:22:42.548342
12756	1-12756	21	68	1608	14	4528	0	14	2012-08-02 11:22:13.191049	2012-08-02 11:22:42.548342
12475	1-12475	37	68	6601	36	5513	0	14	2012-08-02 11:20:23.49027	2012-08-02 11:22:42.548342
12347	1-12347	15	79	4826	86	8242	0	0	2012-08-02 11:19:33.839735	2012-08-02 11:22:47.665344
6017	1-6017	59	79	6414	56	1692	0	0	2012-08-02 05:22:51.586931	2012-08-02 11:22:47.665344
11740	1-11740	83	93	7885	44	8766	0	2	2012-08-02 11:15:37.764672	2012-08-02 11:22:47.665344
12859	1-12859	81	1	438	59	9618	0	0	2012-08-02 11:22:53.543014	2012-08-02 11:22:53.543014
12765	1-12765	3	75	1591	23	4750	0	0	2012-08-02 11:22:16.693227	2012-08-02 11:22:57.509065
12870	1-12870	70	45	1737	75	5965	0	0	2012-08-02 11:22:57.509065	2012-08-02 11:22:57.509065
12572	1-12572	92	12	3953	93	7080	0	11	2012-08-02 11:21:01.687092	2012-08-02 11:22:59.10205
12821	1-12821	67	12	1180	50	7034	0	5	2012-08-02 11:22:38.626019	2012-08-02 11:22:59.10205
11388	1-11388	11	18	4992	53	2121	0	0	2012-08-02 11:13:21.690781	2012-08-02 11:23:01.420239
12880	1-12880	83	7	1680	18	5125	0	5	2012-08-02 11:23:01.420239	2012-08-02 11:23:01.420239
12778	1-12778	6	48	3426	20	4266	0	7	2012-08-02 11:22:21.941505	2012-08-02 11:23:12.249847
12781	1-12781	20	30	3897	53	3799	0	2	2012-08-02 11:22:23.134547	2012-08-02 11:23:13.410418
11888	1-11888	34	74	5934	73	4331	0	2	2012-08-02 11:16:35.841397	2012-08-02 11:23:14.61459
12925	1-12925	17	41	587	79	6149	0	0	2012-08-02 11:23:19.343074	2012-08-02 11:23:19.343074
12928	1-12928	35	4	2197	80	4663	0	0	2012-08-02 11:23:20.514438	2012-08-02 11:23:20.514438
12831	1-12831	18	32	2024	71	9531	0	6	2012-08-02 11:22:42.946689	2012-08-02 11:23:24.171569
12937	1-12937	61	52	652	4	8005	0	4	2012-08-02 11:23:24.171569	2012-08-02 11:23:24.171569
12604	1-12604	35	39	789	3	8236	0	8	2012-08-02 11:21:13.887775	2012-08-02 11:21:13.887775
12613	1-12613	30	80	1220	17	7781	0	6	2012-08-02 11:21:17.40059	2012-08-02 11:21:17.40059
12554	1-12554	66	85	5992	44	9410	0	6	2012-08-02 11:20:54.260199	2012-08-02 11:21:18.572577
11435	1-11435	21	58	6479	38	6680	0	0	2012-08-02 11:13:39.975895	2012-08-02 11:21:18.572577
12617	1-12617	32	6	1817	11	2584	0	4	2012-08-02 11:21:18.572577	2012-08-02 11:21:18.572577
11686	1-11686	53	20	2307	8	1505	0	0	2012-08-02 11:15:17.02225	2012-08-02 11:21:20.141439
12414	1-12414	40	12	9277	66	6322	0	4	2012-08-02 11:19:59.706834	2012-08-02 11:21:23.655203
12629	1-12629	85	50	2318	12	8858	0	2	2012-08-02 11:21:23.655203	2012-08-02 11:21:23.655203
12638	1-12638	83	15	1082	35	9374	0	6	2012-08-02 11:21:27.146426	2012-08-02 11:21:27.146426
12644	1-12644	69	29	1	39	9127	0	0	2012-08-02 11:21:29.466507	2012-08-02 11:21:29.466507
11336	1-11336	83	65	5171	35	3524	0	6	2012-08-02 11:13:01.602358	2012-08-02 11:21:30.659848
11805	1-11805	90	74	2637	35	5567	0	8	2012-08-02 11:16:03.709206	2012-08-02 11:21:30.659848
12559	1-12559	88	39	1071	64	1493	0	8	2012-08-02 11:20:56.250456	2012-08-02 11:21:30.659848
12193	1-12193	70	87	9022	7	7485	0	0	2012-08-02 11:18:34.314206	2012-08-02 11:21:30.659848
12648	1-12648	56	35	4888	39	9687	0	0	2012-08-02 11:21:30.659848	2012-08-02 11:21:30.659848
11930	1-11930	97	87	9708	26	4436	0	4	2012-08-02 11:16:52.165039	2012-08-02 11:21:32.229149
12211	1-12211	62	26	4824	31	9481	0	3	2012-08-02 11:18:41.232527	2012-08-02 11:21:37.75325
12665	1-12665	31	33	2299	1	3075	0	2	2012-08-02 11:21:37.75325	2012-08-02 11:21:37.75325
12463	1-12463	93	66	3404	50	7612	0	3	2012-08-02 11:20:18.422869	2012-08-02 11:21:40.075695
5555	1-5555	98	1	9385	53	9624	0	0	2012-08-02 05:19:43.12443	2012-08-02 11:21:42.45101
12677	1-12677	12	53	1805	33	9125	0	0	2012-08-02 11:21:42.45101	2012-08-02 11:21:42.45101
12688	1-12688	7	6	5536	74	5669	0	5	2012-08-02 11:21:46.816578	2012-08-02 11:21:46.816578
12265	1-12265	53	81	7077	35	7981	0	0	2012-08-02 11:19:01.720992	2012-08-02 11:21:49.169599
12694	1-12694	18	38	3403	100	4994	0	6	2012-08-02 11:21:49.169599	2012-08-02 11:21:49.169599
11588	1-11588	86	22	6754	31	3549	0	0	2012-08-02 11:14:38.847184	2012-08-02 11:21:50.318842
11978	1-11978	41	37	7314	45	6971	0	0	2012-08-02 11:17:11.135517	2012-08-02 11:21:51.468072
7052	1-7052	12	35	8671	38	5142	0	6	2012-08-02 05:29:52.452157	2012-08-02 11:21:53.798983
5333	1-5333	47	35	8779	49	4534	0	0	2012-08-02 05:18:15.238449	2012-08-02 11:21:53.798983
12706	1-12706	46	38	3094	35	7767	0	5	2012-08-02 11:21:53.798983	2012-08-02 11:21:53.798983
12735	1-12735	58	45	1890	86	2417	0	8	2012-08-02 11:22:05.125694	2012-08-02 11:22:05.125694
8961	1-8961	84	45	9106	27	4023	0	0	2012-08-02 05:42:51.641639	2012-08-02 11:22:11.633168
12605	1-12605	27	96	4573	67	9206	0	0	2012-08-02 11:21:14.274664	2012-08-02 11:22:11.633168
3762	1-3762	50	8	5689	56	1866	0	0	2012-08-02 05:07:46.730009	2012-08-02 11:22:12.815234
12701	1-12701	72	48	3329	44	6983	0	12	2012-08-02 11:21:51.843443	2012-08-02 11:22:12.815234
12734	1-12734	6	2	5466	71	6502	0	16	2012-08-02 11:22:04.739061	2012-08-02 11:22:15.135514
12521	1-12521	51	99	1522	72	1987	0	0	2012-08-02 11:20:41.348057	2012-08-02 11:22:15.135514
9225	1-9225	4	55	3052	80	3149	0	3	2012-08-02 05:44:39.668204	2012-08-02 11:22:15.135514
12761	1-12761	86	72	4737	30	6802	0	5	2012-08-02 11:22:15.135514	2012-08-02 11:22:15.135514
9336	1-9336	39	13	6761	31	4273	0	5	2012-08-02 05:45:24.846666	2012-08-02 11:22:18.75946
8584	1-8584	69	39	4735	31	2172	0	0	2012-08-02 05:40:18.89265	2012-08-02 11:22:18.75946
12770	1-12770	87	31	2167	29	9888	0	0	2012-08-02 11:22:18.75946	2012-08-02 11:22:18.75946
10497	1-10497	95	51	8845	21	8552	0	7	2012-08-02 10:31:55.732897	2012-08-02 11:22:19.974844
12773	1-12773	65	5	686	51	3420	0	8	2012-08-02 11:22:19.974844	2012-08-02 11:22:19.974844
2846	1-2846	9	94	2066	5	141	141	0	2012-08-02 05:01:35.716335	2012-08-02 11:22:19.974844
12779	1-12779	65	50	2014	77	4865	0	7	2012-08-02 11:22:22.351083	2012-08-02 11:22:22.351083
9439	1-9439	52	18	7198	70	8851	0	11	2012-08-02 05:46:06.62494	2012-08-02 11:22:24.670211
12785	1-12785	41	70	4624	18	7973	0	4	2012-08-02 11:22:24.670211	2012-08-02 11:22:24.670211
3727	1-3727	45	72	8872	5	4059	0	0	2012-08-02 05:07:32.884957	2012-08-02 11:22:26.990612
12537	1-12537	74	70	6923	32	4145	0	5	2012-08-02 11:20:47.640146	2012-08-02 11:22:27.774733
12793	1-12793	84	32	2518	66	3331	0	9	2012-08-02 11:22:27.774733	2012-08-02 11:22:27.774733
12515	1-12515	31	51	4185	54	3206	0	7	2012-08-02 11:20:39.026873	2012-08-02 11:22:30.105975
12799	1-12799	21	54	815	51	2637	0	7	2012-08-02 11:22:30.105975	2012-08-02 11:22:30.105975
9734	1-9734	40	50	3512	42	2830	0	6	2012-08-02 05:48:07.354758	2012-08-02 11:22:38.626019
9645	1-9645	73	77	2764	57	839	0	0	2012-08-02 05:47:30.647077	2012-08-02 11:22:39.775033
12824	1-12824	24	50	1491	77	7604	0	8	2012-08-02 11:22:39.775033	2012-08-02 11:22:39.775033
12169	1-12169	92	8	7837	24	8141	0	20	2012-08-02 11:18:25.232612	2012-08-02 11:22:45.676422
12832	1-12832	38	16	8284	68	8565	0	6	2012-08-02 11:22:42.548342	2012-08-02 11:22:49.245482
12726	1-12726	74	31	1209	8	2966	0	0	2012-08-02 11:22:01.610705	2012-08-02 11:23:02.582878
12700	1-12700	64	45	5489	61	4910	0	21	2012-08-02 11:21:51.468072	2012-08-02 11:23:02.582878
12882	1-12882	94	38	1191	16	4517	0	0	2012-08-02 11:23:02.582878	2012-08-02 11:23:02.582878
12844	1-12844	82	44	3963	47	4800	0	8	2012-08-02 11:22:47.665344	2012-08-02 11:23:03.742743
12825	1-12825	50	16	7990	31	7644	0	17	2012-08-02 11:22:40.161601	2012-08-02 11:23:09.156232
12899	1-12899	75	12	2212	88	5964	0	0	2012-08-02 11:23:09.156232	2012-08-02 11:23:09.156232
6678	1-6678	36	6	7499	75	2067	0	0	2012-08-02 05:27:21.815264	2012-08-02 11:23:12.636473
12069	1-12069	7	46	7981	75	3761	0	10	2012-08-02 11:17:46.881871	2012-08-02 11:23:12.636473
12128	1-12128	75	54	7038	84	4980	0	7	2012-08-02 11:18:09.555327	2012-08-02 11:23:13.80786
7403	1-7403	9	41	9492	16	8902	0	0	2012-08-02 05:32:14.194754	2012-08-02 11:23:16.569875
12892	1-12892	40	90	1459	22	6279	0	11	2012-08-02 11:23:06.836243	2012-08-02 11:23:17.763025
12921	1-12921	78	22	434	40	5647	0	5	2012-08-02 11:23:17.763025	2012-08-02 11:23:17.763025
12686	1-12686	42	33	2886	35	5469	0	5	2012-08-02 11:21:46.042929	2012-08-02 11:23:19.74081
9373	1-9373	64	49	9480	74	3566	0	8	2012-08-02 05:45:40.005665	2012-08-02 11:23:19.74081
10997	1-10997	60	49	8931	78	3894	0	3	2012-08-02 10:39:21.565378	2012-08-02 11:23:19.74081
12850	1-12850	70	33	4896	49	9474	0	9	2012-08-02 11:22:50.030198	2012-08-02 11:23:19.74081
11180	1-11180	25	75	5402	42	7064	0	4	2012-08-02 11:12:00.693658	2012-08-02 11:23:20.912193
7794	1-7794	71	32	9741	61	4251	0	8	2012-08-02 05:34:56.375583	2012-08-02 11:23:20.912193
12929	1-12929	76	42	2517	75	7270	0	3	2012-08-02 11:23:20.912193	2012-08-02 11:23:20.912193
9441	1-9441	84	93	8282	96	6072	0	0	2012-08-02 05:46:07.564486	2012-08-02 11:23:24.580265
9732	1-9732	27	28	4952	2	7154	0	5	2012-08-02 05:48:06.581392	2012-08-02 11:23:24.580265
12769	1-12769	87	34	4659	6	9328	0	0	2012-08-02 11:22:18.372422	2012-08-02 11:23:25.731152
6780	1-6780	56	75	8383	77	7082	0	13	2012-08-02 05:28:02.313684	2012-08-02 11:23:25.731152
11845	1-11845	21	30	8775	99	9735	0	7	2012-08-02 11:16:19.255494	2012-08-02 11:23:25.731152
8936	1-8936	77	6	7969	18	3669	0	7	2012-08-02 05:42:41.387271	2012-08-02 11:22:47.278678
12820	1-12820	69	54	836	47	1002	0	0	2012-08-02 11:22:38.250409	2012-08-02 11:22:48.450151
5902	1-5902	10	71	8808	79	3452	0	16	2012-08-02 05:22:03.741433	2012-08-02 11:22:53.156836
10414	1-10414	22	71	3777	78	860	0	6	2012-08-02 10:27:50.075143	2012-08-02 11:22:53.156836
12867	1-12867	87	29	1326	60	3328	0	10	2012-08-02 11:22:56.724639	2012-08-02 11:22:56.724639
12272	1-12272	26	60	4668	45	3412	0	0	2012-08-02 11:19:04.747868	2012-08-02 11:22:57.949342
12869	1-12869	31	61	1388	60	9199	0	3	2012-08-02 11:22:57.949342	2012-08-02 11:22:57.949342
11915	1-11915	2	8	6809	46	7655	0	4	2012-08-02 11:16:46.284519	2012-08-02 11:22:58.72632
11531	1-11531	79	18	6277	81	8261	0	0	2012-08-02 11:14:17.012887	2012-08-02 11:22:58.72632
12872	1-12872	26	94	886	18	6574	0	11	2012-08-02 11:22:58.72632	2012-08-02 11:22:58.72632
12826	1-12826	68	85	2136	82	2445	0	5	2012-08-02 11:22:40.604474	2012-08-02 11:22:59.871787
12875	1-12875	51	80	1754	2	7861	0	15	2012-08-02 11:22:59.871787	2012-08-02 11:22:59.871787
11627	1-11627	17	74	5828	67	8855	0	0	2012-08-02 11:14:53.87613	2012-08-02 11:23:01.036268
12878	1-12878	83	7	3502	4	5501	0	0	2012-08-02 11:23:01.036268	2012-08-02 11:23:01.036268
9848	1-9848	94	37	4398	50	2330	0	5	2012-08-02 05:48:53.578593	2012-08-02 11:23:03.356209
11778	1-11778	92	2	8867	85	5556	0	0	2012-08-02 11:15:53.121161	2012-08-02 11:23:05.676201
10775	1-10775	68	10	6213	32	5386	0	0	2012-08-02 10:37:53.877914	2012-08-02 11:23:06.836243
11550	1-11550	32	81	6548	56	1738	0	0	2012-08-02 11:14:24.296489	2012-08-02 11:23:07.609567
12897	1-12897	69	25	2371	23	2630	0	2	2012-08-02 11:23:08.394079	2012-08-02 11:23:08.394079
12702	1-12702	59	56	2207	44	6303	0	12	2012-08-02 11:21:52.241142	2012-08-02 11:23:09.565198
6071	1-6071	98	7	9430	100	3756	0	0	2012-08-02 05:23:13.729081	2012-08-02 11:23:09.565198
12903	1-12903	6	21	301	62	4990	0	0	2012-08-02 11:23:10.725197	2012-08-02 11:23:10.725197
12906	1-12906	45	54	1467	68	8972	0	0	2012-08-02 11:23:11.863359	2012-08-02 11:23:11.863359
12909	1-12909	88	20	1850	19	7475	0	0	2012-08-02 11:23:13.02337	2012-08-02 11:23:13.02337
12651	1-12651	96	95	7308	39	7864	0	3	2012-08-02 11:21:32.229149	2012-08-02 11:23:17.376362
12386	1-12386	55	29	7666	39	5427	0	0	2012-08-02 11:19:48.910609	2012-08-02 11:23:17.376362
12536	1-12536	4	19	5049	26	6334	0	29	2012-08-02 11:20:47.231491	2012-08-02 11:23:17.376362
9770	1-9770	50	19	6402	59	7471	0	0	2012-08-02 05:48:21.830943	2012-08-02 11:23:17.376362
12030	1-12030	16	44	8822	98	7691	0	5	2012-08-02 11:17:31.565594	2012-08-02 11:23:18.547636
7981	1-7981	99	81	7471	41	2231	0	10	2012-08-02 05:36:15.278642	2012-08-02 11:23:18.547636
10774	1-10774	92	28	2475	9	1863	0	3	2012-08-02 10:37:53.480207	2012-08-02 11:23:22.625081
12790	1-12790	74	5	2507	72	8384	0	6	2012-08-02 11:22:26.990612	2012-08-02 11:23:27.333831
12914	1-12914	43	16	5408	26	4205	0	5	2012-08-02 11:23:15.399098	2012-08-02 11:23:27.333831
12948	1-12948	81	35	98	36	2394	0	0	2012-08-02 11:23:28.515498	2012-08-02 11:23:28.515498
12298	1-12298	1	84	579	11	400	0	7	2012-08-02 11:19:14.757324	2012-08-02 11:23:29.686653
12890	1-12890	66	85	3691	2	7154	0	4	2012-08-02 11:23:05.676201	2012-08-02 11:23:29.686653
12332	1-12332	49	13	8231	11	7217	0	0	2012-08-02 11:19:28.09272	2012-08-02 11:23:30.835937
12953	1-12953	79	92	707	13	6778	0	4	2012-08-02 11:23:30.835937	2012-08-02 11:23:30.835937
8177	1-8177	36	11	9498	58	9560	686	7	2012-08-02 05:37:34.699572	2012-08-02 11:23:30.835937
2307	1-2307	34	11	4150	58	484	484	7	2012-08-02 04:57:44.02233	2012-08-02 11:23:30.835937
11887	1-11887	15	8	8997	58	7003	0	3	2012-08-02 11:16:35.06808	2012-08-02 11:23:32.79149
12755	1-12755	32	44	9724	8	9618	0	5	2012-08-02 11:22:12.815234	2012-08-02 11:23:32.79149
12938	1-12938	3	96	1571	70	5173	0	2	2012-08-02 11:23:24.580265	2012-08-02 11:23:39.862557
11969	1-11969	92	66	4977	70	9508	0	2	2012-08-02 11:17:07.644443	2012-08-02 11:23:39.862557
12977	1-12977	10	70	3816	99	5883	0	5	2012-08-02 11:23:39.862557	2012-08-02 11:23:39.862557
12952	1-12952	65	93	5751	70	4290	0	0	2012-08-02 11:23:30.073518	2012-08-02 11:23:41.000653
11527	1-11527	66	40	9283	39	3268	0	10	2012-08-02 11:14:15.488358	2012-08-02 11:23:42.194146
12983	1-12983	57	78	2375	26	4975	0	5	2012-08-02 11:23:42.194146	2012-08-02 11:23:42.194146
8180	1-8180	91	66	4059	88	2729	0	7	2012-08-02 05:37:35.970271	2012-08-02 11:23:43.365977
13012	1-13012	51	83	36	71	7584	0	0	2012-08-02 11:23:53.543515	2012-08-02 11:23:53.543515
8136	1-8136	17	81	9039	79	3763	0	3	2012-08-02 05:37:17.959245	2012-08-02 11:23:54.703668
13016	1-13016	80	70	1702	81	9932	0	9	2012-08-02 11:23:54.703668	2012-08-02 11:23:54.703668
11119	1-11119	41	4	3935	19	1250	0	0	2012-08-02 11:11:36.707517	2012-08-02 11:23:59.785955
13028	1-13028	44	19	1172	42	6517	0	2	2012-08-02 11:23:59.785955	2012-08-02 11:23:59.785955
12025	1-12025	1	12	9481	10	5439	0	0	2012-08-02 11:17:29.610228	2012-08-02 11:24:03.210707
13037	1-13037	90	10	4856	12	8624	0	0	2012-08-02 11:24:03.210707	2012-08-02 11:24:03.210707
5668	1-5668	18	65	9856	10	7917	0	0	2012-08-02 05:20:28.606115	2012-08-02 11:24:09.430813
13053	1-13053	68	58	264	65	8263	0	14	2012-08-02 11:24:09.430813	2012-08-02 11:24:09.430813
10323	1-10323	14	84	6233	16	4409	0	0	2012-08-02 10:25:46.530979	2012-08-02 11:24:12.922075
12978	1-12978	11	1	1787	16	6496	0	4	2012-08-02 11:23:40.238195	2012-08-02 11:24:12.922075
12945	1-12945	28	72	5391	16	7462	0	0	2012-08-02 11:23:27.333831	2012-08-02 11:24:12.922075
9377	1-9377	96	26	4409	16	5657	0	3	2012-08-02 05:45:41.608093	2012-08-02 11:24:12.922075
13062	1-13062	1	16	1578	50	7154	0	0	2012-08-02 11:24:12.922075	2012-08-02 11:24:12.922075
13065	1-13065	90	17	89	74	8668	0	0	2012-08-02 11:24:14.10453	2012-08-02 11:24:14.10453
13030	1-13030	95	83	4110	28	1980	0	0	2012-08-02 11:24:00.537144	2012-08-02 11:24:19.144836
12101	1-12101	54	95	9014	98	9900	0	8	2012-08-02 11:17:59.090529	2012-08-02 11:24:19.939707
12918	1-12918	32	45	6307	46	6942	0	0	2012-08-02 11:23:16.569875	2012-08-02 11:24:19.939707
10958	1-10958	20	37	9970	43	4845	0	3	2012-08-02 10:39:05.921426	2012-08-02 11:24:26.844739
13098	1-13098	85	43	634	38	4689	0	11	2012-08-02 11:24:26.844739	2012-08-02 11:24:26.844739
13101	1-13101	68	7	1983	83	7483	0	2	2012-08-02 11:24:28.004933	2012-08-02 11:24:28.004933
13073	1-13073	66	96	5690	11	9770	0	0	2012-08-02 11:24:17.175544	2012-08-02 11:24:30.358203
10428	1-10428	82	96	7677	64	4968	0	11	2012-08-02 10:27:55.611142	2012-08-02 11:24:30.358203
13107	1-13107	6	1	5901	96	9593	0	6	2012-08-02 11:24:30.358203	2012-08-02 11:24:30.358203
13110	1-13110	56	66	510	18	1044	0	8	2012-08-02 11:24:31.529657	2012-08-02 11:24:31.529657
12911	1-12911	32	85	2707	54	3117	0	10	2012-08-02 11:23:13.80786	2012-08-02 11:24:33.89365
12969	1-12969	91	88	5184	60	5868	0	5	2012-08-02 11:23:36.725301	2012-08-02 11:24:35.065115
13128	1-13128	76	75	651	16	4016	0	0	2012-08-02 11:24:38.986892	2012-08-02 11:24:38.986892
10792	1-10792	97	41	3758	30	2633	0	0	2012-08-02 10:38:00.573459	2012-08-02 11:24:41.318316
13134	1-13134	40	22	421	41	1424	0	0	2012-08-02 11:24:41.318316	2012-08-02 11:24:41.318316
1896	1-1896	83	86	7689	40	8184	0	0	2012-08-02 04:54:45.923304	2012-08-02 11:24:44.12585
12132	1-12132	91	86	3506	3	885	0	6	2012-08-02 11:18:11.06854	2012-08-02 11:24:44.12585
12941	1-12941	64	77	3126	75	8673	0	11	2012-08-02 11:23:25.731152	2012-08-02 11:23:25.731152
12282	1-12282	84	56	3467	81	8496	0	5	2012-08-02 11:19:08.60354	2012-08-02 11:23:29.300201
10529	1-10529	84	56	7819	97	8601	0	6	2012-08-02 10:32:08.367296	2012-08-02 11:23:29.300201
12908	1-12908	56	75	1808	34	7041	0	0	2012-08-02 11:23:12.636473	2012-08-02 11:23:30.449552
12185	1-12185	30	34	9141	38	9967	0	0	2012-08-02 11:18:31.275956	2012-08-02 11:23:30.449552
12234	1-12234	67	7	5248	35	2307	0	6	2012-08-02 11:18:49.730361	2012-08-02 11:23:32.018196
12241	1-12241	80	7	6966	16	8962	0	7	2012-08-02 11:18:52.802473	2012-08-02 11:23:32.018196
12518	1-12518	38	7	1815	2	3487	0	5	2012-08-02 11:20:40.17661	2012-08-02 11:23:32.018196
12775	1-12775	78	7	4115	78	2500	0	10	2012-08-02 11:22:20.770288	2012-08-02 11:23:32.018196
12894	1-12894	10	44	2572	81	6120	0	0	2012-08-02 11:23:07.609567	2012-08-02 11:23:32.018196
10749	1-10749	49	81	7629	7	7253	1921	0	2012-08-02 10:37:44.264527	2012-08-02 11:23:32.018196
12936	1-12936	75	5	4826	4	3362	0	10	2012-08-02 11:23:23.773616	2012-08-02 11:23:37.874152
9198	1-9198	75	5	7265	40	6084	0	6	2012-08-02 05:44:28.682456	2012-08-02 11:23:37.874152
12873	1-12873	99	31	3621	12	9353	0	11	2012-08-02 11:22:59.10205	2012-08-02 11:23:40.238195
10674	1-10674	31	16	6385	80	3567	0	0	2012-08-02 10:36:10.940094	2012-08-02 11:23:40.238195
11750	1-11750	24	16	6106	33	2115	0	0	2012-08-02 11:15:42.063484	2012-08-02 11:23:40.238195
10131	1-10131	20	16	7519	31	741	741	11	2012-08-02 10:21:32.297328	2012-08-02 11:23:40.238195
11578	1-11578	6	75	2678	90	3287	0	12	2012-08-02 11:14:35.013612	2012-08-02 11:23:44.161638
12858	1-12858	92	75	3782	32	8616	0	8	2012-08-02 11:22:53.156836	2012-08-02 11:23:44.161638
12988	1-12988	18	91	1058	75	9939	0	8	2012-08-02 11:23:44.161638	2012-08-02 11:23:44.161638
3671	1-3671	33	71	8788	37	5915	0	0	2012-08-02 05:07:10.622819	2012-08-02 11:23:45.333175
588	1-588	47	52	7484	27	1882	0	0	2012-08-02 04:44:33.720249	2012-08-02 11:23:45.333175
12991	1-12991	90	50	824	71	8740	0	6	2012-08-02 11:23:45.333175	2012-08-02 11:23:45.333175
11827	1-11827	48	41	2358	83	4096	0	5	2012-08-02 11:16:12.394521	2012-08-02 11:23:46.504397
9276	1-9276	97	83	3532	70	1909	0	0	2012-08-02 05:45:00.491696	2012-08-02 11:23:46.504397
6532	1-6532	53	47	3363	9	1028	0	0	2012-08-02 05:26:22.365733	2012-08-02 11:23:46.504397
9343	1-9343	44	41	2310	28	590	0	0	2012-08-02 05:45:27.708591	2012-08-02 11:23:46.504397
12995	1-12995	18	28	2400	41	9595	0	4	2012-08-02 11:23:46.504397	2012-08-02 11:23:46.504397
13013	1-13013	47	81	5376	94	4554	0	3	2012-08-02 11:23:53.941308	2012-08-02 11:23:55.476965
12200	1-12200	83	68	4833	51	8319	0	6	2012-08-02 11:18:36.987947	2012-08-02 11:23:55.476965
10894	1-10894	37	100	6011	59	4325	0	3	2012-08-02 10:38:40.575325	2012-08-02 11:23:55.476965
13017	1-13017	56	17	3467	100	8623	0	3	2012-08-02 11:23:55.476965	2012-08-02 11:23:55.476965
12653	1-12653	58	98	2976	60	7846	0	0	2012-08-02 11:21:33.411452	2012-08-02 11:23:56.637412
12729	1-12729	23	70	6153	28	4066	0	0	2012-08-02 11:22:02.794802	2012-08-02 11:24:04.779544
12980	1-12980	41	70	4617	93	6811	0	9	2012-08-02 11:23:41.000653	2012-08-02 11:24:04.779544
13043	1-13043	80	98	3783	12	4987	0	0	2012-08-02 11:24:05.530796	2012-08-02 11:24:07.111196
13046	1-13046	7	94	1533	98	7391	0	0	2012-08-02 11:24:07.111196	2012-08-02 11:24:07.111196
11365	1-11365	16	80	8982	46	4388	927	5	2012-08-02 11:13:12.827628	2012-08-02 11:24:07.111196
9364	1-9364	49	76	5992	88	3275	0	0	2012-08-02 05:45:36.381909	2012-08-02 11:24:10.204408
8831	1-8831	19	69	8734	70	7348	0	0	2012-08-02 05:41:58.738246	2012-08-02 11:24:10.204408
12440	1-12440	12	88	3205	93	9553	0	0	2012-08-02 11:20:09.761988	2012-08-02 11:24:10.204408
13055	1-13055	56	93	2669	69	8188	0	0	2012-08-02 11:24:10.204408	2012-08-02 11:24:10.204408
12550	1-12550	85	30	2303	2	7990	0	0	2012-08-02 11:20:52.678224	2012-08-02 11:24:17.175544
11179	1-11179	64	29	9752	27	6394	0	7	2012-08-02 11:12:00.284548	2012-08-02 11:24:17.175544
12830	1-12830	39	11	4335	30	5410	0	5	2012-08-02 11:22:42.15062	2012-08-02 11:24:17.175544
7593	1-7593	16	11	2745	80	1772	0	0	2012-08-02 05:33:32.289397	2012-08-02 11:24:17.175544
13094	1-13094	99	59	387	16	7584	0	0	2012-08-02 11:24:25.320192	2012-08-02 11:24:25.320192
13103	1-13103	44	11	1599	78	6746	0	0	2012-08-02 11:24:28.767381	2012-08-02 11:24:28.767381
13080	1-13080	72	44	3292	45	6617	0	2	2012-08-02 11:24:19.939707	2012-08-02 11:24:29.949566
12433	1-12433	98	26	2623	40	5595	0	0	2012-08-02 11:20:06.273364	2012-08-02 11:24:29.949566
13106	1-13106	64	45	4694	19	3796	0	8	2012-08-02 11:24:29.949566	2012-08-02 11:24:29.949566
12796	1-12796	81	60	8997	92	7998	0	0	2012-08-02 11:22:28.93486	2012-08-02 11:24:37.075581
13124	1-13124	8	37	4840	60	5123	0	13	2012-08-02 11:24:37.075581	2012-08-02 11:24:37.075581
11998	1-11998	22	70	5919	84	4423	0	4	2012-08-02 11:17:18.219448	2012-08-02 11:24:38.611274
12625	1-12625	62	14	8157	84	5168	0	10	2012-08-02 11:21:22.120828	2012-08-02 11:24:38.611274
12008	1-12008	61	26	7615	5	8725	0	18	2012-08-02 11:17:22.947904	2012-08-02 11:24:38.611274
12204	1-12204	47	55	9222	1	6465	0	10	2012-08-02 11:18:38.547847	2012-08-02 11:24:40.931417
13083	1-13083	49	9	3526	94	2832	0	0	2012-08-02 11:24:21.099854	2012-08-02 11:24:48.823952
13153	1-13153	53	94	2929	73	5710	0	3	2012-08-02 11:24:48.823952	2012-08-02 11:24:48.823952
12924	1-12924	44	98	5428	44	9118	0	0	2012-08-02 11:23:18.547636	2012-08-02 11:24:49.983851
13156	1-13156	67	7	1560	98	3012	0	3	2012-08-02 11:24:49.983851	2012-08-02 11:24:49.983851
13170	1-13170	9	51	793	85	8548	0	2	2012-08-02 11:24:55.920634	2012-08-02 11:24:55.920634
13171	1-13171	82	72	7940	9	9049	0	9	2012-08-02 11:24:55.53012	2012-08-02 11:25:00.269988
12958	1-12958	51	66	5580	58	9564	0	5	2012-08-02 11:23:32.404853	2012-08-02 11:25:00.269988
12961	1-12961	62	3	5288	66	9428	0	9	2012-08-02 11:23:33.576217	2012-08-02 11:25:00.269988
12496	1-12496	54	87	7737	9	3710	0	3	2012-08-02 11:20:31.675507	2012-08-02 11:25:00.269988
13186	1-13186	75	13	396	2	2855	0	5	2012-08-02 11:25:01.838976	2012-08-02 11:25:01.838976
9740	1-9740	21	95	7169	48	5590	0	0	2012-08-02 05:48:09.774523	2012-08-02 11:25:03.021187
10393	1-10393	1	95	6183	69	4913	0	8	2012-08-02 10:27:41.788038	2012-08-02 11:25:03.021187
12146	1-12146	71	32	3732	48	3413	0	2	2012-08-02 11:18:16.404878	2012-08-02 11:25:03.021187
9726	1-9726	24	54	2289	1	680	0	0	2012-08-02 05:48:04.082647	2012-08-02 11:25:07.330226
13154	1-13154	96	18	907	100	671	0	6	2012-08-02 11:24:49.199126	2012-08-02 11:25:07.330226
12896	1-12896	30	4	8596	1	8958	0	4	2012-08-02 11:23:08.007205	2012-08-02 11:25:08.490508
12471	1-12471	71	82	5262	4	6808	0	5	2012-08-02 11:20:21.929827	2012-08-02 11:25:08.490508
13203	1-13203	45	13	1730	82	7356	0	9	2012-08-02 11:25:08.490508	2012-08-02 11:25:08.490508
12787	1-12787	12	53	6006	46	6740	0	2	2012-08-02 11:22:25.443854	2012-08-02 11:25:11.652325
13211	1-13211	29	46	3449	33	4784	0	15	2012-08-02 11:25:11.652325	2012-08-02 11:25:11.652325
13225	1-13225	68	21	200	54	4217	0	2	2012-08-02 11:25:16.724159	2012-08-02 11:25:16.724159
13237	1-13237	15	43	317	1	2238	0	6	2012-08-02 11:25:21.813307	2012-08-02 11:25:21.813307
9716	1-9716	94	31	6291	80	9198	0	0	2012-08-02 05:47:59.917475	2012-08-02 11:25:23.031427
12226	1-12226	20	1	3639	38	5946	0	17	2012-08-02 11:18:47.034512	2012-08-02 11:25:24.204162
12926	1-12926	55	88	7113	33	9278	0	0	2012-08-02 11:23:19.74081	2012-08-02 11:23:28.902649
11766	1-11766	97	63	7173	90	10000	0	0	2012-08-02 11:15:48.414054	2012-08-02 11:23:28.902649
12949	1-12949	77	91	6273	63	9120	0	18	2012-08-02 11:23:28.902649	2012-08-02 11:23:28.902649
11194	1-11194	40	58	6432	3	2694	0	4	2012-08-02 11:12:06.186467	2012-08-02 11:23:33.576217
11200	1-11200	57	13	8590	27	7497	0	3	2012-08-02 11:12:08.551568	2012-08-02 11:23:33.576217
8583	1-8583	44	91	8067	65	4399	0	0	2012-08-02 05:40:18.494972	2012-08-02 11:23:34.747401
12964	1-12964	69	10	1992	91	8038	0	8	2012-08-02 11:23:34.747401	2012-08-02 11:23:34.747401
294	1-294	90	99	2267	47	52	52	0	2012-08-02 04:43:58.789092	2012-08-02 11:23:34.747401
8676	1-8676	52	78	6488	14	9459	0	4	2012-08-02 05:40:55.929007	2012-08-02 11:23:37.112039
12269	1-12269	63	86	3952	92	7687	0	6	2012-08-02 11:19:03.632029	2012-08-02 11:23:43.77499
12954	1-12954	37	38	6758	78	7150	0	2	2012-08-02 11:23:30.449552	2012-08-02 11:23:44.946237
12950	1-12950	34	46	3121	56	6662	0	0	2012-08-02 11:23:29.300201	2012-08-02 11:23:46.117465
7365	1-7365	40	26	8999	69	9268	0	0	2012-08-02 05:31:58.81422	2012-08-02 11:23:53.941308
12624	1-12624	17	94	170	26	248	0	0	2012-08-02 11:21:21.710808	2012-08-02 11:23:53.941308
12957	1-12957	31	81	2277	7	8888	0	8	2012-08-02 11:23:32.018196	2012-08-02 11:23:58.228472
10027	1-10027	62	36	5884	57	5551	0	7	2012-08-02 10:19:32.987884	2012-08-02 11:23:58.228472
13024	1-13024	71	20	1081	81	6636	0	0	2012-08-02 11:23:58.228472	2012-08-02 11:23:58.228472
12423	1-12423	38	21	3421	5	2692	0	7	2012-08-02 11:20:03.165061	2012-08-02 11:24:00.537144
13039	1-13039	70	13	1258	85	3339	0	0	2012-08-02 11:24:03.98406	2012-08-02 11:24:03.98406
12972	1-12972	77	95	3748	5	8923	0	0	2012-08-02 11:23:37.874152	2012-08-02 11:24:16.777883
11792	1-11792	35	58	5780	25	8297	0	11	2012-08-02 11:15:58.624662	2012-08-02 11:24:17.573325
12000	1-12000	80	80	3518	25	3786	0	0	2012-08-02 11:17:19.799106	2012-08-02 11:24:17.573325
13074	1-13074	71	44	301	58	6832	0	2	2012-08-02 11:24:17.573325	2012-08-02 11:24:17.573325
11842	1-11842	100	82	4704	80	6909	0	11	2012-08-02 11:16:18.0948	2012-08-02 11:24:20.71302
12752	1-12752	61	27	5405	96	7336	0	10	2012-08-02 11:22:11.633168	2012-08-02 11:24:20.71302
12855	1-12855	100	35	5542	76	7034	0	6	2012-08-02 11:22:51.996373	2012-08-02 11:24:20.71302
13082	1-13082	25	96	2223	35	9655	0	17	2012-08-02 11:24:20.71302	2012-08-02 11:24:20.71302
12807	1-12807	14	70	2999	41	2981	0	6	2012-08-02 11:22:33.223422	2012-08-02 11:24:24.160096
13096	1-13096	35	54	49	58	41	0	11	2012-08-02 11:24:26.082758	2012-08-02 11:24:26.082758
13109	1-13109	73	85	2345	75	8167	0	10	2012-08-02 11:24:30.74489	2012-08-02 11:24:30.74489
11892	1-11892	76	93	7806	34	7316	0	6	2012-08-02 11:16:37.399241	2012-08-02 11:24:33.496367
11326	1-11326	34	46	8319	22	7841	0	0	2012-08-02 11:12:57.757696	2012-08-02 11:24:33.496367
13115	1-13115	4	22	2470	93	9463	0	4	2012-08-02 11:24:33.496367	2012-08-02 11:24:33.496367
12277	1-12277	82	3	6312	91	9534	0	4	2012-08-02 11:19:06.670204	2012-08-02 11:24:39.384613
9843	1-9843	33	4	7340	42	8159	0	0	2012-08-02 05:48:51.56609	2012-08-02 11:24:40.555909
12551	1-12551	60	42	3114	57	3766	0	3	2012-08-02 11:20:53.075923	2012-08-02 11:24:40.555909
13132	1-13132	70	20	4362	4	2523	0	9	2012-08-02 11:24:40.555909	2012-08-02 11:24:40.555909
10730	1-10730	92	69	4562	80	4348	0	2	2012-08-02 10:36:32.839409	2012-08-02 11:24:41.705139
13119	1-13119	8	92	5782	88	6780	0	5	2012-08-02 11:24:35.065115	2012-08-02 11:24:42.909484
13137	1-13137	16	55	1873	17	3609	0	11	2012-08-02 11:24:42.909484	2012-08-02 11:24:42.909484
12582	1-12582	69	40	4721	45	5205	0	11	2012-08-02 11:21:05.178781	2012-08-02 11:24:48.415244
12970	1-12970	6	82	2744	38	6352	0	11	2012-08-02 11:23:37.112039	2012-08-02 11:24:48.415244
5325	1-5325	59	4	6831	34	2886	0	0	2012-08-02 05:18:12.065103	2012-08-02 11:24:48.415244
12658	1-12658	34	21	5063	79	3789	0	0	2012-08-02 11:21:34.980127	2012-08-02 11:24:51.961695
10088	1-10088	74	96	5705	8	3641	0	0	2012-08-02 10:21:15.194431	2012-08-02 11:24:51.961695
12889	1-12889	62	42	5800	74	3801	0	0	2012-08-02 11:23:05.278493	2012-08-02 11:24:54.347969
13166	1-13166	11	76	1874	42	7695	0	6	2012-08-02 11:24:54.347969	2012-08-02 11:24:54.347969
11683	1-11683	16	83	8108	99	4521	0	0	2012-08-02 11:15:15.883667	2012-08-02 11:24:55.132544
13169	1-13169	8	99	590	71	3518	0	4	2012-08-02 11:24:55.132544	2012-08-02 11:24:55.132544
13172	1-13172	54	63	34	33	4103	0	6	2012-08-02 11:24:56.292589	2012-08-02 11:24:56.292589
8931	1-8931	29	20	2240	59	873	0	0	2012-08-02 05:42:39.376893	2012-08-02 11:25:01.441006
12304	1-12304	20	5	1385	20	562	0	6	2012-08-02 11:19:17.249311	2012-08-02 11:25:06.92148
13198	1-13198	75	20	232	5	7230	0	0	2012-08-02 11:25:06.92148	2012-08-02 11:25:06.92148
12213	1-12213	31	42	7502	4	5457	0	0	2012-08-02 11:18:41.619389	2012-08-02 11:25:07.717116
13136	1-13136	97	5	3445	96	3998	0	0	2012-08-02 11:24:42.102776	2012-08-02 11:25:19.091902
13230	1-13230	36	96	3332	35	5852	0	21	2012-08-02 11:25:19.091902	2012-08-02 11:25:19.091902
12990	1-12990	88	78	3485	96	7394	0	8	2012-08-02 11:23:44.946237	2012-08-02 11:25:20.254567
12673	1-12673	91	78	1328	5	2591	0	3	2012-08-02 11:21:40.859588	2012-08-02 11:25:20.254567
9868	1-9868	10	23	5669	76	4670	0	0	2012-08-02 05:49:01.73753	2012-08-02 11:25:22.619491
12294	1-12294	46	81	4511	25	3421	0	4	2012-08-02 11:19:13.243989	2012-08-02 11:25:26.164453
13248	1-13248	34	22	459	43	6233	0	12	2012-08-02 11:25:26.164453	2012-08-02 11:25:26.164453
6176	1-6176	89	25	9880	22	2903	819	15	2012-08-02 05:23:55.755308	2012-08-02 11:25:26.164453
13252	1-13252	34	47	2772	29	3543	0	0	2012-08-02 11:25:27.349989	2012-08-02 11:25:28.933653
12545	1-12545	21	47	5230	79	8154	0	7	2012-08-02 11:20:50.711483	2012-08-02 11:25:28.933653
13255	1-13255	6	29	3061	47	6719	0	3	2012-08-02 11:25:28.933653	2012-08-02 11:25:28.933653
12839	1-12839	64	24	5174	26	7100	0	7	2012-08-02 11:22:45.676422	2012-08-02 11:25:31.323161
6013	1-6013	51	45	4792	37	3419	0	0	2012-08-02 05:22:49.896236	2012-08-02 11:25:38.40866
13022	1-13022	61	45	2263	36	751	0	4	2012-08-02 11:23:57.433023	2012-08-02 11:25:38.40866
13279	1-13279	47	66	593	45	9635	0	0	2012-08-02 11:25:38.40866	2012-08-02 11:25:38.40866
11701	1-11701	89	28	6684	6	5677	0	0	2012-08-02 11:15:22.935042	2012-08-02 11:25:41.128704
13007	1-13007	1	12	2046	14	4989	0	2	2012-08-02 11:23:51.577128	2012-08-02 11:25:51.610151
13312	1-13312	45	14	6185	30	5114	0	0	2012-08-02 11:25:51.610151	2012-08-02 11:25:51.610151
9540	1-9540	56	47	2857	35	914	0	0	2012-08-02 05:46:47.817361	2012-08-02 11:25:52.803265
13090	1-13090	30	31	1622	70	7302	0	0	2012-08-02 11:24:24.160096	2012-08-02 11:25:53.587928
12530	1-12530	99	78	3425	77	9472	0	2	2012-08-02 11:20:44.889051	2012-08-02 11:25:53.587928
12837	1-12837	15	4	5898	77	9070	0	5	2012-08-02 11:22:44.903286	2012-08-02 11:25:53.587928
13317	1-13317	96	77	3452	31	9491	0	0	2012-08-02 11:25:53.587928	2012-08-02 11:25:53.587928
13189	1-13189	71	48	5033	95	8516	0	2	2012-08-02 11:25:03.021187	2012-08-02 11:25:55.986268
9287	1-9287	53	43	8534	100	5346	0	3	2012-08-02 05:45:05.032731	2012-08-02 11:26:00.681608
11498	1-11498	98	55	5285	13	4179	0	7	2012-08-02 11:14:04.273767	2012-08-02 11:26:00.681608
11371	1-11371	90	43	2327	89	1559	0	6	2012-08-02 11:13:15.501264	2012-08-02 11:26:00.681608
13140	1-13140	67	30	897	86	9053	0	9	2012-08-02 11:24:44.12585	2012-08-02 11:24:44.12585
12959	1-12959	14	25	7240	44	9110	0	0	2012-08-02 11:23:32.79149	2012-08-02 11:24:44.94585
13142	1-13142	63	44	695	56	9611	0	9	2012-08-02 11:24:44.94585	2012-08-02 11:24:44.94585
10941	1-10941	6	23	3139	86	3113	0	0	2012-08-02 10:38:59.142666	2012-08-02 11:24:45.708282
13148	1-13148	49	87	2679	53	6013	0	4	2012-08-02 11:24:46.868161	2012-08-02 11:24:46.868161
11425	1-11425	45	23	7573	69	6508	0	7	2012-08-02 11:13:36.097532	2012-08-02 11:24:50.381471
12943	1-12943	69	24	4260	90	6283	0	4	2012-08-02 11:23:26.515891	2012-08-02 11:24:50.381471
8354	1-8354	14	33	5327	69	4783	0	0	2012-08-02 05:38:46.367723	2012-08-02 11:24:50.381471
12580	1-12580	99	24	2277	66	2237	0	10	2012-08-02 11:21:04.394262	2012-08-02 11:24:50.381471
13157	1-13157	66	69	5493	24	7397	0	3	2012-08-02 11:24:50.381471	2012-08-02 11:24:50.381471
13160	1-13160	68	20	2105	64	7423	0	4	2012-08-02 11:24:51.575071	2012-08-02 11:24:51.575071
12877	1-12877	85	53	1627	72	3945	0	7	2012-08-02 11:23:00.648243	2012-08-02 11:24:55.53012
12933	1-12933	73	9	5464	28	7945	0	6	2012-08-02 11:23:22.625081	2012-08-02 11:24:55.53012
6664	1-6664	89	9	9183	53	7774	0	6	2012-08-02 05:27:16.180472	2012-08-02 11:24:55.53012
10503	1-10503	19	43	8512	50	7972	0	5	2012-08-02 10:31:58.119548	2012-08-02 11:24:58.314903
13032	1-13032	55	60	3838	43	2329	0	5	2012-08-02 11:24:01.310801	2012-08-02 11:24:58.314903
12989	1-12989	67	60	1533	90	1393	0	10	2012-08-02 11:23:44.559618	2012-08-02 11:24:58.314903
12759	1-12759	59	80	7701	72	9089	0	0	2012-08-02 11:22:14.361965	2012-08-02 11:25:04.93251
12791	1-12791	86	58	2475	85	5910	0	2	2012-08-02 11:22:26.592613	2012-08-02 11:25:04.93251
10218	1-10218	31	79	5978	75	2561	0	8	2012-08-02 10:22:30.455459	2012-08-02 11:25:04.93251
13064	1-13064	37	28	4438	62	7685	0	11	2012-08-02 11:24:13.695679	2012-08-02 11:25:04.93251
13194	1-13194	53	22	3142	58	9574	0	3	2012-08-02 11:25:04.93251	2012-08-02 11:25:04.93251
12751	1-12751	63	39	5196	78	7507	0	0	2012-08-02 11:22:11.235432	2012-08-02 11:25:08.866093
13204	1-13204	22	77	7015	59	2525	0	3	2012-08-02 11:25:08.866093	2012-08-02 11:25:10.048654
11513	1-11513	35	4	3954	66	1261	0	5	2012-08-02 11:14:09.654949	2012-08-02 11:25:10.048654
11282	1-11282	82	31	4713	43	7329	0	7	2012-08-02 11:12:40.649859	2012-08-02 11:25:10.048654
9029	1-9029	7	20	9369	34	5546	0	0	2012-08-02 05:43:19.630385	2012-08-02 11:25:11.265631
11362	1-11362	72	85	8098	45	6939	0	8	2012-08-02 11:13:11.690241	2012-08-02 11:25:11.265631
13133	1-13133	31	50	2757	20	6151	0	0	2012-08-02 11:24:40.931417	2012-08-02 11:25:11.265631
9705	1-9705	30	38	6695	96	2365	0	0	2012-08-02 05:47:55.374277	2012-08-02 11:25:11.265631
13209	1-13209	80	45	4070	50	9632	0	11	2012-08-02 11:25:11.265631	2012-08-02 11:25:11.265631
13182	1-13182	84	35	3555	39	2868	0	6	2012-08-02 11:25:00.645521	2012-08-02 11:25:12.050696
13213	1-13213	37	72	1852	35	7698	0	10	2012-08-02 11:25:12.050696	2012-08-02 11:25:12.050696
2712	1-2712	37	18	8348	8	1494	1171	0	2012-08-02 05:00:39.297379	2012-08-02 11:25:12.050696
11394	1-11394	32	58	9082	79	9305	0	0	2012-08-02 11:13:24.042075	2012-08-02 11:25:17.553175
1598	1-1598	42	80	3062	42	573	0	0	2012-08-02 04:52:29.294909	2012-08-02 11:25:17.553175
13226	1-13226	35	57	1052	58	9735	0	5	2012-08-02 11:25:17.553175	2012-08-02 11:25:17.553175
6769	1-6769	26	87	8159	91	1258	0	0	2012-08-02 05:27:58.016011	2012-08-02 11:25:24.602023
9617	1-9617	34	3	4601	59	2634	0	16	2012-08-02 05:47:19.16651	2012-08-02 11:25:24.602023
13244	1-13244	13	11	1268	63	8146	0	0	2012-08-02 11:25:24.602023	2012-08-02 11:25:24.602023
12979	1-12979	95	54	3103	29	1596	0	0	2012-08-02 11:23:40.613779	2012-08-02 11:25:26.950328
11849	1-11849	24	21	4048	6	2481	0	3	2012-08-02 11:16:20.79122	2012-08-02 11:25:28.160499
13077	1-13077	24	21	6028	13	8097	0	0	2012-08-02 11:24:19.144836	2012-08-02 11:25:28.160499
2602	1-2602	77	9	1225	73	365	0	0	2012-08-02 04:59:52.523864	2012-08-02 11:25:29.719316
7045	1-7045	2	9	3761	19	1005	0	0	2012-08-02 05:29:49.667562	2012-08-02 11:25:29.719316
13268	1-13268	93	10	3623	12	4117	0	3	2012-08-02 11:25:34.086727	2012-08-02 11:25:34.086727
12614	1-12614	49	89	1553	52	4960	0	7	2012-08-02 11:21:17.787498	2012-08-02 11:25:36.461597
11339	1-11339	11	89	8237	81	8352	0	5	2012-08-02 11:13:02.784225	2012-08-02 11:25:36.461597
11518	1-11518	67	89	4786	97	2134	0	8	2012-08-02 11:14:11.99675	2012-08-02 11:25:36.461597
13280	1-13280	64	54	1213	99	8655	0	3	2012-08-02 11:25:39.159937	2012-08-02 11:25:39.159937
13185	1-13185	9	1	3847	10	8403	0	12	2012-08-02 11:25:01.441006	2012-08-02 11:25:41.957032
12071	1-12071	99	14	6236	13	3274	0	9	2012-08-02 11:17:47.644069	2012-08-02 11:25:41.957032
8702	1-8702	49	75	9510	45	9198	0	0	2012-08-02 05:41:06.549156	2012-08-02 11:25:41.957032
11349	1-11349	11	64	4818	80	5214	0	3	2012-08-02 11:13:06.673873	2012-08-02 11:25:41.957032
13288	1-13288	26	45	3515	12	8426	0	0	2012-08-02 11:25:41.957032	2012-08-02 11:25:41.957032
8098	1-8098	23	14	7570	13	2408	115	3	2012-08-02 05:37:02.499979	2012-08-02 11:25:41.957032
13257	1-13257	39	19	511	99	8451	0	0	2012-08-02 11:25:29.719316	2012-08-02 11:25:43.994978
11650	1-11650	100	61	7410	99	7913	0	9	2012-08-02 11:15:02.79246	2012-08-02 11:25:43.994978
13302	1-13302	81	54	1654	74	7520	0	7	2012-08-02 11:25:47.632607	2012-08-02 11:25:47.632607
11694	1-11694	45	60	7103	96	4534	0	7	2012-08-02 11:15:20.194247	2012-08-02 11:25:50.836867
12498	1-12498	86	5	9735	84	6190	0	0	2012-08-02 11:20:32.051483	2012-08-02 11:25:50.836867
13250	1-13250	97	58	2924	54	9766	0	0	2012-08-02 11:25:26.950328	2012-08-02 11:25:52.405473
12910	1-12910	61	53	4858	13	5040	0	0	2012-08-02 11:23:13.410418	2012-08-02 11:25:56.417239
13116	1-13116	38	29	2262	17	5869	0	0	2012-08-02 11:24:33.89365	2012-08-02 11:26:05.444433
9499	1-9499	48	55	5998	16	6360	0	7	2012-08-02 05:46:31.099521	2012-08-02 11:26:06.593594
13350	1-13350	70	16	1017	17	3555	0	0	2012-08-02 11:26:06.593594	2012-08-02 11:26:06.593594
12299	1-12299	50	59	4968	53	5522	0	9	2012-08-02 11:19:15.144134	2012-08-02 11:26:08.173921
12716	1-12716	8	6	7758	97	6906	0	7	2012-08-02 11:21:57.347689	2012-08-02 11:26:11.325063
12592	1-12592	24	73	5031	19	7642	0	6	2012-08-02 11:21:09.156381	2012-08-02 11:26:11.325063
13009	1-13009	4	63	4044	76	4452	0	5	2012-08-02 11:23:52.372498	2012-08-02 11:26:11.325063
13361	1-13361	51	97	488	73	6820	0	0	2012-08-02 11:26:11.325063	2012-08-02 11:26:11.325063
13311	1-13311	67	90	1651	60	6245	0	4	2012-08-02 11:25:50.836867	2012-08-02 11:26:12.563318
13364	1-13364	83	60	367	90	8884	0	11	2012-08-02 11:26:12.563318	2012-08-02 11:26:12.563318
12800	1-12800	77	66	5362	13	9321	0	7	2012-08-02 11:22:30.890589	2012-08-02 11:26:17.480661
12031	1-12031	96	5	4565	52	3231	0	0	2012-08-02 11:17:31.952414	2012-08-02 11:26:18.69543
13380	1-13380	40	15	1724	5	4400	0	7	2012-08-02 11:26:18.69543	2012-08-02 11:26:18.69543
9296	1-9296	67	38	2607	2	2126	0	0	2012-08-02 05:45:08.63603	2012-08-02 11:26:20.242499
11427	1-11427	24	18	1554	2	5101	0	7	2012-08-02 11:13:36.893379	2012-08-02 11:26:20.242499
13384	1-13384	8	2	2867	61	6551	0	5	2012-08-02 11:26:20.242499	2012-08-02 11:26:20.242499
7752	1-7752	77	43	6204	70	5308	0	13	2012-08-02 05:34:38.609274	2012-08-02 11:26:21.833694
12003	1-12003	72	26	6916	33	5153	0	2	2012-08-02 11:17:20.959129	2012-08-02 11:26:21.833694
13243	1-13243	67	7	1511	98	8902	0	5	2012-08-02 11:25:24.204162	2012-08-02 11:25:24.204162
13178	1-13178	45	85	2174	60	6120	0	0	2012-08-02 11:24:58.314903	2012-08-02 11:25:25.377138
8845	1-8845	41	11	4313	59	2871	0	5	2012-08-02 05:42:04.384437	2012-08-02 11:25:26.565595
13127	1-13127	45	84	8343	68	7050	0	8	2012-08-02 11:24:38.611274	2012-08-02 11:25:26.565595
13249	1-13249	63	27	860	11	2635	0	0	2012-08-02 11:25:26.565595	2012-08-02 11:25:26.565595
12018	1-12018	38	41	6536	47	4868	0	0	2012-08-02 11:17:26.892444	2012-08-02 11:25:28.546934
12888	1-12888	74	50	2805	38	5755	0	5	2012-08-02 11:23:04.891912	2012-08-02 11:25:32.107626
13242	1-13242	70	81	3796	38	6021	0	0	2012-08-02 11:25:23.817019	2012-08-02 11:25:32.107626
12429	1-12429	55	42	1431	11	992	0	3	2012-08-02 11:20:05.485712	2012-08-02 11:25:41.526325
13270	1-13270	56	23	1238	70	5717	0	0	2012-08-02 11:25:34.848544	2012-08-02 11:25:41.526325
8075	1-8075	73	23	7419	48	5592	0	0	2012-08-02 05:36:53.230798	2012-08-02 11:25:41.526325
12922	1-12922	32	12	2953	33	2336	0	0	2012-08-02 11:23:18.149589	2012-08-02 11:25:42.77523
12462	1-12462	1	12	6400	64	7634	0	0	2012-08-02 11:20:18.808868	2012-08-02 11:25:42.77523
13292	1-13292	11	92	416	12	6407	0	6	2012-08-02 11:25:42.77523	2012-08-02 11:25:42.77523
13263	1-13263	95	38	8045	94	6149	0	2	2012-08-02 11:25:32.107626	2012-08-02 11:25:44.780263
11861	1-11861	93	58	5378	80	9801	0	0	2012-08-02 11:16:25.80742	2012-08-02 11:25:44.780263
13319	1-13319	45	92	4280	99	6144	0	0	2012-08-02 11:25:53.985582	2012-08-02 11:25:53.985582
10542	1-10542	53	4	7871	43	4342	0	0	2012-08-02 10:35:32.328138	2012-08-02 11:25:56.803419
13325	1-13325	42	56	367	4	8155	0	7	2012-08-02 11:25:56.803419	2012-08-02 11:25:56.803419
10956	1-10956	32	71	7014	72	4232	0	0	2012-08-02 10:39:05.136158	2012-08-02 11:26:01.079355
13335	1-13335	40	62	529	71	8822	0	0	2012-08-02 11:26:01.079355	2012-08-02 11:26:01.079355
13349	1-13349	33	98	247	78	9137	0	0	2012-08-02 11:26:06.969062	2012-08-02 11:26:06.969062
12861	1-12861	31	37	4354	26	1148	0	0	2012-08-02 11:22:54.371683	2012-08-02 11:26:07.754192
13352	1-13352	59	7	2400	37	9144	0	0	2012-08-02 11:26:07.754192	2012-08-02 11:26:07.754192
13355	1-13355	41	46	144	6	6707	0	18	2012-08-02 11:26:08.958394	2012-08-02 11:26:08.958394
6887	1-6887	93	43	9367	5	6447	0	0	2012-08-02 05:28:46.105031	2012-08-02 11:26:12.950056
13365	1-13365	64	83	1469	43	4858	0	10	2012-08-02 11:26:12.950056	2012-08-02 11:26:12.950056
11525	1-11525	86	46	7383	94	3657	0	0	2012-08-02 11:14:14.70427	2012-08-02 11:26:14.165465
13381	1-13381	56	10	242	61	849	0	0	2012-08-02 11:26:19.468651	2012-08-02 11:26:19.468651
11714	1-11714	87	36	5446	25	6107	0	10	2012-08-02 11:15:28.045657	2012-08-02 11:26:21.446797
12632	1-12632	81	36	3101	92	6527	0	9	2012-08-02 11:21:24.804569	2012-08-02 11:26:21.446797
13387	1-13387	80	25	3835	36	7358	0	0	2012-08-02 11:26:21.446797	2012-08-02 11:26:21.446797
12760	1-12760	37	51	1062	83	2980	0	8	2012-08-02 11:22:14.748595	2012-08-02 11:26:26.596176
11839	1-11839	98	24	5496	72	7048	0	6	2012-08-02 11:16:16.945883	2012-08-02 11:26:26.596176
12803	1-12803	27	80	9786	4	8789	0	6	2012-08-02 11:22:31.663789	2012-08-02 11:26:26.596176
12740	1-12740	77	8	5231	91	7301	0	8	2012-08-02 11:22:07.401958	2012-08-02 11:26:27.778526
13397	1-13397	14	88	6417	75	7594	0	11	2012-08-02 11:26:25.822572	2012-08-02 11:26:28.960842
10950	1-10950	18	74	5153	2	5549	0	2	2012-08-02 10:39:02.715548	2012-08-02 11:26:28.960842
13405	1-13405	42	2	4472	74	6193	0	7	2012-08-02 11:26:28.960842	2012-08-02 11:26:28.960842
10349	1-10349	74	89	4741	72	4828	0	11	2012-08-02 10:27:24.261447	2012-08-02 11:26:34.100074
12075	1-12075	10	20	7662	55	7058	0	8	2012-08-02 11:17:49.124408	2012-08-02 11:26:36.46422
9210	1-9210	29	38	6057	28	1345	0	5	2012-08-02 05:44:33.510785	2012-08-02 11:26:36.46422
13434	1-13434	78	31	1154	76	7633	0	6	2012-08-02 11:26:40.430765	2012-08-02 11:26:40.430765
13441	1-13441	98	77	3253	11	8927	0	0	2012-08-02 11:26:42.816986	2012-08-02 11:26:42.816986
13235	1-13235	3	9	7597	28	9843	0	10	2012-08-02 11:25:21.0275	2012-08-02 11:26:45.648723
11832	1-11832	60	37	9332	81	5130	0	3	2012-08-02 11:16:14.283249	2012-08-02 11:26:52.469722
12236	1-12236	80	38	7356	37	6386	0	17	2012-08-02 11:18:50.879976	2012-08-02 11:26:55.196257
5071	1-5071	14	38	8830	92	4966	0	0	2012-08-02 05:16:31.106782	2012-08-02 11:26:55.196257
11784	1-11784	83	55	5923	92	5179	0	0	2012-08-02 11:15:55.464712	2012-08-02 11:26:55.196257
13471	1-13471	48	37	5434	55	9604	0	4	2012-08-02 11:26:55.196257	2012-08-02 11:26:55.196257
9272	1-9272	48	55	4627	38	887	887	4	2012-08-02 05:44:58.856494	2012-08-02 11:26:55.196257
11203	1-11203	6	73	5323	9	5832	0	9	2012-08-02 11:12:09.711611	2012-08-02 11:26:57.671379
12955	1-12955	10	3	4499	87	8532	0	0	2012-08-02 11:23:31.234028	2012-08-02 11:26:57.671379
6392	1-6392	80	71	6859	38	4125	0	4	2012-08-02 05:25:25.69154	2012-08-02 11:26:57.671379
10863	1-10863	21	38	9380	87	7397	0	10	2012-08-02 10:38:28.440531	2012-08-02 11:26:57.671379
12438	1-12438	69	73	9883	63	8668	0	0	2012-08-02 11:20:08.988553	2012-08-02 11:26:57.671379
12652	1-12652	10	76	4345	55	7871	0	7	2012-08-02 11:21:32.627116	2012-08-02 11:26:58.843133
12663	1-12663	82	64	4477	55	6876	0	6	2012-08-02 11:21:37.389028	2012-08-02 11:26:58.843133
11824	1-11824	71	71	9440	52	8222	0	5	2012-08-02 11:16:11.234151	2012-08-02 11:27:02.776472
10022	1-10022	90	71	9411	21	7010	0	5	2012-08-02 10:19:31.042844	2012-08-02 11:27:02.776472
13490	1-13490	97	81	1024	71	7023	0	0	2012-08-02 11:27:02.776472	2012-08-02 11:27:02.776472
13274	1-13274	47	67	3776	89	8713	0	0	2012-08-02 11:25:36.461597	2012-08-02 11:27:05.118969
9558	1-9558	36	76	9981	80	6284	0	0	2012-08-02 05:46:55.054987	2012-08-02 11:27:06.323253
13236	1-13236	40	91	7437	76	9381	0	5	2012-08-02 11:25:21.415457	2012-08-02 11:27:06.323253
11323	1-11323	52	91	9381	99	8467	0	3	2012-08-02 11:12:56.619458	2012-08-02 11:27:06.323253
4994	1-4994	25	52	4393	85	2082	0	6	2012-08-02 05:15:59.999893	2012-08-02 11:27:08.65467
13260	1-13260	4	52	5541	81	3163	0	6	2012-08-02 11:25:30.515029	2012-08-02 11:27:08.65467
10510	1-10510	87	84	5180	85	6843	0	0	2012-08-02 10:32:00.882391	2012-08-02 11:27:08.65467
13505	1-13505	58	85	1695	52	8847	0	2	2012-08-02 11:27:08.65467	2012-08-02 11:27:08.65467
12425	1-12425	46	100	8736	68	6902	0	0	2012-08-02 11:20:03.938555	2012-08-02 11:27:11.019619
11136	1-11136	72	63	9389	23	9537	5629	10	2012-08-02 11:11:43.40729	2012-08-02 11:27:11.019619
13151	1-13151	53	59	1782	72	2379	0	10	2012-08-02 11:24:48.028151	2012-08-02 11:27:12.224339
12917	1-12917	35	93	5737	46	2722	0	0	2012-08-02 11:23:16.183658	2012-08-02 11:27:13.362106
3282	1-3282	8	93	2886	88	930	0	0	2012-08-02 05:04:33.363379	2012-08-02 11:27:13.362106
8824	1-8824	40	93	8999	65	4585	0	0	2012-08-02 05:41:55.954218	2012-08-02 11:27:13.362106
13517	1-13517	20	88	1917	93	7655	0	11	2012-08-02 11:27:13.362106	2012-08-02 11:27:13.362106
13424	1-13424	9	55	2660	38	8473	0	11	2012-08-02 11:26:36.46422	2012-08-02 11:27:15.385591
13385	1-13385	63	100	1992	38	3037	0	8	2012-08-02 11:26:21.062162	2012-08-02 11:27:15.385591
13522	1-13522	98	38	3557	51	2881	0	4	2012-08-02 11:27:15.385591	2012-08-02 11:27:15.385591
12724	1-12724	71	90	9538	91	7745	0	0	2012-08-02 11:22:00.84848	2012-08-02 11:27:16.568012
8072	1-8072	49	90	7987	60	5595	0	0	2012-08-02 05:36:52.026458	2012-08-02 11:27:16.568012
13334	1-13334	98	13	844	43	8836	0	8	2012-08-02 11:26:00.681608	2012-08-02 11:26:00.681608
10109	1-10109	88	3	4311	94	2978	0	0	2012-08-02 10:21:23.562326	2012-08-02 11:26:04.262141
13129	1-13129	38	25	7503	3	6944	0	0	2012-08-02 11:24:39.384613	2012-08-02 11:26:04.262141
13342	1-13342	46	58	554	25	7769	0	13	2012-08-02 11:26:04.262141	2012-08-02 11:26:04.262141
13345	1-13345	11	41	43	19	2778	0	11	2012-08-02 11:26:05.046846	2012-08-02 11:26:05.046846
6161	1-6161	72	82	5346	70	2696	0	9	2012-08-02 05:23:49.735257	2012-08-02 11:26:08.571845
13188	1-13188	80	26	1064	87	4592	0	8	2012-08-02 11:25:02.645589	2012-08-02 11:26:08.571845
12985	1-12985	68	79	4225	63	4166	0	14	2012-08-02 11:23:42.569605	2012-08-02 11:26:08.571845
13354	1-13354	5	70	9828	26	9736	0	0	2012-08-02 11:26:08.571845	2012-08-02 11:26:08.571845
7257	1-7257	97	17	5963	10	2252	0	4	2012-08-02 05:31:16.351965	2012-08-02 11:26:10.938883
12818	1-12818	43	49	3232	17	4136	0	6	2012-08-02 11:22:37.466277	2012-08-02 11:26:10.938883
10680	1-10680	35	92	6521	59	2278	0	4	2012-08-02 10:36:13.329885	2012-08-02 11:26:10.938883
4883	1-4883	57	49	6224	63	1258	0	8	2012-08-02 05:15:16.545072	2012-08-02 11:26:10.938883
13360	1-13360	81	30	1828	49	9833	0	9	2012-08-02 11:26:10.938883	2012-08-02 11:26:10.938883
7251	1-7251	59	30	8346	49	5766	1778	0	2012-08-02 05:31:14.064644	2012-08-02 11:26:10.938883
12074	1-12074	99	69	4285	76	2772	0	0	2012-08-02 11:17:48.749121	2012-08-02 11:26:12.132263
11716	1-11716	91	3	8873	64	7899	0	20	2012-08-02 11:15:28.809227	2012-08-02 11:26:13.358762
12590	1-12590	53	3	1571	20	2832	0	8	2012-08-02 11:21:08.372282	2012-08-02 11:26:13.358762
11417	1-11417	53	91	9990	25	7667	0	4	2012-08-02 11:13:33.048464	2012-08-02 11:26:13.358762
11728	1-11728	77	72	3857	22	1600	0	0	2012-08-02 11:15:33.474112	2012-08-02 11:26:13.358762
13366	1-13366	73	22	3601	3	9877	0	6	2012-08-02 11:26:13.358762	2012-08-02 11:26:13.358762
9702	1-9702	26	11	4591	22	3247	0	0	2012-08-02 05:47:54.170021	2012-08-02 11:26:15.336549
10595	1-10595	37	34	6155	22	4085	0	14	2012-08-02 10:35:36.370076	2012-08-02 11:26:15.336549
7724	1-7724	56	11	5367	29	5745	0	5	2012-08-02 05:34:26.663312	2012-08-02 11:26:15.336549
13287	1-13287	30	11	3155	23	5899	0	7	2012-08-02 11:25:41.526325	2012-08-02 11:26:15.336549
13371	1-13371	38	22	3856	11	8529	0	0	2012-08-02 11:26:15.336549	2012-08-02 11:26:15.336549
3105	1-3105	38	11	3484	23	225	225	0	2012-08-02 05:03:21.846383	2012-08-02 11:26:15.336549
12523	1-12523	18	79	5559	29	5467	0	6	2012-08-02 11:20:42.145201	2012-08-02 11:26:19.855415
13382	1-13382	95	43	3415	30	7743	0	4	2012-08-02 11:26:19.855415	2012-08-02 11:26:19.855415
10885	1-10885	57	4	9347	74	4874	0	9	2012-08-02 10:38:37.060068	2012-08-02 11:26:22.231701
13388	1-13388	6	13	1446	4	5811	0	9	2012-08-02 11:26:22.231701	2012-08-02 11:26:22.231701
9119	1-9119	86	57	4617	12	1806	0	0	2012-08-02 05:43:56.63839	2012-08-02 11:26:24.64072
13161	1-13161	15	92	5000	96	4733	0	2	2012-08-02 11:24:51.961695	2012-08-02 11:26:24.64072
10806	1-10806	94	100	6869	21	9691	0	0	2012-08-02 10:38:06.064779	2012-08-02 11:26:26.20944
12336	1-12336	44	100	3407	81	8154	0	0	2012-08-02 11:19:29.595325	2012-08-02 11:26:26.20944
13401	1-13401	91	15	1486	68	1033	0	5	2012-08-02 11:26:27.391834	2012-08-02 11:26:27.391834
12289	1-12289	63	33	5733	1	5680	0	0	2012-08-02 11:19:11.299369	2012-08-02 11:26:28.56418
13404	1-13404	38	65	241	33	5138	0	0	2012-08-02 11:26:28.56418	2012-08-02 11:26:28.56418
13071	1-13071	58	68	5714	26	9844	0	0	2012-08-02 11:24:16.777883	2012-08-02 11:26:38.055142
13428	1-13428	19	26	2575	31	8657	0	9	2012-08-02 11:26:38.055142	2012-08-02 11:26:38.055142
2178	1-2178	2	26	3918	31	1697	815	0	2012-08-02 04:56:47.84789	2012-08-02 11:26:38.055142
12697	1-12697	39	25	6965	22	8481	0	22	2012-08-02 11:21:50.318842	2012-08-02 11:26:40.03285
13433	1-13433	38	22	4805	100	5586	0	7	2012-08-02 11:26:40.03285	2012-08-02 11:26:40.03285
13353	1-13353	65	100	6450	59	7447	0	7	2012-08-02 11:26:08.173921	2012-08-02 11:26:41.226091
11785	1-11785	7	100	6238	47	5966	0	5	2012-08-02 11:15:55.862637	2012-08-02 11:26:41.226091
13138	1-13138	63	65	6032	9	6381	0	0	2012-08-02 11:24:42.511782	2012-08-02 11:26:43.601757
12349	1-12349	50	63	5406	70	8983	0	3	2012-08-02 11:19:34.602459	2012-08-02 11:26:47.220183
13008	1-13008	76	63	944	87	6695	0	3	2012-08-02 11:23:51.974777	2012-08-02 11:26:47.220183
13041	1-13041	32	72	7240	70	7015	0	7	2012-08-02 11:24:04.779544	2012-08-02 11:26:47.220183
12586	1-12586	72	98	3700	70	1387	0	2	2012-08-02 11:21:06.74773	2012-08-02 11:26:47.220183
13450	1-13450	25	70	3137	63	8531	0	0	2012-08-02 11:26:47.220183	2012-08-02 11:26:47.220183
13253	1-13253	34	90	1973	41	4286	0	8	2012-08-02 11:25:28.546934	2012-08-02 11:26:48.013555
12398	1-12398	18	19	1701	79	2534	0	0	2012-08-02 11:19:53.528598	2012-08-02 11:26:54.810076
13470	1-13470	63	34	2840	65	5797	0	8	2012-08-02 11:26:54.810076	2012-08-02 11:26:54.810076
13286	1-13286	28	6	8689	65	9798	0	0	2012-08-02 11:25:41.128704	2012-08-02 11:26:58.444928
8205	1-8205	67	54	9028	91	3886	0	7	2012-08-02 05:37:46.134829	2012-08-02 11:27:01.95929
13440	1-13440	10	77	2847	44	6205	0	0	2012-08-02 11:26:43.204062	2012-08-02 11:27:06.710134
13500	1-13500	28	50	660	77	6954	0	10	2012-08-02 11:27:06.710134	2012-08-02 11:27:06.710134
7347	1-7347	6	1	7129	95	6620	0	5	2012-08-02 05:31:51.752848	2012-08-02 11:27:12.599056
13200	1-13200	62	26	5599	18	6026	0	11	2012-08-02 11:25:07.330226	2012-08-02 11:27:12.599056
5542	1-5542	86	61	3006	10	1619	0	0	2012-08-02 05:19:37.953378	2012-08-02 11:27:12.599056
13515	1-13515	30	90	2001	1	9073	0	0	2012-08-02 11:27:12.599056	2012-08-02 11:27:12.599056
12207	1-12207	21	99	4680	77	4361	0	0	2012-08-02 11:18:39.707781	2012-08-02 11:27:15.009877
13241	1-13241	73	26	1981	82	8143	0	0	2012-08-02 11:25:23.418435	2012-08-02 11:27:16.943419
13525	1-13525	43	82	2054	28	4887	0	5	2012-08-02 11:27:16.943419	2012-08-02 11:27:16.943419
12583	1-12583	33	68	4480	67	7757	0	2	2012-08-02 11:21:05.565369	2012-08-02 11:27:17.739084
13528	1-13528	56	70	720	3	2835	0	6	2012-08-02 11:27:17.739084	2012-08-02 11:27:17.739084
13191	1-13191	44	86	2408	72	3202	0	8	2012-08-02 11:25:03.396764	2012-08-02 11:27:20.169976
13477	1-13477	63	87	8766	73	8505	0	0	2012-08-02 11:26:57.671379	2012-08-02 11:27:22.677663
13436	1-13436	22	11	5523	100	5636	0	0	2012-08-02 11:26:41.226091	2012-08-02 11:27:27.04239
11210	1-11210	79	44	5138	9	1860	0	0	2012-08-02 11:12:12.451122	2012-08-02 11:27:27.848631
13553	1-13553	74	38	3552	56	5750	0	7	2012-08-02 11:27:27.848631	2012-08-02 11:27:27.848631
13563	1-13563	26	99	3	14	8555	0	0	2012-08-02 11:27:31.75991	2012-08-02 11:27:31.75991
13567	1-13567	25	17	1927	43	7374	0	0	2012-08-02 11:27:32.974984	2012-08-02 11:27:32.974984
12975	1-12975	9	37	6487	99	3047	0	7	2012-08-02 11:23:39.067216	2012-08-02 11:27:38.068659
12403	1-12403	33	39	7832	65	9763	0	5	2012-08-02 11:19:55.462086	2012-08-02 11:27:39.261479
9037	1-9037	37	19	7916	10	8687	0	0	2012-08-02 05:43:22.911479	2012-08-02 11:27:39.261479
12920	1-12920	86	39	6258	19	7340	0	0	2012-08-02 11:23:17.376362	2012-08-02 11:27:39.261479
13002	1-13002	22	9	2270	38	4112	0	0	2012-08-02 11:23:49.632673	2012-08-02 11:27:39.261479
12585	1-12585	59	61	5603	38	7770	0	0	2012-08-02 11:21:06.350101	2012-08-02 11:27:39.261479
13206	1-13206	65	56	2924	44	5561	0	15	2012-08-02 11:25:09.263834	2012-08-02 11:27:41.683599
13386	1-13386	88	70	3643	26	7909	0	5	2012-08-02 11:26:21.833694	2012-08-02 11:26:21.833694
12400	1-12400	44	25	4762	47	3777	0	17	2012-08-02 11:19:54.302253	2012-08-02 11:26:22.651417
5200	1-5200	20	22	9811	50	3842	0	0	2012-08-02 05:17:23.062504	2012-08-02 11:26:22.651417
10468	1-10468	95	57	5235	6	3283	0	4	2012-08-02 10:31:44.191083	2012-08-02 11:26:22.651417
13389	1-13389	46	6	561	25	7223	0	0	2012-08-02 11:26:22.651417	2012-08-02 11:26:22.651417
13392	1-13392	82	1	489	96	2863	0	9	2012-08-02 11:26:23.856512	2012-08-02 11:26:23.856512
10929	1-10929	80	63	4257	88	8759	0	8	2012-08-02 10:38:54.397898	2012-08-02 11:26:25.822572
11458	1-11458	50	75	9023	90	5443	0	11	2012-08-02 11:13:48.814959	2012-08-02 11:26:25.822572
13233	1-13233	86	90	3431	78	5510	0	2	2012-08-02 11:25:20.254567	2012-08-02 11:26:28.176291
13326	1-13326	30	95	1334	32	3467	0	6	2012-08-02 11:25:57.610274	2012-08-02 11:26:32.564429
13426	1-13426	84	77	348	80	5116	0	0	2012-08-02 11:26:37.271162	2012-08-02 11:26:37.271162
11956	1-11956	93	44	9665	9	6878	0	2	2012-08-02 11:17:02.419952	2012-08-02 11:26:43.204062
11921	1-11921	54	44	4964	55	5198	0	13	2012-08-02 11:16:48.67307	2012-08-02 11:26:43.204062
5105	1-5105	91	55	9101	11	629	629	0	2012-08-02 05:16:45.093994	2012-08-02 11:26:43.204062
12474	1-12474	56	22	721	1	235	0	11	2012-08-02 11:20:23.092205	2012-08-02 11:26:44.023142
13374	1-13374	64	22	2291	79	505	0	10	2012-08-02 11:26:16.552352	2012-08-02 11:26:44.023142
13443	1-13443	20	72	1338	22	5337	0	3	2012-08-02 11:26:44.023142	2012-08-02 11:26:44.023142
13254	1-13254	37	97	2905	21	6914	0	0	2012-08-02 11:25:28.160499	2012-08-02 11:26:48.433006
13068	1-13068	9	60	5056	87	6130	0	7	2012-08-02 11:24:15.242404	2012-08-02 11:26:48.433006
11752	1-11752	45	49	5768	3	3799	0	6	2012-08-02 11:15:43.257718	2012-08-02 11:26:48.433006
13454	1-13454	6	87	5211	97	4924	0	3	2012-08-02 11:26:48.433006	2012-08-02 11:26:48.433006
9693	1-9693	79	74	9142	53	7130	0	0	2012-08-02 05:47:50.457537	2012-08-02 11:26:49.593691
13457	1-13457	68	53	3324	74	9037	0	4	2012-08-02 11:26:49.593691	2012-08-02 11:26:49.593691
12750	1-12750	37	96	2860	21	5736	0	0	2012-08-02 11:22:10.837455	2012-08-02 11:26:55.649625
12409	1-12409	15	54	4473	21	3954	0	0	2012-08-02 11:19:57.793276	2012-08-02 11:26:55.649625
11630	1-11630	68	63	7815	53	9233	0	4	2012-08-02 11:14:54.660834	2012-08-02 11:26:56.88707
6193	1-6193	93	73	6342	65	3370	0	0	2012-08-02 05:24:02.781518	2012-08-02 11:26:56.88707
13181	1-13181	19	53	4685	46	4328	0	4	2012-08-02 11:24:59.905458	2012-08-02 11:26:56.88707
12631	1-12631	74	51	5380	46	9345	0	6	2012-08-02 11:21:24.417601	2012-08-02 11:26:56.88707
3196	1-3196	99	53	6431	46	1096	1096	0	2012-08-02 05:03:58.836118	2012-08-02 11:26:56.88707
13481	1-13481	34	37	294	18	342	0	9	2012-08-02 11:26:59.229694	2012-08-02 11:26:59.229694
12720	1-12720	44	53	3028	7	6309	0	5	2012-08-02 11:21:59.268612	2012-08-02 11:27:01.561032
13472	1-13472	19	21	1496	25	6112	0	0	2012-08-02 11:26:55.649625	2012-08-02 11:27:01.561032
8524	1-8524	89	23	4861	40	6531	0	4	2012-08-02 05:39:54.73686	2012-08-02 11:27:01.561032
13486	1-13486	7	7	2528	21	3495	0	10	2012-08-02 11:27:01.561032	2012-08-02 11:27:01.561032
7650	1-7650	27	52	7094	2	6656	0	11	2012-08-02 05:33:55.783382	2012-08-02 11:27:03.56104
11594	1-11594	68	81	7434	40	7928	0	3	2012-08-02 11:14:41.111776	2012-08-02 11:27:03.56104
11759	1-11759	52	72	4756	2	4394	0	2	2012-08-02 11:15:45.56318	2012-08-02 11:27:03.56104
12802	1-12802	48	34	4608	31	2230	0	7	2012-08-02 11:22:31.288271	2012-08-02 11:27:05.925457
13229	1-13229	16	35	4863	68	7523	0	3	2012-08-02 11:25:18.70535	2012-08-02 11:27:05.925457
13323	1-13323	4	77	3383	34	5953	0	3	2012-08-02 11:25:56.417239	2012-08-02 11:27:07.097589
13363	1-13363	58	11	1284	69	2747	0	0	2012-08-02 11:26:12.132263	2012-08-02 11:27:07.097589
13504	1-13504	45	4	439	44	3758	0	6	2012-08-02 11:27:08.268094	2012-08-02 11:27:08.268094
13089	1-13089	5	33	6730	28	5636	0	7	2012-08-02 11:24:23.397931	2012-08-02 11:27:21.750267
12715	1-12715	15	33	2443	49	2226	0	9	2012-08-02 11:21:57.732986	2012-08-02 11:27:21.750267
13538	1-13538	92	48	170	33	6837	0	4	2012-08-02 11:27:21.750267	2012-08-02 11:27:21.750267
12797	1-12797	76	5	4923	83	9571	0	2	2012-08-02 11:22:29.332704	2012-08-02 11:27:28.61077
5899	1-5899	56	32	6286	2	7492	0	0	2012-08-02 05:22:02.537106	2012-08-02 11:27:34.135171
13555	1-13555	78	99	2614	5	8646	0	4	2012-08-02 11:27:28.61077	2012-08-02 11:27:38.853085
13246	1-13246	59	60	3489	5	2187	0	0	2012-08-02 11:25:25.377138	2012-08-02 11:27:38.853085
13240	1-13240	53	96	6988	31	5892	0	0	2012-08-02 11:25:23.031427	2012-08-02 11:27:16.568012
12547	1-12547	78	90	7400	23	9767	0	0	2012-08-02 11:20:51.473831	2012-08-02 11:27:16.568012
4387	1-4387	28	91	9147	96	1700	1700	0	2012-08-02 05:11:56.863464	2012-08-02 11:27:16.568012
13536	1-13536	7	8	872	64	2529	0	5	2012-08-02 11:27:20.943618	2012-08-02 11:27:20.943618
13152	1-13152	97	45	6087	82	9765	0	5	2012-08-02 11:24:48.415244	2012-08-02 11:27:23.488864
8011	1-8011	68	82	9097	80	3254	1486	7	2012-08-02 05:36:27.55314	2012-08-02 11:27:23.488864
4267	1-4267	50	45	3927	82	117	117	0	2012-08-02 05:11:07.563372	2012-08-02 11:27:23.488864
13038	1-13038	46	69	5333	32	4641	0	11	2012-08-02 11:24:03.597316	2012-08-02 11:27:26.633186
13511	1-13511	18	68	5351	63	4640	0	13	2012-08-02 11:27:11.019619	2012-08-02 11:27:29.384122
12574	1-12574	98	68	3074	46	847	0	9	2012-08-02 11:21:02.062718	2012-08-02 11:27:29.384122
6531	1-6531	86	61	7836	73	4298	0	0	2012-08-02 05:26:21.945898	2012-08-02 11:27:29.384122
3999	1-3999	67	65	4132	61	1283	82	0	2012-08-02 05:09:21.263385	2012-08-02 11:27:29.384122
6033	1-6033	8	91	9253	42	4482	0	8	2012-08-02 05:22:58.21553	2012-08-02 11:27:32.588199
13565	1-13565	91	27	293	80	3103	0	2	2012-08-02 11:27:32.588199	2012-08-02 11:27:32.588199
12987	1-12987	10	92	2960	15	9702	0	7	2012-08-02 11:23:43.77499	2012-08-02 11:27:33.74851
13465	1-13465	44	40	3788	15	9164	0	11	2012-08-02 11:26:52.469722	2012-08-02 11:27:33.74851
9295	1-9295	96	20	4861	2	4365	0	5	2012-08-02 05:45:08.238287	2012-08-02 11:27:33.74851
13568	1-13568	85	15	1837	8	7075	0	0	2012-08-02 11:27:33.74851	2012-08-02 11:27:33.74851
13583	1-13583	2	34	3776	79	1361	0	19	2012-08-02 11:27:39.648159	2012-08-02 11:27:39.648159
13590	1-13590	31	25	37	79	722	0	0	2012-08-02 11:27:42.932116	2012-08-02 11:27:42.932116
13475	1-13475	84	46	5635	73	9551	0	0	2012-08-02 11:26:56.88707	2012-08-02 11:27:41.683599
11967	1-11967	7	28	2933	21	4540	0	9	2012-08-02 11:17:06.352061	2012-08-02 11:27:41.683599
13589	1-13589	4	44	4340	28	8173	0	3	2012-08-02 11:27:41.683599	2012-08-02 11:27:41.683599
10344	1-10344	12	4	9820	42	7757	2357	0	2012-08-02 10:27:22.161711	2012-08-02 11:27:41.683599
3489	1-3489	19	15	7343	43	929	0	0	2012-08-02 05:05:57.373449	2012-08-02 11:27:43.340954
12576	1-12576	55	97	3054	48	2493	0	9	2012-08-02 11:21:02.836286	2012-08-02 11:27:43.340954
11374	1-11374	78	97	8278	33	2662	0	14	2012-08-02 11:13:16.265532	2012-08-02 11:27:43.340954
13346	1-13346	84	17	5948	55	9475	0	8	2012-08-02 11:26:05.444433	2012-08-02 11:27:44.513646
13314	1-13314	60	35	1029	47	3368	0	7	2012-08-02 11:25:52.803265	2012-08-02 11:27:44.513646
6155	1-6155	79	3	8276	2	8843	0	0	2012-08-02 05:23:47.321085	2012-08-02 11:27:45.695868
13442	1-13442	93	9	2524	50	4720	0	6	2012-08-02 11:26:43.601757	2012-08-02 11:27:45.695868
13597	1-13597	32	2	999	9	8663	0	3	2012-08-02 11:27:45.695868	2012-08-02 11:27:45.695868
13202	1-13202	22	1	4074	42	7078	0	15	2012-08-02 11:25:07.717116	2012-08-02 11:27:46.491424
13306	1-13306	15	57	3120	92	4104	0	3	2012-08-02 11:25:49.212646	2012-08-02 11:27:46.491424
11011	1-11011	20	26	4663	92	4250	0	0	2012-08-02 10:39:27.061241	2012-08-02 11:27:46.491424
13600	1-13600	13	92	5076	1	6685	0	24	2012-08-02 11:27:46.491424	2012-08-02 11:27:46.491424
1518	1-1518	24	42	2143	57	34	34	0	2012-08-02 04:51:48.303974	2012-08-02 11:27:46.491424
13261	1-13261	11	58	518	24	847	0	5	2012-08-02 11:25:31.323161	2012-08-02 11:27:47.287245
13605	1-13605	22	81	77	69	2509	0	0	2012-08-02 11:27:48.491294	2012-08-02 11:27:48.491294
12331	1-12331	58	76	3077	50	3655	0	5	2012-08-02 11:19:27.695107	2012-08-02 11:27:49.662717
12144	1-12144	94	76	7810	79	6537	0	0	2012-08-02 11:18:15.631649	2012-08-02 11:27:49.662717
13608	1-13608	2	77	92	76	9818	0	6	2012-08-02 11:27:49.662717	2012-08-02 11:27:49.662717
10945	1-10945	66	76	7217	97	3230	1156	14	2012-08-02 10:39:00.760215	2012-08-02 11:27:49.662717
13610	1-13610	69	89	304	64	5479	0	0	2012-08-02 11:27:50.847531	2012-08-02 11:27:50.847531
3236	1-3236	22	59	7020	89	857	98	0	2012-08-02 05:04:15.186765	2012-08-02 11:27:50.847531
12245	1-12245	50	69	8135	81	8114	0	7	2012-08-02 11:18:54.382368	2012-08-02 11:27:52.435896
13615	1-13615	51	95	3212	69	8864	0	16	2012-08-02 11:27:52.435896	2012-08-02 11:27:52.435896
10908	1-10908	50	77	7256	34	2367	0	0	2012-08-02 10:38:46.055248	2012-08-02 11:27:54.733372
11624	1-11624	48	5	4886	68	1965	0	9	2012-08-02 11:14:52.705029	2012-08-02 11:27:54.733372
12443	1-12443	1	75	5709	58	4399	0	0	2012-08-02 11:20:10.955201	2012-08-02 11:27:55.871376
10985	1-10985	65	20	7253	65	3074	0	0	2012-08-02 10:39:16.468513	2012-08-02 11:27:55.871376
6317	1-6317	19	61	4441	5	3420	0	0	2012-08-02 05:24:54.826942	2012-08-02 11:27:56.667174
12374	1-12374	12	42	2470	5	1760	0	5	2012-08-02 11:19:44.237284	2012-08-02 11:27:56.667174
11288	1-11288	90	61	8188	62	8346	0	0	2012-08-02 11:12:43.02552	2012-08-02 11:27:56.667174
13625	1-13625	100	5	2473	61	7393	0	6	2012-08-02 11:27:56.667174	2012-08-02 11:27:56.667174
13313	1-13313	94	21	5257	23	6550	0	6	2012-08-02 11:25:52.405473	2012-08-02 11:27:57.176007
12596	1-12596	24	34	6561	32	9520	0	8	2012-08-02 11:21:10.736867	2012-08-02 11:27:56.258298
12780	1-12780	69	57	3259	20	7221	0	2	2012-08-02 11:22:22.736954	2012-08-02 11:27:56.258298
4905	1-4905	46	20	9304	34	2602	2602	0	2012-08-02 05:15:25.11807	2012-08-02 11:27:56.258298
11872	1-11872	49	34	7016	52	9853	0	0	2012-08-02 11:16:29.687373	2012-08-02 11:27:58.263015
11708	1-11708	71	46	2998	2	2995	0	0	2012-08-02 11:15:25.713717	2012-08-02 11:27:58.263015
13520	1-13520	58	77	5503	60	9296	0	4	2012-08-02 11:27:15.009877	2012-08-02 11:27:57.794761
13631	1-13631	94	75	691	69	4099	0	0	2012-08-02 11:27:58.733665	2012-08-02 11:27:58.733665
11905	1-11905	59	27	6788	50	7863	0	4	2012-08-02 11:16:42.349933	2012-08-02 11:27:59.982053
13533	1-13533	10	28	2009	4	3935	0	9	2012-08-02 11:27:19.36346	2012-08-02 11:27:59.982053
13633	1-13633	71	50	3811	28	6955	0	0	2012-08-02 11:27:59.982053	2012-08-02 11:27:59.982053
11075	1-11075	7	79	6604	100	4030	0	10	2012-08-02 11:11:19.594987	2012-08-02 11:28:00.831823
13635	1-13635	23	2	777	45	2818	0	0	2012-08-02 11:28:00.396833	2012-08-02 11:28:00.396833
11287	1-11287	13	98	5618	55	3550	0	10	2012-08-02 11:12:42.628102	2012-08-02 11:28:02.777661
12719	1-12719	25	98	5146	80	7760	0	15	2012-08-02 11:21:58.89314	2012-08-02 11:28:06.634738
13221	1-13221	72	52	8047	72	8885	0	0	2012-08-02 11:25:15.586519	2012-08-02 11:28:06.634738
12657	1-12657	83	10	3521	20	8753	0	10	2012-08-02 11:21:34.593518	2012-08-02 11:28:07.419662
12421	1-12421	83	45	7072	14	8725	0	7	2012-08-02 11:20:02.402723	2012-08-02 11:28:07.419662
10499	1-10499	43	10	5736	38	4795	0	6	2012-08-02 10:31:56.539513	2012-08-02 11:28:07.419662
7989	1-7989	99	98	5687	37	5463	0	0	2012-08-02 05:36:18.659191	2012-08-02 11:28:07.419662
13557	1-13557	45	73	3204	68	8032	0	4	2012-08-02 11:27:29.384122	2012-08-02 11:28:08.203628
12441	1-12441	9	100	2352	2	3623	0	0	2012-08-02 11:20:10.160025	2012-08-02 11:28:09.375063
11206	1-11206	78	96	8390	99	6794	0	3	2012-08-02 11:12:10.871355	2012-08-02 11:28:09.375063
13634	1-13634	10	100	1170	96	2116	0	0	2012-08-02 11:28:00.831823	2012-08-02 11:28:09.375063
13598	1-13598	66	74	321	7	1232	0	5	2012-08-02 11:27:45.297993	2012-08-02 11:28:09.375063
12602	1-12602	60	15	427	99	126	0	0	2012-08-02 11:21:13.102981	2012-08-02 11:28:09.375063
10742	1-10742	5	100	5342	47	2971	0	2	2012-08-02 10:37:41.457705	2012-08-02 11:28:09.375063
13657	1-13657	75	20	1699	100	9845	0	14	2012-08-02 11:28:09.375063	2012-08-02 11:28:09.375063
13659	1-13659	53	59	2516	77	5375	0	8	2012-08-02 11:28:10.148295	2012-08-02 11:28:10.148295
12505	1-12505	95	17	4872	92	7604	0	3	2012-08-02 11:20:35.168872	2012-08-02 11:28:10.568682
12916	1-12916	19	92	4423	37	7934	0	5	2012-08-02 11:23:15.796603	2012-08-02 11:28:13.275174
13118	1-13118	32	93	6053	62	2333	0	0	2012-08-02 11:24:34.667439	2012-08-02 11:28:13.275174
13666	1-13666	48	37	4213	93	8945	0	2	2012-08-02 11:28:13.275174	2012-08-02 11:28:13.275174
11284	1-11284	73	24	4858	77	9267	0	0	2012-08-02 11:12:41.025309	2012-08-02 11:28:12.888559
13653	1-13653	28	14	7084	10	5575	0	0	2012-08-02 11:28:07.419662	2012-08-02 11:28:12.888559
13667	1-13667	37	84	4513	24	7998	0	0	2012-08-02 11:28:12.888559	2012-08-02 11:28:12.888559
12183	1-12183	34	59	8851	11	8017	0	7	2012-08-02 11:18:30.878229	2012-08-02 11:28:14.457753
13592	1-13592	59	63	2017	97	8395	0	7	2012-08-02 11:27:43.340954	2012-08-02 11:28:14.457753
13514	1-13514	55	67	4211	59	5816	0	8	2012-08-02 11:27:12.224339	2012-08-02 11:28:14.457753
12944	1-12944	83	73	5279	38	9999	0	0	2012-08-02 11:23:26.935786	2012-08-02 11:28:14.457753
13670	1-13670	59	38	3465	67	9583	0	17	2012-08-02 11:28:14.457753	2012-08-02 11:28:14.457753
9308	1-9308	46	95	8263	60	6358	0	4	2012-08-02 05:45:13.464635	2012-08-02 11:28:14.866431
13671	1-13671	28	60	2818	95	7773	0	13	2012-08-02 11:28:14.866431	2012-08-02 11:28:14.866431
12533	1-12533	14	74	2262	84	7250	0	0	2012-08-02 11:20:46.071419	2012-08-02 11:28:16.424944
12388	1-12388	10	97	9297	49	9350	0	3	2012-08-02 11:19:49.673039	2012-08-02 11:28:16.424944
13399	1-13399	89	83	6359	24	6555	0	7	2012-08-02 11:26:26.596176	2012-08-02 11:28:16.811248
13679	1-13679	21	10	752	31	7458	0	0	2012-08-02 11:28:18.391519	2012-08-02 11:28:18.391519
13588	1-13588	45	45	6581	95	9070	0	19	2012-08-02 11:27:42.155272	2012-08-02 11:28:19.186746
13414	1-13414	2	59	6417	95	8895	0	0	2012-08-02 11:26:32.564429	2012-08-02 11:28:22.314052
13464	1-13464	4	17	4983	99	4774	0	8	2012-08-02 11:26:52.854644	2012-08-02 11:28:22.314052
13006	1-13006	22	59	2610	89	3742	0	9	2012-08-02 11:23:51.179364	2012-08-02 11:28:22.314052
13690	1-13690	18	99	4595	59	9439	0	5	2012-08-02 11:28:22.314052	2012-08-02 11:28:22.314052
13699	1-13699	23	30	1543	21	6305	0	13	2012-08-02 11:28:25.826838	2012-08-02 11:28:25.826838
13628	1-13628	51	52	3942	46	4769	0	6	2012-08-02 11:27:58.263015	2012-08-02 11:28:30.157943
13710	1-13710	75	46	3349	24	7924	0	0	2012-08-02 11:28:30.157943	2012-08-02 11:28:30.157943
13651	1-13651	54	72	2128	98	4138	0	0	2012-08-02 11:28:06.634738	2012-08-02 11:28:32.889244
10536	1-10536	45	73	6883	2	7548	0	5	2012-08-02 10:32:11.162677	2012-08-02 11:28:34.027524
13720	1-13720	78	2	3512	73	8015	0	6	2012-08-02 11:28:34.027524	2012-08-02 11:28:34.027524
13739	1-13739	89	54	322	35	5373	0	0	2012-08-02 11:28:41.71681	2012-08-02 11:28:41.71681
13741	1-13741	72	56	408	50	8064	0	13	2012-08-02 11:28:42.910342	2012-08-02 11:28:42.910342
12795	1-12795	38	97	1055	15	4939	0	5	2012-08-02 11:22:28.53702	2012-08-02 11:28:43.735007
12451	1-12451	24	57	4585	26	2184	0	17	2012-08-02 11:20:14.057303	2012-08-02 11:28:44.878308
11944	1-11944	68	36	4470	26	4897	0	4	2012-08-02 11:16:57.7121	2012-08-02 11:28:44.878308
12406	1-12406	28	47	6451	34	8402	0	10	2012-08-02 11:19:56.655416	2012-08-02 11:28:47.243134
12871	1-12871	4	47	5367	8	3644	0	4	2012-08-02 11:22:58.360779	2012-08-02 11:28:47.243134
13627	1-13627	53	20	6882	34	8074	0	9	2012-08-02 11:27:56.258298	2012-08-02 11:28:47.243134
13759	1-13759	6	50	2192	32	3757	0	2	2012-08-02 11:28:49.552546	2012-08-02 11:28:49.552546
13708	1-13708	79	94	712	73	451	0	5	2012-08-02 11:28:29.384639	2012-08-02 11:28:53.156625
7740	1-7740	68	41	4466	11	1583	0	11	2012-08-02 05:34:33.535975	2012-08-02 11:28:53.156625
12404	1-12404	99	94	6820	10	6930	0	4	2012-08-02 11:19:55.848666	2012-08-02 11:28:53.156625
9186	1-9186	65	94	9705	91	4173	0	0	2012-08-02 05:44:23.797541	2012-08-02 11:28:53.156625
13768	1-13768	66	11	1466	94	6138	0	0	2012-08-02 11:28:53.156625	2012-08-02 11:28:53.156625
13526	1-13526	92	31	1848	90	9452	0	0	2012-08-02 11:27:16.568012	2012-08-02 11:28:55.123393
13773	1-13773	86	90	692	21	3696	0	8	2012-08-02 11:28:55.123393	2012-08-02 11:28:55.123393
13091	1-13091	80	18	5116	43	7046	0	0	2012-08-02 11:24:23.773565	2012-08-02 11:29:03.420622
11159	1-11159	25	27	9512	13	8843	0	6	2012-08-02 11:11:52.484001	2012-08-02 11:29:06.182748
12361	1-12361	72	40	9416	18	2564	0	12	2012-08-02 11:19:39.187014	2012-08-02 11:29:06.945209
13802	1-13802	49	18	119	40	8405	0	7	2012-08-02 11:29:06.945209	2012-08-02 11:29:06.945209
5601	1-5601	4	65	7684	88	4805	0	0	2012-08-02 05:20:01.522058	2012-08-02 11:29:11.276055
13814	1-13814	6	88	2046	65	4824	0	0	2012-08-02 11:29:11.276055	2012-08-02 11:29:11.276055
11735	1-11735	46	76	7320	85	8686	0	0	2012-08-02 11:15:36.196242	2012-08-02 11:29:12.470004
13817	1-13817	15	25	3352	48	6010	0	3	2012-08-02 11:29:12.470004	2012-08-02 11:29:12.470004
13624	1-13624	43	26	4517	30	8546	0	0	2012-08-02 11:27:55.871376	2012-08-02 11:29:15.971766
13216	1-13216	43	81	3610	93	8448	0	0	2012-08-02 11:25:13.639537	2012-08-02 11:29:15.971766
13569	1-13569	48	2	7366	34	9859	0	0	2012-08-02 11:27:34.135171	2012-08-02 11:29:15.971766
13826	1-13826	60	40	4282	1	4967	0	0	2012-08-02 11:29:15.971766	2012-08-02 11:29:15.971766
13480	1-13480	80	55	4602	72	7365	0	0	2012-08-02 11:26:58.843133	2012-08-02 11:29:17.18967
13829	1-13829	69	25	37	57	446	0	4	2012-08-02 11:29:17.18967	2012-08-02 11:29:17.18967
13018	1-13018	11	30	3645	12	2122	0	0	2012-08-02 11:23:55.852535	2012-08-02 11:29:18.371887
13766	1-13766	17	71	8634	35	8554	0	7	2012-08-02 11:28:52.382745	2012-08-02 11:29:18.371887
13842	1-13842	71	65	384	49	6952	0	12	2012-08-02 11:29:22.394404	2012-08-02 11:29:22.394404
13112	1-13112	36	79	2868	7	8025	0	13	2012-08-02 11:24:32.314102	2012-08-02 11:29:24.79138
12243	1-12243	10	65	9004	28	6204	0	0	2012-08-02 11:18:53.620239	2012-08-02 11:29:27.940484
12313	1-12313	91	51	6255	91	7172	0	14	2012-08-02 11:19:20.723102	2012-08-02 11:29:27.940484
13534	1-13534	58	53	2210	86	6676	0	2	2012-08-02 11:27:20.169976	2012-08-02 11:29:34.635924
12173	1-12173	59	96	6432	55	5189	0	0	2012-08-02 11:18:26.745679	2012-08-02 11:29:35.795893
7708	1-7708	46	96	9141	42	5498	0	0	2012-08-02 05:34:19.856586	2012-08-02 11:29:35.795893
13052	1-13052	62	70	5412	42	2839	0	7	2012-08-02 11:24:09.044165	2012-08-02 11:29:35.795893
9544	1-9544	51	96	5057	87	3951	0	4	2012-08-02 05:46:49.43067	2012-08-02 11:29:35.795893
13876	1-13876	65	11	635	96	7155	0	11	2012-08-02 11:29:35.795893	2012-08-02 11:29:35.795893
11560	1-11560	37	66	9966	63	4502	0	6	2012-08-02 11:14:28.119108	2012-08-02 11:29:39.376229
13903	1-13903	10	10	143	69	8990	0	4	2012-08-02 11:29:46.549736	2012-08-02 11:29:46.549736
13923	1-13923	54	51	2074	6	5504	0	6	2012-08-02 11:29:54.404978	2012-08-02 11:29:54.404978
13926	1-13926	69	93	3508	7	8190	0	2	2012-08-02 11:29:55.59852	2012-08-02 11:29:55.59852
13042	1-13042	34	73	5500	35	1787	0	0	2012-08-02 11:24:05.155237	2012-08-02 11:29:56.769677
12912	1-12912	70	42	3277	72	4152	0	16	2012-08-02 11:23:14.216752	2012-08-02 11:29:56.769677
13929	1-13929	44	35	1413	73	9327	0	0	2012-08-02 11:29:56.769677	2012-08-02 11:29:56.769677
7461	1-7461	88	73	7888	42	6201	325	0	2012-08-02 05:32:38.137416	2012-08-02 11:29:56.769677
13747	1-13747	27	26	1100	95	5632	0	15	2012-08-02 11:28:44.878308	2012-08-02 11:29:58.029416
13932	1-13932	56	95	4269	35	8083	0	0	2012-08-02 11:29:58.029416	2012-08-02 11:29:58.029416
13347	1-13347	94	51	8871	100	9525	0	10	2012-08-02 11:26:05.819985	2012-08-02 11:30:00.36048
13941	1-13941	70	42	1106	98	3832	0	0	2012-08-02 11:30:01.554338	2012-08-02 11:30:01.554338
12127	1-12127	53	10	5528	30	4913	0	0	2012-08-02 11:18:09.166373	2012-08-02 11:30:02.736062
9208	1-9208	25	13	6990	75	2422	0	0	2012-08-02 05:44:32.726704	2012-08-02 11:30:03.906984
13947	1-13947	47	75	1886	13	8963	0	0	2012-08-02 11:30:03.906984	2012-08-02 11:30:03.906984
13629	1-13629	16	87	5156	77	3056	0	8	2012-08-02 11:27:57.794761	2012-08-02 11:30:05.056251
13974	1-13974	3	46	1499	6	1372	0	19	2012-08-02 11:30:14.435146	2012-08-02 11:30:14.435146
13791	1-13791	69	96	1029	70	1658	0	5	2012-08-02 11:29:02.260988	2012-08-02 11:30:16.762619
13530	1-13530	11	96	6022	59	6424	0	5	2012-08-02 11:27:18.546144	2012-08-02 11:30:16.762619
13980	1-13980	24	13	2062	96	8057	0	14	2012-08-02 11:30:16.762619	2012-08-02 11:30:16.762619
6979	1-6979	88	99	7343	98	3767	0	7	2012-08-02 05:29:23.458052	2012-08-02 11:30:18.003509
13985	1-13985	65	54	977	45	3697	0	7	2012-08-02 11:30:19.190513	2012-08-02 11:30:19.190513
13403	1-13403	41	78	3043	96	4006	0	11	2012-08-02 11:26:28.176291	2012-08-02 11:30:20.7384
13061	1-13061	28	80	9643	96	9537	0	2	2012-08-02 11:24:12.546408	2012-08-02 11:30:20.7384
6553	1-6553	98	78	7866	88	7705	0	0	2012-08-02 05:26:31.03948	2012-08-02 11:30:20.7384
13990	1-13990	13	96	1357	78	8843	0	17	2012-08-02 11:30:20.7384	2012-08-02 11:30:20.7384
12893	1-12893	23	88	1025	7	753	0	2	2012-08-02 11:23:06.051897	2012-08-02 11:30:21.901036
13175	1-13175	4	12	3579	25	8311	0	0	2012-08-02 11:24:57.497844	2012-08-02 11:28:20.347033
13685	1-13685	2	40	3032	12	4283	0	11	2012-08-02 11:28:20.347033	2012-08-02 11:28:20.347033
12103	1-12103	62	60	9530	21	5900	0	0	2012-08-02 11:17:59.830645	2012-08-02 11:28:23.097742
13692	1-13692	90	21	4199	35	5434	0	0	2012-08-02 11:28:23.097742	2012-08-02 11:28:23.097742
12059	1-12059	94	17	9958	23	5475	0	6	2012-08-02 11:17:42.592182	2012-08-02 11:28:26.600446
11603	1-11603	83	7	8722	20	5240	0	0	2012-08-02 11:14:44.593441	2012-08-02 11:28:30.922001
13227	1-13227	80	58	6518	26	5255	0	0	2012-08-02 11:25:17.93958	2012-08-02 11:28:30.922001
10890	1-10890	69	51	3834	48	2040	0	0	2012-08-02 10:38:39.006261	2012-08-02 11:28:30.922001
13711	1-13711	35	71	2035	58	6628	0	0	2012-08-02 11:28:30.922001	2012-08-02 11:28:30.922001
13697	1-13697	58	14	8149	73	4556	0	0	2012-08-02 11:28:25.053385	2012-08-02 11:28:31.707131
13429	1-13429	30	35	3862	40	9932	0	9	2012-08-02 11:26:38.441881	2012-08-02 11:28:32.513694
13716	1-13716	51	40	2486	31	4307	0	2	2012-08-02 11:28:32.513694	2012-08-02 11:28:32.513694
10405	1-10405	19	27	9555	81	8143	0	4	2012-08-02 10:27:46.539469	2012-08-02 11:28:35.983455
13725	1-13725	58	81	4791	27	7627	0	5	2012-08-02 11:28:35.983455	2012-08-02 11:28:35.983455
13737	1-13737	76	92	1559	51	9750	0	5	2012-08-02 11:28:40.943643	2012-08-02 11:28:40.943643
10987	1-10987	64	66	7568	28	3524	0	11	2012-08-02 10:39:17.254011	2012-08-02 11:28:42.103973
11619	1-11619	42	79	3735	51	1570	0	0	2012-08-02 11:14:50.771572	2012-08-02 11:28:44.502577
13402	1-13402	35	79	4720	8	6399	0	10	2012-08-02 11:26:27.778526	2012-08-02 11:28:44.502577
12542	1-12542	23	79	2511	81	1528	0	2	2012-08-02 11:20:49.562594	2012-08-02 11:28:44.502577
8736	1-8736	35	54	6303	10	1350	0	3	2012-08-02 05:41:20.215277	2012-08-02 11:28:44.502577
12900	1-12900	6	100	5015	56	7169	0	0	2012-08-02 11:23:09.565198	2012-08-02 11:28:44.502577
6895	1-6895	45	7	5485	18	1484	0	0	2012-08-02 05:28:49.398224	2012-08-02 11:28:44.502577
13749	1-13749	58	55	2094	48	3246	0	8	2012-08-02 11:28:45.674267	2012-08-02 11:28:45.674267
11030	1-11030	100	44	3474	48	1388	0	0	2012-08-02 10:39:34.559249	2012-08-02 11:28:50.33668
13761	1-13761	3	32	308	44	8946	0	0	2012-08-02 11:28:50.33668	2012-08-02 11:28:50.33668
13394	1-13394	27	84	6207	43	4026	0	7	2012-08-02 11:26:25.027142	2012-08-02 11:28:59.111716
10793	1-10793	100	75	8277	43	4910	0	0	2012-08-02 10:38:00.971303	2012-08-02 11:28:59.111716
13487	1-13487	63	30	4603	20	6068	0	0	2012-08-02 11:27:01.185372	2012-08-02 11:29:02.658287
13019	1-13019	97	30	8795	75	9597	0	9	2012-08-02 11:23:56.25085	2012-08-02 11:29:02.658287
10242	1-10242	55	30	5537	80	4300	0	14	2012-08-02 10:25:14.629759	2012-08-02 11:29:02.658287
10106	1-10106	42	75	8624	20	9927	3630	0	2012-08-02 10:21:22.388892	2012-08-02 11:29:02.658287
8858	1-8858	28	91	9546	28	2619	12	0	2012-08-02 05:42:09.633285	2012-08-02 11:29:02.658287
13801	1-13801	53	32	516	24	8520	0	10	2012-08-02 11:29:05.796114	2012-08-02 11:29:05.796114
12250	1-12250	16	69	8431	81	6303	0	7	2012-08-02 11:18:56.338124	2012-08-02 11:29:12.082503
9314	1-9314	82	59	9676	81	5593	0	0	2012-08-02 05:45:15.895551	2012-08-02 11:29:15.596151
12608	1-12608	45	18	7474	8	9684	0	6	2012-08-02 11:21:15.423262	2012-08-02 11:29:15.596151
12843	1-12843	41	18	2250	54	9339	0	7	2012-08-02 11:22:47.278678	2012-08-02 11:29:15.596151
13262	1-13262	24	18	3821	89	2640	0	7	2012-08-02 11:25:31.720447	2012-08-02 11:29:15.596151
13825	1-13825	23	81	4467	18	8601	0	2	2012-08-02 11:29:15.596151	2012-08-02 11:29:15.596151
13758	1-13758	95	18	2609	92	9781	0	6	2012-08-02 11:28:49.18761	2012-08-02 11:29:19.178344
13398	1-13398	79	23	7221	100	7375	0	8	2012-08-02 11:26:26.20944	2012-08-02 11:29:19.996112
13835	1-13835	69	9	1810	10	9855	0	13	2012-08-02 11:29:19.996112	2012-08-02 11:29:19.996112
13838	1-13838	100	19	594	47	6984	0	5	2012-08-02 11:29:20.814024	2012-08-02 11:29:20.814024
10972	1-10972	45	64	6515	10	3130	0	9	2012-08-02 10:39:11.417565	2012-08-02 11:29:21.974274
9593	1-9593	27	34	8982	53	6032	0	5	2012-08-02 05:47:09.321242	2012-08-02 11:29:21.974274
13841	1-13841	48	53	1191	64	6815	0	0	2012-08-02 11:29:21.974274	2012-08-02 11:29:21.974274
13146	1-13146	63	1	6605	26	7178	0	7	2012-08-02 11:24:46.084051	2012-08-02 11:29:25.576146
10350	1-10350	33	14	3419	9	677	0	9	2012-08-02 10:27:24.669924	2012-08-02 11:29:28.31607
9754	1-9754	3	74	5415	27	4939	0	0	2012-08-02 05:48:15.387318	2012-08-02 11:29:36.591718
12621	1-12621	88	97	6698	27	7876	0	0	2012-08-02 11:21:20.539741	2012-08-02 11:29:36.591718
9454	1-9454	94	78	9125	75	5460	0	4	2012-08-02 05:46:12.902166	2012-08-02 11:29:38.989309
13885	1-13885	90	29	3137	78	9671	0	14	2012-08-02 11:29:38.989309	2012-08-02 11:29:38.989309
13746	1-13746	79	56	2141	79	7425	0	11	2012-08-02 11:28:44.502577	2012-08-02 11:29:44.549699
11810	1-11810	13	18	5753	21	7019	0	0	2012-08-02 11:16:05.67555	2012-08-02 11:29:45.743019
13901	1-13901	81	91	3156	18	8819	0	17	2012-08-02 11:29:45.743019	2012-08-02 11:29:45.743019
4483	1-4483	94	21	4319	48	76	76	8	2012-08-02 05:12:35.019221	2012-08-02 11:29:45.743019
13794	1-13794	53	31	3743	17	9605	0	4	2012-08-02 11:29:03.818897	2012-08-02 11:29:48.152049
13907	1-13907	25	87	5347	74	7416	0	0	2012-08-02 11:29:48.152049	2012-08-02 11:29:48.152049
13832	1-13832	30	10	5948	28	4911	0	6	2012-08-02 11:29:18.371887	2012-08-02 11:29:54.029273
13582	1-13582	64	38	5948	39	7849	0	7	2012-08-02 11:27:39.261479	2012-08-02 11:29:54.029273
13924	1-13924	91	16	488	83	8948	0	0	2012-08-02 11:29:55.200678	2012-08-02 11:29:55.200678
13833	1-13833	54	67	3249	18	5267	0	8	2012-08-02 11:29:19.178344	2012-08-02 11:29:55.996314
9432	1-9432	34	65	9884	29	5567	0	0	2012-08-02 05:46:03.784142	2012-08-02 11:29:59.576353
4530	1-4530	26	65	7078	62	2750	0	0	2012-08-02 05:12:53.649395	2012-08-02 11:29:59.576353
13936	1-13936	14	62	1867	65	5762	0	0	2012-08-02 11:29:59.576353	2012-08-02 11:29:59.576353
12363	1-12363	59	2	6625	15	9190	0	7	2012-08-02 11:19:39.971519	2012-08-02 11:30:06.614302
12636	1-12636	9	99	9372	47	8731	0	10	2012-08-02 11:21:26.362263	2012-08-02 11:30:08.195798
13964	1-13964	89	47	270	36	6836	0	18	2012-08-02 11:30:10.516962	2012-08-02 11:30:10.516962
13664	1-13664	68	56	2979	77	2465	0	5	2012-08-02 11:28:12.104177	2012-08-02 11:30:11.709306
12329	1-12329	71	82	4438	21	1646	0	9	2012-08-02 11:19:26.910627	2012-08-02 11:30:12.885015
13827	1-13827	73	68	1387	82	7698	0	0	2012-08-02 11:29:16.380709	2012-08-02 11:30:12.885015
214	1-214	19	82	8818	21	2527	1208	0	2012-08-02 04:43:56.256207	2012-08-02 11:30:12.885015
13163	1-13163	79	74	1855	55	1583	0	0	2012-08-02 11:24:52.77937	2012-08-02 11:30:14.048472
13973	1-13973	43	40	3296	27	5762	0	6	2012-08-02 11:30:14.048472	2012-08-02 11:30:14.048472
1385	1-1385	61	55	8860	40	7402	1541	0	2012-08-02 04:50:38.572808	2012-08-02 11:30:14.048472
13307	1-13307	21	65	9289	19	7112	0	0	2012-08-02 11:25:49.986341	2012-08-02 11:30:21.125048
13991	1-13991	33	60	1768	65	3444	0	2	2012-08-02 11:30:21.125048	2012-08-02 11:30:21.125048
13453	1-13453	68	46	2228	90	3228	0	0	2012-08-02 11:26:48.013555	2012-08-02 11:30:22.366312
13271	1-13271	81	80	8865	89	5446	0	4	2012-08-02 11:25:35.268508	2012-08-02 11:30:23.150851
12674	1-12674	63	16	7997	45	7971	0	0	2012-08-02 11:21:41.235435	2012-08-02 11:30:23.150851
13593	1-13593	80	80	6901	16	6106	0	0	2012-08-02 11:27:43.728204	2012-08-02 11:30:24.323159
13609	1-13609	33	77	6667	59	3807	0	5	2012-08-02 11:27:50.060294	2012-08-02 11:28:20.733231
11430	1-11430	34	10	5229	45	6294	0	3	2012-08-02 11:13:38.053361	2012-08-02 11:28:20.733231
11856	1-11856	19	77	2335	2	3117	0	4	2012-08-02 11:16:23.498069	2012-08-02 11:28:20.733231
12609	1-12609	68	37	5381	14	4963	0	0	2012-08-02 11:21:15.821256	2012-08-02 11:28:25.053385
13658	1-13658	70	58	3777	39	5368	0	0	2012-08-02 11:28:09.762063	2012-08-02 11:28:27.428723
13705	1-13705	38	29	930	58	7950	0	13	2012-08-02 11:28:27.428723	2012-08-02 11:28:27.428723
13579	1-13579	44	5	9926	29	7419	0	9	2012-08-02 11:27:38.068659	2012-08-02 11:28:30.535322
13712	1-13712	92	88	2649	11	8955	0	16	2012-08-02 11:28:30.535322	2012-08-02 11:28:30.535322
11438	1-11438	57	66	1561	51	842	0	7	2012-08-02 11:13:41.146774	2012-08-02 11:28:34.42507
13721	1-13721	6	51	978	66	3879	0	11	2012-08-02 11:28:34.42507	2012-08-02 11:28:34.42507
11924	1-11924	15	90	5533	29	4030	0	0	2012-08-02 11:16:49.833994	2012-08-02 11:28:35.596534
12713	1-12713	39	97	1635	74	2058	0	0	2012-08-02 11:21:56.573258	2012-08-02 11:28:35.596534
13724	1-13724	62	62	396	31	5354	0	0	2012-08-02 11:28:35.596534	2012-08-02 11:28:35.596534
13196	1-13196	6	43	8460	46	5797	0	7	2012-08-02 11:25:05.694716	2012-08-02 11:28:48.812571
13757	1-13757	40	46	1826	48	6070	0	14	2012-08-02 11:28:48.812571	2012-08-02 11:28:48.812571
13760	1-13760	3	53	4578	62	3002	0	0	2012-08-02 11:28:49.950697	2012-08-02 11:28:49.950697
12482	1-12482	47	67	5944	42	4886	0	5	2012-08-02 11:20:26.22271	2012-08-02 11:28:51.13227
13763	1-13763	48	42	1732	67	9826	0	20	2012-08-02 11:28:51.13227	2012-08-02 11:28:51.13227
3662	1-3662	49	35	5991	60	4550	0	0	2012-08-02 05:07:07.054648	2012-08-02 11:28:52.382745
8979	1-8979	72	35	1999	29	1627	0	3	2012-08-02 05:42:59.055251	2012-08-02 11:28:52.382745
10788	1-10788	12	84	5315	48	3785	0	19	2012-08-02 10:37:59.402576	2012-08-02 11:28:52.382745
13339	1-13339	9	83	4676	71	5441	0	2	2012-08-02 11:26:02.648922	2012-08-02 11:28:53.543122
13573	1-13573	6	62	2251	40	5741	0	12	2012-08-02 11:27:35.737538	2012-08-02 11:28:53.543122
12455	1-12455	38	62	3562	48	2628	0	0	2012-08-02 11:20:15.597333	2012-08-02 11:28:53.543122
13769	1-13769	89	71	4552	62	7447	0	8	2012-08-02 11:28:53.543122	2012-08-02 11:28:53.543122
13207	1-13207	22	89	3395	4	7385	0	6	2012-08-02 11:25:10.048654	2012-08-02 11:28:54.736808
13427	1-13427	73	53	3340	89	4641	0	0	2012-08-02 11:26:37.668581	2012-08-02 11:28:54.736808
12642	1-12642	64	65	5301	53	9835	0	0	2012-08-02 11:21:28.693341	2012-08-02 11:28:54.736808
11582	1-11582	47	20	3844	40	4942	0	3	2012-08-02 11:14:36.548875	2012-08-02 11:28:55.930098
12746	1-12746	38	99	3424	37	6176	0	7	2012-08-02 11:22:09.301915	2012-08-02 11:28:58.338996
12783	1-12783	94	48	4098	37	9807	0	0	2012-08-02 11:22:23.88632	2012-08-02 11:28:58.338996
13781	1-13781	97	37	2181	99	2974	0	11	2012-08-02 11:28:58.338996	2012-08-02 11:28:58.338996
13369	1-13369	2	58	8312	51	9653	0	9	2012-08-02 11:26:14.165465	2012-08-02 11:29:01.863128
13790	1-13790	61	50	506	58	5620	0	2	2012-08-02 11:29:01.863128	2012-08-02 11:29:01.863128
13799	1-13799	51	38	1634	75	7442	0	0	2012-08-02 11:29:05.409457	2012-08-02 11:29:05.409457
12968	1-12968	94	40	6300	32	4470	0	0	2012-08-02 11:23:36.32752	2012-08-02 11:29:09.320977
13809	1-13809	57	77	672	40	4167	0	9	2012-08-02 11:29:09.320977	2012-08-02 11:29:09.320977
13686	1-13686	55	95	1919	77	7954	0	0	2012-08-02 11:28:20.733231	2012-08-02 11:29:10.513766
13812	1-13812	41	6	2972	71	7104	0	5	2012-08-02 11:29:10.513766	2012-08-02 11:29:11.695894
13818	1-13818	53	68	392	29	2369	0	5	2012-08-02 11:29:12.856069	2012-08-02 11:29:12.856069
12078	1-12078	28	49	8249	38	9557	0	0	2012-08-02 11:17:50.284477	2012-08-02 11:29:15.220572
13675	1-13675	42	49	4168	74	7873	0	15	2012-08-02 11:28:16.424944	2012-08-02 11:29:15.220572
13824	1-13824	15	9	1974	49	9601	0	9	2012-08-02 11:29:15.220572	2012-08-02 11:29:15.220572
13537	1-13537	76	92	1376	89	450	0	4	2012-08-02 11:27:21.341054	2012-08-02 11:29:17.587342
13529	1-13529	36	31	3664	57	9646	0	0	2012-08-02 11:27:18.137268	2012-08-02 11:29:18.780703
13640	1-13640	32	10	3902	98	6334	0	5	2012-08-02 11:28:02.777661	2012-08-02 11:29:18.780703
10186	1-10186	10	17	9638	80	4484	0	3	2012-08-02 10:22:17.980408	2012-08-02 11:29:18.780703
13291	1-13291	78	31	6699	97	5744	0	11	2012-08-02 11:25:43.575281	2012-08-02 11:29:18.780703
13834	1-13834	82	16	5640	31	7280	0	22	2012-08-02 11:29:18.780703	2012-08-02 11:29:18.780703
10734	1-10734	59	16	7591	31	2291	1312	13	2012-08-02 10:36:34.397692	2012-08-02 11:29:18.780703
12749	1-12749	97	23	5891	81	5005	0	3	2012-08-02 11:22:10.450899	2012-08-02 11:29:20.4272
13734	1-13734	58	58	1609	38	3014	0	6	2012-08-02 11:28:39.784041	2012-08-02 11:29:20.4272
13837	1-13837	86	81	2122	23	7127	0	0	2012-08-02 11:29:20.4272	2012-08-02 11:29:20.4272
13846	1-13846	5	15	660	66	2190	0	0	2012-08-02 11:29:23.984961	2012-08-02 11:29:23.984961
10805	1-10805	69	69	7924	58	4346	0	0	2012-08-02 10:38:05.689447	2012-08-02 11:29:25.178304
13496	1-13496	14	89	5893	23	5545	0	0	2012-08-02 11:27:05.118969	2012-08-02 11:29:25.178304
9807	1-9807	41	58	9958	9	4089	0	4	2012-08-02 05:48:36.912894	2012-08-02 11:29:25.178304
13849	1-13849	15	9	1842	89	6988	0	9	2012-08-02 11:29:25.178304	2012-08-02 11:29:25.178304
10772	1-10772	61	99	7022	9	3922	0	2	2012-08-02 10:37:52.718032	2012-08-02 11:29:29.884913
13861	1-13861	53	9	1143	30	7370	0	0	2012-08-02 11:29:29.884913	2012-08-02 11:29:29.884913
12748	1-12748	5	33	2541	14	8223	0	0	2012-08-02 11:22:10.064719	2012-08-02 11:29:31.862809
13499	1-13499	22	80	6205	91	9558	0	0	2012-08-02 11:27:06.323253	2012-08-02 11:29:31.862809
13294	1-13294	35	33	6816	63	6132	0	10	2012-08-02 11:25:44.381593	2012-08-02 11:29:31.862809
6195	1-6195	59	33	8642	73	3196	0	0	2012-08-02 05:24:03.576893	2012-08-02 11:29:31.862809
13113	1-13113	46	91	9128	11	8957	0	10	2012-08-02 11:24:32.711727	2012-08-02 11:29:31.862809
11975	1-11975	4	65	3242	45	4300	0	0	2012-08-02 11:17:09.975713	2012-08-02 11:29:35.033431
13258	1-13258	69	64	6422	75	6430	0	0	2012-08-02 11:25:30.139476	2012-08-02 11:29:35.033431
11991	1-11991	46	34	5615	45	6808	0	9	2012-08-02 11:17:16.230534	2012-08-02 11:29:35.033431
13681	1-13681	39	38	1054	45	2977	0	14	2012-08-02 11:28:19.186746	2012-08-02 11:29:35.033431
13179	1-13179	86	64	8640	68	6505	0	3	2012-08-02 11:24:59.087759	2012-08-02 11:29:35.033431
13874	1-13874	60	45	3928	64	9252	0	5	2012-08-02 11:29:35.033431	2012-08-02 11:29:35.033431
12365	1-12365	45	26	8715	75	7918	0	0	2012-08-02 11:19:40.745033	2012-08-02 11:29:36.193916
13886	1-13886	65	87	688	31	2758	0	15	2012-08-02 11:29:39.762563	2012-08-02 11:29:39.762563
13879	1-13879	38	95	4846	32	4267	0	8	2012-08-02 11:29:36.989615	2012-08-02 11:29:42.160175
13893	1-13893	74	69	3964	95	7915	0	11	2012-08-02 11:29:42.160175	2012-08-02 11:29:42.160175
13660	1-13660	12	92	9490	59	8497	0	6	2012-08-02 11:28:10.568682	2012-08-02 11:29:43.786822
13792	1-13792	47	28	1743	30	5912	0	17	2012-08-02 11:29:02.658287	2012-08-02 11:29:44.95862
13272	1-13272	42	28	2976	46	3789	0	2	2012-08-02 11:25:35.666675	2012-08-02 11:29:44.95862
13899	1-13899	83	41	2942	19	3480	0	12	2012-08-02 11:29:44.95862	2012-08-02 11:29:44.95862
11176	1-11176	84	100	6910	16	7437	0	0	2012-08-02 11:11:59.135345	2012-08-02 11:29:46.141041
13622	1-13622	64	32	1096	5	2593	0	7	2012-08-02 11:27:54.733372	2012-08-02 11:29:46.141041
13340	1-13340	24	43	9964	63	8699	0	0	2012-08-02 11:26:03.069241	2012-08-02 11:29:53.255963
12776	1-12776	90	90	6943	43	4630	0	3	2012-08-02 11:22:21.168349	2012-08-02 11:29:53.255963
13921	1-13921	80	23	1416	90	9630	0	3	2012-08-02 11:29:53.255963	2012-08-02 11:29:53.255963
13934	1-13934	2	59	1083	96	2400	0	7	2012-08-02 11:29:58.802918	2012-08-02 11:29:58.802918
13219	1-13219	79	56	7480	73	3764	0	4	2012-08-02 11:25:14.812617	2012-08-02 11:30:03.509658
13946	1-13946	97	71	1255	56	1837	0	2	2012-08-02 11:30:03.509658	2012-08-02 11:30:03.509658
2542	1-2542	67	71	6076	40	3323	0	0	2012-08-02 04:59:26.747349	2012-08-02 11:30:04.669386
12852	1-12852	3	71	3263	40	1172	0	4	2012-08-02 11:22:50.803263	2012-08-02 11:30:04.669386
13949	1-13949	65	7	420	71	6772	0	8	2012-08-02 11:30:04.669386	2012-08-02 11:30:04.669386
13952	1-13952	37	40	88	58	6670	0	5	2012-08-02 11:30:05.840752	2012-08-02 11:30:05.840752
11941	1-11941	80	6	3902	77	4424	0	17	2012-08-02 11:16:56.529621	2012-08-02 11:30:09.753486
13972	1-13972	61	58	542	75	2360	0	7	2012-08-02 11:30:13.647977	2012-08-02 11:30:13.647977
13396	1-13396	72	91	1714	44	2143	0	18	2012-08-02 11:26:25.413857	2012-08-02 11:30:17.213447
13982	1-13982	1	67	335	91	5957	0	0	2012-08-02 11:30:17.213447	2012-08-02 11:30:17.213447
13878	1-13878	58	27	5017	74	7532	0	0	2012-08-02 11:29:36.591718	2012-08-02 11:30:23.924613
14018	1-14018	19	45	437	33	4381	0	0	2012-08-02 11:30:31.832093	2012-08-02 11:30:31.832093
14021	1-14021	35	33	296	58	7179	0	11	2012-08-02 11:30:32.970668	2012-08-02 11:30:32.970668
14003	1-14003	55	65	820	99	1227	0	8	2012-08-02 11:30:25.883989	2012-08-02 11:30:25.883989
13994	1-13994	7	46	3993	88	7581	0	0	2012-08-02 11:30:21.901036	2012-08-02 11:30:27.067488
13309	1-13309	21	46	3469	22	2480	0	7	2012-08-02 11:25:50.412399	2012-08-02 11:30:27.067488
14009	1-14009	44	35	100	53	4641	0	6	2012-08-02 11:30:28.283346	2012-08-02 11:30:28.283346
13123	1-13123	78	65	5264	26	5153	0	0	2012-08-02 11:24:37.451134	2012-08-02 11:30:32.230146
13542	1-13542	13	89	1933	45	5667	0	0	2012-08-02 11:27:23.488864	2012-08-02 11:30:32.230146
14019	1-14019	4	26	1223	89	5696	0	4	2012-08-02 11:30:32.230146	2012-08-02 11:30:32.230146
10568	1-10568	32	84	8442	50	6531	0	6	2012-08-02 10:35:34.335445	2012-08-02 11:30:33.358312
12960	1-12960	9	10	5421	23	5434	0	0	2012-08-02 11:23:33.189669	2012-08-02 11:30:33.358312
14022	1-14022	79	50	756	10	2994	0	0	2012-08-02 11:30:33.358312	2012-08-02 11:30:33.358312
13447	1-13447	61	100	8242	9	6886	0	7	2012-08-02 11:26:45.648723	2012-08-02 11:30:34.543923
13786	1-13786	48	1	3650	85	7901	0	5	2012-08-02 11:29:00.272127	2012-08-02 11:30:34.543923
11190	1-11190	14	17	8948	85	5949	0	6	2012-08-02 11:12:04.639891	2012-08-02 11:30:34.543923
13195	1-13195	70	34	5627	40	8920	0	6	2012-08-02 11:25:05.308106	2012-08-02 11:30:36.876255
14030	1-14030	56	11	3027	34	6991	0	0	2012-08-02 11:30:36.876255	2012-08-02 11:30:36.876255
13581	1-13581	68	5	5126	77	6342	0	0	2012-08-02 11:27:38.853085	2012-08-02 11:30:38.533307
8369	1-8369	69	76	3219	38	956	0	15	2012-08-02 05:38:52.446328	2012-08-02 11:30:39.30669
13986	1-13986	89	38	1251	26	1169	0	8	2012-08-02 11:30:18.790883	2012-08-02 11:30:39.30669
14036	1-14036	85	14	1922	4	746	0	0	2012-08-02 11:30:39.704933	2012-08-02 11:30:39.704933
11854	1-11854	57	58	6647	7	7543	0	4	2012-08-02 11:16:22.735448	2012-08-02 11:30:40.103243
14037	1-14037	93	7	5231	58	8789	0	11	2012-08-02 11:30:40.103243	2012-08-02 11:30:40.103243
10484	1-10484	38	36	7968	95	8934	0	6	2012-08-02 10:31:50.577476	2012-08-02 11:30:40.501224
13816	1-13816	29	32	4540	69	9569	0	17	2012-08-02 11:29:12.082503	2012-08-02 11:30:29.07869
13865	1-13865	80	11	2615	33	6065	0	2	2012-08-02 11:29:31.862809	2012-08-02 11:30:29.07869
\.


--
-- Data for Name: towner; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY towner (id, name, created, updated) FROM stdin;
1	w85	2012-08-02 04:43:53.386942	\N
2	w41	2012-08-02 04:43:53.420296	\N
3	w91	2012-08-02 04:43:53.453622	\N
4	w26	2012-08-02 04:43:53.46468	\N
5	w32	2012-08-02 04:43:53.475826	\N
6	w11	2012-08-02 04:43:53.486878	\N
7	w48	2012-08-02 04:43:53.498021	\N
8	w2	2012-08-02 04:43:53.520288	\N
9	w1	2012-08-02 04:43:53.531666	\N
10	w88	2012-08-02 04:43:53.542641	\N
11	w81	2012-08-02 04:43:53.553726	\N
12	w94	2012-08-02 04:43:53.564997	\N
13	w82	2012-08-02 04:43:53.576163	\N
14	w45	2012-08-02 04:43:53.587324	\N
15	w19	2012-08-02 04:43:53.609491	\N
16	w9	2012-08-02 04:43:53.620584	\N
17	w90	2012-08-02 04:43:53.631876	\N
18	w28	2012-08-02 04:43:53.643058	\N
19	w95	2012-08-02 04:43:53.654241	\N
20	w92	2012-08-02 04:43:53.665317	\N
21	w64	2012-08-02 04:43:53.676535	\N
22	w23	2012-08-02 04:43:53.687681	\N
23	w15	2012-08-02 04:43:53.698799	\N
24	w54	2012-08-02 04:43:53.710082	\N
25	w58	2012-08-02 04:43:53.721261	\N
26	w3	2012-08-02 04:43:53.732376	\N
27	w51	2012-08-02 04:43:53.766106	\N
28	w36	2012-08-02 04:43:53.77712	\N
29	w68	2012-08-02 04:43:53.788368	\N
30	w35	2012-08-02 04:43:53.799464	\N
31	w75	2012-08-02 04:43:53.832979	\N
32	w96	2012-08-02 04:43:53.844114	\N
33	w66	2012-08-02 04:43:53.866974	\N
34	w57	2012-08-02 04:43:53.911096	\N
35	w47	2012-08-02 04:43:53.922299	\N
36	w100	2012-08-02 04:43:53.933607	\N
37	w65	2012-08-02 04:43:53.966903	\N
38	w86	2012-08-02 04:43:53.978762	\N
39	w80	2012-08-02 04:43:53.989285	\N
40	w69	2012-08-02 04:43:54.011694	\N
41	w5	2012-08-02 04:43:54.022831	\N
42	w59	2012-08-02 04:43:54.045179	\N
43	w72	2012-08-02 04:43:54.056205	\N
44	w99	2012-08-02 04:43:54.067904	\N
45	w8	2012-08-02 04:43:54.089723	\N
46	w27	2012-08-02 04:43:54.180956	\N
47	w87	2012-08-02 04:43:54.191985	\N
48	w12	2012-08-02 04:43:54.203214	\N
49	w22	2012-08-02 04:43:54.214343	\N
50	w16	2012-08-02 04:43:54.225441	\N
51	w50	2012-08-02 04:43:54.24792	\N
52	w4	2012-08-02 04:43:54.270496	\N
53	w97	2012-08-02 04:43:54.314959	\N
54	w49	2012-08-02 04:43:54.371088	\N
55	w29	2012-08-02 04:43:54.415108	\N
56	w34	2012-08-02 04:43:54.482126	\N
57	w13	2012-08-02 04:43:54.493319	\N
58	w79	2012-08-02 04:43:54.515684	\N
59	w31	2012-08-02 04:43:54.560158	\N
60	w61	2012-08-02 04:43:54.649613	\N
61	w30	2012-08-02 04:43:54.67202	\N
62	w37	2012-08-02 04:43:54.738691	\N
63	w73	2012-08-02 04:43:54.749987	\N
64	w25	2012-08-02 04:43:54.783379	\N
65	w62	2012-08-02 04:43:54.805703	\N
66	w55	2012-08-02 04:43:54.816993	\N
67	w77	2012-08-02 04:43:54.828083	\N
68	w76	2012-08-02 04:43:54.861484	\N
69	w17	2012-08-02 04:43:54.88386	\N
70	w63	2012-08-02 04:43:54.89511	\N
71	w78	2012-08-02 04:43:54.906199	\N
72	w67	2012-08-02 04:43:54.928662	\N
73	w70	2012-08-02 04:43:54.939623	\N
74	w89	2012-08-02 04:43:54.962075	\N
75	w39	2012-08-02 04:43:55.0634	\N
76	w52	2012-08-02 04:43:55.095738	\N
77	w93	2012-08-02 04:43:55.184511	\N
78	w21	2012-08-02 04:43:55.21805	\N
79	w98	2012-08-02 04:43:55.273851	\N
80	w43	2012-08-02 04:43:55.309333	\N
81	w44	2012-08-02 04:43:55.34261	\N
82	w24	2012-08-02 04:43:55.376697	\N
83	w74	2012-08-02 04:43:55.432798	\N
84	w56	2012-08-02 04:43:55.465407	\N
85	w40	2012-08-02 04:43:55.487364	\N
86	w18	2012-08-02 04:43:55.654608	\N
87	w46	2012-08-02 04:43:55.743773	\N
88	w53	2012-08-02 04:43:55.85513	\N
89	w71	2012-08-02 04:43:56.244963	\N
90	w10	2012-08-02 04:43:56.278456	\N
91	w33	2012-08-02 04:43:56.333794	\N
92	w84	2012-08-02 04:43:56.814064	\N
93	w14	2012-08-02 04:43:57.113934	\N
94	w7	2012-08-02 04:43:58.666804	\N
95	w20	2012-08-02 04:43:59.834488	\N
96	w42	2012-08-02 04:44:00.344457	\N
97	w60	2012-08-02 04:44:00.510896	\N
98	w38	2012-08-02 04:44:00.721529	\N
99	w83	2012-08-02 04:44:02.01968	\N
100	w6	2012-08-02 04:44:12.32043	\N
\.


--
-- Data for Name: tquality; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tquality (id, name, idd, depository, qtt, created, updated) FROM stdin;
72	olivier/q75	\N	\N	648311	2012-08-02 04:43:54.134344	2012-08-02 11:29:58.41599
32	olivier/q74	\N	\N	649823	2012-08-02 04:43:53.676535	2012-08-02 11:30:29.07869
57	olivier/q50	\N	\N	711412	2012-08-02 04:43:53.911096	2012-08-02 11:30:29.466261
6	olivier/q76	\N	\N	794861	2012-08-02 04:43:53.453622	2012-08-02 11:30:14.435146
24	olivier/q66	\N	\N	663229	2012-08-02 04:43:53.609491	2012-08-02 11:30:29.841556
28	olivier/q90	\N	\N	706748	2012-08-02 04:43:53.643058	2012-08-02 11:29:59.200555
13	olivier/q81	\N	\N	662888	2012-08-02 04:43:53.498021	2012-08-02 11:30:30.229797
68	olivier/q35	\N	\N	739336	2012-08-02 04:43:54.100888	2012-08-02 11:30:30.616645
56	olivier/q53	\N	\N	569996	2012-08-02 04:43:53.899939	2012-08-02 11:30:15.624567
1	olivier/q52	\N	\N	729555	2012-08-02 04:43:53.386942	2012-08-02 11:29:15.971766
11	olivier/q92	\N	\N	685866	2012-08-02 04:43:53.486878	2012-08-02 11:30:31.444941
33	olivier/q11	\N	\N	626265	2012-08-02 04:43:53.687681	2012-08-02 11:30:31.832093
89	olivier/q56	\N	\N	571265	2012-08-02 04:43:54.716626	2012-08-02 11:30:32.230146
81	olivier/q63	\N	\N	698797	2012-08-02 04:43:54.448768	2012-08-02 11:27:42.556313
44	olivier/q68	\N	\N	758746	2012-08-02 04:43:53.788368	2012-08-02 11:30:33.745056
14	olivier/q67	\N	\N	788134	2012-08-02 04:43:53.520288	2012-08-02 11:29:33.863491
59	olivier/q72	\N	\N	721466	2012-08-02 04:43:53.94454	2012-08-02 11:30:02.327152
96	olivier/q96	\N	\N	712079	2012-08-02 04:43:55.106895	2012-08-02 11:30:34.154476
45	olivier/q18	\N	\N	743864	2012-08-02 04:43:53.799464	2012-08-02 11:30:19.190513
8	olivier/q82	\N	\N	642607	2012-08-02 04:43:53.46468	2012-08-02 11:30:19.975864
23	olivier/q58	\N	\N	775346	2012-08-02 04:43:53.587324	2012-08-02 11:30:34.919264
29	olivier/q100	\N	\N	630349	2012-08-02 04:43:53.654241	2012-08-02 11:30:19.577604
100	olivier/q23	\N	\N	740751	2012-08-02 04:43:55.531937	2012-08-02 11:30:34.543923
85	olivier/q70	\N	\N	767228	2012-08-02 04:43:54.549154	2012-08-02 11:30:35.713178
78	olivier/q86	\N	\N	595489	2012-08-02 04:43:54.371088	2012-08-02 11:30:20.7384
35	olivier/q4	\N	\N	624717	2012-08-02 04:43:53.721261	2012-08-02 11:30:36.103653
60	olivier/q10	\N	\N	854456	2012-08-02 04:43:53.955801	2012-08-02 11:30:36.489356
15	olivier/q83	\N	\N	818175	2012-08-02 04:43:53.520288	2012-08-02 11:30:04.293571
65	olivier/q21	\N	\N	632549	2012-08-02 04:43:54.034095	2012-08-02 11:30:21.125048
22	olivier/q30	\N	\N	683490	2012-08-02 04:43:53.587324	2012-08-02 11:29:50.870115
54	olivier/q34	\N	\N	803607	2012-08-02 04:43:53.866974	2012-08-02 11:30:21.489428
71	olivier/q7	\N	\N	838486	2012-08-02 04:43:54.134344	2012-08-02 11:30:04.669386
49	olivier/q57	\N	\N	670212	2012-08-02 04:43:53.821705	2012-08-02 11:30:05.056251
10	olivier/q97	\N	\N	729378	2012-08-02 04:43:53.486878	2012-08-02 11:30:37.29615
37	olivier/q22	\N	\N	795104	2012-08-02 04:43:53.732376	2012-08-02 11:30:37.693893
18	olivier/q24	\N	\N	748946	2012-08-02 04:43:53.542641	2012-08-02 11:30:06.238441
47	olivier/q32	\N	\N	589976	2012-08-02 04:43:53.810421	2012-08-02 11:29:20.814024
12	olivier/q55	\N	\N	662226	2012-08-02 04:43:53.498021	2012-08-02 11:29:37.796298
88	olivier/q14	\N	\N	768081	2012-08-02 04:43:54.705389	2012-08-02 11:30:21.901036
5	olivier/q62	\N	\N	784136	2012-08-02 04:43:53.453622	2012-08-02 11:30:38.533307
76	olivier/q9	\N	\N	699499	2012-08-02 04:43:54.281353	2012-08-02 11:30:39.30669
43	olivier/q17	\N	\N	698151	2012-08-02 04:43:53.77712	2012-08-02 11:30:08.195798
31	olivier/q12	\N	\N	731364	2012-08-02 04:43:53.676535	2012-08-02 11:29:39.762563
87	olivier/q3	\N	\N	895225	2012-08-02 04:43:54.638369	2012-08-02 11:29:52.450102
4	olivier/q48	\N	\N	737586	2012-08-02 04:43:53.420296	2012-08-02 11:30:39.704933
41	olivier/q36	\N	\N	666127	2012-08-02 04:43:53.754798	2012-08-02 11:28:18.004813
86	olivier/q8	\N	\N	659072	2012-08-02 04:43:54.604966	2012-08-02 11:30:24.720723
2	olivier/q26	\N	\N	764599	2012-08-02 04:43:53.386942	2012-08-02 11:30:10.140717
36	olivier/q99	\N	\N	631152	2012-08-02 04:43:53.721261	2012-08-02 11:30:40.501224
67	olivier/q79	\N	\N	612953	2012-08-02 04:43:54.045179	2012-08-02 11:30:40.910529
94	olivier/q87	\N	\N	689062	2012-08-02 04:43:54.962075	2012-08-02 11:28:53.156625
52	olivier/q42	\N	\N	658764	2012-08-02 04:43:53.855198	2012-08-02 11:29:40.978213
80	olivier/q16	\N	\N	757852	2012-08-02 04:43:54.426496	2012-08-02 11:30:24.323159
42	olivier/q20	\N	\N	768719	2012-08-02 04:43:53.766106	2012-08-02 11:29:54.029273
82	olivier/q78	\N	\N	702015	2012-08-02 04:43:54.459934	2012-08-02 11:29:41.364594
61	olivier/q98	\N	\N	683285	2012-08-02 04:43:53.966903	2012-08-02 11:29:41.751628
62	olivier/q46	\N	\N	591189	2012-08-02 04:43:53.978762	2012-08-02 11:28:53.543122
99	olivier/q84	\N	\N	660652	2012-08-02 04:43:55.52068	2012-08-02 11:30:41.308491
83	olivier/q47	\N	\N	655103	2012-08-02 04:43:54.471342	2012-08-02 11:29:55.200678
79	olivier/q65	\N	\N	702840	2012-08-02 04:43:54.40405	2012-08-02 11:29:11.695894
7	olivier/q91	\N	\N	719216	2012-08-02 04:43:53.46468	2012-08-02 11:29:55.59852
74	olivier/q41	\N	\N	598312	2012-08-02 04:43:54.225441	2012-08-02 11:29:55.996314
25	olivier/q85	\N	\N	709168	2012-08-02 04:43:53.620584	2012-08-02 11:27:53.19788
77	olivier/q60	\N	\N	704706	2012-08-02 04:43:54.359409	2012-08-02 11:28:20.733231
39	olivier/q43	\N	\N	638242	2012-08-02 04:43:53.743617	2012-08-02 11:29:57.167606
3	olivier/q59	\N	\N	634115	2012-08-02 04:43:53.420296	2012-08-02 11:29:27.565646
46	olivier/q45	\N	\N	739607	2012-08-02 04:43:53.799464	2012-08-02 11:30:27.067488
95	olivier/q6	\N	\N	748354	2012-08-02 04:43:55.0634	2012-08-02 11:29:42.160175
9	olivier/q69	\N	\N	732528	2012-08-02 04:43:53.475826	2012-08-02 11:30:27.497934
75	olivier/q15	\N	\N	663252	2012-08-02 04:43:54.225441	2012-08-02 11:30:13.647977
50	olivier/q29	\N	\N	739296	2012-08-02 04:43:53.832979	2012-08-02 11:28:42.910342
90	olivier/q44	\N	\N	711543	2012-08-02 04:43:54.772413	2012-08-02 11:30:27.884978
27	olivier/q40	\N	\N	652402	2012-08-02 04:43:53.631876	2012-08-02 11:30:23.924613
48	olivier/q2	\N	\N	714339	2012-08-02 04:43:53.821705	2012-08-02 11:30:25.110376
20	olivier/q71	\N	\N	690621	2012-08-02 04:43:53.564997	2012-08-02 11:30:25.497028
53	olivier/q39	\N	\N	697202	2012-08-02 04:43:53.855198	2012-08-02 11:30:28.283346
97	olivier/q89	\N	\N	714215	2012-08-02 04:43:55.184511	2012-08-02 11:30:31.025132
55	olivier/q5	\N	\N	688877	2012-08-02 04:43:53.888807	2012-08-02 11:30:32.594021
34	olivier/q64	\N	\N	708029	2012-08-02 04:43:53.710082	2012-08-02 11:30:38.920121
58	olivier/q73	\N	\N	750051	2012-08-02 04:43:53.933607	2012-08-02 11:30:40.103243
51	olivier/q95	\N	\N	673477	2012-08-02 04:43:53.844114	2012-08-02 11:28:59.50984
70	olivier/q38	\N	\N	792368	2012-08-02 04:43:54.123479	2012-08-02 11:29:25.951658
38	olivier/q13	\N	\N	725733	2012-08-02 04:43:53.732376	2012-08-02 11:29:26.73639
92	olivier/q1	\N	\N	761363	2012-08-02 04:43:54.794615	2012-08-02 11:29:30.658243
19	olivier/q51	\N	\N	641651	2012-08-02 04:43:53.553726	2012-08-02 11:29:44.95862
69	olivier/q31	\N	\N	803213	2012-08-02 04:43:54.111957	2012-08-02 11:29:46.549736
66	olivier/q54	\N	\N	615664	2012-08-02 04:43:54.034095	2012-08-02 11:29:49.720915
63	olivier/q28	\N	\N	674726	2012-08-02 04:43:53.978762	2012-08-02 11:29:54.802891
73	olivier/q37	\N	\N	631582	2012-08-02 04:43:54.16982	2012-08-02 11:29:56.769677
16	olivier/q33	\N	\N	804977	2012-08-02 04:43:53.531666	2012-08-02 11:29:59.973878
98	olivier/q27	\N	\N	729932	2012-08-02 04:43:55.364611	2012-08-02 11:30:01.554338
17	olivier/q25	\N	\N	727828	2012-08-02 04:43:53.531666	2012-08-02 11:30:05.442698
93	olivier/q80	\N	\N	740067	2012-08-02 04:43:54.861484	2012-08-02 11:30:08.958117
40	olivier/q19	\N	\N	882873	2012-08-02 04:43:53.754798	2012-08-02 11:30:11.311415
21	olivier/q61	\N	\N	742337	2012-08-02 04:43:53.576163	2012-08-02 11:30:12.084664
30	olivier/q49	\N	\N	663149	2012-08-02 04:43:53.665317	2012-08-02 11:30:16.000275
84	olivier/q88	\N	\N	673529	2012-08-02 04:43:54.493319	2012-08-02 11:30:17.613312
91	olivier/q77	\N	\N	733276	2012-08-02 04:43:54.772413	2012-08-02 11:30:17.213447
26	olivier/q93	\N	\N	593232	2012-08-02 04:43:53.620584	2012-08-02 11:30:18.790883
64	olivier/q94	\N	\N	625824	2012-08-02 04:43:54.011694	2012-08-02 11:30:22.763602
\.


--
-- Data for Name: tquote; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tquote (id, own, nr, qtt_requ, np, qtt_prov, qtt_in, qtt_out, flows, created, removed) FROM stdin;
\.


--
-- Data for Name: tquoteremoved; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tquoteremoved (id, own, nr, qtt_requ, np, qtt_prov, qtt_in, qtt_out, flows, created, removed) FROM stdin;
\.


--
-- Data for Name: treltried; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY treltried (np, nr, cnt) FROM stdin;
24	100	7
87	16	15
67	16	13
13	19	5
37	38	21
3	4	2
31	42	5
92	64	3
43	11	11
71	72	2
11	17	14
4	5	11
58	35	3
67	81	10
52	53	7
17	32	8
90	28	9
44	4	7
56	45	6
40	8	7
29	14	13
77	41	5
32	41	12
60	76	12
19	40	2
90	91	7
27	100	15
15	12	12
87	59	4
81	77	15
3	28	17
15	77	11
11	23	4
53	72	7
23	1	9
81	5	16
24	56	9
14	15	13
55	92	9
87	88	4
32	82	8
13	55	11
31	60	11
64	55	8
5	6	3
54	80	14
16	11	18
90	87	9
41	53	5
87	73	2
40	41	4
8	28	21
32	97	12
60	37	15
20	98	9
82	53	17
97	53	17
13	45	11
3	21	17
76	71	5
67	54	3
22	12	8
76	3	5
11	38	6
1	88	3
28	11	9
45	46	14
30	5	9
74	34	6
14	20	6
52	74	3
46	76	12
65	66	8
25	44	14
60	52	10
52	13	11
44	65	5
5	31	10
87	81	10
79	3	3
12	13	4
19	16	6
80	44	6
77	52	2
41	59	7
22	23	11
36	73	8
54	15	8
59	37	2
36	58	5
28	80	12
15	20	6
20	27	8
6	25	10
73	49	20
11	35	10
48	32	6
98	66	16
41	86	5
42	6	12
24	43	6
62	41	8
20	5	8
53	13	8
72	24	12
35	36	7
12	20	20
23	82	25
4	22	3
16	17	12
68	31	7
38	87	28
32	60	16
21	65	16
38	79	9
19	68	12
97	30	12
17	79	3
55	14	8
66	41	8
59	24	17
100	65	4
23	19	11
88	77	18
55	3	9
49	8	6
37	20	8
72	29	9
71	41	4
41	90	10
34	22	6
31	32	6
86	46	7
13	5	10
55	48	16
26	77	11
65	9	9
26	96	8
84	35	6
10	18	14
65	12	5
51	25	5
20	28	4
62	63	4
8	75	21
18	87	10
57	6	19
79	65	16
89	88	11
96	6	11
45	50	7
53	37	7
47	19	6
43	9	9
27	61	9
78	47	17
4	24	12
70	5	5
28	54	11
40	4	10
3	76	8
46	22	12
6	52	8
9	30	4
30	46	7
12	70	2
4	87	10
10	13	7
53	34	10
9	32	6
40	56	10
24	75	7
20	95	5
47	30	7
33	61	8
17	43	11
50	17	9
6	17	10
9	67	16
1	2	24
45	40	16
25	67	11
14	63	13
33	4	14
4	9	18
80	29	13
38	19	4
85	90	3
24	5	4
48	10	10
97	33	4
86	15	20
37	85	20
78	10	8
21	34	13
25	15	20
25	26	19
95	40	2
55	75	2
74	58	16
46	37	14
68	84	12
68	64	7
82	9	3
4	52	6
77	75	22
38	57	12
91	79	7
91	86	9
72	89	15
34	90	18
91	74	5
43	55	13
87	57	4
22	4	18
100	44	11
34	26	11
70	38	17
75	48	9
29	89	7
1	22	12
5	91	10
59	51	6
36	50	7
84	83	11
10	11	11
17	51	8
85	29	22
24	58	7
58	85	9
83	33	11
2	80	18
67	56	14
23	36	9
5	29	13
30	56	13
63	91	21
53	55	11
32	1	12
14	90	11
9	8	13
48	91	14
8	61	6
68	61	19
34	83	9
48	49	18
51	30	7
55	61	20
51	68	13
44	1	6
40	67	13
61	70	15
40	34	9
84	3	12
52	44	14
49	70	12
82	38	4
67	62	11
30	72	9
91	2	4
12	40	13
30	17	5
1	78	9
17	50	10
88	29	4
18	53	3
70	24	13
24	61	18
68	40	20
61	76	12
26	80	9
71	23	19
94	68	18
56	23	13
43	46	20
20	3	10
95	69	13
53	73	11
82	50	5
87	49	10
11	52	16
95	34	2
51	37	4
99	21	24
92	32	16
9	89	8
32	69	15
7	22	6
46	90	12
41	39	6
66	30	9
45	61	10
33	8	10
55	56	4
57	100	6
39	83	17
82	91	13
21	82	11
9	94	3
46	68	10
49	5	15
50	23	17
34	67	22
81	7	2
44	20	11
95	36	12
33	39	6
50	46	8
51	55	17
17	15	7
75	91	11
56	2	3
88	1	12
73	69	13
60	7	6
68	72	14
29	92	10
68	19	12
16	34	13
26	73	9
82	95	12
25	28	10
53	26	9
31	97	10
32	34	16
51	41	7
81	70	11
52	35	7
51	60	6
90	34	9
62	90	18
19	1	6
69	51	6
38	51	9
5	26	20
39	56	7
53	33	5
37	57	5
56	99	12
34	18	13
54	31	15
91	84	4
32	74	2
57	84	14
29	39	22
48	55	9
95	7	15
38	82	15
24	59	11
28	86	12
89	37	13
61	71	13
19	82	7
80	28	8
3	60	14
29	71	8
44	62	21
19	44	18
48	84	21
86	66	11
80	30	19
43	29	8
5	57	4
26	51	8
72	36	7
70	42	4
15	48	14
7	41	7
53	80	5
71	45	8
61	53	4
30	78	12
55	53	10
69	100	12
35	89	10
20	49	13
44	15	13
35	82	11
65	43	13
39	85	22
3	52	6
43	67	11
77	83	3
18	97	3
21	85	10
36	1	10
11	30	32
30	2	6
59	26	16
34	46	9
23	20	4
78	3	8
59	38	14
79	86	13
37	1	12
84	11	28
31	79	5
85	92	9
89	43	10
83	28	4
2	33	8
91	82	3
13	32	8
1	52	11
41	73	13
32	39	10
12	97	9
3	99	12
3	13	8
50	90	14
3	44	9
86	26	15
7	64	14
100	27	7
59	93	8
2	59	14
72	5	8
31	47	3
93	35	9
78	29	16
47	2	10
99	64	6
15	69	8
1	64	12
1	46	23
92	70	2
68	35	7
34	81	7
63	8	7
93	94	13
16	7	14
82	74	3
54	3	8
41	7	16
75	93	4
92	79	9
40	35	11
54	48	9
7	8	16
48	46	16
34	3	15
24	32	11
69	70	9
15	54	4
46	3	9
94	53	16
2	47	4
8	24	5
11	8	14
66	76	5
12	15	8
69	6	9
6	39	11
75	92	6
94	42	15
80	60	14
41	64	16
93	2	3
94	36	16
55	11	20
90	43	12
69	41	2
46	59	2
7	12	3
10	38	5
4	2	9
59	17	11
41	58	5
15	31	10
4	73	4
14	30	8
60	26	3
12	42	2
52	49	10
2	66	11
74	4	15
22	98	12
87	94	14
84	88	3
50	2	10
29	22	6
42	12	4
27	1	29
22	50	7
51	7	10
19	28	12
97	94	9
58	20	7
22	73	9
79	39	10
10	90	6
65	13	16
17	57	2
70	88	13
59	14	12
39	84	10
4	10	11
88	79	17
24	92	6
20	65	7
7	34	7
28	29	13
98	16	9
30	74	5
17	44	7
68	95	6
35	19	8
66	87	17
22	16	2
28	4	12
83	54	6
17	26	6
66	55	6
95	22	7
57	54	4
76	70	10
31	16	27
29	99	9
84	98	11
72	8	16
31	95	6
53	41	14
29	24	5
15	79	16
64	37	14
58	98	11
88	24	12
26	1	13
24	45	5
4	31	6
47	57	12
79	13	7
18	88	6
83	15	11
13	4	6
78	93	3
13	2	7
6	44	25
29	25	6
37	93	8
13	99	8
46	88	4
42	99	7
78	87	17
23	14	7
69	21	6
73	41	2
3	16	7
46	54	9
42	26	5
26	57	22
38	62	12
22	21	10
75	90	8
30	62	7
40	62	14
54	4	13
90	74	15
97	50	5
71	19	8
89	96	13
5	84	15
25	72	8
43	45	6
15	27	6
3	89	2
93	50	3
4	50	20
92	98	8
49	32	14
93	12	13
69	71	7
21	78	2
52	63	10
53	52	21
60	27	15
87	43	5
36	96	9
32	93	5
90	41	9
94	15	13
88	98	9
14	78	9
17	39	4
20	61	15
54	74	13
56	77	2
57	7	7
96	54	12
55	54	15
57	96	6
90	11	11
34	79	7
13	97	12
67	83	5
85	52	8
26	19	35
7	36	11
3	33	8
77	14	5
99	43	4
26	13	17
33	7	6
61	75	28
19	29	15
16	6	14
51	22	3
24	40	8
27	57	10
86	63	11
1	60	14
8	82	19
43	94	10
55	84	3
71	54	15
74	75	8
80	85	12
61	23	7
33	97	21
29	33	6
49	61	6
10	27	5
79	49	12
39	19	18
17	68	5
98	26	19
40	58	9
76	60	10
71	67	16
77	44	16
27	6	7
77	56	6
54	10	14
16	93	12
42	27	5
52	26	14
14	42	2
83	75	16
76	46	11
10	48	16
98	67	7
50	58	11
85	89	7
60	71	10
72	37	5
74	55	8
23	100	11
22	88	6
93	37	5
39	54	2
84	24	20
27	76	12
62	10	19
43	60	7
83	85	6
56	48	6
93	42	11
62	28	13
90	84	8
59	66	11
28	55	18
39	57	15
93	29	7
62	59	9
100	56	23
13	77	11
75	99	7
46	30	8
28	59	10
87	27	14
95	66	11
57	32	5
6	27	9
94	12	12
49	6	16
6	47	26
73	5	5
19	41	15
45	39	18
42	88	3
68	51	11
100	94	7
91	73	14
44	6	19
12	5	4
57	55	7
69	63	8
6	99	19
89	33	10
100	95	14
42	89	17
79	66	16
86	30	12
71	70	6
98	92	11
75	58	8
42	39	23
33	95	8
20	10	14
81	38	8
79	71	18
99	13	18
53	7	7
79	88	14
80	41	10
61	44	5
90	55	8
67	25	2
33	22	18
85	20	11
41	95	8
66	51	15
43	65	8
7	79	18
94	49	10
85	91	12
21	42	8
91	75	12
90	44	5
81	32	16
58	47	6
84	44	10
47	58	15
56	62	11
66	94	10
53	81	3
97	19	12
57	62	4
85	12	9
95	99	7
7	98	11
25	40	8
44	38	12
22	25	30
19	5	3
58	25	3
89	5	4
81	27	10
24	71	12
65	86	6
15	40	14
47	12	10
5	9	6
88	80	11
66	25	5
63	68	15
44	45	3
81	98	6
47	94	6
39	4	2
28	63	11
41	13	3
46	83	15
67	85	17
6	33	12
81	58	9
4	38	5
5	65	5
32	15	9
45	51	8
11	37	7
39	41	17
12	74	13
46	11	6
62	46	17
80	25	5
81	33	6
98	89	13
41	34	9
12	80	2
20	90	9
13	78	8
48	79	7
5	53	4
62	45	12
99	81	11
71	90	10
71	4	8
66	39	9
33	94	7
32	26	4
9	33	11
55	13	8
40	100	16
60	86	20
29	82	18
47	35	10
68	88	14
41	19	9
64	44	8
33	63	7
11	26	13
42	69	14
22	34	24
69	90	10
25	50	14
88	31	15
68	5	13
82	55	24
30	42	10
63	57	5
19	84	3
54	42	9
81	6	7
93	7	7
84	90	2
89	25	4
36	74	2
82	97	2
92	8	14
15	3	3
85	18	6
29	16	12
14	91	7
44	48	16
21	31	3
84	94	12
20	81	3
58	38	18
94	5	19
18	25	3
72	53	10
1	97	9
37	31	12
72	60	18
77	17	16
72	81	4
52	77	6
73	72	12
38	80	9
23	56	11
89	61	8
36	56	11
10	97	3
70	43	19
42	1	19
80	8	11
27	89	7
50	92	17
65	64	11
48	45	11
20	79	8
71	86	5
26	44	5
58	36	11
42	93	6
86	95	10
5	50	5
37	28	18
73	71	2
50	63	15
21	40	5
44	37	17
32	66	7
21	93	8
45	8	8
42	64	10
80	42	11
74	21	12
53	66	10
8	5	8
14	47	5
58	83	7
97	72	4
97	80	10
63	84	8
89	44	22
17	89	15
12	52	22
59	88	4
12	95	10
3	84	11
14	41	12
66	59	2
41	49	12
44	91	20
92	45	9
14	22	6
58	91	9
66	100	7
42	86	8
5	59	20
5	73	11
75	74	9
60	47	8
12	8	15
74	49	19
13	49	11
8	38	3
33	3	7
85	35	6
48	24	9
7	33	7
43	71	12
3	14	9
53	44	10
25	58	16
60	19	10
75	77	15
48	60	12
88	50	8
16	99	7
37	74	7
69	68	13
15	57	4
74	64	4
14	4	4
22	42	9
93	14	4
17	92	8
19	58	11
25	9	10
56	88	8
24	6	18
47	52	10
90	65	15
6	76	6
30	80	2
17	62	8
17	22	8
69	10	5
46	2	7
54	8	21
47	78	4
49	99	10
94	69	18
99	97	8
83	76	3
22	37	10
56	1	14
64	3	22
92	41	17
22	65	17
76	32	5
66	24	12
87	21	6
28	22	5
49	14	8
74	77	12
50	20	13
91	45	16
53	2	5
43	34	2
70	18	14
51	63	6
19	30	10
34	36	12
8	39	3
20	66	8
92	63	11
14	72	13
30	4	12
71	92	4
57	10	7
27	18	23
42	95	2
24	8	26
82	80	3
65	85	17
91	59	13
100	64	8
47	15	14
18	31	8
45	20	6
46	36	2
11	78	7
15	45	5
15	52	2
26	49	7
4	1	14
92	61	9
96	56	18
99	22	7
67	100	9
25	99	5
62	15	4
73	88	8
64	36	6
65	49	10
99	3	8
8	72	8
4	20	12
10	16	9
24	98	20
28	94	6
92	49	5
29	95	6
94	64	10
94	17	10
59	100	10
38	63	5
97	60	13
47	99	13
4	89	14
55	25	13
6	61	4
99	9	5
95	11	9
75	82	3
47	1	7
26	45	5
22	7	4
54	53	6
74	22	6
31	37	7
5	96	16
60	67	19
10	93	8
21	66	10
19	21	10
16	30	10
60	49	9
40	12	8
21	43	17
97	29	17
63	41	4
63	99	10
49	7	17
29	34	7
63	15	21
11	19	14
34	57	11
97	93	13
50	34	14
49	67	7
76	90	8
73	65	13
99	70	8
18	20	17
47	27	5
63	40	6
12	82	13
35	22	9
20	94	9
72	43	8
52	73	4
17	24	4
82	16	16
59	30	11
83	89	12
21	30	14
20	47	15
1	99	5
70	21	3
95	52	4
12	25	8
30	95	9
34	48	6
88	95	6
61	50	11
85	36	10
61	20	20
25	62	6
64	78	3
2	54	3
16	70	7
36	32	19
9	76	5
5	22	10
51	21	7
39	12	3
89	11	12
100	92	9
77	68	10
44	41	8
65	67	12
84	30	16
66	65	2
18	4	9
9	49	3
34	71	6
15	22	12
10	85	4
56	54	9
49	92	12
22	44	3
65	61	7
82	67	12
80	93	8
86	87	4
29	67	8
93	28	10
53	63	8
68	90	8
75	23	6
16	79	10
40	32	13
28	78	21
67	84	10
52	84	11
100	8	6
46	70	10
94	34	4
90	69	3
98	58	9
28	46	24
64	54	6
24	86	12
59	75	10
21	35	8
6	94	6
46	56	7
97	91	7
81	57	16
28	45	4
42	98	6
21	50	14
19	23	23
1	55	13
25	4	10
64	96	14
2	75	17
67	19	4
89	23	16
43	2	10
94	25	6
2	60	13
87	31	16
45	81	11
68	46	6
87	100	10
56	98	5
86	21	3
2	48	2
9	85	8
87	32	10
50	15	11
67	86	10
31	52	13
65	44	8
79	80	3
27	47	10
69	92	17
73	2	8
21	95	5
57	13	3
35	40	13
33	64	11
83	94	23
58	11	18
66	33	12
79	54	12
4	55	14
96	20	10
86	100	18
36	81	11
52	57	5
81	68	15
13	52	8
40	73	7
59	3	19
3	79	5
76	77	9
89	55	7
92	42	8
55	97	7
68	81	8
23	92	10
18	75	17
20	11	17
46	72	6
85	7	17
45	83	8
76	33	4
82	87	5
47	98	13
13	24	7
53	99	11
67	2	7
15	13	8
23	63	23
95	2	6
98	34	15
44	11	8
31	88	5
73	3	6
1	72	14
36	27	3
57	93	8
13	9	9
31	82	11
72	93	4
98	15	14
26	36	8
56	78	12
45	15	4
24	77	5
49	96	24
22	56	11
60	6	9
46	52	9
52	41	10
92	87	11
72	100	5
32	22	11
93	90	11
84	27	9
38	67	15
15	18	9
52	91	12
82	54	7
15	86	6
85	93	7
15	9	13
77	59	9
42	92	8
48	21	19
68	12	9
7	61	11
86	59	3
20	69	2
57	94	7
63	96	8
96	1	10
36	31	7
67	42	22
69	25	9
34	47	15
84	45	5
25	84	4
26	90	8
94	71	8
8	79	14
26	91	17
82	96	15
19	72	27
78	38	7
11	61	12
5	63	7
18	23	10
21	5	4
70	72	12
45	23	10
13	68	18
96	13	17
17	1	8
55	73	7
14	96	11
13	86	7
31	65	10
61	38	3
42	80	8
80	14	8
93	41	6
9	48	11
85	17	10
36	77	9
73	100	5
15	28	16
70	71	6
7	53	10
13	64	6
69	58	13
73	29	6
62	32	7
57	71	12
72	58	7
100	20	22
14	69	10
65	40	11
81	94	4
45	35	7
53	70	3
52	71	11
69	57	11
7	74	6
90	59	18
8	23	20
31	69	7
74	82	7
71	73	16
57	65	12
97	8	7
79	26	14
71	52	6
21	73	7
97	6	12
58	7	13
95	70	5
98	52	15
51	33	4
68	97	10
49	26	8
63	45	21
51	5	10
14	87	10
79	53	3
39	96	11
32	52	11
87	51	10
54	32	8
78	61	9
25	94	4
74	81	5
90	21	10
83	68	26
54	35	3
49	51	8
29	83	7
94	45	15
100	74	8
70	66	8
66	70	14
52	32	5
32	71	6
91	88	10
91	64	10
32	67	8
68	10	19
13	25	8
17	33	6
93	21	16
77	12	14
60	32	4
41	21	16
2	27	4
94	73	5
97	67	11
35	16	17
71	31	10
29	93	4
72	94	3
43	80	17
50	65	9
14	75	10
81	72	6
31	17	4
52	37	6
84	70	9
15	83	9
50	9	9
38	10	12
2	77	6
5	14	7
42	53	3
86	42	16
2	62	5
1	69	12
94	82	13
8	12	8
20	100	8
93	33	7
76	73	11
19	14	14
52	11	10
33	46	17
94	20	9
69	22	9
52	62	7
16	78	9
23	34	19
70	36	7
53	69	11
54	9	5
21	4	4
67	21	8
2	18	11
40	25	4
40	77	11
59	33	19
60	17	7
50	3	3
16	5	4
18	17	11
96	94	15
77	86	10
90	12	10
55	57	10
52	70	10
72	12	12
17	97	7
1	11	11
63	32	4
35	53	3
39	44	4
70	28	9
100	40	14
24	13	2
10	67	7
67	5	3
8	87	19
44	61	11
3	31	4
44	56	17
86	8	8
55	9	3
35	45	10
32	98	13
30	44	10
42	70	10
42	100	8
94	84	13
48	52	13
17	4	9
11	73	4
66	35	20
48	93	11
33	70	6
6	32	10
27	86	6
15	95	12
10	9	15
10	79	13
41	23	6
47	60	10
65	59	4
42	44	11
25	14	10
47	55	4
87	2	6
58	86	10
53	35	7
73	13	4
5	10	4
43	16	13
9	97	5
23	18	2
91	85	6
71	61	11
57	47	13
10	1	16
42	85	5
37	70	11
3	11	9
31	30	9
18	41	5
89	34	9
69	1	11
4	57	5
23	13	8
85	42	3
32	29	5
44	36	12
29	87	13
60	21	7
89	32	6
10	63	11
20	33	8
45	11	16
49	93	6
16	54	9
83	72	5
44	93	7
13	50	6
3	15	4
40	57	7
47	71	15
99	48	10
4	18	8
1	13	18
95	93	17
9	81	13
60	28	7
32	68	18
2	13	6
4	28	11
59	18	2
73	17	5
59	11	10
7	35	11
66	43	5
57	83	10
93	99	6
85	69	8
40	44	2
90	52	12
41	29	11
8	68	8
48	77	12
35	59	7
12	55	8
86	32	9
39	60	15
27	48	12
14	54	5
45	31	17
36	66	5
93	86	8
81	47	13
24	73	7
16	43	13
22	49	7
18	65	4
34	89	9
61	93	8
70	40	5
30	91	7
61	91	15
2	50	3
91	39	10
44	22	7
87	50	7
96	21	9
13	91	7
78	7	13
50	68	4
8	52	2
55	49	8
74	35	11
93	56	11
28	32	13
30	67	6
55	76	13
48	31	11
76	98	4
35	31	7
65	17	8
78	22	6
74	16	5
25	48	11
24	7	3
3	18	4
95	56	6
48	64	8
43	96	4
62	3	4
19	81	4
13	14	14
83	23	8
87	58	5
76	25	19
68	53	3
35	50	8
17	82	9
10	77	10
86	67	13
34	15	12
9	1	7
29	58	5
7	10	7
93	32	11
31	93	11
68	24	18
92	71	6
80	50	11
58	60	9
54	59	7
5	90	2
11	70	10
7	23	9
38	21	11
19	62	5
63	67	6
64	34	16
95	45	24
45	6	7
80	55	7
67	88	9
67	38	22
64	19	7
21	53	13
57	67	12
13	89	6
8	36	2
61	85	10
48	38	10
31	25	16
57	4	5
69	43	4
73	57	10
14	16	8
76	58	9
11	80	15
49	97	8
61	90	10
46	98	16
49	73	5
99	40	8
67	34	8
98	21	3
95	78	9
57	97	10
38	15	6
89	57	4
80	99	3
10	25	5
44	81	13
77	63	18
96	7	6
66	49	11
23	87	5
92	40	9
19	69	9
16	74	2
33	44	11
80	54	4
82	48	7
53	76	7
91	28	10
69	4	5
1	23	5
99	37	14
73	45	7
98	76	4
100	68	5
87	8	2
30	45	8
34	69	7
60	44	16
41	69	14
23	93	8
87	85	3
56	96	12
9	64	4
51	98	4
32	4	15
23	41	7
98	11	9
9	16	5
45	38	16
32	86	10
45	55	12
77	50	12
59	63	5
91	35	12
87	82	8
16	62	8
74	47	4
15	84	8
32	70	8
52	9	7
87	69	5
71	56	3
48	66	6
91	92	12
69	67	2
2	28	8
23	27	11
96	23	6
18	48	5
70	65	5
91	48	12
40	48	12
7	48	12
75	60	10
89	70	2
85	63	6
69	39	9
79	34	20
37	84	5
91	50	3
12	43	7
81	3	16
84	81	15
28	75	10
44	17	2
32	17	3
67	31	10
88	47	5
79	4	4
70	98	4
3	81	3
65	15	11
32	14	4
44	84	9
32	99	8
3	26	8
55	22	3
99	33	8
22	87	2
44	60	14
26	60	11
95	96	4
61	26	3
66	61	6
1	38	8
87	36	10
72	56	3
15	96	6
98	19	24
52	15	18
58	97	9
11	33	13
93	79	13
42	74	4
46	63	8
76	40	4
63	18	9
50	43	10
70	89	4
97	73	7
59	72	18
81	99	3
35	38	8
91	70	11
88	55	10
66	92	11
63	95	16
98	64	11
96	9	5
24	29	6
37	89	6
25	56	11
85	74	8
82	46	4
60	25	4
82	28	6
63	78	13
85	88	2
57	59	14
40	14	10
82	70	9
19	27	10
17	40	4
57	24	7
88	66	17
75	96	10
28	66	14
22	97	9
84	48	2
56	57	16
87	93	6
93	66	8
72	95	6
46	19	2
54	67	11
81	15	3
43	22	14
27	46	7
79	99	10
1	35	20
5	33	5
22	19	16
35	79	8
20	34	17
28	49	4
37	19	6
78	88	8
95	85	7
36	12	7
42	17	11
21	57	3
23	50	6
22	82	11
25	31	8
4	65	11
85	9	6
74	2	10
4	79	11
95	88	21
83	31	12
8	78	8
31	50	5
77	94	9
46	49	10
16	18	15
1	34	9
85	23	6
16	37	8
33	42	12
18	79	6
38	41	4
61	1	17
25	17	8
18	76	12
15	56	11
82	43	7
49	39	2
45	28	9
15	42	5
6	12	8
96	34	9
75	59	9
49	56	9
2	88	8
55	85	12
15	6	5
38	96	6
34	87	5
22	1	9
33	80	4
57	28	17
3	94	8
9	72	13
47	63	6
89	90	3
95	18	4
69	74	15
63	80	14
72	85	5
61	98	3
82	4	13
21	99	13
9	45	5
30	82	7
45	21	6
71	27	13
85	28	7
64	40	18
8	19	11
88	22	6
91	89	6
60	82	9
3	23	3
83	24	11
72	50	7
17	47	3
68	8	15
48	9	7
97	81	4
65	58	16
29	79	11
62	71	12
19	94	9
57	48	10
18	47	11
87	77	12
61	28	5
64	26	14
88	92	8
44	33	2
38	20	3
46	82	9
65	55	4
64	31	13
19	74	7
15	91	10
82	18	11
37	64	7
28	72	13
66	83	12
21	80	9
55	52	8
68	45	7
20	2	19
62	6	9
52	6	4
31	26	11
31	68	16
57	90	13
18	21	8
2	78	12
98	88	6
18	94	14
18	35	7
16	22	11
11	41	20
74	54	8
52	54	7
80	22	10
98	38	6
18	24	11
64	87	10
71	9	20
38	65	10
37	51	8
71	7	11
6	4	4
14	57	7
7	9	14
100	38	8
72	38	11
4	46	3
35	65	10
46	91	4
89	82	14
62	37	3
99	18	11
45	64	7
54	57	11
82	66	3
81	2	13
75	66	3
37	15	10
12	22	9
33	15	10
92	100	6
58	18	15
1	89	10
28	71	4
97	77	4
8	14	3
54	87	5
33	79	16
75	16	4
89	66	4
17	54	17
63	54	7
66	89	5
23	43	11
39	78	10
78	46	11
30	18	17
95	4	18
6	64	6
72	33	7
55	24	5
27	40	8
15	7	7
73	82	4
55	78	4
38	76	16
52	46	7
24	30	20
72	87	4
85	11	10
46	80	16
24	3	8
26	32	7
4	40	9
3	8	10
43	3	13
6	85	6
78	85	13
65	39	11
41	72	6
76	67	5
78	35	11
61	27	16
16	24	6
92	11	8
96	39	13
51	94	6
29	60	22
44	97	14
14	13	4
12	89	13
77	29	3
73	20	11
44	89	2
68	62	20
63	1	10
8	91	11
73	44	6
55	1	11
17	70	5
86	84	16
45	80	15
13	51	3
1	15	6
29	51	9
97	28	10
66	16	9
38	74	15
58	12	4
18	93	5
46	21	7
33	47	18
81	62	10
27	88	2
53	46	6
32	87	3
47	85	16
26	3	5
50	84	12
25	95	14
47	86	11
87	91	11
30	29	11
15	29	6
98	59	10
16	55	17
60	81	8
51	66	8
1	59	18
25	20	2
85	77	17
71	50	9
67	97	6
3	90	4
78	86	3
97	79	8
27	63	14
34	30	6
3	29	13
74	18	8
80	57	4
12	73	15
56	84	5
88	61	10
8	7	7
59	67	11
97	71	8
53	59	14
33	29	8
81	79	3
36	79	6
74	14	7
91	1	15
20	44	13
83	53	4
5	70	3
99	93	6
60	89	3
12	69	14
53	62	5
72	2	9
87	46	18
5	21	11
76	50	4
32	11	5
67	45	5
58	37	10
81	24	3
14	17	2
89	62	5
62	5	13
26	20	11
71	25	10
1	4	10
93	48	15
29	26	3
41	63	15
26	29	3
79	43	14
72	91	5
43	48	9
82	10	3
2	14	13
84	49	3
12	11	4
10	60	3
41	77	9
88	30	12
94	32	7
30	94	9
56	9	13
47	14	8
39	53	6
15	16	7
90	96	4
71	76	12
47	91	10
40	55	11
91	7	19
8	54	4
22	30	5
44	34	11
48	36	3
83	52	7
21	38	7
13	72	3
54	85	13
79	6	7
77	26	5
86	16	4
11	88	18
11	49	3
1	73	14
38	72	2
28	47	11
100	36	5
20	93	26
93	92	11
24	33	20
52	89	9
67	7	5
9	74	29
14	11	3
46	57	10
80	83	9
11	82	14
99	62	3
100	4	18
86	98	7
25	74	13
51	81	13
69	85	15
70	15	9
82	85	7
55	46	2
28	42	15
46	40	7
62	22	5
59	87	15
94	19	2
3	55	7
69	94	8
95	72	20
56	5	6
69	99	5
27	50	4
83	17	9
59	57	10
77	62	15
95	46	3
7	85	9
53	9	12
3	70	8
14	77	5
70	86	14
91	9	8
12	83	7
29	28	17
79	56	19
31	35	9
62	100	14
99	56	11
30	25	11
76	22	7
56	38	9
22	69	6
87	17	10
68	41	9
19	31	7
89	49	3
77	87	10
64	69	10
31	2	7
10	36	7
44	87	11
54	76	7
85	49	3
61	24	3
40	53	4
22	92	4
72	31	10
31	15	3
13	73	4
32	18	5
40	93	18
17	100	11
62	97	16
61	72	11
94	44	10
15	59	21
87	13	4
73	91	7
55	50	8
36	5	3
77	82	9
45	69	10
9	41	9
87	14	6
37	55	10
55	47	12
51	92	6
31	99	11
48	70	14
76	87	15
37	79	11
42	54	14
65	8	9
99	71	11
11	100	18
92	4	7
87	37	14
24	23	5
82	64	9
41	11	3
8	11	9
9	19	15
69	15	3
6	35	9
63	79	18
60	46	6
73	31	4
94	54	5
31	12	11
17	55	13
43	56	5
32	42	12
6	20	26
54	51	9
91	10	11
67	33	11
57	16	3
74	62	6
9	55	3
56	22	11
77	57	12
25	63	7
15	67	13
65	99	4
90	17	5
21	69	8
37	99	10
49	29	7
42	62	11
77	54	8
91	62	12
4	13	11
30	32	14
99	77	5
51	58	12
40	87	8
18	3	6
90	4	19
96	35	10
76	29	7
66	18	12
24	76	4
70	68	6
50	57	5
87	65	4
57	37	8
74	80	10
44	12	9
92	24	8
2	15	8
51	54	9
51	61	10
52	43	7
59	96	7
96	42	7
80	63	7
39	11	13
34	95	3
76	39	4
74	44	14
61	48	6
17	95	12
64	91	2
6	43	8
43	7	2
46	84	9
100	66	7
79	74	12
12	23	10
51	80	6
67	95	2
2	89	9
21	72	18
13	59	9
83	63	4
84	82	17
84	41	2
92	67	4
64	74	7
95	55	4
18	63	10
42	50	9
9	84	6
36	97	17
91	51	19
59	92	9
33	2	6
35	9	6
92	15	4
33	34	8
36	18	3
23	94	7
53	17	6
11	84	8
47	34	7
27	69	9
35	72	15
94	76	8
70	27	10
86	36	15
91	15	2
85	6	10
4	62	11
76	9	5
42	24	9
16	32	9
9	90	4
34	61	3
12	51	11
18	96	4
58	33	12
10	39	9
65	36	7
88	10	4
93	57	8
66	71	3
30	39	6
64	21	4
28	31	12
8	53	12
15	47	6
7	60	8
86	1	6
82	15	15
96	98	12
2	53	8
31	64	9
65	94	7
59	52	11
89	97	2
88	5	10
87	26	9
65	1	5
67	90	6
7	59	4
84	21	10
19	87	4
82	71	4
100	19	8
49	41	7
10	7	3
97	45	20
20	92	8
4	86	15
1	10	9
25	65	9
63	36	9
71	77	6
87	95	13
95	26	18
98	97	5
39	48	5
94	27	11
10	71	6
60	96	5
46	75	9
7	91	4
19	53	3
14	6	5
73	23	8
24	88	7
39	77	11
84	4	4
18	91	19
23	7	8
94	92	7
21	71	11
45	86	2
10	4	4
3	39	9
7	25	4
50	76	10
50	67	4
36	46	3
77	100	15
4	61	4
96	12	11
73	63	14
64	98	11
31	98	8
15	49	4
67	27	18
44	5	3
44	96	10
47	89	12
45	19	8
48	86	10
32	100	5
74	38	13
72	64	17
8	43	9
56	27	4
47	88	6
29	55	9
11	64	10
11	5	7
43	6	4
100	70	5
9	75	11
71	100	4
68	9	10
72	51	3
13	28	11
48	42	6
79	17	15
96	51	2
69	60	13
67	30	3
3	57	14
47	23	12
43	49	4
87	22	4
57	82	10
6	57	9
6	11	9
96	72	5
99	60	4
19	6	8
49	10	16
36	47	19
86	48	11
2	86	6
75	88	14
38	43	13
84	15	5
70	80	9
84	53	2
34	6	8
85	33	8
95	39	3
6	55	2
18	19	11
45	95	6
11	25	8
78	51	6
73	32	4
7	46	14
68	11	11
8	81	6
97	37	6
49	21	10
41	14	4
65	97	4
12	4	9
5	47	9
35	18	10
83	36	7
60	65	5
45	87	6
94	2	4
3	96	5
42	47	4
17	9	7
79	22	11
35	55	6
57	11	5
18	13	9
90	85	4
58	61	17
75	42	6
99	6	16
5	49	9
95	23	6
60	61	5
67	29	17
2	87	11
23	16	10
78	57	3
5	83	3
23	72	9
61	9	2
36	45	5
46	65	8
57	78	5
93	63	8
26	47	9
72	42	19
16	61	3
85	45	11
83	10	8
64	100	6
2	79	12
16	71	6
63	22	3
90	42	8
76	53	8
22	59	6
38	90	12
51	96	8
95	27	5
1	94	11
43	52	8
52	95	12
98	30	11
27	66	4
42	38	5
39	91	7
52	40	10
39	43	20
33	23	7
26	95	5
35	7	10
14	70	8
70	9	6
43	14	9
22	63	7
22	43	6
40	98	3
60	100	11
23	84	11
55	51	7
2	40	2
89	38	10
28	61	11
55	99	4
63	29	5
10	94	10
80	36	7
90	19	8
65	25	5
71	2	18
26	81	8
51	29	7
42	61	15
100	30	14
30	31	4
92	12	16
55	35	7
12	63	6
83	2	7
2	68	6
47	31	3
23	85	10
78	53	3
93	44	5
99	59	3
34	97	4
15	75	4
96	71	17
74	13	10
11	3	9
75	65	7
9	69	8
65	14	4
86	27	4
85	22	6
9	53	9
49	58	3
37	88	7
69	46	3
26	100	5
42	20	2
28	33	10
6	81	15
5	1	12
61	68	8
50	73	7
61	88	5
68	69	5
48	37	6
3	12	3
13	83	6
42	60	5
14	29	2
40	33	4
70	90	9
69	32	20
42	91	14
21	89	14
53	92	2
50	79	3
50	32	4
8	65	10
9	82	12
33	9	11
85	50	16
82	25	8
51	11	6
51	44	9
35	3	9
34	54	5
12	14	4
25	93	6
57	18	4
63	51	12
74	61	4
39	88	3
17	61	11
57	58	11
97	70	14
59	80	6
82	42	5
43	68	5
70	3	10
17	35	8
100	69	12
84	97	5
87	98	4
37	69	2
80	65	3
45	79	3
71	89	7
68	23	11
44	50	4
1	92	28
41	18	6
10	29	8
11	85	6
2	10	6
37	5	4
85	16	10
35	5	3
34	9	11
43	85	11
81	22	6
28	76	12
88	42	4
93	9	7
62	51	10
97	25	10
13	85	12
88	62	5
44	39	12
9	2	6
68	37	8
33	10	3
14	3	4
51	9	8
69	34	4
53	14	5
94	13	15
90	22	2
49	90	4
32	54	9
97	85	4
23	98	7
35	74	12
60	29	11
78	49	7
7	100	9
54	21	3
29	100	3
31	74	7
48	3	3
17	83	18
38	66	8
60	48	5
75	87	8
86	6	10
98	75	3
82	44	8
18	82	11
97	83	13
95	74	5
85	53	4
100	5	5
100	33	8
3	38	10
40	49	3
26	27	3
52	23	9
37	92	9
41	10	12
5	7	4
35	4	3
27	78	8
11	90	7
51	43	11
57	69	8
74	86	10
44	29	5
2	12	3
5	51	3
6	67	3
73	50	9
1	53	6
30	6	5
67	82	7
69	53	10
94	52	4
20	63	5
23	45	5
94	61	6
36	2	9
89	41	7
5	38	7
89	40	3
52	55	12
76	27	5
60	66	7
66	57	2
8	77	5
83	71	5
50	83	14
34	92	8
96	88	2
98	96	6
96	52	2
65	18	3
25	75	11
69	9	3
66	36	5
24	37	4
88	96	11
39	59	26
22	100	3
19	37	8
92	18	9
50	8	4
93	68	8
25	66	6
86	5	10
24	78	10
35	1	9
44	30	5
100	77	3
54	75	4
28	88	4
54	19	8
64	50	8
20	52	8
53	94	8
62	26	4
58	51	2
91	69	5
90	47	2
48	1	3
20	35	4
92	95	16
84	54	11
37	8	8
26	28	3
24	94	16
44	75	3
38	86	2
31	92	3
8	67	3
9	73	12
95	63	4
2	20	7
36	22	5
45	9	5
56	26	5
11	59	14
49	87	5
38	75	11
49	54	4
53	32	2
88	75	5
32	28	4
81	11	8
35	20	11
63	74	8
57	17	2
18	90	6
67	99	2
54	78	5
72	1	9
75	1	8
44	19	15
38	26	3
71	36	4
50	24	11
54	37	11
56	61	6
1	67	3
39	2	4
43	97	7
67	4	11
88	17	8
83	64	3
58	1	9
93	23	5
62	43	19
9	50	25
41	54	8
90	67	7
34	64	8
45	14	8
68	39	3
95	98	13
84	17	5
18	28	10
60	59	2
52	87	11
74	90	14
89	93	7
63	31	6
39	26	15
88	94	7
39	80	5
3	75	3
73	22	9
45	100	19
5	58	14
18	74	4
100	39	8
27	96	7
51	88	13
3	67	6
68	89	6
41	40	6
85	61	9
35	6	8
90	75	15
11	6	7
56	75	9
52	8	6
47	81	5
75	79	13
4	48	7
63	94	16
97	55	6
62	77	2
67	53	4
40	39	13
27	77	5
67	80	5
29	20	14
38	35	11
84	76	4
75	68	18
2	23	5
93	87	2
10	44	2
20	16	7
26	55	7
20	54	8
1	31	4
47	39	7
55	98	15
33	11	8
23	99	23
97	34	3
54	2	11
64	1	11
66	93	5
89	91	11
11	92	5
12	44	14
50	49	3
8	98	12
73	7	7
85	5	4
76	8	11
14	94	2
86	89	17
77	70	9
72	6	5
93	84	4
38	24	7
68	15	6
6	58	3
49	20	6
45	97	9
72	57	7
55	71	4
39	67	6
82	92	18
66	29	5
5	69	12
29	10	9
90	45	12
98	83	6
51	70	8
57	63	2
95	86	6
61	51	2
59	29	8
58	93	4
20	78	4
60	3	11
10	100	5
55	79	4
10	41	8
86	43	2
38	6	9
55	68	12
60	62	12
55	12	10
52	88	6
78	54	16
22	66	6
99	38	7
3	56	5
12	32	3
69	86	9
71	78	10
30	19	12
82	29	15
82	7	3
40	21	4
91	57	4
64	62	6
96	97	4
88	70	14
91	95	5
65	68	14
37	77	7
23	95	15
80	23	6
20	24	10
71	49	10
16	25	8
38	71	9
2	97	7
85	4	5
46	60	6
33	14	11
28	20	9
79	46	5
99	67	7
81	8	7
48	99	10
56	6	9
28	90	14
6	51	7
43	84	9
51	89	5
95	60	15
26	63	3
60	70	3
81	91	11
60	95	7
36	40	6
88	90	7
54	26	9
6	38	11
77	6	25
54	95	9
55	63	18
14	60	4
67	73	9
29	27	2
100	6	6
23	55	11
73	30	7
85	39	5
45	54	8
56	44	11
76	2	9
54	38	4
20	23	6
13	93	9
85	64	6
28	9	13
10	82	10
61	45	25
95	38	2
64	97	9
99	80	5
96	32	11
53	27	3
24	93	5
58	28	6
27	75	3
61	67	5
94	81	7
1	21	9
71	60	2
43	83	12
35	88	7
29	80	3
5	79	4
18	34	2
68	91	14
21	90	10
93	27	18
2	96	8
28	43	5
63	53	9
75	30	15
60	22	16
51	59	9
6	30	5
58	100	6
63	13	6
9	18	5
99	51	12
7	70	6
73	15	3
24	52	3
65	74	3
53	12	4
33	73	12
30	22	5
63	58	9
22	2	3
6	42	2
100	93	3
22	54	15
50	94	5
54	14	5
8	2	7
8	73	4
41	28	9
21	68	4
100	79	13
10	75	5
86	4	13
54	25	6
66	20	12
52	76	3
84	43	4
15	24	12
18	26	16
19	71	7
43	59	7
1	65	10
95	5	11
62	23	3
13	58	20
48	27	8
59	50	4
40	69	8
38	14	4
43	13	12
77	1	8
94	99	8
20	59	8
14	38	6
45	85	11
61	95	7
1	36	11
52	33	3
56	12	6
50	45	16
71	1	5
56	3	11
12	72	4
20	68	9
27	12	16
27	59	15
1	93	12
13	36	6
84	42	8
19	91	8
69	87	4
44	13	6
93	5	9
70	82	12
14	23	4
30	99	8
78	69	3
23	60	5
66	32	11
96	95	5
81	74	4
41	75	11
25	34	5
25	79	5
36	65	3
22	15	8
98	49	9
69	49	11
98	3	5
46	74	12
6	72	13
33	49	12
98	60	7
95	94	7
75	73	12
35	94	3
12	88	5
99	12	9
68	48	3
59	22	16
99	61	14
44	9	4
49	22	15
45	94	8
61	92	6
96	40	2
47	24	5
36	71	3
13	43	5
37	53	7
27	20	3
98	45	5
83	25	7
100	35	6
50	89	5
96	82	5
19	85	2
31	40	4
85	19	5
92	96	8
60	45	8
59	55	11
18	12	7
4	32	2
84	65	7
34	86	25
26	4	6
60	33	6
73	76	13
7	40	15
86	76	2
8	76	8
3	19	12
86	92	2
84	1	4
35	60	4
88	49	23
67	28	4
90	15	14
25	47	9
73	83	5
23	62	4
70	19	6
63	83	4
28	14	3
32	95	9
21	92	8
12	35	3
50	19	7
53	89	8
17	12	8
53	47	3
43	77	4
39	73	6
41	79	3
25	18	8
38	30	8
8	59	8
83	57	14
68	27	11
53	95	5
15	50	7
21	23	8
89	80	8
24	34	11
23	21	12
11	58	8
11	68	6
100	32	3
15	38	7
74	45	2
24	25	8
37	27	8
9	60	6
86	11	4
83	80	6
72	27	4
2	21	10
51	36	6
77	66	4
62	13	5
83	90	4
82	37	2
18	98	7
29	64	4
85	21	7
45	42	8
67	63	6
34	41	4
36	67	8
32	35	5
3	63	11
92	16	2
48	94	16
21	13	12
18	40	15
38	59	4
19	25	19
36	43	4
10	23	6
34	55	22
81	9	8
73	68	3
27	24	10
90	73	12
11	57	5
33	78	5
83	8	5
76	34	7
40	61	4
68	56	13
55	86	5
47	29	6
57	8	10
18	29	8
11	81	4
29	21	7
86	71	5
88	36	7
73	9	6
89	59	12
50	77	5
14	7	3
39	72	6
69	73	10
19	83	2
82	47	5
73	54	3
67	51	12
56	16	4
22	90	13
18	58	4
35	32	3
87	18	4
18	49	10
4	56	9
93	67	8
57	5	7
92	86	11
65	90	3
32	55	11
90	61	6
2	52	22
79	81	6
34	35	8
58	23	8
96	11	16
30	23	14
81	37	5
41	94	6
64	81	8
76	5	4
83	37	2
7	78	11
17	2	6
37	3	12
48	78	7
9	23	12
84	86	4
45	1	9
31	3	9
16	23	5
9	88	7
65	79	2
12	21	6
40	97	3
57	56	11
74	98	5
86	35	10
10	32	10
62	86	3
54	84	2
68	49	11
25	51	2
39	36	7
44	55	4
54	56	6
78	63	3
12	66	11
65	95	11
62	48	8
8	66	3
17	8	9
69	82	5
10	57	8
49	65	13
97	43	5
3	9	17
79	69	2
86	18	12
12	81	10
28	64	5
31	34	9
16	26	13
64	70	6
50	96	10
61	41	2
66	17	11
25	97	6
77	65	7
62	73	5
97	57	2
3	24	5
36	6	6
49	91	4
4	6	9
26	16	7
92	52	8
18	9	5
81	92	16
75	94	5
40	86	9
24	74	11
71	68	5
92	54	9
1	62	4
50	97	4
51	13	14
98	99	16
81	97	11
81	56	9
15	64	14
43	24	7
55	45	6
19	12	18
27	79	6
62	75	6
60	88	7
42	31	7
74	50	4
74	88	6
13	60	6
99	65	9
92	39	6
40	68	9
45	13	10
11	95	3
48	43	3
13	29	3
95	9	6
71	6	7
14	68	15
21	81	9
94	38	5
28	13	3
34	96	9
37	26	4
5	56	3
54	68	13
70	95	2
9	95	11
67	69	5
66	67	3
53	36	3
87	1	5
26	21	11
16	87	11
95	16	11
65	26	4
63	86	3
3	5	6
70	92	19
70	41	3
7	27	5
94	86	3
79	89	6
55	10	13
48	68	2
1	43	7
86	50	6
63	49	14
72	79	4
1	98	10
39	28	2
97	88	6
36	49	12
29	63	11
68	67	5
97	66	6
74	96	5
43	12	8
27	42	12
54	98	7
36	29	9
87	12	3
91	97	14
37	35	4
56	71	4
92	28	13
98	73	12
88	38	3
63	12	6
12	9	10
61	2	8
81	46	6
44	58	7
97	26	9
58	40	6
93	31	10
79	97	14
96	37	7
56	24	16
97	44	10
73	10	8
82	14	7
88	53	11
36	30	4
41	81	12
46	39	10
97	89	10
17	23	7
60	50	15
61	65	13
46	61	5
27	98	6
88	11	6
62	17	4
19	77	5
34	42	6
33	89	2
67	71	3
64	23	10
29	45	4
6	66	2
64	39	9
75	55	9
46	96	7
44	86	6
21	24	6
77	61	6
68	3	8
28	48	5
8	49	5
85	68	5
58	59	12
78	36	8
27	9	2
90	6	8
49	72	3
37	50	3
88	41	9
64	13	11
57	43	8
19	11	6
14	80	5
72	86	9
91	76	14
13	98	6
27	71	8
1	81	16
6	98	9
36	41	12
38	50	8
79	59	5
68	2	4
16	20	5
7	86	4
71	46	5
48	74	3
97	31	17
77	60	5
64	41	6
2	93	4
64	84	16
10	64	12
49	94	3
52	66	4
50	18	7
95	43	5
65	71	14
90	56	4
43	72	3
27	37	11
87	7	2
27	30	13
32	65	3
5	62	5
31	53	10
30	54	4
92	29	5
61	40	8
16	19	7
99	7	11
75	35	6
9	99	7
92	14	13
100	22	9
94	78	8
48	12	7
94	60	6
23	33	15
60	72	3
35	71	12
20	96	13
96	41	4
65	75	9
38	16	10
75	10	9
78	65	12
49	9	12
91	47	10
49	86	5
74	76	4
49	74	5
7	97	6
89	94	8
87	78	4
49	31	6
80	95	12
43	41	11
49	83	9
60	77	7
52	85	6
47	20	2
99	27	15
99	30	16
2	1	5
81	19	4
97	95	5
61	86	5
11	62	4
7	52	15
70	96	7
86	99	10
76	42	7
72	3	4
53	45	11
41	26	7
11	97	11
66	75	7
74	43	5
76	97	5
41	27	11
66	6	12
70	79	6
93	47	8
96	64	12
24	39	6
46	71	7
74	83	5
40	1	12
78	64	3
2	17	10
57	12	6
69	18	3
39	79	5
89	39	7
61	47	11
94	29	8
13	100	8
52	16	11
90	26	5
17	38	12
20	83	7
96	70	12
43	10	13
34	21	10
79	83	5
55	28	6
33	77	3
42	72	2
42	22	2
59	70	7
90	30	8
52	47	9
9	86	14
43	61	3
56	93	8
59	25	10
27	29	13
100	82	7
100	52	12
47	64	13
26	38	11
68	98	9
1	75	6
57	30	9
22	52	6
39	42	4
39	63	5
63	33	15
15	94	7
97	99	14
79	5	10
41	93	6
39	9	10
5	68	7
25	96	6
64	24	8
40	79	16
70	35	11
91	63	16
44	68	7
66	52	7
20	62	6
75	18	11
87	5	6
72	39	6
30	20	11
14	27	4
40	71	6
79	100	6
50	11	8
64	29	4
19	51	4
50	35	4
38	100	9
4	36	4
93	22	7
61	97	10
19	32	8
61	32	13
17	75	4
52	80	5
13	61	4
80	40	6
68	60	6
74	28	6
7	76	3
59	79	3
43	15	4
3	43	14
44	18	6
14	44	3
80	98	20
38	4	8
8	51	9
98	91	12
47	3	13
20	1	7
30	93	10
71	3	4
50	95	5
26	39	5
70	75	11
83	74	5
55	66	2
89	17	8
29	3	13
98	77	5
41	44	5
5	72	10
37	52	2
29	37	3
40	20	5
60	24	9
94	95	3
59	74	9
70	47	3
20	9	8
12	31	16
10	17	6
3	51	13
55	80	3
47	46	6
59	71	10
10	30	12
28	30	7
15	87	6
52	100	12
18	46	5
47	50	3
92	76	12
19	17	4
33	55	7
85	34	5
65	81	17
46	23	6
92	75	4
6	22	10
61	31	6
71	80	7
32	24	9
87	40	10
35	84	9
87	92	11
28	85	4
36	62	15
88	20	15
95	58	4
31	46	6
100	26	9
9	11	10
48	25	5
31	72	5
5	75	11
77	43	4
67	41	11
28	38	11
38	33	4
90	53	12
31	38	9
89	10	6
97	7	8
74	63	2
88	19	2
83	78	7
22	48	5
31	55	6
6	37	10
62	95	11
81	89	9
63	28	6
88	9	6
68	7	13
5	36	4
97	75	6
8	45	10
56	31	8
68	77	2
37	21	10
55	44	19
44	88	10
95	76	11
84	61	11
100	87	3
82	1	8
48	13	2
85	8	4
71	38	8
92	38	6
58	46	12
29	78	9
45	27	2
13	90	2
51	56	13
25	83	3
64	7	8
80	17	7
72	98	9
70	8	12
60	94	7
56	20	14
30	59	13
90	97	10
96	31	14
27	17	8
47	61	4
87	48	4
69	37	4
84	77	9
50	61	8
76	47	5
53	5	4
76	10	9
60	53	11
69	5	9
78	52	3
91	25	8
95	28	10
78	12	6
53	23	7
48	18	4
53	88	10
11	54	3
11	67	8
99	90	11
78	71	10
83	42	5
33	30	12
4	34	7
61	55	7
54	71	5
31	8	4
28	58	8
82	45	11
18	8	7
62	38	8
35	49	6
46	33	5
88	64	10
43	40	11
36	85	10
68	36	9
75	4	11
98	70	17
63	71	5
7	16	9
8	48	6
2	84	8
4	53	6
77	80	11
30	85	3
85	25	10
79	27	7
96	14	4
77	90	17
61	11	7
36	20	11
77	92	3
80	82	18
67	89	10
8	1	8
100	9	7
44	49	9
96	63	5
67	77	14
85	14	5
77	4	8
52	29	3
30	81	3
28	79	3
85	59	11
27	81	7
94	26	11
92	31	6
21	17	7
68	6	8
89	48	10
38	84	4
38	1	19
25	23	6
37	97	8
97	100	7
10	20	8
72	82	3
94	63	2
71	11	7
87	68	8
71	88	12
98	44	11
16	60	12
83	49	6
85	1	9
23	78	12
53	85	7
54	62	9
59	47	3
5	97	9
25	90	4
72	74	6
18	73	3
24	97	4
11	75	5
84	63	5
80	21	7
81	69	14
4	59	7
59	76	7
73	11	9
3	22	11
68	21	3
85	94	5
5	34	2
18	36	21
11	55	12
91	31	2
18	42	4
92	30	5
91	21	14
19	18	4
14	19	5
42	59	10
31	58	5
82	31	11
22	74	10
34	59	7
44	85	10
26	41	4
54	18	13
100	50	8
34	51	15
43	81	3
48	56	8
69	95	19
94	87	8
89	73	7
79	35	4
88	26	6
19	54	3
21	26	3
49	43	18
65	76	10
3	97	14
64	22	4
7	84	6
76	16	6
26	6	5
18	66	9
6	7	2
37	24	3
64	66	3
60	12	7
73	85	11
34	100	6
22	68	2
29	43	10
75	69	12
9	91	9
49	77	12
29	84	7
11	18	8
68	78	5
95	14	10
14	67	5
57	46	6
48	90	5
100	51	16
51	57	10
84	37	5
87	90	14
97	56	13
93	75	3
69	48	4
36	69	3
7	83	15
6	5	5
73	60	8
17	77	12
43	95	16
23	47	4
90	38	9
24	15	3
69	50	7
65	46	3
79	87	7
48	80	10
46	7	3
3	32	10
46	87	4
64	11	2
92	25	6
14	81	9
36	38	4
55	64	9
92	35	4
20	48	9
89	95	7
16	29	6
67	35	12
85	60	4
62	16	6
80	62	3
75	7	19
57	23	5
74	51	4
26	50	3
60	30	5
14	55	3
51	27	13
76	66	5
7	65	5
35	51	5
64	95	7
23	38	6
58	39	9
12	34	6
69	89	7
17	99	5
70	84	10
72	68	8
55	91	7
73	94	6
92	9	5
4	29	6
29	85	6
61	35	5
1	7	7
41	70	9
96	61	9
62	55	5
48	28	7
7	77	4
32	13	13
13	11	6
2	25	8
24	68	4
92	81	13
93	26	5
27	13	6
34	58	5
84	57	7
35	10	4
78	2	2
95	48	6
60	35	8
31	87	16
12	62	5
92	57	6
97	76	25
66	85	6
88	65	5
91	49	6
50	70	7
94	72	2
76	75	5
14	53	5
100	80	8
96	27	14
62	25	8
75	61	7
26	34	8
96	50	5
3	72	3
66	96	3
61	96	11
39	10	8
25	30	12
68	55	2
94	22	11
49	15	4
15	63	8
69	72	9
41	62	7
89	83	12
12	96	11
5	93	3
89	12	11
45	17	8
83	50	7
4	68	8
95	12	10
59	7	4
90	18	15
79	16	6
30	97	9
85	32	4
53	90	5
83	65	7
75	43	12
60	39	19
69	13	3
25	68	6
34	53	7
36	24	6
14	35	3
72	65	8
64	82	5
70	63	7
26	98	13
80	33	8
100	42	5
82	83	6
83	95	5
86	22	4
54	40	4
35	93	10
36	68	17
84	14	12
30	76	10
35	57	3
50	37	9
79	62	4
31	70	4
13	26	10
59	56	4
28	36	7
49	53	2
10	96	5
98	78	10
65	10	4
6	40	5
58	21	3
59	64	8
66	68	12
31	13	10
54	20	6
8	33	9
73	98	10
81	82	3
30	12	10
22	5	6
20	21	3
73	26	6
93	3	7
17	69	4
62	31	9
93	60	3
15	17	9
95	49	11
39	49	4
76	62	2
1	100	13
83	29	9
5	76	3
24	1	19
25	88	2
29	70	9
46	55	8
69	38	3
63	87	4
81	34	7
38	32	6
38	18	7
59	81	9
40	38	4
75	57	8
28	6	10
78	68	27
49	88	5
96	89	3
74	70	4
20	87	8
58	24	3
1	58	2
90	60	13
29	11	13
39	7	4
17	66	7
91	93	5
78	96	21
31	48	12
86	93	6
57	36	14
60	1	5
49	17	7
42	23	2
98	25	6
46	17	6
63	66	9
23	42	5
58	49	11
82	39	4
54	5	11
78	90	6
25	37	6
10	37	12
38	92	4
75	62	5
58	26	4
2	74	4
32	37	12
37	91	17
34	8	9
25	46	5
94	30	12
45	59	9
11	79	4
70	60	6
21	2	5
61	5	10
25	98	11
8	92	3
1	80	4
36	3	2
20	32	13
47	17	8
37	36	4
14	45	11
89	21	5
26	66	20
97	63	11
96	2	8
40	23	12
92	47	9
43	19	11
12	54	4
16	98	4
38	47	6
23	11	11
5	37	3
70	64	6
56	14	8
11	28	16
37	68	3
40	9	6
80	79	5
74	20	9
2	3	9
99	95	7
50	87	8
51	16	11
71	87	7
19	42	4
95	29	5
42	37	4
8	41	11
85	75	7
79	47	11
39	95	8
8	71	3
53	82	5
77	48	5
4	71	17
71	66	10
31	41	7
37	41	3
88	63	14
51	19	8
14	56	4
82	40	12
66	79	11
50	1	7
41	6	3
53	1	7
82	61	8
19	10	4
78	72	3
15	85	3
59	46	4
11	91	23
99	8	8
9	63	6
59	8	5
46	38	7
61	99	13
4	37	7
62	60	4
16	13	7
9	78	3
41	1	4
86	61	9
16	40	5
5	77	9
11	7	7
51	71	13
92	80	8
29	12	11
85	98	10
54	28	7
36	84	5
98	90	19
76	44	5
28	23	10
77	74	4
92	89	3
3	77	6
91	8	11
90	25	4
34	56	3
90	9	6
52	51	9
49	82	5
3	58	6
46	31	4
74	60	8
3	48	4
25	87	7
40	5	11
6	24	6
44	28	4
42	87	11
21	44	6
44	10	5
59	60	5
96	3	16
48	61	6
56	53	6
29	77	6
11	24	4
74	26	11
54	23	11
5	80	3
10	24	10
56	21	4
26	93	4
84	31	2
16	86	5
96	87	10
46	20	6
95	1	11
85	15	3
86	77	3
61	25	17
100	18	7
16	88	6
17	13	4
80	87	10
67	17	6
9	52	6
14	66	4
82	20	5
50	74	3
99	50	4
70	73	4
94	56	5
76	94	5
7	13	16
92	5	5
76	92	4
38	55	16
83	45	5
57	81	6
81	53	4
2	32	6
10	2	11
9	80	4
64	90	4
44	80	11
15	88	4
93	36	2
97	98	7
22	85	4
99	20	6
37	39	7
17	58	7
8	58	2
49	98	2
53	29	5
32	2	6
62	74	3
58	8	6
90	3	8
98	69	7
98	43	12
36	42	7
10	95	6
37	60	6
58	76	6
38	17	3
31	5	3
30	60	9
95	73	6
59	15	2
5	87	4
51	15	4
13	82	4
36	33	6
51	42	8
16	92	7
35	8	11
46	34	4
3	82	4
63	34	4
7	20	8
53	93	9
96	86	5
9	15	5
14	25	4
91	58	10
56	36	7
2	95	5
49	95	2
70	54	3
83	84	4
79	45	11
90	24	6
75	54	2
32	38	2
1	49	11
47	72	4
9	27	7
100	71	12
98	95	16
95	71	14
41	100	11
61	18	2
63	37	2
4	81	7
14	89	3
96	73	4
37	9	5
36	39	2
13	44	9
100	61	12
4	83	3
39	27	11
41	46	7
59	9	10
66	7	4
85	51	3
7	15	4
38	37	8
1	24	7
30	49	2
2	36	3
84	75	12
9	4	10
58	96	5
56	90	9
78	21	5
73	33	7
94	14	5
6	46	20
41	96	7
7	89	6
33	32	21
53	3	6
35	12	3
73	39	5
41	2	7
9	87	5
7	88	4
5	92	2
62	67	7
78	55	6
61	84	9
75	9	3
87	10	7
74	56	5
80	45	19
4	93	4
42	36	3
22	95	17
79	7	7
13	18	2
18	37	10
32	81	3
1	19	3
50	21	2
37	23	5
76	78	9
96	24	3
88	39	6
36	44	7
10	65	11
89	9	13
17	31	6
42	76	8
1	9	12
9	71	8
73	61	9
16	1	8
86	13	5
90	81	5
100	31	8
37	78	4
21	51	11
51	91	7
46	35	7
69	98	4
35	96	23
88	87	3
25	60	4
42	21	5
46	58	5
15	92	12
66	3	13
35	92	2
41	51	7
92	36	11
58	5	3
37	45	5
59	12	5
47	93	7
24	2	3
24	49	2
81	14	7
12	6	12
23	66	5
20	51	3
47	25	20
76	52	10
9	17	4
25	76	5
90	100	6
83	35	3
61	77	12
1	42	6
54	45	2
75	15	17
86	65	5
11	65	2
35	62	12
19	45	11
76	36	3
60	8	8
93	73	3
78	15	3
52	78	9
63	30	3
81	64	9
41	91	7
90	8	7
74	23	13
6	8	12
22	28	8
82	52	10
45	96	12
87	47	5
61	36	3
55	6	4
66	47	7
17	20	4
31	14	4
77	69	7
79	64	3
73	74	5
54	90	8
99	86	6
42	66	3
28	10	10
2	76	5
38	54	7
26	23	4
72	59	11
34	4	7
34	74	8
76	35	8
90	99	5
68	26	7
65	34	10
93	70	11
72	97	9
36	83	5
39	64	5
15	1	7
88	93	4
18	2	8
15	97	7
67	43	3
22	94	5
87	6	3
39	69	7
5	15	9
62	88	4
24	36	2
76	89	12
4	51	3
9	7	8
56	19	12
10	92	3
87	11	9
88	32	7
64	61	6
71	22	7
11	42	12
73	47	6
34	91	8
2	31	7
99	11	8
92	34	11
55	38	8
88	74	9
60	54	5
21	11	6
70	61	7
2	67	2
23	67	4
87	38	15
51	31	5
65	47	12
51	74	12
93	91	5
97	1	14
45	70	9
75	25	4
21	87	9
5	78	5
45	98	6
84	62	3
8	88	6
68	16	12
39	5	4
75	47	4
40	10	11
37	83	7
37	14	3
59	77	7
60	13	3
56	64	4
80	48	6
91	5	10
48	53	2
22	77	12
30	28	22
29	86	6
74	53	6
50	6	4
86	62	7
46	24	3
63	100	7
65	70	5
5	94	11
71	18	5
31	28	8
50	51	7
40	22	7
50	80	5
70	78	5
6	93	7
50	56	14
20	18	10
54	36	2
31	90	9
36	14	3
75	41	7
16	15	7
6	56	12
69	31	6
41	76	4
81	12	5
49	35	7
5	32	11
16	89	2
82	63	3
71	91	8
23	86	12
43	86	11
13	56	4
54	16	5
19	67	9
14	97	4
34	44	8
35	66	11
77	35	3
49	75	9
18	7	8
86	83	3
9	40	4
93	17	4
7	42	3
71	5	3
81	84	5
31	61	8
63	47	3
15	2	13
36	92	5
93	40	3
72	54	4
93	18	7
23	83	4
28	77	6
78	25	9
29	19	5
22	75	7
29	9	11
39	65	11
81	76	7
7	28	5
62	94	7
22	18	5
8	34	6
4	77	4
20	14	8
49	71	4
32	23	8
23	76	6
78	70	2
86	34	8
40	65	9
67	68	9
52	48	10
70	39	4
87	66	5
56	29	8
58	9	15
56	8	9
82	30	6
90	39	3
48	58	7
14	51	3
17	6	3
22	24	6
27	21	19
100	7	11
44	67	5
7	37	3
94	1	13
39	74	6
19	97	8
83	41	9
24	47	4
26	69	7
14	59	11
21	91	13
27	83	2
31	100	4
98	10	8
60	78	6
78	100	13
98	9	4
80	27	4
42	67	7
100	73	4
74	46	5
32	7	18
9	6	5
91	78	6
6	65	5
8	100	6
61	8	6
13	40	3
17	14	5
55	20	13
62	79	7
23	17	14
68	32	10
68	75	7
13	92	6
99	39	5
10	91	4
18	14	8
77	98	2
55	16	7
100	89	5
7	57	4
34	10	10
14	40	6
69	17	3
71	43	8
98	23	8
13	35	5
75	46	13
32	3	8
95	31	5
30	27	9
9	70	15
100	53	7
89	30	5
17	74	3
94	18	3
2	7	9
6	96	12
84	46	3
75	49	4
91	55	10
18	100	9
10	43	6
3	83	4
71	65	3
79	93	6
67	36	5
28	37	4
39	50	7
10	61	4
53	42	9
31	36	8
41	12	5
53	98	9
86	57	3
4	17	4
21	22	2
100	47	5
86	78	6
59	4	8
74	95	6
51	8	7
5	61	3
77	58	7
44	79	13
44	100	3
63	69	8
87	72	6
49	34	2
20	71	6
65	4	20
71	94	8
21	76	2
71	34	5
58	67	5
34	27	9
52	69	8
25	71	3
57	39	4
4	54	9
30	16	11
82	21	2
43	37	9
46	43	9
58	4	6
63	59	5
89	54	3
66	40	8
71	32	11
42	14	9
40	59	5
96	10	8
50	22	5
73	25	7
6	16	3
19	9	5
92	56	5
36	16	6
27	16	13
16	42	6
4	58	3
81	25	3
46	81	5
85	43	6
10	26	3
98	40	4
46	41	8
36	8	6
71	53	5
37	75	2
51	65	10
75	44	6
75	6	10
80	90	6
82	94	8
86	55	4
20	41	2
47	100	8
91	41	7
6	18	5
96	57	11
61	39	14
31	7	12
93	45	5
79	11	6
97	12	7
1	3	6
21	19	10
35	27	7
92	17	7
35	39	9
37	58	7
58	41	8
43	99	5
75	37	4
4	84	7
43	35	8
100	75	5
35	37	3
64	9	3
70	20	8
43	70	4
60	79	10
21	70	3
48	35	5
12	28	5
93	64	3
100	23	13
89	71	6
39	34	2
63	77	5
61	16	3
35	23	2
26	25	9
87	67	7
16	4	6
48	34	3
85	46	5
80	64	6
78	1	6
5	43	6
45	26	7
71	58	13
88	27	2
68	99	9
33	100	6
75	22	2
16	33	8
40	2	2
17	49	8
87	97	9
80	26	6
27	87	5
46	8	8
69	20	3
24	50	2
84	47	13
44	71	11
79	30	9
5	42	6
38	85	2
66	26	3
12	38	3
24	11	5
89	77	11
99	4	5
17	18	4
39	76	4
47	70	2
77	88	2
10	40	4
67	66	9
32	73	5
13	27	15
59	28	5
69	26	8
98	37	6
87	23	10
76	21	7
1	20	8
28	27	6
18	67	10
88	40	9
76	24	11
27	8	3
13	46	9
82	79	3
38	89	2
28	82	7
59	10	5
65	29	6
46	16	4
79	95	8
35	86	6
32	5	2
33	1	6
81	80	6
18	70	6
28	62	5
49	63	7
45	34	14
44	27	10
23	9	3
82	98	5
95	62	2
4	43	4
35	33	7
1	61	7
94	43	3
100	21	6
42	96	4
72	18	3
21	88	8
26	9	3
49	45	5
33	54	2
91	68	3
91	87	4
79	44	4
25	64	4
15	82	4
52	75	5
62	82	7
29	53	4
49	44	5
4	98	3
63	92	6
17	11	5
88	81	7
54	33	3
83	92	3
85	27	5
27	64	8
30	15	7
37	65	3
47	75	5
59	36	10
25	3	7
80	24	3
19	50	6
90	70	8
64	20	5
26	24	14
45	5	4
71	35	3
33	56	2
12	93	4
8	18	12
3	92	5
20	53	4
4	75	3
92	2	8
75	78	7
70	45	14
85	70	5
28	81	11
48	97	11
26	10	4
38	12	8
16	76	5
99	35	5
95	83	12
59	43	9
69	83	4
25	85	13
70	57	6
57	42	11
96	69	3
60	93	5
96	26	3
26	87	10
70	51	11
92	90	5
43	57	7
7	18	12
93	19	2
51	64	4
65	73	7
70	1	5
63	27	2
63	3	3
6	54	4
14	93	3
8	83	4
83	62	11
6	63	4
39	31	2
59	69	5
23	40	4
49	78	3
17	42	6
92	6	4
67	57	3
60	31	3
66	11	14
54	50	7
100	29	4
70	97	5
66	72	4
99	96	8
64	15	7
32	59	17
40	70	6
20	58	7
93	97	2
30	71	7
67	98	4
40	63	4
50	28	3
16	90	4
29	35	11
28	40	6
37	18	3
47	87	4
58	6	7
83	87	2
37	29	6
41	25	5
2	73	8
33	74	2
23	31	7
40	85	3
77	23	11
17	80	7
40	99	4
75	40	3
58	17	7
30	89	2
65	19	6
71	84	9
19	100	4
13	1	5
76	54	6
21	83	3
55	5	4
37	16	2
5	41	4
38	31	3
96	91	12
64	32	4
5	28	3
19	95	3
97	61	3
82	11	12
33	98	10
88	83	5
40	18	9
78	18	2
81	26	5
68	96	8
13	81	5
13	54	8
3	54	3
19	70	3
86	10	5
91	66	16
66	44	7
6	100	4
6	79	8
21	3	6
30	24	7
54	44	3
60	15	3
27	99	10
56	86	3
72	92	5
63	46	9
86	69	7
83	46	13
100	43	7
92	82	8
63	90	4
14	88	5
84	25	4
21	7	14
40	19	3
53	58	8
46	67	2
15	60	10
56	60	4
66	37	4
54	81	6
2	26	6
56	41	12
76	17	5
36	10	3
91	52	5
82	35	5
56	35	7
36	99	4
99	88	6
3	86	8
2	45	2
80	13	3
56	18	5
92	33	5
98	33	7
49	36	12
31	1	4
42	9	11
43	25	3
42	57	5
27	2	2
16	66	5
14	73	4
90	1	4
33	24	5
73	27	4
10	66	5
6	45	12
41	48	5
30	57	4
8	89	6
35	85	2
49	33	15
53	22	4
7	80	4
60	4	5
40	81	9
62	1	8
6	50	12
10	47	7
58	66	7
87	96	12
54	22	6
31	44	2
66	58	2
88	13	4
36	21	15
16	41	8
40	83	4
9	83	5
51	12	2
84	28	6
22	76	5
42	5	4
17	64	5
21	15	7
64	28	5
51	45	3
72	21	4
96	59	8
24	9	7
47	18	13
55	60	5
64	94	8
12	84	7
55	72	12
26	17	16
28	34	4
62	81	11
73	55	4
12	91	4
46	66	2
59	48	11
50	47	4
1	18	6
58	34	2
52	99	4
77	78	5
52	12	9
53	71	4
18	99	6
68	52	5
26	59	4
49	50	5
62	57	6
81	65	3
86	52	6
17	19	3
91	83	7
44	14	2
99	44	8
28	92	2
74	33	6
38	27	4
93	6	2
84	38	3
91	34	6
43	31	11
73	56	8
8	64	7
47	5	3
96	84	5
22	33	11
32	16	12
5	44	8
38	98	6
86	75	5
75	24	3
13	63	2
45	52	5
89	52	5
61	87	7
76	26	4
55	23	3
76	68	7
43	88	3
73	75	3
28	99	8
85	99	4
94	7	5
54	13	10
33	19	3
17	93	4
47	74	3
75	11	11
64	68	7
33	38	6
10	52	4
89	74	6
31	33	4
38	52	2
17	37	5
9	93	4
83	99	6
59	35	4
82	36	6
21	49	4
66	12	8
100	83	6
86	54	8
25	53	9
12	68	9
54	97	7
7	69	9
60	55	4
65	27	5
69	76	5
29	38	4
52	81	5
81	29	4
18	57	8
98	63	7
62	8	3
36	11	6
6	86	5
42	2	2
98	6	7
47	66	7
6	21	9
83	97	5
59	94	8
15	65	3
58	22	8
93	98	12
18	6	11
55	19	8
48	76	7
69	42	11
62	40	3
68	65	3
68	80	5
65	50	6
79	61	6
80	5	7
56	15	5
74	94	7
28	96	2
62	84	6
83	47	2
37	66	3
97	42	2
80	12	5
47	42	2
30	52	4
67	3	4
88	86	6
46	13	11
23	44	5
76	13	5
79	9	4
29	15	2
26	82	2
52	20	2
98	51	11
67	76	4
1	44	5
82	89	10
13	10	4
25	92	4
24	67	2
56	79	4
48	41	3
41	30	13
10	53	6
24	63	18
30	69	5
84	89	8
97	39	8
39	40	17
65	24	11
26	64	2
67	24	14
30	100	3
85	100	2
82	84	8
100	91	3
76	84	6
93	53	9
80	38	7
10	51	4
68	50	3
19	96	13
8	35	9
23	64	11
76	56	10
80	51	2
63	7	4
97	27	7
51	82	6
4	76	4
79	48	2
56	37	4
32	27	5
40	6	6
50	30	5
54	88	5
47	36	2
78	4	4
96	92	6
53	19	9
53	16	4
75	5	5
56	55	5
37	80	8
64	79	3
49	18	5
3	10	7
98	2	5
78	17	2
96	76	4
6	97	7
37	98	7
24	83	11
77	49	6
87	19	10
67	13	3
27	82	2
27	60	6
45	84	5
10	54	12
60	42	4
19	8	5
1	27	8
42	28	8
100	48	6
83	27	5
96	28	5
46	9	11
59	91	4
9	77	3
30	7	8
91	100	9
40	30	6
55	70	2
43	47	5
1	71	3
41	5	2
99	53	5
26	76	8
9	100	13
91	42	5
25	32	11
96	18	3
36	51	9
100	14	6
75	84	2
36	88	6
13	87	3
62	24	3
87	62	7
33	60	2
29	36	6
3	53	2
85	80	6
29	75	5
9	58	12
53	67	8
27	80	7
58	29	15
64	47	3
24	89	9
84	22	16
52	65	4
44	57	5
84	51	7
73	48	5
93	25	11
8	44	9
79	23	4
45	12	3
73	51	5
55	26	6
16	56	5
33	18	11
87	15	7
76	37	4
51	3	5
31	67	8
54	24	10
49	48	3
64	30	6
9	39	6
79	24	8
70	25	9
8	69	4
16	3	3
38	11	4
37	95	3
82	93	2
75	51	14
45	68	7
86	53	4
4	97	7
15	14	10
42	73	11
23	88	9
39	71	5
51	99	2
30	48	7
13	23	4
18	22	3
61	42	8
32	85	3
93	46	6
74	24	5
7	68	7
21	28	13
97	16	4
19	93	4
64	88	7
62	65	7
59	16	7
98	84	2
66	53	6
40	75	3
75	12	2
44	54	6
20	88	3
99	54	4
61	15	3
4	49	6
58	54	12
32	47	3
73	64	13
69	78	6
50	91	3
30	64	3
25	61	2
11	12	3
34	28	2
69	45	3
75	76	7
16	94	4
70	67	10
42	18	7
4	12	3
48	23	5
55	36	9
100	17	7
37	87	4
78	58	5
52	22	7
33	59	6
77	21	5
53	25	9
15	10	3
45	75	12
81	23	8
61	60	3
14	12	4
69	8	3
36	100	5
38	36	3
62	80	4
92	58	7
96	55	4
47	28	5
12	58	3
85	79	12
58	2	4
47	44	12
82	69	6
63	9	8
90	48	4
10	69	3
71	59	5
15	62	5
56	42	6
94	55	7
71	42	4
53	15	2
88	85	8
52	50	2
74	8	5
45	10	9
3	91	5
59	99	9
11	4	10
9	47	3
55	77	10
90	57	2
48	17	6
57	22	3
61	33	10
97	92	5
41	38	9
49	3	5
30	51	8
39	99	2
81	41	8
19	34	4
92	91	4
29	5	12
23	46	3
23	54	4
17	94	8
48	72	2
34	93	8
58	77	5
5	8	5
79	96	9
23	61	3
15	23	2
5	46	5
26	42	3
84	87	5
55	7	4
2	24	2
33	86	7
4	88	8
73	99	7
77	84	6
100	59	3
83	39	6
9	14	14
8	13	5
45	73	5
81	86	2
20	12	9
66	88	3
95	91	9
79	75	4
97	90	7
18	84	3
87	28	3
68	29	2
68	33	3
53	86	3
8	6	4
97	65	10
100	25	5
66	4	7
38	3	4
68	85	3
54	63	4
70	87	6
42	35	6
51	69	2
24	51	10
77	36	13
52	3	5
6	69	8
84	18	3
16	8	3
49	100	3
38	94	6
7	72	2
86	45	9
29	68	6
75	86	4
38	8	6
100	96	3
91	94	4
59	6	6
78	20	2
84	39	5
16	57	3
38	53	2
97	87	7
52	30	2
30	35	3
35	26	4
35	63	7
64	75	3
48	98	4
72	34	6
17	86	6
22	51	4
4	30	4
87	79	2
11	53	8
26	53	8
6	15	3
94	62	5
41	35	2
15	46	6
13	88	4
69	59	9
26	67	4
88	82	3
57	61	8
6	84	4
23	49	7
4	33	6
4	82	8
94	93	4
53	30	3
14	5	7
97	21	8
81	60	5
57	79	7
28	67	2
95	61	2
43	26	14
9	62	4
92	22	5
62	83	3
86	9	4
28	87	7
58	81	9
16	96	6
69	23	9
77	7	5
60	90	9
43	76	4
28	91	11
84	92	4
25	82	6
42	77	5
47	33	9
14	39	7
60	68	6
35	34	3
72	75	8
87	29	5
12	57	11
98	4	4
16	69	6
36	75	7
76	63	7
22	38	10
77	91	4
61	6	4
22	6	2
27	52	3
7	93	3
86	97	7
71	95	3
49	68	5
98	74	6
7	58	8
24	16	4
48	82	4
5	54	6
83	93	3
78	84	11
20	15	8
93	69	4
70	34	5
56	25	8
27	45	4
82	62	9
89	68	2
66	46	4
83	3	5
65	91	8
91	40	3
35	52	13
10	45	2
75	27	3
18	52	5
80	59	4
23	96	5
46	32	3
18	43	6
76	72	5
29	61	4
50	27	9
2	34	4
95	82	5
35	58	7
39	90	2
16	73	2
36	86	6
59	83	6
77	47	3
33	43	6
76	48	3
51	95	5
19	86	2
80	9	2
15	5	4
79	37	6
20	56	4
11	1	9
43	78	10
61	80	3
64	99	3
70	52	2
57	50	3
18	72	3
81	59	5
27	74	9
12	27	5
67	78	3
64	6	6
9	28	6
34	49	3
79	78	3
100	49	2
39	18	10
50	62	6
68	73	9
91	60	3
12	17	5
96	75	3
70	44	6
5	55	6
74	6	6
47	56	12
10	42	4
25	36	13
65	54	4
62	18	3
52	10	5
12	48	2
78	97	9
82	27	3
21	64	10
33	13	4
32	62	5
65	33	3
55	90	3
8	62	8
73	16	5
44	52	3
8	3	4
46	64	2
96	33	7
25	42	4
72	84	9
13	42	8
60	91	2
63	56	2
75	67	4
78	6	8
93	95	3
34	20	13
58	82	4
16	39	4
51	20	5
66	86	7
83	51	9
28	56	3
12	56	9
31	49	3
56	67	4
7	82	11
100	60	3
69	56	2
95	54	7
42	58	8
84	2	8
67	61	2
30	13	5
44	35	7
94	4	5
67	72	5
66	34	4
24	66	11
41	47	10
42	34	3
11	2	6
21	6	8
17	91	4
70	58	5
68	22	7
37	34	2
22	9	2
11	43	5
62	50	3
98	24	8
73	77	3
89	63	5
85	31	5
56	83	2
29	17	4
58	50	4
90	16	9
64	16	3
50	66	8
69	27	5
27	11	8
22	70	6
50	81	7
19	90	8
78	75	6
59	2	3
6	83	4
60	10	10
86	41	7
40	88	5
69	84	5
35	11	7
56	51	3
5	88	5
41	50	11
18	50	2
35	15	7
75	19	4
80	47	2
55	17	14
53	31	4
47	40	2
91	71	5
1	54	4
38	44	5
70	37	6
15	53	3
88	16	5
52	31	6
16	64	3
7	17	3
24	28	8
14	61	7
60	14	3
22	41	10
100	41	8
83	96	5
5	23	5
57	95	6
57	80	4
8	84	4
9	51	5
90	13	4
96	60	16
43	100	2
84	13	3
33	76	3
53	64	5
51	34	2
77	38	5
27	22	4
90	49	3
66	64	2
6	2	3
11	66	4
18	10	4
30	68	11
90	80	3
14	18	4
26	99	11
67	18	2
91	38	4
2	70	5
1	28	8
70	69	8
7	95	3
36	53	4
45	72	3
31	39	5
90	51	6
8	25	5
97	78	6
51	78	5
69	14	3
86	7	3
60	69	2
80	31	7
69	88	5
32	72	3
37	67	2
69	29	6
12	41	6
32	53	4
91	81	2
49	62	4
16	81	7
61	34	3
53	74	6
11	46	3
67	22	6
84	52	2
84	69	10
62	54	5
46	51	12
28	18	3
46	78	3
89	98	3
36	64	8
26	92	9
48	95	7
73	4	5
54	1	3
71	79	4
42	75	7
77	67	5
27	49	3
44	23	4
83	7	3
93	24	3
59	19	8
80	86	2
28	68	5
23	80	6
65	93	6
62	98	12
80	49	4
72	14	4
33	81	2
96	38	5
49	47	3
43	98	3
32	75	13
63	25	7
58	14	4
23	22	6
60	2	3
48	83	7
58	65	5
11	74	7
14	82	4
22	79	9
76	4	5
23	59	4
5	25	2
27	26	4
90	32	16
81	36	6
87	53	6
88	6	3
12	78	2
100	45	7
30	11	9
66	45	9
8	96	10
75	56	5
89	26	7
88	78	8
46	73	10
12	33	5
29	40	2
63	20	5
95	68	4
14	83	8
6	62	5
99	46	6
96	93	8
7	71	7
10	5	2
19	24	6
15	36	2
23	25	3
99	17	11
86	19	4
36	91	7
4	66	4
4	41	2
26	48	8
31	86	2
77	40	7
22	71	5
40	89	3
51	97	5
63	61	8
76	51	8
79	73	2
17	53	9
16	27	7
95	97	5
11	16	2
18	39	2
94	89	7
20	75	11
3	40	4
70	62	2
87	42	3
62	70	6
50	71	2
51	90	3
51	87	2
96	78	13
2	85	6
90	95	5
28	2	4
43	82	3
73	19	6
84	85	4
90	20	8
84	16	7
73	78	4
91	29	3
17	56	3
60	41	3
19	47	2
24	17	6
71	69	7
32	6	9
96	17	5
96	43	5
32	78	4
45	90	3
6	31	3
33	48	7
33	26	6
24	48	10
63	14	2
61	66	4
39	33	9
73	42	3
73	24	4
63	64	4
62	4	5
21	86	5
43	38	6
53	54	3
2	44	2
67	87	2
64	43	7
95	41	4
84	40	8
23	91	4
78	94	6
68	94	2
8	32	6
18	69	4
16	48	4
81	43	10
33	28	9
34	85	3
92	48	2
80	39	12
15	89	3
48	2	5
89	92	5
95	37	5
53	40	2
74	29	3
41	16	2
41	24	3
81	90	3
48	75	2
13	12	5
38	97	6
7	47	5
71	26	3
91	33	9
43	1	2
37	71	4
71	28	4
99	55	8
69	16	2
89	50	8
12	71	3
25	73	4
35	61	3
71	75	5
55	37	8
95	92	3
94	97	8
100	90	3
89	79	5
34	31	5
16	9	3
4	35	2
18	62	5
42	41	3
80	69	5
91	77	6
60	16	2
59	20	4
75	13	10
93	51	6
64	8	6
56	28	4
28	41	3
53	87	5
46	94	5
87	61	2
17	85	8
99	45	5
64	93	6
34	60	4
21	59	4
98	7	7
99	31	3
48	67	6
59	32	4
7	50	4
61	46	4
8	47	7
31	51	2
4	95	9
85	84	8
45	67	4
79	2	8
78	23	2
10	8	2
1	16	6
14	34	7
14	10	4
26	79	3
55	88	6
11	93	2
27	62	5
25	91	9
10	19	8
80	34	2
59	65	2
14	1	3
94	80	4
84	80	5
31	24	5
14	62	2
21	100	6
39	94	3
98	87	2
49	64	5
5	66	4
19	73	16
30	84	5
79	85	2
75	8	5
73	58	3
38	69	6
15	35	4
6	41	2
96	80	12
30	70	5
93	8	3
80	89	5
13	33	2
48	51	6
55	67	4
17	90	5
65	72	5
57	29	11
79	70	6
24	10	6
52	19	5
19	79	8
78	41	4
3	49	9
2	61	2
2	35	2
58	73	4
54	11	3
12	79	3
68	34	4
9	12	4
93	30	9
68	86	4
48	22	2
43	92	7
11	39	4
44	2	3
70	7	5
31	43	12
41	68	3
17	88	2
32	94	5
78	27	2
80	96	4
10	58	4
41	78	4
92	44	2
80	16	5
26	12	2
89	18	8
61	79	2
95	21	6
12	60	2
78	79	5
5	99	7
63	5	3
95	79	2
51	1	7
16	36	4
90	23	6
86	23	5
58	57	8
28	89	3
31	59	4
81	71	6
22	17	3
2	72	5
1	47	5
27	32	4
13	3	4
76	31	7
8	29	2
73	93	7
30	41	4
94	28	8
91	3	8
68	18	4
62	11	2
3	45	7
42	13	4
5	4	4
72	22	7
97	18	6
44	7	2
23	12	7
32	10	7
14	65	6
69	33	6
60	83	8
14	2	5
18	1	7
40	17	3
4	85	4
5	100	3
93	62	2
89	87	3
64	58	3
72	40	6
17	73	6
51	10	5
1	87	5
19	63	7
71	98	3
63	48	13
27	68	3
94	3	7
8	93	8
58	16	4
96	25	5
98	5	8
11	63	8
74	93	3
41	4	2
97	41	5
81	63	2
26	35	4
54	72	4
51	6	2
81	88	4
25	29	2
49	37	2
96	16	7
73	12	4
41	71	5
37	47	3
47	49	5
42	84	5
17	65	7
41	17	9
79	94	8
6	74	6
75	85	11
84	72	5
27	93	6
42	4	11
43	90	7
91	36	3
48	44	3
81	55	2
55	81	4
80	74	3
58	44	5
43	27	3
51	83	2
1	51	2
76	14	2
58	94	8
74	40	2
34	14	3
67	52	3
29	88	2
88	58	4
46	89	8
64	5	3
77	97	3
80	78	5
86	33	3
91	17	8
97	48	7
49	25	4
100	98	4
12	2	4
94	6	3
19	92	3
24	87	3
3	88	5
85	58	6
65	20	4
28	7	4
37	73	3
66	42	2
75	3	4
9	56	9
13	62	2
83	100	2
29	32	4
7	81	14
42	30	2
82	60	2
66	5	2
9	96	3
49	52	2
72	76	2
75	95	3
55	32	5
4	80	12
11	50	2
80	68	4
21	58	4
51	67	6
69	80	3
62	61	5
35	75	2
79	58	6
20	57	7
32	84	5
67	64	8
45	3	7
11	72	2
87	54	5
89	15	5
6	48	2
96	79	2
82	73	2
95	20	8
72	83	4
30	90	9
83	22	3
34	82	4
76	80	5
31	21	4
67	20	4
94	85	4
39	98	4
41	42	7
12	49	3
81	61	3
80	52	2
22	84	3
16	31	4
42	63	4
29	13	3
95	50	3
94	16	8
56	95	5
40	28	3
51	38	7
56	72	7
25	81	6
95	13	4
17	87	2
57	15	7
78	5	2
61	81	2
63	35	7
46	53	11
89	60	8
34	73	5
73	80	3
75	39	5
10	55	2
39	100	5
25	5	6
87	64	6
69	19	2
70	46	3
40	42	6
65	60	4
43	8	5
12	90	3
92	97	3
86	14	7
75	21	4
32	58	9
1	50	4
87	60	9
87	39	5
81	21	6
85	57	5
58	56	3
29	65	7
50	12	9
97	20	3
4	91	4
86	29	4
83	77	5
87	56	4
77	42	4
38	58	8
54	30	5
14	79	4
23	29	5
54	69	2
20	40	7
9	44	7
22	46	9
13	39	3
66	23	6
93	61	2
27	94	7
92	37	3
91	12	5
71	8	6
92	99	3
22	40	5
21	67	8
12	92	9
72	71	4
99	91	9
59	85	4
61	3	2
45	25	2
80	71	6
56	11	4
77	24	9
3	64	5
85	83	4
52	28	2
38	49	7
17	28	4
83	5	9
80	58	7
30	98	2
31	22	3
23	3	5
38	42	4
56	39	9
74	25	4
36	37	2
28	5	2
99	83	5
42	48	6
8	20	5
15	81	3
43	30	2
11	21	3
58	69	5
6	89	3
87	63	5
37	62	3
23	65	4
76	23	5
81	52	10
60	34	2
57	87	3
80	56	2
9	10	6
51	28	6
91	72	3
98	80	2
71	99	6
16	44	3
56	17	3
49	12	4
18	55	4
20	46	5
4	74	2
68	13	3
39	38	15
63	44	4
13	69	4
45	82	2
82	100	3
14	21	3
69	66	3
18	60	7
99	75	5
30	43	9
95	19	6
81	18	5
8	9	4
88	21	3
98	85	2
78	73	3
44	66	6
7	54	6
80	32	7
40	74	3
15	73	5
55	69	5
74	1	7
1	39	18
73	89	3
96	49	3
40	82	2
85	81	3
70	48	7
78	24	5
47	95	6
27	91	3
41	88	4
27	92	5
50	69	2
66	97	3
99	29	3
56	50	4
48	73	6
44	46	3
41	33	5
3	65	4
69	97	3
58	45	4
19	36	4
34	76	2
57	44	3
39	55	10
52	5	4
36	34	6
93	74	4
4	8	2
67	26	8
27	53	4
19	99	4
46	86	6
65	42	15
8	94	4
66	50	2
20	70	2
12	53	8
18	51	4
22	8	4
5	98	5
94	10	3
42	19	4
26	8	8
51	93	3
95	65	2
49	40	3
64	42	4
10	28	2
51	24	4
10	68	3
26	78	7
42	52	3
46	99	3
28	19	5
16	100	4
3	7	5
3	46	5
87	34	5
93	58	4
94	77	7
91	18	4
75	100	6
54	55	4
63	23	5
20	7	8
18	32	4
19	76	5
38	83	5
25	24	9
79	10	6
37	42	3
24	99	7
82	99	3
74	59	2
36	9	2
70	77	5
76	38	2
74	10	10
38	34	6
16	84	4
77	16	6
10	12	5
28	100	2
100	76	4
11	15	3
99	32	2
83	66	3
92	1	3
89	27	3
45	29	10
65	28	4
93	85	5
38	25	4
66	14	6
82	6	4
64	59	4
92	46	4
56	81	4
76	69	4
6	49	5
29	90	6
44	92	9
84	74	5
21	33	5
75	26	9
85	95	2
95	44	4
89	22	3
33	16	5
17	5	6
90	63	5
2	30	4
66	84	2
33	17	2
84	9	3
76	100	3
11	51	4
61	21	5
16	68	2
21	54	5
71	74	5
18	64	4
25	80	5
70	14	2
91	53	3
33	82	5
57	2	3
76	43	7
63	38	4
17	21	2
21	29	2
71	17	3
29	62	3
10	84	8
59	68	2
33	31	6
97	59	3
84	5	5
17	84	2
8	16	5
17	27	2
48	71	2
67	49	2
94	67	4
41	32	3
42	33	3
76	15	3
80	76	7
63	73	11
73	37	2
79	1	2
3	69	3
38	61	6
9	36	3
45	44	5
74	52	6
39	13	2
1	86	2
83	19	3
39	35	8
45	16	7
11	13	4
64	12	6
21	60	7
79	20	2
57	33	2
54	17	4
67	96	5
71	20	8
30	10	5
28	65	6
76	41	5
95	57	6
43	39	4
79	33	2
59	89	3
90	77	4
14	26	2
42	16	7
57	98	4
62	64	2
79	76	6
20	8	6
63	98	7
64	85	3
16	12	4
15	100	3
52	34	3
52	72	2
33	69	2
39	46	5
97	74	4
1	33	8
7	87	5
72	44	4
23	2	3
33	53	3
31	20	2
16	67	3
12	19	3
85	2	4
75	32	4
57	19	3
67	48	5
28	93	2
32	63	6
46	28	5
19	66	4
54	41	4
13	66	14
45	37	4
10	88	3
87	4	2
71	63	2
79	60	6
45	65	5
13	94	3
94	90	3
60	51	3
79	21	4
54	34	2
10	22	5
67	74	7
24	54	8
50	100	7
69	61	4
98	71	6
94	79	7
86	79	6
94	46	8
99	25	3
4	39	2
20	99	6
80	67	3
16	53	6
55	96	5
64	80	5
23	90	6
39	3	3
55	8	3
35	81	9
31	6	2
12	50	6
73	92	3
52	21	5
23	6	8
49	42	3
27	54	3
19	60	2
36	70	5
23	75	3
70	83	4
60	98	3
6	36	3
22	11	3
79	19	5
53	65	9
80	53	5
31	29	2
32	31	11
17	48	2
85	54	2
88	4	2
54	93	5
6	34	5
78	39	6
81	100	5
68	100	8
34	77	8
74	97	4
72	99	2
55	30	2
56	46	4
29	31	4
70	91	2
54	46	2
57	77	2
45	56	2
82	24	2
40	26	4
12	47	2
47	54	3
16	47	2
11	96	7
59	1	2
45	60	2
75	45	3
29	42	5
7	67	3
98	41	3
4	7	3
53	18	3
8	31	3
16	38	4
37	48	8
24	69	9
5	60	4
88	12	3
26	94	3
3	25	4
62	21	2
68	54	2
48	6	2
81	75	2
19	20	2
51	53	3
70	94	2
39	29	4
90	40	2
81	20	4
31	96	7
21	18	5
79	41	2
38	81	2
19	4	2
80	4	2
89	67	6
19	39	9
30	75	2
51	79	3
77	32	2
22	62	3
13	53	3
36	35	2
33	88	7
56	63	7
92	60	3
92	27	2
34	75	5
81	16	2
45	7	2
81	44	6
26	68	5
90	86	3
21	96	5
43	4	3
10	99	2
70	93	2
90	10	5
32	50	5
91	32	2
44	73	3
85	13	2
28	70	6
46	93	3
12	98	2
98	94	3
58	10	5
88	76	3
85	76	3
69	93	5
84	50	2
22	39	3
16	72	4
50	16	6
74	17	2
5	95	4
99	19	4
12	10	4
41	22	3
80	11	2
72	80	5
32	40	6
83	13	2
28	83	3
13	21	5
16	75	4
46	45	5
27	35	2
16	59	2
71	83	4
78	11	2
44	25	6
38	73	6
77	99	3
94	9	2
44	98	6
89	13	3
63	11	4
74	42	3
76	74	2
46	97	3
83	21	2
22	72	6
60	85	5
27	97	7
21	97	5
70	31	4
5	20	3
4	42	3
82	13	12
17	29	3
25	43	2
20	50	5
53	28	5
7	38	4
72	19	3
96	5	2
58	75	5
11	27	4
29	54	3
70	23	3
47	41	4
29	47	2
91	90	5
35	47	3
72	62	3
16	82	2
45	66	4
6	28	2
30	14	3
31	77	5
49	30	14
11	22	6
26	70	8
33	65	3
65	6	6
77	11	5
69	11	3
82	26	3
100	11	4
25	12	3
14	37	5
26	58	7
43	75	4
43	18	4
57	25	6
12	30	3
62	29	6
9	65	5
12	45	6
33	12	3
1	12	5
54	58	5
17	16	3
6	19	2
73	97	5
77	13	3
38	5	2
2	38	2
18	61	2
25	6	5
80	77	2
6	9	2
44	77	4
2	100	4
95	59	4
6	77	4
58	74	3
75	64	6
31	10	2
68	47	2
99	92	2
31	4	3
73	46	7
44	32	3
71	62	3
78	98	2
26	37	2
44	42	2
37	7	3
28	44	7
61	10	2
79	25	2
36	25	4
16	95	2
83	70	2
21	12	2
11	77	2
20	29	2
50	48	3
11	32	6
92	26	5
18	81	7
63	70	6
3	27	3
61	74	2
69	81	2
14	33	3
77	19	2
92	55	4
21	55	2
64	89	2
30	26	5
61	73	2
26	65	5
87	3	6
84	10	4
64	2	2
35	21	3
25	21	5
91	54	13
81	1	2
71	81	4
91	14	2
19	65	5
2	46	3
76	91	10
80	97	2
17	60	3
69	75	2
93	81	6
91	80	7
100	28	4
46	42	2
64	53	4
1	90	5
93	88	15
39	58	2
48	4	2
15	76	3
34	2	5
73	87	7
51	100	2
29	94	3
14	99	2
1	40	5
43	17	2
97	24	3
50	75	2
8	15	5
24	46	3
94	11	6
38	9	5
83	60	3
28	50	4
45	2	2
56	100	9
72	55	6
48	62	3
66	98	2
72	52	7
71	48	4
58	71	5
5	11	2
49	11	2
96	100	4
99	15	2
92	93	2
62	93	3
34	68	4
29	74	2
14	24	2
10	14	6
14	9	3
73	14	3
24	84	4
73	67	2
3	95	2
83	4	3
47	73	2
98	72	4
24	62	2
51	23	2
10	35	2
89	53	4
31	62	4
35	54	2
33	50	2
90	31	7
73	8	2
56	73	4
66	15	2
90	46	5
56	10	3
23	79	2
67	44	3
57	86	3
62	53	2
74	27	5
20	30	3
75	38	2
53	61	4
63	43	8
77	95	5
57	31	3
65	88	3
45	89	5
68	63	2
9	46	2
43	54	2
10	31	5
16	80	4
45	78	2
23	81	4
23	89	5
54	79	4
42	78	2
30	9	3
64	45	11
93	96	2
74	100	3
41	87	2
33	85	4
73	18	2
74	87	3
28	39	2
83	16	2
67	1	2
65	62	4
10	72	4
2	37	2
82	68	3
55	74	2
72	45	3
80	72	3
80	66	2
4	99	2
10	50	4
100	85	4
11	40	2
13	34	2
34	11	3
4	14	2
7	44	2
36	61	2
32	20	2
74	71	2
20	77	4
91	67	3
27	85	2
75	33	2
33	45	2
23	10	7
1	74	5
77	5	8
12	3	2
35	73	2
73	35	4
35	95	3
65	56	2
98	42	2
13	75	3
37	59	2
13	37	2
99	89	2
88	46	3
78	33	2
\.


--
-- Data for Name: tuser; Type: TABLE DATA; Schema: public; Owner: olivier
--

COPY tuser (id, name, spent, quota, last_in, created, updated) FROM stdin;
1	olivier	1974422620	0	2012-08-02 11:30:41.308491	2012-08-02 04:43:53.357668	2012-08-02 11:30:41.308491
\.


--
-- Name: tconst_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tconst
    ADD CONSTRAINT tconst_pkey PRIMARY KEY (name);


--
-- Name: tmarket_id_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tmarket
    ADD CONSTRAINT tmarket_id_key UNIQUE (id);


--
-- Name: tmvt_id_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tmvt
    ADD CONSTRAINT tmvt_id_key UNIQUE (id);


--
-- Name: tmvtremoved_id_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tmvtremoved
    ADD CONSTRAINT tmvtremoved_id_key UNIQUE (id);


--
-- Name: torder_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY torder
    ADD CONSTRAINT torder_pkey PRIMARY KEY (id);


--
-- Name: torder_uuid_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY torder
    ADD CONSTRAINT torder_uuid_key UNIQUE (uuid);


--
-- Name: torderremoved_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY torderremoved
    ADD CONSTRAINT torderremoved_pkey PRIMARY KEY (uuid);


--
-- Name: towner_name_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY towner
    ADD CONSTRAINT towner_name_key UNIQUE (name);


--
-- Name: towner_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY towner
    ADD CONSTRAINT towner_pkey PRIMARY KEY (id);


--
-- Name: tquality_name_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tquality
    ADD CONSTRAINT tquality_name_key UNIQUE (name);


--
-- Name: tquote_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tquote
    ADD CONSTRAINT tquote_pkey PRIMARY KEY (id);


--
-- Name: treltried_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY treltried
    ADD CONSTRAINT treltried_pkey PRIMARY KEY (np, nr);


--
-- Name: tuser_id_key; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tuser
    ADD CONSTRAINT tuser_id_key UNIQUE (id);


--
-- Name: tuser_pkey; Type: CONSTRAINT; Schema: public; Owner: olivier; Tablespace: 
--

ALTER TABLE ONLY tuser
    ADD CONSTRAINT tuser_pkey PRIMARY KEY (name);


--
-- Name: tmvt_did_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX tmvt_did_idx ON tmvt USING btree (grp);


--
-- Name: tmvt_nat_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX tmvt_nat_idx ON tmvt USING btree (nat);


--
-- Name: tmvt_own_dst_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX tmvt_own_dst_idx ON tmvt USING btree (own_dst);


--
-- Name: tmvt_own_src_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX tmvt_own_src_idx ON tmvt USING btree (own_src);


--
-- Name: torder_np_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX torder_np_idx ON torder USING btree (np);


--
-- Name: torder_nr_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX torder_nr_idx ON torder USING btree (nr);


--
-- Name: towner_name_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX towner_name_idx ON towner USING btree (name);


--
-- Name: tquality_name_idx; Type: INDEX; Schema: public; Owner: olivier; Tablespace: 
--

CREATE INDEX tquality_name_idx ON tquality USING btree (name);


--
-- Name: trig_befa_towner; Type: TRIGGER; Schema: public; Owner: olivier
--

CREATE TRIGGER trig_befa_towner BEFORE INSERT OR UPDATE ON towner FOR EACH ROW EXECUTE PROCEDURE ftime_updated();


--
-- Name: trig_befa_tquality; Type: TRIGGER; Schema: public; Owner: olivier
--

CREATE TRIGGER trig_befa_tquality BEFORE INSERT OR UPDATE ON tquality FOR EACH ROW EXECUTE PROCEDURE ftime_updated();


--
-- Name: trig_befa_tuser; Type: TRIGGER; Schema: public; Owner: olivier
--

CREATE TRIGGER trig_befa_tuser BEFORE INSERT OR UPDATE ON tuser FOR EACH ROW EXECUTE PROCEDURE ftime_updated();


--
-- Name: ctmvt_grp; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvt
    ADD CONSTRAINT ctmvt_grp FOREIGN KEY (grp) REFERENCES tmvt(id);


--
-- Name: ctmvt_nat; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvt
    ADD CONSTRAINT ctmvt_nat FOREIGN KEY (nat) REFERENCES tquality(id);


--
-- Name: ctmvt_own_dst; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvt
    ADD CONSTRAINT ctmvt_own_dst FOREIGN KEY (own_dst) REFERENCES towner(id);


--
-- Name: ctmvt_own_src; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvt
    ADD CONSTRAINT ctmvt_own_src FOREIGN KEY (own_src) REFERENCES towner(id);


--
-- Name: ctorder_np; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY torder
    ADD CONSTRAINT ctorder_np FOREIGN KEY (np) REFERENCES tquality(id);


--
-- Name: ctorder_nr; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY torder
    ADD CONSTRAINT ctorder_nr FOREIGN KEY (nr) REFERENCES tquality(id);


--
-- Name: ctorder_own; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY torder
    ADD CONSTRAINT ctorder_own FOREIGN KEY (own) REFERENCES towner(id);


--
-- Name: ctquality_depository; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tquality
    ADD CONSTRAINT ctquality_depository FOREIGN KEY (depository) REFERENCES tuser(name);


--
-- Name: ctquality_idd; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tquality
    ADD CONSTRAINT ctquality_idd FOREIGN KEY (idd) REFERENCES tuser(id);


--
-- Name: tmvtremoved_nat_fkey; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvtremoved
    ADD CONSTRAINT tmvtremoved_nat_fkey FOREIGN KEY (nat) REFERENCES tquality(id);


--
-- Name: tmvtremoved_own_dst_fkey; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvtremoved
    ADD CONSTRAINT tmvtremoved_own_dst_fkey FOREIGN KEY (own_dst) REFERENCES towner(id);


--
-- Name: tmvtremoved_own_src_fkey; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY tmvtremoved
    ADD CONSTRAINT tmvtremoved_own_src_fkey FOREIGN KEY (own_src) REFERENCES towner(id);


--
-- Name: treltried_np_fkey; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY treltried
    ADD CONSTRAINT treltried_np_fkey FOREIGN KEY (np) REFERENCES tquality(id);


--
-- Name: treltried_nr_fkey; Type: FK CONSTRAINT; Schema: public; Owner: olivier
--

ALTER TABLE ONLY treltried
    ADD CONSTRAINT treltried_nr_fkey FOREIGN KEY (nr) REFERENCES tquality(id);


--
-- Name: public; Type: ACL; Schema: -; Owner: olivier
--

REVOKE ALL ON SCHEMA public FROM PUBLIC;
REVOKE ALL ON SCHEMA public FROM olivier;
GRANT ALL ON SCHEMA public TO olivier;
GRANT ALL ON SCHEMA public TO PUBLIC;


--
-- Name: fchangestatemarket(boolean); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fchangestatemarket(_execute boolean) FROM PUBLIC;
REVOKE ALL ON FUNCTION fchangestatemarket(_execute boolean) FROM olivier;
GRANT ALL ON FUNCTION fchangestatemarket(_execute boolean) TO olivier;
GRANT ALL ON FUNCTION fchangestatemarket(_execute boolean) TO PUBLIC;
GRANT ALL ON FUNCTION fchangestatemarket(_execute boolean) TO admin;


--
-- Name: fcreateuser(text); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fcreateuser(_name text) FROM PUBLIC;
REVOKE ALL ON FUNCTION fcreateuser(_name text) FROM olivier;
GRANT ALL ON FUNCTION fcreateuser(_name text) TO olivier;
GRANT ALL ON FUNCTION fcreateuser(_name text) TO PUBLIC;
GRANT ALL ON FUNCTION fcreateuser(_name text) TO admin;


--
-- Name: tquote; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tquote FROM PUBLIC;
REVOKE ALL ON TABLE tquote FROM olivier;
GRANT ALL ON TABLE tquote TO olivier;
GRANT SELECT ON TABLE tquote TO client_opened_role;
GRANT SELECT ON TABLE tquote TO client_stopping_role;
GRANT SELECT ON TABLE tquote TO admin;


--
-- Name: fexecquote(text, integer); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fexecquote(_owner text, _idquote integer) FROM PUBLIC;
REVOKE ALL ON FUNCTION fexecquote(_owner text, _idquote integer) FROM olivier;
GRANT ALL ON FUNCTION fexecquote(_owner text, _idquote integer) TO olivier;
GRANT ALL ON FUNCTION fexecquote(_owner text, _idquote integer) TO PUBLIC;
GRANT ALL ON FUNCTION fexecquote(_owner text, _idquote integer) TO client_opened_role;


--
-- Name: fgeterrs(boolean); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fgeterrs(_details boolean) FROM PUBLIC;
REVOKE ALL ON FUNCTION fgeterrs(_details boolean) FROM olivier;
GRANT ALL ON FUNCTION fgeterrs(_details boolean) TO olivier;
GRANT ALL ON FUNCTION fgeterrs(_details boolean) TO PUBLIC;
GRANT ALL ON FUNCTION fgeterrs(_details boolean) TO admin;


--
-- Name: fgetprequote(text, text, bigint, text); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) FROM PUBLIC;
REVOKE ALL ON FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) FROM olivier;
GRANT ALL ON FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) TO olivier;
GRANT ALL ON FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) TO PUBLIC;
GRANT ALL ON FUNCTION fgetprequote(_owner text, _qualityprovided text, _qttprovided bigint, _qualityrequired text) TO client_opened_role;


--
-- Name: fgetquote(text, text, bigint, bigint, text); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) FROM PUBLIC;
REVOKE ALL ON FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) FROM olivier;
GRANT ALL ON FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO olivier;
GRANT ALL ON FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO PUBLIC;
GRANT ALL ON FUNCTION fgetquote(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO client_opened_role;


--
-- Name: fgetstats(boolean); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fgetstats(_details boolean) FROM PUBLIC;
REVOKE ALL ON FUNCTION fgetstats(_details boolean) FROM olivier;
GRANT ALL ON FUNCTION fgetstats(_details boolean) TO olivier;
GRANT ALL ON FUNCTION fgetstats(_details boolean) TO PUBLIC;
GRANT ALL ON FUNCTION fgetstats(_details boolean) TO admin;


--
-- Name: fgetstats(timestamp without time zone, timestamp without time zone); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) FROM PUBLIC;
REVOKE ALL ON FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) FROM olivier;
GRANT ALL ON FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) TO olivier;
GRANT ALL ON FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) TO PUBLIC;
GRANT ALL ON FUNCTION fgetstats(mini timestamp without time zone, maxi timestamp without time zone) TO admin;


--
-- Name: torder; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE torder FROM PUBLIC;
REVOKE ALL ON TABLE torder FROM olivier;
GRANT ALL ON TABLE torder TO olivier;
GRANT SELECT ON TABLE torder TO client_opened_role;
GRANT SELECT ON TABLE torder TO client_stopping_role;
GRANT SELECT ON TABLE torder TO admin;


--
-- Name: finsertorder(text, text, bigint, bigint, text); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) FROM PUBLIC;
REVOKE ALL ON FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) FROM olivier;
GRANT ALL ON FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO olivier;
GRANT ALL ON FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO PUBLIC;
GRANT ALL ON FUNCTION finsertorder(_owner text, _qualityprovided text, _qttprovided bigint, _qttrequired bigint, _qualityrequired text) TO client_opened_role;


--
-- Name: fremoveagreement(integer); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fremoveagreement(_grp integer) FROM PUBLIC;
REVOKE ALL ON FUNCTION fremoveagreement(_grp integer) FROM olivier;
GRANT ALL ON FUNCTION fremoveagreement(_grp integer) TO olivier;
GRANT ALL ON FUNCTION fremoveagreement(_grp integer) TO PUBLIC;
GRANT ALL ON FUNCTION fremoveagreement(_grp integer) TO client_opened_role;
GRANT ALL ON FUNCTION fremoveagreement(_grp integer) TO client_stopping_role;


--
-- Name: towner; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE towner FROM PUBLIC;
REVOKE ALL ON TABLE towner FROM olivier;
GRANT ALL ON TABLE towner TO olivier;
GRANT SELECT ON TABLE towner TO client_opened_role;
GRANT SELECT ON TABLE towner TO client_stopping_role;
GRANT SELECT ON TABLE towner TO admin;


--
-- Name: tquality; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tquality FROM PUBLIC;
REVOKE ALL ON TABLE tquality FROM olivier;
GRANT ALL ON TABLE tquality TO olivier;
GRANT SELECT ON TABLE tquality TO client_opened_role;
GRANT SELECT ON TABLE tquality TO client_stopping_role;
GRANT SELECT ON TABLE tquality TO admin;


--
-- Name: vorder; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vorder FROM PUBLIC;
REVOKE ALL ON TABLE vorder FROM olivier;
GRANT ALL ON TABLE vorder TO olivier;
GRANT SELECT ON TABLE vorder TO client_opened_role;
GRANT SELECT ON TABLE vorder TO client_stopping_role;
GRANT SELECT ON TABLE vorder TO admin;


--
-- Name: fremoveorder(text); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fremoveorder(_uuid text) FROM PUBLIC;
REVOKE ALL ON FUNCTION fremoveorder(_uuid text) FROM olivier;
GRANT ALL ON FUNCTION fremoveorder(_uuid text) TO olivier;
GRANT ALL ON FUNCTION fremoveorder(_uuid text) TO PUBLIC;
GRANT ALL ON FUNCTION fremoveorder(_uuid text) TO client_opened_role;
GRANT ALL ON FUNCTION fremoveorder(_uuid text) TO client_stopping_role;


--
-- Name: fresetmarket(); Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON FUNCTION fresetmarket() FROM PUBLIC;
REVOKE ALL ON FUNCTION fresetmarket() FROM olivier;
GRANT ALL ON FUNCTION fresetmarket() TO olivier;
GRANT ALL ON FUNCTION fresetmarket() TO PUBLIC;
GRANT ALL ON FUNCTION fresetmarket() TO admin;


--
-- Name: tmvt; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tmvt FROM PUBLIC;
REVOKE ALL ON TABLE tmvt FROM olivier;
GRANT ALL ON TABLE tmvt TO olivier;
GRANT SELECT ON TABLE tmvt TO client_opened_role;
GRANT SELECT ON TABLE tmvt TO client_stopping_role;
GRANT SELECT ON TABLE tmvt TO admin;


--
-- Name: tconst; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tconst FROM PUBLIC;
REVOKE ALL ON TABLE tconst FROM olivier;
GRANT ALL ON TABLE tconst TO olivier;
GRANT SELECT ON TABLE tconst TO client_opened_role;
GRANT SELECT ON TABLE tconst TO client_stopping_role;
GRANT SELECT ON TABLE tconst TO admin;


--
-- Name: tmarket; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tmarket FROM PUBLIC;
REVOKE ALL ON TABLE tmarket FROM olivier;
GRANT ALL ON TABLE tmarket TO olivier;
GRANT SELECT ON TABLE tmarket TO client_opened_role;
GRANT SELECT ON TABLE tmarket TO client_stopping_role;
GRANT SELECT ON TABLE tmarket TO admin;


--
-- Name: tmvtremoved; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tmvtremoved FROM PUBLIC;
REVOKE ALL ON TABLE tmvtremoved FROM olivier;
GRANT ALL ON TABLE tmvtremoved TO olivier;
GRANT SELECT ON TABLE tmvtremoved TO client_opened_role;
GRANT SELECT ON TABLE tmvtremoved TO client_stopping_role;
GRANT SELECT ON TABLE tmvtremoved TO admin;


--
-- Name: torderremoved; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE torderremoved FROM PUBLIC;
REVOKE ALL ON TABLE torderremoved FROM olivier;
GRANT ALL ON TABLE torderremoved TO olivier;
GRANT SELECT ON TABLE torderremoved TO client_opened_role;
GRANT SELECT ON TABLE torderremoved TO client_stopping_role;
GRANT SELECT ON TABLE torderremoved TO admin;


--
-- Name: tquoteremoved; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tquoteremoved FROM PUBLIC;
REVOKE ALL ON TABLE tquoteremoved FROM olivier;
GRANT ALL ON TABLE tquoteremoved TO olivier;
GRANT SELECT ON TABLE tquoteremoved TO client_opened_role;
GRANT SELECT ON TABLE tquoteremoved TO client_stopping_role;
GRANT SELECT ON TABLE tquoteremoved TO admin;


--
-- Name: tuser; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE tuser FROM PUBLIC;
REVOKE ALL ON TABLE tuser FROM olivier;
GRANT ALL ON TABLE tuser TO olivier;
GRANT SELECT ON TABLE tuser TO client_opened_role;
GRANT SELECT ON TABLE tuser TO client_stopping_role;
GRANT SELECT ON TABLE tuser TO admin;


--
-- Name: vmarket; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vmarket FROM PUBLIC;
REVOKE ALL ON TABLE vmarket FROM olivier;
GRANT ALL ON TABLE vmarket TO olivier;
GRANT SELECT ON TABLE vmarket TO client_opened_role;
GRANT SELECT ON TABLE vmarket TO client_stopping_role;
GRANT SELECT ON TABLE vmarket TO admin;


--
-- Name: vmvt; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vmvt FROM PUBLIC;
REVOKE ALL ON TABLE vmvt FROM olivier;
GRANT ALL ON TABLE vmvt TO olivier;
GRANT SELECT ON TABLE vmvt TO client_opened_role;
GRANT SELECT ON TABLE vmvt TO client_stopping_role;
GRANT SELECT ON TABLE vmvt TO admin;


--
-- Name: vmvtverif; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vmvtverif FROM PUBLIC;
REVOKE ALL ON TABLE vmvtverif FROM olivier;
GRANT ALL ON TABLE vmvtverif TO olivier;
GRANT SELECT ON TABLE vmvtverif TO client_opened_role;
GRANT SELECT ON TABLE vmvtverif TO client_stopping_role;
GRANT SELECT ON TABLE vmvtverif TO admin;


--
-- Name: vorderremoved; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vorderremoved FROM PUBLIC;
REVOKE ALL ON TABLE vorderremoved FROM olivier;
GRANT ALL ON TABLE vorderremoved TO olivier;
GRANT SELECT ON TABLE vorderremoved TO client_opened_role;
GRANT SELECT ON TABLE vorderremoved TO client_stopping_role;
GRANT SELECT ON TABLE vorderremoved TO admin;


--
-- Name: vorderverif; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vorderverif FROM PUBLIC;
REVOKE ALL ON TABLE vorderverif FROM olivier;
GRANT ALL ON TABLE vorderverif TO olivier;
GRANT SELECT ON TABLE vorderverif TO client_opened_role;
GRANT SELECT ON TABLE vorderverif TO client_stopping_role;
GRANT SELECT ON TABLE vorderverif TO admin;


--
-- Name: vstat; Type: ACL; Schema: public; Owner: olivier
--

REVOKE ALL ON TABLE vstat FROM PUBLIC;
REVOKE ALL ON TABLE vstat FROM olivier;
GRANT ALL ON TABLE vstat TO olivier;
GRANT SELECT ON TABLE vstat TO client_opened_role;
GRANT SELECT ON TABLE vstat TO client_stopping_role;
GRANT SELECT ON TABLE vstat TO admin;


--
-- PostgreSQL database dump complete
--

